#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
from CDTK.Dynamics.IniConSampler import IniConSampler as ics
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#!/usr/bin/env python

import sys
import re
import numpy as np

from CDTK.Dynamics.IniConSampler import IniConSampler as ics
from CDTK.Dynamics.SimulationBox import SimulationBox as sbo
from CDTK.Dynamics.SurfaceHoppingFS import SurfaceHoppingFS as shfs

import CDTK.Dynamics.MINTF_DH as mintf # double hole interfaces

# construct temporal atomic charge evolution
from CDTK.Dynamics.ObservableCharge import ObservableCharge as obs

import QDTK.Tools.Conversion as conv

filename = sys.argv[1]
istate = int(sys.argv[2])
nstates = int(sys.argv[3])

# MOLCAS input parameters
symbols = [ 'N', 'C', 'C', 'O', 'O', 'H', 'H', 'H', 'H', 'H'] 

m = np.array([ 14.0,  14.0,  14.0,
               12.0,  12.0,  12.0,
               12.0,  12.0,  12.0,
               16.0,  16.0,  16.0,
               16.0,  16.0,  16.0,
                1.0,   1.0,   1.0,
                1.0,   1.0,   1.0,
                1.0,   1.0,   1.0,
                1.0,   1.0,   1.0,
                1.0,   1.0,   1.0  ], float)

mintf.symbols = symbols
# setup of ab initio calculations
# Molcas input for molecular energy
#                  analytical gradient
#                  non-adiabatic coupling
mintf.ab_initio_setup['nocc']          = 20 
mintf.ab_initio_setup['scf_charge']    = 0
mintf.ab_initio_setup['basis_set']     = '6-31g*'
mintf.ab_initio_setup['casscf_charge'] = 2
mintf.ab_initio_setup['casscf_spin']   = 1 # and 3
mintf.ab_initio_setup['nactel']        = 28 
mintf.ab_initio_setup['inactive']      = 5 
mintf.ab_initio_setup['ras2']          = 15
mintf.ab_initio_setup['ciroot']        = 120 
mintf.ab_initio_setup['molcas_home']   = MOLCAS_HOME 
mintf.ab_initio_setup['ras1']          = ''

# Initialization of Surface-Hopping module
def getFuncEn():
    def f(x, **opts):
        e = mintf.runMolcasDHEnergyGradient(x, returnStates=range(nstates), **opts) # compute energy and gradient. energy readout. prepare JOB00[x.1.2]
        return e
    return f
potential = getFuncEn()

def getFuncGr():
    def f(x,**opts):
        g = mintf.runMolcasDHGradient(x,**opts) # gradient readout
        return g
    return f
gradients = getFuncGr()

def getFuncVD():
    def f(x_T, x_TpDT, DT, phase_track, state_block, **opts):
        vd = mintf.runMolcasDHCplmat(x_T, x_TpDT, DT, phase_track, state_block, **opts) # compute coupling matrix element
        return vd
    return f
cplmatrix = getFuncVD()

def getFuncCharge():
    def f(current_state, **opts):
        ch = mintf.getCharge_onAtom(current_state) # compute positive charge on each atom
        return ch
    return f
charge = getFuncCharge()

def getFuncAOC():
    def f(x,state_current,**opts):
        g = mintf.runMolcasHFKoopmansAOCoeff(x,state_current,**opts)
        return g
    return f
atomic_orbital=getFuncAOC()

def getFuncTAS():
    def f(x, current_state, **opts):
        de = mintf.runMolcasTASpectra(x, current_state, **opts)
        return de
    return f
absspec = getFuncTAS()


# Load trajectory (initial conditions) from file
sbox = sbo().load(filename)
sbox.islogfile = True
sbox.atomSymbols = symbols
sbox.unit = 'Bohr'
sbox.sh_method = 'FS' # fewest switching method
sbox.isrestart = False 
sbox.phase_track = np.ones(nstates,float)
sbox._has_ed = False
sbox._ed = None
sbox.constraints = None
sbox._has_ob = False
sbox._ob = None

fs_cntrl1 = shfs(initial_state=istate,
                 nstates=nstates,
                 n_q=20,
                 external_E=potential,
                 external_G=gradients,
                 external_VD=cplmatrix,
                 gradient_method='analytical',
                 cplVD_method='wavefunction_overlap',
                 cplBlock_radius=5,
                 DoubleHole=True)

sbox.add_surface_hopping_cntrl(fs_cntrl1)

# Initialize the observable module
ob_cntrl = obs(external_CHARGE=charge)
sbox.add_observable_cntrl(ob_cntrl)

sbox.integrate(tfinal=600.0,dt=20.0)
sbox.save(fname='sb_final')
sbox.traj2VTF()
