#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
from CDTK.Dynamics.IniConSampler import IniConSampler as ics
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#!/usr/bin/env python

import sys
import re
import numpy as np

from CDTK.Dynamics.IniConSampler import IniConSampler as ics
from CDTK.Dynamics.SimulationBox import SimulationBox as sbo

from CDTK.Dynamics.SurfaceHoppingFS import SurfaceHoppingFS as shfs

import CDTK.Tools.Conversion as conv
import CDTK.Dynamics.MolcasInterfaceSH as mintf


filename = sys.argv[1]
istate = int(sys.argv[2])
nstates = int(sys.argv[3])

# MOLCAS input parameters
symbols = ['O','O','H','H','H','H','H']
m = np.array([16.0, 16.0, 1.0, 1.0, 1.0, 1.0, 1.0],float)

mintf.symbols = symbols
# Molcas input for molecular energy
mintf.input_energy = mintf.load_input('MOLCAS_ENERGY_INP.inp') 

# Molcas input for non-adiabatic coupling
mintf.input_cplmat = mintf.load_input('MOLCAS_CPLMAT_INP.inp') 
mintf.input_init_cplmat = mintf.load_input('MOLCAS_CPLMAT_INIT_INP.inp')

# Molcas input for analytical gradient
mintf.input_gradient = mintf.load_input('MOLCAS_GRADIENT_INP.inp')

# Initialization of Surface-Hopping module
def getFuncEn():
    def f(x):
        e = mintf.runMolcasEnergy(x, returnStates=range(nstates))
        return e
    return f
potential = getFuncEn()

def getFuncGr():
    def f(x,**opts):
        g = mintf.runMolcasGradient(x,**opts)
        return g
    return f
gradients = getFuncGr()

def getFuncVD():
    def f(x_T, x_TpDT, DT, phase_track, **opts):
        vd = mintf.runMolcasCplmat(x_T, x_TpDT, DT, phase_track, **opts)
        return vd
    return f
cplmatrix = getFuncVD()

# Load trajectory (initial conditions) from file
sbox = sbo().load(filename)
sbox.islogfile = True
sbox.atomSymbols = symbols
sbox.unit = 'Bohr'
sbox.sh_method = 'FS' # fewest switching method
sbox.isrestart = False 
sbox.phase_track = np.ones(nstates,float)
sbox.constraints = None

#mintf.MOLCAS_ENERGY_INP = '/dev/shm/'+sbox._ID+'_'+str(istate)+'_en.inp'
#mintf.MOLCAS_ENERGY_OUT = '/dev/shm/'+sbox._ID+'_'+str(istate)+'_en.out'

#mintf.MOLCAS_CPLMAT_INP = '/dev/shm/'+sbox._ID+'_'+str(istate)+'_cm.inp'
#mintf.MOLCAS_CPLMAT_OUT = '/dev/shm/'+sbox._ID+'_'+str(istate)+'_cm.out'
#mintf.MOLCAS_T_XYZ    = 'mol_T.xyz'    # wavefunction at T
#mintf.MOLCAS_TpDT_XYZ = 'mol_TpDT.xyz' # wavefunction at T+DT

fs_cntrl1 = shfs(initial_state=istate,
                nstates=nstates,
                n_q=10,
                external_E=potential,
                external_G=gradients,
                external_VD=cplmatrix,
                gradient_method='analytical',
                cplVD_method='wavefunction_overlap')

sbox.add_surface_hopping_cntrl(fs_cntrl1)

sbox.integrate(tfinal=5000.0,dt=5.0)
sbox.save(fname='sb_final')
sbox.traj2VTF()
