#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import sys
import re
import numpy as np
import os

from pyxmol import *
from random import random

AU2SEC=2.418884326500e-17
FS2AU=1e-15/AU2SEC
M2AU=0.5291772083e-10
AU2M=1./M2AU


class PopDyn(object):
    occ=[]
    state=0
    external_P=None
    external_getocc=None
    external_setocc=None
    external_write_chkpt=None
    external_read_chkpt=None
    simbox=None

    factor=8
    pulse_duration=10*FS2AU # 10 fs FWHM
    t_start=- 0.5*factor * pulse_duration
    t_end= 0.5*factor * pulse_duration
    N_time = 200
    width=1.0e-6 / M2AU # 1 micrometer
    nphotons=1e15
    fluence= nphotons/width**2 # fluence in atomic units


    def __init__(self,**opts):
        self.external_P=opts.get('external_P',None)
        self.external_getocc=opts.get('external_getocc',None)
        self.external_setocc=opts.get('external_setocc',None)
        self.external_write_chkpt=opts.get('external_write_chkpt',None)
        self.external_read_chkpt =opts.get('external_read_chkpt', None)

        self.occ=self.external_getocc()

    def write_chkpt(self,fname):
        self.external_write_chkpt()
        EXEC = 'cp chkpt.dat ' + fname 
        os.system(EXEC)

    def read_chkpt(self,fname):
        EXEC = 'cp ' + fname + ' chkpt.dat'
        os.system(EXEC)
        self.external_read_chkpt()

    def pulse_envelope(self,t):
        if t>self.t_end or t<self.t_start:
            return(0.)
        sigma=self.pulse_duration/2.*1/sqrt(2.*log(2.))
        return(exp(- 0.5*(t/sigma)**2) * 1/sqrt(2.*pi) * 1/sigma)


    def _propagatePop(self):
        md_dt=self.simbox.DT
        process=self.external_P()
        tot_decayrate=sum(p[3] for p in process if p[0]=="A" or p[0]=="F")        
        nsteps=max(int(10*tot_decayrate*md_dt),10)
        dt=md_dt/nsteps
        
        #p_env=self.pulse_envelope(t)

        for i in range(0,nsteps):
            p_env=0.
            r=random()      
            took_step=False
            for p in process:
                if(took_step==True):
                    continue
                if(p[0]=="A" or p[0]=="F"):
                    rate=p[3]
                if(p[0]=="P" ):
                    rate=p[3]*p_env
                prob=dt*rate
                if(r<prob ):
                    newocc=self.external_getocc()
                    rem=p[1]
                    add=p[2]
                    for r in rem:
                        if r==0: 
                            continue
                        newocc[r-1]=newocc[r-1]-1
                    if add!=0:    
                        newocc[add-1]=newocc[add-1]+1
                    took_step=True
                    self.external_setocc(newocc)
                else:
                    r=r-prob
            if took_step:
                process=self.external_P()

        self.occ=self.external_getocc()



        


