#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*
#*  Copyright (C) 2020 - 2022
#*  Ludger Inhester, DESY, <ludger.inhester@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import os
import shutil
import sys
import re
import numpy as np

import CDTK
from CDTK.Tools.Utils import getUniqueID
import CDTK.Tools.Utils as uti
import CDTK.Tools.EField as efi
import CDTK.Tools.Conversion as conv

import CDTK.Tools.NumericalDerivatives as nderiv

import pyxmol as XM

TINY = 1.e-8

# - Default definitions
#   !! alteration might lead to unittest failures !!

class xmolecule(object):
    """
    This defines a class to interface with the XMOLECULE program.
    """

    def __init__(self,**opts):
        """
        Initialize and set parameters.
        TODO: Say here something about the possible input parameters
        """

        self.engine      = 'XMolecule' # Used in the wrapper class
        self._ID         = opts.get('ID', getUniqueID(12))
        self._inp        = self._ID+'.inp'
        self._log        = self._ID+'.log'
        self._dat        = self._ID+'.dat'
        self._counter    = 0                          # Execution counter
        self.is_storelog = opts.get('is_storelog',False)
        self.is_storedat = opts.get('is_storedat',False)
        self.is_storeinp = opts.get('is_storeinp',False)
        self.keepdir     = opts.get('keepdir',None)
        # -------------------------------------------------
        self.atomicSymbols = opts.get('atomicSymbols',[])
        self.atomicNumbers = opts.get('atomicNumbers',[])
        self.atomicMasses = opts.get('atomicMasses',[])
        self.efields       = opts.get('efields',[])
        self.calcOverlaps = opts.get('calcOverlaps', False)
        # -------------------------------------------------
        self.cis = False
        self.part = False
        self.hole = False
        self.xmolecule = None
        # -------------------------------------------------
    def hole_efields(self, X, x0):
        """
        Energy and electric field for single-hole Koopmans' theorem calculations.
        
        Input:
        X      -- 3N array of atomic positions, in Bohr (au)
        x0     -- array of dimension 3 of electronic position, in Bohr (au)
        
        Output:
        e      -- Array of energies for all single-hole states in Hartree (au)
        efield -- Electric field at the electron position
        
        """
        self.xmolecule.set_geom(X.tolist())
        l = self.xmolecule.hole_efields(x0.tolist())
        nstates = len(l)
        ncoords = len(x0.tolist())

        occ = []
        e = np.zeros(nstates)
        efield = np.zeros((nstates,ncoords))
        upots = np.zeros(nstates)

        for i, li in enumerate(l):
            occi = li[0]
            ei = li[1]
            efieldi = li[2]
            upoti = li[3]

            occ.append(occi)

            e[i] = ei
            efield[i] = efieldi
            upots[i] = upoti

        return occ, e, efield, upots


    def _setup(self):
        """ 
        This functions initalizes the xmolecule objects after a new input file has been created
        """

        ## Create a interface instance and initialize (e.g. grids etc.)
        self.xmolecule = XM.xmol()
        self.xmolecule.init()
        
    def set_occ(self, state):
        """
        Set occupation number to given state.
        """

        self.xmolecule.set_occ(state)

    def get_occ(self):
        """
        Get list with occupation number for each orbital.
        """

        return self.xmolecule.get_occ()

    def set_geom(self, X):
        """
        Set geometry to X

        X   -- 3N array of atomic positions, in Bohr (au)
        """

        # pass on geometry in Bohr
        self.xmolecule.set_geom(X.tolist())

    def get_geom(self):
        """
        Get geometry to X
        returns  3N array of atomic positions, in Bohr (au)
        """
        return np.array(self.xmolecule.get_geom()).flatten()


    def init_input_options(self, iopts, types, pos, x_pos=None, x_charges=None, x_sigma=None):
        """ 
        Setting input options.

        Input:
        iopts     -- dictonary of options for xmolecule (see XMolecule documentation for reasonable values
        types     -- list of all atoms
        pos       -- list of all positions
        x_pos     -- list of positions of additional point charges
        x_charges -- list of charge values of additional point charges
        """
        for cisOpt in ['cis', 'CIS']:
            if cisOpt in iopts.keys():
                self.cis = conv.stringToBool(iopts[cisOpt][0])

        # Write an appropriate XMolecule input file from all set parameters
        ifile = open('input.in', 'w')
 
       # Coordinates
#        types = cartpos['atomlist']
#        pos = cartpos['coordinates']
        for i in range(len(types)):
            ifile.write(types[i] + '    ' + str.format("{0:+.10f}", pos[i*3]) + '   '
                        + str.format("{0:+.10f}", pos[i*3+1]) + '   ' + str.format("{0:+.10f}", pos[i*3+2]) + ' \n')
        # Additional point charges
        if x_pos is not None and x_sigma is None:
            for i in range(len(x_pos)//3):
                ifile.write('X' + '    ' + str.format("{0:+.10f}", x_pos[i*3]) + '   '
                            + str.format("{0:+.10f}", x_pos[i*3+1]) + '   ' + str.format("{0:+.10f}", x_pos[i*3+2])
                            + '    ' + str.format("{0:+.8f}", x_charges[i]) +  ' \n')
        elif x_pos is not None and x_sigma is not None:
            for i in range(len(x_pos)//3):
                ifile.write('XG' + '    ' + str.format("{0:+.10f}", x_pos[i*3]) + '   '
                            + str.format("{0:+.10f}", x_pos[i*3+1]) + '   ' + str.format("{0:+.10f}", x_pos[i*3+2])
                            + '    ' + str.format("{0:+.8f}", x_charges[i]) + '    ' + str.format("{0:+.8f}", x_sigma[i]) + ' \n')


        # Iterate over entries and set the value read from the input file
        # also write to the xmolecule input file
        for k in iopts.keys():
            setattr(self,k,iopts[k][0])
            ifile.write(k + '=' + iopts[k][0] + ' \n')
        ifile.close()

        self.nstates = -1
        for nstateOpt in ['nstates']:
            if nstateOpt.upper() in [k.upper() for k in iopts.keys()]:
                self.nstates = int(iopts[nstateOpt][0])

        for occPartOpt in ['gs_occ_part', 'valgs_occ_part', 'gs_occ_part', 'gsocc_part']:
            if occPartOpt.upper() in [k.upper() for k in iopts.keys()]:
                if conv.stringToBool(iopts[occPartOpt][0]):
                    self.part = True
                    self.hole = False

        for occOpt in ['gs_occ', 'valgs_occ', 'gs_occ', 'gsocc']:
            if occOpt.upper() in [k.upper() for k in iopts.keys()]:
                if conv.stringToBool(iopts[occOpt][0]):
                    self.part = False
                    self.hole = True

        sys.stdout.flush()
        # do setup for xmolecule
        self._setup()

    def energy_HF(self, X, Time=0.0, **opts):
        """
        Hartree-Fock energy calculation. 

        Input:
        X   -- 3N array of atomic positions, in Bohr (au)

        Output:
        e   -- SCF energy in atomic units
        """

        # pass on geometry in Bohr
        self.xmolecule.set_geom(X.tolist())
        e = self.xmolecule.energy

        return e

    def energy_gradient_HF(self, X, Time=0.0, **opts):
        """
        Hartree-Fock energy and gradient calculation.

        Input:
        X   -- 3N array of atomic positions, in Bohr (au)
        
        Output:
        e   -- SCF energy in atomic units
        g   -- 3N SCF gradient in atomic units
        """

        # pass on geometry in Bohr
        self.xmolecule.set_geom(X.tolist())
        e = self.xmolecule.energy
        g = np.array(self.xmolecule.gradient())

        return e, g

    def part_properties(self, X, Time=0.0, **opts):
        """
        Energy, gradient, non-adiabatic coupling for single-particle Koopmans' theorem calculations.
        The number of single particle states is equal to the number of valence orbitals.

        Input:
        X   -- 3N array of atomic positions, in Bohr (au)
        Output:
        e   -- Array of energies for all single-hole states in atomic units (hartree)
        g   -- Array of 3N gradient for all single-hole states in atomic units
        nac -- 2D array of 3N non-adiabatic coupling between all single-hole states in atomic units
        """
        # pass on geometry in Bohr
        self.xmolecule.set_geom(X)
        occ, e, g, nac = self.xmolecule.part_properties()
        if self.nstates < 0:
            self.nstates = len(e)
        if self.nstates > len(e):
            print(f"Warning: The number of states requested is {self.nstates} but only {len(e)} states are returned. Check your input!")
            self.nstates = len(e)
        # truncate to the number of states specified by user input
        occR = occ[:self.nstates, :occ.shape[1]].copy()
        eR = e[:self.nstates].copy()
        gR = g[:self.nstates, :].copy()
        nacR = nac[:self.nstates, :self.nstates, :].copy()
        del occ, e, g, nac
        #return occ[:self.nstates], e[:self.nstates], g[:self.nstates], nac[:self.nstates, :self.nstates]
        #return occR, e[:self.nstates], gR, nacR
        return occR, eR, gR, nacR

    def hole_properties(self, X, Time=0.0, **opts):
        """
        Energy, gradient, non-adiabatic coupling for single-hole Koopmans' theorem calculations.
        The number of single hole states is equal to the number of valence orbitals.

        Input:
        X   -- 3N array of atomic positions, in Bohr (au)
        
        Output:
        e   -- Array of energies for all single-hole states in atomic units (hartree)
        g   -- Array of 3N gradient for all single-hole states in atomic units
        nac -- 2D array of 3N non-adiabatic coupling between all single-hole states in atomic units
        """
        # pass on geometry in Bohr
        self.xmolecule.set_geom(X.tolist())
        occ, e, g, nac = self.xmolecule.hole_properties()
        if self.nstates < 0:
            self.nstates = len(e)
        if self.nstates > len(e):
            print(f"Warning: The number of states requested is {self.nstates} but only {len(e)} states are returned. Check your input!")
            self.nstates = len(e)
        # truncate to the number of states specified by user input
        return occ[:self.nstates].copy(), e[:self.nstates].copy(), g[:self.nstates].copy(), nac[:self.nstates, :self.nstates].copy()

    def CI_energy(self, X, Time=0.0,state=None, **opts):        
        self.xmolecule.set_geom(X.tolist())
        en=np.array(self.xmolecule.CI_energies())
        if self.nstates < 0:
            self.nstates = len(en)
        if self.nstates > len(en):
            print(f"Warning: The number of states requested is {self.nstates} but only {len(en)} states are returned. Check your input!")
            self.nstates = len(en)
        if(state != None):
            return(en)
        else:
            return(en[state])

    def CI_gradient(self, X, Time=0.0,state=None, **opts):        
        if state==None:
            state=0
        # only along z-direction
        g=nderiv.gradient(lambda x: self.CI_energy(x,state=state), X ,ms=np.array([[0,0,1,0,0,-1]]),step=0.005,fast=True)
        G=np.zeros(len(X))
        G[2]=0.5*g[0]
        G[5]=-0.5*g[0]
        return(G)                

    def CIS_tdens(self, X, i, **opts): 
        """
        returns a list of AO trans density of CIS states at position X
        """
        self.xmolecule.set_geom(X.tolist())
        energies = self.xmolecule.CIS_energies()
        tdens = []
        for i in range(self.nstates):
            tdens.append(self.xmolecule.CIS_transdens(i))
        return tdens

    def CIS_properties(self, X, Time=0.0, state=0, overlap=False, **opts):        
        """
        Energy, gradient, non-adiabatic coupling for CIS states.

        Input:
        X   -- 3N array of atomic positions, in Bohr (au)
        state   -- state index
        
        Output:
        e   -- Array of energies for all single-hole states in atomic units (hartree)
        g   -- Array of 3N gradient for all single-hole states in atomic units
        nac -- 2D array of 3N non-adiabatic coupling between all single-hole states in atomic units
        """
        # pass on geometry in Bohr
        self.xmolecule.set_geom(X.tolist(), overlap=overlap)
        ncoords = len(X.tolist())
        en,gr,nacs = self.xmolecule.CIS_energy_gradient_nacs(state)
        if self.nstates < 0:
            self.nstates = len(en)
        if self.nstates > len(en):
            print(f"Warning: The number of states requested is {self.nstates} but only {len(en)} states are returned. Check your input!")
            self.nstates = len(en)
        nstates = self.nstates
        if overlap:
            overlapMatrix_cis = np.array(self.xmolecule.CI_olap())
        #state_labels = []
        #nac = {}
        #e = en[0:nstates].copy()
        #g = gr[0:nstates,:].copy()

        state_labels = list(range(nstates))
        #for i in range(0,nstates):
        #    state_labels.append(i)
        #    for j in range(0,nstates):
        #        nac[(i,j)] = nacs[i,j,:].copy()
        #del nacs
        #del en
        #del gr

        if not overlap:
            return state_labels, en, gr, nacs
        else:
            return state_labels, en, gr, nacs, overlapMatrix_cis

    def getGamma(self):
        """
        Obtain rates and cross sections at current position.
        """
        gamma = self.xmolecule.process()
        if sys.version_info[0] >= 3: # in python3 we need to decode the byte-string as utf8
            gamma = [(p[0].decode('utf8'), p[1], p[2], p[3]) for p in gamma]
        return gamma


    def getPartialCharges(self, chargetype='mulliken', state=0):
        """
        Returns current partial charges of the molecule accoring to given method.
        """
        if chargetype == 'mulliken':
            if not self.cis:
                return self.xmolecule.mullikencharge()
            else:
                return self.xmolecule.CIS_mullikenpop(state=state, typ='F')
        elif chargetype == 'loewdin':
            if not self.cis:
                return self.xmolecule.loewdincharge()
            else:
                return self.xmolecule.CIS_loewdinpop(state=state, typ='F')
        else:
            print("Charge type not implemented in XMolecule Interface")
            exit(-1)

    def getPopulation(self):
        """
        Returns current Population Analysis data.
        """        
        return self.xmolecule.population()

    def getMOCoefficients(self):
        """
        Returns current molecular orbital coefficients from LCAO
        """
        return self.xmolecule.mo_coef()

    def getOrbitalEnergies(self):
        """
        Returns current Orbital Energies.
        """        
        return self.xmolecule.orbital_energies()

    def getBondOrders(self,botype='Meyer'):
        """
        Returns current BondOrder Matrix of the molecule accoring to given method.
        """        
        if botype == 'Meyer':
            return self.xmolecule.bondorder()
        else:
            print("Charge type not implemented in XMolecule Interface")
            exit(-1)

    def getAbsorptionCrossSections(self):
        """
        Returns absorption cross sections between current molecular orbitals (in fact 2pi**2 * alpha * the oscillator strength)

        aCS[i,j] = absorption cross section from MO i to j

        Spin factor included.
        Note that the unit of the returned value is not exactly cross section as atomic unit, but instead
        crosssection times energy, since the cross section value is only clearly-defined with energy conserving delta function (see formula).
        The returned values are without delta function.
        and thus have to be multiplied with some finite spectral energy
        distribution of the pulse to obtain the cross section for the transition.

        The corresponding energy for the absorption can be obtained from the MO energy difference.
        """
        # the problem with the wrong name "absorptioncs" has been resolved in xmolecule
        # this is a temporary workaroung to keep it back-compatible
        if hasattr(self.xmolecule, "absorptioncs"):
            return self.xmolecule.absorptioncs()
        else:
            os,tdip = self.xmolecule.os()
            os = os * (2.*np.pi**2) * conv.FPC_ALPHA
            return os,tdip

    def getCISEnergies(self):
        """
        Returns CIS energies

        Returns:
        --------
        E[:] list of energies in Hartree:
        E[0] is the HF ground state energy
        E[1] is the first excited state energy
        E[n] is the nth excited state energy
        """
        en = self.xmolecule.CIS_energies()
        if self.nstates < 0:
            self.nstates = len(en)
        if self.nstates > len(en):
            print(f"Warning: The number of states requested is {self.nstates} but only {len(en)} states are returned. Check your input!")
            self.nstates = len(en)
        return en

    def getCISAbsorptionCrossSections(self, state):
        """
        Returns absorption cross sections for a given state with CIS index state
        CIS calculations have to be initialized before by running
        either self.getCISEnergies()  or self.CIS_properties()

        Input:
        -------
        state: int index of initial state: 0 means ground state, 1 means 1st excited state, and so on

        Returns:
        -------
        aCS[j] = absorption cross section to final state j, j=0 (ground state) ... number of cis states
        Note that the unit of the returned value is not exactly cross section as atomic unit, but instead
        crosssection times energy, since the cross section value is only clearly-defined with energy conserving delta function (see formula).
        The returned values are without delta function.
        and thus have to be multiplied with some finite spectral energy
        distribution of the pulse to obtain the cross section for the transition.
        The corresponding energy for the absorption can be obtained from getCISEnergies.
        """
        if hasattr(self.xmolecule, "CIS_absorptioncs"):
            return self.xmolecule.CIS_absorptioncs(state)[0]
        else:
            return self.xmolecule.CIS_os(state)[0] * (2.*np.pi**2) * conv.FPC_ALPHA

    def getFluorescenceRates(self):
        """
        Returns fluorescence rates between current molecular orbitals

        fluoRate[i,j] = fluorescence rate from MO i to j

        Spin orbital factor included.

        The corresponding energy for the fluorescence photon can be obtained from the MO energy difference.
        """

        return self.xmolecule.fluorescencerates()

    def write_mo_cube(self, mo):
        """ Write cube file for molecular orbital `mo`. This is a file named 'mo.cube'
        that can be loaded using VMD to visualise the orbital.

        Parameters
        ----------
        mo : int
            Index of molecular orbital to be used. Note: counting of MO starts from 1.
        """
        self.xmolecule.write_mo_cube(mo)
        return

    def molden(self, filename):
        """ writes molden file

        Args:
            filename (string): file name
        """
        self.xmolecule.write_molden(file=filename)

    def spinAnalysis(self):
       """
       returns spinAnalysis
       """
       if self.cis:
           raise ValueError("Spin density analysis not implemented for CIS")
       return self.xmolecule.spin_analysis()
