#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import unittest
import numpy as np
import os
import CDTK.Interfaces.MolcasInterface as mol
import CDTK.Tools.EField as efi
import CDTK.Tools.Conversion as cv

class MolcasTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_init_input_options(self):
        """
        Tests the correct setting of input options. Only some options are tested.
        """
        
        m = mol.molcas()

        opts = {'basis': ['Everything'], 'ciroot': ['5']}
        m.init_input_options(opts)

        self.assertEqual(m.engine, 'molcas')
        self.assertEqual(m._counter, 0)
        self.assertEqual(m.basis, opts['basis'][0])
        self.assertEqual(m.ciroot, opts['ciroot'][0])


    def test__store_log(self):
        """
        Tests storeing of log file.
        """

        m = mol.molcas()
        m.is_storelog = True

        datfile = open(m._log, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        m._store_log()

        cstr = '{0:0>7d}'.format(m._counter)
        logfile = open(cstr + '_' + m._log, 'r')
        self.assertEqual(logfile.read(), text)
      
    def test_rasci(self):
        """
        Tests the correct return of a rascii energy and gradient.
        """

        m = mol.molcas()
        m.nstates = 4
        m.atomicSymbols = ['O','H','H']
        m.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.7572157*cv.an2au,0.5865358*cv.an2au,
                      0.0,-0.7572157*cv.an2au,0.5865358*cv.an2au])
        ener, grd = m.rasci(x)
        
        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [0., -0., 0.12839, 0., -0.084031, -0.064195, 0., 0.084031, -0.064195]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

        ener, grd = m.rasci(x, root_g=2)    

        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)
        
        grd_compare = [0., 0., 0.050581, 0., -0.125923, -0.025291, 0., 0.125923, -0.025291]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)


    def test_rasci_D(self):
        """
        Tests the correct return of a rascii energy and gradient and coupling matrix.
        """

        m = mol.molcas()
        m.nstates = 4
        m.atomicSymbols = ['O','H','H']
        m.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.7572157*cv.an2au,0.5865358*cv.an2au,
                      0.0,-0.7572157*cv.an2au,0.5865358*cv.an2au])
        v= np.array([0.55643782, -0.32706529, -0.44684414, 0.79078154, 0.44472021, -0.24223186, -0.83286255, 0.08177673, 0.96707229])
        ener, grd,D = m.rasci_D(x, v)
        
        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [0., -0., 0.12839, 0., -0.084031, -0.064195, 0., 0.084031, -0.064195]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

        d_compare = [0., -0.00594596, -0.01479442, -0.00448785, 0.00594596, 0., 0.01447596, 0.01278785, 0.01479442, -0.01447596, 0., 0.00952463, 0.00448785, -0.01278785, -0.00952463,  0.]
        for i, dd in enumerate(D.flatten()):
            self.assertAlmostEqual(dd, d_compare[i], delta=1.e-7)

        v= np.array([0.70613368, 0.61489003, -0.97300109, 0.46928584, 0.81269216, -0.12007034, -0.67366431, 0.92167125, 0.75450382])
        ener,grd, D = m.rasci_D(x, v, root_g=2)
         
        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [0., 0., 0.050581, 0., -0.125923, -0.025291, 0., 0.125923, -0.025291]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

        d_compare = [0., -0.01009658, -0.00987102, -0.0055196, 0.01009658, 0., 0.00592824, 0.0058276, 0.00987102, -0.00592824, 0., 0.00393821, 0.0055196, -0.0058276, -0.00393821, 0.]
        for i, dd in enumerate(D.flatten()):
            self.assertAlmostEqual(dd, d_compare[i], delta=1.e-7)


    def test_rasci_D_num(self):
        """
        Tests the correct return of a rascii energy and gradient and coupling matrix employing numerical gradients.
        """

        m = mol.molcas()
        m.nstates = 4
        m.atomicSymbols = ['O','H','H']
        m.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.7572157*cv.an2au,0.5865358*cv.an2au,
                      0.0,-0.7572157*cv.an2au,0.5865358*cv.an2au])
        v= np.array([0.55643782, -0.32706529, -0.44684414, 0.79078154, 0.44472021, -0.24223186, -0.83286255, 0.08177673, 0.96707229])

        ener, grd,D = m.rasci_D_num(x, v)

        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [-0.000202, 0.005422, 0.047953, -0.000142, -0.049595, -0.02082, -0.000142, 0.055312, -0.02082]
        grda_compare = [0., 0., 0.12839, 0., -0.084031, -0.064195, 0., 0.084031, -0.064195]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

            if grda_compare[i] == 0.0:
                self.assertAlmostEqual(g, 0.0, delta=1.e-2) 
            else:
                ## TODO: this tests does not make any sense as it has a much too large delta, but the
                ## numerical gradient calculation is just too bad -> improve the calculation
                self.assertAlmostEqual(abs((g-grda_compare[i])/grda_compare[i]), 0.0, delta=1.e+2)

        d_compare = [0., -0.00594596, -0.01479442, -0.00448785, 0.00594596, 0., 0.01447596, 0.01278785, 0.01479442, -0.01447596, 0., 0.00952463, 0.00448785, -0.01278785, -0.00952463,  0.]
        for i, dd in enumerate(D.flatten()):
            self.assertAlmostEqual(dd, d_compare[i], delta=1.e-2)

        v= np.array([0.70613368, 0.61489003, -0.97300109, 0.46928584, 0.81269216, -0.12007034, -0.67366431, 0.92167125, 0.75450382])
        ener,grd, D = m.rasci_D_num(x, v, root_g=2)
        print(ener, grd, D)
        e_compare = [-74.57179258, -74.51000574, -74.34548597, -73.69488174]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [9.09999997e-05, 4.99900000e-03, -1.76230000e-02, -1.45000000e-04, -9.30850000e-02, 1.13280000e-02, -1.45000000e-04, 9.85220000e-02, 1.13280000e-02]
        grda_compare = [0., 0., 0.050581, 0., -0.125923, -0.025291, 0., 0.125923, -0.025291]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

            if grda_compare[i] == 0.0:
                self.assertAlmostEqual(g, 0.0, delta=5.e-3) 
            else:
                ## TODO: this tests does not make any sense as it has a much too large delta, but the
                ## numerical gradient calculation is just too bad -> improve the calculation
                self.assertAlmostEqual(abs((g-grda_compare[i])/grda_compare[i]), 0, delta=1.e+2)

        d_compare = [0., -0.01009658, -0.00987102, -0.0055196, 0.01009658, 0., 0.00592824, 0.0058276, 0.00987102, -0.00592824, 0., 0.00393821, 0.0055196, -0.0058276, -0.00393821, 0.]
        for i, dd in enumerate(D.flatten()):
            self.assertAlmostEqual(dd, d_compare[i], delta=1.e-2)

    def test__store_err(self):
        """
        Tests storeing of dat file.
        """

        m = mol.molcas()
        m.is_storeerr = True
        m._err = "blah.err"

        errfile = open(m._err, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        errfile.write(text)
        errfile.close()

        m._store_err()

        cstr = '{0:0>7d}'.format(m._counter)
        errfile = open(cstr + '_' + m._err, 'r')
        self.assertEqual(errfile.read(), text)

    def test__store_inp(self):
        """
        Tests storeing of inp file.
        """
        
        m = mol.molcas()
        m.is_storeinp = True

        datfile = open(m._inp, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        m._store_inp()

        cstr = '{0:0>7d}'.format(m._counter)
        inpfile = open(cstr + '_' + m._inp, 'r')
        self.assertEqual(inpfile.read(), text)


    def test__parse_rasci_energy(self):
        """
        Tests the correct parsing of RASSCF energies.
        """

        e_compare = np.random.rand(5)
        m = mol.molcas()
        m.nstates = 5
        datfile = open(m._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(":: RASSCF root number 1 Total energy: {0:.10f}\n".format(e_compare[0]))
        datfile.write(":: RASSCF root number 2 Total energy: {0:.10f}\n".format(e_compare[1]))
        datfile.write(":: RASSCF root number 3 Total energy: {0:.10f}\n".format(e_compare[2]))
        datfile.write(":: RASSCF root number 4 Total energy: {0:.10f}\n".format(e_compare[3]))
        datfile.write(":: RASSCF root number 5 Total energy: {0:.10f}\n".format(e_compare[4]))

        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        e = m._parse_rasci_energy()
        for i, ee in enumerate(e):
            self.assertAlmostEqual(ee, e_compare[i], delta=1.e-7)

    def test__parse_rasci_gradient(self):
        """
        Tests the correct parsing of HF gradients.
        """

        g_compare = np.random.rand(9)
        m = mol.molcas()
        datfile = open(m._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write("Molecular gradients\n")
        datfile.write("DUM LINE 1\n" + "DUM LINE 2\n" + "DUM LINE 3\n")
        datfile.write("DUM LINE 4\n" + "DUM LINE 5\n" + "DUM LINE 6\n" + "DUM LINE 7\n")
        datfile.write(" A1  {0:.10f}  ".format(g_compare[0]) + " {0:.10f}  ".format(g_compare[1]) + " {0:.10f}\n".format(g_compare[2]))
        datfile.write(" B2  {0:.10f}  ".format(g_compare[3]) + " {0:.10f}  ".format(g_compare[4]) + " {0:.10f}\n".format(g_compare[5]))
        datfile.write(" Fi3  {0:.10f}  ".format(g_compare[6]) + " {0:.10f}  ".format(g_compare[7]) + " {0:.10f}\n".format(g_compare[8]))
        datfile.write(" -")
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")

        datfile.close()

        grd = m._parse_rasci_gradient()
        for i, gg in enumerate(grd):
            self.assertAlmostEqual(gg, g_compare[i], delta=1.e-7)

    def test__parse_rasci_D(self):
        """
        Tests the correct parsing of coupling matrix
        """

        ovl_in = np.random.rand(8,8)
        for i in range(8):
            for j in range(i+1, 8):
                ovl_in[i, j] = ovl_in[j, i]
        ovl_compare = ovl_in[0:4, 4:8]
        d_compare = (ovl_compare - np.transpose(ovl_compare)) / 2

        m = mol.molcas()

        datfile = open(m._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write("      OVERLAP MATRIX FOR THE ORIGINAL STATES:\n")
        datfile.write("DUM LINE 1\n")
        datfile.write(" {0:.10f} \n ".format(ovl_in[0, 0]))
        datfile.write(" {0:.10f}  ".format(ovl_in[1, 0]) + " {0:.10f} \n ".format(ovl_in[1, 1]))
        datfile.write(" {0:.10f}  ".format(ovl_in[2, 0]) + " {0:.10f}  ".format(ovl_in[2, 1]) + " {0:.10f} \n ".format(ovl_in[2, 2]))
        datfile.write(" {0:.10f}  ".format(ovl_in[3, 0]) + " {0:.10f}  ".format(ovl_in[3, 1]) + " {0:.10f}  ".format(ovl_in[3, 2]) + " {0:.10f} \n ".format(ovl_in[3, 3]))
        datfile.write(" {0:.10f}  ".format(ovl_in[4, 0]) + " {0:.10f}  ".format(ovl_in[4, 1]) + " {0:.10f}  ".format(ovl_in[4, 2]) + " {0:.10f}  ".format(ovl_in[4, 3]) + " {0:.10f} \n ".format(ovl_in[4, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_in[5, 0]) + " {0:.10f}  ".format(ovl_in[5, 1]) + " {0:.10f}  ".format(ovl_in[5, 2]) + " {0:.10f}  ".format(ovl_in[5, 3]) + " {0:.10f} \n ".format(ovl_in[5, 4]))
        datfile.write(" {0:.10f} \n ".format(ovl_in[5, 5]))
        datfile.write(" {0:.10f}  ".format(ovl_in[6, 0]) + " {0:.10f}  ".format(ovl_in[6, 1]) + " {0:.10f}  ".format(ovl_in[6, 2]) + " {0:.10f}  ".format(ovl_in[6, 3]) + " {0:.10f} \n ".format(ovl_in[6, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_in[6, 5]) + " {0:.10f} \n ".format(ovl_in[6, 6]))
        datfile.write(" {0:.10f}  ".format(ovl_in[7, 0]) + " {0:.10f}  ".format(ovl_in[7, 1]) + " {0:.10f}  ".format(ovl_in[7, 2]) + " {0:.10f}  ".format(ovl_in[7, 3]) + " {0:.10f} \n ".format(ovl_in[7, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_in[7, 5]) + " {0:.10f}  ".format(ovl_in[7, 6]) + " {0:.10f} \n".format(ovl_in[7, 7]))
        datfile.write("------\n")
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")

        datfile.close()

        d = m._parse_rasci_D(1.0)
        for i, dd in enumerate(d.flatten()):
            self.assertAlmostEqual(dd, d_compare.flatten()[i], delta=1.e-7)

    def test__trackPhases(self):
        """
        Tests the correct adjusting of the sign of the D matrix.
        """

        d_old = 2.0 * np.random.rand(4,4) - 1.0
        
        m = mol.molcas()
        d = m._trackPhases(d_old)

        for i, dd in enumerate(d.flatten()):
            self.assertAlmostEqual(dd, d_old.flatten()[i], delta=1.e-7)

        m._oldD = d_old
        d = d_old.copy() 
        d *= -1.0

        for i, dd in enumerate(d.flatten()):
            self.assertAlmostEqual(dd, -d_old.flatten()[i], delta=1.e-7)

        d_new = m._trackPhases(d)

        for i, dd in enumerate(d_new.flatten()):
            self.assertAlmostEqual(dd, d_old.flatten()[i], delta=1.e-7)

        
    def test_readOverlapMatrix(self):
        """
        Tests the correct parsing of the overlapt matrix.
        """

        ovl_compare = np.random.rand(8,8)
        for i in range(8):
            for j in range(i+1, 8):
                ovl_compare[i, j] = ovl_compare[j, i]

        datfile = open('muh.log', 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write("      OVERLAP MATRIX FOR THE ORIGINAL STATES:\n")
        datfile.write("DUM LINE 1\n")
        datfile.write(" {0:.10f} \n ".format(ovl_compare[0, 0]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[1, 0]) + " {0:.10f} \n ".format(ovl_compare[1, 1]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[2, 0]) + " {0:.10f}  ".format(ovl_compare[2, 1]) + " {0:.10f} \n ".format(ovl_compare[2, 2]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[3, 0]) + " {0:.10f}  ".format(ovl_compare[3, 1]) + " {0:.10f}  ".format(ovl_compare[3, 2]) + " {0:.10f} \n ".format(ovl_compare[3, 3]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[4, 0]) + " {0:.10f}  ".format(ovl_compare[4, 1]) + " {0:.10f}  ".format(ovl_compare[4, 2]) + " {0:.10f}  ".format(ovl_compare[4, 3]) + " {0:.10f} \n ".format(ovl_compare[4, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[5, 0]) + " {0:.10f}  ".format(ovl_compare[5, 1]) + " {0:.10f}  ".format(ovl_compare[5, 2]) + " {0:.10f}  ".format(ovl_compare[5, 3]) + " {0:.10f} \n ".format(ovl_compare[5, 4]))
        datfile.write(" {0:.10f} \n ".format(ovl_compare[5, 5]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[6, 0]) + " {0:.10f}  ".format(ovl_compare[6, 1]) + " {0:.10f}  ".format(ovl_compare[6, 2]) + " {0:.10f}  ".format(ovl_compare[6, 3]) + " {0:.10f} \n ".format(ovl_compare[6, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[6, 5]) + " {0:.10f} \n ".format(ovl_compare[6, 6]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[7, 0]) + " {0:.10f}  ".format(ovl_compare[7, 1]) + " {0:.10f}  ".format(ovl_compare[7, 2]) + " {0:.10f}  ".format(ovl_compare[7, 3]) + " {0:.10f} \n ".format(ovl_compare[7, 4]))
        datfile.write(" {0:.10f}  ".format(ovl_compare[7, 5]) + " {0:.10f}  ".format(ovl_compare[7, 6]) + " {0:.10f} \n".format(ovl_compare[7, 7]))
        datfile.write("------\n")
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")

        datfile.close()

        ovl = mol.readOverlapMatrix('muh.log', 8)
        for i, o in enumerate(ovl.flatten()):
            self.assertAlmostEqual(o, ovl_compare.flatten()[i], delta=1.e-7)

    def test_SmatrixFromOverlapMatrix(self):
        """
        Tests the correct extraction of the s matrix.
        """
        
        ovl_in = np.random.rand(4,4)
        ovl_compare = ovl_in[0:2, 2:4]
        
        ovl = mol.SmatrixFromOverlapMatrix(ovl_in)
        for i, o in enumerate(ovl.flatten()):
            self.assertAlmostEqual(o, ovl_compare.flatten()[i], delta=1.e-7)

    def test__format_coordinates(self):
        """
        Tests correct formatting of coordinates.
        """

        atomicSymbols = ['O','H','H']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.690, 0.60,
                      0.0,-0.690,0.60])

        comp_str = " O    0.0000000    0.0000000    0.0000000\n H    0.0000000    0.6900000    0.6000000\n H    0.0000000   -0.6900000    0.6000000"

        string = mol._format_coordinates(x, atomicSymbols)

        self.assertEqual(comp_str, string)

    def test__format_nrjobiphs(self):
        """
        Tests correct formatting job numbers.
        """

        comp_str = "1 2 3 4 5; 1 2 3 4 5"

        string = mol._format_nrjobiphs(5.0)

        self.assertEqual(comp_str, string)
        

# Add tests above to the suite definition:
def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(MolcasTestCase('test_init_input_options'))
    suite.addTest(MolcasTestCase('test_rasci'))
    suite.addTest(MolcasTestCase('test_rasci_D'))

## TODO make test better once the numerical gradients are better!!
    suite.addTest(MolcasTestCase('test_rasci_D_num'))

    suite.addTest(MolcasTestCase('test__store_log'))
    suite.addTest(MolcasTestCase('test__store_err'))
    suite.addTest(MolcasTestCase('test__store_inp'))

    suite.addTest(MolcasTestCase('test__parse_rasci_energy'))
    suite.addTest(MolcasTestCase('test__parse_rasci_gradient'))
    suite.addTest(MolcasTestCase('test__parse_rasci_D'))

    suite.addTest(MolcasTestCase('test__trackPhases'))
    suite.addTest(MolcasTestCase('test_readOverlapMatrix'))
    suite.addTest(MolcasTestCase('test_SmatrixFromOverlapMatrix'))

    suite.addTest(MolcasTestCase('test__format_coordinates'))
    suite.addTest(MolcasTestCase('test__format_nrjobiphs'))

## TODO just define template input files. Less important to test and implicitly tested through other tests.
    # suite.addTest(MolcasTestCase('test__input_template_rasci_E_D'))
    # suite.addTest(MolcasTestCase('test__input_template_rasci_E'))
    # suite.addTest(MolcasTestCase('test__input_template_rasci_D'))
    # suite.addTest(MolcasTestCase('test__input_template_rasci'))
## TOOD does not do anything, so not tested
    # suite.addTest(MolcasTestCase('test__format_efields'))
## TOOD does not do anything, so not tested
    # suite.addTest(MolcasTestCase('test__parse_execution_error'))
## TOOD does not do anything, so not tested
    # suite.addTest(MolcasTestCase('test__clean'))
## TODO implement tests. However, they are implicitly tested through a lot of other tests and this not so important
    # suite.addTest(MolcasTestCase('test__run_molcas'))

    return suite


if __name__ == "__main__":
    unittest.main()
