import os
import unittest
import numpy as np
from CDTK.Tools.OutputGrabber import suppress_stdout, suppress_stdoutstderr


import CDTK.Interfaces.EmbeddingInterface as emb
import CDTK.Interfaces.XMoleculeInterface as xmol
import CDTK.Interfaces.ForceField as ff
import CDTK.Tools.Utils as uti

class EmbeddingTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        os.remove('input.in')
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_gradient_pcharge_xmolecule(self):
        """
        Test of the partial charge implementation in XMOLECULE.
        Two partial charges with q=+1 separated 1 a.u. should
        yield a gradient of 1. The calculation must include He
        as XMOLECULE does not accept only partial charges.
        """
        
        # Configuration of the He + X + X system
        atomlist_He = ['He']
        pos_He = np.array([0.0, 0.0, 0.0])
        qoptions_dict = {'HF':['yes'], 'gto':['6-31G'], 'verbose':['True']}
        pcharges = [1.0, 1.0]
        pos_pcharges = np.array([100.0, 0.0, 0.0, 100.0, 0.0, 1.0])

        # XMOLECULE calculation
        with suppress_stdout():
            qce_obj = xmol.xmolecule()
            qce_obj.init_input_options(qoptions_dict, atomlist_He, pos_He, 
                                    x_charges=pcharges, x_pos=pos_pcharges)
            en, grad = qce_obj.energy_gradient_HF(np.concatenate((pos_He, pos_pcharges)))
        
        # Check gradient value
        self.assertAlmostEqual(grad[5], 1.0, delta=1e-9)
        self.assertAlmostEqual(grad[8], -1.0, delta=1e-9)

        pass

    def test_EnGrad_QMMMaddition_waterdimer(self):
        """
        Testing that QM and MM energies and gradients are correctly added for
        a QM/MM calculation with mechanical embedding of a water dimer where 
        the two molecules are very separated. The QM/MM results are compared
        with XMOLECULE and ForceField calculations.
        """
        
        # Configuration of the dimer
        atomlist_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomlist_dimer')
        atomlist_H2O = atomlist_dimer[:len(atomlist_dimer)//2]
        atomnums_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomnums_dimer')
        atommass_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atommass_dimer')
        atommass3_dimer = np.repeat(atommass_dimer, 3)
        pos_dimer = np.array(uti.readColumnFile('filesForTests/EmbeddingTest/posFAR_dimer'), dtype=float)
        vel_dimer  = np.zeros(pos_dimer.shape)
        posQM_H2O = pos_dimer[:len(pos_dimer)//2]
        posMM_H2O = pos_dimer[len(pos_dimer)//2:]
        topol_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/topol_dimer')
        topolMM = topol_dimer[len(topol_dimer)//2:]
        
        # Parameters for QCE and ForceField
        qoptions_dict = {'HF':['yes'], 'gto':['6-31G'], 'gs_occ':['yes'], 'occ':['22221'], 'N':['1'], 'verbose':['False']}
        forcefields = ['SPC/E', 'TIP4P/2005f']
       
        # !: Revise criteria
        tolEdimer = 1e-5
        tolGdimer = 1e-5

        # Embedding calculation
        for f in forcefields:
            with suppress_stdout():
                embedding_obj = emb.embedding(atomlist_dimer, atomnums_dimer, atommass3_dimer, pos_dimer, 
                                            topol=topol_dimer, forcefield=[f], region=['3'],
                                            quantum_options=qoptions_dict)
                e_emb, g_emb = embedding_obj.energy_gradient(pos_dimer, V=vel_dimer, SH=True)
                w = embedding_obj.getW(pos_dimer, V=vel_dimer)
                np.testing.assert_allclose(e_emb, np.diagonal(w), atol=1e-8, rtol=1e-2)
                e_emb = e_emb[0]
                g_emb = np.array(g_emb[0])

                # QCE calculation
                qce_obj = xmol.xmolecule()
                qce_obj.init_input_options(qoptions_dict, atomlist_H2O, posQM_H2O)
                e_qce, g_qce = qce_obj.energy_gradient_HF(posQM_H2O)

                # ForceField calculation
                ff_obj = ff.jointForceFields(topolMM, waterForcefield=f)
                e_ff, g_ff = ff_obj.getEnGrad(posMM_H2O)

                # Compare energies and gradients
                self.assertAlmostEqual(e_emb, e_qce + e_ff, delta=tolEdimer)
                np.testing.assert_allclose(g_emb, np.concatenate((g_qce, g_ff)), atol=tolGdimer)
        return        

    def test_EnGrad_MechEmbed_waterdimer(self):
        """
        Test the energy and gradient obtained from a QM/MM calculation with mechanical 
        embedding using the SPC/E and TIP4P/2005f force fields of a water dimer.
        """

        # Configuration of the dimer
        atomlist_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomlist_dimer')
        atomnums_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomnums_dimer')
        atommass_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atommass_dimer')
        atommass3_dimer = np.repeat(atommass_dimer, 3)
        pos_dimer = np.array(uti.readColumnFile('filesForTests/EmbeddingTest/posFAR_dimer'), dtype=float)
        vel_dimer  = np.zeros(pos_dimer.shape)
        topol_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/topol_dimer')
        
        # Parameters for QCE and ForceField
        qoptions_dict = {'HF':['yes'], 'gto':['6-31G'], 'gs_occ':['yes'], 'occ':['22221'], 'N':['1'], 'verbose':['False'],
                         'conv':['1e-14'], 'dens_conv':['1e-12'], 'eps_teint':['0.']}
        forcefields = ['SPC/E', 'TIP4P/2005f']

        # Reference values for QM/MM energies and gradients for a water dimer (FAR case)
        # Values for SCP/E and TIP4P/2005f forcefields respectively
        ref_ener = [-75.48261701, -75.48200512]
        ref_grad = [[-1.89672655e-02, -2.44906634e-02,  4.29136417e-16, -1.05287313e-02,
                      2.77474187e-02,  1.36608399e-16,  2.94958337e-02, -3.25665641e-03,
                     -5.65744816e-16,  3.29089200e-06, -2.01053491e-06,  0.00000000e+00,
                     -1.50445100e-06,  1.00466845e-06,  0.00000000e+00, -1.62326139e-06,
                      9.06937459e-07,  0.00000000e+00],
                    [-1.89658722e-02, -2.44906627e-02,  9.41778864e-16, -1.05294871e-02,
                      2.77474184e-02, -1.13487007e-15,  2.94951522e-02, -3.25662949e-03,
                      1.93091204e-16, -2.32552369e-02,  2.02175963e-05,  0.00000000e+00,
                      1.16208814e-02, -8.57887775e-03,  0.00000000e+00,  1.16345626e-02,
                      8.55853399e-03,  0.00000000e+00]
                     ]

        for ii, f in enumerate(forcefields):
            with suppress_stdout():
                embedding_obj = emb.embedding(atomlist_dimer, atomnums_dimer, atommass3_dimer, pos_dimer, 
                                            topol=topol_dimer, forcefield=[f], region=['3'],
                                            quantum_options=qoptions_dict)
                e_emb, g_emb = embedding_obj.energy_gradient(pos_dimer, V=vel_dimer, SH=True)
                w = embedding_obj.getW(pos_dimer, V=vel_dimer)
                np.testing.assert_allclose(e_emb, np.diagonal(w), atol=1e-8, rtol=1e-2)
                e_emb, g_emb = e_emb[0], np.array(g_emb[0])
                self.assertAlmostEqual(e_emb, ref_ener[ii], delta=1e-8)
                np.testing.assert_allclose(g_emb, ref_grad[ii], atol=1e-8, rtol=1e-4)

    def test_EnGrad_ElecEmbed_waterdimer(self):
        """
        Test the energy and gradient obtained from a QM/MM calculation with electrostatic 
        embedding using the SPC/E and TIP4P/2005f force fields of a water dimer.
        """

        # Configuration of the dimer
        atomlist_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomlist_dimer')
        atomnums_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atomnums_dimer')
        atommass_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/atommass_dimer')
        atommass3_dimer = np.repeat(atommass_dimer, 3)
        pos_dimer = np.array(uti.readColumnFile('filesForTests/EmbeddingTest/posFAR_dimer'), dtype=float)
        topol_dimer = uti.readColumnFile('filesForTests/EmbeddingTest/topol_dimer')
        vel_dimer = np.zeros(pos_dimer.shape)
        vel_dimer[0:3] = 0.
        vel_dimer[2:6] = 0.1
        vel_dimer[5:9] = 0.2
        vel_dimer[8:12] = -10
        vel_dimer[11:15] = -5
        vel_dimer[14:18] = -5
        # Parameters for QCE and ForceField
        qoptions_dict = {'HF':['yes'], 'gto':['6-31G'], 'gs_occ':['yes'], 'occ':['22221'], 'N':['1'], 'verbose':['False'], 'qmmm':['yes']}
        forcefields = ['SPC/E', 'TIP4P/2005f']

        # Reference values for QM/MM energies and gradients for a water dimer (FAR case)
        # Values for SCP/E and TIP4P/2005f forcefields respectively
        ref_ener = [-75.48282486, -75.48220498]
        ref_grad = [[-1.89701389e-02, -2.44917684e-02,  1.70156264e-16, -1.05311076e-02,
                      2.77479219e-02,  1.13587729e-15,  2.94949015e-02, -3.25600655e-03,
                     -1.30603354e-15,  1.89982270e-04, -2.20255830e-06, -2.88766475e-22,
                     -9.17914510e-05,  2.98444413e-06,  1.37644251e-22, -9.19148176e-05,
                     -8.87235849e-07,  1.35453724e-22],
                     [-1.89697655e-02, -2.44915950e-02, -3.68617864e-16, -1.05313655e-02,
                       2.77477540e-02, -2.60503392e-16,  2.94949954e-02, -3.25601628e-03,
                       6.29121240e-16, -2.30773495e-02,  2.06337404e-05,  4.88245556e-22,
                       1.15348471e-02, -8.57661659e-03, -2.32864980e-22,  1.15485712e-02,
                        8.55588054e-03, -2.31850600e-22]
                    ]
        ref_d = [
            [[0.-0.j        , 0.-4.00631552j, 0.+2.89251804j, 0.+0.85696271j],
            [0.+4.00631552j, 0.-0.j        , 0.+0.01008775j, 0.-0.01041286j],
            [0.-2.89251804j, 0.-0.01008775j, 0.-0.j        , 0.+0.01820042j],
            [0.-0.85696271j, 0.+0.01041286j, 0.-0.01820042j, 0.-0.j        ]],
            [[0.-0.j        , 0.-4.00631552j, 0.+2.89251804j, 0.+0.85696271j],
            [0.+4.00631552j, 0.-0.j        , 0.+0.01008775j, 0.-0.01041286j],
            [0.-2.89251804j, 0.-0.01008775j, 0.-0.j        , 0.+0.01820042j],
            [0.-0.85696271j, 0.+0.01041286j, 0.-0.01820042j, 0.-0.j        ]]
            ]
        ref_nac12 = [
            [ 7.07548697e-01, -5.48067875e-01, -5.76255815e-16, -4.14330830e-01,
              1.95858087e-01,  2.66771188e-16, -2.93217832e-01,  3.52209819e-01,
              3.09484358e-16,  5.65493864e-06,  1.74435478e-06,  5.47433499e-21,
             -2.73318187e-06, -7.39616927e-07, -2.61264416e-21, -2.63159170e-06,
             -9.20443612e-07, -2.60293479e-21],
            [ 7.07548537e-01, -5.48067483e-01, -6.36023068e-16, -4.14330665e-01,
              1.95857930e-01,  5.47198037e-16, -2.93217838e-01,  3.52209583e-01,
              8.88247151e-17,  5.39136814e-06,  1.66461538e-06,  6.51310019e-21,
             -2.62201999e-06, -6.72667287e-07, -3.10816040e-21, -2.48864220e-06,
             -9.10071834e-07, -3.09580984e-21]
            ]

        for ii, f in enumerate(forcefields):
            with suppress_stdout():
                embedding_obj = emb.embedding(atomlist_dimer, atomnums_dimer, atommass3_dimer, pos_dimer, 
                                            topol=topol_dimer, forcefield=[f], region=['3'],
                                            quantum_options=qoptions_dict, electrostatic=["True"])
                e_emb, g_emb = embedding_obj.energy_gradient(pos_dimer, V=vel_dimer, SH=True)
                d_emb = embedding_obj.getD(pos_dimer, V=vel_dimer)
                nac12 = embedding_obj.nac(pos_dimer, 1,2, V=vel_dimer)
                w = embedding_obj.getW(pos_dimer, V=vel_dimer)
                np.testing.assert_allclose(e_emb, np.diagonal(w), atol=1e-8, rtol=1e-2)
                e_emb, g_emb = e_emb[0], np.array(g_emb[0])
                self.assertAlmostEqual(e_emb, ref_ener[ii], delta=1e-8)
                np.testing.assert_allclose(g_emb, ref_grad[ii], atol=1e-8, rtol=1e-2)
                np.testing.assert_allclose(d_emb, ref_d[ii], atol=1e-8, rtol=1e-2)
                np.testing.assert_allclose(nac12, ref_nac12[ii], atol=1e-8, rtol=1e-2)

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(EmbeddingTestCase('test_coulomb_MechEmbed_ElecEmbed'))
    suite.addTest(EmbeddingTestCase('test_gradient_pcharge_xmolecule'))
    suite.addTest(EmbeddingTestCase('test_EnGrad_QMMMaddition_waterdimer'))
    suite.addTest(EmbeddingTestCase('test_EnGrad_MechEmbed_waterdimer'))
    suite.addTest(EmbeddingTestCase('test_EnGrad_ElecEmbed_waterdimer'))
    return suite

if __name__ == "__main__":
    unittest.main()
