#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import unittest
import numpy as np
import os
import CDTK.Interfaces.GamessUSInterface as gms
import CDTK.Tools.EField as efi
import CDTK.Tools.Conversion as cv

class GmsTestCase(unittest.TestCase):
    def setUp(self):
        pass
    
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_init_input_options(self):
        """
        Tests the correct setting of input options. Only some options are tested.
        """
        
        g = gms.gamess()

        opts = {'pathgms': ['/not/here/gms'], 'scratch': ['/temp'], 'iroot': ['17']}
        g.init_input_options(opts)

        self.assertEqual(g.engine, 'GamessUS')
        self.assertEqual(g._counter, 0)
        self.assertEqual(g._norb, 1)
        self.assertEqual(g._vec, '')
        self.assertEqual(g.pathgms, opts['pathgms'][0])
        self.assertEqual(g.scratch, opts['scratch'][0])
        self.assertEqual(g.iroot, opts['iroot'][0])
    
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_energy_HF(self):
        """
        Tests the correct return of a Hartree-Fock energy for the case of no
        electric field and an electric field included.
        """

        g = gms.gamess()
        g.is_storedat = True
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        e = g.energy_HF(x)
        self.assertAlmostEqual(e,-75.7447722639,delta=1.e-7)

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01
        ef.A = a
        g.efields = [ef]
        e = g.energy_HF(x)
        self.assertAlmostEqual(e,-75.7353240393,delta=1.e-7)
    
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_energy_HF_HOMO(self):
        """
        Tests the correct return of a Hartree-Fock energy and HOMO energy for the case of no
        electric field and an electric field included.
        """
        
        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        e, homo = g.energy_HF_HOMO(x)
        self.assertAlmostEqual(e,-75.7447722639,delta=1.e-7)

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01
        ef.A = a
        g.efields = [ef]
        e, homo = g.energy_HF_HOMO(x)
        self.assertAlmostEqual(e,-75.7353240393,delta=1.e-7)
    
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_energy_gradient_HF(self):
        """
        Tests the correct return of a Hartree-Fock energy and gradient for the case of no
        electric field and an electric field included.
        """

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        e, grd = g.energy_gradient_HF(x)
        
        self.assertAlmostEqual(e,-75.7447722639,delta=1.e-7)
        grd_compare = [-0.58378632, 0.29189316, 0.29189316, 0.29189316, -0.46831885, 0.17642569, 0.29189316, 0.17642569, -0.46831885]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([1.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,1.0])
        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01
        ef.A = a
        g.efields = [ef]
        e, grd = g.energy_gradient_HF(x)
        
        self.assertAlmostEqual(e,-75.7353240393,delta=1.e-7)
        grd_compare = [-0.57886527, 0.29419973, 0.29419973, 0.28943264, -0.46970384, 0.17550411, 0.28943263, 0.17550411, -0.46970384]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_MCSCF(self):
        """
        Tests the correct return of a MCSCF energy and gradient.
        """

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        g.icharg = 1
        g.mult = 2
        g.ncore = 1
        g.nact = 4
        g.iroot = 1
        g.wstate = '1.0,1.0'
        g.nels = 7
        g.nstate = 4
        g.sz = 0.5
        x = np.array([0.0,0.0,0.0,
                      0.0,0.7572157*cv.an2au,0.5865358*cv.an2au,
                      0.0,-0.7572157*cv.an2au,0.5865358*cv.an2au])
        ener,grd = g.mcscf(x)
        e_compare = [-75.6104751604, -75.5406032915]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [-2.00000000e-09, 3.60000000e-08, 3.75572100e-03, 1.00000000e-09, -4.36248700e-02, -1.87786100e-03, 1.00000000e-09, 4.36248350e-02, -1.87786000e-03, 2.10000000e-08, -3.01000000e-07, -7.40730290e-02, -1.00000000e-08, -7.39816790e-02, 3.70364900e-02, -1.00000000e-08, 7.39819800e-02, 3.70365390e-02, 0.00000000e+00, 3.60000000e-08, 3.75572100e-03, 0.00000000e+00, -4.36248700e-02, -1.87786100e-03, 0.00000000e+00, 4.36248350e-02, -1.87786000e-03]
        
        for i, g in enumerate(grd.flatten()):
            rmsd = np.sqrt(np.sum(np.array(grd_compare)**2))
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-4)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_NACME(self):
        """
        Tests the correct return of a MCSCF energy, gradient, non-adiabatic
        couplings and dipole matrix elements.
        """

        g = gms.gamess()
        g.is_storelog = True
        g.is_storedat = False
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        g.icharg = 1
        g.mult = 2
        g.ncore = 1
        g.nact = 4
        g.nels = 7
        g.nstate = 4
        g.wstate = '1.0,1.0,1.0,1.0'
        g.sz = 0.5
        x = np.array([0.0,0.0,0.0,
                      0.0,0.7572157*cv.an2au,0.5865358*cv.an2au,
                      0.0,-0.7572157*cv.an2au,0.5865358*cv.an2au])
        ener,grd,nac,mu = g.nacme(x)

        e_compare = [-75.6094796312, -75.5386963807, -75.3607528191, -74.7393172728]
        for i, e in enumerate(ener):
            self.assertAlmostEqual(e, e_compare[i], delta=1.e-7)

        grd_compare = [0., 0., 0.00259793, 0., -0.04317889, -0.00129915, 0., 0.04317863, -0.00129943, 0., 0., -0.07465771, 0., -0.07516999, 0.03732882, 0., 0.07516985, 0.03732863, 0., 0., 0.23396378, 0., -0.07753489, -0.11698173, 0., 0.07753521, -0.11698128, 0., 0., 0.10764288, 0.,-0.15435616, -0.05382137, 0., 0.15435624, -0.05382137]
        for i, g in enumerate(grd.flatten()):
            self.assertAlmostEqual(g, grd_compare[i], delta=1.e-6)

        nac_compare = {(0, 1): [-0.9441664, -0., -0., 0.4270456, -0., -0., 0.4270452, -0., -0. ], (1, 2): [-0., 0.6749733, -0., -0., -0.3039814, 0.077105, -0., -0.3039813, -0.0771055], (3, 2): [ 0., -0.6268447, 0., 0., 0.1233816, 0.1798321, 0., 0.1233815, -0.1798326], (1, 3): [-0., -0., -0.3498146, -0., 0.1148667, -0.0204093, -0., -0.114867, -0.0204095], (3, 0): [-0.1461595, 0., 0., -0.1178219, 0., 0., -0.1178219, 0., 0. ], (3, 1): [ 0., 0., 0.3498146, 0., -0.1148667, 0.0204093, 0., 0.114867, 0.0204095], (2, 1): [ 0., -0.6749733, 0., 0., 0.3039814, -0.077105, 0., 0.3039813, 0.0771055], (2, 0): [ 0., 0., 0., 0.3337191, 0., 0., -0.333719, 0., 0. ], (2, 3): [-0., 0.6268447, -0., -0., -0.1233816, -0.1798321, -0., -0.1233815, 0.1798326], (1, 0): [ 0.9441664, 0., 0., -0.4270456, 0., 0., -0.4270452, 0., 0. ], (0, 3): [ 0.1461595, -0., -0., 0.1178219, -0., -0., 0.1178219, -0., -0. ], (0, 2): [-0., -0., -0., -0.3337191, -0., -0., 0.333719, -0., -0. ]}
        for key in nac_compare:
            for i, n in enumerate(nac[key[0],key[1],:]):
                self.assertAlmostEqual(n, nac_compare[key][i], delta=1.e-5)

        mu_compare = {(0, 1): [ 0.1961077, 0., 0. ], (1, 2): [ 0., -0.07206109, 0. ], (3, 2): [ 0., -0.76626336, 0. ], (0, 0): [ 0., 0., 1.16138581], (3, 3): [ 0., 0., 1.54453735], (3, 0): [-0.61196668, 0., 0. ], (3, 1): [ 0., 0., 0.61418169], (2, 1): [ 0., -0.07206109, 0. ], (0, 2): [ 0., 0., 0.], (2, 0): [ 0., 0., 0.], (1, 3): [ 0., 0., 0.61418169], (2, 3): [ 0., -0.76626336, 0. ], (2, 2): [ 0., 0., 1.49760977], (1, 0): [ 0.1961077, 0., 0. ], (0, 3): [-0.61196668, 0., 0. ], (1, 1): [ 0., 0., 0.9073097]}

        for key in mu:
            for i, m in enumerate(mu[key]):
                self.assertAlmostEqual(m, mu_compare[key][i], delta=1.e-5)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_hessian(self):
        """
        Tests the correct calculation of a hessian.
        """

        g = gms.gamess()
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.04*cv.an2au,
                      0.0,0.790*cv.an2au,0.5667*cv.an2au,
                      0.0,-0.790*cv.an2au,0.5667*cv.an2au])
        hessian = g.hessian(x)
        
        hess_compare = np.array([-3.50361772e-04, 0.00000000e+00, 0.00000000e+00, 1.75000096e-04, 0.00000000e+00, 0.00000000e+00, 1.75000096e-04, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 8.29033020e-01, -3.39987737e-15, 0.00000000e+00, -4.14516170e-01, -2.76478074e-01, 0.00000000e+00, -4.14516170e-01, 2.76478074e-01, 0.00000000e+00, -3.39987737e-15, 4.25550186e-01, 0.00000000e+00, -2.03967281e-01, -2.12775433e-01, 0.00000000e+00, 2.03967281e-01, -2.12775433e-01, 1.75000096e-04, 0.00000000e+00, 0.00000000e+00, -1.70419161e-04, 0.00000000e+00, 0.00000000e+00, -4.40014458e-06, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, -4.14516170e-01, -2.03967281e-01, 0.00000000e+00, 4.42457548e-01, 2.40222470e-01, 0.00000000e+00, -2.79417185e-02, -3.62549294e-02, 0.00000000e+00, -2.76478074e-01, -2.12775433e-01, 0.00000000e+00, 2.40222470e-01, 1.98470030e-01, 0.00000000e+00, 3.62549294e-02, 1.43057438e-02, 1.75000096e-04, 0.00000000e+00, 0.00000000e+00, -4.40014458e-06, 0.00000000e+00, 0.00000000e+00, -1.70419161e-04, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00, -4.14516170e-01, 2.03967281e-01, 0.00000000e+00, -2.79417185e-02, 3.62549294e-02, 0.00000000e+00, 4.42457548e-01, -2.40222470e-01, 0.00000000e+00, 2.76478074e-01, -2.12775433e-01, 0.00000000e+00, -3.62549294e-02, 1.43057438e-02, 0.00000000e+00, -2.40222470e-01, 1.98470030e-01])
        hess_compare = hess_compare.reshape((9,9))

        for i, line in enumerate(hessian):
            for j, h in enumerate(line):
                self.assertAlmostEqual(h, hess_compare[i][j], delta=1.e-7)
        
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_optimize(self):
        """
        Tests the correct coordinate optimization.
        """

        g = gms.gamess()
        g.is_storelog = True
        g.atomicSymbols = ['O','H','H']
        g.atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.690*cv.an2au,0.60*cv.an2au,
                      0.0,-0.690*cv.an2au,0.60*cv.an2au])
        x0 = g.optimize(x)
        
        x0_compare = np.array([0.0, 0.0, 0.04867738, 0.0, 0.79015823, 0.57566131, 0.0, -0.79015823, 0.57566131])

        for i, xx in enumerate(x0):
            self.assertAlmostEqual(xx, x0_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_read_input_template(self):
        """
        Tests the correct reading of an input template. Please note that the file is just
        read and no syntax or semantics checks are performed.
        """

        datfile = open('input.inp', 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        g = gms.gamess()
        g.read_input_template('input.inp')
  
        self.assertEqual(g.itemplate, text)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_efield(self):
        """
        Tests the correct return of electric fields. Tests for one field and two fields.
        """

        g = gms.gamess()
        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01
        ef.A = a
        g.efields = [ef]

        e1 = g.efield()

        e1_compare = [0.01, 0.0, 0.0]
        for i, e in enumerate(e1):
            self.assertAlmostEqual(e, e1_compare[i], delta=1.e-7)


        ef2 = efi.EField()
        ef2.P = np.array([0,0,1],float)
        def a2(t): return 1.00
        ef2.A = a2
        g.efields.append(ef2)

        e2_compare = [0.01, 0.0, 1.0]

        e2 = g.efield()
        for i, e in enumerate(e2):
            self.assertAlmostEqual(e, e2_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_dt_efield(self):
        """
        Tests the correct return of time derivative of electric fields. Tests for one field and two fields.
        """

        g = gms.gamess()
        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01*t
        ef.A = a
        g.efields = [ef]

        e1 = g.dt_efield()

        e1_compare = [0.01, 0.0, 0.0]
        for i, e in enumerate(e1):
            self.assertAlmostEqual(e, e1_compare[i], delta=1.e-7)


        ef2 = efi.EField()
        ef2.P = np.array([0,0,1],float)
        def a2(t): return 1.00*t
        ef2.A = a2
        g.efields.append(ef2)

        e2_compare = [0.01, 0.0, 1.0]

        e2 = g.dt_efield()
        for i, e in enumerate(e2):
            self.assertAlmostEqual(e, e2_compare[i], delta=1.e-7)
    
    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__store_log(self):
        """
        Tests storeing of log file.
        """
        g = gms.gamess()
        g.is_storelog = True

        datfile = open(g._log, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        g._store_log()

        cstr = '{0:0>7d}'.format(g._counter)
        logfile = open(cstr + '_' + g._log, 'r')
        self.assertEqual(logfile.read(), text)
        logfile.close()

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__store_dat(self):
        """
        Tests storeing of dat file.
        """
        g = gms.gamess()
        g.is_storedat = True

        datfile = open(g._dat, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        g._store_dat()

        cstr = '{0:0>7d}'.format(g._counter)
        datfile = open(cstr + '_' + g._dat, 'r')
        self.assertEqual(datfile.read(), text)
        datfile.close()

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")                
    def test__store_inp(self):
        """
        Tests storeing of inp file.
        """

        g = gms.gamess()
        g.is_storeinp = True

        datfile = open(g._inp, 'w')
        text = " $CONTRL\n    RUNTYP=HESSIAN\n    SCFTYP=RHF\n    ICHARG={icharg}\n    MULT={mult}\n $END\n\n $SYSTEM\n    MEMDDI={memddi}\n    MWORDS={mwords}\n    PARALL=.TRUE.\n $END\n\n $BASIS\n    GBASIS={gbasis}\n    NGAUSS={ngauss}\n $END\n\n $GUESS\n    GUESS={guess}\n    NORB={norb}\n $END\n\n $DATA\n    Generated by CDTK interface\n    C1\n{xcart}\n $END\n\n{vec}\n"
        datfile.write(text)
        datfile.close()

        g._store_inp()

        cstr = '{0:0>7d}'.format(g._counter)
        inpfile = open(cstr + '_' + g._inp, 'r')
        self.assertEqual(inpfile.read(), text)
        inpfile.close()

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")        
    def test__clean(self):
        """
        Tests cleaning of data.
        """
        g = gms.gamess()
        os.system("touch " + g._inp + " " + g._log + " " + g._dat)

        self.assertTrue(os.path.isfile(g._inp))
        self.assertTrue(os.path.isfile(g._log))
        self.assertTrue(os.path.isfile(g._dat))

        g._clean()

        self.assertFalse(os.path.isfile(g._inp))
        self.assertFalse(os.path.isfile(g._log))
        self.assertFalse(os.path.isfile(g._dat))

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")                
    def test__parse_energy_HF(self):
        """
        Tests the correct parsing of HF energies.
        """

        e_compare = np.random.rand(1)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" TOTAL ENERGY = {0:.10f}\n".format(e_compare[0]))
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        e = g._parse_energy_HF()
        self.assertAlmostEqual(e, e_compare[0], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_energy_CI(self):
        """
        Tests the correct parsing of CI energies.
        """

        e_compare = np.random.rand(5)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        for i, ee in enumerate(e_compare):
            datfile.write(" STATE " + str(i) + " ENERGY= {0:.10f}\n".format(ee))
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        e = g._parse_energy_CI()
        for i, ee in enumerate(e):
            self.assertAlmostEqual(ee, e_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_mcscf_energy(self):
        """
        Tests the correct parsing of MCSCF energies.
        """

        e_compare = np.random.rand(5)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        for i, ee in enumerate(e_compare):
            datfile.write(" STATE " + str(i) + " ENERGY= {0:.10f}\n".format(ee))
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        e = g._parse_mcscf_energy()
        for i, ee in enumerate(e):
            self.assertAlmostEqual(ee, e_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_nacme_energy(self):
        """
        Tests the correct parsing of non-adiabatic coupling energies.
        """

        e_compare = np.random.rand(5)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        for i, ee in enumerate(e_compare):
            datfile.write(" COMPUTING NACME FOR STATE WITH E= {0:.10f}\n".format(ee))
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        e = g._parse_nacme_energy()
        for i, ee in enumerate(e):
            self.assertAlmostEqual(ee, e_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_gradient_HF(self):
        """
        Tests the correct parsing of HF gradients.
        """

        g_compare = np.random.rand(9)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" UNITS ARE HARTREE/BOHR \n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[0]) + " {0:.10f}  ".format(g_compare[1]) + " {0:.10f}\n".format(g_compare[2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[3]) + " {0:.10f}  ".format(g_compare[4]) + " {0:.10f}\n".format(g_compare[5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[6]) + " {0:.10f}  ".format(g_compare[7]) + " {0:.10f}\n".format(g_compare[8]))
        datfile.write("\n")
        datfile.write("DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.close()

        grd = g._parse_gradient_HF()
        for i, gg in enumerate(grd):
            self.assertAlmostEqual(gg, g_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_nacme_gradient(self):
        """
        Tests the correct parsing of non-adiabatic coupling gradients.
        """

        g_compare = np.random.rand(27).reshape((3,9))
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" UNITS ARE HARTREE/BOHR \n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[0][0]) + " {0:.10f}  ".format(g_compare[0][1]) + " {0:.10f}\n".format(g_compare[0][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[0][3]) + " {0:.10f}  ".format(g_compare[0][4]) + " {0:.10f}\n".format(g_compare[0][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[0][6]) + " {0:.10f}  ".format(g_compare[0][7]) + " {0:.10f}\n".format(g_compare[0][8]))
        datfile.write("\n")
        datfile.write("\n")
        datfile.write("DUMMY LINE 3\n" + "DUMMY LINE 4\n" + "DUMMY LINE 5\n")
        datfile.write(" UNITS ARE HARTREE/BOHR \n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[1][0]) + " {0:.10f}  ".format(g_compare[1][1]) + " {0:.10f}\n".format(g_compare[1][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[1][3]) + " {0:.10f}  ".format(g_compare[1][4]) + " {0:.10f}\n".format(g_compare[1][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[1][6]) + " {0:.10f}  ".format(g_compare[1][7]) + " {0:.10f}\n".format(g_compare[1][8]))
        datfile.write("\n")
        datfile.write("\n")
        datfile.write("DUMMY LINE 6\n" + "DUMMY LINE 7\n" + "DUMMY LINE 8\n")
        datfile.write(" UNITS ARE HARTREE/BOHR \n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[2][0]) + " {0:.10f}  ".format(g_compare[2][1]) + " {0:.10f}\n".format(g_compare[2][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[2][3]) + " {0:.10f}  ".format(g_compare[2][4]) + " {0:.10f}\n".format(g_compare[2][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[2][6]) + " {0:.10f}  ".format(g_compare[2][7]) + " {0:.10f}\n".format(g_compare[2][8]))
        datfile.write("\n")
        datfile.write("DUMMY LINE 9\n" + "DUMMY LINE 10\n" + "DUMMY LINE 11\n")
        datfile.close()

        grd = g._parse_nacme_gradient()
        for i, line in enumerate(grd):
            for j, gg in enumerate(line):
                self.assertAlmostEqual(gg, g_compare[i][j], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_nacme_nacme(self):
        """
        Tests the correct parsing of non-adiabatic couplings.
        """

        nacme_compare = {(0, 0): np.random.rand(9), (0, 1): np.random.rand(9), (1, 1): np.random.rand(9)}
        nacme_compare[(1, 0)] = -nacme_compare[(0, 1)]
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" <STATE 1, WEIGHT= 0.5 | D/DQ | STATE 1, WEIGHT= 0.5> \n MUH \n")
        datfile.write(" 1 A  {0:.10f}  ".format(nacme_compare[(0, 0)][0]) + " {0:.10f}  ".format(nacme_compare[(0, 0)][1]) + " {0:.10f}\n".format(nacme_compare[(0, 0)][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(nacme_compare[(0, 0)][3]) + " {0:.10f}  ".format(nacme_compare[(0, 0)][4]) + " {0:.10f}\n".format(nacme_compare[(0, 0)][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(nacme_compare[(0, 0)][6]) + " {0:.10f}  ".format(nacme_compare[(0, 0)][7]) + " {0:.10f}\n".format(nacme_compare[(0, 0)][8]))
        datfile.write("\n")
        datfile.write("\n")
        datfile.write("DUMMY LINE 3\n" + "DUMMY LINE 4\n" + "DUMMY LINE 5\n")
        datfile.write(" <STATE 2, WEIGHT= 0.5 | D/DQ | STATE 1, WEIGHT= 0.5> \n MUH \n")
        datfile.write(" 1 A  {0:.10f}  ".format(nacme_compare[(1, 0)][0]) + " {0:.10f}  ".format(nacme_compare[(1, 0)][1]) + " {0:.10f}\n".format(nacme_compare[(1, 0)][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(nacme_compare[(1, 0)][3]) + " {0:.10f}  ".format(nacme_compare[(1, 0)][4]) + " {0:.10f}\n".format(nacme_compare[(1, 0)][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(nacme_compare[(1, 0)][6]) + " {0:.10f}  ".format(nacme_compare[(1, 0)][7]) + " {0:.10f}\n".format(nacme_compare[(1, 0)][8]))
        datfile.write("\n")
        datfile.write("\n")
        datfile.write("DUMMY LINE 9\n" + "DUMMY LINE 10\n" + "DUMMY LINE 11\n")
        datfile.write(" <STATE 2, WEIGHT= 0.5 | D/DQ | STATE 2, WEIGHT= 0.5> \n MUH \n")
        datfile.write(" 1 A  {0:.10f}  ".format(nacme_compare[(1, 1)][0]) + " {0:.10f}  ".format(nacme_compare[(1, 1)][1]) + " {0:.10f}\n".format(nacme_compare[(1, 1)][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(nacme_compare[(1, 1)][3]) + " {0:.10f}  ".format(nacme_compare[(1, 1)][4]) + " {0:.10f}\n".format(nacme_compare[(1, 1)][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(nacme_compare[(1, 1)][6]) + " {0:.10f}  ".format(nacme_compare[(1, 1)][7]) + " {0:.10f}\n".format(nacme_compare[(1, 1)][8]))
        datfile.write("\n")
        datfile.write("DUMMY LINE 12\n" + "DUMMY LINE 13\n" + "DUMMY LINE 14\n")
        datfile.close()

        nacme = g._parse_nacme_nacme()
        for i, key in enumerate(nacme):
            for j, n in enumerate(nacme[key]):
                self.assertAlmostEqual(n, nacme_compare[key][j], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_nacme_dipole(self):
        """
        Tests the correct parsing of dipole couplings.
        """

        dip_compare = {(0, 0): np.random.rand(3), (0, 1): np.random.rand(3), (1, 1): np.random.rand(3)}
        dip_compare[(1, 0)] = dip_compare[(0, 1)]
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" --- DIPOLE MOMENTS (DEBYE) --- \n")
        datfile.write(" 1, 1 : DIPOLE  {0:.10f}  ".format(dip_compare[(0, 0)][0]) + " {0:.10f}  ".format(dip_compare[(0, 0)][1]) + " {0:.10f}\n".format(dip_compare[(0, 0)][2]))
        datfile.write(" 1, 2 : DIPOLE  {0:.10f}  ".format(dip_compare[(0, 1)][0]) + " {0:.10f}  ".format(dip_compare[(0, 1)][1]) + " {0:.10f}\n".format(dip_compare[(0, 1)][2]))
        datfile.write(" 2, 2 : DIPOLE  {0:.10f}  ".format(dip_compare[(1, 1)][0]) + " {0:.10f}  ".format(dip_compare[(1, 1)][1]) + " {0:.10f}\n".format(dip_compare[(1, 1)][2]))
        datfile.write("\n \n")
        datfile.write("DUMMY LINE 12\n" + "DUMMY LINE 13\n" + "DUMMY LINE 14\n")
        datfile.close()

        dip = g._parse_nacme_dipole()
        for i, key in enumerate(dip):
            for j, n in enumerate(dip[key]):
                self.assertAlmostEqual(n, dip_compare[key][j], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_vec(self):
        """
        Tests the correct parsing of vec data.
        """

        vec_compare = np.random.rand(20,20)
        compare = []
        g = gms.gamess()
        datfile = open(g._dat, 'w')
        datfile.write("DUMMY LINE\n" + " $VEC\n")

        for i in range(20):
            datfile.write(str(i+1) + ' 1 ')
            for j in range(5):
                datfile.write("{0:.10E} ".format(vec_compare[i][j]))
            datfile.write('\n' + str(i+1) + ' 2 ')
            for j in range(5,10):
                datfile.write("{0:.10E} ".format(vec_compare[i][j]))
            datfile.write('\n' + str(i+1) + ' 3 ')
            for j in range(10,15):
                datfile.write("{0:.10E} ".format(vec_compare[i][j]))
            datfile.write('\n' + str(i+1) + ' 4 ')
            for j in range(15,20):
                datfile.write("{0:.10E} ".format(vec_compare[i][j]))
            datfile.write('\n')

        datfile.write(" $END \n")
        datfile.close()

        datfile = open(g._dat, 'r')
        datfile.readline()
        compare = datfile.read()
        datfile.close()

        vec, orb = g._parse_vec()
        
        self.assertEqual(orb, 20)
        self.assertEqual(vec, compare)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_equilibrium_geometry(self):
        """
        Tests the correct parsing of optimized geometry.
        """


        geom_compare = np.random.rand(9)
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write("DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")

        datfile.write("              ***** EQUILIBRIUM GEOMETRY LOCATED ***** \n")
        datfile.write(" COORDINATES OF ALL ATOMS ARE (ANGS) \n")
        datfile.write("   ATOM   CHARGE       X              Y              Z \n")
        datfile.write(" ------------------------------------------------------------ \n")
        datfile.write(" O           8.0   {0:.10f}  ".format(geom_compare[0]) + " {0:.10f}  ".format(geom_compare[1]) + " {0:.10f}\n".format(geom_compare[2]))
        datfile.write(" H           1.0   {0:.10f}  ".format(geom_compare[3]) + " {0:.10f}  ".format(geom_compare[4]) + " {0:.10f}\n".format(geom_compare[5]))
        datfile.write(" H           1.0   {0:.10f}  ".format(geom_compare[6]) + " {0:.10f}  ".format(geom_compare[7]) + " {0:.10f}\n".format(geom_compare[8]))
        datfile.write("\n \n")
        datfile.write("DUMMY LINE 12\n" + "DUMMY LINE 13\n" + "DUMMY LINE 14\n")
        datfile.close()

        geom = g._parse_equilibrium_geometry()
        for i, x in enumerate(geom):
            self.assertAlmostEqual(x, geom_compare[i], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_energy_CI_surf(self):
        """
        Correct parsing of PES scan.
        """

        scan_compare = np.random.rand(10, 5)

        datfile = open('scan.dat', 'w')

        datfile.write(str(0) + "th DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
        datfile.write(" STATE 1 ENERGY= {0:.10f} ".format(scan_compare[0, 1]))
        datfile.write(" S= {0:.10f} ".format(0.0))
        datfile.write(" SZ= {0:.10f} \n".format(0.0))
        datfile.write(" STATE 2 ENERGY= {0:.10f} ".format(scan_compare[0, 2]))
        datfile.write(" S= {0:.10f} ".format(1.0))
        datfile.write(" SZ= {0:.10f} \n".format(1.0))
        datfile.write(" STATE 3 ENERGY= {0:.10f} ".format(scan_compare[0, 3]))
        datfile.write(" S= {0:.10f} ".format(0.5))
        datfile.write(" SZ= {0:.10f} \n".format(0.5))
        datfile.write(" STATE 4 ENERGY= {0:.10f} ".format(scan_compare[0, 4]))
        datfile.write(" S= {0:.10f} ".format(3.0))
        datfile.write(" SZ= {0:.10f} \n".format(0.0))
        datfile.write("\n")
        datfile.write(str(0) + "th DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
        datfile.write(str(0) + "th DUMMY LINE 7\n" + "DUMMY LINE 8\n" + "DUMMY LINE 9\n")

        for i in range(10):
            datfile.write(str(i) + "th DUMMY LINE 1\n" + "DUMMY LINE 2\n" + "DUMMY LINE 3\n")
            datfile.write("\n")
            datfile.write(" COORD 1=  {0:.10f} \n".format(scan_compare[i, 0]))
            datfile.write("\n")
            datfile.write(str(i) + "th DUMMY LINE 4\n" + "DUMMY LINE 5\n" + "DUMMY LINE 6\n")
            datfile.write(str(i) + "th DUMMY LINE 7\n" + "DUMMY LINE 8\n" + "DUMMY LINE 9\n")
            datfile.write("\n")
            datfile.write(" STATE 1 ENERGY= {0:.10f} ".format(scan_compare[i, 1]))
            datfile.write(" S= {0:.10f} ".format(0.0))
            datfile.write(" SZ= {0:.10f} \n".format(0.0))
            datfile.write(" STATE 2 ENERGY= {0:.10f} ".format(scan_compare[i, 2]))
            datfile.write(" S= {0:.10f} ".format(1.0))
            datfile.write(" SZ= {0:.10f} \n".format(1.0))
            datfile.write(" STATE 3 ENERGY= {0:.10f} ".format(scan_compare[i, 3]))
            datfile.write(" S= {0:.10f} ".format(0.5))
            datfile.write(" SZ= {0:.10f} \n".format(0.5))
            datfile.write(" STATE 4 ENERGY= {0:.10f} ".format(scan_compare[i, 4]))
            datfile.write(" S= {0:.10f} ".format(3.0))
            datfile.write(" SZ= {0:.10f} \n".format(0.0))
            datfile.write("\n")
            datfile.write(str(i) + "th DUMMY LINE 10\n" + "DUMMY LINE 11\n" + "DUMMY LINE 12\n")
            datfile.write(str(i) + "th DUMMY LINE 13\n" + "DUMMY LINE 14\n" + "DUMMY LINE 15\n")
            datfile.write("\n")

        datfile.close()

        scan = gms._parse_energy_CI_surf('scan.dat')
        for i, line in enumerate(scan):
            for j, s in enumerate(line):
                self.assertAlmostEqual(s, scan_compare[i][j], delta=1.e-7)

        scan = gms._parse_energy_CI_surf('scan.dat', [1.0, 1.0])
        for i, line in enumerate(scan):
            for j, s in enumerate(line):
                if j == 0:
                    self.assertAlmostEqual(s, scan_compare[i][j], delta=1.e-7)
                else:
                    self.assertAlmostEqual(s, scan_compare[i][j+1], delta=1.e-7)

        scan = gms._parse_energy_CI_surf('scan.dat', [3.0, 0.0])
        for i, line in enumerate(scan):
            for j, s in enumerate(line):
                if j == 0:
                    self.assertAlmostEqual(s, scan_compare[i][j], delta=1.e-7)
                else:
                    self.assertAlmostEqual(s, scan_compare[i][j+3], delta=1.e-7)

        scan = gms._parse_energy_CI_surf('scan.dat', [1.0, 0.0])
        for i, line in enumerate(scan):
            for j, s in enumerate(line):
                if j == 0:
                    self.assertAlmostEqual(s, scan_compare[i][j], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__parse_gradient_block(self):
        """
        Tests the correct parsing of a gradient block.
        """

        g_compare = np.random.rand(27).reshape((3,9))
        g = gms.gamess()
        datfile = open(g._log, 'w')
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[0][0]) + " {0:.10f}  ".format(g_compare[0][1]) + " {0:.10f}\n".format(g_compare[0][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[0][3]) + " {0:.10f}  ".format(g_compare[0][4]) + " {0:.10f}\n".format(g_compare[0][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[0][6]) + " {0:.10f}  ".format(g_compare[0][7]) + " {0:.10f}\n".format(g_compare[0][8]))
        datfile.write("\n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[1][0]) + " {0:.10f}  ".format(g_compare[1][1]) + " {0:.10f}\n".format(g_compare[1][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[1][3]) + " {0:.10f}  ".format(g_compare[1][4]) + " {0:.10f}\n".format(g_compare[1][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[1][6]) + " {0:.10f}  ".format(g_compare[1][7]) + " {0:.10f}\n".format(g_compare[1][8]))
        datfile.write("\n")
        datfile.write(" 1 A  {0:.10f}  ".format(g_compare[2][0]) + " {0:.10f}  ".format(g_compare[2][1]) + " {0:.10f}\n".format(g_compare[2][2]))
        datfile.write(" 2 B  {0:.10f}  ".format(g_compare[2][3]) + " {0:.10f}  ".format(g_compare[2][4]) + " {0:.10f}\n".format(g_compare[2][5]))
        datfile.write(" 3 Fi  {0:.10f}  ".format(g_compare[2][6]) + " {0:.10f}  ".format(g_compare[2][7]) + " {0:.10f}\n".format(g_compare[2][8]))
        datfile.write("\n")
        datfile.close()

        datfile = open(g._log, 'r')
        
        grd = gms._parse_gradient_block(datfile)
        for i, gg in enumerate(grd):
            self.assertAlmostEqual(gg, g_compare[0][i], delta=1.e-7)
        grd = gms._parse_gradient_block(datfile)
        for i, gg in enumerate(grd):
            self.assertAlmostEqual(gg, g_compare[1][i], delta=1.e-7)
        grd = gms._parse_gradient_block(datfile)
        for i, gg in enumerate(grd):
            self.assertAlmostEqual(gg, g_compare[2][i], delta=1.e-7)
        datfile.close()

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test_parse_hessian(self):
        """
        Tests the correct parsing of a hessian.
        """

        hess_compare = np.random.rand(9,9)
        datfile = open('hess.dat', 'w')
        datfile.write("DUMMY LINE\n" + " $HESS\n" + "DUMMY LINE\n")

        for i in range(9):
            datfile.write(str(i) + ' 1 ')
            for j in range(5):
                datfile.write("{0:.10E} ".format(hess_compare[i][j]))
            datfile.write('\n' + str(i) + ' 2 ')
            for j in range(5,9):
                datfile.write("{0:.10E} ".format(hess_compare[i][j]))
            datfile.write('\n')

        datfile.write(" $END \n")
        datfile.close()

        hessian = gms.parse_hessian('hess.dat')

        for i, line in enumerate(hessian):
            for j, h in enumerate(line):
                self.assertAlmostEqual(h, hess_compare[i][j], delta=1.e-7)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__format_coordinates(self):
        """
        Tests correct formatting of coordinates.
        """

        atomicSymbols = ['O','H','H']
        atomicNumbers = ['8.0','1.0','1.0']
        x = np.array([0.0,0.0,0.0,
                      0.0,0.690, 0.60,
                      0.0,-0.690,0.60])

        comp_str = " O 8.0    0.0000000    0.0000000    0.0000000\n H 1.0    0.0000000    0.6900000    0.6000000\n H 1.0    0.0000000   -0.6900000    0.6000000"

        string = gms._format_coordinates(x, atomicSymbols, atomicNumbers)

        self.assertEqual(comp_str, string)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__format_efields(self):
        """
        Tests correct formatting for electric fields.
        """

        ef = efi.EField()
        ef.P = np.array([1,0,0],float)
        def a(t): return 0.01
        ef.A = a
        efields = [ef]
        
        comp_str = " $EFIELD\n   EVEC(1)=   0.010000000    0.000000000    0.000000000\n $END\n"

        string = gms._format_efields(efields, 0.0)

        self.assertEqual(comp_str, string)

    @unittest.skipIf(gms.PATHGMS is None, "environment variable CDTK_GAMESSUS is not set.")
    def test__format_wstate(self):
        """
        Tests correct formatting for weights string.
        """

        comp_str = "1.0"
        string = gms._format_wstate(1)

        self.assertEqual(comp_str, string)

        comp_str = "1.0,1.0,1.0,1.0,1.0"
        string = gms._format_wstate(5)

        self.assertEqual(comp_str, string)

        comp_str = "1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0"
        string = gms._format_wstate(15)

        self.assertEqual(comp_str, string)


# Add tests above to the suite definition:
def getSuite():
    suite = unittest.TestSuite()
    suite.addTest(GmsTestCase('test_init_input_options'))
    suite.addTest(GmsTestCase('test_energy_HF'))
    suite.addTest(GmsTestCase('test_energy_gradient_HF'))
    suite.addTest(GmsTestCase('test_MCSCF'))
    suite.addTest(GmsTestCase('test_NACME'))
    suite.addTest(GmsTestCase('test_hessian'))
    suite.addTest(GmsTestCase('test_optimize'))
    suite.addTest(GmsTestCase('test_read_input_template'))
    suite.addTest(GmsTestCase('test_efield'))
    suite.addTest(GmsTestCase('test_dt_efield'))
    suite.addTest(GmsTestCase('test__store_log'))
    suite.addTest(GmsTestCase('test__store_dat'))
    suite.addTest(GmsTestCase('test__store_inp'))
    suite.addTest(GmsTestCase('test__clean'))
    suite.addTest(GmsTestCase('test__parse_energy_HF'))
    suite.addTest(GmsTestCase('test__parse_energy_CI'))
    suite.addTest(GmsTestCase('test__parse_mcscf_energy'))
    suite.addTest(GmsTestCase('test__parse_nacme_energy'))
    suite.addTest(GmsTestCase('test__parse_gradient_HF'))
    suite.addTest(GmsTestCase('test__parse_nacme_gradient'))
    suite.addTest(GmsTestCase('test__parse_nacme_nacme'))
    suite.addTest(GmsTestCase('test__parse_nacme_dipole'))
    suite.addTest(GmsTestCase('test__parse_vec'))
    suite.addTest(GmsTestCase('test__parse_equilibrium_geometry'))
    suite.addTest(GmsTestCase('test__parse_gradient_block'))
    suite.addTest(GmsTestCase('test_parse_hessian'))
    suite.addTest(GmsTestCase('test__parse_energy_CI_surf'))
    suite.addTest(GmsTestCase('test__format_coordinates'))
    suite.addTest(GmsTestCase('test__format_efields'))
    suite.addTest(GmsTestCase('test__format_wstate'))

## TODO not working correctly as orbitals are not correctly printed - needs to be fixed
#    suite.addTest(GmsTestCase('test_energy_HF_HOMO'))
#    suite.addTest(GmsTestCase('test__parse_homo_energy'))
## TODO implement tests. However, they are implicitly tested through a lot of other tests and this not so important
#    suite.addTest(GmsTestCase('test_run'))
#    suite.addTest(GmsTestCase('test__run_gms'))
## TODO howto test sys.exit(-1)
#    suite.addTest(GmsTestCase('test__parse_execution_error'))
## TODO just read template input files. Less important to test and implicitly tested through other tests.
    # suite.addTest(GmsTestCase('test__input_template_HF_E_EF'))
    # suite.addTest(GmsTestCase('test__input_template_HF_E_EF_HOMO'))
    # suite.addTest(GmsTestCase('test__input_template_HF_EGR_EF'))
    # suite.addTest(GmsTestCase('test__input_template_MCSCF'))
    # suite.addTest(GmsTestCase('test__input_template_HESSIAN'))
    # suite.addTest(GmsTestCase('test__input_template_OPTIMIZE'))

    return suite


if __name__ == "__main__":

   unittest.main()

