#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#! /usr/bin/env python

import unittest
import CDTK.Dynamics.IniConSampler as ico
import numpy as np
import random
import sys
import CDTK.Tools.Conversion as con
import math

# Testpotential
def potential(x):
    v = 0.5 * 0.1 * (x - 1.5)**2
    return v

def gaussian(x, mu, sig):
    return 1./(np.sqrt(2.*np.pi)*sig) * np.exp(- ((x - mu) / sig)**2 / 2.0 )


class IniConSamplerTestCase(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skip('skip burnin')
    def test_burnin(self):
        """ 
        Testing the burnin phase.
        """
        # initial position
        xcart = np.array([0.0])
        m = np.array([1.0])

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        # create a sampler        
        sampler = ico.IniConSampler(xcart, potential, M=m, NSAMPLES=0, RefEnergy=0.0, stepnorm=0.1)
        sampler.burnin(steps=5000, targetRate=0.4)

        stepnorm = sampler._optimalStepNorm
        self.assertAlmostEqual(stepnorm, 0.07, delta=0.01)

        # initial position
        xcart = np.array([0.0])
        m = np.array([10.0])

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        # create a sampler        
        sampler = ico.IniConSampler(xcart, potential, M=m, NSAMPLES=0, RefEnergy=0.0, stepnorm=1.0)
        sampler.burnin(steps=5000, targetRate=0.4)

        stepnorm = sampler._optimalStepNorm
        self.assertAlmostEqual(stepnorm, 0.17, delta=0.05)

    def test_sampler_boltzmann(self):
        """ 
        Testing the boltzmann sampling.        
        """
        # initial position
        xcart = np.array([0.2])
        m = np.array([1.0])

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        # create a sampler and burnin
        sampler = ico.IniConSampler(xcart, potential, M=m, NSAMPLES=500000, RefEnergy=0.0, stepnorm=0.2)
        sampler.burnin(steps=5000, targetRate=0.6)

        # get positions and velocities; also energies
        X,V = sampler.sampler_boltzmann()
        energies = sampler._log['energy']

        ## TODO: test energies
        # hist, bin_edges = np.histogram(energies, bins=100, density=True)
        # for i, val in enumerate(hist):
        #     v = (bin_edges[i] + bin_edges[i+1]) / 2.0
        #     f.write(str(v) + ' ' + str(val) + '\n')

        ## test positions to correct distribution
        beta = 1.0 / (300.0 * con.FPC_K_B_EV * con.ev2au)
        hist, bin_edges = np.histogram(X, bins=100, density=True)
#        f = open('positions', 'w')
        for i, val in enumerate(hist):
            v = (bin_edges[i] + bin_edges[i+1]) / 2.0
            # get correct value from gaussian distribution
            comp = gaussian(v, 1.5, 1.0/math.sqrt(beta*0.1))
            if (v > 1.5-2.0/math.sqrt(beta*0.1) and v < 1.5+2.0/math.sqrt(beta*0.1)):
#                print val/comp, v, val, comp, "pos"
                self.assertAlmostEqual(val/comp, 1.0, delta=0.1)
#            f.write(str(v) + ' ' + str(val) + ' ' + str(comp) + '\n')
#        f.close()
        
        ## test velocities to correct distribution
        hist, bin_edges = np.histogram(V, bins=100, density=True)
#        f = open('velocities', 'w')
        for i, val in enumerate(hist):
            v = (bin_edges[i] + bin_edges[i+1]) / 2.0
            # get correct value from gaussian distribution
            comp = gaussian(v, 0.0, 1.0/math.sqrt(beta*m[0]))
            if (v > -2.0/math.sqrt(beta*m[0]) and v < 2.0/math.sqrt(beta*m[0])):
#                print val/comp, v, val, comp, "vel" 
                self.assertAlmostEqual(val/comp, 1.0, delta=0.1)
#            f.write(str(v) + ' ' + str(val) + ' ' + str(comp)  + '\n')
#        f.close()


    def test_rotation(self):
        """ 
        Testing rotational sampling.
        """
        
        phi = [0.0, np.pi/4, np.pi/2, np.pi]

        x0 = np.array([0.0, 0.0, 0.5, 0.0, 0.0, -0.5])
        v0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        x = [x0, x0, x0, x0]
        v = [v0, v0, v0, v0]

        x1 = np.array([-0.5/np.sqrt(2.0), 0.0, 0.5/np.sqrt(2.0), 0.5/np.sqrt(2.0), 0.0, -0.5/np.sqrt(2.0)])
        x2 = np.array([-0.5, 0.0, 0.0, 0.5, 0.0, 0.0])
        x3 = np.array([0.0, 0.0, -0.5, 0.0, 0.0, 0.5])
        x_compare = [x0, x1, x2, x3]
        v_compare = [v0, v0, v0, v0]

        x_rot, v_rot = ico.rotation(x, v, phi)

        for i, xi in enumerate(x_rot):
            for j, xx in enumerate(xi):
                self.assertAlmostEqual(xx, x_compare[i][j], delta=1e-12) 
                self.assertAlmostEqual(v_rot[i][j], v_compare[i][j], delta=1e-12) 

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(IniConSamplerTestCase('test_burnin'))
    suite.addTest(IniConSamplerTestCase('test_sampler_boltzmann'))
    suite.addTest(IniConSamplerTestCase('test_rotation'))

    return suite

if __name__ == "__main__":
    unittest.main()



