#! /usr/bin/env python
#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import unittest
import CDTK.Dynamics.MCElectronDynamics as mced
import CDTK.Tools.EField as efi
import numpy as np
import random
import sys
import CDTK.Tools.Conversion as con
import math
import os

try:
    import pyxmol as XM
    import CDTK.Interfaces.XMoleculeInterface as xmol
except ImportError:
    print("XMOLECULE interface could not be loaded!")
#    self.fail("XMOLECULE interface could not be loaded!")


class MCElectronDynamicsTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_rate(self):
        """ 
        Testing rate calculation.
        """

        EF = efi.EField()
        EF.init_from_input_options({'envelope': ['gaussian-fluence'], 'center': ['40.0', 'fs'], 'fwhm': ['10.0', 'fs'], 'amplitude': ['1.0e14', 'ppmum']})

        ES = mced.ElectronicState(state='20', dte=0.01, ratio=0.1, pe=2000.0)
        ES.QCE = None
        ES.field = [EF]

        a = 3.0
        p = ('a', None, a, None)
        self.assertEqual(ES.rate(p), a)

        a = 0.1
        p = ('f', None, a, None)
        self.assertEqual(ES.rate(p), a)

        a = 3.0
        p = ('p', None, a, None)
        self.assertAlmostEqual(ES.rate(p), 0.0,delta=1.e-12)

        ES.t=40.0*41.362
        a = 0.01
        p = ('p', None, a, None)
        self.assertAlmostEqual(ES.rate(p), a*636.32710679,delta=1.e-8)

    def test_step(self):
        """ 
        Testing the step
        """

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        EF = efi.EField()
        EF.init_from_input_options({'envelope': ['gaussian-fluence'], 'center': ['20.0', 'fs'], 'fwhm': ['5.0', 'fs'], 'amplitude': ['1.0e16', 'ppmum']})

        xm = xmol.xmolecule()
        atoms = ['Ne']
        pos =  np.array([0.000000, 0.000000, 0.000000])
        opts = {
                'pcs': ["yes"],
                'PE': ["12000.000000"],
                'auger': ["yes"],
                'fluorescence': ["yes"],
                'MOM': ["yes"],
                'average_occ_mo': ["3,4,5"],
                'occupation': ["22222"]}
        xm.init_input_options(opts, atoms, pos)
        
        ES = mced.ElectronicState(state='22222', dte=0.01, ratio=0.1, pe=12000.0)
        ES.QCE = xm
        ES.DIR = './'
        ES.field = [EF]

        ES.step(0.01)
        self.assertEqual(ES.t, 0.01)

        ES.step(1999.99)
        ES.write_logs()
        self.assertEqual(ES.t, 2000.0)

        infile = open('charge.log', 'r')
        self.assertEqual(next(infile), "0.01 0\n")
        self.assertEqual(next(infile), "2000.0 10\n")
        infile.close()
        
        infile = open('event.log', 'r')


        cmpLines=["608.97 P [1, 2, 2, 2, 2] 7.32924115706e-06 11146\n",
                  "629.61 P [0, 2, 2, 2, 2] 3.85922607223e-06 11051\n",
                  "671.3 A [1, 1, 2, 1, 2] 0.00251866351604 858\n",
                  "671.78 P [0, 1, 2, 1, 2] 3.96995263416e-06 10963\n",
                  "679.61 A [1, 1, 2, 1, 0] 0.00290966350597 831\n",
                  "682.24 P [0, 1, 2, 1, 0] 4.08888155916e-06 10872\n",
                  "730.769999999 P [0, 0, 2, 1, 0] 2.97742838185e-07 11775\n",
                  "800.339999999 P [0, 0, 1, 1, 0] 1.69815590888e-08 11753\n",
                  "843.409999999 A [1, 0, 0, 0, 0] 0.000913612209348 739\n",
                  "843.609999999 P [0, 0, 0, 0, 0] 4.33015450845e-06 10694\n"]
        
        for cmpLine in cmpLines:
            line=next(infile)
            self.assertAlmostEqual(float(line.split()[0]), float(cmpLine.split()[0]))
            self.assertEqual(line.split()[1], cmpLine.split()[1])
            self.assertEqual(line.split()[2:6], cmpLine.split()[2:6])
            self.assertAlmostEqual(float(line.split()[7]), float(cmpLine.split()[7]), places=6)
            self.assertAlmostEqual(float(line.split()[8]), float(cmpLine.split()[8]))
        infile.close()

    def test_step_stop(self):
        """ 
        Testing the step and stop capabilities after 1 jump
        """


        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        EF = efi.EField()
        EF.init_from_input_options({'envelope': ['gaussian-fluence'], 'center': ['20.0', 'fs'], 'fwhm': ['5.0', 'fs'], 'amplitude': ['1.0e16', 'ppmum']})
        #EF.plot(ti=0,tf=50,tunit='fs')
        xm = xmol.xmolecule()
        atoms = ['Ne']
        pos =  np.array([0.000000, 0.000000, 0.000000])
        opts = {
                'pcs': ["yes"],
                'PE': ["12000.000000"],
                'auger': ["yes"],
                'fluorescence': ["yes"],
                'MOM': ["yes"],
                'average_occ_mo': ["3,4,5"],
                'occupation': ["22222"]}
        xm.init_input_options(opts, atoms, pos)

        ES = mced.ElectronicState(state='22222', dte=0.01, ratio=0.1, pe=12000.0)
        ES.QCE = xm
        ES.DIR = './'
        ES.field = [EF]
        ES.stop = True

        ES.step(0.01)
        self.assertEqual(ES.t, 0.01)

        ES.step(1999.99)
        ES.write_logs()
        self.assertAlmostEqual(ES.t,  608.97, delta=0.001)

        infile = open('charge.log', 'r')
        l = next(infile).split()
        self.assertAlmostEqual(float(l[0]), 0.01)
        self.assertAlmostEqual(int(l[1]), 0)
        l = next(infile).split()
        self.assertAlmostEqual(float(l[0]), 608.97)
        self.assertAlmostEqual(int(l[1]), 1)
        infile.close()

        infile = open('event.log', 'r')
        line=next(infile)
        cmpLine="608.97 P [1, 2, 2, 2, 2] 7.32925904737e-06 11146\n"
        self.assertAlmostEqual(float(line.split()[0]), float(cmpLine.split()[0]))
        self.assertEqual(line.split()[1:6], cmpLine.split()[1:6])
        self.assertAlmostEqual(float(line.split()[7]), float(cmpLine.split()[7]))
        self.assertAlmostEqual(float(line.split()[8]), float(cmpLine.split()[8]))

        infile.close()

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(MCElectronDynamicsTestCase('test_rate'))
    suite.addTest(MCElectronDynamicsTestCase('test_step'))
    suite.addTest(MCElectronDynamicsTestCase('test_step_stop'))

    return suite

if __name__ == "__main__":
    unittest.main()
