#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#! /usr/bin/env python

import unittest
import CDTK.Interfaces.WrapperInterface as win
import numpy as np

class WrapperInterfaceTestCase(unittest.TestCase):
    def setUp(self):
        pass

    ## TODO: tests for most functions;
    ## all these functions essentially just check which interface to use and pass on the 
    ## corresponding results - thus they are less important to test

    def test_f_E(self):
        """ 
        This function tests
        """

    def test_f_EGrad_GS(self):
        """ 
        This function tests
        """

    def test_f_EGrad_ES(self):
        """ 
        This function tests
        """

    def test_f_EGrad_NA(self):
        """ 
        This function tests
        """

    def test_f_NAC(self):
        """ 
        This function tests
        """

    def test_f_W(self):
        """ 
        This function tests
        """

    def test_f_D(self):
        """ 
        This function tests
        """

    def test_f_Dmu(self):
        """ 
        This function tests
        """

    def test__is_old_geometry(self):
        """ 
        This function tests the correct check for old geometry.
        """

        x = np.random.rand(3)
        xc = x.copy()
        wi = win.wrapperInterface()

        # old geometry not set
        self.assertEqual(wi._is_old_geometry(x), False)

        # exactly the same
        wi._oldX = x.copy()
        self.assertEqual(wi._is_old_geometry(x), True)

        # Tiny change -> the same
        x += 1.e-9
        self.assertEqual(wi._is_old_geometry(x), True)
        x -= 1.e-9

        # bigger change -> not the same
        x += 1.e-7
        self.assertEqual(wi._is_old_geometry(x), False)

        
    def test__kineticCouplingsMatrix(self):
        """ 
        This function tests the correct setting and sign for the non adiabatic coupling matrix.
        """
        
        wi = win.wrapperInterface()
        nac = np.array([[[0.0, 1.0, -1.0], [0.2, -0.23, 1.37]], 
                        [[-0.2, 0.23, -1.37], [-0.2, -1.3, 0.25]]])
        v = np.array([1.0, -2.0, 0.1])

        # 0,0: -2-0.1 0,1: -0.2-0.46-0.137 1,1 -0.2+2.6+0.025

        ds_compare = np.array([[2.1j, -0.797j], [0.797j, -2.425j]])
        ds = wi._kineticCouplingsMatrix(nac, v, 0.0, 2)
        for d1,d2 in zip(ds_compare.flatten().tolist(), ds.flatten().tolist()):
            self.assertAlmostEqual(d1,d2)

    def test__kineticCouplingsMatrixMu(self):
        """ 
        This function tests the correct setting and sign for the non adiabatic coupling matrix
        due to dipole coupling.
        """
        
        import CDTK.Interfaces.GamessUSInterface as gin
        import CDTK.Tools.EField as efi
        if gin.PATHGMS is None:
            self.skipTest("environment variable CDTK_GAMESSUS is not set.")

        QCE = gin.gamess()
        EF = efi.EField()
        EF.init_from_input_options({'fwhm': ['1','fs'], 'amplitude': ['100', 'au']})
        QCE.efields = [EF]

        wi = win.wrapperInterface()
        e = [0.0, 1.0]
        wi._oldE = e
        wi.QCE = QCE
        
        mu = np.array([[[0.0, 1.0, -1.0], [0.2, -0.23, 1.37]], 
                        [[-0.2, -1.3, 0.25], [0.1, 0.0, 0.0]]])
        v = np.array([1.0, -2.0, 0.1])

        ds_compare = [0.0j, 1.14615395j, -0.20915219j, 0.0j]
        ds = wi._kineticCouplingsMatrixMu(mu, v, 41.36, 2)
        
        for i, dij in enumerate(ds.flatten()):
            self.assertAlmostEqual(ds_compare[i], dij, delta=1e-6)

        
def getSuite():
    suite = unittest.TestSuite()

#    suite.addTest(WrapperInterfaceTestCase('test_f_E'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_EGrad_Gs'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_EGrad_ES'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_EGrad_NA'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_NAC'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_W'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_D'))
#    suite.addTest(WrapperInterfaceTestCase('test_f_Dmu'))
    suite.addTest(WrapperInterfaceTestCase('test__is_old_geometry'))
    suite.addTest(WrapperInterfaceTestCase('test__kineticCouplingsMatrix'))
    suite.addTest(WrapperInterfaceTestCase('test__kineticCouplingsMatrixMu'))

    return suite

if __name__ == "__main__":
    unittest.main()
