#! /usr/bin/env python
#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

import unittest
import numpy as np
import os
from CDTK.Tools.OutputGrabber import suppress_stdout, suppress_stdoutstderr

try:
    import pyxmol as XM
    import CDTK.Interfaces.XMoleculeInterface as xmol
except ImportError:
#    unittest.TestCase().fail("XMOLECULE interface could not be loaded!")
    print("XMOLECULE interface could not be loaded!")
#    return

class XmoleculeTestCase(unittest.TestCase):

    def setUp(self):
        # todo: Create optional parameter to redirect outputs of XMOLECULE
        self.verbose = False
        
        return

    def tearDown(self):
        os.remove("input.in")
        
        return

    def test_init_input_option(self):
        """ 
        This function test the correct creation of an Xmolecule input file!
        """
        os.system("rm O.nao*")
        os.system("rm H.nao*")
        os.system("rm O.tdipole*")
        os.system("rm O.auger*")

        with suppress_stdout():
            xm = xmol.xmolecule()
            atoms = ['O', 'H', 'H']
            pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
            opts = {
                    'pcs': ["yes"],
                    'PE': ["10000.000000"],
                    'auger': ["yes"],
                    'fluorescence': ["yes"],
                    'MOM': ["yes"],
                    'occupation': ["12222"]}
            xm.init_input_options(opts, atoms, pos)
        
        inputFileLines = [
                          "pcs=yes \n",
                          "PE=10000.000000 \n",
                          "auger=yes \n",
                          "fluorescence=yes \n",
                          "MOM=yes \n",
                          "occupation=12222 \n"
                      ]
        # check written input file
        infile = open('input.in', 'r')
        self.assertEqual(next(infile), "O    -0.0000000000   -0.0000000000   +0.1167430000 \n")
        self.assertEqual(next(infile), "H    -0.0000000000   -1.4981030000   +1.2057930000 \n")
        self.assertEqual(next(infile), "H    -0.0000000000   +1.4981030000   +1.2057930000 \n")

        while True:
            l = next(infile, '')
            if l == '':
                break
            self.assertTrue(l in inputFileLines,  "{:s} incompatible with expected inputFile {:s}".format(l,inputFileLines[0]))
            
#    def test_energy_HF(self):
#        """ 
#        This function tests the energy calculation in the XMolecule interface for:
#        - H atom 
#        - H2 at minimum
#        - H2O and core-ionized H2O, close to the H2O minimum
#        """
#
#
#        opts = {
#                'pcs': ["yes"],
#                'PE': ["10000.000000"],
#                'auger': ["yes"],
#                'fluorescence': ["yes"],
#                'MOM': ["yes"],
#                'lmax_ang': ['10'],
#                'N': ['50']}
#
## initialize for H
#        xm = xmol.xmolecule()
#        atoms = ['H']
#        num = ['1.0']
#        pos =  np.array([0.0, 0.0, 0.0])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e = xm.energy_HF(pos)
#        self.assertAlmostEqual(e,-0.5061549234803432,delta=1.e-7)
#
## initialize for H2
#        xm = xmol.xmolecule()
#        atoms = ['H', 'H']
#        pos =  np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.574000])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e = xm.energy_HF(pos)
#        self.assertAlmostEqual(e,-1.2603325834495558,delta=1.e-7)
#
## initialize for H2O
#        xm = xmol.xmolecule()
#        atoms = ['O', 'H', 'H']
#        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e = xm.energy_HF(pos)
#        self.assertAlmostEqual(e,-79.10049441161175,delta=1.e-7)
#
## initialize for H2O core-ionized
#        xm = xmol.xmolecule()
#        atoms = ['O', 'H', 'H']
#        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
#
#
#        opts = {
#                'pcs': ["yes"],
#                'PE': ["10000.000000"],
#                'auger': ["yes"],
#                'fluorescence': ["yes"],
#                'MOM': ["yes"],
#                'occupation': ["12222"],
#                'lmax_ang': ['10'],
#                'N': ['50']}
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e = xm.energy_HF(pos)
#        self.assertAlmostEqual(e,-58.323909071517804,delta=1.e-7)
#
#    def test_energy_gradient_HF(self):
#        """ 
#        This function tests the energy and gradient calculation in the XMolecule interface for:
#        - H atom 
#        - H2 at minimum
#        - H2O and core-ionized H2O, close to the H2O minimum
#        """
#
#
#        opts = {
#                'pcs': ["yes"],
#                'PE': ["10000.000000"],
#                'auger': ["yes"],
#                'fluorescence': ["yes"],
#                'MOM': ["yes"],
#                'lmax_ang': ['10'],
#                'N': ['50']}
#
## initialize for H
#        xm = xmol.xmolecule()
#        atoms = ['H']
#        num = ['1.0']
#        pos =  np.array([0.0, 0.0, 0.0])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e, g = xm.energy_gradient_HF(pos)
#        self.assertAlmostEqual(e,-0.5061549234803432,delta=1.e-7)
#        self.assertAlmostEqual(g[0],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[1],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[2],0.0,delta=1.e-7)
#
## initialize for H2
#        xm = xmol.xmolecule()
#        atoms = ['H', 'H']
#        pos =  np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.574000])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e, g = xm.energy_gradient_HF(pos)
#        self.assertAlmostEqual(e,-1.2603325834495558,delta=1.e-7)
#        self.assertAlmostEqual(g[0],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[1],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[2],0.00094184433048566385,delta=1.e-7)
#        self.assertAlmostEqual(g[3],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[4],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[5],-0.00094184433048566385,delta=1.e-7)
#
## initialize for H2O
#        xm = xmol.xmolecule()
#        atoms = ['O', 'H', 'H']
#        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e, g = xm.energy_gradient_HF(pos)
#        self.assertAlmostEqual(e,-79.10049441161175,delta=1.e-7)
#        self.assertAlmostEqual(g[0],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[1],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[2],-1.50122929e-02,delta=1.e-7)
#        self.assertAlmostEqual(g[3],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[4],2.24910866e-03,delta=1.e-7)
#        self.assertAlmostEqual(g[5],-1.68076371e-03,delta=1.e-7)
#        self.assertAlmostEqual(g[6],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[7],-2.24910866e-03,delta=1.e-7)
#        self.assertAlmostEqual(g[8],-1.68076371e-03,delta=1.e-7)
#
## initialize for H2O core-ionized
#        xm = xmol.xmolecule()
#        atoms = ['O', 'H', 'H']
#        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
#
#
#        opts = {
#                'pcs': ["yes"],
#                'PE': ["10000.000000"],
#                'auger': ["yes"],
#                'fluorescence': ["yes"],
#                'MOM': ["yes"],
#                'occupation': ["12222"],
#                'lmax_ang': ['10'],
#                'N': ['50']}
#
#        xm.init_input_options(opts, atoms, pos)
#
#        e, g = xm.energy_gradient_HF(pos)
#        self.assertAlmostEqual(e,-58.323909071517804,delta=1.e-7)
#        self.assertAlmostEqual(g[0],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[1],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[2],-7.61638414e-02,delta=1.e-7)
#        self.assertAlmostEqual(g[3],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[4],1.00869326e-02,delta=1.e-7)
#        self.assertAlmostEqual(g[5],2.61335071e-02,delta=1.e-7)
#        self.assertAlmostEqual(g[6],0.0,delta=1.e-7)
#        self.assertAlmostEqual(g[7],-1.00869326e-02,delta=1.e-7)
#        self.assertAlmostEqual(g[8],2.61335071e-02,delta=1.e-7)


    def test_hole_properties(self):
        """ 
        This function tests the energy, gradient and NAC calculation employing Koopmans' theorem 
        in the XMolecule interface for:
        - H2 at minimum
        - H2O close to the H2O minimum
        """


        opts = {
                'MOM': ["no"],
                'HF': ["yes"],
                'SCF_energy_conv': ["1e-8"],
                'SCF_density_conv': ["1e-8"],
                'gs_occ': ["yes"],
                'gto ': ["STO-3G"],
                'occupation': ["22221"], 
                'eps_teint': ["0.0"],
                'verbose': ["yes"]
            }
                
#               '': [""],
#               '': [""],
#               '': [""],                
#               'lmax_ang': ['10'],
#               'N': ['50']}

# initialize for H2O
        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        xm.init_input_options(opts, atoms, pos)

        occ, e, g, nac = xm.hole_properties(pos)

        occ_comp = [[2, 2, 2, 2, 1, 0, 0], [2, 2, 2, 1, 2, 0, 0], [2, 2, 1, 2, 2, 0, 0], [2, 1, 2, 2, 2, 0, 0]]
        e_comp = [-74.57607305, -74.52131085, -74.35200887, -73.71021335]
        g_comp = [[-3.37070761e-16, -5.47264908e-17, 1.36155857e-02, 1.67981907e-16, 1.99780366e-02, -6.80779286e-03, 1.69088854e-16, -1.99780366e-02, -6.80779286e-03], [1.88860053e-16, 1.05460798e-15, -4.64054994e-02, -1.94324831e-17, 6.32205319e-02, 2.32027497e-02, -1.69427570e-16, -6.32205319e-02, 2.32027497e-02], [1.53225034e-18, 9.83273658e-17, 2.73401823e-01, 1.53904830e-16, 7.91391089e-02, -1.36700911e-01, -1.55437080e-16, -7.91391089e-02, -1.36700911e-01], [-3.24277764e-17, -4.50373250e-16, 1.25845760e-01, 9.87168047e-17, 1.16075385e-01, -6.29228800e-02, -6.62890283e-17, -1.16075385e-01, -6.29228800e-02]]
        nac_comp = {(0, 1): np.array([7.77406876e-01, -1.68053394e-15, 2.24851327e-15, -3.88703438e-01, -1.92600273e-15, -1.21087248e-15, -3.88703438e-01, 3.60653667e-15, -1.03764079e-15]), (1, 2): np.array([-2.08755843e-16, -7.34975165e-01, -4.91780523e-15, -7.00613005e-16, 3.67487583e-01, -6.11635685e-02, 9.09368848e-16, 3.67487583e-01, 6.11635685e-02]), (3, 2): np.array([1.43507258e-16, 4.55859348e-01, 3.56931248e-16, -7.19395726e-17, -2.27929674e-01, 2.02061590e-01, -7.15676853e-17, -2.27929674e-01, -2.02061590e-01]), (0, 0): np.array([0., 0., 0., 0., 0., 0., 0., 0., 0.]), (3, 3): np.array([0., 0., 0., 0., 0., 0., 0., 0., 0.]), (3, 0): np.array([-1.41457455e-01, 2.24492651e-16, 4.47441042e-16, 7.07287273e-02, 3.50346319e-16, -1.18930819e-16, 7.07287273e-02, -5.74838971e-16, -3.28510223e-16]), (3, 1): np.array([-3.93035102e-16, -7.80378972e-16, -1.10326000e-01, 1.89856841e-16, -1.49568716e-01, 5.51630002e-02, 2.03178261e-16, 1.49568716e-01, 5.51630002e-02]), (2, 1): np.array([2.08755843e-16, 7.34975165e-01, 4.91780523e-15, 7.00613005e-16, -3.67487583e-01, 6.11635685e-02, -9.09368848e-16, -3.67487583e-01, -6.11635685e-02]), (0, 2): np.array([-1.04708229e-16, 2.30199784e-15, 8.02135597e-16, -2.46800375e-01, -9.64023772e-16, 1.75110583e-16, 2.46800375e-01, -1.33797407e-15, -9.77246180e-16]), (2, 0): np.array([1.04708229e-16, -2.30199784e-15, -8.02135597e-16, 2.46800375e-01, 9.64023772e-16, -1.75110583e-16, -2.46800375e-01, 1.33797407e-15, 9.77246180e-16]), (1, 3): np.array([3.93035102e-16, 7.80378972e-16, 1.10326000e-01, -1.89856841e-16, 1.49568716e-01, -5.51630002e-02, -2.03178261e-16, -1.49568716e-01, -5.51630002e-02]), (2, 3): np.array([-1.43507258e-16, -4.55859348e-01, -3.56931248e-16, 7.19395726e-17, 2.27929674e-01, -2.02061590e-01, 7.15676853e-17, 2.27929674e-01, 2.02061590e-01]), (2, 2): np.array([0., 0., 0., 0., 0., 0., 0., 0., 0.]), (1, 0): np.array([-7.77406876e-01, 1.68053394e-15, -2.24851327e-15, 3.88703438e-01, 1.92600273e-15, 1.21087248e-15, 3.88703438e-01, -3.60653667e-15, 1.03764079e-15]), (0, 3): np.array([1.41457455e-01, -2.24492651e-16, -4.47441042e-16, -7.07287273e-02, -3.50346319e-16, 1.18930819e-16, -7.07287273e-02, 5.74838971e-16, 3.28510223e-16]), (1, 1): np.array([0., 0., 0., 0., 0., 0., 0., 0., 0.])}

        for i, oi in enumerate(occ_comp):
            for j, oij in enumerate(oi):
                self.assertEqual(oij, occ[i][j])

        for i, ei in enumerate(e_comp):
            self.assertAlmostEqual(ei, e[i], delta=1.e-5)                  

        for i, gi in enumerate(g_comp):
            for j, gij in enumerate(gi):
                self.assertAlmostEqual(gij, g[i,j], delta=1.e-5)

        for ij in nac_comp.keys():
            for k, nacijk in enumerate(nac_comp[ij]):
                self.assertAlmostEqual(nacijk, nac[ij][k], delta=1.e-5)

    def test_absorptionCS(self):
        """ 
        This function tests the implementation of absorption cross sections
        """
        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
#        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        pos = np.array([ 0.0, 0.0, 0.0, 0.0, -1.4309301866, 1.1083919440, 0.0, 1.4309301866, 1.1083919440])
        opts = {
                'pcs': ["yes"],
                'PE': ["10000.000000"],
                'auger': ["yes"],
                'fluorescence': ["yes"],
                'MOM': ["yes"],
                'gto ': ["STO-3G.gbs"],
                'occupation': ["22221"]}
        xm.init_input_options(opts, atoms, pos)

        print('Getting absorption cross-sections')

        aCS, _ = xm.getAbsorptionCrossSections()

        check_aCS = np.zeros([7, 7], dtype='float')
        # fill in values from test_absorption_interface
        check_aCS[0,4] = 6.98146E-03
        check_aCS[0,5] = 4.54228E-03
        check_aCS[0,6] = 8.48164E-03
        check_aCS[1,4] = 2.65598E-02
        check_aCS[1,5] = 5.25475E-03
        check_aCS[1,6] = 9.05025E-03
        check_aCS[2,5] = 8.16368E-02
        check_aCS[2,6] = 5.42683E-02
        check_aCS[3,4] = 9.93594E-05
        check_aCS[3,5] = 1.79386E-02
        check_aCS[3,6] = 3.77493E-02
        check_aCS[4,5] = 4.56789E-05
    
        for i in range(aCS.shape[0]):
            for j in range(aCS.shape[1]):
                #if np.any(aCS[i,j]!=0 or check_aCS[i,j]!=0):
                #    print '{0:d}\t{1:d}\t{2:.5e}\t{3:.5e}'.format(i,j,aCS[i,j],check_aCS[i,j])
                self.assertAlmostEqual( aCS[i,j], check_aCS[i,j], delta=1.e-4)

    def test_dipole_moments(self):
        """ 
        This function tests the implementation of transition dipole moments
        by using symmetry considerations to see which transitions are foribidden,
        they have a 0 transition dipole moment, and which are allowed.
        """


        print("Start XMOLECULE calculation")
        with suppress_stdout():
            xm = xmol.xmolecule()
            atoms = ['O', 'H', 'H']
            pos = np.array([ 0.0, 0.0, 0.0, 0.0, -1.4309301866, 1.1083919440, 0.0, 1.4309301866, 1.1083919440])
            opts = {
                    'pcs': ["yes"],
                    'PE': ["10.000000"],
                    'auger': ["yes"],
                    'fluorescence': ["yes"],
                    'MOM': ["yes"],
                    'gto ': ["STO-3G"],
                    'occupation': ["22222"]}
            xm.init_input_options(opts, atoms, pos)

        print('Getting transition dipole moments')
        _, tdip = xm.getAbsorptionCrossSections()


        def func_assert_tdip(ind_i, ind_j, pol_a, pol_f):
            for p in pol_a: # Assert weather dipole is 0
                self.assertAlmostEqual(tdip[ind_i, ind_j, p], 0.0, delta=1.e-12)
                self.assertAlmostEqual(tdip[ind_j, ind_i, p], 0.0, delta=1.e-12)
            for p in pol_f: # Assert weather dipole is not 0
                self.assertNotAlmostEqual(tdip[ind_i, ind_j, p], 0.0, delta=1.e-5)
                self.assertNotAlmostEqual(tdip[ind_j, ind_i, p], 0.0, delta=1.e-5)

        irr_rep = ["A1", "A1", "B2", "A1", "B1", "A1", "B2"]
        for i in range(tdip.shape[0]):
            for j in range(i, tdip.shape[1]):
                if irr_rep[i] == "A1" and irr_rep[j] == "A1":
                    # Allowed transition only with z (A1)
                    func_assert_tdip(i, j, [0, 1], [2])
                elif irr_rep[i] == "A1" and irr_rep[j] == "B1" or irr_rep[i] == "B1" and irr_rep[j] == "A1":
                    # Allowed transition only with x (B1)
                    func_assert_tdip(i, j, [1, 2], [0])
                elif irr_rep[i] == "A1" and irr_rep[j] == "B2" or irr_rep[i] == "B2" and irr_rep[j] == "A1":
                    # Allowed transition only with y (B2)
                    func_assert_tdip(i, j, [0, 2], [1])
                elif irr_rep[i] == "B1" and irr_rep[j] == "B1":
                    # ! Should be allowed transition with z (A1), virtually 0 because
                    #   B1 is basically an atomic 2p orbital from oxygen, so B1-B1
                    #   would be a p-p transition (forbidden in atoms)
                    func_assert_tdip(i, j, [0, 1, 2], [])
                elif irr_rep[i] == "B1" and irr_rep[j] == "B2" or irr_rep[i] == "B2" and irr_rep[j] == "B1":
                    # Forbidden transition
                    func_assert_tdip(i, j, [0, 1, 2], [])
                elif irr_rep[i] == "B2" and irr_rep[j] == "B2":
                    # Allowed transition only with z (A1)
                    func_assert_tdip(i, j, [0, 1], [2])


    def test_dipole_moments(self):
        """ 
        This function tests the implementation of transition dipole moments
        by using symmetry considerations to see which transitions are foribidden,
        they have a 0 transition dipole moment, and which are allowed.
        """


        print("Start XMOLECULE calculation")
        with suppress_stdout():
            xm = xmol.xmolecule()
            atoms = ['O', 'H', 'H']
            pos = np.array([ 0.0, 0.0, 0.0, 0.0, -1.4309301866, 1.1083919440, 0.0, 1.4309301866, 1.1083919440])
            opts = {
                    'pcs': ["yes"],
                    'PE': ["10.000000"],
                    'auger': ["yes"],
                    'fluorescence': ["yes"],
                    'MOM': ["yes"],
                    'sym': 'yes',
                    'gto ': ["STO-3G"],
                    'occupation': ["22222"]}
            xm.init_input_options(opts, atoms, pos)
        print('Getting transition dipole moments')
        _, tdip = xm.getAbsorptionCrossSections()

        def func_assert_tdip(ind_i, ind_j, pol_a, pol_f):
            for p in pol_a:
                self.assertAlmostEqual(tdip[ind_i, ind_j, p], 0.0, delta=1.e-12)
                self.assertAlmostEqual(tdip[ind_j, ind_i, p], 0.0, delta=1.e-12)
            for p in pol_f:
                self.assertNotAlmostEqual(tdip[ind_i, ind_j, p], 0.0, delta=1.e-5)
                self.assertNotAlmostEqual(tdip[ind_j, ind_i, p], 0.0, delta=1.e-5)

        irr_rep = ["A1", "A1", "B2", "A1", "B1", "A1", "B2"]
        for i in range(tdip.shape[0]):
            for j in range(i, tdip.shape[1]):
                if irr_rep[i] == "A1" and irr_rep[j] == "A1":
                    # Allowed transition only with z (A1)
                    func_assert_tdip(i, j, [0, 1], [2])
                elif irr_rep[i] == "A1" and irr_rep[j] == "B1" or irr_rep[i] == "B1" and irr_rep[j] == "A1":
                    # Allowed transition only with x (B1)
                    func_assert_tdip(i, j, [1, 2], [0])
                elif irr_rep[i] == "A1" and irr_rep[j] == "B2" or irr_rep[i] == "B2" and irr_rep[j] == "A1":
                    # Allowed transition only with y (B2)
                    func_assert_tdip(i, j, [0, 2], [1])
                elif irr_rep[i] == "B1" and irr_rep[j] == "B1":
                    # ? Allowed transition only with z (A1)
                    # func_assert_tdip(i, j, [0, 1], [2])
                    # func_assert_tdip(i, j, [0, 1, 2], [])
                    continue
                elif irr_rep[i] == "B2" and irr_rep[j] == "B2":
                    # Allowed transition only with z (A1)
                    func_assert_tdip(i, j, [0, 1], [2])
                elif (irr_rep[i] == "B1" and irr_rep[j] == "B2") or (irr_rep[i] == "B2" and irr_rep[j] == "B1"):
                    # Forbidden transition
                    func_assert_tdip(i, j, [0, 1, 2], [])


    def test_fluorescence_rates(self):
        """ 
        This function tests the implementation of fluorescence rates
        """



        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos = np.array([ 0.0, 0.0, 0.0, 0.0, -1.4309301866, 1.1083919440, 0.0, 1.4309301866, 1.1083919440])
        opts = {
                'pcs': ["yes"],
                'PE': ["10000.000000"],
                'auger': ["yes"],
                'fluorescence': ["yes"],
                'MOM': ["yes"],
                'gto ': ["STO-3G.gbs"],
                'occupation': ["12222"]}
        xm.init_input_options(opts, atoms, pos)

        fluo_rates = np.asarray( xm.getFluorescenceRates() )

        tmp = np.zeros([7,7])
        tmp[0,1] = 1.0 # somehow the test fails if this line is not present
        tmp[0,1] = 1.17906e-6
        tmp[0,2] = 1.11895E-05
        tmp[0,3] = 1.16002E-05
        tmp[0,4] = 1.41810E-05
#
        for i in range(fluo_rates.shape[0]):
            for j in range(fluo_rates.shape[1]):
                self.assertAlmostEqual( fluo_rates[i,j], tmp[i,j], delta=1.e-8)

    def test_init_input_embedding(self):
        """
        This function tests the input file with embedding

        """
        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")


        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        x_pos = np.array([0.0, 0.0, 0.0, 1.0, 1.0, 1.0, 2.0, 2.0, -2.0])
        x_charges = np.array([-0.82, +0.41, 1.0])
        opts = {'occupation': ["22222"],
                'gto': ["6-31G"],
                'HF' : ["yes"],
                }
        xm.init_input_options(opts, atoms, pos,x_pos=x_pos,x_charges=x_charges)

        inputFileLines = [
                          "occupation=22222 \n",
                          "gto=6-31G \n",
                          "HF=yes \n"]

# check written input file
        infile = open('input.in', 'r')
        self.assertEqual(next(infile), "O    -0.0000000000   -0.0000000000   +0.1167430000 \n")
        self.assertEqual(next(infile), "H    -0.0000000000   -1.4981030000   +1.2057930000 \n")
        self.assertEqual(next(infile), "H    -0.0000000000   +1.4981030000   +1.2057930000 \n")
        self.assertEqual(next(infile), "X    +0.0000000000   +0.0000000000   +0.0000000000    -0.82000000 \n")
        self.assertEqual(next(infile), "X    +1.0000000000   +1.0000000000   +1.0000000000    +0.41000000 \n")
        self.assertEqual(next(infile), "X    +2.0000000000   +2.0000000000   -2.0000000000    +1.00000000 \n")
        print(inputFileLines)

        while True:
            l = next(infile, '')
            print(l)
            if l == '':
                break
            self.assertTrue(l in inputFileLines,  "{:s} incompatible with expected inputFile".format(l))
        infile.close()

    def test_energy_gradient_embedding(self):
        """
        This function tests the gradient and energy functions with embedding
        """

        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")

        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        x_pos = np.array([0.0, 0.0, 0.5])
        x_charges = np.array([-1.0])
        opts = {'occupation': ["22222"],
                'gto': ["6-31G"],
                'HF' : ["yes"]
                }
        xm.init_input_options(opts, atoms, pos,x_pos=x_pos,x_charges=x_charges)

        e,g = xm.energy_gradient_HF( np.append(pos,x_pos) )

        g1 = np.array([-2.06809020e-04,-1.66584400e-05,-4.00104637e+01,-2.77946413e-05,5.09622876e-02,-7.28605812e-02,-4.61895624e-05,-5.09622835e-02,-7.28592708e-02, 7.44955031e-06,-7.88905652e-06, 4.01562253e+01])
        # Compare with pre computed values
        # Numerical Gradients: (E[x+dr] - E[x-dr]) / 2 / dr
        # dr = 0.001 [a.u.]
        self.assertAlmostEqual(e, -86.0564490941,delta=1.e-7)
        for i in range( len(g) ):
            self.assertAlmostEqual(g[i], g1[i], delta=1.e-1)

    def test_hole_properties_embedding(self):
        """
        This function tests Koopmans hole properties (gradient, NACMEs) with embedding
        """

        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")

        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        x_pos = np.array([0.0, 0.0, 0.0])
        x_charges = np.array([-1.0])
        opts = {'occupation': ["22221"],
                'gto': ["6-31G"],
                'gs_occ' : ["yes"],
                'HF' : ["yes"]
                }
        xm.init_input_options(opts, atoms, pos,x_pos=x_pos,x_charges=x_charges)

        occ,e,g,nac = xm.hole_properties( np.append(pos,x_pos) )

#        # TODO compare with pre computed values
#        self.assertAlmostEqual(e, None)
#        self.assertAlmostEqual(g, None)
#        self.assertAlmostEqual(nac, None)

    def test_bond_order(self):
        """
        This function tests the bond order returned via the interface
        """

        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")

        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        x_pos = np.array([0.0, 0.0, 0.0])
        opts = {'occupation': ["22222"],
                'gto': ["6-31G"],
                'gs_occ' : ["yes"],
                'HF' : ["yes"]
                }
        xm.init_input_options(opts, atoms, pos)

        bo = xm.getBondOrders()
        bo = np.asarray(bo)

        bo2 = np.array([[16.01902870280308, 0.7963300032595401, 0.7963300032595261], [0.7963300032595401, 0.3928425448001332, 0.004983097279268163], [0.7963300032595261, 0.004983097279268163, 0.39284254480011577]])
        for i in range(bo.shape[0]):
            for j in range(bo.shape[1]):
                self.assertAlmostEqual(bo[i,j], bo2[i,j])

    def test_population_analysis(self):
        """
        This function tests the population analysis returned via the interface
        """

        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")

        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([0.0000000000, 0.0000000000, 0.0000000000, -1.4309301866, 1.1083919440, 0.0000000000, 1.4309301866, 1.1083919440, 0.0000000000])
        opts = {'occupation': ["22222"],
                'gto': ["6-31G"],
                'gs_occ' : ["yes"],
                'HF' : ["yes"],
                'SCF_energy_conv' : ["1.0e-8"],
                'SCF_density_conv' : ["1.0e-8"],
                }
        xm.init_input_options(opts, atoms, pos)

        pop = xm.getPopulation()

        popRef = np.array([[ 9.99696500e-01,  1.51749924e-04,  1.51749924e-04],
                           [ 8.71670672e-01,  6.41646641e-02,  6.41646641e-02],
                           [ 6.31250288e-01,  1.84374856e-01,  1.84374856e-01],
                           [ 8.93285914e-01,  5.33570428e-02,  5.33570428e-02],
                           [ 1.00000000e+00,  1.68182689e-30,  9.19504747e-31],
                           [-1.23120977e-01,  5.61560489e-01,  5.61560489e-01],
                           [ 5.60233735e-02,  4.71988313e-01,  4.71988313e-01],
                           [ 2.26066514e-01,  3.86966743e-01,  3.86966743e-01],
                           [ 1.00000000e+00, -2.02098716e-31,  1.22179732e-29],
                           [ 3.32541475e-01,  3.33729263e-01,  3.33729263e-01],
                           [ 8.38786904e-01,  8.06065478e-02,  8.06065478e-02],
                           [ 1.08665982e+00, -4.33299123e-02, -4.33299123e-02],
                           [ 1.18713951e+00, -9.35697560e-02, -9.35697560e-02]])
        np.testing.assert_almost_equal(pop, popRef, decimal=5)

    def test_geom_getter_setter(self):
        """
        This function tests setting / getting the geometry via the interface
        """


        PATHXMOLECULEBASISDIR = os.environ.get('XMOLECULE_BASIS_DIR',None)
        if PATHXMOLECULEBASISDIR == None:
            self.fail("XMOLECULE gto basis path not set. Please set environment variablei XMOLECULE_BASIS_DIR")

        xm = xmol.xmolecule()
        atoms = ['O', 'H', 'H']
        pos =  np.array([-0.000000, -0.000000, 0.116743, -0.000000, -1.498103, 1.205793, -0.000000, 1.498103, 1.205793])
        pos2 = np.array([0.0, 0.0, 0.116743, 0.0, -1.7, 1.1, 0.0, 1.7, 1.1])
        x_pos = np.array([0.0, 0.0, 0.0])
        opts = {'occupation': ["22222"],
                'gto': ["6-31G"],
                'gs_occ' : ["yes"],
                'HF' : ["yes"]
                }
        xm.init_input_options(opts, atoms, pos)

        xm.set_geom(pos2)
        
        pos3 = xm.get_geom()
        for i in range(pos2.shape[0]):
            self.assertAlmostEqual(pos2[i],pos3[i])





def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(XmoleculeTestCase('test_init_input_option'))
#    suite.addTest(XmoleculeTestCase('test_energy_HF'))
#    suite.addTest(XmoleculeTestCase('test_energy_gradient_HF'))
    suite.addTest(XmoleculeTestCase('test_hole_properties'))
    suite.addTest(XmoleculeTestCase('test_absorptionCS'))
    suite.addTest(XmoleculeTestCase('test_dipole_moments'))
    suite.addTest(XmoleculeTestCase('test_fluorescence_rates'))
    suite.addTest(XmoleculeTestCase('test_init_input_embedding'))
    suite.addTest(XmoleculeTestCase('test_energy_gradient_embedding'))
    suite.addTest(XmoleculeTestCase('test_hole_properties_embedding'))
    suite.addTest(XmoleculeTestCase('test_bond_order'))
    suite.addTest(XmoleculeTestCase('test_population_analysis'))

    return suite

if __name__ == "__main__":
    unittest.main()
