#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#! /usr/bin/env python

import unittest
import numpy as np
import random
import sys
import math
import CDTK.analyze as analyze

import CDTK.Dynamics.Trajectory as tra
import CDTK.Dynamics.MCElectronDynamics as mced

class analyzeTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_red(self):
        """ 
        Testing the reduction routine for calculating the correlation.
        """

        redString = 'dot'
        c0 = [0.0, 1.0, -1.0]
        c1 = [2.0, 3.0, -1.0]
        comp = 4.0
        
        red = analyze.red(c0, c1, redString)

        self.assertAlmostEqual(comp, red, delta=1e-10)

        redString = 'mult'
        c0 = np.array(c0)
        c1 = np.array(c1)
        comp = [0.0, 3.0, 1.0]
        
        red = analyze.red(c0, c1, redString)

        for i, r in enumerate(red):
            self.assertAlmostEqual(comp[i], r, delta=1e-10)
            
        c0 = 3.0
        c1 = 1.5
        comp = 4.5

        red = analyze.red(c0, c1, redString)

        self.assertAlmostEqual(comp, red, delta=1e-10)

        return

    def test_doParse(self):
        """
        Tests the correct parsing of an command file.
        """

        commands_compare = {0: ("1; +id", "1; +pos -1 0 -2 m,1,2", "0.0", "mult", "0", "mean", "muh"),
                        1: ("1; +pos -1 0 -2 1", "1; +pos -1 0 -2 1", "0.0", "mult", None, "std", "mieh"),
                        2: ("1; +id", "2; +pos -1 0 -2 m,1,2,3 -p >,10.0  +cha -p -n 0", "None", "mult", "0", "mean", "maeh")}

        ofi = open('test.inp', 'w')
        ofi.write("-A '" + commands_compare[0][0] + "' \n")
        ofi.write("-B '" + commands_compare[0][1] + "' \n")
        ofi.write("-r '" + commands_compare[0][3] + "' \n")
        ofi.write("-s '" + commands_compare[0][4] + "' \n")
        ofi.write("-o '" + commands_compare[0][6] + "' \n")              
        ofi.close()

        commands = analyze.doParse('test.inp', './', analyze.getParser())
        for k in commands:
            for i in range(len(commands[k])):
                if i == 6:
                    self.assertEqual(commands[k][i], "./" + commands_compare[k][i])
                else:
                    self.assertEqual(commands[k][i], commands_compare[k][i])

                ofi = open('test.inp', 'w')
        ofi.write("-A '" + commands_compare[0][0] + "' \n")
        ofi.write("-B '" + commands_compare[0][1] + "' \n")
        ofi.write("-r '" + commands_compare[0][3] + "' \n")
        ofi.write("-s '" + commands_compare[0][4] + "' \n")
        ofi.write("-o '" + commands_compare[0][6] + "' \n")
        ofi.write("### \n")
        ofi.write("-A '" + commands_compare[1][0] + "' \n")
        ofi.write("-B '" + commands_compare[1][1] + "' \n")
        ofi.write("-r '" + commands_compare[1][3] + "' \n")
        ofi.write("-v '" + commands_compare[1][5] + "' \n")
        ofi.write("-o '" + commands_compare[1][6] + "' \n")              
        ofi.write("### \n")
        ofi.write("-A '" + commands_compare[2][0] + "' \n")
        ofi.write("-B '" + commands_compare[2][1] + "' \n")
        ofi.write("-r '" + commands_compare[2][3] + "' \n")
        ofi.write("-s '" + commands_compare[2][4] + "' \n")
        ofi.write("-p '" + commands_compare[2][2] + "' \n")
        ofi.write("-o '" + commands_compare[2][6] + "' \n")              
        ofi.close()
        
        commands = analyze.doParse('test.inp', './', analyze.getParser())
        for k in commands:
            for i in range(len(commands[k])):
                if i == 6:
                    self.assertEqual(commands[k][i], "./" + commands_compare[k][i])
                else:
                    self.assertEqual(commands[k][i], commands_compare[k][i])

        return

    def test_position(self):
        """
        Tests the correct calculation of position related quantities.
        """

        dim = 15
        m = np.array([1.00 * 1822.8885300626]*6 + [6.00 * 1822.8885300626]*3 + [8.00 * 1822.8885300626]*6 )
        r0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.574000, 0.0, 0.0, -1.57400, 1.0, -1.0, 0.0, -0.25, 0.5, 1.1])

        self.traj = tra.Trajectory(18)
        self.traj.M = m
        self.traj.logR = [r0]
        
        optionstr = "-x 0"
        val_compare = [0.0]
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        for i, v in enumerate(val):
            self.assertEqual(v, val_compare[i])
        
        optionstr = "-x 3,4,5,12"
        val_compare = [r0[3], r0[4], r0[5], r0[12]]
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        for i, v in enumerate(val):
            self.assertEqual(v, val_compare[i])

        optionstr = "-x m,1"
        val_compare = r0[3:6]
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        for i, v in enumerate(val):
            self.assertEqual(v, val_compare[i])

        optionstr = "-x m,0,2,4"
        val_compare = (m[0]*r0[0:3] + m[6]*r0[6:9] + m[12]*r0[12:15]) / (m[0] + m[6] + m[12])
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        for i, v in enumerate(val):
            self.assertEqual(v, val_compare[i])


        optionstr = "-1 0 -2 1"
        val_compare = r0[5]
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, val_compare)
        
        optionstr = "-1 1 -2 2"
        val_compare = 2.0*r0[5]
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, val_compare)

        optionstr = "-1 3 -2 2"
        val_compare = 2.116004726
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertAlmostEqual(val, val_compare, delta=1e-8)

        optionstr = "-1 0 -2 m,0,2,4"
        val_compare = np.linalg.norm((m[0]*r0[0:3] + m[6]*r0[6:9] + m[12]*r0[12:15]) / (m[0] + m[6] + m[12]))
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, val_compare)

        optionstr = "-1 m,1,3 -2 m,0,2,4"
        a = (m[0]*r0[0:3] + m[6]*r0[6:9] + m[12]*r0[12:15]) / (m[0] + m[6] + m[12])
        b = (m[3]*r0[3:6] + m[9]*r0[9:12]) / (m[3] + m[9])
        val_compare = np.linalg.norm(a-b)
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, val_compare)


        optionstr = "-1 m,1,3 -2 m,0,2,4 -p >,1.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 1.0)

        optionstr = "-1 m,1,3 -2 m,0,2,4 -p >,2.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)

        optionstr = "-1 m,1,3 -2 m,0,2,4 -p <,1.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)

        optionstr = "-1 m,1,3 -2 m,0,2,4 -p <,2.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 1.0)


        optionstr = "-1 m,1,3 -2 m,0,2,4 -p 1.0,<,1.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)

        optionstr = "-1 m,1,3 -2 m,0,2,4 -p 1.0,<,2.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 1.0)

        optionstr = "-1 m,1,3 -2 m,0,2,4 -p 2.0,<,3.5"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)        

        optionstr = "-d 1.0,0"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 1.0)        

        optionstr = "-d 6.0,0"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)        
        
        optionstr = "-d 6.0,m,0,1"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 0.0)        

        optionstr = "-d 0.5,m,0,1"
        val = analyze.position(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertEqual(val, 1.0)        

        
        return

    def test_charge(self):
        """
        Tests the correct calculation of charge related quantities.
        """

        dim = 15
        ES = mced.ElectronicState(state='muh', dte=0.01, ratio=0.1, pe=1000, stop=False)

        charge = [[0.0, 0.0], [1.0, 1.0], [2.0, 1.0], [3.0, 2.0]]
        partial = [(0.0, 0.0, -0.3, -0.7, 0.51, 0.51), (1.0, 0.4, -0.1, -0.51, 0.8, 0.4), (2.0, 0.51, 0.0, -0.6, 0.6, 0.51), (3.0, 0.7, -0.2, -0.3, 1.0, 0.8)]
        
        ES.log['charge'] = charge
        ES.log['partial'] = partial

        self.traj = tra.Trajectory(18)
        self.traj.ES = ES
        
        optionstr = ""
        for i in range(4):
            val_compare = charge[i][1]
            val = analyze.charge(optionstr, 0.0, self.traj, i, 1.0)
            self.assertEqual(val, val_compare)

                    
        optionstr = "-t"
        for i in range(4):
            val_compare = partial[i]
            val = analyze.charge(optionstr, 0.0, self.traj, i, 1.0)
            for j, v in enumerate(val):
                self.assertEqual(v, val_compare[j+1])

                                    
        optionstr = "-t -n 0,2"
        for i in range(4):
            val_compare = [partial[i][1]] + [partial[i][3]]
            val = analyze.charge(optionstr, 0.0, self.traj, i, 1.0)
            for j, v in enumerate(val):
                self.assertEqual(v, val_compare[j])

        partial_compare= [(1.0, 1.0, 0.0, 0.0, 0.0), (1.0, 1.0, 0.0, 0.0, 1.0), (0.0, 1.0, 0.0, 0.0, 0.0), (0.0, 1.0, 1.0, 0.0, 0.0)]                                   
        optionstr = "-t -p 0"
        for i in range(4):
            val_compare = partial_compare[i]
            val = analyze.charge(optionstr, 0.0, self.traj, i, 1.0)
            for j, v in enumerate(val):
                self.assertEqual(v, val_compare[j])

        partial_compare= [(0.0, 0.0, 0.0, 1.0, 1.0), (0.0, 0.0, 0.0, 1.0, 0.0), (1.0, 0.0, 0.0, 1.0, 1.0), (1.0, 0.0, 0.0, 1.0, 1.0)]                                   
        optionstr = "-t -p 1"
        for i in range(4):
            val_compare = partial_compare[i]
            val = analyze.charge(optionstr, 0.0, self.traj, i, 1.0)
            for j, v in enumerate(val):
                self.assertEqual(v, val_compare[j])

        return

    def test_kinetic(self):
        """
        Tests the correct calculation of kinetic energy related quantities.
        """

        dim = 15
        m = np.array([1.00 * 1822.8885300626]*6 + [6.00 * 1822.8885300626]*3 + [8.00 * 1822.8885300626]*6 )
        v0 = np.array([1.0, 0.3, -0.2, 0.0, 0.25, -1.0, 2.0, 1.3, -1.5, 0.02, -0.7, 0.0, -0.25, 0.5, 1.1])

        t_compare = np.array([1.13, 1.0625, 47.64, 3.9232, 12.18]) * 1822.8885300626 / 2.0
        
        self.traj = tra.Trajectory(dim)
        self.traj.M = m
        self.traj.logV = [v0]

        ES = mced.ElectronicState(state='muh', dte=0.01, ratio=0.1, pe=1000, stop=False)
        
        charge = [[0.0, 0.0]]
        partial = [(0.0, 0.0, -0.3, -0.7, 0.5, 0.5)]
        ES.log['charge'] = charge
        ES.log['partial'] = partial
        self.traj.ES=ES

        r0 = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.574000, 0.0, 0.0, -1.57400, 1.0, -1.0, 0.0, -0.25, 0.5, 1.1])
        self.traj.logR = [r0]

        redmass = np.array([ 6.0/7.0, 8.0/9.0, 8.0/9.0 ])
        cha = np.array([ partial[0][2]*partial[0][3], partial[0][2]*partial[0][4], partial[0][2]*partial[0][5] ])
        dis = [3.148, 2.116004726, 0.732922915]

        k_compare = np.array([t_compare[0], t_compare[1] + np.sum(redmass * cha / dis)])

        optionstr = ""
        val_compare = np.sum(t_compare)
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertAlmostEqual(val, val_compare, delta=1e-10)

        optionstr = "-n 0"
        val_compare = t_compare[0]
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertAlmostEqual(val, val_compare, delta=1e-10)

        optionstr = "-n 3"
        val_compare = t_compare[3]
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertAlmostEqual(val, val_compare, delta=1e-10)
        
        optionstr = "-n 0,1,3"
        val_compare = t_compare[0] + t_compare[1] + t_compare[3]
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        self.assertAlmostEqual(val, val_compare, delta=1e-10)

        optionstr = "-s"
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        for j, v in enumerate(val):
            self.assertAlmostEqual(v, t_compare[j], delta=1e-10)

        optionstr = "-n 0,1,3 -s"
        val_compare = [t_compare[0], t_compare[1], t_compare[3]]
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        for j, v in enumerate(val):
            self.assertAlmostEqual(v, val_compare[j], delta=1e-10)

        optionstr = "-k"
        val = analyze.kinetic(optionstr, 0.0, self.traj, 0, 1.0)
        for j, v in enumerate(val):
            self.assertAlmostEqual(v, k_compare[j], delta=1e-8)
            if j == 1:
                break

        return
    
    def test_velocity(self):
        """
        Tests the correct calculation of velocity related quantities.
        """

        return


def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(analyzeTestCase('test_red'))
    suite.addTest(analyzeTestCase('test_doParse'))
    suite.addTest(analyzeTestCase('test_position'))
    suite.addTest(analyzeTestCase('test_charge'))
    suite.addTest(analyzeTestCase('test_kinetic'))



    ## TODO
#    suite.addTest(analyzeTestCase('test_bootstrap'))
    
    return suite

if __name__ == "__main__":
    unittest.main()

