#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#! /usr/bin/env python

import os
import unittest
import numpy as np
import random
import sys
import math

import CDTK.xpyder as xpyder

class xpyderTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_start(self):
        """ 
        Testing the starting of model calculations.
        """

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        # write input file here
        inf = open('input.in', 'w')
        inf.write("$SYSTEM \n")
        inf.write("    qchemistry = xmolecule \n")
        inf.write("    xunit       = an \n")
        inf.write("$END \n \n")

        inf.write("$xmolecule \n")	
        inf.write("	pcs=yes \n")
        inf.write("	PE=10000.000000 \n")
        inf.write("	auger=yes \n")
        inf.write("	fluorescence=yes \n")
        inf.write("	MOM=yes \n")
        inf.write("$end \n \n")

        inf.write("$trajectory \n")
        inf.write("     dt = 0.1 fs \n")
        inf.write("     tf = 0.1 fs \n")
        inf.write("$end \n \n")

        inf.write("$quantum \n")
        inf.write("     type = gs \n")
        inf.write("$end \n \n")

        inf.write("$cartPOS  \n")
        inf.write("O  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.7572157    0.5865358 \n")
        inf.write("H  0.0    -0.7572157    0.5865358 \n")
        inf.write("$end \n")

        inf.write("$cartVEL  \n")
        inf.write("O  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.0          0.0 \n")
        inf.write("$end \n")
        inf.close()

        # basic
        sys.argv = ['dummy', '-i', 'input.in', '-d', 'pyder_test', '-p']

        xpyder.start()
                  
        ## check all files
        compare_file = ["O\n", "H\n", "H\n"]
        # compare created files
        with open('pyder_test/atomlist', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1

        compare_file = ["15.994915\n", "1.007825\n", "1.007825\n"]
        # compare created files
        with open('pyder_test/atommass', 'r') as inf:
            i = 0
            for line in inf:
                self.assertAlmostEqual(float(line), float(compare_file[i]))
                i += 1

        compare_file = ["8\n", "1\n", "1\n"]
        # compare created files
        with open('pyder_test/atomnums', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1

        compare_file = ["    0.000      0.000000    -79.098925    -79.098925\n", "    4.134      0.000009    -79.098934    -79.098925\n"]
        # compare created files
        with open('pyder_test/E.log', 'r') as inf:
            i = 0
            for line in inf:
                np.testing.assert_array_almost_equal([float(e) for e in line.split()], [float(e) for e in compare_file[i].split()])
                i += 1

        compare_file = ["    0.000     0.000000     0.000000     0.000000     0.000000     1.430930     1.108392     0.000000    -1.430930     1.108392\n", "    4.134    -0.000000    -0.000000    -0.000001    -0.000000     1.431061     1.108446    -0.000000    -1.431061     1.108446\n"]
        # compare created files
        with open('pyder_test/R.log', 'r') as inf:
            i = 0
            for line in inf:
                np.testing.assert_array_almost_equal([float(e) for e in line.split()], [float(e) for e in compare_file[i].split()])
                i += 1

        compare_file = ["    0.000     0.000000     0.000000     0.000000     0.000000     0.000000     0.000000     0.000000     0.000000     0.000000\n", "    4.134    -0.000000    -0.000000    -0.000001    -0.000000     0.000063     0.000026     0.000000    -0.000063     0.000026\n"]
        # compare created files
        with open('pyder_test/V.log', 'r') as inf:
            i = 0
            for line in inf:
                np.testing.assert_array_almost_equal([float(e) for e in line.split()], [float(e) for e in compare_file[i].split()])
                i += 1

        os.system('rm -rf input.in pyder_test')

        return

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(xpyderTestCase('test_start'))

    return suite

if __name__ == "__main__":
    unittest.main()

