#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
#! /usr/bin/env python

import os
import unittest
import numpy as np
import random
import sys
import math

import CDTK.xsample as xsample

class xsampleTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_start(self):
        """ 
        Testing the starting of model calculations.
        """

        # seed random number generator
        np.random.seed(0)
        random.seed(0)

        # write input file here
        inf = open('input.in', 'w')
        inf.write("$SYSTEM \n")
        inf.write("    xunit       = an \n")
        inf.write("$END \n \n")

        inf.write("$sampling \n")
        inf.write("type = zero_point_energy \n")
        inf.write("nsample = 2 \n")
        inf.write("$end \n \n")

        inf.write("$cartPOS  \n")
        inf.write("O  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.7572157    0.5865358 \n")
        inf.write("H  0.0    -0.7572157    0.5865358 \n")
        inf.write("$end \n")

        inf.write("$cartVEL  \n")
        inf.write("O  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.0          0.0 \n")
        inf.write("H  0.0     0.0          0.0 \n")
        inf.write("$end \n")
        inf.close()

        # write hessian file here TODO
        inf = open('hessian.dat', 'w')
        inf.write("      2.15284860E-02  0.00000000E+00  0.00000000E+00 -1.07643466E-02  0.00000000E+00  0.00000000E+00 -1.07643466E-02  0.00000000E+00  0.00000000E+00 \n")
        inf.write("      0.00000000E+00  7.21033985E-01  4.19675195E-15  0.00000000E+00 -3.60517450E-01 -2.70916535E-01  0.00000000E+00 -3.60517450E-01  2.70916535E-01 \n")
        inf.write("      0.00000000E+00  4.19675195E-15  4.68775827E-01  0.00000000E+00 -1.96187664E-01 -2.34388396E-01  0.00000000E+00  1.96187664E-01 -2.34388396E-01 \n")
        inf.write("     -1.07643466E-02  0.00000000E+00  0.00000000E+00  3.84904999E-03  0.00000000E+00  0.00000000E+00  6.91540018E-03  0.00000000E+00  0.00000000E+00 \n")
        inf.write("      0.00000000E+00 -3.60517450E-01 -1.96187664E-01  0.00000000E+00  4.03036903E-01  2.33551472E-01  0.00000000E+00 -4.25189954E-02 -3.73633947E-02 \n")
        inf.write("      0.00000000E+00 -2.70916535E-01 -2.34388396E-01  0.00000000E+00  2.33551472E-01  2.20586341E-01  0.00000000E+00  3.73633947E-02  1.38025369E-02 \n")
        inf.write("     -1.07643466E-02  0.00000000E+00  0.00000000E+00  6.91540018E-03  0.00000000E+00  0.00000000E+00  3.84904999E-03  0.00000000E+00  0.00000000E+00 \n")
        inf.write("      0.00000000E+00 -3.60517450E-01  1.96187664E-01  0.00000000E+00 -4.25189954E-02  3.73633947E-02  0.00000000E+00  4.03036903E-01 -2.33551472E-01 \n")
        inf.write("      0.00000000E+00  2.70916535E-01 -2.34388396E-01  0.00000000E+00 -3.73633947E-02  1.38025369E-02  0.00000000E+00 -2.33551472E-01  2.20586341E-01 \n")
        inf.close()

        # basic
        sys.argv = ['dummy', '-i', 'input.in', '-d', 'sample_test', '-H', 'hessian.dat']
        
        xsample.start()
                  
        ## check all files

        compare_file = ["O\n", "H\n", "H\n"]
        # compare created files
        with open('./ensmbl_sample_test/atomlist', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1
        with open('ensmbl_sample_test/trj_0000000/atomlist', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1
        with open('ensmbl_sample_test/trj_0000001/atomlist', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1

        compare_file = ["15.994915\n", "1.007825\n", "1.007825\n"]
        # compare created files
        with open('ensmbl_sample_test/atommass', 'r') as inf:
            i = 0
            for line in inf:
                self.assertAlmostEqual(float(line), float(compare_file[i]))
                i += 1
        with open('ensmbl_sample_test/trj_0000000/atommass', 'r') as inf:
            i = 0
            for line in inf:
                self.assertAlmostEqual(float(line), float(compare_file[i]))
                i += 1
        with open('ensmbl_sample_test/trj_0000001/atommass', 'r') as inf:
            i = 0
            for line in inf:
                self.assertAlmostEqual(float(line), float(compare_file[i]))
                i += 1

        compare_file = ["8\n", "1\n", "1\n"]
        # compare created files
        with open('ensmbl_sample_test/atomnums', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1
        with open('ensmbl_sample_test/trj_0000000/atomnums', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1
        with open('ensmbl_sample_test/trj_0000001/atomnums', 'r') as inf:
            i = 0
            for line in inf:
                self.assertEqual(line, compare_file[i])
                i += 1

        # compare_file = ["-6.42357332676e-21\n", "0.00711300503645\n", "0.0024326184825\n", "1.63386084017e-17\n", "1.24360516255\n", "1.04644898404\n", "-1.00463339224e-17\n", "-1.3564935828\n", "1.13172739472\n"]
        # compare created files
        #with open('ensmbl_sample_test/trj_0000000/atompos', 'r') as inf:
        #    i = 0
        #    for line in inf:
        #        self.assertAlmostEqual(float(line), float(compare_file[i]))
        #        i += 1

        # compare_file = ["-1.55435971402e-22\n", "0.000172118662182\n", "-0.000111586022648\n", "2.39164446218e-19\n", "-0.00255186048075\n", "-0.000146295118732\n", "-8.6905610468e-20\n", "-0.000179784410444\n", "0.00191724987876\n"]
        # compare created files
        # with open('ensmbl_sample_test/trj_0000000/atomvel', 'r') as inf:
        #    i = 0
        #    for line in inf:
        #        self.assertAlmostEqual(float(line), float(compare_file[i]))                
        #        i += 1


        # compare_file = ["-8.47148770611e-21\n", "0.00938071874555\n", "-0.00870909403764\n", "6.24940183593e-18\n", "1.37668974436\n", "1.12126880058\n", "2.04892733194e-18\n", "-1.5255683849\n", "1.23373502007\n"]
        # compare created files
        # with open('ensmbl_sample_test/trj_0000001/atompos', 'r') as inf:
        #    i = 0
        #    for line in inf:
        #        self.assertAlmostEqual(float(line), float(compare_file[i]))
        #        i += 1

        # compare_file = ["-1.1841431389e-22\n", "0.000131123530198\n", "-1.67098768008e-05\n", "3.03518143378e-19\n", "-0.00346636688175\n", "-0.000653426343993\n", "-1.87524236652e-19\n", "0.00138534372691\n", "0.000918624596701\n"]
        # compare created files
        # with open('ensmbl_sample_test/trj_0000001/atomvel', 'r') as inf:
        #    i = 0
        #    for line in inf:
        #        self.assertAlmostEqual(float(line), float(compare_file[i]))
        #        i += 1

        os.system('rm -rf input.in hessian.dat ensmbl_sample_test')

        return

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(xsampleTestCase('test_start'))

    return suite

if __name__ == "__main__":
    unittest.main()

