#!/usr/bin/env python

#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************

"""
Tool to calculate x-ray absorption cross sections and other related quantities
for either core orbitals to valence hole or between excited state transitions
along a CDTK trajectory using XMOLECULE interface.
This should be run from inside a CDTK trajectory folder since it needs the
files 'atomlist', 'R.log' and 'S.log'. Additionally 'topol' file to include
point charges other than just water, if that is desired.

Possible outputs from this script are absorption cross sections, molecular
orbital energies, partial charges, hole and charge centres, excited state
energies, etc.
For a complete list of input and output options, please check the help for
this script using the '-h' option.

Note: For now, calculation of absorption cross section only works for
      molecules with specific atoms such as H, C, O and N.
"""


import os
import sys
from optparse import OptionParser
import numpy as np
import subprocess as sp

import CDTK.Tools.Utils as uti
import CDTK.Tools.Conversion as cv
import CDTK.Interfaces.XMoleculeInterface as xmol
import CDTK.Interfaces.ForceField as ff


def get_parser():
    """ Return parser object to parse command line arguments.

    Returns
    -------
    parser : OptionParser object
    """

    parser=OptionParser()

    parser.add_option('--xmolpath',
            dest='xmolpath',
            type='string',
            help='Absolute path to XMOLECULE directory. This will be used to '
                 'access both "basis_sets" and "src" subdirectories.')

    parser.add_option('--gto',
            dest='gto',
            type='string',
            default='STO-3G',
            help='Gaussian basis set to be used by XMOLECULE. Only basis name '
                 'should be given and not a file name.')

    parser.add_option('--dt',
            dest='dt',
            type='float',
            help='Time step to calculate aCS along trajectory [fs]')

    # TODO: This might be very useful to add to look into for example MO movie for a particular time range.
    #parser.add_option('--t0',
    #        dest='t0',
    #        type='float',
    #        default=0.0,
    #        help='Starting time [fs]')

    parser.add_option('--tmax',
            dest='tmax',
            type='float',
            help='Maximal time to consider [fs]')

    # parser.add_option('--qcut',
    #         dest='qcut',
    #         type='float',
    #         default=0.1,
    #         help='Cutoff for charges')

    parser.add_option('--natom',
            dest='natom',
            type='int',
            help='Number of QM atoms to consider')

    parser.add_option('--elec',
            dest='elec',
            action='store_true',
            default=False,
            help='Classical electron present in trajectory')

    parser.add_option('--embed',
            dest='embed',
            action='store_true',
            default=False,
            help='Whether to include atoms beyond #natom as point charges.'\
                 ' For now, water point charges are using SPC/E water model.'\
                 ' This adds "_w_point_charges" as file suffix to all output'\
                 ' files except for bond order and movie cube files. Default: False')

    parser.add_option('--is_MM_water',
            dest='is_MM_water',
            action='store_true',
            default=False,
            help='Whether the MM region is just water. Default: False')

    parser.add_option('--emnmol',
            dest='emnmol',
            type='int',
            default=None,
            help='Number of water molecules considered for embedding as point charges.'\
                 ' This is an option only if is_MM_water==True.' )

    # Options for outputting required quantities start here. 
    parser.add_option('--absorptionCS',
            dest='absorptionCS',
            action='store_true',
            default=False,
            help='Whether to output absorption cross section with spectral lines'\
                 ' for either core orbital to valence hole transitions,'\
                 ' or excited state transitions. Default: False')

    parser.add_option('--full_hole_CS',
            dest='get_full_hole_CS',
            action='store_true',
            default=False,
            help='Whether to output absorption cross section with spectral lines'\
                 ' for all core orbital to unoccupied orbital (including hole),'\
                 ' transitions instead of only to valence hole in Koopmans theorem. Default: False')

    parser.add_option('--static_CS',
            dest='get_static_CS',
            action='store_true',
            default=False,
            help='Whether to output static (i.e. t=0) absorption cross section'\
                 ' with spectral lines for all core orbital to unoccupied orbital'\
                 ' transitions. This uses "atompos" file instead of "R.log" and'\
                 ' "S.log" from a particular CDTK sampling folder.'\
                 ' Note: this only work for (unperturbed) ground state for now,'\
                 ' and not ionized or excited state.')

    parser.add_option('--acs_mo',
            dest='acs_mo',
            type='string',
            default=None,
            help='List of unoccupied molecular orbitals (separated by comma) for'\
                 ' which absorption cross section with spectral lines should be'\
                 ' outputted, possibly along with the hole orbital if hole is present.'\
                 ' Note: this only works if "full_hole_CS" or "static_CS" is'\
                 ' selected, and counting of MO starts from 1.')
 
    # TODO: how to deal with cases when this leads to different number of possible transitions?
    parser.add_option('--emin_abs',
            dest='emin_abs',
            type='float',
            default=0.0,
            help='Minimum photon energy [eV] to be considered in absorption'\
                 ' cross-section. Default: 0 eV. Note: This only works for'\
                 ' excited states for now and not ionized states.')

    parser.add_option('--orbitalEnergies',
            dest='orbitalEnergies',
            action='store_true',
            default=False,
            help='Whether to output occupied orbital energies. Default: False')

    parser.add_option('--movie',
            dest='movie',
            action='store_true',
            default=False,
            help='Whether to output molecular orbitals (MO) as cube files for movie generation.'\
                 ' For Koopmans theorem, the current hole MO is used and for'\
                 ' excites states, MO selected by "es_movie_mo" is used. Default: False')

    parser.add_option('--cmid',
            dest='cmid',
            action='store_true',
            default=False,
            help='Whether to output the charge center. Default: False')

    parser.add_option('--bord',
            dest='bord',
            action='store_true',
            default=False,
            help='Whether to output the bond order analysis results. Default: False')

    parser.add_option('--hmid',
            dest='hmid',
            action='store_true',
            default=False,
            help='Whether to output the hole center. Default: False')

    parser.add_option('--is_QM_water',
            dest='is_QM_water',
            action='store_true',
            default=False,
            help='Whether the QM region is a water cluster. Default: False')

    parser.add_option('--add_converge_params',
            dest='add_extra_params',
            action='store_true',
            default=False,
            help='Whether additional parameters should be added for faster'\
                 ' convergence of XMOLECULE calculation. Default: False')

    # Options for excited states start here. By default, Koopmans' theorem
    # ionized states are used and excited states have to be turned on explicitly
    # using these options.
    parser.add_option('--use_excited_state',
            dest='use_excited_state',
            action='store_true',
            default=False,
            help='Whether to output excited state energies and cross sections'\
                 ' instead of the default valence hole state ones. Default: False')

    parser.add_option('--es_energy_states',
            dest='es_energy_states',
            type='int',
            default=None,
            help='Output excited state energies (plus the ground state) upto given state.'\
                 ' If not given, excited state energies are not outputted.'\
                 ' Note: ground state and "es_energy_states"-1 states are considered.')

    # For now, select a particular orbital to output cube files for a molecular
    # movie for the excited state case.
    # TODO: maybe it makes sense to output the most dominant MO in the current excited state.
    parser.add_option('--es_movie_mo',
            dest='es_movie_mo',
            type='int',
            default=None,
            help='Index of molecular orbital (MO) to be outputted for movie.'\
                 ' Note: Counting of MO starts from 1, and this is only an option'\
                 ' for excited states since for Koopmans theorem, hole MO is used.')

    return parser


# TODO: need to add headers to output files.

def do_bond_order_analysis(i, xm, natoms):
    """ Perform bond order analysis and write the result to file with name
    'bondorders-<i>.dat' where i is the time step index in 4 spaces.
    Note: The folder 'bondorders' is needed for this function.

    Parameters
    ----------
    i : int
        Time step index.
    xm : XMoleculeInterface object
        Object that interfaces with XMOLECULE code.
    natoms : int
        Number of atoms in QM part.

    """
    outfile = open('./bondorders/bondorders-' + '{0:04d}'.format(i) + '.dat', 'w')
    bondOrders = np.asarray(xm.getBondOrders())
    bondOrders = bondOrders[:natoms, :natoms]    # restrict to QM region

    for ll, bo in enumerate(bondOrders):
        for jj, bbo in enumerate(bo):
            outfile.write('  {0:.4f}'.format(bbo))
        outfile.write('\n')
    outfile.close()

    return


def do_hole_center_analysis(t, xm, natoms, hole_index, r_xyz, atomlist, hole_pop_file, hole_center_file):
    """ Get hole center and write the result to `hole_center_file`
    (here named 'hole_center.dat') with columns being time in fs, xyz coordinates
    of the hole center, and index, symbol and distance to the closest atom.
    Hole population on each atoms is also written to `hole_pop_file`
    (here 'hole_populations.dat') with columns being time in fs, hole
    populations for on each atom and total hole population (which should in theory be 1).
    

    Parameters
    ----------
    t : float
        Current time.
    xm : XMoleculeInterface object
        Object that interfaces with XMOLECULE code.
    natoms : int
        Number of atoms in QM part.
    hole_index : int
        Index of orbital with the hole.
    r_xyz : (natoms, 3) ndarray of floats
        xyz coordinates of QM atoms at time `t`.
    atomlist : (natoms) list of string
        Atom symbol for all QM atoms.
    hole_pop_file : file object
        File to output hole population to.
    hole_center_file : file object
        File to output hole center to.

    """

    # 'gpops' is (n_mo, natoms) list of floats as populations,
    # where n_mo is the number of molecular orbitals.
    # Note: populations for point charges (embeddings) are zero since
    #       they don't have any associated basis functions so they are not considered.)
    gpops = np.array(xm.getPopulation())[:, :natoms]
    hole_pop = gpops[hole_index]

    # center of "mass" of the hole, weighted by hole population.
    # Hole population should be normalized (i.e. sum is 1) since they are populations.
    center_hole = np.average(r_xyz, axis=0, weights=hole_pop)

    # Get index of closest atom to hole center and its distance
    dist_to_center = np.linalg.norm(r_xyz - center_hole, axis=1)
    center_atom_index = np.argmin(dist_to_center)
    min_dist = dist_to_center[center_atom_index]

    hole_pop_file.write(' {0:.4f}'.format(t))
    for p in hole_pop:
        hole_pop_file.write('  {:.6f}'.format(p))
    hole_pop_file.write('  {:.6f}\n'.format(np.sum(hole_pop)))

    hole_center_file.write(' {0:.4f}   {1:.6f}   {2:.6f}   {3:.6f}   {4:d}   {5:s}   {6:.6f}\n'\
                            .format(t, center_hole[0], center_hole[1], center_hole[2],
                                    center_atom_index, atomlist[center_atom_index], min_dist))

    return


def get_charges(t, xm, natoms, charge_file, is_QM_water=False):
    """ Get partial charges for each QM atom and write the result to `charge_file`
    (here 'partial_charges.dat') with columns being time in fs, partial charges
    on each atom and total charge (which should in theory be 1 since there is
    1 initial charge to begin with).

    Parameters
    ----------
    t : float
        Current time.
    xm : XMoleculeInterface object
        Object that interfaces with XMOLECULE code.
    natoms : int
        Number of atoms in QM part.
    charge_file : file object
        File to output partial charges to.
    is_QM_water : bool
        Whether the QM region is a water cluster. If yes, then appropriate water
        model charges are subtracted to get charge differences.

    Returns
    -------
    differential_charges : (natoms) ndarray of floats
        Partial charges on each QM atom (with charge in equilibrium state subtracted,
        if applicable, to get only change in charges).
    """

    # Get (natoms) ndarray of floats as partial charges for each atom.
    # Note: partial charges for point charges (embeddings) are zero since
    #       they don't have any associated basis functions so they are not considered.)
    charges = np.array(xm.getPartialCharges())[:natoms]

    if not is_QM_water:
        differential_charges = charges.copy()

    else:
        # Get changes in charge compared to water model rather than absolute charge.
        differential_charges = np.zeros(len(charges))

        # TODO: adapt to general case rather than just water.
        for ll, c in enumerate(charges):
            if ll%3 == 0:
                # if 'O'
                differential_charges[ll] = c+0.8476
            else:
                # if 'H'
                differential_charges[ll] = c-0.4238

    charge_file.write(' {0:.4f}'.format(t))
    for c in differential_charges:
        charge_file.write('  {:.6f}'.format(c))
    charge_file.write('  {:.6f}\n'.format(np.sum(differential_charges)))

    return differential_charges


def do_charge_center_analysis(t, xm, r_xyz, charges, atomlist, charge_center_file):
    """ Get charge center and write the result to `charge_center_file`
    (here named 'charge_center.dat') with columns being time in fs, xyz coordinates
    of charge center, and index, symbol and distance to the closest atom.

    Parameters
    ----------
    t : float
        Current time.
    xm : XMoleculeInterface object
        Object that interfaces with XMOLECULE code.
    r_xyz : (natoms, 3) ndarray of floats
        xyz coordinates of QM atoms at time `t`.
    charges : (natoms) ndarray of floats
        Partial charges of atom at time `t`.
    atomlist : (natoms) list of string
        Atom symbol for all QM atoms.
    charge_center_file : file object
        File to output charge center to.
    """

    # Get charge center weighted by partial charges.
    charge_center = np.average(r_xyz, axis=0, weights=charges)

    # Get index of closest atom to charge center and its distance
    dist_to_center = np.linalg.norm(r_xyz - charge_center, axis=1)
    center_atom_index = np.argmin(dist_to_center)
    min_dist = dist_to_center[center_atom_index]

    charge_center_file.write(' {0:.4f}   {1:.6f}   {2:.6f}   {3:.6f}   {4:d}   {5:s}   {6:.6f}\n'\
                             .format(t, charge_center[0], charge_center[1], charge_center[2],
                                     center_atom_index, atomlist[center_atom_index], min_dist))

    return


def write_absorption_cs(opts, t_vals, deltaE_vals, a_vals, mo_list=None, file_suffix=None):
    """ Write absorption cross section to file named 'absorptionCS.dat'.
    Each time step is written in a new row with columns being time in fs,
    energy difference in eV (between either hole orbital and a core orbital,
    or excited states) representing the transition line and then its cross section
    (this is repreated for either each core orbital or all excites state transitions).

    Parameters
    ----------
    opts : parser output object
        Object with attributes set by OptionParser object
    t_vals : (nsteps) ndarray of floats
        Times used to evaluate cross section in fs.
    deltaE_vals : (nsteps, ncore) ndarray of floats
        Energy differences representing transition lines between either hole
        orbital and core orbital, or excited states evaluated at `t_vals` in eV.
    a_vals : (nsteps, ncore) ndarray of floats
        Cross sections for the transition lines evaluated at `t_vals` in au.
    mo_list : list of int or None
        Particular MOs requested.
    file_suffix : str or None
        Suffix to add to output file. Can be for example if point charges are included.
    """
    file_ending_w_suffix = 'CS' + file_suffix + '.dat'

    # Add MO list to file name and file header if present.
    if mo_list is not None:
        add_mo_filename = 'mo_'
        add_mo_header = 'MO '
        for mo in mo_list:
            add_mo_filename = add_mo_filename + str(mo) + '_'
            add_mo_header = add_mo_header + str(mo) + ', '

    if opts.get_full_hole_CS:
        if mo_list is None:
            file_name = 'absorption_full_hole_' + file_ending_w_suffix
        else:
            file_name = 'absorption_hole_' + add_mo_filename + file_ending_w_suffix
    elif opts.get_static_CS:
        if mo_list is None:
            file_name = 'absorption_static_full_' + file_ending_w_suffix
        else:
            file_name = 'absorption_static_' + add_mo_filename + file_ending_w_suffix
    else:
        file_name = 'absorption' + file_ending_w_suffix

    with open(file_name,'w') as f:
        # Writing header to file.
        f.write('# Absorption cross-sections along trajectory')
        f.write('\n')
        f.write('# dt [fs]: {0}'.format(opts.dt))
        f.write('\n')
        f.write('# GTO: ' + opts.gto)
        f.write('\n')
        if not opts.use_excited_state:
            if opts.get_full_hole_CS:
                if mo_list is None:
                    f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated' \
                            ' for all core-unoccupied orbital (including hole) transitions.')
                else:
                    f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated' \
                            ' for core to ' + add_mo_header + 'and hole transitions.')
            elif opts.get_static_CS:
                if mo_list is None:
                    f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated' \
                            ' for all core-unoccupied orbital transitions.')
                else:
                    f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated' \
                            ' for core to ' + add_mo_header + 'transitions.')
            else:
                f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated for all core-valence hole transitions.')
        else:
            f.write('#  t[fs] [:  DeltaE[eV]  aCS[a.u.] :] -- repeated for all excited state transitions.')
        f.write('\n')

        for i, t in enumerate(t_vals):
            f.write('  {0:.4f}'.format(t))
            for j in range(deltaE_vals.shape[1]):
                f.write('  {0:.4f}  {1:.4e}'.format(deltaE_vals[i,j], a_vals[i,j]))
            f.write('\n')
    return


def write_orbital_energies(opts, t_vals, e_vals, file_suffix=None):
    """ Write occupied orbital energies to file named 'occupied_orbital_energies.dat'.
    Each time step is written in a new row with columns being time in fs,
    energies for all occupied orbitals in eV.

    Parameters
    ----------
    opts : parser output object
        Object with attributes set by OptionParser object
    t_vals : (nsteps) ndarray of floats
        Times used to evaluate cross section in fs.
    e_vals : (nsteps, n_orbitals) ndarray of floats
        Occupied orbital energies evaluated at `t_vals` in eV.
    file_suffix : str or None
        Suffix to add to output file. Can be for example if point charges are included.
    """
    # TODO: better to use np.savetxt with column concatinated t_vals and e_vals
    with open('occupied_orbital_energies' + file_suffix + '.dat','w') as f:
        # Writing header to file.
        f.write('# Occupied orbital energies along trajectory')
        f.write('\n')
        f.write('# dt [fs]: {0}'.format(opts.dt))
        f.write('\n')
        f.write('# GTO: ' + opts.gto)
        f.write('\n')
        f.write('#  t[fs]      Energy for all occupied orbitals [eV]')
        f.write('\n')

        n_orbitals = e_vals.shape[1]

        for i, t in enumerate(t_vals):
            f.write('  {0:.4f}'.format(t))
            for j in range(n_orbitals):
                f.write('  {0:.4f}'.format(e_vals[i,j]))
            f.write('\n')

    return


def write_es_energies(opts, t_vals, e_vals, file_suffix=None):
    """ Write excited state energies to file named 'excited_state_energies.dat'.
    Each time step is written in a new row with columns being time in fs,
    energies in eV for ground state and 'es_energy_states'-1 excited states.

    Parameters
    ----------
    opts : parser output object
        Object with attributes set by OptionParser object
    t_vals : (nsteps) ndarray of floats
        Times used to evaluate cross section in fs.
    e_vals : (nsteps, es_energy_states) ndarray of floats
        Excited state energies evaluated at `t_vals` in eV.
    file_suffix : str or None
        Suffix to add to output file. Can be for example if point charges are included.
    """
    # TODO: better to use np.savetxt with column concatinated t_vals and e_vals
    with open('excited_state_energies' + file_suffix + '.dat','w') as f:
        # Writing header to file.
        f.write('# Excited state energies along trajectory')
        f.write('\n')
        f.write('# dt [fs]: {0}'.format(opts.dt))
        f.write('\n')
        f.write('# GTO: ' + opts.gto)
        f.write('\n')
        f.write('#  t[fs]      Energy [eV] from ground state upto {0} excited state'.format(opts.es_energy_states-1))
        f.write('\n')

        n_es = e_vals.shape[1]

        for i, t in enumerate(t_vals):
            f.write('  {0:.4f}'.format(t))
            for j in range(n_es):
                f.write('  {0:.4f}'.format(e_vals[i,j]))
            f.write('\n')
    return


def start():
    """ Main part of the program. This structure of starting with a function
    can be useful for unit testing purposes.
    """

    parser = get_parser()
    opts, args = parser.parse_args(sys.argv[1:])

    if (opts.es_energy_states is not None) and (not opts.use_excited_state):
        raise RuntimeError("Excited state energies can only be outputted if"
                           " 'use_excited_state' option is selected.")

    if (opts.es_movie_mo is not None) and (not opts.use_excited_state):
        raise RuntimeError("Specific molecular orbital is only outputted for"
                           " excited states using the 'use_excited_state' option.")

    if opts.hmid and (opts.use_excited_state or opts.get_static_CS):
        raise RuntimeError("Calculating hole centre only makes sense"
                           " when using Koopmans' theorem and not for"
                           " excited states or static ground state.")

    if opts.cmid and (opts.use_excited_state or opts.get_static_CS):
        raise RuntimeError("Calculating charge centre only makes sense"
                           " when there is a hole using Koopmans' theorem, and"
                           " not for excited states or static ground state" 
                           " as there is no overall charge.")

    if opts.get_full_hole_CS and opts.use_excited_state:
        raise RuntimeError("Obtaining full absorption cross section from core"
                           " orbital to all unoccupied orbital only makes sense"
                           " when using Koopmans' theorem and not for"
                           " excited states.")

    if (opts.acs_mo is not None) and (not (opts.get_full_hole_CS or opts.get_static_CS)):
        raise RuntimeError("Selecting unoccupied MOs to be used for outputting"
                           " absorption cross section only makes sense if full"
                           " or static spectra is requested.")


    # Basis set file with path to it.
    basis_file_w_path = opts.xmolpath + '/basis_sets/' + opts.gto + '.gbs'

    # Setup general xmolecule options;
    # Here N is an option for grid points and is only needed for point charges (embeddings).
    xmOpts={'HF':['yes'],
           'gto':[basis_file_w_path],
           'SIni':['yes'],
           'N':['1']}

    if opts.use_excited_state:
        # Options for CIS excited state
        extra_state_opts = {'cis':['yes'],
                            'ncore':['0']}
    elif opts.get_static_CS:
        # For static case, simply use ground state
        extra_state_opts = {}
    else:
        # Options for Koopmans' theorem hole states,
        # Note: only singly charged states are considered here.
        extra_state_opts = {'gs_occ':['yes'],
                            'charge':['1']}
        
    xmOpts.update(extra_state_opts)

    # Additional parameters to help convergence, following Confluence page of XMOLECULE.
    if opts.add_extra_params:
        extra_opts = {'NODIIS': ['yes'],
                      'LevelShift': ['1.0'],
                      'SwitchToDIIS' : ['0.001'],
                      'cphfdiis': ['yes']}
        xmOpts.update(extra_opts)

    # read atomlist
    alist = uti.readColumnFile('atomlist')[:opts.natom]
    natoms = len(alist)

    if opts.embed:
        # Add suffix to files if point charges are used, except for bondorders
        # and movie cube files.
        pc_file_suffix = "_w_point_charges"

        if opts.is_MM_water and (opts.emnmol is not None):
            # to speed up compromise on embedding region size
            last_embed_ix = 3*opts.emnmol
            alist2 = uti.readColumnFile('atomlist')[opts.natom:opts.natom + last_embed_ix]
        else:
            alist2 = uti.readColumnFile('atomlist')[opts.natom:]

        nembed = len(alist2)

        # embed: point charges with SPC/E model for water
        point_charges = []
        if opts.is_MM_water:
            for label in alist2:
                if label=='O':
                    point_charges.append( -0.8476 )
                elif label=='H':
                    point_charges.append( 0.4238 )
        else:
            # If other point charges apart from water are present, get charges
            # from forcefields module with SPC/E water model.
            if os.path.exists('topol'):
                topol = np.array(uti.readColumnFile('topol'))
            else:
                raise ValueError('No file ./topol found.')

            MMFull_ff = ff.jointForceFields(topol, waterForcefield='SPC/E')
            point_charges = list(MMFull_ff.charges[opts.natom:])

            print("MM atoms to include: %d" % topol[opts.natom:].shape[0])
            print("MM point charges included: %d" % len(point_charges))

    else:
        pc_file_suffix = ""

    # Create output folders and files
    if opts.bord:
        if not os.path.exists('bondorders'):
            os.mkdir('bondorders')
    if opts.movie:
        if not os.path.exists('cubes'):
            os.mkdir('cubes')

    if opts.cmid:
        charge_file = open('partial_charges' + pc_file_suffix + '.dat', 'w')
        charge_center_file = open('charge_center' + pc_file_suffix + '.dat', 'w')

    if opts.hmid:
        hole_pop_file = open('hole_populations' + pc_file_suffix + '.dat', 'w')
        hole_center_file = open('hole_center' + pc_file_suffix + '.dat', 'w') 

    if opts.absorptionCS:
        if opts.acs_mo is not None:
            # This gives a list of ints representing the MOs.
            acs_mo_list = [int(mo) for mo in opts.acs_mo.split(',')]
        else:
            acs_mo_list = None

    # Read timesteps (to get array 't', and ints 'step' and 'nsteps'), states (as a list of int) and
    # positions ('r_xyz_full' as (n_times, natoms + nembed, 3) ndarray of floats) over the trajectory.

    if not opts.get_static_CS:
        # read trajectory time steps; have cutoff for times more than tmax.
        t = uti.getTrajectoryTimeSteps('R.log') * cv.au2fs
        dt_traj = (t[1] - t[0])
        t = t[t <= opts.tmax]
        # Obtain total number of steps needed based on requested step size and trajectory step.
        # TODO: what to do for cases when 'step' isn't integer?
        if opts.dt < dt_traj:
            step = 1
        else:
            step = (opts.dt+1e-5)//dt_traj
        nsteps = len(t)// step

        print("Step = ", step)
        print("nsteps = ", nsteps)

        # Read trajectory electronic states;
        # for Koopmans' theorem, state = 0 refers to HOMO, state = 1 to HOMO-1 and so on;
        # for excited states, state = 0 refers to ground state, state = 1 to first excited state and so on.
        states = uti.readColumnFile('S.log',col_ix=1)
        states = list(map(int, states)) 

        # Read trajectory xyz coordinates for all times.
        r_xyz_full = uti.atomicCoordinatesFromRlog('R.log')

    else:
        # For static case, there is only t=0 and single ground state coordiantes
        t = [0]
        step = 1
        nsteps = 1
        states = [0]
        r_xyz_full = np.loadtxt('atompos').reshape(1, -1, 3)

    # Separate coordinates into QM ('r_xyz') and MM ('r2_xyz') part if necessary.
    r_xyz = r_xyz_full[:,:opts.natom]
    
    if opts.embed:
        if opts.is_MM_water and (opts.emnmol is not None):
            r2_xyz = r_xyz_full[:, opts.natom:opts.natom + last_embed_ix]
        else:
            r2_xyz = r_xyz_full[:, opts.natom:]

    # Only needed if classical electron is present.
    if opts.elec:
        elec_xyz = r_xyz[:,-1,:]
        elec_typ = 'X'
        elec_q   = [-1.0]
        r_xyz    = r_xyz[:,:-1,:]


    # Initialize xmolecule interface with first timestep geometry
    xm=xmol.xmolecule()
    if opts.elec:
        xm.init_input_options(xmOpts, alist, r_xyz[0].flatten(),
                              x_pos=elec_xyz[0].flatten(), x_charges=elec_q)
    elif opts.embed:
        xm.init_input_options(xmOpts, alist, r_xyz[0].flatten(),
                              x_pos=r2_xyz[0].flatten(), x_charges=point_charges)
    else:
        xm.init_input_options(xmOpts, alist, r_xyz[0].flatten())

    # From now on use RIni/SIni scheme (i.e. using initial guess wavefunction from last time step)
    # TODO: This isn't initialized in the input file though since only for "movie" that is done.
    xmOpts['RIni']='yes'
    occs=xm.get_occ()


    # Note: This is only implemented for certain elements. For elements with
    #       higher atomic number, what is considered core orbital might vary.
    # TODO: Maybe include this in the CDTK element dictionary?

    # Index for HOMO orbital, indexing starts from core orbital with 0
    homo_ix = len([True for i in occs if i != 0]) - 1  # Use occs till where it goes to 0.
    print("HOMO index: ", homo_ix)

    # Number of core orbitals, this is 0 for H, and 1 for C, N and O
    ncore = 0
    for atom in alist:
        if (atom == 'C' or atom == 'N' or atom =='O'):
            ncore += 1
        elif(atom == 'H'):
            pass
        else:
            raise NotImplementedError("Calculation of core orbitals for "
                                      + atom + " is not implemented yet.")
    print("Number of core orbitals: ", ncore)

    # energy differences and absorption cross-sections
    t_vals = np.zeros(nsteps)
    orbital_e_vals = np.zeros([nsteps, (homo_ix+1)])


    # For Koopmans' theorem, these will be (nsteps, ncore) if only transitions
    # to hole orbital are requested.
    # For excited states, the exact number of states is difficult to estimate
    # so unspecified here but these will be 'nsteps' list in the end.
    deltaE_vals = []
    a_vals = []

    if opts.use_excited_state:
        es_e_vals = []  # This contains ground and excited state energies

    # calculate quantities along trajectory
    for i in range(nsteps):
        current_t = t[i*step]

        if opts.use_excited_state:
            # For excited state, just need current state
            current_es_state = states[i*step]

        elif opts.get_static_CS:
            # TODO: is this even needed for static case?
            new_occs = [0 if occ==0 else 2 for occ in occs]
            xm.set_occ(new_occs)

        else:
            # For Koopmans' theorem, set occupation based on where the hole is.
            hole_ix = homo_ix - states[i*step]
            new_occs = [0 if occ==0 else 2 for occ in occs]
            new_occs[hole_ix] = 1
            xm.set_occ(new_occs)

        # Here 'new_r_xyz' is (natoms, 3) ndarray of floats.
        new_r_xyz = r_xyz[i*step]
        new_r_full = new_r_xyz.flatten()

        if opts.elec:
            new_elec_xyz = elec_xyz[i*step]
            new_r_full = np.append(new_r_full, new_elec_xyz.flatten())
        elif opts.embed:
            new_r2_xyz = r2_xyz[i*step]
            new_r_full = np.append(new_r_full, new_r2_xyz.flatten())

        xm.set_geom(new_r_full)

        if opts.cmid:
            charges = get_charges(current_t, xm, natoms, charge_file, opts.is_QM_water)
            do_charge_center_analysis(current_t, xm, new_r_xyz, charges, alist, charge_center_file)

        if opts.bord:
            do_bond_order_analysis(i, xm, natoms)

        if opts.hmid:
            do_hole_center_analysis(current_t, xm, natoms, hole_ix, new_r_xyz, alist, hole_pop_file, hole_center_file)

        if opts.absorptionCS or opts.orbitalEnergies or opts.es_energy_states:
            t_vals[i] = current_t

            moes = np.asarray(xm.getOrbitalEnergies()) * cv.au2ev
            orbital_e_vals[i] = moes[:(homo_ix+1)]

            if opts.use_excited_state:
                # This contains ground state + all excited state energies.
                # TODO: Add shift from initial ground state since there seems to be an arbritatry shift.
                es_energies = np.array(xm.getCISEnergies()) * cv.au2ev
                # Only append till selected state.
                # TODO: should there be a check for 'opts.es_energy_states'?
                es_e_vals.append(es_energies[:opts.es_energy_states])

            if opts.absorptionCS:
                if not opts.use_excited_state:
                    xacs, tdip = xm.getAbsorptionCrossSections()

                    if not opts.get_static_CS:
                        # Transition to hole from core orbitals.
                        deltaE = moes[hole_ix] - moes[:ncore]
                        selected_xacs = xacs[:ncore, hole_ix]
                    else:
                        deltaE = np.array([])
                        selected_xacs = np.array([])


                    if opts.get_full_hole_CS or opts.get_static_CS:
                        if opts.acs_mo is None:
                            # Add transitions to all other unoccupied orbitals from core orbitals.
                            for core_orb in range(ncore):
                                deltaE = np.append(deltaE, (moes[(homo_ix+1):] - moes[core_orb]))
                                selected_xacs = np.append(selected_xacs, (xacs[core_orb, (homo_ix+1):]))
                        else:
                            # Add transitions to selected unoccupied orbitals from core orbitals.
                            for core_orb in range(ncore):
                                for sel_mo in acs_mo_list:
                                    # Index for counting starts from 0 in python, hence the '-1' is needed.
                                    deltaE = np.append(deltaE, (moes[sel_mo-1] - moes[core_orb]))
                                    selected_xacs = np.append(selected_xacs, (xacs[core_orb, sel_mo-1]))


                    deltaE_vals.append(deltaE)
                    a_vals.append(selected_xacs)

                else:
                    deltaE = es_energies - es_energies[current_es_state]
                    es_acs = np.array(xm.getCISAbsorptionCrossSections(current_es_state))[0]

                    # Select only values more than or equal to cutoff
                    # TODO: do this only at i=0 to get same number of states over time?
                    selection_deltaE = (deltaE >= opts.emin_abs)

                    deltaE_vals.append(deltaE[selection_deltaE])
                    a_vals.append(es_acs[selection_deltaE])

        if opts.movie == True:
            if not opts.use_excited_state:
                # For Koopmans' theorem, select hole MO.
                # Plus 1 is done since MO counting is XMOLECULE starts from 1.
                index_movie_mo = hole_ix + 1
            else:
                # For excited states, select requested MO.
                index_movie_mo = opts.es_movie_mo

            xm.write_mo_cube(index_movie_mo)

            p = sp.Popen("mv mo.cube cubes/mo_t-" + '{0:04d}'.format(i) + ".cube", shell=True, executable="bash")
            p.wait()


    if opts.absorptionCS:
        write_absorption_cs(opts, t_vals, np.array(deltaE_vals), np.array(a_vals), acs_mo_list, pc_file_suffix)

    if opts.orbitalEnergies:
        write_orbital_energies(opts, t_vals, orbital_e_vals, pc_file_suffix)

    if opts.es_energy_states:
        write_es_energies(opts, t_vals, np.array(es_e_vals), pc_file_suffix)

    if opts.hmid:
        hole_pop_file.close()
        hole_center_file.close()

    if opts.cmid:
        charge_file.close()
        charge_center_file.close()

if __name__ == "__main__":
    start()
