#!/usr/bin/env python

#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************


# ---------------------------------------------------------------------------------
# Tool to calculate core orbital energies for (H2O)n+ trajectories
# ---------------------------------------------------------------------------------

import os
import sys
from optparse import OptionParser

import numpy as np

import CDTK.Tools.Utils as uti
import CDTK.Tools.Conversion as cv

import CDTK.Interfaces.XMoleculeInterface as xmol
import pyxmol as XM

def start():
    # --------------------------------------------------------------------------
    # Parse command line options
    # --------------------------------------------------------------------------
    parser=OptionParser()

    parser.add_option('--gto',
            dest='gto',
            type='string',
            default='STO-3G',
            help='Gaussian basis set in XMOLECULE_BASIS_DIR')

    parser.add_option('--dt',
            dest='dt',
            type='float',
            default=10.0,
            help='Time step to calculate aCS along trajectory [fs]')

    parser.add_option('--tmax',
            dest='tmax',
            type='float',
            default=150.0,
            help='Maximal time to consider [fs]')

    parser.add_option('--natom',
            dest='natom',
            type='int',
            default=36,
            help='Atoms to consider')

    parser.add_option('--embed',
            dest='embed',
            action='store_true',
            default=False,
            help='Whether to include atoms beyond #natom as point charges. Default: False')

    parser.add_option('--emnmol',
            dest='emnmol',
            type='int',
            default=116,
            help='Number of molecules considered for embedding as point charges')

    parser.add_option('-q','--quiet',
            action='store_false',
            dest='verbose',
            default=True,
            help='Specify print level')

    opts, args = parser.parse_args(sys.argv[1:])

    # --------------------------------------------------------------------------
    # Program starts here
    # --------------------------------------------------------------------------

    # set general xmolecule options
    xmOpts={'HF':['yes'],
        'gto':[ opts.gto ],
        'MOM':['yes'],
        'gs_occ':['yes'],
        'SIni':['yes'],
        'RIni':['no'],
        'N':['1'],
        'charge':['1']}

    # read atomlist
    alist = uti.readColumnFile('atomlist')[:opts.natom]
    natoms = len(alist)
    # to speed up compromise on embedding region size
    last_embed_ix = 3*opts.emnmol
    alist2 = uti.readColumnFile('atomlist')[opts.natom:opts.natom + last_embed_ix]
    nembed = len(alist2)

    # embed: point charges from SPC model
    spc_charges = []
    for label in alist2:
        if label=='O':
            spc_charges.append( 0.8476 )
        elif label=='H':
            spc_charges.append( -0.4238 )

    # Note: this is true only for (H2O)n trajectories
    nmols  = natoms // 3
    homo_ix = 5 * nmols - 1
    ncore = nmols

    # read trajectory time steps
    t = uti.getTrajectoryTimeSteps('R.log') * cv.au2fs
    dt_traj = (t[1] - t[0])
    t = t[t <= opts.tmax]
    if opts.dt < dt_traj:
        step = 1
    else:
        step = opts.dt // dt_traj
    nsteps = len(t) // step

    # energy differences and absorption cross-sections
    t_vals = np.zeros(nsteps)
    e_vals = np.zeros([nsteps, 5*ncore])

    # read trajectory electronic states
    states = uti.readColumnFile('S.log',col_ix=1)
    states = list(map(int, states)) 

    # read trajectory xyz coordinates
    r_xyz_old = uti.atomicCoordinatesFromRlog('R.log')
    r_xyz = r_xyz_old[:,:opts.natom]
    r2_xyz = r_xyz_old[:,opts.natom:opts.natom + last_embed_ix]

    # initialize xmolecule interface
    xm=xmol.xmolecule()
    if opts.embed:
        xm.init_input_options(xmOpts, alist, r_xyz[0].flatten(), x_pos=r2_xyz[0].flatten(), x_charges=spc_charges)
    else:
        xm.init_input_options(xmOpts, alist, r_xyz[0].flatten())

    # From now on use RIni/SIni scheme
    xmOpts['RIni']='yes'
    occs=xm.get_occ()
    # This line somehow fixes a bug when setting new_occs below
    new_occs = np.zeros(len(xm.get_occ()))

    # calculate quantities along trajectory
    for i in range(nsteps):

        hole_ix = homo_ix - states[i*step]

        new_occs = [ 0 if occ==0 else 2 for occ in occs]
        new_occs[ hole_ix ] = 1

        xm.set_occ( new_occs )

        new_r = r_xyz[i*step].flatten()

        if opts.embed:
            new_r = np.append(new_r, r2_xyz[i*step].flatten())
        xm.set_geom( new_r )
        
        moes = np.asarray( xm.getOrbitalEnergies() ) * cv.au2ev

        t_vals[i] = t[i*step]
        e_vals[i] = moes[:5*ncore]

    # write output files
    # absorption cross section
    with open('core_orbital_energies.dat','w') as f:
        f.write('#  t[fs] MOE [eV]')
        f.write('\n')

        for i in range(nsteps):
            f.write('  {0:.4f}'.format(t_vals[i]))
            for j in range(5*ncore):
                f.write('  {0:.4f}'.format(e_vals[i,j]))
            f.write('\n')

if __name__ == "__main__":
    start()