#!/usr/bin/env python
#*  **************************************************************************
#*
#*  CDTK, Chemical Dynamics Toolkit
#*  A modular system for chemical dynamics applications and more
#*
#*  Copyright (C) 2011, 2012, 2013, 2014, 2015, 2016
#*  Oriol Vendrell, DESY, <oriol.vendrell@desy.de>
#*
#*  Copyright (C) 2017, 2018, 2019
#*  Ralph Welsch, DESY, <ralph.welsch@desy.de>
#*
#*  Copyright (C) 2020, 2021, 2022, 2023
#*  Ludger Inhester, DESY, ludger.inhester@cfel.de
#*
#*  This file is part of CDTK.
#*
#*  CDTK is free software: you can redistribute it and/or modify
#*  it under the terms of the GNU General Public License as published by
#*  the Free Software Foundation, either version 3 of the License, or
#*  (at your option) any later version.
#*
#*  This program is distributed in the hope that it will be useful,
#*  but WITHOUT ANY WARRANTY; without even the implied warranty of
#*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#*  GNU General Public License for more details.
#*
#*  You should have received a copy of the GNU General Public License
#*  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#*
#*  **************************************************************************
"""
This module has several functions to transform coordinates and velocities obtained
from Gromacs calculation into suitable input files for CDTK. Use ``gmx2cdtk -h`` 
to find out how to use this program.


Basic units in GMX, please look at the following web.
https://manual.gromacs.org/documentation/2019/reference-manual/definitions.html#table-basicunits
  
1 Picoseconds = 41341.374575751 Atomic Unit Of Time
1 nm = 18.8973 Bohr
1 nm/ps = 18.8973/41341.374575751 Bohr/a.u. = 0.00045710381413114194
1 nm/ps = 4.571038E-04 Bohr/a.u.
"""

import subprocess
from optparse import OptionParser

import numpy as np
import CDTK.Tools.Conversion as cv
from gmx2cdtk import get_coord_center, atomName2atomNum, atomName2atomMass, \
    atomName2Element, shift_box, sort_molec, print_geom_files



def get_parser():
    parser = OptionParser()
    parser = OptionParser(usage="usage: %prog -x filename.xyz [further options]")
    parser.add_option(
        "-x",
        "--xyz_file",
        dest="xyzFile",
        type="str",
        default=None,
        help=" .xyz input file.",
    )

    parser.add_option(
        "-n",
        "--cen_mol",
        dest="centeredMolecule",
        type="int",
        default=1,
        help="Please enter the index of the molecule that you want to center. It is an integer number and not exceeding the number of water molecules in the box.",
    )

    parser.add_option(
        "--cen_atoms",
        dest="centerAtoms",
        type=str,
        default=None,
        help="Please enter a list of atom indices that you want to center (comma seperated).",
    )

    parser.add_option(
        "-s",
        "--skip",
        dest="skip",
        type="float",
        default=None,
        help="Skip molecules (residues) with center of mass distance from the center large than this value. Value is given in atomic units",
    )

    parser.add_option(
        "--centerMethod",
        type="choice",
        dest="centerMethod",
        default="first",
        choices=["first", "com", "None"],
        help="How to center: according to first atom of residue (first) or center of mass of residue (com) or None",
    )

    parser.add_option(
        "-o",
        type="str",
        dest="xyzOut",
        default=None,
        help=" .xyz output file (optional)",
    )

    parser.add_option(
        "--noreorder",
        action='store_false',
        dest="doReorder",
        default=True,
        help="do not reorder the molecules according to distance (optional, default: do sort the molecules)",
    )
    
    parser.add_option(
        "--noshift",
        action='store_false',
        dest="doShift",
        default=True,
        help="do not shift the molecule's postions",
    )
    return parser


oxygenNames = ['O', 'OW', 'OD', 'O_', 'OS']
hydrogenNames = ['H', 'HW', 'HW1', 'HW2', 'HD', 'HC']
carbonNames = ['C', 'CA', 'CR', 'CW', 'C_', 'CS']

def readxyz(filename, check_water=False):
    """
    Read xyz file and return the topology (guessed), element types, coordinates,
    velocities of all atoms and the box size.

    Parameters
    ----------
    filename : string
        Name of xyz file.

    Returns
    -------
    molec : dict of ndarray.
        Dictionary containing the following arrays:
        - geom : (n_atoms, 3) ndarray of float.
            Coordinates of all atoms.
        - vel : (n_atoms, 3) ndarray of float.
            Velocities of all atoms.
        - residueIndex: (n_atoms) ndarray of ints
        - residueName: (n_atoms) ndarray of str
        - topol : (n_atoms) ndarray of str.
            String identifiers for the topology of the system (atoms in molecules).
        - atomlist : (n_atoms) ndarray of str.
            String identifiers of all atoms types (e.g., "O", "H").
        - atomnums : (n_atoms) ndarray of int.
            Intigers with atomic numners of each element in atomloist.
        - atommass : (n_atoms) ndarray of float.
            Atomic masses of each element in atomlist.
        - boxsize : (3) ndarray of float.
            Box size (len_x, len_y, len_z).
        - residueName : (n_atoms) ndarray of str
            Name of residue for each atom 
        - residueIndex : (n_atoms) ndarray of int
            Index of residue for each atom
        - atomName : (n_atoms) ndarray of str
            Name of atoms (e.g. "OW", "HW1")
   """
    # the .gro file format has fixed column positions as delimiters ("%5d%-5s%5s%5d%8.3f%8.3f%8.3f%8.4f%8.4f%8.4f\n")
    atomName = []
    xyz = []
    with open(filename) as f:
        natoms = int(f.readline())
        comment = f.readline()
        for n in range(0,natoms):
            l = f.readline().split()
            atomName.append(l[0])
            xyz.append(np.array([float(l[1]), float(l[2]), float(l[3])]) * cv.an2au)
        f.close()
    atomName = np.array(atomName, dtype="<U5")
    xyz = np.array(xyz)
    residueName = np.empty((natoms), dtype="<U5")
    residueIndex = np.repeat(0, natoms)
    topol = np.empty((natoms),dtype=str)
    isWater = False
    resIdx = 0
    i = 0
    while i < natoms:
        if i < natoms + 2:
            if atomName[i] == "O" and atomName[i+1] == "H" and atomName[i+2] == "H":
                if np.linalg.norm(xyz[i,:] - xyz[i+1,:]) > 3.0:
                    print(f"Warning atom {i} and {i+1} are more than 3.0 au. apart")
                    print(xyz[i,:] - xyz[i+1,:])
                if np.linalg.norm(xyz[i,:] - xyz[i+2,:]) > 3.0:
                    print(f"Warning atom {i} and {i+2} are more than 3.0 au. apart")
                    print(xyz[i,:] - xyz[i+2,:])
                resIdx = resIdx + 1
                atomName[i] = "OW"
                atomName[i+1] = "HW1"
                atomName[i+2] = "HW2"
                for j in range(i,i+3):
                    residueName[j] = f"SOL"
                    residueIndex[j] = resIdx
                i = i + 3
            else:
                resIdx = resIdx + 1
                residueName[i] = f"UKNWN"
                residueIndex[i] = resIdx
    topol = np.array([str(ri) + rn.strip() + "_" + aN.strip()
                for ri, rn, aN in zip(residueIndex, residueName, atomName)])  # topology with cdtk format
    vxyz = np.zeros(xyz.shape)
    v_box = [np.nan, np.nan, np.nan]

    # Get atomic numbers, atomic masses, and element names
    atomnums = np.array(list(map(atomName2atomNum, atomName)))
    atommass = np.array(list(map(atomName2atomMass, atomName)))
    atomlist = np.array(list(map(atomName2Element, atomName)))
    molec = {
        "geom": xyz,
        "vel": vxyz,
        "topol": topol,
        "atomlist": atomlist,
        "atomnums": atomnums,
        "atommass": atommass,
        "boxsize": v_box,
        "residueIndex": residueIndex,
        "residueName": residueName,
        "atomName": np.array([a.replace('_','') for a in atomName]),
    }
    return molec




def start():
    parser = get_parser()
    args, _ = parser.parse_args()
    if args.xyzFile is None:
        parser.error("You need to provide a .xyz file!")
    molec_data = readxyz(args.xyzFile)

    # Get center of box and shift coordinates.
    if args.centerAtoms is not None:
        centerAtoms = [int(item) for item in args.centerAtoms.split(',')]
        boxCenter = get_coord_center(
            centerAtoms, molec_data, centerMethod=args.centerMethod)
        keepOrder = centerAtoms
    else:
        boxCenter = get_coord_center(
            args.centeredMolecule, molec_data, centerMethod=args.centerMethod)

    if args.doShift:
        shifted_geom = shift_box(
            boxCenter, molec_data["geom"], molec_data["boxsize"])
        molec_data['geom'] = shifted_geom

    # sort molecules according to distance
    if args.doReorder:
        molec_data = sort_molec(molec_data, skipDistance=args.skip, keepOrder=keepOrder)
    ekin = 0.5 * np.sum(molec_data["vel"][:,:]**2 * cv.am2au * molec_data["atommass"][:,np.newaxis])
    print(f"kinetic Energy: {ekin} a.u.")
    print_geom_files(molec_data)

if __name__ == "__main__":
    start()
