#!/bin/bash
# -----------------------------------------------------------------------------
# This script runs XATOM with all possible configurations for a given atom.
# - Only for 5 <= Z <= 10, but it is readily extensible for other cases.
# - This will generate individual outputs for each configuration.
# - This will generate one data file collecting all atomic data.
# - This is /bin/bash because of the C-style do-loop: for (( .. ; .. ; .. ))
#   but it can be run with /bin/sh using "seq" (see below)
# -----------------------------------------------------------------------------

#
# set up the system
#
SYSTEM="Ne";  Z=10      # neon (Z=10)
PE=2000                 # photon energy in eV for cross section calculation

#
# set up XATOM
# - XATOM_EXE: full PATH of the XATOM excutable
# - XATOM_OPT: options with the system setting above
#              other options taken from the default setup as mush as possible
# - XATOM_DATA_FILE: atomic data file (to be used for xatom_req and xatom_mc)
#                    with the above example, the file name will be
#                    "Ne_process_2000eV.dat"
# - XATOM_INPUT: if you want to use an input file, then specify it here
#
XATOM_EXE=../../src/xatom
XATOM_OPT="-Z $Z -pcs -PE $PE -decay"
XATOM_DATA_FILE="${SYSTEM}_process_${PE}eV.dat"
[ -f $XATOM_DATA_FILE ] && rm -f $XATOM_DATA_FILE  # always start from empty
XATOM_INPUT=""

#
# do loop for occupation numbers in 1s, 2s, and 2p subshells
# NOTE: /bin/bash vs. /bin/sh
# - do-loop
#   bash: for (( i = [start] ; i <= [end] ; i += [increment] ))
#   sh:   for i in `seq [start] [increment] [end]`
# - if-then
#   bash: [ A == B ] && ...
#   sh:   [ A -eq B ] && ...
#
max_2p=$(( $Z - 2 - 2 ))                        # max. occupation number in 2p
for (( q = 0; q <= $Z; q++ )); do               # scan over q ( 0 <= q <= Z )
  for (( n_1s = 2; n_1s >= 0; n_1s-- )); do     # occupation number in 1s
  for (( n_2s = 2; n_2s >= 0; n_2s-- )); do     # occupation number in 2s
  for (( n_2p = $max_2p; n_2p >= 0; n_2p-- )); do  # occupation number in 2p
    n_elec=$(( $n_1s + $n_2s + $n_2p ))         # calculate N of electrons
    [ $n_elec == 0 ] && continue                # skip if there is no electron
    charge=$(( $Z - $n_elec ))                  # calculate the charge
    [ $charge != $q ] && continue               # only run below when charge==q
    config="1s${n_1s}_2s${n_2s}_2p${n_2p}"      # set up electron configuration
    FILE="${SYSTEM}_${charge}+_${config}.out"   # filename for each config.
    echo -n "Running for $FILE ... "
    $XATOM_EXE $XATOM_INPUT $XATOM_OPT -config $config -dat $XATOM_DATA_FILE -o $FILE
    echo "done"
  done  # n_2p
  done  # n_2s
  done  # n_1s
done    # q

echo "$XATOM_DATA_FILE has been generated."
