!----------------------------------------------------------------------------
! cfelt_bessel.f90
!----------------------------------------------------------------------------
! Copyright (C) 2015-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2015-2016 Otfried Geffert
!          2017      John Bekx
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
module cfelt_bessel

use constants
implicit none

contains

subroutine cfelt_spherical_bessel(lmax, Q, r, N, jmat)
  !******************************************************************************
  ! The subroutine cfelt_spherical_bessel computes a matrix jmat, where
  ! the columns jmat(:,l) are the values of the l-th spherical bessel
  ! function at positions Q*r
  !******************************************************************************
  implicit none
  integer,    intent(in) :: lmax, N
  real(long), intent(in) :: Q
  real(long), intent(in) :: r(:)
  real(long), dimension(N,lmax), intent(out) :: jmat

  real(long), allocatable, save :: jmatsave(:,:)
  real(long), allocatable, save :: xvec(:)
  integer,    save :: lmaxsave, Nsave, Nlarge
  logical,save :: firstcall = .true.

  real(long),dimension(lmax) :: jvec
  real(long) :: x, x1, x2, y1, y2, delta_y
  integer :: i, f, l, at
  logical :: found

  real(long), dimension(N)      :: extra_x
  real(long), dimension(N,lmax) :: extra_jvec
  
  integer, save :: computed, reused
  logical :: Lentz = .true.
  real(long) :: smallest_x

  if( firstcall ) then
    !write(57,*) 'firstcall N=',N,' lmax', lmax,' Q=',Q
    call first_run
  else
      smallest_x = Q*r(1)
      if( (N.eq.Nsave) .and. (lmax.eq.lmaxsave) .and. (smallest_x .ge. xvec(1)) ) then
        !write(57,*) 'othercall Nlarge=',Nlarge,' lmax', lmax,' Q=',Q
        call other_runs !with interpolation
      else
        ! results from prev run cannot be used --> destroy them
        !write(57,*) 'othercall deallocate :-('
        deallocate(xvec)
        deallocate(jmatsave)
        call first_run
      end if
  end if
 
contains

  subroutine first_run
    firstcall = .false. ! -----> at next call use optimization
    computed = 0
    Nlarge = N
    Nsave = N
    lmaxsave = lmax
    allocate(xvec(N))
    allocate(jmatsave(N,lmax))

    do i = 1, N
      x = Q * r(i)
      xvec(i) = x 
      computed = computed + 1
      if ( x == ZERO ) then
        jmat(i,:) = ZERO
        jmat(i,1) = ONE
        jmatsave(i,:) = jmat(i,:)
        cycle
      endif
      if ( Lentz ) then
        call cfelt_j(lmax, x, jvec)
      else
        call zhangjin_sphj(lmax, x, jvec)
      endif
      jmat    (i,:) = jvec
      jmatsave(i,:) = jvec
    end do  
  end subroutine

  subroutine other_runs
    integer :: newcalls 

      ! the stop statements: checking _all_ conditions for optimizations:

      newcalls = 0 ! do not do this with integer :: newcalls = 0
                       ! since this will be done ONCE only, so it holds OLD
                       ! values accumulated -----> long debugging sessions
      at = 1
      reused   = 0
      all_x: do i = 1, N
        x = Q * r(i)
        found = .false.
        find: do f = at ,Nlarge

          if(f.gt.1) then
            if(xvec(f).lt.xvec(f-1)) stop 'cfelt_bessel: optimization failed - r-array must be monotonously increasing'
          endif

          !try: fill jvec()

          if( x .gt. xvec(f) ) then
            ! keep on searching
            cycle 
          endif

          if( x .lt. xvec(f) ) then
            ! found next higher value: interpolate
            !assumes monotonously increasing x values at interpolation !!!
            if (f == 1) then
              write(STDERR,*) "x=", x, " is smaller than xvec(1) "
              write(STDERR,*) "xvec(1)=", xvec(1)
              stop 'cfelt_bessel: optimization failed - r-array must be monotonously increasing'
            endif
            x1 = xvec(f-1)
            x2 = xvec(f)
            do l=1,lmax
              y1 = jmatsave(f-1,l)
              y2 = jmatsave(f,l)
              delta_y = y2-y1
              jvec(l) = y1 + delta_y * ((x-x1)/(x2-x1))
              !jvec(l) = jmatsave(i,l) !simply copy
            end do
            found = .true.
            at = f
            exit find
          endif

          if(x .eq. xvec(f)) then
            ! equality: simply copy values
            do l=1,lmax
              jvec(l) = jmatsave(i,l) 
            end do
            found = .true.
            at = f
            exit find
          endif

        end do find

        if( found ) then
          if( reused .eq. 0 ) then
            !write(57,*) 'computed: ', computed
            computed = 0
          end if
          reused = reused +1
          jmat(i,:) = jvec
        else
          if( computed .eq. 0 ) then
            !write(57,*) 'reused ,interpolated: ', reused
            reused = 0
          endif
          computed = computed +1 
          ! x outside interpolation interval 
          ! so: 
          ! compute a NEW jvec array (1..lmax)
          if( Lentz ) then
            call cfelt_j(lmax, x, jvec)
          else
            call zhangjin_sphj(lmax, x, jvec)
          endif
          jmat(i,:) = jvec
          newcalls = newcalls + 1
          at = Nlarge+newcalls

          ! save x & jvec
          extra_x(newcalls)      = x
          extra_jvec(newcalls,:) = jvec
        end if

      end do all_x

      if( newcalls .ge. 1) then
         !write(57,*) 'newcalls=',newcalls

         ! enlarge xvec & jmatsave  BY newcalls:
         ! that means our tables of arguments & bessel function values get bigger

         ! append extra_x to xvec
         call array_append( xvec, extra_x, newcalls )

         ! append extra_jvec to jmatsave
         call array_append_dim1( jmatsave, extra_jvec, newcalls )

         Nlarge = Nlarge + newcalls
      end if

  end subroutine

end subroutine cfelt_spherical_bessel

subroutine array_append( arr, append, append_len )
  implicit none
  real(long), allocatable :: arr(:)
  real(long)              :: append(:)
  integer                 :: append_len

  real(long), allocatable :: tmp_arr(:)
  integer                 :: length 

  length = size(arr)
  allocate(tmp_arr(length))
  tmp_arr = arr
  deallocate(arr)
  allocate(arr(length+append_len))

  arr(:length) = tmp_arr
  arr(length+1:) = append
  deallocate(tmp_arr)

end subroutine array_append

subroutine array_append_dim1( arr, append, append_len )
  implicit none
  real(long), allocatable :: arr(:,:)
  real(long)              :: append(:,:)
  integer                 :: append_len

  real(long), allocatable :: tmp_arr(:,:)
  integer                 :: length_1, length_2

  length_1 = size(arr,1)
  length_2 = size(arr,2)
  allocate(tmp_arr(length_1,length_2))
  tmp_arr = arr
  deallocate(arr)
  allocate(arr(length_1+append_len, length_2))

  arr(:length_1,:) = tmp_arr
  arr(length_1+1:,:) = append
  deallocate(tmp_arr)

end subroutine array_append_dim1

subroutine zhangjin_sphj( lmax, z, j )
  implicit none
  integer, intent(in)     :: lmax
  real(long), intent(in)  :: z
  real(long), intent(out) :: j(:)

  real(long),dimension(lmax) :: jprime
  real(long)                 :: largest
  integer                    :: l, max_computed

  call sphj( lmax-1, z, max_computed, j, jprime )

  ! j values may be infinity or NaN
  largest = huge(z);
  do l=1,lmax
    if( j(l) > largest )  then
      write(STDERR, *) 'ERROR: spherical Bessel function sphj(',lmax-1,',',z, '): yields infinity= ',j(l)
      stop
    end if
    if( j(l) /= j(l) )  then
      write(STDERR, *) 'ERROR: spherical Bessel function sphj(',lmax-1,',',z, '): yields NaN=',j(l)
      stop
    endif
  end do

end subroutine zhangjin_sphj

subroutine cfelt_j( lmax, z, j )
  implicit none
  integer, intent(in) :: lmax
  real(long), intent(in) :: z
  real(long), intent(out) :: j(:)
  integer,  dimension(0:lmax-1) :: n_arr
  integer  :: i

  do i = 0,lmax-1 ! fill n_arr with indices
    n_arr(i) = i
  end do

  j = jn( n_arr, z )

end subroutine cfelt_j


elemental function jn( n, z) result(j)
  ! spherical bessel functions j
  ! implemented by Otfried Geffert
  !  2014/09/25
  ! -----------------------------
  ! according to the three papers
  ! of H.W. Lentz 1973...1990
  integer,    intent(in) :: n 
  real(long), intent(in) :: z
  real(long)  :: x
  real(long) :: j, j0, b, bmul, bdiv
  integer :: m 
  integer :: k
  logical :: divflag, mulflag

  divflag = .true.
  mulflag = .true.
  x = z
  m = 0
  j0 = sin(x)/x

  if( n .eq. 0 ) then
    j = j0
    return !....................n=0
  end if

  j = 1
  bmul = 0
  bdiv = 0

  do k=1,n
    m = m+1
    b = (2/x) * ( 1 + m -0.5)

    bmul = b - bmul
    !write(*,*) bmul
    j = j*bmul
    bmul = 1/bmul
  enddo

  do
    m = m+1
    !b  = (2/x) * ( n + m -0.5)
    b  = (2/x) * ( 1 + m -0.5)
    !write(*,*) 
    !write(*,*) b

    bmul = b - bmul
    !write(*,*) bmul

    if( mulflag ) then
      if( bmul .ne. 0 ) then
        j = j*bmul
        bmul = 1/bmul
      else
        mulflag = .false.
        bmul = 0.0
        j = -j
      end if
    else
      mulflag = .true.
      bmul = 0.0
    end if

    bdiv = b - bdiv
    !write(*,*) bdiv

    if( divflag ) then
      if( bdiv .ne. 0 ) then
        j = j/bdiv
        bdiv = 1/bdiv
      else
        divflag = .false.
        bdiv = 0.0
        j = -j
      endif
    else
      divflag = .true.
      bdiv = 0.0
    end if

    !write(*,*) m, j0/j
    !pause
    
    if( .not. divflag ) cycle
    if( .not. mulflag ) cycle

    if( bmul .eq. bdiv )  exit
    !if(abs(bmul-bdiv) < 1E-6) exit
  end do
  !write(*,*) m
  j = j0/j
  return 

end function jn

end module cfelt_bessel
