!----------------------------------------------------------------------------
! compton.f90
!----------------------------------------------------------------------------
! Copyright (C) 2013-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2013-2014 Jan Malte Slowik
!          2014-2016 Otfried Geffert
!          2017      John Bekx
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
module Compton
!******************************************************************************
! Module to calculate the double differential scattering cross section of 
! Compton scattering. This module is based on notes by Gopal Dixit, 
! Manish Jung Thapa, and Robin Santra.
! Written by Jan Malte Slowik, July 2013.
! Modified by John Bekx, May 2017 - September 2017.
! Added the nonuniform section.
!******************************************************************************
  use constants
  use read_param
  use interpolate
  use wavefunction
  use spherical_bessel
  use cfelt_bessel
  use tindex
  implicit none

  contains

subroutine print_total_compton_cross_section( P, F )

  !# Arguments------------------------------------  
  type(Param), intent(in)      :: P
  type(WF_data), intent(inout) :: F
  !type(Coordinate), intent(in) :: G

  !# local variables------------------------------
  integer :: lmax
  integer :: N_theta, N_omega !number of grid points for angle and energy grid
  real(long) :: d_theta, d_omega !stepsize of grid
  real(long) :: min_theta ! starting point of grid

!  type(Spline)                 :: SP
  integer                      :: OUTPUT

  real(long)                   :: omega_in
  real(long),allocatable,dimension(:) :: omega_s(:)
  real(long),allocatable,dimension(:) :: theta(:)

  integer :: i, j, m
  real(long), dimension(F%N_uni, (P%compton_lmax+1)*F%N_orb) :: cont_wf
  real(long), dimension(P%compton_N_omega, P%compton_N_theta, F%N_orb) :: DDSP
  real(long), dimension(P%compton_N_omega, P%compton_N_theta) :: DDSP_sum
  real(long), dimension(P%compton_N_theta) :: S_sum
  real(long), dimension(P%compton_N_theta, F%N_orb) :: S
  real(long), dimension(P%compton_N_theta) :: S_total
  real(long) :: total_CS
  real(long), dimension(P%compton_N_theta) :: sin_theta, scat_fac
  character(len=256) :: file_energy_and_angle, file_energy_integrated, file_total
  real(long), dimension(P%compton_N_theta,F%N_orb) :: b2b

  lmax = P%compton_lmax
  N_theta = P%compton_N_theta
  N_omega = P%compton_N_omega
  d_theta = P%compton_d_theta
  d_omega = P%compton_d_omega
  min_theta = P%compton_min_theta


  OUTPUT = P%OUTPUT
  !do not do a 2nd time: call interpolate_potential( F )
!  SP = F%SP_pot


  !# create energy grid
  omega_in = P%PE_start
  d_omega = dble(d_omega) * EV2AU   !set energy grid spacing
  allocate(omega_s(N_omega))
  omega_s(:) = (/ ( omega_in - dble(j) * d_omega , j=1, N_omega ) /)

  
  d_theta = dble(d_theta) * PI/180.d0   ! set angle grid spacing
  min_theta = dble(min_theta) * PI/180.d0
  allocate(theta(N_theta))
  theta(:) = (/ ( min_theta + dble(j) * d_theta, j=0, N_theta-1 ) /)

  
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module with the full calculation of the doubly differential scattering cross section (DDSCS).'
  write(OUTPUT,"('# ')")

  !#
  !# Calculate the double-differential scattering probability on energy and angle grid
  !# In the loop over the angle grid, the same continuum wavefunction may be used!
  !#  
  
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# Calculating Compton Cross section:')")
  write(OUTPUT,"('# The incoming photon energy is: ',F7.1,' eV')") omega_in/EV2AU
  write(OUTPUT,"('# Using ', I3,' energy steps of ', F4.1,' eV and ', I3,' angular steps of ', F4.1,' degrees.')") N_omega, d_omega/EV2AU, N_theta, d_theta*180./PI
  write(OUTPUT,"('# lmax is ', I3,'.')") lmax
  write(OUTPUT,"('# Now proceeding with energy step:')")
  write(OUTPUT,"('# ....5...10...15...20...25...30...35...40...45...50')")
  write(OUTPUT,"(A$)") '# '
  energy_grid: do j = 1, N_omega
    write(OUTPUT,"(A$)") '*'
    if ( mod(j,50) == 0 ) write(OUTPUT,"(/'# '$)")
    ! create continuum wavefunctions
    cont_wf = ZERO
    cont_wf = get_continuum_wf(F, P, omega_s(j), lmax)
    !
    angle_grid: do m = 1, N_theta
      DDSP(j,m,:) = calculate_DDSP(P, F, cont_wf, theta(m), omega_s(j), lmax)
      !DDSP(j,m,:) = calculate_DDSP_nonlinear(P, F, G, cont_wf, theta(m), omega_s(j), lmax)
    end do angle_grid
  end do energy_grid
  write(OUTPUT,"(' '/'# ')")

  !# print the DDSP for each orbital to a file, such that in every file one has the compton spectrum for a given 
  !# scattering angle in the 2:N_theta+1 columns 
  DDSP_sum(:,:) = ZERO
  do i=1, F%N_orb
  file_energy_and_angle = trim(P%filename)//'.ddsp_'//trim(WF_name(F,i))      
    open(unit=111, file=file_energy_and_angle, status='replace', action='write')
    write(111,"('#',a13,900f13.2)") 'theta (deg)', theta(:)*180./Pi
    write(111,"('#',10x,a)") 'First column: photon energy. Each other column contains the spectrum for a given theta.'
     write(111,"('#',a13,5x,a)") 'omega_s','Spectrum'
    do j = 1, N_omega
    write(111,'(900ES13.3)') omega_s(j), DDSP(j,1:N_theta,i)
    DDSP_sum(j,1:N_theta) = DDSP_sum(j,1:N_theta) + DDSP(j,1:N_theta,i)
    end do
    close(unit=111)
  end do

  ! print DDSP_tot (DDSP_sum)
  file_energy_and_angle = trim(P%filename)//'.ddsp_tot'
  open(unit=111, file=file_energy_and_angle, status='replace', action='write')
  do j = 1, N_omega
    write(111,'(900ES13.3)') omega_s(j), DDSP_sum(j,1:N_theta)
  end do
  close(unit=111)

  !return !!!!!!!!!! Sangkil had introduced this

  write(OUTPUT,"('# ',a)") 'The dynamic structure factor for each orbital are written to the files "DDSP_energy_angle_ORBITALNAME.dat"'
  write(OUTPUT,"('# ',a)") 'The first column in these files is the photon energy, every other column contains the Compton spectrum corresponding to given theta.'
  write(OUTPUT,"('# ')")

  !# bound-to-bound transitions, depend only on angle-grid.
  do m=1, N_theta
    b2b(m,:) = calculate_bound2bound(P, F, theta(m))
  end do
  
  !#
  !# Calculate total scattering probability for all orbitals integrated over energies and angles.
  !#
  
  S_total=ZERO

  !#
  !#Energy integral
  !#
  do i= 1, F%N_orb
    do m = 1, N_theta
      S(m,i) = sum( DDSP(1:N_omega,m,i) )
      S(m,i) = S(m,i) - HALF * DDSP(N_omega,m,i) - HALF * DDSP(1,m,i)           ! according to trapezoid rule
      S(m,i) = S(m,i) * d_omega
      S_total(m) = S_total(m) + S(m,i)          !S_total is gives the total scattering probability for a given angle, i.e. the sum of S(m,i) over all subshells
    end do
  ! print the scattering probability for each orbital to a file, 
  ! such that the column gives the probability scattering a photon of any energy to this angle.  
    file_energy_integrated = trim(P%filename)//'.ddspi_'//trim(WF_name(F,i))      
    open(unit=112, file=file_energy_integrated, status='replace', action='write')
    write(112,"('#',99a13)") 'theta', 'Thomson_CS', 'S(theta)', 'bound2bound'
    do m = 1, N_theta
      write(112,'(99ES13.3)') theta(m), thomson_CS(theta(m)), S(m,i), b2b(m,i)
    end do
    close(unit=112)
  end do

  write(OUTPUT,"('# ',a)") 'The energy integrated DDSCS (static structure factors) for each individual orbital is written &
    &to the file "DDSP_integrated_energy_ORBITALNAME.dat" '
  write(OUTPUT,"('# ',a)") 'The columns in these files are scattering angle (theta), Thomson cross section, S(theta), and in the last column the contribution to'
  write(OUTPUT,"('# ',a)") 'S(theta) coming from the bound-to-bound transitions.'
  write(OUTPUT,"('# ')")

  ! print S_total to file
  file_total = trim(P%filename)//'.ddspi_tot'      
  open(unit=113, file=file_total, status='replace', action='write')
  write(113,"('#',99a13)") 'theta', 'Thomson_CS', 'S(theta)', 'bound2bound', 'free e-   ', 'WallerHartree'
    do m = 1, N_theta
      !write(112,'(99ES13.3)') theta(m), thomson_CS(theta(m)), S_total(m)
      write(113,'(99ES13.3)') theta(m), thomson_CS(theta(m)), S_total(m), sum( b2b(m,1:F%N_orb) ), free_electron_background(F, theta(m)), calculate_waller_hartree(P,F,theta(m),lmax)
    end do
  close(unit=113)

  write(OUTPUT,"('# ',a)") 'The total energy integrated DDSCS (total static structure factors) for all orbitals'
  write(OUTPUT,"('# ',a)") 'is written to the file "DDSP_integrated_energy_total.dat".'
  write(OUTPUT,"('# ',a)") 'The columns in these files are theta, Thomson_CS , S(theta), bound-to-bound contributions,'
  write(OUTPUT,"('# ',a)") 'the background from free electrons, and the result from the waller-hartree calculation.'
  write(OUTPUT,"('# ')")

  ! write results to .ixs file for rate equations
  if ( P%yes_xmdyn ) then
    write( P%IO_IXS, "(a,i10)" ) "ixs", N_theta ! P%IO_IXS=STDOUT for xmdyn
  else
    open( P%IO_IXS, file=P%file_ixs, access="append", form="formatted" )
    call config_to_index( F%config, i )         ! i: index of configuration
    write( P%IO_IXS, "('#',2i10)" ) i, N_theta        
  end if
  do m=1, N_theta
    scat_fac(m)= S_total(m) + sum( b2b(m,1:F%N_orb) )
    !print out to .ixs file: theta in radian and S(Q), i.e. dsigma/dOmega in units of Thomson cross section
    write( P%IO_IXS, "(f23.15,es23.15)" ) theta(m), scat_fac(m)/thomson_CS(theta(m))
  end do
  if ( .not. P%yes_xmdyn ) then
    write( P%IO_IXS, "()" )
    close( P%IO_IXS )
  end if


  ! write results to .ixs file for rate equations INCLUDING FREE ELECTRON BACKGROUND
  open(P%IO_IXS2, file=P%file_ixs2, access="append", form="formatted" )
  !open(unit=114, file='results_with_background.ixs', status='replace', action='write' )
  call config_to_index( F%config, i )   ! i: index of configuration
  write(P%IO_IXS2, "('#',2i10)" ) i, N_theta        
  do m=1, N_theta
    scat_fac(m)= S_total(m) + sum( b2b(m,1:F%N_orb) )
    !print out to .ixs file: theta in radian and S(Q), i.e. dsigma/dOmega in units of Thomson cross section
    write(P%IO_IXS2, "(f23.15,es23.15)" ) theta(m), (scat_fac(m)+free_electron_background(F, theta(m)))/thomson_CS(theta(m))
  end do
  write(P%IO_IXS2, "()" )
  close(P%IO_IXS2)
  
  do m = 1, N_theta
    S_sum(m) = sum( S(m, 1:F%N_orb) ) 
  end do
  
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# Output files for xatom_req are written to: ')")
  write(OUTPUT,"('#   DSP             : ',a)") trim(P%file_ixs)
  write(OUTPUT,"('#   DSP w/ free elec: ',a)") trim(P%file_ixs2)

  !#
  !# angular_integral
  !# 
  sin_theta = (/ (dsin(theta(m) ) , m=1, N_theta) /)
  total_CS = sum( S_sum(1:N_theta)*sin_theta(1:N_theta) )
  total_CS = total_CS - HALF * ( S_sum(N_theta)*sin_theta(N_theta) ) - HALF * ( S_sum(1) * sin_theta(1) )
  total_CS = ( TWO * d_theta * N_theta ) * d_theta * total_CS

!   write(OUTPUT,"('# ')")
!   write(OUTPUT,"('# The total Compton cross section is (in atomic units):')")
!   write(OUTPUT,*) total_CS
!   write(OUTPUT,"('# The total Compton cross section is (in megabarn):')")
!   write(OUTPUT,*) total_CS * AU2M**2 / 1.0e-22_long
!   write(OUTPUT,"('# ')")

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an angular integration over the given theta range from', P%compton_min_theta, ' deg to ', theta(N_theta)*180./PI, ' deg'
  write(OUTPUT,"('# the total Compton cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'


end subroutine print_total_compton_cross_section


subroutine print_compton_spectrum( P, F, theta_deg, unit_no )
  ! Print the compton spectrum for a given scattering angle!

  type(Param), intent(in)      :: P
  type(WF_data), intent(inout) :: F
  real(long), intent(in)       :: theta_deg
  integer, intent(in)          :: unit_no
!  type(Spline)                 :: SP
  integer                      :: OUTPUT

  real(long)                   :: omega_in, d_omega
  real(long)                   :: theta
  real(long),allocatable,dimension(:) :: omega_s(:)
  real(long),allocatable,dimension(:) :: order(:)
  real(long)                   :: DDSP(F%N_orb)
  real(long) :: DDSP_sum
  integer                      :: lmax, N_omega

  integer :: i, j, k, hi, lo

  real(long), dimension(F%N_uni, (P%compton_lmax+1)*F%N_orb) :: cont_wf

  real(long), dimension(0:F%N_orb) :: J_Q
  real(long) :: p_q, Q

  lmax = P%compton_lmax
  N_omega = P%compton_N_omega
  d_omega = P%compton_d_omega

  OUTPUT = unit_no
  call interpolate_potential( F )
!  SP = F%SP_pot


  theta = dble(PI/180.0 * theta_deg)
  omega_in = P%PE_start

  allocate(omega_s(N_omega))
  omega_s(:) = (/ ( omega_in - dble(j)* d_omega*EV2AU , j=1, N_omega ) /)
  


!   open(unit=100, file='compton_spectrum.dat', status='replace', action='write')
!   write(100,'(A)')'#    1            2              3             4             5           6             7             8           9           10             11           12         13             14'
!   write(100,'(A)')'#    omega_s      Q     Thomson_CS(theta)  p_q       J_Q(1s)     J_Q(2s)     J_Q(2p)   J_Q(sum)  DDSP_sum    DDSP(1s)    DDSP(2s)     DDSP(2p)     CS_IA     J(CS_IA)'
  
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module with the Compton spectrum and Compton profile calculation:'
  write(OUTPUT,"('# ')")
  write(OUTPUT,'(A)') '# It calculates the DDSCS (total and each single orbital contributions) for different scattered photon energies (omega_s),'
  write(OUTPUT,"('# at scattering angle ',f4.1,' degrees!')") theta_deg 
  write(OUTPUT,'(A)') '# Additionaly, the Compton profile J (total and single orbital) for the respective p_q is determined (also at fixed scattering angle).'
  write(OUTPUT,"('# ')")
  !
  ! NOTE: $ in the format suppresses a new line, 
  !       so "('#'$)" is equivalent to "('#')", advance="no"
  !
  write(OUTPUT,"('#',99a13$)") 'omega_s', 'Q    ', 'Thomson_CS', 'DDSCS(total)', ( WF_name(F,i), i=1,F%N_orb )
  write(OUTPUT,"(13x,99a13 $)") ,  'p_Q    ', 'J(p_q)(total)', ( WF_name(F,i), i=1,F%N_orb )

  lo = 1
  hi = N_omega
  allocate(order(hi))
  if( P%compton_index_half ) then
    order = divide2_index_order(lo,hi)
  else
    order = linear_index_order(lo,hi)
  end if

  do k = lo, hi
!     write(*,"('#  currently calculating step ',I3,' of ',I3,' omega_s =',f9.2,' eV')" ) j, N_omega, omega_s(j)/EV2AU
    j = int( order(k) )
    !write(*,*) 'j=',j
    !goto 9000
    cont_wf = get_continuum_wf(F, P, omega_s(j), lmax)
     
    DDSP(:) = calculate_DDSP(P, F, cont_wf, theta, omega_s(j), lmax)
     DDSP_sum = sum( DDSP(:) )
    
!     !---------- Now follows some testing stuff
!     CS_IA = impulse_approximation(omega_s(j),omega_in,theta)
!     
     Q = momentum_transfer(omega_in, omega_s(j), theta)
     p_q = (omega_in - omega_s(j))/Q - HALF * Q
!     
    do i = 1, F%N_orb
      J_Q(i) = compton_profile(DDSP(i), omega_s(j), omega_in, theta)
    end do
      J_Q(0) = compton_profile(DDSP_sum, omega_s(j), omega_in, theta)
    write(OUTPUT,"(99ES13.3$)") omega_s(j), Q, Thomson_CS(theta), DDSP_sum, DDSP(:)
    write(OUTPUT,"(13x,99ES13.3 $)") p_q, J_Q(:)
 !9000 continue
  end do


!   close(unit=100)

end subroutine print_compton_spectrum

function calculate_DDSP(P, F, cont_wf, theta, omega_s, lmax) result (DDSP)
  !This function calculates the Double-differential scattering probability (DDSP) for a given angle and final photon energy.

  type(Param), intent(in)             :: P
  type(WF_data), intent(in)           :: F
  real(long), dimension(F%N_uni, (lmax+1)*F%N_orb) :: cont_wf
  integer,intent(in)                  :: lmax
  real(long)                          :: DDSP(F%N_orb)
  real(long)                          :: theta, omega_s, omega_in
  !local variables 
  integer                             :: i, j, k, column, r
  integer                              :: occ_n, l_orb, l2max
  real(long)                          :: E, e_n, Q, CG, jj, S, sum_1, sum_2
  real(long)                           :: u(F%N_uni) !, up(F%N_uni)
  real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes

  
  omega_in = P%PE_start                           !incoming photon energy
  Q = momentum_transfer(omega_in, omega_s, theta) !modulus of the momentum transfer
!   Q = momentum_transfer(omega_in, omega_in, theta) !modulus of the momentum transfer
  DDSP(:) = ZERO                                  !DDSP = double differential scattering probability.  

  l2max = maxval( F%config(:,2) ) + lmax + 1      !lmax + 1 to count also l=0.
  if (l2max < 15) l2max = 15            ! to generate Bessel_function with faster routine
  allocate(j_bes(F%N_uni,l2max))
  j_bes = ZERO
  if( P%compton_cfelt_bessel ) then
    call cfelt_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)  !gives the spherical Besselfunctions j_l(Qr) up to l2max
  else
    call get_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)  !gives the spherical Besselfunctions j_l(Qr) up to l2max
  end if 
  
  orbitals: do i = 1, F%N_orb                 !sum over all orbitals
    e_n = F%e_orb(i)                          !orbital energy
    occ_n = F%config(i,3)                     !occupation number
    l_orb = F%config(i,2)                     !angular quantum number l
    E = e_n + omega_in - omega_s
    sum_2 = ZERO
    if (occ_n == 0) cycle                     !no electron in the orbital
    if ( E < 0 ) cycle                        !electron cannot be ionized --> not a bound to free transition
    do j = 0, lmax                            !sum over l-value of final state
      l2max = l_orb + j
      jj = dble(j)
      sum_1 = ZERO
      do r = 1, F%N_uni
        u(r) = cont_wf(r,(i-1)*(lmax+1) + j+1)
      end do
      do k = 0, l2max                         !sum over l-value of the expansion of exp(iQr)
        CG = cleb( 2*l_orb, 0, 2*j, 0, 2*k, 0)
        if (CG == 0 .or. mod(l_orb+j+k,2) ==1 ) cycle
        column = k+1                        !j_0 is in the first column of j_bes, j_1 in the second, etc.
        S = ZERO
        S = r_integral_linear(F%N_uni, F%dr, u(:), F%u_uni(:,i), j_bes(:,column))
        sum_1 = sum_1 + (S * CG)**2
      end do
       sum_2 = sum_2 + (TWO * jj + ONE) * sum_1
    end do  
      DDSP(i) = sum_2  * Thomson_CS(theta) * omega_s / omega_in * dble(occ_n) 
!       DDSP(i) = sum_2  * dble(occ_n) * Thomson_CS(theta) 
  end do orbitals

end function calculate_DDSP

function calculate_DDSP_nonlinear(P, F, G, cont_wf, theta, omega_s, lmax) result (DDSP)
  !calculates the non-linear double-differential scattering probability (DDSP) 
  !for a given angle and final photon energy.

  type(Param), intent(in)             :: P
  type(WF_data), intent(in)           :: F
  type(Coordinate), intent(in)        :: G
  real(long), dimension(F%N_uni, (lmax+1)*F%N_orb) :: cont_wf
  integer,intent(in)                  :: lmax
  real(long)                          :: DDSP(F%N_orb)
  real(long)                          :: theta, omega_s, omega_in
  !local variables 
  integer                             :: i, j, k, column, r
  integer                             :: occ_n, l_orb, l2max
  real(long)                          :: E, e_n, Q, CG, jj, S, sum_1, sum_2, S_non1, S_non2
  real(long)                          :: u(F%N_uni), deriv_ui(F%N_uni)
  real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes

  
  omega_in = P%PE_start                           !incoming photon energy
  Q = momentum_transfer(omega_in, omega_s, theta) !modulus of the momentum transfer
  DDSP(:) = ZERO                                  

  l2max = maxval( F%config(:,2) ) + lmax + 1      !lmax + 1 to count also l=0.
  if (l2max < 15) l2max = 15            ! to generate Bessel_function with faster routine
  allocate(j_bes(F%N_uni,l2max))
  j_bes = ZERO
  if( P%compton_cfelt_bessel ) then
    call cfelt_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)  !gives the spherical Besselfunctions j_l(Qr) up to l2max
  else
    call get_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)  !gives the spherical Besselfunctions j_l(Qr) up to l2max
  end if 
  
  
  orbitals: do i = 1, F%N_orb                 
    e_n = F%e_orb(i)                          !orbital energy
    occ_n = F%config(i,3)                     !occupation number
    l_orb = F%config(i,2)                     !angular quantum number l
    E = e_n + omega_in - omega_s
    sum_2 = ZERO
    if (occ_n == 0) cycle                     
    if ( E < 0 ) cycle                        !electron cannot be ionized --> not a bound to free transition
    do j = 0, lmax                            !sum over l-value of final state
      l2max = l_orb + j
      jj = dble(j)
      sum_1 = ZERO
      do r = 1, F%N_uni
        u(r) = cont_wf(r,(i-1)*(lmax+1) + j+1)
      end do
      do k = 0, l2max                         !sum over l-value of the expansion of exp(iQr)
        CG = cleb( 2*l_orb, 0, 2*j, 0, 2*k, 0)
        if (CG == 0 .or. mod(l_orb+j+k,2) ==1 ) cycle
        column = k+1                        !j_0 is in the first column of j_bes, j_1 in the second, etc.
        S      = ZERO
        S      = r_integral_linear(   F%N_uni, F%dr, u(:), F%u_uni(:,i), j_bes(:,column))

        S_non1 = ZERO
        deriv_ui  = matmul( G%Deriv(:,:), F%u_uni(:,i))
        S_non1 = r_integral_nonlinear1(F%N_uni, F%dr, u(:), deriv_ui , j_bes(:,column))

        S_non2 = ZERO
        S_non2 = r_integral_nonlinear2(F%N_uni, F%dr, u(:), F%u_uni(:,i), j_bes(:,column))

        S = S + S_non1 + S_non2 ! TODO add factors from Dietrich
        sum_1 = sum_1 + (S * CG)**2
      end do
       sum_2 = sum_2 + (TWO * jj + ONE) * sum_1
    end do  
      DDSP(i) = sum_2  * Thomson_CS(theta) * omega_s / omega_in * dble(occ_n) 
  end do orbitals

end function calculate_DDSP_nonlinear


function calculate_bound2bound(P, F, theta) result (b2b)
!This function calculates the differential scattering probability (DSP) for bound-to-bound excitations a given angle.
! 
  type(Param), intent(in)             :: P
  type(WF_data), intent(in)           :: F
  real(long)                          :: b2b(F%N_orb)
  real(long)                          :: theta
! local variables 
  integer, dimension(F%N_all)          :: occ_final
   real(long)                          :: omega_in, Q
  integer                             :: i, j, m, column
   integer                              :: occ_n, l_in, l_fi, l2max
  real(long)                          :: E, e_n, CG, S, sum_1, sum_2, factor
   real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes
  real(long)                           :: u(F%N_uni)
  integer    :: n,l, occ

  !# Determine occupation number of final bound states
  occ_final = 0
  !write(*,*) 'calculate_bound2bound()========='
  do j = 1, F%N_orb
    n   = F%config(j,1)
    l   = F%config(j,2)
    occ = F%config(j,3)
    !write(*,*) 'j=',j,'-------->', 'n=',n,'; l=',l,'; occ=',occ
    i   = F%N * l + (n-l)
    !write(*,*) 'occ_final(',i,')=',occ
    occ_final(i) = occ
  end do
  !write(*,*) 'sum(occ_final(1:F%N_all)) vs  sum(F%config(1:F%N_orb,3))'
  !write(*,*) sum(occ_final(1:F%N_all)),  sum(F%config(1:F%N_orb,3))
  if ( sum(occ_final(1:F%N_all)) /= sum(F%config(1:F%N_orb,3))) then
    write(STDERR, *) 'Occupation numbers of bound final states do not sum up.'
    stop
  end if

   omega_in = P%PE_start                           !incoming photon energy
  Q = momentum_transfer(omega_in, omega_in, theta)
  b2b(:) = ZERO

  !# Create the spherical Bessel functions: 
  !  NOTE here we neglect the change in Q for different final states
  l2max = maxval( F%config(:,2) ) + F%lmax + 1
  if (l2max < 15) l2max = 15            ! to generate Bessel_function with faster routine
  allocate(j_bes(F%N_uni,l2max))
  if( P%compton_cfelt_bessel ) then
    call cfelt_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)
  else
    call get_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)
  end if

   occupied_orbitals: do i = 1, F%N_orb        !sum over all orbitals
     e_n = F%e_orb(i)                          !orbital energy
     occ_n = F%config(i,3)                     !occupation number
     l_in = F%config(i,2)                     !angular quantum number l
     sum_2 = ZERO
     if (occ_n == 0) cycle                     !no electron in the orbital
    final_states: do m = 1, F%N_all
      E = F%e_all(m) - F%e_orb(i)
      if ( F%e_all(m) > 0 ) cycle          !do not include pseudo-continuum states
      if ( E < 0) cycle
      if ( abs(F%e_all(m) - F%e_orb(i)) < 1.d-8 ) cycle !Do not include bound transitions with (n_f, l_f)=(n_i,l_i)!
      l_fi = ceiling( (real(m)- real(F%N))/real(F%N) )
      sum_1 = ZERO

!      Q = momentum_transfer(omega_in, omega_in-E, theta)
!      call get_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)

      do j = 0, l_in + int(l_fi)                ! sum over l-value of the expansion of exp(iQr)
        CG = cleb( 2*l_in, 0, 2*l_fi, 0, 2*j, 0)
        if (CG == 0 .or. mod( l_in+l_fi+j ,2) ==1 ) cycle
        u(:) = interpolate_bound_wf_on_uniform_grid(F, m)
        column = j+1                        !j_0 is in the first column of j_bes, j_1 in the second, etc.
         S = ZERO
         S = r_integral_linear(F%N_uni, F%dr, u(:), F%u_uni(:,i), j_bes(:,column))
        sum_1 = sum_1 + (S * CG)**2
      end do
      factor = (TWO * dble(l_fi) + ONE) * (omega_in - E)/omega_in * ( ONE - dble(occ_final(m))/( TWO*(TWO*dble(l_fi) +ONE)) )  
      sum_2 = sum_2 + factor * sum_1
    end do final_states
    b2b(i) = sum_2 * Thomson_CS(theta) * dble(occ_n)
   end do occupied_orbitals

end function calculate_bound2bound


function calculate_waller_hartree(P,F,theta,lmax) result (DSP)
  type(Param), intent(in)             :: P
  type(WF_data), intent(in)           :: F
   real(long)                          :: theta
  real(long)                          :: DSP
  integer                              :: lmax
  !local variables
  integer                              :: j,k,l, column, l_k, l_j, occ_k, occ_j, N_det
  real(long)                          :: omega_in
  integer                              :: l2max
  real(long), allocatable, dimension(:,:) :: j_bes
  real(long)                          :: S, Q, CG
  real(long), external                :: cleb
  real(long)                          :: sum_1, sum_2, average
  

  omega_in = P%PE_start               !incoming photon energy
  Q = momentum_transfer(omega_in, omega_in, theta)

  !l2max = maxval( F%config(:,2) ) + F%lmax + 1
  l2max = lmax+1
  if (l2max < 15) l2max = 15            ! to generate Bessel_function with faster routine
  allocate(j_bes(F%N_uni,l2max))
  if( P%compton_cfelt_bessel ) then
    call cfelt_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)
  else
    call get_spherical_bessel(l2max, Q, F%r_uni, F%N_uni, j_bes)
  end if
  
  if (theta==0) then
    DSP = ZERO
    return
  end if

  DSP = ZERO
  do j = 1, F%N_orb
    sum_2 = ZERO
    l_j = F%config(j,2)
    occ_j= F%config(j,3)
    do k=1, F%N_orb
      sum_1 = ZERO
      N_det = 0
      average = 0
      l_k = F%config(k,2)
      occ_k= F%config(k,3)
       do l = 0, l_j + l_k  ! sum over l-value of exp(iQr)
        CG = cleb( 2*l_k, 0, 2*l, 0, 2*l_j, 0)
        column = l+1
        S = r_integral_linear(F%N_uni, F%dr, F%u_uni(:,j), F%u_uni(:,k), j_bes(:,column))
        sum_1 = sum_1 + (TWO * dble(l) + ONE) * S**2 * CG**2
      end do       
      average = ZERO
      if ( 4* l_k + 2 == occ_k .and. 4* l_j + 2 == occ_j) then  !j and k are closed shells
        average = ONE
      else if ( 4* l_j + 2 == occ_j .and. 4* l_k + 2 /= occ_k) then ! j closed, k open
        average = dble( binomial(4*l_k + 1, occ_k-1) )/ dble( binomial(4* l_k + 2, occ_k)  )
      else if ( 4* l_k + 2 == occ_k .and. 4* l_j + 2 /= occ_j) then ! j open, k closed
          average = dble( binomial(4*l_j + 1, occ_j-1) ) / dble( binomial(4* l_j + 2, occ_j)  )
      else if  (j == k .and. 4* l_j + 2 /= occ_j) then  !j=k is open
         average = dble( binomial(4*l_k + 1, occ_k-1) ) / dble( binomial(4* l_k + 2, occ_k)  )
      else if  (j /= k .and. 4* l_j + 2 /= occ_j .and. 4* l_k + 2 /= occ_k) then  !j /= k and j is open, k is open
         average = dble( binomial(4*l_k + 1, occ_k-1) ) / dble( binomial(4* l_k + 2, occ_k)  ) * dble( binomial(4*l_j + 1, occ_j-1) ) / dble( binomial(4* l_j + 2, occ_j)  )
      else
        write(STDERR, *) 'Waller-Hartree: Averaging failed.' 
      end if
      sum_2 = sum_2 + average * TWO * (TWO * dble(l_k) + ONE) *  sum_1 
    end do 
      DSP = DSP + sum_2 
  end do  
 
  DSP = Thomson_CS(theta) * ( dble( sum(F%config(1:F%N_orb,3)) ) -  DSP )

end function calculate_waller_hartree



subroutine print_waller_hartree( P, F )
  type(Param), intent(in)             :: P
  type(WF_data), intent(in)           :: F
  !local
  integer :: OUTPUT
  integer :: lmax, N_theta
  real(long) :: d_theta
  real(long) :: min_theta, omega_in
  real(long),allocatable,dimension(:) :: theta(:)
  integer :: m,j, i
  real(long) :: total_CS
  real(long), dimension(P%compton_N_theta) :: scat_fac, sin_theta
  
  OUTPUT = P%OUTPUT
  
  lmax = P%compton_lmax
  N_theta = P%compton_N_theta
  d_theta = P%compton_d_theta
  min_theta = P%compton_min_theta

  omega_in = P%PE_start               !incoming photon energy

  d_theta = dble(d_theta) * PI/180.d0   ! set angle grid spacing
  min_theta = dble(min_theta) * PI/180.d0
  allocate(theta(N_theta))
  theta(:) = (/ ( min_theta + dble(j) * d_theta, j=0, N_theta-1 ) /)

  
  !Giving some information to standart output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module with the Waller-Hartree method.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates of static structure factor S(theta), w.r.t. scattering angle.'
  write(OUTPUT,"('# ', a)")  'The differential scattering probability (DSP) into a given scattering angle is DSP = Thomson_CS * S(theta).'
  write(OUTPUT,"('# ', a)")  ' "DSP w free e-" gives the static structure factor plus an additional'
  write(OUTPUT,"('# ', a)")  'background contribution from free electrons (all electrons missing to the neutral atom are considered as free).' 
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 4a13)") 'Theta (rad)', 'Thomson_CS  ', 'DSP      ', 'DSP w free e-'
  
  call config_to_index( F%config, i )   ! i: index of configuration
  if ( P%yes_xmdyn ) then
    write( P%IO_IXS, "(a,i10)" ) "ixs", N_theta ! P%IO_IXS=STDOUT for xmdyn
  else
    open(P%IO_IXS, file=P%file_ixs, access="append", form="formatted" )    !Output file for rate equations
    write( P%IO_IXS, "('#',2i10)" ) i, N_theta        
    open(P%IO_IXS2, file=P%file_ixs2, access="append", form="formatted" ) !Output file for rate equations with background from free electrons
    write( P%IO_IXS2, "('#',2i10)" ) i, N_theta
  end if
  
  do m = 1, N_theta
    scat_fac(m) = calculate_waller_hartree(P,F,theta(m),lmax)
    write(OUTPUT,'(99ES13.3)') theta(m), thomson_CS(theta(m)), scat_fac(m), scat_fac(m)+free_electron_background(F, theta(m)) 
      
    !print out to .ixs file: theta in radian and S(Q), i.e. dsigma/dOmega in units of Thomson-cross-section
    if ( P%yes_xmdyn ) then
      write( P%IO_IXS, "(f23.15,es23.15)" ) &
        momentum_transfer( omega_in, omega_in, theta(m) ), &
        scat_fac(m)/thomson_CS(theta(m))
    else
      write( P%IO_IXS, "(f23.15,es23.15)" ) theta(m), scat_fac(m)/thomson_CS(theta(m))
      write( P%IO_IXS2, "(f23.15,es23.15)" ) theta(m), (scat_fac(m)+free_electron_background(F, theta(m)))/thomson_CS(theta(m))
    end if
  end do

  if ( .not. P%yes_xmdyn ) then
    write( P%IO_IXS, "()" )
    close( P%IO_IXS )
    write( P%IO_IXS2, "()" )
    close( unit=P%IO_IXS2 )
  end if
    
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# Output files for xatom_req are written to: ')")
  write(OUTPUT,"('#   DSP             : ',a)") trim(P%file_ixs)
  write(OUTPUT,"('#   DSP w/ free elec: ',a)") trim(P%file_ixs2)

  !#
  !# angular_integral
  !# 
  sin_theta = (/ (dsin(theta(m) ) , m=1, N_theta) /)
  total_CS = sum( scat_fac(1:N_theta)*sin_theta(1:N_theta) )
  total_CS = total_CS - HALF * ( scat_fac(N_theta)*sin_theta(N_theta) ) - HALF * ( scat_fac(1) * sin_theta(1) )
  total_CS = ( TWO * d_theta * N_theta ) * d_theta * total_CS

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an angular integration over the given theta range from', P%compton_min_theta, ' deg to ', theta(N_theta)*180./PI, ' deg'
  write(OUTPUT,"('# the total Compton cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'
!   write(OUTPUT,"('# ')")


end subroutine print_waller_hartree

!=======================================================================================================
!=====================================Start of nonuniform section=======================================
!======================================================================================================= 

function calculate_structure_factor_nonuniform_b2b(P, F, theta_rad) result (DSP_partial) 
! This function calculates the bound-to-bound contribution to S(Q)*Thomson cross-section for a given 
! angle theta(rad). 
  type(Param), intent(in)             :: P               
  type(WF_data), intent(in)           :: F
  real(long)                          :: DSP_partial(F%N_orb)         
  real(long)                          :: theta_rad 
  integer, dimension(F%N_all)         :: occ_all
  real(long)                          :: omega_in, Q
  integer                             :: i, L, i_all, column
  integer                             :: n_i, occ_i, l_i, n_f, l_f, n_all, L_max
  real(long)                          :: E, epsilon_i, epsilon_f, CG, R_integrand, sum_1, sum_2, factor
  real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes  
  
  omega_in = P%PE_start                   
  occ_all(:) = 0
  DSP_partial(:) = ZERO 
  L_max = maxval( F%config(:,2) ) + P%lmax + 1     ! P%lmax has to be larger than ~25 to get decent 
                                                   ! results. I always do -lmax 35. 
  allocate(j_bes(F%N, L_max))

  ! This prepares all initial occupation numbers
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

  ! (8.36) in Jan Malte's thesis.
  occupied_orbitals: do i = 1, F%N_orb 
    n_i   = F%config(i,1)                     
    l_i   = F%config(i,2)     
    occ_i = F%config(i,3)         
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i 
    epsilon_i = F%e_orb(i) 
    sum_2 = ZERO                       
    if (occ_i == 0) cycle              
  
    final_states: do l_f = 0, P%lmax
      grid_points: do n_f = 1, F%N 
        n_all = F%N * l_f + n_f
        epsilon_f = F%e_all(n_all)       
        E = epsilon_f - epsilon_i  
        if (epsilon_f > 0) cycle    ! only b2b contribution  
        if (omega_in-E < 0.) cycle  ! Don't include negative scattering energies
        if (i_all == n_all) cycle   ! Don't include "transitions" to the same state

        sum_1 = ZERO                     
        Q = momentum_transfer(omega_in, omega_in-E, theta_rad)  
        call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
        do L = 0, l_i+l_f    
          CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
          if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
          column = L+1                  
          R_integrand = radial_integral_u_all( F, i_all, n_all, j_bes(:,column)) 
          sum_1 = sum_1 + (R_integrand * CG)**2         
        end do
        factor = (TWO * dble(l_f) + ONE) * (omega_in - E)/omega_in * &
        &(ONE - dble(occ_all(n_all))/dble(2*(2*l_f+1)))
        sum_2 = sum_2 + factor * sum_1  
      end do grid_points
    end do final_states
    DSP_partial(i) = Thomson_CS(theta_rad) * sum_2 * dble(occ_all(i_all))
  end do occupied_orbitals
end function calculate_structure_factor_nonuniform_b2b

!======================================================================================================= 

function calculate_structure_factor_nonuniform_b2free(P, F, theta_rad) result (DSP_partial)
! This function calculates the bound-to-free contribution to S(Q)*Thomson cross-section for a given 
! angle theta(rad). 
  type(Param), intent(in)             :: P               
  type(WF_data), intent(in)           :: F
  real(long)                          :: DSP_partial(F%N_orb)         
  real(long)                          :: theta_rad 
  integer, dimension(F%N_all)         :: occ_all
  real(long)                          :: omega_in, Q
  integer                             :: i, L, i_all, column
  integer                             :: n_i, occ_i, l_i, n_f, l_f, n_all, L_max
  real(long)                          :: E, epsilon_i, epsilon_f, CG, R_integrand, sum_1, sum_2, factor
  real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes  
  
  omega_in = P%PE_start                   
  occ_all(:) = 0
  DSP_partial(:) = ZERO 
  L_max = maxval( F%config(:,2) ) + P%lmax + 1   ! P%lmax has to be larger than ~25 to get decent 
                                                 ! results. I always do -lmax 35. 

  allocate(j_bes(F%N,L_max))

  ! This prepares all initial occupation numbers
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

  ! (8.36) in Jan Malte's thesis.
  occupied_orbitals: do i = 1, F%N_orb 
    n_i   = F%config(i,1)                     
    l_i   = F%config(i,2)     
    occ_i = F%config(i,3)         
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i 
    epsilon_i = F%e_orb(i) 
    sum_2 = ZERO                       
    if (occ_i == 0) cycle              
  
    final_states: do l_f = 0, P%lmax
      grid_points: do n_f = 1, F%N 
        n_all = F%N * l_f + n_f
        epsilon_f = F%e_all(n_all)       
        E = epsilon_f - epsilon_i  
        if (epsilon_f < 0) cycle    ! only b2free contributions
        if (omega_in-E < 0.) cycle  ! Don't include negative scattering energies
        if (i_all == n_all) cycle   ! Don't include "transitions" to the same state

        sum_1 = ZERO                     
        Q = momentum_transfer(omega_in, omega_in-E, theta_rad)  
        call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
        do L = 0, l_i+l_f    
          CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
          if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
          column = L+1                  
          R_integrand = radial_integral_u_all( F, i_all, n_all, j_bes(:,column)) 
          sum_1 = sum_1 + (R_integrand * CG)**2         
        end do
        factor = (TWO * dble(l_f) + ONE) * (omega_in - E)/omega_in * &
        &(ONE - dble(occ_all(n_all))/dble(2*(2*l_f+1)))
        sum_2 = sum_2 + factor * sum_1  
      end do grid_points
    end do final_states
    DSP_partial(i) = Thomson_CS(theta_rad) * sum_2 * dble(occ_all(i_all))
  end do occupied_orbitals
end function calculate_structure_factor_nonuniform_b2free

!======================================================================================================= 

subroutine print_nonuniform_compton_structure_factor_orbital(P, F)
! This subroutine is called by -comp_nonuni_struc_fact_orbi. It returns the individual orbital 
! contributions to S(Q)*Thomson cross-section for a range of angles theta. The first row returns the 
! angle in radians, the second row returns the Thomson cross-section in au. Then follow the bound-to-
! bound contributions, differentiated per orbital; so b2b 1s is the 3rd column, b2b 2s the 4th, followed
! by b2b 2p, etc... After that, the same follows for the b2free contributions per orbital.
! NOTE: This is a bit messy, as the columns for the b2free orbital contributions do NOT have a fixed
! number, i.e. for an arbitrary system under consideration I can't tell you in what column b2free 2s is.
! For this reason, these columns don't have headings.
! A solution could be to write each orbital contribution to a separate file. (TO BE DONE)
  
  type(Param), intent(in)                :: P
  type(WF_data), intent(in)              :: F
  integer                                :: OUTPUT     
  integer                                :: N_theta    
  real(long)                             :: d_theta_deg, d_theta_rad    
  real(long)                             :: min_theta_deg, min_theta_rad, total_CS  
  real(long), allocatable, dimension(:)  :: theta_rad_array, sin_theta_rad, total_scatfac  
  real(long), allocatable, dimension(:,:):: scat_fac1, scat_fac2
  integer                                :: m, j      

  OUTPUT = P%OUTPUT

! Range of theta (in rad) under consideration. Default from 0 deg to 180 deg in steps of 5 deg.
  N_theta = P%compton_N_theta
  d_theta_deg = P%compton_d_theta
  min_theta_deg = P%compton_min_theta

  d_theta_rad = dble(d_theta_deg) * PI/180.d0     
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 
  allocate(theta_rad_array(N_theta))                
  theta_rad_array(:) = (/ ( min_theta_rad + dble(j-1) * d_theta_rad, j=1, N_theta ) /)

  allocate(scat_fac1(N_theta,F%N_orb))
  allocate(scat_fac2(N_theta,F%N_orb))
  allocate(sin_theta_rad(N_theta))
  allocate(total_scatfac(N_theta))                

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates individual orbital contributions to the differential&
                             & scattering probability (DSP) factor, w.r.t. scattering angle.'
  write(OUTPUT,"('# ', a)")  'The DSP is given by Thomson_CS * S(Q).'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'The output is structured as follows:'
  write(OUTPUT,"('# ', a)")  'Columns 3 until (2 + number of subshells of system under consideration) &
                              &are bound-to-bound orbital contributions.'
  write(OUTPUT,"('# ', a)")  'Columns (3 + number of subshells of system under consideration) until &
                              &the end are bound-to-free orbital contributions.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 2a13)") 'Theta (rad)', 'Thomson_CS '
  do m = 1, N_theta                        
    scat_fac1(m,:) = calculate_structure_factor_nonuniform_b2b(P,F,theta_rad_array(m)) 
    scat_fac2(m,:) = calculate_structure_factor_nonuniform_b2free(P,F,theta_rad_array(m))
    write(OUTPUT,'(99ES13.3)') theta_rad_array(m), Thomson_CS(theta_rad_array(m)), scat_fac1(m,:), scat_fac2(m,:)
  end do

! Going to integrate over the angles now, in order to get the total cross-section.

  do m = 1, N_theta
    total_scatfac(m) = sum (scat_fac1(m, 1:F%N_orb) + scat_fac2(m,1:F%N_orb))
  end do
     
  sin_theta_rad = (/ (dsin(theta_rad_array(m) ) , m=1, N_theta) /)
  total_CS = sum( total_scatfac(1:N_theta)*sin_theta_rad(1:N_theta) )
  total_CS = total_CS - HALF * ( total_scatfac(N_theta)*sin_theta_rad(N_theta) ) - &
             &HALF * ( total_scatfac(1) * sin_theta_rad(1) )
  total_CS = ( TWO * PI) * d_theta_rad * total_CS

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an angular integration over the given theta range&
                                           & from', P%compton_min_theta, ' deg to ', &
                                           &theta_rad_array(N_theta)*180./PI, ' deg'
  write(OUTPUT,"('# the total Compton cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'

end subroutine print_nonuniform_compton_structure_factor_orbital

!======================================================================================================= 

subroutine print_nonuniform_structure_factor_compton(P, F)
! This subroutine is called by -compton_nonuni_struc_fact. It returns the total, b2b, and b2free
! contributions to S(Q)*Thomson cross-section for a range of angles theta. The first row returns the 
! angle in radians, the second row returns the Thomson cross-section in au. The 3rd, 4th and 5th row 
! denote the total, b2b, and b2free contributions respectively.
  
  type(Param), intent(in)                :: P
  type(WF_data), intent(in)              :: F
  integer                                :: OUTPUT     
  integer                                :: N_theta    
  real(long)                             :: d_theta_deg, d_theta_rad    
  real(long)                             :: min_theta_deg, min_theta_rad, total_CS  
  real(long), allocatable, dimension(:)  :: theta_rad_array, sin_theta_rad, total_scatfac  
  real(long), allocatable, dimension(:)  :: b2b, b2free
  real(long), allocatable, dimension(:,:):: scat_fac1, scat_fac2
  integer                                :: m, j      

  OUTPUT = P%OUTPUT

! Range of theta (in rad) under consideration. Default from 0 deg to 180 deg in steps of 5 deg.
  N_theta = P%compton_N_theta
  d_theta_deg = P%compton_d_theta
  min_theta_deg = P%compton_min_theta

  d_theta_rad = dble(d_theta_deg) * PI/180.d0     
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 
  allocate(theta_rad_array(N_theta))                
  theta_rad_array(:) = (/ ( min_theta_rad + dble(j-1) * d_theta_rad, j=1, N_theta ) /)

  allocate(scat_fac1(N_theta,F%N_orb))
  allocate(scat_fac2(N_theta,F%N_orb))
  allocate(sin_theta_rad(N_theta))
  allocate(total_scatfac(N_theta))     
  allocate(b2b(N_theta))                
  allocate(b2free(N_theta))                           

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates the differential scattering probability (DSP) factor, &
                              &w.r.t. scattering angle.'
  write(OUTPUT,"('# ', a)")  'The DSP is given by Thomson_CS * S(Q).'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 3a13, 1a8, 1a15 )") 'Theta (rad)', 'Thomson_CS ', 'Total NonUni', 'b2b', 'b2free'
  do m = 1, N_theta                        
    scat_fac1(m,:) = calculate_structure_factor_nonuniform_b2b(P,F,theta_rad_array(m)) 
    b2b(m) = sum (scat_fac1(m, 1:F%N_orb))
    scat_fac2(m,:) = calculate_structure_factor_nonuniform_b2free(P,F,theta_rad_array(m))
    b2free(m) = sum (scat_fac2(m, 1:F%N_orb))
    total_scatfac(m) = sum (scat_fac1(m, 1:F%N_orb) + scat_fac2(m,1:F%N_orb))
    write(OUTPUT,'(99ES13.3)') theta_rad_array(m), Thomson_CS(theta_rad_array(m)), total_scatfac(m), &
                               &b2b(m), b2free(m)
  end do

! Going to integrate over the angles now, in order to get the total cross-section.
     
  sin_theta_rad = (/ (dsin(theta_rad_array(m) ) , m=1, N_theta) /)
  total_CS = sum( total_scatfac(1:N_theta)*sin_theta_rad(1:N_theta) )
  total_CS = total_CS - HALF * ( total_scatfac(N_theta)*sin_theta_rad(N_theta) ) - &
             &HALF * ( total_scatfac(1) * sin_theta_rad(1) )
  total_CS = ( TWO * PI) * d_theta_rad * total_CS

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an angular integration over the given theta range& 
                                           & from', P%compton_min_theta, ' deg to ', &
                                           &theta_rad_array(N_theta)*180./PI, ' deg'
  write(OUTPUT,"('# the total Compton cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'

end subroutine print_nonuniform_structure_factor_compton

!======================================================================================================= 

function calculate_f_n(P, F, theta_rad, omega_s, init_orbital) result(f_n)   
! Calculates for a given orbital init_orbital, scattering energy omega_s, and angle theta the f_n 
! (the DDCS without sum_i, sum_f), returned as an array of dimension F%N * (P%lmax + 1), where each 
! value corresponds to one epsilon_f. These energies are only ordered from smallest to largest per 
! specific l_f.
  type(Param), intent(in)                 :: P               
  type(WF_data), intent(in)               :: F
  real(long), intent(in)                  :: theta_rad
  real(long), intent(in)                  :: omega_s
  integer, intent(in)                     :: init_orbital
  real(long)                              :: f_n(F%N * (P%lmax + 1))
  real(long)                              :: epsilon_f, epsilon_i
  integer                                 :: i, L_max
  real(long), allocatable, dimension(:,:) :: j_bes  
  real(long)                              :: omega_in, Q
  integer, dimension(F%N_all)             :: occ_all
  integer                                 :: n_i, l_i, occ_i, i_all
  integer                                 :: n_f, l_f, n_all, L, column
  real(long)                              :: sum_1, CG, R_integrand, factor
  real(long), external                    :: cleb

  omega_in = P%PE_start 
                  
  L_max = maxval( F%config(:,2) ) + P%lmax + 1 
  allocate(j_bes(F%N,L_max))

  occ_all(:) = 0
  f_n(:) = ZERO

! Don't bother calculating if the energy transfer isn't at least the ionization energy 
  epsilon_i = F%e_orb(init_orbital) 
  if (omega_in - omega_s .lt. -epsilon_i) then 
    return
  end if

! This prepares all initial occupation numbers
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

  n_i   = F%config(init_orbital,1)                     
  l_i   = F%config(init_orbital,2)     
  i_all   = F%N * l_i + (n_i-l_i)

  do l_f = 0, P%lmax
    do n_f = 1, F%N 
      n_all = F%N * l_f + n_f
      epsilon_f = F%e_all(n_all)  ! These energies are all in ascending order per l_f 
      if (epsilon_f < 0) cycle    ! Only b2free contributions
      if (omega_in + epsilon_i - epsilon_f < 0.d0) cycle ! No negative final photon energies
      if (i_all == n_all) cycle   ! Don't include "transitions" to the same state
  
      sum_1 = ZERO                     
      Q = momentum_transfer(omega_in, omega_s, theta_rad)  
      call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
      do L = 0, l_i+l_f    
        CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
        if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
        column = L+1                  
        R_integrand = radial_integral_u_all(F, i_all, n_all, j_bes(:,column)) 
        sum_1 = sum_1 + (R_integrand * CG)**2      ! This does the sum over L         
      end do
      factor = (TWO * dble(l_f) + ONE) * (omega_s/omega_in)     
      f_n(n_all) = Thomson_CS(theta_rad) * occ_all(i_all) * factor * sum_1  
    end do       
  end do 

end function calculate_f_n

!=======================================================================================================

function calculate_compton_nonuniform_DDCS(P,F, theta_rad, omega_s) result(DDCS_orbital)
! Calculates the partial sums of the f_n's from the previous function. First, we specify a single l_f, 
! and then go over the F%N values of n_f. The energies epsilon_f are already ordered from smallest to 
! largest. Per l_f, the cumulative sums over the n_f index are calculated in the 
! array histogram, where element m corresponds to the sum of f_n's up to n = m. Next, the linearly 
! interpolated slopes are calculated through a forward difference method. Finally, this yields, per 
! orbital i, the DDCS value for theta and omega_s, previously specified.
  type(Param), intent(in)             :: P               
  type(WF_data), intent(in)           :: F
  real(long), intent(in)              :: theta_rad
  real(long), intent(in)              :: omega_s
  integer                             :: slope_index, slope_all, n_f, l_f, n_all    
  real(long)                          :: omega_in, epsilon_i, E_cutoff
  integer                             :: i
  real(long)                          :: f_n(F%N * (P%lmax + 1))
  real(long)                          :: epsilon_f(F%N * (P%lmax + 1))
  real(long)                          :: histogram(F%N * (P%lmax + 1)) ! Too big 
  real(long)                          :: slopes(F%N)                   ! Too big 
  integer                             :: E_index(P%lmax + 1)   
  real(long)                          :: slope_at_E
  real(long)                          :: DDCS_orbital(F%N_orb)
                                   
  DDCS_orbital(:) = ZERO
  omega_in = P%PE_start 


! Preparing all final orbital energies
  do l_f = 0, P%lmax 
    do n_f = 1, F%N
      n_all = F%N * l_f + n_f
      epsilon_f(n_all) = F%e_all(n_all)
   end do
  end do

  orbitals: do i = 1, F%N_orb
    epsilon_i = F%e_orb(i) 

! The cutoff energy value dictated by the Dirac-function.
    E_cutoff = omega_in - omega_s + epsilon_i 
 
    f_n = calculate_f_n(P, F, theta_rad, omega_s, i)
    slope_at_E = ZERO
    E_index(:) = F%N

    final_l: do l_f = 0, P%lmax 
!   Finding the cut-off value per l_f 
      do n_f = 1, F%N
        n_all = F%N * l_f + n_f
        if (n_all == F%N * (P%lmax + 1)) cycle
        if ( (E_cutoff .gt. epsilon_f(n_all)) .AND. (E_cutoff .le. epsilon_f(n_all + 1)) ) then
          E_index(l_f + 1) = n_f + 1            ! Index 0 doesn't make sense. Hence the +1.
          exit       
        endif
      end do 

!   Calculating the histograms
      do n_f = 1, E_index(l_f + 1)
        n_all = F%N * l_f + n_f
        histogram(n_all) = sum (f_n(F%N * l_f + 1:F%N * l_f + n_f))   ! sum_{n_f} 
      end do

!   Calculating the slopes 
      do slope_index = 1, E_index(l_f + 1) - 1
        slope_all = F%N * l_f + slope_index
        slopes(slope_index) = ( histogram(slope_all + 1) - histogram(slope_all) ) / &
        &abs( epsilon_f(slope_all + 1) - epsilon_f(slope_all) )
      end do

      slope_at_E = slope_at_E + slopes(E_index(l_f + 1) - 1) 
    end do final_l

    DDCS_orbital(i) = slope_at_E
  end do orbitals

end function calculate_compton_nonuniform_DDCS 

!=======================================================================================================

subroutine print_compton_nonuniform_DDCS(P,F)
! called using -compton_nonuni_DDCS. 
  type(Param), intent(in)             :: P               
  type(WF_data), intent(in)           :: F
  integer                             :: OUTPUT
  integer                             :: N_omega, N_theta
  real(long), allocatable, dimension(:):: omega_s_array, theta_array
  real(long)                          :: omega_in
  real(long)                          :: d_omega_eV, d_omega_au
  real(long)                          :: d_theta_deg, d_theta_rad, min_theta_deg, min_theta_rad   
  integer                             :: a, i, j
  real(long)                          :: DDCS_orbital(F%N_orb)
                                  
  OUTPUT = P%OUTPUT 

  N_omega = P%compton_N_omega
  N_theta = P%compton_N_theta

  d_omega_eV = P%compton_d_omega
  d_omega_au = dble(d_omega_eV) * EV2AU  
         
  d_theta_deg = P%compton_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  min_theta_deg = P%compton_min_theta
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 

! Angular grid
  allocate(theta_array(N_theta))                
  theta_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /)

  omega_in = P%PE_start
! Energy grid
  allocate(omega_s_array(N_omega))
  omega_s_array(:) = (/ ( omega_in - dble(a) * d_omega_au , a=1, N_omega ) /) 

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the Compton scattering module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates the double differential cross-section (DDCS) for a & 
                              & scattering angle theta, and a scattered energy omega_s.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 3a13, 1a12, 1a35 )") 'Theta (rad)', 'Omega_s (eV)', 'Thomson_CS ', 'Total DDCS',&
                                           & 'Orbital contributions to DDCS'

  do i = 1, N_theta
    do j = 1, N_omega
      DDCS_orbital = calculate_compton_nonuniform_DDCS(P,F, theta_array(i), omega_s_array(j)) 
      write(OUTPUT,'(99ES13.3)') theta_array(i), omega_s_array(j) * AU2EV, Thomson_CS(theta_array(i)), &
                                 &sum( DDCS_orbital(1:F%N_orb) ), DDCS_orbital(:)
    end do 
  end do
  
end subroutine print_compton_nonuniform_DDCS 

!=======================================================================================================

subroutine print_compton_nonuniform_energy_DCS(P,F)
! called using -compton_nonuni_energy_resolved. Prints the angle-integrated DCS for a range of omega_s.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  integer                              :: OUTPUT
  integer                              :: N_omega, N_theta
  real(long), allocatable, dimension(:):: omega_s_array, theta_rad_array, sin_theta_rad
  real(long)                           :: omega_in
  real(long)                           :: d_omega_eV, d_omega_au
  real(long)                           :: d_theta_deg, d_theta_rad, min_theta_deg, min_theta_rad   
  integer                              :: a, i, j, m
  real(long)                           :: DDCS_orbital(P%compton_N_theta, F%N_orb)
  real(long)                           :: DCS_energy(P%compton_N_omega, F%N_orb)
  real(long)                           :: DCS_energy_tot(P%compton_N_omega)
  real(long)                           :: total_CS
                                  
  OUTPUT = P%OUTPUT 

  N_omega = P%compton_N_omega
  N_theta = P%compton_N_theta

  d_omega_eV = P%compton_d_omega
  d_omega_au = dble(d_omega_eV) * EV2AU  
         
  d_theta_deg = P%compton_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  min_theta_deg = P%compton_min_theta
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 

! Angular grid
  allocate(theta_rad_array(N_theta))                
  theta_rad_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /)
  allocate(sin_theta_rad(N_theta))  
  sin_theta_rad = (/ (dsin(theta_rad_array(m) ) , m=1, N_theta) /)              

  omega_in = P%PE_start
! Energy grid
  allocate(omega_s_array(N_omega))
  omega_s_array(:) = (/ ( omega_in - dble(a) * d_omega_au , a=1, N_omega ) /) 

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the the Compton scattering module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates the differential cross-section (DCS) for a range of& 
                              & scattered energy omega_s. An angle integration was performed'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 1a15, 1a20, 1a35 )") 'Omega_s(eV)', 'Energy-resolved DCS', &
                                            &'Orbital contributions to DCS'

! Angular integration
  DDCS_orbital(:,:) = ZERO
  do i = 1, N_omega
    do j = 1, N_theta
      DDCS_orbital(j,:) = calculate_compton_nonuniform_DDCS (P,F, theta_rad_array(j), omega_s_array(i)) 
    end do 
    do j = 1, F%N_orb
      DCS_energy(i,j) = sum( DDCS_orbital(1:N_theta,j)*sin_theta_rad(1:N_theta) )
      DCS_energy(i,j) = DCS_energy(i,j) - HALF * ( DDCS_orbital(N_theta,j)*sin_theta_rad(N_theta) ) - &
                 &HALF * ( DDCS_orbital(1,j) * sin_theta_rad(1) )
      DCS_energy(i,j) = ( TWO * PI ) * d_theta_rad * DCS_energy(i,j)
    end do 
    DCS_energy_tot(i) = sum(DCS_energy(i,1:F%N_orb))
    write(OUTPUT,'(99ES13.3)') omega_s_array(i) * AU2EV, DCS_energy_tot(i), DCS_energy(i,:)
  end do

! Energy integration
  total_CS = sum(DCS_energy_tot(1:N_omega)) 
  total_CS = total_CS - HALF * DCS_energy_tot(N_omega) - HALF * DCS_energy_tot(1)
  total_CS = total_CS * d_omega_au

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an energy integration over the given energy range& 
                                           & from', omega_s_array(1) * AU2EV, ' eV to ', &
                                           &omega_s_array(N_omega) * AU2EV, ' eV.'
  write(OUTPUT,"('# the total cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'

end subroutine print_compton_nonuniform_energy_DCS

!=======================================================================================================

subroutine print_compton_nonuniform_angle_DCS(P,F)
! Called using -compton_nonuni_angle_resolved. Prints the angle-integrated DCS for a range of omega_s.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  integer                              :: OUTPUT
  integer                              :: N_omega, N_theta
  real(long), allocatable, dimension(:):: omega_s_array, theta_rad_array, sin_theta_rad
  real(long)                           :: omega_in
  real(long)                           :: d_omega_eV, d_omega_au
  real(long)                           :: d_theta_deg, d_theta_rad, min_theta_deg, min_theta_rad   
  integer                              :: a, i, j, m
  real(long)                           :: DDCS_orbital(P%compton_N_omega, F%N_orb)
  real(long)                           :: DCS_angle(P%compton_N_theta, F%N_orb)
  real(long)                           :: DCS_angle_tot(P%compton_N_theta)
  real(long)                           :: total_CS
                                  
  OUTPUT = P%OUTPUT 

  N_omega = P%compton_N_omega
  N_theta = P%compton_N_theta

  d_omega_eV = P%compton_d_omega
  d_omega_au = dble(d_omega_eV) * EV2AU  
         
  d_theta_deg = P%compton_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  min_theta_deg = P%compton_min_theta
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 

! Angular grid
  allocate(theta_rad_array(N_theta))                
  theta_rad_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /)
  allocate(sin_theta_rad(N_theta))  
  sin_theta_rad = (/ (dsin(theta_rad_array(m) ) , m=1, N_theta) /)              

  omega_in = P%PE_start
! Energy grid
  allocate(omega_s_array(N_omega))
  omega_s_array(:) = (/ ( omega_in - dble(a) * d_omega_au , a=1, N_omega ) /) 

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the the Compton scattering module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates the differential cross-section (DCS) for a range of & 
                              &scattering angles theta. An energy integration was performed'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 1a12, 1a12, 1a20, 1a35 )") 'Theta(rad)', 'Thomson CS', 'Angle-resovled DCS', &
                                            &'Orbital contributions to DCS'

! Energy integration
  DDCS_orbital(:,:) = ZERO
  do i = 1, N_theta
    do j = 1, N_omega
      DDCS_orbital(j,:) = calculate_compton_nonuniform_DDCS(P,F, theta_rad_array(i), omega_s_array(j)) 
    end do 
    do j = 1, F%N_orb
      DCS_angle(i,j) = sum( DDCS_orbital(1:N_omega,j))
      DCS_angle(i,j) = DCS_angle(i,j) - HALF * DDCS_orbital(N_omega,j) - HALF * DDCS_orbital(1,j)
      DCS_angle(i,j) = DCS_angle(i,j) * d_omega_au
    end do
    DCS_angle_tot(i) = sum(DCS_angle(i,1:F%N_orb))
    write(OUTPUT,'(99ES13.3)') theta_rad_array(i), Thomson_CS(theta_rad_array(i)), DCS_angle_tot(i), &
                               &DCS_angle(i,:)
  end do

! Angle integration
  total_CS = sum(DCS_angle_tot(1:N_theta)*sin_theta_rad(1:N_theta)) 
  total_CS = total_CS - HALF * (DCS_angle_tot(N_theta)*sin_theta_rad(N_theta)) & 
             &- HALF * (DCS_angle_tot(1)*sin_theta_rad(1))
  total_CS = (TWO * PI) * total_CS * d_theta_rad

  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a,f4.1,a,f5.1,a)") 'Performing an angular integration over the given theta range& 
                                           & from', min_theta_deg, ' deg to ', &
                                           &theta_rad_array(N_theta)*180./PI, ' deg'
  write(OUTPUT,"('# the total cross section is: ')")
  write(OUTPUT,"('# ', ES13.5,a)") total_CS, ' (atomic units)'
  write(OUTPUT,"('# ', ES13.5,a)")  total_CS * AU2M**2 / 1.0e-22_long, ' (in megabarn)'

end subroutine print_compton_nonuniform_angle_DCS

!======================================================================================================= 
!=====================================End of nonuniform section=========================================
!=======================================================================================================

function momentum_transfer(omega_in, omega_s, theta) result (Q)
  real(long), intent(in) :: omega_in, omega_s, theta
  real(long) :: Q
  real(long) :: alpha
  alpha = ONE / SPEED_C_AU
  Q = omega_in**2  + omega_s**2 - TWO * omega_in * omega_s * dcos(theta)
  Q = alpha * dsqrt(Q)
end function momentum_transfer

function Thomson_CS(theta)
  real(long) :: Thomson_CS, theta, alpha
  alpha = ONE / SPEED_C_AU                        !fine-structure constant = 1 / 137
  Thomson_CS = alpha**4 * HALF * ( ONE + (dcos(theta))**2 )  
!Polarization factor for unpolarized light or for linearly polarized light of symmetric object (e.g. after averaging over orientations).      
end function Thomson_CS

function free_electron_background(F, theta)
  real(long)          :: free_electron_background
  type(WF_data),intent(in) :: F
  real(long)          :: theta
  integer              :: n_elec
  n_elec = int(F%Z) - sum(F%config(1:F%N_orb,3))
  free_electron_background = dble(n_elec)*Thomson_CS(theta)
end function free_electron_background




function r_integral_linear( N, dr, f, g, j ) result (S)
  real(long) :: S
  integer, intent(in) :: N
  real(long), intent(in) :: f(:), g(:), j(:)
  real(long) :: dr
  !
  ! rwf and u are already multiplied by r
  ! integral is evaluated by the trapezoid rule.
  ! 
  S = sum( f(1:N) * g(1:N) * j(1:N) )                                         ! f(1:N) rather than f(:)
  S = S - HALF * ( f(N) * g(N) * j(N) ) !- HALF *( f(1) * g(1) *j(1) )         ! according to trapezoid rule
  S = S * dr
end function r_integral_linear


function r_integral_nonlinear1( n, dr, u_final, u_initial_deriv, j ) result (s)
  integer, intent(in) :: n
  real(long), intent(in) :: dr
  real(long), intent(in) :: u_final(:), u_initial_deriv(:), j(:)
  real(long) :: s,s1,sn
  integer :: i
  
  s = ZERO
  do i = 2,n-1
    s = s +  u_final(i) * j(i) * u_initial_deriv(i)                                 
  end do
  s1 = HALF * u_final(1) * j(1)   * u_initial_deriv(1)                                 
  sn = HALF * u_final(n) * j(n)   * u_initial_deriv(n)                                 
  s = dr * (s + s1 + sn)

end function r_integral_nonlinear1


function r_integral_nonlinear2( n, dr, u_final, u_initial, j ) result (s)
  integer, intent(in) :: n
  real(long), intent(in) :: dr
  real(long), intent(in) :: u_final(:), u_initial(:), j(:)
  real(long) :: s, s1, sn
  integer :: i

  s = ZERO
  do i = 2,n-1
    s = s +  u_final(i) * (1./i*dr)*j(i) * u_initial(i)                                 
  end do
  s1 = HALF * u_final(1) * (1./1*dr)*j(1) * u_initial(1)                                 
  sn = HALF * u_final(n) * (1./n*dr)*j(n) * u_initial(n)                                 
  s = dr * (s + s1 + sn)

end function r_integral_nonlinear2



function get_continuum_wf(F, P, omega_s, lmax) result (cont_wf)
  ! this function creates an array of all continuum wavefunctions,
  ! that are needed to calculate the DDSP at different angles.
  ! the wavefunctions are stored in the columns of the array,
  ! the first lmax+1 columns are functions for the energies of the
  ! 1s subshell, the next lmax+1 columns for the 2s subshell, etc.

  type(WF_data), intent(in)           :: F
  type(Param), intent(in)             :: P
!  type(Spline)                        :: SP
  real(long)                          :: omega_s, E
  real(long), dimension(F%N_uni, (lmax+1)*F%N_orb) :: cont_wf
  integer                              :: lmax, i, l
  real(long)                          :: ll
  real(long)                           :: u(F%N_uni), up(F%N_uni)
  
  do i=1, F%N_orb
    E = F%e_orb(i) + P%PE_start - omega_s
    do l = 0, lmax
      ll = dble(l)
      if( E <= 0 ) then
        cont_wf(:, (i-1)*(lmax+1) + (l+1) ) = ZERO
      end if
      if ( E > 0 ) then
        call unbound(F%N_uni, u, up, F%r_uni, ll, F%Z, E, F )
        cont_wf(:, (i-1)*(lmax+1) + (l+1) ) = u(:)
      end if
    end do                                        
  end do

end function get_continuum_wf


function interpolate_bound_wf_on_uniform_grid(F, m) result(u)  
  ! Interpolates r * wf of u_all(:,m) on the uniform grid 
  
  type(WF_data), intent(in)        :: F
  integer                          :: m
  real(long), dimension(F%N_uni)  :: u
  !local
  type(Spline)                    :: SP
  integer                          :: i

  real(long), dimension(F%N+2)  :: tmp
  
  tmp(:) = ZERO
    SP = Spline_create( F%r_int(:), tmp(:) )
  
!  tmp(2:F%N+1) = F%r(:) * F%u_all(:,m) * sqrt(FOUR*PI) ! YOU NEED TO AVOID sqrt(4pi)
  tmp(2:F%N+1) = F%u_all(:,m)
  call Spline_reuse( SP, tmp(:) )
  u(:) = (/ ( Spline_interpolate( SP, F%r_uni(i) ), i=1,F%N_uni ) /)

end function interpolate_bound_wf_on_uniform_grid



function factorial (n) result (res)
  implicit none
  integer, intent (in) :: n
  integer :: res
  integer :: i
  res = product ((/(i, i = 1, n)/))
 
end function factorial
 
function binomial (n, k) result (res)
  implicit none
  integer, intent (in) :: n
  integer, intent (in) :: k
  integer :: res
  res = factorial (n) / (factorial (k) * factorial (n - k))
 
end function binomial



!====================================================================================================
!====================================================================================================
!====================================================================================================
!FOR TESTING:

function impulse_approximation(omega_s, omega_in, theta) result(DDSP)
!
!IMPULSE APPROXIMATION FOR 1S HYDROGEN ELECTRON ONLY!!!!!!!
!
  real(long), intent(in)   :: omega_s, omega_in, theta
  real(long)               :: DDSP
  real(long)              :: Q, J, p_q !, gamma_1s, gamma_2p
   real(long) :: alpha
   alpha = ONE / SPEED_C_AU
  
  Q = momentum_transfer(omega_in, omega_s, theta)
  p_q = (omega_in - omega_s)/Q - HALF * Q

!---------------------------------------------------
!Hydrogen 1s:
  J = 8.0_long / ( 3.0_long * PI * (ONE + (p_q)**2)**3 )
!Carbon 1s:
  !gamma_1s = 5.54722_long
!  gamma_1s = 6.0_long
  !J = 8.0_long * gamma_1s**5 / ( 3.0_long * PI * ( gamma_1s**2 + (p_q)**2)**3 )
!Carbon 2p:
!  gamma_2p = 1.40396 !from fit in gnuplot with Biggs' data
!  J = 32.0_long * gamma_2p**7 / ( 15.0_long * PI) * ( gamma_2p**2 + 5* (p_q)**2) * ( gamma_2p**2 + p_q**2 )**(-5)
  
 DDSP  = Thomson_CS(theta) * omega_s/omega_in * Q**(-1) * J
  
end function impulse_approximation

function exact_impulse_approximation(omega_s, omega_in, theta) result(DDSP_IA)
!impulse approximation for 1s Hydrogen electron, Eisenberger/Platzmann Eq. (10)
  real(long), intent(in)   :: omega_s, omega_in, theta
  real(long)               :: DDSP_IA
  real(long)              :: p, omega, M, factor, k

!   omega = omega_in - omega_s
!   m_e = 1.0d0
!   a = 1.0d0  !Bohr radius in a.u. 
!   k = momentum_transfer(omega_in, omega_s, theta)
!   
!   p = sqrt( TWO * m *( omega - 13.6d0 ) )
!   
!   factor = PI**2 * 8.0_long**3 * a**2 / p
!   factor = factor * (1 - exp(-TWO * PI /(p*a)))**(-1)
!   factor = factor * ( k**4 * a**4 + ONE / THREE * k**2 * a**2 * (ONE + p**2*a**2))
!   factor = factor * ( (k**2 * a**2 + ONE - p**2 * a**2)**2 + FOUR * p**2 * a**2 )**(-3)
!   
!   M = factor * exp( - TWO/(p*a) * datan( TWO * p * a /(ONE + k**2 * a**2 - p**2 * a**2)))
!   
!   
!    DDSP  = Thomson_CS(theta) * omega_in / omega_s * M
! 
!----------------------------------------------------------
  omega = omega_in - omega_s 
  k = momentum_transfer(omega_in, omega_s, theta) 
  
  p = dsqrt( TWO * ( omega - 13.6d0 * EV2AU ) )
   
  factor = PI**2 * 8.0d0**3 / p
  factor = factor / (ONE - exp( - TWO * PI /p))
  factor = factor * ( k**4 + (ONE / THREE) * k**2 * (ONE + p**2))
  factor = factor * ( (k**2 + ONE - p**2 )**2 + FOUR * p**2 )**(-3)
  factor = factor * exp( - (TWO / p) * abs(datan( TWO * p / (ONE + k**2 - p**2))))
  
  M = FOUR*PI*(TWO*PI)**(-3) * p * factor
  DDSP_IA  = Thomson_CS(theta) * M * omega_s / omega_in  
end function exact_impulse_approximation


function compton_profile(DDSP, omega_s, omega_in, theta) result (J)
  
  real(long), intent(in)   :: DDSP, omega_s, omega_in, theta
  real(long)              :: J, p_q, Q
  
  Q = momentum_transfer(omega_in, omega_s, theta)
  p_q = (omega_in - omega_s)/Q - HALF * Q
  
  J = DDSP / Thomson_CS(theta) * omega_in / omega_s * Q 

end function compton_profile


end module Compton
