!----------------------------------------------------------------------------
! coord_system.f90
!----------------------------------------------------------------------------
! Copyright (C) 2005 Sang-Kil Son
! Copyright (C) 2010 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! 1D spherical coordinate system
!   written by Son, Sang-Kil in Sep. 2005
!
! This module contains all necessary operations in the 1D spherical coord.
! - class Coordinate
!   - Del2(:,:): Laplacian
!   - Del_r(:,:): r component of Gradient, d/dr
!   - weight(:): weights for integration
!   - wf_coeff(:): simply multiply eigenvector by it to get w.f.
! - F generate_Coordinate( r ): make coordinate from dicretized "r"
! - F Integral( G, v ): compute integral
! - F abs_Gradient( G, v ): = | Del v |
!----------------------------------------------------------------------------
module coord_system
  use constants
  use gps
  implicit none
  type Coordinate
    integer :: N_dim
    integer :: N
    real(long) :: rmax, wf_coeff_b
    real(long), allocatable :: r(:)         ! radial coordinate
    real(long), allocatable :: U(:)         ! given potential
    real(long), allocatable :: E(:)         ! energy or eigenvalues
    real(long), allocatable :: H(:,:)       ! Hamiltonian
    real(long), allocatable :: Deriv(:,:)   ! d/dr (:= Deriv in gps.f90)
    real(long), allocatable :: Del_r(:,:)   ! r component of Gradient (:= D1)
    real(long), allocatable :: Del2(:,:)    ! Laplacian
    real(long), allocatable :: weight(:)    ! for integration
    real(long), allocatable :: wf_coeff(:)  ! transforming b/w wf and eigenvec.
    real(long), allocatable :: D2_boundary(:)
  end type Coordinate
contains

!----------------------------------------------------------------------------
! generate 1D spherical coordinate system
!----------------------------------------------------------------------------
function generate_Coordinate( r ) result (G)
  type(Coordinate) :: G
  type(Grid), intent(in) :: r
  integer :: N
  real(long) :: a, rp_b, wf_coeff_b
  !
  ! set dimensions
  !
  G%N_dim = 1
  N = r%N
  G%N = N
  G%rmax = r%Rmax
  !
  ! memory allocation
  !
  allocate( G%r(N) )
  allocate( G%U(N), G%E(N), G%H(N,N), G%Deriv(N,N), G%Del_r(N,N), G%Del2(N,N) )
  allocate( G%weight(N), G%wf_coeff(N) )
  allocate( G%D2_boundary(N) )
  !
  ! construct operators for this system
  !
  G%r(:) = r%r(:)
  G%Deriv(:,:) = r%Deriv(:,:)   ! d/dr (check Deriv in gps.f90)
  G%Del_r(:,:) = r%D1(:,:)      ! r component := d/dr (check d1 in gps.f90)
  G%Del2(:,:)  = r%D2(:,:)
  !
  ! define weight and wf_coeff
  !            4 PI r^2         <-- for spherical coord. integration
  ! w(:) = --------------------
  !        Wavefunction_coeff^2 <-- Gauss-Lobatto integration
  !
  !   Wavefunction_coeff = sqrt( (N+1)(N+2)/2 / r'(x) ) * P_N+1(x)
  !
  !G%weight(:) = FOUR * PI * r%r(:)**2 * r%weight(:)
  G%weight(:) = r%weight(:)     ! just put the same as used in 1D
  !
  ! Wavefunction_coeff: just put the same as used in 1D (Gauss-Lobatto int.)
  ! The following consideration in 3D is disputed.
  ! ------------------------------------------------------------------------
  ! A -> chi = r * psi(r); so, we divide it by Grid_r.
  ! And remember INT r**2 sin theta dr d(theta) d(phi) gives 4PI INT r^2 dr.
  ! Thus, in order to be normalized, wf is divided by sqrt(4PI).
  ! ------------------------------------------------------------------------
  !
  G%wf_coeff(:) = r%wf_coeff(:) !/ r%r(:) / sqrt( FOUR * PI )
  !
  ! set up boundary variables at Rmax
  !
  if ( G%rmax /= ZERO ) then
    a = TWO * r%L / r%Rmax
    rp_b = r%L * (TWO+a) / a**2
    wf_coeff_b = sqrt( dble((N+1)*(N+2)/2) / rp_b )
    G%wf_coeff_b = wf_coeff_b !/ G%rmax / sqrt( FOUR * PI )
    G%D2_boundary(:) = - TWO / ( r%x(:) - ONE )**2 / r%rp(:) / rp_b
  end if
end function generate_Coordinate


!----------------------------------------------------------------------------
! integrate any "radial" function f(r) in 3D
!
!   INT d^3r f(r) = 4 PI INT dr r^2 f(r)
!
! - radial integration with the Gauss-Lobatto integration
! - angular integration is given by 4 PI
! - integration range is given by f(:)
!
! Note that integration of an expectation value, < psi | A | psi >,
! involving the density with u(r) becomes very simple.
! Because our density is always spherically averaged,
!
!                INT dr^3 rho (r,theta,phi)       SUM_i | u_i(r) |^2
!   rho(r) = ---------------------------------- = ------------------
!            INT INT sin(theta) d(theta) d(phi)       4 PI r^2
!
!          = SUM_i rho_i(r),
!
! where rho_i(r) = | u_i(r) / sqrt(2 PI) / r |^2.  In this way, the factor
! of 4 PI r^2 will be cancelled out in the above expression.
! For example, if one evaluates <A> = < psi_i | A(r) | psi_i >, then
!
!   <A> = INT dr^3 rho_i(r) A(r) = INT dr | u_i(r) |^2 A(r)
!
! This is simply evaluated by <A> = sum( G%weight(:) * F%u(:,i)**2 * A(:) ).
! Likewise, < u | A | v > = sum( G%weight(:) * u(:) * A(:) * v(:) )
!           where u and v are radial wavefunctions [ u = r R(r) ].
!
! Here, we may use radial_integral( G, v ) assuming that v involves r**2.
!   radial_integral( G, v ) = sum( G%weight(:) * v(:) )
!----------------------------------------------------------------------------
function integral( G, v ) result (I)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: v(:)
  real(long) :: I
  I = FOUR * PI * sum( v(:) * G%weight(1:size(v)) * G%r(1:size(v))**2 )
end function integral

function radial_integral( G, v ) result (I)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: v(:)
  real(long) :: I
  I = sum( G%weight(1:size(v)) * v(:) )
end function radial_integral

!----------------------------------------------------------------------
! compute | Grad f |
!             d  ^    1     d      ^           1          d    ^
!   Grad f = --- r + --- ------- theta + -------------- ----- phi
!            d r      r  d theta          r sin(theta)  d phi
!----------------------------------------------------------------------
!function abs_Gradient( G, v ) result (I)
!  real(long) :: I
!  type(Coordinate), intent(in) :: G
!  real(long) :: v(:)
!  I = ZERO
!end function abs_Gradient


!----------------------------------------------------------------------
! transform_A_to_wf: convert matrix element A into wavefunction
! transform_wf_to_A: convert wavefunction into matrix element A
!
!   A (computation domain) : eigenvectors of H
!   wf (real space) : wavefunctions containing r, theta and phi
!----------------------------------------------------------------------
!function transform_A_to_wf( G, A ) result (wf)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: A(:)
!  real(long) :: wf(size(A))
!  wf(:) = A(:) * G%wf_coeff(:)
!end function transform_A_to_wf
!
!function transform_wf_to_A( G, wf ) result (A)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: wf(:)
!  real(long) :: A(size(wf))
!  integer :: j
!  A(:) = wf(:) / G%wf_coeff(:)
!end function transform_wf_to_A

end module coord_system
