!----------------------------------------------------------------------------
! coulomb.f90
!----------------------------------------------------------------------------
! Copyright (C) 2020 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2020 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! coulomb integral module
!   calculate double radial integral of 1/|r1-r2|
!   S(i1,i2,i3,i4,K) = int dr1 int dr2 
!     [ u(i1,1) u(i2,2) r_min^K / r_max^(K+1) u(i3,1) u(i4,2) ]
!            |       |                             |       |
!            r1      r2                            r1      r2
!
! double_integral: originally implemented function for Auger (obsolete)
!                  r1, r2 - uniform grid
! double_integral_mod: new implemented function for Auger decay calculation
!                  r1 - non-uniform grid, r2 - uniform grid
! double_integral_nonuniform
!                  r1,r2 - non-uniform grid
!----------------------------------------------------------------------------
module coulomb
  use constants
  use stdio
  use read_param
  use wavefunction
  implicit none
contains

!----------------------------------------------------------------------------
! u1, u2, u3, and u are already multiplied by r
! double integrals are evaluated by the trapezoid rule.
! S = int [ u(i1,1) u(i2,2) r_min^K / r_max^(K+1) u(i3,1) u(2) ] dr1 dr2
! SEE: Eq. (12) in Son, Young & Santra, PRA 83, 033402 (2011).
!----------------------------------------------------------------------------
function double_integral( N, r, u1, u2, u3, u, K ) result (R_K)
  real(long) :: R_K
  integer, intent(in) :: N
  real(long), intent(in) :: r(:), u1(:), u2(:), u3(:)
  real(long), intent(in) :: u(:)        ! u: continuum state
  integer, intent(in) :: K
  integer :: i
  real(long) :: dx, S, r_exp(size(r,1))
  !real(long) :: r_min, r_max
  R_K = ZERO
  dx = r(2) - r(1)      ! equally-spaced
  !
  ! Below is a direct implementation of Eq. (12).
  ! 
  !do i = 2, N           ! it starts from 2 because r(1) = ZERO
  !  do j = 2, N         ! it starts from 2 because r(1) = ZERO
  !    r_min = min( r(i), r(j) )
  !    r_max = max( r(i), r(j) )
  !    S = u1(i)*u2(j)*r_min**dble(K)/r_max**dble(K+1)*u3(i)*u(j)
  !    if ( i == N ) S = S * HALF
  !    if ( j == N ) S = S * HALF
  !    R_K = R_K + S * dx * dx
  !  end do
  !end do
  !
  ! However, it has been much improved as follows...
  ! It's crucial because double_integral is the most time-consuming part.
  !
  do i = 2, N                   ! it starts from 2 because r(1) = zero
    if ( K == 0 ) then
      r_exp(2:i-1) = ONE / r(i)
      r_exp(i+1:N) = ONE / r(i+1:N)
    else if ( K == 1 ) then
      r_exp(2:i-1) = r(2:i-1) / r(i) / r(i)
      r_exp(i+1:N) = r(i) / r(i+1:N) / r(i+1:N)
    else
      r_exp(2:i-1) = ( r(2:i-1) / r(i) )**K / r(i)
      r_exp(i+1:N) = ( r(i) / r(i+1:N) )**K / r(i+1:N)
    end if
    r_exp(1) = ZERO             ! it's always zero
    r_exp(i) = ONE / r(i)       ! r_i == r_j
    r_exp(N) = r_exp(N) * HALF  ! a half value for the trapezoidal rule
    S = u1(i) * u3(i) * sum( u2(1:N) * r_exp(1:N) * u(1:N) )
    R_K = R_K + S * dx**2
  end do
end function double_integral

!----------------------------------------------------------------------------
! Coulomb integral of 1/|r1-r2|
! computed by double radial integral with r1 and r2
!
!   SEE: Eq. (12) in Son, Young & Santra, PRA 83, 033402 (2011).
!   S = int [ u(i1,1) u(i2,2) r_min^K / r_max^(K+1) u(i3,1) u(2) ] dr1 dr2
!             ^^^^^^^ ^^^^^^^                       ^^^^^^^ ^^^^
!             u1(r1)  u2(r2)                        u3(r1)  u(r2)
!
! u1(r1), u3(r1): bound states of u1, u3 are smooth --> non-uniform grid
! u2(r2), u(r2):  rapidly oscillating continuum u --> uniform grid
! u1, u2, u3, and u are already multiplied by r
!----------------------------------------------------------------------------
function double_integral_mod( K, r1, weight, u1, u3, &
                                 r2, u2, u, threshold ) result(R_K)
  integer, intent(in) :: K
  real(long), intent(in) :: r1(:), weight(:), u1(:), u3(:)
  real(long), intent(in) :: r2(:), u2(:), u(:)
  real(long), intent(in) :: threshold
  real(long) :: R_K
  integer :: N1, N2, i, j
  real(long) :: dr, fr12(size(r1)), tmp, r_min, r_max
  N1 = size(r1) ! it's F%N, # of nonuniform grid points
  N2 = size(u2) ! it's F%N_uni, size(r2) might be larger than F%N_uni
  !
  ! truncate uniform grid individually by u2(r2), the same way as cont. rmax
  !
  do j = size(u2), 1, -1        ! size(u2) == F%N_uni
    if ( abs( u2(j) ) > threshold ) exit
  end do
  if ( j == 0 ) call error_message( "in double_integral_mod(): u2(r) = 0" )
  N2 = j                        ! reset up N2, which might be < size(u2)
  dr = r2(2) - r2(1)
  R_K = ZERO
  loop2: do j = N2, 1, -1       ! r2: uniform grid
    loop1: do i = 1, N1         ! r1: non-uniform grid (GPS grid)
      if ( r1(i) >= r2(j) ) then
        r_max = r1(i);  r_min = r2(j)
      else
        r_min = r1(i);  r_max = r2(j)
      end if
      if ( K == 0 ) then  ! separate out K=0 and 1 because a**K is expensive
        fr12(i) = ONE / r_max
      else if ( K == 1 ) then
        fr12(i) = ( r_min / r_max ) / r_max
      else
        fr12(i) = ( r_min / r_max )**K / r_max
      end if
    end do loop1
    tmp = sum( u1(:) * fr12(:)* u3(:) * weight(:) ) ! r1: on non-uniform grid
    R_K = R_K + u(j) * tmp * u2(j)                  ! r2: on uniform grid
  enddo loop2
  R_K = R_K * dr
end function double_integral_mod

!----------------------------------------------------------------------------
! Coulomb integral of 1/|r1-r2|
! computed by double radial integral with r1 and r2
!
!   S = int [ u(i1,1) u(i2,2) r_min^K / r_max^(K+1) u(i3,1) u(i4,2) ] dr1 dr2
!             ^^^^^^^ ^^^^^^^                       ^^^^^^^ ^^^^^^^
!             u1(r1)  u2(r2)                        u3(r1)  u4(r2)
!
! u1, u2, u3, u4: bound states defined on non-uniform grid (r1, r2)
!                 they are already multiplied by r
! ACHTUNG: be careful with ordering of u1, u2, u3, u4
!          different from double_integral_mod( ... u1, u3, ..., u2, u ...)
!----------------------------------------------------------------------------
function double_integral_nonuniform( K, r, weight, u1, u2, u3, u4 ) result(R_K)
  integer, intent(in) :: K
  real(long), intent(in) :: r(:), weight(:)
  real(long), intent(in) :: u1(:), u2(:), u3(:), u4(:)
  real(long) :: R_K
  integer :: N, i
  real(long) :: S, r_exp(size(r,1))
  N = size(r)   ! it's F%N, # of nonuniform grid points
  R_K = ZERO
  do i = 1, N
    if ( K == 0 ) then
      r_exp(1:i-1) = ONE / r(i)
      r_exp(i+1:N) = ONE / r(i+1:N)
    else if ( K == 1 ) then
      r_exp(1:i-1) = r(1:i-1) / r(i) / r(i)
      r_exp(i+1:N) = r(i) / r(i+1:N) / r(i+1:N)
    else
      r_exp(1:i-1) = ( r(1:i-1) / r(i) )**K / r(i)
      r_exp(i+1:N) = ( r(i) / r(i+1:N) )**K / r(i+1:N)
    end if
    r_exp(i) = ONE / r(i)       ! r_i == r_j
    S = sum( u2(1:N) * r_exp(1:N) * u4(1:N) * weight(1:N) )
    R_K = R_K + S * u1(i) * u3(i) * weight(i)
  end do
end function double_integral_nonuniform

!----------------------------------------------------------------------------
! print out all possible double radial integrals within a given configuration
!----------------------------------------------------------------------------
subroutine print_double_integral( P, F )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  real(long) :: R12_K
  integer :: i1, i2, i3, i4
  integer :: n1, n2, n3, n4
  integer :: l1, l2, l3, l4, K
  integer :: n_max, l_max
  integer :: OUTPUT
  character(len=80) :: format_str
  OUTPUT = P%OUTPUT
  n_max = maxval( F%config(1:F%N_orb,1) )
  l_max = maxval( F%config(1:F%N_orb,2) )
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Double radial integral of 1/|r1-r2|:')")
  write(OUTPUT, "('# ',30x,a)") "  r_<^K"
  write(OUTPUT, "('# ',2x,a)") &
    "R_K = int [ u_i1(1) u_i2(2) --------- u_i3(1) u_i4(2) ] dr1 dr2"
  write(OUTPUT, "('# ',30x,a)") "r_>^(K+1)"
  if ( n_max < 10 .and. l_max < 10 ) then
    write(OUTPUT, "('# ',a1,x,4a3,3x,a)") "K", "i1", "i2", "i3", "i4", "R_K"
    format_str = "(i3,x,4(x,2i1),x,es18.10)"
  else
    write(OUTPUT, "('# ',a1,x,4a8,3x,a)") "K", "i1", "i2", "i3", "i4", "R_K"
    format_str = "(i3,x,4(' [',i2,',',i2,']'),x,es18.10)"
  end if
  do i1 =  1, F%N_orb;  n1 = F%config(i1,1);  l1 = F%config(i1,2)
  do i2 =  1, F%N_orb;  n2 = F%config(i2,1);  l2 = F%config(i2,2)
  do i3 = i1, F%N_orb;  n3 = F%config(i3,1);  l3 = F%config(i3,2)
  do i4 = i2, F%N_orb;  n4 = F%config(i4,1);  l4 = F%config(i4,2)
    do K = 0, P%K_max
      R12_K = double_integral_nonuniform( K, F%r(:), F%weight(:), &
                F%u(:,i1), F%u(:,i2), F%u(:,i3), F%u(:,i4) )
      write(OUTPUT, format_str) K, n1, l1, n2, l2, n3, l3, n4, l4, R12_K
    end do
    write(OUTPUT, "()")
  end do
  end do
  end do
  end do
end subroutine print_double_integral

!function double_integral_mod(k, N1, r1, u, u1, N2, r2, weight, u3, u2) result(res)
!  ! rapidly oscillating continuum u is the function of r1.
!  ! integration is done with uniform grid
!  !
!  ! final holes u2, u3 are smooth functions of r2
!  ! integration is done on the GPS grid r2 with its wieght
!  real(long), intent(in) :: r1(:), u(:), u1(:)
!  real(long), intent(in) :: r2(:), weight(:), u3(:), u2(:)
!  integer, intent(in) :: k, N1, N2
!  real(long) :: res
!  integer :: i, j
!  real(long) :: dr, fr12(size(r2,1)), tmp
!
!  res = ZERO
!  dr = r1(2) - r1(1)
!  do i = 1, N1
!     fr12(:) = ZERO
!     do j = 1, N2
!        if ( r1(i) >= r2(j) ) then
!           fr12(j) = ( r2(j) / r1(i) )**k / r1(i)
!        else
!           fr12(j) = ( r1(i) / r2(j) )**k / r2(j)
!        endif
!     enddo
!     ! integral for r2 coordinate on GPS grid
!     tmp = sum( u3(:) * fr12(:)* u2(:) * weight(:) )
!     ! integral for r1 coordinate on uniform grid
!     res = res + u(i) * tmp * u1(i)
!  enddo
!  res = res * dr
!
!end function double_integral_mod


function reduced_matrix( K, l, lp ) result (C)
  real(long) :: C
  integer, intent(in) :: K, l, lp
  real(long), external :: cleb
  !
  ! SEE: Eq. (11) in Son, Young & Santra, PRA 83, 033402 (2011).
  !      Note that there is an ERRATUM in Eq. (11).
  !      The denominator in Eq. (11) must be the numerator.
  !
  ! Clebsch-Gordan coeff.: See anglib.f90 and wikipedia
  !   < j1 m1 j2 m2 | j3 m3 > = (-1)^(j1-j2+m3) sqrt(2*j3+1) (( j1  j2  j3 ))
  !                                                          (( m1  m2 -m3 ))
  ! In comparison with Eq. (11), let j1 = l', j2 = K, j3 = l, and m1=m2=m3=0.
  !   < l' 0 K 0 | l 0 > = (-1)^(l'-K) sqrt(2*l+1) (( l' K  l ))
  !                                                (( 0  0  0 ))
  ! Thus, we need to compute
  !   < l || K || l' > = (-1)^l sqrt( (2*l+1)(2*l'+1) ) (( l' K  l ))
  !                                                     (( 0  0  0 ))
  !                    = (-1)^(l-l'+K) sqrt(2*l'+1) < l' 0 K 0 | l 0 >
  !
  C = sqrt( dble(2*lp+1) ) * cleb( lp*2, 0, K*2, 0, l*2, 0 )
  if ( mod(l-lp+K,2) == 1 ) C = -C
end function reduced_matrix

!----------------------------------------------------------------------------
! v_pqrs( F, p, q, r, s, S_l_ext ): Two-body integral calculation
!                                   with spinor index p, q, r, s
!   S_l_ext is a vector containing pre-calc. radial double integrals (optional)
!                                                      1
! v_pqrs = int d^3x int d^3x' phi^+_p(x) phi^+_q(x') ------ phi_r(x) phi_s(x')
!                                                    |x-x'|
!        = delta_{m_s_p,m_s_r} delta_{m_s_q,m_s_s} *
!          delta_{m_l_p+m_l_q,m_l_r+m_l_s} (-1)^{m_l_p - m_l_r} *
!          sqrt( (2l_r+1)(2l_s+1) / (2l_p+1)(2l_q+1) ) *
!          SUM_{l=0}^{inf} R^l( ip, iq, ir, is ) *
!          C(l_r l l_p; m_l_r, m_l_p - m_l_r, m_l_p) C(l_r l l_p; 0 0 0) *
!          C(l_s l l_q; m_l_s, m_l_q - m_l_s, m_l_q) C(l_s l l_q; 0 0 0)
!    where C is a Clebsch-Gordan coefficient
!          C(l1 l2 l3; ml1, ml2, ml3) = < l1 ml1 l2 ml2 | l3 ml3 >
!----------------------------------------------------------------------------
function v_pqrs( F, p, q, r, s, S_l_ext ) result (v)
  real(long) :: v
  type(WF_data), intent(in) :: F
  integer, intent(in), dimension(1:4) :: p, q, r, s
  real(long), intent(in), optional :: S_l_ext(0:)
  real(long) :: S_l
  integer :: N, l
  integer :: n_p, n_q, n_r, n_s
  integer :: l_p, l_q, l_r, l_s
  integer :: ml_p, ml_q, ml_r, ml_s
  integer :: ms_p, ms_q, ms_r, ms_s
  integer :: ip, iq, ir, is
  real(long), external :: cleb
  N = F%j_s
  n_p = p(1); l_p = p(2); ml_p = p(3); ms_p = p(4); ip = l_p*F%N + (n_p-l_p)
  n_q = q(1); l_q = q(2); ml_q = q(3); ms_q = q(4); iq = l_q*F%N + (n_q-l_q)
  n_r = r(1); l_r = r(2); ml_r = r(3); ms_r = r(4); ir = l_r*F%N + (n_r-l_r)
  n_s = s(1); l_s = s(2); ml_s = s(3); ms_s = s(4); is = l_s*F%N + (n_s-l_s)
  v = ZERO
  if ( ml_p + ml_q /= ml_r + ml_s ) return  ! => delta_{m_l_p+m_l_q,m_l_r+m_l_s}
  if ( ms_p /= ms_r ) return                ! => delta_{m_s_p,m_s_r}
  if ( ms_q /= ms_s ) return                ! => delta_{m_s_q,m_s_s}
  do l = 0, l_p + l_q + l_r + l_s           ! => SUM_{l=0}^{inf}
    if ( present(S_l_ext) ) then
      S_l = S_l_ext(l)
    else
      S_l = double_integral_nonuniform( &     ! => R^l( ip, iq, ir, is )
              l, F%r(1:N), F%weight(1:N), &
              F%u_all(1:N,ip), F%u_all(1:N,iq), &
              F%u_all(1:N,ir), F%u_all(1:N,is) )
    end if
    ! multiplying
    !   C(l_r l l_p; m_l_r, m_l_p - m_l_r, m_l_p) C(l_r l l_p; 0 0 0) *
    !   C(l_s l l_q; m_l_s, m_l_q - m_l_s, m_l_q) C(l_s l l_q; 0 0 0)
    S_l = S_l * &
          cleb( l_r*2, ml_r*2, l*2, (ml_p-ml_r)*2, l_p*2, ml_p*2 ) * &
          cleb( l_r*2,      0, l*2,             0, l_p*2,      0 ) * &
          cleb( l_s*2, ml_s*2, l*2, (ml_q-ml_s)*2, l_q*2, ml_q*2 ) * &
          cleb( l_s*2,      0, l*2,             0, l_q*2,      0 )
    v = v + S_l
  end do
  v = v * &         ! => sqrt( (2l_r+1)(2l_s+1) / (2l_p+1)(2l_q+1) )
      sqrt( dble( (2*l_r+1)*(2*l_s+1) ) / dble( (2*l_p+1)*(2*l_q+1) ) )
  !
  ! WARNING with mod( ) in Fortran 90
  !   mod( 0,2) =  0
  !   mod( 1,2) =  1
  !   mod(-1,2) = -1    <-- what is this?!
  !   mod(-2,2) =  0
  !   mod(-3,2) = -1    <-- ?!
  !     ...
  ! Thus, if ( mod( ml_p - ml_r, 2 ) == 1 ) v = -v gives a wrong answer!!!
  !
  if ( mod( ml_p - ml_r, 2 ) /= 0 ) v = -v      ! => (-1)^{m_l_p - m_l_r}
end function v_pqrs

!----------------------------------------------------------------------------
! To calculate direct Coulomb (E_J) and exchange (E_K) interactions
!    with orbital indexes i and j
! - F v_twobody( F, i, j )       : i and j are index for F%wf(:,i)
! - F v_twobody_all( F, ii, jj ) : ii and jj are index for F%wf_all(:,i)
! - F v_twobody_nl( F, n_i, l_i, n_j, l_j, S_l_J, S_l_K ) :
!       it actually performs E_J and E_K calculations with (n,l)-indexes
!       based on precalculated S_l_J() and S_l_K(), and function of v_pqrs()
!----------------------------------------------------------------------------
function v_twobody( F, i, j ) result (v)
  real(long) :: v(2)    ! v(1) = v_J and v(2) = v_K
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i, j
  real(long), allocatable :: S_l_J(:), S_l_K(:)
  integer :: n_i, l_i, n_j, l_j
  integer :: N, l, lmax
  N = F%j_s
  n_i = F%config(i,1); l_i = F%config(i,2)
  n_j = F%config(j,1); l_j = F%config(j,2)
  lmax = 2 * ( l_i + l_j )
  allocate( S_l_J(0:lmax), S_l_K(0:lmax) )
  !
  ! prepare S_l_J(:) (radial double integrals for direct Coulomb interactions)
  !     and S_l_K(:) (radial double integrals for exchange interactions)
  ! NOTE: the same orbital indexes i and j are not excluded
  !       instead, S_l_K(:) = S_l_J(:)
  !
  do l = 0, lmax
    S_l_J(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                 F%u(1:N,i), F%u(1:N,j), F%u(1:N,i), F%u(1:N,j) )
    if ( i == j ) then
      S_l_K(l) = S_l_J(l)
    else
      S_l_K(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                   F%u(1:N,i), F%u(1:N,j), F%u(1:N,j), F%u(1:N,i) )
    end if
  end do
  !
  ! compute J and K with v_twobody_nl(): v(1) = v_J and v(2) = v_K
  !
  v(:) = v_twobody_nl( F, n_i, l_i, n_j, l_j, S_l_J, S_l_K )
!TEST (taking into account self-interaction; NEED TO RECHECK)
  if ( i == j ) then
    v(:) = v(:) * F%config(i,3) * ( F%config(i,3) - 1 )
  else
    v(:) = v(:) * F%config(i,3) * F%config(j,3)
  end if
!TEST
  deallocate( S_l_J, S_l_K )
end function v_twobody

function v_twobody_all( F, ii, jj ) result (v)
  real(long) :: v(2)    ! v(1) = v_J and v(2) = v_K
  type(WF_data), intent(in) :: F
  integer, intent(in) :: ii, jj
  real(long), allocatable :: S_l_J(:), S_l_K(:)
  integer :: i, j, n_i, l_i, n_j, l_j
  integer :: N, l, lmax
  N = F%j_s
  i = mod(ii,F%N);  l_i = int( (ii-i) / F%N );  n_i = i + l_i
  j = mod(jj,F%N);  l_j = int( (jj-j) / F%N );  n_j = j + l_j
  lmax = 2 * ( l_i + l_j )
  allocate( S_l_J(0:lmax), S_l_K(0:lmax) )
  !
  ! prepare S_l_J(:) (radial double integrals for direct Coulomb interactions)
  !     and S_l_K(:) (radial double integrals for exchange interactions)
  ! NOTE: the same orbital indexes i and j are not excluded
  !       instead, S_l_K(:) = S_l_J(:)
  !
  do l = 0, lmax
    S_l_J(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                 F%u_all(1:N,ii), F%u_all(1:N,jj), &
                 F%u_all(1:N,ii), F%u_all(1:N,jj) )
    if ( ii == jj ) then
      S_l_K(l) = S_l_J(l)
    else
      S_l_K(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                   F%u_all(1:N,ii), F%u_all(1:N,jj), &
                   F%u_all(1:N,jj), F%u_all(1:N,ii) )
    end if
  end do
  !
  ! compute J and K with v_twobody_nl(): v(1) = v_J and v(2) = v_K
  !
  v(:) = v_twobody_nl( F, n_i, l_i, n_j, l_j, S_l_J, S_l_K )
  deallocate( S_l_J, S_l_K )
end function v_twobody_all

function v_twobody_nl( F, n_i, l_i, n_j, l_j, S_l_J, S_l_K ) result (v)
  real(long) :: v(2)    ! v(1) = v_J and v(2) = v_K
  type(WF_data), intent(in) :: F
  integer, intent(in) :: n_i, l_i, n_j, l_j
  real(long), intent(in) :: S_l_J(0:), S_l_K(0:)
  integer :: ml_i, ml_j
  !integer :: ms_i, ms_j
  integer :: p(4), q(4)
  real(long) :: v_J, v_K
  !
  ! summation over all m_l and m_s without considering symmetry...
  ! this is a brute-force way; it should be optimized with an analytic solution
  !
  v_J = ZERO    ! direct Coulomb integral
  v_K = ZERO    ! exchange integral
  do ml_i = -l_i, l_i
  do ml_j = -l_j, l_j
    p(:) = (/ n_i, l_i, ml_i, 0 /)      ! integral doesn't depend on spin
    q(:) = (/ n_j, l_j, ml_j, 0 /)
    !
    ! after taking into account spins (see double do-loop with ms_i and ms_j)
    ! what we need is a factor of 4 for v_J (v_ijij) and 2 for v_K (v_ijji)
    ! NOTE: Here, the same quantum number sets for p and q are not skipped.
    !       Instead, v_ijij and v_ijji are identical in this case, 
    !                so they are cancelled to each other.
    !       In this way, v_J gives the same value of "Hartree E" using density.
    !
    v_J = v_J + v_pqrs( F, p, q, p, q, S_l_J ) * FOUR
    v_K = v_K + v_pqrs( F, p, q, q, p, S_l_K ) * TWO
    !print "(6i3,2es13.5)", n_i, l_i, ml_i, n_j, l_j, ml_j, &
    !  v_pqrs( F, p, q, p, q, S_l_J ), v_pqrs( F, p, q, q, p, S_l_K )
    !
    ! Here is the double do-loop with ms_i and ms_j.
    ! The results are identical to the above with 4 and 2.
    !
    !do ms_i = 0, 1        ! 0: spin down, 1: spin up
    !do ms_j = 0, 1
    !  ! this if-then statement doesn't change the final result,
    !  ! because v_ijij and v_ijji are identical and cancelled to each other.
    !  !if ( n_i == n_j .and. l_i == l_j .and. &
    !  !     ml_i == ml_j .and. ms_i == ms_j ) cycle      ! skip if identical
    !  p(:) = (/ n_i, l_i, ml_i, ms_i /)
    !  q(:) = (/ n_j, l_j, ml_j, ms_j /)
    !  v_J = v_J + v_pqrs( F, p, q, p, q, S_l_J )
    !  v_K = v_K + v_pqrs( F, p, q, q, p, S_l_K )
    !end do
    !end do
  end do
  end do
  !
  ! taking care of averaging over 2*(2l+1)
  !
!TEST (taking into account self-interaction; NEED TO RECHECK)
  if ( n_i == n_j .and. l_i == l_j ) then
    v_J = v_J / dble(4*l_i+2) / dble(4*l_i+1)
    v_K = v_K / dble(4*l_i+2) / dble(4*l_i+1)
  else
    v_J = v_J / dble(4*l_i+2) / dble(4*l_j+2)
    v_K = v_K / dble(4*l_i+2) / dble(4*l_j+2)
  end if
!TEST (in the following, no consideration for self-interaction)
!  v_J = v_J / dble(4*l_i+2) / dble(4*l_j+2)
!  v_K = v_K / dble(4*l_i+2) / dble(4*l_j+2)
!TEST
  !
  ! return ( v_pqpq, v_pqqp )
  !
  v(1) = v_J
  v(2) = v_K
end function v_twobody_nl

!function v_twobody_nl_new( F, n_i, l_i, n_j, l_j, yes_exchange ) result (v)
!  real(long) :: v
!  type(WF_data), intent(in) :: F
!  integer, intent(in) :: n_i, l_i, n_j, l_j
!  logical, intent(in), optional :: yes_exchange
!  integer :: L, K
!  integer :: ii, jj
!  real(long), dimension(0:2*(l_i+l_j)) :: R12, R21
!  real(long) :: T, M0, M1
!  real(long) :: N12, tau
!  real(long), external :: sixj
!  ii = l_i*F%N + (n_i-l_i)
!  jj = l_j*F%N + (n_j-l_j)
!  if ( ii == jj ) then          ! two equivalent electrons
!    !N12 = ONE / dble( (4*l_i+2) * (4*l_i+2-1) )
!    !tau = ONE / sqrt(TWO)
!    N12 = ONE / dble( (4*l_i+2) * (4*l_j+2) )
!    tau = ONE
!  else                          ! two different electrons
!    N12 = ONE / dble( (4*l_i+2) * (4*l_j+2) )
!    tau = ONE
!  end if
!  v = ZERO
!  R12(:) = ZERO
!  R21(:) = ZERO
!  do L = abs(l_i-l_j), l_i+l_j
!    M0 = ZERO
!    M1 = ZERO
!    do K = 0, 2*(l_i+l_j)
!      !
!      ! computing R_K(1,2,3,4)
!      !
!      T = reduced_matrix( K, l_i, l_i ) * reduced_matrix( K, l_j, l_j ) * &
!          sixj( l_i*2, l_j*2, L*2, l_j*2, l_i*2, K*2 )
!      if ( T /= ZERO .and. R12(K) == ZERO ) then
!        R12(K) = double_integral_nonuniform( &
!                 K, F%r(:), F%weight(:), &
!                 F%u_all(:,ii), F%u_all(:,jj), F%u_all(:,ii), F%u_all(:,jj) )
!      end if
!      M0 = M0 + R12(K) * T
!      M1 = M1 + R12(K) * T
!      !
!      ! computing R_K(2,1,3,4)
!      !
!      T = reduced_matrix( K, l_i, l_j ) * reduced_matrix( K, l_j, l_i ) * &
!          sixj( l_i*2, l_j*2, L*2, l_i*2, l_j*2, K*2 )
!      if ( T /= ZERO .and. R21(K) == ZERO ) then
!        if ( ii == jj .and. R12(K) /= ZERO ) then
!          R21(K) = R12(K)
!        else
!          R21(K) = double_integral_nonuniform( &
!                   K, F%r(:), F%weight(:), &
!                   F%u_all(:,jj), F%u_all(:,ii), F%u_all(:,ii), F%u_all(:,jj) )
!        end if
!      end if
!      !
!      ! considering (-1)^(L+S)
!      !
!      if ( mod(L,2) == 0 ) then       ! considering (-1)^(L+S)
!        M0 = M0 + R21(K) * T          ! L+S is even
!        M1 = M1 - R21(K) * T          ! L+S is odd
!      else
!        M0 = M0 - R21(K) * T          ! L+S is odd
!        M1 = M1 + R21(K) * T          ! L+S is even
!      end if
!    end do
!    !
!    ! considering tau and (-1)^( L + l_i + l_j )
!    !
!    M0 = tau * M0;  if ( mod(L+l_i+l_j,2) == 1 ) M0 = -M0
!    M1 = tau * M1;  if ( mod(L+l_i+l_j,2) == 1 ) M1 = -M1
!    !
!    ! summation over L
!    !
!    v = v + dble(2*L+1) * M0 &
!          + dble(3*(2*L+1)) * M1
!  end do
!  v = v * N12 / dble(4*l_i+2)   ! WRONG: averaging over l_i' is missing!!!
!end function v_twobody_nl_new

!----------------------------------------------------------------------------
! SR calculate_E_twobody( P, F, E_J, E_K, config ):
!   calculate E_J and E_K with a given wavefunction F
!   it can be modified with a given config [NOT YET IMPLEMENTED]
!   --> need to store S_l_J(:) and S_l_K(:) to save CPU times
! it originally requires v_twobody() or v_twobody_all(),
! but now it directly calls v_twobody_nl()
!----------------------------------------------------------------------------
subroutine calculate_E_twobody( P, F, E_J, E_K )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  real(long), intent(out) :: E_J, E_K
  !integer, intent(in), optional :: config(:,:)
  real(long) :: S_l_J(0:4*F%lmax), S_l_K(0:4*F%lmax)
  real(long) :: v(2), v_J, v_K
  integer :: i, j
  integer :: n_i, l_i, n_j, l_j
  integer :: N, l, lmax
  N = F%j_s
  E_J = ZERO    ! direct Coulomb interaction (Hartree energy)
  E_K = ZERO    ! exchange interaction (exchange energy)
  if ( P%yes_verbose ) then
    write(P%OUTPUT, "('# ',4a6,2a13)") &
      "i", "j", "occ_i", "occ_j", "v_ijij", "v_ijji"
  end if
  do i = 1, F%N_orb
    do j = i, F%N_orb
      n_i = F%config(i,1); l_i = F%config(i,2)
      n_j = F%config(j,1); l_j = F%config(j,2)
      lmax = 2 * ( l_i + l_j )
      do l = 0, lmax    ! calculate radial double integrals for v_ijij
        S_l_J(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                     F%u(1:N,i), F%u(1:N,j), F%u(1:N,i), F%u(1:N,j) )
      end do
      if ( i == j ) then
        S_l_K(0:lmax) = S_l_J(0:lmax)
      else
        do l = 0, lmax  ! calculate radial double integrals for v_ijji
          S_l_K(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                       F%u(1:N,i), F%u(1:N,j), F%u(1:N,j), F%u(1:N,i) )
        end do
      end if
      v(:) = v_twobody_nl( F, n_i, l_i, n_j, l_j, S_l_J(0:lmax), S_l_K(0:lmax) )
      v_J = v(1); v_K = v(2)
      if ( P%yes_verbose ) then
        write(P%OUTPUT, "('# ',2a6,2i6,2es13.5)") &
          trim(WF_name_nonrel( F, i )), trim(WF_name_nonrel( F, j )), &
          F%config(i,3), F%config(j,3), v_J, v_K
      end if
      if ( i /= j ) then        ! taking unto account off-diagonals
        v_J = v_J * TWO
        v_K = v_K * TWO
      end if
      !
      ! accumulate E_J & E_K with occupation numbers of i-th and j-th orbitals
      !
      E_J = E_J + v_J * F%config(i,3) * F%config(j,3)
      E_K = E_K + v_K * F%config(i,3) * F%config(j,3)
    end do
  end do
  !
  ! direct Coulomb interaction :=  1/2 SUM_ij v_ijij occ_n_i occ_n_j
  ! exchange interaction       := -1/2 SUM_ij v_ijji occ_n_i occ_n_j
  !
  E_J =  HALF * E_J
  E_K = -HALF * E_K
end subroutine calculate_E_twobody

!----------------------------------------------------------------------------
! F exchange_E_twobody( P, F, config ):
!   Since E_J is identical to the Hartree E calculated with density
!   [ see scf.f90; E_J = functional_J( G, rho, v_J ) ],
!   we don't need to repeat E_J calculation.
!   Thus, this function returns E_K only.
! NOTE: I need to think about how to store S_l_K(:) per each (i,j)-pair
!       or v_K per each (i,j)-pair.
!----------------------------------------------------------------------------
function exchange_E_twobody( P, F ) result (E_K)
  real(long) :: E_K
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  !integer, intent(in), optional :: config(:,:)
  real(long) :: S_l_K(0:4*F%lmax)
  real(long) :: v_K
  integer :: i, j
  integer :: n_i, l_i, n_j, l_j
  integer :: N, l, lmax
  N = F%j_s
  E_K = ZERO            ! exchange interaction (exchange energy)
  do i = 1, F%N_orb
    do j = i, F%N_orb
      n_i = F%config(i,1); l_i = F%config(i,2)
      n_j = F%config(j,1); l_j = F%config(j,2)
      lmax = 2 * ( l_i + l_j )
      do l = 0, lmax    ! calculate radial double integrals for v_ijji
        S_l_K(l) = double_integral_nonuniform( l, F%r(1:N), F%weight(1:N), &
                     F%u(1:N,i), F%u(1:N,j), F%u(1:N,j), F%u(1:N,i) )
      end do
      v_K = v_twobody_nl_K( F, n_i, l_i, n_j, l_j, S_l_K(0:lmax) )
      if ( i /= j ) v_K = v_K * TWO     ! taking unto account off-diagonals
      E_K = E_K + v_K * F%config(i,3) * F%config(j,3)   ! accumulate with occ_n
      if ( P%yes_verbose ) then
        write(P%OUTPUT, "('# ',2a6,2i6,es13.5)") &
          trim(WF_name_nonrel( F, i )), trim(WF_name_nonrel( F, j )), &
          F%config(i,3), F%config(j,3), v_K
      end if
    end do
  end do
  E_K = -HALF * E_K  ! exchange E := -1/2 SUM_ij v_ijji occ_n_i occ_n_j
contains
  !
  ! slim version of v_twobody_nl(): see comments in the original function
  !
  function v_twobody_nl_K( F, n_i, l_i, n_j, l_j, S_l_K ) result (v_K)
    real(long) :: v_K
    type(WF_data), intent(in) :: F
    integer, intent(in) :: n_i, l_i, n_j, l_j
    real(long), intent(in) :: S_l_K(0:)
    integer :: ml_i, ml_j
    integer :: p(4), q(4)
    v_K = ZERO    ! exchange integral
    do ml_i = -l_i, l_i
    do ml_j = -l_j, l_j
      p(:) = (/ n_i, l_i, ml_i, 0 /)      ! integral doesn't depend on spin
      q(:) = (/ n_j, l_j, ml_j, 0 /)
      v_K = v_K + v_pqrs( F, p, q, q, p, S_l_K ) * TWO
    end do
    end do
    v_K = v_K / dble(4*l_i+2) / dble(4*l_j+2)     ! averaging over 2*(2l+1)
  end function v_twobody_nl_K
end function exchange_E_twobody

end module coulomb
