!----------------------------------------------------------------------------
! decay.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2021 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2021 Sang-Kil Son <sangkil.son@cfel.de>
!          2015-2016 Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! HFS calculation for the atomic system using the GPS discretization
!   written by Son, Sang-Kil in Jul. 2010 at DESY
!
! This module computes fluorescence rates and Auger rates.
! Don't be screwed(!) by flourescence or fluorescence.  The latter is correct.
! For complete expressions for fluo. and Auger rates, see
!   Bhalla, Folland, and Hein, PRA 8, 649 (1973).
!----------------------------------------------------------------------------
module decay
  use constants
  use coord_system
  use gps
  use read_param
  use wavefunction
  use process
  use coulomb
  implicit none
contains

!----------------------------------------------------------------------------
! SR calculate_fluorescence_rates( P, G, F ):
!   fluorescence rates of the core hole decaying into all possible config.
!   symmetry consideration: Delta l = +/-1
!----------------------------------------------------------------------------
subroutine calculate_fluorescence_rates( P, G, F, S, i_hole )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  type(Process_data), intent(inout) :: S
  integer, intent(in), optional :: i_hole  ! only calc. this subshell if given
  integer :: occ_n, occ_max, n_hole, ind_q, ind_qp
  integer :: q, qp, l_q, j_q, l_qp, j_qp, N_orb
  real(long) :: alpha, IE_qp, IE_q, e_xray
  real(long) :: rate, sum_rates, T, prefactor
  integer :: OUTPUT
  character(len=4) :: str_hole
  character(len=15) :: string
  real(long), external :: sixj
  OUTPUT = P%OUTPUT
  !
  ! print out possible fluorescence rates
  !   rate_Ka = 4 / 3 * alpha^3 * ( I_1s - I_2p )^3 *
  !             | int_0^inf u_1s(r) * r * u_2p(r) dr |^2
  ! for the general case
  ! SEE: Eq. (13) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  ! rate : fluorescence rate per one initial hole
  ! sum_rates : sum of ( rates per one hole times # of initial holes )
  !
  alpha = ONE / SPEED_C_AU      ! fine-structure constant = 1 / 137
  sum_rates = ZERO
  !write(OUTPUT, "('#'/'# Fluorescence (based on orbital energies except []):')")
  write(OUTPUT, "('#'/'# Fluorescence (based on orbital energies):')")
  write(OUTPUT, "('#   I     F     ',a14,2x,a13,6x,a13)") &
    "x-ray E(eV)", "rate(a.u.)", "R_int"
  N_orb = F%N_subshell
  do q = 1, N_orb               ! searching holes
    if ( present(i_hole) .and. q /= i_hole ) cycle  ! skip for q /= i_hole
    occ_max = get_occ_num_max( F, q )
    occ_n = get_occ_num( F, q )
    n_hole = occ_max - occ_n
    if ( n_hole == 0 ) cycle    ! skip if this subshell is full
    str_hole = ""
    if ( n_hole > 1 ) write( str_hole, "(' *',i2)" ) n_hole
    !
    ! if there is a hole, memorize the angular momenta
    !
    if ( P%yes_relativity ) then
      l_q = F%config_rel(q,2)
      j_q = F%config_rel(q,3) ! 2 * j_i
    else
      l_q = F%config(q,2)    
    endif
    IE_q = ionization_E( F, q )
    !
    ! searching electrons which can make a transition to the hole q
    !
    do qp = 1, N_orb
      if ( q == qp ) cycle       ! skip if they are the same subshell
      occ_n = get_occ_num( F, qp )
      if ( occ_n == 0 ) cycle   ! no electron here
      IE_qp = ionization_E( F, qp )
      if ( IE_q <= IE_qp ) cycle     ! skip if E_i is higher than E_j
      !
      ! checking the selectron rules
      !
      if ( P%yes_relativity ) then
        l_qp = F%config_rel(qp,2)
        j_qp = F%config_rel(qp,3) ! 2* j_j
        if ( abs(l_q-l_qp) /= 1 .or.  abs(j_q - j_qp) > 2 ) cycle
      else
        l_qp = F%config(qp,2)       
        if ( abs(l_q-l_qp) /= 1 ) cycle    
      endif
      e_xray = IE_q - IE_qp
      write( string, "(f13.2)" ) e_xray * AU2EV
!      call check_experimental_energy( P, e_xray, string, i, j )
      if ( P%yes_relativity ) then
        ind_q = F%config_rel(q, 5)
        ind_qp = F%config_rel(qp, 5)
      else
        ind_q = q
        ind_qp = qp
      endif
      !
      ! trans. dipole moment for the electron in the subshell qp to the hole q
      !
      T = radial_integral( G, F%u(:,ind_q) * G%r(:) * F%u(:,ind_qp) )
      rate = FOUR / THREE * alpha**3 * e_xray**3 * T**2 
      prefactor = dble( max( l_q, l_qp ) ) * dble(occ_n)
      if ( p%yes_relativity ) then
        prefactor = prefactor * (sixj( 2*l_qp, 1, j_qp, j_q, 2, 2*l_q ))**2
      else
        prefactor = prefactor / dble( 4*l_qp + 2 ) / (2*l_q + 1 )
      endif
      rate = prefactor * rate 
      write(OUTPUT, "('#   ',a3,' - ',a3,' : ',a14,2x,es13.5,a4,2x,es13.5)") &
        WF_name( F, q ), WF_name( F, qp ), trim(string), rate, str_hole, T
      call add_Process( S, "fluorescence", rate * n_hole, e_xray, q, qp )
      sum_rates = sum_rates + rate * n_hole
    end do
  end do
  !
  ! print out the total rate
  ! if i_hole is given, sum_rates becomes the sum of processes with i_hole only
  !
  if ( sum_rates > ZERO ) then
    if ( present(i_hole) ) then
      write(OUTPUT, "('# ',12x,a)") repeat( "-", 31 )
      write(OUTPUT, "('# ',12x,a3,' hole F rate = ',es13.5,a4)") &
        WF_name( F, i_hole ), sum_rates / n_hole, str_hole
    else
      string = lifetime( sum_rates )
      write(OUTPUT, "('# ',15x,a)") repeat( "-", 28 )
      write(OUTPUT, "('# ',15x,'Total F rate = ',es13.5,' a.u.  '," // &
        "'( lifetime = ',a,' )')") sum_rates, trim(string)
    end if
  else
    write(OUTPUT, "('#   NONE')")
  end if
contains
  subroutine check_experimental_energy( P, e_xray, str_energy, i, j )
    type(Param), intent(in) :: P
    real(long), intent(inout) :: e_xray
    character(len=*), intent(inout) :: str_energy
    integer, intent(in) :: i, j
    real(long) :: E
    E = e_xray
    if ( P%EX1 /= ZERO .and. i == 1 .and. j == 3 ) E = P%EX1    ! 1s - 2p
    if ( P%EX2 /= ZERO .and. i == 2 .and. j == 3 ) E = P%EX2    ! 2s - 2p
!    if ( P%EX2 /= ZERO .and. i == 1 .and. j == 5 ) E = P%EX2    ! 1s - 3p
    if ( P%EX3 /= ZERO .and. i == 1 .and. j == 8 ) E = P%EX3    ! 1s - 4p
    if ( E /= e_xray ) then
      e_xray = E
      write( str_energy, "(f11.2)" ) e_xray * AU2EV
      str_energy = '[' // trim(adjustl(str_energy)) // ']'
    end if
  end subroutine check_experimental_energy
end subroutine calculate_fluorescence_rates

subroutine calculate_fluorescence_rates_extended( P, G, F ) !, S )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  !type(Process_data), intent(inout) :: S
  integer :: elec(P%N_elec)
  integer :: i, n_i, l_i, occ_i, hole_i, ind
  integer :: j, n_j, l_j, occ_j
  integer :: n_max
  real(long) :: alpha, e_xray
  real(long) :: T, rate, sum_rates
  character(len=15) :: string
  integer :: OUTPUT
  OUTPUT = P%OUTPUT
  call config_to_elec( F%config, elec, P%N )
  !
  ! print out possible fluorescence rates
  !   rate_Ka = 4 / 3 * alpha^3 * ( I_1s - I_2p )^3 *
  !             | int_0^inf u_1s(r) * r * u_2p(r) dr |^2
  ! for the general case
  ! SEE: Eq. (13) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  ! rate : fluorescence rate per one initial hole
  ! sum_rates : sum of ( rates per one hole times # of initial holes )
  !
  alpha = ONE / SPEED_C_AU      ! fine-structure constant = 1 / 137
  sum_rates = ZERO
  write(OUTPUT, "('#'/'# Fluorescence (based on orbital energies):')")
  write(OUTPUT, "('#   I     F     ',a13,2x,a13,6x,a13)") &
    "x-ray E(eV)", "rate(a.u.)", "R_int"
  !
  ! find out the maximum n of occupied orbitals
  !
  n_max = 0
  do i = 1, F%N_orb
    if ( F%config(i,3) > 0 .and. F%config(i,1) > n_max ) n_max = F%config(i,1)
  end do
  !
  ! scan over initial hole i: initially there must be a hole
  !
  do n_i = 1, n_max
    do l_i = 0, n_i - 1
      ind = F%N * l_i + n_i - l_i
      occ_i = count( elec(:) == ind )   ! occupation number in (n_i,l_i)
      hole_i = 4*l_i + 2 - occ_i        ! # of holes in (n_i,l_i)
      if ( hole_i == 0 ) cycle          ! no hole here
      !
      ! scan over final hole j: initially there must be an electron
      !
      do j = 1, F%N_orb
        n_j = F%config(j,1)         ! quantum number n
        l_j = F%config(j,2)         ! quantum number l
        occ_j = F%config(j,3)       ! occupation number
        if ( occ_j == 0 ) cycle     ! no electron here
        if ( abs(l_i-l_j) /= 1 ) cycle  ! skip unless l_j = l_i +/- 1
        e_xray = F%e_orb(j) - F%e_all(ind)      ! no need about F%e_s
        if ( e_xray <= ZERO ) cycle     ! skip if E_i is higher than E_j
        !
        ! calculate the fluorescence rate
        !
        T = radial_integral( G, F%u_all(:,ind) * G%r(:) * F%u(:,j) )
        rate = FOUR / THREE * alpha**3 * e_xray**3 * T**2 * &
               dble(occ_j) / dble(4*l_j+2) * dble(max(l_i,l_j)) / dble(2*l_i+1) 
        !
        ! print out the rate
        !
        string = ""
        if ( hole_i > 1 ) write( string, "(' *',i2)" ) hole_i
        write(OUTPUT, "('#   ',a3,' - ',a3,' : ',f13.2)", advance="no") &
          WF_name_extended( F, ind ), WF_name( F, j ), e_xray * AU2EV
        write(OUTPUT, "(2x,es13.5,a4,2x,es13.5)") rate, string, T
        !call add_Process
        sum_rates = sum_rates + rate * hole_i
      end do
    end do
  end do
  !
  ! print out the total rate
  !
  if ( sum_rates > ZERO ) then
    string = lifetime( sum_rates )
    write(OUTPUT, "('# ',14x,a)") repeat( "-", 28 )
    write(OUTPUT, "('# ',14x,'Total F rate = ',es13.5,' a.u.  '," // &
      "'( lifetime = ',a,' )')") sum_rates, trim(string)
  else
    write(OUTPUT, "('#   NONE')")
  end if
end subroutine calculate_fluorescence_rates_extended

!----------------------------------------------------------------------------
! SR calculate_Auger_rates( P, F ):
!   compute the Auger rates for all possiblities
!   symmetry consideration within 3-j and 6-j symbols in Auger_group_rate()
!----------------------------------------------------------------------------
subroutine calculate_Auger_rates( P, F, S, i_hole )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  type(Process_data), intent(inout) :: S
  integer, intent(in), optional :: i_hole  ! only calc. this subshell if given
  integer :: i, j, jp, occ_n, occ_max, occ_n_j, occ_n_jp, n_hole
  real(long) :: e_Auger, rate, sum_all, sum_subshell
  real(long) :: e_abs
  character(len=15) :: string
  character(len=4) :: str_hole
  integer :: OUTPUT, N_orb
  N_orb = F%N_subshell
  OUTPUT = P%OUTPUT
  !
  ! print out possible Auger rates within given config(:,:)
  !
  ! rate : Auger rate per each channel
  ! sum_subshell : sum of individual channels per one hole in the subshell
  ! sum_all : sum of (sum_subshell per one hole times # of initial holes)
  !
  !write(OUTPUT, "('#'/'# Auger decay (based on orbital energies except []):')")
  write(OUTPUT, "('#'/'# Auger decay (based on orbital energies):')")
  write(OUTPUT, "('#                   ',a10,2x,a13,9x,a)") &
    "K.E.(eV)", "rate(a.u.)", "K i1 i2 i3 i4          R_K"
  sum_all = ZERO
  do i = 1, N_orb               ! --------- initial hole: i
    if ( present(i_hole) .and. i /= i_hole ) cycle  ! skip i /= i_hole
    occ_max = get_occ_num_max( F, i )
    occ_n = get_occ_num( F, i )
    n_hole = occ_max - occ_n    ! # of initial holes
    if ( n_hole == 0 ) cycle    ! this orbital is full
    str_hole = ""
    if ( n_hole > 1 ) write( str_hole, "(' *',i2)" ) n_hole
    sum_subshell = ZERO
    do j = 1, N_orb             ! ---------- final hole: j
      if ( i == j ) cycle       ! skip if they are the same subshell
      if ( ionization_E( F, i ) <= ionization_E( F, j ) ) &
        cycle                   ! skip if E_i is higher than E_j
      occ_n_j = get_occ_num( F, j )
      do jp = j, N_orb          ! ---------- final hole: j'
        if ( i == jp ) cycle    ! skip if they are the same subshell
        if ( ionization_E( F, i ) <= ionization_E( F, jp ) ) &
          cycle                 ! skip if E_i higher than E_jp
        occ_n_jp = get_occ_num( F, jp )
        if ( ( jp == j .and. occ_n_j >= 2 ) .or. &      ! equivalent elec.
             ( jp /= j .and. occ_n_j >= 1 &
                       .and. occ_n_jp >= 1 ) ) then     ! different elec.
          !
          ! e_Auger: K.E. of Auger electron based on ionization potentials
          !          to be used to check whether it's ionized or not
          ! e_abs: K.E. of Auger electron w.r.t. zero (absolte scale)
          !        to be used for an argument of unbound()
          ! e_Auger and e_abs are the same if the potential tail is zero.
          !
          e_Auger = ionization_E( F, i ) &
                  - ionization_E( F, j ) - ionization_E( F, jp ) 
          if ( F%yes_relativity ) then  ! too much checking yes_rel
            e_abs = F%e_orb_rel(j) + F%e_orb_rel(jp) - F%e_orb_rel(i)
          else
            e_abs = F%e_orb(j) + F%e_orb(jp) - F%e_orb(i)
          end if
          write( string, "(f10.2)" ) e_Auger * AU2EV
!          call check_experimental_energy( P, e_Auger, string, i, j, jp )
          if ( e_Auger > ZERO ) then             
            rate = Auger_group_rate( P, F, j, jp, i, e_abs, OUTPUT )
            sum_subshell = sum_subshell + rate
            write(OUTPUT,"('#   ',a3,' - ',a3,x,a3,' : ',a10,2x,es13.5,a4)") &
              WF_name( F, i ), WF_name( F, j ), WF_name( F, jp ), &
              trim(string), rate, str_hole
            call add_Process( S, "auger", rate * n_hole, e_Auger, i, j, jp )
          end if
        end if
      end do
    end do
    !
    ! print out subshell total
    !
    if ( sum_subshell > ZERO ) then
      write(OUTPUT, "('# ',12x,a)") repeat( "-", 31 )
      write(OUTPUT, "('# ',12x,a3,' hole A rate = ',es13.5,a4)") &
        WF_name( F, i ), sum_subshell, str_hole
      sum_all = sum_all + sum_subshell * n_hole
    end if
  end do
  !
  ! print out the total rate
  !
  if ( sum_all > ZERO ) then
    if ( .not. present(i_hole) ) then
      string = lifetime( sum_all )
      write(OUTPUT, "('# ',15x,'Total A rate = ',es13.5,' a.u.  '," // &
        "'( lifetime = ',a,' )')") sum_all, trim(string)
    end if
  else
    write(OUTPUT, "('#   NONE')")
  end if
contains
  subroutine check_experimental_energy( P, e_Auger, str_energy, i, j, jp )
    type(Param), intent(in) :: P
    real(long), intent(inout) :: e_Auger
    character(len=*), intent(inout) :: str_energy
    integer, intent(in) :: i, j, jp
    real(long) :: E
    E = e_Auger
    !
    ! EA1 for 1s - 2s 2s
    ! EA2 for 1s - 2s 2p
    ! EA3 for 1s - 2p 2p
    !
    if ( P%EA1 /= ZERO .and. i == 1 .and. j == 2 .and. jp == 2 ) E = P%EA1
    if ( P%EA2 /= ZERO .and. i == 1 .and. j == 2 .and. jp == 3 ) E = P%EA2
    if ( P%EA3 /= ZERO .and. i == 1 .and. j == 3 .and. jp == 3 ) E = P%EA3
    if ( E /= e_Auger ) then
      e_Auger = E
      write( str_energy, "(f8.2)" ) e_Auger * AU2EV
      str_energy = '[' // trim(adjustl(str_energy)) // ']'
    end if
  end subroutine check_experimental_energy
end subroutine calculate_Auger_rates

subroutine calculate_Auger_rates_extended( P, F ) !, S )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  !type(Process_data), intent(inout) :: S
  integer :: elec(P%N_elec)
  integer :: i,  n_i,  l_i,  occ_i, hole_i, ind
  integer :: j,  n_j,  l_j,  occ_j
  integer :: jp, n_jp, l_jp, occ_jp
  integer :: n_max
  real(long) :: e_Auger, rate, sum_all, sum_subshell
  real(long) :: e_abs
  character(len=15) :: string
  integer :: OUTPUT
  OUTPUT = P%OUTPUT
  call config_to_elec( F%config, elec, P%N )
  !
  ! print out possible Auger rates within given config(:,:)
  !
  ! rate : Auger rate per each channel
  ! sum_subshell : sum of individual channels per one hole in the subshell
  ! sum_all : sum of (sum_subshell per one hole times # of initial holes)
  !
  write(OUTPUT, "('#'/'# Auger decay (based on orbital energies):')")
  write(OUTPUT, "('#                   ',a10,2x,a13)") "K.E.(eV)", "rate(a.u.)"
!  write(OUTPUT, "('#                   ',a10,2x,a13,7x,a)") &
!    "K.E.(eV)", "rate(a.u.)", "K n1 l1 n2 l2 n3 l3 l4          R_K"
  sum_all = ZERO
  !
  ! find out the maximum n of occupied orbitals
  !
  n_max = 0
  do i = 1, F%N_orb
    if ( F%config(i,3) > 0 .and. F%config(i,1) > n_max ) n_max = F%config(i,1)
  end do
  !
  ! scan over initial hole i: initially there must be a hole
  !
  do n_i = 1, n_max
    do l_i = 0, n_i - 1
      ind = F%N * l_i + n_i - l_i
      occ_i = count( elec(:) == ind )   ! occupation number in (n_i,l_i)
      hole_i = 4*l_i + 2 - occ_i        ! # of holes in (n_i,l_i)
      if ( hole_i == 0 ) cycle          ! no hole here
      sum_subshell = ZERO
      !
      ! scan over final hole j: initially there must be an electron
      !
      do j = 1, F%N_orb
        n_j = F%config(j,1)             ! quantum number n
        l_j = F%config(j,2)             ! quantum number l
        occ_j = F%config(j,3)           ! occupation number
        if ( occ_j == 0 ) cycle         ! no electron here
        if ( n_i == n_j .and. l_i == l_j ) cycle  ! skip if the same subshells
        if ( F%e_all(ind) >= F%e_orb(j) ) cycle   ! skip if E_i >= E_j
        !
        ! scan over final hole jp: initially there must be an electron
        !
        do jp = j, F%N_orb
          n_jp = F%config(jp,1)         ! quantum number n
          l_jp = F%config(jp,2)         ! quantum number l
          occ_jp = F%config(jp,3)       ! occupation number
          if ( occ_jp == 0 ) cycle      ! no electron here
          if ( n_i == n_jp .and. l_i == l_jp ) cycle  ! skip if the same subsh.
          if ( F%e_all(ind) >= F%e_orb(jp) ) cycle    ! skip if E_i >= E_j
          if ( jp == j .and. occ_j < 2 ) cycle        ! equivalent elec.
          !
          ! e_Auger: K.E. of Auger electron based on ionization potentials
          !          to be used to check whether it's ionized or not
          ! e_abs: K.E. of Auger electron w.r.t. zero (absolte scale)
          !        to be used for an argument of unbound()
          ! e_Auger and e_abs are the same if the potential tail is zero.
          !
          e_Auger = ionization_E_extended( F, ind ) &
                  - ionization_E( F, j ) - ionization_E( F, jp )
          if ( F%yes_relativity ) then  ! too much checking yes_rel
            e_abs = F%e_orb_rel(j) + F%e_orb_rel(jp) - F%e_all(ind)
          else
            e_abs = F%e_orb(j) + F%e_orb(jp) - F%e_all(ind)
          end if
          if ( e_Auger < ZERO ) cycle   ! skip with energy criterion
          !
          ! calculate Auger rates
          !
          rate = Auger_group_rate_extended( F, j, jp, ind, e_abs, OUTPUT )
          sum_subshell = sum_subshell + rate
          !
          ! print out the rate
          !
          string = ""
          if ( hole_i > 1 ) write( string, "(' *',i2)" ) hole_i
          write(OUTPUT,"('#   ',a3,' - ',a3,x,a3,' : ',f10.2,2x,es13.5,a4)") &
            WF_name_extended( F, ind ), WF_name( F, j ), WF_name( F, jp ), &
            e_Auger * AU2EV, rate, string
          !call add_Process( S, "auger", rate * n_hole, e_Auger, i, j, jp )
        end do
      end do
      !
      ! print out subshell total
      !
      if ( sum_subshell > ZERO ) then
        write(OUTPUT, "('# ',12x,a)") repeat( "-", 31 )
        write(OUTPUT, "('# ',12x,a3,' hole A rate = ',es13.5,a4)") &
          WF_name_extended( F, ind ), sum_subshell, string
        sum_all = sum_all + sum_subshell * hole_i
      end if
    end do
  end do
  !
  ! print out the total rate
  !
  if ( sum_all > ZERO ) then
    string = lifetime( sum_all )
    write(OUTPUT, "('# ',15x,'Total A rate = ',es13.5,' a.u.  '," // &
      "'( lifetime = ',a,' )')") sum_all, trim(string)
  else
    write(OUTPUT, "('#   NONE')")
  end if
end subroutine calculate_Auger_rates_extended

!----------------------------------------------------------------------------
! compute the Auger rates of the i3-th hole decaying from (i1,i2) to (i3,E)
!   Auger( n3 l3 - n1 l1, n2 l2 )
!     =  pi N_12 SUM_J (2J+1)/(2l3+1) SUM_l4 |M|^2
!     => pi N_12 / (2l3+1) * SUM_l4 SUM_J (2J+1) |M|^2
!     where M is given as Bhalla's paper: PRA 8, 649 (1973).
!
! SEE: Eqs. (7)-(12) in Son, Young & Santra, PRA 83, 033402 (2011).
!      n3 l3 --> i, l_i
!         l4 -->    l_i'
!      n1 l1 --> j, l_j
!      n2 l2 --> j', l_j'
! There are only two cases for S, S=0 or S=1 because of two electrons
!----------------------------------------------------------------------------
function Auger_group_rate( P, F, i1, i2, i3, E, OUTPUT ) result (rate)
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, i3
  real(long), intent(in) :: E
  integer, intent(in) :: OUTPUT
  real(long) :: rate
  integer :: n3, l3, j3, ind3
  !
  ! i1, i2, i3: the index for different subshells
  !   which can refer to the first index of config(:,1:3) or config_rel(:,1:5)
  ! These indexes are the index for u in the non-relativistic case,
  ! but they are NOT the index for u in the relativistic case!
  !
  if ( P%yes_relativity ) then
     !
     ! Packing n, l, and  j into ind3
     ! negative (positive) sign of ind3 is utilized to indicate j=l-s (l+s)
     ! ind3 is the index for u_all:
     !   u_all(:,abs(ind)) := non-relativistic radial function of (n,l)
     !                        where ind = F%N * l + n - l
     !
     n3 = F%config_rel(i3,1)
     l3 = F%config_rel(i3,2)
     j3 = F%config_rel(i3,3)
     ind3 = F%N * l3 + n3 - l3
     if ( j3 == abs(2*l3-1) ) ind3 = -ind3
     rate = Auger_group_rate_extended_jj( F, i1, i2, ind3, E ) !, OUTPUT )
  else
     n3 = F%config(i3,1)
     l3 = F%config(i3,2)
     ind3 = F%N * l3 + n3 - l3
     rate = Auger_group_rate_extended( F, i1, i2, ind3, E, OUTPUT )
  endif
end function Auger_group_rate

function Auger_group_rate_extended( F, i1, i2, ind3, E, OUTPUT ) result (rate)
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, ind3
  real(long), intent(in) :: E
  integer, intent(in) :: OUTPUT
  real(long) :: rate
  integer, parameter :: L_max = 100              ! maximum L
  real(long) :: u(F%N_uni), up(F%N_uni)
  real(long) :: R12(0:L_max), R21(0:L_max)      ! R_K(1,2,3,4) and R_K(2,1,3,4)
  integer :: i, i3
  integer :: n1, n2, n3
  integer :: l1, l2, l3, l4, k, L
  integer :: occ1, occ2
  real(long) :: N12, tau, M0, M1, T, ll
  real(long) :: u3(F%N)
!  real(long) :: u3_uni(F%N_uni)        ! we don't need it anymore
  real(long), external :: sixj
  character(len=80) :: format_str

  n1 = F%config(i1,1);  l1 = F%config(i1,2);  occ1 = F%config(i1,3)
  n2 = F%config(i2,1);  l2 = F%config(i2,2);  occ2 = F%config(i2,3)

  l3 = int( ind3 / F%N )
  n3 = mod( ind3, F%N ) + l3
  !
  ! search (n3,l3) in config(:,:)
  ! if the hole is not in config(:,:), then use F%u_all
  ! when F%u_all is used, solve_SE_all() should be called beforehand
  !
  i3 = 0
  do i = 1, F%N_orb
    if ( n3 == F%config(i,1) .and. l3 == F%config(i,2) ) i3 = i
  end do
  if ( i3 /= 0 ) then
    u3(:) = F%u(:,i3)           ! pick up from occupied bound states
  else
    u3(:) = F%u_all(:,ind3)     ! pick up from F%u_all
  end if
  !
  ! N12 is given by Bhalla's paper
  ! SEE: Eq. (3) in Bhalla, Folland & Hein, PRA 8, 649 (1973).
  ! Equivalently, N_jj' is given by Son's paper
  ! SEE: Eq. (8) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  if ( i1 == i2 ) then          ! two equivalent electrons
    N12 = dble( occ1 * (occ1-1) ) / dble( (4*l1+2) * (4*l1+2-1) )
    tau = ONE / sqrt(TWO)
  else                          ! two different electrons
    N12 = dble( occ1 * occ2 ) / dble( (4*l1+2) * (4*l2+2) )
    tau = ONE
  end if
  if ( l1+l2+l3 > L_max ) then
    write(STDERR,*) "l1 = ", l1
    write(STDERR,*) "l2 = ", l2
    write(STDERR,*) "l3 = ", l3
    write(STDERR,*) "Too large l's: l1+l2+l3 should be <= ", L_max
    stop
  end if
  !
  ! summation over l4
  !   range of l4: |l3-l4| <= L <= l3+l4 --> |L-l3| <= l4 <= L+l3
  !            and |l1-l2| <= L <= l1+l2
  !   so, just take upper bound of l4: l4 <= l1+l2+l3
  !
  rate = ZERO
  loop_l4: do l4 = 0, l1+l2+l3
    ll = dble(l4)               ! unbound.f accepts real*8 rather than integer
    call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, E, F, 'A' )
    R12(:) = ZERO
    R21(:) = ZERO
    do L = abs(l1-l2), l1+l2    ! |l1-l2| <= L <= l1+l2
      M0 = ZERO                 ! M_LS for S=0, SEE: Eq. (9) in PRA 83, 033402
      M1 = ZERO                 ! M_LS for S=1, SEE: Eq. (9) in PRA 83, 033402
      do K = 0, l1+l2+l3
        !
        ! computing R_K(1,2,3,4)
        ! T      -> A_K(j,j',i,i'), SEE: Eq. (10) in PRA 83, 033402
        ! R12(K) -> R_K(j,j',i,i'), SEE: Eq. (12) in PRA 83, 033402
        !
        T = reduced_matrix( K, l3, l1 ) * reduced_matrix( K, l4, l2 ) * &
            sixj( l3*2, l4*2, L*2, l2*2, l1*2, K*2 )
        if ( T /= ZERO .and. R12(K) == ZERO ) then
!          R12(K) = double_integral( F%N_uni, F%r_uni(:), &
!               F%u_uni(:,i1), F%u_uni(:,i2), u3_uni(:), u(:), K )
          R12(K) = double_integral_mod( K, &
                   F%r(:), F%weight(:), F%u(:,i1), u3(:), &
                   F%r_uni(:), F%u_uni(1:F%N_uni,i2), u(1:F%N_uni), F%rm_thres )
          if ( n1 < 10 .and. n2 < 10 .and. n3 < 10 .and. &
               l1 < 10 .and. l2 < 10 .and. l3 < 10 .and. l4 < 10 ) then
            format_str = "('#',51x,i3,3(x,2i1),x,i1,x,es13.5)"
          else
            format_str = "('#',34x,i3,3(' [',i2,',',i2,']'),' [',i2,']',es13.5)"
          end if
          write(OUTPUT,format_str) K, n1, l1, n2, l2, n3, l3, l4, R12(K)
        end if
        M0 = M0 + R12(K) * T
        M1 = M1 + R12(K) * T
        !
        ! computing R_K(2,1,3,4)
        ! T      -> A_K(j',j,i,i'), SEE: Eq. (10) in PRA 83, 033402
        ! R21(K) -> R_K(j',j,i,i'), SEE: Eq. (12) in PRA 83, 033402
        !
        T = reduced_matrix( K, l3, l2 ) * reduced_matrix( K, l4, l1 ) * &
            sixj( l3*2, l4*2, L*2, l1*2, l2*2, K*2 )
        if ( T /= ZERO .and. R21(K) == ZERO ) then
          if ( i1 == i2 .and. R12(K) /= ZERO ) then
            R21(K) = R12(K)
          else
!             R21(K) = double_integral( F%N_uni, F%r_uni(:), &
!                 F%u_uni(:,i2), F%u_uni(:,i1), u3_uni(:), u(:), K )
            R21(K) = double_integral_mod( K, &
                     F%r(:), F%weight(:), F%u(:,i2), u3(:), &
                     F%r_uni(:), F%u_uni(1:F%N_uni,i1), u(1:F%N_uni), &
                     F%rm_thres )
            if ( n1 < 10 .and. n2 < 10 .and. n3 < 10 .and. &
                 l1 < 10 .and. l2 < 10 .and. l3 < 10 .and. l4 < 10 ) then
              format_str = "('#',51x,i3,3(x,2i1),x,i1,x,es13.5)"
            else
              format_str = "('#',34x,i3,3(' [',i2,',',i2,']')," // &
                           "' [',i2,']',es13.5)"
            end if
            write(OUTPUT,format_str) K, n2, l2, n1, l1, n3, l3, l4, R21(K)
          end if
        end if
        !
        ! considering (-1)^(L+S)
        !   for two equivalent electrons and L=0,
        !   in fact S=1 is forbidden by the parity consideration,
        !   but here it can be excluded by a (-1)^(0+1) factor.
        !   i.e. M1 becomes 0 since R_K(1,2,3,4)=R_K(2,1,3,4) and other terms.
        !
        ! M_LS is properly summed over K up according to S=0 (M0) and S=1 (M1)
        ! SEE: Eq. (9) in Son, Young & Santra, PRA 83, 033402 (2011).
        !
        if ( mod(L,2) == 0 ) then       ! considering (-1)^(L+S)
          M0 = M0 + R21(K) * T          ! L+S is even
          M1 = M1 - R21(K) * T          ! L+S is odd
        else
          M0 = M0 - R21(K) * T          ! L+S is odd
          M1 = M1 + R21(K) * T          ! L+S is even
        end if
      end do
      !
      ! considering tau and (-1)^( L + l_j + l_i' )
      ! SEE: Eq. (9) in Son, Young & Santra, PRA 83, 033402 (2011).
      !
      M0 = tau * M0;  if ( mod(L+l1+l4,2) == 1 ) M0 = -M0
      M1 = tau * M1;  if ( mod(L+l1+l4,2) == 1 ) M1 = -M1
      !
      ! summation over L and l4
      ! SEE: Eq. (7) in Son, Young & Santra, PRA 83, 033402 (2011).
      !
      rate = rate + dble(2*L+1) * M0**2         ! S=0: J = L
      rate = rate + dble(3*(2*L+1)) * M1**2     ! S=1: J = L-1, L, L+1 --> x3
    end do
  end do loop_l4
  rate = PI * N12 * rate / dble(2*l3+1)
end function Auger_group_rate_extended

function Auger_group_rate_extended_jj( F, i1, i2, ind3, E ) result (rate)
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, ind3
  real(long), intent(in) :: E
  real(long) :: rate
  integer, parameter :: L_max = 100              ! maximum L
  real(long) :: u(F%N_uni), up(F%N_uni)
  integer :: i, spin
  integer :: n1, n2, n3
  integer :: l1, l2, l3, l4, K, J, j1, j2, j3, j4
  integer :: i1_nr, i2_nr, i3_nr
  integer :: occ1, occ2
  real(long) :: N12, tau, R12(0:L_max), B12, R21(0:L_max), B21, ll
  real(long) :: u3(F%N)
  !real(long) :: u3_uni(F%N_uni)        ! we don't need it anymore
  real(long) :: sum_J, sum_j4 
  real(long) :: sum_dir, sum_exc !, fr(F%N), dr
  real(long), external :: cleb, sixj

  n1 = F%config_rel(i1,1) 
  l1 = F%config_rel(i1,2)
  j1 = F%config_rel(i1,3)
  occ1 = F%config_rel(i1,4)
  i1_nr = F%config_rel(i1,5)    ! index for non-relativistic configuration

  n2 = F%config_rel(i2,1)  
  l2 = F%config_rel(i2,2)
  j2 = F%config_rel(i2,3)
  occ2 = F%config_rel(i2,4)
  i2_nr = F%config_rel(i2,5)    ! index for non-relativistic configuration

  spin = 1
  if ( ind3 < 0 ) spin = -1

  l3 = int( abs(ind3) / F%N )
  j3 = abs( 2*l3 + spin )
  n3 = mod( abs(ind3), F%N ) + l3

  !
  ! search (n3,l3) in config(:,:) to assign i3_nr (non-relativistic index)
  ! if the hole is not in config(:,:), then use F%u_all
  ! when F%u_all is used, solve_SE_all() should be called beforehand
  !
  i3_nr = 0
  do i = 1, F%N_orb
    if ( n3 == F%config(i,1) .and. l3 == F%config(i,2) ) i3_nr = i
  end do
  if ( i3_nr /= 0 ) then
    u3(:) = F%u(:,i3_nr)                ! pick up from occupied bound states
  else
    u3(:) = F%u_all(:,abs(ind3))        ! pick up from F%u_all
  end if
  !
  ! N12 to take the average for the final hole state qq'
  !
  if ( n1 == n2 .and. l1 == l2 .and. j1 == j2 ) then          ! two equivalent electrons
!  if ( i1  == i2 ) then
    N12 = dble( occ1 * (occ1-1) ) * dble(j1+1) / dble(j1)
    tau = ONE / sqrt(TWO)
  else                          ! two different electrons
    N12 = dble( occ1 * occ2 ) 
    tau = ONE
  end if
  if ( l1+l2+l3 > L_max ) then
    write(STDERR,*) "l1 = ", l1
    write(STDERR,*) "l2 = ", l2
    write(STDERR,*) "l3 = ", l3
    write(STDERR,*) "Too large l's: l1+l2+l3 should be <= ", L_max
    stop
  end if
  !
  ! summation over l4
  !   range of l4: |l3-l4| <= L <= l3+l4 --> |L-l3| <= l4 <= L+l3
  !            and |l1-l2| <= L <= l1+l2
  !   so, just take upper bound of l4: l4 <= l1+l2+l3
  !
  rate = ZERO
  loop_l4: do l4 = 0, l1+l2+l3
     R12(:) = ZERO
     R21(:) = ZERO
     ll = dble(l4)               ! unbound.f accepts real*8 rather than integer
     call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, E, F, 'A' )
     sum_j4 = ZERO
     ! instead of j, 2*j is used to make it an integer
     loop_j4: do j4 = abs(2*l4-1), 2*l4+1, 2        
        sum_J = ZERO
        ! J is also multilied by 2
        loop_J: do J = abs(j1-j2), j1+j2, 2 
           sum_dir = ZERO
           sum_exc = ZERO
           ! dirrect & exchange term
           ! For direct   term, k runs from abs(l2-l3) to l2+l3.
           ! For exchange term, k runs from abs(l1-l3) to l1+l3.
           ! To calculate them together, k runs over as below.
           ! B12 or B21 becomes 0, if one of the above triangular relations are not satisfied,
           ! which prevents counting unphysical radial integrals.
           loop_k: do k = min(abs(l2-l3), abs(l1-l3)), max(l2+l3, l1+l3) 
              ! direct term
              B12 = ZERO
              ! Selection rules for the reduced matrix elements.
              ! See Eq. (32) in 8. 5. 2. in Varshalovich
              if ( mod(l4+k+l1, 2) == 0 .and. mod(l3+k+l2, 2) == 0) then
                 B12 = sixj(j4, j1, 2*k, j2, j3, J) * sixj(2*l1, 1, j1, j4, 2*k, 2*l4) &
                      *sixj(2*l2, 1, j2, j3, 2*k, 2*l3) &
                      *reduced_matrix(k, l4, l1) * reduced_matrix(k, l3, l2)
                 if ( B12 /= ZERO .and. R12(k) == ZERO ) then
!                    R12(k) = double_integral( F%N_uni, F%r_uni(:), &
!                         F%u_uni( :,F%config_rel(i1,5) ), F%u_uni( :,F%config_rel(i2,5) ), u(:), u3_uni(:),  k )
!                    R12(k) = double_integral_mod(k, F%N_uni, F%r_uni(:), u(:), F%u_uni( :,F%config_rel(i1,5) ), & 
!                                F%N, F%r(:), F%weight(:), F%u(:, F%config_rel(i3, 5)), F%u(:, F%config_rel(i2, 5)) )
                   !
                   ! WARNING: this R12 uses a different convention from 
                   !          what is used in Auger_group_rate_extended()
                   !
                   R12(k) = double_integral_mod( K, &
                            F%r(:), F%weight(:), F%u(:,i2_nr), u3(:), &
                            F%r_uni(:), F%u_uni(1:F%N_uni,i1_nr), u(1:F%N_uni), F%rm_thres )
                 endif
              endif
              sum_dir = sum_dir + R12(k)*B12

              ! exchange term
              B21 = ZERO
              ! Selection rules for the reduced matrix elements.
              ! See Eq. (32) in 8. 5. 2. in Varshalovich
              if ( mod(l4+k+l2, 2) == 0 .and. mod(l3+k+l1, 2) == 0) then
                 B21 = sixj(j4, j2, 2*k, j1, j3, J) * sixj(2*l2, 1, j2, j4, 2*k, 2*l4) &
                      *sixj(2*l1, 1, j1, j3, 2*k, 2*l3) &
                      *reduced_matrix(k, l4, l2) * reduced_matrix(k, l3, l1)
                 if ( B21 /= ZERO .and. R21(k) == ZERO ) then
                    if ( i1 == i2 )then
                       R21(k) = R12(k)
                    else
!                       R21(k) = double_integral( F%N_uni, F%r_uni(:), &
!                            F%u_uni( :,F%config_rel(i2,5) ), F%u_uni( :,F%config_rel(i1,5) ), u(:), u3_uni(:),  k )
!                       R21(k) = double_integral_mod(k, F%N_uni, F%r_uni(:), u(:), F%u_uni( :,F%config_rel(i2,5) ), & 
!                                F%N, F%r(:), F%weight(:), F%u(:, F%config_rel(i3, 5)), F%u(:, F%config_rel(i1, 5)) )
                   !
                   ! WARNING: this R21 uses a different convention from 
                   !          what is used in Auger_group_rate_extended()
                   !
                   R21(k) = double_integral_mod( K, &
                            F%r(:), F%weight(:), F%u(:,i1_nr), u3(:), &
                            F%r_uni(:), F%u_uni(1:F%N_uni,i2_nr), u(1:F%N_uni), F%rm_thres )
                    endif
                 endif
              endif
              sum_exc = sum_exc + R21(k)*B21
           enddo loop_k
           ! The phase difference between the direct and exchange term is (-1)^J.
           ! This can be also confirmed in Eq. (29) in Asaad, Nucl. Phys. 44, 415, (1963) 
           if ( mod(J/2, 2) /= 0 ) sum_exc = -sum_exc
           sum_J = sum_J + (dble(J+1))*abs(sum_dir + sum_exc)**2
        enddo loop_J
        sum_j4 = sum_j4 + (dble(j4+1))*sum_J
     enddo loop_j4
     rate = rate + sum_j4
  enddo loop_l4
  rate = tau**2*TWO * PI * N12 * rate 

end function Auger_group_rate_extended_jj

!----------------------------------------------------------------------------
! print out lifetime in fs or attosec.
!----------------------------------------------------------------------------
function lifetime( rate ) result (string)
  real(long), intent(in) :: rate                ! in a.u.
  real(long) :: tau
  character(len=15) :: str, string
  tau = ONE / rate * AU2SEC / 1e-15_long        ! to femtosec.
  if ( tau > ONE ) then
    write( str, "(f12.1,' fs')" ) tau
  else
    tau = tau * 1000_long                       ! to attosec.
    write( str, "(f12.1,' as')" ) tau
  end if
  string = trim( adjustl( str ) )
end function lifetime

!----------------------------------------------------------------------------
! create all possible single-hole states and calculate their rates
!----------------------------------------------------------------------------
subroutine create_hole( P, G, F, S )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(inout) :: F
  type(Process_data), intent(inout) :: S
  integer :: i
  integer :: OUTPUT
  real(long) :: e_HOAO  ! highest-occupied-atomic-orbital
  OUTPUT = P%OUTPUT
  e_HOAO = maxval( F%e_orb(1:F%N_orb) )
  !
  ! create single-hole states
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# ',a)") repeat( "=", 75 )
  write(OUTPUT, "('# Create all possible single-hole states')")
  write(OUTPUT, "('#')")
  do i = 1, F%N_orb
    if ( F%config(i,3) == 0 ) cycle     ! no electron here
    if ( F%e_orb(i) == e_HOAO ) cycle   ! skip HOMO... wait, it's HOAO!
    F%config(i,3) = F%config(i,3) - 1   ! create a hole
    write(OUTPUT, "('# - single-hole in ',a,' -> configuration: ',a)") &
      trim( WF_name(F,i) ), trim( config_to_string( F%config(1:F%N_orb,:) ) )
    write(OUTPUT, "('# - calculate decay processes only involving '" // &
                  ",a,' hole')") trim( WF_name(F,i) )
    !
    ! initialize S every time!
    !
    S%config(:,:) = F%config(:,:)
    S%N_process = 0
    S%occ_up(:,:) = 0
    S%occ_down(:,:) = 0
    !
    ! calculate fluorescence and Auger rates, and print them out
    !
!    call calculate_fluorescence_rates_extended( P, G, F, S )
!    call calculate_Auger_rates_extended( P, F, S )
    call calculate_fluorescence_rates( P, G, F, S, i )
    call calculate_Auger_rates( P, F, S, i )
    call print_Process( S, P%IO_DATA )
    write(OUTPUT, "('#')")
    F%config(i,3) = F%config(i,3) + 1   ! recover the original configuration
  end do
end subroutine create_hole

!============================================================================
! For resonant Auger recombination process
!   same as Auger decay, but i' is determined among bound states
!============================================================================
!----------------------------------------------------------------------------
! SR find_Auger_recombination( P, F ):
!   Let's find out possibilities of resonant Auger recombination
!   same as Auger decay, but i' is determined among bound states
!   via energy conservation: e_i + e_i' = e_j + e_j'
!----------------------------------------------------------------------------
subroutine find_Auger_recombination( P, F, linewidth_eV )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  !type(Process_data), intent(inout) :: S
  real(long), optional :: linewidth_eV
  integer :: i,  occ_n_i        ! initially hole
  integer :: ip, occ_n_ip       ! initially hole <- to be determined
  integer :: j,  occ_n_j        ! initially occupied, finally hole
  integer :: jp, occ_n_jp       ! initially occupied, finally hole
  integer :: occ_max, n_hole
  real(long) :: IE_i, IE_ip, IE_j, IE_jp, E_offset      ! in eV
  real(long) :: rate, lw_eV
  !character(len=15) :: string
  character(len=4) :: str_hole
  integer :: OUTPUT, N_orb
  N_orb = F%N_subshell
  OUTPUT = P%OUTPUT
  !
  ! set up linewidth
  !
  if ( present( linewidth_eV) ) then
    lw_eV = linewidth_eV
  else
    lw_eV = P%linewidth * AU2EV
  end if
  !
  ! print out possible Auger recombination processes within given config(:,:)
  !
  write(OUTPUT, "('#'/'# Resonant Auger recombination:')")
  write(OUTPUT, "('# - linewidth (eV) = ',f12.5)") lw_eV
  write(OUTPUT, "('#   resonances to be checked within 4 times of linewidth')")
  write(OUTPUT, "('#   ',2(a4,x,a4,3x),5a10,a13)") &
    "i   ", "ip  ", "j   ", "jp  ", &
    "e_i(eV)", "e_ip(eV)", "e_j(eV)", "e_jp(eV)", "D_E(eV)", "rate(a.u.)"
  do i = 1, N_orb               ! --------- initial hole: i
    occ_max = get_occ_num_max( F, i )
    occ_n_i = get_occ_num( F, i )
    IE_i = ionization_E( F, i ) * AU2EV
    n_hole = occ_max - occ_n_i  ! # of initial holes
    if ( n_hole == 0 ) cycle    ! this orbital is full
    str_hole = ""
!    if ( n_hole > 1 ) write( str_hole, "(' *',i2)" ) n_hole
!    sum_subshell = ZERO
    do j = 1, N_orb             ! ---------- final hole: j
      if ( i == j ) cycle       ! skip if they are the same subshell
      IE_j = ionization_E( F, j ) * AU2EV
!      if ( IE_i <= IE_j ) cycle ! skip if E_i is higher than E_j
      occ_n_j = get_occ_num( F, j )
      if ( occ_n_j == 0 ) cycle ! skip if there is no electron
      do jp = j, N_orb          ! ---------- final hole: j'
        if ( i == jp ) cycle    ! skip if they are the same subshell
        if ( jp == j .and. occ_n_j < 2 ) cycle  ! for equivalent elec.
        IE_jp = ionization_E( F, jp ) * AU2EV
!        if ( IE_i <= IE_jp ) cycle      ! skip if E_i higher than E_jp
        occ_n_jp = get_occ_num( F, jp )
        if ( occ_n_jp == 0 ) cycle      ! skip if there is no electron
        !
        ! find out i'
        !
        if ( IE_i - IE_j - IE_jp > ZERO ) cycle ! this is an Auger decay
        do ip = i + 1, N_orb    ! ---------- initial hole: i' (i' > i)
          if ( ip == j .or. ip == jp ) cycle    ! skip for the same subshells
          IE_ip = ionization_E( F, ip ) * AU2EV
          occ_n_ip = get_occ_num( F, ip )
          if ( occ_max - occ_n_ip == 0 ) cycle  ! this orbital is full
          !
          ! initial zeroth-order energy: E - e_i - e_i'
          ! final   zeroth-order energy: E - e_j - e_j'
          !   where E includes all energies (before creating holes at i & i')
          ! difference = e_i + e_i' - e_j - e_j' = IE_j + IE_jp - IE_i - IE_ip
          !
          E_offset = IE_j + IE_jp - IE_i - IE_ip  ! in eV
          !
          ! apply energy conservation ( 4 times of linewidth )
          !
          if ( abs(E_offset) > FOUR * lw_eV ) cycle    ! within +/- 4 x lw
          rate = Auger_recombination_group_rate( P, F, j, jp, i, ip )
          if ( rate > ZERO ) then
            !
            ! apply a non-normalized Lorentzian profile
            !
            rate = rate * nonnormalized_Lorentz_profile( E_offset, lw_eV )
            write(OUTPUT,"('#   ',a4,x,a4,' - ',a4,x,a4,' : ',5f10.2,es13.5)") &
              WF_name( F, i ), WF_name( F, ip ), &
              WF_name( F, j ), WF_name( F, jp ), &
              -IE_i, -IE_ip, -IE_j, -IE_jp, E_offset, rate
          end if
        end do
      end do
    end do
  end do
contains
!----------------------------------------------------------------------------
!                                                         1
! - non-normalized Lorentz profile: L(x,gamma) = -------------------
!                                                1 + 4 (x / gamma)^2
!   with FWHM = gamma
!----------------------------------------------------------------------------
  function nonnormalized_Lorentz_profile( x, gamma ) result (L)
    real(long) :: L
    real(long), intent(in) :: x, gamma
    L = ONE / ( ONE + FOUR * ( x / gamma )**2 )
    if ( gamma == ZERO ) L = ZERO
  end function nonnormalized_Lorentz_profile
end subroutine find_Auger_recombination

!----------------------------------------------------------------------------
! compute the Auger recombinaton rates decaying from (i1,i2) to (i3,i4)
!   Auger_recomb( n3 l3, n4 l4 - n1 l1, n2 l2 )
! almost the same as Auger, but E --> i4
!----------------------------------------------------------------------------
function Auger_recombination_group_rate( P, F, i1, i2, i3, i4 ) result (rate)
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, i3, i4
  real(long) :: rate
  if ( P%yes_relativity ) then
     rate = Auger_recombination_group_rate_extended_jj( F, i1, i2, i3, i4 )
  else
     rate = Auger_recombination_group_rate_extended( F, i1, i2, i3, i4 )
  endif
end function Auger_recombination_group_rate

!----------------------------------------------------------------------------
! WARNING: This is for resonant Auger recombination, but not for Auger
!----------------------------------------------------------------------------
function Auger_recombination_group_rate_extended( F, i1, i2, i3, i4 ) result (rate)
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, i3, i4
  real(long) :: rate
  integer, parameter :: L_max = 100             ! maximum L
  real(long) :: R12(0:L_max), R21(0:L_max)      ! R_K(1,2,3,4) and R_K(2,1,3,4)
  integer :: n1, n2, n3, n4
  integer :: l1, l2, l3, l4, k, L
  integer :: occ1, occ2, occ3, occ4
  real(long) :: N12, tau, M0, M1, T
  real(long), external :: sixj
  n1 = F%config(i1,1);  l1 = F%config(i1,2);  occ1 = F%config(i1,3)
  n2 = F%config(i2,1);  l2 = F%config(i2,2);  occ2 = F%config(i2,3)
  n3 = F%config(i3,1);  l3 = F%config(i3,2);  occ3 = F%config(i3,3)
  n4 = F%config(i4,1);  l4 = F%config(i4,2);  occ4 = F%config(i4,3)
  !
  ! N12 is given by Bhalla's paper
  ! SEE: Eq. (3) in Bhalla, Folland & Hein, PRA 8, 649 (1973).
  ! Equivalently, N_jj' is given by Son's paper
  ! SEE: Eq. (8) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  if ( i1 == i2 ) then          ! two equivalent electrons
    N12 = dble( occ1 * (occ1-1) ) / dble( (4*l1+2) * (4*l1+2-1) )
    tau = ONE / sqrt(TWO)
  else                          ! two different electrons
    N12 = dble( occ1 * occ2 ) / dble( (4*l1+2) * (4*l2+2) )
    tau = ONE
  end if
  !
  ! calculate at l4
  !   range of l4: |l3-l4| <= L <= l3+l4 --> |L-l3| <= l4 <= L+l3
  !            and |l1-l2| <= L <= l1+l2
  !   so, just take upper bound of l4: l4 <= l1+l2+l3
  !
  rate = ZERO
  R12(:) = ZERO
  R21(:) = ZERO
  if ( l1+l2+l3 < l4 ) return
  do L = abs(l1-l2), l1+l2    ! |l1-l2| <= L <= l1+l2
    M0 = ZERO                 ! M_LS for S=0, SEE: Eq. (9) in PRA 83, 033402
    M1 = ZERO                 ! M_LS for S=1, SEE: Eq. (9) in PRA 83, 033402
    do K = 0, l1+l2+l3
      !
      ! computing R_K(1,2,3,4)
      ! T      -> A_K(j,j',i,i'), SEE: Eq. (10) in PRA 83, 033402
      ! R12(K) -> R_K(j,j',i,i'), SEE: Eq. (12) in PRA 83, 033402
      !
      T = reduced_matrix( K, l3, l1 ) * reduced_matrix( K, l4, l2 ) * &
          sixj( l3*2, l4*2, L*2, l2*2, l1*2, K*2 )
      if ( T /= ZERO .and. R12(K) == ZERO ) then
        R12(K) = double_integral_nonuniform( K, F%r(:), F%weight(:), &
                 F%u(:,i1), F%u(:,i2), F%u(:,i3), F%u(:,i4) )
      end if
      M0 = M0 + R12(K) * T
      M1 = M1 + R12(K) * T
      !
      ! computing R_K(2,1,3,4)
      ! T      -> A_K(j',j,i,i'), SEE: Eq. (10) in PRA 83, 033402
      ! R21(K) -> R_K(j',j,i,i'), SEE: Eq. (12) in PRA 83, 033402
      !
      T = reduced_matrix( K, l3, l2 ) * reduced_matrix( K, l4, l1 ) * &
          sixj( l3*2, l4*2, L*2, l1*2, l2*2, K*2 )
      if ( T /= ZERO .and. R21(K) == ZERO ) then
        if ( i1 == i2 .and. R12(K) /= ZERO ) then
          R21(K) = R12(K)
        else
          R21(K) = double_integral_nonuniform( K, F%r(:), F%weight(:), &
                   F%u(:,i2), F%u(:,i1), F%u(:,i3), F%u(:,i4) )
        end if
      end if
      !
      ! considering (-1)^(L+S)
      !   for two equivalent electrons and L=0,
      !   in fact S=1 is forbidden by the parity consideration,
      !   but here it can be excluded by a (-1)^(0+1) factor.
      !   i.e. M1 becomes 0 since R_K(1,2,3,4)=R_K(2,1,3,4) and other terms.
      !
      ! M_LS is properly summed over K up according to S=0 (M0) and S=1 (M1)
      ! SEE: Eq. (9) in Son, Young & Santra, PRA 83, 033402 (2011).
      !
      if ( mod(L,2) == 0 ) then       ! considering (-1)^(L+S)
        M0 = M0 + R21(K) * T          ! L+S is even
        M1 = M1 - R21(K) * T          ! L+S is odd
      else
        M0 = M0 - R21(K) * T          ! L+S is odd
        M1 = M1 + R21(K) * T          ! L+S is even
      end if
    end do
    !
    ! considering tau and (-1)^( L + l_j + l_i' )
    ! SEE: Eq. (9) in Son, Young & Santra, PRA 83, 033402 (2011).
    !
    M0 = tau * M0;  if ( mod(L+l1+l4,2) == 1 ) M0 = -M0
    M1 = tau * M1;  if ( mod(L+l1+l4,2) == 1 ) M1 = -M1
    !
    ! summation over L and l4
    ! SEE: Eq. (7) in Son, Young & Santra, PRA 83, 033402 (2011).
    !
    rate = rate + dble(2*L+1) * M0**2         ! S=0: J = L
    rate = rate + dble(3*(2*L+1)) * M1**2     ! S=1: J = L-1, L, L+1 --> x3
  end do
  rate = PI * N12 * rate / dble(2*l3+1)
end function Auger_recombination_group_rate_extended

!----------------------------------------------------------------------------
! WARNING: This is for resonant Auger recombination, but not for Auger
!----------------------------------------------------------------------------
function Auger_recombination_group_rate_extended_jj( F, i1, i2, i3, i4 ) result (rate)
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: i1, i2, i3, i4
  real(long) :: rate
  integer, parameter :: L_max = 100              ! maximum L
  !integer :: i, ii, ind3, ind4
  integer :: n1, n2, n3, n4
  integer :: l1, l2, l3, l4, K, J, j1, j2, j3, j4
  integer :: i1_nr, i2_nr, i3_nr, i4_nr
  integer :: occ1, occ2, occ3, occ4
  real(long) :: N12, tau, R12(0:L_max), B12, R21(0:L_max), B21
  !real(long) :: u3(F%N)
  !real(long) :: u3_uni(F%N_uni)        ! we don't need it anymore
  real(long) :: sum_J, sum_j4 
  real(long) :: sum_dir, sum_exc !, fr(F%N), tmp, dr
  real(long), external :: cleb, sixj

  n1 = F%config_rel(i1,1) 
  l1 = F%config_rel(i1,2)
  j1 = F%config_rel(i1,3)
  occ1 = F%config_rel(i1,4)
  i1_nr = F%config_rel(i1,5)    ! index for non-relativistic configuration

  n2 = F%config_rel(i2,1)  
  l2 = F%config_rel(i2,2)
  j2 = F%config_rel(i2,3)
  occ2 = F%config_rel(i2,4)
  i2_nr = F%config_rel(i2,5)    ! index for non-relativistic configuration

  n3 = F%config_rel(i3,1)  
  l3 = F%config_rel(i3,2)
  j3 = F%config_rel(i3,3)
  occ3 = F%config_rel(i3,4)
  i3_nr = F%config_rel(i3,5)    ! index for non-relativistic configuration

  n4 = F%config_rel(i4,1)  
  l4 = F%config_rel(i4,2)
  j4 = F%config_rel(i4,3)
  occ4 = F%config_rel(i4,4)
  i4_nr = F%config_rel(i4,5)    ! index for non-relativistic configuration

  !
  ! N12 to take the average for the final hole state qq'
  !
  if ( n1 == n2 .and. l1 == l2 .and. j1 == j2 ) then          ! two equivalent electrons
!  if ( i1  == i2 ) then
    N12 = dble( occ1 * (occ1-1) ) * dble(j1+1) / dble(j1)
    tau = ONE / sqrt(TWO)
  else                          ! two different electrons
    N12 = dble( occ1 * occ2 ) 
    tau = ONE
  end if
  !
  ! calculate at l4
  !   range of l4: |l3-l4| <= L <= l3+l4 --> |L-l3| <= l4 <= L+l3
  !            and |l1-l2| <= L <= l1+l2
  !   so, just take upper bound of l4: l4 <= l1+l2+l3
  !
  rate = ZERO
  R12(:) = ZERO
  R21(:) = ZERO
  sum_j4 = ZERO
  if ( l1+l2+l3 < l4 ) return
  ! instead of j, 2*j is used to make it an integer
  loop_j4: do j4 = abs(2*l4-1), 2*l4+1, 2        
     sum_J = ZERO
     ! J is also multilied by 2
     loop_J: do J = abs(j1-j2), j1+j2, 2 
        sum_dir = ZERO
        sum_exc = ZERO
        ! dirrect & exchange term
        ! For direct   term, k runs from abs(l2-l3) to l2+l3.
        ! For exchange term, k runs from abs(l1-l3) to l1+l3.
        ! To calculate them together, k runs over as below.
        ! B12 or B21 becomes 0, if one of the above triangular relations are not satisfied,
        ! which prevents counting unphysical radial integrals.
        loop_k: do k = min(abs(l2-l3), abs(l1-l3)), max(l2+l3, l1+l3) 
           ! direct term
           B12 = ZERO
           ! Selection rules for the reduced matrix elements.
           ! See Eq. (32) in 8. 5. 2. in Varshalovich
           if ( mod(l4+k+l1, 2) == 0 .and. mod(l3+k+l2, 2) == 0) then
              B12 = sixj(j4, j1, 2*k, j2, j3, J) * sixj(2*l1, 1, j1, j4, 2*k, 2*l4) &
                   *sixj(2*l2, 1, j2, j3, 2*k, 2*l3) &
                   *reduced_matrix(k, l4, l1) * reduced_matrix(k, l3, l2)
              if ( B12 /= ZERO .and. R12(k) == ZERO ) then
!                 R12(k) = double_integral( F%N_uni, F%r_uni(:), &
!                      F%u_uni( :,F%config_rel(i1,5) ), F%u_uni( :,F%config_rel(i2,5) ), u(:), u3_uni(:),  k )
!                 R12(k) = double_integral_mod(k, F%N_uni, F%r_uni(:), u(:), F%u_uni( :,F%config_rel(i1,5) ), & 
!                             F%N, F%r(:), F%weight(:), F%u(:, F%config_rel(i3, 5)), F%u(:, F%config_rel(i2, 5)) )
                !
                ! WARNING: it seems R12 & R21 are exchanged in Koudai's code
                !
                R12(k) = double_integral_nonuniform( K, &
                         F%r(:), F%weight(:), F%u(:,i2_nr), F%u(:,i1_nr), &
                         F%u(:,i3_nr), F%u(:,i4_nr) )
              endif
           endif
           sum_dir = sum_dir + R12(k)*B12

           ! exchange term
           B21 = ZERO
           ! Selection rules for the reduced matrix elements.
           ! See Eq. (32) in 8. 5. 2. in Varshalovich
           if ( mod(l4+k+l2, 2) == 0 .and. mod(l3+k+l1, 2) == 0) then
              B21 = sixj(j4, j2, 2*k, j1, j3, J) * sixj(2*l2, 1, j2, j4, 2*k, 2*l4) &
                   *sixj(2*l1, 1, j1, j3, 2*k, 2*l3) &
                   *reduced_matrix(k, l4, l2) * reduced_matrix(k, l3, l1)
              if ( B21 /= ZERO .and. R21(k) == ZERO ) then
                 if ( i1 == i2 )then
                    R21(k) = R12(k)
                 else
!                    R21(k) = double_integral( F%N_uni, F%r_uni(:), &
!                         F%u_uni( :,F%config_rel(i2,5) ), F%u_uni( :,F%config_rel(i1,5) ), u(:), u3_uni(:),  k )
!                    R21(k) = double_integral_mod(k, F%N_uni, F%r_uni(:), u(:), F%u_uni( :,F%config_rel(i2,5) ), & 
!                             F%N, F%r(:), F%weight(:), F%u(:, F%config_rel(i3, 5)), F%u(:, F%config_rel(i1, 5)) )
                R21(k) = double_integral_nonuniform( K, &
                         F%r(:), F%weight(:), F%u(:,i1_nr), F%u(:,i2_nr), &
                         F%u(:,i3_nr), F%u(:,i4_nr) )
                 endif
              endif
           endif
           sum_exc = sum_exc + R21(k)*B21
        enddo loop_k
        ! The phase difference between the direct and exchange term is (-1)^J.
        ! This can be also confirmed in Eq. (29) in Asaad, Nucl. Phys. 44, 415, (1963) 
        if ( mod(J/2, 2) /= 0 ) sum_exc = -sum_exc
        sum_J = sum_J + (dble(J+1))*abs(sum_dir + sum_exc)**2
     enddo loop_J
     sum_j4 = sum_j4 + (dble(j4+1))*sum_J
  enddo loop_j4
  rate = rate + sum_j4
  rate = tau**2*TWO * PI * N12 * rate 
end function Auger_recombination_group_rate_extended_jj

end module decay
