!----------------------------------------------------------------------------
! eigen_LAPACK.f90
!----------------------------------------------------------------------------
! Copyright (C) 2005 Sang-Kil Son
! Copyright (C) 2010 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Eigenproblem module : to solve eigenvalue problem using LAPACK
!   written by Son, Sang-Kil in Mar. 2005
!
! USAGE: H should be a regular N x N matrix.
!        Eigenvectors are normalized after each subroutine.
! - SR get_eigenvalues( H, E )          for all ev. / standard eigen problem
!      get_eigenvalues( A, B, E )       for all ev. / generalized eigen problem
!   . SR get_all_ev_of_real_sym( H, E, yes_verbose )
!   . SR get_all_ev_of_real_nonsym( H, E, VL, VR, yes_verbose )
!   . SR get_all_ev_of_real_nonsym_simple( H, E, yes_verbose )
!   . SR get_all_ev_of_complex_Hermitian( H, E, yes_verbose )
!   . SR get_all_ev_of_complex_sym( H, E, yes_verbose )
!   . SR get_all_ev_of_complex_nonsym( H, E, VL, VR, yes_verbose )
!   . SR get_gen_all_ev_real_sym( A, B, E, yes_verbose )
!   . SR get_gen_all_ev_real_nonsym( A, B, E, VL, VR, yes_verbose )
! - SR get_eigenvalues( H, E, m, E0, dE )       for selected ev.
!   *** NOT YET IMPLEMENTED
!
! for real symmetric matrix
! - DSYEV  : all eigenvalues and eigenvectors
! - DSYEVD : all eigenvalues and eigenvectors by divide-and-conquer
! - DSYEVX : selected eigenvalues and eigenvectors
! - DSYEVR : selected eigenvalues and eigenvectors by divide-and-conquer
! - DSYGV  : all eigenvalues/eigenvectors of generalized symmetric-definite
!
! for real nonsymmetric matrix
! - DGEEV  : all eigenvalues and left/right eigenvectors
! - DGEEVX : all eigenvalues and left/right eigenvectors with balancing
! - DGEGV / DGGEV  : all eigenvalues/eigenvectors of generalized non-symmetric
!
! for complex Hermitian matrix
! - ZHEEV  : all eigenvalues and eigenvectors
! - ZHEEVD : all eigenvalues and eigenvectors by divide-and-conquer
! - ZHEEVX : selected eigenvalues and eigenvectors
! - ZHEEVR : selected eigenvalues and eigenvectors by divide-and-conquer
!
! for complex non-Hermitian matrix
! - ZGEEV  : all eigenvalues and left/right eigenvectors
! - ZGEEVX : all eigenvalues and left/right eigenvectors with balancing
!
! URL: http://www.netlib.org/lapack/double/
!      http://www.netlib.org/lapack/complex16/
!----------------------------------------------------------------------------
module eigen_LAPACK
  use constants
  implicit none
  interface get_eigenvalues
    module procedure get_all_ev_of_real_sym, &
                     get_all_ev_of_real_nonsym, &
                     get_all_ev_of_real_nonsym_simple, &
                     get_all_ev_of_complex_Hermitian, &
                     get_all_ev_of_complex_sym, &
                     get_all_ev_of_complex_nonsym, &
                     get_gen_all_ev_real_sym, &
                     get_gen_all_ev_real_nonsym
  end interface
  interface check_normalized
    module procedure check_normal_real_sym, &
                     check_normal_real_nonsym, &
                     check_normal_complex_Hermitian, &
                     check_normal_complex_nonsym
  end interface
contains

!----------------------------------------------------------------------------
! Solve the eigenvalue problem of H(N,N) real symmetric matrix
! - eigenvalues:  real, E(N)
! - eigenvectors: real, H(N,N)
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
subroutine get_all_ev_of_real_sym( H, E, yes_verbose )
  real(long), intent(inout) :: H(:,:)
  real(long), intent(inout) :: E(:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: WORK(size(E)*size(E))
  integer :: N
  integer :: INFO, LDWORK
  N = size(E)
  INFO = 0
  LDWORK = max( N*N, 3*N-1 )
  !
  ! LAPACK:
  !   DSYEV(JOBZ, UPLO, N, A, LDA, W, WORK, LDWORK, INFO)
  !
  call DSYEV( 'V', 'U', N, H, N, E, WORK, LDWORK, INFO )
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in DSYEV : INFO = ", INFO
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in DSYEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_real_sym

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) real nonsym. matrix using LAPACK
! - eigenvalues:  complex, E(N)
! - eigenvectors are not computed.
!----------------------------------------------------------------------------
subroutine get_all_ev_of_real_nonsym_simple( H, E, yes_verbose )
  real(long), intent(inout) :: H(:,:)
  complex(long), intent(out) :: E(:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: WORK(64*size(E)), WR(size(E)), WI(size(E))
  real(long) :: dummy(1)
  complex(long) :: temp
  integer :: INFO, LDA, LDVL, LDVR, LWORK
  integer :: N, k, l, loc(1)
  external DGEEV
  N = size(E)
  INFO = 0      ! INFO must be a variable for NAG in Linux machine
  LDA = size(H,dim=1)
  LDVR = 1;  LDVL = 1
  LWORK = 64*N
  !
  ! LAPACK: http://www.netlib.org/lapack/double/dgeev.f
  !         DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, LDVR,
  !                WORK, LWORK, INFO )
  !
  call DGEEV( 'N', 'N', N, H, LDA, WR, WI, dummy, LDVL, dummy, LDVR, & 
              WORK, LWORK, INFO )
  !
  ! set complex numbers of E
  !
  E(:) = dcmplx( WR(:), WI(:) )
  !
  ! sort eigenvalues and eigenvectors by real(E)
  !
  do k = 1, N-1
    if ( dble( E(k) ) > minval( dble( E(k+1:N) ) ) ) then
      !
      ! swap i and minloc(E)
      !
      loc = minloc( dble( E(k+1:N) ) )
      l = loc(1) + k
      temp = E(k);          E(k) = E(l);        E(l) = temp
    end if
  end do
  !
  ! if there is an error
  !
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in DGEEV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else
      write(STDERR,*) "the QR algorithm failed to compute all the eigenvalues,"
      write(STDERR,*) "and no eigenvectors have been computed."
      if ( INFO < N ) then
        write(STDERR,*) "Some converged eigenvalues:"
        do k = INFO+1, N
          write(STDERR,*) WR(k), WI(k)
        end do
      else
        write(STDERR,*) "There is no converged eigenvalues."
      end if
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in ZGEEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_real_nonsym_simple

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) real nonsym. matrix using LAPACK
! All eigenvalues are stored in E(N) and left eigenvectors in VL(N,N)
!                                        right eigenvectors in VR(N,N)
! 
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL^T A = D VL^T
!
! take transpose of first eq.
!    VR^T A^T = D^T VR^T
! consequently, left eigenvectors of A^T is VR^T.
!----------------------------------------------------------------------------
subroutine get_all_ev_of_real_nonsym( H, E, VL, VR, yes_verbose )
  real(long), intent(inout) :: H(:,:)
  complex(long), intent(out) :: E(:)
  real(long), intent(out) :: VL(:,:), VR(:,:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: WORK(64*size(E)), WR(size(E)), WI(size(E))
  real(long) :: norm, temp_v(size(E))
  complex(long) :: temp
  integer :: INFO, LDA, LDVL, LDVR, LWORK
  integer :: N, k, l, loc(1)
  external DGEEV
  N = size(E)
  INFO = 0      ! INFO must be a variable for NAG in Linux machine
  LDA = size(H,dim=1)
  LDVR = size(VR,dim=1)
  LDVL = size(VL,dim=1)
  LWORK = 64*N
  !
  ! LAPACK: http://www.netlib.org/lapack/double/dgeev.f
  !         DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, LDVR,
  !                WORK, LWORK, INFO )
  !
  call DGEEV( 'V', 'V', N, H, LDA, WR, WI, VL, LDVL, VR, LDVR, & 
              WORK, LWORK, INFO )
  !
  ! We need normalization for w.f., < VL(k) | VR(k) > = 1
  ! and set complex numbers of E
  !
  do k = 1, N
    E(k) = dcmplx( WR(k), WI(k) )
    norm = sqrt( sum( VL(:,k) * VR(:,k) ) )
    VR(:,k) = VR(:,k) / norm
    VL(:,k) = VL(:,k) / norm
  end do
  !
  ! sort eigenvalues and eigenvectors by real(E)
  !
  do k = 1, N-1
    if ( dble( E(k) ) > minval( dble( E(k+1:N) ) ) ) then
      !
      ! swap i and minloc(E)
      !
      loc = minloc( dble( E(k+1:N) ) )
      l = loc(1) + k
      temp = E(k);          E(k) = E(l);        E(l) = temp
      temp_v(:) = VR(:,k);  VR(:,k) = VR(:,l);  VR(:,l) = temp_v(:)
      temp_v(:) = VL(:,k);  VL(:,k) = VL(:,l);  VL(:,l) = temp_v(:)
    end if
  end do
  !
  ! if there is an error
  !
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in DGEEV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else
      write(STDERR,*) "the QR algorithm failed to compute all the eigenvalues,"
      write(STDERR,*) "and no eigenvectors have been computed."
      if ( INFO < N ) then
        write(STDERR,*) "Some converged eigenvalues:"
        do k = INFO+1, N
          write(STDERR,*) WR(k), WI(k)
        end do
      else
        write(STDERR,*) "There is no converged eigenvalues."
      end if
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in ZGEEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_real_nonsym

!----------------------------------------------------------------------------
! Solve the eigenvalue problem of H(N,N) complex Hermitian matrix
! - eigenvalues:  real, E(N)
! - eigenvectors: complex, H(N,N)
!----------------------------------------------------------------------------
subroutine get_all_ev_of_complex_Hermitian( H, E, yes_verbose )
  complex(long), intent(inout) :: H(:,:)
  real(long), intent(inout) :: E(:)
  logical, intent(in), optional :: yes_verbose
  complex(long) :: WORK(size(E)*size(E))
  real(long) :: WORK2(3*size(E)-2)
  integer :: N, INFO, LDWORK
  N = size(E)
  INFO = 0
  LDWORK = max( N*N, 2*N-1 )
  !
  ! LAPACK:
  !   ZHEEV(JOBZ, UPLO, N, A, LDA, W, WORK, LDWORK, WORK2, INFO)
  !
  call ZHEEV( 'V', 'U', N, H, N, E, WORK, LDWORK, WORK2, INFO )
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in ZHEEV : INFO = ", INFO
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in ZHEEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_complex_Hermitian

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) complex symmetric matrix using LAPACK
! All eigenvalues are stored in E(N) and right eigenvectors in H(N,N)
! 
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL A = D VL
!
! take the transpose of first eq.
!    VR^T A^T = D^T VR^T
! here D^T = D; thus, VR^T A^T = D VR^T
! consequently, left eigenvectors of A^T is VR^T.
!
! In this case of the complex symmetric matrix, A = A^T
! so, VL can be simply solved by VR^T.
!----------------------------------------------------------------------------
subroutine get_all_ev_of_complex_sym( H, E, yes_verbose )
  complex(long), intent(inout) :: H(:,:), E(:)
  logical, intent(in), optional :: yes_verbose
  complex(long) :: VR(size(E), size(E)), VL(1,1)
  complex(long) :: WORK(64*size(E))
  real(long) :: RWORK(2*size(E))
  complex(long) :: norm, temp, temp_v(size(E))
  integer :: INFO, LDA, LDVL, LDVR, LWORK
  integer :: N, k, l, loc(1)
  external ZGEEV
  N = size(E)
  INFO = 0      ! INFO must be a variable for NAG in Linux machine
  LDA = N
  LDVR = N
  LDVL = 1
  LWORK = 64*N
  !
  ! LAPACK:
  !   ZGEEV( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR,
  !          WORK, LWORK, RWORK, INFO )
  !
  call ZGEEV( 'N', 'V', N, H, LDA, E, VL, LDVL, VR, LDVR, &
              WORK, LWORK, RWORK, INFO )
  !
  ! We need normalization for w.f., < VL(k) | VR(k) > = 1
  ! Here, H is complex symmetric. VL = VR^T
  !
  do k = 1, N
    norm = sqrt( sum( VR(:,k)**2 ) )
    H(:,k) = VR(:,k) / norm
  end do
  !
  ! sort eigenvalues and eigenvectors by real(E)
  !
  do k = 1, N-1
    if ( real( E(k) ) > minval( real( E(k+1:N) ) ) ) then
      !
      ! swap i and minloc(E)
      !
      loc = minloc( real( E(k+1:N) ) )
      l = loc(1) + k
      temp = E(k);         E(k) = E(l);      E(l) = temp
      temp_v(:) = H(:,k);  H(:,k) = H(:,l);  H(:,l) = temp_v(:)
    end if
  end do
  !
  ! if there is an error
  !
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in ZGEEV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else
      write(STDERR,*) "the QR algorithm failed to compute all the " // &
                      "eigenvalues, and no eigenvectors have been computed."
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in ZGEEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_complex_sym

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) complex nonsym. matrix using LAPACK
! All eigenvalues are stored in E(N) and left eigenvectors in VL(N,N)
!                                        right eigenvectors in VR(N,N)
! 
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL^* A = D VL^*
!
! take transpose + conjugate of first eq.
!    VR^* A^* = D^* VR^*
! consequently, left eigenvectors of A^* is VR^*.
!----------------------------------------------------------------------------
subroutine get_all_ev_of_complex_nonsym( H, E, VL, VR, yes_verbose )
  complex(long), intent(inout) :: H(:,:)
  complex(long), intent(out) :: E(:), VL(:,:), VR(:,:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: RWORK(2*size(E))
  complex(long) :: WORK(64*size(E))
  complex(long) :: norm
  complex(long) :: temp, temp_v(size(E))
  integer :: INFO, LDA, LDVL, LDVR, LWORK
  integer :: N, k, l, loc(1)
  external ZGEEV
  N = size(E)
  INFO = 0      ! INFO must be a variable for NAG in Linux machine
  LDA = size(H,dim=1)
  LDVR = size(VR,dim=1)
  LDVL = size(VL,dim=1)
  LWORK = 64*N
  !
  ! LAPACK: http://www.netlib.org/lapack/complex16/zgeev.f
  !         ZGEEV( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR,
  !                WORK, LWORK, RWORK, INFO )
  !
  call ZGEEV( 'V', 'V', N, H, LDA, E, VL, LDVL, VR, LDVR, & 
              WORK, LWORK, RWORK, INFO )
  !
  ! We need normalization for w.f., < VL(k) | VR(k) > = 1
  !
  do k = 1, N
    norm = sqrt( sum( conjg(VL(:,k)) * VR(:,k) ) )
    VR(:,k) = VR(:,k) / norm
    VL(:,k) = VL(:,k) / conjg(norm)
  end do
  !
  ! sort eigenvalues and eigenvectors by real(E)
  !
  do k = 1, N-1
    if ( dble( E(k) ) > minval( dble( E(k+1:N) ) ) ) then
      !
      ! swap i and minloc(E)
      !
      loc = minloc( dble( E(k+1:N) ) )
      l = loc(1) + k
      temp = E(k);          E(k) = E(l);        E(l) = temp
      temp_v(:) = VR(:,k);  VR(:,k) = VR(:,l);  VR(:,l) = temp_v(:)
      temp_v(:) = VL(:,k);  VL(:,k) = VL(:,l);  VL(:,l) = temp_v(:)
    end if
  end do
  !
  ! if there is an error
  !
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in ZGEEV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else
      write(STDERR,*) "the QR algorithm failed to compute all the " // &
                      "eigenvalues, and no eigenvectors have been computed."
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in ZGEEV = ", WORK(1)
  end if
end subroutine get_all_ev_of_complex_nonsym


!----------------------------------------------------------------------------
! Solve the generalized eigenvalue problem of real symmetric matrix
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
subroutine get_gen_all_ev_real_sym( A, B, E, yes_verbose )
  real(long), intent(inout) :: A(:,:), B(:,:)
  real(long), intent(inout) :: E(:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: WORK(size(E)*size(E))
  integer :: N
  integer :: ITYPE, INFO, LDWORK
  N = size(E)
  ITYPE = 1
  INFO = 0
  LDWORK = max( N*N, 3*N-1 )
  !
  ! LAPACK:
  !   DSYGV(ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, LDWORK, INFO)
  !     ITYPE = 1: A * x = lambda * B * x
  !           = 2: A * B * x = lambda * x
  !           = 3: B * A * x = lambda * x
  !
  call DSYGV( ITYPE, 'V', 'U', N, A, N, B, N, E, WORK, LDWORK, INFO )
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in DSYGV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else 
      write(STDERR,*) "DPOTRF or DSYEV returned an error code:"
      if ( INFO <= N ) then
        write(STDERR,*) "  DSYEV failed to converge;"
        write(STDERR,*) "  ", INFO, "off-diagonal elements of an intermediate"
        write(STDERR,*) "  tridiagonal form did not converge to zero;"
      else
        write(STDERR,*) "  the leading minor of order ", INFO-N, &
                         " of B is not positive definite."
        write(STDERR,*) "  The factorization of B could not be completed and "
        write(STDERR,*) "  no eigenvalues or eigenvectors were computed."
      end if
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in DSYGV = ", WORK(1)
  end if
end subroutine get_gen_all_ev_real_sym


!----------------------------------------------------------------------------
! Solve the generalized eigenvalue problem of real nonsymmetric matrix
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
subroutine get_gen_all_ev_real_nonsym( A, B, E, VL, VR, yes_verbose )
  real(long), intent(inout) :: A(:,:), B(:,:)
  complex(long), intent(inout) :: E(:)
  real(long), intent(inout) :: VL(:,:), VR(:,:)
  logical, intent(in), optional :: yes_verbose
  real(long) :: ALPHAR(size(E)), ALPHAI(size(E)), BETA(size(E))
  real(long) :: WORK(size(E)*size(E))
  real(long) :: temp_v(size(E))
  complex(long) :: temp
  integer :: N, k, l, loc(1)
  integer :: ITYPE, INFO, LDWORK
  N = size(E)
  ITYPE = 1
  INFO = 0
  LDWORK = max( N*N, 8*N )
  !
  ! LAPACK:
  !   DGGEV(JOBVL, JOBVR, N, A, LDA, B, LDB, ALPHAR, ALPHAI, 
  !         BETA, VL, LDVL, VR, LDVR, WORK, LWORK, INFO )
  !
  call DGGEV( 'V', 'V', N, A, N, B, N, ALPHAR, ALPHAI, BETA, VL, N, VR, N, &
              WORK, LDWORK, INFO )
  if ( INFO /= 0 ) then
    write(STDERR,*) "ERROR in DGGEV : INFO = ", INFO
    if ( INFO < 0 ) then
      write(STDERR,*) "the ", -INFO, "th argument had an illegal value."
    else 
      if ( INFO <= N ) then
        write(STDERR,*) "  The QZ iteration failed.  No eigenvectors have been"
        write(STDERR,*) "  calculated, but ALPHAR(j), ALPHAI(j), and BETA(j)"
        write(STDERR,*) "  should be correct for j=INFO+1,...,N."
      else
        write(STDERR,*) "  =N+1: other than QZ iteration failed in DHGEQZ."
        write(STDERR,*) "  =N+2: error return from DTGEVC."
      end if
    end if
    stop
  end if
  if ( present(yes_verbose) ) then
    if ( yes_verbose ) write(STDERR,*) "Optimal LDWORK in DGGEV = ", WORK(1)
  end if
  !
  ! get eigenvalues from ALPHAR, ALPHAI, and BETA
  !
  E(:) = ( ALPHAR(:) + IMAG_ONE * ALPHAI(:) ) / BETA(:)
  !
  ! sort eigenvalues and eigenvectors by real(E), i.e. ALPHAR
  !
  do k = 1, N-1
    if ( dble( E(k) ) > minval( dble( E(k+1:N) ) ) ) then
      !
      ! swap i and minloc(E)
      !
      loc = minloc( dble( E(k+1:N) ) )
      l = loc(1) + k
      temp = E(k);          E(k) = E(l);        E(l) = temp
      temp_v(:) = VR(:,k);  VR(:,k) = VR(:,l);  VR(:,l) = temp_v(:)
      temp_v(:) = VL(:,k);  VL(:,k) = VL(:,l);  VL(:,l) = temp_v(:)
    end if
  end do
end subroutine get_gen_all_ev_real_nonsym


!----------------------------------------------------------------------------
! Check if w.f. is normalized
!   < psi_k | psi_k' > = 1  if k = k'
!                      = 0  if k /= k'
! with the matrix form
!     wf'  wf = I  for a real matrix
!     wf^T wf = I  for a complex matrix
!----------------------------------------------------------------------------
subroutine check_normal_real_sym( wf, tol )
  real(long), intent(in) :: wf(:,:)
  real(long), intent(in) :: tol
  real(long) :: C(size(wf,dim=1),size(wf,dim=2))
  real(long) :: alpha, beta
  logical :: yes_first
  integer :: N, i, j
  alpha = ONE;  beta = ZERO
  N = size(wf,dim=1)
  call DGEMM( 'T', 'N', N, N, N, alpha, wf, N, wf, N, beta, C, N )
  yes_first = .true.
  do i = 1, N
    do j = 1, N
      if ( ( i == j .and. abs( C(i,j) - ONE ) > tol ) .or. &
           ( i /= j .and. abs( C(i,j) ) > tol ) ) then
        if ( yes_first ) then
          write(STDERR,"(a)") "wavefunctions are not normalized:"
          yes_first = .false.
        end if
        write(STDERR,"(2x,'(i,j,value) = ',2i10,es25.15)") i, j, C(i,j)
      end if
    end do
  end do
end subroutine check_normal_real_sym

subroutine check_normal_real_nonsym( wf_L, wf_R, tol )
  real(long), intent(in) :: wf_L(:,:), wf_R(:,:)
  real(long), intent(in) :: tol
  real(long) :: C(size(wf_L,dim=2),size(wf_R,dim=2))
  real(long) :: alpha, beta
  logical :: yes_first
  integer :: N, i, j
  alpha = ONE;  beta = ZERO
  N = size(wf_R,dim=1)
  call DGEMM( 'T', 'N', N, N, N, alpha, wf_L, N, wf_R, N, beta, C, N )
  yes_first = .true.
  do i = 1, N
    do j = 1, N
      if ( ( i == j .and. abs( C(i,j) - ONE ) > tol ) .or. &
           ( i /= j .and. abs( C(i,j) ) > tol ) ) then
        if ( yes_first ) then
          write(STDERR,"(a)") "wavefunctions are not normalized:"
          yes_first = .false.
        end if
        write(STDERR,"(2x,'(i,j,value) = ',2i10,es25.15)") i, j, C(i,j)
      end if
    end do
  end do
end subroutine check_normal_real_nonsym

subroutine check_normal_complex_Hermitian( wf, tol )
  complex(long), intent(in) :: wf(:,:)
  real(long), intent(in) :: tol
  complex(long) :: C(size(wf,dim=1),size(wf,dim=2))
  complex(long) :: alpha, beta
  logical :: yes_first
  integer :: N, i, j
  alpha = ONE;  beta = ZERO
  N = size(wf,dim=1)
  call ZGEMM( 'C', 'N', N, N, N, alpha, wf, N, wf, N, beta, C, N )
  yes_first = .true.
  do i = 1, N
    do j = 1, N
      if ( ( i == j .and. abs( C(i,j) - ONE ) > tol ) .or. &
           ( i /= j .and. abs( C(i,j) ) > tol ) ) then
        if ( yes_first ) then
          write(STDERR,"(a)") "wavefunctions are not normalized:"
          yes_first = .false.
        end if
        write(STDERR,"(2x,'(i,j,value) = ',2i10,es25.15)") i, j, abs(C(i,j))
      end if
    end do
  end do
end subroutine check_normal_complex_Hermitian

subroutine check_normal_complex_nonsym( wf_L, wf_R, tol )
  complex(long), intent(in) :: wf_L(:,:), wf_R(:,:)
  real(long), intent(in) :: tol
  complex(long) :: C(size(wf_L,dim=2),size(wf_R,dim=2))
  complex(long) :: alpha, beta
  logical :: yes_first
  integer :: N, i, j
  alpha = ONE;  beta = ZERO
  N = size(wf_R,dim=1)
  call ZGEMM( 'C', 'N', N, N, N, alpha, wf_L, N, wf_R, N, beta, C, N )
  yes_first = .true.
  do i = 1, N
    do j = 1, N
      if ( ( i == j .and. abs( C(i,j) - ONE ) > tol ) .or. &
           ( i /= j .and. abs( C(i,j) ) > tol ) ) then
        if ( yes_first ) then
          write(STDERR,"(a)") "wavefunctions are not normalized:"
          yes_first = .false.
        end if
        write(STDERR,"(2x,'(i,j,value) = ',2i10,es25.15)") i, j, abs(C(i,j))
      end if
    end do
  end do
end subroutine check_normal_complex_nonsym

end module eigen_LAPACK
