!----------------------------------------------------------------------------
! elec_config.f90
!----------------------------------------------------------------------------
! Copyright (C) 2014-2020 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2014-2020 Sang-Kil Son <sangkil.son@cfel.de>
!          2015      Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! CONFIG module: to specify electronic configuration for atoms or molecules
!   written by Son, Sang-Kil in May 2014
!----------------------------------------------------------------------------
module elec_config
  use constants
  implicit none
  !
  ! specifying electronic configuration for different cases
  !   1) non-relativistic atomic calc. :   2   2   6   1
  !                                      1s  2s  2p  3s
  !
  !   2) relativistic atomic calc. :      2      2      2      4      1
  !                                  1s1/2  2s1/2  2p1/2  2p3/2  3s1/2
  !
  !   3) molecular calc. :    2    2    2    2    2    1
  !                        1st  2nd  3rd  4th  5th  6th 
  !      it's only energy-ordered because we use no molecular symmetry.
  !
  ! Traditional way to define electronic configuration
  ! - config(1:N_orb,1:3) <-- in XATOM, N_orb indicates # of subshells
  !   config(i,1) = n of i-th orbital
  !   config(i,2) = l of i-th orbital
  !   config(i,3) = #occupation of i-th orbital
  !
  ! - N_occ(1:N_orb) <-- in XMOLECULE
  !   N_occ(i) = #occupation of i-th orbital
  !
  ! New way to define electronic configuration with elec(:)
  ! - elec(1:N_elec) contain indexes for individual electrons
  !   elec(i) = index of i-th electron
  !   For XATOM, index = N * l + k      (k-th orbital for a given l, k=[1:N])
  !              index = N * l + n - l  (n & l: principal & orbital quantum #)
  !                                     (N: # of radial grids)
  !              index can be from 1 to N_all = N * ( lmax + 1 ).
  !              l = int( index / N )
  !              n = mod( index, N ) + l
  !   For XMOLECULE, index can be from 1 to N_basis.
  !
  !type :: Config
  !  integer, allocatable :: elec(:)
  !  integer :: N_elec
  !  integer :: N
  !  integer :: lmax
  !end type Config
  interface config_to_index
    module procedure config_to_index4, config_to_index
  end interface
  interface index_to_config
    module procedure index4_to_config, index_to_config
  end interface
contains

!----------------------------------------------------------------------------
! Transform between config(:,:) <--> elec(:)
! It is used for XATOM and requires N_r and possibly lmax.
! - SR elec_to_config( elec, config, N_r )
! - SR config_to_elec( config, elec, N_r )
! - F construct_config_from_elec( elec )
! - F construct_elec_from_config( config )
!
! how to convert config <-> elec
! For C+: 1s1 2s2 2p2
!   1s1 ==> config(1,1:3) = (/ 1, 0, 1 /)
!   2s2 ==> config(2,1:3) = (/ 2, 0, 2 /)
!   2p2 ==> config(3,1:3) = (/ 2, 1, 2 /)
! and with N=200
!   an electron in 2p ==> elec(1) = 200 * 1 + 2 - 1 = 201
!   an electron in 2p ==> elec(2) = 200 * 1 + 2 - 1 = 201
!   an electron in 2s ==> elec(3) = 200 * 0 + 2     = 2
!   an electron in 2s ==> elec(4) = 200 * 0 + 2     = 2
!   an electron in 1s ==> elec(5) = 200 * 0 + 1     = 1
!   the other electron in neutral C is ionized ==> elec(6) = 0
! Well, remember that electrons are indistinguishable.
!----------------------------------------------------------------------------
subroutine elec_to_config( elec, config, N_r )
  integer, intent(in) :: elec(:)
  integer, intent(inout) :: config(:,:)
  integer, intent(in) :: N_r            ! # of radial grids
  integer :: N_orb, N_elec
  integer :: i, j, n, l
  N_orb = size( config, 1 )
  N_elec = size( elec )
  !
  ! let's assume that config(1:N_orb,1 & 2) are set up
  ! only config(1:N_orb,3) == #occupation to be determined
  !
  do j = 1, N_orb
    n = config(j,1)
    l = config(j,2)
    i = N_r * l + n - l
    config(j,3) = count( elec(:) == i )
  end do
  if ( sum( config(:,3) ) /= N_elec ) stop "ERROR in elec_to_config()"
end subroutine elec_to_config

subroutine config_to_elec( config, elec, N_r )
  integer, intent(in) :: config(:,:)
  integer, intent(out) :: elec(:)
  integer, intent(in) :: N_r            ! # of radial grids
  integer :: N_orb, N_elec
  integer :: i, j, n, l, n_occ, temp
  integer :: jj, ang_index, Njp
  logical :: yes_relativity
  if ( size(config,dim=2) == 3 ) then   ! non-relativistic case
    yes_relativity = .false.
  else                                  ! relativistic case
    yes_relativity = .true.
  end if
  N_orb = size(config,dim=1)
  N_elec = size(elec)
  if ( N_elec == 0 ) return
  i = 0                 ! electron index
  do j = N_orb, 1, -1   ! orbital index
    n = config(j,1)     ! quantum number n
    l = config(j,2)     ! quantum number l
    if ( yes_relativity ) then
      jj = config(j,3)          ! quantum number j for the relativistic case
      n_occ = config(j,4)       ! occupation number
      Njp = 0
      if ( l == 0 ) then 
        ang_index = 0
      else
        ang_index = 2*l - 1
        if ( jj /= 2*l - 1 ) Njp = N_r
      endif
      elec(i+1:i+n_occ) = N_r * ang_index + n - l + Njp
    else
      n_occ = config(j,3)       ! occupation number for the non-relat. case
      elec(i+1:i+n_occ) = N_r * l + n - l
    end if
    i = i + n_occ
    if ( i > N_elec ) stop "ERROR in config_to_elec()"
  end do
  if ( i < N_elec ) elec(i+1: ) = 0
  !
  ! sort elec(:) in a descending order by shifting
  !
  do i = 2, size(elec)
    temp = elec(i)
    loop_j: do j = i-1, 1, -1
      if ( elec(j) >= temp ) exit loop_j
      elec(j+1) = elec(j)               ! make a shift
      elec(j) = temp
    end do loop_j
  end do
end subroutine config_to_elec

function construct_config_from_elec( elec, N_r ) result ( config )
  integer, intent(in) :: elec(:)
  integer, intent(in) :: N_r
  integer, allocatable :: config(:,:)
  integer :: N_orb, N_elec
  integer :: i, j, k, n, l, temp
  N_elec = size( elec )
  !
  ! determine N_orb
  !
  N_orb = 0
  temp = 0
  do i = 1, N_elec
    if ( temp /= elec(i) .and. elec(i) /= 0 ) N_orb = N_orb + 1
    temp = elec(i)
  end do
  allocate( config(N_orb,3) )
  !
  ! assign config(:,1 & 2)
  !
  j = 0         ! orbital index
  temp = 0
  do i = 1, N_elec
    if ( elec(i) == 0 ) exit
    if ( temp /= elec(i) .and. elec(i) /= 0 ) then
      j = j + 1
      l = int( elec(i) / N_r )
      k = elec(i) - N_r * l
      n = k + l
      config(j,1) = n
      config(j,2) = l
    end if
    temp = elec(i)
  end do
  !
  ! assign config(:,3)
  !
  call elec_to_config( elec, config, N_r )
end function construct_config_from_elec

function construct_elec_from_config( config, N_r ) result ( elec )
  integer, intent(in) :: config(:,:)
  integer, intent(in) :: N_r
  integer, allocatable :: elec(:)
  integer :: N_elec
  N_elec = sum( config(:,3) )   ! sum of occupation numbers
  allocate( elec(N_elec) )
  call config_to_elec( config, elec, N_r )
end function construct_elec_from_config

subroutine find_N_occ( elec, N_occ )
  integer, intent(in) :: elec(:)
  integer, allocatable, intent(out) :: N_occ(:)
  integer :: i, j
  if ( size(elec) == 0 ) return
  j = 1
  do i = 1, size(elec) - 1
    if ( elec(i) == 0 ) exit
    if ( elec(i) /= elec(i+1) ) j = j + 1
  end do
  if ( elec(i) == 0 ) j = j - 1
  allocate( N_occ(j) )
  N_occ(:) = 0
  j = 1
  do i = 1, size(elec) - 1
    if ( elec(i) == 0 ) exit
    N_occ(j) = N_occ(j) + 1
    if ( elec(i) /= elec(i+1) ) j = j + 1
  end do
  if ( elec(i) /= 0 ) N_occ(j) = N_occ(j) + 1
end subroutine find_N_occ

!----------------------------------------------------------------------------
! Print out a more readable form or read in elec. configuration from a string
!
! WARNING: be careful with occ_index = size(config,dim=2)
!          it's because config(:,:) can be either P%init_config or F%config
!          that has size(config,dim=2)=5 rather than 4
!----------------------------------------------------------------------------
function config_to_string( config, yes_simple ) result (s)
  character(len=2000) :: s
  integer, intent(in) :: config(:,:)
  logical, intent(in), optional :: yes_simple
  character(len=5) :: t
  character(len=7) :: each
  integer :: i, j, l, occ_n
  integer :: occ_index
  logical :: yes_relativity
  if ( size(config,dim=2) == 3 ) then   ! non-relativistic case
    yes_relativity = .false.
    occ_index = 3
  else                                  ! relativistic case
    yes_relativity = .true.
    occ_index = 4
  end if
  j = 1
  s = ""
  i = 0
  do
    i = i + 1
    occ_n = config(i,occ_index)
    if ( present(yes_simple) ) then
      if ( yes_simple .and. occ_n == 0 ) cycle          ! skip no vacancy
    end if
    t = orbital_name( config(i,1), config(i,2) )
    if ( occ_n >= 10 ) then
      write(each, "(a5,i2)") trim(t), occ_n
    else
      write(each, "(a5,i1)") trim(t), occ_n
    end if
    l = len( trim( adjustl(each) ) )
    s(j:j+l-1) = adjustl(each)
    j = j + l
    if ( yes_relativity ) then  ! for the relativistic case, check out higher j
      if ( config(i,2) > 0 ) then       ! j splitting if l/=0
        i = i + 1                       ! check out higher j
        occ_n = config(i,occ_index)
        if ( occ_n >= 10 ) then
          write(each, "(',',i2)") occ_n
        else
          write(each, "(',',i1)") occ_n
        end if
        l = len( trim( adjustl(each) ) )
        s(j:j+l-1) = adjustl(each)
        j = j + l
      end if
    end if
    if ( i == size(config,dim=1) ) exit
    s(j:j) = "_"
    j = j + 1
  end do
end function config_to_string

!
! this is for back-compatibility... used only in shakeoff.f90
!
function config_rel_to_string( config_rel ) result (s)
  character(len=20000) :: s
  integer, intent(in) :: config_rel(:,:)
  character(len=5) :: t
  character(len=7) :: each
  integer :: i, j, l, occ_n
  j = 1
  s = ""
  i = 0
  do
    i = i + 1
    t = orbital_name( config_rel(i,1), config_rel(i,2) )
    !
    ! check out lower j (or s orbital)
    !
    occ_n = config_rel(i,4)
    if ( occ_n >= 10 ) then
      write(each, "(a5,i2)") trim(t), occ_n
    else
      write(each, "(a5,i1)") trim(t), occ_n
    end if
    l = len( trim( adjustl(each) ) )
    s(j:j+l-1) = adjustl(each)
    j = j + l
    if ( config_rel(i,2) > 0 ) then ! j splitting if l/=0
      i = i + 1         ! check out higher j
      occ_n = config_rel(i,4)
      if ( occ_n >= 10 ) then
        write(each, "(',',i2)") occ_n
      else
        write(each, "(',',i1)") occ_n
      end if
      l = len( trim( adjustl(each) ) )
      s(j:j+l-1) = adjustl(each)
      j = j + l
    end if
    if ( i == size(config_rel,dim=1) ) exit
    s(j:j) = "_"
    j = j + 1
  end do
end function config_rel_to_string

!function atomic_elec_to_string( elec ) result ( s )
!  character(len=200) :: s
!  integer, intent(in) :: elec(:)
!  integer, allocatable :: config(:,:)
!  config(:,:) = construct_config_from_elec( elec )
!  s = trim( config_to_string( config ) )
!end function atomic_elec_to_string

!function relat_atomic_elec_to_string( elec ) result ( s )
!  character(len=200) :: s
!  type(Electron), intent(in) :: elec(:)
!end function relat_atomic_elec_to_string

!function molecular_elec_to_string( elec ) result ( s )
!  character(len=200) :: s
!  type(Electron), intent(in) :: elec(:)
!end function molecular_elec_to_string

!----------------------------------------------------------------------------
! subroutines related to config
!   config_to_index( config ): return an index corresponding to the config
!   index_to_config( config ): return a config corresponding to the index
!                              n and l should be already defined in config
!
! config(i,:)
!   config(i,1) -> quantum number n
!   config(i,2) -> quantum number l
!   config(i,3) -> occupation number    for the non-relativistic case
!
!   config(i,1) -> quantum number n
!   config(i,2) -> quantum number l
!   config(i,3) -> quantum number j*2
!   config(i,4) -> occupation number    for the relativistic case
!
! now these F's can handle both non-rel and rel cases.
! you can easily check whether it's relativistic or not by using
!   size(config,dim=2) == 3 => non-relativistic
!   size(config,dim=2) == 4 => relativistic
! WARNING: for P%init_config_rel and F%config_rel, this size becomes 5!!!
!   size(P%init_config_rel,dim=2) == 5
!   size(F%config_rel,dim=2) == 5
!
! how to convert config <-> index
!   for example, C+: 1s1 2s2 2p2 -> l1 = 0, l2 = 0, l3 = 1
!                                   p1 = 1, p2 = 2, p3 = 2 (occ#)
!   index = 1 + p1 + (4l1+3) [ p2 + (4l2+3) p3 ] = 26
!               |      |       |      |     |
!               1      0       2      0     2
!
! For C+: 1s1 2s2 2p2, it can be written as "1 2 2".
!   0 0 0 -> 1          1 0 0 -> 2              2 0 0 -> 3
!   0 1 0 -> 4          1 1 0 -> 5              2 1 0 -> 6
!   0 2 0 -> 7          1 2 0 -> 8              2 2 0 -> 9
!   0 0 1 -> 10         1 0 1 -> 11             2 0 1 -> 12
!   0 1 1 -> 13         1 1 1 -> 14             2 1 1 -> 15
!   0 2 1 -> 16         1 2 1 -> 17             2 2 1 -> 18
!   0 0 2 -> 19         1 0 2 -> 20             2 0 2 -> 21
!   0 1 2 -> 22         1 1 2 -> 23             2 1 2 -> 24
!   0 2 2 -> 25         1 2 2 -> 26             2 2 2 -> 27
!----------------------------------------------------------------------------
subroutine config_to_index( config, i )
  integer, intent(in) :: config(:,:)
  integer*8, intent(out) :: i
  integer :: j, jj, l, occ_n, N_orb
  i = 0
  N_orb = size(config,1)
  if ( size(config,dim=2) == 3 ) then   ! non-relativistic case
    do j = N_orb, 1, -1
      l = config(j,2)
      occ_n = config(j,3)
      i = ( 4 * l + 3 ) * i + occ_n
    end do
  else                                  ! relativistic case
    do j = N_orb, 1, -1
      jj = config(j,3) ! note: 2*jj is stored
      occ_n = config(j,4)
      i = ( jj + 2 ) * i + occ_n
    end do
  end if
  i = i + 1
end subroutine config_to_index

!
! this is for back-compatibility... used only in shakeoff.f90
!
subroutine config_rel_to_index( config_rel, i )
  integer, intent(in) :: config_rel(:,:)
  integer*8, intent(out) :: i
  integer :: j, jj, occ_n, N_orb
  i = 0
  N_orb = size(config_rel,1)
  do j = N_orb, 1, -1
    jj = config_rel(j,3) ! note: 2*jj is stored
    occ_n = config_rel(j,4)
    i = ( jj + 2 ) * i + occ_n
  end do
  i = i + 1
end subroutine config_rel_to_index

subroutine config_to_index4( config, i )
  integer, intent(in) :: config(:,:)
  integer, intent(out) :: i
  integer*8 :: ind
  call config_to_index( config, ind )
  i = int( ind, kind(i) )       ! conversion from integer8 to integer4
end subroutine config_to_index4

subroutine index_to_config( i, config )
  integer*8, intent(in) :: i
  integer, intent(inout) :: config(:,:)
  integer :: j, l, occ_n, N_orb
  integer*8 :: k
  k = i - 1
  N_orb = size(config,dim=1)
  config(:,3) = 0
  do j = 1, N_orb
    l = config(j,2)             ! config(:,2) should be already defined
    if ( k <= 0 ) exit
    occ_n = int( mod( k, 4 * l + 3 ), kind(occ_n) )  ! conversion int8 -> int4
    config(j,3) = occ_n         ! assign occ#
    !k = nint( dble( k - occ_n ) / dble( 4 * l + 3 ) )
    k = ( k - occ_n ) / ( 4 * l + 3 )
  end do
end subroutine index_to_config

subroutine index4_to_config( i, config )
  integer, intent(in) :: i
  integer, intent(inout) :: config(:,:)
  integer*8 :: ind
  ind = i
  call index_to_config( ind, config )
end subroutine index4_to_config

!----------------------------------------------------------------------------
! subroutines related to config (only one configuration)
!   orbital_name( n, l ): print out orbital name for (n,l)
!   config_str( config ): print out elec. config. with orbital name + occ#
!
!   config(i,1) -> quantum number n
!   config(i,2) -> quantum number l
!   config(i,3) -> occupation number
!----------------------------------------------------------------------------
function orbital_name( n, l, j2 ) result (s)
  character(len=5) :: s, str
  integer, intent(in) :: n, l
  integer, intent(in), optional :: j2   ! two times of j
  character(len=1) :: t
  select case ( l )
    case ( 0 );  t = 's'
    case ( 1 );  t = 'p'
    case ( 2 );  t = 'd'
    case default;  t = achar( iachar('f')+l-3 )
  end select
  write(str, "(i4,a1)") n, t
  s = trim( adjustl( str ) )
  if ( present(j2) ) then
    if ( l == 0 ) then
      s = trim( adjustl( str ) ) // "0"
    else if ( j2 == 2*l - 1 ) then
      s = trim( adjustl( str ) ) // "-"
    else if ( j2 == 2*l + 1 ) then
      s = trim( adjustl( str ) ) // "+"
    else
      write(STDERR, *) "ERROR in orbital_name:"
      write(STDERR, *) "Inconsistency in l and j2"
      write(STDERR, *) "  l = ", l
      write(STDERR, *) "  j2 = ", j2
      stop
    end if
  end if
end function orbital_name

function orbital_name_from_config( config, i ) result (s)
  character(len=5) :: s
  integer, intent(in) :: config(:,:)
  integer, intent(in) :: i
  if ( size(config,dim=2) == 3 ) then   ! non-relativistic case
    s = orbital_name( config(i,1), config(i,2) )
  else                                  ! relativistic case
    s = orbital_name( config(i,1), config(i,2), config(i,3) )
  end if
end function orbital_name_from_config

end module elec_config
