!----------------------------------------------------------------------------
! excalib.f90
!----------------------------------------------------------------------------
! Copyright (C) 2021 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2021 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! EXCALIB: Extended X-ray CALIBation
!   use slopes (ion yields as a funciton of pulse energies)
!   to find out the peak fluence
! * This is a preliminary implementation of EXCALIB.
!   To get a full version, see Niels' contribution and
!   Niels Breckwoldt et al., Phys. Rev. Res. 5, 023114 (2023).
!----------------------------------------------------------------------------
program excalib
  use constants
  use stdio
  use interpolate
  use ieee_arithmetic
  implicit none
  type(Argument) :: ARG
  type(Spline), allocatable :: S(:)
  real(long) :: pulse_E
!  real(long) :: Y_avg_th, Y_avg_exp
  real(long) :: diff, xshift, yshift, dx, dy
  real(long) :: diff_min, xshift_min, yshift_min
  integer :: N_charge
  integer :: i, j, k, l, ip
  integer :: INFO, OUTPUT = STDOUT
  logical, allocatable :: yes_charge(:)
  !
  ! foobar_th: theoretical data
  ! foobar_exp: experimental data
  ! both pulse energies and ion yields are stored as a log value
  !
  character(len=255) :: file_th, file_exp, string
  integer :: N_th, N_exp, N_p
  integer :: Nx, Ny
  integer :: Nxmin, Nxmax, Nymin, Nymax
  real(long), allocatable :: pulse_E_th(:), pulse_E_exp(:)
  real(long), allocatable :: Y_th(:,:)          ! theoretical ion yields
  real(long), allocatable :: Y_int(:,:)         ! interpolated ion yields
  real(long), allocatable :: Y_exp(:,:)         ! experimental ion yields
  real(long), allocatable :: Y_exp_error(:,:)   ! experimental yields(error)
  real(long), allocatable :: pulse_E_temp(:), Y_temp(:)
  real(long), allocatable :: temp(:,:)
  real(long), allocatable :: weight(:,:)        ! weighting from Y_exp_error
  real(long) :: weight_coeff                    ! how to treat Y_exp_error
  real(long) :: norm_p                          ! normalization factor
  character(len=1000000) :: each_line           ! enforce to read line by line
  logical :: yes_scan
  integer :: IO_TH_DATA = 98, IO_EXP_DATA = 99
  !
  ! read parameters
  !   if xshift and yshift are assigned, it calculates only at (xshift,yshift).
  !   if not, it scans for a range of xshift and yshift.
  !
  yes_scan = .true.
  xshift = ZERO; yshift = ZERO
  Nx = 20;  dx = 0.1_long;  Nxmin = 0;  Nxmax = 0
  Ny = 60;  dy = 0.2_long;  Nymin = 0;  Nymax = 0
  weight_coeff = ZERO   ! zero means no consideration of Y_exp_error
  string = ""
  call get_input( ARG )
  call parse_option( ARG, "scan", yes_scan )            ! whether scan or not
  call parse_option( ARG, "N", N_charge )               ! number of charges
  call parse_option( ARG, "file_th",  file_th  )        ! theoretical data
  call parse_option( ARG, "file_exp", file_exp )        ! experimental data
  call parse_option( ARG, "Nx", Nx )                    ! (Nx+1): N of x shift
  call parse_option( ARG, "Nxmin", Nxmin )              ! from Nxmin
  call parse_option( ARG, "Nxmax", Nxmax )              ! to Nxmax
  call parse_option( ARG, "Ny", Ny )                    ! (Ny+1): N of y shift
  call parse_option( ARG, "Nymin", Nymin )              ! from Nymin
  call parse_option( ARG, "Nymax", Nymax )              ! to Nymax
  call parse_option( ARG, "dx", dx )                    ! increment in x axis
  call parse_option( ARG, "dy", dy )                    ! increment in y axis
  call parse_option( ARG, "xshift", xshift )            ! fixed value for xshift
  call parse_option( ARG, "yshift", yshift )            ! fixed value for yshift
  call parse_option( ARG, "exclude", string )           ! charges to be excluded
  call parse_option( ARG, "exclude_charge", string )
  call parse_option( ARG, "exclude_Q", string )
  call parse_option( ARG, "error", weight_coeff )       ! coeff. for weight func
  call parse_option( ARG, "weight", weight_coeff )
  if ( yes_scan ) then
    if ( Nx == 0 .or. Ny == 0 ) call error_message( "Nx = 0 or Ny = 0" )
    if ( Nxmin >= Nxmax ) Nxmax = Nxmin + Nx
    if ( Nymin >= Nymax ) then
      Nymin = -Ny / 2
      Nymax =  Ny / 2
    end if
    Nx = Nxmax - Nxmin    ! re-setup Nx
    Ny = Nymax - Nymin    ! re-setup Ny
    write(OUTPUT,"('# N of charge states = ',i5)") N_charge
    write(OUTPUT,"('# shift range (in log values):')")
    write(OUTPUT,"('# - Nx = ',i5,' (',i5,' ... ',i5,'), dx = ',f5.2)") &
      Nx, Nxmin, Nxmax, dx
    write(OUTPUT,"('# - Ny = ',i5,' (',i5,' ... ',i5,'), dy = ',f5.2,x,a)") &
      Ny, Nymin, Nymax, dy, "[OBSOLETE]"
    write(OUTPUT,"('# - shift in the x axis: ',f7.2,' to ',f7.2," // &
                 "' (',i5,' points)')") Nxmin*dx, Nxmax*dx, Nx+1
!    write(OUTPUT,"('# - shift in the y axis: ',f7.2,' to ',f7.2," // &
!                 "' (',i5,' points)')") Nymin*dy, Nymax*dy, Ny+1
!    write(OUTPUT,"('#   w.r.t. diff. between avg. log(ion_yield)')")
  else
    Nx = 1;  Nxmin = 1;  Nxmax = 1;  dx = xshift
    Ny = 1;  Nymin = 1;  Nymax = 1;  dy = yshift
  end if
  !
  ! analyze theoretical data file
  !
  open( IO_TH_DATA, file=file_th, action="read", form="formatted" )
  N_th = 0
  do
    read( IO_TH_DATA, *, iostat=INFO ) pulse_E
    if ( INFO < 0 ) exit                ! exit when it ends
    if ( pulse_E == ZERO ) cycle        ! skip unmeaningful
    N_th = N_th + 1
  end do
  !
  ! analyze theoretical data file
  !
  open( IO_EXP_DATA, file=file_exp, action="read", form="formatted" )
  N_exp = 0
  do
    read( IO_EXP_DATA, *, iostat=INFO ) pulse_E
    if ( INFO < 0 ) exit                ! exit when it ends
    if ( pulse_E == ZERO ) cycle        ! skip unmeaningful
    N_exp = N_exp + 1
  end do
  !
  ! prepare arrays
  !
  allocate( S(N_charge) )
  allocate( yes_charge(N_charge) )
  allocate( pulse_E_th(N_th), pulse_E_exp(N_exp) )
  allocate( Y_th(N_charge,N_th), Y_int(N_charge,N_th), Y_exp(N_charge,N_exp) )
  allocate( Y_exp_error(N_charge,N_exp), temp(2*N_charge,N_exp) )
  allocate( weight(N_charge,N_exp) )
  allocate( pulse_E_temp(N_th), Y_temp(N_th) )
  weight(:,:) = ONE     ! initialize with one
  !
  ! set up yes_charge(:)
  !
  yes_charge(:) = .true.
  if ( string /= "" ) call parse_string( string, yes_charge(:) )
  write(OUTPUT,"('# charge selection:')")
  write(OUTPUT,"('# ',100i4)") (/ ( j, j=1, N_charge ) /)
  write(OUTPUT,"('# ',100l4)") (/ ( yes_charge(j), j=1, N_charge ) /)
  !
  ! read theoretical data
  !
  write(OUTPUT,"('# theoretical data file:  ',a)",advance="no") trim(file_th)
  write(OUTPUT,"('  [',i4,' points]')") N_th
  rewind( IO_TH_DATA )
  i = 1
  do
    read( IO_TH_DATA, *, iostat=INFO ) pulse_E_th(i), Y_th(:,i)
    if ( INFO < 0 ) exit                ! exit when it ends
    if ( pulse_E_th(i) == ZERO ) cycle  ! skip unmeaningful
    do j = 1, N_charge
      if ( Y_th(j,i) == ZERO ) Y_th(j,i) = -ONE ! make it NaN
    end do
    pulse_E_th(i) = log10( pulse_E_th(i) )      ! take a log value
    Y_th(:,i) = log10( Y_th(:,i) )              ! take log values
    i = i + 1
  end do
!  if ( yes_scan ) then
!    Y_avg_th = avg_log_yield( Y_th )
!    write(OUTPUT,"('# - avg. log(ion_yield) = ',es13.5)") Y_avg_th
!  else
!    Y_avg_th = ZERO     ! not used if no scan is made
!  end if
!  do i = 1, N_th
!    write(OUTPUT, "(100es13.5)") pulse_E_th(i), Y_th(:,i)
!  end do
  !
  ! read experimental data
  !
  write(OUTPUT,"('# experimental data file: ',a)",advance="no") trim(file_exp)
  write(OUTPUT,"('  [',i4,' points]')") N_exp
  rewind( IO_EXP_DATA )
  i = 1
  temp(:,:) = -ONE      ! to make its log value NaN
  do
    !
    ! originally, it's written as
    !   read( IO_EXP_DATA, *, iostat=INFO ) pulse_E_exp(i), temp(:,i)
    ! however, Fotran doesn't read data line by line; instead, it tries to 
    ! read from the next line, which gives "garbage" unless the column number 
    ! of the data file is exactly matched with N_charge*2+1.
    ! To avoid this error-prone procedure, it's enforced to read "line by line"
    ! by using "each_line" (lengthy string) and read data from there.
    !
    read( IO_EXP_DATA, "(a)", iostat=INFO ) each_line
    if ( INFO < 0 ) exit                ! exit when it ends
    read( each_line, *, iostat=INFO ) pulse_E_exp(i), temp(:,i)
    if ( INFO < 0 ) then
      write(STDERR,"('# WARNING: In EXP data file, " // &
                   "line ',i4,' does not have enough data')") i
    end if
    if ( pulse_E_exp(i) == ZERO ) cycle ! skip unmeaningful
    do j = 1, N_charge
      Y_exp(j,i)       = temp(j*2-1,i)  ! one column: yield
      Y_exp_error(j,i) = temp(j*2,i)    ! the other column: error
      if ( Y_exp(j,i) == ZERO ) Y_exp(j,i) = -ONE               ! make it NaN
      if ( Y_exp_error(j,i) == ZERO ) Y_exp_error(j,i) = -ONE   ! make it NaN
    end do
    !
    ! prepare weighting factors if weight_coeff > 0
    !
    if ( weight_coeff > ZERO ) then
      weight(:,i) = exp( -weight_coeff * ( Y_exp_error(:,i) / Y_exp(:,i) )**2 )
    end if
    !
    ! take log values for all quantities
    !
    pulse_E_exp(i) = log10( pulse_E_exp(i) )
    Y_exp(:,i) = log10( Y_exp(:,i) )
    Y_exp_error(:,i) = log10( Y_exp_error(:,i) )
    i = i + 1
  end do
  if ( weight_coeff > ZERO ) then
    write(OUTPUT,"('# - EXP errors, exp(-w*[E/Y]^2): w = ',f10.2)") weight_coeff
  else
    write(OUTPUT,"('# - EXP errors are not used')")
  end if
!  if ( yes_scan ) then
!    Y_avg_exp = avg_log_yield( Y_exp )
!    write(OUTPUT,"('# - avg. log(ion_yield) = ',es13.5)") Y_avg_exp
!  else
!    Y_avg_exp = ZERO    ! not used if no scan is made
!  end if
!  do i = 1, N_exp
!    write(OUTPUT, "(100es13.5)") pulse_E_exp(i), Y_exp(:,i)
!  end do
  !
  ! prepare spline for theoretical data
  !
  do j = 1, N_charge
    pulse_E_temp(:) = ZERO
    Y_temp(:) = ZERO
    ip = 0
    do i = 1, N_th
      if ( ieee_is_nan( Y_th(j,i) ) ) cycle     ! skip if Y_th is NaN
      ip = ip + 1
      pulse_E_temp(ip) = pulse_E_th(i)
      Y_temp(ip) = Y_th(j,i)                    ! store Y_th(j,:) except NaN
    end do
    if ( ip == 0 ) then         ! no available data -> interpolating with zeros
      ip = N_th
      pulse_E_temp(1:ip) = pulse_E_th(1:ip)
    end if
    S(j) = Spline_create( pulse_E_temp(1:ip), Y_temp(1:ip) )
  end do
  !
  ! scan pulse energy
  !
  ! yshift is analytically solved from the quadratic function
  ! WARNING: don't forget the weight function
  !   cost function = sum_{selected p} | log Y_exp^p - log Y_th^p - yshift |^2
  !                                    * weight_p
  ! here, we want to find yshift (=log dy) to minimize the cost function
  !             sum_{selected p} ( log Y_exp^p - log Y_th^p ) * weight_p
  ! => yshift = --------------------------------------------------------
  !                       sum_{selected p} weight_p
  !
  write(OUTPUT,"('# - xshift: scanned')")
  write(OUTPUT,"('# - yshift: analytically solved for each xshift')")
  write(OUTPUT,"('#',a12,2a13)") "log_xshift", "log_yshift", "diff"
  diff_min = huge(ZERO)         ! set up the largest number in double precision
  do k = Nxmin, Nxmax           ! x axis
    xshift = dble(k) * dx
    yshift = ZERO
    N_p = 0
    norm_p = ZERO
    do i = 1, N_exp
      pulse_E = pulse_E_exp(i)
      do j = 1, N_charge
        Y_int(j,i) = Spline_interpolate( S(j), pulse_E - xshift )
        if ( .not. yes_charge(j) ) cycle
        if ( ieee_is_nan( Y_exp(j,i) ) ) cycle  ! skip if no exp. data available
        N_p = N_p + 1
        norm_p = norm_p + weight(j,i)
        yshift = yshift + ( Y_exp(j,i) - Y_int(j,i) ) * weight(j,i)
      end do
    end do
    !yshift = yshift / dble(N_p)
    yshift = yshift / norm_p
    !
    ! we don't need to scan dy because we have an analytic solution!
    !
    !do l = Nymin, Nymax         ! yaxis
    !  yshift = ( Y_avg_exp - Y_avg_th ) + dble(l) * dy
    !  diff = ion_yield_difference( Y_exp, Y_int, yshift, yes_charge )
    !  write(OUTPUT,"(3es13.5)") xshift, yshift, diff
    !  if ( diff < diff_min ) then
    !    diff_min = diff
    !    xshift_min = xshift
    !    yshift_min = yshift
    !  end if
    !end do
    diff = ion_yield_difference( Y_exp, Y_int, yshift, yes_charge, weight )
    if ( diff < diff_min ) then
      diff_min = diff
      xshift_min = xshift
      yshift_min = yshift
    end if
    write(OUTPUT,"(3es13.5)") xshift, yshift, diff
    !write(OUTPUT,"()")
  end do
  !
  ! print the global minimum (just one value if no scan is made)
  !
  if ( yes_scan ) then
    write(OUTPUT,"('# minimum diff = ',es13.5)") diff_min
  else
    write(OUTPUT,"('# diff = ',es13.5)") diff
  end if
  write(OUTPUT,"('# xshift in log = ',es13.5,' -> mx (D^2/T) = ',es13.5)") &
    xshift_min, 10**xshift_min
  write(OUTPUT,"('# yshift in log = ',es13.5,' -> my = ',es13.5)") &
    yshift_min, 10**yshift_min
  !
  ! print out warning messages when shift is at the boundary of the range
  !
  if ( yes_scan ) then
    k = Nxmin; xshift = dble(k) * dx
    if ( xshift_min == xshift ) &
      write(OUTPUT,"('# WARNING: xshift is equal to min. of the range.')")
    k = Nxmax; xshift = dble(k) * dx
    if ( xshift_min == xshift ) &
      write(OUTPUT,"('# WARNING: xshift is equal to max. of the range.')")
!    l = Nymin; yshift = ( Y_avg_exp - Y_avg_th ) + dble(l) * dy
!    if ( yshift_min == yshift ) &
!      write(OUTPUT,"('# WARNING: yshift is equal to min. of the range.')")
!    l = Nymax; yshift = ( Y_avg_exp - Y_avg_th ) + dble(l) * dy
!    if ( yshift_min == yshift ) &
!      write(OUTPUT,"('# WARNING: yshift is equal to max. of the range.')")
  end if
  !
  ! print out interpolated data
  !
!  do i = 1, N_exp
!    pulse_E = pulse_E_exp(i)
!    do j = 1, N_charge
!      Y_int(j,i) = Spline_interpolate( S(j), pulse_E - xshift_min ) + yshift_min
!    end do
!  end do
!  do j = 1, N_charge
!    write(100,"('# Q = ',i5)") j
!    write(101,"('# Q = ',i5)") j
!    do i = 1, N_th
!      write(100,"(2es13.5)") pulse_E_th(i) + xshift_min, Y_th(j,i) + yshift_min
!    end do
!    do i = 1, N_exp
!      write(101,"(3es13.5)") pulse_E_exp(i), Y_int(j,i), Y_exp(j,i)
!    end do
!    write(100,"(/)")
!    write(101,"(/)")
!  end do
contains

!function avg_log_yield( Y ) result (Y_avg)
!  real(long) :: Y_avg
!  real(long), intent(in) :: Y(:,:)
!  integer :: i, j, N_count
!  Y_avg = ZERO
!  N_count = 0
!  do i = 1, size(Y,2)
!    do j = 1, size(Y,1)
!      if ( .not. ieee_is_nan( Y(j,i) ) ) then
!        N_count = N_count + 1
!        Y_avg = Y_avg + Y(j,i)
!      end if
!    end do
!  end do
!  Y_avg = Y_avg / dble(N_count)
!end function avg_log_yield

function ion_yield_difference( Y1, Y2, yshift, yes_charge, weight ) result (D)
  real(long) :: D
  real(long), intent(in) :: Y1(:,:)     ! experimental data
  real(long), intent(in) :: Y2(:,:)     ! interpolated data
  real(long), intent(in) :: yshift
  logical, intent(in) :: yes_charge(:)
  real(long), intent(in) :: weight(:,:)     ! weighting factors
  integer :: i, j
  D = ZERO
  do i = 1, size(Y1,2)          ! pulse energy grid
    do j = 1, size(Y1,1)        ! charge
      if ( .not. yes_charge(j) ) cycle
      if ( ieee_is_nan( Y1(j,i) ) ) cycle  ! skip if no exp. data available
      !
      ! here, we need to consider ion yield errors and/or 
      !                           pulse energy distribution
      !
      D = D + ( Y1(j,i) - ( Y2(j,i) + yshift ) )**2 * weight(j,i)
    end do
  end do
end function ion_yield_difference

subroutine parse_string( string, yes_charge )
  character(len=*), intent(in) :: string
  logical, intent(inout) :: yes_charge(:)
  integer :: i, i_max, j, Q, Q_max
  Q_max = size(yes_charge)
  i_max = len(trim(string))
  j = 1
  do i = 1, i_max
    if ( string(i:i) == "," .or. i == i_max ) then
      read( string(j:i), * ) Q
      if ( Q >= 1 .and. Q <= Q_max ) yes_charge(Q) = .false.
      j = i + 1
    end if
  end do
end subroutine parse_string

end program
