!----------------------------------------------------------------------------
! functional_xc.f90
!----------------------------------------------------------------------------
! Copyright (C) 2004 Sang-Kil Son
! Copyright (C) 2010 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! DFT calculation for the atomic system using the GPS discretization
!   written by Son, Sang-Kil in Nov. 2004
!   modified by Son, Sang-Kil in Jul. 2010 implementing more functionals
!
! This module contains potential vectors and energy functional 
! for exchange-correlation parts.
! You may check supported method in readin.f90
! - exchange: LDA, LSDA, Xa, LB, LBa
! - correlation: VWN, PW
!
! Ref) Parr and Yang, "Density-Functional Theory of Atoms and Molecules"
!      Tong and Chu, PRA 55, 3406 (1997)
!      Other references for each functional are listed for each comment.
!----------------------------------------------------------------------------
module functional_xc
  use constants
  use coord_system
  use operations
  use read_param
  implicit none
  interface potential_c
    module procedure potential_c_nospin, potential_c
  end interface
  interface functional_c
    module procedure functional_c_nospin, functional_c
  end interface
contains

!----------------------------------------------------------------------
! General XC functionals
!   potential_x or functional_x( P, G, rho )
!   potential_c or functional_c( P, G, rho_a, rho_b )
!
!   You can easily call these functions with Param P.
!   These fork into appropriate potential or functional according to P,
!   i.e. the type of exchange/correlation functional.
!
! Note: in fact, these potentials and functionals must depend on only rho.
!   However, some functionals need wavefunctions instead of rho,
!   because of the Del rho term.  Of course, Del rho is also dependent
!   on only rho, but I believe there was some numerical errors 
!   when I computed Del rho with only rho.
!   So, these functions require wf as well as rho.
!----------------------------------------------------------------------
function potential_xc( P, G, rho, D_rho ) result (pot)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long), intent(in), optional :: D_rho(:)
  real(long) :: pot(size(rho))
  if ( present(D_rho) ) then
    pot(:) = potential_x( P, G, rho, D_rho ) + potential_c( P, G, rho, D_rho )
  else
    pot(:) = potential_x( P, G, rho ) + potential_c( P, G, rho )
  end if
end function potential_xc

function potential_x( P, G, rho, D_rho ) result (pot)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long), intent(in), optional :: D_rho(:)
  real(long) :: pot(size(rho))
  !
  ! for the exchange potential
  !
  select case ( FUNCTIONAL_X_NAME(P%x_type) )
    case ( "LDA" )
      pot(:) = potential_LDA( rho )
    case ( "LSDA" )
      pot(:) = potential_LSDA( rho )
    case ( "Xa" )
      pot(:) = potential_Xa( rho, P%alpha )
    case ( "Slater" )
      pot(:) = potential_Xa( rho, alpha=ONE )
!    case ( "Xab" )
!      pot(:) = potential_Xab( G, rho, P%alpha, P%beta )
    case ( "LB" )
      if ( .not. present(D_rho) ) stop "LB potential requires D_rho."
      pot(:) = potential_LB( G, rho, D_rho, P%beta )
    case ( "LBa" )
      if ( .not. present(D_rho) ) stop "LB_alpha potential requires D_rho."
      pot(:) = potential_LBalpha( G, rho, D_rho, P%alpha, P%beta )
    case default
      stop "ERROR in exchange functional name"
  end select
end function potential_x

function functional_x( P, G, rho, D_rho ) result (E)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long), intent(in), optional :: D_rho(:) 
  real(long) :: E
  !
  ! for the exchange functional
  !
  select case ( FUNCTIONAL_X_NAME(P%x_type) )
    case ( "LDA" )
      E = functional_LDA( G, rho )
    case ( "LSDA" )
      E = functional_LSDA( G, rho )
    case ( "Xa" )
      E = functional_Xa( G, rho, P%alpha )
    case ( "Slater" )
      E = functional_Xa( G, rho, alpha=ONE )
!    case ( "Xab" )
!      E = functional_Xab( G, rho, P%alpha, P%beta )
    case ( "LB" )
      if ( .not. present(D_rho) ) stop "LB potential requires D_rho."
      E = functional_LB( G, rho, D_rho, P%beta )
    case ( "LBa" )
      if ( .not. present(D_rho) ) stop "LBalpha potential requires D_rho."
      E = functional_LBalpha( G, rho, D_rho, P%alpha, P%beta )
    case default
      stop "ERROR in xc functional name"
  end select
end function functional_x

function potential_c_nospin( P, G, rho ) result (pot) ! for spin-independent
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long) :: pot(size(rho)) 
  pot(:) = potential_c( P, G, rho/TWO, rho/TWO )
end function potential_c_nospin

function functional_c_nospin( P, G, rho ) result (E)  ! for spin-independent
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long) :: E
  E = functional_c( P, G, rho/TWO, rho/TWO )
end function functional_c_nospin

function potential_c( P, G, rho_a, rho_b ) result (pot)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: pot(size(rho_a))
  integer :: N
  N = G%N
  !
  ! for the correlation potential
  !
  select case ( FUNCTIONAL_C_NAME(P%c_type) )
    case ( "none" )
      pot(:) = ZERO
    case ( "VWN" )
      pot(:) = potential_VWN( rho_a, rho_b )
    case ( "PW" )
      pot(:) = potential_PW( rho_a, rho_b )
    case default
      stop "ERROR in correlation functional name"
  end select
end function potential_c

function functional_c( P, G, rho_a, rho_b ) result (E)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: E
  !
  ! for the correlation functional
  !
  select case ( FUNCTIONAL_C_NAME(P%c_type) )
    case ( "none" )
      E = ZERO
    case ( "VWN" )
      E = functional_VWN( G, rho_a, rho_b )
    case ( "PW" )
      E = functional_PW( G, rho_a, rho_b )
  end select
end function functional_c


!----------------------------------------------------------------------
! LDA: Local Density Approximation
!   v_x_LDA(r) = - { 3/PI * rho(r) }^(1/3)
!   E_x_LDA[rho] = -3/4 * (3/PI)^(1/3) INT{ rho(r)^(4/3) }dr
! LSDA: Local Spin Density Approximation
!   v_x_LSDA(r) = 2^(1/3) * v_x_LDA(r)
!   E_x_LDA[rho] = 2^(1/3) * E_x_LDA[rho]
! Xa: Local functional
!   v_Xa(r) = -3/2 * alpha * { 3/PI * rho(r) }^(1/3)
!   E_Xa[rho] = -9/8 * alpha * (3/PI)^(1/3) INT{ rho(r)^(4/3) }dr
!
! All have ONLY exchange part for LDA.
! Xa method == LDA without correlation if alpha=2/3
! HFS == Xa with alpha=1
! SEE: Eq. (4) in Son, Young & Santra, PRA 83, 033402 (2011).
!
! Ref) Slater, PR 81, 385 (1951)
!----------------------------------------------------------------------
function potential_LDA( rho ) result (pot)
  real(long), intent(in) :: rho(:)
  real(long) :: pot(size(rho))
  pot(:) = potential_Xa( rho, alpha=TWO/THREE )
end function potential_LDA

function functional_LDA( G, rho ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long) :: E
  E = functional_Xa( G, rho, alpha=TWO/THREE )
end function functional_LDA

function potential_LSDA( rho ) result (pot)
  real(long), intent(in) :: rho(:)
  real(long) :: pot(size(rho))
  pot(:) = TWO**(ONE/THREE) * potential_Xa( rho, alpha=TWO/THREE )
end function potential_LSDA

function functional_LSDA( G, rho ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long) :: E
  E = TWO**(ONE/THREE) * functional_Xa( G, rho, alpha=TWO/THREE )
end function functional_LSDA

function potential_Xa( rho, alpha ) result (pot)
  real(long), intent(in) :: rho(:), alpha
  real(long) :: pot(size(rho))
  pot(:) = - THREE / TWO * alpha * ( THREE/PI*rho(:) )**(ONE/THREE)
end function potential_Xa

function functional_Xa( G, rho, alpha ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:), alpha
  real(long) :: E
  E = -9.0_long/8.0_long * alpha * (THREE/PI)**(ONE/THREE) * &
       integral( G, rho(:)**(FOUR/THREE) )
end function functional_Xa


!----------------------------------------------------------------------
! X_alpha_beta potential : by F. Herman and K. Schwarz
! Note that the notations are different from original Schwarz75.
! Actual forms are obtained from Bhalla et al, PRA 8, 649 (1973):
!   alpha = 2/3   for LDA
!   beta = 0.0028 for the correction part
! ref) Herman and Schwarz, in Computational Solid State Physics
!                          (Plenum, New York, 1972)
!      Schwarz, Chem. Phys. 7, 94 (1975)
!----------------------------------------------------------------------
!function potential_Xab( G, rho, alpha, beta ) result (pot)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: rho(:), alpha, beta
!  real(long) :: pot(size(rho))
!  real(long) :: GG(G%N), D_rho(G%N), D2_rho(G%N), v(G%N)
!  !
!  ! compute Gradient of rho
!  ! remember this is not accurate for very small rho (for long-range)
!  !
!  v(:) = transform_wf_to_A( G, rho(:) )
!  v(:) = matmul( G%Del_r(:,:), v(:) ) - v(:) / G%r(:)
!  D_rho(:) = transform_A_to_wf( G, v(:) )
!  !
!  ! compute Laplacian of rho
!  !
!  v(:) = transform_wf_to_A( G, rho(:) )
!  v(:) = matmul( G%Del2(:,:), v(:) )
!  D2_rho(:) = transform_A_to_wf( G, v(:) )
!  !
!  ! compute a dimensionless inhomogeneity correction
!  !
!  GG(:) = beta * ( TWO * (D_rho(:)/rho(:))**2 - THREE * D2_rho(:) / rho(:) ) / &
!          rho(:)**(TWO/THREE)
!  pot(:) = potential_Xa( rho, alpha ) * ( ONE + tanh(GG(:)) )
!end function potential_Xab
!
!function functional_Xab( G, rho, alpha, beta ) result (E)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: rho(:), alpha, beta
!  real(long) :: E
!  !
!  ! I'M NOT SURE OF THIS EQUATION!!!
!  ! Just use the simplest form
!  !
!  E = THREE/FOUR * integral( G, potential_Xab( G, rho, alpha, beta )*rho(:) )
!end function functional_Xab


!----------------------------------------------------------------------
! LB model potential : R. van Leeuwen and E. J. Baerends
!   It contains a beta parameter.  Same as LBalpha with alpha=1
!   ref) Leeuwen and Baerends, PRA 49, 2421 (1994)
!
! improved LB ( LBalpha )
!   It contains the alpha and beta empirical parameters.
!   ref) Schipper et al, JCP 112, 1344 (2000)
!
!                                        - beta x^2
! v_LBa = alpha * v_LDA + rho^(1/3) -----------------------
!                                   1 + 3 beta x sinh^-1(x)
!                         where x = | Del rho | / rho^(4/3)
!----------------------------------------------------------------------
function potential_LB( G, rho, D_rho, beta ) result (pot)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:), D_rho(:), beta
  real(long) :: pot(size(rho))
  pot(:) = potential_LBalpha( G, rho, D_rho, alpha=ONE, beta=beta )
end function potential_LB

function functional_LB( G, rho, D_rho, beta ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:), D_rho(:), beta
  real(long) :: E
  E = functional_LBalpha( G, rho, D_rho, alpha=ONE, beta=beta )
end function functional_LB

function potential_LBalpha( G, rho, D_rho, alpha, beta ) result (pot)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:), D_rho(:), alpha, beta
  real(long) :: pot(size(rho))
  integer :: N
  N = G%N
  pot(:) = alpha * potential_LSDA( rho ) + rho(:)**(ONE/THREE) * &
                                           f_LB( rho, D_rho, beta )
  !
  ! attach a long-range Coulomb tail to the LBa potential
  !
!  do i = 1, V%N
!    if ( V%r_c(i) > r0 .and. V%r_c(i) < r_long ) then
!      ratio = ONE / ( ONE + exp( -5.0_long * ( V%r_c(i) - (r0+r_long)/TWO ) ) )
!      pot(i) = pot(i) * ( ONE - ratio ) + ( - ONE / V%r_c(i) ) * ratio
!    else if ( V%r_c(i) >= r_long ) then
!      pot(i) = - ONE / V%r_c(i)
!    end if
!  end do
end function potential_LBalpha

function functional_LBalpha( G, rho, D_rho, alpha, beta ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:), D_rho(:), alpha, beta
  real(long) :: E
  !
  ! I'M NOT SURE OF THIS EQUATION!!!
  !
  E = alpha * functional_LSDA( G, rho ) &
    + THREE/FOUR * integral( G, f_LB( rho, D_rho, beta )*rho(:)**(FOUR/THREE) )
end function functional_LBalpha

function f_LB( rho, D_rho, beta ) result (f)
  real(long), intent(in) :: rho(:), D_rho(:)
  real(long), intent(in) :: beta
  real(long) :: f(size(rho))
  real(long) :: x(size(rho))
  x(:) = D_rho(:) / rho(:)**(FOUR/THREE)
  !
  ! sinh^-1(x) == arcsinh(x) == log( x + sqrt(x^2 + 1 ))
  !
  f(:) = - x(:) / ( ONE / beta/x(:) + THREE * log( x(:) + sqrt(x(:)**2+ONE) ) )
end function f_LB


!----------------------------------------------------------------------
! VWN correlation functional: the 5th formula of VWN
!   SVWN (Slater+VWN in chemistry) is equal to LDA (in physics).
!   ref) Vosko, Wilk, and Nusair, Can. J. Phys. 58, 1200 (1980)
!      * Johnson, Gill, and Pople, J. Chem. Phys. 98, 5612 (1993)
!----------------------------------------------------------------------
function potential_VWN( rho_a, rho_b ) result (pot)
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: pot(size(rho_a))
  real(long) :: rho, z, x, dz, E_c, dE_c
  real(long) :: h, hp, gg, ggp, ep_F, ep_P, ep_A, ep_Fp, ep_Pp, ep_Ap
  integer :: i
  pot(:) = ZERO
  do i = 1, size(rho_a)
    rho = rho_a(i) + rho_b(i)
    if ( rho < EPS ) cycle              ! pot = 0 if rho is very small
    z = ( rho_a(i) - rho_b(i) ) / rho
    dz = ( ONE - z ) / rho              ! only for d z / d rho_a
    x = ( THREE / FOUR / PI / rho )**(ONE/6.0_long)
    !
    ! compute VWN functions
    !
    ep_F = VWN_EP( x, 0 );      ep_Fp = VWN_EPp( x, 0 )
    ep_P = VWN_EP( x, 1 );      ep_Pp = VWN_EPp( x, 1 )
    ep_A = VWN_EP( x, 2 );      ep_Ap = VWN_EPp( x, 2 )
    h  = FOUR/9.0_long/(TWO**(ONE/THREE)-ONE) * ( ep_F - ep_P ) / ep_A - ONE
    hp = FOUR/9.0_long/(TWO**(ONE/THREE)-ONE) / ep_A * &
         ( ep_Fp - ep_Pp - ( ep_F - ep_P ) / ep_A * ep_Ap )
    gg  = 9.0_long / 8.0_long * &
          ( ( ONE + z )**(FOUR/THREE) + ( ONE - z )**(FOUR/THREE) - TWO )
    ggp = THREE / TWO * ( ( ONE + z )**(ONE/THREE) - ( ONE - z )**(ONE/THREE) )
    !
    ! compute potential values
    !
    E_c = ep_P + ep_A * gg * ( ONE + h * z**FOUR )
    dE_c = - x / 6.0_long / rho * &
             (   ep_Pp + ep_Ap * gg * ( ONE + h * z**FOUR ) &
               + ep_A * gg * hp * z**FOUR ) &
           + ep_A * (   ggp * ( ONE + h * z**FOUR ) &
                      + FOUR * gg * h * z**THREE ) * dz
    pot(i) = E_c + rho * dE_c
  end do
end function potential_VWN

function functional_VWN( G, rho_a, rho_b ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: E
  real(long) :: E_VWN(size(rho_a))
  real(long) :: rho, z, x
  real(long) :: h, gg, ep_F, ep_P, ep_A
  integer :: i
  E_VWN(:) = ZERO
  do i = 1, size(rho_a)
    rho = rho_a(i) + rho_b(i)
    if ( rho < EPS ) cycle              ! skip if rho is very small
    z = ( rho_a(i) - rho_b(i) ) / rho
    x = ( THREE / FOUR / PI / rho )**(ONE/6.0_long)
    !
    ! compute VWN functions
    !
    ep_F = VWN_EP( x, 0 )
    ep_P = VWN_EP( x, 1 )
    ep_A = VWN_EP( x, 2 )
    h  = FOUR/9.0_long/(TWO**(ONE/THREE)-ONE) * ( ep_F - ep_P ) / ep_A - ONE
    gg = 9.0_long / 8.0_long * &
         ( ( ONE + z )**(FOUR/THREE) + ( ONE - z )**(FOUR/THREE) - TWO )
    !
    ! compute potential values
    !
    E_VWN(i) = rho * ( ep_P + ep_A * gg * ( ONE + h * z**FOUR ) )
  end do
  E = integral( G, E_VWN(:) )
end function functional_VWN

function VWN_EP( x, f_type ) result (EP)
  real(long), intent(in) :: x
  integer, intent(in) :: f_type
  real(long) :: EP
  real(long) :: A, b, c, x0, Q
  select case ( f_type )
    case ( 0 )  ! EP_F
      A = 0.01553535_long;              b = 7.06042_long
      c = 18.0578_long;                 x0 = -0.32500_long
    case ( 1 )  ! EP_P
      A = 0.0310907_long;               b = 3.72744_long
      c = 12.9352_long;                 x0 = -0.10498_long
    case ( 2 )  ! EP_A
      A = -ONE/(6.0_long*PI**2);        b = 1.13107_long
      c = 13.0045_long;                 x0 = -0.00475840_long
  end select
  Q = sqrt( FOUR * c - b**2 )
  EP = A * (   log( x**2 / (x**2+b*x+c) ) &
             + TWO * b / Q * atan( Q / ( TWO * x + b ) ) &
             - b * x0 / (x0**2+b*x0+c) * &
               (   log( ( x - x0 )**2 / (x**2+b*x+c) ) &
                 + TWO * ( TWO * x0 + b ) / Q * atan( Q / ( TWO * x + b ) ) ) )
end function VWN_EP

function VWN_EPp( x, f_type ) result (EPp)
  real(long), intent(in) :: x
  integer, intent(in) :: f_type
  real(long) :: EPp
  real(long) :: A, b, c, x0, Q
  select case ( f_type )
    case ( 0 )  ! EP_F
      A = 0.01553535_long;              b = 7.06042_long
      c = 18.0578_long;                 x0 = -0.32500_long
    case ( 1 )  ! EP_P
      A = 0.0310907_long;               b = 3.72744_long
      c = 12.9352_long;                 x0 = -0.10498_long
    case ( 2 )  ! EP_A
      A = -ONE/(6.0_long*PI**2);        b = 1.13107_long
      c = 13.0045_long;                 x0 = -0.00475840_long
  end select
  Q = sqrt( FOUR * c - b**2 )
  EPp = A * (   TWO / x - ( TWO * x + b ) / (x**2+b*x+c) &
              - FOUR * b / ( ( TWO * x + b )**2 + Q**2 ) &
              - b * x0 / (x0**2+b*x0+c) * &
                (   TWO / ( x - x0 ) - ( TWO * x + b ) / (x**2+b*x+c) &
                  - FOUR * ( TWO * x0 + b ) / ( ( TWO * x + b )**2 + Q**2 ) ) )
end function VWN_EPp


!----------------------------------------------------------------------
! PW correlation functional: LDA correlation; it doesn't require GGA.
! It's modified from Official PBE correlation code. K. Burke, May 14, 1996.
!   ref) Perdew and Wang, PRB 45, 13244 (1992)
!----------------------------------------------------------------------
function potential_PW( rho_a, rho_b ) result (pot)
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: pot(size(rho_a))
  real(long), parameter :: R13 = (ONE/THREE)
  real(long), parameter :: R43 = (FOUR/THREE)
  real(long), parameter :: GAM = TWO**R43 - TWO
  real(long), parameter :: FZZ = 8.0_long / ( 9.0_long * GAM )
  real(long) :: rho, ZET, RS, RTRS, EU, EURS, EP, EPRS, ALFM, ALFRSM
  real(long) :: ALFC, Z4, FS, EC, ECRS, FZ, ECZET, COMM, v_up, v_down
  integer :: i
  pot(:) = ZERO
  do i = 1, size(rho_a)
    rho = rho_a(i) + rho_b(i)
    if ( rho < EPS ) cycle              ! pot = 0 if rho is very small
    ZET = ( rho_a(i) - rho_b(i) ) / rho
    RS = (0.75D0/PI/rho)**R13
    RTRS = sqrt(RS)
    call PW_gcor2( 0.0310907D0,0.21370D0,7.5957D0,3.5876D0, &
                   1.6382D0,0.49294D0,RTRS,EU,EURS )
    CALL PW_gcor2( 0.01554535D0,0.20548D0,14.1189D0,6.1977D0, &
                   3.3662D0,0.62517D0,RTRS,EP,EPRS )
    CALL PW_gcor2( 0.0168869D0,0.11125D0,10.357D0,3.6231D0, &
                   0.88026D0,0.49671D0,RTRS,ALFM,ALFRSM )
    ALFC = -ALFM
    Z4 = ZET**4
    FS = ((1.D0+ZET)**R43+(1.D0-ZET)**R43-2.D0)/GAM
    EC = EU*(1.D0-FS*Z4)+EP*FS*Z4-ALFM*FS*(1.D0-Z4)/FZZ
    !
    ! compute potential values
    !
    ECRS = EURS*(1.D0-FS*Z4)+EPRS*FS*Z4 - ALFRSM*FS*(1.D0-Z4)/FZZ
    FZ = R43*((1.D0+ZET)**R13-(1.D0-ZET)**R13)/GAM
    ECZET = 4.D0*(ZET**3)*FS*(EP-EU+ALFM/FZZ) &
          + FZ*(Z4*EP-Z4*EU-(1.D0-Z4)*ALFM/FZZ)
    COMM = EC-RS*ECRS/3.D0-ZET*ECZET
    v_up   = COMM + ECZET
    v_down = COMM - ECZET
    pot(i) = v_up       ! only for spin-independent case!!!!
  end do
end function potential_PW

function functional_PW( G, rho_a, rho_b ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho_a(:), rho_b(:)
  real(long) :: E
  real(long) :: E_PW(size(rho_a))
  real(long), parameter :: R13 = (ONE/THREE)
  real(long), parameter :: R43 = (FOUR/THREE)
  real(long), parameter :: GAM = TWO**R43 - TWO
  real(long), parameter :: FZZ = 8.0_long / ( 9.0_long * GAM )
  real(long) :: rho, ZET, RS, RTRS, EU, EURS, EP, EPRS, ALFM, ALFRSM
  real(long) :: ALFC, Z4, FS, EC, ECRS, FZ, ECZET, COMM, v_up, v_down
  integer :: i
  E_PW(:) = ZERO
  do i = 1, size(rho_a)
    rho = rho_a(i) + rho_b(i)
    if ( rho < EPS ) cycle              ! pot = 0 if rho is very small
    ZET = ( rho_a(i) - rho_b(i) ) / rho
    RS = (0.75D0/PI/rho)**R13
    RTRS = sqrt(RS)
    call PW_gcor2( 0.0310907D0,0.21370D0,7.5957D0,3.5876D0, &
                   1.6382D0,0.49294D0,RTRS,EU,EURS )
    CALL PW_gcor2( 0.01554535D0,0.20548D0,14.1189D0,6.1977D0, &
                   3.3662D0,0.62517D0,RTRS,EP,EPRS )
    CALL PW_gcor2( 0.0168869D0,0.11125D0,10.357D0,3.6231D0, &
                   0.88026D0,0.49671D0,RTRS,ALFM,ALFRSM )
    ALFC = -ALFM
    Z4 = ZET**4
    FS = ((1.D0+ZET)**R43+(1.D0-ZET)**R43-2.D0)/GAM
    EC = EU*(1.D0-FS*Z4)+EP*FS*Z4-ALFM*FS*(1.D0-Z4)/FZZ
    !
    ! compute potential values
    !
    ECRS = EURS*(1.D0-FS*Z4)+EPRS*FS*Z4 - ALFRSM*FS*(1.D0-Z4)/FZZ
    FZ = R43*((1.D0+ZET)**R13-(1.D0-ZET)**R13)/GAM
    ECZET = 4.D0*(ZET**3)*FS*(EP-EU+ALFM/FZZ) &
          + FZ*(Z4*EP-Z4*EU-(1.D0-Z4)*ALFM/FZZ)
    COMM = EC-RS*ECRS/3.D0-ZET*ECZET
    v_up   = COMM + ECZET
    v_down = COMM - ECZET
    E_PW(i) = EC * rho
  end do
  E = integral( G, E_PW(:) )
end function functional_PW

subroutine PW_gcor2( A, A1, B1, B2, B3, B4, RTRS, GG, GGRS )
  real(long), intent(in) :: A, A1, B1, B2, B3, B4, RTRS
  real(long), intent(out) :: GG, GGRS
  real(long) :: Q0, Q1, Q2, Q3
  !
  ! slimmed down version of GCOR used in PW91 routines, to interpolate
  ! LSD correlation energy, as given by (10) of PRB 45, 13244 (1992)
  ! made by K. Burke, May 11, 1996.
  !
  Q0 = -2.D0*A*(1.D0+A1*RTRS*RTRS)
  Q1 = 2.D0*A*RTRS*(B1+RTRS*(B2+RTRS*(B3+B4*RTRS)))
  Q2 = DLOG(1.D0+1.D0/Q1)
  GG = Q0*Q2
  Q3 = A*(B1/RTRS+2.D0*B2+RTRS*(3.D0*B3+4.D0*B4*RTRS))
  GGRS = -2.D0*A*A1*Q2-Q0*Q3/(Q1*(1.D0+Q1))
end subroutine PW_gcor2

end module functional_xc
