!----------------------------------------------------------------------------
! gps.f90
!----------------------------------------------------------------------------
! Copyright (C) 2003-2005 Sang-Kil Son
! Copyright (C) 2010      Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Generalized Pseudospectral Method Module
!   written   by Son, Sang-Kil in Dec. 2003 - Feb. 2004
!   rewritten by Son, Sang-Kil in Feb. 2005 - May 2005, Aug. 2005
!   revised   by Son, Sang-Kil in Sep. 2005
!
! This module makes 1D grid points with the following methods:
! - Fourier Grid Hamiltonian method (FGH) : [1...N]
! - Legendre-Gauss collocation method (LG) : [1...N]
! - Legendre-Gauss-Lobatto collocation method (LGL) : [1...N]
! - extended Legendre-Gauss-Lobatto collocation method (ext-LGL) : [0...N+1]
! * Legendre-Gauss-Radau collocation method (LGR) : [1...N]
!
! Also, we can choose any mapping functions like:
! - linear mapping: r = a x + b
! - algebraic (reciprocal) mapping: r = L (1+x)/(1-x)
! - inverse-reciprocal mapping: r = Rmax - L(1-x)/(1+x+alpha)
! - arbitrary mapping: it should be defined by mapping_f and mapping_fp
!
! USAGE:
!   type(Grid) :: G
!   G = generate_Grid( "Legendre-Gauss-Lobatto", N, Rmin, Rmax, L )
!                              -- or --
!   G = generate_Grid( "Legendre-Gauss", N, Rmin, Rmax, L, "algebraic" )
!                              -- or --
!   if you want an arbitrary mapping function,
!     G = generate_Grid( "LGL", N )
!     /* set r(:) and rp(:) as you want */
!     call prepare_GPS_matrix( G )
!
! USEFUL SUBROUTINES
! - SR apply_mapping( G, Rmin, Rmax, L, mapping ): reset a mapping 
! - SR prepare_GPS_matrix( G ): re-construct D1 & D2 after changing a mapping
! - SR impose_LGL_boundary( G, boundary="both" ): impose Neumann B.C.
! - F GPS_interpolate( G, f(:), r ): get a function value at r
! - F GPS_grid_type( G ): return grid type
! - F GPS_mapping_r( G, x ): return r corresponding to x
! - F GPS_mapping_x( G, r ): return x corresponding to r
! - F GPS_mapping_rp( G, x ): return r' corresponding to x
!
! Ref) Yao & Chu, CPL 204, 381 (1993): for GPS
!      Tong & Chu, CP 217, 119 (1997)
!      Telnov & Chu, PRA 59, 2864 (1999)
!      Marston & Balint-Kurti, JCP 91, 3571 (1989)
!      Canuto, et al., Spectral Methods in Fluid Dynamics (1988)
!      Fornberg, A Practical Guide to Pseudospectral Methods (1996)
!
! NOTE: 
!    1) All quantities include r'_i for convenience.
!       "D1" is defined by d1_ij / sqrt( r'_i * r'_j ).
!       "D2" is defined by d2_ij / ( r'_i * r'_j ).
!       "weight" is defined by w_i * r'_i.
!         originally, integral := SUM( f(:) * r'(:) * w(:) )
!         but here,   integral := SUM( f(:) * weight(:) )
!       And "Deriv" is defined by (original_d1)_ij / r'_i.
!         originally, f'(i) := SUM( orig_D1(i,:) * f(:) ) / r'(i)
!         but here,   f'(i) := SUM( Deriv(i,:) * f(:) )
!
!    2) This module is designed to generate N points in (-1,1) for any method,
!       i.e. x1, x2, ..., xN in (-1,1) range.  -1 and 1 are excluded.
!       For Legendre-Gauss-Lobatto, N points come from zeros of P'_N+1(x)=0.
!       For Legendre-Gauss,         N points come from zeros of P'_N(x)=0.
!       Thus, some expressions should be modified from Canuto's book.
!       For Legendre-Gauss-Lobatto, N   in Canuto's notation --> N+1
!       For Legendre-Gauss,         N+1 in Canuto's notation --> N
!----------------------------------------------------------------------------
module gps
  use constants
  use legendre
  implicit none
  !
  ! Grid for 1D
  !   N : # of grid points, 1...N
  !   Rmin, Rmax : the range in real space
  !   x(:) : grid points in the computational space (-1:1)
  !   r(:) : grid points in the real space (Rmin:Rmax), not [ ]
  !   rp(:) : dr/dx = 1 / (dx/dr)
  !   weight(:) : weights for integration
  !   D1(:,:) : first derivatives matrix
  !   D2(:,:) : second derivatives matrix
  !   Deriv(:,:) : original 1st derivatives matrix divided by r'(i)
  !                we can compute f' directly by f'(i) = sum( Deriv(i,:)*f(:) )
  !   L : a mapping parameter for the reciprocal mapping
  !   mapping : mapping type ( 0: arbitrary, 1: linear, 2: algeb., 3: inv.rec. )
  !   grid_type : grid type ( 0: FGH, 1: LG, 2: LGL, 3: ext-LGL )
  !
  type :: Grid
    integer :: N
    real(long) :: Rmin, Rmax, L
    real(long), allocatable :: x(:), r(:), rp(:), weight(:), wf_coeff(:)
    real(long), allocatable :: Deriv(:,:)
    real(long), allocatable :: D1(:,:), D2(:,:)
    integer :: mapping, grid_type
  end type Grid
contains

!----------------------------------------------------------------------------
! generate_Grid: general function to generate grids
!   grid_type: LGL, LG, or FGH
!   mapping_type: linear or algebraic(reciprocal)
! This function does
!   1) set parameters of N and grid_type
!   2) allocate memory for r, rp, D1, D2, and so on
!   2) make computational grids
!   3) make real grids by means of a mapping function
!   4) construct D1 and D2
!----------------------------------------------------------------------------
function generate_Grid( grid_type, N, Rmin, Rmax, L, mapping ) result (G)
  type(Grid) :: G
  character(len=*) :: grid_type
  integer, intent(in) :: N
  real(long), intent(in), optional :: Rmax, Rmin
  real(long), intent(in), optional :: L
  character(len=*), intent(in), optional :: mapping
  integer :: i
  !
  ! 1) set parameters and determine a grid type
  !
  G%N = N
  G%mapping = 0
  select case ( grid_type )
    case ( "Fourier-Grid-Hamiltonian", "FGH", "Fourier", "F" ); G%grid_type = 0
    case ( "Legendre-Gauss", "LG", "Gauss", "Gauss-Legendre" ); G%grid_type = 1
    case ( "Legendre-Gauss-Lobatto", "LGL", "Gauss-Lobatto" );  G%grid_type = 2
    case ( "extended-LGL", "extended_LGL", "ext-LGL", "ext_LGL", &
           "eLGL", "extLGL", "extended-Gauss-Lobatto" );        G%grid_type = 3
    case default
      write(STDERR,*) &
        "ERROR in generate_Grid: "//grid_type//" is not supported."
      stop
  end select
  !
  ! 2) memory allocation
  !
  if ( G%grid_type == 3 ) then          ! extended LGL grid
    allocate( G%x(0:N+1), G%r(0:N+1), G%rp(0:N+1), G%weight(0:N+1) )
    allocate( G%wf_coeff(0:N+1) )
    allocate( G%D1(0:N+1,0:N+1), G%D2(0:N+1,0:N+1) )
    allocate( G%Deriv(0:N+1,0:N+1) )
  else
    allocate( G%x(N), G%r(N), G%rp(N), G%weight(N), G%wf_coeff(N) )
    allocate( G%D1(N,N), G%D2(N,N), G%Deriv(N,N) )
  end if
  !
  ! 3) make computational grids
  !
  if ( G%grid_type == 0 ) then          ! for FGH
    G%Rmin = min( Rmin, Rmax )
    G%Rmax = max( Rmin, Rmax )
    do i = 1, N
      G%x(i) = G%Rmin + ( G%Rmax - G%Rmin ) * dble(i-1) / dble(N)
      !G%x(i) = G%Rmin + ( G%Rmax - G%Rmin ) * dble(i) / dble(N+1)
      G%r(i) = G%x(i)
    end do
    call prepare_GPS_matrix( G )
    return
  else
    call find_GPS_points( G%x(1:N), N, G%grid_type )
    if ( G%grid_type == 3 ) then        ! for ext-LGL
      G%x(0)   = -ONE
      G%x(N+1) =  ONE
    end if
  end if
  !
  ! 4) make real grids, r(:) and rp(:)
  !    if Rmin and Rmax are not specified, i.e. an arbitrary mapping is used,
  !    do nothing.
  !
  if ( present(Rmin) .and. present(Rmax) ) then
    if ( present(L) .and. present(mapping) ) then 
      call apply_mapping( G, Rmin, Rmax, L, mapping )
    else if ( present(L) ) then
      call apply_mapping( G, Rmin, Rmax, L )
    else
      call apply_mapping( G, Rmin, Rmax )
    end if
  end if
  !
  ! 5) make D1(:,:), D2(:,:). Deriv(:,:), and weight(:)
  !
  if ( G%mapping == 0 ) then    ! if it is an arbitrary mapping
    G%Rmin = -ONE
    G%Rmax =  ONE
  else                          ! if it's not an arbitrary mapping
    call prepare_GPS_matrix( G )
  end if
end function generate_Grid


subroutine purge_Grid( G )
  type(Grid), intent(inout) :: G
  deallocate( G%x, G%r, G%rp, G%weight, G%wf_coeff )
  if ( allocated( G%D1 ) ) deallocate( G%D1 )
  if ( allocated( G%D2 ) ) deallocate( G%D2 )
  if ( allocated( G%Deriv ) ) deallocate( G%Deriv )
end subroutine purge_Grid


subroutine apply_mapping( G, Rmin, Rmax, L, mapping )
!
! make real grids ( r(:) and rp(:) ) using a mapping function
!
  type(Grid), intent(inout) :: G
  real(long), intent(in) :: Rmax, Rmin
  real(long), intent(in), optional :: L
  character(len=*), intent(in), optional :: mapping
  integer :: i, N
  !
  ! set parameters
  !
  N = G%N
  G%Rmin = min( Rmin, Rmax )
  G%Rmax = max( Rmin, Rmax )
  G%L = ZERO;   if ( present(L) ) G%L = L
  !
  ! determine a mapping type
  !
  G%mapping = 1
  if ( G%L /= ZERO ) G%mapping = 2      ! default: nonlinear if L is assigned
  if ( present(mapping) ) then
    select case ( mapping )
      case ( "arbitrary", "nothing" );                  G%mapping = 0
      case ( "linear" );                                G%mapping = 1
      case ( "algebraic", "reciprocal", "nonlinear" );  G%mapping = 2
      case ( "inverse_reciprocal", "inverse" );         G%mapping = 3
      case default
        write(STDERR,*) &
          "ERROR in generate_Grid: "//mapping//" is not supported."
        stop
    end select
  end if
  !
  ! construct real grids
  !
  if ( G%mapping /= 0 ) then
    do i = 1, N
      call mapping_function &
           ( G%mapping, G%r(i), G%rp(i), G%x(i), G%Rmin, G%Rmax, G%L )
    end do
    if ( G%grid_type == 3 ) then        ! it requires [0:N+1]
      call mapping_function &
           ( G%mapping, G%r(0), G%rp(0), G%x(0), G%Rmin, G%Rmax, G%L )
      if ( G%Rmin /= G%Rmax ) then      ! if Rmax is not infinite
        call mapping_function &
             ( G%mapping, G%r(N+1), G%rp(N+1), G%x(N+1), G%Rmin, G%Rmax, G%L )
      end if
    end if
  end if
end subroutine apply_mapping


subroutine prepare_GPS_matrix( G )
  type(Grid), intent(inout) :: G
  select case ( G%grid_type )
    case ( 0 ); call construct_FGH( G )
    case ( 1 ); call construct_LG( G )
    case ( 2 ); call construct_LGL( G )
    case ( 3 ); call construct_extended_LGL( G )
  end select
end subroutine prepare_GPS_matrix


function GPS_grid_type( G ) result (string)
  character(len=80) :: string
  type(Grid), intent(in) :: G
  select case ( G%grid_type )
    case ( 0 );  string = "Fourier-Grid-Hamiltonian"
    case ( 1 );  string = "Legendre-Gauss"
    case ( 2 );  string = "Legendre-Gauss-Lobatto"
    case ( 3 );  string = "extended Legendre-Gauss-Lobatto"
    case ( 4 );  string = "Legendre-Gauss-Radau"
  end select
end function GPS_grid_type


!----------------------------------------------------------------------------
! Legendre-Gauss-Lobatto grid: [1...N]
!   collocation points: P'_N+1(x) = 0 ( P'_N(x)=0 in Canuto's notation )
!----------------------------------------------------------------------------
subroutine construct_LGL( G )
  type(Grid), intent(inout) :: G
  integer :: N, i, j
  real(long) :: d1, d2
  N = G%N
  !
  ! construct D1 and D2 matrices, and the first derivatives Deriv
  !   D1_ij := d1_ij / sqrt( r'(x_i) * r'(x_j) )
  !   D2_ij := d2_ij / ( r'(x_i) * r'(x_j) )
  !   Deriv_ij := d1_ij / r'(x_i) * P_N(x_i) / P_N(x_j) in Canuto's notation
  !
  do i = 1, N
    do j = 1, N
      d1 = compute_d1( G, i, j )
      d2 = compute_d2( G, i, j )
      G%D1(i,j) = d1 / sqrt( G%rp(i)*G%rp(j) )
      G%D2(i,j) = d2 / ( G%rp(i)*G%rp(j) )
      G%Deriv(i,j) = d1 / G%rp(i) * &
                     Legendre_P( N+1, G%x(i) ) / Legendre_P( N+1, G%x(j) )
    end do
  end do
  !
  ! determine weights and wf_coeff
  ! Gauss-Lobatto integral in Canuto's notation:
  !   ordinary weights: w_i = 2 / { N(N+1) P_N(x_i)^2 }
  !                     weight = r'_i * w_i
  !   wf_coeff: wf_coeff_i = 1 / sqrt( r'_i w_i ) <- here, +/- problem of P_N
  !             so, it might be safe:
  !               wf_coeff_i = sqrt( N(N+1)/2 / r'_i ) * P_N(x_i)
  !
  ! Why do we define wf_coeff?
  ! Using wf_coeff, we can convert matrix element A (i.e. eigenvector) into wf.
  !   wf_i = wf_coeff_xi * A_i
  !
  do i = 1, N
    G%weight(i) = TWO / dble((N+1)*(N+2)) / Legendre_P(N+1,G%x(i))**2 * G%rp(i)
    G%wf_coeff(i) = sqrt( dble((N+1)*(N+2)/2)/G%rp(i) ) * Legendre_P(N+1,G%x(i))
    !G%wf_coeff(i) = ONE / sqrt( G%weight(i) )
  end do
end subroutine construct_LGL


!----------------------------------------------------------------------------
! extended Legendre-Gauss-Lobatto grid: [0...N+1]
!   collocation points: P'_N+1(x) = 0
!
! It can utilize end points of 0 and N+1, so all elements have [0...N+1].
! If Rmax == Rmin, Rmax means infinity. If so, all elements must have [0...N].
!----------------------------------------------------------------------------
subroutine construct_extended_LGL( G )
  type(Grid), intent(inout) :: G
  integer :: N, i, j
  logical :: yes_R_infinity
  real(long) :: d1, d2
  N = G%N
  yes_R_infinity = .false.
  if ( G%Rmin == G%Rmax ) yes_R_infinity = .true.
  !
  ! construct D1 and D2 matrices, and the first derivatives Deriv
  !   D1_ij := d1_ij / sqrt( r'(x_i) * r'(x_j) )
  !   D2_ij := d2_ij / ( r'(x_i) * r'(x_j) )
  !   Deriv_ij := d1_ij / r'(x_i) * P_N+1(x_i) / P_N+1(x_j)
  !
  do i = 0, N+1
    do j = 0, N+1
      if ( yes_R_infinity .and. ( i == N+1 .or. j == N+1 ) ) cycle
      d1 = compute_d1( G, i, j )
      d2 = compute_d2( G, i, j )
      G%D1(i,j) = d1 / sqrt( G%rp(i)*G%rp(j) )
      G%D2(i,j) = d2 / ( G%rp(i)*G%rp(j) )
      G%Deriv(i,j) = d1 / G%rp(i) * &
                     Legendre_P( N+1, G%x(i) ) / Legendre_P( N+1, G%x(j) )
    end do
  end do
  !
  ! determine weights and wf_coeff
  ! Gauss-Lobatto integral in Canuto's notation:
  !   ordinary weights: w_i = 2 / { N(N+1) P_N(x_i)^2 }
  !                     weight = r'_i * w_i
  !   wf_coeff: wf_coeff_i = 1 / sqrt( r'_i w_i )
  !
  do i = 0, N+1
    if ( yes_R_infinity .and. i == N+1 ) cycle
    G%weight(i) = TWO / dble((N+1)*(N+2)) / Legendre_P(N+1,G%x(i))**2 * G%rp(i)
    G%wf_coeff(i) = sqrt( dble((N+1)*(N+2)/2)/G%rp(i) ) * Legendre_P(N+1,G%x(i))
  end do
end subroutine construct_extended_LGL


!----------------------------------------------------------------------------
! Legendre-Gauss grid
!   collocation points: P_N(x) = 0 ( P_N+1(x)=0 in Canuto's notation )
!----------------------------------------------------------------------------
subroutine construct_LG( G )
  type(Grid), intent(inout) :: G
  integer :: N, i, j
  real(long) :: d1, d2
  N = G%N
  !
  ! construct D1 and D2 matrices, and the first derivatives Deriv
  !
  do i = 1, N
    do j = 1, N
      d1 = compute_d1( G, i, j )
      d2 = compute_d2( G, i, j )
      G%D1(i,j) = d1 / sqrt( G%rp(i)*G%rp(j) )
      G%D2(i,j) = d2 / ( G%rp(i)*G%rp(j) )
      G%Deriv(i,j) = d1 / G%rp(i) * &
                     Legendre_dP( N, G%x(i) ) / Legendre_dP( N, G%x(j) )
    end do
  end do
  !
  ! determine weights and wf_coeff
  ! Gauss integral in Canuto's notation:
  !   ordinary weights: w_i = 2 / { (1-x_i^2) P'_N+1(x_i)^2 }
  !                     weight = r'_i * w_i
  !   wf_coeff: wf_coeff_i = 1 / sqrt( r'_i w_i )
  !
  do i = 1, N
    G%weight(i) = TWO / (ONE-G%x(i)**2) / Legendre_dP(N,G%x(i))**2 * G%rp(i)
    G%wf_coeff(i) = sqrt( (ONE-G%x(i)**2)/TWO/G%rp(i) ) * Legendre_dP(N,G%x(i))
  end do
end subroutine construct_LG


!----------------------------------------------------------------------------
! Fourier-Grid-Hamiltonian method
!----------------------------------------------------------------------------
subroutine construct_FGH( G )
  type(Grid), intent(inout) :: G
  integer :: N, i, j
  real(long) :: s
  if ( G%Rmax == G%Rmin ) stop "ERROR in generate_FGH: Rmax==Rmin"
  N = G%N
  !
  ! construct D2 matrix
  !
  s = ( TWO * PI / (G%Rmax - G%Rmin) )**2
  do i = 1, N
    do j = 1, N
      if ( i == j ) then
        !G%D2(i,j) = - s * dble((N+1)**2 + 2) / 12.0_long
        G%D2(i,j) = - s * dble(N**2 + 2) / 12.0_long
        G%D1(i,j) = ZERO
      else
        !G%D2(i,j) = s / TWO / (sin( PI*dble(i-j)/dble(N+1) ))**2
        G%D2(i,j) = s / TWO / (sin( PI*dble(i-j)/dble(N) ))**2
        if ( modulo( (i+j+1), 2 ) == 1 ) G%D2(i,j) = -G%D2(i,j)
        !G%D1(i,j) = sqrt(s) / TWO / tan( PI*dble(i-j)/dble(N+1) )
        G%D1(i,j) = sqrt(s) / TWO / tan( PI*dble(i-j)/dble(N) )
        if ( modulo( (i-j), 2 ) == 1 ) G%D1(i,j) = -G%D1(i,j)
      end if
    end do
  end do
  !
  ! determine weights and wf_coeff
  !
  G%weight(:) = ( G%Rmax - G%Rmin ) / G%N
  G%wf_coeff(:) = ONE / sqrt( G%weight(:) )
end subroutine construct_FGH


!----------------------------------------------------------------------------
! cardinal function in Canuto's notation
!   for FGH:
!   for LG:  g_j(x) = P_N+1(x) / ( x - x_j ) / P'_N+1(x_j)
!   for LGL: g_j(x) = - ( 1 - x^2 ) P'_N(x) / ( x - x_j ) / N(N+1) P_N(x_j)
!----------------------------------------------------------------------------
function cardinal_f( G, j, x ) result (f)
  real(long) :: f
  type(Grid), intent(in) :: G
  integer, intent(in) :: j
  real(long), intent(in) :: x
  real(long) :: dx
  dx = x - G%x(j)
  if ( abs( dx ) < EPS ) then
    f = ONE
  else
    select case ( G%grid_type )
      case ( 0 )      ! for FGH
        f = sin( dble(G%N)*PI / (G%Rmax - G%Rmin) * dx ) / &
            tan( PI / ( G%Rmax - G%Rmin ) * dx ) / dble(G%N) 
        !f = sin( dble(G%N+1)*PI / (G%Rmax - G%Rmin) * dx ) / &
        !    tan( PI / ( G%Rmax - G%Rmin ) * dx ) / dble(G%N+1) 
      case ( 1 )      ! for LG
        f = Legendre_P( G%N, x ) / dx / Legendre_dP( G%N, G%x(j) )
      case ( 2, 3 )   ! for LGL or ext-LGL
        f = - ( ONE - x**2 ) * Legendre_dP( G%N+1, x ) / dx / &
            dble((G%N+1)*(G%N+2)) / Legendre_P( G%N+1, G%x(j) )
    end select
  end if
end function cardinal_f


!----------------------------------------------------------------------------
! Interpolation of f(:) at r with GPS
!   For LGL, it's assumed that f(x)->0 at end points.
!----------------------------------------------------------------------------
function GPS_interpolate( G, f, r ) result (I)
  real(long) :: I
  type(Grid), intent(in) :: G
  real(long), intent(in) :: f(:), r
  real(long) :: x, alpha
  integer :: j
  !
  ! find x from r
  !
  select case ( G%mapping )
    case ( 1 )          ! linear mapping
      x = ( r - G%Rmin ) / ( G%Rmax - G%Rmin ) * TWO - ONE
    case ( 2 )          ! reciprocal mapping
      alpha = ZERO
      if ( G%Rmax /= G%Rmin ) alpha = TWO * G%L / ( G%Rmax - G%Rmin )
      x = ( alpha + TWO ) / ( ONE + G%L / ( r - G%Rmin ) ) - ONE
    case ( 3 )          ! inverse reciprocal mapping
      alpha = TWO * G%L / ( G%Rmax - G%Rmin )
      x = ONE - ( alpha + TWO ) / ( ONE + G%L / ( G%Rmax - r ) )
    case default
      write(STDERR,*) "GPS_interpolate is not implemented yet " // &
                      "with arbitrary mapping function."
  end select
  !
  ! compute cardinal functions
  !
  I = ZERO
  if ( G%grid_type /= 3 ) then
    do j = 1, G%N
      I = I + cardinal_f( G, j, x ) * f(j)
    end do
  else          ! for ext-LGL
    if ( size(f) /= G%N+2 ) then
      write(STDERR,*) "Extended LGL requires f(x) at end points."
      stop
    end if
    do j = 0, G%N+1
      I = I + cardinal_f( G, j, x ) * f(j+1)
    end do
  end if
end function GPS_interpolate


!----------------------------------------------------------------------------
! compute_d1: compute an element of the 1st derivative matrix
! NOTE: It doesn't contain Legendre polynomials in the D1 matrix.
!       Be careful when you compare it to Canuto's notation.
!----------------------------------------------------------------------------
function compute_d1( G, i, j ) result (d1)
  real(long) :: d1
  type(Grid), intent(in) :: G
  integer, intent(in) :: i, j
  integer :: N
  N = G%N
  select case ( G%grid_type )
    case ( 0 )          ! for FGH
      !
      ! not implemented yet
      !
      d1 = ZERO
    case ( 1 )          ! for LG
      if ( i == j ) then
        d1 = G%x(i) / ( ONE - G%x(i)**2 )
      else
        d1 = ONE / ( G%x(i) - G%x(j) )
      end if
    case ( 2, 3 )       ! for LGL and ext-LGL
                        ! compute d1, putting N+1 in N of the original formula
      if ( i == 0 .and. j == 0 ) then
        d1 = - dble( (N+1)*(N+2) ) / FOUR
      else if ( i == N+1 .and. j == N+1 ) then
        d1 = dble( (N+1)*(N+2) ) / FOUR
      else if ( i == j ) then
        d1 = ZERO
      else 
        d1 = ONE / ( G%x(i) - G%x(j) )
      end if
  end select
end function compute_d1


!----------------------------------------------------------------------------
! compute_d2: compute an element of the 2nd derivative matrix
! NOTE: It doesn't contain Legendre polynomials in the D1 matrix.
!       Be careful when you compare it to Canuto's notation.
!----------------------------------------------------------------------------
function compute_d2( G, i, j ) result (d2)
  real(long) :: d2
  type(Grid), intent(in) :: G
  integer, intent(in) :: i, j
  integer :: N
  N = G%N
  select case ( G%grid_type )
    case ( 0 )          ! for FGH
      if ( i == j ) then
        d2 = PI**2 * dble( (N+1)**2 + 2 ) / 12.0_long
      else
        d2 = - PI**2 * HALF / sin( PI*( G%x(i) - G%x(j) ) )**2
        if ( modulo( (i+j+1), 2 ) == 1 ) d2 = -d2
      end if
    case ( 1 )          ! for LG
      !
      ! not implemented yet, it's just complicated
      !
      d2 = ZERO
    case ( 2, 3 )       ! for LGL and ext-LGL
                        ! compute d1, putting N+1 in N of the original formula
      if ( ( i == 0 .and. j == 0 ) .or. ( i == N+1 .and. j == N+1 ) ) then
        d2 = dble( (N+1)*(N+2) * ( (N+1)*(N+2) - 2 ) ) / 24.0_long
      else if ( ( i == 0 .and. j == N+1 ) .or. ( i == N+1 .and. j == 0 ) ) then
        d2 = dble( (N+1)*(N+2) - 2 ) / FOUR
      else if ( i == j ) then
        d2 = - dble( (N+1)*(N+2) ) / THREE / ( ONE - G%x(i)**2 )
      else 
        d2 = - TWO / ( G%x(i) - G%x(j) )**2
      end if
  end select
end function compute_d2


!----------------------------------------------------------------------------
! Find computational grids for LG or LGL using the Newton-Raphson method
!   LG:  x_j (j=1,2,...,N) : the solutions of P_N(x) = 0
!   LGL: x_j (j=1,2,...,N) : the solutions of P'_N+1(x) = 0
!        Finally, we can get N points (x(1), x(2), .., x(N)) except -1 and 1.
!
! For LG, This Newton-Raphson code gives exactly same values as d01bcf in NAG.
! > call d01bcf( 0, -ONE, ONE, ZERO, ZERO, N, w, x, IFAIL )
!----------------------------------------------------------------------------
subroutine find_GPS_points( x, N, grid_type )
  real(long), intent(out) :: x(:)
  integer, intent(in) :: N, grid_type
  real(long) :: xx, del_x
  integer :: N1, NH
  integer, parameter :: KSTOP = 10
  real(long) :: dth, cd, sd, cs, ss, cssave
  real(long) :: p, dp, ddp, q, dq, recsum
  integer :: i, k
  N1 = N + 1
  NH = N / 2            ! for example, 11 / 2 = 5
  !
  ! set up initial values
  !   PI = 4.0*atan(1.0)
  !
  dth = PI / dble(2*N1 + 1)
  cd = cos(TWO * dth)
  sd = sin(TWO * dth)
  cs = cos(dth)
  ss = sin(dth)
  !
  ! This procedure is derived from Appendix C (Canuto, 1988).
  !
  do i = 1, NH
    xx = cs
    do k = 1, KSTOP
      if ( grid_type == 1 ) then
        !
        ! For LG grid: get P_N(x), P'_N(x)
        ! and we want to solve q(x) = P_N(x)
        !
        call calculate_Legendre(N, xx, p, dp)
        q = p
        dq = dp
      else if ( grid_type == 2 .or. grid_type == 3 ) then
        !
        ! For LGL grid: get P'_N+1(x) and calculate P''_N+1(x)
        ! and we want to solve q(x) = P'_N+1(x)
        !
        call calculate_Legendre(N1, xx, p, dp)
        ddp = ( 2 * xx * dp - N1*(N1+1) * p ) / (1 - xx**2)
        q = dp
        dq = ddp
      end if
      !
      ! calculate the reciprocal summation of previous x values
      !      recsum = 0.0
      !      do j = 1, i-1
      !        recsum = recsum + 1.0 / (xx - x(j))
      !      end do
      !
      recsum = sum( ONE / ( xx - x(1:i-1) ) )
      !
      ! This technique is modified Newton-Raphson method for root-polishing.
      ! It needs the first deriv. (dq) and the reciprocal summation (recsum).
      !                                 q(x_k)
      ! x_k+1 = x_k - ------------------------------------------
      !               q'(x_k) - q(x_k) SUM_i=1^j 1/{ x_k - x_i }
      !
      del_x = -q / (dq - recsum * q)
      xx = xx + del_x
      if ( abs( del_x ) < EPS ) exit
    end do
    x(i) = xx
    cssave = cs * cd - ss * sd
    ss = cs * sd + ss * cd
    cs = cssave
  end do
  !
  ! use symmetry for second half of the roots
  !
  do i = 1, NH
    x(N1-i) = x(i)
    x(i) = -x(i)        ! reverse order: it makes the ascending order for r
  end do
  !
  ! if N is odd, zero is also the root of P'_N+1(x)=0
  !
  if ( modulo( N, 2 ) == 1 ) x(NH+1) = ZERO
end subroutine find_GPS_points


!----------------------------------------------------------------------------
!        d psi |
! impose ----- |              = 0 in D1 matrix for only LGL or ext-LGL grids
!         d r  |r at boundary
!----------------------------------------------------------------------------
subroutine impose_LGL_boundary( G, boundary )
  type(Grid), intent(inout) :: G
  character(len=*), intent(in) :: boundary
  integer :: i, j, N, bc_type
  real(long) :: den
  N = G%N
  select case ( boundary )
    case ( "both" );                            bc_type = 0
    case ( "min", "Rmin", "rmin", "r_min" );    bc_type = 1
    case ( "max", "Rmax", "rmax", "r_max" );    bc_type = 2
    case default;                               bc_type = 0
  end select
  if ( G%Rmax == G%Rmin .and. ( bc_type == 2 .or. bc_type == 0 ) ) then
    write(STDERR,*) &
      "A boundary of Rmax or both is not allowed for Rmax is infinite."
  end if
  select case ( G%grid_type )
    case ( 0, 1 )
      write(STDERR,*) "It's not possible to impose boundaries for FGH or LG."
      stop
    case ( 2 )          ! for LGL
      !
      ! we can construct a new D1 matrix using boundary elements of 0 and N+1.
      !   d1(0,  0  ) = - (N+1)(N+2) / 4
      !   d1(N+1,N+1) = (N+1)(N+2) / 4
      !   d1(0,  N+1) = - 1/2
      !   d1(N+1,0  ) = 1/2
      !   d1(0,  j  ) = 1 / ( - 1 - x(j) )
      !   d1(N+1,j  ) = 1 / (   1 - x(j) )
      !   d1(i,  0  ) = 1 / ( x(i) + 1 )
      !   d1(i,  N+1) = 1 / ( x(i) - 1 )
      ! --> D1(i,j) = d1(i,j) / sqrt( rp(i) * rp(j) )
      !
      den = - ( dble((N+1)*(N+2)) / FOUR )**2 + ONE / FOUR
      do i = 1, N
        do j = 1, N
          if ( bc_type == 0 ) then              ! for both end points
            G%D1(i,j) = G%D1(i,j) - ONE / sqrt( G%rp(i) * G%rp(j) ) * &
                        ( dble((N+1)*(N+2)) / FOUR * &
                            ( - ONE / ( G%x(i)+ONE ) / ( G%x(j)+ONE ) &
                              + ONE / ( G%x(i)-ONE ) / ( G%x(j)-ONE ) ) &
                          - HALF * &
                            (   ONE / ( G%x(i)+ONE ) / ( G%x(j)-ONE ) &
                              - ONE / ( G%x(i)-ONE ) / ( G%x(j)+ONE ) ) ) / den 
          else if ( bc_type == 1 ) then         ! for min end point
            G%D1(i,j) = G%D1(i,j) - ONE / sqrt( G%rp(i) * G%rp(j) ) / &
                                    ( G%x(i)+ONE ) / ( G%x(j)+ONE ) / &
                                    ( dble((N+1)*(N+2)) / FOUR )
          else if ( bc_type == 2 ) then         ! for max end point
            G%D1(i,j) = G%D1(i,j) + ONE / sqrt( G%rp(i) * G%rp(j) ) / &
                                    ( G%x(i)-ONE ) / ( G%x(j)-ONE ) / &
                                    ( dble((N+1)*(N+2)) / FOUR )
          end if
        end do
      end do
    case ( 3 )          ! for ext-LGL
      den = G%D1(0,0) * G%D1(N+1,N+1) - G%D1(0,N+1) * G%D1(N+1,0)
      do i = 1, N
        do j = 1, N
          if ( bc_type == 0 ) then              ! for both end points
            G%D1(i,j) = G%D1(i,j) - &
                        ( G%D1(N+1,N+1) * G%D1(0,j) * G%D1(i,0) + &
                          G%D1(0,0) * G%D1(N+1,j) * G%D1(i,N+1) - &
                          G%D1(0,N+1) * G%D1(N+1,j) * G%D1(i,0) - &
                          G%D1(N+1,0) * G%D1(0,j) * G%D1(i,N+1) ) / den
          else if ( bc_type == 1 ) then         ! for min end point
            G%D1(i,j) = G%D1(i,j) - G%D1(i,0) * G%D1(0,j) / G%D1(0,0)
          else if ( bc_type == 2 ) then         ! for max end point
            G%D1(i,j) = G%D1(i,j) - G%D1(i,N+1) * G%D1(N+1,j) / G%D1(N+1,N+1)
          end if
        end do
      end do
  end select
end subroutine impose_LGL_boundary


!----------------------------------------------------------------------------
! Mapping function:
!   It is implemented as subroutine form because it produces r and rp.
!   mapping = 
!     1: linear mapping
!     2: algebraic(reciprocal) mapping
!     3: reversed reciprocal mapping
!   x := ( -1, 1 )
!   r := ( Rmin, Rmax )  for Rmin /= Rmax
!   r := ( Rmin, inf  )  for Rmin == Rmax
!----------------------------------------------------------------------------
subroutine mapping_function( mapping, r, rp, x, Rmin, Rmax, L )
  integer, intent(in) :: mapping
  real(long), intent(out) :: r, rp
  real(long), intent(in) :: x, Rmin, Rmax
  real(long), intent(in), optional :: L
  real(long) :: alpha
  select case ( mapping )
    case ( 1 )
      !
      ! Linear mapping
      !   r = (Rmax-Rmin)/2*(1+x)
      !   rp = (Rmax-Rmin)/2
      !
      r = Rmin + HALF * ( Rmax - Rmin ) * ( ONE + x )
      rp = HALF * ( Rmax - Rmin )
    case ( 2 )
      !
      ! Algebraic (Reciprocal) mapping
      ! denser grid near Rmin
      !   r = L(1+x)/(1-x) or r = L(1+x)/(1-x+alpha)
      !   rp = 2L/(1-x)**2 or rp = L(2+alpha)/(1-x+alpha)**2
      !     where alpha = 2L/(Rmax-Rmin)
      !
      if ( .not. present(L) ) stop "ERROR in mapping_function: it requires L."
      if ( Rmax == Rmin ) then
        alpha = ZERO        ! Rmax = infinity case
        if ( x == ONE ) stop "ERROR in mapping_reciprocal: r hits infinity."
      else
        alpha = TWO * L / ( Rmax - Rmin )
      end if
      r  = Rmin + L * ( ONE + x ) / ( ONE - x + alpha )
      rp = L * ( TWO + alpha ) / ( ONE - x + alpha )**2
    case ( 3 )
      !
      ! Reciprocal mapping (reversed way)
      ! denser grid near Rmax
      !   r = Rmax - L(1-x)/(1+x+alpha) where alpha = 2L/(Rmax-Rmin)
      !   rp = L(2+alpha)/(1-x+alpha)**2
      !
      if ( .not. present(L) ) stop "ERROR in mapping_function: it requires L."
      if ( Rmax == Rmin ) then
        stop "ERROR in mapping_reciprocal_reversed: Rmax should be finite."
      end if
      alpha = TWO * L / ( Rmax - Rmin )
      r  = Rmax - L * ( ONE - x ) / ( ONE + x + alpha )
      rp = L * ( TWO + alpha ) / ( ONE + x + alpha )**2
  end select
end subroutine mapping_function

function GPS_mapping_r( G, x ) result (r)
  real(long) :: r
  type(Grid), intent(in) :: G
  real(long), intent(in) :: x
  real(long) :: rp              ! discard rp
  call mapping_function( G%mapping, r, rp, x, G%Rmin, G%Rmax, G%L )
end function GPS_mapping_r

function GPS_mapping_rp( G, x ) result (rp)
  real(long) :: rp
  type(Grid), intent(in) :: G
  real(long), intent(in) :: x
  real(long) :: r               ! discard r
  call mapping_function( G%mapping, r, rp, x, G%Rmin, G%Rmax, G%L )
end function GPS_mapping_rp

function GPS_mapping_x( G, r ) result (x)
  real(long) :: x
  type(Grid), intent(in) :: G
  real(long), intent(in) :: r
  real(long) :: Rmin, Rmax, L, alpha
  Rmin = G%Rmin
  Rmax = G%Rmax
  L = G%L
  !
  ! inverse mapping function
  !
  select case ( G%mapping )
    case ( 1 )
      !
      ! Linear mapping: r = (Rmax-Rmin)/2*(1+x)
      !
      x = TWO * ( r - Rmin ) / ( Rmax - Rmin ) - ONE
    case ( 2 )
      !
      ! Algebraic (Reciprocal) mapping (denser grid near Rmin)
      !   r = L(1+x)/(1-x) or r=L(1+x)/(1-x+alpha) where alpha = 2L/(Rmax-Rmin)
      !
      if ( L == ZERO ) stop "ERROR in GPS_mapping_x: L should not be zero."
      if ( Rmax == Rmin ) then
        alpha = ZERO        ! Rmax = infinity case
      else
        alpha = TWO * L / ( Rmax - Rmin )
      end if
      x = ( ( ONE + alpha ) * ( r - Rmin ) - L ) / ( r - Rmin + L )
    case ( 3 )
      !
      ! Reciprocal mapping (reversed way, denser grid near Rmax)
      !   r = Rmax - L(1-x)/(1+x+alpha) where alpha = 2L/(Rmax-Rmin)
      !
      if ( L == ZERO ) stop "ERROR in GPS_mapping_x: L should not be zero."
      if ( Rmax == Rmin ) stop "ERROR in GPS_mapping_x: Rmax should be finite."
      alpha = TWO * L / ( Rmax - Rmin )
      x = ( L - ( Rmax - r ) * ( ONE + alpha ) ) / ( Rmax - r + L )
  end select
end function GPS_mapping_x

end module gps
