!----------------------------------------------------------------------------
! impact_ionization.f90
!----------------------------------------------------------------------------
! Copyright (C) 2017-2018 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2017-2018 John Bekx
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
!******************************************************************************
! Module to calculate the double differential scattering cross section of 
! inelastic electron - electron scattering.
! Written by John Jasper Bekx, September - December 2017.
!******************************************************************************
 module impact_ionization

  use constants
  use read_param
  use wavefunction
  use spherical_bessel
  use cfelt_bessel
  use gps
  implicit none

  contains

!======================================================================================================= 
!-----------------------------------Small functions needed----------------------------------------------
!======================================================================================================= 

function momentum_transfer(energy_in, energy_out, theta_rad) result (Q)
! For electrons, E = (hbar q)^2/(2m), which in au becomes E = q^2 / 2.
! Q = sqrt( q_in**2 + q_out**2 -2*q_in*q_out*cos(theta)) 
!   = sqrt( 2*E_in + 2*E_out - 4*sqrt(E_in*E_out)*cos(theta))
  real(long), intent(in) :: energy_in, energy_out, theta_rad
  real(long) :: Q
  Q = dsqrt(TWO * energy_in + TWO * energy_out - FOUR * dsqrt(energy_in * energy_out) * dcos(theta_rad))
end function momentum_transfer

!======================================================================================================= 

function Rutherford_CS(energy_in, energy_out, theta_rad)
! The Rutherford cross-section is (4m^2e^4)/(hbar^4q^4), which in au becomes 4/q^4.
  real(long), intent(in) :: energy_in, energy_out, theta_rad
  real(long) :: Q, RutherFord_CS
  Q = momentum_transfer(energy_in, energy_out, theta_rad)
  Rutherford_CS = FOUR / (Q)**4 
end function Rutherford_CS

!======================================================================================================= 

!======================================================================================================= 
!-----------------------------DCS(theta) and CS by using the "Manual" method.---------------------------
!------Integral over energies is not explicitly done. The energy conservation is manually imposed.------
!======================================================================================================= 

function calculate_impact_ionization_DCS_angle_res_manual(P, F, energy_in, theta_rad) &
                                                                                  & result (DCS_partial)
! This function calculates the bound-to-free contribution to S(Q)*Rutherford cross-section for a given 
! angle theta (rad). 
  type(Param), intent(in)             :: P               
  type(WF_data), intent(in)           :: F
  real(long), intent(in)              :: theta_rad, energy_in 
  real(long)                          :: DCS_partial(F%N_orb)         
  integer, dimension(F%N_all)         :: occ_all
  real(long)                          :: Q
  integer                             :: i, L, i_all, column
  integer                             :: n_i, occ_i, l_i, n_f, l_f, n_all, L_max
  real(long)                          :: E, epsilon_i, epsilon_f, CG, R_integrand, sum_1, sum_2, factor
  real(long), external                :: cleb
  real(long), allocatable, dimension(:,:) :: j_bes  
  
  occ_all(:) = 0
  DCS_partial(:) = ZERO 
  if (energy_in == 0.d0) then
    return
  end if 
  L_max = maxval( F%config(:,2) ) + P%lmax + 1   

  allocate(j_bes(F%N,L_max))

! This prepares all initial occupation numbers
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

! (8.36) in Jan Malte's thesis.
  occupied_orbitals: do i = 1, F%N_orb 
    n_i   = F%config(i,1)                     
    l_i   = F%config(i,2)     
    occ_i = F%config(i,3)         
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i 
    epsilon_i = F%e_orb(i) 
    sum_2 = ZERO                       
    if (occ_i == 0) cycle                  ! Skip empty orbitals.
    if (energy_in .lt. -epsilon_i) cycle   ! Don't bother if the incoming energy can't ionize the e.
  
    final_states: do l_f = 0, P%lmax
      grid_points: do n_f = 1, F%N 
        n_all = F%N * l_f + n_f
        epsilon_f = F%e_all(n_all)       
        E = epsilon_f - epsilon_i  
        if (epsilon_f .lt. 0.d0) cycle    ! only b2free contributions
        if (energy_in-E .lt. 0.d0) cycle  ! Don't include negative final energies; 
                                          ! Commenting this out will lead call 
                                          ! cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes) to go in
                                          ! an infinite loop. But this also happens in compton.f90
        if (i_all == n_all) cycle         ! Don't include "transitions" to the same state

        sum_1 = ZERO                     
        Q = momentum_transfer(energy_in, energy_in-E, theta_rad)  
        call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
        do L = 0, l_i+l_f    
          CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
          if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
          column = L+1                  
          R_integrand = radial_integral_u_all( F, i_all, n_all, j_bes(:,column)) 
          sum_1 = sum_1 + (R_integrand * CG)**2     
        end do
        factor = (TWO * dble(l_f) + ONE) * dsqrt((energy_in - E)/energy_in) * &
        &Rutherford_CS(energy_in, energy_in - E, theta_rad) * &
        &(ONE - dble(occ_all(n_all))/dble(2*(2*l_f+1)))
        sum_2 = sum_2 + factor * sum_1 
      end do grid_points
    end do final_states
    DCS_partial(i) =  sum_2 *dble(occ_all(i_all)) 
  end do occupied_orbitals
end function calculate_impact_ionization_DCS_angle_res_manual

!======================================================================================================= 

subroutine print_impact_ionization_DCS_angle_res_manual(P, F)
! This subroutine is called by -imp_ion_angle_res_manual. It returns the individual orbital 
! contributions to S(Q)*Rutherford cross section for a range of angles theta. The first row returns the 
! angle in radians, followed by the sum of all orbital contributions. 
! Then follow the bound-to-free contributions, differentiated per orbital; so 1s is 
! the 3rd column, 2s the 4th, followed by 2p, etc... 
! We have NOT considered impact excitation, and so b2b contributions aren't shown. Since the DDCS 
! routine below cannot calculate b2b contributions, we wouldn't have anything to compare it to anyway.
  
  type(Param), intent(in)                :: P
  type(WF_data), intent(in)              :: F
  !type(Grid)                             :: costheta
  integer                                :: OUTPUT     
  integer                                :: N_theta    
  real(long)                             :: energy_in
  real(long), allocatable, dimension(:)  :: b2free
  real(long), allocatable, dimension(:,:):: b2free_orbi
  real(long)                             :: theta_rad_array(P%imp_ion_N_theta) 
  integer                                :: m, a 
  real(long)                             :: d_theta_deg, d_theta_rad, min_theta_deg, min_theta_rad   

  OUTPUT = P%OUTPUT 
  
  energy_in = P%EE_start

! Angular grid
  N_theta = P%imp_ion_n_theta   
  min_theta_deg = P%imp_ion_min_theta
  min_theta_rad = min_theta_deg * PI/180.d0  
  d_theta_deg = P%imp_ion_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  theta_rad_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /)  

  if (PI-dble(N_theta-1)*d_theta_rad .lt. -1.d-6) then
    write(STDERR,*) "You will be looking at angles larger than Pi. Check N_theta and d_theta."
    stop
  endif
    
  allocate(b2free_orbi(N_theta,F%N_orb))
  allocate(b2free(N_theta))                

! This writes out the results.
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the Manual method.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'It calculates individual orbital contributions to the differential&
                            & cross section (DCS) factor, w.r.t. the scattering angle.'
  write(OUTPUT,"('# ', a)") 'The DCS is given by Rutherford_CS * S(Q).'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'The output is structured as follows:'
  write(OUTPUT,"('# ', a)") 'Columns 3 until (2 + number of subshells of system under consideration) &
                             &are orbital contributions. All other columns are properly labeled.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, f10.3)") 'The incoming energy of the electron is (in eV):', &
                                    &P%EE_start*real(AU2EV, 4)
  write(OUTPUT,"('# ', a)") ' '
  write(OUTPUT,"('# ', 1a12, 1a20, 1a26)") 'Theta(rad)', 'Angle resolved DCS', 'Orbital Contributions'
  do m = 1, N_theta                        
    b2free_orbi(m,:) = calculate_impact_ionization_DCS_angle_res_manual(P,F,energy_in, &
                     & theta_rad_array(m))
  end do
  do m = 1, N_theta
    b2free(m) = sum (b2free_orbi(m,1:F%N_orb))
    write(OUTPUT,'(1ES12.3, 1ES15.3, 1a8, 99ES15.3)') theta_rad_array(m), b2free(m), '',b2free_orbi(m,:) 
  end do

  deallocate(b2free_orbi)
  deallocate(b2free)                

end subroutine print_impact_ionization_DCS_angle_res_manual
!======================================================================================================= 

subroutine print_impact_ionization_total_cross_section_manual(P,F)
! Called using -imp_ion_total_cs_manual. Prints the total impact ionization cross section for a range of
! incoming energies energy_in using the manaul method.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  type(Grid)                           :: costheta  
  integer                              :: OUTPUT
  integer                              :: grid_type  ! 0 = log_general, 1 = log_specific, 2 = lin_spaced
  integer                              :: N_energy_in
  integer                              :: N_theta
  real(long), allocatable, dimension(:):: energy_in_array
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  integer                              :: base10exponent, first_num, energy_int, index_cutoff
  integer                              :: base10EE_end, base10EE_start 
  real(long)                           :: rest, restEE_end 
  real(long)                           :: epsilon_i(F%N_orb)
  real(long)                           :: d_energy_in, energy_in_range
  integer                              :: a, i, j, m
  real(long)                           :: DCS_manual_orbi(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: temp_storage(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: total_CS_manual_orbi(F%N_orb)
  real(long)                           :: total_CS_manual

  OUTPUT = P%OUTPUT

! Angular grid
  N_theta = P%imp_ion_N_theta        
  costheta = generate_Grid("LG", N_theta, -ONE, ONE, 5.d-1)
  do m = 1, N_theta
    theta_rad_array(m) = abs(PI-acos(costheta%r(N_theta - m + 1)))  ! this now goes from ZERO to PI
  enddo                                                             ! with the points most dense near 0.

! Initial orbital energies
  do a = 1, F%N_orb
    epsilon_i(a) = F%e_orb(a)
  enddo 

! Energy_in grid. This automatically generates a logarithmically spaced grid (base 10) for an initial 
! energy range. We have defined it in two ways: the normal way for analyzing qualitative behavior 
! (grid_type = 0), and the fine way (grid_type = 1), in order to resolve the structure more for a given 
! energy range.
! grid_type = 2 gives a linearly spaced grid.

! NOTE for possible confusion: 
! 1) In all subroutines P%EE_start > P%EE_end. Though consistent, I can see how this can be confusing.
! 2) In previous subroutines P%EE_end was chosen such as to define a starting point from which to
!    observe energy-resolved behavior in a range of outgoing energies.
!    Now however, P%EE_end is defined as a starting point for incoming energies to look at.
! Again, the fact that I just said "P%EE_end is defined as a starting point" definitely leads to 
! confusion.

  grid_type = P%imp_ion_grid_type
!------------------------------------------------------------------------------------------------------- 
! Energy_in grid; log10 grid. Use this if you want to get some qualitative structure over a large energy
! range
! Treat any given range as if it had started from 0, but continue from the value of P%EE_end.
! Examples for [P%EE_end, P%EE_start]:
! - [0, 100] means 0,1,2,3,4,5,6,7,8,9,10,20,30,40,50,60,70,80,90,100
! - [500, 600] means 500,600 
! - [0, 56] means 0,1,2,3,4,5,6,7,8,9,10,20,30,40,50,56
! - [3, 56] means 3,4,5,6,7,8,9,10,20,30,40,50,56
! - [500, 2000] means 500,600,700,800,900,1000,2000

  if (grid_type == 0) then
    energy_in_range = P%EE_start - P%EE_end   !These CANNOT have the same exponent in base 10.

!   NOTE: the input parameters P%imp_ion_N_energy and P%imp_ion_d_energy refer only to the outgoing 
!   energy grid. 

! We start by decomposing P%EE_start = first_num * 10**(base10EE_start) + rest
    if (P%EE_end == 0.d0) then
      base10EE_end = 0
    else  
      base10EE_end = int(log10(P%EE_end*AU2EV)) 
    endif
    base10EE_start = int(log10(P%EE_start*AU2EV))
    base10exponent = base10EE_start  - base10EE_end
    energy_int = int(P%EE_start*AU2EV) + 1
    first_num = energy_int/10**(base10EE_start)
    if (P%EE_start*AU2EV - dble(first_num * 10**(base10EE_start)) .lt. 1d-3) then
      rest = 0.d0
    else 
      rest = mod(P%EE_start*AU2EV, 10.d0**dble(base10EE_start))
    endif

! Finding how long our energy_in array will be
    if(P%EE_end == 0.d0 .or. base10EE_end == 0) then
      if (rest .ne. 0.d0) then 
        N_energy_in = 1 + 9 * base10exponent + first_num  + 1 - (int(P%EE_end * AU2EV))
      else
        N_energy_in = 1 + 9 * base10exponent + first_num - (int(P%EE_end * AU2EV))
      endif
    else
      if (rest .ne. 0.d0) then 
        N_energy_in =  9 * base10exponent + first_num + 1 - (int(P%EE_end * AU2EV / 10**(base10EE_end))&
                       & - 1 )
      else
        N_energy_in = 9 * base10exponent + first_num - (int(P%EE_end * AU2EV / 10**(base10EE_end)) - 1)
      endif
    endif

! Failsafe for when N_energy_in is negative or zero
    if (N_energy_in .lt. 0) then 
      write(STDERR, *) "The number of incoming-energy steps is N_energy_in =",  N_energy_in
    endif

    allocate(energy_in_array(N_energy_in))
    energy_in_array(:) = P%EE_end
    index_cutoff = 2

! Finding the rest associated to P%EE_end
    if (P%EE_end*AU2EV - dble((int(P%EE_end*AU2EV) + 1)/10**(base10EE_end) * 10**(base10EE_end)) &
        & .lt. 1d-3) then
      restEE_end = 0.d0
    else 
      restEE_end = mod(P%EE_end*AU2EV, 10.d0**dble(base10EE_end))
    endif

! First two elements
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(1)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
      energy_in_array(2) =  energy_in_array(1) + dble(10**(base10exponent))*EV2AU - restEE_end * EV2AU
    enddo

! The rest of energy_in_array
    11 continue
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(index_cutoff - 1 + i)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
        energy_in_array(index_cutoff - 1 + i + 1) =  energy_in_array(index_cutoff) + (i) * & 
                                             &dble(10**(base10exponent))*EV2AU
      if (int(log10(energy_in_array(index_cutoff - 1 + i + 1)*AU2EV)) .ne. base10exponent) then
        index_cutoff = index_cutoff + i
        goto 11                
      endif
    enddo

! The last two elements
    energy_in_array(N_energy_in) = P%EE_start
    if (rest .ne. 0.d0) then
      energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - rest*EV2AU
    else 
      base10exponent = int(log10(energy_in_array(N_energy_in)*AU2EV))
      if (base10exponent .ne. int(log10(energy_in_array(N_energy_in-2)*AU2EV)) .and. first_num == 1)then
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent - 1))*EV2AU
      else 
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent))*EV2AU
      endif      
    endif

    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies (in eV) is as follows :"
    do i = 1, N_energy_in
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo

! Failsafe for negative incoming energies
    do i = 1, N_energy_in
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 

!-------------------------------------------------------------------------------------------------------
! Energy_in grid; fine grid. Use this if you want to resolve some fine-structure at a certain energy 
! range.
! Treat a given range as if it starts from 0, and only in the end transpose the resulting grid on the 
! actual energy range given.
! Examples for [P%EE_end, P%EE_start]:
! - [0, 100] means 0,1,2,3,4,5,6,7,8,9,10,20,30,40,50,60,70,80,90,100
! - [500, 600] means 500,501,502,503,504,505,506,507,508,509,510,520,530,540,550,560,570,580,590,600 
! - [0, 56] means 0,1,2,3,4,5,6,7,8,9,10,20,30,40,50,56
! - [3, 56] means 3,4,5,6,7,8,9,10,11,12,13,23,33,43,53,56
  
  elseif(grid_type == 1) then 
    energy_in_range = P%EE_start - P%EE_end    

!   NOTE: the input parameters P%imp_ion_N_energy and P%imp_ion_d_energy refer only to the outgoing 
!   energy grid. 

    base10exponent= int(log10(energy_in_range*AU2EV))
    energy_int = int(energy_in_range*AU2EV) + 1
    first_num = energy_int/10**(base10exponent)
    if (energy_in_range*AU2EV - dble(first_num * 10**(base10exponent)) .lt. 1d-3) then
      rest = 0.d0
    else 
      rest = mod(energy_in_range*AU2EV, 10.d0**dble(base10exponent))
    endif
    if (rest .ne. 0.d0) then 
      N_energy_in = 1 + 9 * base10exponent + first_num  + 1 
    else
      N_energy_in = 1 + 9 * base10exponent + first_num 
    endif

! Failsafe for when N_energy_in is negative or zero
    if (N_energy_in .lt. 0) then 
      write(STDERR, *) "The number of incoming-energy steps is N_energy_in =",  N_energy_in
    endif

    allocate(energy_in_array(N_energy_in))
    energy_in_array(:) = ZERO
    index_cutoff = 1

    10 continue
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(index_cutoff - 1 + i)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
        energy_in_array(index_cutoff - 1 + i + 1) =  energy_in_array(index_cutoff) + (i) * & 
                                             &dble(10**(base10exponent))*EV2AU
      if (int(log10(energy_in_array(index_cutoff - 1 + i + 1)*AU2EV)) .ne. base10exponent) then
        index_cutoff = index_cutoff + i
        goto 10                
      endif
    enddo

    energy_in_array(N_energy_in) = P%EE_start - P%EE_end
    if (rest .ne. 0.d0) then
      energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - rest*EV2AU
    else 
      base10exponent = int(log10(energy_in_array(N_energy_in)*AU2EV))
      if (base10exponent .eq. int(log10(energy_in_array(N_energy_in-2)*AU2EV))) then
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent))*EV2AU
      else 
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent - 1 ))*EV2AU
      endif      
    endif
    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies is as follows (in eV):"
    do i = 1, N_energy_in
      energy_in_array(i) = energy_in_array(i) + P%EE_end
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo

! Failsafe for negative incoming energies
    do i = 1, N_energy_in
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 

!-------------------------------------------------------------------------------------------------------
! Energy_in grid: linear grid. 
  elseif(grid_type == 2) then
    N_energy_in = P%imp_ion_n_in
    d_energy_in = P%imp_ion_d_in * EV2AU
    energy_in_range = P%EE_start - P%EE_end    
    if (abs(energy_in_range - N_energy_in * d_energy_in) .gt. 1.d-3) then
      write(STDERR, *) "Check your values for N_energy_in and d_energy_in. I have", N_energy_in, "and",&
                       & d_energy_in * AU2EV, "eV."
      stop
    endif
    allocate(energy_in_array(N_energy_in + 1))
    energy_in_array(:) = ZERO
    do i = 1, N_energy_in + 1
      energy_in_array(i) = P%EE_end + dble(i-1) * d_energy_in 
    enddo
    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies (in eV) is as follows :"
    do i = 1, N_energy_in + 1
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo

! Failsafe for negative incoming energies
    do i = 1, N_energy_in + 1
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 
!-------------------------------------------------------------------------------------------------------
  endif

! Writing some declaration output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ', a)")  'It calculates the total cross-section (CS) for a range of & 
                              &incoming energies by an angle integration of the Manual method.' 
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a)")  'Columns 3 to (3 - 1 + number of subshells under consideration) are the &
                              &orbital contributions (1s, 2s, 2p, 3s,...) of the Manual method.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, I5)")'The number of points in the angular grid was:', N_theta
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 1a15, 1a20, 1a30 )") 'Energy_in [eV]','CS Manual [Mb]'&
                                                   &, 'Orbital contributions'

 ! Total cross section calculations
  do i = 1, N_energy_in

! Just print out 0 if there is no ionization
    if (energy_in_array(i)+ maxval(epsilon_i) .lt. 0.d0) then
      total_CS_manual_orbi(:) = ZERO
      total_CS_manual = ZERO
      write(OUTPUT,'(1ES15.3, 1ES18.3, 99ES15.3)') energy_in_array(i) * AU2EV, &
                                                         &total_CS_manual, & 
                                                         &total_CS_manual_orbi(:)
      cycle
    endif

!  Manual method 
    DCS_manual_orbi(:,:) = ZERO
    temp_storage(:,:) = ZERO
    total_CS_manual_orbi(:) = ZERO
    do m = 1, N_theta                        
      DCS_manual_orbi(m,:) = calculate_impact_ionization_DCS_angle_res_manual(P,F,energy_in_array(i)&
                             &,theta_rad_array(m))
      temp_storage(m,:) = DCS_manual_orbi(m,:) * costheta%weight(N_theta - m + 1)
    end do 
    do j = 1, F%N_orb    
      total_CS_manual_orbi(j) = (TWO * PI) *sum( temp_storage(1:N_theta, j)) 
      total_CS_manual_orbi(j) = total_CS_manual_orbi(j)* AU2M**2 / 1.0e-22_long  ! converted to megabarn
    end do
    total_CS_manual = sum(total_CS_manual_orbi(1:F%N_orb))

    write(OUTPUT,'(1ES15.3, 1ES18.3, 99ES15.3)') energy_in_array(i) * AU2EV, & 
                                                         &total_CS_manual, & 
                                                         &total_CS_manual_orbi(:)

  end do 

  deallocate(energy_in_array)
end subroutine print_impact_ionization_total_cross_section_manual

!======================================================================================================= 

!======================================================================================================= 
!---------------DDCS, DCS(energy), DCS(theta), and CS using the nonuniform pseudocontinuum--------------
!======================================================================================================= 

function calculate_impact_ionization_DDCS(P,F, theta_rad, energy_in, energy_out) result(DDCS_orbital)
  type(Param), intent(in)                 :: P               
  type(WF_data), intent(in)               :: F
  real(long), intent(in)                  :: theta_rad
  real(long), intent(in)                  :: energy_in
  real(long), intent(in)                  :: energy_out
  real(long)                              :: f_n(F%N * (P%lmax + 1))   ! Too big 
  integer                                 :: E_index(P%lmax + 1), number_of_neg(P%lmax + 1)
  integer                                 :: E_cutoff_lies_in_epsilon_f_range(P%lmax + 1) 
  real(long)                              :: epsilon_f(F%N * (P%lmax + 1)), epsilon_i, E_cutoff
  integer                                 :: L_max
  real(long), allocatable, dimension(:,:) :: j_bes  
  real(long)                              :: Q
  integer, dimension(F%N_all)             :: occ_all
  integer                                 :: n_i, l_i, occ_i, i_all
  integer                                 :: n_f, l_f, n_all, L, column
  real(long)                              :: sum_1, CG, R_integrand, factor
  real(long), external                    :: cleb
  !integer                                 :: index_all
  !integer                                 :: slope_index, slope_all
  integer                                 :: i, a
  !integer                                 :: j, b
  real(long)                              :: histogram(F%N * (P%lmax + 1)) ! Too big 
  real(long)                              :: slope_at_E
  real(long)                              :: DDCS_orbital(F%N_orb)

  DDCS_orbital(:) = ZERO

! This prepares all initial occupation numbers
  occ_all(:) = 0
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

! Preparing all final orbital energies
  do l_f = 0, P%lmax 
    do n_f = 1, F%N
      n_all = F%N * l_f + n_f
      epsilon_f(n_all) = F%e_all(n_all)
    end do
  end do

! Number of negative energies epsilon_f per l_f. 
! It allows us to access the positive epsilon_f's when used in an index
  number_of_neg(:) = ZERO                    
  do l_f = 0, P%lmax 
    do n_f = 1, F%N
      n_all = F%N * l_f + n_f
      if (epsilon_f(n_all) .lt. 0.d0) then
        number_of_neg(l_f + 1) = number_of_neg(l_f + 1) + 1
      endif
    end do
  end do
 
  L_max = maxval( F%config(:,2) ) + P%lmax + 1 
  allocate(j_bes(F%N,L_max))

  orbitals: do i = 1, F%N_orb

    n_i   = F%config(i,1)                     
    l_i   = F%config(i,2)     
    i_all   = F%N * l_i + (n_i-l_i)
    epsilon_i = F%e_orb(i) 

!  Don't bother calculating if the energy transfer isn't at least the ionization energy 
    if (energy_in - energy_out .lt. -epsilon_i) then 
      cycle
    end if
 ! Don't bother if the incoming energy can't ionize the e.
    if (energy_in .lt. -epsilon_i) then  
      cycle
    end if

    E_cutoff = energy_in - energy_out + epsilon_i 

! E_index is the index value of the epsilon_f (per l_f) which is the smallest value larger than E_cutoff
    E_index(:) = 1

    E_cutoff_lies_in_epsilon_f_range(:) = 1    ! 1 means yes, 0 means no

! Determining the cutoff in n_f per l_f
    do l_f = 0, P%lmax
      if (E_cutoff .lt. epsilon_f(F%N * l_f + 1 + number_of_neg(l_f + 1))) then    ! We don't want 
        E_cutoff_lies_in_epsilon_f_range(l_f + 1) = 0       ! scenarios where the smallest epsilon_f(>0)
      endif                                                 ! is larger than the cutoff energy. 
      do n_f = 1, F%N
        n_all = F%N * l_f + n_f
        if (n_all .eq. F%N * (P%lmax + 1)) cycle
        if (epsilon_f(n_all) .lt. -1.d-6) cycle
        if (E_cutoff .gt. epsilon_f(F%N * l_f + F%N)) then ! If the cutoff energy is larger than any 
          E_index(l_f + 1) = F%N                          ! outgoing energy considered (very unlikely),  
          exit                                           ! approximate with the largest energy available
        endif
        if ( (E_cutoff .gt. epsilon_f(n_all)) .and. (E_cutoff .le. epsilon_f(n_all + 1)) ) then
          E_index(l_f + 1) = n_f + 1            ! Index 0 doesn't make sense. Hence the +1.
          exit                                             
        endif                                            
      end do 
    enddo
    
    do l_f = 0, P%lmax
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .lt. 1 .and. &
                                                &E_cutoff_lies_in_epsilon_f_range(l_f + 1) .eq. 1 ) then   
        write(STDERR,*) "One of the E_index values is less than one, somehow"    
        stop 
      endif                                              
    enddo                                                  

! Calculating the DDCS
    j_bes(:,:) = ZERO
    f_n(:) = ZERO
    do l_f = 0, P%lmax
      if (E_cutoff_lies_in_epsilon_f_range(l_f + 1) .ne. 1) cycle
      do n_f = 1, E_index(l_f + 1) 
        n_all = F%N * l_f + n_f
        if (epsilon_f(n_all) .lt. 0) cycle    ! Only b2free contributions
        if (energy_in + epsilon_i - epsilon_f(n_all) .lt. 0.d0) cycle ! This for makes the difference 
                                                                      ! between 4min and 41 seconds. 
        if (i_all == n_all) cycle   ! Don't include "transitions" to the same state
        sum_1 = ZERO                     
        Q = momentum_transfer(energy_in, energy_out, theta_rad)  
        call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
        do L = 0, l_i+l_f    
          CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
          if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
          column = L+1                  
          R_integrand = radial_integral_u_all(F, i_all, n_all, j_bes(:,column)) 
          sum_1 = sum_1 + (R_integrand * CG)**2      ! This does the sum over L         
        end do
        factor = (TWO * dble(l_f) + ONE) * dsqrt(energy_out/energy_in)  
        f_n(n_all) = Rutherford_CS(energy_in, energy_out, theta_rad) * occ_all(i_all) * factor * sum_1 
      end do    
    end do 

    histogram(:) = ZERO
    slope_at_E = ZERO
    final_l: do l_f = 0, P%lmax 
!   Calculating the histograms

!   If there's only one point of epsilon_f between 0 and E_cutoff, we will pretend that there is a 
!   histogram of height zero at epsilon_f = 0, and take the slope with respect to that. 
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .eq. 1) then          
        histogram(F%N * l_f + 1) = 0.d0      
        histogram(F%N * l_f + 2) =f_n(F%N * l_f + 1 + number_of_neg(l_f + 1))  
      else
        do n_f = 1, E_index(l_f + 1)
          n_all = F%N * l_f + n_f
          histogram(n_all) = sum (f_n(F%N * l_f + 1:F%N * l_f + n_f))   ! sum_{n_f} 
        end do
      endif

!   Calculating the slope between the last two histograms
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .eq. 1) then      
        E_index(l_f + 1) = 2   
        slope_at_E = slope_at_E + ( histogram(F%N * l_f + E_index(l_f + 1)) - & 
                                    &histogram(F%N * l_f + E_index(l_f + 1) - 1 ) ) / &
                            & abs( epsilon_f(F%N * l_f + E_index(l_f + 1) + number_of_neg(l_f + 1)) - & 
                                  &epsilon_f(F%N * l_f + E_index(l_f + 1)- 1) + number_of_neg(l_f + 1) ) 
      else

!   Making sure we have at least two points to interpolate between
        if (E_index(l_f + 1) - number_of_neg(l_f + 1) .lt. 2 .and. &
                                                & E_cutoff_lies_in_epsilon_f_range(l_f + 1) .eq. 1) then          
          write(STDERR,*) "The index of the cutoff energy should be larger than 1. It isn't."
          stop
        endif
        slope_at_E = slope_at_E + ( histogram(F%N * l_f + E_index(l_f + 1)) - & 
                                    &histogram(F%N * l_f + E_index(l_f + 1) - 1 ) ) / &
                                    & abs( epsilon_f(F%N * l_f + E_index(l_f + 1)) - & 
                                    &epsilon_f(F%N * l_f + E_index(l_f + 1)- 1) ) 
      endif
    end do final_l

    DDCS_orbital(i) = slope_at_E
    if (DDCS_orbital(i) .lt. 0.d0 .and. DDCS_orbital(a) .gt. -1.d-100) then
      DDCS_orbital(i) = 0.d0
    endif

  end do orbitals
  deallocate(j_bes)

end function calculate_impact_ionization_DDCS

!=======================================================================================================

subroutine print_impact_ionization_DDCS(P,F)
! Called using -imp_ion_DDCS. Prints the DDCS for a range of theta_rad and enery_out.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  integer                              :: OUTPUT
  integer                              :: N_energy, N_theta
  integer                              :: number_of_edges, N_energy_out
  real(long), allocatable, dimension(:):: energy_out_array
  real(long)                           :: orbital_energies(F%N_orb), temp
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  real(long)                           :: energy_in_start, energy_in_end
  real(long)                           :: d_energy_au, d_energy_eV
  real(long)                           :: d_theta_deg, d_theta_rad, min_theta_deg, min_theta_rad   
  integer                              :: a, b, i, j
  real(long)                           :: DDCS_orbital(F%N_orb)
                                  
  OUTPUT = P%OUTPUT 

  energy_in_start = P%EE_start

  N_energy = P%imp_ion_N_energy
  N_theta = P%imp_ion_N_theta 

  d_energy_eV = P%imp_ion_d_energy 
  d_energy_au = d_energy_eV * EV2AU

  d_theta_deg = P%imp_ion_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  min_theta_deg = P%imp_ion_min_theta
  min_theta_rad = dble(min_theta_deg) * PI/180.d0 

  DDCS_orbital(:) = ZERO
  orbital_energies(:) = ZERO  

  do i = 1, F%N_orb
    orbital_energies(F%N_orb - i + 1) = F%e_orb(i) ! These are now in increasing order, ending with 1s.
  end do

! Angular grid
  theta_rad_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /)

  if (PI-dble(N_theta-1)*d_theta_rad - min_theta_rad .lt. -1.d-6) then
    write(STDERR,*) "You will be looking at angles larger than Pi. Check N_theta and d_theta."
    stop
  endif

! Energy_out grid
  energy_in_end = P%EE_end
! energy_in_end will only define until what point the outgoing energy array energy_out will go. So:

! 0--|--|--|--|--5--|--|...|--|--energy_in_end--|--|...|--energy_in_start. 

! NOTE: it is assumed energy_in_start is NOT a part of the energy_out grid (inelasticity)

! Failsafe for when energy_out becomes negative. 
  if (energy_in_start - dble(N_energy) * d_energy_au .lt. energy_in_end-1.0d-6) then 
    write(STDERR, *) 'Check your choices for energy_in_start, N_energy, and d_energy.'
    stop
  endif

! Adding the edges explicitly
  number_of_edges = F%N_orb
  do a = 1, F%N_orb
    if (energy_in_start - energy_in_end + orbital_energies(a) .lt. 0.d0) then
      number_of_edges = number_of_edges - 1
    endif
  end do
 
  N_energy_out = N_energy + number_of_edges
  allocate(energy_out_array(N_energy_out))
  energy_out_array(:) = ZERO
  if (number_of_edges .ne. 0) then
    do a = 1, number_of_edges
      energy_out_array(a) = energy_in_start + orbital_energies(a)
    end do
  endif
  do a = number_of_edges + 1, N_energy_out
    energy_out_array(a) = energy_in_end + dble(a-number_of_edges-1) * d_energy_au
  end do 

! Sorting the array in decreasing order
  do a = 1, N_energy_out
    do b = a, N_energy_out
      if (energy_out_array(a) .lt. energy_out_array(b)) then 
        temp = energy_out_array(b)
        energy_out_array(b) = energy_out_array(a)
        energy_out_array(a) = temp
      endif
    enddo
  enddo

! Writing standard output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)")  'It calculates the double differential cross-section (DDCS) for a & 
                              & scattering angle theta, and a scattered energy energy_out.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, f10.3)") 'The incoming energy of the electron is (in eV):', &
                                    &P%EE_start*real(AU2EV, 4)

  do i = 1, N_theta
    write(OUTPUT,"('# ', a)") ' '
    write(OUTPUT,"('# ', 1a12, 1a18, 1a16, 1a13, 1a35 )") 'Theta (rad)', 'Energy_out(eV)', &
                                      & 'Rutherford_CS', 'Total DDCS', 'Orbital contributions to DDCS'    
    do j = 1, N_energy_out
      DDCS_orbital = calculate_impact_ionization_DDCS(P,F, theta_rad_array(i), energy_in_start, &
                     &energy_out_array(j)) 
      write(OUTPUT,'(4ES15.3E4, 99ES20.3E4)') theta_rad_array(i), energy_out_array(j) * AU2EV, & 
                                 Rutherford_CS(energy_in_start, energy_out_array(j), &
                                 &theta_rad_array(i)), &
                                 &sum( DDCS_orbital(1:F%N_orb) ), DDCS_orbital(:)
    end do 
  write(OUTPUT,"(a)") ' '
  write(OUTPUT,"('# ', a)") ' '
  end do           
  deallocate(energy_out_array)  
end subroutine print_impact_ionization_DDCS 

!=======================================================================================================

subroutine print_impact_ionization_DCS_energy_res(P,F)
! Called using -imp_ion_energy_resolved. Prints the angle-integrated DCS for a range of energy_out.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  type(Grid)                           :: costheta
  integer                              :: OUTPUT
  integer                              :: N_energy, N_theta
  integer                              :: number_of_edges, N_energy_out
  real(long), allocatable, dimension(:):: energy_out_array
  real(long)                           :: orbital_energies(F%N_orb), temp 
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  real(long)                           :: energy_in_start, energy_in_end
  real(long)                           :: d_energy_au, d_energy_eV
  integer                              :: a, b, i, j, m
  real(long), allocatable, dimension(:, :,:):: DDCS_orbital
  real(long), allocatable, dimension(:,:,:) :: temp_storage
  real(long), allocatable, dimension(:,:)   :: DCS_energy
  real(long), allocatable, dimension(:)     :: DCS_energy_tot
                                  
  OUTPUT = P%OUTPUT 

  do i = 1, F%N_orb
    orbital_energies(F%N_orb - i + 1) = F%e_orb(i) ! These are now in increasing order, ending with 1s.
  end do

! Angular grid
  N_theta = P%imp_ion_N_theta         
  costheta = generate_Grid("LG", N_theta, -ONE, ONE, 5.d-1)
  do m = 1, N_theta
    theta_rad_array(m) = abs(PI-acos(costheta%r(N_theta - m + 1)))   ! this now goes from ZERO to PI
  enddo                                                              ! most dense points on grid near 0.

! Energy_out grid
  energy_in_start = P%EE_start

  N_energy = P%imp_ion_N_energy

  d_energy_eV = P%imp_ion_d_energy 
  d_energy_au = d_energy_eV * EV2AU

  energy_in_end = P%EE_end
! energy_in_end will only define until what point the outgoing energy array energy_out will go. So:

! 0--|--|--|--|--5--|--|...|--|--energy_in_end--|--|...|--energy_in_start.

! NOTE: it is assumed energy_in_start is NOT a part of the energy_out grid. (Inelasticity) 

! Failsafe for when energy_out is out of range. 
  if (energy_in_start - N_energy * d_energy_au .lt. energy_in_end-1.0d-6) then 
    write(STDERR, *) 'Check your choices for energy_in_start, N_energy, and d_energy.'
    stop
  endif

! Adding the edges explicitly
  number_of_edges = F%N_orb
  do a = 1, F%N_orb
    if (energy_in_start - energy_in_end + orbital_energies(a) .lt. 0.d0) then
      number_of_edges = number_of_edges - 1
    endif
  end do

  N_energy_out = N_energy + number_of_edges
  allocate(energy_out_array(N_energy_out))
  energy_out_array(:) = ZERO
  if (number_of_edges .ne. 0) then
    do a = 1, number_of_edges
      energy_out_array(a) = energy_in_start + orbital_energies(a)
    end do
  endif
  do a = number_of_edges + 1, N_energy_out
    energy_out_array(a) = energy_in_end + dble(a-number_of_edges-1) * d_energy_au
  end do 

! Sorting the array in decreasing order
  do a = 1, N_energy_out
    do b = a, N_energy_out
      if (energy_out_array(a) .lt. energy_out_array(b)) then 
        temp = energy_out_array(b)
        energy_out_array(b) = energy_out_array(a)
        energy_out_array(a) = temp
      endif
    enddo
  enddo

! Writing standard output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ', a)")  'It calculates the differential cross-section (DCS) for a range of& 
                              & scattered energy energy_out. An angle integration was performed'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, f10.3)") 'The incoming energy of the electron is (in eV):', &
                                    &P%EE_start*real(AU2EV, 4)
  write(OUTPUT,"('# ', a)") ' '
  write(OUTPUT,"('# ', 1a18, 1a23, 1a35 )") 'Energy_out(eV)', 'Energy-resolved DCS', &
                                            &'Orbital contributions to DCS'

! DDCS calculation
  allocate(DDCS_orbital(N_theta,N_energy_out,F%N_orb))
  allocate(temp_storage(N_theta,N_energy_out,F%N_orb))
  allocate(DCS_energy(N_energy_out,F%N_orb))
  allocate(DCS_energy_tot(N_energy_out))

  DDCS_orbital(:,:,:) = ZERO
  do i = 1, N_theta    
    do j = 1, N_energy_out
      DDCS_orbital(i,j,:) = calculate_impact_ionization_DDCS(P,F, theta_rad_array(i), energy_in_start, &
                     &energy_out_array(j)) 

      do a = 1, F%N_orb
        if (DDCS_orbital(i,j,a) .lt. 0.d0) then
          write(STDERR,*) "Negative DDCS", theta_rad_array(i), energy_out_array(j)*AU2EV, a, &
                                                                                    &DDCS_orbital(i,j,a)
        endif
      enddo

    end do 
  end do  

  temp_storage(:,:,:) = ZERO
  do m = 1, N_theta
    temp_storage(m,:,:) = DDCS_orbital(m, :, :)*costheta%weight(N_theta - m + 1)
  enddo

! Angular integration
  do i = 1, N_energy_out
    do j = 1, F%N_orb
      DCS_energy(i,j) = (TWO * PI) * sum(temp_storage(1:N_theta,i,j))

      if (DCS_energy(i,j) .lt. 0.d0 ) then
        write(STDERR,*) "Negative DCS", energy_out_array(i)*AU2EV, j, DCS_energy(i,j)
        stop
      endif
 
    end do 
    DCS_energy_tot(i) = sum(DCS_energy(i,1:F%N_orb))
    write(OUTPUT,'(2ES18.3, 1ES25.3, 99ES13.3)') energy_out_array(i) * AU2EV, DCS_energy_tot(i), &
                                                 &DCS_energy(i,:)
  end do

  deallocate(energy_out_array)
  deallocate(DDCS_orbital)
  deallocate(temp_storage)
  deallocate(DCS_energy)
  deallocate(DCS_energy_tot)

end subroutine print_impact_ionization_DCS_energy_res

!=======================================================================================================

subroutine print_impact_ionization_DCS_angle_res(P,F)
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  type(Grid)                           :: E_G
  integer                              :: OUTPUT
  integer                              :: N_theta, N_energy_out
  real(long), allocatable, dimension(:):: energy_out_array
  real(long), allocatable, dimension(:):: weight_array
  real(long)                           :: energy_out_range
  integer                              :: number_of_edges
  real(long), allocatable, dimension(:):: fraction_energy_out_range
  integer, allocatable, dimension(:)   :: N_energy_out_orbi
  real(long)                           :: starting_index
  real(long)                           :: orbital_energies(F%N_orb)
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  real(long)                           :: energy_in_start
  real(long)                           :: min_theta_deg, min_theta_rad, d_theta_deg, d_theta_rad   
  integer                              :: a, b, i, j
  real(long), allocatable, dimension(:,:,:):: DDCS_orbital
  real(long)                           :: DCS_angle(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: DCS_angle_tot(P%imp_ion_N_theta)
                                  
  OUTPUT = P%OUTPUT 

! Angular grid
  N_theta = P%imp_ion_n_theta   
  min_theta_deg = P%imp_ion_min_theta
  min_theta_rad = min_theta_deg * PI/180.d0  
  d_theta_deg = P%imp_ion_d_theta
  d_theta_rad = dble(d_theta_deg) * PI/180.d0     

  theta_rad_array(:) = (/ ( min_theta_rad + dble(a-1) * d_theta_rad, a=1, N_theta ) /) 

  if (PI-dble(N_theta-1)*d_theta_rad .lt. -1.d-6) then
    write(STDERR,*) "You will be looking at angles larger than Pi. Check N_theta and d_theta."
    stop
  endif

  do i = 1, F%N_orb
    orbital_energies(F%N_orb - i + 1) = F%e_orb(i) ! These are now in increasing order, ending with 1s.
  end do

! Energy grid
  energy_in_start = P%EE_start

! Writing standard output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ', a)")  'It calculates the differential cross-section (DCS) for a range of & 
                              &scattering angles theta. An energy integration was performed.'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, f10.3)") 'The incoming energy of the electron is (in eV):', &
                                    &P%EE_start*real(AU2EV, 4)
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, I5)")'The number of points in the energy_out grid was:', &
                                                                                     &P%imp_ion_n_energy
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, I5)")'The number of points in the angular grid was:', N_theta

! If we can't ionize the outer-most valence electron, the DDCS is ZERO, and so is the DCS_angle.
  if (energy_in_start + orbital_energies(1) .lt. 0.d0) then
    DCS_angle_tot(:) = ZERO
    DCS_angle(:,:) = ZERO
    do i = 1, N_theta
      write(OUTPUT,'(2ES18.3, 1ES25.3, 99ES13.3)') theta_rad_array(i), DCS_angle_tot(i), DCS_angle(i,:)
    enddo
    return
  endif


!--Energy_out grid--------------------------------------------------------------------------------------
! Between edges a nonuniform grid is defined, and these are glued together in the total energy_out 
! array.

  N_energy_out = P%imp_ion_n_energy
  if (N_energy_out .lt. 10) then              ! Let's assume you need at least 10 energy grid points
    N_energy_out = 10                         ! to make an energy integration for any incoming energy.
  end if

! We go from 0 eV to the outer-most valence edge.
  energy_out_range = energy_in_start + orbital_energies(1)

! Determining the number of edges and how many points will be assigned between two edges
  number_of_edges = F%N_orb
  do a = 1, F%N_orb
    if (energy_in_start + orbital_energies(a) .lt. 0.d0) then
      number_of_edges = number_of_edges - 1
    endif
  end do

  if (number_of_edges .eq. 0) then
    write(STDERR,*) "We shouldn't have been able to get here. The number of edges is zero"  
  endif                       
  allocate(fraction_energy_out_range(number_of_edges))
  allocate(N_energy_out_orbi(number_of_edges))
  fraction_energy_out_range(:) = 1.d0  ! If there's 1 edge, we don't need to split up the energy array
  if (number_of_edges .ge. 2) then
    do a = 1, number_of_edges - 1
      fraction_energy_out_range(a) = abs(orbital_energies(a)-orbital_energies(a+1)) / energy_out_range
    end do
    fraction_energy_out_range(number_of_edges) = (energy_in_start + &
                                                 &orbital_energies(number_of_edges))/ energy_out_range
  endif
  do a = 1, number_of_edges
   N_energy_out_orbi(a) = nint(fraction_energy_out_range(a) * N_energy_out)
    if (N_energy_out_orbi(a) .lt. 20) then   ! to avoid N_energy_out_orbi = 0, let's give it a 
      N_energy_out_orbi(a) = 20              ! minimal value of 20.
    endif
  enddo
 
  N_energy_out = sum(N_energy_out_orbi(1:number_of_edges))  ! This is the new N_energy_out
  allocate(energy_out_array(N_energy_out))
  allocate(weight_array(N_energy_out))

! Defining energy_out_array
  energy_out_array(:) = ZERO
  weight_array(:) = ZERO

! Special case of when number_of_edges = 1
  if (number_of_edges .eq. 1) then
    E_G = generate_Grid("LG", N_energy_out, ZERO, energy_in_start + orbital_energies(1), 5.d-1)
    do a = 1, N_energy_out
      energy_out_array(a) = abs(energy_in_start + orbital_energies(1) - E_G%r(a))  
      weight_array(a) = E_G%weight(a)
    enddo
    call purge_Grid(E_G)
  else 

! First elements
    E_G = generate_Grid("LG", N_energy_out_orbi(1), energy_in_start + orbital_energies(2), &
                                                     &energy_in_start + orbital_energies(1), 5.d-1)
    do a = 1, N_energy_out_orbi(1)
      energy_out_array(a) =  energy_in_start + orbital_energies(2) + abs(energy_in_start + &
                                                                     & orbital_energies(1) - E_G%r(a))
      weight_array(a) = E_G%weight(a)
    enddo
    call purge_Grid(E_G)

! Last elements
    E_G = generate_Grid("LG", N_energy_out_orbi(number_of_edges), ZERO, energy_in_start + &
                                                           & orbital_energies(number_of_edges), 5.d-1)
    starting_index = sum(N_energy_out_orbi(1:number_of_edges - 1))
    do a = 1, N_energy_out_orbi(number_of_edges)
      energy_out_array(starting_index + a) =  abs(energy_in_start + &
                                                        &orbital_energies(number_of_edges) - E_G%r(a))
      weight_array(starting_index + a) = E_G%weight(a)
    enddo
    call purge_Grid(E_G)

! If number_of_edges = 2, we're done, so skip this next part
    if (number_of_edges .eq. 2) goto 100

    do a = 2, number_of_edges - 1
      E_G = generate_Grid("LG", N_energy_out_orbi(a), energy_in_start + orbital_energies(a+1), &
                                                    & energy_in_start + orbital_energies(a), 5.d-1)
      starting_index = sum(N_energy_out_orbi(1:a-1))
      do b = 1 , N_energy_out_orbi(a)
        energy_out_array(starting_index + b) =  energy_in_start + orbital_energies(a+1) + &
                                             &abs(energy_in_start + orbital_energies(a) - E_G%r(b))
        weight_array(starting_index + b) = E_G%weight(b)
      enddo
      call purge_Grid(E_G)
    enddo
   
    100 continue
  endif


!--End of energy_out grid-------------------------------------------------------------------------------

! DDCS calculation
  allocate(DDCS_orbital(N_theta, N_energy_out, F%N_orb))
  DDCS_orbital(:,:,:) = ZERO

  do a = 1, N_theta
    do b = 1, N_energy_out
      DDCS_orbital(a,b,:) = calculate_impact_ionization_DDCS(P,F, theta_rad_array(a),& 
                                                          & energy_in_start, energy_out_array(b)) 
    end do 
  end do 

! Energy integration 
  write(OUTPUT,"('# ', a)") ' '
  write(OUTPUT,"('# ', 1a17, 1a20, 1a35 )") 'Theta(rad)', 'Angle-resovled DCS', &
                                            &'Orbital contributions to DCS'
  DCS_angle(:,:) = ZERO
  DCS_angle_tot(:) = ZERO

  do i = 1, N_theta
    do j = 1, F%N_orb
      DCS_angle(i,j) = sum( DDCS_orbital(i,1:N_energy_out,j) * weight_array(1:N_energy_out))
    enddo
    DCS_angle_tot(i) = sum(DCS_angle(i,1:F%N_orb))
   write(OUTPUT,'(2ES18.3, 1ES25.3, 99ES13.3)') theta_rad_array(i), DCS_angle_tot(i), DCS_angle(i,:)
  enddo

  deallocate(DDCS_orbital)
  deallocate(energy_out_array) 
  deallocate(weight_array) 
  deallocate(fraction_energy_out_range)
  deallocate(N_energy_out_orbi)

end subroutine print_impact_ionization_DCS_angle_res

!======================================================================================================= 

subroutine print_impact_ionization_total_cross_section(P,F)
! Called using -imp_ion_total_cs. Prints the total impact ionization cross section for a range of
! incoming energies energy_in using the two methods above.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  type(Grid)                           :: costheta, E_G  
  integer                              :: OUTPUT
  integer                              :: grid_type  ! 0 = log_general, 1 = log_specific, 2 = lin_spaced
  integer                              :: N_energy_in, N_energy_out
  integer                              :: N_theta
  real(long)                           :: temp
  real(long), allocatable, dimension(:):: energy_in_array
  real(long), allocatable, dimension(:):: energy_out_array
  real(long), allocatable, dimension(:):: weight_array
  real(long)                           :: energy_out_range
  integer                              :: number_of_edges
  real(long)                           :: orbital_energies(F%N_orb)
  real(long), allocatable, dimension(:):: fraction_energy_out_range
  integer, allocatable, dimension(:)   :: N_energy_out_orbi
  real(long)                           :: starting_index
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  integer                              :: base10exponent, first_num, energy_int, index_cutoff
  integer                              :: base10EE_end, base10EE_start 
  real(long)                           :: rest, restEE_end 
  real(long)                           :: epsilon_i(F%N_orb)
  real(long)                           :: d_energy_in, energy_in_range
  integer                              :: a, b, i, j, m
  real(long)                           :: DCS_manual_orbi(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: temp_storage(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: total_CS_manual_orbi(F%N_orb)
  real(long)                           :: total_CS_manual
  real(long), allocatable, dimension(:,:,:):: DDCS_orbi
  real(long), allocatable, dimension(:,:):: DDCS
  real(long)                           :: DCS_orbi(P%imp_ion_N_theta, F%N_orb)
  real(long)                           :: total_CS_from_DDCS_orbi(F%N_orb)
  real(long)                           :: total_CS_from_DDCS

  OUTPUT = P%OUTPUT

! Angular grid
  N_theta = P%imp_ion_N_theta   ! No converged value found yet.  
  costheta = generate_Grid("LG", N_theta, -ONE, ONE, 5.d-1)
  do m = 1, N_theta
    theta_rad_array(m) = abs(PI-acos(costheta%r(N_theta - m + 1)))  ! this now goes from ZERO to PI
  enddo                                                             ! with the points most dense near 0.

! Initial orbital energies
  do a = 1, F%N_orb
    epsilon_i(a) = F%e_orb(a)
  enddo 
  do i = 1, F%N_orb
    orbital_energies(F%N_orb - i + 1) = F%e_orb(i) ! These are now in decreasing order, ending with 1s.
  end do

! Sorting the array in decreasing order (Really making sure they're in increasing order)
  do a = 1, F%N_orb
    do b = a, F%N_orb
      if (orbital_energies(a) .lt. orbital_energies(b)) then 
        temp = orbital_energies(b)
        orbital_energies(b) = orbital_energies(a)
        orbital_energies(a) = temp
      endif
    enddo
  enddo

!=Energy_in grid========================================================================================

  grid_type = P%imp_ion_grid_type
!------------------------------------------------------------------------------------------------------- 
! Energy_in grid; log10 grid. Use this if you want to get some qualitative structure over a large energy
! range
  if (grid_type == 0) then
    energy_in_range = P%EE_start - P%EE_end   !These CANNOT have the same exponent in base 10.

!   NOTE: the input parameters P%imp_ion_N_energy and P%imp_ion_d_energy refer only to the outgoing 
!   energy grid. 

! We start by decomposing P%EE_start = first_num * 10**(base10EE_start) + rest
    if (P%EE_end == 0.d0) then
      base10EE_end = 0
    else  
      base10EE_end = int(log10(P%EE_end*AU2EV)) 
    endif
    base10EE_start = int(log10(P%EE_start*AU2EV))
    base10exponent = base10EE_start  - base10EE_end
    energy_int = int(P%EE_start*AU2EV) + 1
    first_num = energy_int/10**(base10EE_start)
    if (P%EE_start*AU2EV - dble(first_num * 10**(base10EE_start)) .lt. 1d-3) then
      rest = 0.d0
    else 
      rest = mod(P%EE_start*AU2EV, 10.d0**dble(base10EE_start))
    endif

! Finding how long our energy_in array will be
    if(P%EE_end == 0.d0 .or. base10EE_end == 0) then
      if (rest .ne. 0.d0) then 
        N_energy_in = 1 + 9 * base10exponent + first_num  + 1 - (int(P%EE_end * AU2EV))
      else
        N_energy_in = 1 + 9 * base10exponent + first_num - (int(P%EE_end * AU2EV))
      endif
    else
      if (rest .ne. 0.d0) then 
        N_energy_in =  9 * base10exponent + first_num + 1 - (int(P%EE_end * AU2EV / 10**(base10EE_end))&
                       & - 1 )
      else
        N_energy_in = 9 * base10exponent + first_num - (int(P%EE_end * AU2EV / 10**(base10EE_end)) - 1)
      endif
    endif

! Failsafe for when N_energy_in is negative or zero
    if (N_energy_in .lt. 0) then 
      write(STDERR, *) "The number of incoming-energy steps is N_energy_in =",  N_energy_in
    endif

    allocate(energy_in_array(N_energy_in))
    energy_in_array(:) = P%EE_end
    index_cutoff = 2

! Finding the rest associated to P%EE_end
    if (P%EE_end*AU2EV - dble((int(P%EE_end*AU2EV) + 1)/10**(base10EE_end) * 10**(base10EE_end)) &
        & .lt. 1d-3) then
      restEE_end = 0.d0
    else 
      restEE_end = mod(P%EE_end*AU2EV, 10.d0**dble(base10EE_end))
    endif

! First two elements
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(1)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
      energy_in_array(2) =  energy_in_array(1) + dble(10**(base10exponent))*EV2AU - restEE_end * EV2AU
    enddo

! The rest of energy_in_array
    11 continue
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(index_cutoff - 1 + i)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
        energy_in_array(index_cutoff - 1 + i + 1) =  energy_in_array(index_cutoff) + (i) * & 
                                             &dble(10**(base10exponent))*EV2AU
      if (int(log10(energy_in_array(index_cutoff - 1 + i + 1)*AU2EV)) .ne. base10exponent) then
        index_cutoff = index_cutoff + i
        goto 11                
      endif
    enddo

! The last two elements
    energy_in_array(N_energy_in) = P%EE_start
    if (rest .ne. 0.d0) then
      energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - rest*EV2AU
    else 
      base10exponent = int(log10(energy_in_array(N_energy_in)*AU2EV))
      if (base10exponent .ne. int(log10(energy_in_array(N_energy_in-2)*AU2EV)) .and. first_num == 1)then
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent - 1))*EV2AU
      else 
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent))*EV2AU
      endif      
    endif

    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies (in eV) is as follows :"
    do i = 1, N_energy_in
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo

! Failsafe for negative incoming energies
    do i = 1, N_energy_in
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 

!-------------------------------------------------------------------------------------------------------
! Energy_in grid; fine grid. Use this if you want to resolve some fine-structure at a certain energy 
! range.  
  elseif(grid_type == 1) then 
    energy_in_range = P%EE_start - P%EE_end    

    base10exponent= int(log10(energy_in_range*AU2EV))
    energy_int = int(energy_in_range*AU2EV) + 1
    first_num = energy_int/10**(base10exponent)
    if (energy_in_range*AU2EV - dble(first_num * 10**(base10exponent)) .lt. 1d-3) then
      rest = 0.d0
    else 
      rest = mod(energy_in_range*AU2EV, 10.d0**dble(base10exponent))
    endif
    if (rest .ne. 0.d0) then 
      N_energy_in = 1 + 9 * base10exponent + first_num  + 1 
    else
      N_energy_in = 1 + 9 * base10exponent + first_num 
    endif

!   Failsafe for when N_energy_in is negative or zero
    if (N_energy_in .lt. 0) then 
      write(STDERR, *) "The number of incoming-energy steps is N_energy_in =",  N_energy_in
    endif

    allocate(energy_in_array(N_energy_in))
    energy_in_array(:) = ZERO
    index_cutoff = 1

    10 continue
    do i = 1 , N_energy_in - 1 - index_cutoff
      if (energy_in_array(index_cutoff - 1 + i)*AU2EV == 0.d0) then
        base10exponent = 0
      else 
        base10exponent = int(log10(energy_in_array(index_cutoff - 1 + i)*AU2EV))
      endif
        energy_in_array(index_cutoff - 1 + i + 1) =  energy_in_array(index_cutoff) + (i) * & 
                                             &dble(10**(base10exponent))*EV2AU
      if (int(log10(energy_in_array(index_cutoff - 1 + i + 1)*AU2EV)) .ne. base10exponent) then
        index_cutoff = index_cutoff + i
        goto 10                
      endif
    enddo

    energy_in_array(N_energy_in) = P%EE_start - P%EE_end
    if (rest .ne. 0.d0) then
      energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - rest*EV2AU
    else 
      base10exponent = int(log10(energy_in_array(N_energy_in)*AU2EV))
      if (base10exponent .eq. int(log10(energy_in_array(N_energy_in-2)*AU2EV))) then
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent))*EV2AU
      else 
        energy_in_array(N_energy_in - 1) = energy_in_array(N_energy_in) - &
                                           & dble(10**(base10exponent - 1 ))*EV2AU
      endif      
    endif
    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies is as follows (in eV):"
    do i = 1, N_energy_in
      energy_in_array(i) = energy_in_array(i) + P%EE_end
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo

! Failsafe for negative incoming energies
    do i = 1, N_energy_in
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 

!-------------------------------------------------------------------------------------------------------
! Energy_in grid: linear grid. 
  elseif(grid_type == 2) then
    N_energy_in = P%imp_ion_n_in
    d_energy_in = P%imp_ion_d_in * EV2AU
    energy_in_range = P%EE_start - P%EE_end    
    if (abs(energy_in_range - N_energy_in * d_energy_in) .gt. 1.d-3) then
      write(STDERR, *) "Check your values for N_energy_in and d_energy_in. I have", N_energy_in, "and",&
                       & d_energy_in * AU2EV, "eV."
      stop
    endif
    allocate(energy_in_array(N_energy_in + 1))
    energy_in_array(:) = ZERO
    do i = 1, N_energy_in + 1
      energy_in_array(i) = P%EE_end + dble(i-1) * d_energy_in 
    enddo
    write(OUTPUT,"('#',a, i10)") "The number of incoming energies considered is:", N_energy_in
    write(OUTPUT,"('#',a)") "The array of incoming energies (in eV) is as follows :"
    do i = 1, N_energy_in + 1
      write(OUTPUT,"('#',99ES13.3)") energy_in_array(i)*AU2EV
    enddo
    N_energy_in = N_energy_in + 1

! Failsafe for negative incoming energies
    do i = 1, N_energy_in 
      if (energy_in_array(i) .lt. 0d0) then
        write(STDERR, *) "Element number", i, "in the incoming energy array is negative: energy_in =", &
                          &energy_in_array(i)
        stop
      endif
    end do 
!-------------------------------------------------------------------------------------------------------
  endif
!=End of energy_in grid=================================================================================

! Since using a single N_energy_out for each energy_in is sufficient, this parameter is useless now.
  !energy_out_fine_scaling = P%imp_ion_energy_scaling  ! Determines how fine the energy grid is for 
                                                      ! integration. Default at ONE.

! Writing some declaration output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ', a)")  'It calculates the total cross-section (CS) for a range of & 
                              &incoming energies in two different ways:' 
  write(OUTPUT,"('# ', a)")  'first by an angle integration&
                              & of the result of the Manual method, then from the full DDCS calculation&
                              &, doing the energy integral first and the angle integral second.' 
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a)")  'Columns 4 to (4 - 1 + number of subshells under consideration) are the &
                              &orbital contributions (1s, 2s, 2p, 3s,...) of the Manual method. The &
                              & ensuing columns are the orbital contributions of the DDCS method.' 
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, I5)")'The number of points in the energy_out grid was:', &
                                                                                     &P%imp_ion_n_energy
  write(OUTPUT,"('# ', a, I5)")'The number of points in the angular grid was:', N_theta
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', 1a15, 1a20, 1a20, 1a30 )") 'Energy_in [eV]','CS Manual [Mb]','CS from DDCS [Mb]'&
                                                   &, 'Orbital contributions'

!=Total cross section calculations======================================================================
  do i = 1, N_energy_in

! Just print out 0 if there was no ionization
    if (energy_in_array(i) + orbital_energies(1) .lt. 0.d0) then
      total_CS_manual_orbi(:) = ZERO
      total_CS_manual = ZERO
      total_CS_from_DDCS_orbi(:) = ZERO
      total_CS_from_DDCS = ZERO
      write(OUTPUT,'(1ES15.3, 1ES18.3, 1ES18.3, 1ES30.3, 99ES15.3)') energy_in_array(i) * AU2EV, &
                                                         &total_CS_manual, total_CS_from_DDCS, & 
                                                         &total_CS_manual_orbi(:), &
                                                         &total_CS_from_DDCS_orbi(:)
      cycle
    endif

!==Manual method (don't need energy_out for that)=======================================================
    DCS_manual_orbi(:,:) = ZERO
    temp_storage(:,:) = ZERO
    total_CS_manual_orbi(:) = ZERO
    do m = 1, N_theta                        
      DCS_manual_orbi(m,:) = calculate_impact_ionization_DCS_angle_res_manual(P,F,energy_in_array(i)&
                             &,theta_rad_array(m))
      temp_storage(m,:) = DCS_manual_orbi(m,:) * costheta%weight(N_theta - m + 1)
    end do 
    do j = 1, F%N_orb    
      total_CS_manual_orbi(j) = (TWO * PI) *sum( temp_storage(1:N_theta, j) )
      total_CS_manual_orbi(j) = total_CS_manual_orbi(j)* AU2M**2 / 1.0e-22_long  ! converted to megabarn
    end do
    total_CS_manual = sum(total_CS_manual_orbi(1:F%N_orb))

!==Calculation of DDCS==================================================================================

!--Energy_out grid--------------------------------------------------------------------------------------
! Between edges a nonuniform grid is defined, and these are glued together in the total energy_out 
! array.

    N_energy_out = P%imp_ion_n_energy 
    if (N_energy_out .lt. 10) then              ! Let's assume you need at least 10 energy grid points
      N_energy_out = 10                         ! to make an energy integration for any incoming energy.
    end if

!  Now comes a bunch of arbitrarily defined number of steps that will yield converged results.
  !  if (energy_in_array(i) .le. 100.d0*EV2AU .and. energy_in_array(i) .gt. 10.d0*EV2AU) then    
  !    N_energy_out = max(P%imp_ion_n_energy, 100) * energy_out_fine_scaling                
  !  else if (energy_in_array(i) .le. 1000.d0*EV2AU .and. energy_in_array(i) .gt. 100.d0*EV2AU) then   
  !    N_energy_out = max(P%imp_ion_n_energy, 500) * energy_out_fine_scaling                    
  !  else if (energy_in_array(i) .le. 10000.d0*EV2AU .and. energy_in_array(i) .gt. 1000.d0*EV2AU) then   
  !    N_energy_out = max(P%imp_ion_n_energy, 1000) * energy_out_fine_scaling                    
  !  else if (energy_in_array(i) .le. 50000.d0*EV2AU .and. energy_in_array(i) .gt. 10000.d0*EV2AU) then   
  !    N_energy_out = max(P%imp_ion_n_energy, 10000) * energy_out_fine_scaling                    
  !  endif

! We go from 0 eV to the outer-most valence edge.
    energy_out_range = energy_in_array(i) + orbital_energies(1)

! Determining the number of edges and how many points will be assigned between two edges
    number_of_edges = F%N_orb
    do a = 1, F%N_orb
      if (energy_in_array(i) + orbital_energies(a) .lt. 0.d0) then
        number_of_edges = number_of_edges - 1
      endif
    end do

    if (number_of_edges .eq. 0) then
      write (STDERR,*) "We should not have been able to have gotten here."
    endif                        
    allocate(fraction_energy_out_range(number_of_edges))
    allocate(N_energy_out_orbi(number_of_edges))
    fraction_energy_out_range(:) = 1.d0  ! If there's 1 edge, we don't need to split up the energy array
    if (number_of_edges .ge. 2) then
      do a = 1, number_of_edges - 1
        fraction_energy_out_range(a) = abs(orbital_energies(a)-orbital_energies(a+1)) / energy_out_range
      end do
      fraction_energy_out_range(number_of_edges) = (energy_in_array(i) + &
                                                   &orbital_energies(number_of_edges))/ energy_out_range
    endif

    do a = 1, number_of_edges
      N_energy_out_orbi(a) = nint(fraction_energy_out_range(a) * N_energy_out)
      if (N_energy_out_orbi(a) .lt. 20) then   ! to avoid N_energy_out_orbi = 0, let's give it a 
        N_energy_out_orbi(a) = 20              ! minimal value of 20.
      endif
    enddo
 
    N_energy_out = sum(N_energy_out_orbi(1:number_of_edges))  ! This is the new N_energy_out
    allocate(energy_out_array(N_energy_out))
    allocate(weight_array(N_energy_out))

! Defining energy_out_array
    energy_out_array(:) = ZERO
    weight_array(:) = ZERO

! Special case of when number_of_edges = 1
    if (number_of_edges .eq. 1) then
      E_G = generate_Grid("LG", N_energy_out, ZERO, energy_in_array(i) + orbital_energies(1), 5.d-1)
      do a = 1, N_energy_out
        energy_out_array(a) = abs(energy_in_array(i) + orbital_energies(1) - E_G%r(a))  
        weight_array(a) = E_G%weight(a)
      enddo
      call purge_Grid(E_G)
    else 

! First elements
      E_G = generate_Grid("LG", N_energy_out_orbi(1), energy_in_array(i) + orbital_energies(2), &
                                                       &energy_in_array(i) + orbital_energies(1), 5.d-1)
      do a = 1, N_energy_out_orbi(1)
        energy_out_array(a) =  energy_in_array(i) + orbital_energies(2) + abs(energy_in_array(i) + &
                                                                       & orbital_energies(1) - E_G%r(a))
        weight_array(a) = E_G%weight(a)
      enddo
      call purge_Grid(E_G)

! Last elements
      E_G = generate_Grid("LG", N_energy_out_orbi(number_of_edges), ZERO, energy_in_array(i) + &
                                                             & orbital_energies(number_of_edges), 5.d-1)
      starting_index = sum(N_energy_out_orbi(1:number_of_edges - 1))
      do a = 1, N_energy_out_orbi(number_of_edges)
        energy_out_array(starting_index + a) =  abs(energy_in_array(i) + &
                                                          &orbital_energies(number_of_edges) - E_G%r(a))
        weight_array(starting_index + a) = E_G%weight(a)
      enddo
      call purge_Grid(E_G)

! If number_of_edges = 2, we're done, so skip this next part
      if (number_of_edges .eq. 2) goto 100

      do a = 2, number_of_edges - 1
        E_G = generate_Grid("LG", N_energy_out_orbi(a), energy_in_array(i) + orbital_energies(a+1), &
                                                      & energy_in_array(i) + orbital_energies(a), 5.d-1)
        starting_index = sum(N_energy_out_orbi(1:a-1))
        do b = 1 , N_energy_out_orbi(a)
          energy_out_array(starting_index + b) =  energy_in_array(i) + orbital_energies(a+1) + &
                                               &abs(energy_in_array(i) + orbital_energies(a) - E_G%r(b))
          weight_array(starting_index + b) = E_G%weight(b)
        enddo
        call purge_Grid(E_G)
      enddo
   
      100 continue

      do a = 1, N_energy_out
        if(energy_out_array(a) .lt. 0.d0) then
          write(STDERR, *) "One of the outgoing energies negative. Energy_in = ", energy_in_array(i)& 
                                             &*AU2EV, "energy_out = ", energy_out_array(a)*AU2EV
          stop
        endif
      enddo
       
    endif


!--End of energy_out grid-------------------------------------------------------------------------------

! Initialization

    allocate(DDCS_orbi(N_theta, N_energy_out, F%N_orb))
    DDCS_orbi(:,:,:) = ZERO
    allocate(DDCS(N_theta, N_energy_out))
    DDCS(:,:) = ZERO

! DDCS calculation
    do a = 1, N_theta
      do b = 1, N_energy_out
        DDCS_orbi(a,b,:) = calculate_impact_ionization_DDCS(P,F, theta_rad_array(a), & 
                                &energy_in_array(i), energy_out_array(b)) 
      end do 
    end do 

! First energy integration, then angle integration. (Opposite gave the same. I checked.)
    DCS_orbi(:,:) = ZERO
    temp_storage(:,:) = ZERO
    total_CS_from_DDCS_orbi(:) = ZERO
    total_CS_from_DDCS = ZERO
    do m = 1, N_theta
      do j = 1, F%N_orb
        DCS_orbi(m,j) = sum( DDCS_orbi(m,1:N_energy_out,j) * weight_array(1:N_energy_out))
      end do
      temp_storage(m,:) = DCS_orbi(m,:) * costheta%weight(N_theta - m + 1)
    end do 

    do j = 1, F%N_orb    
      total_CS_from_DDCS_orbi(j) = (TWO * PI) *sum( temp_storage(1:N_theta, j))
      total_CS_from_DDCS_orbi(j) = total_CS_from_DDCS_orbi(j)* AU2M**2 / 1.0e-22_long  ! megabarn
    end do
    total_CS_from_DDCS = sum(total_CS_from_DDCS_orbi(1:F%N_orb))
    write(OUTPUT,'(1ES15.3, 1ES18.3, 1ES18.3, 1ES30.3, 99ES15.3)') energy_in_array(i) * AU2EV, & 
                                                         &total_CS_manual, total_CS_from_DDCS, & 
                                                         &total_CS_manual_orbi(:), &
                                                         &total_CS_from_DDCS_orbi(:)
    deallocate(energy_out_array) 
    deallocate(weight_array) 
    deallocate(fraction_energy_out_range)
    deallocate(N_energy_out_orbi)
    deallocate(DDCS_orbi)
    deallocate(DDCS)
  end do 

  deallocate(energy_in_array)
end subroutine print_impact_ionization_total_cross_section

!======================================================================================================= 

!======================================================================================================= 
!--------------------DDCS without considering edges, and symmetric DCS(energy)--------------------------
!======================================================================================================= 

function calculate_impact_ionization_DDCS_edgeless(P,F, theta_rad, energy_in, energy_out) &
                                                                                  &result(DDCS_orbital)
  type(Param), intent(in)                 :: P               
  type(WF_data), intent(in)               :: F
  real(long), intent(in)                  :: theta_rad
  real(long), intent(in)                  :: energy_in
  real(long), intent(in)                  :: energy_out
  real(long)                              :: f_n(F%N * (P%lmax + 1))   ! Too big 
  integer                                 :: E_index(P%lmax + 1), number_of_neg(P%lmax + 1)
  integer                                 :: E_cutoff_lies_in_epsilon_f_range(P%lmax + 1) 
  real(long)                              :: epsilon_f(F%N * (P%lmax + 1)), epsilon_i, E_cutoff
  integer                                 :: L_max
  real(long), allocatable, dimension(:,:) :: j_bes  
  real(long)                              :: Q
  integer, dimension(F%N_all)             :: occ_all
  integer                                 :: n_i, l_i, occ_i, i_all
  integer                                 :: n_f, l_f, n_all, L, column
  real(long)                              :: sum_1, CG, R_integrand, factor
  real(long), external                    :: cleb
  !integer                                 :: index_all
  !integer                                 :: slope_index, slope_all
  integer                                 :: i, a
  !integer                                 :: j, b
  real(long)                              :: histogram(F%N * (P%lmax + 1)) ! Too big 
  real(long)                              :: slope_at_E
  real(long)                              :: DDCS_orbital(F%N_orb)

  DDCS_orbital(:) = ZERO

! This prepares all initial occupation numbers
  occ_all(:) = 0
  do i = 1, F%N_orb
    n_i   = F%config(i,1)              
    l_i   = F%config(i,2)              
    occ_i = F%config(i,3)             
    i_all   = F%N * l_i + (n_i-l_i)     
    occ_all(i_all) = occ_i    
  end do

! Preparing all final orbital energies
  do l_f = 0, P%lmax 
    do n_f = 1, F%N
      n_all = F%N * l_f + n_f
      epsilon_f(n_all) = F%e_all(n_all)
    end do
  end do

! Number of negative energies epsilon_f per l_f. 
! It allows us to access the positive epsilon_f's when used in an index
  number_of_neg(:) = ZERO                    
  do l_f = 0, P%lmax 
    do n_f = 1, F%N
      n_all = F%N * l_f + n_f
      if (epsilon_f(n_all) .lt. 0.d0) then
        number_of_neg(l_f + 1) = number_of_neg(l_f + 1) + 1
      endif
    end do
  end do
 
  L_max = maxval( F%config(:,2) ) + P%lmax + 1 
  allocate(j_bes(F%N,L_max))

  orbitals: do i = 1, F%N_orb

    n_i   = F%config(i,1)                     
    l_i   = F%config(i,2)     
    i_all   = F%N * l_i + (n_i-l_i)
    epsilon_i = 0.d0 

    E_cutoff = energy_in - energy_out + epsilon_i 

! E_index is the index value of the epsilon_f (per l_f) which is the smallest value larger than E_cutoff
    E_index(:) = 1

    E_cutoff_lies_in_epsilon_f_range(:) = 1    ! 1 means yes, 0 means no

! Determining the cutoff in n_f per l_f
    do l_f = 0, P%lmax
      if (E_cutoff .lt. epsilon_f(F%N * l_f + 1 + number_of_neg(l_f + 1))) then    ! We don't want 
        E_cutoff_lies_in_epsilon_f_range(l_f + 1) = 0       ! scenarios where the smallest epsilon_f(>0)
      endif                                                 ! is larger than the cutoff energy. 
      do n_f = 1, F%N
        n_all = F%N * l_f + n_f
        if (n_all .eq. F%N * (P%lmax + 1)) cycle
        if (epsilon_f(n_all) .lt. -1.d-6) cycle
        if (E_cutoff .gt. epsilon_f(F%N * l_f + F%N)) then ! If the cutoff energy is larger than any 
          E_index(l_f + 1) = F%N                          ! outgoing energy considered (very unlikely),  
          exit                                           ! approximate with the largest energy available
        endif
        if ( (E_cutoff .gt. epsilon_f(n_all)) .and. (E_cutoff .le. epsilon_f(n_all + 1)) ) then
          E_index(l_f + 1) = n_f + 1            ! Index 0 doesn't make sense. Hence the +1.
          exit                                             
        endif                                            
      end do 
    enddo
    
    do l_f = 0, P%lmax
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .lt. 1 .and. &
                                                &E_cutoff_lies_in_epsilon_f_range(l_f + 1) .eq. 1 ) then   
        write(STDERR,*) "One of the E_index values is less than one, somehow"    
        stop 
      endif                                              
    enddo                                                  

! Calculating the DDCS
    j_bes(:,:) = ZERO
    f_n(:) = ZERO
    do l_f = 0, P%lmax
      if (E_cutoff_lies_in_epsilon_f_range(l_f + 1) .ne. 1) cycle
      do n_f = 1, E_index(l_f + 1) 
        n_all = F%N * l_f + n_f
        if (epsilon_f(n_all) .lt. 0) cycle    ! Only b2free contributions
        if (energy_in + epsilon_i - epsilon_f(n_all) .lt. 0.d0) cycle ! This for makes the difference 
                                                                      ! between 4min and 41 seconds. 
        if (i_all == n_all) cycle   ! Don't include "transitions" to the same state
        sum_1 = ZERO                     
        Q = momentum_transfer(energy_in, energy_out, theta_rad)  
        call cfelt_spherical_bessel(L_max, Q, F%r, F%N, j_bes)
        do L = 0, l_i+l_f    
          CG = cleb(2*l_i, 0, 2*l_f, 0, 2*L, 0) 
          if (CG == 0 .or. mod(l_i+l_f+L, 2) == 1 ) cycle 
          column = L+1                  
          R_integrand = radial_integral_u_all(F, i_all, n_all, j_bes(:,column)) 
          sum_1 = sum_1 + (R_integrand * CG)**2      ! This does the sum over L         
        end do
        factor = (TWO * dble(l_f) + ONE) * dsqrt(energy_out/energy_in)  
        f_n(n_all) = Rutherford_CS(energy_in, energy_out, theta_rad) * occ_all(i_all) * factor * sum_1 
      end do    
    end do 

    histogram(:) = ZERO
    slope_at_E = ZERO
    final_l: do l_f = 0, P%lmax 
!   Calculating the histograms

!   If there's only one point of epsilon_f between 0 and E_cutoff, we will pretend that there is a 
!   histogram of height zero at epsilon_f = 0, and take the slope with respect to that. 
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .eq. 1) then          
        histogram(F%N * l_f + 1) = 0.d0      
        histogram(F%N * l_f + 2) =f_n(F%N * l_f + 1 + number_of_neg(l_f + 1))  
      else
        do n_f = 1, E_index(l_f + 1)
          n_all = F%N * l_f + n_f
          histogram(n_all) = sum (f_n(F%N * l_f + 1:F%N * l_f + n_f))   ! sum_{n_f} 
        end do
      endif

!   Calculating the slope between the last two histograms
      if (E_index(l_f + 1) - number_of_neg(l_f + 1) .eq. 1) then      
        E_index(l_f + 1) = 2   
        slope_at_E = slope_at_E + ( histogram(F%N * l_f + E_index(l_f + 1)) - & 
                                    &histogram(F%N * l_f + E_index(l_f + 1) - 1 ) ) / &
                            & abs( epsilon_f(F%N * l_f + E_index(l_f + 1) + number_of_neg(l_f + 1)) - & 
                                  &epsilon_f(F%N * l_f + E_index(l_f + 1)- 1) + number_of_neg(l_f + 1) ) 
      else

!   Making sure we have at least two points to interpolate between
        if (E_index(l_f + 1) - number_of_neg(l_f + 1) .lt. 2 .and. &
                                                & E_cutoff_lies_in_epsilon_f_range(l_f + 1) .eq. 1) then          
          write(STDERR,*) "The index of the cutoff energy should be larger than 1. It isn't."
          stop
        endif
        slope_at_E = slope_at_E + ( histogram(F%N * l_f + E_index(l_f + 1)) - & 
                                    &histogram(F%N * l_f + E_index(l_f + 1) - 1 ) ) / &
                                    & abs( epsilon_f(F%N * l_f + E_index(l_f + 1)) - & 
                                    &epsilon_f(F%N * l_f + E_index(l_f + 1)- 1) ) 
      endif
    end do final_l

    DDCS_orbital(i) = slope_at_E
    if (DDCS_orbital(i) .lt. 0.d0 .and. DDCS_orbital(a) .gt. -1.d-100) then
      DDCS_orbital(i) = 0.d0
    endif

  end do orbitals
  deallocate(j_bes)

end function calculate_impact_ionization_DDCS_edgeless

!======================================================================================================= 

subroutine print_impact_ionization_DCS_energy_res_symm(P,F)
! Called using -imp_ion_energy_resolved. Prints the angle-integrated DCS for a range of energy_out.
  type(Param), intent(in)              :: P               
  type(WF_data), intent(in)            :: F
  type(Grid)                           :: costheta
  integer                              :: OUTPUT
  integer                              :: N_theta
  integer                              :: N_energy_out
  real(long)                           :: energy_out_array(P%imp_ion_N_energy)
  real(long)                           :: orbital_energies(F%N_orb), temp 
  real(long)                           :: theta_rad_array(P%imp_ion_N_theta)
  real(long)                           :: energy_in_start, energy_in_end
  real(long)                           :: d_energy_au, d_energy_eV
  integer                              :: a, b, i, j, m
  real(long), allocatable, dimension(:, :,:):: DDCS_orbital, DDCS_orbital_edgeless
  real(long), allocatable, dimension(:,:,:) :: temp_storage, temp_storage_edgeless
  real(long), allocatable, dimension(:,:)   :: DCS_energy, DCS_energy_edgeless
  real(long), allocatable, dimension(:)     :: DCS_energy_tot, DCS_energy_tot_edgeless
  real(long), allocatable, dimension(:,:)   :: DCS_energy_symm
  real(long), allocatable, dimension(:)     :: DCS_energy_tot_symm
                                  
  OUTPUT = P%OUTPUT 

  do i = 1, F%N_orb
    orbital_energies(F%N_orb - i + 1) = F%e_orb(i) ! These are now in increasing order, ending with 1s.
  end do

! Angular grid
  N_theta = P%imp_ion_N_theta         
  costheta = generate_Grid("LG", N_theta, -ONE, ONE, 5.d-1)
  do m = 1, N_theta
    theta_rad_array(m) = abs(PI-acos(costheta%r(N_theta - m + 1)))   ! this now goes from ZERO to PI
  enddo                                                              ! most dense points on grid near 0.

! Energy_out grid
  energy_in_start = P%EE_start

  N_energy_out = P%imp_ion_N_energy

  d_energy_eV = P%imp_ion_d_energy 
  d_energy_au = d_energy_eV * EV2AU

  energy_in_end = P%EE_end
! energy_in_end will only define until what point the outgoing energy array energy_out will go. So:

! 0--|--|--|--|--5--|--|...|--|--energy_in_end--|--|...|--energy_in_start.

! NOTE: it is assumed energy_in_start is NOT a part of the energy_out grid. (Inelasticity) 

! Failsafe for when energy_out is out of range. 
  if (energy_in_start - N_energy_out * d_energy_au .lt. energy_in_end-1.0d-6) then 
    write(STDERR, *) 'Check your choices for energy_in_start, N_energy, and d_energy.'
    stop
  endif

  energy_out_array(:) = ZERO
  do a = 1, N_energy_out
    energy_out_array(a) = energy_in_end + dble(a-1) * d_energy_au
  end do 

! Sorting the array in decreasing order
  do a = 1, N_energy_out
    do b = a, N_energy_out
      if (energy_out_array(a) .lt. energy_out_array(b)) then 
        temp = energy_out_array(b)
        energy_out_array(b) = energy_out_array(a)
        energy_out_array(a) = temp
      endif
    enddo
  enddo

! Writing standard output
  write(OUTPUT,"('# ')")
  write(OUTPUT,"('# ', a)") 'This is the impact ionization module using the non-uniform grid.'
  write(OUTPUT,"('# ', a)")  'It calculates the differential cross-section (DCS) for a range of& 
                              & scattered energy energy_out. An angle integration was performed'
  write(OUTPUT,"('# ', a)")  ' '
  write(OUTPUT,"('# ', a, f10.3)") 'The incoming energy of the electron is (in eV):', &
                                    &P%EE_start*real(AU2EV, 4)
  write(OUTPUT,"('# ', a)") ' '
  write(OUTPUT,"('# ', 1a18, 1a23, 1a35 )") 'Energy_out(eV)', 'Energy-resolved DCS', &
                                            &'Orbital contributions to DCS'

! DDCS calculation
  allocate(DDCS_orbital(N_theta,N_energy_out,F%N_orb))
  allocate(temp_storage(N_theta,N_energy_out,F%N_orb))
  allocate(DCS_energy(N_energy_out,F%N_orb))
  allocate(DCS_energy_tot(N_energy_out))

  allocate(DDCS_orbital_edgeless(N_theta,N_energy_out,F%N_orb))
  allocate(temp_storage_edgeless(N_theta,N_energy_out,F%N_orb))
  allocate(DCS_energy_edgeless(N_energy_out,F%N_orb))
  allocate(DCS_energy_tot_edgeless(N_energy_out))

  allocate(DCS_energy_symm(N_energy_out,F%N_orb))
  allocate(DCS_energy_tot_symm(N_energy_out))

  DDCS_orbital(:,:,:) = ZERO
  do i = 1, N_theta    
    do j = 1, N_energy_out
      DDCS_orbital(i,j,:) = calculate_impact_ionization_DDCS(P,F, theta_rad_array(i), energy_in_start, &
                     &energy_out_array(j)) 
      DDCS_orbital_edgeless(i,j,:) = calculate_impact_ionization_DDCS_edgeless(P,F, theta_rad_array(i),& 
                                                                  &energy_in_start, energy_out_array(j)) 
      do a = 1, F%N_orb
        if (DDCS_orbital(i,j,a) .lt. 0.d0) then
          write(STDERR,*) "Negative DDCS", theta_rad_array(i), energy_out_array(j)*AU2EV, a, &
                                                                                    &DDCS_orbital(i,j,a)
        endif
      enddo

      do a = 1, F%N_orb
        if (DDCS_orbital_edgeless(i,j,a) .lt. 0.d0) then
          write(STDERR,*) "Negative (edgeless) DDCS", theta_rad_array(i), energy_out_array(j)*AU2EV, a,&
                                                                           &DDCS_orbital_edgeless(i,j,a)
        endif
      enddo

    end do 
  end do  

  temp_storage(:,:,:) = ZERO
  do m = 1, N_theta
    temp_storage(m,:,:) = DDCS_orbital(m, :, :)*costheta%weight(N_theta - m + 1)
  enddo

  temp_storage_edgeless(:,:,:) = ZERO
  do m = 1, N_theta
    temp_storage_edgeless(m,:,:) = DDCS_orbital_edgeless(m, :, :)*costheta%weight(N_theta - m + 1)
  enddo

! Angular integration
  do i = 1, N_energy_out
    do j = 1, F%N_orb
      DCS_energy(i,j) = (TWO * PI) * sum(temp_storage(1:N_theta,i,j))
      DCS_energy_edgeless(i,j) = (TWO * PI) * sum(temp_storage_edgeless(1:N_theta,i,j))

      if (DCS_energy(i,j) .lt. 0.d0 ) then
        write(STDERR,*) "Negative DCS", energy_out_array(i)*AU2EV, j, DCS_energy(i,j)
        stop
      endif

      if (DCS_energy_edgeless(i,j) .lt. 0.d0 ) then
        write(STDERR,*) "Negative (edgeless) DCS", energy_out_array(i)*AU2EV, j, &
                                                                              &DCS_energy_edgeless(i,j)
        stop
      endif
 
    end do 
    DCS_energy_tot(i) = sum(DCS_energy(i,1:F%N_orb))
    DCS_energy_tot_edgeless(i) = sum(DCS_energy_edgeless(i,1:F%N_orb))
  end do

! Reversing the order of the edgeless DCS, and adding it to the original DCS
  do i = 1, N_energy_out
    do j = 1, F%N_orb
      DCS_energy_symm(i,j) = DCS_energy(i,j) + DCS_energy_edgeless(N_energy_out - i + 1,j)
    end do 
    DCS_energy_tot_symm(i) = sum(DCS_energy_symm(i,1:F%N_orb))
    write(OUTPUT,'(2ES18.3, 1ES25.3, 99ES13.3)') energy_out_array(i) * AU2EV, DCS_energy_tot_symm(i), &
                                                 &DCS_energy_symm(i,:)
  end do

  deallocate(DDCS_orbital)
  deallocate(temp_storage)
  deallocate(DCS_energy)
  deallocate(DCS_energy_tot)

  deallocate(DDCS_orbital_edgeless)
  deallocate(temp_storage_edgeless)
  deallocate(DCS_energy_edgeless)
  deallocate(DCS_energy_tot_edgeless)

  deallocate(DCS_energy_symm)
  deallocate(DCS_energy_tot_symm)

end subroutine print_impact_ionization_DCS_energy_res_symm

!======================================================================================================= 

end module impact_ionization
