!----------------------------------------------------------------------------
! matrix.f90
!----------------------------------------------------------------------------
! Copyright (C) 2004 Sang-Kil Son
! Copyright (C) 2010 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Matrix Module
!   programmed by Son, Sang-Kil in Feb. 2004
!   for PHSX815 : Computational Physics
!
! Ref)
!   Press et al., "Numerical Recipes in Fortran 77"
!   Press et al., "Numerical Recipes in Fortran 90"
!
! Main data types
! - LU_matrix : LU-decomposed matrix
!   - F LU_decompose(A(N,N)) : create LU matrix by LU-decomposition of A
!   - F LU_solve(LU, b(N)) : solve a linear equation for b with LU
!   - F LU_inverse(LU) : inverse matrix of LU
!   - F LU_determinant(LU) : determinant of LU
!   - F LU_lower(LU) : return L matrix of LU
!   - F LU_upper(LU) : return U matrix of LU
!
! - TRD_matrix : tridiagonal matrix
!   - F TRD_create : create tridiagonal matrix from matrix or vectors
!     . TRD_create_from_vector(a, b, c)
!     . TRD_create_from_matrix(A)
!   - F TRD_restore(TRD) : restore original matrix from tridiagonal matrix
!   - F TRD_solve(TRD, b(N)) : solve a linear equation for b with TRD
!   - SR TRD_eigen(TRD, l(N), v(N,N)) : solve eigenvalues and eigenvertors
!                                       for symmetric tridiagonal matrix
!
! Auxiliary functions and subroutines
! - SR print_matrix(A(N,N)) : print out a matrix
! - SR print_vector(x(N)) : print out a vector
!----------------------------------------------------------------------------
module matrix
  use constants
  implicit none
  !
  ! Data type: LU_matrix 
  !   it contains the dimension of the matrix, N
  !   and LU-decomposed matrix, LU(N,N)
  !   also, permutation(N) and the sign for determinant, d
  !
  type :: LU_matrix
    integer :: n
    real(long), dimension(:,:), allocatable :: LU
    integer, dimension(:), allocatable :: permutation
    integer :: d
  end type LU_matrix
  !
  ! Data_type: TRD_matrix
  !   n : the dimension of the matrix
  !   a : sub-diagonal values, a(i), i=2, ..., n
  !   b : main diagonal values,  b(i), i=1, ..., n
  !   c : super-diagonal values, c(i), i=1, ..., n-1
  !
  type :: TRD_matrix
    integer :: n
    real(long), dimension(:), allocatable :: a, b, c
  end type TRD_matrix
  !
  ! interfaces: it can extend F ans SR with any kind of arguments
  !
  interface TRD_create
    module procedure TRD_create_from_vectors, &
                     TRD_create_from_matrix, &
                     TRD_create_from_values
  end interface
contains

function LU_decompose(A) result (M)
!
! Carry out LU-decomposition for A and put the decomposed matrix into M
!   M: type(LU_matrix)
!   A: N*N matrix
! This program is implementation of Crout's method with partial pivoting.
!
  type(LU_matrix) :: M
  real(long), dimension(:,:), intent(in) :: A
  real(long), dimension(size(A,dim=1)) :: scaling_factor, temp_row
  real(long) :: a_max, temp
  integer :: n, i, j, i_max
  !
  ! Initialize
  !   A is a n * n matrix.
  !   # rows = size(A, dim=1)
  !   # columns = size(A, dim=2)
  ! consequently, size(A,dim=1) = size(A,dim=2)
  !
  n = size(A, dim=1)
  M%n = n
  M%d = 1
  allocate( M%LU(n, n), M%permutation(n) )
  M%LU = A
  !
  ! Implicit scaling procedure
  !
  do i = 1, n
    a_max = maxval( abs( A(i,:) ) )
    if ( a_max < EPS ) stop "singular matrix in LU_decompose"
    scaling_factor(i) = ONE / a_max
  end do
  !
  ! Crout's method
  ! The following equations are exactly same as Press's procedure
  !
  do j = 1, n
    !
    ! equation (2.3.12) except for i=j
    !
    do i = 1, j-1
      M%LU(i,j) = M%LU(i,j) - sum( M%LU(i,1:i-1) * M%LU(1:i-1,j) )
    end do
    !
    ! equation (2.3.12) for i=j and equation (2.3.13)
    !
    a_max = ZERO
    do i = j, n
      M%LU(i,j) = M%LU(i,j) - sum( M%LU(i,1:j-1) * M%LU(1:j-1,j) )
      !
      ! find the maximum element and store its point (i_max)
      ! That point will be the pivot element.
      !
      temp = abs( M%LU(i,j) ) * scaling_factor(i)
      if ( temp > a_max ) then
        i_max = i
        a_max = temp
      end if
    end do
    !
    ! if needed to interchange rows
    !
    if ( j /= i_max ) then
      temp_row = M%LU(j,:)              ! swap j-th row and i_max-th row
      M%LU(j,:) = M%LU(i_max,:)         ! A(j,:) : j-th row
      M%LU(i_max,:) = temp_row
      M%d = -M%d
      scaling_factor(i_max) = scaling_factor(j)
    end if
    !
    ! divide by the pivot element
    !
    M%permutation(j) = i_max
    if ( abs( M%LU(j,j) ) < EPS ) M%LU(j,j) = EPS    ! prevent singular case
    if ( j /= n ) M%LU(j+1:n,j) = M%LU(j+1:n,j) / M%LU(j,j)
  end do
end function LU_decompose


function LU_solve(M, b) result (x)
!
! Solve a linear equation for b with LU-decomposed matrix M
! Return the solution for b as a new vector x
!   b : row vector, not column vector, that is, just 1D array of n
!
  type(LU_matrix), intent(in) :: M
  real(long), dimension(:), intent(in) :: b
  real(long), dimension(M%n) :: x
  real(long) :: temp
  integer :: i, j, n
  n = M%n
  !
  ! rearrange b(n) with the permutation
  !
  x = b
  do i = 1, n
    j = M%permutation(i)
    if ( i /= j ) then
      temp = x(i)
      x(i) = x(j)
      x(j) = temp
    end if
  end do
  !
  ! forward substitution using Press's eq (2.3.6) and alpha_ii = 1
  !
  do i = 1, n
    x(i) = x(i) - sum( M%LU(i,1:i-1) * x(1:i-1) )
  end do
  !
  ! back substitution using Press's eq (2.3.7)
  !
  do i = n, 1, -1
    x(i) = ( x(i) - sum( M%LU(i,i+1:n) * x(i+1:n) ) ) / M%LU(i,i)
  end do
end function LU_solve


function LU_inverse(M) result (X)
!
! Return the inverse matrix of LU-decomposed matrix M
!   using F solve(M, b)
!
  type(LU_matrix), intent(in) :: M
  real(long), dimension(M%n,M%n) :: X
  real(long), dimension(M%n) :: b
  integer :: i, n
  n = M%n
  !
  ! set up identity matrix and solve n linear equations with identity.
  !
  do i = 1, n
    b = ZERO; b(i) = ONE        ! b is the i-th column vector of identity.
    X(:,i) = LU_solve(M, b)     ! A(:,i) is the i-th column vector of A
  end do
end function LU_inverse


function LU_lower(M) result (L)
!
! Return L matrix of LU-decomposed matrix M
!
  type(LU_matrix), intent(in) :: M
  real(long), dimension(M%n,M%n) :: L
  integer :: i, n
  n = M%n
  L = ZERO
  L(1,1) = ONE
  do i=2, n
    L(i,i) = ONE
    L(i,1:i-1) = M%LU(i,1:i-1)
  end do
end function LU_lower


function LU_upper(M) result (U)
!
! Return U matrix of LU-decomposed matrix M
!
  type(LU_matrix), intent(in) :: M
  real(long), dimension(M%n,M%n) :: U
  integer :: i, n
  n = M%n
  U = ZERO
  do i=1, n
    U(i,i:n) = M%LU(i,i:n)
  end do
end function LU_upper


function LU_determinant(M) result (d)
!
! Return the determinant value of LU-decomposed matrix M
!
  type(LU_matrix), intent(in) :: M
  real(long) :: d
  integer :: i
  !
  ! The determinant can be solved by a product of all diagonal elements.
  !
  d = dble(M%d) * product( (/ ( M%LU(i,i), i=1, M%n ) /) )
end function LU_determinant


function TRD_create_from_matrix(A) result (M)
!
! Initialize tridiagonal matrix M from ordinary n x n matrix A
!   A: n x n matrix
!   M: type(TRD_matrix)
!
  type(TRD_matrix) :: M
  real(long), dimension(:,:), intent(in) :: A
  integer :: n, i
  n = size(A, dim=1)
  M%n = n
  allocate( M%a(n), M%b(n), M%c(n) )
  !
  ! a : below diagonal, a(i) = A(i,i-1), i=2, ..., n
  ! b : main diagonal,  b(i) = A(i,i),   i=1, ..., n
  ! c : above diagonal, c(i) = A(i,i+1), i=1, ..., n-1
  ! a(1) and c(n) are useless.
  !
  M%a(1) = ZERO
  M%b(1) = A(1,1)
  M%c(n) = ZERO
  do i = 2, n
    M%a(i)   = A(i,i-1)
    M%b(i)   = A(i,i)
    M%c(i-1) = A(i-1,i)
  end do
end function TRD_create_from_matrix


function TRD_create_from_vectors(a, b, c) result (M)
!
! Initialize tridiagonal matrix M from 3 vectors of diagonal elements
!   a: sub-diagonal (n-1) vector
!   b: main diagonal (n) vector
!   c: super-diagonal (n-1) vector
!   M: type(TRD_matrix)
!
  type(TRD_matrix) :: M
  real(long), dimension(:), intent(in) :: a, b, c
  integer :: n
  n = size(b, dim=1)
  M%n = n
  allocate( M%a(n), M%b(n), M%c(n) )
  !
  ! M%a : sub-diagonal,   M%a(i) = a(i-1), i=2, ..., n
  ! M%b : main diagonal,  M%b(i) = b(i),   i=1, ..., n
  ! M%c : super-diagonal, M%c(i) = c(i), i=1, ..., n-1
  ! M%a(1) and M%c(n) are useless.
  !
  M%a(1) = ZERO
  M%c(n) = ZERO
  M%a(2:n) = a(1:n-1)
  M%b(1:n) = b(1:n)
  M%c(1:n-1) = c(1:n-1)
end function TRD_create_from_vectors


function TRD_create_from_values(a, b, c, n) result (M)
!
! Initialize tridiagonal matrix M of all same diagonal elements from 3 values
!   a: a sub-diagonal value
!   b: a main diagonal value
!   c: a super-diagonal value
!   n : dimension
!   M: type(TRD_matrix)
!
  type(TRD_matrix) :: M
  real(long), intent(in) :: a, b, c
  integer, intent(in) :: n
  M%n = n
  allocate( M%a(n), M%b(n), M%c(n) )
  !
  ! M%a : sub-diagonal,   M%a(i) = a, i=2, ..., n
  ! M%b : main diagonal,  M%b(i) = b, i=1, ..., n
  ! M%c : super-diagonal, M%c(i) = c, i=1, ..., n-1
  ! M%a(1) and M%c(n) are useless.
  !
  M%a(1) = ZERO
  M%c(n) = ZERO
  M%a(2:n) = a
  M%b(1:n) = b
  M%c(1:n-1) = c
end function TRD_create_from_values


function TRD_restore(M) result (A)
!
! Restore original matrix A from tridiagonal matrix M
!   M: type(TRD_matrix)
!   A: n x n matrix
! 
  type(TRD_matrix) :: M
  real(long), dimension(M%n,M%n) :: A
  integer :: i
  A = ZERO
  A(1,1) = M%b(1)
  do i = 2, M%n
    A(i,i-1) = M%a(i)
    A(i,i)   = M%b(i)
    A(i-1,i) = M%c(i-1)
  end do
end function TRD_restore


function TRD_solve(M, r) result (x)
!
! Solve a linear equation for r with tridiagonal matrix M
! Return the solution for r as a new vector x
!   b : row vector, not column vector, that is, just 1D array of n
!
  type(TRD_matrix), intent(in) :: M
  real(long), dimension(:), intent(in) :: r
  real(long), dimension(M%n) :: x, temp
  real(long) :: bet
  integer :: n, j
  n = M%n
  bet = M%b(1)
  if ( M%b(1) == ZERO ) stop "TRD_solve: rewrite equations"
  x(1) = r(1) / bet
  do j = 2, n           ! decomposition and forward substitution
    temp(j) = M%c(j-1) / bet
    bet = M%b(j) - M%a(j) * temp(j)
    if ( bet == ZERO ) stop "TRD_solve: failed"
    x(j) = ( r(j) - M%a(j) * x(j-1) ) / bet
  end do
  do j = n-1, 1, -1     ! backsubstitution
    x(j) = x(j) - temp(j+1) * x(j+1)
  end do
end function TRD_solve


subroutine TRD_eigen(T, d, v)
!
! Find eigenvalues and eigenvectors for symmetric tridoagonal matrix
! Using TQLI (Tridiagonal QL algorithm with Implicit shifts) from Press
!   M: type(TRD_matrix)
!   d(n): eigenvalues
!   v(n,n): eigenvectors, k-th column corresponding to k-th eigenvalue
!
  type(TRD_matrix), intent(in) :: T
  real(long), dimension(:), intent(out) :: d
  real(long), dimension(:,:), intent(out) :: v
  real(long), dimension(T%n) :: e, ff
  real(long) :: b, c, dd, f, g, p, r, s
  integer :: i, iter, l, m, n
  n = T%n
  if ( any( T%a(2:n) /= T%c(1:n-1) ) ) &
    stop "TRD_eigen: requires symmetric matrix"
  if ( n /= size(d) ) stop "TRD_eigen: requires n array for eigenvalues"
  if ( n /= size(v,dim=1) .or. n /= size(v,dim=2) ) &
    stop "TRD_eigen: requires n x n array for eigenvectors"
  !
  ! Initialize
  !
  d = T%b
  e(1:n-1) = T%a(2:n)
  v = ZERO
  do i = 1, n
    v(i,i) = ONE
  end do

  do l = 1, n
    iter = 0
    iterate: do
      !
      ! look for a single small subdiagonal element to split the matrix
      !
      do m = l, n-1
        dd = abs(d(m)) + abs(d(m+1))
        if ( abs(e(m)) + dd == dd ) exit
      end do
      if ( m == l) exit iterate
      if ( iter == 30 ) stop "TRD_eigen: too many iteration"
      iter = iter + 1
      g = (d(l+1) - d(l)) / (TWO * e(l))                ! form shift
      r = pythagoras( g, ONE )
      g = d(m) - d(l) + e(l) / ( g + sign(r,g) )        ! this is d_m - k_s
      s = ONE;  c = ONE;  p = ZERO
      !
      ! A plane rotation as in the original QL, followed by Givens rotations
      ! to restore tridiagonal form.
      !
      do i = m-1, l, -1
        f = s * e(i)
        b = c * e(i)
        r = pythagoras( f, g )
        e(i+1) = r
        if ( r == ZERO ) then           ! Recover from underflow
          d(i+1) = d(i+1) - p
          e(m) = ZERO
          cycle iterate
        end if
        s = f / r
        c = g / r
        g = d(i+1) - p
        r = ( d(i) - g ) * s + TWO * c * b
        p = s * r
        d(i+1) = g + p
        g = c * r - b
        !
        ! form eigenvectors
        !
        ff(1:n) = v(1:n, i+1)
        v(1:n, i+1) = s * v(1:n, i) + c * ff(1:n)
        v(1:n, i)   = c * v(1:n, i) - s * ff(1:n)
      end do
      d(l) = d(l) - p
      e(l) = g
      e(m) = ZERO
    end do iterate
  end do
end subroutine TRD_eigen


subroutine print_matrix(A)
!
! Print arbitrary N*M matrix
!
  real(long), dimension(:,:), intent(in) :: A
  integer :: i
  do i = 1, size(A, dim=1)
    print "(t8,100f10.5)", A(i,:)
  end do
end subroutine print_matrix


subroutine print_vector(x)
!
! Print arbitrary N-vector as if a column vector
!   x : row vector, not column vector, that is, just 1D array of n
!
  real(long), dimension(:), intent(in) :: x
  integer :: i
  do i = 1, size(x)
    print "(t8,f10.5)", x(i)
  end do
end subroutine print_vector


function pythagoras(a, b) result (p)
!
! Compute sqrt(a^2 + b^2) without destructive underflow or overflow
!   which may occur in computation of a^2 and b^2
!
  real(long), intent(in) :: a, b
  real(long) :: p, abs_a, abs_b
  abs_a = abs(a)
  abs_b = abs(b)
  if ( abs_a > abs_b ) then
    p = abs_a * sqrt( ONE + (abs_b / abs_a)**2 )
  else
    if ( abs_b == ZERO ) then
      p = ZERO
    else
      p = abs_b * sqrt( ONE + (abs_a / abs_b)**2 )
    end if
  end if
end function pythagoras

end module matrix
