!----------------------------------------------------------------------------
! operations.f90
!----------------------------------------------------------------------------
! Copyright (C) 2004      Sang-Kil Son
! Copyright (C) 2010-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2017 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! HFS calculation for the atomic system using the GPS discretization
!   rewritten by Son, Sang-Kil in 2010 for XATOM
! Originally this module was for DFT calculations
!   written by Son, Sang-Kil in Nov. 2004 - Dec. 2004
! Therefore, you may find here some reminiscences about DFT.
!
! This operations module contains the core operations such as
! density, KS equation, gradient_rho, and integral as well as
! basic functionals such as kinetic, Hartree, and external potentials.
! Also, it contains transform between matrix element A and wf value.
!
! For extension to the 2D case, it is necessary to modify all functions
! and subroutines in this module except "density".
!
! Notations for functionals
! - E_Ts: kinetic energy functional
! - v_ext: external potential containing only nuclear-electron repulsion
! - v_J, E_J: Hartree potential and energy functional
!----------------------------------------------------------------------------
module operations
  use constants
  use matrix
  use coord_system
  use eigen_LAPACK
  use wavefunction
  implicit none
contains

!----------------------------------------------------------------------
! density: rho = SUM_{i=1}^{N_electrons} { |psi_i|**2 }
! SEE: Eq. (3) in Son, Young & Santra, PRA 83, 033402 (2011).
! SEE: Eq. (5) in Son, Young & Santra, PRA 83, 033402 (2011).
!----------------------------------------------------------------------
function density( G, u, f_occ ) result (rho)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: u(:,:)
  real(long), intent(in) :: f_occ(:)
  !integer, intent(in) :: config(:,:)  ! (:,1)->n, (:,2)->l, (:,3)->occ.#
  real(long) :: rho(size(u,dim=1))
  integer :: i
  !
  ! In fact, wavefunctions must have all r, theta and phi.
  !   psi(r,theta,phi) = R_nl(r) Y_lm(th,ph) = [ u_nl(r) / r ] Y_lm(th,ph)
  !
  ! However, the density depends on only r due to the sphetical symmetry.
  !   rho(r,theta,phi) --> rho(r)
  !
  ! To do so, the density is spherically averaged as
  !
  !                INT dr^3 rho (r,theta,phi)       SUM_i | u_i(r) |^2
  !   rho(r) = ---------------------------------- = ------------------
  !            INT INT sin(theta) d(theta) d(phi)       4 PI r^2
  !
  rho(:) = ZERO
  do i = 1, size(f_occ)
    !if ( config(i,3) == 0 ) cycle
    !rho(:) = rho(:) + abs( u(:,i) )**2 * dble( config(i,3) )
    if ( f_occ(i) == ZERO ) cycle
    rho(:) = rho(:) + abs( u(:,i) )**2 * f_occ(i)
  end do
  rho(:) = rho(:) / G%r(:)**2 / FOUR / PI
end function density


!----------------------------------------------------------------------
! one-electron radial Schroedinger equation
!       1  d^2   l(l+1)
!   ( - - ---- + ------ + v_eff ) u_i = e_i u_i
!       2 dr^2   2 r^2
! put v_eff, solve the eigenvalue problem, 
! and then get radial wavefunctions(u) and eigenvalues(e)
! SEE: Eq. (1) in Son, Young & Santra, PRA 83, 033402 (2011).
!----------------------------------------------------------------------
subroutine solve_SE( F, G, v_eff )
  type(WF_data), intent(inout) :: F
  type(Coordinate), intent(inout) :: G
  real(long), intent(in) :: v_eff(:)
  integer :: N_orb, i, j, n, n_max, l, l_max !, ind
  !
  ! find out N_orb from config(:,:); scanned backward
  !
  do i = size(F%config,dim=1), 1, -1
    if ( any( F%config(i,:) > 0 ) ) exit
  end do
  N_orb = i
  if ( N_orb == 0 ) N_orb = 1
  n_max = maxval( F%config(1:N_orb,1) )
  l_max = maxval( F%config(1:N_orb,2) )
  do l = 0, l_max
    G%H(:,:) = -HALF * G%Del2(:,:)
    do j = 1, G%N
      G%H(j,j) = G%H(j,j) + v_eff(j) + dble(l*(l+1)/2) / G%r(j)**2
    end do
    call get_eigenvalues( G%H, G%E )
    !
    ! set up F%u(:,1:N_orb) according to config(1:N_orb,:)
    !
    do n = l+1, n_max
      do i = 1, N_orb
        if ( F%config(i,2) == l .and. F%config(i,1) == n ) then
          !
          ! because the lowest energy starts n=1 whatever l
          ! i.e. E(1s) = Energy(G,1) with l=0
          !      E(2s) = Energy(G,2) with l=0
          !      E(3s) = Energy(G,3) with l=0
          !      ...
          !      E(2p) = Energy(G,1) with l=1
          !      E(3p) = Energy(G,2) with l=1
          !      ...
          F%u(:,i) = G%H(:,n-l) * G%wf_coeff(:)
          F%e_orb(i) = G%E(n-l)
        end if
      end do
    end do
    !
    ! currently, don't update F%u_all(:,:) at every single SCF iteration
    ! because i) we may need F%u_all with fully SCF-converged u, and
    !         ii) transformation of G%wf_coeff(:) will take some time.
    !
!------------------------------------------------------------------------------
!    !
!    ! set up F%u_all(:,1:N_all), up to l_max defined in the above
!    !                        not up to F%lmax
!    ! F%u_all(:,:) for l > l_max will be define by solve_SE_all()
!    !
!    do j = 1, G%N
!      ind = l * G%N + j
!      F%u_all(:,ind) = G%H(:,j) * G%wf_coeff(:)
!      F%e_all(ind) = G%E(j)
!    end do
!------------------------------------------------------------------------------
  end do
end subroutine solve_SE

!----------------------------------------------------------------------------
! solve the Schroedinger equation for the potential given by F%pot(:)
! store all bound and psuedo-continuum states in F%e_all(:) and F%u_all(:,:)
!   c.f.) use solve_SE() for a simple case
!         use solve_SE_finite_temp() in plasma.f90 for thermal HFS
!           --> it internally calls solve_SE_all()
! NOTE: F%pot(:) should be calculated before calling this subroutine.
!
! The j-th state of the quantum number l is stored in
!   F%e_all(i) and F%u_all(:,i), where i = l * N + j and N is # of grids.
!----------------------------------------------------------------------------
subroutine solve_SE_all( F, G )
  type(WF_data), intent(inout) :: F
  type(Coordinate), intent(inout) :: G
  integer :: i, j, l
  !
  ! calcualte all eigenstates over l = 0, ..., lmax
  !
  do l = 0, F%lmax
     G%H(:,:) = -HALF * G%Del2(:,:)
     do j = 1, G%N
        G%H(j,j) = G%H(j,j) + F%pot(j) + dble(l*(l+1)/2) / G%r(j)**2
     end do
     call get_eigenvalues( G%H, G%E )
     do j = 1, G%N
        i = l * G%N + j
        F%e_all(i) = G%E(j)
        F%u_all(:,i) = G%H(:,j) * G%wf_coeff(:)
     end do
  end do
end subroutine solve_SE_all


!----------------------------------------------------------------------
! Kinetic energy
!   Ts = SUM_i < psi_i | -1/2 Del^2 | psi_i >
!----------------------------------------------------------------------
!function functional_Ts( G, wf, config, i ) result (E)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: wf(:,:)
!  integer, intent(in) :: config(:,:)
!  integer, intent(in), optional :: i
!  real(long) :: E
!  real(long) :: E_j, w(G%N)
!  integer :: j, l, j_start, j_end
!  j_start = 1
!  j_end = size(config,dim=1)    ! # of diff. orbitals
!  if ( present(i) ) then
!    j_start = i
!    j_end = i
!  end if
!  E = ZERO
!  do j = j_start, j_end
!    if ( config(j,3) == 0 ) cycle
!    w(:) = transform_wf_to_A( G, wf(j,:) )
!    w(:) = matmul( G%Del2(:,:), w(:) )
!    w(:) = transform_A_to_wf( G, w(:) )
!    E_j= - HALF * integral( G, wf(j,:) * w(:) )
!    l = config(j,2)
!    if ( l /= 0 ) then
!      E_j = E_j + dble(l*(l+1)/2) * integral( G, wf(j,:)**2 / G%r(:)**2 )
!    end if
!    E = E + E_j * dble(config(j,3))
!  end do
!  if ( present(i) ) E = E_j
!end function functional_Ts


!----------------------------------------------------------------------
! External potential or nucleus-electron interaction
!   v_ext = - SUM_alpha { Z_alpha / r_{i alpha} }
! here is an atomic case, so v_ext = - Z / r
! SEE: part of Eq. (2) in Son, Young & Santra, PRA 83, 033402 (2011).
!----------------------------------------------------------------------
function potential_ext( G, Z ) result (v)
  type(Coordinate) :: G
  integer, intent(in) :: Z
  real(long) :: v(G%N)
  v(:) = - dble(Z) / G%r(:)
end function potential_ext


!----------------------------------------------------------------------
! Hartree potential or classical electron-electron repulsion potential
!   v_J(r) = INT{ rho(r') / |r-r'| }d^3r'
!   there are two methods to compute this potential.
!   1) multipole expansion
!       1/|r-r'| = SUM_{l=0}^{inf} SUM_{m=-l}^{l} 4PI / (2l+1) * 
!                  ( r_<^l / r_>^{l+1} ) Y_lm(theta,phi) Y*_lm(theta',phi')
!      we assume that the density rho(r') is spherical symmetric,
!      so l=0 and m=0, i.e., there is no summation over l and m.
!       --> rho(r') / |r-r'| = rho(r') / r_>
!       --> v_J(r) = INT{ rho(r') / r_> }d^3r' where r_> is max( r, r' )
!      It can be very simple; however you will see several "complicated" 
!      different implementations.  Nonetheless, you won't gain any speed up!
!      What a surprise!!!
!
!   2) Poisson equation.
!       D^2 psi(r) = - 4 PI rho(r')
!       --> v_J(r) = psi(r)
!      It is a little complicated, but it is solvable because we know 
!      how to compute inverse matrix of D^2.  Even more, since we are using
!      LU decomposition for [D^2]^-1, it turns out that CPU time is very 
!      comparable with the multipole expansion.
!
!   Currently we are using the multipole expansion.
!   There are some discrepancies (<< 0.01%) in energy between two methods:
!     Ne total energy (Poisson):   -133.13755646 a.u.
!     Ne total energy (multipole): -133.13362818 a.u.
!
!   Ref) Arfken's book, section 8.7, pp548--556
!        Jiang, Tong, and Chu, PRB 63, 045317 (2001)
!
! Hartree energy functional
!   E[rho] = 1/2 * INT{ rho(r) rho(r') / |r-r'| }d^3r' d^3r
!
! SEE: part of Eq. (2) in Son, Young & Santra, PRA 83, 033402 (2011).
! SEE: note that Eq. (2) needs to be corrected as rho(r) -> rho(r')
!----------------------------------------------------------------------
function potential_J( G, rho ) result (v)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long) :: v(G%N) !, rho_sum
  real(long) :: r_max(G%N)
  integer :: i
  integer :: j_s        ! radial index corresponding to the WS radius, r_s
  j_s = size(rho)       ! it can be the max. # of grids or the index for r_s
  !
  ! the simplest implementation
  !
!  do i = 1, G%N
!    r_max(:) = (/ ( max( G%r(i), G%r(j) ), j = 1, G%N ) /)
!    v(i) = integral( G, rho(:) / r_max(:) )
!  end do
  !
  ! the next simplest implementation
  !
  r_max(:) = G%r(:)
  do i = 1, G%N
    r_max(1:i) = G%r(i)
    v(i) = integral( G, rho(1:j_s) / r_max(1:j_s) )
  end do
  !
  ! the most complicated one; well, there is almost no speed-up, though.
  !
!  v(1) = integral( G, rho(:) / G%r(:) )
!  rho_sum = rho(1) * G%weight(1)
!  do i = 2, G%N
!    v(i) = v(i-1) - rho_sum / G%r(i-1) + rho_sum / G%r(i)
!    rho_sum = rho_sum + rho(i) * G%weight(i)
!  end do
end function potential_J

!function potential_J_solved_by_Poisson( G, Laplacian, rho ) result (v)
!  type(Coordinate), intent(in) :: G
!  type(LU_matrix), intent(in) :: Laplacian
!  real(long), intent(in) :: rho(:)
!  real(long) :: v(G%N)
!  real(long) :: v_BC, wf_coeff, w(G%N), test(G%N)
!  real(long) :: a, rp, d2
!  integer :: i, j, N
!  N = G%N
!  !
!  ! remember Laplacian is for the computation domain (A).
!  ! so, we need to transform into A and back into real space (wf).
!  !
!  w(:) = - FOUR * PI * transform_wf_to_A( G, rho(:) )
!  if ( G%rmax /= ZERO ) then
!    !
!    ! Boundary condition at Rmax:
!    ! If Rmax is infinity, the potential v_J at Rmax is equal to zero.
!    ! However, if Rmax is a finite value, consider the b. c. at Rmax
!    !
!    ! Here, we can consider a sphere of the electron cloud of the radius Rmax.
!    ! On the surface of that sphere (r=Rmax), the potential is given by
!    !   v_J = Q / r
!    !       = INT{ rho(r) }dr / Rmax
!    ! Using this, we may modify the matrix and then calculate the potential
!    ! at other points.
!    !
!    v_BC = integral( G, rho ) / G%rmax
!    !
!    ! This value should be transformed to A_j value.
!    ! Original transformation looks like ( same as transform_wf_to_A )
!    ! but consider the boundary value of r(x(N+1))
!    ! Now, v_BC becomes A_j values to be used for D2.
!    !
!    ! For LGL grids, wf_coeff at N+1 is defined by
!    ! wf_coeff = sqrt( dble((N+1)*(N+2)/2) / G%rp(N+1) ) * Legendre_P(N+1,1)
!    ! Here, Legendre_P(N+1,1) = 1
!    !       G%rp(N+1) = L * ( 2 + alpha ) / alpha**2, alpha = 2 L / Rmax
!    !
!    v_BC = v_BC / G%wf_coeff_b
!    !
!    ! modify the matrix
!    !   SUM_{j=1}^{N-1} D_ij v_j + D_iN v_N = w_i,  i = 1, ..., N-1
!    !   SUM_{j=1}^{N-1} D_Nj v_j + D_NN v_N = w_N
!    !     unknown : v_j ( j = 1, ..., N-1 ), w_N
!    !     known   : w_i ( i = 1, ..., N_1 ), v_N <-- transformed v_BC above
!    !   => SUM_{j=1}^{N-1} D_ij v_j = w_i - D_iN v_N
!    !
!    w(:) = w(:) - G%D2_boundary(:) * v_BC
!  end if
!  w(:) = LU_solve( Laplacian, w(:) )
!  v(:) = transform_A_to_wf( G, w(:) )
!end function potential_J_solved_by_Poisson

function functional_J( G, rho, v_J ) result (E)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long), intent(in), optional :: v_J(:)
  real(long) :: E
  real(long) :: v(G%N)
  integer :: j_s        ! radial index corresponding to the WS radius, r_s
  j_s = size(rho)       ! it can be the max. # of grids or the index for r_s
  !
  ! this function must be called by functional_J( G, rho, v_J )
  !                              or functional_J( G, rho )
  !
  if ( present(v_J) ) then
    v(:) = v_J(:)
  else
    v(:) = potential_J( G, rho(1:j_s) )
  end if
  E = HALF * integral( G, rho(1:j_s) * v(1:j_s) )
end function functional_J


!----------------------------------------------------------------------
! compute | Del rho |
!
! Because of accuracy problem, it is very difficult to get reasonable Del rho.
! For detailed discussion, obsolete.f90 has a lot of choices for Del rho.
! In fact, Del rho should be dependent of only rho.
! But here, we use wf instead of rho because of accuracy problem.
!----------------------------------------------------------------------
!function gradient_rho( G, rho, wf ) result (w)
!  type(Coordinate), intent(in) :: G
!  real(long), intent(in) :: rho(:), wf(:,:)
!  real(long) :: w(G%N)
!  real(long) :: Z_eff, A, v(G%N)
!  integer :: i, j
!  w(:) = ZERO
!  do i = 1, size(wf,dim=1)
!    v(:) = transform_wf_to_A( G, wf(i,:) )
!    v(:) = matmul( G%Del_r(:,:), v(:) ) - v(:) / G%r(:)
!    v(:) = transform_A_to_wf( G, v(:) )
!    w(:) = w(:) + TWO * wf(i,:) * v(:)
!  end do
!  !v(:) = rho(:) / G%wf_coeff(:)
!  !v(:) = matmul( G%Del_r(:,:), v(:) ) - v(:) / G%r(:)
!  !w(:) = v(:) * G%wf_coeff(:)
!  !
!  ! post-process for Del rho
!  ! If we assume rho -> exp(-ar) when r -> infinity
!  ! then Del rho = - a = log(rho)/r.
!  ! So, take -a for Del rho when rho is very small.
!  !
!  do j = 1, G%N
!    if ( abs(rho(j)) < 1.0e-8_long ) exit
!  end do
!  A = log( rho(j) ) - w(j) / rho(j) * G%r(j)
!  do i = j, G%N
!    Z_eff = ( A - log( rho(i) ) ) / G%r(i)
!    w(i) = -Z_eff * rho(i)
!  end do
!end function gradient_rho

end module operations
