!----------------------------------------------------------------------------
! pcs.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2021 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2021 Sang-Kil Son <sangkil.son@cfel.de>
!          2015-2016 Koudai Toyota
!          2016      Dietrich Krebs
!          2018,2020 Ludger Inhester
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Photo-ionization / -absorption cross section (PCS) calculation
!   written by Son, Sang-Kil in Jul. 2010 at DESY
!
! - photoionization: one electron goes into the continuum
! - photoabsorption: broader concept including both photoionization and 
!   resonant photoexcitation (all possible bound-to-bound transitions)
!
! This module requires unbound.f for continuum wavefunctions.
!----------------------------------------------------------------------------
module pcs
  use constants
  use coord_system
  use gps
  use read_param
  use elec_config
  use wavefunction
  use process
  use interpolate
  implicit none
  !
  ! PCS_data: needed if P%yes_resonance = T or P%convolution = T
  !           in other words, needed for calculate_photoabsorption_CS()
  !
  type :: PCS_data
    !
    ! define an energy range
    !
    real(long) :: E_start, E_end, dE
    !
    ! for photoexcitation
    !   we know an analytical expression
    !   since the cross section shape is assumed to be (or forced to be)
    !   narrower than convoluting Gaussian, 
    !   we'd better integrate w.r.t. the cross section for convolution
    !
    integer, allocatable :: N_res(:)            ! N_res(F%N_orb)
    integer, allocatable :: ind(:,:)            ! ind(F%N_all,F%N_orb)
    real(long), allocatable :: CS_res(:,:)      ! CS_res(F%N_all,F%N_orb)
    !real(long), allocatable :: E_trans(:,:)    ! E_trans(F%N_all,F%N_orb)
    integer :: linewidth
    !
    ! for photoionization
    !   prepare interpolation from E_start to E_end with a step size of dE
    !   put E_thres if it is in (E_start,E_end)
    !
    real(long), allocatable :: E_thres(:)       ! E_thres(F%N_orb)
    type(Spline), allocatable :: SP(:)          ! SP(F%N_orb)
  end type PCS_data
contains

!----------------------------------------------------------------------------
! photoionization: only bound-to-continuum
!   this routine is kept for simplicity; it should be used only if
!   P%yes_resonance = F and P%yes_convolution = F
!----------------------------------------------------------------------------
subroutine calculate_photoionization_CS( P, F, S )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  type(Process_data), intent(inout) :: S
  integer    :: i, k, N_PE, OUTPUT
  real(long) :: CS_sum, E
  real(long) :: CS(F%N_subshell), beta(F%N_subshell) 
  real(long) :: omega_in, AU2MB

  OUTPUT = P%OUTPUT
  AU2MB = AU2M**2 / 1.0e-22_long        ! 1 Mb = 10^-22 m^2
  N_PE = nint( ( P%PE_end - P%PE_start ) / P%PE_dE )
  if ( P%yes_xmdyn ) N_PE = 0           ! always one value for XMDYN
  if ( N_PE > 0 ) then
    !
    ! loop over photon energies
    !
    write(OUTPUT, "('#'/'# Photoabsorption cross section (in Mb):')")
    write(OUTPUT, "('#',a11,100a13)") "P.E.(eV)", "Total", &
      (/ ( trim( WF_name( F, i ) ), i=1, F%N_subshell ) /)
    photon_energy: do k = 0, N_PE       ! tabulate PCS along with P.E.
      omega_in = P%PE_start + P%PE_dE * dble(k)
      CS(:) = photoionization_CS( P, F, omega_in )
      CS_sum = sum( CS(:) )
      write(OUTPUT, "(f12.5,100es13.5)") omega_in * AU2EV, &
        CS_sum * AU2MB, CS(:) * AU2MB
    end do photon_energy
    write(OUTPUT, "(/)")
  else
    !
    ! calculate PCS at one photon energy only
    !
    omega_in = max( P%PE_start, P%PE_end )      ! take a larger one
    write(OUTPUT, "('#'/'# Photoabsorption cross section:')")
    write(OUTPUT, "('#   photon energy = ',f9.3,' a.u. = ',f9.2,' eV')") &
      omega_in, omega_in * AU2EV
    CS(:) = photoionization_CS( P, F, omega_in, beta )
    CS_sum = sum( CS(:) )
    if ( P%yes_pcs_beta ) then    ! print with beta parameter
      write(OUTPUT, "('#   ',13x,3a13)") "cs(Mb)", "cs(a.u.)" , "beta"
      do i = 1, F%N_subshell      ! tabulate PCS and beta along with diff. orbitals
        if ( CS(i) > ZERO ) then  ! however, beta values are only sensible if CS > 0
          write(OUTPUT, "('#   ',a10,' : ',3es13.5)") &
            WF_name(F,i), CS(i) * AU2MB, CS(i), beta(i)
        else
          write(OUTPUT, "('#   ',a10,' : ',2es13.5)") &
            WF_name(F,i), CS(i) * AU2MB, CS(i)
        end if
        if ( CS(i) > ZERO ) then  ! print out data in results.dat (in a.u.)
          E = omega_in - ionization_E( F, i )
          call add_Process( S, "photoionization", CS(i), E, i )
        end if
      end do
    else                          ! print without beta parameter
      write(OUTPUT, "('#   ',13x,2a13)") "cs(Mb)", "cs(a.u.)"
      do i = 1, F%N_subshell      ! tabulate PCS along with diff. orbitals
        write(OUTPUT, "('#   ',a10,' : ',2es13.5)") &
          WF_name(F,i), CS(i) * AU2MB, CS(i)
        if ( CS(i) > ZERO ) then  ! print out data in results.dat (in a.u.)
          E = omega_in - ionization_E( F, i )
          call add_Process( S, "photoionization", CS(i), E, i )
        end if
      end do
    end if
    write(OUTPUT, "('#   ',a)") repeat( "-", 39 )
    write(OUTPUT, "('#   Total P cs = ',2es13.5)") CS_sum * AU2MB, CS_sum
  end if
end subroutine calculate_photoionization_CS

!----------------------------------------------------------------------------
! compute cross section of photoionization
! - omega_in: incident photon energy
! - CS(i): cross section for the i-th orbital, in a.u.
! SEE: Eq. (6) in Son, Young & Santra, PRA 83, 033402 (2011).
!----------------------------------------------------------------------------
function photoionization_CS( P, F, omega_in, ao_beta ) result (CS)
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  real(long), intent(in) :: omega_in
  real(long) :: CS(F%N_subshell) 
  ! function changes ao_beta by intended(!) side effect
  ! which is not employed if ao_beta is just omitted
  real(long), optional, intent(inout) :: ao_beta(F%N_subshell)
  integer          :: i, l, lp, l_dip, occ_n
  real(long)       :: S(-1:1), phsshft(-1:1)
  character(len=5) :: wfct_name
  real(long)       :: E, ll, alpha
  real(long)       :: E_abs
  real(long)       :: u(F%N_uni), up(F%N_uni) !, AU2MB
  integer          :: n !, index_rel(F%N_orb_rel)
  alpha = ONE / SPEED_C_AU               ! fine-structure constant = 1 / 137
  CS(:) = ZERO
  if(present(ao_beta)) ao_beta(:) = TWO  ! standard value for s-type orbitals
  if ( P%yes_verbose ) then
    write(P%OUTPUT, "('#'/'#',23x,'P.E. = ',f8.2,5x,a3,' - ',a10,5x,a13)") &
      omega_in * AU2EV, "n,l", "E,l", "R_int"
  end if
  orbital: do i = 1, F%N_subshell
    !
    ! E : kinetic energy of an ionized electron w.r.t. potential tail
    !     ( = orbital energy + photon energy - potential tail )
    !     to be used to check whether it's ionized or not
    ! E_abs : K.E. of an ionized electron w.r.t. zero (in the absolute scale)
    !         ( = orbital energy + photon energy )
    !         to be used for an argument of unbound()
    ! E and E_abs are different only if the potential tail is not zero.
    !
    E = omega_in - ionization_E( F, i )
    if ( F%yes_relativity ) then                ! too much checking yes_rel
      E_abs = omega_in + F%e_orb_rel(i)
    else
      E_abs = omega_in + F%e_orb(i)
    end if
    if ( E < ZERO ) cycle                       ! it cannot be ionized
    if ( P%yes_relativity ) then
       n = F%config_rel(i, 5)
       l = F%config_rel(i, 2)
       occ_n = F%config_rel(i, 4)
    else
       n = i
       l = F%config(i, 2)
       occ_n = F%config(i, 3)
    endif
    if ( occ_n == 0 ) cycle              ! skip if no electron in this orbital
    do l_dip = -1, 1, 2
      lp = l + l_dip
      if ( lp .ge. 0 ) then                ! skip if l-1 < 0
        ll = dble(lp)   ! unbound.f accepts real*8 rather than integer
        !
        ! E_abs is used for unbound(); see the comment in the above
        !
        call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, E_abs, F, 'P' )
        S(l_dip) = radial_integral_uni( F%N_uni, F%r_uni(:), F%u_uni(:,n), u(:) )
        ! only calculate phaseshift, if beta calc. is desired
        if ( (P%yes_pcs_beta).and.(present(ao_beta)) ) then      
          ! obtain name of wfct for helpful error-message
          wfct_name = WF_name( F, i )
          phsshft(l_dip) = get_phase_shift(u, up, F%r_uni, F%N_uni, lp, sqrt(TWO*E), F%q, wfct_name)
        end if
        if ( P%yes_verbose ) then
          write(P%OUTPUT, "('#',43x,i1,x,i1,' - ',f8.2,x,i1,5x,es13.5)") &
            F%config(i,1), l, E * AU2EV, lp, S(l_dip)
        end if
        CS(i) = CS(i) + dble( max(l,lp) ) / dble(2*l+1) * (S(l_dip))**2
      end if
    end do
    ! calculate new beta factor if necessary and required by user
    if ( (P%yes_pcs_beta).and.(present(ao_beta)).and.(l > 0) ) then
      ao_beta(i) = get_beta(l, S(1), phsshft(1), S(-1), phsshft(-1))
    end if
    CS(i) = CS(i) * FOUR * PI**2 / THREE * alpha * omega_in * dble(occ_n)
  end do orbital
  if ( P%yes_verbose ) write(P%OUTPUT, "('#')")
contains
  function radial_integral_uni( N, r, f, g ) result (S)
    real(long) :: S
    integer, intent(in) :: N
    real(long), intent(in) :: r(:), f(:), g(:)
    real(long) :: dr
    dr = r(2) - r(1)    ! equal spacing
    !
    ! f and g are already multiplied by r
    ! integral is evaluated by the trapezoid rule.
    ! at the end points, r(0) = 0 and f(N) or g(N) ~= 0
    !
    S = sum( f(1:N) * r(1:N) * g(1:N) )         ! f(1:N) rather than f(:)
    S = S - HALF * ( f(N) * r(N) * g(N) )       ! according to trapezoid rule
    S = S * dr
  end function radial_integral_uni
end function photoionization_CS





function get_beta(a_l, a_Splus, a_phsshft_p, a_Sminus, a_phsshft_m) result(beta)

  implicit none
  
  !# arguments------------------------------------  
  integer, intent(in)                      :: a_l      ! the angular momentum of the initial state
  real(long), intent(in)                   :: a_Splus, a_Sminus
  real(long), intent(in)                   :: a_phsshft_p, a_phsshft_m

  !# local variables------------------------------
  real(long)                               :: beta
  real(long)                               :: aux_var

  !the actual calculation of beta
  aux_var = dble( a_l * (a_l - 1) )
  aux_var = aux_var * a_Sminus * a_Sminus

  beta = aux_var
  !-------------
  aux_var = dble( (a_l + 1) * (a_l + 2) )
  aux_var = aux_var * a_Splus * a_Splus
  
  beta = beta + aux_var
  !-------------
  aux_var = a_phsshft_p - a_phsshft_m
  aux_var = dcos(aux_var)
  aux_var = aux_var * dble( 6 * a_l * (a_l + 1) )
  aux_var = aux_var * a_Splus * a_Sminus

  beta = beta - aux_var
  !-------------
  aux_var = dble( a_l + 1 )
  aux_var = aux_var * a_Splus * a_Splus
  aux_var = aux_var + ( dble(a_l) * a_Sminus * a_Sminus )
  aux_var = aux_var * dble( (2 * a_l) + 1 )
  ! could have safety check for ZERO here
  
  beta = beta / aux_var
  !-------------
end function get_beta


!----------------------------------------------------------------------------
! photoabsorption: bound-to-continuum and bound-to-bound
!   this routine is more generalized than calculate_photoionization_CS().
!
! PE single point : no PE_scan
!   no  resonance, no  convolution : no E_scan
!   no  resonance, yes convolution : E_scan with +/- bandwidth
!   yes resonance, no  convolution : E_scan with +/- max(bandwidth,linewidth)
!   yes resonance, yes convolution : E_scan with +/- bandwidth
! PE range : PE_scan needed
!   no  resonance, no  convolution : E_scan with PE_range
!   no  resonance, yes convolution : E_scan with PE_range +/- bandwidth
!   yes resonance, no  convolution : E_scan with PE_range +/- max(band,line)
!   yes resonance, yes convolution : E_scan with PE_range +/- bandwidth
!----------------------------------------------------------------------------
subroutine calculate_photoabsorption_CS( P, G, F, S )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  type(Process_data), intent(inout) :: S
  type(PCS_data) :: R
  integer :: i, j, k, kp, li, lf, OUTPUT
  real(long) :: omega_in, alpha, AU2MB
  real(long) :: E, E_trans, T
  real(long) :: PE_scan(F%N*2*3+F%N_subshell) ! *2 for l+/-1, *3 for 0/+/-linewidth
  real(long), allocatable :: E_scan(:), CS_scan(:,:), E_thres_added(:)
  real(long) :: CS(F%N_subshell), CS_sum, step 
  !integer :: elec(P%N_elec)
  integer :: N_hole, N_PE, N_scan, N_p, ind_i, ind_f
  integer :: ji, jf, count_added
  real(long), external :: sixj
  character(len=18) :: format_str
  logical :: flag

  alpha = ONE / SPEED_C_AU      ! fine-structure constant = 1 / 137
  OUTPUT = P%OUTPUT
  AU2MB = AU2M**2 / 1.0e-22_long        ! 1 Mb = 10^-22 m^2
  !
  ! create elec(:); let's use elec(:) rather than F%config(:,:)
  ! - change do-loop with F%N_orb --> P%N_elec == size( elec )
  ! - change F%config(:,:) --> elec(:)
  ! - use l = int( elec(i) / N ) and n = mod( elec(i), N ) + 
  !
  !if ( P%yes_relativity == .true. ) then
  !   call config_rel_to_elec( F%config_rel, elec, P%N )
  !else
  !   call config_to_elec( F%config, elec, P%N )
  !endif
  !
  ! set up PE_scan: used for printing out CS as a function of PE
  !   PE_scan(:) contains only ionization thresholds and resonances.
  !   PE will be scanned from P%PE_start and P%PE_end 
  !   and refined near PE_scan(:).
  !
  N_PE = nint( ( P%PE_end - P%PE_start ) / P%PE_dE )
  if ( P%yes_xmdyn ) N_PE = 0           ! always one value for XMDYN
  if ( N_PE > 0 ) then
    PE_scan(:) = -ONE                   ! put arbitrary negative values
    do i = 1, F%N_subshell
      E = ionization_E( F, i )          ! adding ionization thresholds
      if ( E >= P%PE_start .and. E <= P%PE_end ) then         
         call add_scan( PE_scan, E )
      endif
   end do 
  else
    PE_scan(1) = max( P%PE_start, P%PE_end )    ! take a larger one
  end if
  !
  ! set up PCS_data
  !
  allocate( R%N_res(F%N_subshell) )
  allocate( R%ind(F%N_subshell,F%N_subshell), R%CS_res(F%N_subshell,F%N_subshell) )
  allocate( R%E_thres(F%N_subshell), R%SP(F%N_subshell), E_thres_added(F%N_subshell) )

  !
  ! set up E_scan: used for Spline interpolation for photoionization CS
  !   E_scan(:) contains all energies from R%E_start to R%E_end
  !             and 5 more points near thresholds.
  !
  if ( P%N_convolution > 0 ) then       ! yes convolution
    R%E_start = P%PE_start - P%bandwidth * P%multiple_bandwidth
    R%E_end   = P%PE_end   + P%bandwidth * P%multiple_bandwidth
  else if ( P%yes_resonance ) then
    R%E_start = P%PE_start - max(P%bandwidth,P%linewidth)*P%multiple_bandwidth
    R%E_end   = P%PE_end   + max(P%bandwidth,P%linewidth)*P%multiple_bandwidth
  else
    R%E_start = P%PE_start
    R%E_end   = P%PE_end
  end if
  if ( R%E_start < ZERO ) R%E_start = ZERO
  R%dE = P%pcs_scan_step
  N_scan = nint( ( R%E_end - R%E_start ) / R%dE ) + 1
  if ( R%E_start /= R%E_end .and. N_scan < 5 ) then  ! make N_scan at least 5
    R%dE = ( R%E_end - R%E_start ) / FOUR
    N_scan = 5
  end if
  E_thres_added(:) = ZERO
  count_added = 0
  N_p = N_scan
  do i = 1, F%N_subshell
    R%E_thres(i) = ionization_E( F, i )
    if ( R%E_thres(i) >= R%E_start .and. R%E_thres(i) <= R%E_end ) then 
       ! no counting the same orbital ionization energy twice 
       flag = .false.
       do j = 1, count_added
          if ( E_thres_added(j) == R%E_thres(i) ) then
             flag = .true.
             exit
          endif
       enddo
       if ( .not. flag ) then          
          count_added = count_added + 1
          E_thres_added(count_added) = R%E_thres(i)
          N_scan = N_scan + 5
       endif
    endif
 end do
  allocate( E_scan(N_scan), CS_scan(N_scan,F%N_subshell) )
  E_scan(:) = -ONE      ! put arbitrary negative values
  E_scan(1:N_p) = (/ ( R%E_start + dble(i) * R%dE, i = 0, N_p-1 ) /)
  do i = 1, F%N_subshell
    if ( R%E_thres(i) >= R%E_start .and. R%E_thres(i) <= R%E_end ) then
       ! no counting the same orbital ionization energy twice 
       flag = .false.
       do j = 1, N_scan
          if ( E_scan(j) == R%E_thres(i) ) then
             flag = .true.
             exit
          endif
       enddo
       if ( .not. flag ) then
          step = R%dE * 0.1_long  
          if ( R%E_thres(i) + FOUR * step > R%E_end ) then
             step = ( R%E_end - R%E_thres(i) ) / 5.0_long
          endif
          do k = 0, 4       ! adding 5 points after the ionization threshold
             call add_scan( E_scan, R%E_thres(i) + dble(k) * step )
          end do
       endif
    end if
  end do
!  write(*,*)(i,E_scan(i),i=1,N_scan)

  !
  ! calculate photoionization cross sections at every E_scan
  ! and prepare Spline interpolation for E > E_thres
  !
  do j = 1, N_scan
    CS_scan(j,:) = photoionization_CS( P, F, E_scan(j) )
  end do
  do i = 1, F%N_subshell
    do j = 1, N_scan
      if ( R%E_thres(i) <= E_scan(j) ) exit
    end do
    if ( j <= N_scan ) R%SP(i) = Spline_create( E_scan(j:), CS_scan(j:,i) )
  end do
  !
  ! calculate photoabsorption cross sections:
  ! first, find all possible transitions between R%E_start and R%E_end
  ! and then print out cross sections if N_PE == 0.
  !
  write(OUTPUT, "('#')")
  if ( N_PE == 0 ) then                 ! print out one value of PE
    if ( P%yes_resonance ) then
      write(OUTPUT, "('# Photoabsorption cross section: " // &
                    "ionization and resonant excitation')")
    else
      write(OUTPUT, "('# Photoionization cross section: ')")
    end if
    write(OUTPUT, "('#   photon energy (eV) = ',f9.2)") PE_scan(1) * AU2EV
  else                                  ! print out a range of PE
    if ( P%yes_resonance ) then
      write(OUTPUT, "('# Possible resonances for bound-to-bound transitions:')")
    else
      write(OUTPUT, "('# Photoionization cross section: ')")
    end if
    write(OUTPUT, "('#   photon energy (eV) = ',f9.2,' to ',f9.2,' by ',f9.3)")&
      P%PE_start * AU2EV, P%PE_end * AU2EV, P%PE_dE * AU2EV
  end if
  if ( P%N_convolution > 0 .or. P%yes_resonance ) &
    write(OUTPUT, "('#   bandwidth (eV)     = ',f9.2)") P%bandwidth * AU2EV
  if ( N_scan > 1 ) then
    write(OUTPUT, "('#   energy scan (eV)   = ',f9.2,' to ',f9.2)") &
      R%E_start * AU2EV, R%E_end * AU2EV
  end if
  if ( P%N_convolution > 0 ) then
    write(OUTPUT, "('#   convolution with ',i5,' points " // &
                  "for photoionization CS')") P%N_convolution
  else
    write(OUTPUT, "('#   no convolution performed')")
    if ( N_PE /= 0 .and. P%yes_resonance ) &
      write(OUTPUT, "('#   for a better resonance plot, decrease -dE enough')")
  end if
  if ( P%yes_resonance ) then
    if ( P%linewidth == ZERO ) then
      write(OUTPUT, "('#   natural linewidths are not given; ')", advance="no")
      if ( P%N_convolution > 0 ) then
        write(OUTPUT, "('assumed less than bandwidth')")
      else
        write(OUTPUT, "('the same as no resonance case')")
      end if
    else
      write(OUTPUT, "('#   natural linewidth (eV) = ',f12.5)") P%linewidth*AU2EV
      if ( P%N_convolution > 0 ) write(OUTPUT, "('#   ',a)") &
        "this natural linewidth is assumed to be less than the bandwidth"
    end if
  end if
  if ( P%yes_resonance .or. N_PE == 0 ) then
    write(OUTPUT, "('#'/'#   i    - j     : ')", advance="no")
    if ( N_PE == 0 ) write(OUTPUT, "(2a13)", advance="no") "cs(Mb)", "cs(a.u.)"
    if ( P%yes_resonance ) &
      write(OUTPUT, "(2a13)", advance="no") "E_trans(eV)", "<i|r|j>"
    write(OUTPUT, "()")
  end if
  omega_in = P%PE_start
  if ( N_PE == 0 ) omega_in = PE_scan(1)        ! defined in the above
  CS_sum = ZERO
  ji = 0
  jf = 0
  initial: do i = 1, F%N_subshell            ! initial state
     if ( .not. P%yes_resonance ) exit initial
     if ( P%yes_relativity ) then
        if ( F%e_orb_rel(i) > ZERO ) cycle
        if ( F%config_rel(i,4) == 0 ) cycle
        li = F%config_rel(i,2)
        ji = F%config_rel(i,3) ! 2 * j quantum number of the initial state
     else
        if ( F%e_orb(i) > ZERO ) cycle
        if ( F%config(i,3) == 0 ) cycle
        li = F%config(i,2)
     endif
     if ( li == P%lmax ) cycle
     k = 0
     ! When resonace flag is on, default value of P%lmax is defined as maxval (config(:,2)) + 1 
     ! afeter Aufbau principle. For instance, P%lmax becomes 3 for neutral Xe of the ground state. 
     ! This means that photoexciations from f to g orbital are not considered; see 'if' statement above.     
     final: do j = 1, F%N_subshell
        if ( P%yes_relativity ) then
           lf = F%config_rel(j,2)
           jf = F%config_rel(j,3) ! 2 * j quantum number of the final state
        else
           lf = F%config(j,2)
        endif
        
        ! Having obtained li, ji and lf, jf, we consider the selection rules
        if ( P%yes_relativity ) then
           if( abs(ji-jf) > 2 .or. abs(li - lf) /= 1 ) cycle
           if ( F%e_orb_rel(j) > ZERO ) cycle
        else
           if( abs(li - lf) /= 1 ) cycle
           if ( F%e_orb(j) > ZERO ) cycle
        endif

        E_trans = ionization_E( F, i ) - ionization_E( F, j )
        if ( E_trans < ZERO ) cycle     ! skip if final_E is lower than init_E
        if ( E_trans < R%E_start ) cycle       ! skip if no resonance
        if ( E_trans > R%E_end   ) cycle
        if ( P%yes_relativity ) then
           N_hole = jf + 1 - F%config_rel(j,4) 
        else
           N_hole = 4*lf + 2 - F%config(j,3)
        endif
        if ( N_hole == 0 ) cycle        ! skip if it is occupied

        !
        ! now this transition is possibly resonant to the photon energy
        ! store this index into resonance(1:,i)
        ! resonance(0,i) will be # of bound-bound transitions
        !
        k = k + 1       ! corresponding to # of resonant excitations
        R%ind(k,i) = j
        !
        ! calculate resonant excitation cross section
        ! this is NOT complete; the Lorentzian part not included here
        !                                1           gamma/2
        ! Lorentz profile: L(x,gamma) = -- * ------------------------
        !                               pi   (w - w0)^2 + (gamma/2)^2
        ! Don't forget 1/pi!!!
        !
        if ( P%yes_relativity ) then
           ind_i = F%config_rel(i,5)
           ind_f = F%config_rel(j,5)
        else
           ind_i = i
           ind_f = j
        endif
        T = radial_integral( G, F%u(:,ind_i) * F%r(:) * F%u(:,ind_f) )
        R%CS_res(k,i) = FOUR * PI**2 * alpha / THREE * dble(N_hole) * dble( max( li, lf ) ) * T**2
        if ( P%yes_relativity ) then
           R%CS_res(k,i) = R%CS_res(k,i) * sixj( 2*li, 1,  ji, jf, 2, 2*lf )**2
        else
           R%CS_res(k,i) = R%CS_res(k,i) / dble( 4 * lf + 2 ) / dble( 2 * li + 1 )
        endif

        !
        ! store this resonant energy into PE_scan(:)
        !
        if ( E_trans >= P%PE_start .and. E_trans <= P%PE_end ) then
           if ( P%linewidth /= ZERO ) then
              call add_scan( PE_scan, E_trans - P%linewidth )
              call add_scan( PE_scan, E_trans )
              call add_scan( PE_scan, E_trans + P%linewidth )
           else
              call add_scan( PE_scan, E_trans )
           end if
        end if
        !
        ! print this resonance line 
        ! if N_PE == 0, print it with the resonant absorption cross section
        !
        write(OUTPUT, "('#   ',a4,' - ',a5,' : ')", advance="no") &
             WF_name( F, i ), WF_name( F, j )
        if ( N_PE == 0 ) then   ! print out resonant photoexcitation CS
           CS(i) = partial_PCS( P, F, R, omega_in, i, k )
           CS_sum = CS_sum + CS(i)
           write(OUTPUT, "(2es13.5)", advance="no") CS(i) * AU2MB, CS(i)
           if ( CS(i) > ZERO ) then
              call add_Process( S, "photoexcitation", CS(i), E_trans, i, j )
           end if
        end if
        write(OUTPUT, "(f13.5,es13.5)") E_trans * AU2EV, T
     end do final
     R%N_res(i) = k              ! store # of resonant excitations
     if ( N_PE == 0 ) then       ! print out photoionization CS
        CS(i) = partial_PCS( P, F, R, omega_in, i )
        CS_sum = CS_sum + CS(i)
        !
        ! maybe later on, we need to print out two diff. channels here
        !
        write(OUTPUT, "('#   ',a4,' - ',a5,' : ',2es13.5)") &
             WF_name( F, i ), "cont.", CS(i) * AU2MB, CS(i)
        if ( CS(i) > ZERO ) then  ! print out data in results.dat (in a.u.)
           E = omega_in - ionization_E( F, i )
           call add_Process( S, "photoionization", CS(i), E, i )
        end if
     end if
  end do initial
  if ( N_PE == 0 ) then
     write(OUTPUT, "('#   ',a)") repeat( "-", 41 )
     write(OUTPUT, "('#   Total P cs =   ',2es13.5)") CS_sum * AU2MB, CS_sum
  end if

  !
  ! loop over photon energy
  !
  if ( N_PE > 0 ) then
     write(OUTPUT, "('#'/'# Photoabsorption cross section (in Mb):')")
     write(OUTPUT, "('#',a11,100a13)") "P.E.(eV)", "Total", &
          (/ ( trim(WF_name( F, i )), i=1, F%N_subshell ) /)
     k = 0         ! index for regular grids
     kp = 1        ! index from PE_scan
     write(format_str,'(a7,i4,a7)')'(f12.5,',F%N_subshell + 2,'es13.5)'
     do
        omega_in = P%PE_start + P%PE_dE * dble(k)
        if ( PE_scan(kp) < omega_in .and. PE_scan(kp) >= ZERO ) then
           omega_in = PE_scan(kp)
           kp = kp + 1
        else
           k = k + 1
        end if
        CS(:) = photoabsorption_CS( P, F, R, omega_in )        
        CS_sum = sum( CS(:) )
        write(OUTPUT, format_str) omega_in * AU2EV, &
             CS_sum * AU2MB, CS(:) * AU2MB
        if ( k > N_PE ) exit
     end do
  end if
contains
  subroutine add_scan( E_scan, E )  ! add E in the ascending order
    real(long), intent(inout) :: E_scan(:)
    real(long), intent(in) :: E
    integer :: i, j, i_end
    do i = 1, size(E_scan)
      if ( E_scan(i) > E .or. E_scan(i) < ZERO ) exit
    end do
!    do i_end = i + 1, size(E_scan)
    do i_end = i, size(E_scan)
      if ( E_scan(i_end) < ZERO ) exit
    end do
    if ( i > size(E_scan) .or. i_end > size(E_scan) ) then
      write(STDERR, *) "ERROR in add_scan: No space in E_scan"
      write(STDERR, *) 'i, i_end = ', i, i_end
      write(STDERR, *) "E = ", E * AU2EV
      write(STDERR, *) "size(E_scan) = ", size(E_scan)
      write(STDERR, *) E_scan(:) * AU2EV
      stop
    end if
    if ( E_scan(i) >= ZERO ) then
      do j = i_end, i + 1, -1
        E_scan(j) = E_scan(j-1)
      end do
    end if
    E_scan(i) = E
  end subroutine add_scan
end subroutine calculate_photoabsorption_CS

function partial_PCS( P, F, R, omega_in, i, k ) result ( CS )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  type(PCS_data), intent(in) :: R
  real(long), intent(in) :: omega_in
  integer, intent(in) :: i              ! i-th occ. orbital
  integer, intent(in), optional :: k    ! no k -> photoionization
  real(long) :: CS
  real(long) :: omega, dE, sigma, E_trans, value
  real(long) :: E_min, E_max, profile
  integer :: j, N, N_occ

  if ( omega_in < R%E_start-P%PE_dE .or. omega_in > R%E_end+P%PE_dE ) then
     write(STDERR,*) "ERROR in partial_PCS(): omega_in is out of range."
     write(STDERR,*) "omega_in = ", omega_in * AU2EV
     write(STDERR,*) "E range = ", R%E_start * AU2EV, R%E_end * AU2EV
     write(STDERR,*)

     stop
  end if
  sigma = P%bandwidth / TWO / sqrt( TWO * log( TWO ) )  ! P%bandwidth in FWHM
  CS = ZERO

  if ( present(k) ) then
     if ( P%yes_relativity ) then
        N_occ = F%config_rel(i,4)
     else
        N_occ = F%config(i,3)
     endif
     !
     ! calculate resonant excitation (bound-to-bound) cross section
     ! assuming that the Lorentz profile is approx. to the delta function
     !
     !if ( k < 1 .or. k > R%N_res(i) ) stop "ERROR in partial_PCS()"
     E_trans = ionization_E( F, i ) - ionization_E( F, R%ind(k,i) )
     if ( P%N_convolution == 0 ) then    ! no convolution
        CS = dble( N_occ ) * omega_in * R%CS_res(k,i) * &
             Lorentz_profile( E_trans - omega_in, P%linewidth )
     else
        omega = E_trans   ! resonance between photon E and transition E
        if ( P%yes_flattop ) then
           profile = Flattop_profile( omega_in, omega, P%bandwidth )
        else
           profile = Gauss_profile( omega_in - omega, sigma )
        endif
        CS = dble( N_occ ) * omega * R%CS_res(k,i) * profile
!        write(99,*) CS / TWO / PI**2 * SPEED_C_AU / profile
     end if
  else
    !
    ! calculate photoionization (bound-to-continuum) cross sections from Spline
    ! convolution: (f*g)(t) = int f(t') g(t-t') dt'
    !                       = int f(t-t') g(t') dt'
    ! t = omega_in
    ! t' = omega
    !
     if ( P%N_convolution == 0 ) then    ! no convolution
        if ( omega_in >= R%E_thres(i) ) &
             CS = Spline_interpolate( R%SP(i), omega_in )
     else                                ! make a convolution
        N = int( P%N_convolution / 2 )
        dE = P%bandwidth * P%multiple_bandwidth / dble(N)
        E_max = omega_in + P%bandwidth * P%multiple_bandwidth
        E_min = omega_in - P%bandwidth * P%multiple_bandwidth
        if ( E_min < ZERO ) E_min = ZERO
        if ( R%E_thres(i) < E_min ) then  ! all photon energies are above IP
           do j = -N, N                    ! integrate near omega_in
              omega = omega_in + dble(j) * dE
              if ( P%yes_flattop ) then
                 profile = Flattop_profile(omega_in, omega, P%bandwidth)
              else
                 profile = Gauss_profile( omega_in - omega, sigma )             
              endif
              value = Spline_interpolate( R%SP(i), omega ) * profile
                   
              if ( j == -N .or. j == N ) value = value * HALF  ! trapezoidal rule
              CS = CS + value * dE
           end do
        else if ( R%E_thres(i) <= E_max ) then    ! IP is between E_min and E_max
           do j = 0, 2*N                   ! integrate after IP
              omega = R%E_thres(i) + dble(j) * dE 
              if ( omega > E_max ) exit
              if ( P%yes_flattop ) then
                 profile = Flattop_profile(omega_in, omega, P%bandwidth)
              else
                 profile = Gauss_profile( omega_in - omega, sigma )             
              endif
              value = Spline_interpolate( R%SP(i), omega ) * profile
              CS = CS + value * dE
           end do
           CS = CS - value * dE * HALF     ! considering trapezoidal rule
        else
           CS = ZERO                       ! all photon energies are below IP
        end if
     end if
  end if
  !
  ! if CS is too small, then set it to zero
  !
  if ( CS <= 1.0e-100_long ) CS = ZERO
contains
!----------------------------------------------------------------------------
! Spectral profiles
!                                 1        gamma/2
! - Lorentz profile: L(x,gamma) = -- * -----------------    with FWHM = gamma
!                                 pi   x^2 + (gamma/2)^2
!
!                               exp( -x^2 / 2 sigma^2 )
! - Gauss profile: G(x,sigma) = -----------------------
!                                 sigma * sqrt( 2 pi )
!
!                                          with FWHM = 2*sqrt(2*log(2)) sigma
!----------------------------------------------------------------------------
  function Lorentz_profile( x, gamma ) result (L)
    real(long) :: L
    real(long), intent(in) :: x, gamma
    L = gamma / TWO / PI / ( x**2 + gamma**2 / FOUR )
    if ( x == ZERO .and. gamma == ZERO ) L = ZERO
  end function Lorentz_profile

  function Gauss_profile( x, sigma ) result (G)
    real(long) :: G
    real(long), intent(in) :: x, sigma
    G = exp( -x**2 / TWO / sigma**2 ) / sigma / sqrt( TWO * PI )
  end function Gauss_profile
  
  function Flattop_profile( omega_in, omega, bandwidth ) result (F)
    real(long) :: F
    real(long), intent(in) :: omega_in, omega, bandwidth

    if ( abs(omega_in - omega) <= bandwidth ) then
       F = ONE / bandwidth / TWO
    else
       F = ZERO
    endif
  end function Flattop_profile

end function partial_PCS

function photoabsorption_CS( P, F, R, omega_in ) result ( CS )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  type(PCS_data), intent(in) :: R
  !type(Process_data), intent(inout) :: S
  real(long), intent(in) :: omega_in
  real(long) :: CS(F%N_subshell), AU2MB
  integer :: i, k, l, n_occ

  AU2MB = AU2M**2 / 1.0e-22_long

  CS(:) = ZERO
  do i = 1, F%N_subshell
     if ( P%yes_relativity ) then
        l = F%config_rel(i,2)
        n_occ = F%config_rel(i,4)
     else
        l = F%config(i,2)
        n_occ = F%config(i,3)
     endif
     if ( l == P%lmax .or. n_occ == 0 ) then
        cycle
     endif
     ! photoionization
     CS(i) = partial_PCS( P, F, R, omega_in, i ) 
     if ( P%yes_resonance ) then
        do k = 1, R%N_res(i)
           ! photoexcitation
           CS(i) = CS(i) + partial_PCS( P, F, R, omega_in, i, k ) 
        end do
!        if ( F%config_rel(i,1) == 2 .and. F%config_rel(i,2) == 0 ) then
!           write(99,'(e13.5,1x,e13.5,<R%N_res(i)>e13.5)')omega_in*AU2EV, &
!                partial_PCS( P, F, R, omega_in, i )*AU2MB, &
!                ( (partial_PCS( P, F, R, omega_in, i, k )*AU2MB), k=1, R%N_res(i) )
!        endif
     end if
     if ( CS(i) <= 1e-99_long ) CS(i) = ZERO

  end do
end function photoabsorption_CS

subroutine calculate_oscillator_str( P, G, F )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  real(long) :: E_trans, T, osci_str
  integer :: li, ji, lf, jf, ind_i, ind_f, N_occ, N_hole, l_greater
  integer :: i, j, OUTPUT
  real(long), external :: sixj

  OUTPUT = P%OUTPUT
  write(OUTPUT, "('#'/'# Oscillator strength:')")
  write(OUTPUT, "('#'/'#   i    - j     : ')", advance="no")
  write(OUTPUT, "(3a13)") "E_trans(a.u.)", "<i|r|j>", "oscil. str"

  initial: do i = 1, F%N_subshell            ! initial state
     if ( P%yes_relativity ) then
        if ( F%e_orb_rel(i) > ZERO ) cycle
        N_occ = F%config_rel(i,4) 
        if ( N_occ == 0 ) cycle
        li = F%config_rel(i,2)
        ji = F%config_rel(i,3) ! 2 * j quantum number of the initial state

     else
        if ( F%e_orb(i) > ZERO ) cycle
        N_occ = F%config(i,3)
        if ( N_occ == 0 ) cycle
        li = F%config(i,2)
     endif
     if ( li == P%lmax ) cycle
     final: do j = 1, F%N_subshell
        if ( P%yes_relativity ) then
           lf = F%config_rel(j,2)
           jf = F%config_rel(j,3) ! 2 * j quantum number of the final state
        else
           lf = F%config(j,2)
        endif
        
        ! Having obtained li, ji and lf, jf, we consider the selection rules
        if ( P%yes_relativity ) then
           if( abs(ji-jf) > 2 .or. abs(li - lf) /= 1 ) cycle
           if ( F%e_orb_rel(j) > ZERO ) cycle
        else
           if( abs(li - lf) /= 1 ) cycle
           if ( F%e_orb(j) > ZERO ) cycle
        endif

        E_trans = ionization_E( F, i ) - ionization_E( F, j )
        if ( E_trans < ZERO ) cycle     ! skip if final_E is lower than init_E
        if ( P%yes_relativity ) then
           N_hole = jf + 1 - F%config_rel(j,4) 
        else
           N_hole = 4*lf + 2 - F%config(j,3)
        endif

        if ( P%yes_relativity ) then
           ind_i = F%config_rel(i,5)
           ind_f = F%config_rel(j,5)
        else
           ind_i = i
           ind_f = j
        endif
        
        T = radial_integral( G, F%u(:,ind_i) * F%r(:) * F%u(:,ind_f) )

        ! cross_section = 2 * pi^2 * \alpha * osci_str
        l_greater = max(li,lf)
        if ( P%yes_relativity ) then
           osci_str = T**2 * E_trans * sixj(2*li, 1, ji, jf, 2, 2*lf)**2  &
                     * TWO * dble( N_occ * N_hole * l_greater ) / THREE 
        else
           osci_str = T**2 * E_trans * dble( N_occ * N_hole * l_greater ) &
                     / dble(2*li + 1) / dble(2*lf + 1) / THREE  
        endif

        !
        ! print this resonance line 
        ! if N_PE == 0, print it with the resonant absorption cross section
        !
        write(OUTPUT, "('#   ',a4,' - ',a5,' : ')", advance="no") &
             WF_name( F, i ), WF_name( F, j )
        write(OUTPUT, "(f13.5,2es13.5)") E_trans, T, osci_str
     end do final
  end do initial
end subroutine calculate_oscillator_str

end module pcs
