!----------------------------------------------------------------------------
! plasma.f90
!----------------------------------------------------------------------------
! Copyright (C) 2011-2023 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2011-2023 Sang-Kil Son <sangkil.son@cfel.de>
!          2011      Robert Thiele
!          2020      Rui Jin
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! plasma calculation
!   written by Son, Sang-Kil and Robert Thiele in Jul. 2011 at DESY
!
! Ref) Murillo and Weisheit, Phys. Rep. 302, 1 (1998)
!----------------------------------------------------------------------------
module plasma
  use constants
  use elec_config
  use read_param
  use coord_system
  use wavefunction
  use eigen_LAPACK
  use operations
  use interpolate
  implicit none
contains

!----------------------------------------------------------------------------
! OBSOLETE IMPLEMENTATION
!----------------------------------------------------------------------------
!subroutine plasma_screening( P, v, r )
!  type(Param), intent(in) :: P
!  real(long), intent(inout) :: v(:)
!  real(long), intent(in) :: r(:)
!  real(long) :: lambda, wigner, wioverlam
!  select case ( P%screening_type )
!    case ( 1 )          ! simple Debye model: Eq. (84) in Murillo98
!      lambda = P%debye_length
!      v(:) = v(:) * exp( -r(:) / lambda )
!    case ( 2 )          ! ion-sphere model: Eq. (97) in Murillo98
!      lambda = P%wigner_length
!      v(:) = v(:) * ( ONE - HALF*r(:)/lambda * ( THREE - (r(:)/lambda)**2 ) )
!    case ( 3 )          ! hybrid: Eq. (99-109) in Murillo98
!      lambda = P%debye_length
!      wigner = P%wigner_length
!      wioverlam = (wigner/lambda)**3 + ONE
!      where ( r(:) < P%rprime )
!        v(:) = v(:) * ( ONE - THREE*r(:)*lambda**2/(TWO*wigner**3)*&
!                        ( wioverlam**(TWO/THREE) - ONE ) + (r(:)/wigner)**3 )
!      elsewhere
!        v(:) = v(:) * ( THREE*lambda**2*P%rprime/wigner**3 * &
!                        exp( -( r(:) - P%rprime )/lambda ))
!      end where
!  end select
!end subroutine


!----------------------------------------------------------------------------
! Simple Debye screening factor
!----------------------------------------------------------------------------
subroutine Debye_screening( v, r, lambda )
  real(long), intent(inout) :: v(:)
  real(long), intent(in) :: r(:)
  real(long), intent(in) :: lambda      ! Debye length
  v(:) = v(:) * exp( -r(:) / lambda )
end subroutine Debye_screening

!----------------------------------------------------------------------------
! Screened Slater exchange potential: only HFS potential can be considered.
!
! Ref) Robinson et al., PRL 9, 215 (1962)
!      Muda & Sugata, Phys. Lett. 29A, 647 (1969)
!      Sachs, Acta Cryst. 22, 931 (1967) == Robinson et al. and simple
!----------------------------------------------------------------------------
function screened_potential_xc( rho, lambda ) result (v)
  real(long), intent(in) :: rho(:), lambda
  real(long) :: v(size(rho))
  real(long) :: K_F, alpha
  integer :: i
  !
  ! alpha = K_S / K_F   where K_F = ( 3 PI^2 rho )^(1/3)
  !                           K_S = 1 / lambda
  ! v_screened = -3 ( K_F / 2PI ) F( K_S / K_F )
  !            = v_Slater * F(alpha)
  !
  do i = 1, size(rho)
    if ( rho(i) < 1e-15_long ) then
      v(i) = ZERO
    else
      K_F = ( THREE * PI**2 * rho(i) )**(ONE/THREE)
      alpha = ONE / lambda / K_F
      v(i) = -THREE * K_F / TWO / PI * Func( alpha )
    end if
  end do
contains
  function Func( a ) result (f)
    real(long), intent(in) :: a
    real(long) :: f
    !
    ! Muda & Sugata, Phys. Lett. 29A, 647 (1969)
    !
    !f = HALF * ( ONE + ONE / ( ONE + a**2 ) &
    !             - a**2 / FOUR * ( a**4 + THREE * a**2 - FOUR ) / &
    !               ( ONE + a**2 )**2 * log( ONE + FOUR / a**2 ) &
    !             - a / ( ONE + a**2 )**2 * &
    !               ( PI + a**2 * ( TWO + a**2 ) * atan( TWO / a ) ) )
    !
    ! Robinson et al., PRL 9, 215 (1962)
    !
    !f = ( ONE - FOUR / THREE * a * atan( TWO / a ) &
    !          + HALF * a**2 * log( ONE + FOUR / a**2 ) &
    !          - a**2 / 6_long * ( ONE - a**2/FOUR * log( ONE + FOUR/a**2 ) ) )
    !
    ! Sachs, Acta Cryst. 22, 931 (1967) == Robinson et al.
    !
    f = ONE - a**2 / 6_long - FOUR/THREE * a * atan( TWO / a ) &
            + a**2 / TWO * ( ONE + a**2 / 12_long ) * log( ONE + FOUR / a**2 )
  end function Func
end function screened_potential_xc

!----------------------------------------------------------------------------
! Screened Coulomb potential (Yukawa or Debye-Huckel potential)
! using multipole expansion,
!   exp(-K*|r-r'|)/|r-r'| 
!     = 4 PI K SUM_{l=0}^{inf} i_l(K r<) k_l(K r>) 
!              SUM_{m=-l}^{l} Y_lm(theta,phi) Y*_lm(theta',phi')
!   where i_n(x) = sqrt( PI / 2 x ) I_{n+1/2}(x)
!         k_n(x) = sqrt( 2 / PI x ) K_{n+1/2}(x)
!
! For spherical symmetric rho(r'), i.e. l=0 and m=0
!   rho(r') * exp(-K*|r-r'|)/|r-r'| 
!     = K rho(r') i_0 ( K r< ) k_0( K r> )
!     = rho(r') / { 2 K r< r> } * ( exp[ -K (r>-r<) ] - exp[ -K (r>+r<) ] )
!
! Ref) Arfken's book, p565 and p733
!      Boschitsch, Fenley & Olson, J. Comput. Phys. 151, 212 (1999):
!        for the non-spherical case
!----------------------------------------------------------------------------
function screened_potential_J( G, rho, lambda ) result (v)
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: rho(:)
  real(long), intent(in) :: lambda
  real(long) :: v(G%N), K
  integer :: i, j_s
  j_s = size(rho)       ! it can be the max. # of grids or the index for r_s
  K = ONE / lambda
  do i = 1, G%N
    v(i) = HALF / K * integral( G, rho(1:j_s) / G%r(i) / G%r(1:j_s) * &
           ( exp(-abs(G%r(i)-G%r(1:j_s))*K ) - exp(-(G%r(i)+G%r(1:j_s))*K ) ) )
  end do
end function screened_potential_J

!----------------------------------------------------------------------
! solve the Schroedinger equation 
! for average-atom model or finite-temperature HFS
!----------------------------------------------------------------------
subroutine solve_SE_finite_temp( F, G, v_eff, T_eV )
  type(WF_data), intent(inout) :: F
  type(Coordinate), intent(inout) :: G
  real(long), intent(in) :: v_eff(:)
  real(long), intent(in) :: T_eV                        ! temperature in eV
  real(long) :: ff, ffp, x, dx, beta, exp_X
  integer :: i, j, k, l
  integer, parameter :: max_iter = 1000000  ! max. iteration for root-finding
  real(long), parameter :: EPS = 1e-8_long  ! for root-finding
  logical :: yes_Newton_Raphson
  !beta = ONE / ( BOLTZMANN_K_SI * J2AU * T_e )  ! T_e in K
  beta = ONE / ( T_eV * EV2AU ) ! remember that T_eV = T_e * k (J/K) * J2EV
  !
  ! solve the Schroedinger equation for all bound (and pseudo-continuum) states
  !
  F%pot(:) = v_eff(:)
  call solve_SE_all( F, G )
  !
  ! calculate normalization factor for all states
  ! integral over r < r_s
  !
  do i = 1, F%N_all
    F%norm(i) = radial_integral( G, F%u_all(1:F%j_s,i)**2 )
  end do
  !
  ! determine a chemical potential by using the Newton-Raphson method
  !   which needs f(x) and f'(x) = d f(x) / dx
  !   f(mu) = INT_(r_s) d^3 r rho(r) - N_elec --> root finding of f(mu) = 0
  !   rho(r) = SUM_p | psi_p(r) |^2 * g_p / ( 1 + exp(beta*(e_p-mu)) )
  ! Here, e_p -> e_orb, g_p -> 4l+2, | psi_p(r) |^2 -> norm, mu -> x
  !
  ! working expression:
  !   f(x) = SUM_i norm(i) * (4l+2) / ( 1 + exp(beta*(e_orb(i)-x)) )
  !   f'(x) = SUM_i norm(i) * (4l+2) * ( -beta * exp(beta*(e_orb(i)-x)) )
  !                                  / ( 1 + exp(beta*(e_orb(i)-x)) )**2
  ! in order to evaluate f'(x), there is one numerical issue 
  ! because exp(...) becomes either very close to zero or infinitely large.
  ! so, f'(x) should be rewritten as
  !   f'(x) = SUM_i norm(i) * (4l+2) * ( -beta ) 
  !                                  / ( 1 + exp(beta*(e_orb(i)-x)) )
  !                                  / ( 1 + 1/exp(beta*(e_orb(i)-x)) )
  !
  ! at the same time, determine fractional occupation numbers: n_occ(:)
  ! in fact, it is being updated during Newton-Raphson iterations
  !
  ! NOTE: If T is very close to zero, then the Newton-Raphson method fails,
  ! because f(mu) is more or less a step function, which means 
  ! it's almost flat except the very narrow region near the root.
  ! Therefore, if the Newton-Raphson method fails, it tries to use 
  ! the simplest maybe the dullest method to find a root.
  !
  x = F%chem_pot        ! initial guess from the argument of this function
                        ! we'd better consider calculated energy spectrum!!!
  dx = F%chem_pot * 0.01_long   ! initial step
  yes_Newton_Raphson = .true.   ! use the Newton-Raphson method first!
  root_finding: do k = 1, max_iter
    ff = -F%N_elec_rs   ! remember N_elec can be a fractional number
    ffp = ZERO
    do l = 0, F%lmax      ! compute f
      do j = 1, G%N
        i = l * G%N + j
        exp_X = exp( beta * ( F%e_all(i) - x ) )
        F%n_occ(i) = dble(4*l+2) / ( ONE + exp_X )      ! 4l+2: degeneracy for l
        ff = ff + F%norm(i) * F%n_occ(i)
        !
        ! be careful with ffp calculation
        ! since exp_X can be extremely large, 
        !   exp_X / ( 1 + exp_X ) ==> 1 / ( 1 + 1 / exp_X )
        !
        ffp = ffp + beta * F%norm(i) * F%n_occ(i) / ( ONE + ONE / exp_X )
      end do
    end do
    !
    ! if f' is too small, then turn off the Newton-Raphson method
    !
    if ( abs(ffp) < EPS ) yes_Newton_Raphson = .false.
    !
    ! determine dx according to Newton-Raphson or the dullest method
    !
    if ( yes_Newton_Raphson ) then
      dx = - ff / ffp           ! the core part of the Newton-Raphson method
    else
      !
      ! Here is the core part of the dullest root-finding method
      ! - the sign of dx must be changed according to the sign of f
      !   if ff > 0, then dx must be negative (going backward)
      !   if ff < 0, then dx must be positive (going forward)
      ! - whenever the sign of dx is changed, dx becomes smaller
      !
      if ( ( ff > ZERO .and. dx > ZERO ) .or. &
           ( ff < ZERO .and. dx < ZERO ) ) dx = -dx * 0.1_long
    end if
    if ( abs(dx) < EPS .or. abs(ff) < EPS ) exit root_finding
    if ( F%yes_fix_chem_pot ) exit root_finding  ! no iteration when mu is fixed
    x = x + dx
  end do root_finding
  if ( k >= max_iter ) then
    write(STDERR,*) "Root_finding: not converged after ",max_iter," iterations"
    write(STDERR,*) "  x    = ", x, "  x0 = ", F%chem_pot
    write(STDERR,*) "  f(x) = ", ff
    write(STDERR,*) "  dx   = ", dx
    stop
  end if
  F%chem_pot = x        ! we found a new chemical potential!
  !
  ! The sea level (e_s) is defined by the asymptotic end point of the potential.
  ! If muffin-tin is used, the sea level is defined at r_s(Wigner-Seitz radius).
  !
  F%e_s = v_eff(G%N)
  !
  ! put eigenstates into initially occupied orbitals: F%e_orb(j) and F%u(:,j)
  ! also put F%n_occ into F%f_occ(j)
  !
  do j = 1, F%N_orb
    l = F%config(j,2)
    i = l * F%N + ( F%config(j,1) - l )
    F%e_orb(j) = F%e_all(i)
    F%u(:,j)   = F%u_all(:,i)
    F%f_occ(j) = F%n_occ(i)
  end do
end subroutine solve_SE_finite_temp

subroutine construct_density_finite_temp( F, rho_bound, rho_free, &
                                          yes_fixed_config )
  type(WF_data), intent(in) :: F
  real(long), intent(out) :: rho_bound(:), rho_free(:)  ! densities
  logical, intent(in) :: yes_fixed_config
  integer :: i, j, l
  !
  ! construct density according to the Fermi-Dirac distribution F%n_occ(:)
  ! if yes_fixed_config is yes, bound config. is given by "-config".
  ! otherwise, bound config. is given by comparing e_orb <= e_s
  !
  rho_bound(:) = ZERO
  rho_free(:) = ZERO
  if ( yes_fixed_config ) then
    do i = 1, F%N_all
      rho_free(:) = rho_free(:) + abs( F%u_all(:,i)**2 ) * F%n_occ(i)
    end do
    do j = 1, F%N_orb
      l = F%config(j,2)
      i = l * F%N + ( F%config(j,1) - l )
      rho_free(:)  = rho_free(:)  - abs( F%u_all(:,i)**2 ) * F%n_occ(i)
      rho_bound(:) = rho_bound(:) + abs( F%u_all(:,i)**2 ) * F%n_occ(i)
    end do
  else
    do i = 1, F%N_all
      if ( F%e_all(i) <= F%e_s ) then
        rho_bound(:) = rho_bound(:) + abs( F%u_all(:,i)**2 ) * F%n_occ(i)
      else 
        rho_free(:) = rho_free(:) + abs( F%u_all(:,i)**2 ) * F%n_occ(i)
      end if
    end do
  end if
  !
  ! density should be spherically averaged: see operations.f90
  !
  rho_bound(:) = rho_bound(:) / F%r(:)**2 / FOUR / PI
  rho_free(:)  = rho_free(:)  / F%r(:)**2 / FOUR / PI
end subroutine construct_density_finite_temp

!----------------------------------------------------------------------------
! grand canonical partition function Z_G; a log value is calculated
!----------------------------------------------------------------------------
function log_partition_function( F, T_eV, mu_au ) result (log_ZG)
  real(long) :: log_ZG
  type(WF_data), intent(in) :: F
  real(long), intent(in) :: T_eV        ! temperature in eV
  real(long), intent(in) :: mu_au       ! mu (chemical potential) in a.u.
  integer :: i, j, l
  real(long) :: beta, A, X, exp_X
  beta = ONE / ( T_eV * EV2AU ) ! remember that T_eV = T_e * k (J/K) * J2EV
  log_ZG = ZERO
  !
  ! zeroth-order
  !
  do l = 0, F%lmax
    do j = 1, F%N
      i = l * F%N + j
      X = beta * ( F%e_all(i) - mu_au )
      exp_X = exp( X )
      if ( exp_X < 1e-99_long ) then
        A = - X
      else
        A = log( ONE + ONE / exp_X )
      end if
      log_ZG = log_ZG + A * dble(4*l+2) &       ! (4l+2) : degeneracy for l
                          * F%norm(i)   ! multiplied by norm (EXPERIMENTAL)
    end do
  end do
end function log_partition_function

!----------------------------------------------------------------------------
! Pair correlation function for ionic contribution
!   g(r) = 1 + a*exp(-b*r) * sin(c*r-d)
!----------------------------------------------------------------------------
function pair_correlation( r, q ) result (g)
  real(long), intent(in) :: r(:), q(4)
  real(long) :: g(size(r))
  integer :: i
  g(:) = ONE + q(1) * exp( -q(2) * r(:) ) * sin( q(3) * r(:) - q(4) )
  do i = size(r), 1, -1 ! scanned from the end of r
    if ( g(i) <= ZERO ) exit
  end do
  g(1:i) = ZERO         ! put zero after the crossing point
end function pair_correlation

!----------------------------------------------------------------------------
! Print information of atom-in-plasma method
!----------------------------------------------------------------------------
subroutine print_atom_in_plasma( P, G, g_i, adj_e_dens, adj_r_s )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: g_i(:), adj_e_dens, adj_r_s
  real(long) :: r_m, n_den, factor, dens
  integer :: i, OUTPUT
  OUTPUT = P%OUTPUT
  !
  ! factor: number density --> density for nucleus
  !
  factor = P%mass * AU2KG * 1000_long / ( AU2M * 100_long )**3
  !
  ! print the type of screening model
  !
  write(OUTPUT, "('# Atom-in-Plasma model: ')", advance="no")
  if ( P%yes_ion_screening ) then
    write(OUTPUT, "(a)", advance="no") "neighboring ions + "
  end if
  if ( P%yes_debye_screening .and. P%yes_n_elec_screening ) then
    write(OUTPUT, "(a)") "** non-sense: BOTH Debye and n_elec screenings **"
  else if ( P%yes_debye_screening ) then
    write(OUTPUT, "(a)") "Debye screening"
  else if ( P%yes_n_elec_screening ) then
    write(OUTPUT, "(a)") "n_elec screening"
  else if ( P%yes_average_atom ) then
    write(OUTPUT, "(a)", advance="no") "average-atom"
    if ( P%yes_wigner_seitz ) write(OUTPUT,"(a)",advance="no") " + Wigner-Seitz"
    if ( P%yes_muffin_tin )   write(OUTPUT,"(a)",advance="no") " + muffin-tin"
    write(OUTPUT, "()")
  else if ( P%yes_wigner_seitz ) then
    write(OUTPUT, "(a)", advance="no") "Wigner-Seitz"
    if ( P%yes_muffin_tin )   write(OUTPUT, "(a)", advance="no") " + muffin-tin"
    write(OUTPUT, "()")
  else if ( P%yes_quasineutral ) then
    write(OUTPUT, "(a)") "quasineutral"
  else if ( P%yes_xpot ) then
    write(OUTPUT, "(a)") "XPOT with environmental potential"
  else
    write(OUTPUT, "()")
  end if
  !
  ! for plasma calculation with external environment potential from XPOT
  !
  if ( P%yes_xpot ) then
    if ( P%yes_load_env ) then
      write(OUTPUT, "('#   - load file_env (output of XPOT): ',a)") &
        trim(P%file_env)
    end if
    if ( P%yes_wigner_seitz ) &
      write(OUTPUT, "('#   - yes_wigner_seitz : ',l1)") P%yes_wigner_seitz
    write(OUTPUT, "('#   - yes_fixed_pot    : ',l1)") P%yes_fixed_pot
    write(OUTPUT, "('#   - flat_tail: ')", advance="no")
    select case ( P%flat_tail_type )
      case ( 1 )
        write(OUTPUT, "('WS (wigner-seitz sphere)')")
        write(OUTPUT, "('#     the potential becomes flat after r_WS')") 
      case ( 2 )
        write(OUTPUT, "('TS (touching sphere) with a fixed pot')")
        write(OUTPUT, "('#     the potential becomes flat " // &
                      "with a fixed value V_L after V >= V_L')")
        write(OUTPUT, "('#     V_L: determined by touching-sphere scheme " // &
                      "and specified in file_env')")
      case ( 3 )
        write(OUTPUT, "('EP (effective potential)')")
        if ( P%yes_load_pot .and. P%SCF_max_iter == 1 ) then
          write(OUTPUT, "('#   - load file_pot: ',a)") trim(P%file_pot)
          write(OUTPUT, "('#     after loading effective potential, " // &
                        "no SCF will be performed')")
        end if
        write(OUTPUT, "('#     for r <= rmax (bound states), " // &
                      "the whole effective potential is used')")
        write(OUTPUT, "('#     for r >  rmax (continuum states), " // &
                      "a flat tail is set by V_eff(r=rmax)')")
    end select
    write(OUTPUT, "('#   NOTE: for XPOT calculation, " // &
                  "the options below are ideally all F')", advance="no")
    if ( P%yes_wigner_seitz ) write(OUTPUT, "(' (except WS)')", advance="no")
    write(OUTPUT, "(/'#')")
  end if
  !
  ! print all options related to the screening model
  !
  write(OUTPUT, "('#   checking option variables: -yes_[option]')")
  write(OUTPUT, "('#   - debye_screening  : ',l1,4x," // & 
                     "'- latter       : ',l1,4x," // &
                     "'- average_atom : ',l1)") &
    P%yes_debye_screening, P%yes_latter, P%yes_average_atom
  write(OUTPUT, "('#   - ion_screening    : ',l1,4x," // & 
                     "'- muffin_tin   : ',l1,4x," // &
                     "'- quasineutral : ',l1)") &
    P%yes_ion_screening, P%yes_muffin_tin, P%yes_quasineutral
  write(OUTPUT, "('#   - n_elec_screening : ',l1,4x," // & 
                     "'- wigner_seitz : ',l1,4x," // &
                     "'- load_den     : ',l1)") &
    P%yes_n_elec_screening, P%yes_wigner_seitz, P%yes_load_den
  write(OUTPUT, "('#   - grand_canonical  : ',l1,4x," // & 
                     "'- fixed_config : ',l1,4x," // &
                     "'- scaling      : ',l1)") &
    P%yes_grand_canonical, P%yes_fixed_config, P%yes_scaling
  write(OUTPUT, "('#   - uniform_ion_dens : ',l1,4x," // & 
                     "'- turnoff_V_ne : ',l1,4x," // &
                     "'- turnoff_V_ee : ',l1)") &
    P%yes_uniform_ion_dens, P%yes_turn_off_V_ne, P%yes_turn_off_V_ee
  write(OUTPUT, "('#   - flag (used for anything) : ',l1)") P%yes_flag
  !
  ! print parameters for neighboring ions
  !
  if ( P%yes_ion_screening ) then
    do i = 1, G%N
      if ( g_i(i) > ZERO ) exit
    end do
    r_m = G%r(i)
    n_den = ONE / integral( G, ONE - g_i(:) )
    write(OUTPUT, "('#   pair correlation function for neighboring ions')")
    write(OUTPUT, "('#     g(r) = 1 +',f7.3,' * exp( -',f7.3,' * r ) * '," // &
                  "'sin( ',f7.3,' * r -',f7.3,' )')") P%pcf(:)
    write(OUTPUT, "('#     g(r) = 0 for r < ',f10.5)") r_m
    write(OUTPUT, "('#   effective ionic charge = ',sp,f9.3)") P%Z_eff
    if ( P%yes_quasineutral ) then
      write(OUTPUT, "('#   [quasineutral] elec density = " // & 
                    "ionic density * eff. ionic charge')")
    end if
  end if
  !
  ! print parameters for Debye screening
  !
  if ( P%yes_debye_screening ) then
    write(OUTPUT, "('#   Debye length(a.u.)  = ',f8.1)") P%debye_length
  end if
  !
  ! print ionic density
  !
  if ( P%yes_ion_screening .or. P%yes_wigner_seitz .or. &
       P%yes_n_elec_screening ) then
    write(OUTPUT, "('#   ionic density (g/cm^3)          = ',f10.5)") &
      P%ion_dens * factor
    write(OUTPUT, "('#   ionic number density (1/a.u.^3) = ',es12.3,4x," // &
                  "'[',es11.3,' m^-3 ]')") P%ion_dens, P%ion_dens / AU2M**3
    !
    ! print parameters for Wigner-Seitz radius
    !
    if ( P%yes_wigner_seitz ) then
      dens = THREE / FOUR / PI / adj_r_s**3 
      write(OUTPUT, "('#   Wigner-Seitz radius (a.u.) = ',f8.3)") P%r_s
      write(OUTPUT, "('#   adjusted r_s (on a grid)   = ',f13.8)") adj_r_s
      if ( dens /= P%ion_dens ) &
        write(OUTPUT, "('#   => corres. ion_dens (1/a.u.^3)  = ',es12.3,4x," //&
                      "'[',f10.5,' g/cm^3 ]')") dens, dens * factor
    end if
  end if
  !
  ! print parameters for n_elec screening
  !
  if ( P%yes_n_elec_screening ) then
    write(OUTPUT, "('#   elec. number density (1/a.u.^3) = ',es11.3,5x," // &
                  "'[',es11.3,' m^-3 ]')") P%elec_dens, P%elec_dens / AU2M**3
    if ( P%yes_wigner_seitz ) &
      write(OUTPUT, "('#   adjusted elec_dens ',12x,' = ',es16.8)") adj_e_dens
  end if
  !
  ! print parameters for average_atom
  !
  if ( P%yes_average_atom ) then
    write(OUTPUT, "('#   temperature (eV)',15x,' = ',f8.3,8x," // & 
                  "'[',es10.2,' Kelvin ]')") P%T_eV, P%T_eV * EV2K
  end if
  !
  ! file information
  !
  if ( P%yes_average_atom ) then
    write(OUTPUT, "('#   free elec. density data saved: ',a)") trim(P%file_den)
  else if ( P%yes_load_den ) then
    write(OUTPUT, "('#   free elec. density data loaded: ',a)") trim(P%file_den)
  end if
  !
  ! at the end
  !
  write(OUTPUT, "('#')")
end subroutine print_atom_in_plasma

!----------------------------------------------------------------------------
! Print grand-canonical ensemble analysis
!----------------------------------------------------------------------------
subroutine print_average_atom( P, F, T_eV, OUTPUT )
  type(Param), intent(in) :: P
!  type(WF_DATA), intent(in) :: F
  type(WF_DATA), intent(inout) :: F
  real(long), intent(in) :: T_eV        ! temperature
  integer, intent(in) :: OUTPUT
  real(long) :: beta, prob, prob_sum, avg_charge
  real(long) :: n1      ! charge calc. based on initially occ. orbitals
  real(long) :: n2      ! charge calc. based on bound orbitals in solid
  real(long) :: n3      ! charge calc. based on bound orbitals + r_s
  real(long) :: n_occ_all
  real(long) :: n_l(0:F%lmax)
  integer :: i, j, k, l, m, q
  integer :: N_elec     ! # of electrons == P%Z for neutrality
  integer :: N_bound    ! # of bound spatial orbitals
  integer :: N_config   ! # of all possible configurations for bound electrons
  integer, allocatable :: bound_max_occ(:), bound_occ(:)
  real(long), allocatable :: bound_e_orb(:), prob_charge(:)
  integer, allocatable :: mp_config_occ(:,:) ! most probable config for given q
  real(long), allocatable :: prob_mp_config(:)
  integer :: i_mp_charge, i_mp_config
  real(long) :: mp_charge, mp_config
  real(long) :: binom   ! defined in anglib.f90
  real(long) :: exp_X
  real(long) :: pressure, pressure_bar, volume
  real(long) :: log_ZG   ! log( partition function )
  real(long) :: E_sum, N0
  integer, allocatable :: config(:,:)
  logical :: yes_print, yes
  integer :: OUTPUT_CONFIG = 20
  character(len=500) :: string_mp_charge_config, string_mp_config
  !N_elec = sum( F%config(:,3) ) ! 3 -> occupation number
  N_elec = P%Z          ! to define the atomic charge state
  beta = ONE / ( T_eV * EV2AU ) ! remember that T_eV = T_e * k (J/K) * J2EV
  !
  ! calculate the sum of occupation numbers, i.e., averaged charge
  ! find out the number of bound states ==> N_bound
  !
  write(OUTPUT, "('#')")
  yes_print = .false.
  N_bound = 0
  n1 = ZERO;  n2 = ZERO;  n3 = ZERO
  n_l(0:) = ZERO
  do l = 0, F%lmax
    do j = 1, F%N
      i = l * F%N + j
      n_l(l) = n_l(l) + F%n_occ(i) * F%norm(i)
      if ( F%e_all(i) <= F%e_s ) then   ! this spatial orbital is bound
        N_bound = N_bound + 1
        n2 = n2 + F%n_occ(i)
        n3 = n3 + F%n_occ(i) * F%norm(i)
        yes = .true.
        do k = 1, F%N_orb               ! check if it's already printed
          if ( F%config(k,1) == j+l .and. F%config(k,2) == l ) yes = .false.
        end do
        if ( yes ) then
          write(OUTPUT, "('#   ',a4,f6.3,f16.8,f12.2,a12,f12.5)") &
            orbital_name( j+l, l ), F%n_occ(i), &
            F%e_all(i), F%e_all(i) * AU2EV, "---", F%norm(i)
          yes_print = .true.
        end if
      end if
    end do
  end do
  if ( yes_print ) write(OUTPUT, "('#')")
  do j = 1, F%N_orb
    l = F%config(j,2)
    i = l * F%N + ( F%config(j,1) - l )
    n1 = n1 + F%n_occ(i)
  end do
  n_occ_all = sum( F%n_occ(:) )
  write(OUTPUT, "('# - sum of occupation numbers: ')")
  write(OUTPUT, "('#   in all available orbitals   = ')", advance="no")
  if ( n_occ_all < 1e6_long ) then
    write(OUTPUT, "(f10.4)") n_occ_all
  else
    write(OUTPUT, "(es13.5)") n_occ_all
  end if
  write(OUTPUT, "('#   in initially occ. orbitals  = ',f10.4,4x,'/ '," // &
                "'eff. charge = ',sp,f10.4)") n1, dble(P%Z) - n1
  write(OUTPUT, "('#   in bound orb (E < pot_tail) = ',f10.4,4x,'/ '," // &
                "'eff. charge = ',sp,f10.4)") n2, dble(P%Z) - n2
  write(OUTPUT, "('#   in bound orb and r < r_s    = ',f10.4,4x,'/ '," // &
                "'eff. charge = ',sp,f10.4)") n3, dble(P%Z) - n3
  F%Q_avg = dble(P%Z) - n2      ! use the second definition in the aboves
  write(OUTPUT, "('#   average charge state Q_avg  = ',sp,f10.4)") F%Q_avg
  write(OUTPUT, "('#   sum of occ. numbers (r < r_s) for each l: 0,1,2,...')")
  write(OUTPUT, "('# |OCC_L| ',999es10.3)") n_l(0:)
!TEST
!  !
!  ! find resonance states <--- think about naming!
!  !
!  write(OUTPUT, "('#')")
!  write(OUTPUT, "('# Find resonance states: print norm < 1e-5')")
!  do l = 0, F%lmax
!    write(OUTPUT, "('# - l = ',i5)") l
!    do j = 1, F%N
!      i = l * F%N + j
!      if ( F%norm(i) > 1e-5_long ) then
!        write(OUTPUT, "(2es13.5)") F%e_all(i), F%norm(i)
!      end if
!    end do
!    write(OUTPUT, "(/)")
!  end do
!TEST
  !
  ! print out all orbital information: occ. number, norm. factor, and orbital E
  ! NOTE: it's too large, so it's printed only if the yes_verbose option is on
  !
  if ( P%yes_verbose) then
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# - individual occ. number: i=1,2,...')")
    where ( F%n_occ(:) <= 1e-99_long) F%n_occ(:) = ZERO
    do l = 0, F%lmax
      write(OUTPUT, "('# |OCC   l=',i3,'| ',999es12.4)") l, &
        F%n_occ(l*F%N+1:(l+1)*F%N)
    end do
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# - individual normalization factor: i=1,2,...')")
    do l = 0, F%lmax
      write(OUTPUT, "('# |NORM  l=',i3,'| ',999es12.4)") l, &
        F%norm(l*F%N+1:(l+1)*F%N)
    end do
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# - individual orbital energy: i=1,2,...')")
    do l = 0, F%lmax
      write(OUTPUT, "('# |ORB_E l=',i3,'| ',999es12.4)") l, &
        F%e_all(l*F%N+1:(l+1)*F%N)
    end do
  end if
  !
  ! print out the grand partition function
  !   log_ZG, E_sum, N0: multiplied by norm (EXPERIMENTAL)
  !
  log_ZG = log_partition_function( F, T_eV, F%chem_pot )
  E_sum = sum( F%e_all(:) * F%n_occ(:) * F%norm(:) )
  N0 = sum( F%n_occ(:) * F%norm(:) )
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Grand partition function (zeroth order)')")
  write(OUTPUT, "('# * all of them multiplied by norm. factor " // &
                "before summed up')")
  write(OUTPUT, "('#   - partition func log_ZG(0)   = ',es13.5)") log_ZG
  write(OUTPUT, "('#   - average elec number <N>(0) = ',es13.5)") N0
  write(OUTPUT, "('#   - orbital energy sum E_sum   = ',es13.5)") E_sum
  write(OUTPUT, "('# * units are a.u. if necessary')")
  write(OUTPUT, "('# ',11x,5a13)") "beta", "mu", "log_ZG(0)", "<N>(0)", "E_sum"
  write(OUTPUT, "('# |log_ZG|   ',5es13.5)") &
    beta, F%chem_pot, log_ZG, N0, E_sum
  !
  ! calculate pressure according to Gibbs-Duhem relation
  !
  volume = FOUR * PI / THREE * P%r_s**3
  pressure = log_ZG / volume / beta
  pressure_bar = pressure * AU2KG / AU2M / AU2SEC**2 * 1e-5_long
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# - pressure (a.u.) = ',es13.5,5x,'[ ',es13.5,' bar ]')") &
    pressure, pressure_bar
  !
  ! print out <H>
  !
  !E_avg = sum( F%e_all(:) * F%n_occ(:) )
  !write(OUTPUT, "('#')")
  !write(OUTPUT, "('# - averaged total energy (eV), <H> = ',es13.5)") &
  !  E_avg * AU2EV
  !
  ! The rest is for the grand-canonical ensemble analysis.
  !
  if ( .not. P%yes_grand_canonical ) return
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Grand canonical ensemble analysis: ')", advance="no")
  if ( P%yes_fixed_config ) then
    write(OUTPUT, "('bound states chosen by -config')")
  else
    write(OUTPUT, "('bound states chosen by e_orb < pot_tail')")
  end if
  if ( P%yes_verbose ) then
    open( OUTPUT_CONFIG, file=trim(P%filename) // ".config" )
    write(OUTPUT, "('# - all fixed configurations printed out in ',a)") &
      trim(P%filename) // ".config"
  end if
  !
  ! set up the bound configuration
  !   i) given by the -config option (when P%yes_fixed_config is on)
  !   ii) automatically chosen according to e_orb < potential tail
  !
  if ( P%yes_fixed_config ) N_bound = F%N_orb
  allocate( config(N_bound,3) )
  allocate( bound_max_occ(N_bound), bound_occ(N_bound), bound_e_orb(N_bound) )
  if ( P%yes_fixed_config ) then
    !
    ! i) bound config. is specified by the -config option
    !
    config(:,:) = F%config(:,:)
    do j = 1, F%N_orb
      l = F%config(j,2)
      i = l * F%N + ( F%config(j,1) - l )
      bound_max_occ(j) = 2 * ( 2 * l + 1 )
      bound_e_orb(j) = F%e_all(i)
    end do
  else
    !
    ! ii) bound config. is automatically determined by e_orb < pot_tail
    !
    k = 0
    do l = 0, F%lmax
      do j = 1, F%N
        i = l * F%N + j
        if ( F%e_all(i) <= F%e_s ) then  ! this spatial orbital is bound
          k = k + 1
          bound_max_occ(k) = 2 * ( 2 * l + 1 )
          bound_e_orb(k) = F%e_all(i)
          config(k,1) = j + l
          config(k,2) = l
        end if
      end do
    end do
    !
    ! ordering of bound states according to orbital energies
    !
    do k = 1, N_bound
      do j = k + 1, N_bound
        if ( bound_e_orb(j) < bound_e_orb(k) ) then
          call swap_r( bound_e_orb(j), bound_e_orb(k) )
          call swap_i( bound_max_occ(j), bound_max_occ(k) )
          call swap_i( config(j,1), config(k,1) )
          call swap_i( config(j,2), config(k,2) )
        end if
      end do
    end do
  end if
  write(OUTPUT, "('#',18x,100a3)") &
    (/ ( orbital_name( config(k,1), config(k,2) ), k = 1, N_bound ) /)
  !
  ! find the number of all possible configurations of bound electrons
  ! and make a do-loop with that number
  !
  allocate( prob_charge( N_elec-sum(bound_max_occ(:)):N_elec ) )
  allocate( prob_mp_config( N_elec-sum(bound_max_occ(:)):N_elec ) )
  allocate( mp_config_occ( N_bound, N_elec-sum(bound_max_occ(:)):N_elec ) )
  prob_charge(:) = ZERO
  prob_mp_config(:) = ZERO
  prob_sum = ZERO
  N_config = product( bound_max_occ(:) + 1 )
  do i = 1, N_config
    !
    ! construct all possible config. of bound electrons --> bound_occ(:)
    ! for given bound spatial orbitals
    ! The following do-loop is a kind of MAGIC routine!  Just it works!
    !
    k = i
    do j = N_bound, 1, -1
      m = mod( k-1, bound_max_occ(j)+1 )
      bound_occ(j) = bound_max_occ(j) - m
      k = ( k - m ) / ( bound_max_occ(j)+1 ) + 1
    end do
    !
    ! calculate a probability of finding this configuration
    ! binom(): defined in anglib.f90
    !
    prob = ONE
    do j = 1, N_bound
      exp_X = exp( beta * ( bound_e_orb(j) - F%chem_pot ) )
      prob = prob * binom( bound_max_occ(j), bound_occ(j) ) &
             / ( ONE + exp_X )**bound_occ(j) &
             * ( exp_X / ( ONE+exp_X ) )**(bound_max_occ(j)-bound_occ(j))
      !prob = prob * binom( bound_max_occ(j), bound_occ(j) ) &
      !       * exp( -beta * ( bound_e_orb(j) - F%chem_pot ) )**bound_occ(j) &
      !       / ( ONE+exp(-beta*(bound_e_orb(j)-F%chem_pot)) )**bound_max_occ(j)
    end do
    !
    ! print out this probability
    !
    q = N_elec - sum( bound_occ(:) )
    prob_charge(q) = prob_charge(q) + prob
    if ( prob > prob_mp_config(q) ) then
      prob_mp_config(q) = prob
      mp_config_occ(:,q) = bound_occ(:)
    end if
    prob_sum = prob_sum + prob
    if ( prob >= 1e-4_long ) then
      write(OUTPUT, "('#   config (',sp,i3,s,'): ',100i3)", advance="no") &
        q, bound_occ(:)
      write(OUTPUT, "(3x,f8.4)") prob
    end if
    if ( P%yes_verbose ) then
      config(:,3) = bound_occ(:)
      write(OUTPUT_CONFIG, "(i3,x,a,es16.8e3)") &
        q, trim( config_to_string( config ) ), prob
    end if
  end do
  !
  ! print out probabilities of finding a certain charge
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('#',a9,a12,4x,a12,2x,a)") "charge", "charge_prob", &
    "config_prob", "the most probable config for given q"
  avg_charge = ZERO
  mp_charge = ZERO
  mp_config = ZERO
  do q = N_elec - sum( bound_max_occ(:) ), N_elec
    avg_charge = avg_charge + dble(q) * prob_charge(q)
    if ( prob_charge(q) >= 1e-8_long ) then
      config(:,3) = mp_config_occ(:,q)
      write(OUTPUT, "(sp,i10,s,f12.8,4x,f12.8,2x,a)") q, prob_charge(q), &
        prob_mp_config(q), trim( config_to_string( config ) )
    end if
    if ( prob_charge(q) > mp_charge ) then    ! set up the most probable charge
      mp_charge = prob_charge(q)
      i_mp_charge = q
    end if
    if ( prob_mp_config(q) > mp_config ) then ! set up the most probable config
      mp_config = prob_mp_config(q)
      i_mp_config = q
    end if
  end do
  config(:,3) = mp_config_occ(:,i_mp_charge)  ! MP config of MP charge
  string_mp_charge_config = trim( config_to_string( config ) )
  config(:,3) = mp_config_occ(:,i_mp_config)  ! MP config of all charges
  string_mp_config = trim( config_to_string( config ) )
  write(OUTPUT, "('#',a9,f12.8)") "sum:", prob_sum
  write(OUTPUT, "('#',a9,f12.8,4x,12x,2x,a,' (',sp,i3,')')") "max:", &
    prob_charge(i_mp_charge), trim( string_mp_charge_config ), i_mp_charge
  write(OUTPUT, "('#',9x,12x,a4,f12.8,2x,a,' (',sp,i3,')')") "max:", &
    prob_mp_config(i_mp_config), trim( string_mp_config ), i_mp_config
  write(OUTPUT, "('# - avg. charge = ',sp,f12.8)") avg_charge
  write(OUTPUT, "('# - the most probable charge (MPQ)  = ',sp,i5,s,f12.8)") &
    i_mp_charge, prob_charge(i_mp_charge)
  write(OUTPUT, "('# - the most probable config of MPQ = ',a,2x,f12.8)") &
    trim( string_mp_charge_config ), prob_mp_config(i_mp_charge)
  write(OUTPUT, "('# - the most probable config of all = ',a,2x,f12.8)") &
    trim( string_mp_config ), prob_mp_config(i_mp_config)
contains
  subroutine swap_r( a, b )
    real(long), intent(inout) :: a, b
    real(long) :: temp
    temp = a;  a = b;  b = temp
  end subroutine swap_r
  subroutine swap_i( n, m )
    integer, intent(inout) :: n, m
    integer :: temp
    temp = n;  n = m;  m = temp
  end subroutine swap_i
end subroutine print_average_atom

subroutine determine_WS( G, r_s, j_s, elec_dens )
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: r_s
  integer, intent(inout) :: j_s
  real(long), intent(inout) :: elec_dens
  integer :: i
  do j_s = 1, G%N
    if ( G%r(j_s) >= r_s ) exit
  end do
  j_s = j_s - 1
  !
  ! adjust elec_dens according to (adj)r_s = G%r(j_s)
  !                                        int_{r<=r_s} 1 d^3 r
  !   (adj)elec_dens = (orig)elec_dens * -------------------------
  !                                      int_{r<=(adj)r_s} 1 d^3 r
  !   numerator = 4 pi / 3 * r_s**3
  !   denominator = numerically caculated by integral( G, ... )
  !
  elec_dens = elec_dens * ( FOUR*PI/THREE * r_s**3 ) &
                        / integral( G, (/ ( ONE, i = 1, j_s ) /) )
end subroutine determine_WS

!----------------------------------------------------------------------------
! For plasma environment potential, considering the touching sphere of
! muffin-tin model VL is determined from ab initio XMDYN-XPOT output v_eff
! should be no greater than it.
!
! Historically, this implementation was called "muffin-tin" model because it
! imposes a flat pontential at the end. However, it's different from the
! original muffin-tin implementation, where r_c is given and the potential tail
! is given by V(r_c). In this subroutine, VL is given (typically from the
! touching scheme, but not necessarily) and the potential tail is
! attached with VL. 
! ==> The subroutine name is changed from touch_muffin() to apply_fixed_pot().
!----------------------------------------------------------------------------
subroutine apply_fixed_pot( G, v_eff, VL_fixed_pot, j_s )
  type(Coordinate), intent(in) :: G
  integer, intent(inout) :: j_s
  real(long), intent(inout) :: v_eff(:)
  real(long), intent(in) :: VL_fixed_pot
  do j_s = 1, G%N
    if ( v_eff(j_s) >= VL_fixed_pot ) exit
  end do
  j_s = j_s - 1
  !print *,"r_ws", G%r(j_s)
  v_eff(j_s+1:) = VL_fixed_pot
! Different from AA calculations where the free-electron screening effect are
! self consistent to compensate the bound electron leakage, in our
! XATOM-XPOT-XMDYN calculation, the v_env is calculated externally with xpot,
! there is no free electron density contribution all electrons should be treated 
! as bound(or quasibound). 
  j_s = G%N
end subroutine apply_fixed_pot

!----------------------------------------------------------------------------
! load / save free-electron density
! - save: don't do anything; no scaled or normalized
!         but print out the normalization value
! - load: currently, the density is normalized with respect to r_s
!         and then scaled up to satisfy the neutrality
!----------------------------------------------------------------------------
subroutine load_free_elec( filename, rho_free )
  !type(Coordinate), intent(in) :: G
  character(len=*), intent(in) :: filename
  real(long), intent(out) :: rho_free(:)
  !integer, intent(in) :: j_s, N_free
  real(long) :: value
  integer, parameter :: IO_DEN = 50
  integer :: i
  open( IO_DEN, file=trim(filename), action="read", form="formatted" )
  do i = 1, size(rho_free)
    read( IO_DEN, * ) value, rho_free(i)        ! let's do interpolation later
  end do
  close( IO_DEN )
  !
  ! double-check whether rho_free is normalized w.r.t j_s.
  !
  !norm = integral( G, rho_free(1:j_s) )
  !rho_free(:) = rho_free(:) / norm * dble(N_free)
end subroutine load_free_elec

subroutine save_free_elec( G, filename, rho_free, j_s )
  type(Coordinate), intent(in) :: G
  character(len=*), intent(in) :: filename
  real(long), intent(in) :: rho_free(:)
  integer, intent(in) :: j_s
  real(long) :: norm
  integer, parameter :: IO_DEN = 50
  integer :: i
  norm = integral( G, rho_free(1:j_s) )
  open( IO_DEN, file=trim(filename), action="write", form="formatted" )
  do i = 1, size(rho_free)
    write( IO_DEN, * ) G%r(i), rho_free(i) ! / norm
  end do
  write( IO_DEN, "('# integral within r_s = ',f13.8)" ) norm
  write( IO_DEN, "('# r_s (a.u.) = ',es13.5)" ) G%r(j_s)
  close( IO_DEN )
end subroutine save_free_elec

!subroutine apply_free_elec( G, rho, rho_free, j_s, Z )
!  type(Coordinate), intent(in) :: G
!  real(long), intent(inout) :: rho(:)
!  real(long), intent(in) :: rho_free(:)
!  integer, intent(in) :: j_s    ! index for Wigner-Seitz radius (r_s)
!  integer, intent(in) :: Z      ! # of electrons within r_s = nuclear charge Z
!  real(long) :: factor
!  !
!  ! rho_free(:) is normalized so that INT_0^{r_s} rho_free(r) dr = 1
!  ! In other words, factor is corresponding to the number of free electrons.
!  !
!  factor = dble(Z) - integral( G, rho(1:j_s) )
!  rho(:) = rho(:) + factor * rho_free(:)
!end subroutine apply_free_elec

!----------------------------------------------------------------------------
! Load an external environmental potential file from XPOT
!----------------------------------------------------------------------------
subroutine load_env_pot( P, G, v_env, VL )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long), intent(out) :: v_env(:), VL
  integer, parameter :: IO_POT = 52 
  real(long), allocatable :: v_env0(:), rin(:)
  real(long) :: xq, rc
  integer :: ir, Nr
  logical :: yes_exist
  type(Spline) :: SPOT
  inquire( file = trim(P%file_env), exist = yes_exist )
  if ( .not. yes_exist ) &
    call error_message( "The environmental potential file is not found" )
  !
  ! file format for an environmental potential file
  ! - Nr: number of grid points (uniform), which is given by XPOT
  ! - xq: it's supposed to be the charge 'q' at the end; q = Z - N_e + 1
  !       but it's not really used in the calculation
  ! - rc: touching radius; it's not really used in the calculaiton
  ! - VL: global flat potential value, determined by the touching scheme
  !       if P%yes_fixed_pot is on, the effective potential is given by this
  !       value after a point where v_eff > VL [ SEE apply_fixed_pot() ]
  !
  open( IO_POT, file=trim(P%file_env), action="read", form="formatted" )
  read( IO_POT, * ) Nr, xq, rc, VL
  allocate( rin(Nr), v_env0(Nr) )
  read( IO_POT, * ) ( rin(ir), v_env0(ir), ir = 1, Nr )
  close(IO_POT)
  !
  ! interpolate/extrapolate external environmental potential into radial grid
  ! - v_env0: uniform grid loaded from external env. potential file
  ! - v_env:  nonuniform radial grid to be used in the calculation
  ! NOTE: it should be extrapolated with a constant value if r > rmax_env_pot
  !       rmax_env_pot: the max. radius of the external env. potential file.
  !       This rmax_env_pot must be less than the half of the supercell size
  !       because env_pot is constructed using minimal-image convention.
  !
  SPOT = Spline_create( rin(:), v_env0(:) )
  v_env(:) = Spline_interpolate_vector( SPOT, G%r(:) )
  deallocate(rin, v_env0)
end subroutine load_env_pot

!----------------------------------------------------------------------------
! Thermal reduction factor
!   The form is given in arXiv:1309.3043
!   Originally, Perrot and Dharma-Wardana, PRA 30, 2619 (1984)
!   The original forms in both references are for LDA rather than HFS,
!   but in arXiv it is wittern as 0 < R < 1, which can be applied to HFS.
!----------------------------------------------------------------------------
function thermal_reduction_factor( rho, T_eV ) result (R)
  real(long), intent(in) :: rho(:), T_eV
  real(long) :: R(size(rho))
  real(long) :: theta(size(rho))
  theta(:) = TWO * (T_eV*EV2AU) / ( THREE * PI**2 * rho(:) )**(TWO/THREE)
  R(:) = FOUR / THREE &
         * (   0.75_long + 3.04363_long * theta(:)**2 &
             - 0.092270_long * theta(:)**3 + 1.70350_long * theta(:)**4 ) &
         / ( ONE + 8.31051_long * theta(:)**2 + 5.1105_long * theta(:)**4 ) &
         * tanh( ONE / theta(:) )
  where ( rho(:) == ZERO ) R(:) = ZERO
end function thermal_reduction_factor

!----------------------------------------------------------------------------
!----------------------------------------------------------------------------
subroutine potential_x_finite_temp( v_x, rho, T_eV, thermal_x_type )
  real(long), intent(inout) :: v_x(:)
  real(long), intent(in) :: rho(:), T_eV
  integer, intent(in) :: thermal_x_type
  real(long) :: theta(size(rho))  ! electron degeneracy, lambda in Rozsnyai72
  real(long) :: R(size(rho))      ! thermal reduction factor in Perrot84
  real(long) :: v_M(size(rho))    ! Maxwellian in Rozsnyai72
  theta(:) = TWO * (T_eV*EV2AU) / ( THREE * PI**2 * rho(:) )**(TWO/THREE)
  if ( thermal_x_type == 1 ) then ! Perrot & Dharma-wardana, PRA 30, 2619 (1984)
    R(:) = FOUR / THREE &
           * (   0.75_long + 3.04363_long * theta(:)**2 &
               - 0.092270_long * theta(:)**3 + 1.70350_long * theta(:)**4 ) &
           / ( ONE + 8.31051_long * theta(:)**2 + 5.1105_long * theta(:)**4 ) &
           * tanh( ONE / theta(:) )
    v_x(:) = v_x(:) * R(:)
  else if ( thermal_x_type == 2 ) then  ! Rozsnyai, PRA 5, 1137 (1972)
    v_M(:) = - PI * rho(:) / (T_eV*EV2AU)
    where ( theta(:) < ONE )
      v_x(:) = ( ONE - theta(:)**2 ) * v_x(:) + theta(:)**2 * v_M(:)
    elsewhere
      v_x(:) = v_M(:)
    end where
  end if
  where ( rho(:) == ZERO ) v_x(:) = ZERO
end subroutine potential_x_finite_temp

!----------------------------------------------------------------------------
! calculate E(0) from FT-MBPT:
!   E(0) = mu^(0) * <N>^(0) - d log[Z_G^(0)] / d beta
! - partial derivative of log(Z_G) is evaluated by a simple finite difference
!----------------------------------------------------------------------------
subroutine print_MBPT_E0( P, F, F1, F2, T1_eV, T2_eV, OUTPUT )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F, F1, F2  ! F1, F2: calc. at adjacent T=P%T_eV
  real(long), intent(in) :: T1_eV, T2_eV  ! in eV
  integer, intent(in) :: OUTPUT
  real(long) :: beta, beta1, beta2
  real(long) :: log_ZG, log_ZG1, log_ZG2
  real(long) :: mu0     ! chemical potential calculated at T
  real(long) :: N0      ! <N>^(0) = - d log(Z_G) / d mu / beta = SUM_p n_p
  real(long) :: E0      ! zeroth-order E, E^(0)
  real(long) :: E_sum   ! orbital energy sum, which is NOT equal to E0
  real(long) :: deriv
  mu0 = F%chem_pot                      ! calculated at T
  N0 = sum( F%n_occ(:) * F%norm(:) )    ! take into account norm (EXPERIMENTAL)
  E_sum = sum( F%e_all(:) * F%n_occ(:) * F%norm(:) )
  beta  = ONE / ( P%T_eV * EV2AU )
  beta1 = ONE / ( T1_eV * EV2AU )
  beta2 = ONE / ( T2_eV * EV2AU )
  log_ZG  = log_partition_function( F, P%T_eV, mu0 )    ! using mu0
  log_ZG1 = log_partition_function( F1, T1_eV, mu0 )    ! using mu0
  log_ZG2 = log_partition_function( F2, T2_eV, mu0 )    ! using mu0
  deriv = ( ( log_ZG2 - log_ZG ) / ( beta2 - beta ) + &
            ( log_ZG - log_ZG1 ) / ( beta - beta1 ) ) * HALF
  E0 = mu0 * N0 - deriv
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# FT-MBPT analysis:')")
  write(OUTPUT, "('#   T (eV)     = ',es13.5)") P%T_eV
  write(OUTPUT, "('#   mu0 (a.u.) = ',es13.5)") mu0
  write(OUTPUT, "('#   N0         = ',es13.5)") N0
  write(OUTPUT, "('#   orbital energy sum, E_sum (a.u.) = ',es13.5)") E_sum
  write(OUTPUT, "('#   SCF calculations at three T points:',3es13.5)") &
    T1_eV, P%T_eV, T2_eV
  write(OUTPUT, "('#   log(Z_G0) is calculated using the same mu0:')")
  write(OUTPUT, "('#             ',2a13)") "logZG0", "T(eV)"
  write(OUTPUT, "('#   - left  : ',2es13.5)") log_ZG1, T1_eV
  write(OUTPUT, "('#   - center: ',2es13.5)") log_ZG,  P%T_eV
  write(OUTPUT, "('#   - right : ',2es13.5)") log_ZG2, T2_eV
  write(OUTPUT, "('#   ',a)") &
    "and then the partial derivative evaluated using simple finite difference"
  write(OUTPUT, "('#   zeroth-order E0 = mu0*N0 - dlog(Z_G0)/dbeta = '," // &
    "es13.5,' (a.u.)')") E0
  write(OUTPUT, "('#')")
  write(OUTPUT, "('#          ',3a12)") "T(eV)", "E0(eV)", "E_sum(eV)"
  write(OUTPUT, "('# |MBPT|   ',3f12.3)") P%T_eV, E0 * AU2EV, E_sum * AU2EV
end subroutine print_MBPT_E0

end module plasma
