!----------------------------------------------------------------------------
! print_property.f90
!----------------------------------------------------------------------------
! Copyright (C) 2012 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2012 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Print property module for all xatom dynamics calculations (req, mc, mc_fly)
!   written by Son, Sang-Kil in 2012
!----------------------------------------------------------------------------
module print_property
  use constants
  use atom
contains

subroutine print_charge_states( OUTPUT, p_final, p_avg, n_avg )
  integer, intent(in) :: OUTPUT
  real(long), intent(in) :: p_final(0:) ! charge-state pop. after the pulse
  real(long), intent(in) :: p_avg(0:)   ! charge-state pop. averaged over time
  real(long), intent(in), optional :: n_avg(1:) ! # of absorbed photons
  real(long) :: q_final, q_avg, stdev_final, stdev_avg, sum_ion_yield, sum_n
  integer :: i, Z
  !
  ! print out charge-state distribution
  !
  Z = size(P_final) - 1                 ! max. charge == nuclear charge
  q_final = ZERO
  q_avg   = ZERO
  stdev_final = ZERO
  stdev_avg   = ZERO
  sum_ion_yield = sum( p_final(1:) )    ! sum of ions only, excluding neutral
  write(OUTPUT, "('#   ',a3,2x,3a12)", advance="no") &
    "q", "Pop", "Pop_avg", "Ion_yield"
  if ( present(n_avg) ) write(OUTPUT, "(a12)", advance="no") "N_photon"
  write(OUTPUT, "()")
  do i = 0, Z
    q_final = q_final + dble(i) * p_final(i)
    q_avg   = q_avg   + dble(i) * p_avg(i)
    stdev_final = stdev_final + dble(i**2) * p_final(i)
    stdev_avg   = stdev_avg   + dble(i**2) * p_avg(i)
    write(OUTPUT, "(4x,sp,i3,2x,s,2es12.5)", advance="no") &
      i, p_final(i), p_avg(i)
    if ( i /= 0 ) then
      if ( sum_ion_yield == ZERO ) then
        write(OUTPUT, "(a12)", advance="no") "NaN"
      else
        write(OUTPUT, "(es12.5)", advance="no") p_final(i) / sum_ion_yield
      end if
      if ( present(n_avg) ) write(OUTPUT, "(es12.5)", advance="no") n_avg(i)
    end if
    write(OUTPUT, "()")
  end do
  stdev_final = sqrt( stdev_final - q_final**2 )
  stdev_avg   = sqrt( stdev_avg   - q_avg**2 )
  write(OUTPUT, "('#|CHARGE|',2es12.5)") q_final, q_avg
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Mean charge after the pulse                = '," // &
                "sp,f6.2,5x,'STDEV:',s,f6.2)") q_final, stdev_final
  write(OUTPUT, "('# Time-averaged charge weighted by the pulse = '," // & 
                "sp,f6.2,5x,'STDEV:',s,f6.2)") q_avg, stdev_avg
  if ( present(n_avg) ) then
    sum_n = sum( n_avg(1:) * p_final(1:) )  ! sum of # abs. photons per charge
    write(OUTPUT, "('# Averaged number of absorbed photons        = ',f8.4)") &
                  sum_n
  end if
end subroutine print_charge_states

subroutine print_divider( OUTPUT, title, symbol )
  integer, intent(in) :: OUTPUT
  character(len=*), intent(in), optional :: title
  character(len=1), intent(in), optional :: symbol
  integer :: len_title, len_left, len_right
  character(len=1) :: s
  s = "="
  if ( present( symbol ) ) s = symbol
  if ( present( title ) .and. title /= "" ) then
    len_title = len_trim( title ) + 2
    len_left = ( 80 - 2 - len_title ) / 2
    len_right = 80 - 2 - len_left - len_title
    write(OUTPUT, "('# ',a,x,a,x,a)") &
      repeat( s, len_left ), trim( title ), repeat( s, len_right )
  else
    write(OUTPUT, "('# ',a)") repeat( s, 80 - 2 )
  end if
end subroutine print_divider

subroutine print_all_ions( OUTPUT, Z )
  integer, intent(in) :: OUTPUT, Z
  integer :: k
  do k = 0, Z
    if ( k < 10 ) then
      write(OUTPUT, "(a8,'+',i1)", advance="no") trim(ATOMIC_NAME(Z)), k
    else if ( k < 100 ) then
      write(OUTPUT, "(a7,'+',i2)", advance="no") trim(ATOMIC_NAME(Z)), k
    else
      write(OUTPUT, "(a6,'+',i3)", advance="no") trim(ATOMIC_NAME(Z)), k
    end if
  end do
end subroutine print_all_ions

end module print_property
