!----------------------------------------------------------------------------
! process.f90
!----------------------------------------------------------------------------
! Copyright (C) 2012-2020 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2012-2020 Sang-Kil Son <sangkil.son@cfel.de>
!          2015-2016 Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Process module, especially made for Monte-Carlo on-the-fly,
!                 but used for everywhere
!   written by Son, Sang-Kil in Mar. 2012
!
! This module contains all config. and process information for each config.
! The electronic configuration is defined as a series of subshells.
!   example) 1s1 2s2 2p6
!           +---+---+---+
!           | 1 | 2 | 6 |  each box represents a subshell.
!           +---+---+---+  
! Each subshell is assigned with quantum numbers (n,l) 
! and contains an occupation number.
! The maximum number of subshells is given by the number of electrons
! because all electrons are different in the extreme case.
!
! For each configuration, possible processes are considered.
! Each transition is specified with changes of occupation numbers in subshells
!   examples)
!   - photoionization in 2s: occ# is down in the second subshell.
!   - Auger decay 1s-2s2p: occ# up in 1 subshell, occ# down in 2 & 3 subshells
!   - fluorescence 1s-2p: occ# up in 1 subshell, occ# down in 3 subshell
!   - excitation from 2s to 3p: occ# down in 2 subshell and occ# up in a new 
!     subshell (4th), along with specifying (n=3,l=1) for the new subshell
!
! To make indexes for electronic configurations, we use the "list" or "tree"
! structure.  Each subshell has branches of 0...max_occ.  Each branch is
! pointing to a next subshell.  Once it passes through all (initially-occupied)
! subshells, one can obtain all process information corresponding to the config.
!   example)
!     1s +- 0
!        |- 1 --> 2s +- 0
!        +- 2        |- 1
!                    +- 2 --> 2p +- 0
!                                |- 1
!                                |- 2
!                                 ...
!                                +- 6 =====> 1s1 2s2 2p6
!                                            all process data are assigned here.
!----------------------------------------------------------------------------
module process
  use constants
  use atom
  use elec_config
  implicit none
  !
  ! Process data contains cross sections and rates of all processes.
  !   config(1:N_elec,1:3) : elec. config. info
  !     note that 1...N_elec because it's possible to have all diff. electrons
  !     1:3 indicates 1->n, 2->l, and 3->occupation#
  !   E_tot: total energy
  !   N_process : # of processes possibly generated from this configuration
  !   tag(k) : process name tag, P: photoionization, A: Auger, F: fluorescence
  !            E: resonant excitation
  !   value(k) : cross section or rate (in a.u.) of the k-th process
  !   E_eV(k) : transition energy (in eV) in integer, used for spectra
  !   occ_down(3,k): change in config., occ# will be decreased, three rooms
  !   occ_up(1,k):   change in config., occ# will be increased, only one room
  !   excite_to_n(1,k): if k-th process is excitation, then specify n
  !   excite_to_l(1,k): if k-th process is excitation, then specify l
  !   shakeoff(i,k) : single shake-off prob. from i-th orb for k-th process
  !   shakeoff_E(i,k) : transition energy after shake-off, in integer
  !   NOTE: fortran uses column-major order, 
  !         so spanning of (:,k) is faster than (k,:)
  !
  ! - for photoionization / photoabsorption cross section
  !   ionization_threshold(i) : orbital energy of the i-th subshell
  !   cross_section(i,1:5) : fitting parameters for photoion. cross section
  !   trans_matrix(:,:) : bound-bound transition matrix elements
  !
  ! - for memory efficiency, we use the smallest integer size
  !   integer*1 == integer(kind=1): -128 ~ +127
  !   integer*2 == integer(kind=2): -32768 ~ +32767
  !   integer*4 == integer(kind=4): -2^31 ~ +2^31-1 = 2,147,483,647 (default)
  !   integer*8 == integer(kind=8): -2^63 ~ +2^63-1
  !   for your information, the neutral Pt index goes 28,591,624,215
  !   that's why integer*8 for this index
  !
  type :: Process_data
    integer, allocatable :: config(:,:)     ! elec. config. info
!    integer*8 :: ind                       ! integer for indexing this config.
    integer :: N_process                    ! # of processes
    real(long) :: E_tot                     ! total energy
    real(long) :: E0                        ! zeroth energy (=sum of orb. E)
    character(len=1), allocatable :: tag(:) ! process name tag, 1...N_proc
    integer, allocatable :: p_type(:)       ! process type, 1...N_proc
    !
    ! process type definition
    !   tag     p_type
    !    P or V  1    : photoionization
    !    A       2    : Auger-Meitner decay (see doi:10.1063/PT.3.4281)
    !    F       3    : fluorescence
    !    E       4    : photoexcitation
    !    2       5    : direct two-photon ionization
    !    3       6    : direct three-photon ionization
    !    4       7    : direct four-photon ionization
    !    5       8    : direct five-photon ionization
    !
    real(long), allocatable :: value(:)     ! cross sec or rate in a.u., 1...N_p
    integer, allocatable :: E_eV(:)         ! transition E in eV, 1...N_proc
    integer, allocatable :: occ_up(:,:), occ_down(:,:)
!    integer, allocatable :: excite_to_n(:,:), excite_to_l(:,:)
    integer :: counter_visit            ! counter for visiting this config.
    !
    ! think about shakeoff; there are three different implementations
    ! a) store the calculated potential for the given config
    !    -> simple, but eigenvectors need to be solved at every single time
    ! b) store the calculated branching ratios for all possible transitions
    !    -> straightforward, but SCF has to be calculated for all transitions
    !       even though some transitions won't ever be visited
    ! c) store the calculated orbitals
    !    -> it will take much more memory than a), 
    !       but repeated eigenvalue problem will be avoided.
    !    => Let's choose this one!  F%u(:,:) to be copied to S%u(:,:)
    !
    real(long), allocatable :: u(:,:)   ! will be allocated when it's needed
  end type Process_data
  !
  ! Config_tree: tree or list structure of electronic configuration
  !
  type :: Config_tree
    type(Process_data), allocatable :: S  ! to be allocated only at the tree end
    type(Config_tree), pointer :: p(:) => NULL()
  end type Config_tree
contains

function create_Process( config, N_proc_max ) result (S)
  type(Process_data) :: S
  integer, intent(in) :: config(:,:), N_proc_max
  integer :: N_orb
  !
  ! check whether it's relativistic or not
  !   size(config,dim=2) == 3 => non-relativistic
  !   size(config,dim=2) /= 4 => relativistic
  ! WARNING: for P%init_config_rel and F%config_rel, size() = 5
  ! REMEMBER: 1st arg: n, 2nd arg:l, 3rd arg:N_occ for non-relativistic
  !           1st arg: n, 2nd arg:l, 3rd arg:j, 4th arg: N_occ for relativistic
  !
  ! S%config is defined as S%config(1:N_orb,1:3) for non-relativistic
  !                     or S%config(1:N_orb,1:4) for relativistic
  !
  N_orb = size(config,dim=1)
  if ( size(config,dim=2) == 3 ) then   ! for the non-relativistic case
    allocate( S%config(N_orb,3) )       ! size(S%config,dim=2) = 3
    S%config(:,1:3) = config(:,1:3)     ! copy for 1:3
  else                                  ! for the relativistic case
    allocate( S%config(N_orb,4) )       ! size(S%config,dim=2) = 4
    S%config(:,1:4) = config(:,1:4)     ! copy only for 1:4, not for 1:5
  end if
  S%N_process = 0       ! it will be increased whenever proc. is added
  if ( N_proc_max > 0 ) then
    allocate( S%tag(N_proc_max), S%value(N_proc_max), S%E_eV(N_proc_max) )
    allocate( S%p_type(N_proc_max) )
    allocate( S%occ_up(1,N_proc_max), S%occ_down(3,N_proc_max) )
    S%occ_up(:,:) = 0
    S%occ_down(:,:) = 0    
  end if
end function create_Process

!subroutine copy_Process( S1, S2 )
!  type(Process_data), intent(in) :: S1
!  type(Process_data), intent(out) :: S2
!  integer :: N
!  N = S1%N_process      ! S2 will be prepared to fit to #proc of S1
!  S2 = create_Process( S1%config, N )
!  S2%N_process = N
!  S2%N_visit   = S1%N_visit
!  S2%config(:,:) = S1%config(:,:)
!  S2%tag(:)   = S1%tag(1:N)
!  S2%value(:) = S1%value(1:N)
!  S2%E_eV(:)  = S1%E_eV(1:N)
!  S2%occ_up(:,:)   = S1%occ_up(:,1:N)
!  S2%occ_down(:,:) = S1%occ_down(:,1:N)
!  S2%excite_to_n(:,:) = S1%excite_to_n(:,1:N)
!  S2%excite_to_l(:,:) = S1%excite_to_l(:,1:N)
!end subroutine copy_Process

!----------------------------------------------------------------------------
! add one process to S
! - S%value(k): rate or cross section
! - S%E_eV(k): transition energy in eV, rounded to nearest integer
! - S%N_process is increased by one.
! - S%tag is given by P, V, A, F, 2 (two-photon abs.), 3 (three-photon), ...
!----------------------------------------------------------------------------
subroutine add_Process( S, str, value, E, i, j, jp )
  type(Process_data), intent(inout) :: S
  character(len=*), intent(in) :: str
  real(long), intent(in) :: value, E
  integer, intent(in) :: i      ! i, j, jp: orbital index for config(:,:)
  integer, intent(in), optional :: j, jp
  integer :: k, n_max
  !
  ! increase # of processes
  ! and put this process into Process_data
  !
  S%N_process = S%N_process + 1
  if ( S%N_process > size(S%tag) ) &
    stop "ERROR in add_Process(): N_process exceeds N_proc_max."
  k = S%N_process
  S%value(k) = value
  S%E_eV(k) = nint( E * AU2EV )    ! nint(): round to nearest integer
  !
  ! set up the label for the process
  ! for relation between tag and p_type, see the beginning of this module
  !
  n_max = maxval( S%config(:,1) )       ! max. n of the shells
  select case ( str )
    case ( "P", "photoionization", "photoabsorption" )  ! i ionized
      S%tag(k) = "P"
      S%p_type(k) = 1
      if ( S%config(i,1) == n_max ) S%tag(k) = "V"      ! valence photoion.
      S%occ_down(1,k) = i
    case ( "A", "auger", "Auger" )                      ! j -> i, jp ionized
      if ( .not. ( present(j) .and. present(jp) ) ) &
        stop "ERROR when add_Process() is called with A."
      S%tag(k) = "A"
      S%p_type(k) = 2
      S%occ_up(1,k) = i
      S%occ_down(1,k) = j
      S%occ_down(2,k) = jp
    case ( "F", "fluorescence", "fluo" )        ! j -> i
      if ( .not. present(j) ) stop "ERROR when add_Process() is called with F."
      S%tag(k) = "F"
      S%p_type(k) = 3
      S%occ_up(1,k) = i
      S%occ_down(1,k) = j
    case ( "E", "photoexcitation", "excite" )
      if ( .not. ( present(j) ) ) &
        stop "ERROR when add_Process() is called with E."
      S%tag(k) = "E"
      S%p_type(k) = 4
      S%occ_down(1,k) = i
      S%occ_up(1,k) = j
    case ( "2", "two-photon" )          ! i ionized
      S%tag(k) = "2"
      S%p_type(k) = 5
      S%occ_down(1,k) = i
    case ( "3", "three-photon" )        ! i ionized
      S%tag(k) = "3"
      S%p_type(k) = 6
      S%occ_down(1,k) = i
    case ( "4", "four-photon" )         ! i ionized
      S%tag(k) = "4"
      S%p_type(k) = 7
      S%occ_down(1,k) = i
    case ( "5", "five-photon" )         ! i ionized
      S%tag(k) = "5"
      S%p_type(k) = 8
      S%occ_down(1,k) = i
  end select
end subroutine add_Process

!----------------------------------------------------------------------------
! print out all processes belonging to the current configuration
! for backward compatibility, print out the index using config_to_index()
!----------------------------------------------------------------------------
subroutine print_Process( S, OUTPUT )
  type(Process_data), intent(in) :: S
  integer, intent(in) :: OUTPUT
  integer :: i, k
  integer*8 :: i_initial, i_final
  integer :: config_final(size(S%config,dim=1),size(S%config,dim=2))
  integer :: N_max, occ_index
  character(len=3000) :: s1, s2, s3, s4
  character(len=6000) :: description
  N_max = size(S%config,dim=1)
  occ_index = size(S%config,dim=2)      ! S%config is either 3 or 4
  do i = N_max, 1, -1
    if ( S%config(i,1) /= 0 ) exit
  end do
  write( s1, "(1000i3)" ) S%config(1:i,occ_index)      ! print out initial occ#
  call config_to_index( S%config, i_initial )
  do k = 1, S%N_process
    config_final(:,:) = config_Process( S, k )
    call config_to_index( config_final, i_final )
    do i = N_max, 1, -1
      if ( config_final(i,1) /= 0 ) exit
    end do
    write( s2, "(1000i3)" ) config_final(1:i,occ_index)  ! print out final occ#
    write( s3, "(2i25)" ) i_initial, i_final
    write( s4, * ) S%value(k)
    write( description, "(' # ',a,' -> ',a)" ) trim(s1), trim(s2)
    !
    ! print out shake-off branching ratios if defined
    !
!    if ( S%shakeoff(1,k) == -ONE ) then
      write( OUTPUT, "(x,a,a50,a30,x,i6,a)" ) &
        S%tag(k), trim(s3), s4, S%E_eV(k), trim(description)
!    else
!      write( OUTPUT, "('#',a,a20,a24,x,i6,a)" ) &       ! print original value
!        S%tag(k), trim(s3), s4, S%E_eV(k), trim(description)
!      write( s4, * ) S%value(k) * ( ONE - sum( S%shakeoff(:,k) ) )
!      write( OUTPUT, "(x,a,a20,a24,x,i6,a)" ) &         ! after shakeoff
!        S%tag(k), trim(s3), s4, S%E_eV(k), trim(description)
!      do i = 1, N_orb
!        if ( S%shakeoff(i,k) == ZERO ) cycle
!        config_final(i,3) = config_final(i,3) - 1
!        call config_to_index( config_final, i_final )
!        write( s2, "(100i3)" ) config_final(:,3)
!        write( s3, "(2i10)" ) i_initial, i_final
!        write( s4, * ) S%value(k) * S%shakeoff(i,k)
!        write( description, "(' # ',a,' -> ',a,' ; S1',f5.1,'%')" ) &
!          trim(s1), trim(s2), S%shakeoff(i,k) * 100.0_long
!        write( OUTPUT, "(x,a,a20,a24,x,i6,a)" ) &
!          S%tag(k), trim(s3), trim(s4), S%shakeoff_E(i,k), trim(description)
!        config_final(i,3) = config_final(i,3) + 1
!      end do
!    end if
  end do
end subroutine print_Process

function config_Process( S, k ) result (config_final)
  type(Process_data), intent(in) :: S
  integer, intent(in) :: k
  integer :: config_final(size(S%config,dim=1),size(S%config,dim=2))
  integer :: i, j, occ_index
  ! find the position of the argument of config variable pointing to occupation number
  occ_index = size(S%config,dim=2)      ! S%config is either 3 or 4
  config_final(:,:) = S%config(:,:)
  !
  ! checking a subshell for decreasing occupation #
  !
  do j = 1, size(S%occ_down,dim=1)
    i = S%occ_down(j,k)
    if ( i /= 0 ) config_final(i,occ_index) = config_final(i,occ_index) - 1
  end do
  !
  ! checking a subshell for increasing occupation #
  !
  do j = 1, size(S%occ_up,dim=1)
    i = S%occ_up(j,k)
    if ( i /= 0 ) then
      config_final(i,occ_index) = config_final(i,occ_index) + 1
!      if ( S%excite_to_n(j,k) /= 0 ) then       ! this is excitation
!        config_final(i,1) = S%excite_to_n(j,k)  ! assign a new quantum # n
!        config_final(i,2) = S%excite_to_l(j,k)  ! assign a new quantum # l
!      end if
    end if
  end do
end function config_Process

function string_Process( S, k ) result (string)
  type(Process_data), intent(in) :: S
  integer, intent(in) :: k
  character(len=3000) :: string
  character(len=3000) :: s1, s2
  integer :: config_final(size(S%config,dim=1),size(S%config,dim=2))
  integer :: i, N_max, occ_index
  config_final(:,:) = config_Process( S, k )
  N_max = size(S%config,dim=1)
  occ_index = size(S%config,dim=2)      ! S%config is either 3 or 4
  do i = N_max, 1, -1
    if ( config_final(i,1) /= 0 ) exit
  end do
  write( s1, "(1000i3)" ) S%config(1:i,occ_index)
  write( s2, "(1000i3)" ) config_final(1:i,occ_index)
  write( string, "(a,' -> ',a)" ) trim(s1), trim(s2)
end function string_Process

function label_Process( S, k ) result (string)
  type(Process_data), intent(in) :: S
  integer, intent(in) :: k
  character(len=3000) :: string
  integer :: i, j, jp
  integer :: p_type
  character(len=1) :: tag
  character(len=5) :: ti, tj, tjp
  p_type = S%p_type(k)
  tag = S%tag(k)
  select case ( p_type )
    case ( 1 )          ! photoionization
      i  = S%occ_down(1,k)
      ti = orbital_name_from_config( S%config, i )
      write(string,"(a,' : ',a)") tag, trim(ti)
    case ( 2 )          ! Auger-Meitner decay
      i  = S%occ_up(1,k)
      j  = S%occ_down(1,k)
      jp = S%occ_down(2,k)
      ti = orbital_name_from_config( S%config, i )
      tj = orbital_name_from_config( S%config, j )
      tjp = orbital_name_from_config( S%config, jp )
      write(string,"(a,' : ',a,' -- ',a,' ',a)") &
        tag, trim(ti), trim(tj),trim(tjp)
    case ( 3 )          ! fluorescence
      i  = S%occ_up(1,k)
      j  = S%occ_down(1,k)
      ti = orbital_name_from_config( S%config, i )
      tj = orbital_name_from_config( S%config, j )
      write(string,"(a,' : ',a,' <- ',a)") tag, trim(ti), trim(tj)
    case ( 4 )          ! photoexcitation
      i  = S%occ_down(1,k)
      j  = S%occ_up(1,k)
      ti = orbital_name_from_config( S%config, i )
      tj = orbital_name_from_config( S%config, j )
      write(string,"(a,' : ',a,' -> ',a)") tag, trim(ti), trim(tj)
    case default        ! assuming direct multi-photon ionization
      i  = S%occ_down(1,k)
      ti = orbital_name_from_config( S%config, i )
      write(string,"(a,' : ',a)") tag, trim(ti)
  end select
end function label_Process

!----------------------------------------------------------------------------
! load / save Process data (or called atomic data)
! - load_Process and save_Process must have the same structure.
! - all information of Process_data must be stored.
! - filename is supposed to be given by get_filename().
!----------------------------------------------------------------------------
!subroutine load_Process( S, filename, yes_loaded )
!  type(Process_data), intent(inout) :: S
!  character(len=*), intent(in) :: filename
!  logical, intent(out) :: yes_loaded
!  integer :: IO_DATA = 20, INFO
!  open( IO_DATA, file=filename, &
!    action="read", status="old", form="unformatted" )
!  read( IO_DATA, iostat=INFO ) S%Z, S%N_subshell
!  if ( INFO < 0 ) then  ! if end-of-file, then stop here
!    close( IO_DATA )
!    yes_loaded = .false.
!    return
!  end if
!  read( IO_DATA ) S%config(:,:), S%ext_config(:,:)
!  read( IO_DATA ) S%N_process
!  read( IO_DATA ) S%tag(1:S%N_process)
!  read( IO_DATA ) S%value(1:S%N_process)
!  read( IO_DATA ) S%E_eV(1:S%N_process)
!  read( IO_DATA ) S%shakeoff(:,1:S%N_process)
!  read( IO_DATA ) S%shakeoff_E(:,1:S%N_process)
!  close( IO_DATA )
!  S%charge = S%Z - sum( S%ext_config(:,0) )
!  yes_loaded = .true.
!end subroutine load_Process
!
!subroutine save_Process( S, filename )
!  type(Process_data), intent(inout) :: S
!  character(len=*), intent(in) :: filename
!  integer :: IO_DATA = 20
!  open( IO_DATA, file=filename, &
!    action="write", status="replace", form="unformatted" )
!  write( IO_DATA ) S%Z, S%N_subshell
!  write( IO_DATA ) S%config(:,:), S%ext_config(:,:)
!  write( IO_DATA ) S%N_process
!  write( IO_DATA ) S%tag(1:S%N_process)
!  write( IO_DATA ) S%value(1:S%N_process)
!  write( IO_DATA ) S%E_eV(1:S%N_process)
!  write( IO_DATA ) S%shakeoff(:,1:S%N_process)
!  write( IO_DATA ) S%shakeoff_E(:,1:S%N_process)
!  close( IO_DATA )
!  ! S%charge are obtained by S%Z ad S%ext_config
!end subroutine save_Process
!
!function get_filename( S, data_dir ) result (str)
!  character(len=200) :: str
!  type(Process_data), intent(in) :: S
!  character(len=*) :: data_dir
!  character(len=3) :: str_charge
!  character(len=7) :: str_name
!  integer :: i
!  !
!  ! define a filename for each configuration
!  !
!  write(str_charge, "(i3)") S%charge
!  write(str_name, "(a,'_',a,'+')") trim(ATOMIC_NAME(S%Z)), &
!                                   trim(adjustl(str_charge))
!  write(str,"(a,'/',a,'/',a,'_',100a1)") &
!    trim(data_dir), trim(str_name), trim(str_name), &
!    (/ ( i_to_a(S%ext_config(i,0)), i=1, S%N_subshell ) /)
!contains
!  function i_to_a( n ) result (a)
!    integer*1, intent(in) :: n
!    character(len=1) :: a
!    !
!    ! get a character from a number: 10 -> A, 11 -> B, and so on
!    !
!    if ( n < 0 ) then
!      stop "ERROR in process.f90: occupancy must be > 0."
!    else if ( n < 10 ) then
!      write( a, "(i1)" ) n
!    else
!      a = achar( iachar('A') + n - 10 )
!    end if
!  end function i_to_a
!end function get_filename

!function max_occ( config, i ) result (n)
!  integer :: n
!  integer, intent(in) :: config(:,:), i
!  integer :: l
!  l = config(i,2)
!  n = 2 * ( 2*l + 1 )
!end function max_occ

function get_end_node( Q0, config, N_occ ) result (Q)
  type(Config_tree), pointer :: Q
  type(Config_tree), intent(in), target :: Q0
  integer, intent(in) :: config(:,:), N_occ
  integer :: i, jj, l, occ, max_occ
  Q => Q0       ! it always starts from the beginning, Q0
  if ( size(config,dim=2) == 3 ) then   ! for the non-relativistic case
    do i = 1, N_occ
      l   = config(i,2)
      occ = config(i,3)
      max_occ = 2 * ( 2*l + 1 )
      if ( .not. associated( Q%p ) ) allocate( Q%p(0:max_occ) )
      Q => Q%p(occ)
    end do
  else                                  ! for the relativistic case
    do i = 1, N_occ
      jj = config(i,3)                  ! 2*j instead of j
      occ = config(i,4)
      max_occ = jj + 1
      if ( .not. associated( Q%p ) ) allocate( Q%p(0:max_occ) )
      Q => Q%p(occ)
    end do
  end if
!  do i = 1, N_occ
!     if ( yes_relativity == .true. ) then
!        jj = config(i,3) ! 2*j instead of j
!        occ = config(i,4)
!        max_occ = jj + 1
!     else
!        l   = config(i,2)
!        occ = config(i,3)
!        max_occ = 2 * ( 2*l + 1 )
!     endif
!    if ( .not. associated( Q%p ) ) then
!      allocate( Q%p(0:max_occ) )
!      !do j = 0, max_occ        ! we don't need to nullify these S's
!      !  Q%p(j)%S => NULL()     ! because S is now "allocatable".
!      !end do
!    end if
!    Q => Q%p(occ)
!  end do
end function get_end_node

end module process
