!----------------------------------------------------------------------------
! read_param.f90
!----------------------------------------------------------------------------
! Copyright (C) 2005      Sang-Kil Son
! Copyright (C) 2010-2023 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2023 Sang-Kil Son <sangkil.son@cfel.de>
!          2014-2015 Ludger Inhester <ludger.inhester@cfel.de>
!          2014      Jan Malte Slowik
!          2014-2016 Otfried Geffert
!          2015-2017 Koudai Toyota
!          2017-2018 John Bekx
!          2018-2019 Rui Jin
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! READ_PARAM module for atomic and diatomic molecular system 
! with both time-independent and time-dependent calculations
!   written by Son, Sang-Kil in Jan. 2005, Jun. 2005
!
! This module can be used for any calculation of an atom or 
! a diatomic molecule (homonuclear or heteronuclear), such as
! static calculation (TISE and DFT), dynamic calculation (TDSE and TDDFT),
! and Floquet calculation (Floquet and TDFloquet)
!----------------------------------------------------------------------------
module read_param
  use constants
  use stdio
  use atom
  use elec_config
  implicit none
  character(len=10), parameter :: FUNCTIONAL_X_NAME(7) = (/ &
    "LSDA  ", &   ! Local Spin Density Approximation (N_a /= N_b)
    "LDA   ", &   ! Local Density Approximation (N_a == N_b)
    "Xa    ", &   ! X-alpha == LDA without correlation if alpha = 2/3
    "Xab   ", &   ! X-alpha-beta == X-alpha if beta = 0
    "LB    ", &   ! R. van Leeuwen and E. J. Baerends
    "LBa   ", &   ! improved LB containing alpha and beta empirical parameters
    "Slater" /)   ! Slater potential == X-alpha with alpha = 1.00
  character(len=10), parameter :: FUNCTIONAL_C_NAME(5) = (/ &
    "none", &
    "CA  ", &   ! Ceperley and Alder (1980)
    "VWN ", &   ! Vosko, Wilk, and Nusair potential (1980)
    "PW  ", &   ! Perdew and Wang potential (1992)
    "LYP " /)   ! Lee, Yang, and Parr potential
  ! ------------------------------------------------
  ! Param: all parameters to be used in main program
  ! ------------------------------------------------
  type Param
    !
    ! calc_type: calculation type
    !   - 0 : TISE, time-independent Schoedinger equation
    !   - 1 : TDSE, time-dependent Schoedinger equation
    !   - 2 : Floquet, time-independent non-Hermittian Floquet method
    !   - 3 : TDFloquet, time-dependent non-Hermittian Floquet method
    !   - 4 : HFS, simple model of Hartree-Fock-Slater == LDA
    !   - 5 : REQ, rate equation solver
    !   - 6 : MC, Monte-Carlo calculation
    !   - 7 : MC_FLY, Monte-Carlo on-the-fly calculation
    !   - 8 : TWOSTEP, two-step model calculation
    !
    integer :: calc_type
    character(len=15) :: calc_string
    !
    ! for an atom or a homonuclear/heteronuclear diatomic molecule
    !   N_N : # of nuclei
    !   Z, Z1, Z2 : nuclear charges of atom
    !   isotope, isotope1, isotope2 : isotopes
    !   mass, mass1, mass2 : atomic mass
    !   R : internuclear separation distance    cf) foci coord. a = R / 2
    !   N_a, N_b : numbers of alpha and beta electrons of the molecule
    !   N_orb: # of different orbitals
    !   N_frozen: # of electrons in frozen orbitals
    !
    character(len=20) :: system
    integer :: N_N
    integer :: Z, Z1, Z2
    integer :: isotope, isotope1, isotope2
    real(long) :: mass, mass1, mass2
    integer :: N_a, N_b, N_orb, N_elec, N_frozen
    integer, allocatable :: init_config(:,:) ! (:,1)->n, (:,2)->l, (:,3)->occ#
    real(long), allocatable :: init_f_occ(:) ! fractional occupation numbers
    logical :: yes_fractional   ! true if fractional occ. #'s are used
    !
    ! grid parameters for the spherical coord. or the prolate spheroidal coord.
    ! - N1, L1, Rmax1, Rmin1: grid for coord. 1
    ! - N2, L2, Rmax2, Rmin2: grid for coord. 2
    ! - N3, L3, Rmax3, Rmin3: grid for coord. 3
    ! - N: # of grid points for the total Hamiltonian
    ! - Rmax: maximum radius
    ! - E_limit: the limitation of E to be taken when time propagates
    ! - dr: equally-spaced Delta r for the continuum state
    !
    integer :: N
    integer :: N1, N2, N3
    real(long) :: L1, L2, L3
    real(long) :: Rmax1, Rmax2, Rmax3
    real(long) :: Rmin1, Rmin2, Rmin3
    real(long) :: Rmax, E_limit
    real(long) :: dr, rm_thres, continuum_rmax, diffuse_r2max
    integer :: lmax, nmax       ! for printing of continuum states
    real(long) :: continuum_E   ! continuum energy
    !
    ! Auger electron energies or x-ray energies can be given manually
    !
    real(long) :: EA1, EA2, EA3         ! well... just accepts three energies
    real(long) :: EX1, EX2, EX3         ! well... just accepts three energies
    !
    ! Photon energies
    !
    real(long) :: PE_start, PE_end, PE_dE
    real(long) :: bandwidth     ! bandwidth (FWHM) of the x-ray pulse (in eV)
    real(long) :: multiple_bandwidth
    real(long) :: linewidth     ! spectral linewidth of the system (== Gamma/2)
    integer :: N_convolution    ! convolution with N points
    real(long) :: pcs_scan_step ! pcs calculation scanned with this step
    logical :: yes_harmonics    ! whether harmonics of PE is used
    real(long) :: harmonic2     ! second harmonic ratio w.r.t the main freq
    real(long) :: harmonic3     ! third harmonic ratio w.r.t the main freq
    real(long) :: PE2           ! additional photon energy
    real(long) :: PE2_contrib   ! additional photon energy contribution (ratio)
    !
    ! for elastic scattering
    !
    integer :: N_scd    ! # of grids in the angle or the momentum Q space
    real(long) :: Q_start, Q_end
    real(long) :: resolution    ! derirable resolution
    !
    ! for resonant elastic scattering (dispersion correction)
    !
    integer :: N_disp
    real(long) :: L_disp
    real(long) :: PE_range
    !
    ! for shakeoff (shake-off) process
    !
    logical :: yes_shakeoff
    real(long) :: shakeoff_thres        ! NOT USED
    integer :: shakeoff_type ! <=0 : only photoionization, >0 : all, NOT USED
    !
    ! for printing out coulomb integrals
    !
    logical :: yes_coulomb
    integer :: K_max
    !
    ! threshold for printing process data
    !
    real(long) :: print_threshold
    !
    ! for complex scaling and Floquet
    ! - Mmax : the maximum of Fourier index, -Mmax ... 0 ... Mmax
    ! - theta: rotation angle for complex scaling
    ! - E0: expected ground energy, used for Floquet
    ! - dE: threshold of eigenvalues, use machine precision if dE = ZERO
    ! - nev: # of eigenvalues to be calculated
    !
    integer :: Mmax
    real(long) :: theta, E0, dE
    integer :: nev
    !
    ! for DFT exchange-correlation functional
    !   x_type : exchange functional type
    !   c_type : correlation functional type
    !   x_type and c_type may have the index of x_name(:) and c_name(:).
    !   alpha, beta : parameters for the particular functional
    !   yes_OEP_KLI : turn on or off for OEP-KLI procedure
    !   yes_Latter  : turn on or off for Latter tail correction
    !   yes_ext_pot : turn on or off for exact potential for one- or two-elec.
    !
    integer :: x_type, c_type
    real(long) :: alpha, beta
    logical :: yes_OEP_KLI, yes_Latter, yes_exact_pot, yes_thermal
    integer :: thermal_x_type
    logical :: yes_twobody      ! 1st-order-corrected E w/ two-body interaction
    !
    ! for SCF procedure
    !   SCF_max_iter : the maximum iteration of SCF
    !   SCF_energy_conv : the convergence of SCF with energy difference
    !   SCF_imag_conv : the convergence with relative difference of Im(E)
    !   SCF_density_conv : the convergence of SCF with density difference
    !   SCF_mixing : mixing ratio of old and new potentials when SCF iteration
    !
    integer :: SCF_max_iter
    real(long) :: SCF_energy_conv, SCF_density_conv, SCF_mixing, SCF_imag_conv
    logical :: yes_adaptive_mixing
    logical :: yes_diis
    !
    ! for Laser parameters: (OBSOLETE... but not all of them)
    ! all values are stored in a.u. except some in o.c.(optical cycles)
    !   pulse_type : ramp or chirp
    !                0: ramp, 1: sin^2, 2: chirp, 3: Gaussian, 4: trapezoidal
    !                5: cosinoidal, 6:flat-top
    !                7: pump-probe (two-pulses), 
    !                8: split (double-Gaussian), 9: split (double-flattop)
    !                10: read from file (P%file_pulse)
    !   F : amplitude, F = Frms * sqrt(2) and F = sqrt( I / 3.51e16 )
    !   T : period, T = 2 PI / w0
    !   w0 : frequency
    !   phase : E =  E0 sin( w0 t + phase )
    !           E =  E0 sin( w0 t ) if phase = 0
    !           E = -E0 cos( w0 t ) if phase = -PI/2
    !   pulse_length : in o.c., valid for chirp pulse
    !   ramp_until : in o.c., valid for ramp pulse
    !   chirp_param : cr * 4.86e-8, valid only for chirp pulse
    !   dt : in o.c.
    !   t_start : in o.c., starting time for Laser pumping
    !   t_end : in o.c., ending time for Laser pumping
    !
    ! (NEW for X-ray parameters)
    !   FWHM : in a.u., Full Width at Half Maximum
    !   p_multiple : in multiple of FWHM
    !   N_time : # of iterations for time, N_time = ( p_end - p_start ) / dt
    !   n_photons : # of photons
    !   width : spatial width
    !   fluence : n_photons / width**2
    !   yes_polarized : yes or no about linearly polarization
    !
    integer :: pulse_type
    real(long) :: F, T, w0, chirp_param, FWHM                   ! in a.u.
    real(long) :: dt_au, t_start_au, t_end_au                   ! in a.u.
!    real(long) :: pulse_length, ramp_until, dt, t_start, t_end  ! in o.c.
    real(long) :: n_photons, width, fluence
    real(long) :: p_multiple            ! in multiple of FWHM
    real(long) :: time_delay            ! time delay in a.u.
    real(long) :: fluence_ratio         ! fluence ratio b/w 1st and 2nd pulses
    real(long) :: FWHM_ratio            ! pulse duration ratio b/w 1&2 pulses
    integer :: N_time                   ! # of temporal grid points (per pulse)
    integer :: N_time_ext               ! extension of the temporal grid
    integer*8 :: i0                     ! initial config. for dynamics
    integer :: fluence_start, fluence_end, d_fluence
    logical :: yes_velocity_gauge
    logical :: yes_polarized
    logical :: yes_synchrotron  ! simulate one-photon abs. for the synch. case
    logical :: yes_flattop      ! flattop profile of a pulse in energy domain
    !
    ! manipulate rates and cross sections
    !
    logical :: yes_massage
    integer :: init_occ_max_i   ! max. index for initially occupied orbitals
    real(long) :: massage_autoionization_rate   ! multiply autoionization rates
    real(long) :: massage_auger_rate            ! multiply auger rates
    real(long) :: massage_resonant_pcs          ! multiply photoexcitation cs
    real(long) :: massage_pcs                   ! multiply photoionization cs
    real(long) :: massage_decay_s               ! multiply decay rates for s
    real(long) :: massage_decay_p               ! multiply decay rates for p
    real(long) :: massage_decay_d               ! multiply decay rates for d
    real(long) :: massage_decay_f               ! multiply decay rates for f
    integer    :: massage_auger_q1              ! massage Auger rate for q < q1
    real(long) :: massage_auger_factor1         ! multiply factor1 for q < q1
    integer    :: massage_auger_q2              ! massage Auger rate for q >= q2
    real(long) :: massage_auger_factor2         ! multiply factor1 for q >= q2
    !
    ! scanning parameters: 
    ! to be used for varying fluence or pulse duration in xatom_mc_fly
    !
    logical :: yes_scan
    integer :: scan_type        ! 1: arithmetic, 2: geometric
    real(long) :: scan_common   ! common difference (1) or common ratio (2)
    integer :: scan_start       ! starting index
    integer :: scan_end         ! ending index
    integer :: scan_step        ! increasing step for indexes
    integer :: scan_quantity    ! 1: fluence, 2: pulse duration
    !
    ! Monte-Carlo control
    !
    real(long) :: MC_conv       ! convergence threshold
    integer :: MC_max_events    ! max # of events
    integer :: MC_step          ! print out & check convergence at every MC_step
    real(long) :: MC_dt_ratio   ! dt = 1/(sum of rates) * MC_dt_ratio
    logical :: yes_pathway      ! print out each trajectory in MC
    integer :: N_proc_max       ! max. number of processes
    integer :: max_counter_visit! control # of visits to store atomic data
    integer :: N_config_visited ! NOT PARAMETER; JUST TO BE USED IN MC_FLY
    integer :: N_config_calc    ! NOT PARAMETER; JUST TO BE USED IN MC_FLY
    integer :: N_config_stored  ! NOT PARAMETER; JUST TO BE USED IN MC_FLY
    integer :: N_process_stored ! NOT PARAMETER; JUST TO BE USED IN MC_FLY
    !
    ! Monte-Carlo on-the-fly specific
    !
    logical :: yes_track_res    ! track elec. config. when resonance takes place
    integer :: unit_track_res = 300 
    logical :: yes_track_cs     ! track charge state
    integer :: unit_track_cs = 400
    !
    ! Atom-in-Plasma parameters
    !
    logical :: yes_screening, yes_quasineutral
    logical :: yes_debye_screening, yes_ion_screening, yes_n_elec_screening
    logical :: yes_average_atom, yes_muffin_tin, yes_wigner_seitz
    logical :: yes_grand_canonical  ! whether GC ensemble analysis is printed
    logical :: yes_fixed_config     ! GC analysis & rho_free given by "-config"
    logical :: yes_scaling
    logical :: yes_uniform_ion_dens     ! put a uniform positive ion density
    logical :: yes_turn_off_V_ne        ! turn off v_ext(:)
    logical :: yes_turn_off_V_ee        ! turn off v_J(:) and v_ex(:)
    logical :: yes_fix_chem_pot         ! fix the chemical potential
    real(long) :: chem_pot_fixed        ! chemical potential as a parameter
    logical :: yes_FT_MBPT      ! use FT-MBPT formula, currently (0) only
    real(long) :: debye_length  ! Debye length in a.u.
    real(long) :: pcf(4)        ! fit parameters for pair-correlation func.
    real(long) :: Z_eff         ! effective ionic charge
    real(long) :: ion_dens      ! averaged ionic density
    real(long) :: elec_dens     ! averaged electronic density
    real(long) :: r_s           ! Wigner-Seitz radius
    real(long) :: T_eV          ! temperature (in eV)
    real(long) :: twostep_thres ! threshold for performing 2nd step in two-step
    !
    ! plasma calculation with the external environmental potential from XPOT
    ! XPOT: a bridge between XATOM and XMDYN for IPD-feedback plasma simulation
    !       it can be specified as one of the plasma models treated in XATOM
    ! - yes_xpot: set up separately or determined depending on other options
    ! - yes_load_env: whether the external env. potential file is loaded
    ! - yes_fixed_pot: whether the flat tail is given by external fixed value
    ! - (yes_wigner_seitz): whether the flat tail is given by WS radius
    ! - flat_tail_type: For XPOT calc., the potential tail becomes flat after
    !   a certain radius. This variable defines a type of the flat tail
    !   depending on yes_fixed_pot and yes_wigner_seitz
    !     (0)   : no "flat" tail, the shape is given by a Coulomb potential
    !             even beyond rmax (for continuum-state calculation)
    !     1 [WS]: yes_wigner_seitz=T and yes_fixed_pot=F
    !             muffin-tin approximation, set by a pot value at the WS radius
    !     2 [TS]: yes_wigner_seitz=F and yes_fixed_pot=T
    !             external fixed value, set by a value specified in file_env
    !             and it is usually given from the touching-sphere scheme (TS)
    !             This is also considered as muffin-tin-like approximation
    !     3 [EP]: yes_wigner_seitz=F and yes_fixed_pot=F
    !             given by (loaded or calc.) effective potential (EP), 
    !             which is mainly governed by environmental pot. (also EP).
    !             When r > rmax (for continuum calc.), set by a value at r=rmax
    !
    ! Note that 0 is for test purpose and 1 has not been used in publication.
    !   For 2, the fixed value is determined by the touching-sphere scheme.
    !          SEE Rui et al., PRE 103, 023203 (2021) and PRE 106, 015206 (2022)
    !   For 3, it is developed for Sourav's project on 
    !          radiation damage dynamics influenced by IPD feedback.
    !          Depending on yes_load_pot, the value is determined
    !          from loaded v_eff or from calculated v_eff.
    !
    ! Conditions:
    ! - if flat_tail_type > 0, then automatically yes_xpot = true
    ! - if flat_tail_type = 2, then yes_load_env and yes_fixed_pot must be true
    ! - if flat_tail_type = 3, then yes_load_pot must be true
    ! - if yes_xpot is set to true, then flat_tail_type = 3 as default
    ! - if yes_load_env = true, then flat_tail_type can be any, but > 0
    ! - if yes_fixed_pot = true, then flat_tail_type = 2
    !
    logical :: yes_xpot                 ! turn on for XPOT-type plasma calc.
    logical :: yes_load_env             ! (Rui's orig: yes_v_env)
    logical :: yes_fixed_pot            ! (Rui's orig: use_extern_tail)
    !logical :: yes_plasma_noscf        ! (Rui's orig: yes_se) --> OBSOLETE
    character(len=512) :: file_env      ! (Rui's orig: envpotfile)
    integer :: flat_tail_type           ! defines the flat tail potential
    !
    ! output options
    !   all types of x-ray scattering (photon-in, photon-out):
    !   # elastic scattering: I --> I
    !     - non-resonant: first order of A^2, coherent x-ray scattering
    !     - resonant: second order of p dot A, dispersion correction
    !   # inelastic scattering: I --> F /= I
    !     - non-resonant: first order of A^2, Compton scattering, IXS
    !                     without any intermediate states
    !     - resonant: second order of p dot A, Raman scattering, RIXS
    !                 with intermediate states
    !   c.f.) Basically "Raman" means non-resonant effects.  For examples,
    !         XRS: x-ray Raman scattering (non-resonant inelastic)
    !         RIXS: resonant inelastic x-ray scattering (called resonant Raman)
    !         In our definition (convention), "Raman" means 
    !         inelastic scattering by the second order of p dot A.
    ! In the options,
    !   "-inelastic" for both non-resonant and resonant inelastic in XATOM_REQ
    !   "-compton" for non-resonant inelastic in XATOM
    !   "-raman" to be used for resonant inelastic in XATOM
    !
    integer :: OUTPUT, IO_DATA, IO_SCD, IO_MAD, IO_IXS, IO_IXS2, IO_POP
    integer :: IO_DIR, IO_CONFIG, IO_NULL, IO_NAO, IO_TDIPOLE, IO_AUGER
    integer :: IO_OVERLAP, IO_PULSE
    logical :: yes_verbose, yes_print_pot
    logical :: yes_print_wf, yes_print_unbound
    logical :: yes_HS
    logical :: yes_load_pot, yes_save_pot, yes_load_den
    logical :: yes_fluorescence, yes_auger, yes_pcs, yes_spectra
    logical :: yes_coincidence  ! print out spectra for each final charge
    logical :: yes_occupation   ! print out occupation number dynamics
    logical :: yes_tpcs         ! two-photon cross section (not yet implemented)
    logical :: yes_scattering, yes_dispersion
    logical :: yes_inelastic_scattering
    logical :: yes_create_hole          ! print out all possible single-hole's
    logical :: yes_spline               ! use spline for f' calculation
    logical :: yes_resonance            ! turn on resonant excitation
    logical :: yes_auger_recombination  ! resonant Auger recombination
    logical :: yes_oscillator_str       ! calculate oscillator strength
    logical :: yes_transition_bound_bound       ! bound-bound transition dipole
    logical :: yes_transition_bound_free        ! bound-free transition dipole
    logical :: yes_transition_dipole_bound_free ! bound-free transition dipole
    logical :: yes_transition_auger        ! Auger radial matrix element
    logical :: yes_pcs_beta     ! beta parameter for photoion. cross sec.
    logical :: yes_save_eigen, yes_save_U, yes_save_wf
    logical :: yes_load_eigen, yes_load_U, yes_load_wf
    logical :: yes_silent
    logical :: yes_e_all        ! print out all eigenenergies for diff. l
    logical :: yes_population   ! time-averaged pop. for scattering analysis
    logical :: yes_pop_reduced  ! charge-state pop. instead of config. pop.
    logical :: yes_dat_eV       ! energy in eV loaded from foobar.dat
    logical :: yes_flag         ! it might be useful for various testing!
    logical :: yes_gitversion   ! if yes, print out gitversion and quit
    logical :: yes_load_pulse   ! temporal profile is read from P%file_pulse
    character(len=512) :: filename, file_out
    character(len=512) :: file_data, file_scd, file_mad, file_ixs, file_ixs2
    character(len=512) :: file_pop
    character(len=512) :: file_den, file_pot, data_dir
    character(len=512) :: file_nao, file_tdipole, file_auger, file_overlap
    character(len=512) :: file_pulse
    !
    ! for other programs
    !
    logical :: yes_xmdyn                ! for XMDYN
    logical :: yes_nao                  ! for XMOLECULE
    logical :: yes_overlap              ! for XOVERLAP
    logical :: yes_polarization         ! for XMOLECULE
    !
    ! for Compton scattering
    ! 
    logical :: yes_compton
    integer :: compton_type     ! 0: Waller-Hartree, 1: spectrum, 2: DDSCS
    logical :: compton_theta_list
    real(long), allocatable :: compton_theta_angles(:) ! built from list on command-line (compton_theta_list=.TRUE.)
    real(long) :: compton_theta, compton_d_omega
    real(long) :: compton_d_theta, compton_min_theta
    integer    :: compton_lmax, compton_N_omega, compton_N_theta
    !real(long) :: compton_rmax         ==> replaced with continuum_rmax
    logical :: compton_cfelt_bessel
    logical :: compton_index_half
    !
    ! for impact ionization
    !
    logical :: yes_impact_ionization   
    real(long) :: EE_start, EE_end
    integer :: impact_ionization_type, imp_ion_grid_type
    integer :: imp_ion_N_energy, imp_ion_N_theta, imp_ion_n_in
    real(long) :: imp_ion_d_energy, imp_ion_d_theta, imp_ion_d_in   
    real(long) :: imp_ion_min_theta, imp_ion_energy_scaling
    !
    ! for relativistic calculation
    !
    logical :: yes_relativity
    integer :: N_orb_rel
    integer, allocatable :: init_config_rel(:,:)
    real(long), allocatable :: init_f_occ_rel(:)
  end type Param
contains

!----------------------------------------------------------------------------
! read all parameters from STDIN or input file through get_input( ARG )
! Argument type is defined in stdio.f90
!----------------------------------------------------------------------------
subroutine read_parameters( P, calc_string, yes_wrong_option )
  type(Param), intent(inout) :: P
  character(len=*), intent(in), optional :: calc_string
  logical, intent(in), optional :: yes_wrong_option
  type(Argument) :: ARG
  real(long) :: factor
  character(len=*), parameter :: file_default = "results"
  character(len=*), parameter :: file_null = "/dev/null"
  character(len=*), parameter :: data_dir = "atomic_data"
  integer, parameter :: OUTPUT = 10, IO_DATA = 11, IO_SCD = 12
  integer, parameter :: IO_NAO = 19, IO_TDIPOLE = 20, IO_AUGER = 21, IO_OVERLAP = 999
  integer, parameter :: IO_MAD = 13     ! because you got mad... it should be 13
  integer, parameter :: IO_IXS = 15, IO_IXS2 = 16
  integer, parameter :: IO_POP = 17
  integer, parameter :: IO_NULL = 1, IO_CONFIG = 14    ! 13 is reserved for MAD
  integer, parameter :: IO_PULSE = 18
  integer :: N_e, i, j, l, n, charge, charge2, n_ang
  character(len=20) :: pot_name, x_name, c_name, thermal_x_name, pulse_shape
  character(len=8192) :: file_root, config_str, hole_str
  logical :: yes_charge_override
  logical :: yes_alpha, yes_beta, yes_only_test
  integer, allocatable :: n_occ(:,:)    ! output from parse_config_string()
  real(long), allocatable :: f_occ(:,:), f_occ_tmp(:,:)
  real(long) :: bandwidth_percent
  ! for bound-bound transitions
  integer :: num_vir
  integer, allocatable :: config_tmp(:,:)
  !
  ! set calc_type
  !
  P%calc_type = 0               ! default: TISE
  if ( present( calc_string ) ) then
    select case ( calc_string )
      case ( "TISE", "GPS", "DFT" );            P%calc_type = 0
      case ( "TDSE", "TDGPS", "TDDFT" );        P%calc_type = 1
      case ( "Floquet", "DFT-Floquet", "Floquet-DFT", "Floquet-TDDFT" )
                                                P%calc_type = 2
      case ( "TDFloquet", "TD-Floquet", "TD-DFT-Floquet", "TDDFT-TDFloquet", &
             "TDFloquet-TDDFT", "TDFloquet-DFT" )
                                                P%calc_type = 3
      case ( "HFS", "Herman-Skillman", "HS" );  P%calc_type = 4
      case ( "REQ", "rate_equation", "rate_eq" );       P%calc_type = 5
      case ( "MC", "Monte-Carlo" );                     P%calc_type = 6
      case ( "MC_fly", "mc_fly" );                      P%calc_type = 7
      case ( "TWOSTEP", "twostep", "two-step" );        P%calc_type = 8
      case default;
        call error_message( "calc_type="//calc_string//" is not supported." )
    end select
    P%calc_string = calc_string
  end if
  !
  ! set initial values: define the atomic or molecular system
  !
  P%system = ""
  if ( P%N_N == 0 ) P%N_N = 1   ! default: atom
  if ( P%Z == 0 ) P%Z = 1       ! default: H
  P%isotope = 0                 ! 0 if average atomic weight is used
  if ( P%N_a == 0 ) P%N_a = -1  ! -1 if not assign
  if ( P%N_b == 0 ) P%N_b = -1  ! -1 if not assign
  N_e = -1                      ! total # of electrons
  charge = 0                    ! molecular charge
  yes_charge_override = .false. ! force to use charge from command line (or opt)
  P%N_frozen = 0                ! # of electrons in frozen orbitals
  config_str = ""
  hole_str = ""
  !
  ! set initial values: define grid parameters
  !
  if ( P%N /= 0 ) P%N1 = P%N
  if ( P%N1 == 0 ) P%N1 = 50
  if ( P%L1 == ZERO ) P%L1 = 50.0_long
  if ( P%L2 == ZERO ) P%L2 = 50.0_long
  if ( P%L3 == ZERO ) P%L3 = 50.0_long
  if ( P%Rmax == ZERO ) then
    P%Rmax = 50.0_long          ! maximum radius
  else if ( P%Rmax < ZERO ) then
    P%Rmax = ZERO               ! Rmax -> infinity
  end if
  if ( P%E_limit == ZERO ) P%E_limit = 30.0_long
  P%dr = 0.005_long
  P%rm_thres = 1.0e-8_long
  P%diffuse_r2max=8._long ! for orbitals with <r^2> larger than this we don't calculat transition amplitudes
  P%continuum_rmax = ZERO
  P%lmax = -1
  P%nmax = -1
  !P%Rmax1 = ZERO;  P%Rmax2 = ZERO;  P%Rmax3 = ZERO
  !P%Rmin1 = ZERO;  P%Rmin2 = ZERO;  P%Rmin3 = ZERO
  !
  ! set initial values for complex scaling and Floquet
  !
  if ( P%Mmax == 0 ) P%Mmax = 4
  if ( P%theta == ZERO ) P%theta = 0.45_long
  if ( P%E0 == ZERO ) then
    P%E0 = -0.5_long                                    ! the ground E of H
    if ( P%N_N == 2 ) P%E0 = -1.1026342144949_long      ! the ground E of H2+
  end if
  P%dE = ZERO           ! use machine precision
  if ( P%nev == 0 ) P%nev = 5
  !
  ! set initial values: define the DFT type (XC potential)
  !
  pot_name = ""
  x_name = "Slater"
  c_name = "none"
  thermal_x_name = ""
  P%alpha = ZERO
  P%beta = ZERO
  yes_alpha = .false.
  yes_beta = .false.
  P%yes_OEP_KLI = .false.
  P%yes_Latter = .true.
  P%yes_exact_pot = .false.
  P%yes_thermal = .false. ! no temperature-dependent xc functional
  P%thermal_x_type = 0
  P%yes_twobody = .false.
  !
  ! set initial values: define SCF parameters
  !
  if ( P%SCF_max_iter == 0 ) P%SCF_max_iter = 100
  if ( P%SCF_energy_conv  == ZERO ) P%SCF_energy_conv = 1.0e-7_long
  if ( P%SCF_imag_conv    == ZERO ) P%SCF_imag_conv = 1.0e-3_long
  if ( P%SCF_density_conv == ZERO ) P%SCF_density_conv = 1.0e-6_long
  if ( P%SCF_mixing == ZERO ) P%SCF_mixing = 0.6_long   ! 60% of new
  P%yes_adaptive_mixing = .false.
  P%yes_diis = .false.
  !
  ! set initial laser parameters
  ! for TD-Floquet, t_start = 0 & t_end = 1
  !
  pulse_shape = "Gaussian"
  P%pulse_type = 3      ! default: Gaussian
!  if ( P%t_start == ZERO ) P%t_start = -30.0_long
!  if ( P%t_end   == ZERO ) P%t_end   =  30.0_long
  P%N_time = 20000
  P%N_time_ext = 0
  P%i0 = -1     ! initial config. for dynamics: -1 ==> ground state of neutral
  P%FWHM = 100_long * 1.0e-15_long * SEC2AU     ! == 100 fs
  P%n_photons = 1e12_long                       ! # of photons
  P%width = ONE * 1.0e-6_long * M2AU            ! == 1 micrometer
  P%p_multiple = 8.0_long                       ! 8 times of FWHM
  P%fluence = -ONE      ! will be computed by n_photons / width**2
  P%fluence_start = 0   ! integer number
  P%fluence_end   = 0   ! integer number
  P%d_fluence = 100     ! Delta fluence = 100 a.u. (default)
  P%time_delay = 5.0e-15_long * SEC2AU          ! == 5 fs
  P%fluence_ratio = ONE ! use the same fluences for two pulses
  P%FWHM_ratio    = ONE ! use the same pulse duration for two pulses
  P%bandwidth = -ONE
  bandwidth_percent = ZERO
  P%multiple_bandwidth = TWO
  P%pcs_scan_step = 10 * EV2AU  ! pcs calculations scanned by 10 eV
  P%linewidth = ZERO
  P%N_convolution = 0   ! N=0 means no convolution
  P%yes_harmonics = .false.     ! no harmonics provided
  P%harmonic2 = ZERO
  P%harmonic3 = ZERO
  P%yes_flattop = .false.  ! for flattop profile of a pulse in energy domain
  !
  ! masssage options: manipulate rates and/or cross sections
  !
  P%yes_massage = .false.
  P%massage_autoionization_rate = ONE
  P%massage_auger_rate = ONE
  P%massage_resonant_pcs = ONE
  P%massage_pcs = ONE
  P%massage_decay_s = ONE
  P%massage_decay_p = ONE
  P%massage_decay_d = ONE
  P%massage_decay_f = ONE
  P%massage_auger_q1 = 0
  P%massage_auger_factor1 = ONE
  P%massage_auger_q2 = 0
  P%massage_auger_factor2 = ONE
  !
  ! set initial plasma calculation
  !
  P%yes_screening = .false.
  P%yes_debye_screening = .false.
  P%yes_ion_screening = .false.
  P%yes_n_elec_screening = .false.
  P%yes_average_atom = .false.
  P%yes_muffin_tin = .false.
  P%yes_wigner_seitz = .false.
  P%yes_quasineutral = .true.
  P%yes_grand_canonical = .false.
  P%yes_fixed_config = .false.
  P%yes_scaling = .false.
  P%yes_uniform_ion_dens = .false.
  P%yes_turn_off_V_ne = .false.
  P%yes_turn_off_V_ee = .false.
  P%yes_fix_chem_pot = .false.
  P%chem_pot_fixed = ZERO       ! default value; only used if P%yes_[...]=yes
  P%yes_FT_MBPT = .false.
  P%debye_length = 10.0_long    ! 10 a.u.
  P%pcf(1) = 7.0_long           ! initial values from Robert
  P%pcf(2) = 1.2_long / ANG2AU
  P%pcf(3) = 5.4_long / ANG2AU
  P%pcf(4) = 6.1_long
  P%Z_eff = ZERO
  P%ion_dens = 2.0_long ! default: density is 2.0 g/cm^3
  P%elec_dens = ZERO    ! default: no elec. dens. is specified --> quasineutral
  P%T_eV = ZERO         ! default: temperature T_eV = 0 K = 0 eV
  P%twostep_thres = 1e-6_long   ! default: 2nd step calc. only for prob > 1e-6
  P%r_s = ZERO          ! Wigner-Seitz radius
  !
  ! set initial values: define output options
  !
  P%OUTPUT = STDOUT             ! if not assigned, OUTPUT is STDOUT
  P%IO_DATA = IO_DATA
  P%IO_SCD  = IO_SCD
  P%IO_MAD  = IO_MAD
  P%IO_IXS  = IO_IXS
  P%IO_IXS2 = IO_IXS2
  P%IO_POP  = IO_POP 
  P%IO_NULL = IO_NULL           ! /dev/null
  P%IO_CONFIG = IO_CONFIG       ! process data for each configuration
  P%IO_NAO = IO_NAO
  P%IO_OVERLAP = IO_OVERLAP
  P%IO_TDIPOLE = IO_TDIPOLE
  P%IO_AUGER = IO_AUGER
  P%IO_PULSE = IO_PULSE
  P%yes_print_pot = .false.
  P%yes_print_wf  = .false.
  P%yes_print_unbound = .false.
  P%yes_save_pot = .false.
  P%yes_load_pot = .false.
  P%yes_load_den = .false.
  P%yes_verbose = .false.
  P%yes_synchrotron = .false.
  P%yes_HS = .false.    ! print out the same information as orig. HS code
  P%yes_polarized = .false.     ! default: unpolarized x-ray
  P%yes_save_eigen = .false.;   P%yes_load_eigen = .false.
  P%yes_save_U = .false.;       P%yes_load_U = .false.
  P%yes_save_wf = .false.;      P%yes_load_wf = .false.
  yes_only_test = .false.
  file_root   = ""
  P%file_out  = ""
  P%file_data = ""
  P%file_scd  = ""
  P%file_ixs  = ""
  P%file_ixs2 = ""
  P%file_mad  = ""
  P%file_pop  = ""
  P%file_den  = ""
  P%file_pot  = ""
  P%file_nao  = ""
  P%file_overlap  = ""
  P%file_auger    = ""
  P%file_tdipole  = ""
  P%file_pulse    = ""          ! controlled by P%yes_load_pulse
  P%data_dir  = data_dir
  P%yes_fluorescence = .false.
  P%yes_auger = .false.
  P%yes_pcs = .false.
  P%yes_pcs_beta = .false.
  P%yes_tpcs = .false.
  P%yes_scattering = .false.
  P%yes_inelastic_scattering = .false.
  P%yes_compton = .false.
  P%yes_impact_ionization = .false.                     
  P%yes_dispersion = .false.
  P%yes_resonance = .false.
  P%yes_auger_recombination = .false.
  P%yes_create_hole = .false.
  P%yes_shakeoff = .false.
  P%yes_coulomb = .false.
  P%K_max = 10
  P%yes_oscillator_str = .false.  ! for calculations of oscillator strength
  P%yes_transition_bound_bound = .false.
  P%yes_transition_bound_free = .false.
  P%yes_transition_dipole_bound_free = .false.
  P%yes_transition_auger = .false.
  P%yes_spectra = .false.
  P%yes_coincidence = .false.
  P%yes_occupation = .false.
  P%yes_spline = .true.
  P%yes_load_pulse = .false.            ! file name is given by P%file_pulse
  P%EA1 = ZERO
  P%EA2 = ZERO
  P%EA3 = ZERO
  P%EX1 = ZERO
  P%EX2 = ZERO
  P%EX3 = ZERO
  P%continuum_E = 1000.0_long * EV2AU   ! continuum energy = 1 keV
  P%PE_start = 12000.0_long * EV2AU     ! photon energy = 12 keV
  P%PE_end = P%PE_start
  P%PE_dE = 0.1_long * EV2AU
  P%PE2 = ZERO
  P%PE2_contrib = ZERO
  !
  ! set initial values: for impact ionization parameters
  !
  P%EE_start = 500.0_long * EV2AU                       
  P%EE_end = 0.0_long * EV2AU                           
  P%imp_ion_n_in = 100             ! Only needed if P%imp_ion_grid_type = 2
  P%imp_ion_d_in = 5.0_long        ! Only needed if P%imp_ion_grid_type = 2
  P%imp_ion_n_energy = 200
  P%imp_ion_d_energy = 2.5_long    ! Only needed for -imp_ion_DDCS and -imp_ion_energy_resolved
  P%imp_ion_n_theta = 50
  P%imp_ion_d_theta = 5.0_long     ! Only needed for -imp_ion_DDCS and -imp_ion_angle_resolved
  P%imp_ion_min_theta = 0.0_long
  !P%imp_ion_energy_scaling = 1.0_long
  P%imp_ion_grid_type = 0          ! Only needed for -imp_ion_total_cs and -imp_ion_total_cs_manual
  !
  ! set initial values: for compton scattering parameters
  !
  P%compton_type = 0    ! Waller-Hartree
  P%compton_theta = 60
  P%compton_N_theta = 37
  P%compton_d_theta = 5
  P%compton_min_theta = 0
  P%compton_lmax = 35
  P%compton_N_omega = 64
  P%compton_d_omega = 10
  P%compton_cfelt_bessel = .false.
  P%compton_index_half = .false.
  P%compton_theta_list = .false.
  P%shakeoff_thres = 0.01_long          ! print only shakeoff prob. > 1e-2
  P%shakeoff_type = 0                   ! only for first photoionization
  P%print_threshold = ZERO              ! print all process data
  P%yes_silent = .false.                ! if yes, no print for SCF
  P%yes_gitversion = .false.
  P%yes_flag = .false.
  P%yes_relativity = .false.            ! option for relativistic calculation
  P%yes_dat_eV = .true.                 ! energy in eV loaded from foobar.dat
  P%yes_e_all = .false.                 ! print all eigenenergies for diff. l
  !
  ! print out time-averaged pulse-weighted populations for scattering analysis
  !
  P%yes_population = .false.
  P%yes_pop_reduced = .false.
  !
  ! for other programs
  !
  P%yes_xmdyn = .false.
  P%yes_nao = .false.
  P%yes_overlap = .false.
  P%yes_polarization = .false.
  !
  ! set initial values: for scattering factor calculations
  !
  P%N_scd = 100
  P%Q_start = ZERO
!  P%Q_end = ZERO
  P%Q_end = 20.0_long           ! corresponding to omega = 37.28 keV
  P%resolution = ONE * ANG2AU   ! resolution = 1 Angstrom
  P%N_disp = 100
  P%L_disp = 50.0_long
  P%PE_range = 1000_long * EV2AU        ! scan form omega-1keV to omega+1keV
  !
  ! scanning parameters
  !
  P%yes_scan = .false.
  P%scan_type = 1       ! (1) arithmetic or (2) geometric progression
  P%scan_common = 2     ! common difference (1) or common ratio (2)
  P%scan_start = 0      ! index = scan_start...scan_end, increased by scan_step
  P%scan_end = 0        !   if the initial value is 1, then values with (2) are
  P%scan_step = 1       !   1, 2, 4, 8, 16, 32, 64, 256, 512, ...
  P%scan_quantity = 0   ! 0: none, 1: fluence, 2: pulse duration
  !
  ! Monte-Carlo parameters
  !
  P%MC_conv = 1.0e-4_long
  P%MC_max_events = 100000
  P%MC_step = 100
  P%MC_dt_ratio = 0.1_long
  P%yes_pathway = .false.
  P%N_proc_max = 10000
  P%max_counter_visit = 1       ! once visited, store atomic data right away
  P%yes_track_res = .false.     ! track elec. config. after resonant excitation
  P%yes_track_cs = .false.      ! track charge state of trajectories
  !
  ! plasma calculation with an external environment potential from XPOT
  !
  P%yes_xpot = .false.          ! default: no XPOT-type calculation
  P%file_env = 'pot.txt'        ! default file name for environmental potential
  P%yes_load_env = .false.      ! determine whether env. pot. is used or not
  P%yes_fixed_pot = .false.     ! determine whether touching scheme is used
  !P%yes_plasma_noscf = .false.  ! yes_se = .true. => no SCF is performed
  P%flat_tail_type = 0          ! not assigned
  !
  ! read the parameters
  !
  call get_input( ARG )
  do i = 1, ARG%N
    !call convert_to_lowercase( ARG%option(i) )
    select case ( ARG%option(i) )
      !
      ! determine the atomic or molecular system
      !
      case ( "-s", "-system" );                 P%system = ARG%string(i)
                                                if ( N_e < 0 ) N_e = 0
      case ( "-A", "-atom" );                   P%system = ARG%string(i)
                                                if ( N_e < 0 ) N_e = 0
                                                P%N_N = 1
      case ( "-molecule", "-mole", "-mol" );    P%system = ARG%string(i)
                                                if ( N_e < 0 ) N_e = 0
                                                P%N_N = 2
      case ( "-Z", "-nuclear" );                P%Z = int( ARG%value(i) )
      case ( "-c", "-charge" );                 yes_charge_override = .true.
                                                charge = int( ARG%value(i) )
      case ( "-Na", "-N_a", "-N_alpha" );       P%N_a = int( ARG%value(i) )
      case ( "-Nb", "-N_b", "-N_beta"  );       P%N_b = int( ARG%value(i) )
      case ( "-e", "-N_e", "-N_elec" );         N_e = int( ARG%value(i) )
      case ( "-frozen", "-N_frozen" );          P%N_frozen = int( ARG%value(i) )
      case ( "-isotope" );                      P%isotope = int( ARG%value(i) )
      !
      ! set up the electronic configuration
      !
      case ( "-conf", "-config", "-configuration" );  config_str = ARG%string(i)
      case ( "-hole" );                               hole_str = ARG%string(i)
      !
      ! determine grid parameters
      !
      case ( "-N1", "-N", "-N_r" );     P%N1 = int( ARG%value(i) )
  !    case ( "-N2", "-N_uni" );         P%N2 = ARG%value(i)
  !    case ( "-N3", "-N_R", "-NR" );    P%N3 = ARG%value(i)
      case ( "-L", "-L1" );             P%L1 = ARG%value(i)
  !    case ( "-L2" );                   P%L2 = ARG%value(i)
  !    case ( "-L3" );                   P%L3 = ARG%value(i)
      case ( "-Rmax", "-rmax" );        P%Rmax = ARG%value(i)
  !    case ( "-E_limit" );              P%E_limit = ARG%value(i)
  !    case ( "-Rmax1", "-rmax1" );      P%Rmax1 = ARG%value(i)
      case ( "-Rmax2", "-rmax2" );      P%Rmax2 = ARG%value(i)
  !    case ( "-Rmax3", "-rmax3" );      P%Rmax3 = ARG%value(i)
      case ( "-dr", "-dx" );            P%dr = ARG%value(i)
      case ( "-rm_thres", "-rm_threshold", "-rm", "-r_thres" )
                                        P%rm_thres = ARG%value(i)
      case ( "-diffuse_r2max" )
                                        P%diffuse_r2max = ARG%value(i)
      case ( "-continuum_rmax", "-cont_rmax", "-rmax_cont", "-rmax_continuum" )
                                        P%continuum_rmax = ARG%value(i)
      case ( "-lmax" );                 P%lmax = int( ARG%value(i) )
      case ( "-nmax" );                 P%nmax = int( ARG%value(i) )
      case ( "-nev" );                  P%nev = int( ARG%value(i) )
      !
      ! determine Auger electron energies / X-ray fluorescence energies
      ! input should be in eV!!!
      !
      case ( "-EA1", "-E1", "-EA" );    P%EA1 = ARG%value(i) * EV2AU
      case ( "-EA2", "-E2" );           P%EA2 = ARG%value(i) * EV2AU
      case ( "-EA3", "-E3" );           P%EA3 = ARG%value(i) * EV2AU
      case ( "-EX1", "-EX" );           P%EX1 = ARG%value(i) * EV2AU
      case ( "-EX2" );                  P%EX2 = ARG%value(i) * EV2AU
      case ( "-EX3" );                  P%EX3 = ARG%value(i) * EV2AU
      !
      ! determine photon energy for photoabsorption: input should be in eV!!!
      !                         and elastic scattering
      !
      case ( "-PE", "-photon", "-photon_E", "-photon_energy", &
             "-omega", "-omega_in" )
          P%PE_start = ARG%value(i) * EV2AU
          if ( ARG%value2(i) == ZERO ) then
            P%PE_end = P%PE_start
          else
            P%PE_end = ARG%value2(i) * EV2AU
          end if
      case ( "-PE_start", "-start" )
!          P%yes_pcs = .true.
          P%PE_start = ARG%value(i) * EV2AU
      case ( "-PE_end", "-end" )
!          P%yes_pcs = .true.
          P%PE_end = ARG%value(i) * EV2AU
      case ( "-dE", "-dPE", "-delta", "-D", "-PE_dE" )
!          P%yes_pcs = .true.
          P%PE_dE = ARG%value(i) * EV2AU
          if ( P%PE_dE == ZERO ) P%PE_dE = 0.1_long * EV2AU
      case ( "-PE2" );          P%PE2 = ARG%value(i) * EV2AU      ! in eAV
      case ( "-PE2_contrib" );  P%PE2_contrib = ARG%value(i)      ! ratio
      case ( "-bandwidth" );    P%bandwidth = ARG%value(i)*EV2AU  ! in eV
      case ( "-linewidth" );    P%linewidth = ARG%value(i)*EV2AU  ! in eV
      case ( "-bandwidth_percent" )             ! bandwidth is given as %
        bandwidth_percent = ARG%value(i)
      case ( "-multiple_bandwidth", "-bandwidth_multiple" )
        P%multiple_bandwidth = ARG%value(i)
      case ( "-convolution", "-convol" )        ! turn on convolution
        P%N_convolution = int( ARG%value(i) )
        if ( P%N_convolution == 0 ) P%N_convolution = 200
        if ( .not. ARG%yes(i) ) P%N_convolution = 0
      case ( "-pcs_scan_step", "-pcs_step" )
        P%pcs_scan_step = ARG%value(i) * EV2AU
      case ( "-H2", "-harmonic2", "-HG2" )
        P%yes_harmonics = .true.
        P%harmonic2 = ARG%value(i)      ! ratio for the second harmonic
      case ( "-H3", "-harmonic3", "-HG3" )
        P%yes_harmonics = .true.
        P%harmonic3 = ARG%value(i)      ! ratio for the third harmonic
      !
      ! determine momentum grids for elastic scattering factor
      !
      case ( "-N3", "-N_cff", "-N_Q", "-N_angle", "-N_sc", "-N_scd" )
          P%yes_scattering = .true.
          P%N_scd = int( ARG%value(i) )
      case ( "-Q", "-momentum", "-momentum_transfer" )
          P%yes_scattering = .true.
          if ( ARG%value2(i) == ZERO ) then
            P%Q_end = ARG%value(i)
          else
            P%Q_start = ARG%value(i)
            P%Q_end = ARG%value2(i)
          end if
      case ( "-resolution", "-res" ); P%resolution = ARG%value(i) * ANG2AU
      case ( "-N_disp", "-N_dispersion" ); P%N_disp = int( ARG%value(i) )
      case ( "-L_disp", "-L_dispersion" ); P%L_disp = ARG%value(i)
      case ( "-E_range", "-PE_range", "-DE" ); P%PE_range = ARG%value(i)*EV2AU
      !
      ! determine the DFT type
      !
      case ( "-type", "-xc_type", "-xc", "-DFT", "-functional" )
                                                pot_name = ARG%string(i)
      case ( "-x_type", "-X", "-exchange" );    x_name = ARG%string(i)
      case ( "-c_type", "-C", "-correlation" ); c_name = ARG%string(i)
      case ( "-alpha" );        P%alpha = ARG%value(i);  yes_alpha = .true.
      case ( "-beta"  );        P%beta  = ARG%value(i);  yes_beta  = .true.
      case ( "-KLI", "-OEP", "-SIC", "-OEP-KLI", "-OEP-KLI-SIC" )
                                P%yes_OEP_KLI = ARG%yes(i)
      case ( "-tail", "-Latter", "-latter" );   P%yes_Latter = ARG%yes(i)
      case ( "-exact", "-exact_pot", "-exact_potential" );
                                P%yes_exact_pot = ARG%yes(i)
                                P%yes_latter = .false.  ! we don't need it!
      case ( "-thermal", "-thermal_x_type", "-thermal_exchange", &
             "-finite-temp", "-finite-temperature", &
             "-finite_temp", "-finite_temperature" );
                                P%yes_thermal = .true.
                                thermal_x_name = ARG%string(i)
      case ( "-twobody", "-two-body", "-first-order", &
             "-first-order-correction", "-first-order-E", &
             "-first-order-energy" );           P%yes_twobody = ARG%yes(i)
      !
      ! plasma calculation with an external environment potential from XPOT
      ! NOTE: if a fixed potential tail is NOT used, then it MUST BE 
      !       specified "-fixed_pot off" AFTER "-env", because "-env" always
      !       enforces "-fixed_pot on"
      !       --> To avoid such a confusion, let's NOT use "-env"; instead, 
      !           use "-file_env" to specify a file for environmental potential
      ! NOTE: here is a list of original variable names suggested by Rui
      !   P%yes_v_env       --> P%yes_load_env (=> specialized uniform grid)
      !   P%envpotname      --> P%file_env
      !   P%use_extern_tail --> P%yes_fixed_pot or ext_pot (=> one value)
      !   P%yes_muffin_tin  --> keep using it; needs to be separated from others
      !   P%yes_se          --> P%yes_plasma_noscf; b/c it works only for plasma
      !                         (=> GPS radial grid; same as P%file_pot)
      ! Here is the original comment from Rui
      ! ************************************************************************
      ! Readin in the environmental effect as an external potential
      ! field v_env(:), if only -env specified but not file name try to use
      ! default file pot.txt. If still cannot find file, don't use external
      ! ************************************************************************
      ! ==> HOWEVER, the last has to be reconsidered; currently, if the file is
      !     not found, then it stops with an error message, which is a safer way
      !
      case ( "-xpot", "-XPOT", &
             "-flat_tail", "-flat_tail_type", "-type_flat_tail" )
        !
        ! set up XPOT calculation and determine a type of flat tail potential
        ! USAGE: -xpot or -flat_tail [ NO | WS | TS | EP ]
        !   [NO]: identical to isolated atom (no, zero, or off)
        !   [WS]: given by a value at r_WS within muffin-tin approx. (CASE 1)
        !   [TS]: given by a fixed value at r_T specified in file_env (CASE 2)
        !   [EP]: given by effective potential (mainly env. pot.) (CASE 3)
        !
        P%yes_xpot = .true.             ! turn on this option
        P%yes_quasineutral = .false.    ! enforce to switch off quasineutral
        P%yes_Latter = .false.          ! enforce to switch off Latter
        select case ( ARG%string(i) )
          case ( "NO", "no", "ZERO", "zero", "OFF", "off" )
            P%flat_tail_type = 0
            P%yes_xpot = .false.        ! turn off this option!
          case ( "WS", "wigner_seitz", "wigner" );      P%flat_tail_type = 1
          case ( "TS", "touching", "touching_sphere", &
                 "fixed", "external", "extern" );       P%flat_tail_type = 2
          case ( "EP", "effective_potential", "eff_pot", &
                 "env_pot", "env_potential", &
                 "environment", "environment_potential", &
                 "environmental", "environmental_potential", &
                 "noMT", "no_MT", "no_muffin_tin", "no_muffin" )
                                                        P%flat_tail_type = 3
          case ( "" );                                  P%flat_tail_type = 3
          case default
            call error_message( "flat_tail=" // trim(ARG%string(i)) // & 
              " is not supported; choose [EP|TS|WS]" )
        end select
      case ( "-load_env", "-load_env_pot", "-file_env", "-file_env_pot" )
        P%yes_load_env = .true.         ! turn on this option
        P%file_env = ARG%string(i)      ! set up a filename for env. potential
      case ( "-yes_env", "-yes_env_pot", "-env_pot", &
             "-environment_potential", "-environmental_potential" )
        P%yes_load_env = ARG%yes(i)     ! turn on/off this option
      case ( "-env", "-environment", "-Environment", "-Env" )
        ! This is equivalent to the original implementation from Rui
        ! NOT recommended for further usage; KEEP this for back-compatability
        if ( ARG%string(i) .ne. "" ) P%file_env = ARG%string(i)
        P%yes_load_env = .true.         ! turn on this option
        P%yes_fixed_pot = .true.        ! enforce to switch on extern_tail
        P%yes_wigner_seitz = .false.    ! enforce to switch off wigner-seitz
        P%flat_tail_type = 2            ! yes_wigner_seitz=F & yes_fixed_pot=T
        P%yes_quasineutral = .false.    ! enforce to switch off quasineutral
        P%yes_Latter = .false.          ! enforce to switch off Latter
      case ( "-yes_extern_tail", "-yes_fixed_pot", &
             "-fixed_pot", "-fixed_potential", &
             "-use_extern_tail", "-extern_tail", "-touch_MT" )
        ! This is correponding to the original implementation from Rui
        ! NOT recommended for further usage; KEEP this for back-compatability
        P%yes_fixed_pot = ARG%yes(i)    ! turn on/off this option
        if ( P%yes_fixed_pot ) then
          P%flat_tail_type = 2          ! yes_wigner_seitz=F & yes_fixed_pot=T
        else
          P%flat_tail_type = 3          ! yes_wigner_seitz=F & yes_fixed_pot=F
        end if
      !
      ! There were option names of "-noscf" and "-plasma_noscf", 
      ! which was identical to "-se", but was suggested for better naming.
      ! However, it's still confusing because "noscf" could be used for 
      ! non-plasma calculation, so they have been removed as of Oct. 6, 2023.
      !   ==> "-se [...]" should be equivalent to 
      !       "-flat_tail EP -SCF_max_iter 1 -load_pot [...]"
      !       which looks lengthy, but it is easy to understand...
      !
      case ( "-SE", "-se" )
        ! This is equivalent to the original implementation from Rui
        ! NOT recommended for further usage; KEEP this for back-compatability
        if ( ARG%string(i) .ne. "" ) P%file_pot = ARG%string(i)
        P%SCF_max_iter = 1              ! enforce to turn off SCF; max_iter = 1
        P%yes_load_pot = .true.         ! enforce to load P%file_pot
        P%flat_tail_type = 3            ! yes_wigner_seitz=F & yes_fixed_pot=F
      !
      ! determine SCF parameters
      ! NOTE: "-noscf" should be independent from plasma calculation,
      !       so it has been moved from above to here (sangkil on 2023-10-04)
      !
      case ( "-iteration", "-iter", "-max", "-max_iter", "-max_iteration", &
             "-SCF_max", "-SCF_max_iter", "-SCF_iter", "-SCF_max_iteration" )
                                        P%SCF_max_iter = int( ARG%value(i) )
      case ( "-noscf" );                P%SCF_max_iter = 1  ! it means no SCF
      case ( "-SCF", "-SCF_conv", "-SCF_energy_conv", "-energy_conv", &
                     "-conv", "-real_conv", "-ener_conv" )
                                        P%SCF_energy_conv = ARG%value(i)
      case ( "-SCF_imag_conv", "-SCF_imaginary_conv", &
                               "-imag_conv", "-imaginary_conv" )
                                        P%SCF_imag_conv = ARG%value(i)
      case ( "-SCF_density_conv", "-SCF_dens_conv", "-SCF_den_conv", &
                                  "-density_conv", "-dens_conv", "-den_conv" )
                                        P%SCF_density_conv = ARG%value(i)
      case ( "-SCF_mixing", "-mixing" );  P%SCF_mixing = ARG%value(i)
      case ( "-SCF_adaptive", "-SCF_adaptive_mixing", &
             "-adaptive", "-adaptive_mixing" )
                                        P%yes_adaptive_mixing = ARG%yes(i)
      case ( "-SCF_diis", "-SCF_DIIS", &
             "-diis", "-DIIS" )
                                        P%yes_diis = ARG%yes(i)
      !
      ! determine Laser parameters
      !
      case ( "-laser", "-pulse", "-pulse_shape", "-pulse_type" )
                                        pulse_shape = ARG%string(i)
      case ( "-FWHM", "-fwhm", "-tau", "-temporal_width", "-pulse_duration" )
        P%FWHM = ARG%value(i) * 1.0e-15_long * SEC2AU           ! fs -> a.u.
      case ( "-width", "-spatial_width", "-focal_width" )
        P%width = ARG%value(i) * 1.0e-6_long * M2AU             ! micron -> a.u.
      case ( "-area", "-spatial_area", "-focal_area" )
        P%width = sqrt( ARG%value(i) ) * 1.0e-6_long * M2AU     ! micron -> a.u.
      case ( "-photons", "-N_photons", "-n_photons", "-N_photon", "-n_photon", &
             "-N_ph", "-n_ph", "-Nph", "-nph" )
                                              P%n_photons = ARG%value(i)
      case ( "-pulse_length", "-length", &
             "-multiple", "-pulse_multiple", "-p_multiple", "-p_mult" )
                                              P%p_multiple = ARG%value(i)
      case ( "-fluence" )       ! in a.u.
        P%fluence = ARG%value(i)
        if ( int(ARG%value2(i)) /= 0 ) then  ! fluence can be given as a range
          P%fluence_start = int( P%fluence )      ! NOTE: fluence_start is int
          P%fluence_end   = int( ARG%value2(i) )  ! NOTE: fluence_end is int
        end if
      case ( "-d_fluence", "-delta_fluence", "-df", "-d_f", &
             "-incremenet_fluence", "-fluence_increment", "-f_increment", &
             "-fluence_step", "-f_step" )
        P%d_fluence = int( ARG%value(i) )         ! NOTE: d_fluence is int
      case ( "-fluence_start", "-f_start" ); P%fluence_start = int(ARG%value(i))
      case ( "-fluence_end", "-f_end" );     P%fluence_end   = int(ARG%value(i))
      case ( "-synchrotron", "-synch" ) ! one-photon ionization with synchrotron
        P%yes_synchrotron = ARG%yes(i)
        P%fluence = ZERO                ! make sure zero fluence
      case ( "-N_time", "-Ntime" );             P%N_time = int( ARG%value(i) )
      case ( "-N_time_ext", "-N_time_extension", "-N_time_extra", &
             "-Ntime_ext",  "-Ntime_extension", "-Ntime_extra", &
             "-N_extra_time" )
        P%N_time_ext = int( ARG%value(i) )
      case ( "-initial", "-i0", "-index" );  P%i0 = int(ARG%value(i),kind(P%i0))
      case ( "-delay", "-time_delay", "-split", "-time_split", &
             "-interval", "-time_interval" )
        P%time_delay = ARG%value(i) * 1.0e-15_long * SEC2AU     ! fs -> a.u.
        ! it can be either 7 (two pulses) or 8 (double Gaussian)
        if ( P%pulse_type /= 8 .and. P%pulse_type /= 9 ) &
          P%pulse_type = 7        ! assume two pulses if time_delay is specified
      case ( "-fluence_ratio" );                P%fluence_ratio = ARG%value(i)
      case ( "-FWHM_ratio", "-fwhm_ratio", "-tau_ratio", &
             "-pulse_duration_ratio" )
        P%FWHM_ratio = ARG%value(i)
        if ( P%FWHM_ratio > ONE ) P%FWHM_ratio = ONE / P%FWHM_ratio
      !
      ! scanning parameters: (sequence, series, progression)
      !   define indexes for scanning fluence or pulse duration
      !
      case ( "-scan" )
        if ( ARG%string(i) == "fluence" ) then
          P%scan_quantity = 1
        else if ( ARG%string(i) == "pulse_duration" .or. &
                  ARG%string(i) == "pulse_length" .or. &
                  ARG%string(i) == "pulse" .or. ARG%string(i) == "FWHM" ) then
          P%scan_quantity = 2
        end if
        if ( P%scan_quantity /= 0 ) P%yes_scan = .true.
      case ( "-scan_range", "-scan_index" )
        P%scan_start = int( ARG%value(i) )      ! conversion real -> int
        P%scan_end   = int( ARG%value2(i) )     ! conversion real -> int
      case ( "-scan_start" );           P%scan_start = int( ARG%value(i) )
      case ( "-scan_end" );             P%scan_end   = int( ARG%value(i) )
      case ( "-scan_step" );            P%scan_step  = int( ARG%value(i) )
      case ( "-scan_type" )
        select case ( ARG%string(i) )
          case ( "arithmetic", "linear" );      P%scan_type = 1
          case ( "geometric", "exponential" );  P%scan_type = 2
        end select
      case ( "-scan_difference", "-scan_diff", "-scan_d" )  ! arithmetic prog.
        P%scan_type = 1
        P%scan_common = ARG%value(i)
      case ( "-scan_factor", "-scan_ratio", "-scan_r" )     ! geometric prog.
        P%scan_type = 2
        P%scan_common = ARG%value(i)
      !
      ! Monte-Carlo parameters
      !
      case( "-MC_conv" );                       P%MC_conv = ARG%value(i)
      case( "-MC_max_events", "-max_events" )
                                          P%MC_max_events = int( ARG%value(i) )
      case( "-MC_step", "-step", "-every" );    P%MC_step = int( ARG%value(i) )
      case( "-MC_dt_ratio", "-dt_ratio", "-ratio" )
                                                P%MC_dt_ratio = ARG%value(i)
      case( "-pathway", "-traj", "-trajectory", &
            "-MC_pathway", "-MC_traj", "-MC_trajectory" )
                                                P%yes_pathway = ARG%yes(i)
      case( "-N_process", "-N_process_max", "-N_proc", "-N_proc_max" )
                                             P%N_proc_max = int( ARG%value(i) )
      case( "-max_counter_visit", "-max_visit", "-visit", "-counter_visit" )
        if ( ARG%string(i) == "once" ) then
          P%max_counter_visit = 1
        else if ( ARG%string(i) == "twice" ) then
          P%max_counter_visit = 2
        else
          P%max_counter_visit = int( ARG%value(i) )
        end if
      case( "-visit_once" );    P%max_counter_visit = 1
      case( "-visit_twice" );   P%max_counter_visit = 2
      !
      ! Plasma parameters
      !
      case( "-plasma", "-screen", "-screening" )
        select case ( ARG%string(i) )
          case ( "Debye", "debye" );  P%yes_debye_screening = .true.
          case ( "n_elec", "elec" );  P%yes_n_elec_screening = .true.
          case ( "ion", "ions" );     P%yes_ion_screening = .true.
          case ( "average", "average_atom", "AA" )
                                      P%yes_average_atom = .true.
                                      P%yes_muffin_tin = .true.
                                      P%yes_wigner_seitz = .true.
                                      P%yes_scaling = .false.
          case ( "XPOT", "xpot" );    P%yes_xpot = .true.
                                      P%yes_quasineutral = .false.
                                      P%yes_Latter = .false.
          case default
                                      P%yes_debye_screening = ARG%yes(i)
        end select
      case( "-yes_debye", "-yes_debye_screening", &
            "-yes_Debye", "-yes_Debye_screening" )
        P%yes_debye_screening = .true.
      case( "-yes_ion", "-yes_ion_screening", &
            "-yes_neighbor", "-yes_neighboring", "-yes_neighboring_ions", &
            "-yes_pair", "-yes_pair_corr", "-yes_pcf", &
            "-neighbor", "-neighboring", "-neighboring_ions", &
            "-pair", "-pair_corr", "-pair_correlation", "-pcf" )
        P%yes_ion_screening = ARG%yes(i)
      case ( "-yes_n_elec_density", "-yes_n_elec_dens", "-yes_n_elec", &
             "-yes_elec_density", "-yes_elec_dens", "-yes_elec" )
        P%yes_n_elec_screening = .true.
      case ( "-yes_quasineutral", "-yes_neutral", "-yes_quasi", &
             "-quasineutral", "-neutral", "-quasi" )
        P%yes_quasineutral = ARG%yes(i)
      case ( "-yes_wigner_seitz", "-yes_Wigner_Seitz", "-yes_WS", &
             "-wigner_seitz", "-Wigner_Seitz", "-WS", "-solid" )
        P%yes_wigner_seitz = ARG%yes(i)
      case ( "-yes_muffin_tin", "-yes_mt", "-yes_MT", "-yes_muffin", &
             "-muffin_tin", "-mt", "-MT", "-muffin" )
        P%yes_wigner_seitz = ARG%yes(i) ! muffin-tin should be along with WS
        P%yes_muffin_tin = ARG%yes(i)
      case ( "-yes_grand_canonical", "-yes_GC", "-grand_canonical", "-GC" )
        P%yes_grand_canonical = ARG%yes(i)
        if ( ARG%string(i) == "config" .or. ARG%string(i) == "configure" ) then
          P%yes_grand_canonical = .true.
          P%yes_fixed_config = .true.
        end if
      case ( "-yes_fixed_config", "-yes_fixed", "-fixed_config", &
             "-yes_bound_config", "-yes_bound", "-bound_config" )
        P%yes_fixed_config = ARG%yes(i)
      case ( "-yes_scaling", "-scaling" )
        P%yes_scaling = ARG%yes(i)
      case ( "-yes_fix_chem_pot", "-yes_chem_pot_fix", &
             "-yes_fixed_chem_pot", "-yes_chem_pot_fixed", &
             "-yes_fix_mu", "-yes_mu_fix", "-yes_fixed_mu", "-yes_mu_fixed" )
        P%yes_fix_chem_pot = ARG%yes(i)
      case ( "-chem_pot", "-chemical_potential", "-mu", &
             "-chem_pot_fixed", "-chemical_potential_fixed", "-mu_fixed" )
        P%yes_fix_chem_pot = .true.
        P%chem_pot_fixed = ARG%value(i)         ! given in a.u.
      case ( "-FT_MBPT", "-FTMBPT", "-MBPT" )
        P%yes_FT_MBPT = .true.
      case( "-debye", "-Debye", "-lambda", "-lambda_D", "-lambdaD" )
        P%yes_debye_screening = .true.
        P%debye_length = ARG%value(i)
        if ( P%debye_length == ZERO ) P%yes_debye_screening = .false.
      case( "-pcf_a", "-pair_a", "-pair_corr_a" );      P%pcf(1) = ARG%value(i)
      case( "-pcf_b", "-pair_b", "-pair_corr_b" );      P%pcf(2) = ARG%value(i)
      case( "-pcf_c", "-pair_c", "-pair_corr_c" );      P%pcf(3) = ARG%value(i)
      case( "-pcf_d", "-pair_d", "-pair_corr_d" );      P%pcf(4) = ARG%value(i)
      case( "-ion", "-ions", "-effective_charge", "-eff_charge", "-eff_Z", &
            "-Z_eff", "-Z_ion", "-Z_avg", "-Zeff", "-Zion", "-Zavg" )
        P%Z_eff = ARG%value(i)
      case( "-ion_density", "-ion_dens", "-solid_density" )
        P%ion_dens = ARG%value(i)       ! g/cm^3, to be changed into 1/a.u.^3
      case ( "-n_elec_density", "-n_elec_dens", "-n_elec" )
        P%yes_n_elec_screening = .true.
        P%yes_quasineutral = .false.
        P%elec_dens = ARG%value(i)      ! multiple of ion_dens, to be 1/a.u.^3
        if ( P%elec_dens == ZERO ) P%yes_n_elec_screening = .false.
      case ( "-r_s", "-r_WS", "-WS_radius", "-radius_WS" )
        P%yes_wigner_seitz = .true.
        P%r_s = ARG%value(i)
      case ( "-T_eV", "-temp_eV", "-temperature_eV", &
             "-T_e", "-T_elec", "-temp", "-temperature" )
        P%T_eV = ARG%value(i)
      case ( "-T_K", "-temp_K", "-temperature_K" )
        P%T_eV = ARG%value(i) / EV2K
      case ( "-uniform_ion_dens", "-uniform_ion_density", &
             "-uniform_nuc_charge", "-uniform" )
        P%yes_uniform_ion_dens = ARG%yes(i)
      case ( "-turn_off_V_ne", "-turnoff_V_ne", "-turnoff_Vne" )
        P%yes_turn_off_V_ne = .true.
      case ( "-V_ne", "-nuc_elec", "-V_ext", "-v_ne", "-v_ext", "-Vne", "-vne" )
        P%yes_turn_off_V_ne = .not. ARG%yes(i)
      case ( "-turn_off_V_ee", "-turnoff_V_ee", "-turnoff_Vee" )
        P%yes_turn_off_V_ee = .true.
      case ( "-V_ee", "-elec_elec", "-v_ee", "-Vee", "-vee" )
        P%yes_turn_off_V_ee = .not. ARG%yes(i)
      case ( "-twostep", "-twostep_thres", "-thres_twostep" )
        P%twostep_thres = ARG%value(i)
      !
      ! relativitic calculation
      !
      case ( "-relativity", "-rel", "-relativistic" )
        P%yes_relativity = ARG%yes(i)
      !
      ! determine filename
      !
      case ( "", "-i", "-file", "-input" );     file_root = ARG%string(i)
      case ( "-o", "-output" ); P%OUTPUT = OUTPUT
                                P%file_out = ARG%string(i)
      case ( "-d", "+d", "-dat", "-data", "-process", &
             "-file_data", "-file_dat" )
                                P%file_data = ARG%string(i)
      case ( "-dat_eV", "-data_eV" );           P%yes_dat_eV = ARG%yes(i)
      case ( "-scd", "-sc_data", "-s_data", "-scattering_data", "-file_scd" )
                                P%yes_scattering = .true.
                                P%file_scd = ARG%string(i)
      case ( "-ixs", "-iscd", "-isc_data", "-is_data", "-file_ixs" )
                                P%yes_inelastic_scattering = .true.
                                P%file_ixs = ARG%string(i)
      case ( "-mad", "-mad_data", "-d_data", "-dispersion_data", "-file_mad" )
                                P%yes_dispersion = .true.
                                P%file_mad = ARG%string(i)
      case ( "-file_den", "-file_dens", "-file_density" )
                                P%file_den = ARG%string(i)
      case ( "-den", "-density", "-den_data", "-density_data", &
             "-load_den", "-load_dens", "-load_density" )
                                P%yes_load_den = .true.
                                P%file_den = ARG%string(i)
      case ( "-load_pot", "-load_potential" )
                                P%yes_load_pot = .true.
                                P%file_pot = ARG%string(i)
      case ( "-save_pot", "-save_potential" )
                                P%yes_save_pot = .true.
                                P%file_pot = ARG%string(i)
      case ( "-file_pot", "-file_potential" );  P%file_pot = ARG%string(i)
      case ( "-yes_load_pot" );                 P%yes_load_pot = ARG%yes(i)
      case ( "-yes_save_pot" );                 P%yes_save_pot = ARG%yes(i)
      case ( "-dir" );          P%data_dir = ARG%string(i)

      case ( "-file_nao" );     P%file_nao = ARG%string(i)
      case ( "-file_overlap" ); P%file_overlap = ARG%string(i)
      P%file_overlap = trim(adjustl(P%system))//'.'//trim(adjustl(config_str))//'.xov'
      case ( "-file_auger" );   P%file_auger = ARG%string(i)
      case ( "-file_tdipole" ); P%file_tdipole = ARG%string(i)
      case ( "-file_pulse", "-load_pulse" )
                                P%yes_load_pulse = .true.
                                P%pulse_type = 10       ! read in file
                                P%file_pulse = ARG%string(i)

!---------------------------------Impact Ionization----------------------------
      !
      ! determine electron energy for impact ionization: in eV!!!
      ! and check out all options for impact ionization
      !
      case ("-EE")                                 
          P%EE_start = ARG%value(i) * EV2AU        
      case ("-EE_final", "-EE_end")
          P%EE_end = ARG%value(i) * EV2AU
      case ("-imp_ion_n_energy", "-impact_ion_n_energy", &
            "-imp_ionization_n_energy", "-impact_ionization_n_energy")
          P%imp_ion_N_energy = int( ARG%value(i) )
      case ("-imp_ion_d_energy", "-impact_ion_d_energy", &
            "-imp_ionization_d_energy", "-impact_ionization_d_energy")
          P%imp_ion_d_energy = ARG%value(i)
      case ("-imp_ion_n_in", "-impact_ion_n_in", &
            "-imp_ionization_n_in", "-impact_ionization_n_in")
          P%imp_ion_n_in = int( ARG%value(i) )
      case ("-imp_ion_d_in", "-impact_ion_d_in", &
            "-imp_ionization_d_in", "-impact_ionization_d_in")
          P%imp_ion_d_in = ARG%value(i)
      case ("-imp_ion_n_theta", "-impact_ion_n_theta", &
            "-imp_ionization_n_theta", "-impact_ionization_n_theta")        
          P%imp_ion_N_theta = int( ARG%value(i) )
      case ("-imp_ion_d_theta", "-impact_ion_d_theta", &
            "-imp_ionization_d_theta", "-impact_ionization_d_theta")
          P%imp_ion_d_theta = int( ARG%value(i) )
      case ("-imp_ion_min_theta", "-impact_ion_min_theta", &
            "-imp_ionization_min_theta", "-impact_ionization_min_theta")
          P%imp_ion_min_theta = ARG%value(i)
      !case ("-imp_ion_energy_scaling")      
      !    P%imp_ion_energy_scaling = ARG%value(i)        
      case ("-imp_ion_grid_type")
          P%imp_ion_grid_type = int( ARG%value(i) )
      case ("-impact_ionization_angle_resolved_manual", &
            "-impact_ionization_angle_res_manual", &
            "-imp_ion_angle_resolved_manual", "-imp_ion_angle_res_manual")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 0
      case ("-impact_ionization_total_cross_section_manual", &
            "-impact_ionization_total_cs_manual", &
            "-impact_ionization_tot_cross_section_manual", &
            "-impact_ionization_tot_cs_manual", &
            "-imp_ion_total_cross_section_manual", "-imp_ion_tot_cs_manual", &
            "-imp_ion_tot_cross_section_manual", "-imp_ion_total_cs_manual")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 1
      case ("-impact_ionization_DDCS", "-impact_ion_DDCS", &
            "-imp_ionization_DDCS", "-imp_ion_DDCS", & 
            "-impact_ionization_double", "-impact_ion_double", &
            "-imp_ionization_double", "-imp_ion_double", &
            "-impact_ionization_doubly", "-impact_ion_doubly", &
            "-imp_ionization_doubly", "-imp_ion_doubly")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 2
      case ("-impact_ionization_energy_resolved", &
            "-impact_ionization_energy_res", &       
            "-imp_ion_energy_resolved", "-imp_ion_energy_res")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 3
      case ("-impact_ionization_angle_resolved", &
            "-impact_ionization_angle_res", &       
            "-imp_ion_angle_resolved", "-imp_ion_angle_res")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 4
      case ("-impact_ionization_total_cross_section", &
            "-impact_ionization_total_cs", &
            "-impact_ionization_tot_cross_section", &
            "-impact_ionization_tot_cs", &
            "-imp_ion_total_cross_section", "-imp_ion_total_cs", &
            "-imp_ion_tot_cs", "-imp_ion_tot_cross_section" )
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 5
      case ("-impact_ionization_energy_resolved_symmetric", &
            "-impact_ionization_energy_res_symm", &
            "-imp_ion_energy_resolved_symmetric", "-imp_ion_energy_res_symm")
        P%yes_impact_ionization = .true.
        P%impact_ionization_type = 6
!------------------------------------------------------------------------------
      !
      ! for Compton scattering
      !   P%compton_type == 0 : Waller-Hartree (default)
      !                     1 : spectrum
      !                     2 : doubly differential scattering cross section  
      !
      case ( "-compton", "-Compton", "-compton_scattering", &
             "-Compton_scattering" , "-Comp", "-comp" )
        P%yes_compton = .true.
        select case ( ARG%string(i) )
          case ( "WH", "W_H", "W-H", "waller_hartree", "waller-hartree", &
                 "Waller-Hartree", "Waller_Hartree" );  P%compton_type = 0
          case ( "spectrum", "spectra" );               P%compton_type = 1
          case ( "DDSCS", "DDCS", "total_compton_cross_section", &
                 "doubly", "doubly_differential", "doublydifferential" );
                                                        P%compton_type = 2
          case default;                                 P%compton_type = 0
        end select
      case ( "-waller_hartree", "-compton_waller_hartree", "-WH", &
             "-compton_wh" , "-compton_w_h")
        P%yes_compton = .true.
        P%compton_type = 0
      case ( "-compton_spectrum", "-comptonspectrum");
        P%yes_compton = .true.
        P%compton_type = 1
      case ( "-compton_doubly_differential", "-compton_doublydifferential", &
             "-compton_DDSCS", "-compton_doubly" , &
             "-compton_total_compton_cross_section")
        P%yes_compton = .true.
        P%compton_type = 2
      case ("-compton_nonuniform_structure_factor_orbital", &
            "-compton_nonuniform_struc_fact_orbi", &
            "-comp_nonuni_structure_factor_orbital", &
            "-comp_nonuni_struc_fact_orbi", & 
            "-compton_nonuni_struc_fact_orbi" )    
        P%yes_compton = .true.
        P%compton_type = 3
      case ("-compton_nonuniform_structure_factor", &
            "-compton_nonuniform_struc_fact", &
            "-comp_nonuni_structure_factor", &
            "-comp_nonuni_struc_fact", "-compton_nonuni_struc_fact" )
        P%yes_compton = .true.
        P%compton_type = 4
      case ("-compton_nonuniform_DDCS", "-comp_nonuni_DDCS", &
            "-comp_nonuni_doubly", "-comp_nonuni_double", &
            "-compton_nonuniform_doubly", "-compton_nonuniform_double", &
            "-compton_nonuni_DDCS" )
        P%yes_compton = .true.
        P%compton_type = 5
      case ("-compton_nonuniform_energy_resolved", &
            "-compton_nonuniform_energy_res", &
            "-comp_nonuni_energy_resolved", "-comp_nonuni_energy_res", & 
            "-compton_nonuni_energy_res",  "-compton_nonuni_energy_resovled", &
            "-comp_nonuniform_energy_resolved", "-comp_nonuniform_energy_res")
        P%yes_compton = .true.
        P%compton_type = 6
      case ("-compton_nonuniform_angle_resolved", &
            "-compton_nonuniform_angle_res", &
            "-comp_nonuni_angle_resolved", "-comp_nonuni_angle_res", &
            "-compton_nonuni_angle_res",  "-compton_nonuni_angle_resovled", &
            "-comp_nonuniform_angle_resolved", "-comp_nonuniform_angle_res")
        P%yes_compton = .true.
        P%compton_type = 7
      case ( "-comp_theta", "-compton_theta" );
        if( .not. associated(ARG%val_arr(i)%ptr )) then  
          P%compton_theta_list = .FALSE.
          P%compton_theta     = ARG%value(i)
        else
          P%compton_theta_list = .TRUE.
          ! 
          ! P%compton_theta_angles(:) are filled now
          !
          n_ang = size(ARG%val_arr(i)%ptr(:))   
          allocate(P%compton_theta_angles( n_ang ))  
          P%compton_theta_angles(:) = ARG%val_arr(i)%ptr(:)
        end if
      case ( "-comp_n_theta", "-compton_n_theta" );
                                        P%compton_N_theta   = int(ARG%value(i))
      case ( "-comp_d_theta", "-compton_d_theta" );
                                        P%compton_d_theta   = ARG%value(i)
      case ( "-comp_min_theta", "-compton_min_theta" );
                                        P%compton_min_theta = ARG%value(i)
      case ( "-comp_lmax", "-compton_lmax" );
                                        P%compton_lmax      = int(ARG%value(i))
      case ( "-comp_n_omega", "-compton_n_omega" );
                                        P%compton_N_omega   = int(ARG%value(i))
      case ( "-comp_d_omega", "-compton_d_omega" );
                                        P%compton_d_omega   = ARG%value(i)
      case ( "-comp_rmax", "-compton_rmax" );
                                        P%continuum_rmax    = ARG%value(i)
      case ( "-comp_tbessel", "-compton_tbessel" );
                                        P%compton_cfelt_bessel = ARG%yes(i)
      case ( "-comp_index2", "-compton_index2" );
                                        P%compton_index_half = .true.
      !
      ! manipulate cross sections and rates
      !
      case ( "-massage_autoionization" );
                                P%massage_autoionization_rate = ARG%value(i)
      case ( "-massage_auger" );        P%massage_auger_rate = ARG%value(i)
      case ( "-massage_resonance" );    P%massage_resonant_pcs = ARG%value(i)
      case ( "-massage_pcs" );          P%massage_pcs = ARG%value(i)
      case ( "-massage_decay_s" );      P%massage_decay_s = ARG%value(i)
      case ( "-massage_decay_p" );      P%massage_decay_p = ARG%value(i)
      case ( "-massage_decay_d" );      P%massage_decay_d = ARG%value(i)
      case ( "-massage_decay_f" );      P%massage_decay_f = ARG%value(i)
      case ( "-massage_auger_q1" );      P%massage_auger_q1 = int(ARG%value(i))
      case ( "-massage_auger_factor1" ); P%massage_auger_factor1 = ARG%value(i)
      case ( "-massage_auger_q2" );      P%massage_auger_q2 = int(ARG%value(i))
      case ( "-massage_auger_factor2" ); P%massage_auger_factor2 = ARG%value(i)
      !
      ! determine output options
      !
      case ( "-p", "-plot" );           P%yes_print_pot = .true.
                                        P%yes_print_wf = .true.
                                        P%yes_print_unbound = .true.
      case ( "-pot", "-potential" );    P%yes_print_pot = ARG%yes(i)
      case ( "-bound", "-bound_wf", "-wf_bound", "-occupied", "-occ" )
                                        P%yes_print_wf = .true.
      case ( "-continuum", "-cont", "-continuum_wf", "-unbound", &
             "-wf_continuum", "-wf_cont", "wf_unbound", &
             "-continuum_energy", "-continuum_E", "-cont_E", "-CE" )
                                        P%yes_print_unbound = .true.
                                        P%continuum_E = ARG%value(i) * EV2AU
      case ( "-e_all", "-energy_all", "-all_energies", "-all_eigenenergies", &
             "-e_spectrum", "-energy_spectrum" )
                                        P%yes_e_all = ARG%yes(i)
      case ( "-wf", "-wavefunction" )
        select case ( ARG%string(i) )
          case ( "continuum", "cont", "c", "unbound" )
                                        P%yes_print_unbound = .true.
          case ( "bound", "b" );        P%yes_print_wf = .true.
          case ( "bound+unbound", "bound+continuum", "b+c", "both" )
                                        P%yes_print_wf = .true.
                                        P%yes_print_unbound = .true.
          case default;                 P%yes_print_wf = ARG%yes(i)
        end select
      case ( "-fluo", "-fluorescence" );      P%yes_fluorescence = ARG%yes(i)
      case ( "-auger", "-Auger" );            P%yes_auger = ARG%yes(i)
      case ( "-rate", "-rates", "-decay" );   P%yes_fluorescence = ARG%yes(i)
                                              P%yes_auger = ARG%yes(i)
      case ( "-auger_recombination", "-Auger_recombination", &
             "-auger_recomb", "-Auger_recomb", &
             "-resonant_auger_recombination", "-resonant_Auger_recombination", &
             "-resonant_auger_recomb", "-resonant_Auger_recomb", &
             "-RAR", "-rar" );          P%yes_auger_recombination = ARG%yes(i)
      case ( "-pcs", "-photo", "-photoabsorption", "-photoabs", "-cs", &
             "-photoionization", "-photoion", &
             "-crosssection", "-cross" );     P%yes_pcs = ARG%yes(i)
      case ( "-pcs_beta", "-yes_pcs_beta", "-beta_parameter", "-print_beta" )
                                              P%yes_pcs_beta = ARG%yes(i)
      case ( "-scattering", "-elastic_scattering", &
             "-form_factor", "-formfactor", "-coherent_form_factor", "-cff" )
                                              P%yes_scattering = ARG%yes(i)
      case ( "-resonant", "-resonant_scattering", "-MAD", &
             "-dispersive", "-dispersive_correction", &
             "-dispersion", "-dispersion_correction" )
                                        P%yes_dispersion   = ARG%yes(i)
      case ( "-spline", "-dispersion_spline", "-dispersive_spline", &
             "-MAD_spline" );           P%yes_spline = ARG%yes(i)
      case ( "-inelastic", "-inelastic_scattering" )
                                        P%yes_inelastic_scattering = ARG%yes(i)
      case ( "-tpcs", "-two_photon", "-two_photon_cross_section" )
                                        P%yes_tpcs = ARG%yes(i)
      case ( "-coulomb" );              P%yes_coulomb = ARG%yes(i)
      case ( "-K_max", "-max_K" );      P%K_max = int( ARG%value(i) )
      case ( "-transition", "-trans", "-transition_dipole", "-T", "-t" )
        select case ( ARG%string(i) )
          case ( "bound-bound", "b-b", "bb", "bound", "b" )
            P%yes_transition_bound_bound = .true.
          case ( "bound-free", "b-f", "bf", "free", "f", &
                 "bound-continuum", "b-c", "bc", "continuum", "cont", "c" )
            P%yes_transition_bound_free = .true.
          case ( "transitiondipole" )
            P%yes_transition_dipole_bound_free = .true.
          case ( "auger" )
            P%yes_transition_auger = .true.
          case default
            P%yes_transition_bound_bound = ARG%yes(i)
            P%yes_transition_bound_free  = ARG%yes(i)
        end select
      case ( "-recombination", "-recomb", "-HHG", "-hhg" )
                                        P%yes_transition_bound_free = ARG%yes(i)
      case ( "-resonance" );            P%yes_resonance    = ARG%yes(i)
      case ( "-shakeoff", "-shake_off", "-so" )
                                        P%yes_shakeoff     = ARG%yes(i)
      case ( "-shakeoff_type" )
                                        P%shakeoff_type    = int( ARG%value(i) )
      case ( "-shakeoff_thres", "-shakeoff_threshold", &
             "-so_thres", "-so_threshold", "-SO_thres", "-SO_threshold" )
                                        P%shakeoff_thres = ARG%value(i)
      case ( "-create_hole", "-decay_all", "-decay_hole" )
                                        P%yes_create_hole  = ARG%yes(i)
      case ( "-all" );                  P%yes_fluorescence = ARG%yes(i)
                                        P%yes_auger        = ARG%yes(i)
                                        P%yes_pcs          = ARG%yes(i)
                                        P%yes_scattering   = ARG%yes(i)
                                        P%yes_compton      = ARG%yes(i)
      case ( "-HS", "-compatible", "-Herman", "-Herman-Skillman" )
                                        P%yes_HS = ARG%yes(i)
      case ( "-threshold", "-thres", "-process_thres", &
             "-print", "-print_thres", "-print_threshold", &
             "-cutoff", "-validate" )
                                        P%print_threshold = ARG%value(i)
      case ( "-spectra", "-yes_spectra" );      P%yes_spectra = ARG%yes(i)
      case ( "-coincidence", "-yes_coincidence" )
          P%yes_coincidence = ARG%yes(i)
          if ( P%yes_coincidence ) P%yes_spectra = .true.
      case ( "-occupation", "-yes_occupation", &
             "-occupation_number_dynamics", "-occupation_number", &
             "-occupation_dynamics", "-occ_dynamics", "-occ_n", &
             "-occ_n_dynamics" )
          P%yes_occupation = ARG%yes(i)
      !
      ! determine to save or load data file
      !
      case ( "-U",     "-save_U" );     P%yes_save_U = .true.
      case ( "+U",     "-load_U" );     P%yes_load_U = .true.
!      case ( "-wf",    "-save_wf" );    P%yes_save_wf = .true.
      case ( "+wf",    "-load_wf" );    P%yes_load_wf = .true.
      !
      ! for other programs: XMDYN, XMOLECULE
      !
      case ( "-xmol", "-xmolecule", "-nao" );   P%yes_nao = ARG%yes(i)
      case ( "-xoverlap", "-xovlp", "-overlap" );   P%yes_overlap = ARG%yes(i)
      case ( "-polarization", "-polar" );       P%yes_polarization = ARG%yes(i)
      case ( "-xmdyn" );                        P%yes_xmdyn = ARG%yes(i)
      case ( "-xmdyn_all" )
        P%yes_xmdyn = .true.                    ! -xmdyn on
        P%yes_silent = .true.                   ! -silent on
        P%yes_pcs = .true.                      ! -pcs on
        P%yes_fluorescence = .true.             ! -fluo on
        P%yes_auger = .true.                    ! -auger on
        P%yes_scattering = .true.               ! -scd on
        P%yes_dispersion = .true.               ! -mad on
        P%yes_inelastic_scattering = .true.     ! -ixs on
      !
      ! printing option for time-averaged population
      !
      case ( "-yes_pop", "-yes_population" );   P%yes_population = ARG%yes(i)
      case ( "-yes_reduced_pop", "-yes_reduce_pop", &
             "-yes_pop_reduced", "-yes_pop_reduce", &
             "-yes_charge_pop", "-yes_charge_population", "-yes_pop_charge" )
        P%yes_population = .true.
        P%yes_pop_reduced = ARG%yes(i)
      case ( "-pop", "-population" )
        P%yes_population = .true.
        P%file_pop = ARG%string(i)
      case ( "-reduced_pop", "-reduce_pop", "-pop_reduced", "-pop_reduce", &
             "-charge_pop", "-charge_population", "-pop_charge" )
        P%yes_population = .true.
        P%yes_pop_reduced = .true.
        P%file_pop = ARG%string(i)
      !
      ! misc options
      !
      case ( "-silent", "--silent" );                   P%yes_silent = .true.
      case ( "-v", "-V", "-verbose", "--verbose" );     P%yes_verbose = .true.
      case ( "-test", "--test" );                       yes_only_test = .true.
      case ( "-flag" );                                 P%yes_flag = ARG%yes(i)
      case ( "-help", "-h", "--help" );                 call help_message( P )
      case ( "-version", "--version" );         P%yes_gitversion = .true.
      !
      ! additional options
      !
      case ( "-track_res" );    P%yes_track_res = ARG%yes(i)    
      case ( "-track_cs" );     P%yes_track_cs = ARG%yes(i)   
      case ( "-flattop" );      P%yes_flattop = ARG%yes(i)
      case ( "-oscillator" );   P%yes_oscillator_str = ARG%yes(i)   
      !
      ! error message
      !
      case default
        if ( present(yes_wrong_option) .and. .not. yes_wrong_option ) cycle
        call help_message( P, &
          "there is a wrong option of " // trim(ARG%option(i)) // "." )
    end select
  end do
  !
  ! specify N in the general way
  !
  P%N = P%N1
  !
  ! specify the system to be considered
  !
  if ( P%system /= "" ) then
    call determine_system( P%system, P%N_N, charge2, P%Z, P%Z1, P%Z2 )
    if ( .not. yes_charge_override ) charge = charge2
    N_e = 0
  end if
  if ( P%Z < 1 ) call help_message( P, "Z must be greater than 1." )
  !
  ! set mass of one nucleus <-- this is used in plasma.f90
  !   if isotope is assigned ( not 0 ),
  !     then mass = Z*mass_proton + (isotope-Z)*mass_neutron
  !   otherwise, take average atomic weight from the periodic table
  !
  if ( P%Z /= 0 ) then
    if ( P%isotope == 0 ) then
      if ( P%Z <= size(ATOMIC_MASS) ) then
        P%mass = ATOMIC_MASS(P%Z) * AMU_SI
      else
        write(STDERR,"('# ',a,i3/'# ',a)") &
          "WARNING: isotope of a heavy atom should be assigned for Z>", &
          size(ATOMIC_MASS), &
          "WARNING: atomic mass is set to zero."
        P%mass = ZERO
      end if
    else
      P%mass = ( MASS_P_SI*dble(P%Z) + MASS_N_SI*dble(P%isotope-P%Z) )
    end if
    P%mass = P%mass * KG2AU
  end if
  !
  ! set up the total number of electrons
  !
  if ( N_e == -1 .or. N_e == 0 ) N_e = P%Z - charge
!  if ( N_e == -1 ) then
!    N_e = P%Z                   ! default: neutral ion
!  else if ( N_e == 0 ) then
!    N_e = P%Z - charge
!  end if
  !
  ! set up the electronic configuration using config_str and hole_str
  ! currently, s, p, d, f, g, and h orbitals are supported.
  !
  !   n_occ(:,1) = n                --+
  !   n_occ(:,2) = l                  |--- used for P%init_config(:,1:3)
  !   n_occ(:,3) = n_occ (in total) --+
  !   n_occ(:,4) = n_occ in lower j only --> used for the relativistic case
  !
  !   f_occ(:,1) = fractional n_occ
  !   f_occ(:,2) = fractional n_occ in lower j only
  !
  ! relativistic case: init_config_rel(:,4), init_f_occ_rel(:), N_orb_rel
  ! non-relativistic case: init_config(:,3), init_f_occ(:), N_orb
  !
  
  !
  ! determine the electric configuration
  !
  P%N_orb = find_N_subshell( N_e, config_str )
  allocate( n_occ(P%N_orb,4), f_occ(P%N_orb,2) )
  if ( config_str /= "" ) then
    call parse_config_string( config_str, n_occ, f_occ )
  else
    call Aufbau_principle( N_e, n_occ )    
    f_occ(:,1:2) = dble( n_occ(:,3:4) )
    P%yes_fractional = .false.
  end if

  !
  ! assign lmax if it's not given by a parameter
  !
  if ( P%lmax < 0 ) then
    l = maxval( n_occ(:,2) )     ! l_max
    n = maxval( n_occ(:,1) )     ! l = 0, ..., n_max - 1
    P%lmax = l
    if ( P%yes_resonance .or. P%yes_fluorescence ) P%lmax = l + 1
    if ( P%yes_auger ) P%lmax = max( P%lmax, (n - 1) - 1 )
  end if

  !
  !  assign nmax if it's not given by a parameter
  !
  if ( P%nmax < 0 .and. P%yes_resonance ) then
     P%nmax = maxval( n_occ(:,1) )
  endif

  !
  ! adding virtual and Rydberg states, if resonance flag is true
  !
  if ( P%yes_resonance ) then
     ! back up current n_occ and f_occ
     num_vir = count_virtual_orbital( n_occ, P%nmax, P%lmax )
     allocate( config_tmp(P%N_orb+num_vir,4), f_occ_tmp(P%N_orb+num_vir,2) )
     config_tmp(1:P%N_orb, 1:4) = n_occ(1:P%N_orb, 1:4) 
     f_occ_tmp(1:P%N_orb, 1:2) = f_occ(1:P%N_orb, 1:2)

     ! extend the memory space of n_occ and f_occ, and renew N_orb
     deallocate( n_occ, f_occ )
     allocate( n_occ(P%N_orb + num_vir, 4), f_occ(P%N_orb + num_vir, 2) )
     n_occ(1:P%N_orb, 1:4) = config_tmp(1:P%N_orb, 1:4)
     f_occ(1:P%N_orb, 1:2) = f_occ_tmp(1:P%N_orb, 1:2)
     call add_virtual_orbital( n_occ, f_occ, P%nmax, P%lmax, P%N_orb, num_vir )
     deallocate( config_tmp, f_occ_tmp )
     P%N_orb = P%N_orb + num_vir
     

     ! sort n_occ according to n
!     call sort_config( n_occ, f_occ, P%N_orb )
  endif

  if ( hole_str /= "" ) &
    call parse_config_string( hole_str, n_occ, f_occ, yes_hole = .true. )
  !
  ! convert n_occ into P%init_config and f_occ into P%init_f_occ
  !
  allocate( P%init_config(P%N_orb,3), P%init_f_occ(P%N_orb) )
  P%init_config(:,1:3) = n_occ(:,1:3)
  P%init_f_occ(:) = f_occ(:,1)
  !
  ! relavitistic configuration
  !
  if ( P%yes_relativity ) then
    P%N_orb_rel = count( n_occ(:,2) == 0 ) + count( n_occ(:,2) /= 0 ) * 2
    allocate( P%init_config_rel(P%N_orb_rel,5), P%init_f_occ_rel(P%N_orb_rel) )
    call convert_n_occ_to_config_rel( n_occ, f_occ, &
                                      P%init_config_rel, P%init_f_occ_rel )
  end if
  !
  ! if config_str or hole_str is assigned, 
  ! check out fractional occupation numbers and N_e will be reset.
  !
  if ( config_str /= "" .or. hole_str /= "" ) then
    N_e = sum( P%init_config(:,3) )
    P%N_a = -1;  P%N_b = -1
    P%yes_fractional = any( dble( P%init_config(:,3) ) /= P%init_f_occ(:) )
  end if
  !
  ! sort the electron configuration according to n
  ! --------------------------------------------------------------------------
  ! It's good to have a sorted structure, 
  ! but it causes a lot of trouble when indexing is made.
  ! For correct indexing, the unfilled shell for an atom should be at the end.
  ! That's the reason why I didn't use this part.
  ! --------------------------------------------------------------------------
  !
!  do i = 1, P%N_orb
!    do j = i+1, P%N_orb
!      if ( P%init_config(i,1) > P%init_config(j,1) ) then
!        call swap_index( P%init_config(i,:), P%init_config(j,:) )
!      else if ( P%init_config(i,1) == P%init_config(j,1) .and. &
!                P%init_config(i,2) >  P%init_config(j,2) ) then
!        call swap_index( P%init_config(i,:), P%init_config(j,:) )
!      end if
!    end do
!  end do
  !
  ! specify the numbers of alpha and beta electrons
  !
  if ( P%N_a == -1 .and. P%N_b == -1 ) then
    P%N_b = int( N_e / 2 )
    P%N_a = N_e - P%N_b
  else if ( P%N_a == -1 ) then
    P%N_a = N_e - P%N_b
  else if ( P%N_b == -1 ) then
    P%N_b = N_e - P%N_a
  end if
  if ( P%N_a < 0 .or. P%N_b < 0 ) &
    call help_message( P, "electronic configuration is wrong." )
  if ( P%N_a < P%N_b ) call help_message( P, "choose N(alpha) >= N(beta)" )
  !
  ! FINALLY
  !
  P%N_elec = P%N_a + P%N_b
  if ( P%N_frozen > P%N_elec ) &
    call help_message( P, "N_frozen should be smaller than N_elec." )
  if ( P%i0 > 0 ) then                          ! convert i0 to config
    call index_to_config( P%i0, P%init_config )
    P%N_elec = sum( P%init_config(:,3) )
  else
    call config_to_index( P%init_config, P%i0 )
  end if
  !
  ! set calc_string according to calc_type and N_a+N_b
  !
  select case ( P%calc_type )
    case ( 0 );       P%calc_string = "DFT"
    case ( 1 );       P%calc_string = "TDDFT"
    case ( 2 );       P%calc_string = "Floquet-TDDFT"
    case ( 3 );       P%calc_string = "TDFloquet-TDDFT"
    case ( 4 );       P%calc_string = "HFS"
    case ( 5 );       P%calc_string = "REQ"
    case ( 6 );       P%calc_string = "MC"
    case ( 7 );       P%calc_string = "MC_FLY"
    case ( 8 );       P%calc_string = "TWOSTEP"
  end select
  if ( P%N_elec == 1 ) then
    select case ( P%calc_type )
      case ( 0 );       P%calc_string = "TISE"
      case ( 1 );       P%calc_string = "TDSE"
      case ( 2 );       P%calc_string = "Floquet"
      case ( 3 );       P%calc_string = "TDFloquet"
    end select
  end if
  !
  ! set system name
  !
!  if ( P%system == "" ) then
!    P%Z1 = P%Z;  P%Z2 = 0       ! only atomic case
!    P%isotope1 = P%isotope;  P%isotope2 = 0
!    P%system = system_name( P%Z1, P%Z2, P%isotope1, P%isotope2, N_e )
!  end if
  P%system = system_name( P%Z, 0, P%isotope, 0, P%N_elec )
  !
  ! specify DFT xc name
  !
  if ( pot_name /= "" ) then
    if ( pot_name == "BLYP" ) then
      x_name = "B"
      c_name = "LYP"
    else if ( pot_name == "SVWN" ) then
      x_name = "Slater"
      c_name = "VWN"
    else if ( pot_name == "S" ) then
      x_name = "Slater"
    else if ( pot_name == "thermal" ) then
      P%yes_thermal = .true.
    else
      x_name = pot_name
      c_name = "none"
    end if
  end if
  if ( P%yes_thermal ) then
    if ( thermal_x_name == "Rozsnyai" .or. thermal_x_name == "R" ) then
      P%thermal_x_type = 2      ! Rozsnyai, PRA 5, 1137 (1972)
    else
      P%thermal_x_type = 1      ! Perrot & Dharma-wardana, PRA 30, 2619 (1984)
    end if
  end if
  !
  ! specify DFT xc functional
  !
  P%x_type = 0;  P%c_type = 0
  if ( x_name == "exact" ) then
    P%yes_exact_pot = .true.
    P%x_type = 7        ! put Slater (for plasma electron density)
  end if
  do j = 1, size(FUNCTIONAL_X_NAME)
    if ( x_name == FUNCTIONAL_X_NAME(j) ) P%x_type = j
  end do
  do j = 1, size(FUNCTIONAL_C_NAME)
    if ( c_name == FUNCTIONAL_C_NAME(j) ) P%c_type = j
  end do
  if ( P%x_type == 0 ) call help_message( P, &
    trim(x_name) // " is not applicable for the exchange functional." )
  if ( P%c_type == 0 ) call help_message( P, &
    trim(c_name) // " is not applicable for the correlation functional." )
  if ( P%yes_exact_pot .and. P%N_elec > 2 ) call help_message( P, &
    "The exact potential should be used for N_elec <= 2." );
  if ( P%yes_exact_pot ) P%yes_Latter = .false. ! we don't need it
  !
  ! initialize parameters for the particular functional
  !
  select case ( FUNCTIONAL_X_NAME(P%x_type) )
    case ( "Xa" )
      if ( .not. yes_alpha ) P%alpha = TWO / THREE
    case ( "Xab" )
      if ( .not. yes_alpha ) P%alpha = TWO / THREE
      if ( .not. yes_beta )  P%beta  = 0.0028
    case ( "LB" )
      if ( .not. yes_beta  ) P%beta  = 0.05_long
    case ( "LBa" )
      if ( .not. yes_alpha ) P%alpha = 1.19_long
      if ( .not. yes_beta  ) P%beta  = 0.01_long
  end select
  !
  ! set SCF convergence
  !
  if ( P%SCF_energy_conv > ONE ) &
    P%SCF_energy_conv = 10_long**(-P%SCF_energy_conv)
  if ( P%SCF_imag_conv > ONE ) &
    P%SCF_imag_conv = 10_long**(-P%SCF_imag_conv)
  if ( P%SCF_density_conv > ONE ) &
    P%SCF_density_conv = 10_long**(-P%SCF_density_conv)
  !
  ! specify period T and dt, and set pulse length and N_time
  !
  if ( P%w0 /= ZERO ) P%T = TWO * PI / P%w0     ! DC field if P%w0 = 0
!  if ( P%pulse_length == ZERO ) P%pulse_length = P%t_end - P%t_start
!  if ( P%calc_type == 3 ) then  ! only one o.c. in TD-Floquet
!    P%t_start = ZERO
!    P%t_end = ONE
!  end if
  !
  ! set total time iterations (N_time)
  !
!  if ( P%N_time == 0 ) P%N_time = int( ( P%t_end - P%t_start ) / P%dt )
  !
  ! for dispersion correction on elastic scattering
  !
  !if ( abs( P%PE_end - P%PE_start ) < EPS ) P%yes_spline = .false.
  !
  ! assign pulse_shape according to other parmaeters
  !
  if ( P%fluence == ZERO ) pulse_shape = "flat-top"  ! zero-fluence
  if ( P%yes_load_pulse ) pulse_shape = "file"       ! read from external file
  !
  ! reform pulse shape: either ramp or chirp
  !
  select case ( pulse_shape )
    case ( "ramp", "Ramp", "ramped", "Ramped", "ramping", "Ramping", "turn-on" )
      P%pulse_type = 0          ! 0: ramp
!      if ( P%t_start /= ZERO ) &
!        write(STDERR,*) "t_start is forced to ZERO when laser is ramping"
!      P%t_start = ZERO
    case ( "sin2", "sin^2", "chirp", "chirped", "Chirp", "Chirped" )
      if ( P%chirp_param == ZERO ) then
        P%pulse_type = 1        ! 1: sin^2
      else
        P%pulse_type = 2        ! 2: chirp
      end if
    case ( "Gaussian", "gaussian" )
      P%pulse_type = 3          ! 3: Gaussian
!      if ( P%FWHM == ZERO ) P%FWHM = P%pulse_length * P%T / FOUR
      P%dt_au = P%p_multiple * P%FWHM / dble(P%N_time)
      P%t_start_au = -HALF * P%p_multiple * P%FWHM
      P%t_end_au   =  HALF * P%p_multiple * P%FWHM
    case ( "trapezoidal", "trapezoid", "linear_ramp" )
      P%pulse_type = 4          ! 4: trapezoidal
!      if ( P%t_start /= ZERO ) &
!        write(STDERR,*) "t_start is forced to ZERO when laser is ramping"
!      P%t_start = ZERO
    case ( "cos", "cosine", "cosinoidal" )
      P%pulse_type = 5          ! 5: cosinoidal
    case ( "flat-top", "flattop" )
      P%pulse_type = 6          ! 6: flat-top
      P%dt_au = P%FWHM / dble(P%N_time) ! here P%FWHM is just a pulse length
      P%t_start_au = ZERO
      P%t_end_au   = P%FWHM             ! again, it's just a pulse length
    case ( "two", "two-pulses", "two-pulse", "twopulses", "twopulse", &
           "two-gaussians", "two-gaussian", "twogaussians", "twogaussian", &
           "pump-probe" )
      P%pulse_type = 7
      P%dt_au = P%p_multiple * P%FWHM / dble(P%N_time)
      P%t_start_au = -HALF * P%p_multiple * P%FWHM
      P%t_end_au   = P%time_delay + HALF * P%p_multiple * P%FWHM
    case ( "split", "double", "double-gaussian", "double-Gaussian", "dg" )
      P%pulse_type = 8
      P%dt_au = P%p_multiple * P%FWHM / dble(P%N_time)
      P%t_start_au = -HALF * P%p_multiple * P%FWHM
      P%t_end_au   =  HALF * P%p_multiple * P%FWHM
      if ( P%time_delay < ZERO ) then
        P%t_start_au = P%t_start_au + P%time_delay
      else
        P%t_end_au   = P%t_end_au   + P%time_delay
      end if
    case ( "split_flattop", "double_flattop", "double-flattop", "df" )
      P%pulse_type = 9
      if ( P%FWHM_ratio > ONE ) then    ! FWHW here means just pulse length
        P%FWHM_ratio    = ONE / P%FWHM_ratio  ! set up P%FWHM for a wider pulse
        P%fluence_ratio = ONE / P%fluence_ratio
      end if
      P%dt_au = P%FWHM * P%FWHM_ratio / dble(P%N_time)  ! take the narrower one
      P%t_start_au = -HALF * P%FWHM     ! P%FWHM_ratio always < 1
      P%t_end_au   =  HALF * P%FWHM     ! P%FWHM_ratio always < 1
      if ( P%time_delay < ZERO ) then
        P%t_start_au = P%t_start_au + P%time_delay
      else
        P%t_end_au   = P%t_end_au   + P%time_delay
      end if
    case ( "file" )
      P%pulse_type = 10
      P%dt_au = P%FWHM / dble(P%N_time) ! here P%FWHM is just a pulse length
      P%t_start_au = ZERO
      P%t_end_au   = P%FWHM             ! again, it's just a pulse length
    case ( "DC" )
      P%w0 = ZERO               ! DC field
    case default
      call error_message( trim(pulse_shape)//" is not supported." )
  end select
  !
  ! reset n_photons according to fluence (if fluence is assigned)
  !
  if ( P%fluence < ZERO ) then
    P%fluence = P%n_photons / P%width**2
  else
    P%n_photons = P%fluence * P%width**2
  end if
  if ( P%fluence_start /= P%fluence_end ) then  ! fluence is given as a range
    P%d_fluence = min( P%d_fluence, abs( P%fluence_end - P%fluence_start ) )
    !
    ! set up scanning parameters with arithmetic progression
    !
    P%yes_scan = .true.
    P%scan_type = 1     ! 1: arithmetic
    P%scan_quantity = 1 ! 1: fluence
    P%scan_end = abs( P%fluence_end - P%fluence_start ) / P%d_fluence
    P%scan_common = P%d_fluence
  end if
  if ( P%yes_scan .and. P%scan_start == P%scan_end ) then
    P%scan_end = P%scan_start + P%scan_step * 9         ! make a 10-point scan
  end if
  !
  ! set up photon energy (P%PE_start and P%PE_end)
  !
  if ( P%PE_start < ZERO ) P%PE_start = ZERO
  if ( abs(P%PE_end) < EPS ) P%PE_end = P%PE_start
  !
  ! post-processing for XPOT parameters
  ! - if a flat tail is set up, then it should be XPOT calculation
  ! - if a flat tail is not set up and if XPOT is on, then flat_tail_type=3
  !
  if ( P%flat_tail_type > 0 ) P%yes_xpot = .true.
  if ( P%yes_xpot ) then
    if ( P%flat_tail_type == 0 ) P%flat_tail_type = 3   ! as default
    select case ( P%flat_tail_type )
      case ( 1 )        ! [WS]: muffin-tin approx. with Wigner-Seitz radius
        P%yes_wigner_seitz = .true.
        P%yes_fixed_pot = .false.
      case ( 2 )        ! [TS]: muffin-tin-like approx. with touching-sphere
        P%yes_wigner_seitz = .false.
        P%yes_fixed_pot = .true.
        if ( .not. P%yes_load_env ) then        ! requires file_env
          call error_message( "-flat_tail=TS requires an environmental " // &
                              "potential file (-load_env)..." )
        end if
      case ( 3 )        ! [EP]: no muffin-tin, given by effective potential
        P%yes_wigner_seitz = .false.
        P%yes_fixed_pot = .false.
        if ( P%SCF_max_iter == 1 ) then ! no SCF: used in the fallback strategy
          if ( .not. P%yes_load_pot ) then      ! requires file_pot
            call error_message( "-flat_tail=EP requires an effective " // &
                                "potential file (-load_pot)..." )
          end if
        end if
    end select
  end if
  !
  ! For plasma
  ! - set up P%yes_screening
  ! - convert density (g/cm^3) to number density (1/a.u.^3)
  !   factor: number density --> density
  ! - consider the quasineutral assumption
  ! - this part should be after XPOT, 
  !   because P%yes_wigner_seitz could be changed according to P%flat_tail_type
  !
  P%yes_screening = P%yes_debye_screening .or. P%yes_n_elec_screening .or. &
                    P%yes_ion_screening .or. P%yes_average_atom .or. &
                    P%yes_wigner_seitz .or. P%yes_load_den 

  factor = P%mass * AU2KG * 1000_long / ( AU2M * 100_long )**3
  P%ion_dens  = P%ion_dens / factor             ! from g/cm^3 to 1/a.u.^3
  if ( P%yes_quasineutral ) then
    if ( P%Z_eff == ZERO ) P%Z_eff = P%Z - P%N_elec
    P%elec_dens = P%Z_eff
  end if
  P%elec_dens = P%elec_dens * P%ion_dens        ! from multiple to 1/a.u.^3
  if ( P%r_s == ZERO ) P%r_s = ( THREE / FOUR / PI / P%ion_dens )**(ONE/THREE)
  if ( P%yes_average_atom .and. P%T_eV == ZERO ) then
    write(STDERR,"(a/a)") &
      "The average-atom model at T=0 eV is equal to regular HFS.", &
      "Why don't you use a very small T_eV such as T_eV=0.00001...?"
    stop
  end if
  !
  ! For scattering
  !
  if ( P%yes_inelastic_scattering ) then        ! it's used for XATOM_REQ
    P%yes_scattering = .true.
    P%yes_compton = .true.
  end if
  !
  ! consider continuum rmax
  !
!  if ( P%rmax /= ZERO .and. P%continuum_rmax > P%rmax ) &
!    call error_message( "continuum_rmax is too large!" )
  if ( P%continuum_rmax < ZERO ) &
    call error_message( "continuum_rmax must be > 0" )
  if ( P%yes_compton .and. P%compton_type > 0 ) then    ! if not Waller-Hartree
    if ( P%continuum_rmax == ZERO ) then                ! if not specified
      P%continuum_rmax = 450_long                       ! suggested by Jan Malte
    end if
  end if
  !
  ! consider resonant excitation
  !
  !if ( P%yes_resonance ) P%yes_transition_bound_bound = .true.
  if ( P%bandwidth < ZERO ) then
    if ( P%N_convolution > 0 .or. P%yes_resonance ) then
      if ( bandwidth_percent /= ZERO ) then
        P%bandwidth = P%PE_start * bandwidth_percent / 100_long
      else
        P%bandwidth = P%PE_start * 0.01_long            ! 1% of photon energy
      end if
    else
      P%bandwidth = ZERO
    end if
  end if
  !
  ! check out massage options
  !
  if ( P%massage_autoionization_rate /= ONE .or. &
       P%massage_auger_rate /= ONE .or. &
       P%massage_resonant_pcs /= ONE .or. &
       P%massage_pcs /= ONE .or. &
       P%massage_decay_s /= ONE .or. &
       P%massage_decay_p /= ONE .or. &
       P%massage_decay_d /= ONE .or. &
       P%massage_decay_f /= ONE .or. &
       P%massage_auger_factor1 /= ONE .or. &
       P%massage_auger_factor2 /= ONE ) then
    P%yes_massage = .true.
  end if
  !
  ! set up P%init_occ_max_i : max. index for initially occupied orbitals
  !   actually, this is originally used for P%yes_massage = .true.
  !   but let's use it to differentiate autoionization from Auger decays.
  !
  if ( P%yes_relativity ) then
    do i = 1, P%N_orb_rel
      if ( P%init_config_rel(i,4) > 0 ) P%init_occ_max_i = i
    end do
  else
    do i = 1, P%N_orb
      if ( P%init_config(i,3) > 0 ) P%init_occ_max_i = i
    end do
  end if
  !
  ! check for dynamics problem; system should be given by a neutral atom.
  !
!  if ( P%calc_type == 5 .or. P%calc_type == 6 ) then    ! REQ or MC
!    if ( P%Z /= P%N_elec ) &
!      call error_message( "The system should be a neutral atom." )
!  end if
  !
  ! check save or load options
  !
  if ( P%yes_save_eigen .and. P%yes_load_eigen ) P%yes_save_eigen = .false.
  if ( P%yes_save_U     .and. P%yes_load_U     ) P%yes_save_U     = .false.
  if ( P%yes_save_wf    .and. P%yes_load_wf    ) P%yes_save_wf    = .false.
  !
  ! check yes_only_test
  !
  if ( yes_only_test ) then
    call print_parameters( P )
    stop
  end if
  !
  ! misc.
  !
  if ( P%yes_verbose ) P%yes_silent = .false.
  open( P%IO_NULL, file=file_null )
  !if ( P%yes_xmdyn ) P%yes_silent = .true.
  if ( P%yes_silent ) P%OUTPUT = P%IO_NULL
  !
  ! set filename and open output and data files
  ! process data will be stored in IO_DATA
  !
  if ( P%file_out /= "" ) file_root = P%file_out ! priority for output filename
  P%filename = file_root
  j = index( file_root, ".", .true. )
  if ( j /= 0 ) P%filename = file_root( :j-1)   ! delete the extension
  if ( P%filename  == "" ) P%filename = file_default
  if ( P%file_out  == "" ) P%file_out = trim(P%filename) // ".out"
  if ( P%file_data == "" ) P%file_data = trim(P%filename) // ".dat"
  if ( P%file_scd  == "" ) P%file_scd  = trim(P%filename) // ".scd"
  if ( P%file_ixs  == "" ) P%file_ixs  = trim(P%filename) // ".ixs"
  if ( P%file_mad  == "" ) P%file_mad  = trim(P%filename) // ".mad"
  if ( P%file_pop  == "" ) P%file_pop  = trim(P%filename) // ".pop"
  if ( P%file_den  == "" ) P%file_den  = trim(P%filename) // ".den"
  if ( P%file_pot  == "" ) P%file_pot  = trim(P%filename) // ".pot"
  ! the default for nao, auger, tdipole is to print to stdout
!  if ( P%file_nao  == "" ) P%file_nao  = trim(P%filename) // ".nao"
!  if ( P%file_auger  == "" ) P%file_auger  = trim(P%filename) // ".auger"
!  if ( P%file_tdipole  == "" ) P%file_tdipole  = trim(P%filename) // ".tdipole"
  if ( P%file_ixs2 == "" ) then
    P%file_ixs2 = trim(P%filename) // ".ixs2"
    if ( P%filename == "results" ) P%file_ixs2 = "results_with_background.ixs"
  end if
  i = len_trim( P%data_dir )
  if ( P%data_dir(i:i) == "/" ) P%data_dir = P%data_dir(1:i-1)
  if ( P%OUTPUT /= STDOUT .and. P%OUTPUT /= P%IO_NULL ) &
    open( P%OUTPUT, file=P%file_out )
  if ( P%yes_nao ) call open_file( P%file_nao, P%IO_NAO, P%OUTPUT )
  if ( P%yes_overlap ) call open_file( P%file_overlap, P%IO_OVERLAP, P%OUTPUT )
  if ( P%yes_transition_auger ) &
    call open_file( P%file_auger, P%IO_AUGER, P%OUTPUT )
  if ( P%yes_transition_dipole_bound_free ) &
    call open_file( P%file_tdipole, P%IO_TDIPOLE, P%OUTPUT )
contains
  subroutine open_file( filename, IO, IO2 )
    character(len=*), intent(in) :: filename
    integer, intent(inout) :: IO
    integer, intent(in) :: IO2
    if ( filename == "" ) then
      IO = IO2
    else
      open( IO, file=filename, &
            status="unknown", action="write", form="formatted" )
    end if
  end subroutine open_file
end subroutine read_parameters


!----------------------------------------------------------------------------
! print all parameters according to calc_type
!        1 electron     many electrons
!   - 0: TISE           DFT
!   - 1: TDSE           TDDFT
!   - 2: Floquet        Floquet-TDDFT
!   - 3: TDFloquet      TDFloquet-TDDFT
!   - 4:                HFS
!   - 5:                REQ
!   - 6:                MC
!   - 7:                MC on-the-fly
!   - 8:                TWOSTEP
!----------------------------------------------------------------------------
subroutine print_parameters( P, title )
  type(Param), intent(in) :: P
  character(len=*), intent(in), optional :: title
  character(len=80) :: string, string_today
  integer :: OUTPUT, N, i, j, l
  real(long) :: pulse_duration
  type(Argument) :: ARG_CL
  character(len=8192) :: filename, line
  integer :: INPUT_FILE = 9, status
  if ( P%yes_silent ) return
  OUTPUT = P%OUTPUT
  !
  ! print the title
  !
  string = "XATOM " // trim(P%calc_string)
  if ( present(title) ) string = trim(string) // " : " // title
  write(OUTPUT, "('# ',a)") repeat( "-", len_trim(string) )
  write(OUTPUT, "('# ',a)") trim(string)
  string_today = today()
  write(OUTPUT, "('#'/'# Job started at  : ',a)") trim(string_today)
  write(OUTPUT, "('# ',a)") repeat( "-", len_trim(string) )
  !
  ! print command line and input file if a verbose mode is on
  !
  if ( P%yes_verbose ) then
    call get_options( ARG_CL )
    filename = ""
    if ( ARG_CL%N == 1 .and. &
         ARG_CL%option(1) .eq. "" ) then        ! only input file is given
      filename = ARG_CL%string(1)               ! find out an input file
    else
      write(OUTPUT, "('# ',a)") repeat( "=", 78 )
      write(OUTPUT, "('# DUMP command line: ')", advance="no")
      arg_index: do i = 1, ARG_CL%N             ! print all command-line options
        if ( ARG_CL%option(i) .eq. "" ) then
          filename = ARG_CL%string(i)           ! find out an input file
        else
          write(OUTPUT, "(a,' ')", advance="no") trim(ARG_CL%option(i))
        end if
        if ( ARG_CL%string(i) .ne. "" ) &
          write(OUTPUT, "(a,' ')", advance="no") trim(ARG_CL%string(i))
      end do arg_index
      write(OUTPUT, "()")
    end if
    if ( filename .ne. "" ) then                ! if an input file is given
      write(OUTPUT, "('# ',a)") repeat( "=", 78 )
      write(OUTPUT, "('# DUMP input file: ',a)") trim(filename)
      write(OUTPUT, "('# ',a)") repeat( "=", 78 )
      open(INPUT_FILE, file=filename, action="read", form="formatted")
      do
        read(INPUT_FILE, "(a)", iostat=status) line
        if ( status < 0 ) exit
        write(OUTPUT, "('# ',a)") trim(line)
      end do
      close(INPUT_FILE)
    end if
    if ( ( ARG_CL%N > 1 .or. ARG_CL%option(1) .ne. "" ) .or. &
         filename .ne. "" ) then
      write(OUTPUT, "('# ',a)") repeat( "=", 78 )
      write(OUTPUT, "('#')")
    end if
  end if
  !
  ! atom and electron configuration
  !
  if ( P%Z == P%N_elec ) then
    write(OUTPUT, "('# Atom: ',a)") trim(P%system)
  else
    write(OUTPUT, "('# Atomic ion: ',a)") trim(P%system)
  end if
  write(OUTPUT, "('#   nuclear charge:  ',i3)") P%Z
  if ( .not. P%yes_fractional ) then
    write(OUTPUT, "('#   total electrons: ',i3)") P%N_elec
    write(OUTPUT, "('#   total charge:    ',sp,i3)") P%Z - P%N_elec
    if ( .not. P%yes_relativity ) then
      write(OUTPUT, "('#   configuration: ',a)") &
        trim( config_to_string( P%init_config ) )
    else
      write(OUTPUT, "('#   configuration: ',a)") &
        trim( config_rel_to_string( P%init_config_rel ) )
    end if
  else
    write(OUTPUT, "('#   total electrons: ',f7.3)") sum( P%init_f_occ(:) )
    write(OUTPUT, "('#   total charge:    ',sp,f7.3)") &
      P%Z - sum( P%init_f_occ(:) )
    write(OUTPUT, "('#   configuration: ')", advance="no")
    if ( .not. P%yes_relativity ) then
      do i = 1, P%N_orb
        write(OUTPUT, "(a)", advance="no") &
          trim( orbital_name( P%init_config(i,1), P%init_config(i,2) ) )
        write(string, "(f8.3)") P%init_f_occ(i)
        write(OUTPUT, "(a)", advance="no") trim( adjustl( string ) )
        if ( i /= P%N_orb ) write(OUTPUT, "('_')", advance="no")
      end do
    else
      i = 0
      do
        i = i + 1
        write(OUTPUT, "(a)", advance="no") &
          trim( orbital_name( P%init_config_rel(i,1), P%init_config_rel(i,2) ) )
        write(string, "(f8.3)") P%init_f_occ_rel(i)
        write(OUTPUT, "(a)", advance="no") trim( adjustl( string ) )
        if ( P%init_config_rel(i,2) /= 0 ) then         ! if it's not s orbital
          i = i + 1
          write(OUTPUT, "(',')", advance="no")
          write(string, "(f8.3)") P%init_f_occ_rel(i)
          write(OUTPUT, "(a)", advance="no") trim( adjustl( string ) )
        end if
        if ( i == P%N_orb_rel ) exit
        write(OUTPUT, "('_')", advance="no")
      end do
    end if
    write(OUTPUT, "()")
  end if
  if ( P%yes_fractional ) &
    write(OUTPUT, "('#   fractional configuration is used')")
  if ( P%yes_relativity ) &
    write(OUTPUT, "('#   relativistic configuration is used')")
  write(OUTPUT, "('#')")
  !
  ! computational method
  !
  computational_method: if ( P%calc_type == 4 .or. P%calc_type == 7 .or. &
                             P%calc_type == 8 ) then
    !
    ! DFT xc type and parameters
    !
    write(OUTPUT, "('# Computational method:')")
    write(OUTPUT, "('#   exchange functional: ')", advance="no") 
    if ( P%yes_exact_pot ) then
      if ( P%N_elec == 1 ) then
        write(OUTPUT, "(a)", advance="no") "exact one elec."
      else
        write(OUTPUT, "(a)", advance="no") "exact Hartree-Fock"
      end if
      if ( P%yes_screening ) &
        write(OUTPUT, "(a)", advance="no") " + HFS for free-elec."
    else
      write(OUTPUT, "(a)", advance="no") trim( FUNCTIONAL_X_NAME(P%x_type) )
    end if
    if ( P%yes_OEP_KLI ) write(OUTPUT, "(' (OEP-KLI-SIC)')", advance="no")
    if ( P%yes_Latter  ) &
      write(OUTPUT, "(' (Latter tail correction)')", advance="no")
    if ( P%yes_thermal ) then
      if ( P%thermal_x_type == 1 ) then
        write(OUTPUT, "(' (Perrot-Dharma-wardana functional)')", advance="no")
      else if ( P%thermal_x_type == 2 ) then
        write(OUTPUT, "(' (Rozsnyai functional)')", advance="no")
      end if
    end if
    write(OUTPUT, "()")
    select case ( FUNCTIONAL_X_NAME(P%x_type) )
      case ( "Xa" )
        write(OUTPUT, "('#     alpha = ',f6.3)") P%alpha
      case ( "Xab" )
        write(OUTPUT, "('#     alpha = ',f7.4)") P%alpha
        write(OUTPUT, "('#     beta  = ',f7.4)") P%beta
      case ( "LB" )
        write(OUTPUT, "('#     beta  = ',f6.3)") P%beta
      case ( "LBa" )
        write(OUTPUT, "('#     alpha = ',f6.3)") P%alpha
        write(OUTPUT, "('#     beta  = ',f6.3)") P%beta
    end select
    write(OUTPUT, "('#   correlation functional: ',a)") &
      trim ( FUNCTIONAL_C_NAME(P%c_type) )
    if ( P%yes_thermal ) write(OUTPUT, "('#   finite-temperature calculation')")
    if ( P%yes_relativity ) write(OUTPUT, "('#   ',a)") &
      "relativistic calculation with the first order correction"
    write(OUTPUT, "('#')")
    !
    ! parameters related to SCF
    !
    if ( P%yes_verbose ) then
      write(OUTPUT, "('# SCF procedure:')")
      write(OUTPUT, "('#   max iteration = ',i4)") P%SCF_max_iter
      if ( P%SCF_max_iter > 1 ) then
        if ( P%yes_load_env ) then
          write(OUTPUT, "('#   convergence for zeroth energy " // &
                        "(with file_env) = ',es8.1)") P%SCF_energy_conv
        else
          write(OUTPUT, "('#   convergence for energy  = ',es8.1)") &
            P%SCF_energy_conv
          !if ( P%calc_type >= 2 ) then          ! for Floquet
          !  write(OUTPUT, "('#   convergence for Im(E)   = ',es8.1,a)") &
          !    P%SCF_imag_conv, " (relative)"
          !end if
          write(OUTPUT, "('#   convergence for density = ',es8.1)") &
            P%SCF_density_conv
          if ( P%yes_adaptive_mixing ) then
            write(OUTPUT, "('#   adaptive mixing is used')")
          else
            write(OUTPUT, "('#   mixing ratio = ',f5.2,' of new potentials')") &
              P%SCF_mixing
          end if
        end if
      else
        write(OUTPUT, "('#   no SCF is performed for max iter = 0 or 1')")
      end if
      write(OUTPUT, "('#')")
    end if
    !
    ! grid parameters
    !
    write(OUTPUT, "('# Grid for the spherical coordinate:')")
    write(OUTPUT, "('#   N = ',i4)") P%N1
    write(OUTPUT, "('#   L parameter = ',f8.1)") P%L1
    if ( P%Rmax == ZERO ) then
      write(OUTPUT, "('#   Rmax(a.u.)  = infinity')")
    else
      write(OUTPUT, "('#   Rmax(a.u.)  = ',f8.1)") P%Rmax
    end if
    write(OUTPUT, "('#   lmax  = ',i4)") P%lmax
    write(OUTPUT, "('#')")
    !
    ! compton scattering parameters
    !
    if ( P%yes_compton ) then
      write(OUTPUT, "('# Compton scattering: ')", advance="no")
      select case ( P%compton_type )
        case ( 0 );  write(OUTPUT, "('Waller-Hartree')")
        case ( 1 );  write(OUTPUT, "('spectrum')")
        case ( 2 );  write(OUTPUT, "('doubly differential scattering CS')")
      end select

      if( P%compton_theta_list ) then
        write(OUTPUT, "('#       theta = ',f8.1)")  P%compton_theta_angles(:)
      else
        write(OUTPUT, "('#       theta = ',f8.1)")  P%compton_theta
      end if
      write(OUTPUT, "('#     n_theta = ',i6)")    P%compton_N_theta
      write(OUTPUT, "('#     d_theta = ',f8.1)")  P%compton_d_theta
      write(OUTPUT, "('#   min_theta = ',f8.1)")  P%compton_min_theta
      write(OUTPUT, "('#     lmax    = ',i6)")    P%compton_lmax
      write(OUTPUT, "('#     n_omega = ',i6)")    P%compton_N_omega
      write(OUTPUT, "('#     d_omega = ',f8.1)")  P%compton_d_omega
      write(OUTPUT, "('#     rmax    = ',f8.1)")  P%continuum_rmax
      write(OUTPUT, "('#')")
    end if
  end if computational_method
  !
  ! just emphasize we are including shake-off processes!
  !
  if ( P%yes_shakeoff ) then
    write(OUTPUT, "('# Shakeoff calculation: ')", advance="no")
    !if ( P%shakeoff_type <= 0 ) then
    !  write(OUTPUT, "('for only photoionization')") 
    !else
      write(OUTPUT, "('for all processes')") 
    !end if
    !write(OUTPUT, "('#   charge states upto ',sp,i3)") abs(P%shakeoff_type)
    !write(OUTPUT, "('#   braching ratio >= ',es9.1)") P%shakeoff_thres
    write(OUTPUT, "('#')")
  end if
  !
  ! just emphasize we are calculating the dispersion correction!
  !
  if ( P%yes_dispersion ) &
    write(OUTPUT, "('# Dispersion correction: ',a/'#')") "f = f0 + f' + i f''"
  !
  ! threshold for printing the process data
  !
  if ( P%print_threshold /= ZERO ) then
    write(OUTPUT, "('# Printing process data when number >= ',es10.2/'#')") &
      P%print_threshold
  end if
  !
  ! laser parameters and absorber parameters
  !
  dynamic_parameters: if ( P%calc_type >= 5 ) then
    write(OUTPUT, "('# X-ray parameters:')")
    if ( P%fluence /= ZERO ) then
      write(OUTPUT, "('#   Photon energy (eV)  = ',f9.2)") P%PE_start * AU2EV
      if ( P%bandwidth /= ZERO ) then
        write(OUTPUT, "('#   - bandwidth (eV)    = ',f9.2,'  (',f6.2,'%)')") &
          P%bandwidth * AU2EV, P%bandwidth / P%PE_start * 100_long
        write(OUTPUT, "('#   - multiple of bandwidth (x',f4.1,') will be " // &
          "considered in resonance/convolution')") P%multiple_bandwidth
      end if
      if ( P%yes_resonance ) then
        write(OUTPUT, "('#   - resonant excitation is on ')", advance="no")
        if ( P%N_convolution > 0 ) then
          write(OUTPUT, "(a)") "with convolution"
        else
          write(OUTPUT, "(a)") "with no convolution"
        end if
      end if
      if ( P%yes_harmonics ) then
        if ( P%harmonic2 > ZERO ) &
          write(OUTPUT,"('#   - second harmonic contrib. = ',f7.4)") P%harmonic2
        if ( P%harmonic3 > ZERO ) &
          write(OUTPUT,"('#   - third harmonic contrib.  = ',f7.4)") P%harmonic3
      end if
      if ( P%PE2 /= ZERO ) then
        write(OUTPUT,"('#   - 2nd photon E (eV)  = ',f9.2)") P%PE2 * AU2EV
        write(OUTPUT,"('#   - 2nd photon E contrib. = ',es10.2)") P%PE2_contrib
      end if
      !
      ! single-point fluence
      !
      write(OUTPUT, "('#   Number of photons   = ',es10.2)") P%n_photons
      write(OUTPUT, "('#   Spatial width (micrometer) = ',f8.3)") &
        P%width * AU2M / 1.0e-6_long
      write(OUTPUT, "('#   Fluence (1/a.u.^2)  = ',f12.2)") P%fluence
      !
      ! fluence is given as a range
      !
      if ( P%fluence_start /= P%fluence_end ) then
        write(OUTPUT, "('#   - given range for F = ',i10,' - ',i10)") &
          P%fluence_start, P%fluence_end
        write(OUTPUT, "('#   - inclement for F   = ',i10)") P%d_fluence
      end if
    else
      N = 1; j = 0                      ! find out N (# of all possible config)
      do i = 1, P%N_orb - 1             ! open-shell must locate at the end
        l = P%init_config(i,2)
        j = j + ( 4 * l + 2 )           ! max_occ = 4l+2; added to j
        N = N * ( 4 * l + 2 + 1 )       ! possible config: 0...max_occ
      end do
      N = N * ( ( P%Z - j ) + 1 )
      !
      ! zero fluence cases
      !   a) if there is inner-shell vacancy, then it runs with all relaxations.
      !   b) if yes_synchrotron is on, then it makes single absorption, creates
      !      a inner-shell vacancy according to photoionization cross sections
      !      and runs with all relaxations.
      !   c) if yes_synchrotron is off, then it means the ground-state neutral 
      !      atom without any inner-shell vacancy. This is a trivial case, which
      !      doesn't do any dynamics.
      !
      if ( P%i0 /= N ) then     ! NOT ground-state config. of neutral atom
        write(OUTPUT, "('#   Inner-shell vacancy: '," // &
                      "'no photoionization and decay from given config')")
      else if ( P%yes_synchrotron ) then
        write(OUTPUT, "('#   Synchrotron simulation: '," // &
                      "'only one photoionization')")
        write(OUTPUT, "('#   Photon energy (eV) = ',f9.2)") P%PE_start * AU2EV
      else
        write(OUTPUT, "('#   Number of photons = ',es10.2)") P%n_photons
        write(OUTPUT, "('#   Zero fluence: no photoionization')")
      end if
    end if
    write(OUTPUT, "('#')")
    if ( P%pulse_type == 3 ) then               ! Gaussian
      pulse_duration = P%FWHM * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Gaussian envelope')") 
      if ( pulse_duration >= ONE ) then
        write(OUTPUT, "('#   Temporal width, tau (fs) [FWHM] = ',f8.1)") &
          pulse_duration
      else
        write(OUTPUT, "('#   Temporal width, tau (fs) [FWHM] = ',es13.5)") &
          pulse_duration
      end if
      write(OUTPUT, "('#   Pulse length = ',f5.1,' times of tau')") P%p_multiple
      write(OUTPUT, "('#   Number of temporal grids = ',i8)") P%N_time
    else if ( P%pulse_type == 6 ) then          ! flat-top
      write(OUTPUT, "('#   Flat-top envelope')") 
      write(OUTPUT, "('#   Pulse length (fs) = ',f8.1)") &
        ( P%t_end_au - P%t_start_au ) * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Number of temporal grids = ',i8)") P%N_time
    else if ( P%pulse_type == 7 ) then          ! pump-probe
      write(OUTPUT, "('#   Two Gaussian-envelope pulses')") 
      write(OUTPUT, "('#   Temporal width, tau (fs) [FWHM] = ',f8.1)") &
        P%FWHM * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Pulse length = ',f5.1,' times of tau per pulse')") &
        P%p_multiple
      write(OUTPUT, "('#   Number of temporal grids per pulse = ',i8)") P%N_time
      write(OUTPUT, "('#   Time delay (fs) between two pulses = ',f8.1)") &
        P%time_delay * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Fluence of the pump  pulse (1/a.u.^2) = ',f15.2)") &
        P%fluence * P%fluence_ratio
      write(OUTPUT, "('#   Fluence of the probe pulse (1/a.u.^2) = ',f15.2)") &
        P%fluence
      write(OUTPUT, "('#   Ratio of fluences between two pulses = ',f5.1)") &
        P%fluence_ratio
    else if ( P%pulse_type == 8 ) then          ! double Gaussian
      write(OUTPUT, "('#   Double-Gaussian envelope (split)')") 
      write(OUTPUT, "('#   (W)ider temporal width, tau (fs) [FWHM] = ',f10.3)")&
        P%FWHM * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   (N)arrower temporal width (fs)   [FWHM] = ',f10.3)")&
        P%FWHM * AU2SEC / 1.0e-15_long * P%FWHM_ratio
      write(OUTPUT, "('#   Pulse length = ',f5.1,' times of tau per pulse')") &
        P%p_multiple
      write(OUTPUT, "('#   Number of temporal grids per pulse = ',i8)") P%N_time
      write(OUTPUT, "('#   Time interval (fs) b/w double Gaussians     " // &
                    " = ',f8.1)") P%time_delay * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Ratio (N/W) of fluences b/w double Gaussians" // &
                    " = ',f8.1)") P%fluence_ratio
    else if ( P%pulse_type == 9 ) then          ! double flattop
      write(OUTPUT, "('#   Double-flattop envelope (split)')") 
      write(OUTPUT, "('#   (W)ider pulse length (fs)    = ',f10.3)") &
        P%FWHM * AU2SEC / 1.0e-15_long  ! P%FWHM: pulse length for wider pulse
      write(OUTPUT, "('#   (N)arrower pulse length (fs) = ',f10.3)") &
        P%FWHM * P%FWHM_ratio * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Number of temporal grids for narrow pulse = ',i8)") & 
        P%N_time
      write(OUTPUT, "('#   Time interval (fs) b/w double flattops     " // &
                    " = ',f8.1)") P%time_delay * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Ratio (N/W) of fluences b/w double flattops" // &
                    " = ',f8.1)") P%fluence_ratio
    else if ( P%pulse_type == 10 ) then         ! read from file
      write(OUTPUT, "('#   Temporal profile from file: ',a)") trim(P%file_pulse)
      write(OUTPUT, "('#   Pulse length (fs) = ',f8.1)") &
        ( P%t_end_au - P%t_start_au ) * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   Number of data points from file = ',i8)") P%N_time
    end if
    if ( P%calc_type == 5 .or. P%calc_type == 8 ) then
      write(OUTPUT, "('#   dt = ',f10.5,' (a.u.) = ',f10.5,' (fs)')") &
        P%dt_au, P%dt_au * AU2SEC / 1.0e-15_long
    else if ( P%calc_type == 6 .or. P%calc_type == 7 ) then
      write(OUTPUT, "('#   dt < ',f10.5,' (a.u.) = ',f10.5,' (fs)')") &
        P%dt_au, P%dt_au * AU2SEC / 1.0e-15_long
      write(OUTPUT, "('#   * dt can be adjusted according to the rates.')")
    end if
    write(OUTPUT, "('#')")
    !
    ! for Monte-Carlo calculations
    !
    if ( P%calc_type == 6 .or. P%calc_type == 7 ) then
      write(OUTPUT, "('# Monte-Carlo parameters:')")
      write(OUTPUT, "('#   max events = ',i10)") P%MC_max_events
      write(OUTPUT, "('#   convergence for population = ',es8.1)") P%MC_conv
      write(OUTPUT, "('#   print and check convergence " // &
                    "at every step of ',i5,' events')") P%MC_step
      write(OUTPUT, "('#   dt ratio = ',es8.1)") P%MC_dt_ratio
      write(OUTPUT, "('#   * dt is given by 1/(sum of rates) times dt_ratio.')")
      if ( P%calc_type == 7 ) then
        write(OUTPUT, "('#   atomic data stored in memory after visited ')", &
                      advance="no")
        if ( P%max_counter_visit == 1 ) then
          write(OUTPUT, "('once')")
        else
          write(OUTPUT, "(i2,' times')") P%max_counter_visit
        end if
      end if
      write(OUTPUT, "('#')")
    end if
    !
    ! print out the configuration at the beginning
    !
    write(OUTPUT, "('# Initial configuration:')")
    if ( P%i0 > 0 ) write(OUTPUT, "('#   index  = ',i12)") P%i0
    write(OUTPUT, "('#   config = ',a,' (',sp,i3,')')") &
      trim( config_to_string(P%init_config) ), P%Z - sum( P%init_config(:,3) )
    write(OUTPUT, "('#')")
  end if dynamic_parameters
end subroutine print_parameters


subroutine determine_system( string, N_N, charge, Z, Z1, Z2 )
!
! analysis the atomic or the molecular name: H or H2 or H2+ or H2++
!
  character(len=*), intent(in) :: string
  integer, intent(out) :: N_N, charge, Z, Z1, Z2
  character(len=2) :: atom(2)
  character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
  integer :: i, j, atom_Z(2)
  charge = 0
  N_N = 0
  j = 0
  atom(:) = "";  atom_Z(:) = 0
  analysis_string: do i = 1, len( string )
    select case ( string(i:i) )
      case ( "+" );   charge = charge + 1
      case ( "-" );   charge = charge - 1
      case ( "2" );   N_N = 2
    end select
    if ( index( uppercase, string(i:i) ) /= 0 ) then
      j = j + 1
      if ( j > 2 ) exit analysis_string
      if ( i < len( string ) .and. &
           index( lowercase, string(i+1:i+1) ) /= 0 ) then
        atom(j) = string(i:i+1)
      else
        atom(j) = string(i:i)
      end if
    end if
  end do analysis_string
  if ( j == 0 .or. j > 2 ) then
    write(STDERR,*) "It supports only atoms or diatomic molecules"
    stop
  end if
  if ( N_N == 0 ) N_N = j
  !
  ! find atomic name in the periodic table
  !
  do j = 1, 2
    if ( atom(j) == "" ) cycle
    atom_Z(j) = get_atomic_number( atom(j) )
  end do
  !
  ! assign Z1, Z2, and Z
  !
  if ( N_N == 1 ) then
    Z = atom_Z(1)
  else
    Z1 = atom_Z(1)
    Z2 = atom_Z(2)
    if ( Z2 == 0 ) Z2 = Z1    ! homonuclear case
  end if
end subroutine determine_system


function system_name( Z1, Z2, isotope1, isotope2, N_e ) result (string)
!
! return molecular name like N2++
!
  integer, intent(in) :: Z1, Z2, isotope1, isotope2, N_e
  character(len=20) :: string, atom(2)
  integer :: i, charge, Z(2), isotope(2)
  Z(1) = Z1;  isotope(1) = isotope1
  Z(2) = Z2;  isotope(2) = isotope2
  do i = 1, 2
    if ( Z(i) <= 0 ) cycle
    if ( Z(i) <= size(ATOMIC_NAME) ) then
      atom(i) = ATOMIC_NAME(Z(i))
    else
      write(atom(i),"(i20)") Z(i)
      atom(i) = "(" // trim(adjustl(atom(i))) // ")"
    end if
    if ( isotope(i) /= 0 ) then
      if ( Z(i) == 1 .and. isotope(i) == 1 ) then
        atom(i) = "H"
      else if ( Z(i) == 1 .and. isotope(i) == 2 ) then
        atom(i) = "D"
      else if ( Z(i) == 1 .and. isotope(i) == 3 ) then
        atom(i) = "T"
      else
        write(string,"(i20)") isotope(i)
        atom(i) = "[" // trim(adjustl(string)) // "]" // trim(atom(i))
      end if
    end if
  end do
  !
  ! first atom
  !
  string = atom(1)
  !
  ! second atom
  !
  if ( atom(1) == atom(2) ) then
    string = trim(string) // "2"
  else if ( Z2 /= 0 ) then
    string = trim(string) // trim(atom(2))
  end if
  !
  ! put + or -
  !
  charge = Z1 + Z2 - N_e
  if ( charge > 0 ) then
    string = trim(string) // repeat( "+", abs(charge) )
  else if ( charge < 0 ) then
    string = trim(string) // repeat( "-", abs(charge) )
  end if
end function system_name

!----------------------------------------------------------------------------
! parse config_string into words
!   1s2_2s2_2p6 --> 1s2, 2s2, 2p6
! and each word will be into parse_subshell()
!----------------------------------------------------------------------------
subroutine parse_config_string( string, n_occ, f_occ, yes_hole )
  character(len=*), intent(in) :: string
  integer, intent(inout) :: n_occ(:,:)
  real(long), intent(inout) :: f_occ(:,:)
  logical, intent(in), optional :: yes_hole
  integer :: i, j, js, je, k, jmax
  integer :: n, l, N_subshell
  integer :: num, num_lo, num_hi
  real(long) :: f_num, f_num_lo, f_num_hi
  character(len=*), parameter :: subshell_separator = "_ "  ! by "_" or " "
  N_subshell = size(n_occ,dim=1)
  i = 0         ! subshell index
  js = 1
  jmax = len( trim(string) )
  do j = 1, jmax + 1
    if ( j <= jmax ) then
      if ( index( subshell_separator, string(j:j) ) == 0 ) cycle
    end if
    i = i + 1
    je = j - 1
    call parse_subshell( string(js:je), n, l, &
                         num_lo, num_hi, f_num_lo, f_num_hi )
    num = num_lo + num_hi
    f_num = f_num_lo + f_num_hi
    if ( present( yes_hole ) .and. yes_hole ) then
      !
      ! num and f_num are hole numbers, so they are subtracted from occ.
      !
      do k = 1, N_subshell
        if ( n_occ(k,1) == n .and. n_occ(k,2) == l ) then
          n_occ(k,3) = n_occ(k,3) - num       ! n_occ(:,3) : total occ. #
          n_occ(k,4) = n_occ(k,4) - num_lo    ! n_occ(:,4) : occ# in lower j
          f_occ(k,1)  = f_occ(k,1)  - f_num     ! f_occ(:,1)  : total occ. #
          f_occ(k,2)  = f_occ(k,2)  - f_num_lo  ! f_occ(:,2)  : occ# in lower j
          if ( n_occ(k,3) < 0 .or. f_occ(k,1) < ZERO .or. &
               n_occ(k,4) < 0 .or. f_occ(k,2) < ZERO .or. &
               n_occ(k,3) - n_occ(k,4) < 0 .or. &
               f_occ(k,1)  - f_occ(k,2)  < ZERO ) then
            write(STDERR,*) "n = ", n_occ(k,1)
            write(STDERR,*) "l = ", n_occ(k,2)
            write(STDERR,*) "occ = ", n_occ(k,3), f_occ(k,1)
            write(STDERR,*) "occ in lower j = ", n_occ(k,4), f_occ(k,2)
            call error_message( "occupation number becomes negative!" )
          endif
        end if
      end do
    else
      n_occ(i,1) = n          ! quantum number n
      n_occ(i,2) = l          ! quantum number l
      n_occ(i,3) = num        ! integer occupation number
      n_occ(i,4) = num_lo     ! integer occupation number in lower j
      f_occ(i,1) = f_num      ! fractional occupation number
      f_occ(i,2) = f_num_lo   ! fractional occupation number in lower j
    end if
    js = j + 1
  end do
end subroutine parse_config_string

!----------------------------------------------------------------------------
! parse one word for subshell information (relativity ready!)
!   1s2   --> n=1, l=0, num_lo=0, num_hi=2, f_num_lo=0.0, f_num_hi=2.0
!   2p6   --> n=2, l=1, num_lo=2, num_hi=4, f_num_lo=2.0, f_num_hi=4.0
!   2p1,3 --> n=2, l=1, num_lo=1, num_hi=3, f_num_lo=1.0, f_num_hi=3.0
!   3s1.5 --> n=3, l=0, num_lo=0, num_hi=2, f_num_lo=0.0, f_num_hi=1.5
!----------------------------------------------------------------------------
subroutine parse_subshell( word, n, l, num_lo, num_hi, f_num_lo, f_num_hi )
  character(len=*), intent(in) :: word
  integer, intent(out) :: n, l, num_lo, num_hi
  real(long), intent(out) :: f_num_lo, f_num_hi
  integer :: i, imax, k, INFO
  integer :: num
  real(long) :: f_num
  character(len=*), parameter :: j_separator = ",:|/"
  !
  ! find out quantum number l
  !
  imax = len( trim( word ) )
  l = -1
  search_l: do i = 1, imax
    select case ( word(i:i) )
      case ( "s", "S" );  l = 0
      case ( "p", "P" );  l = 1
      case ( "d", "D" );  l = 2
      case ( "f", "F" );  l = 3
      case ( "g", "G" );  l = 4
      case ( "h", "H" );  l = 5
      case ( "i", "I" );  l = 6
      case ( "j", "J" );  l = 7
      case ( "k", "K" );  l = 8
      case ( "l", "L" );  l = 9
      case ( "m", "M" );  l = 10   
    end select
    if ( l /= -1 ) exit search_l
  end do search_l
  if ( l == -1 ) &
    call error_message( "Wrong configuration string or hole string!" )
  !
  ! find out quantum number n
  !
  read( word(1:i-1), *, iostat=INFO ) n
  !
  ! find out occupation (or hole) number
  !
  f_num_lo = -ONE         ! -1 if nothing is assigned
  f_num_hi = -ONE         ! -1 if nothing is assigned
  do k = i+1, imax
    if ( index( j_separator, word(k:k) ) /= 0 ) then
      if ( l == 0 ) call error_message( "s orbital does not have different j!" )
      !
      ! j-dependent occupation numbers separated by ",", ":", "|", "/"
      !   num_lo and f_num_lo : j = l - 1/2, max. occ. = 2(l-1/2)+1 = 2l
      !   num_hi and f_num_hi : j = l + 1/2, max. occ. = 2(l+1/2)+1 = 2l+2
      !
      read( word(i+1:k-1), *, iostat=INFO ) f_num_lo    ! real(long)
      read( word(k+1:   ), *, iostat=INFO ) f_num_hi    ! real(long)
    end if
  end do
  if ( f_num_lo == -ONE .and. f_num_hi == -ONE ) then
    !
    ! read occupation number (not separated by diff. j)
    !
    read( word(i+1: ), *, iostat=INFO ) f_num     ! real(long)
    if ( f_num > dble(2 * l) ) then
      f_num_lo = dble(2 * l)            ! put max. occ. for j=l-1/2
      f_num_hi = f_num - f_num_lo       ! put the rest  for j=l+1/2
    else
      f_num_lo = f_num
      f_num_hi = ZERO
    end if
  end if
  if ( f_num_lo == -ONE ) f_num_lo = ZERO
  if ( f_num_hi == -ONE ) f_num_hi = ZERO
  num_lo = nint( f_num_lo )     ! rounded up
  num_hi = nint( f_num_hi )     ! rounded up
  !
  ! checking occupation number
  !
  num = num_lo + num_hi
  f_num = f_num_lo + f_num_hi
  if ( num_lo > 2*l .or. f_num_lo > dble(2*l) ) &
    call error_message( "Wrong occ. number of lower j in " // trim(word) )
  if ( num_hi > 2*l+2 .or. f_num_hi > dble(2*l+2) ) &
    call error_message( "Wrong occ. number of higher j in " // trim(word) )
  if ( num > 4*l+2 .or. f_num > dble(4*l+2) ) &
    call error_message( "Wrong occ. number in " // trim(word) )
  !
  ! print for testing
  !
  !write(*,*) trim(word)
  !write(*,*) "- n = ", n
  !write(*,*) "- l = ", l
  !write(*,*) "- lower j  = ", num_lo, f_num_lo
  !write(*,*) "- higher j = ", num_hi, f_num_hi
  !write(*,*)
end subroutine parse_subshell

subroutine convert_n_occ_to_config_rel( n_occ, f_occ, config_rel, f_occ_rel )
  integer, intent(in) :: n_occ(:,:)
  real(long), intent(in) :: f_occ(:,:)
  integer, intent(out) :: config_rel(:,:)
  real(long), intent(out) :: f_occ_rel(:)
  integer :: N_orb, N_orb_rel
  integer :: i, k, l
  N_orb = size(n_occ,dim=1)
  N_orb_rel = count( n_occ(:,2) == 0 ) + count( n_occ(:,2) /= 0 ) * 2
  if ( N_orb_rel /= size(config_rel,dim=1) .or. &
       N_orb_rel /= size(f_occ_rel) ) then
    call error_message( "Inconsistent dim. in convert_n_occ_to_config_rel" )
  end if
  i = 0                 ! index for config_rel
  do k = 1, N_orb       ! index for n_occ
    i = i + 1
    config_rel(i,1:2) = n_occ(k,1:2)
    l = n_occ(k,2)      ! quantum number l
    if ( l == 0 ) then
      config_rel(i,3) = 1                       ! 2*j stored: j = 0 + 1/2
      config_rel(i,4) = n_occ(k,3)
      config_rel(i,5) = k
      f_occ_rel(i)    = f_occ(k,1)
    else
      config_rel(i,3) = 2*l - 1                 ! 2*j stored: j = l - 1/2
      config_rel(i,4) = n_occ(k,4)              ! occ# in lower j
      config_rel(i,5) = k
      f_occ_rel(i)    = f_occ(k,2)
      i = i + 1         ! I don't understand this convention of config_rel...
      config_rel(i,1:2) = n_occ(k,1:2)
      config_rel(i,3) = 2*l + 1                 ! 2*j stored: j = l + 1/2
      config_rel(i,4) = n_occ(k,3) - n_occ(k,4) ! occ# in higher j
      config_rel(i,5) = k
      f_occ_rel(i)    = f_occ(k,1) - f_occ(k,2)
    end if
  end do
end subroutine convert_n_occ_to_config_rel

!
! this is not used in this module, but it might be useful for other programs
!
subroutine parse_config_string_to_P( P, config_str )
  character(len=*), intent(in) :: config_str
  type(Param), intent(inout) :: P
  integer :: n_occ(P%N_orb,4)
  real(long) :: f_occ(P%N_orb,2)
  call parse_config_string( config_str, n_occ, f_occ )
  P%init_config(1:P%N_orb,1:3) = n_occ(:,1:3)
  P%init_f_occ(1:P%N_orb) = f_occ(:,1)
  if ( P%yes_relativity ) then
    P%N_orb_rel = count( n_occ(:,2) == 0 ) + count( n_occ(:,2) /= 0 ) * 2
    if ( .not. allocated( P%init_config_rel ) ) &
      allocate( P%init_config_rel(P%N_orb_rel,5) )
    if ( .not. allocated( P%init_f_occ_rel ) ) &
      allocate( P%init_f_occ_rel(P%N_orb_rel) )
    call convert_n_occ_to_config_rel( n_occ, f_occ, &
                                      P%init_config_rel, P%init_f_occ_rel )
  end if
end subroutine parse_config_string_to_P

!----------------------------------------------------------------------------
! how many subshells in config_string or from Aufbau principle with N_e
!   1s, 2s, 2p, 3s, 3p, 3d, ...
!   These are "subshells", rather than "orbitals".
!   But the variable names are sometimes N_orb instead of N_subshell, so
!   BE CAREFUL!!!
!----------------------------------------------------------------------------
function find_N_subshell( N_e, string ) result (N_subshell)
  integer, intent(in) :: N_e    ! # of electrons, used for Aufbau principle
  character(len=*), intent(in) :: string
  integer :: N_subshell
  integer :: count_s    ! how many s orbitals in string
  integer :: i, j, nl, n, l, num
  if ( string /= "" ) then
    !
    ! according to config_string (it shouldn't be hole_string )
    ! subshells are separated by "_" or " "
    !
    N_subshell = 1
    count_s = 0
    do j = 1, len( trim( string ) )
      select case ( string(j:j) )
        case ( "_", " " );  N_subshell = N_subshell + 1
        case ( "s", "S" );  count_s = count_s + 1
      end select
    end do
  else
    !
    ! Aufbau principle: how to fill up electrons into subshells
    !
    nl = 0; i = 0; num = 0
    count_s = 0
    loop: do
      nl = nl + 1       ! nl = n + l
      do n = 1, nl
        l = nl - n
        if ( l >= n ) cycle
        i = i + 1
        if ( l == 0 ) count_s = count_s + 1
        num = num + (4*l+2)
        if ( num >= N_e ) exit loop
      end do
    end do loop
    N_subshell = i
  end if
  !
  ! think out the relativistic case: because j=l-1/2 and l+1/2, 
  ! N_subshell becomes doubly counted for l > 0 and
  !                    singly counted for l = 0.
  !
  !if ( present(yes_relativity) ) then
  !  if ( yes_relativity ) N_subshell = 2 * N_subshell - count_s
  !end if
end function find_N_subshell

!----------------------------------------------------------------------------
! Aufbau principle: electrons filled up in the order of 
!                   1s, 2s, 2p, 3s, 3p, 4s, 4d, 4p, 5s, ...
!----------------------------------------------------------------------------
subroutine Aufbau_principle( N_e, n_occ )
  integer, intent(in) :: N_e
  integer, intent(out) :: n_occ(:,:)
  integer :: i, n, l, nl, num
  !
  ! if config_str is not assigned, we have to consider Hund's rule!
  ! Hey, it's not the Hund's rule...
  ! It is the Aufbau principle or Madelung's rule.
  !
  nl = 0; i = 0; num = 0
  loop: do
    nl = nl + 1
    do n = 1, nl
      l = nl - n
      if ( l >= n ) cycle
      i = i + 1
      n_occ(i,1) = n
      n_occ(i,2) = l
      n_occ(i,3) = 4*l + 2
      n_occ(i,4) = 2*l          ! max. occ# in lower j
      num = num + n_occ(i,3)
      if ( num >= N_e ) then
        n_occ(i,3) = N_e - num + n_occ(i,3)
        if ( n_occ(i,3) < 2*l ) n_occ(i,4) = n_occ(i,3)
        exit loop
      end if
    end do
  end do loop
end subroutine Aufbau_principle

!----------------------------------------------------------------------------
! Print out process information
!   name  initial_index  final_index  value  energy  # description
!   value can be a cross seciton or rate.
!   value and energy are in a.u.
!----------------------------------------------------------------------------
!subroutine print_process( P, str, value, E, config, i, j, jp )
!  type(Param), intent(in) :: P
!  character(len=*), intent(in) :: str
!  real(long), intent(in) :: value, E
!  integer, intent(in) :: config(:,:), i
!  integer, intent(in), optional :: j, jp
!  integer :: n_max, config_final(size(config,1),3)
!  integer :: i_initial, i_final
!  integer :: k, occ_n
!  character(len=1) :: tag
!  character(len=80) :: s1, s2, s3, s4
!  character(len=100) :: string
!  character(len=300) :: description
!  !
!  ! check the threshold of the value for printing
!  !
!  if ( P%print_threshold /= ZERO .and. value < P%print_threshold ) return
!  !
!  ! set up the label for the process
!  !
!  n_max = maxval( config(:,1) )                 ! max. n of the shells
!  config_final(:,:) = config(:,:)
!  select case ( str )
!    case ( "P", "photoionization", "photoabsorption" )  ! i ionized
!      tag = "P"
!      if ( config(i,1) == n_max ) tag = "V"     ! valence photoionization
!      config_final(i,3) = config_final(i,3) - 1
!    case ( "A", "auger", "Auger" )              ! j -> i, jp ionized
!      tag = "A"
!      config_final(i,3) = config_final(i,3) + 1
!      if ( present(j) )  config_final(j,3)  = config_final(j,3)  - 1
!      if ( present(jp) ) config_final(jp,3) = config_final(jp,3) - 1
!    case ( "F", "fluorescence", "fluo" )        ! j -> i
!      tag = "F"
!      config_final(i,3) = config_final(i,3) + 1
!      if ( present(j) ) config_final(j,3) = config_final(j,3) - 1
!  end select
!  !
!  ! find indexes for initial and final configurations
!  !
!  call config_to_index( config, i_initial )
!  call config_to_index( config_final, i_final )
!  !
!  ! complete the process description
!  !
!  write( s1, "(100i3)" ) config(:,3)            ! print out initial occ#
!  write( s2, "(100i3)" ) config_final(:,3)      ! print out final occ#
!  write( description, "(' # ',a15,a,' -> ',a)" ) str, trim(s1), trim(s2)
!  if ( P%print_threshold /= ZERO ) then
!    string = config_to_string( config_final )
!    description(len(trim(description))+1: ) = " | " // trim(string)
!  end if
!  !
!  ! print out into the DATA file
!  !
!  write( s3, "(2i10)" ) i_initial, i_final
!  write( s4, * ) value
!  write( P%IO_DATA, "(x,a,a20,a24,x,f10.5,a)") &
!    tag, trim(s3), s4, E, trim(description)
!end subroutine print_process

subroutine swap_index_int( a, b )
  integer, intent(inout) :: a(:), b(:)
  integer :: temp(size(a,1))
  temp(:) = a(:);  a(:) = b(:);  b(:) = temp(:)
end subroutine swap_index_int

subroutine swap_index_real( a, b )
  real(long), intent(inout) :: a(:), b(:)
  real(long) :: temp(size(a,1))
  temp(:) = a(:);  a(:) = b(:);  b(:) = temp(:)
end subroutine swap_index_real

function count_virtual_orbital( n_occ, nmax, lmax ) result ( num_vir )
  integer, intent(in) :: n_occ(:,:), nmax, lmax
  integer :: n, l, i, num_vir
  logical :: is_exist

  num_vir = 0

  ! count virtual and Rydberg orbitals in need
  do n = 1, nmax
     do l = 0, lmax
        if ( l > n - 1 ) exit
        ! scan n_occ
        is_exist = .false.
        do i = 1, size(n_occ, dim=1)
           if ( n_occ(i,1) == n .and. n_occ(i,2) == l ) then
              is_exist = .true.
              exit
           endif
        enddo
        if ( .not. is_exist ) num_vir = num_vir + 1
     enddo
  enddo
end function count_virtual_orbital

subroutine add_virtual_orbital( n_occ, f_occ, nmax, lmax, N_orb, num_vir )
  integer, intent(inout) :: n_occ(:,:)
  real(long), intent(inout) :: f_occ(:,:)
  integer, intent(in) :: nmax, lmax, N_orb, num_vir
  integer :: n, l, i, count
  logical :: is_exist

  count = 0
  do n = 1, nmax
     do l = 0, lmax
        is_exist = .false.
        if ( l > n - 1 ) exit
        do i = 1, N_orb
           if ( n_occ(i, 1) == n .and. n_occ(i, 2) == l ) then 
              is_exist = .true.
              exit
           endif
        enddo
        if ( .not. is_exist ) then
           count = count + 1
           n_occ(N_orb + count, 1) = n
           n_occ(N_orb + count, 2) = l
           n_occ(N_orb + count, 3) = 0
           n_occ(N_orb + count, 4) = 0
           
           f_occ(N_orb + count, 1) = 0
           f_occ(N_orb + count, 2) = 0
        endif
     enddo
  enddo
  if ( count /= num_vir ) stop 'add_virtual_orbital in read_param: number of virtialorbitals does not coinside with one estimated in count_virtual_orbital '
end subroutine add_virtual_orbital

subroutine help_message( P, s )
  type(Param), intent(inout) :: P
  character(len=*), optional :: s
  character(len=80) :: string
  if ( present(s) ) write(STDERR, *) "ERROR: " // s
  if ( P%calc_string .eq. "" ) P%calc_string = "Program"
  !string = trim(P%calc_string) // " options"
  string = "XATOM in a nutshell"
  write(STDERR, *) repeat( "-", len_trim(string) )
  write(STDERR, *) trim(string)
  write(STDERR, *) repeat( "-", len_trim(string) )
  write(STDERR, *)
  write(STDERR, *) "Want to calculate electronic structure"
  write(STDERR, *) "- All you need is to specify an atomic species"
  write(STDERR, *) "  ./xatom -s Ne"
  write(STDERR, *) "  ./xatom -s Ne -charge 1"
  write(STDERR, *) "  ./xatom -s Ne -config 1s1_2s2_2p6"
  write(STDERR, *) "  ./xatom -s Ne -hole 1s1"
  write(STDERR, *)
  write(STDERR, *) "Want to know photoionization cross section"
  write(STDERR, *) "- Need to specify a photon energy, let's say 2 keV"
  write(STDERR, *) "  ./xatom -s Ne -pcs -PE 2000"
  write(STDERR, *)
  write(STDERR, *) "Want to know decay processes"
  write(STDERR, *) "- Need to make a core vacancy"
  write(STDERR, *) "  ./xatom -s Ne -hole 1s1 -decay"
  write(STDERR, *) "  ./xatom -s Ne -hole 1s1 -auger"
  write(STDERR, *) "  ./xatom -s Ne -hole 1s1 -fluorescence"
  write(STDERR, *)
  write(STDERR, *) "Want to plot something (with gnuplot)"
  write(STDERR, *) "- For the orbitals (bound and continuum at 10 eV)"
  write(STDERR, *) "  ./xatom -s Ne -bound"
  write(STDERR, *) "  ./xatom -s Ne -continuum 10"
  write(STDERR, *) "- For the potential"
  write(STDERR, *) "  ./xatom -s Ne -potential"
  write(STDERR, *) "- For the photoionization cross sections as a function of the photon energy"
  write(STDERR, *) "  ./xatom -s Ne -pcs -PE 0-1000 -dE 10"
  write(STDERR, *)
  write(STDERR, *) "Troubleshooting"
  write(STDERR, *) "- More detailed output"
  write(STDERR, *) "  ./xatom -s Ne -v"
  write(STDERR, *) "- Change the mixing parameter when a SCF convergence error occurs"
  write(STDERR, *) "  ./xatom -s Ne -mixing 0.5"
  write(STDERR, *) "- Change the grid parameters in order to check numerical convergency"
  write(STDERR, *) "  ./xatom -s Ne -N 300 -rmax 100 -L 2"
  write(STDERR, *)
  stop
end subroutine help_message

subroutine sort_config( n_occ, f_occ, N_orb )
  integer, intent(inout) :: n_occ(:,:)
  real(long), intent(inout) :: f_occ(:,:)
  integer, intent(in) :: N_orb
  integer :: i, j

  do i = 1, N_orb
     do j = i+1, N_orb
        if ( n_occ(i,1) > n_occ(j,1) ) then
           call swap_index_int( n_occ(i,:), n_occ(j,:) )
           call swap_index_real( f_occ(i,:), f_occ(j,:) )
        else if ( n_occ(i,1) == n_occ(j,1) .and. &
                  n_occ(i,2) > n_occ(j,2) ) then
           call swap_index_int( n_occ(i,:), n_occ(j,:) )
           call swap_index_real( f_occ(i,:), f_occ(j,:) )
        end if
     end do
  end do
end subroutine sort_config


end module read_param
