!----------------------------------------------------------------------------
! scattering.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2015 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2015 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! HFS calculation for the atomic system using the GPS discretization
!   written by Son, Sang-Kil in Jul. 2010 at DESY
!
! This module computes coherent scattering form factors.
!   coherent scattering == elastic scattering
!   non-resonant elastic scattering --> f0
!   resonant elastic scattering --> f' + i f''
!
! For XMDYN, scd and mad output formats need to be slightly modified.
! Check if-then with "P%yes_xmdyn".
!----------------------------------------------------------------------------
module scattering
  use constants
  use elec_config
  use read_param
  use coord_system
  use gps
  use wavefunction
  use operations
  use interpolate
  use pcs
  implicit none
contains

subroutine calculate_coherent_form_factor( P, G, F )
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  !type(Grid) :: x                       ! x := cos(theta)
  real(long) :: rho(F%N), f0(0:P%N_scd+1) !, sigma(0:P%N_scd+1)
  real(long) :: Q, dQ, alpha
  !real(long) :: omega_in, theta, sigma_tot
  real(long) :: AU2BARN
  integer :: k, N, OUTPUT
  integer*8 :: ind
  OUTPUT = P%OUTPUT
  rho(:) = density( G, F%u, F%f_occ(1:F%N_orb) )
  alpha = ONE / SPEED_C_AU              ! fine-structure constant, alpha = 1/c
  AU2BARN = AU2M**2 / 1.0e-28_long      ! 1 barns = 10^-28 m^2
  N = P%N_scd                           ! # of grids
  !
  ! set up angular grids
  !   Legendre-Gauss: does not include the end points, x%r(1...N)
  !   extended-LGL: extended Legendre-Gauss-Lobatto includes the end points
  !                 x%r(0...N+1)
  !
  !x = generate_Grid( "Legendre-Gauss", N, -ONE, ONE )
!  x = generate_Grid( "extended-LGL", N, -ONE, ONE )
  !
  ! prepare the SCD file
  !
  N = P%N_scd                           ! N: # of grids in angle or Q
  if ( P%yes_xmdyn ) then
    write( P%IO_SCD, "(a,i10)" ) "scd", P%N_scd+1  ! P%IO_SCD=STDOUT for xmdyn
  else
    open( P%IO_SCD, file=P%file_scd, access="append", form="formatted" )
    call config_to_index( F%config, ind )       ! i: index of configuration
    write( P%IO_SCD, "('#',2i12)" ) ind, P%N_scd+1      ! because 0...P%N_scd
  end if
  !
  ! print out information
  !
  write(OUTPUT, "('#'/'# Elastic scattering factor (coherent form factor):')")
!  if ( P%Q_end /= ZERO ) then
!    omega_in = P%Q_end / TWO / alpha    ! reset P.E. with the maximum Q
    write(OUTPUT, "('#   momentum transfer Q = ',f6.1,' ~ ',f6.1)") &
      P%Q_start, P%Q_end
!  else
!    omega_in = P%PE_start               ! photon energy
!  end if
!  write(OUTPUT, "('#   photon energy = ',f9.3,' a.u. = ',f9.2,' eV')") &
!    omega_in, omega_in * AU2EV
  write(OUTPUT, "('#   N of grids in angular or momentum space = ',i5)") N
  write(OUTPUT, "('#')")
!  write(OUTPUT, "('#',a12,3a13)") "theta", "Q", "f0(Q)", "diff cs(au)"
  write(OUTPUT, "('#',a12,3a13)") "Q", "f0(Q)"
  !
  ! loop over theta (or momentum transfer Q)
  !
!  do k = N+1, 0, -1
!    theta = acos( x%r(k) )
!    Q = TWO * alpha * omega_in * sin( theta / TWO )
  dQ = P%Q_end / dble( P%N_scd )
  do k = 0, P%N_scd
    Q = dQ * dble(k)
    if ( Q == ZERO ) then
      f0(k) = integral( G, rho(:) )    ! # of electrons
    else
      !
      ! f0 = 4 pi * int_0^inf [ r^2 rho(r) sin( Q r ) / ( Q r ) ] dr
      !   4pi is omitted in the following expression
      !   because integral( G, f ) over the whole space (not just dr)
      !   already has a 4pi factor inside.
      !   Note that 4pi in the original expression is not from 
      !     int_0^2pi d(phi) int_0^pi sin theta d(theta).
      !
      ! SEE: Eq. (16) in Son, Young & Santra, PRA 83, 033402 (2011).
      !
      f0(k) = integral( G, rho(:) * sin( Q * F%r(:) ) / Q / F%r(:) )
    end if
!    sigma(k) = alpha**4 * f0(k)**2 * ( ONE + cos(theta)**2 ) / TWO
!    write(OUTPUT, "(3f13.8,es13.5)") theta, Q, f0(k), sigma(k)
    write(OUTPUT, "(2f13.8)") Q, f0(k)
    write( P%IO_SCD, "(f7.2,es23.15)" ) Q, f0(k)
  end do
  !
  ! compute the total cross section
  !   sigma_tot = int_0^2pi d(phi) int_0^pi d(theta) sin(theta) *
  !                 [ alpha^4 |f0(Q)|^2 ( 1+cos(theta)^2 ) / 2 ]
  !
!  sigma_tot = TWO * PI * sum( sigma(:) * x%weight(:) )
!  write(OUTPUT, "('#   Total S cross section (a.u.) = ',4x,es13.5," // &
!                "'  [ ',f13.6' barns ]')") sigma_tot, sigma_tot * AU2BARN
!  call save_scd( P, F%config, f0 )      ! print out cff data as SCD
  write(OUTPUT, "(/)")          ! print two blank lines
  if ( .not. P%yes_xmdyn ) then
    write( P%IO_SCD, "()" )
    close( P%IO_SCD )
  end if
end subroutine calculate_coherent_form_factor


!----------------------------------------------------------------------------
! calculate f' and f'' with the dispersion correction
!   f(Q,w) = f0(Q) + f'(w) + i f''(w)
!   f''(w) = - w / ( 4pi r0 c ) * sigma_P(w)
!   f'(w) = -2 / ( 4 pi^2 r0 c ) *
!           P_int_0^inf [ w'^2 sigma_P(w') / ( w'^2 - w^2 ) ] dw'
!
! f' integral is evaluated with the Gauss-Legendre quadrature
! in the forward (w to w+PE_range) and backward (w-PE_range to w) directions.
!
! i) if a range of photon energy is given,
!    sigma_P is calculated for given range and interpolated with Spline.
!    Then f' can be computed on freq. grids with interpolated sigma_P values.
! ii) if one photon energy is specified,
!     then f' can be computed on grids with calculated sigma_P at each grid.
!     This takes much more times but maybe more accurate... and no other way!
!
! For the case i), it creates a MAD file which has the following structure.
!   #1: (index of given configuration) (N of photon energies given in the range)
!   #2:                            (index after ionization) ...
!   #3: (photon energy) (f') (f'') (1s cross section) ... (nl cross section)
!   #4: (photon energy) ...
!   #N+2: (photon energy) ...
!   #N+3: (blank)
! where photon energy is in eV and cross sections are in a.u.
!----------------------------------------------------------------------------
subroutine calculate_dispersion_correction( P, F )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  type(Spline) :: SP_PE
  type(Grid) :: FG              ! frequency grids
  integer :: i, j, k, N_PE, OUTPUT
  !integer :: N_scan
  integer*8 :: ind
  real(long) :: CS(F%N_orb), CS_tot, r0
  real(long) :: omega, omega_p
  real(long) :: fp, fpp         ! f' and f''
  real(long), allocatable :: fp_scan(:)
  OUTPUT = P%OUTPUT
  N_PE = nint( ( P%PE_end - P%PE_start ) / P%PE_dE )    ! how many points
  !
  ! print information
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Resonant elastic scattering (dispersion correction):')")
  write(OUTPUT, "('#   ',a,2x,a)") &
    "f(Q,omega) = f0(Q) + f'(omega) + i f''(omega)", "where omega is P.E."
  if ( N_PE > 0 ) then
    write(OUTPUT, "('#   photon energy range (eV) = ',f10.2,' to ',f10.2)") &
      P%PE_start * AU2EV, P%PE_end * AU2EV
    write(OUTPUT, "('#   photon energy step (eV)  = ',f10.2)") P%PE_dE * AU2EV
  end if
  write(OUTPUT, "('#   Gauss-Legendre integration range (eV) = +/-',f10.2)") &
    P%PE_range * AU2EV
  write(OUTPUT, "('#   - N of grids  = ',i8)") P%N_disp
  write(OUTPUT, "('#   - L parameter = ',f8.1)") P%L_disp
  if ( P%yes_spline ) then
    write(OUTPUT, "('#   ',a)") &
      "- f'' is interpolated with Spline (less accurate)"
  else
    write(OUTPUT, "('#   ',a)") &
      "- f'' is calculated at every grid (more accurate)"
  end if
  !
  ! prepare integration grid for f'
  !
  FG = generate_Grid( "Gauss-Legendre", P%N_disp, ZERO, P%PE_range, P%L_disp )
  allocate( fp_scan(P%N_disp) )
  !
  ! Thomson scattering length: r0 = e^2 / 4 pi e0 m c^2 = 2.82E-5 Angs
  ! a.k.a. classical electron radius
  !
  r0 = ONE / SPEED_C_AU**2
  !
  ! interpolate the photoabsorption cross section for entire range
  !
  if ( P%yes_spline ) SP_PE = interpolate_PCS( P, F )
  !
  ! prepare the MAD file if there is a range of photon energies
  !
  if ( N_PE > 0 ) then
    if ( P%yes_xmdyn ) then
      write( P%IO_MAD, "(a,i10)" ) "mad", N_PE+1  ! P%IO_MAD=STDOUT for xmdyn
    else
      open( P%IO_MAD, file=P%file_mad, access="append", form="formatted" )
      call config_to_index( F%config, ind )
      write( P%IO_MAD, "('#',2i12)" ) ind, N_PE + 1     ! 0 ... N_PE
      write( P%IO_MAD, "('#',51x)", advance="no" )
      do i = 1, F%N_orb
        F%config(i,3) = F%config(i,3) - 1               ! take one electron
        call config_to_index( F%config, ind )           ! "ind" can be minus,
        write( P%IO_MAD, "(i20)", advance="no" ) ind    ! which will be ignored
        F%config(i,3) = F%config(i,3) + 1               ! back to original
      end do
      write( P%IO_MAD, "()" )
    end if
  end if
  !
  ! loop over the photon energy: from PE_start to PE_end
  !
  if ( N_PE > 0 ) write(OUTPUT, "('#'/'#',a11,100a13)") "P.E.(eV)", "f'", "f''"
  photon_energy: do k = 0, N_PE
    omega = P%PE_start + P%PE_dE * dble(k)
    !
    ! for given omega, basically we need to scan 0~infinity to compute f'
    ! here we scan PE_start-PE_range to PE_end+PE_range.
    !
    do j = 1, P%N_disp
      !
      ! integrating in the backward direction
      !
      omega_p = omega - FG%r(j)
      if ( omega_p <= ZERO ) then
        CS_tot = ZERO
        fp_scan(j) = ZERO
      else
        if ( P%yes_spline ) then
          CS_tot = Spline_interpolate( SP_PE, omega_p )
        else
          CS(:) = photoionization_CS( P, F, omega_p )
          CS_tot = sum( CS(:) )
        end if
        fp_scan(j) = CS_tot * omega_p**2 / ( omega_p**2 - omega**2 )
      end if
      !
      ! integrating in the forward direction
      !
      omega_p = omega + FG%r(j)
      if ( P%yes_spline ) then
        CS_tot = Spline_interpolate( SP_PE, omega_p )
      else
        CS(:) = photoionization_CS( P, F, omega_p )
        CS_tot = sum( CS(:) )
      end if
      fp_scan(j) = fp_scan(j) + CS_tot*omega_p**2 / ( omega_p**2 - omega**2 )
    end do
    !
    ! calculate f' by adding up for the whole range (forward + backward)
    !
    fp = -sum( fp_scan(:) * FG%weight(:) ) / PI**2 / TWO / r0 / SPEED_C_AU
    !
    ! calculate f'' at a given photon energy
    !
    CS(:) = photoionization_CS( P, F, omega )
    fpp = - omega / FOUR / PI / r0 / SPEED_C_AU * sum( CS(:) )
    !
    ! print out f'(omega) and f''(omega)
    ! and save the ".mad" file inclnuding f', f'', and all PCS data
    !
    if ( N_PE > 0 ) then
      write(OUTPUT, "(f12.2,2es13.5)") omega * AU2EV, fp, fpp
      write( P%IO_MAD, "(f12.2,100es20.12)" ) omega * AU2EV, fp, fpp, CS(:)
    else
      write(OUTPUT, "('#   photon energy = ',f9.3,' a.u. = ',f9.2,' eV')") &
        P%PE_start, P%PE_start * AU2EV
      write(OUTPUT, "('#             ',a3,' = ',es13.5)") "f'", fp
      write(OUTPUT, "('#             ',a3,' = ',es13.5)") "f''", fpp
    end if
  end do photon_energy
  !
  ! close the MAD file
  !
  if ( N_PE > 0 .and. .not. P%yes_xmdyn ) then
    write( P%IO_MAD, "()" )
    close( P%IO_MAD )
  end if
end subroutine calculate_dispersion_correction

function interpolate_PCS( P, F ) result (S)
  type(Spline) :: S
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  integer, parameter :: Nmax = 50000
  integer :: i, j, k, n
  real(long), save :: omega(Nmax), sigma(Nmax)
  real(long) :: PE, PE_max
  real(long) :: CS(F%N_orb), IE(F%N_orb), temp
  !
  ! first, calculate photoabsorption cross sections
  ! PE are scanned from PE_start - PE_range to PE_end + PE_range with PE_dE,
  ! and then a cross section will be interpolated at a certain photon energy.
  !
  !omega_min = P%PE_start - nint( P%PE_range / P%PE_dE ) * P%PE_dE
  !if ( omega_min < ZERO ) &
  !  omega_min = P%PE_start - nint( P%PE_start / P%PE_dE ) * P%PE_dE
  !N_scan = nint( ( P%PE_end + P%PE_range - omega_min ) / P%PE_dE )
  !allocate( sigma_scan(N_scan+1), omega_scan(N_scan+1) )
  !allocate( CS_PE(0:N_PE,F%N_orb) )   ! store PCS data to be printed out
  !j = nint( ( P%PE_start - omega_min ) / P%PE_dE )
  !do k = 0, N_scan
  !  omega = omega_min + P%PE_dE * dble(k)
  !  CS(:) = photoionization_CS( P, F, omega )
  !  omega_scan(k+1) = omega
  !  sigma_scan(k+1) = sum( CS(:) )
  !  if ( k >= j .and. k <= j + N_PE ) CS_PE(k-j,:) = CS(:)
  !end do
  !SP_PE = Spline_create( omega_scan(:), sigma_scan(:) )
  !
  ! sort IE(:) as an ascending order
  !
  do i = 1, F%N_orb
    IE(i) = ionization_E( F, i )
  end do
  do i = 1, F%N_orb
    do j = i+1, F%N_orb
      if ( IE(j) < IE(i) ) then
        temp = IE(i); IE(i) = IE(j); IE(j) = temp
      end if
    end do
  end do
  !
  ! scan photon energy (PE) from 0 to infinity (PE_max)
  !
  PE = ZERO
  PE_max = P%PE_end + P%PE_range
  n = 0
  k = 1
  PE_scan: do
    !
    ! choose photon energy step
    ! - near ionization energy: dE * 0.1
    ! - in general: dE
    ! - after 1s ionization energy: dE for 100 more steps
    !                               dE*5 for 100 more steps
    !                               dE*10 afterward
    !
    if ( k > F%N_orb ) then
      if ( PE < IE(F%N_orb) + P%PE_dE * 100.0_long ) then
        PE = PE + P%PE_dE
      else if ( PE < IE(F%N_orb) + P%PE_dE * 600.0_long ) then
        PE = PE + P%PE_dE * 5.0_long
      else
        PE = PE + P%PE_dE * 10.0_long
      end if
    else if ( abs( PE - IE(k) ) < P%PE_dE ) then
      PE = PE + P%PE_dE * 0.1_long
      if ( PE > IE(k) + P%PE_dE ) k = k + 1
    else
      PE = PE + P%PE_dE
    end if
    !
    ! exit condition: 
    !
    if ( PE > PE_max ) exit PE_scan
    if ( n >= Nmax ) stop "Too many PCS points in interpolation"
    !
    ! calculate PCS
    !
    n = n + 1
    CS(:) = photoionization_CS( P, F, PE )
    omega(n) = PE
    sigma(n) = sum( CS(:) )
  end do PE_scan
  !
  ! costruct Spline
  !
  S = Spline_create( omega(1:n), sigma(1:n) )
end function interpolate_PCS

end module scattering
