!----------------------------------------------------------------------------
! scf.f90
!----------------------------------------------------------------------------
! Copyright (C) 2004-2005 Sang-Kil Son
! Copyright (C) 2010-2023 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2023 Sang-Kil Son <sangkil.son@cfel.de>
!          2011      Robert Thiele
!          2015-2016 Koudai Toyota
!          2019-2020 Rui Jin
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! HFS calculation for the atomic system using the GPS discretization
!   written by Son, Sang-Kil in Jul. 2010 at DESY
!   (modified from DFT_atom, originally written by Sang-Kil in 2004-2005)
!   modified by Son, Sang-Kil and Thiele, Robert in Dec. 2011 at DESY
!
! This module contains the core part of DFT calculation, i.e. SCF procedure.
! KS orbitals are calculated self-consistently by means of KS equation
! with xc functional which is specified in P.  After SCF procedure, all
! information of densities, energies, and potentials are stored in S.
!----------------------------------------------------------------------------
module scf
  use constants
  use read_param
  use coord_system
  use functional_xc
  use operations
  use wavefunction
  use plasma
  use coulomb
  implicit none
contains

subroutine SCF_procedure( P, G, F )
  type(Param), intent(in) :: P
  type(Coordinate), intent(inout) :: G
  type(WF_data), intent(inout) :: F
  real(long) :: E, E_old, E_Ts, E_J, E_x, E_c, E_ext, E_sum_orb, E_sum_orb_old
  real(long) :: E_xc_int, E_Coulomb, E_pot, E_Latter, E_1e
  real(long) :: E1_tb, E_tb_J, E_tb_K
  real(long), dimension(G%N) :: rho, rho_old, rho_bound, rho_free, g_i
  !real(long), dimension(G%N) :: rho_outside
  real(long), dimension(G%N) :: v_eff, v_xc, v_J, v_ext, v_Coul, v_old
  real(long), dimension(G%N) :: v_xc_bound, v_xc_free
  real(long) :: alpha   ! SCF mixing parameter
  real(long) :: v_temp(2)
!  real(long), dimension(P%SCF_max_iter,G%N) :: v_reserv
!  integer, parameter :: N_prev = 100
!  real(long), dimension(G%N,N_prev) :: v_prev
!  real(long), dimension(N_prev) :: error_prev
  real(long) :: elec_dens, value, norm
  !real(long) :: adj_ion_dens
  real(long) :: n_elec  ! # of elec. calculated by numerical integral of density
  integer :: i, j, k, l, j_s, OUTPUT
  character(len=80) :: string
  !
  ! The followings are for plasma calculation 
  ! with environmental potential obtained from XPOT
  ! VL_fixed_pot is used only when P%yes_fixed_pot is on
  !
  real(long), dimension(G%N) :: v_env   ! external environment potential
  real(long) :: VL_fixed_pot            ! external fixed potential value
  !
  ! Initialize arrays and variables
  !
  OUTPUT = P%OUTPUT
  j_s = G%N             ! if r_s is not assigned, j_s is just the end of grids
  rho_free(:) = ZERO    ! free-electron density (or external electron density)
  v_J(:) = ZERO
  v_xc(:) = ZERO
  alpha = P%SCF_mixing  ! to be varied during SCF if P%yes_adaptive_mixing is on
  !
  ! if fractional occ. numbers are used, F%f_occ is given from P
  ! otherwise, F%f_occ(:) is the same as F%config(:,3)
  !
  if ( .not. P%yes_fractional ) F%f_occ(:) = F%config(:,3)
  !
  ! define the external potential (== nuclear contribution to the potential)
  !
  if ( P%yes_uniform_ion_dens ) then
    !
    ! v_ext(r) = -INT{ dens/|r-r'| } d^3r' 
    !          = -INT{ dens/r_> } d^3r' where r_> is max(r,r')
    !          = -dens*4pi*( INT_0^r{1/r}r'^2 dr' + INT_r^rmax{1/r'}r'^2 dr' )
    !          = -dens*4pi*( rmax^2 / 2 - r^2 / 6 )
    !   where dens is a uniform ion density
    !         dens = P%ion_dens * Z because charge Z is uniformly distributed
    !
    ! or simply use potential_J()
    !   rho(:) = -P%ion_dens * P%Z; v_ext(:) = potential_J( G, rho(1:G%N) )
    !
    v_ext(:) = -P%ion_dens*P%Z*( G%rmax**2/TWO - G%r(:)**2/6.0_long ) * FOUR*PI
  else if ( P%yes_turn_off_V_ne ) then
    v_ext(:) = ZERO                     ! no interaction between nuc and elec
  else
    v_ext(:) = potential_ext( G, P%Z )
  end if
  !
  ! considering screening effects
  !
  if ( P%yes_screening ) then
    !
    ! set up the radial index (j_s) for the Wigner-Seitz radius (r_s)
    ! and adjust elec_dens accordingly
    !
    elec_dens = P%elec_dens
    if ( P%yes_wigner_seitz ) call determine_WS( G, P%r_s, j_s, elec_dens )
    F%j_s = j_s
    !
    ! set up N_elec_rs: # of electrons within r_s
    !
    F%N_elec_rs = dble(P%N_elec)
    if ( P%yes_average_atom .and. P%yes_quasineutral ) F%N_elec_rs = dble(P%Z)
    if ( P%yes_ion_screening ) &        ! including all elec. in r_s + N_elec
      F%N_elec_rs = F%N_elec_rs + elec_dens * FOUR / THREE * PI * P%r_s**3
    !
    ! Debye screening for v_ext(:)
    !
    if ( P%yes_debye_screening ) &
      call Debye_screening( v_ext(:), G%r(:), P%debye_length )
    !
    ! ion screening is included into v_ext(:)
    !
    if ( P%yes_ion_screening ) then
      g_i(:) = pair_correlation( G%r(:), P%pcf(:) )
      rho(:) = -P%Z_eff * P%ion_dens * g_i(:)
      if ( P%yes_debye_screening ) then
        v_ext(:) = v_ext(:) + &
                   screened_potential_J( G, rho(1:j_s), P%debye_length )
      else
        v_ext(:) = v_ext(:) + potential_J( G, rho(1:j_s) )
      end if
    end if
    !
    ! load a free-electron density from the previous calc or an external file
    !
    if ( F%yes_reuse_rho_free ) then
      rho_free(:) = F%rho_free(:)
    else if ( P%yes_load_den ) then
      if ( .not. P%yes_wigner_seitz ) &
        call error_message( "-load_dens should be used with r_s." )
      call load_free_elec( P%file_den, rho_free )
    end if
    !
    ! In PRA 86, 033411 (2012), we include P%elec_dens in v_J only, not in v_xc.
    ! However, it can be included in both v_J and v_xc, controlled by yes_flag
    !
    if ( P%yes_n_elec_screening ) rho_free(:) = elec_dens
    !
    ! scaling factor: modify rho_free in order to satisfy the neutrality
    !
    if ( P%yes_scaling ) then
      norm = integral( G, rho_free(1:j_s) )
      rho_free(:) = rho_free(:) / norm * dble( P%Z - P%N_elec )
    end if
  end if
  !
  ! print out all plasm-related options
  !
  if ( .not. P%yes_silent .and. ( P%yes_screening .or. P%yes_xpot ) ) &
    call print_atom_in_plasma( P, G, g_i, elec_dens, G%r(j_s) )
  !
  ! load an external env. potential from XPOT if available
  !
  if ( P%yes_load_env ) call load_env_pot( P, G, v_env(:), VL_fixed_pot )
  !
  ! prepare an initial potential to obtain an initial guess density
  !
  if ( F%yes_reuse_pot ) then
    v_eff(:) = F%pot(:)         ! previous calculation, F%pot(:), if available
  else if ( P%yes_load_pot ) then
    call load_potential( P%file_pot, v_eff(:) )         ! loading from file
  else
    v_eff(:) = v_ext(:)                                 ! use 1-e model
    !
    ! add potential due to the free electrons (when P%yes_screening is on)
    !
    if ( P%yes_screening ) then
      v_J(:) = potential_J( G, rho_free(1:j_s) )
      v_xc_free(:) = potential_xc( P, G, rho_free(:) )
      if ( P%yes_thermal ) then
        v_xc_free(:) = v_xc_free(:) * &
                       thermal_reduction_factor( rho_free, P%T_eV )
      end if
      v_eff(:) = v_eff(:) + v_J(:) + v_xc_free(:)
    end if
    !
    ! make a flat potential tail depending on several options
    ! for XPOT calcualtion, there are three cases:
    ! - CASE 1: apply muffin-tin approx. according to wigner-seitz radius
    ! - CASE 2: use a fixed value specified in file_env
    ! - CASE 3: nothing for r<=rmax, but put a constant of V(r=rmax) for r>rmax 
    !
    if ( F%yes_flat_tail ) then
      if ( P%yes_load_env ) v_eff(:) = v_eff(:) + v_env(:)
      if ( P%yes_muffin_tin .or. &                              ! muffin-tin
           P%flat_tail_type == 1 ) v_eff(j_s+1:) = v_eff(j_s)   ! CASE 1 (XPOT)
      if ( P%yes_fixed_pot ) then ! or P%flat_tail_type == 2  ==> CASE 2 (XPOT)
        call apply_fixed_pot( G, v_eff, VL_fixed_pot, j_s )
      end if
    end if
  end if
  F%e_s = v_eff(G%N)    ! preparing for CASE 3 (XPOT) and other plasma calc.
  v_old(:) = v_eff(:)
  !
  ! guess an initial density
  ! if you want alternative initial values, use these
  !   u(:,1) = exp( -P%Z * Grid_r(G) )
  !
  if ( P%yes_average_atom .and. F%yes_reuse_pot ) then
    !
    ! if BOTH average atom AND reuse_pot options are on,
    ! then solve_SE_finite_temp with the given potential
    !
    call solve_SE_finite_temp( F, G, v_eff, P%T_eV )
    call construct_density_finite_temp( F, rho_bound, rho_free, &
                                        P%yes_fixed_config )
    E_sum_orb = sum( F%e_all(:) * F%n_occ(:) * F%norm(:) )
  else
    !
    ! even though average-atom calculation, solve_SE for the initial guess
    !
    call solve_SE( F, G, v_eff )
    E_sum_orb = sum( F%e_orb(1:F%N_orb) * F%f_occ(1:F%N_orb) )
    rho_bound(:) = density( G, F%u, F%f_occ(1:F%N_orb) )
    !
    ! if the chemical potential is NOT fixed, then
    ! choose highest-occupied-atomic-orbital as the chemical potential
    ! HOAO... sounds like Hawaiian!
    !
    if ( .not. F%yes_fix_chem_pot ) F%chem_pot = maxval( F%e_orb(1:F%N_orb) )
  end if
  rho(:) = rho_bound(:) + rho_free(:)
!  if ( P%yes_screening ) rho_outside(:) = calculate_rho_outside( G, rho, j_s )
  !
  ! prepare Latter correction: including all electrons within r_s
  !   pot(:) ~ -( Z - n_elec + 1 ) / r  for r -> infinity
  ! here, n_elec is calculated by a numerical integral of elec. density,
  ! which is the safest way to define # of electrons in the system of interest.
  !
  ! assign F%q that is the effective charge in the asymptotic region of r
  !   F%q = - r * pot           as r -> rmax
  !       ~ ( Z - n_elec + 1 )  when self-interaction correction is 
  !                                  properly considered
  ! F%q is used when pot. on uniform grid is prepared in interpolate_potential()
  ! It has to be reassigned during MC-on-the-fly because n_elec is changing.
  ! So, it is calculated here, rather than create_WF() in wavefunciton.f90
  !
  ! NOTE: With muffin-tin pot, F%q should be re-defined.
  !       It should be done at the end of the scf procedure.
  !
  n_elec = integral( G, rho(1:j_s) )    ! integral for r < r_s
  F%q = dble( P%Z - n_elec )
  if ( P%yes_Latter .or. P%yes_OEP_KLI ) F%q = F%q + ONE
  v_Coul(:) = - F%q / G%r(:)    ! to be used for the Latter correction only
  !v_Coul(:) = - ( dble(P%Z + 1) - n_elec ) / G%r(:)
  if ( P%yes_debye_screening ) then
    call Debye_screening( v_Coul(:), G%r(:), P%debye_length )
    F%q = F%q * exp( -F%r_int(F%N+2) / P%debye_length )
  end if
  !
  ! initialization for SCF procedure
  !
  F%yes_converged = .false.
  E_old = ZERO
  E_sum_orb_old = ZERO
  rho_old(:) = rho(:)
  if ( P%yes_verbose ) then
    write(OUTPUT, "('# ',a4,4a10,a14,a10)", advance="no") &
      "iter", "E(orb)", "E(exch)", "E(Coul)", "E(J)*", "E(total)", "diff."
    if ( P%yes_adaptive_mixing ) write(OUTPUT, "(a10)", advance="no") "mixing"
    write(OUTPUT, "()")
  end if
  ! ===========================================================================
  ! SCF (self-consistence field) procedure
  !   0. guess an initial density (in the above)
  !   1. get a potential
  !   2. modify the potential if necessary (e.g. Latter or muffin-tin)
  !   3. calculate energies
  !   4. check the SCF convergence; if it's converged, exit SCF
  !   5. solve the Schroedinger equation
  !   6. modify the density if necessary (e.g. free-electron density)
  !   7. repeat 1-6 until satisfying 4
  ! ===========================================================================
  SCF: do i = 1, P%SCF_max_iter
    !
    ! 1. get a new potential
    !
    if ( P%yes_debye_screening ) then           ! debye screening
      v_J(:)  = screened_potential_J( G, rho(1:j_s), P%debye_length )
      v_xc(:) = screened_potential_xc( rho(:), P%debye_length )
      v_xc_bound(:) = screened_potential_xc( rho_bound(:), P%debye_length )
      v_xc_free(:)  = screened_potential_xc( rho_free(:), P%debye_length )
    else
      v_J(:)  = potential_J( G, rho(1:j_s) )
      v_xc(:) = potential_xc( P, G, rho(:) )
      v_xc_bound(:) = potential_xc( P, G, rho_bound(:) )
      v_xc_free(:)  = potential_xc( P, G, rho_free(:) )
    end if
    if ( P%yes_exact_pot ) then
      !
      ! exact potential for one- or two-elec. system
      ! for one-elec., v_J = 0 and v_xc = 0
      ! for two-elec., v_J = same and v_xc = - v_J / 2 (exact HF)
      !                or v_J --> v_J / 2 and v_xc = 0 <== No, it gives wrong E
      ! v_J and v_xc from free-electrons are added if necessary
      !
      if ( P%N_elec == 1 ) then
        v_J(:) = ZERO
        v_xc_bound(:) = ZERO
        v_xc_free(:)  = ZERO
      else if ( P%N_elec == 2 ) then
        v_J(:) = potential_J( G, rho_bound(1:j_s) )
        v_xc_bound(:) = -HALF * v_J(:)
        v_xc_free(:)  = ZERO
      end if
      if ( P%yes_screening ) then
        v_J(:) = v_J(:) + potential_J( G, rho_free(1:j_s) )
        v_xc_free(:) = potential_xc( P, G, rho_free(:) )
      end if
      v_xc(:) = v_xc_bound(:) + v_xc_free(:)
    end if
    if ( P%yes_thermal ) then
      if ( P%yes_flag ) then            ! separate v_xc_bound and v_xc_free
        call potential_x_finite_temp( v_xc_free(:), rho_free(:), &
                                      P%T_eV, P%thermal_x_type )
        v_xc(:) = v_xc_bound(:) + v_xc_free(:)
      else
        call potential_x_finite_temp( v_xc(:), rho(:), &
                                      P%T_eV, P%thermal_x_type )
      end if
    end if
!    if ( P%yes_muffin_tin ) &
!      v_J(:) = v_J(:) + integral( G, rho_outside(:) / G%r(:) )
    if ( P%yes_turn_off_V_ee ) then
      v_J(:) = ZERO
      v_xc(:) = ZERO
    end if
    v_eff(:) = v_ext(:) + v_J(:) + v_xc(:)
    !
    ! 2. modify the potential (muffin-tin approximation or Latter correction)
    !    FROM RUI: for plasma environmental calculation, do Latter correction 
    !      for the atom part potential and add environmental potential v_env
    !      --> we'd better turn off the Latter correction when plasma
    !          calculation is performed; OR it shouldn't matter anyway
    !
    if ( P%yes_Latter ) call Latter_correction( v_eff, v_Coul )
    if ( F%yes_flat_tail ) then
      if ( P%yes_load_env ) v_eff(:) = v_eff(:) + v_env(:)
      if ( P%yes_muffin_tin .or. &
           P%flat_tail_type == 2 ) v_eff(j_s+1:) = v_eff(j_s)
      if ( P%yes_fixed_pot ) call apply_fixed_pot( G, v_eff, VL_fixed_pot, j_s )
    end if
    F%e_s = v_eff(G%N)
    !
    ! 3. calculate energies
    !  - E_xc_int = INT{ v_xc(r) * rho(r) }dr
    !
    !  - total energy
    !      E = SUM{e_orbital} - J[rho] + E_xc[rho] - INT{ v_xc(r) * rho(r) }dr
    !        where SUM{e_orbital} = T_s[rho] + INT{ v_eff(r) rho(r) }dr
    !              J[rho] = 1/2 * INT{ rho(r) * rho(r') / |r-r'| }dr dr'
    !      or
    !      E = Ts[rho] + J[rho] + E_xc[rho] + INT{ v_ext(r) * rho(r) }dr
    !    Two diff. calculations are identical.
    !    [ACHTUNG] Be careful with the sign of v_xc and E_xc in the above eqs.
    !              In this code, v_xc and E_xc already contain the minus sign.
    !    [ACHTUNG] the potential modification due to the Latter correction
    !              is not included!!!
    !
    !  - To compare my program with Tong's program, he defined E_coulomb 
    !    like INT{ rho * v }dr, that corresponds to -( E_J + E_xc_int ).
    !
    E_J = functional_J( G, rho(1:j_s), v_J(:) )
    E_x = functional_x( P, G, rho(1:j_s) )
    E_c = functional_c( P, G, rho(1:j_s) )
    E_xc_int = integral( G, rho(1:j_s) * v_xc(1:j_s) )
    if ( P%yes_turn_off_V_ee ) then
      E_J = ZERO
      E_x = ZERO
      E_c = ZERO
      E_xc_int = ZERO
    end if
    !E_sum_orb = sum( F%e_orb(:) * F%config(:,3) )      ! calculated from 5.
    if ( P%yes_exact_pot ) then
      if ( P%N_elec == 1 ) then
        E_x = ZERO
        E_xc_int = ZERO
      else if ( P%N_elec == 2 ) then
        E_x = -HALF * E_J
        !E_xc_int = -E_J  ! calculated with v_xc in the above (v_xc = -v_J/2)
      end if
    end if
    E = E_sum_orb - E_J + ( E_x + E_c ) - E_xc_int
    E_Coulomb = - E_J - E_xc_int
    !
    ! 4. check SCF convegence and exit
    !
    if ( P%N_elec == 1 .and. P%yes_exact_pot ) exit SCF
    if ( P%yes_verbose ) then
      write(OUTPUT, "('# ',i4,4f10.4,f14.7,es10.2)", advance="no") &
        i, E_sum_orb, E_x, E_Coulomb, E_J, E, E_old - E
      if ( P%yes_adaptive_mixing .and. i > 1 ) &
        write(OUTPUT, "(f10.7)", advance="no") alpha
      write(OUTPUT, "()")
    end if
    if ( i >= P%SCF_max_iter ) exit SCF
    if ( P%yes_xpot ) then
      !
      ! for XPOT calculation, we apply a reduced condition:
      ! check only the zeroth energy (E_sum_orb), rather than
      ! the total energy (E) and electron density (rho)
      !
      if ( check_convergence( P, E_sum_orb, E_sum_orb_old ) ) exit SCF
    else
      if ( check_convergence( P, E, E_old, rho, rho_old ) ) exit SCF
    endif
    !
    ! SCF acceleration: linear mixing with newly-calcualted potential, v_eff(:)
    !                                  and previously-used potential, v_old(:)
    !   alpha is evaluated based on v_eff & v_old if P%yes_adaptive_mixing is on
    !   alpha = P%SCF_mixing                      otherwise
    !
    if ( i == 1 ) v_old(:) = v_eff(:) ! 100% new v_eff for the first iteration
    if ( P%yes_adaptive_mixing ) &
      alpha = adaptive_mixing_parameter( v_eff(:), v_old(:) )
    !call potential_mixing( v_eff(:), v_prev(:,:), error_prev(:), i, v_old(:) )
    v_eff(:) = v_eff(:) * alpha + v_old(:) * ( ONE - alpha )
    !
    ! set up "old" quantities: energy E, density rho, potential v
    ! "old" means something calcualted in the previous SCF iteration
    !
    E_old = E
    E_sum_orb_old = E_sum_orb
    rho_old(:) = rho(:)
    v_old(:) = v_eff(:)
    !
    ! 5. solve the Schroedinger equation
    !
    if ( P%yes_average_atom ) then      ! finite-temperature HFS
      call solve_SE_finite_temp( F, G, v_eff, P%T_eV )
      call construct_density_finite_temp( F, rho_bound, rho_free, &
                                          P%yes_fixed_config )
      E_sum_orb = sum( F%e_all(:) * F%n_occ(:) * F%norm(:) )
    else                                ! regular HFS
      call solve_SE( F, G, v_eff )
      E_sum_orb = sum( F%e_orb(1:F%N_orb) * F%f_occ(1:F%N_orb) )
      rho_bound(:) = density( G, F%u, F%f_occ(1:F%N_orb) )
      F%chem_pot = maxval( F%e_orb(1:F%N_orb) ) ! choose HOAO
    end if
    !
    ! 6. add up the external electron density
    !
    if ( P%yes_screening .and. P%yes_scaling ) then
      value = integral( G, rho_bound(1:j_s) )
      if ( value /= dble(P%N_elec) ) then
        norm = integral( G, rho_free(1:j_s) )
        rho_free(:) = rho_free(:) / norm * ( dble(P%Z) - value )
      end if
    end if
    rho(:) = rho_bound(:) + rho_free(:)
!    if ( P%yes_screening ) &
!      rho_outside(:) = calculate_rho_outside( G, rho, j_s )
  end do SCF
  ! ===========================================================================
  ! the end of the SCF procedure
  ! ===========================================================================
  if ( P%SCF_max_iter <= 1 ) F%yes_converged = .true. ! max_iter=0 or 1: no SCF
  if ( i < P%SCF_max_iter ) F%yes_converged = .true.
  F%E_tot = E
  F%E0 = E_sum_orb
  F%yes_reuse_pot = .true.      ! this pot(:) can be used later on
  if ( P%yes_screening ) then
    F%yes_reuse_rho_free = .true.  ! calculated rho_free(:) can be used later on
    F%rho_free(:) = rho_free(:)
  end if
  F%N_iter = i                  ! save # of iterations
  F%pot(:) = v_old(:)           ! save pot(:) for other usages
                                ! here, v_old(:) instead of v_eff(:)
                                ! to keep the same potential used in solve_SE()
  if ( F%yes_flat_tail ) then
    !
    ! if a flat tail is assumed (either P%yes_muffin_tin or P%yes_xpot), then
    ! put -pot*r at r=F%r_int(F%N+2); pot=const.
    !
    F%e_s = F%pot(F%N)          ! put an asymptotic value of the potential tail
    F%q = -F%pot(F%N) * F%r_int(F%N+2)
  end if
  if ( P%yes_save_pot ) call save_potential( P%file_pot, F%pot(:) )
  if ( P%yes_average_atom ) call save_free_elec( G, P%file_den, rho_free, j_s )
  if ( P%yes_silent ) return
  !
  ! Print out SCF convergency information
  !
  if ( P%yes_verbose ) then
    if ( P%yes_adaptive_mixing ) then
      write(OUTPUT, "('# ',a)") repeat( "-", 78 )
    else
      write(OUTPUT, "('# ',a)") repeat( "-", 68 )
    end if
    write(OUTPUT, "('#     ',4f10.4,f14.7,a10)", advance="no") &
      E_sum_orb, E_x, E_Coulomb, E_J, E
    if ( F%yes_converged ) then
      write(OUTPUT, "(a10)") "converged"
    else
      write(OUTPUT, "(a10)") "not conv."
    end if
    write(OUTPUT, "('#')")
  end if
  string = ""
  if ( i > 1 ) then
    if ( F%yes_converged ) then
      write(OUTPUT, "('# SCF converged after ',i4,' iterations')") i
      if ( P%yes_load_pot ) &
        write(OUTPUT, "('#   using the potential file: ',a)") trim(P%file_pot)
      string = ""
    else
      write(OUTPUT, "('# SCF NOT converged after ',i4,' iterations!!!')") i
      string = "*** WARNING: NOT CONVERGED ***"
    end if
    if ( P%yes_xpot ) then      ! only sum of orbital energies is checked
      write(OUTPUT, "('#   convergence (sum of orbital energies) = ',es10.3)") &
        E_sum_orb_old - E_sum_orb
    else                        ! energy AND density are checked
      write(OUTPUT, "('#   energy convergence  = ',es10.3)") E_old - E
      write(OUTPUT, "('#   density convergence = ',es10.3)") &
        maxval( abs( rho_old(:) - rho(:) ) )
    end if
  else
    write(OUTPUT, "('# No SCF performed: N of iterations = ',i5)") F%N_iter
  end if
  write(OUTPUT, "('#')")
  !
  ! print energies
  !   E(0) = zeroth-order energy = sum of occupied orbital energies
  !                                or sum of e_orb * occ# * norm within r_s
  !          It corresponds to one-elec E with the mean field.
  !          E_sum_orb = SUM_of_E_kin + E_pot
  !          E_pot = 2*E_J + E_xc_int + E_ext + E_Latter
  !                                             ~~~~~~~~ <- added due to Latter
  !   E(1) = first-order energy = including first-order correction
  !                               thru the Slater exchange functional
  !          E = E_sum_orb - E_J + ( E_x + E_c ) - E_xc_int
  !          ==> E = SUM_of_E_kin + E_J + E_ext + E_x + E_c + E_Latter
  !          HERE E_Latter is supposed to be taken out!
  !          BUT we just have used E = E_sum_orb - E_J + ...
  !
  !   E(1)_tb = first-order-energy with explicit two-body interaction
  !          E = SUM_of_E_kin + E_ext + E_J + E_K
  !              E_J =  1/2 SUM_ij v_ijij occ_n_i occ_n_j <== same as the above
  !              E_K = -1/2 SUM_ij v_ijji occ_n_i occ_n_j
  !          NOTE that there is no E_Latter contribution
  !
  ! well, one might consider three versions of E_pot
  !   1) E_pot = integral( G, rho(:) * ( v_J(:) + v_xc(:) + v_ext(:) ) )
  !   2) E_pot = integral( G, rho(:) * v_eff(:) )
  !   3) E_pot = integral( G, rho(:) * F%pot(:) )  <-- most reasonable!
  ! F%pot is assigned with v_old to keep the same potential used in SE,
  ! F%pot is supposed to be the same as v_eff, if SCF is fully converged.
  ! In 3 (and 2), the Latter correction and the env. potential are implemented,
  ! while 1 doesn't have such correction or addition.
  !
  if ( P%yes_screening ) string = trim(string) // "++ screening effect ++"
  E_ext = integral( G, rho(1:j_s) * v_ext(1:j_s) )
  E_pot = integral( G, rho(1:j_s) * F%pot(1:j_s) )  ! F%pot instead of v_eff
  E_Ts = E_sum_orb - E_pot      ! equivalent to sum_i <psi_i|Ts|psi_i> n_i
  E_1e = E_Ts + E_ext
  write(OUTPUT, "('# Energies: ', a)") trim(string)
  write(OUTPUT, "('# - zeroth-order energy (a.u.):')")
  write(OUTPUT, "('#   SUM of E_orb   = ',f16.8,'  [ ',f12.2,' eV ]')") &
    E_sum_orb, E_sum_orb * AU2EV
  write(OUTPUT, "('#   (SUM of E_kin) = ',f16.8)") E_Ts
  write(OUTPUT, "('#   (Effec. pot E) = ',f16.8)") E_pot
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# - first-order energy (a.u.): ',a)") "with DFT formula"
  write(OUTPUT, "('#   Kinetic E      = ',f16.8)") E_Ts
  if ( P%yes_Latter ) then
    E_Latter = E - E_Ts - E_J - E_x - E_c - E_ext
    write(OUTPUT, "('#   E Latter corr  = ',f16.8)") E_Latter
    E_1e = E_1e + E_Latter
  end if
  write(OUTPUT, "('#   Nuc-Elec pot E = ',f16.8)", advance="no") E_ext
  write(OUTPUT, "('    One-Elec E = ',f16.8)") E_1e
  if ( P%N_elec > 1 .or. P%yes_screening .or. .not. P%yes_exact_pot ) then
    write(OUTPUT, "('#')")
    write(OUTPUT, "('#   Hartree E      = ',f16.8)") E_J
    write(OUTPUT, "('#   Exchange E     = ',f16.8)") E_x
    write(OUTPUT, "('#   Correlation E  = ',f16.8)", advance="no") E_c
    write(OUTPUT, "('    Two-Elec E = ',f16.8)") E_J + E_x + E_c
  end if
  write(OUTPUT, "('#   -----------------',a)") repeat( "-", 16 )
  write(OUTPUT, "('#   Total Energy   = ',f16.8,'  [ ',f12.2,' eV ]')") &
    E, E * AU2EV
  write(OUTPUT, "('#')")
  if ( P%yes_twobody ) then             ! prit out E with two-body interaction
    if ( P%yes_screening ) stop         ! it's not ready yet
    write(OUTPUT, "('# - first-order energy (a.u.): ',a)") &
      "with explicit two-body interactions"
!TEST (this way E_tb_J is a slightly different from E_J)
! Sang-Kil: NEED TO RECHECK this part (and coulomb.f90)
  E_tb_J = ZERO
  E_tb_K = ZERO
  do i = 1, F%N_orb
    do j = 1, F%N_orb
      v_temp(:) = v_twobody( F, i, j )
      E_tb_J = E_tb_J + v_temp(1)
      E_tb_K = E_tb_K + v_temp(2)
      !E_tb_J = E_tb_J + v_temp(1) * F%config(i,3) * F%config(j,3)
      !E_tb_K = E_tb_K + v_temp(2) * F%config(i,3) * F%config(j,3)
    end do
  end do
  E_tb_J =  HALF * E_tb_J
  E_tb_K = -HALF * E_tb_K
!TEST (this way E_tb_J is equal to the Hartree E)
!    E_tb_K = exchange_E_twobody( P, F )
!    E_tb_J = E_J       ! it must be indentical
!TEST
    E1_tb = E_Ts + E_ext + E_tb_J + E_tb_K
    write(OUTPUT, "('#   Kinetic E      = ',f16.8)") E_Ts
    write(OUTPUT, "('#   Nuc-Elec pot E = ',f16.8)") E_ext
    write(OUTPUT, "('#   E_J (two-body) = ',f16.8,4x,a)") E_tb_J, &
      "<== may differ from Hartree E"
    write(OUTPUT, "('#   E_K (two-body) = ',f16.8,4x,a)") E_tb_K, &
      "<== exact exchange E"
    write(OUTPUT, "('#   -----------------',a)") repeat( "-", 16 )
    write(OUTPUT, "('#   Total E (tb)   = ',f16.8,'  [ ',f12.2,' eV ]')") &
      E1_tb, E1_tb * AU2EV
    write(OUTPUT, "('#')")
  end if
  write(OUTPUT, "('#        ',a12)", advance="no") "E(0)"
  if ( P%N_elec > 1 .or. P%yes_screening .or. .not. P%yes_exact_pot ) then
    write(OUTPUT, "(a12)", advance="no") "E(1)"
    if ( P%yes_twobody ) write(OUTPUT, "(a12)", advance="no") "E(1)_tb"
  end if
  write(OUTPUT, "()")
  write(OUTPUT, "('# |E_TOT|',f12.2)", advance="no") E_sum_orb * AU2EV
  if ( P%N_elec > 1 .or. P%yes_screening .or. .not. P%yes_exact_pot ) then
    write(OUTPUT, "(f12.2)", advance="no") E * AU2EV
    if ( P%yes_twobody ) write(OUTPUT, "(f12.2)", advance="no") E1_tb * AU2EV
  end if
  write(OUTPUT, "()")
  write(OUTPUT, "('#')")
  !
  ! print orbital energies
  !
  write(OUTPUT, "('# Orbital energies: ',a)") trim(string)
  write(OUTPUT, "('#   ',4x,a6,3a15)", advance="no") &
    "n_occ", "E(a.u.)", "E(eV)", "KE(eV)"
  if ( P%yes_wigner_seitz ) write(OUTPUT, "(a11)", advance="no") "norm"
  write(OUTPUT, "()")
  do j = 1, F%N_orb
    ! two different ways give the same results; just one is chosen
    !E_Ts = functional_Ts( G, F%u, F%config, j )
    E_Ts = F%e_orb(j) - radial_integral( G, F%u(:,j)**2 * F%pot(:) )
    l = F%config(j,2)
    i = l * F%N + ( F%config(j,1) - l )
    write(OUTPUT, "('#   ')", advance="no")
    write(OUTPUT, "(a4)", advance="no") WF_name_nonrel( F, j )
    if ( P%yes_average_atom ) then      ! F%n_occ(1:F%N_all)
      write(OUTPUT, "(f6.3)", advance="no") F%n_occ(i)
    else if ( P%yes_fractional ) then   ! F%f_occ(1:F%N_orb_max)
      write(OUTPUT, "(f6.3)", advance="no") F%f_occ(j)
    else
      write(OUTPUT, "(i6)", advance="no") F%config(j,3)
    end if
    write(OUTPUT, "(f15.8,2f15.6)", advance="no") &
      F%e_orb(j), F%e_orb(j) * AU2EV, E_Ts * AU2EV
    if ( P%yes_wigner_seitz ) write(OUTPUT, "(f11.5)", advance="no") &
      radial_integral( G, F%u(1:j_s,j)**2 )
    if ( F%yes_flat_tail .and. F%e_orb(j) > F%e_s ) &
      write(OUTPUT, "(a)", advance="no") " [unbound]"
    write(OUTPUT, "()")
  end do
  !
  ! additional information for plasma claculation 
  !
  if ( F%yes_flat_tail ) write(OUTPUT, &
    "('# - pot_tail =',f15.8,f15.6)") F%e_s, F%e_s*AU2EV
  if ( P%yes_average_atom ) write(OUTPUT, &
    "('# - chem_pot =',f15.8,f15.6)") F%chem_pot, F%chem_pot*AU2EV
  if ( P%yes_wigner_seitz ) then
    write(OUTPUT, "('# - N of electrons within r_s = ',f8.2)") &
      integral( G, rho(1:j_s) )       ! double-check of # elec. within r_s
  end if
  !
  ! additional information for average-atom calculation
  !
  if ( P%yes_average_atom ) call print_average_atom( P, F, P%T_eV, OUTPUT )
  !
  ! print mean-field matrix elements
  !   F%config(j,2) : quantum number l
  !
  if ( P%yes_coulomb ) then
    write(OUTPUT, "('#'/'# Mean-field matrix elements:')")
    write(OUTPUT, "(2a5,2a18)") "j", "k", "norm", "v_bar_jk"
    do j = 1, F%N_orb
      do k = j, F%N_orb
        if ( F%config(j,2) == F%config(k,2) ) then
          write(OUTPUT, "(2i5,2es18.10)") j, k, &
            radial_integral_u( F, j, k ), &
            radial_integral_u( F, j, k, v_eff(:) - v_ext(:) )
        end if
      end do
    end do
    write(OUTPUT, "()")
  end if
  !
  ! print a summary
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('#      ',100a10)", advance="no") &
    (/ ( trim(WF_name_nonrel( F, j )), j=1, F%N_orb ) /)
  if ( F%yes_flat_tail ) write(OUTPUT, "(a10)", advance="no") "pot_tail"
  if ( P%yes_average_atom ) write(OUTPUT, "(a10)", advance="no") "chem_pot"
  write(OUTPUT, "()")
  write(OUTPUT, "('# |SCF|',100f10.2)", advance="no") F%e_orb(1:F%N_orb)*AU2EV
  if ( F%yes_flat_tail ) write(OUTPUT, "(f10.2)", advance="no") F%e_s * AU2EV
  if ( P%yes_average_atom ) &
    write(OUTPUT, "(f10.2)", advance="no") F%chem_pot * AU2EV
  write(OUTPUT, "()")
  !
  ! additional information when a flat potential tail is used
  !
  if ( F%yes_flat_tail ) then
    write(OUTPUT, "('# |IP| ',100f10.2)" ) ( F%e_s - F%e_orb(1:F%N_orb) )*AU2EV
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# - (inner) ionization potentials (IP) = e_s - e_i')")
    write(OUTPUT, "('# - potential tail e_s obtained using ')", advance="no")
    if ( P%yes_muffin_tin ) then
      write(OUTPUT, "('muffin-tin approx.')")
    else if ( P%yes_xpot ) then
      select case ( P%flat_tail_type )
        case ( 1 ); write(OUTPUT, "('a value at Wigner-Seitz radius')")
        case ( 2 )
          write(OUTPUT, "('a fixed value (touching-sphere scheme)')")
        case ( 3 )
          if ( P%SCF_max_iter == 1 .and. P%yes_load_pot ) then
            write(OUTPUT, "('loaded effective potential')")
          else
            write(OUTPUT, "('calculated effective potential')")
          end if
      end select
    end if
  end if
  !
  ! print all eigen-energies for different l's
  !
  if ( P%yes_e_all ) then
    if ( .not. P%yes_average_atom ) call solve_SE_all( F, G )
    call print_e_all( P, F, OUTPUT )
  end if
  !
  ! print potential and density: HERE F%pot(:) might be different from v_eff(:)
  ! NOTE: v_env(:) is already in non-uniform grids
  !
  if ( P%yes_print_pot ) then
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# Plotting of potential and density')")
    write(OUTPUT, "('#',a11,5a12)", advance="no") "r(a.u.)", &
      "v_eff", "v_ext", "v_J", "v_xc", "rho"
    if ( P%yes_screening ) &
      write(OUTPUT, "(2a12)", advance="no") "rho_bound", "rho_free"
    if ( P%yes_load_env ) write(OUTPUT, "(a12)", advance="no") "v_env"
    write(OUTPUT, "()")
    do j = 1, G%N
      write(OUTPUT, "(6es12.4)", advance="no") F%r(j), &
        F%pot(j), v_ext(j), v_J(j), v_xc(j), rho(j)
      if ( P%yes_screening ) &
        write(OUTPUT, "(2es12.4)", advance="no") rho_bound(j), rho_free(j)
      if( P%yes_load_env ) write(OUTPUT, "(es12.4)", advance="no") v_env(j)
      write(OUTPUT, "()")
    end do
    write(OUTPUT, "(/)")        ! print two blank lines
  end if
  !
  ! testing...
  !
  ! printing the pair correlation function
!  adj_ion_dens = THREE / FOUR / PI / G%r(j_s)**3
!  write(OUTPUT, "('#')")
!  write(OUTPUT, "('# nuc-elec pair correlation function:')")
!  write(OUTPUT, "('#   r_s      (input) (adjusted) = ',2f16.8)") P%r_s, G%r(j_s)
!  write(OUTPUT, "('#   ion_dens (input) (adjusted) = ',2es16.8)") &
!    P%ion_dens, adj_ion_dens
!  do j = 1, G%N
!    value = rho_free(j) / ( adj_ion_dens * F%Q_avg )
!    if ( G%r(j) >= G%r(j_s) ) then                ! r_s <= r < 2*r_s
!      value = value * ( ONE - (G%r(j)-G%r(j_s))/G%r(j_s) ) + &
!              ONE * ( (G%r(j)-G%r(j_s))/G%r(j_s) )
!      if ( G%r(j) >= G%r(j_s) * TWO ) value = ONE     ! 2*r_s <= r
!    end if
!    write(OUTPUT, "(2es16.8)") F%r(j), value
!  end do
!  v_eff(:) = transform_wf_to_A( G, F%pot(:) )
!  v_eff(:) = matmul( G%Del_r(:,:), v_eff(:) ) - v_eff(:) / G%r(:)
!  v_eff(:) = transform_A_to_wf( G, v_eff(:) )
!  !v_eff(:) = F%pot(:) * G%r(:)
!  !v_eff(:) = matmul( G%Del_r(:,:), v_eff(:) ) / G%r(:) - F%pot(:) / G%r(:)
!  do i = 1, G%N
!    write(OUTPUT, "(100es11.3)") G%r(i), F%pot(i), v_eff(i)
!  end do
  !
  ! print out total energy with integrals over all r
  !
!  value = functional_J( G, rho(1:j_s), v_J(:) )
!  E_J = functional_J( G, rho(1:G%N), v_J(:) )
!  E_x = functional_x( P, G, rho(1:G%N) )
!  E_c = functional_c( P, G, rho(1:G%N) )
!  E_xc_int = integral( G, rho(1:j_s) * v_xc(1:j_s) )
!  E = E_sum_orb + E_J + ( E_x + E_c ) - TWO*value - E_xc_int
!  write(OUTPUT, "('#')")
!  write(OUTPUT, "('#   E_J within r_s = ',f16.8)") value
!  write(OUTPUT, "('#   E_J with all r = ',f16.8)") E_J
!  write(OUTPUT, "('#   E_x with all r = ',f16.8)") E_x
!  write(OUTPUT, "('#   Total Energy   = ',f16.8)") E
  !
  ! print out IE according to the 1st-order correction using unrelaxed orbitals
  ! actually it prints out the total energy of 1s^-1
  !
!  F%config(1,3) = F%config(1,3) - 1     ! take out one electron from 1s
!  rho(:) = density( F%wf, F%config )
!  E_sum_orb = sum( F%e_orb(:) * F%config(:,3) )
!  E_J = functional_J( G, rho(1:j_s), v_J(:) )
!  E_x = functional_x( P, G, rho(1:j_s) )
!  E_c = functional_c( P, G, rho(1:j_s) )
!  E_xc_int = integral( G, rho(1:j_s) * v_xc(1:j_s) )
!  E = E_sum_orb - E_J + ( E_x + E_c ) - E_xc_int
!  write(OUTPUT, "('#')")
!  write(OUTPUT, "('#   Total Energy (1s^-1) = ',f16.8)") E
end subroutine SCF_procedure

!----------------------------------------------------------------------------
! check convergence with total energy and all points of density
!----------------------------------------------------------------------------
function check_convergence( P, E, E_old, rho, rho_old ) result (yes_conv)
  type(Param), intent(in) :: P
  real(long), intent(in) :: E, E_old
  real(long), intent(in), optional :: rho(:), rho_old(:)
  logical :: yes_conv
  if ( abs( E_old - E ) < P%SCF_energy_conv ) then
    yes_conv = .true.
    if ( present(rho) .and. present(rho_old) ) then
      !
      ! if density is given, 
      ! BOTH energy and density need to satisfy convergence conditions
      !
      if ( maxval( abs( rho_old(:) - rho(:) ) ) < P%SCF_density_conv ) then
        yes_conv = .true.
      else
        yes_conv = .false.
      end if
    end if
  else
    yes_conv = .false.
  end if
end function check_convergence

!----------------------------------------------------------------------------
! Latter tail correction
! SEE: paragraph after Eq. (4) in Son, Young & Santra, PRA 83, 033402 (2011).
!      Also see the commented-out subroutine below.
!----------------------------------------------------------------------------
subroutine Latter_correction( v, v_Coul )
  real(long), intent(inout) :: v(:)
  real(long), intent(in) :: v_Coul(:)
  where ( v(:) > v_Coul(:) ) v(:) = v_Coul(:)
end subroutine Latter_correction

!subroutine Latter_correction( G, F, v, v_Coul, SP, SP2 )
!  type(Coordinate), intent(in) :: G
!  type(WF_data), intent(inout) :: F
!  real(long), intent(inout) :: v(:)
!  real(long), intent(in) :: v_Coul(:)
!  type(Spline), intent(inout) :: SP, SP2
!  integer :: i, j, k, i0, i1, i2
!  real(long) :: r0, r1, r2, p0, p1, p2
!  real(long) :: r, p, t, r_min, r_max
!  !
!  ! originally suggested by Bhalla's paper
!  ! however, it makes a kinked point when it starts and 
!  ! it makes a lot of troubles in pot and w.f. and so on.
!  ! --> however, we decide to use this the simplest implementation.
!  !     it turns out... just one line!!!
!  !
!  where ( v(:) > v_Coul(:) ) v(:) = v_Coul(:)
!  !
!  ! smoothe it with Spline interpolation
!  ! interpolate it on the HS grids
!  ! and interpolate it back on the GPS grids... what a mess!
!  !
!  F%f_int(2:F%N+1) = v(:) * F%r(:)
!  call Spline_reuse( SP, F%f_int(:) )
!  F%f_HS(:) = (/ ( Spline_interpolate( SP, F%r_HS(j) ), j=1, F%N_HS ) /)
!  call Spline_reuse( SP2, F%f_HS(:) )
!  F%pot(:) = (/ ( Spline_interpolate( SP2, F%r(j) ), j=1, F%N ) /)
!  v(:) = F%pot(:) / F%r(:)
!  !
!  ! smoothly attach the Latter tail correction
!  !
!  ! using a quadratic Bezier curve with (r0,p0), (r1,p1), and (r2,p2)
!  !   X(t) = (1-t)^2 X0 + 2(1-t)t X1 + t^2 X2
!  !
!  do j = 1, G%N
!    if ( v(j) >= v_Coul(j) ) exit
!  end do
!  i1 = j;  r1 = G%r(i1);  p1 = v_Coul(i1) * r1
!  r_min = max( r1 - r_w, G%r(1) )
!  r_max = min( r1 + r_w, G%r(G%N) )
!  do k = i1, 1, -1
!    if ( G%r(k) < r_min ) exit
!  end do
!  i0 = k;  r0 = G%r(i0);  p0 = v(i0) * r0
!  do k = i1, G%N
!    if ( G%r(k) > r_max ) exit
!  end do
!  i2 = k;  r2 = G%r(i2);  p2 = v_Coul(i2) * r2
!  do i = i0, i2
!    r = G%r(i)
!    t = ( r0 - r1 + sqrt( (r0-r1)**2 - ( r0-TWO*r1+r2 ) * ( r0-r ) ) ) / &
!        ( r0-TWO*r1+r2 )
!    p = ( ONE - t )**2 * p0 + TWO * ( ONE - t ) * t * p1 + t**2 * p2
!    v(i) = p / r
!  end do
!  v(i2+1:) = v_Coul(i2+1:)
!  !
!  ! using a smooth step function
!  !
!  do j = 1, G%N
!    if ( G%r(j) > r_c ) then
!      v_eff(j) = v_Coul(j)
!    else if ( G%r(j) > r_c - TWO * r_w ) then
!      ratio = ONE / ( ONE + exp( ( G%r(j) - r_c - r_w ) / r_w * TWO ) )
!      v_eff(j) = v_eff(j) * ratio + v_Coul(j) * ( ONE - ratio )
!      if ( v_eff(j) > v_Coul(j) ) v_eff(j) = v_Coul(j)
!    end if
!  end do
!  !
!  ! generate grid points according to the original HS code
!  ! copied from CreaPot.f
!  !===========================================================================
!  u = 0.885341377d0/Z**(1.d0/3.d0)
!  deltax = 0.0025d0
!  x = 0.d0
!  i = 1
!  r(i) = u*x
!  do j = 1,11
!     do k = 1,40
!        i = i + 1
!        x = x + deltax
!        r(i) = u*x
!     end do
!     deltax = 2.d0*deltax
!  end do
!  !===========================================================================
!  !
!  ! interpolate pot(:) at r_i using spline
!  !   at r=0, r*pot = -Z
!  !   at r=r_max, r*pot = -(Z-Z0+1) where Z0=# of electons
!  !
!  N = size(pot,1) + 2
!  r_int(1) = ZERO;              rp_int(1) = -dble(Z_nuc)
!  r_int(2:N-1) = G%r(:);        rp_int(2:N-1) = v(:) * G%r(:)
!  r_int(N) = r(441);            rp_int(N) = -dble(Z_eff)
!  SP = Spline_create( r_int(:), rp_int(:) )
!end subroutine Latter_correction

!----------------------------------------------------------------------------
! Consider the charge density outside the WS radius
!   This is for testing. This should not be used in the SCF iterations.
!----------------------------------------------------------------------------
function calculate_rho_outside( G, rho, j_s ) result (rho_out)
  type(Coordinate), intent(in) :: G
  real(long), intent(inout) :: rho(:)
  integer, intent(in) :: j_s
  real(long) :: rho_out(G%N)
  real(long) :: rho_bar, v_out
  !real(long) :: rmax, r_s
  integer :: i, i_s
  !rmax = G%rmax
  !r_s = G%r(j_s)
  do i = 1, G%N
    if ( G%r(i) >= 10.0_long ) exit
  end do
  i_s = i
  rho(i_s+1:) = rho(i_s)
  i_s = G%N
  rho_out(:) = ONE
  rho_bar = ( integral( G, rho(1:i_s) )     - integral( G, rho(1:j_s) ) ) / &
            ( integral( G, rho_out(1:i_s) ) - integral( G, rho_out(1:j_s) ) )
  rho_out(1:j_s) = ZERO
  rho_out(j_s+1:i_s) = rho(j_s+1:i_s) - rho_bar
  !rho_out(i_s+1:) = ZERO
  v_out = integral( G, rho_out(1:i_s) / G%r(1:i_s) )
  print "(4es13.5)", rho_bar, integral( G, rho_out(1:i_s) ), v_out, &
                     TWO * integral( G, rho(1:j_s) ) * v_out
  do i = 1, G%N
    write(100,*) G%r(i), rho_out(i)
  end do
  write(100,"(/)")
end function calculate_rho_outside

subroutine save_potential( filename, pot )
  character(len=*), intent(in) :: filename
  real(long), intent(in) :: pot(:)
  integer, parameter :: IO_POT = 51
  open( IO_POT, file=trim(filename), action="write", form="formatted" )
  write( IO_POT, * ) pot(:)
end subroutine save_potential

subroutine load_potential( filename, pot )
  character(len=*), intent(in) :: filename
  real(long), intent(out) :: pot(:)
  integer, parameter :: IO_POT = 51
  open( IO_POT, file=trim(filename), action="read", form="formatted" )
  read( IO_POT, * ) pot(:)
end subroutine load_potential

!----------------------------------------------------------------------------
! adaptive_mixing_paramter(): 
!   a simple linear mixing scheme with an adaptive mixing parameter
!   based on the difference between v_i (before mixing) and v_old (after mixing)
! potential_mixing(): for further development, for example, DIIS
!----------------------------------------------------------------------------
function adaptive_mixing_parameter( v_i, v_old ) result (alpha)
  real(long) :: alpha
  real(long), intent(in) :: v_i(:)      ! newly-calculated pot. at i-th iter
  real(long), intent(in) :: v_old(:)    ! previously-used potential
  real(long) :: coeff, error
!TEST
!  real(long), save :: coeff_old = -ONE
!  if ( coeff_old == -ONE ) then         ! first iteration, do nothing for v_i
!    coeff_old = ZERO
!    !coeff_old = ONE / sqrt( dble( size(v_i) ) ) ! 1 / max. error for 100% diff
!    alpha = ONE
!  else                                  ! next iteration, mix them
!TEST
  real(long), save :: coeff_old = ZERO
  if ( coeff_old == ZERO ) then         ! first iteration, do nothing for v_i
    coeff_old = ONE / sqrt( dble( size(v_i) ) ) ! 1 / max. error for 100% diff
    alpha = ONE
  else                                  ! next iteration, mix them
    !
    ! based on John's implementation of "inverse error"
    !
    error = sqrt( sum( ( v_i(:) - v_old(:) )**2 / v_old(:)**2 ) )
    coeff = ONE / error
    alpha = coeff / ( coeff + coeff_old )       ! adaptive mixing parameter
    coeff_old = coeff_old + coeff               ! summation of all prev. coeffs
  end if
end function adaptive_mixing_parameter

!subroutine potential_mixing( v_i, v_prev, error_prev, i, v_old )
!  real(long), intent(inout) :: v_i(:)   ! newly calculated pot. at i-th iter
!                                        ! it will be replaced with v_new
!  real(long), intent(inout) :: v_prev(:,:), error_prev(:)
!  integer, intent(in) :: i              ! index for SCF iteration
!  real(long), intent(in) :: v_old(:)
!  real(long) :: v_new(size(v_i))        ! new potential after mixing
!  real(long) :: mixing_coeff(size(v_prev,2))
!  real(long) :: norm
!  !real(long) :: error, alpha
!  integer :: N, N_max, j, m
!  N = size(v_i)                 ! == size(v_prev,1), number of grids
!  N_max = size(v_prev,2)        ! number of max. previous steps
!  m = i                         ! how many prev. iterations do we have?
!  if ( i == 1 ) then            ! for the first iteration
!    v_prev(:,:) = ZERO          ! initialize
!    error_prev(:) = ZERO        ! initialize
!    v_prev(:,1) = v_i(:)
!    error_prev(1) = sqrt(dble(N))       ! max. error when 100% differences
!  else
!    if ( i > N_max ) then
!      do j = 1, N_max - 1       ! make a shift
!        v_prev(:,j) = v_prev(:,j+1)
!        error_prev(j) = error_prev(j+1)
!      end do
!      m = N_max                 ! m should be <= N_max
!    else
!    end if
!    v_prev(:,m) = v_i(:)        ! store the newly-calculated potential
!    !
!    ! error is defined as r.m.s.
!    !
!    error_prev(m) = sqrt( sum( ( v_prev(:,m) - v_prev(:,m-1) )**2 &
!                               / v_prev(:,m-1)**2 ) )
!  end if
!!TEST --- error defined with the current potential
!!  if ( m > 1 ) then
!!    do j = 1, m - 1
!!      error_prev(j) = sqrt( sum( ( v_i(:) - v_prev(:,j) )**2 &
!!                                 / v_prev(:,j)**2 ) )
!!    end do
!!  end if
!!TEST
!  !
!  ! set up mixing coefficients
!  !
!  mixing_coeff(1:m) = ONE / error_prev(1:m)
!  norm = sum( mixing_coeff(1:m) )
!  mixing_coeff(1:m) = mixing_coeff(1:m) / norm
!  print "(2i3,100es10.2)", i, m, mixing_coeff(1:m)
!  !
!  ! mix previous potentials
!  !
!  v_new(:) = ZERO
!  do j = 1, m
!    v_new(:) = v_new(:) + mixing_coeff(j) * v_prev(:,j)
!  end do
!!TEST --- just adding v_old with (1-alpha)
!!  alpha = mixing_coeff(m)
!!  v_new(:) = alpha * v_i(:) + ( ONE - alpha ) * v_old(:)
!!  error = sqrt( sum( ( v_i(:) - v_old(:) )**2 / v_old(:)**2 ) )
!!  print *, "|ERROR|", i, alpha, error
!!TEST
!  !
!  ! v_eff(:) fed with new potential v_new
!  !
!  v_i(:) = v_new(:)
!!TEST
!!  !
!!  ! HEY, why don't we put the new potential into the latest element in v_prev???
!!  !
!!  if ( m /= 1 ) then
!!    v_prev(:,m) = v_new(:)
!!    error_prev(m) = sqrt( sum( ( v_prev(:,m) - v_prev(:,m-1) )**2 &
!!                               / v_prev(:,m-1)**2 ) )
!!  end if
!!TEST
!end subroutine potential_mixing

subroutine DIIS(G, i, v_eff, v_reserv, v_old)
  use matrix
  type(LU_matrix) :: BLU
  type(Coordinate), intent(in) :: G
  integer, parameter :: Nm = 5
  !integer :: N
  integer, intent(in) :: i ! the current iteration number
  real(long), intent(inout) :: v_reserv(:,:)
  real(long), intent(inout) :: v_eff(:), v_old(:)
  real(long) :: dif(Nm,size(v_eff))
  real(long), allocatable :: B(:, :), wt(:), br(:)
  real(long) :: norm
  integer :: itr, jtr, im
  ! constructing Pulay's least square fit condition matrix. 
  ! B * wt = br
  ! wt = ( c1, c2, ..., lambda)
  v_reserv(i,:) = v_eff(:)
  !write(*,*)i
  !write(*,"(20es10.2)")v_reserv(i,1:20)
  !
  if( i == 1 ) then
    return
  endif
  
  im = min(Nm, i)
  !
  allocate( B(im + 1, im + 1), wt(im + 1), br(im + 1) )
  do itr = i-im + 1, i   ! within latest im elements
    jtr = itr - i + im
    if( itr == 1 ) then
      dif(jtr, :) = (v_reserv(itr,:) - v_old(:)) !&
!                    / (v_reserv(itr, :) + 1.d-20)
  !    write(*,"(20es10.2)") dif(jtr,1:20)
    else
      dif(jtr, :) = (v_reserv(itr,:) - v_reserv(itr - 1, :)) !&
!                    / (v_reserv(itr, :) + 1.d-20)
   !   write(*,"(20es10.2)") dif(jtr,1:20)
    endif
    !dif(jtr,:) = abs(dif(jtr,:)) / integral( G, dif(jtr,:) )
    dif(jtr,:) = dif(jtr,:) / integral( G, dif(jtr,:) )
  enddo
  
  do itr = 1, im
    do jtr = 1, im
!      B(itr, jtr) = sum(dif(itr,:) * dif(jtr,:)) 
      B(itr, jtr) = integral( G, dif(itr,:) * dif(jtr,:) )
    enddo 
  enddo 
  B(im + 1, :) = -1.d0
  B(:, im + 1) = -1.d0
  B(im + 1, im + 1) = 0.d0
  !write(*,"(<im+1>es10.2)")B   ! disabled this line to make it gfortran compatible
  br(1:im) = 0.d0
  br(im + 1) = -1.d0 
  ! LU decompose B and solve the linear equation set.
  BLU = LU_decompose(B)
  wt = LU_solve(BLU, br)
  norm = sum(wt(:))
  wt(1:im+1) = wt(1:im+1)/norm
  v_eff(:) = 0.d0
  do itr = 1, im
    v_eff(:) = v_eff(:) + v_reserv(itr, :) * wt(itr)
  enddo 

  write(*,"(i3,'# wt' 50f10.3)")i, wt(1:im)
  deallocate(B, wt, br)
  
end subroutine
end module scf
