!----------------------------------------------------------------------------
! shakeoff.f90
!----------------------------------------------------------------------------
! Copyright (C) 2012-2019 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2012-2019 Sang-Kil Son <sangkil.son@cfel.de>
!          2016-2018 Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! shake-off branching ratio calculation
!   written by Son, Sang-Kil in Mar. 2012 at DESY
!----------------------------------------------------------------------------
module shakeoff
  use constants
  use read_param
  use coord_system
  use gps
  use wavefunction
  use process
  use scf
  use track_elec_config
  use random_num
  implicit none

  type :: ShakeOff_data
     type(WF_data) :: F(2)
     real(long), allocatable :: shakeoff(:,:)
     integer, allocatable :: shakeoff_E(:,:), config(:,:) 
     real(long), allocatable :: sum_br(:)
     integer :: occ_index, OUTPUT
     integer :: N_subshell
     logical :: yes_relativity
     logical :: is_so_happened 
  end type ShakeOff_data
contains
  function CreateShakeOffData( P, G ) result( SO )
    type(Param), intent(in) :: P
    type(Coordinate), intent(in) :: G
    type(ShakeOff_data) :: SO
    
    SO%is_so_happened = .false.

    SO%OUTPUT = P%OUTPUT

    ! wave function before and after the SO 
    SO%F(1) = create_WF( P, G )
    SO%F(2) = create_WF( P, G )

    SO%yes_relativity = P%yes_relativity
    if ( SO%yes_relativity ) then
       SO%occ_index = 4
       SO%N_subshell = P%N_orb_rel
    else
       SO%occ_index = 3
       SO%N_subshell = P%N_orb
    endif
    allocate( SO%config( SO%N_subshell, SO%occ_index ) )
    SO%config(:,:) = 0

    allocate( SO%shakeoff(P%N_proc_max, SO%N_subshell), &
              SO%shakeoff_E(P%N_proc_max, SO%N_subshell), &
              SO%sum_br( P%N_proc_max ) )
    SO%shakeoff(:,:) = ZERO
    SO%shakeoff_E(:,:) = ZERO

    SO%sum_br(:) = ZERO

    ! for xatom_mc_fly
    call initialize_random()

  end function CreateShakeOffData

  subroutine RegisterWF( F, SO, flag_str )
    type(WF_data), intent(in), target :: F
    type(ShakeOff_data), intent(inout) :: SO
    character(len=*), intent(in) :: flag_str
    integer :: n

    if ( flag_str == "before" ) then
       n = 1
    elseif ( flag_str == "middle" ) then
       n = 2
    else
       stop 'in RegisterWF in shakeoff.f90: flag_str must be before or middle'
    endif

    ! copy orbital wave function and orbital energies
    SO%F(n)%u(:,:) = F%u(:,:)
    if ( SO%yes_relativity ) then
       SO%F(n)%e_orb_rel(:) = F%e_orb_rel(:)
    else
       SO%F(n)%e_orb(:) = F%e_orb(:)
    endif
  end subroutine RegisterWF

  subroutine ConfigProcessAndRegister( P, G, S, k, SO, flag_str )
    type(Param), intent(inout) :: P
    type(Coordinate), intent(inout) :: G
    type(Process_data), intent(in) :: S
    integer, intent(in) :: k
    type(ShakeOff_data), intent(inout) :: SO
    character(len=*), intent(in) :: flag_str
    integer :: n, output

    if ( flag_str == "before" ) then
       n = 1
    elseif ( flag_str == "middle" ) then
       n = 2
    else
       stop 'in ConfigProcessAndRegister in shakeoff.f90: flag_str must be before or after'
    endif

    ! proceed decay of the process id of k
    SO%config = config_Process( S, k )
    SO%F(n)%config(:,:) = get_non_relativistic_config( P, SO%config )

    call SCF_procedure( P, G, SO%F(n) )
    if ( SO%yes_relativity ) then 
       output = P%OUTPUT
       P%OUTPUT = P%IO_NULL
       call relativity_calc_first_order_corr( P, G, SO%F(n) )
       P%OUTPUT = output
    endif

  end subroutine ConfigProcessAndRegister
    
  subroutine CalculateShakeoffBR( P, G, S, id, SO )
    type(Param), intent(in) :: P
    type(Coordinate), intent(in) :: G
    type(Process_data), intent(inout) :: S
    type(Shakeoff_data), intent(inout) :: SO
    integer, intent(in) :: id
    integer :: i, occ_ip, wf_ind_i
    real(long) :: overlap, prob

    if ( .not. P%yes_shakeoff ) &
      stop "No CalculateShakeoffBR is called when P%yes_shakeoff is off"

    ! shake-off probability by simple orbital overlap
    ! T. A. Carlson and C. W. Nestor Jr, PRA 8, 2887 (1973)
    prob = 0.d0
    SO%shakeoff( id, :) = ZERO
    SO%shakeoff_E( id,:) = ZERO
    SO%sum_br( id ) = ZERO

    do i = 1, SO%N_subshell
       if ( SO%yes_relativity ) then
          wf_ind_i = SO%F(2)%config_rel( i, 5 )
          occ_ip = SO%F(2)%config_rel( i, 4 )
       else
          wf_ind_i = i
          occ_ip = SO%F(2)%config( i, 3 )
       endif

       if ( occ_ip == 0 ) cycle     ! no electrons in this subshell
       SO%shakeoff_E( id, i )  = S%E_eV(id) - &
                                 int( ionization_E( SO%F(2), i ) * AU2EV )
       if ( SO%shakeoff_E( id, i ) < ZERO ) cycle ! shake-off electron must have at least zero energy

       ! calcuate the orbital overlap
       overlap = radial_integral( G, SO%F(1)%u(:,wf_ind_i) * SO%F(2)%u(:,wf_ind_i) )

       !
       ! calculate single shake-off probabilities
       !
       SO%shakeoff( id, i ) = dble( occ_ip )*( ONE - overlap**2 ) 
       SO%sum_br( id ) = SO%sum_br( id ) + SO%shakeoff( id, i )
       SO%shakeoff( id, i ) = SO%shakeoff( id, i ) * S%value(id)
    end do
    ! correct rate/pcs
    S%value(id) = ( 1.d0 - SO%sum_br(id) ) * S%value(id)
  end subroutine CalculateShakeoffBR

  subroutine ShowResult( S, SO, id )
    type(Process_data), intent(in) :: S
    type(Shakeoff_data), intent(inout) :: SO
    integer, intent(in) :: id
    !integer, allocatable :: config(:,:)
    integer :: i !, occ_index

    ! rate without shakeoff
    call PrintOut( S, id, SO, 0, 0 )

    do i = 1, SO%N_subshell
       if ( SO%yes_relativity ) then
          SO%config(:,:) = SO%F(2)%config_rel(:,:)
       else
          SO%config(:,:) = SO%F(2)%config(:,:)
       endif

       if ( SO%config(i, SO%occ_index) == 0 .or. SO%shakeoff_E(id, i) < 0 ) cycle

       ! shakeoff happens
       call PrintOut( S, id, SO, i, 1 ) 
    enddo
    
  end subroutine ShowResult

  subroutine PrintOut( S, id, SO, orb_index, SO_type ) 
    type(Process_data), intent(in) :: S
    integer, intent(in) :: id
    type(Shakeoff_data), intent(inout) :: SO
    integer, intent(in) :: orb_index, SO_type
    integer*8 :: before, middle, after
    character(len=10000) :: string

    ! configuration index 
    if ( SO%yes_relativity ) then
       call config_rel_to_index( SO%F(1)%config_rel, before ) ! before the primal decay
       call config_rel_to_index( SO%F(2)%config_rel, middle ) ! after the primal decay
    else
       call config_to_index( SO%F(1)%config, before )
       call config_to_index( SO%F(2)%config, middle )
    endif

    select case ( SO_type )
    case ( 0 )
       write(string, "('#',x,a1,2i10,10x,es13.5,x,i10,a)") &
            S%tag(id), before, middle, &
            S%value(id), S%E_eV(id)
    case ( 1 )
       if ( SO%yes_relativity ) then
          SO%config(:,:) = SO%F(2)%config_rel(:,:)
          SO%config(orb_index,4) = SO%config(orb_index,4) - 1 ! after the shakeoff
          call config_rel_to_index( SO%config, after )
       else
          SO%config(:,:) = SO%F(2)%config(:,:)
          SO%config(orb_index,3) = SO%config(orb_index,3) - 1 
          call config_to_index( SO%config, after )
       endif

       write(string, "('#',x,a1,3i10,es13.5,x,i10,a,a5,a2)") &
            S%tag(id), before, middle, after, &
            SO%shakeoff(id, orb_index), SO%shakeoff_E(id, orb_index), " ;S1"
    case ( 2 ); string = " ;S2"
    end select
    write(SO%OUTPUT,"(a)")trim( adjustl(string) )
  end subroutine PrintOut

  ! After a primal decay such as P, A, F etc, proceed shakeoff process
  subroutine ProceedShakeoff( orbital, id, SO ) !, prob )
    integer, intent(inout) :: orbital(:,:)
    integer, intent(in) :: id
    type(Shakeoff_data), intent(inout) :: SO
    !real(long), intent(in) :: prob
    real(long) :: rnd, br_sum
    integer :: i

    SO%is_so_happened = .false.

    rnd = random()

    ! no shakeoff
    if ( rnd > SO%sum_br( id ) ) return

    br_sum = ZERO
    do i = 1, SO%N_subshell
       if ( SO%shakeoff( id, i ) == ZERO ) cycle
       br_sum = br_sum + SO%shakeoff( id, i )
       if ( br_sum > rnd ) then
          orbital( i, SO%occ_index ) = orbital( i, SO%occ_index ) - 1
          SO%is_so_happened = .true.
          exit
       endif
    enddo

  end subroutine ProceedShakeoff

end module shakeoff
