!----------------------------------------------------------------------------
! spectra.f90
!----------------------------------------------------------------------------
! Copyright (C) 2017 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2017 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Analyze (fluorescence) spectra generated from xatom_twostep.f90
!----------------------------------------------------------------------------
program analyze_spectra
  use constants
  use stdio
  implicit none
  type(Argument) :: ARG
  integer :: i, j, k, i_max
  integer :: N, N_col, N_col_ref
  real(long) :: FWHM, s, s0
  real(long) :: E, dE, E_k, E_max, E_min
  real(long) :: t, t_k, t_min, t_max, dt
  real(long) :: value
  real(long), allocatable :: f(:,:)
  real(long), allocatable :: spec_conv(:)       ! spectrum convoluted
  real(long), allocatable :: spec_orig(:)       ! spectrum not-convoluted
  real(long), allocatable :: spec_E(:)          ! energy bin
  character(len=1024) :: filename
  character(len=1000000) :: line
  integer :: INPUT = 10, OUTPUT = STDOUT, status
  logical :: yes_TD         ! time-dependent spectrum
  logical :: yes_TD_matrix  ! TD spectrum matrix form to be used with gnuplot
  !
  ! get some parameters
  !
  E_min = ZERO
  E_max = -ONE
  dE = ONE      ! 1 eV
  dt = ONE      ! 1 fs
  FWHM = 5_long ! 5 eV
  yes_TD = .false.
  call get_input( ARG )
  call convert_option_lowercase( ARG )
  call parse_option( ARG, "fwhm", FWHM )
  call parse_option( ARG, "bandwidth", FWHM )
  call parse_option( ARG, "resolution", FWHM )
  call parse_option( ARG, "file", filename )
  call parse_option( ARG, "data", filename )
  call parse_option( ARG, "e", E_min, E_max )
  call parse_option( ARG, "e_min", E_min )
  call parse_option( ARG, "e_max", E_max )
  call parse_option( ARG, "de", dE )
  call parse_option( ARG, "td", yes_TD )
  call parse_option( ARG, "td_matrix", yes_TD_matrix )
  !
  ! find # of data
  !
  open( INPUT, file=filename, action="read", form="formatted" )
  N = 0         ! # of rows
  N_col = 0     ! # of columns
  N_col_ref = -1
  do
    read( INPUT, "(a)", iostat=status ) line
    if ( status < 0 ) exit
    !
    ! checking # of rows: lines starting with '#' are comments
    !
    if ( len_trim( line ) /= 0 .and. line(1:1) /= "#" ) N = N + 1
    do j = 1, len_trim( line )
      if ( line(j:j) == "#" ) then      ! comment on the right side of #
        line(j:) = ""
        exit
      end if
    end do
    !
    ! checking # of columns: copied from Zoltan's vol_int
    !
    N_col = count( [ len_trim(line) > 0, &
            ( line(i:i) /= ' ' .and. line(i:i) /= ',' .and. &
              line(i+1:i+1) == ' ' .or. line(i+1:i+1) == ',', &
              i=1, len_trim(line)-1 ) ] )
    if ( N_col_ref == -1 ) N_col_ref = N_col
    if ( N_col /= N_col_ref ) &
      stop "Inconsistency for the number of columns in datafile."
  end do
  allocate( f(N,N_col) )
  rewind( INPUT )
  !
  ! read data
  !
  i = 0
  do
    read( INPUT, "(a)", iostat=status ) line
    if ( status < 0 ) exit
    if ( len_trim( line ) /= 0 .and. line(1:1) /= "#" ) then
!      do j = 1, len_trim( line )
!        if ( line(j:j) == "#" ) then
!          line(j:) = ""
!          exit
!        end if
!      end do
      i = i + 1
      read( line, * ) f(i,1:N_col)
    end if
  end do
  close( INPUT )
  ! --------------------------------------------------------------------------
  ! analyze data
  ! - f(:,1) : time (in fs)
  ! - f(:,2) : fluorescence energy (in eV)
  ! - f(:,3) : probability to have a signal at a given energy a given time
  ! - f(:,4) : fluorescence rate... but it's not used
  ! --------------------------------------------------------------------------
  !
  ! set up variables
  !
  t_min = minval( f(:,1) )
  t_max = maxval( f(:,1) )
  dt = ZERO
  do k = 1, N - 1
    if ( ( f(k+1,1) - f(k,1) ) > dt ) dt = ( f(k+1,1) - f(k,1) )
  end do
  call parse_option( ARG, "t", t_min, t_max )
  call parse_option( ARG, "time", t_min, t_max )
  call parse_option( ARG, "t_min", t_min )
  call parse_option( ARG, "t_max", t_max )
  call parse_option( ARG, "until", t_max )
  call parse_option( ARG, "dt", dt )
  if ( E_max == -ONE ) E_max = maxval( f(:,2) ) + FWHM * TWO
  !
  ! print out the header
  !
  write( OUTPUT, "('# Analyzing spectra')")
  write( OUTPUT, "('# - energy range (eV) = ',f8.1,' - ',f8.1)") E_min, E_max
  write( OUTPUT, "('# - energy step (eV)  = ',f8.1)") dE
  if ( dE == ONE .and. FWHM == ZERO ) then
    write( OUTPUT, "('# - no convolution is made')")
  else
    write( OUTPUT, "('# - convolution with Gaussian, FWHM (eV) = ',f7.2)" ) FWHM
  end if
  write( OUTPUT, "('# - time range (fs) = ',f8.1,' - ',f8.1)") t_min, t_max
  if ( yes_TD ) write( OUTPUT, "('# - time-depedent spectrum')" )
  if ( yes_TD_matrix ) &
    write( OUTPUT, "('# - time-depedent spectrum: matrix form')" )
  if ( yes_TD .or. yes_TD_matrix ) then
    write( OUTPUT, "('#   dt (fs) = ',f8.1)" ) dt
  end if
  write( OUTPUT, "('#')" )
  !
  ! scan energy and calculate spectrum
  !
  i_max = max( int( ( E_max - E_min ) / dE ), 1 )
  allocate( spec_conv(1:i_max), spec_orig(1:i_max), spec_E(1:i_max) )
  spec_conv(:) = ZERO
  spec_orig(:) = ZERO
  spec_E(:) = (/ ( E_min + dble(i) * dE, i=1,i_max ) /)
  j = 3         ! only one column is considered now
  t = t_min
  do k = 1, N   ! scan all data
    t_k = f(k,1)        ! time of the k-th data
    E_k = f(k,2)        ! energy of the k-th data
    if ( t_k < t .or. t_k > t_max ) cycle       ! skip if not in the time range
    do i = 1, i_max
      if ( abs( E_k - spec_E(i) ) < HALF * dE ) &
        spec_orig(i) = spec_orig(i) + f(k,j)
      value = f(k,j) * exp( -(spec_E(i)-E_k)**2 / FWHM**2 *FOUR*log(TWO) ) &
                     / FWHM / sqrt( PI / FOUR / log(TWO) ) * dE
      if ( value > 1e-99_long ) spec_conv(i) = spec_conv(i) + value
    end do
    !
    ! print out time-dependent spectrum
    !
    if ( yes_TD .or. yes_TD_matrix ) then
      if ( k == N .or. f(k+1,1) >= t + dt ) then
        if ( yes_TD ) then
          write( OUTPUT, "('# time(fs) = ',es13.5)" ) t
          do i = 1, i_max
            if ( ( i /= 1 .and. i /= i_max ) .and. &
                 ( spec_conv(i) == ZERO .and. spec_orig(i) == ZERO ) ) cycle
            write( OUTPUT, "(i6,2es16.8)" ) &
              int(spec_E(i)), spec_conv(i), spec_orig(i)
          end do
          write( OUTPUT, "(/)" )
        else if ( yes_TD_matrix ) then
          write(OUTPUT,"(100000es16.8)") spec_conv(:)
        end if
        spec_conv(:) = ZERO
        spec_orig(:) = ZERO
        t = t + dt
      end if
    end if
  end do
  !
  ! print out time-integrated spectrum
  !
  if ( ( yes_TD .or. yes_TD_matrix ) == .false. ) then
    do i = 1, i_max
      write( OUTPUT, "(i6,2es16.8)" ) int(spec_E(i)), spec_conv(i), spec_orig(i)
    end do
  end if

!  do i = 1, i_max
!    E = E_min + dble(i) * dE
!    write( OUTPUT, "(i6)", advance="no" ) int(E)
!    do j = 3, 3 ! N_col - 1
!      !
!      ! perform integral
!      !
!      s  = ZERO
!      s0 = ZERO
!      do k = 1, N
!        t_k = f(k,1)    ! first column:  time in fs
!        E_k = f(k,2)    ! second column: energy in eV
!        if ( t_k < t_min .or. t_k > t_max ) cycle  ! skip if not in time range
!        if ( abs( E_k - E ) < EPS ) s0 = s0 + f(k,j)
!        s = s + f(k,j) * exp( -( E - E_k )**2 / FWHM**2 *FOUR*log(TWO) )
!      end do
!      s = s / FWHM / sqrt( PI / FOUR / log(TWO) ) * dE
!      !s0 = sum( f(:,j), abs( f(:,2) - E ) < EPS )
!      !if ( dE == ONE .and. FWHM == ZERO ) then
!      !  s = s0
!      !else
!      !  s = sum( f(:,j) * exp( -( E - f(:,2) )**2 / FWHM**2 *FOUR*log(TWO) ) ) &
!      !           / FWHM / sqrt( PI / FOUR / log(TWO) ) * dE
!      !end if
!      if ( s  < 1e-99_long ) s  = ZERO
!      if ( s0 < 1e-99_long ) s0 = ZERO
!      write( OUTPUT, "(2es16.8)", advance="no" ) s, s0
!    end do
!    write( OUTPUT, "()" )
!  end do
end program analyze_spectra
