!----------------------------------------------------------------------------
! spherical_bessel.f90
!----------------------------------------------------------------------------
! Copyright (C) 2013-2014 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2013 Jan Malte Slowik
!          2014 Otfried Geffert
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
!******************************************************************************
! This subroutine calculates the spherical bessel function.
! The subroutine get_spherical_bessel gives out a matrix j, where
! the columns j(:,l) are the values of the l-th spherical bessel function at positions Qr.
! The subroutine sphj is contained in the file m_spherical_bessel.f and was written by
!	Shanjie Zhang and Jianming Jin and was found at 
! http://jin.ece.illinois.edu/routines/routines.html.
!                                               Jan Malte Slowik, July 2013.
!******************************************************************************
module spherical_bessel
  use constants
  implicit none
contains

subroutine get_spherical_bessel(lmax, Q, r, N, j)
  integer, intent(in) :: lmax, N
  integer :: lmaxcomp 
  real(long), intent(in) :: Q
  real(long), intent(in) :: r(:)
  real(long) :: x
  real(long),dimension(lmax) :: bes, besprime
  real(long), dimension(N,lmax), intent(out) :: j
  integer :: i
  real(long) :: infinity
  integer :: test
  infinity = huge(Q)    !largest value of double precision variable
  do i = 1, N
    x = Q * r(i)
    if (x == ZERO) then
      j(i,:) = ZERO
      j(i,1) = ONE
    else
      call sphj(lmax-1, x, lmaxcomp, bes, besprime)
      do test = 1,lmax
        if( bes(test) > infinity .or. bes(test) /= bes(test) ) then     !check whether infinity or NaN
          write(STDERR, *) 'Calculation of spherical Bessel functions failed.' 
          stop
        end if 
      end do
      j(i,:) = bes
    end if
  end do
end subroutine get_spherical_bessel

end module spherical_bessel
