!----------------------------------------------------------------------------
! stdio.f90
!----------------------------------------------------------------------------
! Copyright (C) 2005      Sang-Kil Son
! Copyright (C) 2010-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2017 Sang-Kil Son <sangkil.son@cfel.de>
!          2016      Otfried Geffert
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! STDIO : Standard Input/Output module
!   written by Son, Sang-Kil in Mar. 2005
!
! This module requires two intrinsic subroutines: getarg() and getenv()
!
! Data type
! - Argument : for command-line & input file options
!
! Functions(F) and subroutines(SR)
! - SR error_message( string ) : print error messages and stop
!
! - SR get_options( ARG ) : get options from command line
! - SR get_input( ARG )   : get options from input file and command line
!   ex) foobar -N 30  or  -N:30  or -N=30  or -N30  or +N 30
!       foobar -N 10-30
!       foobar file.in  or  -N 30 file.in
!       * if no option, file.in will be taken as input file
! - SR parse_option( ARG, option_name, string or value )
! - SR show_input( ARG )  : show options from input file and command line
!
! - SR footnote() : print out CPU_time, current time, and hostname
!
! - SR check_time( counter ) : check elapsed time
! - F job_time() : return job time converted from counter with hostname
!   . F job_time_integer( counter ) : integer value from check_time()
!   . F job_time_real( CPU_time ) : real value from CPU_time()
!
! - F today() : return the date string of today
! - F hostname() : return the hostname
!----------------------------------------------------------------------------
module stdio
  use constants
  implicit none

  type :: arrptr
    real(long), pointer, dimension(:) :: ptr
  end type arrptr

  type :: Argument
    integer :: N
    character(len=8192), dimension(:), allocatable :: option, string
    real(long), dimension(:), allocatable :: value, value2
    logical, dimension(:), allocatable :: yes
    type(arrptr), dimension(:), allocatable :: val_arr
  end type Argument

  interface job_time
    module procedure job_time_integer8, job_time_real
  end interface
  interface parse_option
    module procedure parse_option_string, &
                     parse_option_real, parse_option_integer, &
                     parse_option_logical
  end interface
contains

!----------------------------------------------------------------------------
! show error message and stop the program
!----------------------------------------------------------------------------
subroutine error_message( s )
  character(len=*), optional :: s
  if ( present(s) ) write(STDERR, *) "ERROR: " // s
  stop
end subroutine error_message


!----------------------------------------------------------------------------
! get options from command lines and the input file and put them into ARG
!----------------------------------------------------------------------------
subroutine get_input( ARG )
  type(Argument), intent(out) :: ARG
  type(Argument) :: ARG_CL
  integer :: i, j, k, n, n_line, status
  character(len=8192) :: filename, line, s
  integer :: INPUT_FILE = 9
  integer, parameter :: MAX_N_LINE = 20000
  logical :: yes_exist
  !
  ! First, get options from command line
  !
  call get_options( ARG_CL )
  !
  ! find filename in ARG_CL
  !
  n_line = 0    ! total # line of file
  arg_index: do i = 1, ARG_CL%N
    select case ( ARG_CL%option(i) )
      case ( "" )
        filename = ARG_CL%string(i)
        if ( filename == "" ) then
          INPUT_FILE = STDIN
          n_line = MAX_N_LINE   ! add additional lines of input file
        else
          inquire(file=filename, exist=yes_exist)
          if ( .not. yes_exist ) &
            call error_message( trim( filename ) // " is not found." )
          open(INPUT_FILE, file=filename, action="read", form="formatted")
          n_line = 0
          do
            read(INPUT_FILE, "(a)", iostat=status) line
            if ( status < 0 ) exit
            n_line = n_line + 1
          end do
          rewind(INPUT_FILE)
        end if
        exit arg_index
    end select
  end do arg_index
  !
  ! memory allocation
  !
  n = ARG_CL%N + n_line
  allocate( ARG%option(n), ARG%string(n), &
            ARG%value(n), ARG%value2(n), ARG%yes(n), ARG%val_arr(n) )
  !
  ! read data from input file or STDIN
  !
  n = 0         ! # meaningful line 
  if ( n_line > 0 ) then
    do
      read(INPUT_FILE, "(a)", iostat=status) line
      if ( status < 0 ) exit
      !
      ! put this line into ARG
      !
      if ( len_trim( line ) /= 0 .and. line(1:1) /= "#" ) then
        n = n + 1
        if ( n > ARG_CL%N + MAX_N_LINE ) &
          call error_message( "The length of input file is too long." )
        !
        ! replace tab characters into spaces: ichar(TAB) = 9
        !
        do j = 1, len_trim(line)
          if ( ichar( line(j:j) ) == 9 ) line(j:j) = " "
        end do
        !
        ! parse an option name
        !
        j = index(line, "=")
        if ( j == 0 ) j = index(line, ":")
        ARG%option(n) = "-" // adjustl( line(1:j-1) )
        !
        ! remove a trailing comment
        !
        k = index(line, "#")    ! find out a comment with "#"
        if ( k == 0 ) then
          ARG%string(n) = adjustl( line(j+1: ) )
        else
          ARG%string(n) = adjustl( line(j+1:k-1) )
        end if
        !
        ! parse an argument (or arguments)
        !
        s = ARG%string(n)
        read(s, *, iostat=status) ARG%value(n)
        !
        ! check the range such as "30-40", "30:40", "30~40"
        ! exclude exponential expressions like "1e-10" or "1E-10"
        !
        k = scan( s, "-~,:;" )
        if ( k /= 0 ) then
          if ( s(k-1:k) /= "e-" .and. s(k-1:k) /= "E-" ) then
            read(s( :k-1), *, iostat=status) ARG%value(n)
            read(s(k+1: ), *, iostat=status) ARG%value2(n)
          end if
        end if
        ARG%yes(n) = determine_yes_no( s )
      end if
    end do
    rewind(INPUT_FILE)
  end if
  !
  ! copy ARG_CL to ARG
  !
  ARG%N = ARG_CL%N + n
  do i = n + 1, ARG%N
    ARG%option(i) = ARG_CL%option(i - n)
    ARG%string(i) = ARG_CL%string(i - n)
    ARG%value(i)  = ARG_CL%value(i - n)
    ARG%value2(i) = ARG_CL%value2(i - n)
    ARG%yes(i) = ARG_CL%yes(i - n)
    ARG%val_arr(i) = ARG_CL%val_arr(i - n)
  end do
  !
  ! deallocate ARG_CL
  !
  deallocate( ARG_CL%option, ARG_CL%string, &
              ARG_CL%value, ARG_CL%value2, ARG_CL%yes, ARG_CL%val_arr )
end subroutine get_input


!----------------------------------------------------------------------------
! get command-line options and put them into ARG
!----------------------------------------------------------------------------
subroutine get_options( ARG )
  type(Argument), intent(out) :: ARG
  integer :: narg, i, k, n
  integer :: iargc, status
  real(long) :: value
  character(len=8192) :: s, s2, s3

  narg = iargc()
  allocate( ARG%option(narg), ARG%string(narg) )
  allocate( ARG%value(narg), ARG%value2(narg), ARG%yes(narg) )
  ARG%option(:) = ""
  ARG%value(:) = ZERO
  ARG%value2(:) = ZERO
  ARG%string(:) = ""
  ARG%yes(:) = .false.
  allocate( ARG%val_arr(narg) )
  do n = 1, narg
    nullify( ARG%val_arr(n)%ptr )
  end do

  i = 0
  do n = 1, narg
    i = i + 1
    if ( i > narg ) exit
    call getarg( i, s )
    if ( s(1:1) == "-" .or. s(1:1) == "+" ) then
      !
      ! option sign started with - or +
      !
      s2 = "";  value = ZERO
      if ( i+1 <= narg ) call getarg( i+1, s2 )
      read(s2, *, iostat=status) value
      if ( i+1 > narg .or. &
           ( s2(1:1) == "-" .and. value == ZERO ) .or. &
           ( s2(1:1) == "+" .and. value == ZERO ) ) then
        !
        ! next option follows the option
        ! so, we need to find some arguments inside the option
        !
        k = index( s, "=" )
        if ( k == 0 ) k = index( s, ":" )
        if ( k /= 0 ) then
          ARG%option(n) = s( :k-1)      ! the case of -N=30 or -N:30
          ARG%string(n) = s(k+1: )
          read(s(k+1: ), *, iostat=status) ARG%value(n)
        else
          read(s(3: ), *, iostat=status) ARG%value(n)
          if ( ARG%value(n) /= ZERO ) then
            ARG%option(n) = s(1:2)      ! the case of -N30
            ARG%string(n) = s(3: )
          else if ( s(3:3) == "0" ) then
            ARG%option(n) = s(1:2)      ! the case of -N0
            ARG%string(n) = s(3: )
          else
            ARG%option(n) = s           ! the case of -N
            ARG%string(n) = ""
          end if
        end if
      else
        !
        ! there is an argument before next option
        !
        ARG%option(n) = s               ! the case of -N 30
        ARG%string(n) = s2
        read(s2, *, iostat=status) ARG%value(n)
        i = i + 1
      end if
    else
      !
      ! no option sign
      !
      ARG%option(n) = ""
      ARG%string(n) = s
      read(s, *, iostat=status) ARG%value(n)
    end if


    s3 = ARG%string(n)
    option_argument: if( s3(1:1) == "(") then
      call process_list()
    else
      !
      ! check the range such as "30-40", "30:40", "30~40"
      ! exclude exponential expressions like "1e-10" or "1E-10"
      !
      k = scan( s3, "-~,:;" )
      if ( k /= 0 ) then
        if ( s3(k-1:k) /= "e-" .and. s3(k-1:k) /= "E-" ) then
          read(s3( :k-1), *, iostat=status) ARG%value(n)
          read(s3(k+1: ), *, iostat=status) ARG%value2(n)
        end if
      end if
    end if option_argument
    !
    ! set ARG%yes
    !
    ARG%yes(n) = determine_yes_no( ARG%string(n) )
  end do
  ARG%N = n - 1
  return

contains

subroutine process_list
    integer :: k, pos, prev_pos
    integer :: nvalues 
    real(long) :: value
    opening: if( s3(1:1) == "(") then
      pos = index( s3, ")" )
      if ( pos == 0 ) then
        call error_message( "a list of values starting with ( must end with )")
      end if
      !
      ! how many values are in the list ( count kommas !)
      !
      pos = 1
      nvalues = 0
      do
        prev_pos = pos
        pos = index( s3(pos:), "," )
        if( pos .eq. 0) exit
        ! increment the number of values in the list
        nvalues = nvalues+1
        pos = prev_pos + pos
      end do
      if( nvalues .eq. 0 ) then
        call error_message( "a list of values must consist of ','" // &
                            " separated items" )
      end if
      nvalues = nvalues + 1 ! because we counted  the number of ','
      !
      ! allocate some memory for the list
      !
      allocate( ARG%val_arr(n)%ptr( nvalues ))
      !
      ! convert every single value from the list --> and put into ARG%val_arr(n)
      !
      pos = 1
      do k=1,nvalues
        read(s3(pos:), *, iostat=status) value
        ARG%val_arr(n)%ptr( k ) = value

        prev_pos = pos
        pos = index( s3(pos:), "," )
        if( pos .eq. 0) exit
        pos = prev_pos + pos
      end do

    end if opening
end subroutine process_list

end subroutine get_options


!----------------------------------------------------------------------------
! show all options from the input file and the command line
!----------------------------------------------------------------------------
subroutine show_input( ARG )
  type(Argument), intent(in) :: ARG
  integer :: i
  do i = 1, ARG%N
    if ( ARG%option(i) == "-" ) then
      write(STDERR, "(a)") trim( ARG%string(i) )
    else if ( ARG%option(i)(1:1) == "-" ) then
      write(STDERR, "(a,'= ',a)") trim( ARG%option(i)(2:) ), &
                                  trim( ARG%string(i) )
    else if ( ARG%option(i)(1:1) == "+" ) then
      write(STDERR, "('# command option: ',a,' ',a)") trim( ARG%option(i) ), &
                                                      trim( ARG%string(i) )
    else
      write(STDERR, "('# command option: ',a)") trim( ARG%string(i) )
    end if
  end do
end subroutine show_input


!----------------------------------------------------------------------------
! get one option value or string from ARG
!   The first chracter of ARG%option(i) should be + or -.
!   That's why it is taken by ARG%option(i)(2:).
!----------------------------------------------------------------------------
subroutine parse_option_string( ARG, option_name, string )
  type(Argument), intent(in) :: ARG
  character(len=*), intent(in) :: option_name
  character(len=*), intent(out) :: string
  integer :: i
  do i = 1, ARG%N
    if ( ARG%option(i)(2:) == option_name ) string = ARG%string(i)
  end do
end subroutine parse_option_string

subroutine parse_option_real( ARG, option_name, value, value2 )
  type(Argument), intent(in) :: ARG
  character(len=*), intent(in) :: option_name
  real(long), intent(out) :: value
  real(long), intent(out), optional :: value2
  integer :: i
  do i = 1, ARG%N
    if ( ARG%option(i)(2:) == option_name ) then
      value = ARG%value(i)
      if ( present(value2) ) value2 = ARG%value2(i)
    end if
  end do
end subroutine parse_option_real

subroutine parse_option_integer( ARG, option_name, value, value2 )
  type(Argument), intent(in) :: ARG
  character(len=*), intent(in) :: option_name
  integer, intent(out) :: value
  integer, intent(out), optional :: value2
  integer :: i
  !
  ! ARG%value(:) and ARG%value2(:) are real, but value and value2 are integer
  ! Thus, INT = int( REAL, kind(INT) ) enforces conversion real -> int
  !
  do i = 1, ARG%N
    if ( ARG%option(i)(2:) == option_name ) then
      value = int( ARG%value(i), kind(value) )
      if ( present(value2) ) value2 = int( ARG%value2(i), kind(value2) )
    end if
  end do
end subroutine parse_option_integer

subroutine parse_option_logical( ARG, option_name, yes )
  type(Argument), intent(in) :: ARG
  character(len=*), intent(in) :: option_name
  logical, intent(inout) :: yes
  integer :: i
  do i = 1, ARG%N
    if ( ARG%option(i)(2:) == option_name ) &
      yes = determine_yes_no( ARG%string(i), yes )
  end do 
end subroutine parse_option_logical

function determine_yes_no( string, yes_default ) result (yes)
  logical :: yes
  character(len=*), intent(in) :: string
  logical, intent(in), optional :: yes_default
  character(len=5) :: word
  if ( present(yes_default) ) then
     yes = yes_default
  else
     yes = .false.
  end if
  read( string, "(a)" ) word
  select case ( word )
    case ( "yes", "YES", "Yes", "Y", "on", "ON", "On", &
           "true", "TRUE", "True", "T", "" );                   yes = .true.
    case ( "no", "NO", "No", "N", "off", "OFF", "Off", &
           "false", "FALSE", "False", "F" );                    yes = .false.
    case default;                                               ! no change
  end select
end function determine_yes_no

!----------------------------------------------------------------------------
! convert string into lowercase, or into uppercase
!----------------------------------------------------------------------------
subroutine convert_to_lowercase( string )
  character(len=*), intent(inout) :: string
  integer :: i, j
  character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
  do i = 1, len_trim(string)
    j = index( uppercase, string(i:i) )
    if ( j > 0 ) string(i:i) = lowercase(j:j)
  end do
end subroutine convert_to_lowercase

subroutine convert_to_uppercase( string )
  character(len=*), intent(inout) :: string
  integer :: i, j
  character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
  do i = 1, len_trim(string)
    j = index( lowercase, string(i:i) )
    if ( j > 0 ) string(i:i) = uppercase(j:j)
  end do
end subroutine convert_to_uppercase

subroutine convert_option_lowercase( ARG )
  type(Argument), intent(inout) :: ARG
  integer :: i
  do i = 1, ARG%N
    call convert_to_lowercase( ARG%option(i) )
  end do
end subroutine convert_option_lowercase

!----------------------------------------------------------------------------
! return the date string of today  ex) 26-Jun-2004
!----------------------------------------------------------------------------
function today() result (string)
  integer :: v(8), y, m, d, hh, mm, ss
  character(len=20) :: string
  character(len=3), dimension(12), parameter :: month = &
    (/ "Jan", "Feb", "Mar", "Apr", "May", "Jun", &
       "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" /)
  call date_and_time( values=v )
  y = v(1);     m = v(2);       d = v(3)
  hh = v(5);    mm = v(6);      ss = v(7)
  write(string, "(i2,'-',a3,'-',i4,x,i2,':',i2.2,':',i2.2)") &
    d, month(m), y, hh, mm, ss
end function today


!----------------------------------------------------------------------------
! Use this procedure when you want a check point.
! But, for initilization, you must put this without counter 
!      at the very begining of your program.
! ex)    check_time()
!        ...
!        check_time( c, t )
!        print *, job_time( c )
!
! Here is an issue of SYSTEM_CLOCK( count, count_rate, count_max )
!
! From the GNU Fortran Compiler webpage: 
! "Note that the millisecond resolution of the kind=4 version implies that
!  the COUNT will wrap around in roughly 25 days."
!
! In the Intel Fortran Compiler, count_rate is 10000 for integer(4).
! Since the range of integer(4) is -2147483648 ~ 2147483647,
! count in integer(4) can handle only 429496 seconds, i.e., about 5 days!
!
! To avoid this issue, let's use the kind=8 version, i.e., integer(8).
!----------------------------------------------------------------------------
subroutine check_time( c, t )
  integer*8, intent(out), optional :: c ! used for system_clock
  real, intent(out), optional :: t      ! used for CPU_time
  integer*8 :: c1
  real :: t1
  integer*8, save :: c0 = 0
  real, save :: t0 = ZERO
  logical, save :: yes_started = .false.
  call system_clock( c1 )       ! it must give real elapsed time
  call CPU_time( t1 )           ! it must give user(CPU) time, but not with PGI
  if ( present(c) .and. present(t) ) then
    if ( .not. yes_started ) &
      call error_message( "check_time should be used without counter " // &
                          "at the begining of your program." )
    c = c1 - c0
    t = t1 - t0
  else
    yes_started = .true.        ! initialization: save c1 and t1 into c0 and t0
    c0 = c1
    t0 = t1
  end if
end subroutine check_time


!----------------------------------------------------------------------------
! return the time string converted from counter
! ex) counter=100000000 with integer(8)
!     --> 100 sec --> 1 minute 40.00 seconds
!----------------------------------------------------------------------------
function job_time_integer8( counter ) result (string)
  integer*8, intent(in) :: counter
  integer*8 :: c, c_rate
  integer :: hh, mm, mss, time
  real :: ss
  character(len=60) :: string
  call system_clock( c, c_rate )
  mss = int( mod( counter, c_rate ), kind(mss) )  ! conversion int8 -> int4
  time = int( counter / c_rate, kind(time) )      ! conversion int8 -> int4
                                ss = mod( time, 60 ) + real(mss) / real(c_rate)
  time = time / 60;             mm = mod( time, 60 )
  time = time / 60;             hh = time
  string = get_job_time( hh, mm, ss )
end function job_time_integer8

function job_time_real( time ) result (string)
  real, intent(in) :: time
  integer :: hh, mm, itime
  real :: ss
  character(len=60) :: string
  itime = int(time) / 60;       mm = mod( itime, 60 )
  itime = itime / 60;           hh = itime
  ss = time - real( hh*60*60 + mm*60 )
  string = get_job_time( hh, mm, ss )
end function job_time_real

function get_job_time( hh, mm, ss ) result(string)
  character(len=60) :: string
  integer :: hh, mm
  real :: ss
  if ( hh /= 0 ) then
    write(string, "(i8,x,a,x,i2,x,a,x,f5.2,x,'seconds')") &
      hh, trim( plural( "hour", hh ) ), mm, trim( plural( "minute", mm ) ), ss
    string = adjustl( string )
  else if ( mm /= 0 ) then
    write(string, "(i2,x,a,x,f5.2,x,'seconds')") &
      mm, trim( plural( "minute", mm ) ), ss
  else
    write(string, "(f5.2,x,'seconds')") ss
  end if
  string = trim(string) // " on " // trim( hostname() )
contains
  function plural( s, i ) result ( p )
    character(len=*), intent(in) :: s
    integer, intent(in) :: i
    character(len=len(s)+1) :: p
    p = s
    if ( i > 1 ) p = trim(p) // "s"
  end function plural
end function get_job_time


!----------------------------------------------------------------------------
! return hostname from environment variable of HOSTNAME
!----------------------------------------------------------------------------
function hostname() result (string)
  character(len=30) :: string
  call hostnm( string )
  if ( string == "" ) call getenv( "HOSTNAME", string )
  if ( string == "" ) call getenv( "HOST", string )
end function hostname


!----------------------------------------------------------------------------
! print out CPU_time, current time, and hostname
!   With PGI compiler, CPU_time gives real time instead of CPU time.
!   What can I do for it? Huh?
! USAGE)  call init_time()
!         ...
!         call footnote()
!----------------------------------------------------------------------------
subroutine init_time()
  call check_time()
end subroutine init_time

subroutine footnote( IOUT )
  integer, intent(in), optional :: IOUT
  integer :: OUTPUT
  integer*8 :: c        ! used for system_clock
  real :: t             ! used for CPU_time
  character(len=80) :: string_today, string_real, string_CPU
  OUTPUT = STDOUT;  if ( present(IOUT) ) OUTPUT = IOUT
  call check_time( c, t )
  string_today = today()
  string_real  = job_time( c )
  string_CPU   = job_time( t )
  write(OUTPUT, "('#'/'# Current date/time  : ',a)") trim(string_today)
  write(OUTPUT, "('# elapsed real time  : ',a)") trim(string_real)
  write(OUTPUT, "('# Job time(CPU_time) : ',a)") trim(string_CPU)
  !
  ! for OpenMP machines
  !
  !call getenv( "OMP_NUM_THREADS", string )
  !write(OUTPUT, "('# Job time(CPU_time) : ',a,' (x ',a,' CPUs)')") &
  !  trim(string_CPU), trim(string)
end subroutine footnote


!subroutine flush( OUTPUT )
!  integer, intent(in) :: OUTPUT
!  character(len=512) :: filename
!  logical :: yes_file
!  inquire( unit=OUTPUT, named=yes_file, name=filename )
!  close( OUTPUT )
!  if ( yes_file .and. filename /= "stdout" ) then
!    open( OUTPUT, file=trim(filename), status="old", position="append" )
!  else
!    open( OUTPUT )
!  end if
!end subroutine flush

end module stdio
