!----------------------------------------------------------------------------
! transition.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2020 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2017 Sang-Kil Son <sangkil.son@cfel.de>
!          2014-2020 Ludger Inhester <ludger.inhester@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! HFS calculation for the atomic system using the GPS discretization
!   written by Son, Sang-Kil in Sep. 2010 at DESY
!----------------------------------------------------------------------------
module transition
  use constants
  use coord_system
  use read_param
  use wavefunction
  use decay
  use pcs
  implicit none
contains

!----------------------------------------------------------------------------
! calculate transition matrix elements of bound-free transition
!   to be used for recombination amplitudes in HHG
! collaboration with Siddharth Bhardwaj
! omega for continuum states is given by P%PE_start ... P%PE_end
!   with an energy step of P%PE_dE
!----------------------------------------------------------------------------
subroutine print_transition_bound_free( P, F )
  type(Param), intent(in) :: P
  type(WF_data), intent(inout) :: F
  real(long), allocatable :: unbound_u(:,:)
  real(long) :: u(F%N_uni), up(F%N_uni)
  integer :: i, l, lp, OUTPUT
  integer :: lmin, lmax
  real(long) :: omega, ll, S, S_A
  OUTPUT = P%OUTPUT
  !
  ! determine lmin and lmax of occupied orbitals and prepare unbound_u(:,:)
  !
  lmin = minval( abs( F%config(1:F%N_orb,2) - 1 ) )
  lmax = maxval( F%config(1:F%N_orb,2) ) + 1
  allocate( unbound_u(F%N_uni,lmin:lmax) )
  !
  ! scan omega... here, omega is a continuum-state energy, not a photon energy!
  !               even though it scans from P%PE_start to P%PE_end by P%PE_dE!
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Radial matrix elements of bound-free transition')")
  write(OUTPUT, "('# - length (L) form: ',6x,a)") "< P_n,l(r) | r | P_E,l'(r) >"
  write(OUTPUT, "('# - acceleration (A) form: ',a)") &
    "< P_n,l(r) | Z/r^2 | P_E,l'(r) >"
  write(OUTPUT, "('#',a8)", advance="no") "E(eV)"
  do i = 1, F%N_orb
    lp = F%config(i,2)
    do l = abs( lp - 1 ), lp + 1, 2
      write(OUTPUT,"(a6,'-E,',i1,a3)",advance="no") trim(WF_name(F,i)), l, "(L)"
      write(OUTPUT,"(a6,'-E,',i1,a3)",advance="no") trim(WF_name(F,i)), l, "(A)"
    end do
  end do
  write(OUTPUT, "()")
  omega = P%PE_start
  energy_scan: do
    !
    ! prepare continuum state corresponding to omega
    !
    do l = lmin, lmax
      ll = dble(l)      ! unbound.f accepts real*8 rather than integer
      call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, omega, F )
      unbound_u(:,l) = u(:)
    end do
    !
    ! calculate bound-free transition amplitudes
    !
    write(OUTPUT, "(f9.2)", advance="no") omega * AU2EV
    do i = 1, F%N_orb
      lp = F%config(i,2)
      do l = abs( lp - 1 ), lp + 1, 2
        !
        ! calculate the length form: < P_n,l(r) | r | P_E,l'(r) >
        ! u are already multiplied by r
        ! integral is evaluated by the trapezoid rule.
        ! at the end points, r(0) = 0 and u_uni(N) ~= 0
        !
        S = sum( F%u_uni(:,i) * F%r_uni(:) * unbound_u(:,l) ) * F%dr
        !
        ! calculate the accelaration form: < P_n,l(r) | Z/r^2 | P_E,l'(r) >
        ! don't include r(1), which is zero because f(1) = g(1) = 0.
        !
        S_A = sum( F%u_uni(2:,i) * unbound_u(2:,l) / F%r_uni(2:)**2 ) * F%dr
        S_A = F%Z * S_A
        write(OUTPUT, "(2es13.5)", advance="no") S, S_A
      end do
    end do
    write(OUTPUT, "()")
    !
    ! increase omega by P%PE_dE
    !
    omega = omega + P%PE_dE
    if ( omega > P%PE_end ) exit energy_scan
  end do energy_scan
  deallocate( unbound_u )
  write(OUTPUT, "(/)")          ! print two blank lines
end subroutine print_transition_bound_free

!----------------------------------------------------------------------------
! calculate transition matrix elements of bound-bound transition
!----------------------------------------------------------------------------
subroutine print_transition_bound_bound( G, F, OUTPUT )
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  integer, intent(in) :: OUTPUT
  integer :: i, ip, l, lp, l_small, N_orb
  real(long) :: T
  N_orb = F%N_subshell
  !
  ! just for testing purpose: non-relativistic cases only
  ! TAKE CARE OF ANGULAR INTEGRAL! < cos theta > = sqrt( (l+1)*(l+1) / (2l+1) ! (2l+3) )
  !
  do i = 1, N_orb
    l = F%config(i,2)
    do ip = i+1, N_orb
      lp = F%config(ip,2)
      if ( abs( l - lp ) /= 1 ) cycle
      l_small = min( l, lp )
      T = radial_integral( G, F%u(:,i) * G%r(:) * F%u(:,ip) ) * &
          sqrt( dble(l_small+1)**2 / dble(2*l_small+1) / dble(2*l_small+3) )
      write(OUTPUT, "('#  B2B ',a3,' - ',a3,' : ',es13.5,2i5)") &
        WF_name( F, i ), WF_name( F, ip ), abs(T)
    end do
  end do
end subroutine print_transition_bound_bound

!----------------------------------------------------------------------------
! calculate transition matrix elements of bound-free transition
!   to be used for molecular calculations
! omega for continuum states is given by P%PE_start ... P%PE_end
!   with an energy step of P%PE_dE
!----------------------------------------------------------------------------
subroutine print_transitiondipole_bound_free( P,G, F ,OUTPUT)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(in) :: F
  integer, intent(in) :: OUTPUT

  real(long), allocatable :: unbound_u(:,:)
  real(long) :: u(F%N_uni), up(F%N_uni)
  integer :: i, l, lp, j ,n
  integer :: lmin, lmax, count
  real(long) :: omega, ll, S

  real(long) :: u_grid(G%N), norm
  !real(long) :: u_uni(F%N_uni)
  real(long) :: r2_avg(F%N_orb)
  !
  ! calculate <r^2> to check whether the radial function is too diffuse
  ! it's diffuse if r2_avg > P%diffuse_r2max
  !
  do i = 1, F%N_orb
    r2_avg(i) = radial_integral_u( F, i, i, F%r(:)**2 )
  end do
  !
  ! determine lmin and lmax of occupied orbitals and prepare unbound_u(:,:)
  !
  lmin = 0
  lmax = F%lmax + 1
  allocate( unbound_u(F%N_uni,lmin:lmax) )
  !
  ! scan omega... here, omega is a continuum-state energy, not a photon energy!
  !               even though it scans from P%PE_start to P%PE_end by P%PE_dE!
  !
  write(P%OUTPUT, "('#')")
  write(P%OUTPUT, "('# Radial matrix elements of bound-free transition')")
  write(P%OUTPUT, "('# - length (L) form: ',6x,a)") "< P_n,l(r) | r | P_E,l'(r) >"
  write(P%OUTPUT, "('#',a8)", advance="no") "E(eV)"
  count=0

  do lp=0, F%lmax  
    do j = 1, P%nev
      if ( is_occupied( F%config(1:F%N_orb,1:3), lp, j ) ) then      
        n = lp + j        
        do i = 1, size(F%config,dim=1)
          if ( F%config(i,1) == n .and. F%config(i,2) == lp ) &
            exit
        end do
        if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
        do l = abs( lp - 1 ), lp + 1, 2
          write(P%OUTPUT,"(a6,'-E,',i1,a3)",advance="no") trim(WF_name(F,i)), l, "(L)"
          count=count+1
        end do
      else if(l > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), lp-1 , j )) then
        n = lp + j  
        do i = 1, size(F%config,dim=1)
          if ( F%config(i,1) == n-1 .and. F%config(i,2) == lp -1) &
            exit
        end do
        if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
        do l = abs( lp - 1 ), lp + 1, 2
          write(P%OUTPUT,"(a6,'-E,',i1,a3)",advance="no") trim(WF_name(F,i)), l, "(L)"
          count=count+1
        end do
      end if
    end do
  end do
  write(P%OUTPUT, "()")

  write(OUTPUT, "(i10)") count
  ! write indexing line
  do lp=0, F%lmax  
    do j = 1, P%nev
      if ( is_occupied( F%config(1:F%N_orb,1:3), lp, j ) ) then      
        n = lp + j        
        do i = 1, size(F%config,dim=1)
          if ( F%config(i,1) == n .and. F%config(i,2) == lp ) &
            exit
        end do
        if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
        do l = abs( lp - 1 ), lp + 1, 2
          write(OUTPUT, "(i4,i4,i4)",advance="no") n,lp,l
        end do
      else if(lp > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), lp-1 , j )) then
        n = lp + j
        do i = 1, size(F%config,dim=1)
          if ( F%config(i,1) == n-1 .and. F%config(i,2) == lp-1 ) &
            exit
        end do
        if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
        do l = abs( lp - 1 ), lp + 1, 2
          write(OUTPUT, "(i4,i4,i4)",advance="no") n,lp,l
        end do
      end if
    end do
  end do
  write(OUTPUT, "()")


  omega = P%PE_start
  energy_scan: do
    !
    ! prepare continuum state corresponding to omega
    !
    do l = lmin, lmax
      ll = dble(l)      ! unbound.f accepts real*8 rather than integer
      call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, omega, F )
      unbound_u(:,l) = u(:)
    end do
    !
    ! calculate bound-free transition amplitudes
    !
    write(OUTPUT, "(f9.2)", advance="no") omega * AU2EV

    do lp=0, F%lmax  
      do j = 1, P%nev
        if ( is_occupied( F%config(1:F%N_orb,1:3), lp, j ) ) then      
          n = lp + j        
          do i = 1, size(F%config,dim=1)
            if ( F%config(i,1) == n .and. F%config(i,2) == lp ) &
              exit
          end do
          if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
          do l = abs( lp - 1 ), lp + 1, 2
            S = sum( F%u_uni(:,i) * F%r_uni(:) * unbound_u(:,l) ) * F%dr
            write(OUTPUT, "(es13.5)", advance="no") S
          end do
        else if(lp > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), lp-1 , j )) then
          n = lp + j
          do i = 1, size(F%config,dim=1)
            if ( F%config(i,1) == n-1 .and. F%config(i,2) == lp-1 ) &
              exit
          end do
          if ( r2_avg(i) > P%diffuse_r2max ) cycle  ! skip if it's too diffuse
          ! additional factors for the polarization basis function: r * 1/sqrt(norm)
          u_grid(:)=F%u(:,i) * F%r
          norm=radial_integral(G,u_grid(:)**2)
          do l = abs( lp - 1 ), lp + 1, 2
            S = sum( F%r_uni(:)* F%u_uni(:,i)/sqrt(norm) * F%r_uni(:) * unbound_u(:,l) ) * F%dr
            write(OUTPUT, "(es13.5)", advance="no") S
          end do
        end if
      end do! j loop
    end do! lp loop
    write(OUTPUT, "()")
    !
    ! increase omega by P%PE_dE
    !
    omega = omega + P%PE_dE
    if ( omega > P%PE_end ) exit energy_scan
  end do energy_scan
  deallocate( unbound_u )
  write(OUTPUT, "(/)")          ! print two blank lines

end subroutine print_transitiondipole_bound_free


function is_occupied( config, l, j ) result (yes_occupied)
  logical :: yes_occupied
  integer, intent(in) :: config(:,:), l, j
  integer :: n, i
  yes_occupied = .false.
  n = l + j
  do i = 1, size(config,dim=1)
    if ( config(i,1) == n .and. config(i,2) == l ) &
      yes_occupied = .true.
  end do
end function is_occupied

!function is_diffuse( F , i, diffuse_r2max ) result (yes_diffuse)
!  logical :: yes_diffuse
!  type(WF_data), intent(in) :: F
!  integer,intent(in) :: i    
!  real(long),intent(in) :: diffuse_r2max
!  integer :: j
!  real(long) :: r_m
!  yes_diffuse = .false.
!  
!  if(F%r2_orb(i)> diffuse_r2max)then
!    yes_diffuse = .true.
!  else
!    yes_diffuse = .false.
!  end if
!
!end function is_diffuse

function has_hole( config, l, j ) result (yes_hole)
  logical :: yes_hole
  integer, intent(in) :: config(:,:), l, j
  integer :: n, i
  yes_hole = .false.
  n = l + j
  do i = 1, size(config,dim=1)
    if ( config(i,1) == n .and. config(i,2) == l ) then
      if(config(i,3) /= 2* (2*l+1))then
        yes_hole = .true.
      else
        yes_hole = .false.
      end if
      return
    end if
  end do
end function has_hole
  


!----------------------------------------------------------------------------
! calculate and print radial matrix elements for Auger transitions
!   to be used for molecular calculations
!----------------------------------------------------------------------------

subroutine print_Auger_R( P, G, F, OUTPUT)
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: OUTPUT

  real(long), external :: sixj

  real(long) :: u(F%N_uni), up(F%N_uni)


  real(long) :: u1_uni(F%N_uni)
  real(long) :: u2_uni(F%N_uni)
  real(long) :: u3_uni(F%N_uni)


  real(long) :: u1(F%N)
  real(long) :: u2(F%N)
  real(long) :: u3(F%N)
  
  

  real(long) :: u_grid(G%N), norm
  integer :: i, l
  real(long) :: T,ll,e_Auger
  real(long), allocatable :: R(:)
  real(long) :: e1,e2,e3
  integer :: n1,n2,n3
  integer :: i1,i2,i3
  integer :: j1,j2,j3
  integer :: l1,l2,l3,l4
  integer :: K
  integer :: orbitals_per_l(0:F%lmax)

  !integer :: polar1,polar2,polar3
  real(long) :: r2_avg(F%N_orb)

  real(long) :: delta_l

  !
  ! calculate <r^2> to check whether the radial function is too diffuse
  ! it's diffuse if r2_avg > P%diffuse_r2max
  !
  do i = 1, F%N_orb
    r2_avg(i) = radial_integral_u( F, i, i, F%r(:)**2 )
  end do

  allocate(R(0:3*F%lmax))
  orbitals_per_l(:)=0
  do i1 = 1, F%N_orb
     n1 = F%config(i1,1)
     l1 = F%config(i1,2)
     orbitals_per_l(l1)=orbitals_per_l(l1)+1
  end do

  write(P%OUTPUT, "('#')")
  write(P%OUTPUT, "('# Radial matrix elements for Auger transition')")
  write(P%OUTPUT, "('# ',a)") "< P_n1,l1;P_n2,l2 | r_min^K/r_max^(K+1) | P_n3,l3;P_E,l' >"
  write(P%OUTPUT, "('#',a1,3(x,2a2),x,a2,a13,a13,x)", advance="no") 'K','n1','l1','n2','l2','n3','l3','l4','value','scatt._phase'
  write(P%OUTPUT, "('#',a8)" ) "E(eV)"
  
  ! Description:
  ! The loop structure is a little complicated because we also consider polarization basis functions, here:
  ! In general index 1 and 2 describe the final holes, index 3 the initial hole, and index 4 the continuum wave function
  ! 1. On the outside loop over all angular momenta of the bound electron l1,l2,l3
  ! 2. For each angular momentum l loop over the index j that describe the orbital index within the set of 
  !    orbitals / basis functions. From the index j it is indicated whether we have in fact an actual atomic orbital
  !    or a polarization function, constructed from the atomic orbital with index j-1.
  !    The respective radial functions are copied into u1,u2,u3, respectively. Same procedure with orbital eigenvalues e1,e2,e3. 
  !    The main quantum number n is derived from index j as well.
  ! 3. Loop over continuum angular momentum l4.
  !    For each angular momentum l4 and the calculated auger energy e_Auger calculate the continuum wave function
  ! 4. Loop over the 'angular momentum transfer' K, L
  !    and calculate all matrix elements R(K,n1,l1,n2,l3,n3,l3,n4,l4)
  !    As the matrix elements is independent of L, skip computation of R if it has already been calculated for a specific K
  !

  do l1=0, F%lmax  
    do l2=0, F%lmax 
      do l3=0, F%lmax
        
        do j1 = 1, P%nev
          if ( is_occupied( F%config(1:F%N_orb,1:3), l1, j1 ) ) then
            n1 = l1 + j1
            do i1 = 1, size(F%config,dim=1)
              if ( F%config(i1,1) == n1 .and. F%config(i1,2) == l1 ) &
                exit
            end do
            if ( r2_avg(i1) > P%diffuse_r2max ) cycle  ! skip if too diffuse
            u1_uni(:)=F%u_uni(:,i1)
            u1(:)=F%u(:,i1)
            e1=F%e_orb( i1 ) 
          else if(l1 > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), l1-1 , j1 )) then
            n1 = l1 + j1
            do i1 = 1, size(F%config,dim=1)
              if ( F%config(i1,1) == n1-1 .and. F%config(i1,2) == l1-1 ) &
                exit
            end do
            if ( r2_avg(i1) > P%diffuse_r2max ) cycle  ! skip if too diffuse
            u_grid(:)=F%u(:,i1) * F%r
            norm=radial_integral(G,u_grid(:)**2)
            u1_uni(:)=F%r_uni(:)*F%u_uni(:,i1)/sqrt(norm)
            u1(:)=F%r(:)*F%u(:,i1)/sqrt(norm)
            if( u1(maxloc(abs(u1),1)) * u1_uni(maxloc(abs(u1_uni),1)) < 0 )&
                 stop "u and u_uni have different sign!"
            e1=F%e_orb( i1 )
          else
            cycle 
          end if

          do j2 = 1, P%nev
            if ( is_occupied( F%config(1:F%N_orb,1:3), l2, j2 ) ) then
              n2 = l2 + j2
              do i2 = 1, size(F%config,dim=1)
                if ( F%config(i2,1) == n2 .and. F%config(i2,2) == l2 ) &
                  exit
              end do
              if ( r2_avg(i2) > P%diffuse_r2max ) cycle  ! skip if too diffuse
              u2_uni(:)=F%u_uni(:,i2)
              u2(:)=F%u(:,i2)
              e2=F%e_orb( i2 ) 
            else if(l2 > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), l2-1 , j2 )) then
              n2 = l2 + j2
              do i2 = 1, size(F%config,dim=1)
                if ( F%config(i2,1) == n2-1 .and. F%config(i2,2) == l2-1 ) &
                  exit
              end do
              if ( r2_avg(i2) > P%diffuse_r2max ) cycle  ! skip if too diffuse
              u_grid(:)=F%u(:,i2) * F%r
              norm=radial_integral(G,u_grid(:)**2)
              u2_uni=F%r_uni(:)*F%u_uni(:,i2)/sqrt(norm)
              u2=F%r(:)*F%u(:,i2)/sqrt(norm)
              if( u2(maxloc(abs(u2),1)) * u2_uni(maxloc(abs(u2_uni),1)) < 0 )&
                   stop "u and u_uni have different sign!"

              e2=F%e_orb( i2 )
            else
              cycle 
            end if
            if (j1 < j2)cycle ! to avoid unnecessary calculations
            do j3 = 1, P%nev
              if ( is_occupied( F%config(1:F%N_orb,1:3), l3, j3 ) ) then                
                n3 = l3 + j3
                do i3 = 1, size(F%config,dim=1)
                  if ( F%config(i3,1) == n3 .and. F%config(i3,2) == l3 ) &
                    exit
                end do
                if ( r2_avg(i3) > P%diffuse_r2max ) cycle  ! skip if too diffuse
                u3_uni(:)=F%u_uni(:,i3)
                u3(:)=F%u(:,i3)
                if( u3(maxloc(abs(u3),1)) * u3_uni(maxloc(abs(u3_uni),1)) < 0)&
                     stop "u and u_uni have different sign!"

                e3=F%e_orb( i3 ) 
              else if(l3 > 0 .and. P%yes_polarization .and. is_occupied( F%config(1:F%N_orb,1:3), l3-1 , j3 ) )then
                cycle
                ! No core hole in polarization function
                !                       n3 = l3 + j3
                !                       do i3 = 1, size(F%config,dim=1)
                !                          if ( F%config(i3,1) == n3-1 .and. F%config(i3,2) == l3-1 ) &
                !                               exit
                !                       end do
                !                       u_grid(:)=F%u(:,i3) * F%r(:)
                !                       norm=radial_integral(G,u_grid(:)**2)
                !                       u=F%r_uni(:)*F%u_uni(:,i3)/sqrt(norm)                       
                !                       e3=F%e_orb( i3 ) 
              else
                cycle 
              end if
              
              if(.not. has_hole( F%config(1:F%N_orb,1:3), l3, j3 ))&
                cycle

              e_Auger = e1 + e2 - e3 
              if ( e_Auger <= ZERO ) cycle   ! skip with energy criterion
              
              do l4=0,l1 + l2 + l3
                ll = dble(l4)               ! unbound.f accepts real*8 rather than integer
                call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, e_Auger, F )
                ! here we have to add the -l*pi/2 to the phase
                ! to get the consistent phase between two ang. momenta
                delta_l= modulo(get_phase_shift(u, up, F%r_uni, F%N_uni, l4, sqrt(TWO*e_Auger), F%q, "Auger")-dble(l4)*HALF*PI,TWO*PI)
                
                R(:) = ZERO
                do L = abs(l1-l2), l1+l2 
                  do K = 0, l1+l2+l3
                    T = ZERO
                    T = reduced_matrix( K, l3, l1 ) * reduced_matrix( K, l4, l2 ) * &
                      sixj( l3*2, l4*2, L*2, l2*2, l1*2, K*2 )
                    if ( T /= ZERO .and. R(K) == ZERO ) then                      

!                      R(K) = double_integral( F%N_uni, F%r_uni(:), &
!                        u1(:), u2(:), u3(:), u(:), K )
                      R(K) = double_integral_mod( K, &
                        F%r(:), F%weight(:), u1(:), u3(:), &
                        F%r_uni(:), u2_uni(:), u(:), F%rm_thres )

                      write(OUTPUT,"(x,i1,3(x,2i2),x,i2,es13.5,e13.5,' # ',es13.5)") &
                        K, n1, l1, n2, l2, n3, l3, l4, R(K), delta_l, e_Auger*AU2EV
                    end if                    
                  end do ! K loop
                end do ! L loop

                ! j1 <-> j2
                if(j1/=j2)then
                   R(:) = ZERO
                   do L = abs(l1-l2), l1+l2 
                     do K = 0, l1+l2+l3
                       T = ZERO
                       T = reduced_matrix( K, l3, l2 ) * reduced_matrix( K, l4, l1 ) * &
                         sixj( l3*2, l4*2, L*2, l1*2, l2*2, K*2 )
                       if ( T /= ZERO .and. R(K) == ZERO ) then                        
                         R(K) = double_integral_mod( K, &
                           F%r(:), F%weight(:), u2(:), u3(:), &
                           F%r_uni(:), u1_uni(:), u(:), F%rm_thres )
!                         R(K) = double_integral( F%N_uni, F%r_uni(:), &
!                           u2(:), u1(:), u3(:), u(:), K )
                      write(OUTPUT,"(x,i1,3(x,2i2),x,i2,es13.5,e13.5,' # ',es13.5)") &
                           K, n2, l2, n1, l1, n3, l3, l4, R(K), delta_l, e_Auger*AU2EV
                       end if
                     end do ! K loop
                   end do ! L loop                  
                 end if !  j1 /= j2

              end do! l4 loop
              
              
            end do ! j3 loop
          end do ! j2 loop
        end do ! j1 loop
        
      end do ! l3
    end do ! l2 
  end do ! l1

  deallocate(R)
          



end subroutine print_Auger_R




end module transition
