!----------------------------------------------------------------------------
! unbound.f
!----------------------------------------------------------------------------
! Copyright (C) 2010-2016 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2016 Sang-Kil Son <sangkil.son@cfel.de>
!          2016      Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Calculation of unbound states for a given positive energy (requires rk4.f)
!   originally written by Robin Santra (RadialWaveFunctions/Continuum/)
!
!   Son, Sang-Kil included SP in arguments in unbound, rk4, and RSE
!           to avoid common /potpar/
!   also modified to give more information when normalization fails
!   eventually you can delete q in arguments --> it's done on April 20, 2016.
!   Son, Sang-Kil changed to include F instead of SP
!----------------------------------------------------------------------------
      subroutine unbound(n,u,up,r,l,Z,energy, F, str)
      use wavefunction 
      use elec_config

      implicit none

      integer n,nnorm
      parameter (nnorm = 10)
      real*8 u(n),up(n),r(n)
      real*8 l,Z,energy
      real*8 q          ! used as an internal variable; not an argument anymore
      real*8 upara(0:3),uppara(0:3)
      real*8 uin(2),uout(2)
      real*8 A,Ap,App
      real*8 x(2),C(nnorm)
      real*8 alpha,norm,pihalf,relerr
      type(WF_data) :: F
      character(len=1), intent(in), optional :: str
      parameter (pihalf = 1.570796327d0)
      integer i, index

c     determination of expansion coefficients of wave function u(r) about r=0

! Rui's addition, following Equation (5.1) on page 81 of 
! Douglas Hartree "The calculation of atomic structure"
! ---------------------------------------------------------------------------
!      real*8 test(n)
!      real*8 v0
!      test(:) = Spline_interpolate_vector(SP, r)
!      v0 = test(2) / r(2) + Z / r(2)
!      v0 = -v0
!      ==> upara(2) = Z*Z - (v0 + (l+1.d0)*energy)
!      ==> upara(3) = Z*Z - (v0+(3.d0*l+4.d0)*energy)
! ---------------------------------------------------------------------------
! The above addition makes almost no change for the result, so we skip this.
      upara(0) = 1.d0
      upara(1) = -Z/(l+1.d0)
      upara(2) = Z*Z - (l+1.d0)*energy
      upara(2) = upara(2)/(l+1.d0)
      upara(2) = upara(2)/(2.d0*l+3.d0)
      upara(3) = Z*Z - (3.d0*l+4.d0)*energy
      upara(3) = -upara(3)*Z/3.d0
      upara(3) = upara(3)/(l+1.d0)
      upara(3) = upara(3)/(l+2.d0)
      upara(3) = upara(3)/(2.d0*l+3.d0)

c     determination of expansion coefficients of u'(r) about r=0

      uppara(0) = l+1.d0
      uppara(1) = upara(1)*(l+2.d0)
      uppara(2) = upara(2)*(l+3.d0)
      uppara(3) = upara(3)*(l+4.d0)

c     evaluation of unnormalized wave function at the first two grid points

      u(1) = 0.d0

      if (l.eq.0) then
         up(1) = 1.d0
      else
         up(1) = 0.d0
      end if

      u(2) = upara(3)
      u(2) = upara(2) + u(2)*r(2)
      u(2) = upara(1) + u(2)*r(2)
      u(2) = upara(0) + u(2)*r(2)
      u(2) = u(2)*r(2)**(l+1.d0)
      
      up(2) = uppara(3)
      up(2) = uppara(2) + up(2)*r(2)
      up(2) = uppara(1) + up(2)*r(2)
      up(2) = uppara(0) + up(2)*r(2)
      up(2) = up(2)*r(2)**l

c     numerical integration of radial Schroedinger equation

      do i = 2, n-1
         uin(1) = u(i)
         uin(2) = up(i)
         call rk4(uin,r(i),uout,r(i+1),l,energy,F)
         u(i+1) = uout(1)
         up(i+1) = uout(2)
      end do

c     energy normalization

      upara(0) = 2.d0*energy
      upara(1) = l*(l+1.d0)
!      upara(2) = 2.d0*q        ! upara(2) is not really used after this point
      upara(3) = upara(1)*3.d0

      i = n - nnorm
      !
      ! since F%SP_pot contains r * pot, we need a special care for
      ! extrapolation if a constant potential is needed for plasma calc.
      !
      if ( r(i) <= F%rmax ) then                        ! interpolation
        q = -Spline_interpolate( F%SP_pot, r(i) )       ! q means -r * pot
      else                                              ! extrapolation
        !
        ! F%q is the negative value of Spline_interpolate(SP,r=F%rmax)
        ! - for isolated atom calc., q should be constant (pot ~ - q / r)
        ! - for plasma calc., potential should be constant if muffin-tin
        !   approximation or external environmental potential is applied
        !
        if ( F%yes_flat_tail ) then     ! plasma case (flat pot. tail)
          q = F%q / F%rmax * r(i)
        else                            ! isolated atom case
          q = F%q
        end if
      end if
      uppara(0) = upara(1)/r(i)
      uppara(1) = uppara(0) - q
      uppara(2) = uppara(1) - q
      uppara(3) = 3.d0*uppara(0) - 2.d0*q
      A = upara(0) - uppara(2)/r(i)
      Ap = 0.5d0*uppara(1)/r(i)**2
      Ap = Ap/A
      Ap = 5.d0*Ap**2
      App = 0.5d0*uppara(3)/r(i)**3
      App = App/A
      x(1) = dabs(A + Ap + App)
      x(1) = dsqrt(x(1))
      uin(1) = dsqrt(x(1))*u(i)

      norm = 0.d0
      do i = n - nnorm + 1, n
         if ( r(i) <= F%rmax ) then                     ! interpolation
           q = -Spline_interpolate( F%SP_pot, r(i) )    ! q means -r * pot
         else                                           ! extrapolation
           if ( F%yes_flat_tail ) then  ! plasma case (flat pot. tail)
             q = F%q / F%rmax * r(i)
           else                         ! isolated atom case: q constant
             q = F%q
           end if
         end if
         uppara(0) = upara(1)/r(i)       
         uppara(1) = uppara(0) - q       
         uppara(2) = uppara(1) - q       
         uppara(3) = 3.d0*uppara(0) - 2.d0*q
         A = upara(0) - uppara(2)/r(i)
         Ap = 0.5d0*uppara(1)/r(i)**2
         Ap = Ap/A
         Ap = 5.d0*Ap**2
         App = 0.5d0*uppara(3)/r(i)**3
         App = App/A
         x(2) = dabs(A + Ap + App)
         x(2) = dsqrt(x(2))
         uin(2) = dsqrt(x(2))*u(i)
         alpha = 0.5d0*(x(1) + x(2))*(r(i) - r(i-1))
         index = n + 1 - i
         C(index) = uin(1)**2 + uin(2)**2 
         C(index) = C(index) - 2.d0*uin(1)*uin(2)*dcos(alpha)
         C(index) = C(index)*pihalf
         C(index) = dabs(dsqrt(C(index))/dsin(alpha))
c         write(98,*) index,r(i),C(index)
         norm = norm + C(index)
         x(1) = x(2) 
         uin(1) = uin(2)
      end do

      norm = norm/dble(nnorm)
      relerr = dabs(norm - C(1))/norm
      if (relerr .lt. 1.d-4) then
         norm = C(1)
      else
         write(*,*) '# *** WARNING Normailization failed'
         write(*,*) "relerr = ", relerr
         write(*,*) "energy = ", energy
         write(*,*) "l = ", l
         write(*,*) "Z = ", Z
         write(*,*) "q = ", q
         write(*,*) "r = ", r(n)
         write(*,*) 'normalization failed in unbound.f'

         if ( present(str) ) then
            write(*,*) 'process = ',str
         endif

         write(*,*) 'electric config'
         if ( F%yes_relativity ) then
           write(*,'(a)')
     &     trim(config_to_string(F%config_rel(1:F%N_orb_rel,1:4)))
         else
           write(*,'(a)')
     &     trim(config_to_string(F%config(1:F%N_orb,1:3)))
         endif

         stop
      end if
      norm = 1.d0/norm

      do i = 1, n
         u(i) = u(i)*norm
         up(i) = up(i)*norm
      end do

      return 

      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

