!----------------------------------------------------------------------------
! validate_dat.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2010 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Utility to validate the process data file
!   written by Son, Sang-Kil in Dec. 2010 at DESY
!
! - check out duplicated processes
! - remove small contribution processes (w/ small rates) > P%print_threshold
!   ONLY for Auger and fluoresence rates
!----------------------------------------------------------------------------
program validate_dat
  use constants
  use stdio
  use read_param
  use elec_config
  implicit none
  type(Argument) :: ARG
  type(Param) :: P
  integer, allocatable :: p_ind(:,:)
  logical, allocatable :: yes_duplicated(:)
  real, allocatable :: p_value(:,:), p_energy(:,:)      ! just single precision!
  character(len=1), allocatable :: p_tag(:,:)
  integer, parameter :: max_process = 167
  integer :: i, j, k, l, N, N_count
  real(long) :: value, energy, factor
  character(len=1) :: tag
  integer :: INFO
  logical :: yes_validate =.true.
  factor = ZERO                 ! multiply this factor if it's nonzero
  call read_parameters( P, yes_wrong_option=.false. )
  call get_input( ARG )         ! read additional parameters
  call parse_option( ARG, "factor", factor )
  call parse_option( ARG, "validate", yes_validate )
  if ( factor /= ZERO ) then
    write(STDERR, "('# factor = ',es10.2)") factor
    write(STDERR, "('# NOTE: this factor will be multiplied for ALL values,')")
    write(STDERR, "('#       no matter whether it is cross section or rate.')")
  end if
  !
  ! allocate memory
  ! p_ind(i,0):  # of (non-duplicated) processes for the i-th configuration
  ! p_ind(i,1:): 
  !
  N = product( P%init_config(:,3) + 1 )
  write(STDERR, "('# of all configurations = ',i10)") N
  allocate( p_ind(N,0:max_process), yes_duplicated(N) )
  allocate( p_value(N,max_process), p_energy(N,max_process) )
  allocate( p_tag(N,max_process) )
  p_ind(:,:) = 0
  yes_duplicated(:) = .false.
  !
  ! read data file
  !
  open( P%IO_DATA, file=P%file_data, action="read", form="formatted" )
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value, energy
    if ( factor /= ZERO ) value = value * factor
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i == 0 .or. j == 0 .or. value == ZERO ) cycle  ! skip unmeaningful
    if ( yes_validate == .false. ) then ! do nothing but multiplying the factor
      write(STDOUT, "(a1,2i10,es13.5,x,f10.5)") tag, i, j, value, energy
      cycle
    end if
    if ( i > N .or. j > N ) then
      write(STDERR, *) "Index should be less than ", N, ": ", tag, i, j
    else if ( i < 0 .or. j < 0 ) then
      write(STDERR, *) "Index should be greater than 0: ", tag, i, j
    else if ( any( p_ind(i,1:p_ind(i,0)) == j ) ) then
      write(STDERR, *) "Duplicated: ", tag, i, j, value
      yes_duplicated(i) = .true.
    else
      k = p_ind(i,0) + 1
      p_ind(i,k) = j
      p_value(i,k) = value
      p_energy(i,k) = energy
      p_tag(i,k) = tag
      p_ind(i,0) = p_ind(i,0) + 1
      if ( p_ind(i,0) >= max_process ) then
        write(STDERR,"('Too many process for one index, max_process = ',i15)") &
          max_process
        stop
      end if
    end if
  end do
  if ( yes_validate == .false. ) call exit(0)
  N_count = count( yes_duplicated(:) )
  write(STDERR, "('# of duplicated indexes = ',i10)") N_count
  !
  ! okay, now produce a refined data file...
  ! remember that p_value and p_energy are single precision, 
  ! but it should be okay because only 5 decimal points will be printed out.
  !
  do i = 1, N
    !
    ! sort them as the ascending order
    !
    do j = 1, p_ind(i,0)
      do k = j+1, p_ind(i,0)
        if ( p_ind(i,j) > p_ind(i,k) ) then
          l = p_ind(i,j)                ! p_ind(i,j) <-> p_ind(i,k)
          p_ind(i,j) = p_ind(i,k)
          p_ind(i,k) = l

          tag = p_tag(i,j)              ! p_ind(i,j) <-> p_ind(i,k)
          p_tag(i,j) = p_tag(i,k)
          p_tag(i,k) = tag

          value = p_value(i,j)          ! p_value(i,j) <-> p_value(i,k)
          p_value(i,j) = p_value(i,k)
          p_value(i,k) = value

          energy = p_energy(i,j)        ! p_energy(i,j) <-> p_energy(i,k)
          p_energy(i,j) = p_energy(i,k)
          p_energy(i,k) = energy
        end if
      end do
    end do
    !
    ! print
    !
    do k = 1, p_ind(i,0)
      j = p_ind(i,k)
      if ( ( p_tag(i,k) == "A" .or. p_tag(i,k) == "F" ) .and. &
           ( p_value(i,k) <= P%print_threshold ) ) cycle
      write(STDOUT, "(a1,2i10,es13.5,x,f10.5)") &
        p_tag(i,k), i, j, p_value(i,k), p_energy(i,k)
    end do
  end do
  !
  ! exit code
  !
  call exit(N_count)
end program validate_dat
