!----------------------------------------------------------------------------
! wavefunction.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2021 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2021 Sang-Kil Son <sangkil.son@cfel.de>
!          2015-2020 Ludger Inhester <ludger.inhester@cfel.de>
!          2015-2018 Koudai Toyota
!          2019      Rui Jin
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! All about wavefunctions for the atomic system
!   written by Son, Sang-Kil in Jul. 2010
!   significantly modified by Son, Sang-Kil in Jun. 2014
! Implementation of relativy was done by Koudai Toyota in Apr. 2015 at CFEL
! (originally relativity.f90 but now merged here)
!
! Wavefunction in the atomic system (in the spherical coord.) is expressed as
!                                              u_nl(r)
! Psi(r,theta,phi) = R_nl(r) Y_ml(theta,phi) = ------- Y_ml(theta,phi).
!                                                 r
! Here, u_nl(r) is stored in u(:,i) where i = (n,l) and
!       (n,l) information is stored in config(i,1) -> n and config(i,2) -> l.
!
! Here is a summary of significant changes in Jun. 2014 (Rev. 1090)
! - F%wf(i,:)       --> F%u(:,i)      ; previously wf = u / r / sqrt(4PI)
! - F%rwf_uni(i,:)  --> F%u_uni(:,i)
! - F%SP_wf         --> F%SP_u
! - F%wf_all(:,ind) --> F%u_all(:,ind)
!----------------------------------------------------------------------------
module wavefunction
  use constants
  use read_param
  use coord_system
  use interpolate
  use eigen_LAPACK
  use elec_config
  implicit none
  !
  ! define WF_data
  !   u(j,i): radial wavefunction, u(r) = r R(r), 
  !           at j-th grid (j=1...N) of i-th orbital (i=1...N_orb)
  !           In many case, u can be scanned as u(:,i).
  !           Remember this column-major order is much efficient in Fortran.
  !   config(i,:): configuration information of the i-th orbital
  !     config(i,1) -> quantum number n
  !     config(i,2) -> quantum number l
  !     config(i,3) -> occupation number
  !   e_orb(i): orbital energy of the i-th orbital
  !   pot(j): potential at j-th grid
  !
  ! ACHTUNG: F%config(1:F%N_orb,:)
  !          F%u(:,1:F%N_orb)
  !          F%e_orb(1:F%N_orb)     for all occupied orbitals
  !                                 (or for the configuration given in P).
  !          all of them are extendable to 1:F%N_orb_max.
  !
  type :: WF_data
    integer :: N        ! # of nonuniform grids
    integer :: N_orb    ! # of subshells (different orbitals)
    integer :: N_orb_max        ! max. N_orb; to be used for mc_fly
    integer :: N_uni    ! # of uniform grids, internally determined
    integer :: N_HS     ! it's always N_HS = 441 (HS: Herman-Skillman)
    integer, allocatable :: config(:,:)                 ! elec. config. info
    real(long), allocatable :: e_orb(:)                 ! orbital energies
    real(long), allocatable :: u(:,:), r(:), pot(:)     ! GPS grids
    real(long), allocatable :: weight(:)                ! GPS weights
    real(long), allocatable :: r_int(:), f_int(:)       ! grids for interpol.
    real(long), allocatable :: r_HS(:),  f_HS(:)        ! Herman-Skillman
    real(long), allocatable :: r_uni(:), u_uni(:,:)     ! uniform grids
    real(long), allocatable :: f_occ(:)                 ! fractional occ. #
    real(long) :: E_tot ! total energy
    real(long) :: E0    ! total energy (zeroth order = sum of all orbital E's)
    real(long) :: Z     ! Z for HS grids, nuclear charge,  -r*pot at r=0
    real(long) :: q     ! q for HS grids, q = Z - N_e + 1, -r*pot at r=r_max
    real(long) :: dr    ! for uniform grids
    real(long) :: rmax                  ! rmax for bound states
    real(long) :: continuum_rmax        ! rmax for continuum states
    type(Spline) :: SP_u        ! used for interpolating u [ = r R(r) ]
    type(Spline) :: SP_pot      ! used for interpolating r*pot
    logical :: yes_reuse_pot    ! yes if pot is given & used for initialization
    integer :: N_iter   ! detect # of iterations; Kota's suggestion!
    logical :: yes_converged
    real(long) :: rm_thres      ! to be used to determine rmax for continuum
    !
    ! for relativistic calculations and resonance
    !
    logical :: yes_relativity, yes_resonance
    integer :: N_orb_rel
    integer, allocatable :: config_rel(:,:)
    real(long), allocatable :: e_orb_rel(:), f_occ_rel(:)
    real(long), allocatable :: Emass(:), Edar(:), Eso(:)
    integer :: N_subshell       ! # of subshells; this should be used
                                ! instead of F%N_orb or F%N_orb_rel
    !
    ! extended variables: to store all bound states 
    !   calculated by NxN matrices for l=0 upto l=lmax
    !   They are used for finite-temperature (FT) HFS, bound-bound transitions,
    !   direct two-photon ionization cross sections, or Compton scattering.
    !
    integer :: N_all, lmax
    real(long), allocatable :: e_all(:)    ! all eigen-energies, (1:N*(lmax+1))
    real(long), allocatable :: u_all(:,:)  ! all w.f.'s, (1:N,1:N*(lmax+1))
    !
    ! below are for finite-temperature HFS (and plasma calculation)
    !
    !real(long) :: r_s                   ! Wigner-Seitz radius (NOT USED)
    real(long) :: N_elec_rs             ! # of electrons within r_s
    integer :: j_s                      ! grid index corresponding to r_s
    real(long) :: e_s                   ! potential tail
    real(long) :: chem_pot              ! chemical potential
    real(long), allocatable :: n_occ(:) ! fractional occ. # for FT-HFS
    real(long), allocatable :: norm(:)  ! normalization factor within r_s
    real(long) :: Q_avg                 ! average charge state
    real(long), allocatable :: rho_free(:)      ! free elec. density
    logical :: yes_reuse_rho_free       ! yes if rho_free is used for initial.
    logical :: yes_fix_chem_pot         ! yes if chemical potential is fixed
    logical :: yes_flat_tail            ! yes if potential tail is constant
  end type
  interface radial_integral_u_uni
    module procedure radial_integral_u_uni_cont, radial_integral_u_uni
  end interface
contains

function create_WF( P, G ) result (F)
  type(WF_data) :: F
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  real(long) :: dr
  integer :: i
  !
  ! initialize yes/no variables... 
  ! this is critical! if you don't do so, those variables are randomly assigned
  !
  F%yes_reuse_pot = .false.      ! whether you want to reuse the previous pot.
  F%yes_reuse_rho_free = .false. ! whether you reuse the previous rho_free
  F%yes_converged = .false.      ! check if SCF procedure is converged or not
  F%yes_flat_tail = .false.      ! whether the system has a flat potential tail
                                 ! it will be assigned according to 
                                 ! P%yes_muffin_tin and P%yes_xpot
  F%yes_resonance = P%yes_resonance             ! assigned after input
  F%yes_relativity = P%yes_relativity           ! assigned after input
  F%yes_fix_chem_pot = P%yes_fix_chem_pot       ! assigned after input
  !
  ! set up the dimension of arrays and memory allocation
  !
  F%N_orb = P%N_orb     ! # of orbitals (shells)
  F%N = G%N             ! # of grids in GPS
  F%j_s = F%N           ! index for WS radius (used for plasma calculation)
  F%N_HS = 441          ! # of grids in the Herman-Skillman code, always = 441
  F%dr = P%dr           ! for uniform mesh
  F%rm_thres = P%rm_thres
  allocate( F%r(F%N), F%pot(F%N), F%weight(F%N) )    ! in GPS grids
  allocate( F%r_int(F%N+2), F%f_int(F%N+2) )         ! for interpol. of GPS
  allocate( F%r_HS(F%N_HS), F%f_HS(F%N_HS) )         ! in HS grids
  if ( P%yes_screening ) allocate( F%rho_free(F%N) )
  !
  ! during mc_fly including not only ionization but also excitation,
  ! N_orb can reach to N_elec as the maximum case.
  !
  ! ACHTUNG: F%n_occ vs. F%f_occ
  ! - F%f_occ(1:F%N_orb_max): used for fractional occ. number given from P
  ! - F%n_occ(1:F%N_all): used for finite-temperature calculation
  !                       calculated in plasma.f90
  !
!  F%N_orb_max = max( P%N_elec, P%N_orb ) <-- this is NOT used YET! (Dec. 2016)
  F%N_orb_max = P%N_orb
  allocate( F%config(F%N_orb_max,3) )
  allocate( F%f_occ(F%N_orb_max) )
  allocate( F%e_orb(F%N_orb_max) )
  allocate( F%u(F%N,F%N_orb_max) )
  F%config(:,:) = 0
  F%e_orb(:) = ZERO
  F%f_occ(:) = ZERO
  !
  ! for all bound states
  !
  F%lmax = P%lmax
  if ( P%yes_relativity ) then
     F%N_all = F%N * ( 2 * P%lmax + 1 ) !  factor 2 for different j index for given l
  else
     F%N_all = F%N * ( P%lmax + 1 )
  endif
  allocate( F%e_all(F%N_all), F%u_all(F%N,F%N_all) )
  if ( P%yes_average_atom ) allocate( F%n_occ(F%N_all), F%norm(F%N_all) )
  !
  ! copy configuration infomation from P
  !
  F%config(1:F%N_orb,:) = P%init_config(:,:)
  F%f_occ(1:F%N_orb) = P%init_f_occ(:)
  !
  ! for relativistic calculations
  ! obtain the number of orbitals
  !
  if ( F%yes_relativity ) then
    F%N_orb_rel = P%N_orb_rel
    allocate( F%config_rel(F%N_orb_rel, 5) ) 
    allocate( F%e_orb_rel(F%N_orb_rel) )
    allocate( F%f_occ_rel(F%N_orb_rel) )
    allocate( F%Emass(F%N_orb) ) 
    allocate( F%Edar(F%N_orb) ) 
    allocate( F%Eso(F%N_orb) ) 
    F%config_rel(:,:) = P%init_config_rel(:,:)
    F%f_occ_rel(:) = P%init_f_occ_rel(:)
    F%N_subshell = F%N_orb_rel
  else
    F%N_subshell = F%N_orb
  endif
  !
  ! copy GPS grids from G
  !
  F%r(:) = G%r(:)
  F%weight(:) = G%weight(:)
  F%rmax = G%Rmax
  !
  ! generate HS grid points according to the original HS code
  ! copied from CreaPot.f
  !===========================================================================
  !u = 0.885341377d0/F%Z**(1.d0/3.d0)
  !deltax = 0.0025d0
  !x = 0.d0
  !i = 1
  !F%r_HS(i) = u*x
  !do j = 1,11
  !   do k = 1,40
  !      i = i + 1
  !      x = x + deltax
  !      F%r_HS(i) = u*x
  !   end do
  !   deltax = 2.d0*deltax
  !end do
  !===========================================================================
  ! copied from pcs.f
  !===========================================================================
  F%r_HS(1) = 0.0d0
  DR = 0.00125d0 * 0.885341377d0 / EXP(LOG(dble(P%Z))/3.0d0)
  DO i = 2,F%N_HS       ! ngrid == 441
     IF (MOD(i,40) .EQ. 2) DR = DR + DR
     F%r_HS(i) = F%r_HS(i-1) + DR
  END DO 
  !===========================================================================
  !
  ! generate interpolation grids based on GPS grids
  ! just adding one more point at both ends
  !
  F%r_int(1) = ZERO
  F%r_int(2:F%N+1) = F%r(:)
  F%r_int(F%N+2) = P%rmax
  if ( P%rmax == ZERO ) F%r_int(F%N+2) = F%r(G%N)*TWO   ! put an enough large r
  !
  ! set up F%Z and F%q
  ! assigning F%q in the scf procedure is much safer than here; see scf.f90
  !
  F%Z = dble(P%Z)                       ! nuclear charge,   r*pot = -Z at r=0
!  F%q = dble(P%Z - P%N_elec)            ! effective charge, r*pot = -q at r_max
!  if ( P%yes_Latter .or. P%yes_OEP_KLI ) then
!    F%q = F%q + ONE                     ! self-interaction correction
!  end if
!  if ( P%yes_debye_screening ) then     ! it's supposed to be in plasma.f90...
!    F%q = F%q * exp( -F%r_int(F%N+2) / P%debye_length )
!  end if
  !
  ! set up for finite-temperature HF and plasma calculation
  !
  if ( F%yes_fix_chem_pot ) F%chem_pot = P%chem_pot_fixed
  !
  ! flat pot. tail is to be used if muffin-tin or XPOT calculation is used
  ! - P%yes_muffin_tin : attach a constant tail after the Wigner-Seitz radius
  ! - P%yes_xpot       : attach a potential tail according to P%flat_tail_type
  !
  if ( P%yes_muffin_tin .or. P%yes_xpot ) F%yes_flat_tail = .true.
end function create_WF

!----------------------------------------------------------------------------
! prepare WF for continuum states
! interpolate r*w.f. on uniform grids
!----------------------------------------------------------------------------
subroutine extend_WF_continuum( F, OUTPUT, opt_rmax )
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: OUTPUT
  real(long), intent(in), optional :: opt_rmax
  real(long) :: rmax, r_avg, r2_avg, r_m, norm_error
  integer :: i, j, N_elec
  integer, parameter :: IO_NULL = 1     ! /dev/null
  !
  ! rmax for continuum states is internally determined with F%rm_thres
  ! _or_ set from opt_rmax
  !
  if( present(opt_rmax) ) then
    rmax = opt_rmax
  else
    rmax = ZERO
    do i = 1, F%N_orb       
      if ( F%config(i,3) == 0 ) cycle   
      do j = F%N, 1, -1                 ! look over nonuniform grids
        if ( abs( F%u(j,i) ) > F%rm_thres ) exit
      end do
      if ( j == F%N ) then              ! diffuse orbital is detected!
        rmax = F%r_int(F%N+2)           ! rmax is set to P%rmax(=F%r_int(F%N+2))
        N_elec = sum( F%config(:,3) )   ! non-rel orbital info will be printed
        write(STDERR,"('# ',a)") repeat("*",34) // " WARNING " // repeat("*",34)
        write(STDERR,"('# A very diffuse orbital is detected!')")
        write(STDERR,"('# charge state : ',sp,i3)") int( F%Z - N_elec )
!TEST --- it might cause "BUS ERROR" during runtime, but it needs DOUBLE-CHECK
!        write(STDERR,"('# elec. config.: ',a)") &
!          trim( adjustl( config_to_string(F%config, yes_simple=.true.) ) )
!TEST
        write(STDERR,"('# orbital name : ',a)") trim(WF_name_nonrel(F,i))
        write(STDERR,"('# orbital amplitude, u(r_m) = ',es13.5)") F%u(j,i)
        write(STDERR,"('#                    at r_m = ',es13.5)") F%r(j)
        write(STDERR,"('#         rmax in the input = ',es13.5)") rmax
        write(STDERR,"('#              rm_threshold = ',es13.5)") F%rm_thres
        write(STDERR,"('# So, r_m is too close to rmax, which means that')")
        write(STDERR,"('# rmax is too small to describe the diffuse orbital.')")
        write(STDERR,"('# Possible solutions:')")
        write(STDERR,"('# - increase -rmax or')")
        write(STDERR,"('# - increase -rm_threshold or')")
        write(STDERR,"('# - use a fixed continuum_rmax with -continuum_rmax')")
        write(STDERR,"('# ',a)") "In the current situation, " // &
                                 "let continuum_rmax=rmax, and keep going..."
        write(STDERR,"('# ',a)") repeat( "*", 77 )
        !
        ! instead of stopping here, we may use rmax given by P%rmax,
        ! which is the maximum of maximum radius for given parameters
        !
        !stop
      else
        if ( F%r(j+1) > rmax ) rmax = F%r(j+1)  ! max. r among all subshells
      end if
    end do
  end if
  F%continuum_rmax = rmax
  !
  ! max. # of grids in uniform mesh
  ! this F%N_uni will be changed whenever this subroutine is called.
  ! therefore, it's very important to use F%u_uni(1:F%N_uni) and 
  ! F%r_uni(1:F%N_uni) to avoid unnecessary operations
  !
  F%N_uni = int( rmax / F%dr )
  if ( F%N_uni == 0 ) then
    !call error_message( "N_uni becomes zero... no bound state detected." )
    F%N_uni = 1
    write(STDERR,"('# ',a)") repeat("*",34) // " WARNING " // repeat("*",34)
    write(STDERR,"('# N_uni becomes zero, i.e., no bound state is detected.')")
    write(STDERR,"('# Probably no electron remains or -dr is too large.')")
    write(STDERR,"('# Enforce N_uni = 1.')")
    write(STDERR,"('# ',a)") repeat( "*", 77 )
  end if
  !
  ! check if this is the first time to be called
  ! ********** ACHTUNG **********
  ! this subroutine will be first called for the ground-state configuration
  ! of the neutral atom, which should be the largest rmax.
  ! rmax and F%N_uni for all other configurations will be smaller than 
  ! those for the ground-state configuration of the neutral atom.
  ! that's why it is safe to memory allocation once from the beginning.
  ! if you consider "resonantly-excited states", 
  ! then this one may not be enough!!!
  !
  if ( .not. allocated(F%r_uni) .and. .not. allocated(F%u_uni) ) then
    !
    ! memory allocation for uniform grids
    ! determine N_uni by using rmax
    !
    allocate( F%r_uni(F%N_uni), F%u_uni(F%N_uni,F%N_orb) )
    F%r_uni(:) = (/ ( F%dr * dble(j), j=0, F%N_uni-1 ) /)  ! uniform grids
    !
    ! prepare interpolation
    ! r_int(:) are already defined in create_WF() that will never be changed.
    ! f_int(:) will be changed according to what you want to fit.
    !
    F%f_int(:) = ZERO
    F%SP_u   = Spline_create( F%r_int(:), F%f_int(:) )
    F%SP_pot = Spline_create( F%r_int(:), F%f_int(:) )
  else
    !
    ! initialize F%u_uni(:,:), which will be used only for 1:F%N_uni
    !
    if ( F%N_uni > size( F%r_uni(:) ) ) &
      call error_message( "N_uni becomes larger than originally assigned!" )
  end if
  F%u_uni(:,:) = ZERO
  !
  ! interpolate u = r * R(r) on uniform grids using Spline
  !
  do i = 1, F%N_orb
    call interpolate_u( F, F%u(:,i) )
    F%u_uni(1:F%N_uni,i) = Spline_interpolate( F%SP_u, F%r_uni(1:F%N_uni) )
  end do
  !
  ! interpolate r * pot on uniform grids using Spline
  ! this potential can be used for Auger rates, pcs, and so on
  !
  call interpolate_potential( F )
  !
  ! print uniform grid information
  ! remember that all radial wavefunctions are non-relativistic
  ! so, be careful with labelling
  !
  if ( OUTPUT == IO_NULL ) return       ! skip printing, especially for mc_fly
  write(OUTPUT, "('#'/'# Radial wavefunction information (a.u.): " // &
                "(r_m: r_max where |u(r)| > ',es8.1,')')") F%rm_thres
  write(OUTPUT, "('#   ',5x,a9,3a12)") "<r>", "<r^2>", "r_m", "norm. err"
  do i = 1, F%N_orb
    do j = F%N_uni, 1, -1
      if ( abs( F%u_uni(j,i) ) > F%rm_thres ) exit
    end do
    if ( j < 1 ) j = 1  ! avoid j=0 after full iteration of do-loop
    r_m = F%r_uni(j)
    !r_avg  = sum( F%u_uni(:,i)**2 * F%r_uni(:)    ) * F%dr
    !r2_avg = sum( F%u_uni(:,i)**2 * F%r_uni(:)**2 ) * F%dr
    r_avg  = radial_integral_u( F, i, i, F%r(:) )
    r2_avg = radial_integral_u( F, i, i, F%r(:)**2 )
    norm_error = ONE - sum( F%u_uni(1:j,i)**2 ) * F%dr
    write(OUTPUT, "('#   ',a5,f9.3,f12.3)", advance="no") &
      WF_name_nonrel( F, i ), r_avg, r2_avg
    if ( F%config(i,3) == 0 ) then
      write(OUTPUT, "(a24)") "[skip for n_occ=0]"
    else
      write(OUTPUT, "(f12.3,es12.1)") r_m, norm_error
    end if
  end do
  write(OUTPUT, "('#'/'# Uniform grid parameters for continuum states:')")
  write(OUTPUT, "('#   N_uniform             = ',i7)") F%N_uni
  write(OUTPUT, "('#   continuum_rmax(a.u.)  = ',f11.3)", advance="no") rmax
  if ( present(opt_rmax) ) then
    write(OUTPUT, "(' (external parameter)')")
  else
    write(OUTPUT, "(' (internally determined)')")
  end if
  write(OUTPUT, "('#   dr(a.u.)              = ',es10.2)") F%dr
  if ( .not. present(opt_rmax) ) then
    if ( rmax == F%r_int(F%N+2) ) then
      write(OUTPUT, "('#   ',a)") &
        "continuum_rmax is given as equal to rmax " // &
        "-- WARNING: rmax might be too small"
    else
      write(OUTPUT, "('#   threshold to determine r_m = ',es10.2)") F%rm_thres
      write(OUTPUT, "('#   ',a)") &
        "continuum_rmax is just right next to r_m in radial grid."
    end if
  end if
end subroutine extend_WF_continuum

!----------------------------------------------------------------------------
! calculate ionization energy or ionization threshold or
!           ionization potential (IP) from Koopman's theorem
! consider the potential tail if muffin-tin approx. is used (and plasma case)
!----------------------------------------------------------------------------
function ionization_E( F, i ) result (E)
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i
  real(long) :: E, e_orb
  if ( F%yes_relativity ) then
    e_orb = F%e_orb_rel(i) 
!     e_orb = F%e_orb(F%config_rel(i,5))
  else
    e_orb = F%e_orb(i)
  endif
  !
  ! consider the potential tail also for plasma environment case.
  !
  if ( F%yes_flat_tail ) then
    E = F%e_s - e_orb
  else
    E = - e_orb
  end if
end function ionization_E

function ionization_E_extended( F, ind ) result (E)
  type(WF_data), intent(in) :: F
  integer, intent(in) :: ind
  real(long) :: E
  if ( F%yes_flat_tail ) then
    E = F%e_s - F%e_all(ind)
  else
    E = - F%e_all(ind)
  end if
end function ionization_E_extended

!----------------------------------------------------------------------------
! obtain a string for the subshell (n,l) like 1s, 2s, 2p, and so on
! orbital_name() is defined in elec_config.f90
!----------------------------------------------------------------------------
function WF_name( F, i ) result (s)
  character(len=5) :: s
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i
  if ( F%yes_relativity ) then
    s = orbital_name( F%config_rel(i,1), F%config_rel(i,2), F%config_rel(i,3) )
  else
    s = orbital_name( F%config(i,1), F%config(i,2) )
  end if
end function WF_name

function WF_name_nonrel( F, i ) result (s)      ! non-relativistic: only n & l
  character(len=5) :: s
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i
  s = orbital_name( F%config(i,1), F%config(i,2) )
end function WF_name_nonrel

function WF_name_extended( F, ind ) result (s)
  character(len=5) :: s
  type(WF_data), intent(in) :: F
  integer, intent(in) :: ind
  integer :: n, l
  l = int( ind / F%N )
  n = mod( ind, F%N ) + l
  s = orbital_name( n, l )
end function WF_name_extended


!----------------------------------------------------------------------------
! interpolate_u:
!   preparation of interpolating u on uniform grids using Spline
!   * using GPS_interpolate() : it's too slow
!
! interpolate_potential: 
!   preparation of interpolating r * pot on uniform grids using Spline
!   since r*pot is not zero at the end points, GPS interpol. is not convenient.
!   at r=0, r*pot = -Z
!   at r=r_max, r*pot = -(Z-Z0+1) where Z0=# of electons
!----------------------------------------------------------------------------
subroutine interpolate_u( F, u )
  type(WF_data), intent(inout) :: F
  real(long), intent(inout) :: u(:)
  F%f_int(1) = ZERO
  F%f_int(2:F%N+1) = u(:)
  F%f_int(F%N+2) = ZERO
  call Spline_reuse( F%SP_u, F%f_int(:) )
end subroutine interpolate_u

subroutine interpolate_potential( F )
  type(WF_data), intent(inout) :: F
  F%f_int(1) = -F%Z
  F%f_int(2:F%N+1) = F%pot(:) * F%r(:)
  F%f_int(F%N+2) = -F%q ! it is calculated at the beginning of SCF; see scf.f90
  call Spline_reuse( F%SP_pot, F%f_int(:) )
end subroutine interpolate_potential


!----------------------------------------------------------------------------
! process_name: print out a kind of process, and initial and final states
!   Unfortunately, it works only for s and p orbitals which have < 10 elec.
!----------------------------------------------------------------------------
!function process_name( F, str, i, j, jp ) result (s)
!  character(len=20) :: s
!  type(WF_data), intent(in) :: F
!  character(len=*), intent(in) :: str
!  integer, intent(in) :: i
!  integer, intent(in), optional :: j, jp
!  integer :: n_max, occ_before(F%N_orb), occ_after(F%N_orb)
!  character(len=1) :: tag
!  character(len=10) :: s1, s2
!  n_max = maxval( F%config(:,1) )             ! max. n of the shells
!  occ_before(:) = F%config(:,3)               ! occupation #'s
!  occ_after(:) = occ_before(:)
!  select case ( str )
!    case ( "P", "photoionization", "photoabsorption" )  ! i ionized
!      tag = "P"
!      if ( F%config(i,1) == n_max ) tag = "V" ! valence photoionization
!      occ_after(i) = occ_after(i) - 1
!    case ( "A", "auger", "Auger" )              ! j -> i, jp ionized
!      tag = "A"
!      occ_after(i) = occ_after(i) + 1
!      if ( present(j) )  occ_after(j)  = occ_after(j)  - 1
!      if ( present(jp) ) occ_after(jp) = occ_after(jp) - 1
!    case ( "F", "fluorescence", "fluo" )        ! j -> i
!      tag = "F"
!      occ_after(i) = occ_after(i) + 1
!      if ( present(j) )  occ_after(j) = occ_after(j) - 1
!  end select
!  write( s1, "(10i1)" ) occ_before(:)
!  write( s2, "(10i1)" ) occ_after(:)
!  !
!  ! complete the process name
!  !
!  write( s, "(a,a,'_',a,' = ')" ) tag, trim(s1), trim(s2)
!end function process_name


!----------------------------------------------------------------------------
! print wavefunctions on nonuniform / uniform grids
! - print_bound(): occupied w.f. on nonuniform grids
! - print_unbound(): continuum w.f. on uniform grids
!----------------------------------------------------------------------------
subroutine print_bound( F, OUTPUT )
  type(WF_data), intent(in) :: F
  integer, intent(in) :: OUTPUT
  integer :: i, j
  real(long) :: u_sign(F%N_subshell)
  do i = 1, F%N_subshell
    u_sign(i) = sign( ONE, F%u(i,1) )   ! sign determined by the first element
  end do
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Plotting of radial bound states " // &
                "[ u(r) = r R(r) ] on nonuniform grids')")
  write(OUTPUT, "('#',a10,99a11)") "r(a.u.)", &
    (/ ( trim(WF_name(F,i)), i=1, F%N_subshell ) /)
  do j = 1, F%N
    write(OUTPUT, "(100es11.3)") F%r(j), F%u(j,1:F%N_subshell) * u_sign(1:F%N_subshell)
  end do
  write(OUTPUT, "(/)")          ! print two blank lines
end subroutine print_bound

subroutine print_unbound( F, E, OUTPUT )
  type(WF_data), intent(inout) :: F
  real(long), intent(in) :: E   ! positive (continuum) energy in a.u.
  integer, intent(in) :: OUTPUT
  real(long) :: unbound_u(F%N_uni,0:F%lmax)
  real(long) :: u(F%N_uni), up(F%N_uni)
  integer :: j, l
  real(long) :: ll
  real(long) :: phase(0:F%lmax), q
  q=-Spline_interpolate( F%SP_pot, F%r_uni(F%N_uni) )
  do l = 0, F%lmax
    ll = dble(l)        ! unbound.f accepts real*8 rather than integer
    call unbound( F%N_uni, u, up, F%r_uni, ll, F%Z, E, F )
    unbound_u(:,l) = u(:)
    phase(l)=get_phase_shift(u, up, F%r_uni, F%N_uni, l, sqrt(TWO*E), F%q, 'cont ')
  end do
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# Plotting of radial continuum states " // &
                "[ u(r) = r R(r) ] on uniform grids')")
  write(OUTPUT, "('#   continuum energy = ',f9.3,' a.u. = ',f9.2,' eV')") &
    E, E * AU2EV
  write(OUTPUT, "('#',a10)",advance="no") "phases="
  do l=0,F%lmax
     write(OUTPUT, "(es11.3)",advance="no") modulo(phase(l)-dble(l)*HALF*PI,TWO*PI)
  end do
  write(OUTPUT, "()",advance="yes")
  write(OUTPUT, "('#',a10)",advance="no") "C. Phases="
  do l=0,F%lmax
     write(OUTPUT, "(es11.3)",advance="no") coulomb_phase(q,l,sqrt(TWO*E))
  end do
  write(OUTPUT, "()",advance="yes")
  write(OUTPUT, "('#',a10,a11,99i11)") "r(a.u.)", "l=0", (/ ( l,l=1,F%lmax ) /)
  do j = 1, F%N_uni
     write(OUTPUT, "(f11.4)",advance="no") F%r_uni(j)
     do l=0,F%lmax       
        write(OUTPUT, "(es11.3)", advance="no") unbound_u(j,l)
     end do
     write(OUTPUT, "()",advance="yes")
  end do
!TEST -- for Laura's calculation
!  write(OUTPUT, "('#',10x,100es11.3)") (/ ( radial_integral_uniform( F, unbound_u(:,l)**2 ), l=0,F%lmax ) /)
!TEST
  write(OUTPUT, "(/)")          ! print two blank lines




end subroutine print_unbound

! original implementation by Dietrich Krebs,
! revised 2020 by Ludger Inhester
function get_phase_shift(u, up, r_uni, N, l, k, q, wfct_name) result(phase_avg)
  real(long), intent(in) :: u(N)
  real(long), intent(in) :: up(N)
  real(long), intent(in) :: r_uni(N)
  integer, intent(in) :: N
  integer, intent(in) :: l
  real(long), intent(in) :: k
  real(long), intent(in) :: q
  character(len=5) :: wfct_name
  
  real(long), parameter :: conv_crit = 1.d-2 ! std of phase shift (in rad)  must be less than this
  integer :: i
  real(long) :: phase, phase_avg
  real(long) :: phase_std
  real(long) :: r
  real(long) :: gr
  real(long) :: fr
  integer  :: N_avg ! number of averaging points
  integer :: pi_shift = 0

  ! N_avg contains one wavelength
  N_avg = floor( (2*PI / k) / (r_uni(N)-r_uni(N-1)))    
  phase_avg=ZERO
  phase_std=ZERO
  do i=N,max(N-N_avg+1,1),-1
    r = r_uni(i)
    ! asymotic form is sin(k * r + q / k * log(2 * k * r) - l * pi/2 + phase) 
    phase = mod(&
         datan2( (q / (k * r) + k) * u(i), up(i))&
         - k * r &
         - q / k * log(2 * k * r)&
         + dble(l)*HALF*PI, &
         TWO * PI)
    if (phase<0) phase = phase + TWO*PI
    ! Up to now the phase is determined up to modulo pi.
    ! To adjust for a shift of pi, we check for opposite sign
    fr = sin(k*r + q/k * log(2*k*r) -l * pi/2 + phase)
    gr = (q / (k * r) + k) * (cos(k*r + q/k * log(2*k*r) - l * pi/2 + phase))
    if ( (fr * u(i) < 0) .and. (gr * up(i) < 0) )&
         phase = phase + PI    
    ! We average over a number of radial points.
    ! If the phase is close to 0 or close to 2pi, the phase may jump within the averaged values
    ! to prevent this, we average over a shifted phase and later shift back
    ! if the phase is > 1.5 pi, phase is shifted by -pi
    ! if the phase is < 0.5 pi, phase is shifted by +pi
    if (i==N .and. phase > 1.5 * PI )then
      pi_shift=1
    else if (i==N .and. phase < 0.5 * PI )then
      pi_shift=-1
    end if
    if(pi_shift/=0)then
      phase = phase - pi_shift * PI
      phase = mod(phase, TWO * PI)   
    end if
    phase_avg = phase_avg + phase
    phase_std = phase_std + phase * phase
  end do
  phase_avg = mod(phase_avg/N_avg, TWO * PI)
  phase_std = phase_std/N_avg - phase_avg*phase_avg
  phase_std = sqrt(phase_std)
  if(pi_shift/=0)then
    phase_avg = phase_avg + pi_shift * PI
    phase_avg = mod(phase_avg, TWO * PI)   
  end if
  if (phase_std > conv_crit) then
    write(STDERR, "('# -------------------------------------')")
    write(STDERR, "('# ERROR in: get_phase_shift')")
    write(STDERR, "('#           phaseshift has large fluctuations. StdDev=', f10.5 ,'rad.')") phase_std
    write(STDERR, "('#           for ',A5,'orbital')") wfct_name
    write(STDERR, "('#           at final state l = ',I4,' ')") l
  end if
end function get_phase_shift


!
! Written by Ludger Inhester
! returns the coulomb phase
function coulomb_phase(q,l,k) result(phase)
  real(long), intent(in) :: q
  integer, intent(in):: l
  real(long), intent(in) :: k
  real(long) :: phase
  real(long) :: eta
  ! the coulomb function is asymptotically
  ! sin( k*r - 1/k*log(2*k*r) - l/2*pi + arg( Gamma( l + 1 + i*1/k )) )
  ! Taylor: "scattering theory"
  ! We ignore the logarithmic part and return the phase that depends
  ! on different l:  
  eta=-q/k
  phase=-HALF*l*PI + arg_gamma(l,eta)
  phase=modulo(phase,TWO*PI)
  if(phase < ZERO)phase=phase+TWO*PI
contains
  ! this is an approximation for the argument of the Gamma function
  ! Gamma(l+1+i*eta) Eq. 6 of Barata, J.C.A., Canto, L.F. & Hussein, M.S. 
  ! Braz J Phys (2011) 41: 50. https://doi.org/10.1007/s13538-011-0005-6 
  function arg_gamma(l,eta ) result(arg)
    integer, intent(in):: l
    real(long), intent(in) :: eta
    real(long) :: arg
    arg=(dble(l)+HALF) * atan( eta/dble(l+1) ) + &
      eta*( log(  sqrt( dble(l+1)**2 + eta**2 )) - ONE )&
      -eta/(dble(12)*(dble(l+1)**2+eta**2))
    arg=modulo(arg,TWO*PI)
  end function arg_gamma
  
end function coulomb_phase


!----------------------------------------------------------------------------
! back-compatibility for the Herman-Skillman (HS) code
! print information with the HS format in a specific file
! also print the same information with a more-understandable format in OUTPUT
! - print_potdat()    : print potentials to pot.dat
! - print_hswfout()   : print wavefunctions to in hswf.out
! - print_energydat() : print orbital energies to energy.dat
!----------------------------------------------------------------------------
subroutine print_potdat( F, OUTPUT )
  type(WF_data), intent(inout) :: F
  integer, intent(in) :: OUTPUT
  real(long) :: r
  integer :: i, POTDAT = 21
  open( POTDAT, file="pot.dat", form="formatted" )
  write(OUTPUT, "('#'/'# External file: pot.dat')")
  write(OUTPUT, "('# Plotting of the potential in HS grids')")
  do i = 1, F%N_HS
    r = F%r_HS(i)
    write( POTDAT, "(2e25.16)" ) r, Spline_interpolate( F%SP_pot, r )
    write( OUTPUT, "(2es16.8)" ) r, Spline_interpolate( F%SP_pot, r )
  end do
  write(OUTPUT, "(/)")
  close( POTDAT )
end subroutine print_potdat

subroutine print_hswfout( F, string, OUTPUT )
  type(WF_data), intent(inout) :: F
  character(len=*), intent(in) :: string
  integer, intent(in) :: OUTPUT
  integer :: i, j, l, occ_n, KKK, HSWFOUT = 22
  real(long) :: WWNL, q
  real(long) :: u_HS(F%N_HS,F%N_subshell)
  open( HSWFOUT, file="hswf.out", form="formatted" )
  write(OUTPUT, "('#'/'# External file: hswf.out')")
  !
  ! print out titles
  !
  write( HSWFOUT, "(a)" ) "HERMAN-S-GPS"        ! title
  write( HSWFOUT, "(a)" ) string                ! title
  write( HSWFOUT, "(f9.4)" ) F%Z                ! Z; nuclear charge
  write( HSWFOUT, "(i4)" ) F%N_subshell              ! NCSPVS; # of orbitals (shells)
  !
  ! print out interpolated w.f. in hswf.out
  !
  do i = 1, F%N_subshell
    l = F%config(i,2)
    occ_n = F%config(i,3)
    WWNL = dble(occ_n) / dble(4*l+2)
    !
    ! interpolation of the i-th w.f. on the HS grids
    !
    call interpolate_u( F, F%u(:,i) )
    do j = F%N_HS, 1, -1
      q = Spline_interpolate( F%SP_u, F%r_HS(j) )
      if ( abs(q) > 1.0e-8_long ) exit
    end do
    KKK = j
    F%f_HS(1) = ZERO
    do j = 2, KKK
      F%f_HS(j) = Spline_interpolate( F%SP_u, F%r_HS(j) )
    end do
    F%f_HS(KKK+1:) = ZERO
    write( HSWFOUT, "(i4)" ) l                  ! L(i)
    write( HSWFOUT, "(i4)" ) KKK                ! KKK(i); print all points
    write( HSWFOUT, "(f9.4)" ) WWNL             ! WWNL(i)
    do j = 1, KKK-1, 5
      write( HSWFOUT, "(1p5e14.7)" ) F%f_HS(j:j+4)
    end do
    write( HSWFOUT, "(1pe14.7)" ) F%f_HS(KKK)
    !
    ! store for printing in OUTPUT
    !
    u_HS(:,i) = F%f_HS(:)
  end do
  !  real(long) :: e_kin
  ! more understandable forms
  !
  write(OUTPUT, "('# Plotting of radial wavefunctions (u) in HS grids')")
  write(OUTPUT, "('#',a15)", advance="no") "r_HS(a.u.)"
  write(OUTPUT, "(100a16)", advance="no") &
    (/ ( trim(WF_name(F,i)), i=1, F%N_subshell ) /)
  write(OUTPUT, "()")
  do j = 1, F%N_HS
    write(OUTPUT, "(100es16.8)") F%r_HS(j), u_HS(j,:)
  end do
  write(OUTPUT, "(/)")
end subroutine print_hswfout

subroutine print_energydat( F, OUTPUT )
  type(WF_data), intent(in) :: F
  integer, intent(in) :: OUTPUT
  integer :: i, n, l, m, ENERGYDAT = 23
  real(long) :: occ, E
  open( ENERGYDAT, file="energy.dat", form="formatted" )
  write(OUTPUT, "('#'/'# External file: energy.dat')")
  write(OUTPUT, "('# Orbital Energies in the Rydberg unit:')")
  do i = 1, F%N_subshell
    n = F%config(i,1)
    l = F%config(i,2)
    m = 0
    occ = dble( F%config(i,3) )
    E = F%e_orb(i) * TWO        ! Rydberg unit
    write( ENERGYDAT, "(7x,3i1,f10.6,f14.3)" ) n, l, m, occ, E
    write( OUTPUT, "(7x,3i1,f10.6,f14.3)" ) n, l, m, occ, E
  end do
  write(OUTPUT, "(/)")
end subroutine print_energydat

!----------------------------------------------------------------------------
! print atomic data for XMDYN (activated with a "-xmdyn" option)
!----------------------------------------------------------------------------
subroutine print_xmdyn( F, G, OUTPUT )
  type(WF_data), intent(in) :: F
  type(Coordinate), intent(in) :: G
  integer, intent(in) :: OUTPUT
  integer :: i, n, l, wf_ind
  real(long) :: e_orb, e_kin, r_avg
  !
  ! use the same format in the output as much as possible
  !   [total energy in a.u.]
  !   [n] [l] [orbital E in eV] [kinetic E in eV] [avg. radius in a.u.]
  !   ...
  !
  !write(OUTPUT, "(es18.10)") F%E_tot
  write(OUTPUT, "(f16.8)") F%E_tot
  do i = 1, F%N_subshell
    if ( F%yes_relativity ) then        ! I NEED TO FIX THIS REL. IMPLEMENTATION
      n = F%config_rel(i,1)
      l = F%config_rel(i,2)
      e_orb = F%e_orb_rel(i)
      wf_ind = F%config_rel(i,5)
      e_kin = F%e_orb_rel(i) - radial_integral( G, F%u(:,wf_ind)**2 * F%pot(:) )
      r_avg = radial_integral( G, F%u(:,wf_ind)**2 * F%r(:) )
    else
      n = F%config(i,1)
      l = F%config(i,2)
      e_orb = F%e_orb(i)
      e_kin = F%e_orb(i) - radial_integral( G, F%u(:,i)**2 * F%pot(:) )
      r_avg = radial_integral( G, F%u(:,i)**2 * F%r(:) )
    endif
    if ( F%yes_flat_tail ) then ! take into account F%e_s for plasma calc.
      e_orb = e_orb - F%e_s
      e_kin = e_kin + F%e_s
    end if
    write(OUTPUT, "(2i3,2f12.2,f9.3)") n, l, e_orb*AU2EV, e_kin*AU2EV, r_avg
  end do
end subroutine print_xmdyn

!----------------------------------------------------------------------------
! print out the NAO (numerical atomic orbital) file for XMOLECULE
! (activated with a "-nao" or "-xmolecule" or "-xmol" option)
!
! EXAMPLE of the NAO basis function file:
!     200  1.0  50.0  # N_AG, L_AG, rmax_AG --> used for radial grid param.
!     2  5            # lmax and how many NAOs for given l
!     3               # the number of orbitals
!     1  0  2         # config(1,1:3) --> 1s2
!     2  0  2         # config(2,1:3) --> 2s2
!     2  1  3         # config(3,1:3) --> 2p3
!     ... (5 sets of 200 NAO values at every radial grid for l=0)
!     ... (5 sets of 200 NAO values at every radial grid for l=1)
!     ... (5 sets of 200 NAO values at every radial grid for l=2)
!     EOF
!----------------------------------------------------------------------------
subroutine print_nao( P, F, G, OUTPUT )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  type(Coordinate), intent(in) :: G
  real(long) :: u(G%N), norm
  integer, intent(in) :: OUTPUT
  integer :: i, j, l
  !
  ! print out a warning message about the Latter correction
  !
  if ( P%yes_latter ) write(OUTPUT, *) "# Latter correction is used!!!"
  !
  ! print out radial grid parameters
  !
  write(OUTPUT, *) P%N, P%L1, P%rmax
  !
  ! print out electronic configuration for this atom
  !
  write(OUTPUT, *) F%N_subshell              ! how many different radial w.f.
  do j = 1, F%N_subshell
    write(OUTPUT, *) F%config(j,1:3)    ! configuration info
  end do
  !
  ! print out numerical atomic orbitals
  !
  write(OUTPUT, *) F%lmax, P%nev        ! how many w.f. in this file
  do l = 0, F%lmax
    do j = 1, P%nev
       if(is_occupied( F%config(1:F%N_subshell,1:3), l, j ) ) then
          i = l * F%N + j
          write(OUTPUT, *) F%u_all(:,i)   ! print atomic orbitals
       else if(l > 0 .and. P%yes_polarization &
            .and. is_occupied( F%config(1:F%N_subshell,1:3), l-1 , j ) ) then
          !
          ! polar. func. (l'=l+1) is constructed by r times radial func. of l
          !
          i = ( l - 1 ) * F%N + j
          u(:) = F%u_all(:,i) * F%r(:)
          norm = radial_integral( G, u(:)**2 )
          u(:) = u(:) / sqrt(norm) ! sqrt was missing before
          write(OUTPUT, *) u(:)
       else
         ! write virtuals otherwise
         i = l * F%N + j
         write(OUTPUT, *) F%u_all(:,i)   
!          ! write zeros otherwise
!          u(:)=ZERO
!          write(OUTPUT, *) u(:)
       end if
    end do
 end do
contains
  function is_occupied( config, l, j ) result (yes_occupied)
    logical :: yes_occupied
    integer, intent(in) :: config(:,:), l, j
    integer :: n, i
    yes_occupied = .false.
    n = l + j
    do i = 1, size(config,dim=1)
      if ( config(i,1) == n .and. config(i,2) == l ) &
        yes_occupied = .true.
    end do
  end function is_occupied
end subroutine print_nao

!----------------------------------------------------------------------------
! modified from print_nao to print orbital energy and effective potential for
! xoverlap to reconstruct the atomic Hamiltonian.
!  (numerical atomic orbital) file for XMOLECULE
! (activated with a "-nao" or "-xmolecule" or "-xmol" option)
!
! EXAMPLE of the NAO basis function file:
!     200  1.0  50.0  # N_AG, L_AG, rmax_AG --> used for radial grid param.
!     2  5            # lmax and how many NAOs for given l
!     3               # the number of orbitals
!     1  0  2 Eorb_1s # config(1,1:3) --> 1s2
!     2  0  2 Eorb_2s # config(2,1:3) --> 2s2
!     2  1  3 Eorb_2p # config(3,1:3) --> 2p3
!     ... (Effective potential at every radial grid)
!     ... (5 sets of 200 NAO values at every radial grid for l=0)
!     ... (5 sets of 200 NAO values at every radial grid for l=1)
!     ... (5 sets of 200 NAO values at every radial grid for l=2)
!     EOF
!----------------------------------------------------------------------------
 subroutine print_overlap( P, F, G, OUTPUT )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  type(Coordinate), intent(in) :: G
  real(long) :: u(G%N), norm
  integer, intent(in) :: OUTPUT
  integer :: i, j, l
  character aeorb*20000
  !
  ! print out a warning message about the Latter correction
  !
  if ( P%yes_latter ) write(OUTPUT, *) "# Latter correction is used!!!"
  !
  ! print out radial grid parameters
  !
  write(OUTPUT, *) P%N, P%L1, P%rmax
  !
  ! print out electronic configuration for this atom
  !
  write(OUTPUT, *) F%N_subshell              ! how many different radial w.f.
  do j = 1, F%N_subshell
    write(OUTPUT, *) F%config(j,1:3), F%e_orb(j)  ! configuration info,E_orbital
  end do
  !
  ! print out numerical atomic orbitals
  !
  write(OUTPUT, *) F%lmax, P%nev        ! how many w.f. in this file
  write(aeorb, '(1000E20.10)')((F%e_all(l*F%N+j),j=1,P%nev),l=0,F%lmax)
  write(OUTPUT, '(A)')trim(aeorb)
  write(OUTPUT,*)F%pot(1:G%N)
  
  do l = 0, F%lmax
    do j = 1, P%nev
       if(is_occupied( F%config(1:F%N_subshell,1:3), l, j ) ) then
          i = l * F%N + j
          write(OUTPUT, *) F%u_all(:,i)   ! print atomic orbitals
       else if(l > 0 .and. P%yes_polarization &
            .and. is_occupied( F%config(1:F%N_subshell,1:3), l-1 , j ) ) then
          !
          ! polar. func. (l'=l+1) is constructed by r times radial func. of l
          !
          i = ( l - 1 ) * F%N + j
          u(:) = F%u_all(:,i) * F%r(:)
          norm = radial_integral( G, u(:)**2 )
          u(:) = u(:) / sqrt(norm) ! sqrt was missing before
          write(OUTPUT, *) u(:)
       else
         ! write virtuals otherwise
         i = l * F%N + j
         write(OUTPUT, *) F%u_all(:,i)   
!          ! write zeros otherwise
!          u(:)=ZERO
!          write(OUTPUT, *) u(:)
       end if
    end do
 end do
 contains
  function is_occupied( config, l, j ) result (yes_occupied)
    logical :: yes_occupied
    integer, intent(in) :: config(:,:), l, j
    integer :: n, i
    yes_occupied = .false.
    n = l + j
    do i = 1, size(config,dim=1)
      if ( config(i,1) == n .and. config(i,2) == l ) &
        yes_occupied = .true.
    end do
  end function is_occupied
end subroutine print_overlap

subroutine print_e_all( P, F, OUTPUT )
  type(Param), intent(in) :: P
  type(WF_data), intent(in) :: F
  integer, intent(in) :: OUTPUT
  integer :: i, j, l
  write(OUTPUT, "('#'/'# Print out all eigenvalues:')")
  write(OUTPUT, "('#',a4)", advance="no") "idx"
  if ( P%yes_average_atom ) write(OUTPUT, "(a16)", advance="no") "occ. num."
  write(OUTPUT, "(a16)") "energy (a.u.)"
  do l = 0, F%lmax
    write(OUTPUT, "('# l = ',i5)") l
    do j = 1, F%N
      i = l * F%N + j
      write(OUTPUT, "(i5)", advance="no") j
      if ( P%yes_average_atom ) write(OUTPUT,"(f16.8)",advance="no") F%n_occ(i)
      write(OUTPUT, "(es16.8)") F%e_all(i)
    end do
    write(OUTPUT, "(/)")
  end do
end subroutine print_e_all

!----------------------------------------------------------------------------
! get_occ_num(F,i):     to obtain the occupation # of the i-th subshell
! get_occ_num_max(F,i): to obtain the max. occ. # of the i-th subshell
! get_N_elec(F):        to obtain the number of electrons
!----------------------------------------------------------------------------
function get_occ_num_max( F, i ) result (num)
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i
  integer :: num
  if ( F%yes_relativity ) then
    num = F%config_rel(i,3) + 1         ! max. occ = 2j+1
  else
    num = 4 * F%config(i,2) + 2         ! max. occ = 2(2l+1)
  endif
end function get_occ_num_max

function get_occ_num( F, i ) result (num)
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i
  integer :: num
  if ( F%yes_relativity ) then
    num = F%config_rel(i,4)
  else
    num = F%config(i,3)
  endif
end function get_occ_num

function get_N_elec( F ) result (N_elec)
  type(WF_data), intent(in) :: F
  integer :: N_elec
  if ( F%yes_relativity ) then
    N_elec = sum( F%config_rel(:,4) )
  else
    N_elec = sum( F%config(:,3) )
  end if
end function get_N_elec

!-------------------------------------------------------------------------------
! Koudai Toyota's implementation copied from relativity.f90
! - SR relativity_calc_first_order_corr()
! - SR calc_mass_dar_so_term()
! - F calc_rel_orb_ene()
!-------------------------------------------------------------------------------
! 1st order corrections for the orbital energies
subroutine relativity_calc_first_order_corr( P, G, F )    
  type(Param), intent(in) :: P
  type(Coordinate), intent(in) :: G
  type(WF_data), intent(inout) :: F
  real(long) :: E_sum_orb, Eso
  integer :: l, n_occ
  integer :: i, j, OUTPUT

  OUTPUT = P%OUTPUT

  ! Emass: mass correction, Edar: Darwin term, Eso: spin-orbit coupling 
  do i = 1, P%N_orb
     ! calculate mass, Darwin, and so terms
     call calc_mass_dar_so_term( G, F%u(:,i), F%e_orb(i), F%pot, F%config(i, 2), F%Emass(i), F%Edar(i), F%Eso(i) )
  enddo

  F%e_orb_rel(:) = calc_rel_orb_ene( F%e_orb, F%config(:, 2), F%Emass, F%Edar, F%Eso, F%N_orb, F%N_orb_rel )
 
  !
  ! print energies
  !
  E_sum_orb = sum( F%e_orb_rel(:) * F%f_occ_rel(:) )
  write(OUTPUT, "('#')")
  write(OUTPUT, "('# ',a)") repeat( "=", 77 )
  write(OUTPUT, "('# Energies with relativistic correction: ')")
  write(OUTPUT, "('# - zeroth-order energy (a.u.):')")
  write(OUTPUT, "('#   SUM of E_orb_rel = ',f16.8,'  [ ',f12.2,' eV ]')") &
       E_sum_orb, E_sum_orb * AU2EV
  !
  ! update F%E0 and F%E_tot
  !   the sum of non-relativistic orbital energies (F%E0) will be replaced with
  !   the sum of relativistic orbital energies (E_sum_orb)
  !
  F%E_tot = F%E_tot - F%E0 + E_sum_orb  ! update the 1st-order-corrected E_tot
  F%E0 = E_sum_orb                      ! update the 0th-order E_tot
  !
  ! print out orbital energies
  !
  write(OUTPUT,"('#')")
  write(OUTPUT,"('# Orbital energies with relativistic correction:')")
  write(OUTPUT,"('#   ',4x,a6,a16,4a12)") "n_occ", "E(eV)", "E0(eV)", &
       "Em(eV)", "Ed(eV)", "Eso(eV)"
  i = 1
  do j = 1, F%N_orb_rel
     n_occ = int( F%f_occ_rel(j) )
     Eso = F%Eso(i)
     l = F%config_rel(j,2)
     if ( F%config_rel(j,3) == 2*l - 1 ) then
        Eso = Eso * dble(-l-1)
     else if ( F%config_rel(j,3) == 2*l + 1 ) then
        Eso = Eso * dble(l)
     end if
     write(OUTPUT,"('#   ',a4,i6,f16.8,4f12.4)") & 
          WF_name( F, j ), n_occ, F%e_orb_rel(j) * AU2EV, &
          F%e_orb(i) * AU2EV, F%Emass(i)* AU2EV, F%Edar(i) * AU2EV, Eso * AU2EV
     if ( F%config_rel(j,3) == 2*l + 1 ) i = i + 1
  end do
  !
  ! print a summary
  !
  write(OUTPUT, "('#')")
  write(OUTPUT, "('#          ',100a10)") &
       (/ ( trim(WF_name( F, j )), j=1, F%N_orb_rel ) /)
  write(OUTPUT, "('# |SCF_REL|',100f10.2)") F%e_orb_rel(:)*AU2EV
  if ( P%N_elec > 1 .or. P%yes_screening .or. .not. P%yes_exact_pot ) &
    write(OUTPUT, "('#            ',2a12)") "E(0)", "E(1)"
  write(OUTPUT, "('# |E_TOT_REL|',f12.2)", advance="no") F%E0 * AU2EV
  if ( P%N_elec > 1 .or. P%yes_screening .or. .not. P%yes_exact_pot ) &
    write(OUTPUT, "(f12.2)", advance="no") F%E_TOT * AU2EV
  write(OUTPUT, "()")
  write(OUTPUT, "('# ',a)") repeat( "=", 77 )
end subroutine relativity_calc_first_order_corr

! This subroutine calculates individual energy correction
subroutine calc_mass_dar_so_term( G, u, e_orb, pot, l, Emass, Edar, Eso )
  real(long), parameter :: alpha = ONE/SPEED_C_AU, alpha2 = alpha * alpha
  type(Coordinate), intent(in) :: G
  real(long), intent(in) :: u(:), e_orb, pot(:)
  integer, intent(in) :: l
  real(long), intent(inout) :: Emass, Edar, Eso
  real(long) :: tmp1, tmp2
  real(long), allocatable :: vec(:)

  allocate( vec(G%N) )

  ! Emass
  Emass = -alpha2 / TWO * radial_integral( G, (u(:)*(e_orb - pot(:)))**2)

  ! E_so 
  ! Partial integration is done once.
  ! Surface term does not vanish for s-state, but we do not need to know it
  ! since Eso is zero for s-state.
  !vec(:) = transform_wf_to_A( G, F%u(:,i) )
  !vec(:) = matmul( G%Del_r(:,:), vec(:) )    
  !vec(:) = transform_A_to_wf( G, vec(:) )
  vec(:) = u(:) / G%wf_coeff(:)
  vec(:) = matmul( G%Del_r(:,:), vec(:) )    
  vec(:) = vec(:) * G%wf_coeff(:)

  tmp1 = radial_integral( G, pot(:) / G%r(:)**2 * u(:)**2 ) &
       - radial_integral( G, TWO*pot(:) / G%r(:) * u(:) * vec(:) )
  tmp1 = tmp1 * alpha2 / FOUR
    
  ! preparations for Edar
  tmp2 = radial_integral( G, pot(:) * vec(:)**2 )

  if ( l == 0 ) then
     Eso = ZERO
  else
     Eso = tmp1
  endif

  ! Edar
  ! Partial integrations are done once. 
  ! 2nd derivative is eliminated by Schroedinger eq.
  tmp2 = tmp2 + radial_integral( G, pot(:) * u(:)**2 * &
                ( dble(l*(l+1)) / G%r(:)**2 - TWO * (e_orb - pot(:)) ) )
  tmp2 = tmp2 * alpha2 / FOUR
  Edar = tmp2 + tmp1
end subroutine calc_mass_dar_so_term

function calc_rel_orb_ene( e_orb, l, Emass, Edar, Eso, N_orb, N_tot ) result ( e_orb_rel )
  real(long), intent(in) :: e_orb(:), Emass(:), Edar(:), Eso(:)
  integer, intent(in) :: l(:), N_orb, N_tot
  real(long), allocatable :: e_orb_rel(:)
  integer :: i, j, jj, ll

  allocate( e_orb_rel(N_tot) )

  j = 1
  do i = 1, N_orb
     ! Add the correction terms to the 0th orbital energies
     if ( l(i) == 0 ) then
        e_orb_rel(j) = e_orb(i) + Emass(i) + Edar(i) 
        j = j + 1
     else
        ll = -l(i)-1 
        do jj = j, j+1
           e_orb_rel(jj) = e_orb(i) + Emass(i) + Edar(i) + dble(ll)*Eso(i)
           ll = l(i)
        enddo
        j = j + 2
     endif
  enddo
end function calc_rel_orb_ene

function get_non_relativistic_config( P, orbital ) result( config_non )
  type(Param), intent(in) :: P
  integer, intent(in) :: orbital(:,:)
  integer :: config_non(P%N_orb,3)
  integer :: i, j, l
  if ( P%yes_relativity ) then  ! for the relativistic case
    j = 1
    do i = 1, P%N_orb
      config_non(i, 1) = orbital(j, 1) ! principal quantum number n
      l = orbital(j, 2) ! angular momentum l
      config_non(i, 2) = l
      config_non(i, 3) = orbital(j, 4) ! occupation numbers
      j = j + 1
      if ( l /= 0 ) then
        config_non(i, 3) = config_non(i, 3) + orbital(j, 4)
        j = j + 1
      endif
    enddo
  else
    config_non(:,:) = orbital(:,:)
  endif
end function get_non_relativistic_config

!----------------------------------------------------------------------------
! Radial integral routines: 
!   Pick up the i-th and j-th radial wavefunctions
!   and calculate radial integration <i|v|j> = int r^2 [ R_i * v * R_j ] dr
!                                            = int [ r R_i * v * r R_j ] dr
!                                            = int [ u_i * v * u_j ] dr
!   Remember that all of F%u, F%u_all, F%u_uni contain a "r" factor.
!   Therefore no r^2 is needed in any of radial integral.
!   Note that these functions can be called only with "F", without "G".
!
! - for F%u on a non-uniform grid
!   USAGE)    S = radial_integral_u( F, i, j, v(:) )
!          or S = radial_integral_u( F, i, j )
!          where i and j are the index for occupied (n,l) states
!
! - for F%u_all on a non-uniform grid --> call solve_SE_all() required
!   USAGE)    S = radial_integral_u_all( F, ii, jj, v(:) )
!          or S = radial_integral_u_all( F, ii, jj )
!          where ii and jj are the index for all (n,l) states
!
! - for F%u_uni on a uniform grid     --> call extend_WF_continuum() required
!   USAGE)    S = radial_integral_u_uni( F, i, j, v_uni(:) )
!          or S = radial_integral_u_uni( F, i, j )
!                  ==> equivalent to radial_integral_u( F, i, j )
!          or S = radial_integral_u_uni( F, i, u_uni_j(:), v_uni(:) )
!          or S = radial_integral_u_uni( F, i, u_uni_j(:) )
!          where i and j are the index for occupied (n,l) states
! 
! for non-uniform-grid integration, see radial_integral() in coord_system.f90
! radial_integral_u_uni() can handle all diff. 4 cases with help of interface
!----------------------------------------------------------------------------
function radial_integral_u( F, i, j, v ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i, j   ! use F%u(:,i) and F%u(:,j)
  real(long), intent(in), optional :: v(:)
  integer :: N
  !
  ! integral performed on a non-uniform grid with F%u
  !
  if ( present(v) ) then
    N = min( F%N, size(v) )     ! # of grid points is given by F%N or v(:)
    S = radial_integral_nonuniform( F, F%u(1:N,i) * F%u(1:N,j) * v(1:N) )
  else
    S = radial_integral_nonuniform( F, F%u(:,i) * F%u(:,j) )
  end if
end function radial_integral_u

function radial_integral_u_all( F, ii, jj, v ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  integer, intent(in) :: ii, jj ! use F%u_all(:,ii) and F%u_all(:,jj)
  real(long), intent(in), optional :: v(:)
  integer :: N
  !
  ! integral performed on a non-uniform grid with F%u_all
  !
  if ( present(v) ) then
    N = min( F%N, size(v) )     ! # of grid points is given by F%N or v(:)
    S = radial_integral_nonuniform( F, F%u_all(1:N,ii) * F%u_all(1:N,jj) &
                                                       * v(1:N) )
  else
    S = radial_integral_nonuniform( F, F%u_all(:,ii) * F%u_all(:,jj) )
  end if
end function radial_integral_u_all

function radial_integral_u_uni( F, i, j, v_uni ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i, j           ! use F%u_uni(:,i) and F%u_uni(:,j)
  real(long), intent(in), optional :: v_uni(:)
  integer :: N
  if ( present(v_uni) ) then
    N = min( F%N_uni, size(v_uni) )     ! # of grid is given by v(:) or F%N_uni
    S = radial_integral_uniform( F, F%u_uni(1:N,i) * F%u_uni(1:N,j) &
                                                   * v_uni(1:N) )
  else
    S = radial_integral_uniform( F, F%u_uni(:,i) * F%u_uni(:,j) )
  end if
end function radial_integral_u_uni

function radial_integral_u_uni_cont( F, i, u_uni_j, v_uni ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  integer, intent(in) :: i              ! use F%u_uni(:,i)
  real(long), intent(in) :: u_uni_j(:)  ! and u_uni_j is given separately
  real(long), intent(in), optional :: v_uni(:)
  integer :: N  ! # of grid is given by v_uni(:) or u_uni_j(:) or F%N_uni
  !
  ! integral performed on a uniform grid with F%u_uni
  !
  if ( present(v_uni) ) then
    N = min( F%N_uni, size(v_uni) )
    S = radial_integral_uniform( F, F%u_uni(1:N,i) * u_uni_j(1:N) * v_uni(1:N) )
  else
    N = min( F%N_uni, size(u_uni_j) )
    S = radial_integral_uniform( F, F%u_uni(1:N,i) * u_uni_j(1:N) )
  end if
end function radial_integral_u_uni_cont

!----------------------------------------------------------------------------
! core integral routines
!----------------------------------------------------------------------------
function radial_integral_nonuniform( F, v ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  real(long), intent(in) :: v(:)        ! v(:) should contain the r^2 factor
  integer :: N                          ! # of grid is given by v(:) or F%N
  N = min( F%N, size(v) )
  S = sum( F%weight(1:N) * v(1:N) )     ! F%weight(:): GPS quadrature
end function radial_integral_nonuniform

function radial_integral_uniform( F, v_uni ) result (S)
  real(long) :: S
  type(WF_data), intent(in) :: F
  real(long), intent(in) :: v_uni(:)    ! v_uni(:) should contain the r^2 factor
  integer :: N_uni                      ! # of grid is given by v(:) or F%N_uni
  N_uni = min( F%N_uni, size(v_uni) )
  S = sum( v_uni(1:N_uni) )             ! summation by the trapezoid rule
  S = S - HALF * v_uni(N_uni) - HALF * v_uni(1)              
  S = S * F%dr
end function radial_integral_uniform

end module wavefunction
