!!----------------------------------------------------------------------------
!! XATOM: an integrated toolkit for X-ray and atomic physics
!!============================================================================
!
!----------------------------------------------------------------------------
! xatom.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2023 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Main author: 2010-2023 Sang-Kil Son <sangkil.son@cfel.de>
!              (see all authors in CONTRIBUTING.md)
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
!
!! HFS calculation for atoms using the GPS discretization
!!   written by Son, Sang-Kil in Jul. 2010 at DESY
!!   (modified from DFT_atom, originally written by Sang-Kil in 2004-2005)
!! - HFS: Hartree-Fock-Slater method
!! - GPS: generalized pseudospectral method
!!
!! This program consists of the following modules,
!!   constants.f90:     contains precision information and basic numbers
!!   stdio.f90:         functions to manage standard input and output
!!   atom.f90:          atomic name and mass
!!   read_param.f90:    read parameters from STDIN and/or input file
!!   gps.f90:           grid data for the radial part using the GPS method
!!   coord_[].f90:      coordinate system including derivatives and integral
!!   interpolate.f90:   linear and cubic spline interpolation
!!   legendre.f90:      legendre polynomial
!!   eigen_LAPACK.f90:  eigenvalue solver with LAPACK
!!   functional_xc.f90: exchange-correlation functionals
!!   operations.f90:    core operations such as Schroedinger Eq, func_Ts, func_J
!!   wavefunction.f90:  define wavefunction-related data
!!   SCF.f90:           self-consistent-field calculation
!!
!!   decay.f90:         fluorescence and Auger rates
!!   pcs.f90:           photoabsorption cross sections (ionization & excitation)
!!   shakeoff.f90:      shake-off branching ratios
!!   scattering.f90:    non-resonant / resonant elastic scattering form factors
!!   transition.f90:    bound-bound / bound-free transition amplitude
!!   process.f90:       define process-related data
!!
!!   git.f90:           git version number (automatically created by Makefile)
!!
!!   [followings are given by Robin Santra (and slightly modified by Sang-Kil)]
!!   anglib.f90:        angular momentum coupling coefficients
!!   unbound.f:         energy-normalized continuum wavefunction
!!   rk4.f:             4th-order Runge-Kutta integrator, needed for unbound.f
!!   RSE.f:             solving radial Schroedinger equation, needed for rk4.f
!!
!! The XATOM package consists of the following programs,
!!   xatom:             this program
!!   xatom_req:         direct solver for rate equations (REQ)
!!   xatom_mc:          Monte-Carlo solver for REQ (to be merged with xatom_req)
!!   xatom_shakeoff:    shake-off branching ratio (to be merged with xatom)
!!   xatom_mc_fly:      Monte-Carlo on-the-fly
!!   find_config:       utility to find an index for a given config., vice versa
!!   validate_dat:      utility to validate a data file
!!
!! NOTE:
!!   As of Rev. 298, DFT except HFS may not be properly implemented,
!!   so you may use them very cautiously.
!!   For extension to the 2D case (homonuclear diatomic case),
!!   you may need coord_r_x.f90.
!!
!! Refs for DFT and HFS: 
!!   Parr and Yang, "Density-Functional Theory of Atoms and Molecules", 
!!     Oxford University Press, 1989.
!!   Tong and Chu, Phys. Rev. A 55, 3406 (1997).
!!   Herman and Skillman, "Atomic Structure Calculations", Prentice-Hall, 1963.
!!
!! Refs for x-ray fluorescence rates and Auger rates:
!!   Santra, J. Phys. B 42, 023001 (2009).
!!   Walters and Bhalla, At. Data 3, 301 (1971).
!!   Bhalla, Folland, and Hein, Phys. Rev. A 8, 649 (1973).
!!
!! Papers where calculations are performed by this XATOM package:
!!   Son, Young, and Santra, Phys. Rev. A 83, 033402 (2011).
!!     - hollow-atom project; basic implementation of photoionization, Auger, 
!!       fluorescence, and scattering processes and rate equations
!!   Doumy et al., Phys. Rev. Lett. 106, 083002 (2011).
!!     - Ne project; shake-off process added
!!   Son, Chapman, and Santra, submitted.
!!     - MAD project; resonant elastic scattering (dispersion correction) added
!!   Rudek et al., in preparation.
!!     - Xe project; Monte-Carlo solver for rate eqs, elec./fluorescence spectra
!!
!! To see implementations in the XATOM package from equations in the papers,
!! > grep "SEE:" *.f90
!!----------------------------------------------------------------------------
program xatom
  use constants
  use stdio
  use read_param
  use gps
  use coord_system
  use wavefunction
  use process
  use scf
  use decay
  use pcs
  use shakeoff
  use scattering
  use transition
  use git
  use compton
  use impact_ionization         
  implicit none
  !
  ! Grid: defined in gps.f90
  ! Coordinate: defined in coord_system.f90
  !   grid data for the radial part, r
  !   values in real space and computational space
  !   first and second derivative matrix
  !
  type(Grid) :: r
  type(Coordinate) :: G
  !
  ! Param: defined in read_param.f90
  !   parameters to be used in calculation
  !   such as # of grids, DFT type, SCF parameters, etc
  !
  type(Param) :: P
  !
  ! WF_data: defined in wavefunction.f90
  !   varibles for wavefunctions, orbital energies, and potential
  !
  type(WF_data) :: F, F1, F2
  !
  ! Process_data: defined in process.f90
  !
  type(Process_data) :: S
  !
  ! Shakeoff_data: shakeoff results used for xmdyn inputs 
  !
  type(Shakeoff_data) :: SO
  !
  ! misc. variables
  !
  real(long) :: sum_rates
  character(len=15) :: string
  integer :: k, idx
  real(long) :: angle
  character(len=256) :: filename
  character(len=256) :: num_str
  real(long) :: T0, T1, T2
  !
  ! start your program with time checking!
  !
  call init_time()
  !
  ! set up initial values
  !
  P%N = 200
  P%L1 = 1.0_long
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "HFS" )
  if ( P%yes_gitversion ) stop GITVERSION
  call print_parameters( P, "electronic structure of atoms and more, " // &
                            GITVERSION )
  !
  ! build the Legendre grid
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  !
  ! build WF_data and Process_data
  !
  F = create_WF( P, G )
  if ( P%yes_relativity ) then
     S = create_Process( P%init_config_rel, P%N_proc_max )
  else
     S = create_Process( P%init_config, P%N_proc_max )
  endif
  !
  ! SCF procedure: outputs are wf, e_orb, and pot
  !
  call SCF_procedure( P, G, F )
  if ( P%yes_relativity ) call relativity_calc_first_order_corr( P, G, F )
  
  if ( P%yes_shakeoff ) then
     SO = CreateShakeOffData( P, G )
     call RegisterWF( F, SO, 'before' )
  endif

  !
  ! create pot.dat and hswf.out (same as the original HS code) and energy.dat
  ! for further calculations
  !
  if ( P%yes_HS ) then
    call print_potdat( F, P%OUTPUT )
    call print_hswfout( F, P%system, P%OUTPUT )
    call print_energydat( F, P%OUTPUT )
  end if
  !
  ! interpolate r*w.f. on uniform grids
  ! print out uniform grid information for continuum states
  !
  if ( P%yes_auger .or. P%yes_pcs &
       .or. P%yes_dispersion &
       .or. P%yes_compton &
       .or. P%yes_print_unbound &
       .or. P%yes_resonance &
       .or. P%yes_create_hole &
       .or. P%yes_transition_bound_free &
       .or. P%yes_transition_dipole_bound_free &
       .or. P%yes_transition_auger &
       .or. P%yes_impact_ionization &
     ) then  
    if ( P%continuum_rmax > ZERO ) then         ! if it is given as a parameter
      !
      ! if P%continuum_rmax is given as a parameter, then use it
      ! the default for P%continuum_rmax is 450 a.u. for compton scattering
      ! see read_param.f90 for setting the default value
      !
      call extend_WF_continuum( F, P%OUTPUT, P%continuum_rmax )
    else
      call extend_WF_continuum( F, P%OUTPUT )
    end if
  end if
  !
  ! solve all bound (and pseudo-continuum) states
  !
  if ( P%yes_auger .or. P%yes_fluorescence &
       .or. P%yes_compton &
       .or. P%yes_transition_bound_bound &
       .or. P%yes_nao &
       .or. P%yes_overlap &
       .or. P%yes_create_hole &
       .or. P%yes_impact_ionization &
       .or. P%yes_transition_auger &
       .or. P%yes_transition_dipole_bound_free &
     ) then
    call solve_SE_all( F, G )
  end if
  !=========================================================================
  ! For XMOLECULE
  !   print out NAO (numerical atomic orbitals) and stop
  !
  if ( P%yes_nao ) call print_nao( P, F, G, P%IO_NAO )
  !=========================================================================
  ! For XOVERLAP
  !   print out XOVERLAP (orbit energy, effective potential and numerical atomic orbitals) and stop
  !
  if ( P%yes_overlap ) call print_overlap( P, F, G, P%IO_OVERLAP )
  !=========================================================================
  !
  ! print out possible fluorescence rates
  !
  if ( P%yes_fluorescence ) call calculate_fluorescence_rates( P, G, F, S )
!  if ( P%yes_fluorescence ) &
!    call calculate_fluorescence_rates_extended( P, G, F, S )
  !
  ! print out possible Auger (Coster-Kronig) rates
  !
  if ( P%yes_auger ) call calculate_Auger_rates( P, F, S )
!  if ( P%yes_auger ) call calculate_Auger_rates_extended( P, F, S )
  !
  ! hole-state lifetime calculated from Auger and fluorescence rates
  !
  if ( P%yes_fluorescence .and. P%yes_auger ) then
    sum_rates = sum( S%value(1:S%N_process) )
    if ( sum_rates > ZERO ) then
      string = lifetime( sum_rates )
      write(P%OUTPUT, "('#'/'# ',11x,'Total decay rate = ',es13.5,' a.u.'," // &
        "'  ( lifetime = ',a,' )')") sum_rates, trim(string)
      write(P%OUTPUT, "('#',11x,'Total decay width = ',es13.5,' eV')") &
        sum_rates * AU2EV
    end if
  end if
  !
  ! find out possible resonant Auger recombination processes
  ! based on a given linewidth (P%linewidth)
  !
  if ( P%yes_auger_recombination ) call find_Auger_recombination( P, F )
  !
  ! print out scillator strength
  !
  if ( P%yes_oscillator_str ) call calculate_oscillator_str( P, G, F )
  !
  ! print out photoabsorption cross sections
  !
  if ( P%yes_pcs ) then
    if ( P%yes_resonance ) then ! print out both resonant and non-resonant
       call calculate_photoabsorption_CS( P, G, F, S )
    else                        ! print out only non-resonant
       call calculate_photoionization_CS( P, F, S )
    end if
  end if

  !
  ! calculate shake-off branching ratios
  !
  if ( P%yes_shakeoff ) then
     P%yes_silent = .true.
     ! make WF_data after decay
     if ( .not. P%yes_xmdyn ) then
        write(SO%OUTPUT,"(a)")"#"
        write(SO%OUTPUT,"(a)")"# Shakeoff branching ratio"
        do k = 1, S%N_process
           call ConfigProcessAndRegister( P, G, S, k, SO, 'middle' )
           call CalculateShakeoffBR( P, G, S, k, SO )
           call ShowResult( S, SO, k )
       enddo
    endif
    P%yes_silent = .false.
  end if
  !=========================================================================
  ! For XMDYN
  !   print out all atomic data and scattering data into the standard output
  !   for the "energetics" format, see print_xmdyn() in wavefunction.f90
  !   for the other formats,
  !     "output_type"  "# of lines"
  !   output_type = dat: atomic data including pcs, Auger & fluorescence rates
  !               = scd: nonresonant elastic x-ray scattering form factor, f0
  !               = mad: resonant elastic x-ray scattering, f' and f''
  !               = ixs: inelastic x-ray scattering
  !   see individual modules for each output type
  !
  if ( P%yes_xmdyn ) then
    call print_xmdyn( F, G, STDOUT )    ! print out energetics
    if ( S%N_process > 0 ) then
      write( STDOUT, "(a,i10)" ) "dat", S%N_process
      if ( P%yes_shakeoff ) then
!         do k = 1, S%N_process
!            call calculate_shakeoff_ratios( P, G, F, S, k )
!         enddo
      else
         call print_Process( S, STDOUT )   ! print pcs, Auger and fluo. rates
      endif
    end if
    !
    ! replace SCD/MAD/IXS files with STDOUT
    ! IXS2 (DSP w/ free electrons) is not necessary for XMDYN.
    ! and then, we use the same subroutines for printing out SCD/MAD/IXS data,
    ! but the corresponding subroutines are slightly modified to print out
    ! "output_type" and "# of lines".
    !
    if ( P%yes_scattering ) P%IO_SCD = STDOUT
    if ( P%yes_dispersion ) P%IO_MAD = STDOUT
    if ( P%yes_compton    ) P%IO_IXS = STDOUT
  end if
  !=========================================================================
  !
  ! print Process_data
  ! IO_DATA = 11 and file_data would be "results.dat" if not assigned
  ! file_data will be accumulated (access="append")
  !
  if ( S%N_process > 0 .and. .not. P%yes_xmdyn ) then
    open( P%IO_DATA, file=P%file_data, access="append", form="formatted" )
    call print_Process( S, P%IO_DATA )
    write( P%IO_DATA, "()" )       ! print a blank line
    close( P%IO_DATA )
  end if
  !
  ! print out coherent form factor (non-resonant elastic scattering factor)
  !
  if ( P%yes_scattering ) call calculate_coherent_form_factor( P, G, F )
  !
  ! print out resonant elastic scattering factor correction, f' and f''
  !
  if ( P%yes_dispersion ) call calculate_dispersion_correction( P, F )
  !
  ! print out all possible single-hole states and their decay rates
  !
  if ( P%yes_create_hole ) then
    open( P%IO_DATA, file=P%file_data, access="append", form="formatted" )
    call create_hole( P, G, F, S )  ! S will be changed inside create_hole()
    write( P%IO_DATA, "()" )        ! print a blank line
    close( P%IO_DATA )
  end if
  !
  ! print out potential / density / wavefunctions
  !
  if ( P%yes_print_wf ) call print_bound( F, P%OUTPUT )
  if ( P%yes_print_unbound ) call print_unbound( F, P%continuum_E, P%OUTPUT )
  !
  ! print out coulomb integrals (double radial integrals)
  !
  if ( P%yes_coulomb ) call print_double_integral( P, F )
  !
  ! print out bound-bound / bound-free transition amplitudes
  !
  if ( P%yes_transition_bound_bound ) &
    call print_transition_bound_bound( G, F, P%OUTPUT )
  if ( P%yes_transition_bound_free ) call print_transition_bound_free( P, F )
  if ( P%yes_transition_dipole_bound_free ) &
    call print_transitiondipole_bound_free( P, G, F, P%IO_TDIPOLE )
  if ( P%yes_transition_auger ) call print_Auger_R( P, G, F, P%IO_AUGER )
  !
  !=========================================================================
  ! Jans Addition for Compton Code
  !------------------------------------------------------------------------
  if ( P%yes_compton ) then
    select case ( P%compton_type )
      case ( 0 );  call print_waller_hartree( P, F )    ! Waller-Hartree
      case ( 1 ); 
         if( .not. P%compton_theta_list ) then
            call print_compton_spectrum( P, F, P%compton_theta, P%OUTPUT )  !  spectrum for ONE angle
         else
            do idx=1,size(P%compton_theta_angles)
              angle = P%compton_theta_angles(idx)
              write(num_str,'(F10.3)') angle
              filename = trim(P%filename)//'.spec_'//trim(adjustl(num_str))
              open(unit=111, file=filename, status='replace', action='write')
              call print_compton_spectrum( P, F, P%compton_theta_angles(idx), 111 )  
              close(unit=111)
            end do
         endif   
      case ( 2 );  call print_total_compton_cross_section( P, F )  ! DDSCS
  !=========================================================================
  ! John's Addition for Compton Code
  !------------------------------------------------------------------------
      case ( 3 ); call print_nonuniform_compton_structure_factor_orbital(P, F)
      case ( 4 ); call print_nonuniform_structure_factor_compton(P, F)
      case ( 5 ); call print_compton_nonuniform_DDCS(P, F)
      case ( 6 ); call print_compton_nonuniform_energy_DCS(P, F)
      case ( 7 ); call print_compton_nonuniform_angle_DCS(P, F)
    end select
  end if 

  if ( P%yes_impact_ionization ) then                                        
    select case (P%impact_ionization_type)                                   
      case ( 0 ); call print_impact_ionization_DCS_angle_res_manual(P, F)
      case ( 1 ); call print_impact_ionization_total_cross_section_manual(P,F)
      case ( 2 ); call print_impact_ionization_DDCS(P,F)
      case ( 3 ); call print_impact_ionization_DCS_energy_res(P,F)
      case ( 4 ); call print_impact_ionization_DCS_angle_res(P,F)
      case ( 5 ); call print_impact_ionization_total_cross_section(P,F)
      case ( 6 ); call print_impact_ionization_DCS_energy_res_symm(P,F)
    end select                                                               
  end if                                                                     
  !=========================================================================
  ! For FT-MBPT; currently only E(0) is implemented
  !=========================================================================  
  if ( P%yes_FT_MBPT .and. F%yes_converged ) then
    P%yes_silent = .true.
    T0 = P%T_eV
    !
    ! perform SCF at a slightly smaller T
    !
    T1 = T0 * 0.99_long;  P%T_eV = T1
    F1 = create_WF( P, G )
    F1%yes_reuse_pot = .true.
    F1%pot(:) = F%pot(:)
    call SCF_procedure( P, G, F1 )
    !
    ! perform SCF at a slightly smaller T
    !
    T2 = T0 * 1.01_long;  P%T_eV = T2
    F2 = create_WF( P, G )
    F2%yes_reuse_pot = .true.
    F2%pot(:) = F%pot(:)
    call SCF_procedure( P, G, F2 )
    !
    ! print out E0 following the FT-MBPT formulation
    ! FT-MBPT: finite-temperature many-body perturbation theory
    !
    P%T_eV = T0         ! go back to the original temperature
    call print_MBPT_E0( P, F, F1, F2, T1, T2, P%OUTPUT )
    P%yes_silent = .false.
  end if
  !
  ! print out a foot note (current time / CPU time / elapse time...)
  !
  call footnote( P%OUTPUT )
  !
  ! check if the current SCF calculation is converged or not
  !
  if ( .not. F%yes_converged ) &
    stop "*** WARNING: THE CURRENT SCF CALCULATION IS NOT CONVERGED! ***"

!  do k=1, F%N_subshell
!     if (F%yes_relativity == .true. ) then
!        write(*,'(i,1x,a,1x,4i3,1x,es13.6)')k-1,WF_name(F,k),(F%config_rel(k,l),l=1,4),F%u(F%N,F%config_rel(k,5))
!        do l=1,F%N
!           write(99,*)F%r(l),F%u(l,F%config_rel(k,5))
!        enddo
!     else
!        write(*,*)k-1,WF_name(F,k),(F%config(k,l),l=1,3),F%u(F%N,k)
!        do l=1,F%N
!           write(99,*)F%r(l),F%u(l,k)
!        enddo
!     endif
!     write(99,*)
!     write(99,*)
!  enddo


end program xatom
