!----------------------------------------------------------------------------
! xatom_imaging.f90
!----------------------------------------------------------------------------
! Copyright (C) 2016-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2016-2017 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! xatom_imaging: calculate elastic and inelastic scattering
!                including radiation damage dynamics
!   Be careful with units!
!   - foobar.ddsp_tot : the first column is energy in a.u.
!   - foobar.pop      : the first column of the second part is energy in eV
!   - omega(:)        : energy in a.u.
!   - output          : energy in a.u. --> consistent with compton output
!----------------------------------------------------------------------------
program xatom_imaging
  use constants
  use stdio
  use read_param
  use scf
  use compton
  use git
  implicit none
  type(Param) :: P
  type(Coordinate) :: G
  type(WF_data) :: F
  type(Grid) :: r
  integer :: i, j, k, charge, N, E_eV, N_orb
  integer :: IO_COMPTON = 23
  integer :: OUTPUT
  real(long) :: pop, omega_in, omega_temp
  character(len=512) :: line, config_str, file_ddsp_tot
  real(long), allocatable :: omega(:), theta(:)
  real(long), allocatable :: DDSCS_value(:), DDSCS(:,:), TSCS(:)
  real(long), allocatable :: DDSCS_free(:,:)
  integer, allocatable :: config(:,:)
  real(long), allocatable :: f_occ(:,:)
  integer :: status
  integer*8 :: int_time
  real :: timer, timer_old
  logical :: yes_exist
  !
  ! start your program with time checking!
  !
  call init_time()
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "HFS" )
  if ( P%yes_gitversion ) stop GITVERSION
  call print_parameters( P, "elastic and inelastic scattering of atoms, " // &
                            GITVERSION )
  OUTPUT = P%OUTPUT
  !
  ! set up a filename to be read all the time
  !
  file_ddsp_tot = trim(P%filename) // ".ddsp_tot"
  !
  ! memory allocation: 2 times of a given range for omega(:) and DDSCS(:,:)
  !   calculated DDSCS's are descendingly ordered
  !   so, DDSCS(N+1:2*N) --> downward shifts: read from files
  !       DDSCS(1:N)     --> upward shifts: for free-electrons
  !
  N = P%compton_N_omega
  allocate( omega(N*2), theta(P%compton_N_theta) )
  allocate( DDSCS(N*2,P%compton_N_theta) )
  allocate( DDSCS_free(N*2,P%compton_N_theta) )
  allocate( DDSCS_value(P%compton_N_theta) )
  allocate( TSCS(P%compton_N_theta) )
  DDSCS(:,:) = ZERO
  !
  ! calculate Thomson scattering cross sections
  ! BE CAREFUL: P%compton_*_theta are in degree!
  !
  omega_in = P%PE_start
  do j = 1, P%compton_N_theta
    theta(j) = ( P%compton_min_theta + dble(j-1) * P%compton_d_theta ) * &
               PI / 180_long
    TSCS(j) = Thomson_CS( theta(j) )    ! defined in compton.f90
  end do
  !
  ! extend omega(:) to higher energy shifts
  ! BE CAREFUL: P%compton_d_omega seems to be given in eV!
  ! BE CAREFUL: omega in *.ddsp_* is given in a descending order
  ! 
  ! omega(N+1:2*N) are the same as omega_s(:) in compton.f90
  ! and omega(1:N) are extended to the higher energy shifts
  ! => omega(N+j) = omega_in - dble(j)*P%compton_d_omega*EV2AU in compton.f90
  ! => omega(i) = omega_in - dble(i-N)*P%compton_d_omega*EV2AU
  !
  do i = 2 * N, 1, -1
    omega(i) = omega_in - dble(i-N) * P%compton_d_omega * EV2AU
  end do
!  !
!  ! sort configurations: swap_index_int() is defined in read_param.f90
!  !
!  do i = 1, P%N_orb
!    do j = i+1, P%N_orb
!      if ( P%init_config(i,1) > P%init_config(j,1) ) then
!        call swap_index_int( P%init_config(i,:), P%init_config(j,:) )
!      else if ( P%init_config(i,1) == P%init_config(j,1) .and. &
!                P%init_config(i,2) >  P%init_config(j,2) ) then
!        call swap_index_int( P%init_config(i,:), P%init_config(j,:) )
!      end if
!    end do
!  end do
  !
  ! prepare configuration arrays to be used for parse_config_string()
  !
  allocate( config(P%N_orb,4), f_occ(P%N_orb,2) )
  !
  ! build the Legendre grid and WF_data
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  F = create_WF( P, G )
  !
  ! read foobar.pop: configuration string and its time-averaged population
  !
  open( P%IO_POP, file=P%file_pop, form="formatted", action="read" )
  loop_config: do
    read( P%IO_POP, "(a)", iostat=status ) line
    if ( line == "" .or. status < 0 ) exit              ! exit the loop
    if ( line(1:1) == "#" ) cycle                       ! skip for # (comment)
    read( line, * ) config_str, pop
    !
    ! put config_str into P%init_config
    ! parse_config_string() is no more compatible with the non-rel case!
    !
    config(:,3) = 0
    call parse_config_string( config_str, config, f_occ )
    do i = P%N_orb, 1, -1
      if ( config(i,3) /= 0 ) exit
    end do
    N_orb = i
    if ( sum( config(1:N_orb,3) ) == 0 ) cycle          ! skip if N_elec=0
    !
    ! check whether this configuration has been calculated or not
    ! how? check the directory whose name is config_str... and check the file
    !
    inquire( directory=trim(config_str), exist=yes_exist )
    if ( .not. yes_exist ) call system( "mkdir " // trim(config_str) )
    inquire( file=trim(config_str)//"/"//trim(file_ddsp_tot), exist=yes_exist )
    if ( .not. yes_exist ) then
      !
      ! create the file
      !
      open( IO_COMPTON, file=trim(config_str) // "/" // trim(file_ddsp_tot), &
            form="formatted", action="write" )
      close( IO_COMPTON )
      !
      ! move this directory
      !
      call chdir( trim(config_str) )    ! GNU extension, but it works for ifort
      !
      ! run SCF and compton scattering calculations
      !
      F%config(:,3) = 0
      F%N_orb = N_orb
      F%config(1:N_orb,1:3) = config(1:N_orb,1:3) ! ignore the relativistic case
      write(OUTPUT, "('# ',a)") repeat( "=", 77 )
      write(OUTPUT, "('# configuration: ',a/'#')") trim(config_str)
      call SCF_procedure( P, G, F )
      if ( P%yes_compton ) then
        if ( P%continuum_rmax > ZERO ) then
          call extend_WF_continuum( F, P%OUTPUT, P%continuum_rmax )
        else
          call extend_WF_continuum( F, P%OUTPUT )
        end if
        call solve_SE_all( F, G )
        call print_total_compton_cross_section( P, F )
      end if
      !
      ! return the original directory
      !
      call chdir( ".." )                ! GNU extension, but it works for ifort
      !
      ! check time
      !
      call check_time( int_time, timer )
      write(OUTPUT, "('#'/'# CPU_time checkpoint: ',a/'#')") &
        trim( job_time( timer - timer_old ) )
      timer_old = timer
    end if
    !
    ! read compton scattering data from foobar.ddsp_tot
    !
    write(OUTPUT,"('# reading a configuration: ',a)") trim(config_str)
    open( IO_COMPTON, file=trim(config_str) // "/" // trim(file_ddsp_tot), &
          form="formatted", action="read" )
    do i = N + 1, N*2
      read( IO_COMPTON, *, iostat=status ) omega_temp, DDSCS_value(:)
      if ( abs( omega_temp - omega(i) ) / omega(i) > 0.001_long ) then
        write(STDERR,*) "omega from calculation = ", omega(i)
        write(STDERR,*) "omega from file        = ", omega_temp
        stop
      end if
!      if ( status < 0 ) call error_message( "when reading a file" )
      DDSCS(i,:) = DDSCS(i,:) + pop * DDSCS_value(:)
    end do
  end do loop_config
  !
  ! read one blank line
  !
  read( P%IO_POP, "(a)", iostat=status ) line
  !
  ! read foobar.pop: free-electron populations
  !
  DDSCS_free(:,:) = ZERO
  loop_free_elec: do
    read( P%IO_POP, "(a)", iostat=status ) line
    if ( status < 0 ) exit                              ! exit the loop
    if ( line(1:1) == "#" .or. line == "" ) cycle       ! skip for # (comment)
    read( line, * ) E_eV, pop
    do j = 1, P%compton_N_theta
      DDSCS_free(:,j) = DDSCS_free(:,j) + pop * &
        compton_cs_free_elec( dble(E_eV)*EV2AU, omega_in, omega(:), theta(j) )
    end do
  end do loop_free_elec
  close( P%IO_POP )
  !
  ! print DDSCS (bound + free, bound, free)
  !
  write(OUTPUT, "('# ',a)") "both bound and free electrons"
  do i = 1, N * 2       ! bound + free
    where ( abs( DDSCS(i,:) ) < 1e-99_long ) DDSCS(i,:) = ZERO
    where ( abs( DDSCS_free(i,:) ) < 1e-99_long ) DDSCS_free(i,:) = ZERO
    write(OUTPUT,"(999es13.5)") omega(i), DDSCS(i,:) + DDSCS_free(i,:)
  end do
  write(OUTPUT, "(/)")  ! two blank lines
  write(OUTPUT, "('# ',a)") "bound electrons only"
  do i = 1, N * 2       ! bound only
    write(OUTPUT,"(999es13.5)") omega(i), DDSCS(i,:)
  end do
  write(OUTPUT, "(/)")  ! two blank lines
  write(OUTPUT, "('# ',a)") "ionized electrons only"
  do i = 1, N * 2       ! free only
    write(OUTPUT,"(999es13.5)") omega(i), DDSCS_free(i,:)
  end do
  !
  ! print out a foot note (current time / CPU time / elapse time...)
  !
  call footnote( P%OUTPUT )
contains

function compton_cs_free_elec( E, omega_in, omega, theta ) result (cs)
  real(long), intent(in) :: E, omega_in, theta
  real(long), intent(in) :: omega(:)
  real(long) :: cs(size(omega))
  real(long) :: shift0, lowerbound, upperbound
  real(long) :: Q, Q_old, D
  real(long) :: D_lower, D_upper, Q_lower, Q_upper
  real(long), parameter :: alpha = ONE / SPEED_C_AU
  integer :: i, k
  Q = TWO * alpha * omega_in * sin(HALF*theta)
  shift0 = omega_in - Q**2 / TWO
  lowerbound = shift0 - Q * sqrt( TWO*E )
  upperbound = shift0 + Q * sqrt( TWO*E )
  !shift0 = omega_in - alpha**2 * omega_in**2 * ( ONE - cos(theta) )
  !lowerbound = shift0 - TWO * alpha * omega_in * sqrt( TWO*E ) * sin(HALF*theta)
  !upperbound = shift0 + TWO * alpha * omega_in * sqrt( TWO*E ) * sin(HALF*theta)
  !
  ! solve the full expression in an iterative way
  !
!  D = ZERO
!  Q_old = ZERO
!  do k = 1, 50          ! determine D_lower
!    Q = alpha * sqrt( TWO * omega_in * ( omega_in + D ) * ( ONE-cos(theta) ) + &
!                      D**2 )
!    D = - Q**2 / TWO - Q * sqrt( TWO * E )
!    if ( abs( Q - Q_old ) < 1e-5_long ) exit
!    Q_old = Q
!  end do
!  D_lower = D
!  Q_lower = Q
!  print *, "lower k = ", k, abs( Q - Q_old )
!
!  D = ZERO
!  Q_old = ZERO
!  do k = 1, 50          ! determine D_upper
!    Q = alpha * sqrt( TWO * omega_in * ( omega_in + D ) * ( ONE-cos(theta) ) + &
!                      D**2 )
!    D = - Q**2 / TWO + Q * sqrt( TWO * E )
!    if ( abs( Q - Q_old ) < 1e-5_long ) exit
!    Q_old = Q
!  end do
!  D_upper = D
!  Q_upper = Q
!  print *, "upper k = ", k, abs( Q - Q_old )
!  !print *, shift0, lowerbound, upperbound
!  !print *, omega_in - Q_lower**2 / TWO, omega_in + D_lower
!  !print *, omega_in - Q_upper**2 / TWO, omega_in + D_upper
!  write(OUTPUT, "('C0 = ',f8.1,2x,'D0 = ',f8.1,5x,2f8.1)") &
!    ( omega_in - shift0 ) * AU2EV, ( upperbound - lowerbound )/TWO * AU2EV, &
!    lowerbound * AU2EV, upperbound * AU2EV
!  write(OUTPUT, "('Cl = ',f8.1,2x,'Dl = ',f8.1,5x,f8.1,8x,5x,f5.1)") &
!    Q_lower**2 / TWO * AU2EV, -( D_lower + Q_lower**2 / TWO ) * AU2EV, &
!    ( omega_in + D_lower ) * AU2EV, &
!    ( omega_in + D_lower - lowerbound ) / lowerbound * 100_long
!  write(OUTPUT, "('Cu = ',f8.1,2x,'Du = ',f8.1,5x,8x,f8.1,5x,f5.1)") &
!    Q_upper**2 / TWO * AU2EV, ( D_upper + Q_upper**2 / TWO ) * AU2EV, &
!    ( omega_in + D_upper ) * AU2EV, &
!    ( omega_in + D_upper - upperbound ) / upperbound * 100_long
!  !print *, "Q0      = ", sqrt( TWO * ( omega_in - shift0 ) ), &
!  !       "  D0      = ", omega_in - lowerbound, omega_in - upperbound
!  !print *, "Q_lower = ", Q_lower, "  D_lower = ", D_lower
!  !print *, "Q_upper = ", Q_upper, "  D_upper = ", D_upper
!  print *
!  Q = ( Q_lower + Q_upper ) / TWO
!  lowerbound = omega_in + D_lower
!  upperbound = omega_in + D_upper
  !
  ! calculate cross section from free electrons
  !
  do i = 1, size(omega)
    if ( omega(i) >= lowerbound .and. omega(i) <= upperbound ) then
!      cs(i) = Thomson_CS( theta ) * omega(i) / omega_in / &
!              ( FOUR * alpha * omega_in * sin(HALF*theta) * sqrt( TWO*E ) )
      cs(i) = Thomson_CS( theta ) * omega(i) / omega_in / &
              ( TWO * Q * sqrt( TWO*E ) )
    else
      cs(i) = ZERO
    end if
  end do
end function compton_cs_free_elec

end program xatom_imaging
