!----------------------------------------------------------------------------
! xatom_mc.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2019 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2019 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Monte-Carlo simulation for x-ray damaged atoms
!   written by Son, Sang-Kil in Dec. 2010 at DESY
!
! fort.200- : used for detailed pathways output when "-pathway" is on
! fort.300- : used for spectra output when "-spectra" is on
!----------------------------------------------------------------------------
program xatom_mc
  use constants
  use stdio
  use elec_config
  use read_param
  use gps
  use interpolate
  use print_property
  use git
  use random_num
  implicit none
  type(Param) :: P
  type(Grid) :: x               ! angular grids, x := cos(theta)
  !type(VSL_STREAM_STATE) :: stream
  real(long) :: rand, prob, diff
  integer :: ic                 ! current index during MC
  integer :: ic_old, ic_new
  integer, allocatable :: C_count(:)
  real(long), allocatable :: C(:)       ! probabilities of all configurations
  real(long), allocatable :: pop(:)     ! probabilities of all species
  real(long), allocatable :: pop_old(:)
  real(long), allocatable :: C_avg(:), pop_avg(:)   ! time-averaged quan.
  integer, allocatable :: charge(:)     ! charges of all configurations
  !
  ! Process data contains cross sections and rates of all processes.
  ! this is optimized for MC calculations, i.e.,
  ! easy to point out processes for one initial configuration.
  ! - A(i,j) indicates a process from i-th config to j-th config.
  ! - this A matrix is stored as a modified sparse matrix as follow,
  !   * Note: it's different from the standard sparse matrix!
  !
  !       | 1.0 -1.0      |
  !   A = |      2.0      |      NNZ = 5
  !       |               |
  !       |      3.0 -2.0 |
  !
  !   ind(:,1) = (/  1    3    0    4  /)       # indexing in A(:) for i
  !                                             # 0 indicates no elements for i
  !   ind(:,2) = (/  2    1    0    2  /)       # how many numbers of i-th conf.
  !   col(:)   = (/  1    2    2    2    3    /)
  !   A(:)     = (/  1.0 -1.0  2.0  3.0 -2.0  /)
  !
  !   N: the size of the matrix, N x N
  !   NNZ: the # of non-zero elements
  !   ind(:,1): the index in A(:) and col(:) of the 1st element of the i-th row
  !   col: the column index of an element
  !   # of nonzeros in the i-th row is stored in ind(:,2)
  !   * Note: In the original scheme, # of nonzeros in i is ind(i+1) - ind(i)
  !
  ! Here, value(:), cs_order(:), p_type(:), energy_eV(:) are sparse matrices.
  !
  type :: Process_data
    integer :: N                        ! # of dimension = # of config indexes
    integer :: N_process                ! # of nonzero elements = # of processes
    integer, allocatable :: ind(:,:)    ! for sparse matrices
    integer, allocatable :: col(:)      ! for sparse matrices
    real(long), allocatable :: value(:) ! cross section or rate in a.u.
    integer, allocatable :: cs_order(:) ! 0:rate, 1:one-photon cs, 2:two-ph. cs.
                                        ! hey, it's related to #photons!!!
    !
    ! spectra of electron and photon
    !   p_type(:) : type of the process
    !   energy_eV(:) : energy of elec. or photon of the process
    !                  values are rounded to nearst integer
    !   spectra_max: max. energy for spectra
    !   spectra(k,j) : counts for the k eV bin, here resolution becomes eV.
    !                  k = 0...max_spectra
    !                  j = 1 : photoelec, 2 : Auger, 3 : fluorescence
    !
    real(long), allocatable :: spectra(:,:)
    integer, allocatable :: p_type(:), energy_eV(:) ! hey, why p_type & cs_order!!!
    integer :: max_spectra
  end type
  type(Process_data) :: S
  integer :: i, j, k, l, N, k_E
  integer :: N_event, N_visit, N_config
  integer, allocatable :: config(:,:)    ! electronic configuration
  integer :: N_ph               ! N_ph: # of photons involved in the processes
  integer, allocatable :: N_ph_count(:)
  real(long) :: sum_spectra(3)  ! j = 1: photoelec, 2: Auger, 3: fluorescence
  real(long), allocatable :: N_ph_avg(:) ! avg. of N_ph over trajectories per Q
  real(long) :: t, dt, p_env, charge_avg, rate, rate_sum
  real(long) :: f2_avg, f0_avg, fp_avg, fpp_avg, ff_avg
  real(long), parameter :: AU2FS = AU2SEC / 1e-15_long
  integer :: OUTPUT
  !
  ! for charge-resolved spectra
  !
  integer, allocatable :: count_spectra_traj(:,:), count_spectra_charge(:,:,:)
  character(len=512) :: filename
  !
  ! for plotting of population dynamics
  !
  integer :: N_time
  real(long), allocatable :: time_grid(:)       ! temporal grids
  real(long), allocatable :: pop_t(:)
  integer, allocatable :: pop_t_count(:,:)
  !
  ! for scattering process
  !
  real(long), allocatable :: scd(:,:)   ! scattering data for all indexes
  real(long), allocatable :: scd_f(:,:) ! f0(Q=0), f', and f'' for all indexes
  real(long), allocatable :: cs(:)      ! scattering cross section at time t
  real(long), allocatable :: cs0(:)     ! scattering cross section of undamaged
  real(long), allocatable :: n_dist(:)  ! ang. dist. of # of scattered photons
  real(long), allocatable :: n0_dist(:) ! ang. dist. of # from unmadaged
  real(long) :: n_tot, n0_tot           ! total # of scattered photons
  real(long) :: R_factor                ! R factor
  real(long) :: I_tot, I0_tot           ! for Neutze(2000)'s definition
  real(long) :: theta
  real(long) :: alpha = ONE / SPEED_C_AU  ! fine-structure constant, alpha = 1/c
  !
  ! start your program with time checking!
  !
  call init_time()
  call initialize_random()
  !call initialize_random( stream )
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "MC" )
  call print_parameters( P, "Monte-Carlo simulation, " // GITVERSION )
  OUTPUT = P%OUTPUT
  !
  ! memory allocation
  ! N: # of all possible configurations, i.e. # of equations
  ! originally, N is given by occupation numbers := PI_i [ occ_n(i) + 1 ]
  ! now, N is given by the ground-state of the neutral atom of Z
  !
!  N = product( P%init_config(:,3) + 1 )
  N = 1; j = 0
  do i = 1, P%N_orb - 1         ! open-shell must locate at the end
    l = P%init_config(i,2)
    j = j + ( 4 * l + 2 )       ! max_occ = 4l+2; added to j
    N = N * ( 4 * l + 2 + 1 )   ! possible config: 0...max_occ ==> x(max_occ+1)
  end do
  N = N * ( ( P%Z - j ) + 1 )
  S%N = N
  allocate( C_count(N) )        ! counting of all configurations
  allocate( C(N) )              ! probabilities of all configurations
  allocate( charge(N) )         ! charges of all configurations
  allocate( pop(0:P%Z) )        ! prob. of all species; C, C+, C++, and so on
  allocate( pop_old(0:P%Z) )
  allocate( C_avg(N), pop_avg(0:P%Z) )      ! time-averaged quantities
  allocate( N_ph_count(1:P%Z) ) ! sum of # abs. photons for each charge
  allocate( N_ph_avg(1:P%Z) )   ! avg. # abs. photons for each charge
  C_avg(:) = ZERO
  C_count(:) = 0
  pop(0:) = ZERO
  N_ph_count(:) = 0
  N_ph_avg(:) = ZERO
  !
  ! define temporal grids
  !
  N_time = 100          ! let's put a fixed number for the time being
  allocate( time_grid(0:N_time) )
  allocate( pop_t(0:P%Z) )
  allocate( pop_t_count(0:N_time,0:P%Z) )
  pop_t(0:) = ZERO
  pop_t_count(0:,0:) = 0
  dt = ( P%t_end_au - P%t_start_au ) / dble(N_time)     ! during one pulse
  do i = 0, N_time
    time_grid(i) = P%t_start_au + dble(i) * dt
  end do
  !
  ! set up N_config from N_frozen
  ! N_config: # of configurations excluding electrons in frozen shell
  !
  N_config = N
  j = 0
  do i = 1, P%N_orb
    if ( j >= P%N_frozen ) exit
    j = j + P%init_config(i,3)          ! 3 is the occupation number
    N_config = N_config / ( P%init_config(i,3) + 1 )
  end do
  !
  ! determine charge(:) for each configuration
  !
  allocate( config(P%N_orb,3) )
  config(:,:) = P%init_config(:,:)      ! copy config from Param P
  do j = 1, N
    call index_to_config( j, config )
    charge(j) = P%Z - sum( config(:,3) )        ! Z - ( sum up all occ# )
  end do
  !
  ! read rate data from the input file
  ! and prepare the matrix containing all rates
  !
  write(OUTPUT, "('# Filenames:')")
  call load_rates( P, S )
  !
  ! prepare spectra data
  !
  if ( P%yes_spectra ) then
    allocate( S%spectra(0:S%max_spectra,3) )
    S%spectra(:,:) = ZERO
    allocate( count_spectra_traj(0:S%max_spectra,3) )
    allocate( count_spectra_charge(0:S%max_spectra,3,0:P%Z) )
    count_spectra_traj(0:,:) = 0
    count_spectra_charge(0:,:,0:) = 0
  end if
  !
  ! prepare the elastic scattering calculations
  ! non-resonant: requires only the SCD file
  ! resonant: requires both the SCD and MAD files
  !
  if ( P%yes_scattering ) then
    !
    ! prepare angular grids for integrating cross sections
    !   extended-LGL: x%r(0...N+1) because it includes the end points.
    ! CAUTION: it should be the same as the definition of x in scattering.f90
    !
    theta = TWO * asin( PI / P%resolution / alpha / P%PE_start )
    x = generate_Grid( "extended-LGL", P%N_scd, cos(theta), ONE )
    !
    ! memory allocation
    !
    allocate( scd(N,0:P%N_scd+1) )
    allocate( scd_f(N,0:2) )      ! f0(Q=0), f', f''
    allocate( cs(0:P%N_scd+1), cs0(0:P%N_scd+1) )
    allocate( n_dist(0:P%N_scd+1), n0_dist(0:P%N_scd+1) )
    n_dist(:) = ZERO
    !
    ! read SCD (and MAD) files
    !
    call load_scd( P, S, x, scd, scd_f )
    !
    ! set up cross section of the neutral atom as undamaged one
    !
    cs0(:) = scd(N,:)
  end if
  !
  ! set up the initial configuration
  !
  C(:) = ZERO
  if ( P%i0 < 0 .or. P%i0 > N ) P%i0 = N  ! initial = neutral if not specified
  C(P%i0) = ONE
  !
  ! consider synchrotron simulation if ground-state config. of neutral atom
  !
  if ( P%yes_synchrotron .and. P%i0 == N ) then
    rate_sum = ZERO
    do k = S%ind(P%i0,1), S%ind(P%i0,2)
      if ( S%cs_order(k) /= 1 ) cycle ! look for P only
      prob = S%value(k)               ! treated as a shake-off branching ratio
      rate_sum = rate_sum + prob      ! it's not a rate_sum, but just a sum
    end do
    if ( rate_sum /= ZERO ) then
      do k = S%ind(P%i0,1), S%ind(P%i0,2)
        if ( S%cs_order(k) /= 1 ) cycle
        S%value(k) = S%value(k) / rate_sum      ! make it as a probability
      end do
    else
      stop "No pcs data for the given configuration"
    end if
  end if
  !
  ! prepare some additional files
  !
  if ( P%yes_pathway ) then
    open( 200, file=trim(P%file_out) // ".path",   action="write" )
    open( 201, file=trim(P%file_out) // ".path.P", action="write" )
    open( 202, file=trim(P%file_out) // ".path.A", action="write" )
    open( 203, file=trim(P%file_out) // ".path.F", action="write" )
  end if
  if ( P%yes_spectra ) then
    open( 300, file=trim(P%file_out) // ".transition", action="write" )
    if ( P%yes_coincidence ) then
      do i = 0, P%Z
        write( filename, "(a,'.spectra.',i2.2)" ) trim(P%file_out), i
        open( 301+i, file=trim(filename), action="write" )
      end do
    end if
  end if
  !
  ! print out the heading
  !
  call print_divider( OUTPUT, "Monte Carlo calculation" )
  write(OUTPUT, "('#',a9,2a10)", advance="no") "event", "diff", "1-P_tot"
  call print_all_ions( OUTPUT, P%Z )
  write(OUTPUT, "()")
  !============================================================================
  ! Monte-Carlo method
  !============================================================================
  event: do j = 1, P%MC_max_events
    ic = int( P%i0, kind(ic) )  ! start from the initial config at the beginning
    dt = P%dt_au        ! it's supposed to be changed during time_evolution.
    t = P%t_start_au    ! time_evolution from t_start_au until no further decay
    N_ph = 0            ! # of photons involved in this trajectory
    if ( P%yes_pathway ) write(200,"(es13.5,i4)") t * AU2FS, charge(ic) 
    if ( P%yes_spectra ) count_spectra_traj(0:,:) = 0
    !
    ! consider synchrotron simulation
    !
    if ( P%yes_synchrotron ) then
      rand = random()
      process_P: do k = S%ind(ic,1), S%ind(ic,2)
        if ( S%cs_order(k) /= 1 ) cycle ! look for P only
        prob = S%value(k)               ! treated as a shake-off branching ratio
        if ( rand < prob ) then         ! hurrah! we want to go this pathway
          ic = S%col(k)                 ! set up a new configuration
          exit process_P
        end if
        rand = rand - prob
      end do process_P
    end if
    !
    ! time evolution
    !
    time_evolution: do
      !
      ! compute time-averaged quantity
      !
      if ( t <= P%t_end_au ) then
        p_env = pulse_envelope( P, t + dt * HALF )
        C_avg(ic) = C_avg(ic) + p_env * dt
        where( time_grid(:) >= t .and. time_grid(:) < t + dt ) &
          pop_t_count(:,charge(ic)) = pop_t_count(:,charge(ic)) + 1
      end if
      !
      ! time propagation: t is in a.u. and dt is varied during time propagation
      !
      t = t + dt
      !
      ! get a random number
      !
      !rand = random( stream )
      rand = random()
      rate_sum = ZERO
      !
      ! examine each process and determine "go" or "no-go"
      !
      loop_process: do k = S%ind(ic,1), S%ind(ic,2)
        rate = compute_rate( P, S%value(k), S%cs_order(k), t )
        prob = rate * dt
        ! ---------------------------------------------------------------------
        ! hurrah! let's move! we want to go this pathway
        ! ---------------------------------------------------------------------
        if ( rand < prob ) then
          if ( S%cs_order(k) > 0 ) N_ph = N_ph + S%cs_order(k)
          do i = k, S%ind(ic,2)         ! calculate remaining rates for rate_sum
            rate_sum = rate_sum + &
              compute_rate( P, S%value(i), S%cs_order(i), t )
          end do
          ic_old = ic
          ic_new = S%col(k)
          ic = ic_new                   ! set up a new configuration
          !
          ! print out electron/photon spectra
          !   fort.300         : all transition channels
          !   fort.301 ~ 300+q+1 : spectra for the final charge state +q,
          !                        which are print out at the end.
          !
          if ( P%yes_spectra ) then
            k_E = S%energy_eV(k)
            S%spectra( k_E, S%p_type(k) ) = S%spectra( k_E, S%p_type(k) ) + ONE
            count_spectra_traj(k_E,S%p_type(k)) = &
              count_spectra_traj(k_E,S%p_type(k)) + 1
            call index_to_config( ic_old, config )
            write(300,"(i2,i6,i2,' # ',a)",advance="no") charge(ic_old), k_E, &
              S%p_type(k), trim( config_to_string( config ) )
            call index_to_config( ic_new, config )
            write(300,"(' - ',a)") trim( config_to_string( config ) )
          end if
          !
          ! print out detailed pathway
          !   fort.200: all
          !   fort.201: photoionization
          !   fort.202: Auger decay
          !   fort.203: fluorescence
          !
          if ( P%yes_pathway ) then
            write(200,"(es13.5,i4,/)") t * AU2FS, charge(ic_old) 
            call index_to_config( ic_new, config )
            write(200,"(es13.5,i4,2x,'# ',a)") &
              t * AU2FS, charge(ic_new), trim( config_to_string( config ) )
            if ( S%p_type(k) /= 3 ) then
              write(200+S%p_type(k),"(es13.5,2i4,/,es13.5,2i4,/)") &
                t * AU2FS - ONE, charge(ic_old), charge(ic_new), &
                t * AU2FS + ONE, charge(ic_old), charge(ic_new)
            else
              write(200+S%p_type(k),"(es13.5,2i4)") t * AU2FS, charge(ic_old)
            end if
          end if
          !
          ! we don't need to be in this loop any more; let's escape
          !
          exit loop_process
        end if
        rand = rand - prob
        rate_sum = rate_sum + rate
      end do loop_process
      !
      ! stop this time evolution if there is no further decay available
      !
      !if ( rate_sum == ZERO ) then
      if ( abs(rate_sum) <= 1e-99_long ) then
        where( time_grid(:) >= t + dt ) &
          pop_t_count(:,charge(ic)) = pop_t_count(:,charge(ic)) + 1
        time_evol_end: do       ! time propagation until the end
          p_env = pulse_envelope( P, t + dt * HALF )
          C_avg(ic) = C_avg(ic) + p_env * dt
          t = t + dt
          if ( t > P%t_end_au ) exit time_evol_end
        end do time_evol_end
        if ( P%yes_spectra ) then  ! put spectra info with the final charge
          count_spectra_charge(0:,:,charge(ic)) = &
            count_spectra_charge(0:,:,charge(ic)) + count_spectra_traj(0:,:)
        end if
        exit time_evolution
      end if
      !
      ! adjust dt according to the sum of rates
      !
      if ( t <= P%t_end_au ) then
        dt = min( P%MC_dt_ratio / rate_sum, P%dt_au )
      else
        dt = P%MC_dt_ratio / rate_sum                   ! no need to have dt_max
      end if
    end do time_evolution
    !
    ! print out pathway information if it's turned on
    !
    if ( P%yes_pathway ) then
      write(200,"('# j = ',i8,5x,'N_ph = ',i5,5x,'final charge = ',i5,//)") &
        j, N_ph, charge(ic)
      write(201,"(/)")
      write(202,"(/)")
      write(203,"('#',/)")
    end if
    !
    ! compute the number of scattered photons (angular distribution)
    !     d N                                            d sigma
    !   ------- = INT dt [ flux(t) * SUM P_{conf}(t) * [ ------- ]_{conf} ]
    !   d Omega                     {conf}               d Omega
    !
!    if ( P%yes_scattering ) then
!      do k = 0, P%N_scd + 1
!        cs(k) = sum( C(:) * scd(:,k) )
!        n_dist(k) = n_dist(k) + P%fluence * p_env * cs(k) * dt
!      end do
      !
      ! the above implementation is almost 2 times faster than the below
      !
      !cs(:) = ZERO
      !do j = 1, N
      !  if ( C(j) > 1.0e-8_long ) cs(:) = cs(:) + C(j) * scd(j,:)
      !end do
      !n_dist(:) = n_dist(:) + P%fluence * p_env * cs(:) * dt
!    end if
    !
    ! counter increased
    !
    C_count(ic) = C_count(ic) + 1
    if ( N_ph > 0 .and. charge(ic) > 0 ) &
      N_ph_count(charge(ic)) = N_ph_count(charge(ic)) + N_ph
    !
    ! print out and check convergence
    !
    if ( mod(j,P%MC_step) == 0 ) then
      pop_old(0:) = pop(0:)
      pop(0:) = ZERO
      C(:) = dble(C_count(:)) / dble(j)
      do l = 1, N
        pop( charge(l) ) = pop( charge(l) ) + C(l)
      end do
      where ( pop(0:) < 1.0e-99_long ) pop(0:) = ZERO
      diff = maxval( abs( pop_old(0:) - pop(0:) ) )
      write(OUTPUT, "(i10,es10.3,sp,es10.2,s,100es10.3)") &
        j, diff, ONE - sum( pop(0:) ), pop(0:)
      if ( diff <= P%MC_conv ) exit event
    end if
  end do event
  !============================================================================
  !
  ! average all quantities for many events
  !
  N_event = j
  C(:) = dble(C_count(:)) / dble(N_event)
  C_avg(:) = C_avg(:) / dble(N_event)
  !
  ! compute probabilities of species and averaged values
  !
  where ( C(:)     < 1.0e-99_long ) C(:) = ZERO
  where ( C_avg(:) < 1.0e-99_long ) C_avg(:) = ZERO
  pop(0:) = ZERO
  pop_avg(0:) = ZERO
  do l = 1, N
    pop( charge(l) ) = pop( charge(l) ) + C(l)
    pop_avg( charge(l) ) = pop_avg( charge(l) ) + C_avg(l)
  end do
  charge_avg  = sum( (/ ( dble(i) * pop_avg(i), i = 0, P%Z ) /) )
  where ( pop(1:) /= ZERO ) &
    N_ph_avg(1:) = dble( N_ph_count(1:) ) / ( pop(1:) * dble(N_event) )
  !
  ! print out summary of populations (or probabilities)
  ! - |POP|     ones after the pulse and a long long time
  ! - |POP_AVG| ones averaged with weights by the normalized pulse
  ! - |POP_PULSE_END| ones at the end of the pulse; for back-compatibility only
  !
  write(OUTPUT, "('#',a9,20x,100es10.3)") "|POP|    ", pop(0:)
  write(OUTPUT, "('#',a9,20x,100es10.3)") "|POP_AVG|", pop_avg(0:)
  write(OUTPUT, "(/)")
  !
  ! print out charge-state distribution
  !
  call print_divider( OUTPUT, "Charge-state distribution" )
  call print_charge_states( OUTPUT, pop, pop_avg, N_ph_avg )
  !
  ! print out some important configurations (all if verbose mode)
  !
  write(OUTPUT, "('#')")
  call print_divider( OUTPUT, &
    "Population analysis of individual configurations" )
  write(OUTPUT, "('#   ',21x,2a12,4x,a)") "Pop", "Pop_avg", "description"
  do j = 1, N
    if ( .not. P%yes_verbose .and. &
         C(j) < 1.0e-5_long .and. C_avg(j) < 1.0e-5_long ) cycle
    if ( P%yes_verbose .and. C_avg(j) == ZERO ) cycle
    call index_to_config( j, config )
    write(OUTPUT, "('#   config = ',i10,2x,2es12.5,' ; ',sp,i3,': ',a)") &
      j, C(j), C_avg(j), charge(j), trim( config_to_string( config ) )
  end do
  write(OUTPUT, "(/)")
  !
  ! print out population dynamics
  !
  call print_divider( OUTPUT, "Population dynamics during the pulse" )
  write(OUTPUT, "('#',a9,2a10)", advance="no") "t(fs)", "pulse_env", "1-P_tot"
  call print_all_ions( OUTPUT, P%Z )
  write(OUTPUT, "()")
  do i = 0, N_time
    t = time_grid(i)
    p_env = pulse_envelope( P, t )
    pop_t(0:) = dble( pop_t_count(i,0:) ) / dble(N_event)
    write(OUTPUT, "(f10.3,es10.3,sp,es10.2,s,100es10.3)") &
      t * AU2FS, p_env, ONE - sum( pop_t(0:) ), pop_t(0:)
  end do
  write(OUTPUT, "('#',a15,14x,100es10.3)") "|POP_PULSE_END|", pop_t(0:)
  write(OUTPUT, "(/)")
  !
  ! print out electron and photon spectra
  !
  if ( P%yes_spectra ) then
    call print_divider( OUTPUT, "Electron/photon spectra" )
    write(OUTPUT, "('#   ',a6,3a13)") &
      "E(eV)", "Photoelec.", "Auger elec.", "fluorescence"
    !
    ! 1 eV bin
    !
    S%spectra(:,1:3) = S%spectra(:,1:3) / dble(N_event)
    sum_spectra(1:3) = ZERO
    do j = 0, S%max_spectra
      where ( S%spectra(j,1:3) < 1.0e-99_long ) S%spectra(j,1:3) = ZERO
      write(OUTPUT, "(4x,i6,3es13.5)") j, S%spectra(j,1:3)
      sum_spectra(1:3) = sum_spectra(1:3) + S%spectra(j,1:3) 
    end do
    write(OUTPUT, "('#|SPECTRA|',3es13.5)") sum_spectra(1:3)
    write(OUTPUT, "(/)")
    !
    ! 10 eV bin, optionally
    ! OLD code)
    !    do j = 10, S%max_spectra, 10
    !      write(OUTPUT, "(4x,i6,3es13.5)") j, &
    !        (/ ( sum( S%spectra(j-9:j,k) ), k=1,3 ) /)
    !    end do
    !
    !sum_spectra(1:3) = ZERO
    !do j = 0, S%max_spectra
    !  if ( mod( j, 10 ) == 0 ) then
    !    write(OUTPUT, "(4x,i6,3es13.5)") j, sum_spectra(1:3)
    !    sum_spectra(1:3) = ZERO
    !  end if
    !  sum_spectra(1:3) = sum_spectra(1:3) + S%spectra(j,1:3)
    !end do
    !write(OUTPUT, "(/)")
    !
    ! print out spectra for given charge states
    !
    if ( P%yes_coincidence ) then       ! charge-resolved spectra
      do i = 0, P%Z
        do j = 0, S%max_spectra
          write(301+i, "(4x,i6,3es13.5)") j, &
            count_spectra_charge(j,:,i) / dble(N_event)
        end do
      end do
    end if
  end if
  !
  ! print out # of scattered photons and R factor
  !
  if ( P%yes_scattering ) then
    call print_divider( OUTPUT, "Scattering and/or MAD analysis"  )
    n0_dist(:) = P%fluence * cs0(:)     ! # from undamaged
    n_tot  = TWO*PI * sum( n_dist(0:)  * x%weight(0:) ) ! 2pi from INT d(phi)
    n0_tot = TWO*PI * sum( n0_dist(0:) * x%weight(0:) ) ! 2pi from INT d(phi)
    !
    ! R-factor defined in Neutze, et al., Nature 406, 752 (2000)
    ! R = SUM_Omega | sqrt(I(Omega)) / I_tot - sqrt(I0(Omega)) / I0_tot |
    !   where I_tot  = SUM_Omega sqrt( I(Omega)  )    I:  damaged
    !         I0_tot = SUM_Omega sqrt( I0(Omega) )    I0: undamaged
    !
    I_tot  = TWO*PI * sum( sqrt(n_dist(0:))  * x%weight(0:) )
    I0_tot = TWO*PI * sum( sqrt(n0_dist(0:)) * x%weight(0:) )
    R_factor = TWO*PI * sum( abs( sqrt(n_dist(0:))/I_tot - &
                                  sqrt(n0_dist(0:))/I0_tot ) * x%weight(0:) )
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# Scattering process with x-ray damage')")
    write(OUTPUT, "('#   Resolution d (angstrom) = ',f8.3)") &
      P%resolution * AU2ANG
    write(OUTPUT, "('#   Q_max = 2pi / d (a.u.)  = ',es14.7)") &
      TWO*PI / P%resolution
    write(OUTPUT, "('#   N_scattered [undamaged] = ',es14.7)") n0_tot
    write(OUTPUT, "('#   N_scattered [damaged]   = ',es14.7)") n_tot
    write(OUTPUT, "('#   R-factor                = ',es14.7)") R_factor
    !
    ! time-averaged form factors with the dispersion correction
    !   a(0,omega) = f2_avg
    !   b(0,omega) = ( f0_avg + fp_avg ) * 2
    !   c(omega) = fpp_avg * 2
    !   a~(0,omega) = ff_avg
    !
    ! f2_avg = sum_I1 [ int dt J(t) P_I1(t) ] | f_1,I1(0,omega) |^2
    ! ff_avg = int dt J(t) | sum_I1 P_I1(t) f_1,I1(0,omega) |^2
    !
    if ( P%yes_dispersion ) then
      f2_avg  = sum( ( ( scd_f(:,0) + scd_f(:,1) )**2 + scd_f(:,2)**2 ) * &
                     C_avg(:) )
      f0_avg  = sum( scd_f(:,0) * C_avg(:) )
      fp_avg  = sum( scd_f(:,1) * C_avg(:) )
      fpp_avg = sum( scd_f(:,2) * C_avg(:) )
      write(OUTPUT, "('#')")
      write(OUTPUT, "('# Time-averaged form factors (Q=0)')")
      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f0",    f0_avg
      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f'",    fp_avg
      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f''",   fpp_avg
      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "|f|^2", f2_avg
      write(OUTPUT, "('#')")
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a(0,omega)", f2_avg
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "b(0,omega)", &
        TWO * ( f0_avg + fp_avg )
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "c(omega)", TWO * fpp_avg
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a~(0,omega)", ff_avg
    end if
    !
    ! print out summary
    !
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# PUNCH ',6a12)") "fluence(au)", "FWHM(au)", &
      "N(undam.)", "N(damaged)", "R-factor", "avg. charge"
    write(OUTPUT, "('#|PUNCH|',6es12.5)") &
      P%fluence, P%FWHM, n0_tot, n_tot, R_factor, charge_avg
    if ( P%yes_dispersion ) then
      !write(OUTPUT, "('# PUNC2 ',6a12)") "PE(eV)", &
      !  "avg. f0", "avg. f'", "avg. f''", "avg. f_real", "avg. |f|^2"
      !write(OUTPUT, "('#|PUNC2|',f12.2,5es12.4)") P%PE_start * AU2EV, &
      !  f0_avg, fp_avg, fpp_avg, f0_avg + fp_avg, f2_avg
      write(OUTPUT, "('# PUNC2 ',6a12)") "PE(eV)", &
        "a(0,omega)", "b(0,omega)", "c(omega)", "a~(0,omega)"
      write(OUTPUT, "('#|PUNC2|',f12.2,4es12.4)") P%PE_start * AU2EV, &
        f2_avg, TWO * ( f0_avg + fp_avg ), TWO * fpp_avg, ff_avg
    end if
  end if
  !
  ! print out how many configurations are involved
  ! total # of configurations: we need to consider N_frozen, so it becomes N-m.
  !
  call print_divider( OUTPUT, "Monte Carlo statistics" )
  N_visit = count( C_avg(:) /= ZERO )
  write(OUTPUT, "('# - number of trajectories           = ',i15)") N_event
  write(OUTPUT, "('# - max. diff. of charge populations = ',es15.5)") diff
  write(OUTPUT, "('# - diff. checked out per ',i5,' trajectories')") P%MC_step
  write(OUTPUT, "('# - N of all possible q-hole config. = ',i15)") N_config
  write(OUTPUT, "('# - N of config. visited in this run = ',i15," // &
                "'  (',f6.2,'%)')") &
    N_visit, dble(N_visit) / dble(N_config) * 100_long
  !
  ! print out the foot note
  !
  call footnote( OUTPUT )
contains

subroutine load_rates( P, S )
  type(Param), intent(in) :: P
  type(Process_data), intent(inout) :: S
  integer :: i, j, l, i_old
  integer :: INFO
  real(long) :: value, energy
  character(len=1) :: tag
  !
  ! # of all possible configurations for a given atom
  !
  allocate( S%ind(S%N,2) )
  !
  ! find total # of processes in the data file
  ! CAUTION: the i index in the data file should be in the ascending order,
  !          otherwise it gives an error!
  ! CAUTION: elements of the i index should be grouped.
  !
  open( P%IO_DATA, file=P%file_data, action="read", form="formatted" )
  l = 0; i_old = 0
  S%ind(:,:) = 0
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i < 0 .or. j < 0 .or. value == ZERO ) cycle    ! skip unmeaningful
    if ( i > S%N .or. j > S%N ) then
      write(STDERR, *) "In the data file,"
      write(STDERR, *) tag, i, j, value
      write(STDERR, *) "Index should not be greater than ", S%N
      stop
    end if
    l = l + 1
    if ( l == 1 ) then
      S%ind(i,1) = l
      i_old = i
    else if ( i /= i_old ) then
      if ( S%ind(i,2) /= 0 ) then
        !
        ! if this is not zero, 
        ! then there is another i-th config element in different runs.
        !
        write(STDERR, *) "All elements of the i-th configurations " // &
                         "must be grouped."
        write(STDERR, *) "Please run the validate_dat tool."
        stop
      end if
      S%ind(i,1) = l            ! put the current index for i
      S%ind(i_old,2) = l - 1    ! put the ending index for i_old
      i_old = i
    end if
  end do
  S%ind(i_old,2) = l            ! put the ending index for i_old
  S%N_process = l                     ! # of non-zero elements
  allocate( S%col(S%N_process), S%value(S%N_process) )
  allocate( S%cs_order(S%N_process), S%p_type(S%N_process) )
  allocate( S%energy_eV(S%N_process) )
  rewind( P%IO_DATA )
  write(OUTPUT, "('#   considering how many processes: ',i10)") S%N_process
  !
  ! read rate data and save into S
  !   i: initial configuation index
  !   j: final configuration index
  !   value: rate or cross section; both should be in a.u.
  !
  l = 0
  S%max_spectra = 0
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value, energy
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i < 0 .or. j < 0 .or. value == ZERO ) cycle    ! skip unmeaningful
    l = l + 1
    S%col(l) = j
    S%value(l) = value
    S%energy_eV(l) = nint(energy)               ! energy is given in eV already
    !S%energy_eV(l) = nint(energy * AU2EV)       ! nint(): round to nearest int.
    if ( S%max_spectra < S%energy_eV(l) ) S%max_spectra = S%energy_eV(l)
    if ( tag == "P" .or. tag == "V" ) then      ! for photoionization
      S%p_type(l) = 1;  S%cs_order(l) = 1
    else if ( tag == "T" ) then                 ! for two-photon ionization
      S%p_type(l) = 1;  S%cs_order(l) = 2
    else if ( tag == "A" ) then                 ! for Auger decay
      S%p_type(l) = 2;  S%cs_order(l) = 0
    else                                        ! for fluorescence
      S%p_type(l) = 3;  S%cs_order(l) = 0
    end if
  end do
  close( P%IO_DATA )
  write(OUTPUT, "('#   all process data loaded from: ',a)") trim(P%file_data)
  if ( P%yes_spectra ) then
    write(OUTPUT, "('#   max. energy (eV) in spectra from data file: ',i8)") &
      S%max_spectra
  end if
  write(OUTPUT, "('#')")
end subroutine load_rates

subroutine load_scd( P, S, x, scd, scd_f )
  type(Param), intent(in) :: P
  type(Process_data), intent(inout) :: S
  type(Grid), intent(in) :: x
  real(long), intent(out) :: scd(:,0:)          ! WARNING: starting from 0!!!
  real(long), intent(out) :: scd_f(:,0:)        ! WARNING: starting from 0!!!
  real(long), allocatable :: Q(:), f(:)         ! dim. determined by SCD file
  real(long) :: fp(size(scd,1))                 ! fp(i) or fpp(i): i-th config.
  real(long) :: fpp(size(scd,1))
  real(long) :: i_final(P%N_orb), pcs(P%N_orb)
  real(long) :: Q_k, omega, alpha, f0, w, temp1, temp2
  type(Spline) :: SP
  complex(long) :: cff          ! coherent form factor: COMPLEX!!!
  integer :: i, j, k, l, m, N, N_orb, INFO, OUTPUT
  character(len=1) :: tag
  logical :: yes_exist
  alpha = ONE / SPEED_C_AU      ! fine-structure constant, alpha = 1/c
  N_orb = P%N_orb               ! this should be N_shell.  # of diff. shells
  OUTPUT = P%OUTPUT
  !
  ! read the dispersion correction from the MAD file
  ! ACHTUNG: even without the dispersion correction,
  !          sometimes it's convenient to read this MAD file 
  !          because it contains PCS data for different photon energies.
  !          so, read the MAD file no matter how yes_dispersion is on or not.
  !
  fp(:) = ZERO
  fpp(:) = ZERO
  scd_f(:,1) = ZERO     ! f'
  scd_f(:,2) = ZERO     ! f''
  inquire( file=P%file_mad, exist=yes_exist )
  if ( yes_exist ) then
    open( P%IO_MAD, file=P%file_mad, action="read", form="formatted" )
    omega = P%PE_start * AU2EV  ! be careful, MAD has P.E. in eV
    m = 0
    loop_MAD: do
      !
      ! i: configuration index, see config_to_index()
      ! N: # of data for given index
      ! tag: must be a character of '#'
      !      why do I use '#'?  well... you may use gnuplot with the SCD file .
      !
      read( P%IO_MAD, *, iostat=INFO ) tag, i, N
      if ( INFO < 0 ) exit loop_MAD             ! exit when it ends
      m = m + 1
      read( P%IO_MAD, * ) tag, i_final(:)       ! forget about (-) i_final's...
      do j = 1, N
        read( P%IO_MAD, * ) w, temp1, temp2, pcs(:)
        if ( abs( omega - w ) < 1.0e-8_long ) then  ! set up data for given P.E.
          !
          ! set up the dispersion correction for the i-th configuration
          !
          if ( P%yes_dispersion ) then
            scd_f(i,1) = temp1          ! f'
            scd_f(i,2) = temp2          ! f''
          end if
          !
          ! change the PCS data for the i-th configuration with given P.E.
          !
          do l = 1, N_orb
            do k = S%ind(i,1), S%ind(i,2)
              if ( S%col(k) == i_final(l) ) S%value(k) = pcs(l)
            end do
          end do
        end if
      end do
      read( P%IO_MAD, * )                       ! skip one blank line
    end do loop_MAD
    close( P%IO_MAD )
    if ( m /= 0 ) then
      write(OUTPUT, "('# ... all MAD data are loaded from ',a)") &
        trim(P%file_mad)
    else if ( .not. P%yes_dispersion ) then
      write(OUTPUT, "('# ... only PCS data are loaded from ',a)") &
        trim(P%file_mad)
    end if
  else
    write(OUTPUT, "('# ... no MAD file is available')")
    write(OUTPUT, "('# ... PCS data are fixed at a certain photon energy.')")
  end if
  !
  ! read the scattering data from the SCD file
  ! convert them from the coferent form factor into the cross section
  !
  !   d sigma
  !   ------- = alpha^4 |f0(Q)|^2 [ the polarization-dependent factor ]
  !   d Omega
  !             where [ ] = ( 1 + cos(theta)^2 ) / 2 for unpolarized x-rays
  !
  open( P%IO_SCD, file=P%file_scd, action="read", form="formatted" )
  m = 0
  scd(1,:) = ZERO     ! no electron in this configuration (bare atom)
  loop_SCD: do
    !
    ! i: configuration index, see config_to_index()
    ! N: # of data for given index
    ! tag: must be a character of '#'
    !
    read( P%IO_SCD, *, iostat=INFO ) tag, i, N
    if ( INFO < 0 ) exit loop_SCD               ! exit when it ends
    m = m + 1
    if ( m == 1 ) allocate( Q(N), f(N) )        ! for the first data set
    do k = 1, N
      read( P%IO_SCD, * ) Q(k), f(k)            ! read the i-th config. data
    end do
    read( P%IO_SCD, * )                         ! skip one blank line
    if ( m == 1 ) then
      SP = Spline_create( Q, f )                ! for the first data set
    else
      call Spline_reuse( SP, f )                ! otherwise, just reuse
    end if
    !
    ! set up the i-th scd data
    !
    do k = 0, P%N_scd + 1
      Q_k = TWO * alpha * P%PE_start * sin( acos( x%r(k) ) / TWO )
      f0 = Spline_interpolate( SP, Q_k )
      cff = f0                                  ! non-resonant form factor
      if ( Q_k == ZERO ) scd_f(i,0) = f0        ! f0(Q=0)
      if ( P%yes_dispersion ) then              ! dispersion correction
        cff = cff + scd_f(i,1) + IMAG_ONE * scd_f(i,2)
      end if
      scd(i,k) = alpha**4 * cdabs( cff )**2 * ( ONE + x%r(k)**2 ) / TWO
    end do
  end do loop_SCD
  close( P%IO_SCD )
  write(OUTPUT, "('# ... all SCD data are loaded from ',a)") trim(P%file_scd)
  write(OUTPUT, "('#')")
  deallocate( Q, f )
end subroutine load_scd


function compute_rate( P, value, order, t ) result (rate)
  real(long) :: rate
  type(Param), intent(in) :: P
  real(long), intent(in) :: value, t
  integer, intent(in) :: order
  real(long) :: flux
  rate = value
  if ( order > 0 ) then
    !
    ! compute flux
    !   fluence = # of photons / spatial_width^2
    !
    flux = P%fluence * pulse_envelope( P, t )
    rate = rate * flux**order
  end if
end function compute_rate

function pulse_envelope( P, t ) result (f)
  real(long) :: f
  type(Param), intent(in) :: P
  real(long), intent(in) :: t
  real(long) :: sigma
  !
  ! normalized Gaussian distribution
  !   f = e^( -t^2 / 2 sigma^2 ) / ( sigma sqrt(2pi) )
  ! where sigma: standard deviation
  !       FWHM (Full Width at Half Maximum), tau = 2 sigma sqrt( 2 ln 2 )
  !
  if ( t <= P%t_start_au .or. t >= P%t_end_au ) then
    f = ZERO
  else if ( P%pulse_type == 3 ) then    ! Gaussian
    sigma = P%FWHM / TWO / sqrt( TWO * log(TWO) )
    f = exp( -t**2 / TWO / sigma**2 ) / sqrt( TWO * PI ) / sigma
  else if ( P%pulse_type == 6 ) then    ! flat-top, FWHM here is pulse duration
    f = ONE / P%FWHM
  end if
end function pulse_envelope

end program xatom_mc
