!----------------------------------------------------------------------------
! xatom_mc_fly.f90
!----------------------------------------------------------------------------
! Copyright (C) 2013-2021 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2013-2021 Sang-Kil Son <sangkil.son@cfel.de>
!          2015-2018 Koudai Toyota
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Monte-Carlo On-the-fly simulation for x-ray damaged atoms
!   originally written (MC) by Son, Sang-Kil in Dec. 2010 at DESY
!   written (MC_fly) by Son, Sang-Kil in Mar. 2012 at DESY
!
! ISSUE TO BE RESOLVED SOON:
!   config(:,1:3) for nonrelativistic
!   config(:,1:4) for relativistic
!   F%config(:,1:3)     for nonrelativistic
!   F%config_rel(:,1:5) for relativistic
! For the relativistic case, 
!   config and F%config_rel have different formats from the non-rel case, 
!   which are very confusing, especially for config!!!
! 
! - F%config_rel(i,1)   config(i,1)     : quantum number n
! - F%config_rel(i,2)   config(i,2)     : quantum number l
! - F%config_rel(i,3)   config(i,3)     : quantum number j
! - F%config_rel(i,4)   config(i,4)     : occupation number
! - F%config_rel(i,5)                   : (nonrel) orbital index
!----------------------------------------------------------------------------
program xatom_mc_fly
  use constants
  use stdio
  use atom
  use elec_config
  use read_param
  use gps
  use wavefunction
  use interpolate
  use process
  use scf
  use decay
  use pcs
  use random_num
!  use shakeoff
  use track_elec_config
  use print_property
  use git
  implicit none
  type(Config_tree), target :: Q0       ! starting point of the tree
  type(Param) :: P
  type(Coordinate) :: G
  type(WF_data) :: F
  type(Process_data), pointer :: S
  type(Grid) :: r
  !type(Grid) :: x               ! angular grids, x := cos(theta)
  type(track_config) :: tec
  !type(VSL_STREAM_STATE) :: stream
  real(long) :: rand, prob, diff
  integer, allocatable :: config(:,:)
  real(long), allocatable :: pop(:)         ! population of all species
  integer, allocatable :: pop_count(:)
  real(long), allocatable :: pop_old(:)
  real(long), allocatable :: pop_avg(:)     ! time-averaged quan.
  real(long), allocatable :: init_pot(:)
  integer :: i, j, k, l
  integer :: ii
  integer :: N_event, N_ph, charge, charge_old
  integer, allocatable :: N_ph_count(:)
  integer, allocatable :: occ_count(:)      ! occupation number
  real(long), allocatable :: N_ph_avg(:)
  real(long) :: N_ph_abs                ! # of absorbed photons
  real(long) :: E_absorbed              ! absorbed energy via photons
  real(long) :: E_emitted               ! emitted energy via photons
  real(long) :: t, tp, dt, p_env, charge_avg, rate, rate_sum, E_old
  real(long) :: E_eV_before_relax, E_eV_after_relax
  !real(long) :: f2_avg, f0_avg, fp_avg, fpp_avg, ff_avg
  real(long), parameter :: AU2FS = AU2SEC / 1e-15_long
  integer :: OUTPUT, OUTPUT_MAIN
  !logical :: yes_exist
  integer :: p_type
  character(len=1) :: tag               ! indicates a process type
  !character(len=3) :: str_charge
  !character(len=7) :: str_name
  character(len=25) :: str_process
  character(len=512) :: file_scan       ! filename for scanning fluences
  character(len=512) :: filename
  character(len=10) :: format_string
  integer*8 :: int_time
  real :: timer, timer_old
  integer*8 :: N_config
  integer :: N_acc_traj         ! N of traj. accumulated for a range of fluence
  integer :: N_acc_visit, N_acc_calc, N_acc_store, N_acc_proc
  integer :: N_acc_res          ! accumulated counter for resonance processes
  integer :: N_acc_so           ! accumulated counter for shakeoff processes
  integer :: N_acc_time         ! CPU time accumulated for a range of fluence
  integer :: N_res_excitation   ! counter for resonance processes
  integer :: N_shakeoff         ! counter for shakeoff processes
  real(long) :: value
  real(long) :: scan_v0, scan_v ! for scanning fluence or pulse duration
  integer, allocatable :: N_proc(:,:)   ! counter for all processes
  integer :: N_PI(5)                    ! N of (multi)photon ionization up to 5
  integer :: i_core, quantum_n          ! temp. indexes used for N_proc
  !
  ! for plotting of population dynamics
  !
  integer :: N_time, Np_time, k_time            ! k_time: index for time_grid(:)
  real(long), allocatable :: time_grid(:)       ! temporal grids
  real(long), allocatable :: pop_t(:)
  integer, allocatable :: pop_t_count(:,:)      ! charge state population
  integer, allocatable :: occ_t_count(:,:)      ! occupation number
  !
  ! for scattering process
  !
  !real(long), allocatable :: scd(:,:)   ! scattering data for all indexes
  !real(long), allocatable :: scd_f(:,:) ! f0(Q=0), f', and f'' for all indexes
  !real(long), allocatable :: cs(:)      ! scattering cross section at time t
  !real(long), allocatable :: cs0(:)     ! scattering cross section of undamaged
  !real(long), allocatable :: n_dist(:)  ! ang. dist. of # of scattered photons
  !real(long), allocatable :: n0_dist(:) ! ang. dist. of # from unmadaged
  !real(long) :: n_tot, n0_tot           ! total # of scattered photons
  !real(long) :: R_factor                ! R factor
  !real(long) :: I_tot, I0_tot           ! for Neutze(2000)'s definition
  !real(long) :: Q, resolution, theta
  !real(long) :: alpha = ONE / SPEED_C_AU  ! fine-structure constant, alpha = 1/c
  !
  ! for relativistic effects:
  !   occupation number is stored in config(:,3) when yes_relativity is off
  !   occupation number is stored in config(:,4) when yes_relativity is on
  !   thus, occ_index will have 3 or 4 depending on yes_relativity.
  ! this convention has to be reconsidered...
  ! => April 10, 2019 Let's make this way
  !    size(config,dim=2) == 3 ==> non-relativistic
  !    size(config,dim=2) == 4 ==> relativistic
  !    from config(:,:), we can recognize whether it's relativistic or not
  !    then, occ_index = size(config(:,:),dim=2)
  ! => WARNING!!! -- April 16, 2019
  !    size(config,dim=2) = 3 or 4
  !    size(P%init_config,dim=2) = 3
  !    size(P%init_config_rel,dim=2) = 0 or 5
  !    size(F%config,dim=2) = 3
  !    size(F%config_rel,dim=2) = 0 or 5
  !    Don't use size(config,dim=2) == 4!!!
  !
  integer :: occ_index
  !
  ! for spectra
  !   count_spectra_traj(k,j) : counts for the k-th eV bin, per trajectory
  !     it is initialized at the beginning of each trajectory
  !     k = 0...max_E_spectra
  !     j = 1 : photoelec, 2 : Auger, 3 : fluorescence
  !   count_spectra_charge(k,j,q) : accumulated counts for each charge q
  !     it is used for coincident measurement along with charge state
  !     however, it is also used no matter whether P%yes_coincidence is on/off
  !     because total spectra is calculated by summation of it over all q's
  !     => spectra(k,j) = sum( count_spectra_charge(k,j,0:) ) / N_event
  !   pop_e_avg(k) : time-averaged pulse-weighted pop. of k-th-eV electrons
  !                  pop_e_avg = pop_e_avg / N_event at the end
  !
  integer, allocatable :: count_spectra_traj(:,:)
  integer, allocatable :: count_spectra_charge(:,:,:)   ! for coincidence
  real(long), allocatable :: pop_e_avg(:)
  real(long) :: sum_spectra(3)
  integer :: max_E_spectra
  !
  ! Spline interpolation for the temporal profile
  !
  type(Spline) :: SP    ! only meaningful if P%yes_load_pulse is on
  real(long), allocatable :: temporal_profile(:,:)

  ! a flag used when shake-off is calculaed
  !logical :: is_dynamical = .true.

!  ! for shakeoff
!  type(Shakeoff_data) :: SO

  ! for track_cs
  logical :: is_decay_happened

  !
  ! start your program with time checking!
  !
  call init_time()
  call initialize_random()
  !call initialize_random( stream )
  !
  ! set up initial values
  !
  P%N = 200
  P%L1 = 1.0_long
  P%Rmax = ZERO
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "MC_fly" )
  call print_parameters( P, "Monte-Carlo on-the-fly simulation, "// GITVERSION )
  OUTPUT = P%OUTPUT
  !
  ! print out some printing options
  !
  write(OUTPUT, "('# Printing options:')")
  write(OUTPUT, "('#   -yes_spectra:     ',l)") P%yes_spectra
  write(OUTPUT, "('#   -yes_coincidence: ',l)", advance="no") P%yes_coincidence
  if ( P%yes_coincidence ) then
    write(OUTPUT, "(2x,'=> output: ',a,'.spectra.Q')" ) trim(P%file_out)
  else
    write(OUTPUT, "()")
  end if
  write(OUTPUT, "('#   -yes_population:  ',l)", advance="no") P%yes_population
  if ( P%yes_population ) then
    write(OUTPUT, "(2x,'=> output: ',a)") trim(P%file_pop)
  else
    write(OUTPUT, "()")
  end if
  write(OUTPUT, "('#   -yes_occupation:  ',l)") P%yes_occupation
  write(OUTPUT, "('#')")
  !
  ! print out massage options --- manipulating cross sections and rates
  !
  if ( P%yes_massage ) then
    write(OUTPUT, "('# massage options: multiplication factor')")
    write(OUTPUT, "('# - autoionization rates                 x ',f5.2)") &
      P%massage_autoionization_rate
    write(OUTPUT, "('# - all auger rates (excluding autoion.) x ',f5.2)") &
      P%massage_auger_rate
    write(OUTPUT, "('# - resonant excitation cross section    x ',f5.2)") &
      P%massage_resonant_pcs
    write(OUTPUT, "('# - photoionization cross section        x ',f5.2)") &
      P%massage_pcs
    write(OUTPUT, "('# - Auger for s hole',21x,'x ',f5.2)") P%massage_decay_s
    write(OUTPUT, "('# - Auger for p hole',21x,'x ',f5.2)") P%massage_decay_p
    write(OUTPUT, "('# - Auger for d hole',21x,'x ',f5.2)") P%massage_decay_d
    write(OUTPUT, "('# - Auger for f hole',21x,'x ',f5.2)") P%massage_decay_f
    write(OUTPUT, "('# - Auger rates multiplied by ',f5.2,' if q <  ',i3)") &
      P%massage_auger_factor1, P%massage_auger_q1
    write(OUTPUT, "('# - Auger rates multiplied by ',f5.2,' if q >= ',i3)") &
      P%massage_auger_factor2, P%massage_auger_q2
    write(OUTPUT, "('#')")
  end if
  !
  ! set up a directory for atomic data
  ! in regular fortran, there is no way to check if a directory exists.
  ! here I just use an ifort feature, inquire( directory="" )
  !
!  inquire( directory=trim(P%data_dir), exist=yes_exist )
!  if ( .not. yes_exist ) call system( "mkdir " // trim(P%data_dir) )
!  do i = 0, P%Z
!    write(str_charge, "(i3)") i
!    write(str_name, "(a,'_',a,'+')") trim(ATOMIC_NAME(P%Z)), &
!                                     trim(adjustl(str_charge))
!    inquire( directory=trim(P%data_dir)//"/"//trim(str_name), exist=yes_exist )
!    if ( .not. yes_exist ) &
!      call system( "mkdir " // trim(P%data_dir) // "/" // trim(str_name) )
!  end do
!  write(OUTPUT, "('# Filenames:')")
!  write(OUTPUT, "('#   all atomic data are stored in ',a,'/.')") &
!    trim(P%data_dir)
  !
  ! .dat file: for back-compatibility to xatom_mc and xatom_req
  !
!  open( P%IO_DATA, file=P%file_data, access="append", form="formatted" )
!  write(OUTPUT, "('#   one data file for back-compatibility: ',a)") &
!    trim(P%file_data)
!  write(OUTPUT, "('#   ',a/'#   ',a)") &
!    "WARNING: be careful to use xatom_mc with this data file,", &
!    "         especially for heavy atom and high fluence cases"
!  write(OUTPUT, "('#')")
  !
  ! build the Legendre grid
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  !
  ! build WF_data
  !
  F = create_WF( P, G )
  !
  ! memory allocation
  !
  if ( P%yes_relativity ) then
    occ_index = 4 ! 1st arg:n, 2nd arg: l, 3rd arg:j, 4th arg: occ 
  else
    occ_index = 3 ! 1st arg:n, 2nd arg: l, 3rd arg: occ 
  endif
  allocate( config(F%N_subshell,occ_index) ) ! max. # diff subshells == N_elec
  allocate( pop(0:P%Z) )                ! pop. of all species; C, C+, C++, ...
  allocate( pop_count(0:P%Z) )          ! integer count for each trajectory
  allocate( pop_old(0:P%Z) )
  allocate( pop_avg(0:P%Z) )            ! time-averaged quantities
  allocate( init_pot(F%N) )             ! potential for the initial config.
  allocate( N_ph_count(1:P%Z) )         ! sum of # abs. photons for each charge
  allocate( N_ph_avg(1:P%Z) )           ! avg. # abs. photons for each charge
  !
  ! counter for all processes for each subshell
  !   N_proc(k,i), k : process index; k=1 for P, k=2 for A, k=3 for F, ...
  !                i : core orbital index
  !
  allocate( N_proc(10,maxval(P%init_config(:,1))) )
  !
  ! define the temporal grid
  !   N_time = N' + N_ext
  !            N' : # of temporal grid points given by P%N_time / 100
  !            N_ext : more grid points extended by P%N_time_ext
  !   equal spacing up to N' (arithmetic sequence)
  !   then geometric sequence for the rest
  ! if P%yes_load_pulse, then N_time = P%N_time-1
  !
  Np_time = max( 100, P%N_time / 100 )  ! let's put a fixed # for the time being
  N_time = Np_time + P%N_time_ext       ! total number of temporal grid points
  allocate( time_grid(0:N_time) )
  allocate( pop_t(0:P%Z) )
  allocate( pop_t_count(0:N_time,0:P%Z) )
  if ( P%yes_occupation ) then
    allocate( occ_count(F%N_subshell) )
    allocate( occ_t_count(0:N_time,F%N_subshell) )
  end if
  dt = ( P%t_end_au - P%t_start_au ) / dble(Np_time)    ! common difference
  do i = 0, Np_time             ! arithmetic sequence during one pulse
    time_grid(i) = P%t_start_au + dble(i) * dt
  end do
  do i = Np_time + 1, N_time    ! geometric sequence for the rest
    time_grid(i) = time_grid(i-1) * 2.00_long           ! common ratio = 2
  end do
  !
  ! prepare temporal profile data (spline) if P%yes_load_pulse is on
  !
  if ( P%yes_load_pulse ) then
    if ( P%pulse_type /= 10 ) &
      call error_message( "Please set up pulse=file" )
    if ( P%MC_max_events /= 1 ) &       ! only one trajetory is meaningful
      call error_message( "MC_max_events=1 if the temporal profile is read in" )
    allocate( temporal_profile(P%N_time,2) )
    open( P%IO_PULSE, file=P%file_pulse )
    dt = ( P%t_end_au - P%t_start_au ) / dble(P%N_time - 1)
    do i = 1, P%N_time
      temporal_profile(i,1) = P%t_start_au + dble(i-1) * dt
      read( P%IO_PULSE, * ) temporal_profile(i,2)
    end do
    close( P%IO_PULSE )
    SP = Spline_create( temporal_profile(:,1), temporal_profile(:,2) )
  end if
  !
  ! prepare spectra data
  ! max. energy in the spectra is given by the max. photon energy
  !   1: photoelectrons, 2: Auger electrons, 3: fluorescence
  !
  if ( P%yes_spectra ) then
    max_E_spectra = nint( P%PE_start * AU2EV ) + 1
    if ( P%yes_harmonics ) then         ! consider higher E for harmonics
      if ( P%harmonic3 > ZERO ) then
        max_E_spectra = max_E_spectra * 3       ! third harmonic
      else if ( P%harmonic2 > ZERO ) then
        max_E_spectra = max_E_spectra * 2       ! second harmonic
      end if
    end if
!TEST   including four-photon process...
!    max_E_spectra = ( nint( P%PE_start * AU2EV ) + 1 ) * 4
!TEST
    allocate( count_spectra_traj(0:max_E_spectra,3) )
    allocate( count_spectra_charge(0:max_E_spectra,3,0:P%Z) )
    allocate( pop_e_avg(0:max_E_spectra) )
    count_spectra_traj(0:,:) = 0
    count_spectra_charge(0:,:,0:) = 0
    pop_e_avg(0:) = ZERO
  end if
  !
  ! set up the ground-state configuration of the neutral atom
  ! run SCF calculation for the first time and prepare WF_continuum
  ! It's necessary because it gives the largest rmax for 
  !   the continuum-state calculation in SR extend_WF_continuum().
  !
  j = 0
  do i = 1, P%N_orb             ! open-shell must locate at the end
    l = P%init_config(i,2)      ! quantum number l
    j = j + 4 * l + 2
    if ( j <= P%Z ) then 
       F%config(i,3) = 4 * l + 2   ! put max. occupation number
    else 
       F%config(i,3) = (4 * l + 2) - (j - P%Z)
       exit
    endif
  end do
  if ( sum( F%config(:,3) ) /= P%Z ) then      
    call error_message( "Number of electron is not equal to Z" )
  end if
  if ( any( F%config(:,3) /= P%init_config(:,3) ) ) then
    write(OUTPUT, "('# --- SCF calc. with the neutral ground-state " // &
      "config: ',a)") trim( config_to_string(F%config) )
    write(OUTPUT, "('#')")
  end if
  call SCF_procedure( P, G, F ) ! <-- always neutral ground-state config.
  !
  ! calculate N_config: # of all possible configurations, i.e. # of equations
  ! N_config is given by the ground-state of the neutral atom of Z
  !
  N_config = 1; j = 0
  do i = 1, F%N_subshell - 1    ! F%N_subshell = F%N_orb or F%N_orb_rel
    k = get_occ_num_max( F, i )
    j = j + k
    N_config = N_config * (k+1) ! possible config: 0...max_occ ==> x(max_occ+1)
  end do
  N_config = N_config * ( ( P%Z - j ) + 1 )
  !
  ! prepare uniform grids with the ground-state config. of the neutral
  ! if continuum_rmax is specified, then use it as a fixed continuum_rmax
  !
  if ( P%continuum_rmax > ZERO ) then
    call extend_WF_continuum( F, P%OUTPUT, P%continuum_rmax )
  else
    call extend_WF_continuum( F, P%OUTPUT )
  end if
  write(OUTPUT, "('#')")
  !
  ! run SCF calculation with given initial configuration,
  ! interpolation of the new potential [in extend_WF_continuum()]
  ! and get all atomic data for the initial configuration
  !
  call initialize_config( P, config )
  if ( any( F%config(:,3) /= P%init_config(:,3) ) ) then
    F%config(1:F%N_orb,:) = get_non_relativistic_config( P, config(:,:) )
    write(OUTPUT, "('# --- SCF calc. with the given initial config: ',a)") &
      trim( config_to_string(F%config) )
    write(OUTPUT, "('#')")
    call SCF_procedure( P, G, F )
    call extend_WF_continuum( F, P%OUTPUT )
    write(OUTPUT, "('#')")
  end if
  init_pot(:) = F%pot(:)        ! keep this potential for the initial config.
  P%OUTPUT = P%IO_NULL          ! yes_silent: turn off SCF information printing
!  ! shakeoff module
!  SO = CreateShakeoffData( P, G )
  call get_atomic_data( Q0, P, G, F, config, S )
  !
  ! prepare the elastic scattering calculations
  ! non-resonant: requires only the SCD file
  ! resonant: requires both the SCD and MAD files
  !
!  if ( P%yes_scattering ) then
!    !
!    ! prepare angular grids for integrating cross sections
!    !   extended-LGL: x%r(0...N+1) because it includes the end points.
!    ! CAUTION: it should be the same as the definition of x in scattering.f90
!    !
!    theta = TWO * asin( PI / P%resolution / alpha / P%PE_start )
!    x = generate_Grid( "extended-LGL", P%N_scd, cos(theta), ONE )
!    !
!    ! memory allocation
!    !
!    allocate( scd(N,0:P%N_scd+1) )
!    allocate( scd_f(N,0:2) )      ! f0(Q=0), f', f''
!    allocate( cs(0:P%N_scd+1), cs0(0:P%N_scd+1) )
!    allocate( n_dist(0:P%N_scd+1), n0_dist(0:P%N_scd+1) )
!    n_dist(:) = ZERO
!    !
!    ! read SCD (and MAD) files
!    !
!    call load_scd( P, S, x, scd, scd_f )
!    !
!    ! set up cross section of the neutral atom as undamaged one
!    !
!    cs0(:) = scd(N,:)
!  end if
  !
  ! consider synchrotron simulation
  !
!  if ( P%yes_synchrotron .and. P%i0 == N ) then
!    rate_sum = ZERO
!    do k = S%ind(P%i0,1), S%ind(P%i0,2)
!      if ( S%cs_order(k) /= 1 ) cycle ! look for P only
!      prob = S%value(k)               ! treated as a shake-off branching ratio
!      rate_sum = rate_sum + prob      ! it's not a rate_sum, but just a sum
!    end do
!    if ( rate_sum /= ZERO ) then
!      do k = S%ind(P%i0,1), S%ind(P%i0,2)
!        if ( S%cs_order(k) /= 1 ) cycle
!        S%value(k) = S%value(k) / rate_sum      ! make it as a probability
!      end do
!    end if
!  end if
  !
  ! prepare some additional files
  !
  if ( P%yes_pathway ) then
    open( 200, file=trim(P%file_out) // ".path", action="write" )
  end if
!
! foobar.occ becomes a pretty big file, but it's rather not complete.
! instead of foobar.occ, we need to think about a full trajectory file...
!
!  if ( P%yes_occupation ) then
!    open( 300, file=trim(P%file_out) // ".occ", action="write" )
!    write( 300, "('#',a4)", advance="no" ) "Q"
!    do i = 1, F%N_subshell
!      write( 300, "(a5)", advance="no" ) &
!        trim( orbital_name_from_config( config, i ) )
!    end do
!    write( 300, "()" )
!  end if
  !
  ! When P%yes_track_res is on, xatom_mc_fly starts to track electric
  ! configurations after resonant excitation.  
  !
  if ( P%yes_track_res ) call open_track_res( P, tec )
  !
  ! if P%yes_track_cs is on, xatom_mc_fly tracks a charge state of trajectories
  !
  if ( P%yes_track_cs ) call open_track_cs( P, tec )
  !
  ! set up filename format and print more info. when scanning is on
  !
  N_acc_traj  = 0       ! N of trajectories accumulated for a series
  N_acc_visit = 0       ! N of config. visited accumulated for a series
  N_acc_store = 0       ! N of config. stored in memory for a series
  N_acc_calc  = 0       ! N of config. visited accumulated for a series
  N_acc_proc  = 0       ! N of processes stored in memory for a series
  if ( P%yes_scan ) then
    call check_time( int_time, timer )
    timer_old = timer
    N_acc_time  = 0     ! CPU time accumulated for a series
    N_acc_res = 0       ! N of resonant excitations for a series
    N_acc_so = 0        ! N of shakeoff processes for a series
    OUTPUT_MAIN = OUTPUT
    OUTPUT = 999        ! IO number for each scanning file
    i = int( log10( dble(P%scan_end) ) ) + 1
    write( format_string, "('(3(a,i',i1,'))')" ) i
    write( OUTPUT_MAIN, "(a)", advance="no" ) "# Scanning "
    if ( P%scan_quantity == 1 ) then            ! varying fluence
      write( OUTPUT_MAIN, "(a)", advance="no" ) "fluence"
      scan_v0 = P%fluence
    else if ( P%scan_quantity == 2 ) then       ! varying pulse duration
      write( OUTPUT_MAIN, "(a)", advance="no" ) "pulse duration"
      scan_v0 = P%FWHM
    end if
    if ( P%scan_type == 1 ) then                ! linear increment
      write( OUTPUT_MAIN, "(a)" ) " in arithmetic progression"
    else if ( P%scan_type == 2 ) then           ! exponential increment
      write( OUTPUT_MAIN, "(a)" ) " in geometric progression"
    end if
    write( OUTPUT_MAIN, format_string ) "#   index scanned from ", &
      P%scan_start, " to ", P%scan_end, " by ", P%scan_step
    if ( P%scan_type == 1 ) then
      write( OUTPUT_MAIN,"('#   value given by v = v0 + d * index (in a.u.)')" )
      write( OUTPUT_MAIN,"('#   common diff.   d = ',f14.3)" ) P%scan_common
    else if ( P%scan_type == 2 ) then
      write( OUTPUT_MAIN,"('#   value given by v = v0 * r**index (in a.u.)')" )
      write( OUTPUT_MAIN,"('#   common ratio   r = ',f14.3)" ) P%scan_common
    end if
    write( OUTPUT_MAIN, "('#   initial value v0 = ',f14.3,'  [ ')", &
      advance="no" ) scan_v0
    if ( P%scan_quantity == 1 ) then
      write( OUTPUT_MAIN, "(es11.3,' ph/micron^2 ]')" ) scan_v0 / &
        ( AU2M / 1.0e-6_long )**2
    else if ( P%scan_quantity == 2 ) then
      write( OUTPUT_MAIN, "(f14.3,' fs ]')" ) scan_v0 * AU2SEC / 1.0e-15_long
    end if
    write( OUTPUT_MAIN, "('#')" )
    write( OUTPUT_MAIN, "('#   ',a)" ) "- N_traj  : N of trajectories"
    write( OUTPUT_MAIN, "('#   ',a)" ) &
      "- N_visit : N of configurations newly visited"
    write( OUTPUT_MAIN, "('#   ',a)" ) &
      "- N_store : N of configurations stored in memory"
    write( OUTPUT_MAIN, "('#   ',a)" ) &
      "- N_calc  : N of configurations calculated"
    write( OUTPUT_MAIN, "('#   ',a)" ) "- CPU: CPU time in seconds"
    write( OUTPUT_MAIN, "('#')" )
    write( OUTPUT_MAIN, "('#',a7,5a8,2x,a)" ) "index", &
      "N_traj", "N_visit", "N_store", "N_calc", "CPU", "filename"
    write( format_string, "('(a,i',i1,'.',i1,',a)')" ) i, i     ! used later
  end if
  !=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  ! SCAN with different values: fluence or pulse duration
  ! BE CAREFUL WITH INDENTATION!!!
  !=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  SCAN: do ii = P%scan_start, P%scan_end, P%scan_step
  !--------------------------------------------------------------------------
  ! initialize for each calculation
  !--------------------------------------------------------------------------
  pop_count(0:) = 0
  pop(0:) = ZERO
  pop_avg(0:) = ZERO
  N_ph_count(:) = 0
  N_ph_avg(:) = ZERO
  pop_t(0:) = ZERO
  pop_t_count(0:,0:) = 0
  if ( P%yes_occupation ) then
    occ_count(:) = 0
    occ_t_count(0:,:) = 0
  end if
  if ( P%yes_spectra ) then
    pop_e_avg(0:) = ZERO
    count_spectra_charge(0:,:,0:) = 0
  end if
  P%N_config_visited = 0        ! counter for N_config newly visited
  P%N_config_stored = 0         ! counter for N_config stored in memory
  P%N_config_calc = 0           ! counter for N_config calculated
  P%N_process_stored = 0        ! counter for N_process stored in memory
  N_proc(:,:) = 0               ! counter for all processes for each subshell
  N_shakeoff = 0                ! counter for shakeoff processes
  !
  ! set up a fluence value or a pulse duration when scanning is on
  !
  if ( P%yes_scan ) then
    if ( P%scan_type == 1 ) then                ! arithmetic progression
      scan_v = scan_v0 + P%scan_common * dble(ii)
    else if ( P%scan_type == 2 ) then           ! geometric progression
      scan_v = scan_v0 * P%scan_common**ii
    end if
    if ( P%scan_quantity == 1 ) then
      !
      ! varying fluence: set up P%fluence with scan_v
      !
      P%fluence = scan_v
    else if ( P%scan_quantity == 2 ) then
      !
      ! varying pulse duration: set up not only P%FWHM, but also time-related 
      !   quantities ( P%dt_au, P%t_start_au, P%t_end_au, and time_grid(:) )
      ! At this point,
      !   scan_v must contain the new tau, where tau is the pulse duration.
      !   P%FWHM must contain the old tau.
      ! So, all time-related quantities should be re-scaled with scan_v/P%FWHM.
      ! Then, P%FWHM needs to be assigned with the new value, i.e., scan_v.
      !
      value = scan_v / P%FWHM   ! rescaling factor
      P%dt_au      = P%dt_au      * value
      P%t_start_au = P%t_start_au * value
      P%t_end_au   = P%t_end_au   * value
      time_grid(:) = time_grid(:) * value
      P%FWHM = scan_v
    end if
    !
    ! open up a new file for each scanning point
    ! and then print out outputs there
    !
    write( file_scan, format_string ) trim(P%filename) // "_i", ii, ".out"
    open( OUTPUT, file=trim(file_scan) )
    write(OUTPUT, "('# ',a)" ) repeat( "-", 77 )
    write(OUTPUT, "('# index = ',i5)") ii
    if ( P%scan_quantity == 1 ) then            ! varying fluence
      write(OUTPUT, "('# fluence (1/a.u.^2 and ph/micron^2) = ',2es18.10)") &
        P%fluence, P%fluence / ( AU2M / 1.0e-6_long )**2
    else if ( P%scan_quantity == 2 ) then       ! varying pulse duration
      write(OUTPUT, "('# pulse duration (a.u. and fs) = ',2es18.10)") &
        P%FWHM, P%FWHM * AU2SEC / 1.0e-15_long
    end if
    write(OUTPUT, "('# ',a)" ) repeat( "-", 77 )
    write(OUTPUT, "('#')")
  end if
  !
  ! print out the heading
  !
  call print_divider( OUTPUT, "Monte Carlo calculation" )
  write(OUTPUT, "('#',a9,2a10)", advance="no") "event", "diff", "1-P_tot"
  call print_all_ions( OUTPUT, P%Z )
  write(OUTPUT, "()")
  !============================================================================
  ! Monte-Carlo iterations
  !============================================================================
  event: do j = 1, P%MC_max_events
    is_decay_happened = .false. 
    call initialize_config( P, config ) ! copy the initial elec. config from P
    charge = P%Z - P%N_elec
    F%pot(:) = init_pot(:)              ! use potential from the initial config
    call get_atomic_data( Q0, P, G, F, config, S )
    dt = P%dt_au        ! it's supposed to be changed during time_evolution.
    t = P%t_start_au    ! time_evolution from t_start_au until no further decay
    N_ph = 0            ! # of photons involved in this trajectory
    k_time = 0          ! k_time: index for time_grid(:)
!    if ( P%yes_pathway ) then
!      write(200,"(es13.5,i4)") t * AU2FS, charge
!    end if
    if ( P%yes_spectra ) count_spectra_traj(0:,:) = 0 ! initialize at every traj
    !
    ! consider synchrotron simulation
    !
!    if ( P%yes_synchrotron ) then
!      rand = random()
!      process_P: do k = 1, S%N_process
!        if ( S%cs_order(k) /= 1 ) cycle ! look for P only
!        prob = S%value(k)               ! treated as a shake-off branching ratio
!        if ( rand < prob ) then         ! hurrah! we want to go this pathway
!          ic = S%col(k)                 ! set up a new configuration
!          exit process_P
!        end if
!        rand = rand - prob
!      end do process_P
!    end if
    !
    ! time evolution
    !
    time_evolution: do
      !
      ! compute time-averaged quantity
      !
      if ( t <= P%t_end_au ) then
        p_env = pulse_envelope( P, SP, t + dt * HALF )
        pop_avg(charge) = pop_avg(charge) + p_env * dt
      end if
      !
      ! compute time-dependent quantities from t to t+dt
      ! - pop_t_count: time-dependent population
      ! - occ_t_count: time-dependent occupation number
      !
      do k = k_time, N_time     ! time_grid(0:N_time)
        if ( k /= N_time .and. time_grid(k) >= t + dt ) exit
        pop_t_count(k,charge) = pop_t_count(k,charge) + 1
        if ( P%yes_occupation ) then
          occ_t_count(k,:) = occ_t_count(k,:) + config(:,occ_index)
        end if
      end do
      k_time = k
      if ( k_time == N_time ) k_time = N_time + 1       ! avoid double-counting 
        !
        ! WHERE statement below is much slower than just DO...LOOP statement
        !
        !if ( P%yes_occupation == .false. ) then  ! only population counted
        !  where( time_grid(:) >= t .and. time_grid(:) < t + dt ) &
        !    pop_t_count(:,charge) = pop_t_count(:,charge) + 1
        !else                    ! population and occupation number counted
        !  do k = 0, N_time
        !    if ( time_grid(k) < t ) cycle
        !    if ( time_grid(k) >= t + dt ) exit
        !    pop_t_count(k,charge) = pop_t_count(k,charge) + 1
        !    if ( P%yes_relativity ) then
        !      occ_t_count(k,:) = occ_t_count(k,:) + config(:,4)
        !    else
        !      occ_t_count(k,:) = occ_t_count(k,:) + config(:,3)
        !    end if
        !  end do
        !end if
      !
      ! time propagation: t is in a.u. and dt is varied during time propagation
      !
      t = t + dt
      !
      ! get a random number
      !
      !rand = random( stream )
      rand = random()
      rate_sum = ZERO
      !
      ! examine each process and determine "go" or "no-go"
      !
      loop_process: do k = 1, S%N_process
        rate = compute_rate( P, SP, S%value(k), S%tag(k), t )
        prob = rate * dt
        ! ---------------------------------------------------------------------
        ! hurrah! let's move! we want to go this pathway
        ! ---------------------------------------------------------------------
        if ( rand < prob ) then
          is_decay_happened = .true.
          p_type = S%p_type(k)
          tag = S%tag(k)
          N_ph = N_ph + number_of_photons( tag )
          !
          ! counter for all processes for each subshell
          !   i_core: index for core subshell
          !   quantum_n: quantum number n for the i_core-th subshell
          !
          if ( p_type == 2 .or. p_type == 3 ) then      ! A or F
            i_core = S%occ_up(1,k)    ! A or F: occ# of core orb. will be up
          else
            i_core = S%occ_down(1,k)  ! P or E: occ# of core orb. will be down
          end if
          quantum_n  = S%config(i_core,1)               ! quantum number n
          N_proc( p_type, quantum_n ) = N_proc( p_type, quantum_n ) + 1
          !
          ! calculate remaining rates for rate_sum
          !
          do i = k, S%N_process
            rate_sum = rate_sum + compute_rate( P, SP, S%value(i), S%tag(i), t )
          end do
          !
          ! consider electron/photon spectra
          !
          if ( P%yes_spectra ) then
            call record_spectra( P, SP, S, k, t, max_E_spectra, &
                                 count_spectra_traj, pop_e_avg )
          end if
          !
          ! save old data for printing pathways or calculating shakeoff ratios
          !
          charge_old = charge           ! old charge for pathways
          E_old = S%E_tot               ! old total energy for pathways
          E_eV_before_relax = S%E0 * AU2EV - S%E_eV(k)
          if ( S%p_type(k) == 1 .or. S%p_type(k) == 4 ) then
            E_eV_before_relax = E_eV_before_relax + P%PE_start * AU2EV
          else if ( S%p_type(k) > 4 ) then
            E_eV_before_relax = E_eV_before_relax + &
                                P%PE_start * AU2EV * dble( S%p_type(k) - 3 )
          end if
          str_process = label_Process( S, k )
          !
          ! set up a new configuration: k-th process
          ! and obtain its atomic data
          !
          config(:,:) = config_Process( S, k ) ! proceed the k-th decay channel
          !
          ! consider shakeoff process
          ! if shakeoff happens, then config(:,:) will be updated
          !
          if ( P%yes_shakeoff ) &
            call consider_shakeoff( Q0, P, G, F, S, config, N_shakeoff )

          ! if yes_track_cs is on, record a charge state and time
          if ( P%yes_track_cs ) call track_cs( P, tag, config, t, tec )
          ! if yes_track_res is on, record the electric configuration
          if ( P%yes_track_res ) call track_res( P, tag, config, t, tec )
!          ! shake-off
!          if ( P%yes_shakeoff ) then
!             call ProceedShakeoff( config, k, SO, prob )
!             if ( SO%is_so_happened ) then
!                ! if yes_track_cs is on, record a charge state and time
!                if ( P%yes_track_cs ) call track_cs( P, 'S', config, t, tec )
!                ! if yes_track_res is on, record the electric configuration
!                if ( P%yes_track_res ) call track_res( P, 'S', config, t, tec )
!             endif
!          endif
          !
          ! ionic charge
          !
          charge = P%Z - sum( config(:,occ_index) )
          !
          ! get atomic data for a new config. specified with "config(:,:)"
          !
          call get_atomic_data( Q0, P, G, F, config, S )
          !
          ! print out pathway details
          ! NOTE: It looks like yes_pathway doesn't work with yes_shakeoff!!!
          !
          if ( P%yes_pathway ) then
            E_eV_after_relax = S%E0 * AU2EV
            write(200,"(es13.5,2(i4,f16.8),' |',3i8,' |',2x,a)") &
              t * AU2FS, charge_old, E_old, charge, S%E_tot, &
              int( E_eV_before_relax ), int( E_eV_after_relax ), &
              int( E_eV_after_relax - E_eV_before_relax ), &
              trim(str_process)
          end if
          !
          ! we don't need to be in this loop any more; let's escape
          !
          exit loop_process
        end if
        rand = rand - prob
        rate_sum = rate_sum + rate
      end do loop_process
      !
      ! stop this time evolution if there is no further decay available
      !
      if ( rate_sum == ZERO ) then      ! pop. and occ. number counted
        do k = k_time, N_time
          pop_t_count(k,charge) = pop_t_count(k,charge) + 1
          if ( P%yes_occupation ) then
            occ_t_count(k,:) = occ_t_count(k,:) + config(:,occ_index)
          end if
        end do
        if ( P%yes_occupation ) then    ! final occ. numbers into occ_count(:)
          occ_count(:) = occ_count(:) + config(:,occ_index)
!          write( 300, "(999i5)" ) charge, config(:,occ_index)
        end if
        time_evol_end: do               ! time propagation until the end
          p_env = pulse_envelope( P, SP, t + dt * HALF )
          pop_avg(charge) = pop_avg(charge) + p_env * dt
          t = t + dt
          if ( t > P%t_end_au ) exit time_evol_end
        end do time_evol_end
        if ( P%yes_spectra ) then       ! put spectra info with the final charge
          count_spectra_charge(0:,1:3,charge) = &
            count_spectra_charge(0:,1:3,charge) + count_spectra_traj(0:,1:3)
        end if
        exit time_evolution
      end if
      !
      ! adjust dt according to the sum of rates
      !
      if ( t <= P%t_end_au ) then
        dt = min( P%MC_dt_ratio / rate_sum, P%dt_au )
      else
        dt = P%MC_dt_ratio / rate_sum                   ! no need to have dt_max
      end if
    end do time_evolution
    !
    ! print out pathway information if it's turned on
    !
    if ( P%yes_pathway ) then
      write(200,"('# j = ',i8,5x,'N_ph = ',i5,5x,'final charge = ',i5,//)") &
        j, N_ph, charge
      !write(200,"('# final config = ',a,//)") trim( config_to_string( config ) )
    end if
    !
    ! compute the number of scattered photons (angular distribution)
    !     d N                                            d sigma
    !   ------- = INT dt [ flux(t) * SUM P_{conf}(t) * [ ------- ]_{conf} ]
    !   d Omega                     {conf}               d Omega
    !
!    if ( P%yes_scattering ) then
!      do k = 0, P%N_scd + 1
!        cs(k) = sum( C(:) * scd(:,k) )
!        n_dist(k) = n_dist(k) + P%fluence * p_env * cs(k) * dt
!      end do
!      !
!      ! the above implementation is almost 2 times faster than the below
!      !
!      !cs(:) = ZERO
!      !do j = 1, N
!      !  if ( C(j) > 1.0e-8_long ) cs(:) = cs(:) + C(j) * scd(j,:)
!      !end do
!      !n_dist(:) = n_dist(:) + P%fluence * p_env * cs(:) * dt
!    end if
    !
    ! counter increased
    !
    pop_count(charge)  = pop_count(charge)  + 1
    if ( N_ph > 0 .and. charge > 0 ) &
      N_ph_count(charge) = N_ph_count(charge) + N_ph
    !
    ! print out and check convergence
    !
    if ( mod(j,P%MC_step) == 0 ) then
      pop_old(0:) = pop(0:)
      pop(0:) = dble( pop_count(0:) ) / dble(j)
      where ( pop(0:) < 1.0e-99_long ) pop(0:) = ZERO
      diff = maxval( abs( pop_old(0:) - pop(0:) ) )
      write(OUTPUT, "(i10,es10.3,sp,es10.2,s,100es10.3)") &
        j, diff, ONE - sum( pop(0:) ), pop(0:)
      if ( diff < P%MC_conv ) exit event ! never exit until the end if MC_conv=0
    end if
    !
    ! track resonant excitations or trajectories of charge states
    !
    if ( P%yes_track_res .and. tec%track_res_started .and. is_decay_happened ) then
      write(P%unit_track_res,*)
      write(P%unit_track_res,*)
      tec%track_res_started = .false.
    endif
    if ( P%yes_track_cs .and. tec%track_cs_started .and. is_decay_happened ) then 
      write(P%unit_track_cs,*)
      write(P%unit_track_cs,*)
    endif
  end do event
  !============================================================================
  !
  ! close files for tracking resonant excitations or traj. of charge states
  !
  if ( P%yes_track_res ) call close_track_res( P )
  if ( P%yes_track_cs ) call close_track_cs( P )
  !
  ! average all quantities for many events
  ! NOTE: j is # of events or trajectories
  !       but, j becomes max+1 after full iterations of the do-loop,
  !       which needs to be corrected to max.
  !
  if ( j > P%MC_max_events ) j = P%MC_max_events
  N_event = j                   ! assign # of events or trajectories
  pop(0:) = dble( pop_count(0:) ) / dble(N_event)
  pop_avg(0:) = pop_avg(0:) / dble(N_event)
  where ( pop_count(1:) /= 0 ) &
    N_ph_avg(1:) = dble( N_ph_count(1:) ) / dble( pop_count(1:) )
  where ( pop(0:)     < 1.0e-99_long ) pop(0:) = ZERO
  where ( pop_avg(0:) < 1.0e-99_long ) pop_avg(0:) = ZERO
  charge_avg  = sum( (/ ( dble(i) * pop_avg(i), i = 0, P%Z ) /) )
  !
  ! print out summary of populations (or probabilities)
  ! - |POP|     ones after the pulse and a long long time
  ! - |POP_AVG| ones averaged with weights by the normalized pulse
  ! - |POP_PULSE_END| ones at the end of the pulse; for back-compatibility only
  !
  write(OUTPUT, "('#',a9,20x,100es10.3)") "|POP|    ", pop(0:)
  write(OUTPUT, "('#',a9,20x,100es10.3)") "|POP_AVG|", pop_avg(0:)
  write(OUTPUT, "(/)")
  !
  ! print out charge-state distribution
  !
  call print_divider( OUTPUT, "Charge-state distribution" )
  call print_charge_states( OUTPUT, pop, pop_avg, N_ph_avg )
  write(OUTPUT, "(/)")
  !
  ! print out population dynamics
  !
  call print_divider( OUTPUT, "Population dynamics during the pulse" )
  write(OUTPUT, "('#',a9,2a10)", advance="no") "t(fs)", "pulse_env", "1-P_tot"
  call print_all_ions( OUTPUT, P%Z )
  write(OUTPUT, "(a10)") "charge"
  do i = 0, N_time
    t = time_grid(i)
    p_env = pulse_envelope( P, SP, t )
    if ( p_env < 1e-99_long ) p_env = ZERO
    pop_t(0:) = dble( pop_t_count(i,0:) ) / dble(N_event)
    write(OUTPUT, "(es10.2,es10.3,sp,es10.2,s,100es10.3)") &
      t * AU2FS, p_env, ONE - sum( pop_t(0:) ), pop_t(0:), &
      sum( (/ ( pop_t(k) * dble(k), k=1, P%Z ) /) )
    if ( P%fluence /= ZERO .and. i == Np_time ) &  ! Np_time: at the pulse end
      write(OUTPUT, "('#',a15,14x,100es10.3)") "|POP_PULSE_END|", pop_t(0:)
  end do
  write(OUTPUT, "(/)")
  !
  ! print out some important configurations (all if verbose mode)
  !
!  write(OUTPUT, "('# Probabilities of individual configurations:')")
!  write(OUTPUT, "('#   ',21x,2a12,4x,a)") "P(t_end)", "P_avg", "description"
!  do j = 1, N
!    if ( .not. P%yes_verbose .and. &
!         C(j) < 1.0e-5_long .and. C_avg(j) < 1.0e-5_long ) cycle
!    if ( P%yes_verbose .and. C_avg(j) == ZERO ) cycle
!    call index_to_config( j, config )
!    write(OUTPUT, "('#   config = ',i10,2x,2es12.5,' ; ',sp,i3,': ',a)") &
!      j, C(j), C_avg(j), charge(j), trim( config_to_string( config ) )
!  end do
  !
  ! print out electron and photon spectra
  !
  if ( P%yes_spectra ) then
    pop_e_avg(0:) = pop_e_avg(0:) / dble(N_event)
    call print_divider( OUTPUT, "Electron/photon spectra" )
    write(OUTPUT, "('#   ',a6,3a13)") &
      "E(eV)", "Photoelec.", "Auger elec.", "fluorescence"
    N_ph_abs = sum( N_ph_avg(1:) * pop(1:) )    ! pop(0) is not included!
    E_absorbed = N_ph_abs * P%PE_start * AU2EV
    E_emitted = ZERO
    sum_spectra(:) = ZERO
    do j = 0, max_E_spectra
      write(OUTPUT, "(4x,i6)", advance="no") j
      do k = 1, 3
        value = dble( sum( count_spectra_charge(j,k,0:) ) ) / dble(N_event)
        if ( value < 1.0e-99_long ) value = ZERO
        write(OUTPUT, "(es13.5)", advance="no") value
        sum_spectra(k) = sum_spectra(k) + value
        if ( k == 3 ) E_emitted = E_emitted + dble(j) * value ! fluo. only in eV
      end do
      write(OUTPUT, "()")
    end do
    write(OUTPUT, "('#|SPECTRA|',3es13.5)") sum_spectra(:)
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# Energy gain from incident photons (eV) = ',f10.2)") &
                  E_absorbed
    write(OUTPUT, "('# Energy loss by emitted photons (eV)    = ',f10.2)") &
                  E_emitted
    write(OUTPUT, "(/)")
    !
    ! print out spectra for given charge states (coincident measurement)
    !
    if ( P%yes_coincidence ) then       ! charge-resolved spectra
      do i = 0, P%Z
        write( filename, "(a,'.spectra.',i2.2)" ) trim(P%file_out), i
        open( 301+i, file=trim(filename), action="write" )
        do j = 0, max_E_spectra
          write(301+i, "(4x,i6,3es13.5)") j, &
            count_spectra_charge(j,1:3,i) / dble(N_event)
        end do
        close( 301+i )
      end do
    end if
  end if
  !
  ! print out foobar.pop for post-process of scattering
  !
  if ( P%yes_population ) then
    open( P%IO_POP, file=P%file_pop, form="formatted" )
    !
    ! print out pulse-weighted time-averaged populations 
    ! for all charge states, not for all configurations
    !
    write(P%IO_POP, "('# Pulse-weighted time-averaged populations'/'#')")
    write(P%IO_POP, "('# Charge-state populations:')")
    write(P%IO_POP, "('# printed with the ground-state config. only')")
    write(P%IO_POP, "('# Charge    Pop_avg')")
    do i = 0, P%Z
      if ( pop_avg(i) > ZERO ) &
        write(P%IO_POP, "(i3,2x,es18.10)") i, pop_avg(i)
    end do
    write(P%IO_POP, "(/)")
    !
    ! print out time-averaged population for free electrons
    !
    if ( P%yes_spectra ) then
      write(P%IO_POP, "('# Free electron populations')")
      write(P%IO_POP, "('#',a5,a18)") "E(eV)", "Pop_avg"
      do j = 0, max_E_spectra
        if ( pop_e_avg(j) > 1e-99_long ) &
          write(P%IO_POP, "(i6,es18.10)") j, pop_e_avg(j)
      end do
      write(P%IO_POP, "('# sum:',es18.10)") sum( pop_e_avg(0:) )
    end if
  end if
  !
  ! print out occupation number dynamics
  !
  if ( P%yes_occupation ) then
    call print_divider( OUTPUT, "Occupation number dynamics during the pulse" )
    write(OUTPUT, "('#',a9,a10)", advance="no") "t(fs)", "pulse_env"
    do i = 1, F%N_subshell
      write(OUTPUT, "(a10)", advance="no") &
        trim( orbital_name_from_config( config, i ) )
    end do
    write(OUTPUT, "(a10)") "charge"
    do k = 0, N_time
      t = time_grid(k)
      p_env = pulse_envelope( P, SP, t )
      if ( p_env < 1e-99_long ) p_env = ZERO
      write(OUTPUT, "(es10.2,es10.3,999f10.6)") &
        t * AU2FS, p_env, dble( occ_t_count(k,:) ) / dble(N_event), &
        dble(P%Z) - sum( dble( occ_t_count(k,:) ) / dble(N_event) )
      if ( P%fluence /= ZERO .and. k == Np_time ) then  ! Np_time: at pulse end
        write(OUTPUT, "('#',a15,4x,999f10.6)") "|OCC_PULSE_END|", &
          dble( occ_t_count(k,:) ) / dble(N_event)
      end if
    end do
    write(OUTPUT, "('#',a9,10x,999f10.6)") "|OCC|    ", &
      dble( occ_count(:) ) / dble(N_event)
    write(OUTPUT, "(/)")
  end if
  !
  ! print out # of scattered photons and R factor
  !
!  if ( P%yes_scattering ) then
!    n0_dist(:) = P%fluence * cs0(:)     ! # from undamaged
!    n_tot  = TWO*PI * sum( n_dist(0:)  * x%weight(0:) ) ! 2pi from INT d(phi)
!    n0_tot = TWO*PI * sum( n0_dist(0:) * x%weight(0:) ) ! 2pi from INT d(phi)
!    !
!    ! R-factor defined in Neutze, et al., Nature 406, 752 (2000)
!    ! R = SUM_Omega | sqrt(I(Omega)) / I_tot - sqrt(I0(Omega)) / I0_tot |
!    !   where I_tot  = SUM_Omega sqrt( I(Omega)  )    I:  damaged
!    !         I0_tot = SUM_Omega sqrt( I0(Omega) )    I0: undamaged
!    !
!    I_tot  = TWO*PI * sum( sqrt(n_dist(0:))  * x%weight(0:) )
!    I0_tot = TWO*PI * sum( sqrt(n0_dist(0:)) * x%weight(0:) )
!    R_factor = TWO*PI * sum( abs( sqrt(n_dist(0:))/I_tot - &
!                                  sqrt(n0_dist(0:))/I0_tot ) * x%weight(0:) )
!    write(OUTPUT, "('#')")
!    write(OUTPUT, "('# Scattering process with x-ray damage')")
!    write(OUTPUT, "('#   Resolution d (angstrom) = ',f8.3)") &
!      P%resolution * AU2ANG
!    write(OUTPUT, "('#   Q_max = 2pi / d (a.u.)  = ',es14.7)") &
!      TWO*PI / P%resolution
!    write(OUTPUT, "('#   N_scattered [undamaged] = ',es14.7)") n0_tot
!    write(OUTPUT, "('#   N_scattered [damaged]   = ',es14.7)") n_tot
!    write(OUTPUT, "('#   R-factor                = ',es14.7)") R_factor
!    !
!    ! time-averaged form factors with the dispersion correction
!    !   a(0,omega) = f2_avg
!    !   b(0,omega) = ( f0_avg + fp_avg ) * 2
!    !   c(omega) = fpp_avg * 2
!    !   a~(0,omega) = ff_avg
!    !
!    ! f2_avg = sum_I1 [ int dt J(t) P_I1(t) ] | f_1,I1(0,omega) |^2
!    ! ff_avg = int dt J(t) | sum_I1 P_I1(t) f_1,I1(0,omega) |^2
!    !
!    if ( P%yes_dispersion ) then
!      f2_avg  = sum( ( ( scd_f(:,0) + scd_f(:,1) )**2 + scd_f(:,2)**2 ) * &
!                     C_avg(:) )
!      f0_avg  = sum( scd_f(:,0) * C_avg(:) )
!      fp_avg  = sum( scd_f(:,1) * C_avg(:) )
!      fpp_avg = sum( scd_f(:,2) * C_avg(:) )
!      write(OUTPUT, "('#')")
!      write(OUTPUT, "('# Time-averaged form factors (Q=0)')")
!      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f0",    f0_avg
!      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f'",    fp_avg
!      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "f''",   fpp_avg
!      write(OUTPUT, "('#   ',a5,' = ',es14.7)") "|f|^2", f2_avg
!      write(OUTPUT, "('#')")
!      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a(0,omega)", f2_avg
!      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "b(0,omega)", &
!        TWO * ( f0_avg + fp_avg )
!      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "c(omega)", TWO * fpp_avg
!      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a~(0,omega)", ff_avg
!    end if
!    !
!    ! print out summary
!    !
!    write(OUTPUT, "('#')")
!    write(OUTPUT, "('# PUNCH ',6a12)") "fluence(au)", "FWHM(au)", &
!      "N(undam.)", "N(damaged)", "R-factor", "avg. charge"
!    write(OUTPUT, "('#|PUNCH|',6es12.5)") &
!      P%fluence, P%FWHM, n0_tot, n_tot, R_factor, charge_avg
!    if ( P%yes_dispersion ) then
!      !write(OUTPUT, "('# PUNC2 ',6a12)") "PE(eV)", &
!      !  "avg. f0", "avg. f'", "avg. f''", "avg. f_real", "avg. |f|^2"
!      !write(OUTPUT, "('#|PUNC2|',f12.2,5es12.4)") P%PE_start * AU2EV, &
!      !  f0_avg, fp_avg, fpp_avg, f0_avg + fp_avg, f2_avg
!      write(OUTPUT, "('# PUNC2 ',6a12)") "PE(eV)", &
!        "a(0,omega)", "b(0,omega)", "c(omega)", "a~(0,omega)"
!      write(OUTPUT, "('#|PUNC2|',f12.2,4es12.4)") P%PE_start * AU2EV, &
!        f2_avg, TWO * ( f0_avg + fp_avg ), TWO * fpp_avg, ff_avg
!    end if
!  end if
  !
  ! print out how many config. are visited and stored during MC calculations
  !
  call print_divider( OUTPUT, "Monte Carlo statistics" )
  write(OUTPUT, "('# - number of trajectories            = ',i15)") N_event
  write(OUTPUT, "('#   max. diff. of charge populations  = ',es15.5)") diff
  write(OUTPUT, "('#   diff. checked out per ',i5,' trajectories')") P%MC_step
  write(OUTPUT, "('# - N of all possible q-hole config.  = ',i15)") N_config
  write(OUTPUT, "('#   N of config. newly visited        = ',i15," // &
                "'  (',f6.2,'%)')") &
    P%N_config_visited, dble(P%N_config_visited) / dble(N_config) * 100_long
  write(OUTPUT, "('#   N of config. stored in this run   = ',i15," // &
                "'  (',f6.2,'%)')") &
    P%N_config_stored, dble(P%N_config_stored) / dble(N_config) * 100_long
  write(OUTPUT, "('#   N of config. calculated           = ',i15," // &
                "'  (',f6.2,'%)')") &
    P%N_config_calc, dble(P%N_config_calc) / dble(N_config) * 100_long
  write(OUTPUT, "('#   N of processes stored in this run = ',i15)") &
    P%N_process_stored
  write(OUTPUT, "('#   atomic data are stored in memory " // &
                "after visited ')", advance="no" )
  if ( P%max_counter_visit == 1 ) then
    write(OUTPUT, "('once')")
  else
    write(OUTPUT, "(i2,' times')") P%max_counter_visit
  end if
  !
  ! print out process info
  !
  if ( P%yes_resonance ) then
    N_res_excitation = sum( N_proc(4,:) )       ! 4: photoexcitation
    write(OUTPUT, "('# - occurence of resonant excitations = ',i15)") &
      N_res_excitation
    write(OUTPUT, "('#   N of resonance per trajectory     = ',f15.3)") &
      dble( N_res_excitation ) / dble( N_event )
  end if
  if ( P%yes_shakeoff ) then
    write(OUTPUT, "('# - occurence of shakeoff processes   = ',i15)") &
      N_shakeoff
    write(OUTPUT, "('#   N of shakeoff per trajectory      = ',f15.3)") &
      dble( N_shakeoff ) / dble( N_event )
  end if
  N_PI(1) = sum( N_proc(1,:) )          ! single-photon ionization
  do k = 2, 5                           ! direct multiphoton ionization
    N_PI(k) = sum( N_proc(k+3,:) )      ! N_proc(5,:) -> two-ph, ...
  end do
  if ( any( N_PI(2:) > 0 ) ) then
    write(OUTPUT,"('# - occurence of one-photon ionization proc.   = ',i15)") &
      N_PI(1)
    write(OUTPUT,"('#   N of one-photon ionization per trajectory  = ',f15.3)")&
      dble( N_PI(1) ) / dble( N_event )
    write(OUTPUT,"('# - occurence of direct multiphoton ionization = ',4i7)") &
      N_PI(2:)
    write(OUTPUT,"('#   N of direct (2/3/4/5)-photon per traj.     = ',4f7.3)")&
      dble( N_PI(2:) ) / dble( N_event )
  end if
  write(OUTPUT, "('#         ',3a14)", advance="no") &
    "N_photoion", "N_Auger", "N_fluo"
  if ( P%yes_resonance ) write(OUTPUT, "(a14)", advance="no") "N_res_exc"
  if ( P%yes_shakeoff  ) write(OUTPUT, "(a14)", advance="no") "N_shakeoff"
  write(OUTPUT, "()")
  write(OUTPUT, "('# |N_PROC|',3f14.3)", advance="no") &
    dble(sum(N_proc(1,:))) / dble(N_event), &   ! N_proc(1,:) : photoionization
    dble(sum(N_proc(2,:))) / dble(N_event), &   ! N_proc(2,:) : Auger-Meitner
    dble(sum(N_proc(3,:))) / dble(N_event)      ! N_proc(3,:) : fluorescence
  if ( P%yes_resonance ) &
    write(OUTPUT,"(f14.3)",advance="no") dble(sum(N_proc(4,:))) / dble(N_event)
  if ( P%yes_shakeoff ) &
    write(OUTPUT,"(f14.3)",advance="no") dble(N_shakeoff) / dble(N_event)
  write(OUTPUT, "()")
  write(OUTPUT, "('#         ',7a10)") "tot", "K", "L", "M", "N", "O", "..."
  k = 1; write(OUTPUT, "('# |NUM_P| ',100f10.3)") &
    dble(sum(N_proc(k,:))) / dble(N_event), dble(N_proc(k,:)) / dble(N_event)
  k = 2; write(OUTPUT, "('# |NUM_A| ',100f10.3)") &
    dble(sum(N_proc(k,:))) / dble(N_event), dble(N_proc(k,:)) / dble(N_event)
  k = 3; write(OUTPUT, "('# |NUM_F| ',100f10.3)") &
    dble(sum(N_proc(k,:))) / dble(N_event), dble(N_proc(k,:)) / dble(N_event)
  if ( P%yes_resonance ) then
    k = 4; write(OUTPUT, "('# |NUM_E| ',100f10.3)") &
      dble(sum(N_proc(k,:))) / dble(N_event), dble(N_proc(k,:)) / dble(N_event)
  end if
  do k = 2, 5   ! direct multiphoton ionization; N_proc(5,:) -> two-ph, ...
    if ( N_PI(k) > 0 ) then
      write(OUTPUT, "('# |NUM_',1d,'P|',100f10.3)") k, &
        dble(N_PI(k)) / dble(N_event), dble(N_proc(k+3,:)) / dble(N_event)
    end if
  end do
  !
  ! print out info for each scanning point in a single line
  ! and calculate accumulated counters when a range of fluence is given
  !
  N_acc_traj  = N_acc_traj  + N_event
  N_acc_visit = N_acc_visit + P%N_config_visited
  N_acc_store = N_acc_store + P%N_config_stored
  N_acc_calc  = N_acc_calc  + P%N_config_calc
  N_acc_proc  = N_acc_proc  + P%N_process_stored
  if ( P%yes_scan ) then
    call check_time( int_time, timer )
    close( OUTPUT )
    write( OUTPUT_MAIN, "('#',i7,5i8,2x,a)" ) ii, &
      N_event, P%N_config_visited, P%N_config_stored, P%N_config_calc, &
      int( timer - timer_old ), trim(file_scan)
    N_acc_time  = N_acc_time  + int( timer - timer_old )        ! in seconds
    timer_old = timer
    if ( P%yes_resonance ) N_acc_res = N_acc_res + N_res_excitation
    if ( P%yes_shakeoff  ) N_acc_so  = N_acc_so  + N_shakeoff
  end if
  end do SCAN
  !=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  ! END SCAN
  !=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
  !
  ! print info of total counters
  !
  if ( P%yes_scan ) then
    OUTPUT = OUTPUT_MAIN
    write(OUTPUT, "('#       ',a)" ) repeat( "-", 40 )
    write(OUTPUT, "('#|MCFLY|',5i8)") N_acc_traj, N_acc_visit, N_acc_store, &
      N_acc_calc, N_acc_time
    write(OUTPUT, "('#')")
    call print_divider( OUTPUT, "global Monte Carlo statistics" )
    write(OUTPUT, "('# - N of all possible q-hole config.        = ',i15)") &
      N_config
    write(OUTPUT, "('#   total N of config. visited              = '," // &
                  "i15,'  (',f6.2,'%)')") &
      N_acc_visit, dble(N_acc_visit) / dble(N_config) * 100_long
    write(OUTPUT, "('#   total N of config. stored in memory     = '," // &
                  "i15,'  (',f6.2,'%)')") &
      N_acc_store, dble(N_acc_store) / dble(N_config) * 100_long
    write(OUTPUT, "('#   total N of config. calculated           = ',i15)") &
      N_acc_calc
    write(OUTPUT, "('#   total N of processes stored in memory   = ',i15)") &
      N_acc_proc
    write(OUTPUT, "('#   atomic data are stored in memory " // &
                  "after visited ')", advance="no" )
    if ( P%max_counter_visit == 1 ) then
      write(OUTPUT, "('once')")
    else
      write(OUTPUT, "(i2,' times')") P%max_counter_visit
    end if
    if ( P%yes_resonance ) then
      write(OUTPUT, "('# - total occurence of resonant excitations = ',i15)") &
        N_acc_res
    end if
    if ( P%yes_shakeoff ) then
      write(OUTPUT, "('# - total occurence of shakeoff processes   = ',i15)") &
        N_acc_so
    end if
  end if
  !
  ! print out memory usage info
  !
  call initialize_config( P, config )           ! set up S with initial config
  call get_atomic_data( Q0, P, G, F, config, S )
  write(OUTPUT, "('#')")
  call print_divider( OUTPUT, "Memory usage (experimental)" )
  call print_memory_usage( S, N_acc_store, N_acc_proc, OUTPUT )
  call footnote( OUTPUT )
contains

!subroutine load_scd( P, S, x, scd, scd_f )
!  type(Param), intent(in) :: P
!  type(Process_data), intent(inout) :: S
!  type(Grid), intent(in) :: x
!  real(long), intent(out) :: scd(:,0:)          ! WARNING: starting from 0!!!
!  real(long), intent(out) :: scd_f(:,0:)        ! WARNING: starting from 0!!!
!  real(long), allocatable :: Q(:), f(:)         ! dim. determined by SCD file
!  real(long) :: fp(size(scd,1))                 ! fp(i) or fpp(i): i-th config.
!  real(long) :: fpp(size(scd,1))
!  real(long) :: i_final(P%N_orb), pcs(P%N_orb)
!  real(long) :: Q_k, omega, alpha, f0, w, temp1, temp2
!  type(Spline) :: SP
!  complex(long) :: cff          ! coherent form factor: COMPLEX!!!
!  integer :: i, j, k, l, m, N, N_orb, INFO, OUTPUT
!  character(len=1) :: tag
!  logical :: yes_exist
!  alpha = ONE / SPEED_C_AU      ! fine-structure constant, alpha = 1/c
!  N_orb = P%N_orb               ! this should be N_shell.  # of diff. shells
!  OUTPUT = P%OUTPUT
!  !
!  ! read the dispersion correction from the MAD file
!  ! ACHTUNG: even without the dispersion correction,
!  !          sometimes it's convenient to read this MAD file 
!  !          because it contains PCS data for different photon energies.
!  !          so, read the MAD file no matter how yes_dispersion is on or not.
!  !
!  fp(:) = ZERO
!  fpp(:) = ZERO
!  scd_f(:,1) = ZERO     ! f'
!  scd_f(:,2) = ZERO     ! f''
!  inquire( file=P%file_mad, exist=yes_exist )
!  if ( yes_exist ) then
!    open( P%IO_MAD, file=P%file_mad, action="read", form="formatted" )
!    omega = P%PE_start * AU2EV  ! be careful, MAD has P.E. in eV
!    m = 0
!    loop_MAD: do
!      !
!      ! i: configuration index, see config_to_index()
!      ! N: # of data for given index
!      ! tag: must be a character of '#'
!      !      why do I use '#'?  well... you may use gnuplot with the SCD file .
!      !
!      read( P%IO_MAD, *, iostat=INFO ) tag, i, N
!      if ( INFO < 0 ) exit loop_MAD             ! exit when it ends
!      m = m + 1
!      read( P%IO_MAD, * ) tag, i_final(:)       ! forget about (-) i_final's...
!      do j = 1, N
!        read( P%IO_MAD, * ) w, temp1, temp2, pcs(:)
!        if ( abs( omega - w ) < 1.0e-8_long ) then  ! set up data for given P.E.
!          !
!          ! set up the dispersion correction for the i-th configuration
!          !
!          if ( P%yes_dispersion ) then
!            scd_f(i,1) = temp1          ! f'
!            scd_f(i,2) = temp2          ! f''
!          end if
!          !
!          ! change the PCS data for the i-th configuration with given P.E.
!          !
!          do l = 1, N_orb
!            do k = S%ind(i,1), S%ind(i,2)
!              if ( S%col(k) == i_final(l) ) S%value(k) = pcs(l)
!            end do
!          end do
!        end if
!      end do
!      read( P%IO_MAD, * )                       ! skip one blank line
!    end do loop_MAD
!    close( P%IO_MAD )
!    if ( m /= 0 ) then
!      write(OUTPUT, "('# ... all MAD data are loaded from ',a)") &
!        trim(P%file_mad)
!    else if ( .not. P%yes_dispersion ) then
!      write(OUTPUT, "('# ... only PCS data are loaded from ',a)") &
!        trim(P%file_mad)
!    end if
!  else
!    write(OUTPUT, "('# ... no MAD file is available')")
!    write(OUTPUT, "('# ... PCS data are fixed at a certain photon energy.')")
!  end if
!  !
!  ! read the scattering data from the SCD file
!  ! convert them from the coferent form factor into the cross section
!  !
!  !   d sigma
!  !   ------- = alpha^4 |f0(Q)|^2 [ the polarization-dependent factor ]
!  !   d Omega
!  !             where [ ] = ( 1 + cos(theta)^2 ) / 2 for unpolarized x-rays
!  !
!  open( P%IO_SCD, file=P%file_scd, action="read", form="formatted" )
!  m = 0
!  scd(1,:) = ZERO     ! no electron in this configuration (bare atom)
!  loop_SCD: do
!    !
!    ! i: configuration index, see config_to_index()
!    ! N: # of data for given index
!    ! tag: must be a character of '#'
!    !
!    read( P%IO_SCD, *, iostat=INFO ) tag, i, N
!    if ( INFO < 0 ) exit loop_SCD               ! exit when it ends
!    m = m + 1
!    if ( m == 1 ) allocate( Q(N), f(N) )        ! for the first data set
!    do k = 1, N
!      read( P%IO_SCD, * ) Q(k), f(k)            ! read the i-th config. data
!    end do
!    read( P%IO_SCD, * )                         ! skip one blank line
!    if ( m == 1 ) then
!      SP = Spline_create( Q, f )                ! for the first data set
!    else
!      call Spline_reuse( SP, f )                ! otherwise, just reuse
!    end if
!    !
!    ! set up the i-th scd data
!    !
!    do k = 0, P%N_scd + 1
!      Q_k = TWO * alpha * P%PE_start * sin( acos( x%r(k) ) / TWO )
!      f0 = Spline_interpolate( SP, Q_k )
!      cff = f0                                  ! non-resonant form factor
!      if ( Q_k == ZERO ) scd_f(i,0) = f0        ! f0(Q=0)
!      if ( P%yes_dispersion ) then              ! dispersion correction
!        cff = cff + scd_f(i,1) + IMAG_ONE * scd_f(i,2)
!      end if
!      scd(i,k) = alpha**4 * cdabs( cff )**2 * ( ONE + x%r(k)**2 ) / TWO
!    end do
!  end do loop_SCD
!  close( P%IO_SCD )
!  write(OUTPUT, "('# ... all SCD data are loaded from ',a)") trim(P%file_scd)
!  write(OUTPUT, "('#')")
!  deallocate( Q, f )
!end subroutine load_scd

function pulse_envelope( P, SP, t ) result (f)
  real(long) :: f
  type(Param), intent(in) :: P
  type(Spline), intent(in) :: SP
  real(long), intent(in) :: t
  real(long) :: sigma, f1, f2, sigma2, r1, r2
  !
  ! normalized Gaussian distribution
  !   f = e^( -t^2 / 2 sigma^2 ) / ( sigma sqrt(2pi) )
  ! where sigma: standard deviation
  !       FWHM (Full Width at Half Maximum), tau = 2 sigma sqrt( 2 ln 2 )
  !
  if ( t <= P%t_start_au .or. t >= P%t_end_au ) then
    f = ZERO
  else if ( P%pulse_type == 3 ) then    ! Gaussian
    sigma = P%FWHM / TWO / sqrt( TWO * log(TWO) )
    f = exp( -t**2 / TWO / sigma**2 ) / sqrt( TWO * PI ) / sigma
  else if ( P%pulse_type == 6 ) then    ! flat-top, FWHM here is pulse duration
    f = ONE / P%FWHM
  else if ( P%pulse_type == 8 ) then    ! double Gaussian (split)
    sigma = P%FWHM / TWO / sqrt( TWO * log(TWO) )
    sigma2 = sigma * P%FWHM_ratio
    f1 = exp( -t**2 / TWO / sigma**2 ) / sqrt( TWO * PI ) / sigma
    f2 = exp( -( t - P%time_delay )**2 / TWO / sigma2**2 ) &
         / sqrt( TWO * PI ) / sigma2
    r1 = ONE / ( ONE + P%fluence_ratio )
    r2 = P%fluence_ratio / ( ONE + P%fluence_ratio )
    f = f1 * r1 + f2 * r2
  else if ( P%pulse_type == 9 ) then      ! double flattops (split)
    sigma  = P%FWHM * HALF                ! 1/2 pulse length for the wider
    sigma2 = P%FWHM * P%FWHM_ratio * HALF ! 1/2 pulse length for the narrower
    f = ZERO
    if ( t > -sigma .and. t <= sigma ) then
      f1 = ONE / P%FWHM                   ! pulse length for the wider
      r1 = ONE / ( ONE + P%fluence_ratio )
      f = f1 * r1
    end if
    if ( t > P%time_delay - sigma2 .and. t <= P%time_delay + sigma2 ) then
      f2 = ONE / ( P%FWHM * P%FWHM_ratio )  ! remember P%FWHM_ratio always < 1
      r2 = P%fluence_ratio / ( ONE + P%fluence_ratio )
      f = f + f2 * r2
    end if
  else if ( P%pulse_type == 10 ) then
    f = Spline_interpolate( SP, t )
  end if
end function pulse_envelope

!----------------------------------------------------------------------------
! SR get_atomic_data(): to obtain all necessary atomic data for a given config.
! - config. is given as "orbital" --> Now it's "config" as of Apr. 6, 2019
! - all atomic data are returned via S
!
! internal variables:
! - Q0: staring point of the Config_tree, which is defined in process.f90
! - Q: ending point of the Config_tree, pointer
! - S0: temporary space for Process_data, memory allocated with P%N_proc_max
!
! There are three different cases for calculating atomic data
! (A) S => Q%S : use atomic data that are already calculated & stored in Q%S
! (B) S => S0  : calculate atomic data, but do not store them
! (C) copy S0 to Q%S and S => Q%S : calculate atomic data & store them in Q%S
!----------------------------------------------------------------------------
subroutine get_atomic_data( Q0, P, G, F, config, S )
  type(Config_tree), target, intent(in) :: Q0
  type(Param), intent(inout) :: P       ! (inout) because of P%N_config_*
  type(Coordinate), intent(inout) :: G  ! (inout) because of SCF_procedure()
  type(WF_data), intent(inout) :: F     ! (inout) because of SCF_procedure()
  integer, intent(in) :: config(:,:)
  type(Process_data), pointer, intent(out) :: S
  integer :: occ_index
  integer :: core_hole_index
  integer :: N_process
  type(Config_tree), pointer :: Q               ! ending point of the tree
  type(Process_data), save, target :: S0        ! Process_data w/ P%N_proc_max
!  type(Shakeoff_data), intent(inout) :: SO
  logical, save :: yes_first_call = .true.
  integer :: i, N0
  real(long) :: omega, bandwidth, charge
!TEST
!
! hard-wired direct two-photon cross sections for Xe@1200 eV --> to be removed
!
!  integer :: n_occ
!  real(long), parameter :: TAU = 3.45e-18               ! period (sec) @1200 eV
!  real(long), parameter :: PCS_3s12 = 8.99404E-02 / TWO    * 1e-22  ! pcs of
!  real(long), parameter :: PCS_3p12 = 1.22968E-01 / TWO    * 1e-22  ! neutral Xe
!  real(long), parameter :: PCS_3p32 = 2.19311E-01 / FOUR   * 1e-22  ! unit: m^2
!  real(long), parameter :: PCS_3d32 = 3.79970E-01 / FOUR   * 1e-22
!  real(long), parameter :: PCS_3d52 = 5.47418E-01 / 6_long * 1e-22
!  real(long), parameter :: PCS_4s12 = 2.20862E-02 / TWO    * 1e-22
!  real(long), parameter :: PCS_4p12 = 2.48612E-02 / TWO    * 1e-22
!  real(long), parameter :: PCS_4p32 = 4.84224E-02 / FOUR   * 1e-22
!  real(long), parameter :: PCS_4d32 = 4.39624E-02 / FOUR   * 1e-22
!  real(long), parameter :: PCS_4d52 = 6.55488E-02 / 6_long * 1e-22
!  real(long), parameter :: TWOPCS(5:14) = (/ &
!    ( PCS_3s12*M2AU**2 )**2 * TAU*SEC2AU, &   ! 3s0
!    ( PCS_3p12*M2AU**2 )**2 * TAU*SEC2AU, &   ! 3p-
!    ( PCS_3p32*M2AU**2 )**2 * TAU*SEC2AU, &   ! 3p+
!    ( PCS_3d32*M2AU**2 )**2 * TAU*SEC2AU, &   ! 3d-
!    ( PCS_3d52*M2AU**2 )**2 * TAU*SEC2AU, &   ! 3d+
!    ( PCS_4s12*M2AU**2 )**2 * TAU*SEC2AU, &   ! 4s0
!    ( PCS_4p12*M2AU**2 )**2 * TAU*SEC2AU, &   ! 4p-
!    ( PCS_4p32*M2AU**2 )**2 * TAU*SEC2AU, &   ! 4p+
!    ( PCS_4d32*M2AU**2 )**2 * TAU*SEC2AU, &   ! 4d-
!    ( PCS_4d52*M2AU**2 )**2 * TAU*SEC2AU /)   ! 4d+
!  real(long), parameter :: THREEPCS(5:9) = (/ &
!    ( PCS_3s12*M2AU**2 )**3 * (TAU*SEC2AU)**2, &   ! 3s0
!    ( PCS_3p12*M2AU**2 )**3 * (TAU*SEC2AU)**2, &   ! 3p-
!    ( PCS_3p32*M2AU**2 )**3 * (TAU*SEC2AU)**2, &   ! 3p+
!    ( PCS_3d32*M2AU**2 )**3 * (TAU*SEC2AU)**2, &   ! 3d-
!    ( PCS_3d52*M2AU**2 )**3 * (TAU*SEC2AU)**2 /)   ! 3d+
!TEST
  !
  ! check whether this subroutine is called for the first time
  ! if yes, then prepare S0 with the P%N_proc_max
  ! 
  if ( yes_first_call ) then
    S0 = create_Process( config, P%N_proc_max )
    P%N_config_visited = 0
    P%N_config_stored  = 0
    P%N_config_calc    = 0
    P%N_process_stored = 0
    yes_first_call = .false.
  endif
  !
  ! find out the end node of the Q0 tree
  ! the Q0 tree is automatically generated within SR get_end_node()
  !
  Q => get_end_node( Q0, config, F%N_subshell )
  !
  ! check out Q%S
  !
  if ( allocated(Q%S) ) then                    ! if yes, it is already visited
    if ( allocated(Q%S%config) ) then           ! Q%S%config already exists
      !
      ! CASE (A): all calculated rates are already stored in Q%S
      !           then use this Q%S for the atomic data
      !
      !Q%S%counter_visit = Q%S%counter_visit + 1        ! no need to increase...
      S => Q%S

!      ! shakeoff 
!      if ( P%yes_shakeoff ) then
!         ! wave function for config(:,:)
!         F%config(1:F%N_orb,:) = get_non_relativistic_config( P, config(:,:) )
!         call SCF_procedure( P, G, F )
!         if ( P%yes_relativity ) then
!            F%config_rel(1:F%N_subshell,1:4) = config(1:F%N_subshell,1:4)
!            F%f_occ_rel(:) = F%config_rel(:,4)
!            call relativity_calc_first_order_corr( P, G, F )
!         endif
!         call get_shakeoff_br( P, G, F, S, SO )
!      endif

      return    ! exit this routine for the case (A)!
    end if
  else                          ! this is a newly visited config.
    allocate( Q%S )             ! prepare Q%S, but no memory alloc. inside Q%S
    Q%S%counter_visit = 0       ! initialize the counter
    P%N_config_visited = P%N_config_visited + 1
  end if
  Q%S%counter_visit = Q%S%counter_visit + 1     ! increase the counter
  !
  ! check if no electron remains; if yes, don't do any calculation
  !
  occ_index = size(config,dim=2)
  if ( sum(config(:,occ_index)) == 0 ) then    ! no calc. but do counting
    if ( Q%S%counter_visit >= P%max_counter_visit ) then
      P%N_config_stored = P%N_config_stored + 1
      N_process = 0
      Q%S = create_Process( config, N_process )
    end if
    S => Q%S    ! S should be assigned right before before "return"
    return
  end if
  !
  ! calculate SCF: SCF is done with non-relativistic electronic config
  !
  P%N_config_calc = P%N_config_calc + 1
  F%config(1:F%N_orb,:) = get_non_relativistic_config( P, config(:,:) )
  call SCF_procedure( P, G, F )
  if ( P%yes_relativity ) then
    !
    ! F%config_rel(:,4) vs. F%f_occ_rel(:)
    ! The former is used for cross section and rate calculations,
    ! while the latter is currently used only for printing.
    !
    F%config_rel(1:F%N_subshell,1:4) = config(1:F%N_subshell,1:4)
    F%f_occ_rel(:) = F%config_rel(:,4)
    call relativity_calc_first_order_corr( P, G, F )
  endif
  call extend_WF_continuum( F, P%OUTPUT )
  !
  ! reset S0
  !
  S0%config(:,:) = config(:,:)
  S0%E_tot = F%E_tot    ! set up total energy for given configuration
  S0%E0    = F%E0       ! set up zeroth energy
  S0%N_process = 0      ! must be initialized; to be accumulated for rate calc.
  S0%occ_up(:,:) = 0
  S0%occ_down(:,:) = 0
  !
  ! photoionization, Auger, and fluorescence with F
  ! all calculated data will be stored in S0
  !
  N0 = S0%N_process     ! to be changed after calculate_Auger_rates()
  call calculate_Auger_rates( P, F, S0 )
  do i = N0 + 1, S0%N_process   ! for newly-added Auger processes
    if ( S0%occ_down(1,i) > P%init_occ_max_i .or. &
         S0%occ_down(2,i) > P%init_occ_max_i ) then
      S0%tag(i) = "a"           ! it's assigned as autoionization
    end if
    if ( P%yes_massage ) then
      core_hole_index = S0%occ_up(1,i)          ! orbital index for core hole
      select case ( S0%config(core_hole_index,2) )  ! config(:,2): quantum # l
        case ( 0 ); S0%value(i) = S0%value(i) * P%massage_decay_s
        case ( 1 ); S0%value(i) = S0%value(i) * P%massage_decay_p
        case ( 2 ); S0%value(i) = S0%value(i) * P%massage_decay_d
        case ( 3 ); S0%value(i) = S0%value(i) * P%massage_decay_f
      end select
      charge = P%Z - sum( config(:,occ_index) )
      if ( charge < P%massage_auger_q1 ) &
        S0%value(i) = S0%value(i) * P%massage_auger_factor1
      if ( charge >= P%massage_auger_q2 ) &
        S0%value(i) = S0%value(i) * P%massage_auger_factor2
    end if
  end do
  N0 = S0%N_process     ! to be changed after calculate_fluorescence_rates()
  call calculate_fluorescence_rates( P, G, F, S0 )
  do i = N0 + 1, S0%N_process   ! for newly-added fluorescence processes
    if ( P%yes_massage ) then
      core_hole_index = S0%occ_up(1,i)          ! orbital index for core hole
      select case ( S0%config(core_hole_index,2) )  ! config(:,2): quantum # l
        case ( 0 ); S0%value(i) = S0%value(i) * P%massage_decay_s
        case ( 1 ); S0%value(i) = S0%value(i) * P%massage_decay_p
        case ( 2 ); S0%value(i) = S0%value(i) * P%massage_decay_d
        case ( 3 ); S0%value(i) = S0%value(i) * P%massage_decay_f
      end select
    end if
  end do
  if ( P%fluence /= ZERO ) then
    if ( .not. P%yes_resonance ) then
      call calculate_photoionization_CS( P, F, S0 )
    else
      call calculate_photoabsorption_CS( P, G, F, S0 )
    endif
  endif
  !
  ! consider harmonics of P%PE_start
  !
  if ( P%yes_harmonics ) then
    omega = P%PE_start                  ! the main frequency is stored
    bandwidth = P%bandwidth             ! the main bandwidth is stored
    if ( P%harmonic2 > ZERO ) then      ! second harmonic
      P%PE_start = omega * TWO
      P%PE_end   = omega * TWO
      P%bandwidth = bandwidth * TWO
      N0 = S0%N_process     ! to be changed after calculate_PCS()
      if ( .not. P%yes_resonance ) then
        call calculate_photoionization_CS( P, F, S0 )
      else
        call calculate_photoabsorption_CS( P, G, F, S0 )
      endif
      do i = N0 + 1, S0%N_process
        S0%value(i) = S0%value(i) * P%harmonic2 ! reduce cross section
      end do
    end if
    if ( P%harmonic3 > ZERO ) then      ! third harmonic
      P%PE_start = omega * THREE
      P%PE_end   = omega * THREE
      P%bandwidth = bandwidth * THREE
      N0 = S0%N_process
      if ( .not. P%yes_resonance ) then
        call calculate_photoionization_CS( P, F, S0 )
      else
        call calculate_photoabsorption_CS( P, G, F, S0 )
      endif
      do i = N0 + 1, S0%N_process
        S0%value(i) = S0%value(i) * P%harmonic3 ! reduce cross section
      end do
    end if
    P%PE_start = omega          ! restore the original value
    P%PE_end   = omega          ! restore the original value
    P%bandwidth = bandwidth     ! restore the original value 
  end if
  !
  ! consider an additional photon energy
  !
  if ( P%PE2 /= ZERO ) then
    omega = P%PE_start                  ! the main frequency is stored
    bandwidth = P%bandwidth             ! the main bandwidth is stored
    P%PE_start = P%PE2
    P%PE_end   = P%PE2
    P%bandwidth = bandwidth / omega * P%PE2
    N0 = S0%N_process     ! to be changed after calculate_PCS()
    if ( .not. P%yes_resonance ) then
      call calculate_photoionization_CS( P, F, S0 )
    else
      call calculate_photoabsorption_CS( P, G, F, S0 )
    endif
    do i = N0 + 1, S0%N_process
      S0%value(i) = S0%value(i) * P%PE2_contrib ! reduce cross section
    end do
    P%PE_start = omega          ! restore the original value
    P%PE_end   = omega          ! restore the original value
    P%bandwidth = bandwidth     ! restore the original value 
  end if
!TEST
!  ADD extra cross section taken from Makris et al., PRL 102, 033002 (2009)
!  - two-photon cross section:   7x10^-49 cm^4 sec = 37 a.u.
!  - three-photon cross section: 2x10^-79 cm^6 sec^2 = 1.6e4 a.u.
!  - four-photon cross section:  10^-112 cm^8 sec^3 = 1.1e4 a.u.
!  13 => Xe 4d3/2
!  14 => Xe 4d5/2
!----
!  charge = P%Z - sum( config(:,occ_index) )
!  omega = P%PE_start
!  if ( charge == 7 .or. charge == 8 ) then      ! two-photon cross section
!    if ( S0%config(13,occ_index) > 0 ) then
!      E = omega * TWO - F%e_orb_rel(13)
!      call add_Process( S0, "two-photon", 36.904_long*0.4_long, E, 13 )
!    end if
!    if ( S0%config(14,occ_index) > 0 ) then
!      E = omega * TWO - F%e_orb_rel(14)
!      call add_Process( S0, "two-photon", 36.904_long*0.6_long, E, 14 )
!    end if
!  end if
!  if ( charge == 9 .or. charge == 10 .or. charge == 11 ) then   ! three-photon
!    if ( S0%config(13,occ_index) > 0 ) then
!      E = omega * THREE - F%e_orb_rel(13)
!      call add_Process( S0, "three-photon", 15567_long*0.4_long, E, 13 )
!    end if
!    if ( S0%config(14,occ_index) > 0 ) then
!      E = omega * THREE - F%e_orb_rel(14)
!      call add_Process( S0, "three-photon", 15567_long*0.6_long, E, 14 )
!    end if
!  end if
!  if ( charge == 12 ) then                      ! four-photon
!    if ( S0%config(13,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(13)
!      call add_Process( S0, "four-photon", 344.72_long*0.4_long, E, 13 )
!    end if
!    if ( S0%config(14,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(14)
!      call add_Process( S0, "four-photon", 344.72_long*0.6_long, E, 14 )
!    end if
!  end if
!  if ( charge == 13 .or. charge == 14 ) then    ! four-photon
!    if ( S0%config(13,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(13)
!      call add_Process( S0, "four-photon", 11491_long*0.4_long, E, 13 )
!    end if
!    if ( S0%config(14,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(14)
!      call add_Process( S0, "four-photon", 11491_long*0.6_long, E, 14 )
!    end if
!  end if
!  if ( charge == 15 ) then                      ! four-photon
!    if ( S0%config(13,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(13)
!      call add_Process( S0, "four-photon", 45963_long*0.4_long, E, 13 )
!    end if
!    if ( S0%config(14,occ_index) > 0 ) then
!      E = omega * FOUR - F%e_orb_rel(14)
!      call add_Process( S0, "four-photon", 45963_long*0.6_long, E, 14 )
!    end if
!  end if
!----
!  write(STDERR, "('CHARGE = ',i5)") P%Z - sum( config(:,occ_index) )
!  call print_Process( S0, STDERR )
!  write(STDERR, "()")
!TEST
  ! For Xe@1200 eV
  ! add extra process for direct 2- or 3-photon ionization when n=3 or n=4
!  omega = P%PE_start
!  do i = 1, F%N_subshell
!    n_occ = S0%config(i,occ_index)
!    if ( i >= 5 .and. i <= 14 .and. n_occ > 0 ) then
!      !
!      ! add direct two-photon ionization
!      !
!      E = omega * TWO - ionization_E( F, i )
!      if ( E > ZERO ) then
!        call add_Process( S0, "two-photon", &
!          P%boost * TWOPCS(i) * n_occ, E, i )
!      end if
!      if ( i <= 9 ) then
!        E = omega * THREE - ionization_E( F, i )
!        if ( E > ZERO ) then
!          call add_Process( S0, "three-photon", &
!            P%boost * THREEPCS(i) * n_occ, E, i )
!        end if
!      end if
!    end if
!  end do
!TEST
  !
  ! shakeoff process: store all orbitals for calculating branching ratios
  !
  if ( P%yes_shakeoff ) then
    if ( .not. allocated( S0%u ) ) &  ! mem alloc here, not in create_Process
      allocate( S0%u(F%N,F%N_orb) )
    S0%u(:,:) = F%u(:,:)
  end if
 
!  ! shakeoff 
!  if ( P%yes_shakeoff ) then
!     call get_shakeoff_br( P, G, F, S0, SO )
!  endif

  !
  ! check out whether counter_visit < max.
  !
  if ( Q%S%counter_visit < P%max_counter_visit ) then
    !
    ! CASE (B): use S0 for the atomic data
    !
    S => S0
    return    ! exit this routine for the case (B)!
  else
    !
    ! CASE (C): store S0 into Q%S
    ! memory allocation for Q%S, only with N_process, rather than P%N_proc_max
    ! then make a copy from S0 to Q%S, only for 1:N_process
    !
    if ( allocated(Q%S%config) ) then   ! something wrong! stop here!!
      call error_message( "in get_atomic_data()" )
    end if
    P%N_config_stored = P%N_config_stored + 1   ! this is newly stored config.
    P%N_process_stored = P%N_process_stored + S0%N_process
    N_process = S0%N_process    ! S0%N_process determined from rate calc.
    Q%S = create_Process( config, N_process )
    if ( N_process > 0 ) then
      Q%S%tag(1:N_process)        = S0%tag(1:N_process)
      Q%S%p_type(1:N_process)     = S0%p_type(1:N_process)
      Q%S%value(1:N_process)      = S0%value(1:N_process) 
      Q%S%E_eV(1:N_process)       = S0%E_eV(1:N_process) 
      Q%S%occ_up(:,1:N_process)   = S0%occ_up(:,1:N_process) 
      Q%S%occ_down(:,1:N_process) = S0%occ_down(:,1:N_process)
      Q%S%N_process = N_process ! BE CAREFUL: S%N_process=0 after create_Process
    endif
    Q%S%E_tot = S0%E_tot        ! This should be assigned even if N_process = 1
    Q%S%E0    = S0%E0
    if ( P%yes_shakeoff ) then        ! store orbitals
      allocate( Q%S%u(F%N,F%N_orb) )  ! BE CAREFUL: don't forget mem-alloc
      Q%S%u(:,:) = S0%u(:,:)
    end if
    S => Q%S                    ! S should be assigned right before "return"
    return    ! exit this routine for the case (C)!
  end if
  !
  ! generate dat file
  !
!  call print_Process( S, P%IO_DATA )
end subroutine get_atomic_data

!subroutine get_shakeoff_br( P, G, F, S, SO )
!  type(Param), intent(inout) :: P
!  type(Coordinate), intent(inout) :: G
!  type(WF_data), intent(in) :: F
!  type(Process_data), intent(inout) :: S
!  type(Shakeoff_data), intent(inout) :: SO
!  integer :: id
!
!    ! wave function before a primal decay of id-th channel 
!    call RegisterWF( F, SO, 'before' )
!    do id = 1, S%N_process
!       ! wave function after the primald decay
!       call ConfigProcessAndRegister( P, G, S, id, SO, 'middle' )
!
!       ! correct rates and pcs, calculate branching ratios
!       call CalculateShakeoffBR( P, G, S, id, SO )
!       !SO%OUTPUT = OUTPUT
!       !call ShowResult( S0, SO, id )
!    enddo
!end subroutine get_shakeoff_br

!----------------------------------------------------------------------------
! consider shakeoff process
! - calculate shakeoff branching ratios for the given transition
!   and proceed shakeoff via a Monte Carlo scheme
! - it turns out that SCF calc. for initial & final ALL THE TIME are too slow
!   therefore, S%u(:,:) store orbitals after SCF of each config.
!   and they are used for calculating shakeoff braching ratios
!
! To specify the transition,
!   the initial config. is given in S%config and
!   the final config. is given in config(:,:)
! After shakeoff branching ratios are calculated,
!   one random number is used to determine whether shakeoff happens or not.
! If shakeoff happens, the new config. will be stored in config(:,:)
!----------------------------------------------------------------------------
subroutine consider_shakeoff( Q0, P, G, F, S, config, N_shakeoff )
  type(Config_tree), target, intent(in) :: Q0
  type(Param), intent(in) :: P
  type(Coordinate), intent(inout) :: G  ! (inout) because of SCF_procedure()
  type(WF_data), intent(inout) :: F     ! (inout) because of SCF_procedure()
  type(Process_data), intent(in) :: S   ! contains initial config & orbitals
  integer, intent(inout) :: config(:,:) ! final config. but will be new config.
  integer, intent(inout) :: N_shakeoff  ! counter for shakeoff processes
  type(Config_tree), pointer :: Q       ! ending point of the tree
  real(long) :: prob(size(config,dim=1))          ! shakeoff branching ratios
  real(long), dimension(F%N,F%N_orb) :: u_i, u_f  ! initial & final orbitals
  integer :: i, k, occ_n, occ_index
  real(long) :: overlap, rand
  character(len=100) :: string1, string2
  logical :: u_f_loaded
  !
  ! obtain WF for the initial configuration (non-relativistic calculation)
  !
  if ( size(u_i,dim=1) /= size(S%u,dim=1) ) &
    call error_message( "SR consider_shakeoff: u_i and S%u are not matched." )
  u_i(:,:) = S%u(:,:)
  !
  ! obtain WF for the final configuration (non-relativistic calculation)
  ! search it in Q; otherwise calculate it with F
  !
  F%config(:,:) = get_non_relativistic_config( P, config )  ! final config.
  Q => get_end_node( Q0, config, F%N_subshell )
  u_f_loaded = .false.
  if ( allocated(Q%S) ) then            ! it's been visited
    if ( allocated(Q%S%u) ) then        ! AND u(:,:) is stored
      u_f(:,:) = Q%S%u(:,:)
      u_f_loaded = .true.
    end if
  end if
  if ( .not. u_f_loaded ) then
    call SCF_procedure( P, G, F )  ! this final config is new, so solve SCF
    u_f(:,:) = F%u(:,:)
  end if
  !
  ! calculate branching ratios... considering only single shakeoff
  !
  prob(:) = ZERO
  do k = 1, size(prob)
    if ( F%yes_relativity ) then
      occ_n = config(k,4)       ! occupation number
      i = F%config_rel(k,5)     ! index for non-relativistic orbital
    else
      occ_n = config(k,3)       ! occupation number
      i = k
    end if
    if ( occ_n == 0 ) cycle     ! no electron in this subshell
    overlap = radial_integral( G, u_i(:,i) * u_f(:,i) )
    prob(k) = ( ONE - overlap**2 ) * occ_n
  end do
  if ( sum( prob(:) ) > ONE ) then      ! ERROR if sum of branching ratios > 1
    string1 = config_to_string( S%config )      ! either non-rel or rel
    string2 = config_to_string( F%config )      ! always non-rel
    if ( F%yes_relativity ) string2 = config_to_string( F%config_rel )
    write(STDERR, "('Something wrong in the shakeoff routine:')")
    write(STDERR, "('- initial config: ',a)") trim(string1)
    write(STDERR, "('- final config  : ',a)") trim(string2)
    write(STDERR, "(a5,2a17,3x,a13)") "orb_i", "initial", "final", "prob"
    write(STDERR, "(5x,2a17)") repeat("-",17), repeat("-",17)
    do k = 1, size(prob)
      write(STDERR, "(i5,2(2x,3i5),3x,es13.5)") k, &
        S%config(k,1), S%config(k,2), S%config(k,3), &
        F%config(k,1), F%config(k,2), F%config(k,3), &
        prob(k)
    end do
    stop
  end if
  !
  ! let's do a Monte Carlo way
  ! config(:,:) contains the final config. of the given transition
  !             will be the new config. after shakeoff (if any)
  ! if no shakeoff happens, then there will be no change in config(:,:)
  !
  occ_index = size(config,dim=2)
  rand = random()                       ! choose a random number for shakeoff
  loop_shakeoff: do k = 1, size(prob)
    if ( prob(k) == ZERO ) cycle        ! skip if probability is zero
    if ( rand < prob(k) ) then
      !
      ! hurray! shakeoff happens at the k-th subshell
      ! take out one electron from the k-th subshell and exit the loop
      !
      config(k,occ_index) = config(k,occ_index) - 1
      N_shakeoff = N_shakeoff + 1       ! increase the shakeoff counter
      exit loop_shakeoff
    end if
    rand = rand - prob(k)
  end do loop_shakeoff
end subroutine consider_shakeoff

subroutine initialize_config( P, config )
  integer, intent(inout) :: config(:,:)
  type(Param), intent(in) :: P
  integer :: num
  if ( P%yes_relativity ) then
     num = P%N_orb_rel
     config(1:num,:) = P%init_config_rel(:,1:4)
  else
     num = P%N_orb
     config(1:num,:) = P%init_config(:,:)
  endif
  config(num+1:,:) = 0
end subroutine initialize_config

function compute_rate( P, SP, value, tag, t ) result (rate)
  real(long) :: rate
  type(Param), intent(in) :: P
  type(Spline), intent(in) :: SP        ! valid only if P%yes_load_pulse
  real(long), intent(in) :: value, t
  character(len=1), intent(in) :: tag
  integer :: n
  real(long) :: flux
  n = number_of_photons( tag )
  if ( n == 0 ) then
    rate = value
  else
    flux = P%fluence * pulse_envelope( P, SP, t )
    if ( n == 1 ) then
      rate = value * flux
    else
      rate = value * flux**n
    end if
  end if
  !
  ! manipulate autoionization / Auger / resonant excitation / ionizaation
  !
  if ( P%yes_massage ) then
    if ( tag == "a" ) rate = rate * P%massage_autoionization_rate
    if ( tag == "A" ) rate = rate * P%massage_auger_rate
    if ( tag == "E" ) rate = rate * P%massage_resonant_pcs
    if ( tag == "P" .or. tag == "V" .or. n >= 2 ) rate = rate * P%massage_pcs
  end if
end function compute_rate

function number_of_photons( tag ) result (n)
  integer :: n
  character(len=1), intent(in) :: tag
  if ( tag == "P" .or. tag == "V"  .or. tag == "E" ) then
    n = 1
  else if ( tag == "2" ) then
    n = 2
  else if ( tag == "3" ) then
    n = 3
  else if ( tag == "4" ) then
    n = 4
  else if ( tag == "5" ) then
    n = 5
  else          ! tag == "A" .or. tag == "a" .or. tag == "F"
    n = 0
  end if
end function number_of_photons

subroutine print_memory_usage( S, N_config, N_process, OUTPUT )
  type(Process_data), intent(in) :: S
  integer, intent(in) :: N_config, N_process, OUTPUT
  integer*8 :: M, M2, M3
  real(long) :: a
  integer :: b
  character(len=1) :: c
  logical(kind=1) :: d
  !
  ! based on S structure, estimate a memory size with N_config and N_process
  !
  M = N_process * ( sizeof(c) + sizeof(a) + sizeof(b) + 4*sizeof(b) )
  !                 ---------   ---------   ---------   -----------
  !                 tag         value       E_eV        occ_up/occ_down
  !
  M = M + N_config * ( sizeof(b) + sizeof(a) +  sizeof(b) +     sizeof(d) * 2 )
  !                    ---------   ---------    ---------       -------------
  !                    N_process   E_tot      counter_visit  yes_rel/pot_stored
  !
  M2 = N_config * size(S%config) * sizeof(b)
  M3 = 0
  if ( allocated( S%u ) ) M3 = N_config * size( S%u ) * sizeof(a)
  write(OUTPUT, "('# - memory usage (kilobytes):')")
  write(OUTPUT, "('#   used for all stored atomic data   = ',f15.1)") &
    dble( M + M2 + M3 ) / 1024_long
  write(OUTPUT, "('#   used for S%config(:,:) only       = ',f15.1)") &
    dble( M2 ) / 1024_long
  if ( allocated( S%u ) ) then
    write(OUTPUT, "('#   used for S%u(:) only              = ',f15.1)") &
      dble( M3 ) / 1024_long
  end if
end subroutine print_memory_usage

subroutine record_spectra( P, SP, S, k, t, &
                           max_E_spectra, count_spectra, pop_e_avg )
  type(Param), intent(in) :: P
  type(Spline), intent(in) :: SP        ! valid only if P%yes_load_pulse
  type(Process_data), pointer, intent(in) :: S
  integer, intent(in) :: k, max_E_spectra
  real(long), intent(in) :: t
  integer, intent(inout) :: count_spectra(0:,:)
  real(long), intent(inout) :: pop_e_avg(0:)
  integer :: p_type, k_E
  real(long) :: p_env
  p_type = S%p_type(k)
  if ( p_type >= 5 ) p_type = 1 ! multiphoton ionization is treated as P
!  select case ( S%tag(k) )                      ! checking process type
!  case ( "P", "V" );            p_type = 1      ! photoionization
!  case ( "2", "3", "4", "5" );  p_type = 1      ! multiphoton ionization
!  case ( "A" );                 p_type = 2      ! Auger
!  case ( "F" );                 p_type = 3      ! fluorescence
!  case ( "E" );                 p_type = 4      ! resonant excitation
!  case default
!     call error_message( S%tag(k) // " is not supported." )
!  end select
  k_E = S%E_eV(k)
  !
  ! checking transition energy (k_E) carefully!!!
  ! - k_E can be < 0 or > max_E_spectra because of convolution
  ! - negative case with 'P': negative K.E. of photoelectron
  !   cross section is not zero because it's convoluted
  !   therefore, let's put K.E.=0 in this case
  ! - larger than the upper_bound with 'E':
  !   for 'E', k_E is given by the transition energy
  !   because of convolution, it can be larger than photon energy
  !   in this case, don't count any spectra because it's excitation
  !
  if ( k_E < 0  .and. p_type == 1 ) k_E = 0    ! put K.E.=0
  if ( k_E >= 0 .and. k_E <= max_E_spectra ) then
    if ( p_type <= 3 ) then   ! only for P/V, A, and F
      count_spectra( k_E, p_type ) = count_spectra( k_E, p_type ) + 1
    end if
    if ( p_type <= 2 ) then   ! only for electron spectra (P/V and A)
      !
      ! once electron or photon is recorded in spectra(:,:)
      ! it won't be changed during the time propagation.
      ! therefore, pop_e_avg(:) can be calculated at this moment.
      !
      tp = t
      spec_time_evol_end: do       ! time propagation until the end
        p_env = pulse_envelope( P, SP, tp + P%dt_au * HALF )
        pop_e_avg( k_E ) = pop_e_avg( k_E ) + p_env * P%dt_au
        tp = tp + P%dt_au
        if ( tp > P%t_end_au ) exit spec_time_evol_end
      end do spec_time_evol_end
    end if
  end if
end subroutine record_spectra

end program xatom_mc_fly
