!----------------------------------------------------------------------------
! xatom_req.f90
!----------------------------------------------------------------------------
! Copyright (C) 2010-2017 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2010-2017 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Rate equation solver for x-ray damaged atoms
!   written by Son, Sang-Kil in Jul. 2010 at DESY
!----------------------------------------------------------------------------
program xatom_req
  use constants
  use stdio
  use elec_config
  use read_param
  use gps
  use interpolate
  use print_property
  use git
  implicit none
  type(Param) :: P
  type(Grid) :: x                       ! angular grids, x := cos(theta)
  !
  ! C(:) : configurational populations
  ! pop(:) : charge-state populations
  !
  real(long), allocatable :: C(:)       ! probabilities of all configurations
  real(long), allocatable :: Cdot(:)    ! rates of C(:)
  real(long), allocatable :: C_bak(:)   ! back up of C(:)
  real(long), allocatable :: pop(:)     ! probabilities of all ion species
  real(long), allocatable :: C_end(:), pop_end(:)   ! at the end of pulse
  real(long), allocatable :: C_avg(:), pop_avg(:)   ! time-averaged
  integer, allocatable :: charge(:)     ! charges of all configurations
  integer, allocatable :: ground_config(:)  ! ground config. for given charges
  !
  ! Process data contains cross sections and rates of all processes.
  ! this is optimized for REQ calculations just with a sparse matrix scheme.
  !
  type :: Process_data
    integer :: N                        ! max. index = # of config indexes
    integer :: N_process                ! # of diff. rates = # of processes
    integer, allocatable :: initial(:), final(:)
    real(long), allocatable :: value(:) ! cross section or rate in a.u.
    integer, allocatable :: cs_order(:) ! 0:rate, 1:one-photon cs, 2:two-ph. cs.
    !
    ! spectra of electron and photon
    !   p_type(i) : type of the i-th process
    !   energy_eV(i) : energy of elec. or photon of the i-th process
    !               values are rounded to nearst integer
    !   spectra_max: max. energy for spectra
    !   spectra(k,j) : counts for the k eV bin, here resolution becomes eV.
    !                  k = 0...max_spectra
    !                  j = 1 : photoelec, 2 : Auger, 3 : fluorescence
    !   pop_free_elec(k) : time-averaged pulse-weighted pop. of k-eV electrons
    !
    real(long), allocatable :: spectra(:,:)
    real(long), allocatable :: pop_free_elec(:)
    integer, allocatable :: p_type(:), energy_eV(:)
    integer :: max_spectra
  end type
  type(Process_data) :: S
  integer :: i, j, k, l, N, N_t, i_step
  integer, allocatable :: config(:,:)   ! electronic configuration
  real(long), allocatable :: work(:,:)  ! for RK4 routine
  real(long) :: t, t0, dt, p_env, charge_avg, rate
  real(long) :: N_ph_abs                ! # of absorbed photons
  real(long) :: E_absorbed              ! absorbed energy via photons
  real(long) :: E_emitted               ! emitted energy via photons
  real(long) :: sum_spectra(3)  ! j = 1: photoelec, 2: Auger, 3: fluorescence
  integer :: energy_eV
  integer :: OUTPUT, OUTPUT_DFF = 30
  integer :: lower_bound
  integer :: i0                         ! the ground-state of neutral atom
  !
  ! for scattering process
  !
  real(long), allocatable :: scd(:,:)   ! scattering data for all indexes
  real(long), allocatable :: ff(:,:)    ! f0(Q=0), f', and f'' for all indexes
  real(long), allocatable :: ff0(:,:)   ! f0(Q) for all indexes
  real(long), allocatable :: ixs(:,:)   ! inelastic scattering for all indexes
  real(long), allocatable :: cs(:)      ! scattering cross section at time t
  real(long), allocatable :: n_dist(:)  ! ang. dist. of # of scattered photons
  real(long), allocatable :: n0_dist(:) ! ang. dist. of # from unmadaged
  real(long), allocatable :: ni_dist(:) ! ang. dist. of # of inelastic scat. ph.
  real(long), allocatable :: ni0_dist(:)! ang. dist. of # from unmadaged
  real(long) :: n_tot, n0_tot           ! total # of scattered photons
  real(long) :: ni_tot, ni0_tot         ! total # of inelastically scat. ph.
  real(long) :: R_factor                ! R factor
  real(long) :: I_tot, I0_tot           ! for Neutze(2000)'s definition
  real(long) :: Q, theta, omega_in, ratio
  real(long) :: alpha = ONE / SPEED_C_AU  ! fine-structure constant, alpha = 1/c
  real(long) :: f2_avg, f0_avg, fp_avg, fpp_avg, f0_Q
  real(long) :: ff_t_real, ff_t_cmplx
  real(long), allocatable :: ff_avg(:)
  real(long) :: MAD_a, MAD_b, MAD_c, MAD_a_tilde, Z
  !
  ! start your program with time checking!
  !
  call init_time()
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "REQ" )
  call print_parameters( P, "direct rate equation solver, " // GITVERSION )
  OUTPUT = P%OUTPUT
  if ( P%yes_dispersion ) then
    P%yes_scattering = .true.   ! always mad with scat.  it's really true.
    !
    ! store dynamical form factors for further analysis
    !
    open( OUTPUT_DFF, file=trim(P%filename) // ".dff" )
  end if
  omega_in = P%PE_start         ! incident photon energy
  !
  ! memory allocation
  ! N: # of all possible configurations, i.e. # of equations
  ! originally, N is given by occupation numbers := PI_i [ occ_n(i) + 1 ]
  !
  !  N = product( P%init_config(:,3) + 1 )
  !
  ! now, N is given by the ground-state of the neutral atom of Z
  !
  !  do i = 1, P%N_orb - 1         ! open-shell must locate at the end
  !    l = P%init_config(i,2)
  !    j = j + ( 4 * l + 2 )       ! max_occ = 4l+2; added to j
  !    N = N * ( 4 * l + 2 + 1 )   ! possible config: 0...max_occ ==> x(max_occ+1)
  !  end do
  !  N = N * ( ( P%Z - j ) + 1 )
  !
  ! here, we consider a general case, including initially unoccupied subshells
  ! then N is given by the maximum occupation numbers 
  !   for all P%init_config(:,:) subshells := PI_i [ max_occ_n(i) + 1 ]
  !
  N = 1; j = 0
  do i = 1, P%N_orb
    l = P%init_config(i,2)
    j = j + ( 4 * l + 2 )       ! max_occ = 4l+2; added to j
    N = N * ( 4 * l + 2 + 1 )   ! possible config: 0...max_occ ==> x(max_occ+1)
  end do
  i0 = N * ( ( P%Z - j ) + 1 )  ! index for the ground-state of the neutral atom
  S%N = N
  allocate( config(P%N_orb,3) )
  config(:,:) = P%init_config(:,:)
  call index_to_config( N, config )
  lower_bound = P%Z - sum( config(:,3) ) ! can be negative
  allocate( C(N), Cdot(N) )     ! probabilities of all configurations
  allocate( C_bak(N) )          ! backup for C; used if P%yes_pop_reduced is on
  allocate( charge(N) )         ! charges of all configurations
  allocate( pop(lower_bound:P%Z) )        ! prob. of all species; C, C+, C++, and so on
  allocate( C_end(N), pop_end(lower_bound:P%Z) )  ! quantities at the end of the pulse
  allocate( C_avg(N), pop_avg(lower_bound:P%Z) )  ! time-averaged quantities
  allocate( work(N,5) )         ! for RK4 routine
  !
  ! determine charge(:) for each configuration
  !
  config(:,:) = P%init_config(:,:)              ! copy config from Param P
  do j = 1, N
    call index_to_config( j, config )
    charge(j) = P%Z - sum( config(:,3) )        ! Z - ( sum up all occ# )
  end do
  !
  ! if ( P%yes_pop_reduced ) is on, then
  ! - set up ground-state configuration for given charges
  ! - modify C_avg for elastic and inelastic scattering and for MAD
  ! - modify ff_avg for MAD
  !
  if ( P%yes_pop_reduced ) then
    allocate( ground_config(0:P%Z) )    ! store indexes for the ground config.
    do i = 0, P%Z                       ! scan over all charge states
      ground_config(i) = find_ground_config( i, charge(:), config(:,:) )
    end do
  end if
  !
  ! check the max. resolution from a given photon energy (PE_start)
  !   Q = 2 alpha omega sin(theta/2) ==> Qmax = 2 alpha omega, where theta=pi
  !   and we have another expression for Qmax related to the resolution
  !   Qmax = 2 pi / d
  !   thus, 2 alpha omega = 2 pi / d
  !     ==> d_max = pi / ( alpha omega )
  !
  if ( P%yes_scattering ) then
    if ( P%resolution < PI * SPEED_C_AU / omega_in ) then
      write(OUTPUT, "('# [WARNING] Adjust the resolution according to '," // &
                    "'a given photon energy')")
      write(OUTPUT, "('#   photon energy (eV) = ',f9.2)") omega_in * AU2EV 
      write(OUTPUT, "('#   given resolution (angstrom)    = ',f10.5)") &
            P%resolution * AU2ANG
      P%resolution = PI * SPEED_C_AU / omega_in
      write(OUTPUT, "('#   adjusted resolution (angstrom) = ',f10.5)") &
            P%resolution * AU2ANG
      write(OUTPUT, "('#')")
      theta = PI                        ! angle from 0 to pi
    else
      theta = TWO * asin( PI / P%resolution / alpha / omega_in )
    end if
    !
    ! prepare angular grids for integrating cross sections
    !   extended-LGL: x%r(0...N+1) because it includes the end points.
    !
    x = generate_Grid( "extended-LGL", P%N_scd, cos(theta), ONE )
  end if
  !
  ! read rate data from the input file
  ! and prepare the matrix containing all rates
  !
  write(OUTPUT, "('# Filenames:')")
  call load_rates( P, S )
  !
  ! prepare spectra data
  !
  if ( P%yes_spectra ) then
    allocate( S%spectra(lower_bound:S%max_spectra,3) )
    allocate( S%pop_free_elec(lower_bound:S%max_spectra) )
    S%spectra(:,:) = ZERO
    S%pop_free_elec(:) = ZERO
  end if
  !
  ! set up an initial value
  !
  C_avg(:) = ZERO
  i_step = max( 10, P%N_time / 100 )
  N_ph_abs = ZERO
  !
  ! set up the initial configuration
  !
  C(:) = ZERO
  if ( P%i0 < 0 .or. P%i0 > N ) P%i0 = i0  ! initial = neutral if not specified
  C(P%i0) = ONE
  !
  ! consider synchrotron simulation if ground-state config. of neutral atom
  !
  if ( P%yes_synchrotron .and. P%i0 == i0 ) then
    C(P%i0) = ZERO            ! make a vacancy at P%i0
    do l = 1, S%N_process
      if ( S%initial(l) == P%i0 .and. S%p_type(l) == 1 ) &  ! p_type=1: pcs
        C(S%final(l)) = S%value(l)      ! put a pcs value into C(:)
    end do
    if ( sum( C(:) ) == ZERO ) stop "No pcs data for the given configuration"
    C(:) = C(:) / sum( C(:) )   ! C(:) contain shake-off braching ratios.
  end if
  !
  ! prepare the elastic scattering calculations
  ! non-resonant: requires only the SCD file
  ! resonant: requires both the SCD and MAD files
  !
  if ( P%yes_scattering ) then
    !
    ! memory allocation
    !
    allocate( scd(N,0:P%N_scd+1) )
    allocate( ff(N,0:2) )      ! f0(Q=0), f', f''
    allocate( ff0(N,0:P%N_scd+1) )
    allocate( ff_avg(0:P%N_scd+1) )
    allocate( cs(0:P%N_scd+1) )
    allocate( n_dist(0:P%N_scd+1), n0_dist(0:P%N_scd+1) )
    n_dist(:) = ZERO
    ff_avg(:) = ZERO
    if ( P%yes_inelastic_scattering ) then
      allocate( ixs(N,0:P%N_scd+1) )
      allocate( ni_dist(0:P%N_scd+1), ni0_dist(0:P%N_scd+1) )
      ni_dist(:) = ZERO
      call load_scd( P, S, x, scd, ff, ff0, ixs ) ! read (MAD), SCD, IXS
    else
      call load_scd( P, S, x, scd, ff, ff0 )      ! read (MAD), SCD
    end if
  end if
  write(OUTPUT, "('#')")
  !
  ! print out the heading
  !
  call print_divider( OUTPUT, "Time propagation" )
  write(OUTPUT, "('#   column 1: time(fs)')")
  write(OUTPUT, "('#   column 2: pulse envelope (normalized)')")
  write(OUTPUT, "('#   column 3: 1 - total population')")
  write(OUTPUT, "('#   column 4: population of the neutral atom')")
  if ( P%Z == 1 ) then
    write(OUTPUT, "('#   column 5: population of +1')")
  else
    write(OUTPUT, "('#   column 5-',i2,': population of +1 ~ ',sp,i3)") &
      P%Z+4, P%Z
  end if
  write(OUTPUT, "('#   column',i2,': number of absorbed photons')") P%Z+5
  if ( P%yes_dispersion ) then
    write(OUTPUT, "('#   column',i2,': Re[ dynamical form factor ]')") P%Z+6
    write(OUTPUT, "('#   column',i2,': Im[ dynamical form factor ]')") P%Z+7
  end if
  write(OUTPUT, "('#')")
  write(OUTPUT, "('#',a9,2a10)", advance="no") "t(fs)", "pulse_env", "1-P_tot"
  call print_all_ions( OUTPUT, P%Z )
  write(OUTPUT, "(a10)", advance="no") "N_ph_abs"
  if ( P%yes_dispersion ) then
    write(OUTPUT, "(2a15)", advance="no") "Re[f~(t)]", "Im[f~(t)]"
    write(OUTPUT_DFF, "('# ',a8,2x,a)") "time(fs)", &
      "Re[f~(Qmax,t)] ... Re[f~(Q=0,t)]  Im[f~(t)]"
    write(OUTPUT_DFF, "(a10,500es15.7)") "#[f^0(Q)] ", ff0(N,0:P%N_scd+1)
  end if
  write(OUTPUT, "()")
  !============================================================================
  ! time-propagation: numerical integration of rate equations
  !============================================================================
  N_t = int( ( P%t_end_au - P%t_start_au ) / P%dt_au )
  t0 = P%t_start_au             ! in a.u.
  dt = P%dt_au                  ! in a.u.
  time_evolution: do i = 0, N_t
    t = t0 + dt * dble(i)
    p_env = pulse_envelope( P, t, yes_probing = .true. )
    !
    ! time propagation
    !
    call RK4( N, C, t, t+dt, work, S, P )
!    if ( any( C(:) > ONE ) .or. any( C(:) < ZERO ) ) then
!      write(STDERR, *) "Probability should be 0 <= P <= 1."
!      write(STDERR, *) C(:)
!    end if
    !
    ! compute the average probability weighted by the normalized pulse
    ! SEE: Eq. (25) in Son, Young & Santra, PRA 83, 033402 (2011).
    !      C_avg(I) -> \bar{P}_I
    !      C(I)     -> P_I(t)
    !      p_env    -> f(t)
    !
    C_avg(:) = C_avg(:) + p_env * C(:) * dt
    !
    ! compute the spectra of electrons or photons
    ! - S%p_type(l) = 1 : photoelectron
    ! - S%p_type(l) = 2 : Auger electron
    !
    if ( P%yes_spectra ) then
      do l = 1, S%N_process
        rate = compute_rate( P, S%value(l), S%cs_order(l), t )
        energy_eV = S%energy_eV(l)
        S%spectra( energy_eV, S%p_type(l) ) = &
          S%spectra( energy_eV, S%p_type(l) ) + C( S%initial(l) ) * rate * dt
      end do
      S%pop_free_elec(:) = S%pop_free_elec(:) + &
        p_env * ( S%spectra(:,1) + S%spectra(:,2) ) * dt
    end if
    !
    ! compute the number of absorbed photons
    !
    do l = 1, S%N_process
      if ( S%p_type(l) /= 1 ) cycle
      rate = compute_rate( P, S%value(l), S%cs_order(l), t )
      N_ph_abs = N_ph_abs + C( S%initial(l) ) * rate * dt * dble(S%cs_order(l))
    end do
    !
    ! compute the number of scattered photons (angular distribution)
    !     d N                                            d sigma
    !   ------- = INT dt [ flux(t) * SUM P_{conf}(t) * [ ------- ]_{conf} ]
    !   d Omega                     {conf}               d Omega
    ! SEE: Eq. (20) [ before integral over Q ] in PRA 83, 033402 (2011).
    ! SEE: Eq. (22) in Son, Young & Santra, PRA 83, 033402 (2011).
    !
    if ( P%yes_scattering ) then
      !
      ! if P%yes_pop_reduced is on, modify C(:) according to pop(:)
      ! this modified C(:) is only used for scattering calculations
      ! ESPECIALLY for dispersion correction
      !
      if ( P%yes_pop_reduced ) then
        C_bak(:) = C(:)         ! make a backup
        pop(:) = ZERO
        do l = 1, N             ! pop(:) : charge-state populations
          pop( charge(l) ) = pop( charge(l) ) + C(l)
        end do
        C(:) = ZERO             ! ignore configurational population C(:)!!!
        do j = 0, P%Z           ! scan over all charge states
          l = ground_config(j)
          C(l) = pop(j)         ! recalculate C(:) using pop(:)
        end do
      end if
      !
      ! print f''(t)
      !
      if ( P%yes_dispersion ) then
        ff_t_cmplx = sum( ff(:,2) * C(:) )   ! imag. of dynamical form factor
        write(OUTPUT_DFF, "(f10.3)", advance="no") t * AU2SEC / 1.0e-15_long
      end if
      !
      ! scan over Q grid
      !
      do k = 0, P%N_scd + 1
        !
        ! calculate the time-dependent diff. scattering cross section at Q_k
        ! SEE: Eq. (19) in Son, Young & Santra, PRA 83, 033402 (2011).
        !
        cs(k) = sum( C(:) * scd(:,k) )
        n_dist(k) = n_dist(k) + P%fluence * p_env * cs(k) * dt
        if ( P%yes_inelastic_scattering ) then
          cs(k) = sum( C(:) * ixs(:,k) )
          ni_dist(k) = ni_dist(k) + P%fluence * p_env * cs(k) * dt
        end if
        !
        ! dynamical form factor for the MAD analysis
        ! ff_avg --> MAD a_tilde coefficient
        ! ff_t_real, ff_t_cmplx: real and imag. parts of dynamical form factor
        ! - ff_t_cmplx is calculated only once because it doesn't depends on Q
        ! - ff_t_real will have a value at Q=0 (k=P%N_scd+1) at the end
        !
        if ( P%yes_dispersion ) then
          ff_t_real = sum( ( ff0(:,k) + ff(:,1) ) * C(:) )
          ff_avg(k) = ff_avg(k) + dt * p_env * ( ff_t_real**2 + ff_t_cmplx**2 )
          write(OUTPUT_DFF, "(es15.7)", advance="no") ff_t_real
        end if
      end do
      if ( P%yes_dispersion ) write(OUTPUT_DFF, "(es15.7)") ff_t_cmplx
      !
      ! the above implementation is almost 2 times faster than the below
      !
      !cs(:) = ZERO
      !do j = 1, N
      !  if ( C(j) > 1.0e-8_long ) cs(:) = cs(:) + C(j) * scd(j,:)
      !end do
      !n_dist(:) = n_dist(:) + P%fluence * p_env * cs(:) * dt
      !
      ! restore C(:) if P%yes_pop_reduced is used
      !
      if ( P%yes_pop_reduced ) C(:) = C_bak(:)
    end if
    !
    ! print out at every i_step
    !
    if ( mod(i,i_step) == 0 ) then
      pop(:) = ZERO
      do l = 1, N
        pop( charge(l) ) = pop( charge(l) ) + C(l)
      end do
      where ( pop(:) < 1.0e-99_long ) pop(:) = ZERO
      write(OUTPUT, "(f10.3,es10.3,sp,es10.2,s,100es10.3)", advance="no") &
        t * AU2SEC / 1.0e-15_long, p_env, ONE - sum( pop(:) ), pop(0:)
      write(OUTPUT, "(es10.3)", advance="no") N_ph_abs
      if ( P%yes_dispersion ) &
        write(OUTPUT, "(2es15.7)", advance="no") ff_t_real, ff_t_cmplx ! at Q=0
      write(OUTPUT, "()")
    end if
  end do time_evolution
  !============================================================================
  ! post-time-evolution: further relaxation via Auger and fluorescence
  !============================================================================
  C_end(:) = C(:)               ! store information at the end of the pulse
  call post_time_evolution( P, S, C )
  !
  ! compute probabilities of species and averaged values
  !
  where ( C(:)     < 1.0e-99_long ) C(:) = ZERO
  where ( C_end(:) < 1.0e-99_long ) C_end(:) = ZERO
  where ( C_avg(:) < 1.0e-99_long ) C_avg(:) = ZERO
  pop(:) = ZERO
  pop_end(:) = ZERO
  pop_avg(:) = ZERO
  do l = 1, N
    pop( charge(l) ) = pop( charge(l) ) + C(l)
    pop_end( charge(l) ) = pop_end( charge(l) ) + C_end(l)
    pop_avg( charge(l) ) = pop_avg( charge(l) ) + C_avg(l)
  end do
  charge_avg = sum( (/ ( dble(i) * pop_avg(i), i = 0, P%Z ) /) )
  !
  ! double-check simulation
  !
  do i = lower_bound, -1
    if ( pop(i) /= ZERO ) then
      write(STDERR,*) "There are some populations in negative charges."
      write(STDERR,*) "charge = ", i, "pop = ", pop(i)
      stop
    end if
  end do
  !
  ! print out summary of populations (or probabilities)
  ! - |POP|     ones after the pulse and a long long time
  ! - |POP_AVG| ones averaged with weights by the normalized pulse
  ! - |POP_PULSE_END| ones at the end of the pulse; for back-compatibility only
  !
  write(OUTPUT, "('#',a15,14x,100es10.3)") "|POP_PULSE_END|", pop_end(0:)
  write(OUTPUT, "('#',a9,20x,100es10.3)")  "|POP_AVG|", pop_avg(0:)
  write(OUTPUT, "('#',a9,20x,100es10.3)")  "|POP|    ", pop(0:)
  write(OUTPUT, "('#',a9,20x,100es11.3E3)")  "|POPFULL|", pop(0:)
  write(OUTPUT, "(/)")
  !
  ! print out charge-state distribution
  !
  call print_divider( OUTPUT, "Charge-state distribution" )
  call print_charge_states( OUTPUT, pop(0:P%Z), pop_avg(0:P%Z) )
  write(OUTPUT, "('# Averaged number of absorbed photons        = ',f8.4)") &
                N_ph_abs
  write(OUTPUT, "(/)")
  !
  ! print out some important configurations (all if verbose mode)
  !
  call print_divider( OUTPUT, &
    "Population analysis of individual configurations" )
  write(OUTPUT, "('#   ',21x,2a12,4x,a)") "Pop", "Pop_avg", "description"
  do j = 1, N
    if ( .not. P%yes_verbose .and. &
         C(j) < 1.0e-5_long .and. C_avg(j) < 1.0e-5_long ) cycle
    call index_to_config( j, config )
    write(OUTPUT, "('#   config = ',i10,2x,2es12.5,' ; ',sp,i3,': ',a)") &
      j, C(j), C_avg(j), charge(j), trim( config_to_string( config ) )
  end do
  write(OUTPUT, "(/)")
  !
  ! print out electron and photon spectra
  !
  if ( P%yes_spectra ) then
    call print_divider( OUTPUT, "Electron/photon spectra" )
    write(OUTPUT, "('#   ',a6,3a13)") &
      "E(eV)", "Photoelec.", "Auger elec.", "fluorescence"
    E_absorbed = N_ph_abs * P%PE_start * AU2EV
    E_emitted = ZERO
    sum_spectra(1:3) = ZERO
    do j = 0, S%max_spectra
      where ( S%spectra(j,1:3) < 1.0e-99_long ) S%spectra(j,1:3) = ZERO
      write(OUTPUT, "(4x,i6,3es13.5)") j, S%spectra(j,1:3)
      sum_spectra(1:3) = sum_spectra(1:3) + S%spectra(j,1:3) 
      E_emitted = E_emitted + dble(j) * S%spectra(j,3)  ! fluo. only, in eV
    end do
    write(OUTPUT, "('#|SPECTRA|',3es13.5)") sum_spectra(1:3)
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# Energy gain from incident photons (eV) = ',f10.2)") &
                  E_absorbed
    write(OUTPUT, "('# Energy loss by emitted photons (eV)    = ',f10.2)") &
                  E_emitted
    write(OUTPUT, "(/)")
  end if
  !
  ! reduce C_avg(:) according to charge populations if P%yes_pop_reduced is on
  !
  if ( P%yes_pop_reduced ) then
    C_avg(:) = ZERO             ! ignore configuration populations!!!
    do i = 0, P%Z               ! scan over all charge states
      j = ground_config(i)
      C_avg(j) = pop_avg(i)     ! recalculate C_avg(:) using pop_avg(:)
    end do
    !
    ! n_dist and n_tot should be simply calculated from C_avg(:)
    ! they are recalculated with reduced C_avg(:)
    ! BUT they are already calculated during the time propagation
    !
    !do k = 0, P%N_scd + 1
    !  n_dist(k)  = P%fluence * sum( C_avg(:) * scd(:,k) )
    !  ni_dist(k) = P%fluence * sum( C_avg(:) * ixs(:,k) )
    !end do
  end if
  !
  ! print out # of scattered photons and R factor
  !
  if ( P%yes_scattering ) then
    call print_divider( OUTPUT, "Scattering and/or MAD analysis" )
    write(OUTPUT, "('# All calculations are based on ')", advance="no")
    if ( P%yes_pop_reduced ) then
      write(OUTPUT, "('charge-state')", advance="no")
    else
      write(OUTPUT, "('configuration')", advance="no")
    end if
    write(OUTPUT, "(' populations'/'#')")
    !
    ! set up cross section of the neutral atom as undamaged one
    !
    n0_dist(:) = P%fluence*scd(P%i0,:)  ! # from undamaged
                                        ! SEE: Eq. (23) in PRA 83, 033402 (2011)
    n_tot  = TWO*PI * sum( n_dist(0:)  * x%weight(0:) ) ! 2pi from INT d(phi)
    n0_tot = TWO*PI * sum( n0_dist(0:) * x%weight(0:) ) ! 2pi from INT d(phi)
    !
    ! R-factor defined in Neutze et al., Nature 406, 752 (2000)
    ! R = SUM_Omega | sqrt(I(Omega)) / I_tot - sqrt(I0(Omega)) / I0_tot |
    !   where I_tot  = SUM_Omega sqrt( I(Omega)  )    I:  damaged
    !         I0_tot = SUM_Omega sqrt( I0(Omega) )    I0: undamaged
    ! SEE: Eq. (21) in Son, Young & Santra, PRA 83, 033402 (2011).
    !
    I_tot  = TWO*PI * sum( sqrt(n_dist(0:))  * x%weight(0:) )
    I0_tot = TWO*PI * sum( sqrt(n0_dist(0:)) * x%weight(0:) )
    R_factor = TWO*PI * sum( abs( sqrt(n_dist(0:))/I_tot - &
                                  sqrt(n0_dist(0:))/I0_tot ) * x%weight(0:) )
    theta = acos( x%r(0) )      ! at Q_max
    write(OUTPUT, "('# Scattering process with x-ray damage')")
    write(OUTPUT, "('#   Resolution d (angstrom) = ',f8.3)") &
      P%resolution * AU2ANG
    write(OUTPUT, "('#   Q_max = 2pi / d (a.u.)  = ',es14.7,5x," // &
                  "'[ max. angle = ',f6.2,' (deg) ]')") &
      TWO*PI / P%resolution, theta / pi * 180.0_long
    write(OUTPUT, "('#   (elastic) N_sc_tot [undamaged]   = ',es14.7)") n0_tot
    write(OUTPUT, "('#   (elastic) N_sc_tot [damaged]     = ',es14.7)") n_tot
    write(OUTPUT, "('#   R-factor                         = ',es14.7)") R_factor
    if ( P%yes_inelastic_scattering ) then
      ni0_dist(:) = P%fluence * ixs(P%i0,:)     ! # from undamaged
      ni_tot  = TWO*PI * sum( ni_dist(0:)  * x%weight(0:) )
      ni0_tot = TWO*PI * sum( ni0_dist(0:) * x%weight(0:) )
      write(OUTPUT,"('#   (inelastic) N_sc_tot [undamaged] = ',es14.7)") ni0_tot
      write(OUTPUT,"('#   (inelastic) N_sc_tot [damaged]   = ',es14.7)") ni_tot
      write(OUTPUT,"('#   alpha^2 omega_in = ',es14.7)") alpha**2 * omega_in
    end if
    write(OUTPUT, "('#')")
    !
    ! print out scattering patterns as a function of Q
    !
    write(OUTPUT, "('# Angular distribution of N of scattered photons:')")
    write(OUTPUT, "('#   sigma_Thompson = alpha^4 [ 1+cos(theta)^2 ]/2'," // &
                  "'  (for unpolarized x-rays)')")
    write(OUTPUT, "('#   Q = 2alpha omega_in sin( theta / 2 )')")
    write(OUTPUT, "('#   es: elastic, is: inelastic, '," // &
                  "'n0: undamaged, n: damaged')")
    write(OUTPUT, "('#',a9,a11,2a12)", advance="no") &
      "Q(au^-1)", "theta(deg)", "n0_es", "n_es"
    if ( P%yes_inelastic_scattering ) &
      write(OUTPUT, "(2a12,a11)", advance="no") "n0_is", "n_is", "w_F/w_in"
    write(OUTPUT, "()")
    do k = P%N_scd + 1, 0, -1
      theta = acos( x%r(k) )
      Q = TWO * alpha * omega_in * sin( theta / TWO )
      write(OUTPUT, "(f10.6,f11.6,2es12.5)", advance="no") &
        Q, theta / pi * 180.0_long, n0_dist(k), n_dist(k)
      if ( P%yes_inelastic_scattering ) then
        ratio = ( sqrt( ( ONE / alpha**2 - omega_in * cos(theta) )**2 &
                        + TWO * omega_in / alpha**2 - omega_in**2 ) &
                  - ONE / alpha**2 + omega_in * cos(theta) ) / omega_in
        write(OUTPUT, "(2es12.5,f11.8)", advance="no") &
          ni0_dist(k), ni_dist(k), ratio
      end if
      write(OUTPUT, "()")
    end do
    !
    ! time-averaged form factors with the dispersion correction
    !   f2_avg = sum_I1 [ int dt J(t) P_I1(t) ] | f_1,I1(0,omega) |^2 --> a
    !   ff_avg = int dt J(t) | sum_I1 P_I1(t) f_1,I1(0,omega) |^2 --> a~
    !
    !   a(0,omega)  = f2_avg / Z**2
    !   b(0,omega)  = 2 * ( f0_avg + fp_avg ) / Z
    !   c(omega)    = 2 * fpp_avg / Z
    !   a~(0,omega) = ff_avg / Z**2             where Z = f^0_H(Q) at Q=0
    !
    ! SEE: Eq. (5) in Son, Chapman & Santra, submitted
    !
    if ( P%yes_dispersion ) then
      !
      ! print out at Q=0
      !
      f2_avg  = sum( ( ( ff(:,0) + ff(:,1) )**2 + ff(:,2)**2 ) * C_avg(:) )
      f0_avg  = sum( ff(:,0) * C_avg(:) )
      fp_avg  = sum( ff(:,1) * C_avg(:) )
      fpp_avg = sum( ff(:,2) * C_avg(:) )
      write(OUTPUT, "(/)")
      write(OUTPUT, "('# Time-averaged form factors (Q=0):')")
      write(OUTPUT, "('#   ',a14,' = ',es14.7)") "avg. f0",    f0_avg
      write(OUTPUT, "('#   ',a14,' = ',es14.7)") "avg. f'",    fp_avg
      write(OUTPUT, "('#   ',a14,' = ',es14.7)") "avg. f''",   fpp_avg
      write(OUTPUT, "('#   ',a14,' = ',es14.7)") "abs(avg. f)^2", &
        (f0_avg+fp_avg)**2 + fpp_avg**2
      write(OUTPUT, "('#   ',a14,' = ',es14.7,2x,a)") "avg. abs(f)^2", &
        f2_avg, "--> a * Z^2"
      write(OUTPUT, "('#   ',a14,' = ',es14.7,2x,a)") "avg. abs(f~)^2", &
        ff_avg(P%N_scd+1), "--> a~ * Z^2"       ! k=P%N_asc+1 --> Q=0
      !
      ! print out MAD coefficients at all Q
      ! ff0(i,k) : normal form factor of i-th configuration at Q_k
      !   i=N         ==> neutral
      !   k=P%N_scd+1 ==> Q=0
      !
      write(OUTPUT, "('#')")
      write(OUTPUT, "('# MAD coefficients for all Q:')")
      write(OUTPUT, "('#',a12,5a13)") &
        "Q (a.u.^-1)", "f0[neutral]", "a", "b", "c", "a~"
      MAD_c = TWO * ( fpp_avg ) / Z
      do k = P%N_scd + 1, 0, -1
        Q = TWO * alpha * omega_in * sin( acos( x%r(k) ) / TWO )
        f0_Q = ff0(N,k)      ! f0 of the neutral atom at Q
        MAD_a = sum( ( ( ff0(:,k) + ff(:,1) )**2 + ff(:,2)**2 ) * &
                     C_avg(:) ) / f0_Q**2
        MAD_b = TWO * ( sum( ff0(:,k) * C_avg(:) ) + fp_avg ) / f0_Q
        MAD_c = TWO * ( fpp_avg ) / f0_Q
        MAD_a_tilde = ff_avg(k) / f0_Q**2
        write(OUTPUT, "(f13.8,5es13.5)") &
          Q, f0_Q, MAD_a, MAD_b, MAD_c, MAD_a_tilde
      end do
      !
      ! print out MAD coefficients at Q=0
      !
      Z = ff(N,0)                    ! f0 of the neutral atom
      MAD_a = f2_avg / Z**2
      MAD_b = TWO * ( f0_avg + fp_avg ) / Z
      MAD_c = TWO * ( fpp_avg ) / Z
      MAD_a_tilde = ff_avg(P%N_scd+1) / Z**2    ! Q=0 at k=P%N_scd+1
      write(OUTPUT, "('#')")
      write(OUTPUT, "('# MAD coefficients (Q=0):')")
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a(0,omega)", MAD_a
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "b(0,omega)", MAD_b
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "c(omega)",   MAD_c
      write(OUTPUT, "('#   ',a11,' = ',es14.7)") "a~(0,omega)", MAD_a_tilde
   end if
    !
    ! print out summary
    !
    write(OUTPUT, "('#')")
    write(OUTPUT, "('# PUNCH ',6a12)") "fluence(au)", "FWHM(au)", &
      "N(undam.)", "N(damaged)", "R-factor", "avg. charge"
    write(OUTPUT, "('#|PUNCH|',6es12.5)") &
      P%fluence, P%FWHM, n0_tot, n_tot, R_factor, charge_avg
    if ( P%yes_dispersion ) then
      write(OUTPUT, "('# MAD   ',6a12)") "PE(eV)", &
        "a(0,omega)", "b(0,omega)", "c(omega)", "a~(0,omega)"
      write(OUTPUT, "('#|MAD|  ',f12.2,4es12.4)") omega_in * AU2EV, &
        MAD_a, MAD_b, MAD_c, MAD_a_tilde
    end if
    write(OUTPUT, "(/)")
  end if
  !
  ! print out foobar.pop for post-process of scattering
  !
  if ( P%yes_population ) then
    open( P%IO_POP, file=P%file_pop, form="formatted" )
    !
    ! print out pulse-weighted time-averaged populations for all configurations
    !
    write(P%IO_POP, "('# Pulse-weighted time-averaged populations'/'#')")
    if ( P%yes_pop_reduced ) then
      write(P%IO_POP, "('# Charge-state populations:')")
      write(P%IO_POP, "('# printed with the ground-state config. only')")
    else
      write(P%IO_POP, "('# Configurational populations:')")
    end if
    write(P%IO_POP, "('# Config    Pop_avg')")
    do i = 1, N
      if ( C_avg(i) > ZERO ) then
        call index_to_config( i, config )       ! fing config corresponding to i
        write(P%IO_POP, "(a,2x,es18.10)") &
          trim( config_to_string( config ) ), C_avg(i)
      end if
    end do
    write(P%IO_POP, "(/)")
    !
    ! print out time-averaged population for free electrons
    !
    if ( P%yes_spectra ) then
      write(P%IO_POP, "('# Free electron populations')")
      write(P%IO_POP, "('#',a5,a18)") "E(eV)", "Pop_avg"
      do j = 0, S%max_spectra
        if ( S%pop_free_elec(j) > 1e-99_long ) &
          write(P%IO_POP, "(i6,es18.10)") j, S%pop_free_elec(j)
      end do
      write(P%IO_POP, "('# sum:',es18.10)") sum( S%pop_free_elec(:) )
    end if
  end if
  !
  ! print out the foot note
  !
  call footnote( OUTPUT )
contains

subroutine load_rates( P, S )
  type(Param), intent(in) :: P
  type(Process_data), intent(inout) :: S
  integer :: i, j, l
  integer :: INFO
  real(long) :: value, energy
  character(len=1) :: tag
  !
  ! find total # of processes in the data file
  !
  open( P%IO_DATA, file=P%file_data, action="read", form="formatted" )
  l = 0
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i == 0 .or. j == 0 .or. value == ZERO ) cycle  ! skip unmeaningful
    l = l + 1
  end do
  S%N_process = l
  allocate( S%initial(S%N_process), S%final(S%N_process), S%value(S%N_process) )
  allocate( S%cs_order(S%N_process), S%p_type(S%N_process) )
  allocate( S%energy_eV(S%N_process) )
  rewind( P%IO_DATA )
  write(OUTPUT, "('#   considering how many processes: ',i10)") S%N_process
  !
  ! read rate data and save into S
  !   i: initial configuation index
  !   j: final configuration index
  !   value: rate or cross section; both should be in a.u.
  !
  l = 0
  S%max_spectra = 0
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value, energy
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i == 0 .or. j == 0 .or. value == ZERO ) cycle  ! skip unmeaningful
!    if ( i > S%N .or. j > S%N ) then
!      write(STDERR, *) "In the data file,"
!      write(STDERR, *) tag, i, j, value
!      write(STDERR, *) "Index should not be greater than ", S%N
!      stop
!    end if
    !
    ! if the same pair of (i,j) already exists, it creates a new item (new l)
    ! so the prev. data are NOT overriden; effectively new data are accumulated
    !
    l = l + 1
    S%initial(l) = i
    S%final(l) = j
    S%value(l) = value
    if ( P%yes_dat_eV ) then                    ! nint(): round to nearest int.
      S%energy_eV(l) = nint(energy)             ! energy is given in eV already
    else
      S%energy_eV(l) = nint(energy * AU2EV)     ! convert a.u. into eV
    end if
    if ( S%max_spectra < S%energy_eV(l) ) S%max_spectra = S%energy_eV(l)
    if ( tag == "P" .or. tag == "V" .or. tag == "E" ) then      ! for photoabsorption
      S%p_type(l) = 1;  S%cs_order(l) = 1
    else if ( tag == "T" ) then                 ! for two-photon ionization
      S%p_type(l) = 1;  S%cs_order(l) = 2
    else if ( tag == "A" ) then                 ! for Auger decay
      S%p_type(l) = 2;  S%cs_order(l) = 0
    else                                        ! for fluorescence
      S%p_type(l) = 3;  S%cs_order(l) = 0
    end if
  end do
  close( P%IO_DATA )
  write(OUTPUT, "('#   all process data loaded from: ',a)") trim(P%file_data)
  if ( P%yes_spectra ) then
    write(OUTPUT, "('#   max. energy (eV) in spectra from data file: ',i8)") &
      S%max_spectra
  end if
end subroutine load_rates

subroutine load_scd( P, S, x, scd, ff, ff0, ixs )
  type(Param), intent(in) :: P
  type(Process_data), intent(inout) :: S
  type(Grid), intent(in) :: x
  real(long), intent(out) :: scd(:,0:)          ! WARNING: starting from 0!!!
  real(long), intent(out) :: ff(:,0:)           ! WARNING: starting from 0!!!
  real(long), intent(out) :: ff0(:,0:)          ! WARNING: starting from 0!!!
  real(long), intent(out), optional :: ixs(:,0:)        !  starting from 0!!!
  real(long), allocatable :: Q(:), f(:)         ! dim. determined by SCD file
  real(long), allocatable :: theta(:)
  real(long) :: fp(size(scd,1))                 ! fp(i) or fpp(i): i-th config.
  real(long) :: fpp(size(scd,1))
  real(long) :: i_final(P%N_orb), pcs(P%N_orb)
  real(long) :: Q_k, omega, alpha, f0, w, temp1, temp2, sigma_T
  type(Spline) :: SP
  complex(long) :: cff          ! coherent form factor: COMPLEX!!!
  integer :: i, j, k, l, m, N, N_orb, INFO, OUTPUT
  character(len=1) :: tag
  logical :: yes_exist, yes_omega_found
  alpha = ONE / SPEED_C_AU      ! fine-structure constant, alpha = 1/c
  N_orb = P%N_orb               ! this should be N_shell.  # of diff. shells
  OUTPUT = P%OUTPUT
  yes_omega_found = .false.
  !
  ! load MAD ------------------------------------------------------------------
  ! read the dispersion correction from the MAD file
  ! ACHTUNG: even without the dispersion correction,
  !          sometimes it's convenient to read this MAD file 
  !          because it contains PCS data for different photon energies.
  !          so, read the MAD file no matter how yes_dispersion is on or not.
  !
  fp(:) = ZERO
  fpp(:) = ZERO
  ff(:,1) = ZERO     ! f'
  ff(:,2) = ZERO     ! f''
  inquire( file=P%file_mad, exist=yes_exist )
  if ( yes_exist ) then
    open( P%IO_MAD, file=P%file_mad, action="read", form="formatted" )
    omega = P%PE_start * AU2EV  ! be careful, MAD has P.E. in eV
    m = 0
    loop_MAD: do
      !
      ! i: configuration index, see config_to_index()
      ! N: # of data for given index
      ! tag: must be a character of '#'
      !      why do I use '#'?  well... you may use gnuplot with the SCD file .
      !
      read( P%IO_MAD, *, iostat=INFO ) tag, i, N
      if ( INFO < 0 ) exit loop_MAD             ! exit when it ends
      m = m + 1
      read( P%IO_MAD, * ) tag, i_final(:)       ! forget about (-) i_final's...
      do k = 1, N
        read( P%IO_MAD, * ) w, temp1, temp2, pcs(:)
        !
        ! ACHTUNG: What are you doing here!!!
        !          you can interpolate values at any omega.
        !          but here you pick up only one omega from the table.
        !          you must fix it soon.
        !          ... well, it's necessary to avoid reading the whole MAD file.
        !
        if ( abs( omega - w ) < 1.0e-8_long ) then  ! set up data for given P.E.
          yes_omega_found = .true.
          !
          ! set up the dispersion correction for the i-th configuration
          !
          if ( P%yes_dispersion ) then
            ff(i,1) = temp1          ! f'
            ff(i,2) = temp2          ! f''
          end if
          !
          ! change the PCS data for the i-th configuration with given P.E.
          !
          loop_process: do j = 1, S%N_process
            if ( S%initial(j) == i ) then
              do l = 1, N_orb
                if ( S%final(j) == i_final(l) ) S%value(j) = pcs(l)
              end do
            end if
          end do loop_process
        end if
      end do
      read( P%IO_MAD, * )                       ! skip one blank line
      !
      ! check if the photon energy is correctly chosen from the MAD file
      !
      if ( m == 1 .and. .not. yes_omega_found ) then
        write(STDERR, "('ERROR: ',f8.1,' eV is not found in the MAD file.')") &
          omega
        stop
      end if
    end do loop_MAD
    close( P%IO_MAD )
    if ( m /= 0 ) then
      write(OUTPUT, "('#   all MAD data are loaded from: ',a)") &
        trim(P%file_mad)
    else if ( .not. P%yes_dispersion ) then
      write(OUTPUT, "('#   only PCS data are loaded from: ',a)") &
        trim(P%file_mad)
    end if
  else
    write(OUTPUT, "('#   no MAD file is available')")
    write(OUTPUT, "('#   PCS data are fixed at a certain photon energy.')")
  end if
  !
  ! load SCD ------------------------------------------------------------------
  ! read the scattering data from the SCD file
  ! convert them from the coferent form factor into the cross section
  ! SEE: Eqs. (17) and (18) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  !   d sigma
  !   ------- = alpha^4 |f0(Q)|^2 [ the polarization-dependent factor ]
  !   d Omega
  !             where [ ] = ( 1 + cos(theta)^2 ) / 2 for unpolarized x-rays
  !                       = ( 1 - cos(phi)^2 sin(theta)^2 ) for linearly pol.
  !                   however, it doesn't matter for the atomic case.
  !  sigma_T = alpha^4 [ polar. factor ] : Thompson scattering
  !
  open( P%IO_SCD, file=P%file_scd, action="read", form="formatted" )
  m = 0
  scd(1,:) = ZERO     ! no electron in this configuration (bare atom)
  loop_SCD: do
    !
    ! i: configuration index, see config_to_index()
    ! N: # of data for given index
    ! tag: must be a character of '#'
    !
    read( P%IO_SCD, *, iostat=INFO ) tag, i, N
    if ( INFO < 0 ) exit loop_SCD               ! exit when it ends
    m = m + 1
    if ( m == 1 ) allocate( Q(N), f(N) )        ! for the first data set
    do k = 1, N
      read( P%IO_SCD, * ) Q(k), f(k)            ! read the i-th config. data
    end do
    read( P%IO_SCD, * )                         ! skip one blank line
    if ( m == 1 ) then
      SP = Spline_create( Q, f )                ! for the first data set
    else
      call Spline_reuse( SP, f )                ! otherwise, just reuse
    end if
    !
    ! set up the i-th scd data
    !   ff(i,0) = f0_i(Q=0), ff(i,1) = f'_i and ff(i,2) = f''_i
    !   ff0(i,k) = f0_i(Q_k)
    !   scd(i,k) = sigma_T | f0_i(Q_k) + f'_i + i f''_i |^2
    !
    do k = 0, P%N_scd + 1       ! Q=0 at k=P%N_scd+1
      Q_k = TWO * alpha * P%PE_start * sin( acos( x%r(k) ) / TWO )
      f0 = Spline_interpolate( SP, Q_k )
      cff = f0                                  ! non-resonant form factor
      if ( k == P%N_scd + 1 ) ff(i,0) = f0   ! set up f0(Q=0)
      if ( P%yes_dispersion ) cff = cff + ff(i,1) + IMAG_ONE * ff(i,2)
      sigma_T = alpha**4 * ( ONE + x%r(k)**2 ) / TWO    ! Thompson scattering
      scd(i,k) = sigma_T * cdabs( cff )**2
      ff0(i,k) = f0
    end do
  end do loop_SCD
  close( P%IO_SCD )
  write(OUTPUT, "('#   all SCD data are loaded from: ',a)") trim(P%file_scd)
  deallocate( Q, f )
  call purge_Spline( SP )
  !
  ! load IXS ------------------------------------------------------------------
  ! read the inelastic x-ray scattering data from the IXS file
  !
  inquire( file=P%file_ixs, exist=yes_exist )
  if ( present(ixs) .and. yes_exist ) then
    open( P%IO_IXS, file=P%file_ixs, action="read", form="formatted" )
    m = 0
    ixs(1,:) = ZERO     ! no electron in this configuration (bare atom)
    loop_IXS: do
      !
      ! i: configuration index, see config_to_index()
      ! N: # of data for given index
      ! tag: must be a character of '#'
      !
      read( P%IO_IXS, *, iostat=INFO ) tag, i, N
      if ( INFO < 0 ) exit loop_IXS             ! exit when it ends
      m = m + 1
      if ( m == 1 ) allocate( theta(N), f(N) )  ! for the first data set
      do k = 1, N
        read( P%IO_IXS, * ) theta(k), f(k)      ! read the i-th config. data
                                                ! theta in radian, f in sigma_T
      end do
      read( P%IO_IXS, * )                       ! skip one blank line
      if ( m == 1 ) then
        SP = Spline_create( theta, f )          ! for the first data set
      else
        call Spline_reuse( SP, f )              ! otherwise, just reuse
      end if
      !
      ! set up the i-th ixs data
      !
      do k = 0, P%N_scd + 1
        sigma_T = alpha**4 * ( ONE + x%r(k)**2 ) / TWO  ! Thompson scattering
        ixs(i,k) = sigma_T * Spline_interpolate( SP, acos( x%r(k) ) )
      end do
    end do loop_IXS
    close( P%IO_IXS )
    write(OUTPUT, "('#   all IXS data are loaded from: ',a)") trim(P%file_ixs)
    deallocate( theta, f )
    call purge_Spline( SP )
  end if
end subroutine load_scd


!----------------------------------------------------------------------------
! post_time_evolution:
!   After time-evolution during the pulse, the system can further relax.
!   All relaxation processes occur according to the branching ratios.
!   They are repeating until no further relaxation, i.e.,
!   the sum of all relaxation rates becomes zero.
!----------------------------------------------------------------------------
subroutine post_time_evolution( P, S, C )
  type(Param), intent(in) :: P
  type(Process_data), intent(inout) :: S
  real(long), intent(inout) :: C(:)
  real(long) :: sum_rate(S%N), C_old(S%N), branching
  integer :: i, j, l

  !
  ! add up all Auger and fluo. rates for given config.
  !
  sum_rate(:) = ZERO
  do l = 1, S%N_process
    if ( S%cs_order(l) == 0 ) then
      i = S%initial(l)
      if ( C(i) >= 0 ) then
         sum_rate(i) = sum_rate(i) + S%value(l)
      endif
    end if
  end do
  !
  ! further relaxation according to branching ratios, until no more relaxation
  !
  relaxation: do
    C_old(:) = C(:)
    do l = 1, S%N_process
      if ( S%cs_order(l) == 0 ) then
        i = S%initial(l)
        j = S%final(l)
        if ( sum_rate(i) == ZERO  .or. C(i) < 0 ) cycle        ! no relaxation for this config.
        branching = S%value(l) / sum_rate(i)
        if ( P%yes_spectra ) then
          S%spectra( S%energy_eV(l), S%p_type(l) ) = &
            S%spectra( S%energy_eV(l), S%p_type(l) ) + C_old(i) * branching
        end if
        C(i) = C(i) - C_old(i) * branching      ! C(i) decreases by same amount
        C(j) = C(j) + C_old(i) * branching      ! C(j) increases by this amount
      end if
    end do
    do i = 1, S%N                               ! clean up population C(:)
      if ( C(i) < 1.0e-99_long ) then
        C(i) = ZERO
        sum_rate(i) = ZERO                      ! don't consider if C(i) = 0
      end if
    end do
    if ( sum( sum_rate(:) ) == ZERO ) then 
       exit relaxation   ! no further relaxation
    endif
  end do relaxation
end subroutine post_time_evolution


!----------------------------------------------------------------------------
! find out the ground-state configuration for a given charge
! if the sum of the quantum number n for all subshells gives the minumum,
! then it should be the ground-state configuration.
!----------------------------------------------------------------------------
function find_ground_config( charge, charge_array, config ) result (ind)
  integer, intent(in) :: charge
  integer, intent(in) :: charge_array(:)
  integer, intent(inout) :: config(:,:)
  integer :: ind, n, l
  integer :: n_sum, n_sum_min
  integer :: l_sum, l_sum_min
  integer :: i, j
  n_sum_min = 0
  l_sum_min = 0
  do j = 1, size(config,dim=1)
    n = config(j,1)
    l = config(j,2)
    n_sum_min = n_sum_min + n * ( 4 * l + 2 )   ! considering max. occ.
    l_sum_min = l_sum_min + l * ( 4 * l + 2 )   ! considering max. occ.
  end do
  do i = 1, size(charge_array)
    if ( charge == charge_array(i) ) then
      call index_to_config( i, config ) ! fing config corresponding to i
      n_sum = sum( config(:,1) * config(:,3) )
      l_sum = sum( config(:,2) * config(:,3) )
!      print "(i10,x,a,2x,'n_sum=',i4,2x,'l_sum=',i4)", &
!        i, trim( config_to_string( config ) ), n_sum, l_sum
      !
      ! find out the min. sum of n (or the min. sum of l)
      !
      if ( n_sum < n_sum_min .or. &
           ( n_sum == n_sum_min .and. l_sum <= l_sum_min ) ) then 
        ind = i
        n_sum_min = n_sum
        l_sum_min = l_sum
      end if
    end if
  end do
!  print *, "charge = ", charge
!  print *, "index found = ", ind
!  print *
end function find_ground_config


!----------------------------------------------------------------------------
! REQ: rate equations
!      it will be called by RK4
!----------------------------------------------------------------------------
subroutine REQ( t, C, Cdot, S, P )
  real(long), intent(in) :: t, C(:)
  real(long), intent(out) :: Cdot(:)
  type(Process_data), intent(in) :: S
  type(Param), intent(in) :: P
  real(long) :: rate
  integer :: i, j, k
  !
  ! compute Cdot := dC / dt = M C
  ! SEE: Eq. (14) in Son, Young & Santra, PRA 83, 033402 (2011).
  !
  Cdot(:) = ZERO
  do k = 1, S%N_process
    rate = compute_rate( P, S%value(k), S%cs_order(k), t )
    i = S%initial(k)
    j = S%final(k)
    Cdot(i) = Cdot(i) - rate * C(i)     ! C(i) decreases by rate * C(i)
    Cdot(j) = Cdot(j) + rate * C(i)     ! C(j) increases by rate * C(i)
  end do
end subroutine REQ

function compute_rate( P, value, order, t ) result (rate)
  real(long) :: rate
  type(Param), intent(in) :: P
  real(long), intent(in) :: value, t
  integer, intent(in) :: order
  real(long) :: flux
  rate = value
  if ( order > 0 ) then
    !
    ! compute flux
    !   fluence = # of photons / spatial_width^2
    !
    flux = P%fluence * pulse_envelope( P, t )
    rate = rate * flux**order
  end if
end function compute_rate

function pulse_envelope( P, t, yes_probing ) result (f)
  real(long) :: f
  type(Param), intent(in) :: P
  real(long), intent(in) :: t
  logical, intent(in), optional :: yes_probing
  real(long) :: sigma, f1, f2, sigma2, r1, r2
  if ( t < P%t_start_au .or. t > P%t_end_au ) then
    f = ZERO
    return
  end if
  !
  ! normalized Gaussian distribution
  !   f = e^( -t^2 / 2 sigma^2 ) / ( sigma sqrt(2pi) )
  ! where sigma: standard deviation
  !       FWHM (Full Width at Half Maximum), tau = 2 sigma sqrt( 2 ln 2 )
  ! at t=0
  !   f = 2 * sqrt( ln 2 / pi ) / FWHM
  !
  if ( P%pulse_type == 3 ) then         ! Gaussian
    sigma = P%FWHM / TWO / sqrt( TWO*log(TWO) )
    f = exp( -t**2 / TWO / sigma**2 ) / sqrt(TWO*PI) / sigma
  else if ( P%pulse_type == 6 ) then    ! flat-top
    f = ONE / ( P%FWHM )
  else if ( P%pulse_type == 7 ) then    ! pump-probe (two-pulses)
    sigma = P%FWHM / TWO / sqrt( TWO*log(TWO) )
    f = P%fluence_ratio * exp( -t**2 /TWO/sigma**2 ) / sqrt(TWO*PI) / sigma + &
        exp( -( t - P%time_delay )**2 /TWO/sigma**2 ) / sqrt(TWO*PI) / sigma
    if ( present( yes_probing ) ) then
      if ( yes_probing ) &                ! only take the second pulse
        f = exp( -( t-P%time_delay )**2 /TWO/sigma**2 ) / sqrt(TWO*PI) / sigma
    end if
  else if ( P%pulse_type == 8 ) then    ! double Gaussian (split)
    sigma = P%FWHM / TWO / sqrt( TWO * log(TWO) )
    sigma2 = sigma * P%FWHM_ratio
    f1 = exp( -t**2 / TWO / sigma**2 ) / sqrt( TWO * PI ) / sigma
    f2 = exp( -( t - P%time_delay )**2 / TWO / sigma2**2 ) &
         / sqrt( TWO * PI ) / sigma2
    r1 = ONE / ( ONE + P%fluence_ratio )
    r2 = P%fluence_ratio / ( ONE + P%fluence_ratio )
    f = f1 * r1 + f2 * r2
  end if
end function pulse_envelope


!============================================================================
! RK4 copied from Nina's code
!     modified to accept S and P
!============================================================================
      subroutine rk4(neq,n,t0,tf,work,S,P)
      implicit none
      integer :: neq
      real(long) :: n(neq)
      real(long) :: t0,tc,tf,dt
      real(long) :: work(neq,5)
      type(Process_data), intent(in) :: S
      type(Param), intent(in) :: P
      integer :: i

      tc = 0.5d0*(t0+tf)
      dt = tf - t0

      call REQ(t0,n,work(:,1),S,P)

      do i = 1, neq
        work(i,5) = n(i) + 0.5d0*dt*work(i,1)
      end do

      call REQ(tc,work(:,5),work(:,2),S,P)

      do i = 1, neq
        work(i,5) = n(i) + 0.5d0*dt*work(i,2)
      end do

      call REQ(tc,work(:,5),work(:,3),S,P)

      do i = 1, neq
        work(i,5) = n(i) + dt*work(i,3)
      end do

      call REQ(tf,work(:,5),work(:,4),S,P)

      do i = 1, neq
        work(i,5) = work(i,1) + 2.d0*work(i,2) + 2.d0*work(i,3) + work(i,4)
        n(i) = n(i) + work(i,5)*dt/6.d0
      end do 

!      t0 = tf

      return 
      end subroutine
!============================================================================

end program xatom_req
