!----------------------------------------------------------------------------
! xatom_shakeoff.f90
!----------------------------------------------------------------------------
! Copyright (C) 2011-2019 Deutsches Elektronen-Synchrotron DESY,
!                         a research centre of the Helmholtz Association.
! Authors: 2011-2019 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Shake-off calculations between initial and final configurations
!   written by Son, Sang-Kil in Jan. 2011 at DESY
!----------------------------------------------------------------------------
program xatom_shakeoff
  use constants
  use stdio
  use elec_config
  use read_param
  use gps
  use coord_system
  use wavefunction
  use scf
  use eigen_LAPACK
  implicit none
  type(Argument) :: ARG
  type(Param) :: P
  type(Grid) :: r
  type(Coordinate) :: G
  type(WF_data) :: F1, F2
  integer :: i, j, k, l, occ_n, N, N_orb
  integer :: ik, ij, lk, lj, occ_n_init
  integer :: ip, jp, kp
  real(long) :: value, energy, shakeoff_E
  real(long) :: S, p0, p1, p2
  real(long), allocatable :: prob(:)
  real(long) :: threshold       ! threshold for absolute A and F rates
  real(long) :: p_threshold     ! threshold for percentage of SO probability
  character(len=1) :: tag
  character(len=100) :: string1, string2, string3
  character(len=8192) :: config_str
  integer :: OUTPUT
  !
  ! start your program with time checking!
  !
  call init_time()
  !
  ! set up initial values
  !
  P%N = 200
  P%L1 = 1.0_long
  config_str = ""
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "HFS", yes_wrong_option = .false. )
  call print_parameters( P, "shake-off branching ratios" )
  call get_input( ARG )         ! read additional parameters
  call parse_option( ARG, "final_config", config_str )
  call parse_option( ARG, "config_final", config_str )
  call parse_option( ARG, "final",        config_str )
  if ( config_str == "" ) &
    call error_message( "A final_config should be specified." )
  !
  ! build the Legendre grid
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  !
  ! calculate the initial state: F1
  ! the initial configuration is given by P, i.e., using "-config"
  !
  F1 = create_WF( P, G )
  string1 = config_to_string( F1%config )
  write(P%OUTPUT, "('# Non-rel calc of the initial config: ',a)") trim(string1)
  write(P%OUTPUT, "('# ',a)") repeat( "=", 78 )
  call SCF_procedure( P, G, F1 )
  write(P%OUTPUT, "('#')")
  !
  ! calculate the final state: F2
  ! the final configuration is given by "-final_config"
  !
  N_orb = find_N_subshell( P%N_elec, config_str )
  if ( P%N_orb /= N_orb ) call error_message( &
    "The numbers of subshells in -config and -final_config are not matched." )
  call parse_config_string_to_P( P, config_str )
  F2 = create_WF( P, G )
  string2 = config_to_string( F2%config )
  write(P%OUTPUT, "('# Non-rel calc. of the final config: ',a)") trim(string2)
  write(P%OUTPUT, "('# ',a)") repeat( "=", 78 )
  call SCF_procedure( P, G, F2 )
  write(P%OUTPUT, "('#')")
  !
  ! calculate and print out branching ratios
  !
  OUTPUT = P%OUTPUT
  if ( P%yes_xmdyn ) OUTPUT = STDOUT
  allocate( prob(F2%N_subshell) )
  write(P%OUTPUT, "('# Shake-off branching ratio ',a)") &
    "(only single shake-off considered)"
  if ( P%yes_relativity ) then
    string1 = config_to_string( F1%config_rel )
    string2 = config_to_string( F2%config_rel )
  else
    string1 = config_to_string( F1%config )
    string2 = config_to_string( F2%config )
  end if
  write(P%OUTPUT, "('# - initial config:                  ',a)") trim( string1 )
  write(P%OUTPUT, "('# - final config (before shake-off): ',a)") trim( string2 )
  write(P%OUTPUT, "('# ',a)") repeat( "=", 78 )
  write(P%OUTPUT, "('# ',a)") "orbital index, branching ratio, " // &
                              "and final config (after shake-off)"
  p0 = ONE
  do k = 1, F2%N_subshell
    if ( P%yes_relativity ) then
      occ_n = F2%config_rel(k,4)        ! occupation number
      i = F2%config_rel(k,5)            ! index for non-relativistic orbital
      F2%config_rel(k,4) = F2%config_rel(k,4) - 1       ! take out one elec.
      string3 = config_to_string( F2%config_rel )       ! get a string
      F2%config_rel(k,4) = F2%config_rel(k,4) + 1       ! put it back
    else
      occ_n = F2%config(k,3)            ! occupation number
      i = k                             ! index for non-rel. orbital == k
      F2%config(k,3) = F2%config(k,3) - 1               ! take out one elec.
      string3 = config_to_string( F2%config )           ! get a string
      F2%config(k,3) = F2%config(k,3) + 1               ! put it back
    end if
    if ( occ_n == 0 ) cycle             ! no electron in this subshell
    S = radial_integral( G, F1%u(:,i) * F2%u(:,i) )     ! overlap b/w F1 and F2
    prob(k) = ONE - S**2
    p1 = prob(k) * occ_n                ! for single shake-off
    p0 = p0 - p1
    write(OUTPUT, "(i6,es13.5,'  # ',a)") k, p1, trim(string3)
  end do
  write(P%OUTPUT, "('# ',a4,es13.5,'  # ',a,' (no shake-off)')") &
    "|SO|", p0, trim(string2)
!  !
!  ! think about shake-up
!  !
!  write(P%OUTPUT, "()")
!  do j = 1, F1%N_subshell       ! scan F1
!    if ( P%yes_relativity ) then
!      occ_n_init = F1%config_rel(j,4)
!      ij = F1%config_rel(j,5)
!      lj = F1%config_rel(j,2)
!    else
!      occ_n_init = F1%config(j,3)
!      ij = j
!      lj = F1%config(j,2)
!    end if
!    if ( occ_n_init == 0 ) cycle
!    do k = 1, F2%N_subshell     ! scan F2
!      if ( P%yes_relativity ) then
!        occ_n = F2%config_rel(k,4)
!        ik = F2%config_rel(k,5)
!        lk = F2%config_rel(k,2)
!      else
!        occ_n = F2%config(k,3)
!        ik = k
!        lk = F2%config(k,2)
!      end if
!      if ( lj /= lk ) cycle
!      S = radial_integral( G, F1%u(:,ij) * F2%u(:,ik) ) ! overlap b/w F1 and F2
!      p1 = S**2
!      write(P%OUTPUT, "(2i6,es13.5,'  # ',2a5)") j, k, p1, &
!        WF_name(F1,j), WF_name(F2,k)
!    end do
!  end do
  call footnote( P%OUTPUT )
end program xatom_shakeoff
