!----------------------------------------------------------------------------
! xatom_shakeoff_dat.f90
!----------------------------------------------------------------------------
! Copyright (C) 2021 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2021 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Shake-off calculations on the top of the process data file
!   written by Son, Sang-Kil in Jan. 2011 at DESY
!----------------------------------------------------------------------------
program xatom_shakeoff_dat
  use constants
  use stdio
  use elec_config
  use read_param
  use gps
  use coord_system
  use wavefunction
  use scf
  use eigen_LAPACK
  implicit none
  !
  ! Grid: defined in gps.f90
  ! Coordinate: defined in coord_r.f90
  !   grid data for the radial part, r
  !   values in real space and computational space
  !   first and second derivative matrix
  !
  type(Grid) :: r
  type(Coordinate) :: G
  !
  ! Param: defined in readin.f90
  !   parameters to be used in calculation
  !   such as # of grids, DFT type, SCF parameters, etc
  !
  type(Param) :: P
  !
  ! WF_data: defined in wavefunction.f90
  !   varibles for wavefunctions, orbital energies, and potential
  !
  type(WF_data) :: F1, F2
  !
  ! misc. variables
  !
  integer, allocatable :: config(:,:)   ! electronic configuration
  integer :: OUTPUT
  integer :: i, j, k, l, occ_n, N
  integer :: ip, jp, kp
  real(long) :: value, energy, shakeoff_E
  real(long) :: S, p0, p1, p2
  real(long), allocatable :: prob(:)
  real(long) :: threshold       ! threshold for absolute A and F rates
  real(long) :: p_threshold     ! threshold for percentage of SO probability
  character(len=1) :: tag
  character(len=100) :: string1, string2
  integer :: INFO

  !
  ! start your program with time checking!
  !
  call init_time()
  !
  ! set up initial values
  !
  P%system = ""
  P%N = 200
  P%L1 = 1.0_long
  !
  ! read data from STDIN and/or input file
  !
  call read_parameters( P, "HFS" )
  P%yes_silent = .true.         ! force not to print out during SCF procedure
  threshold = P%print_threshold
  !p_threshold = 0.001           ! 0.1 %
  p_threshold = ZERO
  !
  ! the new data file is always printed in STDOUT
  ! other useful information is printed in OUTPUT
  ! if OUTPUT is not specified, print the output in fort.100
  !
  if ( P%OUTPUT == STDOUT ) P%OUTPUT = 100
  OUTPUT = P%OUTPUT
  call print_parameters( P, "shake-off branching ratios" )
  !
  ! build the Legendre grid
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  !
  ! build WF_data
  !
  F1 = create_WF( P, G )
  F2 = create_WF( P, G )
  !
  ! read the process data file and iterate processes
  !
  allocate( prob(P%N_orb) )
  allocate( config(P%N_orb,3) )
  N = product( P%init_config(:,3) + 1 )
  open( P%IO_DATA, file=P%file_data, action="read", form="formatted" )
  ip = 0
  write(OUTPUT, "('# Shake-off probabilities (%)')")
  write(OUTPUT, "('#   sudden approx: P_nl = [ 1 - |<nl,C1|nl,C2>|^2 ]*#occ')")
  write(OUTPUT, "('#')")
  do k = 1, F2%N_orb
    write(OUTPUT, "('    P',a2)", advance="no") WF_name( F2, k )
  end do
  do k = 1, F2%N_orb
    do kp = k, F2%N_orb
      write(OUTPUT, "(' P',a2,'-',a2)", advance="no") &
        WF_name( F2, k ), WF_name( F2, kp )
    end do
  end do
  write(OUTPUT, "(a7)") "P0"
  do
    i = 0;  j = 0;  value = ZERO
    read( P%IO_DATA, *, iostat=INFO ) tag, i, j, value, energy
    if ( INFO < 0 ) exit                                ! exit when it ends
    if ( tag == "#" ) cycle                             ! skip comments
    if ( i == 0 .or. j == 0 .or. value == ZERO ) cycle  ! skip unmeaningful
    if ( i > N .or. j > N ) &
      write(STDERR, *) "Index should be less than ", N, ": ", tag, i, j
    if ( i < 0 .or. j < 0 ) &
      write(STDERR, *) "Index should be greater than 0: ", tag, i, j
    !
    ! skip if it's too small values
    !
!
    if ( ( tag == "F" .or. tag == "A" ) .and. value <= 1.0E-4_long ) then
      call print_one( tag, i, j, value, energy, threshold, 0 )
      cycle
    end if
!
    !
    ! find out elec. configurations from indexes (i and j)
    !   index i -> F1 (before transition)
    !   index j -> F2 (after transition)
    !
    call index_to_config( i, F1%config )
    call index_to_config( j, F2%config )
    !
    ! calculate w.f. for each configuration
    !
    if ( i /= ip ) call SCF_procedure( P, G, F1 )
    call SCF_procedure( P, G, F2 )  ! get the potential for this ion
    !
    ! let's detach another electron from F2 (shake-off)
    ! print out single shake-off
    !
    p0 = ONE
    do k = 1, F2%N_orb
      shakeoff_E = energy - ( -F2%e_orb(k) * AU2EV )
      occ_n = F2%config(k,3)    ! occupation number
      if ( occ_n == 0 ) then    ! no electron in this subshell
        write(OUTPUT, "(7x)", advance="no")
        cycle
      end if
      S = radial_integral( G, F1%u(:,k) * F2%u(:,k) )   ! overlap b/w F1 and F2
      prob(k) = ONE - S**2
      p1 = prob(k) * occ_n      ! for single shake-off
      write(OUTPUT, "(f7.2)", advance="no") p1 * 100.0_long
      if ( p1 >= p_threshold .and. shakeoff_E >= ZERO ) then
        p0 = p0 - p1
        config(:,:) = F2%config(:,:)
        config(k,3) = config(k,3) - 1
        call config_to_index( config, jp )
        call print_one( tag, i, jp, value*p1, shakeoff_E, threshold, 1 )
      end if
    end do
    !
    ! print out double shake-off
    !
!    do k = 1, F2%N_orb
!      occ_n = F2%config(k,3)    ! occupation number
!      if ( occ_n <= 1 ) then    ! double shake-off from 1 electron is no way.
!        do kp = k, F2%N_orb; write(OUTPUT, "(7x)", advance="no"); end do
!        cycle
!      end if
!      do kp = k, F2%N_orb
!        shakeoff_E = energy - ( -F2%e_orb(k) ) - ( -F2%e_orb(kp) )
!        p2 = prob(k) * prob(kp)
!        if ( k == kp ) then
!          p2 = p2 * dble( occ_n * (occ_n-1) / 2 )
!        else
!          p2 = p2 * occ_n * F2%config(kp,3)
!        end if
!        write(OUTPUT, "(f7.2)", advance="no") p2 * 100.0_long
!        if ( p2 >= p_threshold .and. shakeoff_E > ZERO ) then
!          p0 = p0 - p2
!          config(:,:) = F2%config(:,:)
!          config(k,3) = config(k,3) - 1
!          config(kp,3) = config(kp,3) - 1
!          call config_to_index( config, jp )
!          call print_one( tag, i, jp, value*p2, shakeoff_E, threshold, 2 )
!        end if
!      end do
!    end do
    !
    ! print out P0 (probability of no shake-off)
    !
    string1 = config_to_string( F1%config )
    string2 = config_to_string( F2%config )
    write(OUTPUT, "(f7.2)", advance="no") p0 * 100.0_long
    write(OUTPUT, "(' : ',a,x,a,' -> ',a)") tag, trim(string1), trim(string2)
    call print_one( tag, i, j, value*p0, energy, threshold, 0 )
    !
    ! when i doesn't change, we don't need to solve w.f. of F1
    !
    ip = i
  end do
  call footnote( OUTPUT )
contains

subroutine print_one( tag, i, j, value, energy, threshold, SO_type )
  character(len=1), intent(in) :: tag
  integer, intent(in) :: i, j
  real(long), intent(in) :: value, energy, threshold
  integer :: SO_type
  character(len=5) :: string
  if ( ( tag == "A" .or. tag == "F" ) .and. ( value <= threshold ) ) return
  select case ( SO_type )
    case ( 0 ); string = ""
    case ( 1 ); string = " ; S1"
    case ( 2 ); string = " ; S2"
  end select
  write(STDOUT, "(a1,2i10,es13.5,x,f10.5,a)") tag, i, j, value, energy, string
end subroutine print_one

end program xatom_shakeoff_dat
