!----------------------------------------------------------------------------
! xatom_twostep.f90
!----------------------------------------------------------------------------
! Copyright (C) 2017 Deutsches Elektronen-Synchrotron DESY,
!                    a research centre of the Helmholtz Association.
! Authors: 2017 Sang-Kil Son <sangkil.son@cfel.de>
! License: GPL-3.0-or-later
!
! This file is part of XATOM.
!
! XATOM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! XATOM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE. See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! XATOM. If not, see <https://www.gnu.org/licenses/>.
!----------------------------------------------------------------------------
! Rate equation solver for plasma ionization dynamics
! based on the two-step model
!   written by Son, Sang-Kil in Jan. 2017 at DESY
!----------------------------------------------------------------------------
program xatom_twostep
  use constants
  use stdio
  use read_param
  use gps
  use wavefunction
  use process
  use plasma
  use scf
  use pcs
  use decay
  use print_property
  use git
  use interpolate
  implicit none
  type(Argument) :: ARG
  type(Param) :: P
  type(Coordinate) :: G
  type(WF_data) :: F, F2        ! F: first step, F2: second step
  type(Process_data) :: S
  type(Grid) :: r
  integer :: i_time, N_t        ! time variables
  integer :: i, j, k, l
  real(long) :: t, t0, dt       ! time, initial time, and time step
  real(long) :: p_env           ! pulse envelope at t
  real(long) :: total_pcs       ! total cross section
  real(long) :: flux            ! photon flux
  real(long) :: rate            ! total rate at t = flux * total_pcs
  real(long) :: N_ph_abs        ! N of absorbed photons
  real(long) :: E_absorption    ! absorbed energy = (Del N_ph_abs) * photon E
  real(long) :: E_emission      ! emitted energy: calculated from second-step
  real(long) :: U, U0           ! total energy
  integer :: OUTPUT
  integer, parameter :: IO_SPECTRA = 22
  character(len=512) :: file_spectra, file_UT
  type(Linear) :: SP
  real(long) :: AU2MB
  logical :: yes_reabsorption   ! fluo. E is reabsorbed by the system or not
  AU2MB = AU2M**2 / 1.0e-22_long        ! 1 Mb = 10^-22 m^2
  !
  ! start your program with time checking!
  !
  call init_time()
  call read_parameters( P, "TWOSTEP" )
  !
  ! parameter setup for finite-temperature calculations
  !
  yes_reabsorption = .false.
  file_UT = trim(P%filename) // ".ut"
  call get_input( ARG )
  call parse_option( ARG, "relation", file_UT )
  call parse_option( ARG, "TU",       file_UT )
  call parse_option( ARG, "reabsorption", yes_reabsorption )
  !
  ! special initialization for this program
  !
  P%yes_average_atom = .true.   ! enforce an average-atom calculation
  if ( P%T_eV == ZERO ) P%T_eV = 0.0001_long    ! put a small value
  do j = 1, P%N_orb
    l = P%init_config(j,2)
    P%init_config(j,3) = 2 * ( 2 * l + 1 )      ! put max. occupation number
  end do
  call print_parameters( P, "direct rate equation solver for plasma, " // &
                            GITVERSION )
  OUTPUT = P%OUTPUT
  !
  ! set up a file for time-dependent fluorescence spectra
  ! FORMAT: [ t in fs ] [ fluo. energy ] [ probability at given t ] [ rate ]
  !
  file_spectra = trim(P%filename) // ".spectra"
  open( IO_SPECTRA, file=file_spectra, action="write", form="formatted" )
  !
  ! print out additional infomation
  !
  write(OUTPUT, "('# time-dependent spectra data: ',a)") trim(file_spectra)
  write(OUTPUT, "('# config_prob threshold for the 2nd step = ',es9.1)") &
    P%twostep_thres
  write(OUTPUT, "('#')")
  !
  ! build the Legendre grid and WF_data
  !
  r = generate_Grid( "LGL", P%N, ZERO, P%Rmax, P%L1 )
  G = generate_Coordinate( r )
  F  = create_WF( P, G )
  F2 = create_WF( P, G )        ! for the second step
  S = create_Process( P%init_config, P%N_proc_max )
  !
  ! run SCF calculation for the first time and prepare WF_continuum
  !
  call SCF_procedure( P, G, F )
  call print_average_atom( P, F, P%T_eV, OUTPUT )  ! grand-canonical analysis
  U0 = F%E0     ! set up total energy at initial T
  !
  ! prepare continuum states for pcs calculation
  !
  if ( P%continuum_rmax /= P%rmax ) then
    write(STDERR,"('# WARNING: P%continuum is forced to be equal to P%rmax')")
    P%continuum_rmax = P%rmax
  end if
  call extend_WF_continuum( F, P%OUTPUT, P%continuum_rmax )
  call calculate_photoionization_CS( P, F, S )  ! run for printing
  S%N_process = 0     ! nullify S
  !
  ! set P%OUTPUT to NULL
  !
  P%OUTPUT = P%IO_NULL
  !
  ! set up SP for the T(temperature)-U(energy) relation
  ! skip this set up if no time propagation is needed (P%N_time == 0 )
  !
  if ( P%N_time /= 0 ) SP = setup_UT_relation( P, G, F, file_UT )
!  U = U0
!  do i = 1, 100
!    U = U + ONE
!    P%T_eV = Linear_interpolate( SP, U * AU2EV )
!    write(99,*) U*AU2EV, P%T_eV
!  end do
!  stop
  !
  ! initialize variables before time-propagation
  !
  N_t = int( ( P%t_end_au - P%t_start_au ) / P%dt_au )
  if ( P%N_time == 0 ) N_t = 0  ! no time-propagation
  t0 = P%t_start_au             ! in a.u.
  dt = P%dt_au                  ! in a.u.
  N_ph_abs = ZERO
  U = U0
  if ( P%N_time /= 0 ) then     ! no time-propagation
    write(OUTPUT, "('#')")
    call print_divider( OUTPUT, "Time propagation" )
    write(OUTPUT, "('#   column 1: time(fs)')")
    write(OUTPUT, "('#   column 2: pulse envelope (normalized)')")
    write(OUTPUT, "('#   column 3: total photoionization cross section(Mb)')")
    write(OUTPUT, "('#   column 4: number of absorbed photons')")
    write(OUTPUT, "('#   column 5: Del U(eV), calculated by E_abs & E_emit')")
    write(OUTPUT, "('#   column 6: Del E0(eV), calculated by SCF for given T')")
    write(OUTPUT, "('#   column 7: temperature(eV)')")
    write(OUTPUT, "('#   column 8: charge')")
    !write(OUTPUT, "('#',a9,3a10,2a11,a10,a8)") "t(fs)", "pulse_env", &
    !  "total_pcs", "N_ph_abs", "Del_U(eV)", "Del_E0(eV)", "T(eV)", "charge"
  end if
  !============================================================================
  ! time-propagation: numerical integration of rate equations to obtain T(t)
  !   only N_ph_abs is solved based on the cross sections
  !   and T(t) is evaluated accordingly --> P%T_eV varies during time
  !============================================================================
  time_evolution: do i_time = 0, N_t
    t = t0 + dt * dble(i_time)
    p_env = pulse_envelope( P, t )
    flux = P%fluence * p_env
    !
    ! at every single t, perform SCF assuming that it's thermalized at T(t)
    !
    call SCF_procedure( P, G, F )       ! P%T_eV varies during time
    !
    ! print out the first-step information
    !
    call print_divider( OUTPUT, "", "-" )
    write(OUTPUT, "('#        ',a11,a5,100a10)") "T(eV)", "scf", &
      (/ ( trim( WF_name( F, j ) ), j=1, F%N_orb ) /), "pot_tail", "chem_pot"
    write(OUTPUT, "('# |STEP1|',f11.3,i5,100f10.2)") P%T_eV, F%N_iter, &
      F%e_orb(:) * AU2EV, F%e_s * AU2EV, F%chem_pot * AU2EV
    write(OUTPUT, "('# |STEP1-OCC|',12x,100f10.3)", advance="no") F%f_occ(:)
    write(OUTPUT, "('      charge=',sp,f7.2)") F%Q_avg
    !
    ! Now, perform the second step and calculate fluorescence processes
    ! temperature is given by P%T_eV
    !
    call perform_second_step( P, G, F, F2, S, flux, total_pcs, E_emission, &
                              IO_SPECTRA, OUTPUT )
    call print_divider( OUTPUT, "", "-" )
    !
    ! P%N_time == 0 is used for the two-step model calculaiton at a fixed T
    !
    if ( P%N_time == 0 ) exit time_evolution
    !
    ! print out time propagation of T(t)
    !
    write(OUTPUT, "(f10.3,3es10.3,2es11.3,es10.3,sp,f8.2)") &
      t * AU2SEC / 1.0e-15_long, p_env, total_pcs * AU2MB, N_ph_abs, &
      ( U - U0 ) * AU2EV, ( F%E0 - U0 ) * AU2EV, P%T_eV, F%Q_avg
    !
    ! CASE (1): calculate pcs based on averaged orbitals
    ! CASE (2): calculate pcs by averaging ensembles from the second-step calc.
    ! ---> We choose CASE (2)
    !
    !call calculate_photoionization_CS( P, F, S )
    !total_pcs = get_total_pcs( F, S )
    !
    ! calculate N of absored photons; this is the simplest rate equation ever!
    !
    rate = flux * total_pcs
    N_ph_abs = N_ph_abs + rate * dt     ! only single-photon abs. considered
    E_absorption = P%PE_start * ( rate * dt )
    !
    ! evaluate T(t) by using interpolation of the relation between T and E_tot
    !
    U = U + E_absorption        ! add absorbed E (in a.u.)
    U = U - E_emission          ! substract emitted E (a.u.) via fluorescence
    if ( yes_reabsorption ) U = U + E_emission  ! fluo. is reabsorbed in plasma
    P%T_eV = Linear_interpolate( SP, U * AU2EV )
    !
    ! nullify S
    !
    S%N_process = 0
  end do time_evolution
  !============================================================================
  ! post-time-evolution: further relaxation via Auger and fluorescence
  !                      is it really necessary???
  !============================================================================
  !
  ! print out the foot note
  !
  close( IO_SPECTRA )
  call footnote( OUTPUT )
contains

subroutine perform_second_step( P, G, F, F2, S, flux, total_pcs, E_emission, &
                                IO_SPECTRA, OUTPUT )
  type(Param), intent(inout) :: P       ! it will be changed for F2
  type(Coordinate), intent(inout) :: G
  type(WF_data), intent(inout) :: F, F2 ! F: first step, F2: second step
  type(Process_data), intent(inout) :: S
  real(long), intent(in) :: flux        ! photon flux
  real(long), intent(out) :: total_pcs  ! total photoionization cross section
  real(long), intent(out) :: E_emission ! emitted energy (in a.u.)
  integer, intent(in) :: IO_SPECTRA, OUTPUT
  integer :: i, j, k, l, m
  integer :: ip, jp                     ! index for S
  integer :: N_config
  integer :: bound_max_occ(F%N_orb), bound_occ(F%N_orb)
  real(long) :: bound_e_orb(F%N_orb)
  real(long) :: binom   ! defined in anglib.f90
  real(long) :: prob, prob_sum, exp_beta, beta
  real(long) :: pcs(F%N_orb)
  real(long) :: weight, value
  real(long), parameter :: AU2FS = AU2SEC / 1e-15_long
  character(len=3000) :: string
  !
  ! read out photoionization cross section (pcs) from S <== CASE (1)
  !
!  pcs(:) = ZERO
!  do jp = 1, S%N_process
!    if ( S%tag(jp) /= "P" .and. S%tag(jp) /= "V" ) &
!      call error_message( "in perform_second_step()" )
!    i = S%occ_down(1,jp)        ! i-th orbital is ionized
!    pcs(i) = S%value(jp)
!  end do
  !
  ! set up some variables
  !
  beta = ONE / ( P%T_eV * EV2AU ) ! remember that T_eV = T_e * k (J/K) * J2EV
  prob_sum = ZERO
  total_pcs = ZERO      ! in a.u.
  E_emission = ZERO     ! in a.u.
  !
  ! put plasma density and potential into F2, which are calculated from F
  ! and turn off the average-atom option
  !
  F2%rho_free(:) = F%rho_free(:)
  F2%pot(:) = F%pot(:)
  F2%yes_reuse_rho_free = .true.
  F2%yes_reuse_pot = .true.
  P%yes_average_atom = .false.  ! second step: regular HFS calculation
  !
  ! calculate the maximum # of all possible bound configurations
  ! here, we consider bound configurations specified in P%config(:,:)
  ! and remember that P%config(:,:) should be fully occupied
  !
  do j = 1, F%N_orb
    l = F%config(j,2)
    i = l * F%N + ( F%config(j,1) - l )
    bound_max_occ(j) = get_occ_num_max( F, j )
    bound_e_orb(j) = F%e_all(i)
  end do
  N_config = product( bound_max_occ(:) + 1 )    ! all possible bound config.
  !
  ! print the header for the second step
  !
  write(OUTPUT, "('#        ',a11,a5,100a10)") "config_prob", "scf", &
    (/ ( trim( WF_name( F, j ) ), j=1, F%N_orb ) /), "pot_tail", "config  "
  !
  ! scan over all possible bound configurations
  !
  do i = 1, N_config
    !
    ! construct all possible config. of bound electrons --> bound_occ(:)
    ! for given bound spatial orbitals
    ! it's copied from plasma.f90 --- MAGIC routine
    !
    k = i
    do j = F%N_orb, 1, -1
      m = mod( k-1, bound_max_occ(j)+1 )
      bound_occ(j) = bound_max_occ(j) - m
      k = ( k - m ) / ( bound_max_occ(j)+1 ) + 1
    end do
    !
    ! calculate a probability of finding this configuration
    ! binom(): defined in anglib.f90
    !
    prob = ONE
    do j = 1, F%N_orb
      exp_beta = exp( beta * ( bound_e_orb(j) - F%chem_pot ) )
      prob = prob * binom( bound_max_occ(j), bound_occ(j) ) &
             / ( ONE + exp_beta )**bound_occ(j) &
             * ( exp_beta / ( ONE+exp_beta ) )**(bound_max_occ(j)-bound_occ(j))
    end do
    if ( prob < P%twostep_thres ) cycle ! skip if prob is too small
    prob_sum = prob_sum + prob          ! sum up when prob >= 1E-4
    !
    ! set up F2 and run SCF
    !
    F2%config(:,3) = bound_occ(:)
    call SCF_procedure( P, G, F2 )
    string = config_to_string( F2%config )
    write(OUTPUT, "('# |STEP2|')", advance="no")
    if ( prob >= 1E-6_long) then
      write(OUTPUT, "(f11.6)", advance="no") prob
    else
      write(OUTPUT, "(es11.3)", advance="no") prob
    end if
    write(OUTPUT, "(i5,100f10.2)", advance="no") &
      F2%N_iter, F2%e_orb(:) * AU2EV, F2%e_s * AU2EV
    write(OUTPUT, "(2x,a)") trim(string)
    !
    ! calculate photoionization cross section
    !
    S%config(:,3) = F2%config(:,3)      ! set up configuration
    S%N_process = 0                     ! nullify S
    if ( P%continuum_rmax > ZERO ) then
      call extend_WF_continuum( F2, P%IO_NULL, P%continuum_rmax )
    else
      call extend_WF_continuum( F2, P%IO_NULL )
    end if
    call calculate_photoionization_CS( P, F2, S )
    pcs(:) = ZERO
    do ip = 1, S%N_process
      jp = S%occ_down(1,ip)             ! jp-th orbital is ionized
      pcs(jp) = S%value(ip)
    end do
    total_pcs = total_pcs + sum( pcs(:) ) * prob
    !
    ! calculate fluorescence: consider TWO cases
    ! (a) fluo. without any additional absorption <-- NO by thermalization
    ! (b) make an absorption and then fluorescence to that core hole
    ! remember that all rates and transition energies are stored in S
    !
    do j = 1, F2%N_orb
      !
      ! case (a): calculate fluo. only for a hole (in any) in j-th orbital
      !   weight: consider only probability of this configuration
      !
!      call calculate_fluorescence_rates( P, G, F2, S, j )  ! only for j-th orb.
!      weight = prob
!      do ip = jp, S%N_process
!        write(IO_SPECTRA, "(f10.3,i10,es13.5)", advance="no") &
!          t * AU2FS, S%E_eV(ip), S%value(ip) * weight * P%dt_au
!        write(IO_SPECTRA, "('  # ',a,' : ',a,2x,es13.5,f10.3)") &
!          trim(string), trim( string_Process( S, ip ) ), S%value(ip), P%T_eV
!      end do
!      jp = S%N_process + 1
      !
      ! case (b): create a hole in j-th orb. and then calculate fluo.
      !   weight: consider both prob. of this configuration (= prob)
      !           photoionization prob. at the current time (= flux * pcs)
      !
!      weight = prob * ( flux * pcs(j) )
      weight = prob * ( flux * pcs(j) ) * P%dt_au
      if ( weight == ZERO ) cycle       ! skip if pcs=0 (ionization forbidden)
      if ( F2%config(j,3) == 0 ) cycle  ! skip if no electron in this subshell
      F2%config(j,3) = F2%config(j,3) - 1     ! creating a hole in j-th orb.
      S%config(j,3)  = F2%config(j,3)
      S%N_process = 0                   ! nullify S
      call calculate_fluorescence_rates( P, G, F2, S, j ) ! only for j-th orb.
      do ip = 1, S%N_process
        !
        ! value   := w_{E,t_i} to be printed out in IO_SPECTRA
        !            P%dt_au is multiplied here, 
        !            so these printed values can be simply added up
        ! S%value := fluorescence rate for the given configuration
        ! weight  := config_prob * ( photoionization probability at given t)
        ! 
        value = S%value(ip) * weight * P%dt_au
        E_emission = E_emission + S%E_eV(ip) * EV2AU * value
        if ( value < 1E-99_long ) cycle ! skip it's too small
        write(IO_SPECTRA, "(f10.3,i10,es13.5)", advance="no") &
          t * AU2FS, S%E_eV(ip), value
        write(IO_SPECTRA, "('  # ',a,' : ',a,2x,es13.5,f10.3)") &
          trim(string), trim( string_Process( S, ip ) ), S%value(ip), P%T_eV
      end do
      F2%config(j,3) = F2%config(j,3) + 1     ! go back to the orig. config.
      S%config(j,3)  = F2%config(j,3)
    end do
  end do
  !
  ! print all fluorescence energies and rates from S
  !    S%tag(1...N_orb) = "P" or "V"
  !    S%tag(N_orb+1...) : should be all "F"
  !
!  write(OUTPUT, "('#   ',a)") "all processes:"
!  do k = 1, S%N_process
!    write(OUTPUT, "('#   ',10x,a,i10,es13.5)") &
!      S%tag(k), S%E_eV(k), S%value(k)
!  end do
  !
  ! go back to the original setup for the first step calculation
  !
  P%yes_average_atom = .true.
  !
  ! print the footer
  ! prob_sum is used to check whether showing of prob >= 1E-4 is enough or not
  !
  write(OUTPUT, "('#     ',a4,f10.6)") "sum=", prob_sum
end subroutine perform_second_step

function pulse_envelope( P, t ) result (f)
  real(long) :: f
  type(Param), intent(in) :: P
  real(long), intent(in) :: t
  real(long) :: sigma
  !
  ! normalized Gaussian distribution
  !   f = e^( -t^2 / 2 sigma^2 ) / ( sigma sqrt(2pi) )
  ! where sigma: standard deviation
  !       FWHM (Full Width at Half Maximum), tau = 2 sigma sqrt( 2 ln 2 )
  !
  if ( t < P%t_start_au .or. t > P%t_end_au ) then
    f = ZERO
  else if ( P%pulse_type == 3 ) then    ! Gaussian
    sigma = P%FWHM / TWO / sqrt( TWO * log(TWO) )
    f = exp( -t**2 / TWO / sigma**2 ) / sqrt( TWO * PI ) / sigma
  else if ( P%pulse_type == 6 ) then    ! flat-top, FWHM here is pulse duration
    f = ONE / P%FWHM
  end if
end function pulse_envelope

function get_total_pcs( F, S ) result (cs)
  type(WF_data), intent(in) :: F
  type(Process_data), intent(in) :: S
  integer :: i, j, N_max_occ
  real(long) :: cs
  real(long) :: ratio
  !
  ! calculate total photoionization cross section
  ! taking into account fractional occupation numbers
  !
  ! Remember that photoionization_CS() in pcs.f90 calculates PCS based on 
  ! occupation numbers of F%config(:,3).
  ! Therefore, fractional occupation numbers from the average-atom calculation
  ! should be taken info account.
  !
  cs = ZERO
  do j = 1, S%N_process
    if ( S%tag(j) == "P" .or. S%tag(j) == "V" ) then
      do i = 1, F%N_orb
        if ( i == S%occ_down(1,j) ) then
          N_max_occ = get_occ_num_max( F, i )
          cs = cs + S%value(j) * F%f_occ(i) / dble(N_max_occ)
        end if
      end do
    end if
  end do
end function get_total_pcs

function setup_UT_relation( P, G, F, filename ) result (SP)
  type(Linear) :: SP
  type(Param), intent(inout) :: P
  type(Coordinate), intent(inout) :: G
  type(WF_data), intent(inout) :: F
  character(len=*), intent(in) :: filename
  real(long) :: T0, dT, T_max, T_restore
  real(long), allocatable :: U(:), T(:)
  real(long) :: Temp
  integer :: i, N
  integer :: IO_FILE = 23, OUTPUT, status
  logical :: yes_exist
  character(len=1000000) :: line
  OUTPUT = P%OUTPUT
  T0 = ZERO             ! in eV
  dT = 20_long          ! in eV
  T_max = 1000_long     ! in eV
  T_restore = P%T_eV
  !
  ! run finite-T calculations for the T-U relation or load data from a file
  !
  write(OUTPUT, "('#')")
  inquire( file=filename, exist=yes_exist )
  if ( yes_exist ) then
    !
    ! if the file exists, then read out this file
    !
    write(OUTPUT, "('# Temperature(T) - Energy(U) relation loaded from:',a)") &
      trim(filename)
    open( IO_FILE, file=filename, action="read", form="formatted" )
    !
    ! determine N (# of rows)
    !
    N = 0
    do
      read( IO_FILE, "(a)", iostat=status ) line
      if ( status < 0 ) exit
      if ( len_trim( line ) == 0 .or. line(1:1) == "#" ) cycle
      N = N + 1
    end do
    allocate( U(N), T(N) )
    rewind( IO_FILE )
    !
    ! read data
    !
    i = 0
    do
      read( IO_FILE, "(a)", iostat=status ) line
      if ( status < 0 ) exit
      if ( len_trim( line ) == 0 .or. line(1:1) == "#" ) cycle
      i = i + 1
      read( line, * ) U(i), T(i)        ! their units are in eV
    end do
  else
    !
    ! if there is no file, then generate one
    !
    open( IO_FILE, file=filename, action="write", form="formatted" )
    N = int( ( T_max - T0 ) / dT + 0.5_long ) + 1
    allocate( U(N), T(N) )
    write(OUTPUT, "('# Temperature(T) - Energy(U) relation')")
    write(OUTPUT, "('#',2a11)") "T(eV)", "U(eV)"
    i = 1
    Temp = T0
    T(i) = Temp
    U(i) = F%E0 * AU2EV                 ! assuming P%T_eV calc. is already done
    write( IO_FILE, * ) U(i), T(i)      ! and it's printed out as i=1
    FT_loop: do
      Temp = Temp + dT
      if ( Temp > T_max ) exit FT_loop
      P%T_eV = Temp
      if ( P%T_eV == ZERO ) P%T_eV = 0.001_long ! put a small value
      call SCF_procedure( P, G, F )
      i = i + 1
      T(i) = Temp               ! unit of T: eV
      U(i) = F%E0 * AU2EV       ! unit of U: eV
      ! THIS IS EXPERIMENTAL ... remove the negative heat capacity ...
      if ( U(i) < U(i-1) ) U(i) = U(1) + 0.05_long * T(i)**2
      ! END
      write(OUTPUT, "('#',2f11.3,i5,100f10.2)") T(i), U(i), &
        F%N_iter, F%e_orb(:) * AU2EV, F%e_s * AU2EV, F%chem_pot * AU2EV
      write( IO_FILE, * ) U(i), T(i)
    end do FT_loop
    P%T_eV = T_restore          ! restore the initial temperature
  end if
  !
  ! prepare a linear interpolation, 
  ! instead of Spline that has spurious oscillations at a kink position
  !
  SP = Linear_create( U(:), T(:) )
end function setup_UT_relation

!function restore_potential( T_eV_step, pot_step, T_eV ) result (pot)
!  real(long), intent(in) :: T_eV_step(:), pot_step(:,:), T_eV
!  real(long) :: pot(size(pot_step,1))
!  integer :: i
!  if ( T_eV < T_eV_step(1) ) then
!    pot(:) = pot_step(:,1)
!  else
!    do i = size(T_eV_step), 1, -1
!      if ( T_eV_step(i) <= T_eV ) exit
!    end do
!    pot(:) = pot_step(:,i)
!  end if
!end function restore_potential

end program xatom_twostep
