///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>



/* #################################################### */

/* This is a workaround to remove never occupied 4f from XATOM econf string */
/* so that XATOM calculation is faster ; originally used for Xe */
/* This way the active space is limited by excluding 4f */
#define WORKAROUND_XE_NO_4f   

// ########################################################
#define MODULE_NAME               XATOM

#define XATOM_ARG_MAXQ            20
#define XATOM_ARG_NUMQ            100
#define XATOM_ARG_MAXE            15000
#define XATOM_ARG_DE              10

#define XATOM_EXE                 xatom
#define XATOM_ATM_FLAGS           -decay
#define XATOM_PCS_FLAGS           -pcs
#define XATOM_SCD_FLAGS           -scd -Q XATOM_ARG_MAXQ -N_Q XATOM_ARG_NUMQ
#define XATOM_IXS_FLAGS           -ixs
#define XATOM_MAD_FLAGS           -mad -PE 0-XATOM_ARG_MAXE -dE XATOM_ARG_DE

/* #################################################### */

#ifdef MODULE_MAIN
    #include "common.h"
#else
    #include "common.h"
    #include "globalconsts.h"
    #include "endltabs.h"
    #include "xmdyn_io.h"
#endif

#include "Xatom.h"


// ########################################################
// ########################################################

struct g_struct_xatom_tree
{
    struct g_struct_xatom_tree *prnt ;
    struct g_struct_xatom_tree **chldrn ;
    int num_chldrn ;

    long long int id ;
    long long int calls ;
    long long int type_id ;

    int orbit ;
    int occ ;

    struct g_struct_xatom_data *data ;

    char *id_str ;

    int numQ ;
    double *Q ;
} ;


struct g_struct_tree_prop
{
    int maxdepth ;
//    int num_conf ;
    int max_type_id ;
} ;



static struct g_struct_xatom_tree    *g_xatom_tree = NULL ;
static struct g_struct_xatom_tree    *g_xatom_shadow_tree = NULL ;

static struct g_struct_tree_prop     g_tree_prop ;

char s_dbase_dir_load[LSTRLTH], s_dbase_dir_save[LSTRLTH] , s_xatom_opts[LSTRLTH] ;


// ########################################################
// ########################################################


#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
    //return a_return ;
}




//############################## FUNCTION #################################

int convert_nlj2endl( int a_n , int a_l , double a_j )
{

    int id , nn , ll , jj2 ;
    int maxid = 10000 ;
    id = 0 ;

    for( nn = 1 ; nn <= a_n ; nn++ )
    {
        if ( maxid < id ) { return -1 ; }
        if ( 1 < nn ) { id++ ; }
        for ( ll = 0 ; ll < nn ; ll++ )
        {
            if ( 0 < ll ) { id++ ; }
            for ( jj2 = abs( 2*ll - 1 ) ; jj2 <= ( 2*ll + 1 ) ; jj2+=2 )
            {
                id++ ;
                if ( ( a_n == nn ) && ( a_l == ll ) && ( a_j == jj2/2.0 ) )
                { return id ; }
            }
        }
    }


  //fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  MODULE_ERROR( "You should not be here! id=1!" , -1 ) ;
  return 0 ;
}



//############################## FUNCTION #################################

int convert_endl2nlj( int a_id , int *a_n , int *a_l , double *a_j )
{

    int id , nn , ll , jj2 ;
    int maxid = 10000 ;
    id = 0 ;

    for( nn = 1 ; nn <= maxid ; nn++ )
    {
        if ( maxid < id ) { return -1 ; }
        if ( 1 < nn ) { id++ ; }
        for ( ll = 0 ; ll < nn ; ll++ )
        {
            if ( 0 < ll ) { id++ ; }
            for ( jj2 = abs( 2*ll - 1 ) ; jj2 <= ( 2*ll + 1 ) ; jj2+=2 )
            {
                id++ ;
                if ( id == a_id )
                {
                    *a_n = nn ;
                    *a_l = ll ;
                    *a_j = jj2 / 2.0 ;
                    return 0 ;
                }
            }
        }
    }

  return 0 ;
}




//############################## FUNCTION #################################

int convert_shellid_endl2xatom( int a_max_occ_shell , int *a_endl_conf , 
                                int **a_xatom_conf , int *a_indexes )
{

    int nn , ll , jj2 ;
    int maxid = 10000  , maxL ;
    char x_str[STRLTH] , l_str[STRLTH]  ;
    char l2char[STRLTH] = "spdf" ;
    int dim ;

    /* We allocate a bit (approx. 3 times more) memory than necessary */
    if ( NULL == *a_xatom_conf ) 
    {
        printf( "xatom_conf has been  allocated " ) ; printf("sdfasdfasfg\n") ;
        *a_xatom_conf = calloc( a_max_occ_shell , sizeof(int) ) ;
    }

    int id = 0 ;

    maxL = strlen( l2char ) ;
    dim = 0 ;

    for( nn = 1 ; id <= maxid ; nn++ )
    {
        if ( maxid < id ) { return -1 ; }
        if ( 1 < nn ) { id++ ; }
        for ( ll = 0 ; ll < nn ; ll++ )
        {
            if ( 0 < ll ) { id++ ; }
            for ( jj2 = abs( 2*ll - 1 ) ; jj2 <= ( 2*ll + 1 ) ; jj2+=2 )
            {
                id++ ;

                if ( jj2 == abs( 2*ll - 1 ) )
                {
                    dim++ ;
                    (*a_xatom_conf)[dim] = a_endl_conf[id] ;
                    if ( NULL != a_indexes ) { a_indexes[dim-1] = id ; } 
                }

                if ( id == a_max_occ_shell )
                {
                    (*a_xatom_conf)[0] = dim ;
                    return 0 ;
                }
            }
        }
    }

    return -1 ;
}



//############################## FUNCTION #################################

int convert_shellid_xatom2endl( int *a_max_occ_shell , int *a_xatom_conf , 
                                int **a_endl_conf , int *a_indexes )
{

    int nn , ll , jj2 ;
    int maxid = 10000  , maxL ;
    char x_str[STRLTH] , l_str[STRLTH]  ;
    char l2char[STRLTH] = "spdf" ;
    int dim ;

    /* We allocate a bit more memory than necessary */
    if ( NULL == *a_endl_conf )
    {
        //debug_print( "endl_conf has been allocated" ) ;
        printf( "endl_conf has been allocated\n" ) ;
        *a_endl_conf = calloc( 3 * a_xatom_conf[0] , sizeof(int) ) ;
    }

    int id = 0 ;

    maxL = strlen( l2char ) ;
    dim = 0 ;

    for( nn = 1 ; id <= maxid ; nn++ )
    {
        if ( maxid < id ) { return -1 ; }
        if ( 1 < nn ) { id++ ; }
        for ( ll = 0 ; ll < nn ; ll++ )
        {
            if ( 0 < ll ) { id++ ; }
            for ( jj2 = abs( 2*ll - 1 ) ; jj2 <= ( 2*ll + 1 ) ; jj2+=2 )
            {
                id++ ;

                if ( jj2 == abs( 2*ll - 1 ) )
                {
                    dim++ ;
                    (*a_endl_conf)[id] = a_xatom_conf[dim] ;
                    if ( NULL != a_indexes ) { a_indexes[dim-1] = id ; } 
                }

                if ( dim == a_xatom_conf[0] )
                {   
                    if ( NULL != a_max_occ_shell ) { *a_max_occ_shell = id ; }
                    return 0 ;
                }
            }
        }
    }

    return -1 ;
}



//############################## FUNCTION #################################

int get_orbit_max_occupancy( int a_orbit )
    /* TODO: change the table to formula */
{
  switch ( a_orbit ) {
   // 1s
   case 1:    return(2);
   // 2s     
   case 3:    return(2);
   // 2p     
   case 5:    return(6);
   case 6:    return(0);
   // 3s     
   case 8:    return(2);
   // 3p     
   case 10:   return(6);
   case 11:   return(0);
   // 3d     
   case 13:   return(10);
   case 14:   return(0);
   // 4s     
   case 16:   return(2);
   // 4p     
   case 18:   return(6);
   case 19:   return(0);
   // 4d     
   case 21:   return(10);
   case 22:   return(0);
   // 4f     
   case 24:   return(14);
   case 25:   return(0);
   // 5s     
   case 27:   return(2);
   // 5p     
   case 29:   return(6);
   case 30:   return(0);
   // 5d     
   case 32:   return(10);
   case 33:   return(0);
   // 5f     
   case 35:   return(14);
   case 36:   return(0);
   // 5g     
   case 38:   return(18);
   case 39:   return(0);
   // 6s     
   case 41:   return(2);
   // 6p     
   case 43:   return(6);
   case 44:   return(0);
   // 6d     
   case 46:   return(10);
   case 47:   return(0);
   // 6f     
   case 49:   return(14);
   case 50:   return(0);
   // 6g     
   case 52:   return(18);
   case 53:   return(0);
   // 6h     
   case 55:   return(22);
   case 56:   return(0);
   // 7s     
   case 58:   return(2);
   // 7p     
   case 60:   return(6);
   case 61:   return(0);
   // otherwise
   default:   return(0);
  }
}




//############################## FUNCTION #################################

int get_highest_occupied_orbital( int *a_config , int a_max_orbit ) 
{
    int ii ;
    for ( ii = a_max_orbit ; ( ( ii > 0 ) && ( 0 == a_config[ii] ) ) ; --ii )
        ;
    return( ii ) ;    
} 



//############################## FUNCTION #################################

int print_econfig( int *a_config , int a_max_orbit )
{
    int ii ; 
    printf( "%3d:" , a_max_orbit ) ;
    for ( ii=0 ; ii <= a_max_orbit ; ii++ ) 
    {
        printf( "%d " , a_config[ii] ) ;
    }
    printf( "\n" ) ;
    return 0 ;
}





//############################## FUNCTION #################################

int do_econf_diff_list( int *a_conf1 , int *a_conf2 , int a_max_occ_shell , 
                        int *a_result , int a_exp_num_diff )
{
    int ii , jj , cc , xdiff ;
    cc = 0 ;
    for ( ii = 0 ; ii <= a_max_occ_shell ; ++ii ) 
    {
        xdiff = abs( a_conf1[ii] - a_conf2[ii] ) ;
        for ( jj = 0 ; jj < xdiff ; jj++ )
        {
            a_result[cc] = ii ;
            cc++ ;
        }
    }
    if ( cc != a_exp_num_diff ) 
    {
        fprintf( stderr , "\n!!! Expt: %d, curr:%d .  " , a_exp_num_diff , cc ) ;
        MODULE_ERROR( "" , -1 ) ;
        return -1 ;
    }

    return 0 ;
}



//############################## FUNCTION #################################

int get_xatom_conf_str( int a_max_occ_shell , int *a_config , char *a_str , char type )
{
    /* C -> calc , F -> file , L -> log  */

    int nn , ll , jj2 ;
    int maxid = 10000  , maxL ;
    char x_str[STRLTH+64] , l_str[STRLTH]  ;
    char l2char[STRLTH] = "spdfghijkl" ;

    if ( ( 'C' != type ) && ( 'L' != type ) &&( 'F' != type ) ) { return( MODULE_ERROR("Wrong type in get_xatom_conf_str",-1) ) ; } ;

    if ( 'L' == type ) { type = 'F' ; } 

    strcpy( a_str , "" ) ;

    int id = 0 ;

    maxL = strlen( l2char ) ;

    for( nn = 1 ; id <= maxid ; nn++ )
    {
        if ( maxid < id ) { return -1 ; }
        if ( 1 < nn ) { id++ ; }
        for ( ll = 0 ; ll < nn ; ll++ )
        {
            if ( 0 < ll ) { id++ ; }
            for ( jj2 = abs( 2*ll - 1 ) ; jj2 <= ( 2*ll + 1 ) ; jj2+=2 )
            {
                id++ ;

                if ( jj2 == abs( 2*ll - 1 ) )
                {
                    strncpy( l_str , l2char + ll , 1 ) ; l_str[1] = '\0' ;
                    sprintf( x_str , "%d%s%d" , nn , l_str , a_config[id] ) ;
#ifdef WORKAROUND_XE_NO_4f
                    if ( 24 != id )  //* NOTE: 24 is the EADL designator of 4f5/2 (4f7/2 is merged into this) *//
#endif
                    {
                        strcat( strcat( a_str , x_str ) , "_" ) ;
                    }
                }

                if ( id == a_max_occ_shell )
                {
                    if ( 0 == strlen( a_str ) ) { strcpy( a_str , "1s0_" ) ; } 
                    a_str[ strlen( a_str ) - 1 ] = '\0' ; 
                    return 0 ;
                }
            }
        }
    }

    MODULE_ERROR( "You should not be here!" , -1 ) ;
    return -1 ;
}



// ##########################################################################
// ##########################################################################
// ##########################################################################



//############################## FUNCTION #################################

int alloc_xatom_data_content( struct g_struct_xatom_data *atomic_econfig  , 
                              char *conf_str , char *what , int num )
{
    if ( NULL != conf_str )
    {
        atomic_econfig->str = calloc( strlen( conf_str ) + 1 , sizeof(char) ) ;
        strcpy( atomic_econfig->str , conf_str ) ;
    }

    if ( ( strcmp( what , "max_occ_shell" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->NE_orbit = calloc( num+1 , sizeof(int) ) ;

        atomic_econfig->r = calloc( num+1 , sizeof(double) ) ;

        atomic_econfig->B = calloc( num+1 , sizeof(double) ) ;
        atomic_econfig->K = calloc( num+1 , sizeof(double) ) ;
        atomic_econfig->V = calloc( num+1 , sizeof(double) ) ;

        atomic_econfig->ph_E = calloc( num+1 , sizeof(double) ) ;
        atomic_econfig->ph_sigma = calloc( num+1 , sizeof(double) ) ;

        atomic_econfig->dE_recomb = calloc( num+1 , sizeof(int) ) ;

        atomic_econfig->num_fl = 0 ;
        atomic_econfig->num_au = 0 ;
        atomic_econfig->num_fij = 0 ;
        atomic_econfig->num_sample = 0 ;

        atomic_econfig->fprime = 0 ;
        atomic_econfig->fdprime = 0 ;
        atomic_econfig->fdprime_pcs = 0 ;

    }

    if ( ( strcmp( what , "fl" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->fl_orbit = calloc( 2 * num , sizeof(int) ) ;
        atomic_econfig->fl_w = calloc( num , sizeof(double) ) ;
        atomic_econfig->fl_E = calloc( num , sizeof(double) ) ;
    }

    if ( ( strcmp( what , "au" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->au_orbit = calloc( 3 * num , sizeof(int) ) ;
        atomic_econfig->au_w = calloc( num , sizeof(double) ) ;
        atomic_econfig->au_E = calloc( num , sizeof(double) ) ;
    }

    if ( ( strcmp( what , "fij" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->fij = calloc( num , sizeof(int) ) ;
    }

    if ( ( strcmp( what , "scd" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->f0  = calloc( num , sizeof(double) ) ;
        if ( NULL == g_xatom_tree[atomic_econfig->Z].Q ) 
        { 
            g_xatom_tree[atomic_econfig->Z].Q = calloc( num , sizeof(double) ) ; 
        }
    }

    if ( ( strcmp( what , "ixs" ) == 0 ) && ( num > 0 ) )
    {
        atomic_econfig->ixs  = calloc( num , sizeof(double) ) ;
        if ( NULL == g_xatom_tree[atomic_econfig->Z].Q ) 
        { 
            g_xatom_tree[atomic_econfig->Z].Q = calloc( num , sizeof(double) ) ; 
        }
    }


    return 0 ;
}




//############################## FUNCTION #################################

int cp_xatom_data( struct g_struct_xatom_data *a_dest , struct g_struct_xatom_data *a_source )
{
    int ii ;

    a_dest->max_orbit = a_source->max_orbit ;
    a_dest->NE = a_source->NE ;
    a_dest->E = a_source->E ;

    a_dest->NE_orbit = calloc( (a_dest->max_orbit) + 1 , sizeof(int) ) ;

    a_dest->r = calloc( (a_dest->max_orbit) + 1 , sizeof(int) ) ;

    a_dest->B = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;
    a_dest->K = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;
    a_dest->V = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;
    a_dest->ph_E = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;
    a_dest->ph_sigma = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;
    a_dest->dE_recomb = calloc( (a_dest->max_orbit) + 1 , sizeof(double) ) ;

    for ( ii = 0 ; ii <= a_source->max_orbit ; ii++ )
    {
        a_dest->NE_orbit[ii] = a_source->NE_orbit[ii] ;
        a_dest->r[ii] = a_source->r[ii] ;
        a_dest->B[ii] = a_source->B[ii] ;
        a_dest->K[ii] = a_source->K[ii] ;
        a_dest->V[ii] = a_source->V[ii] ;
        a_dest->ph_E[ii] = a_source->ph_E[ii] ;
        a_dest->ph_sigma[ii] = a_source->ph_sigma[ii] ;
        a_dest->dE_recomb[ii] = a_source->dE_recomb[ii] ;
    }

    a_dest->num_fl = a_source->num_fl ;
    a_dest->fl_orbit = calloc( 2 * a_dest->num_fl , sizeof(int) ) ;
    a_dest->fl_w     = calloc( 1 * a_dest->num_fl , sizeof(double) ) ;
    a_dest->fl_E     = calloc( 1 * a_dest->num_fl , sizeof(double) ) ;

    for ( ii = 0 ; ii <= a_source->num_fl ; ii++ )
    {
        a_dest->fl_orbit[ii][0] = a_source->fl_orbit[ii][0] ;
        a_dest->fl_orbit[ii][1] = a_source->fl_orbit[ii][1] ;
        a_dest->fl_w[ii] = a_source->fl_w[ii] ;
        a_dest->fl_E[ii] = a_source->fl_E[ii] ;
    }

    a_dest->num_au = a_source->num_au ;
    a_dest->au_orbit = calloc( 3 * a_dest->num_au , sizeof(int) ) ;
    a_dest->au_w     = calloc( 1 * a_dest->num_au , sizeof(double) ) ;
    a_dest->au_E     = calloc( 1 * a_dest->num_au , sizeof(double) ) ;

    for ( ii = 0 ; ii <= a_source->num_au ; ii++ )
    {
        a_dest->au_orbit[ii][0] = a_source->au_orbit[ii][0] ;
        a_dest->au_orbit[ii][1] = a_source->au_orbit[ii][1] ;
        a_dest->au_orbit[ii][2] = a_source->au_orbit[ii][2] ;
        a_dest->au_w[ii] = a_source->au_w[ii] ;
        a_dest->au_E[ii] = a_source->au_E[ii] ;
    }


    a_dest->num_sample = a_source->num_sample ;
    a_dest->q_sample = calloc( a_dest->num_sample , sizeof(double) ) ;
    a_dest->f0  = calloc( a_dest->num_sample , sizeof(double) ) ;
    for ( ii = 0 ; ii <= a_source->num_sample ; ii++ )
    {
        a_dest->q_sample[ii] = a_source->q_sample[ii] ;
        a_dest->f0[ii] = a_source->f0[ii] ;
    }

    a_dest->num_fij = a_source->num_fij;
    a_dest->fij = calloc( a_dest->num_fij , sizeof(double) ) ;
    for ( ii = 0 ; ii <= a_source->num_fij ; ii++ )
    {
        a_dest->fij[ii] = a_source->fij[ii] ;
    }

    return 0 ;
}



//############################## FUNCTION #################################

int create_xatom_data(  struct g_struct_xatom_data *a_data )
{
    a_data = calloc( 1 , sizeof(  struct g_struct_xatom_data ) ) ;
    return 0 ;

}



//############################## FUNCTION #################################

int destroy_xatom_data( struct g_struct_xatom_data **a_dest )
{
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->NE_orbit ) ;

    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->r ) ;

    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->B ) ;
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->K ) ;
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->V ) ;
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->ph_E ) ;
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->ph_sigma ) ;
    if ( (*a_dest)->max_orbit > 0 ) free( (*a_dest)->dE_recomb ) ;

    if ( (*a_dest)->num_fl > 0 ) free( (*a_dest)->fl_orbit ) ;
    if ( (*a_dest)->num_fl > 0 ) free( (*a_dest)->fl_E ) ;
    if ( (*a_dest)->num_fl > 0 ) free( (*a_dest)->fl_w ) ;

    if ( (*a_dest)->num_au > 0 ) free( (*a_dest)->au_orbit ) ;
    if ( (*a_dest)->num_au > 0 ) free( (*a_dest)->au_E ) ;
    if ( (*a_dest)->num_au > 0 ) free( (*a_dest)->au_w ) ;

    if ( (*a_dest)->num_fij > 0 ) free( (*a_dest)->fij ) ;
    if ( (*a_dest)->num_sample > 0 ) free( (*a_dest)->q_sample ) ;
    if ( (*a_dest)->num_sample > 0 ) free( (*a_dest)->f0 ) ;

    free( *a_dest ) ;
    *a_dest = NULL ;

    return 0 ;
}




//############################## FUNCTION #################################

int xatom_tree_zero_newchld(  struct g_struct_xatom_tree *a_chld , 
                              struct g_struct_xatom_tree *a_prnt ,
                              int a_orbit ,
                              int a_occ )
{
    char x[128] ;
    a_chld->prnt = a_prnt ;
    a_chld->chldrn = NULL ;
    a_chld->num_chldrn = 0 ;

    a_chld->id = 0 ;
    a_chld->calls = 0 ;

    a_chld->orbit = a_orbit ;
    a_chld->occ = a_occ ;

    a_chld->data = NULL ;
    a_chld->id_str = malloc( 1024 * sizeof(char) ) ;
    strcpy( a_chld->id_str , a_chld->prnt->id_str ) ;

    sprintf( x , "%2d " , a_occ ) ;
    strcat( a_chld->id_str , x ) ;

  return 0 ;
}



//############################## FUNCTION #################################

int xatom_tree_create_newchld( struct g_struct_xatom_tree *sel_unit, int next_orbit )
{
    int  jj , num_deg ;

    num_deg = get_orbit_max_occupancy( next_orbit ) ;
    sel_unit->num_chldrn = num_deg +1 ;
    sel_unit->chldrn = malloc( ( 1 + num_deg ) * 
                               sizeof( struct g_struct_xatom_tree* ) ) ;
    for ( jj=0 ; jj < num_deg+1 ; jj++ ) 
    {
        sel_unit->chldrn[jj] = malloc( sizeof( struct g_struct_xatom_tree ) ) ;
        xatom_tree_zero_newchld( sel_unit->chldrn[jj] , sel_unit , 
                                 next_orbit , jj ) ;
    }

  return 0 ;
}



//############################## FUNCTION #################################

int xatom_tree_add( int a_Z , int a_num_orbit , int *a_config , struct g_struct_xatom_data **a_data_out )
{
    int ii , next_orbit  ;
    struct g_struct_xatom_tree  *sel_unit ;

    sel_unit = &(g_xatom_tree[a_Z]) ; 

    for ( ii = 1 ; ii <= a_num_orbit ; ii++ )
    {
        if ( NULL == sel_unit->chldrn )
        {
            next_orbit = sel_unit->orbit + 1 ;
            xatom_tree_create_newchld( sel_unit , next_orbit ) ;
        }
        sel_unit = sel_unit->chldrn[ a_config[ii] ] ;
    }

    if ( NULL == sel_unit->data )
    {
        sel_unit->type_id =  g_tree_prop.max_type_id + 1 ; g_tree_prop.max_type_id ++ ;
        sel_unit->id = 10000 * sel_unit->orbit + sel_unit->occ ;
        sel_unit->data = malloc( sizeof( struct g_struct_xatom_data ) ) ;
        sel_unit->data->parent = (void*)sel_unit ;
        sel_unit->data->NE_orbit = NULL ;
        sel_unit->data->NE = -1 ;     /* to indicate that it still has to be filled up !! */
        if ( g_tree_prop.maxdepth < sel_unit->orbit  )
        {
            g_tree_prop.maxdepth = sel_unit->orbit ; 
        }
        sel_unit->data->Z = a_Z ;
    }
    (sel_unit->calls)++ ;
    *a_data_out = sel_unit->data ; 

  return 0 ;
}




//############################## FUNCTION #################################

int xatom_tree_print_status(  struct g_struct_xatom_tree *sel_unit , int level  , char *id )
{
    int ii ;


    for ( ii = 0 ; ii < sel_unit->num_chldrn ; ii++ )
    {
        xatom_tree_print_status( sel_unit->chldrn[ii] , level + 1  , id ) ;
    }

    if ( NULL != sel_unit->data ) 
    { 
        printf( "%s" , id ) ; 
        printf( "### %s ###" , sel_unit->id_str ) ;

        printf( "    " ) ; 
        printf( " %3d %3d calls: %lld\n" , 
                sel_unit->data->max_orbit, sel_unit->data->NE, 
                sel_unit->calls) ;

            printf( "%s" , id ) ; 
            printf( "    max_orbit  " ) ; 
            printf( "%3d " , sel_unit->data->max_orbit  ) ; 
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    NE         " ) ; 
            printf( "%3d " , sel_unit->data->NE ) ; 
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    E  " ) ; 
            printf( "%+11.2f " , sel_unit->data->E / g_phys_math_const.e1  ) ; 
            printf( "\n" ) ;

        if ( NULL != sel_unit->data->NE_orbit )
        {

            printf( "%s" , id ) ; 
            printf( "    NE " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%11d " , sel_unit->data->NE_orbit[ii] ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    B  " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.2f " , sel_unit->data->B[ii] / g_phys_math_const.e1  ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    K  " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.2f " , sel_unit->data->K[ii] / g_phys_math_const.e1  ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    V  " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.2f " , sel_unit->data->V[ii] / g_phys_math_const.e1  ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    r  " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.2e " , sel_unit->data->r[ii] ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    ps " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.3e " , 
                        convert_phys_unit_scalar( "si2au-area" ,
                            sel_unit->data->ph_sigma[ii] ) ) ; }
            printf( "\n" ) ;

            printf( "%s" , id ) ; 
            printf( "    pE " ) ; 
            for ( ii = 0 ; ii <= sel_unit->data->max_orbit ; ii++ )
                { printf( "%+11.2f " , sel_unit->data->ph_E[ii] / g_phys_math_const.e1  ) ; }
            printf( "\n" ) ;

            for ( ii = 0 ; ii < sel_unit->data->num_fl ; ii++ )
                {
                    printf( "%s" , id ) ; 
                    printf( "    fl " ) ; 
                    printf( "%5d %5d %+11.2e  %+11.2f\n" , 
                        sel_unit->data->fl_orbit[ii][0], 
                        sel_unit->data->fl_orbit[ii][1],
                        convert_phys_unit_scalar( "si2au-rate" , 
                            sel_unit->data->fl_w[ii] ) ,
                        sel_unit->data->fl_E[ii] / g_phys_math_const.e1 ) ; 
                }

            for ( ii = 0 ; ii < sel_unit->data->num_au ; ii++ )
                { 
                    printf( "%s" , id ) ; 
                    printf( "    au " ) ; 
                    printf( "%5d %5d %5d %+11.2e  %+11.2f\n" , 
                        sel_unit->data->au_orbit[ii][0], 
                        sel_unit->data->au_orbit[ii][1],
                        sel_unit->data->au_orbit[ii][2],
                        convert_phys_unit_scalar( "si2au-rate" , 
                            sel_unit->data->au_w[ii] ) , 
                        sel_unit->data->au_E[ii] / g_phys_math_const.e1 ) ; 
                }
            printf( "\n" ) ;
        }


    }
    else
    {
        if ( 0 == sel_unit->num_chldrn ) 
        {
            printf( "%s    %s\n" , id , sel_unit->id_str  ) ; 
        }
    }


  return 0 ;
}




// #######################################################################
// #######################################################################
// #######################################################################



struct proc_data_chain
{
    struct proc_data_chain *prev ;
    struct proc_data_chain *next ;

    char id ;

    double v ;
    double E ;

    int *conf1 ;
    int *conf2 ;
} ;




//#define PRINT_LOADED_DATA 1

//############################## FUNCTION #################################

int load_xatom_raw_data_version1( int a_max_occ_shell , int *a_xmdyn_conf , 
                         int *a_xatom_conf, int *a_indexes ,
                         FILE **a_fp, FILE **a_fp_save , struct g_struct_xatom_data *x_data , double a_Eph )

{

    int nn , ll , jj2 ;
    int maxid = 10000  , maxL ;
    char x_str[STRLTH] , l_str[STRLTH]  ;
    char l2char[STRLTH] = "spdf" ;
    char input_line[LSTRLTH] , *str_pos ;

    int *a_config ;
    char *a_str ;

    int just_cont = 0 ;
   
    int ii , jj , idat ;
    char *token  , *saveptr ;
    FILE *fp = *a_fp ;
    const char delimiters[] = " ";

    int id = 0 ;

    int dimXc ;
    double *loaded_B = NULL, *loaded_K = NULL , *loaded_r = NULL ;
    int dimP , dimF , dimA  , dimT ;

    int *temp1 , *temp2 , tempi;

    char *proc_id  = NULL;
    double (*proc_val)[2] = NULL ;
    int *proc_orbits_i = NULL , *proc_orbits_f = NULL ;

    enum g_option_string print_logmsg ;

    char datablock[STRLTH] ;
    int  datadim ;

    dimXc = a_xatom_conf[0] ; 
    dimP = dimF = dimA = dimT = 0 ;

    // READ DATA BLOCK starts
    while (fgets( input_line , sizeof( input_line )-1, fp) != NULL ) 
    {  

#if PRINT_LOADED_DATA == 1 
        printf( "%s" , input_line ) ;
#endif

        if ( NULL != *a_fp_save ) { fprintf( *a_fp_save , "%s" , input_line ) ; }

        sscanf( input_line , "%s" , datablock ) ;

        // ORBIT DATA block starts -- if it is coming
        if ( ! is_str_all_alpha( datablock ) )
        {  

            sscanf( input_line , "%le" , &(x_data->E) ) ;

            loaded_B = calloc( dimXc + 1 , sizeof(double) ) ;
            loaded_K = calloc( dimXc + 1 , sizeof(double) ) ;
            loaded_r = calloc( dimXc + 1 , sizeof(double) ) ;

            for ( ii = 0 ; ii < dimXc ; ii++ )
            {
#ifdef WORKAROUND_XE_NO_4f
                if ( 9 == ii )   { continue ; } //* NOTE: ii==9 is the 10th orbit in XMDYN, i.e. EADL designator=24 (4f), see, e.g., eqshell *//
#endif
                if (fgets( input_line , sizeof( input_line )-1, fp) != NULL ) 
                {
#if PRINT_LOADED_DATA == 1 
                    printf( "%s" , input_line ) ;
#endif
                    if ( NULL != *a_fp_save ) { fprintf( *a_fp_save , "%s" , input_line ) ; }
                    sscanf( input_line , "%*s %*s %le %le %le" , 
                            loaded_B+ii , loaded_K+ii , loaded_r+ii ) ;
                 }
                 else
                 {
                     MODULE_ERROR( "ERROR in loading B/K/r from XATOM (or from table)!" , -1 ) ;
                 }
            }

            for ( ii = 1 ; ii <= a_max_occ_shell ; ii++ )
            {
                x_data->NE_orbit[ii] = a_xmdyn_conf[ii] ; 
            }

            if ( NULL != loaded_B )
            {
            for ( ii = 0 ; ii < dimXc ; ii++ )
            {
                x_data->B[ a_indexes[ii] ] = loaded_B[ii] ; 
                x_data->K[ a_indexes[ii] ] = loaded_K[ii] ; 
                x_data->r[ a_indexes[ii] ] = loaded_r[ii] ; 
            }
            }
            free( loaded_B) ; free( loaded_K ) ; free( loaded_r ) ;
        }  
        // ORBIT DATA block ends 

        // dat block starts 
        else if ( strcmp( datablock , "dat" ) == 0 )
        {  

            sscanf( input_line , "%s %d" , datablock , &datadim ) ;

            for ( idat = 0 ; idat < datadim ; idat++ )
            {
                fgets( input_line , sizeof( input_line )-1, fp) ;
#if PRINT_LOADED_DATA == 1 
                printf( "%d %s" , strlen( input_line ) , input_line ) ; 
#endif
                if ( NULL != *a_fp_save ) { fprintf( *a_fp_save , "%s" , input_line ) ; }

                if ( 1 ==  strlen( input_line ) ) { continue ; }

                proc_id  = realloc( proc_id  , (dimT+1) * sizeof(char) ) ;
                proc_val = realloc( proc_val , (dimT+1) * ( 2 * sizeof(double) ) ) ;
                proc_orbits_i = realloc( proc_orbits_i , 
                                         (dimT+1) *( dimXc + 1 ) * sizeof(int) ) ;
                proc_orbits_f = realloc( proc_orbits_f , 
                                         (dimT+1) *( dimXc + 1 ) * sizeof(int) ) ;
                proc_orbits_i[ dimT*(dimXc+1) + 0 ] =  dimXc ;
                proc_orbits_f[ dimT*(dimXc+1) + 0 ] =  dimXc ;

                token = strtok_r ( input_line , delimiters , &saveptr ); 
                sscanf( token , "%c" , proc_id + dimT ) ;

#define  AAAAA   NULL
                token = strtok_r (AAAAA, delimiters , &saveptr );  /* xatom conf1*/
                token = strtok_r (AAAAA, delimiters , &saveptr );  /* xatom conf2*/
                token = strtok_r (AAAAA, delimiters , &saveptr ); 
                sscanf( token , "%le" , &(proc_val[dimT][0]) ) ;   // rate/cross_section
                token = strtok_r (AAAAA, delimiters , &saveptr ); 
                sscanf( token , "%le" , &(proc_val[dimT][1]) ) ;   // Energy

                token = strtok_r (AAAAA, delimiters , &saveptr);   /* `#` in xatom output */

                for ( ii = 1 ; ii <= dimXc ; ++ii )
                {
#ifdef WORKAROUND_XE_NO_4f
                    if ( 10 == ii )   { proc_orbits_i[ dimT*(dimXc+1) + ii] = 0 ; continue ; }  //* NOTE: 10th orbit in XMDYN, i.e. EADL designator=24 (4f), see, e.g., eqshell *//
#endif
                    token = strtok_r (AAAAA, delimiters , &saveptr);  
                    if ( NULL == token ) /* Inconsistent xatom output */
                    {
                        MODULE_ERROR( "xatom-onthefly error: Inconsistent xatom output" , -1 ) ;
                    }
                    sscanf( token , "%d" , &(proc_orbits_i[ dimT*(dimXc+1) + ii]) ) ;
                }

                token = strtok_r (AAAAA, delimiters , &saveptr ); /*  `->` in xatom output  */

                for ( ii = 1 ; ii <= dimXc ; ++ii )
                {
#ifdef WORKAROUND_XE_NO_4f
                    if ( 10 == ii )   { proc_orbits_f[ dimT*(dimXc+1) + ii] = 0 ; continue ; }   //* NOTE: 10th orbit in XMDYN, i.e. EADL designator=24 (4f), see, e.g., eqshell *//
#endif
                    token = strtok_r (AAAAA, delimiters , &saveptr);  
                    if ( NULL == token ) /* Inconsistent xatom output */
                    {
                        MODULE_ERROR( "xatom-onthefly error: Inconsistent xatom output" , -1 ) ;
                    }
                    sscanf( token , "%d" , &(proc_orbits_f[ dimT*(dimXc+1) + ii]) ) ;
                }

                if    ( ( 'P' == proc_id[dimT] ) || ( 'V' == proc_id[dimT] ) ) { dimP++ ; }
                else if ( 'F' == proc_id[dimT] ) { dimF++ ; }
                else if ( 'A' == proc_id[dimT] ) { dimA++ ; }
                dimT++ ;

            }

    /** Now we copy data to xdata */


            alloc_xatom_data_content( x_data , NULL , "fl" , dimF ) ;
            alloc_xatom_data_content( x_data , NULL , "au" , dimA ) ;
            x_data->fdprime_pcs = 0 ;

            int *orbit_0 , *orbit_tmpl ;
            orbit_tmpl = calloc( ( a_max_occ_shell + 1 ) , sizeof(int) ) ;
            orbit_0 = calloc( ( a_max_occ_shell + 1 ) , sizeof(int) ) ;

            convert_shellid_xatom2endl( NULL , a_xatom_conf , &orbit_0 , NULL ) ;

            int ccA , ccF ;
            ccA = ccF = 0 ;

            for ( ii = 0 ; ii < dimT ; ii++ )
            {
                convert_shellid_xatom2endl( NULL , proc_orbits_i + ii*(dimXc+1) , &orbit_tmpl , NULL ) ;
                    convert_shellid_xatom2endl( NULL , proc_orbits_f + ii*(dimXc+1) , &orbit_tmpl , NULL ) ;

                if ( ( 'P' == proc_id[ii] ) || ( 'V' == proc_id[ii] ) )
                {
                    do_econf_diff_list( orbit_0 , orbit_tmpl , a_max_occ_shell , 
                                        &tempi , 1 ) ;
                    x_data->ph_sigma[tempi] = proc_val[ii][0] ;
                    x_data->ph_E[tempi]     = proc_val[ii][1] ;
                    x_data->fdprime_pcs    += proc_val[ii][0] ; // Here we collect f_double_prime
                }

                else if ( 'F' == proc_id[ii] ) 
                { 
                    do_econf_diff_list( orbit_0 , orbit_tmpl , a_max_occ_shell , 
                                        x_data->fl_orbit[ccF] , 2 ) ;
                    x_data->fl_w[ccF] = proc_val[ii][0] ;
                    x_data->fl_E[ccF] = proc_val[ii][1] ;
                    ccF++ ;
                }

                else if ( 'A' == proc_id[ii] ) 
                {  
                    do_econf_diff_list( orbit_0 , orbit_tmpl , a_max_occ_shell , 
                                        x_data->au_orbit[ccA] , 3 ) ;
                    x_data->au_w[ccA] = proc_val[ii][0] ;
                    x_data->au_E[ccA] = proc_val[ii][1] ;
                    ccA++ ;
                }
            }

            if ( dimF > 0 ) { x_data->num_fl = dimF ; }
            if ( dimA > 0 ) { x_data->num_au = dimA ; } 

            free( orbit_tmpl ) ; free( orbit_0 ) ;
            if ( dimT > 0 )
            {
                free( proc_id ) ; 
                free( proc_val ) ; 
                free( proc_orbits_i ) ; 
                free( proc_orbits_f ) ;
            }

        }  
        // dat block ends

        // scd block starts
        else if ( strcmp( datablock , "scd" ) == 0 )
        {  
            double *bufferQ ;
            sscanf( input_line , "%s %d" , datablock , &datadim ) ;
            //alloc_xatom_data_content_old( x_data , NULL , 0 , 0 , 0 , 0 , datadim ) ;
            alloc_xatom_data_content( x_data , NULL , "scd" , datadim ) ;
            bufferQ = calloc( datadim , sizeof(double) ) ;
            x_data->num_sample = datadim ;
            x_data->q_sample = g_xatom_tree[x_data->Z].Q ;
            // We load Q and f0 data
            for ( idat = 0 ; idat < datadim ; idat++ )
            {
                fgets( input_line , sizeof( input_line )-1, fp) ;
                sscanf( input_line , "%le %le" , bufferQ+idat , (x_data->f0)+idat ) ;
            }
            // if 1st dbase loaded, we fill up Q
            if ( 0 == g_xatom_tree[x_data->Z].numQ ) 
            {
                for ( idat = 0 ; idat < datadim ; idat++ )
                {
                    g_xatom_tree[x_data->Z].Q[idat] = bufferQ[idat] ;
                    g_xatom_tree[x_data->Z].numQ = datadim ;
                }
            }
            // We check that really same Q 
            if ( datadim != g_xatom_tree[x_data->Z].numQ ) { MODULE_ERROR( "scd numQ inconsistency" , -1 ) ; }
            for ( idat = 0 ; idat < datadim ; idat++ ) 
            { 
                if ( g_xatom_tree[x_data->Z].Q[idat] != bufferQ[idat] )
                {
                   //xmdyn_error( "scd Q inconsistency" ) ;
                   MODULE_ERROR( "scd Q inconsistency" , -1 ) ;
                }
            }
            free(bufferQ) ;
        }
        // scd block ends

        //mad block starts
        else if ( strcmp( datablock , "mad" ) == 0 )
        {  
            double *vectorE, *vectorfp, *vectorfdp ;
            sscanf( input_line , "%s %d" , datablock , &datadim ) ;
            vectorE = calloc( datadim , sizeof(double) ) ;
            vectorfp = calloc( datadim , sizeof(double) ) ;
            vectorfdp = calloc( datadim , sizeof(double) ) ;
            // We load Q and f0 data
            for ( idat = 0 ; idat < datadim ; idat++ )
            {
                fgets( input_line , sizeof( input_line )-1, fp) ;
                sscanf( input_line , "%le %le %le" , vectorE+idat , vectorfp+idat , vectorfdp+idat ) ;
            }
            full_interp1( vectorE  , vectorfp , datadim , 
                          &a_Eph , &(x_data->fprime)  , 1 ,   "def"   ) ;
            full_interp1( vectorE  , vectorfdp , datadim , 
                          &a_Eph , &(x_data->fdprime) , 1 ,   "def"   ) ;

            free(vectorE) , free(vectorfp) , free(vectorfdp) ;
        }
        // mad block ends

        // ixs block starts
        else if ( strcmp( datablock , "ixs" ) == 0 )
        {  
            double *bufferQ , *bufferD ; int numQ ;
            sscanf( input_line , "%s %d" , datablock , &datadim ) ; numQ = g_xatom_tree[x_data->Z].numQ ;
            alloc_xatom_data_content( x_data , NULL , "ixs" , numQ ) ;
            bufferQ = calloc( datadim , sizeof(double) ) ;
            bufferD = calloc( datadim , sizeof(double) ) ;
            x_data->num_sample = numQ ;
            // We load Q and f0 data
            for ( idat = 0 ; idat < datadim ; idat++ )
            {
                fgets( input_line , sizeof( input_line )-1, fp) ;
                sscanf( input_line , "%le %le" , bufferQ+idat , bufferD+idat ) ;
            }
            // We check that really same Q 
            for ( idat = 0 ; idat < numQ ; idat++ )
            {
                if ( x_data->q_sample[idat] <= bufferQ[datadim-1] )
                {
                    full_interp1( bufferQ , bufferD , datadim , 
                                  (x_data->q_sample)+idat , (x_data->ixs)+idat  , 1 ,   "def"   ) ;
                }
                else { x_data->ixs[idat] = bufferD[datadim-1] ; }
            }
            free(bufferQ) ; free( bufferD ) ;
        }
        // ixs block ends

        else
        {  
            fprintf( stdout, "Unknown block: %s " , datablock ) ; break ;
            MODULE_ERROR( "" , -1 ) ;
        }


    } // READ DATA BLOCK ends

    return 0 ;
}




//############################## FUNCTION #################################

int correct_xatom_loaded_data( struct g_struct_xatom_data *x_data )
{

    int ii ;


    x_data->NE = 0 ;


    /* This part refers to database version 10002 */

    double t_au2si = 2.418884326505e-17 , l_au2si = 5.2917720859e-11 ;
    double e1 =  g_phys_math_const.e1 ;
    double epsilon = 1e-6 ;
    double alpha = 1. / ( 4. * PI * g_phys_math_const.eps0 ) * e1*e1 / ( g_phys_math_const.hbar * g_phys_math_const.c ) ;
    double Eph = g_inputs.Ephoton ;

    x_data->E *= 27.21138 * e1  ;
    for ( ii = 0 ; ii <= x_data->max_orbit ; ii++ ) 
    {
        // NOTE: We get rid of zero kinetic energy by adding epsilon
        if ( 0 == x_data->ph_E[ii] ) { x_data->ph_E[ii] = epsilon ; }

        x_data->NE +=  x_data->NE_orbit[ii] ;
        x_data->B[ii] *=  - e1 ;
        x_data->K[ii] *= e1 ;
        x_data->V[ii] = - x_data->B[ii] - x_data->K[ii] ;

        x_data->r[ii] = convert_unit_au2m( x_data->r[ii] ) ;

        x_data->ph_E[ii] *= 27.21138 * e1 ;
        x_data->ph_sigma[ii] *=  l_au2si * l_au2si  ;
    }

    for ( ii = 0 ; ii < x_data->num_fl ; ii++ ) 
    {
        // NOTE: We get rid of zero kinetic energy by adding epsilon
        if ( 0 == x_data->fl_E[ii] ) { x_data->fl_E[ii] = epsilon ; }

        x_data->fl_w[ii] /= t_au2si   ;
        x_data->fl_E[ii] *= e1 ;
    }

    for ( ii = 0 ; ii < x_data->num_au ; ii++ ) 
    {
        // NOTE: We get rid of zero kinetic energy by adding epsilon
        if ( 0 == x_data->au_E[ii] ) { x_data->au_E[ii] = epsilon ; }

        x_data->au_w[ii] /= t_au2si ;
        x_data->au_E[ii] *= e1 ;
    }

    double x = - convert_unit_J2au( 1. ) / ( 4.0 * PI * alpha ) ;
    x_data->fdprime_pcs *= Eph * x  ;

    return 0 ;

}



//############################## FUNCTION #################################

int get_dbase_filename( int dbase_version , 
                        char *dbase , double a_Eph , int a_Z , char *a_conf_str , char *target ,
                        char *a_fname , char *a_log , char *version_file, char *version_str ,
                        int Edependence ) 
{
    char dir_dbase[LSTRLTH] = "" ;
    char dir_dbase_log[LSTRLTH] = "" ;

    char dbase_id[STRLTH] = "" , xlocal[STRLTH] ; 
    char empty_str[1] = "" ;

    char ext[10] = "" , *sel_dir ; 

    char EphE_str[STRLTH] , Zstr[STRLTH] , dbase_version_str[STRLTH] ;

    sprintf( dbase_version_str , "%02d" , dbase_version ) ;

    strcpy( a_fname , "" ) ;       strcpy( a_log , "" ) ;
    strcpy( version_file , "" ) ;  strcpy( version_str , "" ) ;

    sprintf( Zstr , "%d" , a_Z ) ;
    sprintf( EphE_str, "%.3f" ,  a_Eph / g_phys_math_const.e1 )  ;

    strcat( strcat( strcat( strcat( strcat( version_str , "on-the-fly-" ) , EphE_str ), "eV-" ) , Zstr ) , ":" ) ;

    strcpy( xlocal , "" ) ;   sel_dir = s_dbase_dir_load ;
    if ( strcmp( target , "save" ) == 0  ) { sel_dir = s_dbase_dir_save ; }

    strcat(strcat(strcat(strcat(strcat(strcat( dir_dbase , sel_dir ) , "/") , dbase_version_str ) , "-" ) , xlocal) , dbase )  ;
    if ( Edependence ) { strcat(strcat( dir_dbase ,  "/" ) , EphE_str ) ; }
    strcat(strcat( dir_dbase , "/" ) , Zstr ) ;
    strcpy( dbase_id , "/" ) ;
    strcpy( ext , ".dat" ) ;

    strcpy( dir_dbase_log , dir_dbase ) ; strcat( dir_dbase_log , "/log" ) ;

    {
        if ( strcmp( target , "load" ) != 0  ) { dircreate( dir_dbase ) ; dircreate( dir_dbase_log ) ; } 
    }

    strcat(strcat(strcat(strcat( a_fname  , dir_dbase ) , dbase_id ) ,  a_conf_str ) , ext ) ;
    strcat(strcat(strcat(strcat( a_log    , dir_dbase_log ) , dbase_id ) ,  a_conf_str ) , ".log" ) ;
    strcat(strcat( version_file   , dir_dbase_log ) , "/xatom-version.log" ) ;

    return 0 ;
}



//############################## FUNCTION #################################

int xatom_call_on_the_fly_version3( int a_Z , int a_max_occ_shell , int *a_config , double a_Eph ,
                           struct g_struct_xatom_data *x_data ) 
{

    // inputs
    char *input_dbases = g_inputs.dbase_types ;
    ////

    char Z_str[STRLTH] , PE_str[STRLTH] , conf_str_calc[LSTRLTH] , conf_str_full[LSTRLTH] , fname[LSTRLTH] , version_file[LSTRLTH] ;
    int ii ;

    // NOTE! `scd' MUST come before `ixs', because ixs is interpolated to scd grid!
    int dbase_type_num = 5 ;
    char * dbase_types [] = {
        "atm" ,         // 0
        "pcs" ,         // 1
        "scd" ,         // 2
        "ixs" ,         // 3
        "mad" ,         // 4
    } ;
    char * xatom_opts [] = {
        STR( XATOM_ATM_FLAGS ) ,
        STR( XATOM_PCS_FLAGS ) ,
        STR( XATOM_SCD_FLAGS ) ,
        STR( XATOM_IXS_FLAGS ) ,
        STR( XATOM_MAD_FLAGS ) ,
    } ;
    int dbase_Edependence[] = { 0 , 1 , 0 , 1 , 0 } ;

    char **dbase_buffer ; 

    int save_xatom_log = 1 ;

    // If the data have already been calculated, we can simply exit:
    if ( 0 <= x_data->NE ) { return 0 ; } 

    get_xatom_conf_str( a_max_occ_shell , a_config , conf_str_calc , 'F' ) ; // Cannot be 'C' !!!
    get_xatom_conf_str( a_max_occ_shell , a_config , conf_str_full , 'F' ) ;

    int *xatom_conf = calloc( a_max_occ_shell + 1 , sizeof(int) ) ;
    int *orb_conv_ind = calloc( a_max_occ_shell + 1 , sizeof(int) ) ;
    convert_shellid_endl2xatom( a_max_occ_shell , a_config , &xatom_conf , orb_conv_ind ) ;

    FILE *fp ;
    char log_file[LSTRLTH] = ""  , version_str[LSTRLTH] = "" ;   
    char cmd[STRLTH] = ""  , opt[STRLTH] = ""  ;

    dbase_buffer = calloc( dbase_type_num , sizeof( char* ) ) ; 
    for ( ii=0 ; ii < dbase_type_num ; ii++ )  {  dbase_buffer[ii] = NULL ; }

    sprintf( Z_str , "%02d" , a_Z ) ;
    sprintf( PE_str , "%.3f" , a_Eph / g_phys_math_const.e1 ) ;

    // What dbase is to be processed and how? (none/load/calc)
    for ( ii=0 ; ii < dbase_type_num ; ii++ ) 
    {
        if ( NULL != strstr( input_dbases , dbase_types[ii] ) ) 
        { 
            get_dbase_filename( 2 , dbase_types[ii] , a_Eph ,  a_Z , conf_str_full , "load" ,
                                fname , log_file , version_file, version_str , dbase_Edependence[ii] ) ;

            if ( get_filesystem_object_type( fname ) == 'f' ) 
            {  
                fprintf( stdout , "Loading %s" , fname ) ; fflush( stdout ) ; 
                load_stream( dbase_buffer+ii , fname , 'f' ) ;
                fprintf( stdout , "\n" ) ;
            }

            else  
            {
                strcpy( opt, " " ) ; strcpy( cmd, "" ) ;
                strcat(strcat(strcat(strcat(strcat(strcat( opt , s_xatom_opts ) , 
                                         " -Z " ), Z_str ) , " -conf " ) , conf_str_calc ) , " " ) ;
                strcat( strcat( opt , xatom_opts[ii] ) , " " ) ;
                if ( 1 == dbase_Edependence[ii] ) { strcat(strcat( opt , " -PE " ) , PE_str ) ; } 
                strcat(strcat( cmd , STR( XATOM_EXE ) ) , opt ) ;
                if ( save_xatom_log != 0 ) 
                { 
                    get_dbase_filename( 2 ,dbase_types[ii] , a_Eph ,  a_Z , conf_str_full , "save" ,
                                        fname , log_file , version_file, version_str , dbase_Edependence[ii] ) ;
                    strcat(strcat( cmd , " -o " ) , log_file ) ; 
                }
                else 
                { 
                    strcat( cmd , " -silent " ) ; 
                }
                printf( "%s" , cmd ) ; fflush( stdout ) ;
                strcat( cmd , " 2>/dev/null" ) ;
                load_stream( dbase_buffer+ii , cmd , 'p' ) ; 
                fp = fopen( fname , "w" ) ; fprintf( fp , "%s" , dbase_buffer[ii] ) ; fclose(fp) ;
                printf( "   DONE\n" ) ;

            }

        }
    }
        
    x_data->max_orbit = a_max_occ_shell ;
    alloc_xatom_data_content( x_data , conf_str_full , "max_occ_shell" , x_data->max_orbit  ) ;

    for ( ii = 0 ; ii < dbase_type_num ; ii++ )
    {
        if ( NULL != dbase_buffer[ii] )
        {
            FILE *stream , *nullstream = NULL ;
            if ( ( strlen( dbase_buffer[ii] ) == 0 ) || ( -1 == sscanf( dbase_buffer[ii] , "%*s") ) )
            { 
                char x[LSTRLTH] = "" ; strcat(strcat(strcat(strcat(x,"Empty Z = " ) , Z_str ) , " atomic config: " ) , conf_str_full ) ; MODULE_ERROR( x , -1 ) ;
            }
            stream = fmemopen( dbase_buffer[ii] , strlen (dbase_buffer[ii]) , "r" ) ;
            load_xatom_raw_data_version1( a_max_occ_shell , a_config , xatom_conf , orb_conv_ind , 
                                          &stream , &nullstream , x_data , convert_unit_J2eV( a_Eph ) ) ;
            fclose( stream ) ;
        }
    }

    correct_xatom_loaded_data( x_data ) ;

    free( xatom_conf ) ;
    free( orb_conv_ind ) ;

    for ( ii=0 ; ii < dbase_type_num ; ii++ ) { if ( NULL != dbase_buffer[ii] ) { free( dbase_buffer[ii] ) ; } }
    free( dbase_buffer ) ;

    return 0 ;
}





//############################## FUNCTION #################################

int xmdyn_get_atom_types( int *num_T , int **T , struct g_struct_xatom_data ***x_data )
{

    int NN = g_inputs.N ;

    int ii , tt  ;

    *num_T = 0 ;
    *T = calloc( NN , sizeof(int) ) ; for (ii=0; ii<NN ; ii++) { (*T)[ii] = -1 ; }

    for ( ii = 0 ; ii < NN ; ii++ )
    {
        for ( tt=0 ; (tt < *num_T) && (-1 == (*T)[ii]) ; tt++ )
        {
            if ( (*x_data)[tt] == g_atomic_data[ii].atomic_econfig ) 
            {
                (*T)[ii] = tt ;
            }
        }
        if ( -1 == (*T)[ii])
        {
            *num_T += 1 ;
            (*x_data) =  realloc( *x_data , *num_T * sizeof( struct g_struct_xatom_data* ) ) ;
            (*x_data)[*num_T-1] = g_atomic_data[ii].atomic_econfig ;
            (*T)[ii] = *num_T - 1 ;
        }
    }

    return 0 ;


}


// #######################################################################
// #######################################################################
// #######################################################################

//############################## FUNCTION #################################

int xatom__alloc_confarray( int **a_e_config , int a_max_occ_shell ) 
{
    *a_e_config = calloc( a_max_occ_shell + 1 , sizeof(int) ) ;
    if ( NULL == *a_e_config ) { MODULE_ERROR( "No memory" , -1 ) ; }
    return 0 ;
}



//############################## FUNCTION #################################

int xatom__get_tree_tab( int a_Z , int a_max_occ_shell , int *a_e_config , struct g_struct_xatom_data **a_atomic_econfig )
{
    xatom_tree_add( a_Z , a_max_occ_shell , a_e_config , a_atomic_econfig ) ;
    xatom_call_on_the_fly( a_Z , a_max_occ_shell , a_e_config , g_inputs.Ephoton , *a_atomic_econfig ) ;
    return 0 ;
}


//############################## FUNCTION #################################

int xatom__get_transitionE_conf2conf( double *a_E , int a_Z , int a_max_occ_shell_i , int *a_e_config_i , int a_max_occ_shell_f , int *a_e_config_f )
{
    
    struct g_struct_xatom_data *atomic_econfig_i = NULL ;
    struct g_struct_xatom_data *atomic_econfig_f = NULL ;
    struct g_struct_xatom_data *atomic_econfig_x = NULL ;
    int ii , sig ;

    if ( a_max_occ_shell_i != a_max_occ_shell_f ) { MODULE_ERROR( "Not yet prepared for this." , -1 ) ; }

    int *x_e_config ;
    xatom__alloc_confarray( &x_e_config , a_max_occ_shell_i ) ; 

    for ( ii=0 ; ii <= a_max_occ_shell_i + 1 ; ii++ ) { x_e_config[ii] = a_e_config_i[ii] ; }

    xatom__get_tree_tab( a_Z , a_max_occ_shell_i , a_e_config_i , &atomic_econfig_i ) ;
    if ( NULL != a_e_config_f ) 
    {
        xatom__get_tree_tab( a_Z , a_max_occ_shell_f , a_e_config_f , &atomic_econfig_f ) ;
    }

    {
        *a_E = 0 ;
        int found = 0 ;
        for ( ii = 0 ; ii <= a_max_occ_shell_i ; ii++ )
        {
           if ( a_e_config_i[ii] != a_e_config_f[ii] )
           {
               if ( abs( a_e_config_i[ii] - a_e_config_f[ii] ) != 1 ) { MODULE_ERROR( "Not prepared." , -1 ) ;} ;
               found++ ;
               x_e_config[ii] = a_e_config_f[ii] ;
               xatom__get_tree_tab( a_Z , a_max_occ_shell_i , x_e_config , &atomic_econfig_x ) ;
               if   ( a_e_config_i[ii] > a_e_config_f[ii] ) { *a_E += atomic_econfig_i->B[ii] ; }
               else                                         { *a_E -= atomic_econfig_x->B[ii] ; }

           }
        }
        if ( 1 < found ) { MODULE_ERROR( "Not yet prepared for this" , -1 ) ; }
    }


    free( x_e_config ) ;
    return 0 ;
}



//############################## FUNCTION #################################

int xatom__getconf_change2final( int *a_e_config_f , int *a_max_occ_shell_f , int *a_e_config_i , int a_max_occ_shell_i , int shell , int diff )
{
    int ii ;

    if ( *a_max_occ_shell_f != a_max_occ_shell_i ) { MODULE_ERROR("Not prepared.",-1) ; } 

    for ( ii = 0 ; ii < MIN( a_max_occ_shell_i , *a_max_occ_shell_f ) ; ii++ )
    {
        a_e_config_f[ii] = a_e_config_i[ii] ;
    }

    a_e_config_f[shell] += diff ;
    if ( a_e_config_f[shell] < 0 ) { MODULE_ERROR( "Negative occupation number!" , -1 ) ; } // CrossCheck
    return 0 ;
}




// #######################################################################
// #######################################################################
// #######################################################################



//############################## FUNCTION #################################

int xatom_tree_print_whole( void ) 
{
    int ii , tt , num_trees = 1 ;
    char tree_id_str[STRLTH] ;
    struct g_struct_xatom_tree  **p_tree ;

    num_trees = 1 ; // 2 ;

    for ( tt = 0 ; tt < num_trees ; tt++ )
    {
        if ( tt == 0 ) { p_tree = &g_xatom_tree ; }
        if ( tt == 1 ) { p_tree = &g_xatom_shadow_tree ; }
        sprintf( tree_id_str , "T%d| " , tt ) ; 
        sprintf( tree_id_str , "T:  " ) ; 
        for ( ii = 0 ; ii < maxZvalue ; ++ii )
        {
            if ( NULL != g_xatom_tree[ii].chldrn ) 
            {
                printf( "%s" , tree_id_str ) ;
                printf( "\n") ;
                printf( "%s" , tree_id_str ) ;
                printf( "###############\n") ;
                printf( "%s" , tree_id_str ) ;
                printf( "###  Z = %2d ###\n" , ii ) ;
                printf( "%s" , tree_id_str ) ;
                printf( "###############\n") ;
                printf( "\n") ;
                xatom_tree_print_status( &( (*p_tree)[ii] ) , 0 , tree_id_str ) ;
            }
        }

    }

return 0 ;
}





//############################## FUNCTION #################################

int xatom_call_on_the_fly( int a_Z , int a_max_occ_shell , int *a_config , double a_Eph ,
                           struct g_struct_xatom_data *x_data ) 
{
    int ret ;
    {
         ret = xatom_call_on_the_fly_version3( a_Z , a_max_occ_shell ,
                                               a_config , a_Eph , x_data ) ; 
    }
    return ret ;
}



// ###################################################################


static struct g_struct_xatom_data **s_gstate_neutral_cfgs = NULL ;


//############################## FUNCTION #################################

static double get_atomic_mass( int a_Z )
{

//*  Atom masses are from ENDL tables  */

    double allZ[] = { 
                    1.0079e+00 ,
                    4.0026e+00 ,
                    6.9410e+00 ,
                    9.0121e+00 ,
                    1.0810e+01 ,
                    1.2011e+01 ,
                    1.4006e+01 ,
                    1.5999e+01 ,
                    1.8998e+01 ,
                    2.0179e+01 ,
                    2.2989e+01 ,
                    2.4305e+01 ,
                    2.6981e+01 ,
                    2.8085e+01 ,
                    3.0973e+01 ,
                    3.2060e+01 ,
                    3.5453e+01 ,
                    3.9948e+01 ,
                    3.9098e+01 ,
                    4.0080e+01 ,
                    4.4955e+01 ,
                    4.7900e+01 ,
                    5.0941e+01 ,
                    5.1996e+01 ,
                    5.4938e+01 ,
                    5.5847e+01 ,
                    5.8933e+01 ,
                    5.8700e+01 ,
                    6.3546e+01 ,
                    6.5380e+01 ,
                    6.9720e+01 ,
                    7.2590e+01 ,
                    7.4921e+01 ,
                    7.8960e+01 ,
                    7.9904e+01 ,
                    8.3800e+01 ,
                    8.5467e+01 ,
                    8.7620e+01 ,
                    8.8905e+01 ,
                    9.1220e+01 ,
                    9.2906e+01 ,
                    9.5940e+01 ,
                    9.8000e+01 ,
                    1.0107e+02 ,
                    1.0290e+02 ,
                    1.0640e+02 ,
                    1.0786e+02 ,
                    1.1241e+02 ,
                    1.1482e+02 ,
                    1.1869e+02 ,
                    1.2175e+02 ,
                    1.2760e+02 ,
                    1.2690e+02 ,
                    1.3130e+02 ,
                    1.3290e+02 ,
                    1.3733e+02 ,
                    1.3890e+02 ,
                    1.4012e+02 ,
                    1.4090e+02 ,
                    1.4424e+02 ,
                    1.4500e+02 ,
                    1.5040e+02 ,
                    1.5196e+02 ,
                    1.5725e+02 ,
                    1.5892e+02 ,
                    1.6250e+02 ,
                    1.6493e+02 ,
                    1.6726e+02 ,
                    1.6893e+02 ,
                    1.7304e+02 ,
                    1.7496e+02 ,
                    1.7849e+02 ,
                    1.8094e+02 ,
                    1.8385e+02 ,
                    1.8620e+02 ,
                    1.9020e+02 ,
                    1.9222e+02 ,
                    1.9509e+02 ,
                    1.9696e+02 ,
                    2.0059e+02 ,
                    2.0437e+02 ,
                    2.0720e+02 ,
                    2.0898e+02 ,
                    2.0900e+02 ,
                    2.1000e+02 ,
                    2.2200e+02 ,
                    2.2300e+02 ,
                    2.2600e+02 ,
                    2.2700e+02 ,
                    2.3203e+02 ,
                    2.3100e+02 ,
                    2.3802e+02 ,
                    2.3700e+02 ,
                    2.4400e+02 ,
                    2.4300e+02 ,
                    2.4700e+02 ,
                    2.4700e+02 ,
                    2.5100e+02 ,
                    2.5200e+02    
    } ;

    return( allZ[a_Z-1] ) ;
}



//############################## FUNCTION #################################

static int get_grstate_neutral_endl( int a_Z , int *a_maxshellnum , int **a_cfg ) 
{
    int *sel_cfg ;

    int cfg1 [] = { 0, 1 } ;
    int cfg2 [] = { 0, 2 } ;
    int cfg3 [] = { 0, 2, 0, 1 } ;
    int cfg4 [] = { 0, 2, 0, 2 } ;
    int cfg5 [] = { 0, 2, 0, 2, 0, 1, 0 } ;
    int cfg6 [] = { 0, 2, 0, 2, 0, 2, 0 } ;
    int cfg7 [] = { 0, 2, 0, 2, 0, 3, 0 } ;
    int cfg8 [] = { 0, 2, 0, 2, 0, 4, 0 } ;
    int cfg9 [] = { 0, 2, 0, 2, 0, 5, 0 } ;
    int cfg10[] = { 0, 2, 0, 2, 0, 6, 0 } ;
    int cfg11[] = { 0, 2, 0, 2, 0, 6, 0, 0, 1 } ;
    int cfg12[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2 } ;
    int cfg13[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 1, 0 } ;
    int cfg14[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 2, 0 } ;
    int cfg15[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 3, 0 } ;
    int cfg16[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 4, 0 } ;
    int cfg17[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 5, 0 } ;
#ifdef    WORKAROUND_AR_4S
    int cfg18[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0 } ;
#else  // WORKAROUND_AR_4S
    int cfg18[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0 } ;
#endif // WORKAROUND_AR_4S
    int cfg19[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 1 } ;
    int cfg20[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 2 } ;
    int cfg21[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 2 } ;
    int cfg22[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 2, 0, 0, 2 } ;
    int cfg23[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 3, 0, 0, 2 } ;
    int cfg24[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 5, 0, 0, 1 } ;
    int cfg25[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 5, 0, 0, 2 } ;
    int cfg26[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 6, 0, 0, 2 } ;
    int cfg27[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 7, 0, 0, 2 } ;
    int cfg28[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 8, 0, 0, 2 } ;
    int cfg29[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 1 } ;
    int cfg30[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2 } ;
    int cfg31[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 1, 0 } ;
    int cfg32[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 2, 0 } ;
    int cfg33[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 3, 0 } ;
    int cfg34[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 4, 0 } ;
    int cfg35[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 5, 0 } ;
    int cfg36[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0 } ;
    int cfg37[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 1 } ;
    int cfg38[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg39[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 2 } ;
    int cfg40[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 2, 0, 0, 0, 0, 0, 2 } ;
    int cfg41[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 4, 0, 0, 0, 0, 0, 1 } ;
    int cfg42[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 5, 0, 0, 0, 0, 0, 1 } ;
    int cfg43[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 6, 0, 0, 0, 0, 0, 1 } ;
    int cfg44[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 7, 0, 0, 0, 0, 0, 1 } ;
    int cfg45[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 8, 0, 0, 0, 0, 0, 1 } ;
    int cfg46[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 1 } ;
    int cfg47[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 1 } ;
    int cfg48[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2 } ;
    int cfg49[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 1, 0 } ;
    int cfg50[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 2, 0 } ;
    int cfg51[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 3, 0 } ;
    int cfg52[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 4, 0 } ;
    int cfg53[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 5, 0 } ;
    int cfg54[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 6, 0 } ;
    int cfg55[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 } ;
    int cfg56[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg57[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg58[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg59[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 3, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg60[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 4, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg61[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 5, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg62[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 6, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg63[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 7, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg64[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 7, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg65[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 9, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg66[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg67[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 11, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg68[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 12, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg69[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 13, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg70[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg71[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg72[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg73[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg74[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg75[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg76[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg77[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg78[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 0, 1 } ;
    int cfg79[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 1 } ;
    int cfg80[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg81[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 1, 0 } ;
    int cfg82[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 0 } ;
    int cfg83[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 3, 0 } ;
    int cfg84[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 4, 0 } ;
    int cfg85[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 5, 0 } ;
    int cfg86[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 6, 0 } ;
    int cfg87[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 } ;
    int cfg88[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg89[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg90[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg91[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg92[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 3, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg93[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 4, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg94[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 6, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg95[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 7, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg96[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 7, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg97[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 8, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg98[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 10, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    int cfg99[] = { 0, 2, 0, 2, 0, 6, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 14, 0, 0, 2, 0, 6, 0, 0, 10, 0, 0, 11, 0, 0, 0, 0, 0, 2, 0, 6, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2 } ;
    
    
    int all_maxshellnum[] = { 1, 1, 3, 3, 6, 6, 6, 6, 6, 6, 8, 8, 11, 11, 11, 11, 11, 11, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 19, 19, 19, 19, 19, 19, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 30, 30, 30, 30, 30, 30, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 44, 44, 44, 44, 44, 44, 58, 58, 58, 58, 58, 58, 58, 58, 58, 58, 58, 58, 58 } ;

#ifdef    WORKAROUND_AR_4S
    all_maxshellnum[17] = 16 ;
#endif // WORKAROUND_AR_4S

    switch ( a_Z ) {
        case 1 : sel_cfg = cfg1  ; break;  
        case 2 : sel_cfg = cfg2  ; break;  
        case 3 : sel_cfg = cfg3  ; break;  
        case 4 : sel_cfg = cfg4  ; break;  
        case 5 : sel_cfg = cfg5  ; break;  
        case 6 : sel_cfg = cfg6  ; break;  
        case 7 : sel_cfg = cfg7  ; break;  
        case 8 : sel_cfg = cfg8  ; break;  
        case 9 : sel_cfg = cfg9  ; break;  
        case 10: sel_cfg = cfg10 ; break;  
        case 11: sel_cfg = cfg11 ; break;  
        case 12: sel_cfg = cfg12 ; break;  
        case 13: sel_cfg = cfg13 ; break;  
        case 14: sel_cfg = cfg14 ; break;  
        case 15: sel_cfg = cfg15 ; break;  
        case 16: sel_cfg = cfg16 ; break;  
        case 17: sel_cfg = cfg17 ; break;  
        case 18: sel_cfg = cfg18 ; break;  
        case 19: sel_cfg = cfg19 ; break;  
        case 20: sel_cfg = cfg20 ; break;  
        case 21: sel_cfg = cfg21 ; break;  
        case 22: sel_cfg = cfg22 ; break;  
        case 23: sel_cfg = cfg23 ; break;  
        case 24: sel_cfg = cfg24 ; break;  
        case 25: sel_cfg = cfg25 ; break;  
        case 26: sel_cfg = cfg26 ; break;  
        case 27: sel_cfg = cfg27 ; break;  
        case 28: sel_cfg = cfg28 ; break;  
        case 29: sel_cfg = cfg29 ; break;  
        case 30: sel_cfg = cfg30 ; break;  
        case 31: sel_cfg = cfg31 ; break;  
        case 32: sel_cfg = cfg32 ; break;  
        case 33: sel_cfg = cfg33 ; break;  
        case 34: sel_cfg = cfg34 ; break;  
        case 35: sel_cfg = cfg35 ; break;  
        case 36: sel_cfg = cfg36 ; break;  
        case 37: sel_cfg = cfg37 ; break;  
        case 38: sel_cfg = cfg38 ; break;  
        case 39: sel_cfg = cfg39 ; break;  
        case 40: sel_cfg = cfg40 ; break;  
        case 41: sel_cfg = cfg41 ; break;  
        case 42: sel_cfg = cfg42 ; break;  
        case 43: sel_cfg = cfg43 ; break;  
        case 44: sel_cfg = cfg44 ; break;  
        case 45: sel_cfg = cfg45 ; break;  
        case 46: sel_cfg = cfg46 ; break;  
        case 47: sel_cfg = cfg47 ; break;  
        case 48: sel_cfg = cfg48 ; break;  
        case 49: sel_cfg = cfg49 ; break;  
        case 50: sel_cfg = cfg50 ; break;  
        case 51: sel_cfg = cfg51 ; break;  
        case 52: sel_cfg = cfg52 ; break;  
        case 53: sel_cfg = cfg53 ; break;  
        case 54: sel_cfg = cfg54 ; break;  
        case 55: sel_cfg = cfg55 ; break;  
        case 56: sel_cfg = cfg56 ; break;  
        case 57: sel_cfg = cfg57 ; break;  
        case 58: sel_cfg = cfg58 ; break;  
        case 59: sel_cfg = cfg59 ; break;  
        case 60: sel_cfg = cfg60 ; break;  
        case 61: sel_cfg = cfg61 ; break;  
        case 62: sel_cfg = cfg62 ; break;  
        case 63: sel_cfg = cfg63 ; break;  
        case 64: sel_cfg = cfg64 ; break;  
        case 65: sel_cfg = cfg65 ; break;  
        case 66: sel_cfg = cfg66 ; break;  
        case 67: sel_cfg = cfg67 ; break;  
        case 68: sel_cfg = cfg68 ; break;  
        case 69: sel_cfg = cfg69 ; break;  
        case 70: sel_cfg = cfg70 ; break;  
        case 71: sel_cfg = cfg71 ; break;  
        case 72: sel_cfg = cfg72 ; break;  
        case 73: sel_cfg = cfg73 ; break;  
        case 74: sel_cfg = cfg74 ; break;  
        case 75: sel_cfg = cfg75 ; break;  
        case 76: sel_cfg = cfg76 ; break;  
        case 77: sel_cfg = cfg77 ; break;  
        case 78: sel_cfg = cfg78 ; break;  
        case 79: sel_cfg = cfg79 ; break;  
        case 80: sel_cfg = cfg80 ; break;  
        case 81: sel_cfg = cfg81 ; break;  
        case 82: sel_cfg = cfg82 ; break;  
        case 83: sel_cfg = cfg83 ; break;  
        case 84: sel_cfg = cfg84 ; break;  
        case 85: sel_cfg = cfg85 ; break;  
        case 86: sel_cfg = cfg86 ; break;  
        case 87: sel_cfg = cfg87 ; break;  
        case 88: sel_cfg = cfg88 ; break;  
        case 89: sel_cfg = cfg89 ; break;  
        case 90: sel_cfg = cfg90 ; break;  
        case 91: sel_cfg = cfg91 ; break;  
        case 92: sel_cfg = cfg92 ; break;  
        case 93: sel_cfg = cfg93 ; break;  
        case 94: sel_cfg = cfg94 ; break;  
        case 95: sel_cfg = cfg95 ; break;  
        case 96: sel_cfg = cfg96 ; break;  
        case 97: sel_cfg = cfg97 ; break;  
        case 98: sel_cfg = cfg98 ; break;  
        case 99: sel_cfg = cfg99 ; break;  
    }

    int ii ;

    *a_maxshellnum = all_maxshellnum[a_Z-1] ;
    *a_cfg = calloc( *a_maxshellnum + 1 , sizeof(int) ) ;
    for ( ii = 0 ; ii < *a_maxshellnum + 1 ; ii++ )
    {
        (*a_cfg)[ii] = sel_cfg[ii] ; 
    }

    return 0 ;

}


//############################## FUNCTION #################################

int xmdyn_get_econfig_properties( char *a_comm, int a_x , int a_nargs , ... )
{

    struct g_struct_xatom_data  *a_cfg = NULL ;
    va_list valist;
    char *opt ;

    int *xi , **vi ;
    double *xd, **vd ;
    char *xc , **xs ;

    int cc = 0 ;

    /* initialize valist for num number of arguments */
    va_start( valist, a_nargs );

    if ( strcmp( a_comm , "Z" ) == 0 )
    {
        /** This if can be removed once mass is not initialized before econfig  */
        if ( NULL != s_gstate_neutral_cfgs ) 
        { 
            a_cfg = s_gstate_neutral_cfgs[a_x] ; 
        }
    }
    if ( strcmp( a_comm , "cfg" ) == 0 )
    {
        a_cfg = va_arg(valist, struct g_struct_xatom_data* ) ; cc++ ;
    }

    while ( cc < a_nargs )
    {
        opt = va_arg(valist, char*) ; cc++ ;

        if ( strcmp( opt , "str" ) == 0 )
        {
            xs = va_arg(valist, char** ) ; cc++ ;
            *xs = a_cfg->str ;
        }

        if ( strcmp( opt , "max_orbit" ) == 0 )
        {
            xi = va_arg(valist, int* ) ; cc++ ;
            *xi = a_cfg->max_orbit ;
        }

        if ( strcmp( opt , "NE" ) == 0 )
        {
            xi = va_arg(valist, int* ) ; cc++ ;
            *xi = a_cfg->NE ;
        }

        else if ( strcmp( opt , "occupancy" ) == 0 )
        {
            vi = va_arg(valist, int** ) ; cc++ ;
            *vi = a_cfg->NE_orbit ;
        }

        else if ( strcmp( opt , "E" ) == 0 )
        {
            xd = va_arg(valist, double* ) ; cc++ ;
            *xd = a_cfg->E ;
        }

        else if ( strcmp( opt , "r" ) == 0 )
        {
            vd = va_arg(valist, double** ) ; cc++ ;
            *vd = a_cfg->r ;
        }

        else if ( strcmp( opt , "B" ) == 0 )
        {
            vd = va_arg(valist, double** ) ; cc++ ;
            *vd = a_cfg->B ;
        }

        else if ( strcmp( opt , "K" ) == 0 )
        {
            vd = va_arg(valist, double** ) ; cc++ ;
            *vd = a_cfg->K ;
        }

        else if ( strcmp( opt , "V" ) == 0 )
        {
            vd = va_arg(valist, double** ) ; cc++ ;
            *vd = a_cfg->V ;
        }

        else if ( strcmp( opt , "mass" ) == 0 )
        {
            xd = va_arg(valist, double* ) ; cc++ ;
            *xd = get_atomic_mass( a_x ) * g_phys_math_const.mp ;
        }

        else if ( strcmp( opt , "max_orbit" ) == 0 ) {;}
        else { printf("%s\n",opt); MODULE_ERROR("Unknown opt.",-1) ; }


    }

    /* clean memory reserved for valist */
    va_end(valist);

    return 0 ;
}





//############################## FUNCTION #################################

static int xatom_setup_grstate_neutral_endl( int *a_all_Z , int a_size )
{

    int maxZ = 99 ;
    int selZ , ii; 

    static int init = 1 ;

    if ( ! init ) { return -1 ; } 

    s_gstate_neutral_cfgs = calloc( maxZ + 1 , sizeof( struct g_struct_xatom_data* ) ) ;

    int maxshellnum , *config ;
    if ( NULL == a_all_Z ) { a_size = 0 ; printf( "Empty array of Z.\n" ) ; }

    for ( ii = 0 ; ii < a_size ; ii++ )
    {
        selZ = a_all_Z[ii]; 
        get_grstate_neutral_endl( selZ , &maxshellnum , &config ) ;
        xatom_tree_add( selZ , maxshellnum , config , s_gstate_neutral_cfgs + selZ ) ;
        xatom_call_on_the_fly( selZ , maxshellnum , config , g_inputs.Ephoton , s_gstate_neutral_cfgs[selZ] ) ;
        free( config ) ;
    }

    init = 0 ;
    return 0 ;

}





/** Interface  */
//############################## FUNCTION #################################

int init_Xatom( char *dir1 , char *dir2 , char *xatom_opts ) 
{

    strcpy( s_dbase_dir_load , dir1 ) ;
    strcpy( s_dbase_dir_save , dir2 ) ; strcat( s_dbase_dir_save , "/dbase/" ) ;
    strcpy( s_xatom_opts , xatom_opts ) ;

    char dbases[LSTRLTH] = "f0" ;

    int ii , tt , num_trees = 1 ;
    struct g_struct_xatom_tree    **p_tree ;
    
    num_trees = 2 ;

    for ( tt = 0 ; tt < num_trees ; tt++ )
    {
        if ( tt == 0 ) { p_tree = &g_xatom_tree ; }
        if ( tt == 1 ) { p_tree = &g_xatom_shadow_tree ; }

        *p_tree = calloc( maxZvalue + 1 , sizeof( struct g_struct_xatom_data ) ) ;

        for ( ii = 0 ; ii < maxZvalue ; ++ii )
        {
            (*p_tree)[ii].prnt   = NULL ;
            (*p_tree)[ii].chldrn = NULL ;
            (*p_tree)[ii].num_chldrn = 0 ;
            (*p_tree)[ii].id = 0 ;
            (*p_tree)[ii].calls = 0 ;
            (*p_tree)[ii].orbit = 0 ;
            (*p_tree)[ii].occ = 0 ;
            (*p_tree)[ii].data = NULL ;
            (*p_tree)[ii].id_str = malloc( 1024 * sizeof(char) ) ; 
            strcpy( (*p_tree)[ii].id_str , " 0 " ) ;
            (*p_tree)[ii].numQ = 0 ;
            (*p_tree)[ii].Q = NULL ;
        }

        g_tree_prop.maxdepth = 0 ;
        g_tree_prop.max_type_id   = -1 ;

        if ( strstr( dbases , "f0" ) != NULL ) 
        {
            (*p_tree)[ii].numQ = 1 ; // Q is still NULL, that means database to be loaded
        }

    }

  return 0 ;
}



//############################## FUNCTION #################################

int xmdyn_load_grstate_neutral_econfigs( int a_N , int *a_allZ )
{
    xatom_setup_grstate_neutral_endl( a_allZ , a_N ) ;
    return 0 ;
}





//############################## FUNCTION #################################

int module_main_Xatom( void ) 
{
    int xi = 0, *vi = NULL ;
    double xd = 0 , *vd = NULL ;
    char xc , *xs = NULL ;

    xmdyn_get_econfig_properties( "Z", 6 , 2 , "str" , &xs ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "max_orbit" , &xi ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "NE" , &xi ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "occupancy" , &vi ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "E" , &xd ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "r" , &vd ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "B" , &vd ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "K" , &vd ) ;
    xmdyn_get_econfig_properties( "Z", 6 , 2 , "V" , &vd ) ;

    return 0 ;   
}

 




#ifdef MODULE_MAIN
int _main( void )
#else
//############################## FUNCTION #################################

int xatom_test_main( void )
#endif
{

    int c0[] = { 0 , 0 } ;
    int c1[] = { 0 , 1 } ;
    int c2[] = { 0 , 2 , 0 , 1 } ;
    int c3[] = { 0 , 2 , 0 , 1 , 0 , 2 } ;
    int c4[] = { 0 , 2 , 0 , 2 , 0 , 4 , 0 , 0 , 1 } ;
    int c5[] = { 0 , 0 , 0 , 0 , 0 , 2  } ;
    int c6[] = { 0 , 2 , 0 , 2 , 0 , 2 } ;

    struct g_struct_xatom_data *x_data ;

    init_Xatom( "../dbase" , "./dbase"  , "-xmdyn" ) ; 

    xatom_tree_add( 6 , 5 , c6 , &x_data ) ;
    printf( "XXX %d\n",  x_data->NE );
    xatom_call_on_the_fly( 6 , 5 , c6 , 500*g_phys_math_const.e1 , x_data )  ;
    printf( "XXX %d\n",  x_data->NE );

    xatom_tree_add( 6 , 5 , c6 , &x_data ) ;
    printf( "XXX %d\n",  x_data->NE );
    xatom_call_on_the_fly( 6 , 5 , c6 , 500*g_phys_math_const.e1 , x_data )  ;

  return 0 ;
}




