///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


/* ************************************************************************/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
//#include <sys/utsname.h>

#include "common.h"

#include "global_declare.h"

#include "globalconsts.h"

#include "input_cmdline_help.h"
#include "init_system.h"

#include "Xatom.h"

#include "xmdyn_filenames.h"

#include "filetests.h"
#include "xmdyn_io.h"
#include "classdyn.h"
#include "endltabs.h"
#include "tests.h"

#include "mc/quantumproc.h"
#include "secel.h"


#include "../build/obj/Xmdyn_version_updated.h"

//#define MAXKNUM 2
//#define MAXLNUM 4



/* for debugging  */

#define todebugfile(var,format,file) {FILE *fp;char debfile[STRLTH]; \
 if (strcmp(#file,"x")==0)\
   strcpy(debfile,"debugfile.txt");\
 else \
   strcpy(debfile,#file);\
   fp=fopen(debfile,"a");\
   fprintf(fp,#format,var);\
   fclose(fp);\
}

#define debugprint(var,format) {printf(#format,var);printf("\n");}
#define debugsign(text) {printf(#text"\n");}

static enum g_option_string  l_running_status = RUNNING ;
static int INPUT_REREAD=0;
static char current_process='i'; // 'i':init, 'r':running, 's':saving

char commandline[LSTRLTH];
char programname[STRLTH];
 

//############################## FUNCTION #################################

void constans(void){
  //e1=1.602189e-19, eps0=8.854187e-12, pi=3.14159265358979;
  MESSAGES.unit_l=1e-10;
  MESSAGES.unit_t=1e-15;
  MESSAGES.unit_E= g_phys_math_const.e1 ;

  // g_phys_math_const .mp = 


}



//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//  *******************  MAIN  **************************
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX



//############################## FUNCTION #################################

int main(int argc,char **argv)
{

  int i,j,k, ii , jj ;

  int timestep,timestepstart;
  int exitcode=0;
  
  int pid_number,*pid_numbers;
  char hostname[STRLTH],*(*hostnames);
  
  int ans;

  double DTtotal ;

 
  double (*memo_r1)[3] , (*memo_r2)[3] , (*memo_v1)[3] , (*memo_v2)[3]  , (*save_r2)[3]; //(*prev_r2)[3] ;

  strcpy(programname,argv[0]);  
  
  g_inputs.step = 0 ; // init for debugging puposes


  sprintf( g_inputs.xmdyn_version , "%s" , SVN_VERSION ) ;


  init_phys_mat_const() ; 

  if (argc==1) 
  {
    helptext( SVN_VERSION , VERSIONNUMBER , COMPILE_COMMAND ) ;
    exit(0) ;
  }
 
  if ( strcmp( argv[1] , "++pause++" ) == 0  ) 
  {
      pause() ;
      exit( 0 ) ;
  }

  if ( strcmp( argv[1] , "--help" ) == 0  ) 
  {
      if ( argc > 2 ) {
          if ( strcmp( argv[2] , "settings" ) == 0 )
          {
              input_defaults();
              save_summaryfile( -1 ) ;
          }
      }
      else if ( 2 == argc )
      {
          fprintf( stderr , "Options: \n" ) ;
          fprintf( stderr , "        --help settings\n" ) ;
      }
      exit( 0 ) ;
  }



  constans();    //ALLNODES!!!

 
  int My_Rank = 0, Root_Rank = 0 , Event_Rank = 0 , My_Calc_Rank = 0 ,  Num_Calc_Rank = 1 ,  Num_Rank = 1 ;
 
  pid_number=getpid();
  if (My_Rank==Root_Rank) {
    pid_numbers=calloc(Num_Rank,sizeof(int));
  }
  pid_numbers[0]=pid_number;

  
  gethostname(hostname,STRLTH);
  if (My_Rank==Root_Rank) {
    hostnames=calloc(Num_Rank,sizeof(char*));
    for (i=0;i<Num_Rank;++i) {
      hostnames[i]=calloc(MSTRLTH,sizeof(char));
    }
    strcpy(hostnames[0],hostname);
  }
    
  clearstr(commandline);
  for (i=0;i<argc;++i)
    strcat(strcat(commandline,argv[i])," ");

  


 g_inputs.randseed = -1 ;

// **************************************************
// **************************************************
/*********************************
**  Main command line argument  **
**********************************/


  strcpy( g_inputs.inputfile,argv[1]);
  argv++; argc--;

 
  //if ( 0 == strcmp( g_inputs.inputfile , "-v" ) ) { printf( "%s\n" , VERSIONNUMBER ) ; exit( 0 ) ; }
  if ( 0 == strcmp( g_inputs.inputfile , "-v" ) ) { printf( "%s\n" , SVN_VERSION ) ; exit( 0 ) ; }
  //if ( 0 == strcmp( g_inputs.inputfile , "-v-svn" ) ) { printf( "%s\n" , g_inputs.xmdyn_version ) ; exit( 0 ) ; }


  printf(  "### start: %s",cd_currdate());


  input_defaults();
  
  if ( 'd' == get_filesystem_object_type( g_inputs.inputfile ) )
  {
      while ( '/' == g_inputs.inputfile[strlen( g_inputs.inputfile )] ) 
      {
          g_inputs.inputfile[strlen( g_inputs.inputfile ) - 1 ] = '\0' ;
      }
      strcat( g_inputs.inputfile , "/input.txt" ) ;
      strcpy( g_inputs.format_loadsave , "001" ) ;

  }



// **************************************************
// **************************************************
// *******************
// LOAD INPUT & INIT
// *******************

  
  if (My_Rank==Root_Rank) {

    ii = 0 ;
    exitcode=inputin();
    input_postproc() ;

    char *dir_dbase , *dir_root ;
    init_xmdyn_loadsave( &dir_dbase , &dir_root ) ;
    init_init_system() ;
    init_Xatom( dir_dbase , dir_root , g_inputs.xatom_commandline_opt ) ;
    init_module_all() ;

  }  
  error_exit("Cannot load input file. Exit.",-1,exitcode);

  // **************************************************
  // **************************************************
  // **************************************************


  if (My_Rank==Root_Rank)
    exitcode = flags( argc , argv , SVN_VERSION , VERSIONNUMBER ) ;
  error_exit("Wrong flag!\n",-1,exitcode);

  if ( (My_Rank==Root_Rank) && (strcmp( g_inputs.runnumber,"")==0) )
    exitcode=-1;
  error_exit("No runnumber specified!\n",-1,exitcode);

  if ( (Num_Rank==1)  ) {
    Num_Calc_Rank=Num_Rank;
    My_Calc_Rank=My_Rank;
  }
  else {
    Num_Calc_Rank=Num_Rank-1;
    My_Calc_Rank=My_Rank-1;
  }
 

  //    ****    ****    ****    ****    ****    ****


  varibles(2);

  if (My_Rank==Root_Rank) {
    exitcode=0;
    
  if (My_Rank==Root_Rank) {
    filenames();
  }

    if ( i = load_sample( &timestepstart ) , i < 0 ) 
    {
      if ( i == -2 )
      {
        fprintf( stderr , "You cannot have default and specified input at the same time!\n") ;
        fprintf( stderr , "   %s\n   %s\n " , g_inputs.sample_dir , g_inputs.datadir ) ;
      }
      else { printf("Error in startingconfig: %d\n",i); }
      exitcode=-1;
    }


  if (My_Rank==Root_Rank) {
    dirhiercreate();

    exitcode=makingfotomoments();
  }

  error_exit("Problem in makingfotomoments!\n",-1,exitcode);

     g_inputs.step = timestepstart ;

    if ( (exitcode>=0) && (timestepstart< g_inputs.nstep) ) {
      print_extrainfo_out(programname,VERSIONNUMBER,'v',pid_numbers,hostnames);
      print_extrainfo_out(programname,VERSIONNUMBER,'s',pid_numbers,hostnames);
    }
    
    if ( (exitcode>=0) && (timestepstart==0) ) {
      cp_source_input(commandline,programname,VERSIONNUMBER);
    }

    if ( (exitcode>=0) && (classdintableinitialisations('i')<0) ) {
      printf("Error in classdintableinitialisations: %d\n",i);
      exitcode=-2;
    }  

    if (exitcode>=0) {   cache_filenames(20,0) ;   }  // init

    if ( (exitcode>=0) && ( g_inputs.initfoto==1) ) {
      print_extrainfo_out(programname,VERSIONNUMBER,'r',pid_numbers,hostnames);
      cp_source_input("rm",programname,VERSIONNUMBER);
      exitcode=-3;
    }
    if (exitcode>=0) {
      check_if_continued(timestepstart, g_inputs.nstep);
    }
    
    if (exitcode>=0) {
      save_summaryfile(timestepstart);
    }
  }
  error_exit("Error!\n",-1,exitcode);


  xmdyn_md_energy( "system-E" , "calc" ) ; //X calc_tot_init_ener(12) ;  // 
  if (timestepstart==0) {
     g_ENERGY_VALUES.Einitial = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(-1);  // 
  }


  randcalls_counts('s');
  randcalls_counts('i' ,  g_inputs.randseed );
  randcalls_counts('c');
 

//#define  XTEST_INEL_CROSS_SECTION
#ifdef  XTEST_INEL_CROSS_SECTION
  testsuite_inel_cross_section( g_inputs.impion_model ) ;
  exit(0);
#endif



  if (My_Rank==Root_Rank) {
    if (startingdataindex()==-99)  {
      data_loadsave( "snp_full" , g_inputs.format_loadsave , -9 , l_running_status ) ;
    }
    data_loadsave( "snp_full" , g_inputs.format_loadsave , -8 , l_running_status ) ;
  }
  

  current_process='r';
  

  memo_r1 = calloc( ( g_inputs.N+ g_inputs.E) * 3 , sizeof(double) ) ;
  memo_v1 = calloc( ( g_inputs.N+ g_inputs.E) * 3 , sizeof(double) ) ;
  memo_r2 = calloc( ( g_inputs.N+ g_inputs.E) * 3 , sizeof(double) ) ;
  memo_v2 = calloc( ( g_inputs.N+ g_inputs.E) * 3 , sizeof(double) ) ;


cd_tictoc(10,"test1",My_Rank,Num_Rank);

  long long int memo_randcalls ;
  double Nph_in_longstep ;


   {
       if ( timestepstart == 0 ) 
       { 
           if ( 1 == is_snp_save_now( -1 , "snp" ) )
           {
               data_loadsave( "snp" , g_inputs.format_loadsave , -1 , l_running_status ) ;
               data_loadsave( "snp_full_restart" , g_inputs.format_loadsave , 0 , l_running_status ) ;
           }
       }
   }



  // **************************************
  // ##########   TIME-LOOP    ############
  // **************************************

  for ( timestep = timestepstart ; ( timestep < g_inputs.nstep ) && ( l_running_status != EXIT_AT_NEXT_STEP ) ; ++timestep ) 
  {

    xmdyn_clock( "tic" , "---ts" ) ;

    if ( PAUSE100 == l_running_status ) 
    {
        actsteptofile( -1 , PAUSE100 , NULL ) ;
    }

    DUMP_TIMES_START(fs)

    g_inputs.step = timestep ;
    Nph_in_longstep = 0 ;  

    
    if ( 2 != xmdyn_apply_shortcut( "check" ) ) 
    {

    xmdyn_check_opts_while_running( "si-off" , NULL ) ;
    xmdyn_check_opts_while_running( "irec-on" , NULL ) ;
 
    DTtotal =  g_inputs.DT ;
    

    cd_tictoc( 0 , "rec" , My_Rank , Num_Rank ) ;
    xmdyn_clock( "tic" , "---re" ) ;
    instant_recombination( timestep );
    xmdyn_clock( "toc-add" , "---re" ) ;
    cd_tictoc( 1 , "rec" , My_Rank , Num_Rank ) ; 
   

    save_r2 = g_system.r2 ;
    g_system.r2 = g_system.prev_r2 ;
 

    xmdyn_clock( "tic" , "---bo-update" ) ;
    xmdyn_clock( "toc-add" , "---bo-update" ) ;

    g_system.r2 = save_r2 ;

    for ( ii=0 ; ii < g_inputs.N + g_inputs.E; ii++) { for ( jj=0; jj<3; jj++ ) {
        g_system.prev_r2[ii][jj] = g_system.r1[ii][jj] ;
    }}

    {

      { Nph_in_longstep = 0 ; }

       g_inputs.step = timestep;


      {
        cd_tictoc(0,"si",My_Rank,Num_Rank);
        xmdyn_clock( "tic" , "---si" ) ;
        electronscattering( timestep );    //ALLNODES!!!
        xmdyn_clock( "toc-add" , "---si" ) ;
        cd_tictoc(1,"si",My_Rank,Num_Rank);
      }


      {
         g_inputs.subDT =  g_inputs.DT ; 
        cd_tictoc(0,"qp",My_Rank,Num_Rank);
        xmdyn_clock( "tic" , "---qp" ) ;
        quantum_stochast_proc( timestep , DTtotal );
        xmdyn_clock( "toc-add" , "---qp" ) ;
        cd_tictoc(1,"qp",My_Rank,Num_Rank);
        Nph_in_longstep += beam_photon_num( cd_time_now() ,  g_inputs.DT ) ;
      }



      {
        
	if (strcmp( g_inputs.timestepOK,"y")==0){
          cd_tictoc(0,"ts",My_Rank,Num_Rank);
          xmdyn_clock( "tic" , "---md" ) ;
          classicaltimestep( timestep ) ;
          xmdyn_clock( "toc-add" , "---md" ) ;
          cd_tictoc(1,"ts",My_Rank,Num_Rank);
        }

        recombiantion_follow_captured_electrons() ;

        ans=0;


         g_inputs.DT = DTtotal ;

      }
     

      randcalls_counts('c');

    } 

     g_inputs.subDT =  g_inputs.DT ; 

    if (strcmp( g_inputs.timestepOK,"y")==0){
      xmdyn_md_energy( "system-E" , "calc;nextstep" ) ; //X calc_tot_init_ener(112);
    }


    if (My_Rank==Root_Rank) {
      self_consistency_check(timestep,"elconfig");
      self_consistency_check(timestep,"q");
    }

    } // END OF xmdyn_apply_shortcut


    DUMP_TIMES_END(fs)
    xmdyn_clock( "toc-add" , "---ts" ) ;
          
xmdyn_update_atoms( NULL , "all,bond,-env" ) ; // DBGDBG

    //  *** SAVING start ***
//#if  0
    {

    DUMP_TIMES_START(wr)
    xmdyn_clock( "tic" , "---wr" ) ;

    if (My_Rank==Root_Rank) 
    {
      current_process='s';
 
      data_loadsave( "snp_full" , g_inputs.format_loadsave , timestep , l_running_status ) ;

      
      if ( EXIT_AT_NEXT_STEP == l_running_status ) 
      {
          data_loadsave( "snp_full" , g_inputs.format_loadsave , timestep , l_running_status ) ;
          data_loadsave( "snp_full_restart" , g_inputs.format_loadsave , timestep , l_running_status ) ;
      }

      if ( ( 1 == is_snp_save_now( g_inputs.step ,"snp" ) ) || ( EXIT_FORCE_SAVE_AT_NEXT_STEP == l_running_status ) )
      {
          xmdyn_update_atoms( NULL , "all,bond,-env" ) ;
          data_loadsave( "snp" , g_inputs.format_loadsave , timestep , l_running_status ) ;
          if ( EXIT_AT_NEXT_SNP == l_running_status ) { l_running_status = EXIT_AT_NEXT_STEP ; } 
          if ( EXIT_FORCE_SAVE_AT_NEXT_STEP == l_running_status ) { l_running_status = EXIT_AT_NEXT_STEP ; } 
          data_loadsave( "snp_full_restart" , g_inputs.format_loadsave , timestep , l_running_status ) ;
          xmdyn_apply_shortcut( "execute" ) ;
      }
      else
      {
          data_loadsave( "snp_full_restart" , g_inputs.format_loadsave , timestep , DUMMY ) ;
      }
     
      if  ( 1 == is_snp_save_now( g_inputs.step ,"E" ) ) 
      {
          data_loadsave( "E" , g_inputs.format_loadsave , timestep , l_running_status ) ;
      }
     
     
      xmdyn_clock( "toc-add" , "---wr" ) ;
      data_loadsave( "progress" , g_inputs.format_loadsave , timestep ,RUNNING ) ;
      //  *** SAVING end ***
      
      if (INPUT_REREAD==1) {

        char *dir_dbase , *dir_root ;

        inputin();
        input_postproc() ;
        init_xmdyn_loadsave( &dir_dbase , &dir_root ) ;
        init_init_system() ;

        save_summaryfile(timestep);
        INPUT_REREAD=0;
      }

      current_process='r';
    }
    DUMP_TIMES_END(wr)

    }



  }

  // ### ### ### TIME-LOOP END ### ### ### ### ###

  free( memo_r1 ) ; free( memo_v1 ) ;  free( memo_r2 ) ; free( memo_v2 ) ; //free( prev_v2 ) ;


  if ( timestep == g_inputs.nstep ) { l_running_status = DONE ; } 
  else { l_running_status = INTERRUPTED ; }

  xmdyn_apply_shortcut( "cancel" ) ;
  data_loadsave( "snp_full_restart" , g_inputs.format_loadsave , timestep - 1  , l_running_status ) ;

  current_process='s';
  if (My_Rank==Root_Rank) {
    instant_recombination( -timestep ) ; // log

    data_loadsave( "cache-close" , g_inputs.format_loadsave , timestep , l_running_status ) ;
    data_loadsave( "progress" , g_inputs.format_loadsave , timestep , l_running_status ) ;

    classdintableinitialisations('f');
    freememory();
  
    print_log( "\n" , NULL , "" , "flush" ) ;
 
    if (timestepstart< g_inputs.nstep) {
      if ( DONE == l_running_status ) {
          print_extrainfo_out(programname,VERSIONNUMBER,'e',pid_numbers,hostnames);
      }
      else
      {
          print_extrainfo_out(programname,VERSIONNUMBER,'i',pid_numbers,hostnames);
      }
    }

  }

  instant_recombination_final();
  
  if  ( DONE != l_running_status ) { printf(  "\n### stopped:   %s" , cd_currdate() ) ; }
  else {  printf( "### end:   %s" , cd_currdate() ) ;  }

  if ( DONE != l_running_status ) { exitcode = 1 ; }   // EXIT_FAILURE ; }


  exit( exitcode ) ;

}

