///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "common.h"

#include "globalconsts.h"
#include "endltabs.h"
#include "classdyn.h"
#include "tests.h"
#include "filetests.h"
#include "xmdyn_io.h"
#include "mc/e_release.h"

/*  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  */

// ########################################################


#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
    //return a_return ;
}

// ########################################################

// --------------------------------------------------------
// ########################################################

//   t= g_phys_math_const.C_k * g_phys_math_const.e2 *(q1)*(q2)/(pow(((double)norm2)+potencrb2,1.5));

#define Cpairinteractionforces(force1,force2,dr12,q1,q2,norm2,t) {\
   t= g_phys_math_const.C_k * g_phys_math_const.e2 *(q1)*(q2) / ( sqrt(((double)norm2)+potencrb2) * (((double)norm2)+potencrb2) ) ;\
   (force1)[0]+=(t)*(dr12)[0]; \
   (force1)[1]+=(t)*(dr12)[1]; \
   (force1)[2]+=(t)*(dr12)[2]; \
   (force2)[0]-=(t)*(dr12)[0]; \
   (force2)[1]-=(t)*(dr12)[1]; \
   (force2)[2]-=(t)*(dr12)[2]; \
}

//   t= g_phys_math_const.C_k * g_phys_math_const.e2 *(q1)*(q2)/(pow(((double)norm2)+potencrb2,1.5));

#define Cpairinteractionforcesfixatoms(force1,force2,dr12,q1,q2,norm2,t) {\
   t= g_phys_math_const.C_k * g_phys_math_const.e2 *(q1)*(q2) / ( sqrt(((double)norm2)+potencrb2) * (((double)norm2)+potencrb2) ) ;\
   (force1)[0]+=(((fixatomsOK=='y')&&(i<N)) ? 0 : (t*dr12[0])); \
   (force1)[1]+=(((fixatomsOK=='y')&&(i<N)) ? 0 : (t*dr12[1])); \
   (force1)[2]+=(((fixatomsOK=='y')&&(i<N)) ? 0 : (t*dr12[2])); \
   (force2)[0]-=(t)*(dr12)[0]; \
   (force2)[1]-=(t)*(dr12)[1]; \
   (force2)[2]-=(t)*(dr12)[2]; \
}





/*  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  --  */

static double *potcut;


// ########################################################
// ########################################################



// ########################################################
// ########################################################
// ########################################################



//############################## FUNCTION #################################

int is_class_part(int part) {

    if ( COULOMB == g_inputs.classdyn_partsel )
    {
        if ( ( g_system.qi [part]!=0 ) &&
             ( g_system.ptype [part]!=  g_inputs.nots) && ( g_system.ptype [part]!=  g_inputs.noall) ) {
          return( 1 ) ;
        }
    }

    // old default:
    else
    {
        if ( ( (part< g_inputs.N) || ( g_system.qi [part]!=0) ) &&
             ( g_system.ptype [part]!=  g_inputs.nots) && ( g_system.ptype [part]!=  g_inputs.noall) ) {
          return(1);
        }
    }
    return(0) ;
}







//############################## FUNCTION #################################

int calc_class_parts_dim(void){
  int i,class_parts_dim;
  //if (My_Rank==Root_Rank) {
    if ( g_inputs.tselectrons=='n') {
      class_parts_dim= g_inputs.N;
    }
    else {
      class_parts_dim=0;
      for (i=0;i< g_inputs.N+ g_inputs.E;++i) {
        if (is_class_part(i)) {
          class_parts_dim++;
        }
      }
    }
  //}
  return(class_parts_dim);
}




//############################## FUNCTION #################################

void set_class_parts(int *class_parts,int class_parts_dim) {
  int c,i;
//  if (My_Rank==Root_Rank) {
    c=0;
    for (i=0;((i< g_inputs.N+ g_inputs.E)&&(c<class_parts_dim));++i) {
      if (is_class_part(i)) {
        class_parts[c]=i;
        c++;
      }
    }
//  }
}





// ###############################
// ########## TIMESTEP  ##########
// ###############################


//############################## FUNCTION #################################

void classicaltimestep( int step )
{
  double  dt ;
  int i,j,k;
  double *r1_0,*v1_0,*pr,*pv;
  static int class_parts_dim, *class_parts  ;
  int gstep = 0 , maxgstep = 1 ;
        class_parts_dim=calc_class_parts_dim();
   
        class_parts=calloc(class_parts_dim,sizeof(int));
        set_class_parts(class_parts,class_parts_dim);

  
  r1_0 = calloc( ( g_inputs.N + g_inputs.E ) * 3 , sizeof(double) ) ;
  v1_0 = calloc( ( g_inputs.N + g_inputs.E ) * 3 , sizeof(double) ) ;  
  
  pr = &( g_system.r1 [0][0] ) ;  pv = &( g_system.v1 [0][0] ) ;
  
  for (i=0;i<( g_inputs.N+ g_inputs.E)*3;++i) {
    r1_0[i] = pr[i] ;
    v1_0[i] = pv[i] ;
  }
  

  for (i=0;i< g_inputs.tsfactor;++i) 
  {

      double time_now = cd_time_now() + i * g_inputs.DT / (double) maxgstep / (double) g_inputs.tsfactor ;

    if (strcmp( g_inputs.int_method,"DUMMY")==0) {;} 
        
     else if  (strcmp( g_inputs.int_method,"verlet_plain")==0){
      md_integrator_plain_verlet(  g_inputs.DT / (double) g_inputs.tsfactor / maxgstep ,
                             class_parts , class_parts_dim ) ;
    }

    else if  (strcmp( g_inputs.int_method,"verlet")==0){
      md_integrator_verlet(  time_now , g_inputs.DT / (double) g_inputs.tsfactor / maxgstep ,
                             class_parts , class_parts_dim ) ;
    }

    else { MODULE_ERROR( "Wrong MD intergator!" , -1 ) ; } ;


    
    
    if ( g_inputs.fixatomsOK=='y') {
      for (j=0;j< g_inputs.N;++j) { for (k=0;k<3;++k) {
	   g_system.v1 [j][k]=0;
	   g_system.r2 [j][k]= g_system.r1 [j][k];
      }}	
    }
     
    
    vectorswap( g_system.r1 , g_system.r2 );
    vectorswap( g_system.v1 , g_system.v2 );

  }

  if ( 0 )
  {
      // because r1 and r2 are swaped, now r1(v1) is the actual pos(vel),
      //         r2(v2) is the old one
      pr=&( g_system.r2 [0][0]);  pv=&( g_system.v2 [0][0]);
      for (i=0;i<( g_inputs.N+ g_inputs.E)*3;++i){
        pr[i]=r1_0[i];
        pv[i]=v1_0[i];
      }
  }
        
  free(r1_0), free(v1_0);

  free(class_parts);
  // }

}







// ###############################
// ########### TUNINGS ###########
// ###############################





// ########################################################



#ifdef COULOMB_NUM_TYPE
#undef COULOMB_NUM_TYPE
#endif

#define COULOMB_NUM_TYPE double





//############################## FUNCTION #################################

double calc_coulomb_force_substep_raw(COULOMB_NUM_TYPE rj[][3], int qj[], 
                                      COULOMB_NUM_TYPE ri[][3], int qi[],
                                      COULOMB_NUM_TYPE params[], COULOMB_NUM_TYPE fi[][3],
                                      int ni, int nj)
{


  int ii,jj;
  COULOMB_NUM_TYPE norm2,xxx, potencrb2;
  COULOMB_NUM_TYPE dr12[3] ;
  int calcmethod=0; 
  COULOMB_NUM_TYPE prefactor=1;

  char  *coulomb_calc   = g_inputs.coulomb_calc_method ;

  double qi_all = 0 , qj_all = 0 ;

for ( ii = 0 ; ii < ni ; ii++ ) { qi_all += fabs( qi[ii]) ; }
for ( ii = 0 ; ii < nj ; ii++ ) { qj_all += fabs( qj[ii]) ; }

if ( ( 0 == ni*nj ) || ( 0 == qi_all ) || ( 0 == qj_all ) )
{
    for ( ii = 0 ; ii < ni ; ++ii ) { fi[ii][0] = fi[ii][1] = fi[ii][2] = 0 ; }
    return 0 ;
}

calcmethod = 0 ;

potencrb2 = (COULOMB_NUM_TYPE) params[0] ;

calcmethod = 3;
switch (calcmethod)
{

  case 3:
    prefactor = (COULOMB_NUM_TYPE) (  g_phys_math_const.C_k  *  g_phys_math_const.e2  ) ;
    for (ii=0;ii<ni;++ii) {
     for (jj=0;jj<nj;++jj) {
      vectminus(dr12,ri[ii],rj[jj]);
      norm2=vectnorm2(dr12);
      xxx = 1.0 *
            qi[ii]*qj[jj] /
            ( sqrt(norm2+potencrb2) * (norm2+potencrb2) ) ;
      fi[ii][0] += xxx * dr12[0] ;
      fi[ii][1] += xxx * dr12[1] ;
      fi[ii][2] += xxx * dr12[2] ;
     }
    }
    for ( ii = 0 ; ii < ni ; ii++ )  { for ( jj = 0 ; jj < 3 ; jj++ ) {
        fi[ii][jj] *= (COULOMB_NUM_TYPE)prefactor ;
    }}
    break;


    
    default:
      break ;
      
}



  return 0 ;
}


//########################################################################
//########################################################################
//########################################################################


//############################## FUNCTION #################################

int md_integrator_verlet( double tnow , double DT , int *class_parts , int class_parts_dim )
{
  int i,j,ii,pp;

  double (*r1all)[3] , (*r2all)[3] , (*v1all)[3] , (*v2all)[3] , *mall ;
  int *qall , *qall_corr;
  double DT2 ;


  COULOMB_NUM_TYPE (*f1)[3] ,(*f2)[3] , (*xr)[3] , (*xv)[3] , *xm ; 
  COULOMB_NUM_TYPE (*f1_ext)[3] ,(*f2_ext)[3] ;
  int  *xq , *xqc ;
  int dim_tot , Natoms ;
  COULOMB_NUM_TYPE params[1] ;

  double (*xr0)[3] ;

  DT2 = DT * DT ;
  params[0] = (COULOMB_NUM_TYPE)  g_inputs.potencrb2; 

  Natoms = g_inputs.N ;
  r1all = g_system.r1 ;
  r2all = g_system.r2 ;
  v1all = g_system.v1 ;
  v2all = g_system.v2 ;
  qall  = g_system.qi ;
  qall_corr  = g_system.qi_corr ;
  mall  = g_system.m ;
  dim_tot = class_parts_dim ;

  if ( 0 == dim_tot ) { return 0 ; } 


  xr =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  xv =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  xq =  calloc( dim_tot , sizeof(int) );
  xqc =  calloc( dim_tot , sizeof(int) );
  xm =  calloc( dim_tot , sizeof(COULOMB_NUM_TYPE));
  xr0 = calloc( dim_tot , 3*sizeof(double) );


  f1 =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  f2 =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  
  f1_ext =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  f2_ext =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  
  for ( i=0 ; i < dim_tot ; ++i ){
    xm[i] = mall[class_parts[i]];
    xq[i] = qall[class_parts[i]] ;
    xqc[i] = qall_corr[class_parts[i]] ;
    for (j=0;j<3;++j) {
      xr[i][j]  = r1all[class_parts[i]][j] ;
      xv[i][j]  = v1all[class_parts[i]][j] ;
      xr0[i][j] = r1all[class_parts[i]][j] ;
    }
  }


  calc_coulomb_force_substep_raw( xr , xq , xr , xq , params , f1 , dim_tot , dim_tot ) ;



  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      xr[ii][j] += xv[ii][j] * DT + 0.5 * f1[ii][j] / xm[ii] * DT2 ;
  } }


  calc_coulomb_force_substep_raw( xr , xq , xr , xq , params , f2 , dim_tot , dim_tot ) ;


  double W = 0 , fcons_sum , fext_sum ; 
  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      //W += f1[ii][j] * ( r2all[ class_parts[ii] ][j] - r1all[ class_parts[ii] ][j]  ) ; 
      fcons_sum = ( f1[ii][j] - f1_ext[ii][j] ) + ( f2[ii][j] - f2_ext[ii][j] ) ;
      fext_sum  = f1_ext[ii][j] + f2_ext[ii][j] ;
      W += (   xv[ii][j] * fext_sum * DT
             + 0.5  * ( fcons_sum * fext_sum ) * DT * DT
             + 0.25 * ( fext_sum * fext_sum )  * DT * DT ) * 0.5 ;
  }}
  g_ENERGY_VALUES.Work += W ;

 
  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      xv[ii][j] += 0.5 * ( f1[ii][j] + f2[ii][j] ) / xm[ii] * DT ;
  } }


  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      pp = class_parts[ii] ;
      r2all[pp][j] = xr[ii][j] ;
      v2all[pp][j] = xv[ii][j] ;
      r1all[pp][j] = xr0[ii][j] ;
  } }


  free(f1); 
  free(f2); 
  free(f1_ext); 
  free(f2_ext); 
  free(xr);
  free(xv);
  free(xq);
  free(xqc);
  free(xm);

  free(xr0);

  return 0 ;

}



//############################## FUNCTION #################################

int md_integrator_plain_verlet( double DT , int *class_parts , int class_parts_dim )
{
  int i,j,ii,pp;

  double (*r1all)[3] , (*r2all)[3] , (*v1all)[3] , (*v2all)[3] , *mall ;
  int *qall , *qall_corr ;
  double DT2 ;
  static int init = 1 ;

  static COULOMB_NUM_TYPE (*f1)[3] ,(*f2)[3] ;
  COULOMB_NUM_TYPE (*xr)[3] , (*xv)[3] , *xm , (*ff)[3] ;
  int  *xq , *xqc ;
  int dim_tot , Natoms ;
  COULOMB_NUM_TYPE params[1] ;

  DT2 = DT * DT ;
  params[0] = (COULOMB_NUM_TYPE)  g_inputs.potencrb2; 

  Natoms = g_inputs.N ;
  r1all = g_system.r1 ;
  r2all = g_system.r2 ;
  v1all = g_system.v1 ;
  v2all = g_system.v2 ;
  qall  = g_system.qi ;
  qall_corr  = g_system.qi_corr ;
  mall  = g_system.m ;
  dim_tot = class_parts_dim ;


  xr =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  xv =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  xq =  calloc( dim_tot , sizeof(int) );
  xqc =  calloc( dim_tot , sizeof(int) );
  xm =  calloc( dim_tot , sizeof(COULOMB_NUM_TYPE));


  for ( i=0 ; i < dim_tot ; ++i ){
    xm[i] = mall[class_parts[i]];
    xq[i] = qall[class_parts[i]] ;
    xqc[i] = qall_corr[class_parts[i]] ;
    for (j=0;j<3;++j) {
      xr[i][j] = r1all[class_parts[i]][j] ;
      xv[i][j] = v1all[class_parts[i]][j] ;
    }
  }

  if ( 1 == init ) 
  {
      init = 0 ;
      f1 =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
      f2 =  calloc( dim_tot , 3*sizeof(COULOMB_NUM_TYPE) );
  
      calc_coulomb_force_substep_raw( xr , xq , xr , xq , params , f1 , dim_tot , dim_tot ) ;

  }

  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      xr[ii][j] += xv[ii][j] * DT + 0.5 * f1[ii][j] / xm[ii] * DT2 ;
  } }


  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) { f2[ii][j] = 0 ; }}

  calc_coulomb_force_substep_raw( xr , xq , xr , xq , params , f2 , dim_tot , dim_tot ) ;


  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      xv[ii][j] += 0.5 * ( f1[ii][j] + f2[ii][j] ) / xm[ii] * DT ;
  } }


  for ( ii = 0 ; ii < dim_tot ; ++ii ) {   for (j=0;j<3;++j) {
      pp = class_parts[ii] ;
      r2all[pp][j] = xr[ii][j] ;
      v2all[pp][j] = xv[ii][j] ;
  } }

  ff = f1 ; f1 = f2 ; f2 = ff ; // We swap f1 and f2 ; 
  //free(f1); 
  //free(f2); 
  free(xr);
  free(xv);
  free(xq);
  free(xqc);
  free(xm);

  return 0 ;

}



// velocity VERLET:
// F1 = f( g_system.r1 )
// r2 =  g_system.r1  +  g_system.v1 *h + 0.5*F1*h^2/m
// F2 = f(r2)
// v2 = v1 + 0.5*(F1+F2)*h/m

// #############################################################



#ifdef COULOMB_NUM_TYPE
#undef COULOMB_NUM_TYPE
#endif
#define COULOMB_NUM_TYPE double




//############################## FUNCTION #################################

int classdintableinitialisations(char comm){
  if (potcutmanage(comm)==-1){
    potcutmanage('f');
    return(-1);
  }
  return(0);
}


//############################## FUNCTION #################################

int potcutmanage(char comm){

int i;

  double Eintotal;
  int N,Z,E,potcutmaxnum;
  Eintotal= g_ENERGY_VALUES.Eintotal;
  N= g_inputs.N; Z= g_inputs.Z; E= g_inputs.E;

  potcutmaxnum=maxZvalue+1;

switch (comm) {
 case 'i':
  potcut=calloc(potcutmaxnum+1,sizeof(double));
  for (i=0;i<potcutmaxnum+1;++i) {
    potcut[i]=0.0e-10;
  }
  break;
 case 'f':
  free(potcut);
  break;
 }
  return(0);
}






//############################## FUNCTION #################################

double get_part_part_distance( double *a_dr , double *r1 , double *r2 )
{
    double dr[3] ;
    int ii  ;

    for ( ii=0 ; ii<3 ; ii++ ) { dr[ii] = r1[ii] - r2[ii] ; }
    if ( NULL != a_dr ) { for ( ii=0 ; ii<3 ; ii++ ) { a_dr[ii] = dr[ii] ; }}
    return( sqrt( vectnorm2(dr) ) ) ;
}



//############################## FUNCTION #################################

double xmdyn_system_get_particle_distance( int part1 , int part2 , char sel )
{
    double dr[3] ;
    double (*rr)[3] ;
    rr = g_system.r1 ;  if ( 'o' == sel ) { rr = g_system.r2 ; } ;

    return get_part_part_distance( NULL , rr[part1] , rr[part2] ) ;
}




//############################## FUNCTION #################################

int nearestatom2electron( int electron, double **r_at , double *value , char type) {
  int ii, jj , cell , nearestatom;
  double dr[3],mindist2,potvalue;
  int N;
  double r_test[3]  , dist ;

  double (*r_sys)[3] = g_system.r1 ;

  double xxx ;

  N= g_inputs.N;

  //type = 'c'  => Coulomb
  //     = 'd'  => distance

  switch (type) {

  case 'd':
    vectminus(dr, g_system.r1[electron] , g_system.r1 [0]);
    mindist2=vectnorm2(dr);
    nearestatom=0;
    for (ii=0;ii<N;++ii) {
      vectminus(dr, g_system.r1[electron] , g_system.r1 [ii]);
      if (vectnorm2(dr)<mindist2) {
	mindist2=vectnorm2(dr);
	nearestatom=ii;
      }
    }
    break;

  case 'c':
    vectminus( dr , g_system.r1[electron] , g_system.r1[0] ) ;
    potvalue = enerregcut( sqrt(vectnorm2(dr)) , g_system.qi[0] , -1 , 0 ) ;
    //potvalue = enerregcut( dist , g_system.qi[0] , -1 , 0 ) ;
    nearestatom = 0 ;    
    if ( NULL != r_at ) { for ( jj=0 ; jj<3 ; jj++ ) { (*r_at)[jj] = r_sys[0][jj] ; } }

    for ( ii = 0 ; ii < N ; ++ii ) {  
      for ( jj = 0 ; jj < 3 ; jj++ ) { r_test[jj] = r_sys[ii][jj] ; }
      vectminus( dr , g_system.r1[electron] , r_test ) ;
      if ( xxx = enerregcut( sqrt(vectnorm2(dr)) , g_system.qi[ii] , -1 , ii ) , xxx < potvalue ) 
      {
	potvalue = xxx ;
	nearestatom = ii ;   
        if ( NULL != r_at ) { for ( jj=0 ; jj<3 ; jj++ ) { (*r_at)[jj] = r_test[jj] ; } }
      }
    }
    if ( NULL != value ) { *value = potvalue ; }
    break;

  default:
    printf("Wrong nearestatom2electron() type value!\n");
    exit(-1);
  }

  return(nearestatom);
}


// ########################################################
// ###################        ENERGY         ##############
// ########################################################



//############################## FUNCTION #################################

double enerreg(double dist, int q1, int q2){
  return( g_phys_math_const.C_k * g_phys_math_const.e2 *q1*q2/sqrt(dist*dist+ g_inputs.potencrb2));
}



//############################## FUNCTION #################################

double enerregcut(double dist, int q1, int q2,int actatom){

  if (actatom< g_inputs.N) {
    return( (dist>potcut[ g_system.qi [actatom]]) ?
	     g_phys_math_const.C_k * g_phys_math_const.e2 *q1*q2/sqrt(dist*dist+ g_inputs.potencrb2) :
	     g_phys_math_const.C_k * g_phys_math_const.e2 *q1*q2/sqrt((potcut[ g_system.qi [actatom]])*(potcut[ g_system.qi [actatom]])+ g_inputs.potencrb2) );
  }
  else {
    return( g_phys_math_const.C_k * g_phys_math_const.e2 *q1*q2/sqrt(dist*dist+ g_inputs.potencrb2));
  }
}



//############################## FUNCTION #################################

double potenerreal(double dist, int q1, int q2){
  return( g_phys_math_const.C_k * g_phys_math_const.e2 *q1*q2/dist);
}




//############################## FUNCTION #################################

int calc_potE_sys_014( double *a_Epot )
{
    int i , j , ii , jj ;
    int act_j_from , act_j_to ;
    int class_parts_dim , *class_parts ;
    int i_from, j_from, i_to, j_to ;
    double Epot , Etemp ;
    double *Epot_th,*Etemp_th;
    double rel[3] ;

    char  *coulomb_calc   = g_inputs.coulomb_calc_method ;
    double potencrb       = g_inputs.potencrb ;
    int    N = g_inputs.N ;
    int    E = g_inputs.E ;

    Epot = 0 ; *a_Epot = 0 ;

    class_parts_dim=calc_class_parts_dim();

 
    class_parts=calloc(class_parts_dim,sizeof(int));
    set_class_parts(class_parts,class_parts_dim);


    {
    int Num_Threads = 1 ;
    Epot_th=calloc(Num_Threads,sizeof(double));


    for ( ii = 0 ; ii < class_parts_dim ; ii++ )
    {
        double dr3[3] , norm2 ;
        int My_Thread = 0 ;
        for ( jj = 0 ; jj < ii ; jj++ )
        {
            vectminus( dr3 , g_system.r1[class_parts[ii]] , g_system.r1[class_parts[jj]] ) ;
            norm2 = vectnorm2( dr3 ) ;
            Epot_th[My_Thread] += enerreg( sqrt(norm2), g_system.qi[class_parts[ii]], g_system.qi[class_parts[jj]] ) ;
        }
    }
   
    Epot=0;
    for (i=0;i<Num_Threads;++i) {
      Epot+=Epot_th[i];
    }

    free(Epot_th); 
    }

  *a_Epot = Epot ;

  free(class_parts);

  return 0 ;


}





//############################## FUNCTION #################################

double totalpotenergy(int comm) 
{
  int i,j;
  double rel[3],Epot,EpotNL,Etemp;
  int class_parts_dim,*class_parts,i_from,j_from,i_to,j_to;
  int ii,jj,act_j_from,act_j_to;

  double *Epot_th,*Etemp_th;

  char  *coulomb_calc   = g_inputs.coulomb_calc_method ;
  double potencrb = g_inputs.potencrb ;
  int N = g_inputs.N ;
  int E = g_inputs.E ;


  int use_method;

  if ( strstr(  g_inputs.misc_opt , "no-energy-calc" ) != NULL ) {
      return(0) ;
  }


  Epot=0;  

  use_method=0;

  if (comm==12)  
  {   
      calc_potE_sys_014( &Epot ) ;   
  }


  return(Epot);
}




//############################## FUNCTION #################################

double potEactual(int atom,int el1,int el2, int n_atoms_affected , int *atoms_affected ) {
  int i,j;
  double Epot,EpotAA,Epot1,Epot2,Epot3,EpotNL=0,rel[3];
  int N,E;
  int nonloc;


  if ( strstr(  g_inputs.misc_opt , "no-energy-calc" ) != NULL ) {
      return(0) ;
  }

  N= g_inputs.N;E= g_inputs.E;

  Epot=EpotAA=EpotNL=Epot1=Epot2=Epot3=0;

  nonloc=0;
  
  if (atom<N) {
    
    if (nonloc) {
      for (i=0;i<N-1;++i) {
        for (j=i+1;j<N;++j) {
          vectminus(rel, g_system.r1 [i], g_system.r1 [j]);
          EpotAA+=enerreg(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [j]);
        }
      }
    }

    else {
      for (i=0;i<N;++i) {
        if (i!=atom) {
          vectminus(rel, g_system.r1 [i], g_system.r1 [atom]);
          EpotAA+=enerreg(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [atom]);
        }
      }
    }
    
  }



  
  for (i = (atom<N ? N : 0); i<N+E; ++i) {
    if ( ( (( g_system.qi [i]!=0) || (i<N)) && (i!=atom) ) &&
         ( g_system.ptype [i]!=  g_inputs.nots) && ( g_system.ptype [i]!=  g_inputs.noall) ) {
      vectminus(rel, g_system.r1 [i], g_system.r1 [atom]);
      if  ( (i<N) && (atom>N-1) )
	Epot1+=enerregcut(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [atom],i);
      else
	Epot1+=enerreg(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [atom]);
    }
  }
  
  if (el1>=0) {
    for (i=0;i<N+E;++i) {
      if ( ( (( g_system.qi [i]!=0) || (i<N)) && (i!=atom) && (i!=el1) ) && 
           ( g_system.ptype [i]!=  g_inputs.nots) && ( g_system.ptype [i]!=  g_inputs.noall) ) {
	vectminus(rel, g_system.r1 [i], g_system.r1 [el1]);
	if  ( (i<N) && (el1>N-1) )
	  Epot2+=enerregcut(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [el1],i);
	else
	  Epot2+=enerreg(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [el1]);
      }
    }
  }

  if (el2>=0) {
    for (i=0;i<N+E;++i) {
      if ( ( (( g_system.qi [i]!=0) || (i<N)) && (i!=atom) && (i!=el1) && (i!=el2) ) &&
           ( g_system.ptype [i]!=  g_inputs.nots) && ( g_system.ptype [i]!=  g_inputs.noall) ) {
	vectminus(rel, g_system.r1 [i], g_system.r1 [el2]);
	if  ( (i<N) && (el2>N-1) )
	  Epot3+=enerregcut(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [el2],i);
	else
	  Epot3+=enerreg(sqrt(vectnorm2(rel)), g_system.qi [i], g_system.qi [el2]);
      }
    }
  }
  
  Epot += ( EpotAA + EpotNL + Epot1 + Epot2 + Epot3 ) ;

  return(Epot);
}

#define ENERGY_VARIABLE  g_ENERGY_VALUES

//############################## FUNCTION #################################

double xmdyn_md_energy(char *a_comm , char *a_option )
{
  static int l_step=-1 ;
  int sstep = 0 ;
  static double storedE[20] = { 0 } ;
  double returnE ;
  int ii , jj , calc_forced = 0 ;

  if ( strstr(  g_inputs.misc_opt , "no-energy-calc" ) != NULL ) { return(0) ; }

  //* Debug/Monitor purpose
  if ( strcmp( a_comm , "test" ) == 0 )
  {
      if ( strstr( a_option , "Ediff-cache" )  != NULL ) 
      { 
          returnE = xmdyn_md_energy( "system-E" , "getcache" ) - 
                    ( g_ENERGY_VALUES.Eintotal + g_ENERGY_VALUES.Einitial ) ; 
      }
      if ( strstr( a_option , "Ediff-calc" )  != NULL ) 
      { 
          returnE = totalpotenergy(12) + xmdyn_md_energy( "system-K" , "calc" ) - 
                    ( g_ENERGY_VALUES.Eintotal  + g_ENERGY_VALUES.Einitial ) ; 
      }
      if ( strstr( a_option , "Etot-calc" )  != NULL ) 
          { returnE = totalpotenergy(12) + xmdyn_md_energy( "system-K" , "calc" ) ; }
      if ( strstr( a_option , "eV" )  != NULL ) { returnE = convert_unit_J2eV( returnE ) ; }
      if ( strstr( a_option , "rel" ) != NULL ) 
      { 
          if ( 0 == g_ENERGY_VALUES.Eintotal ) { returnE = -99 ; }
          else { returnE = returnE / g_ENERGY_VALUES.Eintotal ; }
      }
      return ( returnE ) ;
  }


  if ( strcmp( a_comm , "store" ) == 0 )
  {
      storedE[0] = ENERGY_VARIABLE.totalpotE ;
      storedE[1] = ENERGY_VARIABLE.KE ;
      storedE[0] = g_ENERGY_VALUES.totalpotE ;
      storedE[1] = g_ENERGY_VALUES.KE ;
      //printf( "DBG2 %le\n" ,  ENERGY_VARIABLE.totalpotE/1.6022e-19 ) ;
      return 0 ;
  }

  if ( strcmp( a_comm , "restore" ) == 0 )
  {
      g_ENERGY_VALUES.totalpotE = storedE[0] ;
      g_ENERGY_VALUES.KE = storedE[1] ;
      return 0 ;
  }

  returnE = 0 ;

  if ( strstr( a_comm , "system" ) != NULL )
  {

      if ( ( ( strstr( a_option , "calc" ) != NULL ) ||
             ( g_inputs.step != l_step ) ) && 
             ( strstr( a_option , "getcache" ) == NULL )
         )
      {
          calc_forced = 1 ;
      }

      if ( ( strstr( a_comm , "K" ) != NULL ) || ( strstr( a_comm , "E" ) != NULL ) )
      {
          {
              ENERGY_VARIABLE.KE = 0 ;
              for ( ii=0 ; ii< g_inputs.N + g_inputs.E; ++ii ) { for ( jj = 0 ; jj < 3 ;++jj ) {
                  ENERGY_VARIABLE.KE += ( 0.5 * g_system.m[ii]* g_system.v1[ii][jj] * g_system.v1[ii][jj] ) ;
              }}
          }
          returnE += ENERGY_VARIABLE.KE ; 
      }

      if ( ( strstr( a_comm , "V" ) != NULL ) || ( strstr( a_comm , "E" ) != NULL ) )
      {
          if ( calc_forced ) 
          {  
              if ( strstr( a_option , "calc" ) == NULL ) 
              { 
              }
              else { 
              }
              ENERGY_VARIABLE.totalpotE = totalpotenergy(12) ;  
              if ( strstr( a_option , "nextstep" ) != NULL ) 
                   {  l_step = g_inputs.step + 1 ; }
              else {  l_step = g_inputs.step     ; }
          }
          else { ENERGY_VARIABLE.totalpotE = g_ENERGY_VALUES.totalpotE ; }
          returnE += ENERGY_VARIABLE.totalpotE ; 
      }

      return( returnE ) ;
  }

  MODULE_ERROR( "Wrong argument." , -1 ) ;
  return 0 ;

}

#undef ENERGY_VARIABLE

// #####################################################
// #####################################################
// #################  DELOC TEST  ######################
// #####################################################



//############################## FUNCTION #################################

double nnbor_condition(int el,int at,char comm){
  double dr[3] , dist ;

  {
      vectminus(dr, g_system.r1 [at], g_system.r1 [el]);
      dist = sqrt( vectnorm2(dr) ) ;
  }

  switch (comm){
  case 'n':
    return( dist * dist );
    break;
  case 'c':
    return( enerreg(dist, g_system.qi [el], g_system.qi [at]) ) ;
    break;
  }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0 ;
}

//############################## FUNCTION #################################

int* get_nnbors(int el,int numnbors,char comm,int init){
  static int *nbors_list, first_call=1;
  static double *nbors_dist;
  int i,j,k,jj;
  double act_cond;
  
  init=1;
  
  if (first_call){
    first_call=0;
    nbors_list=calloc(numnbors,sizeof(int));
    nbors_dist=calloc(numnbors,sizeof(double));
  }
  if (init){
    for (i=0;i<numnbors;++i){
      nbors_list[i]=-1;
      nbors_dist[i]=0;
    }
    for (i=0;i< g_inputs.N;++i){
      if ( g_system.qi [i]!=0) {
        act_cond=nnbor_condition(el,i,comm);
        j=jj=0;
        while ( (jj<numnbors) && (jj< g_inputs.N-1) && (nbors_list[jj]!=-1) ){
          if ( (act_cond>nbors_dist[jj]) ) {
              j++;
           } 
           jj++;
        }
        for ( k=((numnbors< g_inputs.N)?numnbors-1: g_inputs.N-1); k>j; --k) {
          nbors_list[k]=nbors_list[k-1];
          nbors_dist[k]=nbors_dist[k-1];
        }
        if ((j<numnbors)&&(j< g_inputs.N)) {
          nbors_list[j]=i;
          nbors_dist[j]=act_cond;
        }
      }
    }
  }
  return(nbors_list);
}

//############################## FUNCTION #################################

int e_deloc_pair(int el,int at1,int at2,int id,int numnbors,
                 double pot0,double T,int init){
  static double memorel[3],potE,*maxEv;
  static int first_call=1;
  int i,j,np;
  double drat[3],rx[3],maxE;
  
  init=1;
  
  if (first_call){
    first_call=0;
    maxEv=calloc(numnbors,sizeof(double));
  }
  vect3equal(memorel, g_system.r1 [el]);
  vectminus(drat, g_system.r1 [at2], g_system.r1 [at1]);
  np= g_inputs.deloctest_ninterp*ceil(sqrt(vectnorm2(drat))/.1e-10)+3; // >3 always
  maxE=0;
  for (i=0;i<np;++i){
    for (j=0;j<3;++j){
      rx[j] =  g_system.r1 [at1][j] + ((double)i/(double)(np-1))*( g_system.r1 [at2][j]- g_system.r1 [at1][j]);
    }
    vect3equal( g_system.r1 [el],rx);
    if ( ( (potE=potEactual(el,-1,-1,0,NULL),potE>maxE) && (maxE!=0) ) || (maxE==0) ) {
      maxE=potE;
    }
  }
  vect3equal( g_system.r1 [el],memorel);
  if (T+pot0-maxE>0) {
    return(1);
  }
  else{
    return(0);
  }
}

//############################## FUNCTION #################################

int e_deloc_nbors(int el,int numnbors,int *nbors_list,char comm,double T,int init) {
  int i,result;
  double pot0,dr[3];

  result=1;
  pot0=potEactual(el,-1,-1,0,NULL);
  if (numnbors==0){
    return(1);
  }
  if (numnbors==1) {
    vectminus(dr, g_system.r1 [nbors_list[0]], g_system.r1 [el]);
    if (T + enerreg(sqrt(vectnorm2(dr)), g_system.qi [nbors_list[0]], g_system.qi [el])>0) {
      return(1);
    }
    else{
      return(0);
    }
  }
  switch (comm){
  case 'c':
    for (i=1;i<numnbors;++i){
      result*=e_deloc_pair(el,nbors_list[0],nbors_list[i],i,numnbors,pot0,T,init);
    }
    break;
  case 'n':
    break;
  }
  return(result);
}

//############################## FUNCTION #################################

int e_deloc_test(int el,double T,int init){
  int *nbors_list,i,numnbors, x;
  
  nbors_list=get_nnbors(el, g_inputs.deloctest_nn, g_inputs.deloctest_comm,init);
  numnbors=0;
  for (i=0;i< g_inputs.deloctest_nn;++i){
    if (nbors_list[i]!=-1){
      numnbors++;
    }
  }
x = e_deloc_nbors(el,numnbors,nbors_list, g_inputs.deloctest_comm,T,init);
  return(x-1);
  // so return value is 0/-1 instead of 1/0
}


// ########################################################
// ########################################################
// ########################################################




//############################## FUNCTION #################################

int testsuite_classdyn( char *comm )
{
    printf( "ENTERING testsuite_classdyn\n" ) ;
    int ii ; 
    for ( ii = 0 ; ii < maxZvalue ; ii++ ) 
        printf( "PotCut % 3d : %le\n" , ii , potcut[ii] ) ; 
    printf( "LEAVING testsuite_classdyn\n" ) ;

  return 0 ;
}



