///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>

#include "common.h"

#include "globalconsts.h"

#include "Xatom.h"

#include "xmdyn_io.h"

#include "endltabs.h"
#include "classdyn.h"
#include "tests.h"


#include "init_system.h"
#include "xmdyn_filenames.h"

#include "secel.h"

// ########################################################

//* NOTE: to be uncommented to deactivate 4f orbital in order to speed up the calculation  *//
//#define  WORKAROUND_XE_NO_4f

// ########################################################

#define ATOMDATA_TABLE_XATOM 

// ########################################################


#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}


// ########################################################

extern int getatomicdata(int Z) ;
extern int correctatomicdata(struct ATOMDATA_STRUCT *act,int Z) ;


#ifdef ATOMDATA_TABLE_XATOM



static int eqshellnum=24;
static int eqshell[][4] = {
			   { 0, 0, 0, 0},
			   { 1, 0, 2, 0},
			   { 3, 0, 2, 0},
			   { 5, 0, 6, 0},
			   { 8, 0, 2, 0},
			   {10, 0, 6, 0},
			   {13, 0,10, 0},
			   {16, 0, 2, 0},
			   {18, 0, 6, 0},
			   {21, 0,10, 0},
			   {24, 0,14, 0},
			   {27, 0, 2, 0},
			   {29, 0, 6, 0},
			   {32, 0,10, 0},
			   {35, 0,14, 0},
			   {38, 0,18, 0},
			   {41, 0, 2, 0},
			   {43, 0, 6, 0},
			   {46, 0,10, 0},
			   {49, 0,14, 0},
			   {52, 0,18, 0},
			   {55, 0,22, 0},
			   {58, 0, 2, 0},
			   {60, 0, 6, 0}
}; //{subshell1,subshell2,max_num_of_el_subshell1,max_num_of_el_subshell2}


static int eqshell_all[][4] = {
			   { 0, 0, 0, 0},
			   { 1, 0, 2, 0},
			   { 2, 0, 0, 0},
			   { 3, 0, 2, 0},
			   { 4, 0, 0, 0},
                     
			   { 5, 6, 6, 0},
			   { 6, 5, 0, 6},
                     
			   { 7, 0, 0, 0},
			   { 8, 0, 2, 0},
			   { 9, 0, 0, 0},
                     
			   {10,11, 6, 0},
			   {11,10, 0, 6},
                     
			   {12, 0, 0, 0},
                     
			   {13,14,10, 0},
			   {14,13, 0,10},
                     
			   {15, 0, 0, 0},
			   {16, 0, 2, 0},
			   {17, 0, 0, 0},
                     
			   {18,19, 6, 0},
			   {19,18, 0, 6},
                     
			   {20, 0, 0, 0},
                     
			   {21,22,10, 0},
			   {22,21, 0,10},
                     
			   {23, 0, 0, 0},
                     
			   {24,25,14, 0},
			   {25,24, 0,14},
                     
			   {26, 0, 0, 0},
			   {27, 0, 2, 0},
			   {28, 0, 0, 0},
                     
			   {29,30, 6, 0},
			   {30,29, 0, 6},
                     
			   {31, 0, 0, 0},
                     
			   {32,33,10, 0},
			   {33,32, 0,10},
                     
			   {34, 0, 0, 0},
                     
			   {35,36,14, 0},
			   {36,35, 0,14},
                     
			   {37, 0, 0, 0},
                     
			   {38,39,18, 0},
			   {39,38, 0,18},
                     
			   {40, 0, 0, 0},
			   {41, 0, 2, 0},
			   {42, 0, 0, 0},
                     
			   {43,44, 6, 0},
			   {44,43, 0, 6},
                     
			   {45, 0, 0, 0},
                     
			   {46,47,10, 0},
			   {47,46, 0,10},
                     
			   {48, 0, 0, 0},
                     
			   {49,50,14, 0},
			   {50,49, 0,14},
                     
			   {51, 0, 0, 0},
                     
			   {52,53,18, 0},
			   {53,52, 0,18},
                     
			   {54, 0, 0, 0},
                     
			   {55,56,22, 0},
			   {56,55, 0,22},
                     
			   {57, 0, 0, 0},
			   {58, 0, 2, 0},
			   {59, 0, 0, 0},
                     
			   {60,61, 6, 0},
			   {61,60, 0, 6}
}; //{subshell1,subshell2,max_num_of_el_subshell1,max_num_of_el_subshell2}


#else		// EADL format

static int eqshellnum=24;
static int eqshell[][4] = {
			   { 0, 0, 0, 0},
			   { 1, 0, 2, 0},
			   { 3, 0, 2, 0},
			   { 5, 6, 2, 4},
			   { 8, 0, 2, 0},
			   {10,11, 2, 4},
			   {13,14, 4, 6},
			   {16, 0, 2, 0},
			   {18,19, 2, 4},
			   {21,22, 4, 6},
			   {24,25, 6, 8},
			   {27, 0, 2, 0},
			   {29,30, 2, 4},
			   {32,33, 4, 6},
			   {35,36, 6, 8},
			   {38,39, 8,10},
			   {41, 0, 2, 0},
			   {43,44, 2, 4},
			   {46,47, 4, 6},
			   {49,50, 6, 8},
			   {52,53, 8,10},
			   {55,56,10,12},
			   {58, 0, 2, 0},
			   {60,61, 2, 4}
}; //{subshell1,subshell2,max_num_of_el_subshell1,max_num_of_el_subshell2}


static int eqshell_all[][4] = {
			   { 0, 0, 0, 0},
			   { 1, 0, 2, 0},
			   { 2, 0, 0, 0},
			   { 3, 0, 2, 0},
			   { 4, 0, 0, 0},

			   { 5, 6, 2, 4},
			   { 6, 5, 2, 4},

			   { 7, 0, 0, 0},
			   { 8, 0, 2, 0},
			   { 9, 0, 0, 0},

			   {10,11, 2, 4},
			   {11,10, 2, 4},

			   {12, 0, 0, 0},

			   {13,14, 4, 6},
			   {14,13, 4, 6},

			   {15, 0, 0, 0},
			   {16, 0, 2, 0},
			   {17, 0, 0, 0},

			   {18,19, 2, 4},
			   {19,18, 2, 4},

			   {20, 0, 0, 0},

			   {21,22, 4, 6},
			   {22,21, 4, 6},

			   {23, 0, 0, 0},

			   {24,25, 6, 8},
			   {25,24, 6, 8},

			   {26, 0, 0, 0},
			   {27, 0, 2, 0},
			   {28, 0, 0, 0},

			   {29,30, 2, 4},
			   {30,29, 2, 4},

			   {31, 0, 0, 0},

			   {32,33, 4, 6},
			   {33,32, 4, 6},

			   {34, 0, 0, 0},

			   {35,36, 6, 8},
			   {36,35, 6, 8},

			   {37, 0, 0, 0},

			   {38,39, 8,10},
			   {39,38, 8,10},

			   {40, 0, 0, 0},
			   {41, 0, 2, 0},
			   {42, 0, 0, 0},

			   {43,44, 2, 4},
			   {44,43, 2, 4},

			   {45, 0, 0, 0},

			   {46,47, 4, 6},
			   {47,46, 4, 6},

			   {48, 0, 0, 0},

			   {49,50, 6, 8},
			   {50,49, 6, 8},

			   {51, 0, 0, 0},

			   {52,53, 8,10},
			   {53,52, 8,10},

			   {54, 0, 0, 0},

			   {55,56,10,12},
			   {56,55,10,12},

			   {57, 0, 0, 0},
			   {58, 0, 2, 0},
			   {59, 0, 0, 0},

			   {60,61, 2, 4},
			   {61,60, 2, 4}
}; //{subshell1,subshell2,max_num_of_el_subshell1,max_num_of_el_subshell2}

#endif

static int shellid_qnum_dim = 39 ;
static double shellid_qnum[][4] = {
                                    { 1, 1, 0, 1/2.0},   
                                    { 3, 2, 0, 1/2.0},  
                                    { 5, 2, 1, 1/2.0},  
                                    { 6, 2, 1, 3/2.0},  
                                    { 8, 3, 0, 1/2.0},     
                                    {10, 3, 1, 1/2.0},  
                                    {11, 3, 1, 3/2.0},   
                                    {13, 3, 2, 3/2.0},  
                                    {14, 3, 2, 5/2.0},   
                                    {16, 4, 0, 1/2.0},  
                                    {18, 4, 1, 1/2.0},  
                                    {19, 4, 1, 3/2.0},  
                                    {21, 4, 2, 3/2.0},  
                                    {22, 4, 2, 5/2.0},   
                                    {24, 4, 3, 5/2.0},  
                                    {25, 4, 3, 7/2.0},   
                                    {27, 5, 0, 1/2.0},  
                                    {29, 5, 1, 1/2.0},  
                                    {30, 5, 1, 3/2.0},  
                                    {32, 5, 2, 3/2.0},  
                                    {33, 5, 2, 5/2.0},  
                                    {35, 5, 3, 5/2.0},  
                                    {36, 5, 3, 7/2.0},  
                                    {38, 5, 4, 7/2.0},  
                                    {39, 5, 4, 9/2.0},   
                                    {41, 6, 0, 1/2.0}, 
                                    {43, 6, 1, 1/2.0}, 
                                    {44, 6, 1, 3/2.0}, 
                                    {46, 6, 2, 3/2.0}, 
                                    {47, 6, 2, 5/2.0}, 
                                    {49, 6, 3, 5/2.0}, 
                                    {50, 6, 3, 7/2.0}, 
                                    {52, 6, 4, 7/2.0}, 
                                    {53, 6, 4, 9/2.0}, 
                                    {55, 6, 5, 9/2.0}, 
                                    {56, 6, 5, 11/2.0}, 
                                    {58, 7, 0, 1/2.0}, 
                                    {60, 7, 1, 1/2.0},   
                                    {61, 7, 1, 3/2.0} 
};






// ########################################################
// ########################################################
// ########################################################





//############################## FUNCTION #################################

int getallatomicdata(int *Z){
  int i;
  static int warning_missing_xatom_tab = 1 , warning_missing_xatom_KE = 1 ;

for (i=0;i< g_inputs.N;++i){

    getatomicdata(Z[i]);

    if ( XATOM_TAB == g_inputs.atomic_data_source )
    {
    }

  }

return(0);
}



//############################## FUNCTION #################################

int getatomicdata(int Z){
  static int tabsinit=0;
  struct ATOMDATA_STRUCT *act;
  int i;
  int h,w;
  int tmpi , *ptmpi ;

  if (tabsinit==0) {
    ATOMDATA=calloc(maxZvalue+1,sizeof(struct ATOMDATA_STRUCT));
    for (i=0;i<maxZvalue+1;++i){
      ATOMDATA[i].maxshellnum=0;
    }
    tabsinit=1;
  }
  act=ATOMDATA+Z;

  if (act->maxshellnum>0) {
    return(2);
  }


  xmdyn_get_econfig_properties( "Z" , Z , 2 , "max_orbit" , &tmpi ) ; 
  xmdyn_get_econfig_properties( "Z" , Z , 2 , "occupancy" , &ptmpi ) ;

  act->maxshellnum = tmpi ;
  act->shell=calloc(act->maxshellnum+1,sizeof(struct SHELLDATA_STRUCT));
  for ( i=0 ; i < (act->maxshellnum) + 1 ; ++i ) { act->shell[i].NE = ptmpi[i] ; } ;

  return 0 ;
}


//############################## FUNCTION #################################

int subshell_maxenum(int shell)
{

  if ( ( XATOM_TAB == g_inputs.atomic_data_source ) || ( XATOM_ONTHEFLY == g_inputs.atomic_data_source ) )
  {
      switch (shell) {
       // 1s
       case 1:
        return(2);
       // 2s
       case 3:
        return(2);
       // 2p
       case 5:
        return(6);
       case 6:
        return(0);
       // 3s
       case 8:
        return(2);
       // 3p
       case 10:
        return(6);
       case 11:
        return(0);
       // 3d
       case 13:
        return(10);
       case 14:
        return(0);
       // 4s
       case 16:
        return(2);
       // 4p
       case 18:
        return(6);
       case 19:
        return(0);
       // 4d
       case 21:
        return(10);
       case 22:
        return(0);
       // 4f
       case 24:
        return(14);
       case 25:
        return(0);
       // 5s
       case 27:
        return(2);
       // 5p
       case 29:
        return(6);
       case 30:
        return(0);
       // 5d
       case 32:
        return(10);
       case 33:
        return(0);
       // 5f
       case 35:
        return(14);
       case 36:
        return(0);
       // 5g
       case 38:
        return(18);
       case 39:
        return(0);
       // 6s
       case 41:
        return(2);
       // 6p
       case 43:
        return(6);
       case 44:
        return(0);
       // 6d
       case 46:
        return(10);
       case 47:
        return(0);
       // 6f
       case 49:
        return(14);
       case 50:
        return(0);
       // 6g
       case 52:
        return(18);
       case 53:
        return(0);
       // 6h
       case 55:
        return(22);
       case 56:
        return(0);
       // 7s
       case 58:
        return(2);
       // 7p
       case 60:
        return(6);
       case 61:
        return(0);
       default:
        return(0);
      }
  }

  else
  {
      switch (shell) {
       // 1s
       case 1:
        return(2);
       // 2s
       case 3:
        return(2);
       // 2p
       case 5:
        return(2);
       case 6:
        return(4);
       // 3s
       case 8:
        return(2);
       // 3p
       case 10:
        return(2);
       case 11:
        return(4);
       // 3d
       case 13:
        return(4);
       case 14:
        return(6);
       // 4s
       case 16:
        return(2);
       // 4p
       case 18:
        return(2);
       case 19:
        return(4);
       // 4d
       case 21:
        return(4);
       case 22:
        return(6);
       // 4f
       case 24:
        return(6);
       case 25:
        return(8);
       // 5s
       case 27:
        return(2);
       // 5p
       case 29:
        return(2);
       case 30:
        return(4);
       // 5d
       case 32:
        return(4);
       case 33:
        return(6);
       // 5f
       case 35:
        return(6);
       case 36:
        return(8);
       // 5g
       case 38:
        return(8);
       case 39:
        return(10);
       // 6s
       case 41:
        return(2);
       // 6p
       case 43:
        return(2);
       case 44:
        return(4);
       // 6d
       case 46:
        return(4);
       case 47:
        return(6);
       // 6f
       case 49:
        return(6);
       case 50:
        return(8);
       // 6g
       case 52:
        return(8);
       case 53:
        return(10);
       // 6h
       case 55:
        return(10);
       case 56:
        return(12);
       // 7s
       case 58:
        return(2);
       // 7p
       case 60:
        return(2);
       case 61:
        return(4);
       default:
        return(0);
      }
  }

}


//############################## FUNCTION #################################

int shellholes(int NE,int shell){
  // TODO: get rid of this
  return(subshell_maxenum(shell)-NE);
}


//############################## FUNCTION #################################

int cumsums_shell_states(int Z){
  int i,numel;
  numel=0;
  for (i=1;i<=ATOMDATA[Z].maxshellnum;++i){
    if ( subshell_maxenum(i) > 0 )
      numel+=subshell_maxenum(i);
  }
  return(numel);
}

//############################## FUNCTION #################################

int get_equviv_shell(int shell1){
  return( eqshell_all[ shell1 ][1] ) ;

  int i;
  for (i=0;i<eqshellnum;++i){
    if ( (shell1==eqshell[i][0]) || (shell1==eqshell[i][1]) ){
      if (shell1==eqshell[i][0])
        return(eqshell[i][1]);
      else
        return(eqshell[i][0]);
    }
  }
  return(0);
}

//############################## FUNCTION #################################

char subshellint2char(int subshell){
  if (subshell<=1)
    return('K');
  else if (subshell<=6)
    return('L');
  else if (subshell<=14)
    return('M');
  else if (subshell<=25)
    return('N');
  else if (subshell<=39)
    return('O');
  else if (subshell<=56)
    return('P');
  else if (subshell<=61)
    return('Q');

  return 'Z' ;
}


//############################## FUNCTION #################################

int n_quantum_number(int subshell){
  if (subshell==0)
    return(0);
  else if (subshell<=1)
    return(1);
  else if (subshell<=6)
    return(2);
  else if (subshell<=14)
    return(3);
  else if (subshell<=25)
    return(4);
  else if (subshell<=39)
    return(5);
  else if (subshell<=56)
    return(6);
  else if (subshell<=61)
    return(7);
  else
    return(0);
}


//############################## FUNCTION #################################

double subshell2qnum(int shell,int qnum) {
  int i;
  for (i=0;i<shellid_qnum_dim;++i) {
    if (shellid_qnum[i][0]==shell) {
      return(shellid_qnum[i][qnum]);
    }  
  } 
  return(-1);
}


//############################## FUNCTION #################################

int highest_unoccupied_shell(int at)
{
    int vacancy_shell = g_atomic_data[at].maxshellnum ;
    while ( ( g_atomic_data[at].shell[vacancy_shell].NE == 
              subshell_maxenum( vacancy_shell ) 
            )
            && ( vacancy_shell > 0 ) 
          )
    {
        vacancy_shell-- ;
    }
    return vacancy_shell ;
}


//############################## FUNCTION #################################

int setsubshellNE(int Z,int subshell){
/*  You have to init the equviv subshells directly one after the other!!!
    First the one with smaller shell-id!!!  */
  static int n1,n2,maxn1,maxn2,nsum;
  int i,j,n1start;
  double d,probsum;


  for (i=0;i<eqshellnum;++i){
    for (j=0;j<2;++j){
      if (subshell==eqshell[i][j]){
        if (j==0){
	  if (eqshell[i][1]==0){
	    return(ATOMDATA[Z].shell[subshell].NE);
	  }
	  else {
            nsum=(int)(ATOMDATA[Z].shell[eqshell[i][0]].NE+
                       ATOMDATA[Z].shell[eqshell[i][1]].NE);
            maxn1=eqshell[i][2];maxn2=eqshell[i][3];
	    d=randnum(0);probsum=0;
	    n1start= (maxn2-nsum>=0) ? -1 : (nsum-maxn2-1) ;
	    n1=n1start;
            while ( (probsum<d) && (n1<maxn1) && (n1<nsum) ){
              n1++;
	      probsum+=(binom(maxn1,n1)*binom(maxn2,nsum-n1)/
	               (double)binom(maxn1+maxn2,nsum));
	    }
	    if (n1==n1start)
	      n1++;
	    n2=nsum-n1;
	    return(n1);
	  }
	}
	else {
	  return(n2); // it was set just at the previous call; 'n2' is static! 
	}
      }
    }
  }


  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0 ;
}


//############################## FUNCTION #################################

int getrand_shell_el(int actatom,int actshell){
  int i,i_rand,counter;
  //i_rand=floor(g_atomic_data[actatom].shell[actshell].NE * randnum(0));
   if ( 0 == g_atomic_data[actatom].atomic_econfig->NE_orbit[actshell] )
      { MODULE_ERROR( "Search for electron when there is none." , -1 ); }

  i_rand = floor( g_atomic_data[actatom].atomic_econfig->NE_orbit[actshell] * randnum(0) );
  counter=0;
  i=g_atomic_data[actatom].shell[actshell].e_from;
  while ( (counter<=i_rand) &&
                           (counter<=g_atomic_data[actatom].shell[actshell].e_to) ){
    if (isitshellelectron(i))
      counter++;
    if (counter<=i_rand)
      i++;
  }
  if (i>g_atomic_data[actatom].shell[actshell].e_to)
    return(-1);  // error
  else
    return(i);  // OK
}


//############################## FUNCTION #################################

int getrand_shell_hole(int actatom,int actshell){
  int i,i_rand,counter,numofholes;
  numofholes =  shellholes(g_atomic_data[actatom].shell[actshell].NE,actshell) ;
   if ( 0 == numofholes )
      { MODULE_ERROR( "Search for hole when there is none." , -1 ); }

  i_rand=floor( numofholes * randnum(0) );
  counter=0;
  i=g_atomic_data[actatom].shell[actshell].e_from;
  while ( (counter<=i_rand) &&
                           (counter<=g_atomic_data[actatom].shell[actshell].e_to) ){
    if (!isitshellelectron(i))
      counter++;
    if (counter<=i_rand)
      i++;
  }
  if (i>g_atomic_data[actatom].shell[actshell].e_to)
    return(-1);  // error
  else
    return(i);  // OK
}


//############################## FUNCTION #################################

int get_hole_on_shell(int actatom,int actshell){
  int i,i_rand,counter,numofholes;
  numofholes =  shellholes(g_atomic_data[actatom].shell[actshell].NE,actshell) ;
   if ( 0 == numofholes )
      { MODULE_ERROR( "Search for hole when there is none." , -1 ); }
  i_rand = 0 ; //floor( numofholes * randnum(0) );
  counter=0;
  i=g_atomic_data[actatom].shell[actshell].e_from;
  while ( (counter<=i_rand) &&
                           (counter<=g_atomic_data[actatom].shell[actshell].e_to) ){
    if (!isitshellelectron(i))
      counter++;
    if (counter<=i_rand)
      i++;
  }
  if (i>g_atomic_data[actatom].shell[actshell].e_to)
    return(-1);  // error
  else
    return(i);  // OK
}


//############################## FUNCTION #################################

int isitfreeelectron(int acte){
  if ( ( g_system.ptype [acte]=='f') || ( g_system.ptype [acte]=='g') || 
       ( g_system.ptype [acte]=='a') || 
       ( g_system.ptype [acte]=='s') || ( g_system.ptype [acte]=='t') || 
       ( g_system.ptype [acte]=='u') || ( g_system.ptype [acte]=='v') ||
       ( g_system.ptype [acte]=='l') ||
       ( g_system.ptype [acte]=='e')
       // add new type of electron here!
     ) {
    return(1);
  }
  else{
    return(0);
  }
}

//############################## FUNCTION #################################

int isitshellelectron(int acte){
  if ( ( g_system.ptype [acte]=='K') || ( g_system.ptype [acte]=='L') || ( g_system.ptype [acte]=='M') || \
       ( g_system.ptype [acte]=='N') || ( g_system.ptype [acte]=='O') || ( g_system.ptype [acte]=='P') || \
       ( g_system.ptype [acte]=='Q') || ( g_system.ptype [acte]=='Z') ){
    return(1);
  }
  else{
    return(0);
  }
}

//############################## FUNCTION #################################

int is_real_subshell(int shell){
  int i;
  if (shell==0)
    return(0);
  for (i=0;i<eqshellnum;++i){
    if ( (shell==eqshell[i][0]) || (shell==eqshell[i][1]) )
      return(1);
  }
  return(0);
}

//############################## FUNCTION #################################

int particle_params_swap(int p1,int p2){
  int i;

  swap_char( g_system.ptype [p1], g_system.ptype [p2]);
  swap_int( g_system.qi [p1], g_system.qi [p2]);
  swap_double( g_system.qd [p1], g_system.qd [p2]);
  swap_int( g_system.qi_corr [p1], g_system.qi_corr [p2]);
  swap_int( g_system .motherion[p1] , g_system .motherion[p2] ) ;
  for (i=0;i<3;++i){
    swap_double( g_system.r1 [p1][i], g_system.r1 [p2][i]);
    swap_double( g_system.r2 [p1][i], g_system.r2 [p2][i]);
    swap_double( g_system.v1 [p1][i], g_system.v1 [p2][i]);
    swap_double( g_system.v2 [p1][i], g_system.v2 [p2][i]);
    swap_double( g_system.prev_r2 [p1][i], g_system.prev_r2 [p2][i]);
  }
  instant_recombination_swap( p1 , p2 ) ;
  SWAP_VAR( g_system.uid[p1] ,  g_system.uid[p2] , long long int ) ;

  return 0 ;
}



//############################## FUNCTION #################################

int particle_rvq_zero(int p){
  int i;

   g_system.qi[p] = 0;
   g_system.qd[p] = 0;
   g_system.qi_corr[p] = 0;
  for (i=0;i<3;++i){
     g_system.r1[p][i] = g_system.r2[p][i] = 
     g_system.v1[p][i] = g_system.v2[p][i] =
     g_system.prev_r2[p][i] =  0 ;
  }

  g_system.uid[p] = -1 ;

  return 0 ;
}



//############################## FUNCTION #################################

double realrb(int actatom,int actshell){
  static double RB0=0.5e-10;
  return( (double)RB0/(double) g_system.Z [actatom] *
           (n_quantum_number(actshell))*(n_quantum_number(actshell)) );
}

//############################## FUNCTION #################################

double initvreal(int actatom,int actshell){
  double dist,Ekin;

  dist=realrb(actatom,actshell);
  
  Ekin= g_inputs.Ephoton - g_atomic_data[actatom].atomic_econfig->B[actshell]  -
       potenerreal(dist,-1, g_system.Z [actatom]);
  return(sqrt(2.0*Ekin/ g_phys_math_const.me ));

}


//############################## FUNCTION #################################

int el_config_update_orbital_occupancies(int actatom){
  int actsh;
  struct ATOMS_STRUCT *actatomp;
  struct ATOMDATA_STRUCT *actdata;

  actatomp=g_atomic_data+actatom;
  actdata=ATOMDATA+*(actatomp->Z);
  actatomp->NE=0;
  for (actsh=1;actsh<=actatomp->maxshellnum;++actsh){
    (actatomp->NE)+=actatomp->shell[actsh].NE;
  }
  return(0);
}


//############################## FUNCTION #################################

int netcharge_config_update( int a_id )
{
  int ii_from , ii_to , ii  , qq ;
  if ( a_id < 0 ) {
    ii_from = 0 ;
    ii_to =  g_inputs.N ;
  } 
  else {
    ii_from = a_id ;
    ii_to = a_id+1 ;
  }

  for ( ii=ii_from ; ii<ii_to ; ii++ ) 
  {

     g_system.qi_corr[ii] = g_system.qi[ii] + get_class_recomb_delta_charge( ii ) ;
     if ( g_system.qi_corr[ii] < 0 ) 
     { 
        qq = get_class_recomb_delta_charge( ii ) ;
        g_system.qi_corr[ii] = 0 ; 
     }

   }

   return 0 ;
}


//############################## FUNCTION #################################

int elec_config_update( int actatom )
{

  el_config_update_orbital_occupancies(actatom);

  if ( XATOM_TAB == g_inputs.atomic_data_source )
  {
      xatom_tab_update_atoms( actatom ) ;
  }

  else if ( XATOM_ONTHEFLY == g_inputs.atomic_data_source )
  {
      xatom_onthefly_update_atoms( actatom ) ;
  }

  netcharge_config_update( actatom ) ;



  return(0);
}

//############################## FUNCTION #################################

int xmdyn_update_atoms(char **a_logstr, char *a_option, ...)
{
   int num_atoms = g_inputs.N ;

   int *atom_list = NULL ;
   va_list valist;
   int ii ;
   int opt_econf = 0 , opt_bond = 0 , opt_ionization = 0 ;
   double dV , *out_dV = NULL ;
   char xenv[16] = "+env" ;

   /* initialize valist for num number of arguments */
   va_start(valist, a_option);


    /** Choose Atoms **/

   if ( strstr( a_option , "array" ) != NULL )
   {
       num_atoms = va_arg(valist, int) ; 
       atom_list = va_arg(valist, int*) ; 
   }
   else if ( strstr( a_option , "list" ) != NULL )
   {
       num_atoms = va_arg(valist, int) ; 
       if ( 0 < num_atoms ) { atom_list = XMDYN_ALLOCATE_MEM( num_atoms * sizeof(int) ) ; }
       for (ii = 0; ii < num_atoms; ii++) { atom_list[ii] = va_arg(valist, int) ; } 
   }
   else if ( strstr( a_option , "all" ) != NULL ) { }
   else { MODULE_ERROR( "Wrong particle list definition!" , -1 ) ; }

   if ( strstr( a_option , "getE" ) != NULL ) { out_dV = va_arg(valist, double* ) ; }
   if ( strstr( a_option , "-env" ) != NULL ) { strcpy( xenv , "-env" ) ; }
   if ( strstr( a_option , "ionization" ) != NULL ) { opt_ionization = 1 ; }


   /** Choose what to Update */

   if      ( strstr( a_option , "econfig" ) != NULL ) { opt_econf = 1 ; }
   else if ( strstr( a_option , "bond" ) != NULL )    { opt_bond = 1 ; }
   else if ( strstr( a_option , "full" ) != NULL )    { opt_econf = 1 ; opt_bond = 1 ; }
   else { MODULE_ERROR( "Wrong/No Specification of what to update!" , -1 ) ; }


   /** Do Update **/

   if ( strstr( a_option , "all" ) != NULL )
   {
       if ( opt_econf )
       {
           for (ii = 0; ii < num_atoms; ii++) { elec_config_update( ii ) ; }
       }
       if ( opt_bond )
       {
           // TODO: bond
       }
   }

   else
   {
       if ( opt_econf ) { for (ii = 0; ii < num_atoms; ii++) { elec_config_update( atom_list[ii] ) ; } }
   }
 

   /** Deallocate Mem if needed **/
   if ( strstr( a_option , "list" ) != NULL ) {
       if ( 0 < num_atoms ) { XMDYN_DEALLOCATE_MEM( atom_list ) ; }
   }
        
   /* clean memory reserved for valist */
   va_end(valist);

   return 0;
}


//############################## FUNCTION #################################

double actualrb(int aktatom,int actshell){
  if ( g_inputs.rstarte == 1e-99 ) { return( 0 ) ; }
  if ( 0 <= g_inputs.rstarte ) { return( g_inputs.rstarte ) ; }

  if (  g_inputs.rstarte < 0 )
      return( -  g_inputs.rstarte ) ;

    return 0 ;
}


//############################## FUNCTION #################################

double beam_photon_num(double t,double DT) {
  static int init=1 , dim ;
  static double T0,T_1 , T_2,Thalf_1, Thalf_2,Ti,Tf,sigma_1, sigma_2,x1, x2,norm_1, norm_2,Nav;
  static double N , t_memo=-1 , DT_memo=-1 , *t_ref , *dN_dAdt_ref ;
  static double t_xrsd_delay ;
  static char strid[LSTRLTH]="" ;
  int ii , num_lines ;
  static mytype_interp1 ref_data ;
  double intensity ;

  if (init) 
  {
	char xfile[LSTRLTH]="" ;
	FILE *fp ;
    init=0;
    T0= g_inputs.T0;
    T_1 =  g_inputs.T_1 ;
    T_2 =  g_inputs.T_2 ;
    Thalf_1 = T_1 /2.0;
    Thalf_2 = T_2 /2.0;
    strncpy( strid ,  g_inputs.beamtype , 5 ) ;

    if ( (strcmp( g_inputs.beamtype,"gauss")==0) || (strcmp( g_inputs.beamtype,"gaussian")==0)   ) 
    {
      sigma_1 = T_1/sqrt(8.0*log(2.0)) ; x1=2.0*sigma_1*sigma_1;
      norm_1 =  g_inputs.Nphoton / (sqrt(2.0*PI)*sigma_1) ;
      sigma_2 = T_2/sqrt(8.0*log(2.0)) ; x2=2.0*sigma_2*sigma_2;
      norm_2 =  g_inputs.Nphoton / (sqrt(2.0*PI)*sigma_2) ;
    }

    else if (strcmp( g_inputs.beamtype,"step")==0)
    {
      Nav =  g_inputs.Nphoton/T_1;
      Ti = T0-Thalf_1 ;
      Tf = T0+Thalf_1 ;
    }

    else if ( strcmp( strid ,"file:")==0) 
    {
      strcat( strcat( strcat( xfile ,  g_inputs.rootdir ) , "/" ) ,  g_inputs.beamtype+5 ) ;
	  num_lines = fileop_num_of_lines( xfile ) ;
	  fprintf(stdout , "Loading beamprofile %s (%d lines)\n" , xfile , num_lines ) ;
      t_ref = calloc( num_lines , sizeof(double) ) ;
      dN_dAdt_ref = calloc( num_lines , sizeof(double) ) ;
	  fp = fopen( xfile , "r" ) ;
	  for ( ii=0 ; ii < num_lines ; ii++ ) {
		  fscanf( fp , "%le %le" , t_ref+ii , dN_dAdt_ref+ii) ;
	  }
	  fclose(fp) ;
	  dim = num_lines ;
      init_interp1( &ref_data , t_ref , dN_dAdt_ref , dim , "cspline" ) ; // cspline or linear
    }

    if ( strcmp( strid ,"file:") != 0 )
    {
        if ( 0 != g_inputs.xrsd_delay ) 
        { 
            norm_2 =       g_inputs.xrsd_ratio_2toT    * norm_2 ; 
            norm_1 = ( 1 - g_inputs.xrsd_ratio_2toT )  * norm_1 ; 
        }
        else { norm_2 = 0 ; }
	//norm_1 = norm_1 - norm_2 ;
	t_xrsd_delay = g_inputs.xrsd_delay ;
    }
  }

  if ( ( t != t_memo ) || ( DT != DT_memo ) ) {

    if ( (strcmp( g_inputs.beamtype,"gauss")==0) || (strcmp( g_inputs.beamtype,"gaussian")==0)  )  {
  	N = norm_1 * DT * exp( -( t - T0 ) * ( t - T0 ) / x1 ) +
  	    norm_2 * DT * exp( -( t - T0 - t_xrsd_delay ) * ( t - T0 - t_xrsd_delay ) / x2 ) ;
    }

    else if (strcmp( g_inputs.beamtype,"step")==0) {
      N = Nav * DT  * ( Heaviside1(t-Ti) - Heaviside1(t-Tf) );
    }

	else if (strcmp( strid ,"file:")==0) {
	  if ( ( t < t_ref[0] ) || ( t > t_ref[dim-1] ) ) {
	    N = 0 ;
      }
	  else {
        N = do_interp1( &ref_data , t ) ; 
        if ( N < 0 ) { N = 0 ; } 
                N *= ( 1. / ( g_inputs.Ephoton) * DT ) ;
	  }
	}


    else {
  	  N = -1 ;
    }

	t_memo = t ; DT_memo = DT ;

  }

  return(N);
}



//############################## FUNCTION #################################

int beamparams(int step,double *nph){
  double dt , tnow ;
  int qpxfactor = 0 ;

  tnow = cd_time_now() ;
  dt =  g_inputs.DT / pow( 2 ,  qpxfactor) ;

  *nph = beam_photon_num( tnow , dt ) ;

  return(0) ;

}

//TODO:RM
//############################## FUNCTION #################################

int insidecluster(int part,double dist){
  static double R, M, C[3], MC[3];
  double dr[3];
  int i,j;
  
  if (part==-1){    
    R=M=C[0]=C[1]=C[2]=MC[0]=MC[1]=MC[2]=0;
    for (i=0;i< g_inputs.N;++i){
      M+= g_system.m[i];
      for (j=0;j<3;++j){
        C[j]+= g_system.r1 [i][j];
        MC[j]+= g_system.m[i]* g_system.r1 [i][j];
      }    
    }
    for (j=0;j<3;++j){
      C[j]/=(double) g_inputs.N;
      MC[j]/=M;
    }            
    for (i=0;i< g_inputs.N;++i){
      vectminus(dr, g_system.r1 [i],C);
      if (sqrt(vectnorm2(dr))>R){
        R=sqrt(vectnorm2(dr));
      }    
    }
    return(-1);
  }
  
  vectminus(dr, g_system.r1 [part],C);
  if (sqrt(vectnorm2(dr))>R+dist){
    return(0);
  }
  else{
    return(1);
  }
  
}

//############################## FUNCTION #################################

int self_consistency_check(int step,char *comm) {
  int at,sh,i,NE,subNE;
  static int no_error_detected = 1 ;

  if (strcmp(comm,"elconfig")==0) {
    for (at=0;at< g_inputs.N;++at) {
      NE=0;
      for (sh=0;sh<=g_atomic_data[at].maxshellnum;++sh) {
        if (is_real_subshell(sh)) {
          subNE=0;
	  for (i=g_atomic_data[at].shell[sh].e_from;i<=g_atomic_data[at].shell[sh].e_to;++i){
            if (isitshellelectron(i)) {
	      subNE++;
	    }
	  }
	  if ( no_error_detected && (subNE!=g_atomic_data[at].shell[sh].NE) ) {
            no_error_detected = 0 ;

            print_log( "" , (void*) "ERROR-check-elcfg-sub:" , "str" , "noflush" ) ;
            print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
            print_log( " atom:" , (void*) &at  , "int" , "noflush" ) ;
            print_log( " shell:" , (void*) &sh  , "int" , "noflush" ) ;
            print_log( " NE:" , (void*) &subNE  , "int" , "noflush" ) ;
            print_log( " NEbooked:" , (void*) &(g_atomic_data[at].shell[sh].NE)  , "int" , "noflush" ) ;
            print_log( "\n" , NULL , "" , "noflush" ) ;
	  }
	  NE+=subNE;
	}
      }
      if ( no_error_detected && (NE!=g_atomic_data[at].NE) ) {
            no_error_detected = 0 ;

            print_log( "" , (void*) "ERROR-check-elcfg:" , "str" , "noflush" ) ;
            print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
            print_log( " atom:" , (void*) &at  , "int" , "noflush" ) ;
            print_log( " NE:" , (void*) &subNE  , "int" , "noflush" ) ;
            print_log( " NEbooked:" , (void*) &(g_atomic_data[at].shell[sh].NE)  , "int" , "noflush" ) ;
            print_log( "\n" , NULL , "" , "noflush" ) ;
      }
    }
  }

  if ( (strcmp(comm,"q")==0) && ( strstr(  g_inputs.misc_opt , "nolog:qerror" ) == NULL ) ) {
    for (at=0;at< g_inputs.N;++at) {
      if ( no_error_detected && ( g_system.qi [at]!= g_system.Z [at]-g_atomic_data[at].NE) ) {
            no_error_detected = 0 ;
           
            int xxx ;
            print_log( "" , (void*) "ERROR-check-q:" , "str" , "noflush" ) ;
            print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
            print_log( " atom:" , (void*) &at  , "int" , "noflush" ) ;
            print_log( " q:" , (void*) &(g_system.qi[at])  , "int" , "noflush" ) ;
            xxx = g_system.Z [at]-g_atomic_data[at].NE ;
            print_log( " qNE:" , (void*) &xxx  , "int" , "noflush" ) ;
            print_log( "\n" , NULL , "" , "noflush" ) ;
 }
    }
  }  


  return 0 ;
}


//  ###############  UPDATE  ##################

//############################## FUNCTION #################################

int xatom_onthefly_update_atoms( int a_id ) {
  int ii , jj , sh, maxshellnum , ii_from , ii_to , ZZ , gotit , num_e ;
  int numradproc, numnonradproc, curr_radproc , curr_nonradproc  ;
  struct XATOM_DBASE_STRUCT *actdbase ;
  struct XATOM_DBASE_PROC_STRUCT *actproc ;
  struct ATOMS_STRUCT *x_atomdata  ;
  int *e_config  ;
  int max_occ_shell = 0 ;
  
  if ( a_id < 0 ) {
    ii_from = 0 ;
    ii_to =  g_inputs.N ;
  } 
  else {
    ii_from = a_id ;
    ii_to = a_id+1 ;
  }

  for ( ii=ii_from ; ii<ii_to ; ii++ ) 
  {

    /* ########################################### */
   
    x_atomdata = g_atomic_data + ii ;

    e_config = calloc( x_atomdata->maxshellnum + 1 , sizeof(int) ) ;
    num_e = 0 ;
    for ( jj = 0 ; jj <= x_atomdata->maxshellnum ; jj++ )
    {
        e_config[jj] = x_atomdata->shell[jj].NE ;
        num_e += e_config[jj] ;
        if ( 0 < e_config[jj] ) { max_occ_shell = jj ; }

    }

    // In case there's no electron, we still have to specify a max_shell > 0
    if ( 0 == max_occ_shell ) { max_occ_shell = 1 ; } 

    if ( max_occ_shell < get_equviv_shell(max_occ_shell) ) { max_occ_shell++ ; }

    xatom_tree_add( *(x_atomdata->Z) , max_occ_shell , e_config , &(x_atomdata->atomic_econfig) ) ;

    xatom_call_on_the_fly( *(x_atomdata->Z) , max_occ_shell , 
                           e_config , g_inputs.Ephoton , 
                           x_atomdata->atomic_econfig )  ;

/*** recombination energy */

    x_atomdata->atomic_econfig->Ecrit_cl2q = 0 ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if ( strstr( g_inputs.irec_method , "FIX" ) != NULL ) 
    {
        if ( num_e < *(x_atomdata->Z) )
        {
            char *xstr = g_inputs.irec_method + 3 ;
            sscanf( xstr , "%le" , &(x_atomdata->atomic_econfig->Ecrit_cl2q) ) ;
            x_atomdata->atomic_econfig->Ecrit_cl2q *= g_phys_math_const.e1 ;
        }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if ( strcmp( g_inputs.irec_method , "B0" ) == 0 ) 
    {
        if ( num_e < *(x_atomdata->Z) )
        {
            x_atomdata->atomic_econfig->Ecrit_cl2q = x_atomdata->atomic_econfig->B[max_occ_shell] ;
        }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if ( strcmp( g_inputs.irec_method , "HUMO" ) == 0 ) 
    {
        if ( num_e < *(x_atomdata->Z) )
        {
            x_atomdata->atomic_econfig->Ecrit_cl2q = x_atomdata->atomic_econfig->B[ highest_unoccupied_shell( ii ) ] ; 
        }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

#ifdef WORKAROUND_XE_NO_4f
    DBG_WARNING_NOTE_ON_CODE( "4f orbital was deactivated at compilation to speed up calculation." );
#endif

    if ( strcmp( g_inputs.irec_method , "B+" ) == 0 ) 
    {
        if ( num_e < *(x_atomdata->Z) )
        //if ( *(x_atomdata->q) > 0 )
        {
            int sel_occ , sel_hole ;
            struct g_struct_xatom_data *temp_atomic_econfig = NULL ;

            sel_hole = x_atomdata->maxshellnum ; 
            while ( ( 0 <= sel_hole ) &&
                    ( ( get_orbit_max_occupancy(sel_hole) == 0 ) ||
                      ( ( e_config[sel_hole] == get_orbit_max_occupancy(sel_hole) ) && ( e_config[sel_hole] >0 ) )
                    )
#ifdef WORKAROUND_XE_NO_4f
                    || ( 24 == sel_hole )
#endif
                  )
            { sel_hole-- ; }

            if ( sel_hole < 0 ) { MODULE_ERROR( "ERROR!!! sel_hole < 0" , -1 ) ; }


            sel_occ = x_atomdata->maxshellnum ; 
            while (  ( 0 <= sel_occ )  &&  ( get_orbit_max_occupancy(sel_occ) == 0 )  )
            { sel_occ-- ; }

            if ( sel_occ < 0 ) { MODULE_ERROR( "ERROR!!! sel_occ < 0" , -1 ) ; }


            e_config[sel_hole]++ ;

            max_occ_shell = 1 ;
            for ( jj = 0 ; jj <= x_atomdata->maxshellnum ; jj++ )
            {
                if ( 0 < e_config[jj] ) { max_occ_shell = jj ; }
            }

            xatom_tree_add( *(x_atomdata->Z) , max_occ_shell , e_config , &(temp_atomic_econfig) ) ;
            xatom_call_on_the_fly( *(x_atomdata->Z) , max_occ_shell , 
                                   e_config , g_inputs.Ephoton , 
                                   temp_atomic_econfig ) ;
            x_atomdata->atomic_econfig->Ecrit_cl2q = temp_atomic_econfig->B[sel_occ] ;
        }
    }

    x_atomdata->atomic_econfig->B_rec = x_atomdata->atomic_econfig->Ecrit_cl2q ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if ( NULL != strstr( g_inputs.irec_method , "B-" ) ) 
    {
        double dE_cr2B = 0 ;
        if ( 2 < strlen( g_inputs.irec_method ) ) 
        { 
            char *xxxs = g_inputs.irec_method + 2 ; 
            sscanf( xxxs , "%le" , &dE_cr2B ) ;
        }
        if ( num_e < *(x_atomdata->Z) )
        {
            int sel_occ , sel_hole ;
            struct g_struct_xatom_data *temp_atomic_econfig = NULL ;

            sel_hole = x_atomdata->maxshellnum ; 
            while ( ( 0 <= sel_hole ) &&
                    ( ( get_orbit_max_occupancy(sel_hole) == 0 ) ||
                      ( ( e_config[sel_hole] == get_orbit_max_occupancy(sel_hole) ) && ( e_config[sel_hole] >0 ) )
                    )
#ifdef WORKAROUND_XE_NO_4f
                    || ( 24 == sel_hole )
#endif
                  )
            { sel_hole-- ; }

            if ( sel_hole < 0 ) { MODULE_ERROR( "ERROR!!! sel_hole < 0" , -1 ) ; }

            sel_occ = x_atomdata->maxshellnum ; 
            while (  ( 0 <= sel_occ )  &&  ( get_orbit_max_occupancy(sel_occ) == 0 )  )
            { sel_occ-- ; }

            if ( sel_occ < 0 ) { MODULE_ERROR( "ERROR!!! sel_occ < 0" , -1 ) ; }


            e_config[sel_hole]++ ;

            max_occ_shell = 1 ;
            for ( jj = 0 ; jj <= x_atomdata->maxshellnum ; jj++ )
            {
                if ( 0 < e_config[jj] ) { max_occ_shell = jj ; }
            }

            xatom_tree_add( *(x_atomdata->Z) , max_occ_shell , e_config , &(temp_atomic_econfig) ) ;
            xatom_call_on_the_fly( *(x_atomdata->Z) , max_occ_shell , e_config , 
                                   g_inputs.Ephoton , temp_atomic_econfig ) ;

            x_atomdata->atomic_econfig->B_rec = temp_atomic_econfig->B[sel_occ] ;
            x_atomdata->atomic_econfig->Ecrit_cl2q = x_atomdata->atomic_econfig->B_rec - 
                                                     convert_unit_eV2J( dE_cr2B ) ;
            if ( x_atomdata->atomic_econfig->Ecrit_cl2q < 0 ) { x_atomdata->atomic_econfig->Ecrit_cl2q = 0 ; }
        }
        else { x_atomdata->atomic_econfig->Ecrit_cl2q = x_atomdata->atomic_econfig->B_rec = -1 ; } 
    }



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


    free( e_config ) ;

    
  }


  return 0 ;

}


//############################## FUNCTION #################################

int xatom_tab_update_atoms( int a_id ) {
  int ii , jj , sh, maxshellnum , ii_from , ii_to , ZZ , gotit ;
  int numradproc, numnonradproc, curr_radproc , curr_nonradproc  ;
  struct XATOM_DBASE_STRUCT *actdbase ;
  struct XATOM_DBASE_PROC_STRUCT *actproc ;

  return 0 ;
  
  return 0 ;
}



// ########################################################
// ########################################################
// ########################################################

//############################## FUNCTION #################################

double cd_time_now(void) 
{
  int sstep = 0 ; 
  int fsxfactor = 0 ;
  return( (  g_inputs.step +  sstep * ( 1. / pow( 2. ,  fsxfactor ) ) ) *  g_inputs.DT  ) ;
}


//############################## FUNCTION #################################

int get_classical_electrons( int *Nel , int *el_id , int Npresel , int *el_presel )
{
    int ii , cc ;

   if ( NULL == el_presel )
   {
       el_presel = calloc( g_inputs.E , sizeof(int) ) ;
       for ( ii = 0 ; ii < g_inputs.E ; ii++ ) { el_presel[ii] = g_inputs.N + ii ; }
       Npresel = g_inputs.E ;
   }


   *Nel = 0 ;

   for ( ii = 0 ; ii < Npresel ; ++ii ) 
   {
     if  ( isitfreeelectron( el_presel[ii] ) ) { ( *Nel )++; }
   }

   if ( *Nel == 0 ) { el_id = NULL ; return 0 ; }


   el_id = calloc( *Nel , sizeof(int) ) ;
   cc = 0 ;
   for ( ii = 0 ; ii < Npresel ; ++ii ) 
   {
     if  ( isitfreeelectron( el_presel[ii] ) ) 
     {
       el_id[cc] = el_presel[ii] ;
       cc++ ;
     }
   }

   if ( cc != *Nel ) 
   {

       return -1 ;
   }

   return 0 ;

}



// ##################################################################
// ##################################################################
// ##################################################################

//############################## FUNCTION #################################

int xmdyn_store_restore_subsystem( char *a_comm , char *a_options , int a_num_sel_parts , ... )
{

   struct g_struct_system_data  *class_system = &g_system ;
   struct ATOMS_STRUCT           *atomic_data = g_atomic_data ;
   int NN = g_inputs.N ;

   static int *l_sel_parts = NULL ;
   static int l_num_parts = 0 ;

   static int *stored_maxshellnum , *stored_NE , **stored_shellNE ;
   static struct g_struct_xatom_data **stored_econfig ;
   static double (*stored_r1)[3] ,  (*stored_r2)[3] ,  (*stored_v1)[3] ,  (*stored_v2)[3] , *stored_qd ;
   static int *stored_qi , *stored_qi_corr , *stored_motherion ;
   static long long int *stored_uid ;
   static char *stored_ptype ;

   va_list valist;

   int ii, pp , cc , id , sh ;

   // ---------------------------------------------------------------

   if ( strcmp( a_comm , "store" )  == 0 )
   {

       if ( NULL != l_sel_parts ) { MODULE_ERROR( "Storing double subsystem snp at the same time is not allowed!" , -1 ) ; } ;

       l_num_parts = a_num_sel_parts ;

       if ( strcmp( "a_option" , "atoms" ) == 0 ) { l_num_parts += NN ; }

       l_sel_parts = calloc( l_num_parts , sizeof(int) ) ;

       cc = 0 ;
       if ( strcmp( "a_option" , "atoms" ) == 0 ) { for ( ii=0;ii<NN; ii++) { l_sel_parts[cc] = ii ; cc++ ; }}
       /* initialize valist for num number of arguments */
       va_start( valist, a_num_sel_parts );
       for (ii = 0; ii < a_num_sel_parts; ii++) { l_sel_parts[cc] = va_arg(valist, int) ; cc++ ; } 
       /* clean memory reserved for valist */
       va_end(valist);

       // ECONFIG:
       stored_maxshellnum = calloc( l_num_parts , sizeof(int) ) ;
       stored_econfig = calloc( l_num_parts , sizeof( struct g_struct_xatom_data* ) ) ;
       stored_NE = calloc( l_num_parts , sizeof(int) ) ;
       stored_shellNE = calloc( l_num_parts , sizeof(int*) ) ;
       for ( pp = 0 ; pp < l_num_parts ; pp++ ) { if ( l_sel_parts[pp] < NN ) { 
           stored_shellNE[pp] = calloc( atomic_data[l_sel_parts[pp]].maxshellnum + 1 , sizeof(int) ) ;
       }}
       //stored_ = calloc( num_parts , sizeof( ) ) ;

       // MD
       stored_r1   = calloc( l_num_parts , 3 * sizeof(double) ) ;
       stored_r2   = calloc( l_num_parts , 3 * sizeof(double) ) ;
       stored_v1   = calloc( l_num_parts , 3 * sizeof(double) ) ;
       stored_v2   = calloc( l_num_parts , 3 * sizeof(double) ) ;
       stored_qd   = calloc( l_num_parts , sizeof(double) ) ;
       stored_qi   = calloc( l_num_parts , sizeof(int) ) ;
       stored_qi_corr   = calloc( l_num_parts , sizeof(int) ) ;
       stored_ptype   = calloc( l_num_parts , sizeof(char) ) ;
       stored_uid   = calloc( l_num_parts , sizeof(long long int) ) ;
       stored_motherion   = calloc( l_num_parts , sizeof(int) ) ;

       for ( pp = 0 ; pp < l_num_parts ; pp++ ) 
       { 
           id = l_sel_parts[pp] ;
           if ( id < NN ) 
           { 
               stored_maxshellnum[pp] = atomic_data[id].maxshellnum ;
               stored_econfig[pp] = atomic_data[id].atomic_econfig ;
               stored_NE[pp] =  atomic_data[id].NE ;
               for ( sh=0 ; sh < atomic_data[id].maxshellnum + 1 ; sh++ ) {
                   stored_shellNE[pp][sh] = atomic_data[id].shell[sh].NE ;
               }
           }
           for ( ii=0 ; ii<3; ii++ )
           {
               stored_r1[pp][ii] = (*class_system).r1[id][ii] ;
               stored_r2[pp][ii] = (*class_system).r2[id][ii] ;
               stored_v1[pp][ii] = (*class_system).v1[id][ii] ;
               stored_v2[pp][ii] = (*class_system).v2[id][ii] ;
           }
           stored_qi[pp] = (*class_system).qi[id] ;
           stored_qi_corr[pp] = (*class_system).qi_corr[id] ;
           stored_qd[pp] = (*class_system).qd[id] ;
           stored_ptype[pp] = (*class_system).ptype[id] ;
           stored_uid[pp] = (*class_system).uid[id] ;
           stored_motherion[pp] = (*class_system).motherion[id] ;
       }
       xmdyn_store_restore_recombination( "store" , l_num_parts , l_sel_parts ) ;
       xmdyn_md_energy( "store" , "" ) ;
   }

   // ---------------------------------------------------------------

   if ( strcmp( a_comm , "restore" )  == 0 )
   {

       for ( pp = 0 ; pp < l_num_parts ; pp++ ) 
       { 
           id = l_sel_parts[pp] ;
           if ( id < NN ) 
           { 
               atomic_data[id].maxshellnum = stored_maxshellnum[pp] ;
               atomic_data[id].atomic_econfig = stored_econfig[pp] ;
               atomic_data[id].NE = stored_NE[pp] ;
               for ( sh=0 ; sh < atomic_data[id].maxshellnum + 1 ; sh++ ) {
                   atomic_data[id].shell[sh].NE = stored_shellNE[pp][sh] ;
               }
           }
           for ( ii=0 ; ii<3; ii++ )
           {
               (*class_system).r1[id][ii] = stored_r1[pp][ii] ;
               (*class_system).r2[id][ii] = stored_r2[pp][ii] ;
               (*class_system).v1[id][ii] = stored_v1[pp][ii] ;
               (*class_system).v2[id][ii] = stored_v2[pp][ii] ;
           }
           (*class_system).qi[id] = stored_qi[pp] ;
           (*class_system).qi_corr[id] = stored_qi_corr[pp] ;
           (*class_system).qd[id] = stored_qd[pp] ;
           (*class_system).ptype[id] = stored_ptype[pp] ;
           (*class_system).uid[id] = stored_uid[pp] ;
           (*class_system).motherion[id] = stored_motherion[pp] ;
       }

       xmdyn_store_restore_recombination( "restore" , 0 , NULL ) ;
       xmdyn_md_energy( "restore" , "" ) ;
   }

   // ---------------------------------------------------------------

   if ( ( strcmp( a_comm , "restore" )  == 0 ) ||  ( strcmp( a_comm , "clear" )  == 0 ) )
   {
       for ( pp = 0 ; pp < l_num_parts ; pp++ ) { if ( l_sel_parts[pp] < NN ) { 
           free( stored_shellNE[pp] ) ;
       }}
       free( stored_shellNE ) ; free( stored_NE ) ;  free( stored_econfig ) ;  free( stored_maxshellnum ) ;
       free( stored_r1 ) ;  free( stored_r2 ) ;  free( stored_v1 ) ;  free( stored_v2 ) ; 
       free( stored_qd ) ;  free( stored_qi ) ;  free( stored_qi_corr ) ;  free( stored_ptype) ;
       free( stored_uid ) ;  free( stored_motherion ) ;  
       free( l_sel_parts ) ; l_sel_parts = NULL ; 
       xmdyn_store_restore_recombination( "clear" , 0 , NULL ) ;
   }

   // ---------------------------------------------------------------

   return 0 ;

}


// #############################################


