///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

 #include <stdarg.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>

#include <unistd.h>
#include <malloc.h>

#include "basics.h"
#include "datatypes.h"


//############################## FUNCTION #################################

int xmdyn_remove_element_x(void **array, int sizeOfArray, int indexToRemove , char *xtype )
{

    int unit ;

    if      ( strcmp( xtype , "int" ) == 0 )   { unit = sizeof(int) ; }
    else if ( strcmp( xtype , "int*" ) == 0 )  { unit = sizeof(int*) ; }
    else if ( strcmp( xtype , "char*" ) == 0 ) { unit = sizeof(char*) ; }
    else if ( strcmp( xtype , "double" ) == 0 ) { unit = sizeof(double) ; }
    else { fprintf( stderr , "Unknown type: %s\n" , xtype ) ; }

    void* temp = malloc( (sizeOfArray - 1) * unit ); // allocate an array with a size 1 less than the current one

    if (indexToRemove != 0)
        memcpy( temp, *array, indexToRemove * unit ); // copy everything BEFORE the index

    if (indexToRemove != (sizeOfArray - 1))
	memmove( temp + (indexToRemove) * unit , (*array) + (indexToRemove+1) * unit , (sizeOfArray - indexToRemove -1 ) * unit ); // copy everything AFTER the index

    free (*array);
    *array = temp;
    return 0 ;
}


//############################## FUNCTION #################################

int xmdyn_remove_element(void ***array, int sizeOfArray, int indexToRemove)
{
    void** temp = malloc( (sizeOfArray - 1) * sizeof(void*) ); // allocate an array with a size 1 less than the current one

    if (indexToRemove != 0)
        memcpy( temp, *array, indexToRemove * sizeof(void*) ); // copy everything BEFORE the index

    if (indexToRemove != (sizeOfArray - 1))
	memmove(temp+(indexToRemove), (*array)+(indexToRemove+1), (sizeOfArray - indexToRemove -1 ) * sizeof(void*) ); // copy everything AFTER the index

    free (*array);
    *array = temp;
    return 0 ;
}



//############################## FUNCTION #################################

int xmdyn_remove_element_int(int **array, int sizeOfArray, int indexToRemove)
{
    int* temp = malloc( (sizeOfArray - 1) * sizeof(int) ); // allocate an array with a size 1 less than the current one

    if (indexToRemove != 0)
        memcpy( temp, *array, indexToRemove * sizeof(int) ); // copy everything BEFORE the index

    if (indexToRemove != (sizeOfArray - 1))
	memmove(temp+(indexToRemove), (*array)+(indexToRemove+1), (sizeOfArray - indexToRemove -1 ) * sizeof(int) ); // copy everything AFTER the index

    free (*array);
    *array = temp;
    return 0 ;
}


//############################## FUNCTION #################################

int xmdyn_get_meminfo( char *comm )
{
    struct mallinfo mi = mallinfo();

    if ( 0 == strcmp( comm , "malloc"   ) ) { return mi.uordblks ; }

    if ( 0 == strcmp( comm , "arena"    ) ) { return mi.arena    ; } 
    if ( 0 == strcmp( comm , "ordblks"  ) ) { return mi.ordblks  ; } 
    if ( 0 == strcmp( comm , "smblks"   ) ) { return mi.smblks   ; } 
    if ( 0 == strcmp( comm , "hblks"    ) ) { return mi.hblks    ; } 
    if ( 0 == strcmp( comm , "hblkhd"   ) ) { return mi.hblkhd   ; } 
    if ( 0 == strcmp( comm , "usmblks"  ) ) { return mi.usmblks  ; } 
    if ( 0 == strcmp( comm , "fsmblks"  ) ) { return mi.fsmblks  ; } 
    if ( 0 == strcmp( comm , "uordblks" ) ) { return mi.uordblks  ; } 
    if ( 0 == strcmp( comm , "fordblks" ) ) { return mi.fordblks  ; } 
    if ( 0 == strcmp( comm , "keepcost" ) ) { return mi.keepcost  ; } 

    return -1 ;
}

//############################## FUNCTION #################################

int xmdyn_print_malloc_info( char *misc , int step , int substep ,char *fname , const char *func  , int xline )
{
    //return 0 ;
    int option = 0 ;
    static int prev_size = 0 ;
    static char prev_str[512] = "" ;
    static char curr_str[512] = "" ;
    int xsize = xmdyn_get_meminfo( "malloc" ) ;
    int diff = xsize - prev_size ;
    sprintf( curr_str , "Step: %6d   SubStep: %6d   Mem: %9d   Diff: %09d %s Function: %s   File: %s Line: %d\n" , step , substep , xsize, diff , misc , func , fname , xline  ) ;

    if ( ( 1 == option ) && ( 0 != diff ) ) { printf( "%s%s\n" , prev_str, curr_str ) ; }
    if ( ( 0 == option ) )                        { printf( "%s\n" , curr_str ) ; }

    prev_size = xsize ;
    strcpy( prev_str, curr_str ) ;
    return 0 ;
}

//############################## FUNCTION #################################

int xmdyn_get_malloc_info( char *info , int len_info , char *misc , char *fname , const char *func )
{
    char *s = malloc(sizeof(char)) ; strcpy( s , "" ) ;
    int xsize = xmdyn_get_meminfo( "malloc" ) ;
    char s2[64] ; sprintf( s2 , "%d" , xsize ) ;
    s = str_concat_realloc(str_concat_realloc(str_concat_realloc(str_concat_realloc( s , "Mem: " ) , misc ) , " " ) , s2 ) ; 
    strncpy( info , s , len_info-1 ) ;
    free( s ) ;
    return 0 ;
}

/** NOTE: DBG_MEMORY_USAGE mode is NOT thread-safe  !!! */

static long long int total_memory_in_use = 0 ;
static int *alloc_block_size = NULL ;
static char **manage_mem_files = NULL ;
static char **manage_mem_funcs = NULL ;
static int *manage_mem_lines = NULL ;
static int num_alloc_blocks = 0 ;


//############################## FUNCTION #################################

void* xmdyn_allocate_mem(int nb, char *a_file , int a_line , const char *a_func )
{
    void *mem;
    int fnamesize = 255 ; 

    if ( 0 == nb ) { return NULL ; } 
    
    mem = calloc(nb, 1);                /* allocate, also clearing memory   */
    if (mem == NULL) { fprintf( stderr , "allocate in %s line %d: not enough memory (%d bytes)\n" , a_file , a_line , nb ) ; exit(-1) ;}
#ifdef DBG_MEMORY_USAGE
    total_memory_in_use += (long long int) nb ;

    alloc_blocks = realloc( alloc_blocks , ( num_alloc_blocks + 1 ) * sizeof(void*) ) ;
    alloc_blocks[num_alloc_blocks] = (void*) mem ;

    alloc_block_size = realloc( alloc_block_size , ( num_alloc_blocks + 1 ) * sizeof(int) ) ;
    alloc_block_size[num_alloc_blocks] = nb ;

    manage_mem_files = realloc( manage_mem_files , ( num_alloc_blocks + 1 ) * sizeof(char*) ) ;
    manage_mem_files[num_alloc_blocks] = calloc( fnamesize + 1 , sizeof(char) ) ;
    strncpy( manage_mem_files[num_alloc_blocks] , a_file , fnamesize ) ;

    manage_mem_funcs = realloc( manage_mem_funcs , ( num_alloc_blocks + 1 ) * sizeof(char*) ) ;
    manage_mem_funcs[num_alloc_blocks] = calloc( fnamesize + 1 , sizeof(char) ) ;
    strncpy( manage_mem_funcs[num_alloc_blocks] , a_func , fnamesize ) ;

    manage_mem_lines = realloc( manage_mem_lines , ( num_alloc_blocks + 1 ) * sizeof(int) ) ;
    manage_mem_lines[num_alloc_blocks] = a_line ;

    alloc_block_size[num_alloc_blocks] = nb ;

    num_alloc_blocks++ ;
#endif

    return (mem);
}


//############################## FUNCTION #################################

void xmdyn_deallocate_mem( void *p )
{
  //total_memory_in_use -= sizeof( p ) ;
  
  int c = 0 ;

  if ( NULL == p ) { return ; } 
  
#ifdef DBG_MEMORY_USAGE
  while ( ( alloc_blocks[c] != p ) && ( c < num_alloc_blocks ) ) { c++ ; }
  if ( c == num_alloc_blocks ) { error("Memory pointer not found."); }
  else { 
    num_alloc_blocks-- ;
    xmdyn_remove_element( &alloc_blocks , num_alloc_blocks + 1 , c ) ; 
    total_memory_in_use -= alloc_block_size[c] ;
    free( manage_mem_files[c] ) ;
    xmdyn_remove_element_x( (void**)&manage_mem_files , num_alloc_blocks + 1 , c , "char*" ) ;
    free( manage_mem_funcs[c] ) ;
    xmdyn_remove_element_x( (void**)&manage_mem_funcs , num_alloc_blocks + 1 , c , "char*" ) ;
    xmdyn_remove_element_x( (void**)&manage_mem_lines , num_alloc_blocks + 1 , c , "int" ) ;
    xmdyn_remove_element_int( &alloc_block_size , num_alloc_blocks + 1 , c ) ; 
  } 
  //num_alloc_blocks-- ;
#endif

  free( p ) ;
}



//############################## FUNCTION #################################

int xmdyn_list_allocated_mem( char *list )
{
    int ii ;
    printf( "----------------------------------------------\n" ) ;
    printf("Number of allocated arrays: %d , Allocated bytes: %lld\n" , num_alloc_blocks , total_memory_in_use ) ;
    for ( ii= 0 ; ii < num_alloc_blocks ; ii++ )
    {
        if ( ( strstr( list , manage_mem_files[ii] ) != NULL ) || ( strcmp(list,"") == 0 ) )
        {
            printf( "MEM: File: %-15s Line: %-6d Func: %-30s Bytes: %d\n" , manage_mem_files[ii] , manage_mem_lines[ii] , manage_mem_funcs[ii] , alloc_block_size[ii] ) ;
        }
    }
    printf( "----------------------------------------------\n" ) ;
    return 0 ;
}



//############################## FUNCTION #################################

void xmdyn_dealloc_all( void )
{
#ifdef DBG_MEMORY_USAGE
  int c ;
  for  ( c = 0 ; c < num_alloc_blocks ; c++ )  
  { 
//      printf("MEM todo: %d %d\n" ,num_alloc_blocks ,  c ) ;
      free( alloc_blocks[c] ) ; 
//      printf("MEM done: %d %d\n" ,num_alloc_blocks ,  c ) ;
  }
  free( alloc_blocks ) ;
  free( alloc_block_size ) ;
  alloc_blocks = NULL ;
  num_alloc_blocks = 0 ;
  alloc_block_size = NULL ;
  total_memory_in_use = 0 ;
#else
  printf("`dealloc_all_st` works only if DBG_MEMORY_USAGE defined at compilation.\n") ;
#endif
}



//############################## FUNCTION #################################

void xmdyn_print_mem_usage( void ) 
{ 
//  printf( "mem: %lld\n" , total_memory_in_use ) ; 
  printf( "num_alloc_blocks: %6d  ---  mem usage: %12lld B\n" , num_alloc_blocks , total_memory_in_use ) ; 
}

//############################## FUNCTION #################################

void* my_malloc( int a_size , void *a_idstr ) 
{
    static int total_dyn_mem = 0 ;
    void *p ;

    if ( a_size < 0 )    /*  we are going to free  */
    {
        free( a_idstr ) ;
        total_dyn_mem -= a_size ; 

    }

    if ( a_size > 0 )    /*  we are going to allocate  */
    {
        if ( p = malloc( a_size ) )
        {
            fprintf( stderr , "Memory allocation problem: %s !!!\n" , (char*)a_idstr ) ;
        }
        total_dyn_mem += a_size ; 
        return( p ) ;
    }

    return NULL ;
}


//############################## FUNCTION #################################

int error_exit(char *message,double num,int returnvalue) {
  if (returnvalue!=0) {
    //if ( (My_Rank==Root_Rank) && (strlen(message) != 0) ) 
    if (strlen(message) != 0)  
    {
        fprintf( stderr , "%s %le\n" , message , num ) ;
    }
    exit(returnvalue);
  }

  return 0 ;
}



