///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#define STRRSTR_MISSING

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>
#include <ctype.h>
#include <stdarg.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_interp.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "basics.h"
#include "datatypes.h"



#include <libgen.h>
      
// #####################################################################

#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}


//############################## FUNCTION #################################

char* my_basename( char *xname )
{
    char *bname = xname;
    
    while ( *xname )
    {
        if ( '/' == *xname ) { bname = xname + 1 ; }
        xname++ ;
    }
    return ( bname ) ;
}




//############################## FUNCTION #################################

void dirstrendcorrector(char *dir){
  if ( (dir[strlen(dir)-1]=='\\') || (dir[strlen(dir)-1]=='/') )
    dir[strlen(dir)-1]='\0';
}


//############################## FUNCTION #################################

int iswhitespace(char c) {
  char wsp[STRLTH]=" \t";
  int i;
  for (i=0;(unsigned)i<strlen(wsp);++i){
    if (c==wsp[i]){
      return(1);
    }
  }
  return(0);
}


//############################## FUNCTION #################################

int char_replace(char *s,char c,char rep) {
  char *p;
  
  p=s;
  while (p=strchr(p,c),p!=NULL) {
    p[0]=rep;
  }
  return 0 ;
}


//############################## FUNCTION #################################

int dim_input_pars(char *s) {
  int i,d;
  d=0;
  i=0;
  while (s[i]!='\0') {
    if (s[i]=='_')
      d++;
    i++;
  }
  return(d);
}


//############################## FUNCTION #################################

char* str_remove_comment(char *str, ... )
{
  char *p ;
  p = strchr( str , COMMENTCARACTER ) ;
  if ( p != NULL ) {
    *p = '\0' ;
  }
  return str ;
}


//############################## FUNCTION #################################

double* load_input_pars(char *s,int d) {
  int i;
  char *p;
  static double *par;

  if (d==-1) {
    free(par);
    return(NULL);
  }
  
  par=calloc(d,sizeof(double));
  char_replace(s,'_',' ');   
  i=0; p=s;
  while ( (i<d) && (p=strchr(p,' '), p!=NULL)) {
    p++;
    sscanf(p,"%le",par+i);
    i++;
  }
  char_replace(s,' ','_');   
  return(par);
}

//############################## FUNCTION #################################

double* load_input_pars_ddot(char *s,int d) {
  int i;
  char *p;
  static double *par;

  if (d==-1) {
    free(par);
    return(NULL);
  }
  
  par=calloc(d,sizeof(double));
  char_replace(s,':',' ');   
  i=0; p=s;
  while ( (i<d) && (p=strchr(p,' '), p!=NULL)) {
    p++;
    sscanf(p,"%le",par+i);
    i++;
  }
  char_replace(s,' ',':');   
  return(par);
}


//############################## FUNCTION #################################

int  param_string_get_value( char *a_str , char a_delim , int a_num , char a_type , void *a_out )
{

  char buff[MSTRLTH] , sel[MSTRLTH] , *p ;
  int ii , x ;

  if ( MSTRLTH-1 < strlen( a_str ) ) { return -1 ; }

  strcpy( buff , a_str ) ;

  char_replace( buff , a_delim , ' ' ) ;   
  ii = 0 ; p = buff ;
  while ( (ii<a_num) && (p=strchr(p,' '), p!=NULL)) { p++;  ii++; }
  if ( ( ii<a_num ) || ( x = sscanf(p,"%s",sel), (( EOF==x )||(0==x)) ) ) { return -2 ; }  ;
  switch (a_type)
  {
      case 'f': x = sscanf( p , "%e" ,  (float*)a_out  ) ; break ;
      case 'd': x = sscanf( p , "%le" , (double*)a_out ) ; break ;
      case 'i': x = sscanf( p , "%i" , (int*)a_out     ) ; break ;
      case 'c': x = sscanf( p , "%c" , (char*)a_out    ) ; break ;
      case 's': x = sscanf( p , "%s" , (char*)a_out    ) ; break ;
  }
  
  return(x);

}


//############################## FUNCTION #################################

static int get_opt_ver0( char *a_optstr , char *a_field , char *a_type , void *o_value )
{
    char sep[8] = ";" , eq[8] = "=" ;
    char xfield[SSTRLTH] = "" ;
    char xoptstr[LSTRLTH] = "" ;
    char *loc , *loc2 ;

    if ( SSTRLTH - 2  < strlen( a_field  ) ) { MODULE_ERROR("Too long field string.",-1) ; } 
    if ( LSTRLTH - 3  < strlen( a_optstr ) ) { MODULE_ERROR("Too long option string.",-1) ; } 
    strcpy( xfield , a_field ) ; strcat( xfield , eq ) ;
    strcat( strcat( strcat( xoptstr , sep ) , a_optstr ) , sep ) ;
    char_replace( xoptstr , ':'  , eq[0]  ) ;
    char_replace( xoptstr , ','  , sep[0] ) ;

    loc = strstr( xoptstr , xfield ) ;
    if ( NULL == loc ) { return -1 ; } ;

    loc2 = strstr( loc , eq  ) ; loc2[0] = ' ' ;
    loc2 = strstr( loc , sep ) ; loc2[0] = ' ' ;
   
    if ( NULL != o_value )
    {
        if      ( 0 == strcmp( a_type , "int" ) )    { sscanf( loc , "%*s %d"  , (int*)o_value ) ; }
        else if ( 0 == strcmp( a_type , "str" ) )    { sscanf( loc , "%*s %s"  , (char*)o_value ) ; }
        else if ( 0 == strcmp( a_type , "double" ) ) { sscanf( loc , "%*s %le" , (double*)o_value ) ; }
        else { MODULE_ERROR("No such type defined.",-1) ; } 
    }

   return 0 ; 

 }

//############################## FUNCTION #################################

int get_opt( char *a_optstr , char *a_field  , char *a_type, void *o_value )
{
    return get_opt_ver0( a_optstr , a_field , a_type , o_value ) ;
}


//############################## FUNCTION #################################

int is_str_all_alpha( char *s )
{
    int ii;
    for ( ii = 0 ; (unsigned)ii < strlen(s) ; ii++ ) 
    {
        if ( ! isalpha( s[ii] ) ) { return 0 ; }
    }
    return 1 ;
}



//############################## FUNCTION #################################

int append_to_int_array( int **a_array , int *a_dim , int a_from , int a_diff , int a_to )
{
    int NN ;
    int x , d ;
    int dim_old ;
    int ii ;

    dim_old = *a_dim ;

    d = a_to - a_from ;
    NN = d / a_diff + 1 ;
    
    *a_dim = dim_old + NN ;
    *a_array = realloc( *a_array , *a_dim * sizeof(int) ) ; 

    for ( ii = 0 ; ii < NN ; ii++ )
    {
        (*a_array)[dim_old+ii] =  a_from + a_diff * ii ;
    }

    return 0 ;
}


#define COLUMN_CHAR ":"

//############################## FUNCTION #################################

int extract_from_diff_to( char *a_str , int *a_from , int *a_diff , int *a_to )
{
    char *token  , *saveptr , str_cp[LSTRLTH] ;
    const char delimiters[] = COLUMN_CHAR ;
    int diff ;
    char opt_num_snp = 'n' ;

    strcpy( str_cp , a_str ) ;

    if ( NULL == strstr( str_cp , COLUMN_CHAR ) )
    {
        strcat(strcat( str_cp , ":1:" ) , a_str ) ; 
    }

    token = strtok_r ( str_cp , delimiters , &saveptr ); 
    sscanf( token , "%d" , a_from ) ;
    token = strtok_r ( NULL , COLUMN_CHAR , &saveptr ); 
    if ( '@' == token[0] )
    {
        sscanf( token+1 , "%d" , a_diff ) ;
        opt_num_snp = 'y' ;
    }
    else
    {
        sscanf( token , "%d" , a_diff ) ;
    }
    token = strtok_r ( NULL , COLUMN_CHAR , &saveptr );
    sscanf( token , "%d" , a_to ) ;

    if ( 'y' == opt_num_snp ) { *a_diff = ( *a_to - *a_from ) / ( *a_diff - 1 ) ; }

    return 0 ;
}



#define SEPARATOR_0_CHAR ","
//############################## FUNCTION #################################

int convert_strlist2intarray( char *a_strlist , int **a_array , int *a_dim ) 
{
    int *arr = NULL ;

    char *token  , *saveptr ;
    const char delimiters[] = SEPARATOR_0_CHAR ;
    char str_cp[LSTRLTH] ;
    int from , diff , to ;

    strcpy( str_cp , a_strlist ) ;

    token = strtok_r ( str_cp , delimiters , &saveptr ) ; 
    while ( NULL != token )
    {
        //printf( "SUB: %s\n" , token ) ;
        extract_from_diff_to( token , &from , &diff , &to ) ;
        append_to_int_array( a_array , a_dim , from , diff , to ) ;
        token = strtok_r ( NULL , delimiters , &saveptr ) ;
    }

    return 0 ;
   

}




//############################## FUNCTION #################################

char* str_concat_realloc( char *a_dest , char *a_src )
{
    if ( NULL == a_src ) { return a_dest ; }
    if ( NULL == a_dest )
    {
        a_dest = calloc( 1 , sizeof(char) ) ; strcpy( a_dest , "" ) ;
    }
    a_dest = realloc( a_dest , ( strlen( a_dest ) + strlen( a_src ) + 1 ) * sizeof(char) ) ;
    if (!a_dest) { fprintf( stderr , "No memory in str_concat_realloc.\n" ) ; exit(-1) ;} ;
    strcat( a_dest , a_src ) ;
    return a_dest ;
}



//#define MODULE_MAIN

#ifdef MODULE_MAIN
int main( void )
#else
//############################## FUNCTION #################################

int datatypes_test_main( void )
#endif
{
    int *test = NULL ; 
    int dim = 0 ;
    int ii ;


    append_to_int_array( &test , &dim , 2 , 3 , 7 ) ;
    append_to_int_array( &test , &dim , 2 , 3 , 8 ) ;
    append_to_int_array( &test , &dim , 2 , 3 , 9 ) ;
    append_to_int_array( &test , &dim , 1 , 4 , 11 ) ;
    printf( "DIM: %4d\n" , dim ) ;
    for ( ii = 0 ; ii < dim ; ii++ ) { printf( "%4d " , test[ii] ) ; }
    printf( "\n" ) ;


    int from , diff , to ; 
    char teststr[] = "1:3:7" ;
    //char teststr[] = "0:3:7" ;

    extract_from_diff_to( teststr , &from , &diff , &to ) ;
    printf( "%s %d : %d : %d\n" , teststr , from , diff , to ) ;

    // as to be implemented: 
    char input_snp[] = "-1,13:13:13,1:3:8,234,1:3:9,1:3:10,20:30:120,140:30:" ;
    int *snp_array = NULL , snp_dim = 0 ;
    int the_end = 200 ;

    char end_str[STRLTH] ;

    sprintf( end_str  , "%d\n" , the_end ) ;
    if ( 0 == strcmp( COLUMN_CHAR , input_snp + strlen(input_snp) - 1  ) ) { strcat( input_snp , end_str ) ; }

    convert_strlist2intarray( input_snp , &snp_array , &snp_dim ) ;

    printf( "DIM2: %4d\n" , snp_dim ) ;
    for ( ii = 0 ; ii < snp_dim ; ii++ ) { printf( "%4d " , snp_array[ii] ) ; }
    printf( "\n" ) ;

    return 0 ;

}





// ################################################################
// ####################   TEST FUNCTIONS   ########################
// ################################################################

//############################## FUNCTION #################################

int manip_track_dbase( char *a_todo , struct g_struct_track_dbase *a_dbase , ... )
{
    size_t size;
    FILE *stream ;
    char input_line[LSTRLTH] , xstr[256] , ystr[256] ;
    char *xtypes , *idstr , *x_str ;
    int ii , x_int , *x_out ;
    int xreturn = 0 ;
    double x_dbl ;

    va_list argp ;

    if ( NULL != a_dbase ) { va_start( argp , a_dbase ) ; } 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    if ( NULL != strstr( a_todo , "init" ) )
    {
        a_dbase->str = NULL ;
        a_dbase->dim = 0 ;
        if ( NULL != strstr( a_todo , "verbose" ) ) { printf( "Init dase.\n" ) ; }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    else if ( NULL != strstr( a_todo , "get" ) )
    {
        x_out = va_arg( argp , int* ) ;
        xtypes = va_arg( argp , char* ) ;
        idstr = NULL ;
        stream = fmemopen( xtypes , strlen(xtypes) , "r" ) ;
        while ( EOF != fscanf( stream , "%s" , xstr ) ) 
        { 
            if ( 0 == strcmp( xstr , "d" ) )
            {
                x_int = va_arg( argp , int ) ;
                sprintf( ystr , "%d" , x_int ) ;
            }
            if ( 0 == strcmp( xstr , "le" ) )
            {
                x_dbl = va_arg( argp , double ) ;
                sprintf( ystr , "%le" , x_dbl ) ;
            }
            if ( 0 == strcmp( xstr , "s" ) )
            {
                x_str = va_arg( argp , char* ) ;
                sprintf( ystr , "%s" , x_str ) ;
            }
            idstr = str_concat_realloc( str_concat_realloc( idstr , ystr ) , " " ) ; 
        }
        fclose( stream ) ;
        for ( ii = 0 ; ii < a_dbase->dim ; ii++ )
        {
            if ( 0 == strcmp( idstr , a_dbase->str[ii] ) )
            {
                break ;
            }
        }
        if ( ii < a_dbase->dim ) 
        { 
            if ( NULL != strstr( a_todo , "verbose" ) ) { printf( "Found:  num %d str \"%s\"\n" , (a_dbase->dim) , idstr ) ; }
            free( idstr ) ; *x_out = ii ; 
        }
        else
        {
            if ( NULL != strstr( a_todo , "verbose" ) ) { printf( "Adding: num %d str \"%s\"\n" , (a_dbase->dim) , idstr ) ; }
            a_dbase->str = realloc( a_dbase->str , ( (a_dbase->dim) + 1 ) * sizeof(char**) ) ;
            a_dbase->str[a_dbase->dim] = idstr ;
            (a_dbase->dim)++ ;
            *x_out = ( a_dbase->dim ) - 1 ;
        }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    else if ( NULL != strstr( a_todo , "status" ) )
    {
        printf( "Status:\n" ) ;
        for ( ii = 0 ; ii < a_dbase->dim ; ii++ ) { printf( "   %5d: \"%s\"\n" , ii , a_dbase->str[ii] ) ; }
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    else if ( NULL != strstr( a_todo , "test" ) )
    {
        struct g_struct_track_dbase xdbase ; 
        int loc ;
        ii = 3 ;
        manip_track_dbase( "init,verbose"   , &xdbase ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s" , "s1"  ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s" , "s1"  ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s d" , "s2" , ii ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s" , "s2"  ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s d" , "s2" , ii ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ; 
        manip_track_dbase( "get,verbose"    , &xdbase , &loc , "s" , "s2"  ) ;
        printf("%d\n" , loc) ; manip_track_dbase( "status" , &xdbase ) ;
    }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    else { xreturn = -1 ; } 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    va_end( argp ) ;
    return xreturn ;
}


#undef MODULE_ERROR

