///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


/* The following is needed because of timespec, CLOCK_REALTIME, ... , if you compile with -std=c99 */
#if __STDC_VERSION__ >= 199901L
#define _XOPEN_SOURCE 600
#else
#define _XOPEN_SOURCE 500
#endif /* __STDC_VERSION__ */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>

#include "basics.h"
#include "datatypes.h"



// ##########################################################################


#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}


//############################## FUNCTION #################################

static double diff_time(struct timespec start,struct timespec end) 
{
  return( (double)(end.tv_sec - start.tv_sec) + 
          (double)(end.tv_nsec - start.tv_nsec) / 1000000000 );
}

//############################## FUNCTION #################################

double myclock(double runtime){
/* runtime: -1 : gives back the act time
   else: sets the clock zero, and set def. timevalue to runtime
*/
  //static clock_t cl_start, cl_act;
  static double realruntime;
  int intruntime;
  static struct timespec t_start, t_act;

  if (floor(runtime)==runtime)
    intruntime=floor(runtime);
  else
    intruntime=1;

  switch (intruntime) {
  case -1:                                // gives back the real runtime
    //cl_act=clock();
    //return((double)(cl_act-cl_start)/CLOCKS_PER_SEC+realruntime);
    clock_gettime(CLOCK_REALTIME, &t_act);
    return( diff_time(t_start,t_act) );
  default:                                // sets the runtime
    //cl_start=clock();
    clock_gettime(CLOCK_REALTIME, &t_start);
    realruntime=runtime;
    return(runtime);
  }
}


//############################## FUNCTION #################################

char* cd_currdate(void) {
  time_t curtime;
  struct tm *loctime;
  /* Get the current time.  */
  curtime = time (NULL);
  /* Convert it to local time representation.  */
  loctime = localtime (&curtime);
  /* Print out the date and time in the standard format.  */
  return(asctime (loctime));
}

static int  TICTOC_NUM_JOBS=500 ;



//############################## FUNCTION #################################

double cd_tictoc(int tic,char *job,int thread,int maxthreads) {
  static int init=1;
  int id;
  static double *tfrom, *tcum, *rtcum, dt, rdt;
  double *pfrom, *pcum, act, *rpcum;
  static struct timespec *rtfrom;
  struct timespec *rpfrom, ract;

  if (init) {
    tfrom  = calloc(TICTOC_NUM_JOBS*maxthreads,sizeof(double));
    tcum   = calloc(TICTOC_NUM_JOBS*maxthreads,sizeof(double));
    rtfrom  = calloc(TICTOC_NUM_JOBS*maxthreads,
                     sizeof(struct timespec));
    rtcum   = calloc(TICTOC_NUM_JOBS*maxthreads,
              sizeof(double));
    init = 0;
  }
  
  if (strcmp(job,"ts")==0) {
    id=0;
  }
  if (strcmp(job,"qp")==0) {
    id=1;
  }
  if (strcmp(job,"si")==0) {
    id=2;
  }
  if (strcmp(job,"--")==0) {
    id=3;
  }
  if (strcmp(job,"fs")==0) {
    id=4;
  }
  if (strcmp(job,"wr")==0) {
    id=5;
  }

  if ( strcmp ( job , "rec" ) == 0 ) { id = 6 ; }


  if (strcmp(job,"test1")==0) {
    id=201;
  }
  if (strcmp(job,"test2")==0) {
    id=202;
  }
  if (strcmp(job,"test3")==0) {
    id=203;
  }
  if (strcmp(job,"test4")==0) {
    id=204;
  }
  if (strcmp(job,"test5")==0) {
    id=205;
  }
  if (strcmp(job,"test6")==0) {
    id=206;
  }
  if (strcmp(job,"test7")==0) {
    id=207;
  }
  if (strcmp(job,"test8")==0) {
    id=208;
  }
  if (strcmp(job,"test9")==0) {
    id=209;
  }
  if (strcmp(job,"sub1")==0) {
    id=301;
  }
  if (strcmp(job,"sub2")==0) {
    id=302;
  }
  if (strcmp(job,"sub3")==0) {
    id=303;
  }
  pfrom  = tfrom  + (TICTOC_NUM_JOBS*thread+id);
  pcum   = tcum   + (TICTOC_NUM_JOBS*thread+id);
  rpfrom  = rtfrom  + (TICTOC_NUM_JOBS*thread+id);
  rpcum   = rtcum   + (TICTOC_NUM_JOBS*thread+id);

  switch (tic) {

    case 0: // tic normal
      *pfrom = (double)(clock())/CLOCKS_PER_SEC;
      clock_gettime(CLOCK_REALTIME, rpfrom);
      return(0);
      break;

    case 1: // toc normal
      act = (double)(clock())/CLOCKS_PER_SEC;
      dt = act - *pfrom ;
      *pcum += dt ; 
      clock_gettime(CLOCK_REALTIME, &ract);
      rdt = diff_time(*rpfrom,ract) ;
      *rpcum += rdt ; 
      return ( rdt ) ;
      break;
    
    case 2:
      return(rdt);

    case 10: // tic cumulative & normal
      *pfrom = (double)(clock())/CLOCKS_PER_SEC;
      *pcum  = 0; 
      *rpcum = 0;
      return(0);
      break;

    case 11: // toc cumulative & normal
      return ( *rpcum ) ;
      break;

    case 12: // toc cumulative & normal
      return ( *rpcum ) ;
      break;

  }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0 ;
}


//############################## FUNCTION #################################

int xmdyn_clock( const char *comm , ... ) 
{
/** Example usage sequence: */
/**                         */
/**  xmdyn_clock( "tic"     , "xy" ) ;  */
/**  xmdyn_clock( "toc-add" , "xy" ) ;  */
/**      xmdyn_clock( "tic"     , "xy" ) ;  */
/**      xmdyn_clock( "toc-add" , "xy" ) ;  */
/**  xmdyn_clock( "report"  , &my_string ) ;  */
/**  xmdyn_clock( "clear" ) ;  */
/**                         */
//*  NOTE: a `clear' is always needed between two `report's */

   static char **list = NULL ; 
   static double *elapsed = NULL ;
   static double *from = NULL ;
   static int num = 0 ;
   static char *logstr = NULL ;

   va_list valist;

   int strlth = 64 ;
   int ii , found ;
   char *xs, **pps , xstr1[strlth], xstr2[strlth], xstr3[strlth] ; 
   double xtime ;


   /* initialize valist for num number of arguments */
   va_start( valist, comm );
//   va_end(valist); return 0 ;

   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
   if ( strstr( comm , "tic" ) != NULL )
   {
      xs = va_arg(valist, char*) ;
      found = -1 ; for ( ii = 0 ; ii < num ; ii++ ) { if ( strcmp( list[ii] , xs ) == 0 ) { found = ii ; } }
      if ( -1 == found ) 
      { 
          found = num ; num++ ; 
          list = realloc( list , num * sizeof(char*) ) ;  
          list[found] = calloc( strlth , sizeof(char) ) ;
          strncpy( list[found] , xs , strlth - 1 ) ;
          from = realloc( from , num * sizeof(double) ) ;
          elapsed = realloc( elapsed , num * sizeof(double) ) ;
          elapsed[found] = 0 ;
      }
      from[found] = (double)(clock()) / CLOCKS_PER_SEC ; // clock_gettime(CLOCK_REALTIME, rpfrom);
   }
   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
   else if ( strstr( comm , "toc" ) != NULL )
   {
      xs = va_arg(valist, char*) ;
      found = -1 ; for ( ii = 0 ; ii < num ; ii++ ) { if ( strcmp( list[ii] , xs ) == 0 ) { found = ii ; } }
      if ( -1 == found ) { MODULE_ERROR("toc is called before tic." , -1 ) ; }
      xtime = (double)(clock()) / CLOCKS_PER_SEC ;

   }
   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
   if ( strstr( comm , "add" ) != NULL )
   {
       elapsed[found] += ( xtime - from[found] ) ;
   }
   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
   if ( strstr( comm , "report" ) != NULL )
   {
       if ( logstr != NULL ) { MODULE_ERROR( "logstr has not been deallocated." , -1 ) ; }
       logstr = calloc( 2 , sizeof(char) ) ; strcpy( logstr , "" ) ;
       for ( ii = 0 ; ii < num ; ii++ ) 
       {
           sprintf( xstr1 , "% 12.6f" , elapsed[ii] ) ;
           sprintf( xstr2 , "%-15s:" , list[ii] ) ;
           sprintf( xstr3 , "%-9s" , "" ) ;
           logstr = str_concat_realloc( str_concat_realloc( str_concat_realloc( str_concat_realloc( str_concat_realloc( logstr , xstr3 ) , xstr2 ) , xstr1 ) , "s" ) , "\n" ) ;
       }
       pps = va_arg(valist, char**) ;
       *pps = logstr ;
   }
   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
   if ( strstr( comm , "clear" ) != NULL )
   {
       for ( ii = 0 ; ii < num ; ii++ ) { elapsed[ii] = 0 ; }
       if ( NULL != logstr ) { free( logstr ) ; logstr = NULL ; } 
   }
   // = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

   /* clean memory reserved for valist */
   va_end(valist);
   return 0 ;

}

