///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "basics.h"

#define   STRLTH    1024


//############################## FUNCTION #################################

int dbg_stopper ( int x ) 
{
    if ( 0 < x ) { printf("We stop\n") ; }
    switch (x)
    { 
    case 1:
        x = x ; 
    case 2:
        x = x ; 
    case -1:
        x = x ; 
    case -2:
        x = x ; 
    }
    return 0 ;
}


//############################## FUNCTION #################################

int dbg_print(char typ,...) {
// dbg_print(type,file,mode,variable)
//  type: 'i','d','c','s',  or 'p'(pause)
//  mode: "" or "w"(write) or "n"(newline) or "L"(long prec) or 
//        string of any combination of these
//  file: "":stdout , else filename

  int newline=0, clearf=0, pause=0;
  char *f,*appnl,prec[STRLTH];
  FILE *fp;
  
  int vint;
  double vdouble;
  char vchar,*vstr;
  strcpy(prec,"%le ");
  
  va_list ap;
  
  if (typ=='p') {
    printf("Hit enter to continue!\n");
    getchar();
    return(0);
  }
  
  va_start(ap,typ);
  f=va_arg(ap,char*);
  appnl=va_arg(ap,char*);
  switch (typ){
   case 'i':
    vint=va_arg(ap,int);
    break;
   case 'd':
    vdouble=va_arg(ap,double);
    break;
   case 'c':
    vchar=va_arg(ap,int);
    break;
   case 's':
    vstr=va_arg(ap,char*);
    break;
  }
  va_end(ap);
   
  if (strchr(appnl,'n')!=NULL){
    newline=1;
  }
  if (strchr(appnl,'w')!=NULL){
    clearf=1;
  }
  if (strchr(appnl,'p')!=NULL){
    pause=1;
  }
  if (strchr(appnl,'L')!=NULL){
    strcpy(prec,"%1.15le ");
  }
  
  if (strlen(f)==0) {
    switch (typ) {
     case 'd':
       printf(prec,vdouble);
       break;
     case 'i':
       printf("%d ",vint);
       break;
     case 'c':
       printf("%c ",vchar);
       break;
     case 's':
       printf("%s ",vstr);
       break;
    }
    if (newline) {
      printf("\n");
    }
  }
  else {
    if (clearf) {
      fp=fopen(f,"w");
    }
    else {
      fp=fopen(f,"a");
    }
    switch (typ) {
     case 'd':
       fprintf(fp,prec,vdouble);
       break;
     case 'i':
       fprintf(fp,"%d ",vint);
       break;
     case 'c':
       fprintf(fp,"%c ",vchar);
       break;
     case 's':
       fprintf(fp,"%s ",vstr);
       break;
    }
    if (newline) {
      fprintf(fp,"\n");
    }
    fclose(fp);
  } 
  if (pause) {
    printf("Hit enter to continue!\n");
    getchar();
  }

  return 0 ;
}


//############################## FUNCTION #################################

int debug_pause( char *a_message ) 
{
    fprintf( stderr , "%s " , a_message ) ;
    fflush( stderr ) ;
    scanf( "%*c" ) ;
    //fprintf( stderr , "\n" ) ;
    return 0 ;
}


//############################## FUNCTION #################################

int debug_print( char *a_message ) 
{
    fprintf( stderr , "%s\n" , a_message ) ;
    fflush( stderr ) ;
    //scanf( "%*c" ) ;
    //fprintf( stderr , "\n" ) ;
    return 0 ;
}


//############################## FUNCTION #################################

int dbg_warning_note_on_code( char *fname , int line , char *note )
{
    static int   *dbase_line = NULL ;
    static char **dbase_file = NULL ;
    static int    dbase_num  = 0 ;
    static int init = 1 ;
    int ii  , size = 500 ;

//    return 0 ;
//
    if ( init ) 
    {
        dbase_line = calloc( size+1 , sizeof(int) ) ;
        dbase_file = calloc( size+1 , sizeof(char*) ) ;
        for ( ii = 0 ; ii < size ; ii++ ) { dbase_file[ii] = calloc( STRLTH , sizeof(char) ) ;  }
        init = 0 ;
    }

    if ( dbase_num == size ) { printf("DBG print is full\n") ; return -1 ; }
    for  ( ii = 0 ; ii < dbase_num ; ii++ ) 
    {
        if (  ( line == dbase_line[ii] ) && ( 0 == strcmp( dbase_file[ii] , fname ) ) ) { return -1 ; }
    }
    dbase_line[dbase_num] = line ;
    strcpy( dbase_file[dbase_num] , fname ) ;
    dbase_num++ ;
    fprintf( stderr , "NOTE: %s\n" , note ) ;
    //fprintf( stderr , "WARNING! -- %s / %d : %s\n" , fname , line , note ) ;
    //LOG_RED(stderr,"WARNING!") ; fprintf( stderr , " -- %s / %d : %s\n" , fname , line , note ) ;
    return 0 ;
}





