///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <assert.h>

#include "basics.h"

#include "filesystem.h"



//############################## FUNCTION #################################

int sys_path_operations( char *a_todo , char *a_str ) 
{
    if ( strcmp( a_todo , "~" ) == 0 ) 
    {
        wordexp_t res ;
        wordexp( a_str , &res , 0 ) ;
        strcpy( a_str , res.we_wordv[0] ) ;
        wordfree( &res ) ; 
    }
    return 0 ;
}


//############################## FUNCTION #################################

int nofile(char *s){
  printf("There's no %s file to open!\n",s);
  return(-1);
}


//############################## FUNCTION #################################

int check_directory_exists( char *a_path )
{
    struct stat s ;
    int err = stat( a_path , &s ) ;
    if ( -1 == err ) 
    {
        if ( ENOENT == errno ) { return(0) ; }         /* does not exist */
        else { perror("stat") ;  exit(1) ;  }
    } 
    else 
    {
        if ( S_ISDIR( s.st_mode ) ) { return(1) ; }    /* it is a folder */
        else { return( -1 ) ; }                        /* exists but not a folder */
    }
    return( -2 ) ;
}



//############################## FUNCTION #################################

int mkpath1( char *path , mode_t mode ) 
{
    char *p ;
    assert( path && *path) ;
    for ( p = strchr( path + 1 , '/' ) ; p ; p = strchr( p + 1 , '/' ) ) 
    {
        *p = '\0' ;
        if ( mkdir( path , mode ) == -1 ) 
        {
            if ( errno != EEXIST ) { *p = '/' ; return -1 ; }
        }
        *p = '/' ;
    }
    return 0;
}


//############################## FUNCTION #################################

int mkpath( char *path )
{
    //return( mkpath1( path , S_IRWXU ) ) ;
    return( mkpath1( path , S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH ) ) ;
    //return( mkpath1( path , S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IWGRP | S_IXGRP | S_IROTH | S_IXOTH ) ) ;
    //return( mkpath1( path , S_IRUSR | S_IWUSR | S_IXUSR ) ) ;
}


//############################## FUNCTION #################################

int dircreate(char *diri){

  char path[strlen(diri)+2] ;
  strcpy( path , diri ) ;
  strcat( path , "/" ) ;
  if ( check_directory_exists( path ) == 0 ) {  return( mkpath( path ) ) ; }
  return -1 ;

}


//############################## FUNCTION #################################

int fileop_num_of_lines( char *fname ) 
{
    FILE *fp ;
    fp = fopen ( fname , "r" ) ;
    if ( ! fp ) { fprintf( stderr , "No such file : %s !!\n" , fname ) ; return(-1) ; }

    char xstr[3] ;
    int cc, lines = 0 ;
 
    while ( ( cc = fgetc(fp) ) != EOF )
    {
      sprintf( xstr , "%x" , cc ) ;
      if ( ( strlen( xstr ) ) == 1 ) { ++lines ; }
    }
    fclose( fp ) ;
 
    return lines;
}


//############################## FUNCTION #################################

char get_filesystem_object_type( char *path )
{
    struct stat s ;
    if ( stat( path , &s ) == 0 )
    {
        if ( 0 ) {}
        else if ( s.st_mode & S_IFDIR ) { return( 'd' ) ; } // directory
        else if ( s.st_mode & S_IFREG ) { return( 'f' ) ; } // file
        else                            { return( '-' ) ; } // otherwise
    }
    else { return( 'E' ) ; }  // error
}


//############################## FUNCTION #################################

int load_stream( char **pbuffer , char *cmd , char what )
{

  size_t size;
  FILE *stream , *fp ;
  char input_line[LSTRLTH] ;

  // Load in bp
  stream = open_memstream( pbuffer , &size ) ;

  if      ( what == 'p' ) { fp = popen( cmd , "r" ) ; }
  else if ( what == 'f' ) { fp = fopen( cmd , "r" ) ; }
  else    { fprintf( stderr , "Neither fopne nor popen\n" ) ; return -1 ; } 

  if ( NULL == fp ) 
  {
      fprintf( stderr , "Failed to open file / run command %s\n" , cmd );
      return -1 ;
  }

  while ( fgets( input_line , sizeof( input_line )-1 , fp ) != NULL )
  {
      fprintf( stream , "%s" , input_line ) ;
  }
  fclose ( fp ) ;
  fclose(stream) ; 

  return 0;

 }



