///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_interp.h>

#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "basics.h"

#include "random.h"
#include "numeric.h"





//############################## FUNCTION #################################

int coordtrans_xyz2sphere(double x,double y,double z,
                          double *r,double *phi,double *th)
{
  *r = sqrt( x*x + y*y + z*z );
  //*th = atan2 (z,sqrt(x*x+y*y));
  //*phi = atan2 (y,x);
  *th = acos(z/(*r));
  *phi = ( 0<=x ? 
           asin(y/sqrt(x*x+y*y)) :
	   PI - asin(y/sqrt(x*x+y*y)) ) ;
  return(0);
}


//############################## FUNCTION #################################

double linapprox(double x1,double x2,double y1,double y2,double x){
  return( (x1==x2) ? ((y1+y2)/2.0) : ((y2-y1)/(double)(x2-x1)*(x-x1)+y1) );
}


//############################## FUNCTION #################################

void inversematrix3(double (*Em)[3],double (*Fm)[3]){
  double deter;

  deter=Em[0][0]*Em[1][1]*Em[2][2]+Em[1][0]*Em[2][1]*Em[0][2]+
    Em[2][0]*Em[0][1]*Em[1][2]-Em[0][0]*Em[2][1]*Em[1][2]-
    Em[1][0]*Em[0][1]*Em[2][2]-Em[2][0]*Em[1][1]*Em[0][2];

  Fm[0][0]=(Em[1][1]*Em[2][2]-Em[2][1]*Em[1][2])/deter;
  Fm[0][1]=-(Em[0][1]*Em[2][2]-Em[2][1]*Em[0][2])/deter;
  Fm[0][2]=(Em[0][1]*Em[1][2]-Em[1][1]*Em[0][2])/deter;
  Fm[1][0]=-(Em[1][0]*Em[2][2]-Em[2][0]*Em[1][2])/deter;
  Fm[1][1]=(Em[0][0]*Em[2][2]-Em[2][0]*Em[0][2])/deter;
  Fm[1][2]=-(Em[0][0]*Em[1][2]-Em[1][0]*Em[0][2])/deter;
  Fm[2][0]=(Em[1][0]*Em[2][1]-Em[2][0]*Em[1][1])/deter;
  Fm[2][1]=-(Em[0][0]*Em[2][1]-Em[2][0]*Em[0][1])/deter;
  Fm[2][2]=(Em[0][0]*Em[1][1]-Em[1][0]*Em[0][1])/deter;
}


//############################## FUNCTION #################################

void  coordtrafo(double *to,double *from,double *zaxes,double *xaxes,
		 char where){

  double zaxisnorm,e1norm;
  double *e1,*e2,*e3;
  double *f1,*f2,*f3;
  double (*Em)[3],(*Fm)[3];
  int i;

  Em=calloc(9,sizeof(double));  //if (!Em) nomem();  // TODO
  Fm=calloc(9,sizeof(double));  //if (!Fm) nomem();  // TODO
  e1=Em[0]; e2=Em[1]; e3=Em[2];
  f1=Fm[0]; f2=Fm[1]; f3=Fm[2];

  zaxisnorm=sqrt(vectnorm2(zaxes));
  vecttimesscalar(e3,zaxes,1/zaxisnorm);
  for (i=0;i<3;++i)
    e1[i]=xaxes[i]-(zaxes[i]/vectnorm2(zaxes))*scalarproduct(zaxes,xaxes);
  e1norm=sqrt(vectnorm2(e1));
  vecttimesscalar(e1,e1,1/e1norm);
  crossproduct(e2,e3,e1);

  if (where=='b'){
    to[0]=e1[0]*from[0]+e2[0]*from[1]+e3[0]*from[2];
    to[1]=e1[1]*from[0]+e2[1]*from[1]+e3[1]*from[2];
    to[2]=e1[2]*from[0]+e2[2]*from[1]+e3[2]*from[2];
  }

  if (where=='t'){
    inversematrix3(Em,Fm);
    to[0]=f1[0]*from[0]+f2[0]*from[1]+f3[0]*from[2];
    to[1]=f1[1]*from[0]+f2[1]*from[1]+f3[1]*from[2];
    to[2]=f1[2]*from[0]+f2[2]*from[1]+f3[2]*from[2];
  }

  free(Em); free(Fm);
}





//############################## FUNCTION #################################

void rotatepos( double (*r)[3] , int N , double phiin , double thetain ){
  double d[3],*pr,PHI,THETA;
  int i;

  PHI = phiin*PI/180.0;
  THETA = thetain*PI/180.0;

  for ( i=0 ; i < N ; ++i ) {
    pr=r[i];

    d[0]=cos(PHI)*pr[0]+sin(PHI)*pr[1];
    d[1]=-sin(PHI)*pr[0]+cos(PHI)*pr[1];
    d[2]=pr[2];

    pr[0]=d[0];
    pr[1]=cos(THETA)*d[1]+sin(THETA)*d[2];
    pr[2]=-sin(THETA)*d[1]+cos(THETA)*d[2];
  }
}


//############################## FUNCTION #################################

int binom(int n,int k){
  double result=1,i;

  for (i=0;i<k;++i)
    result*=( (double)(n-i)/(double)(k-i) );

  return((int)result);
}


//############################## FUNCTION #################################

double interpolate(double *x,double *y,int dim,double xx,char method){
  int i,low,high,maxind=0,minind=0;

  switch (method) {

   case 'l':
    low=high=-1;
    for (i=0;i<dim;++i){
      if ( (x[i]<=xx) && ( (low==-1) || (x[low]<x[i]) )  )
        low=i;
      if ( (xx<=x[i]) && ( (high==-1) || (x[i]<x[high]) ) )
        high=i;
      if (x[i]<x[minind])
        minind=i;
      if (x[maxind]<x[i])
        maxind=i;
    }

    if (low==-1){
      //low=high=minind;
      return(y[minind]);
    }
    if (high==-1){
      //low=high=maxind;
      return(y[maxind]);
    }

    return(linapprox(x[low],x[high],y[low],y[high],xx));
    break;

   }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0;
}


//############################## FUNCTION #################################

double draw_poisson(double w){
  double y;

  if (w==0)
    return(0);

  y=randnum(0);
  if (y==1)
    return(draw_poisson(w));
  else
    return(-1.0/w*log(-y+1));
}



//############################## FUNCTION #################################

double drawrandomdistr(double (*actdistr)(double,double*),double *params,
                       int type,int nrm){
  int iternum=20,numpoints=100000;
  double dlimid=1,ulimid=2,params0;;

  int i,dlimind,ulimind;
  double *elems,res,r,nlim,ulim,dlim;

  params0=params[0];

  switch (type){
   case 0 :  //reserved, but not implemented yet
   //
    return(0);
    break;

   case 1 :
   // x=actdistr^(-1)(F)
    return((*actdistr)(randnum(0),params));
    break;

   case 2 :
   // F=actdistr(x) ; F[max] can differ from 1
    params[0]=dlimid;
    dlim=(*actdistr)(0,params);
    params[0]=ulimid;
    ulim=(*actdistr)(0,params);
    params[0]=params0;

    params0=(*actdistr)(ulim,params);
    params0=(*actdistr)(dlim,params);

    r=( (*actdistr)(ulim,params)-(*actdistr)(dlim,params)) * randnum(0)+
         (*actdistr)(dlim,params);

    for (i=0;i<iternum;++i){
      nlim=dlim+(ulim-dlim)/2.0;
      res=(*actdistr)(nlim,params);
      if ((*actdistr)(nlim,params)>=r)
        ulim=nlim;
      else
        dlim=nlim;
    }
    params0=linapprox( (*actdistr)(dlim,params),(*actdistr)(ulim,params),
                        dlim,ulim,r);
    return( linapprox( (*actdistr)(dlim,params),(*actdistr)(ulim,params),
                        dlim,ulim,r) );
    break;

   case 3 :
   // f=actdistr(x) ; int(f) can differ from 1
    params[0]=dlimid;
    dlim=(*actdistr)(0,params);
    params[0]=ulimid;
    ulim=(*actdistr)(0,params);
    params[0]=params0;
    elems=calloc(numpoints,sizeof(double));
    elems[0]=(*actdistr)(dlim+(ulim-dlim)/(numpoints-1)*0,params)*0;
    for (i=1;i<numpoints;++i){
      elems[i]=elems[i-1]+(*actdistr)(dlim+(ulim-dlim)/(numpoints-1)*i,
                                      params);
    }
    r=randnum(0)*elems[numpoints-1];

    ulimind=0;
    while ( (elems[ulimind]<=r) && (ulimind<numpoints-1) ){
      ulimind++;
    }
    dlimind=ulimind-1;

    res=linapprox( (*actdistr)(dlim+(ulim-dlim)/
                    (double)(numpoints-1)*dlimind,params),
                   (*actdistr)(dlim+(ulim-dlim)/
		    (double)(numpoints-1)*ulimind,params),
	           dlim+(ulim-dlim)/(double)(numpoints-1)*dlimind,
		   dlim+(ulim-dlim)/(double)(numpoints-1)*dlimind,r );

    free(elems);
    return(res);
    break;
   default:
     ;
  }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0 ;
}




//############################## FUNCTION #################################

double vmean(int dim,char *type,void* vdata)
{
  double  M,*ddata;
  int i;
  M=0;
  if (strcmp(type,"double")==0)
  {
    ddata = (double*)vdata;
    for (i=0;i<dim;++i) { M += ddata[i]; }
    M /= (double)dim;
  }
  else if (strcmp(type,"int")==0)
  {
  }
  return(M);
}


//############################## FUNCTION #################################

double vstd1(int dim,char *type,void* vdata)
{
  int i;
  if (strcmp(type,"double")==0)
  {
    double M, S, *data, dx;
    data = (double*)vdata;
    M = vmean(dim,type,data);
    S=0;
    for (i=0;i<dim;++i) { dx = (data[i]-M) ; S += dx*dx ;}
    S = sqrt( S / (dim-1.0) );
    return(S);
  }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return 0 ;
}


//############################## FUNCTION #################################

double vstd(int dim,char *type,void* vdata)
{
  return( vstd1(dim,type,vdata) );
}


 
//############################## FUNCTION #################################

int init_interp1( mytype_interp1 *data , double *x , double *y , int dim , char* interp1_type ) 
{
  int ii ;
  data -> dim = dim ;
  data -> x = calloc( dim , sizeof( double ) ) ;
  data -> y = calloc( dim , sizeof( double ) ) ;
  for ( ii=0 ; ii<dim ; ++ii ) {
         ( data -> x )[ii] = x[ii] ;
         ( data -> y )[ii] = y[ii] ;
  }
  strcpy( data -> type , interp1_type ) ;
  data -> acc = gsl_interp_accel_alloc ();
  if ( strcmp( data -> type , "cspline" ) == 0 ) {
    data -> spline = gsl_spline_alloc( gsl_interp_cspline , dim ) ;
    gsl_spline_init ( data -> spline , x , y , dim );
    return( 0 ) ;
  }
  if ( strcmp( data -> type , "linear" ) == 0 ) {
    data -> interpolation = gsl_interp_alloc( gsl_interp_linear , dim );
    gsl_interp_init( data -> interpolation , data -> x , data -> y , dim );
    return( 0 ) ;
  }
  return( -1 ) ;
}


//############################## FUNCTION #################################

int free_interp1( mytype_interp1 *data )
{
  free( data -> x ) ;
  free( data -> y ) ;
  if ( strcmp( data -> type , "cspline" ) == 0 )
    gsl_spline_free ( data -> spline );
  if ( strcmp( data -> type , "linear" ) == 0 )
        gsl_interp_free( data -> interpolation ) ;
  gsl_interp_accel_free (data -> acc );

  /* TODO: return gsl_... error */
  return 0 ;

}


//############################## FUNCTION #################################

double do_interp1( mytype_interp1 *data , double xi ) 
{
 double yi ;
  if ( strcmp( data -> type , "cspline" ) == 0 )
    yi = gsl_spline_eval ( data -> spline , xi , data -> acc ) ;
  if ( strcmp( data -> type , "linear" ) == 0 )
    yi = gsl_interp_eval( data -> interpolation , data -> x , data -> y , xi , data -> acc );
  return( yi ) ;
}



//############################## FUNCTION #################################

int full_interp1( double *x , double *y , int dim , double *xi , double *yi , int dimi , char* interp1_type ) 
{
    char itype[100] ;
    mytype_interp1 ref_data ;
    int ii ;
    if ( strcmp( interp1_type , "def" ) == 0 ) { strcpy( itype , "cspline" ) ; }
    init_interp1( &ref_data , x , y , dim , itype ) ; // cspline or linear
    for ( ii = 0 ; ii < dimi ; ii++ ) { yi[ii] = do_interp1( &ref_data , xi[ii] ) ; }
    free_interp1( &ref_data ) ;
    return 0 ;
}




//############################## FUNCTION #################################

int array_max( void *a_m , void *a_x , long long int dim , char *type )
{
    long long int ii ;

    if ( strcmp( type , "lli" ) == 0 ) 
    {
        long long int m , *v = (long long int*)a_x ;
        m = v[0] ;
        for ( ii = 0 ; ii < dim ; ii++ ) { if ( m < v[ii] ) { m = v[ii] ; } }
        *((long long int*)a_m) = m ;
    }

    return 0 ;

}


//############################## FUNCTION #################################

int math_solve_sec_order_alg_eq( double a , double b , double c , double *x1 , double *x2 )
{
 
    double D ; 
    //*  Equation: a * x^2 + b * x + c = 0    */

    D = b*b - 4*a*c ;
    if ( D < 0 ) { return 0 ; }
    if ( 0 == a ) 
    { 
       if ( 0 == b ) { return 0 ; }
       else { *x1 = *x2 = - c / b ; return 1 ; }
    }
    if ( 0 == D ) { *x1 = *x2 = -b / ( 2. * a ) ; return 1 ; }
    else
    {
        *x1 = ( - b - sqrt(D) ) / ( 2. * a ) ;
        *x2 = ( - b + sqrt(D) ) / ( 2. * a ) ;
        if ( *x1 > *x2 ) { double xx ; xx = *x1 ; *x1 = *x2 ; *x2 = xx ; } 
        return 2 ; 
    }
    return 0 ;
}

