///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <gsl/gsl_errno.h>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_interp.h>



#ifndef M_PI
#define M_PI 3.141592653589793238
#endif

#ifndef PI
#define PI 3.141592653589793238
#endif

#ifndef M_E
#define M_E 2.71828182845905
#endif

#ifndef NAN
#define NAN (sqrt(-1))
#endif



#define vectorswap(x1,x2) { void *p ; p=x1 ; x1=x2 ; x2=p ; }

#define vectnorm2(x) ((x)[0]*(x)[0]+(x)[1]*(x)[1]+(x)[2]*(x)[2])
#define vect3equal(x1,x2) {x1[0]=x2[0];x1[1]=x2[1];x1[2]=x2[2];}
#define vectequal(x1,x2,dim) {int i; for (i=0;i<dim;++i) x1[i]=x2[i];}
#define vector_get_maxindex_nonzero(v,dim,ind) {int i; ind=-1 ; for (i=0; i<dim ; ++i) if (v[i]>0) ind=i; } 

#define vectndiff(xe,x1,x2,dim) {int i; for (i=0;i<dim;++i) xe[i]=x1[i]-x2[i];}
#define vectndiv(xe,x1,x2,dim) {int i; for (i=0;i<dim;++i) xe[i]=x1[i]/x2[i];}
#define vectnfabs(xe,x1,dim) {int i; for (i=0;i<dim;++i) xe[i]=fabs(x1[i]);}
#define vectnequal(xe,x1,dim) {int i; for (i=0;i<dim;++i) xe[i]=(x1[i]);}
#define vectnnozero(xe,dim2,x1,dim) {int i; dim2=0; for (i=0;i<dim;++i) if ((x1)[i]!=0) {(xe)[dim2]=(x1)[i];dim2++;}}

#define vectminus(y,x1,x2) {(y)[0]=(x1)[0]-(x2)[0];(y)[1]=(x1)[1]-(x2)[1];\
                            (y)[2]=(x1)[2]-(x2)[2];}

#define VECTOR3_DIFF(y,x1,x2) { (y)[0]=(x1)[0]-(x2)[0]; (y)[1]=(x1)[1]-(x2)[1]; (y)[2]=(x1)[2]-(x2)[2]; }
#define VECTOR3_ADD(y,x1,x2)  { (y)[0]=(x1)[0]+(x2)[0]; (y)[1]=(x1)[1]+(x2)[1]; (y)[2]=(x1)[2]+(x2)[2]; }
#define VECTOR3_TIMES_SCALAR(vs,v,s) {(vs)[0]=(s)*(v)[0];(vs)[1]=(s)*(v)[1];(vs)[2]=(s)*(v)[2];}
#define VECTOR3_NORM2(v) ( (v)[0] * (v)[0] + (v)[1] * (v)[1] + (v)[2] * (v)[2] )
#define VECTOR3_NORM(v) sqrt( (v)[0] * (v)[0] + (v)[1] * (v)[1] + (v)[2] * (v)[2] )
#define VECTOR3_DOT_PRODUCT( v1 , v2 ) ( (v1)[0] * (v2)[0] + (v1)[1] * (v2)[1] + (v1)[2] * (v2)[2] )
#define VECTOR3_CROSS_PRODUCT(z,x,y) { (z)[0]=(x)[1]*(y)[2]-(x)[2]*(y)[1] ;\
                                       (z)[1]=(x)[2]*(y)[0]-(x)[0]*(y)[2] ;\
                                       (z)[2]=(x)[0]*(y)[1]-(x)[1]*(y)[0] ; }
#define VECTORN_NORM2(x,v,dim) { int i; x=0; for (i=0;i<dim;++i) x += v[i] * v[i] ; }
#define VECTOR3_EQUAL( x1 , x2 ) { x1[0] = x2[0] ; x1[1] = x2[1] ; x1[2] = x2[2] ; }

#define vecttimesscalar(vs,v,s) {(vs)[0]=(s)*(v)[0];(vs)[1]=(s)*(v)[1];(vs)[2]=(s)*(v)[2];}
#define scalarproduct(v1,v2) ((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2])
#define crossproduct(z,x,y) {(z)[0]=(x)[1]*(y)[2]-(x)[2]*(y)[1];\
                             (z)[1]=(x)[2]*(y)[0]-(x)[0]*(y)[2];\
                             (z)[2]=(x)[0]*(y)[1]-(x)[1]*(y)[0];}
#define frac(x) ((x)-floor(x))
#define round(x) (sgn(x)*floor(fabs((double)(x))+0.5))
//#define sgn(x) ((x)==0 ? 0 : ((x)/abs((double)(x))))
#define sgn(x) ( (x)==0 ? 0 : ( (x)>0 ? 1 : -1 ) )
#define max(x,y) ((x)>(y) ? (x) : (y))
#define min(x,y) ((x)<(y) ? (x) : (y))
#define Heaviside1(x) ((x)>=0 ? 1.0 : 0)
#define Heaviside0(x) ((x)>0 ? 1.0 : 0)
#define vect3fprintf(fp,v) {int i; for (i=0;i<3;++i)\
 fprintf(fp,"%le",(double)(v)[i]);fprintf(fp,"\n");}


#define deltaij(x,y)  ( (x)==(y) ? 1 : 0 ) 
#define epsijk(i,j,k) ( sgn((i)-(j)) * sgn((j)-(k)) * sgn((k)-(i)) )

#define DELTA_IJ(x,y)  ( (x)==(y) ? 1 : 0 ) 
#define EPS_IJK(i,j,k) ( sgn((i)-(j)) * sgn((j)-(k)) * sgn((k)-(i)) )

#define SIGN(x) ( (x < 0) ? -1 : (x > 0) )

#define MAX(x,y) ((x)>(y) ? (x) : (y))
#define MIN(x,y) ((x)<(y) ? (x) : (y))




typedef struct mytype_interp1 {
  double *x ;
  double *y ;
  int dim ;
  char type[32] ;
  gsl_interp_accel *acc ;
  gsl_spline *spline ;
  gsl_interp *interpolation ;
} mytype_interp1 ;



extern int coordtrans_xyz2sphere(double x,double y,double z,
                            double *r,double *phi,double *th);

extern double linapprox(double x1,double x2,double y1,double y2,double x);
extern void inversematrix3(double (*Em)[3],double (*Fm)[3]);
extern void  coordtrafo(double *to,double *from,double *zaxes,double *xaxes,
		 char where);

extern void rotatepos(double (*r)[3] , int N , double phiin , double thetain );

extern int binom(int n,int k);
extern double interpolate(double *x,double *y,int dim,double xx,char method);
extern double draw_poisson(double w);
extern double drawrandomdistr(double (*actdistr)(),double *params,
                              int type,int nrm);


extern double vmean(int dim,char *type,void* vdata); 
extern double vstd(int dim,char *type,void* vdata);

extern int init_interp1( mytype_interp1 *data , double *x , double *y , int dim , char* interp1_type ) ;
extern int free_interp1( mytype_interp1 *data ) ;
extern double do_interp1( mytype_interp1 *data , double xi ) ;
extern int full_interp1( double *x , double *y , int dim , double *xi , double *yi , int dimi , char* interp1_type ) ;

extern int array_max( void *a_m , void *a_x , long long int dim , char *type ) ;

extern int math_solve_sec_order_alg_eq( double a , double b , double c , double *x1 , double *x2 ) ;


