///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdio.h>
#include <string.h>

#include "basics.h"

#include "phys_mat_const.h"

struct g_struct_phys_math_const   g_phys_math_const  ;

//############################## FUNCTION #################################

int init_phys_mat_const( void ) 
{
    double PI = 3.141592653589793238 ;

  g_phys_math_const .e1   = 1.602189e-19 ;
  g_phys_math_const .me   = 9.109534e-31 ;
  g_phys_math_const .mp   = 1.672648e-27 ;
  g_phys_math_const .eps0 = 8.854187e-12 ;
  g_phys_math_const .hbar = 6.626176e-34 / 2.0 / PI ;
  g_phys_math_const .c    = 299792458.0 ;
  g_phys_math_const .N_A  = 6.022136736e23 ; 
  g_phys_math_const .a_B  = 0.529e-10 ;
  g_phys_math_const .k_B  = 1.3806505e-23 ;

  g_phys_math_const .e2 = g_phys_math_const .e1 * g_phys_math_const .e1 ;

  g_phys_math_const .C_k = 1.0 /4.0 / PI  / g_phys_math_const.eps0 ;

  g_phys_math_const .alpha =  g_phys_math_const .e2 / 
      ( ( 4 * PI * g_phys_math_const .eps0 ) * g_phys_math_const .hbar * g_phys_math_const .c ) ;

  g_phys_math_const .Hartree = g_phys_math_const.alpha * g_phys_math_const.alpha * 
                               g_phys_math_const.me * 
                               g_phys_math_const.c * g_phys_math_const.c ; 

  g_phys_math_const .au_time = g_phys_math_const.hbar / g_phys_math_const.Hartree ; 

  g_phys_math_const .e1_C_k = g_phys_math_const.e1 * g_phys_math_const.C_k ;
  g_phys_math_const .e2_C_k = g_phys_math_const.e2 * g_phys_math_const.C_k ;

  return 0 ;
 }

//############################## FUNCTION #################################

double convert_phys_unit_scalar( char *a_command , double a_x )
{

    /* FIX THESE CONSTANTS!!! */
    const double t_au2si = 2.418884326505e-17 , l_au2si = 5.2917720859e-11 ;

    if ( strcmp( a_command , "au2si-rate" ) == 0 )
    { return( a_x / t_au2si ) ; }

    if ( strcmp( a_command , "si2au-rate" ) == 0 )
    { return( a_x * t_au2si ) ; }

    if ( strcmp( a_command , "au2si-len" ) == 0 )
    { return( a_x * l_au2si ) ; }

    if ( strcmp( a_command , "si2au-len" ) == 0 )
    { return( a_x / l_au2si ) ; }

    if ( strcmp( a_command , "au2si-area" ) == 0 )
    { return( a_x * l_au2si * l_au2si ) ; }

    if ( strcmp( a_command , "si2au-area" ) == 0 )
    { return( a_x / l_au2si / l_au2si ) ; }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return( 0 ) ;
}


//############################## FUNCTION #################################

double convert_unit_au2m( double a_x )
{
    //return( a_x * g_phys_math_const.a_B ) ;
    return( a_x * 5.291772192e-11 ) ;
}

//############################## FUNCTION #################################

double convert_unit_m2au( double a_x )
{
    //return( a_x / g_phys_math_const.a_B ) ;
    return( a_x / 5.291772192e-11 ) ;
}

//############################## FUNCTION #################################

double convert_unit_au2invsec ( double a_x )
{
    return( a_x / g_phys_math_const.au_time ) ;
}

//############################## FUNCTION #################################

double convert_unit_invsec2au ( double a_x )
{
    return( a_x * g_phys_math_const.au_time ) ;
}

//############################## FUNCTION #################################

double convert_unit_J2eV ( double a_x )
{
    return( a_x / g_phys_math_const.e1 ) ;
}

//############################## FUNCTION #################################

double convert_unit_eV2J ( double a_x )
{
    return( a_x * g_phys_math_const.e1 ) ;
}

//############################## FUNCTION #################################

double convert_unit_au2J( double a_x )
{
    return( a_x * g_phys_math_const.Hartree ) ;
}

//############################## FUNCTION #################################

double convert_unit_J2au( double a_x )
{
    return( a_x / g_phys_math_const.Hartree ) ;
}

//############################## FUNCTION #################################

double convert_unit_au2eV( double a_x )
{
    return( a_x * g_phys_math_const.Hartree / g_phys_math_const.e1 ) ;
}

//############################## FUNCTION #################################

double convert_unit_eV2au( double a_x )
{
    return( a_x  * g_phys_math_const.e1 / g_phys_math_const.Hartree ) ;
}

//############################## FUNCTION #################################

double convert_unit_au2m2( double a_x )
{
    return( a_x * g_phys_math_const.a_B * g_phys_math_const.a_B ) ;
}


