///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>
#include <time.h>

#include "basics.h"

#ifndef PI
#define PI 3.141592653589793238
#endif



struct rand_params_struct {
  long long int calls;
  int seed;
  long long int *node_calls;
};

static struct rand_params_struct rand_params;



//############################## FUNCTION #################################

long long int get_randcalls(void){
  return(rand_params.calls);
}


//############################## FUNCTION #################################

double randnum(int indicator){
  static int RandSeed=1;
  static int randcalls;

  switch (indicator) {
  case 0 :                       // gives back a 0<random number<1
    ++randcalls;
    ++rand_params.calls;
    return(((double)rand()+1)/((double)(RAND_MAX)+2));
  case -4:                       // gives back the seed
    return((double)RandSeed);
  case -3:                       // sets a random seed value
    RandSeed=time(NULL);
    srand(RandSeed);
    return((double)randcalls);
  case -2 :                      // sets randcalls to 0
    randcalls=0;
    rand_params.calls=0;
    return((double)randcalls);
  case -1 :                      // gives back randcalls
    return((double)randcalls);
  default:                       // sets the randomseed
    //    randcalls=0;
    RandSeed=indicator;
    rand_params.seed=indicator;

    srand( indicator ) ;
    
    return(0);
  }
}


//############################## FUNCTION #################################

long long int* randcalls_counts(char comm , ... ) {
  static int init=1;
  static long long int *allcalls , rnd ;
  long long int i;
  int vint;
  va_list ap;


  if (init) {
    init=0;
    allcalls=calloc(1,sizeof(long long int));
  }

  switch (comm) {
  
    case 'g':
      return(allcalls);
      break;
      
    case 'p':
      printf("p %lld\n",rnd);
      break;

    case 'c':
      rnd=get_randcalls();
      allcalls[0]=rnd;
      break;

    case 's':
      rnd=allcalls[0];
      break;

    case 'i':

      va_start(ap,comm);
      vint = va_arg(ap,int);

      randnum(-2);
      randnum( vint );
      for (i=0;i<rnd;++i) {
        randnum(0);
      }
      break;
   
    case 'f':
      free(allcalls);
      init=1;
      break;
  }

  return 0 ;
}


//############################## FUNCTION #################################

void randvec(double *p){
  double phi,costh,sinth;

  phi=2*randnum(0)*PI;
  costh=2*randnum(0)-1;
  sinth=sqrt(1-costh*costh);
  *p=sinth*cos(phi);
  *(p+1)=sinth*sin(phi);
  *(p+2)=costh;
}



//############################## FUNCTION #################################

void randvecsphere(double *phi, double *sinth,double *costh){

  *phi=2*randnum(0)*PI;
  *costh=2*randnum(0)-1;
  *sinth=sqrt(1-(*costh)*(*costh));
}



//############################## FUNCTION #################################

int rpermute_noalloc( int n , int *a ) 
{
    int j, k, t ;
    for ( k = 0 ; k < n ; k++ ) { a[k] = k ; }
    for ( k = n - 1 ; 0 < k ; k-- ) 
    {
        j = ( (int) floor( randnum( 0 ) * n ) )  % ( k + 1 ) ;
        t    = a[j] ;
        a[j] = a[k] ;
        a[k] = t ;
    }
    return( 0 ) ;
}



