///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>

#include "common.h"

#include "globalconsts.h"

#include "xmdyn_filenames.h"

#include "endltabs.h"
#include "Xatom.h"


#include "tests.h"
#include "filetests.h"

#include "secel.h"

#include "init_system.h"
#include "xmdyn_io.h"

#include "classdyn.h"   /* TODO: we should not need it!!! */



static struct allfilenamesstruct *ALLFILES;
static struct save_prec_struct   *SAVE_PREC ;



//############################## FUNCTION #################################

int init_init_system( void ) 
{
    xmdyn_filenames_load( &ALLFILES , &SAVE_PREC ) ;

  return 0 ;
}



//############################## FUNCTION #################################

int init_module_all( void )
{
    static int init = 1 ;
    if ( 0 == init ) { return 0 ; }

    init = 0 ;
    return 0 ;
}




//############################## FUNCTION #################################

int set_default_init_values(char *comm) {
  int i;
  if (strcmp(comm,"qatom")) {
    for (i=0;i< g_inputs.N;++i)
      g_system.qi[i]= g_inputs.qatom;
    return 0 ;
  }
  if (strcmp(comm,"qel")) {
    for (i= g_inputs.N;i< g_inputs.N+ g_inputs.E;++i)
       g_system.qi [i]=0;
    for (i= g_inputs.N+ g_inputs.E- g_inputs.Eextra;i< g_inputs.N+ g_inputs.E;++i)
       g_system.qi [i]=-1;
    return 0 ;
  }

  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ;
  return -1 ;
}







//############################## FUNCTION #################################

void default_init_params(char command){
  int n,sz,i,j,k,maxplace;
  double distfcc;
  char latticetype[SSTRLTH];
  int *tempZall;
  int N,E,Z;
  //double TAV,SZORAS;
  double DIST,DEVI;

  N= g_inputs.N;E= g_inputs.E;Z= g_inputs.Z;DIST= g_inputs.lattice_const;DEVI= g_inputs.lattice_devi;
  strcpy(latticetype, g_inputs.r1input);

  switch (command) {
  
   case 'R': //  RANDOM-INIT    <OK>
    randnum(-2);//randcalls=0;
    randnum( g_inputs.randseed);  //srand(RandSeed);
    break;
    
   case 't':  //  TIME-INIT    <OK>
    myclock(0);//runtime=0;
    break;
    
   case 'D':  //  ATOMIC-DATA-LOAD    <OK>
    tempZall=calloc( g_inputs.N,sizeof(int));
    for (i=0;i<N;++i){
      tempZall[i]=Z;
    }
    getallatomicdata(tempZall);
    free(tempZall);
    break;
    
   case 'E': //  E-INIT    <OK>
    tempZall=calloc( g_inputs.N,sizeof(int));
    for (i=0;i<N;++i){
      tempZall[i]=Z;
    }
     g_inputs.E=0;
    for (i=0;i< g_inputs.N;++i){
       g_inputs.E+=cumsums_shell_states(tempZall[i]);
    }
    free(tempZall);
    break;
     
   case 'v':  // VARIABLES-INIT    <OK>
    varibles(1);
    break;
    
   case 'Z': // Zall-INIT    <OK>
    for (i=0;i<N;++i){
       g_system.Z [i]=Z;
    }
    break;
    
   case 'P': // PHOTOEFF-DATA-LAOD    <OK>
    break;
    
   case 'c':  // ELECTRON-CONFIG-INIT    <OK>
    electronconfiginit();
    E= g_inputs.E;
    break;

   case 'r': // POSITIONS-INIT ( g_system.r1 , g_system.r2 )
    //E= g_inputs.E;
    if (strcmp(latticetype,"sc")==0){
      if ( (pow((double)N,1.0/3.0)-floor(pow((double)N,1.0/3.0))) != 0)
        n=(int)floor(pow((double)N,1.0/3.0))+1;
      else
        n=(int)floor(pow((double)N,1.0/3.0));
      sz=0;
      for (i=0;i<n;++i){
        for (j=0;j<n;++j){
  	  for (k=0;k<n;++k){
	     g_system.r1 [sz][0]=DIST*(i+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	     g_system.r1 [sz][1]=DIST*(j+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	     g_system.r1 [sz][2]=DIST*(k+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	    if ( g_inputs.centeratom==-1)
	      if ( (i==(int)floor(n/2)) && (j==(int)floor(n/2))
		   && (k==(int)floor(n/2)) )
	         g_inputs.centeratom=sz;

	    sz++;
	    if (sz==N)
	      break;
	  }
	  if (sz==N)
	    break;
        }
        if (sz==N)
	  break;  
      }
    }

    if (strcmp(latticetype,"fcc")==0){
      maxplace=0;n=0;
      while (maxplace<N){
        n++;
        maxplace = (int) (n%2==0 ? pow((double)n,3.0)/2.0 : (pow((double)n,3.0)+1)/2.0);
      }

      distfcc=DIST/sqrt(2.0);

      sz=0;
      for (i=0;i<n;++i){
        for (j=0;j<n;++j){
	  for (k=0;k<n;++k){
	    if ((i+j+k)%2==0){
	       g_system.r1 [sz][0]=distfcc*(i+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	       g_system.r1 [sz][1]=distfcc*(j+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	       g_system.r1 [sz][2]=distfcc*(k+DEVI*(randnum(0)-1.0/2.0)-(n-1)/2.0);
	      if ( g_inputs.centeratom==-1)
	        if ( (i==(int)floor(n/2)) && (j==(int)floor(n/2))
		     && (k==(int)floor(n/2)) )
		   g_inputs.centeratom=sz;

	      sz++;
	    }
	    if (sz==N)
	      break;
	  }
	  if (sz==N)
	    break;
        }
        if (sz==N)
	  break;
      }
    }
    if (strcmp( g_inputs.anizotropy,"y")==0)
      rotatepos(  g_system.r1  ,  g_inputs.N +  g_inputs.E ,  g_inputs.phiin ,  g_inputs.thetain );
    for (i=0;i<N+E;++i)
    {
      vect3equal( g_system.r2 [i], g_system.r1 [i]);    
      vect3equal( g_system.prev_r2 [i], g_system.r2 [i]);    
    }
    break;    
  
  }
}



//############################## FUNCTION #################################

int el_config_subshell_init(int elstart,int maxenum,int NE,char typ){
  int remainE,i,j;
  int relpos;
  double x ;
  remainE=NE;
  while (remainE>0) {
    // x = randnum(0) ; // orbital is filled up randomly
    x = 0 ;             // orbital is filled up sequentially, it is better
    relpos=floor(x * (maxenum-(NE-remainE)) );
    i=j=0;
    while (i<=relpos){
      while ( g_system.ptype [elstart+j]==typ) {
        j++;
      }
      if (i<relpos)
        j++;
      i++;
    }
    if (j>=maxenum)
      return(-1);

     g_system.ptype [elstart+j]=typ;
    remainE--;
  }
  for (i=elstart;i<elstart+maxenum;++i){
    if (!isitshellelectron(i))
       g_system.ptype [i]='-';
  }
  return(0);
}


//############################## FUNCTION #################################

int electronconfiginit(void){
  int N,E,actZ,i,j,ecounter;
  struct SHELL_STRUCT *actshell;

  N= g_inputs.N; E= g_inputs.E;

  set_default_init_values("qatom");
  ecounter=N;
  for (i=0;i<N;++i){
    actZ= g_system.Z [i];

    xmdyn_get_econfig_properties( "Z" , g_system.Z[i] , 2 , "mass" , g_system.m + i ) ;
     g_system.ptype [i]='n';

    g_atomic_data[i].Z= g_system.Z +i;

    g_atomic_data[i].maxshellnum=ATOMDATA[actZ].maxshellnum;
    g_atomic_data[i].shell=calloc(g_atomic_data[i].maxshellnum+1,sizeof(struct SHELL_STRUCT));
    g_atomic_data[i].e_from=ecounter;
    for (j=0;j<=g_atomic_data[i].maxshellnum;++j){
      if ( subshell_maxenum(j) == 0 )
      {
        actshell=g_atomic_data[i].shell+j;
        actshell->e_from=-1;
        actshell->e_to=-1;
	actshell->NE=0;
      }
      else{
        actshell=g_atomic_data[i].shell+j;
        actshell->e_from=ecounter;
	actshell->NE=setsubshellNE(actZ,j);
	el_config_subshell_init(ecounter,subshell_maxenum(j),
	                        actshell->NE,subshellint2char(j));
	ecounter+=subshell_maxenum(j);
        actshell->e_to=ecounter-1;
      }
    }
    g_atomic_data[i].e_to=ecounter-1;
  }

  set_default_init_values("qel");
  for (i=N;i<N+ g_inputs.E;++i){
    g_system.m[i]= g_phys_math_const.me ;
  }
  for (i=N+ g_inputs.E- g_inputs.Eextra;i<N+ g_inputs.E;++i){
     g_system.ptype [i]='g';
  }


  return 0 ;
}






//############################## FUNCTION #################################

int startingdataindex(void){
  FILE *fp;
  int indstar,ind1,ind2,foto1exists,foto2exists;
  char strin[STRLTH],c;

  indstar=ind1=ind2=0;
  
  if ( g_inputs.startindex!=0){
    return( g_inputs.startindex); // startfiles' index from commandline
  }

  if (!(fp=fopen(ALLFILES -> data2diskindicator,"r"))){
    fotofilenames(-99);
    return(-99); // indicatorfile does not exist
  }
  else {
    fseek(fp,0,SEEK_SET);
    while ( (fscanf(fp,"%c",&c)==1) && indstar<2)
      if (c=='*')
        indstar++;
    	
    fseek(fp,0,SEEK_SET);
    while ( (fscanf(fp,"%c",strin)==1) && (fscanf(fp,"%c",strin+1)==1) && ind1<2){
      strin[2]='\0';
      if (strcmp(strin,"-1")==0)
        ind1++;
      fseek(fp,-1,SEEK_CUR);
    }	
    fseek(fp,0,SEEK_SET);
    while ( (fscanf(fp,"%c",strin)==1) && (fscanf(fp,"%c",strin+1)==1) && ind2<2){
      strin[2]='\0';
      if (strcmp(strin,"-2")==0)
        ind2++;
      fseek(fp,-1,SEEK_CUR);	
    }	
    fclose(fp);
    
    foto1exists=foto2exists=0;
    fotofilenames(-1);
    if ((fp=fopen(ALLFILES -> r1foto,"r"))){
      fclose(fp);
      foto1exists=1;
    } 
    fotofilenames(-2);
    if ((fp=fopen(ALLFILES -> r2foto,"r"))){
      fclose(fp);
      foto2exists=1;
    } 
    
    if (indstar>0){
      if (ind1==2){
        return(-1);
      }	
      else if (foto2exists){
        return(-2);
      }
    }
    return(-12); //problem with interrupted calc: wrong data2diskindicator file
  }
}



//############################## FUNCTION #################################

int startfromfiles_start_cont(int startind,int *stepstart)
  {
  int i,j,randcalls,d , ii ;
  FILE *fp,*fpZ,*fpf;
  double runtime;
  int N,E,counter,*tempZall,actatom;
  char indstr[LSTRLTH]="",indstr2[STRLTH]="", tempch;
  long long int randcalls_long,*allrandcalls;
  int ir_nn;
  double ir_t,ir_T,*par;

  int Num_Rank = 1 ;

  strcat(strcat(strcat(indstr,"@ i: "),indstr2)," ");
 
  fotofilenames(startind); 
  g_system.Z = NULL ;

  
  if ( (fpf=fopen(ALLFILES -> stepfoto,"r")) ) {
    fscanf(fpf,"%d",stepstart);
    (*stepstart)++;
    fclose(fpf);
    strcat(indstr,"stf ");
  }
  else{
    *stepstart=0;
    strcat(indstr,"st0 ");
  }
  sprintf(indstr2,"%d",startind);
  strcat(strcat(strcat(indstr,"StartIndex:"),indstr2)," ");
  sprintf(indstr2,"%d",*stepstart);
  strcat(strcat(strcat(indstr,"Step:"),indstr2)," ");
  clearstr(indstr2);
 
  if (*stepstart== g_inputs.nstep) {
    printf("Calculation is already finished.\n");
    return(-1);
  }  
  else if (*stepstart> g_inputs.nstep) {
    printf("First step is greater than input last step?!\n");
    return(-1);
  }  
 
  allrandcalls=randcalls_counts('g');
  if ( (fpf=fopen(ALLFILES -> randcallsfoto,"r")) ) {
    fscanf(fpf,"%d",&i);
    if (i!=Num_Rank) {
      printf("Specified foto_rand (startindex %d) is for %d machines!\n",
             startind,i);  //TODO ERROR
      printf("Remove or modify it!\n");
      return(-1);
    }
    for (i=0;i<Num_Rank;++i) {
      fscanf(fpf,"%lld",allrandcalls+i);
    }
    fclose(fpf);
    strcat(indstr,"raf ");
  }
  else{
    randcalls=0;
    randcalls_long=0;
    for (i=0;i<Num_Rank;++i) {
      allrandcalls[i]=0;
    }
    strcat(indstr,"ra0 ");
  }
  randnum(-2); srand( g_inputs.randseed);

  if ( (fpf=fopen(ALLFILES -> timefoto,"r")) ) {
    fscanf(fpf,"%le",&runtime);
    fclose(fpf);
    strcat(indstr,"tif ");
  }
  else{
    runtime=0;
    strcat(indstr,"ti0 ");
  }
  myclock(runtime);
  strcat(indstr,"\n@ i: ");


  // LOADING/SETTING Z: Zall
  counter=0;
  if (!(fp=fopen(ALLFILES -> Zallfoto,"r"))){
    tempZall=calloc( g_inputs.N,sizeof(int));
    for (i=0;i< g_inputs.N;++i){
      tempZall[i]= g_inputs.Z;
    }
    strcat(indstr,"Zinput ");
  }
  else{
    while (fscanf(fp,"%*s") != EOF){
      counter++;
    }
    fclose(fp);
     g_inputs.N=counter;

    tempZall=calloc( g_inputs.N,sizeof(int));
    fp=fopen(ALLFILES -> Zallfoto,"r");
    for (i=0;i< g_inputs.N;++i){
      fscanf(fp,"%d",tempZall+i);
    }
    fclose(fp);
    strcat(indstr,"Zfile ");
  }

  // Getting atomic data
  xmdyn_load_grstate_neutral_econfigs( g_inputs.N , tempZall ) ;
  getallatomicdata(tempZall);


   g_inputs.E=0;
  for (i=0;i< g_inputs.N;++i){
     g_inputs.E+=cumsums_shell_states(tempZall[i]);
  }
  free(tempZall);
   g_inputs.E+= g_inputs.Eextra;

  N= g_inputs.N;E= g_inputs.E;

  varibles(1);
  //
  for ( ii = 0; ii <  g_inputs.N +  g_inputs.E ; ii++ ) 
      g_system .motherion[ii] = -1 ;


  if (!(fpZ=fopen(ALLFILES -> Zallfoto,"r"))){
    for (i=0;i< g_inputs.N;++i){
       g_system.Z [i]= g_inputs.Z;
    }
  }
  else {
    for (i=0;i< g_inputs.N;++i){
      fscanf(fpZ,"%d", g_system.Z +i);
    }
    fclose(fpZ);
  }
  instant_recombination_init() ;

  electronconfiginit();
  for (i=0;i<N;++i) 
  {
//    el_config_update(i);
  }


  if (strstr(  g_inputs.no_realspace_dyn,"fixZ")) {
    d=dim_input_pars(  g_inputs.no_realspace_dyn);
    par=load_input_pars(  g_inputs.no_realspace_dyn,d);
    for (i=0;i<d;++i) {
      for (j=0;j< g_inputs.N;++j) {
        if ( g_system.Z [j]==(int)par[i]) {
	   g_system.ptype [j]=  g_inputs.fixts;
	}
      }
    }
    par=load_input_pars(  g_inputs.no_realspace_dyn,-1);
  }


  if ( (fp=fopen(ALLFILES -> r1fotob,"r")) ) {
    strcat(indstr,"r1b ");
    i=fread( g_system.r1 ,sizeof(double),3*(N+E),fp);
    if (i==N) {
      strcat(indstr,"a ");
    }
    else if (i!=3*(N+E)){
      printf("Problem with r1 binary file!\n");
      strcat(indstr,"r1bP ");
    }
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> r1foto,"r")) ) {
    strcat(indstr,"r1f ");
    counter=0;
    while (fscanf(fpf,"%le",*( g_system.r1 +counter)+0) != EOF){
      fscanf(fpf,"%le",*( g_system.r1 +counter)+1);
      fscanf(fpf,"%le",*( g_system.r1 +counter)+2);
      counter++;
    }
    if (counter==N) {
      strcat(indstr,"a ");
    }
    else if (counter!=N+E){
      printf("Problem with r1 text file!\n");
      strcat(indstr,"r1fP ");
    }
    fclose(fpf);
  }
  else{
    default_init_params('r');
    strcat(indstr,"r1d ");
  }

  if ( (fp=fopen(ALLFILES -> r2fotob,"r")) ) {
    strcat(indstr,"r2b ");
    if (fread( g_system.r2 ,sizeof(double),3*(N+E),fp) != (unsigned)(3*(N+E))){
      printf("Problem with r2 binary file!\n");
      strcat(indstr,"r2bP ");
    }
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> r2foto,"r")) ) {
    strcat(indstr,"r2f ");
    counter=0;
    while (fscanf(fpf,"%le",*( g_system.r2 +counter)+0) != EOF){
      fscanf(fpf,"%le",*( g_system.r2 +counter)+1);
      fscanf(fpf,"%le",*( g_system.r2 +counter)+2);
      counter++;
    }
    if (counter!=N+E){
      printf("Problem with r2 text file!\n");
      strcat(indstr,"r2fP ");
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"r2d ");
    for (i=0;i<N;++i)
      for (j=0;j<3;++j)
         g_system.r2 [i][j]= g_system.r1 [i][j];
  }


  if ( (fp=fopen(ALLFILES -> prev_r2fotob,"r")) ) {
    strcat(indstr,"prev_r2b ");
    if (fread( g_system.prev_r2 ,sizeof(double),3*(N+E),fp)!= (unsigned)(3*(N+E)) ) {
      printf("Problem with r2 binary file!\n");
      strcat(indstr,"prev_r2bP ");
    }
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> prev_r2foto,"r")) ) {
    strcat(indstr,"prev_r2f ");
    counter=0;
    while (fscanf(fpf,"%le",*( g_system.prev_r2 +counter)+0) != EOF){
      fscanf(fpf,"%le",*( g_system.prev_r2 +counter)+1);
      fscanf(fpf,"%le",*( g_system.prev_r2 +counter)+2);
      counter++;
    }
    if (counter!=N+E){
      printf("Problem with r2 text file!\n");
      strcat(indstr,"prev_r2fP ");
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"prev_r2d ");
    for (i=0;i<N;++i)
      for (j=0;j<3;++j)
         g_system.prev_r2 [i][j]= g_system.r2 [i][j];
  }



  if ( (fp=fopen(ALLFILES -> v1fotob,"r")) ) {
    strcat(indstr,"v1b ");
    if (fread( g_system.v1 ,sizeof(double),3*(N+E),fp)!= (unsigned)(3*(N+E)) ) {
      printf("Problem with v1 binary file!\n");
      strcat(indstr,"v1bP ");
    }
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> v1foto,"r")) ) {
    strcat(indstr,"v1f ");
    counter=0;
    while (fscanf(fpf,"%le",*( g_system.v1 +counter)+0) != EOF){
      fscanf(fpf,"%le",*( g_system.v1 +counter)+1);
      fscanf(fpf,"%le",*( g_system.v1 +counter)+2);
      counter++;
    }
    if (counter==N) {
      strcat(indstr,"a ");
    }
    else if (counter!=N+E){
      printf("Problem with v1 text file!\n");
      strcat(indstr,"v1fP ");
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"v1d ");
  }

  if (( fp=fopen(ALLFILES -> v2fotob,"r")) ) {
    strcat(indstr,"v2b ");
    if (fread( g_system.v2 ,sizeof(double),3*(N+E),fp) != (unsigned)(3*(N+E)) ) {
      printf("Problem with v2 binary file!\n");
      strcat(indstr,"v2bP ");
    }
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> v2foto,"r")) ) {
    strcat(indstr,"v2f ");
    counter=0;
    while (fscanf(fpf,"%le",*( g_system.v2 +counter)+0) != EOF){
      fscanf(fpf,"%le",*( g_system.v2 +counter)+1);
      fscanf(fpf,"%le",*( g_system.v2 +counter)+2);
      counter++;
    }
    if (counter!=N+E){
      printf("Problem with v2 text file!\n");
      strcat(indstr,"v2fP ");
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"v2d ");
    for (i=0;i<N;++i)
      for (j=0;j<3;++j)
         g_system.v2 [i][j]= g_system.v1 [i][j];
  }


  if (( fp=fopen(ALLFILES -> massfotob,"r")) ) {
    strcat(indstr,"mb ");
    if (fread( g_system.m,sizeof(double),(N+E),fp) != (unsigned)(N+E) ){
      printf("Problem with m file!\n");
      strcat(indstr,"mbP ");
    }
    fclose(fp);
  }
  else if ( ( fpf=fopen(ALLFILES -> massfoto,"r")) ) {
    strcat(indstr,"mf ");
    counter=0;
    while (fscanf(fpf,"%le",( g_system.m +counter)) != EOF){
      counter++;
    }
    if (counter!=N+E){
      printf("Problem with m text file!\n");
      strcat(indstr,"mfP ");
    }
    fclose(fpf);
  }


  if ( ( fp=fopen(ALLFILES -> energyfotob,"r")) ) {
    strcat(indstr,"enb ");
    if ( (fread(& g_ENERGY_VALUES.Eintotal,sizeof(double),1,fp)!=1) ||
         (fread(& g_ENERGY_VALUES.Eintotal_th,sizeof(double),1,fp)!=1) ||
         (fread(& g_ENERGY_VALUES.Einph,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Einfl,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Einau,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Einsi,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Einrec,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Einitial,sizeof(double),1,fp)!=1) ||
	 (fread(& g_ENERGY_VALUES.Elost,sizeof(double),1,fp)!=1) 
	 || (fread(& g_ENERGY_VALUES.Einfi,sizeof(double),1,fp)!=1) 
       )
    {
      printf("Problem with energy binary file!\n");
      strcat(indstr,"enbP ");
    }
    if (fread(& g_ENERGY_VALUES.Einchmig,sizeof(double),1,fp)!=1) {;}
    if (fread(& g_ENERGY_VALUES.Ebound,sizeof(double),1,fp)!=1) {;}
    if (fread(& g_ENERGY_VALUES.Efl_photon,sizeof(double),1,fp)!=1) {;}
    if (fread(& g_ENERGY_VALUES.Work,sizeof(double),1,fp)!=1) {;}
    fclose(fp);
  }
  else if ( (fpf=fopen(ALLFILES -> energyfoto,"r")) ) {
    strcat(indstr,"enf ");
    counter=0;
    fscanf(fpf,"%le %le %le %le %le %le %le %le %le %le %le %le %le", & g_ENERGY_VALUES.Eintotal,& g_ENERGY_VALUES.Eintotal_th,& g_ENERGY_VALUES.Einph,& g_ENERGY_VALUES.Einfl,& g_ENERGY_VALUES.Einau,& g_ENERGY_VALUES.Einsi,& g_ENERGY_VALUES.Einrec,& g_ENERGY_VALUES.Einitial,& g_ENERGY_VALUES.Elost,& g_ENERGY_VALUES.Einfi , &g_ENERGY_VALUES.Einchmig , &g_ENERGY_VALUES.Ebound , &g_ENERGY_VALUES.Efl_photon );
    fclose(fpf);
  }
  else{
    strcat(indstr,"end ");
     g_ENERGY_VALUES.Eintotal=0;
     g_ENERGY_VALUES.Eintotal_th=0;
     g_ENERGY_VALUES.Einph=0;
     g_ENERGY_VALUES.Einfl=0;
     g_ENERGY_VALUES.Einau=0;
     g_ENERGY_VALUES.Einsi=0;
     g_ENERGY_VALUES.Einrec=0;
     g_ENERGY_VALUES.Einitial=0;
     g_ENERGY_VALUES.Elost=0;
     g_ENERGY_VALUES.Einfi=0;
     g_ENERGY_VALUES.Einchmig=0;
     g_ENERGY_VALUES.Ebound=0;
     g_ENERGY_VALUES.Efl_photon=0;
     g_ENERGY_VALUES.KE = -1 ;
  }


  if ( (fpf=fopen(ALLFILES -> chargefoto,"r")) ) {
    strcat(indstr,"chf ");
    counter=0;
    while ( fscanf(fpf,"%d", g_system.qi +counter) != EOF){
      counter++;
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"chd ");
    set_default_init_values("qatom");
    set_default_init_values("qel");
  }

  if ( ( fpf=fopen(ALLFILES -> motherionfoto,"r")) ) {
    strcat(indstr,"mif ");
    counter=0;
    while ( fscanf(fpf,"%d", (g_system .motherion) + counter) != EOF){
      counter++;
    }
    fclose(fpf);
    if ( counter !=  g_inputs.N +  g_inputs.E ) {
        strcat(indstr,"ERROR:mif!!!  ");
    }
  }
  else{
    strcat(indstr,"mid ");
  }

  if ( (fpf=fopen(ALLFILES -> etypefoto,"r")) ) {
    strcat(indstr,"etf ");
    counter=0;
    while ( fscanf(fpf,"%1s", g_system.ptype +counter) != EOF){
      counter++;
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"etd ");
  }

  if ( (fpf=fopen(ALLFILES -> el_configfoto,"r")) ) {
    strcat(indstr,"ecf ");
    actatom = 0 ;
    while ( fscanf(fpf,"%*s") != EOF){
      fscanf(fpf,"%*s %*s %*s %*s");
      for (j=0;j<=g_atomic_data[actatom].maxshellnum;++j){
        fscanf(fpf,"%*s %d",&(g_atomic_data[actatom].shell[j].NE));
      }
      actatom++ ;
    }
    fclose(fpf);
  }
  else{
    strcat(indstr,"ecd ");
  }


  
  if ( (fpf=fopen(ALLFILES -> instrecfotob,"r")) ) {
      module__recombination_loadsave( "load" , &fpf ) ;
    fclose(fpf);
  }

  
  if ( fpf = fopen( ALLFILES -> chargetransfer_b , "r" ) ) 
  {
      fclose( fpf ) ;
  }


  if ( (fpf=fopen(ALLFILES -> uidfoto,"r")) ) {
    strcat(indstr,"uidf ");
    counter=0;
    while ( fscanf(fpf,"%lld", g_system.uid +counter) != EOF){
      counter++;
    }
    fclose(fpf);
  }

  

 
  xmdyn_update_atoms( NULL , "all,econfig" ) ; // xmdyn_update_atoms( NULL , -2 ) ; 

  xmdyn_update_atoms( NULL , "all,bond" ) ; // xmdyn_update_atoms( NULL , -3 ) ; 


  print_extrainfo_out(indstr2,indstr2,'v',&counter,NULL);
  sprintf(indstr2,"%d",*stepstart);

   
#ifdef XMDYN_XRAYPAC_MORELOG
  fp = fopen( ALLFILES -> progress_file , "a" ) ; 
  fprintf( fp , "%s\n" , indstr ) ;
  fclose( fp ) ;
#endif // XMDYN_XRAYPAC_MORELOG
  return(0);
}




//############################## FUNCTION #################################

int check_snp_status( char *a_dir )
{
    FILE *fp ;
    char fname[LSTRLTH] = "" ;

    if ( 'd' != get_filesystem_object_type( a_dir ) ) { return 0 ; } 

    strcat( strcat( fname , a_dir ) , "/save_status.log" ) ;
    fp = fopen( fname , "r" ) ;
    if ( NULL == fp ) { return -1 ; }
    printf( "%s\n" , fname ) ;
    if ( ( fscanf( fp , "%s" , fname ) ) &&
         ( 0 == strcmp( fname , "saving_finished" ) )
       )
    {
        return 1 ;
    }

    fclose( fp ) ;

    return -1 ;
}




//############################## FUNCTION #################################

enum g_option_string  detect_sample()
{
    int legacy_id ;
    enum g_option_string ret ;
    int status_snp1 , status_snp2 ;

    // old data format
    if ( NULL != strstr( g_inputs.format_loadsave , "000" ) )
    {
       legacy_id = startingdataindex();
       switch  ( legacy_id ) 
       {
           case -1 : return( SNP1_SAMPLE ) ; 
           case -2 : return( SNP2_SAMPLE ) ; 
           case -99 : return( DEF_SAMPLE ) ; 
       }
    }

    // new data format
    else 
    {
        fotofilenames( -1 ) ;
        status_snp1 = check_snp_status( ALLFILES -> fotodir ) ; 
        if ( 1 == check_snp_status( ALLFILES -> fotodir ) ) { return SNP1_SAMPLE ; } 

        fotofilenames( -2 ) ;
        status_snp2 = check_snp_status( ALLFILES -> fotodir ) ; 
        if ( 1 == check_snp_status( ALLFILES -> fotodir ) ) { return SNP2_SAMPLE ; } 

        if ( ( 0 != status_snp1 ) || ( 0 != status_snp2 ) )
        {
            fprintf( stderr , "Problem with both snp_full snapshots!\n" ) ;
            return ERROR ;
        }

        fotofilenames( -99 ) ;
        if ( 'd' == get_filesystem_object_type( ALLFILES -> fotodir ) )
        {
            if ( ( 0 != strcmp( g_inputs.sample_dir , "-" ) ) && 
                 ( 0 != strcmp( g_inputs.sample_dir , "local" ) ) )
            {
                fprintf( stderr , "You cannot have both samples!\n" ) ;
                exit(-1 ) ;
                return ERROR ;
            }
            else { strcpy( g_inputs.sample_dir , "local" ) ; return DEF_SAMPLE ; }
        }
        fotofilenames( -98 ) ;
        return EXT_SAMPLE ;
    }

    return NONE ;
}





//############################## FUNCTION #################################

void varibles(int stage){


  switch (stage) {
   case 1:
    g_atomic_data=calloc( g_inputs.N,sizeof(struct ATOMS_STRUCT));
     g_system.Z =calloc(( g_inputs.N+ g_inputs.E),sizeof(int)); if (! g_system.Z ) nomem();
     g_system.r1 =calloc(( g_inputs.N+ g_inputs.E)*3,sizeof(double)); if (! g_system.r1 ) nomem();
     g_system.r2 =calloc(( g_inputs.N+ g_inputs.E)*3,sizeof(double)); if (! g_system.r2 ) nomem();
     g_system.prev_r2 =calloc(( g_inputs.N+ g_inputs.E)*3,sizeof(double)); if (! g_system.prev_r2 ) nomem();
     g_system.v1 =calloc(( g_inputs.N+ g_inputs.E)*3,sizeof(double)); if (! g_system.v1 ) nomem();
     g_system.v2 =calloc(( g_inputs.N+ g_inputs.E)*3,sizeof(double)); if (! g_system.v2 ) nomem();
     g_system.m =calloc(( g_inputs.N+ g_inputs.E),sizeof(double)); if (! g_system.m ) nomem();
     g_system.qd =calloc(( g_inputs.N+ g_inputs.E),sizeof(double)); if (! g_system.qd ) nomem();
     g_system.qi =calloc(( g_inputs.N+ g_inputs.E),sizeof(int)); if (! g_system.qi ) nomem();
     g_system.qi_corr =calloc(( g_inputs.N+ g_inputs.E),sizeof(int)); if (! g_system.qi_corr ) nomem();
     g_system.ptype =calloc(( g_inputs.N+ g_inputs.E)+1,sizeof(char));
     g_system.uid = XMDYN_ALLOCATE_MEM( (g_inputs.N+g_inputs.E) * sizeof( long long int ) ) ;
     long long int lii ;
     for ( lii = 0          ; lii < g_inputs.N              ; lii++ ) { g_system.uid[lii] = lii ; }
     for ( lii = g_inputs.N ; lii < g_inputs.N + g_inputs.E ; lii++ ) { g_system.uid[lii] =  -1 ; }


    g_system .motherion = calloc( ( g_inputs.N+ g_inputs.E) , sizeof(int) ); 
    if ( ! g_system.motherion ) nomem();

    if (! g_system.ptype )
      nomem();
    else
       g_system.ptype [ g_inputs.N+ g_inputs.E]='\n';
    break;

   case 2:
     g_ENERGY_VALUES.Eintotal=0;
     g_ENERGY_VALUES.Eintotal_th=0;
     g_ENERGY_VALUES.Einph=0;
     g_ENERGY_VALUES.Einfl=0;
     g_ENERGY_VALUES.Einau=0;
     g_ENERGY_VALUES.Einsi=0;
     g_ENERGY_VALUES.Einrec=0;
     g_ENERGY_VALUES.Einitial=0;
     g_ENERGY_VALUES.totalpotE=0;
     g_ENERGY_VALUES.Elost=0;
     g_ENERGY_VALUES.Einfi=0;
     g_ENERGY_VALUES.Einchmig=0;
     g_ENERGY_VALUES.Ebound=0;
     g_ENERGY_VALUES.Efl_photon=0;
     g_ENERGY_VALUES.KE = -1 ;
    break;
 }


}


//############################## FUNCTION #################################

void freememory(void){
  int Z;

  Z= g_inputs.Z;

  free( g_system.r1 ),free( g_system.r2 ),free( g_system.prev_r2 ),free( g_system.v1 ),free( g_system.v2 ),free( g_system.m ),free( g_system.qi ); free(g_system.qi_corr) ;
  free( g_system.ptype );
}




//############################## FUNCTION #################################

int load_sample(int *stepstart)
{
    enum g_option_string sample_type ;

    sample_type = detect_sample() ;

    if ( DEF_SAMPLE == sample_type )
    {
        if (startfromfiles_start_cont(-99,stepstart)==-1) { return(-1); }
    }

    if ( EXT_SAMPLE == sample_type )
    {
        if (startfromfiles_start_cont(-98,stepstart)==-1) { return(-1); }
    }

    if ( SNP1_SAMPLE == sample_type )
    {
        if (startfromfiles_start_cont(-1,stepstart)==-1) { return(-1); }
    }

    if ( SNP2_SAMPLE == sample_type )
    {
        if (startfromfiles_start_cont(-2,stepstart)==-1) { return(-1); }
    }

    return 0 ;

}



