///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>
#include <libgen.h>
#include <float.h>

#include "common.h"

#include "globalconsts.h"

#include "endltabs.h"

#include "input_cmdline_help.h"

#include "tests.h"

#include "xmdyn_filenames.h"

#define XMDYN_XRAYPAC_EXT


static struct allfilenamesstruct *ALLFILES;
static struct save_prec_struct   *SAVE_PREC ;



//############################## FUNCTION #################################

int init_input_cmdline_help( void ) 
{
    xmdyn_filenames_load( &ALLFILES , &SAVE_PREC ) ;

  return 0 ;
}



//############################## FUNCTION #################################

int input_defaults(void){

    char temp_str[LSTRLTH] ;

  strcpy( g_inputs.input_version , "-" ) ;


// ###############################
    // RUN ID
// ###############################
   strcpy( g_inputs.runnumber,"XMDYN");


// ###############################
   // SYSTEM
// ###############################
   g_inputs.N = 0 ; // 50;
   g_inputs.Z = 0 ; // 6;
   g_inputs.Eextra = 0 ;
   g_inputs.lattice_const = 2.4e-10 ;
   g_inputs.lattice_devi = 0.0 ;
   strcpy( g_inputs.r1input , "sc" ) ;
   g_inputs.phiin=0;
   g_inputs.thetain=0;


// ###############################
   // ###   PULSE   ###
// ###############################
   g_inputs.Ephoton = 0 ;  g_inputs.Ephoton*= g_phys_math_const.e1 ; // 12e3 ; 
   g_inputs.Nphoton = 0 ; // 5e12;
   g_inputs.diameter = sqrt( 4 * 1e-6*1e-6 / PI ) ; // 1um2 focus  // 1000e-10 ;
   g_inputs.T_1 = -1 ; //10e-15 ;
   g_inputs.T_2 = g_inputs.T_1 ;
   g_inputs.T0 = -1 ; //15e-15;
   g_inputs.xrsd_delay = 0 ;
   g_inputs.xrsd_ratio_2toT = 0 ; // 0.5 ;
   strcpy( g_inputs.beamtype,"gaussian");



// ###############################
   // REAL SPACE SOLVER
// ###############################
   g_inputs.DT = 1e-18 ;
   g_inputs.nstep = 0 ; //30000 ;
   strcpy( g_inputs.int_method,"verlet");


   g_inputs.tsfactor=1;

   g_inputs.classdyn_partsel = COULOMB_ATOMS ;

   strcpy( g_inputs.coulomb_calc_method , "direct" ) ;


// ###############################
 // SAVINGS / SNAPSHOTS
// ###############################
   strcpy( g_inputs.snp_input , "" ) ;
   strcpy( g_inputs.snp_full_input , "" ) ;
   g_inputs.chkp_dt = 3600.0 ;   // saves a checkpoint in every hour
   strcpy( g_inputs.snp_E_input , "0:1:" ) ; 
   g_inputs.foto = 0 ; 
   g_inputs.progress_dump = 50 ;
   g_inputs.fotomomentsdim=10;
   g_inputs.fotofromfile='n';
   g_inputs.savev3='n';
   g_inputs.onelastsnp = 0 ;



// ###############################
   // ###   MD INTERACTIONS
// ###############################
   g_inputs.potencrb = 0.5e-10 ; 

// ###############################
   // ###   ACTIVE MODULES   ###
// ###############################
  strcpy( g_inputs.timestepOK,"y");
  strcpy( g_inputs.photoionOK,"y");
  strcpy( g_inputs.relaxflOK,"y");
  strcpy( g_inputs.relaxauOK,"y");
  strcpy( g_inputs.scatteringOK,"y");
  strcpy( g_inputs.elscattOK,"n");
  strcpy( g_inputs.elimpionOK,"y");
  strcpy( g_inputs.anizotropy,"y");
  strcpy( g_inputs.anizotropy,"n");
  strcpy( g_inputs.instrecOK,"y"); 


// ###############################
  // ###   XATOM / TABLES   ###
// ###############################
  
   strcpy( g_inputs.atomic_data_source_string , "on_the_fly" ) ;
   strcpy( g_inputs.atomic_data_eval_string , "tree" ) ;
   strcpy( g_inputs.xatom_onthefly_savedir_string , "local" ) ;
   g_inputs.atomic_data_source = XATOM_ONTHEFLY ;
   g_inputs.atomic_data_eval   = ATOMDATA_USE_TREE ;

   g_inputs.xatom_onthefly_savedir = DATADIR ;

   g_inputs.dbase_xatom_version = 2 ;

   g_inputs.xatom_silent = NO ;
   g_inputs.xatom_load_silent = NO ;
   g_inputs.xatom_calc_silent = NO ;

   strcpy( g_inputs.xatom_version , "" ) ;
   strcpy( g_inputs.dbase_types , "atm;pcs" ) ;

// ###############################
 // ###   ATOMIC / PROCESS  PROPERTIES   ###
// ###############################
   //g_inputs.quantumproc_coupled = COUPLED ;
   g_inputs.quantumproc_coupled = DECOUPLED ;
  strcpy( g_inputs.bindE,"linear");
  strcpy(  g_inputs.xatom_energies , "Koopmans" ) ;

   g_inputs.one_photoion = NO ;

// ###############################
  // ###   DEFAULT CHARGES   ###
// ###############################
   g_inputs.qphoto = -1 ; 
   g_inputs.qauger=-1;
   g_inputs.qsecphoto=-1;
   g_inputs.qsecauger=-1;
   g_inputs.qsecsec=-1;
   g_inputs.qfion=-1;
   g_inputs.qatom=0;
   g_inputs.dqatomphotoion=1;
   g_inputs.dqatomauger=1;
   g_inputs.dqatomsec=1;
   g_inputs.dqatomfion=1;

// ###############################
   // ###   DEFAULT DIRECTORIES ###
// ###############################
   strcpy( g_inputs.rootdir, "/NOT_USED_ANYMORE");
   strcpy( g_inputs.dbasedir, "~/xmdyn/dbase");

   strcpy( g_inputs.cachedatadir,"-");
   strcpy( g_inputs.sample_dir,"-");

   strcpy( g_inputs.shortcut_comm,"-") ;
   g_inputs.shortcut_ref = -1  ;



// ###############################
   // ###   MD -- EXTRA    ###
// ###############################
   g_inputs.fixatomsOK='n';
   g_inputs.tselectrons='y';

// ###############################
   // ###   EXTRA    ###
// ###############################
   g_inputs.print_buff_store_size = 10000;
   strcpy( g_inputs.kiirasprec,"% 1.5e ");

   strcpy( g_inputs.avaltype,"none");
   g_inputs.centeratom=-1;

// ###############################
   // ###   XMDYN INTERNALS    ###
// ###############################
  strcpy(  g_inputs.misc_opt ,  "-" ) ;
  //strcpy(  g_inputs.misc_opt ,  "MC:decoupled" ) ; 

   g_inputs.rstarte = 0.5e-10 ; 

   // -- impact ionization --
   strcpy(  g_inputs.impion_model , "BEB" ) ;
   g_inputs.maxscattertry=10;
   strcpy( g_inputs.si_type,"1atom");

   strcpy( g_inputs.impioncriterium,"c");
   g_inputs.maximpdist = 100e-10 ; //* NOTE: not in use *//
   g_inputs.inel_dsdW='r';
   strcpy(  g_inputs.no_realspace_dyn,"-");


   g_inputs.si_model = 1 ;
   g_inputs.si_dsig_dome = 'i';
   g_inputs.si_restricted = 0 ;

   g_inputs.impion_scale = 1.0 ;

   g_inputs.deloctest_nn=2;
   g_inputs.deloctest_comm='c';
   g_inputs.deloctest_ninterp=1;



   // -- recombination --
   g_inputs.maxrecdist=100e-10;
   g_inputs.ireccond='c';
   g_inputs.irecDT = 1 ;
   g_inputs.irecperiods = 1.0 ;
   strcpy( g_inputs.irec_method , "0" ) ;
   strcpy( g_inputs.irec_method , "B-1" ) ;
   strcpy( g_inputs.irec_switchon , "0" ) ;
   strcpy( g_inputs.irec_misc , "" ) ;

// ###############################
   // ###   MISC    ###
// ###############################
   g_inputs.event_dest = LOGFILE ;
   g_inputs.printmessage='n';
   g_inputs.daemon=0;
   g_inputs.startindex=0;
   g_inputs.started=0;
   g_inputs.initfoto=0;


   g_inputs.maxcachetime=1;


  strcpy( g_inputs.extraopts,"");
    

  if ( g_inputs.foto==0)  g_inputs.foto= g_inputs.nstep+1;

  strcpy( g_inputs.format_loadsave , "001" ) ;
  strcpy( g_inputs.txtbin_loadsave , "txt" ) ;

// ###############################
// ###############################
// ###############################

    g_inputs.nots='=';
    g_inputs.nosec='0';
    g_inputs.noall='_';
    g_inputs.fixts='!';


// ###############################
// ###############################
// ###############################

    strcpy( g_inputs.xatom_commandline_opt , "-xmdyn" ) ;

 init_input_cmdline_help( )  ;

 return(0);
}



//############################## FUNCTION #################################

void comments(FILE **fpp){
  char ch;
  ch='a';
  while (ch!=EoL)
    fscanf(*fpp,"%c",&ch);
}

//############################## FUNCTION #################################

int ifcomments(FILE **fpp){
  char c1[2];

  fscanf(*fpp,"%1s",&c1[0]);
  if (c1[0]==COMMENTCARACTER)	{
    while (c1[0]!=EoL)
      fscanf(*fpp,"%c",&c1[0]);
  }
  else{
    fseek(*fpp,-(long)sizeof(char),SEEK_CUR);
    return(-1);
  }
  return(1);
}

//############################## FUNCTION #################################

int oldinputfilecompatible(char *str){
  static char (*oldinputs)[STRLTH];
  static int init=1,i;
  static int num;


  return(-1);

  if (init==1){
    num=2;
    oldinputs=calloc(num*STRLTH,sizeof(char));
    strcpy(oldinputs[0],"fddf");
    strcpy(oldinputs[1],"hghs");
    init=0;
  }

  if (strcmp(str,"Theend")){
  }

  for (i=0;i<num;++i){
    if (strcmp(oldinputs[i],str)) {
      return(1);
    }
  }
  return(0);
}


//############################## FUNCTION #################################

void readinstr(char *in, FILE **fpp){
  char tempstr[STRLTH],*line=NULL,*pch,*start;
  //size_t len = 0;
  int len = 0;
  while (ifcomments(fpp)==1)
    ;
  line = calloc( LSTRLTH , sizeof(char) ) ;

  line = fgets(line, LSTRLTH, *fpp);
  line[strlen(line)-1]='\0';
  pch = strchr ( line , COMMENTCARACTER );
  if (pch != NULL)
    *pch='\0';
  else
    pch=line+strlen(line);
  while ( (strlen(line)!=0) && (pch!=NULL) &&
           ( (*(pch-1)==' ') || (*(pch-1)=='\t') ) ) {
    pch--;
    *pch='\0';
  }
  start=line;
  while ( (*start==' ') || (*start=='\t') )
    start++;

  strcpy(in,start);
  if (oldinputfilecompatible(tempstr)==1) {
    readinstr(in,fpp);
  }
  free(line);
}


//############################## FUNCTION #################################

void readinch(char *in, FILE **fpp){
  char tempstr[STRLTH];
  while (ifcomments(fpp)==1)
    ;

  fscanf(*fpp,"%c",in);
  comments(fpp);
	if (oldinputfilecompatible(tempstr)==1) {
		readinch(in,fpp);
	}

}


//############################## FUNCTION #################################

void readinle(double *in, FILE **fpp){
	char tempstr[STRLTH];
  while (ifcomments(fpp)==1)
    ;

  fscanf(*fpp,"%le",in);
  comments(fpp);
	if (oldinputfilecompatible(tempstr)==1) {
		readinle(in,fpp);
	}

}

//############################## FUNCTION #################################

void readind(int *in, FILE **fpp){
	char tempstr[STRLTH];
  while (ifcomments(fpp)==1)
    ;

  fscanf(*fpp,"%d",in);
  comments(fpp);
	if (oldinputfilecompatible(tempstr)==1) {
		readind(in,fpp);
	}

}


//############################## FUNCTION #################################

void readinle_vector(double *in, FILE **fpp , int a_size ) {
	char tempstr[STRLTH];
        int ii ;
  while (ifcomments(fpp)==1)
    ;

  for ( ii = 0 ; ii < a_size ; ii++ ) { fscanf( *fpp , "%le" , in + ii ) ; }
  comments(fpp);
	if (oldinputfilecompatible(tempstr)==1) {
		readinle(in,fpp);
	}

}

// ########################
//      LOAD INPUT
// ########################


//############################## FUNCTION #################################

int inputin(void) {
  FILE *fpin;
  char inputfield[STRLTH];
  static int counter=0;
  char temp_str[LSTRLTH] ;
  
  if (!(fpin=fopen( g_inputs.inputfile,"r"))) {
    printf("%s does not exist.\n", g_inputs.inputfile);
    return(-1);
  }

  while (fscanf(fpin,"%s",inputfield)==1) {
  
  if ( inputfield[0] == COMMENTCARACTER ) {
      comments(&fpin);
      continue;
  }



  if (counter==0) {
    strcpy( g_inputs.progress_target,"tofile");
    strcpy( g_inputs.paramsout,"tofile");
    strcpy( g_inputs.datadir,"def");
  }



  if (counter==0) 
  {
  
    if ( 0 ) { } // PLACEHOLDER for initial 'if'
    else if (strcmp(inputfield,"N")==0) {
      readind(& g_inputs.N,&fpin);
    }
    else if (strcmp(inputfield,"Z")==0) {
      readind(& g_inputs.Z,&fpin);
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"Eextra")==0) {
      readind(& g_inputs.Eextra,&fpin);
    }
#endif // XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"DT")==0) {
      readinle(& g_inputs.DT,&fpin);
    }
    else if (strcmp(inputfield,"STEPS")==0) {
      readind(& g_inputs.nstep,&fpin);
    }
    else if (strcmp(inputfield,"EPH")==0) {
      readinle(& g_inputs.Ephoton,&fpin);                      
    } 
    else if (strcmp(inputfield,"NPH")==0) { 
      readinle(& g_inputs.Nphoton,&fpin);
    }
    else if (strcmp(inputfield,"FLU_MAX")==0) { 
      readinle(& g_inputs.flu_max,&fpin);
    }
    else if (strcmp(inputfield,"DIAM")==0) {
      readinle(& g_inputs.diameter,&fpin);
    }
    else if ( (strcmp(inputfield,"T")==0) || (strcmp(inputfield,"T_1")==0) ) {
      readinle(& g_inputs.T_1,&fpin);
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"T_2")==0) {
      readinle(& g_inputs.T_2,&fpin);
    }
#endif // XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"T0")==0) {
      readinle(& g_inputs.T0,&fpin);
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"XRSD_DELAY")==0) {
      readinle(& g_inputs.xrsd_delay,&fpin);
    }
    else if ( (strcmp(inputfield,"XRSD_2to1")==0) || (strcmp(inputfield,"XRSD_2toT")==0) )  {
      readinle(& g_inputs.xrsd_ratio_2toT,&fpin);
    }
#endif // XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"BEAM")==0) || (strcmp(inputfield,"PROFILE")==0)   ) {
      readinstr( g_inputs.beamtype,&fpin);
    }



    else if ( (strcmp(inputfield,"SAVE")==0) || (strcmp(inputfield,"FOTO")==0)  ) {
      readind(& g_inputs.foto,&fpin);
    }
    else if (  (strcmp(inputfield,"PROGRESS")==0) || (strcmp(inputfield,"PROG")==0) ||
               (strcmp(inputfield,"POS")==0) || (strcmp(inputfield,"POS_INDIC")==0)  ) {
      readind(& g_inputs.progress_dump,&fpin);
    }

    else if ( ( strcmp(inputfield,"LAT_PARAM")==0 ) ||  ( strcmp(inputfield,"DIST")==0 ) ) {
      readinle(& g_inputs.lattice_const,&fpin);
    }
    else if (strcmp(inputfield,"DEVI")==0) {
      readinle(& g_inputs.lattice_devi,&fpin);
    }
    else if (strcmp(inputfield,"LATTICE")==0) {
      readinstr( g_inputs.r1input,&fpin);
    }

    else if ( (strcmp(inputfield,"R0")==0) || (strcmp(inputfield,"POTENCRB")==0)  ) {
      readinle(& g_inputs.potencrb,&fpin); 
    }
    else if ( (strcmp(inputfield,"RKOK")==0) || (strcmp(inputfield,"tsOK")==0) || (strcmp(inputfield,"ts")==0) || (strcmp(inputfield,"md")==0) ) {
      readinstr( g_inputs.timestepOK,&fpin);
    }
    else if ( (strcmp(inputfield,"phOK")==0) || (strcmp(inputfield,"ph")==0)  )  {
      readinstr( g_inputs.photoionOK,&fpin);
    }
    else if ( (strcmp(inputfield,"flOK")==0) || (strcmp(inputfield,"fl")==0)  ) {
      readinstr( g_inputs.relaxflOK,&fpin);
    }
    else if ( (strcmp(inputfield,"auOK")==0) || (strcmp(inputfield,"au")==0)  ) {
      readinstr( g_inputs.relaxauOK,&fpin);
    } 
#ifdef XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"e_scatt")==0)||(strcmp(inputfield,"scOK")==0) ) {
      readinstr( g_inputs.scatteringOK,&fpin);
     }
    else if ( (strcmp(inputfield,"elOK")==0) || (strcmp(inputfield,"elsc")==0)  ) {
      readinstr( g_inputs.elscattOK,&fpin);
     } 
#endif // XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"siOK")==0) || (strcmp(inputfield,"si")==0)  ) {
      readinstr( g_inputs.elimpionOK,&fpin);
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"phANIZOTROPY")==0) || (strcmp(inputfield,"ph_aniz")==0)  ) {
      readinstr( g_inputs.anizotropy,&fpin);
    }

    else if (strcmp(inputfield,"Qphel")==0) {
      readind(& g_inputs.qphoto,&fpin);
    }
    else if (strcmp(inputfield,"Qauel")==0) {
      readind(& g_inputs.qauger,&fpin);
    }
    else if ( (strcmp(inputfield,"Qsecph")==0) || (strcmp(inputfield,"Qsiel<-phel")==0)  ) {
      readind(& g_inputs.qsecphoto,&fpin);
    }
    else if ( (strcmp(inputfield,"Qsecau")==0) || (strcmp(inputfield,"Qsiel<-auel")==0)  ) {
      readind(& g_inputs.qsecauger,&fpin);
    }
    else if ( (strcmp(inputfield,"Qsecsec")==0) || (strcmp(inputfield,"Qsiel<-siel")==0)  ) {
      readind(& g_inputs.qsecsec,&fpin);
    }
    else if (strcmp(inputfield,"Qfiel")==0) {
      readind(& g_inputs.qfion,&fpin);
    }
    else if (strcmp(inputfield,"Qatom")==0) {
      readind(& g_inputs.qatom,&fpin);
    }
    else if ( (strcmp(inputfield,"dQph")==0) || (strcmp(inputfield,"dQphion")==0)  ) {
      readind(& g_inputs.dqatomphotoion,&fpin);
    }
    else if ( (strcmp(inputfield,"dQau")==0) || (strcmp(inputfield,"dQauion")==0)  ) {
      readind(& g_inputs.dqatomauger,&fpin);
    }
    else if ( (strcmp(inputfield,"dQsi")==0) || (strcmp(inputfield,"dQsiion")==0)  ) {
      readind(& g_inputs.dqatomsec,&fpin);
    }
    else if ( (strcmp(inputfield,"dQfi")==0) || (strcmp(inputfield,"dQfion")==0)  ) {
      readind(& g_inputs.dqatomfion,&fpin);
    }


    else if (strcmp(inputfield,"PHI")==0) {
      readinle(& g_inputs.phiin,&fpin);
    }
    else if (strcmp(inputfield,"THETA")==0) {
      readinle(& g_inputs.thetain,&fpin);
    }
#endif // XMDYN_XRAYPAC_EXT
    
    else if ( (strcmp(inputfield,"RANDSEED")==0) || (strcmp(inputfield,"RANDOMSEED")==0)  ) {
      readind(& g_inputs.randseed,&fpin);
    }

    else if (strcmp(inputfield,"DBASE_TYPE")==0) {
      readinstr( g_inputs.dbase_types,&fpin);;
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"DBASE_XATOM")==0) {
      readind(& g_inputs.dbase_xatom_version,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"DBASEDIR")==0) {
      readinstr( g_inputs.dbasedir,&fpin);  
    }
    // *** FLAGS ***
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"TSFACTOR")==0) {
      readind(& g_inputs.tsfactor,&fpin);;
    }
    else if (strcmp(inputfield,"MAXSCATTRY")==0) {
      readind(& g_inputs.maxscattertry,&fpin);;
    }
    else if (strcmp(inputfield,"FIXATOMS")==0) {
      readinch(& g_inputs.fixatomsOK,&fpin);;
    }
    else if (strcmp(inputfield,"TSEL")==0) {
      readinch(& g_inputs.tselectrons,&fpin);;
    }
    else if (strcmp(inputfield,"SAVEPREC")==0) {
      readinstr( g_inputs.kiirasprec,&fpin);;
      strcat( g_inputs.kiirasprec," ");
    }
    else if (strcmp(inputfield,"AVALTYPE")==0) {
      readinstr( g_inputs.avaltype,&fpin);;
    }

    else if (strcmp(inputfield,"CENTERATOM")==0) {
      readind(& g_inputs.centeratom,&fpin);;
    }
    else if (strcmp(inputfield,"SAVEV3")==0) {
      readinch(& g_inputs.savev3,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"FOTODIM")==0)  || (strcmp(inputfield,"SNP_FULL")==0) || (strcmp(inputfield,"CHKP")==0)  ) {
      readinstr( g_inputs.snp_full_input , &fpin ) ;
    }
    else if ( strcmp(inputfield,"CHKP_DT")==0 ) {
      readinle( &g_inputs.chkp_dt,&fpin ) ;
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"IMPIONCRIT")==0) {
      readinstr( g_inputs.impioncriterium,&fpin);;
    }
    else if (strcmp(inputfield,"PRINTMSG")==0) {
      readinch(& g_inputs.printmessage,&fpin);;
    }
    else if (strcmp(inputfield,"DAEMON")==0) {
      readind(& g_inputs.daemon,&fpin);;
    }
    else if (strcmp(inputfield,"STARTINDEX")==0) {
      readind(& g_inputs.startindex,&fpin);;
    }
    else if (strcmp(inputfield,"STARTED")==0) {
      readind(& g_inputs.started,&fpin);;
    }
    else if (strcmp(inputfield,"INITFOTO")==0) {
      readind(& g_inputs.initfoto,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"INT_METHOD")==0) || (strcmp(inputfield,"MD_INTEGRATOR")==0)  ) {
      readinstr( g_inputs.int_method,&fpin);;
    }

#ifdef XMDYN_XRAYPAC_EXT
    else if ( strcmp( inputfield , "COULOMB_CALC" ) == 0 )  {
      readinstr( g_inputs.coulomb_calc_method , &fpin ) ;
    }

    else if ( (strcmp(inputfield,"LOGBUFFSIZE")==0) || (strcmp(inputfield,"PRINTBUFFSTORESIZE")==0)  ) {
      readind(& g_inputs.print_buff_store_size,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"MAXIMPDIST")==0) {
      readinle(& g_inputs.maximpdist,&fpin);;
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"DELOCTEST_NN")==0) {
      readind(& g_inputs.deloctest_nn,&fpin);;
    }
    else if (strcmp(inputfield,"DELOCTEST_COMM")==0) {
      readinch(& g_inputs.deloctest_comm,&fpin);;
    }
    else if (strcmp(inputfield,"DELOCTEST_NINTERP")==0) {
      readind(& g_inputs.deloctest_ninterp,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"RSTARTE")==0) {
      readinle(& g_inputs.rstarte,&fpin);;
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"CACHEDIR")==0) {
      readinstr( g_inputs.cachedatadir,&fpin);;
    }
    else if (strcmp(inputfield,"SI_TYPE")==0) {
      readinstr( g_inputs.si_type,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT
    else if ( (strcmp(inputfield,"ir")==0) || (strcmp(inputfield,"re")==0) )  {
      readinstr( g_inputs.instrecOK,&fpin);
    }
    else if (strcmp(inputfield,"MAXRECDIST")==0) {
      readinle(& g_inputs.maxrecdist,&fpin);;
    }
    else if (strcmp(inputfield,"IRECCOND")==0) {
      readinch(& g_inputs.ireccond,&fpin);;
    }
    else if ( (strcmp(inputfield,"IRECDT")==0) || (strcmp(inputfield,"RE_STEP")==0)  )  {
      readind(& g_inputs.irecDT,&fpin);;
    }
    else if ( (strcmp(inputfield,"IRECPER")==0) || (strcmp(inputfield,"RE_PER")==0) ) {
      readinle(& g_inputs.irecperiods,&fpin);;
    }
    else if ( (strcmp(inputfield,"IRECMETHOD")==0) || (strcmp(inputfield,"RE_METHOD")==0) ) {
      readinstr( g_inputs.irec_method , &fpin ) ;
    }
    else if ( (strcmp(inputfield,"IREC_ON")==0) || (strcmp(inputfield,"RE_ON")==0) ) {
      readinstr( g_inputs.irec_switchon , &fpin ) ;
    }
    else if ( (strcmp(inputfield,"IREC_MISC")==0) || (strcmp(inputfield,"RE_MISC")==0)   ) {
      readinstr( g_inputs.irec_misc , &fpin ) ;
    }
#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"SI_dsdW")==0) {
      readinch(& g_inputs.inel_dsdW,&fpin);;
    }
    else if (strcmp(inputfield,"NOTS")==0) {
      readinstr(  g_inputs.no_realspace_dyn,&fpin);;
    }
    else if (strcmp(inputfield,"CACHETIME")==0) {
      readinle(& g_inputs.maxcachetime,&fpin);;
    }
    else if (strcmp(inputfield,"bindE")==0) {
      readinstr( g_inputs.bindE,&fpin);;
    }
    else if (strcmp(inputfield,"ONESNP")==0) {
      readind(& g_inputs.onelastsnp,&fpin);;
    }
    else if (strcmp(inputfield,"SI_MODEL")==0) {
      readind(& g_inputs.si_model,&fpin);;
    }
    else if (strcmp(inputfield,"MISC_OPT")==0) {
      readinstr(  g_inputs.misc_opt , &fpin ) ;
    }
    else if (strcmp(inputfield,"IMPION_SCALE")==0) {
      readinle(& g_inputs.impion_scale,&fpin);;
    }
    else if (strcmp(inputfield,"IMPION_MODEL")==0) {
      readinstr(  g_inputs.impion_model , &fpin ) ;
    }
    else if (strcmp(inputfield,"XATOM_E")==0) {
      readinstr( g_inputs.xatom_energies,&fpin);;
    }
#endif // XMDYN_XRAYPAC_EXT

    else if (strcmp(inputfield,"SNP")==0) {
      readinstr( g_inputs.snp_input , &fpin ) ;
    }

#ifdef XMDYN_XRAYPAC_EXT
    else if (strcmp(inputfield,"SNP_E")==0) {
      readinstr( g_inputs.snp_E_input , &fpin ) ;
    }

#endif // XMDYN_XRAYPAC_EXT

    else if (strcmp(inputfield,"SAMPLE")==0) {
      readinstr( g_inputs.sample_dir , &fpin ) ;
    }

    else if ( 0 == strcmp( inputfield , "XATOM_OPTS" ) ) {
      readinstr( g_inputs.xatom_commandline_opt , &fpin ) ;
    }

#ifdef XMDYN_XRAYPAC_EXT
    else if ( 0 == strcmp( inputfield , "XATOM_SAVE" ) ) {
      readinstr( g_inputs.xatom_onthefly_savedir_string , &fpin ) ;
    }



    else if (strcmp(inputfield,"SHORTCUT")==0) {
      readinstr(  g_inputs.shortcut_comm , &fpin ) ;
    }


    else if (strcmp(inputfield,"FORMAT_BIN")==0) {
      readinstr(  g_inputs.txtbin_loadsave , &fpin ) ;
    }

#endif // XMDYN_XRAYPAC_EXT

  }
  
  else 
  {
    if (strcmp(inputfield,"STEPS")==0) {
      readind(& g_inputs.nstep,&fpin);
    }
    else if (strcmp(inputfield,"POS_INDIC")==0) {
      readind(& g_inputs.progress_dump,&fpin);
    }
  }
  
  }
  
  fclose(fpin);
 
  counter++;
  return(0);

}


//############################## FUNCTION #################################

int input_postproc( void) 
{
  char temp_str[LSTRLTH] ;
  double x ;

  if ( -1 == g_inputs.T_2 ) { g_inputs.T_2 = g_inputs.T_1 ; } 

  // MORE
  
  g_inputs.potencrb2=pow( g_inputs.potencrb,2);


  // MISC_OPT
  //
  if ( 0 != strlen( g_inputs.misc_opt ) )
  {
      strcpy( temp_str , g_inputs.misc_opt ) ;
      strcpy( g_inputs.misc_opt , ";")  ;
      strcat( strcat( g_inputs.misc_opt , temp_str ) , ";" ) ;
  }

  if ( ( 0 == strcmp( g_inputs.xatom_onthefly_savedir_string , "dbase" ) ) ||
       ( 0 == strcmp( g_inputs.xatom_onthefly_savedir_string , "global" ) ) )
  {
      g_inputs.xatom_onthefly_savedir = DATABASE ;
  }


   if (  g_inputs.randseed == -1 ) { g_inputs.randseed = 1 ; }


   // FILE/FOLDER 
   strcpy( g_inputs.datadir , "" ) ;
   strcat( strcat( strcat( g_inputs.datadir , g_inputs.rootdir ) , "/data/" ) , g_inputs.runnumber ) ;
   sys_path_operations( "~" ,  g_inputs.datadir )  ;

   if ( 0 == strcmp( "input.txt" , my_basename( g_inputs.inputfile ) ) )
   {
       strcpy( temp_str ,  g_inputs.inputfile ) ; 
       dirname( temp_str ) ;
       strcpy( g_inputs.datadir , temp_str ) ;
   }

   dirstrendcorrector( g_inputs.rootdir);
   dirstrendcorrector( g_inputs.dbasedir);

   // XFEL PARAMS
   g_inputs.Ephoton*= g_phys_math_const.e1 ; 
   if ( 0 == g_inputs.flu_max ) 
   { 
     g_inputs.flu_max = g_inputs.Nphoton / ( g_inputs.diameter * g_inputs.diameter *PI / 4. ) ;
   }

   if ( 0 == g_inputs.diameter )
   {
       g_inputs.diameter = sqrt( g_inputs.Nphoton / g_inputs.flu_max * 4. / PI ) ;
   }

   if ( 0 == g_inputs.Nphoton ) 
   {
       g_inputs.Nphoton = g_inputs.flu_max * ( g_inputs.diameter * g_inputs.diameter *PI / 4. ) ;
   }
#ifdef XMDYN_XRAYPAC_EXT
//   printf( "%le %le %le\n" , g_inputs.Nphoton , g_inputs.diameter , g_inputs.flu_max ) ;
#endif // XMDYN_XRAYPAC_EXT



   // PROGRESS
   if ( g_inputs.foto==0 )  g_inputs.foto= g_inputs.nstep+1;                                     


   if ( 0 == strlen( g_inputs.snp_input ) ) 
   {
       sprintf( g_inputs.snp_input , "%d:%d:%d" , g_inputs.foto, g_inputs.foto , g_inputs.nstep ) ;
   }


   char end_str[STRLTH] ;
   sprintf( end_str  , "%d" , g_inputs.nstep ) ;

   if ( 0 == strcmp( ":" , g_inputs.snp_input + strlen(g_inputs.snp_input) - 1  ) ) 
   { 
       strcat( g_inputs.snp_input , end_str ) ; 
   }

   if ( 0 == strcmp( ":" , g_inputs.snp_full_input + strlen(g_inputs.snp_full_input) - 1  ) ) 
   { 
       strcat( g_inputs.snp_full_input , end_str ) ; 
   }

   if ( 0 == strcmp( ":" , g_inputs.snp_E_input + strlen(g_inputs.snp_E_input) - 1  ) ) 
   { 
       strcat( g_inputs.snp_E_input , end_str ) ; 
   }

   
   convert_strlist2intarray( g_inputs.snp_input     , &(g_inputs.snp)     , &(g_inputs.snp_dim)     ) ;
   convert_strlist2intarray( g_inputs.snp_full_input, &(g_inputs.snp_full), &(g_inputs.snp_full_dim)) ;
   convert_strlist2intarray( g_inputs.snp_E_input   , &(g_inputs.snp_E)   , &(g_inputs.snp_E_dim)   ) ;

  
  if (  g_inputs.si_model == 1 )   
  {
       g_inputs.si_dsig_dome = 'i' ;
       g_inputs.si_restricted = 0 ;
  }
  if (  g_inputs.si_model == 2 )   
  {
       g_inputs.si_dsig_dome = 'i' ;
       g_inputs.si_restricted = 1 ;
  }
  if (  g_inputs.si_model == 3 )   
  {
       g_inputs.si_dsig_dome = 'f' ;
       g_inputs.si_restricted = 0 ;
  }
  if (  g_inputs.si_model == 4 )   
  {
       g_inputs.si_dsig_dome = 'f' ;
       g_inputs.si_restricted = 1 ;
  }

  if ( strstr(  g_inputs.misc_opt , "MC:atomic3" ) != NULL ) 
  {
      strcpy(  g_inputs.elimpionOK , "n" ) ;
      strcat(  g_inputs.misc_opt ,  ";no-energy-calc" ) ;
       g_inputs.rstarte = -1 ;

       g_inputs.qphoto =  g_inputs.qauger =  g_inputs.qsecphoto =  g_inputs.qsecauger = 
         g_inputs.qsecsec =  g_inputs.qfion = -1 ;
         g_inputs.dqatomphotoion =  g_inputs.dqatomauger = 
         g_inputs.dqatomsec =  g_inputs.dqatomfion = 1 ;
  }


  if ( strstr(  g_inputs.misc_opt , "MC:atomic2" ) != NULL ) 
  {
      strcpy(  g_inputs.timestepOK , "n" ) ;
       g_inputs.rstarte = 1 ;
       g_inputs.qphoto =  g_inputs.qauger =  g_inputs.qsecphoto =  g_inputs.qsecauger = 
         g_inputs.qsecsec =  g_inputs.qfion = 
         g_inputs.qatom =  g_inputs.dqatomphotoion =  g_inputs.dqatomauger = 
         g_inputs.dqatomsec =  g_inputs.dqatomfion = 0 ;
      strcat(  g_inputs.misc_opt , "-nolog:qerror" ) ;
  }

  if ( ( strstr( g_inputs.misc_opt , "MC:atomic" ) != NULL ) ||
       ( strstr( g_inputs.misc_opt , "MConly" ) != NULL ) )
  {
      strcpy(  g_inputs.timestepOK , "n" ) ;
      strcpy(  g_inputs.elimpionOK , "n" ) ;
      strcpy(  g_inputs.instrecOK  , "n" ) ;
      strcat(  g_inputs.misc_opt ,  ";no-energy-calc;MC:decoupled;" ) ;
      g_inputs.rstarte = -1 ;
      g_inputs.lattice_const = 10 ;

      g_inputs.qphoto =  g_inputs.qauger =  g_inputs.qsecphoto =  g_inputs.qsecauger = 
                         g_inputs.qsecsec =  g_inputs.qfion = -1 ;
      g_inputs.dqatomphotoion = g_inputs.dqatomauger = g_inputs.dqatomsec = g_inputs.dqatomfion = 1 ;
  }

  if ( strstr(  g_inputs.misc_opt , ";MC:decoupled;" ) != NULL )
  {
     g_inputs.quantumproc_coupled = DECOUPLED ;
  }

  if ( strstr(  g_inputs.misc_opt , ";synchrotron;" ) != NULL )
  {
     g_inputs.one_photoion = YES ;
  }

  if ( strstr(  g_inputs.misc_opt , ";DYN:Coulomb;" ) != NULL )
  {
     g_inputs.classdyn_partsel = COULOMB ;
  }


  sys_path_operations( "~" ,  g_inputs.dbasedir ) ; 
  sys_path_operations( "~" ,  g_inputs.rootdir ) ;

  return 0 ;

}



#define FIELDLTH 20
#define SAVE_SUMMARY(type,field,var) { fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %"MACRO_TO_STRING(type)"\n" , MACRO_TO_STRING(field) , var ) ; } 

//############################## FUNCTION #################################

int save_summaryfile(int step)
{
  FILE *fp;
  static int counter=0;
  char sfile[LSTRLTH],cstr[SSTRLTH],sstr[STRLTH];
  
  if ( 0 <= step )
  {
      strcpy(sfile,ALLFILES -> summaryfile);
      if (counter>0) {
        sprintf(cstr,"%d",counter);
        sprintf(sstr,"%d",step);
        strcat(strcat(sfile,"."),cstr);
      }

      fp=fopen(sfile,"w");
  }
  else
  { 
      fp = stderr ; 
  }

#ifdef XMDYN_XRAYPAC_EXT
  fprintf( fp , "MISC_OPT %s\n" ,  g_inputs.misc_opt ) ;
  fprintf(fp,"\n");
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp, "############# SYSTEM #############\n" ) ;
  SAVE_SUMMARY( d , N , g_inputs.N ) ;
  SAVE_SUMMARY( d , Z , g_inputs.Z ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"Eextra %d\n", g_inputs.Eextra);
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp,"\n");
  SAVE_SUMMARY( s , SAMPLE , g_inputs.sample_dir ) ;
  SAVE_SUMMARY( le , DIST , g_inputs.lattice_const ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"DEVI %le\n", g_inputs.lattice_devi);
  fprintf(fp,"LATTICE %s\n", g_inputs.r1input);
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp,"\n");
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"SHORTCUT %s\n", g_inputs.shortcut_comm );
  fprintf(fp,"\n");
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp, "############# PULSE ##############\n" ) ;
  SAVE_SUMMARY( le , EPH , g_inputs.Ephoton/ g_phys_math_const.e1 ) ;
  SAVE_SUMMARY( le , NPH , g_inputs.Nphoton ) ;
  SAVE_SUMMARY( le , DIAM , g_inputs.diameter ) ;
  SAVE_SUMMARY( le , FLU_MAX , g_inputs.flu_max ) ;
  SAVE_SUMMARY( le , T , g_inputs.T_1 ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"T_2 %le\n", g_inputs.T_2);
#endif // XMDYN_XRAYPAC_EXT
  SAVE_SUMMARY( le , T0 , g_inputs.T0 ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"XRSD_DELAY %le\n", g_inputs.xrsd_delay);
  fprintf(fp,"XRSD_2toT %le\n", g_inputs.xrsd_ratio_2toT);
#endif // XMDYN_XRAYPAC_EXT
  SAVE_SUMMARY( s , PROFILE , g_inputs.beamtype ) ;
  fprintf(fp, "\n########### PROCESSES ############\n" ) ;
  SAVE_SUMMARY( s , md , g_inputs.timestepOK ) ;
  SAVE_SUMMARY( s , ph , g_inputs.photoionOK ) ;
  SAVE_SUMMARY( s , fl , g_inputs.relaxflOK ) ;
  SAVE_SUMMARY( s , au , g_inputs.relaxauOK ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"elsc %s\n", g_inputs.elscattOK);//fprintf(fp,"elOK %s\n", g_inputs.elscattOK);
#endif // XMDYN_XRAYPAC_EXT
  SAVE_SUMMARY( s , si , g_inputs.elimpionOK ) ;
  SAVE_SUMMARY( s , re , g_inputs.instrecOK ) ;
#ifdef XMDYN_XRAYPAC_EXT
   fprintf(fp,"ph_aniz %s\n", g_inputs.anizotropy);//fprintf(fp,"phANIZOTROPY %s\n", g_inputs.anizotropy);

#endif // XMDYN_XRAYPAC_EXT

 

  fprintf(fp, "\n############ COULOMB #############\n" ) ;
  SAVE_SUMMARY( le , R0 , g_inputs.potencrb ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"COULOMB_CALC %s\n", g_inputs.coulomb_calc_method ) ;
  fprintf(fp,"\n");
#endif // XMDYN_XRAYPAC_EXT


  fprintf(fp, "\n########### INTEGRATOR ###########\n" ) ;
  SAVE_SUMMARY( le , DT , g_inputs.DT ) ;
  SAVE_SUMMARY( d , STEPS , g_inputs.nstep ) ;
  fprintf(fp,"\n");
  SAVE_SUMMARY( s , MD_INTEGRATOR , g_inputs.int_method ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"\n");
  fprintf(fp,"TSFACTOR %d\n", g_inputs.tsfactor);
  fprintf(fp,"\n");
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp, "\n########## DATA SAVING ###########\n" ) ;
  
  fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %s\n" , "SNP" , g_inputs.snp_input ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %s\n" , "CHKP" , g_inputs.snp_full_input ) ;
  fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %le\n" , "CHKP_DT" , g_inputs.chkp_dt ) ;
#endif // XMDYN_XRAYPAC_EXT
  fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %s\n" , "SNP_E" , g_inputs.snp_E_input ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf( fp , "%-"MACRO_TO_STRING(FIELDLTH)"s %s\n" , "FORMAT_BIN" , g_inputs.txtbin_loadsave ) ;
#endif // XMDYN_XRAYPAC_EXT
  fprintf( fp , "\n" ) ;

  SAVE_SUMMARY( d , PROGRESS , g_inputs.progress_dump ) ;
  //
  fprintf(fp, "\n########## RANDOM SEED  ##########\n" ) ;
  SAVE_SUMMARY( d , RANDSEED , g_inputs.randseed ) ;
  fprintf(fp, "\n############## MISC ##############\n" ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"bindE %s\n", g_inputs.bindE);
#endif // XMDYN_XRAYPAC_EXT
  SAVE_SUMMARY( le , RSTARTE , g_inputs.rstarte ) ;

#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp, "\n########## DAFAULT CHARGES / CHANGE OF CHARGES ##########\n" ) ;
  fprintf(fp,"Qphel %d\n", g_inputs.qphoto);
  fprintf(fp,"Qauel %d\n", g_inputs.qauger);
  fprintf(fp,"Qsecph %d\n", g_inputs.qsecphoto);//fprintf(fp,"Qsiel<-phel %d\n", g_inputs.qsecphoto);
  fprintf(fp,"Qsecau %d\n", g_inputs.qsecauger);//fprintf(fp,"Qsiel<-auel %d\n", g_inputs.qsecauger);
  fprintf(fp,"Qsecsec %d\n", g_inputs.qsecsec);//fprintf(fp,"Qsiel<-siel\n", g_inputs.qsecsec);
  fprintf(fp,"Qfiel %d\n", g_inputs.qfion);
  fprintf(fp,"Qatom %d\n", g_inputs.qatom);
  fprintf(fp,"dQph %d\n", g_inputs.dqatomphotoion);//fprintf(fp,"dQphion %d\n", g_inputs.dqatomphotoion);
  fprintf(fp,"dQau %d\n", g_inputs.dqatomauger);//fprintf(fp,"dQauion %d\n", g_inputs.dqatomauger);
  fprintf(fp,"dQsi %d\n", g_inputs.dqatomsec);//fprintf(fp,"dQsiion %d\n", g_inputs.dqatomsec);
  fprintf(fp,"dQfi %d\n", g_inputs.dqatomfion);//fprintf(fp,"dQsiion %d\n", g_inputs.dqatomsec);
  fprintf(fp,"\n");
  fprintf(fp,"PHI %le\n", g_inputs.phiin);
  fprintf(fp,"THETA %le\n", g_inputs.thetain);
#endif // XMDYN_XRAYPAC_EXT

  fprintf(fp, "\n####### XATOM AND DATABASE #######\n" ) ;
  SAVE_SUMMARY( s , XATOM_OPTS , g_inputs.xatom_commandline_opt ) ;
  SAVE_SUMMARY( s , DBASE_TYPE , g_inputs.dbase_types ) ;
  SAVE_SUMMARY( s , DBASEDIR , g_inputs.dbasedir ) ;
#ifdef XMDYN_XRAYPAC_EXT
  fprintf(fp,"DBASE_XATOM %d\n", g_inputs.dbase_xatom_version);
  fprintf(fp,"CACHEDIR %s\n", g_inputs.cachedatadir);
  fprintf(fp,"CACHETIME %le\n", g_inputs.maxcachetime);
  fprintf(fp, "\n########## DATA SAVEING ##########\n" ) ;
  fprintf(fp,"NOTS %s\n",  g_inputs.no_realspace_dyn);
  fprintf(fp,"TSEL %c\n", g_inputs.tselectrons);
  fprintf(fp,"FIXATOMS %c\n", g_inputs.fixatomsOK);
  //

  fprintf(fp, "\n########## OTHER CONDITIONS ##########\n" ) ;
  fprintf(fp,"XATOM_E %s\n", g_inputs.xatom_energies);
  fprintf(fp,"\n");
  fprintf(fp,"SI_TYPE %s\n", g_inputs.si_type);
  fprintf(fp,"SI_MODEL %d\n", g_inputs.si_model);
  fprintf(fp,"IMPION_SCALE %le\n" ,  g_inputs.impion_scale ) ;
  fprintf(fp,"IMPION_MODEL %s\n", g_inputs.impion_model);
  fprintf(fp,"MAXSCATTRY %d\n", g_inputs.maxscattertry);
  fprintf(fp,"MAXIMPDIST %le\n", g_inputs.maximpdist);
  fprintf(fp,"IMPIONCRIT %s\n", g_inputs.impioncriterium);
  fprintf(fp,"SI_dsdW %c\n", g_inputs.inel_dsdW);
  fprintf(fp,"\n");
  fprintf(fp,"IRECMETHOD %s\n", g_inputs.irec_method ) ;
  fprintf(fp,"IRECCOND %c\n", g_inputs.ireccond);
  fprintf(fp,"IRECDT %d\n", g_inputs.irecDT);
  fprintf(fp,"IRECPER %le\n", g_inputs.irecperiods);
  SAVE_SUMMARY( s , IREC_ON     , g_inputs.irec_switchon ) ; 
  SAVE_SUMMARY( s , IREC_MISC   , g_inputs.irec_misc ) ; 
  fprintf(fp,"MAXRECDIST %le\n", g_inputs.maxrecdist);
  fprintf(fp,"\n");
  fprintf(fp,"DELOCTEST_NN %d\n", g_inputs.deloctest_nn);
  fprintf(fp,"DELOCTEST_COMM %c\n", g_inputs.deloctest_comm);
  fprintf(fp,"DELOCTEST_NINTERP %d\n", g_inputs.deloctest_ninterp);
  fprintf(fp,"\n");
  //
  fprintf(fp, "\n########## SPEC. ##########\n" ) ;
  fprintf(fp,"AVALTYPE %s\n", g_inputs.avaltype);
  fprintf(fp,"CENTERATOM %d\n", g_inputs.centeratom);
  fprintf(fp, "\n########## OPTIONS ##########\n" ) ;
  fprintf(fp,"PRINTMSG %c\n", g_inputs.printmessage);
  fprintf(fp,"LOGBUFFSIZE %d\n", g_inputs.print_buff_store_size);//fprintf(fp,"PRINTBUFFSTORESIZE %d\n", g_inputs.print_buff_store_size);
  fprintf(fp,"DAEMON %d\n", g_inputs.daemon);
  fprintf(fp,"STARTINDEX %d\n", g_inputs.startindex);
  fprintf(fp,"STARTED %d\n", g_inputs.started);
  fprintf(fp,"INITFOTO %d\n", g_inputs.initfoto);
  fprintf(fp,"ONESNP %d\n", g_inputs.onelastsnp);
  fprintf(fp,"SAVEPREC %s\n", g_inputs.kiirasprec);
  fprintf(fp,"\n\n\n");
  fprintf(fp,"ATOM_DATA_SOURCE %s\n", g_inputs.atomic_data_source_string );
  fprintf(fp,"ATOM_DATA_EVAL %s\n", g_inputs.atomic_data_eval_string );
  fprintf(fp,"XATOM_SAVE %s\n", g_inputs.xatom_onthefly_savedir_string );

  fprintf(fp,"\n\n\n");
  fprintf(fp,"\n\n\n");
#endif // XMDYN_XRAYPAC_EXT
  fprintf(fp,"\n");
  
  if ( 0 <= step )
  {
      fclose(fp);
      counter++;

  }

  return 0 ;
}

#undef FIELDLTH


//############################## FUNCTION #################################

void helptext(char *svn_version , char *release_version , char *compile_command ) {
  printf("Version: %s\n", svn_version ) ;
  printf("Compiled with: %s\n", compile_command ) ;
  exit(0);
}


//############################## FUNCTION #################################

int flag_single_optstr( char ***p_argv , int *p_argc , char a_type , char *a_opt , void *a_dest )
{
    int found_error = 0 ;

    if ( ( (**p_argv)[2] != EoS ) && ( strcmp( (**p_argv) + 2 , a_opt ) == 0 ) ) 
    {
       found_error = 1 ;
       if ( 's' == a_type )
       {
           if ( ( --(*p_argc) == 0 ) || ( sscanf( (*++(*p_argv)) , "%s" , (char*) a_dest ) == 0 ) ) 
           { 
               found_error = -1000 ;
           } 
       }
    }

     if ( found_error < 0 ) { printf( "Wrong value for %s!\n" , a_opt ) ; }

     return found_error ;
}


//HELP: FLAGS
//############################## FUNCTION #################################

int flags(int argc,char **argv,char *svn_version ,char *version){
  char *s;
  int opt_found ;
  int zz ;

  while ( ( --argc > 0 ) && ( ( *++argv )[0] == '-' ) )
  {
    if ((*argv)[1]=='-') 
    {

      opt_found = 0 ; 

      opt_found += flag_single_optstr( &argv , &argc , 's' , "data_format" , (void*) g_inputs.format_loadsave ) ;
      opt_found += flag_single_optstr( &argv , &argc , 's' , "sample" , (void*) g_inputs.sample_dir ) ;

      printf( "%d %s\n" , opt_found , g_inputs.format_loadsave ) ;

      if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"maxscattertry")==0) ){
	if ((--argc==0) ||  (sscanf((*++argv),"%d",& g_inputs.maxscattertry)==0)){
	  printf("Wrong maxscattertry value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"new_dataformat")==0) )
      {
          strcpy( g_inputs.format_loadsave , "001" ) ; 
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"version")==0) )
      {
          printf( "%s\n" , svn_version ) ; exit( 0 ) ; 
      }


      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"centeratom")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.centeratom)==0)){
	  printf("Wrong centeratom value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"tsfactor")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.tsfactor)==0)){
	  printf("Wrong tsfactor value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"qatom")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.qatom)==0)){
	  printf("Wrong qatom value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"pbssize")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.print_buff_store_size)==0)){
	  printf("Wrong pbssize value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"maximpdist")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%le",& g_inputs.maximpdist)==0)){
	  printf("Wrong maximpdist value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"deloctest_nn")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.deloctest_nn)==0)){
	  printf("Wrong deloctest_nn value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"deloctest_ninterp")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.deloctest_ninterp)==0)){
	  printf("Wrong deloctest_ninterp value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"deloctest_comm")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%c",& g_inputs.deloctest_comm)==0)){
	  printf("Wrong deloctest_comm value!\n");
	  return(-1);
	}
      }




      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"startfile")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.startindex)==0)){
	  printf("Wrong startindex value!\n");
	  return(-1);
	}
      }




      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"screen")==0) ){
	 g_inputs.printmessage='y';
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"xatom-silent")==0) ){
	 g_inputs.xatom_silent = YES ;
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"xatom-silent-load")==0) ){
	 g_inputs.xatom_load_silent = YES ;
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"xatom-silent-calc")==0) ){
	 g_inputs.xatom_calc_silent = YES ;
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"daemon")==0) ){
	 g_inputs.daemon=1;
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"verlet")==0) ){
	strcpy( g_inputs.int_method,"verlet");
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"RK4")==0) ){
	strcpy( g_inputs.int_method,"RK4");
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"noelectrons")==0) ){
	 g_inputs.tselectrons='n';
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"initfoto")==0) ){
	 g_inputs.initfoto=1;
      }


      else if (((*argv)[2]!=EoS) && (strcmp((*argv)+2,"impioncriterium")==0)){
	if ((--argc==0) || (sscanf((*++argv),"%s", g_inputs.impioncriterium)==0)){
	  printf("Wrong impioncriterium value!\n");
	  return(-1);
	}
      }

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"avalanche")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%s", g_inputs.avaltype)==0)){
	  printf("Wrong avalanche value!\n");
	  return(-1);
	}
      }


      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"foto")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%d",& g_inputs.fotomomentsdim)==0)){
	  printf("Wrong foto value!\n");
	  return(-1);
	}
      }


      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"extra")==0) ){
	if ((--argc==0) || (sscanf((*++argv),"%s", g_inputs.extraopts)==0)){
	  printf("Wrong extra value!\n");
	  return(-1);
	}
      }


      else if ( ( (*argv)[2]!=EoS ) && ( strcmp((*argv)+2,"") == 0 ) ) {
	if ( (--argc==0) || ( sscanf((*++argv) , "%s" , g_inputs.format_loadsave ) == 0 ) ) {
	  printf( "Wrong  value!\n " ) ;
	  return -1  ;
	}
      }

      else if ( ( (*argv)[2]!=EoS ) && 
                ( ( strcmp((*argv)+2,"TEMPLATE1") == 0 ) || ( strcmp((*argv)+2,"TEMPLATE2") == 0 ) )  ) {
	if ( (--argc==0) || ( sscanf((*++argv) , "%s" , g_inputs.extraopts ) == 0 ) ) {
	  printf( "Wrong TEMPLATE value!\n " ) ;
	  return -1  ;
	}
      }



      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"fotofile")==0) )
	 g_inputs.fotofromfile='y';

      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"fixatoms")==0) )
	 g_inputs.fixatomsOK='y';


      else if ( ((*argv)[2]!=EoS) && (strcmp((*argv)+2,"help")==0) ){
        fprintf( stderr , "Execute without aruments!\n" ) ;
      }

      else if ( opt_found == 0 ) {
	printf("There's no such '--' flag!\n");
	return(-1);
      }

      else if ( opt_found < 0 ) {
	printf("Error in handling a flag!\n");
	return(-1);
      }

    }

    else{

      if((*argv)[1]==EoS){
	printf("There's no flag after '-'!\n");
	return(-1);
      }

      for(s=argv[0]+1;*s!=EoS;++s)

	switch(*s){

	case 'p':
          strcpy( g_inputs.kiirasprec,"% 1.12e ");
	  break;

	case 'v':
          g_inputs.event_dest = SCREEN_AND_LOGFILE ;
	  break;

	default:
	  printf("There's no such '-' flag!\n");
	  return(-1);

	}

    }

  }

  return(0);

}



//############################## FUNCTION #################################

int xmdyn_check_opts_while_running( char *a_cmd , void *a_par1 )
{
    int ii , dd ;
    double xx ;

    if ( 0 == strcmp( a_cmd , "irec-on" ) )
    {
        int xstep = 0 , xshift = 0 ;
        double xtime , ontime = 0 ;
        if ( strcmp( g_inputs.irec_switchon , "0" ) != 0 )
        {
        if ( strstr( g_inputs.instrecOK , "Y" ) != NULL )
        {
            if ( strstr( g_inputs.irec_switchon , "++" ) != NULL )
            {
                xshift = 2 ;
                ontime = g_inputs.T0 + g_inputs.xrsd_delay ;
            }
            sscanf( g_inputs.irec_switchon + xshift , "%le" , &xtime ) ;
            ontime += xtime ; 
            if ( ontime <= g_inputs.step * g_inputs.DT )
            {
                *(strstr( g_inputs.instrecOK , "Y" )) = 'y' ;
            }
        }}
    }



    return 0 ; // NOT YET FINISHED

    if ( 0 == strcmp( a_cmd , "si-off" ) )
    {
        int step ;
        step = *((int*)a_par1) ;
        if ( xx * 1e-15 < step * g_inputs.DT )
        {

        }
    }
}

