///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "../common.h"

#include "../globalconsts.h"
#include "../filetests.h"
#include "../endltabs.h"
#include "../classdyn.h"
#include "quantumproc.h"

#include "../xmdyn_io.h"

#include "../Xatom.h"

#include "e_release.h"

#include "auger.h"


#include "../tests.h"




// ########################
//  ATOMIC AUGER PROCESS
// ########################


//############################## FUNCTION #################################

int getaugerelectrons(int actatom,int eventid, int *i,int *j,int *k, int ishell , int jshell , int kshell ){
  int i_rand,j_rand,k_rand,ii,jj,kk,counter,returnval=0;
    struct g_struct_xatom_data *data ;

    data = g_atomic_data[actatom].atomic_econfig ;

  i_rand = floor( ( get_orbit_max_occupancy( ishell ) -  data->NE_orbit[ishell] ) * randnum(0) ) ;

  counter=0;
  ii = g_atomic_data[actatom].shell[ishell].e_from ;
  while ( ( counter <= i_rand ) &&
          ( ii <= g_atomic_data[actatom].shell[ishell].e_to ) ) {
    if ( !isitshellelectron(ii) )
      counter++;
    if ( counter<=i_rand )
      ii++;
  }
  *i = ii;
  if ( ii > g_atomic_data[actatom].shell[ishell].e_to )
    returnval -= 1;  // error



  j_rand = floor( g_atomic_data[actatom].atomic_econfig->NE_orbit[jshell] * randnum(0) ) ;
  counter = 0 ;
  jj = g_atomic_data[actatom].shell[jshell].e_from ;
  while ( ( counter <= j_rand ) &&
          ( jj <= g_atomic_data[actatom].shell[jshell].e_to ) ) {
    if ( isitshellelectron(jj) ) 
      counter++ ;
    if ( counter <= j_rand )
      jj++;
  }
  *j = jj;
  if ( jj > g_atomic_data[actatom].shell[jshell].e_to )
    returnval -= 1;  // error


  k_rand = floor( ( jshell == kshell ?
                    g_atomic_data[actatom].atomic_econfig->NE_orbit[kshell] - 1 :
	 	    g_atomic_data[actatom].atomic_econfig->NE_orbit[kshell] )
                   * randnum(0) ) ;
  counter = 0 ;
  kk = g_atomic_data[actatom].shell[kshell].e_from ;
  while ( ( counter <= k_rand ) &&
           ( kk <= g_atomic_data[actatom].shell[kshell].e_to ) ) {
    if ( isitshellelectron(kk) ) {
      if ( ( jshell == kshell ) && ( kk == *j ) )
        ;
      else
        counter++ ;
    }
    if ( counter <= k_rand )
      kk++ ;
  }
  *k = kk ;
  if ( kk > g_atomic_data[actatom].shell[kshell].e_to )
    returnval -= 1;  // error

  return( returnval );
}



//############################## FUNCTION #################################

int augeranizotropy(double *phi,double *sint, double *cost,int actatom,
                    int i,int j,int k){
  randvecsphere(phi,sint,cost);

  return 0 ;

}



//############################## FUNCTION #################################

int augerenergygain(int actatom,int i,int j,int k,int eventid , double *Egain, double Ediff ){
  *Egain = + g_atomic_data[actatom].atomic_econfig->B[i] 
           - g_atomic_data[actatom].atomic_econfig->B[j] 
           - g_atomic_data[actatom].atomic_econfig->B[k] ;

   if ( strcmp(  g_inputs.xatom_energies , "table" ) == 0 ) 
   {
       Ediff = g_atomic_data[actatom].atomic_econfig->au_E[eventid] ;
   }

   if ( strcmp(  g_inputs.xatom_energies , "Koopmans" ) == 0 ) 
   {
      Ediff  = + g_atomic_data[actatom].atomic_econfig->B[i] 
               - g_atomic_data[actatom].atomic_econfig->B[j] 
               - g_atomic_data[actatom].atomic_econfig->B[k] ;
   }

   if ( ( strcmp(  g_inputs.xatom_energies , "totalE" ) == 0 ) || ( strcmp(  g_inputs.xatom_energies , "mixed" ) == 0 ) )
   {


       Ediff =  g_atomic_data[actatom].atomic_econfig->E ;
       g_atomic_data[actatom].shell[i].NE ++ ;
       g_atomic_data[actatom].shell[j].NE -- ;
       g_atomic_data[actatom].shell[k].NE -- ;
       xmdyn_update_atoms( NULL , "list,econfig" , 1 , actatom ) ; // xmdyn_update_atoms( NULL , 1 , actatom );
       Ediff -=   g_atomic_data[actatom].atomic_econfig->E ;
       g_atomic_data[actatom].shell[i].NE -- ;
       g_atomic_data[actatom].shell[j].NE ++ ;
       g_atomic_data[actatom].shell[k].NE ++ ;
       xmdyn_update_atoms( NULL , "list,econfig" , 1 , actatom ) ; // xmdyn_update_atoms( NULL , 1 , actatom );

   }

  *Egain = Ediff ;

  return(0);

}


//############################## FUNCTION #################################

int augermomentumgain(int actatom,int i ,int j,int k,double *pgain) {
  pgain[0]=0;
  pgain[1]=0;
  pgain[2]=0;
  return(0);
}



//############################## FUNCTION #################################

int do_augerrelax(int actatom,int eventid,int step,double T1){
  int i,ishell,j,jshell,k,kshell,  reltry;
  double subpotEbef,subpotEaft,subpotEth,Ekinbef,Ekinaft ,  dEbound = 0 ;
  double Egain,pgain[3], Ediff=0 , new_re[3] , new_ve[3] , new_va[3] ;
  double phi,cost,sint ; //,Elostaft,Elostbef;
  int res , ii_mess ;
  int q_log ;

  static int maxreltry=5;
  static int relmissed=0;
  double temp_d ;

  char cfgstr_f[LSTRLTH] ;
  char cfgstr_i[LSTRLTH] ;

  int Ecalc_full = 0 ;
  char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;

  xmdyn_md_energy( "system-V" , "usecache" ) ;

  int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = actatom ;

  get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                      g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr_i, 'L'  ) ;

    Ediff = -1 ;  // TODO: to be removed finally

  ishell =  g_atomic_data[actatom].atomic_econfig->au_orbit[eventid][0] ;
  jshell =  g_atomic_data[actatom].atomic_econfig->au_orbit[eventid][1] ;
  kshell =  g_atomic_data[actatom].atomic_econfig->au_orbit[eventid][2] ;

  res = getaugerelectrons(actatom,eventid,&i,&j,&k, ishell , jshell , kshell );

  if ( res<0 ) {
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        print_log( " ERROR_au_e:" , NULL , "" , "noflush" ) ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        print_log_full( " q:" , (void*) &(g_system.qi[actatom])  , "int" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " i:" , (void*) &ishell  , "int" , "noflush" ) ;
        print_log( " j:" , (void*) &jshell  , "int" , "noflush" ) ;
        print_log( " k:" , (void*) &kshell  , "int" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
        temp_d = T1/MESSAGES.unit_t ;
        print_log( " T:" , (void*) &temp_d  , "double" , "noflush" ) ;
        print_log( " \n" , NULL , "" , "noflush" ) ;
        print_log( "" , NULL , "" , "flush" ) ;
#endif
    free( atoms_affected ) ; return(-2);
  }

  reltry=0;
  Ekinbef = xmdyn_md_energy( "system-K" , "calc" ) ; //X calc_tot_init_ener(2);
  augerenergygain( actatom , ishell , jshell , kshell , eventid , &Egain, Ediff );
  augermomentumgain(actatom, ishell , jshell , kshell , pgain ) ;

  if (Egain<=0){
    reltry=maxreltry;
  }

    q_log = g_system.qi[actatom] ;

  int indic = -1 ;
  while ( ( reltry < maxreltry ) && ( indic < 0  ) )
  {
      xmdyn_store_restore_subsystem( "store" , "-" , 2 , actatom , k ) ;
      g_atomic_data[actatom].shell[(int)ishell].NE++;
      g_atomic_data[actatom].shell[(int)jshell].NE--;

      augeranizotropy(&phi,&sint,&cost,actatom, ishell , jshell , kshell ) ;
      indic = xmdyn_electron_release_single( 'a' , actatom , k , kshell , 
                   g_inputs.qauger, g_inputs.dqatomauger , &subpotEbef , &subpotEth ,
                   new_re , new_ve , new_va , phi , cost , sint , Egain , pgain , &logstr) ;

      if ( indic < 0 ) {  xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;  }
      reltry++;
  }

  
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
  if ( indic < 0 ) 
  {
        print_log( "-au:" , NULL , "" , "noflush" ) ;
  }
  else
  {
        print_log( "+au:" , NULL , "" , "noflush" ) ;
  }

        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        //print_log_full( " q:" , (void*) &(g_system.qi[actatom])  , "int" , "noflush" ) ;
        print_log_full( " q:" , (void*) &q_log  , "int" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " i:" , (void*) &ishell  , "int" , "noflush" ) ;
        print_log( " j:" , (void*) &jshell  , "int" , "noflush" ) ;
        print_log( " k:" , (void*) &kshell  , "int" , "noflush" ) ;
        print_log( " hole:" , (void*) &i  , "int" , "noflush" ) ;
        print_log( " el1:" , (void*) &j  , "int" , "noflush" ) ;
        print_log( " el2:" , (void*) &k  , "int" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
#endif

  if ( indic < 0 ) 
  {
    relmissed++;
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        temp_d = T1/MESSAGES.unit_t ;
        print_log( " T:" , (void*) &temp_d  , "double" , "noflush" ) ;
        temp_d =  Egain / MESSAGES.unit_E ;
        print_log( " Egain:" , (void*) &temp_d  , "double" , "noflush" ) ;
        print_log_full( " cfg_i:" , (void*)cfgstr_i , "str" , "noflush" ) ;
        print_log( " \n" , NULL , "" , "noflush" ) ;
        print_log( "" , NULL , "" , "flush" ) ;
#endif
    free(logstr) ; free( atoms_affected ) ; return(-1);

  }
  else{
     xmdyn_store_restore_subsystem( "clear" , "-" , 0 ) ;  // q_log = 

    particle_params_swap(i,j);
     g_system.ptype [i]=subshellint2char( ishell );
    particle_rvq_zero(i);
     g_system.ptype [k]='a';

    array_max( g_system.uid + k , g_system.uid , g_inputs.N + g_inputs.E , "lli" ) ; g_system.uid[k]++ ;

  if ( ! Ecalc_full ) 
  {
    subpotEaft = potEactual(actatom,k,-1 ,num_atoms_affected,atoms_affected );
    xmdyn_md_energy( "system-V" , "getcache" ) ; //X calc_tot_init_ener(212);
    g_ENERGY_VALUES.totalpotE+=(subpotEaft-subpotEbef);
  }
  else
  {
    subpotEaft = xmdyn_md_energy( "system-V" , "calc" ) ; //X calc_tot_init_ener(12); subpotEaft=g_ENERGY_VALUES.totalpotE ;
  }

    Ekinaft = xmdyn_md_energy( "system-K" , "calc" ) ; //X calc_tot_init_ener(2);

     g_ENERGY_VALUES.Eintotal+=(Ekinaft-Ekinbef+subpotEaft-subpotEbef);
     g_ENERGY_VALUES.Eintotal_th+=Egain;
     g_ENERGY_VALUES.Einau+=Egain;
     g_ENERGY_VALUES.Elost+=(subpotEth-subpotEaft);
     g_ENERGY_VALUES.Ebound -= Egain ;
    
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        print_log( " uid:" , (void*) (g_system.uid + k) , "llint" , "noflush" ) ;
        temp_d = T1/MESSAGES.unit_t ;
        print_log( " T:" , (void*) &temp_d  , "double" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        temp_d =  Egain / MESSAGES.unit_E ;
        print_log( " Egain:" , (void*) &temp_d  , "double" , "noflush" ) ;
        temp_d = (Ekinaft - Ekinbef) / MESSAGES.unit_E ;
        print_log( " Tlocal:" , (void*) &temp_d  , "double" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG

        print_log_full( " cfg_i:" , (void*)cfgstr_i , "str" , "noflush" ) ;
        get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                            g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr_f, 'L'  ) ;
        print_log_full( " cfg_f:" , (void*)cfgstr_f , "str" , "noflush" ) ;

        print_log( " \n" , NULL , "" , "noflush" ) ; 
        print_log( logstr , NULL , "" , "noflush" ) ; 
        print_log( "" , NULL , "" , "flush" ) ; 
        xmdyn_apply_shortcut( "cancel" ) ;
#endif
  }


    free( logstr ) ; free( atoms_affected ) ; return(0);
}



//############################## FUNCTION #################################

int try_augerrelax(int actatom,double DT,int step, double *T, int *event )
{
  int sheme;
  double Tnew;

  *T = DT * 1.01 ;

  for ( sheme = 0 ; sheme < g_atomic_data[actatom].atomic_econfig->num_au ; ++sheme )
  {
    Tnew = draw_poisson( g_atomic_data[actatom].atomic_econfig->au_w[sheme] ) ;
    if ( ( Tnew > 0 ) && ( Tnew < *T ) )
    {
      *T = Tnew ;
      *event = sheme ;
    }
  }
  return 0 ;
}




