///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "../common.h"

#include "../globalconsts.h"
#include "../filetests.h"
#include "../endltabs.h"
#include "../classdyn.h"

#include "../xmdyn_io.h"

#include "e_release.h"


// #####################################################################
#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}


//############################## FUNCTION #################################

int electron_release_ver01(char effect,int actatom,int el1,int actshell,
             int qelectron,int dqatom,
	     double *subpotEbef_out,double *subpotEaft_out, 
             double *new_re , double *new_ve, double *new_va,
	     double phi,double cost,double sint,
	     double Egain,double *pgain, char **plogstr ) 
{
  double Ekin,Eatom,vv,VV,px,py,pz,sintn,costn,Mn;
  double sinf,cosf,sint2,cost2,sinf2,cosf2,sinfn,cosfn;
  double totalpotEbefore,totalpotEafter;
  double subpotEbef,subpotEaft;

  double me =  g_phys_math_const.me ;

  int Ecalc_full = 0 ;

  //* INIT ***************************************//
  int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = actatom ;
  totalpotEbefore = xmdyn_md_energy( "system-V" , "usecache" ) ; //X calc_tot_init_ener(212); totalpotEbefore= g_ENERGY_VALUES.totalpotE;

  if ( ! Ecalc_full ) {   subpotEbef = potEactual(actatom,-1,-1 , num_atoms_affected , atoms_affected ) ; }
  else                {   subpotEbef = totalpotEbefore ; }

  //* INIT ***************************************//

  Mn = g_system.m[actatom];
  Eatom=0.5*Mn*vectnorm2( g_system.v1 [actatom]);

  cosf=cos(phi);   sinf=sin(phi);
  sinf2=sinf*sinf; cosf2=cosf*cosf;
  cost2=cost*cost; sint2=sint*sint;

   g_system.r1 [el1][0]= g_system.r2 [el1][0]= g_system.r1 [actatom][0]+actualrb(actatom,actshell)*cosf*sint;
   g_system.r1 [el1][1]= g_system.r2 [el1][1]= g_system.r1 [actatom][1]+actualrb(actatom,actshell)*sinf*sint;
   g_system.r1 [el1][2]= g_system.r2 [el1][2]= g_system.r1 [actatom][2]+actualrb(actatom,actshell)*cost;


   g_system.qi [el1] = qelectron ;
   g_system.qi [actatom] += dqatom ;
   g_atomic_data[actatom].shell[actshell].NE--;
   g_atomic_data[actatom].NE--;
   g_system .motherion[el1] = actatom ;

   xmdyn_update_atoms( plogstr , "array,full,ionization" , num_atoms_affected, atoms_affected ) ; // xmdyn_update_atoms( NULL , -11 , num_atoms_affected, atoms_affected ) ;

  if ( ! Ecalc_full ) 
  {
  subpotEaft=potEactual(actatom,el1,-1 , num_atoms_affected , atoms_affected );
  totalpotEafter=totalpotEbefore-subpotEbef+subpotEaft;
  }
  else
  {
      subpotEaft = xmdyn_md_energy( "system-V" ,  "calc" ) ; //X calc_tot_init_ener(12) ; subpotEaft = g_ENERGY_VALUES.totalpotE ;
      totalpotEafter = subpotEaft;
  }

  pz=Mn* g_system.v1 [actatom][2]+pgain[2];
  py=Mn* g_system.v1 [actatom][1]+pgain[1];
  px=Mn* g_system.v1 [actatom][0]+pgain[0];

  Ekin=Eatom+totalpotEbefore-totalpotEafter+Egain;

  if ( (me*me*(py*py*sinf2*sint2+2*py*pz*cost*sint*sinf+2*px*py*sint2*cosf*sinf+2*px*pz*cosf*sint*cost+pz*pz*cost2+px*px*sint2*cosf2)+me*(me+Mn)*(2*Mn*Ekin-px*px-py*py-pz*pz))<0 ) 
  {
      free( atoms_affected ) ; return(-1);
  }


  vv=(me*(py*sint*sinf+pz*cost+px*sint*cosf)+sqrt(me*me*(py*py*sinf2*sint2+2*py*pz*cost*sint*sinf+2*px*py*sint2*cosf*sinf+2*px*pz*cosf*sint*cost+pz*pz*cost2+px*px*sint2*cosf2)+me*(me+Mn)*(2*Mn*Ekin-px*px-py*py-pz*pz)))/(me*(me+Mn));

  if ( (2*Ekin-me*vv*vv)/Mn < 0 ) 
  {
      free( atoms_affected ) ; return(-2);
  }

  VV=sqrt((2*Ekin-me*vv*vv)/Mn);
  costn=(pz-me*vv*cost)/Mn/VV;

  if ( (costn<-1) || (costn>1) ) 
  {
      free( atoms_affected ) ; return(-3);
  }
  sintn=sqrt(1-costn*costn);
  if (sintn==0){
    sinfn=0;
    cosfn=0;
  }
  else{
    sinfn=(py-me*vv*sint*sinf)/Mn/VV/sintn;
    cosfn=(px-me*vv*sint*cosf)/Mn/VV/sintn;
  }

  if (criteria_for_ionization(effect,actatom,el1,.5*me*vv*vv)<0)
  {
      free( atoms_affected ) ; return(-9);
  }

   g_system.v1 [el1][0]=vv*cosf*sint;
   g_system.v1 [el1][1]=vv*sinf*sint;
   g_system.v1 [el1][2]=vv*cost;
   g_system.v1 [actatom][0]=VV*cosfn*sintn;
   g_system.v1 [actatom][1]=VV*sinfn*sintn;
   g_system.v1 [actatom][2]=VV*costn;

   VECTOR3_EQUAL( g_system.r2[el1] , g_system.r1[el1] ) ;
  free( atoms_affected ) ;

  *subpotEbef_out=subpotEbef;
  *subpotEaft_out=subpotEaft;

  return(0);

}


//############################## FUNCTION #################################

int xmdyn_electron_release_single( char effect,int actatom,int el1,int actshell,
             int qelectron,int dqatom,
	     double *subpotEbef_out,double *subpotEaft_out, 
             double *new_re , double *new_ve, double *new_va,
	     double phi,double cost,double sint,
	     double Egain,double *pgain, char **plogstr) 
{

    int ret ;

    ret = electron_release_ver01( effect , actatom , el1 , actshell , qelectron , dqatom ,
                                  subpotEbef_out , subpotEaft_out , 
                                  new_re , new_ve , new_va , phi , cost , sint , Egain , pgain , plogstr ) ;

    return ret ;
}


//############################## FUNCTION #################################

int criteria_for_ionization(char effect,int actatom,int actel,double T) {
  int nearatom;
  double drnearatom[3];
  char criteriastr[STRLTH];
  double E0=0;
  double r_atom[3] ;
  
  strcpy( criteriastr ,  g_inputs.impioncriterium ) ;
  /* criteria: 'd': distance
               'c': Coulomb
	       'n': no extra criteria
  */

  if (T<0)
    return(-2);

  //* TODO: additional criterion may be used for the decision upon electron release, not in use currently *//
  if ( ( strstr( criteriastr , "Ccccccccccc" ) != NULL ) ||  ( strstr( criteriastr , "cccccccccccc" ) != 0 ) ) {
    nearatom = nearestatom2electron( actel , (double**)(&r_atom) , NULL , 'c' ) ;
    vectminus(drnearatom, g_system.r1 [actel], r_atom ) ;
    if ((T<0) ||
	(T+enerregcut(sqrt(vectnorm2(drnearatom)),
			  g_system.qi [nearatom],-1,nearatom)<0) )
      return(-3);
  }

  return(0);
}


// ##############################################################




#ifdef DO_DEBUG
#undef DO_DEBUG
#endif
#undef MODULE_ERROR

