///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "../common.h"

#include "../globalconsts.h"
#include "../filetests.h"
#include "../endltabs.h"
#include "../Xatom.h"

#include "../xmdyn_io.h"


#include "../classdyn.h"
#include "e_release.h"

#include "fluor.h"

#include "../tests.h"



//############################## FUNCTION #################################

int backpushingfluor(int aktatom, int i,int j, int ishell,int jshell , double Ediff )
{
  double *v1 = g_system.v1[aktatom]  ;
  double Mn  = g_system.m[aktatom] ;
  double c = g_phys_math_const.c ;

  double En,Eatom,Eatomf,VV;
  double phi,cost,sint,cosf,sinf , v_e , hb_ome ;

  double VV_old ;

  Eatom = 0.5 * Mn * vectnorm2( v1 ) ;

  En = Ediff ;

  randvecsphere( &phi , &sint , &cost ) ;
  cosf = cos( phi ) ; sinf = sin( phi ) ; 

  VV_old = sqrt( c * c +2.0 * En / Mn ) - c ;

  v_e = ( v1[0] * cosf * sint + v1[1] * sinf * sint + v1[2] * cost ) ;
  VV = + c - v_e
       - sqrt( ( c - v_e ) * ( c - v_e ) + 2.0 * Ediff / Mn ) ;
  hb_ome = Mn * c * ( -( c - v_e ) + sqrt( ( c - v_e ) * ( c - v_e ) + 2.0 * Ediff / Mn ) ) ; 
  
  v1[0] += VV * cosf * sint ;
  v1[1] += VV * sinf * sint ;
  v1[2] += VV * cost ;

  Eatomf = 0.5 * Mn * vectnorm2( v1 ) ;
  g_ENERGY_VALUES.Eintotal += ( Eatomf - Eatom ) ;  //<=ENERGY
  g_ENERGY_VALUES.Eintotal_th += ( Eatomf - Eatom ) ;  //<=ENERGY
  g_ENERGY_VALUES.Einfl += ( Eatomf - Eatom ) ;  //<=ENERGY
  g_ENERGY_VALUES.Efl_photon += hb_ome ;
  g_ENERGY_VALUES.Ebound -= Ediff ;

  return(0);
}



//############################## FUNCTION #################################

int getfluorelectrons(int actatom,int eventid,int *i,int *j) {
  int i_rand,j_rand,ii,jj,counter,returnval=0;
  int ishell , jshell ;
  struct g_struct_xatom_data *data ;

  data = g_atomic_data[actatom].atomic_econfig ;

  ishell = data->fl_orbit[eventid][0] ;
  jshell = data->fl_orbit[eventid][1] ; 

  i_rand = floor( ( get_orbit_max_occupancy( ishell ) -  data->NE_orbit[ishell] ) * randnum(0) );

  counter=0;
  ii=g_atomic_data[actatom].shell[ishell].e_from;
  while ( (counter<=i_rand) &&
          (ii<=g_atomic_data[actatom].shell[ishell].e_to) ){
    if (!isitshellelectron(ii))
      counter++;
    if (counter<=i_rand)
      ii++;
  }
  *i=ii;
  if (ii>g_atomic_data[actatom].shell[ishell].e_to)
    returnval-=1;  // error



  j_rand = floor( g_atomic_data[actatom].atomic_econfig->NE_orbit[jshell] * randnum(0) ) ;
  counter=0;
  jj=g_atomic_data[actatom].shell[jshell].e_from;
  while ( (counter<=j_rand) &&
          (jj<=g_atomic_data[actatom].shell[jshell].e_to) ){
    if (isitshellelectron(jj))
      counter++;
    if (counter<=j_rand)
      jj++;
  }
  *j=jj;
  if (jj>g_atomic_data[actatom].shell[jshell].e_to)
    returnval-=1;  // error

  return(returnval);

}




//############################## FUNCTION #################################

int do_fluorrelax(int actatom,int eventid,int step,double T1){
  int i,ishell,j,jshell,  reltry;
  double  Ediff=0 ;
  int res , ii_mess ;

  static int maxreltry=5;
  static int relmissed=0;

  double temp_d ;

  xmdyn_md_energy( "system-V" , "usecache" ) ;

  if ( ATOMDATA_USE_TAB == g_inputs.atomic_data_eval )
  {
     fprintf( stderr , "ERROR ERROR ERROR ERROR ERROR ERROR!!!!!!!!!!!!!!!!!!!!\n" ) ; exit(-1) ;
    
  }
  else if ( ATOMDATA_USE_TREE == g_inputs.atomic_data_eval )
  {
      Ediff = g_atomic_data[actatom].atomic_econfig->fl_E[eventid] ;
      ishell = g_atomic_data[actatom].atomic_econfig->fl_orbit[eventid][0] ;
      jshell = g_atomic_data[actatom].atomic_econfig->fl_orbit[eventid][1] ;
  }



  res = getfluorelectrons( actatom , eventid , &i , &j ) ; 

  if (res<0){
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        int sstep = 0 ;
        print_log( "ERROR_fl_e:" , NULL , "" , "noflush" ) ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " sstep:" , (void*) &sstep  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        print_log( " i:" , (void*) &ishell  , "int" , "noflush" ) ;
        print_log( " j:" , (void*) &jshell  , "int" , "noflush" ) ;
        temp_d = T1/MESSAGES.unit_t ;
        print_log( " T:" , (void*) &temp_d  , "double" , "noflush" ) ;
        print_log( " \n" , NULL , "" , "noflush" ) ;
        print_log( "" , NULL , "" , "flush" ) ;
#endif
    return(-2);
  }


  reltry=0;

  while ( (reltry<maxreltry) &&
                       (backpushingfluor(actatom,i,j,ishell,jshell,Ediff)==-1) )
    reltry++;

#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
  if (reltry==maxreltry) {
        print_log( "-fl:" , NULL , "" , "noflush" ) ;
  }
  else
  {
        print_log( "+fl:" , NULL , "" , "noflush" ) ;
  }

        int sstep = 0 ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " sstep:" , (void*) &sstep  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        print_log_full( " q:" , (void*) &(g_system.qi[actatom])  , "int" , "noflush" ) ;
        print_log( " i:" , (void*) &ishell  , "int" , "noflush" ) ;
        print_log( " j:" , (void*) &jshell  , "int" , "noflush" ) ;
        print_log( " hole:" , (void*) &i  , "int" , "noflush" ) ;
        print_log( " el2:" , (void*) &j  , "int" , "noflush" ) ;
        temp_d = Ediff /  g_phys_math_const.e1 ;
        print_log( " Egain:" , (void*) &temp_d  , "double" , "noflush" ) ;
        temp_d = T1/MESSAGES.unit_t ;
        print_log( " T:" , (void*) &temp_d  , "double" , "noflush" ) ;
#endif

  if (reltry==maxreltry) {
    relmissed++;
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        print_log( " \n" , NULL , "" , "noflush" ) ;
        print_log( "" , NULL , "" , "flush" ) ;
#endif
    return(-1);
  }

  else{
    int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = actatom ;

    particle_params_swap(i,j);
    g_system.ptype [i] = subshellint2char( ishell );
    particle_rvq_zero(i);
    g_atomic_data[actatom].shell[ishell].NE++ ;
    g_atomic_data[actatom].shell[jshell].NE-- ;

#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        char cfgstr[LSTRLTH] ;
        char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;
        get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                            g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr, 'L'  ) ;
        print_log_full( " cfg_i:" , (void*)cfgstr , "str" , "noflush" ) ;
        xmdyn_update_atoms( &logstr , "array,full" , num_atoms_affected, atoms_affected ) ;
        get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                            g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr, 'L'  ) ;
        print_log_full( " cfg_f:" , (void*)cfgstr , "str" , "noflush" ) ;

        print_log( " \n" , NULL , "" , "noflush" ) ; 
        print_log( logstr , NULL , "" , "noflush" ) ; 
        print_log( "" , NULL , "" , "flush" ) ; 

        free( logstr ) ;

#endif

  }

  return(0);
}



//############################## FUNCTION #################################

int try_fluorrelax(int actatom,double DT,int step, double *T, int *event)
{
  int sheme ;
  double Tnew;
  int num_event ;

  *T = DT * 1.01 ;

  if ( ATOMDATA_USE_TAB == g_inputs.atomic_data_eval )
  {
     fprintf( stderr , "ERROR ERROR ERROR ERROR ERROR ERROR!!!!!!!!!!!!!!!!!!!!\n" ) ; exit(-1) ;
  }

  else if ( ATOMDATA_USE_TREE == g_inputs.atomic_data_eval )
  {
      num_event = g_atomic_data[actatom].atomic_econfig->num_fl ;
      for ( sheme = 0 ; sheme < num_event ; ++sheme ) {

        Tnew = draw_poisson( g_atomic_data[actatom].atomic_econfig->fl_w[sheme] ) ; 
        if ( ( Tnew > 0 ) && ( Tnew < *T ) )
        {
          *T = Tnew ;
          *event = sheme ;
        }
      }
  }


  return 0 ;
}


