///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "../common.h"

#include "../globalconsts.h"
#include "../filetests.h"
#include "../endltabs.h"
#include "../classdyn.h"

#include "../xmdyn_io.h"
#include "../Xatom.h"

#include "e_release.h"

#include "pheff.h"

#include "../tests.h"


// ##########################################################################

static int s_do_debug = 0 ;


//############################## FUNCTION #################################

static double getexactphotoeffsigma(int actatom,int actshell)
{

    if ( ATOMDATA_USE_TAB == g_inputs.atomic_data_eval )
    {
        fprintf( stderr , "ERROR ERROR ERROR ERROR ERROR ERROR!!!!!!!!!!!!!!!!!!!!\n" ) ; exit(-1) ;
    }
    else if ( ATOMDATA_USE_TREE == g_inputs.atomic_data_eval )
    {
         return g_atomic_data[actatom].atomic_econfig->ph_sigma[actshell] ;
    }

    return -1 ; // Error

}


//############################## FUNCTION #################################

static double pheff_xphi(double z,double *params){
  // F , F[max]=1
  if (params[0]==1)
    return(0);
  if (params[0]==2)
    return(2*PI);

  return(.5*(z+sin(z)*cos(z))/PI);
}


//############################## FUNCTION #################################

double pheff_xth_distr(double z, double *params ) {
  double c1,c2,beta;
  if (params[0]==1)
    return(-1);
  if (params[0]==2)
    return(1);
  beta = params[1] ;
  c1=4.0/(3.0*pow(1+beta,2.0)*pow(beta-1,2.0));
  c2=-(2*beta+1)/(3.0*pow(beta,3.0)*pow(1+beta,2.0));
  return(((1-pow(beta,2.0)-3*beta*z+3*pow(beta,2.0)*pow(z,2.0))/
	  (3*pow(beta,3.0)*pow(beta*z-1,3.0))-c2)/c1);
}


//############################## FUNCTION #################################

static double pheff_xth(double z, double *params){
  // F , F[max]=1
  double beta, c1,c2 , params_th[3];

  if (params[0]==1)
    return(-1);
  if (params[0]==2)
    return(1);

  beta=initvreal(params[1],params[2])/ g_phys_math_const.c ;
  params_th[1] = beta ;
  return( pheff_xth_distr ( z,  params_th ) ) ;

  c1=4.0/(3.0*pow(1+beta,2.0)*pow(beta-1,2.0));
  c2=-(2*beta+1)/(3.0*pow(beta,3.0)*pow(1+beta,2.0));
  return(((1-pow(beta,2.0)-3*beta*z+3*pow(beta,2.0)*pow(z,2.0))/
	  (3*pow(beta,3.0)*pow(beta*z-1,3.0))-c2)/c1);

  // Distr.function for theta; gives back theta, not cos(theta)!!!
  // Not normalized: F(PI)!=1  
  if (params[0]==1)
    return(0);
  if (params[0]==2)
    return(PI);
  beta=initvreal(params[1],params[2])/ g_phys_math_const.c ;          
  return(-4.0/3.0 * (1-beta+3*pow(tan(0.5*z),2)+3*beta*pow(tan(1/2*z),2)) / pow(1+beta,2) / pow( pow(tan(0.5*z),2)+1+beta*pow(tan(0.5*z),2)-beta, 3 ));
  
}




//############################## FUNCTION #################################

static int photoeffanizotropy(double *phi,double *sinth, double *costh,
                       int actatom,int actshell){
  static double params_phi[2],params_th[5];
  double tempd[2];
  tempd[0]=0;tempd[1]=5e14;

  params_phi[1]=0;
  params_th[1]=actatom;
  params_th[2]=actshell;

  if (strcmp( g_inputs.anizotropy,"n")==0){
    randvecsphere(phi,sinth,costh);
    return(0);
  }
  if (strcmp( g_inputs.anizotropy,"y")==0){
    *phi=drawrandomdistr(pheff_xphi,params_phi,2,0);
    *costh=drawrandomdistr(pheff_xth,params_th,2,0);
    *sinth=sqrt(1-(*costh)*(*costh));
    return(0);
  }
  return 0 ;
}


//############################## FUNCTION #################################

static int photoeffenergygain(int actatom,int actshell,double *Egain , double *dEbound ){
    if ( ATOMDATA_USE_TAB == g_inputs.atomic_data_eval )
    {
        fprintf( stderr , "ERROR ERROR ERROR ERROR ERROR ERROR!!!!!!!!!!!!!!!!!!!!\n" ) ; exit(-1) ;
    }
    else if ( ATOMDATA_USE_TREE == g_inputs.atomic_data_eval )
    {
        *Egain = g_inputs.Ephoton - 
                 g_atomic_data[actatom].atomic_econfig->B[actshell] ;
        *dEbound = g_atomic_data[actatom].atomic_econfig->B[actshell] ;
    }
    return(0);
}


//############################## FUNCTION #################################

static int photoeffmomentumgain(int actatom,int actshell,double *pgain){
  pgain[0]=0;
  pgain[1]=0;
  pgain[2]= g_inputs.Ephoton/ g_phys_math_const.c ;
  return(0);
}


//############################## FUNCTION #################################

int do_pheffect(int actatom,int actshell,int step,double T1) {
  int actel,  phtry,i,j;
  double subpotEbef,subpotEaft,subpotEth,Ekinbef,Ekinaft,Egain , dEbound = 0 ;
  double pgain[3],phi,cost,sint;
  double tempd1,tempd2,tempd3,rel[3] , new_re[3] , new_ve[3] , new_va[3] ;
  int ii_mess = 0 ;
  int q_log ;

  static int phmissed=0;
  static int maxphtry=5;

  int Ecalc_full = 0 ;
  char cfgstr_i[LSTRLTH] ;

  char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;

  int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = actatom ;

  xmdyn_md_energy( "system-V" , "usecache" ) ;

  actel=getrand_shell_el(actatom,actshell);

  if (actel==-1){
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
    {
        double xxx ;
        print_log( "" , (void*) " ERROR_ph_e:" , "str" , "noflush" ) ;
        print_log( "step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( "atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( "Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        print_log( "shell:" , (void*) &actshell  , "int" , "noflush" ) ;
        xxx = T1 / MESSAGES.unit_t ;
        print_log( "T:" , (void*) &xxx , "double" ,  "noflush" ) ;
        print_log( "\n" , NULL , "" , "noflush" ) ;
    }
#endif
    free( atoms_affected ) ; return(-2);
  }

  get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                      g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr_i, 'L' ) ;
  phtry=0;
  Ekinbef = xmdyn_md_energy( "system-K" , "calc" ) ; //Xcalc_tot_init_ener(2);
  photoeffenergygain(actatom,actshell,&Egain, &dEbound );
  photoeffmomentumgain(actatom,actshell,pgain);

  q_log = g_system.qi[actatom] ;

  int indic = -1 ;
  while ( ( phtry < maxphtry ) && ( indic < 0  ) )
  {
      xmdyn_store_restore_subsystem( "store" , "-" , 2 , actatom , actel ) ;
      photoeffanizotropy( &phi , &sint , &cost , actatom , actshell ) ;
      indic = xmdyn_electron_release_single( 'p' , actatom , actel , actshell , 
                   g_inputs.qphoto , g_inputs.dqatomphotoion , &subpotEbef , &subpotEth , 
                   new_re , new_ve , new_va , phi , cost , sint , Egain , pgain , &logstr ) ;
      if ( indic < 0 ) {  xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;  }
      phtry++;
  }

#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
  if ( indic < 0 ) 
  {
        print_log( " -ph:" , NULL , "str" , "noflush" ) ;
  }
  else
  {
        print_log( "+ph:" , NULL , "str" , "noflush" ) ;
  }
  {
        double xxx ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        print_log_full( " q:" , (void*) &q_log , "int" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " shell:" , (void*) &actshell  , "int" , "noflush" ) ;
        print_log( " electron:" , (void*) &actel  , "int" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
  }
#endif

  if ( indic < 0 ) 
  {
    phmissed++;
        
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
    {
        double xxx ;
        xxx = T1 / MESSAGES.unit_t ;
        print_log( " T:" , (void*) &xxx , "double" , "noflush" ) ; 
        print_log( "\n" , NULL , "" , "noflush" ) ;
    }
#endif
    free(logstr) ; free( atoms_affected ) ; return(-1);
  }

  else
  {
     xmdyn_store_restore_subsystem( "clear" , "-" , 0 ) ;  // q_log = 

     g_system.ptype [actel]='f';

    array_max( g_system.uid + actel , g_system.uid , g_inputs.N + g_inputs.E , "lli" ) ; g_system.uid[actel]++ ;


  if ( ! Ecalc_full ) 
  {
    subpotEaft=potEactual(actatom,actel,-1 , num_atoms_affected , atoms_affected );
    g_ENERGY_VALUES.totalpotE = xmdyn_md_energy( "system-V" , "getcache" ) - subpotEbef + subpotEaft ; //X calc_tot_init_ener(212); g_ENERGY_VALUES.totalpotE= g_ENERGY_VALUES.totalpotE-subpotEbef+subpotEaft;
  }
  else
  {
    subpotEaft = xmdyn_md_energy( "system-V" , "calc" ) ; //X calc_tot_init_ener(12); subpotEaft=g_ENERGY_VALUES.totalpotE ;
  }
    Ekinaft = xmdyn_md_energy( "system-K" , "calc" ) ;   //X calc_tot_init_ener(2);
if ( 0 != s_do_debug ) { printf("DDD %le\n" , convert_unit_J2eV(subpotEaft) ); }

     g_ENERGY_VALUES.Eintotal+=(Ekinaft-Ekinbef-subpotEbef+subpotEaft);
     g_ENERGY_VALUES.Eintotal_th+=Egain;
     g_ENERGY_VALUES.Einph+=Egain;
     g_ENERGY_VALUES.Elost+=(subpotEth-subpotEaft);
     g_ENERGY_VALUES.Ebound += dEbound ;
    
#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
    {
        double xxx ;
        print_log( " uid:" , (void*) (g_system.uid + actel) , "llint" , "noflush" ) ;
        xxx = T1 / MESSAGES.unit_t ;
        print_log( " T:" , (void*) &xxx , "double" , "noflush" ) ; 
#ifdef XMDYN_XRAYPAC_MORELOG
        xxx = Egain / MESSAGES.unit_E ;
        print_log( " Egain:" , (void*) &xxx , "double" , "noflush" ) ;
        xxx = (Ekinaft - Ekinbef) / MESSAGES.unit_E ;
        print_log( " Tlocal:" , (void*)&xxx , "double" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG

        char cfgstr[LSTRLTH] ;
        print_log_full( " cfg_i:" , (void*)cfgstr_i , "str" , "noflush" ) ;
        get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                            g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr, 'L'  ) ;
        print_log_full( " cfg_f:" , (void*)cfgstr , "str" , "noflush" ) ;
        print_log( " \n" , NULL , "" , "noflush" ) ; 
        print_log( logstr , NULL , "" , "noflush" ) ; 
        print_log( "" , NULL , "" , "flush" ) ; 


    }
#endif
   if ( 0 != s_do_debug ) { int ii ; char *xxx=malloc(2*sizeof(char)); strcpy(xxx,""); printf("  UPDATE: "); for (ii=0;ii<num_atoms_affected;ii++) { printf("%4d ",atoms_affected[ii] ); } printf("\n"); free(xxx); } ; 

     
  }

  xmdyn_apply_shortcut( "cancel" ) ;
        
  free( logstr ) ; free( atoms_affected ) ; return(0);
}


//############################## FUNCTION #################################

int try_photoeffect(int actatom,double DT,int step,double *T,int *event)
{
  static int oldstep=-1 ;
  static double nph;
  double photosig, Tnew,w;
  int  actshell, maxshell ;
  struct ATOMS_STRUCT *actatomp;

  if ( ( step!=oldstep ) ) {
    beamparams( step , &nph ) ;
    oldstep = step ;
  }

  *T = DT * 1.01 ;
  actatomp = g_atomic_data + actatom ;


  if ( ATOMDATA_USE_TAB == g_inputs.atomic_data_eval )
  {
      maxshell = g_atomic_data[actatom].maxshellnum ;
  }
  else if ( ATOMDATA_USE_TREE == g_inputs.atomic_data_eval )
  {
      maxshell = g_atomic_data[actatom].atomic_econfig->max_orbit ;
  }


  for ( actshell = 1 ; actshell <= maxshell ; ++actshell )
  {

      photosig = getexactphotoeffsigma( actatom , actshell ) ;

      w = photosig * nph / (  g_inputs.diameter *  g_inputs.diameter / 4.0 * PI *  g_inputs.subDT ) ;

      Tnew = draw_poisson( w ) ;
      if ( ( Tnew > 0 ) && ( Tnew <*T ) ) 
      {
          *T = Tnew ;
          *event = actshell ;
      }
  }

  return 0 ;
}

