///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "../common.h"

#include "../globalconsts.h"
#include "../endltabs.h"

#include "../xmdyn_io.h"

#include "e_release.h"
#include "pheff.h"
#include "fluor.h"
#include "auger.h"

#include "quantumproc.h"

#include "../tests.h"


//############################## FUNCTION #################################

int get_qstoch_limits(int *from_atom,int *to_atom){
  static int f,t;
  int maxevents,i,c;
  int np,crank;
  static char method[LSTRLTH]="exact"; // "exact" "trivial"
  int My_Rank = 0 , Event_Rank = 0 ;
  
*from_atom = 0 ; *to_atom = g_inputs.N ; return (0) ;
}



//############################## FUNCTION #################################

int quantum_stochast_proc_entangled(int step , double DTtotal ) {
  double DT,T1;
  int actatom;
  char event;
  // different kinds:
  double Tph,Tfl,Tau;
  int phevent,flevent,auevent,eventid,targetatom,esuccess;
  int from_atom,to_atom;
  int My_Rank = 0 , Event_Rank = 0 ;

  get_qstoch_limits(&from_atom,&to_atom);
  
  DT =  g_inputs.subDT;

  while (DT!=0){
    T1=DT;
    targetatom=0; eventid=-1;
     
    for (actatom=from_atom;actatom<to_atom;++actatom){
      Tph=0;
      if (strcmp( g_inputs.photoionOK,"y")==0)
        try_photoeffect(actatom,DT,step,&Tph,&phevent);
      Tfl=0;
      if (strcmp( g_inputs.relaxflOK,"y")==0)
        try_fluorrelax(actatom,DT,step,&Tfl,&flevent);
      Tau=0;
      if (strcmp( g_inputs.relaxauOK,"y")==0)
        try_augerrelax(actatom,DT,step,&Tau,&auevent);

      if ( (Tph!=0) && (Tph<=T1) ){
        T1=Tph;
	event='p'; targetatom=actatom; eventid=phevent;
      }
      if ( (Tfl!=0) && (Tfl<=T1) ){
        T1=Tfl;
	event='f'; targetatom=actatom; eventid=flevent;
      }
      if ( (Tau!=0) && (Tau<=T1) ){
        T1=Tau;
	event='a'; targetatom=actatom; eventid=auevent;
      }
      Tph=Tfl=Tau=0;
    }

    if (My_Rank==Event_Rank){
      if ( (T1==0) || (T1>=DT) ) {
        DT=0;
      }
      else{
        switch (event) {
          case 'p':
  	    esuccess = do_pheffect(targetatom,eventid,step,
  	                           cd_time_now() +  g_inputs.subDT - DT + T1 );
            if ( YES == g_inputs.one_photoion ) {
                if ( esuccess == 0 ) {
                    strcpy(  g_inputs.photoionOK , "n" ) ;
                }
            }
  	    break;
	   case 'f':
	    esuccess=do_fluorrelax(targetatom,eventid,step,
  	                           cd_time_now() +  g_inputs.subDT - DT + T1 );
	    break;
	   case 'a':
	        esuccess=do_augerrelax(targetatom,eventid,step,
  	                               cd_time_now() +  g_inputs.subDT - DT + T1 );
	    break;
        }
        if (esuccess==0) {
          xmdyn_update_atoms( NULL , "list,full" , 1 , targetatom ) ; // xmdyn_update_atoms( NULL , 1 , targetatom );
        }
        DT-=T1;
      }
    }


  }

  return( 0 ) ;

}


//############################## FUNCTION #################################

int quantum_stochast_proc_disentangled(int step , double DTtotal ) {
  double DT,T1;
  int actatom;
  char event;
  // different kinds:
  double Tph,Tfl,Tau;
  int phevent,flevent,auevent,eventid,targetatom,esuccess;
  int from_atom,to_atom;
  int ii_atoms ;
  int *atom_permute_list , Natoms ;
  int My_Rank = 0 , Event_Rank = 0 ;

 
  get_qstoch_limits(&from_atom,&to_atom);
  Natoms = to_atom - from_atom ;
  
  atom_permute_list = calloc( Natoms , sizeof(int) ) ;

  rpermute_noalloc( Natoms , atom_permute_list  ) ;

  for ( ii_atoms = from_atom ; ii_atoms < to_atom ; ++ii_atoms ) {
    
    actatom = from_atom + atom_permute_list [ ii_atoms ] ;
    targetatom=0 ; 
  
  DT =  g_inputs.subDT;

  while (DT!=0){
    
    T1=DT;
    eventid=-1;
     
      Tph=0;
      if (strcmp( g_inputs.photoionOK,"y")==0)
        try_photoeffect(actatom,DT,step,&Tph,&phevent);
      Tfl=0;
      if (strcmp( g_inputs.relaxflOK,"y")==0)
        try_fluorrelax(actatom,DT,step,&Tfl,&flevent);
      Tau=0;
      if (strcmp( g_inputs.relaxauOK,"y")==0)
        try_augerrelax(actatom,DT,step,&Tau,&auevent);

      if ( (Tph!=0) && (Tph<=T1) ){
        T1=Tph;
	event='p'; targetatom=actatom; eventid=phevent;
      }
      if ( (Tfl!=0) && (Tfl<=T1) ){
        T1=Tfl;
	event='f'; targetatom=actatom; eventid=flevent;
      }
      if ( (Tau!=0) && (Tau<=T1) ){
        T1=Tau;
	event='a'; targetatom=actatom; eventid=auevent;
      }
      Tph=Tfl=Tau=0;


    if (My_Rank==Event_Rank){
      if ( (T1==0) || (T1>=DT) ) {
        DT=0;
      }
      else{
        switch (event) {
          case 'p':
  	    esuccess = do_pheffect(targetatom,eventid,step,
  	                           cd_time_now() +  g_inputs.subDT - DT + T1 );
            if ( YES == g_inputs.one_photoion ) {
                if ( esuccess == 0 ) {
                    strcpy(  g_inputs.photoionOK , "n" ) ;
                }
            }
  	    break;
	   case 'f':
	    esuccess=do_fluorrelax(targetatom,eventid,step,
  	                           cd_time_now() +  g_inputs.subDT - DT + T1 );
	    break;
	   case 'a':
	    esuccess=do_augerrelax(targetatom,eventid,step,
  	                           cd_time_now() +  g_inputs.subDT - DT + T1 );
	    break;
        }
        if (esuccess==0){
          xmdyn_update_atoms( NULL , "list,full" , 1 , targetatom ) ; // xmdyn_update_atoms( NULL , 1 , targetatom );
        }
        DT-=T1;
      }
    }


    }

  }
 
  free( atom_permute_list ) ;
  return( 0 ) ;

}



//############################## FUNCTION #################################

int quantum_stochast_proc(int step , double DTtotal ) {
    int ii ;

    if ( DECOUPLED == g_inputs.quantumproc_coupled ) {
        ii = quantum_stochast_proc_disentangled( step , DTtotal ) ;
    }
    else {
        ii = quantum_stochast_proc_entangled( step , DTtotal ) ;
    }
    return( ii ) ;

}


