///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <float.h>


#include "common.h"

#include "globalconsts.h"
#include "filetests.h"
#include "endltabs.h"
#include "Xatom.h"
#include "xmdyn_io.h"
//#include "elstart.h"
#include "classdyn.h"
#include "tests.h"
#include "secel.h"


// #####################################################################

#define SECEL_TABLE_DIM 5

// #define DBG_ENERGY

//#define MODULE_DEBUG

// #####################################################################

#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}


//############################## FUNCTION #################################

int xmdyn_get_classical_electrons( int *num , int **list )
{
    int N = g_inputs.N ;
    int E = g_inputs.E ;

    int ii ;

    *num = 0 ;
    for ( ii = N ; ii < N + E ; ii++ ) {  if ( isitfreeelectron( ii ) ) { (*num)++ ; } }

    (*list) = XMDYN_ALLOCATE_MEM( *num * sizeof(int) ) ;

    *num = 0 ;
    for ( ii = N ; ii < N + E ; ii++ ) {  if ( isitfreeelectron( ii ) ) {
        (*list)[*num] = ii ;
        (*num)++ ; 
    } }

    return 0 ;
}




//############################## FUNCTION #################################

double pot_of_2nd_nearest_ion( int el0 , int atom0 ) 
{
    int ii , atom1 ;
    double dr[3] , *re , potE , potEx ;

    re =  g_system.r1 [el0] ;

    if (  g_inputs.N == 1 ) { return( 0 ) ; }
    if ( atom0 == 0 ) { atom1 = 1 ; } else { atom1 = 0 ; } 

    vectminus( dr , re ,  g_system.r1 [atom1] ) ;
    potE = enerregcut( sqrt( vectnorm2( dr ) ) ,  g_system.qi [atom1],  g_system.qi [el0] , atom1 ) ;

    for ( ii = 0 ; ii < atom1 ; ii++ )
    {
        vectminus( dr , re ,  g_system.r1 [ii] ) ;
        potEx = enerregcut( sqrt( vectnorm2( dr ) ) ,  g_system.qi [ii],  g_system.qi [el0] , ii ) ;
        if ( potEx < potE ) { potE = potEx ; atom1 = ii ; }
    }
    for ( ii = atom1+1 ; ii <  g_inputs.N ; ii++ )
    {
        vectminus( dr , re ,  g_system.r1 [ii] ) ;
        potEx = enerregcut( sqrt( vectnorm2( dr ) ) ,  g_system.qi [ii],  g_system.qi [el0] , ii ) ;
        if ( potEx < potE ) { potE = potEx ; atom1 = ii ; }
    }

    return( potE ) ;

}


static struct instant_rec_struct
{
    double *Tall ;
    double *time ;
    int *nnatoms ;
    double *Ekin ;
    double *Epot ;
    double *Etot ;
    double *Etot0 ;
    double *phi ;
    double *phi_cum ;
    int *num_e_on_atom ;
}  l_instant_recomb ;





//############################## FUNCTION #################################

int xmdyn_store_restore_recombination( char *todo , int a_n , int *sel_par )
{
    static int npart = 0 ;
    static int *parts = NULL ;
    static double  *Tall = NULL, *time = NULL, 
                   *Ekin = NULL, *Epot = NULL, *Etot = NULL,  *Etot0 = NULL , 
                   *phi = NULL, *phi_cum = NULL ;
    static int *nnatoms = NULL, *num_e_on_atom = NULL;

    int Natom = g_inputs.N ;

    int ii , sel ;

    if ( strcmp( todo , "store" ) == 0 )
    {
        if ( NULL != parts ) { MODULE_ERROR( "ERROR: Storing when data already stored. Exit." , -1 ) ; }

        npart = a_n ;
        parts = XMDYN_ALLOCATE_MEM( a_n * sizeof(int) ) ;
        Tall = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        time = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        Ekin = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        Epot = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        Etot = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        Etot0 = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        phi = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        phi_cum = XMDYN_ALLOCATE_MEM( a_n * sizeof(double) ) ;
        nnatoms = XMDYN_ALLOCATE_MEM( a_n * sizeof(int) ) ;
        num_e_on_atom = XMDYN_ALLOCATE_MEM( a_n * sizeof(int) ) ;

        for ( ii = 0 ; ii < npart ; ii++ ) { parts[ii] = sel_par[ii] ; } 

        for ( ii = 0 ; ii < npart ; ii++ ) 
        {
            sel = parts[ii] ;
            if ( sel < Natom ) 
            {
                num_e_on_atom[ii] = l_instant_recomb.num_e_on_atom[sel] ;
            }
            else
            {
                Tall[ii] = l_instant_recomb.Tall[sel] ;
                time[ii] = l_instant_recomb.time[sel] ;
                Ekin[ii] = l_instant_recomb.Ekin[sel] ;
                Epot[ii] = l_instant_recomb.Epot[sel] ;
                Etot[ii] = l_instant_recomb.Etot[sel] ;
                Etot0[ii] = l_instant_recomb.Etot0[sel] ;
                phi[ii] = l_instant_recomb.phi[sel] ;
                phi_cum[ii] = l_instant_recomb.phi_cum[sel] ;
                nnatoms[ii] = l_instant_recomb.nnatoms[sel] ;
            }
        }
        return 0 ;
    }

    if ( strcmp( todo , "restore" ) == 0 )
    {
        if ( NULL == parts ) {  MODULE_ERROR( "ERROR: No data stored to be resored. Exit." , -1 ) ; }

        for ( ii = 0 ; ii < npart ; ii++ ) 
        {
            sel = parts[ii] ;
            if ( sel < Natom ) 
            {
                num_e_on_atom[ii] = l_instant_recomb.num_e_on_atom[sel] ;
            }
            else
            {
                l_instant_recomb.Tall[sel] = Tall[ii] ;
                l_instant_recomb.time[sel] = time[ii] ;
                l_instant_recomb.Ekin[sel] = Ekin[ii] ;
                l_instant_recomb.Epot[sel] = Epot[ii] ;
                l_instant_recomb.Etot[sel] = Etot[ii] ;
                l_instant_recomb.Etot0[sel] = Etot0[ii] ;
                l_instant_recomb.phi[sel] = phi[ii] ;
                l_instant_recomb.phi_cum[sel] = phi_cum[ii] ;
                l_instant_recomb.nnatoms[sel] = nnatoms[ii] ;
            }
        }
        return 0 ;
    }

    if ( strcmp( todo , "clear" ) == 0 ) 
    {
        if ( NULL == parts ) {  MODULE_ERROR( "ERROR: No data stored to deallocate. Exit." , -1 ) ; }
        XMDYN_DEALLOCATE_MEM( parts ) ; parts = NULL ;
        XMDYN_DEALLOCATE_MEM( Tall ) ; Tall = NULL ;
        XMDYN_DEALLOCATE_MEM( time ) ; time = NULL ;
        XMDYN_DEALLOCATE_MEM( Ekin ) ; Ekin = NULL ;
        XMDYN_DEALLOCATE_MEM( Epot ) ; Epot = NULL ;
        XMDYN_DEALLOCATE_MEM( Etot0 ) ; Etot0 = NULL ;
        XMDYN_DEALLOCATE_MEM( phi ) ; phi = NULL ;
        XMDYN_DEALLOCATE_MEM( phi_cum ) ; phi_cum = NULL ;
        XMDYN_DEALLOCATE_MEM( nnatoms ) ; nnatoms = NULL ;
        XMDYN_DEALLOCATE_MEM( num_e_on_atom ) ; num_e_on_atom = NULL ;
        return 0 ;
    }

    MODULE_ERROR( "ERROR: Misuse of function. Exit." , -1 ) ; 

    return 0 ;
}




//############################## FUNCTION #################################

int get_class_recomb_delta_charge( int atom )
{
    if ( g_inputs.N <= atom ) { return 0 ; } 
    return( - l_instant_recomb.num_e_on_atom[ atom ] ) ;
}




//############################## FUNCTION #################################

int module__recombination_loadsave( char *comm , FILE **fpp )
{
    int N = g_inputs.N + g_inputs.E ;
    int ii ; 

    if ( 0 == strcmp( comm , "load" ) )
    {
      if ( 0 )  // TOBE NEW
      {
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.time + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.Tall + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.nnatoms + ii , sizeof(int)    , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.Ekin + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.Epot + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.Etot + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.phi + ii     , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fread( l_instant_recomb.phi_cum + ii , sizeof(double) , 1 , *fpp ) ; }
      }
      else // OLD
      {
          for ( ii = 0 ; ii < N ; ii++) 
          {
               fread( l_instant_recomb.time + ii    , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.Tall + ii    , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.nnatoms + ii , sizeof(int)    , 1 , *fpp ) ;
               fread( l_instant_recomb.Ekin + ii    , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.Epot + ii    , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.Etot + ii    , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.phi + ii     , sizeof(double) , 1 , *fpp ) ;
               fread( l_instant_recomb.phi_cum + ii , sizeof(double) , 1 , *fpp ) ;
          }
      }
      for ( ii = 0 ; ii < N ; ii++) 
      {
          if ( ( l_instant_recomb.Tall[ii] < 0 ) && 
               ( -1 < l_instant_recomb.nnatoms[ii] ) 
             ) 
          { l_instant_recomb.num_e_on_atom[ l_instant_recomb.nnatoms[ii] ]++ ; } 
      }

    }

    else if ( 0 == strcmp( comm , "save" ) )
    {
      if ( 0 ) // TOBE NEW
      {
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.time + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.Tall + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.nnatoms + ii , sizeof(int)    , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.Ekin + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.Epot + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.Etot + ii    , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.phi + ii     , sizeof(double) , 1 , *fpp ) ; }
        for ( ii = 0 ; ii < N ; ii++) { fwrite( l_instant_recomb.phi_cum + ii , sizeof(double) , 1 , *fpp ) ; }
      }
      else // OLD
      {
          for ( ii = 0 ; ii < N ; ii++) 
          {
              fwrite( l_instant_recomb.time + ii    , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.Tall + ii    , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.nnatoms + ii , sizeof(int)    , 1 , *fpp ) ;
              fwrite( l_instant_recomb.Ekin + ii    , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.Epot + ii    , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.Etot + ii    , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.phi + ii     , sizeof(double) , 1 , *fpp ) ;
              fwrite( l_instant_recomb.phi_cum + ii , sizeof(double) , 1 , *fpp ) ;
          }
      }
    }
    else { MODULE_ERROR( "Wrong command." , -1 ) ; }
    return 0 ;
}


//############################## FUNCTION #################################

int calc_relphi_2parts( int at , int el , double *phi )
{
    double dr[3] , rr , sinphi , cosphi , (*r1)[3] ;

    r1 = g_system.r1 ;

    {
        vectminus( dr , r1[ l_instant_recomb.nnatoms[el] ] , r1[el] ) ;
        rr = sqrt( vectnorm2( dr ) ) ;
    }

    if ( rr > 0 )
    {
        cosphi = dr[0] / rr ;
        sinphi = dr[1] / rr ;
        *phi = SIGN( sinphi ) * acos( cosphi ) ;
        *phi  = *phi  - floor( *phi  / (2.*PI) ) * (2.*PI) ;
        return 0 ;
    }
    else
    {
        *phi = -1e50 ;
        return -1 ;
    }
    return 0 ;
}




#define DPHI_ACCUMULATION fabs
//############################## FUNCTION #################################

int recombiantion_follow_captured_electrons( void )
{
    int N = g_inputs.N ;
    int E = g_inputs.E ;

    int ii ;
    double phi , dphi ;

    for ( ii = N ; ii < N+E ; ii++ )
    {
//        if ( 1068 == ii ) { ii *= 1 ; printf("%le %le %le %le\n" , acos(0) , acos(0) , acos(-0.1) , acos(0.5) ) ; }
        if ( 0 <= l_instant_recomb.nnatoms[ii] )
        {
            if ( calc_relphi_2parts( l_instant_recomb.nnatoms[ii] , ii, &phi ) == 0 )
            {
                dphi = phi - l_instant_recomb.phi[ii] ;
                //dphi = dphi - floor( dphi / (2.*PI) ) * (2.*PI) ;
                if ( fabs( dphi ) > PI ) { dphi = dphi - SIGN( dphi ) * (2.*PI) ; }
                l_instant_recomb.phi_cum[ii] += DPHI_ACCUMULATION( dphi ) ;
                l_instant_recomb.phi[ii] = phi ;
            }
            else { l_instant_recomb.phi_cum[ii] = -1e50 ; }
        }
    }

#ifdef MODULE_DEBUG
if ( 1 ) //** DEBUG **//
{
    FILE *fp = fopen( "./recomb-debug-dyn.txt" , "a" ) ;
    for ( ii = N ; ii < N+E ; ii++ )
    {
        if ( l_instant_recomb.Etot[ii] < 0 )
        {
        int sstep = 0 ;
            fprintf( fp , "%6d %3d %6lld %6d   % le % le % le   % le % le\n" , 
                     g_inputs.step , sstep , 
                     g_system.uid[ii] , l_instant_recomb.nnatoms[ii] , 
                     convert_unit_J2eV( l_instant_recomb.Etot[ii] ) , 
                     convert_unit_J2eV( g_atomic_data[l_instant_recomb.nnatoms[ii]].atomic_econfig->B_rec ) , 
                     convert_unit_J2eV( g_atomic_data[l_instant_recomb.nnatoms[ii]].atomic_econfig->Ecrit_cl2q ) , 
                     l_instant_recomb.phi_cum[ii] / (2.*PI) , 
                     l_instant_recomb.phi[ii] / (2*PI) );
        }
    }
    fclose(fp) ;
}
#endif

    return 0 ;
}
#undef DPHI_ACCUMULATION





//############################## FUNCTION #################################

int save_snp_recombination_log( FILE **fp ) 
{
    int el ;
    double xxx ;
    double Ekin , Epot , Etot , Etot0 ;
    int atom ;

    int N = g_inputs.N ;
    int E = g_inputs.E ;
    long long int *uid = g_system.uid ;

    int width = floor( log10( N + E ) + 1 ) ;
    char wstr[STRLTH] , str[STRLTH] , llstr[STRLTH] ;

    sprintf( wstr, "%d" , width ) ;
    clearstr(str)   ;  strcat( strcat( strcat( str   , "%" ) , wstr ) , "d "   ) ;
    clearstr(llstr) ;  strcat( strcat( strcat( llstr , "%" ) , wstr ) , "lld " ) ;

    //fprintf( *fp , "# %s " ,  ) ;

    for ( el = N ; el < N + E ; el++ ) 
    {
        if ( l_instant_recomb.time[el] != 0 )
        {
            /*  trap columns: atom Z q el uid time #periodsT #periodsC Tperiod Ekin Epot Etot Etot_now recombined  */
            Ekin = l_instant_recomb.Ekin[el] ;
            Epot = l_instant_recomb.Epot[el] ;
            Etot = l_instant_recomb.Etot[el] ;
            Etot0 = l_instant_recomb.Etot0[el] ;
            atom = l_instant_recomb.nnatoms[el] ;
            fprintf( *fp , " " ) ;
            fprintf( *fp , str , atom ) ;
            fprintf( *fp , "%3d %3d " , g_system.Z[atom] ,g_system.qi[atom] ) ;
            fprintf( *fp , str , el ) ;
            fprintf( *fp , llstr , uid[el] ) ;
            xxx = l_instant_recomb.time[el] / MESSAGES.unit_t ;
            fprintf( *fp , "% le " , xxx ) ;
            if    ( 0 == l_instant_recomb.Tall[el]  )    { xxx = -1 ; }
            else    {  xxx = l_instant_recomb.time[el] /  l_instant_recomb.Tall[el]  ; }
            fprintf( *fp , "% le " , xxx ) ;
            xxx = l_instant_recomb.phi_cum[el] / (2.*PI) ;
            fprintf( *fp , " % le " , xxx ) ;
            xxx = l_instant_recomb.Tall[el] / MESSAGES.unit_t ;
            fprintf( *fp , "% le " , xxx ) ;
            Ekin /= MESSAGES.unit_E ;
            Epot /= MESSAGES.unit_E ;
            Etot /= MESSAGES.unit_E ;
            Etot0 /= MESSAGES.unit_E ;
            fprintf( *fp , "  % le % le % le   % le" , Ekin, Epot , Etot , Etot0 ) ;
            if ( ( l_instant_recomb.Tall[el] < 0 ) && ( 0 <= l_instant_recomb.nnatoms[el] ) ) 
                { fprintf( *fp , "   1" ); }
            else if ( ( 1 < l_instant_recomb.time[el] / l_instant_recomb.Tall[el] ) &&
                      ( 1 < l_instant_recomb.phi_cum[el] / (2.*PI) )
                    )
                { fprintf( *fp , "   2" ); }
            else 
                { fprintf( *fp , "   0" ); }
            fprintf( *fp , "\n" ) ;
        }
    }

    return 0 ;
}





//############################## FUNCTION #################################

static int get_classical_binding_energy( char *level , int at , int el , double *Etot , double *Ekin , double *Epot )
{
  double *rel = g_system.r1[el] ; 
  double *vel = g_system.v1[el] ;
  double *rat = g_system.r1[at] ; 
  double *vat = g_system.v1[at] ;

  double Rv1[3] , Rr1[3] , rr ;


  if ( strcmp( level , "0" ) == 0 )
  {
      vectminus(Rv1,vel,vat);
    
      { 
          vectminus(Rr1,rel,rat); 
          *Epot = enerregcut( sqrt(vectnorm2(Rr1)) , g_system.qi [el] , g_system.qi [at] , at ) ;
      }
    
      *Ekin = 0.5 *  g_phys_math_const.me  * vectnorm2(Rv1);
      *Etot = (*Ekin) + (*Epot) ;
  }

  else { MODULE_ERROR( "Unknown level." , -1 ) ; }

  return 0 ;

}




//############################## FUNCTION #################################

int try_instrec(int el,int at,char cond,int step,
                double *t_out,double *T_out, double *phiC , double *a_Bclass ) {

  double  fullcircle= g_inputs.irecperiods;

  double Epot = 0 , Ekin = 0 , Etot = 0 , alpha , T , x ;
  int i; 
  double Ecrit = 0 ;

  int do_log = 1 ;


  Ecrit = - g_atomic_data[at].atomic_econfig->Ecrit_cl2q ;

  // TODO: do we need this?
  if ( cond == 'F' ) { fullcircle = 1 ; }


  get_classical_binding_energy( "0" , at , el , &Etot , &Ekin , &Epot ) ;
  *a_Bclass = Etot ;


  if ( 0 <= Etot ) { at = -1 ; T = 0 ; }
  else
  {
      alpha =  g_phys_math_const.C_k * g_system.qi [at]* g_phys_math_const.e1 *  g_phys_math_const.e1  ;
      T = PI * alpha * sqrt( g_phys_math_const.me / ( 2.0 * (-1) * Etot * Etot * Etot ) );
   }

  if ( ( 't' == cond ) && 
       ( ( Ecrit <= Etot ) || ( at != l_instant_recomb.nnatoms[el] ) || ( cond == 'f' ) )
     ) 
  { return 0 ; }


      char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;
      int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = l_instant_recomb.nnatoms[el] ;

      // LOG if no recombination will happen of Last log
      //if ( ( ( at != l_instant_recomb.nnatoms[el] ) && ( l_instant_recomb.nnatoms[el] != -1 ) ) || 
      //     ( cond == 'f' ) )
      if ( ( at != l_instant_recomb.nnatoms[el] ) || ( cond == 'f' ) )
      {        

    #if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
        if ( ( cond == 'f' ) || ( strstr( g_inputs.instrecOK , "F" ) != NULL )  )
        {
          if ( cond == 'f' )
          {
              Ekin = l_instant_recomb.Ekin[el] ;
              Epot = l_instant_recomb.Epot[el] ;
              Etot = l_instant_recomb.Etot[el] ;
          }
          double xxx ; char idstr[STRLTH] = "" ;

#if 1
          /* We finish -- last log */
          if ( cond == 'f' )         { strcpy( idstr , "Lre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }
          /** Going from recombined to other: Reburn   */ 
          else if  ( l_instant_recomb.Tall[el] < 0  )  { strcpy( idstr , "Rre:" ) ; }
          /** Becoming Free  */
          else if ( 0 <= Etot )      { strcpy( idstr , "Fre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }
          /** Becoming Trapped  */
          else if ( ( Ecrit <= Etot ) && ( -1 == l_instant_recomb.nnatoms[el] ) )  { strcpy( idstr , "Tre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }
          /** Hopping  */
          else if ( ( Ecrit <= Etot ) && ( -1 != l_instant_recomb.nnatoms[el] ) )  { strcpy( idstr , "Hre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }
          /** Free -> free  */
          else if ( -1 == l_instant_recomb.nnatoms[el] ) { strcpy( idstr , "Gre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }
          /**  */
          else  { strcpy( idstr , "Dre:" ) ; if ( strstr(g_inputs.instrecOK , "F1" ) == NULL ) { do_log = 0 ; } }  
#endif
#ifndef XMDYN_XRAYPAC_MORELOG
          do_log = 0 ;
#endif // XMDYN_XRAYPAC_MORELOG

          if ( do_log )
          {
              print_log( "" , (void*) idstr , "str" , "noflush" ) ; 
              print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
              print_log( " atom:" , (void*) &(l_instant_recomb.nnatoms[el])  , "int" , "noflush" ) ;
              if ( -1 != l_instant_recomb.nnatoms[el] ) 
              {
              print_log( " Z:" , (void*) &(g_system.Z[l_instant_recomb.nnatoms[el]]) , "int" , "noflush" ) ;
              print_log( " q:" , (void*) &g_system.qi[l_instant_recomb.nnatoms[el]] , "int" , "noflush" ) ;
              print_log( " qcorr:" , (void*) &g_system.qi_corr[l_instant_recomb.nnatoms[el]] , "int" , "noflush" ) ;
              }
              if ( ! ( cond == 'f' ) )
              {
                  print_log( " atom2:" , (void*) &at  , "int" , "noflush" ) ;
                  if ( -1 != at ) 
                  {
                      print_log( " Z2:" , (void*) &(g_system.Z[at]) , "int" , "noflush" ) ;
                      print_log( " q2:" , (void*) &g_system.qi[at] , "int" , "noflush" ) ;
                  }
              }
              print_log( " el:" , (void*) &el , "int" , "noflush" ) ;
              print_log( " uid:" , (void*) (g_system.uid + el) , "llint" , "noflush" ) ;
              if ( -1 != l_instant_recomb.nnatoms[el] ) 
              {
                  xxx = Ekin / MESSAGES.unit_E ;
                  print_log( " Ekin:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  xxx = Epot / MESSAGES.unit_E ;
                  print_log( " Epot:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  xxx = Etot / MESSAGES.unit_E ;
                  print_log( " Etot:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  if ( ! (  ( Ecrit <= Etot )  ||  ( cond == 'f' )  ) )
                  {
                      xxx = l_instant_recomb.Epot[el] / MESSAGES.unit_E ;
                      print_log( " Epot0:" , (void*) &xxx , "double" ,  "noflush" ) ;
                      xxx = l_instant_recomb.Etot[el] / MESSAGES.unit_E ;
                      print_log( " Etot0:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  }
                  xxx = l_instant_recomb.Tall[el] / MESSAGES.unit_t ;
                  print_log( " T:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  xxx = l_instant_recomb.time[el] / MESSAGES.unit_t ;
                  print_log( " t:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  if    ( 0 == l_instant_recomb.Tall[el]  )    { xxx = -1 ; }
                  else    {  xxx = l_instant_recomb.time[el] /  l_instant_recomb.Tall[el]  ; }
                  print_log( " perT:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  xxx = l_instant_recomb.phi_cum[el] / (2.*PI) ;
                  print_log( " phiC:" , (void*) &xxx , "double" ,  "noflush" ) ;
                  //print_log( " \n" , NULL , "" , "noflush" ) ;
                  //print_log( "" , NULL , "" , "flush" ) ;
              }
          }
        }
        else { do_log = 0 ;  } ;
    #endif

      }

      if ( cond == 'f' ) { if (do_log) { print_log( " \n" , NULL , "" , "flush" ) ; } free( logstr ) ; free( atoms_affected ) ; return 0 ; }
  


      if ( ( at != l_instant_recomb.nnatoms[el]  ) )
      { 
          if ( ( strstr( g_inputs.instrecOK , "F" ) != NULL ) && 
               ( l_instant_recomb.Tall[el] < 0 ) && ( 0 <= l_instant_recomb.nnatoms[el] ) )  
          { 
              double Vi=0, Vf=0 ;

              if ( 0 <= l_instant_recomb.nnatoms[el] )
              {
                { Vi = potEactual(l_instant_recomb.nnatoms[el],-1,-1,0,NULL); } 
              }
              l_instant_recomb.num_e_on_atom[l_instant_recomb.nnatoms[el]]-- ; 
              if ( 0 <= l_instant_recomb.nnatoms[el] )
              {
                //el_config_update( l_instant_recomb.nnatoms[el] ) ;
                xmdyn_update_atoms( &logstr , "list,full" , 1 , l_instant_recomb.nnatoms[el] ) ; // xmdyn_update_atoms( &logstr , 1 , l_instant_recomb.nnatoms[el] );
                { Vf = potEactual(l_instant_recomb.nnatoms[el],-1,-1,0,NULL); } 
                g_ENERGY_VALUES.Eintotal    += ( Vf - Vi ) ;
                g_ENERGY_VALUES.Eintotal_th += ( Vf - Vi ) ;
                g_ENERGY_VALUES.Einrec      += ( Vf - Vi ) ;
                Vf = ( Vf - Vi ) / MESSAGES.unit_E ; 
              }
              if ( do_log ) { print_log( " dV:" , (void*) &Vf , "double" ,  "noflush" ) ; }
          }
          //if ( ( strstr( g_inputs.instrecOK , "F" ) != NULL ) && ( l_instant_recomb.Tall[el] < 0 ) && (  l_instant_recomb.nnatoms[el] < 0  ) )  { printf("1 %d %d\n" , l_instant_recomb.nnatoms[el] , el );  } ; // DEBUG
          l_instant_recomb.nnatoms[el] = at ; 
          l_instant_recomb.time[el] = 0 ;
          l_instant_recomb.Tall[el] = 0 ;
          l_instant_recomb.Ekin[el] = 0 ;
          l_instant_recomb.Epot[el] = 0 ;
          l_instant_recomb.Etot[el] = 0 ;
          l_instant_recomb.Etot0[el] = 0 ;
          if ( 0 <= at ) 
          { 
              calc_relphi_2parts( at , el , &(l_instant_recomb.phi[el]) ) ; 
              l_instant_recomb.phi_cum[el] = 0 ;
          }
          else 
          { 
              l_instant_recomb.phi[el] = -1 ; 
              l_instant_recomb.phi_cum[el] = 0 ;
          }

          if ( do_log ) 
          { 
              print_log( " \n" , NULL , "" , "noflush" ) ; 
              print_log( logstr , NULL , "" , "noflush" ) ; 
              print_log( "" , NULL , "" , "flush" ) ; 
          }
          free( logstr ) ;
          free( atoms_affected ) ;

          return(0) ;
      }

      free( logstr ) ;
      free( atoms_affected ) ;


      //l_instant_recomb.Tall[el] = T ;

      if ( 't' == cond )
      {
         if  ( ( Ecrit > Etot ) && ( fullcircle > 0 ) && ( l_instant_recomb.time[el] > T * fullcircle ) ) { return 1 ; }
         else { return 0 ; }
      }

      if ( l_instant_recomb.Tall[el] < 0 )
      {
          l_instant_recomb.time[el] += g_inputs.DT * g_inputs.irecDT ;
          l_instant_recomb.Etot0[el] = Etot ;
          return 0 ;
      }

      l_instant_recomb.Tall[el] = T ; 
      l_instant_recomb.Ekin[el] = Ekin ; 
      l_instant_recomb.Epot[el] = Epot ; 
      l_instant_recomb.Etot[el] = Etot ;
      l_instant_recomb.Etot0[el] = Etot ;
      
      if ( 0 <= Etot ) { return 0 ; } 

      if ( Ecrit <= Etot ) { l_instant_recomb.time[el] += g_inputs.DT * g_inputs.irecDT ; return 0 ; } 
      
      if ( ( fullcircle > 0 ) && 
           ( l_instant_recomb.time[el] > T * fullcircle ) && 
           ( fabs( l_instant_recomb.phi_cum[el] ) >  2*PI * fullcircle ) 
         ) 
      {
          *t_out = l_instant_recomb.time[el] ;
	  *T_out = T ; //*fullcircle;
          *phiC = l_instant_recomb.phi_cum[el] ; 
          //if ( strcmp( g_inputs.instrecOK , "L" ) == 0 )
          l_instant_recomb.time[el] += g_inputs.DT * g_inputs.irecDT ;
          if ( strstr( g_inputs.instrecOK , "F" ) != NULL )
          {
  	      l_instant_recomb.Tall[el] *= -1 ;
              if ( 0 <= at ) 
              { 
                  l_instant_recomb.num_e_on_atom[at]++ ; 
                  //el_config_update( l_instant_recomb.nnatoms[el] ) ; 
              }
              //if ( at < 0  ) { printf( "2 %d %d\n" , at , el ) ; } // DEBUG
          }
          else
          {
          //*  the lines below went to function  do_instrec *//
//   	      l_instant_recomb.time[el] = 0 ;
// 	      l_instant_recomb.Tall[el] = 0 ;
//               l_instant_recomb.phi_cum[el] = 0 ;
//               l_instant_recomb.Ekin[el] = 0 ;
//               l_instant_recomb.Epot[el] = 0 ;
//               l_instant_recomb.Etot[el] = 0 ;
//               l_instant_recomb.Etot0[el] = 0 ;
           }
          return 1 ;
      }
      else 
      {
          l_instant_recomb.time[el] += g_inputs.DT * g_inputs.irecDT ;
          return 0 ;
      }
 
  return(0);
}


 
#if 1
//############################## FUNCTION #################################

static int xmdyn_3Brecomb_add_pE_to_el( double Mat , double Mel ,
                                        double *v0_at , double *v0_el , double *v_at , double *v_ep , 
                                        double KE_extra  , double *a_afactor ) 
{
    double vx2 ;
    double xxx[3] ;
    double a_1 , a_2 , aa , coeff_a , coeff_b , coeff_c ;
    int res ;

    //** For testing only: all energy to electron, no momentum conservation **//
    if ( 0 )  
    {
        vx2 = VECTOR3_NORM2( v_ep ) ; 
        // KE_extra contains KE_ion that has to be removed as it does not change
        //double vnew2 = vx2 + 2. * ( KE_extra - 0.5 * Mat * VECTOR3_NORM2(v0_at) ) / Mel ;
        double vnew2 = vx2 + 2. * KE_extra / Mel ;
        if ( vnew2 < 0 ) { return -1 ; }
        VECTOR3_TIMES_SCALAR( v_ep , v_ep , sqrt( vnew2 / vx2 ) ) ;
        vx2 = VECTOR3_NORM2( v_ep ) ; 
        return 0 ;
    }

    //** Energy and momentum conservation with inclusion of the atom **//

    vx2 = VECTOR3_NORM2( v_ep ) ;

    coeff_a = -0.5 * Mel*Mel * vx2 / Mat - 0.5 * Mel * vx2 ;
    coeff_b = Mel * VECTOR3_DOT_PRODUCT( v0_at , v_ep ) + 
              Mel*Mel/Mat * ( VECTOR3_DOT_PRODUCT( v0_el , v_ep ) + vx2 ) ;
        VECTOR3_ADD( xxx , v0_el , v_ep ) ;
        VECTOR3_TIMES_SCALAR( xxx , xxx , Mel / Mat ) ;
        VECTOR3_ADD( xxx , v0_at , xxx ) ;
    coeff_c = KE_extra + 0.5 * Mat * VECTOR3_NORM2( v0_at ) + 0.5 * Mel * vx2 - 0.5 * Mat * VECTOR3_NORM2( xxx ) ;

    res = math_solve_sec_order_alg_eq( coeff_a , coeff_b , coeff_c , &a_1 , &a_2 ) ;
    if ( 0 == res ) { return -1 ; }

    if ( fabs( a_1 ) < fabs( a_2 ) ) { aa = a_2 ; } else { aa = a_1 ; }

    VECTOR3_TIMES_SCALAR( xxx , v_ep , (1-aa) ) ;
    VECTOR3_ADD( xxx , xxx , v0_el ) ;
    VECTOR3_TIMES_SCALAR( xxx , xxx , Mel/Mat ) ;
    VECTOR3_ADD( xxx , xxx , v0_at ) ;

    //**  We change the coordinates here  *//
    VECTOR3_EQUAL( v_at , xxx ) ;
    VECTOR3_TIMES_SCALAR( v_ep , v_ep , aa ) ;

    *a_afactor = aa ;

    return 0 ;
}




//############################## FUNCTION #################################

static int get_nearest_electron( int p_ref , int el_rec , double *dlim , int *nbor )
{

     int num , *list = NULL , ii , found ;
     double  dmin = DBL_MAX , dd ;

     xmdyn_get_classical_electrons( &num , &list ) ;

     *nbor = -1 ;

     found = 0 ;
     for ( ii = 0 ; ii < num ; ii++ ) 
     {
         if ( el_rec != list[ii] )
         {
             dd = xmdyn_system_get_particle_distance( p_ref , list[ii] , 'n' ) ;
             if ( ( *dlim < dd ) && ( dd < dmin  ) ) 
             { 
                 *nbor = list[ii] ; dmin = dd ;  found = 1 ; 
             }
         }
     }
     XMDYN_DEALLOCATE_MEM(list) ;
     *dlim = dmin ;
     return 0 ;
}



#endif




//############################## FUNCTION #################################

int do_instrec(int el0,int at,double t,double T, double phiC , double Bclass , int step) 
{

 double (*sr1)[3] = g_system.r1 ;
 double (*sv1)[3] = g_system.v1 ;
 double *smass    = g_system.m ;
 int    *sqi      = g_system.qi ;
 int    *sqi_corr = g_system.qi_corr ;
 char   *sptype   = g_system.ptype ;
 long long int *suid = g_system.uid ;
 struct ATOMS_STRUCT *atomic_data = g_atomic_data ;
 char *instrecOK = g_inputs.instrecOK ;

 int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = at ;

 double rel[3],vtr[3],vat[3],vel[3],e[3],v_e,v_a,Vi,Vf,Ei,Ef, B ; //, Vo = 0  ;
 double x,ori_r1[3],ori_v1[3],ori_r2[3],ori_v2[3] , ori_v1_el0[3] ;
 char ori_etype;
 int i , ori_q , ori_qcorr , sh , hole ;
 char cfgstr_i[LSTRLTH] = "-" ,  cfgstr_f[LSTRLTH] = "-" ;
 double T_period , t_capture ;
 double KE_extra  , KE_el_i , KE_at_i , KE_at_f , KE_el2_i , KE_el2_f , KE_elx_i , KE_elx_f ;

 long long int el_uid = g_system.uid[el0] ;

 static long long int *log_skipped_ions = NULL , log_skipped_size = 0 ;

 //array_max( &log_skipped_size , suid , g_inputs.N + g_inputs.E , "lli" ) ;
 if ( log_skipped_size < el_uid + 1 )
 {
     long long int new_log_skipped_size = el_uid + 1 ; int iii ;
     log_skipped_ions = realloc( log_skipped_ions , ((int) new_log_skipped_size) * sizeof(long long int) ) ;
     for ( iii = log_skipped_size ; iii < new_log_skipped_size ; iii++ ) { log_skipped_ions[iii] = -1 ; }
     log_skipped_size = new_log_skipped_size ;
 }
 

 KE_at_i = KE_at_f = KE_el2_i = KE_el2_f = KE_elx_i = KE_elx_f = 0 ;

  int nbor ;
  int el2 = -1 , recomb_e_partner = -1 ; 
  double v_partner = 0 , dist_partner = 0 ;


 sh = highest_unoccupied_shell( at ) ;
 hole = get_hole_on_shell( at , sh ) ; // getrand_shell_hole( at , sh ) ;
  
 struct g_struct_xatom_data *old_atomic_econfig = atomic_data[at].atomic_econfig ;


#ifdef DBG_ENERGY
  DBG_WARNING_NOTE_ON_CODE( "DBG. Energy ref.calc. To Be Removed." ) ;
  double dbg_Vbef = xmdyn_md_energy( "system-V" , "calc" ), dbg_Tbef = xmdyn_md_energy( "system-K" , "calc" ), dbg_Ebef = dbg_Tbef + dbg_Vbef ;
#endif  
 xmdyn_store_restore_subsystem( "store" , "-" , 3 , at , el0 , hole ) ;

 t_capture = l_instant_recomb.time[el0] ;
 T_period  = l_instant_recomb.Tall[el0]  ;

 //**  Initial kinetic properties  **/

 v_e  = sqrt(vectnorm2( sv1 [el0]));
 v_a  = sqrt(vectnorm2( sv1 [at]));

 { Vi = potEactual(at,el0,-1,num_atoms_affected,atoms_affected); }
 KE_el_i = 0.5 * g_phys_math_const.me * v_e * v_e ;
 KE_at_i = 0.5 * smass[at] * v_a * v_a ;
 KE_at_i = 0.5 * smass[at] * VECTOR3_NORM2( sv1[at] ) ;
 Ei = KE_at_i + KE_el_i + Vi ;
 //KE_e_rec  = 0.5* g_phys_math_const.me *v_e*v_e ; 

 // NOTE: Update B in exec_recomb
 B  =  g_atomic_data[at].atomic_econfig->B_rec ;

 get_xatom_conf_str( atomic_data[at].atomic_econfig->max_orbit , 
                     atomic_data[at].atomic_econfig->NE_orbit , cfgstr_i, 'L'  ) ;
 ori_q = sqi[at] ;
 ori_qcorr = sqi_corr[at] ;
 VECTOR3_EQUAL( ori_v1_el0 , sv1[el0] ) ;

 char re_id[SSTRLTH] = "=re:" ;
 int electron_remains = 1 ;
 char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;
 char *logstr2 ; logstr2 = calloc(2,sizeof(char) ) ; strcpy( logstr2 , "" ) ;


 //** if we are not in log-only mode: **/
 if ( strstr( instrecOK , "y" ) != NULL )
 {
     strcpy( re_id , "+re:" ) ;
     electron_remains = 0 ;
//     xmdyn_classical_check_ionization() ;
     particle_params_swap(el0,hole);
     sptype[hole] = subshellint2char(sh) ;
     particle_rvq_zero(hole) ;
     atomic_data[at].shell[sh].NE++ ;
     atomic_data[at].NE++ ;
     // g_system.qi [at]-= g_inputs.dqatomrec;  // TODO
     sqi [at] -= 1 ;
     //if ( el2 == hole ) { el2 = el ; } // because of the swap above!!
  }


  xmdyn_update_atoms( &logstr , "array,full" , num_atoms_affected , atoms_affected ) ; // xmdyn_update_atoms( &logstr , "list,full" , 1 , at ) ; // xmdyn_update_atoms( &logstr , 1 , at );

  get_xatom_conf_str( atomic_data[at].atomic_econfig->max_orbit , 
                      atomic_data[at].atomic_econfig->NE_orbit , cfgstr_f, 'L'  ) ;

  { 
      Vf = potEactual( at , (1 - electron_remains) * (-1) + electron_remains * el0 , -1 , num_atoms_affected , atoms_affected ) ; 
      xmdyn_md_energy( "system-V" , "usecache" ) ; //X calc_tot_init_ener(212);
  }

  KE_extra = KE_el_i + Vi - ( Vf + (-B) ) ; 

  double aratio = 0 ;
  int nbor_id = 0 ;
  Ef = KE_el2_i = KE_el2_f = 0 ;

//  KE_extra -= KE_at_i ; //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  //if ( opt_energy_to_partner )
  if ( strstr( instrecOK , "y" ) != NULL )
  {
     double v0_at[3] , v0_el[3] , v0_el2[3] ;
     VECTOR3_EQUAL( v0_el , ori_v1_el0 ) ;

     //* TODO: store_system `add` to be added *//
    
     el2 = 0 ; int found = -1 ;

//******************************************
     double K_add , K_add0 , K_todo, v_add[3] , v_add0[3] , v_todo[3] ;
     int curr_iter , max_iter ;
     double dmin = -1 ;
     K_add = KE_extra ; VECTOR3_EQUAL( v_add , v0_el ) ; 

     //while ( ( -1 != el2 ) && ( -1 == found ) )
    #ifdef MODULE_DEBUG
    {
                 FILE *fp = fopen( "./recomb-debug-event.txt" , "a" ) ;
                 fprintf( fp , "RECOMB: %d\n" , hole ) ;
                 fclose(fp);
    }
    #endif
     while ( ( -1 != el2 ) && ( K_add != 0 ) )
     {
         get_nearest_electron( at , hole , &dmin , &el2 ) ; nbor_id++ ;
         recomb_e_partner = el2 ;
         double xfactor = 1 ;
         get_opt( g_inputs.irec_misc , "reduceKE" , "double" , &xfactor ) ;

         if ( 0 <= el2 )
         {
             KE_elx_i = 0.5 * smass[el2] * VECTOR3_NORM2( sv1[el2] ) ;
             KE_el2_i += KE_elx_i ; // 0.5 * smass[el2] * VECTOR3_NORM2( sv1[el2] ) ;
             K_add0 = K_add ; VECTOR3_EQUAL( v_add0 , v_add ) ;
             K_todo = 0 ; v_todo[0] = v_todo[1] = v_todo[2] = 0 ; 
             curr_iter = 0 ; max_iter = 10 ; 
             while ( ( K_add != 0 ) && ( curr_iter < max_iter ) )
             {
                 VECTOR3_EQUAL( v0_at  , sv1[at]  ) ;
                 VECTOR3_EQUAL( v0_el2 , sv1[el2] ) ;
                 found = -1 ;
                 found = xmdyn_3Brecomb_add_pE_to_el( smass[at] , smass[el2] , 
                              v0_at , v_add , sv1[at] , sv1[el2] , K_add , &aratio ) ;
                 if (  
                       ( get_opt( g_inputs.irec_misc , "skip-re" , "" , NULL ) == 0 ) ||
                       ( ( get_opt( g_inputs.irec_misc , "posKEdistr" , "" , NULL ) == 0 ) && 
                         ( K_add < 0 ) ) ||
                       ( ( get_opt( g_inputs.irec_misc , "posE" , "" , NULL ) == 0 ) &&
                         ( K_add + l_instant_recomb.Etot[el2] < 0 ) ) ||
                       ( xfactor * (0.5*smass[el2]*VECTOR3_NORM2(v0_el2)) < -K_add )
                    )
                 { found = -1 ; }
    #ifdef MODULE_DEBUG
                 FILE *fp = fopen( "./recomb-debug-event.txt" , "a" ) ;
                 int sstep = 0 ;
                 fprintf(fp,"S: %3d SS: %3d el2: %6d iter: %3d KEp: % le KEp+Etot: % le Etot: % le\n", g_inputs.step , sstep , el2 , curr_iter , convert_unit_J2eV(K_add) , convert_unit_J2eV(K_add + l_instant_recomb.Etot[el2]) , convert_unit_J2eV(l_instant_recomb.Etot[el2]) ) ;
                 fclose(fp) ;
    #endif
                  if ( -1 == found ) 
                  {
                      VECTOR3_EQUAL( sv1[at]  , v0_at  ) ; VECTOR3_EQUAL( sv1[el2] , v0_el2 ) ;
                  }

                  if ( ( 0 == curr_iter ) && ( -1 != found ) )
                  {
                      K_add = v_add[0] = v_add[1] = v_add[2] = 0 ;
                  }

                  curr_iter++ ;
                  if (  curr_iter == max_iter )
                  {
                      if ( -1 != found )
                      {
                          K_add  = K_todo ; VECTOR3_EQUAL( v_add , v_todo ) ;
                      }
                      else
                      {
                          K_add += K_todo ; VECTOR3_ADD( v_add , v_add , v_todo ) ;
                      }
                  }
                  else
                  {
                      K_add *= 0.5 ; VECTOR3_TIMES_SCALAR( v_add , v_add , 0.5 ) ;
                      if ( -1 != found ) 
                      {
                          K_todo -= K_add ; VECTOR3_DIFF( v_todo , v_todo , v_add ) ;
                      }
                      else 
                      {
                          K_todo += K_add ; VECTOR3_ADD( v_todo , v_todo , v_add ) ;
                      }
                  }
             }
             KE_elx_f = 0.5 * smass[el2] * VECTOR3_NORM2( sv1[el2] ) ;
             KE_el2_f += KE_elx_f ; // 0.5 * smass[el2] * VECTOR3_NORM2( sv1[el2] ) ;
             
             if ( 0 != KE_elx_f - KE_elx_i ) 
             {
                  char xxxstr[32];
                  KE_elx_f -= KE_elx_i ; KE_elx_f = convert_unit_J2eV( KE_elx_f ) ; 
                  sprintf(xxxstr," dKE2: %le", KE_elx_f) ; logstr2 = str_concat_realloc( logstr2 , xxxstr ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
                  sprintf(xxxstr," el2: %d", el2) ; logstr2 = str_concat_realloc( logstr2 , xxxstr ) ;
#endif // XMDYN_XRAYPAC_MORELOG
                  sprintf(xxxstr," uid2: %lld", g_system.uid[el2]); logstr2 = str_concat_realloc( logstr2 , xxxstr ) ;
                  sprintf(xxxstr," dist: %le", dmin) ; logstr2 = str_concat_realloc( logstr2 , xxxstr ) ;
              }

         }
     }
//******************************************

      if ( -1 == el2 ) 
      {
          KE_at_f = 0.5 * smass[at] * VECTOR3_NORM2( sv1[at] ) ;
          
          strcpy( re_id , "-re:" ) ;
          if ( suid[at] == log_skipped_ions[el_uid] ) { strcpy( re_id , "Nre:" ) ; strcpy(logstr,"") ; }
          log_skipped_ions[el_uid] = suid[at] ;
      }
      else
      {
          l_instant_recomb.time[hole] = 0 ;
	  l_instant_recomb.Tall[hole] = 0 ;
          l_instant_recomb.phi_cum[hole] = 0 ;
          l_instant_recomb.Ekin[hole] = 0 ;
          l_instant_recomb.Epot[hole] = 0 ;
          l_instant_recomb.Etot[hole] = 0 ;
          l_instant_recomb.Etot0[hole] = 0 ;

          KE_at_f = 0.5 * smass[at] * VECTOR3_NORM2( sv1[at] ) ;
          Ef += 0.5* smass[at]*VECTOR3_NORM2(sv1[at])  +  
                electron_remains * 0.5 * g_phys_math_const.me * VECTOR3_NORM2(sv1[el0]) +  Vf ;
          Ei += KE_el2_i ;
          Ef += KE_el2_f ;
    
          g_ENERGY_VALUES.Eintotal += ( Ef - Ei ) ; 
          g_ENERGY_VALUES.Eintotal_th += B ;
          g_ENERGY_VALUES.Einrec += ( Ef - Ei ) ;
          g_ENERGY_VALUES.Ebound += - B ;
      }
  }

 
  if ( ( strcmp( re_id , "Nre:" ) != 0 ) || ( get_opt( g_inputs.irec_misc , "logNre" , "" , NULL ) == 0 ) )
  {
       double xxx ;
       print_log( "" , (void*) re_id , "str" , "noflush" ) ;
       print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
       print_log( " atom:" , (void*) &at  , "int" , "noflush" ) ;
       print_log( " Z:" , (void*) &(g_system.Z[at]) , "int" , "noflush" ) ;
       print_log( " q:" , (void*) &ori_q , "int" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
       print_log( " qcorr:" , (void*) &ori_qcorr , "int" , "noflush" ) ;
       print_log( " shell:" , (void*) &sh , "int" , "noflush" ) ;
       print_log( " el:" , (void*) &el0 , "int" , "noflush" ) ;
       print_log( " hole:" , (void*) &hole , "int" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
       print_log( " uid:" , (void*) &el_uid , "llint" , "noflush" ) ;
       //if ( opt_energy_to_partner ) 
       if ( strstr( instrecOK , "y" ) != NULL )
       {
       }
       B = convert_unit_J2eV( B ) ;
       print_log( " Bqm:" , (void*) &B , "double" ,  "noflush" ) ;
       Bclass = convert_unit_J2eV( Bclass ) ;
       print_log( " Bcl:" , (void*) &Bclass , "double" ,  "noflush" ) ;
       xxx = convert_unit_J2eV( Vf -Vi ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
       print_log( " dV:" , (void*) &xxx , "double" ,  "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
       xxx = phiC / (2.*PI) ;
       print_log( " perC:" , (void*) &xxx , "double" ,  "noflush" ) ;
       xxx = T / MESSAGES.unit_t ;
#ifdef XMDYN_XRAYPAC_MORELOG
       print_log( " T:" , (void*) &xxx , "double" ,  "noflush" ) ;
       xxx = t / MESSAGES.unit_t ;
       print_log( " t:" , (void*) &xxx , "double" ,  "noflush" ) ;
       xxx = t / T  ;
       print_log( " perT:" , (void*) &xxx , "double" ,  "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
       print_log_full( " cfg_i:" , (void*)cfgstr_i , "str" , "noflush" ) ;
       print_log_full( " cfg_f:" , (void*)cfgstr_f , "str" , "noflush" ) ;
       KE_at_f -= KE_at_i ; KE_at_f = convert_unit_J2eV( KE_at_f ) ;
       print_log( " dKEat:" , (void*) &KE_at_f , "double" , "noflush" ) ;
       print_log( logstr2 , NULL , "" , "noflush" ) ; 
       print_log( " \n" , NULL , "" , "noflush" ) ;
       print_log( logstr , NULL , "" , "noflush" ) ; 
       print_log( "" , NULL , "" , "flush" ) ;
   }

#ifdef DBG_ENERGY
  DBG_WARNING_NOTE_ON_CODE( "DBG. Energy ref.calc. To Be Removed." ) ;
  double dbg_Vaft = xmdyn_md_energy( "system-V" , "calc" ) , dbg_Taft = xmdyn_md_energy( "system-K" , "calc" ), dbg_Eaft = dbg_Taft + dbg_Vaft ;
  FILE *fp=fopen( "./dbg-Egain_Rec.txt" , "a" ); fprintf( fp, "%1.12le %1.12le % 1.6le % 1.6le % 6d % 6d % 6d % 14s % 14s\n",  
  (dbg_Eaft-dbg_Ebef)/g_phys_math_const.e1  ,  
  B/g_phys_math_const.e1  ,  
  old_atomic_econfig->B_rec/ g_phys_math_const.e1  ,  
  old_atomic_econfig->Ecrit_cl2q/g_phys_math_const.e1  ,  
  step , at , el , cfgstr_i , cfgstr_f ), fclose(fp);
#endif  

 
  if ( -1 == el2 ) { xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ; }
  else             { xmdyn_store_restore_subsystem( "clear" , "-" , 0 ) ; }
  free( logstr ) ;
  free( logstr2 ) ;
  free( atoms_affected ) ;

  return 0 ;
}




//############################## FUNCTION #################################

int instant_recombination_init( void )
{
    int ii ;
    l_instant_recomb.nnatoms = calloc( g_inputs.N+ g_inputs.E,sizeof(int)) ;
    l_instant_recomb.time = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.Tall = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.Ekin = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.Epot = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.Etot = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.Etot0 = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.phi = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    l_instant_recomb.phi_cum = calloc( g_inputs.N+ g_inputs.E,sizeof(double));
    for ( ii = 0 ; ii < g_inputs.N + g_inputs.E ; ii++ )
    {
        l_instant_recomb.nnatoms[ii] = -1 ;
    }
    l_instant_recomb.num_e_on_atom = calloc( g_inputs.N , sizeof(int) ) ;
    return(0);
}



//############################## FUNCTION #################################

int instant_recombination_final( void )
{
    free(l_instant_recomb.nnatoms);
    free( l_instant_recomb.time ) ;
    free( l_instant_recomb.Tall ) ;
    free( l_instant_recomb.Ekin ) ;
    free( l_instant_recomb.Epot ) ;
    free( l_instant_recomb.Etot ) ;
    free( l_instant_recomb.Etot0 ) ;
    free( l_instant_recomb.num_e_on_atom ) ;
    return 0 ;
}





//############################## FUNCTION #################################

int instant_recombination_swap( int at , int el )
{
    double x ;
    int i ;

    x = l_instant_recomb.time[at];
    l_instant_recomb.time[at] = l_instant_recomb.time[el];
    l_instant_recomb.time[el] = x ;

    x = l_instant_recomb.Tall[at] ;
    l_instant_recomb.Tall[at] = l_instant_recomb.Tall[el];
    l_instant_recomb.Tall[el] = x;

    i = l_instant_recomb.nnatoms [at] ;
    l_instant_recomb.nnatoms[at] = l_instant_recomb.nnatoms[el] ;
    l_instant_recomb.nnatoms[el] = i ;

    x = l_instant_recomb.Ekin[at] ;
    l_instant_recomb.Ekin[at] = l_instant_recomb.Ekin[el];
    l_instant_recomb.Ekin[el] = x;

    x = l_instant_recomb.Epot[at] ;
    l_instant_recomb.Epot[at] = l_instant_recomb.Epot[el];
    l_instant_recomb.Epot[el] = x;

    x = l_instant_recomb.Etot[at] ;
    l_instant_recomb.Etot[at] = l_instant_recomb.Etot[el];
    l_instant_recomb.Etot[el] = x;

    x = l_instant_recomb.Etot0[at] ;
    l_instant_recomb.Etot0[at] = l_instant_recomb.Etot0[el];
    l_instant_recomb.Etot0[el] = x;

    x = l_instant_recomb.phi[at] ;
    l_instant_recomb.phi[at] = l_instant_recomb.phi[el];
    l_instant_recomb.phi[el] = x;

    x = l_instant_recomb.phi_cum[at] ;
    l_instant_recomb.phi_cum[at] = l_instant_recomb.phi_cum[el];
    l_instant_recomb.phi_cum[el] = x;

    return(0);
 }






//############################## FUNCTION #################################

int xmdyn_free_el_for_recomb( int **free_el , int **nnatoms, double **mindists , int *numel ) 
{
    int NN = g_inputs.N ;
    int EE = g_inputs.E ;
    int ii , at , ee ;
    double x ;

    xmdyn_get_classical_electrons( numel , free_el ) ; 

    if ( 0 < *numel )
    {
        *nnatoms  = calloc( *numel,sizeof(int));
        *mindists = calloc( *numel,sizeof(double));
    }

    for ( ii = 0 ; ii < *numel ; ++ii ) {
      ee = (*free_el)[ii];
      (*nnatoms)[ii] = 0 ; // from_atom ;
      (*mindists)[ii] = nnbor_condition( ee , 0 , g_inputs.ireccond ) ;
    }
  
    for ( ii = 0 ; ( ii < *numel ) && ( *numel > 0) ; ++ii ) {
      ee = (*free_el)[ii] ;
      //counter=0;
      if (1) {
        for ( at = 0 ; at < NN ; ++at ) {
	  x = nnbor_condition( ee, at , g_inputs.ireccond ) ;
	  if ( x < (*mindists)[ii]) {
	    (*mindists)[ii] = x ; 
	    (*nnatoms)[ii] = at ;
	  }
        }
      }
    }

    return 0 ;
}
      
// #####################################################################

#ifndef COULOMB_ION_E_POT
#define COULOMB_ION_E_POT( q1 , q2 , r ) ( enerreg( r , q1 , q2 ) )
#endif

//############################## FUNCTION #################################

double get_max_barrier_2ions( int at1 , int at2 ) 
{
    int ngrid = 101 ;
    int q1 = g_system.qi[at1] ;
    int q2 = g_system.qi[at2] ;
    double dvect[3] ; VECTOR3_DIFF( dvect , g_system.r1[at1] , g_system.r1[at2] ) ;
    double dist = VECTOR3_NORM( dvect ) ;
    double dr = dist / ( ngrid - 1.0 ) ;

    int ii , ii_peak ;
    double potpeak , xpot ;

    ii_peak = 1 ; potpeak = COULOMB_ION_E_POT( q1 , -1 , 1*dr ) + COULOMB_ION_E_POT( q2 , -1 , dist * 1*dr ) ;
    for ( ii = 1 ; ii < ngrid - 1 ; ii++ )
    {
        xpot = COULOMB_ION_E_POT( q1 , -1, ii*dr ) + COULOMB_ION_E_POT( q2 , -1 , dist - ii*dr ) ;
        if ( potpeak < xpot ) { potpeak = xpot ; ii_peak = ii ; }
    }
    return( potpeak ) ;
}


//############################## FUNCTION #################################

int  irec_find_nondeloc( int *free_el , int *numel , int *nnatoms1, int *nnatoms2 , double *mindists1 , double *mindists2 ) 
{
    int ii , cc , el , yyi  , swap;
    double kinE , potE , yyd ;

    cc = 0 ; ii = 0 ;
    while ( cc < *numel )
    {
        el = free_el[ii] ;
        kinE = 0.5 *g_system.m[el] * VECTOR3_NORM2( g_system.v1[el] ) ;
        potE = get_max_barrier_2ions( nnatoms1[ii] , nnatoms2[ii] ) ;
        if ( kinE + potE < 0 ) 
        {
            ii++ ;
        }
        else
        {
            swap = *numel - (cc-ii) - 1 ;
            yyi = free_el[ii]   ; free_el[ii]   = free_el[swap]   ; free_el[swap]  = yyi ;
            yyi = nnatoms1[ii]  ; nnatoms1[ii]  = nnatoms1[swap]  ; nnatoms1[swap] = yyi ;
            yyi = nnatoms2[ii]  ; nnatoms2[ii]  = nnatoms2[swap]  ; nnatoms2[swap] = yyi ;
            yyd = mindists1[ii] ; mindists1[ii] = mindists1[swap] ; mindists1[swap] = yyd ;
            yyd = mindists2[ii] ; mindists2[ii] = mindists2[swap] ; mindists2[swap] = yyd ;
        }
        cc++ ;
    }
    *numel = ii ;
    return 0 ;
}


//############################## FUNCTION #################################

static int process_nonloc_electrons( int *el_loc , int numloc , char *logid )
{
    int Na = g_inputs.N ;
    int Ne = g_inputs.E ;
    int step = g_inputs.step ;

    int ii ;

    char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;
    int *tab_isloc = calloc( Na+Ne , sizeof(int) ) ;

    for ( ii = 0 ; ii < numloc ; ii++ ) {   tab_isloc[ el_loc[ii] ] = 1 ;   }

    for ( ii = Na ; ii < Na+Ne ; ii++ )
    {
        if ( ! tab_isloc[ii] )
        {
            if ( 0 <= l_instant_recomb.nnatoms[ii] ) 
            {
                int el = ii ;
                //char idstr[] = "re" ;
                char idstr[16] = "" ;
                double xxx ;
                if  ( l_instant_recomb.Tall[el] < 0  )  
                { 
                    strcpy( idstr , "rre:" ) ; 
                    l_instant_recomb.num_e_on_atom[l_instant_recomb.nnatoms[ii]]-- ; 
                }
                else
                {
                    strcat( strcat( idstr , logid )  , "re:" ) ;
                }
                //print_log( "" , (void*) logid , "str" , "noflush" ) ; 
#ifdef XMDYN_XRAYPAC_MORELOG
                print_log( "" , (void*) idstr , "str" , "noflush" ) ; 
                print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
                print_log( " atom:" , (void*) &(l_instant_recomb.nnatoms[el])  , "int" , "noflush" ) ;
                print_log( " Z:" , (void*) &(g_system.Z[l_instant_recomb.nnatoms[el]]) , "int" , "noflush" ) ;
                print_log( " q:" , (void*) &g_system.qi[l_instant_recomb.nnatoms[el]] , "int" , "noflush" ) ;
                print_log( " qcorr:" , (void*) &g_system.qi_corr[l_instant_recomb.nnatoms[el]] , "int" , "noflush" ) ;
                print_log( " el:" , (void*) &el , "int" , "noflush" ) ;
                print_log( " uid:" , (void*) (g_system.uid + el) , "llint" , "noflush" ) ;
                xxx = l_instant_recomb.Epot[el] / MESSAGES.unit_E ;
                print_log( " Epot0:" , (void*) &xxx , "double" ,  "noflush" ) ;
                xxx = l_instant_recomb.Etot[el] / MESSAGES.unit_E ;
                print_log( " Etot0:" , (void*) &xxx , "double" ,  "noflush" ) ;
                xxx = l_instant_recomb.Tall[el] / MESSAGES.unit_t ;
                print_log( " T:" , (void*) &xxx , "double" ,  "noflush" ) ;
                xxx = l_instant_recomb.time[el] / MESSAGES.unit_t ;
                print_log( " t:" , (void*) &xxx , "double" ,  "noflush" ) ;
                if    ( 0 == l_instant_recomb.Tall[el]  )    { xxx = -1 ; }
                else    {  xxx = l_instant_recomb.time[el] /  l_instant_recomb.Tall[el]  ; }
                print_log( " perT:" , (void*) &xxx , "double" ,  "noflush" ) ;
                xxx = l_instant_recomb.phi_cum[el] / (2.*PI) ;
                print_log( " phiC:" , (void*) &xxx , "double" ,  "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
                xmdyn_update_atoms( &logstr , "list,full" , 1 , l_instant_recomb.nnatoms[el] ) ; // xmdyn_update_atoms( &logstr , 1 , l_instant_recomb.nnatoms[el] );
#ifdef XMDYN_XRAYPAC_MORELOG
                print_log( " \n" , NULL , "" , "noflush" ) ; 
                print_log( logstr , NULL , "" , "noflush" ) ; 
                print_log( "" , NULL , "" , "flush" ) ; 
#endif // XMDYN_XRAYPAC_MORELOG
                l_instant_recomb.nnatoms[ii] = -1 ; 
                l_instant_recomb.time[ii] = 0 ;
                l_instant_recomb.Tall[ii] = 0 ;
                l_instant_recomb.Ekin[ii] = 0 ;
                l_instant_recomb.Epot[ii] = 0 ;
                l_instant_recomb.Etot[ii] = 0 ;
                l_instant_recomb.Etot0[ii] = 0 ;
                l_instant_recomb.phi[ii] = -1 ; 
                l_instant_recomb.phi_cum[ii] = 0 ;
           }
        }
    }      
    free(logstr) ; free(tab_isloc);
    return 0 ;
}


//############################## FUNCTION #################################

int instant_recombination( int step ) {
  int *free_el,numel, num_loc_el , *nnatoms1 , *nnatoms2 ;
  int el , at , ii ,from_atom,to_atom,from_el,to_el;
  double *mindists1, *mindists2, x, t,T,phiC, Bclass ;
  
  if ( ( strstr( g_inputs.instrecOK , "y" ) == NULL ) && ( strstr( g_inputs.instrecOK , "F" ) == NULL )  ) 
  {      return 0 ; }


#if 1
    if ( step < 0 ) 
    {
      for ( el = g_inputs.N ; el < g_inputs.N + g_inputs.E ; ++el ) {
        if ( isitfreeelectron( el ) ) {
          try_instrec( el , 0 , 'f' , step , &t , &T , &phiC , &Bclass ) ;
	}
      }
      return(0);
    }
    // ############
#endif


    if ( (step+1) % g_inputs.irecDT != 0 )  { return 0 ; }
      

 if ( 1 )
 {
    insidecluster(-1,0);  
    free_el=get_free_electrons(&numel, g_inputs.maxrecdist);
    get_secel_limits(&from_atom,&to_atom,&from_el,&to_el,numel);

    if (numel>0) 
    { 
        nnatoms1 = calloc( numel , sizeof(int));
        mindists1 = calloc( numel , sizeof(double));
        nnatoms2 = calloc( numel , sizeof(int));
        mindists2 = calloc( numel , sizeof(double));
    }

    for ( ii = from_el ; ii < to_el ; ++ii ) 
    {
      el = free_el[ii];
      int nn0 , nn1 ;
      double xx0 , xx1 ;
      nn0 = from_atom ;
      if ( from_atom < to_atom ) { nn1 = from_atom + 1 ; }
      else { nn1 = nn0 ; }
      xx0 = nnbor_condition( el , nn0 , g_inputs.ireccond ) ;
      xx1 = nnbor_condition( el , nn1 , g_inputs.ireccond ) ;
      if ( xx0 < xx1 )
      {
          nnatoms1[ii] = nn0 ; mindists1[ii] = xx0 ;
          nnatoms2[ii] = nn1 ; mindists2[ii] = xx1 ;
      }
      else
      {
          nnatoms1[ii] = nn1 ; mindists1[ii] = xx1 ;
          nnatoms2[ii] = nn0 ; mindists2[ii] = xx0 ;
      }
    }
  
    for ( ii = from_el ; ( ii < to_el ) && ( numel > 0 ) ; ++ii ) {
      el = free_el[ii];
      //counter=0;
      if (1) {
        for ( at = from_atom ; at < to_atom ; ++at ) 
        {
	  x = nnbor_condition( el , at , g_inputs.ireccond ) ;
	  if ( x < mindists1[ii] ) {
              mindists2[ii] = mindists1[ii] ;  nnatoms2[ii] = nnatoms1[ii] ; mindists1[ii] = x ; nnatoms1[ii] = at ;
	  }
          else if  ( x < mindists2[ii] ) {
              mindists2[ii] = x ; nnatoms2[ii] = at ;
          }
        }
      }
    }
  }

  else 
  { 
      xmdyn_free_el_for_recomb( &free_el , &nnatoms1, &mindists1 , &numel ) ;
  } 

  
    process_nonloc_electrons( free_el , numel , "f" ) ;
  {
    num_loc_el = numel ;
    irec_find_nondeloc( free_el , &num_loc_el , nnatoms1, nnatoms2 , mindists1 , mindists2 ) ;
//    printf("DBG:LOC % 6d %d %d \n" , g_inputs.step , numel , num_loc_el );
    numel = num_loc_el ;
    process_nonloc_electrons( free_el , numel , "d" ) ;
  }

  int My_Rank = 0 , Event_Rank = 0 ;
    if (My_Rank==Event_Rank) {

      if ( 0 )
      {
          for ( ii = 0 ; ii < numel ; ++ii ) {
            el = free_el[ii];
            try_instrec( el , nnatoms1[ii] , 't' , step , &t , &T , &phiC , &Bclass ) ;
          }
      }

      for ( ii = 0 ; ii < numel ; ++ii ) {
        el = free_el[ii] ;
        if ( try_instrec( el , nnatoms1[ii] , 'd' , step , &t , &T , &phiC , &Bclass ) ) {
          do_instrec( el , nnatoms1[ii] , t , T , phiC , Bclass , step ) ;
        }
      }

    }  
   
    if (numel>0) { free(nnatoms1); free(mindists1);  free(nnatoms2); free(mindists2);  free(free_el); }

  return 0 ;
}





// #####################################################################
// #####################################################################
// #####################################################################




//############################## FUNCTION #################################

double electronbackstepping(int actatom,int actelectron){
  int i,N,E;
  double oldr[3],newr[3],relold[3],relnew[3],Epot0,Epotnew;
  double temp1;

  N= g_inputs.N; E= g_inputs.E;
  Epot0=Epotnew=0;
  vect3equal(oldr, g_system.r1 [actelectron]);
  vect3equal(newr, g_system.r2 [actelectron]);
  for (i=0;i<N+E;++i)
    if ( (i!=actelectron) )  { 
      vectminus(relold, g_system.r1 [i],oldr);
      vectminus(relnew, g_system.r1 [i],newr);
      if (i<N){
	Epot0+=enerregcut(sqrt(vectnorm2(relold)), g_system.qi [i], g_system.qi [actelectron],actatom);
	Epotnew+=enerregcut(sqrt(vectnorm2(relnew)), g_system.qi [i], g_system.qi [actelectron],actatom);
      }
      else{
	Epot0+=enerreg(sqrt(vectnorm2(relold)), g_system.qi [i], g_system.qi [actelectron]);
	Epotnew+=enerreg(sqrt(vectnorm2(relnew)), g_system.qi [i], g_system.qi [actelectron]);
      }
    }
  vect3equal( g_system.r1 [actelectron],newr);
  vect3equal( g_system.r2 [actelectron],oldr);

  temp1=Epot0-Epotnew;
  return(Epot0-Epotnew);
}





//############################## FUNCTION #################################

double thetascatt_old(void){
  return(0);
}




//############################## FUNCTION #################################

int elasticscattering(int actatom,int actelectron){
  double phi,sinf,cosf,theta,sint,cost,Ekinplus=0;
  double Mn,px,py,pz,vv,VV;
  double vel[3],veltr[3],rel[3]; //,vel1[3],vel1lab[3],vat1[3],vat1lab[3];
  double zaxes[3],reltr[3],vtr[3],vattr[3],tempvect[3];
  double En,cosT,sinT,cosF,sinF;
  char backstepindicator[SSTRLTH];
  double me =  g_phys_math_const.me ;

  Mn= g_system.m[actatom];

  vect3equal(zaxes, g_system.v1 [actelectron]);
  vectminus(vel, g_system.v1 [actelectron], g_system.v1 [actatom]);
  vectminus(rel, g_system.r1 [actelectron], g_system.r1 [actatom]);
  coordtrafo(reltr,rel,zaxes,rel,'t');
  coordtrafo(vtr, g_system.v1 [actelectron],zaxes,rel,'t');
  coordtrafo(vattr, g_system.v1 [actatom],zaxes,rel,'t');

  phi=(randnum(0)-.5)*PI;
  sinf=sin(phi); cosf=cos(phi);

  theta=thetascatt_old();
  sint=sin(theta); cost=cos(theta);

  strcpy(backstepindicator,"n");
  if (theta>PI/2){
    strcpy(backstepindicator,"y");
    Ekinplus+=electronbackstepping(actatom,actelectron);
  }

 px=Mn*vattr[0]+ g_phys_math_const.me *vtr[0];
 py=Mn*vattr[1]+ g_phys_math_const.me *vtr[1];
 pz=Mn*vattr[2]+ g_phys_math_const.me *vtr[2];
 En=.5*Mn*vectnorm2( g_system.v1 [actatom])+.5*me*vectnorm2( g_system.v1 [actelectron])+Ekinplus;

 if ( (-me*pow(Mn,2.0)*(2.0*Mn*pow(pz,3.0)*px*sin(theta)*cos(theta)-2.0*me*pow(px,3.0)*sin(theta)*cos(theta)*pz+2.0*pow(pz,3.0)*me*px*sin(theta)*cos(theta)+2.0*Mn*pow(py,2.0)*px*sin(theta)*cos(theta)*pz-4.0*Mn*me*px*sin(theta)*cos(theta)*pz*En-4.0*En*px*sin(theta)*cos(theta)*pz*pow(Mn,2.0)+2.0*me*px*sin(theta)*cos(theta)*pz*pow(py,2.0)+2.0*Mn*pow(px,3.0)*sin(theta)*cos(theta)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2.0*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2.0*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(theta),3.0)*pz*sin(theta)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(theta),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(px,4.0)*pow(cos(theta),2.0)+2.0*Mn*En*me*pow(px,2.0)*pow(cos(theta),2.0)-4.0*me*px*pow(cos(theta),3.0)*pow(pz,3.0)*sin(theta)-2.0*Mn*pow(pz,2.0)*En*me*pow(cos(theta),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),4.0)+Mn*pow(pz,4.0)*pow(cos(theta),2.0)-me*pow(px,4.0)*pow(cos(theta),4.0)-pow(pz,4.0)*me*pow(cos(theta),4.0)+pow(pz,4.0)*me*pow(cos(theta),2.0)+me*pow(px,4.0)*pow(cos(theta),2.0)-2.0*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(theta),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(theta),2.0)+2.0*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(theta),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(theta),2.0))) <0 ){
   if ( (strcmp(backstepindicator,"y")==0) ){
     vect3equal(tempvect, g_system.r1 [actelectron]);
     vect3equal( g_system.r1 [actelectron], g_system.r2 [actelectron]);
     vect3equal( g_system.r2 [actelectron],tempvect);
   }
   return(-1);
 }


 vv=-(-Mn*pow(pz,2.0)*me*pow(cos(theta),2.0)-Mn*me*pow(px,2.0)+Mn*me*pow(px,2.0)*pow(cos(theta),2.0)-2.0*Mn*me*px*sin(theta)*cos(theta)*pz-
	sqrt(-me*pow(Mn,2.0)*(2.0*Mn*pow(pz,3.0)*px*sin(theta)*cos(theta)-2.0*me*pow(px,3.0)*sin(theta)*cos(theta)*pz+2.0*pow(pz,3.0)*me*px*sin(theta)*cos(theta)+2.0*Mn*pow(py,2.0)*px*sin(theta)*cos(theta)*pz-4.0*Mn*me*px*sin(theta)*cos(theta)*pz*En-4.0*En*px*sin(theta)*cos(theta)*pz*pow(Mn,2.0)+2.0*me*px*sin(theta)*cos(theta)*pz*pow(py,2.0)+2.0*Mn*pow(px,3.0)*sin(theta)*cos(theta)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2.0*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2.0*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(theta),3.0)*pz*sin(theta)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(theta),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(px,4.0)*pow(cos(theta),2.0)+2.0*Mn*En*me*pow(px,2.0)*pow(cos(theta),2.0)-4.0*me*px*pow(cos(theta),3.0)*pow(pz,3.0)*sin(theta)-2.0*Mn*pow(pz,2.0)*En*me*pow(cos(theta),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),4.0)+Mn*pow(pz,4.0)*pow(cos(theta),2.0)-me*pow(px,4.0)*pow(cos(theta),4.0)-pow(pz,4.0)*me*pow(cos(theta),4.0)+pow(pz,4.0)*me*pow(cos(theta),2.0)+me*pow(px,4.0)*pow(cos(theta),2.0)-2.0*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(theta),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(theta),2.0)+2.0*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(theta),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(theta),2.0))))/(Mn*me*(Mn*px*sin(theta)+me*px*sin(theta)+Mn*pz*cos(theta)+me*pz*cos(theta)));

 if (vv<0)
   vv=-(-Mn*pow(pz,2.0)*me*pow(cos(theta),2.0)-Mn*me*pow(px,2.0)+Mn*me*pow(px,2.0)*pow(cos(theta),2.0)-2.0*Mn*me*px*sin(theta)*cos(theta)*pz+
	sqrt(-me*pow(Mn,2.0)*(2.0*Mn*pow(pz,3.0)*px*sin(theta)*cos(theta)-2.0*me*pow(px,3.0)*sin(theta)*cos(theta)*pz+2.0*pow(pz,3.0)*me*px*sin(theta)*cos(theta)+2.0*Mn*pow(py,2.0)*px*sin(theta)*cos(theta)*pz-4.0*Mn*me*px*sin(theta)*cos(theta)*pz*En-4.0*En*px*sin(theta)*cos(theta)*pz*pow(Mn,2.0)+2.0*me*px*sin(theta)*cos(theta)*pz*pow(py,2.0)+2.0*Mn*pow(px,3.0)*sin(theta)*cos(theta)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2.0*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2.0*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(theta),3.0)*pz*sin(theta)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(theta),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(px,4.0)*pow(cos(theta),2.0)+2.0*Mn*En*me*pow(px,2.0)*pow(cos(theta),2.0)-4.0*me*px*pow(cos(theta),3.0)*pow(pz,3.0)*sin(theta)-2.0*Mn*pow(pz,2.0)*En*me*pow(cos(theta),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),4.0)+Mn*pow(pz,4.0)*pow(cos(theta),2.0)-me*pow(px,4.0)*pow(cos(theta),4.0)-pow(pz,4.0)*me*pow(cos(theta),4.0)+pow(pz,4.0)*me*pow(cos(theta),2.0)+me*pow(px,4.0)*pow(cos(theta),2.0)-2.0*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(theta),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(theta),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(theta),2.0)+2.0*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(theta),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(theta),2.0))))/(Mn*me*(Mn*px*sin(theta)+me*px*sin(theta)+Mn*pz*cos(theta)+me*pz*cos(theta)));
 

 VV=sqrt((En-.5*me*vv*vv)*2/Mn);
 cosT=(pz-me*vv*cos(theta))/Mn/VV;
 sinT=sqrt(1-cosT*cosT);
 sinF=py/(Mn*VV*sinT);
 cosF=(px-me*vv*sin(theta))/(Mn*VV*sinT);
 
 veltr[0]=vv*sin(theta);
 veltr[1]=0;
 veltr[2]=vv*cos(theta);
 vattr[0]=VV*sinT*cosF;
 vattr[1]=VV*sinT*sinF;
 vattr[2]=VV*cosT;

 coordtrafo( g_system.v1 [actatom],vattr,zaxes,rel,'b');
 coordtrafo( g_system.v1 [actelectron],veltr,zaxes,rel,'b');

 return(0);
}





//############################## FUNCTION #################################

int criteriaforimpion(char *criteria , int actatom,
                      int actelectron, double (*r_pe)[3] , int secel, double (*r_se)[3] ,
		      double E0,double Tpri,double Tsec,int init) {
  int nearatompri,nearatomsec, mion_pri , mion_sec;
  double drnearatompri[3],drnearatomsec[3], mdistpri[3] , mdistsec[3] ;
  double rat_sel_pe[3] ,  rat_sel_se[3] ;
  double (*p_rat_sel_pe)[3] = (double (*)[3])rat_sel_pe ,  (*p_rat_sel_se)[3] = (double (*)[3])rat_sel_se ;
  
  /* criteria: 'd': distance
               'c': Coulomb
	       'n': no extra criteria
  */

  //return(-99) ;

  if ( (Tpri<0) || (Tsec<0) )  {  return(-2) ;  }


  if  ( strstr( criteria , "d" ) != NULL )  {
    nearatompri = nearestatom2electron( actelectron , (double **)(&p_rat_sel_pe) , NULL , 'd' ) ;
    nearatomsec=nearestatom2electron( secel , (double **)(&p_rat_sel_se) , NULL ,'d');
    vectminus( drnearatompri, g_system.r1 [actelectron], g_system.r1 [nearatompri]);
    vectminus(drnearatomsec, g_system.r1 [secel], g_system.r1 [nearatomsec]);
    if ((Tpri<0) || (Tsec<0) ||  
	(Tpri+enerregcut(sqrt(vectnorm2(drnearatompri)),
			  g_system.qi [nearatompri],-1,nearatompri)<0) ||
	(Tsec+enerregcut(sqrt(vectnorm2(drnearatomsec)),
			  g_system.qi [nearatomsec],-1,nearatomsec)<0) )
      return(-3);
  }


  if  ( strstr( criteria , "c" ) != NULL )  {
    double dr_ee[3] ;  
    vectminus( dr_ee , r_pe , r_se ) ; 
    nearatompri = nearestatom2electron( actelectron , (double **)(&p_rat_sel_pe) , NULL , 'c' ) ;
    nearatomsec = nearestatom2electron( secel , (double **)(&p_rat_sel_se) , NULL , 'c' ) ;
    vectminus( drnearatompri , g_system.r1[actelectron] , rat_sel_pe ) ;
    vectminus( drnearatomsec , g_system.r1[secel] , rat_sel_se ) ;

    double Vee = enerregcut( sqrt(vectnorm2(dr_ee)) , -1 , -1 , actelectron ) ;
    double V1 =  enerregcut( sqrt(vectnorm2(drnearatompri)) , g_system.qi[nearatompri] , -1 , nearatompri ) ;
    double V2 =  enerregcut( sqrt(vectnorm2(drnearatomsec)) , g_system.qi[nearatomsec] , -1 , nearatomsec ) ;
    double E1 =  Tpri + V1 + ( actatom == nearatompri ? Vee * 0 : 0 ) ;
    double E2 =  Tsec + V2 ;


    if ( ( E1 < 0 ) || ( E2 < 0 ) ) {  return(-3);  }

  }

  if  ( strstr( criteria , "C" ) != NULL )  {
    if ( (e_deloc_test(actelectron,Tpri,init)<0) ||
         (e_deloc_test(secel,Tsec,init)<0) )
      return(-4);
  }

  if  ( strstr( criteria , "M" ) != NULL )  {
    mion_pri = g_system.motherion[actelectron] ;
    mion_sec = g_system.motherion[secel] ;
    vectminus( mdistpri ,  g_system.r1 [actelectron] ,  g_system.r1 [mion_pri] ) ;
    vectminus( mdistsec ,  g_system.r1 [secel] ,  g_system.r1 [mion_sec] ) ;
    if ( 
         ( Tpri + COULOMB_ION_E_POT(  g_system.qi [mion_pri] ,  g_system.qi [actelectron] , sqrt( vectnorm2( mdistpri ) ) ) < 0 ) || 
         ( Tpri + COULOMB_ION_E_POT(  g_system.qi [mion_sec] ,  g_system.qi [secel]       , sqrt( vectnorm2( mdistsec ) ) ) < 0 ) 
       )
      return(-5);
  }



  return(0);

}






//############################## FUNCTION #################################

double thetainel(double arg,double *params){
  double k,kv,ka,q,cosg,g;
  double E0,E1,E2,th1,out;
  char type[2];
  double th2;
  double result=0;
  
  if (params[0]==1)
    return(0);
  if (params[0]==2)
    return(PI);

  E0=params[1];
  E1=params[2];
  E2=params[3];
  th1=params[4];

  while ( (result==0) || (result==PI) ) {
    if (strcmp(type,"Landau3p736")==0) {
      k=sqrt(2* g_phys_math_const.me *E0)/ g_phys_math_const.hbar     *( g_phys_math_const.hbar * g_phys_math_const.hbar / g_phys_math_const.me /( g_phys_math_const.e1 *  g_phys_math_const.e1  ));
      kv=sqrt(2* g_phys_math_const.me *E1)/ g_phys_math_const.hbar    *( g_phys_math_const.hbar * g_phys_math_const.hbar / g_phys_math_const.me /( g_phys_math_const.e1 *  g_phys_math_const.e1  ));
      ka=sqrt(2* g_phys_math_const.me *E2)/ g_phys_math_const.hbar    *( g_phys_math_const.hbar * g_phys_math_const.hbar / g_phys_math_const.me /( g_phys_math_const.e1 *  g_phys_math_const.e1  ));

      q=sqrt(k*k+kv*kv-2*k*kv*cos(th1));
      cosg=(ka*kv*cos(th1-th2)-ka*k*cos(th2))/(ka*q);
      g=acos(cosg);

      out= kv*ka* (q*q+2*q*ka*cos(g)+(ka*ka+1)*(pow(cos(g),2.0))) /
        (k*q*q* pow(q*q+2*q*ka*cos(g)+1+ka*ka,4) * (pow(q+ka,2.0)+1) *
         (pow(q-ka,2.0)+1) * (1-exp(-2*PI/ka)) )  *
        exp(-2.0/ka*atan(2*ka/(q*q-ka*ka+1))) * sin(th1) * sin(th2);

      result = out  * pow( g_phys_math_const.hbar * g_phys_math_const.hbar / g_phys_math_const.me / g_phys_math_const.e1 /  g_phys_math_const.e1  ,2.0) ;
    }
    else {
      result=randnum(0)*PI;
    }
  }

  return(result);
}






//############################## FUNCTION #################################

int setthetainelscatt(double *th1,double *th2,
                      double E0,double Epri,double Esec){
  static double params[10];
  static char type='i';  // i: isotropically; a: anizo

//#ifdef WORKAROUND_SECEL_PRIMARY_FWD_SCATT
  type =  g_inputs.si_dsig_dome ; 
//#endif

  switch (type){
   case 'f':
    *th1 = 0;
    *th2 = randnum(0)*PI;
    break;
   case 'i':
    *th1=randnum(0)*PI;
    *th2=randnum(0)*PI;
    break;
   case 'a':
    params[1]=E0;
    params[2]=Epri;
    params[3]=Esec;
    params[4]=*th1;
    *th1=drawrandomdistr(&thetainel,params,1,0);
    *th2=drawrandomdistr(&thetainel,params,1,0);
    break;
  }
  return(0);

}





//############################## FUNCTION #################################

double dsigdW_inel_int(double W,double *params){
  int N,n;
  double B,U,T,Ni,t,u,w, a,b,c,d,e;

  N=params[1];
  B=params[2];
  U=params[3];
  T=params[4];
  n=params[5];
  Ni=params[6];

  t=T/B;
  u=U/B;
  w=W/B;

  if (params[0]==1)
    return(0);
  if (params[0]==2)
    return(T-B);

  a=0;
  b=0;
  c=0;
  d=0;
  e=0;

  return( (Ni/N-2)/(t+1)*(log(w+1)-log(t-w)) +
          (2-Ni/N)* (-1/(w+1) + 1/(t-w)) +
	  ( log(t)/N*log(w+1) *
	   (a*(-1/(w+1)) + b*(-1/2.0/pow((w+1),2.0)) + c*(-1/3.0/pow((w+1),3.0))+
	    d*(-1/4.0/pow((w+1),4.0)) + e*(-1/5.0/pow((w+1),5.0) )) ) );  // ???
}



//############################## FUNCTION #################################

int setineleE(double *Epri,double *Esec,double E0,
              double B,double K,double NE) {
  static double params[20];
  //static char type='r';  // h: half-half; e: equally; r:real

  switch ( g_inputs.inel_dsdW){
   case 'h':
    *Epri = 0.5 * ( E0 - B ) ;
    break;
   case 'e':
    *Epri = randnum(0) * ( E0 - B ) ;
    break;
   case '3':
    //*Epri=0.5*(E0-actshell->B);
    if ( E0 - B > 2 * 30 *  g_phys_math_const.e1  ) {
        *Epri = 30 *  g_phys_math_const.e1  ;
    }
    else {
        *Epri = randnum(0) * ( E0 - B ) ;
    }
    break;
   case 'r':
    params[0] = 0 ;
    params[1] = NE ;
    params[2] = B ;
    params[3] = K ;
    params[4] = E0 ;
    params[5] = 1 ;  // n
    params[6] = 1 ;  // Ni
    *Epri = drawrandomdistr( dsigdW_inel_int , params , 2 , 0 ) ;
    break;
  }

  if (*Epri < ( E0 - B ) / 2.0 )
    *Epri = E0 - B - *Epri ;

  *Esec = E0 - B - *Epri ;

  return 0 ;
}






//############################## FUNCTION #################################

int do_si_effect(int step , int actatom , int actelectron , int impshell ,
                 double *r1e_pr , double *r2e_pr ,
                 int *secondaryelectron , double *dbginfo , int init  , char **logstr ) 
{
  double Mn,Ebind,Eatom,Eprimary0=0,Eprimary0rel,Esecondary,Eprimary, pot2;
  double Epri,Esec,E0,Ekinplus=0;
  double Ttotal,En,Tpri,Tsec,vv;
  double th1,th2;
  double vel[3],veltr[3],vel1tr[3],vel1lab[3];
  double velsectr[3],velseclab[3],vattr[3],vatlab[3];
  double rel[3],reltr[3],relsectr[3],relseclab[3],rat0[3];
  double zaxes[3],xaxes[3],tempvect[3];
  double eedist[3];
  double thetapossec,dtheta=0.01;
  double px,py,pz,vsec,VV;
  //double intactfactor1,scale;
  double sinT,cosT,sinF,cosF;
  char backstepindicator[SSTRLTH];
  int i,secel;
  FILE *fptemp;
  double potEbefore,potEafter,Etotbefore,Etotafter, potEee, potEe1, potEe2;
  double potEloc1,potEloc2;
  double intactfactor;
  int critvalue;
  double rstartsec;
  char impioncriterium[STRLTH];
  int dQatom,dQelectron;

  static char scatterparticlesdataout='n';

  double subpotEbef,subpotEaft,Ekinbef,Ekinaft;

  struct g_struct_xatom_data *atomic_econfig ;

  double me =  g_phys_math_const.me ;

  double r_cellshift[3] ;

  int Ecalc_full = 0 ;

  
  int *atoms_affected = calloc( 1 , sizeof(int) ) , num_atoms_affected = 1 ; atoms_affected[0] = actatom ;


  vectminus( r_cellshift , r1e_pr , g_system.r1[actelectron] ) ;

  strcpy( impioncriterium ,  g_inputs.impioncriterium ) ;

  atomic_econfig = g_atomic_data[actatom].atomic_econfig ;

  Ebind = atomic_econfig->B[impshell] ; // Ebinding_shell(actatom,impshell);
  dbginfo[0] = Ebind / MESSAGES.unit_E ; // dbginfo
  secel = getrand_shell_el( actatom , impshell ) ;

  xmdyn_store_restore_subsystem( "store" , "-" , 3 , actatom , actelectron, secel ) ;

  *secondaryelectron = secel ;

  Mn = g_system.m[actatom] ;

  if (scatterparticlesdataout=='y')
    foto2disk(0,-3);



  if ( Ecalc_full )
  {
      xmdyn_md_energy( "system-V" , "calc" ) ; //X calc_tot_init_ener(12) ;
      potEbefore = g_ENERGY_VALUES.totalpotE ;
      subpotEbef = potEbefore ;
      Ekinbef = xmdyn_md_energy( "system-K" , "calc" ) ; //X calc_tot_init_ener(2) ;
  }
  else
  {
      xmdyn_md_energy( "system-V" , "usecache" ) ; //X  calc_tot_init_ener(212);
      potEbefore = g_ENERGY_VALUES.totalpotE ;
      subpotEbef = potEactual( actatom , actelectron , -1 , num_atoms_affected , atoms_affected ) ;
      Ekinbef = xmdyn_md_energy( "system-K" , "calc" ) ; //X calc_tot_init_ener(2) ;
  //Etotcalcbef=calc_tot_init_ener(1);  // DEBUG
  }

//  printf( "DBG3y %5d %4d %4d %le\n" , step , actatom, actelectron , g_ENERGY_VALUES.totalpotE/1.6022e-19 ) ;

  Eatom = .5 * g_system.m[actatom] * vectnorm2( g_system.v1[actatom] ) ;
  dbginfo[1] = Eatom / MESSAGES.unit_E ; // dbginfo
  Eprimary0 = .5 * g_system.m[actelectron] * vectnorm2( g_system.v1[actelectron] ) ;
  dbginfo[2] = Eprimary0 / MESSAGES.unit_E ; // dbginfo
  Etotbefore = Eatom + Eprimary0 + potEbefore ;

  vectminus( rel, r1e_pr, g_system.r1[actatom] ) ;
  vectminus( vel, g_system.v1[actelectron], g_system.v1[actatom] ) ;
  vect3equal( zaxes , vel ) ;
  vect3equal( xaxes , rel ) ;
  rat0[0] = rat0[1] = rat0[2] = 0 ;

  coordtrafo( reltr , rel , zaxes , xaxes , 't' ) ;
  coordtrafo( veltr , vel , zaxes , xaxes , 't' ) ;
  
  Eprimary0rel = .5 * g_system.m[actelectron] * vectnorm2(vel) ;


  double rat_sel[3] , xxx1  ;  double *ppp ; ppp = rat_sel ;
  int nearatompri = nearestatom2electron( actelectron , &ppp , &xxx1 , 'c' ) ;
  double drnearatompri[3] ;
  vectminus( drnearatompri , r1e_pr , rat_sel ) ;
  double V1 = enerregcut( sqrt(vectnorm2(drnearatompri)) , g_system.qi[nearatompri] , -1 , nearatompri ) ;
  //* TODO: consider uncommenting below, but breaks backward compatibility *//
//  if ( Eprimary0 + V1 < Ebind ) {   return( -99 ) ;   }
  

  pot2 = 0 ;
  if ( strcmp( g_inputs.si_type , "2atom" ) == 0 ) {
    pot2 = pot_of_2nd_nearest_ion( actelectron , actatom ) ;
  }

  E0 = .5* g_phys_math_const.me *vectnorm2(vel) + 
       enerregcut(sqrt(vectnorm2(rel)), g_system.qi [actelectron], g_system.qi [actatom],actatom) + pot2 ;
  setineleE( &Epri , &Esec , E0 , 
             atomic_econfig->B[impshell] , atomic_econfig->K[impshell] , 
             atomic_econfig->NE_orbit[impshell] ) ;
  setthetainelscatt( &th1 , &th2 , E0 , Epri , Esec ) ;
  //}  

  dbginfo[3] = E0 / MESSAGES.unit_E ;   // dbginfo
  dbginfo[4] = Epri / MESSAGES.unit_E ; // dbginfo
  dbginfo[5] = Esec / MESSAGES.unit_E ; // dbginfo


  double r1e_sec[3] ;
  thetapossec = -PI/2.0 - ( th2 - PI/2.0 ) / fabs( th2 - PI/2.0 ) * dtheta ;
  thetapossec = -th2 ;
  rstartsec = actualrb( actatom , impshell ) ;
  relsectr[0] = rstartsec * sin( thetapossec ) ;
  relsectr[1] = 0 ;
  relsectr[2] = rstartsec * cos( thetapossec ) ;
  coordtrafo(relseclab,relsectr,zaxes,xaxes,'b');
  r1e_sec[0] = relseclab[0] + g_system.r1[actatom][0] ;
  r1e_sec[1] = relseclab[1] + g_system.r1[actatom][1] ;
  r1e_sec[2] = relseclab[2] + g_system.r1[actatom][2] ; 

  vect3equal( g_system.r1[secel] , r1e_sec ) ;
  vect3equal( g_system.r2[secel] , r1e_sec ) ;


  strcpy( backstepindicator , "n" ) ;  // TODO : possibly restict to si_type 3 and 4, to be on the safe side
  if ( th1 > PI/2.0 ) 
  {
    Ekinplus*=1;;
  }

  if ( ( g_system.ptype[actelectron] == 'f' ) || 
       ( g_system.ptype[actelectron] == 'g' ) ) 
     g_system.qi[secel] = dQelectron = g_inputs.qsecphoto ;
  else if ( g_system.ptype[actelectron] == 'a' )
     g_system.qi[secel] = dQelectron = g_inputs.qsecauger ;
  else if ( ( g_system.ptype[actelectron] == 's' ) || 
            ( g_system.ptype[actelectron] == 't' ) || 
            ( g_system.ptype[actelectron] == 'u' ) || 
            ( g_system.ptype[actelectron] == 'v' )  
          ) 
     g_system.qi[secel] = dQelectron = g_inputs.qsecsec ;
  else
     g_system.qi[secel] = dQelectron =-1 ;
  // add new type of electron here !!!  
   g_system.qi[actatom] += g_inputs.dqatomsec ;
  dQatom = g_inputs.dqatomsec ;
  g_atomic_data[actatom].NE-- ;
  g_atomic_data[actatom].shell[impshell].NE-- ;
  //el_config_update( actatom ) ;
  xmdyn_update_atoms( logstr , "array,full,ionization" , num_atoms_affected , atoms_affected ) ; // xmdyn_update_atoms( logstr , 1 , actatom );



  if ( Ecalc_full )
  {
      xmdyn_md_energy( "system-V" , "calc" ) ; //X calc_tot_init_ener(12) ;
      potEafter = g_ENERGY_VALUES.totalpotE ;
      subpotEaft = potEafter ;
  }
  else
  {
      subpotEaft = potEactual( actatom , actelectron , secel , num_atoms_affected , atoms_affected ) ;
      potEafter = potEbefore - ( subpotEbef - subpotEaft ) ;
  }


//  printf( "DBG3z %5d %4d %4d %le\n" , step , actatom, actelectron , g_ENERGY_VALUES.totalpotE/1.6022e-19 ) ;


  // The old way: `intactfactor`
  vectminus( eedist , r1e_pr , r1e_sec ) ;
  intactfactor = 0.5 ;     // the ratio of the el-el pot. the faster el. gets modified. 
  intactfactor = 1.0 ;     // the ratio of the el-el pot. the faster el. gets modified with. 

  potEe1 = enerregcut( sqrt( vectnorm2(rel) ) , g_system.qi[actelectron] , g_system.qi[actatom] , actatom ) ;
  potEe2 = enerregcut( rstartsec , g_system.qi[secel] , g_system.qi[actatom] , actatom ) ;
  potEee = enerreg( sqrt( vectnorm2(eedist) ) , g_system.qi[secel] , g_system.qi[actelectron] ) ;


  Tpri = Epri - 
         enerregcut( sqrt( vectnorm2(rel) ) , 
                     g_system.qi[actelectron] , g_system.qi[actatom] , actatom )
         - pot2 - intactfactor * potEee ;


  Tsec = Eprimary0rel - Ebind + potEbefore - potEafter - Tpri ;

//fprintf( stderr , "%d %d %d %d %le %le %le %le %le %le  %le %le %le %le %le \n " , step , actatom, actelectron, secel, Ebind , subpotEbef/e1 , subpotEaft/e1 , potEbefore/e1 , potEafter/e1 , Eprimary0/e1 , Epri/e1 , Esec/e1 , E0/e1 , Tpri/e1 , Tsec/e1 ) ;

  dbginfo[6] = Tpri / MESSAGES.unit_E ; // dbginfo
  dbginfo[7] = Tsec / MESSAGES.unit_E ; // dbginfo
  
  dbginfo[8]  = sqrt( vectnorm2(rel) ) / MESSAGES.unit_l ; // dbginfo
  dbginfo[9]  = sqrt( vectnorm2(relseclab) ) / MESSAGES.unit_l ; // dbginfo
  dbginfo[10] = sqrt( vectnorm2(eedist) ) / MESSAGES.unit_l ; // dbginfo

  if ( ( Tpri < 0 ) || ( Tsec < 0 ) ) {
     free( atoms_affected ) ; xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;
    return(-10);
  }

  g_system.motherion[secel] = actatom ;

  critvalue = criteriaforimpion( impioncriterium , 
                                 actatom , actelectron , (double (*)[3])r1e_pr , secel , (double (*)[3])r1e_sec ,
                                 E0 , Tpri , Tsec , init ) ;

  /* criteria: 'd':distance; 'c':Coulomb; 'n':no extra criteria  */
  if ( critvalue < 0 ) {
     free( atoms_affected ) ; xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;
    return(critvalue);
  }
  if (( g_system.ptype [actelectron]=='f') || ( g_system.ptype [actelectron]=='g'))
     g_system.ptype [secel]='s';
  if (( g_system.ptype [actelectron]=='s') || ( g_system.ptype [actelectron]=='t'))
     g_system.ptype [secel]='s';
  if ( g_system.ptype [actelectron]=='a')
     g_system.ptype [secel]='s';
  if (( g_system.ptype [actelectron]=='u') || ( g_system.ptype [actelectron]=='v'))
     g_system.ptype [secel]='s';
   g_system.ptype [secel]='s';

  vv = sqrt( 2.0 * Tpri / me ) ;
  vel1tr[0] = vv * sin( th1 ) ;
  vel1tr[1] = 0 ;
  vel1tr[2] = vv * cos( th1 ) ;

  th2 *= (-1) ;

  px = - me * vel1tr[0] ;
  py = - me * vel1tr[1] ;
  pz = me * veltr[2] - me * vel1tr[2] ;
  En = Tsec ;
  
  if ( (Tsec<0) || ( (-me*pow(Mn,2.0)*(2*Mn*pow(pz,3.0)*px*sin(th2)*cos(th2)-2*me*pow(px,3.0)*sin(th2)*cos(th2)*pz+2*pow(pz,3.0)*me*px*sin(th2)*cos(th2)+2*Mn*pow(py,2.0)*px*sin(th2)*cos(th2)*pz-4.0*Mn*me*px*sin(th2)*cos(th2)*pz*En-4.0*En*px*sin(th2)*cos(th2)*pz*pow(Mn,2.0)+2*me*px*sin(th2)*cos(th2)*pz*pow(py,2.0)+2*Mn*pow(px,3.0)*sin(th2)*cos(th2)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(th2),3.0)*pz*sin(th2)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(th2),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(px,4.0)*pow(cos(th2),2.0)+2*Mn*En*me*pow(px,2.0)*pow(cos(th2),2.0)-4.0*me*px*pow(cos(th2),3.0)*pow(pz,3.0)*sin(th2)-2*Mn*pow(pz,2.0)*En*me*pow(cos(th2),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),4.0)+Mn*pow(pz,4.0)*pow(cos(th2),2.0)-me*pow(px,4.0)*pow(cos(th2),4.0)-pow(pz,4.0)*me*pow(cos(th2),4.0)+pow(pz,4.0)*me*pow(cos(th2),2.0)+me*pow(px,4.0)*pow(cos(th2),2.0)-2*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(th2),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(th2),2.0)+2*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(th2),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(th2),2.0))) <0) ){
     free( atoms_affected ) ; xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;
    return(-1);
  }


  vsec=-(-Mn*pow(pz,2.0)*me*pow(cos(th2),2.0)-Mn*me*pow(px,2.0)+Mn*me*pow(px,2.0)*pow(cos(th2),2.0)-2.0*Mn*me*px*sin(th2)*cos(th2)*pz-
	 sqrt(-me*pow(Mn,2.0)*(2.0*Mn*pow(pz,3.0)*px*sin(th2)*cos(th2)-2.0*me*pow(px,3.0)*sin(th2)*cos(th2)*pz+2.0*pow(pz,3.0)*me*px*sin(th2)*cos(th2)+2.0*Mn*pow(py,2.0)*px*sin(th2)*cos(th2)*pz-4.0*Mn*me*px*sin(th2)*cos(th2)*pz*En-4.0*En*px*sin(th2)*cos(th2)*pz*pow(Mn,2.0)+2.0*me*px*sin(th2)*cos(th2)*pz*pow(py,2.0)+2.0*Mn*pow(px,3.0)*sin(th2)*cos(th2)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2.0*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2.0*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(th2),3.0)*pz*sin(th2)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(th2),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(px,4.0)*pow(cos(th2),2.0)+2*Mn*En*me*pow(px,2.0)*pow(cos(th2),2.0)-4.0*me*px*pow(cos(th2),3.0)*pow(pz,3.0)*sin(th2)-2*Mn*pow(pz,2.0)*En*me*pow(cos(th2),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),4.0)+Mn*pow(pz,4.0)*pow(cos(th2),2.0)-me*pow(px,4.0)*pow(cos(th2),4.0)-pow(pz,4.0)*me*pow(cos(th2),4.0)+pow(pz,4.0)*me*pow(cos(th2),2.0)+me*pow(px,4.0)*pow(cos(th2),2.0)-2*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(th2),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(th2),2.0)+2*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(th2),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(th2),2.0))))/(Mn*me*(Mn*px*sin(th2)+me*px*sin(th2)+Mn*pz*cos(th2)+me*pz*cos(th2)));

  if (vsec<0)
    vsec=-(-Mn*pow(pz,2.0)*me*pow(cos(th2),2.0)-Mn*me*pow(px,2.0)+Mn*me*pow(px,2.0)*pow(cos(th2),2.0)-2*Mn*me*px*sin(th2)*cos(th2)*pz+
	   sqrt(-me*pow(Mn,2.0)*(2*Mn*pow(pz,3.0)*px*sin(th2)*cos(th2)-2*me*pow(px,3.0)*sin(th2)*cos(th2)*pz+2*pow(pz,3.0)*me*px*sin(th2)*cos(th2)+2*Mn*pow(py,2.0)*px*sin(th2)*cos(th2)*pz-4.0*Mn*me*px*sin(th2)*cos(th2)*pz*En-4.0*En*px*sin(th2)*cos(th2)*pz*pow(Mn,2.0)+2*me*px*sin(th2)*cos(th2)*pz*pow(py,2.0)+2*Mn*pow(px,3.0)*sin(th2)*cos(th2)*pz+Mn*pow(px,2.0)*pow(pz,2.0)+me*pow(px,2.0)*pow(pz,2.0)+Mn*pow(py,2.0)*pow(px,2.0)-2*En*pow(px,2.0)*pow(Mn,2.0)+me*pow(px,2.0)*pow(py,2.0)-2*Mn*En*me*pow(px,2.0)+Mn*pow(px,4.0)+4.0*me*pow(px,3.0)*pow(cos(th2),3.0)*pz*sin(th2)-me*pow(px,2.0)*pow(py,2.0)*pow(cos(th2),2.0)+Mn*pow(py,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(px,4.0)*pow(cos(th2),2.0)+2*Mn*En*me*pow(px,2.0)*pow(cos(th2),2.0)-4.0*me*px*pow(cos(th2),3.0)*pow(pz,3.0)*sin(th2)-2*Mn*pow(pz,2.0)*En*me*pow(cos(th2),2.0)+6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),4.0)+Mn*pow(pz,4.0)*pow(cos(th2),2.0)-me*pow(px,4.0)*pow(cos(th2),4.0)-pow(pz,4.0)*me*pow(cos(th2),4.0)+pow(pz,4.0)*me*pow(cos(th2),2.0)+me*pow(px,4.0)*pow(cos(th2),2.0)-2*pow(Mn,2.0)*pow(pz,2.0)*En*pow(cos(th2),2.0)-6*me*pow(px,2.0)*pow(pz,2.0)*pow(cos(th2),2.0)-Mn*pow(py,2.0)*pow(px,2.0)*pow(cos(th2),2.0)+2*En*pow(px,2.0)*pow(Mn,2.0)*pow(cos(th2),2.0)+pow(pz,2.0)*me*pow(py,2.0)*pow(cos(th2),2.0))))/(Mn*me*(Mn*px*sin(th2)+me*px*sin(th2)+Mn*pz*cos(th2)+me*pz*cos(th2)));

  if (En-.5*me*vsec*vsec<0) {
    free( atoms_affected ) ; xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;
    return(-11);
  }
  
  VV = sqrt( ( En - .5 * me * vsec * vsec ) * 2.0 / Mn ) ;
  cosT = ( pz - me * vsec * cos( th2 ) ) / Mn / VV ; 
  if ( ( cosT > 1 ) || ( cosT < -1 ) ) 
  {
    free( atoms_affected ) ; xmdyn_store_restore_subsystem( "restore" , "-" , 0 ) ;
    return(-12);
  }

  sinT = sqrt( 1.0 - cosT * cosT ) ;
  sinF = py / ( Mn * VV * sinT ) ;
  cosF = ( px - me * vsec * sin( th2 ) ) / ( Mn * VV * sinT ) ;

  velsectr[0] = vsec * sin( th2 ) ;
  velsectr[1] = 0 ; 
  velsectr[2] = vsec * cos( th2 ) ;
  vattr[0] = VV * sinT * cosF ;
  vattr[1] = VV * sinT * sinF ;
  vattr[2] = VV * cosT ;



  coordtrafo( vel1lab , vel1tr , zaxes , xaxes , 'b' ) ;
  g_system.v1[actelectron][0] = vel1lab[0] + g_system.v1[actatom][0] ;
  g_system.v1[actelectron][1] = vel1lab[1] + g_system.v1[actatom][1] ;
  g_system.v1[actelectron][2] = vel1lab[2] + g_system.v1[actatom][2] ;

  coordtrafo( velseclab , velsectr , zaxes , xaxes, 'b' ) ;
  g_system.v1[secel][0] = velseclab[0] + g_system.v1[actatom][0] ;
  g_system.v1[secel][1] = velseclab[1] + g_system.v1[actatom][1] ;
  g_system.v1[secel][2] = velseclab[2] + g_system.v1[actatom][2] ;

  coordtrafo( vatlab , vattr , zaxes , xaxes , 'b' ) ;
  g_system.v1[actatom][0] = vatlab[0] + g_system.v1[actatom][0] ;
  g_system.v1[actatom][1] = vatlab[1] + g_system.v1[actatom][1] ;
  g_system.v1[actatom][2] = vatlab[2] + g_system.v1[actatom][2] ;


    array_max( g_system.uid + secel , g_system.uid , g_inputs.N + g_inputs.E , "lli" ) ; g_system.uid[secel]++ ;

  Etotafter = .5 * me * ( vectnorm2( g_system.v1[actelectron] ) + 
                          vectnorm2( g_system.v1[secel] )
                        )
              + .5 * Mn * vectnorm2( g_system.v1[actatom] ) + potEafter ;

  if (scatterparticlesdataout=='y')
    foto2disk(0,-4);

  if ( Ecalc_full )
  {
      xmdyn_md_energy( "system-V" , "calc" ) ; //X calc_tot_init_ener(12) ;
      potEafter = g_ENERGY_VALUES.totalpotE ;
      subpotEaft = potEafter ;
      Ekinaft = xmdyn_md_energy( "system-K" , "calc" ) ; //X  calc_tot_init_ener(2) ;
  }
  else
  {
      subpotEaft = potEactual( actatom , actelectron , secel, num_atoms_affected , atoms_affected  ) ;
      Ekinaft = xmdyn_md_energy( "system-K" , "calc" ) ; //X calc_tot_init_ener(2) ;
      xmdyn_md_energy( "system-V" , "usecache" ) ; //X calc_tot_init_ener(212);
  }


  g_ENERGY_VALUES.totalpotE = potEbefore - subpotEbef + subpotEaft ;
  g_ENERGY_VALUES.Eintotal += ( Ekinaft - Ekinbef - subpotEbef + subpotEaft ) ;
  g_ENERGY_VALUES.Eintotal_th += (- Ebind ) ;
  g_ENERGY_VALUES.Einsi += ( - Ebind ) ;
  g_ENERGY_VALUES.Ebound -= ( - Ebind ) ;

  //Etotcalcaft=calc_tot_init_ener(1);  // DEBUG

  Esecondary = .5 * g_system.m[secel] * vectnorm2( g_system.v1[secel] ) ;
  Eprimary   = .5 * g_system.m[actelectron]  *vectnorm2( g_system.v1[actelectron] ) ;


  free( atoms_affected ) ; xmdyn_store_restore_subsystem( "clear" , "-" , 0 ) ;
  return(0);
}




//############################## FUNCTION #################################

void scatter( int actatom , int actelectron , int step , double *r1e , double *r2e , double *v1e , double *v2e ) 
{
  double *subsigmacum,temp;
  int scattersuccess=-1,scattertry=0;
  int i,gotevent,impshell,secel,ii;
  static int maxscattindic=1;
  double dbginfo[20] , T ;
  int init;
  int ii_mess ; 
  char *logstr ; logstr = calloc(2,sizeof(char) ) ; strcpy( logstr , "" ) ;
  char cfgstr_f[LSTRLTH] ;
  char cfgstr_i[LSTRLTH] ;

  if ( maxscattindic )
  {
     g_inputs.maxscattertry = 5 ;
    maxscattindic = 0; 
  }
  
  init = 1 ;

  get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                      g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr_i, 'L'  ) ;

  T = eimpion_effective_KE( actatom , r1e , v1e ) ;

  subsigmacum = calloc( g_atomic_data[actatom].atomic_econfig->max_orbit + 1 , sizeof(double) ) ;
  subsigmacum[0] = elastsigma( actatom , T ) ;
  for ( i = 1 ; i <= g_atomic_data[actatom].atomic_econfig->max_orbit ; ++i ) 
  {
    subsigmacum[i] = subsigmacum[i-1] +
                     eimpion_sigma_orbit_as_KE( actatom , i , T ) ;
  }

  temp = randnum(0) * subsigmacum[ g_atomic_data[actatom].atomic_econfig->max_orbit ] ;

  gotevent = 0 ; impshell = 0 ;
  while ( !gotevent )
  {
    if ( temp < subsigmacum[impshell] ) 
    {
      gotevent = 1 ;
    }
    else
    {
      impshell++ ;
    }
    if ( impshell == g_atomic_data[actatom].atomic_econfig->max_orbit )
    {
      gotevent = 1 ;
    }
  }

  if ( impshell == 0 )
  {  // -> elastic scattering
  }

  else {    // -> inelastic scattering

    while ( ( scattersuccess < 0 ) &&
            (scattertry < g_inputs.maxscattertry ) )
    {
      if ( ( strcmp( g_inputs.si_type , "1atom" ) == 0 ) || 
           ( strcmp( g_inputs.si_type , "2atom" ) == 0 ) 
          ) 
      {
        strcpy( logstr , "" ) ;
        scattersuccess=
        do_si_effect( step , actatom , actelectron , impshell , r1e , r2e , &secel , dbginfo , init , &logstr ) ;
      }
      scattertry++ ;
      init = 0 ;
    }

    if ( 0 <= scattersuccess )
    {
        xmdyn_apply_shortcut( "cancel" ) ;
     }


#if LOG_SYSTEM == 2 ||  LOG_SYSTEM == 0
    if ( scattersuccess < 0 )
    {
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( "-si:" , NULL , "" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
#ifndef XMDYN_XRAYPAC_MORELOG
        free( logstr) ; free( subsigmacum ) ; return ;
#endif // XMDYN_XRAYPAC_MORELOG
    }
    else
    {
        print_log( "+si:" , NULL , "" , "noflush" ) ;
    }
        int temp_i ;
        double temp_d ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " atom:" , (void*) &actatom  , "int" , "noflush" ) ;
        print_log( " Z:" , (void*) &(g_system.Z[actatom])  , "int" , "noflush" ) ;
        temp_i = g_system.qi[actatom] - ( scattersuccess >= 0 ? g_inputs.dqatomsec : 0  ) ; 
        print_log( " q:" , (void*) &temp_i  , "int" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " shell:" , (void*) &impshell  , "int" , "noflush" ) ;
        print_log( " el0:" , (void*) &actelectron  , "int" , "noflush" ) ;
        print_log( " el:" , (void*) &secel  , "int" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
        print_log( " uid0:" , (void*) (g_system.uid + actelectron) , "llint" , "noflush" ) ;
        print_log( " uid:" , (void*) (g_system.uid + secel) , "llint" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " Eatom:" , (void*) &(dbginfo[1]) , "double" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
        print_log( " Ebind:" , (void*) &(dbginfo[0]) , "double" , "noflush" ) ;
        print_log( " Epri0:" , (void*) &(dbginfo[2]) , "double" , "noflush" ) ;
        print_log( " E0:" , (void*) &(dbginfo[3]) , "double" , "noflush" ) ;
        print_log( " Epri:" , (void*) &(dbginfo[4]) , "double" , "noflush" ) ;
        print_log( " Esec:" , (void*) &(dbginfo[5]) , "double" , "noflush" ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
        print_log( " Tpri:" , (void*) &(dbginfo[6]) , "double" , "noflush" ) ;
        print_log( " Tsec:" , (void*) &(dbginfo[7]) , "double" , "noflush" ) ;
        print_log( " rel:" , (void*) &(dbginfo[8]) , "double" , "noflush" ) ;
        print_log( " rsec:" , (void*) &(dbginfo[9]) , "double" , "noflush" ) ;
        print_log( " ree:" , (void*) &(dbginfo[10]) , "double" , "noflush" ) ;
        temp_d = scattersuccess ;
        print_log( " critval:" , (void*) &(temp_d) , "double" , "noflush" ) ;
#endif // XMDYN_XRAYPAC_MORELOG
        
        get_xatom_conf_str( g_atomic_data[actatom].atomic_econfig->max_orbit , 
                            g_atomic_data[actatom].atomic_econfig->NE_orbit , cfgstr_f, 'L'  ) ;
        print_log_full( " cfg_i:" , (void*)cfgstr_i , "str" , "noflush" ) ;
        print_log_full( " cfg_f:" , (void*)cfgstr_f , "str" , "noflush" ) ;

        print_log( " \n" , NULL , "" , "noflush" ) ; 
        if ( 0 <= scattersuccess ) { print_log( logstr , NULL , "" , "noflush" ) ; }
        print_log( "" , NULL , "" , "flush" ) ;


#endif

  }

  free( logstr) ;
  free( subsigmacum ) ;
}





//############################## FUNCTION #################################

double elastsigma( int actatom , double T )
{
  return(0);
}



//############################## FUNCTION #################################

double elastsigma_max( int Z , double T ) 
{
  return(0);
}



//############################## FUNCTION #################################

double elastsigma_atomic_max( double T ,int Z ) 
{
  return(0);
}




//############################## FUNCTION #################################

double inelastsigma_YKKIM_PRA50( double T, double B , double U ,
                                 int NE , double Q , double n )
{
  static double R,rb0=0.52918e-10;
  double t,u,S,result;

  R = 13.6057 * g_phys_math_const.e1 ;

  if ( ( NE == 0 ) || ( T <= B ) ) 
  {
    return(0);
  }

  S = 4.0 * PI * rb0 * rb0 * NE * ( R / B ) * ( R / B ) ;
  t = T / B ; 
  u = U / B ;
  
  result = ( S / ( t + ( u + 1.0 ) / n ) * 
             ( Q * log(t) / 2.0 * ( 1.0 - 1.0 / ( t * t ) ) +
               ( 2.0 - Q ) * ( 1.0 - 1.0 / t - log(t) / ( t + 1.0 ) ) 
             ) 
           ) ;
  return( result ) ;
}





//############################## FUNCTION #################################

double inelastsigma_Lotz( double T , double B , double U ,
                          int NE , double Q , double n )
{
  static double R , rb0 = 0.52918e-10 ;
  double t,S,result;

  R = 13.6057 * g_phys_math_const.e1 ;
  //Q=1.0;  // ???
  //n=1.0;  // ???

  if ( ( NE == 0 ) || ( T <= B ) ) 
  {
    return(0);
  }

  S = 4.0 * PI * rb0 * rb0 * NE * ( R / B ) * ( R / B ) ;
  t = T / B ; //u=U/B;
  
  result = S * log(t) / t ;

  return( result ) ;
}






//############################## FUNCTION #################################

double inelastsigma_solid_fit_diamond(double T,double BindE,double U,
                                      int NE,double Q,double n_input)
{
    double E, A,B,a,b,Eth,E0,  rho,n,Mmol,  lam,sigma  ;
    double ps_ratio ;

    if ( fabs( BindE ) > 20* g_phys_math_const.e1  )
    {
        return( 0 ) ;
    }

    A=14.0 ; B=-38.0 ; a=4.78e-6 ; b=4.5 ; Eth=5.47 ; E0=1.0 ;
    rho = 3.53 ; // [g/cm^3]
    Mmol = 12.0 ;  // [g]
    n = rho * 6.02e23 / Mmol * 1e-24 ; // [1/A^3]

    E = T ;
    E /=  g_phys_math_const.e1  ;
    lam = sqrt(E) / ( a * pow(E-Eth,b) )  +  ( E - E0 * exp(-B/A) ) / ( A * log(E/E0) + B ) ;

    sigma = 1.0 / ( n * lam ) * 1e-20 ;   // [m^2]


    ps_ratio = 3.0 ;

    if ( fabs( BindE ) < 15* g_phys_math_const.e1  )     //  2p electron
    {
        sigma = NE * sigma / 2.0 / ( ps_ratio + 1.0 ) * ps_ratio ;
    }
    else                             //  2s electron
    {
        sigma = NE * sigma / 2.0 / ( ps_ratio + 1.0 ) ;
    }


    return( sigma ) ;


}





//############################## FUNCTION #################################

double eimpion_sigma_orbit_as_KE( int actatom , int actshell , double T ) 
{
  double Q,B,U,n , scale_factor ;
  int NE;
  //struct SHELL_STRUCT *shell;
  double  (*sel_inelastsigma) (double,double,double,int,double,double) ;

  //shell=g_atomic_data[actatom].shell+actshell;
  NE = g_atomic_data[actatom].atomic_econfig->NE_orbit[actshell] ;    //shell->NE;
  B  = g_atomic_data[actatom].atomic_econfig->B[actshell] ;           //shell->B;
  U  = g_atomic_data[actatom].atomic_econfig->K[actshell] ;           //shell->K;
  Q  = 1.0 ;  //* NOTE: osc. strength unknown, therefore Q=1 is used *//
  n  = 1.0 ;  // 

  /* This is the default */
  sel_inelastsigma = &inelastsigma_YKKIM_PRA50 ;
  scale_factor = g_inputs.impion_scale ;

  /* These are other cross sections */
  if ( strstr( g_inputs.impion_model , "BEB-neutral" ) != NULL ) 
  {
     if ( g_atomic_data[actatom].atomic_econfig->NE < *(g_atomic_data[actatom].Z) )
     {   
         return( 0 ) ;
     }
  }

  if ( strstr(  g_inputs.impion_model , "BEB-C0only-real" ) != NULL ) 
  {
      if ( g_atomic_data[actatom].atomic_econfig->NE < *(g_atomic_data[actatom].Z) )
      {   
          return( 0 ) ;
      }
      if ( actshell == 1 )
      {   
          B = 308 *  g_phys_math_const.e1  ;
          U = 437 *  g_phys_math_const.e1 ;
          Q = 1;
      }
      else if ( actshell == 3 )
      {   
          B = 19.2 *  g_phys_math_const.e1 ;
          U = 41.9 *  g_phys_math_const.e1 ;
          Q = 1;
      }
      else if ( ( actshell == 5 ) || ( actshell == 6 ) )
      {   
          B = 11.5 * g_phys_math_const.e1 ;
          U = 34.1 * g_phys_math_const.e1  ;
          Q = 1 ;
      }
      else
      {   
          if ( NE != 0 )
              printf("Not definied %d %d\n" , actshell , NE ) ;
      }
  }


  if ( strstr(  g_inputs.impion_model , "diamond" ) != NULL ) 
  {
      sel_inelastsigma = &inelastsigma_solid_fit_diamond ;
  }


  if ( ( strstr(  g_inputs.impion_model , "lotz" ) != NULL ) || 
       ( strstr(  g_inputs.impion_model , "Lotz" ) != NULL )     )
  {
      sel_inelastsigma = &inelastsigma_Lotz ;
  }


  /* Now we evaluate and return the cross section value  */
  return( scale_factor * sel_inelastsigma( T , B , U , NE , Q , n ) ) ;
  
}




//############################## FUNCTION #################################

double eimpion_sigma_total_as_KE( int atomid , double a_E ){
    double csect , subshell_csect ;
    int iish ;
    csect = 0 ;
    for ( iish = 0 ; iish <= g_atomic_data[atomid].atomic_econfig->max_orbit ; iish++ )
    {
        subshell_csect = eimpion_sigma_orbit_as_KE( atomid , iish , a_E ) ;
        csect += subshell_csect ; 
    }
    return( csect ) ;
}





//############################## FUNCTION #################################

double eimpion_effective_KE( int actatom , double *re , double *ve ) {
  double dr[3] , pot2 ;
  double T ;

  double q_el = -1 ;

  double (*sys_r)[3] = g_system.r1 ;
  int (*sys_qi)   = g_system.qi ;


  T = 0 ;

  if ( strcmp( g_inputs.si_type , "localold" ) == 0 ) 
  {
    T = .5 * g_phys_math_const.me  *vectnorm2(ve) ;
  }

  if ( strcmp( g_inputs.si_type , "1atom" ) == 0 ) 
  {
    vectminus( dr , re , sys_r[actatom] ) ;
    T = .5* g_phys_math_const.me * vectnorm2(ve) +
         enerregcut( sqrt( vectnorm2(dr) ) , 
                     sys_qi[actatom] , q_el , actatom ) ;
  }

   if ( strcmp( g_inputs.si_type , "2atom" ) == 0 ) 
   {
     printf( "si_type = 2atom is off\n" ) ; exit(-1) ;
     return(0) ;
  }

  return( T ) ;
}





//############################## FUNCTION #################################

int  eimpion_screening_nbor_list( double a_nbor_dist , double *ret_nborlist , int *ret_nbornum , int *ret_max_numnbors ) 
{
    static int init=1 , *nbornum ;
    static double *nborlist ;
    int max_numnbors = 20 , ii , jj ;
    double cutoff2  , dr[3] ;

    if ( a_nbor_dist < 0 ) 
    {
        ret_nborlist = nborlist ;
        ret_nbornum  = nbornum  ;
        *ret_max_numnbors = max_numnbors ;
        return  0;
    }

    cutoff2 = a_nbor_dist * a_nbor_dist ;
    
    if ( init == 1 ) 
    {
        nborlist = calloc(  g_inputs.N * (max_numnbors+1) , sizeof(int) ) ;
        nbornum  = calloc(  g_inputs.N , sizeof(int) ) ;
        init = 0 ;
    }

    for ( ii = 0 ; ii <  g_inputs.N                    ; ii++ )  { nbornum[ii]    = 0  ; }
    for ( ii = 0 ; ii <  g_inputs.N * (max_numnbors+1) ; ii++ )  { nborlist[ ii ] = -1 ; }

    for ( ii = 0 ; ii <  g_inputs.N-1 ; ii++ )
    {
        for ( jj = ii+1 ;  jj <  g_inputs.N ; jj++ )
        {
            vectminus( dr ,  g_system.r1 [ii] ,  g_system.r1 [jj] ) ;
            if ( vectnorm2( dr ) < cutoff2 )
            {
                if( ii != jj )
                {
                    if ( nbornum[ii] < max_numnbors )
                    {
                        nborlist[ ii*max_numnbors + nbornum[ii] ] = jj ;
                        nbornum[ii]++;
                    }
                    else { fprintf( stderr , "Too many neighbors in routine eimpion_screening_nbor_list: %d!\n" , ii ) ; }
                    if ( nbornum[jj] < max_numnbors )
                    {
                        nborlist[ jj*max_numnbors + nbornum[jj] ] = ii ;
                        nbornum[jj]++;
                    }
                    else { fprintf( stderr , "Too many neighbors in routine eimpion_screening_nbor_list: %d!\n" , jj ) ; }
                }
            }
        }
    }

    ret_nborlist = nborlist ;
    ret_nbornum  = nbornum  ;
    *ret_max_numnbors = max_numnbors ;


  return 0 ;
}





//############################## FUNCTION #################################

int get_my_neighbors( int a_atom , double a_dist , int *Nnbors  , int *my_newid , int **i_out , double **r_out )
{
    int  ii , cc , *i_nbor ;
    double (*r_nbor)[3] , dr[3] , cutoff2  ;

    cutoff2 = a_dist * a_dist ;

    *Nnbors = 0 ;
    for ( ii = 0 ; ii <  g_inputs.N ; ii++ ) 
    {
        vectminus( dr ,  g_system.r1 [ii] ,  g_system.r1 [a_atom] ) ;
        if ( vectnorm2( dr ) < cutoff2 ) { (*Nnbors)++ ; }
    }

    i_nbor = calloc( (*Nnbors) * 1 , sizeof( int    ) ) ;
    r_nbor = calloc( (*Nnbors) * 3 , sizeof( double ) ) ;

    cc = 0 ;
    for ( ii = 0 ; ii <  g_inputs.N ; ii++ ) 
    {
            vectminus( dr ,  g_system.r1 [ii] ,  g_system.r1 [a_atom] ) ;
            if ( vectnorm2( dr ) < cutoff2 ) 
            {
                if ( ii == a_atom ) { *my_newid = cc ; } 
                i_nbor[cc]    = ii ;
                r_nbor[cc][0] =  g_system.r1 [ii][0] ;
                r_nbor[cc][1] =  g_system.r1 [ii][1] ;
                r_nbor[cc][2] =  g_system.r1 [ii][2] ;
                cc++ ;
            }
    }

    *r_out = (double*)r_nbor ;
    *i_out = i_nbor ;

    return(0) ;

}



//############################## FUNCTION #################################

double eimpion_sigma_total_in_system( int actatom , double *re , double *ve )
{
  int i;
  double siginel , T ;

  T = eimpion_effective_KE( actatom , re , ve ) ;

  siginel = 0 ;
  for ( i = 1 ; i <= g_atomic_data[actatom].atomic_econfig->max_orbit ; ++i )
  {
    siginel += eimpion_sigma_orbit_as_KE( actatom , i , T  ) ;
  }


  return(siginel);
}






//############################## FUNCTION #################################

double inelastsigma_atomic_max(double T,int Z) {
  int sh, NE ;
  double sigma, Q, n;
  int max_orbit , *occupancy ;
  double *B, *V ;

  //WARNING_WORKAROUND_TOBE_REMOVED( "FIXME HERE." ) ;
  //return( 10e-10 * 10e-10 * PI ) ;

  xmdyn_get_econfig_properties( "Z", Z , 8 , 
                                "max_orbit" , &max_orbit ,
                                "occupancy" , &occupancy ,
                                "B" , &B ,
                                "V" , &V 
                              ) ;

  sigma=0;
  for ( sh = 0 ; sh < max_orbit + 1 ; sh++ )
  {
      NE = occupancy[sh] ;
      if ( 0 < NE ) 
      {
        Q = 1;
        n = 1;
        sigma += inelastsigma_Lotz( T , B[sh] , V[sh] , NE , Q , n ) ;
      }
  }
  return(sigma);


}






//############################## FUNCTION #################################

int try_si_effect( int atom , double *re , double *ve , double totalb2 , double dist2 ) 
 {

  double (*sys_r)[3] = g_system.r1 ;
  double (*sys_v)[3] = g_system.v1 ;
  int *sys_qi = g_system.qi ;

  int q_el = -1 ;

  double b2,dr[3],v[3],L[3],rsq,vsq,Lsq,vinfsq, vinfsq_mod , pot2 ;
  //double vv,vv1,vv2;
  

  if ( strcmp( g_inputs.si_type , "localold" ) == 0 ) 
  {
    if ( dist2 <= totalb2 ) 
    {
      return(1) ;
    }
    else 
    {
      return(0) ;
    }
  }
  
  if ( strcmp( g_inputs.si_type , "1atom" ) == 0 ) 
  {

    vectminus( dr , re , sys_r[atom] ) ; rsq = vectnorm2(dr);
    vectminus( v,   ve , sys_v[atom] ) ; vsq = vectnorm2(v);
    crossproduct( L , dr , v ) ; 
    Lsq = vectnorm2(L) ;
    vinfsq = 2.0 / g_phys_math_const.me  * 
             ( .5 * g_phys_math_const.me * vsq + 
               enerregcut( sqrt( rsq ) , sys_qi[atom] , q_el , atom ) 
             ) ;
    
    b2 = Lsq / vinfsq ;

    if ( ( vinfsq > 0 ) && ( b2 < totalb2 ) ) 
    {
      return(1) ;
    }
    else 
    {
      return(0) ;
    }
  }


  fprintf( stderr , "\n!!! You should not be here in function %s , id: 1 !!!\n\n" , __FUNCTION_NAME__  ) ; exit(-1) ;
  return 0 ;

}



// **************************************************************


struct secel_chain_struct {
  struct secel_chain_struct *prev;
  struct secel_chain_struct *next;
  int atom;
  int el;
  int dim;
}; 





//############################## FUNCTION #################################

int* get_free_electrons(int *numel,double maxdist){
  int i,c,*free_el;
  double d[3];
  *numel=0;
  
  for (i= g_inputs.N;i< g_inputs.N+ g_inputs.E;++i) {
    vectminus(d, g_system.r1 [i], g_system.r2 [i]);
    if ( (isitfreeelectron(i)) && 
         ( g_system.ptype [i]!=  g_inputs.nosec ) && ( g_system.ptype [i]!=  g_inputs.noall) &&
	 (vectnorm2(d)!=0) && 
	 ( (maxdist==0) || (insidecluster(i,maxdist))) ) {
      (*numel)++;
    }
  }
  if (*numel==0) {
    free_el=NULL;
    return(free_el);
  }
  
  free_el=calloc(*numel,sizeof(int));
  c=0;
  for (i= g_inputs.N;i< g_inputs.N+ g_inputs.E;++i) {
    vectminus(d, g_system.r1 [i], g_system.r2 [i]);
    if ( (isitfreeelectron(i)) && 
         ( g_system.ptype [i]!=  g_inputs.nosec) && ( g_system.ptype [i]!=  g_inputs.noall) &&
	 (vectnorm2(d)!=0) && 
	 ( (maxdist==0) || (insidecluster(i,maxdist))) ) {
      free_el[c]=i;
      c++;
    }
  }
  return(free_el);
}




//############################## FUNCTION #################################

int get_secel_limits(int *from_atom,int *to_atom,int *from_el,int *to_el,
                      int numel){
  int np,crank,atomevents,maxevents,x;
  int i,c;
  static int fa,ta,fe,te,f,t;
  //static char method[LSTRLTH]="exact"; // "trivial"  "exact"
  static char method[LSTRLTH]="trivial"; // "trivial"  "exact"
  int My_Rank = 0 , Root_Rank = 0 , Num_Rank = 1 ;
  
  np=Num_Rank-1;
  crank=My_Rank-1;
  
  if (numel>0){
    ;
  }
  
  maxevents=0;atomevents=0;
  for (i=0;i< g_inputs.N;++i) {
    atomevents += g_atomic_data[i].atomic_econfig->max_orbit ;
  }
  maxevents = atomevents * numel ;

  c=0;
  if (My_Rank==Root_Rank) {
    if (Num_Rank==1) {
      fe=0;
      fa=0;
      te=numel;
      ta= g_inputs.N;
    }
    else {
      fa=0;
      ta=0;
      fe=0;
      te=0;
    }
  }
  else {
    if (My_Rank==1)  {
      fe=0; fa=0;
    }
    else {
    }
    
    // atoms are equal:
    if ( (strcmp(method,"trivial")==0) && (My_Rank!=Root_Rank) ) {
      f = round(numel* g_inputs.N/(double)np*(My_Rank-1));
      t = round(numel* g_inputs.N/(double)np*(My_Rank))-1;
      x = t-f+1;
      fe=floor(f/(double) g_inputs.N);
      fa=f- g_inputs.N*fe;
      te=floor(t/(double) g_inputs.N);
      ta=t- g_inputs.N*te;
      //
      fa=0;
      ta= g_inputs.N;
      fe=floor(numel/(double)np*crank);
      te=floor(numel/(double)np*(crank+1));
    }
  }
  
  *from_el=fe;
  *from_atom=fa;
  *to_el=te;
  *to_atom=ta;
  //printf("dbg_ %d %d %d %d %d %d\n",My_Rank,numel,fe,te,fa,ta); // LIMITS
  
  return(0);
}



// #####################################################################
// #####################################################################
// #####################################################################

//############################## FUNCTION #################################

int get_center_of_mass( double *R,   double (*r)[3] , double *m ,int dim )
{
    int ii , jj ;

    double M = dim ;
    if ( NULL != m ) { M = 0 ; for ( ii = 0 ; ii < dim ; ii++ ) { M += m[ii] ; } }

    R[0] = R[1] = R[2] = 0 ;

    for ( ii = 0 ; ii < dim ; ii++ ) { for ( jj = 0 ; jj < 3 ; jj++ ) {
        R[jj] += ( NULL == m ? 1 : m[jj] ) * r[ii][jj] ;
    }}

    for ( jj = 0 ; jj < 3 ; jj++ ) { R[jj] /= M ; } 
    return 0 ;
}


//############################## FUNCTION #################################

double get_cluster_radius( double *Rcom , double (*r)[3] , int N )
{
    double r_com[3] , r_diff[3] ;
    double R2 = 0 ;
    int ii ;
    get_center_of_mass( r_com , r , NULL , N ) ;
    for ( ii = 0 ; ii < N ; ii++ )
    {
        vectndiff( r_diff , r[ii] , r_com , 3 ) ;
        if ( R2 < vectnorm2( r_diff ) ) { R2 = vectnorm2( r_diff ) ; }
    }
    if ( NULL != Rcom ) { for ( ii=0 ; ii<3 ; ii++) { Rcom[ii] = r_com[ii] ; } }
    return( sqrt( R2 ) ) ;
}


//############################## FUNCTION #################################

int do_select_electrons_for_secel( int **e_selected , int *N_selected , 
                                   double *dist_cl ,
                                   int N_a , double (*r1_a)[3] , double (*r2_a)[3] , 
                                   int N_e , double (*r1_e)[3] , double (*r2_e)[3]  )
{
    int ii , jj ;
    double Rcluster , r_com[3] , re_cl[3] , dr[3] ;

    *e_selected = XMDYN_ALLOCATE_MEM( N_e * sizeof(int) ) ;
    *N_selected = 0 ;

    // get electrons inside the cluster
    get_center_of_mass( r_com , r1_a , NULL , N_a ) ;
    Rcluster = get_cluster_radius( r_com , r1_a , N_a ) ; 
    for ( ii = 0 ; ii < N_e ; ii++ )
    {
        vectminus( re_cl , r1_e[ii] , r_com );
        vectminus( dr    , r1_e[ii] , r2_e[ii] );
        // TO_BE_OPTIM
        //if ( vectnorm2(re_cl) < (Rcluster + dist[ii]) * (Rcluster + dist[ii] ) ) 
        if ( ( sqrt(vectnorm2(re_cl)) < Rcluster + dist_cl[ii] ) &&
             ( vectnorm2( dr ) != 0 )
           )
        {
            (*e_selected)[*N_selected] = ii ;   (*N_selected)++ ;
        }
     }

    return 0 ;
}


//############################## FUNCTION #################################

int get_secel_maxsigma2( double *maxs2 , 
                         int numel , int *Ztypes , 
                         double (*re1)[3] , double (*re2)[3] , 
                         double (*ve1)[3] , double (*ve2)[3] , 
                         double (*ra1)[3] , double (*ra2)[3] , 
                         double (*va1)[3] , double (*va2)[3]  
                         )
{

  // GLOBALS:
  int maxZ = maxZvalue ;

  int ii , jj , selZ ;
  double *vel2 , *vel1 , vv1 , vv2 , T1 , T2 , ss1 , ss2 , es1 , es2 , newsigma ;

  for ( ii = 0 ; ii < numel ; ++ii ) 
  {
    vel2 =  ve2[ii];
    vel1 =  ve1[ii];
    vv2 = vectnorm2( vel2 ) ;
    vv1 = vectnorm2( vel1 ) ;
    T1 = 0.5 * g_phys_math_const.me * vv1 ; 
    T2 = 0.5 * g_phys_math_const.me * vv2 ;
    for ( selZ = 1 ; selZ <= maxZ ; ++selZ ) 
    {
      if ( Ztypes[selZ] > 0 ) 
      {
        ss1 = inelastsigma_atomic_max( T1 , selZ ) ;
        ss2 = inelastsigma_atomic_max( T2 , selZ ) ;
        es1 = elastsigma_atomic_max( T1 , selZ ) ;
        es2 = elastsigma_atomic_max( T2 , selZ ) ;
	newsigma = max( ss1 , ss2 ) + max( es1 , es2 ) ;
	if ( maxs2[ii] < newsigma / PI ) { maxs2[ii] = newsigma / PI ; }
      }
    }
  }

  return 0 ;

}


//############################## FUNCTION #################################

int build_secel_raw_table( int (*table)[SECEL_TABLE_DIM] , int numel , int N , double *r2limit ,
                           double (*xra)[3] , double (*xr1)[3] , double (*xr2)[3]
                         )
{

{
  int tt = SECEL_TABLE_DIM - 1 ;
  int ii , jj , cc ;
  double sig2 , *re1 , *re2 , rd[3] , *ra , R1[3] ;
  double n2rd , rmax , R1d , R2d , dist2  ;

  for ( ii = 0 ; ii < numel ; ++ii ) 
  //for (ii=from_el;ii<to_el;++ii) 
  {
     double rd[3] , R1[3] , R2[3] ;
     //while (1) { ///////
     cc = 0;
     sig2 = r2limit[ii];
     re1 = xr1[ii];
     re2 = xr2[ii];
     vectminus( rd , re1 , re2 ) ;
     n2rd = vectnorm2( rd ) ; 
     rmax = ( sig2 + n2rd ) ;
     for ( jj = 0 ; jj < N ; ++jj ) 
     {
        ra =  xra[jj] ; 
        vectminus( R1 , ra , re1 ) ;
        if ( vectnorm2( R1 ) <= rmax )
        {
            R1d = scalarproduct( R1 , rd ) ;
            if ( dist2 = vectnorm2(R1) - R1d * R1d / n2rd , dist2 < sig2 )
            {
              vectminus( R2 , ra , re2 ) ;
              R2d = scalarproduct( R2 , rd ) ;
              if ( R1d * R2d <= 0 )
              {
                table[ii][(cc)%(tt)] = jj ;
                cc++;
              }
            }
        }
     }
     table[ii][SECEL_TABLE_DIM-1] = cc ;
     //} ///////
  }

}


    return 0 ;

}





//############################## FUNCTION #################################

int xmdyn_preselect_electrons_for_secel( int (**a_table)[SECEL_TABLE_DIM], int **a_free_el , int *a_numel ,
                                         double (**a_r1e)[3] , double (**a_r2e)[3] , 
                                         double (**a_v1e)[3] , double (**a_v2e)[3] )
// output: r1e, r2e,v1e,v2e
{

    // GLOBAL -> LOCAL :
    int *sysZ = g_system.Z ;
    int N = g_inputs.N ;
    double (*r1)[3] = g_system.r1 ;
    double (*r2)[3] = g_system.r2 ;
    double (*v1)[3] = g_system.v1 ;
    double (*v2)[3] = g_system.v2 ;
    int maxZ = maxZvalue ;

    static int *Ztypes;
    static int init = 1 ;

    double (*r1e)[3] , (*r2e)[3] ;

    double *dist_cl ;
    int ii , jj ;
    double (*v1e)[3] , (*v2e)[3] ;

    double *maxs2 ;
    int nume , *list_class_el ;
   
    if (init)
    {
        Ztypes = XMDYN_ALLOCATE_MEM( (maxZ+1) * sizeof(int) ) ; //init; not deallocated
        for ( ii = 0 ; ii < N ; ++ii ) { Ztypes[sysZ[ii]] = 1 ; }
        init=0;
    }

    // Allocated inside: list_class_el
    xmdyn_get_classical_electrons( &nume , &list_class_el ) ; 

    if ( 0 == nume ) { *a_numel = 0 ; return 0 ; } 

    r1e = XMDYN_ALLOCATE_MEM( nume * 3 * sizeof(double) ) ;
    r2e = XMDYN_ALLOCATE_MEM( nume * 3 * sizeof(double) ) ;
    v1e = XMDYN_ALLOCATE_MEM( nume * 3 * sizeof(double) ) ;
    v2e = XMDYN_ALLOCATE_MEM( nume * 3 * sizeof(double) ) ;
    for ( ii = 0 ; ii < nume ; ii++ ) {  for ( jj = 0 ; jj < 3 ; jj++ ) {
        r1e[ii][jj] = r1[list_class_el[ii]][jj] ;
        r2e[ii][jj] = r2[list_class_el[ii]][jj] ;
        v1e[ii][jj] = v1[list_class_el[ii]][jj] ;
        v2e[ii][jj] = v2[list_class_el[ii]][jj] ;
    }}

    {  // DEBUGGING BLOCK
        int (*dbg_p)[nume] =  (int (*)[nume])list_class_el ;
        nume *= 1 ;
    }

    {
        int *e_selected, N_selected ;

        // TODO: electron (KE) dependent cluster_dist ; though sqrt(crossection/pi) is typically much less than 10 Angstrom
        dist_cl = XMDYN_ALLOCATE_MEM( nume * sizeof(double) ) ;
        for ( ii = 0 ; ii < nume ; ii++ ) { dist_cl[ii] = 10e-10 * 1.0001 ; }

        // Allocated inside: e_selected
        do_select_electrons_for_secel( &e_selected , &N_selected , 
                                       dist_cl , N , r1 , r2 , nume , r1e , r2e ) ;

        XMDYN_DEALLOCATE_MEM( dist_cl ) ;

        for ( ii = 0 ; ii < N_selected ; ii++ ) 
        {  
            for ( jj = 0 ; jj < 3 ; jj++ ) {
                r1e[ii][jj] = r1e[e_selected[ii]][jj] ; r2e[ii][jj] = r2e[e_selected[ii]][jj] ;
                v1e[ii][jj] = v1e[e_selected[ii]][jj] ; v2e[ii][jj] = v2e[e_selected[ii]][jj] ;
            }
            list_class_el[ii] = list_class_el[e_selected[ii]] ;
        }
        XMDYN_DEALLOCATE_MEM( e_selected ) ; 
        nume = N_selected ;

        if ( 0 == nume ) 
        {
            *a_numel = nume ;   
            XMDYN_DEALLOCATE_MEM(r1e) ; XMDYN_DEALLOCATE_MEM(r2e) ; 
            XMDYN_DEALLOCATE_MEM(v1e) ; XMDYN_DEALLOCATE_MEM(v2e) ; XMDYN_DEALLOCATE_MEM(list_class_el) ;
            return 0 ;
        }

        maxs2 = XMDYN_ALLOCATE_MEM( nume * sizeof(double) ) ;
        get_secel_maxsigma2(  maxs2 , nume , Ztypes , 
                              r1e , r2e , v1e , v2e , r1 , r2 , v1 , v2 ) ;

    }
  
      
    int (*table)[SECEL_TABLE_DIM]  ;
    table = XMDYN_ALLOCATE_MEM( nume * SECEL_TABLE_DIM * sizeof(int) ) ;

    build_secel_raw_table( table , nume , N , maxs2 , r1 , r1e , r2e ) ;

    XMDYN_DEALLOCATE_MEM( maxs2 ) ;
 
    *a_table = table ;
    *a_free_el = list_class_el ;
    *a_numel = nume ;
    *a_r1e = r1e ;
    *a_r2e = r2e ;
    *a_v1e = v1e ;
    *a_v2e = v2e ;

    return 0 ;
}



//############################## FUNCTION #################################

int electronscattering_02 (int step ) {

  int i,j,icounter , ii_mess ;
  double *rel1,*rel2,*vel1,*vel2,*rat,d[3],Rr1[3];
  double elastb2,inelastb2,totalb2,dist2,sp_Rr1d;
  
  int *free_el,numel,actatom,actel;
  static int init=1;
  int from_atom,to_atom,from_el,to_el;
  
  int N,E;
  N= g_inputs.N; E= g_inputs.E;

  static int *Ztypes;

  double maxr2,sig2,vv1,vv2,T1,T2,ss1,ss2,es1,es2,newsigma;
  double *r2limit;
  double *re1,*re2,n2rd,rmax,*ra;
  //double rd[3],R1[3],R2[3];
  double R1d,R2d;
  int ii,jj,tt,c;
  int *multiscatt_table,(*table)[SECEL_TABLE_DIM];
  double (*xr1)[3],(*xr2)[3],(*xra)[3];
  int *xq,*xZ;
  int My_Rank=0 , Event_Rank = 0 ;

  double (*r1e)[3] , (*r2e)[3] , (*v1e)[3] , (*v2e)[3] ;


    if  ( (strcmp( g_inputs.elscattOK,"y")!=0) && 
        (strcmp( g_inputs.elimpionOK,"y")!=0)   )
    return(0);

    // ALLOCATED INSIDE: table, free_el , r1e, r2e , v1e , v2e
    xmdyn_preselect_electrons_for_secel( &table, &free_el , &numel , &r1e, &r2e , &v1e , &v2e ) ;

    if ( 0 == numel ) { return( 0 ) ; } 


//#############  SELECTING ELECTRONS: 2nd ROUND ##################

//printf("XXX01 %d %d\n", N,numel);

  multiscatt_table = XMDYN_ALLOCATE_MEM( N * sizeof(int) ) ;

  if (My_Rank==Event_Rank) 
  {
    int try_result;
    for (i=0;i<numel;++i)
    {
      if ( SECEL_TABLE_DIM - 1 < table[i][SECEL_TABLE_DIM-1] )
      {
        actel = free_el[i];
        int sstep = 0 ;
        print_log( "+Ms:" , NULL , "str" , "noflush" ) ;
        print_log( " step:" , (void*) &step  , "int" , "noflush" ) ;
        print_log( " sstep:" , (void*) &sstep  , "int" , "noflush" ) ;
        print_log( " el:" , (void*) &actel  , "int" , "noflush" ) ;
        print_log( " Ms:" , (void*) &(table[i][SECEL_TABLE_DIM-1])  , "int" , "noflush" ) ;
        print_log( " \n" , NULL , "" , "flush" ) ;
        table[i][SECEL_TABLE_DIM-1] = SECEL_TABLE_DIM - 1 ;
      }
    }
    for (ii=0;ii<numel;++ii) 
    {
      for (jj=0;jj<table[ii][SECEL_TABLE_DIM-1];++jj) 
      {
        actel = free_el[ii] ;
        actatom = table[ii][jj] ;
        rel1 = r1e[ii]; rel2 = r2e[ii];
        vectminus( d , rel1 , rel2 ) ;
        rat = g_system.r1[actatom] ;
        vectminus( Rr1 , rat , rel1 ) ;
        sp_Rr1d = scalarproduct( Rr1 , d ) ;
        dist2 = vectnorm2( Rr1 ) - sp_Rr1d * sp_Rr1d / vectnorm2(d) ;
        //
        //elastb2 = elastb2calc( actatom , actel ) ;
        // NOTE: External (cross-section based) elastic e-ion scattering is off
        elastb2 = 0 ;
        inelastb2 = eimpion_sigma_total_in_system( actatom , rel1 , v1e[ii] ) / PI ;
        totalb2 = inelastb2 + elastb2 ;
        //
//	printf("XXXXX %e %e \n",totalb2,dist2);
        try_result = try_si_effect( actatom , rel1 , v1e[ii] , totalb2 , dist2 ) ;
        if (try_result == 1 ) 
	{
          if (multiscatt_table[actatom]==0) 
	  {
            scatter( actatom , actel , step , rel1 , rel2 , v1e[ii] , v2e[ii] ) ;
	    multiscatt_table[actatom]++;
          }
          else 
	  {
	    ii_mess = 0 ;
          }
        }
      }
    }  
  }
  


  if (numel>0)
  {
//    free(xra);
//    free(xZ);
//    free(xr1);
//    free(xr2);
//    free(xq);
//    free(r2limit);
    XMDYN_DEALLOCATE_MEM(free_el);
    XMDYN_DEALLOCATE_MEM(table);
    XMDYN_DEALLOCATE_MEM(multiscatt_table);
    XMDYN_DEALLOCATE_MEM(r1e) ; XMDYN_DEALLOCATE_MEM(r2e) ; 
    XMDYN_DEALLOCATE_MEM(v1e) ; XMDYN_DEALLOCATE_MEM(v2e) ;
  }  

  return( 0 ) ;

}


//############################## FUNCTION #################################

int  electronscattering(int step ){
  return( electronscattering_02(step ) ) ;
}


#undef MODULE_DEBUG


