///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>

#include "common.h"

#include "globalconsts.h"

#include "xmdyn_io.h"
#include "xmdyn_filenames.h"


static struct allfilenamesstruct ALLFILES ;

static struct save_prec_struct  SAVE_PREC ;


//############################## FUNCTION #################################

int xmdyn_filenames_load( struct allfilenamesstruct **allfiles , struct save_prec_struct **save_prec ) 
{
    SAVE_PREC.r=5;
    SAVE_PREC.E=18;
    SAVE_PREC.foto_r=12;
    SAVE_PREC.foto_E=18;
  
    *allfiles = &ALLFILES;
    *save_prec = &SAVE_PREC ;
    return( 0 ) ;
}




//############################## FUNCTION #################################

int cache_filenames(int s,int step) {
  static char root[MSTRLTH],cache[MSTRLTH],rootdatadir[MSTRLTH],cachedatadir[MSTRLTH],app[MSTRLTH];
  static int init=0, num=0,counter=0,tmpint1,tmpint2;
  char *p,comm[LSTRLTH];
  double *par;
  FILE *fp;
  int N,E;
  static double time, maxcachetime=0.5;  // hour(s) of max cache time

  maxcachetime= g_inputs.maxcachetime*3600;

  if (strcmp( g_inputs.cachedatadir,"-")==0)
    return(0);
  
  N= g_inputs.N; E= g_inputs.E;
  
  if (init==0) {
    strcpy(root, g_inputs.rootdir);
    strcpy(rootdatadir,ALLFILES.datadir);
    strcpy(cache, g_inputs.cachedatadir);
    p=strchr(cache,'_'); *p='\0';
    par=load_input_pars( g_inputs.cachedatadir,1);
    num=(int)par[0];
    load_input_pars( g_inputs.cachedatadir,-1);
    tmpint1 = N*3 + (8+SAVE_PREC.foto_E)*10 + (N+E)*2 + (N+E)*2 + (N+E)*4 +
           // Z     ENER                       et       farel      q
	      (N+E)*13*(8+SAVE_PREC.foto_r+sizeof(double));
	   // m+3*r1+3*r2+3*v1+3*v2
    tmpint1 *=2;
    tmpint2 += (N+E)*2 + (N+E)*4 + (N+E)*11 + 
            // et         q        m
               ( g_inputs.savev3=='y'?7:4)*(N+E)*(8+SAVE_PREC.r);
	   // r,v,v3(if needed)  
    num*=1024*1024;
    num-=tmpint1;
    num/=tmpint2;

    strcpy(app,"");
    strcat(strcat(app,ALLFILES.datadir),"/app.sh");
    fp=fopen(app,"w");
    fprintf(fp," \
#!/bin/bash\n\
\n\
S=\"$1\"\n\
T=\"$2\"\n\
\n\
C=`pwd`\n\
cd \"$T\"\n\
T=`pwd`\n\
cd \"$C\"; cd \"$S\"\n\
\n\
find -type d | while read x ; do\n\
  mkdir -p \"$T/$x\"\n\
done\n\
\n\
find -type f | grep -v \"./foto/\" | while read x ; do\n\
  cat \"$x\" >> \"$T/$x\"\n\
done\n\
cp ./foto/* \"$T/foto/\" \n\
    ");
    fclose(fp);
    strcpy(comm,"");
    strcat(strcat(comm,"chmod +x "),app);
    system(comm);
    time=myclock(-1);
    init=1;
  }
  
  if (s==20) {
    strcpy( g_inputs.rootdir,cache);
    filenames();
    strcpy(cachedatadir,ALLFILES.datadir);
    strcpy(comm,"");
    strcat(strcat(comm,"rm -rf "),cachedatadir);
    system(comm);
    dirhiercreate();
    strcpy( g_inputs.rootdir,root);
    filenames();
    return(0);
  }
 
  if ( (s==0) || (s==10) ) 
  {
    strcpy( g_inputs.rootdir,cache);
    filenames();
  }

  if ( (s==1) ||(s==11) ) {
    strcpy(cachedatadir,ALLFILES.datadir);
    if ( (num==0) || ((counter+1)%num==0) || (s==11) ||
         (myclock(-1)-time>maxcachetime) )  {
      time = myclock(-1);
      strcpy(comm,"");
      strcat(strcat(strcat(strcat(strcat(comm,app)," "),cachedatadir)," "),rootdatadir);
      system(comm);
      strcpy(comm,"");
      strcat(strcat(comm,"rm -rf "),cachedatadir);
      system(comm);
      dirhiercreate();
    }
    counter++;
  }

  if ( (s==2) || (s==12) ) {
    strcpy( g_inputs.rootdir,root);
    filenames();
  }

  if (s==13)  {
    strcpy(comm,"");
    strcat(strcat(comm,"rm -rf "),cachedatadir);
    system(comm);
  }
    

  return 0 ;
}





//############################## FUNCTION #################################

void filenames(void){
  char mutual[MSTRLTH];

  clearstr(ALLFILES.datadir);
  clearstr(mutual);

  if ( 0 == strcmp( g_inputs.format_loadsave , "000" ) )
  {
      strcat(strcat(strcat(strcat(ALLFILES.datadir, g_inputs.rootdir),"/data/"), g_inputs.runnumber),"/");
      strcat(strcat(strcat(mutual,ALLFILES.datadir),"bio"), g_inputs.runnumber);
  }
  else
  {
      strcat(strcat(ALLFILES.datadir, g_inputs.datadir) , "/");
      strcat(mutual,ALLFILES.datadir);
  }

  clearstr(ALLFILES.fotodir);
  clearstr(ALLFILES.fotostartdir);
  clearstr(ALLFILES.archivedir);
  clearstr(ALLFILES.tempdir);
  strcat(strcat(ALLFILES.fotodir,ALLFILES.datadir),"foto/");
  strcat(strcat(ALLFILES.fotostartdir,ALLFILES.datadir),"fotostart/");
  strcat(strcat(ALLFILES.archivedir,ALLFILES.datadir),"");
  strcat(strcat(ALLFILES.tempdir,ALLFILES.datadir),"temp/");

  clearstr(ALLFILES.dbase_overwrite);
  clearstr(ALLFILES.gopt);
  clearstr(ALLFILES.snptimes);
  clearstr(ALLFILES.summaryfile);
  clearstr(ALLFILES.data2diskindicator);
  clearstr(ALLFILES.snptimes);
  clearstr(ALLFILES.uid);
  clearstr(ALLFILES.dimer);
  clearstr(ALLFILES.r);
  clearstr(ALLFILES.v);
  clearstr(ALLFILES.v3);
  clearstr(ALLFILES.q);
  clearstr(ALLFILES.et);
  clearstr(ALLFILES.m);
  clearstr(ALLFILES.Eintotal);
  clearstr(ALLFILES.infofile);
  clearstr(ALLFILES.progress_file);
  clearstr(ALLFILES.params);
  clearstr(ALLFILES.beamfile);
  clearstr(ALLFILES.dbg1);
  clearstr(ALLFILES.ireclog);
  clearstr(ALLFILES.recomb_snpfile);
  strcat(strcat(ALLFILES.summaryfile,ALLFILES.archivedir),"/xparams.txt");
  strcat(strcat(strcat(strcat(ALLFILES.data2diskindicator,ALLFILES.datadir),"foto/foto"), g_inputs.runnumber),"zindicator.txt");
  strcat(strcat(ALLFILES.gopt,mutual),"gopt.txt");
  strcat(strcat(ALLFILES.snptimes,mutual),"snp_times.txt");
  strcat(strcat(ALLFILES.uid,mutual),"uid.txt");
  strcat(strcat(ALLFILES.dimer,mutual),"dimer.txt");
  strcat(strcat(ALLFILES.r,mutual),"r.txt");
  strcat(strcat(ALLFILES.v,mutual),"v.txt");
  strcat(strcat(ALLFILES.v3,mutual),"v3.txt");
  strcat(strcat(ALLFILES.q,mutual),"q.txt");
  strcat(strcat(ALLFILES.et,mutual),"et.txt");
  strcat(strcat(ALLFILES.recomb_snpfile,mutual),"trap.txt");
  strcat(strcat(ALLFILES.m,mutual),"m.txt");
  strcat(strcat(ALLFILES.Eintotal,mutual),"Etot.txt");
  strcat(strcat(ALLFILES.infofile,mutual),"zinfo.txt");
  if ( 0 == strcmp( g_inputs.format_loadsave , "000" ) ) 
      { strcat(strcat(ALLFILES.progress_file,mutual),"hol.txt") ; }
  else
      { strcat(strcat(ALLFILES.progress_file,mutual),"progress.log") ; }
  strcat(strcat(ALLFILES.params,mutual),"zpar.txt");
  strcat(strcat(ALLFILES.beamfile,mutual),"beam.txt");
  strcat(strcat(ALLFILES.dbg1,mutual),"dbg1.txt");
  strcat(strcat(ALLFILES.ireclog,mutual),"irlog.txt");

  strcat(strcat(ALLFILES.dbase_overwrite,mutual),"dbase.dat");

  fotofilenames(0);
  strcpy(ALLFILES.r1input,ALLFILES.r1foto);
  strcpy(ALLFILES.v1input,ALLFILES.v1foto);
  strcpy(ALLFILES.etypeinput,ALLFILES.etypefoto);
  strcpy(ALLFILES.qinput,ALLFILES.chargefoto);
  strcpy(ALLFILES.minput,ALLFILES.massfoto);

}






//############################## FUNCTION #################################

void fotofilenames(int step){
  char mutual[MSTRLTH]="";
  char stepstr[MSTRLTH]="";

  char sample_dir[LSTRLTH] ;

  strcpy( sample_dir , g_inputs.sample_dir ) ;

  strcpy(ALLFILES.uidfoto,"");
  strcpy(ALLFILES.dimerfoto,"");
  strcpy(ALLFILES.r1foto,"");
  strcpy(ALLFILES.v1foto,"");
  strcpy(ALLFILES.r2foto,"");
  strcpy(ALLFILES.prev_r2foto,"");
  strcpy(ALLFILES.v2foto,"");
  strcpy(ALLFILES.chargefoto,"");
  strcpy(ALLFILES.massfoto,"");
  strcpy(ALLFILES.etypefoto,"");
  strcpy(ALLFILES.stepfoto,"");
  strcpy(ALLFILES.energyfoto,"");
  strcpy(ALLFILES.farelfoto,"");
  strcpy(ALLFILES.randcallsfoto,"");
  strcpy(ALLFILES.energyfoto,"");
  strcpy(ALLFILES.timefoto,"");
  strcpy(ALLFILES.el_configfoto,"");
  strcpy(ALLFILES.Zallfoto,"");
  strcpy(ALLFILES.motherionfoto,"");

  strcpy(ALLFILES.r1fotob,"");
  strcpy(ALLFILES.v1fotob,"");
  strcpy(ALLFILES.r2fotob,"");
  strcpy(ALLFILES.prev_r2fotob,"");
  strcpy(ALLFILES.v2fotob,"");
  strcpy(ALLFILES.massfotob,"");
  strcpy(ALLFILES.energyfotob,"");
  
  strcpy(ALLFILES.instrecfotob,"");
  strcpy(ALLFILES.instrecfotol,"");

  strcpy(ALLFILES.chargetransfer_b,"");

  strcpy(ALLFILES.dbase_overwrite,"");

  strcpy(ALLFILES.savestatus,"");


  sprintf(stepstr,"%d",step);

  if (step!=-99)
    strcat(strcat(strcat(strcat(mutual,ALLFILES.fotodir),"foto"), g_inputs.runnumber),stepstr);    
  else
    strcat(strcat(mutual,ALLFILES.fotostartdir),"foto_");

  if ( NULL != strstr( g_inputs.format_loadsave , "001" ) )
  {
    strcpy( mutual, "" ) ;
    strcpy( ALLFILES.fotodir, "" ) ;
    sprintf( stepstr , "%08d" , step ) ;
    if ( step == -99 )
    {
        strcat( strcat( strcat( strcat( mutual , g_inputs.datadir ) , "/" ) , "sample" ) , "/" ) ; 
    }
    else if ( step == -98 )
    {
        strcat( strcat( mutual , g_inputs.sample_dir ) , "/" ) ; 
    }
    else
    {
        strcat(  strcat( strcat( strcat( strcat( strcat( mutual , g_inputs.datadir ) , "/" ) , "chkp" ) , "/" ) , stepstr ) , "/" ) ; 
    }
    strcpy( ALLFILES.fotodir , mutual ) ;
        
  }
    

  strcat(strcat(ALLFILES.uidfoto,mutual),"uid.txt");
  strcat(strcat(ALLFILES.dimerfoto,mutual),"dimer.txt");
  strcat(strcat(ALLFILES.r1foto,mutual),"r.txt");
  strcat(strcat(ALLFILES.v1foto,mutual),"v.txt");
  strcat(strcat(ALLFILES.r2foto,mutual),"r2.txt");
  strcat(strcat(ALLFILES.prev_r2foto,mutual),"prev_r2.txt");
  strcat(strcat(ALLFILES.v2foto,mutual),"v2.txt");
  strcat(strcat(ALLFILES.massfoto,mutual),"m.txt");
  strcat(strcat(ALLFILES.etypefoto,mutual),"et.txt");
  strcat(strcat(ALLFILES.chargefoto,mutual),"q.txt");
  strcat(strcat(ALLFILES.stepfoto,mutual),"step.txt");
  strcat(strcat(ALLFILES.energyfoto,mutual),"en.txt");
  strcat(strcat(ALLFILES.farelfoto,mutual),"farel.txt");
  strcat(strcat(ALLFILES.randcallsfoto,mutual),"rand.txt");
  strcat(strcat(ALLFILES.timefoto,mutual),"time.txt");
  strcat(strcat(ALLFILES.el_configfoto,mutual),"econf.txt");
  strcat(strcat(ALLFILES.Zallfoto,mutual),"Z.txt");
  strcat(strcat(ALLFILES.motherionfoto,mutual),"motherion.txt");

  strcat(strcat(ALLFILES.r1fotob,mutual),"rb.txt");
  strcat(strcat(ALLFILES.v1fotob,mutual),"vb.txt");
  strcat(strcat(ALLFILES.r2fotob,mutual),"r2b.txt");
  strcat(strcat(ALLFILES.prev_r2fotob,mutual),"prev_r2b.txt");
  strcat(strcat(ALLFILES.v2fotob,mutual),"v2b.txt");
  strcat(strcat(ALLFILES.massfotob,mutual),"mb.txt");
  strcat(strcat(ALLFILES.energyfotob,mutual),"enb.txt");
  strcat(strcat(ALLFILES.instrecfotob,mutual),"irecb.txt");
  strcat(strcat(ALLFILES.instrecfotol,mutual),"irecl.txt");
  strcat(strcat(ALLFILES.chargetransfer_b,mutual),"chtr.b");
  strcat(strcat(ALLFILES.dbase_overwrite,mutual),"dbase.dat");

  strcat(strcat(ALLFILES.savestatus,mutual),"save_status.log");
}





