///
///  XMDYN - a computer simulation tool for modeling x-ray induced dynamics of matter
///
///  Copyright (C) 2012-2024 Deutsches Elektronen-Synchrotron DESY,
///                          a research centre of the Helmholtz Association.
///  Email: xraypac@mail.desy.de 
/// 
///  This file is part of XMDYN, which is distributed under the terms of the GNU General Public License version 3.
///  See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
///
///  Authors: Zoltan Jurek, Center for Free-Electron Laser Science, DESY
///


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <stdarg.h>
#include <wordexp.h>
#include <libgen.h>

#include <unistd.h>

#include "common.h"

#include "globalconsts.h"
#include "filetests.h"
#include "endltabs.h"
#include "classdyn.h"
#include "xmdyn_io.h"
#include "tests.h"
#include "secel.h"
#include "init_system.h"

#include "Xatom.h"

#include "xmdyn_filenames.h"


#define LOG_BUFFER_SIZE    100000
//#define LOG_BUFFER_SIZE    1100

static int *fotomoments;



static struct allfilenamesstruct *ALLFILES;
static struct save_prec_struct   *SAVE_PREC ;


static struct s_struct_rel_fnames 
{
    char log_mem[STRLTH] ;
} s_rel_fnames ;



static struct snp_fileid_struct
{
    char data_dir[LSTRLTH] ;
    char snp_dir[LSTRLTH] ;
    char snpfull_dir[LSTRLTH] ;
    char archive_dir[LSTRLTH] ;
    char xatom_dir[LSTRLTH] ;
    char tmp_dir[LSTRLTH] ;
    char snp_root[LSTRLTH];

    char Z[SSTRLTH]  ;
    char r[SSTRLTH]  ;
    char v[SSTRLTH]  ;
    char vv[SSTRLTH]  ;
    char qi[SSTRLTH]  ;
    char m[SSTRLTH]  ;
    char ptype[SSTRLTH]  ;
    char trap[SSTRLTH]  ;
    char uid[SSTRLTH]  ;
    char dimer[SSTRLTH]  ;

    char econf[SSTRLTH]  ;

    char atomtype[SSTRLTH]  ;
    char Q[SSTRLTH]  ;
    char f0[SSTRLTH]  ;
    char scd[SSTRLTH]  ;
    char mad[SSTRLTH]  ;
    char ixs[SSTRLTH]  ;
    char fpp[SSTRLTH]  ;

    char energy_new[SSTRLTH] ;
    char energy[SSTRLTH] ;
    char beam[SSTRLTH] ;
    char snp_times[SSTRLTH] ;

    char irlog[SSTRLTH] ;


} SNP_FILEID ; 



// ##########################################################################


#define MODULE_ERROR(xmessage,xreturn) { s_module_error( __FILE__ , __LINE__ , xmessage  , xreturn ) ; }

//############################## FUNCTION #################################

static int s_module_error( char *a_fname , int a_line , char *a_message , int a_return )
{
    fprintf( stderr , "ERROR -- file: %s , line: %d : %s\n" , a_fname , a_line , a_message ) ;
    exit( a_return ) ;
}



//############################## FUNCTION #################################

int init_xmdyn2_loadsave( void )
{
    strcpy( SNP_FILEID.data_dir , g_inputs.datadir ) ;
    strcpy( SNP_FILEID.snp_dir , "/snp" ) ;
    strcpy( SNP_FILEID.snpfull_dir , "/chkp" ) ;
    strcpy( SNP_FILEID.archive_dir , "" ) ;
    strcpy( SNP_FILEID.tmp_dir , "/tmp" ) ;
    strcpy( SNP_FILEID.snp_root, SNP_FILEID.data_dir ) ; 
    strcat( SNP_FILEID.snp_root, SNP_FILEID.snp_dir ) ;

    strcpy( SNP_FILEID.Z , "Z.dat" ) ;
    strcpy( SNP_FILEID.r , "r.dat" ) ;
    strcpy( SNP_FILEID.v , "v.dat" ) ;
    strcpy( SNP_FILEID.vv , "vv.dat" ) ;
    strcpy( SNP_FILEID.qi , "q.dat" ) ;
    strcpy( SNP_FILEID.m , "m.dat" ) ;
    strcpy( SNP_FILEID.ptype , "pt.dat" ) ;
    strcpy( SNP_FILEID.trap , "trap.dat" ) ;
    strcpy( SNP_FILEID.uid , "uid.dat" ) ;
    strcpy( SNP_FILEID.dimer , "dimer.dat" ) ;
    strcpy( SNP_FILEID.econf , "econf.dat" ) ;

    strcpy( SNP_FILEID.atomtype , "T.dat" ) ;
    strcpy( SNP_FILEID.Q , "Q.dat" ) ;
    strcpy( SNP_FILEID.f0 , "f0.dat" ) ;
    strcpy( SNP_FILEID.scd , "scd.dat" ) ;
    strcpy( SNP_FILEID.mad , "mad.dat" ) ;
    strcpy( SNP_FILEID.ixs , "ixs.dat" ) ;
    strcpy( SNP_FILEID.fpp , "fpp.dat" ) ;

    strcpy( SNP_FILEID.energy_new , "all_energy.dat" ) ;
    strcpy( SNP_FILEID.energy , "energy.dat" ) ;
    strcpy( SNP_FILEID.beam , "beam.dat" ) ;
    strcpy( SNP_FILEID.snp_times , "snp_times.dat" ) ;
    strcpy( SNP_FILEID.irlog , "irlog.dat" ) ;

  return 0 ;
}



//############################## FUNCTION #################################

int init_xmdyn_loadsave( char **dir_dbase , char **dir_root ) 
{
    xmdyn_filenames_load( &ALLFILES , &SAVE_PREC ) ;

    init_xmdyn2_loadsave( ) ;

    *dir_dbase = g_inputs.dbasedir ;
    *dir_root = g_inputs.datadir ;

    strcpy( s_rel_fnames.log_mem , "/runtime/malloc.log" ) ;
  return 0 ;
}




//############################## FUNCTION #################################

int create_xmdyn_dirs(void) {
    char xstr[LSTRLTH] ;
  clearstr( xstr ) ; 
  strcat( strcat( xstr , SNP_FILEID.data_dir ) , "" ) ; dircreate( xstr ) ;
  clearstr( xstr ) ; 
  strcat( strcat( xstr , SNP_FILEID.data_dir ) , SNP_FILEID.snp_dir ) ; dircreate( xstr ) ;
  clearstr( xstr ) ; 
#ifdef XMDYN_XRAYPAC_MORELOG
  strcat( strcat( xstr , SNP_FILEID.data_dir ) , SNP_FILEID.snpfull_dir ) ; dircreate( xstr ) ;
  clearstr( xstr ) ;
#endif // XMDYN_XRAYPAC_MORELOG
  strcat( strcat( xstr , SNP_FILEID.data_dir ) , s_rel_fnames.log_mem ) ; dirname( xstr ) ; dircreate( xstr  ) ;

  return 0 ;
}





//############################## FUNCTION #################################

int dirhiercreate(void){
  dircreate( ALLFILES -> datadir ) ;
//  dircreate(ALLFILES -> fotodir);

  create_xmdyn_dirs() ;

  return 0 ;
}





//############################## FUNCTION #################################

int is_snp_save_now( int a_step , char *todo )
{
    int dim ;
    int *snps ;
    int ii ;
    if ( 0 == strcmp( todo , "snp" ) ) {  snps = g_inputs.snp ; dim = g_inputs.snp_dim     ; }
    if ( 0 == strcmp( todo , "snp_full" ) ) {  snps = g_inputs.snp_full ; dim = g_inputs.snp_full_dim  ; }
    if ( 0 == strcmp( todo , "E" ) )   {  snps = g_inputs.snp_E ; dim = g_inputs.snp_E_dim   ; }

    for ( ii = 0 ; ii < dim ; ii++ )
    {
        if ( a_step + 1 == snps[ii]  ) { return 1 ; } ;
    }

    return 0 ;
}





//############################## FUNCTION #################################

int print_log_new( char *a_str , void *a_data , char *a_type , char* a_flush )
{

    static char *log_buffer = NULL ;
    char *new_buffer = NULL , x_buffer[LSTRLTH] = "" ;
    char logfile[LSTRLTH] = "" ;
    FILE *fp ;
    static  int saved_pos = -1 ;


    strcat(strcat(strcat(logfile, g_inputs.datadir), "/"), "events.log") ;

    //* FIXME program stops when buffer flushed + deallocated before finishig log line  *//
    
    if ( NULL == log_buffer ) { log_buffer = calloc( 2 , sizeof(char) ) ; strcpy( log_buffer , "" ) ; }


    if ( 0 == strcmp( a_type , "savepos" ) ) { saved_pos = strlen( log_buffer ) ;  return 0 ; }

    if ( 0 == strcmp( a_type , "replace_1st_char_in_saved" ) ) 
    { 
        if ( -1 == saved_pos ) { MODULE_ERROR( "ERROR: Referring to position lost in log_buffer! Exit." , -1 ) ; } ;
        if ( strlen( log_buffer ) <  (unsigned)saved_pos ) { MODULE_ERROR( "ERROR: Referring to position beyond log_buffer! Exit." , -1 ) ; } ;
        log_buffer[saved_pos] = a_str[0] ; return 0 ; 
    }

    if ( 0 == strcmp( a_type , "remove_saved" ) ) {
        if ( -1 == saved_pos ) { MODULE_ERROR( "ERROR: Referring to position lost in log_buffer! Exit." , -1 ) ; } ;
        if ( strlen( log_buffer ) <  (unsigned)saved_pos ) { MODULE_ERROR( "ERROR: Referring to position beyond log_buffer! Exit." , -1 ) ; } ;
       log_buffer[saved_pos] = EoS ; return 0 ; 
    }

    if ( 0 == strcmp( a_type , "replace_1st_char" ) ) { log_buffer[0] = a_str[0] ; return 0 ; }


    new_buffer = calloc( 2 , sizeof(char) ) ; strcpy( new_buffer , "" ) ;

    new_buffer = str_concat_realloc( new_buffer , a_str ) ;

    if ( a_data != NULL )
    {
        if ( 0 == strcmp( a_type , "llint" ) )
        {
            sprintf( x_buffer , "%lld" , *((long long int*)a_data) ) ;
        }
        else if ( 0 == strcmp( a_type , "int" ) )
        {
            sprintf( x_buffer , "%d" , *((int*)a_data) ) ;
        }
        else if ( 0 == strcmp( a_type , "double" ) )
        {
            sprintf( x_buffer , "%1.5le" , *((double*)a_data) ) ;
        }
        else if ( 0 == strcmp( a_type , "str" ) )
        {
            sprintf( x_buffer , "%s" , ((char*)a_data) ) ;
        }
        else if ( 0 == strcmp( a_type , "" ) )
        {
            strcpy( x_buffer , "" ) ;
        }
        else { MODULE_ERROR( "Wrong a_type" , -1 ) ; }

        if ( strlen( new_buffer ) > 0 )
        {
            new_buffer = str_concat_realloc( new_buffer , " " ) ;
        }
        new_buffer = str_concat_realloc( new_buffer , x_buffer ) ;
    }

    if ( SCREEN_AND_LOGFILE == g_inputs.event_dest )
    {
        fprintf( stdout, "%s" , new_buffer ) ;
    }

    log_buffer = str_concat_realloc( log_buffer , new_buffer ) ;

    if ( ( strlen( log_buffer ) > LOG_BUFFER_SIZE ) || 
         ( strcmp( a_flush , "flush" ) == 0 ) )
    {
        fp = fopen( logfile , "a" ) ;
        fprintf( fp , "%s" , log_buffer ) ;
        fclose( fp ) ;
        free( log_buffer ) ; log_buffer = NULL ; saved_pos = -1 ;
    }

    free( new_buffer ) ;

  return 0 ;
}




//############################## FUNCTION #################################

int print_log_compat( char *a_str , void *a_data , char *a_type , char* a_flush )
{
    //DBG_WARNING_NOTE_ON_CODE( "print_log_compat is off!" );    
    return 0 ;

    static char log_buffer[LOG_BUFFER_SIZE] = ""  , *saved_pos ;
    char new_buffer[LSTRLTH] = "" , x_buffer[LSTRLTH] = "" ;
    char logfile[LSTRLTH] = "" ;
    FILE *fp ;


    strcat(strcat(strcat(logfile, g_inputs.datadir), "/"), "dbg-events_compat.log") ;

    if ( 0 == strcmp( a_type , "savepos" ) ) { saved_pos = log_buffer + strlen( log_buffer ) ;  return 0 ; }

    if ( 0 == strcmp( a_type , "replace_1st_char_in_saved" ) ) { saved_pos[0] = a_str[0] ; return 0 ; }

    if ( 0 == strcmp( a_type , "remove_saved" ) ) { saved_pos[0] = EoS ; return 0 ; }

    if ( 0 == strcmp( a_type , "replace_1st_char" ) ) { log_buffer[0] = a_str[0] ; return 0 ; }


    strcat( new_buffer , a_str ) ;

    if ( a_data != NULL )
    {
        if ( 0 == strcmp( a_type , "int" ) )
        {
            sprintf( x_buffer , "%d" , *((int*)a_data) ) ;
        }

        if ( 0 == strcmp( a_type , "double" ) )
        {
            sprintf( x_buffer , "%1.10le" , *((double*)a_data) ) ;
        }

        if ( 0 == strcmp( a_type , "str" ) )
        {
            sprintf( x_buffer , "%s" , ((char*)a_data) ) ;
        }
        if ( strlen( new_buffer ) > 0 )
        {
            strcat( new_buffer , " " ) ;
        }
        strcat( new_buffer , x_buffer ) ;

    }

    if ( strlen( log_buffer ) + strlen( new_buffer ) + 5  > LOG_BUFFER_SIZE )
    {
        fp = fopen( logfile , "a" ) ;
        fprintf( fp , "%s" , log_buffer ) ;
        fclose( fp ) ;
        strcpy( log_buffer , "" ) ;
    }
    strcat( log_buffer , new_buffer ) ;

    if ( strcmp( a_flush , "flush" ) == 0 ) 
    {
        fp = fopen( logfile , "a" ) ;
        fprintf( fp , "%s" , log_buffer ) ;
        fclose( fp ) ;
        strcpy( log_buffer , "" ) ;
    }

  return 0 ;
}





//############################## FUNCTION #################################

int print_log( char *a_str , void *a_data , char *a_type , char* a_flush )
{
//return -1 ; // TODO TODO TODO
    print_log_new( a_str , a_data , a_type , a_flush ) ;
    print_log_compat( a_str , a_data , a_type , a_flush ) ;
    return 0 ;
}




//############################## FUNCTION #################################

int print_log_full( char *a_str , void *a_data , char *a_type , char* a_flush )
{
    print_log_new( a_str , a_data , a_type , a_flush ) ;
    return 0 ;
}





//############################## FUNCTION #################################

void paramsout(char *programname,char *VERSIONNUMBER){
  FILE *fp;


  fp=fopen(ALLFILES -> params,"a");
  fprintf(fp,"Prob./%1.2esec of photoeff for K electron: %e\n", g_inputs.DT, g_inputs.sprob);
  fprintf(fp,"Prob./%1.2esec of photoeff for L electron: %e\n\n", g_inputs.DT, g_inputs.pprob);

  fclose(fp);


  fp=fopen(ALLFILES -> infofile,"a");
  fprintf(fp,"This calculation is done by %s based on cd%s.c\n",programname,VERSIONNUMBER);
  fprintf(fp,"Impioncriterium: %s\n", g_inputs.impioncriterium);
  if ( g_inputs.fixatomsOK=='y')
    fprintf(fp,"The atoms are not moving, they are fixed.\n");
  if (strcmp( g_inputs.avaltype,"none")!=0){
    fprintf(fp,"Type of avalanche: %s.\n", g_inputs.avaltype);
    fprintf(fp,"Atom in the centre: %d\n", g_inputs.centeratom);
  }

  fclose(fp);
}






//############################## FUNCTION #################################

int get_xmdyn_snp_fname( char *a_type , int a_step , char *a_fname ) 
{
    char stepstr[STRLTH] , tmpstr[LSTRLTH] ;

    clearstr( a_fname ) ;
    sprintf( stepstr , "%08d" , a_step ) ;

    if ( 0 == strcmp( a_type , "r" ) ) { strcpy( tmpstr ,  SNP_FILEID.r ) ; } 
    else if ( 0 == strcmp( a_type , "basename" ) ) { strcpy( a_fname ,  stepstr ) ; return 0 ; } 
    else if ( 0 == strcmp( a_type , "root" ) ) { strcpy( tmpstr , "" ) ; } 
    else if ( 0 == strcmp( a_type , "Z" ) ) { strcpy( tmpstr ,  SNP_FILEID.Z ) ; } 
    else if ( 0 == strcmp( a_type , "v" ) ) { strcpy( tmpstr ,  SNP_FILEID.v ) ; } 
    else if ( 0 == strcmp( a_type , "q" ) ) { strcpy( tmpstr ,  SNP_FILEID.qi ) ; } 
    else if ( 0 == strcmp( a_type , "ptype" ) ) { strcpy( tmpstr ,  SNP_FILEID.ptype ) ; } 
    else if ( 0 == strcmp( a_type , "trap" ) ) { strcpy( tmpstr ,  SNP_FILEID.trap ) ; } 
    else if ( 0 == strcmp( a_type , "uid" ) ) { strcpy( tmpstr ,  SNP_FILEID.uid ) ; } 
    else if ( 0 == strcmp( a_type , "dimer" ) ) { strcpy( tmpstr ,  SNP_FILEID.dimer ) ; } 
    else if ( 0 == strcmp( a_type , "econf" ) ) { strcpy( tmpstr ,  SNP_FILEID.econf ) ; } 
    else if ( 0 == strcmp( a_type , "atomtype" ) ) { strcpy( tmpstr ,  SNP_FILEID.atomtype ) ; } 
    else if ( 0 == strcmp( a_type , "Q" ) ) { strcpy( tmpstr ,  SNP_FILEID.Q ) ; } 
    else if ( 0 == strcmp( a_type , "f0" ) ) { strcpy( tmpstr ,  SNP_FILEID.f0 ) ; } 
    else if ( 0 == strcmp( a_type , "scd" ) ) { strcpy( tmpstr ,  SNP_FILEID.scd ) ; } 
    else if ( 0 == strcmp( a_type , "mad" ) ) { strcpy( tmpstr ,  SNP_FILEID.mad ) ; } 
    else if ( 0 == strcmp( a_type , "ixs" ) ) { strcpy( tmpstr ,  SNP_FILEID.ixs ) ; } 
    else if ( 0 == strcmp( a_type , "fpp" ) ) { strcpy( tmpstr ,  SNP_FILEID.fpp ) ; } 
    else if ( 0 == strcmp( a_type , "m" ) ) { strcpy( tmpstr ,  SNP_FILEID.m ) ; } 

    else if ( 0 == strcmp( a_type , "E" ) ) { strcpy( tmpstr ,  SNP_FILEID.energy_new ) ; strcpy( stepstr , "." ) ; } 
    else if ( 0 == strcmp( a_type , "energy" ) ) { strcpy( tmpstr ,  SNP_FILEID.energy ) ; strcpy( stepstr , "." ) ;  } 
    else if ( 0 == strcmp( a_type , "beam" ) ) { strcpy( tmpstr ,  SNP_FILEID.beam ) ; strcpy( stepstr , "." ) ;  } 
    else if ( 0 == strcmp( a_type , "snp_times" ) ) { strcpy( tmpstr ,  SNP_FILEID.snp_times ) ; strcpy( stepstr , "." ) ;  } 

    else if ( 0 == strcmp( a_type , "snpfull_dir" ) ) 
    { 
        strcpy( tmpstr , "" ) ; 
        strcat( strcat( strcat( strcat( strcat( a_fname , SNP_FILEID.data_dir ) , "/" ) ,SNP_FILEID.snpfull_dir ) , "/" ) , stepstr ) ; 
        return 0 ;
    } 

    else { return -1 ; }

    if ( strlen( tmpstr ) > 0 )
    {
        strcat( strcat(  strcat( strcat( strcat( strcat( strcat( a_fname , SNP_FILEID.data_dir ) , "/" ) ,SNP_FILEID.snp_dir ) , "/" ) , stepstr ) , "/" ) ,  tmpstr ) ; 
    }
    else 
    {
        strcat( strcat(  strcat( strcat( strcat( strcat( strcat( a_fname , SNP_FILEID.data_dir ) , "/" ) ,SNP_FILEID.snp_dir ) , "/" ) , stepstr ) , "" ) ,  tmpstr ) ; 
    }

    return 0 ;
}



//############################## FUNCTION #################################

int print_xmdyn_data_tofile( char *a_fname , char *a_type , char *txtbin_format , void *a_p , int d1 , int d2 )
{
  FILE *fp ;
  int ii, kk;
  double *pd ;
  int *pi ;
  char *pc ;
  char tmpstr[LSTRLTH] ,  *p_ch ;
  long long int *plli ;

  strcpy( tmpstr , a_fname ) ;
  dirname( tmpstr )  ;
  dircreate( tmpstr  ) ;

  fp = fopen( a_fname , "w" ) ;

if ( 0 == strcmp( txtbin_format , "txt" ) )
{

  if ( 0 == strcmp( a_type , "d" ) )
  {
      pd = (double*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% le  " , pd[ ii*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
  if ( 0 == strcmp( a_type , "lli" ) )
  {
      plli = (long long int*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% 3lld  " , plli[ ii*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
   if ( 0 == strcmp( a_type , "i" ) )
  {
      pi = (int*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% 3d  " , pi[ ii*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
  if ( 0 == strcmp( a_type , "c" ) )
  {
      pc = (char*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "%c  " , pc[ ii*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
}

else
{
  int unitsize ;
  if ( 0 == strcmp( a_type , "d" ) ) { unitsize = sizeof(double) ; }
  if ( 0 == strcmp( a_type , "i" ) ) { unitsize = sizeof(int) ; }
  if ( 0 == strcmp( a_type , "lli" ) ) { unitsize = sizeof(long long int) ; }
  if ( 0 == strcmp( a_type , "c" ) ) { unitsize = sizeof(char) ; }
  fwrite( a_p , unitsize , d1 * d2 , fp ) ;
}



  fclose( fp ) ;

  return 0 ;

} 



//############################## FUNCTION #################################

int print_xmdyn_data_tofile_reduced( char *a_fname , char *a_type , char *txtbin_format , void *a_p , int *sel_parts , int d1 , int d2 )
{
  FILE *fp ;
  int ii, kk;
  double *pd ;
  int *pi ;
  char *pc ;
  char tmpstr[LSTRLTH] ,  *p_ch ;
  long long int *plli ;

  strcpy( tmpstr , a_fname ) ;
  dirname( tmpstr )  ;
  dircreate( tmpstr  ) ;

  fp = fopen( a_fname , "w" ) ;

if ( 0 == strcmp( txtbin_format , "txt" ) )
{

  if ( 0 == strcmp( a_type , "d" ) )
  {
      pd = (double*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% le  " , pd[ sel_parts[ii]*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
  if ( 0 == strcmp( a_type , "lli" ) )
  {
      plli = (long long int*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% 3lld  " , plli[ sel_parts[ii]*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
   if ( 0 == strcmp( a_type , "i" ) )
  {
      pi = (int*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "% 3d  " , pi[ sel_parts[ii]*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
 
  if ( 0 == strcmp( a_type , "c" ) )
  {
      pc = (char*)a_p ;
      for ( ii=0 ; ii < d1 ; ++ii ) {
        for ( kk=0 ; kk < d2 ; ++kk ) { fprintf( fp , "%c  " , pc[ sel_parts[ii]*d2 + kk ] ) ; }
        fprintf(fp,"\n");
      }
  }
}

else
{
  int unitsize ;
  if ( 0 == strcmp( a_type , "d" ) ) { unitsize = sizeof(double) ; }
  if ( 0 == strcmp( a_type , "i" ) ) { unitsize = sizeof(int) ; }
  if ( 0 == strcmp( a_type , "lli" ) ) { unitsize = sizeof(long long int) ; }
  if ( 0 == strcmp( a_type , "c" ) ) { unitsize = sizeof(char) ; }
  for ( ii=0 ; ii < d1 ; ++ii ) { fwrite( (void*)((char*)a_p + sel_parts[ii]*d2) , unitsize , d2 , fp ) ; }
}


  fclose( fp ) ;

  return 0 ;

} 



//############################## FUNCTION #################################

int link_xmdyn_snp( int step , int ref )
{
    int ii ;
    char fname_ref[LSTRLTH] , fname_step[LSTRLTH] ;
    get_xmdyn_snp_fname( "root" , step , fname_step ) ;
    get_xmdyn_snp_fname( "basename" , ref  , fname_ref  ) ;
    ii = symlink( fname_ref , fname_step ) ;
    return 0 ;
}



//############################## FUNCTION #################################

int link_xmdyn_snpfull( int step , int ref )
{
    int ii ;
    char fname_ref[LSTRLTH] , fname_step[LSTRLTH] ;
    get_xmdyn_snp_fname( "snpfull_dir" , step , fname_step ) ;
    get_xmdyn_snp_fname( "basename" , ref  , fname_ref  ) ;
    ii = symlink( fname_ref , fname_step ) ;
    return 0 ;
}




//############################## FUNCTION #################################

int save_xmdyn_snp_full( char *a_snp_type , int step )
{

  FILE *fp ;
  int ii, jj , kk;
  double *pr,*pv,vv,nph;
  int fototab;

  int N,E,Z;
  double T_1,T0,DT;
  double tempd1;
  char fname[LSTRLTH] ;
  char xstr[LSTRLTH] ;

  char *txtbin_format = g_inputs.txtbin_loadsave ;
  char *dbase_types = g_inputs.dbase_types ;


  if ( strstr(  g_inputs.misc_opt , "nodiskwrite" ) != NULL ) { return 0 ; } ;


  N = g_inputs.N ; Z = g_inputs.Z ; E = g_inputs.E ; 
  T_1 = g_inputs.T_1 ; T0 = g_inputs.T0 ; DT = g_inputs.DT ;

  if ( 0 == strcmp( a_snp_type , "E" ) ) 
  { 
      step++ ; 
      tempd1 = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(3);
      get_xmdyn_snp_fname( "E" , step , fname ) ;
      fp=fopen( fname ,"a");
      fprintf(fp, "% 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e %d\n" , 
                  g_ENERGY_VALUES.Eintotal , g_ENERGY_VALUES.Eintotal_th , 
                  g_ENERGY_VALUES.Einph , g_ENERGY_VALUES.Einfl, g_ENERGY_VALUES.Einau, g_ENERGY_VALUES.Einsi, g_ENERGY_VALUES.Einrec, 
                  tempd1, g_ENERGY_VALUES.Einitial, g_ENERGY_VALUES.Elost, g_ENERGY_VALUES.Einfi , g_ENERGY_VALUES.Einchmig , g_ENERGY_VALUES.Ebound , g_ENERGY_VALUES.Efl_photon , g_inputs.step) ; //, g_ENERGY_VALUES.Work ) ;
      fclose(fp);
      return 0 ;
  } 
  


  if ( 0 == strcmp( a_snp_type , "snp" ) ) { step++ ; } 

#ifdef XMDYN_XRAYPAC_MORELOG
  tempd1 = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(3);
  get_xmdyn_snp_fname( "energy" , step , fname ) ;
  fp=fopen( fname ,"a");
  fprintf(fp, "% 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e %d\n" , 
              g_ENERGY_VALUES.Eintotal , g_ENERGY_VALUES.Eintotal_th , 
              g_ENERGY_VALUES.Einph , g_ENERGY_VALUES.Einfl, g_ENERGY_VALUES.Einau, g_ENERGY_VALUES.Einsi, g_ENERGY_VALUES.Einrec, 
              tempd1, g_ENERGY_VALUES.Einitial, g_ENERGY_VALUES.Elost, g_ENERGY_VALUES.Einfi , g_ENERGY_VALUES.Einchmig  , g_ENERGY_VALUES.Ebound , g_ENERGY_VALUES.Efl_photon , g_inputs.step ) ; //, g_ENERGY_VALUES.Work ) ;
  fclose(fp);
#endif // XMDYN_XRAYPAC_MORELOG

  beamparams(step,&nph);
  get_xmdyn_snp_fname( "beam" , step , fname ) ;
  fp=fopen( fname ,"a");   fprintf(fp,"%1.8e ",nph);   fclose(fp);

  get_xmdyn_snp_fname( "snp_times" , step , fname ) ;
  fp = fopen( fname , "a" ) ; fprintf( fp , "%d " ,  step+1 ) ; fclose( fp ) ;

  print_log( "" , NULL , "" , "flush" ) ;
 

  if (  ( 2 == xmdyn_apply_shortcut( "check" ) )  &&  ( -1 != g_inputs.shortcut_ref )  ) 
  {  
      link_xmdyn_snp( step , g_inputs.shortcut_ref ) ;   return 0 ;  
  } 

  g_inputs.shortcut_ref = step ;

  get_xmdyn_snp_fname( "uid" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "lli" , txtbin_format ,(void*)g_system.uid , N+E , 1 ) ;

  get_xmdyn_snp_fname( "Z" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "i" , txtbin_format ,(void*)g_system.Z , N+E , 1 ) ;

  get_xmdyn_snp_fname( "r" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "d" , txtbin_format , (void*)&(g_system.r1[0][0]) , N+E , 3 ) ;

  get_xmdyn_snp_fname( "v" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "d" , txtbin_format , (void*)&(g_system.v1[0][0]) , N+E , 3 ) ;

  get_xmdyn_snp_fname( "q" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "i" , txtbin_format , (void*)g_system.qi , N+E , 1 ) ;

#ifdef XMDYN_XRAYPAC_MORELOG
  get_xmdyn_snp_fname( "ptype" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "c" , txtbin_format , (void*)g_system.ptype , N+E , 1 ) ;
#endif // XMDYN_XRAYPAC_MORELOG

  get_xmdyn_snp_fname( "m" , step , fname ) ;
  print_xmdyn_data_tofile( fname , "d" , txtbin_format , (void*)g_system.m , N+E , 1 ) ;

  get_xmdyn_snp_fname( "trap" , step , fname ) ;
  fp = fopen( fname , "w" ) ; save_snp_recombination_log( &fp ) ; fclose(fp) ;

  int *atom_type_label = NULL , num_atom_type , num_Q = -1 ;
  struct g_struct_xatom_data **atom_type_xatom = NULL ;

  xmdyn_get_atom_types( &num_atom_type , &atom_type_label , &atom_type_xatom ) ;
  //
  get_xmdyn_snp_fname( "econf" , step , fname ) ;
  fp = fopen( fname , "w" ) ;
  for ( ii=0 ; ii < num_atom_type ; ii++) { fprintf( fp , "%02d %s\n" , atom_type_xatom[ii]->Z, atom_type_xatom[ii]->str ) ; }
  fclose(fp) ;
  //
  get_xmdyn_snp_fname( "atomtype" , step , fname ) ;
  fp = fopen( fname , "w" ) ;
  for ( ii=0 ; ii < N ; ii++) { fprintf( fp , "%d " , atom_type_label[ii] ) ; }
  fprintf( fp , "\n" ) ;
  fclose(fp) ;
  //

  if ( ( NULL != strstr( dbase_types , "f0" ) )  ||
       ( NULL != strstr( dbase_types , "scd" ) ) ||
       ( NULL != strstr( dbase_types , "ixs" ) )   )
  {
      get_xmdyn_snp_fname( "Q" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( jj=0 ; jj < atom_type_xatom[0]->num_sample ; jj++) 
      { 
          fprintf( fp , "%le " , atom_type_xatom[0]->q_sample[jj] ) ; 
      }
      fprintf( fp , "\n" ) ;
      fclose(fp) ;
  }
  //

  if ( ( NULL != strstr( dbase_types , "f0" ) )  ||
       ( NULL != strstr( dbase_types , "scd" ) ) )
  {
      get_xmdyn_snp_fname( "f0" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          for ( jj=0 ; jj < atom_type_xatom[ii]->num_sample ; jj++) 
          {
              fprintf( fp , "%le " , atom_type_xatom[ii]->f0[jj] ) ; 
          }
          fprintf( fp , "\n" ) ;
      }
      fclose(fp) ;
  }

  //
  if ( NULL != strstr( dbase_types , "mad" ) ) 
  {
      get_xmdyn_snp_fname( "mad" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          fprintf( fp , "%+le %+le\n" , atom_type_xatom[ii]->fprime ,  atom_type_xatom[ii]->fdprime ) ; 
      }
      fclose(fp) ;
  }

  //
  if ( NULL != strstr( dbase_types , "fpp" ) ) 
  {
      get_xmdyn_snp_fname( "fpp" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          fprintf( fp , "%+le\n" , atom_type_xatom[ii]->fdprime_pcs ) ; 
      }
      fclose(fp) ;
  }
  //

  if ( ( NULL != strstr( dbase_types , "ixs" ) ) )
  {
      get_xmdyn_snp_fname( "ixs" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          for ( jj=0 ; jj < atom_type_xatom[ii]->num_sample ; jj++) 
          {
              fprintf( fp , "%le " , atom_type_xatom[ii]->ixs[jj] ) ; 
          }
          fprintf( fp , "\n" ) ;
      }
      fclose(fp) ;
  }

  //

  free( atom_type_label ) ; free( atom_type_xatom ) ;


  return 0 ;

}


//############################## FUNCTION #################################

int save_xmdyn_snp_reduced( char *a_snp_type , int step )
{

  FILE *fp ;
  int ii, jj , kk;
  double *pr,*pv,vv,nph;
  int fototab;

  int N,E,Z;
  double T_1,T0,DT;
  double tempd1;
  char fname[LSTRLTH] ;
  char xstr[LSTRLTH] ;

  char *txtbin_format = g_inputs.txtbin_loadsave ;
  char *dbase_types = g_inputs.dbase_types ;


  if ( strstr(  g_inputs.misc_opt , "nodiskwrite" ) != NULL ) { return 0 ; } ;


  N = g_inputs.N ; Z = g_inputs.Z ; E = g_inputs.E ; 
  T_1 = g_inputs.T_1 ; T0 = g_inputs.T0 ; DT = g_inputs.DT ;


  if ( 0 == strcmp( a_snp_type , "E" ) ) 
  { 
      step++ ; 
      tempd1 = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(3);
      get_xmdyn_snp_fname( "E" , step , fname ) ;
      fp=fopen( fname ,"a");
      fprintf(fp, "% 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e %d\n" , 
                  g_ENERGY_VALUES.Eintotal , g_ENERGY_VALUES.Eintotal_th , 
                  g_ENERGY_VALUES.Einph , g_ENERGY_VALUES.Einfl, g_ENERGY_VALUES.Einau, g_ENERGY_VALUES.Einsi, g_ENERGY_VALUES.Einrec, 
                  tempd1, g_ENERGY_VALUES.Einitial, g_ENERGY_VALUES.Elost, g_ENERGY_VALUES.Einfi , g_ENERGY_VALUES.Einchmig , g_ENERGY_VALUES.Ebound , g_ENERGY_VALUES.Efl_photon , g_inputs.step) ; //, g_ENERGY_VALUES.Work ) ;
      fclose(fp);
      return 0 ;
  } 
  



  if ( 0 == strcmp( a_snp_type , "snp" ) ) { step++ ; } 

#ifdef XMDYN_XRAYPAC_MORELOG
  tempd1 = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(3);
  get_xmdyn_snp_fname( "energy" , step , fname ) ;
  fp=fopen( fname ,"a");
  fprintf(fp, "% 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e % 1.18e %d\n" , 
              g_ENERGY_VALUES.Eintotal , g_ENERGY_VALUES.Eintotal_th , 
              g_ENERGY_VALUES.Einph , g_ENERGY_VALUES.Einfl, g_ENERGY_VALUES.Einau, g_ENERGY_VALUES.Einsi, g_ENERGY_VALUES.Einrec, 
              tempd1, g_ENERGY_VALUES.Einitial, g_ENERGY_VALUES.Elost, g_ENERGY_VALUES.Einfi , g_ENERGY_VALUES.Einchmig  , g_ENERGY_VALUES.Ebound , g_ENERGY_VALUES.Efl_photon , g_inputs.step ) ; //, g_ENERGY_VALUES.Work ) ;
  fclose(fp);
#endif // XMDYN_XRAYPAC_MORELOG

  beamparams(step,&nph);
  get_xmdyn_snp_fname( "beam" , step , fname ) ;
  fp=fopen( fname ,"a");   fprintf(fp,"%1.8e ",nph);   fclose(fp);

  get_xmdyn_snp_fname( "snp_times" , step , fname ) ;
  fp = fopen( fname , "a" ) ; fprintf( fp , "%d " ,  step ) ; fclose( fp ) ;

  print_log( "" , NULL , "" , "flush" ) ;
 

  if (  ( 2 == xmdyn_apply_shortcut( "check" ) )  &&  ( -1 != g_inputs.shortcut_ref )  ) 
  {  
      link_xmdyn_snp( step , g_inputs.shortcut_ref ) ;   return 0 ;  
  } 

  int classpartdim = calc_class_parts_dim() ;
  int* classparts = calloc( classpartdim , sizeof(int) ) ;
  set_class_parts( classparts , classpartdim ) ;

  g_inputs.shortcut_ref = step ;

  get_xmdyn_snp_fname( "uid" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "lli" , txtbin_format ,(void*)g_system.uid , classparts , classpartdim , 1 ) ;


  get_xmdyn_snp_fname( "Z" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "i" , txtbin_format ,(void*)g_system.Z , classparts , classpartdim , 1 ) ;

  get_xmdyn_snp_fname( "r" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "d" , txtbin_format , (void*)&(g_system.r1[0][0]) , classparts , classpartdim , 3 ) ;

  get_xmdyn_snp_fname( "v" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "d" , txtbin_format , (void*)&(g_system.v1[0][0]) , classparts , classpartdim , 3 ) ;

  get_xmdyn_snp_fname( "q" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "i" , txtbin_format , (void*)g_system.qi , classparts , classpartdim , 1 ) ;

#ifdef XMDYN_XRAYPAC_MORELOG
  get_xmdyn_snp_fname( "ptype" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "c" , txtbin_format , (void*)g_system.ptype , classparts , classpartdim , 1 ) ;
#endif // XMDYN_XRAYPAC_MORELOG

  get_xmdyn_snp_fname( "m" , step , fname ) ;
  print_xmdyn_data_tofile_reduced( fname , "d" , txtbin_format , (void*)g_system.m , classparts , classpartdim , 1 ) ;

#ifdef XMDYN_XRAYPAC_MORELOG
  get_xmdyn_snp_fname( "trap" , step , fname ) ;
  fp = fopen( fname , "w" ) ; save_snp_recombination_log( &fp ) ; fclose(fp) ;
#endif // XMDYN_XRAYPAC_MORELOG

  int *atom_type_label = NULL , num_atom_type , num_Q = -1 ;
  struct g_struct_xatom_data **atom_type_xatom = NULL ;

  xmdyn_get_atom_types( &num_atom_type , &atom_type_label , &atom_type_xatom ) ;
  //
  get_xmdyn_snp_fname( "econf" , step , fname ) ;
  fp = fopen( fname , "w" ) ;
  for ( ii=0 ; ii < num_atom_type ; ii++) { fprintf( fp , "%02d %s\n" , atom_type_xatom[ii]->Z, atom_type_xatom[ii]->str ) ; }
  fclose(fp) ;
  //
  get_xmdyn_snp_fname( "atomtype" , step , fname ) ;
  fp = fopen( fname , "w" ) ;
  for ( ii=0 ; ii < N ; ii++) { fprintf( fp , "%d " , atom_type_label[ii] ) ; }
  fprintf( fp , "\n" ) ;
  fclose(fp) ;
  //

  if ( ( NULL != strstr( dbase_types , "f0" ) )  ||
       ( NULL != strstr( dbase_types , "scd" ) ) ||
       ( NULL != strstr( dbase_types , "ixs" ) )   )
  {
      get_xmdyn_snp_fname( "Q" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( jj=0 ; jj < atom_type_xatom[0]->num_sample ; jj++) 
      { 
          fprintf( fp , "%le " , atom_type_xatom[0]->q_sample[jj] ) ; 
      }
      fprintf( fp , "\n" ) ;
      fclose(fp) ;
  }
  //

  if ( ( NULL != strstr( dbase_types , "f0" ) )  ||
       ( NULL != strstr( dbase_types , "scd" ) ) )
  {
      get_xmdyn_snp_fname( "f0" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          for ( jj=0 ; jj < atom_type_xatom[ii]->num_sample ; jj++) 
          {
              fprintf( fp , "%le " , atom_type_xatom[ii]->f0[jj] ) ; 
          }
          fprintf( fp , "\n" ) ;
      }
      fclose(fp) ;
  }

  //
  if ( NULL != strstr( dbase_types , "mad" ) ) 
  {
      get_xmdyn_snp_fname( "mad" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          fprintf( fp , "%+le %+le\n" , atom_type_xatom[ii]->fprime ,  atom_type_xatom[ii]->fdprime ) ; 
      }
      fclose(fp) ;
  }

  //
  if ( NULL != strstr( dbase_types , "fpp" ) ) 
  {
      get_xmdyn_snp_fname( "fpp" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          fprintf( fp , "%+le\n" , atom_type_xatom[ii]->fdprime_pcs ) ; 
      }
      fclose(fp) ;
  }
  //

  if ( ( NULL != strstr( dbase_types , "ixs" ) ) )
  {
      get_xmdyn_snp_fname( "ixs" , step , fname ) ;
      fp = fopen( fname , "w" ) ;
      for ( ii=0 ; ii < num_atom_type ; ii++) 
      { 
          for ( jj=0 ; jj < atom_type_xatom[ii]->num_sample ; jj++) 
          {
              fprintf( fp , "%le " , atom_type_xatom[ii]->ixs[jj] ) ; 
          }
          fprintf( fp , "\n" ) ;
      }
      fclose(fp) ;
  }

  //

  free( atom_type_label ) ; free( atom_type_xatom ) ;


  free( classparts ) ;
  return 0 ;

}


//############################## FUNCTION #################################

int save_xmdyn_snp( char *a_snp_type , int step )
{
    int version = 1 ;
#ifdef XMDYN_XRAYPAC_MORELOG
    version = 0 ;
#endif // XMDYN_XRAYPAC_MORELOG
    if ( 0 == version ) { return save_xmdyn_snp_full( a_snp_type , step ) ; }
    if ( 1 == version ) { return save_xmdyn_snp_reduced( a_snp_type , step ) ; }
    return -1 ;
}



//############################## FUNCTION #################################

void data2disk(int step){
  FILE *fp1,*fp2,*fp3,*fp4,*fp5,*fp6,*fp7,*fpv3, *fp ;
  FILE *fpbeam;
  int i,k;
  double *pr,*pv,vv,nph;
  int fototab;

  int N,E,Z;
  double T_1,T0,DT;
  double tempd1;
  char saveprec[STRLTH]="", saveprecE[STRLTH]="";

  if ( strstr(  g_inputs.misc_opt , "nodiskwrite" ) != NULL ) { return ; } ;
	  
  N= g_inputs.N;Z= g_inputs.Z;E= g_inputs.E;T_1= g_inputs.T_1;T0= g_inputs.T0;DT= g_inputs.DT;

  sprintf(saveprec,"%% 1.%de ",SAVE_PREC -> r);
  sprintf(saveprecE,"%% 1.%de ",SAVE_PREC -> E);



  if ( g_inputs.foto> g_inputs.nstep)
    fototab=1;
  else
    fototab= g_inputs.foto;

  fp = fopen( ALLFILES->snptimes , "a" ) ;
  fprintf( fp , "%d " ,  step+1 ) ;
  fclose( fp ) ;

    
  fp6=fopen(ALLFILES -> data2diskindicator,"w");
  fprintf(fp6,"*\n");
  fclose(fp6);

  fp1=fopen(ALLFILES -> r,"a");
  if ( g_inputs.savev3=='y')
    fpv3=fopen(ALLFILES -> v3,"a");
  fp2=fopen(ALLFILES -> v,"a");
  fp3=fopen(ALLFILES -> q,"a");
  fp4=fopen(ALLFILES -> m,"a");
  fp5=fopen(ALLFILES -> et,"a");
  fp7=fopen(ALLFILES -> Eintotal,"a");
  tempd1 = xmdyn_md_energy( "system-E" , "getcache" ) ; //X calc_tot_init_ener(3);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Eintotal);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Eintotal_th);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einph);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einfl);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einau);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einsi);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einrec);
  fprintf(fp7,saveprecE,tempd1);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einitial);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Elost);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einfi);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Einchmig);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Ebound);
  fprintf(fp7,saveprecE, g_ENERGY_VALUES.Efl_photon);
  fprintf(fp7,"\n");
  fclose(fp7);
    //}

  for (i=0;i<N+E;++i){
    pr= g_system.r1 [i],  pv= g_system.v1 [i];
    vv=sqrt(pv[0]*pv[0]+pv[1]*pv[1]+pv[2]*pv[2]);
    fprintf(fp2,saveprec,vv);
    fprintf(fp3,"% 1.2d ", g_system.qi [i]);
    fprintf(fp4,"% 1.3e ", g_system.m [i]);
    fprintf(fp5,"%c ", g_system.ptype [i]);
    for (k=0;k<3;++k) {
      fprintf(fp1,saveprec,pr[k]);
      if ( g_inputs.savev3=='y')
	fprintf(fpv3,saveprec,pv[k]);
    }
  }

  fprintf(fp1,"\n");
  fprintf(fp2,"\n");
  fprintf(fp3,"\n");
  fprintf(fp4,"\n");
  fprintf(fp5,"\n");
  if ( g_inputs.savev3=='y')
    fprintf(fpv3,"\n");

  fclose(fp1), fclose(fp2), fclose(fp3),fclose(fp5);fclose(fp4);
  if ( g_inputs.savev3=='y')
    fclose(fpv3);

  beamparams(step,&nph);
  fpbeam=fopen(ALLFILES -> beamfile,"a");
  fprintf(fpbeam,"%1.8e ",nph);
  fclose(fpbeam);

  fp6=fopen(ALLFILES -> data2diskindicator,"a");
  fprintf(fp6,"*\n\n");
  fclose(fp6);

}





//############################## FUNCTION #################################

int foto2disk(int step,int indic) {
  /* indic: fotoindicator*/
  int i,j, ii ;
  FILE *fpr1,*fpr2,*fpv1,*fpv2,*fpm,*fpq,*fpetype,*fpstep,*fptime,*fprand;
  FILE *fp,*fpindic,*fpenergy;
  int N,E;
  char saveprec[STRLTH]="", saveprecE[STRLTH]="";
  int ir_nn;
  double ir_t,ir_T;
  long long int *allrandcalls;

  char data_format[STRLTH] , sample_dir [LSTRLTH] ;

  strcpy( data_format , g_inputs.format_loadsave ) ;

  if ( strstr(  g_inputs.misc_opt , "nodiskwrite" ) != NULL ) { return 0; } ;

  if ( ( (  g_inputs.onelastsnp == 1 ) && ( indic != -1 ) ) || 
       ( g_inputs.onelastsnp == -1 )  ) {
    return( 0 ) ;
  }

  N= g_inputs.N;E= g_inputs.E;

  sprintf(saveprec,"%% 1.%de ",SAVE_PREC -> foto_r);
  sprintf(saveprecE,"%% 1.%de ",SAVE_PREC -> foto_E);
  
  if  (indic>=0){
    if ( is_snp_save_now( g_inputs.step , "snp_full" ) )
      indic=step;
    else
      return(0);
  }

  if ( 2 == xmdyn_apply_shortcut( "check" ) ) 
  {
      link_xmdyn_snpfull( step , -1 ) ;
      return 0 ;
  }



  fotofilenames(indic);
  dircreate( ALLFILES->fotodir ) ;


  if ( 0 == strcmp( data_format , "000" ) )
  {
  if ( (indic==-1) || (indic==-2) ){
    fpindic=fopen(ALLFILES -> data2diskindicator,"a");
    fprintf(fpindic,"%d\n",indic);
    fclose(fpindic);
  }
  }
  else
  {
    fp = fopen( ALLFILES -> savestatus , "w" ) ;
    fprintf(fp,"saving_started\n");
    fclose(fp);
  }

  
  fpstep=fopen(ALLFILES -> stepfoto,"w") ;
  fprintf(fpstep,"% d\n",step);
  fclose(fpstep);

  fptime=fopen(ALLFILES -> timefoto,"w") ;
  fprintf(fptime,"%le :time used for calculation in sec.\n",
	  myclock(-1));
  fclose(fptime);

  if ( ! xmdyn_apply_shortcut( "check" ) ) 
  {
 
  fpr1=fopen(ALLFILES -> r1foto,"w");
  fpv1=fopen(ALLFILES -> v1foto,"w");
  fpm=fopen(ALLFILES -> massfoto,"w"),
  fpetype=fopen(ALLFILES -> etypefoto,"w");
  fpq=fopen(ALLFILES -> chargefoto,"w");
  for (i=0;i<N+E;++i){
    fprintf(fpm,saveprec, g_system.m[i]);
    fprintf(fpq,"% 2d ", g_system.qi [i]);
    fprintf(fpetype,"%c ", g_system.ptype [i]);
    for (j=0;j<3;++j){
      fprintf(fpr1,saveprec, g_system.r1 [i][j]);
      fprintf(fpv1,saveprec, g_system.v1 [i][j]);
    }
    fprintf(fpr1,"\n");
    fprintf(fpv1,"\n");
  }
  fprintf(fpm,"\n");
  fprintf(fpq,"\n");
  fprintf(fpetype,"\n");
  fclose(fpr1),fclose(fpv1),fclose(fpm),fclose(fpq),fclose(fpetype);

  fpenergy=fopen(ALLFILES -> energyfoto,"w");
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Eintotal);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Eintotal_th);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einph);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einfl);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einau);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einsi);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einrec);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einitial);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Elost);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einfi);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Einchmig);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Ebound);
  fprintf(fpenergy,saveprecE, g_ENERGY_VALUES.Efl_photon);
  fclose(fpenergy);

	  
 
  fprand=fopen(ALLFILES -> randcallsfoto,"w") ;
  allrandcalls=randcalls_counts('g');
  int Num_Rank = 1 ;
  fprintf(fprand,"%d ",Num_Rank);
  for (i=0;i<Num_Rank;++i){
    fprintf(fprand,"%lld ",allrandcalls[i]);
  }
  fprintf(fprand,"\n");
  fclose(fprand);

  //if (indic<0) {
    fp=fopen(ALLFILES -> prev_r2foto,"w");
    fpr2=fopen(ALLFILES -> r2foto,"w");
    fpv2=fopen(ALLFILES -> v2foto,"w");
    for (i=0;i<N+E;++i){
      for (j=0;j<3;++j){
	fprintf(fp,saveprec, g_system.prev_r2 [i][j]);
	fprintf(fpr2,saveprec, g_system.r2 [i][j]);
	fprintf(fpv2,saveprec, g_system.v2 [i][j]);
      }
      fprintf(fp,"\n");
      fprintf(fpr2,"\n");
      fprintf(fpv2,"\n");
    }
    fclose(fpr2);fclose(fpv2);fclose(fp);
  //}

  fp=fopen(ALLFILES -> el_configfoto,"w");
  for (i=0;i<N;++i){
    fprintf(fp,"## %d %d %d %d\n",i, g_system.Z [i],g_atomic_data[i].e_from,g_atomic_data[i].e_to);
    for (j=0;j<=g_atomic_data[i].maxshellnum;++j){
      fprintf(fp,"%2d %2d\n",j,g_atomic_data[i].shell[j].NE);
    }
  }
  fclose(fp);

  fp=fopen(ALLFILES -> Zallfoto,"w");
  for (i=0;i<N;++i){
    fprintf(fp,"%d ", g_system.Z [i]);
  }
  fclose(fp);

  fp=fopen(ALLFILES -> r1fotob,"w");
  fwrite( g_system.r1 ,sizeof(double),3*( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> r2fotob,"w");
  fwrite( g_system.r2 ,sizeof(double),3*( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> prev_r2fotob,"w");
  fwrite( g_system.prev_r2 ,sizeof(double),3*( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> v1fotob,"w");
  fwrite( g_system.v1 ,sizeof(double),3*( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> v2fotob,"w");
  fwrite( g_system.v2 ,sizeof(double),3*( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> massfotob,"w");
  fwrite( g_system.m ,sizeof(double),( g_inputs.N+ g_inputs.E),fp);
  fclose(fp);
  fp=fopen(ALLFILES -> energyfotob,"w");
  fwrite(& g_ENERGY_VALUES.Eintotal,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Eintotal_th,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einph,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einfl,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einau,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einsi,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einrec,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einitial,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Elost,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einfi,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Einchmig,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Ebound,sizeof(double),1,fp);
  fwrite(& g_ENERGY_VALUES.Efl_photon,sizeof(double),1,fp);
  fclose(fp);


  fp=fopen(ALLFILES -> motherionfoto,"w");
  for ( ii = 0 ; ii <  g_inputs.N+ g_inputs.E ; ++ii ) {
      fprintf( fp , "% d " , g_system .motherion[ii] ) ;
  }
  fclose(fp);


  fp=fopen(ALLFILES -> uidfoto,"w");
  for ( ii = 0 ; ii <  g_inputs.N+ g_inputs.E ; ++ii ) {
      fprintf( fp , "% lld " , g_system .uid[ii] ) ;
  }
  fclose(fp);



  fp=fopen(ALLFILES -> instrecfotob,"w");
  module__recombination_loadsave( "save" , &fp ) ;
  fclose(fp);

  } // shortcut end


  fp = fopen( ALLFILES -> chargetransfer_b , "w" ) ;
  fclose( fp ) ;
  


  if ( 0 == strcmp( data_format , "000" ) )
  {
  if ( (indic==-1) || (indic==-2) ){
    fpindic=fopen(ALLFILES -> data2diskindicator,"a");
    fprintf(fpindic,"%d\n",indic);
    fclose(fpindic);
  }
  }
  else
  {
    fp = fopen( ALLFILES -> savestatus , "w" ) ;
    fprintf(fp,"saving_finished\n");
    fclose(fp);
  }

 
  return(0);
}


//############################## FUNCTION #################################

int data_loadsave( char *a_todo , char *a_format , int a_step , enum g_option_string  a_status )
{
    double chkp_dt = g_inputs.chkp_dt ;

    static unsigned int t_start = 0 ;
    unsigned int t_now ;
    char tmpstr[LSTRLTH] ;

    cache_filenames(0,a_step);

    if ( 0 == strcmp( a_todo , "progress" ) )
    {
      char *plog ;
      xmdyn_clock( "report" , &plog ) ;
      actsteptofile( a_step , a_status , plog );
      xmdyn_clock( "clear" ) ;
    }


    if ( 0 == strcmp( a_todo , "snp" ) )
    {
        if ( NULL != strstr( a_format , "000" ) ) { 
            if (startingdataindex()!=-2) { data2disk( a_step ); }
        }
        if ( NULL != strstr( a_format , "001" ) ) {  save_xmdyn_snp( "snp" , a_step ) ;  }
    }

    if ( 0 == strcmp( a_todo , "E" ) ) { save_xmdyn_snp( "E" , a_step ) ; }



    if ( 0 == strcmp( a_todo , "snp_full" ) )
    {
        get_xmdyn_snp_fname( "snpfull_dir" , a_step , tmpstr ) ;
        foto2disk( a_step , a_step );
    }

    if ( 0 == strcmp( a_todo , "snp_full_restart" ) )
    {
        if ( 0 == t_start ) { t_start = (unsigned int)time(NULL) ; }
        if ( DUMMY == a_status )
        {
            t_now = (unsigned int)time(NULL) ;
            if ( (double)(t_now - t_start) < chkp_dt ) { return 0 ; } 
        }
        foto2disk(a_step,-1);
        foto2disk(a_step,-2);
        t_start = (unsigned int)time(NULL) ;
    }


    cache_filenames(1,a_step);
    cache_filenames(2,a_step);

    if ( 0 == strcmp( a_todo , "cache-close" ) )
    {
        cache_filenames(10, a_step );
        cache_filenames(11, a_step );
        cache_filenames(12, a_step );
        cache_filenames(13, a_step ); // finalize
    }


    return 0 ;
}






//############################## FUNCTION #################################

int actsteptofile( int step , enum g_option_string a_status , char *timestr )
{
  FILE *fp;
  time_t t; time(&t);


  {

  if ( a_status == PAUSE100 )  {
    fp=fopen(ALLFILES -> progress_file,"a");
    if ( fp )   { fprintf(fp,"  CONTROL: PAUSE100\n") ; fclose(fp) ; return 0 ; }
    return -1 ;
  }
  if ( a_status == DONE )  {
    fp=fopen(ALLFILES -> progress_file,"a");
    fprintf(fp,"  DONE\n");
    fclose(fp);
    return 0 ;
  }
  if ( a_status == INTERRUPTED )  {
    fp=fopen(ALLFILES -> progress_file,"a");
    fprintf(fp,"  INTERRUPTED\n");
    fclose(fp);
    return 0 ;
  }
  if  ((step+1)% g_inputs.progress_dump==0){
    fp=fopen(ALLFILES -> progress_file,"a");
    fprintf(fp,"% 10d   ---   M: %12d   D: %s", step, xmdyn_get_meminfo("malloc") , asctime(localtime(&t)) ) ;
#ifdef XMDYN_XRAYPAC_MORELOG
    if ( NULL != timestr ) { fprintf(fp,"%s", timestr ) ; }
#endif // XMDYN_XRAYPAC_MORELOG
    fclose(fp) ;
  }

  }

  return 0 ;

}




//############################## FUNCTION #################################

void check_if_continued(int step,int laststep) {
  FILE *fp;
  char s[STRLTH];
  if ( (fp=fopen(ALLFILES -> progress_file,"r")) ) {
    while (fscanf(fp,"%s",s) != EOF)
      ;
    fclose(fp);
    if ( (step!=0) && (step!=laststep) && (strcmp(s,"INTERRUPTED")!=0) ) {
      actsteptofile(-2 , INTERRUPTED , NULL );
    }
    return;
  }
}







//############################## FUNCTION #################################

int makingfotomoments(void){

  int i;
  double dfoto;
  char fotofile[MSTRLTH]="",fotodir[MSTRLTH]="";
  FILE *fp;

  int fotomomentsdim;
  char fotofromfile;


  fotomomentsdim= g_inputs.fotomomentsdim;
  fotofromfile= g_inputs.fotofromfile;
  strcat(strcat(strcat(strcat(strcat(strcat(fotofile, g_inputs.rootdir),"/data/"),
			       g_inputs.runnumber),"/bio"), g_inputs.runnumber),"zfoto.txt");
  strcat(strcat(strcat(strcat(fotodir, g_inputs.rootdir),"/data/"), g_inputs.runnumber),
	 "/foto");

  if ( NULL != strstr( g_inputs.format_loadsave , "000" ) ) { 
      dircreate(fotodir);
  }

  if (fotofromfile=='y'){
    if (!(fp=fopen(fotofile,"r"))) return(nofile(fotofile));
    if ( (fscanf(fp,"%d",& g_inputs.fotomomentsdim)==0)) {
      printf("There's problem with fotofile in calc. %s!\n", g_inputs.runnumber);
      return(-1);
    }
    fotomomentsdim= g_inputs.fotomomentsdim;
    fotomoments=calloc(fotomomentsdim,sizeof(int));
    for (i=0;i<fotomomentsdim;++i)
      if ( (fscanf(fp,"%d",&((fotomoments)[i]))==0) ||
	   ((fotomoments)[i]==0) ) {
	printf("There's problem with fotofile in calc. %s!\n", g_inputs.runnumber);
	return(-1);
      }
  }
  else {
    if (fotomomentsdim==0) {
      fotomomentsdim= g_inputs.fotomomentsdim=1;
      fotomoments=calloc(1,sizeof(int));
      (fotomoments)[0]= g_inputs.nstep+2;
    }
    else {
      fotomoments=calloc(fotomomentsdim,sizeof(int));
      dfoto= g_inputs.nstep/((double)(fotomomentsdim));
      for (i=0;i<fotomomentsdim;++i)
	(fotomoments)[i]=floor((i+1)*dfoto);
    }
  }

  return(0);
}



//############################## FUNCTION #################################

void cp_source_input(char *commandline,char *programname,char *version){
  char cp_input[LSTRLTH]="",cp_source[LSTRLTH]="",cp_exe[LSTRLTH]="";
  char mkdir_archive[LSTRLTH]="",archive_dir[MSTRLTH]="",source_dir[MSTRLTH]="";
  char commline_write[LSTRLTH]="";
  char xstr[LSTRLTH] ;
  
  strcat(strcat(strcat(archive_dir," "),ALLFILES -> datadir),"/archive");
  strcpy(archive_dir,ALLFILES -> archivedir);
  strcat(strcat(strcat(source_dir, g_inputs.rootdir),"/cd_source/"),version);

  if (strcmp(commandline,"rm")==0){
    strcat(strcat(mkdir_archive,"rm -r "),archive_dir);
    system(mkdir_archive);
    return;
  }
    
  if (strcmp(commandline,"input.txt-link")==0){
    strcpy( xstr , "" ) ;
    strcat( strcat(strcat( xstr ,"cd "), ALLFILES -> datadir ) , "; ln -s xparams.txt input.txt" ) ;
    system( xstr );
    return;
  }
    
  strcat(strcat(mkdir_archive,"mkdir -p "),archive_dir);
  strcat(strcat(strcat(strcat(cp_input,"cp "), g_inputs.inputfile)," "),archive_dir);
  strcat(strcat(strcat(strcat(cp_exe,"cp "),programname)," "),archive_dir);
  strcat(strcat(strcat(strcat(strcat(cp_source,"tar -czf "),archive_dir),
                       "/src.tgz "),source_dir)," &>/dev/null ");
  strcat(strcat(strcat(strcat(strcat(commline_write,"echo \" "),commandline),
                       "\" >"),archive_dir),"/zzz-commandline.txt");
  
  system(commline_write);
}





//############################## FUNCTION #################################

int cd_filename(char* fname,char *id,long long int step)
{
  int i;
  char st[LSTRLTH], st1[LSTRLTH],  
       ddir[LSTRLTH];
  sprintf(st1,"%lld",step);
  strcpy(st,"");
  for (i=0;(unsigned)i<15-strlen(st1);++i) { strcat(st,"0"); }
  strcat(st,st1);
  strcpy(fname,"");
  if (strstr(id,"tmp")!=NULL) 
  {
    strcpy(ddir,ALLFILES -> tempdir);
  }
  strcat(strcat(strcat(strcat(strcat(fname,ddir),"/"),id),"-"),st);

  return 0 ;
}





//############################## FUNCTION #################################

int data_save(char *fname,char *dformat,int dim1,int dim2,void *data,char *opt) 
{
  FILE *fp;
  char optopen[2],dtype[SSTRLTH];
  int optbinary=0, opttext=0;
  int ii,jj;
  double *ddata;

  strcpy(optopen,"a"); optbinary=0; opttext=0;
  if (strstr(opt,"a")!=NULL) { strcpy(optopen,"a"); }	//append 
  if (strstr(opt,"w")!=NULL) { strcpy(optopen,"w"); }	//write
  if (strstr(opt,"b")!=NULL) { optbinary=1; }	//binary
  if (strstr(opt,"t")!=NULL) { opttext=1; }	//text-ascii
  
  if (strstr(dformat,"ld")!=NULL) 
  {
    strcpy(dtype,"longint");
  }
  else if (strstr(dformat,"d")!=NULL)
  {
    strcpy(dtype,"int");
  }
  else if (strstr(dformat,"le")!=NULL)
  {
    strcpy(dtype,"double");
  }

  if ( (strstr(opt,"N")!=NULL) && (dim1*dim2==0) )
  {
    return(0);
  }

  if (opttext==1)
  {
    if ( fp=fopen(fname,optopen),fp==NULL ) 
    {
      return(-1);
    }
    ddata = (double*) data;
      for (ii=0;ii<dim1;++ii) {
        for (jj=0;jj<dim2;++jj) {
          fprintf(fp,"%1.20le ",ddata[ii*dim2+jj]);
        }
	fprintf(fp,"\n");
      }
    fclose(fp);
  }


  return 0 ;
}




//############################## FUNCTION #################################

int cd_data_save(char *id,long long int step,char *dformat,int dim1,int dim2,void *data,char *opt) 
{
  char fname[LSTRLTH];
  cd_filename(fname,id,step);  
  data_save(fname,dformat,dim1,dim2,data,opt); 

  return 0 ;
}





//############################## FUNCTION #################################

int dbg_filedump(char *id,int subid,int dim1, int dim2, void* data,char *opt) 
{
  static int fmemostep=-1, pmemostep=-1, xxxmemostep=-1;
  static long long int fs=0, ps=0, xxxs=0;
  static int finit=1, pinit=1, xxxinit=1;
  long long int step;

  if (strstr(id,"xxx")!=NULL)
  {
    if ( (xxxinit==1) || (xxxmemostep!= g_inputs.step) )
    {
      xxxmemostep =  g_inputs.step;
      xxxs=0;
      xxxinit=0;
    }
    if (subid==-1) { xxxs++; return(0); }
    step = (long long int) g_inputs.step * (long long int)10000 +
           xxxs;
    step = step*100 + subid;
    cd_data_save(id,step,"%1.20le",dim1,dim2,data,opt);
  }
  if (strstr(id,"force")!=NULL)
  {
    if ( (finit==1) || (fmemostep!= g_inputs.step) )
    {
      fmemostep =  g_inputs.step;
      fs=0;
      finit=0;
    }
    if (subid==-1) { fs++; return(0); }
    step = (long long int) g_inputs.step * (long long int)10000 +
           fs;
    step = step*100 + subid;
    cd_data_save(id,step,"%1.20le",dim1,dim2,data,opt);
    //fs++;
  }
  if (strstr(id,"pot")!=NULL)
  {
    if ( (pinit==1) || (step!=pmemostep) )
    {
      pmemostep = step;
      ps=0;
      pinit=0;
    }
    step = (long long int) subid;
    cd_data_save(id,step,"%1.20le",dim1,dim2,data,opt);
    ps++;
  }
  if ( (strstr(id,"mean")!=NULL) || 
       (strstr(id,"max")!=NULL)  ||
       (strstr(id,"std")!=NULL) 
     ) 
  {
    step = (long long int) subid;
    cd_data_save(id,step,"%1.20le",dim1,dim2,data,opt);
  }
  return(0);
}






//############################## FUNCTION #################################

int print_extrainfo_out(char *programname,char *versionnumber,char startend,int *pid_numbers,char **hostnames){
  int i;
  FILE *fp;
  char messagefile[LSTRLTH]="";


  strcpy( messagefile , ALLFILES -> progress_file ) ;


  switch (startend){

   case 'v':
    if ( g_inputs.started==0) {
      fp = fopen(messagefile,"a");
      fclose(fp);
       g_inputs.started=1;
    }
    break;

   case 's':
    fp=fopen(messagefile,"a");
    fprintf(fp,"@ Executable: %s, version: %s\n",programname, g_inputs.xmdyn_version );
    fprintf(fp,"@ PIDs:\n");
    int Num_Rank = 1 ;
    for (i=0;i<Num_Rank;++i){
      fprintf(fp,"@   process %d: host: %s pid: %d \n",i,hostnames[i],pid_numbers[i]);
    }
    fclose(fp);

    fp = fopen(messagefile,"a");
    fprintf(fp,"@ start: %s\n",cd_currdate());
    fprintf(fp,"@ ##### ##### ##### ##### ##### #####\n");
    fclose(fp);
    break;

   case 'e':
    fp = fopen(messagefile,"a");
    fprintf(fp,"@ ##### ##### ##### ##### ##### #####\n");
    fprintf(fp,"@ end: %s\n",cd_currdate());
    fclose(fp);
    break;

   case 'i':
    fp = fopen(messagefile,"a");
    fprintf(fp,"@ ##### ##### ##### ##### ##### #####\n");
    fprintf(fp,"@ stopped: %s\n",cd_currdate());
    fclose(fp);
    break;

   case 'r':
    remove(messagefile);
    break;
    
  }


  return 0 ;
}






//############################## FUNCTION #################################

int xmdyn_apply_shortcut( char *command )
{
    int N = g_inputs.N ;
    int E = g_inputs.E ;
    double DT = g_inputs.DT ;
    int *qsys = g_system.qi ;

    // shortcut_status:  0: undecided (return 0) ; 
    //                  -1: cancelled (return 0) ; 
    //                  +1: apply (return 1) ;
    static int shortcut_status = 0 ;  // Undecided
    static int step = -1 ; 
    static int step_saved = 0 ;
    int value = -1 , wait_step ;
    int ii , Qtot ;


    // CANCEL:
    if ( 0 == strcmp( command , "cancel" ) )
    {
        if ( NULL == strstr( g_inputs.shortcut_comm , "fragQ:" ) )
        { 
            shortcut_status = -1 ; 
        }
        return 0 ; 
    }

    if ( 0 == strcmp( command , "execute" ) ) { if ( 0 < shortcut_status ) { shortcut_status = 2 ; } } 

    // If decided already:
    if ( -1 == shortcut_status ) { return 0 ; }
    if (  0 <  shortcut_status ) { return shortcut_status ; }

    if ( step == g_inputs.step ) { if ( 0 == shortcut_status ) { return 0 ; } }
    else { step = g_inputs.step ; }

    // Check:
    if ( 0 == strcmp( command , "check" ) )
    {

        if ( 0 == strcmp( g_inputs.shortcut_comm , "full" ) )
        {
            shortcut_status = 1 ;   return 1 ;  
        }


        else if ( NULL != strstr( g_inputs.shortcut_comm , "s" ) ) 
        { 
            double xxx ;
            sscanf( g_inputs.shortcut_comm , "%le" , &xxx ) ;
            value = (int) floor( ( g_inputs.T0 + xxx * ( g_inputs.T_1 / sqrt( 8*log(2.) ) ) ) / DT ) + 
                    (int) floor( ( g_inputs.T0 + xxx * ( g_inputs.T_2 / sqrt( 8*log(2.) ) ) + g_inputs.xrsd_delay) / DT ) ; 
        }

        /**  Below must be the last but one **/
        else if ( 0 != strcmp( g_inputs.shortcut_comm , "-" ) )
        { 
            sscanf( g_inputs.shortcut_comm , "%d" , &value ) ; 
        }

        /**  Below must be the last one **/
        else {  return 0 ;  }

        if ( value != -1 ) 
        {
            Qtot = 0 ;
            for ( ii = 0 ; ii < N+E ; ii++ ) { Qtot += abs( qsys[ii] ) ; }
            if ( ( g_inputs.step > value ) && ( 0 == Qtot ) )  
            {  
                shortcut_status = 1 ; print_log("Shortcut no-ionization fulfilled at step ",(void*)&step,"int","flush"); return 1 ;  
            }
        }
    }

    return 0 ;
}

