#!/usr/bin/python 
# ANALYZATION SCRIPT FOR XFSSH/XMOLECULE_MC OUTPUT
# Ludger Inhester  OCTOBER 2015

# This script provides a python module that reads in relevant trajectory data from xFSSH/xmolecule_mc output
# run as standalone with multiple output files a argument and it will print out final states (charges, KER, N absorbed photons, distances)
# i.e.  python analyze.py outputfile
#
# you can use it also as a python module to import the relevant routines and data structures
# use the python statement  "from analyze import *" to have them available in your python script
# most important routines are
#
# read_trajectories(filename):
#   reads in all trajectories in a list from file filename
# read_parameters(filename):
#   reads in pulse parameters
# trajectory data is stored as a list of trajectories
# each trajectory is a list of steps, which can be of class Sstep,step,Nstep:
# Sstep for starting step, step for el step , Nstep for nuclear step
# these are specific classes contain relevant information about the specific step
# and routines to calculate distances, kinetic energy release, charge population ...

import sys
from math import exp,sqrt,log,pi
import gzip

AMU2AU=1822.939417794
FS2AU=0.0242


massC=12.0110*AMU2AU
massI=126.9*AMU2AU
massH=1.0*AMU2AU

mass=[massI,massC,massH,massH,massH]
atoms =["I","C","H","H","H"]
atomsZ=[ 53,  6,  1,  1,  1]

class gross:
        """ This class stores the gross population data
        """

        mo=0
        orbe=0.0
        focc=0.0

        p=[0. for i in range(0,len(atoms))]
        def __init__(self,mo,orbe,focc,p):
                self.mo=mo
                self.orbe=orbe
                self.focc=focc
                self.p=p

class Nstep:
        """ This class stores the data for a MD step.
        methods:
        KER(prev_step): returns the kinetic energy release derived by numerical differentiation to a previous Nstep prev_step
        CI_distance: returns the carbon iodine distance
        CH1_distance: returns the carbon H1 distance
        ...
        """
        occ=[]
        time=0.0

        pos=[0. for i in range(0,3*len(atoms))]
        vel=[0 for i in range(0,3*len(atoms))]
        q=[0. for i in range(0,len(atoms))]

        MullikenCharge=[]

        def __init__(self, time,occ,pos,vel,MullikenCharge=None):
                self.time=time
                self.occ=occ
                self.MullikenCharge=MullikenCharge                

                self.vel=vel
                self.q=MullikenCharge
                self.pos=pos

        def __str__(self):
                return "time=% 5.2f occ=%s d_CI=%f d_CH=%f " % (self.time,self.occ, self.atom_distance(0,1), self.atom_distance(1,2) )

        def KER_atom(self,atom):                
                #dt=(self.time-prev_step.time)/FS2AU
                #vel=[(a - b)/dt for a, b in zip(self.Ipos, prev_step.Ipos)]
                vel=self.vel[3*atom:3*(atom+1)]
                vel2=sum(vel[i]*vel[i] for i in [0,1,2] )
                return(0.5*mass[atom]*vel2)

        def mom(self):                
                # dt=(self.time-prev_step.time)/FS2AU
                # mom=[massI*(a - b)/dt for a, b in zip(self.Ipos, prev_step.Ipos)]
                mom=[[0.,0.,0.] for i in [0,1,2] for a in atoms]
                for i in range(0,len(atoms)):
                        for j in [0,1,2]:
                                mom[i][j]=mass[i] * self.vel[3*i+j]
                return(mom)               

        def mom_sum(self):
                mom=self.mom()
                mom_s=[0. for i in range(0,len(atoms))]
                for j in [0,1,2]:
                        mom_s[i]=sum(mom[i][j] for i in range(0,len(atoms)))
                return(mom_s)
        
        def KER(self):
                "kinetic energy per atom"
                mom=self.mom()
                KER=[ sum([mom[i][j]*mom[i][j]/(2.*mass[i]) for j in range(0,3)]) for i in range(0,len(atoms))]
                return(KER)

        def KER_COM_substracted(self):
                "kinetic energy per atom with center of mass momentum substracted"

                mom=self.mom()
                COM_mom=[ sum([ mom[i][j] for i in range(0,len(atoms)) ]) for j in [0,1,2] ]
                COM_vel=[ COM_mom[j]/sum(mass) for j in [0,1,2] ]
                mom=[ [mom[i][j]-COM_vel[j]*mass[i] for j in [0,1,2] ] for i in range(0,len(atoms))]
                KER=[ sum([mom[i][j]*mom[i][j]/(2*mass[i]) for j in [0,1,2] ]) for i in range(0,len(atoms))]
                return(KER)

        def atom_distance(self,atom1,atom2):
                d=sqrt(sum( (self.pos[3*atom1+i]-self.pos[3*atom2+i])*(self.pos[3*atom1+i]-self.pos[3*atom2+i]) for i in [0,1,2] ))
                return(d)



class Sstep:
        """ This class stores the data for the starting step in each trajectory 
        """
        occ=[]
        gross_mo_pop=[]
        time=-10000.0 # something very negative
        energy=0.0
        typ="S"

        q=[1.0*Z for Z in atomsZ]

        def __init__(self, time,energy,occ,gross_mo_pop):
                self.energy=energy
                self.occ=occ
                self.gross_mo_pop=gross_mo_pop
                self.time=time

                self.q=[1.0*Z for Z in atomsZ]
                for gross in gross_mo_pop:
                        for i in range(0,len(atoms)):
                                self.q[i]=self.q[i]-gross.focc*gross.p[i]

        def __str__(self):
                return "time=% 5.2f occ=%s energy=%f q=%s " % (self.time,self.occ,self.energy,self.q)
                

class step:
        """ This class stores the data for electronic step 
        object 
        qI,qC,qH1,qH2,qH3 contain the Mulliken charges
        """
        energy=0.0
        focc=[]
        occ=[]
        gross_mo_pop=[]
        time=0.0

        typ="?"
	olap=0.0
        MullikenCharge=[]

        q=[1.0*Z for Z in atomsZ]

        def __init__(self, time,energy,occ,gross_mo_pop,typ,olap,MullikenCharge=None):
                self.energy=energy
                self.occ=occ
                self.gross_mo_pop=gross_mo_pop
                self.time=time
                self.MullikenCharge=MullikenCharge
                self.typ=typ
		self.olap=olap

                self.q=[1.0*Z for Z in atomsZ]
                for gross in gross_mo_pop:
                        for i in range(0,len(atoms)):
                                self.q[i]=self.q[i]-gross.focc*gross.p[i]
#                self.q=MullikenCharge
        def __str__(self):
                return "time=% 5.2f occ=%s energy=%f q=%s" % (self.time,self.occ,self.energy,self,q)
        


def check_COM_KER(traj):
        nucstep_index=[i  for i in range(0,len(traj)) if isinstance(traj[i],Nstep)]
        M=sum(mass)
        mom_sum=traj[nucstep_index[-1]].mom_sum()
        print '# COM KER: ',sum(p*p/(2*M) for p in mom_sum)*27.21,'eV'
        if ( sum(p*p/(2*M) for p in mom_sum) > 3/27.21 ):
                print '# Warning: trajectory has more than 3eV center of mass motion:',sum(p*p/(2*M) for p in mom_sum)*27.21,"eV"
                return(False)
        else:
                return(True)





def read_parameters(filename):
        """ returns fluence and pulse duration from outputfile filename
        trajectory data is returned as a list of trajectories, containing a list of steps
        """
        f=open(filename,'r')
        for line in f:
                if(line.startswith('#   Number of photons = ')):
                        a=line.split('=')
                        num_photons=float(a[1])
                if(line.startswith('#   Fluence ')):
                        a=line.split('=')
                        fluence=float(a[1].split()[0])
                if(line.startswith('#   Temporal width')):
                        a=line.split('=')
                        FWHM=float(a[1])
        f.close()
        return((fluence,FWHM))



def read_filter(filename,filefilter=[]):
        """ returns array of filenames and valid trajectories
        """
        f=open(filename,'r')
        file_filter=[]
        for line in f:
                if(line.startswith('#FWHM')):
                        a=line.rstrip().split('_')
                        idnum=a[2]
                        filter_per_id=[]
                elif(len(line)==1):
                        if(len(filter_per_id)>0):
                                file_filter.append((idnum,filter_per_id))
                elif(len(line.split())==3):
                        a=line.split()                        
                        filter_per_id.append((int(a[0]),a[1]))
                
        f.close()

        return file_filter


def read_trajectories(filename,traj_filter=[]):
        """ reads in trajectory from outputfile filename
        
        """

        print "#Reading file", filename
        if(filename.endswith(".gz")):
                f=gzip.open(filename,'r')
        else:
                f=open(filename,'r')
        traj_list=[]
        traj=[]
        traj_gross_pop=[]
        
        # some flags
        has_ended=0
        has_dissociated=0
        is_neutral_gross_pop=1
        has_failed=0

        neutralE=0.
        neutralocc="222222222222222222222222222222200000000"
        neutraltime=-100000

        traj_MullikenCharge=[]
        for line in f:
                if(line.startswith('#MullikenCharge:') and not "Final" in line ):
                        a=line.split()
                        #print a 
                        #print a[2],a[18],a[19],a[20],a[21],a[22]
                        
                        try:
                                t=float(a[2])
                                q=[0. for i in range(0,len(atoms))]
                                for i in range(0,len(atoms)):
                                        q[i]=float(a[18+i])

                                traj_MullikenCharge.append((t,q))
                        except:
                                print "#Warning: MullikenCharge not parsable:", line

                if(line.startswith('#GROSS')  ):
                        a=line.split()
                        if(a[1]=="MO"):
                                gross_pop=[]
                                continue
                        mo=int(a[1])
                        orbe=float(a[2])
                        occ=float(a[3])

                        p=[ float(a[4+i]) for i in range(0,len(atoms))]
                        gross_pop.append(gross(mo,orbe,occ,p))
                        # when we reach mo=39 store the gross_population into gross object
                        if(mo==39): 
                                # the first gross population is the neutral state gross pop: store it differently (it will never appear again)
                                if(is_neutral_gross_pop==1):
                                        neutral_gross_pop=gross_pop 
                                        is_neutral_gross_pop=0
                                else:
                                        traj_gross_pop.append(gross_pop)

                if(line.startswith('#genFSSH:END:')):
                        has_ended=1
                        print line.rstrip()
                        a=line.split()                                           

                if(line.startswith(' #dump MCtrajectory: status:MCstep failed due to Error')):
                        print line.rstrip()
                        a=line.split()
                        has_failed=1

                if(line.startswith('#genFSSH:Start')):
                        # store the old trajectory:
                        if(has_ended==1):
                                if(has_failed==0 and filter_out==0 and parsing_error==0):
                                        traj.insert(0,Sstep(neutraltime ,neutralE , neutralocc, neutral_gross_pop ))
#                                        if(check_COM_KER(traj) or True):
                                        traj_list.append(traj)
                                                
                                        print "# done",len(traj),"steps "
                                if(filter_out==1):
                                        print "#traj filtered"
                                if(has_failed==1 ):
                                        print "#traj failed"
                                if(parsing_error==1 ):
                                        print "#parsing error"

#                        else:
#                                print "# trajectory has Started but never ended"

                        print line.rstrip()
                        a=line.rstrip().split()
                        traj_index=int(a[3])
                        # reset flags and lists for new trajectory
                        has_failed=0        
                        has_ended=0
                        has_dissociated=0
                        parsing_error=0
                        filter_out=0
                        traj=[]
                        traj_gross_pop=[]
                        traj_MullikenCharge=[]

                        if(len(traj_filter)==0):
                                filter_out=0
                        elif(traj_index > len(traj_filter)  ):
                                print "#filter too short"
                                filter_out=1
                        elif( traj_index-1 != traj_filter[traj_index-1][0 ] ):
                                print "something is wrong!", traj_filter[traj_index-1][0], traj_index-1
                                quit()
                        else:
                                if(traj_filter[traj_index-1][1]=="Y"):
                                        filter_out=0
                                else:
                                        filter_out=1


                if(line.startswith('#MCtraj:N') and has_dissociated==0  and has_failed==0 and filter_out==0):
                        a=line.split()
                        time=float(a[1])
                        occ=a[2]

                        try:
                                pos=[float(a[ 3              + i ]) for i in range(0,3*len(atoms))]
                                vel=[float(a[ 3+3*len(atoms) + i ]) for i in range(0,3*len(atoms))]
                        except:
                                print "************ parsing error ***********"
                                parsing_error=1
                                break

                        # when the coordinates become larger than 1000 a.u. we have problems to parse the line
                        # solution: we skip parsing lines if any coordinate is large than 800 a.u.
                        if (any( [ abs(c) >900 for c in pos ])):
                                has_dissociated=1
                        
                        Mullikenstep=min(sum(  M[0]<time for M in traj_MullikenCharge),len(traj_MullikenCharge)-1)
                        if(len(traj_MullikenCharge)==0):
                                traj.append(Nstep(time,occ,pos,vel,None))
                        else:
                                traj.append(Nstep(time,occ,pos,vel,traj_MullikenCharge[Mullikenstep][1]))
        
                if(line.startswith('#MCtraj::E') and has_failed==0 and filter_out==0):
                        a=line.split()
                        time=float(a[1])
                        occ=a[2]
                        i=3
                        if(len(a)==6):
                                typ=a[i]
                                i=i+1
                        else:
                                typ="S"
                        olap=float(a[i])
                        i=i+1
                        energy=float(a[i])                        
                        # in case of Starting entry (which is only given for static nuclei calculations, save the values)
                        if(typ=="S"):
                                neutraltime=time
                                neutralE=energy
                                neutralocc=occ                                
                        else:
                                # we need the number of electronic step to assign the right gross_pop for this step
                                elstep=sum(isinstance(s,step) for s in traj)
                                Mullikenstep=min(sum(  M[0]<time for M in traj_MullikenCharge),len(traj_MullikenCharge))
                                if(len(traj_MullikenCharge)==0):
                                        traj.append(step(time,energy,occ,traj_gross_pop[elstep],typ,olap))
                                else:
                                        traj.append(step(time,energy,occ,traj_gross_pop[elstep],typ,olap,traj_MullikenCharge[Mullikenstep][1]))
                        
        if(has_ended==1):
                if(has_failed==0 and filter_out==0 and parsing_error==0):
                        traj.insert(0,Sstep(neutraltime ,neutralE , neutralocc, neutral_gross_pop ))
#                        if(check_COM_KER(traj) or True):
                        traj_list.append(traj)
                        
                        print "# done",len(traj),"steps "
                if(filter_out==1):
                        print "#traj filtered"
                if(has_failed==1 ):
                        print "#traj failed"
                if(parsing_error==1 ):
                        print "#parsing error"


        f.close()
        return(traj_list)


def pulse_envelope(FWHM,t):
#   f = e^( -t^2 / 2 sigma^2 ) / ( sigma sqrt(2pi) )
        sigma = FWHM / 2. *1./ sqrt( 2 * log(2.) )
        f=exp(-t*t/(2.*sigma*sigma))/(sigma*sqrt(2.*pi))
        return f


if __name__ == "__main__":
        if (len(sys.argv)<1):
                print "usage",sys.argv[0]," xFSSH outputfiles "
                print "or",sys.argv[0]," -f filter_file xFSSH outputfiles "
                exit(1)

        if(sys.argv[1]=="-f"):
                filterfile=sys.argv[2]
                trajfilter=read_filter(filterfile)                
                outputfiles=sys.argv[3:]
        else:
                outputfiles=sys.argv[1:]

        # Read in all trajectories and append them to traj_list
        traj_list=[]
        for outputfile in outputfiles:
                print "#Reading ",outputfile
                #fluence,FWHM=read_parameters(outputfile)        
                #print '#Fluence',fluence
                #print '#FWHM',FWHM

                if(sys.argv[1]=="-f"):
                        for tf in trajfilter:
                                if outputfile.rstrip('.gz').endswith(tf[0]):
                                        break
                        if not outputfile.rstrip('.gz').endswith(tf[0]):
                                print "Warning: filter for file",outputfile,"not found"
                                exit(-1)
                                tf=["?",[]]

                        traj_list_file=read_trajectories(outputfile,tf[1])
                else:
                        traj_list_file=read_trajectories(outputfile)

                print '# ntraj=',len(traj_list_file)

                                

                traj_list.extend(traj_list_file)

# Do some analysis:                
        final_steps=[]
        t120_steps=[]


        print "# Final Mulliken Charges and kinetic energy releases:"
        print "#",
        for i in range(0,len(atoms)):
                print "%6s" % ("q"+atoms[i]),
        print "  ",
        for i in range(0,len(atoms)):
                print "%7s" % ("KER_"+atoms[i]),
        print "  ",
        print "%7s" % ("N_photon"),
        print "%8s %8s %8s %8s" % ("d_"+atoms[0]+atoms[1],"d_"+atoms[1]+atoms[2],"d_"+atoms[1]+atoms[3],"d_"+atoms[1]+atoms[4]),
        print "  "


        for traj in traj_list:    
                # find the indices for the nuclear / electronic steps
                nucstep_index=[i  for i in range(0,len(traj)) if isinstance(traj[i],Nstep)]
                elstep_index=[i  for i in range(0,len(traj)) if isinstance(traj[i],step) or isinstance(traj[i],Sstep)]
                # the indices for the photoionization steps
                photoionization_index=[i for i in elstep_index if traj[i].typ=="P"]
                # append to list of final electronic steps

                #last_step_index=elstep_index[-1]
                last_step_index=len(traj)-1
                final_steps.append(traj[last_step_index])
                within_pulse_index=[i for i in elstep_index if traj[i].time<120.0 and (isinstance(traj[i],step) or isinstance(traj[i],Sstep))]
                t120_steps.append(traj[within_pulse_index[-1]])

                # in case we have more than one nuclear step: derive the kinetic energy per atom and the final distances of the atoms
                if(len(nucstep_index)>=2):
                        #KER=traj[nucstep_index[-1]].KER()
                        KER=traj[nucstep_index[-1]].KER_COM_substracted()
                        # calculate the remaining potential energy and add it up to kinetic energy  (assume two particle Coulomb potential)

                        KER_remain=[0. for i in range(0,len(atoms))]
                        for i in range(0,len(atoms)):
                                for j in range(0,len(atoms)):
                                        if(i==j):
                                                continue
                                        KER_remain[i]=KER_remain[i]+\
                                            mass[j]/(mass[i]+mass[j])*\
                                            traj[last_step_index].q[i] * traj[last_step_index].q[j]/ traj[nucstep_index[-1]].atom_distance(i,j)
                        for i in range(0,len(atoms)):
                                KER[i]=KER[i]+KER_remain[i]

                        distances=[traj[nucstep_index[-1]].atom_distance(0,1),\
                                           traj[nucstep_index[-1]].atom_distance(1,2),\
                                           traj[nucstep_index[-1]].atom_distance(1,3),\
                                           traj[nucstep_index[-1]].atom_distance(1,4)]
                else:
                        KER=[0. for i in range(0,len(atoms))]
                        KER_remain=[0. for i in range(0,len(atoms))]
                        distances=[0.,0.,0.,0.]

                # print for every trajectory
                print " ",
                for i in range(0,len(atoms)):
                        print "%6.3f" % traj[last_step_index].q[i],
                print "  ",
                for i in range(0,len(atoms)):
                        print "%7.2f" % KER[i],
                print " ",
                print  "%8d  " % len(photoionization_index),
                print "%8.2f %8.2f %8.2f %8.2f" % (distances[0],distances[1],distances[2],distances[3]),
                print ""

        if(len(traj_list)==0):
                print "no trajectories"
                sys.exit()
        # calculate average final state values        
        Qtotal= sum([ sum(s.q) for s in final_steps ])/len(traj_list)
        Q=[0. for i in range(0,len(atoms))]
        for i in range(0,len(atoms)):
                Q[i]= sum([ s.q[i] for s in final_steps ])/len(traj_list)
                        
        print ""
        print ""
        print "# Average charges:"
        print "#mean charge q ",
        for a in atoms:
                print "mean q"+a+" ",
        print ""
        print "#total ntraj",len(traj_list)
        print Qtotal,
        for Qx in Q:
                print Qx,
        print ""


        # Qtotal= sum([ sum(s.q) for s in t120_steps ])/len(traj_list)
        # Q=[0. for i in range(0,len(atoms))]
        # for i in range(0,len(atoms)):
        #         Q[i]= sum([ s.q[i] for s in t120_steps ])/len(traj_list)

        # print ""
        # print ""
        # print "# Average charges at 120 fs:"
        # print "#mean charge q ",
        # for a in atoms:
        #         print "mean q"+a+" ",
        # print ""
        # print "#total ntraj",len(traj_list)
        # print Qtotal,
        # for Qx in Q:
        #         print Qx,
        # print ""


