#!/bin/bash
#----------------------------------------------------------------------------
# maketransitions - script to calculate all ionized configurations of a molecule 
#                   at given geometry 
# Needs user and system specifications from file
#      - maketransitions_setup.sh or setup.sh (looks for them in that order
#        in directory of execution. Else uses examples for CO from the script
#        directory)
#
# uses functions from 
#      - maketransitions_functions.sh
#      - string_functions.sh
# 
# sources init.sh to setup aliases and other settings shared by maketransitions.sh
# and the scripts it generates.
#
# $Id$
#----------------------------------------------------------------------------

# record date and time of starting the script for automated backup of the used input files and created output files
# with save_files function 
date_start=$(date +%Y-%m-%d)
time_start=$(date +%k-%m-%S)

script_dir="$(dirname $(readlink -f $0))"; 
echo "script_dir=${script_dir}"

# source setup file
# look for setup file first in wd, which is equivalent to ./, then in maketransitions path
# look first for setup file under the name maketransitions_setup.sh, then setup.sh
if [ "$(pwd)" != "$script_dir" ]; then
    if [ -f "$(pwd)/maketransitions_setup.sh" ]; then
        setup_file="$(pwd)/maketransitions_setup.sh"
        # . ./maketransitions_setup.sh
        echo "setup file is ./maketransitions_setup.sh"
    elif [ -f "$(pwd)/setup.sh" ]; then
        setup_file="$(pwd)/setup.sh"
        # . ./setup.sh
        echo "setup file is ./setup.sh."
    else
        setup_file="${script_dir}/maketransitions_setup.sh"
        # . ${script_dir}/maketransitions_setup.sh
        # for some reason bash adds a second slash in front of filename in this variant
        # echo "setup file is ${script_dir}/maketransitions_setup.sh"
        echo "setup file is example file from script directory."
    fi
else
    if [ -f "./maketransitions_setup.sh" ]; then
        setup_file="${script_dir}/maketransitions_setup.sh"
        # .  ./maketransitions_setup.sh
        echo "setup file is ${script_dir}/maketransition_setup.sh"
    elif [ -f "./setup.sh" ]; then
        setup_file="$(pwd)/setup.sh"
        # . ./setup.sh
        echo "setup file is ./setup.sh."
    else
        >&2 echo "no setup file found -> exit"
        exit 1
    fi
fi
. ${setup_file} || exit 1

# source init.sh
init_file="${script_dir}/init.sh"
. ${init_file} || exit 1

# look for input file in working directory, otherwise use example input from script dir
if ! check_existence_of_input_files ${input_file} &> /dev/null; then
    echo "No input file in working directory. Use example input file from script directory for CO."
    input_file="${script_dir}/${input_file}"
    check_existence_of_input_files ${input_file} || exit 1
else
    echo "input file is ${input_file}"
    input_file="$(readlink -f ${input_file})"
fi

save_files ${setup_file} ${input_file}

# empty files supposed for output
echo '' > notconvergedlist.txt
echo '' > converged_to_wrong_states.txt

echo "Using Rev. $rev"

{
    calculate_basis_and_atomic_transitions || exit 1

    ### calculate neutral CO
    occ=$occ_neutral
    set_atom_conf || exit 1
    link_basis_and_atomic_transitions  || exit 1
    calculate_neutral_conf || exit 1

    # iteration through all ionization states 
    while get_next_occ; do
        if exclude_occ_due_to_sym ${occ}; then continue; fi

        # set array parents
        get_parents || exit 1
        # remove parents excluded due to symmetry
        i=0
        while ! [ -z ${parents[$i]+x} ]; do
            if exclude_occ_due_to_sym ${parents[$i]}; then 
                unset parents[$i]
            fi
            let i+=1
        done
        parents=(${parents[@]})

        set_atom_conf || exit 1
        link_basis_and_atomic_transitions || exit 1

        # try to converge to occ from every parent
        for ((i=0; i<${#parents[@]}; i++)); do          
            parentocc=${parents[$i]}

            calculate_parentocc_occ || exit 1
            # choose best method
            analyze_parentocc_occ || exit 1
        done

        # choose best parent
        analyze_results_all_parents || exit 1

        # Calculate process data with the best parent
        if ! [ -z ${options_transitions+x} ]; then
            calculate_transitions || exit 1
        fi        
    done 
} > maketransitions.out

# Collect process data from all calculated configurations
if ! [ -z ${options_transitions+x} ]; then
    cat [012]*/process.dat > process.dat
fi

# Statistics
get_stats_of_output_files maketransitions.out notconvergedlist.txt converged_to_wrong_states.txt  || exit 1

write_stats_of_output_files | tee statistics.txt || exit 1

# Save output files
save_files notconvergedlist.txt converged_to_wrong_states.txt maketransitions.out statistics.txt || exit 1

