#!/bin/bash
#----------------------------------------------------------------------------
# maketransitions_functions.sh
#
# $Id$
#----------------------------------------------------------------------------
# . ./string_functions.sh

check_GLOBAL_IN() {
    ################################################################################
    # checks all entries of array GLOBAL_IN if they are variables/arrays to
    # which values have been assigned
    # returns: #####################################################################
    # 1 if one or more variables haven't been assigned
    # 2 if caller is not a funciton and no name was passed as argument
    # 0 if succesfull
    ################################################################################
    local var var_info yes_exit calling_function
    if [ -z "$1" ]; then
        calling_function=${FUNCNAME[1]}
        if [ -z "$calling_function" ]; then
            calling_function="unknown caller"
            >&2 echo "Caller is not a function. Pass the caller name as argument."
            return 2
        fi
    else
        calling_function="$1"
    fi
    yes_exit="no"
    for var in ${GLOBAL_IN[*]}; do
        var_info=$(declare -p $var 2> /dev/null)
        if [ -z "$var_info" ]; then
            >&2 echo "$calling_function requires \$${var} as input by global variable"
            yes_exit="yes"
        fi
        # would only recognize arrays if element 0 set
        # if [ -z "${!var+x}" ]; then
        #     2>1 echo "\$${var} is required as input by setting global variable"
        #     yes_exit="yes"
        # fi
    done
    if [ "$yes_exit" == "yes" ]; then
        >&2 echo "Not all global variables used as input by $calling_function were defined -> exit"
        return 1
    fi
    return 0
}

check_occupation_variables() {
    ################################################################################
    # check if $parentocc and $occ are set and have same length
    ################################################################################
    local GLOBAL_IN=(parentocc occ)
    check_GLOBAL_IN || return 1 
    
    if [ $(echo -n "${parentocc})" | wc -c) -ne $(echo -n "${parentocc})" | wc -c) ]
    then
        >&2 echo "Var. occ and parentocc do not have same length."
        return 1
    fi
    echo "Occupation variables have same length."
    return 0
}

# set_occ_arrays() {
#     ### OUTPUTs global variables: ##################################################
#     # p_occ_array - array of $parentocc entries as floats
#     # occ_array   - array of $occ entries as floats
#     ################################################################################
#     local GLOBAL_IN=(occ parentocc)
#     check_GLOBAL_IN || return 1

#     local j
#     j=0; 
#     while [ $j -lt ${#occ} ]
#     do
#         p_occ_array[$j]=$(printf "%1.3f" ${parentocc:${j}:1})
#         occ_array[$j]=$(printf "%1.3f" ${occ:${j}:1})
#         let j=$j+1
#     done
# }
set_occ_array() {
    ### OUTPUTs global variables: ##################################################
    # occ_array   - array of $occ entries
    ################################################################################
    local GLOBAL_IN=(occ)
    check_GLOBAL_IN || return 1

    local j
    j=0; 
    while [ $j -lt ${#occ} ]
    do
        occ_array[$j]=${occ:${j}:1}
        # occ_array[$j]=$(printf "%1.3f" ${occ:${j}:1})
        let j=$j+1
    done
}
set_p_occ_array() {
    ### OUTPUTs global variables: ##################################################
    # p_occ_array - array of $parentocc entries
    ################################################################################
    local GLOBAL_IN=(parentocc)
    check_GLOBAL_IN || return 1

    local j
    j=0; 
    while [ $j -lt ${#parentocc} ]
    do
        p_occ_array[$j]=${parentocc:${j}:1}
        # p_occ_array[$j]=$(printf "%1.3f" ${parentocc:${j}:1})
        let j=$j+1
    done
}
set_occ_from_array() {
    ### OUTPUTs global variables: ##################################################
    # occ - occupation string
    ################################################################################
    local GLOBAL_IN=(occ_array)
    check_GLOBAL_IN || return 1
    
    occ=$(echo -n "${occ_array[*]}" | tr -d " ")
}
set_parentocc_from_array() {
    ### OUTPUTs global variables: ##################################################
    # parentocc - parent occupation string
    ################################################################################
    local GLOBAL_IN=(p_occ_array)
    check_GLOBAL_IN || return 1
    
    parentocc=$(echo -n "${p_occ_array[*]}" | tr -d " ")
}

set_frac_occ() {
    ################################################################################
    # interpolates between parentocc and occ. 
    # fraction=1 -> frac_occ=p_occ_array, fraction=0 -> frac_occ=occ_array
    ### OUTPUTs global variables: ##################################################
    # frac_occ - array of fractional occupation constructed from occ_array, p_occ_array
    #            and the fraction provided as $1
    ################################################################################
    local GLOBAL_IN=(occ_array p_occ_array)
    check_GLOBAL_IN || return 1 
    local j  tmp fraction
    if ! [ -z "${1+x}" ]
    then
        fraction="${1}"
    else
        echo "Pass ${FUNCNAME[0]} the fraction between occ and parentocc as argument."
        return 1
    fi

    frac_occ=""
    j=0
    tmp=$(printf "%1.3f" $(echo "${occ_array[$j]} + ${fraction} * (${p_occ_array[$j]} - ${occ_array[$j]})" | bc))
    frac_occ="${tmp}"
    j=1
    while [ $j -lt ${#occ} ]
    do
        tmp=$(printf "%1.3f" $(echo "${occ_array[$j]} + ${fraction} * (${p_occ_array[$j]} - ${occ_array[$j]})" | bc))
        frac_occ="${frac_occ},${tmp}"
        let j=$j+1
    done
    return 0

    # frac_occ=`echo "${p_occ_array[0]} + $fraction * (${occ_array[0]} - ${p_occ_array[0]})" | bc `
    # for j in 1 2 3 4 5 6; do
    #     tmp=`echo "${p_occ_array[$j]} + $fraction * (${occ_array[$j]} - ${p_occ_array[$j]})" | bc `
    #     frac_occ="$frac_occ,$tmp"
    # done
}

set_atom_conf() {
    ### OUTPUTs global variables: ##################################################
    # atom_conf - array of xatom conf strings
    ################################################################################
    local GLOBAL_IN=(atoms occ atoms_Kshell_MO_index atoms_conf_neutral)
    check_GLOBAL_IN || return 1

    local i
    for ((i=0;i<${#atoms[*]};i++)); do
        if [ ${atoms_Kshell_MO_index[$i]} -ge 0 ]; then 
            atom_conf[$i]=$(echo ${atoms_conf_neutral[$i]} | sed -r "s/1s./1s${occ:${atoms_Kshell_MO_index[$i]}:1}/")
        else # keep neutral conf
            atom_conf[$i]=${atoms_conf_neutral[$i]}
        fi
    done
    # atomconfO="1s${occ:0:1}_2s2_2p4"
    # atomconfC="1s${occ:1:1}_2s2_2p2"
}

get_nel() { 
    ### returns: ###################################################################
    # number of electrons in occupation=$1
    ### examples of use: ###########################################################
    # occ_array=(2 2)
    # get_nel ${occ_array[@]} # 4
    # occupation=22
    # get_nel $occupation # 4
    ################################################################################
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1

    local occupation count
    if [ -z ${1+x} ]
    then
        >&2 echo "Pass ${FUNCNAME[0]} occupation string as argumentor multiple arguments to be summed -> exit"
        return 1
    fi
    if [ ${#@} -gt 1 ]; then
        count=$1
        while shift; do
            if [ -z ${1+x} ]; then break; fi
            let count+=$1
        done
    else # -eq 1
        occupation=$1
        count=0
        for (( i=0; i<${#occupation}; i++ )); do
            let count+=${occupation:$i:1} 
        done
    fi
    echo -n $count
    return 0    
}

get_parents() {
    ################################################################################
    # take occ=$1 and construct all parents that are not excluded by symmetry
    # and have non negative occupation of each MO
    ### OUTPUTs global variables: ##################################################
    # parents - array of parent occupation strings
    ################################################################################
    GLOBAL_IN=(occ)
    check_GLOBAL_IN || return 1
    local i pos new_occ parentocc

    set_occ_array
    parents=()
    for (( i = 0; i < ${#occ}; i++)); do
        # parentocc=$occ
        p_occ_array=(${occ_array[*]})
        # is there an occupation pattern from which we can ionize ?
        # let new_occ=${parentocc:${i}:1}+1
        let new_occ=${p_occ_array[$i]}+1
        if [ $new_occ -le 2  ]; then
            # let pos=${i}+1
            # parentocc=$(echo ${parentocc} | sed "s/./${new_occ}/${pos}")
            p_occ_array[$i]=$new_occ
            set_parentocc_from_array
            parents[${#parents[@]}]=${parentocc}
        fi  
    done
}

get_next_occ() {
################################################################################
# method to get next $occ as with multiple for loops:
#
# from $occ generate next occupation in the following fashion
# choose the last MO that has occ. greater 0, take out one el. while filling
# all higher MOs in with 2 electrons
# this always ensures that parents of the next occ. have already been calculated
# but doesn't iterate through occupations with same number of electrons
# to get next occupation with same number of electrons, use get_next_occ_nel $nel
### returns: ###################################################################
# 0 as long as next occ contains electrons
# 1 if unsuccessful
# 2 if next occ contains no electrons
################################################################################
    GLOBAL_IN=(occ)
    check_GLOBAL_IN || return 1
    local i pos new_occ return_value
    
    set_occ_array

    return_value=2
    let i=${#occ}-1
    while [ $i -ge 0 ]; do
        # let new_occ=${occ:${i}:1}-1
        let new_occ=${occ_array[$i]}-1
        if [ $new_occ -ge 0 ]; then
            # let pos=${i}+1
            # occ=$(echo ${occ} | sed "s/./${new_occ}/${pos}")
            occ_array[$i]=$new_occ
            for ((pos=$i+1; pos<${#occ}; pos++)); do
                occ_array[$pos]=2
            done
            # while [ $pos -lt ${#occ} ]; do
            #     let pos++
            #     occ=$(echo ${occ} | sed "s/./2/${pos}")
            # done

            # exclude the occupation without any electrons
            if [ $(get_nel ${occ_array[@]}) -eq 0 ]; then
                break
            else
                return_value=0
                break
            fi
        else
            let i--
        fi
    done

    set_occ_from_array
    return $return_value
}

get_next_occ_nel() {
################################################################################
# method to get next $occ as with multiple for loops and choose only those 
# with same number of electrons
#
# tends to be much slower then get_next_occ
# (this might be due to the fact that get_next_occ which is called many times uses string operations;
#  changing this to working with integer arrays $occ_array and $p_occ_array should improve speed!)
#
### returns: ###################################################################
# 2 if new occ = old occ
# 0 else
################################################################################
    GLOBAL_IN=(occ)
    check_GLOBAL_IN || return 1

    local number_of_electrons
    if [ -z "${1+x}" ]
    then
        echo "Pass ${FUNCNAME[0]} number of electrons as argument -> exit"
        return 1
    fi    
    number_of_electrons=$1

    while get_next_occ; do
        if [ $(get_nel $occ) -ne ${number_of_electrons} ] ; then
            continue
        else
            return 0
        fi
    done
    return 2
}

write_all_occs_with_parents() {
################################################################################
# test the method get_next_occ
# writes: ######################################################################
# all occupation numbers and their parents in file $occ_file
################################################################################
    # GLOBAL_IN=(occ)
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1

    local occ=2222
    local occ_file
    occ_file="all_occs_with_parents"

    { 
        echo "# all occupation numbers followed by their parents" 
        # get_parents # symmetry on
        get_parents 1 # symmetry off
        echo "${occ} ${parents[*]}"
        while get_next_occ ; do
            # if exclude_occ_due_to_sym ${occ}; then continue; fi

            # get_parents # symmetry on
            get_parents 1 # symmetry off
            echo "${occ} ${parents[*]}" 
        done
    } > $occ_file
}

write_all_occs_sorted_with_parents() {
################################################################################
# test the method get_next_occ_nel
# writes: ######################################################################
# all occupation numbers and their parents in file $occ_file 
# sorted by numbers of electrons
################################################################################
    # GLOBAL_IN=(occ)
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1
    
    local occ=2222
    local occ_file start_occ
    occ_file="all_occs_sorted_with_parents"
    
    {
        echo "# all occupation numbers followed by their parents"
        # get_parents # symmetry on
        get_parents 1 # symmetry off
        echo "${occ} ${parents[*]}"

        start_occ=$occ
        let nel=$(get_nel $occ)-1
        while [ $nel -gt 0 ]; do
            occ=$start_occ
            while get_next_occ_nel $nel; do
                # get_parents # symmetry on
                get_parents 1 # symmetry off
                echo "${occ} ${parents[*]}"
            done
            let nel--
        done
    } > $occ_file            
}

write_input_file() {
################################################################################
# writes input file to current directory
################################################################################
    GLOBAL_IN=(input_file)
    check_GLOBAL_IN || return 1

    if ! [ -f "$(basename $input_file)" ]; then
        cp "${input_file}" "$(basename $input_file)"
# echo 'system=CO
# N_nuc= 2
# C      0.0                 0.0                 0.0     2s1p C.nao
# O      0.0                 0.0                 2.132   2s1p O.nao

# average_occ_MO=5,6

# EPS1=0.0

# SCF_max_iter= 100
# SCF_energy_conv= 1E-7
# SCF_density_conv= 1E-7
# verbose=yes
# ' > input.in
    fi

}

calculate_basis_and_atomic_transitions(){
################################################################################
# function to create nao and auger and tdipole files for all K shell occupations of C and O
### writes: ####################################################################
# files with commands to generate basis
################################################################################
    GLOBAL_IN=(atoms atoms_conf_neutral atoms_options atoms_file_prefix_format setup_file init_file)
    check_GLOBAL_IN || return 1

    local i j conf file_prefix filename atoms_uniq
    
    atoms_uniq=()
    for ((i=0; i<${#atoms[*]}; i++)); do
        # skip double entries
        if [ "${atoms_uniq[*]#${atoms[$i]}}" == "${atoms_uniq[*]}"  ]; then
            atoms_uniq+=(${atoms[$i]})
        else
            continue
        fi

        filename=$(printf "xatom%s.cmd" "${atoms[$i]}")
        {     echo  "#!/bin/bash 
rev=$rev
setup_file=\"${setup_file}\"
 . \${setup_file} 
init_file=\"${init_file}\"
. \${init_file}"
            for ((j=2; j>=0; j--)); do
                conf=$(echo ${atoms_conf_neutral[$i]} | sed "s/1s2/1s${j}/")
                file_prefix=$(printf "${atoms_file_prefix_format}" "${atoms[$i]}" "$conf")
                echo -n "\${xatom_path} -s ${atoms[$i]} -conf ${conf} -file_nao ${file_prefix}.nao"
                echo -n " -file_tdipole ${file_prefix}.tdipole -file_auger ${file_prefix}.auger"
                echo    " ${atoms_options[$i]}"
            done 
        } > $filename
        bash $filename
    done
} # end calculate_basis_and_atomic_transitions


link_basis_and_atomic_transitions(){
################################################################################
# create directory for given occ and link .nao, .tdipole and .auger files for given atom_conf
################################################################################
    GLOBAL_IN=(occ atoms atom_conf atoms_file_prefix_format)
    check_GLOBAL_IN || return 1

    # occ=$1
    # atomconfC=$2
    # atomconfO=$3
    local i suffix file_prefix atoms_uniq

    if ! [ -d "$occ" ]; then
        mkdir $occ    
    fi
    cd $occ

    atoms_uniq=()
    for ((i=0; i<${#atoms[*]}; i++)); do
        # skip double entries
        if [ "${atoms_uniq[*]#${atoms[$i]}}" == "${atoms_uniq[*]}"  ]; then
            atoms_uniq+=(${atoms[$i]})
        else
            continue
        fi

        file_prefix=$(printf "${atoms_file_prefix_format}" "${atoms[i]}" "${atom_conf[$i]}")
        for suffix in .nao .auger .tdipole; do
            if [ -e "../${file_prefix}${suffix}" ] && ! [ -e "${atoms[i]}${suffix}" ]; then
                ln -s "../${file_prefix}${suffix}" "${atoms[i]}${suffix}"
            fi
        done
    done

    # if ! [ -e "C.nao" ]; then
    #     ln -s ../C.${atomconfC}.nao C.nao
    # fi
    # if ! [ -e "O.nao" ]; then
    #     ln -s ../O.${atomconfO}.nao O.nao
    # fi
    # ln -s ../C.${atomconfC}.auger C.auger
    # ln -s ../O.${atomconfO}.auger O.auger
    # ln -s ../C.${atomconfC}.tdipole C.tdipole
    # ln -s ../O.${atomconfO}.tdipole O.tdipole
    cd ..

} # end link_basis_and_atomic_transitions

cp_parent_data() {
################################################################################
# cp basis and wf of parent occupation to current directory
################################################################################
    GLOBAL_IN=(parentocc)
    check_GLOBAL_IN || return 1

    cp ../${parentocc}/basis.dat basis.dat 
    cp ../${parentocc}/wf.dat wf.dat
}


save_basis_wf() {
################################################################################
# save basis.dat and wf.dat specific to the command $c and the parent occupation $parentocc
################################################################################
    GLOBAL_IN=(c parentocc)
    check_GLOBAL_IN || return 1

    cp basis.dat basis.${c}.${parentocc}.dat
    cp wf.dat wf.${c}.${parentocc}.dat
}

calculate_neutral_conf(){
################################################################################
# calculate neutral molecule
# writes: ######################################################################
# xmolecule.cmd
################################################################################
    GLOBAL_IN=(occ optionsneutral rev setup_file init_file input_file xmolecule_path)
    check_GLOBAL_IN || return 1

    cd $occ
    
    write_input_file
    echo  "#!/bin/bash 
rev=$rev
setup_file=\"${setup_file}\"
 . \${setup_file} 
init_file=\"${init_file}\"
. \${init_file}
\${xmolecule_path} input.in -occ ${occ}  ${optionsneutral} -xatom_path=\${xatom_path}
    " > xmolecule.cmd
    sh ./xmolecule.cmd  > output.${rev}

    if [ -z "$(grep "SCF converged" output.${rev})" ]; then 
	>&2 echo "${FUNCNAME[0]} not converged -> exit"
        cd ..
        return 1
    fi
        
    cd ..    
    return 0
} # end calculate_neutral_conf

calculate_ionized_conf_DIIS(){
    ################################################################################
    # Calculate the electronic structure of occ starting with guess
    # constructed from parentocc using $optionsionized
    # appends to file: #############################################################
    # notconvergedlist.txt
    # writes: ######################################################################
    # xmolecule.cmd.${parentocc}
    # wf.cmd.${parentocc}.dat
    # basis.cmd.${parentocc}.dat
    # olap.cmd.${parentocc}.${occ}
    # output.${rev}.cmd.${parentocc}
    ################################################################################
    GLOBAL_IN=(occ rev parentocc setup_file init_file input_file xmolecule_path)
    check_GLOBAL_IN || return 1

    local c
    c="cmd"

    cd $occ

    # if results for parentocc are not found, write note into notconvergedlist.txt
    if ! [ -e "../${parentocc}/wf.dat" ]; then 
	echo "guess orbitals not present"
        echo "${parentocc} -> ${occ} (guess)" >> ../notconvergedlist.txt
        cd ..
        return 1
    fi
    write_input_file
    # write script that will call xmolecule
    echo  "#!/bin/bash 
rev=$rev
setup_file=\"${setup_file}\"
 . \${setup_file} 
init_file=\"${init_file}\"
. \${init_file}

parentocc=\"${parentocc}\"
occ=\"${occ}\"

cp_parent_data

\${xmolecule_path} input.in -occ \${occ}  \${optionsionized} -xatom_path=\${xatom_path}" > xmolecule.${c}.${parentocc} 

    bash ./xmolecule.${c}.${parentocc}  > output.${rev}.${c}.${parentocc}

    # if xmol didn't converge note in notconvergedlist.txt, else write olap to file and
    # save wf.dat and basis.dat
    if [ -z "$(grep 'SCF converged' output.${rev}.${c}.${parentocc})" ]; then
        echo "${parentocc} -> ${occ} (DIIS)" >> ../notconvergedlist.txt
    else
    #olap=`grep 'GUESSINFO: Overlap with initial guess (end):' output.${rev}.${c}.${parentocc} | cut -f 3 -d:`    
       olap=$(get_last_num "$(grep 'GUESSINFO: Overlap with initial guess (end):' output.${rev}.${c}.${parentocc})")
        echo $olap > olap.${c}.${parentocc}.${occ}
        save_basis_wf
    fi
    cd ..
} # end calculate_ionized_conf_DIIS

calculate_ionized_conf_fractions(){
    ################################################################################
    # fall back option: description see function def. of calculate_ionized_conf_smooth
    # 
    # this function adds the possiblity to calculate a specific subset of fractional 
    # occupations as specified by the variable fractions
    # also the final state given by variable occ will only be calculated if yes_final -eq 0 
    # assumes: #####################################################################
    # wf.dat and basis.dat contain a starting guess for xmolecule
    # appends to file: #############################################################
    # notconvergedlist.txt
    # writes: ######################################################################
    # xmolecule.cmd5.${parentocc}
    # wf.cmd5.${parentocc}.dat
    # basis.cmd5.${parentocc}.dat
    # olap.cmd5.${parentocc}.${occ}
    # output.${rev}.cmd5.${parentocc}
    ################################################################################
    GLOBAL_IN=(occ rev parentocc optionsionized0 input_file fractions yes_final xmolecule_path overlap_path)
    check_GLOBAL_IN || return 1

    local i 

    write_input_file

    set_occ_array
    set_p_occ_array
    
    for i in ${fractions[*]}; do
        echo "step f=$i"
        set_frac_occ $i

        ${xmolecule_path} input.in -occ ${occ} -f_occ ${frac_occ} ${optionsionized0} -xatom_path=${xatom_path}
        echo -n 'GUESSINFO: Overlap with initial guess:'
        ${overlap_path} input.in -occ ${occ} -RIni -ref ../${parentocc}/wf.dat -ref_basis ../${parentocc}/basis.dat | sed s/WF_overlap//
        echo "finish f=$i"
    done

    # if ! [ -z ${yes_final+x} ] && [ ${yes_final} -eq 0 ] ; then
    #     echo "step final"
    #     ${xmolecule_path} input.in -occ ${occ}  ${optionsionized0} -xatom_path=${xatom_path} | grep -v GUESSINFO
    #     echo -n 'GUESSINFO: Overlap with initial guess (end):'
    #     ${overlap_path} input.in -occ ${occ} -RIni -ref ../${parentocc}/wf.dat -ref_basis ../${parentocc}/basis.dat | sed s/WF_overlap//
    #     echo "finish final"

    #     # if xmol didn't converge note in notconvergedlist.txt, else write olap to file and
    #     # save wf.dat and basis.dat
    #     if [ -z "$(cat output.${rev}.${c}.${parentocc} | sed -n "/step final/,/finish final/p" | grep "SCF converged")" ]; then
    #         echo "${parentocc} -> ${occ} (smooth)" >> ../notconvergedlist.txt
    #     else
    #         olap=$(get_last_num "$(grep 'GUESSINFO: Overlap with initial guess (end):' output.${rev}.${c}.${parentocc})")
    #         echo $olap > olap.${c}.${parentocc}.${occ}
    #         save_basis_wf
    #     fi
    # fi

} # end calculate_ionized_conf_fractions

calculate_ionized_conf_smooth(){
    ################################################################################
    # fall back option: try to vary the occupation between occ and parentocc
    # in ten steps as fractional occupations and converge SCF everytime
    # use $optionsionized0
    # appends to file: #############################################################
    # notconvergedlist.txt
    # writes: ######################################################################
    # xmolecule.cmd5.${parentocc}
    # wf.cmd5.${parentocc}.dat
    # basis.cmd5.${parentocc}.dat
    # olap.cmd5.${parentocc}.${occ}
    # output.${rev}.cmd5.${parentocc}
    ################################################################################
    GLOBAL_IN=(occ rev parentocc setup_file init_file input_file  xmolecule_path overlap_path)
    check_GLOBAL_IN || return 1

    local c i 
    c="cmd5"

    cd $occ

    if ! [ -f "../${parentocc}/wf.dat" ]; then 
	echo "guess orbitals not present"
        echo "${parentocc} -> ${occ} (guess)" >> ../notconvergedlist.txt
        cd ..
        return 1
    fi
    write_input_file
            
    # write script that will call xmolecule
    echo  "#!/bin/bash 
rev=$rev
setup_file=\"${setup_file}\"
 . \${setup_file} 
init_file=\"${init_file}\"
. \${init_file}

parentocc=\"$parentocc\"
occ=\"$occ\"

cp_parent_data
set_occ_array
set_p_occ_array

for i in 1.0 0.9 0.8 0.7 0.6 0.5 0.4 0.3 0.2 0.1 0.0; do
  echo \"step f=\$i\"
  set_frac_occ \$i

  \${xmolecule_path} input.in -occ \${occ} -f_occ \${frac_occ} \${optionsionized0} -xatom_path=\${xatom_path}
  echo -n 'GUESSINFO: Overlap with initial guess:'
  \${overlap_path} input.in -occ \${occ} -RIni -ref ../\${parentocc}/wf.dat -ref_basis ../\${parentocc}/basis.dat | sed s/WF_overlap//
  echo \"finish f=\$i\"
done
echo \"step final\"
\${xmolecule_path} input.in -occ \${occ}  \${optionsionized0} -xatom_path=\${xatom_path} | grep -v GUESSINFO

echo -n 'GUESSINFO: Overlap with initial guess (end):'
\${overlap_path} input.in -occ \${occ} -RIni -ref ../\${parentocc}/wf.dat -ref_basis ../\${parentocc}/basis.dat | sed s/WF_overlap//
echo \"finish final\"
" > xmolecule.${c}.${parentocc}

    bash ./xmolecule.${c}.${parentocc}  > output.${rev}.${c}.${parentocc}

    # if xmol didn't converge note in notconvergedlist.txt, else write olap to file and
    # save wf.dat and basis.dat
    if [ -z "$(cat output.${rev}.${c}.${parentocc} | sed -n "/step final/,/finish final/p" | grep "SCF converged")" ]; then
        echo "${parentocc} -> ${occ} (smooth)" >> ../notconvergedlist.txt
    else
       olap=$(get_last_num "$(grep 'GUESSINFO: Overlap with initial guess (end):' output.${rev}.${c}.${parentocc})")
        echo $olap > olap.${c}.${parentocc}.${occ}
        save_basis_wf
    fi
    cd ..
} # end calculate_ionized_conf_smooth

calculate_parentocc_occ() {
    ################################################################################
    # Solve for electronic structure of occ starting with wavefunction
    # constructed from parentocc. Check the overlap and in case try
    # to converge to better result with fallback option
    #
    # First try calculate_ionized_conf_DIIS (c=cmd)
    # if (overlap with initial guess)^2 < 0.5 (or not converged)
    # try calculate_ionized_conf_smooth (c=cmd5)
    # compare which method gave the best overlap with initial guess
    # save the wf to  ${occ}/wf.${parentocc}.dat
    # and basis to    ${occ}/basis.${parentocc}.dat
    # and output to   ${occ}/output.${rev}.${parentocc}
    # olap of individual method is saved in files ${occ}/olap.${c}.${parentocc}.${occ}
    ################################################################################
    GLOBAL_IN=(parentocc occ rev optionsionized optionsionized0 xmolecule_path overlap_path)
    check_GLOBAL_IN || return 1

    local olap

    echo "Calculating $occ with parent occ $parentocc"

    calculate_ionized_conf_DIIS
    if [ -f $occ/olap.cmd.${parentocc}.${occ} ]; then
        olap=`cat $occ/olap.cmd.${parentocc}.${occ} | tr -d " "`
        if [ -z "$olap" ]; then olap=0; fi
    else
        olap=0
    fi
    
    # if the olap^2 is smaller than 0.5, 
    # try to converge also by changing the occ by fractions
    if (( `echo "${olap}*${olap} < 0.5" | bc` == 1 ));then
        calculate_ionized_conf_smooth
        if [ -f $occ/olap.cmd5.${parentocc}.${occ} ];then
            olap=`cat $occ/olap.cmd5.${parentocc}.${occ} | tr -d " "`
            if [ -z "$olap" ]; then olap=0; fi
        else
            olap=0
        fi
    fi

} # end calculate_parentocc_occ

analyze_parentocc_occ() {
    ################################################################################
    # choose which of two methods gave best overlap after calling
    # calculate_parentocc_occ and make note in notconvergedlist.txt if one method 
    # didn't converge
    # writes: ######################################################################
    # ${occ}/output.${rev}.${parentocc}
    # appends to: ##################################################################
    # notconvergedlist.txt
    ################################################################################
    GLOBAL_IN=(occ parentocc rev)
    check_GLOBAL_IN || return 1

    local command_array olap maxolap c cmax converged
    command_array=("cmd" "cmd5")
    maxolap=0
    cmax="cmd"
    converged=0
    for c in ${command_array[*]}; do 
        if [ -f $occ/olap.${c}.${parentocc}.${occ} ];then
            olap=`cat $occ/olap.${c}.${parentocc}.${occ} | tr -d " "`
            if [ -z "$olap" ]; then
                olap=0
            else
                converged=1                
                echo "   $c $olap"
            fi
        else
            olap=0
        fi
        if (( `echo "${olap}*${olap} > ${maxolap}*${maxolap}" | bc` == 1 ));then
            cmax=$c
            maxolap=$olap
        fi
    done
    # c=$cmax

    # if no method converged, write to stdout
    if [ "$converged" == "0"  ]; then
        echo "${parentocc} -> ${occ} not converged with any method" 
    fi
    # if some method converged, copy files of best method to the 
    # standard filenames without reference to the command used
    if [ $converged -eq 1 ]; then
        if [ "$cmax" == "cmd" ]; then
            cp ${occ}/output.${rev}.${cmax}.${parentocc} ${occ}/output.${rev}.${parentocc}
        else # "$cmax" == "cmd5"
            cat ${occ}/output.${rev}.${cmax}.${parentocc} | \
                sed -n "/step final/,/finish final/p" > ${occ}/output.${rev}.${parentocc}
        fi
        cp ${occ}/basis.${cmax}.${parentocc}.dat ${occ}/basis.${parentocc}.dat
        cp ${occ}/wf.${cmax}.${parentocc}.dat ${occ}/wf.${parentocc}.dat
    fi
    # # delete wf and basis files with reference to cmds
    # # (for loop and if clause make sure only existing files will be tried to be removed)
    # for file in ${occ}/wf.cmd*.dat ${occ}/basis.cmd*.dat; do
    #     if [ -f "$file" ]; then
    #         rm $file
    #     fi
    # done
} # end analyze_parentocc_occ

analyze_results_all_parents() {
    ################################################################################
    # analyze results
    #
    # If there are different results, choose best result and copy its data to ${occ}/wf.dat and ${occ}/basis.dat.
    # Note the cases that differ from best result in energy in converged_to_wrong_states.txt.
    # Generate ${occ}/results.txt, which lists energy and olap to initial guess
    # for each parent.
    # Writes: ######################################################################
    # ${occ}/wf.dat
    # ${occ}/basis.dat
    # ${occ}/results.txt
    # Appends to: ##################################################################
    # converged_to_worong_states.txt
    ################################################################################
    GLOBAL_IN=(parents occ rev)
    check_GLOBAL_IN || return 1

    local olap E f j yes_diff j_maxolap diff
    olap=()
    E=()
    j=0
    # collect overlap and energy for all parents in arrays olap and E, if not converged set 
    # the parents entry to 0.0
    for ((j=0; j<${#parents[@]}; j++)); do
	f=${occ}/output.${rev}.${parents[$j]}
        
        if ! [ -f "$f" ] || [[ `grep -i 'NOT converged' $f ` ]]; then
            echo "${parents[$j]} -> $occ not converged"
            E[$j]=0.00
            olap[$j]=0.0
        else
            # E[$j]=`grep -i 'Total energy' $f|awk '{print $5}'`
            E[$j]=$(get_nth_num "$(grep -i 'Total energy' $f)" 1)
            if [ -z "${E[$j]}" ]; then E[$j]=0; fi
        # olap[$j]=`grep 'GUESSINFO: Overlap with initial guess (end):' $f | cut -f 3 -d: | awk '{printf "%f",$1}'` 
            olap[$j]=$(get_last_num "$(grep 'GUESSINFO: Overlap with initial guess (end):' $f)")           
            if [ -z "${olap[$j]}" ]; then olap[$j]=0; fi
        fi
    done
    # are all energies the same? if not set yes_diff=1
    yes_diff=0
    for (( j=0; j<${#parents[@]}; j++ )); do
        # replace this test of absolute difference by test for difference relative to max of both energies?
        diff=0
        diff=`echo "(${E[0]#-} - ${E[${j}]#-})^2 > 0.0001^2" | bc`
        if (( $diff  == 1 )); then
            yes_diff=1
            break
        fi
    done
    
    # if not all parents converged to a solution with the same energy, 
    # find the parentocc which gave initial guess with highest overlap to final solution
    # and output to ${occ}/results.txt and converged_to_wrong_states.txt
    # else output to ${occ}/results.txt
    if (( ${yes_diff} == 1 )); then
        echo "Different energies for different parentoccupations:"                                    
        j_maxolap=0
        for (( j=0; j<${#parents[@]}; j++ )); do
            if (( `echo "${olap[$j]}*${olap[$j]} > ${olap[$j_maxolap]}*${olap[$j_maxolap]}" | bc` == 1 ));then
                j_maxolap=$j
            fi
        done

        # summarize calculations from each parent in results.txt
        for (( j=0; j<${#parents[@]}; j++ )); do
            if (( $j==$j_maxolap )); then
                echo "* ${parents[$j]} -> $occ ${E[$j]} ${olap[$j]}"
            else
                echo "  ${parents[$j]} -> $occ ${E[$j]} ${olap[$j]}"
            fi
        done | tee ${occ}/results.txt
        
        # for each parentocc that converged to a different energy than the parent with maximum overlap
        # to the final solution, make an entry in converged_to_wrong_states.txt
        for (( j=0; j<${#parents[@]}; j++ )); do
            diff=0
            diff=`echo "(${E[$j_maxolap]#-} - ${E[$j]#-})^2 > 0.0001^2" | bc`
            if (( $diff  == 1 )); then
                echo "  ${parents[$j]} -> $occ ${E[$j]} ${olap[$j]}"
            fi
        done >> converged_to_wrong_states.txt


        if [ -e "$occ/wf.${parents[$j_maxolap]}.dat" ]; then
            cp $occ/wf.${parents[$j_maxolap]}.dat $occ/wf.dat
        fi
        if [ -e "$occ/basis.${parents[$j_maxolap]}.dat" ]; then
            cp $occ/basis.${parents[$j_maxolap]}.dat $occ/basis.dat
        fi
        cp $occ/output.${rev}.${parents[$j_maxolap]} $occ/output.${rev}                                    
    else 
        echo "Same energies for different parentoccupations:"
        for (( j=0; j<${#parents[@]}; j++ )); do
            if (( $j==0 )); then
                echo "* ${parents[$j]} -> $occ ${E[$j]} ${olap[$j]}"
            else
                echo "  ${parents[$j]} -> $occ ${E[$j]} ${olap[$j]}"
            fi
        done | tee ${occ}/results.txt
        
        if [ -e "$occ/wf.${parents[0]}.dat" ]; then
            cp $occ/wf.${parents[0]}.dat $occ/wf.dat
        fi
        if [ -e "$occ/basis.${parents[0]}.dat" ]; then
            cp $occ/basis.${parents[0]}.dat $occ/basis.dat
        fi
        cp $occ/output.${rev}.${parents[0]} $occ/output.${rev}
    fi
    return 0
} # end analyze_results_all_parents

calculate_transitions() {
################################################################################
# Calculate transition rates 
# Writes: ######################################################################
# $occ/process.dat
# $occ/xmolecule.transitions
# $occ/output.${rev}.transitions
################################################################################
    GLOBAL_IN=(occ rev)
    check_GLOBAL_IN || return 1

    cd $occ

    # check if any parent converged
    if ! [ -e "wf.dat" ]; then 
        >&2 echo "calculate_transitions: No converged solution found for occ=$occ."        
        cd ..
        return 1
    fi

    write_input_file
    # write script that will call xmolecule
    echo  "#!/bin/bash 
rev=$rev
setup_file=\"${setup_file}\"
 . \${setup_file} 
init_file=\"${init_file}\"
. \${init_file}

occ=\"${occ}\"

\${xmolecule_path} input.in -occ \${occ} \${options_transitions} -xatom_path=\${xatom_path}" > xmolecule.transitions

    # call xmolecule
    bash ./xmolecule.transitions  > output.${rev}.transitions

    # check convergence
    if [ -z "$(grep 'SCF converged' output.${rev}.transitions)" ]; then
        echo "${parentocc} -> ${occ} (transitions)" >> ../notconvergedlist.txt
    fi


    cd ..
}

check_existence_of_input_files() {
    ################################################################################
    # Checks all passed strings if they are a file
    ################################################################################
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1

    local yes_exit file
    yes_exit="no" # unix false
    for file in "$@"; do
        if ! [ -f "${file}" ]; then
            echo "File ${file} doesn't exist, but is necessary for execution."
            yes_exit="yes"
        fi
    done
    if [ "$yes_exit" == "yes" ]; then
        echo "Not all input files present -> exit"
        return 1
    fi
    return 0
}

save_files() {
    ################################################################################
    # Cps all passed files to current directory using the filename 
    # suffixed with passed date and time
    ################################################################################
    GLOBAL_IN=(date_start time_start)
    check_GLOBAL_IN || return 1
    local file

    for file in "$@"; do
        if [ -f "${file}" ]; then
            cp $file $(basename ${file})_${date_start}_${time_start}
        fi
    done
    return 0
}

clean_files_and_directories() {
    ################################################################################
    # REMOVE all files and directories created by maketransitions.sh
    # to be executed in directory of setup.sh
    ################################################################################
    local dir file

    # remove directories for all occupation numbers
    for dir in *; do 
        if [ -d "${dir}" ] && [[ "${dir}" =~ [0-2]{${#occ_neutral}} ]]; then 
            rm -r "$dir" 
        fi
    done
    # empty directories for all occupation numbers
    # for file in 0*/*; do
    #     if [ -f "${file}" ]; then rm "$file"; fi
    # done
    # for file in 1*/*; do
    #     if [ -f "${file}" ]; then rm "$file"; fi
    # done
    # for file in 2*/*; do
    #     if [ -f "${file}" ]; then rm "$file"; fi
    # done
    # # remove directories
    # for dir in 0*; do
    #     if [ -d "${dir}" ]; then rmdir "${dir}"; fi
    # done
    # for dir in 1*; do
    #     if [ -d "${dir}" ]; then rmdir "${dir}"; fi
    # done
    # for dir in 2*; do
    #     if [ -d "${dir}" ]; then rmdir "${dir}"; fi
    # done
    # rm other files created
    for file in *.nao *.tdipole *.auger xatom*.cmd convergedlist.txt converged_to_wrong_states.txt notconvergedlist.txt statistics.txt process.dat maketransitions.out; do
        if [ -f "${file}" ]; then rm "$file"; fi
    done
}

write_stats_of_output_files() {
    ################################################################################
    # echos stats to stdout
    ################################################################################
    GLOBAL_IN=(stats_total stats_DIIS stats_DIIS_small stats_DIIS_improved stats_DIIS_improved_big \
        stats_DIIS_not_converged stats_DIIS_not_converged_smooth stats_DIIS_not_converged_smooth_big \
        stats_smooth_converged stats_smooth_not_converged stats_overlap_small stats_converged_wrong \
        stats_different_energies stats_not_converged)
    check_GLOBAL_IN || return 1

    local string_width int_width format_string
    # format string for making a table
    string_width=60
    int_width=10
    format_string="%-${string_width}s (%${int_width}d): %${int_width}d\n"

    echo "# Statistics of maketransitions.sh"
    echo -e "${stats_files}"
    echo "# in () ${stats_text}"

        # echo "Combinations of occ and parentocc calculated (${stats_total[0]}): ${stats_total}"
    printf "${format_string}" \
        "Combinations of occ and parentocc calculated" ${stats_total[0]} ${stats_total[2]}

        # echo "- of those converged with DIIS (${stats_DIIS[0]}): ${stats_DIIS[2]}"
    printf "${format_string}" \
        "- of those converged with DIIS" ${stats_DIIS[0]} ${stats_DIIS[2]}

        # echo "-- of those overlap^2 < 0.5 (${stats_DIIS_small[0]}): ${stats_DIIS_small[2]}"
    printf "${format_string}" \
        "-- of those overlap^2 < 0.5" ${stats_DIIS_small[0]}  ${stats_DIIS_small[2]}

        # echo "--- of those improved by smooth (${stats_DIIS_improved[0]}): ${stats_DIIS_improved[2]}"
    printf "${format_string}" \
        "--- of those improved by smooth" ${stats_DIIS_improved[0]} ${stats_DIIS_improved[2]}

        # echo "---- of those overlap^2 > 0.5 (${stats_DIIS_improved_big[0]}): ${stats_DIIS_improved_big[2]}"
    printf "${format_string}" \
        "---- of those overlap^2 > 0.5" ${stats_DIIS_improved_big[0]} ${stats_DIIS_improved_big[2]}

        # echo "- cases in which DIIS didn't converge (${stats_DIIS_not_converged[0]}): ${stats_DIIS_not_converged[2]}"
    printf "${format_string}" \
        "- cases in which DIIS didn't converge" ${stats_DIIS_not_converged[0]} ${stats_DIIS_not_converged[2]}

        # echo "-- of those converged with smooth (${stats_DIIS_not_converged_smooth[0]}): ${stats_DIIS_not_converged_smooth[2]}"
    printf "${format_string}" \
        "-- of those converged with smooth" ${stats_DIIS_not_converged_smooth[0]} ${stats_DIIS_not_converged_smooth[2]}

        # echo "--- of those with overlap^2 > 0.5 (${stats_DIIS_not_converged_smooth_big[0]}): ${stats_DIIS_not_converged_smooth_big[2]}"
    printf "${format_string}" \
        "--- of those with overlap^2 > 0.5" ${stats_DIIS_not_converged_smooth_big[0]} ${stats_DIIS_not_converged_smooth_big[2]}

        # echo "- cases in which smooth was called and solution obtained(${stats_smooth_converged[0]}): ${stats_smooth_converged[2]}"
    printf "${format_string}" \
        "- cases in which smooth was called and solution obtained" ${stats_smooth_converged[0]} ${stats_smooth_converged[2]}

        # echo "- cases in which smooth didn't converge (${stats_smooth_not_converged[0]}): ${stats_smooth_not_converged[2]}"
    printf "${format_string}" \
        "- cases in which smooth didn't converge" ${stats_smooth_not_converged[0]} ${stats_smooth_not_converged[2]}
    
        # echo "- cases not converged with any method (${stats_not_converged[0]}): ${stats_not_converged[2]}"
    printf "${format_string}" \
        "- cases not converged with any method" ${stats_not_converged[0]} ${stats_not_converged[2]}
    
        # echo "- Different energies for different parentoccs (${stats_different_energies[0]}): ${stats_different_energies[2]}"
    printf "${format_string}" \
        "- Different energies for different parentoccs" ${stats_different_energies[0]} ${stats_different_energies[2]}

        # echo "- Combinations of occ and parentocc, that didn't converge to the solution with best overlap (${stats_converged_wrong[0]}: ${stats_converged_wrong[2]}"
    printf "${format_string}" \
        "- combinations of parentocc and occ that converged wrong" ${stats_converged_wrong[0]} ${stats_converged_wrong[2]}

        # echo "- cases of overlap^2 < 0.5 in total (${stats_overlap_small[0]}): ${stats_overlap_small[2]}"
    printf "${format_string}" \
        "- cases of overlap^2 < 0.5 in total" ${stats_overlap_small[0]} ${stats_overlap_small[2]}
    # else
    #     echo "Combinations of occ and parentocc calculated: ${stats_total[2]}"

    #     echo "- of those converged with DIIS: ${stats_DIIS[2]}"

    #     echo "-- of those overlap^2 < 0.5: ${stats_DIIS_small[2]}"

    #     echo "--- of those improved by smooth: ${stats_DIIS_improved[2]}"
        
    #     echo "---- of those overlap^2 > 0.5: ${stats_DIIS_improved_big[2]}"

    #     echo "- cases in which DIIS didn't converge: ${stats_DIIS_not_converged[2]}"

    #     echo "-- of those converged with smooth: ${stats_DIIS_not_converged_smooth[2]}"

    #     echo "--- of those with overlap^2 > 0.5: ${stats_DIIS_not_converged_psmooth_big[2]}"

    #     echo "- cases in which smooth was called and solution obtained: ${stats_smooth_converged[2]}"
        
    #     echo "- cases in which smooth didn't converge: ${stats_smooth_not_converged[2]}"
        
    #     echo "- cases not converged with any method: ${stats_not_converged[2]}"
        
    #     echo "- Different energies for different parentoccs: ${stats_different_energies[2]}"
        
    #     echo -n "- Combinations of occ and parentocc, that didn't converge to the solution with best overlap:"
    #     echo ${stats_converged_wrong[2]}

    #     echo "- cases of overlap^2 < 0.5 in total: ${stats_overlap_small[2]}"
    # fi
}

get_stats_of_output_files() {
    ################################################################################
    # Save stats to respective arrays
    ################################################################################
    # OUTPUTs global variables:
    # stats_total[2] 
    # stats_DIIS[2] 
    # stats_DIIS_small[2] 
    # stats_DIIS_improved[2] 
    # stats_DIIS_improved_big[2] 
    # stats_DIIS_not_converged[2] 
    # stats_DIIS_not_converged_smooth[2] 
    # stats_DIIS_not_converged_smooth_big[2] 
    # stats_smooth_converged[2] 
    # stats_smooth_not_converged[2] 
    # stats_overlap_small[2] 
    # stats_converged_wrong[2]
    # stats_different_energies[2] 
    # stats_not_converged[2] 
    ################################################################################
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1
    
    local outfile notconvergedlist converged_wrong
    if [ -z "${1}" ] || [ -z "${2}" ] || [ -z "${3}" ]; then
        >&2 echo -e "Pass ${FUNCNAME[0]} the filenames of output of maketransitions.sh/maketransitions.out,\nnotconvergedlist.txt and converged_to_wrong_states.txt as argument -> exit"
        return 1
    fi    
    outfile=$1
    notconvergedlist=$2
    converged_wrong=$3

    # echo "# Statistics of maketransitions.sh"
    stats_files[2]="# used files: \n# output of maketransitions.sh: ${1} \n# notconvergedlist: ${2} \n# converged wrong: ${3}"
    # echo "# in () are reference values for CO"

    # echo "Combinations of occ and parentocc calculated (5016):"
    stats_total[2]=$(grep "Calculating" "${outfile}" | wc -l)

    # echo "- of those converged with DIIS (4996):"
    stats_DIIS[2]=$(cat ${outfile} | sed -n -r "s/ *cmd +(.*)/\1*\1>0.01/p" | bc | grep "1" | wc -l)

    # echo "-- of those overlap^2 < 0.5 (151/149):"
    stats_DIIS_small[2]=$(cat ${outfile} | sed -n -r "s/ *cmd +(.*)/\1*\1<0.5/p" | bc | grep "1" | wc -l)

    # echo "--- of those improved by smooth:"
    stats_DIIS_improved[2]=$(grep -B 1 --group-separator "--" "cmd5" ${outfile} | \
        tr -d "\n" | (sed "s/--/\n/g"; echo "") | \
        sed -n -r "s/cmd (-?[01]{1}\.[0-9]*) +cmd5 (-?[01]{1}\.[0-9]*).*/\1*\1 < \2*\2/p" | \
        bc | grep "1" | wc -l)
   
    # echo "---- of those overlap^2 > 0.5 (56):"
    stats_DIIS_improved_big[2]=$(grep -B 1 --group-separator "--" "cmd5" ${outfile} | \
        tr -d "\n" | (sed "s/--/\n/g"; echo "") | \
        sed -n -r "s/cmd (-?[01]{1}\.[0-9]*) +cmd5 (-?[01]{1}\.[0-9]*).*/\2*\2>0.5/p" | \
        bc | grep "1" | wc -l)    

    # echo "- cases in which DIIS didn't converge (20):"
    stats_DIIS_not_converged[2]=$(grep "DIIS" "${notconvergedlist}" | wc -l)\

    # echo "-- of those converged with smooth (15):"
    stats_DIIS_not_converged_smooth[2]=$(grep -B 1 --group-separator "--" "cmd5" ${outfile} | \
        tr -d "\n" | (sed "s/--/\n/g"; echo "") | \
        sed -n "/Calculating/p" | sed -n -r "s/.*Calculating.*cmd5 (.*)/\1*\1>0.01/p" | \
        bc | grep "1" | wc -l)

    # echo "--- of those with overlap^2 > 0.5 (1):"
    stats_DIIS_not_converged_smooth_big[2]=$(grep -B 1 --group-separator "--" "cmd5" ${outfile} | \
        tr -d "\n" | (sed "s/--/\n/g";  echo "") | \
        sed -n "/Calculating/p" | sed -n -r "s/.*Calculating.*cmd5 (.*)/\1*\1>0.5/p" | \
        bc | grep "1" | wc -l)

    # echo "- cases in which smooth was called and solution obtained:"
    stats_smooth_converged[2]=$(grep "cmd5" "${outfile}" | wc -l)
    
    # echo "- cases in which smooth didn't converge:"
    stats_smooth_not_converged[2]=$(grep "smooth" "${notconvergedlist}" | wc -l)
 
    # echo "- cases not converged with any method:"
    stats_not_converged[2]=$(grep "not converged with any method" "${outfile}" | wc -l)
    
    # echo "- Different energies for different parentoccs:"
    stats_different_energies[2]=$(grep "Different energies for different parentoccupations:" "${outfile}" | wc -l)
    
    # echo "- Combinations of occ and parentocc, that didn't converge to the solution with best overlap:"
    stats_converged_wrong[2]=$(cat ${converged_wrong} | wc -l)

    # echo "- cases of overlap^2 < 0.5 in total (107/109):"
    stats_overlap_small[2]=$(cat ${outfile} | \
        sed -n -r "s/[*]? *[0-9]+ *-> *[0-9]+ *-?[0-9]+\.[0-9]* +(-?[0-9]+\.[0-9]*)/\1*\1<0.5/p" | \
        bc | grep "1" | wc -l)

    # echo "-- # of those that has  overlap^2 < 0.5 (151/149):"
    # number=0
    # grep "cmd " "${outfile}" | \
    #   while read line; do
    #     olap=$(get_last_num "$line")
    #     if [ $(echo "$olap*$olap < 0.5" | bc) -eq 1 ]; then echo " "; fi
    #   done | \
    #   wc -l
    # echo "-- # of those that has  overlap^2 < 0.5 (151/149):"
    # cat "${outfile}" | sed -n -r "s/ *cmd +(.*)/\1/p"  \
    #         | xargs -n1 | \
    #     while read number; 
    #     do 
    #         if [ $(echo "${number}*${number}<0.5" | bc) -eq 1 ]; then echo "yes"; fi; 
    #     done | wc -l
    # echo "- # of those converged with DIIS (4996):"    
    # grep "cmd " "${outfile}" | \
    #   while read line; do
    #     olap=$(get_last_num "$line")
    #     if [ $(echo "$olap*$olap > 0.01" | bc) -eq 1 ]; then echo " "; fi
    #   done | \
    #   wc -l    
}

occ_with_different_energies(){
    # Get occupations for which different parentoccs. converged to different energies
    # from output of maketransitions.sh
    #
    # search for lines in file $1 with "Different energies from different parentocc:"
    # and print the occ from the following line containing parentocc -> occ
    #
    # e.g. make file with two columns for two files from output with
    # paste <(occ_with_different_energies maketransitions.out | sort) \
    # <(occ_with_different_energies ~/Calculation/CO/transitions/maketransitions_output_ | sort)
    GLOBAL_IN=()
    check_GLOBAL_IN || return 1

    local file
    if [ -z "${1+x}" ]; then
        >&2 echo "Pass ${FUNCNAME[0]} filename as argument -> exit"
        return 1
    fi    
    file=$1
    cat $file | sed -n -r "/Different/I{n; s/[* ] [0-9]{7} -> ([0-9]{7}).*/\1/p}"
}

