!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module absorption
  !! author: Caroline Arnold (2016)
  !! author: Ludger Inhester (2017-2023)
  !! Module for calculating bound-bound aborption transition cross sections
  use constants
  use wavefunction
  use fluorescence
  use atom_in_mol
  use molecular_grid
  use operations
  use process
  use averaging
  use mcscf
  use spin_coupling
  use basis_func
  use cubefile
  use matrix
  use molden
  use rwfile
  use scf
  use hartreefock
  use cis
  use mergesort
  use population

  implicit none
  private
  public :: calc_oscillator_strength_CI, calc_oscillator_strength_CIS, calc_oscillator_strength
contains
  subroutine calc_oscillator_strength(P, B, F, G, OS_SO, tDip)
    !! The subroutine calc_absorption_CS calculates the absorption cross sections based on orbital energies
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Grid3D), intent(in) :: G
    real(long), intent(out), optional :: OS_SO(B%Nr_basis, B%Nr_basis)    !! absorption CS between MOs including factor from spin orbitals
    real(long), intent(out), optional :: tDip(B%Nr_basis, B%Nr_basis, 3)   !! transition dipole moment matrix elements in mo space
    real(long), dimension(B%N_basis, B%N_basis, 3) :: tdipole_bf ! bf space
    real(long), dimension(B%Nr_basis, B%Nr_basis, 3) :: tdipole_mo ! mo space
    integer :: hole, part ! molecular orbitals: hole and particle loop idx
    real(long) :: energy_transition ! energy of transition
    real(long) :: TotalOS ! total cross section
    real(long) :: PartOS ! cross section per particle
    real(long) :: OS ! cross section per hole and particle
    integer :: pol ! polarization
    !real(long) :: n_hole, n_part
    character(len=15) :: string
    character(len=6) :: str_hole
    integer :: OUTPUT
    real(long) :: factor
    OUTPUT = P%OUTPUT

    ! initialize optional output array
    if (present(OS_SO)) OS_SO(:, :) = ZERO

    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do
    if (P%print_level .gt. 1) then
      if(P%yes_absorptioncs) then
        write (OUTPUT, "('# Cross sections assuming photon energy', f13.2, x, 'eV')") P%PE * AU2EV
        write (OUTPUT, "('#                     bandwidth (stdev)', f13.2, x, 'eV')") P%PE_width * AU2EV
        write (OUTPUT, "('#                     bandwidth (FWHM) ', f13.2, x, 'eV')") &
          P%PE_width *  TWO*sqrt(TWO*log(TWO)) * AU2EV
      end if
      write (OUTPUT, "('#'/'# Absorption to bound states (based on orbital energies):')")
      if(P%yes_absorptioncs) then
        write (OUTPUT, "('#    I     F    ',a13,2x,a13,8x,3a13,x,a13)") &
        "x-ray E(eV)", "oscill. strength", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z", "cs (a.u.)"
      else
        write (OUTPUT, "('#    I     F    ',a13,2x,a13,8x,3a13)") &
        "x-ray E(eV)", "oscill. strength", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
      end if
    end if

    ! Save transition dipole moments for passing it to other interfaces
    if (present(tDip)) tDip = tdipole_mo

    ! Absorption starts from an occupied hole state and goes to an unoccupied
    ! particle state

    TotalOS = ZERO
    ! Absorption from MO i to MO j
    do part = 1, size(F%N_occ)
      PartOS = ZERO
      if (.NOT. is_occupied(P, F, F%N_occ, part)) cycle  ! Only from occupied MO
      do hole = 1, size(F%N_occ)
        OS = ZERO
        ! Only to MO with negative energies because the quality
        ! of molecular Rydberg states is questionable due to missing
        ! Latter correction
        !          if( F%CE(hole) > 0.0 ) cycle               ! Comment out this line to
        !          INCLUDE Rydberg states
        if (.NOT. has_hole(P, F, F%N_occ, hole)) cycle  ! Only to empty MO
        if (is_equivalent(P, F, hole, part)) cycle     ! i != j
        energy_transition = F%CE(hole) - F%CE(part)
        if (energy_transition .le. 0) cycle
        factor = stat_factor_not_occupied(P, F, F%N_occ, hole)*TWO*stat_factor_occupied(P, F, F%N_occ, part)
        write (str_hole, "('*',f5.2)") factor
        write (string, "(f13.2)") energy_transition*AU2EV
        ! transition: part <- hole
        OS = oscillator_strength(tdipole_mo(part, hole, :), energy_transition)
        if (present(OS_SO)) OS_SO(hole, part) = OS * factor
        if (P%print_level .gt. 1) then
          if(P%yes_absorptioncs) then
            write (OUTPUT, "('#    ',i3,' - 'i3,' : ',a13,2x,es13.5,a6,2x,3es13.5,es13.5)") &
              part, hole, trim(string), OS, str_hole, tdipole_mo(part, hole, :),&
              TWO * PI**2 / SPEED_C_AU * OS * factor * gauss_profile(energy_transition - P%PE, P%PE_width)
          else
            write (OUTPUT, "('#    ',i3,' - 'i3,' : ',a13,2x,es13.5,a6,2x,3es13.5)") &
              part, hole, trim(string), OS, str_hole, tdipole_mo(part, hole, :)
          end if
        end if
        PartOS = PartOS + factor * OS
      end do ! hole loop
      TotalOS = TotalOS + PartOS
    end do ! part loop
    if (P%print_level .gt. 1) then
      if (TotalOS .gt. ZERO) then
        write (OUTPUT, "('# ',12x,'Total oscill. strength = ',es13.5,' a.u.')") TotalOS
      else
        write (OUTPUT, "('#   NONE')")
      end if
    end if
  end subroutine calc_oscillator_strength

  subroutine calc_oscillator_strength_CIS(P, A, B, F, G, state_in, os, tDip)
    !!calculates the oscillator strenght based on CIS energies
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Grid3D), intent(in) :: G

    integer, intent(in), optional :: state_in ! given state number
    ! absorption CS between MOs including factor from spin orbitals
    real(long), intent(out), optional :: os(size(F%E_cis) + 1)
    real(long), intent(out), optional :: tDip(size(F%E_cis) + 1, 3) ! mo space

    ! transition dipole moment matrix elements
    real(long), dimension(B%N_basis, B%N_basis, 3) :: tdipole_bf ! bf space
    real(long), dimension(B%Nr_basis, B%Nr_basis, 3) :: tdipole_mo ! mo space
    real(long), allocatable :: tdipole_states(:, :) ! state space
    real(long) :: energy_transition ! energy of transition
    real(long) :: TotalOS ! total cross section
    real(long) :: PartOS ! cross section per particle
    real(long) :: thisOS ! cross section per hole and particle
    integer :: pol ! polarization
    character(len=16) ::string
    integer :: OUTPUT
    integer :: nstates, l, m, m2, n
    integer :: r, s, t, u
    integer :: state
    type(WF_data) :: F_temp
    real(long), allocatable :: tdens(:, :)
    real(long), allocatable :: LC_hole(:), LC_part(:)
    character(len=512) :: moldenfilename ! filename
    real(long) :: traceDS
    OUTPUT = P%OUTPUT

    if (present(state_in)) then
      state = state_in
    else
      state = P%state
    end if
    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do
    if (P%print_level .gt. 1) then
      if(P%yes_absorptioncs) then
        write (OUTPUT, "('# Cross sections assuming photon energy', f13.2, x, 'eV')") P%PE * AU2EV
        write (OUTPUT, "('#                     bandwidth (stdev)', f13.2, x, 'eV')") P%PE_width * AU2EV
        write (OUTPUT, "('#                     bandwidth (FWHM) ', f13.2, x, 'eV')") &
          P%PE_width *  TWO*sqrt(TWO*log(TWO)) * AU2EV
      end if
      write (OUTPUT, "('#'/'# Absorption to bound states (based on CIS energies):')")
      if(P%yes_absorptioncs) then
        write (OUTPUT, "('#    I           F          ',a13,2x,a13,8x,3a17,x,a17)") &
          "x-ray E(eV)", "oscill. strength (a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z", "cs (a.u.)"
      else
        write (OUTPUT, "('#    I           F          ',a13,2x,a13,8x,3a17)") &
          "x-ray E(eV)", "oscill. strength (a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
      end if
    end if

    ! Absorption starts from an occupied hole state and goes to an unoccupied
    ! particle state

    nstates = size(F%E_cis)
    n = size(F%CIS_excitation, 2)
    allocate (tdipole_states(0:nstates, 3))
    tdipole_states(:, :) = ZERO

    do l = 0, nstates
      if (state .eq. 0) then
        if (l .eq. 0) then
          do r = 1,size(F%N_occ,1)
            if(F%N_occ(r)==2)&
              tdipole_states(l, :) = tdipole_states(l, :) + TWO * tdipole_mo(r, r, :)
          end do
        else
          ! 0 -> l
          do m = 1, n
            r = F%CIS_excitation(1, m)
            t = F%CIS_excitation(2, m)
            tdipole_states(l, :) = tdipole_states(l, :) + sqrt(TWO)*F%C_CIS(m, l)*tdipole_mo(r, t, :)
          end do
        end if! l==0
      else
        if (l .eq. 0) then
          ! state -> 0
          do m = 1, n
            r = F%CIS_excitation(1, m)
            t = F%CIS_excitation(2, m)
            tdipole_states(0, :) = tdipole_states(0, :) + sqrt(TWO)*F%C_CIS(m, state)*tdipole_mo(r, t, :)
          end do
        else
          ! state -> l
          do m = 1, n
            do m2 = 1, n
              if (m .eq. m2) then ! hole and part are the same
                do r = 1,size(F%N_occ,1)
                  if(F%N_occ(r)==2)&
                    tdipole_states(l, :) = tdipole_states(l, :) + TWO * F%C_CIS(m, l)*F%C_CIS(m2, state) * tdipole_mo(r, r, :)
                end do
                r = F%CIS_excitation(2, m)
                tdipole_states(l, :) = tdipole_states(l, :) + F%C_CIS(m, l)*F%C_CIS(m2, state) * tdipole_mo(r, r, :)
                t = F%CIS_excitation(1, m)
                tdipole_states(l, :) = tdipole_states(l, :) - F%C_CIS(m, l)*F%C_CIS(m2, state) * tdipole_mo(t, t, :)
              else if (F%CIS_excitation(1, m) .eq. F%CIS_excitation(1, m2)) then ! hole is same
                r = F%CIS_excitation(2, m)
                s = F%CIS_excitation(2, m2)
                tdipole_states(l, :) = tdipole_states(l, :) + F%C_CIS(m, l)*F%C_CIS(m2, state) * tdipole_mo(r, s, :)
              else if (F%CIS_excitation(2, m) .eq. F%CIS_excitation(2, m2)) then ! part is same
                t = F%CIS_excitation(1, m)
                u = F%CIS_excitation(1, m2)
                tdipole_states(l, :) = tdipole_states(l, :) - F%C_CIS(m, l)*F%C_CIS(m2, state) * tdipole_mo(t, u, :)
              end if
            end do
          end do
        end if! l=0
      end if
    end do

    if (present(os)) &
      os(:) = ZERO
    if (present(tDIP)) &
      tDIP(1:nstates + 1, :) = tdipole_states(0:nstates, :)

    ! correct tdipole to the state itself so it gives the charge dipole matrix element (with respect to the center of charge)
    traceDS = matTrace(F%D(:,:).m.B%S(:,:))
    tdipole_states(state, 1) = sum(A(:)%fcharge * A(:)%x) - tdipole_states(state, 1) - B%env(2) * traceDS
    tdipole_states(state, 2) = sum(A(:)%fcharge * A(:)%y) - tdipole_states(state, 2) - B%env(3) * traceDS
    tdipole_states(state, 3) = sum(A(:)%fcharge * A(:)%z) - tdipole_states(state, 3) - B%env(4) * traceDS

    TotalOS = ZERO
    do l = 0, nstates
      !if (l .eq. state) cycle
      PartOS = ZERO
      if (state .eq. 0 .and. l .eq. 0) then
        energy_transition = ZERO
      else if (state .eq. 0 .and. l .gt. 0) then
        energy_transition = F%E_CIS(l)
      else if (state .gt. 0 .and. l .eq. 0) then
        energy_transition = -F%E_CIS(state)
      else
        energy_transition = F%E_CIS(l) - F%E_CIS(state)
      end if
      thisOS = oscillator_strength(tdipole_states(l, :), abs(energy_transition))
      TotalOS = TotalOS + thisOS
      if (present(os)) &
        os(l + 1) = thisOS
      if (P%print_level .gt. 1) then
        write (string, "(f13.2)") energy_transition*AU2EV
        if(P%yes_absorptioncs) then
          write (OUTPUT, "('#    ',i9,' - 'i9,' : ',2x,a13,2x,es17.5,2x,3es17.5,x,es17.5)") &
            state, l, trim(string), thisOS, tdipole_states(l, :), &
            TWO * PI**2 / SPEED_C_AU * thisOS * gauss_profile(energy_transition - P%PE, P%PE_width)
        else
          write (OUTPUT, "('#    ',i9,' - 'i9,' : ',2x,a13,2x,es17.5,2x,3es17.5)") &
            state, l, trim(string), thisOS, tdipole_states(l, :)
        end if
      end if
    end do

    if (P%print_level .gt. 1) then
      if (TotalOS .gt. ZERO) then
        write (OUTPUT, "('# ',12x,'Total oscillator strength = ',es13.5,' a.u.')") TotalOS
      else
        write (OUTPUT, "('#   NONE')")
      end if
    end if
    deallocate (tdipole_states)

    if (allocated(P%nat_trans_orb)) then
      allocate(tdens(size(F%LC,2), size(F%LC,2)))
      do l = 0, nstates
        if(all(P%nat_trans_orb /= l)) cycle
        if(state == l) cycle
        if (P%print_level > 1) then
          write (OUTPUT, '("# building transition-density matrix for ", i8," to excited state ",i8)') state,l
        end if
        if(state == 0 .and. l>0) then
          tdens = ZERO
          if(F%multp==1) then
            do r = 1, size(F%CIS_excitation, 2)
              tdens(F%CIS_excitation(1, r), F%CIS_excitation(2, r)) = sqrt(TWO) * F%C_cis(r, l)
            end do
            tdens = tdens / sqrt(TWO) ! This factor results in a normalized transition density Tr(Tdens tdens^T) = 1
          end if
        else if(state /=0 .and. l == 0) then
          tdens = ZERO
          if(F%multp==1)then
            do r = 1, size(F%CIS_excitation, 2)
              tdens(F%CIS_excitation(2, r), F%CIS_excitation(1, r)) = sqrt(TWO) * F%C_cis(r, state)
            end do
            tdens = tdens / sqrt(TWO) ! This factor results in a normalized transition density Tr(Tdens tdens^T) = 1
          end if
        else if(state > 0 .and. l > 0) then
          tdens = ZERO
          if(F%multp==3)then
            stop 'not implemented'
          end if
          do m = 1, n
            do m2 = 1, n
              if (m .eq. m2) then
                do r=1,size(F%N_occ)
                  if(F%N_occ(r)==2)&
                    tdens(r, r) = tdens(r, r) + F%C_CIS(m, l) * F%C_CIS(m2, state)
                end do
                r = F%CIS_excitation(2, m)
                tdens(r, r) = tdens(r, r) + F%C_CIS(m, l) * F%C_CIS(m2, state)
                t = F%CIS_excitation(1, m)
                tdens(t, t) = tdens(t, t) + F%C_CIS(m, l) * F%C_CIS(m2, state)
              else if (F%CIS_excitation(1, m) .eq. F%CIS_excitation(1, m2)) then ! hole is same
                r = F%CIS_excitation(2, m2)
                s = F%CIS_excitation(2, m)
                tdens(r, s) = tdens(r, s) + F%C_CIS(m, l) * F%C_CIS(m2, state)
              else if (F%CIS_excitation(2, m) .eq. F%CIS_excitation(2, m2)) then ! part is same
                t = F%CIS_excitation(1, m2)
                u = F%CIS_excitation(1, m)
                tdens(t, u) = tdens(t, u) - F%C_CIS(m, l) * F%C_CIS(m2, state)
              end if
            end do
          end do
        end if! state>0 and l>0

        allocate (LC_hole(size(tdens, 1)))
        allocate (LC_part(size(tdens, 2)))
        write(OUTPUT,'("# Trace of (tdens * tdens^T) = ",f10.5)') sum(tdens(:,:)**2)

        call make_nat_trans_orbitals(OUTPUT, tdens, LC_hole, LC_part)

        if (P%molden_file .ne. "") then
          F_temp = F
          F_temp%LC = ZERO
          F_temp%LC(:, 1) = F%LC(:, :) .m. LC_hole(:)
          F_temp%LC(:, 2) = F%LC(:, :) .m. LC_part(:)
          write (moldenfilename, '("trans_",i0.4,".molden")') l
          call write_molden_file(P, A, B, F_temp, moldenfilename)
        end if
        deallocate(LC_hole)
        deallocate(LC_part)
      end do! l
      deallocate(tdens)
    end if

  end subroutine calc_oscillator_strength_CIS

  subroutine calc_oscillator_strength_CI(P, A, B, F, G)
    !! calculates the oscillator strenght employing a CI calculation
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    type(Grid3D), intent(in) :: G
    integer :: OUTPUT
    real(long), dimension(B%N_basis, B%N_basis, 3) :: tdipole_bf ! bf space
    real(long), dimension(B%Nr_basis, B%Nr_basis, 3) :: tdipole_mo ! mo space
    integer, allocatable :: occs_temp(:, :)
    integer, allocatable, target :: initial_occ(:, :), final_occ(:, :) ! initial occupation
    integer, allocatable, target :: initial_occ_orig(:, :) ! initial occupation, no symmetry restrictions
    real(long), allocatable :: initial_csf_c(:)
    real(long), allocatable :: initial_CI_matrix(:, :), final_CI_matrix(:, :) ! CI matrix for initial states
    real(long), allocatable :: initial_CI_energy(:), final_CI_energy(:) ! energy of initial states
    type(CSF), allocatable :: initial_csf_array(:), final_csf_array(:) ! array for csfs initial state
    real(long), allocatable :: csf_c2(:, :)
    real(long), allocatable :: CI_diagonal(:)
    real(long), allocatable :: dm(:,:,:,:)
    integer :: k, l, m, n
    integer :: nev
    integer, allocatable :: idx(:)
    integer :: N_initial_csfs, N_final_csfs ! N of initial csfs
    integer :: final_Nopen
    integer :: initial_multp, final_multp
    real(long) :: E_ini
    integer :: N_final_occ
    logical :: redundant_occ, skip_sym, skip_spin
    integer :: N_skip_sym, N_skip_spin, N_redundant
    real(long) :: abs_energy, norm
    integer :: Ncore
    integer :: irrep_occ, n_irrep, f_irrep
    integer :: ic, io2, fc
    integer :: i, j, pol
    integer :: hole, ihole, virt, hole2, virt2
    integer :: i_final_occ
    real(long), allocatable :: csf_tdipole(:, :, :) ! tdipole amplitude for each initial - final csf  pair
    real(long), allocatable :: state_tdipole(:, :) ! tdipole amplitude - initial state to all final states
    real(long), allocatable :: temp(:, :) !
    real(long), allocatable :: temp2(:) !
    real(long) :: ChannelOS, TotalOS
    real(long), allocatable :: csf_O(:, :) !overlap MO MO' for each csf
    type(WF_data) :: F_final
    !type(Basis) :: B_final
    integer dipole_irrep(3)
    real(long), allocatable :: tdens(:, :)
    real(long), allocatable :: densMatrix_hole(:, :), densMatrix_part(:, :)
    real(long), allocatable :: tdens_csf(:, :, :)
    real(long), allocatable :: tdens_csf2(:, :, :)
    real(long), allocatable :: LC_hole(:), LC_part(:)
    integer, allocatable :: final_hole(:)
    integer, allocatable :: core(:)
    integer, allocatable :: occupied(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    type(UmatrixTable), target :: UTini, UTfin
    type(UmatrixTable), pointer :: UTjoint
    integer :: NOpenMax, nintern
    character(len=512) :: moldenfilename ! filename
    real(long), allocatable :: n_occ(:) ! natural occupation numbers of final state
    real(long), allocatable :: opdm(:, :)
    real(long), allocatable :: Dmatrix(:, :)
    type(WF_data) :: F_temp
    real(long) :: os
    real(long) :: pop(size(A,1))
    OUTPUT = P%OUTPUT
    Ncore = F%Ncore
    ! set the symmetry properties of the dipole operator
    if (P%yes_sym) then
      if (B%sym_label .eq. "D2h") then
        dipole_irrep = (/8, 7, 6/)
      else if (B%sym_label .eq. "D2") then
        dipole_irrep = (/4, 3, 2/)
      else if (B%sym_label .eq. "C2v") then
        dipole_irrep = (/3, 4, 1/)
      else if (B%sym_label .eq. "C2") then
        dipole_irrep = (/2, 2, 1/)
      else if (B%sym_label .eq. "Cs") then
        dipole_irrep = (/2, 2, 1/)
      else if (B%sym_label .eq. "C1") then
        dipole_irrep = (/1, 1, 1/)
      else
        stop 'unrecognized symmetry label'
      end if
    end if
    ! set the array of final_holes: by default all core orbitals
    if (allocated(P%final_hole)) then
      allocate (final_hole(size(P%final_hole)))
      final_hole = P%final_hole
    else
      allocate (final_hole(Ncore))
      final_hole = (/(j, j=1, Ncore)/)
    end if
    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do
    ! obtain initial_occ array
    if (P%yes_cis) then
      allocate (initial_occ(size(F%N_occ), size(F%CIS_excitation, 2) + 1))
      initial_occ(:, 1) = F%N_occ(:)
      do i = 1, size(initial_occ, 2) - 1
        initial_occ(:, i + 1) = F%N_occ(:)
        initial_occ(F%CIS_excitation(1, i), i + 1) = initial_occ(F%CIS_excitation(1, i), i + 1) - 1
        initial_occ(F%CIS_excitation(2, i), i + 1) = initial_occ(F%CIS_excitation(2, i), i + 1) + 1
      end do
    else if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, initial_occ)
    else if (allocated(P%active_space_MO)) then
      call active_space_occs(F%N_occ, P%active_space_MO, initial_occ)
    else
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    end if
    !
    if (allocated(P%final_active_space_MO)) then
      ! initial_occ_orig from final active space
      call active_space_occs(F%N_occ, P%final_active_space_MO, initial_occ_orig)
    else
      ! initial_occ_orig = initial_occ
      allocate (initial_occ_orig, source=initial_occ)
    end if
    if (P%yes_sym .and. F%state_irrep .gt. 0) then
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, initial_occ, occs_temp)
      deallocate (initial_occ)
      allocate (initial_occ(size(occs_temp, 1), size(occs_temp, 2)))
      initial_occ = occs_temp
      deallocate (occs_temp)
    end if

    N_initial_csfs = 0
    initial_multp = P%multp
    write (P%OUTPUT, "('# Initial configurations for spin multp: ',i4)") initial_multp
    call construct_csf_array(P, B, F, initial_multp, P%state_irrep, initial_occ, initial_csf_array)
    N_initial_csfs = size(initial_csf_array)
    write (P%OUTPUT, "('# N_initial_csfs: ',i7)") N_initial_csfs
    allocate (initial_CI_matrix(N_initial_csfs, N_initial_csfs))
    allocate (initial_CI_energy(N_initial_csfs))
    allocate (initial_csf_c(N_initial_csfs))
    NOpenMax = maxval(count(initial_occ .eq. 1, dim=1))
    call constructUMatrixTable(NOpenMax, initial_multp, UTini)

    !
    allocate (active(count((/((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0)), j=1, size(initial_occ, 1))/))))
    allocate (core(count((/((all(initial_occ(j, :) .eq. 2)), j=1, size(initial_occ, 1))/))))
    allocate (virtual(count((/(all(initial_occ(j, :) .eq. 0), j=1, size(initial_occ, 1))/))))
    allocate (occupied(size(active, 1) + size(core, 1)))
    active = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0))), j=1, size(initial_occ, 1))/))
    core = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(all(initial_occ(j, :) .eq. 2), j=1, size(initial_occ, 1))/))
    virtual = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(all(initial_occ(j, :) .eq. 0), j=1, size(initial_occ, 1))/))
    occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))
    do j = size(active), 1, -1
      if (any(sum(initial_occ(active(j:), :), dim=1) .gt. 1)) exit
    end do
    nintern = j
    write (6, '("#     core:",99i4)') core
    write (6, '("#   active:",99i4)') active
    write (6, '("# internal:",99i4)') active(:nintern)
    write (6, '("# external:",99i4)') active(nintern + 1:)

    allocate (F_core(size(F%LC, 2), size(F%LC, 2)))
    call construct_h0_matrix(G, B, F)
    F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)
    if (P%yes_direct_scf) then
      call calculate_F_inactive_direct(B, core, F%LC, F_core)
    else
      call calculate_F_inactive(P, B, core, F%LC, F_core)
    end if
    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do
    write (P%OUTPUT, '("# computing CI matrix")')
    call transform_twoel_allactive(P, B, F, active(:nintern), P%OUTPUT)
    if (nintern .lt. size(active)) then
      if (nintern .gt. 0) then
        call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
        call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      end if
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)
      if (nintern .gt. 0) then
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        deallocate (F%teint_cdei)
      end if
    else
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix)
    end if
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# done")')
    deallocate (F%teint_abcd)
    if (allocated(P%ci_coef)) then
      initial_csf_c(:) = P%ci_coef(:)
      norm = sqrt(sum(initial_csf_c(:)**2))
      initial_csf_c(:) = initial_csf_c(:)/norm
      if (P%yes_verbose) then
        write (OUTPUT, '("# fixed coefficients:")')
      end if
      E_ini = dot_product(initial_csf_c, matmul(initial_CI_matrix, initial_csf_c))
      initial_CI_matrix(:, P%state) = initial_csf_c
    else
      ! solve it
      write (P%OUTPUT, '("# solving CI matrix")')
      call get_eigenvalues(initial_CI_matrix, initial_CI_energy)
      write (P%OUTPUT, '("# done")')
      E_ini = initial_CI_energy(P%state)
      initial_csf_c(:) = initial_CI_matrix(:, P%state)
    end if
    write (P%OUTPUT, "('# Initial state ',i4,x,'spin multp=',i4,x,'energy=',f15.5)") &
      P%state, P%multp, E_ini + nuclear_repulsion_energy(A, P)
    do ic = 1, size(initial_CI_matrix, 1)
      if (abs(initial_csf_c(ic)) .lt. 0.1) cycle
      write (P%OUTPUT, '("# csf ",i7,x,f15.5,x,a)') &
        ic, initial_csf_c(ic), trim(csfstring(initial_csf_array(ic)%occ, initial_csf_array(ic)%csf))
    end do
    ! prepare OS
    TotalOS = ZERO
    final_multp = initial_multp
    write (P%OUTPUT, "('# Final Spin multiplicity ',i3)") final_multp
    if (P%yes_sym) then
      n_irrep = B%nirrep
    else
      n_irrep = 1
    end if
    do f_irrep = 1, n_irrep
      if (P%yes_sym .and. P%state_irrep .gt. 0) then
        if (B%irrep_prod(P%state_irrep, dipole_irrep(1)) .ne. f_irrep .and. &
          B%irrep_prod(P%state_irrep, dipole_irrep(2)) .ne. f_irrep .and. &
          B%irrep_prod(P%state_irrep, dipole_irrep(3)) .ne. f_irrep) then
          write (P%OUTPUT, "('# Final state irrep skipped because of symmetry',x,a,x,'* [ ',a,x,a,x,a,' ] /= ',a)") &
            trim(B%irrep_name(P%state_irrep)), &
            trim(B%irrep_name(dipole_irrep(1))), trim(B%irrep_name(dipole_irrep(2))), trim(B%irrep_name(dipole_irrep(3))), &
            trim(B%irrep_name(f_irrep))
          cycle
        end if
        write (P%OUTPUT, "('# Final state irrep ',i3,x,a)") f_irrep, trim(B%irrep_name(f_irrep))
      else if (P%yes_sym) then
        write (P%OUTPUT, "('# Final state irrep ',i3,x,a)") f_irrep, trim(B%irrep_name(f_irrep))
      end if

      N_final_occ = 0
      if (allocated(P%final_gas)) then
        call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%final_gas, final_occ)
        N_final_occ = size(final_occ, 2)
      else
        ! Loop over hole and valence pair combinations: count configurations
        do io2 = 1, size(initial_occ_orig, 2)
          !        do hole=1,size(initial_occ,1)
          do ihole = 1, size(final_hole)
            hole = final_hole(ihole)
            if (initial_occ_orig(hole, io2) .eq. 0) cycle
            do virt = 1, size(initial_occ_orig(:, io2))
              if (initial_occ_orig(virt, io2) .eq. 2) cycle
              if (virt .eq. hole) cycle
              if (P%CIS_limit(2) .gt. 0 .and. virt .gt. P%CIS_limit(2)) cycle
              if (P%CIS_limit(1) .gt. 0 .and. hole .gt. P%CIS_limit(1)) cycle
              N_final_occ = N_final_occ + 1
              !
              if (P%yes_process_CISD) then
                do hole2 = Ncore + 1, size(initial_occ_orig(:, io2))
                  if (initial_occ_orig(hole2, io2) .eq. 0) cycle
                  if (initial_occ_orig(hole2, io2) .eq. 1 .and. hole2 .eq. virt) cycle
                  do virt2 = hole2 + 1, size(initial_occ_orig(:, io2))
                    if (initial_occ_orig(virt2, io2) .eq. 2) cycle
                    if (initial_occ_orig(virt2, io2) .eq. 1 .and. virt .eq. virt2) cycle
                    N_final_occ = N_final_occ + 1
                  end do
                end do
              end if
              !
            end do! virt
          end do! hole
        end do
        allocate (final_occ(size(F%N_occ), N_final_occ))
        write (P%OUTPUT, "('# N_final_occ count ',i15)") N_final_occ
        N_final_csfs = 0
        ! write final_occ and count csfs
        i_final_occ = 0
        N_skip_sym = 0
        N_skip_spin = 0
        N_redundant = 0
        do io2 = 1, size(initial_occ_orig, 2)
          do ihole = 1, size(final_hole)
            hole = final_hole(ihole)
            if (initial_occ_orig(hole, io2) .eq. 0) cycle
            do virt = 1, size(initial_occ_orig(:, io2))
              if (initial_occ_orig(virt, io2) .eq. 2) cycle
              if (virt .eq. hole) cycle
              if (P%CIS_limit(2) .gt. 0 .and. virt .gt. P%CIS_limit(2)) cycle
              if (P%CIS_limit(1) .gt. 0 .and. hole .gt. P%CIS_limit(1)) cycle
              i_final_occ = i_final_occ + 1
              ! Final occupation
              final_occ(:, i_final_occ) = initial_occ_orig(:, io2)
              final_occ(virt, i_final_occ) = final_occ(virt, i_final_occ) + 1
              final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) - 1
              redundant_occ = .FALSE.
              do j = 1, i_final_occ - 1
                if (all(final_occ(:, j) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
                if (redundant_occ) exit
              end do
              skip_spin = .FALSE.
              skip_sym = .FALSE.
              final_Nopen = count(final_occ(:, i_final_occ) .eq. 1)
              if (final_multp - 1 .gt. final_Nopen) skip_spin = .TRUE.
              if (P%yes_sym) then
                irrep_occ = 1
                do i = 1, size(initial_occ_orig(:, i_final_occ))
                  if (final_occ(i, i_final_occ) .eq. 1) &
                       irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
                end do
                if (irrep_occ .ne. f_irrep) skip_sym = .TRUE.
              end if
              if (skip_sym .or. skip_spin .or. redundant_occ) then
                if (redundant_occ) N_redundant = N_redundant + 1
                if (skip_sym) N_skip_sym = N_skip_sym + 1
                if (skip_spin) N_skip_spin = N_skip_spin + 1
                i_final_occ = i_final_occ - 1
              end if
              if (P%yes_process_CISD) then
                do hole2 = Ncore + 1, size(initial_occ_orig(:, io2))
                  if (initial_occ_orig(hole2, io2) .eq. 0) cycle
                  if (initial_occ_orig(hole2, io2) .eq. 1 .and. hole2 .eq. virt) cycle
                  do virt2 = hole2 + 1, size(initial_occ_orig(:, io2))
                    if (initial_occ_orig(virt2, io2) .eq. 2) cycle
                    if (initial_occ_orig(virt2, io2) .eq. 1 .and. virt .eq. virt2) cycle
                    if (P%CIS_limit(2) .gt. 0 .and. virt2 .gt. P%CIS_limit(2)) cycle
                    if (P%CIS_limit(1) .gt. 0 .and. hole2 .gt. P%CIS_limit(1)) cycle
                    i_final_occ = i_final_occ + 1
                    final_occ(:, i_final_occ) = initial_occ_orig(:, io2)
                    final_occ(virt, i_final_occ) = final_occ(virt, i_final_occ) + 1
                    final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) - 1
                    final_occ(hole2, i_final_occ) = final_occ(hole2, i_final_occ) - 1
                    final_occ(virt2, i_final_occ) = final_occ(virt2, i_final_occ) + 1
                    final_Nopen = count(final_occ(:, i_final_occ) .eq. 1)
                    skip_sym = .FALSE.
                    skip_spin = .FALSE.
                    if (final_multp - 1 .gt. final_Nopen) skip_spin = .TRUE.
                    if (P%yes_sym) then
                      irrep_occ = 1
                      do i = 1, size(final_occ, 1)
                        if (final_occ(i, i_final_occ) .eq. 1) &
                             irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
                      end do
                      if (irrep_occ .ne. f_irrep) skip_sym = .TRUE.
                    end if
                    redundant_occ = .FALSE.
                    do j = 1, i_final_occ - 1
                      if (all(final_occ(:, j) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
                      if (redundant_occ) exit
                    end do
                    if (redundant_occ .or. skip_sym .or. skip_spin) then
                      if (redundant_occ) N_redundant = N_redundant + 1
                      if (skip_sym) N_skip_sym = N_skip_sym + 1
                      if (skip_spin) N_skip_spin = N_skip_spin + 1
                      i_final_occ = i_final_occ - 1
                    end if
                  end do
                end do
              end if
            end do! virt
          end do! hole
        end do! io2
        write (P%OUTPUT, "('# excluded because symmetry  ',i7)") N_skip_sym
        write (P%OUTPUT, "('# excluded because of spin  ',i7)") N_skip_spin
        write (P%OUTPUT, "('# redundant ',i7)") N_redundant
        N_final_occ = i_final_occ
      end if! final_gas

      if (P%yes_sym) then
        call filter_sym_occs(B%irrep_prod, F%irrep_MO, f_irrep, final_occ(:, :N_final_occ), occs_temp)
        write (P%OUTPUT, "('# excluded because of symmetry ',i7)") N_final_occ - size(occs_temp, 2)
      else
        allocate (occs_temp(size(final_occ, 1), 1:N_final_occ))
        occs_temp = final_occ(:, 1:N_final_occ)
      end if
      N_final_occ = size(occs_temp, 2)
      deallocate (final_occ)
      allocate (final_occ(size(occs_temp, 1), size(occs_temp, 2)))
      final_occ(:, :) = occs_temp(:, :)
      deallocate (occs_temp)
      write (P%OUTPUT, "('# N_final_occ ',i7)") N_final_occ
      call construct_csf_array(P, B, F, final_multp, f_irrep, final_occ(:, 1:N_final_occ), final_csf_array)
      N_final_csfs = size(final_csf_array)
      write (P%OUTPUT, "('# N_final_csfs ',i7)") N_final_csfs
      ! if there are no spin configurations for this spin continue with next spin multiplicity
      if (N_final_csfs .eq. 0) then
        deallocate (final_occ)
        cycle
      end if
      NOpenMax = maxval(count(final_occ .eq. 1, dim=1))
      call constructUMatrixTable(NOpenMax, final_multp, UTfin)
      allocate (final_CI_energy(N_final_csfs))
      allocate (csf_tdipole(N_initial_csfs, N_final_csfs, 3))
      ! Loop over final configuration state functions
      ! Calculate for every configuration the transtion dipole matrix element
      csf_tdipole(:, :, :) = ZERO
      if (allocated(P%nat_trans_orb)) then
        allocate (tdens_csf(size(initial_occ, 1), size(final_occ, 1), size(final_csf_array, 1)))
        tdens_csf(:, :, :) = ZERO
      end if
      if(P%yes_verbose) &
        write(P%OUTPUT, "('# computing the tdipoles')")
      if(UTfin%NopenMax>UTini%NopenMax)then
        UTjoint => UTfin
      else
        UTjoint => UTini
      end if
      csf_tdipole(:, :, :) = ZERO
      ic = 0
      do i = 1, size(initial_occ,2)
        if(UTjoint%f(count(initial_occ(:,i)==1))==0) cycle
        fc = 0
        do j = 1, size(final_occ,2)
          if(UTjoint%f(count(final_occ(:,j)==1))==0) cycle
          if(sum(abs(initial_occ(:,i) - final_occ(:,j))) <= 2) then
            if (count(initial_occ(:,i)==1) >= count(final_occ(:,j)==1))then
              allocate(dm(size(initial_occ,1),size(initial_occ,2), &
                UTjoint%f(count(initial_occ(:,i)==1)), &
                UTjoint%f(count(final_occ(:,j)==1))))
              call opdm_matrix_subblock(initial_occ(:,i), final_occ(:,j), final_multp, UTjoint, dm)
              do k=1, UTjoint%f(count(initial_occ(:,i)==1))
                do l=1, UTjoint%f(count(final_occ(:,j)==1))
                  if (allocated(P%nat_trans_orb)) &
                    tdens_csf(:, :, fc+l) = tdens_csf(:, :, fc+l) + initial_CI_matrix(ic+k, P%state) * dm(k,l,:,:)
                  do m = 1, size(dm, 3)
                    do n = 1, size(dm, 4)
                      csf_tdipole(ic+k, fc+l, :) = csf_tdipole(ic+k, fc+l, :) + dm(k,l,m,n) * tdipole_mo(m, n, :)
                    end do! n
                  end do! m
                end do! l
              end do! k
            else ! swap initial and final
              allocate(dm(size(initial_occ,1),size(initial_occ,2), &
                num_csf(count(final_occ(:,j)==1), final_multp), &
                num_csf(count(initial_occ(:,i)==1), initial_multp)))
              call opdm_matrix_subblock(final_occ(:,j), initial_occ(:,i), final_multp, UTjoint, dm)
              do k=1, num_csf(count(initial_occ(:,i)==1), initial_multp)
                do l=1, num_csf(count(final_occ(:,j)==1), final_multp)
                  if (allocated(P%nat_trans_orb)) &
                    tdens_csf(:, :, fc+l) = tdens_csf(:, :, fc+l) + initial_CI_matrix(ic+k, P%state) * transpose(dm(l,k,:,:))
                  do m = 1, size(dm, 3)
                    do n = 1, size(dm, 4)
                      csf_tdipole(ic+k, fc+l, :) = csf_tdipole(ic+k, fc+l, :) + dm(l,k,n,m) * tdipole_mo(m, n, :)
                    end do! n
                  end do! m
                end do! l
              end do! k
            end if
            deallocate(dm)
          end if
          fc = fc + UTjoint%f(count(final_occ(:,j)==1))
        end do! j
        ic = ic + UTjoint%f(count(initial_occ(:,i)==1))
      end do! i
      nullify(UTjoint)

      if(P%yes_verbose) &
        write(P%OUTPUT, "('# done')")
      if (P%yes_process_CI_relaxed) then
        if (P%process_CI_relaxed_variant .ne. 2 .and. P%process_CI_relaxed_variant .ne. 1) then
          write (P%OUTPUT, '("#",a)') 'process_CI_relaxed_variant', P%process_CI_relaxed_variant
          stop 'process_CI_relaxed_variant must be 1 or 2'
        end if
        F_final = F
        ! it's a really bad idea to copy B!
        !B_final=B
        F_final%LC_ini = F%LC
        call read_wf(F_final, F_final%LC, "wf_relaxed.dat")

        if (P%process_CI_relaxed_variant .eq. 2 .or. allocated(P%nat_trans_orb)) then
          allocate (csf_O(size(final_csf_array), size(final_csf_array)))
          call overlap_csfs(P, final_csf_array, F%LC, F_final%LC, B%S, csf_O)
          if (allocated(P%nat_trans_orb)) then
            allocate (tdens_csf2(size(initial_occ, 1), size(final_occ, 1), size(final_csf_array, 1)))
            tdens_csf2 = ZERO
            do j = 1, N_final_csfs
              do i = 1, N_final_csfs
                tdens_csf2(:, :, j) = tdens_csf2(:, :, j) + tdens_csf(:, :, i)*csf_O(i, j)
              end do
            end do
            tdens_csf = tdens_csf2
            deallocate (tdens_csf2)
          end if
          ! project transition dipole into new basis
          allocate (temp(N_initial_csfs, N_final_csfs))
          do pol = 1, 3
            do i = 1, N_initial_csfs
              do j = 1, N_final_csfs
                temp(i, j) = sum(csf_tdipole(i, :, pol)*csf_O(:, j))
              end do
            end do
            csf_tdipole(:, :, pol) = temp
          end do
          deallocate (temp)
          write (P%OUTPUT, '(a)') '# Overlap to new basis:'
          do fc = 1, size(final_csf_array, 1)
            write (P%OUTPUT, '("# csf ",i7,x,a,x,"tdip=",es20.7,x,f7.3,f7.3)') fc, &
              trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), &
              sum(csf_tdipole(P%state, fc, :)**2), &
              sum(csf_O(:, fc)**2), sum(csf_O(fc, :)**2)
          end do
          deallocate (csf_O)
        else if (P%process_CI_relaxed_variant .eq. 1) then
          csf_tdipole = ZERO
          call overlap_operator_csfs(P, initial_csf_array(1:N_initial_csfs), &
               final_csf_array(1:N_final_csfs),&
               F%LC,&
               F_final%LC, B%S, tdipole_bf(:, :, :), &
               csf_tdipole(:, :, :))
        end if ! process_ci_relaxed_)variant
        call construct_h0_matrix(G, B, F_final)
        F_final%H0_MO = F_final%LC(:, :) .tm.F_final%H0(:, :) .m.F_final%LC(:, :)
      else
        F_final = F
        !B_final=B
      end if!
      deallocate (active, core)
      allocate (active(count((/((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. &
        any(final_occ(j, 1:N_final_occ) .gt. 0) ), j=1, size(final_occ, 1))/))))
      allocate (core(count((/((all(final_occ(j, 1:N_final_occ) .eq. 2)), j=1, size(final_occ, 1))/))))
      active = pack((/(j, j=1, size(final_occ, 1))/), &
        (/(((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0))), &
        j=1, size(final_occ, 1))/))
      core = pack((/(j, j=1, size(final_occ, 1))/), &
        (/(all(final_occ(j, 1:N_final_occ) .eq. 2), j=1, size(final_occ, 1))/))
      nintern = size(active)
      do j = size(active), 1, -1
        if (any(sum(final_occ(active(j:), :N_final_occ), dim=1) .gt. 1)) exit
      end do
      nintern = j
      write (P%OUTPUT, '("#     core:",99i4)') core
      write (P%OUTPUT, '("# fin_hole:",99i4)') final_hole
      write (P%OUTPUT, '("#   active:",99i4)') active
      write (P%OUTPUT, '("# internal:",99i4)') active(:nintern)
      write (P%OUTPUT, '("# external:",99i4)') active(nintern + 1:)
      ! number of states to consider
      if (P%nstates .gt. 0) then
        nev = P%nstates
      else
        nev = N_final_csfs
      end if
      nev = max(nev, 1)
      nev = min(nev, N_final_csfs)
      if (P%yes_direct_scf) then
        call calculate_F_inactive_direct(B, core, F_final%LC, F_core)
      else
        call calculate_F_inactive(P, B, core, F_final%LC, F_core)
      end if
      E_core = ZERO
      do i = 1, size(core)
        E_core = E_core + F_core(core(i), core(i)) + F_final%H0_MO(core(i), core(i))
      end do
      call transform_twoel_allactive(P, B, F_final, active(:nintern), P%OUTPUT)
      if (nintern .lt. size(active)) then
        call transform_twoel_two_core_two_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
        call transform_twoel_three_core_one_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
      end if
      if (P%yes_use_arpack .or. P%yes_use_davidson .or. P%yes_use_krylov) then
        allocate (CI_diagonal(N_final_csfs))      
        write (P%OUTPUT, '("# calculating CI diagonal entries")')
        call calc_CIMatrix_diagonal(UTfin, E_core, F_core(active(:), active(:)), &
          F_final%teint_abcd, &
          final_occ(active(:), 1:N_final_occ), &
          CI_diagonal, nintern=nintern, &
          teint_iiee=F_final%teint_cdij, &
          teint_ieie=F_final%teint_cidj, &
          teint_iiie=F_final%teint_cdei)
        write (P%OUTPUT, '("# done")')
        allocate (csf_c2(N_final_csfs, nev))
        csf_c2 = ZERO
        ! take the configuration with lowest diagonal element from those entries that are in the primary space
        allocate (idx(size(CI_diagonal)))
        call merge_argsort(CI_diagonal, idx)
        j = 0
        do while (j .lt. nev)
          j = j + 1
          csf_c2(idx(j), j) = ONE
          if (j .eq. nev) exit
        end do
        deallocate (idx)
        ! print summary
        do i = 1, nev
          do j = 1, size(csf_c2, 1)
            if (csf_c2(j, i) .eq. ZERO) cycle
            write (P%OUTPUT, '("# Starting guess: ",i4,x,es20.7)', advance="yes") i, ci_diagonal(j)
            write (P%OUTPUT, '("# ",es20.7,x,a)', advance="yes") &
              csf_c2(j, i), trim(csfstring(final_csf_array(j)%occ, final_csf_array(j)%csf(:)))
          end do
        end do        
        if (P%yes_use_arpack) then
          ! call arpack
#ifdef WITH_ARPACK
          call get_some_eigenvalues_arpack(N_final_csfs, nev, final_CI_energy(1:nev), csf_c2, matrix_vector_multp)
#else
          stop "not compiled with arpack support"
#endif
        else if (P%yes_use_davidson) then
          ! call Davidson
          ! call get_some_eigenvalues_VariableBlockDavidson(N_final_csfs, nev, P%davblock, max(P%davMax,nev), final_CI_energy(1:nev), csf_c2,  ci_diagonal, matrix_matrix_multp, P%EPS_Auger, P%OUTPUT)
          call get_some_eigenvalues_davpack(N_final_csfs, nev, min(P%davblock, nev), &
            min(max(P%davMax, nev),N_final_csfs), final_CI_energy(1:nev), &
            csf_c2, ci_diagonal, matrix_matrix_multp, &
            P%eps_teint, P%OUTPUT)
        else if (P%yes_use_krylov) then
#ifdef WITH_KRYLOV
          call get_some_eigenvalues_libKrylov(N_final_csfs, nev, nev, &
            final_CI_energy(1:nev), &
            csf_c2, ci_diagonal, matrix_matrix_multp)
#else
          stop "not compiled with libkrylov support"
#endif
        end if
        deallocate (CI_diagonal)
        deallocate (F_final%teint_abcd)
        if (nintern .lt. size(active)) then
          if(allocated(F_final%teint_cdij))&
            deallocate (F_final%teint_cdij)
          if(allocated(F_final%teint_cidj))&
            deallocate (F_final%teint_cidj)
          if(allocated(F_final%teint_cdei))&
            deallocate (F_final%teint_cdei)
        end if
      else
        allocate (final_CI_matrix(N_final_csfs, N_final_csfs))
        write (P%OUTPUT, '("# calculating CI matrix")')
        if (nintern .lt. size(active)) then
          call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
            F_final%teint_abcd, &
            final_occ(active(:), 1:N_final_occ), &
            final_CI_matrix, nintern=nintern, &
            teint_iiee=F_final%teint_cdij, &
            teint_ieie=F_final%teint_cidj, &
            teint_iiie=F_final%teint_cdei)
          if(allocated(F_final%teint_cdij))&
            deallocate (F_final%teint_cdij)
          if(allocated(F_final%teint_cidj))&
            deallocate (F_final%teint_cidj)
          if(allocated(F_final%teint_cdei))&
            deallocate (F_final%teint_cdei)
        else
          call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
          F_final%teint_abcd, &
          final_occ(active(:), 1:N_final_occ), &
          final_CI_matrix)
        end if
        deallocate (F_final%teint_abcd)
        write (P%OUTPUT, '("# done")')          
        ! solve it
        write (P%OUTPUT, '("# solving CI matrix")')
        allocate (csf_c2(N_final_csfs, nev))
        call get_some_ev_of_real_sym(final_CI_matrix, final_CI_energy, csf_c2, nev, P%EPS_teint, yes_verbose=.False.)
        write (P%OUTPUT, '("# done")')
        deallocate (final_CI_matrix)
      end if
      ! transfrom the transition dipole from csf space to CI-state space
      allocate (temp2(N_final_csfs))
      allocate(state_tdipole(nev, 3))
      do pol = 1, 3
        temp2 = csf_tdipole(:, :, pol) .tm. initial_CI_matrix(:, P%state)
        state_tdipole(:, pol) = csf_c2(:, :) .tm. temp2
      end do
      deallocate(temp2)
      deallocate(csf_tdipole)
      do i = 1, nev
        abs_energy = final_CI_energy(i) - E_ini
        ChannelOS = oscillator_strength(state_tdipole(i,:), abs_energy)
        TotalOS = TotalOS + ChannelOS
        write (P%OUTPUT, '("# ",i7," E=",f20.5," a.u."," Abs Energy=",f15.5," eV")') &
          i, final_CI_energy(i) + nuclear_repulsion_energy(A, P), &
          AU2EV*abs_energy
        do fc = 1, size(final_csf_array, 1)
          if (abs(csf_c2(fc, i)) .gt. 1e-1) &
            write (P%OUTPUT, '("# ",7x,f10.5,x,i7,x,a)', advance="yes") &
            csf_c2(fc, i), fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf))
        end do! fc/
        if (allocated(P%nat_trans_orb)) then
          if (any(P%nat_trans_orb(:) .eq. i)) then
            ! make natural transition orbitals from transition density
            allocate (tdens(size(initial_occ, 1), size(final_occ, 1)))
            tdens(:, :) = ZERO
            do fc = 1, size(final_csf_array, 1)
              tdens(:, :) = tdens(:, :) + tdens_csf(:, :, fc) * csf_c2(fc, i)
            end do
            allocate (LC_hole(size(tdens, 1)))
            allocate (LC_part(size(tdens, 2)))
            tdens = tdens/sqrt(2.)
            call make_nat_trans_orbitals(P%OUTPUT, tdens, LC_hole, LC_part)
            ! nat. trans density expressed in initial state orbital basis
            if (P%molden_file .ne. "") then
              F_temp = F_final
              F_temp%LC = ZERO
              F_temp%LC(:, 1) = F%LC(:, :) .m. LC_hole(:)
              F_temp%LC(:, 2) = F%LC(:, :) .m. LC_part(:)
              if(P%yes_sym)then
                write (moldenfilename, '("trans_",a,"_",i0.4,".molden")') trim(B%irrep_name(f_irrep)), i
              else
                write (moldenfilename, '("trans_",i0.4,".molden")') i
              end if
              call write_molden_file(P, A, B, F_temp, moldenfilename)
            end if
            ! population analysis of the hole/ particle density
            allocate (densMatrix_hole(size(F%LC, 1), size(F%LC, 1))) ! tdens_tmp is in AO basis
            densMatrix_hole = F%LC .m. (tdens .mt. tdens) .mt. F%LC
            allocate (densMatrix_part(size(F%LC, 1), size(F%LC, 1))) ! tdens_tmp is in AO basis
            densMatrix_part = F%LC .m. (tdens .tm. tdens) .mt. F%LC
            write(P%OUTPUT,'("#        Mulliken Pop. analysis:")')
            write(P%OUTPUT,'("# ",a13,x)', advance='no') '      Atom'
            do j=1,size(A,1)
              if(A(j)%charge == 0) cycle
              write(P%OUTPUT,'(a7)', advance='no') A(j)%label
            end do
            write(P%OUTPUT,'()')
            call mulliken_population(A, B, F, pop,  densMatrix_hole)
            write(P%OUTPUT,'("# ",a13,x)', advance='no') '      Hole'
            do j=1,size(A,1)
              if(A(j)%charge == 0) cycle
              write(P%OUTPUT, '(f7.2)',  advance='no') pop(j)
            end do
            write(P%OUTPUT,'()')
            call mulliken_population(A, B, F, pop,  densMatrix_part)
            write(P%OUTPUT,'("# ",a13,x)', advance='no') '      Part'
            do j=1,size(A,1)
              if(A(j)%charge == 0) cycle
              write(P%OUTPUT, '(f7.2)',  advance='no') pop(j)
            end do
            write(P%OUTPUT,'()')
            write(P%OUTPUT,'("#        Loewdin Pop. analysis:")')
            call loewdin_population(A, B, F, pop,  densMatrix_hole)
            write(P%OUTPUT,'("# ",a13,x)', advance='no') '      Hole'
            do j=1,size(A,1)
              if(A(j)%charge == 0) cycle
              write(P%OUTPUT, '(f7.2)',  advance='no') pop(j)
            end do
            write(P%OUTPUT,'()')
            call loewdin_population(A, B, F, pop,  densMatrix_part)
            write(P%OUTPUT,'("# ",a13,x)', advance='no') '      Part'
            do j=1,size(A,1)
              if(A(j)%charge == 0) cycle
              write(P%OUTPUT, '(f7.2)',  advance='no') pop(j)
            end do
            write(P%OUTPUT,'()')
            call density_analysis(P, G, B, densMatrix_hole, P%OUTPUT, description='       Hole density:')
            call density_analysis(P, G, B, densMatrix_part, P%OUTPUT, description='       Part density:')
            deallocate (densMatrix_part)
            deallocate (densMatrix_hole)
            deallocate (tdens)
            deallocate (lc_hole)
            deallocate (lc_part)
            !! make natural orbitals
            allocate (opdm(size(F%LC, 2), size(F%LC, 2)))
            allocate (Dmatrix(size(F%LC, 1), size(F%LC, 1)))
            allocate (n_occ(size(F%LC, 2)))
            opdm = ZERO
            call one_particle_dm_CI2(UTfin, final_occ(:,1:N_final_occ), csf_c2(:, i), opdm)
            F_final%D = F_final%LC .m. opdm .mt. F_final%LC
            opdm = -opdm
            call get_eigenvalues(opdm, n_occ)
            n_occ = -n_occ
            call reorder_RSTRCT(opdm, n_occ)
            if (P%molden_file .ne. "") then
              F_temp%LC = F_final%LC.m.opdm
              F_temp%f_occ = n_occ
              write (moldenfilename, '("final_",i0.4,".molden")') i
              call write_molden_file(P, A, B, F_temp, moldenfilename)
            end if
            deallocate (Dmatrix)
            deallocate (opdm)
            deallocate (n_occ)
          end if
        end if ! nat_trans_orb
        write (P%OUTPUT, "('# ',7x,'channel energy/rate:' ,x,es13.5,es13.5,3es13.5)") &
          AU2EV*abs_energy, ChannelOS, state_tdipole(i,:)
      end do! i loop over final states

      if (P%print_level .gt. 1) then
        if(P%yes_absorptioncs) then
          write (OUTPUT, "('# Cross sections assuming photon energy', f13.2, x, 'eV')") P%PE * AU2EV
          write (OUTPUT, "('#                     bandwidth (stdev)', f13.2, x, 'eV')") P%PE_width * AU2EV
          write (OUTPUT, "('#                     bandwidth (FWHM) ', f13.2, x, 'eV')") &
            P%PE_width *  TWO*sqrt(TWO*log(TWO)) * AU2EV
        end if
        if (P%yes_sym) then
          write (OUTPUT, "('#'/'# Absorption to bound states (based on CI states) irrep ',i4,x,a,':')") &
            f_irrep, trim(B%irrep_name(f_irrep))
        else
          write (OUTPUT, "('#'/'# Absorption to bound states (based on CI states) ',':')")
        end if
        if(P%yes_absorptioncs) then
          write (OUTPUT, "('#    I     F       ',a13,2x,a13,8x,3a13,x,a13)") &
            "x-ray E(eV)", "oscill. strength(a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z","cs (a.u.)"
        else
          write (OUTPUT, "('#    I     F       ',a13,2x,a13,8x,3a13)") &
            "x-ray E(eV)", "oscill. strength(a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
        end if
      end if

      do i = 1, nev
        abs_energy = final_CI_energy(i) - E_ini
        os = oscillator_strength(state_tdipole(i,:), abs_energy)
        if(P%yes_absorptioncs) then
          write (OUTPUT, "('#    ',i3,' - 'i8,' : ',f13.2,2x,es13.5,6x,2x,4es13.5,x,es13.5)") &
            P%state, i, abs_energy*AU2EV, os, &
            state_tdipole(i,:), TWO * PI**2 / SPEED_C_AU * oS * gauss_profile(abs_energy - P%PE, P%PE_width)
        else
          write (OUTPUT, "('#    ',i3,' - 'i8,' : ',f13.2,2x,es13.5,6x,2x,4es13.5)") &
            P%state, i, abs_energy*AU2EV, os, &
            state_tdipole(i,:)
        end if
      end do
      ! clean up arrays for this final spin multp:
      call purge_csf_array(final_csf_array)
      deallocate (csf_c2)
      deallocate (final_CI_energy)
      deallocate (state_tdipole)
      deallocate (final_occ)
      if (allocated(P%nat_trans_orb)) &
           deallocate(tdens_csf)

    end do ! f_irrep
    deallocate (final_hole)
    call purge_csf_array(initial_csf_array)
    call purge_UMatrixTable(UTini)
    call purge_UMatrixTable(UTfin)
    deallocate (initial_occ)
    deallocate (initial_occ_orig)
    if (P%print_level .gt. 1) then
      if (TotalOS .gt. ZERO) then
        write (OUTPUT, "('# ',12x,'Total oscillator strength = ',es13.5,' a.u.')") TotalOS

      else
        write (OUTPUT, "('#   NONE')")
      end if
    end if
  contains

    function matrix_matrix_multp(n, ncol, Cin, Cout) result(error)
      integer, intent(in) :: n, ncol
      real(long), intent(in) :: Cin(n, ncol)
      real(long), intent(out) :: Cout(n, ncol)
      integer :: error
      call calc_CIMatrix_matrix_multp(UTfin, E_core, F_core(active(:), active(:)), &
        F_final%teint_abcd, &
        final_occ(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F_final%teint_cdij, &
        teint_ieie=F_final%teint_cidj, &
        teint_iiie=F_final%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_matrix_multp
  end subroutine calc_oscillator_strength_CI

!>------------------------------------------------------------------------------
!! function oscillator_strength multiplies the transition dipole vector with prefactors
!! \[ \frac{2}{3} \Delta E |\mathbf{d}|^2 \]
!! where \(d\) is the transition dipole, and \(\Delta E\) is energy_transition
!!
  function oscillator_strength(tdipole, energy_transition) result(os)
    real(long), intent(in) :: tdipole(3) !! the transition dipole
    real(long), intent(in) :: energy_transition !! transition energy
    real(long) :: os !! absorption cross section (return value)
    os = TWO/THREE * energy_transition * sum(tdipole(:)**2)
    if (os .lt. 1e-15_long) then
      os = 0.0
    end if
  end function oscillator_strength

  function gauss_profile(x, stdev) result(g)
    real(long), intent(in) :: x !! x
    real(long), intent(in) :: stdev !! stdev
    real(long) :: g !! gauss
    g = exp( - HALF * x**2 / stdev**2 ) / stdev / sqrt( TWO * PI )
  end function gauss_profile

end module absorption
