!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module anglib
  !! author: Paul Stevenson, Oxford University/Oak Ridge National Laboratory.
  !! author: Robin Santra
  !!
  !! Library of angular momentum coupling coefficient routines in fortran 90
  !! Paul Stevenson, Oxford University/Oak Ridge National Laboratory.
  !! spaul@mail.phy.ornl.gov
  !! Modified by Robin Santra on 10/10/02 from single to double precision.
  !! Paul D. Stevenson's page (as of August 2023):
  !! https://openresearch.surrey.ac.uk/esploro/profile/paul_denis_stevenson/
  !! ORCiD: 0000-0003-2645-2569
  !! p.stevenson@surrey.ac.uk
  !! 
  !! Further information
  !! - Cleblet: Java implementation
  !!   * P. D. Stevenson, "Analytic angular momentum coupling coefficient 
  !!     calculators," Computer Physics Communications 147(3), 853-858 (2002)
  !!     doi: https://doi.org/10.1016/S0010-4655(02)00462-9
  !!   * source code: Elsevier Data Repository, doi: 10.17632/6v8ggwt3c6.1
  !! - anglib_improved.f90: Fortran implementation modified by Emanuele Ballarin
  !!   https://gist.github.com/emaballarin/c7a9dd0eba68f0ebb2f7f546ba1b93ac
  !!
  !! Here is the license statement from anglib_improved.f90 (LGPLv3):
  !!  anglib.f90: Angular momentum coupling coefficients for Q.M., in Fortran 9x
  !!  Copyright (C) 1998  Paul Stevenson     (idea, mainline code)
  !!                2018  Emanuele Ballarin  (optimizations, porting)
  !!
  !!  This library is free software. You can redistribute it and/or modify it
  !!  under the terms of the GNU Lesser General Public License as published by
  !!  the Free Software Foundation, either version 3, or (at your option) any
  !!  later version.
  !!
  !!  This library is distributed in the hope that it could be useful, but WITHOUT
  !!  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  !!  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
  !!  for more details.
  !!
  !!  You should have received a copy of the GNU Lesser General Public License
  !!  along with this library. If not, write to the Free Software Foundation,
  !!  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
  !!----------------------------------------------------------------------------
contains  
  pure function cleb(j1, m1, j2, m2, j, m)
    !! calculate a clebsch-gordan coefficient \( \langle \frac{j_1}{2}, \frac{m_1}{2}, \frac{j_2}{2}, \frac{m_2}{2} | \frac{j}{2}, \frac{m}{2} \rangle \)
    !! arguments are integer and twice the true value.  
    implicit none
    real*8    :: cleb
    real*8    :: factor, sum
    integer, intent(in) :: j1, m1, j2, m2, j, m
    integer :: par, z, zmin, zmax

    ! some checks for validity (let's just return zero for bogus arguments)

    if (2*(j1/2) - int(2*(j1/2.0)) .ne. 2*(abs(m1)/2) - int(2*(abs(m1)/2.0)) .or. &
        2*(j2/2) - int(2*(j2/2.0)) .ne. 2*(abs(m2)/2) - int(2*(abs(m2)/2.0)) .or. &
        2*(j/2) - int(2*(j/2.0)) .ne. 2*(abs(m)/2) - int(2*(abs(m)/2.0)) .or. &
        j1 .lt. 0 .or. &
        j2 .lt. 0 .or. &
        j .lt. 0 .or. &
        abs(m1) .gt. j1 .or. &
        abs(m2) .gt. j2 .or. &
        abs(m) .gt. j .or. &
        j1 + j2 .lt. j .or. &
        abs(j1 - j2) .gt. j .or. &
        m1 + m2 .ne. m) then
      cleb = 0.0d0
    else

      factor = 0.0d0
      factor = binom(j1, (j1 + j2 - j)/2)/binom((j1 + j2 + j + 2)/2, (j1 + j2 - j)/2)
      factor = factor*binom(j2, (j1 + j2 - j)/2)/binom(j1, (j1 - m1)/2)
      factor = factor/binom(j2, (j2 - m2)/2)/binom(j, (j - m)/2)
      factor = dsqrt(factor)

      zmin = max(0, j2 + (j1 - m1)/2 - (j1 + j2 + j)/2, j1 + (j2 + m2)/2 - (j1 + j2 + j)/2)
      zmax = min((j1 + j2 - j)/2, (j1 - m1)/2, (j2 + m2)/2)

      sum = 0.0d0
      do z = zmin, zmax
        par = 1
        if (2*(z/2) - int(2*(z/2.0)) .ne. 0) par = -1
        sum = sum + par*binom((j1 + j2 - j)/2, z)*binom((j1 - j2 + j)/2, (j1 - m1)/2 - z)* &
              binom((-j1 + j2 + j)/2, (j2 + m2)/2 - z)
      end do

      cleb = factor*sum
    end if
  end function cleb

  pure function sixj(a, b, c, d, e, f)
    !! calculates a Wigner 6-j symbol. Argument a-f are integer and are
    !! twice the true value of the 6-j's arguments, in the form
    !! { a b c }
    !! { d e f }  
    implicit none
    integer, intent(in) :: a, b, c, d, e, f
    real*8 :: sixj
    integer :: phase, nlo, nhi, n
    real*8 :: outfactors, sum, sumterm


    ! First check for consistency of arguments:
    sixj = 0.0d0
    if (mod(a + b, 2) .ne. mod(c, 2)) return
    if (mod(c + d, 2) .ne. mod(e, 2)) return
    if (mod(a + e, 2) .ne. mod(f, 2)) return
    if (mod(b + d, 2) .ne. mod(f, 2)) return
    if (abs(a - b) .gt. c .or. a + b .lt. c) return
    if (abs(c - d) .gt. e .or. c + d .lt. e) return
    if (abs(a - e) .gt. f .or. a + e .lt. f) return
    if (abs(b - d) .gt. f .or. b + d .lt. f) return

    phase = (-1)**((a + c + d + f)/2)

    outfactors = angdelta(a, b, c)/angdelta(a, e, f)
    outfactors = outfactors*angdelta(b, d, f)/angdelta(c, d, e)

    nlo = 0
    nhi = min((a - b + d + e)/2, (c + e + f - b)/2, (a + c + d + f)/2 + 1, (a - b + c)/2, &
         & (d + f - b)/2, (a + e + f)/2 + 1, (c + d + e)/2 + 1)

    sum = 0.0d0
    do n = nlo, nhi
      sumterm = (-1)**n
      sumterm = sumterm*factorial((a - b + d + e)/2 - n)/factorial(n)/ &
                factorial((a - b + c)/2 - n)
      sumterm = sumterm*factorial((-b + c + e + f)/2 - n)/factorial((d - b + f)/2 - n)
      sumterm = sumterm*factorial((a + c + d + f)/2 + 1 - n)/factorial((a + e + f)/2 + 1 - n)
      sumterm = sumterm/factorial((c + d + e)/2 + 1 - n)
      sum = sum + sumterm
    end do

    sixj = phase*sum*outfactors

  end function sixj

  pure function angdelta(a, b, c)
    !! calculate the function delta as defined in varshalovich et al. for
    !! use in 6-j symbol:
    implicit none
    integer, intent(in) :: a, b, c
    real*8 :: angdelta, scr1
    scr1 = factorial((a + b - c)/2)
    scr1 = scr1/factorial((a + b + c)/2 + 1)
    scr1 = scr1*factorial((a - b + c)/2)
    scr1 = scr1*factorial((-a + b + c)/2)
    angdelta = dsqrt(scr1)
  end function angdelta

  pure function ninej(a, b, c, d, e, f, g, h, i)
    !! calculates a 9-j symbol. The arguments are given as integers twice the
    !! value of the true arguments in the form
    !! { a b c }
    !! { d e f }
    !! { g h i }    
    implicit none
    integer, intent(in) :: a, b, c, d, e, f, g, h, i
    real*8 :: ninej, sum
    integer :: xlo, xhi
    integer :: x
    ninej = 0.0d0
    ! first check for bogus arguments (and return zero if so)
    if (abs(a - b) .gt. c .or. a + b .lt. c) return
    if (abs(d - e) .gt. f .or. d + e .lt. f) return
    if (abs(g - h) .gt. i .or. g + h .lt. i) return
    if (abs(a - d) .gt. g .or. a + d .lt. g) return
    if (abs(b - e) .gt. h .or. b + e .lt. h) return
    if (abs(c - f) .gt. i .or. c + f .lt. i) return

    xlo = max(abs(b - f), abs(a - i), abs(h - d))
    xhi = min(b + f, a + i, h + d)

    sum = 0.0d0
    do x = xlo, xhi, 2
      sum = sum + (-1)**x*(x + 1)*sixj(a, b, c, f, i, x)*sixj(d, e, f, b, x, h)* &
            sixj(g, h, i, x, a, d)
    end do
    ninej = sum

  end function ninej

  recursive pure function factorial(n) result(res)
    !! calculated factorial
    implicit none
    integer, intent(in) :: n !! integer number
    real*8 :: res !! result = n!

    if (n .eq. 0 .or. n .eq. 1) then
      res = 1.0d0
    else
      res = dble(n*factorial(n - 1))
    end if
  end function factorial

  recursive pure function binom(n, r) result(res)
    !! calculates binomial
    implicit none
    integer, intent(in) :: n, r
    real*8 :: res

    if (n .eq. r .or. r .eq. 0) then
      res = 1.0d0
    else if (r .eq. 1) then
      res = dble(n)
    else
      res = dble(n)/dble(n - r) * binom(n - 1, r)
    end if
  end function binom
end module anglib
