!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module angular
  !! author: Sang-Kil Son
  !!
  !! Angular 2D Grid module
  !!   written by Son, Sang-Kil in May 2007
  !!
  !! This molule is a simplified version from lebedev.f90 and womersley.f90.
  !! It generates angular 2D grids only, and doesn't have any angular GPS formula.
  !!
  !! USAGE:
  !!   type(Angular_Grid) :: G
  !!   G = generate_Angular_Grid( "Lebedev", lmax )
  !!   G = generate_Angular_Grid( "Womersley", lmax )
  !!     where lmax is the maximum angular momentum
  !!   G = generate_Angular_Grid( "icosahedral", N )
  !!     where N is the number of grids
  !!
  !! ENVIRONMENT VARIABLES:
  !! - VFD_WOMERSLEY_DIR: directory for Womersley grid data
  !!
  !! Lebedev Ref)
  !!      Lebedev, Comput. Math. Math. Phys. 15, 44 (1975)
  !!               Zh. Vychisl. Mat. Mat. Fiz. 15, 48 (1975) - Russian
  !!      Lebedev, Comput. Math. Math. Phys. 16, 10 (1976)
  !!               Zh. Vychisl. Mat. Mat. Fiz. 16, 293 (1976) - Russian
  !!      Lebedev & Laikov, Dokl. Math. 59, 477 (1999)
  !!      Wang & Carrington, J. Theoret. Comput. Chem. 2, 599 (2003):
  !!           comparison between Lebedev and Gauss grids
  !!      Becke, JCP 88, 2547 (1988): introduced Lebedev grids into quant. chem.
  !! Womersley Ref)
  !!      Sloan & Womersley, Adv. Comput. Math. 21, 102 (2004)
  !!      Womersley & Sloan, Adv. Comput. Math. 14, 195 (2001)
  !!      http://web.maths.unsw.edu.au/~rsw/Sphere/
  !! Icosahedral Ref)
  !!      http://en.wikipedia.org/wiki/Icosahedral_symmetry
  !!
  !!----------------------------------------------------------------------------
  use constants
  implicit none
  type Angular_Grid
    !! angular grid information
    integer :: N, lmax, l_trunc, grid_type
    real(long), allocatable :: x(:), y(:), z(:), weight(:)
    real(long) :: theta, phi
  end type Angular_Grid
  private
  public :: Angular_Grid, construct_Angular_Grid, purge_Angular_Grid
contains

  subroutine construct_Angular_Grid(grid_type, lmax, G, l_tr)
    !! angular grid
    character(len=*), intent(in) :: grid_type
    integer, intent(in) :: lmax
    type(Angular_Grid), intent(out) :: G
    integer, intent(in), optional :: l_tr
    integer, parameter :: N_table(65) = & ! information N and lmax for Lebedev
      (/6, 14, 26, 38, 50, 74, 86, 110, 146, 170, &
      194, 230, 266, 302, 350, 0, 434, 0, 0, 590, &
      0, 0, 770, 0, 0, 974, 0, 0, 1202, 0, &
      0, 1454, 0, 0, 1730, 0, 0, 2030, 0, 0, &
      2354, 0, 0, 2702, 0, 0, 3074, 0, 0, 3470, &
      0, 0, 3890, 0, 0, 4334, 0, 0, 4802, 0, &
      0, 5294, 0, 0, 5810/)
    real(long) :: norm
    integer :: N, l_trunc, i
    integer :: WOMERSLEY_DATA = 8
    character(len=80) :: filename, dir_womersley
    l_trunc = 0

    !
    ! 1) determine a grid_type
    !
    select case (grid_type)
     case ("Lebedev", "lebedev", "Leb", "leb", "L", "LBD"); G%grid_type = 0
     case ("Womersley", "womersley", "Wom", "wom", "W", "WM"); G%grid_type = 1
     case ("Icosahedral", "icosahedral", "ico", "Ico", "I", "Ih", "IH", &
       "Fullerene", "fullerene", "C60"); G%grid_type = 2
     case default
      write (STDERR, *) &
        "Error in generate_Grid: "//grid_type//" is not supported."
      stop
    end select
    !
    ! 2) set N and l_trunc
    !
    if (G%grid_type .eq. 0) then                  ! ---- for Lebedev grids ----
      if (present(l_tr)) then
        l_trunc = l_tr
        if (l_tr .eq. 0) l_trunc = lmax
      else
        !select case ( lmax )
        !  case ( 4 );               l_trunc = lmax + 2
        !  case ( 5 );               l_trunc = lmax + 3
        !  case ( 6, 7, 8, 9, 10 );  l_trunc = lmax + 4
        !  case ( 13, 14, 15 );      l_trunc = lmax + 5
        !  case default;     stop "Lebedev grids: Please specify l_trunc."
        !end select
      end if
      N = N_table(lmax)
      if (lmax .lt. 1 .or. lmax .gt. size(N_table) .or. N .eq. 0) then
        write (STDERR, "('Lebedev grids: lmax = ',i3,' is not supported.')") lmax
        call print_N_Lebedev_Grid(N_table)
        stop
      end if
    else if (G%grid_type .eq. 1) then             ! ---- for Womersley grids ----
      N = (lmax + 1)**2         ! note that N_SH = N = ( lmax + 1 )^2
      l_trunc = lmax              ! they're exactly the same for Wom. grids
    else if (G%grid_type .eq. 2) then             ! ---- for Fullerene grids ----
      N = lmax                    ! take N from lmax
      if (N .ne. 12 .and. N .ne. 20 .and. N .ne. 60) then
        write (STDERR, "('Icosahedral grids: N = ',i3,' is not supperted.')") N
        stop
      end if
      l_trunc = lmax              ! actually, icosahedral don't depend on lmax
    end if
    !
    ! 3) set parameters and allocate memory
    !
    G%N = N
    G%lmax = lmax
    G%l_trunc = l_trunc
    allocate (G%x(N), G%y(N), G%z(N), G%weight(N))
    !
    ! 4) prepare grid data
    !    - Lebedev grid: call lebedev-laikov.f
    !    - Womersley grid: read from files
    !    - Icosahedral grid: build up according to the symmetry
    !    NOTE: Leb. grid is already normalized, i.e., sum( weight(:) ) = 1
    !          Wom. grid has a factor, i.e., sum( wieght(:) ) = 4 pi
    !            --> it is forced to be normalized at the end!
    !          Ico. grid does not have any well-defined weights --> don't use it!
    !
    if (G%grid_type .eq. 0) then
      !---------------
      ! Lebedev grids
      !---------------
      select case (N)
       case (6); call LD0006(G%x, G%y, G%z, G%weight, N) ! L=3,   lmax=1
       case (14); call LD0014(G%x, G%y, G%z, G%weight, N) ! L=5,   lmax=2
       case (26); call LD0026(G%x, G%y, G%z, G%weight, N) ! L=7,   lmax=3
       case (38); call LD0038(G%x, G%y, G%z, G%weight, N) ! L=9,   lmax=4
       case (50); call LD0050(G%x, G%y, G%z, G%weight, N) ! L=11,  lmax=5
       case (74); call LD0074(G%x, G%y, G%z, G%weight, N) ! L=13,  lmax=6
       case (86); call LD0086(G%x, G%y, G%z, G%weight, N) ! L=15,  lmax=7
       case (110); call LD0110(G%x, G%y, G%z, G%weight, N) ! L=17,  lmax=8
       case (146); call LD0146(G%x, G%y, G%z, G%weight, N) ! L=19,  lmax=9
       case (170); call LD0170(G%x, G%y, G%z, G%weight, N) ! L=21,  lmax=10
       case (194); call LD0194(G%x, G%y, G%z, G%weight, N) ! L=23,  lmax=11
       case (230); call LD0230(G%x, G%y, G%z, G%weight, N) ! L=25,  lmax=12
       case (266); call LD0266(G%x, G%y, G%z, G%weight, N) ! L=27,  lmax=13
       case (302); call LD0302(G%x, G%y, G%z, G%weight, N) ! L=29,  lmax=14
       case (350); call LD0350(G%x, G%y, G%z, G%weight, N) ! L=31,  lmax=15
       case (434); call LD0434(G%x, G%y, G%z, G%weight, N) ! L=35,  lmax=17
       case (590); call LD0590(G%x, G%y, G%z, G%weight, N) ! L=41,  lmax=20
       case (770); call LD0770(G%x, G%y, G%z, G%weight, N) ! L=47,  lmax=23
       case (974); call LD0974(G%x, G%y, G%z, G%weight, N) ! L=53,  lmax=26
       case (1202); call LD1202(G%x, G%y, G%z, G%weight, N) ! L=59,  lmax=29
       case (1454); call LD1454(G%x, G%y, G%z, G%weight, N) ! L=65,  lmax=32
       case (1730); call LD1730(G%x, G%y, G%z, G%weight, N) ! L=71,  lmax=35
       case (2030); call LD2030(G%x, G%y, G%z, G%weight, N) ! L=77,  lmax=38
       case (2354); call LD2354(G%x, G%y, G%z, G%weight, N) ! L=83,  lmax=41
       case (2702); call LD2702(G%x, G%y, G%z, G%weight, N) ! L=89,  lmax=44
       case (3074); call LD3074(G%x, G%y, G%z, G%weight, N) ! L=95,  lmax=47
       case (3470); call LD3470(G%x, G%y, G%z, G%weight, N) ! L=101, lmax=50
       case (3890); call LD3890(G%x, G%y, G%z, G%weight, N) ! L=107, lmax=53
       case (4334); call LD4334(G%x, G%y, G%z, G%weight, N) ! L=113, lmax=56
       case (4802); call LD4802(G%x, G%y, G%z, G%weight, N) ! L=119, lmax=59
       case (5294); call LD5294(G%x, G%y, G%z, G%weight, N) ! L=125, lmax=62
       case (5810); call LD5810(G%x, G%y, G%z, G%weight, N) ! L=131, lmax=65
      end select
    else if (G%grid_type .eq. 1) then
      !-----------------
      ! Womersley grids
      !-----------------
      call getenv("VFD_WOMERSLEY_DIR", dir_womersley)
      if (dir_womersley .eq. "") dir_womersley = "womersley_grid" ! at current dir
      write (filename, "(a,'/md',i2.2,'.',i4.4)") trim(dir_womersley), lmax, N
      open (WOMERSLEY_DATA, file=filename, action="read", form="formatted")
      do i = 1, N
        read (WOMERSLEY_DATA, *) G%x(i), G%y(i), G%z(i), G%weight(i)
      end do
      close (WOMERSLEY_DATA)
      norm = sum(G%weight(:))           ! Wom. grid has a factor of 4*pi
      G%weight(:) = G%weight(:)/norm    ! thus it has to be normalized
    else if (G%grid_type .eq. 2) then
      !-------------------
      ! Icosahedral grids
      ! --> weights are not defined in this grid type, so it cannot be used.
      !-------------------
      stop "The icosahedral type is currently NOT supported."
      !G%x(:) = ZERO;  G%y(:) = ZERO;  G%z(:) = ZERO
      !phi = ( ONE + sqrt( 5.0_long ) ) / TWO
      !if ( N == 12 ) then         ! icosahedron model
      !  call seed_polyhedron( G, ZERO, ONE, phi )
      !  r0 = sqrt( phi + TWO )
      !else if ( N == 20 ) then    ! dodecahedron model
      !  call seed_polyhedron( G, ONE, ONE, ONE )
      !  call seed_polyhedron( G, ZERO, ONE/phi, phi )
      !  r0 = sqrt( THREE )
      !else if ( N == 60 ) then    ! fullerene model (truncated icosahedron)
      !  call seed_polyhedron( G, ZERO, ONE, THREE*phi )
      !  call seed_polyhedron( G, TWO, ONE+TWO*phi, phi )
      !  call seed_polyhedron( G, ONE, TWO+phi, TWO*phi )
      !  r0 = sqrt( 9.0_long * phi + 10.0_long )
      !end if
      !G%x(:) = G%x(:) / r0
      !G%y(:) = G%y(:) / r0
      !G%z(:) = G%z(:) / r0
    end if
    !
    ! find triangle among nearst points with respect to z-axis
    !
!  do i = 1, N
!    if ( G%x(i) == ZERO .and. G%y(i) == ZERO ) exit
!  end do
!  i1 = i        ! set the first point
!  d_min = TWO
!  do i = 1, N
!    if ( i == i1 ) cycle
!    d = distance_Angular_Grid( G, i, i1 )
!    if ( d < d_min ) then
!      d_min = d
!      i2 = i    ! set the second point
!    end if
!  end do
!  d_min = TWO
!  do i = 1, N
!    if ( i == i1 .or. i == i2 ) cycle
!    d = distance_Angular_Grid( G, i, i1 ) + distance_Angular_Grid( G, i, i2 )
!    if ( d < d_min ) then
!      d_min = d
!      i3 = i    ! set the third point
!    end if
!  end do
    !
    ! set theta and phi from the center of gravity of triangle (i1,i2,i3)
    !
!  x = ( G%x(i1) + G%x(i2) + G%x(i3) ) / THREE
!  y = ( G%y(i1) + G%y(i2) + G%y(i3) ) / THREE
!  z = ( G%z(i1) + G%z(i2) + G%z(i3) ) / THREE
!  G%theta = acos( z / sqrt( x**2 + y**2 + z**2 ) )
!  G%phi   = atan( y / x )
  end subroutine construct_Angular_Grid

  subroutine purge_Angular_Grid(G)
    !! purges angular grid
    type(Angular_Grid), intent(inout) :: G
    G%N = 0
    G%lmax = 0
    G%l_trunc = 0
    if(allocated(G%x)) deallocate (G%x)
    if(allocated(G%y)) deallocate (G%y)
    if(allocated(G%z)) deallocate (G%z)
    if(allocated(G%weight)) deallocate (G%weight)
  end subroutine purge_Angular_Grid

  function distance_Angular_Grid(G, i, j) result(d)
    !! @bug unused
    real(long) :: d
    type(Angular_Grid), intent(in) :: G
    integer, intent(in) :: i, j
    d = sqrt((G%x(i) - G%x(j))**2 + (G%y(i) - G%y(j))**2 + (G%z(i) - G%z(j))**2)
  end function distance_Angular_Grid

  subroutine seed_polyhedron(G, x, y, z)
    !! @bug unused
    type(Angular_grid), intent(inout) :: G
    real(long), intent(in) :: x, y, z
    real(long) :: px, py, pz, sx, sy, sz
    integer :: i, j, k, n, N_v
    logical :: yes_rotate
    !
    ! make px < py < pz
    !
    px = min(x, y, z)
    pz = max(x, y, z)
    py = x + y + z - px - pz
    N_v = G%N
    yes_rotate = .true.
    if (px .eq. py .and. py .eq. pz) yes_rotate = .false.
    !
    ! find i such that G(i,:) is not null
    !
    do i = 1, N_v
      if (G%x(i) .eq. ZERO .and. G%y(i) .eq. ZERO .and. G%z(i) .eq. ZERO) exit
    end do
    k = i
    !
    ! generate those points:
    ! (  px,  py,  pz ), (  px,  py, -pz ), (  px, -py,  pz ), (  px, -py, -pz )
    ! ( -px,  py,  pz ), ( -px,  py, -pz ), ( -px, -py,  pz ), ( -px, -py, -pz )
    ! then, rotate x -> y -> z
    !
    n = 8
    if (px .eq. ZERO) n = 4
    if (py .eq. ZERO .and. px .eq. ZERO) n = 2
    do j = 0, n - 1
      sx = px*(-1)**int(j/4)
      sy = py*(-1)**int(j/2)
      sz = pz*(-1)**j
      i = k + j; G%x(i) = sx; G%y(i) = sy; G%z(i) = sz
      if (yes_rotate) then
        i = k + n + j; G%x(i) = sy; G%y(i) = sz; G%z(i) = sx
        i = k + n + n + j; G%x(i) = sz; G%y(i) = sx; G%z(i) = sy
      end if
    end do
  end subroutine seed_polyhedron

!----------------------------------------------------------------------------
! print_N_Lebedev_Grid prints all supported N for Lebedev grids
!   lmax     L     N
!      1     3     6 |    2     5    14 |    3     7    26 |    4     9    38
!      5    11    50 |    6    13    74 |    7    15    86 |    8    17   110
!      9    19   146 |   10    21   170 |   11    23   194 |   12    25   230
!     13    27   266 |   14    29   302 |   15    31   350 |   17    35   434
!     20    41   590 |   23    47   770 |   26    53   974 |   29    59  1202
!     32    65  1454 |   35    71  1730 |   38    77  2030 |   41    83  2354
!     44    89  2702 |   47    95  3074 |   50   101  3470 |   53   107  3890
!     56   113  4334 |   59   119  4802 |   62   125  5294 |   65   131  5810
!----------------------------------------------------------------------------
  subroutine print_N_Lebedev_Grid(N_table)
    integer, intent(in) :: N_table(:)
    integer :: l, k
    write (STDERR, "()")
    write (STDERR, "(a)") "lmax : max. of quantum number l"
    write (STDERR, "(7x,a)") "Lebedev grids can be computed accurately up to lmax"
    write (STDERR, "(a)") "L : order of Lebedev grid"
    write (STDERR, "(a)") "N : number of Lebedev grid points"
    write (STDERR, "()")
    write (STDERR, "(a)") "Supported lmax and N:"
    write (STDERR, "(3a6)") "lmax", "L", "N"
    k = 0
    do l = 1, size(N_table)
      if (N_table(l) .ne. 0) then
        k = k + 1
        write (STDERR, '(3i6)', advance="no") l, 2*l + 1, N_table(l)
        if (mod(k, 4) .eq. 0) then
          write (STDERR, '()')
        else
          write (STDERR, "(' |')", advance="no")
        end if
      end if
    end do
    if (mod(k, 4) .ne. 0) write (STDERR, '()')
  end subroutine print_N_Lebedev_Grid

end module angular
