!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module AngularIntegration
  !! author: Ludger Inhester
  !! module to perform angular integration  
  use constants
  use anglib
  implicit none
  type :: atomorbital
    !! atomic orbital quantum numbers
    integer :: n
    integer :: l
    integer :: m
  end type atomorbital
  private
  public :: atomorbital, realGauntCoef, realAngularIntegral
contains

  !>
  !! This function computes the double angular integral (real spherical harmonics)
  !!
  !!   4pi/(2 l+1) sum_m
  !!    int d Omega Y_lm^* Y_l1m1 Y_l3m3
  !!    int d Omega Y_lm   Y_l2m2 Y_l4m4
  !!
  pure function realAngularIntegral(a_l, a_i1, a_i2, a_i3, a_i4) result(aintegral)
    implicit none
    integer, intent(in) :: a_l !! l quantum number
    type(atomorbital), intent(in) :: a_i1 !! set of l,m quantum numbers
    type(atomorbital), intent(in) :: a_i2 !! set of l,m quantum numbers
    type(atomorbital), intent(in) :: a_i3 !! set of l,m quantum numbers
    type(atomorbital), intent(in) :: a_i4 !! set of l,m quantum numbers
    integer ::m
    real(long) :: aintegral
    aintegral = ZERO
    do m = -a_l, a_l
      aintegral = aintegral + &
                  realGauntCoef(a_l, m, a_i1%l, a_i1%m, a_i3%l, a_i3%m)* &
                  realGauntCoef(a_l, m, a_i2%l, a_i2%m, a_i4%l, a_i4%m)
    end do
    aintegral = aintegral*FOUR*PI/(TWO*a_l + ONE)
    return
  end function realAngularIntegral

  pure function realGauntCoef(a_l, a_m, a_l1, a_m1, a_l2, a_m2) result(rGaunt)
    !! real Gaunt coefficient
    integer, intent(in) :: a_l, a_m, a_l1, a_m1, a_l2, a_m2
    real(long) :: rGaunt
    real(long), parameter :: SQRT_HALF = sqrt(HALF)

    integer :: s, s1, s2 ! signs
    integer :: c, c1, c2 ! condon shortley phase

    complex(long) :: a, a1, a2; ! expansion coefficients
    complex(long) :: cGaunt; 
    if (a_m .lt. 0) then
      a = CMPLX(ZERO, -SQRT_HALF, kind=long)
    else if (a_m .eq. 0) then
      a = CMPLX(HALF, ZERO, kind=long)
    else if (a_m .gt. 0) then
      a = CMPLX(SQRT_HALF, ZERO, kind=long)
    end if

    if (a_m1 .lt. 0) then
      a1 = CMPLX(ZERO, -SQRT_HALF, kind=long)
    else if (a_m1 .eq. 0) then
      a1 = HALF
    else if (a_m1 .gt. 0) then
      a1 = CMPLX(SQRT_HALF, ZERO, kind=long)
    end if

    if (a_m2 .lt. 0) then
      a2 = CMPLX(ZERO, -SQRT_HALF, kind=long)
    else if (a_m2 .eq. 0) then
      a2 = HALF
    else if (a_m2 .gt. 0) then
      a2 = CMPLX(SQRT_HALF, ZERO, kind=long)
    end if

    if (a_m .lt. 0) then
      s = -1
    else
      s = 1
    end if

    if (a_m1 .lt. 0) then
      s1 = -1
    else
      s1 = 1
    end if

    if (a_m2 .lt. 0) then
      s2 = -1
    else
      s2 = 1
    end if

    if (MOD(a_m, 2) .eq. 0) then
      c = 1
    else
      c = -1
    end if

    if (MOD(a_m1, 2) .eq. 0) then
      c1 = 1
    else
      c1 = -1
    end if

    if (MOD(a_m2, 2) .eq. 0) then
      c2 = 1
    else
      c2 = -1
    end if

    cGaunt = CMPLX(ZERO, ZERO, kind=long)
    if (abs(a_m) + abs(a_m1) + abs(a_m2) .eq. 0) then
      cGaunt = cGaunt + (s1*s2*c*c1*c2 + s)*GauntCoef(a_l, abs(a_m), a_l1, abs(a_m1), a_l2, abs(a_m2))
    end if
    if (-abs(a_m) + abs(a_m1) + abs(a_m2) .eq. 0) then
      cGaunt = cGaunt + (c + s*s1*s2*c1*c2)*GauntCoef(a_l, -abs(a_m), a_l1, abs(a_m1), a_l2, abs(a_m2))
    end if
    if (-abs(a_m) - abs(a_m1) + abs(a_m2) .eq. 0) then
      cGaunt = cGaunt + (s1*c*c1 + s*s2*c2)*GauntCoef(a_l, -abs(a_m), a_l1, -abs(a_m1), a_l2, abs(a_m2))
    end if
    if (-abs(a_m) + abs(a_m1) - abs(a_m2) .eq. 0) then
      cGaunt = cGaunt + (s2*c*c2 + s*s1*c1)*GauntCoef(a_l, -abs(a_m), a_l1, abs(a_m1), a_l2, -abs(a_m2))
    end if
    cGaunt = cGaunt*conjg(a)*a1*a2
    rGaunt = real(CGaunt); 
  end function realGauntCoef

  pure function GauntCoef(a_l, a_m, a_l1, a_m1, a_l2, a_m2) result(Gaunt)
    !! complex Gaunt coefficient
    integer, intent(in) :: a_l, a_m, a_l1, a_m1, a_l2, a_m2
    real(long) Gaunt
    integer :: sign
    Gaunt = ZERO
    if (a_m + a_m1 + a_m2 .ne. 0) return; 
    sign = 1
    if (MOD(a_m, 2) .ne. 0) sign = -1
    Gaunt = sign*sqrt((TWO*a_l1 + ONE)*(TWO*a_l2 + ONE)/(TWO*a_l + ONE))*ONE/sqrt(4*PI) &
            *cleb(2*a_l1, 2*a_m1, 2*a_l2, 2*a_m2, 2*a_l, -2*a_m) &
            *cleb(2*a_l1, 0, 2*a_l2, 0, 2*a_l, 0)
    return
  end function GauntCoef

end module angularintegration
