!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module atom
  !! author: Sang-Kil Son
  !!
  !! Atomic data 
  !! extracted from OCTOPUS, http://www.tddft.org/programs/octopus/
  !!   written by Son, Sang-Kil in Nov. 2004 - Dec. 2004
  !!
  !! - ATOMIC_NAME(1...109)
  !! - ATOMIC_MASS(1...54)
  !!
  use constants
  implicit none
  character(len=2), parameter :: ATOMIC_NAME(109) = (/ &
                                 'H ', 'He', &                                                  ! period 1
                                 'Li', 'Be', 'B ', 'C ', 'N ', 'O ', 'F ', 'Ne', &                    ! period 2
                                 'Na', 'Mg', 'Al', 'Si', 'P ', 'S ', 'Cl', 'Ar', &                    ! period 3
                                 'K ', 'Ca', 'Sc', 'Ti', 'V ', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', &
                                 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', &  ! 4
                                 'Rb', 'Sr', 'Y ', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', &
                                 'In', 'Sn', 'Sb', 'Te', 'I ', 'Xe', &  ! 5
                                 'Cs', 'Ba', &
                                 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', &
                                 'Gd', 'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', &
                                 'Lu', 'Hf', 'Ta', 'W ', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg', &
                                 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', &  ! 6
                                 'Fr', 'Ra', &
                                 'Ac', 'Th', 'Pa', 'U ', 'Np', 'Pu', 'Am', &
                                 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', &
                                 'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt'/)               ! period 7
  real(long), parameter :: ATOMIC_MASS(54) = (/ &
                           1.00790D+00, 4.00260D+00, 6.94000D+00, 9.01218D+00, & !  H, He, Li, Be
                           10.81000D+00, 12.01100D+00, 14.00670D+00, 15.99940D+00, & !  B,  C,  N,  O
                           18.99840D+00, 20.17900D+00, 22.98977D+00, 24.30500D+00, & !  F, Ne, Na, Mg
                           26.98154D+00, 28.08550D+00, 30.97376D+00, 32.06000D+00, & ! Al, Si,  P,  S
                           35.45300D+00, 39.94800D+00, 39.09830D+00, 40.08000D+00, & ! Cl, Ar,  K, Ca
                           44.95590D+00, 47.90000D+00, 50.94150D+00, 51.99600D+00, & ! Sc, Ti,  V, Cr
                           54.93800D+00, 55.84700D+00, 58.93320D+00, 58.71000D+00, & ! Mn, Fe, Co, Ni
                           63.54600D+00, 65.38000D+00, 69.73500D+00, 72.59000D+00, & ! Cu, Zn, Ga, Ge
                           74.92160D+00, 78.96000D+00, 79.90400D+00, 83.80000D+00, & ! As, Se, Br, Kr
                           85.46780D+00, 87.62000D+00, 88.90590D+00, 91.22000D+00, & ! Rb, Sc,  Y, Zr
                           92.90640D+00, 95.94000D+00, 98.90620D+00, 101.0700D+00, & ! Nb, Mo, Tc, Ru
                           102.9055D+00, 106.4000D+00, 107.8680D+00, 112.4100D+00, & ! Rh, Pd, Ag, Cd
                           114.8200D+00, 118.6900D+00, 121.7500D+00, 127.6000D+00, & ! In, Sn, Sb, Te
                           126.9045D+00, 131.3000D+00/)                             !  I, Xe
private
public get_atomic_number, get_atomic_species, ATOMIC_NAME, ATOMIC_MASS
contains

  function get_atomic_number(s) result(n)
    !! returns atomic number of a given element
    integer :: n
    character(len=2), intent(in) :: s
    character(len=2) :: atom
    character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    integer :: i
    atom = s
    !
    ! make the second character lowercase
    !
    if (index(uppercase, atom(2:2)) .ne. 0 &
        ) atom(2:2) = char(ichar(s(2:2)) - ichar("A") + ichar("a"))
    !
    ! find a chemical element in ATOMIC_NAME
    !
    do i = 1, size(ATOMIC_NAME)
      if (atom .eq. ATOMIC_NAME(i)) then
        n = i             ! return the atomic number
        return
      end if
    end do
    write (STDERR, *) "There is no atomic name like "//trim(atom)//"."
    stop
  end function get_atomic_number

  subroutine get_atomic_species(s, charge, mass)
    !! returns charge and mass of a given element
    character(len=*), intent(in) :: s
    integer, intent(out) :: charge
    real(long), intent(out) :: mass
    character(len=2) :: atom
    character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
    integer :: i, is, ie, INFO
    !
    ! figure out the atomic mass
    !
    mass = ZERO
    is = 0; ie = 0
    do i = 1, len(s)                      ! mass can be assigned as Br[80]
      if (s(i:i) .eq. "[") is = i
      if (s(i:i) .eq. "]") ie = i
    end do
    if (is .eq. 0) is = len(s) + 1
    if (is .ne. 0 .and. ie .ne. 0) then     ! set up mass
      read (s(is + 1:ie - 1), *, iostat=INFO) mass
    end if
    !
    ! figure out the charge
    ! atom name is
    !
    charge = 0
    atom = s(1:2)
    if (index(uppercase, atom(2:2)) .ne. 0) then
      !
      ! if the second character is uppercase, make it lowercase
      !
      atom(2:2) = char(ichar(s(2:2)) - ichar("A") + ichar("a"))
    else if (index(lowercase, atom(2:2)) .eq. 0) then
      !
      ! if the second character is not lowercase, ignore it
      !
      atom = s(1:1)
    end if
    if (atom .eq. "D") then                       ! D: deutrium
      charge = 1
      mass = TWO
    else if (atom .eq. "T") then                  ! T: tritium
      charge = 1
      mass = THREE
    else
      do i = 1, size(ATOMIC_NAME)                 ! set up charge
        if (atom .eq. ATOMIC_NAME(i)) charge = i
      end do
      if (mass .eq. ZERO) mass = ATOMIC_MASS(charge)  ! if mass is not assigned
      if (charge .eq. 0) then                           ! if charge not assigned
        write (STDERR, *) "There is no atomic name like "//trim(s)//"."
        stop
      end if
    end if
  end subroutine get_atomic_species

end module atom
