!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module atom_in_mol
  !! author: Sang-Kil Son
  !!
  !! ATOM_IN_MOLECULE module to define atomic data within a molecule
  !!   written by Son, Sang-Kil in Mar. 2014
  !!
  !! USAGE:
  !!   type(Atom_set) :: A(:)
  !!   call construct_Atom_set( P, A )
  use constants
  use read_param
  use gps
  use angular
  use sph_harmonics
  use atom
  use stdio
  implicit none
  type :: Atom_set
    !! contains atom information
    integer :: i_nuc    ! atomic index
    real(long) :: x, y, z
    integer :: charge
    real(long) :: fcharge
    real(long) :: sigma ! for nuclear gaussian distribution distribution model (in libcint zeta=1/(2*sigma^2))
    character(len=6) :: label
    !
    ! atomic grids (AG)
    !
    integer :: N_r      ! # of radial grids in atomic grids
    integer :: N_ang    ! # of angular grids in atomic grids
    integer :: lmax_ang ! lmax to be used for angular grids
    real(long) :: rmax, mapping_L               ! radial grid parameters
    real(long), allocatable :: Y_lm(:, :)        ! store sph_harmonics, Y_lm
    type(Grid) :: radial
    type(Angular_Grid) :: angular
    !
    ! atomic orbitals (AO), to be used as basis functions in molecular calc.
    !
    integer :: N_nao            ! # of different orbitals: (n,l)-subshells
    real(long), allocatable :: u_AO(:, :)        ! radial f. NAO, (1:N_r,1:N_nao)
    real(long), allocatable :: D2_AO(:, :)       ! radial f. D2_, (1:N_r,1:N_nao)
    real(long), allocatable :: dAO_dr(:, :)      ! radial derivative f. NAO, (1:N_r,1:N_nao)
    integer, allocatable :: n_AO(:)             ! principal quantum #, (1:N_nao)
    integer, allocatable :: l_AO(:)             ! angular quantum #, (1:N_nao)
    integer :: N_orb    ! how many different orbitals (radial w.f.'s)
    integer, allocatable :: config(:, :)         ! config(1:N_orb,1:3)
    !!
    !! atomic configuration
    !!   config(i,1) -> quantum number n
    !!   config(i,2) -> quantum number l
    !!   config(i,3) -> occupation number
    !!
    character(len=80) :: nao_name 
    !! the basis name printed out    
    !! Basis functions to be read in read_molecule_param() --------------------
    !!   We employ NAO (Numerical Atomic Orbital) as basis functions
    !!     for molecular calculations.
    !!   N_nao_table(:,:) : specifying # of NAO functions per l and each atom
    !!     (0:MAX_lmax,1:N_nuc) where MAX_lmax = max. of lmax for all atoms
    !!     Remember that 2s1p means two s orbitals and one p orbital,
    !!     so # of basis functions becomes 5!!!
    !!   N_basis : total # of basis functions, considering l
    !!
    !!   EXAMPLE of basis_string
    !!     2s1p at i-th atom: two s orbitals and one p orbitals (px, py, pz)
    !!       N_nao = 3;  N_basis = 2*1 + 1*3 = 5
    !!       N_nao_table(l=0,i) = 2;  N_nao_table(l=1,i) = 1
    !!     5s3p1d at i-th atom: 5 s orbitals, 3 p orbitals, and 1 d orbital
    !!       N_nao = 9;  N_basis = 5*1 + 3*3 + 1*5 = 19
    !!       N_nao_table(0:2,i) = ( 5 3 1 )
    !!
    !!   EXAMPLE of the NAO basis function file:
    !!     200  1.0  50.0        # N_AG, L_AG, rmax_AG
    !!     2  5                  # lmax and how many NAOs for given l
    !!     ... (5 lines with 200 NAO values at 200 radial grid points for l=0)
    !!     ... (5 lines with 200 NAO values at 200 radial grid points for l=1)
    !!     ... (5 lines with 200 NAO values at 200 radial grid points for l=2)
    !!     EOF
    !!
    integer, allocatable :: N_nao_table(:)            ! (0:lmax)
    character(len=512) :: hole_string ! the string that defines the atomic hole configuration
    character(len=512) :: conf_string ! the string that defines the atomic hole configuration
    character(len=512) :: nao_file_nuc  ! filename, (1:N_nuc)
    character(len=512) :: auger_file_nuc  ! filename, (1:N_nuc)
    character(len=512) :: tdipole_file_nuc  ! filename, (1:N_nuc)

    integer :: sym_op_pair(8)
    !! symmetry operations pairs
    !! contains for every defined symmetry operation (i=1..8)
    !! the atom index to which the specific atom is translated under
    !! the symmetry operation i ( or -1 if symmetry operation does change the molecule ).
    !! Symmetry operations are:
    !! 1 E    : identity
    !! 2 C2X  : 180 degree rotation around X axis
    !! 3 C2Y  : 180 degree rotation around Y axis
    !! 4 C2Z  : 180 degree rotation around Z axis
    !! 5 i    : inversion
    !! 6 SIGXY: reflection on XY plane
    !! 7 SIGXZ: reflection on XZ plane
    !! 8 SIGYZ: reflection on YZ plane
  end type Atom_set
  private
  public :: Atom_set, purge_Atom_set, construct_Atom_set, update_Atom_set_with_NAO, update_Atom_set_with_displacement, &
    dump_Atom_set, subshell_name

contains
  subroutine purge_Atom_set(A)
    !! deallocates array of atom_sets
    !!
    type(Atom_set), allocatable, intent(inout) :: A(:)
    integer :: i
    do i = 1, size(A, 1)
      deallocate (A(i)%N_nao_table)
      deallocate (A(i)%n_AO)
      deallocate (A(i)%l_AO)
      deallocate (A(i)%config)
      if (allocated(A(i)%Y_lm)) deallocate (A(i)%Y_lm)
      if (allocated(A(i)%u_AO)) deallocate (A(i)%u_AO)
      if (allocated(A(i)%D2_AO)) deallocate (A(i)%D2_AO)
      if (allocated(A(i)%DAO_dr)) deallocate (A(i)%DAO_dr)
      call purge_grid(A(i)%radial)
      call purge_Angular_Grid(A(i)%angular)
    end do
    deallocate (A)
  end subroutine purge_Atom_set

  subroutine construct_Atom_set(P, A)
    !! constructs atom set
    !!
    type(Param), intent(in) :: P
    type(Atom_set), allocatable, intent(out) :: A(:)
    integer :: i
    character(len=6) :: str
    !
    ! Do you need more than s, p, d, f orbitals?
    ! Now, the limit is l <= 10: s, p, d, f, g, h, i, j, k, l, m orbitals
    !
    integer, parameter :: MAX_lmax_nao = 10
    integer :: ind
    logical :: file_exist
    character(len=1500) :: cmd_line
    !
    ! we have N_nuc atoms in a molecule
    !
    allocate (A(P%N_nuc))
    !
    ! set up atomic positions, AG (atomic grids), and AO (atomic orbitals)
    !
    do i = 1, P%N_nuc
      allocate (A(i)%N_nao_table(0:MAX_lmax_nao), source=P%N_nao_table(0:MAX_lmax_nao, i))
      A(i)%nao_file_nuc = P%nao_file_nuc(i)
      A(i)%nao_name = P%nao_name(i)
      !A(i)%N_nao_table(0:MAX_lmax_nao) = P%N_nao_table(0:MAX_lmax_nao, i)

      ind = index(P%nao_file_nuc(i), ".nao")
      A(i)%auger_file_nuc = trim(P%nao_file_nuc(i) (:ind - 1)//".auger")
      A(i)%tdipole_file_nuc = trim(P%nao_file_nuc(i) (:ind - 1)//".tdipole")
      A(i)%hole_string = ""

      !
      ! put an atomic position and a charge over there
      !
      A(i)%i_nuc = i
      A(i)%x = P%x_nuc(i)
      A(i)%y = P%y_nuc(i)
      A(i)%z = P%z_nuc(i)
      A(i)%charge = P%charge_nuc(i)
      A(i)%fcharge = P%fcharge_nuc(i)
      A(i)%sigma = P%sigma_nuc(i)
      write (str, "(i6)") A(i)%i_nuc
      if (A(i)%charge .eq. 0) then
        if (A(i)%sigma .gt. ZERO) then
          A(i)%label = "XG"
        else
          A(i)%label = "X"
        end if
      else
        A(i)%label = trim(ATOMIC_NAME(A(i)%charge))//trim(adjustl(str))
      end if
      !
      ! set up # of different atomic orbitals
      ! this is DIFFERENT from # of basis functions.
      !
      A(i)%N_nao = sum(P%N_nao_table(:, i))

      !
      ! In case we have no nao file (if we use gaussians) use default grid parameters
      !
      if (A(i)%N_nao .eq. 0) then
        A(i)%N_r = 0
        A(i)%N_ang = 0
        A(i)%lmax_ang = 0
        A(i)%mapping_L = 0
        A(i)%rmax = 0
      else if (P%nao_file_nuc(i) .eq. "" .and. P%yes_gaussian) then
        A(i)%N_r = P%N
        A(i)%mapping_L = P%L
        A(i)%rmax = P%rmax
        A(i)%nao_file_nuc = trim(P%nuc_name(i))//".nao"
        A(i)%auger_file_nuc = trim(P%nuc_name(i))//".auger"
        A(i)%tdipole_file_nuc = trim(P%nuc_name(i))//".tdipole"
      else
        ! check whether .nao file exists. if not, call xatom and create one with default grid parameters
        inquire (file=A(i)%nao_file_nuc, EXIST=file_exist)
        if (.NOT. file_exist) then
          inquire (file=P%XATOM_PATH, EXIST=file_exist)
          if (.not. file_exist) then
            write (STDERR, '("Not found: ",a)') P%XATOM_PATH
            stop 'Please specify xatom path'
          end if
          ! default grid parameter:
          A(i)%N_r = P%N
          A(i)%mapping_L = P%L
          A(i)%rmax = P%rmax
          write (cmd_line, '(a,1x,"-silent -latter off",1x,"-s ",a,1x,"-rmax ",f15.3,1x,"-N ",i4, " -nao -file_nao ",a)') &
            trim(P%XATOM_PATH), trim(ATOMIC_NAME(A(i)%charge)), A(i)%rmax, A(i)%N_r, A(i)%nao_file_nuc
          write (STDERR, "('# CALLING:',100a)") trim(cmd_line)
          call system(cmd_line)
        end if

      end if

      !
      ! read AO from a file specified in P
      ! radial grid parameters are also read from the file
      !
      if (P%nao_file_nuc(i) .eq. "" .and. P%yes_gaussian) then
        call prepare_AO(A(i), "", A(i)%N_nao_table(:))
      else
        call prepare_AO(A(i), A(i)%nao_file_nuc, A(i)%N_nao_table(:))
      end if
      !
      ! construct AG
      ! - radial grid param.: A%N_r, A%mapping_L, A%rmax, loaded from the file
      ! - angular grid param.: A%lmax_ang, specified by P
      !
      if (A(i)%N_nao .gt. 0) &
        call prepare_AG(A(i), P%lmax_AG(i), P%type_radial, P%type_angular)
      !
      ! construct D2_AO using radial grid points
      !
      if (A(i)%N_nao .gt. 0 .and. .not. P%yes_gaussian) then
        call prepare_D2_AO(A(i))
        call prepare_dAO_dr(A(i))
      end if
    end do
    call sym_op_pairs(A)

  end subroutine construct_Atom_set

  subroutine update_Atom_set_with_NAO(A)
    !! update Atom_set with new NAOs
    !!   it rereads the NAO files assigned in A(:).
    type(Atom_set), intent(inout) :: A(:)
    integer :: i_nuc, N_nuc
    N_nuc = size(A)
    do i_nuc = 1, N_nuc
      call prepare_AO(A(i_nuc), A(i_nuc)%nao_file_nuc, A(i_nuc)%N_nao_table(:))
      if (A(i_nuc)%nao_file_nuc .ne. "") then
        call prepare_D2_AO(A(i_nuc))
        call prepare_dAO_dr(A(i_nuc))
      end if
    end do
  end subroutine update_Atom_set_with_NAO

  subroutine update_Atom_set_with_displacement(A, displacement)
    !! update Atom_set with displacement
    type(Atom_set), intent(inout) :: A(:)
    real(long), intent(in) :: displacement(:)
    integer :: i_nuc, N_nuc
    N_nuc = size(A)
    do i_nuc = 1, N_nuc
      A(i_nuc)%x = A(i_nuc)%x + displacement(i_nuc * 3 - 2)   ! x-coordinate
      A(i_nuc)%y = A(i_nuc)%y + displacement(i_nuc * 3 - 1)   ! y-coordinate
      A(i_nuc)%z = A(i_nuc)%z + displacement(i_nuc * 3)   ! z-coordinate
    end do
  end subroutine update_Atom_set_with_displacement

!>----------------------------------------------------------------------------
!! prepare atomic orbitals
!!   read in numerical radial orbitals from the NAO files
!!   set up the following arrays: A%u_AO, A%D2_AO, A%n_AO, A%l_AO & A%config
!!   also read in radial grid information: A%N_r, A%mapping_L, A%rmax
!! NOTE: A%D2_AO is not really prepared in this routine.
!!       therefore, you should call prepare_D2_AO!
!!----------------------------------------------------------------------------
  subroutine prepare_AO(A, nao_file, N_nao_each)
    type(Atom_set), intent(inout) :: A
    character(len=*), intent(in) :: nao_file !! file name of nao files
    integer, intent(in) :: N_nao_each(0:)    ! always be careful
    integer :: j, k, l, lmax, nev
    real(long), allocatable :: u(:)
    integer, parameter :: IO_BASIS = 20   ! no one knows why I'm using this #...
    integer :: INFO
    integer :: num
    if (nao_file .eq. "") then
      if (.not. allocated(A%n_AO)) allocate (A%n_AO(A%N_nao))
      if (.not. allocated(A%l_AO)) allocate (A%l_AO(A%N_nao))
      if (.not. allocated(A%config)) allocate (A%config(A%N_nao, 3))
      lmax = size(N_nao_each) - 1
      k = 0
      num = 0
      do l = 0, lmax
        do j = 1, N_nao_each(l)

          k = k + 1
          A%config(k, 1) = j + l
          A%config(k, 2) = l
          A%config(k, 3) = neutral_occ_aufbau(A%charge, j + l, l)
          num = num + A%config(k, 3)
          A%n_AO(k) = j + l       ! principal quantum #
          A%l_AO(k) = l           ! angular quantum #
        end do
      end do
      A%N_orb = k
    else

      !
      ! EXAMPLE of the NAO basis function file:
      !     200  1.0  50.0  # N_AG, L_AG, rmax_AG --> used for radial grid param.
      !     2  5            # lmax and how many NAOs for given l
      !     3               # the number of orbitals
      !     1  0  2         # config(1,1:3) --> 1s2
      !     2  0  2         # config(2,1:3) --> 2s2
      !     2  1  3         # config(3,1:3) --> 2p3
      !     ... (5 sets of 200 NAO values at every radial grid for l=0)
      !     ... (5 sets of 200 NAO values at every radial grid for l=1)
      !     ... (5 sets of 200 NAO values at every radial grid for l=2)
      !     EOF
      ! WARNING: A radial function, u(r), is used here!
      !
      open (IO_BASIS, file=nao_file, action="read", iostat=INFO)
      if (INFO .ne. 0) then
        write (STDERR, *) "File not found: ", trim(nao_file)
        stop
      end if
      read (IO_BASIS, *) A%N_r, A%mapping_L, A%rmax      ! radial grid parameters
      if (.not. allocated(A%u_AO)) allocate (A%u_AO(A%N_r, A%N_nao))
      if (.not. allocated(A%D2_AO)) allocate (A%D2_AO(A%N_r, A%N_nao))
      if (.not. allocated(A%dAO_dr)) allocate (A%dAO_dr(A%N_r, A%N_nao))
      if (.not. allocated(A%n_AO)) allocate (A%n_AO(A%N_nao))
      if (.not. allocated(A%l_AO)) allocate (A%l_AO(A%N_nao))
      allocate (u(A%N_r))
      !
      ! read configuration info
      !
      read (IO_BASIS, *) A%N_orb
      if (.not. allocated(A%config)) allocate (A%config(A%N_orb, 3))
      A%config(:,:) = 0
      do j = 1, A%N_orb
        read (IO_BASIS, *) A%config(j, 1:3)
      end do
      !
      ! read wavefunctions
      !
      read (IO_BASIS, *) lmax, nev                       ! related to # of NAOs
      k = 0
      do l = 0, lmax
        do j = 1, nev
          if (j .le. N_nao_each(l)) then
            k = k + 1
            read (IO_BASIS, *) A%u_AO(:, k)
            A%n_AO(k) = j + l       ! principal quantum #
            A%l_AO(k) = l           ! angular quantum #
          else
            read (IO_BASIS, *) u(:)
          end if
        end do
      end do
      deallocate (u)
      close (IO_BASIS)
    end if
  contains
    function neutral_occ_aufbau(Z, n, l) result(occ)
      integer, intent(in):: Z, n, l
      integer :: occ
      integer :: n_i, l_i
      integer :: num
      num = 0
      do n_i = 1, n
        do l_i = 0, n_i - 1
          if (num .eq. Z) then
            occ = 0
          else
            occ = 2*(2*l_i + 1)
            if (num + occ .gt. Z) occ = Z - num
          end if
          if (n_i .eq. n .and. l_i .eq. l) return
          num = num + occ
        end do
      end do
    end function neutral_occ_aufbau
  end subroutine prepare_AO

  !>----------------------------------------------------------------------------
  !! prepare atomic grid points
  !!   set up radial grids ( A%radial ) and angular grids ( A%angular )
  !!   A%radial requires radial grid parameters which are defined in prepare_AO
  !!   therefore, call prepare_AO first before call this routine!
  !!----------------------------------------------------------------------------
  subroutine prepare_AG(A, lmax_ang, type_radial, type_angular)
    type(Atom_set), intent(inout) :: A
    integer, intent(in) :: lmax_ang
    character(len=*), intent(in) :: type_radial, type_angular !! describe the angular grid type
    !type(Grid) :: r
    !type(Angular_Grid) :: ang
    integer :: i, l, m, lm
    real(long) :: x, y, z, theta, phi
    !
    ! create radial grids: parameters are given from the basis file
    !
    call construct_Grid(type_radial, A%N_r, A%radial, Rmin=ZERO, Rmax=A%rmax, L=A%mapping_L)
    ! create angular grids: parameters are given from P
    !
    call construct_Angular_Grid(type_angular,lmax_ang, A%angular )
    if (.not. allocated(A%Y_lm)) allocate (A%Y_lm(A%angular%N, (lmax_ang + 1)**2))
    !
    ! put angular grids: positions and weights
    !
    A%N_ang = A%angular%N
    A%lmax_ang = lmax_ang
    do i = 1, A%N_ang
      x = A%angular%x(i); y = A%angular%y(i); z = A%angular%z(i)
      theta = acos(z/sqrt(x**2 + y**2 + z**2))  ! 0 <= theta <= pi
      phi = atan2(y, x)                           ! 0 <= phi < 2 pi
      if (phi .lt. ZERO) phi = phi + TWO * PI
      !
      ! set up sph_harmonics Y_lm
      !
      do l = 0, A%lmax_ang
        do m = -l, l
          lm = l**2 + (m + l) + 1       ! lm = 1, 2, 3, ...
          A%Y_lm(i, lm) = sph_harmonics_real(l, m, theta, phi)
        end do
      end do
    end do
  end subroutine prepare_AG

  !> subroutine prepare_dAO_dr
  !! calculates the radial derivative of basis functions and stores it in A%D_ao
  !! Note: this is only the radial derivative, since the spherical harmonic part
  !! is not eigenfunction to the gradient
  subroutine prepare_dAO_dr(A)
    type(Atom_set), intent(inout) :: A
    integer :: i
    if (.not. allocated(A%radial%D1)) then
      call error_message("Error in prepare_D2_AO: A%radial is not defined; "// &
        "call prepare_AG() first.")
    end if
    forall (i=1:A%N_nao)
      A%dAO_dr(:, i) = A%radial%wf_coeff(:)*matmul(A%radial%D1(:, :), (A%u_AO(:, i)/A%radial%wf_coeff(:)))
    end forall
  end subroutine prepare_dAO_dr
  !>----------------------------------------------------------------------------
  !! subroutine prepare_D2_AO(
  !!  prepare the second derivative of w.f. for the kinetic operator: u''
  !! NOTE: A%radial should be defined
  !!       therefore call prepare_AG() first before call this routine!
  !! NOTE: basically, this routine is identical to prepare_D2f() in XMOLset.f90
  !!----------------------------------------------------------------------------
  subroutine prepare_D2_AO(A)
    type(Atom_set), intent(inout) :: A
    integer :: i
    !
    ! remember u'' means d^2 u / d r^2.
    !   Since r%D2 provides d^2 / d x^2, u has to be transformed into x coord by
    !      u / r%wf_coeff.
    !   Then it should be transformed back into r coord by
    !      matmul( r%D2, u / r%wf_coeff ) * r%wf_coeff.
    !
    ! HERE, D2_AO does not include the 1/r factor,
    ! in contrast to the original implementation!!!
    !   current implementation: D2_AO = u''(r)
    !   original implementation (read from the NAO file): D2_AO = u''(r) / r
    !
    if (.not. allocated(A%radial%D2)) then
      call error_message("Error in prepare_D2_AO: A%radial is not defined; "// &
        "call prepare_AG() first.")
    end if



    ! I have no idea why, but for some reason the above way is indicated to produce mem leak, 22.05.2015, Ludger
    forall (i=1:A%N_nao)
      A%D2_AO(:, i) = A%radial%wf_coeff(:)*matmul(A%radial%D2(:, :), (A%u_AO(:, i)/A%radial%wf_coeff(:)))
      A%D2_AO(:, i) = A%D2_AO(:, i) - dble(A%l_AO(i)*(A%l_AO(i) + 1))/A%radial%r(:)**2*A%u_AO(:, i)
    end forall
  end subroutine prepare_D2_AO

  subroutine dump_Atom_set(A, IOUT)
    !! prints atom set information
    type(Atom_set), intent(in) :: A(:)
    integer, intent(in), optional :: IOUT
    integer :: i, j, N_nuc, OUTPUT
    real(long) :: x, y, z, theta, phi
    OUTPUT = STDOUT; if (present(IOUT)) OUTPUT = IOUT
    N_nuc = size(A)
    write (OUTPUT, "(a,x,i4)") "N_nuc = ", N_nuc
    do i = 1, N_nuc
      write (OUTPUT, "(a,x,i4)") "Atom_index: ", i
      write (OUTPUT, "(a,x,f10.3)") "  x = ", A(i)%x
      write (OUTPUT, "(a,x,f10.3)") "  y = ", A(i)%y
      write (OUTPUT, "(a,x,f10.3)") "  z = ", A(i)%z
      write (OUTPUT, "(a,x,i4)") "  charge = ", A(i)%charge
      write (OUTPUT, "(a)") "  Atomic grids:"
      write (OUTPUT, "(a,x,i6)") "    N_r = ", A(i)%N_r
      write (OUTPUT, "(a,x,f15.3)") "    mapping_L = ", A(i)%mapping_L
      write (OUTPUT, "(a,x,f15.3)") "    rmax = ", A(i)%rmax
      write (OUTPUT, "(2a15)") "r", "weight"
      do j = 1, A(i)%N_r
        write (OUTPUT, "(f15.5,x,f15.5)") A(i)%radial%r(j), A(i)%radial%weight(j)
      end do
      write (OUTPUT, "(a,x,i4)") "    N_ang = ", A(i)%N_ang
      write (OUTPUT, "(a,x,i4)") "    lmax_ang = ", A(i)%lmax_ang
      write (OUTPUT, "(3a15)") "theta", "phi", "weight"
      do j = 1, A(i)%N_ang
        x = A(i)%angular%x(j); y = A(i)%angular%y(j); z = A(i)%angular%z(j)
        theta = acos(z/sqrt(x**2 + y**2 + z**2))  ! 0 <= theta <= pi
        phi = atan2(y, x)                           ! 0 <= phi < 2 pi
        if (phi .lt. ZERO) phi = phi + TWO * PI
        write (OUTPUT, "(f15.5,f15.5,f15.5)") theta, phi, A(i)%angular%weight(j)
      end do
      write (OUTPUT, "(a)") "  Atomic orbitals:"
      write (OUTPUT, "(a,x,i4)") "    N_nao = ", A(i)%N_nao
      if (allocated(A(i)%u_AO)) then
        do j = 1, size(A(i)%u_AO, 2)
          write (OUTPUT, "(5000es24.5e3)") A(i)%u_AO(:, j)
          write (OUTPUT, "(5000es24.5e3)") A(i)%D2_AO(:, j)
        end do
      end if
      write (OUTPUT, "()")
    end do
  end subroutine dump_Atom_set

  function subshell_name(n, l) result(s)
    !! subshell_name: return 1s, 2s, 2p, ... according to (n,l)
    character(len=5) :: s, str
    integer, intent(in) :: n, l
    character(len=1) :: t
    select case (l)
     case (0); t = 's'
     case (1); t = 'p'
     case (2); t = 'd'
     case default; t = achar(iachar('f') + l - 3)
    end select
    write (str, "(i4,a1)") n, t
    s = trim(adjustl(str))
  end function subshell_name

  !> for each atom in array A fill the array sym_op_pair(i)
  !! that contains for every defined symmetry operation (i=1..8)
  !! the atom index to which the specific atom is translated under
  !! the symmetry operation ( or -1 if symmetry operation does change the molecule )
  !! parameter THRESHOLD defines the threshold for the coordinates
  subroutine sym_op_pairs(A)
    type(Atom_set), intent(inout) :: A(:)
    integer :: i, atom1, atom2
    real(long), parameter :: THRESHOLD = 1e-10
    integer :: symx(8)
    integer :: symy(8)
    integer :: symz(8)

    ! The symmetry operations
    ! 1 E
    ! 2 C2X
    ! 3 C2Y
    ! 4 C2Z
    ! 5 i
    ! 6 SIGXY
    ! 7 SIGXZ
    ! 8 SIGYZ
    ! this defines how coordinates x,y,z change under the symmetry operation:
    symx(1) = 1; symy(1) = 1; symz(1) = 1
    symx(2) = 1; symy(2) = -1; symz(2) = -1
    symx(3) = -1; symy(3) = 1; symz(3) = -1
    symx(4) = -1; symy(4) = -1; symz(4) = 1
    symx(5) = -1; symy(5) = -1; symz(5) = -1
    symx(6) = 1; symy(6) = 1; symz(6) = -1
    symx(7) = 1; symy(7) = -1; symz(7) = 1
    symx(8) = -1; symy(8) = 1; symz(8) = 1

    do i = 1, 8
      A(:)%sym_op_pair(i) = -1
    end do

    do atom1 = 1, size(A)
      do atom2 = atom1, size(A)
        if (A(atom1)%charge .ne. A(atom2)%charge) cycle
        do i = 1, 8
          if ( &
            abs(symx(i)*A(atom1)%x - A(atom2)%x) .lt. THRESHOLD .and. &
            abs(symy(i)*A(atom1)%y - A(atom2)%y) .lt. THRESHOLD .and. &
            abs(symz(i)*A(atom1)%z - A(atom2)%z) .lt. THRESHOLD) then
            A(atom1)%sym_op_pair(i) = atom2
            A(atom2)%sym_op_pair(i) = atom1
          end if
        end do
      end do
    end do

  end subroutine sym_op_pairs

end module atom_in_mol
