!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module augerrates
  !! author: Ludger Inhester
  !!
  !! This module is for the calculation of Auger rates via one-center approximation
  use angularintegration
  use constants
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use process
  use averaging
  use spin_coupling
  use elec_config
  use mcscf
  use population
  use mergesort
  use spin_coupling
  use hartreefock
  use rwfile
  use operations
  use basis_func
  use scf
  use anglib
  implicit none

  type :: RadialIntegrals
    !! Radial continuum-bound integral information from xatom
    integer :: N ! number of of entries
    integer, dimension(:), allocatable :: K
    integer, dimension(:), allocatable :: n1
    integer, dimension(:), allocatable :: l1
    integer, dimension(:), allocatable :: n2
    integer, dimension(:), allocatable :: l2
    integer, dimension(:), allocatable :: n3
    integer, dimension(:), allocatable :: l3
    integer, dimension(:), allocatable :: l4
    real(long), dimension(:), allocatable :: R

    real(long), dimension(:), allocatable :: delta ! phase shifts

! In case we use a generic basis set,
! we need info how the basis functions
! translate into the minimal atomic orbitals
! min basis data
! how the basis set translates into the minimal basis
! is given by Pmin, minimal basis information is provided by min_info
! Pmin and min_info have to be provided in .auger file
! and created in advance by the project_basis program

    real(long), allocatable :: Pmin(:, :) ! basis to min basis projection
    real(long), allocatable :: Omin(:, :) ! keeps the orbital expansion in minimal basis
    real(long), allocatable :: Smin(:, :) ! keeps the S matrix of the minimal basis
    integer, allocatable :: min_info(:, :) ! info on minimal basis:
    ! min_info(:,1)=atom index
    ! min_info(:,2)=n
    ! min_info(:,3)=l
    ! min_info(:,4)=m
  end type RadialIntegrals

  type :: teint_contributions
    !! bound-continuum integral contribution
    integer :: val1 !! valence hole 1
    integer :: val2 !! valence hole 2
    integer :: hole !! refilled core hole
    real(long) :: coef !! amplitude
  end type teint_contributions
  private
  public calcAugerRates, calcAugerRates_spin_avg, calcAugerRates_spin, calcAugerRates_CI
contains
  subroutine calcAugerRates(P, A, B, F, S, auger_fpath)
    !! calculates Auger-Meitner decay transitions based on orbital energy differences
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S

    integer :: hole, val1, val2 ! molecular orbitals
    integer :: j ! counter for MOs
    integer :: m, l ! counter for angular momentum
    integer :: at ! counter for atoms
    integer :: lmax ! this is the maximum angular momentum in the continuum which we consider

    real(long) :: TotalRate, Rate1, Rate3, Rate, ChannelRate, HoleRate, A1, A2 ! Rates & Amplitudes
    type(atomorbital) :: continuumAO ! continuum orbital parameter structure
    type(RadialIntegrals), dimension(P%N_nuc) :: RI ! Radialintegral structure to be filled by xatom output
    character(len=512) :: auger_file_nuc ! filename

    !! -- Mar.10.2015 : read-in specified fpath ---
    character*(*), optional :: auger_fpath(:)       !! Currently we use char*(128) in MC code
    real(long) :: AE_energy ! Auger electron energy

    if (P%yes_MCSCF) then
      stop "with MCSCF use additional -process_CI option"
    end if
    if (P%yes_auger_random) then
      write (P%OUTPUT, *) '# using random re-scale:', F%random_scale
    end if
    TotalRate = ZERO
    RI(:)%N = 0
    lmax = 0
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger file
      if (A(at)%charge .eq. 1) cycle
      if (present(auger_fpath)) then              ! extension Mar.10.2015
        auger_file_nuc = ""                        !
        auger_file_nuc = trim(auger_fpath(at))   !
      else                                       !
        auger_file_nuc = A(at)%auger_file_nuc
      end if
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Reading Auger Radial integrals from file ',a)") trim(auger_file_nuc)
      end if
      call readRIntegral(auger_file_nuc, RI(at))
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# atom: ',i4,' ',a, ' : ',i4,' entries')") &
          at, trim(auger_file_nuc), size(RI(at)%R)
      end if
      lmax = max(lmax, maxval(RI(at)%l4))

      ! prepare the projection of MO's into minimal basis set
      if (allocated(RI(at)%Pmin)) then
        allocate (RI(at)%Smin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        forall (j=1:size(RI(at)%Pmin, 1)) RI(at)%Smin(j, j) = ONE
        allocate (RI(at)%Omin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        RI(at)%Omin = matmul(RI(at)%Pmin, F%LC)
      else if (P%yes_gaussian) then
        allocate (RI(at)%Omin(size(F%Omin, 1), size(F%Omin, 2)))
        allocate (RI(at)%Smin(size(F%Smin, 1), size(F%Smin, 2)))
        allocate (RI(at)%min_info(size(F%min_info, 1), size(F%min_info, 2)))
        RI(at)%Omin = F%Omin
        RI(at)%Smin = F%Smin
        RI(at)%min_info = F%min_info
      end if! if Pmin allocated
      if (P%yes_verbose .and. allocated(RI(at)%min_info)) then
        call print_projection_info(RI(at)%min_info, RI(at)%Omin, RI(at)%Smin, at, P%OUTPUT)
      end if! yes_verbose
    end do! atom

    do hole = 1, size(F%N_occ)
      if (.NOT. has_hole(P, F, F%N_occ, hole)) cycle
      HoleRate = ZERO
      do val1 = 1, size(F%N_occ)
        if (.NOT. is_occupied(P, F, F%N_occ, val1)) cycle
        do val2 = val1, size(F%N_occ)
          if (.NOT. is_pair_occupied(P, F, F%N_occ, val1, val2)) cycle

          ! Auger electron energy is derived from the mo orbital energy difference
          AE_energy = F%CE(val1) + F%CE(val2) - F%CE(hole)

          if (AE_energy .lt. 0) cycle
          continuumAO%n = 99 ! dummy index for the electronic continuum

          ! loop over all continuum channel
          ChannelRate = ZERO
          do l = 0, lmax
            continuumAO%l = l
            do m = -l, l
              continuumAO%m = m
              if (val1 .eq. val2) then
                ! Singlet contributions
                ! only possible in doubly occupied MOs
                Rate1 = ZERO
                do at = 1, P%N_nuc
                  if (allocated(RI(at)%Omin)) then
                    A1 = AugerTransitionAmplitude(P, RI(at)%min_info, RI(at)%Omin(:, val1), &
                      RI(at)%Omin(:, val2), RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                  else
                    A1 = AugerTransitionAmplitude(P, B%info, F%LC(:, val1), F%LC(:, val2), &
                      F%LC(:, hole), at, continuumAO, RI(:))
                  end if
                  Rate1 = Rate1 + TWO*PI*(A1)**2
                end do
                Rate = Rate1
                ! multiply with the chance that we have a val1,val2 combination
                Rate = Rate*stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)
                ! statistical factor for the hole
                Rate = TWO*stat_factor_not_occupied(P, F, F%N_occ, hole)*Rate
              else
                ! singlet contribution + 3*triplet contribution

                Rate1 = ZERO
                Rate3 = ZERO
                do at = 1, P%N_nuc
                  ! special case for hydrogen: no auger
                  if (A(at)%charge .eq. 1) cycle

                  if (allocated(RI(at)%Omin)) then
                    A1 = AugerTransitionAmplitude(P, RI(at)%min_info, &
                      RI(at)%Omin(:, val1), RI(at)%Omin(:, val2), &
                      RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                    A2 = AugerTransitionAmplitude(P, RI(at)%min_info, &
                      RI(at)%Omin(:, val2), RI(at)%Omin(:, val1), &
                      RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                  else
                    A1 = AugerTransitionAmplitude(P, B%info, F%LC(:, val1), F%LC(:, val2), F%LC(:, hole), &
                      at, continuumAO, RI(:))
                    A2 = AugerTransitionAmplitude(P, B%info, F%LC(:, val2), F%LC(:, val1), F%LC(:, hole), &
                      at, continuumAO, RI(:))
                  end if
                  Rate1 = Rate1 + ONE*PI*(A1 + A2)**2
                  Rate3 = Rate3 + THREE*PI*(A1 - A2)**2
                end do
                ! statistical factors:
                ! multiply by the chance that this specific spin orbital pair is occupied
                ! multiply with the chance that we have a val1,val2 combination
                Rate = (Rate1 + Rate3)*stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)

                ! statistical factor for the hole
                Rate = TWO*stat_factor_not_occupied(P, F, F%N_occ, hole)*Rate

                if (P%print_level .gt. 2 .and. Rate3 .gt. ZERO) &
                  write (P%OUTPUT, "('#      Auger triplet:  ',i4,' +', i4,' -> ',i4,' + ','(',i4,',',i4,')',es13.5,' eV ', es13.5,f12.5)") &
                  val1, val2, hole, continuumAO%l, continuumAO%m, AE_energy*AU2EV, &
                  Rate3*stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)*TWO*stat_factor_not_occupied(P, F, F%N_occ, hole), &
                  stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)*TWO*stat_factor_not_occupied(P, F, F%N_occ, hole)
                if (P%print_level .gt. 2 .and. Rate1 .gt. ZERO) &
                  write (P%OUTPUT, &
                  "('#      Auger singlet:  ',i4,' +', i4,' -> ',i4,' + ','(',i4,',',i4,')',es13.5,' eV ', es13.5,f12.5)") &
                  val1, val2, hole, continuumAO%l, continuumAO%m, AE_energy*AU2EV, &
                  Rate1*stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)*TWO*stat_factor_not_occupied(P, F, F%N_occ, hole), &
                  stat_factor_pair_occupied(P, F, F%N_occ, val1, val2)*TWO*stat_factor_not_occupied(P, F, F%N_occ, hole)
              end if

              if (P%yes_auger_random) then
                if (P%yes_auger_random_2) then
                  ! throw a dice for every rate
                  Rate = Rate + Rate*random_gaussian()
                else
                  Rate = Rate + Rate*F%random_scale
                end if
                if (Rate .lt. ZERO) Rate = ZERO
              end if

              ChannelRate = ChannelRate + Rate
              HoleRate = HoleRate + Rate

            end do! m loop
          end do! l loop

          if (P%print_level .gt. 1 ) &
            write (P%OUTPUT, "('#   Auger channel:  ',i4,' +', i4,' -> ',i4,' + ',es13.5,' eV ', es13.5) ") &
            val1, val2, hole, AE_energy*AU2EV, ChannelRate
          call add_Process(P, F, S, "Auger", ChannelRate, AE_energy, hole, val1, val2)

        end do ! val2 loop
      end do ! val1 loop

      if (P%print_level .gt. 1 .and. HoleRate .ne. ZERO) &
        write (P%OUTPUT, "('#  Hole Augerrate ',es13.5)") HoleRate
      TotalRate = TotalRate + HoleRate

    end do ! hole loop

    if (P%print_level .gt. 1 &
      ) write (P%OUTPUT, "('# Total Augerrate ',es13.5,' a.u.  (lifetime= ',a,')')") TotalRate, trim(lifetime(TotalRate))
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger
      if (A(at)%charge .eq. 1) cycle
      deallocate (RI(at)%K)
      deallocate (RI(at)%n1)
      deallocate (RI(at)%l1)
      deallocate (RI(at)%n2)
      deallocate (RI(at)%l2)
      deallocate (RI(at)%n3)
      deallocate (RI(at)%l3)
      deallocate (RI(at)%l4)
      deallocate (RI(at)%R)
      if (allocated(RI(at)%Pmin)) then
        deallocate (RI(at)%Pmin)
        deallocate (RI(at)%Omin)
        deallocate (RI(at)%Smin)
        deallocate (RI(at)%min_info)
      end if
    end do

  contains
    !----------------------------------------------------------------------------
    ! print out lifetime in fs or attosec.
    !----------------------------------------------------------------------------
    function lifetime(decay_rate) result(string)
      real(long), intent(in) :: decay_rate                ! in a.u.
      real(long) :: tau
      character(len=15) :: str, string
      tau = ONE/decay_rate*AU2SEC/1e-15_long        ! to femtosec.
      if (tau .gt. ONE) then
        write (str, "(f12.1,' fs')") tau
      else
        tau = tau*1000_long                       ! to attosec.
        write (str, "(f12.1,' as')") tau
      end if
      string = trim(adjustl(str))
    end function lifetime
  end subroutine calcAugerRates

  ! This function returns the Augertransitionamplitudes (teint) for molecular orbitals
  ! i.e. <val1,val2 | hole, continuumsAO>
  function AugerTransitionAmplitude(P, Binfo, LCval1, LCval2, LChole, at, continuumAO, RI) result(amplitude)
    type(Param), intent(in) :: P
    integer, intent(in) :: Binfo(:, :)
    real(long) :: LCval1(:), LCval2(:), LChole(:)
    integer, intent(in) :: at
    type(atomorbital), intent(in) :: continuumAO
    type(RadialIntegrals), dimension(P%N_Nuc), intent(in) :: RI(:)
    real(long) :: amplitude
    integer :: bf1, bf2, bf3

    amplitude = ZERO
    ! loop over all basis function combinations on the atom
    do bf1 = 1, size(LCval1, 1)
      if (Binfo(bf1, 1) .ne. at) cycle
      if (abs(LCval1(bf1)) .lt. P%EPS_Auger) cycle
      do bf2 = 1, size(LCval2, 1)
        if (Binfo(bf2, 1) .ne. at) cycle
        if (abs(LCval2(bf2)) .lt. P%EPS_Auger) cycle
        do bf3 = 1, size(LChole, 1)
          if (Binfo(bf3, 1) .ne. at) cycle
          if (abs(LChole(bf3)) .lt. P%EPS_Auger) cycle
          if (abs(LCval1(bf1)*LCval2(bf2)*LChole(bf3)) .lt. P%EPS_Auger) cycle

          ! prescreen according to selection rules
          ! we have sum_l (lm l_1m_1 l_2m_2) (lm l_3m_3 l_4m_4)
          ! where (lm l_1m_1 l_2m_2) are real Gaunt coefficient
          ! Selection rules:
          ! 1. the sum of l+l_1+l_2 and the sum of l+l_3+l_4 must be even or odd
          ! => mod(l_1+l_2,2) == mod(l_3+l_4,2)
          if (mod(Binfo(bf1, 4) + Binfo(bf2, 4), 2) .ne. mod(Binfo(bf3, 4) + continuumAO%l, 2)) cycle
          ! 2. the m's must sum up to zero with any combination of alternating signs
          if ( &
            abs(Binfo(bf1, 5) + Binfo(bf2, 5)) .ne. abs(Binfo(bf3, 5) + continuumAO%m) .and. &
            abs(Binfo(bf1, 5) - Binfo(bf2, 5)) .ne. abs(Binfo(bf3, 5) + continuumAO%m) .and. &
            abs(Binfo(bf1, 5) + Binfo(bf2, 5)) .ne. abs(Binfo(bf3, 5) - continuumAO%m) .and. &
            abs(Binfo(bf1, 5) - Binfo(bf2, 5)) .ne. abs(Binfo(bf3, 5) - continuumAO%m)) cycle
          ! 3. the l's must fullfill the triangle relations
          ! |l_1-l_2| <= l <= |l_1 + l_2|
          ! |l_3-l_4| <= l <= |l_3 + l_4|
          ! => |l_1 - l_2| <= |l_3 + l_4| && |l_3 - l_4| <= |l_1 + l_2|
          if ( &
            abs(Binfo(bf1, 4) - Binfo(bf2, 4)) .gt. abs(Binfo(bf3, 4) + continuumAO%l) .or.&
            abs(Binfo(bf1, 4) + Binfo(bf2, 4)) .lt. abs(Binfo(bf3, 4) - continuumAO%l)) cycle
          amplitude = amplitude + LCval1(bf1)*LCval2(bf2)*LChole(bf3)* &
            AugerTransitionAmplitudeBF(P, Binfo, bf1, bf2, bf3, continuumAO, RI(:))
        end do
      end do
    end do
    return
  end function AugerTransitionAmplitude

  ! This function returns the Augertransitionamplitudes (teint) for basis functions
  ! i.e. <bf1,bf2 | bf3, ao4>
  ! TODO: vectorize this function?
  function AugerTransitionAmplitudeBF(P, Binfo, bf1, bf2, bf3, ao4, RI) result(amplitude)
    type(Param), intent(in) :: P
    integer, intent(in) :: Binfo(:, :)
    integer, intent(in) :: bf1, bf2, bf3
    type(atomorbital), intent(in) :: ao4
    type(RadialIntegrals), dimension(P%N_Nuc), intent(in) :: RI(:)

    type(atomorbital) :: ao1, ao2, ao3

    real(long) :: amplitude

    integer :: at, i

    at = Binfo(bf3, 1)

    ao1%n = Binfo(bf1, 3)
    ao2%n = Binfo(bf2, 3)
    ao3%n = Binfo(bf3, 3)

    ao1%l = Binfo(bf1, 4)
    ao2%l = Binfo(bf2, 4)
    ao3%l = Binfo(bf3, 4)

    ao1%m = Binfo(bf1, 5)
    ao2%m = Binfo(bf2, 5)
    ao3%m = Binfo(bf3, 5)

!    auger_file_nuc = trim( ATOMIC_NAME(P%charge_nuc(atom)) ) // ".auger"
!    call readRIntegral( P, auger_file_nuc, RI )

    ! this if clauses in the loop are terribly expensive
    amplitude = ZERO
    do i = 1, RI(at)%N
      ! if (&
      !   ao3%n /= RI(atom)%n3(i) .OR. &
      !   ao3%l /= RI(atom)%l3(i) .OR. &
      !   ao2%n /= RI(atom)%n2(i) .OR. &
      !   ao2%l /= RI(atom)%l2(i) .OR. &
      !   ao1%n /= RI(atom)%n1(i) .OR. &
      !   ao1%l /= RI(atom)%l1(i) .OR. &
      !   ao4%l /= RI(atom)%l4(i) &
      !   ) cycle
      ! seperate if clauses are faster
      ! look for the integral n1==RI%n1
      if (ao1%n .ne. RI(at)%n1(i)) cycle
      ! look for the integral n2==RI%n2
      if (ao2%n .ne. RI(at)%n2(i)) cycle
      ! look for the integral n3==RI%n3
      if (ao3%n .ne. RI(at)%n3(i)) cycle
      ! look for the integral l1==RI%l1
      if (ao1%l .ne. RI(at)%l1(i)) cycle
      ! look for the integral l2==RI%l2
      if (ao2%l .ne. RI(at)%l2(i)) cycle
      ! look for the integral l3==RI%l3
      if (ao3%l .ne. RI(at)%l3(i)) cycle
      ! look for the integral l4==RI%l4
      if (ao4%l .ne. RI(at)%l4(i)) cycle
      amplitude = amplitude + RI(at)%R(i)*RealAngularIntegral(RI(at)%K(i), ao1, ao2, ao3, ao4)

    end do
  end function AugerTransitionAmplitudeBF

  function AugerPhase(q, l, E, RI) result(delta)
    real(long), intent(in) :: Q
    integer, intent(in) :: l
    real(long), intent(in) :: E

    type(RadialIntegrals), intent(in) :: RI
    integer :: i
    real(long) :: delta
    delta = 11111._long
    do i = 1, RI%N
      if (l .eq. RI%l4(i) .and. RI%delta(i) .lt. TWO*PI) then
        delta = RI%delta(i)
      end if
    end do
    if (delta .lt. TWO*PI) return
!    write(STDERR,*) 'Coulomb Phase!'
    delta = coulomb_phase(q, l, sqrt(2*E))
    return
  end function AugerPhase

  ! This functions reads in the Atomic Radial Auger Integrals from the xatom output
  subroutine readRIntegral(filename, RI)
    !    type(Param),intent(in) :: P
    character(len=512), intent(in) :: filename

    type(RadialIntegrals), intent(inout) :: RI
    integer ::  j, RWStatus, num_lines
    integer :: K
    integer :: n1
    integer :: l1
    integer :: n2
    integer :: l2
    integer :: n3
    integer :: l3
    integer :: l4
    real(long) :: R
    character(len=512) :: rest, rest2
    integer :: i
    integer :: num_min, num_bf
    num_lines = 0
    open (unit=31, file=filename, IOSTAT=RWStatus, status='OLD', ACTION='READ')
    ! count the number of lines
    if (RWStatus .eq. 0) then
      num_lines = 0
      READ (31, '(a)', IOSTAT=RWStatus) rest
      if (RWStatus .eq. 0 .and. rest(1:5) .eq. "basis") then
        read (31, *) num_min, num_bf

        allocate (RI%Pmin(num_min, num_bf))
        allocate (RI%min_info(num_min, 5))
        do i = 1, num_min
          read (31, *) RI%min_info(i, 1), &
            RI%min_info(i, 3), &
            RI%min_info(i, 4), &
            RI%min_info(i, 5), &
            RI%Pmin(i, 1:num_bf)
        end do
      else
        rewind (31)
      end if
      do
        READ (31, "(x,i1,3(x,2i2),x,i2,es13.5,a)", IOSTAT=RWStatus) K, n1, l1, n2, l2, n3, l3, l4, R, rest
        IF (RWStatus .eq. 0) THEN
          !             write(P%OUTPUT,"(x,i1,3(x,2i2),x,i2,es13.5,a)") K, n1, l1, n2, l2, n3, l3, l4, R, rest
          num_lines = num_lines + 1
        else
          exit
        end if
      end do
    end if
    RI%N = num_lines
    allocate (RI%K(num_lines))
    allocate (RI%n1(num_lines))
    allocate (RI%l1(num_lines))
    allocate (RI%n2(num_lines))
    allocate (RI%l2(num_lines))
    allocate (RI%n3(num_lines))
    allocate (RI%l3(num_lines))

    allocate (RI%l4(num_lines))
    allocate (RI%R(num_lines))
    allocate (RI%delta(num_lines))
    if (num_lines .gt. 0) then
      do j = 1, num_lines + 1
        backspace (unit=31)
      end do
    end if
    do j = 1, num_lines
      READ (31, "(x,i1,3(x,2i2),x,i2,es13.5,a)", IOSTAT=RWStatus) &
        RI%K(j), RI%n1(j), RI%l1(j), RI%n2(j), RI%l2(j), RI%n3(j), RI%l3(j), RI%l4(j), RI%R(j), rest
      read (rest, '(es13.5,a)', IOSTAT=RWStatus) RI%delta(j), rest2
      if (RWStatus .ne. 0) then
        ! in case the phase is not provided in .auger file, set the phase to something very large
        rest2 = rest
        RI%delta(j) = 11111._long
      end if
    end do
    close (31)
  end subroutine readRIntegral

  subroutine calcAugerRates_spin(P, A, B, F, S, auger_fpath)
    !! calculates Auger-Meitner decay transitions based on orbital energy differences resolving individual spin configurations    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S

    integer, parameter :: MAX_SPIN = 20
    integer :: hole, val1, val2 ! molecular orbital indices
    integer :: i, j ! counter
    integer :: m, l ! counter for angular momentum
    integer :: at ! counter for atoms
    integer :: lmax ! this is the maximum angular momentum in the continuum which we consider
    real(long) :: AvgTotalRate, ChannelRate, Amplitude, HighSpinRate ! Rates & Amplitudes

    real(long) :: confAvgTotalRate(MAX_SPIN)
    integer :: confpermultp(MAX_SPIN)
    type(atomorbital) :: continuumAO ! continuum orbital parameter structure
    type(RadialIntegrals), dimension(P%N_nuc) :: RI ! Radialintegral structure to be filled by xatom output
    character(len=512) :: auger_file_nuc ! filename
    !! -- Mar.10.2015 : read-in specified fpath ---
    character*(*), optional :: auger_fpath(:)       !! Currently we use char*(128) in MC code

    real(long) :: AE_energy ! Auger electron energy
    integer, allocatable :: final_occ(:) ! final occupation
    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer, allocatable :: final_csfs(:, :) ! final spin configurations
    integer, allocatable :: final_dets(:, :) ! final determinants
    integer, allocatable :: initial_csfs(:, :) ! initial configurations
    integer, allocatable :: initial_dets(:, :) ! initial determinants
    integer :: initial_multp ! initial multp
    integer :: Nic ! counter for initial configurations
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: final_two_M, final_two_S, initial_two_S, initial_two_M ! final and intial spin quantum numbers
    integer :: continuum_spin ! quantum number of continuum spin
    integer :: only_left(2), only_left_spin(2) ! array for orbital indices for determinant comparison
    integer :: only_right(2), only_right_spin(2) ! array for spin indices for determinant comparison
    integer :: coef_sign ! sign for determinant comparisno
    real(long) :: coef ! coeficient for CSF expansion

    integer, parameter :: MAX_TEINTS = 1000 ! maximum number of two electron integrals that
    ! can contribute to one pair of initial and final configuration
    type(teint_contributions), allocatable :: teints(:) ! array of teints structure
    integer :: Nteints ! counter for teints
    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: io ! index for initial occupation
    integer :: irrep_occ

    real(long), allocatable :: TotalRate(:) ! rates between initial conf to final conf
    real(long), allocatable :: rateic2fc(:, :) ! rates between initial conf to final conf
    real(long), allocatable :: ampl_12(:, :) ! rates between initial conf to final conf
    real(long), allocatable :: ampl_21(:, :) ! rates between initial conf to final conf
    real(long) :: WE_factor
    real(long), allocatable :: coef_i(:, :) ! coeficient for CSF expansion
    real(long), allocatable :: coef_f(:, :) ! coeficient for CSF expansion

    if (P%yes_MCSCF) then
      stop "with MCSCF use additional -process_CI option"
    end if

    allocate (final_occ(size(F%N_occ)))
    ! in case of MO occupation averaging be render all occupation patterns; otherwise just take F%N_occ
    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if

    allocate (teints(MAX_TEINTS))
    ! read all radial integrals
    lmax = 0
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger file
      if (A(at)%charge .eq. 1) cycle
      if (present(auger_fpath)) then              ! extension Mar.10.2015
        auger_file_nuc = ""                        !
        auger_file_nuc = trim(auger_fpath(at))   !
      else                                       !
        auger_file_nuc = A(at)%auger_file_nuc
      end if
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Reading Auger Radial integrals from file ',a)") trim(auger_file_nuc)
      end if
      call readRIntegral(auger_file_nuc, RI(at))
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# atom: ',i4,' ',a, ' : ',i4,' entries')") &
          at, trim(auger_file_nuc), size(RI(at)%R)
      end if
      lmax = max(lmax, maxval(RI(at)%l4))
      ! prepare the projection of MO's into minimal basis set
      if (allocated(RI(at)%Pmin)) then
        allocate (RI(at)%Smin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        forall (j=1:size(RI(at)%Pmin, 1)) RI(at)%Smin(j, j) = ONE
        allocate (RI(at)%Omin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        RI(at)%Omin = matmul(RI(at)%Pmin, F%LC)
      else if (P%yes_gaussian) then
        allocate (RI(at)%Omin(size(F%Omin, 1), size(F%Omin, 2)))
        allocate (RI(at)%Smin(size(F%Smin, 1), size(F%Smin, 2)))
        allocate (RI(at)%min_info(size(F%min_info, 1), size(F%min_info, 2)))
        RI(at)%Omin = F%Omin
        RI(at)%Smin = F%Smin
        RI(at)%min_info = F%min_info
      end if! if Pmin allocated
      if (P%yes_verbose .and. allocated(RI(at)%min_info)) then
        call print_projection_info(RI(at)%min_info, RI(at)%Omin, RI(at)%Smin, at, P%OUTPUT)
      end if! yes_verbose
    end do! atom

    Nic = 0
    AvgTotalRate = ZERO
    ConfAvgTotalRate(:) = ZERO
    Confpermultp(:) = 0

    do io = 1, size(initial_occ, 2)
      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        Confpermultp(initial_multp) = Confpermultp(initial_multp) + size(initial_csfs, 2)
        deallocate (initial_csfs)
      end do
    end do
    ! loop over initial occupation
    do io = 1, size(initial_occ, 2)
      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      irrep_occ = 1
      do i = 1, size(initial_occ(:, io))
        if (P%yes_sym .and. initial_occ(i, io) .eq. &
          1) irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
      end do
      if (P%print_level .gt. 1 .or. P%yes_verbose) then
        write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)
        if (P%yes_sym) &
          write (P%OUTPUT, "('# irrep: ',i4)") irrep_occ
      end if
      ! loop over all possible initial state spin multiplicities
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        ! prepare the representations of csf and det according to genealogical coupling scheme
        ! for the initial state configuration(s)
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        call Nopen2dets(initial_Nopen, initial_multp, initial_dets)
        allocate (coef_i(size(initial_dets, 2), size(initial_csfs, 2)))
        do ic = 1, size(initial_csfs, 2)
          if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
            if (ic .ne. F%csf_num) cycle
          end if
          do id = 1, size(initial_dets, 2)
            coef_i(id, ic) = csfdet(initial_csfs(:, ic), initial_dets(:, id))
          end do
        end do

        if (P%print_level .gt. 1 .or. P%yes_verbose) then
          write (P%OUTPUT, "('# Initial Spin multiplicity ',i3,' has ',i3, ' spin configurations:')") &
            initial_multp, size(initial_csfs, 2)
          if (P%print_level .gt. 4 .or. P%yes_verbose) then
            do j = 1, size(initial_csfs, 2)
              write (P%OUTPUT, "('# ',a )") csfstring(initial_occ(:, io), initial_csfs(:, j))
            end do
          end if
        end if

        allocate (TotalRate(size(initial_csfs, 2)))
        TotalRate(:) = ZERO

        ! Loop over hole and valence pair combinations
        do hole = 1, size(initial_occ(:, io))
          !            if( .NOT. has_hole(P,F,initial_occ(:,io),hole) )cycle
          if (initial_occ(hole, io) .eq. 2) cycle
          do val1 = 1, size(initial_occ(:, io))
            if (initial_occ(val1, io) .eq. 0) cycle
            !              if( .NOT. is_occupied(P,F,initial_occ(:,io),val1) ) cycle
            do val2 = val1, size(initial_occ(:, io))
              if (initial_occ(val2, io) .eq. 0) cycle
              if (val1 .eq. val2 .and. initial_occ(val1, io) .lt. 2) cycle
              !                if( .NOT. is_pair_occupied(P,F,initial_occ(:,io),val1,val2) ) cycle
              ! Auger electron energy is derived from the mo orbital energy difference
              AE_energy = F%CE(val1) + F%CE(val2) - F%CE(hole)
              if (AE_energy .lt. 0) cycle

              allocate (ampl_12(size(A), (lmax + 1)**2))
              allocate (ampl_21(size(A), (lmax + 1)**2))
              do l = 0, lmax
                continuumAO%l = l
                do m = -l, l
                  continuumAO%m = m
                  ! incoherently sum over different atom contribution
                  do at = 1, P%N_nuc
                    ! special case for hydrogen: no auger
                    if (A(at)%charge .eq. 1) cycle
                    if (allocated(RI(at)%Omin)) then
                      ampl_12(at, l*(l + 1) + m + 1) = &
                        AugerTransitionAmplitude(P, RI(at)%min_info, RI(at)%Omin(:, val1), RI(at)%Omin(:, val2), &
                        RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                      ampl_21(at, l*(l + 1) + m + 1) = &
                        AugerTransitionAmplitude(P, RI(at)%min_info, RI(at)%Omin(:, val2), RI(at)%Omin(:, val1), &
                        RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                    else
                      ampl_12(at, l*(l + 1) + m + 1) = &
                        AugerTransitionAmplitude(P, B%info, F%LC(:, val1), F%LC(:, val2), F%LC(:, hole), at, continuumAO, RI(:))
                      ampl_21(at, l*(l + 1) + m + 1) = &
                        AugerTransitionAmplitude(P, B%info, F%LC(:, val2), F%LC(:, val1), F%LC(:, hole), at, continuumAO, RI(:))
                    end if
                  end do! atom
                end do!m
              end do! l

              ! Final occupation
              final_occ = initial_occ(:, io)
              final_occ(val1) = final_occ(val1) - 1
              final_occ(val2) = final_occ(val2) - 1
              final_occ(hole) = final_occ(hole) + 1
              final_Nopen = 0
              do i = 1, size(final_occ, 1)
                if (final_occ(i) .eq. 1) final_Nopen = final_Nopen + 1
              end do
              ! Create spin configurations for the possible final spin states
              do final_multp = initial_multp - 1, initial_multp + 1, 2
                if (final_multp .eq. 0) cycle
                if (final_multp - 1 .gt. final_Nopen) cycle

                ! We calculate transition rate to high spin state
                ! Apply Wigner Eckert to get factor for transition rates to other MS configurations:
                initial_two_S = initial_multp - 1
                initial_two_M = initial_two_S
                final_two_S = final_multp - 1
                WE_factor = ZERO
                do final_two_M = -final_two_S, final_two_S, 2
                  WE_factor = WE_factor + &
                    cleb(final_two_S, final_two_M, 1, initial_two_S - final_two_M, initial_two_S, initial_two_S)**2 &
                    /cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2
                end do

                call Nopen2csfs(final_Nopen, final_multp, final_csfs)
                call Nopen2dets(final_Nopen, final_multp, final_dets)
                allocate (coef_f(size(final_dets, 2), size(final_csfs, 2)))
                do fc = 1, size(final_csfs, 2)
                  do fd = 1, size(final_dets, 2)
                    coef_f(fd, fc) = csfdet(final_csfs(:, fc), final_dets(:, fd))
                  end do
                end do

                ! Loop over configurations within spin multiplicity and occupation
                allocate (rateic2fc(size(initial_csfs, 2), size(final_csfs, 2)))
                rateic2fc(:, :) = ZERO

                ! There are only two types of integrals that can appear in this setup:
                ! <val1 val2 | hole k>
                ! or <val2 val1 | hole k>
                Nteints = 2
                teints(1)%val1 = val1
                teints(1)%val2 = val2
                teints(1)%hole = hole
                teints(1)%coef = ZERO
                teints(2)%val1 = val2
                teints(2)%val2 = val1
                teints(2)%hole = hole
                teints(2)%coef = ZERO

                ! Loop over initial configurations
                do ic = 1, size(initial_csfs, 2)
                  if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                    if (ic .ne. F%csf_num) cycle
                  end if
                  if (P%print_level .gt. 2) &
                    write (P%OUTPUT, "('# initial configuration with multiplicity',i3,2x,':',a)") &
                    initial_multp, trim(csfstring(initial_occ(:, io), initial_csfs(:, ic)))
                  if (P%print_level .gt. 3) then
                    do id = 1, size(initial_dets, 2)
                      write (P%OUTPUT, *) '# ', &
                        csfdet(initial_csfs(:, ic), initial_dets(:, id)), &
                        trim(detstring(initial_occ(:, io), initial_dets(:, id)))
                    end do
                  end if
                  do fc = 1, size(final_csfs, 2)
                    if (P%print_level .gt. 3) &
                      write (P%OUTPUT, "('# ',2x,'-> final configuration ',a,1x, 'multiplicity ',i3)") &
                      trim(csfstring(final_occ, final_csfs(:, fc))), final_multp
                    if (P%print_level .gt. 4) then
                      do fd = 1, size(final_dets, 2)
                        write (P%OUTPUT, *) '# ', &
                          csfdet(final_csfs(:, fc), final_dets(:, fd)), &
                          detstring(final_occ, final_dets(:, fd))
                      end do
                    end if
                    teints(1)%coef = ZERO
                    teints(2)%coef = ZERO

                    ! Loop over determinant pairs
                    do id = 1, size(initial_dets, 2)
                      do fd = 1, size(final_dets, 2)
                        if (P%print_level .gt. 4) then
                          write (P%OUTPUT, *) '#   ', &
                            csfdet(initial_csfs(:, ic), initial_dets(:, id)), &
                            trim(detstring(initial_occ(:, io), initial_dets(:, id)))
                          write (P%OUTPUT, *) '# ->', &
                            csfdet(final_csfs(:, fc), final_dets(:, fd)), &
                            trim(detstring(final_occ, final_dets(:, fd)))
                        end if

                        coef = coef_f(fd, fc)*coef_i(id, ic)
                        if (coef .eq. ZERO) cycle
                        call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
                          initial_occ(:, io), initial_dets(:, id), final_occ, final_dets(:, fd))
                        if (P%print_level .gt. 4) then
                          write (P%OUTPUT, *) '# ', coef_i(id, ic), coef_f(fd, fc), coef_sign
                        end if
                        coef = coef*coef_sign
                        if (coef_sign .ne. 0 .and. count(only_left(:) .ne. 0) .eq. 2 .and. count(only_right(:) .ne. 0) .eq. 1) then
                          continuum_spin = sum(only_left_spin) - sum(only_right_spin)
                          if (abs(continuum_spin) .eq. 1) then
                            if (Nteints .gt. MAX_TEINTS) then
                              write (P%OUTPUT, *) '# too much teints! increase MAX_TEINTS'
                              stop
                            end if
                            if ( &
                              only_left_spin(1) .eq. only_right_spin(1) .and. &
                              only_left_spin(1) .eq. only_left_spin(2) .and. &
                              only_left_spin(1) .eq. continuum_spin) then
                              ! a a -> a a
                              ! b b -> b b
                              if (only_left(1) .eq. val1 .and. only_left(2) .eq. val2) then
                                teints(1)%coef = teints(1)%coef + coef
                                teints(2)%coef = teints(2)%coef - coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    coef, teints(1)%val1, teints(1)%hole, teints(1)%val2, 'k'
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    - coef, teints(2)%val1, teints(2)%hole, teints(2)%val2, 'k'
                                end if
                              else if (only_left(1) .eq. val2 .and. only_left(2) .eq. val1) then
                                teints(2)%coef = teints(2)%coef + coef
                                teints(1)%coef = teints(1)%coef - coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    - coef, teints(1)%val1, teints(1)%hole, teints(1)%val2, 'k'
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    coef, teints(2)%val1, teints(2)%hole, teints(2)%val2, 'k'
                                end if
                              end if
                            else if ( &
                              only_left_spin(1) .eq. only_right_spin(1) .and. &
                              only_left_spin(2) .eq. continuum_spin) then
                              ! a b -> a b
                              ! b a -> b a
                              if (only_left(1) .eq. val1 .and. only_left(2) .eq. val2) then
                                teints(1)%coef = teints(1)%coef + coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    coef, teints(1)%val1, teints(1)%hole, teints(1)%val2, 'k'
                                end if
                              else if (only_left(1) .eq. val2 .and. only_left(2) .eq. val1) then
                                teints(2)%coef = teints(2)%coef + coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    coef, teints(2)%val1, teints(2)%hole, teints(2)%val2, 'k'
                                end if
                              end if
                            else if ( &
                              only_left_spin(2) .eq. only_right_spin(1) .and. &
                              only_left_spin(1) .eq. continuum_spin) then
                              ! a b -> b a
                              ! b a -> a b
                              if (only_left(1) .eq. val1 .and. only_left(2) .eq. val2) then
                                teints(2)%coef = teints(2)%coef - coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    - coef, teints(2)%val1, teints(2)%hole, teints(2)%val2, 'k'
                                end if
                              else if (only_left(1) .eq. val2 .and. only_left(2) .eq. val1) then
                                teints(1)%coef = teints(1)%coef - coef
                                if (P%print_level .gt. 4) then
                                  write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='yes') &
                                    - coef, teints(1)%val1, teints(1)%hole, teints(1)%val2, 'k'
                                end if
                              end if
                            end if
                          end if
                        end if
                      end do ! id
                    end do ! fd

                    if (P%print_level .gt. 2) then
                      write (P%OUTPUT, "('# ',10x,a2)", advance='no') '|'
                      do i = 1, Nteints
                        write (P%OUTPUT, "('+', es10.2,'* (',i2,i2,'|',i2,a2, ')')", advance='no') &
                          sqrt(WE_factor)*teints(i)%coef, teints(i)%val1, teints(i)%hole, teints(i)%val2, 'k'
                      end do
                      write (P%OUTPUT, *) '|^2'
                    end if

                    continuumAO%n = 99
                    ! loop over all continuum channel
                    HighSpinRate = ZERO
                    do l = 0, lmax
                      continuumAO%l = l
                      do m = -l, l
                        continuumAO%m = m
                        ! incoherently sum over different atom contribution
                        do at = 1, P%N_nuc
                          Amplitude = ZERO
                          ! special case for hydrogen: no auger
                          if (A(at)%charge .eq. 1) cycle
                          Amplitude = teints(1)%coef*ampl_12(at, l*(l + 1) + m + 1) + &
                            teints(2)%coef*ampl_21(at, l*(l + 1) + m + 1)
                          HighSpinRate = HighSpinRate + TWO*PI*Amplitude**2
                        end do! atom
                      end do!m
                    end do!l

                    ChannelRate = HighSpinRate*WE_factor
                    if (P%print_level .gt. 2 &
                      ) write (P%OUTPUT, "('# ',5x,'channel rate' ,es15.5)") ChannelRate
                    rateic2fc(ic, fc) = ChannelRate
                  end do ! fc final configuration
                end do! ic

                if (P%yes_average_rates .or. (P%yes_average_csf_rates .and. F%multp .eq. initial_multp)) then
                  call add_Process(P, F, S, "Auger", sum(rateic2fc(:, :))/size(rateic2fc, 1), AE_energy, hole, val1, val2, &
                    initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=1, &
                    final_multp=final_multp, final_csf_num=1)
                else
                  ic = F%csf_num
                  if (F%multp .eq. initial_multp) then
                    do fc = 1, size(final_csfs, 2)
                      if (rateic2fc(ic, fc) .eq. ZERO) cycle
                      call add_Process(P, F, S, "Auger", rateic2fc(ic, fc), AE_energy, hole, val1, val2, &
                        initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=ic, &
                        final_multp=final_multp, final_csf_num=fc)
                    end do
                  end if
                end if

                do ic = 1, size(initial_csfs, 2)
                  if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                    if (ic .ne. F%csf_num) cycle
                  end if
                  TotalRate(ic) = TotalRate(ic) + sum(rateic2fc(ic, :))
                end do

                deallocate (rateic2fc)
                deallocate (final_csfs)
                deallocate (final_dets)
                deallocate (coef_f)
              end do! final_multp loop
              deallocate (ampl_12)
              deallocate (ampl_21)
            end do ! val2 loop
          end do ! val1 loop
        end do ! hole loop

        if (P%print_level .gt. 1 .or. P%yes_verbose) then

          do ic = 1, size(initial_csfs, 2)
            if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
              if (ic .ne. F%csf_num) cycle
            end if
            write (P%OUTPUT, "('#  Total Augerrate for multp ',i4,' ic ',i4,' : ',es13.5,' a.u. (lifetime=',a,')')") &
              initial_multp, ic, TotalRate(ic), trim(lifetime(TotalRate(ic)))
          end do
          if (P%yes_average_csf_rates) &
            write (P%OUTPUT, "('#  Total Augerrate for multp ',i4,' avg_ic ',' : ',es13.5,' a.u. (lifetime=',a,')')") &
            initial_multp, sum(TotalRate(:))/size(TotalRate(:), 1), trim(lifetime(sum(TotalRate(:))/size(TotalRate(:), 1)))
        end if

        AvgTotalRate = AvgTotalRate + initial_multp*sum(TotalRate(:))
        Nic = Nic + initial_multp*size(TotalRate(:), 1)
        deallocate (initial_dets)
        deallocate (initial_csfs)
        deallocate (coef_i)
        deallocate (TotalRate)
      end do ! initial_multp
    end do ! io
    AvgTotalRate = AvgTotalRate/Nic

    if (P%yes_average_csf_rates) then
      if (P%print_level .gt. 1 .or. P%yes_verbose) then
        write (P%OUTPUT, "('# Average Total Augerrate ',es13.5,' a.u.  (lifetime= ',a,')')") &
          AvgTotalRate, trim(lifetime(AvgTotalRate))
        write (P%OUTPUT, "('# averaged over ',i3,' spin states' )") Nic
      end if
    end if
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger
      if (A(at)%charge .eq. 1) cycle
      deallocate (RI(at)%K)
      deallocate (RI(at)%n1)
      deallocate (RI(at)%l1)
      deallocate (RI(at)%n2)
      deallocate (RI(at)%l2)
      deallocate (RI(at)%n3)
      deallocate (RI(at)%l3)
      deallocate (RI(at)%l4)
      deallocate (RI(at)%R)
      if (allocated(RI(at)%Pmin)) then
        deallocate (RI(at)%Pmin)
        deallocate (RI(at)%Omin)
        deallocate (RI(at)%min_info)
      end if
    end do
    deallocate (teints)
    deallocate (final_occ)
    deallocate (initial_occ)
  contains
    !----------------------------------------------------------------------------
    ! print out lifetime in fs or attosec.
    !----------------------------------------------------------------------------
    function lifetime(rate) result(string)
      real(long), intent(in) :: rate                ! in a.u.
      real(long) :: tau
      character(len=15) :: str, string
      tau = ONE/rate*AU2SEC/1e-15_long        ! to femtosec.
      if (tau .gt. ONE) then
        write (str, "(f12.1,' fs')") tau
      else
        tau = tau*1000_long                       ! to attosec.
        write (str, "(f12.1,' as')") tau
      end if
      string = trim(adjustl(str))
    end function lifetime
  end subroutine calcAugerRates_spin

  subroutine calcAugerRates_CI(P, A, B, F, G, auger_fpath)
    !! calculates Auger-Meitner decay transitions based on configuration interaction calculation    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    type(Grid3D), intent(in) :: G

    type(WF_data) :: F_final
    !type(Basis) :: B_final

    integer :: hole, val1, val2 ! molecular orbital indices
    integer :: virt, part
    integer :: i, j, k ! counter
    integer :: m, l ! counter for angular momentum
    integer :: at ! counter for atoms
    integer :: lmax ! this is the maximum angular momentum in the continuum which we consider
    real(long) :: AvgTotalRate, TotalRate, HoleRate, ChannelRate, Amplitude, HighSpinRate ! Rates & Amplitudes
    type(atomorbital) :: continuumAO ! continuum orbital parameter structure
    type(RadialIntegrals), dimension(P%N_nuc) :: RI ! Radialintegral structure to be filled by xatom output
    character(len=512) :: auger_file_nuc ! filename
    character(len=512) :: csfStr
    character*(*), optional :: auger_fpath(:)       !! Currently we use char*(128) in MC code
    real(long) :: AE_energy, csfRate ! Auger electron energy
    real(long) :: E_ini ! initial configuration energy
    real(long), allocatable :: csf_c2(:, :)
    integer, allocatable, target :: final_occ(:, :) ! final occupation
    integer, allocatable, target :: initial_occ(:, :) ! initial occupation
    integer, allocatable, target :: initial_occ_orig(:, :) ! initial occupation, no symmetry restrictions
    integer, allocatable :: flag_final_csfs(:) ! final csf flag
    integer :: initial_multp ! initial spin multp
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: final_two_M, final_two_S, initial_two_S, initial_two_M ! final and intial spin quantum numbers
    integer :: continuum_spin ! quantum number of continuum spin
    integer :: only_left(2), only_left_spin(2) ! array for orbital indices for determinant comparison
    integer :: only_right(2), only_right_spin(2) ! array for spin indices for determinant comparison
    integer :: coef_sign ! fermionic sign for determinant comparison
    real(long) :: coef ! coeficient for CSF expansion
    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: io2
    integer :: N_initial_csfs ! N of initial csfs
    integer :: N_final_occ ! N of final occupation
    integer :: N_final_csfs ! N of final csfs
    integer :: i_final_occ ! index for final occupation
    integer :: l1, l2, m1, m2
    integer :: f_irrep, n_irrep
    integer, allocatable :: occs_temp(:, :) ! temp occupation
    type(CSF), allocatable :: initial_csf_array(:) ! array for csfs initial state
    type(CSF), allocatable :: final_csf_array(:) ! array for csfs final state
    real(long), allocatable :: final_csf_Auger_amplitude(:, :) ! Auger amplitude for each csf and continuum channel
    real(long), allocatable :: Auger_intensity_ang_dist(:) ! Auger intensity for each l,m channel
    real(long), allocatable :: auger_teint_coef(:, :, :) ! coefficients for Auger two-electron-integrals
    real(long), allocatable :: final_CI_matrix(:, :) ! CI matrix for final states
    real(long), allocatable :: final_CI_energy(:) ! energy of final states
    real(long), allocatable :: initial_CI_matrix(:, :) ! CI matrix for initial states
    real(long), allocatable :: initial_CI_energy(:) ! energy of initial states
    real(long), allocatable :: opdm(:, :) ! one-particle density matrix of final state
    real(long), allocatable :: n_occ(:) ! natural occupation numbers of final state
    real(long), allocatable :: charge(:) ! mulliken charges of final state
    integer :: Ncore ! number of core orbitals in the molecule
    logical :: redundant_occ ! number of redundant occupations
    type(UmatrixTable) :: UTini, UTfin
    real(long), allocatable :: CI_diagonal(:)
    integer, allocatable :: core(:)
    integer, allocatable :: occupied(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    integer :: NOpenMax, nintern
    real(long), allocatable :: csf_O(:, :) ! Auger amplitude for each csf and continuum channel
    real(long), allocatable :: final_csf_Auger_amplitude_temp(:) ! Auger amplitude for each csf and continuum channel
    integer, allocatable :: nz_el(:)
    real(long) :: delta_l1, delta_l2, gaunt_int, Coul_charge
    real(long), allocatable :: Amplitude_ang_dist(:)
    integer :: ini_state
    integer :: nev
    real :: timer
    real :: time_last_step
    integer*8 :: int_time      ! timer
    integer, allocatable :: idx(:)
    ! read all radial integrals
    lmax = 0
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger file
      if (A(at)%charge .eq. 1) cycle
      if (present(auger_fpath)) then              ! extension Mar.10.2015
        auger_file_nuc = ""                        !
        auger_file_nuc = trim(auger_fpath(at))   !
      else                                       !
        auger_file_nuc = A(at)%auger_file_nuc
      end if
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Reading Auger Radial integrals from file ',a)") trim(auger_file_nuc)
      end if
      call readRIntegral(auger_file_nuc, RI(at))
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# atom: ',i4,' ',a, ' : ',i4,' entries')") &
          at, trim(auger_file_nuc), size(RI(at)%R)
      end if
      lmax = max(lmax, maxval(RI(at)%l4))

      ! prepare the projection of MO's into minimal basis set
      if (allocated(RI(at)%Pmin)) then
        allocate (RI(at)%Smin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        forall (j=1:size(RI(at)%Pmin, 1)) RI(at)%Smin(j, j) = ONE
        allocate (RI(at)%Omin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        RI(at)%Omin = matmul(RI(at)%Pmin, F%LC)
      else if (P%yes_gaussian) then
        allocate (RI(at)%Omin(size(F%Omin, 1), size(F%Omin, 2)))
        allocate (RI(at)%Smin(size(F%Smin, 1), size(F%Smin, 2)))
        allocate (RI(at)%min_info(size(F%min_info, 1), size(F%min_info, 2)))
        RI(at)%Omin = F%Omin
        RI(at)%Smin = F%Smin
        RI(at)%min_info = F%min_info
      end if! if Pmin allocated
      if (P%yes_verbose .and. allocated(RI(at)%min_info)) then
        call print_projection_info(RI(at)%min_info, RI(at)%Omin, RI(at)%Smin, at, P%OUTPUT)
      end if! yes_verbose
    end do! atom
    AvgTotalRate = ZERO
    ini_state = F%state
    Ncore = F%Ncore
    ! in case of MO occupation averaging we render all occupation patterns; otherwise just take F%N_occ

    if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, initial_occ)
    else if (allocated(P%active_space_MO)) then
      call active_space_occs(F%N_occ, P%active_space_MO, initial_occ)
    else if (P%yes_process_cis_ini) then
      allocate (initial_occ(size(F%N_occ), count(F%N_occ(F%Ncore + 1:) .gt. 0)*count(F%N_occ(F%Ncore + 1:) .lt. 2) + 1))
      initial_occ(:, 1) = F%N_occ(:)
      i = 1
      do hole = F%Ncore + 1, size(F%N_occ, 1)
        if (F%N_occ(hole) .eq. 0) cycle
        do part = F%Ncore + 1, size(F%N_occ, 1)
          if (F%N_occ(part) .eq. 2) cycle
          initial_occ(:, i + 1) = F%N_occ(:)
          initial_occ(hole, i + 1) = initial_occ(hole, i + 1) - 1
          initial_occ(part, i + 1) = initial_occ(part, i + 1) + 1
          i = i + 1
        end do
      end do
    else
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    end if

    ! initial_occ_orig = initial_occ
    allocate (initial_occ_orig, source=initial_occ)
    ! filter symmetry
    if (P%yes_sym .and. F%state_irrep .gt. 0) then
      deallocate (initial_occ)
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, initial_occ_orig, initial_occ)
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Number of configurations before symmetry:',i6)") size(initial_occ_orig, 2)
        write (P%OUTPUT, "('# Number of configurations after symmetry :',i6)") size(initial_occ, 2)
      end if
    end if

    initial_multp = F%multp
    write (P%OUTPUT, "('# Initial configurations for spin multp: ',i4)") initial_multp
    call construct_csf_array(P, B, F, initial_multp, P%state_irrep, initial_occ, initial_csf_array)
    N_initial_csfs = size(initial_csf_array)
    allocate (initial_CI_matrix(N_initial_csfs, N_initial_csfs))
    allocate (initial_CI_energy(N_initial_csfs))
    write (P%OUTPUT, "('# N_initial_csfs: ',i7)") N_initial_csfs
    write (P%OUTPUT, '("# computing CI matrix")')
    NOpenMax = maxval(count(initial_occ .eq. 1, dim=1))
    call constructUMatrixTable(NOpenMax, initial_multp, UTini)

    ! determine active/core/virtual MOs
    allocate (active(count((/((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0)), j=1, size(initial_occ, 1))/))))
    allocate (core(count((/((all(initial_occ(j, :) .eq. 2)), j=1, size(initial_occ, 1))/))))
    allocate (virtual(count((/(all(initial_occ(j, :) .eq. 0), j=1, size(initial_occ, 1))/))))
    allocate (occupied(size(active, 1) + size(core, 1)))
    active = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0))), j=1, size(initial_occ, 1))/))
    core = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(all(initial_occ(j, :) .eq. 2), j=1, size(initial_occ, 1))/))
    virtual = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(all(initial_occ(j, :) .eq. 0), j=1, size(initial_occ, 1))/))
    occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))
    nintern = size(active)
    do j = size(active), 1, -1
      if (any(sum(initial_occ(active(j:), :), dim=1) .gt. 1)) exit
    end do
    nintern = j
    write (6, '("#     core:",99i4)') core
    write (6, '("#   active:",99i4)') active
    write (6, '("# internal:",99i4)') active(:nintern)
    write (6, '("# external:",99i4)') active(nintern + 1:)

    allocate (F_core(size(F%LC, 2), size(F%LC, 2)))
    call construct_h0_matrix(G, B, F)
    F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)
    call calculate_F_inactive(P, B, core, F%LC, F_core)
    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do
    if (nintern .gt. 0) call transform_twoel_allactive(P, B, F, active(:nintern), P%OUTPUT)
    if (nintern .lt. size(active)) then
      if (nintern .gt. 0) then
        call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
        call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      end if
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)
      if (nintern .gt. 0) then
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        deallocate (F%teint_cdei)
      end if
    else
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix)
    end if
    if (nintern .gt. 0) deallocate (F%teint_abcd)
    write (P%OUTPUT, '("# done")')
    ! solve it
    write (P%OUTPUT, '("# solving CI matrix")')
    call get_eigenvalues(initial_CI_matrix, initial_CI_energy)
    write (P%OUTPUT, '("# done")')

    E_ini = initial_CI_energy(ini_state)

    write (P%OUTPUT, "('# Initial state ',i4,x,'spin multp=',i4,x,'energy=',f15.7)") &
      ini_state, F%multp, E_ini + nuclear_repulsion_energy(A, P)
    do ic = 1, size(initial_CI_matrix, 1)
      if (abs(initial_CI_matrix(ic, ini_state)) .lt. 0.1) cycle
      write (P%OUTPUT, '("# csf ",i7,x,f15.5,x,a)') &
        ic, initial_CI_matrix(ic, ini_state), trim(csfstring(initial_csf_array(ic)%occ, initial_csf_array(ic)%csf))
    end do

    ! Loop over final state spin
    TotalRate = ZERO
    do final_multp = initial_multp - 1, initial_multp + 1, 2
      if (final_multp .eq. 0) cycle
      write (P%OUTPUT, "('# Final Spin multiplicity ',i3)") final_multp
      if (P%yes_sym) then
        n_irrep = B%nirrep
      else
        n_irrep = 1
      end if
      do f_irrep = 1, n_irrep
        write (P%OUTPUT, "('# Final state irrep ',i3)") f_irrep
        N_final_occ = 0

        if (allocated(P%final_gas)) then
          call general_active_space_occs(size(F%N_occ), sum(F%N_occ) - 1, P%final_gas, final_occ)
          N_final_occ = size(final_occ, 2)
        else
          ! Loop over hole and valence pair combinations: count configurations
          do io2 = 1, size(initial_occ_orig, 2)
            do hole = 1, Ncore
              if (initial_occ_orig(hole, io2) .eq. 2) cycle
              do val1 = Ncore + 1, size(initial_occ_orig(:, io2))
                if (initial_occ_orig(val1, io2) .eq. 0) cycle
                if (val1 .eq. hole) cycle
                do val2 = val1, size(initial_occ_orig(:, io2))
                  if (initial_occ_orig(val2, io2) .eq. 0) cycle
                  if (val2 .eq. hole) cycle
                  if (val1 .eq. val2 .and. initial_occ_orig(val1, io2) .lt. 2) cycle
                  AE_energy = F%CE(val1) + F%CE(val2) - F%CE(hole)
                  N_final_occ = N_final_occ + 1
                end do! val2
              end do! val1
            end do! hole
          end do
          allocate (final_occ(size(F%N_occ), N_final_occ))
          write (P%OUTPUT, "('# N_final_occ ',i12)") N_final_occ
          N_final_csfs = 0
          ! write final_occ and count csfs
          i_final_occ = 0
          do io2 = 1, size(initial_occ_orig, 2)
            do hole = 1, Ncore
              if (initial_occ_orig(hole, io2) .eq. 2) cycle
              HoleRate = ZERO
              do val1 = Ncore + 1, size(initial_occ_orig(:, io2))
                if (initial_occ_orig(val1, io2) .eq. 0) cycle
                if (val1 .eq. hole) cycle
                do val2 = val1, size(initial_occ_orig(:, io2))
                  if (val2 .eq. hole) cycle
                  if (initial_occ_orig(val2, io2) .eq. 0) cycle
                  if (val1 .eq. val2 .and. initial_occ_orig(val1, io2) .lt. 2) cycle
                  AE_energy = F%CE(val1) + F%CE(val2) - F%CE(hole)
                  !if(AE_energy < 0) cycle
                  i_final_occ = i_final_occ + 1
                  ! Final occupation
                  final_occ(:, i_final_occ) = initial_occ_orig(:, io2)
                  final_occ(val1, i_final_occ) = final_occ(val1, i_final_occ) - 1
                  final_occ(val2, i_final_occ) = final_occ(val2, i_final_occ) - 1
                  final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) + 1
                  redundant_occ = .FALSE.
                  do j = 1, i_final_occ - 1
                    if (all(final_occ(:, j) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
                    if (redundant_occ) exit
                  end do
                  if (redundant_occ) then
                    i_final_occ = i_final_occ - 1
                    cycle
                  end if
                  !final_Nopen=count(final_occ(:,i_final_occ)==1)
                  !if(final_multp-1 > final_Nopen)cycle
                end do! val2
              end do! val1
            end do! hole
          end do! io2
          write (P%OUTPUT, "('# redundant occs ',i7)") N_final_occ - i_final_occ
          N_final_occ = i_final_occ
          ! add single excitations
          if (P%yes_process_CIS) then
            j = 0
            do i_final_occ = 1, size(final_occ, 2)
              j = j + count(final_occ(Ncore + 1:, i_final_occ) .gt. 0)*count(final_occ(Ncore + 1:, :) .lt. 2)
            end do
            allocate (occs_temp(size(final_occ, 1), N_final_occ + j + 1))
            occs_temp(:, 1:N_final_occ) = final_occ(:, 1:N_final_occ)
            j = N_final_occ
            do i_final_occ = 1, size(final_occ, 2)
              do val1 = Ncore + 1, size(final_occ, 1)
                do virt = val1 + 1, size(final_occ, 1)
                  if (final_occ(val1, i_final_occ) .gt. 0 .and. final_occ(virt, i_final_occ) .lt. 2) then
                    j = j + 1
                    occs_temp(:, j) = final_occ(:, i_final_occ)
                    occs_temp(val1, j) = occs_temp(val1, j) - 1
                    occs_temp(virt, j) = occs_temp(virt, j) + 1
                    ! check if spin multiplicity allows for excitation
                    final_Nopen = count(occs_temp(:, j) .eq. 1)
                    if (final_multp - 1 .gt. final_Nopen) then
                      j = j - 1
                      cycle
                    end if
                    redundant_occ = .FALSE.
                    do k = 1, j - 1
                      if (all(occs_temp(:, j) .eq. occs_temp(:, k))) redundant_occ = .TRUE.
                      if (redundant_occ) exit
                    end do
                    if (redundant_occ) then
                      j = j - 1
                      cycle
                    end if
                  end if
                end do! virt
              end do! val1
            end do! i_final_occ
            write (P%OUTPUT, "('# redundant excitations ',i7)") size(occs_temp, 2) - N_final_occ - j
            deallocate (final_occ)
            allocate (final_occ(size(occs_temp, 1), 1:j))
            final_occ(:, :) = occs_temp(:, 1:j)
            write (P%OUTPUT, *) 'Added ', j - N_final_occ, 'single excitations'
            deallocate (occs_temp)
            N_final_occ = size(final_occ, 2)
          end if! process_CIS
        end if! final_gas

        NOpenMax = maxval(count(final_occ .eq. 1, dim=1))
        call constructUMatrixTable(NOpenMax, final_multp, UTfin)

        if (P%yes_sym) then
          call filter_sym_occs(B%irrep_prod, F%irrep_MO, f_irrep, final_occ(:, :N_final_occ), occs_temp)
          write (P%OUTPUT, "('# excluded because of symmetry ',i7)") N_final_occ - size(occs_temp, 2)
        else
          allocate (occs_temp(size(final_occ, 1), 1:N_final_occ))
          occs_temp = final_occ(:, 1:N_final_occ)
        end if
        N_final_occ = size(occs_temp, 2)
        deallocate (final_occ)
        allocate (final_occ(size(occs_temp, 1), size(occs_temp, 2)))
        final_occ(:, :) = occs_temp(:, :)
        deallocate (occs_temp)

        ! fill in all spin configurations into final_csf_array
        call construct_csf_array(P, B, F, final_multp, f_irrep, final_occ, final_csf_array)
        N_final_csfs = size(final_csf_array, 1)

        if (N_final_csfs .eq. 0) then
          ! if there are no spin configurations for this spin continue with next spin multiplicity
          deallocate (final_occ)
          cycle
        end if

        allocate (flag_final_csfs(N_final_csfs))
        allocate (final_csf_Auger_amplitude((lmax + 1)**2, N_final_csfs))
        allocate (Auger_intensity_ang_dist((4*lmax + 1)**2))
        allocate (Amplitude_ang_dist((2*lmax + 1)**2))

        ! fill flag_final_csfs to prescreen configurations that are different by 4 occupations to any initial configuration
        flag_final_csfs(:) = 1
        fc = 1
        do
          k = num_csf(count(final_csf_array(fc)%occ(:) .eq. 1), final_multp)
          do ic = 1, size(initial_csf_array, 1)
            if (sum(abs(initial_csf_array(ic)%occ(:) - final_csf_array(fc)%occ(:))) .le. 3) then
              flag_final_csfs(fc:fc + k - 1) = 0
              cycle
            end if
            if (flag_final_csfs(fc) .eq. 0) exit
          end do
          fc = fc + k
          if (fc .gt. N_final_csfs) exit
        end do!

        write (P%OUTPUT, "('# N_final_occ ',i12)") N_final_occ
        write (P%OUTPUT, "('# N_final_csfs ',i12)") N_final_csfs

        write (P%OUTPUT, "('# primary   csf ',i12)") count(flag_final_csfs(:N_final_csfs) .eq. 0)
        write (P%OUTPUT, "('# secondary csf ',i12)") count(flag_final_csfs(:N_final_csfs) .eq. 1)
        ! number of states to consider
        if (P%nstates .gt. 0) then
          nev = P%nstates
        else
          nev = count(flag_final_csfs(:N_final_csfs) .eq. 0)
        end if
        nev = max(nev, 1)
        nev = min(nev, N_final_csfs)
        ! Loop over final configuration state functions
        ! Calculate for every configuration / continuum channel the Auger amplitude
        !$omp parallel private(csfRate,csfStr,fc,auger_teint_coef,ic,coef,id,fd,coef_sign,only_left,only_right,only_left_spin,only_right_spin,continuum_spin,HighSpinRate,continuumAO,val1,val2,hole,Amplitude,at,l,m) shared(final_csf_array,final_csf_Auger_amplitude,initial_csf_array,ini_state,lmax,P,RI,flag_final_csfs,initial_CI_matrix,F,B,A,final_occ) default(none)
        allocate (auger_teint_coef(size(final_occ, 1), size(final_occ, 1), size(final_occ, 1)))
        !$omp do schedule(static,1)
        do fc = 1, size(final_csf_array, 1)
          !write(P%OUTPUT,'("# csf",i5)') fc
          !do fd=1,size(final_csf_array(fc)%dets,2)
          !write(P%OUTPUT,'("# ",f7.3,x,a)') CSFDET(final_csf_array(fc)%csf,final_csf_array(fc)%dets(:,fd)),trim(detstring(final_csf_array(fc)%occ,final_csf_array(fc)%dets(:,fd)))
          !end do

          ! TODO: a way to obtain Auger amplitudes is
          ! through the symmetric group approach
          ! instead of comparing individual determinants
          final_csf_Auger_amplitude(:, fc) = ZERO
          if (flag_final_csfs(fc) .eq. 0) then
            auger_teint_coef(:, :, :) = ZERO
            do ic = 1, size(initial_csf_array, 1)
              if (sum(abs(initial_csf_array(ic)%occ(:) - final_csf_array(fc)%occ(:))) .gt. 3) cycle
              if (abs(initial_CI_matrix(ic, ini_state)) .lt. P%EPS_Auger) cycle
              ! Loop over determinant pairs
              do id = 1, size(initial_csf_array(ic)%dets, 2)
                do fd = 1, size(final_csf_array(fc)%dets, 2)
                  coef = initial_CI_matrix(ic, ini_state)* &
                    csfdet(final_csf_array(fc)%csf, final_csf_array(fc)%dets(:, fd))* &
                    csfdet(initial_csf_array(ic)%csf, initial_csf_array(ic)%dets(:, id))
                  if (coef .eq. ZERO) cycle
                  call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
                    initial_csf_array(ic)%occ, initial_csf_array(ic)%dets(:, id), &
                    final_csf_array(fc)%occ, final_csf_array(fc)%dets(:, fd))
                  coef = coef*coef_sign
                  if (coef_sign .ne. 0 .and. count(only_left(:) .ne. 0) .eq. 2 .and. count(only_right(:) .ne. 0) .eq. 1) then
                    continuum_spin = sum(only_left_spin) - sum(only_right_spin)
                    if (abs(continuum_spin) .eq. 1) then
                      if ( &
                        only_left_spin(1) .eq. only_right_spin(1) .and. &
                        only_left_spin(1) .eq. only_left_spin(2) .and. &
                        only_left_spin(1) .eq. continuum_spin) then
                        ! a a -> a a
                        ! b b -> b b
                        auger_teint_coef(only_left(1), only_left(2), only_right(1)) = &
                          auger_teint_coef(only_left(1), only_left(2), only_right(1)) + coef
                        auger_teint_coef(only_left(2), only_left(1), only_right(1)) = &
                          auger_teint_coef(only_left(2), only_left(1), only_right(1)) - coef
                      else if ( &
                        only_left_spin(1) .eq. only_right_spin(1) .and. &
                        only_left_spin(2) .eq. continuum_spin) then
                        ! a b -> a b
                        ! b a -> b a
                        auger_teint_coef(only_left(1), only_left(2), only_right(1)) = &
                          auger_teint_coef(only_left(1), only_left(2), only_right(1)) + coef
                      else if ( &
                        only_left_spin(2) .eq. only_right_spin(1) .and. &
                        only_left_spin(1) .eq. continuum_spin) then
                        ! a b -> b a
                        ! b a -> a b
                        auger_teint_coef(only_left(2), only_left(1), only_right(1)) = &
                          auger_teint_coef(only_left(2), only_left(1), only_right(1)) - coef
                      end if
                    end if
                  end if
                end do! id
              end do! fd
            end do! ic

            HighSpinRate = ZERO
            final_csf_Auger_amplitude(:, fc) = ZERO
            do l = 0, lmax
              continuumAO%l = l
              do m = -l, l
                continuumAO%m = m
                do at = 1, P%N_nuc
                  if (.not. allocated(RI(at)%R)) cycle
                  if (size(RI(at)%R, 1) .eq. 0) cycle
                  Amplitude = ZERO
                  !  special case for hydrogen: no auger
                  if (A(at)%charge .eq. 1) cycle
                  do val1 = 1, size(auger_teint_coef, 1)
                    do val2 = 1, size(auger_teint_coef, 2)
                      do hole = 1, size(auger_teint_coef, 3)
                        if (abs(auger_teint_coef(val1, val2, hole)) .lt. P%EPS_Auger) cycle
                        if (allocated(RI(at)%Omin)) then
                          Amplitude = AugerTransitionAmplitude(P, RI(at)%min_info, &
                            RI(at)%Omin(:, val1), RI(at)%Omin(:, val2), &
                            RI(at)%Omin(:, hole), at, continuumAO, RI(:))
                        else
                          Amplitude = AugerTransitionAmplitude(P, B%info, &
                            F%LC(:, val1), F%LC(:, val2), &
                            F%LC(:, hole), at, continuumAO, RI(:))
                        end if
                        final_csf_Auger_amplitude((l)**2 + l + m + 1, fc) = &
                          final_csf_Auger_amplitude((l)**2 + l + m + 1, fc) + &
                          Amplitude*auger_teint_coef(val1, val2, hole)

                        !if(abs(Amplitude) > 1e-3)&
                        !     write(P%OUTPUT,'(10x,"<",i4,x,i4,",",x,i4,x,i2,x,i2,">",x,f8.3,"*",es,"=",es)') ,&
                        !     val1,val2,hole,l,m,auger_teint_coef(val1,val2,hole),Amplitude,auger_teint_coef(val1,val2,hole)*Amplitude
                      end do! hole
                    end do! val2
                  end do! val1
                  HighSpinRate = HighSpinRate + TWO*PI*final_csf_Auger_amplitude((l)**2 + l + m + 1, fc)**2
                  !                    if(final_csf_Auger_amplitude( (l)**2 + l+m+1 ,fc) > 1e-3)&
                  !                      write(P%OUTPUT,'("# l=",i4,"m=",i4,"Amplitude :",es,"Rate :",es)') l,m,final_csf_Auger_amplitude( (l)**2 + l+m+1 ,fc),TWO*PI*final_csf_Auger_amplitude( (l)**2 + l+m+1 ,fc)**2
                end do! atom
              end do! m
            end do! l
          end if! flag_final_csfs==0
          !csfStr = trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf))
          !csfRate = sum(final_csf_Auger_amplitude(:, fc)**2)*TWO*PI
          !write(P%OUTPUT,'("# csf ",i7,x,a,x,"rate=",es20.7)') fc, trim(csfStr), csfRate
        end do! fc
        !$omp end do
        deallocate (auger_teint_coef)
        !$omp end parallel
        if (P%yes_verbose) then
          write (P%OUTPUT, "('# Primary csfs (2h-1p): ')")
          do fc = 1, size(final_csf_array, 1)
            if (flag_final_csfs(fc) .eq. 0 )&
              write (P%OUTPUT, "('# ',a,x,es12.5 )") &
              csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf), &
              sum(final_csf_Auger_amplitude(:, fc)**2)*TWO*PI
          end do
        end if
        if (P%yes_process_CI_relaxed) then
          ! obtain double ionized gs orbitals
          F_final = F
          !B_final = B
          call read_wf(F, F_final%LC, "wf_relaxed.dat")
          allocate (csf_O(size(final_csf_array), size(final_csf_array)))
          call overlap_csfs(P, final_csf_array, F%LC, F_final%LC, B%S, csf_O)
          allocate (final_csf_Auger_amplitude_temp(N_final_csfs))
          do i = 1, (lmax + 1)**2
            do j = 1, N_final_csfs
              final_csf_Auger_amplitude_temp(j) = sum(final_csf_Auger_amplitude(i, :)*csf_O(:, j))
              if (final_csf_Auger_amplitude_temp(j)**2 .lt. 1e-30) &
                final_csf_Auger_amplitude_temp(j) = ZERO
            end do
            final_csf_Auger_amplitude(i, :) = final_csf_Auger_amplitude_temp(:)
          end do
          deallocate (final_csf_Auger_amplitude_temp)
          write (6, *) 'Overlap to new basis:'
          do fc = 1, size(final_csf_array, 1)
            write (P%OUTPUT, '("# csf ",i7,x,a,x,"rate=",es20.7,x,f7.3,f7.3)') &
              fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), &
              sum(final_csf_Auger_amplitude(:, fc)**2)*TWO*PI, &
              sum(csf_O(:, fc)**2), sum(csf_O(fc, :)**2)
          end do
          deallocate (csf_O)
        else
          F_final = F
          !B_final = B
        end if!
        allocate (nz_el(count((/(sum(final_csf_Auger_amplitude(:, fc)**2) .gt. ZERO, fc=1, size(final_csf_array, 1))/))))
        nz_el(:) = pack((/(fc, fc=1, size(final_csf_array, 1))/), &
          (/(sum(final_csf_Auger_amplitude(:, fc)**2) .gt. ZERO, fc=1, size(final_csf_array, 1))/))
        !        write(6,*) nz_el

        allocate (final_CI_energy(N_final_csfs))
        deallocate (active, core)
        allocate (active( &
          count((/((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0) &
          ), j=1, size(final_occ, 1))/))))
        allocate (core(count((/((all(final_occ(j, 1:N_final_occ) .eq. 2)), j=1, size(final_occ, 1))/))))
        active = pack((/(j, j=1, size(final_occ, 1))/), &
          (/(((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0))), &
          j=1, size(final_occ, 1))/))
        core = pack((/(j, j=1, size(final_occ, 1))/), &
          (/(all(final_occ(j, 1:N_final_occ) .eq. 2), j=1, size(final_occ, 1))/))
        nintern = size(active)
        do j = size(active), 1, -1
          if (any(sum(final_occ(active(j:), :N_final_occ), dim=1) .gt. 1)) exit
        end do
        nintern = j
        write (6, '("#     core:",99i4)') core
        write (6, '("#   active:",99i4)') active
        write (6, '("# internal:",99i4)') active(:nintern)
        write (6, '("# external:",99i4)') active(nintern + 1:)
        call construct_h0_matrix(G, B, F_final)
        F_final%H0_MO = F_final%LC(:, :) .tm.B%H0(:, :) .m.F_final%LC(:, :)
        call calculate_F_inactive(P, B, core, F_final%LC, F_core)
        E_core = ZERO
        do i = 1, size(core)
          E_core = E_core + F_core(core(i), core(i)) + F_final%H0_MO(core(i), core(i))
        end do
        call transform_twoel_allactive(P, B, F_final, active(:nintern), P%OUTPUT)
        if (nintern .lt. size(active)) then
          call transform_twoel_two_core_two_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
          call transform_twoel_three_core_one_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
        end if

        if (P%yes_use_arpack .or. P%yes_use_davidson .or. P%yes_use_krylov) then
          allocate (CI_diagonal(N_final_csfs))

          write (P%OUTPUT, '("# calculating CI diagonal entries")')
          call calc_CIMatrix_diagonal(UTfin, E_core, F_core(active(:), active(:)), &
            F_final%teint_abcd, &
            final_occ(active(:), 1:N_final_occ), &
            CI_diagonal, nintern=nintern, &
            teint_iiee=F_final%teint_cdij, &
            teint_ieie=F_final%teint_cidj, &
            teint_iiie=F_final%teint_cdei)
          write (P%OUTPUT, '("# done")')
          nev = max(P%nstates, 1)
          allocate (csf_c2(N_final_csfs, nev))
          csf_c2 = ZERO
          ! take the configuration with lowest diagonal element from those entries that are in the primary space
          allocate (idx(size(CI_diagonal)))
          call merge_argsort(CI_diagonal, idx)
          j = 0
          k = 0
          do while (j .lt. nev .and. k .lt. size(idx))
            k = k + 1
            if (flag_final_csfs(idx(k)) .ne. 0) cycle
            j = j + 1
            csf_c2(idx(k), j) = ONE
            if (j .eq. nev) exit
          end do
          ! if there is still space, fill up with lowest energy from secondart space
          if (nev .gt. j) then
            k = 0
            do while (j .lt. nev .and. k .lt. size(idx))
              k = k + 1
              if (flag_final_csfs(idx(k)) .eq. 0) cycle
              j = j + 1
              csf_c2(idx(k), j) = ONE
            end do
          end if
          deallocate (idx)
          ! print summary
          do i = 1, nev
            do j = 1, size(csf_c2, 1)
              if (csf_c2(j, i) .eq. ZERO) cycle
              write (P%OUTPUT, '("# Starting guess: ",i4,x,es20.7)', advance="yes") i, ci_diagonal(j)
              write (P%OUTPUT, '("# ",es20.7,x,a)', advance="yes") &
                csf_c2(j, i), trim(csfstring(final_csf_array(j)%occ, final_csf_array(j)%csf(:)))
            end do
          end do
          if (P%yes_use_arpack) then
            ! call arpack
#ifdef WITH_ARPACK
            call get_some_eigenvalues_arpack(N_final_csfs, nev, final_CI_energy(1:nev), csf_c2, matrix_vector_multp)
#else
            stop "not compiled with arpack support"
#endif
          else if (P%yes_use_davidson) then
            ! call Davidson
            ! call get_some_eigenvalues_VariableBlockDavidson(N_final_csfs, nev, P%davblock, max(P%davMax,nev), final_CI_energy(1:nev), csf_c2,  ci_diagonal, matrix_matrix_multp, P%EPS_Auger, P%OUTPUT)
            call get_some_eigenvalues_davpack(N_final_csfs, nev, min(P%davblock, nev), &
              min(max(P%davMax, nev),N_final_csfs), final_CI_energy(1:nev), &
              csf_c2, ci_diagonal, matrix_matrix_multp, &
              P%EPS_Auger, P%OUTPUT)
          else if (P%yes_use_krylov) then
#ifdef WITH_KRYLOV
            call get_some_eigenvalues_libKrylov(N_final_csfs, nev, nev, &
              final_CI_energy(1:nev), &
              csf_c2, ci_diagonal, matrix_matrix_multp, &
              epsilon=P%EPS_Auger, OUTPUT=P%OUTPUT)
#else
            stop "not compiled with libkrylov support"
#endif
          end if
          deallocate (CI_diagonal)
          deallocate (F_final%teint_abcd)
          if (nintern .lt. size(active)) then
            if(allocated(F_final%teint_cdij))&
              deallocate (F_final%teint_cdij)
            if(allocated(F_final%teint_cidj))&
              deallocate (F_final%teint_cidj)
            if(allocated(F_final%teint_cdei))&
              deallocate (F_final%teint_cdei)
          end if
        else
          allocate (final_CI_matrix(N_final_csfs, N_final_csfs))
          write (P%OUTPUT, '("# calculating CI matrix")')
          if (nintern .lt. size(active)) then
            call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
              F_final%teint_abcd, &
              final_occ(active(:), 1:N_final_occ), &
              final_CI_matrix, nintern=nintern, &
              teint_iiee=F_final%teint_cdij, &
              teint_ieie=F_final%teint_cidj, &
              teint_iiie=F_final%teint_cdei)
            if(allocated(F_final%teint_cdij))&
              deallocate (F_final%teint_cdij)
            if(allocated(F_final%teint_cidj))&
              deallocate (F_final%teint_cidj)
            if(allocated(F_final%teint_cdei))&
              deallocate (F_final%teint_cdei)
          else
            call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
              F_final%teint_abcd, &
              final_occ(active(:), 1:N_final_occ), &
              final_CI_matrix)
            write (P%OUTPUT, '("# done")')
          end if

          deallocate (F_final%teint_abcd)
          write (P%OUTPUT, '("# done")')
          ! solve it
          write (P%OUTPUT, '("# solving CI matrix")')
          allocate (csf_c2(N_final_csfs, nev))
          call get_some_ev_of_real_sym(final_CI_matrix, final_CI_energy, csf_c2, nev, P%EPS_Auger, yes_verbose=.False.)

          !call get_eigenvalues( final_CI_matrix, final_CI_energy )
          !csf_c2(:,:) = final_CI_matrix(:,1:nev)
          write (P%OUTPUT, '("# done")')

          deallocate (final_CI_matrix)
        end if! use arpack

        ! For every CI vector assemble the Auger amplitudes:
        do i = 1, size(csf_c2, 2)

          AE_energy = E_ini - final_CI_energy(i)
          Coul_charge = TWO
          write (P%OUTPUT, '("# ",i4," E=",f20.5," a.u."," Auger Energy=",f15.5," eV")') &
            i, final_CI_energy(i) + nuclear_repulsion_energy(A, P), 27.211*AE_energy
          if (sum(csf_c2(nz_el(:), i)**2) .lt. 0.01) cycle ! skipping states which have very poor contribution in any meaningfull csfs
          do k = 1, size(final_csf_array, 1)
            if (abs(csf_c2(k, i)) .gt. 1e-1) then
              write (P%OUTPUT, '("# ",5x,f10.5,x,a,es20.7)', advance="yes") &
                csf_c2(k, i), trim(csfstring(final_csf_array(k)%occ, final_csf_array(k)%csf)), &
                sum(final_csf_Auger_amplitude(:, k)**2)*TWO*PI
              !do l=0,lmax
              !  do m=-l,l
              !    if(abs(final_csf_Auger_amplitude( (l)**2 + l+m+1 ,k ))>ZERO)&
              !         write(P%OUTPUT,'("# ",10x,i4,i4,es)',advance="yes") &
              !         l, m,&
              !         final_csf_Auger_amplitude( (l)**2 + l+m+1 ,k )
              !  end do
              !end do
            end if
          end do
          ! for all the angular distribution, we assume that only one .auger file is provided that
          ! contains radial Auger integrals
          do at = 1, size(RI)
            if (size(RI(at)%R) .gt. 0) exit
          end do

          call check_time(int_time, timer)
          time_last_step = timer

          write (P%OUTPUT, '("# angular distribution from phases and amplitudes for atom  ",i4)') at
          HighSpinRate = ZERO
          do l = 0, lmax
            do m = -l, l
              Amplitude = ZERO
              do fc = 1, size(final_csf_array, 1)
                Amplitude = Amplitude + csf_c2(fc, i)*final_csf_Auger_amplitude((l)**2 + l + m + 1, fc)
              end do
              Amplitude_ang_dist(l**2 + l + m + 1) = Amplitude
              HighSpinRate = HighSpinRate + TWO*PI*Amplitude**2
            end do!m
          end do!l
          call check_time(int_time, timer)
          if (P%yes_verbose) &
            write (P%OUTPUT, '("# Auger CI: ","time for amplitudes",f15.3,"sec ")') timer - time_last_step
          time_last_step = timer

          Auger_intensity_ang_dist(:) = ZERO
          do l = 0, 2*lmax
            do m = -l, l
              do l1 = 0, lmax
                !                delta_l1=coulomb_phase(Coul_charge,l1,sqrt(TWO*AE_energy))
                delta_l1 = AugerPhase(Coul_charge, l1, AE_energy, RI(at))
                do m1 = -l1, l1
                  do l2 = 0, lmax
                        !                    delta_l2=coulomb_phase(Coul_charge,l2,sqrt(TWO*AE_energy))
                    delta_l2 = AugerPhase(Coul_charge, l2, AE_energy, RI(at))
                    do m2 = -l2, l2
                      gaunt_int = realGauntCoef(l, m, l1, m1, l2, m2)

                      Auger_intensity_ang_dist((l)**2 + l + m + 1) = &
                        Auger_intensity_ang_dist((l)**2 + l + m + 1) + &
                        TWO*PI* &
                        Amplitude_ang_dist(l1**2 + l1 + m1 + 1)* &
                        Amplitude_ang_dist(l2**2 + l2 + m2 + 1)* &
                        cos(delta_l2 - delta_l1)* &
                        gaunt_int
                      !                      if( abs(Amplitude_ang_dist(l1**2+l1+m1+1))>ZERO &
                      !                           .and. abs(Amplitude_ang_dist(l2**2+l2+m2+1))>ZERO &
                      !                           .and. abs(gaunt_int)>ZERO &
                      !                      )&
                      !                           write(6,'("# ",i4,i4,es,i4,i4,es,i4,i4)') l,m, Amplitude_ang_dist(l1**2+l1+m1+1), l1,m1, Amplitude_ang_dist(l2**2+l2+m2+1),l2,m2
                    end do! m2
                  end do! l2
                end do! m1
              end do!l1
            end do!m
          end do!l

          call check_time(int_time, timer)
          if (P%yes_verbose) &
            write (P%OUTPUT, '("# Auger CI: ","time for angular phases",f15.3,"sec ")') timer - time_last_step
          time_last_step = timer

          ! Apply Wigner Eckert to get the transition rates to other MS configurations:
          initial_two_S = initial_multp - 1
          initial_two_M = initial_two_S
          final_two_S = final_multp - 1
          write (P%OUTPUT, "('# ',5x,'rate for MS=',f5.2,'  =',es20.7)") final_two_S/2., HighSpinRate
          ChannelRate = ZERO
          do final_two_M = -final_two_S, final_two_S, 2
            ChannelRate = ChannelRate + HighSpinRate* &
              cleb(final_two_S, final_two_M, 1, initial_two_S - final_two_M, initial_two_S, initial_two_S)**2 &
              /cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2
          end do
          write (P%OUTPUT, "('# ',5x,'channel energy/rate:' ,x,es20.7,es20.7)") 27.211*AE_energy, ChannelRate

          call check_time(int_time, timer)
          if (P%yes_verbose) &
            write (P%OUTPUT, '("# Auger CI: ","time for rate",f15.3,"sec ")') timer - time_last_step
          time_last_step = timer

          Auger_intensity_ang_dist(:) = Auger_intensity_ang_dist(:)*ChannelRate/HighSpinRate

          Amplitude_ang_dist(:) = Amplitude_ang_dist(:)*sqrt(ChannelRate/HighSpinRate)

          write (P%OUTPUT, '("# ", a4, a4, a15, a15)') 'l', 'm', 'amplitude', 'phase'
          do l = 0, lmax
            do m = -l, l
              if (abs(Amplitude_ang_dist(l**2 + l + m + 1)) .gt. ZERO ) &
                write (P%OUTPUT, '("# ",i4,i4, es20.7, es20.7)') &
                l, m, Amplitude_ang_dist(l**2 + l + m + 1), AugerPhase(Coul_charge, l, AE_energy, RI(at))
            end do
          end do

          write (P%OUTPUT, '("# Angular distribution: I(theta,phi) = \sum_{lm} a_{lm} Y_{lm} (theta,phi) ")')
          write (P%OUTPUT, '("# ",5x,a4,a4,5x,a10,a15,a15,a15)') "l", "m", "a_lm/a_00", " a_lm", &
            "a_lm*sqrt((2*l+1)/(4pi))", "a_lm*sqrt((2*l+1)*(4pi))"
          do l = 0, 2*lmax
            do m = -l, l
              if (abs(Auger_intensity_ang_dist((l)**2 + l + m + 1))/abs(Auger_intensity_ang_dist(1)) .gt. 1e-3) &
                write (P%OUTPUT, '("# ",5x,i4,i4,5x,f10.5,es15.3,es15.3,es15.3,es15.3)') &
                l, m, Auger_intensity_ang_dist((l)**2 + l + m + 1)/Auger_intensity_ang_dist(1), &
                Auger_intensity_ang_dist((l)**2 + l + m + 1), &
                Auger_intensity_ang_dist((l)**2 + l + m + 1)*sqrt((TWO*l + 1))*ONE/sqrt(FOUR*PI), &
                Auger_intensity_ang_dist((l)**2 + l + m + 1)*sqrt((TWO*l + 1))*sqrt(FOUR*PI)
            end do!m
          end do!l

          call check_time(int_time, timer)
          if (P%yes_verbose) &
            write (P%OUTPUT, '("# Auger CI: ","time for angular distribution",f15.3,"sec ")') timer - time_last_step
          time_last_step = timer

          !call add_Process( P,F,S, "Auger", ChannelRate, AE_energy, 0, 0, 0 ,&
          !  initial_irrep=F%state_irrep,initial_multp=initial_multp, initial_state_num=ini_state,&
          !  final_irrep=f_irrep,final_multp=final_multp,final_state_num=i)

          TotalRate = TotalRate + ChannelRate
          if (P%yes_population) then
            ! Do some additional analysis on the CI eigenvectors
            ! calculate opdm
            allocate (opdm(size(F%LC, 2), size(F%LC, 2)))
            allocate (n_occ(size(F%LC, 2)))
            allocate (charge(size(A, 1)))
            !call one_particle_dm_CI(opdm,final_CI_matrix(:,i),final_csf_array )
            call check_time(int_time, timer)
            time_last_step = timer

            call one_particle_dm_CI2(UTfin, final_occ, csf_c2(:, i), opdm)
            call check_time(int_time, timer)
            if (P%yes_verbose) &
              write (P%OUTPUT, '("# Auger CI: ","time for calculating DM",f15.3,"sec ")') timer - time_last_step
            time_last_step = timer

            F_final%D = F_final%LC.m.opdm.mt.F_final%LC

            call check_time(int_time, timer)
            if (P%yes_verbose) &
              write (P%OUTPUT, '("# Auger CI: ","time for transforming DM",f15.3,"sec ")') timer - time_last_step
            time_last_step = timer
            do j = 1, size(A)
              charge(j) = dble(A(j)%charge)
            end do
            do j = 1, B%N_basis
              charge(B%info(j, 1)) = charge(B%info(j, 1)) - sum(F_final%D(j, :)*B%S(j, :))
            end do
            write (P%OUTPUT, '("# Mulliken Charge: ",99f10.5)') charge(:)
            deallocate (charge)
            call check_time(int_time, timer)
            if (P%yes_verbose) &
              write (P%OUTPUT, '("# Auger CI: ","time for Mulliken",f15.3,"sec ")') timer - time_last_step
            time_last_step = timer

            call bondOrder_analysis(P, B, F_final%D, P%OUTPUT)
            call check_time(int_time, timer)
            if (P%yes_verbose) &
              write (P%OUTPUT, '("# Auger CI: ","time for BondOrder",f15.3,"sec ")') timer - time_last_step
            time_last_step = timer

            call get_eigenvalues(opdm, n_occ)
            call check_time(int_time, timer)
            if (P%yes_verbose) &
              write (P%OUTPUT, '("# Auger CI: ","time for natural orbitals",f15.3,"sec ")') timer - time_last_step
            time_last_step = timer

            write (P%OUTPUT, '("# natural occ.: ",99f7.2)') n_occ(size(n_occ):1:-1)
            deallocate (opdm)
            deallocate (n_occ)
          end if! yes_population
        end do! i loop over final states
        deallocate (csf_c2)

        ! clean up arrays for this final spin multp:
        call purge_csf_array(final_csf_array)
        deallocate (Auger_intensity_ang_dist)
        deallocate (Amplitude_ang_dist)
        deallocate (final_csf_array)
        deallocate (flag_final_csfs)
        !deallocate(final_CI_matrix)
        deallocate (final_CI_energy)
        deallocate (final_csf_Auger_amplitude)

        deallocate (final_occ)
        deallocate (nz_el)
      end do ! f_irrep
      call purge_UMatrixTable(UTfin)
    end do ! final_multp

    write (P%OUTPUT, "('#  Total Augerrate: ',es13.5,' a.u. (lifetime=',a,')')") TotalRate, trim(lifetime(TotalRate))
    ! cleanup for initial state
    call purge_csf_array(initial_csf_array)
    deallocate (initial_CI_matrix)
    deallocate (initial_CI_energy)
    deallocate (initial_occ)
    deallocate (initial_occ_orig)
    call purge_UMatrixTable(UTini)
    deallocate (core)
    deallocate (occupied)
    deallocate (active)
    deallocate (virtual)

    ! cleanup aux. data
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger
      if (A(at)%charge .eq. 1) cycle
      deallocate (RI(at)%K)
      deallocate (RI(at)%n1)
      deallocate (RI(at)%l1)
      deallocate (RI(at)%n2)
      deallocate (RI(at)%l2)
      deallocate (RI(at)%n3)
      deallocate (RI(at)%l3)
      deallocate (RI(at)%l4)
      deallocate (RI(at)%R)
      if (allocated(RI(at)%Pmin)) then
        deallocate (RI(at)%Pmin)
        deallocate (RI(at)%Omin)
        deallocate (RI(at)%min_info)
      end if
    end do
  contains
    !----------------------------------------------------------------------------
    ! print out lifetime in fs or attosec.
    !----------------------------------------------------------------------------
    function lifetime(rate) result(string)
      real(long), intent(in) :: rate                ! in a.u.
      real(long) :: tau
      character(len=15) :: str, string
      tau = ONE/rate*AU2SEC/1e-15_long        ! to femtosec.
      if (tau .gt. ONE) then
        write (str, "(f12.1,' fs')") tau
      else
        tau = tau*1000_long                       ! to attosec.
        write (str, "(f12.1,' as')") tau
      end if
      string = trim(adjustl(str))
    end function lifetime

    function matrix_vector_multp(n, Cin, Cout) result(error)
      integer, intent(in) :: n
      real(long), intent(in) :: Cin(n)
      real(long), intent(out) :: Cout(n)
      integer :: error
      Cout = Cin
      call calc_CIMatrix_vector_multp(UTfin, E_core, F_core(active(:), active(:)), &
        F_final%teint_abcd, &
        final_occ(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F_final%teint_cdij, &
        teint_ieie=F_final%teint_cidj, &
        teint_iiie=F_final%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_vector_multp

    function matrix_matrix_multp(n, ncol, Cin, Cout) result(error)
      integer, intent(in) :: n, ncol
      real(long), intent(in) :: Cin(n, ncol)
      real(long), intent(out) :: Cout(n, ncol)
      integer :: error
      call calc_CIMatrix_matrix_multp(UTfin, E_core, F_core(active(:), active(:)), &
        F_final%teint_abcd, &
        final_occ(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F_final%teint_cdij, &
        teint_ieie=F_final%teint_cidj, &
        teint_iiie=F_final%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_matrix_multp
  end subroutine calcAugerRates_CI

  function coulomb_phase(q, l, k) result(phase)
    real(long), intent(in) :: q
    integer, intent(in):: l
    real(long), intent(in) :: k
    real(long) :: phase
    ! the Coulomb function is asymptotically
    ! sin( k*r - eta*log(2*k*r) - l/2*pi + arg( Gamma( l + 1 + i * eta )) )
    ! Taylor: "scattering theory"
    ! We ignore the logarithmic part and return the phase that depends
    ! on different l:
    ! eta = -q/k ! Z is positive charge, electron has negative charge, so eta is negative!
    phase = -HALF*l*PI + arg_gamma(l, -q/k)
    phase = modulo(phase, TWO*PI)
    if (phase .lt. ZERO) phase = phase + TWO*PI
  contains
    ! this is an approximation for the argument of the Gamma function
    ! Gamma(l+1+i*eta) Eq. 6 of Barata, J.C.A., Canto, L.F. & Hussein, M.S.
    ! Braz J Phys (2011) 41: 50. https://doi.org/10.1007/s13538-011-0005-6
    function arg_gamma(l, eta) result(arg)
      integer, intent(in):: l
      real(long), intent(in) :: eta
      real(long) :: arg
      arg = (dble(l) + HALF)*atan(eta/dble(l + 1)) + &
        eta*(log(sqrt(dble(l + 1)**2 + eta**2)) - ONE) &
        - eta/(dble(12)*(dble(l + 1)**2 + eta**2))
      arg = modulo(arg, TWO*PI)
    end function arg_gamma
  end function coulomb_phase

  subroutine calcAugerRates_spin_avg(P, A, B, F, S, auger_fpath)
    !! calculates Auger-Meitner decay transitions based on orbital energy differences resolving individual spin configurations and considering averaging over configuration    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S

    integer, parameter :: MAX_SPIN = 20
    integer :: hole, val1, val2 ! molecular orbital indices
    integer :: j ! counter
    integer :: m, l ! counter for angular momentum
    integer :: at ! counter for atoms
    integer :: lmax ! this is the maximum angular momentum in the continuum which we consider
    real(long) :: AvgTotalRate, ChannelRate, Rate, TotalRate, SpinavgRate ! Rates & Amplitudes
    integer, allocatable :: weight_io(:)
    integer :: Spinavg_n, Total_n, AvgTotal_n
    type(atomorbital) :: continuumAO ! continuum orbital parameter structure
    type(RadialIntegrals), dimension(P%N_nuc) :: RI ! Radialintegral structure to be filled by xatom output
    character(len=512) :: auger_file_nuc ! filename
    !! -- Mar.10.2015 : read-in specified fpath ---
    character*(*), optional :: auger_fpath(:)       !! Currently we use char*(128) in MC code
    real(long) :: AE_energy ! Auger electron energy
    integer :: initial_multp ! initial multp
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: rest_multp ! multipicity in remaining electron configuration
    integer :: rest_Nopen ! open shells in remaining electron configuration
    integer :: parti_multp ! initial multipicity of hole,val1,val2 electron configuration
    integer :: parti_Nopen ! open shells in remaining electron configuration
    integer :: partf_multp ! final multipicity of hole,val1,val2 electron configuration
    integer :: partf_Nopen ! open shells in remaining electron configuration
    integer :: rest_Ncsf, parti_Ncsf, partf_Ncsf, initial_Ncsf ! number of spin configuration state functions
    integer :: final_two_S, initial_two_S, initial_two_M ! two times final and intial spin quantum numbers
    integer :: parti_two_M, parti_two_S ! two times spin quantum numbers of initial participator part,
    integer :: partf_two_M, partf_two_S ! final participator part,
    integer :: rest_two_M, rest_two_S   ! and rest
    real(long):: WE_factor ! factor from Wigner Eckhard theorem
    real(long):: spin_factor ! another spin factor
    real(long), allocatable :: ampl_12(:, :) ! amplitudes between initial conf to final conf
    real(long), allocatable :: ampl_21(:, :) ! amplitudes between initial conf to final conf
    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer :: io

    ! read all radial integrals
    lmax = 0
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger file
      if (A(at)%charge .eq. 1) cycle
      if (present(auger_fpath)) then              ! extension Mar.10.2015
        auger_file_nuc = ""                        !
        auger_file_nuc = trim(auger_fpath(at))   !
      else                                       !
        auger_file_nuc = A(at)%auger_file_nuc
      end if
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Reading Auger Radial integrals from file ',a)") trim(auger_file_nuc)
      end if
      call readRIntegral(auger_file_nuc, RI(at))
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# atom: ',i4,' ',a, ' : ',i4,' entries')") &
          at, trim(auger_file_nuc), size(RI(at)%R)
      end if
      lmax = max(lmax, maxval(RI(at)%l4))
      ! prepare the projection of MO's into minimal basis set
      if (allocated(RI(at)%Pmin)) then
        allocate (RI(at)%Smin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        forall (j=1:size(RI(at)%Pmin, 1)) RI(at)%Smin(j, j) = ONE
        allocate (RI(at)%Omin(size(RI(at)%Pmin, 1), size(RI(at)%Pmin, 2)))
        RI(at)%Omin = matmul(RI(at)%Pmin, F%LC)
      else if (P%yes_gaussian) then
        allocate (RI(at)%Omin(size(F%Omin, 1), size(F%Omin, 2)))
        allocate (RI(at)%Smin(size(F%Smin, 1), size(F%Smin, 2)))
        allocate (RI(at)%min_info(size(F%min_info, 1), size(F%min_info, 2)))
        RI(at)%Omin = F%Omin
        RI(at)%Smin = F%Smin
        RI(at)%min_info = F%min_info
      end if! if Pmin allocated
      if (P%yes_verbose .and. allocated(RI(at)%min_info)) then
        call print_projection_info(RI(at)%min_info, RI(at)%Omin, RI(at)%Smin, at, P%OUTPUT)
      end if! yes_verbose
    end do! atom

    allocate (ampl_12(size(A), (lmax + 1)**2))
    allocate (ampl_21(size(A), (lmax + 1)**2))

    ! in case of MO occupation averaging be render all occupation patterns; otherwise just take F%N_occ
    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if
    allocate (weight_io(size(initial_occ, 2)))
    weight_io(:) = 0

    AvgTotalRate = ZERO
    AvgTotal_n = 0
    do io = 1, size(initial_occ, 2)
      TotalRate = ZERO
      Total_n = 0
      if (P%print_level .gt. 1 &
        ) write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)

      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        weight_io(io) = weight_io(io) + initial_multp*num_csf(initial_Nopen, initial_multp)
      end do

      ! Loop over hole and valence pair combinations
      do hole = 1, size(initial_occ, 1)
        if (initial_occ(hole, io) .eq. 2) cycle
        do val1 = 1, size(initial_occ, 1)
          if (initial_occ(val1, io) .eq. 0) cycle
          do val2 = val1, size(initial_occ, 1)
            if (initial_occ(val2, io) .eq. 0) cycle
            if (val1 .eq. val2 .and. initial_occ(val2, io) .lt. 2) cycle
            AE_energy = F%CE(val1) + F%CE(val2) - F%CE(hole)
            if (AE_energy .lt. 0) cycle
            ampl_12(:, :) = ZERO
            ampl_21(:, :) = ZERO
            do l = 0, lmax
              continuumAO%l = l
              do m = -l, l
                continuumAO%m = m
                ! we incoherently sum over different atom contribution
                do at = 1, P%N_nuc
                  ! special case for hydrogen: no auger
                  if (A(at)%charge .eq. 1) cycle
                  if (allocated(RI(at)%Omin)) then
                    ampl_12(at, l*(l + 1) + m + 1) = AugerTransitionAmplitude(P, RI(at)%min_info, RI(at)%Omin(:, val1), &
                      RI(at)%Omin(:, val2), RI(at)%Omin(:, hole), &
                      at, continuumAO, RI(:))
                    ampl_21(at, l*(l + 1) + m + 1) = AugerTransitionAmplitude(P, RI(at)%min_info, RI(at)%Omin(:, val2), &
                      RI(at)%Omin(:, val1), RI(at)%Omin(:, hole), &
                      at, continuumAO, RI(:))
                  else
                    ampl_12(at, l*(l + 1) + m + 1) = AugerTransitionAmplitude(P, B%info, F%LC(:, val1), F%LC(:, val2), &
                      F%LC(:, hole), at, continuumAO, RI(:))
                    ampl_21(at, l*(l + 1) + m + 1) = AugerTransitionAmplitude(P, B%info, F%LC(:, val2), F%LC(:, val1), &
                      F%LC(:, hole), at, continuumAO, RI(:))
                  end if
                end do! atom
              end do!m
            end do! l
            parti_Nopen = count(initial_occ((/hole, val1, val2/), io) .eq. 1)
            rest_Nopen = initial_Nopen - parti_Nopen
            if (val1 .ne. val2) then
              partf_Nopen = count(initial_occ((/hole, val1, val2/), io) + (/1, -1, -1/) .eq. 1)
            else
              partf_Nopen = count(initial_occ((/hole/), io) + (/1/) .eq. 1)
            end if
            final_Nopen = rest_Nopen + partf_Nopen

            SpinavgRate = ZERO
            Spinavg_n = 0
            do initial_multp = initial_Nopen + 1, 1, -2
              if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
              ChannelRate = ZERO
              initial_Ncsf = num_csf(initial_Nopen, initial_multp)
              if (P%print_level .gt. 2) then
                write (P%OUTPUT, "('#    ')")
                write (P%OUTPUT, "('# ','imultp',2x,'  p ','  r ','  t ',x,'fmultp',2x,'  p ','  r ','  t ',' factor ')")
              end if
              ! loop over possible final spin multiplicities
              do final_multp = initial_multp - 1, initial_multp + 1, 2
                if (final_multp .eq. 0) cycle
                if (final_multp - 1 .gt. final_Nopen) cycle
                Rate = ZERO
                ! We calculate transition rate to high spin state
                ! Apply Wigner Eckert to get factor for transition rates to other MS configurations:
                initial_two_S = initial_multp - 1
                initial_two_M = initial_two_S
                final_two_S = final_multp - 1
                WE_factor = ONE/cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2
                ! we split up the electrons into participator + rest:
                ! loop over multiplicity for the rest
                do rest_multp = rest_Nopen + 1, 1, -2
                  rest_Ncsf = num_csf(rest_Nopen, rest_multp)
                  if (rest_Ncsf .eq. 0) cycle
                  do parti_multp = parti_Nopen + 1, 1, -2
                    ! loop over initial multiplicity for the participator part
                    parti_Ncsf = num_csf(parti_Nopen, parti_multp)
                    do partf_multp = parti_multp - 1, parti_multp + 1, 2
                      ! loop over final multiplicity for the participator part
                      if (partf_multp .eq. 0) cycle
                      partf_Ncsf = num_csf(partf_Nopen, partf_multp)
                      if (partf_Ncsf .eq. 0) cycle
                      parti_two_S = parti_multp - 1
                      partf_two_S = partf_multp - 1
                      rest_two_S = rest_multp - 1
                      ! triangular inequality
                      if (parti_two_S + rest_two_S .lt. initial_two_S) cycle
                      if (abs(parti_two_S - rest_two_S) .gt. initial_two_S) cycle
                      if (partf_two_S + rest_two_S .lt. final_two_S) cycle
                      if (abs(partf_two_S - rest_two_S) .gt. final_two_S) cycle
                      ! this is the big factor that couples all angular momenta
                      spin_factor = ZERO
                      do parti_two_M = -parti_two_S, parti_two_S
                        ! do partf_two_M=-partf_two_S,partf_two_S
                        partf_two_M = parti_two_M + final_two_S - initial_two_S
                        ! do rest_two_M=-rest_two_S,rest_two_S
                        rest_two_M = final_two_S - partf_two_M
                        spin_factor = spin_factor + &
                          cleb(partf_two_S, partf_two_M, rest_two_S, rest_two_M, final_two_S, final_two_S)* &
                          cleb(parti_two_S, parti_two_M, rest_two_S, rest_two_M, initial_two_S, initial_two_S)* &
                          cleb(partf_two_S, partf_two_M, 1, parti_two_M - partf_two_M, parti_two_S, parti_two_M)/ &
                          cleb(partf_two_S, partf_two_S, 1, parti_two_S - partf_two_S, parti_two_S, parti_two_S)
                        ! end do
                        ! end do
                      end do
                      spin_factor = spin_factor**2
                      if (spin_factor .eq. ZERO) cycle

                      if (P%print_level .gt. 2) then
                        write (P%OUTPUT, "('# ','imultp',2x,i4,i4,i4,x,'fmultp',2x,i4,i4,i4,es20.7)") &
                          parti_multp, rest_multp, initial_multp, partf_multp, rest_multp, final_multp, &
                          spin_factor*WE_factor*dble(rest_Ncsf)/dble(initial_Ncsf)
                      end if
                      Rate = Rate + TWO*PI*WE_factor*dble(rest_Ncsf)/dble(initial_Ncsf)*spin_factor &
                        *auger_rate_spin_minimal(val1, val2, initial_occ(hole, io), initial_occ(val1, io), &
                        initial_occ(val2, io), parti_multp, partf_multp, ampl_12, ampl_21)
                    end do ! partf_multp
                  end do ! parti_multp
                end do! rest_multp

                ChannelRate = ChannelRate + Rate
                if (P%print_level .gt. 1 &
                  ) write (P%OUTPUT, "('#   Auger channel:  ','imultp',i4,x,'fmultp',i4,3x,i4,' +', i4, ' -> ',i4,' + ',es13.5,' eV ', es13.5) &
                &                ") initial_multp, final_multp, val1, val2, hole, AE_energy*AU2EV, Rate
                if (.not. P%yes_all_multps) &
                  TotalRate = TotalRate + Rate
                call add_Process(P, F, S, "Auger", Rate, AE_energy, hole, val1, val2, &
                  initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=1, &
                  final_multp=final_multp, final_csf_num=1)
              end do! final_multp
              if (P%print_level .gt. 1 &
                ) write (P%OUTPUT, "('#   Auger channel:  ','imultp',i4,x,'fmultp','  * ',3x,i4,' +', i4, ' -> ',i4,' + ',es13.5,' eV ', es13.5) &
              &              ") initial_multp, val1, val2, hole, AE_energy*AU2EV, ChannelRate
              if (P%yes_all_multps) then
                SpinavgRate = SpinavgRate + initial_multp*initial_Ncsf*ChannelRate
                Spinavg_n = Spinavg_n + initial_multp*initial_Ncsf
              end if
            end do! initial_multp

            if (P%yes_all_multps) then
              SpinavgRate = SpinavgRate/Spinavg_n
              TotalRate = TotalRate + SpinavgRate
            end if

            if (P%print_level .gt. 1 .and. P%yes_all_multps) then
              write (P%OUTPUT, &
                "('#   Auger channel:  ','(spin averaged)       ',3x,i4,' +', i4, ' -> ',i4,' + ',es13.5,' eV ', es13.5)") &
                val1, val2, hole, AE_energy*AU2EV, SpinavgRate
            end if

          end do ! val2 loop
        end do ! val1 loop
      end do ! hole loop

      AvgTotalRate = AvgTotalRate + TotalRate*weight_io(io)
      if (P%print_level .gt. 1) then
        if (P%yes_all_multps) then
          write (P%OUTPUT, "('#  Spin Averaged Total Augerrate: ',es13.5,' a.u. (lifetime=',a,') ',i4)") &
            TotalRate, trim(lifetime(TotalRate)), weight_io(io)
        else
          write (P%OUTPUT, "('#  Total Augerrate: ',es13.5,' a.u. (lifetime=',a,')')") &
            TotalRate, trim(lifetime(TotalRate))
        end if
      end if

    end do! io

    AvgTotalRate = AvgTotalRate/sum(weight_io(:))
    if (P%print_level .gt. 1) &
      write (P%OUTPUT, "('# Average Total Augerrate: ',es13.5,' a.u. (lifetime=',a,') ')") &
      AvgTotalRate, trim(lifetime(AvgTotalRate))

    deallocate (initial_occ)
    deallocate (weight_io)
    deallocate (ampl_12)
    deallocate (ampl_21)
    do at = 1, P%N_nuc
      ! special case for hydrogen: no auger
      if (A(at)%charge .eq. 1) cycle
      deallocate (RI(at)%K)
      deallocate (RI(at)%n1)
      deallocate (RI(at)%l1)
      deallocate (RI(at)%n2)
      deallocate (RI(at)%l2)
      deallocate (RI(at)%n3)
      deallocate (RI(at)%l3)
      deallocate (RI(at)%l4)
      deallocate (RI(at)%R)
      if (allocated(RI(at)%Pmin)) then
        deallocate (RI(at)%Pmin)
        deallocate (RI(at)%Omin)
        deallocate (RI(at)%Smin)
        deallocate (RI(at)%min_info)
      end if
    end do
  contains
    !----------------------------------------------------------------------------
    ! print out lifetime in fs or attosec.
    !----------------------------------------------------------------------------
    function lifetime(decay_rate) result(string)
      real(long), intent(in) :: decay_rate                ! in a.u.
      real(long) :: tau
      character(len=15) :: str, string
      tau = ONE/decay_rate*AU2SEC/1e-15_long        ! to femtosec.
      if (tau .gt. ONE) then
        write (str, "(f12.1,' fs')") tau
      else
        tau = tau*1000_long                       ! to attosec.
        write (str, "(f12.1,' as')") tau
      end if
      string = trim(adjustl(str))
    end function lifetime
    !>----------------------------------------------------------------------------
    !! function auger_rate_spin_minimal
    !! returns the auger rate for a minimal electronic spin configuration hole,val1,val2 -> hole+1,val1-1,val2-1
    !! with initial and final spin multiplicity given by initital_multp and final_multp
    !! auger integrals <val1,val2|hole, l,m > and <val2,val1|hole, l,m > are given in ampl_12 and ampl_21 arrays, respectively.
    !! ampl_xy is a double array of dimension (Natom,(lmax+1)**2)
    !----------------------------------------------------------------------------
    function auger_rate_spin_minimal(val1, val2, occ_hole, &
      occ_val1, occ_val2, initial_multp, final_multp, ampl_12, ampl_21) result(minimal_rate)
      real(long) :: minimal_rate
      integer, intent(in) :: val1, val2, occ_hole, occ_val1, occ_val2, initial_multp, final_multp
      real(long), intent(in) :: ampl_12(:, :), ampl_21(:, :)
      integer :: Natom
      minimal_rate = ZERO
      Natom = size(ampl_12, 1)
      if (val1 .eq. val2) then
        if ((initial_multp .eq. 1 .and. final_multp .eq. 2) .or. &
          (initial_multp .eq. 2 .and. final_multp .eq. 1)) then
          minimal_rate = minimal_rate + ONE*sum(sum(ampl_12(:, :)**2, dim=1), dim=1)
        end if
      else
        if (occ_hole .eq. 0 .and. occ_val1 .eq. 2 .and. occ_val2 .eq. 2) then
          if ((initial_multp .eq. 1 .and. final_multp .eq. 2)) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_12(:, :) + ampl_21(:, :))**2 + &
              THREE*(ampl_12(:, :) - ampl_21(:, :))**2, dim=1))
          end if
        elseif (occ_hole .eq. 0 .and. occ_val1 .eq. 1 .and. occ_val2 .eq. 1) then
          if (initial_multp .eq. 1 .and. final_multp .eq. 2) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_12(:, :) + ampl_21(:, :))**2, dim=1))
          else if (initial_multp .eq. 3 .and. final_multp .eq. 2) then
            minimal_rate = minimal_rate + ONE*sum(sum((ampl_12(:, :) - ampl_21(:, :))**2, dim=1))
          end if
        else if (occ_hole .eq. 0 .and. occ_val1 .eq. 2 .and. occ_val2 .eq. 1) then
          if (initial_multp .eq. 2 .and. final_multp .eq. 1) then
            minimal_rate = minimal_rate + HALF*sum(sum((TWO*ampl_12(:, :) - ampl_21(:, :))**2, dim=1))
          elseif (initial_multp .eq. 2 .and. final_multp .eq. 3) then
            minimal_rate = minimal_rate + ONE*sum(sum((ampl_21(:, :))**2, dim=1))
          end if
        elseif (occ_hole .eq. 0 .and. occ_val1 .eq. 1 .and. occ_val2 .eq. 2) then
          if (initial_multp .eq. 2 .and. final_multp .eq. 1) then
            minimal_rate = minimal_rate + HALF*sum(sum((TWO*ampl_21(:, :) - ampl_12(:, :))**2, dim=1))
          elseif (initial_multp .eq. 2 .and. final_multp .eq. 3) then
            minimal_rate = minimal_rate + ONE*sum(sum((ampl_12(:, :))**2, dim=1))
          end if
        else if (occ_hole .eq. 1 .and. occ_val1 .eq. 2 .and. occ_val2 .eq. 2) then
          if ((initial_multp .eq. 2 .and. final_multp .eq. 1)) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_12(:, :) + ampl_21(:, :))**2, dim=1))
          else if ((initial_multp .eq. 2 .and. final_multp .eq. 3)) then
            minimal_rate = minimal_rate + ONE*sum(sum((ampl_12(:, :) - ampl_21(:, :))**2, dim=1))
          end if
        else if (occ_hole .eq. 1 .and. occ_val1 .eq. 1 .and. occ_val2 .eq. 1) then
          if ((initial_multp .eq. 2 .and. final_multp .eq. 1)) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_12(:, :) + ampl_21(:, :))**2 + THREE*(ampl_12(:, :) - ampl_21(:, :))**2, dim=1))
          end if
        else if (occ_hole .eq. 1 .and. occ_val1 .eq. 2 .and. occ_val2 .eq. 1) then
          if ((initial_multp .eq. 1 .and. final_multp .eq. 2)) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_12(:, :) - TWO*ampl_21(:, :))**2, dim=1))
          else if ((initial_multp .eq. 3 .and. final_multp .eq. 2)) then
            minimal_rate = minimal_rate + ONE*sum(sum((ampl_12(:, :))**2, dim=1))
          end if
        else if (occ_hole .eq. 1 .and. occ_val1 .eq. 1 .and. occ_val2 .eq. 2) then
          if ((initial_multp .eq. 1 .and. final_multp .eq. 2)) then
            minimal_rate = minimal_rate + HALF*sum(sum((ampl_21(:, :) - TWO*ampl_12(:, :))**2, dim=1))
          else if ((initial_multp .eq. 3 .and. final_multp .eq. 2)) then
            minimal_rate = minimal_rate + ONE*sum(sum(ampl_21(:, :)**2, dim=1))
          end if
        end if ! occ_val1 ==  .and. occ_val2 ==
      end if ! val1==val2
    end function auger_rate_spin_minimal

  end subroutine calcAugerRates_spin_avg

end module
