!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module averaging
  !! author: Ludger Inhester
  !!
  !! averaging module 
  !! This module consists of tools that are required for averaging occupations scheme
  !! for averaging within different MO subspaces as specified by the strutures P%av_occ(:)

  use read_param
  use wavefunction
  use elec_config
  use anglib
  use constants
  implicit none
  private
  public averaged_occs, is_equivalent, is_occupied, stat_factor_occupied, stat_factor_not_occupied, &
  stat_factor_pair_occupied, has_hole, is_pair_occupied, is_equivalent_occ
contains
  !> subroutine averaged_occs:
  !! returns an array of occupations
  !! that describe the occupation patterns
  !! which are averaged, using the averaging scheme
  !! given in F%av_occ(:) and the occupation N_occ(:)
  subroutine averaged_occs(F, N_occ, occs) 
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer,allocatable, intent(out) :: occs(:,:)
    integer :: Noccs, Nel, Norb, Ndouble, Noccs_av
    integer :: j,k,kold,l
    Noccs=1
    if(allocated(F%av_occ))then    
      do j=1,size(F%av_occ)
        Noccs_av=0
        Nel=sum(F%N_occ(F%av_occ(j)%MO))
        Norb=size(F%av_occ(j)%MO,1)
        do Ndouble=max(Nel-Norb,0),floor(HALF*Nel)
          Noccs_av=Noccs_av+nint(max(binom(Norb,Ndouble)*binom(Norb-Ndouble,Nel-2*Ndouble),ONE))
        end do                        
        Noccs=Noccs*Noccs_av
      end do
      !write(STDERR,*) 'Noccs',Noccs
      allocate(occs(size(F%N_occ,1),Noccs))
      k=1
      occs(:,k)=N_occ(:)
      do j=1,size(F%av_occ)
        ! write(STDERR,*) 'av_occ',F%av_occ(j)%MO
        ! do i=1,k
        !   write(STDERR,"(99i1)") occs(:,i)
        ! end do
        kold=k
        do l=1,kold
          call excitations(occs(:,l),F%av_occ(j)%MO(:), occs,k)
        end do
      end do

    end if
  contains
    !> just render all occupations by performing all sorts of excitations
    !! in the av_occ space
    recursive subroutine excitations(Nocc,MO, occs,index)
      integer, intent(in)  :: Nocc(:)       
      integer, intent(in)  :: MO(:)       
      integer, intent(inout) :: occs(:,:)     
      integer, intent(inout) :: index
      integer :: temp_occ(size(occs,1))
      integer :: i1,i2
      integer :: k
      do i1 =1,size(MO)
        if(Nocc(MO(i1)) == 0)cycle
        do i2 =1,size(MO)
          if(i1==i2)cycle
          if(Nocc(MO(i2)) == 2)cycle
          if(index > size(occs,2)) exit
          temp_occ(:)=Nocc(:)
          temp_occ(MO(i1))=Nocc(MO(i1))-1
          temp_occ(MO(i2))=Nocc(MO(i2))+1
          do k=1,index
            if (ALL(temp_occ(:)==occs(:,k)) ) exit
          end do
          if(ALL(temp_occ(:)==occs(:,k)) )cycle
          index=index+1
          occs(:,index)=temp_occ
          call excitations(occs(:,index),MO,occs(:,:),index)          
        end do
      end do      
    end subroutine excitations    
  end subroutine averaged_occs

  !>function is_occupied:
  !! P: Parameter structure 
  !! N_occ: occ array 
  !! oi: orbital index
  !! returns TRUE if there might be a statistical chance for occupancy in orbital oi
  !! otherwise FALSE
  function is_occupied(P,F,N_occ,oi) result(res)
    type(WF_data), intent(in) :: F
    type(Param),intent(in) :: P
    integer,intent(in) :: N_occ(:)
    integer :: oi
    logical :: res
    integer :: j
    res=.FALSE.
    if(allocated(F%av_occ) .and. P%yes_average_rates)then
      do j=1,size(F%av_occ)
        if(ANY(F%av_occ(j)%MO==oi) .and. sum(N_occ(F%av_occ(j)%MO(:))) > 0) res=.TRUE.
      end do
      if(.NOT. res .and. N_occ(oi) > 0) res=.TRUE.
    else
      if(N_occ(oi) == 0) res=.FALSE.
      if(N_occ(oi) > 0) res=.TRUE.
    end if
  end function is_occupied

  !>function has_hole:
  !! P: Parameter structure 
  !! N_occ: occ array 
  !! oi: orbital index
  !! returns TRUE if there might be a statistical chance for a hole in orbital oi
  !! otherwise FALSE
  function has_hole(P,F,N_occ,oi) result(res)
    ! returns if there is statistically a hole in MO oi 
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer, intent(in) :: oi
    logical :: res
    integer :: j
    res=.FALSE.
    if(allocated(F%av_occ) .and. P%yes_average_rates)then
      do j=1,size(F%av_occ)
        if(ANY(F%av_occ(j)%MO==oi) .and. sum(N_occ(F%av_occ(j)%MO(:))) /= 2*size(F%av_occ(j)%MO(:)) ) then
          res=.TRUE.
          return
        end if
      end do
    end if
    if(N_occ(oi) /= 2) then
      res=.TRUE.
      return
    end if
  end function has_hole

  !> returns true if occ1 and occ2 are equivalent through averaging
  function is_equivalent_occ(F,occ1,occ2) result(res)
    type(WF_data), intent(in) :: F
    integer, intent(in) :: occ1(:),occ2(:)
    logical :: res
    integer ::i,j
    res=.TRUE.
    if (any(occ1(:)/=occ2(:)))then
      res=.FALSE.
    end if
    if(allocated(F%av_occ) .and. .not. res)then
      res=.TRUE.
      compare_loop: do i=1,size(occ1)
        do j=1,size(F%av_occ)
          if(any(F%av_occ(j)%MO(:)==i))then
            if(sum(occ1(F%av_occ(j)%MO(:)))/= sum(occ2(F%av_occ(j)%MO(:))) )then
              res=.FALSE.
              exit compare_loop
            end if
          else
            if(occ1(i)/=occ2(i))then
              res=.FALSE.
              exit compare_loop
            end if
          end if
        end do
      end do compare_loop
    end if
  end function is_equivalent_occ

  !> function is_equivalent:  
  !! N_occ: occ array 
  !! part: orbital index
  !! hole: orbital index
  !! returns TRUE if orbital part and orbital hole are equivalent
  !! otherwise FALSE
  function is_equivalent(P,F,part,hole) result(res)
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F

    integer :: part,hole
    logical :: res
    integer :: j
    res=.FALSE.    
    if( hole==part ) then
      res=.TRUE.
      return
    end if
    if(allocated(F%av_occ) .and. P%yes_average_rates)then
      do j=1,size(F%av_occ)
        if(ANY(F%av_occ(j)%MO==hole) .and. ANY(F%av_occ(j)%MO==part)  ) then
          res=.TRUE.
          return
        end if
      end do
    end if
  end function is_equivalent

  !> function is_pair_occupied:  
  !! P: Parameter structure 
  !! N_occ: occ array 
  !! oi: orbital index
  !! oj: orbital index
  !! returns TRUE if there is a statistical chance that orbital oi and oj are both occupied
  !! otherwise FALSE
  function is_pair_occupied(P,F,N_occ,oi,oj) result(res)
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer, intent(in) :: oi,oj
    logical :: res
    integer :: i,j
    if(.NOT. is_occupied(P,F,N_occ,oi) .or. .NOT. is_occupied(P,F,N_occ,oj) ) then
      res=.FALSE.
      return
    end if
    if(allocated(F%av_occ).and. P%yes_average_rates)then       
      ! check if both are in an av_occ space
      do i=1,size(F%av_occ)
        ! check if both are in the same av_occ space
        if(ANY(F%av_occ(i)%MO==oi) .and. ANY(F%av_occ(i)%MO==oj))then
          if( sum(N_occ(F%av_occ(i)%MO(:))) < 2 )then
            res=.FALSE.
          else
            res=.TRUE.
          end if
          return
          ! check if oi is av_occ space i oj is in av_occ space j
        else if(ANY(F%av_occ(i)%MO==oi) )then
          do j=1,size(F%av_occ)
            if(ANY(F%av_occ(j)%MO==oj))then
              if( sum(N_occ(F%av_occ(i)%MO(:))) < 1 .or. &
                sum(N_occ(F%av_occ(j)%MO(:))) < 1 ) then
                res=.FALSE.
              else
                res=.TRUE.
              end if
              return
            end if
          end do ! j loop
        end if
      end do! i loop

      ! check if one is in ov_occ space
      do i=1,size(F%av_occ)
        if(ANY(F%av_occ(i)%MO==oi) )then
          if( sum(N_occ(F%av_occ(i)%MO(:))) < 1 .or.  N_occ(oj) < 1 )then
            res=.FALSE.
          else
            res=.TRUE.
          end if
          return
        else if (ANY(F%av_occ(i)%MO==oj) )then
          if( sum(N_occ(F%av_occ(i)%MO(:))) < 1 .or.  N_occ(oi) < 1 )then
            res=.FALSE.
          else
            res=.TRUE.
          end if
          return
        end if
      end do! i loop
    end if! if av_occ allocated

    ! neither oi nor oj are in an av_occ space
    if( oi==oj .and. N_occ(oi)<2) then
      res=.FALSE.
    else if( oi/=oj .and. N_occ(oi)+N_occ(oj) < 2 ) then
      res=.FALSE.
    else
      res=.TRUE.
    end if
    return
  end function is_pair_occupied


  !> function stat_factor_pair_occupied
  !! P: Parameter structure 
  !! N_occ: occ array 
  !! oi: orbital index
  !! oj: orbital index
  !! returns the statistical factor that
  !! a spin-orbital pair oi, oj is occupied
  function stat_factor_pair_occupied(P,F,N_occ,oi,oj) result(prob)
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer :: oi,oj
    real(long) :: F_occ(size(N_occ))
    real(long) :: prob
    integer :: i

    ! if i and j are non equivalent
    ! chance that oi and oj are occupied is given by
    ! Nocc_i/Nspinorb  * Nocc_j/Nspinorb 
    ! if i and j are equivalent
    ! chance that oi and oj are occupied is given by
    ! Nocci/Nspinorb * (Nocci-1)/(Nspinorb-1)

    ! prepare fractional occupation number
    f_occ=ONE*N_occ
    if(allocated(F%av_occ))then
      do i=1,size(F%av_occ)
        F_occ(F%av_occ(i)%MO(:))=ONE/size(F%av_occ(i)%MO)*(sum(N_occ(F%av_occ(i)%MO)))
      end do
    end if


    ! orbitals are the same oi=oj
    if(oi==oj)then
      if(allocated(F%av_occ).and. P%yes_average_rates)then
        do i=1,size(F%av_occ)
          ! oi is in av_occ(i)
          if(ANY(F%av_occ(i)%MO==oi))then
            prob=( sum(N_occ(F%av_occ(i)%MO(:)))    ) / ( TWO * size(F%av_occ(i)%MO(:))      )*&
              ( sum(N_occ(F%av_occ(i)%MO(:))) - 1) / ( TWO * size(F%av_occ(i)%MO(:)) - ONE) 
            return
          end if
        end do
      end if ! P%av_occ allocated
      ! if we have reached here oi is not in any av_occ(i) n/2 * (n-1)/(2-1) = n/2 * (n-1)
      prob= HALF*N_occ(oi) * ( N_occ(oi)- ONE) 
      return      
      ! orbitals are not the same oi/=oj
    else 
      if(allocated(F%av_occ) .and. P%yes_average_rates)then
        do i=1,size(F%av_occ)
          ! the orbitals are in the same av_occ
          if(ANY(F%av_occ(i)%MO==oi) .and. ANY(F%av_occ(i)%MO==oj))then                
            prob=( sum(N_occ(F%av_occ(i)%MO(:)))    ) / ( TWO * size(F%av_occ(i)%MO(:))      )*&
              ( sum(N_occ(F%av_occ(i)%MO(:)))  - 1  ) / ( TWO * size(F%av_occ(i)%MO(:)) -ONE ) 
            return
          end if ! if oi in av_occ(i)                    
        end do ! i loop        
        ! if we reach to this point orbitals are not in the same av_occ
        prob=HALF*F_occ(oi)*HALF*F_occ(oj)
      else        
        prob= HALF*N_occ(oi) * HALF * N_occ(oj)
      end if ! if allocate(P%av_occ)
      return
    end if ! oi /= oj

  end function stat_factor_pair_occupied

  !> return the statistical factor that orbital oi is not occupied
  !! basically 1 - f_occ/2
  function stat_factor_not_occupied(P,F,N_occ,oi) result(factor)  
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer, intent(in) :: oi
    integer :: j
    real(long) :: factor
    if(allocated(F%av_occ) .and. P%yes_average_rates)then
      do j=1,size(F%av_occ)
        if(ANY(F%av_occ(j)%MO==oi) ) then
          factor=ONE-HALF*sum(N_occ(F%av_occ(j)%MO))/size(F%av_occ(j)%MO)
          return
        end if
      end do
    end if
    factor=ONE-HALF*N_occ(oi)
  end function stat_factor_not_occupied


  !> return the statistical factor that orbital oi is not occupied
  !! basically f_occ/2
  function stat_factor_occupied(P,F,N_occ,oi) result(factor)  
    type(Param),intent(in) :: P
    type(WF_data), intent(in) :: F
    integer,intent(in) :: N_occ(:)
    integer, intent(in) :: oi
    integer :: j
    real(long) :: factor
    if(allocated(F%av_occ) .and. P%yes_average_rates)then
      do j=1,size(F%av_occ)
        if(ANY(F%av_occ(j)%MO==oi) ) then
          factor=HALF*sum(N_occ(F%av_occ(j)%MO))/size(F%av_occ(j)%MO)
          return
        end if
      end do
    end if
    factor=HALF*N_occ(oi)
  end function stat_factor_occupied

end module averaging
