!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module basis_func
  !! author: Sang-Kil Son
  !!
  !! BASIS_FUNC module to define basis-set functions from atomic orbitals
  !! represented on multicenter molecular grids
  !!   written by Son, Sang-Kil and Hao, Yajiang in Mar. 2014
  !!
  !! USAGE:
  !!   type(Basis) :: B
  !!
  !! Ref) Becke, J. Chem. Phys. 88, 2547 (1988)
  !!      Becke & Dickson, J. Chem. Phys. 89, 2993 (1988)
  !!      Becke & Dickson, J. Chem. Phys. 92, 3610 (1990)
  !!      Dickson & Becke, J. Chem. Phys. 99, 3898 (1993)
  !!
  use constants
  use read_param
  use atom
  use atom_in_mol
  use molecular_grid
  use eigen_lapack
  use interpolate
  use sph_harmonics
  use matrix
  use gaussian
  use packed_index
  implicit none
  ! constants for symmetry
  integer, parameter :: N_SYMOP = 8
  integer, parameter :: SYMOP_E = 1       !! 1 E
  integer, parameter :: SYMOP_C2X = 2     !! 2 C2X
  integer, parameter :: SYMOP_C2Y = 3     !! 3 C2Y
  integer, parameter :: SYMOP_C2Z = 4     !! 4 C2Z
  integer, parameter :: SYMOP_I = 5       !! 5 i
  integer, parameter :: SYMOP_SIGXY = 6   !! 6 SIGXY
  integer, parameter :: SYMOP_SIGXZ = 7   !! 7 SIGXZ
  integer, parameter :: SYMOP_SIGYZ = 8   !! 8 SIGYZ
  character(len=5), parameter :: sym_op_name(8) = (/'E    ', 'C2X  ', 'C2Y  ', 'C2Z  ', 'i    ', 'SIGXY', 'SIGXZ', 'SIGYZ'/)
  interface 
  function memory_map_file(n)  bind(c, name="memory_map_file") result(ptr)
    use :: iso_c_binding, only:c_long, c_int, c_ptr, c_size_t
    implicit none
    integer(c_size_t) :: n
    type(c_ptr) :: ptr
  end function memory_map_file

  function memory_unmap_file(n, ptr)  bind(c, name="memory_unmap_file") result(ret)
    use :: iso_c_binding, only:c_long, c_int, c_ptr, c_size_t
    implicit none
    integer(c_size_t) :: n
    type(c_ptr) :: ptr
    integer(c_int) :: ret
  end function memory_unmap_file
  end interface
  type :: Basis
    !! basis information
    logical :: yes_initialized = .false.
    integer :: N_basis                          !! total # of basis functions
    integer :: Nr_basis                      !! total # of basis functions projected out
    real(long), allocatable :: func(:, :)        !! (1:N_grid,1:N_basis)
    real(long), allocatable :: dfdR(:, :, :)      !! (1:N_grid,1:N_basis,3)
    real(long), allocatable :: D2f(:, :)         !! (1:N_grid,1:N_basis)
    logical, allocatable :: yes_skip(:, :)       !! (1:N_nuc,1:N_basis)
    !! skip if all basis functions (and its deriv.) are zero for a given atom A
    !! it helps to speed up because we know that B%func are zero if r > A%rmax
    !! and we don't need to evaluate anything beyond A%rmax
    !! --> initialized in place_AO_on_Grid3D()
    integer, allocatable :: info(:, :)           ! (1:N_basis,5)
    !! basis function info
    !!   B%info(i,1) : atomic index for the i-th basis
    !!   B%info(i,2) : nuclear charge for the i-th basis
    !!   B%info(i,3) : n of AO for the i-th basis
    !!   B%info(i,4) : l of AO for the i-th basis
    !!   B%info(i,5) : m of AO for the i-th basis
    character(len=10), allocatable :: label(:)  ! (1:N_basis)
    !
    ! Overlap matrix and quantities derived from it, (1:N_basis,1:N_basis)
    !
    real(long), allocatable :: S(:, :) !! overlap matrix
    real(long), allocatable :: SSqrt(:, :) !! sqrt overlap matrix
    real(long), allocatable :: SInvSqrt(:, :) !! sqrt of inverse overlap matrix
    real(long), allocatable :: SInv(:, :) !! inverse overlap matrix 
    real(long), allocatable :: X(:, :)
    real(long), allocatable :: XXtS(:, :)
    real(long), allocatable :: HKin(:, :) !! Kinetic operator matrix, (1:N_basis,1:N_basis)
    real(long), allocatable :: Vext(:, :) !! Nuclear potential (1:N_basis,1:N_basis)
    real(long), allocatable :: H0(:, :) !! Kinetic + nuclear potential
    real(long), pointer :: teint(:) => NULL() !! Two-Electron integrals
    logical :: yes_mmap = .False.
    !>
    !! yes_overlap(k_nuc,ij):
    !!   whether a multiplication of two basis functions (i and j)
    !!   contributes to the k_nuc-th atomic sphere [k_nuc=1...N_nuc]
    !!   If this variable is yes, then the k_nuc-th atom will be included in
    !!   multicenter integration involving the i-th and j-th basis functions.
    !!   If no, then it is not included, which means "save the CPU time!"
    !!
    !!   Here k_nuc=0 means consideration of all atomic spheres.
    !!   It can be used for a simple and quick condition.
    !!
    !!   The index ij is defined by ij = i+j(j-1)/2 from a packed storage scheme
    !!   for an upper trianglular matrix, [ij=1...N_basis(N_basis+1)/2].
    !!   SEE: http://www.netlib.org/lapack/lug/node123.html
    !!
    !! When i == j, yes_overlap is determined by EPS_overlap_single.
    !! When i /= j, yes_overlap is determined by EPS_overlap.
    !!!   Most of cases, EPS_overlap_single should be smaller than EPS_overlap.
    !!   Therefore, a small number (enough for EPS_overlap_single) is used for
    !!   both EPS_overlap_single and EPS_overlap even though they can be
    !!   chosen separately.
    !! - default in read_param.f90: EPS_overlap = EPS_overlap_single = 1e-6
    !!
    logical, allocatable :: yes_overlap(:, :)
    !    logical, allocatable :: yes_elec_stat(:,:)
    !    logical, allocatable :: yes_coulomb(:,:)
    !    real(long), allocatable :: coulomb(:,:)
    real(long), allocatable :: overlap(:, :)
    real(long) :: EPS_overlap, EPS_overlap_single
    real(long) :: R_max_overlap
    !    real(long) :: R_max_coulomb
    !>
    !! numerical accuracy for single-center and multicenter integration
    !!                        and integration optimization by truncation
    !!
    real(long) :: error_single, error_multi, error_trunc

    integer, allocatable :: first_bf_on_atom(:)
    integer, allocatable :: num_bf_on_atom(:)

    ! symmetry part
    logical :: yes_symmetry(N_SYMOP)
    integer, allocatable :: irrep_prod(:, :)
    real(long), allocatable :: sym_bf(:, :)      !! conversion matrix from symmetric bf to atomic orbitals
    integer, allocatable    :: irrep_sym_bf(:)  !! index of irreducible representation per sym bf
    integer, allocatable    :: irrep_sym_bfr(:)  !! index of irreducible representation per sym bf
    character(len=5) :: sym_label !! Schoenflies symbol
    integer :: nirrep  !! number of irrep
    character(len=4), allocatable :: irrep_name(:) !! name of irrep

    real(long) :: minimal_S_eigenvalue !! minimal overlap matrix eigenvalue
    real(long) :: minimal_S_remaining_eigenvalue !! minimal overlap matrix eigenval that is not projected out

    ! Gaussian basis set data
    real(long), allocatable :: env(:)
    integer, allocatable :: atm(:, :)
    integer, allocatable :: bas(:, :)
    integer, allocatable :: first_bf_of_shell(:)
  end type Basis

  private
  public :: Basis, construct_Basis, purge_Basis, braket, orthonormalizeOrbitals,&
    calc_trans_dipole_moment_bf, precalculate_twoel_GTO, print_calc_info, calc_quadrupole_moment_bf
contains

!>----------------------------------------------------------------------------
!! construct (or reconstruct) B from A and G
!! all necessary informations are obtained from
!! - A : atomic orbitals of individual atoms
!! - G : molecular grids
!!----------------------------------------------------------------------------
  subroutine construct_Basis(A, G, B, EPS_overlap, EPS_overlap_single, yes_sym, gto_file_path, yes_gradient, basisSetOverlapLim)
    type(Atom_set), intent(in) :: A(:) !! atom set
    type(Grid3D), intent(in) :: G !! grid 
    type(Basis), intent(out) :: B !! basis
    real(long), intent(in), optional :: EPS_overlap, EPS_overlap_single, basisSetOverlapLim !! truncation parameters
    logical, intent(in), optional :: yes_sym !! use symmetry?
    logical, intent(in), optional :: yes_gradient !! prepare for calculating gradients?
    character(len=*), intent(in), optional :: gto_file_path !! path to file with gto informations 

    logical :: sym, gradient
    integer :: i, j, js, je, N_nuc
    integer :: bf, sh
    real(long) :: basisSetLim
    integer, allocatable :: infoT(:, :)
    basisSetLim = ZERO
    if (present(basisSetOverlapLim)) &
      basisSetLim = basisSetOverlapLim

    sym = .FALSE.
    if (present(yes_sym)) sym = yes_sym
    gradient = .FALSE.
    if (present(yes_gradient)) gradient = yes_gradient
    N_nuc = size(A)
    if (B%yes_initialized .eqv. .false.) then
      B%error_single = ZERO       ! to be decided in construct_overlap_diagonal
      B%error_multi = ZERO       ! to be decided in construct_overlap_diagonal
      B%error_trunc = ZERO       ! to be decided in construct_overlap_*
      if (present(EPS_overlap)) B%EPS_overlap = EPS_overlap
      if (present(EPS_overlap_single)) B%EPS_overlap_single = EPS_overlap_single
      !
      ! count the number of basis functions
      !
      if (present(gto_file_path)) then
        call testLibCint
        call construct_Gaussian_basis_set(A, B%env, B%atm, B%bas, gto_file_path)
        B%N_basis = sum(2*B%bas(2, :) + 1)
        ! set first_bf_of_shell
        if (B%yes_initialized .eqv. .false.) &
          allocate (B%first_bf_of_shell(size(B%bas, 2)))
        bf = 1
        do sh = 1, size(B%bas, 2) ! loop over shells
          B%first_bf_of_shell(sh) = bf
          bf = bf + 2*B%bas(2, sh) + 1
        end do
      else
        B%N_basis = 0
        do i = 1, N_nuc
          B%N_basis = B%N_basis + sum(2*A(i)%l_AO(:) + 1)
        end do
      end if
      !
      ! memory allocation for the basis functions and derived quantities
      !
      allocate (B%func(G%N_grid, B%N_basis), B%D2f(G%N_grid, B%N_basis))
      allocate (B%yes_skip(G%N_nuc, B%N_basis))
      allocate (B%info(B%N_basis, 5), B%label(B%N_basis))
      if (gradient) &
        allocate (B%dfdR(G%N_grid, B%N_basis, 3))
      !
      ! memory allocation for the overlap matrix and derived quantities, and
      !                       the kinetic operator matrix
      !
      allocate (B%S(B%N_basis, B%N_basis))
      allocate (B%Sinv(B%N_basis, B%N_basis))
      allocate (B%SInvSqrt(B%N_basis, B%N_basis))
      allocate (B%SSqrt(B%N_basis, B%N_basis))
      allocate (B%HKin(B%N_basis, B%N_basis))
      allocate (B%yes_overlap(0:G%N_nuc, B%N_basis*(B%N_basis + 1)/2))
      allocate (B%overlap(1:G%N_nuc, B%N_basis*(B%N_basis + 1)/2))

      allocate (B%first_bf_on_atom(N_nuc))
      allocate (B%num_bf_on_atom(N_nuc))

      if (sym) then
        allocate (B%sym_bf(B%N_basis, B%N_basis))
        allocate (B%irrep_sym_bf(B%N_basis))
      end if
    end if
    !
    ! set up basis functions
    !
    B%yes_skip(:, :) = .false.     ! as default, don't skip any for basis f. eval.
    ! initialize in place_AO_on_Grid3D()
    if (present(gto_file_path)) then
      call place_GTO_on_Grid3D(G, A, B, yes_gradient=gradient)
      call construct_overlap_GTO(B)
    else
      B%first_bf_on_atom(:) = 0
      B%num_bf_on_atom(:) = 0
      j = 1
      do i = 1, N_nuc
        if (A(i)%N_Nao .eq. 0) cycle
        call place_AO_on_Grid3D(G, A(i), B, j, yes_gradient=gradient)
        js = j; je = j + sum(2*A(i)%l_AO(:) + 1) - 1
        B%info(js:je, 1) = i         ! i-th nucleus
        j = je + 1
        B%first_bf_on_atom(i) = js
        B%num_bf_on_atom(i) = je - js + 1
      end do
      !
      ! construct the overlap matrix (B%S and B%yes_overlap)
      !
      call construct_overlap_diagonal(G, B)
      call construct_overlap_offdiagonal(G, B)
      !
      ! construct yes_elec_stat matrix based on internuclear distances
      !
    end if

    !
    ! prepare a label for each basis function
    !
    if (B%yes_initialized .eqv. .false.) then
      allocate (infoT(size(B%info, 2), size(B%info, 1)))
      infoT = transpose(B%info)
      do j = 1, B%N_basis
        !B%label(j) = prepare_label( B%info(j,:) )
        B%label(j) = prepare_label(infoT(:, j))
      end do
      deallocate (infoT)
    end if

    if (sym) then
      call construct_symmetry_bf(A, B)
    end if

    !call construct_InvSqrtOverlapMatrix( B )
    !call construct_SqrtOverlapMatrix( B )
    !B%SInv(:,:) = MATMUL( B%SInvSqrt(:,:), B%SInvSqrt(:,:) )

    call construct_SqrtOverlapMatrix(B, basisSetLim)
    !
    ! construct kinetic operator matrix
    !
    if (present(gto_file_path)) then
      call construct_hkin_matrix_GTO(B)
      allocate (B%Vext(B%N_basis, B%N_basis))
      call construct_Vext_matrix_GTO(B)
      allocate (B%H0(B%N_basis, B%N_basis))
      B%H0 = B%Hkin + B%Vext
    else
      call construct_hkin_matrix(G, B)
    end if

    ! write(STDERR,'("# S")')
    !do i=1,B%N_basis
    !  write(STDERR,'("# ",99f7.3)') B%S(i,:)
    !end do


    !
    ! now this has been initialized
    ! next time we don't need to allocate memory for B
    !
    B%yes_initialized = .true.

  contains
    function prepare_label(info) result(s)
      integer, intent(in) :: info(5)
      character(len=10) :: s, str
      integer :: i, Z, n, l, m
      i = info(1)         ! nuclear index (atomic index)
      Z = info(2)         ! nuclear charge
      n = info(3)
      l = info(4)
      m = info(5)
      write (str, "(i10)") i
      s = trim(ATOMIC_NAME(Z))//trim(adjustl(str))//' '// &
        subshell_name(n, l)
      !if ( l == 1 ) then
      !  select case ( m )
      !    case ( -1 );  s = trim(s) // 'x'
      !    case (  0 );  s = trim(s) // 'z'
      !    case (  1 );  s = trim(s) // 'y'
      !  end select
      !else if ( l > 1 ) then
      if (l .ge. 1) then
        write (str, "(i5)") abs(m)
        s = trim(s)//trim(adjustl(str))
        if (m .gt. 0) s = trim(s)//'+'
        if (m .lt. 0) s = trim(s)//'-'
      end if
    end function prepare_label

    subroutine construct_overlap_GTO(B)
      !! calculates the overlap matrix S using GTO for basis B
      type(Basis), intent(inout) :: B !! basis
      call calculate_overlap_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, B%S)
      B%yes_overlap(:, :) = .TRUE.
    end subroutine construct_overlap_GTO

    subroutine construct_hkin_matrix_GTO(B)
      !! calculates the kinetic energy matrix for given basis B
      type(Basis), intent(inout) :: B
      call calculate_Hkin_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, B%hkin)
    end subroutine construct_hkin_matrix_GTO

    subroutine construct_Vext_matrix_GTO(B)
      !! construct Vext matrix
      type(Basis), intent(inout) :: B
      call calculate_Vext_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, B%Vext)
    end subroutine construct_Vext_matrix_GTO

    subroutine construct_overlap_diagonal(G, B)
      !! calculate the overlap matrix for diagonal parts (S_ii)
      type(Grid3D), intent(in) :: G
      type(Basis), intent(inout) :: B
      real(long) :: value, error !, R_AB
      !real(long) :: r(G%N_grid)
      integer :: j, j_nuc, k_nuc, ks, ke
      integer(8) :: jj

      do j = 1, B%N_basis
        j_nuc = B%info(j, 1)
        jj = index_packed(j, j)
        B%S(j, j) = ZERO
        B%overlap(:, jj) = ZERO
        B%yes_overlap(:, jj) = .false.           ! default: not included
        error = ZERO
        loop_k_nuc: do k_nuc = 1, G%N_nuc
          if (B%yes_skip(k_nuc, j)) cycle loop_k_nuc
          !
          ! decompose the overlap function, phi_j * phi_j, into individual atoms
          !   here, value is an integrated density for each atom sphere
          !   if value >= EPS_overlap_single, then this atom sphere is included
          !   in the further multicenter integration ==> B%yes_overlap is yes
          !   otherwise, it is not included ==> B%yes_overlap is no
          !
          ks = G%i_start(k_nuc)
          ke = G%i_end(k_nuc)
          value = sum(B%func(ks:ke, j)**2*G%weight(ks:ke))
          if (j_nuc .eq. k_nuc .or. &   ! always included for the same atom
            value .ge. B%EPS_overlap_single) then
            B%S(j, j) = B%S(j, j) + value
            B%yes_overlap(k_nuc, jj) = .true.
            B%overlap(k_nuc, jj) = value

          else
            error = error + value

          end if
          !
          ! set up yes_coulomb
          !

        end do loop_k_nuc
        !
        ! assess numerical accuracy for integration optimization
        !
        if (error .gt. B%error_trunc) B%error_trunc = error
        !
        ! B%yes_overlap(0,:) can be used for a simple condition
        !
        B%yes_overlap(0, jj) = any(B%yes_overlap(1:G%N_nuc, jj))
      end do
    end subroutine construct_overlap_diagonal

    subroutine construct_overlap_offdiagonal(G, B)
      !! calculate the overlap matrix for off-diagonal parts (S_ij)
      type(Grid3D), intent(in) :: G
      type(Basis), intent(inout) :: B
      real(long) :: func(G%N_grid)
      real(long) :: value, error, overlap_value
      integer :: i, j, k_nuc, ks, ke
      integer(8) :: ij
      integer :: i_nuc, j_nuc
      logical :: yes_overlap_all
      do i = 1, B%N_basis
        i_nuc = B%info(i, 1)
        loop_j: do j = i + 1, B%N_basis           ! NOTE: i == j excluded
          j_nuc = B%info(j, 1)
          ij = index_packed(i, j)
          B%S(i, j) = ZERO
          B%yes_overlap(:, ij) = .false.           ! default: not included
          B%overlap(:, ij) = ZERO
          !
          ! analyze this overlap function (its abs value) on individual atoms
          !
          ! The following operation is the most time-consuming part
          !   func(:) = B%func(:,i) * B%func(:,j)
          !   ==> func(ks:ke) = B%func(ks:ke,i) * B%func(ks:ke,j)
          !       i.e., evaluated only for the relevant atom
          !
          error = ZERO
          yes_overlap_all = .false.
          loop_k_nuc: do k_nuc = 1, G%N_nuc
            if (B%yes_skip(k_nuc, i) .or. B%yes_skip(k_nuc, j)) cycle loop_k_nuc
            ks = G%i_start(k_nuc)
            ke = G%i_end(k_nuc)
            func(ks:ke) = B%func(ks:ke, i)*B%func(ks:ke, j)
            value = abs(sum(abs(func(ks:ke))*G%weight(ks:ke)))
            !
            ! three conditions: this atomic sphere must contain
            !   1) a large contribution from i-th basis function
            !   2) a large contribution from j-th basis function
            !   3) a large contribution from a pair of i-th and j-th basis func.
            ! WHY? just make a simple rule: take only 3) ... Sang-Kil on 2019/5/7
            !

            if (value .ge. B%EPS_overlap) then
              overlap_value = sum(func(ks:ke)*G%weight(ks:ke))
              B%S(i, j) = B%S(i, j) + overlap_value
              B%overlap(k_nuc, ij) = overlap_value
              B%yes_overlap(k_nuc, ij) = .true.
              yes_overlap_all = .true.

            else
              error = error + value

            end if
            !
            ! set up yes_coulomb
            !

          end do loop_k_nuc
          B%S(j, i) = B%S(i, j)  ! considering the symmetric matrix, always i /= j
          if (error .gt. B%error_trunc) B%error_trunc = error
          !
          ! B%yes_overlap(0,:) can be used for a simple condition
          !
          B%yes_overlap(0, ij) = yes_overlap_all

        end do loop_j
      end do
    end subroutine construct_overlap_offdiagonal

    subroutine construct_hkin_matrix(G, B)
      !! calculate the kinetic energy matrix for given basis B
      type(Grid3D), intent(in) :: G
      type(Basis), intent(inout) :: B
      integer i, j
      do i = 1, B%N_basis
        do j = i, B%N_basis
          B%HKin(i, j) = -HALF*braket_Laplacian(G, B, i, j)
          if (i .ne. j) B%HKin(j, i) = B%HKin(i, j)
        end do
      end do
    end subroutine construct_hkin_matrix

    
    !> Calculates the square root, inverse, and inverse sqrt
    !! of the symmetric matrix S
    !! and the canonical orthogonalization matrix X
    !! input: Basis and Matrix S (N_basisxN_basis)
    !! optional: projLim if overlap matrix elements are smaller than this
    !! threshold, basis set is truncated in the canonical orthogonalization
    !! by neglecting eigenvectors with eigenvalues smaller than projLim
    !!
    !! if S is symmetric, then S = V D V^T (spectral decomposition)
    !! ==> S^(1/2)  = V D^(1/2) V^T        square root
    !!     S^(-1/2) = V D^(-1/2) V^T       inverse of square root
    !!     S^(-1)   = V D^(-1) V^T         inverse
    !! matrix elements:
    !!     S^m_ij = SUM_k V_ik D^m_kk V^T_kj
    !!            = SUM_k V_ik D^m_kk V_jk
    !!
    subroutine construct_SqrtOverlapMatrix(B, projLim)
      type(Basis), intent(inout) :: B
      real(long), intent(in), optional :: projLim
      integer :: N, i, j, irrep, ni
      real(long), dimension(B%N_basis) :: Sev
      real(long), dimension(B%N_basis, B%N_basis) :: SV
      N = B%N_basis
      SV(:, :) = B%S(:, :)
      call get_eigenvalues(SV, Sev)
      B%minimal_S_eigenvalue = minval(Sev(:))
      B%minimal_S_remaining_eigenvalue = minval(Sev(:), Sev(:) .gt. projLim)
      do i = 1, N
        do j = i, N
          B%SSqrt(i, j) = sum(SV(i, :)*sqrt(Sev(:))*SV(j, :))
          B%SInvSqrt(i, j) = sum(SV(i, :)/sqrt(Sev(:))*SV(j, :))
          B%SInv(i, j) = sum(SV(i, :)/Sev(:)*SV(j, :))
          if (i .ne. j) then
            B%SSqrt(j, i) = B%SSqrt(i, j)
            B%SInvSqrt(j, i) = B%SInvSqrt(i, j)
            B%SInv(j, i) = B%SInv(i, j)
          end if
        end do
      end do
      if (allocated(B%sym_bf)) then
        ! calculate the overlap matrix in symmetrized AO's
        SV = B%sym_bf.tm.B%S.m.B%sym_bf
        call get_eigenvalues_symmetry(SV, SeV, B%irrep_sym_bf)
      end if
      ! X for canonical orthogonalization
      B%Nr_basis = B%N_basis
      if (present(projLim)) &
        B%Nr_basis = count(Sev .gt. projLim)
      allocate (B%X(B%N_basis, B%Nr_basis))
      allocate (B%XXtS(B%N_basis, B%N_basis))
      B%X = ZERO
      B%XXtS = ZERO

      if (.not. allocated(B%sym_bf)) then
        j = 1
        do
          i = maxloc(Sev, dim=1) ! determine the next largest eigenvalue
          B%X(:, j) = ONE/sqrt(Sev(i))*SV(:, i)
          Sev(i) = -HUGE(Sev(i))
          j = j + 1
          if (j .gt. B%Nr_basis) exit
        end do
      else
        allocate (B%irrep_sym_bfr(B%Nr_basis))
        j = 1
        ! determine the eigenvalue in order of size within irrep
        ni = 0
        do irrep = 1, B%nirrep
          if (all(B%irrep_sym_bf(:) .ne. irrep)) cycle
          if (present(projLim)) &
            ni = ni + count((Sev(:) .gt. projLim) .and. (B%irrep_sym_bf(:) .eq. irrep))
          if (.not. present(projLim)) &
            ni = ni + count((B%irrep_sym_bf(:) .eq. irrep))
          do
            i = maxloc(Sev, dim=1, mask=B%irrep_sym_bf(:) .eq. irrep)
            if (present(projLim)) then
              if (Sev(i) .le. projLim) exit
            end if
            B%X(:, j) = ONE/sqrt(Sev(i))*SV(:, i)
            B%irrep_sym_bfr(j) = B%irrep_sym_bf(i)
            j = j + 1
            Sev(i) = -HUGE(Sev(i))
            if (j .gt. B%Nr_basis) exit
            if (j .gt. ni) exit
          end do
        end do! irrep
      end if! sym
      if (allocated(B%sym_bf)) then
        ! go back to non-symmetrized basis
        B%X = B%sym_bf.m.B%X
      end if
      B%XXtS(:, :) = B%X.mt.B%X.m.B%S

    end subroutine construct_SqrtOverlapMatrix
  end subroutine construct_Basis

  subroutine reconstruct_Basis(A, G, B, EPS_overlap, EPS_overlap_single, yes_sym, yes_gradient)
    !! update basis structure with new geometry and or basis set
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(out) :: B
    real(long), intent(in), optional :: EPS_overlap, EPS_overlap_single
    logical, intent(in), optional :: yes_sym
    logical, intent(in), optional :: yes_gradient
    call construct_Basis(A, G, B, EPS_overlap, EPS_overlap_single, yes_sym=yes_sym, yes_gradient=yes_gradient)
  end subroutine reconstruct_Basis

  subroutine place_AO_on_Grid3D(G, A, B, i_start, yes_gradient)
    implicit none
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A
    type(Basis), intent(inout) :: B
    integer, intent(in) :: i_start
    logical, intent(in), optional :: yes_gradient
    logical :: gradient = .FALSE.
    integer :: i, j, k, l, m, N_r
    integer :: ks, ke, k_nuc
    real(long) :: x, y, z
    real(long) :: r(G%N_grid), theta(G%N_grid), phi(G%N_grid)
    real(long) :: radial_f(G%N_grid), radial_d2f(G%N_grid), radial_df(G%N_grid)
    real(long) :: r_int(A%N_r + 2), f_int(A%N_r + 2)  ! for interpolation with zeros
    real(long) :: d2f_int(A%N_r + 2)                ! for interpolation with zeros
    real(long) :: df_int(A%N_r + 2)                 ! for interpolation with zeros
    type(Spline) :: SP, SPD, SPD2
    real(long) :: sphh_real, gY(3)
    real(long) :: sin_theta, cos_theta, sin_phi, cos_phi
    logical :: yes_skip_each_grid(G%N_grid)
    if (present(yes_gradient)) gradient = yes_gradient
    !
    ! prepare Spline interpolation with zeros at both ends
    ! why spline instead of GPS_interpolate()? --> see wavefunction.f90 in XATOM
    !
    N_r = A%N_r
    r_int(1) = ZERO
    r_int(2:N_r + 1) = A%radial%r(:)
    r_int(N_r + 2) = A%rmax
    if (A%rmax .eq. ZERO) r_int(N_r + 2) = A%radial%r(N_r)*TWO     ! put an enough large r
    f_int(:) = ZERO
    df_int(:) = ZERO
    d2f_int(:) = ZERO
    call spline_create(r_int(:), f_int(:), SP)
    call spline_create(r_int(:), df_int(:), SPD)
    call spline_create(r_int(:), d2f_int(:), SPD2)
    !
    ! Cartesian coordinate -> spherical coordinate
    !
    do j = 1, G%N_grid
      x = G%x(j) - A%x
      y = G%y(j) - A%y
      z = G%z(j) - A%z
      !
      ! when periodicity is considered, shift grid positions
      ! taking into account "minimum image convention"
      ! ACHTUNG: here, translate all grid points w.r.t their position and
      !   nuclear position; otherwise, it cannot describe a torus structure!
      !   this should be applied in the same way
      !   for calculate_weight_Grid3D in molecular_grid.f90
      !
      if (G%yes_periodic) call translation_in_crystal(G, x, y, z)
      call xyz_to_sph(x, y, z, r(j), theta(j), phi(j))
    end do
    !
    ! i : basis index, i_start ... ( i_start + N_basis - 1 ) for this atom A
    ! k : index of different NAOs, 1 ... N_nao
    !
    i = i_start
    do k = 1, A%N_nao     ! N_nao: # of different orbitals
      l = A%l_AO(k)       ! l: the orbital angular momentum quantum number
      f_int(2:N_r + 1) = A%u_AO(:, k)                ! prepare interpolation of u(r)
      call spline_reuse(SP, f_int(:))           ! remember zeros at both ends
      df_int(2:N_r + 1) = A%dAO_dr(:, k)             ! prepare interpolation of du/dr(r)
      call spline_reuse(SPD, df_int(:))         ! remember zeros at both ends
      !
      ! for s orbitals (l=0), u''(0) is not zero.
      ! current remedy is to interpolate u''*r instead of u''
      !
      if (l .eq. 0) then
        d2f_int(2:N_r + 1) = A%D2_AO(:, k)*A%radial%r(:)  ! interpolate u'' * r for l=0
      else
        d2f_int(2:N_r + 1) = A%D2_AO(:, k)           ! interpolate u''
      end if
      call spline_reuse(SPD2, d2f_int(:))       ! remember zeros at both ends
      !
      ! calculation of R(r) and Del^2 R(r) with the Spline interpolation
      !   R(r) = u(r) / r
      !                        u(r)      1  d        d    u(r)       u''(r)
      !   Del^2 R(r) = Del^2 ( ---- ) = --- -- [ r^2 -- ( ---- ) ] = ------
      !                         r       r^2 dr       dr    r           r
      !
      ! Spline_interpolate( SP, r )   := u(r)
      ! Spline_interpolate( SPD2, r ) := u''(r)   for l /= 0
      !                               := u''(r)*r for l = 0
      !
      ! if you want to use GPS interpolation...
      !    radial_f(j) = GPS_interpolate( A%radial, A%u_AO(:,k), r(j) ) / r(j)
      !
      ! We are interpolating u(r), so it is always zero when r >= A%rmax.
      !
      yes_skip_each_grid(:) = .false.
      where (r(:) .ge. A%rmax) yes_skip_each_grid(:) = .true.
      do j = 1, G%N_grid
        if (yes_skip_each_grid(j)) then         ! if r >= A%rmax
          radial_f(j) = ZERO
          radial_df(j) = ZERO
          radial_d2f(j) = ZERO
        else
          radial_f(j) = spline_interpolate(SP, r(j))/r(j)
          ! d(u/r)/dr = u'/r - u/r**2 = u'/r - radial_f/r
          radial_df(j) = spline_interpolate(SPD, r(j))/r(j) - radial_f(j)/r(j)
          radial_d2f(j) = spline_interpolate(SPD2, r(j))/r(j)
          if (l .eq. 0) radial_d2f(j) = radial_d2f(j)/r(j)  ! for s orbitals
        end if
      end do
      !
      ! attach the angular part
      !
      do m = -l, l        ! m: the associated projection quantum number
        do j = 1, G%N_grid
          if (yes_skip_each_grid(j)) then       ! if r >= A%rmax
            B%func(j, i) = ZERO
            B%D2f(j, i) = ZERO
            if (gradient) B%dfdR(j, i, :) = ZERO
          else
            sphh_real = sph_harmonics_real(l, m, theta(j), phi(j))
            B%func(j, i) = radial_f(j)*sphh_real
            B%D2f(j, i) = radial_d2f(j)*sphh_real
            if (gradient) then
              sin_theta = sin(theta(j))
              cos_theta = cos(theta(j))
              sin_phi = sin(phi(j))
              cos_phi = cos(phi(j))
              ! df /dR = e_r * df/dR * Y_lm + f(r)/r * grad Y_lm
              ! for historical reasons df/dR has a negative sign
              ! e_r part
              B%dfdR(j, i, 1) = -radial_df(j)*sphh_real*sin_theta*cos_phi
              B%dfdR(j, i, 2) = -radial_df(j)*sphh_real*sin_theta*sin_phi
              B%dfdR(j, i, 3) = -radial_df(j)*sphh_real*cos_theta
              ! grad Y part
              gY(:) = grad_sph_harmonic_Y_real(l, m, theta(j), phi(j))
              B%dfdR(j, i, :) = B%dfdR(j, i, :) - radial_f(j)/r(j)*gY(:)
            end if! gradient
          end if
        end do
        !
        ! set up B%yes_skip for a given k_nuc
        ! by checking whether r(j) >= A%rmax at every grid point j within k_nuc
        ! For r > A%rmax, the basis function and its derivative are zero, so that
        !   we can skip any operation.
        !
        do k_nuc = 1, G%N_nuc
          ks = G%i_start(k_nuc)
          ke = G%i_end(k_nuc)
          if (all(yes_skip_each_grid(ks:ke))) B%yes_skip(k_nuc, i) = .true.
        end do
        !
        ! set up basis info
        !
        B%info(i, 2) = A%charge
        B%info(i, 3) = A%n_AO(k)
        B%info(i, 4) = l
        B%info(i, 5) = m
        i = i + 1         ! increase the basis index for different m
      end do
    end do
    
    call purge_Spline(SP)
    call purge_Spline(SPD)
    call purge_Spline(SPD2)
    
  end subroutine place_AO_on_Grid3D

  !> This subroutine places Gaussian orbitals on the grid
  !! and sets the array first_bf_on_atom and B%info
  subroutine place_GTO_on_Grid3D(G, A, B, yes_gradient)
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(inout) :: B
    logical, intent(in), optional :: yes_gradient
    logical :: gradient = .FALSE.
    integer :: i, j, l, m
    integer, parameter :: LMAX = 10
    integer :: sh, atom, bf
    integer, allocatable :: shells_on_atom(:, :)
    real(long) :: x, y, z
    real(long) :: r(G%N_grid), theta(G%N_grid), phi(G%N_grid)
    real(long) :: radial_f, radial_d2f, radial_df
    real(long) :: alpha, coef
    integer :: n_prim
    real(long) :: sphh_real
    real(long) :: sin_theta, cos_theta, sin_phi, cos_phi
    real(long) :: gY(3)
    if (present(yes_gradient)) gradient = yes_gradient
    allocate (shells_on_atom(size(A), 0:LMAX))
    bf = 1
    shells_on_atom(:, :) = 0
    do sh = 1, size(B%bas, 2) ! loop over shells
      atom = B%bas(1, sh) + 1 ! atom: atom index
      l = B%bas(2, sh)    ! l: the orbital angular momentum quantum number
      shells_on_atom(atom, l) = shells_on_atom(atom, l) + 1
      !
      ! shift grid positions, taking into account "minimum image convention"
      !
      do j = 1, G%N_grid
        x = G%x(j) - A(atom)%x
        y = G%y(j) - A(atom)%y
        z = G%z(j) - A(atom)%z
        if (G%yes_periodic) call translation_in_crystal(G, x, y, z)
        call xyz_to_sph(x, y, z, r(j), theta(j), phi(j))
      end do
      n_prim = B%bas(3, sh) ! number of primitives
      do m = -l, l
        do j = 1, G%N_grid ! for all grid points
          radial_f = ZERO
          radial_d2f = ZERO
          radial_df = ZERO
          do i = 1, n_prim
            alpha = B%env(B%bas(6, sh) + i)
            coef = B%env(B%bas(7, sh) + i)
            radial_f = radial_f + coef*r(j)**l*exp(-alpha*r(j)**2)
            radial_df = radial_df + coef*-2*alpha*r(j)**(l + 1)*exp(-alpha*r(j)**2)
            if (l .gt. 0) radial_df = radial_df + coef*l*r(j)**(l - 1)*exp(-alpha*r(j)**2)
            !          1/r* d2( r*f(r) ) - l*(l+1)/r^2* f(r)
            radial_d2f = radial_d2f + coef* &
              r(j)**(l)*exp(-alpha*r(j)**2) &
              *1/r(j)**2* &
              (dble(l*(l + 1)) - TWO*dble(3 + 2*l)*alpha*r(j)**2 + FOUR*r(j)**4*alpha**2)
            radial_d2f = radial_d2f + coef* &
              r(j)**(l)*exp(-alpha*r(j)**2) &
              *-dble(l*(l + 1))/r(j)**2
          end do
          sphh_real = sph_harmonics_real(l, m, theta(j), phi(j))
          B%func(j, bf) = radial_f*sphh_real
          B%D2f(j, bf) = radial_d2f*sphh_real
          if (gradient) then
            !
            ! df /dR = e_r * df/dr * Y_lm + f(r)/r * grad Y_lm
            ! for historical reasons df/dr has a negative sign
            ! e_r part
            sin_theta = sin(theta(j))
            cos_theta = cos(theta(j))
            sin_phi = sin(phi(j))
            cos_phi = cos(phi(j))
            B%dfdR(j, bf, 1) = -radial_df*sphh_real*sin_theta*cos_phi
            B%dfdR(j, bf, 2) = -radial_df*sphh_real*sin_theta*sin_phi
            B%dfdR(j, bf, 3) = -radial_df*sphh_real*cos_theta
            ! grad Y part
            gY(:) = grad_sph_harmonic_Y_real(l, m, theta(j), phi(j))
            B%dfdR(j, bf, :) = B%dfdR(j, bf, :) - radial_f/r(j)*gY(:)
          end if
        end do
        B%info(bf, 1) = atom
        B%info(bf, 2) = A(atom)%charge
        B%info(bf, 3) = shells_on_atom(atom, l) + l ! this should be the atomic n quantum number
        B%info(bf, 4) = l
        B%info(bf, 5) = m
        bf = bf + 1
      end do ! m loop
    end do! sh loop
    do i = 1, size(A)
      B%num_bf_on_atom(i) = count(B%info(:, 1) .eq. i)
      do j = 1, B%N_basis
        if (B%info(j, 1) .eq. i) then
          B%first_bf_on_atom(i) = j
          exit
        end if
      end do
    end do
    deallocate (shells_on_atom)

  end subroutine place_GTO_on_Grid3D

  subroutine dump_Basis(B, IOUT)
    type(Basis), intent(in) :: B
    integer, intent(in), optional :: IOUT
    integer :: i, OUTPUT
    OUTPUT = STDOUT; if (present(IOUT)) OUTPUT = IOUT
    write (OUTPUT, *) "Basis functions:"
    write (OUTPUT, *) "  N_basis = ", B%N_basis
    do i = 1, B%N_basis
      write (OUTPUT, *) i, B%func(:, i)
    end do
  end subroutine dump_Basis

  subroutine print_overlap_accuracy(G, B, IOUT)    
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    integer, intent(in), optional :: IOUT
    integer :: i, j, i_nuc, j_nuc, k_nuc, OUTPUT
    integer :: count_single, count_double, count_multi
    real(long) :: error, func(G%N_grid)
    real(long) :: avg_single, max_single
    real(long) :: avg_double, max_double
    real(long) :: avg_multi, max_multi
    OUTPUT = STDOUT; if (present(IOUT)) OUTPUT = IOUT
    avg_single = ZERO; max_single = ZERO; count_single = 0
    avg_double = ZERO; max_double = ZERO; count_double = 0
    avg_multi = ZERO; max_multi = ZERO; count_multi = 0
    do i = 1, B%N_basis
      i_nuc = B%info(i, 1)
      func(:) = B%func(:, i)**2
      !
      ! multicenter
      !
      error = ONE - integral(G, func(:))
      call add_error(avg_multi, max_multi, count_multi, error)
      !
      ! single-center
      !
      error = ONE - integral_single_center(G, func(:), i_nuc)
      call add_error(avg_single, max_single, count_single, error)
      !
      ! double-center; scan over all nuclei
      !
      do k_nuc = 1, G%N_nuc
        error = ONE - integral_double_center(G, func(:), i_nuc, k_nuc)
        call add_error(avg_double, max_double, count_double, error)
      end do
      !
      ! considering off-diagonal; only for the same atom (i_nuc == j_nuc)
      !
      do j = i + 1, B%N_basis
        j_nuc = B%info(j, 1)
        if (i_nuc .ne. j_nuc) cycle
        func(:) = B%func(:, i)*B%func(:, j)
        !
        ! multicenter
        !
        error = integral(G, func(:))
        call add_error(avg_multi, max_multi, count_multi, error)
        !
        ! single-center
        !
        error = integral_single_center(G, func(:), i_nuc)
        call add_error(avg_single, max_single, count_single, error)
        !
        ! double-center; scan over all nuclei
        !
        do k_nuc = 1, G%N_nuc
          error = integral_double_center(G, func(:), i_nuc, k_nuc)
          call add_error(avg_double, max_double, count_double, error)
        end do
      end do
    end do
    !
    ! calculate average
    !
    avg_single = avg_single/dble(count_single)
    avg_double = avg_double/dble(count_double)
    avg_multi = avg_multi/dble(count_multi)
    !
    ! print accuracy data
    !
    write (OUTPUT, "('# Multicenter integration accuracy: ',a)") &
      "valid only for normalized basis set"
    write (OUTPUT, "('#   ',25x,2a13)") "avg. error", "max. error"
    write (OUTPUT, "('#   single-center for S_ii : ',2es13.5)") &
      avg_single, max_single
    write (OUTPUT, "('#   double-center for S_ij : ',2es13.5)") &
      avg_double, max_double
    write (OUTPUT, "('#   multicenter   for all  : ',2es13.5)") &
      avg_multi, max_multi
    write (OUTPUT, "('#   minimal overlap matrix eigenvalue = ',es13.5)") &
      B%minimal_S_eigenvalue
    if (B%minimal_S_eigenvalue .lt. ZERO) then
      write (STDERR, "(a)") "ERROR: Negative overlap matrix eigenvalue is found."
      write (STDERR, "(a)") "       It may be related to linear dependence."
      stop
    end if
    write (OUTPUT, "('#')")
  contains
    subroutine add_error(e_avg, e_max, counter, error)
      real(long), intent(inout) :: e_avg, e_max
      integer, intent(inout) :: counter
      real(long), intent(in) :: error
      e_avg = e_avg + abs(error)                          ! accumulate avg
      if (abs(error) .gt. e_max) e_max = abs(error)        ! find max.
      counter = counter + 1                               ! increase counter
    end subroutine add_error
  end subroutine print_overlap_accuracy

  !>----------------------------------------------------------------------------
  !! braket(): \( < i | O | j > \) where i, j indicate i-th and j-th basis function
  !! braket_Laplacian(): \( < i | \nabla^2 | j > \)
  !!
  !! This integration is crucial to reduce computational time.
  !!   There are many overlaps among atomic grids. Note that a multiplication of
  !!   two AOs is NOT decomposed into a sum of two atomic-centered grid sets.
  !!   Even one AO is NOT represented with one atomic-centered grid set, because
  !!   of many overlaps.  Thus, a multiplication of two AOs (overlap matrix)
  !!   SHOULD NOT BE CALCULATED simply with integral_double_center()!!!
  !! Here is how to calculate multicenter integration. --- Method (d)
  !!   First, decompose into individual atomic-centered grid set (atomic sphere).
  !!   Second, check if i-th basis function and j-th basis function have a
  !!           large contribution (overlap) to each atomic sphere.
  !!           Also check if a multiplication of i-th and j-th basis functions
  !!           has a large contribution to each atomic sphere.
  !!           All of them are taken into account via B%yes_overlap(k_nuc,ij).
  !!   Third, sum up only if there is a large contribution.
  !!----------------------------------------------------------------------------
  function braket(G, B, i, j, f) result(value)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    integer, intent(in) :: i, j
    real(long), intent(in) :: f(:)
    real(long) :: value

    !integer :: i_nuc, j_nuc
    !integer :: is, ie
    integer :: k_nuc
    integer(8) :: ij
    integer :: ks, ke

    !
    ! (a) over all molecular grids: the simplest one
    !     surprisingly the simplest one is the fastest one...
    !
    !value = integral( G, B%func(:,i) * B%func(:,j) * f(:) )
    !
    ! (b) over only two atomic grids: as the simplest approximation,
    !     we can use double-center integral no matter how it's implemented
    !     --- currently, it integrates over only two atomic grids
    !
    !i_nuc = B%info(i,1)
    !j_nuc = B%info(j,1)
    !value = integral_double_center( G, B%func(:,i) * B%func(:,j) * f(:), &
    !                                i_nuc, j_nuc )
    !
    ! (c) using neighboring atoms for given grid points
    !
    !value = sum( B%func(:,i) * B%func(:,j) * f(:) * G%weight(:), &
    !        MASK=( G%yes_neighbor(:,i_nuc) .or. G%yes_neighbor(:,j_nuc) ) )
    !
    ! (d) using overlap arguments for individual atomic spheres
    !
    value = ZERO
    ij = index_packed(i, j)   ! index for packed storage of upper triang.
    do k_nuc = 1, G%N_nuc
      if (B%yes_overlap(k_nuc, ij)) then
        ks = G%i_start(k_nuc)
        ke = G%i_end(k_nuc)
        value = value + sum(G%weight(ks:ke)* &
          B%func(ks:ke, i)*B%func(ks:ke, j)*f(ks:ke))
      end if
    end do
  end function braket

  function braket_Laplacian(G, B, i, j) result(value)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    integer, intent(in) :: i, j
    real(long) :: value
    !integer :: i_nuc, j_nuc
    integer :: k_nuc, ks, ke
    integer(8) :: ij
    !
    ! because the current Laplacian implementation is NOT symmetric,
    ! we are taking an average of < i | Del^2 j > and < Del^2 i | j >
    !
    !
    ! (a) over all molecular grids
    !     Here evaluations of Laplacian matrix elements are done only once.
    !     Therefore, we may use the most accurate and reliable one: (a)
    !
    !value = integral( G, B%func(:,i)*B%D2f(:,j) + B%D2f(:,i)*B%func(:,j) ) / TWO
    !
    ! (b) over only two atomic grids
    !
    !i_nuc = B%info(i,1)
    !j_nuc = B%info(j,1)
    !value = integral_double_center( G, &
    !          B%func(:,i) * B%D2f(:,j) + B%D2f(:,i) * B%func(:,j), &
    !          i_nuc, j_nuc ) / TWO
    !
    ! (c) using neighboring atoms for given grid points
    !
    !value = sum( (B%func(:,i)*B%D2f(:,j)+B%D2f(:,i)*B%func(:,j)) * G%weight(:), &
    !        MASK=( G%yes_neighbor(:,i_nuc) .or. G%yes_neighbor(:,j_nuc) ) ) / TWO
    !
    ! (d) using overlap arguments for individual atomic spheres
    !
    value = ZERO
    ij = index_packed(i, j)   ! index for packed storage of upper triang.
    do k_nuc = 1, G%N_nuc
      if (B%yes_overlap(k_nuc, ij)) then
        ks = G%i_start(k_nuc)
        ke = G%i_end(k_nuc)
        value = value + sum(G%weight(ks:ke)* &
          (B%func(ks:ke, i)*B%D2f(ks:ke, j) + &
          B%D2f(ks:ke, i)*B%func(ks:ke, j)))
      end if
    end do
    value = value/TWO
  end function braket_Laplacian

  subroutine memmap(n, fptr)
    use, intrinsic :: iso_c_binding
    integer(c_size_t), intent(in) :: n
    real(long), pointer :: fptr(:)
    type(c_ptr) :: cptr
    integer :: i
    real(long) :: dummy
    cptr =  memory_map_file(n * c_sizeof(dummy))
    CALL C_F_POINTER(cptr, fptr, [n])
  end subroutine memmap

  subroutine memunmap(fptr)
    use, intrinsic :: iso_c_binding
    integer(c_size_t) :: n
    real(long), pointer, intent(inout) :: fptr(:)
    integer :: ret
    type(c_ptr) :: cptr
    n = size(fptr)
    cptr = C_LOC(fptr)
    ret = memory_unmap_file(n, cptr)
  end subroutine memunmap

  subroutine purge_Basis(B)
    !! deallocate basis structure    
    type(Basis), intent(inout) :: B
    if (allocated(B%func)) deallocate (B%func)
    if (allocated(B%D2f)) deallocate (B%D2f)
    if (allocated(B%info)) deallocate (B%info)
    if (allocated(B%label)) deallocate (B%label)
    if (allocated(B%S)) deallocate (B%S)
    if (allocated(B%Sinv)) deallocate (B%Sinv)
    if (allocated(B%X)) deallocate (B%X)
    if (allocated(B%Sinvsqrt)) deallocate (B%SInvSqrt)
    if (allocated(B%SSqrt)) deallocate (B%SSqrt)
    if (allocated(B%Hkin)) deallocate (B%HKin)
    if (allocated(B%Vext)) deallocate (B%Vext)
    if (allocated(B%H0)) deallocate (B%H0)
    if (associated(B%teint)) then
      if (B%yes_mmap) then
        call memunmap(B%teint)
        nullify(B%teint)
      else
        deallocate (B%teint)
        nullify(B%teint)
      end if
    end if
    if (allocated(B%yes_overlap)) deallocate (B%yes_overlap)
    if (allocated(B%sym_bf)) deallocate (B%sym_bf)
    if (allocated(B%irrep_sym_bf)) deallocate (B%irrep_sym_bf)
    if (allocated(B%irrep_name)) deallocate (B%irrep_name)
    if (allocated(B%first_bf_on_atom)) deallocate (B%first_bf_on_atom)
    if (allocated(B%num_bf_on_atom)) deallocate (B%num_bf_on_atom)

    if (allocated(B%atm)) deallocate (B%atm)
    if (allocated(B%bas)) deallocate (B%bas)
    if (allocated(B%env)) deallocate (B%env)
    if (allocated(B%vext)) deallocate (B%vext)
    if (allocated(B%first_bf_of_shell)) deallocate (B%first_bf_of_shell)
    if (allocated(B%dfdR)) deallocate (B%dfdR)
    if (allocated(B%overlap)) deallocate (B%overlap)
    if (allocated(B%XXtS)) deallocate (B%XXtS)
    if (allocated(B%yes_skip)) deallocate (B%yes_skip)
  end subroutine purge_Basis

  !> This subroutine construct symmetric basis functions and stores them in B%sym_bf
  !! The following steps are done
  !! 1. define how the atomic orbitals transform under symmetry operations
  !! 2. investigate what kind of symmetries hold for our molecule
  !! 3. identify the Schoenflies symbol and specify the characters of the irrep's
  !! 4. construct the symmetry orbitals for each ired. rep. B%sym_bf(:,:)
  !!
  !!    After construction:
  !!    B%yes_symmetry: array of length N_SYMOP, =.TRUE. if molecule is symmetric under this sym. op.
  !!    B%sym_label: contains the Schoenflies symbol
  !!    B%nirrep:    number of irred. rep.
  !!    B%irrep_name: (array of length B%nirrep ) contains the name of each irred. rep.
  !!    B%sym_bf(i,j): (B%N_basisxB%N_basis) ith (unsymmetric) basis function's coefficient
  !!                  for the jth symmetric basis function
  !!    B%irrep_sym_bf: (array of length B%N_basis) B%irrep_sym_bf(j) specifies for symmetric bf j
  !!                     with coefficients sym_bf(:j) the index of the irred. rep.
  !!
  subroutine construct_symmetry_bf(A, B)
    type(Basis), intent(inout) :: B
    type(Atom_set), intent(in) :: A(:)

    integer :: N_nuc, i, j

    integer, parameter :: lmax = 4
    integer :: l, lm, op, bf, atom, atom_pair, bf_pair, n_sym_bf
    integer, allocatable :: sym_lm(:, :)
    integer, allocatable :: irrep_char(:, :)

    real(long) :: norm
    logical :: duplicate
    integer, allocatable :: bf_combination(:, :)

    integer :: symx(N_SYMOP)
    integer :: symy(N_SYMOP)
    integer :: symz(N_SYMOP)
    integer, allocatable :: char_symop(:) ! symmetry operations in character table
    integer :: IOUT
    IOUT = STDOUT         ! currently, printing here is fixed with STDOUT

    allocate (bf_combination(B%N_basis, B%N_basis))

    N_nuc = size(A)
    allocate (sym_lm(N_SYMOP, (lmax + 1)**2))

    ! We use N_SYMOP=8 symmetry operations:
    ! Define how primitive functions transform under these 8 sym operations
    symx(SYMOP_E) = 1; symy(SYMOP_E) = 1; symz(SYMOP_E) = 1
    symx(SYMOP_C2X) = 1; symy(SYMOP_C2X) = -1; symz(SYMOP_C2X) = -1
    symx(SYMOP_C2Y) = -1; symy(SYMOP_C2Y) = 1; symz(SYMOP_C2Y) = -1
    symx(SYMOP_C2Z) = -1; symy(SYMOP_C2Z) = -1; symz(SYMOP_C2Z) = 1
    symx(SYMOP_I) = -1; symy(SYMOP_I) = -1; symz(SYMOP_I) = -1
    symx(SYMOP_SIGXY) = 1; symy(SYMOP_SIGXY) = 1; symz(SYMOP_SIGXY) = -1
    symx(SYMOP_SIGXZ) = 1; symy(SYMOP_SIGXZ) = -1; symz(SYMOP_SIGXZ) = 1
    symx(SYMOP_SIGYZ) = -1; symy(SYMOP_SIGYZ) = 1; symz(SYMOP_SIGYZ) = 1

    ! Specify how atomic orbitals transform:
    sym_lm(:, :) = 0
    lm = 0
    sym_lm(:, lm + 1) = 1 ! s
    lm = lm + 1
    do op = 1, 8
      sym_lm(op, lm + 1) = symy(op)  ! py
      sym_lm(op, lm + 2) = symz(op)  ! pz
      sym_lm(op, lm + 3) = symx(op)  ! px
    end do
    lm = lm + 2*(1) + 1
    do op = 1, 8
      sym_lm(op, lm + 1) = symx(op)*symy(op)  ! dxy
      sym_lm(op, lm + 2) = symy(op)*symz(op)  ! dyz
      sym_lm(op, lm + 3) = 1                  ! dz2
      sym_lm(op, lm + 4) = symx(op)*symz(op)  ! dxz
      sym_lm(op, lm + 5) = 1                  ! dx2-y2
    end do
    lm = lm + 2*(2) + 1
    do op = 1, 8
      sym_lm(op, lm + 1) = symy(op)                    ! fy(3x2-y2)
      sym_lm(op, lm + 2) = symx(op)*symy(op)*symz(op)  ! fxyz
      sym_lm(op, lm + 3) = symy(op)                    ! fyz2
      sym_lm(op, lm + 4) = symz(op)                    ! fz3
      sym_lm(op, lm + 5) = symx(op)                    ! fxz2
      sym_lm(op, lm + 6) = symz(op)                    ! fz(x2-y2)
      sym_lm(op, lm + 7) = symx(op)                    ! fx(x2-3y2)
    end do
    lm = lm + 2*(3) + 1
    do op = 1, 8
      sym_lm(op, lm + 1) = symx(op) * symy(op)         ! g xy(x**2-y**2)
      sym_lm(op, lm + 2) = symy(op) * symz(op)         ! g yz(3x**2-y**2)
      sym_lm(op, lm + 3) = symx(op) * symy(op)         ! g xy(7z**2-r**2)
      sym_lm(op, lm + 4) = symy(op)                    ! g y (7x**3 -r**2)
      sym_lm(op, lm + 5) = 1                           ! g (35z**4-30z**2r**2+3r**4)
      sym_lm(op, lm + 6) = symx(op)                    ! g x (7x**3 - 3zr**2)
      sym_lm(op, lm + 7) = 1                           ! g (x**2-y**2) * (7x**2-r**2)
      sym_lm(op, lm + 8) = symx(op) * symz(op)         ! g xz (x**2-3y**2)
      sym_lm(op, lm + 9) = 1                           ! g x**2(x**2*-3y**2) - y**2(3x**2-y**2)
    end do
    ! investigate what kind of symmetries hold for our molecule
    B%yes_symmetry(:) = .TRUE.
    do bf = 1, B%N_basis
      atom = B%info(bf, 1)
      do op = 1, N_SYMOP
        if (.NOT. B%yes_symmetry(op)) cycle
        atom_pair = A(atom)%sym_op_pair(op)
        if (atom_pair .eq. -1) then
          B%yes_symmetry(op) = .FALSE.
          cycle
        end if
      end do
    end do
    ! write(IOUT,'("# ",a)',advance="no") 'The following symmetry operations are used:'
    ! do i=1,8
    !   if(.NOT. B%yes_symmetry(i)) cycle
    !   write(IOUT,'(1x,a5)',advance="no") adjustr(sym_op_name(i))
    ! end do
    ! write(IOUT,'()')
    ! identify the Schoenflies symbol and specify the characters of the irrep's
    if (B%yes_symmetry(SYMOP_C2X) .and. B%yes_symmetry(SYMOP_C2Y) .and. B%yes_symmetry(SYMOP_C2Z) .and. &
      B%yes_symmetry(SYMOP_I) .and. B%yes_symmetry(SYMOP_SIGXY) .and. &
      B%yes_symmetry(SYMOP_SIGXZ) .and. B%yes_symmetry(SYMOP_SIGYZ) &
      ) then
      ! D2h symmetry
      B%nirrep = 8
      B%sym_label = "D2h"
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_C2Z, SYMOP_C2Y, SYMOP_C2X, SYMOP_I, SYMOP_SIGXY, SYMOP_SIGXZ, SYMOP_SIGYZ/)
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1, 1, 1, 1, 1, 1, 1/)
      irrep_char(2, :) = (/1, 1, -1, -1, 1, 1, -1, -1/)
      irrep_char(3, :) = (/1, -1, 1, -1, 1, -1, 1, -1/)
      irrep_char(4, :) = (/1, -1, -1, 1, 1, -1, -1, 1/)
      irrep_char(5, :) = (/1, 1, 1, 1, -1, -1, -1, -1/)
      irrep_char(6, :) = (/1, 1, -1, -1, -1, -1, 1, 1/)
      irrep_char(7, :) = (/1, -1, 1, -1, -1, 1, -1, 1/)
      irrep_char(8, :) = (/1, -1, -1, 1, -1, 1, 1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2, 3, 4, 5, 6, 7, 8/)
      B%irrep_prod(2, :) = (/2, 1, 4, 3, 6, 5, 8, 7/)
      B%irrep_prod(3, :) = (/3, 4, 1, 2, 7, 8, 5, 6/)
      B%irrep_prod(4, :) = (/4, 3, 2, 1, 8, 7, 6, 5/)
      B%irrep_prod(5, :) = (/5, 6, 7, 8, 1, 2, 3, 4/)
      B%irrep_prod(6, :) = (/6, 5, 8, 7, 2, 1, 4, 3/)
      B%irrep_prod(7, :) = (/7, 8, 5, 6, 3, 4, 1, 2/)
      B%irrep_prod(8, :) = (/8, 7, 6, 5, 4, 3, 2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'Ag ', 'B1g', 'B2g', 'B3g', 'Au ', 'B1u', 'B2u', 'B3u'/)
    else if (B%yes_symmetry(SYMOP_C2Z) .and. B%yes_symmetry(SYMOP_C2Y) .and. B%yes_symmetry(SYMOP_C2X)) then
      ! D2 symmetry
      B%nirrep = 4
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_C2Z, SYMOP_C2Y, SYMOP_C2X/)
      B%sym_label = "D2 "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1, 1, 1/)
      irrep_char(2, :) = (/1, 1, -1, -1/)
      irrep_char(3, :) = (/1, -1, 1, -1/)
      irrep_char(4, :) = (/1, -1, -1, 1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2, 3, 4/)
      B%irrep_prod(2, :) = (/2, 1, 4, 3/)
      B%irrep_prod(3, :) = (/3, 4, 1, 2/)
      B%irrep_prod(4, :) = (/4, 3, 2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'A1', 'B1', 'B2', 'B3'/)
    else if (B%yes_symmetry(SYMOP_C2Z) .and. B%yes_symmetry(SYMOP_SIGXZ) .and. B%yes_symmetry(SYMOP_SIGYZ)) then
      ! C2v symmetry (assuming z axis aligned)
      B%nirrep = 4
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_C2Z, SYMOP_SIGXZ, SYMOP_SIGYZ/)
      B%sym_label = "C2v"
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1, 1, 1/)
      irrep_char(2, :) = (/1, 1, -1, -1/)
      irrep_char(3, :) = (/1, -1, 1, -1/)
      irrep_char(4, :) = (/1, -1, -1, 1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2, 3, 4/)
      B%irrep_prod(2, :) = (/2, 1, 4, 3/)
      B%irrep_prod(3, :) = (/3, 4, 1, 2/)
      B%irrep_prod(4, :) = (/4, 3, 2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'A1', 'A2', 'B1', 'B2'/)
    else if (B%yes_symmetry(SYMOP_C2Z)) then
      ! C2 symmetry (assuming z axis aligned)
      B%nirrep = 2
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_C2Z/)
      B%sym_label = "C2 "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1/)
      irrep_char(2, :) = (/1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2/)
      B%irrep_prod(2, :) = (/2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'A', 'B'/)
    else if (B%yes_symmetry(SYMOP_I)) then
      ! Ci symmetry
      B%nirrep = 2
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_I/)
      B%sym_label = "Ci "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1/)
      irrep_char(2, :) = (/1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2/)
      B%irrep_prod(2, :) = (/2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'Ag', 'Au'/)
    else if (B%yes_symmetry(SYMOP_SIGXZ)) then
      ! Cs symmetry xz plane
      B%nirrep = 2
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_SIGXZ/)
      B%sym_label = "Cs "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1/)
      irrep_char(2, :) = (/1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2/)
      B%irrep_prod(2, :) = (/2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'Ap ', 'App'/)
    else if (B%yes_symmetry(SYMOP_SIGXY)) then
      ! Cs symmetry xy plane
      B%nirrep = 2
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_SIGXY/)
      B%sym_label = "Cs "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1/)
      irrep_char(2, :) = (/1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2/)
      B%irrep_prod(2, :) = (/2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'Ap ', 'App'/)
    else if (B%yes_symmetry(SYMOP_SIGYZ)) then
      ! Cs symmetry yz plane
      B%nirrep = 2
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E, SYMOP_SIGYZ/)
      B%sym_label = "Cs "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1, 1/)
      irrep_char(2, :) = (/1, -1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1, 2/)
      B%irrep_prod(2, :) = (/2, 1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'Ap ', 'App'/)
    else
      ! C1 symmetry
      B%nirrep = 1
      allocate (char_symop(B%nirrep))
      char_symop(:) = (/SYMOP_E/)
      B%sym_label = "C1 "
      allocate (irrep_char(B%nirrep, B%nirrep))
      irrep_char(1, :) = (/1/)
      allocate (B%irrep_prod(B%nirrep, B%nirrep))
      B%irrep_prod(1, :) = (/1/)
      allocate (B%irrep_name(B%nirrep))
      B%irrep_name(:) = (/'A'/)
    end if

    ! write(IOUT,'("# ",a,a)') 'Character table:',B%sym_label
    ! write(IOUT,'("# ",a5,8(1x,a5))') "irrep",adjustr(sym_op_name(char_symop(:)))
    ! do i=1,B%nirrep
    !   write(IOUT,'("# ",a5,8(1x,i5))') B%irrep_name(i),irrep_char(i,:)
    ! end do

    ! construct the symmetry orbitals:
    ! for each ired. rep. i
    ! project out the basis function (bf) combinations according the character of the ired. rep. given in irrep_char(i,op)

    ! loop over bf, apply all sym operations, i.e.
    ! bf -> sym_lm( char_symop(op), lm ) * bf_pair
    ! add them up multiplied with the character of the sym. op. in this ired. rep. irrep_char(i,op)
    ! -> bf_combination(n_sym_bf,:) contains integer combination of basisfunction
    ! Example: bf_combination(j,2)=bf_combination(j,4)=2 otherwise bf_combination=0
    !          -> linear combination of the 4th and the 6th bf: sym_bf(j,2) = sym_bf(j,4) = 1/sqrt(2)
    !

    n_sym_bf = 1
    do i = 1, B%nirrep ! for each irred. rep.
      do bf = 1, B%N_basis ! loop over each basis function
        lm = 0
        do l = 0, B%info(bf, 4) - 1
          lm = lm + 2*l + 1
        end do
        lm = lm + l + B%info(bf, 5) + 1
        if (lm .gt. (lmax + 1)**2) then
          write (STDERR, *) 'symmetry yet only works for basisfunction lmax <= ', lmax
          stop
        end if
        if (n_sym_bf .gt. B%N_basis) exit
        atom = B%info(bf, 1)
        bf_combination(n_sym_bf, :) = 0
        duplicate = .FALSE.
        do op = 1, size(char_symop)
          ! apply symmetry operation op:
          atom_pair = A(atom)%sym_op_pair(char_symop(op))
          bf_pair = bf - B%first_bf_on_atom(atom) + B%first_bf_on_atom(atom_pair)
          ! add up bf multiplied with irrep_char(i,op)
          bf_combination(n_sym_bf, bf_pair) = bf_combination(n_sym_bf, bf_pair) + sym_lm(char_symop(op), lm)*irrep_char(i, op)
        end do
        ! We may have no symmetric function for the current irrep from this bf
        if (ALL(bf_combination(n_sym_bf, :) .eq. 0)) cycle
        ! We may have also a duplicate symmetric function based on this bf
        do j = 1, n_sym_bf - 1
          if (ALL(bf_combination(n_sym_bf, :) .eq. bf_combination(j, :) )) duplicate = .TRUE.
          if (ALL(bf_combination(n_sym_bf, :) .eq. -bf_combination(j, :) )) duplicate = .TRUE.
        end do
        if (duplicate) cycle
        ! finally: contruct the linear combination in sym_bf
        norm = sqrt(dble(sum(bf_combination(n_sym_bf, :)**2)))
        B%irrep_sym_bf(n_sym_bf) = i
        B%sym_bf(:, n_sym_bf) = dble(bf_combination(n_sym_bf, :))/norm

        n_sym_bf = n_sym_bf + 1
      end do
    end do

    deallocate (char_symop)
    deallocate (irrep_char)
    deallocate (sym_lm)
    deallocate (bf_combination)
  end subroutine construct_symmetry_bf

  subroutine precalculate_twoel_GTO(P, B)
    !! subroutine to precalculate all 2el matrix elements
    type(Param), intent(in) :: P
    type(Basis), intent(inout) :: B
    integer(long) :: nij, nijkl
    integer :: err
    real(long) :: dummy
    nij = B%N_basis*(B%N_basis + 1)/2
    nijkl = nij*(nij + 1)/2
    if(P%yes_verbose)&
      write(P%OUTPUT, "('# allocating ',f20.1, ' Mb for teints')" ) nijkl * sizeof(dummy) / 1024. / 1024.
    if(P%yes_mmap_teints)then
      call memmap(nijkl, B%teint)
      B%yes_mmap = .TRUE.
    else
      if (associated(B%teint)) then
        deallocate (B%teint)
        nullify(B%teint)
      end if
      allocate (B%teint(nijkl), stat=err)
      if (err .gt. 0) then
        write (STDERR, *) 'Error allocating', int8(nij)*int8((nij + 1)/2), '* double memory'
        stop
      end if
      B%yes_mmap = .False.
    end if
    if(P%yes_verbose)&
      write(P%OUTPUT, "('# done allocating ')" )
    call calculate_teints_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, B%N_basis, B%teint, &
      accuracy=P%EPS_teint)
  end subroutine precalculate_twoel_GTO

  !>----------------------------------------------------------------------------
  !! print grid parameters and basis function (numerical atomic orbital) info
  !! it is supposed to be in read_param.f90, but it requires A(:) and G
  !! wait a minute... even it doesn't contain F!!!  why should it be here?
  !! it has beend moved to basis_func
  !!----------------------------------------------------------------------------
  subroutine print_calc_info(P, A, G, B)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    integer :: OUTPUT
    integer :: i, l, N_basis, MAX_len
    integer :: N_total, N_eval
    character(len=3) :: s
    OUTPUT = P%OUTPUT
    !
    ! information of multicenter molecular grids
    !
    if (P%print_level .gt. 1) then
      write (OUTPUT, "('# Multicenter molecular grids: ')")
      write (OUTPUT, "('#   radial grid type:  ',a)") trim(P%type_radial)
      write (OUTPUT, "('#   angular grid type: ',a)") trim(P%type_angular)
      write (OUTPUT, "('#   total number of molecular grids = ',i10)") G%N_grid
      write (OUTPUT, "('#   total number of basis functions = ',i10)") B%N_basis
      write (OUTPUT, "('#   atomic grids and numerical atomic orbitals info:')")
      MAX_len = 0
      do i = 1, G%N_nuc
        l = len(trim(P%nao_name(i)))
        if (l .gt. MAX_len) MAX_len = l
      end do
      write (s, "(i3)") max(MAX_len + 2, 10)
      write (OUTPUT, "('#   ',6x,a6,2a6,a8,a6,a"//trim(s)//",a5,2x,a)") &
        "N_r", "L", "rmax", "lmax", "N_ang", "basis", "N_b", "file"

      !N_basis(:) = 0
      !MAX_lmax = size(P%N_nao_table,dim=1) - 1
      do i = 1, G%N_nuc
        !do l = 0, MAX_lmax  ! counting # of basis functions per atom
        !  N_basis(i) = N_basis(i) + P%N_nao_table(l,i) * ( 2*l + 1 )
        !end do
        N_basis = sum(2*A(i)%l_AO(:) + 1)
        write (OUTPUT, "('#   ',a6,i6,f6.2,f6.1,i8,i6,a"//trim(s)//",i5,2x,a)") &
          A(i)%label, &                             ! atomic element and index
          A(i)%N_r, A(i)%mapping_L, A(i)%rmax, &    ! radial grid info
          A(i)%lmax_ang, A(i)%N_ang, &              ! angular grid info
          trim(P%nao_name(i)), N_basis, trim(P%nao_file_nuc(i))
      end do
      write (OUTPUT, "('#')")
    end if
    !if ( P%N_basis /= sum( N_basis(:) ) ) stop "Something wrong in N_basis!"
    !
    ! information of multicenter integration accuracy
    ! this is printed out only in the verbose mode
    ! because it requires to perform additional integrations
    ! F print_overlap_accuracy(), defined in basis_func.f90
    !
    if (P%print_level .gt. 1 .and. .not. P%yes_gaussian) then
      if (P%yes_verbose) call print_overlap_accuracy(G, B, P%OUTPUT)
    end if
    !
    ! information of how to optimize multicenter integration
    !                when the overlap matrix is constructed
    !
    if (P%print_level .gt. 1) then
      write (OUTPUT, "('# Multicenter integration optimization using Q_ij^A:')")
      write (OUTPUT, "('#   epsilon for Q_ij on each atomic grid    = ',es13.1)") &
        B%EPS_overlap
      write (OUTPUT, "('#   epsilon for Q_ii on each atomic grid    = ',es13.1)") &
        B%EPS_overlap_single
      write (OUTPUT, "('#   max error in integration optimization   = ',es13.5)") &
        B%error_trunc

      write (OUTPUT, "('#')")
      write (OUTPUT, "('#   - N_atom   : N of neighboring atoms per each basis function')")
      write (OUTPUT, "('#   - N_overlap: N of overlap matrix elements = N_basis * (N_basis + 1) / 2')")
      write (OUTPUT, "('#   - N_sphere : N of atomic grids used in integration = N_b*(N_b+1)/2*N_atom')")
      write (OUTPUT, "('#   - N_all    : all numbers without integration optimization')")
      write (OUTPUT, "('#   - N_eval   : N of actually evaluated with integration optimization')")
      write (OUTPUT, "('#   ',10x,3a10)") "N_all", "N_eval", "%"
      !
      ! average neighboring atoms
      !
      N_total = G%N_nuc
      N_eval = 0
      do i = 1, B%N_basis
        N_eval = N_eval + count(B%yes_overlap(1:, index_packed(i, i)))
      end do
      write (OUTPUT, "('#   ',a10,i10,2f10.2)") "N_atom", &
        N_total, dble(N_eval)/dble(B%N_basis), &
        dble(N_eval)/dble(B%N_basis)/dble(N_total)*100_long
      !
      ! overlap matrix argument
      !
      !N_total = B%N_basis**2                 ! considering all matrix elements
      N_total = size(B%yes_overlap, dim=2)  ! considering only upper triangular
      N_eval = count(B%yes_overlap(0, :))
      write (OUTPUT, "('#   ',a10,2i10,f10.2,2x,a)") "N_overlap", &
        N_total, N_eval, dble(N_eval)/dble(N_total)*100_long, "(sparsity)"
      !
      ! more details: overlap value on each atomic sphere
      !
      !N_total = B%N_basis**2 * G%N_nuc      ! considering all matrix elements
      N_total = size(B%yes_overlap, dim=2)*G%N_nuc
      N_eval = count(B%yes_overlap(1:, :))
      write (OUTPUT, "('#   ',a10,2i10,f10.2)") "N_sphere", &
        N_total, N_eval, dble(N_eval)/dble(N_total)*100_long
      write (OUTPUT, "('#')")
    end if
    if (P%print_level .gt. 1 .and. len(trim(P%gto_file_path)) .gt. 0) then
      call print_GTO_info(B%bas, B%atm, B%env, P%OUTPUT)

    end if
    if (P%print_level .gt. 1) then
      write (OUTPUT, "('# Smallest overlap matrix eigenvalue = ',es13.5)") &
        B%minimal_S_eigenvalue
      write (OUTPUT, "('# Smallest remaining overlap matrix eigenvalue = ',es13.5)") &
        B%minimal_S_remaining_eigenvalue
      write (OUTPUT, '("# Number of basis functions after overlap projection",i9)') B%Nr_basis
    end if
    if (P%yes_sym .and. P%print_level .gt. 1) then
      write (OUTPUT, '("# Symmetric basis functions")')
      write (OUTPUT, '("# ",2x,a6, 2x,a6,x,a16)') 'IRREP', '#bf', '#bf after projection'
      do i = 1, B%nirrep
        write (OUTPUT, '("# ",2x,a6, 2x,i6,x,i6)') &
          B%irrep_name(i), count(B%irrep_sym_bf(:) .eq. i), count(B%irrep_sym_bfr(:) .eq. i)
      end do
    end if
  end subroutine print_calc_info

  subroutine calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    !! calculates the transition dipole matrix elements between basisfunctions
    !! and stores them in tdipole_bf( mu, nu, p)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(Grid3D), intent(in) :: G

    real(long), dimension(P%N_basis, P%N_basis, 3), intent(out) :: tdipole_bf
    ! transition dipole matrix element (cartesian: x,y,z) return value
    integer :: mu, nu

    if (P%yes_gaussian) then
      call calculate_dipole_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, tdipole_bf)
    else
      tdipole_bf(:, :, :) = ZERO
      do mu = 1, B%N_basis
        ! compute < mu | x/y/z | mu >
        tdipole_bf(mu, mu, 1) = braket(G, B, mu, mu, G%x(:))
        tdipole_bf(mu, mu, 2) = braket(G, B, mu, mu, G%y(:))
        tdipole_bf(mu, mu, 3) = braket(G, B, mu, mu, G%z(:))
        do nu = mu + 1, B%N_basis
          ! for nu .ne. mu, compute < mu | x/y/z | nu >
          tdipole_bf(mu, nu, 1) = braket(G, B, mu, nu, G%x(:))
          tdipole_bf(mu, nu, 2) = braket(G, B, mu, nu, G%y(:))
          tdipole_bf(mu, nu, 3) = braket(G, B, mu, nu, G%z(:))
          tdipole_bf(nu, mu, :) = tdipole_bf(mu, nu, :)
        end do
      end do
    end if

  end subroutine calc_trans_dipole_moment_bf

  subroutine calc_quadrupole_moment_bf(P, B, G, quadrupole_bf)
    !! calculates the transition dipole matrix elements between basisfunctions
    !! and stores them in tdipole_bf( mu, nu, p)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(Grid3D), intent(in) :: G

    real(long), dimension(P%N_basis, P%N_basis, 3, 3), intent(out) :: quadrupole_bf
    ! quadrupole matrix element (cartesian: x,y,z, x,y,z) return value
    integer :: mu, nu
    if (P%yes_gaussian) then
      call calculate_quadrupole_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, quadrupole_bf)
    else
      quadrupole_bf(:, :, :, :) = ZERO
      do mu = 1, B%N_basis
        ! compute < mu | x/y/z | mu >
        quadrupole_bf(mu, mu, 1, 1) = braket(G, B, mu, mu, G%x(:) * G%x(:) )
        quadrupole_bf(mu, mu, 1, 2) = braket(G, B, mu, mu, G%x(:) * G%y(:) )
        quadrupole_bf(mu, mu, 1, 3) = braket(G, B, mu, mu, G%x(:) * G%z(:) )
        quadrupole_bf(mu, mu, 2, 1) = quadrupole_bf(mu, mu, 1, 2)
        quadrupole_bf(mu, mu, 2, 2) = braket(G, B, mu, mu, G%y(:) * G%y(:) )
        quadrupole_bf(mu, mu, 2, 3) = braket(G, B, mu, mu, G%y(:) * G%z(:) )
        quadrupole_bf(mu, mu, 3, 1) = quadrupole_bf(mu, mu, 1, 3)
        quadrupole_bf(mu, mu, 3, 2) = quadrupole_bf(mu, mu, 2, 3)
        quadrupole_bf(mu, mu, 3, 3) = braket(G, B, mu, mu, G%z(:) * G%z(:) )
        do nu = mu + 1, B%N_basis
          ! for nu .ne. mu, compute < mu | x/y/z | nu >
          quadrupole_bf(mu, nu, 1, 1) = braket(G, B, mu, nu, G%x(:) * G%x(:) )
          quadrupole_bf(mu, nu, 1, 2) = braket(G, B, mu, nu, G%x(:) * G%y(:) )
          quadrupole_bf(mu, nu, 1, 3) = braket(G, B, mu, nu, G%x(:) * G%z(:) )
          quadrupole_bf(mu, nu, 2, 1) = quadrupole_bf(mu, nu, 1, 2)
          quadrupole_bf(mu, nu, 2, 2) = braket(G, B, mu, nu, G%y(:) * G%y(:) )
          quadrupole_bf(mu, nu, 2, 3) = braket(G, B, mu, nu, G%y(:) * G%z(:) )
          quadrupole_bf(mu, nu, 3, 1) = quadrupole_bf(mu, nu, 1, 3)
          quadrupole_bf(mu, nu, 3, 2) = quadrupole_bf(mu, nu, 2, 3)
          quadrupole_bf(mu, nu, 3, 3) = braket(G, B, mu, nu, G%z(:) * G%z(:) )
          quadrupole_bf(nu, mu, :,: ) = quadrupole_bf(mu, nu, :, :)
        end do! nu
      end do! mu
    end if

  end subroutine calc_quadrupole_moment_bf

  subroutine orthonormalizeOrbitals(B, C, irrep_MO)
    !! Do a Gram-Schmidt orthogonalization for the coefficients given by C
    !! according to the metric given by the overlap matrix B%S,
    !! i.e. such that < C(:,i) S(:,:) C(:,j) > = delta_i,j
    !! B are the basis set parameters
    type(Basis), intent(in) :: B
    real(long), dimension(:, :), intent(inout) :: C ! coefficients
    integer, intent(in), optional :: irrep_MO(:)
    real(long) :: olap
    integer i, j, N, k, ir
    real(long), parameter :: threshold = 1e-8

    ! project into reduced basis
    C = B%XXtS.m.C

    N = size(C, 2)
    if (present(irrep_MO)) then
      do ir = 1, B%nirrep
        k = 1
        do i = 1, N
          if (irrep_MO(i) .ne. ir) cycle
          do
            do j = 1, i - 1
              if (irrep_MO(j) .ne. ir) cycle
              olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, i)))
              if (olap .gt. threshold) then
                C(:, i) = C(:, i)/sqrt(olap)
              else
                exit
              end if
              olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, j)))
              C(:, i) = C(:, i) - olap*C(:, j)
            end do
            olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, i)))
            if (olap .gt. threshold) then
              C(:, i) = C(:, i)/sqrt(olap)
              exit
            else
              ! go to the next k of matching irrep
              do
                if (B%irrep_sym_bfr(k) .eq. ir) exit
                k = k + 1
                if (k .gt. size(B%X, 2)) exit
              end do
              if (k .gt. size(B%X, 2)) stop "# Orthonormalization failed"
              if (B%irrep_sym_bfr(k) .eq. ir) then
                write (6, '("# WARNING: orbital ",i4," linear dependent." // &
                  & "Inserting from default orthogonalization contribution",i4)') i, k
                C(:, i) = B%X(:, k)
              end if
              k = k + 1
              cycle
            end if
          end do!
        end do! i
      end do! ir
    else
      k = 1
      do i = 1, N
        do
          do j = 1, i - 1
            olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, i)))
            if (olap .gt. threshold) then
              C(:, i) = C(:, i)/sqrt(olap)
            else
              exit
            end if
            olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, j)))
            C(:, i) = C(:, i) - olap*C(:, j)
          end do
          ! renormalize
          olap = dot_product(C(:, i), MATMUL(B%S(:, :), C(:, i)))
          if (olap .gt. threshold) then
            C(:, i) = C(:, i)/sqrt(olap)
            exit
          else
            ! if olap is too small, take a vector from B%X and orthogonalize it against the other ones
            ! this is a _very_stupid procedure and something more efficient should be done (?)
            if (k .gt. size(B%X, 2)) stop "# Orthonormalization failed"
            write (6, '("# WARNING: orbital ",i4," linear dependent."// &
            &              " Inserting from default orthogonalization contribution",i4)') i, k
            C(:, i) = B%X(:, k)
            k = k + 1
            cycle
          end if
        end do
      end do ! i
    end if
  end subroutine orthonormalizeOrbitals

end module basis_func
