!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module cis
  !! author: Ludger Inhester
  !! CIS module to compute CIS excited states
  !!
  use constants
  use basis_func
  use read_param
  use wavefunction
  use atom_in_mol
  use packed_index
  use hartreefock
  use mcscf
  use molden
  use population
  implicit none
  private
  public :: cis_procedure, cis_densitymatrix
contains

  subroutine cis_procedure(P, A, G, B, F, OUTPUT)
    !! calculates cis states
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in), optional:: OUTPUT

    integer :: n_occ, n_unocc, n
    real(long), allocatable :: H_cis(:, :)
    real(long), allocatable :: E_cis(:)
    integer :: i, j ! general index for orbitals
    integer :: l, m ! index for excitations
    integer :: r, s ! index for virtual orbitals
    integer :: t, u ! index for occupied orbitals
    integer(8) :: rs, ut, rt, us, rsut, rtus
    integer :: IOUT
    integer :: nstates
    integer, allocatable :: core(:)
    integer, allocatable :: virtual(:)
    integer, allocatable :: index2core(:)
    integer, allocatable :: index2virtual(:)
    integer(8) :: nij
    integer :: nc, ni
    integer(8) :: cd, ij, ci, dj, cdij, cidj
    real :: timer
    real :: time_last_step
    integer*8 :: int_time      ! timer
    type(WF_data) :: F_temp
    real(long), allocatable :: tdens(:,:)
    real(long), allocatable :: LC_hole(:), LC_part(:)
    character(len=512) :: moldenfilename ! filename
    type(CSF), allocatable :: csf_array(:)
    integer, allocatable :: occs(:, :)
    real(long), allocatable :: csf_O(:, :)
    if (P%yes_openshell .or. .not. P%yes_HF) then
      stop 'CIS requires closed-shell HF'
    end if

    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT

    if (allocated(P%cis_hole)) then
      ! if cis_hole is specified do CIS within sub space where hole is within of P%cis_hole

      n_occ = count(F%N_occ(P%cis_hole(:)) .eq. 2)
      n_unocc = count(F%N_occ(F%Ncore + 1:) .eq. 0)
      if (P%yes_sym .and. P%state_irrep .gt. 0) then
        n = 0
        do i = 1, size(P%cis_hole, 1)
          if (F%N_occ(P%cis_hole(i)) .ne. 2) cycle
          do j = F%Ncore + 1, size(F%CE, 1)
            if (F%N_occ(j) .ne. 0) cycle
            if (B%irrep_prod(F%irrep_MO(P%cis_hole(i)), F%irrep_MO(j)) .eq. P%state_irrep) n = n + 1
          end do
        end do
      else
        n = n_occ*n_unocc
      end if
      if (P%nstates .eq. -1) then
        nstates = n
      else
        nstates = P%nstates - 1
      end if

      if (.not. allocated(F%CIS_excitation)) &
        allocate (F%CIS_excitation(2, n))
      if (.not. allocated(F%E_cis)) &
        allocate (F%E_cis(nstates))
      if (.not. allocated(F%C_cis)) &
        allocate (F%C_cis(n, nstates))
      allocate (H_CIS(n, n))
      allocate (E_CIS(n))
      m = 1
      do i = 1, size(P%cis_hole)
        if (F%N_occ(P%cis_hole(i)) .ne. 2) cycle
        do j = F%Ncore + 1, size(F%CE, 1)
          if (F%N_occ(j) .ne. 0) cycle

          if (P%yes_sym .and. P%state_irrep .gt. 0) then
            if (B%irrep_prod(F%irrep_MO(P%cis_hole(i)), F%irrep_MO(j)) .ne. P%state_irrep) cycle
          end if

          F%CIS_excitation(1, m) = P%cis_hole(i)
          F%CIS_excitation(2, m) = j
          m = m + 1
        end do
      end do
    else! if(allocated(P%cis_hole))
      n_occ = count(F%N_occ(F%Ncore + 1:) .eq. 2)
      n_unocc = count(F%N_occ(F%Ncore + 1:) .eq. 0)
      if (P%yes_sym .and. P%state_irrep .gt. 0) then
        n = 0
        do i = F%Ncore + 1, size(F%CE)
          if (F%N_occ(i) .ne. 2) cycle
          do j = F%Ncore + 1, size(F%CE, 1)
            if (F%N_occ(j) .ne. 0) cycle
            if (B%irrep_prod(F%irrep_MO(i), F%irrep_MO(j)) .eq. P%state_irrep) n = n + 1
          end do
        end do
      else
        n = n_occ*n_unocc
      end if
      if (P%nstates .eq. -1) then
        nstates = n
      else
        nstates = P%nstates - 1
      end if
      if (allocated(F%CIS_excitation)) deallocate (F%CIS_excitation)
      allocate (F%CIS_excitation(2, n))
      if (allocated(F%E_cis)) deallocate (F%E_cis)
      allocate (F%E_cis(nstates))
      if (allocated(F%C_cis)) deallocate (F%C_cis)
      allocate (F%C_cis(n, nstates))
      allocate (H_CIS(n, n))
      allocate (E_CIS(n))
      m = 1
      do i = F%Ncore + 1, size(F%CE)
        if (F%N_occ(i) .ne. 2) cycle
        do j = F%Ncore + 1, size(F%CE, 1)
          if (F%N_occ(j) .ne. 0) cycle
          if (P%yes_sym .and. P%state_irrep .gt. 0) then
            if (B%irrep_prod(F%irrep_MO(i), F%irrep_MO(j)) .ne. P%state_irrep) cycle
          end if
          F%CIS_excitation(1, m) = i
          F%CIS_excitation(2, m) = j
          m = m + 1
        end do
      end do
    end if! if(allocated(P%cis_hole)

    allocate (core(count((/(any(F%CIS_excitation(1, :) .eq. j), j=1, size(F%N_occ, 1))/))))
    allocate (virtual(count((/(any(F%CIS_excitation(2, :) .eq. j), j=1, size(F%N_occ, 1))/))))
    core = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/(any(F%CIS_excitation(1, :) .eq. j), j=1, size(F%N_occ, 1))/))
    virtual = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/(any(F%CIS_excitation(2, :) .eq. j), j=1, size(F%N_occ, 1))/))

    allocate (index2core(size(F%N_occ, 1)))
    allocate (index2virtual(size(F%N_occ, 1)))
    index2core = (/(minloc(abs(core - j)), j=1, size(F%N_occ, 1))/)
    index2virtual = (/(minloc(abs(virtual - j)), j=1, size(F%N_occ, 1))/)
    call transform_twoel_two_core_two_virtual(P, B, F, core, virtual, P%OUTPUT)

    n_occ = count(F%N_occ(F%Ncore + 1:) .eq. 2)
    n_unocc = count(F%N_occ(F%Ncore + 1:) .eq. 0)

    ni = size(virtual, 1)
    nc = size(core, 1)
    nij = ni*(ni + 1)/2
    H_cis = ZERO
    If (P%yes_verbose .and. P%yes_sym) &
      write (IOUT, '("# CIS for symmetry irrep ",i4)') P%state_irrep
    do l = 1, n
      t = F%CIS_excitation(1, l) ! occupied
      r = F%CIS_excitation(2, l) ! unoccupied
      rt = index_packed(r, t)
      rs = index_packed(r, r)
      ut = index_packed(t, t)
      rtus = index_packed(rt, rt)
      rsut = index_packed(rs, ut)
      ij = index_packed(index2virtual(r), index2virtual(r))
      cd = index_packed(index2core(t), index2core(t))
      dj = (index2core(t) - 1)*ni + index2virtual(r)
      ci = (index2core(t) - 1)*ni + index2virtual(r)
      cdij = (cd - 1)*nij + ij
      cidj = index_packed(dj, ci)

      if (P%multp .eq. 1) then
        H_CIS(l, l) = F%CE(r) - F%CE(t) + TWO*F%teint_cidj(cidj) - F%teint_cdij(cdij)
      else if (P%multp .eq. 3) then
        H_CIS(l, l) = F%CE(r) - F%CE(t) - F%teint_cdij(cdij)
      end if

      do m = l + 1, n
        u = F%CIS_excitation(1, m) ! occupied
        s = F%CIS_excitation(2, m) ! unoccupied

        us = index_packed(u, s)
        rtus = index_packed(rt, us)
        rs = index_packed(r, s)
        ut = index_packed(u, t)
        rsut = index_packed(rs, ut)

        ij = index_packed(index2virtual(r), index2virtual(s))
        cd = index_packed(index2core(t), index2core(u))
        dj = (index2core(t) - 1)*ni + index2virtual(r)
        ci = (index2core(u) - 1)*ni + index2virtual(s)
        cdij = (cd - 1)*nij + ij
        cidj = index_packed(dj, ci)
        if (P%multp .eq. 1) then
          H_CIS(l, m) = TWO*F%teint_cidj(cidj) - F%teint_cdij(cdij)
        else if (P%multp .eq. 3) then
          H_CIS(l, m) = -F%teint_cdij(cdij)
        end if
        H_CIS(m, l) = H_CIS(l, m)
      end do
    end do
    !call get_eigenvalues(H_CIS, E_cis)
    !F%E_cis(:) = E_cis(1:nstates)
    !F%C_cis(:, 1:nstates) = H_CIS(:, 1:nstates) ! save coefficients for later usage
    call get_some_ev_of_real_sym(H_CIS, F%E_CIS, F%C_CIS, nstates, P%SCF_energy_conv, yes_verbose=.False.)
    ! calculate the density matrix for state F%state
    if (allocated(F%D_CI)) deallocate (F%D_CI)
    allocate (F%D_CI(size(F%D, 1), size(F%D, 2)))
    if (F%state .gt. 0) then
      if (P%print_level > 1) then
        write (IOUT, '("# building density matrix for excited state ",i8)') F%state
      end if
      F%D_CI = cis_densitymatrix(B, F, F%state, f_occ=F%F_occ, type='F')
    else
      F%D_CI = F%D
    end if

    do l = 1, nstates
      if (P%print_level > 1) then
        write (IOUT, '("# excited state ",i8)') l
        write (IOUT, '("# E-E_0=",f15.8, "eV ")') F%E_CIS(l)*AU2EV
        write (IOUT, '("# E= ",f15.8, "H ")') F%E_tot + F%E_CIS(l)
      end if
      do m = 1, n
        if (abs(F%C_CIS(m, l)) .gt. 0.05 .and. P%print_level > 1) &
          write (IOUT, '("#  ",i4,"->",i4, " ",f10.4)') &
          F%CIS_excitation(1, m), &
          F%CIS_excitation(2, m), &
          F%C_cis(m, l)
      end do
      if (allocated(P%nat_trans_orb)) then
        if (any(P%nat_trans_orb(:) .eq. l)) then
          allocate(tdens(size(F%LC,2), size(F%LC,2)))
          tdens = transpose(cis_densitymatrix(B, F, l, type='T', basisD='M'))
          allocate (LC_hole(size(tdens, 1)))
          allocate (LC_part(size(tdens, 2)))
          tdens = tdens / sqrt(TWO)
          call make_nat_trans_orbitals(P%OUTPUT, tdens, LC_hole, LC_part)
          if (P%molden_file .ne. "") then
            F_temp = F
            F_temp%LC = ZERO
            F_temp%LC(:, 1) = F%LC(:, :) .m. LC_hole(:)
            F_temp%LC(:, 2) = F%LC(:, :) .m. LC_part(:)
            if(P%yes_sym)then
              write (moldenfilename, '("trans_cis_",a,"_",i0.4,".molden")') trim(B%irrep_name(P%state_irrep)), l
            else
              write (moldenfilename, '("trans_cis_",i0.4,".molden")') l
            end if
            call write_molden_file(P, A, B, F_temp, moldenfilename)
          end if
          deallocate(tdens)
          deallocate(LC_hole)
          deallocate(LC_part)

          allocate(tdens(size(F%LC,1), size(F%LC,1)))
          tdens = cis_densitymatrix(B, F, l, type='F', basisD='A')
          call density_analysis(P, G, B, tdens, P%OUTPUT, description='     Electron density:')
          tdens = -cis_densitymatrix(B, F, l, type='H', basisD='A')
          call density_analysis(P, G, B, tdens, P%OUTPUT, description='     Hole density:')
          tdens = cis_densitymatrix(B, F, l, type='P', basisD='A')
          call density_analysis(P, G, B, tdens, P%OUTPUT, description='     Particle density:')
          deallocate(tdens)
        end if
      end if

    end do
    if (P%print_level > 1) write (IOUT, "('# |CIS| ',2000f10.2)") F%E_CIS(1:nstates)*AU2EV

    if (F%yes_calculate_CI_overlap .and. allocated(F%C_cis_old)) then
      call check_time(int_time, timer)
      time_last_step = timer
      allocate (occs(size(F%LC, 2), 1:size(H_CIS, 1) + 1))
      occs(:, 1) = F%N_occ
      do i = 2, size(H_CIS, 1) + 1
        occs(:, i) = F%N_occ
        occs(F%CIS_excitation(1, i - 1), i) = occs(F%CIS_excitation(1, i - 1), i) - 1
        occs(F%CIS_excitation(2, i - 1), i) = occs(F%CIS_excitation(2, i - 1), i) + 1
      end do

      ! with symmetry switched on this seems to not work!
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CIS ","time for occs",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
      call construct_csf_array(P, B, F, P%multp, P%state_irrep, occs, csf_array)
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CIS ","time for csfs",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
      allocate (csf_O(size(csf_array, 1), size(csf_array, 1)))
      csf_O = ZERO
      call overlap_csfs(P, csf_array, F%LC, F%LC_old, B%S, csf_O)
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CIS ","time for overlap csf",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
      if (allocated(F%O_ci)) deallocate (F%O_ci)
      if (all(csf_array(1)%occ .eq. F%N_occ)) then
        allocate (F%O_CI(nstates + 1, nstates + 1))
        F%O_CI(1, 1) = csf_O(1, 1)
        F%O_CI(1, 2:) = matmul(csf_O(1, 2:), F%C_cis_old(:, :))
        F%O_CI(2:, 1) = matmul(transpose(F%C_cis(:, :)), csf_O(2:, 1))
        F%O_CI(2:, 2:) = F%C_cis(:, :) .tm. csf_O(2:, 2:) .m. F%C_cis_old(:, :)
      else
        ! if symmetry has been used, the closed ground state may not be part of the csf_O matrix. In this case, we assume 1 for the overlap of the ground state
        allocate (F%O_CI(nstates + 1, nstates + 1))
        F%O_CI(1, 1) = 1.
        F%O_CI(1, 2:) = 0.
        F%O_CI(2:, 1) = 0.
        F%O_CI(2:, 2:) = F%C_cis(:, :) .tm.csf_O(1:, 1:) .m.F%C_cis_old(:, :)
      end if
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CIS ","time for overlap cis",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
      deallocate (csf_O)
      call purge_csf_array(csf_array)
      deallocate (occs)
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CIS ","time for overlap",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
    end if

    deallocate (H_CIS)
    deallocate (core)
    deallocate (virtual)
    deallocate (index2core)
    deallocate (index2virtual)
  end subroutine cis_procedure


  function cis_densitymatrix(B, F, state, f_occ, type, basisD) result(D)
    !! calculates different types of cis density matrix for stateindex state
    !! state: state index 1...n
    !! typ (optional argument):
    !! 'F' for full density matrix
    !! 'P' for particle density matrix
    !! 'H' for hole density matrix
    !! 'T' for transition-density matrix
    !! basisD (optional argument):
    !! 'M' for MO basis
    !! 'A' for AO basis (default)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: state
    character, intent(in), optional :: type
    character, intent(in), optional :: basisD
    real(long), intent(out), optional :: f_occ(size(F%LC, 2))
    character :: type_p, basis_p
    real(long), allocatable :: D(:, :)
    real(long) :: Dmo(size(F%LC, 2), size(F%LC, 2))
    integer :: i, j
    type_p = 'F'
    if (present(type)) type_p = type
    basis_p = 'A'
    if (present(basisD)) basis_p = basisD
    if (.not. allocated(F%CIS_excitation)) &
      stop 'Warning CIS has not been calculated yet ?'
    if (state .eq. 0) stop 'cis_density_matrix: state index must be > 0'
    DMO = ZERO
    if (type_p .eq. 'T') then
      do i = 1, size(F%CIS_excitation, 2)
        DMO(F%CIS_excitation(2, i), F%CIS_excitation(1, i)) = sqrt(TWO) * F%C_cis(i, state)
      end do
    else
      do i = 1, size(F%N_occ)
        if (F%N_occ(i) .eq. 0) cycle
        if (type_p .eq. 'F') DMO(i, i) = dble(F%N_occ(i))
      end do
      do i = 1, size(F%CIS_excitation, 2)
        do j = 1, size(F%CIS_excitation, 2)
          if (F%CIS_excitation(2, i) .eq. F%CIS_excitation(2, j) .and. (type_p .eq. 'F' .or. type_p .eq. 'H')) then
            DMO(F%CIS_excitation(1, i), F%CIS_excitation(1, j)) = DMO(F%CIS_excitation(1, i), F%CIS_excitation(1, j)) &
              - F%C_cis(i, state)*F%C_cis(j, state)
          end if
          if (F%CIS_excitation(1, i) .eq. F%CIS_excitation(1, j) .and. (type_p .eq. 'F' .or. type_p .eq. 'P')) then
            DMO(F%CIS_excitation(2, i), F%CIS_excitation(2, j)) = DMO(F%CIS_excitation(2, i), F%CIS_excitation(2, j)) &
              + F%C_cis(i, state)*F%C_cis(j, state)
          end if
        end do
      end do
      if (present(f_occ)) then
        do i = 1, size(Dmo, 1)
          f_occ(i) = Dmo(i, i)
        end do
      end if
    end if
    if (basis_p=='M') then
      allocate(D(size(DMO,1),size(DMO,2)))
      D = DMO
    else if (basis_p=='A') then
      allocate(D(size(F%LC, 1),size(F%LC, 1)))
      D = F%LC.m.Dmo.mt.F%LC
    else
      stop 'basis must be either "A" or "M"'
    end if
  end function cis_densitymatrix

end module cis
