!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!


module cubefile
  !! author: Ludger Inhester
  !!
  !! This module provides output in the Gaussian cube file format  
  use constants
  use interpolate
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use wavefunction
  use sph_harmonics

  implicit none
  private
  public :: write_cube_file
contains


  subroutine write_cube_file(P, A, B, F, filename, mo, D)
    !! write_density_cube_file(P,A,B,G,F, filename)
    !!
    !! write a cubefile into file filename
    !! of quantity rho on the molecular grid
    !!    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    character(len=*), intent(in) :: filename !! output filename for the cube file
    integer, intent(in), optional :: mo !! if given an mo cube file is written
    real(long), intent(in), optional :: D(size(F%LC, 1), size(F%LC, 1))
    !! if given, this density matrix is used, otherwise F%D
    real(long) :: DM(size(F%LC, 1), size(F%LC, 1))
    integer :: ix, iy, iz
    integer :: bf1, bf2, atom, i, j
    real(long) :: origin(3)
    integer :: nv(3)
    real(long) :: step(3)
    real(long) :: min_coord(3)
    real(long) :: max_coord(3)
    real(long) :: lx, ly, lz
    real(long) :: res
    integer :: n1, n2, sh1, sh2
    real(long), allocatable::  gridpoint_x(:)
    real(long), allocatable::  gridpoint_y(:)
    real(long), allocatable::  gridpoint_z(:)
    real(long), allocatable::  values(:, :, :)
    real(long), allocatable::  bf1_values(:, :, :)
    real(long), allocatable::  bf2_values(:, :, :)
    real(long), allocatable::  sh1_values(:, :, :, :)
    real(long), allocatable::  sh2_values(:, :, :, :)
    DM(:, :) = F%D(:, :)
    if (present(D)) &
      DM(:, :) = D(:, :)
    ! The parameters of the cube grid:
    ! the edges of the grid cube are at least 3.0 bohr length
    ! away from all nuclei as specified in P structure
    min_coord = HUGE(min_coord)
    max_coord = -HUGE(max_coord)
    ! do atom=1,size(A)
    !    if(A(atom)%x<min_coord(1)) min_coord(1)=A(atom)%x
    !    if(A(atom)%y<min_coord(2)) min_coord(2)=A(atom)%y
    !    if(A(atom)%z<min_coord(3)) min_coord(3)=A(atom)%z
    !    if(A(atom)%x>max_coord(1)) max_coord(1)=A(atom)%x
    !    if(A(atom)%y>max_coord(2)) max_coord(2)=A(atom)%y
    !    if(A(atom)%z>max_coord(3)) max_coord(3)=A(atom)%z
    ! end do
    do atom = 1, size(A)
      if (P%charge_nuc(atom) .eq. 0) cycle
      if (P%x_nuc(atom) .lt. min_coord(1)) min_coord(1) = P%x_nuc(atom)
      if (P%y_nuc(atom) .lt. min_coord(2)) min_coord(2) = P%y_nuc(atom)
      if (P%z_nuc(atom) .lt. min_coord(3)) min_coord(3) = P%z_nuc(atom)
      if (P%x_nuc(atom) .gt. max_coord(1)) max_coord(1) = P%x_nuc(atom)
      if (P%y_nuc(atom) .gt. max_coord(2)) max_coord(2) = P%y_nuc(atom)
      if (P%z_nuc(atom) .gt. max_coord(3)) max_coord(3) = P%z_nuc(atom)
    end do
    ! we distribute P%Nvoxel voxels
    ! (max_coord(1)-min_coord(1))*(max_coord(2)-min_coord(2))*(max_coord(3)-min_coord(3))
    lx = max_coord(1) - min_coord(1) + 10_long
    ly = max_coord(2) - min_coord(2) + 10_long
    lz = max_coord(3) - min_coord(3) + 10_long
!      write(P%OUTPUT,'("# Volume: ",es)') dble(lx*ly*lz)
!      write(P%OUTPUT,'("# Voxel per Volume: ",es)') dble(P%Nvoxel)/dble(lx*ly*lz)
!      write(P%OUTPUT,'("# Voxel per Volume^1/3: ",es)') (dble(P%Nvoxel)/dble(lx*ly*lz))**dble(1./3)
    res = (dble(P%Nvoxel)/dble(lx*ly*lz))**(dble(1./3))
    ! number of voxel per dimension
    nv(1) = int(res*lx)
    nv(2) = int(res*ly)
    nv(3) = int(res*lz)
!      nv(1)=100
!      nv(2)=100
!      nv(3)=100
    write (P%OUTPUT, '("# Total number of voxels: ",i8)') nv(1)*nv(2)*nv(3)
    write (P%OUTPUT, '("# points along x directions",i4," box length=",f10.5," (",f10.5," points per a.u.)")') nv(1), lx, nv(1)/lx
    write (P%OUTPUT, '("# points along y directions",i4," box length=",f10.5," (",f10.5," points per a.u.)")') nv(2), ly, nv(2)/ly
    write (P%OUTPUT, '("# points along z directions",i4," box length=",f10.5," (",f10.5," points per a.u.)")') nv(3), lz, nv(3)/lz
    ! origin
    origin(1) = min_coord(1) - 5.0_long
    origin(2) = min_coord(2) - 5.0_long
    origin(3) = min_coord(3) - 5.0_long
    ! step size
    step(1) = (max_coord(1) + 5.0_long - origin(1))/nv(1)
    step(2) = (max_coord(2) + 5.0_long - origin(2))/nv(2)
    step(3) = (max_coord(3) + 5.0_long - origin(3))/nv(3)
    ! write cube file header
    open (unit=66, file=filename, status='unknown', ACTION='write')
    write (66, '("# cube file generated by xmolecule")')
    if (present(mo)) then
      write (66, '(" molecular orbital ",i4)') mo
    else
      write (66, '(" total electron density ")')
    end if
    write (66, '(i4,3f12.6)') P%N_nuc, origin(:)
    write (66, '(i4,3f12.6)') nv(1), step(1), ZERO, ZERO
    write (66, '(i4,3f12.6)') nv(2), ZERO, step(2), ZERO
    write (66, '(i4,3f12.6)') nv(3), ZERO, ZERO, step(3)
    do j = 1, P%N_nuc
      write (66, '(i4,4f12.6)') P%charge_nuc(j), ZERO, A(j)%x, A(j)%y, A(j)%z
    end do
    allocate (gridpoint_x(nv(1)))
    allocate (gridpoint_y(nv(2)))
    allocate (gridpoint_z(nv(3)))
    allocate (values(nv(3), nv(2), nv(1)))

    do ix = 1, nv(1)
      gridpoint_x(ix) = origin(1) + (ix - 1)*step(1)
    end Do
    do iy = 1, nv(2)
      gridpoint_y(iy) = origin(2) + (iy - 1)*step(2)
    end Do
    do iz = 1, nv(3)
      gridpoint_z(iz) = origin(3) + (iz - 1)*step(3)
    end Do
    values(:, :, :) = ZERO
    if (P%yes_gaussian) then
      if (present(mo)) then
        do sh1 = 1, size(B%bas, 2)
          n1 = 2 * B%bas(2, sh1) + 1
          allocate (sh1_values(n1, nv(3), nv(2), nv(1)))
          sh1_values = ZERO
          call calculate_bf_values_shell(P, A, B, sh1_values, gridpoint_x, gridpoint_y, gridpoint_z, sh1)
          do i = 1, n1
            values(:, :, :) = values(:, :, :) + F%LC(B%first_bf_of_shell(sh1) + i - 1, mo) * sh1_values(i, :, :, :)
          end do
          deallocate (sh1_values)
        end do! sh1
      else
        do sh1 = 1, size(B%bas, 2)
          n1 = 2 * B%bas(2, sh1) + 1
          allocate (sh1_values(n1, nv(3), nv(2), nv(1)))
          sh1_values = ZERO
          call calculate_bf_values_shell(P, A, B, sh1_values, gridpoint_x, gridpoint_y, gridpoint_z, sh1)
          do i = 1, n1
            do j = 1, n1
            values(:, :, :) = values(:, :, :) &
              + DM(B%first_bf_of_shell(sh1) + i - 1, B%first_bf_of_shell(sh1) + j - 1) &
              * sh1_values(i, :, :, :) * sh1_values(j, :, :, :)
            end do! j
          end do!i
          do sh2 = sh1 + 1, size(B%bas, 2)
            n2 = 2 * B%bas(2, sh2) + 1
            if( all(abs(DM(B%first_bf_of_shell(sh1): B%first_bf_of_shell(sh1) + n1 - 1, B%first_bf_of_shell(sh2):B%first_bf_of_shell(sh2) + n2 - 1)) < P%SCF_density_conv)) &
              cycle
            allocate (sh2_values(n2, nv(3), nv(2), nv(1)))
            sh2_values = ZERO
            call calculate_bf_values_shell(P, A, B, sh2_values, gridpoint_x, gridpoint_y, gridpoint_z, sh2)
            do i = 1, n1
              do j = 1, n2
              values(:, :, :) = values(:, :, :) &
                + TWO * DM(B%first_bf_of_shell(sh1) + i - 1, B%first_bf_of_shell(sh2) + j - 1) &
                * sh1_values(i, :, :, :) * sh2_values(j, :, :, :)
              end do! j
            end do!i
            deallocate(sh2_values)
          end do! sh2
          deallocate(sh1_values)
        end do! sh1
      end if! present(mo)
    else
      allocate (bf1_values(nv(3), nv(2), nv(1)))
      allocate (bf2_values(nv(3), nv(2), nv(1)))
      if (present(mo)) then
        do bf1 = 1, B%N_basis
          bf1_values = ZERO
          call calculate_bf_values(P, A, B, bf1_values, gridpoint_x, gridpoint_y, gridpoint_z, bf1)
          values(:, :, :) = values(:, :, :) + F%LC(bf1, mo)*bf1_values(:, :, :)
        end do
      else
        do bf1 = 1, B%N_basis
          bf1_values = ZERO
          call calculate_bf_values(P, A, B, bf1_values, gridpoint_x, gridpoint_y, gridpoint_z, bf1)
          do bf2 = bf1 + 1, B%N_basis
            bf2_values = ZERO
            call calculate_bf_values(P, A, B, bf2_values, gridpoint_x, gridpoint_y, gridpoint_z, bf2)
            values(:, :, :) = values(:, :, :) + TWO*DM(bf1, bf2)*bf1_values(:, :, :)*bf2_values(:, :, :)
          end do
          values(:, :, :) = values(:, :, :) + DM(bf1, bf1)*bf1_values(:, :, :)*bf1_values(:, :, :)
        end do
      end if
      deallocate (bf1_values)
      deallocate (bf2_values)
    end if
    ! write cube file data
    do ix = 1, nv(1)
      do iy = 1, nv(2)
        do iz = 1, nv(3)
          write (66, '(E14.5)', advance="no") values(iz, iy, ix)
          if (mod(iz, 6) .eq. 0) write (66, '()')
        end do
        write (66, '()')
      end do
    end do
    close (66)
    deallocate (gridpoint_x)
    deallocate (gridpoint_y)
    deallocate (gridpoint_z)
    deallocate (values)

  end subroutine write_cube_file

  subroutine calculate_bf_values(P, A, B, values, gridpoint_x, gridpoint_y, gridpoint_z, mu)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B

    real(long), intent(in) ::  gridpoint_x(:)
    real(long), intent(in) ::  gridpoint_y(:)
    real(long), intent(in) ::  gridpoint_z(:)
    real(long), intent(out) ::  values(:, :, :)
    integer, intent(in) :: mu
    integer :: atom, n, l, m, N_r, k

    integer :: ix, iy, iz
    real(long) :: x, y, z, theta, phi, r
    real(long) :: radial_f
    real(long), allocatable :: r_int(:), f_int(:)  ! for interpolation with zeros
    type(Spline) :: SP

    integer :: i, sh
    integer :: nprim
    real(long) :: alpha
    real(long) :: coef
    real(long) :: sphh
    real(long) :: gto

    atom = B%info(mu, 1)
    n = B%info(mu, 3)
    l = B%info(mu, 4)
    m = B%info(mu, 5)
    do k = 1, A(atom)%N_nao
      if (A(atom)%n_AO(k) .eq. n .and. A(atom)%l_AO(k) .eq. l) exit
    end do

    if (.not. P%yes_gaussian) then
      N_r = A(atom)%N_r
      allocate (r_int(N_r + 2), f_int(N_r + 2))
      r_int(1) = ZERO
      r_int(2:N_r + 1) = A(atom)%radial%r(:)
      r_int(N_r + 2) = A(atom)%rmax
      if (A(atom)%rmax .eq. ZERO) r_int(N_r + 2) = A(atom)%radial%r(N_r)*FOUR     ! put an enough large r
      f_int(1) = ZERO
      f_int(N_r + 2) = ZERO
      f_int(2:N_r + 1) = A(atom)%u_AO(:, k)                ! prepare interpolation of u(r)
      call spline_create(r_int(:), f_int(:), SP)
    else
      do sh = 1, size(B%bas, 2) ! loop over shells
        if (B%first_bf_of_shell(sh) .gt. mu) exit
      end do
      sh = sh - 1
      if (l .ne. B%bas(2, sh )) &
        stop "something is wrong: l/= B%bas(4,sh)!"
    end if
    do ix = 1, size(gridpoint_x)
      x = gridpoint_x(ix) - A(atom)%x
      do iy = 1, size(gridpoint_y)
        y = gridpoint_y(iy) - A(atom)%y
        do iz = 1, size(gridpoint_z)
          z = gridpoint_z(iz) - A(atom)%z
          values(iz, iy, ix) = ZERO
          call xyz_to_sph(x, y, z, r, theta, phi)
          if (.not. P%yes_gaussian) then
            radial_f = spline_interpolate(SP, r)/r
            values(iz, iy, ix) = radial_f * sph_harmonics_real(A(atom)%l_AO(k), m, theta, phi)
          else
            sphh = sph_harmonics_real(l, m, theta, phi)
            nprim = B%bas(3, sh)
            gto = ZERO
            radial_f = ZERO
            do i = 1, nprim
              alpha = B%env(B%bas(6, sh) + i)
              coef = B%env(B%bas(7, sh) + i)
              radial_f = radial_f + coef * exp(-alpha*r**2)
            end do
            gto = radial_f*r**l*sphh
            values(iz, iy, ix) = values(iz, iy, ix) + gto
          end if
        end do! iz
      end do! iy
    end do! ix
    if (.not. P%yes_gaussian) then
      deallocate (r_int)
      deallocate (f_int)
      call purge_Spline(SP)
    end if
  end subroutine calculate_bf_values


  subroutine calculate_bf_values_shell(P, A, B, values, gridpoint_x, gridpoint_y, gridpoint_z, sh)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B

    real(long), intent(in) ::  gridpoint_x(:)
    real(long), intent(in) ::  gridpoint_y(:)
    real(long), intent(in) ::  gridpoint_z(:)
    real(long), intent(out) ::  values(:, :, :, :)
    integer, intent(in) :: sh
    real(long), parameter :: exp_threshold = 20_long
    integer :: i, atom, n, l, m, N_r, k, mu

    integer :: ix, iy, iz
    real(long) :: x, y, z, theta, phi, r
    real(long) :: radial_f
    real(long), allocatable :: r_int(:), f_int(:)  ! for interpolation with zeros
    type(Spline) :: SP

    integer :: nprim
    real(long) :: alpha
    real(long) :: coef
    real(long) :: sphh
    real(long) :: gto
    if (P%yes_gaussian) then
      nprim = B%bas(3, sh)
      atom = B%bas(1, sh) + 1
      l = B%bas(2, sh)
    else
      stop 'calculate_bf_values_shell does not work for non-GTO basis sets: first_bf_of_shell is not filled for NAOs'
      mu = B%first_bf_of_shell(sh)
      atom = B%info(mu, 1)
      n = B%info(mu, 3)
      l = B%info(mu, 4)
      N_r = A(atom)%N_r
      allocate (r_int(N_r + 2), f_int(N_r + 2))
      r_int(1) = ZERO
      r_int(2:N_r + 1) = A(atom)%radial%r(:)
      r_int(N_r + 2) = A(atom)%rmax
      if (A(atom)%rmax .eq. ZERO) r_int(N_r + 2) = A(atom)%radial%r(N_r)*FOUR     ! put an enough large r
      f_int(1) = ZERO
      f_int(N_r + 2) = ZERO
      f_int(2:N_r + 1) = A(atom)%u_AO(:, k)                ! prepare interpolation of u(r)
      call spline_create(r_int(:), f_int(:), SP)
    end if
    values(:,:,:,:) = ZERO

    !$omp parallel do default(none) &
    !$omp shared(values, gridpoint_x, gridpoint_y, gridpoint_z, A, B, P, atom, SP, k, l, sh, nprim) &
    !$omp private(ix, iy, iz, x, y, z, theta, phi, r, gto, sphh, coef, alpha, m, radial_f) &
    !$omp schedule(static) collapse(3)
    do ix = 1, size(gridpoint_x)
      do iy = 1, size(gridpoint_y)
        do iz = 1, size(gridpoint_z)
          x = gridpoint_x(ix) - A(atom)%x
          y = gridpoint_y(iy) - A(atom)%y
          z = gridpoint_z(iz) - A(atom)%z
          call xyz_to_sph(x, y, z, r, theta, phi)
          if (.not. P%yes_gaussian) then
            radial_f = spline_interpolate(SP, r)/r
            do m = 1, 2 * l + 1
              values(m, iz, iy, ix) = radial_f * sph_harmonics_real(l, m-l-1, theta, phi)
            end do
          else
            if( all( B%env( B%bas(6, sh) + 1: B%bas(6, sh) + 1 + nprim) * r * r > exp_threshold) ) cycle
            gto = ZERO
            radial_f = ZERO
            do i = 1, nprim
              alpha = B%env(B%bas(6, sh) + i)
              coef = B%env(B%bas(7, sh) + i)
              radial_f = radial_f + coef * exp(-alpha*r**2)
            end do
            do m = 1, 2 * l + 1
              sphh = sph_harmonics_real(l, m - l - 1, theta, phi)
              gto = radial_f*r**l * sphh
              values(m, iz, iy, ix) = gto
            end do!m
          end if
        end do! iz
      end do! iy
    end do! ix
    !$omp end parallel do
    if (.not. P%yes_gaussian) then
      deallocate (r_int)
      deallocate (f_int)
      call purge_Spline(SP)
    end if

  end subroutine calculate_bf_values_shell





end module cubefile
