!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module eigen_lapack
  !! author: Sang-Kil Son
  !! author: Ludger Inhester

  !! Eigenproblem module : to solve eigenvalue problem using LAPACK
  !!   written by Son, Sang-Kil in Mar. 2005
  !!   davidson algorithm and interface to krylov library added by Ludger Inhester 2023
  !! USAGE: H should be a regular N x N matrix.
  !!        Eigenvectors are normalized after each subroutine.
  !! - SR get_eigenvalues( H, E )          for all ev. / standard eigen problem
  !!      get_eigenvalues( A, B, E )       for all ev. / generalized eigen problem
  !!   . SR get_all_ev_of_real_sym( H, E, yes_verbose )
  !!   . SR get_all_ev_of_real_nonsym( H, E, VL, VR, yes_verbose )
  !!   . SR get_all_ev_of_real_nonsym_simple( H, E, yes_verbose )
  !!   . SR get_all_ev_of_complex_Hermitian( H, E, yes_verbose )
  !!   . SR get_all_ev_of_complex_sym( H, E, yes_verbose )
  !!   . SR get_all_ev_of_complex_nonsym( H, E, VL, VR, yes_verbose )
  !!   . SR get_gen_all_ev_real_sym( A, B, E, yes_verbose )
  !!   . SR get_gen_all_ev_real_nonsym( A, B, E, VL, VR, yes_verbose )
  !! - SR get_eigenvalues( H, E, m, E0, dE )       for selected ev.
  !!   *** NOT YET IMPLEMENTED
  !!
  !! for real symmetric matrix
  !! - DSYEV  : all eigenvalues and eigenvectors
  !! - DSYEVD : all eigenvalues and eigenvectors by divide-and-conquer
  !! - DSYEVX : selected eigenvalues and eigenvectors
  !! - DSYEVR : selected eigenvalues and eigenvectors by divide-and-conquer
  !! - DSYGV  : all eigenvalues/eigenvectors of generalized symmetric-definite
  !!
  !! for real nonsymmetric matrix
  !! - DGEEV  : all eigenvalues and left/right eigenvectors
  !! - DGEEVX : all eigenvalues and left/right eigenvectors with balancing
  !! - DGEGV / DGGEV  : all eigenvalues/eigenvectors of generalized non-symmetric
  !!
  !! for complex Hermitian matrix
  !! - ZHEEV  : all eigenvalues and eigenvectors
  !! - ZHEEVD : all eigenvalues and eigenvectors by divide-and-conquer
  !! - ZHEEVX : selected eigenvalues and eigenvectors
  !! - ZHEEVR : selected eigenvalues and eigenvectors by divide-and-conquer
  !!
  !! for complex non-Hermitian matrix
  !! - ZGEEV  : all eigenvalues and left/right eigenvectors
  !! - ZGEEVX : all eigenvalues and left/right eigenvectors with balancing
  !!
  !! URL: http://www.netlib.org/lapack/double/
  !!      http://www.netlib.org/lapack/complex16/
  !!
  !!----------------------------------------------------------------------------
  use constants
  use matrix
  implicit none
  interface get_eigenvalues
    !! returns eigenvalues of square matrices
    module procedure get_all_ev_of_real_sym, &
      get_all_ev_of_real_nonsym, &
      get_all_ev_of_real_nonsym_simple, &
      get_all_ev_of_complex_Hermitian, &
      get_all_ev_of_complex_sym, &
      get_all_ev_of_complex_nonsym, &
      get_gen_all_ev_real_sym, &
      get_gen_all_ev_real_nonsym
  end interface
  interface check_normalized
    !!----------------------------------------------------------------------------
    !! Check if w.f. is normalized
    !!  \( < \psi_k | \psi_k' > \) = 1  if k = k'
    !!                             = 0  if k /= k'
    !! with the matrix form
    !!     wf'  wf = I  for a real matrix
    !!     wf^T wf = I  for a complex matrix
    !!----------------------------------------------------------------------------
    module procedure check_normal_real_sym, &
      check_normal_real_nonsym, &
      check_normal_complex_Hermitian, &
      check_normal_complex_nonsym
  end interface
  private
  public :: get_eigenvalues_symmetry, get_eigenvalues, get_some_ev_of_real_sym, &
    get_some_eigenvalues_davpack
#ifdef WITH_ARPACK
  public :: get_some_eigenvalues_arpack
#endif
#ifdef WITH_KRYLOV
  public :: get_some_eigenvalues_libKrylov
#endif

contains

!----------------------------------------------------------------------------
! Solve the eigenvalue problem of H(N,N) real symmetric matrix
! - eigenvalues:  real, E(N)
! - eigenvectors: real, H(N,N)
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_real_sym(H, E, yes_verbose)
    real(long), intent(inout) :: H(:, :)
    real(long), intent(inout) :: E(:)
    logical, intent(in), optional :: yes_verbose
    real(long) :: WORK(size(E)*size(E))
    integer :: N
    integer :: INFO, LDWORK
    real(long) :: temp, b, d1, d2
    N = size(E)
    INFO = 0
    LDWORK = max(N*N, 3*N - 1)
    if (N==1) then
      E(1) = H(1,1)
      H(1,1) = ONE
    else if (N==2)then
      if (H(1,2) /= ZERO) then
        b = H(1,2)
        d1 = H(1,1)
        d2 = H(2,2)
        temp = sqrt(FOUR * b**2 + (d1 - d2)**2 )
        H(1,1) = - (d2 - d1 + temp) / (TWO * b)
        H(1,2) = (d1 - d2 + temp) / (TWO * b)
        H(2,1) = ONE
        H(2,2) = ONE
        H(:,1) = H(:,1) / sqrt(sum(H(:,1)**2))
        H(:,2) = H(:,2) / sqrt(sum(H(:,2)**2))
        E(1) = HALF * (d1 + d2 - temp)
        E(2) = HALF * (d1 + d2 + temp)
      else
        E(1) = H(1,1)
        E(2) = H(2,2)
        H(:, :)=ZERO
        H(1, 1) = ONE
        H(2, 2) = ONE
      end if
    else
      !
      ! LAPACK:
      !   DSYEV(JOBZ, UPLO, N, A, LDA, W, WORK, LDWORK, INFO)
      !
      call DSYEV('V', 'U', N, H, N, E, WORK, LDWORK, INFO)
      if (INFO .ne. 0) then
        write (STDERR, *) "Error in DSYEV : INFO = ", INFO
        stop
      end if
      if (present(yes_verbose)) then
        if (yes_verbose) write (STDERR, *) "Optimal LDWORK in DSYEV = ", WORK(1)
      end if
    end if
  end subroutine get_all_ev_of_real_sym

!>----------------------------------------------------------------------------
!! Solve the eigenvalue problem of H(N,N) real symmetric matrix using dsyevx
!! - matrix: real, H(N,N)
!! - eigenvalues:  real, E(num)
!! - eigenvectors: real, Z(N,num)
!! - num: number of eigenvectors and eigenvalues with lowest eigenvalues
!! - epsilon: (optional) absolute tolerance
!! Eigenvalues are stored in E(N)
!!----------------------------------------------------------------------------
  subroutine get_some_ev_of_real_sym(H, E, Z, num, epsilon, yes_verbose)
    real(long), intent(inout) :: H(:, :)
    real(long), intent(out) :: E(num)
    real(long), intent(out) :: Z(size(H, 1), num)
    integer, intent(in) :: num
    real(long), intent(in), optional :: epsilon
    logical, intent(in), optional :: yes_verbose
    real(long) :: WORK(max(8*size(H, 1), 2))
    integer :: N, M, IL, IU
    integer :: INFO, LWORK
    integer :: IWORK(max(5*size(H, 1), 1)), IFAIL(size(H, 1))
    real(long) :: VL, VU
    real(long) :: abstol
    abstol = 1e-10_long ! default absolute tolerance
    if (present(epsilon)) &
      abstol = epsilon
    VL = ZERO ! not referenced
    VU = ZERO ! not referenced
    IL = 1
    IU = num
    N = size(H, 1)
    INFO = 0
    LWORK = size(WORK)
    !
    ! LAPACK:
    !   DSYEVX(JOBZ, UPLO, N, A, LDA, W, WORK, LDWORK, INFO)
    !L DSYEVX( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU, ABSTOL, M, W, Z, LDZ, WORK, LWORK, IWORK, IFAIL, INFO )
    call DSYEVX('V', 'I', 'U', N, H, N, VL, VU, IL, IU, abstol, M, E, Z, N, WORK, LWORK, IWORK, IFAIL, INFO)
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DSYEVX : INFO = ", INFO
      if (INFO .gt. 0) then
        write (STDERR, *) "Failed to obtain eigenvalues index ", IFAIL(:)
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in DSYEVX = ", WORK(1)
    end if
  end subroutine get_some_ev_of_real_sym

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) real nonsym. matrix using LAPACK
! - eigenvalues:  complex, E(N)
! - eigenvectors are not computed.
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_real_nonsym_simple(H, E, yes_verbose)
    real(long), intent(inout) :: H(:, :)
    complex(long), intent(out) :: E(:)
    logical, intent(in), optional :: yes_verbose
    real(long) :: WORK(64*size(E)), WR(size(E)), WI(size(E))
    real(long) :: dummy(1)
    complex(long) :: temp
    integer :: INFO, LDA, LDVL, LDVR, LWORK
    integer :: N, k, l, loc(1)
    external DGEEV
    N = size(E)
    INFO = 0      ! INFO must be a variable for NAG in Linux machine
    LDA = size(H, dim=1)
    LDVR = 1; LDVL = 1
    LWORK = 64*N
    !
    ! LAPACK: http://www.netlib.org/lapack/double/dgeev.f
    !         DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, LDVR,
    !                WORK, LWORK, INFO )
    !
    call DGEEV('N', 'N', N, H, LDA, WR, WI, dummy, LDVL, dummy, LDVR, &
               WORK, LWORK, INFO)
    !
    ! set complex numbers of E
    !
    E(:) = dcmplx(WR(:), WI(:))
    !
    ! sort eigenvalues and eigenvectors by real(E)
    !
    do k = 1, N - 1
      if (dble(E(k)) .gt. minval(dble(E(k + 1:N)))) then
        !
        ! swap i and minloc(E)
        !
        loc = minloc(dble(E(k + 1:N)))
        l = loc(1) + k
        temp = E(k)
        E(k) = E(l)
        E(l) = temp
      end if
    end do
    !
    ! if there is an error
    !
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DGEEV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        write (STDERR, *) "the QR algorithm failed to compute all the eigenvalues,"
        write (STDERR, *) "and no eigenvectors have been computed."
        if (INFO .lt. N) then
          write (STDERR, *) "Some converged eigenvalues:"
          do k = INFO + 1, N
            write (STDERR, *) WR(k), WI(k)
          end do
        else
          write (STDERR, *) "There is no converged eigenvalues."
        end if
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in ZGEEV = ", WORK(1)
    end if
  end subroutine get_all_ev_of_real_nonsym_simple

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) real nonsym. matrix using LAPACK
! All eigenvalues are stored in E(N) and left eigenvectors in VL(N,N)
!                                        right eigenvectors in VR(N,N)
!
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL^T A = D VL^T
!
! take transpose of first eq.
!    VR^T A^T = D^T VR^T
! consequently, left eigenvectors of A^T is VR^T.
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_real_nonsym(H, E, QL, QR, yes_verbose)
    real(long), intent(inout) :: H(:, :)
    complex(long), intent(out) :: E(:)
    complex(long), intent(out) :: QL(:, :), QR(:, :)
    logical, intent(in), optional :: yes_verbose
    real(long) :: WORK(64*size(E)), WR(size(E)), WI(size(E))
    real(long) :: VL(size(E), size(E)), VR(size(E), size(E))
    integer :: INFO, LDA, LDVL, LDVR, LWORK
    integer :: N, k
    external DGEEV
    N = size(E)
    INFO = 0      ! INFO must be a variable for NAG in Linux machine
    LDA = size(H, dim=1)
    LDVR = size(VR, dim=1)
    LDVL = size(VL, dim=1)
    LWORK = 64*N
    !
    ! LAPACK: http://www.netlib.org/lapack/double/dgeev.f
    !         DGEEV( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, LDVR,
    !                WORK, LWORK, INFO )
    !
    call DGEEV('V', 'V', N, H, LDA, WR, WI, VL, LDVL, VR, LDVR, &
               WORK, LWORK, INFO)
    !
    ! set eigenvalues and eigenvectors
    !
    k = 1
    do
      if (WI(k) .eq. ZERO) then
        E(k) = WR(k)
        QR(:, k) = VR(:, k)
        QL(:, k) = VL(:, k)
      else
        E(k) = dcmplx(WR(k), WI(k))
        E(k + 1) = dcmplx(WR(k), -WI(k))
        QR(:, k) = VR(:, k) + IMAG_ONE*VR(:, k + 1)
        QR(:, k + 1) = VR(:, k) - IMAG_ONE*VR(:, k + 1)
        QL(:, k) = VL(:, k) + IMAG_ONE*VL(:, k + 1)
        QL(:, k + 1) = VL(:, k) - IMAG_ONE*VL(:, k + 1)
        k = k + 1
      end if
      k = k + 1
      if (k .gt. N) exit
    end do
    !
    ! We need normalization for w.f., < VL(k) | VR(k) > = 1
    !
!  do k = 1, N
!    norm = sqrt( sum( conjg( QL(:,k) ) * QR(:,k) ) )
!    QR(:,k) = QR(:,k) / norm
!    QL(:,k) = QL(:,k) / norm
!  end do
    !
    ! if there is an error
    !
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DGEEV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        write (STDERR, *) "the QR algorithm failed to compute all the eigenvalues,"
        write (STDERR, *) "and no eigenvectors have been computed."
        if (INFO .lt. N) then
          write (STDERR, *) "Some converged eigenvalues:"
          do k = INFO + 1, N
            write (STDERR, *) WR(k), WI(k)
          end do
        else
          write (STDERR, *) "There is no converged eigenvalues."
        end if
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in ZGEEV = ", WORK(1)
    end if
  end subroutine get_all_ev_of_real_nonsym

!----------------------------------------------------------------------------
! Solve the eigenvalue problem of H(N,N) complex Hermitian matrix
! - eigenvalues:  real, E(N)
! - eigenvectors: complex, H(N,N)
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_complex_Hermitian(H, E, yes_verbose)
    complex(long), intent(inout) :: H(:, :)
    real(long), intent(inout) :: E(:)
    logical, intent(in), optional :: yes_verbose
    complex(long) :: WORK(size(E)*size(E))
    real(long) :: WORK2(3*size(E) - 2)
    integer :: N, INFO, LDWORK
    N = size(E)
    INFO = 0
    LDWORK = max(N*N, 2*N - 1)
    !
    ! LAPACK:
    !   ZHEEV(JOBZ, UPLO, N, A, LDA, W, WORK, LDWORK, WORK2, INFO)
    !
    call ZHEEV('V', 'U', N, H, N, E, WORK, LDWORK, WORK2, INFO)
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in ZHEEV : INFO = ", INFO
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in ZHEEV = ", WORK(1)
    end if
  end subroutine get_all_ev_of_complex_Hermitian

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) complex symmetric matrix using LAPACK
! All eigenvalues are stored in E(N) and right eigenvectors in H(N,N)
!
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL A = D VL
!
! take the transpose of first eq.
!    VR^T A^T = D^T VR^T
! here D^T = D; thus, VR^T A^T = D VR^T
! consequently, left eigenvectors of A^T is VR^T.
!
! In this case of the complex symmetric matrix, A = A^T
! so, VL can be simply solved by VR^T.
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_complex_sym(H, E, yes_verbose)
    complex(long), intent(inout) :: H(:, :), E(:)
    logical, intent(in), optional :: yes_verbose
    complex(long) :: VR(size(E), size(E)), VL(1, 1)
    complex(long) :: WORK(64*size(E))
    real(long) :: RWORK(2*size(E))
    complex(long) :: norm, temp, temp_v(size(E))
    integer :: INFO, LDA, LDVL, LDVR, LWORK
    integer :: N, k, l, loc(1)
    external ZGEEV
    N = size(E)
    INFO = 0      ! INFO must be a variable for NAG in Linux machine
    LDA = N
    LDVR = N
    LDVL = 1
    LWORK = 64*N
    !
    ! LAPACK:
    !   ZGEEV( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR,
    !          WORK, LWORK, RWORK, INFO )
    !
    call ZGEEV('N', 'V', N, H, LDA, E, VL, LDVL, VR, LDVR, &
               WORK, LWORK, RWORK, INFO)
    !
    ! We need normalization for w.f., < VL(k) | VR(k) > = 1
    ! Here, H is complex symmetric. VL = VR^T
    !
    do k = 1, N
      norm = sqrt(sum(VR(:, k)**2))
      H(:, k) = VR(:, k)/norm
    end do
    !
    ! sort eigenvalues and eigenvectors by real(E)
    !
    do k = 1, N - 1
      if (real(E(k)) .gt. minval(real(E(k + 1:N)))) then
        !
        ! swap i and minloc(E)
        !
        loc = minloc(real(E(k + 1:N)))
        l = loc(1) + k
        temp = E(k); E(k) = E(l); E(l) = temp
        temp_v(:) = H(:, k); H(:, k) = H(:, l); H(:, l) = temp_v(:)
      end if
    end do
    !
    ! if there is an error
    !
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in ZGEEV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        write (STDERR, *) "the QR algorithm failed to compute all the "// &
          "eigenvalues, and no eigenvectors have been computed."
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in ZGEEV = ", WORK(1)
    end if
  end subroutine get_all_ev_of_complex_sym

!----------------------------------------------------------------------------
! Solve the eigen. problem of H(N,N) complex nonsym. matrix using LAPACK
! All eigenvalues are stored in E(N) and left eigenvectors in VL(N,N)
!                                        right eigenvectors in VR(N,N)
!
! Left and right eigenvectors are defined by
!    A VR = VR D        where D = diag( eigenvalues )
!    VL^* A = D VL^*
!
! take transpose + conjugate of first eq.
!    VR^* A^* = D^* VR^*
! consequently, left eigenvectors of A^* is VR^*.
!----------------------------------------------------------------------------
  subroutine get_all_ev_of_complex_nonsym(H, E, VL, VR, yes_verbose)
    complex(long), intent(inout) :: H(:, :)
    complex(long), intent(out) :: E(:), VL(:, :), VR(:, :)
    logical, intent(in), optional :: yes_verbose
    real(long) :: RWORK(2*size(E))
    complex(long) :: WORK(64*size(E))
    complex(long) :: norm
    complex(long) :: temp, temp_v(size(E))
    integer :: INFO, LDA, LDVL, LDVR, LWORK
    integer :: N, k, l, loc(1)
    external ZGEEV
    N = size(E)
    INFO = 0      ! INFO must be a variable for NAG in Linux machine
    LDA = size(H, dim=1)
    LDVR = size(VR, dim=1)
    LDVL = size(VL, dim=1)
    LWORK = 64*N
    !
    ! LAPACK: http://www.netlib.org/lapack/complex16/zgeev.f
    !         ZGEEV( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR,
    !                WORK, LWORK, RWORK, INFO )
    !
    call ZGEEV('V', 'V', N, H, LDA, E, VL, LDVL, VR, LDVR, &
               WORK, LWORK, RWORK, INFO)
    !
    ! We need normalization for w.f., < VL(k) | VR(k) > = 1
    !
    do k = 1, N
      norm = sqrt(sum(conjg(VL(:, k))*VR(:, k)))
      VR(:, k) = VR(:, k)/norm
      VL(:, k) = VL(:, k)/conjg(norm)
    end do
    !
    ! sort eigenvalues and eigenvectors by real(E)
    !
    do k = 1, N - 1
      if (dble(E(k)) .gt. minval(dble(E(k + 1:N)))) then
        !
        ! swap i and minloc(E)
        !
        loc = minloc(dble(E(k + 1:N)))
        l = loc(1) + k
        temp = E(k); E(k) = E(l); E(l) = temp
        temp_v(:) = VR(:, k); VR(:, k) = VR(:, l); VR(:, l) = temp_v(:)
        temp_v(:) = VL(:, k); VL(:, k) = VL(:, l); VL(:, l) = temp_v(:)
      end if
    end do
    !
    ! if there is an error
    !
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in ZGEEV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        write (STDERR, *) "the QR algorithm failed to compute all the "// &
          "eigenvalues, and no eigenvectors have been computed."
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in ZGEEV = ", WORK(1)
    end if
  end subroutine get_all_ev_of_complex_nonsym

!----------------------------------------------------------------------------
! Solve the generalized eigenvalue problem of real symmetric matrix
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
  subroutine get_gen_all_ev_real_sym(A, B, E, yes_verbose)
    real(long), intent(inout) :: A(:, :), B(:, :)
    real(long), intent(inout) :: E(:)
    logical, intent(in), optional :: yes_verbose
    real(long) :: WORK(size(E)*size(E))
    integer :: N
    integer :: ITYPE, INFO, LDWORK
    N = size(E)
    ITYPE = 1
    INFO = 0
    LDWORK = max(N*N, 3*N - 1)
    !
    ! LAPACK:
    !   DSYGV(ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, LDWORK, INFO)
    !     ITYPE = 1: A * x = lambda * B * x
    !           = 2: A * B * x = lambda * x
    !           = 3: B * A * x = lambda * x
    !
    call DSYGV(ITYPE, 'V', 'U', N, A, N, B, N, E, WORK, LDWORK, INFO)
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DSYGV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        write (STDERR, *) "DPOTRF or DSYEV returned an error code:"
        if (INFO .le. N) then
          write (STDERR, *) "  DSYEV failed to converge;"
          write (STDERR, *) "  ", INFO, "off-diagonal elements of an intermediate"
          write (STDERR, *) "  tridiagonal form did not converge to zero;"
        else
          write (STDERR, *) "  the leading minor of order ", INFO - N, &
            " of B is not positive definite."
          write (STDERR, *) "  The factorization of B could not be completed and "
          write (STDERR, *) "  no eigenvalues or eigenvectors were computed."
        end if
      end if
      stop
    end if
    if (present(yes_verbose)) then 
     if (yes_verbose) write (STDERR, *) "Optimal LDWORK in DSYGV = ", WORK(1)
    end if
  end subroutine get_gen_all_ev_real_sym

!----------------------------------------------------------------------------
! Solve the generalized eigenvalue problem of real nonsymmetric matrix
! Eigenvalues are stored in E(N)
!----------------------------------------------------------------------------
  subroutine get_gen_all_ev_real_nonsym(A, B, E, VL, VR, yes_verbose)
    real(long), intent(inout) :: A(:, :), B(:, :)
    complex(long), intent(inout) :: E(:)
    real(long), intent(inout) :: VL(:, :), VR(:, :)
    logical, intent(in), optional :: yes_verbose
    real(long) :: ALPHAR(size(E)), ALPHAI(size(E)), BETA(size(E))
    real(long) :: WORK(size(E)*size(E))
    real(long) :: temp_v(size(E))
    complex(long) :: temp
    integer :: N, k, l, loc(1)
    integer :: ITYPE, INFO, LDWORK
    N = size(E)
    ITYPE = 1
    INFO = 0
    LDWORK = max(N*N, 8*N)
    !
    ! LAPACK:
    !   DGGEV(JOBVL, JOBVR, N, A, LDA, B, LDB, ALPHAR, ALPHAI,
    !         BETA, VL, LDVL, VR, LDVR, WORK, LWORK, INFO )
    !
    call DGGEV('V', 'V', N, A, N, B, N, ALPHAR, ALPHAI, BETA, VL, N, VR, N, &
               WORK, LDWORK, INFO)
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DGGEV : INFO = ", INFO
      if (INFO .lt. 0) then
        write (STDERR, *) "the ", -INFO, "th argument had an illegal value."
      else
        if (INFO .le. N) then
          write (STDERR, *) "  The QZ iteration failed.  No eigenvectors have been"
          write (STDERR, *) "  calculated, but ALPHAR(j), ALPHAI(j), and BETA(j)"
          write (STDERR, *) "  should be correct for j=INFO+1,...,N."
        else
          write (STDERR, *) "  =N+1: other than QZ iteration failed in DHGEQZ."
          write (STDERR, *) "  =N+2: error return from DTGEVC."
        end if
      end if
      stop
    end if
    if (present(yes_verbose)) then
      if (yes_verbose) write (STDERR, *) "Optimal LDWORK in DGGEV = ", WORK(1)
    end if
    !
    ! get eigenvalues from ALPHAR, ALPHAI, and BETA
    !
    E(:) = (ALPHAR(:) + IMAG_ONE*ALPHAI(:))/BETA(:)
    !
    ! sort eigenvalues and eigenvectors by real(E), i.e. ALPHAR
    !
    do k = 1, N - 1
      if (dble(E(k)) .gt. minval(dble(E(k + 1:N)))) then
        !
        ! swap i and minloc(E)
        !
        loc = minloc(dble(E(k + 1:N)))
        l = loc(1) + k
        temp = E(k); E(k) = E(l); E(l) = temp
        temp_v(:) = VR(:, k); VR(:, k) = VR(:, l); VR(:, l) = temp_v(:)
        temp_v(:) = VL(:, k); VL(:, k) = VL(:, l); VL(:, l) = temp_v(:)
      end if
    end do
  end subroutine get_gen_all_ev_real_nonsym

  subroutine check_normal_real_sym(wf, tol)
    real(long), intent(in) :: wf(:, :)
    real(long), intent(in) :: tol
    real(long) :: C(size(wf, dim=1), size(wf, dim=2))
    real(long) :: alpha, beta
    logical :: yes_first
    integer :: N, i, j
    alpha = ONE; beta = ZERO
    N = size(wf, dim=1)
    call DGEMM('T', 'N', N, N, N, alpha, wf, N, wf, N, beta, C, N)
    yes_first = .true.
    do i = 1, N
      do j = 1, N
        if ((i .eq. j .and. abs(C(i, j) - ONE) .gt. tol) &
            .or. (i .ne. j .and. abs(C(i, j)) .gt. tol)) then
          if (yes_first) then
            write (STDERR, "(a)") "wavefunctions are not normalized:"
            yes_first = .false.
          end if
          write (STDERR, "(2x,'(i,j,value) = ',2i10,es25.15)") i, j, C(i, j)
        end if
      end do
    end do
  end subroutine check_normal_real_sym

  subroutine check_normal_real_nonsym(wf_L, wf_R, tol)
    real(long), intent(in) :: wf_L(:, :), wf_R(:, :)
    real(long), intent(in) :: tol
    real(long) :: C(size(wf_L, dim=2), size(wf_R, dim=2))
    real(long) :: alpha, beta
    logical :: yes_first
    integer :: N, i, j
    alpha = ONE; beta = ZERO
    N = size(wf_R, dim=1)
    call DGEMM('T', 'N', N, N, N, alpha, wf_L, N, wf_R, N, beta, C, N)
    yes_first = .true.
    do i = 1, N
      do j = 1, N
        if ((i .eq. j .and. abs(C(i, j) - ONE) .gt. tol) &
            .or. (i .ne. j .and. abs(C(i, j)) .gt. tol)) then
          if (yes_first) then
            write (STDERR, "(a)") "wavefunctions are not normalized:"
            yes_first = .false.
          end if
          write (STDERR, "(2x,'(i,j,value) = ',2i10,es25.15)") i, j, C(i, j)
        end if
      end do
    end do
  end subroutine check_normal_real_nonsym

  subroutine check_normal_complex_Hermitian(wf, tol)
    complex(long), intent(in) :: wf(:, :)
    real(long), intent(in) :: tol
    complex(long) :: C(size(wf, dim=1), size(wf, dim=2))
    complex(long) :: alpha, beta
    logical :: yes_first
    integer :: N, i, j
    alpha = ONE; beta = ZERO
    N = size(wf, dim=1)
    call ZGEMM('C', 'N', N, N, N, alpha, wf, N, wf, N, beta, C, N)
    yes_first = .true.
    do i = 1, N
      do j = 1, N
        if ((i .eq. j .and. abs(C(i, j) - ONE) .gt. tol) &
            .or. (i .ne. j .and. abs(C(i, j)) .gt. tol)) then
          if (yes_first) then
            write (STDERR, "(a)") "wavefunctions are not normalized:"
            yes_first = .false.
          end if
          write (STDERR, "(2x,'(i,j,value) = ',2i10,es25.15)") i, j, abs(C(i, j))
        end if
      end do
    end do
  end subroutine check_normal_complex_Hermitian

  subroutine check_normal_complex_nonsym(wf_L, wf_R, tol)
    complex(long), intent(in) :: wf_L(:, :), wf_R(:, :)
    real(long), intent(in) :: tol
    complex(long) :: C(size(wf_L, dim=2), size(wf_R, dim=2))
    complex(long) :: alpha, beta
    logical :: yes_first
    integer :: N, i, j
    alpha = ONE; beta = ZERO
    N = size(wf_R, dim=1)
    call ZGEMM('C', 'N', N, N, N, alpha, wf_L, N, wf_R, N, beta, C, N)
    yes_first = .true.
    do i = 1, N
      do j = 1, N
        if ((i .eq. j .and. abs(C(i, j) - ONE) .gt. tol) &
            .or. (i .ne. j .and. abs(C(i, j)) .gt. tol)) then
          if (yes_first) then
            write (STDERR, "(a)") "wavefunctions are not normalized:"
            yes_first = .false.
          end if
          write (STDERR, "(2x,'(i,j,value) = ',2i10,es25.15)") i, j, abs(C(i, j))
        end if
      end do
    end do
  end subroutine check_normal_complex_nonsym

  !>----------------------------------------------------------------------------
  !! Solve the eigenvalue problem of a tridiagonal symmetric matrix
  !! - input: diagonal entries
  !! - input: sub-diagonal entries, input: real, E(N-1)
  !! - output: eigenvalues, real, D(N)
  !! - output: eigenvectors, real, Z(N,N)
  !!----------------------------------------------------------------------------
  subroutine get_all_ev_of_tridiagonal_sym(D, E, Z, yes_get_ev)
    real(long), intent(inout) :: D(:) ! diagonal entries (in) eigenvalues (out)
    real(long), intent(inout) :: E(size(D, 1) - 1) ! subdiagonal (in)
    real(long), intent(inout) :: Z(size(D, 1), size(D, 1)) ! eigenvectors (out)
    logical, intent(in), optional :: yes_get_ev
    logical :: get_ev
    real(long) :: WORK(size(D)*2)
    integer :: N, LDZ
    integer :: INFO
    N = size(D)
    LDZ = N
    INFO = 0
    get_ev = .True.
    if (present(yes_get_ev)) &
      get_ev = yes_get_ev
    Z = ZERO
    if (get_ev) then
      call DSTEQR('I', N, D, E, Z, LDZ, WORK, INFO)
    else
      call DSTEQR('N', N, D, E, Z, LDZ, WORK, INFO)
    end if
    if (INFO .ne. 0) then
      write (STDERR, *) "Error in DSTEGR : INFO = ", INFO
      stop
    end if
  end subroutine get_all_ev_of_tridiagonal_sym

  subroutine get_eigenvalues_symmetry(M, E, irrep)
    !! computes eigenvalues of square matrices
    !! under consideration that there a extra symmetries
    real(long), intent(inout) :: M(:, :)
    real(long), intent(inout) :: E(:)
    integer, intent(in) :: irrep(:)
    integer :: i, j
    real(long), allocatable :: M_irrep(:, :)
    real(long), allocatable :: E_irrep(:)
    integer, allocatable :: idx(:)
    integer :: nirrep
    integer :: N
    N = size(irrep, 1)
    nirrep = maxval(irrep)

    do i = 1, N
      do j = 1, i
        if (irrep(i) .ne. irrep(j)) then
          if (abs(M(i, j)) .gt. 1e-10 .or. abs(M(j, i)) .gt. 1e-10) then
            write (STDERR, *) 'get_eigenvalues_symmetry: Warning unsymmetric matrix entries found ', i, j, M(i, j), M(j, i)
            write (STDERR, *) '-> set to zero !!!'
          end if
          M(i, j) = ZERO
          M(j, i) = ZERO
        end if
      end do
    end do

    do i = 1, nirrep
      if (count(irrep .eq. i) .eq. 0) cycle
      allocate (idx(count(irrep .eq. i)))
      idx = PACK([(j, j=1, N)], [(irrep(j) .eq. i, j=1, N)])
      allocate (M_irrep(size(idx), size(idx)))
      allocate (E_irrep(size(idx)))
      M_irrep = M(idx, idx)

!    write(STDERR,'("# ",a,i4)') 'irrep',i
!    do j=1,size(idx)
!      write(STDERR,'("# ",10f5.1)') M_irrep(j,:)
!    end do
      call get_eigenvalues(M_irrep, E_irrep)
      M(idx, idx) = M_irrep
      E(idx) = E_irrep
      deallocate (M_irrep)
      deallocate (E_irrep)
      deallocate (idx)
    end do

  end subroutine get_eigenvalues_symmetry

!> Eigenvalues using the Davidson algorithm
!! n: dimension
!! nev: number of eigenvalues to obtain
!! ndim: size of subspace to keep (iterations after which to do a restart)
!! E(1:nev) returned eigenvalues
!! C(1:n,1:nev) on input: guessed eigenvectors, on output: found eigenvectors
!! mv: matrix vector multiplication subroutine
!! epsilon: convergence criterion (for the rms of residue)
!! OUTPUT (optional): output
  subroutine get_some_eigenvalues_Davidson(n, nev, ndim, E, C, diag, mv, epsilon, OUTPUT)
    integer, intent(in) :: n, nev, ndim
    real(long), intent(out) :: E(nev)
    real(long), intent(inout) :: C(n, nev)
    real(long), intent(in) :: diag(n)
    real(long), intent(in) :: epsilon
    integer :: mv
    integer, intent(in), optional :: OUTPUT
    integer :: IOUT
    integer :: k, j
    integer :: error
    real(long) :: Vk(n, ndim)
    real(long) :: mVk(n, ndim)
    real(long) :: hk(ndim, ndim)
    real(long) :: ek(ndim, ndim)
    real(long) :: muk(ndim)
    real(long) :: xk(n)
    real(long) :: rk(n)
    real(long) :: tkp(n)
    real(long) :: proj(ndim)
    IOUT = 6
    if (present(OUTPUT)) &
      IOUT = OUTPUT
    Vk(:, 1) = C(:, 1)/sqrt(sum(C(:, 1)**2))
    do
      do k = 1, ndim
        error = mv(n, Vk(:, k), mVk(:, k))
        do j = 1, k - 1
          hk(k, j) = sum(Vk(:, k)*mVk(:, j))
          hk(j, k) = sum(Vk(:, j)*mVk(:, k))
        end do
        hk(k, k) = sum(Vk(:, k)*mVk(:, k))
        ek(:, :) = hk(:, :)
        call get_eigenvalues(ek(1:k, 1:k), muk(1:k))
        write (IOUT, *) 'lowest ev', muk(1)
        xk(:) = matmul(Vk(:, 1:k), ek(1:k, 1))
        !mv(n, xk(:), rk(:))
        rk(:) = muk(1)*xk(:) - (mVk(:, 1:k) .m.ek(1:k, 1))
        write (6, *) 'res', sqrt(sum(rk(:)**2))
        if (sqrt(sum(rk(:)**2)) .lt. epsilon) exit
        if (k .eq. ndim) exit
        ! tkp = (muk I - diag)^-1 rk
        tkp(:) = ONE/(muk(1) - diag(:) + 1e-10)*rk(:)

        if (.False.) then
          call gram_schmidt(Vk(:, 1:k), tkp)
        else
          proj(1:k) = Vk(:, 1:k) .tm.tkp(:)
          tkp(:) = tkp(:) - (Vk(:, 1:k) .m.proj(1:k))
          tkp(:) = tkp(:)/sqrt(sum(tkp(:)**2))
        end if
        Vk(:, k + 1) = tkp
      end do
      if (sqrt(sum(rk(:)**2)) .lt. epsilon) exit
      write (IOUT, *) 'restart'
      Vk(:, 1) = xk(:)
    end do
    mVk(:, 1:k) = Vk(:, 1:k) .mt.ek(1:k, 1:k)
    C(:, 1:nev) = mVk(:, 1:nev)
    E(1:nev) = muk(1:nev)
  end subroutine get_some_eigenvalues_Davidson

  subroutine get_some_eigenvalues_BlockDavidson(n, nev, ndim, E, C, diag, mm, epsilon, OUTPUT)
    !! Implementation of the Block-Davidson algorithm
    integer, intent(in) :: n !! dimension
    integer, intent(in) :: nev !! number of eigenvalues to obtain = size of the block
    integer, intent(in) :: ndim !! size of subspace to keep, iterations after which to perform a restart
    real(long), intent(out) :: E(nev) !! returned eigenvalues
    real(long), intent(inout) :: C(n, nev) !! on input: guessed eigenvectors, on output: found eigenvectors
    real(long), intent(in) :: diag(n) !! diagonal matrix elements
    real(long), intent(in) :: epsilon
    integer :: mm !! matrix matrix multiplication subroutine
    integer, intent(in), optional :: OUTPUT !! output
    integer :: IOUT
    integer :: k, i, j
    real(long) :: Vk(n, ndim*nev)
    real(long) :: Wk(n, ndim*nev)
    real(long) :: hk(ndim*nev, ndim*nev)
    real(long) :: ek(ndim*nev, ndim*nev)
    real(long) :: diagHSave(ndim*nev)
    real(long) :: muk(ndim*nev)
    real(long) :: xk(n, nev)
    real(long) :: rk(n, nev)
    real(long) :: tkp(n, nev)
    real(long) :: Ak(n, nev)
    logical :: restarted = .False.
    integer :: error

    if (ndim*nev .gt. n*0.75) then
      stop 'ndim*nev too large'
    end if
    IOUT = 6
    if (present(OUTPUT)) &
      IOUT = OUTPUT
    do j = 1, nev
      if (j .eq. 1) then
        Vk(:, j) = C(:, j)/sqrt(sum(C(:, j)**2))
      else
        call gram_schmidt(Vk(:, 1:j - 1), C(:, j))
        Vk(:, j) = C(:, j)
      end if
    end do
    do
      do k = 1, ndim
        if (restarted) then
          restarted = .False.
          cycle
        end if
        error = mm(n, nev, Vk(:, (k - 1)*nev + 1:(k)*nev), Wk(:, (k - 1)*nev + 1:(k)*nev))
        !hk(1:k*nev,1:k*nev) = Vk(:,1:k*nev).tm.Wk(:,1:k*nev)
        hk(1:k*nev, (k - 1)*nev + 1:k*nev) = Vk(:, 1:k*nev) .tm.Wk(:, (k - 1)*nev + 1:k*nev)
        hk((k - 1)*nev + 1:k*nev, 1:k*nev) = Vk(:, (k - 1)*nev + 1:k*nev) .tm.Wk(:, 1:k*nev)
        hk((k - 1)*nev + 1:k*nev, (k - 1)*nev + 1:k*nev) = Vk(:, (k - 1)*nev + 1:k*nev) .tm.Wk(:, (k - 1)*nev + 1:k*nev)

        do j = 1, k*nev
          diagHSave(j) = hk(j, j)
        end do

        call get_some_ev_of_real_sym(hk(1:k*nev, 1:k*nev), muk(1:nev), ek(1:k*nev, 1:nev), nev, yes_verbose=.False.)
        ! restore hk because the upper triangle and diagonal was modified by dsyevx
        do j = 1, k*nev
          hk(j, j) = diagHSave(j)
          hk(j, j + 1:k*nev) = hk(j + 1:k*nev, j)
        end do

        xk(:, 1:nev) = Vk(:, 1:k*nev) .m.ek(1:k*nev, 1:nev)

        Ak(:, :) = Wk(:, 1:k*nev) .m.ek(1:k*nev, 1:nev)
        do i = 1, nev
          !rk(:,i) = xk(:,i) * muk(i) - (Wk(:,1:k*nev) .m. ek(1:k*nev,i))
          rk(:, i) = xk(:, i)*muk(i) - Ak(:, i)
          write (IOUT, *) 'res', i, sqrt(sum(rk(:, i)**2))
        end do

        if (sqrt(sum(rk(:, :)**2)) .lt. epsilon) exit

        ! tkp = (muk I - diag)^-1 rk
        do i = 1, nev
          !tkp(:,i) = ONE/(muk(i) - diag(:) + 1e-10) * rk(:,i)
          !preconditioner from CROUZEIX et al. SIAM J. ScI. COMPUT. Vol. 15, No. 1, pp. 62-76, January 1994
          do j = 1, size(tkp, 1)
            tkp(j, i) = min(ONE/abs(muk(i) - diag(j)), 1e10)*rk(j, i)
          end do
        end do

        if (k .eq. ndim) then
          exit
        else
          do i = 1, nev
            call gram_schmidt(Vk(:, 1:k*nev + (i - 1)), tkp(:, i))
            Vk(:, k*nev + i) = tkp(:, i)
          end do
        end if
      end do
      if (sqrt(sum(rk(:, :)**2)) .lt. epsilon) exit
      write (IOUT, *) 'restart'
      ! Place xk in the first column
      Vk(:, 1:nev) = xk(:, 1:nev)
      ! fill W1 = M xk = Ak
      Wk(:, 1:nev) = Ak(:, 1:nev)
      ! fill h1 = V1 W1
      hk(1:nev, 1:nev) = Vk(:, 1:k*nev) .tm.Wk(:, 1:nev)
      ! add V2
      do i = 1, nev
        call gram_schmidt(Vk(:, 1:nev + (i - 1)), tkp(:, i))
        Vk(:, nev + i) = tkp(:, i)
      end do
      restarted = .True.
    end do
    C(:, 1:nev) = xk(:, 1:nev)
    E(1:nev) = muk(1:nev)
  end subroutine get_some_eigenvalues_BlockDavidson

  subroutine get_some_eigenvalues_davpack(n, nev, nb, m, E, C, diag, mm, epsilon, OUTPUT)
    !! get some eigenvalues and eigenvectors using davpack
    integer, intent(in) :: n !! full dimension
    integer, intent(in) :: nev !! number of ev to be found
    integer, intent(in) :: nb !! initial block size
    integer, intent(in) :: m !! maximum dimension <n
    real(long), intent(out) :: E(nev) !! returned eigenvalues
    real(long), intent(inout) :: C(n, nev) !! returned eigenvectors
    real(long), intent(in) :: diag(n) !! diagonal elements of the matrix
    real(long), intent(in) :: epsilon !! convergence criterion
    integer, intent(in), optional :: OUTPUT !! output
    interface
      function mm(n, k, Cin, Cout) result(error)
        integer, intent(in) :: n
        integer, intent(in) :: k
        real(8), intent(in) :: Cin(n, k)
        real(8), intent(out) :: Cout(n, k)
        integer :: error
      end function mm
    end interface
    integer :: itmax, iter, nmult, lwrk, mb, liwrk
    real(long) :: tol, anorm, res(nev), tmp
    real(long), allocatable :: wrk(:)
    integer, allocatable :: iwrk(:)
    integer :: ilevel, ifail, i

    ilevel = 2
    if (nb .gt. nev) then
      stop 'nev must be >= block size '
    end if
    stop "variable block davidson algorithm not implemented"
    !mb = max(nb, nev)
    !lwrk = (n + m)*(mb + m) + (n + 8)*m

    !liwrk = 6*m
    !allocate (wrk(lwrk))
    !allocate (iwrk(liwrk))
    !itmax = 100
    !tol = epsilon
    !anorm = 0.0d0
    !write (6, *) 'n=', n
    !write (6, *) 'm=', m
    !write (6, *) 'nb=', nb
    !write (6, *) 'nev=', nev
    !call davson(n, m, nb, -nev, itmax, iter, nmult, tol, &
    !            anorm, res, E, C, n, wrk, lwrk, iwrk, liwrk, &
    !            mymatv, mycorr, ilevel, OUTPUT, ifail)

    !write (6, *) 'ifail=', ifail
    !write (6, *) 'iter=', iter
    !write (6, *) 'nmult=', nmult
    !write (6, *) 'nev=', nev

    !deallocate (wrk)
    !deallocate (iwrk)

    ! sort eigenvalues and eigenvectors
    i = 1
    do
      if (i .eq. nev) exit
      if (E(i + 1) .ge. E(i)) then
        i = i + 1
        if (i .eq. nev) exit
      else
        ! swap i and i+1
        call dswap(n, C(:, i), 1, C(:, i + 1), 1)
        tmp = E(i + 1)
        E(i + 1) = E(i)
        E(i) = tmp
        ! go back to i-1
        i = max(i - 1, 1)
      end if
    end do

  contains
    subroutine mymatv(m, x, ldx, y, ldy)
      integer :: m, ldx, ldy
      real(long) :: x(ldx, m), y(ldy, m)
      integer :: error

      error = mm(ldx, m, x, y)

    end subroutine mymatv
    subroutine mycorr(m, ritzv, x, ldx, r, ldr)
      integer :: m, ldx, ldr
      real(long) :: x(ldx, m), ritzv(m)
      real(long) :: r(ldr, m)
      integer :: i, j
      real(long) :: tmp

      do j = 1, m
        do i = 1, n
          tmp = diag(i) - ritzv(j)
          if (tmp .ne. 0.0d0) r(i, j) = r(i, j)/tmp
        end do
      end do
    end subroutine mycorr

  end subroutine get_some_eigenvalues_davpack

  subroutine gram_schmidt(v, vi, yes_norm)
    !! Gram Schmidt Ortogonalization
    !! orthonormalize vector vi against orthonormal set v
    real(long), intent(in) :: v(:, :)
    real(long), intent(inout) :: vi(:)
    logical, intent(in), optional :: yes_norm
    logical :: do_norm
    real(long) :: norm
    integer :: j
    do_norm = .True.
    if (present(yes_norm)) do_norm = yes_norm
    norm = sqrt(sum(vi(:)**2))
    vi(:) = vi(:)/norm
    do j = 1, size(v, 2)
      !norm = sqrt(sum(vi(:)**2))
      !vi(:) = vi(:)/norm
      vi(:) = vi(:) - sum(vi(:)*v(:, j))*v(:, j)
    end do
    if (do_norm) then
      norm = sqrt(sum(vi(:)**2))
      vi(:) = vi(:)/norm
    end if
  end subroutine gram_schmidt

#ifdef WITH_ARPACK
  subroutine get_some_eigenvalues_arpack(n, nev, E, C, mv)
    !! Eigenvalues and eigenvectors using arpack
    integer, intent(in) :: n, nev
    real(long), intent(out) :: E(n)
    real(long), intent(inout) :: C(n, nev)
    interface
      function mv(n, Cin, Cout) result(error)
        integer, intent(in) :: n
        real(8), intent(in) :: Cin(n)
        real(8), intent(out) :: Cout(n)
        integer :: error
      end function mv
    end interface

    integer :: ido
    character :: bmat
    character(len=2) :: which
    integer :: ldv
    real(long) :: tol
    real(long) :: resid(n)
    integer :: ncv
    real(long), allocatable :: V(:, :)
    integer :: iparam(11)
    integer :: ipntr(11)
    real(long) :: workd(3*n)
    integer :: lworkl
    real(long), allocatable :: workl(:)
    integer :: info, ierr
    logical :: rvec
    logical, allocatable :: select(:)
    real(long), allocatable :: d(:, :)
    real(long) :: sigma
    integer :: error
    ncv = 10*nev ! must be <=n, recommendation is >= 2*nev.
    ldv = n
    bmat = 'I'
    allocate (V(ldv, ncv))
    lworkl = ncv**2 + 8*ncv
    allocate (workl(lworkl))
    ido = 0
    lworkl = ncv*(ncv + 8)
    which = 'SA'
    tol = 1e-10
    info = 0

    allocate (select(ncv))
    allocate (d(ncv, 2))

    resid = ZERO
    resid(:) = C(:, 1)

    iparam(1) = 1 ! ishift
    iparam(2) = 0 ! levec
    iparam(3) = 500 !mxiter
    iparam(4) = -1 ! nb
    iparam(5) = nev !nconv
    iparam(6) = -1 ! iupd
    iparam(7) = 1 !mode=1 means standard ev problem
    iparam(8) = 0 ! np relevant for iparam1(1)=0
    iparam(9) = 0 !numop (output)
    iparam(10) = 0 ! numopb (output)
    iparam(11) = 0 !numreo (output)

    rvec = .True. ! obtain eigenvalues and eigenvectors
    do
      call dsaupd(ido, bmat, n, which, nev, tol, resid, &
                  ncv, v, ldv, iparam, ipntr, workd, workl, &
                  lworkl, info)
      if (ido .eq. -1 .or. ido .eq. 1) then
        error = mv(n, workd(ipntr(1):ipntr(1) + n - 1), workd(ipntr(2):ipntr(2) + n - 1))
        cycle
      else
        exit
      end if
    end do
    if (info .lt. 0) then
      print *, ' '
      print *, ' Error with _saupd, info = ', info
      print *, ' Check documentation in _saupd '
      print *, ' '
    else
      write (6, '("# ARPACK step it took ",i4," matrix-vector multiplications ")') iparam(9)
      write (6, '("# ARPACK step it took ",i4," re-orthogonalizations ")') iparam(11)
      call dseupd(rvec, 'All', select, d, v, ldv, sigma, &
                  bmat, n, which, nev, tol, resid, ncv, v, ldv, &
                  iparam, ipntr, workd, workl, lworkl, ierr)
      if (ierr .ne. 0) then
        print *, ' '
        print *, ' Error with _seupd, info = ', ierr
        print *, ' Check the documentation of _seupd. '
        print *, ' '
      else
        E(1:nev) = d(1:nev, 1)
        C(:, 1:nev) = V(:, 1:nev)
      end if
    end if

    deallocate (V)
    deallocate (workl)
    deallocate (select)
    deallocate (d)

  end subroutine get_some_eigenvalues_arpack
#endif

#ifdef WITH_KRYLOV
  subroutine get_some_eigenvalues_libKrylov(n, nev, ndim, E, C, diag, mm, epsilon, OUTPUT)
    use iso_fortran_env, only: real32, real64, int32, int64
    !! eigenvalues from libkrylov
    integer, intent(in) :: n !! dimension
    integer, intent(in) :: nev !! number of eigenvalues to obtain = size of the block
    integer, intent(in) :: ndim !! ndim: size of subspace to keep, iterations after which to perform a restart
    real(long), intent(out) :: E(nev) !! returned eigenvalues
    real(long), intent(inout) :: C(n, nev) !! on input: guessed eigenvectors, on output: found eigenvectors
    real(long), intent(in) :: diag(n) !! diagonal matrix elements
    interface
      function mm(n, k, Cin, Cout) result(error)
        !! mm: matrix matrix multiplication subroutine        
        integer, intent(in) :: n
        integer, intent(in) :: k
        real(8), intent(in) :: Cin(n, k)
        real(8), intent(out) :: Cout(n, k)
        integer :: error
      end function mm
    end interface
    real(long), intent(in), optional :: epsilon !! convergence criterion (for the rms of residue)
    integer, intent(in), optional :: OUTPUT !! output
    real(long) :: eps
    integer(int64) :: krylov_initialize, krylov_add_space, krylov_set_real_space_vectors, &
               krylov_solve_real_equation, krylov_get_space_eigenvalues, &
               krylov_get_real_space_solutions, krylov_finalize, krylov_set_space_preconditioner, &
               krylov_set_space_diagonal, krylov_set_integer_option, krylov_get_integer_option, &
               krylov_set_space_orthonormalizer, krylov_get_space_last_residual_norms, &
               krylov_set_real_option, krylov_header, krylov_footer
    real(real64) :: krylov_get_real_option

    integer :: IOUT, i
    integer(int64) :: error, nl, nevl, ndiml, index, maxitl
    real(long) :: last_residual_norms(nev)
    eps = 1e-6_long
    if (present(epsilon)) eps = epsilon
    IOUT = 6
    if (present(OUTPUT)) &
      IOUT = OUTPUT
    nl = n
    nevl = nev
    ndiml = ndim
    maxitl = 100
    error = krylov_header()
    error = krylov_initialize()
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov kryloc_initialize: ",i5)') error
    index = krylov_add_space('r', 's', 'e', nl, nevl, ndiml) ! real, symmetric, eigenproblem
    ! nl = full dimension
    ! nevl = number of eigenvectors to be found
    ! ndiml = number of starting vectors

    if (index .lt. 0) write (IOUT, '("# Error in libkrylov kryloc_add_space: ",i5)') index

    error = krylov_set_integer_option('max_iterations', maxitl)
    error = krylov_set_real_option('min_basis_vector_norm', dble(1e-8))
    error = krylov_set_real_option('max_residual_norm', eps)
    error = krylov_set_real_option('min_gram_rcond', dble(1e-8))

    write (IOUT, '("# libkrylov parameter",x,a25,i6)') 'max_iterations', krylov_get_integer_option('max_iterations')
    write (IOUT, '("# libkrylov parameter",x,a25,es20.3)') 'min_basis_vector_norm', krylov_get_real_option('min_basis_vector_norm')
    write (IOUT, '("# libkrylov parameter",x,a25,es20.3)') 'max_residual_norm', krylov_get_real_option('max_residual_norm')
    write (IOUT, '("# libkrylov parameter",x,a25,es20.3)') 'min_gram_rcond', krylov_get_real_option('min_gram_rcond')

    error = krylov_set_space_preconditioner(index, 'j') !
    ! 'n' null preconditioner
    ! 'c' conjugate gradient (diagonal approximation)
    ! 'd' Davidson (k-th iterate as diagonal)
    ! 'j' Jacobi-Davidson 1
    ! 'a' Jacobi-Davidson 2
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov set_real_preconditioner: ",i5)') error

    ! only if before 'd'/'c' was selected
    error = krylov_set_space_diagonal(index, nl, diag)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov set_space_diagonal: ",i5)') error
    error = krylov_set_space_orthonormalizer(index, 'o')
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov set_space_orthonormalizer: ",i5)') error
    error = krylov_set_real_space_vectors(index, nl, ndiml, C)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov set_real_space: ",i5)') error
    error = krylov_solve_real_equation(index, mm)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov set_solve_real_equation: ",i5)') error
    error = krylov_get_space_eigenvalues(index, nevl, E)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov get_space_eigenvalues: ",i5)') error
    error = krylov_get_real_space_solutions(index, nl, nevl, C)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov get_solution: ",i5)') error

    error = krylov_get_space_last_residual_norms(index, nevl, last_residual_norms)
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov get_residual_norms: ",i5)') error
    write (IOUT, '("# Residual norms:")')
    do i = 1, int(nevl)
      write (IOUT, '("# ",i6,es15.3)') i, last_residual_norms(i)
    end do

    error = krylov_finalize()
    if (error .lt. 0) write (IOUT, '("# Error in libkrylov finalize: ",i5)') error

    error = krylov_footer()
 end subroutine get_some_eigenvalues_libKrylov
#endif

end module eigen_lapack
