!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module finitetemp
  !! author: Ludger Inhester
  !! author: Kota Hanasaki
  !! this module contains all subroutines related to finite temperature calculation
  use constants
  use wavefunction
  use read_param
  implicit none
  private
  public :: fermi_occupation, calculate_entropy, occ_broadening
contains

  subroutine occ_broadening(P, F, temp)
    !! washes out fractional occupation number for orbitals with simialr energy
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    integer :: i, j
    real(long), intent(in) :: temp ! Temperature == sigma width of Gaussians in Hartree
    real(long) :: sigma, norm, norm_old
    logical :: recheck
    real(long) :: rescale, diff, gauss
    real(long) :: f_occ_old(size(F%f_occ, 1))
    real(long) :: f_occ_save(size(F%f_occ, 1))
    sigma = temp
    norm = ONE
    norm_old = ZERO
    rescale = ZERO
    f_occ_old = F%f_occ
    do
      f_occ_save = F%F_occ
      recheck = .FALSE.
      do i = F%Ncore + 1, size(F%n_occ)
        do j = i + 1, size(F%f_occ)
          gauss = exp(-(F%CE(j) - F%CE(i))**2/(TWO*sigma**2))
          !        gauss=exp(-(F%CE(j)-F%CE(i))/(sigma))
          diff = dble(f%F_occ(i) - f%F_occ(j))
          !        write(6,'("# BROAD ",i4,i4,f15.5,f15.5,f15.5)') i,j,ONE-gauss,HALF*abs(diff)
          if (HALF*abs(diff) .gt. ONE - gauss) then
            F%f_occ(i) = f%F_occ(i) - diff*HALF*gauss
            F%f_occ(j) = f%F_occ(j) + diff*HALF*gauss
            recheck = .TRUE.
          end if
        end do
      end do
      if (.NOT. recheck) exit
    end do

    if (P%yes_verbose) then
      do i = F%Ncore + 1, size(F%n_occ)
        write (6, '("# BROAD ",i4,f15.5,f15.5,f15.5)') i, F%F_occ(i), F_occ_old(i), F%CE(i)*AU2EV
      end do

    end if

  end subroutine occ_broadening

  !> this subroutine return calculates the chemical potential
  !! such that we have sum(F%N_occ) electrons at temperature P%T_eV
  !! with orbital energies F%CE
  !! fractional occupations are saved in F%f_occ. by Ludger Inhester
  !
  !! 26/11/2014 Kota Hanasaki : revision 1406 to adapt to low temperature
  !! 01/12/2014 Kota Hanasaki : revision 1413 fix initial guess
  !! Mar 2017 Ludger Inhester : cleaned up code, restricted it to obitals F%Ncore+1, choose better variable names
  subroutine fermi_occupation(P, F, yes_initialize)
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    logical, optional :: yes_initialize    ! Optional argument to control -initialize- option
    logical, save :: doIntlGuess           ! flag to set initial guess

    integer, parameter :: NR_iterMax = 999, N_trial_max = 999     !
    integer, parameter :: Loglvl = -2  ! only <0 : (error) messages are printed out

    integer :: i
    integer ::  iter, iHomoSomo
    real(long) :: Ne, Nrest, T_AU
    real(long) :: mu_l, mu_h ! lower and upper bound for chem. potential
    real(long) :: nd_l, nd_h ! lower and upper bound for paricle number offset
    real(long) :: mu ! chemical potential
    real(long) :: dx, dxold, v, dv ! step in potential, old step in potential, part. number offset, part. number offset derivative
    data doIntlGuess/.true./              !
    real(Long) :: v_improv_small, vOld, xStep
    integer    :: nTry
    !
    ! the chemical potential is adjusted such that we have the targeted particle number
    ! N=sum(F%N_occ)
    ! Newton-Raphson procedure is used to solve
    !
    !  sum_i  2 * f(e_i,mu) - N = 0
    !
    ! where f(e,mu) is the Fermi function with chemical potential mu.
    ! The convergence criteria is taken from the SCF density convergence criteria
    ! abs( (sum_i  2 * f(e_i,mu) - N) / N ) < P%SCF_density_conv

    ! Here I adopt nrsafe ;  Numerical Recipes in C pp366
    ! To use this, we have to set  nd_l : guess of mu that gives < Ne
    !                              nd_h : guess of mu that gives > Ne
    ! if we are sure about mu is already good enough, we start from  Old_mu \pm T
    ! whereas if not, we look for E_{HOMO}...

    if (present(yes_initialize)) doIntlGuess = yes_initialize
    iter = 0
    Ne = sum(F%N_occ(F%Ncore + 1:))
    if (F%F_occ_given) then
      Ne = sum(F%F_occ(F%Ncore + 1:))
    end if

    ! if the number of electrons is smaller then one
    ! return and don't set fractional occupation number
    if (Ne .lt. ONE) then
      return
    end if

    T_AU = P%T_eV*EV2AU
    if (.not. doIntlGuess) then
      mu_l = F%chem_pot - T_AU
      mu_h = F%chem_pot + T_AU
      nd_l = -Ne
      nd_h = -Ne
      do i = F%Ncore + 1, size(F%N_occ)
        nd_l = nd_l + TWO*fermi(F%CE(i), T_AU, mu_l);
        nd_h = nd_h + TWO*fermi(F%CE(i), T_AU, mu_h)
      end do;
      ! when lower and upper bound have the same sign, do initial guess. why?
      if (nd_l*nd_h .gt. ZERO) then;
        doIntlGuess = .true.
      end if
    end if
    if (doIntlGuess) then
      Nrest = Ne;
      iHOMOSOMO = -1;
      ! find homo index
      do i = F%Ncore + 1, F%N_basis;
        if (Nrest .gt. TWO) then
          Nrest = Nrest - TWO
        else
          iHOMOSOMO = i
          exit
        end if
      end do
      ! mu_h = LUMO orbital energy
      if (iHomoSomo .eq. F%N_basis) then
        mu_h = F%CE(iHomoSomo) + T_au
      else
        mu_h = F%CE(iHomoSomo + 1)
      end if
      ! mu_l = HOMO-1 orbital energy
      if (iHomoSomo .eq. 1) then
        mu_l = F%CE(iHomoSomo) - T_au
      else
        mu_l = F%CE(iHomoSomo - 1)
      end if
      nd_l = -Ne
      nd_h = -Ne
      do i = F%Ncore + 1, size(F%N_occ)
        nd_l = nd_l + TWO*fermi(F%CE(i), T_AU, mu_l);
        nd_h = nd_h + TWO*fermi(F%CE(i), T_AU, mu_h)
      end do;
      if (nd_l .gt. ZERO) then
        ! At our lower bound for chem. pot. we have too many electrons
        ! we have to decrease mu_l stepwise, until it nd_l gets smaller than zero
        v_improv_small = max(1.0E-04, 4.0*nd_l/N_trial_max);  !! if improvement is smaller than this, increase trial step
        xStep = max(T_AU, F%CE(iHOMOSOMO) - F%CE(max(1, iHomoSomo - 1)));
        vOLD = nd_l;
        nTry = 0
        do while (nTry .lt. N_trial_max)
          mu_l = mu_l - xStep; vOLD = nd_l;
          nd_l = -Ne
          do i = F%Ncore + 1, F%N_basis;
            nd_l = nd_l + TWO*fermi(F%CE(i), T_AU, mu_l);
          end do;
          if (nd_l .lt. ZERO) exit
          nTry = nTry + 1
          if (dabs(nd_l - vOLD) .lt. v_improv_small) xStep = xStep*2
        end do
      end if

      if (nd_h .lt. 0.0) then
        ! At our upper bound for chem. pot we have too few electrons
        ! we have to increase mu_h stepwise, until nd_h gets larger than zero
        v_improv_small = max(1.0E-04, 4.0*(-nd_h)/N_trial_max);
        xStep = max(T_AU, F%CE(min(F%N_basis, iHomoSomo + 1)) - F%CE(iHOMOSOMO));
        vOLD = nd_h;
        nTry = 0
        do while (nTry .lt. N_trial_max)
          mu_h = mu_h + xStep;
          vOLD = nd_h;
          nd_h = -Ne
          do i = F%Ncore + 1, F%N_basis
            nd_h = nd_h + TWO*fermi(F%CE(i), T_AU, mu_h)
          end do
          if (nd_h .gt. 0.0) exit
          nTry = nTry + 1
          if (dabs(nd_h - vOLD) .lt. v_improv_small) xStep = xStep*2; ! write(*,*)"#fermi_occupation(intz):",nd_h,vOLD,xStep;
        end do
      end if

      if (nd_l*nd_h .gt. ZERO) then;
        write (*, *) "#fermi_occupation:IntlGuess failed..:", nd_l, mu_l, nd_h, mu_h, T_AU, xStep;
        STOP;
      end if

      !      if (0 .lt. Loglvl) Write (*, *) &
      !        "#fermi_occupation:Initial Guess:", iHomoSomo, F%CE(max(1, iHomoSomo - 1 &)), &
      !           F%CE(min(F%N_basis, iHomoSomo + 1)), Half*(mu_l + mu_h)
      doIntlGuess = .false.
    end if! do IntlGuess

    ! --- rtsafe in Numerical Recipes in C ----
    ! take the half between mu_h and mu_l
    mu = Half*(mu_l + mu_h)
    dxold = dabs(mu_h - mu_l)
    dx = dxold
    !! v = TWO* sum( fermi( F%CE(:),T_AU, mu ) ) - Ne
    v = -Ne
    dv = ZERO
    do i = F%Ncore + 1, F%N_basis;
      v = v + TWO*fermi(F%CE(i), T_AU, mu)
      dv = dv + TWO*fermi_mu_deriv(F%CE(i), T_AU, mu)
    end do
    ! Iterations
    do
      iter = iter + 1
      !! Convergence Check...
      if (dabs(v)/Ne .lt. P%SCF_density_conv .and. dabs(dx) .lt. P%SCF_energy_conv) exit
      if (((mu - mu_h)*dv - v)*((mu - mu_l)*dv - v) .gt. ZERO .or. dabs(v + v) .gt. dabs(dxold*dv)) then
        dxold = dx;
        dx = Half*(mu_h - mu_l)
        mu = mu_l + dx;
      else
        dxold = dx;
        dx = v/dv;
        mu = mu - dx;
        !      if( dabs(dx) < EPS ) then;
        !        if( -1 <Loglvl ) Write(*,*)"#fermi_occupation: Fermi NR search stops at relative dev:",dabs(v)/Ne
        !        exit
        !      end if
      end if
      if (iter .gt. NR_iterMax) STOP "#fermi_occupation:NR-iteration at density_matrix_fermi did not reach convergence.."

      v = -Ne;
      dv = ZERO
      do i = F%Ncore + 1, F%N_basis;
        v = v + TWO*fermi(F%CE(i), T_AU, mu)
        dv = dv + TWO*fermi_mu_deriv(F%CE(i), T_AU, mu)
      end do;
      if (iter .gt. NR_iterMax - 5) Write (*, *) "#fermi_occupation:NRsearch:", iter, mu, v, dv
      if (v .lt. ZERO) then;
        mu_l = mu;
      else;
        mu_h = mu;
      end if
    end do
    F%chem_pot = mu

    forall (i=F%Ncore + 1:F%N_basis) F%f_occ(i) = TWO*fermi(F%CE(i), T_AU, F%chem_pot)
    ! Write(*,*)"ChemicalPotential:", F%chem_pot,sum(occ),"/",Ne

  contains

    !> Fermi function
    !! rev. 1406 : avoids exp(+Big)
    pure real(long) function fermi(E, T, mu) result(ret)
      implicit none
      real(long), intent(in) :: E, T, mu
      real(long) :: xi, expo
      xi = (E - mu)/T;
      if (xi .gt. 0.0) then; expo = exp(-xi); ret = expo/(1 + expo)
      else; expo = exp(xi); ret = ONE/(1 + expo); end if
    end function

    pure real(long) function fermi_mu_deriv(E, T, mu) result(ret)
      implicit none
      real(long), intent(in) :: E, T, mu
      real(long) :: xi, expo
      xi = (E - mu)/T;
      if (xi .gt. 0.0) then; expo = exp(-xi); ret = expo/(T*(1 + expo)**2)
      else; expo = exp(xi); ret = expo/(T*(1 + expo)**2); end if
    end function
  end subroutine fermi_occupation

  real(long) function calculate_entropy(f_occ) result(ret)
    !! function returns the electronic entropy - $$ f*ln(f) - (1-f)*ln(1-f) $$
    real(long), intent(in) :: f_occ(:)
    real(long) :: entropy
    integer :: i
    entropy = ZERO
    do i = 1, size(f_occ)
      ! x*log(x) -> -x for small x
      if (HALF*f_occ(i) .lt. 1e-5) then
        entropy = entropy + HALF*f_occ(i) - (ONE - HALF*f_occ(i))*log(ONE - HALF*f_occ(i))
      else if ((ONE - HALF*f_occ(i)) .lt. 1e-5) then
        entropy = entropy - (HALF*f_occ(i))*log(HALF*f_occ(i)) + (ONE - HALF*f_occ(i))
      else
        entropy = entropy - (HALF*f_occ(i))*log(HALF*f_occ(i)) - (ONE - HALF*f_occ(i))*log(ONE - HALF*f_occ(i))
      end if
    end do
    entropy = TWO*entropy ! twice because of two spin populations
    ret = entropy
  end function calculate_entropy

  !> calculate_grandPotential : calculates grand potential (Omega) using
  !!                            -T \sum log(1+e^{-beta(E-mu)}) - meanField + <H_{int}>
  !! here (thermal) expectation value of mean-field interaction is    F%E_J+F%E_xc_int
  !! while exact (thermal) expectation value of interaction energy is F%E_x + F%E_c
  real(long) function calculate_grandPotential(P, F) result(ret)
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    real(long) :: T_AU, xi
    integer    :: i
    ret = ZERO
    T_AU = P%T_eV*EV2AU
    do i = 1, F%N_basis
      xi = (F%CE(i) - F%chem_pot)/T_AU
      !! ln( 1+e^{-xi} ) = ln( e^{-xi} ( 1+e^{xi} ) )
      if (xi .gt. 0.0) then; ret = ret + log(1 + exp(-xi))
      else; ret = ret + log(exp(xi) + 1) - xi; end if
    end do
    ret = -2*ret*T_AU;    !! 2 for spin
    ret = ret - F%E_xc_int - F%E_J + (F%E_x + F%E_c) + P%E_nuc
  end function

end module
