!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!


module fluorescence
  !! author: Ludger Inhester
  !! This module is for the calculation of Fluorescence rates
  !! written by Ludger Inhester in 2014/2015 at DESY CFEL

  use constants
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use operations
  use basis_func
  use process
  use averaging
  use spin_coupling
  use mcscf
  use gaussian
  use population
  use hartreefock
  use basis_func
  use rwfile
  use scf
  use mergesort
  implicit none
  private
  public :: calc_Fluorescence_rates, calc_Fluorescence_rates_CI, calc_Fluorescence_rates_spin, &
  calc_Fluorescence_rates_spin_avg
contains
  subroutine calc_Fluorescence_rates(P, B, F, G, S, fluo_SO)
    !! calculated fluorescence rates between MOs based on independent el. approximation (i.e. based on orbital energies)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Grid3D), intent(in) :: G
    type(Process_data), intent(inout) :: S

    real(long), intent(out), optional :: fluo_SO(P%N_basis, P%N_basis)
    ! transition dipole moment matrix elements
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_bf  ! bf space
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_mo  ! mo space
    integer :: hole, part ! molecular orbitals: hole and particle
    real(long) :: energy ! energy of transition
    real(long) :: TotalRate ! total rate
    real(long) :: HoleRate ! rate per hole
    real(long) :: Rate ! rate per hole and electron
    integer :: pol ! polarization
    character(len=15) :: string
    character(len=6) :: str_hole
    integer :: OUTPUT
    real(long) :: factor
    OUTPUT = P%OUTPUT

    ! initialize optional output array
    if (present(fluo_SO)) fluo_SO(:, :) = ZERO

    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do
    if (P%print_level .gt. 1) then
      write (OUTPUT, "('#'/'# Fluorescence (based on orbital energies):')")
      write (OUTPUT, "('#    I    F   ',a13,2x,a13,8x,3a13)") &
        "x-ray E(eV)", "rate(a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
    end if
    TotalRate = ZERO
    ! transition from MO i to MO j
    do hole = 1, size(F%N_occ)
      HoleRate = ZERO
      if (.NOT. has_hole(P, F, F%N_occ, hole)) cycle
      do part = 1, size(F%N_occ)
        Rate = ZERO
        if (.NOT. is_occupied(P, F, F%N_occ, part)) cycle
        if (is_equivalent(P, F, part, hole)) cycle
        energy = F%CE(part) - F%CE(hole)
        if (energy .le. 0) cycle
        ! Statistical factor: TWO*stat_factor_not_occupied(P,F,F%N_occ,hole) * stat_factor_occupied(P,F,F%N_occ,part)
        factor = stat_factor_not_occupied(P, F, F%N_occ, hole)*TWO*stat_factor_occupied(P, F, F%N_occ, part)
        write (str_hole, "('*',f5.2)") factor
        write (string, "(f13.2)") energy*AU2EV
        ! transition: hole <- part
        Rate = fluorescence_rate(tdipole_mo, hole, part, energy)
        if (present(fluo_SO)) fluo_SO(part, hole) = Rate*factor
        if (P%print_level .gt. 1 ) &
          write (OUTPUT, "('#   ',i2,' - ',i2,' : ',a13,2x,es13.5,a6,2x,3es13.5) ") &
            hole, part, trim(string), rate, str_hole, tdipole_mo(hole, part, :)
        HoleRate = HoleRate + factor*Rate
        ! Use process structure only if no optional output argument provided
        ! (else the python interfaces crashes? CA 27.03.2018)
        if (.NOT. present(fluo_SO)) then
          call add_Process(P, F, S, "fluorescence", factor*Rate, energy, hole, part)
        end if
      end do ! part loop
      TotalRate = TotalRate + HoleRate
    end do ! i loop
    if (P%print_level .gt. 1) then
      if (TotalRate .gt. ZERO) then
        string = trim(lifetime(TotalRate))
        write (OUTPUT, "('# ',12x,a)") repeat("-", 28)
        write (OUTPUT, "('# ',12x,'Total F rate = ',es13.5,' a.u.  ',"// &
               "'( lifetime = ',a,' )')") TotalRate, trim(string)
      else
        write (OUTPUT, "('#   NONE')")
      end if
    end if
  end subroutine calc_Fluorescence_rates

  subroutine calc_Fluorescence_rates_spin(P, B, F, G, S)
    !! calculated fluorescence rates between MOs based on independent el. approximation (i.e. based on orbital energies) but resolves individual spin configurations   
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Grid3D), intent(in) :: G
    type(Process_data), intent(inout) :: S
    ! transition dipole moment matrix elements
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_bf  ! bf space
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_mo  ! mo space
    integer :: hole, part ! molecular orbitals: hole and particle
    real(long) :: energy ! energy of transition
    real(long) :: alpha = ONE/SPEED_C_AU ! fine structure constant
    real(long) :: AvgTotalRate ! total rate
    real(long) :: ChannelRate ! rate per hole and electron
    integer :: pol ! polarization
    integer :: OUTPUT
    real(long) :: tdipole_spin_cfg(3) ! tdipole of spin configurations
    integer, allocatable :: final_occ(:) ! final occupation
    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer, allocatable :: final_csfs(:, :) ! final spin configurations
    integer, allocatable :: final_dets(:, :) ! final determinants
    integer, allocatable :: initial_csfs(:, :) ! initial configurations
    integer, allocatable :: initial_dets(:, :) ! initial determinants
    integer :: initial_multp ! initial multp
    integer :: Nic ! counter for initial configurations
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: only_left(1), only_left_spin(1) ! array for orbital indices for determinant comparison
    integer :: only_right(1), only_right_spin(1) ! array for spin indices for determinant comparison
    integer :: coef_sign ! sign for determinant comparisno
    real(long) :: coef ! coeficient for CSF expansion
    real(long) :: coef_sum
    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: io ! index for initial occupation
    integer :: i, j

    integer, parameter :: MAX_SPIN = 20
    real(long) :: confAvgTotalrate(MAX_SPIN)
    integer :: confpermultp(MAX_SPIN)
    real(long), allocatable :: rateic2fc(:, :) ! rates between initial conf to final conf
    real(long), allocatable :: TotalRate(:) ! rates between initial conf to final conf
    real(long), allocatable :: coef_i(:, :) ! coeficient for CSF expansion
    real(long), allocatable :: coef_f(:, :) ! coeficient for CSF expansion
    real(long), allocatable :: det_olap(:, :) ! coeficient for det olap
    real(long) :: factor
    character(len=15) :: string

    OUTPUT = P%OUTPUT
    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do
    if (P%print_level .gt. 1) then
      write (OUTPUT, "('#'/'# Fluorescence (based on orbital energies):')")
      write (OUTPUT, "('#    I    F   ',a13,2x,a13,8x,3a13)") &
        "x-ray E(eV)", "rate(a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
    end if

    allocate (final_occ(size(F%N_occ)))
    ! in case of MO occupation averaging be render all occupation patterns; otherwise just take F%N_occ
    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if

    Confpermultp(:) = 0
    do io = 1, size(initial_occ, 2)
      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        Confpermultp(initial_multp) = Confpermultp(initial_multp) + size(initial_csfs, 2)
        deallocate (initial_csfs)
      end do
    end do

    ConfAvgTotalRate(:) = ZERO
    AvgTotalRate = ZERO
    Nic = 0
    ! loop over initial occupation
    do io = 1, size(initial_occ, 2)
      if (P%print_level .gt. 1 &
          ) write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)
      initial_Nopen = 0
      do i = 1, size(initial_occ(:, io))
        if (initial_occ(i, io) .eq. 1) initial_Nopen = initial_Nopen + 1
      end do
      ! loop over all possible initial state spin multiplicities
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        ! prepare the representations of csf and det according to genealogical coupling scheme
        ! for the initial state configuration(s)
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        call Nopen2dets(initial_Nopen, initial_multp, initial_dets)
        allocate (coef_i(size(initial_dets, 2), size(initial_csfs, 2)))

        do ic = 1, size(initial_csfs, 2)
          if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
            if (ic .ne. F%csf_num) cycle
          end if
          do id = 1, size(initial_dets, 2)
            coef_i(id, ic) = csfdet(initial_csfs(:, ic), initial_dets(:, id))
          end do
        end do

        if (P%print_level .gt. 1 .or. P%yes_verbose) then
          write (P%OUTPUT, "('# Initial Spin multiplicity ',i3,' has ',i3, ' spin configurations:')") &
            initial_multp, size(initial_csfs, 2)
          do j = 1, size(initial_csfs, 2)
            if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
              if (j .ne. F%csf_num) cycle
            end if

            write (P%OUTPUT, "('# ',i6,x,a )") j, csfstring(initial_occ(:, io), initial_csfs(:, j))
          end do
        end if

        allocate (TotalRate(size(initial_csfs, 2)))
        TotalRate(:) = ZERO

        ! transition from MO i to MO j
        do hole = 1, size(F%N_occ)
          if (initial_occ(hole, io) .eq. 2) cycle
          if (hole .gt. F%Ncore) cycle  ! only fill orbitals that belong to core orbitals (only valence->core or core->core fluorescence)
          do part = 1, size(F%N_occ)
            if (initial_occ(part, io) .eq. 0) cycle
            if (part .eq. hole) cycle
            energy = F%CE(part) - F%CE(hole)
            if (energy .le. ZERO) cycle ! positive energy
            if (sum(tdipole_mo(hole, part, :)**2) .lt. 1e-40_long) cycle ! skip part hole combinations where  <part | r | hole> is really small
            if (P%print_level .gt. 2 .or. P%yes_verbose) write (P%OUTPUT, "('# ',i4,2x,'->',i4)") part, hole
            write (string, "(f13.2)") energy*AU2EV
            ! Final occupation
            final_occ = initial_occ(:, io)
            final_occ(part) = final_occ(part) - 1
            final_occ(hole) = final_occ(hole) + 1
            final_Nopen = 0
            do i = 1, size(final_occ, 1)
              if (final_occ(i) .eq. 1) final_Nopen = final_Nopen + 1
            end do
            final_multp = initial_multp
            if (final_multp - 1 .gt. final_Nopen) cycle
            call Nopen2csfs(final_Nopen, final_multp, final_csfs)
            call Nopen2dets(final_Nopen, final_multp, final_dets)
            allocate (rateic2fc(size(initial_csfs, 2), size(final_csfs, 2)))

            allocate (coef_f(size(final_dets, 2), size(final_csfs, 2)))
            do fc = 1, size(final_csfs, 2)
              do fd = 1, size(final_dets, 2)
                coef_f(fd, fc) = csfdet(final_csfs(:, fc), final_dets(:, fd))
              end do
            end do

            allocate (det_olap(size(initial_dets, 2), size(final_dets, 2)))
            do id = 1, size(initial_dets, 2)
              do fd = 1, size(final_dets, 2)
                call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
                                 initial_occ(:, io), initial_dets(:, id), final_occ, final_dets(:, fd))
                if (coef_sign .ne. 0 .and.&
                 count(only_left(:) .ne. 0) .eq. 1 .and. count(only_right(:) .ne. 0) .eq. 1 .and. &
                 only_left_spin(1) .eq. only_right_spin(1)) then
                  det_olap(id, fd) = coef_sign
                else
                  det_olap(id, fd) = ZERO
                end if
              end do
            end do

            rateic2fc(:, :) = ZERO
            factor = alpha**3*energy**3*FOUR/THREE
            do ic = 1, size(initial_csfs, 2)
              if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                if (ic .ne. F%csf_num) cycle
              end if
              if (P%print_level .gt. 2 .or. P%yes_verbose) &
                write (P%OUTPUT, "('# initial configuration with multiplicity',i3,2x,':',i6,x,a)") &
                initial_multp, ic, trim(csfstring(initial_occ(:, io), initial_csfs(:, ic)))
              do fc = 1, size(final_csfs, 2)
                tdipole_spin_cfg(:) = ZERO
                ! Loop over determinant pairs
                coef_sum = ZERO
                do id = 1, size(initial_dets, 2)
                  do fd = 1, size(final_dets, 2)
                    coef = coef_f(fd, fc)*coef_i(id, ic)*det_olap(id, fd)
                    if (coef .ne. ZERO) then
                      tdipole_spin_cfg(:) = tdipole_spin_cfg(:) + coef*tdipole_mo(hole, part, :)
                      coef_sum = coef_sum + coef
                    end if
                  end do
                end do
                ChannelRate = factor*sum(tdipole_spin_cfg(:)**2)
                if (P%print_level .gt. 2) then
                  write (P%OUTPUT, "('# ',2x,'-> final configuration ',a,1x, 'multiplicity ',i3)") &
                    trim(csfstring(final_occ, final_csfs(:, fc))), final_multp
                  write (P%OUTPUT, "('# ',5x,'channel rate' ,es15.5,2x,3es15.5)") &
                    ChannelRate, tdipole_spin_cfg(:)**2
                end if
                rateic2fc(ic, fc) = ChannelRate
              end do ! fc
            end do ! ic
            deallocate (det_olap)
            if (P%yes_average_rates .or. (P%yes_average_csf_rates .and. F%multp .eq. initial_multp)) then
              call add_Process(P, F, S, "fluorescence", &
                               sum(rateic2fc(:, :))/size(rateic2fc, 1), energy, hole, part, &
                               initial_occ=initial_occ(:, io), initial_multp=initial_multp, &
                               initial_csf_num=1, &
                               final_multp=final_multp, final_csf_num=1)
            else
              ic = F%csf_num
              if (F%multp .eq. initial_multp) then
                do fc = 1, size(final_csfs, 2)
                  if (rateic2fc(ic, fc) .eq. ZERO) cycle
                  call add_Process(P, F, S, "fluorescence", rateic2fc(ic, fc), energy, hole, part, &
                                   initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=ic, &
                                   final_multp=final_multp, final_csf_num=fc)
                end do
              end if
            end if

            do ic = 1, size(initial_csfs, 2)
              if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                if (ic .ne. F%csf_num) cycle
              end if
              TotalRate(ic) = TotalRate(ic) + sum(rateic2fc(ic, :))
            end do

            deallocate (rateic2fc)
            deallocate (final_csfs)
            deallocate (final_dets)
            deallocate (coef_f)
          end do ! part loop
        end do ! hole loop
        if (P%print_level .gt. 1) then
          do ic = 1, size(initial_csfs, 2)
            if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
              if (ic .ne. F%csf_num) cycle
            end if
            write (P%OUTPUT, "('#  Total Fluorescence for multp ',i4,' ic ',i4,' : ',es13.5,' a.u. (lifetime=',a,')')") &
              initial_multp, ic, TotalRate(ic), trim(lifetime(TotalRate(ic)))
          end do
          if (P%yes_average_csf_rates) &
            write (P%OUTPUT, "('#  Total Fluorescence for multp ',i4,' avg_ic ',' : ',es13.5,' a.u. (lifetime=',a,')')") &
            initial_multp, sum(TotalRate(:))/size(TotalRate(:), 1), &
            trim(lifetime(sum(TotalRate(:))/size(TotalRate(:), 1)))
        end if
        AvgTotalRate = AvgTotalRate + initial_multp*sum(TotalRate(:))
        Nic = Nic + initial_multp*size(TotalRate(:), 1)

        deallocate (initial_dets)
        deallocate (initial_csfs)
        deallocate (coef_i)
        deallocate (TotalRate)
      end do ! initial_multp
    end do ! initial_occ
    AvgTotalRate = AvgTotalRate/Nic

    if (P%yes_average_csf_rates) then
      if (P%print_level .gt. 1 .or. P%yes_verbose) then
        write (P%OUTPUT, "('# Average Total Fluorescence rate ',es13.5,' a.u.  (lifetime= ',a,')')") &
          AvgTotalRate, trim(lifetime(AvgTotalRate))
        write (P%OUTPUT, "('# averaged over ',i3,' spin states' )") Nic
      end if
    end if

    deallocate (final_occ)
    deallocate (initial_occ)
  end subroutine calc_Fluorescence_rates_spin

  subroutine calc_Fluorescence_rates_spin_avg(P, B, F, G, S)
    !! calculated fluorescence rates between MOs based on independent el. approximation (i.e. based on orbital energies) but resolves individual spin configurations and considers averaging over configurations
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Grid3D), intent(in) :: G
    type(Process_data), intent(inout) :: S
    ! to be done

    ! transition dipole moment matrix elements
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_bf  ! bf space
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_mo  ! mo space
    integer :: hole, part ! molecular orbitals: hole and particle
    real(long) :: energy ! energy of transition
    integer :: pol ! polarization
    integer :: OUTPUT

    real(long) :: AvgTotalRate, ChannelRate, Rate, TotalRate, SpinavgRate ! Rates & Amplitudes
    integer :: Spinavg_n, Total_n, AvgTotal_n

    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer :: initial_multp ! initial multp
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: parti_Nopen ! open shells in remaining electron configuration
    integer :: partf_Nopen ! open shells in remaining electron configuration
    integer :: rest_multp ! multipicity in remaining electron configuration
    integer :: rest_Nopen ! open shells in remaining electron configuration
    integer :: rest_Ncsf, initial_Ncsf ! number of spin configuration state functions
    integer :: final_two_S, initial_two_S, initial_two_M, rest_two_S ! two times final and intial spin quantum numbers
    integer :: io ! index for initial occupation
    real(long) :: spin_factor
    integer, allocatable :: weight_io(:)

    OUTPUT = P%OUTPUT
    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do

    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if

    allocate (weight_io(size(initial_occ, 2)))
    weight_io(:) = 0

    AvgTotalRate = ZERO
    AvgTotal_n = 0
    do io = 1, size(initial_occ, 2)
      TotalRate = ZERO
      Total_n = 0
      if (P%print_level .gt. 1) &
      write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)

      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        weight_io(io) = weight_io(io) + initial_multp*num_csf(initial_Nopen, initial_multp)
      end do

      if (P%print_level .gt. 1) &
        write (OUTPUT, "('#   ',a2,' - ',a2,' : ',a13,2x,a13,2x,a10,a10,a10)") &
        'h', 'p', 'X-ray E/eV', 'rate', 'spin_multp', 'factor', 'n_csf'

      ! Loop over hole and valence pair combinations
      do hole = 1, size(initial_occ, 1)
        if (initial_occ(hole, io) .eq. 2) cycle
        do part = 1, size(F%N_occ)
          if (is_equivalent(P, F, part, hole)) cycle
          if (initial_occ(part, io) .eq. 0) cycle
          energy = F%CE(part) - F%CE(hole)
          if (energy .le. 0) cycle

          parti_Nopen = count(initial_occ((/hole, part/), io) .eq. 1)
          partf_Nopen = count(initial_occ((/hole, part/), io) + (/1, -1/) .eq. 1)
          rest_Nopen = initial_Nopen - parti_Nopen
          final_Nopen = rest_Nopen + partf_Nopen

          SpinavgRate = ZERO
          Spinavg_n = 0
          do initial_multp = initial_Nopen + 1, 1, -2
            if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
            ChannelRate = ZERO
            initial_Ncsf = num_csf(initial_Nopen, initial_multp)


            final_multp = initial_multp
            ! We calculate transition rate to high spin state
            ! Apply Wigner Eckert to get factor for transition rates to other MS configurations:
            initial_two_S = initial_multp - 1
            initial_two_M = initial_two_S
            final_two_S = final_multp - 1

            spin_factor = ZERO
            ! we split up the electrons into participator + rest:
            ! loop over multiplicity for the rest
            do rest_multp = rest_Nopen + 1, 1, -2
              rest_two_S = rest_multp - 1
              rest_Ncsf = num_csf(rest_Nopen, rest_multp)
              if (rest_Ncsf .eq. 0) cycle

              if (initial_occ(hole, io) .eq. 1 .and. initial_occ(part, io) .eq.2 .and. &
                rest_two_S + 1 .ge. initial_two_S .and. abs(rest_two_S - 1) .le. initial_two_S) then
                spin_factor = spin_factor + ONE*dble(rest_Ncsf)
              else if (initial_occ(hole, io) .eq. 0 .and. initial_occ(part, io) .eq. 2 .and. rest_two_S .eq. initial_two_S) then
                spin_factor = spin_factor + TWO*dble(rest_Ncsf)
              else if (initial_occ(hole, io) .eq. 1 .and. initial_occ(part, io) .eq. 1 .and. rest_two_S .eq. initial_two_S) then
                spin_factor = spin_factor + TWO*dble(rest_Ncsf)
              else if (initial_occ(hole, io) .eq. 0 .and. initial_occ(part, io) .eq. 1 .and. rest_two_S + 1 .ge. initial_two_S &
                .and. abs(rest_two_S - 1) .le. initial_two_S) then
                spin_factor = spin_factor + ONE*dble(rest_Ncsf)
              else
                spin_factor = spin_factor + ZERO
              end if
            end do! rest_multp
            spin_factor = spin_factor/dble(initial_Ncsf)

            ! transition: hole <- part
            Rate = spin_factor*fluorescence_rate(tdipole_mo, hole, part, energy)
            ChannelRate = Rate

            if (P%print_level .gt. 1 ) &
              write (OUTPUT, "('#   ',i2,' - ',i2,' : ',f13.2,2x,es13.5,2x,i10,es13.5,i4)") &
              hole, part, energy*AU2EV, rate, initial_multp, spin_factor, initial_Ncsf
            if (P%yes_all_multps .eqv. .FALSE.) &
              TotalRate = TotalRate + Rate
            call add_Process(P, F, S, "fluorescence", Rate, energy, hole, part, 0, &
                             initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=1, &
                             final_multp=final_multp, final_csf_num=1)
            if (P%yes_all_multps) then
              SpinavgRate = SpinavgRate + initial_multp*initial_Ncsf*ChannelRate
              Spinavg_n = Spinavg_n + initial_multp*initial_Ncsf
            end if
          end do! initial_multp
          SpinavgRate = SpinavgRate/Spinavg_n
          TotalRate = TotalRate + SpinavgRate
          if (P%print_level .gt. 1 ) write (OUTPUT, "('#   ',i2,' - ',i2,' : ',f13.2,2x,es13.5,2x,a10) ") &
          hole, part, energy*AU2EV, SpinavgRate, 'spin avg'

        end do! part
      end do! hole
      AvgTotalRate = AvgTotalRate + TotalRate*weight_io(io)
      if (P%print_level .gt. 1) then
        if (P%yes_all_multps) then
          write (P%OUTPUT, "('#  Spin Averaged Total Fluorescence rate: ',es13.5,' a.u. (lifetime=',a,') ','weight=',i4)") &
            TotalRate, trim(lifetime(TotalRate)), weight_io(io)
        else
          write (P%OUTPUT, "('#  Total Fluorescence rate: ',es13.5,' a.u. (lifetime=',a,')')") &
            TotalRate, trim(lifetime(TotalRate))
        end if
      end if
    end do! io

    AvgTotalRate = AvgTotalRate/sum(weight_io(:))
    if (P%print_level .gt. 1) &
      write (P%OUTPUT, "('# Average Total Fluorescence rate: ',es13.5,' a.u. (lifetime=',a,') ')") &
      AvgTotalRate, trim(lifetime(AvgTotalRate))

    deallocate (weight_io)
    deallocate (initial_occ)

  end subroutine calc_Fluorescence_rates_spin_avg

  subroutine calc_Fluorescence_rates_CI(P, A, B, F, G)
    !! calculated fluorescence rates based on configuration interaction calculation
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    type(Grid3D), intent(in) :: G
    ! transition dipole moment matrix elements
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_bf  ! bf space
    real(long), dimension(P%N_basis, P%N_basis, 3) :: tdipole_mo  ! mo space
    real(long) :: alpha = ONE/SPEED_C_AU ! fine structure constant
    integer :: pol ! polarization
    integer :: OUTPUT
    integer, allocatable, target :: final_occ(:, :) ! final occupation
    integer, allocatable, target :: initial_occ(:, :) ! initial occupation
    integer, allocatable, target :: initial_occ_orig(:, :) ! initial occupation, no symmetry restrictions
    integer, allocatable :: occs_temp(:, :) ! initial occupation
    integer :: initial_multp ! initial multp
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: only_left(1), only_left_spin(1) ! array for orbital indices for determinant comparison
    integer :: only_right(1), only_right_spin(1) ! array for spin indices for determinant comparison
    integer :: coef_sign ! sign for determinant comparisno
    real(long) :: coef ! coeficient for CSF expansion
    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: N_initial_csfs ! N of initial csfs
    integer :: N_final_occ ! N of final occ
    integer :: N_final_csfs ! N of final csfs
    integer :: i_irrep, irrep_occ, f_irrep, n_irrep, i_final_occ
    integer :: i, j, k, io2
    integer :: val1, val2, virt, hole, part
    real(long) :: E_ini ! initial configuration energy
    logical :: redundant_occ
    type(CSF), allocatable :: initial_csf_array(:) ! array for csfs initial state
    type(CSF), allocatable :: final_csf_array(:) ! array for csfs final state
    real(long), allocatable :: final_csf_tdipole(:, :) ! tdipole between initial conf to final conf
    real(long), allocatable :: final_state_tdipole(:, :) ! tdipole between initial conf to final state
    real(long), allocatable :: csf_tdipole(:, :, :) ! tdipole amplitude for each initial - final csf  pair
    integer, allocatable :: flag_final_csfs(:) ! final csf flag
    real(long), allocatable :: final_CI_matrix(:, :) ! CI matrix for final states
    real(long), allocatable :: csf_c2(:, :)
    real(long), allocatable :: CI_diagonal(:) ! CI matrix diagonal
    real(long), allocatable :: final_CI_energy(:) ! energy of final states
    real(long), allocatable :: initial_CI_matrix(:, :) ! CI matrix for initial states
    real(long), allocatable :: initial_CI_energy(:) ! energy of initial states
    real(long), allocatable :: tdens(:, :)

    real(long), allocatable :: TotalRate(:) ! total rates between initial conf to final conf
    real(long) :: factor
    real(long) :: F_energy
    integer :: Ncore

    type(WF_data) :: F_final
    real(long), allocatable :: csf_O(:, :) ! Fluorescence amplitude for each csf and continuum channel
    real(long), allocatable :: temp(:, :)
    real(long) :: tdipole(3)

    integer, allocatable :: active(:)
    integer, allocatable :: core(:)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    integer :: NOpenMax
    type(UmatrixTable) :: UTini, UTfin
    integer :: nintern
    integer :: nev
    integer, allocatable :: idx(:)

    OUTPUT = P%OUTPUT

    Ncore = F%Ncore
    ! in case of MO occupation averaging we render all occupation patterns; otherwise just take F%N_occ
    i_irrep = P%state_irrep

    if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, initial_occ)
    else if (allocated(P%active_space_MO)) then
      call active_space_occs(F%N_occ, P%active_space_MO, initial_occ)
    else if (P%yes_process_cis_ini) then
      allocate (initial_occ(size(F%N_occ), count(F%N_occ(F%Ncore + 1:) .gt. 0)*count(F%N_occ(F%Ncore + 1:) .lt. 2) + 1))
      initial_occ(:, 1) = F%N_occ(:)
      i = 1
      do hole = F%Ncore + 1, size(F%N_occ, 1)
        if (F%N_occ(hole) .eq. 0) cycle
        do part = F%Ncore + 1, size(F%N_occ, 1)
          if (F%N_occ(part) .eq. 2) cycle
          initial_occ(:, i + 1) = F%N_occ(:)
          initial_occ(hole, i + 1) = initial_occ(hole, i + 1) - 1
          initial_occ(part, i + 1) = initial_occ(part, i + 1) + 1
          i = i + 1
        end do
      end do
    else
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    end if

    ! initial_occ_orig = initial_occ
    allocate (initial_occ_orig, source=initial_occ)
    ! filter symmetry
    if (P%yes_sym .and. F%state_irrep .gt. 0) then
      deallocate (initial_occ)
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, initial_occ_orig, initial_occ)
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Number of configurations before symmetry:',i6)") size(initial_occ_orig, 2)
        write (P%OUTPUT, "('# Number of configurations after symmetry :',i6)") size(initial_occ, 2)
      end if
    end if

    initial_multp = P%multp
    write (P%OUTPUT, "('# Initial configurations for spin multp: ',i4)") initial_multp
    call construct_csf_array(P, B, F, initial_multp, i_irrep, initial_occ, initial_csf_array)
    N_initial_csfs = size(initial_csf_array, 1)
    write (P%OUTPUT, "('# N_initial_csfs: ',i7)") N_initial_csfs
    if(P%state > N_initial_csfs) stop '# state larger than number of configurations'
    allocate (initial_CI_matrix(N_initial_csfs, N_initial_csfs))
    allocate (initial_CI_energy(N_initial_csfs))
    write (P%OUTPUT, '("# computing CI matrix")')
    NOpenMax = maxval(count(initial_occ .eq. 1, dim=1))
    call constructUMatrixTable(NOpenMax, initial_multp, UTini)

    allocate (active(count((/((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0)), j=1, size(initial_occ, 1))/))))
    active = pack((/(j, j=1, size(initial_occ, 1))/), &
                  (/(((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0))), j=1, size(initial_occ, 1))/))
    allocate (core(count((/((all(initial_occ(j, :) .eq. 2)), j=1, size(initial_occ, 1))/))))
    core = pack((/(j, j=1, size(initial_occ, 1))/), &
                (/(all(initial_occ(j, :) .eq. 2), j=1, size(initial_occ, 1))/))
    nintern = size(active)
    do j = size(active), 1, -1
      if (any(sum(initial_occ(active(j:), :), dim=1) .gt. 1)) exit
    end do
    nintern = j
    write (P%OUTPUT, '("#     core:",99i4)') core
    write (P%OUTPUT, '("#   active:",99i4)') active
    write (P%OUTPUT, '("# internal:",99i4)') active(:nintern)
    write (P%OUTPUT, '("# external:",99i4)') active(nintern + 1:)

    ! calculate transition dipole moments in bf space
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    ! transform transition dipole moments into mo space
    do pol = 1, 3 ! loop over polarizations
      tdipole_mo(:, :, pol) = F%LC(:, :) .tm.tdipole_bf(:, :, pol) .m.F%LC(:, :)
    end do

    allocate (F_core(size(F%LC, 2), size(F%LC, 2)))
    call construct_h0_matrix(G, B, F)
    F%H0_MO = F%LC(:, :) .tm. B%H0(:, :) .m. F%LC(:, :)
    call calculate_F_inactive(P, B, core, F%LC, F_core)
    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do
    if (nintern .gt. 0) call transform_twoel_allactive(P, B, F, active(:nintern), P%OUTPUT)
    if (nintern .lt. size(active)) then
      if (nintern .gt. 0) then
        call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
        call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      end if
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)
      if (nintern .gt. 0) then
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        deallocate (F%teint_cdei)
      end if
    else
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix)
    end if
    if (nintern .gt. 0) deallocate (F%teint_abcd)
    write (P%OUTPUT, '("# done")')
    ! solve it
    write (P%OUTPUT, '("# solving CI matrix")')
    call get_eigenvalues(initial_CI_matrix, initial_CI_energy)
    write (P%OUTPUT, '("# done")')

    E_ini = initial_CI_energy(P%state)


    write (P%OUTPUT, "('# Initial state ',i4,x,'spin multp=',i4,x,'energy=',f15.5)") &
      P%state, P%multp, E_ini + nuclear_repulsion_energy(A, P)
    do ic = 1, size(initial_CI_matrix, 1)
      if (abs(initial_CI_matrix(ic, P%state)) .lt. 0.1) cycle
      write (P%OUTPUT, '("# csf ",i7,x,f15.5,x,a)') &
        ic, initial_CI_matrix(ic, P%state), trim(csfstring(initial_csf_array(ic)%occ, initial_csf_array(ic)%csf))
    end do

    if (P%print_level .gt. 1) then
      write (OUTPUT, "('#'/'# Fluorescence (based on CI states):')")
      write (OUTPUT, "('#    I    F   ',a13,2x,a13,8x,3a13)") &
        "x-ray E(eV)", "rate(a.u.)", "tdip.(a.u)_x", "tdip.(a.u)_y", "tdip.(a.u)_z"
    end if
    final_multp = initial_multp
    if (P%yes_sym) then
      n_irrep = B%nirrep
    else
      n_irrep = 1
    end if
    allocate (Totalrate(n_irrep))
    Totalrate(:) = ZERO

    do f_irrep = 1, n_irrep
      write (P%OUTPUT, "('# Final state irrep ',i3)") f_irrep
      N_final_occ = 0
      ! Loop over hole and valence pair combinations: count configurations
      do io2 = 1, size(initial_occ_orig, 2)
        !if(flag_initial_occ(io2)>1)cycle
        do hole = 1, Ncore
          if (initial_occ_orig(hole, io2) .eq. 2) cycle
          do val1 = Ncore + 1, size(initial_occ_orig(:, io2))
            if (initial_occ_orig(val1, io2) .eq. 0) cycle
            if (val1 .eq. hole) cycle
            N_final_occ = N_final_occ + 1
            if (P%yes_process_CIS) then
              do val2 = val1, size(initial_occ_orig(:, io2))
                if (val2 .eq. val1 .and. initial_occ_orig(val2, io2) .lt. 2) cycle
                if (initial_occ_orig(val2, io2) .eq. 0) cycle
                do virt = val2 + 1, size(initial_occ_orig(:, io2))
                  if (initial_occ_orig(virt, io2) .eq. 2) cycle
                  N_final_occ = N_final_occ + 1
                end do
              end do
            end if! process_CIs
          end do! val1
        end do! hole
      end do
      allocate (final_occ(size(F%N_occ), N_final_occ))

      write (P%OUTPUT, "('# N_final_occ ',i5)") N_final_occ
      N_final_csfs = 0
      ! write final_occ and count csfs
      i_final_occ = 0
      do io2 = 1, size(initial_occ_orig, 2)
        !if(flag_initial_occ(io2)>1)cycle
        do hole = 1, Ncore
          if (initial_occ_orig(hole, io2) .eq. 2) cycle
          do val1 = Ncore + 1, size(initial_occ_orig(:, io2))
            if (initial_occ_orig(val1, io2) .eq. 0) cycle
            if (val1 .eq. hole) cycle
            i_final_occ = i_final_occ + 1
            ! Final occupation
            final_occ(:, i_final_occ) = initial_occ_orig(:, io2)
            final_occ(val1, i_final_occ) = final_occ(val1, i_final_occ) - 1
            final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) + 1
            redundant_occ = .FALSE.
            do j = 1, i_final_occ - 1
              if (all(final_occ(:, j) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
              if (redundant_occ) exit
            end do
            if (redundant_occ) then
              i_final_occ = i_final_occ - 1
              cycle
            end if
            final_Nopen = count(final_occ(:, i_final_occ) .eq. 1)
            if (final_multp - 1 .gt. final_Nopen) cycle
            if (P%yes_sym) then
              irrep_occ = 1
              do i = 1, size(final_occ(:, i_final_occ))
                if (final_occ(i, i_final_occ) .eq. &
                    1) irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
              end do
              if (irrep_occ .ne. f_irrep) then
                i_final_occ = i_final_occ - 1
                cycle
              end if
            end if
            if (P%yes_process_CIS) then
              do val2 = val1, size(initial_occ_orig(:, io2))
                if (val2 .eq. val1 .and. initial_occ_orig(val2, io2) .lt. 2) cycle
                if (initial_occ_orig(val2, io2) .eq. 0) cycle
                do virt = val2 + 1, size(initial_occ_orig(:, io2))
                  if (initial_occ_orig(virt, io2) .eq. 2) cycle

                  i_final_occ = i_final_occ + 1
                  ! Final occupation
                  final_occ(:, i_final_occ) = initial_occ_orig(:, io2)
                  final_occ(val1, i_final_occ) = final_occ(val1, i_final_occ) - 1
                  final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) + 1
                  final_occ(val2, i_final_occ) = final_occ(val2, i_final_occ) - 1
                  final_occ(virt, i_final_occ) = final_occ(virt, i_final_occ) + 1
                  redundant_occ = .FALSE.
                  do j = 1, i_final_occ - 1
                    if (all(final_occ(:, j) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
                    if (redundant_occ) exit
                  end do
                  if (redundant_occ) then
                    i_final_occ = i_final_occ - 1
                    cycle
                  end if
                  final_Nopen = count(final_occ(:, i_final_occ) .eq. 1)
                  if (final_multp - 1 .gt. final_Nopen) cycle
                  if (P%yes_sym) then
                    irrep_occ = 1
                    do i = 1, size(final_occ(:, i_final_occ))
                      if (final_occ(i, i_final_occ) .eq. 1) &
                        irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
                    end do
                    if (irrep_occ .ne. f_irrep) then
                      i_final_occ = i_final_occ - 1
                      cycle
                    end if
                  end if
                end do! virt
              end do! val2
            end if! process_CIs

          end do! val1
        end do! hole
      end do! io2

      write (P%OUTPUT, "('# redundant/excluded because of symmetry occs ',i7)") N_final_occ - i_final_occ
      N_final_occ = i_final_occ
      write (P%OUTPUT, "('# N_final_occ ',i5)") N_final_occ

      allocate (occs_temp(size(final_occ, 1), N_final_occ))
      occs_temp = final_occ(:, 1:N_final_occ)
      deallocate (final_occ)
      allocate (final_occ(size(occs_temp, 1), size(occs_temp, 2)))
      final_occ = occs_temp
      deallocate (occs_temp)

      call construct_csf_array(P, B, F, final_multp, f_irrep, final_occ(:, :), final_csf_array)
      N_final_csfs = size(final_csf_array, 1)
      if (N_final_csfs .eq. 0) then
        ! if there are no spin configurations for this irrep continue with next irrep
        deallocate (final_occ)
        cycle
      end if
      allocate (flag_final_csfs(N_final_csfs))
      NOpenMax = maxval(count(final_occ .eq. 1, dim=1))
      call constructUMatrixTable(NOpenMax, final_multp, UTfin)

      ! fill flag_final_csfs to prescreen configurations that are different by 4 occupations to any initial configuration
      flag_final_csfs(:) = 1
      fc = 1
      do
        k = num_csf(count(final_csf_array(fc)%occ(:) .eq. 1), final_multp)
        do ic = 1, size(initial_csf_array, 1)
          if (sum(abs(initial_csf_array(ic)%occ(:) - final_csf_array(fc)%occ(:))) .le. 2) then
            flag_final_csfs(fc:fc + k - 1) = 0
            cycle
          end if
          if (flag_final_csfs(fc) .eq. 0) exit
        end do
        fc = fc + k
        if (fc .gt. N_final_csfs) exit
      end do!
      write (P%OUTPUT, "('# primary   csf ',i12)") count(flag_final_csfs(:N_final_csfs) .eq. 0)
      write (P%OUTPUT, "('# secondary csf ',i12)") count(flag_final_csfs(:N_final_csfs) .eq. 1)
      ! number of states to consider
      if (P%nstates .gt. 0) then
        nev = P%nstates
      else
        nev = count(flag_final_csfs(:N_final_csfs) .eq. 0)
      end if
      nev = max(nev, 1)
      nev = min(nev, N_final_csfs)

      if (N_final_csfs .gt. 0) then

        allocate (final_csf_tdipole(N_final_csfs, 3))
        allocate (csf_tdipole(N_initial_csfs, N_final_csfs, 3))
        allocate (final_state_tdipole(nev, 3))

      ! Loop over final configuration state functions
      ! Calculate for every configuration the transtion dipole matrix element
        csf_tdipole(:, :, :) = ZERO
        do fc = 1, size(final_csf_array, 1)
          do ic = 1, size(initial_csf_array, 1)
            csf_tdipole(ic, fc, :) = ZERO
            allocate (tdens(size(initial_occ, 1), size(final_occ, 1)))
            call trans_dens_csf(initial_csf_array(ic), final_csf_array(fc), tdens)
            do i = 1, size(tdens, 1)
              do j = 1, size(tdens, 2)
                csf_tdipole(ic, fc, :) = csf_tdipole(ic, fc, :) + &
                  tdens(i, j)*tdipole_mo(i, j, :)
              end do! j
            end do! i
            deallocate (tdens)
          end do! ic
        end do! fc

        if (P%yes_process_CI_relaxed) then
          if (P%process_CI_relaxed_variant .ne. 2 .and. P%process_CI_relaxed_variant .ne. 1) then
            write (P%OUTPUT, *) 'process_CI_relaxed_variant', P%process_CI_relaxed_variant
            stop 'process_CI_relaxed_variant must be 1 or 2'
          end if
          ! obtain ionized gs orbitals
          F_final = F
          call read_wf(F, F_final%LC, P%process_relaxed_wf_filename)
          if (P%process_CI_relaxed_variant .eq. 2) then
            !! print the configurations in the initial state basis (without projection)
            do pol = 1, 3
              do j = 1, N_final_csfs
                final_csf_tdipole(j, pol) = sum(initial_CI_matrix(:, P%state) * csf_tdipole(:, j, pol))
              end do ! j
            end do ! pol
            do fc = 1, N_final_csfs
              write (P%OUTPUT, '("# csf ",i7,x,a,x,"tdipole=",3es20.7)') &
                fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), final_csf_tdipole(fc, :)
            end do
            !!
            allocate (csf_O(size(final_csf_array), size(final_csf_array)))
            call overlap_csfs(P, final_csf_array, F%LC, F_final%LC, B%S, csf_O)
            allocate (temp(N_initial_csfs, N_final_csfs))
            do pol = 1, 3
              do i = 1, N_initial_csfs
                do j = 1, N_final_csfs
                  temp(i, j) = sum(csf_tdipole(i, :, pol) * csf_O(:, j))
                end do
              end do
              csf_tdipole(:, :, pol) = temp
              do j = 1, N_final_csfs
                final_csf_tdipole(j, pol) = sum(initial_CI_matrix(:, P%state) * csf_tdipole(:, j, pol))
              end do! j
            end do! pol
            deallocate (temp)
            write (P%OUTPUT, '(a)') '# Overlap to new basis:'
            do fc = 1, size(final_csf_array, 1)
              write (P%OUTPUT, '("# csf ",i7,x,a,x,"tdipole=",3es20.7,x,f7.3,f7.3)') &
                fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), &
                final_csf_tdipole(fc, :), &
                sum(csf_O(:, fc)**2), sum(csf_O(fc, :)**2)
            end do
            deallocate (csf_O)
          else if (P%process_CI_relaxed_variant .eq. 1) then
            csf_tdipole = ZERO
            call overlap_operator_csfs(P, initial_csf_array, &
            final_csf_array, F%LC, F_final%LC, B%S, tdipole_bf(:, :, :), &
            csf_tdipole(:, :, :))
            do pol = 1, 3
              do j = 1, N_final_csfs
                final_csf_tdipole(j, pol) = sum(initial_CI_matrix(:, P%state) * csf_tdipole(:, j, pol))
              end do! j
            end do! pol
            do fc = 1, N_final_csfs
              write (P%OUTPUT, '("# csf ",i7,x,a,x,"tdipole=",3es20.7)') &
                fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), final_csf_tdipole(fc, :)
            end do
          end if
          call construct_h0_matrix(G, B, F_final)
          F_final%H0 = F_final%LC(:, :) .tm.F_final%H0(:, :) .m.F_final%LC(:, :)
        else
          do pol = 1, 3
            do j = 1, N_final_csfs
              final_csf_tdipole(j, pol) = sum(initial_CI_matrix(:, P%state) * csf_tdipole(:, j, pol))
            end do ! j
          end do ! pol
          do fc = 1, N_final_csfs
            write (P%OUTPUT, '("# csf ",i7,x,a,x,"tdipole=",3es20.7)') &
              fc, trim(csfstring(final_csf_array(fc)%occ, final_csf_array(fc)%csf)), final_csf_tdipole(fc, :)
          end do
          F_final = F
          call construct_h0_matrix(G, B, F_final)
          F_final%H0 = F_final%LC(:, :) .tm.F_final%H0(:, :) .m.F_final%LC(:, :)
        end if! P%yes_process_CI_relaxed
        deallocate(csf_tdipole)

        allocate (final_CI_energy(N_final_csfs))

        deallocate (active, core)
        allocate (active(count((/ &
                               ((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0) &
                                 ), j=1, size(final_occ, 1))/))))
        allocate (core(count((/((all(final_occ(j, 1:N_final_occ) .eq. 2)), j=1, size(final_occ, 1))/))))
        active = pack((/(j, j=1, size(final_occ, 1))/), &
                      (/(((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0))), &
                         j=1, size(final_occ, 1))/))
        core = pack((/(j, j=1, size(final_occ, 1))/), &
                    (/(all(final_occ(j, 1:N_final_occ) .eq. 2), j=1, size(final_occ, 1))/))
        nintern = size(active)
        do j = size(active), 1, -1
          if (any(sum(final_occ(active(j:), :N_final_occ), dim=1) .gt. 1)) exit
        end do
        nintern = j
        write (P%OUTPUT, '("#     core:",99i4)') core
        write (P%OUTPUT, '("#   active:",99i4)') active
        write (P%OUTPUT, '("# internal:",99i4)') active(:nintern)
        write (P%OUTPUT, '("# external:",99i4)') active(nintern + 1:)
        call calculate_F_inactive(P, B, core, F_final%LC, F_core)
        E_core = ZERO
        do i = 1, size(core)
          E_core = E_core + F_core(core(i), core(i)) + F_final%H0(core(i), core(i))
        end do

        call transform_twoel_allactive(P, B, F_final, active(1:nintern), P%OUTPUT)
        if (nintern .lt. size(active) .and. nintern .gt. 0) then
          call transform_twoel_two_core_two_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
          call transform_twoel_three_core_one_virtual(P, B, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
        end if  

        if (P%yes_use_arpack .or. P%yes_use_davidson .or. P%yes_use_krylov) then
          allocate (CI_diagonal(N_final_csfs))
          write (P%OUTPUT, '("# calculating CI diagonal entries")')
          call calc_CIMatrix_diagonal(UTfin, E_core, F_core(active(:), active(:)), &
            F_final%teint_abcd, &
            final_occ(active(:), 1:N_final_occ), &
            CI_diagonal, nintern=nintern, &
            teint_iiee=F_final%teint_cdij, &
            teint_ieie=F_final%teint_cidj, &
            teint_iiie=F_final%teint_cdei)
          write (P%OUTPUT, '("# done")')
          nev = max(P%nstates, 1)
          allocate (csf_c2(N_final_csfs, nev))
          csf_c2 = ZERO
          ! take the configuration with lowest diagonal element from those entries that are in the primary space
          allocate (idx(count((flag_final_csfs(:) .eq. 0))))
          j = 0
          do i = 1, min(size(idx), nev)
            j = j + 1
            csf_c2(i, j) = ONE
          end do
          deallocate (idx)
          if (nev .gt. j) then
            write (P%OUTPUT, '("# sort CI_diagonal")')
            allocate (idx(size(CI_diagonal)))
            call merge_argsort(CI_diagonal, idx)
            write (P%OUTPUT, '("# done")')
            k = 0
            do while (j .lt. nev .and. k .lt. size(idx))
              k = k + 1
              if (flag_final_csfs(idx(k)) .eq. 0) cycle
              j = j + 1
              csf_c2(idx(k), j) = ONE
            end do
            deallocate (idx)
          end if
          ! print summary
          do i = 1, nev
            do j = 1, size(csf_c2, 1)
              if (csf_c2(j, i) .eq. ZERO) cycle
              write (P%OUTPUT, '("# Starting guess: ",i4,x,es20.7)', advance="yes") i, ci_diagonal(j)
              write (P%OUTPUT, '("# ",es20.7,x,a)', advance="yes") &
                csf_c2(j, i), trim(csfstring(final_csf_array(j)%occ, final_csf_array(j)%csf(:)))
            end do
          end do
          if (P%yes_use_arpack) then
            ! call arpack
#ifdef WITH_ARPACK
            call get_some_eigenvalues_arpack(N_final_csfs, nev, final_CI_energy(1:nev), csf_c2, matrix_vector_multp)
#else
            stop "not compiled with arpack support"
#endif
          else if (P%yes_use_davidson) then
            ! call Davidson
            ! call get_some_eigenvalues_VariableBlockDavidson(N_final_csfs, nev, P%davblock, max(P%davMax,nev), final_CI_energy(1:nev), csf_c2,  ci_diagonal, matrix_matrix_multp, P%EPS_Auger, P%OUTPUT)
            call get_some_eigenvalues_davpack(N_final_csfs, nev, min(P%davblock, nev), &
              min(max(P%davMax, nev),N_final_csfs), final_CI_energy(1:nev), &
              csf_c2, ci_diagonal, matrix_matrix_multp, &
              P%EPS_Auger, P%OUTPUT)
          else if (P%yes_use_krylov) then
#ifdef WITH_KRYLOV
            call get_some_eigenvalues_libKrylov(N_final_csfs, nev, nev, &
              final_CI_energy(1:nev), &
              csf_c2, ci_diagonal, matrix_matrix_multp, &
              epsilon=P%EPS_Auger, OUTPUT=P%OUTPUT)
#else
            stop "not compiled with libkrylov support"
#endif
          end if
          deallocate (CI_diagonal)
          deallocate (F_final%teint_abcd)
          if (nintern .lt. size(active) .and. nintern .gt. 0) then
            deallocate (F_final%teint_cdij)
            deallocate (F_final%teint_cidj)
            deallocate (F_final%teint_cdei)
          end if
        else
          allocate (final_CI_matrix(N_final_csfs, N_final_csfs))
          if (nintern .lt. size(active) .and. nintern .gt. 0) then
            call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
                             F_final%teint_abcd, &
                             final_occ(active(:), 1:N_final_occ), &
                             final_CI_matrix, nintern=nintern, &
                             teint_iiee=F_final%teint_cdij, &
                             teint_ieie=F_final%teint_cidj, &
                             teint_iiie=F_final%teint_cdei)
            deallocate (F_final%teint_cdij)
            deallocate (F_final%teint_cidj)
            deallocate (F_final%teint_cdei)
          else
            call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
                               F_final%teint_abcd, final_occ(active(:), 1:N_final_occ), final_CI_matrix)
          end if
          deallocate (F_final%teint_abcd)
          write (P%OUTPUT, '("# done")')
          ! solve it
          write (P%OUTPUT, '("# solving CI matrix")')
          allocate (csf_c2(N_final_csfs, nev))
          call get_some_ev_of_real_sym(final_CI_matrix, final_CI_energy, csf_c2, nev, P%EPS_Auger, yes_verbose=.False.)
          write (P%OUTPUT, '("# done")')
          deallocate (final_CI_matrix)
        end if

        do pol = 1,3
          do i = 1, size(csf_c2, 2)
            final_state_tdipole(i, pol) = sum(csf_c2(:, i) * final_csf_tdipole(:, pol))
          end do
        end do ! pol
        do i = 1, size(csf_c2,2)
          F_energy = E_ini - final_CI_energy(i)
          factor = alpha**3*F_energy**3*FOUR/THREE
          write (P%OUTPUT, '("# ",i4," E=",f20.5," a.u."," Fluor. Energy=",f15.5," eV")') &
            i, final_CI_energy(i) + nuclear_repulsion_energy(A, P), &
            AU2EV*F_energy
          do k = 1, size(final_csf_array, 1)
            if (abs(csf_c2(k, i)) .gt. 1e-1) &
              write (P%OUTPUT, '("# ",5x,f10.5,x,a,3es15.5)', advance="yes") &
              csf_c2(k, i), trim(csfstring(final_csf_array(k)%occ, final_csf_array(k)%csf)), final_csf_tdipole(k, :)
          end do
          write (P%OUTPUT, "('# ',5x,'Fluorescence energy+rate:',f15.5,2x,es15.5,4x,3es15.5)") &
            AU2EV*F_energy, factor*sum(final_state_tdipole(i, :)**2), final_state_tdipole(i, :)
          Totalrate(f_irrep) = Totalrate(f_irrep) + factor*sum(final_state_tdipole(i, :)**2)
        end do! i

        write (P%OUTPUT, "('#  Total Fluorescence for irrep ',i4,' : ',es13.5,' a.u. (lifetime=',a,')')") &
          f_irrep, Totalrate(f_irrep), trim(lifetime(TotalRate(f_irrep)))
        deallocate (final_csf_tdipole)
        deallocate (final_state_tdipole)
        deallocate (csf_c2)
        deallocate (final_CI_energy)
        call purge_UMatrixTable(UTfin)
      end if! N_final_csfs > 0
      deallocate (final_occ)
      call purge_csf_array(final_csf_array)
      deallocate(flag_final_csfs)
    end do! f_irrep

    call purge_csf_array(initial_csf_array)
    deallocate (initial_occ)
    deallocate (initial_occ_orig)
    deallocate (initial_CI_matrix)
    deallocate (initial_CI_energy)
    deallocate (core)
    deallocate (active)
    !deallocate(flag_initial_occ)
    call purge_UMatrixTable(UTini)
    write (P%OUTPUT, "('#  Total Fluorescence  ',4x,' : ',es13.5,' a.u. (lifetime=',a,')')") &
      sum(Totalrate(:)), trim(lifetime(sum(TotalRate(:))))
  contains
    function matrix_matrix_multp(n, ncol, Cin, Cout) result(error)
      integer, intent(in) :: n, ncol
      real(long), intent(in) :: Cin(n, ncol)
      real(long), intent(out) :: Cout(n, ncol)
      integer :: error
      call calc_CIMatrix_matrix_multp(UTfin, E_core, F_core(active(:), active(:)), &
        F_final%teint_abcd, &
        final_occ(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F_final%teint_cdij, &
        teint_ieie=F_final%teint_cidj, &
        teint_iiie=F_final%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_matrix_multp

    function matrix_vector_multp(n, Cin, Cout) result(error)
      integer, intent(in) :: n
      real(long), intent(in) :: Cin(n)
      real(long), intent(out) :: Cout(n)
      integer :: error
      Cout = Cin
      call calc_CIMatrix_vector_multp(UTfin, E_core, F_core(active(:), active(:)), &
        F_final%teint_abcd, &
        final_occ(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F_final%teint_cdij, &
        teint_ieie=F_final%teint_cidj, &
        teint_iiie=F_final%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_vector_multp
  end subroutine calc_Fluorescence_rates_CI

  ! This function calculates the fluorescence rate based on the independent particle approximation
  ! i.e. 'part' -> 'hole' for a specific molecular orbital part and hole
  ! energy is the transition energy (i.e. the energy eigenvalue differences of the part - hole MOs)
  function fluorescence_rate(tdipole_mo, hole, part, energy) result(frate)
    real(long), dimension(:, :, :), intent(in) :: tdipole_mo
    integer, intent(in) :: hole, part
    real(long) :: energy ! energy of transition

    real(long) :: frate ! fluorescence rate return value
    real(long) :: alpha = ONE/SPEED_C_AU ! fine structure constant

    frate = alpha**3*energy**3*FOUR/THREE*sum(tdipole_mo(hole, part, :)**2)

  end function fluorescence_rate

!----------------------------------------------------------------------------
! print out lifetime in fs or attosec.
!----------------------------------------------------------------------------
  function lifetime(rate) result(string)
    real(long), intent(in) :: rate                ! in a.u.
    real(long) :: tau
    character(len=20) :: str, string
    tau = ONE/rate*AU2SEC/1e-15_long        ! to femtosec.

    if (tau .gt. 1e18_long) then ! For us more than 100 second is infinte
      write (str, "('Infinity')")
    else if (tau .gt. 1e15_long) then
      tau = tau/1e15_long                       ! to sec.
      write (str, "(f15.1,' s')") tau
    else if (tau .gt. 1e12_long) then
      tau = tau/1e12_long                       ! to milli sec.
      write (str, "(f15.1,' ms')") tau
    else if (tau .gt. 1e9_long) then
      tau = tau/1e9_long                       ! to micro sec.
      write (str, "(f15.1,' mu s')") tau
    else if (tau .gt. 1e6_long) then
      tau = tau/1e6_long                       ! to nanosec.
      write (str, "(f15.1,' ns')") tau
    else if (tau .gt. 1000_long) then
      tau = tau/1000_long                       ! to pikosec.
      write (str, "(f15.1,' ps')") tau
    else if (tau .gt. ONE) then
      write (str, "(f15.1,' fs')") tau           ! to femtosec.
    else
      tau = tau*1000_long                       ! to attosec.
      write (str, "(f15.1,' as')") tau
    end if
    string = trim(adjustl(str))
  end function lifetime

end module fluorescence
