!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!


module gaussian_cphf
  !! author: Caroline Arnold
  !! author: Ludger Inhester
  !! Solves the Coupled Perturbed Hartree-Fock Equations
  !! HF case, Gaussian basis
  !! max_iter and convergence criterion from SCF parameters
  !! 2017,2018 @ CFEL
  use constants
  use basis_func
  use wavefunction
  use atom_in_mol
  use matrix
  use gaussian
  use hartreefock
  use stdio

  implicit none
  private
  public :: gaussian_cphf_iteration
contains
  !> this subroutine performs a coupled perturbed hartree fock calculation
  !! see details in Gerrat & Mills, JCP 49 1719
  !!
  subroutine gaussian_cphf_iteration(P, A, B, F, DU, orb_grad, E_grad, O_grad_out, CIS_state)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(out), dimension(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3) :: DU ! du/dr vector
    real(long), intent(out), dimension(size(F%LC, 2), size(B%atm, 2)*3) :: orb_grad ! orbital gradient
    real(long), intent(out), optional, dimension(size(B%atm, 2)*3) :: E_grad ! total energy gradient, optional
    real(long), intent(out), optional, dimension(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3) :: O_grad_out ! overlap gradient, optional
    integer, intent(in), optional :: CIS_state
    logical :: CIS
    logical :: yes_converged   ! True if converged
    integer :: i, j, k, l
    integer :: iter_steps, n_diis
    integer :: atom
    !real(long) :: DU_last(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3)
    real(long) :: DU_last_r(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3)
    real(long) :: DU_old(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3, P%DIIS_nvector)
    real(long) :: err_old(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3, P%DIIS_nvector)
    real(long) :: Mov_r(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3)
    real(long) :: Moo_r(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3)
    real(long) :: B_r(size(F%LC, 2), size(F%LC, 2), count(A(:)%charge /= 0) * 3)

    real(long) :: O_grad(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3)
    real(long) :: h0_grad(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3)
    real(long) :: J_grad(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3)
    real(long) :: K_grad(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3)

    real(long) :: O_grad_MO(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3)
    real(long) :: h0_grad_MO(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3)
    real(long) :: J_grad_MO(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3)
    real(long) :: K_grad_MO(size(F%LC, 2), size(F%LC, 2), size(B%atm, 2)*3)

    real(long), allocatable :: J_cis(:, :)
    real(long), allocatable :: K_cis(:, :)
    real(long), allocatable :: DD_cis(:, :)
    real(long), allocatable :: DD_cis_d(:, :, :)
    real(long), allocatable :: DLC_CIS_part_hole(:, :, :)
    real(long), allocatable :: DLC_CIS_hole_part(:, :, :)

    real(long) :: LC_T(size(F%LC, 2), size(F%LC, 1))
    real(long) :: DU_r(size(DU,1), size(DU,2), count(A(:)%charge /= 0) * 3)
    real(long) :: DU_i(size(DU,1), size(DU,2), count(A(:)%charge == 0) * 3)
    real(long), allocatable, save :: DU_r_guess(:, :, :)
    logical, save :: first_run = .True.
    real(long) :: residue
    real :: timer
    real :: time_init, time_done, time_step, time_last_step
    integer*8 :: int_time, int_time_last, int_time_step      ! timer
    integer :: dimr, rind_thresh(count(A(:)%charge /= 0) * 3), rind_thresh_i(count(A(:)%charge == 0) * 3), dimri
    CIS = .False.
    if (present(CIS_state)) CIS = .True.
    if (CIS) then
      if (CIS_state .lt. 1) CIS = .False.
    end if
    ! array index of coordinates belonging to true atoms (not point charges)
    dimr = 0
    dimri = 0
    rind_thresh(:) = 0
    rind_thresh_i(:) = 0
    do i = 1, size(A, 1)
      if (A(i)%charge .ne. 0) then
        dimr = dimr + 3
        rind_thresh(dimr - 2) = 3*i - 2
        rind_thresh(dimr - 1) = 3*i - 1
        rind_thresh(dimr) = 3*i
      else
        dimri = dimri + 3
        rind_thresh_i(dimri - 2) = 3*i - 2
        rind_thresh_i(dimri - 1) = 3*i - 1
        rind_thresh_i(dimri) = 3*i
      end if
    end do
    call check_time(int_time, timer)
    time_init = timer

    if (P%yes_verbose) then
      write (P%OUTPUT, "('###### Coupled Perturbed Hartree Fock Equations #####')")
    end if

    ! in case we have stored the previous result, reuse this as an initial guess
    if(.not. first_run .and. allocated(DU_r_guess)) then
      DU_r(:, :, :) = DU_r_guess(:, :, :)
    else
      DU_r(:, :, :) = ZERO
    end if
    if (P%yes_HF .eqv. .False. .or. any(F%N_occ .eq. 1)) then
      stop 'CPHF does only work with closed shell HF!'
    end if
    ! ===========================================================
    ! Iterative procedure for solving the CPHFE
    !
    ! In memory:
    !  * Fock matrix F in MO basis
    !  * Derivative of Fock matrix F_grad in MO basis
    !  * Derivative of overlap matrix O_grad
    !  * 2ee integrals (ij|kl) in MO basis
    !
    ! (0) Start with du/dR = 0
    ! (1) Calculate M
    ! (2) Calculate dF/dR
    ! (3) Calculate du_ij/dR and du_ii/dR
    ! (4) Check to see if du/dR is converged, else go back to (1)
    ! ===========================================================

    LC_T = transpose(F%LC)
    call check_time(int_time, timer)
    time_last_step = timer
    int_time_last = int_time

    ! Calculate h0_grad, J_grad, K_grad
    call h0_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, h0_grad)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","time for h0_grad",f15.3," s Real time:",f15.3," s")') &
        timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time

    call jk_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, F%D, J_grad, K_grad, symmetricD=.True., accuracy=P%EPS_teint)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","CPU time for JK_grad",f15.3," s Real time:",f15.3," s")') &
      timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time

    ! Calculate overlap matrix O_grad
    call overlap_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, O_grad)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","CPU time for olap_grad",f15.3," s Real time:",f15.3," s")') &
      timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time

    ! Transform gradient quantitites to MO space
    !$omp parallel default(none) shared(F,h0_grad,J_grad,K_grad,O_grad,h0_grad_MO,J_grad_MO,K_grad_MO,O_grad_MO,LC_T) private(k)
    !$omp do
    do k = 1, size(h0_grad, 3)
      h0_grad_MO(:, :, k) = LC_T(:, :) .m. h0_grad(:, :, k) .m. F%LC(:, :)
      J_grad_MO(:, :, k) = LC_T(:, :) .m. J_grad(:, :, k) .m. F%LC(:, :)
      K_grad_MO(:, :, k) = LC_T(:, :) .m. K_grad(:, :, k) .m. F%LC(:, :)
      O_grad_MO(:, :, k) = LC_T(:, :) .m. O_grad(:, :, k) .m. F%LC(:, :)
    end do
    !$omp end do
    !$omp end parallel
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","CPU time for transform",f15.3," s Real time:",f15.3," s")') &
      timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time
    ! Construct the occupied,occupied part of the M matrix
    call construct_Moo_Matrix2(P, A, B, F, O_grad_MO(:,:,rind_thresh(:)), Moo_r)
    ! (0) Start with du/dR = 0
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","time for Moo Matrix",f15.3," s Real time:",f15.3," s")') &
      timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time
    do k = 1, dimr
      do j = 1, size(F%LC, 2)
        do i = 1, size(F%LC, 2)
          B_r(i, j, k) = O_grad_MO(i, j, rind_thresh(k)) * F%CE(j) - h0_grad_MO(i, j, rind_thresh(k)) &
            - J_grad_MO(i, j, rind_thresh(k)) + HALF*K_grad_MO(i, j, rind_thresh(k)) -  Moo_r(i, j, k)
        end do! i
      end do! j
    end do! k

    yes_converged = .False.
    iter_steps = 0
    n_diis = 0
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CPHF ","CPU time for preparations",f15.3,"sec ")') timer - time_init
    int_time_last = int_time
    time_last_step = timer
    !!!!!!!!!!!!!!!!!!!!!!!!!! beginning of  CPHF loop
    ! Iterate until converged
    do while ((.not. yes_converged) .and. (iter_steps .lt. P%SCF_max_iter))
      call check_time(int_time, timer)
      time_last_step = timer
      time_step = timer
      int_time_step = int_time
      int_time_last = int_time
      ! Keep last value of DU for convergence test
      DU_last_r(:, :, :) = DU_r(:, :, :)
      iter_steps = iter_steps + 1
      call construct_Mov_Matrix2(P, A, B, F, DU_r, Mov_r)
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CPHF ","CPU time for Mov Matrix",f15.3," s  Real time:",f15.3," s")') &
        timer - time_last_step, user_time(int_time - int_time_last)
      int_time_last = int_time
      time_last_step = timer
      ! (2) Calculate dF/dR
      !F_grad_r(:, :, :) = h0_grad_MO(:, :, rind_thresh(:)) + J_grad_MO(:, :, rind_thresh(:))&
      !                    - HALF*K_grad_MO(:, :, rind_thresh(:)) + Moo_r(:, :, :) + Mov_r(:, :, :)
      ! (3) Calculate du_ij/dR and du_ii/dR
      do k = 1, dimr
        do j = 1, size(F%LC, 2)
          DU_r(j, j, k) = - HALF * O_grad_MO(j, j, rind_thresh(k))
          do i = 1, size(F%LC, 2)
            if (i .eq. j) then
              cycle
            else
              if (abs(F%CE(i) - F%CE(j)) .lt. 1e-5) then ! ugly work-around if orbitals are degenerate
                DU_r(i, j, k) = - HALF * O_grad_MO(i, j, rind_thresh(k))
              else
                DU_r(i, j, k) = (B_r(i, j, k) - Mov_r(i, j, k))/(F%CE(i) - F%CE(j))
              end if
            end if
          end do ! j
        end do ! i
      end do ! k

      ! (4) Convergence test: only check virtual / occupied pairs ?
      residue = maxval(abs(DU_r(:, :, rind_thresh(:)) - DU_last_r(:, :, rind_thresh(:))))
      yes_converged = .False.
      if (residue .lt. P%SCF_energy_conv) yes_converged = .True.

      if (P%yes_CPHFDIIS) then ! store old DU values in diis array
        if (n_diis .lt. P%DIIS_nvector) n_diis = n_diis + 1
        do i = n_diis, 2, -1
          DU_old(:, :, :, i) = DU_old(:, :, :, i - 1)
          err_old(:, :, :, i) = err_old(:, :, :, i - 1)
        end do
        DU_old(:, :, :, 1) = DU_r(:, :, :)
        err_old(:, :, :, 1) = DU_r(:, :, :) - DU_last_r(:, :, :) ! the error estimate of iteration DU is DU-DU_last
        if (n_diis .gt. 1) then ! do an extrapolation using the stored previous iterations
          call diis_step(DU_old, err_old, n_diis, DU_r(:,:,:))
        end if
      end if! yes_CPHFDIIS

      call check_time(int_time, timer)
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# CPHF iteration step:  ',i4,x,'RMS du/dr',es20.7, x,'CPU time=',f15.5,' s Real time: ',f15.5,' s')") &
          iter_steps, sqrt(sum((DU_r(:, :, :) - DU_last_r(:, :, :))**2)), &
          timer - time_step, user_time(int_time - int_time_step)
      end if
    end do ! while (not yes_converged and iter_steps < max_iter )
    !!!!!!!!!!!!!!!!!!!!!!! end of loop !!!!!!!!!!!!!!!!!!
    if (.not. allocated(DU_r_guess)) allocate(DU_r_guess(size(DU_r,1), size(DU_r,2), size(DU_r,3)))
    DU_r_guess(:, :, :) = DU_r(:,:,:)
    first_run = .False.
    ! compute the remaining part of DU (those that involved point charges)
    do k = 1, dimri
      do i = 1, size(F%LC, 2)
        do j = 1, size(F%LC, 2)
          if (i .eq. j) then
            DU_i(i, j, k) = - HALF * O_grad_MO(i, j, rind_thresh_i(k))
          else
            if (abs(F%CE(i) - F%CE(j)) .lt. 1e-5) then ! ugly work-around if orbitals are degenerate
              DU_i(i, j, k) = - HALF * O_grad_MO(i, j, rind_thresh_i(k))
            else
              DU_i(i, j, k) = (&
                O_grad_MO(i, j, rind_thresh_i(k)) * F%CE(j) &
                - h0_grad_MO(i, j, rind_thresh_i(k)) &
                - J_grad_MO(i, j, rind_thresh_i(k)) &
                + HALF*K_grad_MO(i, j, rind_thresh_i(k))&
                ) / (F%CE(i) - F%CE(j))
            end if
          end if
        end do ! j
      end do ! i
    end do ! k
    DU(:, :, rind_thresh(:)) = DU_r(:, :, :)
    DU(:, :, rind_thresh_i(:)) = DU_i(:, :, :)
    if (.not. yes_converged) &
      write (P%OUTPUT, '("# CPHF NOT converged!")')
    do k = 1, dimr
      do i = 1, size(F%LC, 2)
        orb_grad(i, rind_thresh(k)) =  - B_r(i, i, k) + Mov_r(i, i, k)
      end do
    end do
    do k = 1, size(B%atm, 2) * 3 - dimr
      do i = 1, size(F%LC, 2)
        orb_grad(i, rind_thresh_i(k)) = h0_grad_MO(i, i, rind_thresh_i(k))&
                                         + J_grad_MO(i, i, rind_thresh_i(k))&
                                         - HALF*K_grad_MO(i, i, rind_thresh_i(k))&
                                         - O_grad_MO(i, i, rind_thresh_i(k)) * F%CE(i)
      end do
    end do
    ! also return total energy derivative if requested
    if (present(E_grad)) then
      E_grad(:) = ZERO
      do k = 1, size(B%atm, 2)*3
        do i = 1, size(F%LC, 2)
          if (F%N_occ(i) .eq. 2) then
            E_grad(k) = E_grad(k) + TWO * h0_grad_MO(i, i, k) + J_grad_MO(i, i, k) - &
                        HALF * K_grad_MO(i, i, k) - TWO * O_grad_MO(i, i, k) * F%CE(i)
          end if
        end do
      end do
    end if
    if (present(O_grad_out)) &
      O_grad_out = O_grad_MO
    ! This calculates the gradient of CIS vector with coefficients stored in F%CIS_C
    ! excitations described in F%CIS_excitation
    if (CIS) then
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_last_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient state :  ',i4)") CIS_state
      end if
      allocate (DD_cis(size(F%LC, 1), size(F%LC, 1)))
      allocate (DD_cis_d(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3))
      allocate (J_cis(size(F%LC, 1), size(F%LC, 1)))
      allocate (K_cis(size(F%LC, 1), size(F%LC, 1)))

      ! this is the ground->excited state transition density matrix (AO basis
      forall (k=1:size(F%LC, 1), l=1:size(F%LC, 1)) &
        DD_cis(k, l) = sum(F%C_cis(:, CIS_state)*F%LC(k, F%CIS_excitation(2, :))*F%LC(l, F%CIS_excitation(1, :)))

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: build DD_cis   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if

      ! newer, fast variant to calculate DD_cis_d:
      allocate (DLC_CIS_part_hole(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3))
      allocate (DLC_CIS_hole_part(size(F%LC, 1), size(F%LC, 1), size(B%atm, 2)*3))

      ! the following steps take a large share of computing resources
      DLC_CIS_part_hole(:, :, :) = ZERO
      DLC_CIS_hole_part(:, :, :) = ZERO
      !$omp parallel default(none) private(k,l,j) shared(F,DU,LC_T,DLC_CIS_hole_part,DLC_CIS_part_hole,CIS_state)
      !$omp do
      do k = 1, size(DU, 3)
        forall (l=1:size(F%LC, 1), j=1:size(F%LC, 1))
          DLC_CIS_part_hole(j, l, k) = DLC_CIS_part_hole(j, l, k) + &
                                       sum(F%C_cis(:, CIS_state)* &
                                           LC_T(F%CIS_excitation(1, :), l)*DU(j, F%CIS_excitation(2, :), k))
          DLC_CIS_hole_part(j, l, k) = DLC_CIS_hole_part(j, l, k) + &
                                       sum(F%C_cis(:, CIS_state)* &
                                           LC_T(F%CIS_excitation(2, :), l)*DU(j, F%CIS_excitation(1, :), k))
        end forall !l,j
      end do!k
      !$omp end do
      !$omp end parallel

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: build DLC_cis_part_hole / hole_part   ',f15.5,' s')") &
          time_step - time_last_step
        time_last_step = timer
      end if

      DD_cis_d(:, :, :) = ZERO
      do i = 1, size(DD_cis_d, 3)
        forall (l=1:size(F%LC, 1), k=1:size(F%LC, 1))
          DD_cis_d(k, l, i) = DD_cis_d(k, l, i) + &
                              sum(LC_T(:, k)*DLC_CIS_part_hole(:, l, i)) &
                              + sum(LC_T(:, l)*DLC_CIS_hole_part(:, k, i))
        end forall !l,k
      end do

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: build DD_cis_d   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if

      deallocate (DLC_CIS_hole_part)
      deallocate (DLC_CIS_part_hole)
      ! older, slower variant of the code above
      ! DD_cis_d(:,:,:)=ZERO
      ! do i=1,size(F%CIS_excitation,2)
      !   do j=1,size(F%LC,1)
      !     forall (k=1:size(F%LC,1))
      !       forall (l=1:size(F%LC,1))
      !         DD_cis_d(k,l,:) = DD_cis_d(k,l,:) + F%C_cis(i,CIS_state) * &
      !           (F%LC(k,j) * F%LC(l,F%CIS_excitation(1,i)) *  DU(j,F%CIS_excitation(2,i),:) &
      !           +F%LC(k,F%CIS_excitation(2,i)) * F%LC(l,j) *  DU(j,F%CIS_excitation(1,i),:))
      !       end forall! l
      !     end forall! k
      !   end do! j
      ! end do! i

      J_grad = ZERO
      K_grad = ZERO
      call jk_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, DD_cis, J_grad, K_grad, symmetricD=.False., accuracy=P%EPS_teint)

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: JK gradient   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if

      call construct_J_Matrix(B%teint, DD_cis, J_cis)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: J Matrix   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if
      call construct_K_Matrix(B%teint, DD_cis, K_cis)

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: K Matrix   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if

      do i = 1, size(F%CIS_excitation, 2)
        E_grad(:) = E_grad(:) + F%C_CIS(i, CIS_state)**2*(orb_grad(F%CIS_excitation(2, i), :) - orb_grad(F%CIS_excitation(1, i), :))
      end do

      if (P%multp .eq. 1) then
        do k = 1, size(F%LC, 1)
          do l = 1, size(F%LC, 1)
            E_grad(:) = E_grad(:) + TWO*DD_cis(k, l)*J_grad(k, l, :)
            E_grad(:) = E_grad(:) - ONE*DD_cis(k, l)*K_grad(k, l, :)
            E_grad(:) = E_grad(:) + FOUR*DD_cis_d(k, l, :)*J_cis(k, l)
            E_grad(:) = E_grad(:) - TWO*DD_cis_d(k, l, :)*K_cis(k, l)
          end do
        end do
      else if (P%multp .eq. 3) then
        do k = 1, size(F%LC, 1)
          do l = 1, size(F%LC, 1)
            E_grad(:) = E_grad(:) - ONE*DD_cis(k, l)*K_grad(k, l, :)
            E_grad(:) = E_grad(:) - TWO*DD_cis_d(k, l, :)*K_cis(k, l)
          end do
        end do
      end if

      if (P%yes_verbose) then
        call check_time(int_time, timer)
        time_step = timer
        write (P%OUTPUT, "('# CPHF - CIS gradient: E_grad   ',f15.5,' s')") time_step - time_last_step
        time_last_step = timer
      end if

      deallocate (DD_cis)
      deallocate (DD_cis_d)
      deallocate (J_cis)
      deallocate (K_cis)


      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CPHF time for CIS gradient:  ',f15.5,'s')") timer - time_last_step
      end if
    end if! CIS

    if (P%yes_verbose) then
      call check_time(int_time, timer)
      time_done = timer
      write (P%OUTPUT, '("############## ","CPHF total CPU time",f15.3," s")') time_done - time_init
    end if
  end subroutine gaussian_cphf_iteration

!> this routine does a diis step for provided vectors vec(1:n,:) and err_vec(1:n,:)
!! returns the interpolated result
!! the matrix array u contins u(:,:,:,n+1) matrices
  subroutine diis_step(u, err, n, new_u)
    real(long), intent(in) :: u(:, :, :, :)
    real(long), intent(in) :: err(:, :, :, :)
    real(long), intent(out) :: new_u(:, :, :)
    integer, intent(in) :: n
    integer :: i, j, ierr
    real(long) :: factor
    real(long), allocatable :: coef(:)
    real(long), allocatable :: B(:, :)
    real(long), allocatable :: RHS(:)
    allocate (B(n + 1, n + 1))
    allocate (coef(n + 1))
    allocate (RHS(n + 1))
    RHS = ZERO
    RHS(n + 1) = -ONE
    B(1:n, n + 1) = -ONE
    B(n + 1, 1:n) = -ONE
    B(n + 1, n + 1) = ZERO

    do i = 1, n
      do j = i, n
        B(i, j) = sum(err(:, :, :, i)*err(:, :, :, j))
        B(j, i) = B(i, j)
      end do
    end do

    ! Rescale B matrix
    factor = ONE/B(1, 1)
    do i = 1, n
      do j = 1, n
        B(i, j) = B(i, j)*factor
      end do
    end do

    call solve_AxB(B, RHS, coef, ierr)

    ! if something went wrong: return the last vec
    if (ierr .lt. 0) then
      coef(1:n - 1) = ZERO
      coef(1) = ONE
    end if ! ( ierr < 0 )
    new_u(:, :, :) = ZERO
    do i = 1, n
      new_u(:, :, :) = new_u(:, :, :) + coef(i)*u(:, :, :, i)
    end do

    deallocate (B)
    deallocate (coef)
    deallocate (RHS)
  end subroutine diis_step

end module gaussian_cphf

