#!/usr/bin/env python
#
# XMOLECULE electronic structure toolkit
# Copyright 2023 Ludger Inhester
# Ludger Inhester, Center for Free-Electron Laser Science, DESY
# ludger.inhester@cfel.de
#
# XMOLECULE is distributed under the terms of the xxx licence.
#
# program to compute the molecular Hessian
# using analytical first order derivatives
# and numerical second order derivatives
# author: Ludger Inhester
import argparse
import re
import numpy as np
import pyxmol

BOHR2ANGSTROM = 0.529177249
AMU2AU = 1822.877332827

def richardson(f, hi, con=1.4, niter=10, safe=2.0, fast=False):
    """
    Implements Richardson extrapolation to calculate numerical estimates
    of values obtained by finite difference formulas.
    f -- function depending on 1 variable, the finite difference, what f is,
         a derivative, a numerical integral... does not play any role here
    hi -- initial estimate for the finite difference
    con -- constant factor by which hi is divided at each step
    niter -- maximum number of iterations
    safe -- when error grows by a factor more than safe, terminate and return
    """
    if fast:
        if hi > 0.05:
            return f(0.05), 0.0, hi
        else:
            return f(hi), 0.0, hi
    con2 = con*con
    big = 1.e30
    a = np.zeros((niter, niter), float)
    hh = hi
    a[0, 0] = f(hh)
    err = big
    for i in range(1, niter):
        hh = hh/con
        a[0, i] = f(hh)
        fac = con2
        for j in range(1, i+1):
            a[j, i] = (a[j-1, i]*fac - a[j-1, i-1])/(fac-1.0)
            fac = con2*fac
            errt = max(abs(a[j, i]-a[j-1, i]), abs(a[j, i]-a[j-1, i-1]))
            if errt <= err:
                err = errt
                res = a[j, i]
                hf = hh
        if abs(a[i, i]-a[i-1, i-1]) >= safe*err:
            return res, err, hf
    return res, err, hf



def gradient(f, x, ms=None, step=0.1, fast=True):
    """Return an array with the gradient in x along the directions in modes

    f -- function with respect the gradient is taken
    x -- coordinates to take the gradient
    ms -- list of arrays with orthonormal directions
    step -- step to take in the numerical differentiation
    """
    if type(ms) == type(None):
        ms = np.identity(len(x))
    g = []
    for m in ms:
        g.append(_directionalDerivative(f, x, m, step, fast=fast))
    return np.array(g)


def hessian(f, x, ms=None, step=0.1, fast=True):
    """Return an array with the hessian in x along the directions in modes

    f -- function with respect the hessian taken
    x -- coordinates to take the hessian
    ms -- list of arrays with orthonormal directions
    step -- step to take in the numerical differentiation
    """
    if type(ms) == type(None):
        ms = np.identity(len(x))
    h = np.zeros((len(ms), len(ms)), float)
    for i in range(len(ms)):
        for j in range(0, i+1):
            if i == j:
                h[i][j] = _secondDerivative(f, x, ms[i], step, fast=fast)
            else:
                h[i][j] = _crossedDerivative(
                    f, x, ms[i], ms[j], step, fast=fast)
                h[j][i] = h[i][j]
    return h


def _directionalDerivative(func, x, m, step=0.1, fast=True):
    """Here m is just an array representing the direction to take
    the derivative
    """
    def findiffd1(hh):
        return (func(x+m*hh)-func(x-m*hh))/(2.0*hh)
    d, err, hf = richardson(findiffd1, step, fast=fast)
    return d


def _secondDerivative(func, x, m, step=0.1, fast=True):
    def findiffd2(hh):
        return (func(x+m*hh) + func(x-m*hh) - 2*func(x))/(hh**2)
    d, err, hf = richardson(findiffd2, step, fast=fast)
    return d


def _crossedDerivative(func, x, m1, m2, step=0.1, fast=True):
    def findiffd2d2(hh):
        x1i2i = x - hh*m1 - hh*m2
        x1i2f = x - hh*m1 + hh*m2
        x1f2i = x + hh*m1 - hh*m2
        x1f2f = x + hh*m1 + hh*m2
        e1i2i = func(x1i2i)
        e1i2f = func(x1i2f)
        e1f2i = func(x1f2i)
        e1f2f = func(x1f2f)
        return ((e1f2f - e1f2i) - (e1i2f - e1i2i))/(4.0*hh**2)
    d, err, hf = richardson(findiffd2d2, step)
    return d




def parseArguments():
    parser = argparse.ArgumentParser(
                    description='computes the hessian matrix')
    parser.add_argument('filename', help="xmolecule input filename")
    args = parser.parse_args()
    return args


def get_hessian():
    args = parseArguments()

    XM = pyxmol.xmol(args.filename)
    XM.init()
    X = np.array(XM.get_geom())
    mass_simple = XM.mass()
    mass = np.zeros(len(X))
    for i in range(0, len(X)//3):
        for j in range(0, 3):
            mass[3*i+j] = mass_simple[i]
    print("#Masses: %s\n", mass)
    mass *= AMU2AU
    mass = np.array(mass)


    def getFuncEn():
        def f(x):
            XM.set_geom(x.tolist())
            E = XM.energy
            return (E)
        return f
    potential = getFuncEn()

    def getFuncGr():
        def f(x):
            XM.set_geom(x.tolist())
            G = XM.gradient()
            return np.array(G)
        return f
    potential_gradient = getFuncGr()
    
    hessian = gradient(potential_gradient, X, fast=True)
    symhessian = np.zeros((len(X), len(X)))
    asymhessian = np.zeros((len(X), len(X)))
    for i in range(0, len(X)):
        for j in range(0, len(X)):
            symhessian[i, j] = 0.5*(hessian[i, j]+hessian[j, i])
            asymhessian[i, j] = 0.5*(hessian[i, j]-hessian[j, i])
    print("#maximum asymmetry in Hessian %e" % np.amax(abs(asymhessian)))
    hessian = symhessian

    print("#writing hessian to %s" % "hessian.dat")
    with open("hessian.dat", "w") as file:
        for i in range(0, len(X)):
            for j in range(0, len(X)):
                file.write("%15.5e " % hessian[i, j])
            file.write("\n")
        file.close()


    oneoversqrtmass = np.diag(1.0/np.sqrt(mass))
    hessianw = np.dot(oneoversqrtmass, np.dot(hessian, oneoversqrtmass))
    eigen_val, eigen_vec = np.linalg.eigh(hessianw)


    print("Eigenvalues")
    for i in range(0, len(X)):
        imag = 0
        print("Eigenmode %d %f:" % (i, eigen_val[i]))
        if (eigen_val[i] < 0):
            print("%f H %f eV %f 1/cm %s" % (np.sqrt(-eigen_val[i]), np.sqrt(-eigen_val[i])*27.21, np.sqrt(-eigen_val[i])/(
                2*np.pi)*1e7*1/(2.418884)*1/(2.9979245800), "imaginary frequency"))
        elif (eigen_val[i] < 1e-6):
            print("%f H %f eV %f 1/cm %s" % (np.sqrt(eigen_val[i]), np.sqrt(eigen_val[i])*27.21, np.sqrt(
                eigen_val[i])/(2*np.pi)*1e7*1/(2.418884)*1/(2.9979245800), "very small frequency"))
        else:
            print("%f H %f eV %f 1/cm" % (np.sqrt(eigen_val[i]), np.sqrt(
                eigen_val[i])*27.21, np.sqrt(eigen_val[i])/(2*np.pi)*1e7*1/(2.418884)*1/(2.9979245800)))
        print(eigen_vec[:, i])
   
if __name__ == "__main__":
    get_hessian()
