!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module gradient
  !! author: Ludger Inhester
  !!
  !! contains various gradients subroutines
  use constants
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use wavefunction
  use nuclear_derivatives
  use gaussian_cphf
  use gaussian
  use operations
  use hartreefock
  implicit none
contains

  !> this subroutine calculates CIS gradient
  subroutine calculate_CIS_gradient_GTO(P, A, B, F, state, gradients_out, nacs_out, yes_calc_nacs, OUTPUT)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: state
    integer, optional, intent(in) :: OUTPUT

    real(long), optional :: gradients_out(:)
    real(long), optional :: nacs_out(:, :, :)
    logical, optional :: yes_calc_nacs
    logical :: yes_nacs
    real(long), allocatable :: nacs(:, :)
    real(long), allocatable :: nacel_halfTransformed(:, :)
    real(long), allocatable :: nacel_halfTransformedT(:, :)
    real(long), allocatable :: nacs_transformed(:, :, :)

    real(long) :: DU(size(F%LC, 2), size(F%LC, 2), 3*size(A))
    real(long) :: orb_grad(size(F%LC, 2), 3*size(A))
    real(long) :: E_grad(3*size(A))
    real(long) :: nuc_gradient(3*size(A))

    integer :: IOUT
    integer :: i, j, k, at
    integer :: nstates

    real :: timer
    real :: time_init, time_done, time_last_step
    integer*8 :: int_time      ! timer

    call check_time(int_time, timer)
    time_init = timer
    time_last_step = timer

    yes_nacs = .False.
    if (present(yes_calc_nacs)) yes_nacs = yes_calc_nacs

    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT

    call gaussian_cphf_iteration(P, A, B, F, DU, orb_grad, E_grad=E_grad, CIS_state=state)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# gradient ","time for cphf",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    nuc_gradient = nuclear_repulsion_gradient(A, P)
    E_grad = E_grad(:) + nuc_gradient(:)
    if (P%print_level>1) then
      if (state > 0) then
        write (IOUT, "('# CIS energy gradient in a.u. for state ',i8,' excitation energy ',f15.5 ,' eV:')") state, F%E_CIS(state)*AU2EV
      else
        write (IOUT, "('# energy gradient in a.u. for state ',i8,' (ground state) ')") state
      end if
      write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ",7x,a6,3f15.8)') A(i)%label, E_grad((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
      if (P%yes_verbose) then
        write (IOUT, '("# ","=================")')
        write (IOUT, '("# ",a13,3f15.8)') "gradient sum:", grad_sum(E_grad)
        write (IOUT, '("# ","=================")')
      end if
    end if

    if (.not. allocated(F%grad)) allocate (F%grad(3*size(A)))
    F%grad = E_grad
    if (present(gradients_out)) &
      gradients_out = E_grad
    if (yes_nacs) then
      call check_time(int_time, timer)
      time_last_step = timer

      allocate (nacs_transformed(size(F%C_CIS, 2) + 1, size(F%C_CIS, 2) + 1, 3 * size(A)))
      nacs_transformed(:, :, :) = ZERO
      !$omp parallel default(none) private(i,j,k,at,nacel_halfTransformed,nacel_halfTransformedT,nacs) shared(F,A,DU,nacs_transformed)
      allocate (nacs(size(F%CIS_excitation, 2) + 1, size(F%CIS_excitation, 2) + 1))
      allocate (nacel_halfTransformed(size(F%C_CIS, 2) + 1, size(nacs, 2)))
      allocate (nacel_halfTransformedT(size(nacs, 2), size(F%C_CIS, 2) + 1))
      !$omp do
      do at = 1, 3*size(A)
        nacs = ZERO
        do i = 1, size(F%C_cis, 1)
          nacs(i + 1, 1) = nacs(i + 1, 1) + &
            sqrt(HALF) * (DU(F%CIS_excitation(1, i), F%CIS_excitation(2, i), at) - DU(F%CIS_excitation(2, i), F%CIS_excitation(1, i), at))
          nacs(1, i + 1) = nacs(1, i + 1) + &
            sqrt(HALF) * (DU(F%CIS_excitation(2, i), F%CIS_excitation(1, i), at) - DU(F%CIS_excitation(1, i), F%CIS_excitation(2, i), at))
          do j = 1, size(F%C_cis, 1)
            if (F%CIS_excitation(1, i) .eq. F%CIS_excitation(1, j)) then
              nacs(i + 1, j + 1) = nacs(i + 1, j + 1) + &
                  HALF*(DU(F%CIS_excitation(2, i), F%CIS_excitation(2, j), at) - DU(F%CIS_excitation(2, j), F%CIS_excitation(2, i), at))
            else if (F%CIS_excitation(2, i) .eq. F%CIS_excitation(2, j)) then
              nacs(i + 1, j + 1) = nacs(i + 1, j + 1) + &
                  HALF*(DU(F%CIS_excitation(1, i), F%CIS_excitation(1, j), at) - DU(F%CIS_excitation(1, j), F%CIS_excitation(1, i), at))
            end if
          end do! i
        end do! j
        forall (j=2:size(nacel_halfTransformed, 2), i=2:size(nacel_halfTransformed, 1))
          nacel_halfTransformed(i, j) = sum(nacs(2:, j) * F%C_CIS(:, i - 1))
        end forall
        nacel_halfTransformedT = transpose(nacel_halfTransformed)
        forall (j=2:size(nacs_transformed, 2), i=2:size(nacs_transformed, 1))
          nacs_transformed(i, j, at) = sum(nacel_halfTransformedT(2:, i) * F%C_CIS(:, j - 1))
        end forall
        forall (j=2:size(nacs_transformed, 2))
          nacs_transformed(1, j, at) = sum(nacs(1, 2:) * F%C_CIS(:, j - 1))
          nacs_transformed(j, 1, at) = sum(nacs(2:, 1) * F%C_CIS(:, j - 1))
        end forall
      end do! at
      !$omp end do
      deallocate(nacel_halfTransformed)
      deallocate(nacel_halfTransformedT)
      deallocate(nacs)
      !$omp end parallel

      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# gradient ","time for transforming nacs",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer

      if (P%nstates .eq. -1) then
        nstates = size(F%CIS_excitation, 2) + 1
      else
        nstates = P%nstates
      end if

      if (P%print_level>1) then
        do j = 1, nstates
          if (j .eq. 1) then
            write (IOUT, "('# ',2x,'non-adiabatic coupling to ground state  ',' exc. energy ',f15.5,'eV')") ZERO
          else
            write (IOUT, "('# ',2x,'non-adiabatic coupling to state ',i8,' exc. energy ',f15.5,'eV')") j - 1, F%E_CIS(j - 1)*AU2EV
          end if
          write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
          do k = 1, size(A)
            write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, nacs_transformed(j, state + 1, (k - 1)*3 + 1:(k - 1)*3 + 3)
          end do
        end do
      end if

      if (present(nacs_out)) &
        nacs_out(:, :, :) = reshape(nacs_transformed(1:nstates, 1:nstates, :), &
                                    (/size(nacs_transformed, 3), size(F%C_CIS, 2) + 1, size(F%C_CIS, 2) + 1/), order=(/2, 3, 1/))

      deallocate (nacs_transformed)
    end if

    if (P%yes_verbose) then
      call check_time(int_time, timer)
      time_done = timer
      write (P%OUTPUT, '("############## ","gradient total time",f15.3,"sec ")') time_done - time_init
    end if

  end subroutine calculate_CIS_gradient_GTO

  !> this subroutine calculated the Koopmann's Theorem gradient
  !! assuming that an ionized state energy gradient is
  !! $$E_{i} = E_\text{gs} - e_i$$
  !! or and anionic state energy is
  !! $$E_{a} = E_\text{gs} + e_a$$
  !! it even calculates gradient for excited states assuming that
  !! $$E_{i->a} = E_\text{gs} +e_a - e_i,$$ which ignores mutual interaction between orbital a and i
  subroutine calculate_Koopmann_energy_gradient(P, A, B, F, OUTPUT)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, optional, intent(in) :: OUTPUT
    integer :: i, k
    real(long) :: DU(size(F%LC, 2), size(F%LC, 2), 3*size(A))
    real(long) :: orb_grad(size(F%LC, 2), 3*size(A))
    real(long) :: E_grad(3*size(A))
    real(long) :: nuc_gradient(3*size(A))
    integer :: actual_occ(size(F%N_occ))
    integer :: IOUT
    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT
    if (P%yes_valgs_occ) then
      actual_occ = F%N_occ
      if (.not. P%yes_valgs_occ_part) then
        F%N_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.True.)
      else
        F%N_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.False.)
      end if
    end if
    call gaussian_cphf_iteration(P, A, B, F, DU, orb_grad, E_grad=E_grad)
    nuc_gradient = nuclear_repulsion_gradient(A, P)
    E_grad = E_grad(:) + nuc_gradient(:)
    if (P%print_level > 1) then
      write (IOUT, "('# GS energy gradients (a.u.):')")
      write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ",7x,a6,3f15.8)') A(i)%label, E_grad((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
    end if
    if (P%yes_valgs_occ) then
      do i = 1, size(F%N_occ)
        if (actual_occ(i) - F%N_occ(i) .ne. 0) then
          E_grad(:) = E_grad(:) + orb_grad(i, :)*(actual_occ(i) - F%N_occ(i))
          if (P%print_level > 1) then
            write (IOUT, '("# ",sp,i2,ss,x,"orbital:",i5,x,"orbital energy:",f15.3)') actual_occ(i) - F%N_occ(i), i, F%CE(i)
            do k = 1, size(A)
              write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, orb_grad(i, (k - 1)*3 + 1:(k - 1)*3 + 3)
            end do
          end if
        end if
      end do
      F%N_occ = actual_occ
    end if

    if (.not. allocated(F%grad)) allocate (F%grad(3*size(A)))
    F%grad(:) = E_grad(:)
    if (P%yes_verbose) then
      write (IOUT, "('# Total Koopmann energy gradients (a.u.):')")
      write (IOUT, "('# ',7x,6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ","|GRAD| ",a6,3f15.8)') A(i)%label, E_grad((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
    end if

    if (P%yes_verbose) then
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ",a13,3f15.8)') "gradient sum:", grad_sum(E_grad)
      write (IOUT, '("# ","=================")')
    end if
  end subroutine calculate_Koopmann_energy_gradient

  subroutine calculate_hole_energies_Efield(P, A, B, F, pos, occs_out, energies_out, Efield_out, Upot_out, OUTPUT)
    !! calculates Koopmann's hole energies in the presence of an electric field
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long), intent(in) :: pos(3)
    real(long), optional :: Efield_out(:, :)

    real(long), optional :: energies_out(:)
    real(long), optional :: Upot_out(:)
    integer, optional :: occs_out(:, :)

    integer, optional, intent(in) :: OUTPUT

    real(long) :: D(size(F%LC, 1), size(F%LC, 1))
    real(long) :: vcoul(size(F%LC, 1), size(F%LC, 1))
    real(long) :: Upot, nuc_pot
    real(long) :: nuc_Efield(3)
    real(long) :: Efield(3)
    real(long), allocatable :: Rinv_grad(:, :, :)
    integer, allocatable :: occs(:, :)
    real(long), allocatable :: energies(:)
    integer :: actual_occ(size(F%N_occ))
    integer :: gs_occ(size(F%N_occ))
    integer :: i, j, Nval, Ncore, mu, nu, hole
    real(long) :: R_ij, R_corr
    real(long) :: Egs
    integer :: IOUT
    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT

    ! go to groundstate occupation
    if (P%yes_valgs_occ) then
      actual_occ = F%N_occ
      gs_occ = groundstate_occ(actual_occ, Ncore=F%Ncore)
      if (sum(gs_occ) .ne. sum(actual_occ(:)) + 1) stop 'only works for single hole configurations'
      F%N_occ = gs_occ
    end if
    Egs = F%E_tot + sum(F%CE(:)*(gs_occ(:) - actual_occ(:)))
    Nval = count(F%N_occ .eq. 2) - F%Ncore
    Ncore = F%Ncore
    allocate (occs(size(F%N_occ), Nval))
    allocate (energies(Nval))
    allocate (Rinv_grad(B%N_basis, B%N_basis, 3))
    call rinv_gradient_int(B%env, B%atm, B%bas, B%first_bf_of_shell, pos, Rinv_grad)
    ! for calculating the potential
    call calculate_Vcoul_Gaussian_basis_set(B%env, B%atm, B%bas, B%first_bf_of_shell, pos, vcoul)
    nuc_pot = ZERO
    nuc_Efield = ZERO
    do j = 1, size(B%atm, 2)
      R_ij = sqrt((A(j)%x - pos(1))**2 + &
                  (A(j)%y - pos(2))**2 + &
                  (A(j)%z - pos(3))**2)
      R_corr = sqrt(R_ij**2 + P%R_softcore**2)
      nuc_Efield(1) = nuc_Efield(1) + ONE*dble(A(j)%fcharge)/R_corr**3*(pos(1) - A(j)%x)
      nuc_Efield(2) = nuc_Efield(2) + ONE*dble(A(j)%fcharge)/R_corr**3*(pos(2) - A(j)%y)
      nuc_Efield(3) = nuc_Efield(3) + ONE*dble(A(j)%fcharge)/R_corr**3*(pos(3) - A(j)%z)
      nuc_pot = nuc_pot + dble(A(j)%fcharge)/R_corr
    end do
    if (P%yes_verbose .and. P%print_level .gt. 1) then
      write (IOUT, '("# ")')
      write (IOUT, '("# Electric field calculation at pos ",3es20.7)') pos
      write (IOUT, '("# ","nuclear  Efield:",3es20.7)') nuc_Efield
      write (IOUT, '("# ","nuclear  E-pot :",es20.7)') nuc_pot
    end if

    call density_matrix(F%LC, dble(gs_occ), D)
    Efield = nuc_Efield(:)
    Upot = nuc_pot
    do mu = 1, size(D, 1)
      do nu = 1, size(D, 1)
        Efield(:) = Efield(:) + D(mu, nu)*Rinv_grad(mu, nu, :)
        Upot = Upot - D(mu, nu)*vcoul(mu, nu)
      end do
    end do
    if (P%print_level .gt. 1) then
      write (IOUT, '("# ground state: ")')
      write (IOUT, '("# ","               ",5x,x," energy:",f15.3)') Egs
      write (IOUT, '("# ",2x,"occ=",99i1)') gs_occ
      write (IOUT, '("# ","E-pot     :",es20.7)') Upot
      write (IOUT, '("# ","Efield    :",3es20.7)') Efield
    end if

    do i = 1, Nval
      hole = Nval + Ncore - i + 1
      occs(:, i) = gs_occ(:)
      occs(hole, i) = 1
      energies(i) = Egs - F%CE(hole)
      call density_matrix(F%LC, dble(occs(:, i)), D)
      Efield = nuc_Efield(:)
      Upot = nuc_pot
      do mu = 1, size(D, 1)
        do nu = 1, size(D, 1)
          Efield(:) = Efield(:) + D(mu, nu)*Rinv_grad(mu, nu, :)
          Upot = Upot - D(mu, nu)*vcoul(mu, nu)
        end do
      end do
      if (P%print_level .gt. 1) then
        write (IOUT, '("# ")')
        write (IOUT, '("# state: ",i4)') i
        write (IOUT, '("# ","hole in orbital",i5,x," energy:",f15.3)') hole, energies(i)
        write (IOUT, '("# ",2x,"occ=",99i1)') occs(:, i)
        write (IOUT, '("# ","E-pot     :",es20.7)') Upot
        write (IOUT, '("# ","Efield    :",3es20.7)') Efield
      end if
      if (present(Efield_out)) then
        Efield_out(:, i) = Efield(:)
      end if
      if (present(Upot_out)) then
        Upot_out(i) = Upot
      end if
    end do !i

    if (present(occs_out)) then
      do i = 1, Nval
        occs_out(:, i) = occs(:, i)
      end do
      do i = Nval + 1, size(occs_out, 2)
        occs_out(:, i) = -1
      end do
    end if
    if (present(energies_out)) energies_out = energies

    deallocate (occs)
    deallocate (energies)
    deallocate (Rinv_grad)
    if (P%yes_valgs_occ) then
      F%N_occ = actual_occ
    end if
  end subroutine calculate_hole_energies_Efield

  subroutine calculate_hole_energies_gradients_nacs(P, A, B, F, occs_out, energies_out, gradients_out, nacs_out, particles, OUTPUT)
    !! calculates the Koopmann's Theorem gradient
    !! assuming that an excited state energy gradient is
    !! $$d/dR E_{i->a} = E_\text{gs} - e_i $$
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F

    integer, optional, intent(out) :: occs_out(:, :)
    real(long), optional, intent(out) :: energies_out(:)
    real(long), optional, intent(out) :: gradients_out(:, :)
    real(long), optional, intent(out) :: nacs_out(:, :, :)

    integer, optional, intent(in) :: OUTPUT
    logical, optional, intent(in) :: particles

    integer :: occ(size(F%N_occ))
    real(long) :: energy
    real(long), allocatable :: gradient(:)
    real(long), allocatable :: nacs(:, :)
    real(long) :: DU(size(F%LC, 2), size(F%LC, 2), 3*size(A))
    real(long) :: orb_grad(size(F%LC, 2), 3*size(A))
    !    real(long) :: O_grad(B%N_basis,B%N_basis,3*size(A))
    real(long) :: Egs_grad(3*size(A))
    real(long) :: nuc_gradient(3*size(A))
    real(long) :: Egs

    integer :: actual_occ(size(F%N_occ))
    integer :: gs_occ(size(F%N_occ))
    integer :: IOUT
    integer :: Nval, Ncore, Nvirt
    integer :: hole, holej, part, partj
    integer :: i, j, k
    logical :: yes_particle
    integer :: nOccs

    yes_particle = .False.
    if (present(particles)) yes_particle = particles
    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT
    ! go to groundstate occupation
    if (P%yes_valgs_occ) then
      actual_occ = F%N_occ
      if (yes_particle) then
        gs_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.False.)
        if (sum(gs_occ) .ne. sum(actual_occ(:)) - 1) stop 'only works for single particle configurations'
      else
        gs_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.True.)
        if (sum(gs_occ) .ne. sum(actual_occ(:)) + 1) stop 'only works for single hole configurations'
      end if
      F%N_occ = gs_occ
    end if
    call gaussian_cphf_iteration(P, A, B, F, DU, orb_grad, E_grad=Egs_grad)
    Ncore = F%Ncore
    Nval = count(gs_occ .eq. 2) - F%Ncore
    Nvirt = count(gs_occ .eq. 0)
    ! go back to original occupation
    if (P%yes_valgs_occ) then
      F%N_occ = actual_occ
    end if
    nuc_gradient = nuclear_repulsion_gradient(A, P)
    Egs_grad = Egs_grad(:) + nuc_gradient(:)
    if (maxval(abs(grad_sum(Egs_grad))) .gt. 1e-5) stop "wrong grad 1"
    if (P%print_level .gt. 1) then
      write (IOUT, "('# GS energy gradients (a.u.):')")
      write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ",7x,a6,3f15.8)') A(i)%label, Egs_grad((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
    end if
    if (P%yes_verbose) then
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ",a13,3f15.8)') "gradient sum:", grad_sum(Egs_grad)
      write (IOUT, '("# ",a13,3f15.8)') "gradient nuc sum:", grad_sum(nuc_gradient)
      write (IOUT, '("# ","=================")')
    end if
    if (maxval(abs(grad_sum(Egs_grad))) .gt. 1e-5) stop "wrong grad"

    if (.not. allocated(F%grad)) allocate (F%grad(3*size(A)))
    Egs = F%E_tot + sum(F%CE(:)*(gs_occ(:) - actual_occ(:)))
    if (.not. yes_particle) then
      nOccs = nVal
    else
      nOccs = Nvirt
    end if
    if (P%print_level .gt. 1) then
      allocate (gradient(3*size(A)))
      allocate (nacs(3*size(A), nOccs))
    end if
    if (.not. yes_particle) then
      do i = 1, Nval
        hole = Nval + Ncore - i + 1
        if (present(occs_out)) then
          occs_out(:, i) = gs_occ(:)
          occs_out(hole, i) = 1
        end if
        if (present(energies_out)) &
          energies_out(i) = Egs - F%CE(hole)
        if (present(gradients_out)) &
          gradients_out(:, i) = Egs_grad(:) - orb_grad(hole, :)
        if (present(nacs_out)) then
          do j = 1, Nval
            holej = Ncore + Nval + 1 - j
            nacs_out(:, j, i) = HALF*(DU(hole, holej, :) - DU(holej, hole, :))
          end do !j
        end if
        if (P%print_level .gt. 1) then
          occ(:) = gs_occ(:)
          occ(hole) = 1
          energy = Egs - F%CE(hole)
          gradient(:) = Egs_grad(:) - orb_grad(hole, :)
          do j = 1, Nval
            holej = Nval + Ncore + 1 - j
            nacs(:, j) = HALF*(DU(hole, holej, :) - DU(holej, hole, :))
          end do
          gradient(:) = Egs_grad(:) - orb_grad(hole, :)
          write (IOUT, '("# ")')
          write (IOUT, '("# state: ",i4)') i
          write (IOUT, '("# ","hole in orbital",i5,x," energy:",f15.3)') hole, energy
          write (IOUT, '("# ",2x,"occ=",99i1)') occ(:)
          write (IOUT, "('# ',2x,'gradient:')")
          write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
          do k = 1, size(A)
            write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, gradient((k - 1)*3 + 1:(k - 1)*3 + 3)
          end do
          if (P%yes_verbose) then
            write (IOUT, '("# ","=================")')
            write (IOUT, '("# ",a13,3f15.8)') "gradient sum:", grad_sum(gradient(:))
            write (IOUT, '("# ","=================")')
          end if
          do j = 1, Nval
            write (IOUT, "('# ',2x,'non-adiabatic coupling to state ',i4)") j
            write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
            do k = 1, size(A)
              write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, nacs((k - 1)*3 + 1:(k - 1)*3 + 3, j)
            end do
          end do
        end if
      end do
    else! if yes_particle
      do i = 1, Nvirt
        part = Ncore + Nval + i
        if (present(occs_out)) then
          occs_out(:, i) = gs_occ(:)
          occs_out(part, i) = 1
        end if
        if (present(energies_out)) &
          energies_out(i) = Egs + F%CE(part)
        if (present(gradients_out))&
          gradients_out(:, i) = Egs_grad(:) + orb_grad(part, :)
        if (present(nacs_out)) then
          do j = 1, Nvirt
            partj = Ncore + Nval + j
            nacs_out(:, j, i) = HALF*(DU(part, partj, :) - DU(partj, part, :))
          end do !j
        end if
        if (P%print_level .gt. 1) then
          occ(:) = gs_occ(:)
          occ(part) = 1
          energy = Egs + F%CE(part)
          gradient(:) = Egs_grad(:) + orb_grad(part, :)
          do j = 1, Nvirt
            partj = Ncore + Nval + j
            nacs(:, j) = HALF*(DU(part, partj, :) - DU(partj, part, :))
          end do
          write (IOUT, '("# ")')
          write (IOUT, '("# state: ",i4)') i
          write (IOUT, '("# ","particle in orbital",i5,x," energy:",f15.3)') part, energy
          write (IOUT, '("# ",2x,"occ=",99i1)') occ(:)
          write (IOUT, "('# ',2x,'gradient:')")
          write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
          do k = 1, size(A)
            write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, gradient((k - 1)*3 + 1:(k - 1)*3 + 3)
          end do
          if (P%yes_verbose) then
            write (IOUT, '("# ","=================")')
            write (IOUT, '("# ",a13,3f15.8)') "gradient sum:", grad_sum(gradient(:))
            write (IOUT, '("# ","=================")')
          end if
          do j = 1, Nvirt
            write (IOUT, "('# ',2x,'non-adiabatic coupling to state ',i4)") j
            write (IOUT, "('# ',6x,3a15)") "x", "y", "z"
            do k = 1, size(A)
              write (IOUT, '("# ",7x,a6,3f15.8)') A(k)%label, nacs((k - 1)*3 + 1:(k - 1)*3 + 3, j)
            end do
          end do
        end if
      end do! i
    end if! yes_particle

    if (P%print_level .gt. 1) then
      deallocate (nacs)
      deallocate (gradient)
    end if

    if (P%yes_valgs_occ) then
      F%N_occ = actual_occ
    end if
  end subroutine calculate_hole_energies_gradients_nacs

  subroutine calculate_energy_gradient_GTO(P, A, G, B, F, OUTPUT)
    !! calculates the RHF/ ROHF / HFS gradient when GTO are used
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, optional, intent(in) :: OUTPUT
    integer :: IOUT
    integer :: i
    !    real(long) :: grad_sum(3)
    real(long), allocatable :: gradient(:)
    real(long), allocatable :: olap_gradient(:)
    real(long), allocatable :: oe_gradient(:)
    real(long), allocatable :: vc_gradient(:)
    real(long), allocatable :: te_gradient(:)
    real(long), allocatable :: Ex_gradient(:)
    real(long), allocatable :: nuc_gradient(:)

    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT
    allocate (gradient(3*size(B%atm, 2)))
    allocate (olap_gradient(3*size(B%atm, 2)))
    allocate (oe_gradient(3*size(B%atm, 2)))
    allocate (vc_gradient(3*size(B%atm, 2)))
    allocate (te_gradient(3*size(B%atm, 2)))
    allocate (nuc_gradient(3*size(B%atm, 2)))
    allocate (Ex_gradient(3*size(B%atm, 2)))
    nuc_gradient = nuclear_repulsion_gradient(A, P)
    if (P%yes_openshell .and. P%yes_HF) then
      if(P%yes_unrestricted)then
        call overlap_energy_gradient_UHF(B%env, B%atm, B%bas, B%first_bf_of_shell, F%LC_A, F%LC_B, F%FockA, F%FockB, F%n_occ, olap_gradient)
      else
        call overlap_energy_gradient_ROHF(B%env, B%atm, B%bas, B%first_bf_of_shell, F%LC, F%FockA, F%FockB, F%n_occ, olap_gradient)
      end if
    else
      if (any(F%N_occ .eq. 1) .and. P%yes_HF) stop 'use -openshell for gradients of open shell configurations (ROHF)'
      call overlap_energy_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, F%LC, F%CE, F%f_occ, olap_gradient)
    end if
    call one_electron_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, F%D, oe_gradient)
    if (P%yes_openshell .and. P%yes_HF) then
      call two_electron_gradient_ROHF(B%env, B%atm, B%bas, B%first_bf_of_shell, F%Da, F%Db, te_gradient)
      !call two_electron_gradient(B%env,B%atm,B%bas,B%first_bf_of_shell,F%D,te_gradient, Fock=.True.)
    else if (P%yes_HF) then
      call two_electron_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, F%D, te_gradient, Fock=.True.)
    else
      call two_electron_gradient(B%env, B%atm, B%bas, B%first_bf_of_shell, F%D, te_gradient, Fock=.False.)
      call exchange_gradient(A, G, B, F, Ex_gradient)
    end if
    if (P%yes_verbose .and. P%print_level .gt. 1) then
      write (IOUT, '("# ","nuclear  gradient:",3es20.7)') nuc_gradient
      write (IOUT, '("# ","nuclear  grad.sum:",3es20.7)') grad_sum(nuc_gradient)
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ","overlap  gradient:",3es20.7)') olap_gradient
      write (IOUT, '("# ","overlap  grad.sum:",3es20.7)') grad_sum(olap_gradient)
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ","one elec gradient:",3es20.7)') oe_gradient
      write (IOUT, '("# ","one elec grad.sum:",3es20.7)') grad_sum(oe_gradient)
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ","two elec gradient:",3es20.7)') te_gradient
      write (IOUT, '("# ","two elec grad.sum:",3es20.7)') grad_sum(te_gradient)
      write (IOUT, '("# ","=================")')

    end if
    if (P%yes_HF) then
      gradient = nuc_gradient + olap_gradient + oe_gradient + te_gradient
    else
      if (P%yes_verbose .and. P%print_level .gt. 1) then
        write (IOUT, '("# ","exchange gradient:",3es20.7)') Ex_gradient
        write (IOUT, '("# ","exchange grad.sum:",3es20.7)') grad_sum(Ex_gradient)
      end if
      gradient = nuc_gradient + olap_gradient + oe_gradient + te_gradient + Ex_gradient
    end if

    if (P%yes_verbose .and. P%print_level .gt. 1) then
      write (IOUT, '("# ","=================")')
      write (IOUT, '("# ","total    grad.sum:",3es20.7)') grad_sum(gradient)
      write (IOUT, '("# ","=================")')
    end if
    if (P%print_level .gt. 1) then
      write (IOUT, "('# Total energy gradients (a.u.):')")
      write (IOUT, "('# ',7x,6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ","|GRAD| ",a6,3f15.8)') A(i)%label, gradient((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
    end if
    if (.not. allocated(F%grad)) allocate (F%grad(size(gradient, 1)))
    F%grad = gradient
    deallocate (olap_gradient)
    deallocate (oe_gradient)
    deallocate (te_gradient)
    deallocate (nuc_gradient)
    deallocate (Ex_gradient)
    deallocate (gradient)
  end subroutine calculate_energy_gradient_GTO

  subroutine calculate_orbital_gradient_GTO(P, A, B, F, OUTPUT)
    !! calculates the gradients of orbital energies
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, optional, intent(in) :: OUTPUT
    real(long) :: DU(size(F%LC, 2), size(F%LC, 2), 3*size(A))
    real(long) :: orb_grad(size(F%LC, 2), 3*size(A))
    real(long) :: E_grad(3*size(A))
    real(long) :: nuc_gradient(3*size(A))
    integer :: IOUT
    integer :: i, k
    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT
    call gaussian_cphf_iteration(P, A, B, F, DU, orb_grad, E_grad=E_grad)
    ! printout:
    write (IOUT, '("# Orbital energy gradients")')
    write (IOUT, '("# ",a5,a15,x,a15,x,a15,x,a15)') "index", "orb.E", "dE/dx", "dE/dy", "dE/dz"
    do i = 1, B%N_basis
      write (IOUT, '("# ",i5,f15.3)') i, F%CE(i)
      do k = 1, size(A)
        write (IOUT, '("# ",5x,a15,x,es15.8,x,es15.8,x,es15.8)') A(i)%label, orb_grad(i, (k - 1)*3 + 1:(k - 1)*3 + 3)
      end do
      if (P%yes_verbose) then
        write (IOUT, '("# ",5x,a15  ,x,es15.8,x,es15.8,x,es15.8)') "sum ", grad_sum(orb_grad(i, :))
      end if
    end do
    ! also print out total energy gradient
    nuc_gradient = nuclear_repulsion_gradient(A, P)
    E_grad = E_grad(:) + nuc_gradient(:)
    if (P%print_level > 1) then
      write (IOUT, "('# Total energy gradients (a.u.):')")
      write (IOUT, "('# ',7x,6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, '("# ","|GRAD| ",a6,3f15.8)') A(i)%label, E_grad((i - 1)*3 + 1:(i - 1)*3 + 3)
      end do
    end if
  end subroutine calculate_orbital_gradient_GTO

  function grad_sum(grad) result(g)
    !! sum the gradients along coordinates x,y,z
    real(long), intent(in) :: grad(:)
    real(long) :: g(3)
    integer :: N, i, j
    g(:) = ZERO
    N = size(grad)/3
    do i = 1, N
      do j = 1, 3
        g(j) = g(j) + grad((i - 1)*3 + j)
      end do
    end do
  end function grad_sum

  subroutine exchange_gradient(A, G, B, F, Ex_gradient)
    !! computes the gradient of the exchange potential part for HFS
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(out) :: Ex_gradient(size(A)*3)
    integer :: mu, nu, i, atom
    Ex_gradient(:) = ZERO
    do mu = 1, size(F%D, 1)
      do nu = 1, size(F%D, 2)
        atom = B%info(nu, 1)
        ! this is the gradient of rho with respect to the coordinate atom,i
        do i = 1, 3
          Ex_gradient((atom - 1)*3 + i) = Ex_gradient((atom - 1)*3 + i) + &
                                          TWO*F%D(mu, nu)*integral(G, B%func(:, mu)*F%V_xc(:)*B%dfdR(:, nu, i))
        end do
      end do
    end do
  end subroutine Exchange_gradient

  subroutine calculate_energy_gradient(P, A, G, B, F, OUTPUT)
    !! a wrapper for calculate_dEdR defined in nuclear_derivatives.f90
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in), optional:: OUTPUT
    real(long) :: dEdR(3, size(A))
    integer :: Jnuc_want(size(A)), Nwant
    integer :: i, j
    integer :: IOUT

    IOUT = P%OUTPUT
    if (present(OUTPUT)) IOUT = OUTPUT
    !
    ! set up for using calculate_dEdR in nuclear_derivatives.f90
    !
    Nwant = size(A)
    do i = 1, Nwant
      Jnuc_want(i) = i
    end do
    !
    ! calling calculate_dEdR()
    ! then store gradient into F%grad(:)
    !   For the i-th atom, dE/dR_i_x = F%grad( i*3 - 2 )
    !                      dE/dR_i_y = F%grad( i*3 - 1 )
    !                      dE/dR_i_z = F%grad( i*3     )
    !
    if (.not. allocated(F%grad)) &
      allocate (F%grad(size(A)*3))
    if (sum(F%N_occ) .eq. 0) then
      F%grad = nuclear_repulsion_gradient(A, P)
    else
      call calculate_dEdR(dEdR, P, A, G, B, F, Jnuc_want, Nwant)
      do i = 1, size(A)
        do j = 1, 3         ! x, y, z
          F%grad((i - 1)*3 + j) = dEdR(j, i)
        end do
      end do
    end if
    !
    ! print gradients
    !
    if (P%print_level > 1) then
      write (IOUT, "('# Total energy gradients (a.u.):')")
      write (IOUT, "('# ',7x,6x,3a15)") "x", "y", "z"
      do i = 1, size(A)
        write (IOUT, "('# |GRAD| ',a6,3f15.8)") A(i)%label, F%grad(i*3 - 2:i*3)
      end do
      write (IOUT, "('#')")
    end if
  end subroutine calculate_energy_gradient

end module gradient
