!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!


module localization
  !! author: Ludger Inhester
  !!
  !! this module contains routines for orbital localization
  !! parts of the code has been copied from Kotas original code
  use constants
  use read_param
  use wavefunction
  use basis_func
  use molecular_grid
  implicit none
  private
  public fix_orbitals_on_nuclei, localize_orbitals
contains

!> find the best permutation of orbitals specified in MO_index such that
!! sum(gross_pop(MO_index(:),nuc_index(:))) is maximized
  subroutine fix_orbitals_on_nuclei(P, B, F, MO_index, nuc_index)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in):: MO_index(:), nuc_index(:)
    integer :: i, j, n, mu, nu, moi, moj
    logical :: opt
    real(long) :: temp
    real(long), dimension(size(MO_index), P%n_nuc) :: Pop  ! Population on atom
    integer :: OUTPUT
    OUTPUT = P%OUTPUT
    if (P%yes_verbose) then
      write (OUTPUT, '("# Fixing orbitals on nuclei")')
      write (OUTPUT, '("# orbitals:",10i4)') MO_index
      write (OUTPUT, '("# nuclei  :",10i4)') nuc_index
    end if
    if (size(MO_index) .ne. size(nuc_index)) stop 'Error in fix_orbitals_on_nuclei'

    do i = 1, size(MO_index)
      Pop(i, :) = ZERO
      moi = MO_index(i)
      do mu = 1, P%N_basis

        do nu = 1, P%N_basis
          Pop(i, B%info(mu, 1)) = Pop(i, B%info(mu, 1)) + F%LC(mu, moi)*F%LC(nu, moi)*B%S(mu, nu)
        end do
      end do

    end do

    ! sum Pop(i,nuc_index(i))

    do
      opt = .TRUE.
      do i = 1, size(MO_index)
        moi = MO_index(i)

        do j = i + 1, size(MO_index)
          moj = MO_index(j)
          if(P%yes_sym)then
            if(F%irrep_MO(moi) .ne. F%irrep_MO(moj)) cycle
          end if
          if (Pop(i, nuc_index(i)) + Pop(j, nuc_index(j)) .lt. Pop(j, nuc_index(i)) + Pop(i, nuc_index(j))) then

            do mu = 1, P%N_basis
              temp = F%LC(mu, moi)
              F%LC(mu, moi) = F%LC(mu, moj)
              F%LC(mu, moj) = temp
            end do

            temp = F%CE(moi)
            F%CE(moi) = F%CE(moj)
            F%CE(moj) = temp

            do n = 1, P%N_nuc
              temp = Pop(i, n)
              Pop(i, n) = Pop(j, n)
              Pop(j, n) = temp
            end do

            opt = .FALSE.
          end if
          if (opt .eqv. .FALSE.) exit
        end do
        if (opt .eqv. .FALSE.) exit
      end do

      if (opt) exit
    end do
  end subroutine fix_orbitals_on_nuclei

  subroutine localize_orbitals(P, G, B, F, MO_index)
    !! localizes orbitals in MO_index array
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer :: MO_index(:)
    real(long), dimension(B%N_basis, B%N_basis, 3) :: tdipole_bf  ! bf space
    real(long), dimension(B%N_basis, size(MO_index)) :: LC      ! new coef
    !real(long), dimension(P%N_basis) :: CE      ! new orbital eigenvalues
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    call orbRot_PairSep(LC, F%LC(:, MO_index), tdipole_bf)
    F%LC(:, MO_index(:)) = LC
    ! compute new orbital eigenvalues as expectation value of one particale hamiltonian
  end subroutine localize_orbitals

!>----------------------------------------------------------------------------
!! orbRot_PairSepr : maximizes the measure given by fnOrbLocality_PairSepr
!!   the quantity to be maximized is
!! autho: Kota Hanasaki
!!    Boys' self-extension  \sum_{i} \int d^{3}r1 \int d^{3}r2 phi^{2}_{i}(1) phi^{2}_{i}(2) (r1-r2)**2
!!      or equivalently
!!    Pair separation      \sum_{i<j}  [ (i| \mathbf{r} |i) - (j| \mathbf{r} |j) ]**2
!!                                                             -- summation runs through orbitals you want to localize --
!!       refr.: D. A. Kleier et.al.     J. Chem. Phys. 61  3905 (1974)
!!  LC :    (OUTPUT) output Linear Combination coefficients ... those are not specified in  Imo array is just copied
!!  LCsrc : (INPUT)  input  Linear Combination coefficients
!!  Sp1:    (INPUT)  dimension(Lb,Lb,3) :  AO 1st order moment                <\mu| x_{k} | \nu>
!!                   3 stands for the spatial dimension (x,y,z)
!!
!!  Here   braket notation like   (i| X |j) stands for  C_{i}^{\mu}  <\mu| X | \nu> C_{j}^{\nu}
!!
!!  PARAMETER(1) :: tiny_update : if (locality update) < tiny_update , we stop optimization
!!  PARAMETER(2) :: NstepMax    : # of max iteration
!!----------------------------------------------------------------------------
  subroutine orbRot_PairSep(LC, LCsrc, Sp1)
    implicit none
    Real(Long), intent(in) :: LCsrc(:, :)
    real(long) :: Sp1(size(LCsrc, 1), size(LCsrc, 1), 3)
    Real(Long), intent(inout) ::  LC(size(LCsrc, 1), size(LCsrc, 2))
    real(long) :: LCloc(size(LCsrc, 1), 2), LClocSV(size(LCsrc, 1), 2)
    real(long) :: vTOTini, vTOT, vOld, rII(3), rjj(3), rIJ(3), AA, AB, BB, P, Q, theta, vLoc, vLOCold, dum, &
                  tiny_update = 1.0e-06   !! PARAMETER(1) :: tiny_update : if update< this value, we stop optimization
    integer    :: is, i, j, k, NprtOu, &
                  NstepMax = 50               !! PARAMETER(2) :: NstepMax    : # of max iteration
    integer    :: idxPair(2)
    LC = LCsrc
    !vTOTini = fnOrbLocality_PairSepr( LC,Lb,Sp1,Imo,Nmo)
    vTOTini = fnOrbLocality_PairSepr(LC(:, :), Sp1)
    vTOT = vTOTini


    NprtOu = 0
    do is = 1, NstepMax
      vOld = vTot
      !
      ! We proceed pairwise optimization through (i>j) pair.
      !    for each pair, consider rotation
      !        \psi_{I} =  cos_theta \phi_{i} + sin_theta \phi_{j}
      !        \psi_{J} = -sin_theta \phi_{i} + cos_theta \phi_{j}
      !   to maximize
      !       [ (I| \mathbf{r} |I) - (J| \mathbf{r} |J) ] **2
      !       = [ A**2+4 B**2]/2  + cos_4theta [ A**2 -4 B**2]/2 + 2 sin_4theta [ A.B ]  ----(*)
      !   where
      !        A:=(i|r|i)-(j|r|j)  B:=(i|r|j)       (3-dim vector)
      !
      ! The solution has arbitrarity up to   4\theta = \alpha + 2 N \PI             N \in INTEGER
      !     (because of the positive constant term in Eq.(*), there is no further freedom)
      !                                       \theta = 0.25 \alpha + 0.50 N \PI     N \in INTEGER
      !                                      but this means just rotation around final solutions (I<=>J with \pm signs)
      do i = 1, size(LC, 2)
        forall (k=1:3)
          rii(k) = dot_product(LC(:, i), MATMUL(Sp1(:, :, k), LC(:, i))); 
        end forall
        do j = 1, i - 1
          idxPair(1) = i
          idxPair(2) = j

          !vLOCold= fnOrbLocality_PairSepr( LC,Lb,Sp1,(/Imo(i),Imo(j)/),2 )
          vLOCold = fnOrbLocality_PairSepr(LC(:, idxPair), Sp1)
          forall (k=1:3)
            rjj(k) = dot_product(LC(:, j), MATMUL(Sp1(:, :, k), LC(:, j)))
          end forall
          forall (k=1:3)
            rij(k) = dot_product(LC(:, i), MATMUL(Sp1(:, :, k), LC(:, j)))
          end forall
          AA = sum((rII(:) - rjj(:))**2)
          AB = dot_product(rII(:) - rjj(:), rIJ)
          BB = dot_product(rIJ, rIJ)
          P = (AA - 4*BB)/2
          Q = 2*AB
          theta = 0.25_long*atan2(Q, P); 
          !vloc= fnVrot(LC,LCloc,Lb,Sp1,theta,Imo(i),Imo(j))
          vloc = fnVrot(LC(:, idxPair(:)), Sp1, theta, LCloc)
          !write(6,*) 'rot',i,j,theta*180./PI

          !
          ! Here we naively confirm that this pairwise optimization contributes the total locality or not.
          ! XXX TODO : Obviously, this is not efficient. One can store pairwise
          !               [ (i| \mathbf{r} |i) - (j| \mathbf{r} |j) ]**2
          !            to evaluate total locality more efficently.
          !
          if (vloc .gt. vLOCold) then
            LClocSV(:, 1) = LC(:, i)
            LClocSV(:, 2) = LC(:, j)
            LC(:, i) = LCloc(:, 1)
            LC(:, j) = LCloc(:, 2)
            !dum= fnOrbLocality_PairSepr( LC,Lb,Sp1,Imo,Nmo)
            dum = fnOrbLocality_PairSepr(LC(:, :), Sp1)
            if (dum .gt. vTot) then; 
              vTot = dum
            else
              if (NprtOu .lt. 20) THEN; 
                Write (*, *) "#orbRot_PS:REJECT:", is, i, j, dum, dum - vTot, vloc - vLOCold; NprtOu = NprtOu + 1; 
              end if
              LC(:, i) = LClocSV(:, 1)
              LC(:, j) = LClocSV(:, 2)
            end if
          end if
        end do
      end do
      !write(6,*) is,vTot,vTot-vOld
      if (dabs(vTot - vOld) .lt. tiny_update) exit
      dum = fnOrbLocality_PairSepr(LC(:, :), Sp1)
    end do
    if (dabs(vTot - vOld) .gt. tiny_update) then
      write (*, *) 'Warning: Localization failed!'
    end if
  contains
    !--------------------------------------------------------
    ! (internal) function fnVrot :: calculates pair localization measure using rotated LC
    !--------------------------------------------------------
    real(long) function fnVrot(LCorg, tdipole, theta, LCnew) result(vRot)
      implicit none
      real(Long), intent(in) :: LCorg(:, :)
      real(long), intent(in) :: tdipole(size(LCorg, 1), size(LCorg, 1), 3)
      real(long), intent(in) :: theta
      real(Long), intent(inout) :: LCnew(size(LCorg, 1), size(LCorg, 2))
      real(long) :: co, sn
      if (size(LCorg, 2) .ne. &
          2) stop 'wrong array length in fnVrot'
      co = cos(theta)
      sn = sin(theta)
      LCnew(:, 1) = LCorg(:, 1)*co + LCorg(:, 2)*sn
      LCnew(:, 2) = -LCorg(:, 1)*sn + LCorg(:, 2)*co
      vRot = fnOrbLocality_PairSepr(LCnew, tdipole)
    end function
  end subroutine

!>----------------------------------------------------------------------------
!! fnOrbLocality_PairSepr : currently adopted measure of orbital locality : \sum_{<i,j>} ( <i|r|i>-<j|r|j> )**2
!! author: Kota Hanasaki
!!----------------------------------------------------------------------------
  real(long) function fnOrbLocality_PairSepr(LC, tdipole) result(ret)
    implicit none
    real(long), intent(in) :: LC(:, :)
    real(long), intent(in) :: tdipole(size(LC, 1), size(LC, 1), 3)
    real(long) :: Ri(3), rJ(3)
    integer :: i, j, k
    ret = 0
    do i = 1, size(LC, 2); 
      forall (k=1:3)
        Ri(k) = Dot_product(LC(:, i), MATMUL(tdipole(:, :, k), LC(:, i))); 
      end forall
      do j = 1, i - 1; 
        forall (k=1:3)
          Rj(k) = Dot_product(LC(:, j), MATMUL(tdipole(:, :, k), LC(:, j))); 
        end forall
        ret = ret + sum((Ri(:) - Rj(:))**2)
      end do
    end do
  end function

end module localization
