!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module matrix
  !! author: Sang-Kil Son
  !!
  !! Matrix Module
  !!   programmed by Son, Sang-Kil in Feb. 2004
  !!   for PHSX815 : Computational Physics
  !!
  !! Ref)
  !!   Press et al., "Numerical Recipes in Fortran 77"
  !!   Press et al., "Numerical Recipes in Fortran 90"
  !!
  !! Main data types
  !! - LU_matrix : LU-decomposed matrix
  !!   - F LU_decompose(A(N,N)) : create LU matrix by LU-decomposition of A
  !!   - F LU_solve(LU, b(N)) : solve a linear equation for b with LU
  !!   - F LU_inverse(LU) : inverse matrix of LU
  !!   - F LU_determinant(LU) : determinant of LU
  !!   - F LU_lower(LU) : return L matrix of LU
  !!   - F LU_upper(LU) : return U matrix of LU
  !!
  !!
  !! Auxiliary functions and subroutines
  !! - SR print_matrix(A(N,N)) : print out a matrix
  !! - SR print_vector(x(N)) : print out a vector
  use constants
  implicit none
  !>
  !! Data type: LU_matrix
  !!   it contains the dimension of the matrix, N
  !!   and LU-decomposed matrix, LU(N,N)
  !!   also, permutation(N) and the sign for determinant, d
  !!
  type :: LU_matrix
    integer :: n
    ! Why pointer instead of allocatable? Ludger
    real(long), dimension(:, :), pointer :: LU
    integer, dimension(:), pointer :: permutation
    integer :: d
  end type LU_matrix
  !
  !> Operators for efficient matrix multiplication with LAPACK, but
  !! better readibility. (It is assumed that the dimensionality fits.)
  !!
  !! multiply two real square matrices or a square matrix times a vector
  !!          A*B= A .m. B             or   A*v= A .m. v
  !!
  interface operator(.m.)
    module procedure mult_real_matrix_matrix
    module procedure mult_real_matrix_vector
  end interface ! operator(.m.)
  !
  !> multiply two real square matrices, the second is transposed
  !!          A*(B**T)= A .mt. B
  !!
  interface operator(.mt.)
    module procedure mult_real_matrix_trans_matrix
  end interface ! operator(.mt.)
  !
  !> multiply two real square matrices, the first is transposed
  !!          (A**T)*B= A .tm. B
  !!
  interface operator(.tm.)
    module procedure mult_real_trans_matrix_matrix
    module procedure mult_real_trans_matrix_vector
  end interface ! operator(.tm.)
  !
  !> substract two real square matrices element wise
  !!          A-B = A .sub. B
  !! This functionality is provided by f90. Element wise, arrays
  !! (of same dimensionality) can be added, substracted, multiplied and divided.
  !! sub_real_matrix_matrix is not a LAPACK feature. (Michael)
  !!
  interface operator(.sub.)
    module procedure sub_real_matrix_matrix
  end interface ! operator(.sub.)
  private
  public :: operator(.mt.), operator(.m.), operator(.tm.), operator(.sub.), &
    LU_decompose, LU_matrix, LU_solve, LU_inverse, LU_determinant, LU_matrix_purge, &
    matTrace, matrix_determinant, &
    solve_AxB, print_matrix, matTraceProduct
contains

  subroutine LU_matrix_purge(M)
    !! purges LU_matrix structure
    type(LU_matrix) :: M

    deallocate (M%LU)
    deallocate (M%permutation)
    nullify (M%LU)
    nullify (M%permutation)
  end subroutine LU_matrix_purge

  function LU_decompose(A) result(M)    
    !! Carry out LU-decomposition for A and put the decomposed matrix into M
    !! implementation of Crout's method with partial pivoting.
    type(LU_matrix) :: M
    real(long), dimension(:, :), intent(in) :: A !! n*N matrix
    real(long), dimension(size(A, dim=1)) :: scaling_factor, temp_row
    real(long) :: a_max, temp
    integer :: n, i, j, i_max
    !
    ! Initialize
    !   A is a n * n matrix.
    !   # rows = size(A, dim=1)
    !   # columns = size(A, dim=2)
    ! consequently, size(A,dim=1) = size(A,dim=2)
    !
    n = size(A, dim=1)
    M%n = n
    M%d = 1
    allocate (M%LU(n, n), M%permutation(n))
    M%LU = A
    !
    ! Implicit scaling procedure
    !
    do i = 1, n
      a_max = maxval(abs(A(i, :)))
      if (a_max .lt. EPS) stop "singular matrix in LU_decompose"
      scaling_factor(i) = ONE/a_max
    end do
    !
    ! Crout's method
    ! The following equations are exactly same as Press's procedure
    !
    do j = 1, n
      !
      ! equation (2.3.12) except for i=j
      !
      do i = 1, j - 1
        M%LU(i, j) = M%LU(i, j) - sum(M%LU(i, 1:i - 1)*M%LU(1:i - 1, j))
      end do
      !
      ! equation (2.3.12) for i=j and equation (2.3.13)
      !
      a_max = ZERO
      i_max = 1
      do i = j, n
        M%LU(i, j) = M%LU(i, j) - sum(M%LU(i, 1:j - 1)*M%LU(1:j - 1, j))
        !
        ! find the maximum element and store its point (i_max)
        ! That point will be the pivot element.
        !
        temp = abs(M%LU(i, j))*scaling_factor(i)
        if (temp .gt. a_max) then
          i_max = i
          a_max = temp
        end if
      end do
      !
      ! if needed to interchange rows
      !
      if (j .ne. i_max) then
        temp_row = M%LU(j, :)              ! swap j-th row and i_max-th row
        M%LU(j, :) = M%LU(i_max, :)         ! A(j,:) : j-th row
        M%LU(i_max, :) = temp_row
        M%d = -M%d
        scaling_factor(i_max) = scaling_factor(j)
      end if
      !
      ! divide by the pivot element
      !
      M%permutation(j) = i_max
      if (abs(M%LU(j, j)) .lt. EPS) M%LU(j, j) = EPS    ! prevent singular case
      if (j .ne. n) M%LU(j + 1:n, j) = M%LU(j + 1:n, j)/M%LU(j, j)
    end do
  end function LU_decompose

  function LU_solve(M, b) result(x)
    !! Solve a linear equation for b with LU-decomposed matrix M
    !! Return the solution for b as a new vector x
    !!   b : row vector, not column vector, that is, just 1D array of n
    !!
    type(LU_matrix), intent(in) :: M
    real(long), dimension(:), intent(in) :: b
    real(long), dimension(M%n) :: x
    real(long) :: temp
    integer :: i, j, n
    n = M%n
    !
    ! rearrange b(n) with the permutation
    !
    x = b
    do i = 1, n
      j = M%permutation(i)
      if (i .ne. j) then
        temp = x(i)
        x(i) = x(j)
        x(j) = temp
      end if
    end do
    !
    ! forward substitution using Press's eq (2.3.6) and alpha_ii = 1
    !
    do i = 1, n
      x(i) = x(i) - sum(M%LU(i, 1:i - 1)*x(1:i - 1))
    end do
    !
    ! back substitution using Press's eq (2.3.7)
    !
    do i = n, 1, -1
      x(i) = (x(i) - sum(M%LU(i, i + 1:n)*x(i + 1:n)))/M%LU(i, i)
    end do
  end function LU_solve

  function LU_inverse(M) result(X)
    !! Return the inverse matrix of LU-decomposed matrix M
    !!   using F solve(M, b)
    !!
    type(LU_matrix), intent(in) :: M
    real(long), dimension(M%n, M%n) :: X
    real(long), dimension(M%n) :: b
    integer :: i, n
    n = M%n
    !
    ! set up identity matrix and solve n linear equations with identity.
    !
    do i = 1, n
      b = ZERO; b(i) = ONE        ! b is the i-th column vector of identity.
      X(:, i) = LU_solve(M, b)     ! A(:,i) is the i-th column vector of A
    end do
  end function LU_inverse

  function LU_determinant(M) result(d)
    !! returns the determinant value of LU-decomposed matrix M
    !!
    type(LU_matrix), intent(in) :: M
    real(long) :: d
    integer :: i
    !
    ! The determinant can be solved by a product of all diagonal elements.
    !
    d = dble(M%d)*product((/(M%LU(i, i), i=1, M%n)/))
  end function LU_determinant

  subroutine print_matrix(A)
    !! Print arbitrary N*M matrix
    !!
    real(long), dimension(:, :), intent(in) :: A
    integer :: i
    print "('Matrix',i4,i4)", size(A, 1), size(A, 2)
    do i = 1, size(A, dim=1)
      print "(t8,100f10.5)", A(i, :)
    end do
  end subroutine print_matrix

  subroutine print_matrixf(A)
    !! Print arbitrary N*M matrix
    !!
    real, dimension(:, :), intent(in) :: A
    integer :: i
    do i = 1, size(A, dim=1)
      print "(t8,100f10.5)", A(i, :)
    end do
  end subroutine print_matrixf

  subroutine print_vector(x)
    !! Print arbitrary N-vector as if a column vector
    !!   x : row vector, not column vector, that is, just 1D array of n
    !!
    real(long), dimension(:), intent(in) :: x
    integer :: i
    do i = 1, size(x)
      print "(t8,f10.5)", x(i)
    end do
  end subroutine print_vector

  subroutine print_vectorf(x)
    !! Print arbitrary N-vector as if a column vector
    !!   x : row vector, not column vector, that is, just 1D array of n
    !!
    real, dimension(:), intent(in) :: x
    integer :: i
    do i = 1, size(x)
      print "(t8,f10.5)", x(i)
    end do
  end subroutine print_vectorf

  function mult_real_matrix_matrix(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:, :)
    integer :: N, M, K
    real(long) :: C(size(A, dim=1), size(B, dim=2))
    real(long) :: alpha, beta
    alpha = ONE; beta = ZERO
    M = size(A, dim=1)
    N = size(B, dim=2)
    K = size(A, dim=2)
    !if (size(A,2) /= size(B,1) )&
    !     stop 'error in matrix multp'
    call DGEMM('N', 'N', M, N, K, alpha, A, M, B, K, beta, C, M)
  end function mult_real_matrix_matrix

  function mult_real_matrix_trans_matrix(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:, :)
    integer :: M, N, K
    real(long) :: C(size(A, dim=1), size(B, dim=1))
    real(long) :: alpha, beta
    alpha = ONE; beta = ZERO
    M = size(A, dim=1)
    N = size(B, dim=1)
    K = size(A, dim=2)
    !if (size(A,2) /= size(B,2) )&
    !     stop 'error in matrix multp'
    call DGEMM('N', 'T', M, N, K, alpha, A, M, B, N, beta, C, M)
  end function mult_real_matrix_trans_matrix

  function mult_real_trans_matrix_matrix(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:, :)
    integer :: M, N, K
    real(long) :: C(size(A, dim=2), size(B, dim=2))
    real(long) :: alpha, beta
    alpha = ONE; beta = ZERO
    M = size(A, dim=2)
    N = size(B, dim=2)
    K = size(A, dim=1)
    !if (size(A,1) /= size(B,1) )&
    !     stop 'error in matrix multp'
    call DGEMM('T', 'N', M, N, K, alpha, A, K, B, K, beta, C, M)
  end function mult_real_trans_matrix_matrix

  function mult_real_matrix_vector(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:)
    integer :: M, N
    real(long) :: C(size(A, dim=1))
    real(long) :: alpha, beta
    alpha = ONE; beta = ZERO
    M = size(A, dim=1)
    N = size(A, dim=2)
    if (size(A, 2) .ne. size(B, 1) ) &
      stop 'error in matrix vec. multp'
    call DGEMV('N', M, N, alpha, A, M, B, 1, beta, C, 1)
  end function mult_real_matrix_vector

  function mult_real_trans_matrix_vector(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:)
    integer :: M, N
    real(long) :: C(size(A, dim=2))
    real(long) :: alpha, beta
    alpha = ONE; beta = ZERO
    M = size(A, dim=1)
    N = size(A, dim=2)
    if (size(A, 1) .ne. size(B, 1) ) &
      stop 'error in matrix^t vec. multp'
    call DGEMV('T', M, N, alpha, A, M, B, 1, beta, C, 1)
  end function mult_real_trans_matrix_vector

  function matrix_determinant(A) result(det)
    !! returns the determinant of square matrix A
    !! slightly redundant to LU_determiant but allows also to return zero as a determinant
    real(long), intent(in) :: A(:, :)
    real(long) :: A_copy(size(A, 1), size(A, 2))
    real(long) :: det
    integer :: info, lda, m, n
    integer :: ipiv(min(size(A, 1), size(A, 2)))
    integer :: i
    A_copy(:, :) = A(:, :)
    m = size(A, 1)
    n = size(A, 2)
    lda = m
    if (m .ne. n) then
      det = ZERO
    else if (m .eq. 0) then
      det = ONE
    else
      call DGETRF(m, n, A_copy, lda, ipiv, info)
      if (info .gt. 0) then
        det = ZERO
        return
      else
        det = ONE
        do i = 1, size(A, 1)
          if (ipiv(i) .ne. i) then
            det = -det*A_copy(i, i)
          else
            det = det*A_copy(i, i)
          end if
        end do
      end if
    end if
  end function matrix_determinant

  function sub_real_matrix_matrix(A, B) result(C)
    real(long), intent(in) :: A(:, :), B(:, :)
    real(long) :: C(size(A, dim=1), size(A, dim=1))
    integer :: N
    integer :: i, j
    N = size(A, dim=1)

    do i = 1, N
      do j = 1, N
        C(j, i) = A(j, i) - B(j, i)
      end do ! j
    end do ! i
  end function sub_real_matrix_matrix

  subroutine solve_AxB(A, B, D, ierr, yes_verbose)
    !! solves AxB=D for D
    real(long), intent(in) ::  A(:, :) !! matrix A of shape NxN
    real(long), intent(in) ::  B(:) !! vector B of len N
    real(long), intent(out) :: D(size(B,1)) !! result vector of len N
    integer, intent(out) :: ierr !! output status -1 is solution could not be found (A is singular)
    logical, intent(in), optional :: yes_verbose !! print message if A is singular
    logical :: verbose    
    integer :: N, NRHS
    real(long) :: C(size(A,1), size(A,2))
    integer :: IPIV(size(A,2))
    integer :: INFO
    verbose = .False.
    if(present(yes_verbose)) verbose = yes_verbose
    N = size(A,2)
    ierr = 1
    NRHS = 1
    C(:, :) = A(:, :)
    D(:) = B(:)
    call dgesv(N, NRHS, C, N, IPIV, D, N, INFO)
    if (INFO .gt. 0 ) then
      if(verbose)&
        write (STDOUT, *) " A is singular "
      ierr = -1
    end if
    return
  end subroutine solve_AxB

  function matTrace(A) result(B)
    !! returns the trace of matrix A
    integer :: i, M
    real(long), intent(in) :: A(:, :)
    real(long) :: B
    M = size(A, dim=1)

    B = ZERO
    do i = 1, M
      B = B + A(i, i)
    end do ! i
    return
  end function ! MatTrace

  function matTraceProduct(A, B) result(trace)
    !! computes the trace of the matrix A * B
    !! first dimension of A has to match second dimension of B
    !! second dimension of A has to match first dimension of B
    real(long), intent(in) :: A(:, :)
    real(long), intent(in) :: B(:, :)
    real(long) :: trace
    integer :: i
    if(size(A,1) .ne. size(B,2)) stop 'wrong matrix dimensions'
    if(size(A,2) .ne. size(B,1)) stop 'wrong matrix dimensions'
    trace = ZERO
    do i=1, size(A,1)
      trace = trace + dot_product(A(i,:), B(:,i))
    end do
    return
  end function matTraceProduct

end module matrix
