!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!
module mergesort
  !! author: Ludger Inhester
  !!
  !! module implementing merge sort algorithm  
  use constants
  implicit none

  interface merge_sort
    !! sorts array argument
    module procedure merge_sort_int
  end interface merge_sort

  interface merge_argsort
    !! returns the indices that sorts the array
    module procedure merge_argsort_int, merge_argsort_double
  end interface merge_argsort
  private
  public :: merge_sort_int, merge_argsort_int, merge_argsort_double, merge_argsort, merge_sort 
  contains

  subroutine merge_sort_int(A)
    !! sorts integer array A
    implicit none
    integer, intent(inout) :: A(:)
    integer :: N
    integer, allocatable :: W(:)
    N = size(A)
    allocate (W(N))
    call merge_sort_int_sub(A, 1, N, W)
    deallocate (W)
  end subroutine merge_sort_int

  recursive subroutine merge_sort_int_sub(A, Li, Lf, W, idx)
    !! sorts integer array A
    integer, dimension(*) :: A, W
    integer, intent(in) :: Li, Lf
    integer, dimension(*), optional :: idx
    integer :: M
    integer :: N
    N = Lf - Li + 1
    if (N .le. 1) then
      return
    end if
    !write(6,*) 'sorting',A(Li:Lf)
    if (N .eq. 2) then
      if (present(idx)) then
        if (A(idx(Li)) .gt. A(idx(Lf))) then
          M = idx(Li)
          idx(Li) = idx(Lf)
          idx(Lf) = M
        end if
        return
      else
        !write(6,*) 'sorting', A(Li), A(Lf)
        if (A(Li) .gt. A(Lf)) then
          !write(6,*) 'swapping',Li,Lf
          M = A(Li)
          A(Li) = A(Lf)
          A(Lf) = M
        end if
        !write(6,*) 'A sorted',A(Li:Lf)
        return
      end if
    end if
    ! split up into two parts, sort individual parts and merge
    M = (Li + Lf)/2
    if (present(idx)) then
      call merge_sort_int_sub(A, Li, M, W, idx)
      call merge_sort_int_sub(A, M + 1, Lf, W, idx)
      call merge_arg_int(A, Li, M, M + 1, Lf, W, idx)
    else
      call merge_sort_int_sub(A, Li, M, W)
      call merge_sort_int_sub(A, M + 1, Lf, W)
      call merge_int(A, Li, M, M + 1, Lf, W)
    end if
  end subroutine merge_sort_int_sub


  subroutine merge_int(A, Li, Lf, Ri, Rf, W)
    !! merger step for sorting integer array A
    integer, dimension(*) :: A, W
    integer, intent(in) :: Li, Lf, Ri, Rf
    integer N, l, r, d
    N = Lf - Li + Rf - Ri + 1
    l = Li
    r = Ri
    d = 1
    !write(6,*) 'merging',A(Li:Lf)
    !write(6,*) 'and',A(Ri:Rf)
    do while (l .le. Lf .OR. r .le. Rf)
      do while (l .le. Lf .AND. (r .gt. Rf .OR. A(l) .le. A(r)))
        W(d) = A(l)
        l = l + 1
        d = d + 1
      end do
      do while (r .le. Rf .AND. (l .gt. Lf .OR. A(l) .gt. A(r)))
        W(d) = A(r)
        r = r + 1
        d = d + 1
      end do
    end do
    !write(6,*) 'res', W(1:N+1)
    A(Li:Li + N) = W(1:N + 1)
  end subroutine merge_int

  subroutine merge_arg_int(A, Li, Lf, Ri, Rf, W, idx)
    !! merger step for argument-sorting integer array A
    implicit none
    integer, dimension(*) :: A, W
    integer, dimension(*) :: idx
    integer, intent(in) :: Li, Lf, Ri, Rf
    integer N, l, r, d
    N = Lf - Li + Rf - Ri + 1
    l = Li
    r = Ri
    d = 1
    do while (l .le. Lf .OR. r .le. Rf)
      do while (l .le. Lf)
        if (r .gt. Rf) then
          W(d) = idx(l)
          l = l + 1
          d = d + 1
        else if (A(idx(l)) .le. A(idx(r))) then
          W(d) = idx(l)
          l = l + 1
          d = d + 1
        else
          exit
        end if
        if (l .gt. lf) exit
      end do
      do while (r .le. Rf)
        if (l .gt. Lf) then
          W(d) = idx(r)
          r = r + 1
          d = d + 1
        else if (A(idx(l)) .gt. A(idx(r))) then
          W(d) = idx(r)
          r = r + 1
          d = d + 1
        else
          exit
        end if
        if (r .gt. rf) exit
      end do
    end do
    idx(Li:Li + N) = W(1:N + 1)
  end subroutine merge_arg_int

  subroutine merge_argsort_int(A, idx)
    !! arg sorts integer array A
    implicit none
    integer, intent(inout) :: A(:)
    integer, intent(inout) :: idx(size(A, 1))
    integer :: N, i
    integer, allocatable :: W(:)
    N = size(A)
    allocate (W(N))
    idx = (/(i, i=1, N)/)
    call merge_sort_int_sub(A, 1, N, W, idx)
    deallocate (W)
  end subroutine merge_argsort_int

  subroutine merge_argsort_double(A, idx)
    !! arg sorts integer array A
    implicit none
    real(long), intent(inout) :: A(:)
    integer, intent(inout) :: idx(size(A, 1))
    integer :: N, i
    integer, allocatable :: W(:)
    N = size(A)
    allocate (W(N))
    idx = (/(i, i=1, size(A))/)
    call merge_argsort_double_sub(A, 1, N, W, idx)
    deallocate (W)
  end subroutine merge_argsort_double

  recursive subroutine merge_argsort_double_sub(A, Li, Lf, W, idx)
    !! sorts integer array A
    implicit none
    real(long), dimension(*) :: A
    integer, dimension(*) :: W
    integer, intent(in) :: Li, Lf
    integer, dimension(*) :: idx
    integer :: M
    integer :: N
    N = Lf - Li + 1
    if (N .le. 1) then
      return
    end if
    !write(6,*) 'sorting',A(Li:Lf)
    if (N .eq. 2) then
      if (A(idx(Li)) .gt. A(idx(Lf))) then
        M = idx(Li)
        idx(Li) = idx(Lf)
        idx(Lf) = M
      end if
      return
    end if
    ! split up into two parts, sort individual parts and merge
    M = (Li + Lf)/2
    call merge_argsort_double_sub(A, Li, M, W, idx)
    call merge_argsort_double_sub(A, M + 1, Lf, W, idx)
    call merge_arg_double(A, Li, M, M + 1, Lf, W, idx)
  end subroutine merge_argsort_double_sub

  subroutine merge_arg_double(A, Li, Lf, Ri, Rf, W, idx)
    !! merger step for argument-sorting integer array A
    implicit none
    real(long), dimension(*) :: A
    integer, dimension(*) :: idx, W
    integer, intent(in) :: Li, Lf, Ri, Rf
    integer N, l, r, d
    N = Lf - Li + Rf - Ri + 1
    l = Li
    r = Ri
    d = 1
    !write(6,*) 'merging',A(idx(Li:Lf))
    !write(6,*) 'and',A(idx(Ri:Rf))
    do while (l .le. Lf .OR. r .le. Rf)
      do while (l .le. Lf)
        if (r .gt. Rf) then
          W(d) = idx(l)
          l = l + 1
          d = d + 1
        else if (A(idx(l)) .le. A(idx(r))) then
          W(d) = idx(l)
          l = l + 1
          d = d + 1
        else
          exit
        end if
        if (l .gt. lf) exit
      end do
      do while (r .le. Rf)
        if (l .gt. Lf) then
          W(d) = idx(r)
          r = r + 1
          d = d + 1
        else if (A(idx(l)) .gt. A(idx(r))) then
          W(d) = idx(r)
          r = r + 1
          d = d + 1
        else
          exit
        end if
        if (r .gt. rf) exit
      end do
    end do
    idx(Li:Li + N) = W(1:N + 1)
  end subroutine merge_arg_double

end module mergesort
