!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module molden
  !! author: Ludger Inhester
  !!
  !!   This module is for writing out molden files
  !!   written by Ludger Inhester in 2021 at DESY CFEL
  use constants
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use wavefunction
  use cart2sphh
  use matrix

  implicit none
  private
  public :: write_molden_file
contains
  subroutine write_molden_file(P, A, B, F, filename, cartesian)
    !! writes a molden file containing the basis set definition and the linear combination of orbitals
    real(long), external :: CINTgto_norm
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    character(len=*), intent(in) :: filename !! output filename
    logical, intent(in), optional :: cartesian !! write a molden file in cartesian basis functions?
    integer :: sh, bf, at, l, m, n_prim, i, mo, bfc
    character(len=5) :: strl
    real(long) :: alpha, coef
    logical :: yes_cartesian
    real(long), allocatable :: sphh2cart_matrix(:, :)
    integer :: spin, nspin
    real(long) :: orbital(size(F%LC,1))
    yes_cartesian = .True.
    !yes_cartesian = .False.
    if (present(cartesian)) &
      yes_cartesian = cartesian

    if (.NOT. P%yes_gaussian) then
      stop "molden output only works with gaussian orbital basis set!"
    end if

    write (6, '("# ",a,x,a)') "writing molden file", trim(filename)
    open (unit=66, file=filename, status='unknown', ACTION='write')
    write (66, '("[Molden Format]")')
    write (66, '("[Atoms] AU")')
    do at = 1, P%N_nuc
      if (A(at)%charge .eq. 0) cycle ! skip point charges
      write (66, '(a,i8,i8,3es15.5)') P%nuc_name(at), at, P%charge_nuc(at), A(at)%x, A(at)%y, A(at)%z
    end do
    ! we only use spherical harmonic-type orbitals
    if (.not. yes_cartesian) then
      write (66, '("[5D]")')
      write (66, '("[7F]")')
      write (66, '("[9G]")')
    end if
    write (66, '("[GTO]")')
    bf = 1
    do sh = 1, size(B%bas, 2) ! loop over shells
      at = B%bas(1, sh) + 1
      l = B%bas(2, sh)    ! l: the orbital angular momentum quantum number
      n_prim = B%bas(3, sh) ! number of primitives
      if (B%first_bf_on_atom(at) .eq. bf) then
        if (bf .ne. 1 ) write (66, '()')
        write (66, '(i4,x,i1)') at, 0
      end if
      select case (l)
      case (0); strl = "s"
      case (1); strl = "p"
      case (2); strl = "d"
      case (3); strl = "f"
      case (4); strl = "g"
      case (5); strl = "h"
      case (6); strl = "i"
      case default; write (strl, '("i2")') l
      end select
      write (66, '(a2,x,i4,x,f5.2)') strl, n_prim, 1.0
      do i = 1, n_prim
        alpha = B%env(B%bas(6, sh) + i)
        coef = B%env(B%bas(7, sh) + i)/CINTgto_norm(l, alpha)
        write (66, '(es17.10,x,es17.10)') alpha, coef
      end do
      bf = bf + 2 * l + 1
    end do! sh
    write (66, '()')

    write (66, '("[MO]")')
    nspin = 1
    if (P%yes_unrestricted) &
      nspin = 2
    do spin = 1, nspin
      do mo = 1, size(F%LC, 2) ! loop over mos

        bf = 1
        bfc = 1
        if (P%yes_unrestricted) then
          if (spin==1) then
            write (66, '(" Ene= ",f10.4)') F%CE_a(mo)
            write (66, '(" Spin= ",a)') "Alpha"
            write (66, '(" Occup= ",f10.6)') F%f_occ_a(mo)
            orbital = F%LC_a(:, mo)
          else
            write (66, '(" Ene= ",f10.4)') F%CE_b(mo)
            write (66, '(" Spin= ",a)') "Beta"
            write (66, '(" Occup= ",f10.6)') F%f_occ_b(mo)
            orbital = F%LC_b(:, mo)
          end if
        else
          write (66, '(" Ene= ",f10.4)') F%CE(mo)
          write (66, '(" Spin= ",a)') "Alpha"
          write (66, '(" Occup= ",f10.6)') F%f_occ(mo)
          orbital = F%LC(:, mo)
        end if
        do sh = 1, size(B%bas, 2) ! loop over shells
          at = B%bas(1, sh) + 1
          l = B%bas(2, sh)    ! l: the orbital angular momentum quantum number
          if (yes_cartesian) then
            allocate(sphh2cart_matrix( (l + 1) * (l + 2) / 2, 2 * l + 1))
            call sphh2cart(sphh2cart_matrix, l)
          end if
          select case (l)
          case (0)
            if (yes_cartesian) then
              do i= 1, (l + 1) * (l + 2) / 2
                write (66, '(i4,x,es17.10)') bfc + i - 1, sum(sphh2cart_matrix(i,:) * orbital(bf:bf + (2 * l + 1) - 1) )
              end do
            else  
              write (66, '(i4,x,es17.10)') bf, orbital(bf) ! s
            end if
          case (1)
            if (yes_cartesian) then
              do i= 1, (l + 1) * (l + 2) / 2
                write (66, '(i4,x,es17.10)') bfc + i - 1, sum(sphh2cart_matrix(i,:) * orbital(bf:bf + (2 * l + 1) - 1) )
              end do
            else
              write (66, '(i4,x,es17.10)') bf + 0, -orbital(bf + 2) ! x = p+
              write (66, '(i4,x,es17.10)') bf + 1, -orbital(bf + 0) ! y = p-
              write (66, '(i4,x,es17.10)') bf + 2,  orbital(bf + 1) ! z = p0
            end if
          case (2)
            if (yes_cartesian) then
              ! convert to molden order
              ! xx xy xz yy yz zz -> xx yy zz xy xz yz
              write (66, '(i4,x,es17.10)') bfc + 0, sum(sphh2cart_matrix(1,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dxx
              write (66, '(i4,x,es17.10)') bfc + 1, sum(sphh2cart_matrix(4,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dyy
              write (66, '(i4,x,es17.10)') bfc + 2, sum(sphh2cart_matrix(6,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dzz
              write (66, '(i4,x,es17.10)') bfc + 3, sum(sphh2cart_matrix(2,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dxy
              write (66, '(i4,x,es17.10)') bfc + 4, sum(sphh2cart_matrix(3,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dxz
              write (66, '(i4,x,es17.10)') bfc + 5, sum(sphh2cart_matrix(5,:) * orbital(bf:bf + (2 * l + 1) - 1) ) ! dyz
            else
              write (66, '(i4,x,es17.10)') bf + 0,  orbital(bf + 2) ! d0
              write (66, '(i4,x,es17.10)') bf + 1, -orbital(bf + 3) ! d+1
              write (66, '(i4,x,es17.10)') bf + 2, -orbital(bf + 1) ! d-1
              write (66, '(i4,x,es17.10)') bf + 3,  orbital(bf + 4) ! d+2
              write (66, '(i4,x,es17.10)') bf + 4,  orbital(bf + 0) ! d-2
            end if
          case (3)
            if (yes_cartesian) then
              ! convert to molden order
              ! xxx xxy xxz xyy xyz xzz yyy yyz yzz zzz -> xxx, yyy, zzz, xyy, xxy, xxz, xzz, yzz, yyz, xyz
              write (66, '(i4,x,es17.10)') bfc + 0, sum(sphh2cart_matrix( 1, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxxx
              write (66, '(i4,x,es17.10)') bfc + 1, sum(sphh2cart_matrix( 7, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fyyy
              write (66, '(i4,x,es17.10)') bfc + 2, sum(sphh2cart_matrix(10, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fzzz
              write (66, '(i4,x,es17.10)') bfc + 3, sum(sphh2cart_matrix( 4, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxyy
              write (66, '(i4,x,es17.10)') bfc + 4, sum(sphh2cart_matrix( 2, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxxy
              write (66, '(i4,x,es17.10)') bfc + 5, sum(sphh2cart_matrix( 3, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxxz
              write (66, '(i4,x,es17.10)') bfc + 6, sum(sphh2cart_matrix( 6, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxzz
              write (66, '(i4,x,es17.10)') bfc + 7, sum(sphh2cart_matrix( 9, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fyzz
              write (66, '(i4,x,es17.10)') bfc + 8, sum(sphh2cart_matrix( 8, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fyyz
              write (66, '(i4,x,es17.10)') bfc + 9, sum(sphh2cart_matrix( 5, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! fxyz
            else
              write (66, '(i4,x,es17.10)') bf + 0,  orbital(bf + 3) ! f0
              write (66, '(i4,x,es17.10)') bf + 1, -orbital(bf + 4) ! f+1
              write (66, '(i4,x,es17.10)') bf + 2, -orbital(bf + 2) ! f-1
              write (66, '(i4,x,es17.10)') bf + 3,  orbital(bf + 5) ! f+2
              write (66, '(i4,x,es17.10)') bf + 4,  orbital(bf + 1) ! f-2
              write (66, '(i4,x,es17.10)') bf + 5, -orbital(bf + 6) ! f+3
              write (66, '(i4,x,es17.10)') bf + 6, -orbital(bf + 0) ! f-3
            end if
          case (4)
            if (yes_cartesian) then
              ! convert to molden order
              !    1    2    3    4    5    6    7    8    9   10   11   12   13   14   15
              ! xxxx xxxy xxxz xxyy xxyz xxzz xyyy xyyz xyzz xzzz yyyy yyyz yyzz yzzz zzzz
              ! -> 
              ! xxxx yyyy zzzz xxxy xxxz yyyx yyyz zzzx zzzy xxyy xxzz yyzz xxyz yyxz zzxy
              write (66, '(i4,x,es17.10)') bfc + 0, sum(sphh2cart_matrix( 1, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxxx
              write (66, '(i4,x,es17.10)') bfc + 1, sum(sphh2cart_matrix(11, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gyyyy
              write (66, '(i4,x,es17.10)') bfc + 2, sum(sphh2cart_matrix(15, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gzzzz
              write (66, '(i4,x,es17.10)') bfc + 3, sum(sphh2cart_matrix( 2, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxxy
              write (66, '(i4,x,es17.10)') bfc + 4, sum(sphh2cart_matrix( 3, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxxz
              write (66, '(i4,x,es17.10)') bfc + 5, sum(sphh2cart_matrix( 7, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gyyyx
              write (66, '(i4,x,es17.10)') bfc + 6, sum(sphh2cart_matrix(12, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gyyyz
              write (66, '(i4,x,es17.10)') bfc + 7, sum(sphh2cart_matrix(10, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gzzzx
              write (66, '(i4,x,es17.10)') bfc + 8, sum(sphh2cart_matrix(14, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gzzzy
              write (66, '(i4,x,es17.10)') bfc + 9, sum(sphh2cart_matrix( 4, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxyy
              write (66, '(i4,x,es17.10)') bfc +10, sum(sphh2cart_matrix( 6, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxzz
              write (66, '(i4,x,es17.10)') bfc +11, sum(sphh2cart_matrix(13, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gyyzz
              write (66, '(i4,x,es17.10)') bfc +12, sum(sphh2cart_matrix( 5, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gxxyz
              write (66, '(i4,x,es17.10)') bfc +13, sum(sphh2cart_matrix( 8, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gyyxz
              write (66, '(i4,x,es17.10)') bfc +14, sum(sphh2cart_matrix( 9, :) * orbital(bf:bf + (2 * l + 1) - 1) ) ! gzzxy
            else
              write (66, '(i4,x,es17.10)') bf + 0,  orbital(bf + 4) ! g0
              write (66, '(i4,x,es17.10)') bf + 1, -orbital(bf + 5) ! g+1
              write (66, '(i4,x,es17.10)') bf + 2, -orbital(bf + 3) ! g-1
              write (66, '(i4,x,es17.10)') bf + 3,  orbital(bf + 6) ! g+2
              write (66, '(i4,x,es17.10)') bf + 4,  orbital(bf + 2) ! g-2
              write (66, '(i4,x,es17.10)') bf + 5, -orbital(bf + 7) ! g+3
              write (66, '(i4,x,es17.10)') bf + 6, -orbital(bf + 1) ! g-3
              write (66, '(i4,x,es17.10)') bf + 7,  orbital(bf + 8) ! g+4
              write (66, '(i4,x,es17.10)') bf + 8,  orbital(bf + 0) ! g-4
            end if
          case default
            stop 'angular momentum beyond g not supported!'
          end select
          if(yes_cartesian)&
            deallocate(sphh2cart_matrix)
          bfc = bfc + ((l + 1) * (l + 2)) / 2
          bf = bf + 2 * l + 1
        end do! coef
      end do! mo
    end do! spin

    close (unit=66)

  end subroutine write_molden_file

end module molden
