!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module molecular_grid
  !! author: Sang-Kil Son
  !!
  !! MOLECULAR_GRID module to define a multicenter molecular grid system
  !!   written by Son, Sang-Kil in Mar. 2014
  !!   author: Sang-Kil Son
  !! USAGE:
  !!   type(Grid3D) :: G
  !!   call construct_Grid3D( P, A, G )
  !!
  use constants
  use read_param
  use atom_in_mol
  use sph_harmonics
  use packed_index
  use anglib
  use stdio
  use legendre
  implicit none
  type :: Grid3D
    !! molecular grid information
    integer :: N_grid   ! # of molecular grids
    integer :: N_nuc    ! # of nuclei
    real(long), allocatable :: x(:), y(:), z(:), weight(:)
    !
    ! each grid might have additional information
    !   i_nuc(1:N_grid):  the grid belonging to this atomic index
    !   weight2c(1:N_grid,1:N_nuc):
    !     For weight2c( k, j_nuc ),
    !       i_nuc(k) == j_nuc --> a weight for single-center integration
    !       i_nuc(k) /= j_nuc --> a weight for double-center integration
    !   i_r(1:N_grid): store the radial grid index of jth MG for AG
    !   i_ang(1:N_grid): store the angular grid index of jth MG for AG
    !   i_start(i_nuc): staring index of i_nuc-th atom, i_nuc = 1:N_nuc
    !   i_end(i_nuc): ending index of i_nuc-th atom, i_nuc = 1:N_nuc
    !   w_decomp(1:N_grid): nuclear weight for multicenter decomposition
    !
    integer, allocatable :: i_nuc(:), i_r(:), i_ang(:)
    integer, allocatable :: i_start(:), i_end(:)
    real(long), allocatable :: weight2c(:, :), weight_nuc(:, :), w_decomp(:)
    !
    ! radial grids and Y_lm wat the center of A
    !   r_A(i,i_nuc): a distance of i-th MG from i_nuc-th atomic center
    !   Y_lm(i,lm,i_nuc): Y_lm of i-th MG from i_nuc-th atomic center
    !   Y_lm_prefactor(:): Pfefactor of Y_lm used in V_lm(:)
    !
    ! On April 29, Sang-Kil commented out r_A and Y_lm_A
    ! because they are not used any more.
    !
    !    real(long), allocatable :: r_A(:,:)
    real(long), allocatable :: r_A(:)   ! r of i-th MG from atomic center itself
    !    real(long), allocatable :: Y_lm_A(:,:,:)
    real(long), allocatable :: Y_lm_prefactor(:)
    !    real(long) :: P_lm_max(0:20,0:20)
    !
    ! Crystalline parameters copied from P
    !
    logical :: yes_periodic                             ! use periodicity
    logical :: yes_periodic_Fuzzy
    real(long) :: unitcell_x, unitcell_y, unitcell_z    ! size of unitcell
    !
    ! cutoff parameters: only applicable for nuc-nuc distances
    !   determines whether the pair of i_nuc and j_nuc is included or not
    ! RR(1:NN_nuc): NN_nuc = N_nuc * ( N_nuc + 1 ) / 2 using index_packed()
    ! RP(1:3,1:NN_nuc): 1..3 for x, y, z
    ! ** ACHTUNG **
    !   - G%RP(:,ij) returns an internuclear position without translation
    !   - remember that G%RP(:,ji) = - G%RP(:,ij)
    !   - remember that it has to be translated if periodic
    !   - F obtain_nuclear_position() takes care of both sign and translation
    !
    logical :: yes_cutoff
    real(long) :: r_cutoff
    real(long), allocatable :: Z_A(:) ! nuclear charge... why should it be here?
    real(long), allocatable :: XX(:), YY(:), ZZ(:)      ! nuclear positions
    real(long), allocatable :: RR(:)    ! internuclear distances w/o translation
    real(long), allocatable :: RP(:, :)  ! internuclear positions w/o translation
    !
    ! for translation if yes_periodic is on
    ! RR_tr(k,j_nuc,i_nuc): the k-th closest distance between i_nuc and j_nuc
    ! ix_tr(k,j_nuc,i_nuc): translation index for the k-th closest distance
    !   XX_tr = XX + dble(G%ix_tr(k,j_nuc,i_nuc)) * G%unitcell_x
    !           where XX = G%XX(i_nuc) - G%XX(j_nuc)
    !
    integer :: N_tr
    real(long), allocatable :: RR_tr(:, :, :)
    integer, allocatable :: ix_tr(:, :, :), iy_tr(:, :, :), iz_tr(:, :, :)
  end type Grid3D
  private
  public :: Grid3D, construct_Grid3D, reconstruct_Grid3D, update_Grid3D_with_displacement, &
    place_AG_on_Grid3D, calculate_weight_Grid3D, purge_Grid3D, &
    integral_single_center, integral_double_center, integral, &
    bond_distance, obtain_nuclear_position, obtain_grid_position, translation_in_crystal_Fuzzy, &
    translation_in_crystal, find_neighboring_atoms
contains

  subroutine construct_Grid3D(P, A, G)
    !! creates the molecular Grid G
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(out) :: G
    integer :: j, N_nuc, lmax, i_nuc
    integer :: NN_nuc
    integer :: N_grid     ! total number of molecular grids
    N_nuc = size(A)
    N_grid = sum(A(:)%N_r*A(:)%N_ang)         ! add up # of all AG
    G%N_grid = N_grid
    G%N_nuc = N_nuc
    lmax = maxval(A(:)%lmax_ang)
    !
    ! periodicity consideration
    !
    G%yes_cutoff = P%yes_cutoff
    G%r_cutoff = P%r_cutoff
    G%yes_periodic = P%yes_periodic     ! of course, yes
    if (P%yes_periodic) then
      G%yes_periodic_Fuzzy = P%yes_periodic_fuzzy
      G%unitcell_x = P%unitcell_x
      G%unitcell_y = P%unitcell_y
      G%unitcell_z = P%unitcell_z
    end if
    !
    ! memory allocation
    !
    allocate (G%x(N_grid), G%y(N_grid), G%z(N_grid), G%weight(N_grid))
    allocate (G%i_nuc(N_grid), G%i_r(N_grid), G%i_ang(N_grid))
    allocate (G%i_start(N_nuc), G%i_end(N_nuc))
    allocate (G%weight2c(N_grid, N_nuc), G%weight_nuc(N_nuc, N_grid))
    allocate (G%w_decomp(N_grid))
    !  allocate( G%r_A(N_grid,N_nuc) )
    allocate (G%Y_lm_prefactor((lmax + 1)**2))
    !
    ! nuclear position information
    ! G%RR and G%RP (and all _tr) will be set up in calculate_weight_Grid3D()
    !
    allocate (G%Z_A(N_nuc))
    G%Z_A(:) = A(:)%fcharge
    allocate (G%XX(N_nuc), G%YY(N_nuc), G%ZZ(N_nuc))
    G%XX(:) = A(:)%x
    G%YY(:) = A(:)%y
    G%ZZ(:) = A(:)%z
    NN_nuc = N_nuc*(N_nuc + 1)/2
    allocate (G%RR(NN_nuc))      ! distances between i_nuc and j_nuc
    allocate (G%RP(3, NN_nuc))    ! (X,Y,Z) positions between i_nuc and j_nuc
    if (P%yes_periodic) then
      G%N_tr = P%N_tr_max
      !
      ! RR_tr is usually scanned with k=1...N_tr and then j_nuc=1...N_nuc
      ! so k comes first and j_nuc comes later
      !
      allocate (G%RR_tr(G%N_tr, N_nuc, N_nuc))
      allocate (G%ix_tr(G%N_tr, N_nuc, N_nuc), G%iy_tr(G%N_tr, N_nuc, N_nuc), &
        G%iz_tr(G%N_tr, N_nuc, N_nuc))
    end if
    !
    ! GPG (Grid plus Grid): combine individual atomic spherical grids
    !
    j = 1
    do i_nuc = 1, N_nuc       ! i: atomic index, (1:N_nuc)
      G%i_start(i_nuc) = j
      call place_AG_on_Grid3D(A(i_nuc), G, j)
      j = j + A(i_nuc)%N_r*A(i_nuc)%N_ang
      G%i_end(i_nuc) = j - 1
    end do
    !
    ! calculate weights according to Becke's fuzzy atom concept
    !
    call calculate_weight_Grid3D(A(:), G)
    !
    ! calculate r_A and Y_lm_prefactor
    !
    !  call calculate_r_A( A, G )
    call sph_harmonics_prefactor(G, lmax)
    !  call calculate_sph_on_A( A, G, lmax )
  end subroutine construct_Grid3D

!>----------------------------------------------------------------------------
!! reconstruct Grid3D from A
!!----------------------------------------------------------------------------
  subroutine reconstruct_Grid3D(A, G)
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(inout) :: G
    integer :: i_nuc, N_nuc
    N_nuc = size(A)
    !
    ! assign all grid points according to new geometry in A
    !
    do i_nuc = 1, N_nuc
      call place_AG_on_Grid3D(A(i_nuc), G, G%i_start(i_nuc))
      G%XX(i_nuc) = A(i_nuc)%x
      G%YY(i_nuc) = A(i_nuc)%y
      G%ZZ(i_nuc) = A(i_nuc)%z
      G%Z_A(i_nuc) = A(i_nuc)%fcharge
    end do
    !
    ! re-calculate all weights and
    !
    call calculate_weight_Grid3D(A(:), G)  ! recalculate all weights
  end subroutine reconstruct_Grid3D

!>----------------------------------------------------------------------------
!! update Grid3D with displacement
!!   use this subroutine if only molecular geometry in A has been changed.
!!   note that A should be already updated with
!!   update_Atom_set_with_displacement() in atom_in_mol.f90.
!!----------------------------------------------------------------------------
  subroutine update_Grid3D_with_displacement(A, G, displacement)
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(inout) :: G
    real(long), intent(in) :: displacement(:)
    integer :: N_nuc, i_nuc, is, ie
    N_nuc = size(A)
    do i_nuc = 1, N_nuc   ! move all grid points according to displacement(:)
      is = G%i_start(i_nuc)
      ie = G%i_end(i_nuc)
      G%x(is:ie) = G%x(is:ie) + displacement(i_nuc*3 - 2)   ! x-coordinate
      G%y(is:ie) = G%y(is:ie) + displacement(i_nuc*3 - 1)   ! y-coordinate
      G%z(is:ie) = G%z(is:ie) + displacement(i_nuc*3)   ! z-coordinate
      G%XX(i_nuc) = A(i_nuc)%x
      G%YY(i_nuc) = A(i_nuc)%y
      G%ZZ(i_nuc) = A(i_nuc)%z
    end do
    call calculate_weight_Grid3D(A(:), G)  ! recalculate all weights
  end subroutine update_Grid3D_with_displacement

!----------------------------------------------------------------------------
! update both A and G with displacement
!----------------------------------------------------------------------------
  subroutine update_A_and_G_with_displacement(A, G, displacement)
    type(Grid3D), intent(inout) :: G
    type(Atom_set), intent(inout) :: A(:)
    real(long), intent(in) :: displacement(:)
    call update_Atom_set_with_displacement(A, displacement)
    call update_Grid3D_with_displacement(A, G, displacement)
  end subroutine update_A_and_G_with_displacement

!>----------------------------------------------------------------------------
!! place atomic grids for each atom on molecular grids
!!----------------------------------------------------------------------------
  subroutine place_AG_on_Grid3D(A, G, i_start)
    type(Atom_set), intent(in) :: A
    type(Grid3D), intent(inout) :: G
    integer, intent(in) :: i_start
    integer :: i, j, k
    real(long) :: x, y, z, A_x, A_y, A_z
    real(long) :: Rot(3, 3)
    !
    ! rotate angular grids to align to the direction of corres. atom position
    !
    !rr = sqrt( A%x**2 + A%y**2 + A%z**2 )
    !th = acos( A%z / rr )
    !ph = atan2( A%y, A%x )
    !if ( rr == ZERO ) then
    !  th = ZERO
    !  ph = ZERO
    !end if
    !Rot(1,1)=cos(ph)*cos(th); Rot(1,2)=-sin(ph); Rot(1,3)=-cos(ph)*sin(th)
    !Rot(2,1)=sin(ph)*cos(th); Rot(2,2)= cos(ph); Rot(2,3)=-sin(ph)*sin(th)
    !Rot(3,1)=sin(th);         Rot(3,2)= ZERO;    Rot(3,3)= cos(th)
    !if ( N_nuc == 2 .and. i == 2 ) then        ! for diatomic case
    !  Rot(1,1) = -ONE;  Rot(1,2) = ZERO;  Rot(1,3) = ZERO
    !  Rot(2,1) = ZERO;  Rot(2,2) = -ONE;  Rot(2,3) = ZERO
    !  Rot(3,1) = ZERO;  Rot(3,2) = ZERO;  Rot(3,3) = -ONE
    !end if
    !
    ! From Yajiang's request, AG is not rotated at this moment.
    ! So, Rot(:,:) is an identity matrix.
    !
    Rot(1, 1) = ONE; Rot(1, 2) = ZERO; Rot(1, 3) = ZERO
    Rot(2, 1) = ZERO; Rot(2, 2) = ONE; Rot(2, 3) = ZERO
    Rot(3, 1) = ZERO; Rot(3, 2) = ZERO; Rot(3, 3) = ONE
    !
    ! place AG into molecular grids (i_start:i_start+N_AG-1)
    !
    k = i_start
    atomic_angular: do j = 1, A%N_ang
      x = A%angular%x(j)
      y = A%angular%y(j)
      z = A%angular%z(j)
      A_x = Rot(1, 1)*x + Rot(1, 2)*y + Rot(1, 3)*z
      A_y = Rot(2, 1)*x + Rot(2, 2)*y + Rot(2, 3)*z
      A_z = Rot(3, 1)*x + Rot(3, 2)*y + Rot(3, 3)*z
      atomic_radial: do i = 1, A%N_r
        G%x(k) = A%radial%r(i)*A_x + A%x
        G%y(k) = A%radial%r(i)*A_y + A%y
        G%z(k) = A%radial%r(i)*A_z + A%z
        G%i_r(k) = i  ! Yajiang's addition: store the radial grid index for AG
        G%i_ang(k) = j  ! Yajiang's addition: store the angular grid index for AG
        k = k + 1
      end do atomic_radial
    end do atomic_angular
  end subroutine place_AG_on_Grid3D

  subroutine calculate_weight_Grid3D(A, G)
    !! calculate all integration weights on Grid3D and
    implicit none
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(inout) :: G
    integer :: i, i_r, i_ang, k
    integer :: i_nuc, j_nuc, k_nuc, N_nuc
    integer :: ij_nuc, ii_nuc
    real(long) :: R_ij, R_ki, s, s_ki, s_ik
    real(long) :: P(size(A)), r(size(A))
    real(long) :: X, Y, Z
    N_nuc = size(A)
    !
    ! set AG weights
    ! WARNING: you must keep the same order of do-loops as place_AG_on_Grid3D
    !
    i = 0
    G%weight2c(:, :) = ZERO
    do i_nuc = 1, N_nuc
      if (A(i_nuc)%N_nao .eq. 0) cycle
      atomic_angular: do i_ang = 1, A(i_nuc)%N_ang
        atomic_radial: do i_r = 1, A(i_nuc)%N_r
          i = i + 1
          G%i_nuc(i) = i_nuc              ! store the atomic index
          G%weight2c(i, :) = ZERO
          G%weight2c(i, i_nuc) = &         ! store a weight for AG (single-center)
            A(i_nuc)%radial%weight(i_r) * &    ! radial weight
            A(i_nuc)%radial%r(i_r)**2 * &        ! * r^2
            FOUR*PI* &                 ! * 4 pi (missing in ang. w.)
            A(i_nuc)%angular%weight(i_ang)    ! * angular weight
        end do atomic_radial
      end do atomic_angular
    end do
    !
    ! set up internuclear distances between i_nuc and j_nuc --> G%RR(ij_nuc)
    ! later on, bond distance can be obtained by calling bond_distance()
    !
    ! NOTE: NEVER forget initialize variables and arrays!
    !       G%RP(:,ii_nuc) was not properly initialized before Nov. 26, 2024,
    !       which **randomly** caused garbage results for periodic calc...
    !
    do i_nuc = 1, N_nuc
      ii_nuc = int(index_packed(i_nuc, i_nuc))
      G%RR(ii_nuc) = ZERO       ! initialize: distance=0 for the same nuc index
      G%RP(1:3,ii_nuc) = ZERO   ! initialize: (0,0,0)    for the same nuc index
      do j_nuc = i_nuc + 1, N_nuc
        ij_nuc = int(index_packed(i_nuc, j_nuc))
        X = A(i_nuc)%x - A(j_nuc)%x
        Y = A(i_nuc)%y - A(j_nuc)%y
        Z = A(i_nuc)%z - A(j_nuc)%z
        G%RR(ij_nuc) = sqrt(X**2 + Y**2 + Z**2)
        G%RP(:, ij_nuc) = (/X, Y, Z/)  ! NOTE: -G%RP when i & j exchanged
      end do
    end do
    !
    ! G%RR_tr and all _tr will be prepared in setup_translation( G )
    !
    if (G%yes_periodic) call setup_translation(G)
    !-------------------------------------------------------------------------
    ! after this point, you may use bond_distance() rather than G%RR
    !   remember that G%RR without translation
    !   and F bond_distance() with translation
    ! BUT don't use grid_distance() to evaluate nuclear weights
    !     --> see the below comment
    !-------------------------------------------------------------------------
    !
    ! set nuclear weights
    !
    do k = 1, G%N_grid
      k_nuc = G%i_nuc(k)
      !
      ! set up distances from individual atoms
      !   r(i_nuc) = distance between k-th grid point and i_nuc-th atom
      !
      do i_nuc = 1, N_nuc
        ! if ( A(i_nuc)%N_nao == 0 ) cycle
        X = A(i_nuc)%x - G%x(k)
        Y = A(i_nuc)%y - G%y(k)
        Z = A(i_nuc)%z - G%z(k)
        !
        ! when periodicity is considered, shift grid positions
        ! taking into account "minimum image convention"
        ! ACHTUNG: here, translate all grid points w.r.t their position and
        !   nuclear position; otherwise, it cannot describe a torus structure!
        !   this should be applied in the same way
        !   for place_AO_on_Grid3D in basis_func.f90
        !   again, you shouldn't use grid_distance() here,
        !   which translates a nuclear position (not a grid position)
        !   and then shift the grid position accordingly
        !
        if (G%yes_periodic) call translation_in_crystal(G, X, Y, Z)
        r(i_nuc) = sqrt(X**2 + Y**2 + Z**2)
      end do
      !
      ! assign a weight for the two-center case: k_nuc and i_nuc
      !
      !G%weight2c(k,:) = ZERO
      do i_nuc = 1, N_nuc
        if (A(i_nuc)%N_nao .eq. 0) cycle
        if (i_nuc .ne. k_nuc) then
          R_ki = bond_distance(G, k_nuc, i_nuc)
          s_ki = nuclear_weight_s(r(k_nuc), r(i_nuc), R_ki)
          s_ik = nuclear_weight_s(r(i_nuc), r(k_nuc), R_ki)     ! R_ik == R_ki
          G%weight2c(k, i_nuc) = s_ki/(s_ki + s_ik) * &        ! nuc. weight
            G%weight2c(k, k_nuc)               ! AG weight
        end if
      end do
      !
      ! assign a weight for the multicenter case
      !   P(k_nuc) / sum( P(:) )  : nuclear weight
      !   G%weight2c(k,k_nuc)     : AG weight (single-center)
      ! we need only one nuclear weight (here, k_nuc) rather than all i_nuc's
      ! at each grid point (k-th grid).
      !
      P(:) = ZERO
      do i_nuc = 1, N_nuc
        if (A(i_nuc)%N_nao .eq. 0) cycle
        s = ONE
        do j_nuc = 1, N_nuc
          if (A(j_nuc)%N_nao .eq. 0) cycle
          if (i_nuc .ne. j_nuc) then
            R_ij = bond_distance(G, i_nuc, j_nuc)
            s = s*nuclear_weight_s(r(i_nuc), r(j_nuc), R_ij)
          end if
        end do
        P(i_nuc) = s      ! assign for all i_nuc's: not normalized yet
      end do
      G%w_decomp(k) = P(k_nuc)/sum(P(:))      ! store nuclear weight
      G%weight(k) = G%w_decomp(k)*G%weight2c(k, k_nuc)
      G%weight_nuc(:, k) = P(:)/sum(P(:))      ! Yajiang's addition
    end do
  end subroutine calculate_weight_Grid3D

  function nuclear_weight_s(r_i, r_j, R_ij) result(s)
    real(long) :: s
    real(long), intent(in) :: r_i, r_j, R_ij
    real(long) :: mu, p1, p2, p3, p4, p5
    !
    ! this function is defined only for diatomic molecules
    !
    mu = (r_i - r_j)/R_ij
    p1 = HALF*mu*(THREE - mu**2)
    p2 = HALF*p1*(THREE - p1**2)
    p3 = HALF*p2*(THREE - p2**2)
    p4 = HALF*p3*(THREE - p3**2)
    p5 = HALF*p4*(THREE - p4**2)
    s = HALF*(ONE - p3)       ! let's use k = 3
  end function nuclear_weight_s

!subroutine calculate_sph_on_A( A, G, lmax )
!  type(Atom_set), intent(in) :: A(:)
!  type(Grid3D), intent(inout) :: G
!  integer, intent(in) :: lmax
!  integer :: i_nuc, i, l, m, lm
!  real(long) :: x, y, z, r, theta, phi
!  do i_nuc = 1, G%N_nuc
!    do i = 1, G%N_grid
!      x = G%x(i) - A(i_nuc)%x
!      y = G%y(i) - A(i_nuc)%y
!      z = G%z(i) - A(i_nuc)%z
!      call xyz_to_sph( x, y, z, r, theta, phi )
!      G%r_A(i,i_nuc) = r
!      do l = 0, lmax
!        do m = -l, l
!          lm = l**2 + ( m + l ) + 1     ! lm = 1, 2, 3, ...
!          G%Y_lm_A(i,lm,i_nuc) = sph_harmonics_real( l, m, theta, phi )
!        end do
!      end do
!    end do
!  end do
!end subroutine calculate_sph_on_A

!subroutine calculate_r_A( A, G )
!  type(Atom_set), intent(in) :: A(:)
!  type(Grid3D), intent(inout) :: G
!  integer :: i_nuc, i
!  real(long) :: x, y, z, r, theta, phi
!  do i_nuc = 1, G%N_nuc
!    do i = 1, G%N_grid
!      x = G%x(i) - A(i_nuc)%x
!      y = G%y(i) - A(i_nuc)%y
!      z = G%z(i) - A(i_nuc)%z
!      call xyz_to_sph( x, y, z, r, theta, phi )
!      G%r_A(i,i_nuc) = r
!    end do
!  end do
!end subroutine calculate_r_A

!----------------------------------------------------------------------------
! setup G%Y_lm_prefactor
! this is closely connected to sph_harmonics_real() in sph_harmonics.f90
!   sph_harmonics_real( l, m, theta, phi ) :=
!     G%Y_lm_prefactor(lm) * (2l+1) / 4pi *
!     associated_Legendre_P( l, |m|, cos(theta) ) * [ ... ]
!     [ ... ] = 1                 for m = 0
!             = cos( |m| * phi )  for m > 0
!             = sin( |m| * phi )  for m < 0
!
!   G%Y_lm_prefactor(1) = 3.54490770181103 = sqrt(4pi)          l=0, m= 0
!   G%Y_lm_prefactor(2) = 2.04665341589298 = sqrt(4pi/3)        l=1, m=-1
!   G%Y_lm_prefactor(3) = 2.04665341589298 = sqrt(4pi/3)        l=1, m= 0
!   G%Y_lm_prefactor(4) = 2.04665341589298 = sqrt(4pi/3)        l=1, m=+1
!   G%Y_lm_prefactor(5) = 0.45764561643189 = sqrt( pi/15)       l=2, m=-2
!   G%Y_lm_prefactor(6) = 0.91529123286377 = sqrt(4pi/15)       l=2, m=-1
!   G%Y_lm_prefactor(7) = 1.58533091904240 = sqrt(4pi/5)        l=2, m= 0
!   G%Y_lm_prefactor(8) = 0.91529123286377 = sqrt(4pi/15)       l=2, m=+1
!   G%Y_lm_prefactor(9) = 0.45764561643189 = sqrt( pi/15)       l=2, m=+2
!   ...
!----------------------------------------------------------------------------
  subroutine sph_harmonics_prefactor(G, lmax)
    type(Grid3D), intent(inout) :: G
    integer, intent(in) :: lmax
    integer :: l, m, lm
    real(long) :: Y_pre
    real(long) :: p
    do l = 0, lmax
      do m = -l, l
        lm = l**2 + (m + l) + 1     ! lm = 1, 2, 3, ...
        !
        !                 p               Y_pre           Y_lm_prefactor
        ! l=0,m= 0        sqrt(1/2)       1/sqrt(4pi)     sqrt(4pi)
        ! l=1,m=-1        sqrt(3/2/2!)    sqrt(3/4pi)     sqrt(4pi/3)
        ! l=1,m= 0        sqrt(3/2)       sqrt(3/4pi)     sqrt(4pi/3)
        ! l=1,m=+1        sqrt(3/2/2!)    sqrt(3/4pi)     sqrt(4pi/3)
        !
        ! NOTE: factorial(n,m) = m! / n!  for n < m
        !                      = 1        otherwise
        !       thus, factorial(1,2) = 2; factorial(2,1) = 1
        !
        p = sqrt(dble(2*l + 1)/TWO/factorial2(l - abs(m) + 1, l + abs(m)))
        if (m .eq. 0) then
          Y_pre = p/sqrt(TWO*PI)
        else
          Y_pre = p/sqrt(PI)
        end if
        G%Y_lm_prefactor(lm) = Y_pre*FOUR*PI/dble(2*l + 1)
      end do
    end do
  end subroutine sph_harmonics_prefactor

  function calculate_sph_on_i_nuc(G, i, i_nuc, l, m, k_tr) result(Ylm)
    type(Grid3D), intent(in) :: G
    integer, intent(in) :: i, i_nuc, l, m
    integer, intent(in), optional :: k_tr
    real(long) :: x, y, z, r, theta, phi, Ylm, r_tr(3)
    integer :: k  ! translation index     NOTE: Don't ever use "integer :: k = 1"
    k = 1         ! default: k = 1              unless k is a fixed parameter
    if (present(k_tr)) k = k_tr
    r_tr(:) = obtain_grid_position(G, i, i_nuc, k)
    x = r_tr(1)   ! x = G%x(i) - A%x
    y = r_tr(2)   ! y = G%y(i) - A%y
    z = r_tr(3)   ! z = G%z(i) - A%z
    call xyz_to_sph(x, y, z, r, theta, phi)
    Ylm = associated_Legendre_P(l, abs(m), cos(theta))
    if (m .eq. 0) then
      Ylm = Ylm
    else if (m .gt. 0) then
      Ylm = Ylm*cos(dble(m)*phi)
    else
      Ylm = Ylm*sin(dble(-m)*phi)
    end if
  end function calculate_sph_on_i_nuc

  function integral(G, f) result(value)
    !! integral on the grid
    type(Grid3D), intent(in) :: G
    real(long), intent(in) :: f(G%N_grid)
    real(long) :: value
    value = sum(G%weight(:)*f(:))
  end function integral

  function integral_single_center(G, f, i_nuc) result(value)
    !! single center grid integral    
    type(Grid3D), intent(in) :: G
    real(long), intent(in) :: f(G%N_grid)
    integer, intent(in) :: i_nuc
    real(long) :: value
    integer :: is, ie
    !
    ! original implementation
    !
    !value = ZERO
    !do k = 1, G%N_grid
    !  if ( G%i_nuc(k) == i_nuc ) value = value + G%weight2c(k,i_nuc) * f(k)
    !end do
    !
    ! new implementation with G%i_start and G%i_end
    !
    is = G%i_start(i_nuc)
    ie = G%i_end(i_nuc)
    value = sum(f(is:ie)*G%weight2c(is:ie, i_nuc))
  end function integral_single_center

  function integral_double_center(G, f, i_nuc, j_nuc) result(value)
    !! double center grid integral
    type(Grid3D), intent(in) :: G
    real(long), intent(in) :: f(G%N_grid)
    integer, intent(in) :: i_nuc, j_nuc
    real(long) :: value
    integer :: is, ie, js, je
    if (i_nuc .eq. j_nuc) then
      value = integral_single_center(G, f, i_nuc)
      return
    end if
    !value = ZERO
    !do k = 1, G%N_grid
    !  if ( G%i_nuc(k) == i_nuc ) then
    !    value = value + G%weight2c(k,j_nuc) * f(k)
    !  else if ( G%i_nuc(k) == j_nuc ) then
    !    value = value + G%weight2c(k,i_nuc) * f(k)
    !  end if
    !end do
    !
    ! integration with two-center weights
    !
    is = G%i_start(i_nuc) ! collect grid points for the i_nuc-th atom
    ie = G%i_end(i_nuc)
    js = G%i_start(j_nuc) ! collect grid points for the j_nuc-th atom
    je = G%i_end(j_nuc)
    value = sum(f(is:ie)*G%weight2c(is:ie, j_nuc)) + &
      sum(f(js:je)*G%weight2c(js:je, i_nuc))
  end function integral_double_center

!subroutine find_AG_in_Grid3D( G, A, j, k, x, y, z )
!  type(Grid3D), intent(in) :: G
!  type(Atom_set), intent(in) :: A(:)
!  integer, intent(in) :: j              ! INPUT: j-th molecular grid point
!  integer, intent(out) :: k             ! OUTPUT: k-th atom index
!  real(long), intent(out) :: x, y, z
!  integer :: i, N_nuc
!  N_nuc = size(A)
!  do i = 1, N_nuc
!end subroutine find_AG_in_Grid3D

  subroutine dump_Grid3D(G, IOUT)
    type(Grid3D), intent(in) :: G
    integer, intent(in), optional :: IOUT
    integer :: i, OUTPUT
    OUTPUT = STDOUT; if (present(IOUT)) OUTPUT = IOUT
    write (OUTPUT, *) "Molecular grids:"
    write (OUTPUT, *) "  N_grid = ", G%N_grid
    do i = 1, G%N_grid
      write (OUTPUT, "(4es24.15e3)") G%x(i), G%y(i), G%z(i), G%weight(i)
    end do
  end subroutine dump_Grid3D

  subroutine purge_Grid3D(G)
    type(Grid3D), intent(inout) :: G
    if (allocated(G%x)) deallocate (G%x)
    if (allocated(G%y)) deallocate (G%y)
    if (allocated(G%z)) deallocate (G%z)
    if (allocated(G%weight)) deallocate (G%weight)
    if (allocated(G%i_nuc)) deallocate (G%i_nuc)
    if (allocated(G%i_r)) deallocate (G%i_r)
    if (allocated(G%i_ang)) deallocate (G%i_ang)
    if (allocated(G%i_start)) deallocate (G%i_start)
    if (allocated(G%i_end)) deallocate (G%i_end)
    if (allocated(G%weight2c)) deallocate (G%weight2c)
    if (allocated(G%weight_nuc)) deallocate (G%weight_nuc)
    if (allocated(G%w_decomp)) deallocate (G%w_decomp)
    if (allocated(G%Y_lm_prefactor)) deallocate (G%Y_lm_prefactor)
    if (allocated(G%RP)) deallocate (G%RP)
    if (allocated(G%RR)) deallocate (G%RR)
    if (allocated(G%XX)) deallocate (G%XX)
    if (allocated(G%YY)) deallocate (G%YY)
    if (allocated(G%ZZ)) deallocate (G%ZZ)
    if (allocated(G%Z_A)) deallocate (G%Z_A)

  end subroutine purge_Grid3D

!----------------------------------------------------------------------------
! SR translation_in_crystal( G, x, y, z )
!   the periodic boundary condition is considered here.
!   find out a position of (x,y,z) in neighboring unitcells,
!   which would give the smallest r = sqrt( x**2 + y**2 + z**2 )
!
! ==> Minimum Image Convention:
!     Allen & Tildesley, Computer Simulation of Liquids (Oxford, 1987)
!     first mentioned in Metropolis et al., JCP 21, 1087 (1953).
!
! F obtain_grid_position( G, i, j_nuc, k ) result (r(3))
!   returns r(3) = (/ x, y, z /) of the i-th grid point
!   with respect to the j_nuc-th nuclear position.
!   if periodic is used, it's translated such that a distance between
!   the i_nuc-th atom and the j_nuc-th atom is the k-th closest one.
!   NOTE that it's internuclear distance and the translation operation is
!   only made for the nuclear positions.
!
! F obtain_nuclear_position( G, i_nuc, j_nuc, k ) result (R(3))
!   returns R(3) = (/ X, Y, Z /) of the i_nuc-th nuclear position
!   with respect to the j_nuc-th nuclear position.
!   if periodic is used, it's translated to produce the k-th closest distance.
!
! F bond_distance( G, i_nuc, j_nuc, k ) result (R)
!   returns a bond distance between the i_nuc-th atom and the j_nuc-th atom.
!   if it's the periodic case, it returns the k-th distance after translation.
!
! NOTE: obtain_grid_position(), obtain_nuclear_position(), and bond_distance()
!       can be used for both periodic and non-periodic cases.
!----------------------------------------------------------------------------
  subroutine translation_in_crystal(G, x, y, z)
    type(Grid3D), intent(in) :: G
    real(long), intent(inout) :: x, y, z
    real(long) :: x_tr, y_tr, z_tr
    !
    ! make a translation in a given crystal
    ! if x > 0, then x_tr = x - G%unitcell_x
    ! if x < 0, then x_tr = x + G%unitcell_x
    ! ==> x_tr = x - sign( G%unitcell_x, x ), where G%unitcell_x must be > 0.
    !
    x_tr = x - sign(G%unitcell_x, x)
    y_tr = y - sign(G%unitcell_y, y)
    z_tr = z - sign(G%unitcell_z, z)
    !
    ! find out (x,y,z) corresponding to the smallest |x|, |y|, and |z|
    !
    if (abs(x_tr) .lt. abs(x)) x = x_tr
    if (abs(y_tr) .lt. abs(y)) y = y_tr
    if (abs(z_tr) .lt. abs(z)) z = z_tr
  end subroutine translation_in_crystal

  function obtain_grid_position(G, i, j_nuc, k_tr) result(r)
    real(long) :: r(3)                    !! (x, y, z) of i-th grid point
    type(Grid3D), intent(in) :: G
    integer, intent(in) :: i              !! i-th grid point
    integer, intent(in) :: j_nuc          !! j_nuc-th atom
    integer, intent(in), optional :: k_tr !! k-th closest point if periodic
    integer :: i_nuc, k
    real(long) :: x, y, z !, XX, YY, ZZ
    real(long) :: RR(3)    ! (X, Y, Z)
    if (G%yes_periodic .eqv. .false.) then
      !
      ! if it's not periodic, then
      ! simply provide the i-th grid position w.r.t. the j_nuc-th atom
      !
      x = G%x(i) - G%XX(j_nuc)
      y = G%y(i) - G%YY(j_nuc)
      z = G%z(i) - G%ZZ(j_nuc)
    else
      !
      ! if it's periodic, then
      !   translate the i-th grid point such that the distance
      !   between i_nuc and j_nuc is the k-th closest one,
      !   where i_nuc is the nucleus corresponding to the i-th grid
      !
      ! REMEMBER: translation w.r.t. nuclear positions only
      !   XX = G%XX(i_nuc) - G%XX(j_nuc) <-- make translation first
      !   XX_tr = XX + dble(G%ix_tr(k,j_nuc,i_nuc)) * G%unitcell_x <-- 2nd tr.
      !   x = G%x(i) - G%XX(i_nuc) + XX_tr
      !
      k = 1
      if (present(k_tr)) k = k_tr
      i_nuc = G%i_nuc(i)
      !
      ! original expressions
      !
      !XX = G%XX(i_nuc) - G%XX(j_nuc)
      !YY = G%YY(i_nuc) - G%YY(j_nuc)
      !ZZ = G%ZZ(i_nuc) - G%ZZ(j_nuc)
      !call translation_in_crystal( G, XX, YY, ZZ )
      !x = G%x(i) - G%XX(i_nuc) + XX + dble(G%ix_tr(k,j_nuc,i_nuc)) * G%unitcell_x
      !y = G%y(i) - G%YY(i_nuc) + YY + dble(G%iy_tr(k,j_nuc,i_nuc)) * G%unitcell_y
      !z = G%z(i) - G%ZZ(i_nuc) + ZZ + dble(G%iz_tr(k,j_nuc,i_nuc)) * G%unitcell_z
      !
      ! simplified one using pre-calculated G%RP
      !
      RR(1:3) = obtain_nuclear_position(G, i_nuc, j_nuc, k)
      x = G%x(i) - G%XX(i_nuc) + RR(1)
      y = G%y(i) - G%YY(i_nuc) + RR(2)
      z = G%z(i) - G%ZZ(i_nuc) + RR(3)
    end if
    r(:) = (/x, y, z/)
  end function obtain_grid_position

  function grid_distance(G, i, j_nuc, k_tr) result(r_i)
    real(long) :: r_i
    type(Grid3D), intent(in) :: G
    integer, intent(in) :: i              ! i-th grid point
    integer, intent(in) :: j_nuc          ! j_nuc-th atom
    integer, intent(in), optional :: k_tr ! k-th closest point if periodic
    real(long) :: r(3)                    ! (x, y, z) of i-th grid point
    !real(long) :: x, y, z, XX, YY, ZZ
    r(:) = obtain_grid_position(G, i, j_nuc, k_tr)
    r_i = sqrt(sum(r(:)**2))
  end function grid_distance

  function obtain_nuclear_position(G, i_nuc, j_nuc, k_tr) result(R)
    real(long) :: R(3)                    ! (X, Y, Z) of i_nuc - j_nuc
    type(Grid3D), intent(in) :: G
    integer, intent(in) :: i_nuc          ! i_nuc-th atom
    integer, intent(in) :: j_nuc          ! j_nuc-th atom
    integer, intent(in), optional :: k_tr ! k-th closest point if periodic
    integer :: k
    integer :: ij_nuc
    !real(long) :: XX, YY, ZZ
    ij_nuc = int(index_packed(i_nuc, j_nuc))
    R(:) = G%RP(:, ij_nuc)
    if (j_nuc .lt. i_nuc) R(:) = -R(:)
    if (G%yes_periodic) then
      k = 1
      if (present(k_tr)) k = k_tr
      call translation_in_crystal(G, R(1), R(2), R(3))
      R(1) = R(1) + dble(G%ix_tr(k, j_nuc, i_nuc))*G%unitcell_x
      R(2) = R(2) + dble(G%iy_tr(k, j_nuc, i_nuc))*G%unitcell_y
      R(3) = R(3) + dble(G%iz_tr(k, j_nuc, i_nuc))*G%unitcell_z
    end if
    !XX = G%XX(i_nuc) - G%XX(j_nuc)
    !YY = G%YY(i_nuc) - G%YY(j_nuc)
    !ZZ = G%ZZ(i_nuc) - G%ZZ(j_nuc)
    !if ( G%yes_periodic ) then
    !  k = 1
    !  if ( present(k_tr) ) k = k_tr
    !  call translation_in_crystal( G, XX, YY, ZZ )
    !  XX = XX + dble(G%ix_tr(k,j_nuc,i_nuc)) * G%unitcell_x
    !  YY = YY + dble(G%iy_tr(k,j_nuc,i_nuc)) * G%unitcell_y
    !  ZZ = ZZ + dble(G%iz_tr(k,j_nuc,i_nuc)) * G%unitcell_z
    !end if
    !R(:) = (/ XX, YY, ZZ /)
  end function obtain_nuclear_position

  function bond_distance(G, i_nuc, j_nuc, k_tr) result(R_AB)
    !! calculated distances between atoms (employing periodic boundary conditions)    
    real(long) :: R_AB
    type(Grid3D), intent(in) :: G
    integer, intent(in) :: i_nuc, j_nuc   ! i_nuc-th and j_nuc-th atom
    integer, intent(in), optional :: k_tr ! k-th closest point if periodic
    integer :: ij_nuc
    integer :: k  ! translation index     NOTE: Don't ever use "integer :: k = 1"
    k = 1         ! default: k = 1              unless k is a fixed parameter
    !
    ! periodic case: scanning order is pretty much important
    !                k_tr first, j_nuc later, i_nuc at last --> optimal speed
    ! non-periodic case: order of i_nuc and j_nuc doesn't matter
    !
    if (G%yes_periodic) then
      if (present(k_tr)) k = k_tr
      R_AB = G%RR_tr(k, j_nuc, i_nuc)  ! ACHTUNG: index order for k,j_nuc,i_nuc
    else
      ij_nuc = int(index_packed(i_nuc, j_nuc))
      R_AB = G%RR(ij_nuc)
    end if
  end function bond_distance

  subroutine setup_translation(G)
    type(Grid3D), intent(inout) :: G
    real(long) :: X, Y, Z
    integer :: i_nuc, j_nuc
    integer :: ix, iy, iz, i, n, N_tr
    integer :: ij_nuc
    real(long) :: R(3)    ! X, Y, Z
    real(long) :: X_tr, Y_tr, Z_tr
    real(long), allocatable :: R_tr(:)
    integer, allocatable :: i_tr(:, :)
    !
    ! determine max. number of translation for each direction
    ! N_tr: max. number of translated coordinates is given by size(r)
    !       = ( 2*n + 1 )**3
    !
    n = ceiling((dble(G%N_tr)**(ONE/THREE) - ONE)/2)
    N_tr = (2*n + 1)**3
    if (N_tr .lt. G%N_tr ) &
      call error_message("Error: translation_in_crystal_extended")
    allocate (R_tr(N_tr), i_tr(N_tr, 3))
    do i_nuc = 1, G%N_nuc
      do j_nuc = 1, G%N_nuc
        ij_nuc = int(index_packed(i_nuc, j_nuc))
        R(1:3) = G%RP(:, ij_nuc)                     ! (X, Y, Z)
        if (j_nuc .lt. i_nuc) R(1:3) = -R(1:3)       ! consider the right direction
        X = R(1)  !X = G%XX(i_nuc) - G%XX(j_nuc)
        Y = R(2)  !Y = G%YY(i_nuc) - G%YY(j_nuc)
        Z = R(3)  !Z = G%ZZ(i_nuc) - G%ZZ(j_nuc)
        call translation_in_crystal(G, X, Y, Z)   ! x,y,z such that |R| smallest
        !
        ! generate translated coordinates scanning from -n to n
        ! if yes_cutoff and r > r_cutoff, then put a huge number into r,
        ! which won't be sorted in the following procedure
        !
        i = 0
        do ix = -n, n; X_tr = X + dble(ix)*G%unitcell_x
          do iy = -n, n; Y_tr = Y + dble(iy)*G%unitcell_y
            do iz = -n, n; Z_tr = Z + dble(iz)*G%unitcell_z
              i = i + 1
              R_tr(i) = sqrt(X_tr**2 + Y_tr**2 + Z_tr**2)
              i_tr(i, 1:3) = (/ix, iy, iz/)
            end do
          end do
        end do
        !
        ! quick sort r_tr(:) in ascending order
        !
        call quicksort(R_tr, i_tr, 1, N_tr)
        !
        ! bubble sort r_tr(:) in ascending order
        !
        !do i = 1, N_tr
        !  do j = i + 1, N_tr
        !    if ( R_tr(j) < R_tr(i) ) then
        !      temp     = R_tr(j);    R_tr(j)   = R_tr(i);    R_tr(i)   = temp
        !      itemp(:) = i_tr(j,:);  i_tr(j,:) = i_tr(i,:);  i_tr(i,:) = itemp(:)
        !    end if
        !  end do
        !end do
        !
        ! return all _tr variables
        !
        G%RR_tr(:, j_nuc, i_nuc) = R_tr(1:G%N_tr)
        G%ix_tr(:, j_nuc, i_nuc) = i_tr(1:G%N_tr, 1)
        G%iy_tr(:, j_nuc, i_nuc) = i_tr(1:G%N_tr, 2)
        G%iz_tr(:, j_nuc, i_nuc) = i_tr(1:G%N_tr, 3)
      end do
    end do
    deallocate (R_tr, i_tr)
  contains
    recursive subroutine quicksort(R, ixyz, start, end)
      real(long), intent(inout) :: R(:)
      integer, intent(inout) :: ixyz(size(R), 3)
      integer, intent(in) :: start, end
      integer :: i, j, itemp(3)
      real(long) :: mid, temp
      mid = R((start + end)/2)
      i = start
      j = end
      SORTLOOP: do
        do while (R(i) .lt. mid)
          i = i + 1
        end do
        do while (mid .lt. R(j))
          j = j - 1
        end do
        if (i .ge. j) exit SORTLOOP
        temp = R(i)
        R(i) = R(j)
        R(j) = temp
        itemp(:) = ixyz(i, :)
        ixyz(i, :) = ixyz(j, :)
        ixyz(j, :) = itemp(:)
        i = i + 1
        j = j - 1
      end do SORTLOOP
      if (start .lt. i - 1) call quicksort(R, ixyz, start, i - 1)
      if (j + 1 .lt. end) call quicksort(R, ixyz, j + 1, end)
    end subroutine quicksort
  end subroutine setup_translation

  function translation_in_crystal_Fuzzy(G, x, y, z, n_Fuzzy) result(ret4x8)
    implicit none
    type(Grid3d), intent(in) :: G
    real(long), intent(in) :: x, y, z
    real(long) :: ret4x8(4, 8)
    integer, optional, intent(in) :: n_Fuzzy; integer :: nFuz

    logical, parameter :: dbgng = .true.
    real(long) :: L(3), H(3), xA(2), yA(2), zA(2), wX, wY, wZ, wa(3)
    integer :: ijk, i, j, k

    real(long), save :: maxSum = -ONE, minSum = TWO
    nFuz = 3; if (present(n_Fuzzy)) nFuz = n_Fuzzy

    L = (/G%unitcell_x, G%unitcell_y, G%unitcell_z/)
    H = L*half

    !call assertf( abs(x)<=L(1), "#translation_in_crystal_Fuzzy:|x|",-1)
    !call assertf( abs(y)<=L(2), "translation_in_crystal_Fuzzy:|y|",-1)
    !call assertf( abs(z)<=L(3), "translation_in_crystal_Fuzzy:|z|",-1)

    ! |x|>L(1) is very exceptional case but there likely to be. Correct translation is x-L(1) and x-2L(1)
    xA = (/x, x - L(1)/); if (x .gt. L(1)) then; xA = (/x - L(1), x - 2*L(1)/); else if (x .lt. -L(1)) then; xA = (/x + L(1), x + 2*L(1)/); else if (x .lt. 0) then; xA = (/x, x + L(1)/); end if
    yA = (/y, y - L(2)/); if (y .gt. L(2)) then; yA = (/y - L(2), y - 2*L(2)/); else if (y .lt. -L(2)) then; yA = (/y + L(2), y + 2*L(2)/); else if (y .lt. 0) then; yA = (/y, y + L(2)/); end if
    zA = (/z, z - L(3)/); if (z .gt. L(3)) then; zA = (/z - L(3), z - 2*L(3)/); else if (z .lt. -L(3)) then; zA = (/z + L(3), z + 2*L(3)/); else if (z .lt. 0) then; zA = (/z, z + L(3)/); end if

    !if( dbgng ) then
    ! call assertf(  dabs(one-sum( fuzzy1(abs(xA(:)/H(1))) )) .lt. 1.0e-06, "translation_in_crystalXX:1", -1)
    ! call assertf(  dabs(one-sum( fuzzy1(abs(yA(:)/H(2))) )) .lt. 1.0e-06, "translation_in_crystalXX:2", -1)
    ! call assertf(  dabs(one-sum( fuzzy1(abs(zA(:)/H(4))) )) .lt. 1.0e-06, "translation_in_crystalXX:3", -1)
    !endif

    if (nFuz .lt. 0) then
      do k = 2, 8; ret4x8(:, k) = (/x, y, z, ZERO/); end do
      ret4x8(:, 1) = (/merge(xA(2), xA(1), abs(xA(2)) .lt. abs(xA(1)) ), &
        merge(yA(2), yA(1), abs(yA(2)) .lt. abs(yA(1))), & 
        merge(zA(2), zA(1), abs(zA(2)) .lt. abs(zA(1))), ONE/)
      return
    end if

    wa(1) = fuzzy1(abs(xA(1))/H(1), nFuz) + fuzzy1(abs(xA(2))/H(1), nFuz);
    wa(2) = fuzzy1(abs(yA(1))/H(2), nFuz) + fuzzy1(abs(yA(2))/H(2), nFuz);
    wa(3) = fuzzy1(abs(zA(1))/H(3), nFuz) + fuzzy1(abs(zA(2))/H(3), nFuz);
    ijk = 0
    do i = 1, 2; wX = fuzzy1(abs(xA(i))/H(1), nFuz); do j = 1, 2; wY = fuzzy1(abs(yA(j))/H(2), nFuz);
        do k = 1, 2; ijk = ijk + 1; wZ = fuzzy1(abs(zA(k))/H(3), nFuz);
          ret4x8(:, ijk) = (/xA(i), yA(j), zA(k), wX*wY*wZ/);
        end do
      end do; end do

    if (dabs(one - sum(ret4x8(4, :))) .gt. 1.0e-05) then
      !
      ! Check weightsum deviation ... you almost never come here
      !
      maxSum = max(maxSum, sum(ret4x8(4, :)))
      minSum = min(minSum, sum(ret4x8(4, :)))

      write (STDERR, '(A,f16.8,5x,999f16.8)') "#translation_in_crystalXX:!W wgtsum deviates:", sum(ret4x8(4, :)), ret4x8(4, :)
      write (STDERR, '(A,f16.8,5x,999f16.8)') "#translation_in_crystalXX:!W check 1st order dev:", maxval(dabs(wa(:) - ONE)), wa(:); flush (6)
      write (STDERR, '(A,f16.8,5x,999f16.8)') "#translation_in_crystalXX:!W check current min/maxSum:", minSum, maxSum
      write(STDERR,'(A,999f16.8)')"#translation_in_crystalXX:!W wgtsum deviates:" ,xA(:),yA(:),zA(:), abs(xA(:))/H(1), abs(yA(:))/H(2), abs(zA(:))/H(3)
      write(STDERR,'(A,999f16.8)')"#translation_in_crystalXX:!W wgtsum deviates:" ,fuzzy1( abs(xA(1)/H(1)), nFuz), fuzzy1( abs(xA(2)/H(1)), nFuz)
      !     write(STDERR,'(A,999f16.8)')"#translation_in_crystalXX:!W wgtsum deviates:" ,( fuzzy1(0.025_long*j,nFuz)+fuzzy1(TWO-0.025_long*j,nFuz), j=0,40 )
      stop "#translation_in_crystalXX:!E wgtsum deviates:from 1"
    end if

  contains

    !---------------------------------------------------------------------
    ! n-th order Becke's fuzzy function such that
    ! f(0)=1.0 f(1)=0.5 f(2)=0.0
    !---------------------------------------------------------------------
    real(long) function fuzzy1(x, n) result(y)
      implicit none
      real(long) :: x
      integer :: i
      integer, optional :: n
      integer :: m
      m = 3; if (present(n)) m = n;
      if (m .eq. (-1)) then
        y = ONE
        if (x .gt. 0) y = ZERO
        return
      end if
      y = x - ONE
      do i = 1, m
        y = (three - y**2)*y*half
      end do
      y = half*(one - y)
    end function
  end function

  subroutine find_neighboring_atoms(G, A, OUTPUT)
    !! count how many neighboring atoms for each atom and find the bond distances
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    integer :: OUTPUT
    integer :: i, j, ij, k, n, N_nuc, N_tr
    real(long) :: temp, R_max
    real(long), allocatable :: R(:)
    integer :: i_min, i_max
    integer, allocatable :: icount(:)
    integer :: N_c(size(A))
    real(long) :: N_c_avg
    !
    ! set up arrays
    !
    N_nuc = size(A(:))
    if (G%yes_cutoff) then
      do i = 1, N_nuc
        N_c(i) = count(G%RR_tr(:, :, i) .lt. G%r_cutoff)
      end do
      N_c_avg = dble(sum(N_c(:)))/dble(N_nuc)
    else
      N_c(:) = N_nuc
      N_c_avg = dble(N_nuc)
    end if
    if (G%yes_periodic) then
      N_tr = G%N_tr
      allocate (R(N_nuc*N_tr))
    else
      N_tr = 0
      allocate (R(N_nuc))
    end if
    !
    ! print a header
    !
    if (G%yes_cutoff) then
      write (OUTPUT, "('# ',a)") &
        "Connectivity among atoms based on r_cutoff"
      write (OUTPUT, "('#   - (avg) N_c       = ',f10.3)") N_c_avg
      write (OUTPUT, "('#   - R_cutoff (a.u.) = ',f10.3,x,a)") G%r_cutoff, &
        "(applied only for nuclear positions)"
    else
      write (OUTPUT, "('# Connectivity among atoms:')")
      write (OUTPUT, "('#   - N_c ( = N_nuc ) = ',i6)") N_nuc
    end if
    if (G%yes_periodic) then
      write (OUTPUT, "('#   - periodic boundary condition is used')")
      write (OUTPUT, "('#     N of max translation operations = ',i3)") G%N_tr
    end if
    !
    ! find out R_max
    !
    if (G%yes_periodic) then
      R_max = maxval(G%RR_tr(N_tr, :, :))
    else
      R_max = maxval(G%RR(:))
    end if
    if (G%yes_cutoff .and. R_max .gt. G%r_cutoff) R_max = G%r_cutoff
    i_min = 1
    i_max = int(R_max + 0.99)
    allocate (icount(i_min:i_max))
    write (OUTPUT, "('#   R: distance (a.u.), "// &
      "N_c: N_atom to be included in calculations')")
    write (OUTPUT, "('#   ',4x,' ',a5,4a7,' R<',999i4)") "N_c", &
      "1st_R", "2nd_R", "3rd_R", "4th_R", (/(n, n=i_min, i_max)/)
    write (OUTPUT, "('#   ',4x,' ',a)") repeat("-", 36 + (i_max - i_min + 1)*4)
    do i = 1, N_nuc
      !
      ! store all R values into R(:)
      !
      do j = 1, N_nuc
        if (G%yes_periodic) then
          R((j - 1)*N_tr + 1:j*N_tr) = G%RR_tr(:, j, i)
        else
          ij = int(index_packed(i, j))
          R(j) = G%RR(ij)
        end if
      end do
      !
      ! sort R_tr(:) in ascending order
      !
      do j = 1, size(R)
        do k = j + 1, size(R)
          if (R(k) .lt. R(j)) then
            temp = R(k); R(k) = R(j); R(j) = temp
          end if
        end do
      end do
      !
      ! print out first 4 internuclear distances
      !
      write (OUTPUT, "('#   ',a4,':',i5)", advance="no") A(i)%label, N_c(i)
      n = 0
      loop_k: do k = 2, size(R)     ! R(1) must be 0, so skip it
        if (abs(R(k) - R(k - 1)) .gt. 1e-6_long) then
          write (OUTPUT, "(f7.3)", advance="no") R(k)
          n = n + 1
          if (n .ge. 4) exit loop_k
        end if
      end do loop_k
      !
      ! print out counts corresponding to the given distance
      !
      write (OUTPUT, "(3x)", advance="no")
      icount(:) = 0
      do k = 1, size(R)
        n = int(R(k) + 0.99)    ! rounding up
        if (n .lt. i_min) n = i_min
        if (n .gt. i_max) cycle
        icount(n) = icount(n) + 1
      end do
      j = 0
      loop_distance: do n = i_min, i_max
        j = j + icount(n)
        write (OUTPUT, "(i4)", advance="no") icount(n)
        if (j .ge. size(R)) exit loop_distance
      end do loop_distance
      write (OUTPUT, "()")
    end do
    write (OUTPUT, "('#')")
    deallocate (R, icount)
  end subroutine find_neighboring_atoms

end module molecular_grid
