!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module neweffpot
  !! author: Yajiang Hao
  !! author: Ludger Inhester
  !!
  !! This module implement the DIIS in two ways.
  !! (1) construct_Veff(): Simple relaxation + DIIS. The error vector can be
  !! effective or potential; (2) DIIS. DIIS from the begining. The error vector vector
  !! is density matrix. (2) is recommended for the calculation.
  !! @todo: give this module better name

  use constants
  use matrix
  use read_param
  use wavefunction
  use atom_in_mol
  use basis_func
  implicit none
  private
  public :: diis, diis_unrestricted
contains

  subroutine DIIS(P, B, F, NIter)
    !! Direct inversion of iterative subspace [DIIS] procedure
    !! Implement Eqs. (1)-(4) Pulay J. Comp. Chem. 3. 556 (1982)    
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: NIter

    integer :: i, j, k, N

    real(long), allocatable :: BMatrix(:, :) ! B matrix
    real(long), allocatable :: RHS(:) ! Right hand side matrix of our linear equation

    real(long) :: coef(P%DIIS_nvector + 1) ! coefficients

    logical, parameter :: yes_print_diis = .false.
    integer :: nvec, dim, ierr
    real(long) :: factor, trace
    real(long) :: DIISerr(B%N_basis, B%N_basis)
    nvec = P%DIIS_nvector

    DIISerr = (F%H(:, :) .m.F%D(:, :) .m.B%S(:, :)) .sub. (B%S(:, :) .m.F%D(:, :) .m.F%H(:, :))
    if (allocated(F%frozen_MO)) &
      DIISerr(:, :) = F%P_not_frozen.tm.DIISerr(:, :) .m.F%P_not_frozen
    if (B%Nr_basis .lt. B%N_basis) DIISerr(:, :) = B%XXtS.tm.DIISerr(:, :) .m.B%XXtS

    if (NIter .eq. 1) then
      ! if we are in the first iteration allocate arrays save Errormatrix and potential and return
      F%Hsave = ZERO
      F%Errsave(:, :, 1) = DIISerr
      F%Hsave(:, :, 1) = F%H(:, :)
      !     F%Dsave(:,:,1)=F%D(:,:)
      return
    else if (NIter .le. nvec) then
      ! number of iterations is smaller than nvec:
      ! add the new potential and Errormatrix
      F%Errsave(:, :, NIter) = DIISerr(:, :)
      F%Hsave(:, :, NIter) = F%H(:, :)
      !     F%Dsave(:,:,NIter)=F%D(:,:)
      dim = NIter
    else ! Niter <= nvec
      ! number of iterations is larger than nvec:
      ! sweep Errsave and Vsave and add the new Error and potential
      do i = 1, nvec - 1
        F%Hsave(:, :, i) = F%Hsave(:, :, i + 1)
        !        F%Dsave(:,:,i)=F%Dsave(:,:,i+1)
        F%Errsave(:, :, i) = F%Errsave(:, :, i + 1)
      end do
      F%Errsave(:, :, nvec) = DIISerr(:, :)
      F%Hsave(:, :, nvec) = F%H(:, :)
      !     F%Dsave(:,:,nvec)=F%D(:,:)
      dim = nvec
    end if ! Niter <= nvec

    ! we have dim vectors -> the DIIS problem has dimension dim+1
    allocate (BMatrix(dim + 1, dim + 1))
    allocate (RHS(dim + 1))

    ! Fill matrix B and RHS

    RHS(:) = ZERO
    RHS(dim + 1) = -ONE

    !! Feb.25.2015 : K.H : (1) Half of operation (w.r. i and j ) can be skipped using symmetry.
    !!                     (2) You do not need full matrix multiplication (dimension=B%N_basis) since we just need the trace
    !!               this amounts to ~10 sec gain per iteration for 180 atom (660basis) calculations

    Bmatrix(1:dim, dim + 1) = -ONE
    Bmatrix(dim + 1, 1:dim) = -ONE
    Bmatrix(dim + 1, dim + 1) = ZERO
    N = P%N_basis
    do i = 1, dim
      do j = 1, i
        trace = ZERO;
        do k = 1, N
          trace = trace + dot_product(F%Errsave(k, :, i), F%Errsave(:, k, j))
        end do
        BMatrix(j, i) = trace
        BMatrix(i, j) = trace
        !! BMatrix(j,i)=MatTrace( F%Errsave(:,:,i) .mt. F%Errsave(:,:,j) )
        !! BMatrix(i,j)=BMatrix(j,i)
      end do
    end do
    ! Rescaling
    if (BMatrix(1, 1) .gt. ONE) then
      factor = ONE/BMatrix(1, 1)
      do i = 1, dim
        do j = 1, dim
          BMatrix(i, j) = BMatrix(i, j)*factor
        end do
      end do
    end if

    ! solve the linear equation B*coef=RHS
    coef = ZERO
    call solve_AxB(BMatrix, RHS, coef, ierr)
    ! if something went wrong don't just use simple relaxation
    if (ierr .lt. 0) then
      coef = ZERO
      coef(dim) = ONE
    end if ! ( ierr < 0 )

    ! build up the potential from the coefficients
    F%H = ZERO
    !  F%D=ZERO
    do i = 1, dim
      F%H(:, :) = F%H(:, :) + coef(i)*F%Hsave(:, :, i)
      !     F%D(:,:) = F%D(:,:) + coef(i) * F%Dsave(:,:,i)
    end do ! i = 1, nvec

    ! deallocate
    deallocate (BMatrix)
    deallocate (RHS)
  end subroutine DIIS

  subroutine DIIS_unrestricted( P, B, F, NIter) 
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: NIter
    
    integer :: i, j, k, N
    
    real(long), allocatable :: BMatrix(:,:) ! B matrix
    real(long), allocatable :: RHS(:) ! Right hand side matrix of our linear equation
  
    real(long) :: coef(P%DIIS_nvector+1) ! coefficients 
    
    logical, parameter :: yes_print_diis = .false.
    integer :: nvec, dim, ierr
    real(long) :: factor, trace
    real(long) :: DIISerra(B%N_basis, B%N_basis)
    real(long) :: DIISerrb(B%N_basis, B%N_basis) 
    nvec = P%DIIS_nvector
    
    DIISerra = (F%focka(:,:) .m. F%Da(:,:)  .m. B%S(:,:)) .sub. (B%S(:,:) .m. F%Da(:,:) .m. F%focka(:,:))
    DIISerrb = (F%fockb(:,:) .m. F%Db(:,:)  .m. B%S(:,:)) .sub. (B%S(:,:) .m. F%Db(:,:) .m. F%fockb(:,:))
    if(allocated(F%frozen_MO))then
      DIISerra(:,:) = F%P_not_frozen .tm. DIISerra(:,:) .m. F%P_not_frozen
      DIISerrb(:,:) = F%P_not_frozen .tm. DIISerrb(:,:) .m. F%P_not_frozen     
    end if
    if(B%Nr_basis < B%N_basis)then
      DIISerra(:,:) = B%XXtS .tm. DIISerra(:,:) .m. B%XXtS
      DIISerrb(:,:) = B%XXtS .tm. DIISerrb(:,:) .m. B%XXtS
    end if
    if ( NIter == 1 ) then
      ! if we are in the first iteration allocate arrays save Errormatrix and potential and return
      F%Hsave=ZERO
      F%Errsave(:,:,1) = DIISerra
      F%Hsave(:,:,1)=F%focka(:,:)
      F%Errsave(:,:,2) = DIISerrb
      F%Hsave(:,:,2)=F%fockb(:,:)
      return
    else if( NIter .le. nvec  ) then
      ! number of iterations is smaller than nvec:
      ! add the new potential and Errormatrix
      F%Errsave(:,:,NIter * 2 - 1) = DIISerra(:,:)
      F%Hsave(:,:,NIter * 2 - 1) = F%focka(:,:)
      F%Errsave(:,:,NIter * 2) = DIISerrb(:,:)
      F%Hsave(:,:,NIter * 2) = F%fockb(:,:)
      dim=NIter
    else ! 2*Niter <= nvec
      ! number of iterations is larger than nvec:
      ! sweep Errsave and Vsave and add the new Error and potential
      do i=1,nvec-1
        F%Hsave(:, :, 2 * i - 1) = F%Hsave(:, :, 2 * i + 1)
        F%Errsave(:, :, 2 * i - 1) = F%Errsave(: ,: , 2 * i + 1)
        F%Hsave(:, :, 2 * i) = F%Hsave(: , :, 2 * i + 2)
        F%Errsave(:, :, 2 * i) = F%Errsave(:, :, 2 * i + 2)
      end do
      F%Errsave(:, :, 2 * nvec - 1) = DIISerra(:, :)
      F%Hsave(:, :, 2 * nvec - 1) = F%focka(:, :)
      F%Errsave(:, :, 2 * nvec) = DIISerrb(:, :)
      F%Hsave(:, :, 2 * nvec) = F%fockb(:, :)
      dim = nvec
    end if ! 2*Niter <= nvec
    
    ! we have dim vectors -> the DIIS problem has dimension dim+1
    allocate(BMatrix(dim+1,dim+1))
    allocate(RHS(dim+1))
    
    ! Fill matrix B and RHS
    
    RHS(:)=ZERO
    RHS(dim+1)=-ONE
    
    !! Feb.25.2015 : K.H : (1) Half of operation (w.r. i and j ) can be skipped using symmetry.
    !!                     (2) You do not need full matrix multiplication (dimension=B%N_basis) since we just need the trace
    !!               this amounts to ~10 sec gain per iteration for 180 atom (660basis) calculations 
    
    Bmatrix(1:dim, dim+1) = -ONE
    Bmatrix(dim+1, 1:dim) = -ONE
    Bmatrix(dim+1,dim+1)  = ZERO
    N=B%N_basis 
    do i=1,dim
      do j=1,i
        trace=ZERO;
        do k=1,N
          trace=trace+ dot_product( F%Errsave(k,:,2 * i - 1), F%Errsave(:,k, 2 * j - 1) ) ! alpha part
          trace=trace+ dot_product( F%Errsave(k,:,2 * i), F%Errsave(:,k, 2 * j) ) ! beta part
        enddo
        BMatrix(j,i) = trace
        BMatrix(i,j) = trace
      enddo
    enddo
    ! Rescaling
    if(BMatrix(1,1) > ONE) then 
      factor = ONE/BMatrix(1,1)
      do i=1,dim
        do j=1,dim
          BMatrix(i,j) = BMatrix(i,j) * factor
        end do
      end do
    end if
    !call print_matrix(4*BMatrix)
    
    ! solve the linear equation B*coef=RHS 
    coef=ZERO
    call solve_AxB(BMatrix, RHS, coef, ierr)
    ! if something went wrong don't just use simple relaxation
    if ( ierr < 0 ) then
      coef      = ZERO
      coef(dim) = ONE
    end if ! ( ierr < 0 )
    
    ! build up the potential from the coefficients
    F%focka=ZERO
    do i = 1, dim
      F%focka(:,:) = F%focka(:,:) + coef(i) * F%Hsave(:,:,2 * i - 1)
    end do ! i = 1, nvec

    F%fockb=ZERO
    do i = 1, dim
      F%fockb(:,:) = F%fockb(:,:) + coef(i) * F%Hsave(:,:,2 * i)
    end do ! i = 1, nvec
    ! deallocate
    deallocate(BMatrix)
    deallocate(RHS)
  end subroutine DIIS_unrestricted

! optimal damping algorithm
  subroutine ODA(H, Hold, D, Dold)
    real(long), intent(inout) :: H(:, :)
    real(long), intent(in) :: Hold(:, :)
    real(long), intent(inout) :: D(:, :)
    real(long), intent(in) :: Dold(:, :)

    real(long) :: lambda

    lambda = -HALF* &
      matTrace(Matmul(Hold(:, :), D(:, :) - Dold(:, :))) &
      /matTrace(MAtmul(H(:, :) - Hold(:, :), D(:, :) - Dold(:, :)))

    if (lambda .ge. 1) lambda = 1
    write (STDERR, *) 'lambda=', lambda, &
      matTrace(Matmul(H(:, :) - Hold(:, :), (D(:, :) - Dold(:, :)))), &
      matTrace(matmul(Hold(:, :), (D(:, :) - Dold(:, :))))
    D = lambda*D + (1 - lambda)*Dold
    H = lambda*H + (1 - lambda)*Hold

  end subroutine

! calculates the extra/interpolation coefficient according to
! Computers & Chemistry
! Volume 2, Issue 1, 1978, Pages 31–33
  function damping_coefficient(D, D_old, D_old2) result(lambda)
    real(long), intent(in) :: D(:, :)
    real(long), intent(in) :: D_old(:, :)
    real(long), intent(in) :: D_old2(:, :)

    real(long), allocatable :: DD(:, :)
    real(long), allocatable :: DD_old(:, :)
    real(long) :: s1, s2, s3, lambda
    integer :: N, i
    N = size(D, 1)
    allocate (DD(N, N))
    allocate (DD_old(N, N))

    DD(:, :) = D(:, :) - D_old(:, :)
    DD_old(:, :) = D_old(:, :) - D_old2(:, :)

    s1 = ZERO
    s2 = ZERO
    s3 = ZERO
    do i = 1, N
      s1 = s1 + sum(DD(i, :)*DD_old(i, :))
      s2 = s2 + sum(DD_old(i, :)*DD_old(i, :))
      s3 = s3 + sum(DD(i, :)*DD(i, :))
    end do
    deallocate (DD_old)
    deallocate (DD)
    lambda = s1/s2

    if (abs(lambda) .gt. 1) then
      lambda = s1/s3
    end if
  end function damping_coefficient


end module neweffpot
