!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module nuclear_derivatives
  !! author: Kota Hanasaki
  !!
  !! This module calculates nuclear coordinate derivatives of total Energy (T=0) or the grand potential (T/=0)
  !!   written by Kota Hanasaki in 2014 at DESY CFEL
  !!   being modified by Sang-Kil Son since 2019
  !!
  use constants
  use sph_harmonics
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use interpolate
  use anglib
  use operations
  use gps
  use wavefunction
  use scf
  use legendre
  use packed_index
  use matrix
  implicit none

  !>
  !!  Data Structure to calculate 3-center integrals
  !!
  type localGrid
    integer                :: Ngloc_Max, NnucUnq, Ng_loc
    integer, allocatable    :: iOff(:), nuc_indices(:)
    real(long), allocatable :: wgDcmp(:), weight(:)
  end type localGrid

  private 
  public :: calculate_dEdR 
contains

!! -------------------------------------------------------------------
!! fd2toSD2 : converts a full(L,L) matrix to packed (L*(L+1)/2) array
!! -------------------------------------------------------------------
  subroutine fd2toSD2(F2, SD2, L)
    implicit none
    integer :: L, i, j, ij
    real(long) :: F2(L, L), SD2(L*(L + 1)/2)
    ij = 0; do i = 1, L; do j = 1, i; ij = ij + 1; 
      SD2(ij) = F2(i, j); 
    end do; end do
  end subroutine

!! -------------------------------------------------------------------
!! calculate_dEnuc_dR :  derivative of Nuc-Nuc repulsion energy
!!                A  A+B Z_j Z_i
!! E_nuc = (1/2) SUM SUM -------  NOTE: convention of i and j are flipped over
!!                j i/=j  R_ji          DON'T FORGET a factor of 1/2!!!
!!
!! d E_nuc    A+B (X_i - X_j)
!! ------- =  SUM ----------- Z_j Z_i   NOTE: be careful with a sign of
!!  d X_j    i/=j   R_ji^3                    ( X_i - X_j )
!!                                      WHERE IS 1/2?  SEE BELOW
!! -------------------------------------------------------------------
  function calculate_dEnuc_dR(A, j_nuc) result(ret)
    implicit none
    real(long), dimension(3) :: ret
    type(Atom_set), intent(in) :: A(:)
    integer, intent(in) :: j_nuc
    integer :: i_nuc
    real(long) :: XX, YY, ZZ, fac
    ret(:) = ZERO
    do i_nuc = 1, size(A)
      if (i_nuc .eq. j_nuc) cycle
      XX = A(i_nuc)%x - A(j_nuc)%x
      YY = A(i_nuc)%y - A(j_nuc)%y
      ZZ = A(i_nuc)%z - A(j_nuc)%z
      fac = dble(A(i_nuc)%fcharge*A(j_nuc)%fcharge) &
            /sqrt(XX**2 + YY**2 + ZZ**2)**3
      ret(:) = ret(:) + (/XX, YY, ZZ/)*fac
    end do
  end function calculate_dEnuc_dR

  function calculate_dEnuc_dR_periodic(G, A, j_nuc) result(ret)
    implicit none
    real(long), dimension(3) :: ret
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    integer, intent(in) :: j_nuc
    integer :: i_nuc, k, N_tr
    real(long) :: R_AB, R(3)
    !if( G%yes_periodic_Fuzzy ) then
    !  E_nuc = ZERO
    !  ret(:) = ZERO
    !  do i_nuc = 1, size(A)
    !    ZiZj = dble( A(i_nuc)%fcharge * A(j_nuc)%fcharge )
    !    if( i_nuc == j_nuc ) cycle
    !    XX = A(i_nuc)%x - A(j_nuc)%x
    !    YY = A(i_nuc)%y - A(j_nuc)%y
    !    ZZ = A(i_nuc)%z - A(j_nuc)%z
    !    xT4x8(:,:) = translation_in_crystal_Fuzzy( G, XX, YY, ZZ )
    !    do k=1,8; r8(k)= sqrt( sum( xT4x8(1:3,k)**2 ) ); enddo
    !    if ( G%yes_cutoff .and. minval(r8(:)) > G%r_cutoff ) cycle
    !    do k=1,8;  E_nuc = E_nuc + xT4x8(4,k)*ZiZj/r8(k);
    !       ret(1:3)=ret(1:3) + xT4x8(4,k)*xT4x8(1:3,k)*ZiZj/ (r8(k)**3); enddo
    !  end do
    !  return
    !end if
    !
    ! added by Sang-Kil on Apr. 25, 2019
    ! NOTE: bond_distance() and obtain_nuclear_position() can be used
    !       not only for periodic but also for non-periodic
    !
    ret(:) = ZERO
    N_tr = 1
    if (G%yes_periodic) N_tr = G%N_tr
    loop_i: do i_nuc = 1, size(A) ! i_nuc <-> j_nuc in operations.f90
      loop_tr: do k = 1, N_tr     ! N_tr = 1 for the non-periodic case
        !
        ! skip itself but DON'T SKIP LOWER-TRIANGLES; see the equation below
        !
        if (k .eq. 1 .and. j_nuc .eq. i_nuc) cycle loop_tr
        !
        ! skip the rest if a bond distance is larger than r_cutoff
        !
        R_AB = bond_distance(G, j_nuc, i_nuc, k)  ! <-- good scanning order
        if (G%yes_cutoff .and. R_AB .gt. G%r_cutoff) exit loop_tr
        ! NOTE: R(:) = R_i_nuc(:) - R_j_nuc(:)
        !
        !R(:) = obtain_nuclear_position( G, i_nuc, j_nuc, k )
        R(:) = -obtain_nuclear_position(G, j_nuc, i_nuc, k) ! <-- keep order
        ret(:) = ret(:) + R(:)*A(j_nuc)%fcharge*A(i_nuc)%fcharge/R_AB**3
      end do loop_tr
    end do loop_i
  end function calculate_dEnuc_dR_periodic

  real(long) function Ftrace(A, L)
    implicit none
    integer :: L
    real(long) :: A(L, L)
    integer :: i
    Ftrace = ZERO; 
    do i = 1, L; Ftrace = Ftrace + A(i, i); End do
  end function
!> -------------------------------------------------------------------
!! calculate_dEdR : MAIN routine
!!     Jnuc_want(Nwant) : iNuc indices of your need
  !integer :: m_Nnuc = -1
  !data m_Nnuc/-1/
!!     ret(3,Nwant)     : contains dE/dR vectors
!! -------------------------------------------------------------------
  subroutine calculate_dEdR(ret, P, A, G, B, F, Jnuc_want, Nwant, &
                            dS_out, dH_out, dJH_out, dVxc_out, Ja_out, &
                            yes_printout, read_VeeD, yes_FockExchange)
    use constants
    implicit none
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in)       :: Nwant
    integer, dimension(Nwant) :: Jnuc_want
    !integer, optional :: iKEonly
    real(long), dimension(3, Nwant) :: ret
    real(long), dimension(3, Nwant, 5) :: parts
    real(long), allocatable :: dOdR(:, :, :), dOdR_0(:, :, :), &
                               wks_Nb(:), E_occ(:)
    real(long), parameter :: Vex_factor_LDA = (THREE/PI)**(ONE/THREE), &
                             Vex_factor_Slater = (THREE/TWO)*(THREE/PI)**(ONE/THREE)
    real(long), allocatable :: dvdR_J(:, :)
    real(long) :: Vex_factor, dvdR_terms(3)
    Logical, Optional, intent(in) :: yes_printOut; logical :: prtOut
    logical :: yes_d2AO_incl_angl, yes_real_sph_wo_sign
    integer :: iflag_d1d3, iflag_dhdR0, iflag_dSdR0, Jnuc_d, &
               jwant, k, Nbasis, kk, Nadd, out
  !! VeeD ----------------------
    integer :: Nb2, i
    logical, optional :: yes_FockExchange; logical :: fockExchange
    logical, optional :: read_VeeD; logical :: useVeeD
  !! 08.12.2014 ---
    real(long), allocatable :: dOdR_p(:, :)
    logical :: dbgng = .FALSE., checkTime = .FALSE.   !! <DEBUG OPTIONS>
    integer :: Nocc

  !! debug options --------------
    INTEGER :: KOUT
    INTEGER, OPTIONAL :: Ja_out(2)
    REAL(LONG), OPTIONAL, INTENT(out) :: dS_out(B%N_basis*(B%N_basis + 1)/2), dH_out(B%N_basis*(B%N_basis + 1)/2), &
                                         dJH_out(B%N_basis*(B%N_basis + 1)/2), dVxc_out(B%N_basis*(B%N_basis + 1)/2)
    INTEGER, PARAMETER :: iDBGMODE = 0    !!  1=Time Flush  2= CompToOldOne
  !! Time
    REAL :: tm0, tm1, tmes(0:5), tmLas, tmCur
!  Write(*,*)"#calculate_dEdR:rev1412.1 : uses F-quantities"
  !! Write(*,*)"#calculate_dEdR:Rnuc:",(A(i)%x,A(i)%y,A(i)%z,i=1,P%N_nuc)
  !! Write(*,*)"#calculate_dEdR:Nocc:",(F%N_occ(:))
  !! Write(*,*)"#calculate_dEdR:F%CE:",(F%CE(:))
  !! Write(*,*)"#OrbEngSum:",Ftrace(matmul(F%H,F%D),F%N_basis),sum(F%CE(:)*F%N_occ(:))
    if (P%yes_verbose) checkTime = .true.
    call cpu_time(tm0); 
    tmLas = tm0

!! -- Constants --
    Nbasis = B%N_basis
    Nb2 = Nbasis*(Nbasis + 1)/2
  !! we want N upto where we have to sum over
!  if(P%yes_thermal) then; Nocc=Nbasis
!  else;                   do i=Nbasis,1,-1; if(F%n_occ(i)/=0)then; Nocc=i; exit; endif; enddo;        endif
    ! < Here we check f_occ for either case >
    if (P%yes_thermal) then;
      Nocc = Nbasis
    else; 
      do i = Nbasis, 1, -1; 
        if (F%f_occ(i) .gt. EPS) then; 
          Nocc = i; 
          exit; 
        end if; 
      end do; 
    end if

    if( maxval( dabs(F%f_occ(:)-F%n_occ(:)) )>EPS .and. dbgng ) &
      write(*,'(a,999f8.2)')"#calculate_dEdR:dbgng:fractional occ:",F%f_occ(:)

!! -- Options --
    FockExchange = .false.; 
    if (present(yes_FockExchange)) fockExchange = yes_FockExchange
    useVeeD = .false.
    if (present(read_VeeD)) useVeeD = read_VeeD
    If (FockExchange) useVeeD = .true.

    KOUT = -999; if (present(Ja_out)) KOUT = Ja_out(2)

!! --Check Input --
!  if( useVeeD ) then
!    if(.not. FockExchange ) then; ldum=isVeeD_available(Atoms=A)
!    else;                         ldum=isVeeD_available(Atoms=A, yes_Exchange=.true.); end if
!    Call assertf(ldum,"calculate_dEdR:VeeD unavailable...");
!  end if

    IF (PRESENT(Ja_out) .or. PRESENT(dS_out) .or. PRESENT(dH_out) .or. PRESENT(dJH_out)) then; 
      write (*, *) "#calculate_dEdR: !WW DEBUG OPTION EXISTS..", &
        PRESENT(Ja_out), PRESENT(dS_out), PRESENT(dH_out), PRESENT(dJH_out)
    END IF
    !IF (PRESENT(iKEonly)) write (*, *) "! XXX iKEonly Selected ---", iKEonly, "---"

    !Exchange_type_ = FUNCTIONAL_X_NAME(P%x_type)

    prtOut = .False.; if (present(yes_printout)) prtOut = yes_printout

  !! Just a reminder ...
    Vex_factor = Vex_factor_LDA
    if (trim(FUNCTIONAL_X_NAME(P%x_type)) .ne. "LDA") then
      if (trim(FUNCTIONAL_X_NAME(P%x_type)) .eq. "Slater") then
      !! write(0,*)"#calculate_dEdR: Switch to Slater...";
        Vex_factor = Vex_factor_Slater
      else
        write (*, *) "!E //// calculate_dEdR : WARNING ////////////////////////////////////////////////////////////////////"
        write (*, *) "!E -- non-LDA/Slater exchange is not supported...."; STOP
      end if
    end if
    yes_d2AO_incl_angl = .true.
    yes_real_sph_wo_sign = .true.
    iflag_d1d3 = 1       !  do not calculate d3fdR

    allocate (dOdR(B%N_basis, B%N_basis, 3), &
              dOdR_0(B%N_basis, B%N_basis, 3), wks_Nb(Nbasis))

    dOdR = ZERO
    dOdR_0 = ZERO
    wks_Nb = ZERO

    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(0) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(0):", tmes(0)

    allocate (dvdR_J(G%N_grid, 3))
    iflag_dhdR0 = 0; 
    do jwant = 1, Nwant
      Jnuc_d = Jnuc_want(jwant)
      call calculate_dhdR_nucJ_X(dOdR, dOdR_0, iflag_dhdR0, F%v_ext, A, &
                                   P, G, B, F, &
                                   B%dfdR, Jnuc_d, yes_skip_dvdR=.true.)
      iflag_dhdR0 = 1
      !
      ! calculate the external potential part
      !
      if (P%yes_periodic) then
        call calculate_dv_ext_dR_periodic(dvdR_J, G, A(Jnuc_d))
      else
        call calculate_dv_ext_dR(dvdR_J, G, A(Jnuc_d))
      end if
      forall (k=1:3)
        dvdR_terms(k) = sum(G%weight(:)*F%rho(:)*dvdR_J(:, k))
      end forall
      !
      ! construct the part 1
      !   contract_S2( A, B, WKS, N ) = sum_i (AB)_ii when A and B are symmetric
      !
      do k = 1, 3         ! x, y, z
        parts(k, jwant, 1) = contract_S2(F%D, dOdR(:, :, k), wks_Nb, Nbasis) + &
                             dvdR_terms(k)
      end do
      !
      ! debugging purpose
      !
      IF (PRESENT(dH_out) .and. present(Ja_out)) then
        if (Jnuc_d .eq. Ja_out(1 &
                               )) call fd2toSD2(dOdR(:, :, kOut), dH_out, Nbasis)
      end if ! XXX debugging XXX
    end do
    deallocate (dvdR_J)
    !
    ! check cpu time
    !
    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(1) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(1):", tmes(1)

    !----------------------------------------------------------------------------
    ! 2.  Hartree energy ---
    !    V_eff = electronic Hartree potential
    !----------------------------------------------------------------------------
    if (useVeeD) then
      !    do k=1,3;    Call get_VeeD(k,D3wrk,Dp,Nbasis)
      !      do jwant=1,Nwant; Jnuc_d = Jnuc_want(jwant)
      !        VDHp=mask_Matr(D3wrk,Nbasis,Jnuc_d)
      !        parts(k,jwant,2)=DpTrace( Dp,VDHp, Nbasis)
      !      end do
      !    end do
    else
      if (P%yes_gaussian) &
        F%v_J = potential_J(P, G, A, F%rho) ! work around for gaussians
    !! V_eff = potential_J_beta( P, G, A, rho)  !! <revision 5/11/2014>
      do jwant = 1, Nwant
        Jnuc_d = Jnuc_want(jwant)
        call calculate_Gmunu_F(dOdR, G, B, B%dfdR, Jnuc_d, F%v_J)
        !
        ! construct the part 2
        !   contract_S2( A, B, WKS, N ) = sum_i (AB)_ii when A and B are symmetric
        !   actually, contract_S2 = sum_i sum_j A_ij B_ij
        !
        do k = 1, 3       ! x, y, z
          parts(k, jwant, 2) = contract_S2(F%D, dOdR(:, :, k), wks_Nb, Nbasis)
        end do
        !
        ! debugging purpose
        !
        IF (PRESENT(dJH_out) .and. present(Ja_out)) then
          if (Jnuc_d .eq. Ja_out(1 &
                                 )) call fd2toSD2(dOdR(:, :, kOut), dJH_out, Nbasis)
        end if ! XXX debugging XXX
      end do
    end if
    !
    ! check cpu time
    !
    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(2) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(2):", tmes(2)

    !----------------------------------------------------------------------------
    ! 3.  Exchange energy ---
    !    V_eff = exchange potential
    !----------------------------------------------------------------------------
    if (FockExchange) then
      !  do k=1,3;    Call get_VeeD(k,D3wrk,Dp,Nbasis,yes_Exchange=.true.)
      !    do jwant=1,Nwant; Jnuc_d = Jnuc_want(jwant)
      !      VDHp = mask_Matr(D3wrk,Nbasis,Jnuc_d)
      !      parts(k,jwant,3)=DpTrace( Dp,VDHp, Nbasis);
      !    end do
      !  end do
    else
    !! <12/02:Use_Fstructure>  V_eff(:) = - Vex_factor * rho(:)**(ONE/THREE)      !! (THREE/PI)**(ONE/THREE)
      do jwant = 1, Nwant
        Jnuc_d = Jnuc_want(jwant)
        call calculate_dVeffdR_nucJ(dOdR, F%v_xc, G, B, B%dfdR, Jnuc_d)
        !
        ! construct the part 3
        !   contract_S2( A, B, WKS, N ) = sum_i (AB)_ii when A and B are symmetric
        !   actually, contract_S2 = sum_i sum_j A_ij B_ij
        !
        do k = 1, 3
          parts(k, jwant, 3) = contract_S2(F%D, dOdR(:, :, k), wks_Nb, Nbasis)
        end do
        !
        ! debugging purpose
        !
        IF (PRESENT(dVxc_out) .and. present(Ja_out)) then
          if (Jnuc_d .eq. Ja_out(1 &
                                 )) call fd2toSD2(dOdR(:, :, kOut), dVxc_out, Nbasis)
        end if ! XXX debugging XXX
      end do
    end if
    !
    ! check cpu time
    !
    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(3) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(3):", tmes(3)

!!!! -- Sep.11 fix --
!!!!    - TWO* \sum_{i} f_{i} S_{ii} e_{i} = -\sum_{i} (nOcc_{i}e_{i}) S_{ii}
  !! <12/02:Use_Fstructure>     deallocate(V_eff)
    allocate (E_occ(Nbasis))
    E_occ = ZERO
    if (P%yes_thermal) then         !! 27/11/2014:fix use f_occ if THERMAL
      E_occ(1:Nbasis) = F%f_occ(1:Nbasis)*F%CE(1:Nbasis)
    else
      Nadd = min(Nbasis, F%N_basis)
      if (Nbasis .ne. F%N_basis) then; write (0, '(A,2I7)') "#calculate_dEdR:N_orb/=Nbasis: adds up to the smaller one:", F%N_basis, Nbasis; end if
      E_occ(1:Nadd) = F%f_occ(1:Nadd)*F%CE(1:Nadd)
    end if

    !----------------------------------------------------------------------------
    ! 4.  - n_{i} C^{mu}_{i} C^{nu}_{i} e_{i} S'_{ii}  ---
    !----------------------------------------------------------------------------
    deallocate (dOdR, WKS_Nb)
    allocate (dOdR_p(Nb2, 3), WKS_Nb(Nb2)); 
    dOdR_p = ZERO
    iflag_dSdR0 = 0; 
    call prepare_EWDM(WKS_Nb, F%LC, E_occ, Nbasis, Nocc)
    do jwant = 1, Nwant
      Jnuc_d = Jnuc_want(jwant)
      call calculate_dSdR_nucJ_p(dOdR_p, dOdR_0, iflag_dSdR0, &
                                 G, B, B%dfdR, Jnuc_d)
      iflag_dSdR0 = 1
     !!!! --Sep.11 fix    do k=1,3; parts(k,jwant,4)=-TWO*contract_S3(F%LC,dOdR(:,:,k),F%CE,wks_Nb,Nbasis);end do
      if (sum(F%N_occ) .gt. 0) then
        !
        ! construct the part 4
        !
        do k = 1, 3
          parts(k, jwant, 4) = -contract_S3p(dOdR_p(:, k), WKS_Nb, Nbasis)
!original code by Kota:
!        parts(k,jwant,4)= -contract_S3p( F%LC, dOdR_p(:,k), E_occ, &
!                                         WKS_Nb, Nbasis, Nocc )
        end do
        !! contract_S3(F%LC,dOdR(:,:,k),E_occ,wks_Nb,Nbasis);
      end if
      !IF( dbgng ) then
      !   if(.not. allocated(dOdR) )then; allocate(dOdR(Nbasis,Nbasis,3)); endif
      !   call calculate_dSdR_nucJ(dOdR, dOdR_0,iflag_dSdR0, G, B, B%dfdR, Jnuc_d)
      !   write(*,*)"#dEdR:dbgng:NEW:",parts(1:3,jwant,4)
      !   write(*,*)"#dEdR:dbgng:OLD:",( contract_S3(F%LC,dOdR(:,:,k),E_occ,wks_Nb,Nbasis),k=1,3)
      !
      ! debugging purpose
      !
      IF (PRESENT(dS_out) .and. present(Ja_out)) then
        if (Jnuc_d .eq. Ja_out(1 &
                               )) call fd2toSD2(dOdR(:, :, kOut), dS_out, Nbasis)
      end if ! XXX debugging XXX
    end do
    !
    ! check cpu time
    !
!TEST
! contract_S3p takes the most of computational time
!  write(STDERR, "('4: calculate_dSdR_nucJ_p ',a)") trim(timestamp(time_table(1)))
!  write(STDERR, "('4: contract_S3p          ',a)") trim(timestamp(time_table(2)))
!  write(STDERR, "('4: fd2toSD2              ',a)") trim(timestamp(time_table(3)))
!TEST
    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(4) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(4):", tmes(4)

    !----------------------------------------------------------------------------
    ! 5. dEnuc/dR
    !----------------------------------------------------------------------------
    if (P%yes_periodic) then    ! Periodic extension : dEnuc/dR ..
      do jwant = 1, Nwant
        Jnuc_d = Jnuc_want(jwant)
        parts(:, jwant, 5) = calculate_dEnuc_dR_periodic(G, A, Jnuc_d)
      end do
    else
      do jwant = 1, Nwant
        Jnuc_d = Jnuc_want(jwant)
        parts(:, jwant, 5) = calculate_dEnuc_dR(A, Jnuc_d)
      end do
    end if
    !
    ! check cpu time
    !
    call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(5) = tmCur
    if (iDBGMODE .eq. 1) Write (0, *) "#calculate_dEdR:dbgng:Time:(5):", tmes(5)

    !----------------------------------------------------------------------------
    ! sum up all contributions
    !----------------------------------------------------------------------------
    do jwant = 1, Nwant
      do k = 1, 3         ! x, y, z
        ret(k, jwant) = &
          parts(k, jwant, 1) + & ! sp contribution...  first term in Eq. (4)
          parts(k, jwant, 2) + & ! vH contribution... second term in Eq. (4)
          parts(k, jwant, 3) + & ! vEx contribution... third term in Eq. (4)
          parts(k, jwant, 4) + & ! dS contribution... fourth term in Eq. (4)
          parts(k, jwant, 5)     ! Enuc contribution
      !!if(present(non_interacting_part) .and. non_interacting_part/=0 )then
      !!            ret(k,jwant)= parts(k,jwant,1) + parts(k,jwant,4) + parts(k,jwant,5)
      !!end if
      end do
    end do

    if (prtOut) then
    !!if(present(non_interacting_part) .and. non_interacting_part/=0 )then
    !!    write(*,*)"#Energy Derivatives:!WWW non-interacting part"
    !!else
      out = P%OUTPUT
      if (P%yes_thermal) then; write (out, '(A,es16.8,A,999f9.3)') "#SCF Free Energy Derivatives T=", P%T_eV, "(eV)  average occ=", F%f_occ(:)
      else; write (out, '(A,999i1)') "#SCF Ground State Energy Derivatives: for N_occ=", F%n_occ(:)  ! only N_occ in this code for printing
      end if
     write(out,'(A,es16.8,A,9999f12.6)')            "#                                         E_SCF=",F%E_tot,"  nuclear coordinates=",(A(kk)%x,A(kk)%y,A(kk)%z,kk=1,P%N_nuc)
    !!end if
      do jWant = 1, nWant; 
        write (out, '(A,I4,A1,3es18.10)') "#Nuc          ", Jnuc_want(jWant), ":", ret(1:3, jWant)
        if (P%yes_verbose) then
          write (out, '(A)') "#                   ------------------------------------------------------"
          write (out, '(A,3es18.10)') "#  sp -contribution:", parts(1:3, jWant, 1)
          write (out, '(A,3es18.10)') "#  vH -contribution:", parts(1:3, jWant, 2)
          write (out, '(A,3es18.10)') "#  vEx-contribution:", parts(1:3, jWant, 3)
          write (out, '(A,3es18.10)') "#  dS -contribution:", parts(1:3, jWant, 4)
          write (out, '(A,3es18.10)') "#Enuc -contribution:", parts(1:3, jWant, 5)
        end if
        !write(out,*)
      end do

    end if
! suppress unnecessary output
!  if(checkTime) Write(*,'(A,f12.6,A,6f12.6)')"#calculate_dEdR:time:",sum(tmes(0:5))," dfdR/h1P/JH/Vx/dS/check-integ:",tmes(0:5)
    if (iDBGMODE .eq. 1) Then; Write (0, '(A,f12.6,A,6f12.6)') "#calculate_dEdR:DBGNG:Time:", sum(tmes(0:5)), &
 " dfdR/h1P/JH/Vx/dS/check-integ:", tmes(0:5); 
      FLUSH (0); End if
  !! <12/02:Use_Fstructure>  deallocate(rho, Dmat, dfdR, dOdR, dOdR_0, wks_Nb, E_occ)
    deallocate (dOdR_p, dOdR_0, wks_Nb, E_occ)
    !if (allocated(yes_dS_overlap__)) deallocate (yes_dS_overlap__); ! write(*,*)"#calculate_dEdR:DBGNG:de-allocate yes_overlap";
    !Exchange_type_ = "--"
  end subroutine

!! -------------------------------------------------------------------
!! contract_S2 : internal use only
!!     \sum \sum D_{i,j} G_{i,j} where D and G are both SYMMETRIC real matrices
!!  we use summation over dot products sum_{j} DOT( D(:,j) G(:,j) )
!! -------------------------------------------------------------------
  real(long) function contract_S2(D, G, WKS, N) result(ret)
    implicit none
    integer, intent(in) :: N
    real(long), dimension(N, N), intent(in) :: D, G
    real(long), dimension(N) :: WKS
    integer :: i
    forall (i=1:N) WKS(i) = dot_product(D(:, i), G(:, i))
    ret = sum(WKS)
  end function contract_S2

!----------------------------------------------------------------------
! calculate EWDM (energy-weighted density matrix)
!   := SUM_i f_i epsilon_i C_i^mu C_i^nu
!
!   C    : MO coefficient matrix  <== F%LC(:,:)
!   E    : f_i epsilon_i          <== E_occ(:)= F%f_occ(:) * F%CE(:)
!   L    : N of basis functions   <== Nbasis
!   Nocc : N of occupied orbitals <== Nocc
!----------------------------------------------------------------------
  subroutine prepare_EWDM(EWDM, C, E, L, Nocc)
    implicit none
    integer, intent(in) :: L, Nocc
    real(long), dimension(L*(L + 1)/2), intent(out) :: EWDM
    real(long), dimension(L, Nocc), intent(in) :: C        ! MO coeff.
    real(long), dimension(Nocc), intent(in) :: E          ! orbital E * #occ.
    integer :: i, j, ij, k
    EWDM(:) = ZERO
    do k = 1, Nocc        ! index i
      ij = 0
      do i = 1, L         ! mu
        do j = 1, i       ! nu
          ij = ij + 1     ! mu,nu index (lower triangular)
          EWDM(ij) = EWDM(ij) + C(i, k)*C(j, k)*E(k)
        end do
      end do
    end do
  end subroutine prepare_EWDM

!----------------------------------------------------------------------
! calculate the fourth part of Eq. (4)
!                                          d < mu | nu >
!   2 SUM f_i epsilon_i  SUM C_i^mu C_i^nu -------------
!      i                mu,nu                   d q
!             d < mu | nu >
!   = 2  SUM  ------------- SUM f_i epsilon_i C_i^mu C_i^nu
!       mu,nu      d q       i
!                           -------------------------------
!               this is the most time-consuming part in the gradient calc.
!               ==> it must be calculated ONCE!!!
!               ==> separated out with prepare_EWDM
!                   (energy-weighted density matrix)
!
! This routine is called by
!   -contract_S3p( dOdR_p(:,k), WKS_Nb, Nbasis, Nocc )
! Note that "-" at the beginning.
!
! Gp   : d<mu|nu>/dq            <== dOdR_p(:,k)
! DE   : the most time-consuming part <== WKS_Nb
! L    : N of basis functions   <== Nbasis
! Nocc : N of occupied orbitals <== Nocc
!----------------------------------------------------------------------
  real(long) function contract_S3p(Gp, DE, L) result(ret)
    implicit none
    integer, intent(in) :: L
    real(long), dimension(L*(L + 1)/2), intent(in) :: Gp
    real(long), dimension(L*(L + 1)/2), intent(in) :: DE
    integer :: i, ij
    real(long) :: dag
    dag = ZERO
    ij = 0
    do i = 1, L           ! mu
      ij = ij + i         ! mu,mu index (diagonal)
      dag = dag + DE(ij)*Gp(ij)
    end do
    ret = 2*dot_product(DE, Gp) - dag
  end function contract_S3p
! -----------------------------------------------------
! -- Part II :: derivative-matrix calculation codes ---
! -----------------------------------------------------

! --- II-[1] :: dh/dR ---
! if( iflag_dhdR <=0 ) calculates dhdR0(\mu,nu) := < nabla \mu | h | \nu >
! after that it returns correct combination of dhdR0 according to Jnuc_d.
!
! 05/12/2014 : separated out dv/dR part : you have to add \int \rho dv/dR separately...
!
  subroutine calculate_dhdR_nucJ_X(ret, dhdR0, iflag_dhdR, v_ext, A, P, G, B, F, dfdR, Jnuc_d, &
                                  yes_skip_dvdR, &
                                  iKEonly)
  !!use nuclear_derivatives_Add, only:calculate_dhdR_0_xtn
    implicit none
    integer, intent(in) :: Jnuc_d, iflag_dhdR
    type(Atom_set), intent(in) :: A(:)
    !type(Atom_set), pointer :: Aj
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Param), intent(in) :: P

    real(long), dimension(G%N_grid, B%N_basis, 3), intent(in) :: dfdR
    real(long), dimension(B%N_basis), intent(in) :: v_ext
    real(long), dimension(B%N_basis, B%N_basis, 3), intent(out) :: ret

    LOGICAL, OPTIONAL :: YES_skip_dvdR

    real(long), dimension(B%N_basis, B%N_basis, 3), intent(inout) :: dhdR0

    real(long), allocatable :: dvdR_J(:, :)
    !real(long), dimension(:, :, :), pointer :: p_dhdR0
    !real(long), dimension(:, :, :), allocatable, target :: temp_dhdR0
    logical :: yes_dealloc
    integer :: kxyz, MU, nu_, jb0, jb1
    real(long) :: add
    Real :: tm1, tm0, tmlas, tmCur, tmes(3); logical :: checkTime = .false.
  !! option-1 : restrict to KE
    integer, intent(in), optional :: iKEonly  ! if( present(iKEonly).and. iKEonly==1,-1,-2,0 : KE,PE,dPE,sum)
  !! option-2 : iSymOption=0(none), 1, 2
    !integer, intent(in), optional :: iSymOption
    !real(long), intent(in), optional, dimension(G%N_grid, B%N_basis, 3) :: d3fdR
    !real(long), intent(in), optional, dimension(G%N_grid, B%N_basis, 6) :: d2fdR
    integer :: N_nuc
    N_nuc=size(A)

    !  IF(NCALL==0) THEN; WRITE(*,*)"### dhdR0: EXCLUDES 3-Center Integ"; NCALL=NCALL+1; ENDIF
    !Aj => A_all(Jnuc_d)
  !! if(Jnuc_d<3) checkTime=.true.
    if (checkTime) then; call cpu_time(tm0); tmLas = tm0; end if

    !p_dhdR0 => dhdR0; yes_dealloc = .false.
    !if (iflag_dhdR .lt. 0) then
    !  allocate (temp_dhdR0(B%N_basis, B%N_basis, 3)); yes_dealloc = .true.
    !  p_dhdR0 => temp_dhdR0
    !end if

    if (iflag_dhdR .le. 0 .or. present(iKEonly)) then

      !if (present(iSymoption)) then; 
        !call calculate_dhdR_0_xtn(dhdR0, v_ext, G,B,dfdR,iKEonly,d2fdR,d3fdR,iSymOption) ! future extension
      !  stop 'not implemented'
      !else; 
        !if (allocated(yes_dS_overlap__)) then; 
        !  call calculate_dhdR_0(p_dhdR0, v_ext, G, B, dfdR, yes_dS_overlap__, iKEonly); 
        !else; 
        !  if (N_nuc .gt. 0) then
             !! allocate( yes_dS_overlap__(N_nuc,B%N_basis,B%N_basis,3) )
            !call gen_dS_ovlp(yes_dS_overlap__, P, A_all, G, B, F, dfdR)                                ! Sep.24.2015:yes_periodic/yes_cutoff   parameter changed

        call calculate_dhdR_0(dhdR0, v_ext, G, B, dfdR)

        !  else
        !    call calculate_dhdR_0(p_dhdR0, v_ext, G, B, dfdR, iKEonly=iKEonly); 
        !  end if
        !end if
      !end if
    end if
    if (checkTime) then; call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(1) = tmCur; end if
!! ---   allocate(dvdR_J(G%N_grid,3))
!! ---   call calculate_dv_ext_dR(dvdR_J, G, Aj )

    if (checkTime) then; call cpu_time(tm1); tmCur = tm1 - tmLas; tmLas = tm1; tmes(2) = tmCur; end if
    ret = ZERO
    jb0 = B%first_bf_on_atom(Jnuc_d)       ! first basis index that belongs to Jnuc_d
    jb1 = B%first_bf_on_atom(Jnuc_d) + B%num_bf_on_atom(Jnuc_d) - 1   ! last  basis index that belongs to Jnuc_d
    do kxyz = 1, 3
      do MU = jb0, jb1
        do nu_ = 1, B%N_basis
          ret(nu_, MU, kxyz) = dhdR0(nu_, MU, kxyz)  ! MU is differenciated
          !if(MU==B%N_basis .and. nu_==B%N_basis .and. kxyz==1 )&
          !        write(*,'(A,2I4,e18.9)')"dhdR0 contribution:",MU,nu_,ret(nu_,MU,kxyz)
          !if(MU==B%N_basis .and. nu_==B%N_basis .and. kxyz==1 )&
          !        write(*,'(A,2I4,e18.9)')"dhdR0 contribution:",MU,nu_,ret(nu_,MU,kxyz)
        end do
      end do
      do MU = 1, B%N_basis
        do nu_ = 1, MU
!! ---        add= integral(G, B%func(:,MU)* dvdR_J(:,kxyz) *B%func(:,nu_) )
!! ---        if( present(iKEonly) .and. (iKEonly==1 .or. iKEonly==-1 ))add=ZERO
          ret(MU, nu_, kxyz) = ret(MU, nu_, kxyz) + ret(nu_, MU, kxyz)  !!! +add
          ret(nu_, MU, kxyz) = ret(MU, nu_, kxyz)
        end do
      end do
    end do
 !! if( checkTime )then; call cpu_time(tm1);tmCur=tm1-tmLas;tmLas=tm1; tmes(3)=tmCur; endif
    if( iflag_dhdR<=0 .and. checkTime ) Write(*,'(A,f9.3,A,3f9.3)')"#dhdR_0::time:",sum(tmes(1:3))," dhdR0,dvdR,integ:",tmes(1:3)

    if (present(yes_skip_dvdR))then 
      if(yes_skip_dvdR) return
    end if

  !! you still need this for CPHF calculations...
    allocate (dvdR_J(G%N_grid, 3))
    call calculate_dv_ext_dR(dvdR_J, G, A(Jnuc_d))
    do kxyz = 1, 3; 
      do MU = 1, B%N_basis; 
        do nu_ = 1, MU
          add = braket(G, B, MU, nu_, dvdR_J(:, kxyz))
          ret(MU, nu_, kxyz) = ret(MU, nu_, kxyz) + add
          ret(nu_, MU, kxyz) = ret(MU, nu_, kxyz)
        end do; 
      end do; 
    end do
    !integral(G, B%func(:,MU)* dvdR_J(:,kxyz) *B%func(:,nu_) )

    !if (yes_dealloc) deallocate (temp_dhdR0)
  end subroutine

  subroutine calculate_dv_ext_dR(dv_ext_dR, G, Aj)
    implicit none
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: Aj
    real(long), dimension(G%N_grid, 3) :: dv_ext_dR
    real(long) :: Zj
    !
    ! V = sum_i ( - Z_i / r_i )
    ! dV/dX_J^{a} = - ( x^{a}_{el} - X^{a}_{J} ) Z_{J} / | r_{el}-R_{J} |^{3}
    !
    !
    Zj = dble(Aj%fcharge)
    dv_ext_dR(:, 1) = -Zj/ &
                      sqrt((G%x(:) - Aj%x)**2 + (G%y(:) - Aj%y)**2 + (G%z(:) - Aj%z)**2)**3

    dv_ext_dR(:, 2) = (G%y(:) - Aj%y)*dv_ext_dR(:, 1)
    dv_ext_dR(:, 3) = (G%z(:) - Aj%z)*dv_ext_dR(:, 1)

    dv_ext_dR(:, 1) = (G%x(:) - Aj%x)*dv_ext_dR(:, 1)
!  dv_ext_dR(:,1) = - (G%x(:)-Aj%x)/ sqrt( ( G%x(i)-Aj%x )**2 + ( G%y(i)-Aj%y )**2 + ( G%z(i)-Aj%z )**2  )**3
!  dv_ext_dR(:,2) = - (G%y(:)-Aj%y)/ sqrt( ( G%x(i)-Aj%x )**2 + ( G%y(i)-Aj%y )**2 + ( G%z(i)-Aj%z )**2  )**3
!  dv_ext_dR(:,3) = - (G%z(:)-Aj%z)/ sqrt( ( G%x(i)-Aj%x )**2 + ( G%y(i)-Aj%y )**2 + ( G%z(i)-Aj%z )**2  )**3
  end subroutine calculate_dv_ext_dR

!
! this is asymmetric matrix with  dhdR0_{nu,mu} = <  chi_{nu} | h | (- d/dr^{a}_{e}) chi_{mu} >
!                                               = <  (- d/dr^{a}_{e}) chi_{mu} | h | chi_{nu} >
!
  subroutine calculate_dhdR_0(dhdR0, v_ext, G, B, dfdR, iKEonly)
    implicit none
!!!  integer,intent(in),optional :: iKEonly
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in), dimension(G%N_grid) :: v_ext
    real(long), intent(in), dimension(G%N_grid, B%N_basis, 3) :: dfdR
    real(long), dimension(B%N_basis, B%N_basis, 3) :: dhdR0
    integer :: kxyz, mu, nu, Nbasis
    !logical, optional, intent(in) :: yes_overlap(G%N_nuc, B%N_basis, B%N_basis, 3)
    integer, optional :: iKEonly
    logical, allocatable :: dummy_ovlp(:)
    dhdR0 = ZERO
    Nbasis = B%N_basis

    if (present(iKEonly)) stop "calculate_dhdR_0: No KEonly option" !! although its easy to implement

    !if (.not. present(yes_overlap)) then
    !  write (*, *) "#calculate_dhdR_0:uses all-ovlp scheme"
    !  allocate (dummy_ovlp(G%N_nuc)); dummy_ovlp = .true.; 
    !  forall (nu=1:Nbasis, mu=1:Nbasis, kxyz=1:3)
    !    dhdR0(nu, mu, kxyz) = braket_df_h1P_f_b(G, B, v_ext, dfdR(:, :, kXYZ), mu, nu, G%N_nuc, G%i_start, G%i_end, dummy_ovlp)
    !  end forall
    !  return
    !end if

    forall (nu=1:Nbasis, mu=1:Nbasis, kxyz=1:3)
!!         KE=-HALF* integral(G, dfdR(:,mu,kxyz)*B%D2f(:,nu))
!!         PE=integral( G, dfdR(:,mu,kxyz)*v_ext(:)*B%func(:,nu) )
      !dhdR0(nu, mu, kxyz) = braket_df_h1P_f_b(G, B, v_ext, dfdR(:, :, kXYZ), mu, nu, G%N_nuc, G%i_start, G%i_end, &      
      !                                        yes_overlap(:, nu, mu, kxyz))
      dhdR0(nu, mu, kxyz) = braket_df_h1P_f_b(G, B, v_ext, dfdR(:, :, kXYZ), mu, nu, G%N_nuc, G%i_start, G%i_end, &
                                              B%yes_overlap(:,index_packed(mu, nu)))
!        dhdR0(nu,mu)= -HALF* integral(G, dfdR(:,mu,kxyz)*B%D2f(:,nu)) &
!                      + integral( G, dfdR(:,mu,kxyz)*v_ext(:)*B%func(:,nu) )
    end forall
  !!stop
  end subroutine

!! ------ Optimization -------
!! pure real(long) braket_df_###_f_b (G,B,(V,), dfdR, i,j, N_nuc, iStart, iEnd) result(value)
!!
!!    istart/iend are G%i_start/G%i_end
!!
  pure real(long) function braket_df_h1P_f_b(G, B, V, dfdR, i, j, N_nuc, istart, iend, yes_overlap) result(value)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in) :: V(G%N_grid)
    integer, intent(in) :: N_nuc, istart(N_nuc), iend(N_nuc)
    real(long), intent(in) :: dfdR(G%N_grid, B%N_basis)   !! this is dfdR(:,:,kXYZ)
    integer, intent(in) :: i, j
    real(long) :: wks(N_nuc)
    integer :: k_nuc
    logical, intent(in) :: yes_overlap(N_nuc)

    !if (dS_full_sum__) then
    !  VALUE = sum(G%weight(:)*dfdR(:, i)*(-HALF*B%D2f(:, j) + v(:)*B%func(:, j)))
    !  RETURN
    !end if

    !
    wks = ZERO
  !! ij = index_packed( i, j )   ! index for packed storage of upper triang.
    ! IMaj=max(i,j);jmno=min(i,j);  ij=jmno+ IMaj*(IMaj-1)/2
    ! forall( k_nuc=1:N_nuc,B%yes_overlap(k_nuc,ij) )
    do k_nuc = 1, N_nuc
      if (.NOT. yes_overlap(k_nuc)) cycle
      !forall( k_nuc=1:N_nuc,yes_overlap(k_nuc) )
      wks(k_nuc) = sum(G%weight(istart(k_nuc):iend(k_nuc))* &
                       dfdR(istart(k_nuc):iend(k_nuc), i)*(-HALF*B%D2f(istart(k_nuc):iend(k_nuc), j) &
                                                           + v(istart(k_nuc):iend(k_nuc))*B%func(istart(k_nuc):iend(k_nuc), j)))
    end do
!end forall
    value = sum(wks)
  end function braket_df_h1P_f_b

!! real(long) function braket_df_h1P_f_s( G, B, V, dfdR, i, j, N_nuc,istart,iend) result (value)

  pure real(long) function braket_df_V_f_b(G, B, V, dfdR, i, j, N_nuc, istart, iend, yes_overlap) result(value)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in) :: V(G%N_grid)
    integer, intent(in) :: N_nuc, istart(N_nuc), iend(N_nuc)
    real(long), intent(in) :: dfdR(G%N_grid, B%N_basis)   !! this is dfdR(:,:,kXYZ)
    integer, intent(in) :: i, j
    real(long) :: wks(N_nuc)
    integer :: k_nuc
    logical, intent(in) :: yes_overlap(N_nuc)
    !

    !if (dS_full_sum__) then
    !  VALUE = sum(G%weight(:)*dfdR(:, i)*v(:)*B%func(:, j))
    !  RETURN
    !end if

    wks = ZERO
  !! ij = index_packed( i, j )   ! index for packed storage of upper triang.
  !! IMaj=max(i,j);jmno=min(i,j);  ij=jmno+ IMaj*(IMaj-1)/2
    do k_nuc = 1, N_nuc
      if (.NOT. yes_overlap(k_nuc)) cycle

      !forall( k_nuc=1:N_nuc,yes_overlap(k_nuc) )
      wks(k_nuc) = sum(G%weight(istart(k_nuc):iend(k_nuc))* &
                       dfdR(istart(k_nuc):iend(k_nuc), i)* &
                       v(istart(k_nuc):iend(k_nuc))*B%func(istart(k_nuc):iend(k_nuc), j))
      !  end forall
    end do
    value = sum(wks)
  end function braket_df_V_f_b

!! FN : braket_df_ovlp_f_b( G, B, dfdR, i, j, N_nuc,istart,iend,yes_overlap)
!!   calculates   dfdR(:,i) B%func(:,j)
!!   this uses  yes_overlap  to truncate dfdR * f integral.
!!
  pure real(long) function braket_df_ovlp_f_b(G, B, dfdR, i, j, N_nuc, istart, iend, yes_overlap) result(value)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    integer, intent(in) :: N_nuc, istart(N_nuc), iend(N_nuc)
    real(long), intent(in) :: dfdR(G%N_grid, B%N_basis)   !! this is dfdR(:,:,kXYZ)
    integer, intent(in) :: i, j
    real(long) :: wks(N_nuc)
    integer :: k_nuc, ij, iMaj, jmno
    logical, intent(in) :: yes_overlap(N_nuc)
    !
    !if (dS_full_sum__) then
    !  VALUE = sum(G%weight(:)*dfdR(:, i)*B%func(:, j))
    !  RETURN
    !end if

    wks = ZERO
  !! ij = index_packed( i, j )   ! index for packed storage of upper triang.
    IMaj = max(i, j); jmno = min(i, j); ij = jmno + IMaj*(IMaj - 1)/2
    do k_nuc = 1, N_nuc
      if (.NOT. yes_overlap(k_nuc)) cycle
      !forall( k_nuc=1:N_nuc,yes_overlap(k_nuc) )
      wks(k_nuc) = sum(G%weight(istart(k_nuc):iend(k_nuc))* &
                       dfdR(istart(k_nuc):iend(k_nuc), i)*B%func(istart(k_nuc):iend(k_nuc), j))
      !  end forall
    end do
    value = sum(wks)

  end function braket_df_ovlp_f_b
!! real(long) function braket_df_h1P_f( G, B, V, dfdR, i, j) result (value)

!
! calculates packed  dSdR_nucJ
!            out of dSdR0
!
  subroutine calculate_dSdR_nucJ_p(ret, dSdR0, iflag_dSdR0, G, B, dfdR, Jnuc_d)
    implicit none
    integer, intent(in) :: Jnuc_d, iflag_dSdR0
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), dimension(G%N_grid, B%N_basis, 3), intent(in) :: dfdR
    real(long), dimension(B%N_basis*(B%N_basis + 1)/2, 3) :: ret
    real(long), dimension(B%N_basis, B%N_basis, 3), target :: dSdR0
    real(long), dimension(:, :, :), pointer :: p_dSdR0
    real(long), allocatable, target :: temp_dSdR0(:, :, :)
    logical :: yes_dealloc
    integer :: kxyz, MU, nu_, jb0, jb1, mnu, mnuOff, jb0_1, Nbasis

    Nbasis = B%N_basis

    p_dSdR0 => dSdR0; yes_dealloc = .false.
    if (iflag_dSdR0 .lt. 0) then
      allocate (temp_dSdR0(B%N_basis, B%N_basis, 3)); 
      yes_dealloc = .true.
      p_dSdR0 => temp_dSdR0
    end if
    if (iflag_dSdR0 .le. 0) then
      call calculate_dSdR_0(p_dSdR0, G, B, dfdR)
    end if

    ret = ZERO
    jb0 = B%first_bf_on_atom(Jnuc_d)       ! first basis index that belongs to Jnuc_d
    jb1 = B%first_bf_on_atom(Jnuc_d) + B%num_bf_on_atom(Jnuc_d) - 1   ! last  basis index that belongs to Jnuc_d
    jb0_1 = jb0 - 1
    mnuOff = jb0*(jb0 - 1)/2
    do kxyz = 1, 3
      mnu = mnuOff; 
      do MU = jb0, jb1; 
        do nu_ = 1, MU; mnu = mnu + 1
          ret(mnu, kXYZ) = p_dSdR0(nu_, MU, kxyz)
        end do
      end do
      mnu = mnuOff; 
      do MU = jb0, jb1; 
        mnu = mnu + jb0_1; 
        do nu_ = jb0, MU; mnu = mnu + 1
          ret(mnu, kXYZ) = ret(mnu, kXYZ) + p_dSdR0(MU, nu_, kXYZ); 
          !call ASSERTF(mnu==ijSym(nu_,MU),"dSdR-1",-1)
        end do; 
      end do
      if (jb1 .lt. Nbasis) then
        do MU = jb1 + 1, Nbasis
          mnu = mnu + jb0_1; 
          do nu_ = jb0, jb1; mnu = mnu + 1
            ret(mnu, kXYZ) = p_dSdR0(MU, nu_, kXYZ); 
            !call ASSERTF(mnu==ijSym(nu_,MU),"dSdR-2",-1)
          end do
          mnu = mnu + (MU - jb1)
        end do
      end if
    end do

    if (yes_dealloc) deallocate (temp_dSdR0)
  end subroutine

!
! this is asymmetric matrix with  dSdR0_{nu,mu} = <chi_{nu}|- d/dr^{a}_{e} chi_{mu}  >
!
  subroutine calculate_dSdR_0(dSdR0, G, B, dfdR, no_opt)   !! no_opt performs default (high-cost but safe) integral
                                                      !! with-opt requires some preparation..  look up "yes_dS_overlap__" etc.
    implicit none
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in), dimension(G%N_grid, B%N_basis, 3) :: dfdR
    real(long), dimension(B%N_basis, B%N_basis, 3) :: dSdR0
    LOGICAL, OPTIONAL :: no_opt
    integer :: kxyz, mu, nu, Nbasis
    Nbasis = B%N_basis

    if (present(no_opt)) then
      if (no_opt) then
        do kxyz = 1, 3; do mu = 1, B%N_basis; do Nu = 1, B%N_basis; 
            dSdR0(nu, mu, kxyz) = integral(G, dfdR(:, mu, kxyz)*B%func(:, nu))
          end do; end do; end do
        return
      end if
    end if

!  do kxyz=1,3
!    do mu=1,B%N_basis
!      do nu=1,B%N_basis
    forall (nu=1:Nbasis, mu=1:Nbasis, kxyz=1:3)
      !dSdR0(nu, mu, kxyz) = braket_df_ovlp_f_b(G, B, dfdR(:, :, kXYZ), mu, nu, G%N_nuc, G%i_start, G%i_end, &      
      !                                         yes_dS_overlap__(:, nu, mu, kXYZ))
      dSdR0(nu, mu, kxyz) = braket_df_ovlp_f_b(G, B, dfdR(:, :, kXYZ), mu, nu, G%N_nuc, G%i_start, G%i_end, &
                                               B%yes_overlap(:,index_packed(mu,nu)))
    end forall
!= integral( G, dfdR(:,mu,kxyz)*B%func(:,nu) )
!      end do
!    end do
!  end do
  end subroutine

! --- II-[3] :: d { mu nu || rho sigma }/dR  where {..||..} is Chemists' notation of two electron integral ---
!
! Here we calculate  G_{munu}[F] := \int dr1 dr2 { d[ chi_{mu}(r1) chi_{nu}(r1) ]/dR_{J} 1/r_{12} F(r2)
! using Ya-jian's code to integrate over r_{2} :  \int dr2 1/r_{12} F(r_{2}) = \sum_{lm} Y_{lm}(omega_{1}) V_{lm}(r1)
!
! for Energy derivative, you can simply use  F(r2)=rho(r2) and then contract as G_{munu} by half of the density matrix D_{munu}
! whereas for CPHF,      it is    Ci^{mu} Cj^{nu} G_{munu} [rho] + D_{munu} G_{munu} [ rho_{ij} ]
!                                                                                      with   rho_{ij}(r) = MO_{i}(r) MO_{j}(r)
!

! Note added : You are NOT supposed to make four-leg matrices... always try some contraction with respect to -F-

  subroutine calculate_Gmunu_F(ret, G, B, dfdR, Jnuc_d, vF_0)
    !use Vee_core, only:potential_J_beta
    implicit none
    integer, intent(in) :: Jnuc_d
    type(Basis), intent(in) :: B
    type(Grid3D), intent(in) :: G
    real(long), intent(in), dimension(G%N_grid, B%N_basis, 3) :: dfdR
    real(long), dimension(B%N_basis, B%N_basis, 3) :: ret
    real(long), dimension(G%N_grid), optional, intent(in), target :: vF_0
    real(long), allocatable, dimension(:), target :: vF_loc
    real(long), pointer :: p_vF(:)
    integer :: kxyz, mU, nu_, jb0, jb1

    ret = ZERO
    jb0 = B%first_bf_on_atom(Jnuc_d)       ! first basis index that belongs to Jnuc_d
    jb1 = B%first_bf_on_atom(Jnuc_d) + B%num_bf_on_atom(Jnuc_d) - 1   ! last  basis index that belongs to Jnuc_d
!!write(*,*)"calculate_Gmunu_F:",present(vF_0),Jnuc_d,jb0,jb1
    if (present(vF_0)) then; p_vF => vF_0
    else; 
!    allocate( vF_loc(G%N_grid) ); p_vF=>vF_loc; p_vF=potential_J_beta( P, G, A, F );
      stop "something is wrong here!"
    end if !! <revision 5/11/2014>

    do kxyz = 1, 3
      do MU = jb0, jb1
        do nu_ = 1, B%N_basis
          !ret(nu_, MU, kxyz) = braket_df_V_f_b(G, B, p_vF, dfdR(:, :, kXYZ), mu, nu_, G%N_nuc, G%i_start, G%i_end, &          
          !                                     yes_dS_overlap__(:, nu_, MU, kxyz))
          ret(nu_, MU, kxyz) = braket_df_V_f_b(G, B, p_vF, dfdR(:, :, kXYZ), mu, nu_, G%N_nuc, G%i_start, G%i_end, &
                                               B%yes_overlap(:,index_packed(mu,nu_)))
!!integral(G, dfdR(:,MU,kxyz)* p_vF(:) *B%func(:,nu_) )
        end do
      end do
      do MU = 1, B%N_basis
        do nu_ = 1, MU
          ret(MU, nu_, kxyz) = ret(MU, nu_, kxyz) + ret(nu_, MU, kxyz)
          ret(nu_, MU, kxyz) = ret(MU, nu_, kxyz)
        end do
      end do
    end do

    if (.not. present(vF_0)) deallocate (vF_loc)
  end subroutine

! --- II-[4] :: d <mu|Veff|nu>/dR  ** with no derivative operation on V ** ---
!     Veff=Vex is required for dE/dR
!
  subroutine calculate_dVeffdR_nucJ(ret, Veff, G, B, dfdR, Jnuc_d)
    implicit none
    integer, intent(in) :: Jnuc_d
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in), dimension(G%N_grid) :: Veff
    real(long), dimension(B%N_basis, B%N_basis, 3) :: ret
    real(long), dimension(G%N_grid, B%N_basis, 3), intent(in) :: dfdR
    integer :: kxyz, mU, nu_, jb0, jb1, Nbasis
    Nbasis = B%N_basis
    ret = ZERO
    jb0 = B%first_bf_on_atom(Jnuc_d)       ! first basis index that belongs to Jnuc_d
    jb1 = B%first_bf_on_atom(Jnuc_d) + B%num_bf_on_atom(Jnuc_d) - 1   ! last  basis index that belongs to Jnuc_d

    do kxyz = 1, 3
      forall (nu_=1:Nbasis, MU=jb0:jb1)
        ret(nu_, MU, kxyz) = braket_df_V_f_b(G, B, Veff, dfdR(:, :, kXYZ), mu, nu_, G%N_nuc, G%i_start, G%i_end, &
                                             B%yes_overlap(:,index_packed(mu,nu_)))

      end forall
      do MU = 1, B%N_basis
        do nu_ = 1, MU
          ret(MU, nu_, kxyz) = ret(MU, nu_, kxyz) + ret(nu_, MU, kxyz)
          ret(nu_, MU, kxyz) = ret(MU, nu_, kxyz)
        end do
      end do
    end do

  end subroutine

  subroutine calculate_dv_ext_dR_periodic(dv_ext_dR, G, A)
    implicit none
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A
    real(long), dimension(G%N_grid, 3) :: dv_ext_dR
    real(long) :: x, y, z, r, fac
    real(long) :: R_AB, r_tr(3)
    real(long) :: xT4x8(4, 8), r8(8)
    integer :: i, k, N_tr
    integer :: i_nuc
    !
    ! V = sum_J ( - Z_J / r_J )
    ! dV/dX_J^{a} =  - ( x^{a}_{el} - trans(R_{J},r)^{a} ) Z_{J} /
    !                  | r_{el}- trans(R_{J},r) |^{3}
    ! ... we NEGLECT singular behavior of d trans / dX
    !
    dv_ext_dR(:, :) = ZERO
    if (G%yes_periodic_Fuzzy) then
      do i = 1, G%N_grid
        x = G%x(i) - A%x; y = G%y(i) - A%y; z = G%z(i) - A%z; 
        xT4x8(:, :) = translation_in_crystal_Fuzzy(G, x, y, z)
        do k = 1, 8; r8(k) = sqrt(sum(xT4x8(1:3, k)**2)); end do
        IF (G%yes_cutoff) then
          if (minval(r8(:)) .gt. G%r_cutoff) cycle; 
        END IF
        do k = 1, 8; fac = -xT4x8(4, k)*dble(A%fcharge)/r8(k)**3
          dv_ext_dR(i, 1) = dv_ext_dR(i, 1) + xT4x8(1, k)*fac; 
          dv_ext_dR(i, 2) = dv_ext_dR(i, 2) + xT4x8(2, k)*fac; 
          dv_ext_dR(i, 3) = dv_ext_dR(i, 3) + xT4x8(3, k)*fac; end do
      end do
      return
    end if
    loop_i: do i = 1, G%N_grid
      !
      ! added by Sang-Kil on Apr. 25, 2019 and Apr. 26, 2019
      ! use the same procedure as potential_ext() in operations.f90
      !
      i_nuc = G%i_nuc(i)
      N_tr = 1            ! no translation if G%yes_periodic is false
      if (G%yes_periodic) N_tr = G%N_tr
      loop_tr: do k = 1, N_tr
        R_AB = bond_distance(G, A%i_nuc, i_nuc, k)  ! <-- good scanning order
        if (G%yes_cutoff .and. R_AB .gt. G%r_cutoff) exit loop_tr  ! skip the rest
        r_tr(:) = obtain_grid_position(G, i, A%i_nuc, k)        ! (x,y,z)
        r = sqrt(sum(r_tr(:)**2))     ! r=sqrt( x**2 + y**2 + z**2 )
        dv_ext_dR(i, :) = dv_ext_dR(i, :) - r_tr(:)*A%fcharge/r**3
      end do loop_tr
    end do loop_i
  end subroutine calculate_dv_ext_dR_periodic

end module

