!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module operations
  !!   author: Yajiang Hao
  !!   author: Ludger Inhester
  !!
  !! This module contain initial_density, density, density matrix, molecular
  !! orbial density, wavefunction of molecular orbital, external potential,
  !! direct Coulomb potential, single-center integral, single-center Coulomb
  !! potential, Checking fidelity of molecular orbitals, and some other things
  
  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use matrix
  use eigen_lapack
  use averaging
  use packed_index
  use anglib
  use stdio
  use legendre
  use basis_func
  use sph_harmonics
  implicit none
  private
  public :: get_complement_combinations, get_combinations, potential_J, potential_ext, density, occ_space, &
    initial_density, prodOverlap, initial_density_matrix, groundstate_occ, fidelity, &
    nuclear_repulsion_energy, nuclear_repulsion_energy_crystal, long_range_energies, &
    single_excitation_occs, single_excitation_occs2, nuclear_repulsion_gradient, &
    auto_active_space, occ_overlap2_highspin_core, find_optimal_deltaD, &
    general_active_space_occs, active_space_occs, filter_sym_occs, density_matrix, filter_ras_occs, &
    rotate_av_orbitals, print_atomic_quantity, potential_J_A, potential_J_MC, integral, integral_a, &
    prodOverlapUnrestricted

contains

  subroutine occ_overlap2_highspin_core(occ, occs, LC1, LC2, S12, Ncore, overlap2)
    !! returns overlap square between configurations in occ
    !! to configurations in occs with different molecular orbitals and basis sets
    !! considering only high-spin configurations
    integer, intent(in) :: occ(:)
    real(long), intent(in) :: LC1(:, :)
    real(long), intent(in) :: LC2(:, :)
    real(long), intent(in) :: S12(:, :)
    integer, intent(in) :: occs(:, :)
    integer, intent(in) :: Ncore
    real(long), intent(out) :: overlap2(size(occs, 2))
    integer :: i, j, k
    real(long), dimension(:, :), allocatable :: Oalpha! one particle overlap matrix for the double occupied space
    real(long), dimension(:, :), allocatable :: Obeta! one particle overlap matrix for the double+single occupied space
    real(long), dimension(:, :), allocatable :: O! one particle overlap matrix all orbitals
    integer :: Nclosed1, Nopen1, Nalpha, Nbeta ! Number of single / double occupied MOs, alpha/beta MOs
    integer :: Nclosed2, Nopen2 ! Number of single / double occupied MOs
    integer, allocatable :: index_alpha1(:)
    integer, allocatable :: index_beta1(:)
    integer, allocatable :: index_alpha2(:)
    integer, allocatable :: index_beta2(:)
    integer:: multp
    real(long):: olap
    integer :: N
    integer :: Nopencore1
    integer :: Nopencore2

    integer, allocatable :: combinations1(:, :)
    integer, allocatable :: complement1(:, :)
    integer, allocatable :: index_open1(:)
    integer, allocatable :: combinations2(:, :)
    integer, allocatable :: complement2(:, :)
    integer, allocatable :: index_open2(:)

    allocate (O(size(occ, 1), size(occ, 1)))
    ! calculate all MO overlaps
    do i = 1, size(occ, 1)
      do j = 1, size(occ, 1)
        O(i, j) = dot_product(LC1(:, i), MATMUL(S12, LC2(:, j)))
      end do
    end do
    Nclosed1 = count(occ(:) .eq. 2)
    Nopen1 = count(occ(:) .eq. 1)
    Nopencore1 = count(occ(:Ncore) .eq. 1)
    N = 0
    overlap2 = ZERO
    Nopen1 = count(occ(:) .eq. 1)
    ! loop over all possible spin multiplicities for elec. configuration occ
    N = 0
    do multp = Nopen1 + 1, Nopencore1 + 1, -2
      if (Nopen1 .gt. 0) then
        Nalpha = Nclosed1 + (Nopen1 + multp - 1)/2
        Nbeta = Nclosed1 + (Nopen1 - multp + 1)/2
      else
        Nalpha = Nclosed1
        Nbeta = Nclosed1
      end if
      allocate (index_open1(Nopen1))
      allocate (Oalpha(Nalpha, Nalpha))
      allocate (Obeta(Nbeta, Nbeta))
      allocate (index_alpha1(Nalpha))
      allocate (index_beta1(Nbeta))
      allocate (index_alpha2(Nalpha))
      allocate (index_beta2(Nbeta))

      index_open1(1:Nopen1) = pack((/(j, j=1, size(occ, 1))/), occ(:) .eq. 1)
      index_beta1(1:Nclosed1) = pack((/(j, j=1, size(occ, 1))/), occ(:) .eq. 2)
      index_alpha1(1:Nclosed1) = pack((/(j, j=1, size(occ, 1))/), occ(:) .eq. 2)

      call get_combinations(Nopen1 - Nopencore1, Nalpha - Nclosed1 - Nopencore1, combinations1)
      call get_complement_combinations(Nopen1 - Nopencore1, Nalpha - Nclosed1 - Nopencore1, combinations1, complement1)

      !write(6,*) multp,size(combinations1,2),Nopencore1

      if (Nalpha .ne. Nbeta) then
        N = N + 2*size(combinations1, 2)
      else
        N = N + size(combinations1, 2)
      end if

      do j = 1, size(combinations1, 2)

        index_alpha1(Nclosed1 + 1:Nclosed1 + Nopencore1) = index_open1(1:Nopencore1)
        index_alpha1(Nclosed1 + Nopencore1 + 1:) = index_open1(Nopencore1 + combinations1(:, j))
        index_beta1(Nclosed1 + 1:) = index_open1(Nopencore1 + complement1(:, j))
        do i = 1, size(occs, 2)

          Nclosed2 = count(occs(:, i) .eq. 2)
          Nopen2 = count(occs(:, i) .eq. 1)
          Nopencore2 = count(occs(:Ncore, i) .eq. 1)

          if (Nopen2 .ge. Nalpha - Nbeta .and. Nopencore2 .eq. Nopencore1) then
            allocate (index_open2(Nopen2))
            index_open2(1:Nopen2) = pack((/(j, j=1, size(occ, 1))/), occs(:, i) .eq. 1)

            index_alpha2(1:Nclosed2) = pack((/(j, j=1, size(occ, 1))/), occs(:, i) .eq. 2)
            index_beta2(1:Nclosed2) = pack((/(j, j=1, size(occ, 1))/), occs(:, i) .eq. 2)

            call get_combinations(Nopen2 - Nopencore2, Nalpha - Nclosed2 - Nopencore2, combinations2)
            call get_complement_combinations(Nopen2 - Nopencore2, Nalpha - Nclosed2 - Nopencore2, combinations2, complement2)
            do k = 1, size(combinations2, 2)
              index_alpha2(Nclosed2 + 1:Nclosed2 + Nopencore2) = index_open2(1:Nopencore2)
              index_alpha2(Nclosed2 + Nopencore2 + 1:) = index_open2(Nopencore2 + combinations2(:, k))
              index_beta2(Nclosed2 + 1:) = index_open2(Nopencore2 + complement2(:, k))

              !write(6,'(99i1)') occ(:)
              !write(6,'(a,99i3)') "alpha1 ",index_alpha1
              !write(6,'(a,99i3)') "beta1  ",index_beta1
              !write(6,'(a)') "->"
              !write(6,'(99i1)') occs(:,i)
              !write(6,'(a,99i3)') "alpha2 ",index_alpha2
              !write(6,'(a,99i3)') "beta2  ",index_beta2
              Oalpha(1:Nalpha, 1:Nalpha) = O(index_alpha1(:), index_alpha2(:))
              Obeta(1:Nbeta, 1:Nbeta) = O(index_beta1(:), index_beta2(:))
              olap = matrix_determinant(Oalpha)*matrix_determinant(Obeta)
              !write(6,*) olap
              if (Nalpha .ne. Nbeta) then
                overlap2(i) = overlap2(i) + TWO*olap**2
              else
                overlap2(i) = overlap2(i) + olap**2
              end if
            end do
            deallocate (index_open2)
            deallocate (combinations2)
            deallocate (complement2)
          end if
        end do! i loop
      end do! j loop
      deallocate (Oalpha)
      deallocate (Obeta)
      deallocate (index_alpha1)
      deallocate (index_alpha2)
      deallocate (index_beta1)
      deallocate (index_beta2)
      deallocate (index_open1)
      deallocate (combinations1)
      deallocate (complement1)

    end do! multp
    ! we have summed over all initial and all final spin configurations
    ! Now, divide by number of initial spin configurations
    overlap2(:) = overlap2(:)/N
    deallocate (O)
  end subroutine occ_overlap2_highspin_core

  subroutine get_complement_combinations(n, k, combinations, complement)
    !! given a list of combinations k out of n this subroutine returns
    !! the complement combinations
    integer, intent(in) ::n
    integer, intent(in) ::k
    integer, intent(in):: combinations(:, :)
    integer, intent(out), allocatable :: complement(:, :)
    integer :: num
    integer :: i, j, c1, c2
    num = size(combinations, 2)
    allocate (complement(n - k, num))
    do i = 1, num
      c1 = 1
      c2 = 1
      do j = 1, n
        if (c2 .gt. n - k) then
          cycle
        end if
        if (c1 .gt. k) then
          complement(c2, i) = j
          c2 = c2 + 1
          cycle
        end if
        if (combinations(c1, i) .ne. j) then
          complement(c2, i) = j
          c2 = c2 + 1
        else
          c1 = c1 + 1
        end if
      end do
    end do
  end subroutine get_complement_combinations

  subroutine get_combinations(n, k, combinations)
    !! returns the list of combinations k out of 1..n
    !! in lexicographical order    
    integer ::n
    integer ::k
    integer ::j
    integer, allocatable :: combinations(:, :)
    integer :: num
    num = nint(binom(n, k))
    allocate (combinations(k, num))
    combinations(:, 1) = (/(j, j=1, k)/)

    j = 1
    call iterate_combinations(n, k, combinations, j)
  contains
    recursive subroutine iterate_combinations(end, k, comb, ind)
      integer, intent(in) :: end
      integer, intent(inout) :: comb(:, :)
      integer, intent(inout) :: ind
      integer, intent(in) :: k
      integer :: i, j

      do i = k, 1, -1
        if (comb(i, ind) .lt. end - (k - i)) then
          if (ind + 1 .gt. size(comb, 2)) stop "index to large"
          comb(1:i - 1, ind + 1) = comb(1:i - 1, ind) ! all before i are copied
          comb(i, ind + 1) = comb(i, ind) + 1 ! position i is increased by one
          !write(6,*) 'combination before',comb(:,ind+1)
          !write(6,*) 'limits',i,comb(i,ind+1)+1, k+comb(i,ind+1)
          !write(6,*) 'i+1',i+1, size(comb,1), comb(i,ind+1)+1, comb(i,ind+1)+1+size(comb,1)-i
          !if(i+1 .le. size(comb,1))&
          !     comb(i+1:,ind+1) =(/ (j,j=comb(i,ind+1)+1,k+comb(i,ind+1) ) /) ! right from i counted upwards
          if (i + 1 .le. size(comb, 1)) &
            comb(i + 1:, ind + 1) = (/(j, j=comb(i, ind + 1) + 1, comb(i, ind + 1) + size(comb, 1) - i)/) ! right from i counted upwards
          !write(6,*) comb(:,ind+1)
          ind = ind + 1
          call iterate_combinations(end, k, comb, ind)
        end if
      end do
    end subroutine iterate_combinations
  end subroutine get_combinations


  subroutine single_excitation_occs(N_occ, occs, lower_limit, upper_limit)
    !! returns a list of occupations, constructed from N_occ
    !! by exciting one electron
    integer, intent(in) :: N_occ(:)
    integer, allocatable, intent(out) :: occs(:, :)
    integer, optional :: lower_limit, upper_limit
    integer :: Noccs, Nel, Norb, Ndouble, Nsingle, Nvacant
    integer :: i, j, k
    integer :: start, end

    if (present(lower_limit)) then
      start = lower_limit
    else
      start = 1
    end if
    if (present(upper_limit)) then
      end = upper_limit
    else
      end = size(N_occ)
    end if

    Nel = sum(N_occ(start:end))
    Norb = size(N_occ, 1)
    Ndouble = count(N_occ(start:end) .eq. 2)
    Nsingle = count(N_occ(start:end) .eq. 1)
    Nvacant = count(N_occ(start:end) .eq. 0)

    Noccs = Ndouble*(Nvacant + Nsingle) + Nsingle*(Nvacant + Nsingle - 1) + 1
    allocate (occs(Norb, Noccs))
    k = 1
    occs(:, k) = N_occ
    do i = start, end
      if (N_occ(i) .gt. 0) then
        do j = start, end
          if (j .eq. i) cycle
          if (N_occ(j) .lt. 2) then
            k = k + 1
            occs(:, k) = N_occ
            occs(i, k) = occs(i, k) - 1
            occs(j, k) = occs(j, k) + 1
          end if
        end do
      end if
    end do
  end subroutine single_excitation_occs

  subroutine single_excitation_occs2(in_occs, occs, lower_limit, upper_limit)
    !! returns a list of occupations, constructed from N_occ
    !! by exciting one electron    
    integer, intent(in) :: in_occs(:, :)
    integer, allocatable, intent(out) :: occs(:, :)
    integer, optional :: lower_limit, upper_limit
    integer :: Noccs, Nel, Norb, Ndouble, Nsingle, Nvacant
    integer :: i, j, k, l, o
    integer :: start, end
    logical :: redundant_occ
    integer, allocatable :: occs_t(:, :)
    if (present(lower_limit)) then
      start = lower_limit
    else
      start = 1
    end if
    if (present(upper_limit)) then
      end = upper_limit
    else
      end = size(in_occs, 1)
    end if
    if (start .eq. -1) start = 1
    if (end .eq. -1) end = size(in_occs, 1)
    Nel = sum(in_occs(start:end, 1))
    Norb = size(in_occs, 1)
    Noccs = 0
    do o = 1, size(in_occs, 2)
      Ndouble = count(in_occs(start:end, o) .eq. 2)
      Nsingle = count(in_occs(start:end, o) .eq. 1)
      Nvacant = count(in_occs(start:end, o) .eq. 0)
      Noccs = Noccs + Ndouble*(Nvacant + Nsingle) + Nsingle*(Nvacant + Nsingle - 1) + 1
    end do
    allocate (occs_t(Norb, Noccs))
    k = 0
    do o = 1, size(in_occs, 2)
      k = k + 1
      occs_t(:, k) = in_occs(:, o)
    end do
    do o = 1, size(in_occs, 2)
      do i = start, end
        if (in_occs(i, o) .gt. 0) then
          do j = start, end
            if (j .eq. i) cycle
            if (in_occs(j, o) .lt. 2) then
              k = k + 1
              occs_t(:, k) = in_occs(:, o)
              occs_t(i, k) = occs_t(i, k) - 1
              occs_t(j, k) = occs_t(j, k) + 1
              redundant_occ = .FALSE.
              do l = 1, k - 1
                if (all(occs_t(:, k) .eq. occs_t(:, l))) redundant_occ = .TRUE.
                if (redundant_occ) exit
              end do
              if (redundant_occ) then
                k = k - 1
              end if
            end if
          end do! j
        end if
      end do! i
    end do! o
    allocate (occs(Norb, k))
    occs(:, :) = occs_t(:, 1:k)
    deallocate (occs_t)
  end subroutine single_excitation_occs2

  subroutine auto_active_space(occ, LC1, LC2, S12, active_space, Ncore)
    !! from an occupation occ and a new, LC1, and old, LC2, orbital set
    !! plus given basis set overlap matrix S12 return the orbital indices that
    !! are involved in the changes/rotations of orbitals
    !! Ncore is the number of core orbitals
    integer, intent(in) :: occ(:)
    real(long), intent(in) :: LC1(:, :)
    real(long), intent(in) :: LC2(:, :)
    real(long), intent(in) :: S12(:, :)
    integer, intent(out), allocatable :: active_space(:)
    integer, intent(in), optional :: Ncore

    integer :: s
    integer :: i, j, n
    real(long), allocatable :: O(:, :)
    real(long), parameter :: olap_criterion = 0.02
    s = 1
    if (present(Ncore)) s = Ncore

    allocate (O(size(occ, 1), size(occ, 1)))
    ! calculate all MO overlaps
    do i = 1, size(occ, 1)
      do j = 1, size(occ, 1)
        O(i, j) = dot_product(LC1(:, i), MATMUL(S12, LC2(:, j)))
      end do
    end do
    ! include all valence orbitals that have overlap to occupied orbitals in the new or old orbital set
    ! plus include those indices that are occupied
    n = 0
    do i = s + 1, size(occ, 1)
      if (sum(O(i, s + 1:)**2, mask=occ(s + 1:) .gt. 0) .gt. olap_criterion) then
        n = n + 1
      else if (sum(O(s + 1:, i)**2, mask=occ(s + 1:) .gt. 0) .gt. olap_criterion) then
        n = n + 1
      else if (occ(i) .gt. 0) then
        n = n + 1
      end if
    end do
    allocate (active_space(n))
    n = 0
    do i = s + 1, size(occ, 1)
      if (sum(O(i, s + 1:)**2, mask=occ(s + 1:) .gt. 0) .gt. olap_criterion) then
        n = n + 1
        active_space(n) = i
      else if (sum(O(s + 1:, i)**2, mask=occ(s + 1:) .gt. 0) .gt. olap_criterion) then
        n = n + 1
        active_space(n) = i
      else if (occ(i) .gt. 0) then
        n = n + 1
        active_space(n) = i
      end if
    end do
    deallocate (O)

  end subroutine auto_active_space

  subroutine general_active_space_occs(Norbtot, Nelectot, gass, occs)
    !! returns a list of occupations in occs that
    !! distributes Nelectot electrons in Norbtot orbitals
    !! where the occupations comply with the
    !! general active space definitions provides by the list of gass:
    !! gass(i)%orbitals(:) is a list defining an orbital subset
    !! gass(i)%nmin the minimum number of electrons in this orbital subset
    !! gass(i)%nmax the maximum number of electrons in this orbital subset
    integer, intent(in) :: Norbtot
    integer, intent(in) :: Nelectot
    type(GAS), intent(in) :: gass(:)
    integer, allocatable, intent(out) :: occs(:, :)

    integer :: Noccs, Nel, Norb, Ndouble, Noccs_set
    integer, allocatable :: Nelec_as(:, :)
    integer ::Noccs_as(size(gass))
    integer :: i, j, k, l, ll, li, s, ls
    integer, allocatable :: closed_combinations(:, :)
    integer, allocatable :: not_closed_combinations(:, :)
    integer, allocatable :: open_combinations(:, :)
    integer :: nset, ngas
    integer, parameter :: MAX_COMBINATIONS = 100
    ! count the combinations of distributing the total number of electrons onto the orbital subsets
    ! these are 'sets' and Nelec_as(i,j) gives for set j the number of electrons in active space i

    ! make a few checks
    do j = 1, size(gass, 1)
      write (6, '("# number of electrons from",i4,x,"to",i4,x,"in orbitals")') gass(j)%nmin, gass(j)%nmax
      if (gass(j)%nmax .gt. 2*size(gass(j)%orbitals(:), 1)) then
        stop 'check your gas definition'
      end if
      write (6, '("# ")', advance='no')
      do i = 1, size(gass(j)%orbitals(:), 1)
        write (6, '(i4)', advance='no') gass(j)%orbitals(i)
        do k = j+1, size(gass, 1)
          if (any(gass(k)%orbitals(:)==gass(j)%orbitals(i)))&
            stop 'check your gas definition: this orbital is in more than one occupation space'
        end do
      end do
      write (6, '()')
    end do

    if (Nelectot .lt. sum(gass(:)%nmin) .or. Nelectot .gt. sum(gass(:)%nmax)) then
      stop 'check your gas definition: total number of electrons does not match'
    end if

    ! we assume an upper total number of combinations MAX_COMBINATIONS
    allocate (Nelec_as(size(gass), MAX_COMBINATIONS))
    nset = 1
    ngas = 1
    Nel = Nelectot
    Nelec_as = 0

    nset = el_per_gas(Nelec_as, gass, ngas, nset, Nel)

    ! count the number of occupations
    Noccs = 0
    do j = 1, nset
      Noccs_set = 1
      do i = 1, size(gass)
        Noccs_as(i) = 0
        Nel = Nelec_as(i, j)
        Norb = size(gass(i)%orbitals, 1)
        do Ndouble = max(Nel - Norb, 0), floor(HALF*Nel), 1
          Noccs_as(i) = Noccs_as(i) + nint(max(binom(Norb, Ndouble)*binom(Norb - Ndouble, Nel - 2*Ndouble), ONE))
        end do
        Noccs_set = Noccs_set*Noccs_as(i) ! each active space combination are multiplied
      end do! as
      Noccs = Noccs + Noccs_set
    end do
    allocate (occs(Norbtot, Noccs))
    occs(:, :) = 0
    l = 1
    do s = 1, nset
      ls = l
      do i = 1, size(gass)
        li = l - 1
        if (li .lt. ls) li = ls
        Nel = Nelec_as(i, s)
        Norb = size(gass(i)%orbitals, 1)
        l = ls
        do Ndouble = floor(HALF*Nel), max(Nel - Norb, 0), -1
          call get_combinations(Norb, Ndouble, closed_combinations)
          call get_complement_combinations(Norb, Ndouble, closed_combinations, not_closed_combinations)
          call get_combinations(Norb - Ndouble, Nel - 2*Ndouble, open_combinations)
          do j = 1, size(closed_combinations, 2)
            do k = 1, size(open_combinations, 2)
              do ll = ls, li ! we loop through the occs and copy the previous active spaces
                occs(:, l) = occs(:, ll)
                occs(gass(i)%orbitals(:), l) = 0
                occs(gass(i)%orbitals(closed_combinations(:, j)), l) = 2
                occs(gass(i)%orbitals(not_closed_combinations(open_combinations(:, k), j)), l) = 1
                l = l + 1
              end do ! ll
            end do! k
          end do! j
          deallocate (closed_combinations)
          deallocate (not_closed_combinations)
          deallocate (open_combinations)
        end do! Ndouble
      end do! i
    end do! s
    deallocate (Nelec_as)

  contains
    ! distritbute electrons on general active spaces
    recursive function el_per_gas(Nelec_per_as, gass, gas_it, set_it, Nel) result(j)
      type(GAS), intent(in) :: gass(:)
      integer, intent(inout) :: Nelec_per_as(:, :)
      integer, intent(in) :: gas_it, Nel
      integer, intent(inout) :: set_it
      integer :: i, j
      j = 0
      if (set_it .gt. size(Nelec_per_as, 2)) stop "Too many combinations for ras space"
      if (Nel - sum(Nelec_per_as(1:gas_it - 1, set_it)) .ge. 0 .and. gas_it .le. size(gass)) then
        do i = gass(gas_it)%nmin, gass(gas_it)%nmax
          Nelec_per_as(gas_it, set_it) = i
          j = el_per_gas(Nelec_per_as, gass, gas_it + 1, set_it, Nel)
        end do
      else if (Nel - sum(Nelec_per_as(1:gas_it - 1, set_it)) .eq. 0 .and. gas_it .gt. size(gass)) then
        j = set_it
        set_it = set_it + 1
        Nelec_per_as(:, set_it) = Nelec_per_as(:, set_it - 1)
      else
        j = set_it - 1
      end if
      return
    end function el_per_gas
  end subroutine general_active_space_occs

  subroutine filter_sym_occs(irrep_prod, irrep_MO, irrep_state, occs_in, occs_out)
    !! filters set of occupation numbers to meet symmetry restrictions
    integer, intent(in) :: irrep_prod(:, :)
    integer, intent(in) :: irrep_MO(:)
    integer, intent(in) :: irrep_state
    integer, intent(in) :: occs_in(:, :)
    integer, intent(out), allocatable :: occs_out(:, :)
    integer :: occs_irrep(size(occs_in, 2))
    integer :: i, k, irrep
    do k = 1, size(occs_in, 2)
      irrep = 1
      do i = 1, size(occs_in, 1)
        if (occs_in(i, k) .eq. 1) then
          irrep = irrep_prod(irrep, irrep_MO(i))
        end if
      end do
      occs_irrep(k) = irrep
    end do

    allocate (occs_out(size(occs_in, 1), count(occs_irrep(:) .eq. irrep_state)))
    occs_out(:, :) = occs_in(:, pack((/(i, i=1, size(occs_irrep, 1))/), (occs_irrep(:) .eq. irrep_state)))

  end subroutine filter_sym_occs

  subroutine filter_ras_occs(ras, occs_in, occs_out)
    !! filters set of occupation numbers to meet ras restrictions
    type(GAS), intent(in) :: ras(:)
    integer, allocatable, intent(in) :: occs_in(:, :)
    integer, allocatable, intent(out) :: occs_out(:, :)
    integer :: i, j, k
    logical :: accepted
    j = 1
    do i = 1, size(occs_in, 2)
      accepted = .True.
      do k = 1, size(ras)
        if (sum(occs_in(ras(k)%orbitals(:), i)) .lt. ras(k)%nmin .or. &
          sum(occs_in(ras(k)%orbitals(:), i)) .gt. ras(k)%nmax) then
          accepted = .False.
        end if
      end do
      if (accepted) &
        j = j + 1
    end do
    j = j - 1
    allocate (occs_out(size(occs_in, 1), j))
    j = 1
    do i = 1, size(occs_in, 2)
      accepted = .True.
      do k = 1, size(ras)
        if (sum(occs_in(ras(k)%orbitals(:), i)) .lt. ras(k)%nmin .or. &
          sum(occs_in(ras(k)%orbitals(:), i)) .gt. ras(k)%nmax) then
          accepted = .False.
        end if
      end do
      if (accepted) then

        occs_out(:, j) = occs_in(:, i)
        j = j + 1
      end if
    end do
  end subroutine filter_ras_occs

  subroutine active_space_occs(N_occ, active_space_MO, occs)
    !! returns a list of occupations, constructed from N_occ
    !! by redistributing electrons wihtin the orbitals given in active_space_MO
    integer, intent(in) :: active_space_MO(:)
    integer, intent(in) :: N_occ(:)
    integer, allocatable, intent(out) :: occs(:, :)
    integer :: Noccs, Nel, Norb, Ndouble
    integer :: j, k, l
    integer, allocatable :: closed_combinations(:, :)
    integer, allocatable :: not_closed_combinations(:, :)
    integer, allocatable :: open_combinations(:, :)
    Noccs = 0
    Nel = sum(N_occ(active_space_MO))
    Norb = size(active_space_MO, 1)
    do Ndouble = max(Nel - Norb, 0), floor(HALF*Nel), 1
      Noccs = Noccs + nint(max(binom(Norb, Ndouble)*binom(Norb - Ndouble, Nel - 2*Ndouble), ONE))
    end do
    allocate (occs(size(N_occ, 1), Noccs))
    l = 1
    do Ndouble = floor(HALF*Nel), max(Nel - Norb, 0), -1
      call get_combinations(Norb, Ndouble, closed_combinations)
      call get_complement_combinations(Norb, Ndouble, closed_combinations, not_closed_combinations)
      call get_combinations(Norb - Ndouble, Nel - 2*Ndouble, open_combinations)
      do j = 1, size(closed_combinations, 2)
        do k = 1, size(open_combinations, 2)
          occs(:, l) = N_occ(:)
          occs(active_space_MO(:), l) = 0
          occs(active_space_MO(closed_combinations(:, j)), l) = 2
          occs(active_space_MO(not_closed_combinations(open_combinations(:, k), j)), l) = 1
          l = l + 1
        end do
      end do
      deallocate (closed_combinations)
      deallocate (not_closed_combinations)
      deallocate (open_combinations)
    end do
  end subroutine active_space_occs

  function initial_density(P, G, A, B) result(rho)
    !! prepares an initial electron density F%rho
    !! based on atomic electron densities
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    real(long) :: rho(G%N_grid)
    integer :: max_N_nao
    integer ::  k, i, n, l, m, j
    max_N_nao = maxval(A(:)%N_nao)
    rho(:) = ZERO
    k = 1
    do i = 1, P%N_nuc     ! 1 ... # of nuclear
      do l = 0, 10    ! angular quantum number
        if (P%N_nao_table(l, i) .eq. ZERO) cycle
        do n = 1, P%N_nao_table(l, i)
          do m = -l, l
            do j = 1, A(i)%N_orb
              if (A(i)%config(j, 1) .eq. n + l .and. A(i)%config(j, 2) .eq. l) then
                rho(:) = rho(:) + A(i)%config(j, 3)/dble(2*l + 1)*B%func(:, k)**2
              end if
            end do
            k = k + 1
          end do
        end do
      end do
    end do
  end function initial_density

  function initial_density_matrix(P, A, B) result(D)
    !! this is the equivalent to initial_density
    !! but returns a density matrix instead of a grid density
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    real(long) :: D(B%N_basis, B%N_basis)
    integer :: max_N_nao
    integer ::  k, i, n, l, m, j

    max_N_nao = maxval(A(:)%N_nao)
    D = ZERO
    k = 1
    do i = 1, size(A)     ! 1 ... # of nuclear
      do l = 0, 10    ! angular quantum number
        if (P%N_nao_table(l, i) .eq. ZERO) cycle
        do n = 1, P%N_nao_table(l, i)
          do m = -l, l
            do j = 1, A(i)%N_orb
              if (A(i)%config(j, 1) .eq. n + l .and. A(i)%config(j, 2) .eq. l) then
                D(k, k) = D(k, k) + A(i)%config(j, 3)/dble(2*l + 1)
              end if
            end do
            k = k + 1
          end do
        end do
      end do
    end do
  end function initial_density_matrix


  function density(G, B, Dmatrix) result(rho)
    !! computes electronic density on the grid
    !! density: $$\rho(x) = \sum_{i=1}^{N_\text{MO}} n_i | \psi_i(x) |^2$$
    !!          $$\psi_i(x) = \sum_{j=1}^{N_\text{basis}} C_{ij} \phi_j(x) $$
    !! where \( phi(x) \) is a basis function
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long) :: Dmatrix(B%N_basis, B%N_basis)
    real(long) :: rho(G%N_grid)
    !real(long) :: psi(G%N_grid,B%N_basis)
    real(long) :: factor
    integer :: i, j, k_nuc, ks, ke
    integer(8) :: ij
    !
    ! sixth implementation: using B%yes_overlap
    !

    !
    ! bring back the 4th implementation: using B%yes_overlap and density matrix
    !
    rho(:) = ZERO
    do i = 1, B%N_basis
      do j = i, B%N_basis
        ij = index_packed(i, j)
        if (B%yes_overlap(0, ij) .eqv. .false.) cycle
        factor = TWO
        if (i .eq. j) factor = ONE
        do k_nuc = 1, G%N_nuc
          if (B%yes_overlap(k_nuc, ij) .eqv. .false.) cycle
          ks = G%i_start(k_nuc)
          ke = G%i_end(k_nuc)
          rho(ks:ke) = rho(ks:ke) + Dmatrix(i, j) * B%func(ks:ke, i) * &
            B%func(ks:ke, j)*factor
        end do
      end do
    end do
    !
    ! the implementation below takes about 5 times slower than the above
    !
    !do k_nuc = 1, G%N_nuc
    !  ks = G%i_start(k_nuc)
    !  ke = G%i_end(k_nuc)
    !  do i = 1, B%N_basis
    !    do j = i, B%N_basis
    !      ij = index_packed( i, j )
    !      if ( B%yes_overlap(k_nuc,ij) ) then
    !        if ( i == j ) then
    !          rho(ks:ke) = rho(ks:ke) + F%D(i,i) * B%func(ks:ke,i)**2
    !        else
    !          rho(ks:ke) = rho(ks:ke) + TWO * F%D(i,j) * B%func(ks:ke,i) * &
    !                                                     B%func(ks:ke,j)
    !        end if
    !      end if
    !    end do
    !  end do
    !end do
    !
    ! check negative density
    !
    !  if ( any( rho(:) < ZERO ) ) stop "negative density found!!!"
    if (any(rho(:) .lt. ZERO)) then
      ! write(6 --> write(STDERR : sangkil on Jan. 5, 2017
      write (STDERR, '("# !!! negative density found !!!",es15.5)') minval(rho)
      WHERE (rho(:) .lt. ZERO) rho(:) = ZERO
    end if
  end function density

  function spin_density(G, B, Dmatrix) result(rho)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    !  type(WF_data), intent(in) :: F
    real(long) :: Dmatrix(B%N_basis, B%N_basis)
    real(long) :: rho(G%N_grid)
    real(long) :: factor
    integer :: i, j, k_nuc, ks, ke
    integer(8) :: ij
    rho(:) = ZERO
    do i = 1, B%N_basis
      do j = i, B%N_basis
        ij = index_packed(i, j)
        if (B%yes_overlap(0, ij) .eqv. .false.) cycle
        factor = TWO
        if (i .eq. j) factor = ONE
        do k_nuc = 1, G%N_nuc
          if (B%yes_overlap(k_nuc, ij) .eqv. .false.) cycle
          ks = G%i_start(k_nuc)
          ke = G%i_end(k_nuc)
          rho(ks:ke) = rho(ks:ke) + Dmatrix(i, j) * B%func(ks:ke, i) * &
            B%func(ks:ke, j)*factor
        end do
      end do
    end do

  end function spin_density

  subroutine density_matrix(LC, occ, D)
    !! computes the density matrix
    real(long), intent(in) :: LC(:, :)
    real(long), intent(in) :: occ(:)
    real(long), intent(out) :: D(size(LC, 1), size(LC, 1))
    integer :: i, j, k, N
    N = size(LC, 1)
    D = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, size(LC, 2)
      !     if(F%n_occ(k)==0 .and. .not. P%yes_thermal .and. .not. allocated(P%av_occ) )cycle   ! skip unoccupied orbitals
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          D(i, j) = D(i, j) + LC(i, k)*LC(j, k)*occ(k)
          if (i .ne. j) D(j, i) = D(i, j)
        end do
      end do
    end do
  end subroutine density_matrix

  subroutine density_matrix_unrestricted(F)
    type(WF_data), intent(inout) :: F
    integer :: i, j, k, N

    N = F%N_basis
    F%D_a = ZERO
    F%D_b = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, N
      !     if(F%n_occ(k)==0 .and. .not. P%yes_thermal .and. .not. allocated(P%av_occ) )cycle   ! skip unoccupied orbitals
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          F%D_a(i, j) = F%D_a(i, j) + F%LC_a(i, k)*F%LC_a(j, k)*F%f_occ_a(k)
          if (i .ne. j) F%D_a(j, i) = F%D_a(i, j)
          F%D_b(i, j) = F%D_b(i, j) + F%LC_b(i, k)*F%LC_b(j, k)*F%f_occ_b(k)
          if (i .ne. j) F%D_b(j, i) = F%D_b(i, j)
        end do
      end do
    end do

    F%D = F%D_a + F%D_b

  end subroutine density_matrix_unrestricted


  function prodOverlap(P, F, Strans, LCref) result(prodOlap)
    !!  Calculate the product of all occupied orbital projections      
    ! to be done: rotate orbitals within average_occ_space such that they best overlap
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    real(long), dimension(F%N_basis, F%N_basis), intent(in) :: Strans   !! The overlap matrix SI(i,j) between the
    real(long), dimension(F%N_basis, F%Nr_basis), intent(in) :: LCref !! The reference MO coefficients
    real(long), dimension(F%Nr_basis, F%Nr_basis) :: O  !! the MO-MO overlap matrix
    real(long), allocatable :: Proj(:, :)       
    real(long) :: prodOlap
    integer :: N ! number of bf
    integer :: i, j ! counters
    integer :: occ
    if (allocated(F%av_occ)) then
      allocate (Proj(F%Nr_basis, 3 + size(F%av_occ)))
    else
      allocate (Proj(F%Nr_basis, 3))
    end if
    N = F%Nr_basis
    ! compute the overlap matrix to the reference MO
    do i = 1, N
      do j = 1, N
        O(i, j) = dot_product(F%LC(:, i), MATMUL(Strans, LCref(:, j)))
      end do
    end do
    ! compute the Projections
    do i = 1, N
      Proj(i, :) = ZERO
      do j = 1, N
        occ = occ_space(F, j)
        ! projection into the orbital subspace of respective occupation
        !
        if (P%MOMtype .eq. 1) then
          ! original version
          Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)
        else if (P%MOMtype .eq. 2) then
          ! GAMESS version
          Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)
        else if (P%MOMtype .eq. 3) then
          ! Q-chem version
          Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)*O(i, j)
        end if
      end do
    end do
    if (P%MOMtype .eq. 1) then
      ! original version
      Proj(:, :) = abs(Proj(:, :))
    end if
    !prodOlap = ONE
    prodOlap = ZERO
    do i = 1, N
      occ = occ_space(F, i)
      prodOlap = prodOlap + F%f_occ(i)*log(Proj(i, occ + 1))
    end do
    prodOlap = prodOlap
    prodOlap = exp(prodOlap)
  end function prodOverlap

  function prodOverlapUnrestricted(P, F, Strans, LCrefa, LCrefb) result(prodOlap)
    !!  Calculate the product of all occupied orbital projections      
    ! to be done: rotate orbitals within average_occ_space such that they best overlap
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), dimension(F%N_basis, F%N_basis), intent(in) :: Strans   !! The overlap matrix SI(i,j) between the
    real(long), dimension(F%N_basis, F%Nr_basis), intent(in) :: LCrefa !! The reference MO coefficients
    real(long), dimension(F%N_basis, F%Nr_basis), intent(in) :: LCrefb !! The reference MO coefficients
    real(long) :: prodOlap
    F%LC = F%LC_b
    prodOlap = prodOverlap(P, F, Strans, LCrefb)
    F%LC = F%LC_a
    prodOlap = sqrt(prodOlap * prodOverlap(P, F, Strans, LCrefa))
  end function prodOverlapUnrestricted


  pure function occ_space(F, mo) result(n)
    !! returns
    !! 0 for unoccupied MO
    !! 1 for singly occupied MO
    !! 2 for doubly occupied MO
    !! 2+i for orbital in average_occ_MO subspace i=1,...n if the orbitals are partially occupied  
    type(WF_data), intent(in) :: F
    integer, intent(in) :: mo
    integer :: i, n
    n = F%N_occ(mo)
    if (allocated(F%av_occ)) then
      do i = 1, size(F%av_occ)
        if (ANY(F%av_occ(i)%MO(:) .eq. mo) .and. &
          sum(F%f_occ(F%av_occ(i)%MO(:))) .gt. 0 .and. &
          sum(F%f_occ(F%av_occ(i)%MO(:))) .lt. 2*size(F%av_occ(i)%MO(:))) then
          n = 2 + i
          exit
        end if
      end do
    end if
  end function occ_space

  function WFOverlap(F, Strans, LCref) result(overlap)
    !! this function returns the overlap of two many particle wave functions
    !! assuming that we have high spin single determinant, respectively.
    !! Thus, the overlap <psi_1|psi_2> = det( <i|j> ) * det ( <k|l> )
    !! where i and j are alpha and k and l beta spin-orbitals
    !! i and k are orbitals from Slaterdet psi_1, j and l orbitals form Slaterdet psi_2
    ! to be done: rotate orbitals within average_occ_space such that they best overlap
    type(WF_data), intent(in) :: F

    real(long), dimension(F%N_basis, F%N_basis), intent(in) :: Strans   !! The overlap matrix SI(i,j) between the
    real(long), dimension(F%N_basis, F%N_basis), intent(in) :: LCref !! The reference MO coefficients
    ! real(long) :: overlap
    ! integer,allocatable :: occs(:,:)
    ! real(long),allocatable :: overlap2(:)

    ! if(allocated(F%av_occ))then
    !   write(6,*) '#averaged occs', size(occs,2)
    !   call averaged_occs(P,F,F%N_occ,occs)
    !   write(6,*) '#size average', size(occs,2)
    ! else
    !   allocate(occs(size(F%N_occ),1))
    !   occs(:,1)=F%N_occ
    ! end if
    ! allocate(overlap2(size(occs,2)))
    ! overlap2=occ_overlap2(F%N_occ,occs,F%LC,LCref,Strans)

    ! overlap=sqrt(sum(overlap2))
    ! deallocate(overlap2)
    ! deallocate(occs)

    real(long), dimension(F%N_basis, F%N_basis) :: LCrot ! The reference MO coefficients
    real(long), dimension(:, :), allocatable :: Oalpha! one particle overlap matrix for the double occupied space
    real(long), dimension(:, :), allocatable :: Obeta! one particle overlap matrix for the double+single occupied space

    real(long) :: overlap ! result
    integer :: Ndocc, Nsocc, Nalpha, Nbeta ! Number of single / double occupied MOs
    integer :: N ! number of bf
    integer :: i, j, ialpha, jalpha, ibeta, jbeta ! counters

    overlap = ZERO
    N = F%N_basis
    Ndocc = 0
    Nsocc = 0

    LCrot = F%LC

    if (allocated(F%av_occ)) &
      call rotate_av_orbitals(F%av_occ, LCref, Strans, LCrot)
    do i = 1, size(F%n_occ)
      if (F%N_occ(i) .eq. 2) Ndocc = Ndocc + 1
      if (F%N_occ(i) .eq. 1) Nsocc = Nsocc + 1
    end do
    Nalpha = Ndocc
    Nbeta = Ndocc + Nsocc
    ! we assume that we have Ndocc alpha and Ndocc+Nsocc beta orbitals (high-spin beta state)
    allocate (Oalpha(Nalpha, Nalpha), Obeta(Nbeta, Nbeta))

    ! compute the overlap matrices to the reference MO for the alpha and beta orbitals separately
    ! Note: Left index is actual orbital, right index is reference orbital
    ialpha = 1
    ibeta = 1
    do i = 1, size(F%N_occ)
      jalpha = 1
      jbeta = 1
      do j = 1, size(F%N_occ)
        if (F%N_occ(i) .ge. 1 .and. F%N_occ(j) .ge. 1) then
          overlap = dot_product(LCrot(:, i), MATMUL(Strans, LCref(:, j)))
          Obeta(ibeta, jbeta) = overlap
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 2) then
            Oalpha(ialpha, jalpha) = overlap
          end if
        end if
        if (F%N_occ(j) .eq. 2) then
          jalpha = jalpha + 1
          jbeta = jbeta + 1
        else if (F%N_occ(j) .ge. 1) then
          jbeta = jbeta + 1
        end if
      end do
      if (F%N_occ(i) .eq. 2) then
        ialpha = ialpha + 1
        ibeta = ibeta + 1
      else if (F%N_occ(i) .ge. 1) then
        ibeta = ibeta + 1
      end if
    end do

    overlap = matrix_determinant(Oalpha)*matrix_determinant(Obeta)

    deallocate (Obeta)
    deallocate (Oalpha)

  end function WFOverlap

  subroutine rotate_av_orbitals(av_occ, Cref, SI, Crot)
    !! rotates the orbitals F%LC inside the av_occ(:)%MO subspace
    !! such that they best overlap with the orbitals in the av_occ(:)%MO subspace in Cref    
    type(average_occ), intent(in) :: av_occ(:)
    real(long), dimension(:, :), intent(in) :: Cref
    real(long), dimension(:, :), intent(in) :: SI
    real(long), dimension(:, :), intent(inout) :: Crot

    real(long), allocatable :: Ctmp(:, :)
    real(long), allocatable :: O(:, :) ! overlap to reference

    real(long), allocatable :: O_av(:, :) ! av_occ(:)%MO subspace submatrix of O
    real(long), allocatable :: O_av_inv(:, :) ! inverse of O_av

    real(long) :: O_av_det
    integer :: i, j, l
    type(LU_matrix) ::  LU
    integer :: N

    N = size(Cref, 1)
    allocate (Ctmp(N, N))
    allocate (O(N, N))

    ! calculate the overlap matrix
    do i = 1, N
      do j = 1, N
        O(i, j) = dot_product(Crot(:, i), MATMUL(SI, Cref(:, j)))
      end do
    end do

    Ctmp = Crot
    do i = 1, size(av_occ) ! for each subspace
      allocate (O_av(size(av_occ(i)%MO), size(av_occ(i)%MO)))
      allocate (O_av_inv(size(av_occ(i)%MO), size(av_occ(i)%MO)))
      O_av = O(av_occ(i)%MO, av_occ(i)%MO) ! set the overlap submatrix
      LU = LU_decompose(O_av)
      O_av_inv = LU_inverse(LU)  ! calculate its inverse
      O_av_det = matrix_determinant(O_av)
      call LU_matrix_purge(LU)
      !    Crot(:,av_occ(i)%MO(:))=abs(O_av_det)**(ONE/size(O_av,1))*MATMUL(Ctmp(:,av_occ(i)%MO(:)), transpose(O_av_inv(:,:)) )
      ! Rotate orbitals, rotation matrix is inverse of overlap * determinant of overlap ^(1/dim), such that the determinant of the
      ! rotated overlap matrix is exactly the same
      forall (l=1:size(Crot, 1), j=1:size(av_occ(i)%MO))
        Crot(l, av_occ(i)%MO(j)) = abs(O_av_det)**(ONE/size(O_av, 1))*sum(O_av_inv(j, :)*Ctmp(l, av_occ(i)%MO(:)))
      end forall
      Ctmp = Crot
      deallocate (O_av)
      deallocate (O_av_inv)
    end do
    deallocate (Ctmp)
    deallocate (O)
  end subroutine rotate_av_orbitals

  function fidelity(LC, Strans, LCref) result(fid)
    !! computes the overlaps to reference orbitals squared
    real(long), dimension(:, :), intent(in) :: LC ! the MO coefficients
    real(long), dimension(:, :), intent(in) :: Strans   ! The overlap matrix SI(i,j) between the
    real(long), dimension(:, :), intent(in) :: LCref ! The reference MO coefficients
    real(long), dimension(size(LC,2)) :: fid
    integer i
    do i = 1, size(LC,2)
      fid(i) = dot_product(LC(:, i), MATMUL(Strans, LCref(:, i)))
      fid(i) = fid(i)*fid(i)
    end do
  end function fidelity

  function nuclear_repulsion_energy(A, P) result(E_nuc)
    !! nuclear repulsion energy
    type(Atom_set), intent(in) :: A(:)
    type(Param), intent(in) :: P
    real(long) :: E_nuc
    real(long) :: R_ij
    integer :: i, j, N_nuc
    real(long) :: R_corr
    real(long) :: sigma_i, sigma_j, alpha, gauss_corr
    N_nuc = size(A)
    E_nuc = ZERO
    do i = 1, N_nuc
      do j = i + 1, N_nuc
        ! Skip if both nuclei are partial charges (only for QM/MM calculations)
        if ((A(i)%charge .eq. 0 .and. A(j)%charge .eq. 0) .and. P%yes_qmmm) cycle

        R_ij = sqrt((A(i)%x - A(j)%x)**2 + &
          (A(i)%y - A(j)%y)**2 + &
          (A(i)%z - A(j)%z)**2)

        ! Correction for guassian nuclear distribution model
        if (A(i)%sigma .gt. ZERO .or. A(j)%sigma .gt. ZERO) then
          sigma_i = max(0.0, A(i)%sigma)
          sigma_j = max(0.0, A(j)%sigma)
          alpha = 1 / (2*(sigma_i**2 + sigma_j**2))
          gauss_corr = erf(R_ij*sqrt(alpha))
          R_corr = R_ij ! no softcore correction for gaussian charges
        else
          gauss_corr = ONE
          R_corr = sqrt(R_ij**2 + P%R_softcore**2)
        end if

        E_nuc = E_nuc + (dble(A(i)%fcharge*A(j)%fcharge)/R_corr)*gauss_corr

      end do
    end do
  end function nuclear_repulsion_energy

  function nuclear_repulsion_gradient(A, P) result(grad_E)
    !! nuclear repulsion gradient
    type(Atom_set), intent(in) :: A(:)
    type(Param), intent(in) :: P
    real(long) :: grad_E(size(A, 1)*3)
    real(long) :: R_ij
    integer :: i, j, N_nuc
    real(long) :: R_corr
    real(long) :: sigma_i, sigma_j, alpha, gauss_corr

    N_nuc = size(A)
    grad_E = zero
    do i = 1, N_nuc
      do j = i + 1, N_nuc
        ! Skip if both atoms are partial charges and calc_enuc_2fcharge is false (for QM/MM)
        if ((A(i)%charge .eq. 0 .and. A(j)%charge .eq. 0) .and. P%yes_qmmm) cycle

        R_ij = sqrt((A(i)%x - A(j)%x)**2 + &
          (A(i)%y - A(j)%y)**2 + &
          (A(i)%z - A(j)%z)**2)

        ! Correction for guassian nuclear distribution model
        if (A(i)%sigma .gt. ZERO .or. A(j)%sigma .gt. ZERO) then
          sigma_i = max(0.0, A(i)%sigma)
          sigma_j = max(0.0, A(j)%sigma)
          alpha = 1 / (TWO * (sigma_i**2 + sigma_j**2))
          gauss_corr = erf(sqrt(alpha)*R_ij) - TWO * R_ij * dsqrt(alpha/PI) * dexp(-alpha*R_ij**2)
          !gauss_corr = R_corr**2 * TWO * dsqrt(alpha/PI) * dexp(-alpha*R_ij**2)

          R_corr = R_ij ! no softcore correction for gaussian charges
        else
          gauss_corr = ONE
          R_corr = sqrt(R_ij**2 + P%R_softcore**2)
        end if

        grad_E((i - 1)*3 + 1) = grad_E((i - 1)*3 + 1) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(A(i)%x - A(j)%x)/R_corr**3
        grad_E((i - 1)*3 + 2) = grad_E((i - 1)*3 + 2) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(A(i)%y - A(j)%y)/R_corr**3
        grad_E((i - 1)*3 + 3) = grad_E((i - 1)*3 + 3) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(A(i)%z - A(j)%z)/R_corr**3
        grad_E((j - 1)*3 + 1) = grad_E((j - 1)*3 + 1) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(-ONE)*(A(i)%x - A(j)%x)/R_corr**3
        grad_E((j - 1)*3 + 2) = grad_E((j - 1)*3 + 2) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(-ONE)*(A(i)%y - A(j)%y)/R_corr**3
        grad_E((j - 1)*3 + 3) = grad_E((j - 1)*3 + 3) - &
          gauss_corr*dble(A(i)%fcharge*A(j)%fcharge)*(-ONE)*(A(i)%z - A(j)%z)/R_corr**3
      end do
    end do
  end function nuclear_repulsion_gradient

  function nuclear_repulsion_energy_crystal(G, A) result(E_nuc)
    !! nuclear_repulsion_energy_crystal() for the crystalline case

    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    real(long) :: E_nuc
    integer :: i_nuc, j_nuc, k
    real(long) :: R_AB, E
    !  real(long) :: r, R_tr(3), Z_A, Z_B, R_AB, temp
    !  real(long) :: R_ij, X, Y, Z, E
    !  real(long), dimension(size(A)) :: XX, YY, ZZ
    !  integer :: i, j, ij, k
    !  real(long) :: Xij(3), ZiZj, Dij8(8)
    !  real(long) :: XijT4x8(4,8)    ! after "fuzzy" translation
    !
    ! XijI4x8(:,:)   : introduced by Kota, translated x,y,z and fuzzy weight
    !   XijI4x8(1,k) : x coordinate after k-th translation
    !   XijI4x8(2,k) : y coordinate after k-th translation
    !   XijI4x8(3,k) : z coordinate after k-th translation
    !   XijI4x8(4,k) : fuzzy weight for the k-th translation
    !   k=1...8     : it has 8 different translations for one (x,y,z)
    !
    E_nuc = ZERO
    if (G%yes_periodic .eqv. .false.) then
      call error_message("ERROR: call nuclear_repulsion_energy_crystal() "// &
        "without the periodic option")
    !  else if ( G%yes_cutoff == .false. ) then
    !    call error_message( "ERROR: call nuclear_repulsion_energy_crystal() " // &
    !                        "without the cut-off option" )
    end if
    loop_A: do i_nuc = 1, G%N_nuc
      loop_B: do j_nuc = 1, G%N_nuc       ! NOTE: it's not j_nuc = i_nuc+1, N_nuc
        loop_tr: do k = 1, G%N_tr
          !
          ! skip itself and lower-triangles if A & B are within the same copy (k=1)
          !
          if (k .eq. 1 .and. j_nuc .le. i_nuc) cycle loop_tr
          !
          ! skip the rest if a bond distance is larger than r_cutoff
          !
          R_AB = bond_distance(G, i_nuc, j_nuc, k)  ! <-- good scanning order
          if (G%yes_cutoff .and. R_AB .gt. G%r_cutoff) exit loop_tr
          !
          ! WARNING: this is an asymmetric sum
          !
          !               U_s U_s+U_c Z_A Z_B
          ! E_nuc = (1/2) SUM   SUM   -------
          !                A   B/=A    R_AB
          !
          !   A runs every atom within the supercell, A in U_s
          !   B runs every atom within the cutoff radius, B in U_c
          !
          !             U_s Z_A Z_B         U_s U_c Z_A Z_B
          ! ==> E_nuc = SUM ------- + (1/2) SUM SUM -------
          !             A<B   R_AB           A   B    R_AB
          !
          ! within the same supercell:  add Z_A Z_B / R_AB
          ! with different copies:      add (1/2) Z_A Z_B / R_AB
          !
          E = A(i_nuc)%fcharge*A(j_nuc)%fcharge/R_AB
          if (k .ge. 2) E = E*HALF        ! A and B are in different copies
          E_nuc = E_nuc + E
        end do loop_tr
      end do loop_B
    end do loop_A
  end function nuclear_repulsion_energy_crystal

  function long_range_energies(P, G, F) result(E)
    !! returns long range interaction energies
    real(long) :: E(3)    ! 1: nuc-elec, 2: elec-elec, 3: nuc-nuc
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(WF_data), intent(in) :: F
    integer :: i_nuc, j_nuc, k
    real(long) :: R_AB, n_vec(3), dipole, r_c_EPS_00
    if (G%yes_cutoff .eqv. .false.) &
      call error_message("electrostatic_energies() is called without cutoff")
    E(:) = ZERO
    r_c_EPS_00 = huge(ONE)
    loop_A: do i_nuc = 1, G%N_nuc
      loop_B: do j_nuc = 1, G%N_nuc       ! NOTE: it's not j_nuc = i_nuc+1, N_nuc
        if (P%EPS_00 .gt. ZERO &  ! screening consistently in potential_J_atom
          ) r_c_EPS_00 = F%Q_A(j_nuc)/P%EPS_00
        loop_tr: do k = 1, G%N_tr
          R_AB = bond_distance(G, i_nuc, j_nuc, k)  ! <-- good scanning order
          if (R_AB .le. G%r_cutoff) cycle loop_tr   ! skip the short-range region
          !
          ! monopole-monopole interactions
          ! NOTE: a factor of 1/2 for avoiding double-counting
          !       it is necessary because this is an asymmetric sum
          !       for A=1...N_nuc and B=1...N_nuc(+translated)
          !
          E(1) = E(1) - (G%Z_A(i_nuc)*F%Q_A(j_nuc)/R_AB + &    ! nuc-elec
            F%Q_A(i_nuc)*G%Z_A(j_nuc)/R_AB)*HALF
          E(2) = E(2) + F%Q_A(i_nuc)*F%Q_A(j_nuc)/R_AB*HALF    ! elec-elec
          E(3) = E(3) + G%Z_A(i_nuc)*G%Z_A(j_nuc)/R_AB*HALF    ! nuc-nuc
          !
          ! monopole-dipole interactions
          ! - monopole: i_nuc-th atom
          ! - dipole  : j_nuc-th atom
          ! NOTE: there is no double-counting for monopole-dipole interactions
          !   n_vec := a normalized vector from j_nuc pointing i_nuc
          !
          if (R_AB .gt. r_c_EPS_00) exit loop_tr ! skip R_AB > r_c given by EPS_00
          n_vec(:) = obtain_nuclear_position(G, i_nuc, j_nuc, k)/R_AB
          dipole = sum(n_vec(:)*F%dipole_A(:, j_nuc))/R_AB**2
          E(1) = E(1) - G%Z_A(i_nuc)*dipole
          E(2) = E(2) + F%Q_A(i_nuc)*dipole
        end do loop_tr
      end do loop_B
    end do loop_A
  end function long_range_energies

  function potential_ext(G, A) result(v)
    !! Nuc-elec potential
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    real(long) :: v(G%N_grid)
    integer :: i, i_nuc, j_nuc, k, N_tr
    real(long) :: r, x, y, z, Z_A
    real(long) :: R_tr(3), R_AB
    !
    ! V = sum_alpha ( - Z_alpha / r_alpha )
    !     where r_alpha = | (x-x_alpha)^2 + (y-y_alpha)^2 + (z-z_alpha)^2 |
    !
    ! maybe we can try forall() here.
    !
    !do i = 1, G%N_grid
    !  v(i) = sum( - dble(A(:)%charge) / sqrt( ( G%x(i) - A(:)%x )**2 + &
    !                                          ( G%y(i) - A(:)%y )**2 + &
    !                                          ( G%z(i) - A(:)%z )**2 ) )
    !end do

    do i = 1, G%N_grid
      v(i) = ZERO
      i_nuc = G%i_nuc(i)
      loop_nuc: do j_nuc = 1, size(A)
        Z_A = A(j_nuc)%fcharge
        !
        ! adding up -Z/r for given i-th grid point
        ! where r is a distance of the i-th grid point from the j_nuc-th atom
        !
        N_tr = 1          ! no translation if G%yes_periodic is false
        if (G%yes_periodic) N_tr = G%N_tr
        loop_tr: do k = 1, N_tr           ! N_tr=1 if no periodic
          !
          ! r_cutoff applied: REMEMBER it's always applied only for nuc. positions
          !
          R_AB = bond_distance(G, i_nuc, j_nuc, k)  ! <-- good scanning order
          if (G%yes_cutoff .and. R_AB .gt. G%r_cutoff) exit loop_tr ! skip the rest
          !
          ! obtain a grid position that is translated w.r.t. i_nuc and j_nuc
          !
          R_tr(:) = obtain_grid_position(G, i, j_nuc, k)
          x = R_tr(1)     ! x = G%x(i) - A(j_nuc)%x when no periodic is used
          y = R_tr(2)     ! y = G%y(i) - A(j_nuc)%y when no periodic is used
          z = R_tr(3)     ! z = G%z(i) - A(j_nuc)%z when no periodic is used
          r = sqrt(x**2 + y**2 + z**2)
          v(i) = v(i) - Z_A/r
        end do loop_tr
      end do loop_nuc
    end do
  end function potential_ext

  !>-----------------------------------------------------------------------------
  !! calculating the direct Coulomb potential
  !! a.k.a. Hartree potential or classical electron-electron repulsion potential
  !!
  !!   v_J(x) = INT{ rho(x') / |x-x'| } d^3x'
  !!   where x and x' indicate a vector in 3D.
  !!
  !! - potential_J( P, G, A(:), rho ): rho is represented on molecular grids
  !! - potential_J_atom( P, G, A, rho_A ): rho_A is represented on atomic grids
  !!
  !! Basically, its complexity is O(N^2), where N is # of atoms.
  !! When molecular grids are used, it becomes O(N_MG^2), where N_MG is # of MG.
  !! Here, we aim for linear-scaling, i.e., ~O(N_MG) or ~O(N), by utilizing
  !! (a) multicenter decomposition, (b) spherical harmonics expansion, and
  !! (c) multipole expansion techniques, (d) screening w.r.t. distance.
  !!
  !! (a) Multicenter decomposition:
  !!   implemented in potential_J via rho_A
  !!   rho(x) = SUM_A rho_A(x)
  !!     where rho_A(x) = w_A(x) rho(x) and SUM_A w_A(x) = 1 for all x.
  !!   complexity: O(N_MG^2) ==> O(N_NG) O(N) O(N_AG)
  !!
  !! (b) Spherical harmonics expansion:
  !!   implemented in potential_J_atom via rho_lm
  !!   rho_A(x) = SUM_{l=0}^{inf} SUM_{m=-l}^{l} rho_A,lm(r) Y_l^m(th,phi)
  !!     where rho_A,lm(r) = INT d{Omega} rho_A(x) Y_l^m*(th,phi)
  !!   complexity: O(N_MG) O(N) O(N_AG) ==> O(N_MG) O(N) O(N_r) O(lmax^2)
  !!   Practically, it helps a lot because lmax can be quickly trucated.
  !!   Truncation of lmax can be controlled by absolute accuracy, "-EPS_abs".
  !!
  !! (c) Multipole expansion:
  !!   implemented in potential_J_atom by finding rmax
  !!   1/|x-x'| = SUM_{l=0}^{inf}( r_<^l / r_>^{l+1} ) P_l( cos(gamma) )
  !!     where cos(gamma) = cos(th)cos(th') - sin(th)sin(th')cos(phi-phi')
  !!   complexity: O(N_MG) O(N) O(N_r) O(lmax^2)
  !!               -------      ------
  !!                ==> O(N) O(lmax^2) [ O(N_r^2) + O(N_r) + O(N_MG) ]
  !!     depending on how much MGs are close to A (r of MG < rmax).
  !!   The O(N_MG) at the end is trivial, just assigning a value rather than
  !!   performing integrals, so it would be greatly helpful for a large system.
  !!   Choosing rmax can be controlled by relative accuracy, "-EPS_rel".
  !!
  !! (d) Screening with respect to the distance
  !!   if r_A < r_J (cut-off radius), then just use a monopole (l=0 and m=0).
  !!-----------------------------------------------------------------------------
  function potential_J(P, G, A, rho, Q_A, dipole_A) result(v_J)
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    real(long), intent(in) :: rho(:)
    real(long), intent(out), optional :: Q_A(:)        ! a point charge (monopole)
    real(long), intent(out), optional :: dipole_A(:, :) ! dipole_A(1:3,i_nuc)
    real(long) :: v_J(G%N_grid)
    real(long) :: rho_A(G%N_grid) ! atomic density
    integer :: i_nuc, is, ie
    !
    ! (a) Multicenter decomposition
    !   convert rho(1:G%N_grid) into rho_A( G%i_start(i) : G%i_end(i) )
    !   ==> atomic density represented on A atomic grids, where i = 1:G%N_nuc
    !
    !do i = 1, G%N_grid
    !  rho_A(i) = G%weight_nuc( G%i_nuc(i), i ) * rho(i)
    !  v_J(i) = ZERO
    !end do
    v_J(:) = ZERO
    rho_A(:) = G%w_decomp(:)*rho(:)
    !
    ! calculate v_J for individual atomic densities
    !                rho(x')  3             rho_A(x')  3
    !   v_J(x) = INT ------- d x' = SUM INT --------- d x' = SUM v_J,A(x)
    !                 |x-x'|         A   A    |x-x'|          A
    !
    do i_nuc = 1, G%N_nuc
      if (A(i_nuc)%N_nao .eq. 0) cycle
      is = G%i_start(i_nuc)
      ie = G%i_end(i_nuc)
      if (present(Q_A) .and. present(dipole_A)) then
        v_J(:) = v_J(:) + potential_J_atom(P, G, A(i_nuc), rho_A(is:ie), &
          Q_A(i_nuc), dipole_A(1:3, i_nuc))
      else
        v_J(:) = v_J(:) + potential_J_atom(P, G, A(i_nuc), rho_A(is:ie))
      end if

    end do
  end function potential_J

  function potential_J_atom(P, G, A, rho_A, Q_A, dipole_A) result(v_J)
    implicit none
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A
    real(long), intent(in) :: rho_A(:)                    ! rho_A(1:A%N_r*A%N_ang)
    real(long), intent(out), optional :: Q_A              ! store a point charge
    real(long), intent(out), optional :: dipole_A(3)      ! store a dipole
    real(long) :: v_J(G%N_grid)
    real(long) :: r, theta, phi
    real(long), dimension(A%N_r) :: rho_lm, rj_l
    real(long) :: I_tot, d, d_00, r_c, r_l
    real(long) :: x, y, z
    integer :: i, i_r, j_r, k_r, l, m, lm
    !  integer :: is, ie
    integer :: i_nuc, N_tr, k
    real(long) :: R_tr(3), R_AB
    real(long), dimension((A%lmax_ang + 1)**2, A%N_r) :: I_inside, I_outside
    real(long), dimension((A%lmax_ang + 1)**2) :: rmax, d_lm
    real(long) :: Plm((A%lmax_ang + 1)*(A%lmax_ang + 2)/2) ! array for legendre polynomials
    real(long) :: Ylm
    !
    ! initialization
    !
    I_inside(:, :) = ZERO
    I_outside(:, :) = ZERO
    rmax(:) = ZERO
    d_lm(:) = ZERO
    d_00 = ZERO
    !
    ! calculate # of electrons in rho_A: Q_A = int_A rho_A(:) d^3 r_A
    ! Q_A will be used as a point charge
    !
    !  is = G%i_start(A%i_nuc)
    !  ie = G%i_end(A%i_nuc)
    !Q_A = sum( rho_A(:) * G%weight2c(is:ie,A%i_nuc) )
    !
    ! (b) Spherical harmonics expansion
    ! For each rho_lm, we perform double integrals based on multipole expansion.
    !
    loop_l: do l = 0, A%lmax_ang
      if (l .eq. 0) then
        rj_l(:) = ONE
      else
        rj_l(:) = rj_l(:)*A%radial%r(:)        ! rj_l(:) = A%r(:)**l
      end if
      loop_m: do m = -l, l
        lm = l**2 + (m + l) + 1         ! lm = 1, 2, 3, ...
        !
        ! construct rho_lm at every single j_r
        !
        !   rho_A,lm(r') = INT sin theta' d theta' INT d phi' *
        !                  rho_A(vec_r') Y_lm(theta',phi')
        !
        ! WARNING: be careful with the index of rho_A(:)
        !          this index should be consistent with molecular_grid.f90
        !
        ! here is more understandable version...
        !   rho_lm(:) = ZERO;  j = 1
        !   atomic_angular: do j_ang = 1, A%N_ang
        !     atomic_radial: do j_r = 1, A%N_r
        !       rho_lm(j_r) = rho_lm(j_r) + rho_A(j) * A%Y_lm(j_ang,lm) * &
        !                                   FOUR*PI * A%ang_weight(j_ang)
        !       j = j + 1
        !     end do atomic_radial
        !   end do atomic_angular
        !
        do j_r = 1, A%N_r
          rho_lm(j_r) = sum(rho_A(j_r::A%N_r)*A%Y_lm(:, lm)*A%angular%weight(:))
          rho_lm(j_r) = rho_lm(j_r)*FOUR*PI   ! 4 pi missed in A%ang_weight
        end do
        !
        ! KEY POINT of (b)
        ! neglecting values based on how small d_lm is,
        !              rmax
        ! where d_lm = INT dr | rho_lm(r) | r^2
        !               0
        ! i.e., on checking ( d_lm / d_00 ) < EPS_lm
        !
        ! Since it is assumed that rho_A is very much atom-centered density,
        ! rho_00 would be dominant and only a few (l,m) sets are needed.
        ! Therefore, lmax is determined here by checking d_lm / d_00.
        !
        d_lm(lm) = sum(abs(rho_lm(:))*A%radial%r(:)**2*A%radial%weight(:))
        if (l .eq. 0) then
          d_00 = d_lm(lm)
        else
          if (abs(d_lm(lm)) .lt. P%EPS_lm*d_00) cycle loop_m
        end if
        !
        ! (c) Perform double integrals based on multipole expansion
        !                    4pi   m                   2           r<^l
        !   v_J,A(x) = SUM ------ Y (th,phi) INT dr' r'  rho(r') --------
        !              l,m (2l+1)  l                       A,lm  r>^(l+1)
        !                                    ----------------------------
        !                                      => I(r)
        !
        !             1     r        2          l    l rmax      2           1
        !   I(r) = ------- INT dr' r' rho(r') r'  + r  INT dr' r' rho(r') --------
        !          r^(l+1)  0           A,lm            r           A,lm  r'^(l+1)
        !                  ----------------------      ---------------------------
        !                     => I_inside(lm,r)              => I_outside(lm,r)
        !
        !   i: molecular grid index for vec{r}
        !   j: atomic grid index for vec{r'}
        !   i_r: radial grid index of AG for vec{r}
        !   j_r: radial grid index of AG for vec{r'}
        !
        ! if l is large, then we need to worry about accuracy of huge numbers!!!
        !
        ! calculate I_inside in the forward direction until rmax
        ! NOTE: A%r(j_r)**(2+l) --> A%r(j_r)**2 * rj_l(j_r)
        !
        d = abs(rho_lm(1))*A%radial%r(1)**2*A%radial%weight(1)
        I_inside(lm, 1) = rho_lm(1)*A%radial%r(1)**2*rj_l(1)*A%radial%weight(1)
        loop_rp: do j_r = 2, A%N_r         ! performing integral with r'
          I_inside(lm, j_r) = I_inside(lm, j_r - 1) + &
            rho_lm(j_r)*A%radial%r(j_r)**2*rj_l(j_r)*A%radial%weight(j_r)
          !
          ! KEY POINT of (c)
          ! find out rmax and skip all calculations for r > rmax
          ! rmax is determined by how close d_lm(r) is to d_lm(rmax), i.e.,
          ! by checking [ 1 - d_lm(r) / d_lm(rmax) ] < EPS_r
          !
          d = d + abs(rho_lm(j_r))*A%radial%r(j_r)**2*A%radial%weight(j_r)
          if (d_lm(lm) - d .lt. P%EPS_r*d_lm(lm)) exit loop_rp
        end do loop_rp
        !
        ! set up rmax such that
        !   INT_0^rmax dr r^2 rho_lm(r)
        !   ~= INT_0^A%rmax dr r^2 rho_lm(r)
        ! k_r is the radial grid index such that A%r(k_r) = rmax
        !
        k_r = min(j_r, A%N_r)
        rmax(lm) = A%radial%r(k_r)           ! in the worst case, rmax = A%r(A%N_r)
        if (k_r .lt. A%N_r) then
          I_inside(lm, k_r + 1:) = I_inside(lm, k_r)
          I_outside(lm, k_r + 1:) = ZERO
        end if
        !
        ! calculate I_outside in the backward direction from rmax
        ! NOTE: A%r(j_r)**(2-(l+1)) --> A%r(j_r) / rj_l(j_r)
        !
        I_outside(lm, k_r) = rho_lm(k_r)*A%radial%r(k_r)/rj_l(k_r)*A%radial%weight(k_r)
        do j_r = k_r - 1, 1, -1
          I_outside(lm, j_r) = I_outside(lm, j_r + 1) + &
            rho_lm(j_r)*A%radial%r(j_r)/rj_l(j_r)*A%radial%weight(j_r)
        end do
        !
        ! print out some useful information in potential_J
        !
        !if ( P%yes_verbose ) then
        !        write(P%OUTPUT, "('# pot_J: i_nuc=',i3,', (l,m)=',i2,i3,', " // &
        !          "d_lm=',es10.2,', rmax=',f6.2)") A%i_nuc, l, m, d_lm(lm), rmax(lm)
        !end if
      end do loop_m
    end do loop_l
    !
    ! r_c is given by
    !   4 pi   d_00                            d_00
    !   ---- * ---- * Y_00 < EPS_00 ==> r_c = ------ * 4 pi * Y_00
    !    1      r                             EPS_00
    !
    !   Y_00 = 1 / sqrt( 4 pi )
    !   G%Y_lm_prefactor(1) = sqrt( 4 pi ), i.e. r_c = Q_A / EPS_00
    !
    if (P%EPS_00 .le. TINY(P%EPS_00)) then
      r_c = HUGE(r_c)                     ! avoid division by zero
    else
      r_c = d_00*G%Y_lm_prefactor(1)/P%EPS_00
    end if
    !
    ! Q_A: the number of electrons in the A-th atomic electronic density
    !      or the monopole
    ! Q_A = INT d^3 r' rho_A(r') = sqrt(4PI) * d_00^A
    ! and I_inside(1,A%N_r) means I_inside(00,rmax_A) ==> angular INT ==> Q_A
    !     Y_0,0 = sqrt(1/4PI)             defined in sph_harmonics.f90
    !     G%Y_lm_prefactor(1)             defined in molecular_grid.f90
    !           = sqrt(1/4PI) * 4PI / (2l+1) = sqrt(4PI)
    !                           ------------ <== angular INT with l
    !
    if (present(Q_A)) Q_A = I_inside(1, A%N_r)*G%Y_lm_prefactor(1)
    !
    ! dipole_A: the dipole of the A-th atomic electronic density
    !   I_inside(2,A%N_r) = INT dr' r'^3 rho_1,-1^A(r'_A)
    !   I_inside(3,A%N_r) = INT dr' r'^3 rho_1,0^A(r'_A)
    !   I_inside(4,A%N_r) = INT dr' r'^3 rho_1,+1^A(r'_A)
    !   dipole_A_x = INT d^3 r' rho_A(vec_r') x
    !              = INT d^3 r' rho_A(vec_r') r' sin theta' cos phi'
    !              = - INT d^3 r' rho_A(vec_r') r' Y_1,+1 sqrt(4pi/3)
    !              = - INT dr' r'^3 rho_1,+1^A(r') sqrt(4pi/3)
    !              = - I_inside(4,A%N_r) * G%Y_lm_prefactor(4)
    !   dipole_A_y = INT d^3 r' rho_A(vec_r') y
    !              = INT d^3 r' rho_A(vec_r') r' sin theta' sin phi'
    !              = - INT d^3 r' rho_A(vec_r') r' Y_1,-1 sqrt(4pi/3)
    !              = - INT dr' r'^3 rho_1,-1^A(r') sqrt(4pi/3)
    !              = - I_inside(2,A%N_r) * G%Y_lm_prefactor(2)
    !   dipole_A_z = INT d^3 r' rho_A(vec_r') z
    !              = INT d^3 r' rho_A(vec_r') r' cos theta'
    !              = INT d^3 r' rho_A(vec_r') r' Y_1,0 sqrt(4pi/3)
    !              = INT dr' r'^3 rho_1,0^A(r') sqrt(4pi/3)
    !              = I_inside(3,A%N_r) * G%Y_lm_prefactor(3)
    ! NOTE:
    !   sph_harmonics_real is defined in sph_harmonics.f90
    !     Y_1,0  = sqrt(3/4pi) cos theta
    !     Y_1,-1 = sqrt(3/4pi) (-sin theta) sin phi
    !     Y_1,+1 = sqrt(3/4pi) (-sin theta) cos phi
    !              ----------- ==> * 4pi/(2l+1) ==> Y_lm_prefactor
    !   Y_lm_prefactor is defined in molecular_grid.f90
    !     G%Y_lm_prefactor(2) = G%Y_lm_prefactor(4) = sqrt(4pi/3)
    !     G%Y_lm_prefactor(3) = sqrt(4pi/3)
    !
    if (present(dipole_A)) then
      dipole_A(1) = -I_inside(4, A%N_r)*G%Y_lm_prefactor(4)     ! dipole_x
      dipole_A(2) = -I_inside(2, A%N_r)*G%Y_lm_prefactor(2)     ! dipole_y
      dipole_A(3) = I_inside(3, A%N_r)*G%Y_lm_prefactor(3)     ! dipole_z
    end if
    !
    ! construct v_J(r) for each molecular grid point r_i
    !
    v_J(:) = ZERO
    N_tr = 1      ! no translation if G%yes_periodic is false
    if (G%yes_periodic) N_tr = G%N_tr
    loop_i: do i = 1, G%N_grid
      i_nuc = G%i_nuc(i)
      loop_tr: do k = 1, N_tr  ! N_tr=1 if no periodic; BE CAREFUL WITH INDENTATION
        !
        ! r_cutoff applied: REMEMBER it's always applied only for nuc. positions
        !
        R_AB = bond_distance(G, A%i_nuc, i_nuc, k)  ! <-- good scanning order
        if (G%yes_cutoff .and. R_AB .gt. G%r_cutoff) exit loop_tr  ! skip the rest
        !
        ! obtain a grid position that is translated w.r.t. i_nuc and A%i_nuc
        !
        R_tr(:) = obtain_grid_position(G, i, A%i_nuc, k)
        x = R_tr(1)         ! x = G%x(i) - A%x when no periodic is used
        y = R_tr(2)         ! y = G%y(i) - A%y when no periodic is used
        z = R_tr(3)         ! z = G%z(i) - A%z when no periodic is used
        call xyz_to_sph(x, y, z, r, theta, phi)
        !
        ! KEY POINT of (d)
        ! if r_A is large enough ( r_A >= r_c ),
        ! then treat as a point charge: v_J(i) = Q_A / r (or just the monopole)
        !
        if (r .ge. r_c) then
          v_J(i) = v_J(i) + I_inside(1, A%N_r)*G%Y_lm_prefactor(1)/r  ! monopole
          cycle loop_tr     ! skip if r >= r_c and go next translation
        end if
        !
        ! do-loop for l = 0, 1, 2, ..., lmax and for m = -l, ..., l
        !
        r_l = ONE/r
        call associated_Legendre_P_array(A%lmax_ang, cos(theta), Plm)
        loop2_l: do l = 0, A%lmax_ang
          r_l = r_l*r                     ! r_l := r**l
          loop2_m: do m = -l, l
            lm = l**2 + (m + l) + 1       ! lm = 1, 2, 3, ...
            if (abs(d_lm(lm)/d_00) .lt. P%EPS_lm) cycle loop2_m
            !
            ! calculate v_J for given l and m at every single vec{r},
            ! utilizing pre-calcaulted double integrals, I_inside and I_outside.
            !
            ! WARNING: we can further optimize sph_harmonic_real,
            !          which are calculated by recursive relation w.r.t. l.
            !
            if (r .ge. rmax(lm)) then       ! vec{r} is outside rmax
              I_tot = I_inside(lm, A%N_r)/r_l/r  ! ... / r**(l+1)
            else if (r .le. A%radial%r(1)) then    ! even smaller than closest point in AG
              I_tot = I_outside(lm, 1)*r_l
            else            ! otherwise, find i_r where A%r(i_r-1) < r <= A%r(i_r)
              do i_r = 2, A%N_r - 1
                if (A%radial%r(i_r) .ge. r) exit
              end do
              I_tot = I_inside(lm, i_r - 1)/r_l/r + I_outside(lm, i_r)*r_l
            end if
            !
            ! compute spherical harmonics using precalculated legendre polynomials
            !
            if (m .eq. 0) then
              Ylm = G%Y_lm_prefactor(lm)*Plm(l*(l + 1)/2 + abs(m) + 1)
            else if (m .gt. 0) then
              Ylm = G%Y_lm_prefactor(lm)*Plm(l*(l + 1)/2 + abs(m) + 1) * cos(dble(m)*phi)
            else
              Ylm = G%Y_lm_prefactor(lm)*Plm(l*(l + 1)/2 + abs(m) + 1) * sin(dble(-m)*phi)
            end if
            v_J(i) = v_J(i) + I_tot*Ylm
          !        v_J(i) = v_J(i) + I_tot*calculate_sph_on_A_i_nuc(A,G,i,A%i_nuc,l,m)
          end do loop2_m
        end do loop2_l
      end do loop_tr
    end do loop_i
  end function potential_J_atom

  !>-----------------------------------------------------------------------------
  !! Yajiang's original implementation for the direct Coulomb potential
  !! - potential_J_MC()
  !! - potential_J_A_R()
  !! - integral_A()
  !! - potential_J_A()
  !!-----------------------------------------------------------------------------
  function potential_J_MC(P, G, A, rho) result(v)
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    real(long), intent(in) :: rho(:)
    real(long), dimension(G%N_grid) :: r, theta, phi, v_nlm, v, radial_f
    real(long) :: rho_n(P%N_nuc, G%N_grid)
    real(long), allocatable :: rho_A(:, :), rho_lm(:), v_lm(:)
    integer, allocatable :: ang_A(:, :)
    real(long) :: x, y, z
    integer :: l, m, lmax, Nr_max, Nang_max, N_r, N_ang
    integer :: i_nuc, i_ang, i_r, j
    lmax = P%lmax_MP
    Nr_max = maxval(A(:)%N_r)
    Nang_max = maxval(A(:)%N_ang)
    !
    ! no pointer and no allocate --> we should return all of pointers at the end
    ! instead --> real(long) :: r(G%N_grid)
    !
    allocate (rho_A(Nr_max, Nang_max), ang_A(Nr_max, Nang_max), rho_lm(Nr_max), v_lm(Nr_max))

    v(:) = ZERO
    do i_nuc = 1, P%N_nuc
      rho_n(i_nuc, :) = G%weight_nuc(i_nuc, :)*rho(:)
      do j = 1, G%N_grid  !coordinate of MG in ith nuclear coordinate system
        x = G%x(j) - A(i_nuc)%x
        y = G%y(j) - A(i_nuc)%y
        z = G%z(j) - A(i_nuc)%z
        call xyz_to_sph(x, y, z, r(j), theta(j), phi(j))
      end do
      N_r = A(i_nuc)%N_r
      N_ang = A(i_nuc)%N_ang
      do j = 1, G%N_grid
        if (G%i_nuc(j) .eq. i_nuc) then
          i_r = G%i_r(j)
          i_ang = G%i_ang(j)
          rho_A(i_r, i_ang) = rho_n(i_nuc, j)
          ang_A(i_r, i_ang) = j
        end if
      end do
      do l = 0, lmax  !expanding as spherical harmonics for each rho_A(r,angle)
        do m = -l, l
          do i_r = 1, N_r
            rho_lm(i_r) = ZERO
            do i_ang = 1, N_ang
              j = ang_A(i_r, i_ang)
              rho_lm(i_r) = rho_lm(i_r) + rho_A(i_r, i_ang) &  !for each radial, integrate about angular part, solid angle
                * sph_harmonics_real(l, m, theta(j), phi(j)) &  !rho_lm(r): spherical symmetry
                * FOUR * PI * A(i_nuc)%angular%weight(i_ang)
            end do
          end do
          do j = 1, G%N_grid
            radial_f(j) = potential_J_A_R(A(i_nuc), rho_lm(1:N_r), r(j), l)  !calculate Coulomb potential for rho_lm(r) on AG
            v_nlm(j) = radial_f(j)*sph_harmonics_real(l, m, theta(j), phi(j))  !complete the angular part
            v(j) = v(j) + v_nlm(j)  !summing the potential v_nlm(r)
          end do
        end do
      end do
    end do
    deallocate (rho_A, ang_A, rho_lm, v_lm)
  end function potential_J_MC

  function potential_J_A_R(A, rho, r, l) result(v)
    type(Atom_set), intent(in) :: A
    real(long), intent(in) :: rho(:)
    real(long), intent(in) :: r
    integer, intent(in) :: l
    real(long) :: v
    real(long) :: r_max(A%N_r)
    integer :: i, j_s
    j_s = A%N_r
    do i = 1, A%N_r
      if (A%radial%r(i) .lt. r) then
        r_max(i) = (A%radial%r(i)/r)**l/r
      else
        r_max(i) = (r/A%radial%r(i))**l/A%radial%r(i)
      end if
    end do
    v = integral_A(A, rho(1:j_s)*r_max(1:j_s))/(TWO*l + 1)
  end function potential_J_A_R


  function integral_A(A, v) result(I)
    !! integral on atomic grid    
    type(Atom_set), intent(in) :: A
    real(long), intent(in) :: v(:)
    real(long) :: I
    I = sum(v(:)*FOUR*PI*A%radial%r(:)**2*A%radial%weight(1:A%N_r))  ! the range is given by v(:)
  end function integral_A


  function potential_J_A(A, rho) result(v)
    !! Calculating Coulomb potential on the atomic grid
    type(Atom_set), intent(in) :: A
    real(long), intent(in) :: rho(:)
    real(long) :: v(A%N_r)
    real(long) :: r_max(A%N_r)
    integer :: i, j_s

    j_s = A%N_r
    r_max(:) = A%radial%r(:)
    do i = 1, A%N_r
      r_max(1:i) = A%radial%r(i)
      v(i) = integral_A(A, rho(1:j_s)/r_max(1:j_s))
    end do
  end function potential_J_A



  !>-----------------------------------------------------------------------------
  !! construct atomic potentials from a molecular potential
  !! then make them spherical symmetric
  !! then calculate eigenstates (NAOs) for each atomic potential
  !! then modify A(:) with newly calculated NAOs.
  !!-----------------------------------------------------------------------------
  subroutine construct_NAO_from_molecular_potential(A, G, pot)
    type(Atom_set), intent(inout) :: A(:)
    type(Grid3D), intent(in) :: G
    real(long), intent(in) :: pot(:)
    real(long) :: pot_A(G%N_grid)
    integer :: i_nuc
    !
    ! decompose the molecular potential into the atomic components
    !
    pot_A(:) = G%w_decomp(:)*pot(:)
    !
    ! make a spherically-averaged (atomic) potential for each atomic component
    !
    do i_nuc = 1, G%N_nuc
      call treat_each_atom(A(i_nuc), pot_A(G%i_start(i_nuc):G%i_end(i_nuc)))
    end do
  contains
    subroutine treat_each_atom(A, pot_A)
      type(Atom_set), intent(inout) :: A
      real(long), intent(in) :: pot_A(:)  ! pot_A(1:A%N_r*A%N_ang)
      real(long) :: pot_r(A%N_r)
      real(long) :: H(A%N_r, A%N_r), E(A%N_r)
      integer :: i, j, l, l_max, n
      !
      ! the atomic potential is spherically averaged as
      !
      !                INT dr^3 pot (r,theta,phi)
      !   pot(r) = ----------------------------------
      !            INT INT sin(theta) d(theta) d(phi)
      !
      ! A%ang_weight(:) doesn't include a factor of 4 pi.
      ! However, pot(r) is divided by INT INT sin(theta) d(theta) d(phi) = 4 pi.
      ! So, this factor is not shown up at all in the following expression.
      !
      do j = 1, A%N_r
        pot_r(j) = sum(pot_A(j::A%N_r)*A%angular%weight(:))
      end do
      !
      ! print out the atomic potential
      !
        !    open( 100, file=trim(A%label)//".pot" )
      !    write( 100, * ) pot_r(:)
      !
      ! get eigenstates and put them into A
      !
      l_max = maxval(A%l_AO(:))
      do l = 0, l_max
        H(:, :) = -HALF*A%radial%D2(:, :)
        do j = 1, A%N_r
          H(j, j) = H(j, j) + pot_r(j) + dble(l*(l + 1)/2)/A%radial%r(j)**2
        end do
        call get_eigenvalues(H, E)
        do i = 1, A%N_nao
          if (A%l_AO(i) .eq. l) then
            n = A%n_AO(i)
            A%u_AO(:, i) = H(:, n - l)*A%radial%wf_coeff(:)
            A%D2_AO(:, i) = matmul(A%radial%D2(:, :), &
              A%u_AO(:, i)/A%radial%wf_coeff(:)) &
              * A%radial%wf_coeff(:) &
              - dble(A%l_AO(i)*(A%l_AO(i) + 1))/A%radial%r(:)**2*A%u_AO(:, i)
          end if
        end do
      end do
    end subroutine treat_each_atom
  end subroutine construct_NAO_from_molecular_potential

  !>----------------------------------------------------------------------------
  !! print spherically averaged quantity (potential or density) for each atom
  !!       with and without multicenter decomposition
  !!----------------------------------------------------------------------------
  subroutine print_atomic_quantity(A, G, v, OUTPUT)
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    real(long), intent(in) :: v(:)        ! it can be potential or density
    integer, intent(in) :: OUTPUT
    real(long) :: v_A(G%N_grid)
    integer :: i_nuc, is, ie
    !
    ! decompose the molecular quantity into the atomic components
    !
    v_A(:) = G%w_decomp(:)*v(:)
    !
    ! make a spherically-averaged (atomic) quantity for each atomic component
    !
    write (OUTPUT, "('#   ',a)") "v:   spherically-averaged quantity"
    write (OUTPUT, "('#   ',a)") &
      "v_A: spherically-averaged quantity with multicenter decomposition"
    write (OUTPUT, "('#',a15,2a16)") "r(a.u.)", "v", "v_A"
    do i_nuc = 1, G%N_nuc
      write (OUTPUT, "('# i_nuc = ',i5)") i_nuc
      is = G%i_start(i_nuc)
      ie = G%i_end(i_nuc)
      call treat_each_atom(A(i_nuc), v(is:ie), v_A(is:ie), OUTPUT)
      write (OUTPUT, "(/)")
    end do
  contains
    subroutine treat_each_atom(A, v, v_A, OUTPUT)
      type(Atom_set), intent(in) :: A
      real(long), intent(in) :: v(:), v_A(:)    ! v_A(1:A%N_r*A%N_ang)
      integer, intent(in) :: OUTPUT
      real(long) :: v_r, v_A_r
      integer :: j
      !
      ! the atomic quantity is spherically averaged.
      ! See construct_NAO_from_molecular_potential() for details.
      !
      do j = 1, A%N_r
        v_r = sum(v(j::A%N_r)*A%angular%weight(:))
        v_A_r = sum(v_A(j::A%N_r)*A%angular%weight(:))
        write (OUTPUT, "(3es16.8)") A%radial%r(j), v_r, v_A_r
      end do
    end subroutine treat_each_atom
  end subroutine print_atomic_quantity

  subroutine find_optimal_deltaD(P, F, coef)
    !! This subroutines returns a set of coefficients for 'optimal'
    !! representing the current Density matrix F%D as a linear combination
    !! of the saved density matrices in the F%Dsave array
    !! F%D(:,:) = sum_i  coef(i) * F%Dsave(:,:,i)  + deltaD(:,:)
    !! 'optimal' means that the rest is small in that sense that
    !! it has minimal number of nonzero elements.
    !! To achieve this, we use unconstrained Newton minimizer to minimize the function
    !!
    !!   sum_{i<=j} ( 1 - f(  (D_{i,j} - sum_k  c_k * Dsave_{i,j,k})^2 - eps^2 ) )
    !!
    !!   with respect to the coefficients c_k
    !!   where f is the fermi function
    !!
    !!         f=1/(1+exp(x/T))
    !!
    !!   with some chosen temperature T
    !!
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F

    real(long), dimension(F%nsave), intent(inout) :: coef ! the coefficients

    real(long), dimension(P%N_basis, P%N_basis) :: DeltaDmat

    real(long) :: nonzero
    real(long), dimension(F%nsave) :: delta_coef
    real(long), dimension(F%nsave) :: nonzero_deriv
    real(long), dimension(F%nsave, F%nsave) :: nonzero_hessian

    integer :: ierr, count, iter
    integer :: i, j, k, l
    real(long) :: temp
    real(long) ::  fderiv, f2deriv
    real(long) :: epsilon

    epsilon = P%SCF_density_conv/10.

    ! temperature
    temp = 100*epsilon

    ! initial guess for coef
    coef(:) = ZERO
    coef(F%nsave) = ONE
    iter = 0
    ! Minimization loop
    do
      ! calculate deltaD
      DeltaDmat = F%D
      do k = 1, F%nsave
        DeltaDmat = DeltaDmat - coef(k)*F%Dsave(:, :, k)
      end do

      ! calculate nonzero = the function to be minimize, its derivatives, and hessian
      nonzero = ZERO
      nonzero_deriv(:) = ZERO
      nonzero_hessian(:, :) = ZERO

      count = 0
      do i = 1, P%N_basis
        do j = i, P%N_basis
          if (abs(DeltaDmat(i, j)) .gt. epsilon) count = count + 1

          nonzero = nonzero + ONE - fermi(DeltaDmat(i, j)**2 - (epsilon)**2, temp)
          fderiv = fermi_deriv(DeltaDmat(i, j)**2 - epsilon**2, temp)
          f2deriv = fermi_2ndderiv(DeltaDmat(i, j)**2 - epsilon**2, temp)
          do k = 1, F%nsave
            nonzero_deriv(k) = nonzero_deriv(k) + fderiv &
              * TWO * DeltaDmat(i, j)*F%Dsave(i, j, k)
            do l = 1, F%nsave
              nonzero_hessian(k, l) = nonzero_hessian(k, l) - &
                TWO * F%Dsave(i, j, k)*F%Dsave(i, j, l) * &
                (TWO * f2deriv * DeltaDmat(i, j)**2 + fderiv)
            end do ! l loop
          end do ! k loop
        end do ! j loop
      end do ! i loop

      ! delta_coef = hessian^-1 * deriv
      call solve_AxB(nonzero_hessian, nonzero_deriv, delta_coef, ierr)
      coef = coef - delta_coef

      if (ALL(abs(nonzero_deriv(:)) .lt. epsilon)) exit
      if (iter .gt. 200) then
        write (STDERR, *) 'stuck in find_optimal_deltaD'
        coef = ZERO
        exit
      end if

      iter = iter + 1
    end do ! minimizer loop

  contains
    ! Fermi function and its derivatives
    pure real(long) function fermi(E, T) result(ret)
      implicit none
      real(long), intent(in) :: E, T
      real(long) :: xi, expo
      xi = E/T;
      if (xi .gt. 0.0) then;
        expo = exp(-xi); ret = expo/(1 + expo)
      else;
        expo = exp(xi); ret = ONE/(1 + expo);
      end if
      ret = ONE/(ONE + exp(xi));
    end function fermi

    pure real(long) function fermi_deriv(E, T) result(ret)
      implicit none
      real(long), intent(in) :: E, T
      real(long) :: xi, expo
      xi = E/T;
      if (xi .gt. 0.0) then;
        expo = exp(-xi); ret = -expo/(T*(1 + expo)**2)
      else;
        expo = exp(xi); ret = -expo/(T*(1 + expo)**2)
      end if
    end function fermi_deriv

    pure real(long) function fermi_2ndderiv(E, T) result(ret)
      implicit none
      real(long), intent(in) :: E, T
      real(long) :: xi, expo
      xi = E/T;
      !     expo=exp(xi);   ret=TWO/(1+expo)*expo/(T**2*(1+expo)**2) - expo/(T**2*(1+expo)**2)
      expo = exp(xi); ret = (ONE - TWO*fermi(E, T))*fermi_deriv(E, T)/T
    end function fermi_2ndderiv

  end subroutine find_optimal_deltaD

  function groundstate_occ(occ, Ncore, open_shell, add_electrons) result(gs_occ)
    !! returns ground state occupation numbers
    integer, intent(in) :: occ(:)
    integer, optional :: Ncore
    logical, optional :: open_shell
    logical, optional :: add_electrons
    logical :: yes_open_shell, yes_add_electrons
    integer :: gs_occ(size(occ))
    integer :: i, j
    yes_open_shell = .False.
    yes_add_electrons = .True.
    if (present(open_shell)) &
      yes_open_shell = open_shell
    if (present(add_electrons)) &
      yes_add_electrons = add_electrons
    i = 1
    if (present(Ncore)) then
      i = Ncore + 1
    end if
    gs_occ = occ
    i_loop: do
      if (sum(gs_occ(i + 1:)) .eq. 0) exit i_loop
      if (gs_occ(i) .eq. 2) then
        i = i + 1
        cycle i_loop
      end if
      if (gs_occ(i) .lt. 2) then
        j_loop: do j = i + 1, size(occ)
          if (gs_occ(j) .gt. 0) then
            gs_occ(i) = gs_occ(i) + 1
            gs_occ(j) = gs_occ(j) - 1
            exit j_loop
          end if
        end do j_loop
      end if
    end do i_loop
    ! add an electron, such that we have a closed valence shell
    if (.NOT. yes_open_shell .and. gs_occ(i) .eq. 1) then
      if (yes_add_electrons) then
        gs_occ(i) = 2
      else
        gs_occ(i) = 0
      end if
    end if
  end function groundstate_occ

end module operations
