#!/usr/bin/env python
#
# XMOLECULE electronic structure toolkit
# Copyright 2023 Ludger Inhester
# Ludger Inhester, Center for Free-Electron Laser Science, DESY
# ludger.inhester@cfel.de
#
# XMOLECULE is distributed under the terms of the xxx licence.
#
# program for geometry optimization
# author: Ludger Inhester

import sys
import re
import numpy as np
import argparse
from scipy.optimize import fmin_bfgs, minimize
import pyxmol

AMU2AU = 1822.877332827
BOHR2ANGSTROM = 0.529177249

geom = None


def optimize_geom():
    args = parseArguments()
    XM = pyxmol.xmol(args.filename)
    XM.init()

    def getFuncEn():
        def f(x):
            global geom
            XM.set_geom(x.tolist())
            print(" geom: ")
            for i in range(0, len(x)//3):
                print(" %s %10.5f %10.5f %10.5f \b" %
                      (atomLabels[i], x[i*3], x[i*3+1], x[i*3+2]))
            geom = x
            sys.stdout.flush()
            E = XM.energy
            return (E)
        return f
    potential = getFuncEn()


    def getFuncGr():
        def f(x):
            global geom
            XM.set_geom(x.tolist())
            geom = x
            G = XM.gradient()
            print(" Gradient: (RMS=%e)" %
                  np.sqrt(sum(G[i]**2 for i in range(0, len(G)))))
            for i in range(0, len(G)//3):
                print(" %s %10.5f %10.5f %10.5f \b" %
                      (atomLabels[i], G[i*3], G[i*3+1], G[i*3+2]))
            return np.array(G)
        return f
    gradients = getFuncGr()


    atomLabels = XM.atom_labels()
    pos = np.array(XM.get_geom())

    for i in range(0, len(pos)//3):
        print(" %s %10.5f %10.5f %10.5f \b" %
              (atomLabels[i], pos[i*3], pos[i*3+1], pos[i*3+2]))


    res = fmin_bfgs(potential, pos, fprime=gradients, gtol=args.tolerance,
                    disp=True, full_output=True,  retall=True)


    # res = minimize(potential, pos, method="CG", jac=gradients , tol=1e-8 , options={"disp": True} )
    print(res)

    G = res[2]
    E = res[1]
    pos = res[0]

    BO = XM.bondorder()
    for i in range(0, len(pos)//3):
        for j in range(0, i):
            print("BO atom %d (%s) - %d (%s ) %f\n" %
                  (i, atomLabels[i], j, atomLabels[j], BO[i][j]))


    print("Final Gradient: (RMS=%e)" %
          np.sqrt(sum(G[i]**2 for i in range(0, len(G)))))
    for i in range(0, len(pos)//3):
        print(" atom %d %10.5f %10.5f %10.5f \b" % (i, G[i*3], G[i*3+1], G[i*3+2]))
    print("Final Energy: %f " % E)
    print("Optimized Geometry (in Bohr):")
    for i in range(0, len(pos)//3):
        print(" %s %10.5f %10.5f %10.5f \b" %
              (atomLabels[i], pos[i*3], pos[i*3+1], pos[i*3+2]))

    print("Optimized Geometry (in Angstrom):")
    for i in range(0, len(pos)//3):
        print(" %s %10.5f %10.5f %10.5f \b" % (
            atomLabels[i],
            pos[i*3] * BOHR2ANGSTROM,
            pos[i*3+1] * BOHR2ANGSTROM,
            pos[i*3+2] * BOHR2ANGSTROM))


    geom = pos.reshape((len(pos)//3, 3))
    print(geom)
    distances = np.array([[np.linalg.norm(geom[i, :] - geom[j, :])
                     for i in range(geom.shape[0])] for j in range(geom.shape[0])])
    print("Distance matrix in bohr:")
    print(distances)
    print("Distance matrix in Angstrom")
    print(distances * BOHR2ANGSTROM)

def parseArguments():
    parser = argparse.ArgumentParser(
                    description='optimize the molecular geometry')
    parser.add_argument('filename', help="xmolecule input filename")
    parser.add_argument('--tolerance', help="gradient tolerance",
                        type=float,
                        default=1e-5)
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    optimize_geom()
