!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module photoionization
  !!  author: Ludger Inhester
  !!
  !! This module is for the calculation of Ionization rates via independent-atomic-center approximation

  use angularintegration
  use constants
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use interpolate
  use process
  use averaging
  use spin_coupling
  use mcscf
  use hartreefock
  use anglib
  use operations
  use basis_func
  use rwfile
  use scf
  implicit none
  type :: TransitionDipoleMoments
    !! bound-continuum transition dipoles information (read-in from xatom)
    real(long), allocatable :: energy(:) !! energy
    real(long), allocatable :: trans_dipole(:, :) !! transition dipole
    integer, allocatable :: config(:, :) !! atomic configuration

    real(long), allocatable :: Pmin(:, :) !! basis to min basis projection
    real(long), allocatable :: Omin(:, :) !! keeps the orbital expansion in minimal basis
    real(long), allocatable :: Smin(:, :) !! keeps the overlap matrix in minimal basis
    integer, allocatable :: min_info(:, :) !! info on minimal basis:
    !! min_info(:,1)=atom index
    !! min_info(:,2)=n
    !! min_info(:,3)=l
    !! min_info(:,4)=m
  end type TransitionDipoleMoments
  private
  public :: calculate_photoIonization_CS, calculate_photoIonization_CS_CI, &
    calculate_photoIonization_CS_spin, calculate_photoIonization_CS_spin_avg, shakeup_analysis
contains

  subroutine calculate_photoIonization_CS(P, A, B, F, S, fpath)
    !! calculates photoionization cross sections based on orbital energies
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S
    character*(*), optional, intent(in) :: fpath(P%N_nuc)  !! Mar.19.2015: optional parameter that specifies tdipole-filepath to be read in

    real(long) :: omega
    ! the total  cross section (cartesian: x,y,z)
    real(long), dimension(F%N_basis, 3) :: CS
    integer :: i
    !real(long) :: AU2MB = AU2M**2 / 1.0e-22_long

    ! photon energy
    omega = P%PE

    CS = photoIonization_CS(P, A, B, F, omega, fpath)
    ! add statistical factor
    do i = 1, F%N_basis
      CS(i, :) = CS(i, :)*TWO*stat_factor_occupied(P, F, F%N_occ, i)
    end do
    do i = 1, F%N_basis
      if (.NOT. is_occupied(P, F, F%N_occ, i)) cycle
      if (omega + F%CE(i) .lt. 0) cycle
      call add_Process(P, F, S, "photoionization", sum(CS(i, :))/THREE, omega + F%CE(i), i)
    end do

    if (P%print_level .gt. 1) then
      write (P%OUTPUT, "('# Photoion cross sections for omega=',e13.5,' eV in a.u.')") &
        omega*AU2EV
      write (P%OUTPUT, "('# ',a4,3(2x,a13),4x,a13,4x,a13)") &
        'MO', 'x-pol', 'y-pol', 'z-pol', 'averaged', 'pe energy/eV'
      do i = 1, F%N_basis
        if (.NOT. is_occupied(P, F, F%N_occ, i)) cycle
        write (P%OUTPUT, "('# ',i4,3(2x,es13.5),4x,es13.5,4x,es13.5)") &
          i, CS(i, :), sum(CS(i, :))/THREE, AU2EV*(omega + F%CE(i))
      end do
      write (P%OUTPUT, "('# --------------------------')")
      write (P%OUTPUT, "('# ',a4,3(2x,es13.5),4x,es13.5)") &
        'sum:', sum(CS(:, 1)), sum(CS(:, 2)), sum(CS(:, 3)), sum(CS(:, :))/THREE
      write (P%OUTPUT, "('# ')")
    end if
    ! write(P%OUTPUT,"('# Photoion cross sections for omega=',e13.5,' eV in Mb')") &
    !      omega * AU2EV

    ! do i=1,F%N_basis
    !    if( .NOT. is_occupied(P,F,F%N_occ,i)) cycle
    !    if( omega + F%CE(i) < 0 )cycle
    !    write(P%OUTPUT,"('# ',i4,3(2x,es13.5),4x,es13.5,4x,es13.5)")&
    !         i,CS(i,:)*AU2MB,sum(CS(i,:))*AU2MB/THREE,AU2EV*(omega+F%CE(i))
    !      call add_Process( P,S, "photoionization", sum(CS(i,:))/THREE, omega+F%CE(i), i )
    !   end do
    ! write(P%OUTPUT,"('# --------------------------')")
    ! write(P%OUTPUT,"('# ',a4,3(2x,es13.5),4x,es13.5)")&
    !      'sum:',sum(CS(:,1))*AU2MB,sum(CS(:,2))*AU2MB,sum(CS(:,3))*AU2MB,sum(CS(:,:))*AU2MB/THREE
  end subroutine calculate_photoIonization_CS


  subroutine calculate_photoIonization_CS_spin(P, A, B, F, S, fpath)
    !! calculates photoionization cross sections based on orbital energies
    !! resolving individual spin configurations
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S
    character*(*), optional, intent(in) :: fpath(P%N_nuc)  !! Mar.19.2015: optional parameter that specifies tdipole-filepath to be read in
    real(long) :: omega
    real(long) :: E
    ! the total  cross section (cartesian: x,y,z)
    real(long), dimension(F%N_basis, 3) :: CS
    !real(long) :: AU2MB = AU2M**2 / 1.0e-22_long

    real(long), allocatable :: TotalCS(:) ! total cs
    real(long):: AvgTotalCS ! total cs
    real(long) :: ChannelCS ! CS per hole and electron
    real(long) :: HighSpinCS ! cs for high spin

    integer, allocatable :: final_occ(:) ! final occupation
    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer, allocatable :: final_csfs(:, :) ! final spin configurations
    integer, allocatable :: final_dets(:, :) ! final determinants
    integer, allocatable :: initial_csfs(:, :) ! initial configurations
    integer, allocatable :: initial_dets(:, :) ! initial determinants
    integer :: initial_multp ! initial multp
    integer :: Nic ! counter for initial configurations
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: final_multp ! multipicity in final configuration
    integer :: final_two_M, final_two_S, initial_two_S, initial_two_M ! final and intial spin quantum numbers
    integer :: continuum_spin ! quantum number of continuum spin
    integer :: only_left(1), only_left_spin(1) ! array for orbital indices for determinant comparison
    integer :: only_right(1), only_right_spin(1) ! array for spin indices for determinant comparison
    integer :: coef_sign ! sign for determinant comparisno
    real(long) :: coef ! coeficient for CSF expansion
    real(long) :: factor(2) ! factor from coefficients
    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: io ! index for initial occupation
    integer :: i, j, part

    integer, parameter :: MAX_SPIN = 20
    real(long) :: confAvgTotalCS(MAX_SPIN)
    integer :: confpermultp(MAX_SPIN)
    real(long) :: WE_factor
    real(long), allocatable :: csic2fc(:, :) ! cross sections between initial conf to final conf
    real(long), allocatable :: coef_i(:, :) ! coeficient for CSF expansion
    real(long), allocatable :: coef_f(:, :) ! coeficient for CSF expansion

    ! photon energy
    omega = P%PE
    CS = photoIonization_CS(P, A, B, F, omega, fpath)

    if (P%print_level .gt. 1) then
      write (P%OUTPUT, "('# Photoion cross sections for omega=',e13.5,' eV in a.u.')") &
        omega*AU2EV
      write (P%OUTPUT, "('# ',a4,3(2x,a13),4x,a13,4x,a13)") &
        'MO', 'x-pol', 'y-pol', 'z-pol', 'averaged', 'pe energy/eV'
    end if

    allocate (final_occ(size(F%N_occ)))
    ! in case of MO occupation averaging we render all occupation patterns; otherwise just take F%N_occ
    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if

    Confpermultp(:) = 0
    do io = 1, size(initial_occ, 2)
      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        Confpermultp(initial_multp) = Confpermultp(initial_multp) + size(initial_csfs, 2)
        deallocate (initial_csfs)
      end do
    end do

    Nic = 0
    AvgTotalCS = ZERO
    ConfAvgTotalCS(:) = ZERO
    do io = 1, size(initial_occ, 2)
      if (P%print_level .gt. 1 ) &
        write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)
      initial_Nopen = 0
      do i = 1, size(initial_occ(:, io))
        if (initial_occ(i, io) .eq. 1) initial_Nopen = initial_Nopen + 1
      end do
      ! loop over all possible initial state spin multiplicities
      do initial_multp = initial_Nopen + 1, 1, -2
        if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
        ! prepare the representations of csf and det according to genealogical coupling scheme
        ! for the initial state configuration(s)
        call Nopen2csfs(initial_Nopen, initial_multp, initial_csfs)
        call Nopen2dets(initial_Nopen, initial_multp, initial_dets)
        allocate (coef_i(size(initial_dets, 2), size(initial_csfs, 2)))
        if (P%print_level .gt. 1) then
          write (P%OUTPUT, "('# Initial Spin multiplicity ',i3,' has ',i3, ' spin configurations:')") initial_multp, size(initial_csfs, 2)
          do j = 1, size(initial_csfs, 2)
            write (P%OUTPUT, "('# ',a )") csfstring(initial_occ(:, io), initial_csfs(:, j))
          end do
        end if
        ! Loop over initial configurations
        do ic = 1, size(initial_csfs, 2)
          if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
            if (ic .ne. F%csf_num) cycle
          end if
          do id = 1, size(initial_dets, 2)
            coef_i(id, ic) = csfdet(initial_csfs(:, ic), initial_dets(:, id))
          end do
        end do

        ! if( P%print_level> 1 ) &
        !   write(P%OUTPUT,"('# initial configuration with multiplicity',i3,2x,':',a)") initial_multp,trim(csfstring(initial_occ(:,io),initial_csfs(:,ic)))
        allocate (TotalCS(size(initial_csfs, 2)))
        TotalCS(:) = ZERO
        do part = 1, F%N_basis
          if (initial_occ(part, io) .eq. 0) cycle
          E = omega + F%CE(part)
          if (E .lt. 0) cycle
          ! Final occupation
          final_occ = initial_occ(:, io)
          final_occ(part) = final_occ(part) - 1
          final_Nopen = 0
          do i = 1, size(final_occ, 1)
            if (final_occ(i) .eq. 1) final_Nopen = final_Nopen + 1
          end do
          do final_multp = initial_multp - 1, initial_multp + 1, 2
            if (final_multp .eq. 0) cycle
            if (final_multp - 1 .gt. final_Nopen) cycle

            ! We calculate transition rate to high spin state
            ! Apply Wigner Eckert to get factor for transition rates to other MS configurations:
            initial_two_S = initial_multp - 1
            initial_two_M = initial_two_S
            final_two_S = final_multp - 1
            WE_factor = ZERO
            do final_two_M = -final_two_S, final_two_S, 2
              WE_factor = WE_factor + &
                cleb(final_two_S, final_two_M, 1, initial_two_S - final_two_M, initial_two_S, initial_two_S)**2 &
                /cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2
            end do

            call Nopen2csfs(final_Nopen, final_multp, final_csfs)
            call Nopen2dets(final_Nopen, final_multp, final_dets)
            allocate (coef_f(size(final_dets, 2), size(final_csfs, 2)))
            do fc = 1, size(final_csfs, 2)
              do fd = 1, size(final_dets, 2)
                coef_f(fd, fc) = csfdet(final_csfs(:, fc), final_dets(:, fd))
              end do
            end do

            allocate (csic2fc(size(initial_csfs, 2), size(final_csfs, 2)))
            csic2fc(:, :) = ZERO

            do ic = 1, size(initial_csfs, 2)
              if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                if (ic .ne. F%csf_num) cycle
              end if
              ! Loop over configurations within spin multiplicity and occupation
              do fc = 1, size(final_csfs, 2)
                if (P%print_level .gt. 1) &
                  write(P%OUTPUT,"('# ',2x,'-> final configuration ',a,1x, 'multiplicity ',i3)") trim(csfstring(final_occ,final_csfs(:,fc))),final_multp
                factor(:) = ZERO
                ! Loop over determinant pairs
                do id = 1, size(initial_dets, 2)
                  do fd = 1, size(final_dets, 2)
                    coef = coef_f(fd, fc)*coef_i(id, ic)
                    if (coef .eq. ZERO) cycle
                    !                    write(P%OUTPUT,*) '# ',CSFDET(initial_csfs(:,ic),initial_dets(:,id)),trim(detstring(initial_occ(:,io),initial_dets(:,id))),'->',&
                    !                      CSFDET(final_csfs(:,fc),final_dets(:,fd)),trim(detstring(final_occ(:),final_dets(:,fd)))
                    call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
                      initial_occ(:, io), initial_dets(:, id), final_occ, final_dets(:, fd))
                    coef = coef*coef_sign
                    if (coef_sign .ne. 0 .and. count(only_left(:) .ne. 0) .eq. 1 .and. count(only_right(:) .ne. 0) .eq. 0) then
                      continuum_spin = only_left_spin(1)
                      !                      write(P%OUTPUT,"('# ',5x,es15.5,x,i3,'-> k')") coef,only_left(1)
                      if (continuum_spin .eq. -1) then
                        factor(1) = factor(1) + coef
                      else
                        factor(2) = factor(2) + coef
                      end if
                    end if
                  end do
                end do
                factor(:) = factor(:)**2
                HighSpinCS = sum(factor)*sum(CS(part, :))/THREE
                ChannelCS = WE_factor*HighSpinCS

                csic2fc(ic, fc) = ChannelCS
                if (P%print_level .gt. 1 .and. ChannelCS .gt. ZERO) &
                  write (P%OUTPUT, "('# ',5x,'channel cs' ,es15.5,' pe energy',f15.3)") ChannelCS, E*AU2EV
              end do ! final configuration
            end do! ic

            if (P%yes_average_rates .or. (P%yes_average_csf_rates .and. F%multp .eq. initial_multp)) then
              call add_Process(P, F, S, "photoionization", sum(csic2fc(:, :))/size(csic2fc, 1), omega + F%CE(part), part, &
                initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=ic, &
                final_multp=final_multp, final_csf_num=1)
            else
              ic = F%csf_num
              if (F%multp .eq. initial_multp) then
                do fc = 1, size(final_csfs, 2)
                  if (csic2fc(ic, fc) .eq. ZERO) cycle
                  call add_Process(P, F, S, "photoionization", csic2fc(ic, fc), omega + F%CE(part), part, &
                    initial_occ=initial_occ(:, io), initial_multp=initial_multp, initial_csf_num=ic, &
                    final_multp=final_multp, final_csf_num=fc)
                end do
              end if
            end if
            do ic = 1, size(initial_csfs, 2)
              if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
                if (ic .ne. F%csf_num) cycle
              end if
              TotalCS(ic) = TotalCS(ic) + sum(csic2fc(ic, :))
            end do
            deallocate (csic2fc)
            deallocate (final_csfs)
            deallocate (final_dets)
            deallocate (coef_f)
          end do! final multp
        end do ! part
        if (P%print_level .gt. 1) then
          do ic = 1, size(initial_csfs, 2)
            if (.not. P%yes_average_csf_rates .and. F%csf_num .gt. 0) then
              if (ic .ne. F%csf_num) cycle
            end if
            write (P%OUTPUT, "('#  Total CS for multp ',i4,' ic ',i4,' : ',es13.5,' a.u.')") initial_multp, ic, TotalCS(ic)
          end do
          if (P%yes_average_csf_rates) &
            write (P%OUTPUT, "('#  Total CS for multp ',i4,' avg_ic ',' : ',es13.5,' a.u.')") &
            initial_multp, sum(TotalCS(:))/size(TotalCS(:), 1)

        end if
        AvgTotalCS = AvgTotalCS + initial_multp*sum(TotalCS(:))
        Nic = Nic + initial_multp*size(TotalCS, 1)
        deallocate (initial_dets)
        deallocate (initial_csfs)
        deallocate (coef_i)
        deallocate (TotalCS)
      end do ! initial multiplicity
    end do ! initial occupations
    AvgTotalCS = AvgTotalCS/Nic
    if ((P%yes_average_csf_rates .or. P%yes_average_rates) .and. P%print_level .gt. 1) then
      write (P%OUTPUT, "('# Average Total CS ',es13.5,' a.u. ')") AvgTotalCS
      write (P%OUTPUT, "('# averaged over ',i3,' spin configurations' )") Nic
    end if
    deallocate (final_occ)
    deallocate (initial_occ)
  end subroutine calculate_photoIonization_CS_spin

!>
!! function photoIonization_CS(P,A,B,F,omega)
!!
!! computes the ionization cross sections according to
!! independent atomic center approximation
!! using the transition dipoles read in from xatom
!! Arguments:
!!  omega = photon energy in a.u.
!! Return value:
!!  CS (mo,d): 2 dimensional array with
!!  cross section per MO and per polarization d=x,y,z
  function photoIonization_CS(P, A, B, F, omega, fpath) result(CS)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(in) :: omega ! photon energy

    real(long), dimension(F%N_basis, 3) :: CS
    ! the mo -> continuum  cross section (cartesian: x,y,z)

    real(long) :: E ! photoelectron energy
    character(len=512) :: tdipole_file_nuc ! filename
    integer :: atom1, atom2 ! counter for atoms
    integer :: mo ! counter for mos
    type(TransitionDipoleMoments), dimension(P%N_nuc) :: DI
    ! TransitionDipoleMoments for every atom to be read in from xatom output

    integer :: lc, mc, lmax ! continuum lc, mc and maximum l
    real(long) :: alpha = ONE/SPEED_C_AU ! fine structure constant

    character*(*), optional, intent(in) :: fpath(P%N_nuc)  !! Mar.19.2015: optional parameter that specifies tdipole-filepath to be read in
    integer :: j
    integer :: bf, bf2 ! bf index
    real(long), dimension(3) :: tdipole, tdipole2
    ! the bf -> continuum trans dipole matrix element (cartesian: x,y,z)
    logical :: file_exist
    real(long):: Emax, Emin


    ! read in all transition dipole matrix elements for each atom
    ! and determine the maximum continuum angular quantum number lmax
    lmax = 0
    Emin = HUGE(Emin)
    Emax = ZERO
    do atom1 = 1, P%N_nuc
      ! special case for hydrogen: for hydrogen no photoionization cross section calculation
      !if(A(atom)%charge==1)cycle

      if (present(fpath)) then
        tdipole_file_nuc = trim(fpath(atom1))   !! Mar.19.2015 optional file path specification
      else
        tdipole_file_nuc = A(atom1)%tdipole_file_nuc
      end if

      !     dipole_file_nuc = trim( ATOMIC_NAME(P%charge_nuc(atom)) ) // ".tdipole"
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# Reading tdipole integrals from file ',a)") trim(tdipole_file_nuc)
      end if
      inquire (file=A(atom1)%tdipole_file_nuc, EXIST=file_exist)
      if (.NOT. file_exist) then ! Warnings for H atoms are ignored
        if (A(atom1)%charge .ne. 1) &
          write (P%OUTPUT, "('# Warning file ',a,' not found! Skipping contributions for this atom')") trim(A(atom1)%tdipole_file_nuc)
      else
        call readDIntegral(P, tdipole_file_nuc, DI(atom1))
        ! prepare the projection of MO's into minimal basis set
        if (allocated(DI(atom1)%Pmin)) then
          allocate (DI(atom1)%Smin(size(DI(atom1)%Pmin, 1), size(DI(atom1)%Pmin, 2)))
          forall (j=1:size(DI(atom1)%Pmin, 1)) DI(atom1)%Smin(j, j) = ONE
          allocate (DI(atom1)%Omin(size(DI(atom1)%Pmin, 1), size(DI(atom1)%Pmin, 2)))
          DI(atom1)%Omin = matmul(DI(atom1)%Pmin, F%LC)
        else if (P%yes_gaussian) then
          allocate (DI(atom1)%Omin(size(F%Omin, 1), size(F%Omin, 2)))
          allocate (DI(atom1)%Smin(size(F%Smin, 1), size(F%Smin, 2)))
          allocate (DI(atom1)%min_info(size(F%min_info, 1), size(F%min_info, 2)))
          DI(atom1)%Omin(:, :) = F%Omin(:, :)
          DI(atom1)%Smin = F%Smin
          DI(atom1)%min_info(:, :) = F%min_info(:, :)
        end if! if Pmin allocated
        if (P%yes_verbose .and. allocated(DI(atom1)%min_info)) then
          call print_projection_info(DI(atom1)%min_info, DI(atom1)%Omin, DI(atom1)%Smin, atom1, P%OUTPUT)
        end if! yes_verbose

        lmax = max(maxval(A(atom1)%l_AO) + 1, lmax)
        Emax = max(maxval(DI(atom1)%energy), Emax)
        Emin = min(minval(DI(atom1)%energy), Emin)
      end if
    end do! atom

    CS(:, :) = ZERO

    do mo = 1, F%N_basis
      ! calculate the gross population for each basis function from this mo

      E = omega + F%CE(mo)  ! E is the energy of the photoelectron
      if (E .lt. 0) cycle ! orbital cannot be ionized
      ! In case of yes_process_spin and yes_all_occs we need all transition dipole moments
      if (.NOT. is_occupied(P, F, F%N_occ, mo) .and. .not. (P%yes_process_spin .and. P%yes_all_occs)) cycle
      if (E .gt. Emax .or. E .lt. Emin) then
        write (P%OUTPUT, '("#WARNING: Photo electron energy for MO ",'// &
          'i3,es12.3," eV is too ")', advance="no") mo, E*AU2EV
        if (E .gt. Emax) then
          write (P%OUTPUT, '("large")', advance="no")
        else if (E .lt. Emin) then
          write (P%OUTPUT, '("small")', advance="no")
        end if
        write (P%OUTPUT, '(" for the calculated transition dipole moments. '// &
          'Extrapolating.")')
      end if

      do lc = 0, lmax ! loop over all l,m continua
        do mc = -lc, lc
          if (allocated(DI(1)%Omin)) then
            do bf = 1, size(DI(1)%Omin, 1)
              atom1 = DI(1)%min_info(bf, 1)
              if (.not. allocated(DI(atom1)%config)) cycle

              tdipole = trans_dipole_bf_cont(DI(atom1)%min_info(bf, :), DI, E, lc, mc)
              CS(mo, :) = CS(mo, :) + DI(atom1)%Omin(bf, mo)**2*tdipole**2
              if (DI(atom1)%Omin(bf, mo)**2 .gt. 0.1 .and. sum(tdipole**2) .gt. ZERO .and. P%yes_verbose) &
                write(P%OUTPUT,'("# mo",i4,x,"atom",i4,x,"bf",i4,x,f15.5,x,"lc",i4,x,"mc",i4,x,es20.7)') mo,atom1,bf, DI(atom1)%Omin(bf,mo)**2, lc, mc, TWO/THREE * FOUR*PI**2*alpha*omega*sum(tdipole**2)
              do bf2 = bf + 1, size(DI(atom1)%Omin, 1)
                atom2 = DI(1)%min_info(bf2, 1)
                if (atom2 .ne. atom1) cycle ! only cross terms on the same atom
                if (.not. allocated(DI(atom2)%config)) cycle
                tdipole2 = trans_dipole_bf_cont(DI(atom1)%min_info(bf2, :), DI, E, lc, mc)
                CS(mo, :) = CS(mo, :) + TWO*DI(atom1)%Omin(bf, mo)*DI(atom1)%Omin(bf2, mo)*tdipole*tdipole2
              end do !bf 2
            end do! bf
          else
            ! variant 1: for each bf incoherent summation
            do bf = 1, P%N_basis
              atom1 = B%info(bf, 1)
              if (.not. allocated(DI(atom1)%config)) cycle
              tdipole = trans_dipole_bf_cont(B%info(bf, :), DI, E, lc, mc)
              ! This is Gelius original approach (Gelius & Siegbahn, Faraday Discuss 54 257):
              ! use Gross Population for weighing atomic cross sections
              !             CS(mo,:)=CS(mo,:)+gross_pop(bf) * tdipole**2
              ! This is the formula derived by Huang et al using net atomic populations
              ! [C^2 instead of Gross pop]
              ! see Huang & Ellision, J. Electron Spec. and Related Phenomena 4 (1974) 233
              !              CS(mo,:)=CS(mo,:)+ F%LC(bf,mo)**2 *F%N_occ(mo)* tdipole**2
              CS(mo, :) = CS(mo, :) + F%LC(bf, mo)**2*tdipole**2
              ! cross terms:
              ! at the moment, we consider only cross terms on the same atom
              ! the phases of the cross terms are already in the transition dipoles
              do bf2 = bf + 1, P%N_basis
                if (B%info(bf, 1) .ne. B%info(bf2, 1)) cycle ! only cross terms on the same atom
                tdipole2 = trans_dipole_bf_cont(B%info(bf2, :), DI, E, lc, mc)
                !                 CS(mo,:)=CS(mo,:) + F%N_occ(mo)*TWO* F%LC(bf,mo)*F%LC(bf2,mo) * tdipole * tdipole2
                CS(mo, :) = CS(mo, :) + TWO*F%LC(bf, mo)*F%LC(bf2, mo)*tdipole*tdipole2
              end do !bf 2
            end do ! bf
          end if
        end do ! mc
      end do ! lc
      CS(mo, :) = FOUR*PI**2*alpha*omega*CS(mo, :)
    end do! mo loop

    do atom1 = 1, P%N_nuc
      if (allocated(DI(atom1)%config)) &
        call purgeDIntegral(DI(atom1))
    end do
  end function photoIonization_CS

  subroutine calculate_photoIonization_CS_CI(P, A, B, F, G, fpath)
    !! computes photoionizatio cross sections employing configuration interaction calculations
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    type(Grid3D), intent(in) :: G
    character*(*), optional, intent(in) :: fpath(P%N_nuc)  !! Mar.19.2015: optiona

    integer :: i, j
    real(long) :: PE_energy ! Photo electron energy
    real(long) :: omega
    real(long) :: E_ini ! initial configuration energy
    real(long) :: ChannelCS ! CS per hole and electron
    real(long) :: HighSpinCS ! cs for high spin

    integer, allocatable, target :: final_occ(:, :) ! final occupation
    integer, allocatable, target :: initial_occ(:, :) ! initial occupation

    integer :: initial_multp ! initial spin multp
    integer :: final_multp ! multipicity in final configuration
    integer :: final_two_M, final_two_S, initial_two_S, initial_two_M ! final and intial spin quantum numbers
    integer :: continuum_spin ! quantum number of continuum spin

    integer :: ic, id, fc, fd ! index for initial/final spin configuration/deterimant
    integer :: io ! index for initial occupation
    integer :: io2
    integer :: N_initial_csfs ! N of initial csfs
    integer :: N_final_occ ! N of final occupation
    integer :: N_final_csfs ! N of final csfs
    integer :: i_final_occ ! index for final occupation
    integer :: j_final_occ ! index for final occupation
    integer :: hole, ih, ih2, virt
    integer :: irrep_occ, f_irrep, n_irrep, initial_irrep
    integer, allocatable :: occs_temp(:, :) ! temp occupation
    real(long), dimension(F%N_basis, 3) :: CS
    type(CSF), allocatable :: initial_csf_array(:) ! array for csfs initial state
    type(CSF), allocatable :: final_csf_array(:) ! array for csfs final state
    real(long), allocatable :: ionize_coef(:, :)
    real(long), allocatable :: final_csf_ionize_coef(:, :, :)
    real(long), allocatable :: final_csf_ionize_coef2(:, :, :)
    real(long), allocatable :: final_state_ionize_coef(:, :, :)
    real(long), allocatable :: final_CI_matrix(:, :) ! CI matrix for final states
    real(long), allocatable :: final_CI_energy(:) ! energy of final states
    real(long), allocatable :: initial_CI_matrix(:, :) ! CI matrix for initial states
    real(long), allocatable :: initial_CI_energy(:) ! energy of initial states
    logical :: redundant_occ ! number of redundant occupations

    integer :: only_left(1), only_left_spin(1) ! array for orbital indices for determinant comparison
    integer :: only_right(1), only_right_spin(1) ! array for spin indices for determinant comparison
    integer, allocatable :: final_hole(:)

    real(long) :: coef
    integer :: coef_sign
    real(long) :: WE_coef

    type(WF_data) :: F_final
    type(Basis) :: B_final
    real(long), allocatable :: csf_O(:, :) ! Auger amplitude for each csf and continuum channel
    integer, allocatable :: active(:)
    integer, allocatable :: core(:)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    integer :: NOpenMax
    type(UmatrixTable) :: UTini, UTfin
    integer :: nintern

    if (allocated(P%final_hole)) then
      allocate (final_hole(size(P%final_hole)))
      final_hole(:) = P%final_hole(:)
    else
      allocate (final_hole(F%Ncore))
      final_hole = (/(i, i=1, F%Ncore)/)
    end if

    if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, initial_occ)
    else if (P%yes_cis) then
      allocate (initial_occ(size(F%N_occ), size(F%CIS_excitation, 2) + 1))
      initial_occ(:, 1) = F%N_occ(:)
      do i = 1, size(initial_occ, 2) - 1
        initial_occ(:, i + 1) = F%N_occ(:)
        initial_occ(F%CIS_excitation(1, i), i + 1) = initial_occ(F%CIS_excitation(1, i), i + 1) - 1
        initial_occ(F%CIS_excitation(2, i), i + 1) = initial_occ(F%CIS_excitation(2, i), i + 1) + 1
      end do
    else if (.NOT. allocated(P%active_space_MO)) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      do i = 1, size(final_hole)
        if (any(final_hole(i) .eq. P%active_space_MO)) then
          stop 'final_hole must not be active space!'
        end if
      end do
      call active_space_occs(F%N_occ, P%active_space_MO, initial_occ)
    end if

    if (P%yes_sym .and. F%state_irrep .gt. 0) then
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, initial_occ, occs_temp)
      deallocate (initial_occ)
      allocate (initial_occ(size(occs_temp, 1), size(occs_temp, 2)))
      initial_occ = occs_temp
      deallocate (occs_temp)
    end if

    allocate(active( count((/ (  ( any(initial_occ(j,:)<2) .and. any(initial_occ(j,:)>0)   ) .or. any(final_hole==j) , j=1,size(initial_occ,1))  /)) ))

    active = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(((any(initial_occ(j, :) .lt. 2) .and. any(initial_occ(j, :) .gt. 0))) .or. any(final_hole .eq. j), j=1, size(initial_occ, 1))/))
    allocate (core(count((/((all(initial_occ(j, :) .eq. 2)) .and. .not. any(final_hole .eq. j), j=1, size(initial_occ, 1))/))))
    core = pack((/(j, j=1, size(initial_occ, 1))/), &
      (/(all(initial_occ(j, :) .eq. 2) .and. .not. any(final_hole .eq. j), j=1, size(initial_occ, 1))/))

    nintern = size(active)
    do j = size(active), 1, -1
      if (any(sum(initial_occ(active(j:), :), dim=1) .gt. 1)) exit
    end do
    nintern = j
    write (P%OUTPUT, '("#     core:",99i4)') core
    write (P%OUTPUT, '("#   active:",99i4)') active
    write (P%OUTPUT, '("# internal:",99i4)') active(:nintern)
    write (P%OUTPUT, '("# external:",99i4)') active(nintern + 1:)

    ! photon energy
    omega = P%PE

    CS = photoIonization_CS(P, A, B, F, omega, fpath)

    ! loop over initial occupation
    N_initial_csfs = 0
    initial_multp = P%multp
    initial_irrep = P%state_irrep
    write (P%OUTPUT, "('# Initial configurations for spin multp: ',i4)") initial_multp

    call construct_csf_array(P, B, F, initial_multp, initial_irrep, initial_occ, initial_csf_array)

    allocate (F_core(B%N_basis, B%N_basis))
    call calculate_F_inactive(P, B, core, F%LC, F_core)
    if (.not. allocated(B%H0)) then
      F%v_ext(:) = potential_ext(G, A)
      call construct_h0_matrix(G, B, F)
    end if
    F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)

    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do

    N_initial_csfs = size(initial_csf_array, 1)
    write (P%OUTPUT, "('# N_initial_csfs: ',i7)") N_initial_csfs
    allocate (initial_CI_matrix(N_initial_csfs, N_initial_csfs))
    allocate (initial_CI_energy(N_initial_csfs))

    write (P%OUTPUT, '("# computing CI matrix")')
    NOpenMax = maxval(count(initial_occ .eq. 1, dim=1))
    call constructUMatrixTable(NOpenMax, initial_multp, UTini)
    call transform_twoel_allactive(P, B, F, active(:nintern), P%OUTPUT)
    if (nintern .lt. size(active)) then
      call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        initial_occ(active(:), :), &
        initial_CI_matrix, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)
      deallocate (F%teint_cdij)
      deallocate (F%teint_cidj)
      deallocate (F%teint_cdei)
    else
      call calc_CIMatrix(UTini, E_core, F_core(active(:), active(:)), F%teint_abcd, initial_occ(active(:), :), initial_CI_matrix)
    end if

    write (P%OUTPUT, '("# done")')
    ! solve it
    write (P%OUTPUT, '("# solving CI matrix")')
    call get_eigenvalues(initial_CI_matrix, initial_CI_energy)
    write (P%OUTPUT, '("# done")')

    E_ini = initial_CI_energy(P%state)
    write(P%OUTPUT,"('# Initial state ',i4,x,'spin multp=',i4,x,'energy=',f15.5)") P%state,P%multp,E_ini+nuclear_repulsion_energy( A, P )
    if (P%yes_sym) then
      write (P%OUTPUT, "('# Initial state irrep ',i4)") P%state_irrep
    end if
    do ic = 1, size(initial_CI_matrix, 1)
      if (abs(initial_CI_matrix(ic, P%state)) .lt. 0.1) cycle
      write(P%OUTPUT,'("# csf ",i7,x,f15.5,x,a)') ic,initial_CI_matrix(ic,P%state),trim(csfstring(initial_csf_array(ic)%occ,initial_csf_array(ic)%csf))
    end do

    ! Loop over final state spin
    do final_multp = initial_multp - 1, initial_multp + 1, 2
      if (final_multp .eq. 0) cycle
      write (P%OUTPUT, "('# Final Spin multiplicity ',i3)") final_multp
      N_final_occ = 0
      ! Loop over hole and valence pair combinations: count configurations
      !do ih=1,size(final_hole)
      !hole=final_hole(ih)
      !write(P%OUTPUT,"('# Core hole in orbital ',i3)") hole
      if (P%yes_sym) then
        n_irrep = B%nirrep
      else
        n_irrep = 1
      end if
      do f_irrep = 1, n_irrep
        write (P%OUTPUT, "('# Final state irrep ',i3)") f_irrep
        if (P%yes_process_CIS) then
          ! count possible configurations TODO make a better estimat including symmetry
          i_final_occ = 0
          do io2 = 1, size(initial_occ, 2)
            do ih = 1, size(final_hole)
              hole = final_hole(ih)
              if (initial_occ(hole, io2) .eq. 0) cycle
              i_final_occ = i_final_occ + 1
              do ih2 = F%Ncore + 1, size(initial_occ, 1)
                if (initial_occ(ih2, io2) .eq. 0) cycle
                do virt = ih2 + 1, size(initial_occ, 1)
                  if (initial_occ(virt, io2) .eq. 2) cycle
                  i_final_occ = i_final_occ + 1
                end do!virt
              end do! ih2
            end do
          end do
          allocate (final_occ(size(F%N_occ), i_final_occ))
        else
          allocate (final_occ(size(F%N_occ), size(initial_occ, 2)*size(final_hole, 1)))
        end if
        i_final_occ = 0
        do io2 = 1, size(initial_occ, 2)
          do ih = 1, size(final_hole)
            hole = final_hole(ih)
            if (initial_occ(hole, io2) .eq. 0) cycle
            i_final_occ = i_final_occ + 1
            final_occ(:, i_final_occ) = initial_occ(:, io2)
            final_occ(hole, i_final_occ) = final_occ(hole, i_final_occ) - 1
            j_final_occ = i_final_occ
            if (P%yes_process_cis) then
              do ih2 = F%Ncore + 1, size(initial_occ, 1)
                if (final_occ(ih2, j_final_occ) .eq. 0) cycle
                do virt = ih2 + 1, size(initial_occ, 1)
                  if (final_occ(virt, j_final_occ) .eq. 2) cycle
                  i_final_occ = i_final_occ + 1
                  final_occ(:, i_final_occ) = final_occ(:, j_final_occ)
                  final_occ(ih2, i_final_occ) = final_occ(ih2, i_final_occ) - 1
                  final_occ(virt, i_final_occ) = final_occ(virt, i_final_occ) + 1
                  if (P%yes_sym) then
                    irrep_occ = 1
                    do i = 1, size(final_occ(:, i_final_occ))
                      if (final_occ(i, i_final_occ) .eq. 1) &
                        irrep_occ = B%irrep_prod(irrep_occ, F%irrep_MO(i))
                    end do
                    if (irrep_occ .ne. f_irrep) then
                      i_final_occ = i_final_occ - 1
                      cycle
                    end if
                  end if

                  redundant_occ = .FALSE.
                  do i = 1, i_final_occ - 1
                    if (all(final_occ(:, i) .eq. final_occ(:, i_final_occ))) redundant_occ = .TRUE.
                    if (redundant_occ) exit
                  end do
                  if (redundant_occ) then
                    i_final_occ = i_final_occ - 1
                    cycle
                  end if
                end do ! virt
              end do! hole 2
            end if
          end do! hole
        end do! io2

        if (P%yes_sym) then
          call filter_sym_occs(B%irrep_prod, F%irrep_MO, f_irrep, final_occ(:, 1:i_final_occ), occs_temp)
        else
          allocate (occs_temp(size(final_occ, 1), i_final_occ))
          occs_temp = final_occ(:, 1:i_final_occ)
        end if
        deallocate (final_occ)
        allocate (final_occ(size(occs_temp, 1), size(occs_temp, 2)))
        final_occ = occs_temp
        deallocate (occs_temp)

        N_final_occ = size(final_occ, 2)

        call construct_csf_array(P, B, F, final_multp, f_irrep, final_occ(:, 1:N_final_occ), final_csf_array)

        N_final_csfs = size(final_csf_array, 1)

        write (P%OUTPUT, "('# N_final_occ ',i5)") N_final_occ
        write (P%OUTPUT, "('# N_final_csfs ',i7)") N_final_csfs
        ! if there are no spin configurations for this spin continue with next spin multiplicity
        if (N_final_csfs .eq. 0) then
          deallocate (final_occ)
          cycle
        end if

        allocate (final_CI_matrix(N_final_csfs, N_final_csfs))
        allocate (final_CI_energy(N_final_csfs))
        allocate (final_csf_ionize_coef(N_final_csfs, size(final_hole, 1), 2))

        ! Apply Wigner Eckert to get the transition rates to other MS configurations:
        initial_two_S = initial_multp - 1
        initial_two_M = initial_two_S
        final_two_S = final_multp - 1
        ChannelCS = ZERO
        WE_coef = ZERO
        do final_two_M = -final_two_S, final_two_S, 2
          WE_coef = WE_coef + cleb(final_two_S, final_two_M, 1, initial_two_M - final_two_M, initial_two_S, initial_two_S)**2 &
            /cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2
        end do
        write (P%OUTPUT, *) '# Wigner Eckard factor', WE_coef
        allocate (ionize_coef(size(final_hole, 1), 2))
        do fc = 1, size(final_csf_array, 1)
          final_csf_ionize_coef(fc, :, :) = ZERO
          do ic = 1, size(initial_csf_array, 1)
            ionize_coef(:, :) = ZERO
            if (initial_CI_matrix(ic, P%state) .eq. ZERO) cycle
            ! Loop over determinant pairs
            do id = 1, size(initial_csf_array(ic)%dets, 2)
              do fd = 1, size(final_csf_array(fc)%dets, 2)
                !coef=initial_CI_matrix(ic,P%state)*&
                coef = &
                  csfdet(final_csf_array(fc)%csf, final_csf_array(fc)%dets(:, fd))* &
                  csfdet(initial_csf_array(ic)%csf, initial_csf_array(ic)%dets(:, id))
                if (coef .eq. ZERO) cycle
                call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
                  initial_csf_array(ic)%occ, initial_csf_array(ic)%dets(:, id), &
                  final_csf_array(fc)%occ, final_csf_array(fc)%dets(:, fd))
                coef = coef*coef_sign
                if (coef_sign .ne. 0 .and. count(only_left(:) .ne. 0) .eq. 1 .and. count(only_right(:) .ne. 0) .eq. 0) then
                  continuum_spin = only_left_spin(1)
                  ih = findloc(final_hole, only_left(1), dim=1)
                  if (ih .gt. 0) then
                    if (continuum_spin .eq. -1) then
                      ionize_coef(ih, 1) = ionize_coef(ih, 1) + coef
                    else
                      ionize_coef(ih, 2) = ionize_coef(ih, 2) + coef
                    end if
                  end if
                end if! if coef_sign/=0 ...
              end do! fd
            end do! id
            !            write(P%OUTPUT,'("# initial csf ",i7,x,f15.5,x,a)') ic,initial_CI_matrix(ic,P%state),trim(csfstring(initial_csf_array(ic)%occ,initial_csf_array(ic)%csf))
            !            write(P%OUTPUT,'("# final csf ",i7,x,f15.5,x,a)') fc,trim(csfstring(final_csf_array(fc)%occ,final_csf_array(fc)%csf))
            !            write(P%OUTPUT,'("#  ",2f15.5)') ionize_coef(1,:)

            final_csf_ionize_coef(fc, :, :) = final_csf_ionize_coef(fc, :, :) + ionize_coef(:, :)*initial_CI_matrix(ic, P%state)
          end do! ic
        end do!fc

        do ih = 1, size(final_hole, 1)
          write(P%OUTPUT,'("# Sum for MO (csfs)",i4,": to fcsf ",x,2f15.5)') final_hole(ih),WE_coef*sum(final_csf_ionize_coef(:,ih,1)**2),WE_coef*sum(final_csf_ionize_coef(:,ih,2)**2)
        end do

        if (P%yes_process_CI_relaxed) then
          F_final = F
          B_final = B
          call read_wf(F, F_final%LC, "wf_relaxed.dat")
          allocate (csf_O(size(final_csf_array), size(final_csf_array)))
          call overlap_csfs(P, final_csf_array, F%LC, F_final%LC, B%S, csf_O)
          allocate (final_csf_ionize_coef2(N_final_csfs, size(final_hole, 1), 2))
          do j = 1, N_final_csfs
            do io = 1, size(final_hole, 1)
              final_csf_ionize_coef2(j, io, 1) = sum(final_csf_ionize_coef(:, io, 1)*csf_O(:, j))
              final_csf_ionize_coef2(j, io, 2) = sum(final_csf_ionize_coef(:, io, 2)*csf_O(:, j))
            end do
          end do
          write (6, *) 'Overlap to new basis:'
          do fc = 1, size(final_csf_array, 1)
            write(P%OUTPUT,'("# csf ",i7,x,a,x,"ionize_coef=",es20.7,x,f7.3,f7.3)') fc,trim(csfstring(final_csf_array(fc)%occ,final_csf_array(fc)%csf)),&
              sum(final_csf_ionize_coef(fc, :, :)**2), &
              sum(csf_O(:, fc)**2), sum(csf_O(fc, :)**2)
          end do
          do ih = 1, size(final_hole)
            write(P%OUTPUT,'("# Sum for MO (basis transform)",i4,": to fcsf ",x,2f15.5)') final_hole(ih),WE_coef*sum(final_csf_ionize_coef2(:,ih,1)**2),WE_coef*sum(final_csf_ionize_coef2(:,ih,2)**2)
          end do
          final_csf_ionize_coef = final_csf_ionize_coef2
          deallocate (final_csf_ionize_coef2)
          !call construct_h0_matrix( G, B_final, F_final )
          F_final%H0 = F_final%LC(:, :) .tm.B_final%H0(:, :) .m.F_final%LC(:, :)

          deallocate (csf_O)
        else
          F_final = F
          B_final = B
          call construct_h0_matrix(G, B, F_final)
          F_final%H0 = F_final%LC(:, :) .tm.F_final%H0(:, :) .m.F_final%LC(:, :)
        end if!

        ! Compute the final state CI matrix
        write (P%OUTPUT, '("# computing CI matrix")')
        ! compute CI matrix
        NOpenMax = maxval(count(final_occ(:, 1:N_final_occ) .eq. 1, dim=1))
        call constructUMatrixTable(NOpenMax, final_multp, UTfin)
        deallocate (active, core)
        allocate(active( count((/ (  ( any(final_occ(j,1:N_final_occ)<2) .and. any(final_occ(j,1:N_final_occ)>0)   ), j=1,size(final_occ,1))  /)) ))
        allocate (core(count((/((all(final_occ(j, 1:N_final_occ) .eq. 2)), j=1, size(final_occ, 1))/))))
        active = pack((/(j, j=1, size(final_occ, 1))/), &
          (/(((any(final_occ(j, 1:N_final_occ) .lt. 2) .and. any(final_occ(j, 1:N_final_occ) .gt. 0))), j=1, size(final_occ, 1))/))
        core = pack((/(j, j=1, size(final_occ, 1))/), &
          (/(all(final_occ(j, 1:N_final_occ) .eq. 2), j=1, size(final_occ, 1))/))
        nintern = size(active)
        do j = size(active), 1, -1
          if (any(sum(final_occ(active(j:), :N_final_occ), dim=1) .gt. 1)) exit
        end do
        nintern = j
        write (6, '("#     core:",99i4)') core
        write (6, '("#   active:",99i4)') active
        write (6, '("# internal:",99i4)') active(:nintern)
        write (6, '("# external:",99i4)') active(nintern + 1:)
        call calculate_F_inactive(P, B_final, core, F_final%LC, F_core)
        E_core = ZERO
        do i = 1, size(core)
          E_core = E_core + F_core(core(i), core(i)) + F_final%H0(core(i), core(i))
        end do

        call transform_twoel_allactive(P, B_final, F_final, active(1:nintern), P%OUTPUT)
        if (nintern .lt. size(active)) then
          call transform_twoel_two_core_two_virtual(P, B_final, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
          call transform_twoel_three_core_one_virtual(P, B_final, F_final, active(:nintern), active(nintern + 1:), P%OUTPUT)
          call calc_CIMatrix(UTfin, E_core, F_core(active(:), active(:)), &
            F_final%teint_abcd, &
            final_occ(active(:), 1:N_final_occ), &
            final_CI_matrix, nintern=nintern, &
            teint_iiee=F_final%teint_cdij, &
            teint_ieie=F_final%teint_cidj, &
            teint_iiie=F_final%teint_cdei)
          deallocate (F_final%teint_cdij)
          deallocate (F_final%teint_cidj)
          deallocate (F_final%teint_cdei)
        else
          call calc_CIMatrix(UTfin, E_core, F_core(active(:),active(:)), F_final%teint_abcd, final_occ(active(:),1:N_final_occ), final_CI_matrix)
        end if
        write (P%OUTPUT, '("# done")')

        ! solve it
        write (P%OUTPUT, '("# solving CI matrix")')
        call get_eigenvalues(final_CI_matrix, final_CI_energy)
        write (P%OUTPUT, '("# done")')
        ! transform ionization coeficients
        allocate (final_state_ionize_coef(N_final_csfs, size(final_hole, 1), 2))
        do ih = 1, size(final_hole)
          do i = 1, size(final_csf_array)
            final_state_ionize_coef(i, ih, 1) = sum(final_CI_matrix(:, i)*final_csf_ionize_coef(:, ih, 1))
            final_state_ionize_coef(i, ih, 2) = sum(final_CI_matrix(:, i)*final_csf_ionize_coef(:, ih, 2))
          end do
        end do
        deallocate (final_csf_ionize_coef)
        do ih = 1, size(final_hole, 1)
          write(P%OUTPUT,'("# Sum for MO (final states)",i4,": to fcsf ",x,2f15.5)') final_hole(ih),WE_coef*sum(final_state_ionize_coef(:,ih,1)**2),WE_coef*sum(final_state_ionize_coef(:,ih,2)**2)
        end do
        do i = 1, size(final_csf_array, 1)
          ionize_coef(:, :) = final_state_ionize_coef(i, :, :)
          PE_energy = E_ini - final_CI_energy(i) + omega
          if (PE_energy .lt. ZERO) cycle
          write(P%OUTPUT,"('# ','STATE:',x,i5,x,'energy=',x,f15.5,x,'photo electron energy=',x,f15.5,x)") i,final_CI_energy(i)+nuclear_repulsion_energy( A, P ),PE_energy*AU2EV
          do fc = 1, size(final_csf_array, 1)
            if (abs(final_CI_matrix(fc, i)) .lt. 0.1) cycle
            write(P%OUTPUT,'("# csf ",i7,x,f15.5,x,a)') fc,final_CI_matrix(fc,i),trim(csfstring(final_csf_array(fc)%occ,final_csf_array(fc)%csf))
          end do
          !      do j=1,size(ionize_coef,1)
          !        if(abs(ionize_coef(j,1))>1e-3)&
          !          write(P%OUTPUT,"('# ',27x,x,es15.5,x,'<',i5,'|r| psi_k_a >')") ionize_coef(j,1),j
          !        if(abs(ionize_coef(j,2))>1e-3)&
          !          write(P%OUTPUT,"('# ',27x,x,es15.5,x,'<',i5,'|r| psi_k_b >')") ionize_coef(j,2),j
          !      end do
          !      write(P%OUTPUT,"('# ',25x,'==============')")
          HighSpinCS = ZERO
          do j = 1, size(ionize_coef, 1)
            HighSpinCs = HighSpinCs + sum(ionize_coef(j, :)**2)*sum(CS(j, :))/THREE
            if (sum(ionize_coef(j, :)**2) .gt. 1e-6) &
              write(P%OUTPUT,"('# ',20x,f7.3,'*',es15.5,x,'<',i5,'|r| psi_k >^2',x,es15.5)") WE_coef,sum(ionize_coef(j,:)**2),final_hole(j),WE_coef*sum(ionize_coef(j,:)**2)*sum(CS(final_hole(j),:))/THREE
          end do
          ChannelCS = WE_coef*HighSpinCS
          write(P%OUTPUT,"('# PHOTOELECTRON',5x,f15.5,7x,x,es15.5,x,es15.5)") PE_energy*AU2EV,WE_coef*sum(ionize_coef(:,1)**2+ionize_coef(:,2)**2),&
            WE_coef*sum( (ionize_coef(:,1)**2+ionize_coef(:,2)**2) * (CS(final_hole(:),1)+CS(final_hole(:),2)+CS(final_hole(:),3))/THREE )
          write (P%OUTPUT, "('# ',25x,'==============')")

        end do! i
        deallocate (ionize_coef)

        call purge_csf_array(final_csf_array)
        deallocate (final_csf_array)
        deallocate (final_CI_matrix)
        deallocate (final_state_ionize_coef)
        deallocate (final_CI_energy)
        deallocate (final_occ)
        call purge_UMatrixTable(UTfin)
      end do! f_irrep
      !end do! hole
    end do! final_multp

    deallocate (final_hole)
    call purge_csf_array(initial_csf_array)
    deallocate (initial_occ)
    call purge_UMatrixTable(UTini)
    deallocate (initial_CI_matrix)
    deallocate (initial_CI_energy)

    deallocate (F_core)
    deallocate (active)
    deallocate (core)
  end subroutine calculate_photoIonization_CS_CI

  subroutine calculate_photoIonization_CS_spin_avg(P, A, B, F, S, fpath)
    !! calculates photoionization cross sections based on orbital energies
    !! resolving individual spin configurations and averaging over configurations    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S
    character*(*), optional, intent(in) :: fpath(P%N_nuc)  !! Mar.19.2015: optional parameter that specifies tdipole-filepath to be read in

    real(long), dimension(F%N_basis, 3) :: CS
    integer :: initial_multp ! initial multp
    integer :: final_multp ! multipicity in final configuration
    integer :: initial_Nopen ! number of open shells in initial state
    integer :: final_Nopen ! number of open shells in final state
    integer :: rest_Nopen ! open shells in remaining electron configuration
    integer :: parti_Nopen ! open shells in remaining electron configuration
    integer :: partf_Nopen ! open shells in remaining electron configuration
    integer :: rest_Ncsf, initial_Ncsf ! number of spin configuration state functions
    integer :: final_two_S, initial_two_S, initial_two_M ! two times final and intial spin quantum numbers
    integer :: parti_two_S ! two times spin quantum numbers of initial participator part,
    integer :: partf_two_S ! final participator part,
    real(long):: WE_factor ! factor from Wigner Eckhard theorem
    real(long):: spin_factor ! another spin factor
    integer, allocatable :: initial_occ(:, :) ! initial occupation
    integer :: io

    real(long) :: AvgTotalCS, ChannelCS, SpinCS, TotalCS, SpinavgCS ! CS
    integer, allocatable :: weight_io(:)
    integer :: Spinavg_n, Total_n, AvgTotal_n
    integer :: mo
    real(long) :: omega

    ! photon energy
    omega = P%PE
    CS = photoIonization_CS(P, A, B, F, omega, fpath)

    if (P%print_level .gt. 1) then
      write (P%OUTPUT, "('# Photoion cross sections for omega=',e13.5,' eV in a.u.')") &
        omega*AU2EV
      write (P%OUTPUT, "('# ',a4,(2x,a13),4x,a13,4x,a13)") &
        'MO', 'orient. averaged cs', 'pe energy/eV'
    end if

    ! in case of MO occupation averaging we render all occupation patterns; otherwise just take F%N_occ
    if (.NOT. P%yes_all_occs) then
      allocate (initial_occ(size(F%N_occ, 1), 1))
      initial_occ(:, 1) = F%N_occ(:)
    else
      call averaged_occs(F, F%N_occ, initial_occ)
    end if

    allocate (weight_io(size(initial_occ, 2)))
    weight_io(:) = 0

    AvgTotalCS = ZERO
    AvgTotal_n = 0
    do io = 1, size(initial_occ, 2)
      TotalCS = ZERO
      Total_n = 0
      if (P%print_level .gt. 1) &
        write (P%OUTPUT, "('# Initial occupation: ',99i1)") initial_occ(:, io)
      initial_Nopen = count(initial_occ(:, io) .eq. 1)
      do initial_multp = initial_Nopen + 1, 1, -2
        weight_io(io) = weight_io(io) + initial_multp*num_csf(initial_Nopen, initial_multp)
      end do

      if (P%print_level .gt. 1 ) &
        write (P%OUTPUT, "('# ',a4,4x,a13,4x,a13,2x,a10,a10)") &
        'mo', 'CS', 'E_Photoel.', 'multp_i', 'multp_f'

      do mo = 1, size(initial_occ, 1)
        if (initial_occ(mo, io) .eq. 0) cycle
        if (omega + F%CE(mo) .lt. 0) cycle

        if (initial_occ(mo, io) .eq. 1) then
          parti_Nopen = 1
          partf_Nopen = 0
        else if (initial_occ(mo, io) .eq. 2) then
          parti_Nopen = 0
          partf_Nopen = 1
        end if
        rest_Nopen = initial_Nopen - parti_Nopen
        final_Nopen = rest_Nopen + partf_Nopen

        SpinavgCS = ZERO
        Spinavg_n = 0
        do initial_multp = initial_Nopen + 1, 1, -2
          if (.not. P%yes_all_multps .and. initial_multp .ne. F%multp) cycle
          initial_Ncsf = num_csf(initial_Nopen, initial_multp)
          ChannelCS = ZERO
          do final_multp = initial_multp - 1, initial_multp + 1, 2
            if (final_multp .eq. 0) cycle
            if (final_multp - 1 .gt. final_Nopen) cycle
            ! We calculate transition rate to high spin state
            ! Apply Wigner Eckert to get factor for transition rates to other MS configurations:
            initial_two_S = initial_multp - 1
            initial_two_M = initial_two_S
            final_two_S = final_multp - 1
            WE_factor = ONE/cleb(final_two_S, final_two_S, 1, initial_two_S - final_two_S, initial_two_S, initial_two_S)**2

            if (initial_occ(mo, io) .eq. 1) then
              parti_two_S = 1
              partf_two_S = 0
            else
              parti_two_S = 0
              partf_two_S = 1
            end if

            if (initial_occ(mo, io) .eq. 2) then
              rest_Ncsf = num_csf(initial_Nopen, initial_multp)
              spin_factor = cleb(1, final_two_S - initial_two_S, initial_two_S, initial_two_S, final_two_S, final_two_S)**2
            else if (initial_occ(mo, io) .eq. 1) then
              rest_Ncsf = num_csf(final_Nopen, final_multp)
              spin_factor = cleb(1, initial_two_S - final_two_S, final_two_S, final_two_S, initial_two_S, initial_two_S)**2
            end if
            SpinCS = spin_factor*WE_factor*rest_Ncsf/initial_Ncsf*sum(CS(mo, :))/THREE

            if (P%print_level .gt. 3) then
              write (P%OUTPUT, "('# ',8x,'CS',es20.7,x,'factor',es20.7,x)") &
                sum(CS(mo, :))/THREE, rest_Ncsf/initial_Ncsf**WE_factor*spin_factor
            end if

            ChannelCS = ChannelCS + SpinCS
            if (.not. P%yes_all_multps) &
              TotalCS = TotalCS + SpinCS
            call add_Process(P, F, S, "photoionization", SpinCS, omega + F%CE(mo), mo, 0, 0, &
              initial_occ=initial_occ(:, io), initial_multp=initial_multp, &
              initial_csf_num=1, &
              final_multp=final_multp, final_csf_num=1)
            if (P%print_level .gt. 1) then
              write (P%OUTPUT, "('# ',i4,4x,es13.5,4x,es13.5,2x,i10,i10)") &
                mo, SpinCS, AU2EV*(omega + F%CE(mo)), initial_multp, final_multp
            end if
          end do! final_multp
          if (P%yes_all_multps) then
            SpinavgCS = SpinavgCS + initial_multp*initial_Ncsf*ChannelCS
            Spinavg_n = Spinavg_n + initial_multp*initial_Ncsf
          end if
        end do! initial_multp
        if (P%yes_all_multps) then
          SpinavgCS = SpinavgCS/Spinavg_n
          TotalCS = TotalCS + SpinavgCS
        end if

        if (P%print_level .gt. 1 .and. P%yes_all_multps) then
          write (P%OUTPUT, "('# ',i4,4x,es13.5,4x,es13.5,2x,a20)") &
            mo, SpinavgCS, AU2EV*(omega + F%CE(mo)), 'spin average'
        end if
      end do! mo

      AvgTotalCS = AvgTotalCS + TotalCS*weight_io(io)
      if (P%print_level .gt. 1) then
        if (P%yes_all_multps) then
          write (P%OUTPUT, "('# ',a4,4x,es13.5,4x,a13,2x,a20,a10,i5)") &
            'tot', TotalCS, '*', 'spin average', 'weight', weight_io(io)
        else
          write (P%OUTPUT, "('# ',a4,4x,es13.5,4x,a13,2x,a20)") &
            'tot', TotalCS, '*', '*'
        end if
      end if
    end do! io

    AvgTotalCS = AvgTotalCS/sum(weight_io(:))
    if (P%print_level .gt. 1 ) &
      write (P%OUTPUT, "('# Average Total CS ',es13.5,' a.u. ')") AvgTotalCS

    deallocate (weight_io)
    deallocate (initial_occ)
  end subroutine calculate_photoIonization_CS_spin_avg

  !>
  !! trans_dipole_bf_cont(P, A, B, F, bf, energy,lc, mc, DI) result(tdipole)
  !!
  !! this function assembles the transition dipole matrix element
  !!
  !!   <  basis function | d | continuum_lc,mc >
  !! with d = x,y,z
  !! where x=sqrt(4*pi/3) * r * Y_1,1
  !!       y=sqrt(4*pi/3) * r * Y_1,-1
  !!       z=sqrt(4*pi/3) * r * Y_1,0
  !!
  !! The radial integral over the atomic orbitals given in DI is
  !!            infty
  !!           /
  !!DI%value = | dr   f  (r) * r * cont  (r)
  !!           /       l               lc
  !!          0
  !!
  !! the angular integral is calculated by
  !!
  !!                 /
  !! RealGauntcoef = | d \Omega  Y    Y     Y
  !!                 /            l,m  1,d   lc,mc
  !!
  !! with d = -1,0,1
  !!
  !! input parameters:
  !! DI is the structure in which the atomic dipole integrals are given
  !! bf is the basisfunction index
  !! energy is the continuum energy
  !! lc,mc are the continuum l and m angular quantum numbers
  !!
  !! result tdipole is an array of dimension 3
  !! tdipole(1) = transition dipole in x direction
  !! tdipole(2) = transition dipole in y direction
  !! tdipole(3) = transition dipole in z direction
  function trans_dipole_bf_cont(Binfo, DI, energy, lc, mc) result(tdipole)
    integer, intent(in) :: Binfo(:)
    type(TransitionDipoleMoments), intent(in) :: DI(:)

    real(long), intent(in) :: energy
    integer, intent(in) :: lc
    integer, intent(in) :: mc

    real(long), dimension(3) :: tdipole ! transition dipole matrix element to return

    integer :: at ! atom on which the bf is centered
    integer :: n, l, m    ! atomic orbital quantum numbers
    integer :: i, ifound, k

    ! for interpolation
    type(Linear) :: di_linear
    real(long) :: factor, dipval

    factor = sqrt(FOUR*PI/THREE)
    tdipole(:) = ZERO

    at = Binfo(1)
    n = Binfo(3)
    l = Binfo(4)
    m = Binfo(5)

    ! special case for hydrogen: no photoionization from hydrogen
    ! if(A(atom)%charge==1)return

    if (l .ne. lc + 1 .and. l .ne. lc - 1) return  ! return if there is no transition

    ! find the correct entry in DI that matches the description of the bf
    ifound = 0
    do i = 1, size(DI(at)%config(:, 1))      ! loop overall DI
      if (DI(at)%config(i, 1) .eq. n .and. &
        DI(at)%config(i, 2) .eq. l .and. &
        DI(at)%config(i, 3) .eq. lc) then
        ifound = i
        exit
      end if
    end do

    ! no dipole integral is available for this basisfunction: return ZERO
    if (ifound .eq. 0) return

    if (energy .gt. maxval(DI(at)%energy)) then
      ! do extrapolation for larger values
      k = size(DI(at)%energy)
      dipval = DI(at)%trans_dipole(k, ifound) + &
        (DI(at)%trans_dipole(k, ifound) - DI(at)%trans_dipole(k - 1, ifound))/ &
        (DI(at)%energy(k) - DI(at)%energy(k - 1)) &
        *(energy - DI(at)%energy(k))

    else if (energy .lt. minval(DI(at)%energy)) then
      ! do extrapolation for smaller values
      k = 1
      dipval = DI(at)%trans_dipole(k, ifound) + &
        (DI(at)%trans_dipole(k + 1, ifound) - DI(at)%trans_dipole(k, ifound))/ &
        (DI(at)%energy(k + 1) - DI(at)%energy(k)) &
        *(energy - DI(at)%energy(k))
    else
      ! do interpolation
      di_linear = linear_create(DI(at)%energy, DI(at)%trans_dipole(:, ifound))
      dipval = linear_interpolate(di_linear, energy)
      call purge_Linear(di_linear)
    end if

    ! <n,l,m | x | E,lc,mc >
    tdipole(1) = &
      factor*realGauntCoef(l, m, 1, 1, lc, mc) &
      *dipval
    ! <n,l,m | y | E,lc,mc >
    tdipole(2) = &
      factor*realGauntCoef(l, m, 1, -1, lc, mc) &
      *dipval
    ! <n,l,m | z | E,lc,mc >
    tdipole(3) = &
      factor*realGauntCoef(l, m, 1, 0, lc, mc) &
      *dipval

  end function trans_dipole_bf_cont

  !> This subroutine reads in the output of xatom
  !! called with '-transition b-f' option
  !! and stores the resulting radial transition dipole matrix
  !! elements (length gauge) in the array DI%trans_dipole(:,k) and the corresponding
  !! energies in DI%energy(:). k is the index representing the transition (n,l,lc)
  !! quantum numbers.
  subroutine readDIntegral(P, filename, DI)
    type(Param), intent(in) :: P
    character(len=512), intent(in) :: filename
    type(TransitionDipoleMoments), intent(inout) :: DI
    integer :: RWStatus
    integer :: i, num_entries, num_lines, j, k, num_head_lines
    real(long) :: E, Eold
    real(long), allocatable :: S(:)
    character(len=512) :: rest
    character(50) :: frmt
    integer :: num_min, num_bf

    ! set up the format for a line in the data file:
    ! energy in eV (f9.2)
    ! for every n,l,l' triple
    !   < P_n,l(r) | r | P_E,l'(r) > (es13.5)
    !   < P_n,l(r) | Z/r^2 | P_E,l'(r) > (es13.5)
    !
    !  frmt = '(f9.2,       (es13.5,es13.5))'
    !  write(frmt(7:13),"(i7)") num_entries
    num_head_lines = 0
    open (unit=32, file=trim(filename), IOSTAT=RWStatus, status='OLD', ACTION='READ')
    if (RWStatus .eq. 0) then
      READ (32, '(a)', IOSTAT=RWStatus) rest
      if (RWStatus .eq. 0 .and. rest(1:5) .eq. "basis") then
        num_head_lines = num_head_lines + 1
        write (P%OUTPUT, '("# Reading minimal orbital coefficient for ",a)') trim(rest)
        read (32, *) num_min, num_bf
        num_head_lines = num_head_lines + 1

        allocate (DI%Pmin(num_min, num_bf))
        allocate (DI%min_info(num_min, 5))
        do i = 1, num_min
          num_head_lines = num_head_lines + 1
          read (32, *) DI%min_info(i, 1), &
            DI%min_info(i, 3), &
            DI%min_info(i, 4), &
            DI%min_info(i, 5), &
            DI%Pmin(i, 1:num_bf)
        end do
      else
        rewind (32)
      end if
    end if

    ! count the number of lines = number of energies
    if (RWStatus .ne. 0) then
      write (STDERR, *) filename
      stop 'cannot open file'
    else
      num_lines = 0
      ! first line is number of configurations=transitions
      READ (32, "(i10)") num_entries
      num_head_lines = num_head_lines + 1
      allocate (S(num_entries))
      allocate (DI%config(num_entries, 3))

      !10   FORMAT( <num_entries>(i4,i4,i4) )
      !READ(32, 10) (DI%config(k,1),DI%config(k,2),DI%config(k,3),k=1,num_entries)
      WRITE (frmt, '("(",I0, "(i4,i4,i4) )")') num_entries
      READ (32, frmt) (DI%config(k, 1), DI%config(k, 2), DI%config(k, 3), k=1, num_entries)
      num_head_lines = num_head_lines + 1
      WRITE (frmt, '("( f9.2, ",I0, "(es13.5) )")') num_entries

      !20   FORMAT( f9.2, <num_entries>(es13.5) )
      Eold = -ONE
      do
        !READ(32, 20, IOSTAT=RWStatus) E,(S(k), k = 1,num_entries)
        READ (32, frmt, IOSTAT=RWStatus) E, (S(k), k=1, num_entries)
        IF (RWStatus .eq. 0 .and. E .gt. Eold) THEN
          Eold = E
          num_lines = num_lines + 1
        elseif (RWStatus .lt. 0) THEN
          exit ! end of file
        else
          cycle ! the line doesn't fit the format
        end if
      end do
    end if

    if (P%print_level .gt. 1) then
      write (P%OUTPUT, "('# Reading ',i4,' transition dipole matrix elements at ',i4,' different energies')") num_entries, num_lines
      do i = 1, num_entries
        write (P%OUTPUT, "('# n=',i4,'l=',i4,'->','E, l=',i4)") DI%config(i, 1), DI%config(i, 2), DI%config(i, 3)
      end do
    end if

    allocate (DI%energy(num_lines))
    allocate (DI%trans_dipole(num_lines, num_entries))
    WRITE (frmt, '("( f9.2, ",I0, "(es13.5) )")') num_entries
    ! start reading the file from the beginning, skipping the head
    rewind (32)
    do j = 1, num_head_lines
      READ (32, *)
    end do
    j = 0
    Eold = -ONE
    do
      !READ(32, 40, IOSTAT=RWStatus) E,(S(k), k = 1,num_entries)
      READ (32, frmt, IOSTAT=RWStatus) E, (S(k), k=1, num_entries)
      IF (RWStatus .eq. 0 .and. E .gt. Eold) THEN
        j = j + 1
        Eold = E
        DI%trans_dipole(j, :) = S(:)
        DI%energy(j) = E*EV2AU ! energies are in eV, we want au
      elseif (RWStatus .lt. 0) THEN
        exit ! end of file
      else
        cycle ! the line doesn't fit the format
      end if
    end do

    close (32)
    deallocate (S)
    !  write(P%OUTPUT,*) filename,'maximum energy',maxval(DI%energy)*AU2EV
  end subroutine readDIntegral

  subroutine purgeDIntegral(DI)
    type(TransitionDipoleMoments), intent(inout) :: DI
    deallocate (DI%config)
    deallocate (DI%energy)
    deallocate (DI%trans_dipole)
    if (allocated(DI%Pmin)) deallocate (DI%Pmin)
    if (allocated(DI%Omin)) deallocate (DI%Omin)
    if (allocated(DI%Smin)) deallocate (DI%Smin)
    if (allocated(DI%min_info)) deallocate (DI%min_info)
  end subroutine purgeDIntegral

  subroutine shakeup_analysis(OUTPUT, P, B, F)
    !! performs an analysis of shakeup transitions based on mutual overlap of molecular orbitals
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: OUTPUT

    integer, allocatable :: occs(:, :), occs_temp(:,:)
    real(long), allocatable :: overlap2(:)
    integer, allocatable :: active_space(:)
    integer :: i,j,k
    real(long), parameter :: PRINT_THRESHOLD = 1e-4
    type(CSF), allocatable :: csf_array(:) ! array for csfs
    real(long), allocatable :: csf_O(:, :) ! overlap

    integer, allocatable :: active(:)
    integer, allocatable :: core(:)
    real(long) :: F_core(B%N_basis, B%N_basis)
    real(long) :: E_core
    integer :: NOpenMax
    type(UmatrixTable) :: UT
    integer :: nintern
    real(long), allocatable :: CI_matrix(:,:), CI_energy(:)
    integer :: f_irrep, n_irrep

    if (P%yes_sym) then
      n_irrep = B%nirrep
    else
      n_irrep = 1
    end if
    do f_irrep = 1, n_irrep
      if( P%state_irrep > 0 .and. P%state_irrep/=f_irrep)then
        write (P%OUTPUT, "('# Skipping final state irrep ',i3, ' - not matching the initial state irrep', i3)") &
          f_irrep, P%state_irrep
        cycle
      end if
      write (P%OUTPUT, "('# Final state irrep ',i3)") f_irrep

      if (allocated(P%final_gas)) then
        call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%final_gas, occs)
      else if (allocated(P%active_space_MO)) then
        call active_space_occs(F%N_occ, P%active_space_MO, occs)
        write (OUTPUT, '("# ",a,99i3)') 'active_space', P%active_space_MO
      else
        call auto_active_space(F%N_occ, F%LC, F%LC_ini, F%Strans, active_space, F%Ncore)
        write (OUTPUT, '("# ",a,99i3)') 'active_space', active_space
        call active_space_occs(F%N_occ, active_space, occs)
      end if
      if (P%yes_sym) then
        call filter_sym_occs(B%irrep_prod, F%irrep_MO, f_irrep, occs(:, :), occs_temp)
        deallocate (occs)
        allocate (occs(size(occs_temp, 1), size(occs_temp, 2)))
        occs = occs_temp
        deallocate (occs_temp)
      end if

      call construct_csf_array(P, B, F, F%multp, F%state_irrep, occs(:,:), csf_array)
      allocate (csf_O(size(csf_array), size(csf_array)))
      call overlap_csfs(P, csf_array, F%LC, F%LC_ini, F%STrans, csf_O)

      NOpenMax = maxval(count(occs(:, :) .eq. 1, dim=1))
      call constructUMatrixTable(NOpenMax, F%multp, UT)
      allocate (active( count((/ (  ( any(occs(j,:)<2) .and. any(occs(j,:)>0) ), j=1,size(occs,1)) /)) ))
      allocate (core(count((/((all(occs(j,:) .eq. 2)), j=1, size(occs, 1))/))))
      active = pack((/(j, j=1, size(occs, 1))/), &
            (/(((any(occs(j, :) .lt. 2) .and. any(occs(j, :) .gt. 0))), j=1, size(occs, 1))/))
      core = pack((/(j, j=1, size(occs, 1))/), &
            (/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))
      nintern = size(active)
      do j = size(active), 1, -1
        if (any(sum(occs(active(j:), :), dim=1) .gt. 1)) exit
      end do
      nintern = j
      write (OUTPUT, '("#     core:",99i4)') core
      write (OUTPUT, '("#   active:",99i4)') active
      write (OUTPUT, '("# internal:",99i4)') active(:nintern)
      write (OUTPUT, '("# external:",99i4)') active(nintern + 1:)
      call calculate_F_inactive(P, B, core, F%LC, F_core)
      E_core = ZERO
      do i = 1, size(core)
        E_core = E_core + F_core(core(i), core(i)) + F%H0(core(i), core(i))
      end do

      allocate (CI_matrix(size(csf_array), size(csf_array)))
      allocate (CI_energy(size(csf_array)))
      call transform_twoel_allactive(P, B, F, active(1:nintern), OUTPUT)
      if (nintern .lt. size(active)) then
        call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), OUTPUT)
        call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), OUTPUT)
        call calc_CIMatrix(UT, E_core, F_core(active(:), active(:)), &
            F%teint_abcd, &
            occs(active(:), 1:), &
            CI_matrix, nintern=nintern, &
            teint_iiee=F%teint_cdij, &
            teint_ieie=F%teint_cidj, &
            teint_iiie=F%teint_cdei)
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        if(allocated(F%teint_cdei)) deallocate (F%teint_cdei)
      else
        call calc_CIMatrix(UT, E_core, F_core(active(:),active(:)), F%teint_abcd, occs(active(:), :), CI_matrix)
      end if
      write (OUTPUT, '("# done")')
      ! solve it
      write (OUTPUT, '("# solving CI matrix")')
      call get_eigenvalues(CI_matrix, CI_energy)
      write (OUTPUT, '("# done")')
      allocate (overlap2(size(csf_O, 1)))
      do j=1,size(overlap2,1)
        if(any(csf_array(j)%occ(:) /= F%N_occ(:)))&
            cycle
        overlap2(:) = (CI_matrix .tm. csf_O(:, j))**2
        write (OUTPUT, '("# ",a)') "Shakeup probabilities"
        write (OUTPUT, '("# ",a,x,a)') "main configuration", csfstring(csf_array(j)%occ, csf_array(j)%csf)
        write (OUTPUT, '("# ",a,x,f15.5)') "Reference energy (a.u.) ", F%E
        write (OUTPUT, '("# ",a15)') "sum of squared overlaps:"
        write (OUTPUT, '("# ",f15.5,x,f15.5)') sum(overlap2(:)), sum(csf_O(:, j)**2)
        write (OUTPUT, '("# ",a12,x,a15,x,a15,x,a15)') "state index","excitation (eV)", "probability","total energy"
        do i=1,size(CI_matrix,2)
          if (overlap2(i) .gt. PRINT_THRESHOLD ) then
            write (OUTPUT, '("# SHAKE ",i6,x,f15.5,x,es15.5,x,f15.5)') i, (CI_energy(i)-F%E)*AU2EV, overlap2(i),&
              CI_energy(i) + F%E_nuc
            do k=1,size(csf_array)
              if( abs(CI_matrix(k,i))**2 .gt. PRINT_THRESHOLD * 100_long)&
                  write(OUTPUT, '("## ",10x,es15.5,x,a)') CI_matrix(k,i), csfstring(csf_array(k)%occ, csf_array(k)%csf)
            end do
          end if
        end do
      end do! j

      !  do i=1,size(occs,2)
      !    if(overlap2(i)>PRINT_THRESHOLD)&
      !      write(OUTPUT,'("# ",es15.5,x,999i1)') overlap2(i),occs(:,i)
      !  end do

      if (allocated(active_space)) deallocate (active_space)
      call purge_csf_array (csf_array)
      deallocate (overlap2)
      deallocate (occs)
      deallocate (core)
      deallocate (active)
      deallocate (CI_matrix)
      deallocate (CI_energy)
      deallocate (csf_O)
    end do! f_irrep
  end subroutine shakeup_analysis

end module photoionization
