!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module population
  !! author: Ludger Inhester
  !! author: Sang-Kil Son
  !!
  !! module for various kinds of popilation analysis
  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use operations
  use basis_func
  use matrix
  use eigen_LAPACK
  implicit none
  private
  public :: mulliken_population_analysis, loewdin_population_analysis, gross_population_analysis, &
    bondOrder_analysis, atomic_density_population_analysis, spin_analysis, fragment_analysis, &
    loewdin_population, mulliken_population, dominant_subshell_mo, density_analysis
contains

  !>----------------------------------------------------------------------------
  !! Mulliken atomic population analysis
  !!   Ref) Mulliken, JCP 23, 1833 (1955)
  !!
  !!                     N_basis
  !!   N(i;j) = N(i) c_ij  SUM  c_ik S_jk ...... Eq. (6') in Mulliken55
  !!                       k=1
  !!
  !!   where nonzero S_jj must be included in Eq. (6').
  !!   Then the population of each AO is evaluated by
  !!
  !!          N_MO
  !!   N(j) = SUM N(i;j) ...... Eq. (7) in Mulliken
  !!          i=1
  !!----------------------------------------------------------------------------
  subroutine mulliken_population_analysis(A, B, F, OUTPUT, D_given)
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    real(long), optional :: D_given(B%N_basis, B%N_basis)
    integer :: j, k, nl_shell, j_nuc
    real(long) :: atomic_charge(F%N_nuc)
    real(long) :: pop, pop_sum, pop_nl, pop_atom
    character(len=180) :: config_string(F%N_nuc)
    character(len=8) :: str
    real(long) :: D(B%N_basis, B%N_basis)
    real(long) :: DS(B%N_basis, B%N_basis)
    if (present(D_given)) then
      D = D_given
    else
      call density_matrix(F%LC, dble(F%N_occ), D)
    end if
    DS = D(:, :) .m.B%S(:, :)

    write (OUTPUT, "('# Mulliken atomic population analysis: ')")
    write (OUTPUT, "('#   ',12x,4a10)") "S(j,j)", "Pop", "P(n,l)", "P(atom)"
    write (OUTPUT, "('#   ',13x,a)") repeat('-', 39)
    pop_sum = ZERO
    pop_nl = ZERO
    pop_atom = ZERO
    k = 1
    nl_shell = 1
    config_string(:) = ""
    do j = 1, B%N_basis   ! j: AO (basis func) index
      pop = ZERO
      j_nuc = B%info(j, 1) ! atomic index corresponding to j-th AO
      pop = DS(j, j)
      write (OUTPUT, "('#   ',a10,2x,f10.7,f10.3)", advance="no") &
        B%label(j), B%S(j, j), pop
      pop_nl = pop_nl + pop
      pop_atom = pop_atom + pop
      pop_sum = pop_sum + pop
      if (j .lt. B%N_basis) then
        if (B%info(j + 1, 1) .ne. B%info(j, 1)) then  ! the next will be a new atom
          write (OUTPUT, "(2f10.3)", advance="no") pop_nl, pop_atom
          str = subshell_name(B%info(j, 3), B%info(j, 4))
          call append_string(config_string(k), str)
          call append_string_real(config_string(k), pop_nl)
          atomic_charge(k) = A(k)%charge - pop_atom
          pop_nl = ZERO
          pop_atom = ZERO
          nl_shell = nl_shell + 1
          k = k + 1
        else if (B%info(j + 1, 3) .ne. B%info(j, 3) .or. &     ! if next is diff. n
                 B%info(j + 1, 4) .ne. B%info(j, 4)) then     !        and diff. l
          write (OUTPUT, "(f10.3)", advance="no") pop_nl
          str = subshell_name(B%info(j, 3), B%info(j, 4))
          call append_string(config_string(k), str)
          call append_string_real(config_string(k), pop_nl)
          call append_string(config_string(k), "_")
          pop_nl = ZERO
          nl_shell = nl_shell + 1
        end if
      else
        write (OUTPUT, "(2f10.3)", advance="no") pop_nl, pop_atom
        nl_shell = nl_shell + 1
        str = subshell_name(B%info(j, 3), B%info(j, 4))
        call append_string(config_string(k), str)
        call append_string_real(config_string(k), pop_nl)
        atomic_charge(k) = A(k)%charge - pop_atom
      end if
      write (OUTPUT, "()")
      if (pop_atom .eq. ZERO) write (OUTPUT, "('#')")
    end do
    write (OUTPUT, "('#   ',13x,a)") repeat('-', 39)
    write (OUTPUT, "('#   ',29x,a13,f10.3)") "SUM of pop = ", pop_sum
    write (OUTPUT, "('#')")
    do k = 1, F%N_nuc
      if (A(k)%charge .eq. 0) cycle
      write (OUTPUT, "('#   ',a,' atomic charge of ',a5,' = ',sp,f10.3,4x,a)") &
        "|MULLIKEN|", A(k)%label, atomic_charge(k), trim(config_string(k))
    end do
    write (OUTPUT, "('#')")
  contains
    subroutine append_string(s, substr) ! add substr at the end of s
      character(len=80), intent(inout) :: s
      character(len=*), intent(in) :: substr
      s = trim(s)//adjustl(substr)
    end subroutine append_string

    subroutine append_string_real(s, f) ! add a real number at the end of s
      character(len=80), intent(inout) :: s
      real(long), intent(in) :: f
      character(len=8) :: substr
      write (substr, "(f8.2)") f
      if (f .lt. ZERO) write (substr, "(f8.2)") ZERO      ! make population > 0
      s = trim(s)//adjustl(substr)
    end subroutine append_string_real
  end subroutine mulliken_population_analysis

  !>----------------------------------------------------------------------
  !! Calculates bond-order btw atoms.
  !!   -diagonal elements- should be ignored (in practice they are juat the Mulliken charges)
  !! revised 13.09.2016 by Ludger:
  !! Bond order is 
  !! $$ B_{AB} = \sum_{\mu on atom A} \sum_{\nu on atom B} (DS)_\mu\nu * (DS)_\nu\mu $$
  !! See e.g. INTERNATIONAL JOURNAL OF QUANTUM CHEMISTRY, VOL. XXVI, I S 1-154 ( 1984)
  !!
  !! if( OUTPUT <0 )  no print-out
  subroutine bondOrder_analysis(P, B, D, OUTPUT, bondOrder)
    implicit none
    Type(Param) :: P
    Type(Basis) :: B
    real(long), intent(in) :: D(B%N_basis, B%N_basis)
    integer, intent(in) :: OUTPUT
    real(long), Optional, intent(out), target :: bondOrder(P%N_nuc, P%N_nuc)
    real(long), pointer :: BO(:, :)
    real(long), allocatable, target :: boLoc(:, :)
    integer :: i, j
    ! call density_matrix(F%LC,dble(F%N_occ),D)
    ! complicated pointer issue
    if (present(bondOrder)) then; 
      BO => bondOrder
    else
      allocate (boLoc(P%N_nuc, P%N_nuc))
      BO => boLoc
    end if
    BO = calculate_BondOrder(B%S, D, B%N_basis, B%info(:, 1), P%N_nuc)
    if (output .gt. 0) then
      do i = 1, P%N_nuc
        if (P%charge_nuc(i) .eq. 0) cycle
        write (OUTPUT, '(a,i3,999f12.6)') "#BondOrder:", i, BO(i, :)
      end do
      do i = 1, P%N_nuc
        if (P%charge_nuc(i) .eq. 0) cycle
        do j = i + 1, P%N_nuc
          if (P%charge_nuc(j) .eq. 0) cycle
          if (BO(i, j) .gt. 0.5) then
            write (OUTPUT, '("# ",i4,i4,f7.3)') i, j, BO(i, j)
          end if
        end do
      end do
    end if
    if (allocated(boLoc)) then; deallocate (boLoc); end if
    Nullify (BO)

  contains
    function calculate_BondOrder(S, D, Nb, jBtoInuc, Nnuc) result(ret)
      implicit none
      integer, intent(in) :: Nnuc, Nb, jBtoInuc(Nb)
      real(long) :: S(Nb, Nb), D(Nb, Nb)
      real(long) :: DS(Nb, Nb)
      real(long) :: ret(Nnuc, Nnuc)
      integer :: i, j, jB, kB
      DS = matmul(D, S)
      ret(:, :) = ZERO
      do i = 1, Nnuc
        do j = 1, i; 
          do jB = 1, Nb; 
            if (jBtoInuc(jB) .ne. i) cycle; 
            do kB = 1, Nb; 
              if (jBtoInuc(kB) .ne. j) cycle; 
              ret(j, i) = ret(j, i) + DS(jB, kB)*DS(kB, jB)
            end do
          end do
          ret(i, j) = ret(j, i)
        end do
      end do
    end Function calculate_BondOrder
  end subroutine bondOrder_analysis

  subroutine fragment_analysis(P, B, F, OUTPUT)
    !! analysis possible fragments based on bond order values
    Type(Param) :: P
    Type(Basis) :: B
    Type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT

    integer :: fragment(P%N_nuc, P%N_nuc)
    integer :: fragment_c
    integer :: fragment_n(P%N_nuc)
    real(long) :: BO(P%N_nuc, P%N_nuc)
    integer ::i, j

    integer :: mu, nu
    real(long) :: D(B%N_basis, B%N_basis)
    real(long) :: Q(P%N_Nuc)

    call density_matrix(F%LC, dble(F%N_occ), D)

    do i = 1, P%N_Nuc
      Q(i) = dble(P%charge_nuc(i))
    end do
    do mu = 1, size(D, 1)
      do nu = 1, size(D, 1)
        Q(B%info(mu, 1)) = Q(B%info(mu, 1)) - D(mu, nu)*B%S(mu, nu)
      end do
    end do

    call bondOrder_analysis(P, B, D, 0, BO)

    fragment(:, :) = 0
    fragment_c = 1
    fragment(1, 1) = 1
    fragment_n(1) = 1

    fragment_loop: do
      call add_atoms_to_fragment(BO, fragment(:, fragment_c), fragment_n(fragment_c))

      do i = 1, P%N_nuc
        if( P%charge_nuc(i) == 0) cycle ! skip point charges
        if (ANY(i .eq. fragment(:, :))) cycle
        fragment_c = fragment_c + 1
        fragment_n(fragment_c) = 1
        fragment(1, fragment_c) = i
        cycle fragment_loop
      end do
      exit
    end do fragment_loop

    write (OUTPUT, '("# Summary of fragment analysis ")')
    write (OUTPUT, '("# Number of fragments ",i4)') fragment_c
    do j = 1, fragment_c
      write (OUTPUT, '("# Atom indices in fragment ",i4," : ")') j
      do i = 1, fragment_n(j)
        write (OUTPUT, '("# ",i4)') fragment(i, j)
      end do
      write (OUTPUT, '("# Fragment charge ",f7.3)') sum(Q(fragment(:fragment_n(j), j)))
    end do

  contains
    recursive subroutine add_atoms_to_fragment(BO, fragment, index)
      real(long), intent(in) :: BO(:, :)
      integer, intent(inout) :: fragment(:)
      integer, intent(inout) :: index
      real(long), parameter :: bond_order_criterion = 0.5
      integer :: i, j
      j = fragment(index)
      do i = 1, size(BO, 1)
        if (any(fragment(:) .eq. i)) cycle
        if (BO(i, j) .gt. bond_order_criterion) then
          index = index + 1
          fragment(index) = i
          call add_atoms_to_fragment(BO, fragment, index)
        end if
      end do
    end subroutine add_atoms_to_fragment
  end subroutine fragment_analysis


  subroutine gross_population_analysis(P, A, B, F, OUTPUT, Gross_pop, sym_l)
    !! gross orbital population analysis
    Type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: pop(:, :)
    real(long), optional, intent(out) :: Gross_pop(B%N_basis, size(A))
    integer, optional, intent(in) :: sym_l
    integer :: i, j, atom
    character(len=11) :: string
    if (present(sym_l)) then
      if (sym_l .ge. 0) then
        write (OUTPUT, '("#",x,"POPULATION FOR BASISFUNCTION of l=",i1," SYMMETRY")') sym_l
      else
        stop "wrong sym argument"
      end if
    end if

    allocate (pop(B%N_basis, size(A)))
    pop = ZERO
    do j = 1, B%N_basis   ! j: AO (basis func) index
      atom = B%info(j, 1)
      if (present(sym_l)) then
        if (sym_l .ne. B%info(j, 4)) cycle
      end if
      do i = 1, F%N_basis    ! i: MO index
        pop(i, atom) = pop(i, atom) + F%LC(j, i)*sum(F%LC(:, i)*B%S(j, :))
      end do
    end do
    if (present(Gross_pop)) Gross_pop(:, :) = pop(:, :)

    if (P%print_level > 1) then
      write (OUTPUT, '("#GROSS ",a4,1x,a8,1x,a8,1x,a4)', advance="no") "MO", "MO_energy", "f_occ", "occ"
      do atom = 1, size(A)
        if (A(atom)%charge .gt. 0) write (OUTPUT, '(2x,a8)', advance="no") A(atom)%label
      end do
      write (OUTPUT, '()')
      do i = 1, F%N_basis    ! i: MO index
        if (i .eq. F%NCORE + 1 &
            ) write (OUTPUT, '("# ------ core/valence ----  ")')
        write (OUTPUT, '("#GROSS ",i4,1x,f8.2,1x,f8.5,1x,i4)', advance="no") i, F%CE(i), F%f_occ(i), F%N_occ(i)
        do atom = 1, size(A)
          if (A(atom)%charge .gt. 0) write (OUTPUT, '(2x,f8.5)', advance="no") pop(i, atom)
        end do
        string = ""
        do j = 1, F%N_basis
          if (F%LC(j, i)*F%LC(j, i) .gt. 0.1) then
            write (OUTPUT, '(1x,f5.1,1x,a,1x)', advance="no") F%LC(j, i), B%label(j)
          end if
        end do
        write (OUTPUT, '()')
      end do
      write (OUTPUT, "('#')")
    end if
    deallocate (pop)

  end subroutine gross_population_analysis

  subroutine mulliken_population(A, B, F, pop, dmatrix)
    !! mulliken population analysis
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(out) :: pop(size(A, 1))
    real(long), intent(in), optional:: dmatrix(B%N_basis, B%N_basis)
    integer :: i
    real(long) :: D(B%N_basis, B%N_basis)
    real(long) :: DS(B%N_basis, B%N_basis)
    if (present(dmatrix)) then
      D = dmatrix
    else
      call density_matrix(F%LC, dble(F%N_occ), D)
    end if
    DS = D(:, :) .m.B%S(:, :)
    pop = ZERO
    do i = 1, size(D, 1)
      pop(B%info(i, 1)) = pop(B%info(i, 1)) + DS(i, i)
    end do
  end subroutine mulliken_population

  subroutine loewdin_population(A, B, F, pop, dmatrix)
    !! loewdin population analysis
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(out) :: pop(size(A, 1))
    real(long), intent(in), optional:: dmatrix(B%N_basis, B%N_basis)
    real(long) :: DS(B%N_basis, B%N_basis)
    real(long) :: D(B%N_basis, B%N_basis)
    integer :: i
    if (present(dmatrix)) then
      D = dmatrix
    else
      call density_matrix(F%LC, dble(F%N_occ), D)
    end if
    DS = D(:, :) .m.B%SSqrt(:, :)

    pop = ZERO
    do i = 1, size(D, 1)
      pop(B%info(i, 1)) = pop(B%info(i, 1)) + sum(B%SSqrt(i, :)*DS(:, i))
    end do
  end subroutine loewdin_population

  subroutine loewdin_population_analysis(A, B, F, OUTPUT, D_given)
    !! loewdin population analysis
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    integer :: j, k
    real(long) :: atomic_charge(F%N_nuc)
    real(long) :: pop, pop_sum, pop_nl, pop_atom
    character(len=180) :: config_string(F%N_nuc)
    character(len=8) :: str

    real(long) :: DSSqrt(B%N_basis, B%N_basis)
    real(long) :: D(B%N_basis, B%N_basis)
    real(long), optional :: D_given(B%N_basis, B%N_basis)

    if (present(D_given)) then
      D = D_given
    else
      call density_matrix(F%LC, dble(F%N_occ), D)
    end if
    DSSqrt = D(:, :) .m.B%SSqrt(:, :)

    write (OUTPUT, "('# Loewdin atomic population analysis: ')")
    write (OUTPUT, "('#   ',12x,4a10)") "S(j,j)", "Pop", "P(n,l)", "P(atom)"
    write (OUTPUT, "('#   ',13x,a)") repeat('-', 39)

    pop_sum = ZERO
    pop_nl = ZERO
    pop_atom = ZERO
    k = 1
    config_string(:) = ""
    do j = 1, B%N_basis   ! j: AO (basis func) index
      pop = ZERO
      pop = pop + sum(B%SSqrt(j, :)*DSSqrt(:, j))
      write (OUTPUT, "('#   ',a10,2x,f10.7,f10.3)", advance="no") &
        B%label(j), B%S(j, j), pop
      pop_nl = pop_nl + pop
      pop_atom = pop_atom + pop
      pop_sum = pop_sum + pop
      if (j .lt. B%N_basis) then
        if (B%info(j + 1, 1) .ne. B%info(j, 1)) then  ! the next will be a new atom
          write (OUTPUT, "(2f10.3)", advance="no") pop_nl, pop_atom
          str = subshell_name(B%info(j, 3), B%info(j, 4))
          call append_string(config_string(k), str)
          call append_string_real(config_string(k), pop_nl)
          atomic_charge(k) = A(k)%charge - pop_atom
          pop_nl = ZERO
          pop_atom = ZERO
          k = k + 1
        else if (B%info(j + 1, 3) .ne. B%info(j, 3) .or. &     ! if next is diff. n
                 B%info(j + 1, 4) .ne. B%info(j, 4)) then     !        and diff. l
          write (OUTPUT, "(f10.3)", advance="no") pop_nl
          str = subshell_name(B%info(j, 3), B%info(j, 4))
          call append_string(config_string(k), str)
          call append_string_real(config_string(k), pop_nl)
          call append_string(config_string(k), "_")
          pop_nl = ZERO
        end if
      else
        write (OUTPUT, "(2f10.3)", advance="no") pop_nl, pop_atom
        str = subshell_name(B%info(j, 3), B%info(j, 4))
        call append_string(config_string(k), str)
        call append_string_real(config_string(k), pop_nl)
        atomic_charge(k) = A(k)%charge - pop_atom
      end if
      write (OUTPUT, "()")
      if (pop_atom .eq. ZERO) write (OUTPUT, "('#')")
    end do
    write (OUTPUT, "('#   ',13x,a)") repeat('-', 39)
    write (OUTPUT, "('#   ',29x,a13,f10.3)") "SUM of pop = ", pop_sum
    write (OUTPUT, "('#')")
    do k = 1, F%N_nuc
      if (A(k)%charge .eq. 0) cycle
      write (str, "(i8)") k
      write (OUTPUT, "('#   ',a,' atomic charge of ',a5,' = ',sp,f10.3,4x,a)") &
        "|LOEWDIN| ", A(k)%label, atomic_charge(k), trim(config_string(k))
    end do
    write (OUTPUT, "('#')")
  contains
    subroutine append_string(s, substr) ! add substr at the end of s
      character(len=180), intent(inout) :: s
      character(len=*), intent(in) :: substr
      s = trim(s)//adjustl(substr)
    end subroutine append_string

    subroutine append_string_real(s, f) ! add a real number at the end of s
      character(len=180), intent(inout) :: s
      real(long), intent(in) :: f
      character(len=8) :: substr
      write (substr, "(f8.2)") f
      if (f .lt. ZERO) write (substr, "(f8.2)") ZERO      ! make population > 0
      s = trim(s)//adjustl(substr)
    end subroutine append_string_real
  end subroutine loewdin_population_analysis

  subroutine dominant_subshell_MO(B, LC, max_atom, max_n, max_l, max_popnl)
    !! for a given MO return the dominant atomic subshell
    type(Basis), intent(in) :: B
    real(long), intent(in) :: LC(:)
    integer, intent(out) :: max_atom, max_n, max_l
    real(long), intent(out) :: max_popnl

    integer :: i, atom, n, l
    real(long) :: popnl

    max_popnl = ZERO
    atom = B%info(1, 1)
    n = B%info(1, 3)
    l = B%info(1, 4)
    popnl = LC(1)**2
    do i = 2, B%N_basis
      if (B%info(i, 1) .eq. atom .and. B%info(i, 3) .eq. n .and. B%info(i, 4) .eq. l) then
        popnl = popnl + LC(i)**2
      else

        if (popnl .gt. max_popnl) then
          max_popnl = popnl
          max_atom = atom
          max_n = n
          max_l = l
        end if

        popnl = LC(i)**2
        atom = B%info(i, 1)
        n = B%info(i, 3)
        l = B%info(i, 4)
      end if
      if (popnl .gt. max_popnl) then
        max_popnl = popnl
        max_atom = atom
        max_n = n
        max_l = l
      end if

    end do

  end subroutine dominant_subshell_MO

  subroutine atomic_density_population_analysis(G, A, F, OUTPUT)
    !! population analysis based on the grid points
    !!
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    type(WF_data), intent(inout) :: F     ! F%partial_charge(:) updated
    integer, intent(in) :: OUTPUT
    real(long) :: rho_A(G%N_grid) ! atomic density
    real(long) :: pop_atom(G%N_nuc)
    real(long) :: pop_smooth(G%N_nuc), pop_step(G%N_nuc)
    integer :: i_nuc, is, ie
    integer :: i, ibuf(1), nearest_nuc
    !
    ! multicenter decomposition
    !
    rho_A(:) = G%w_decomp(:)*F%rho(:)
    !
    ! calculate population of the atomic density
    !   rho_A(is:ie) : atomic density corresponding to the i_nuc-th atom
    !   G%weight2c(is:ie,i_nuc) : weights for single-center integration
    !
    do i_nuc = 1, G%N_nuc
      is = G%i_start(i_nuc)       ! starting index for i_nuc-th atom
      ie = G%i_end(i_nuc)         ! ending index for i_nuc-th atom
      pop_atom(i_nuc) = sum(rho_A(is:ie)*G%weight2c(is:ie, i_nuc))
    end do
    !
    ! test Kota's implementations
    !
    pop_smooth(:) = ZERO  ! using fuzzy Voronoi cells, equivalent to pop_atom(:)
    pop_step(:) = ZERO    ! using sharp Voronoi cells
    do i = 1, G%N_grid
      pop_smooth(:) = pop_smooth(:) + F%rho(i)*G%weight(i)*G%weight_nuc(:, i)
      ibuf = maxloc(G%weight_nuc(:, i)); nearest_nuc = ibuf(1)
      pop_step(nearest_nuc) = pop_step(nearest_nuc) + F%rho(i)*G%weight(i)
    end do
    !
    ! print out partial charges
    !
    if (OUTPUT .ge. 0) then            ! 28.07.2015 added : no printouts if output<0
      write (OUTPUT, "('# Atomic density population analysis "// &
             "based on Voronoi cells: ')")
      write (OUTPUT, "('#   ',36x,3a12)") "pop_atom", "pop_smooth", "pop_step"
      do i_nuc = 1, G%N_nuc
        write (OUTPUT, "('#   ',a,' atomic charge of ',a5,' = ',sp,3f12.3)") &
          "|ATOM_DEN|", A(i_nuc)%label, &
          A(i_nuc)%charge - pop_atom(i_nuc), &
          A(i_nuc)%charge - pop_smooth(i_nuc), &
          A(i_nuc)%charge - pop_step(i_nuc)
      end do
      write (OUTPUT, "('#')")
    end if
    !
    ! update F%partial_charge(:) using pop_atom(:)
    !
    F%partial_charge(:) = A(:)%charge - pop_atom(:)

    if (allocated(F%partialPopulation_G)) F%partialPopulation_G(:) = pop_atom(:)
    if (allocated(F%partialPopulation_G1)) F%partialPopulation_G1(:) = pop_step(:)
  end subroutine atomic_density_population_analysis

  subroutine spin_analysis(P, G, B, Da, Db, OUTPUT, center_out, rog_out)
    !! performs analysis of the spin center
    Type(Param) :: P
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B

    real(long), intent(in) :: Da(:, :)
    real(long), intent(in) :: Db(:, :)
    integer, intent(in) :: OUTPUT
    real(long), intent(out), optional :: center_out(3)
    real(long), intent(out), optional :: rog_out
    real(long), parameter :: threshold = 1e-9_long
    real(long) :: tdipole_bf(size(Da, 1), size(Da, 1), 3)
    real(long) :: quadrupole_bf(size(Da, 1), size(Da, 1), 3, 3)
    real(long) :: DSa(size(Da, 1), size(Da, 1))
    real(long) :: DSb(size(Da, 1), size(Da, 1))
    real(long) :: spin_D(size(Da, 1), size(Da, 1))
    real(long) :: center(3)
    real(long) :: spin_quad(3, 3), spin_quad_ev(3,3), sev(3)
    real(long) :: sumA, sumB
    integer :: i, j
    real(long) :: traceDSa, traceDSb
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    call calc_quadrupole_moment_bf(P, B, G, quadrupole_bf)
    sumA = ZERO
    sumB = ZERO
    DSa = Da.m.B%S
    DSb = Db.m.B%S
    traceDSa = matTrace(DSa)
    traceDSb = matTrace(DSb)
    sumA = traceDSa
    sumB = traceDSb
    if (abs(traceDSa-traceDSb) < threshold) then
      write (OUTPUT, '("# It seems all electrons are paired. Skipping spin analysis.")')
      return
    end if
    spin_D = (Da(:, :) - Db(:,:))/ (traceDSa - traceDSb)
    do i = 1, 3
      center(i) = sum(tdipole_bf(:, :, i) * spin_D(:,:) )
      if (P%yes_gaussian) &
        center(i) = center(i) + B%env(1+i) * sum(B%S(:,:) * spin_D(:,:))
    end do
    do i = 1, 3
      do j = 1, 3
        spin_quad(i,j) = sum(quadrupole_bf(:, :, i, j) * spin_D(:,:) )
        if (P%yes_gaussian) &
          spin_quad(i,j) = spin_quad(i,j) &
            + B%env(1+i) * sum(tdipole_bf(:, :, j) * spin_D(:,:) ) &
            + B%env(1+j) * sum(tdipole_bf(:, :, i) * spin_D(:,:) ) &
            + B%env(1+i) * B%env(1+j) * sum(B%S(:,:) * spin_D(:,:))
      end do
    end do
    do i = 1, 3
      do j = 1, 3
        spin_quad(i, j) = spin_quad(i, j) - center(i) * center(j)
      end do
    end do
    spin_quad_ev = spin_quad
    call get_eigenvalues(spin_quad_ev, sev)
    if (OUTPUT .gt. 0) then 
      write (OUTPUT, '("# Spin analysis:")')
      write (OUTPUT, '("#")')
      write (OUTPUT, '("# ", a26, x,  f15.5)') "alpha electrons :", traceDSa
      write (OUTPUT, '("# ", a26, x,  f15.5)') "beta electrons  :", traceDSb
      write (OUTPUT, '("# ", a26, x,  f15.5)') "N_alpha - N_beta:", traceDSa - traceDSb
      write (OUTPUT, '("# ", a26, x, 3f15.5)') "center (a.u.):", center(:)
      write (OUTPUT, '("# ", a26, x, 3f15.5)') "tensor eigenvalues (a.u.):", sev(:)
      write (OUTPUT, '("# ", a26, x,  f15.5)') "Radius of gyr. (a.u.):", sqrt(sum(sev(:)))
      write (OUTPUT, '("# ", a26, x,  f15.5)') "Asphericity b (a.u.):", sev(3) - (sev(1) + sev(2))/TWO
      write (OUTPUT, '("#")')
      !write (OUTPUT, '("# ", a20, x,  f15.5)') "Anisotropy kappa (a.u.)", THREE * HALF * sum(sev(:)**4) / sum(sev(:)**2)**2 - HALF
    end if
    if(present(center_out))&
      center_out(:) = center(:)
    if(present(rog_out))&
      rog_out = sqrt(sum(sev(:)))
  end subroutine spin_analysis

  subroutine density_analysis(P, G, B, D, OUTPUT, center_out, rog_out, description)
    !! performs analysis of the spin center
    Type(Param) :: P
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    real(long), intent(in) :: D(:, :)
    integer, intent(in) :: OUTPUT
    real(long), intent(out), optional :: center_out(3)
    real(long), intent(out), optional :: rog_out
    character(*), intent(in), optional :: description

    real(long), parameter :: threshold = 1e-9_long
    real(long) :: tdipole_bf(size(D, 1), size(D, 1), 3)
    real(long) :: quadrupole_bf(size(D, 1), size(D, 1), 3, 3)
    real(long) :: DS(size(D, 1), size(D, 1))
    real(long) :: Dn(size(D, 1), size(D, 1))

    real(long) :: center(3)
    real(long) :: quad(3, 3), quad_ev(3,3), sev(3)
    integer :: i, j
    real(long) :: traceDS
    call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
    call calc_quadrupole_moment_bf(P, B, G, quadrupole_bf)
    DS = D.m.B%S
    traceDS = matTrace(DS)
    Dn = D(:,:) / traceDS
    if (abs(traceDS) < threshold) then
      write (OUTPUT, '("# It seems there is no density. Skipping density analysis.")')
      return
    end if
    do i = 1, 3
      center(i) = sum(tdipole_bf(:, :, i) * Dn(:,:) )
      if (P%yes_gaussian) &
        center(i) = center(i) + B%env(1+i) * sum(B%S(:,:) * Dn(:,:))
    end do
    do i = 1, 3
      do j = 1, 3
        quad(i,j) = sum(quadrupole_bf(:, :, i, j) * Dn(:,:) )
        if (P%yes_gaussian) &
          quad(i,j) = quad(i,j) &
            + B%env(1+i) * sum(tdipole_bf(:, :, j) * Dn(:,:) ) &
            + B%env(1+j) * sum(tdipole_bf(:, :, i) * Dn(:,:) ) &
            + B%env(1+i) * B%env(1+j) * sum(B%S(:,:) * Dn(:,:))
      end do
    end do
    do i = 1, 3
      do j = 1, 3
        quad(i, j) = quad(i, j) - center(i) * center(j)
      end do
    end do
    quad_ev = quad
    call get_eigenvalues(quad_ev, sev)
    if (OUTPUT .gt. 0) then
      if(present(description))then
        write (OUTPUT, '("# ",a)') trim(description)
      else
        write (OUTPUT, '("# Density analysis:")')
      end if
      write (OUTPUT, '("#")')
      write (OUTPUT, '("# ", a26, x,  f15.5)') "N electrons :", traceDS
      write (OUTPUT, '("# ", a26, x, 3f15.5)') "center (a.u.):", center(:)
      write (OUTPUT, '("# ", a26, x, 3f15.5)') "tensor eigenvalues (a.u.):", sev(:)
      write (OUTPUT, '("# ", a26, x,  f15.5)') "Radius of gyr. (a.u.):", sqrt(sum(sev(:)))
      write (OUTPUT, '("# ", a26, x,  f15.5)') "Asphericity b (a.u.):", sev(3) - (sev(1) + sev(2))/TWO
      write (OUTPUT, '("#")')
    end if
    if(present(center_out))&
      center_out(:) = center(:)
    if(present(rog_out))&
      rog_out = sqrt(sum(sev(:)))
  end subroutine density_analysis


end module population
