!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module process
  !! author: Ludger Inhester
  !!
  !! process module adapted from xatom
  use constants
  use read_param
  use wavefunction
  use elec_config
  implicit none

  type :: Process_data
    !! process information
    integer, allocatable :: occ(:)          ! elec. occ. pattern
    integer :: N_process                    ! # of processes
    real(long) :: E_tot                     ! total energy

    character(len=1), allocatable :: tag(:) ! process name tag, 1...N_proc
    real(long), allocatable :: value(:)     ! cross sec or rate in a.u., 1...N_p
    integer, allocatable :: E_eV(:)         ! transition energy in eV, 1...N_proc
    integer, allocatable :: occ_up(:, :), occ_down(:, :)

    ! for tracing spin

    integer, allocatable :: initial_occ(:, :)

    integer, allocatable :: initial_csf_num(:)
    integer, allocatable :: final_csf_num(:)

    integer, allocatable :: initial_irrep(:)
    integer, allocatable :: initial_spin_multp(:)
    integer, allocatable :: initial_state_num(:)
    integer, allocatable :: final_irrep(:)
    integer, allocatable :: final_spin_multp(:)
    integer, allocatable :: final_state_num(:)
    real(long), allocatable :: E_au(:)

  end type Process_data
  private
  public :: Process_data, create_Process, purge_Process, add_Process, print_Process, occ_Process, &
    sum_equivalent_configuration
contains

!>----------------------------------------------------------------------------
!! add one process to S
!! - S%value(k): rate or cross section
!! - S%E_eV(k): transition energy in eV, rounded to nearest integer
!! - S%N_process is increased by one.
!! - S%tag is given by P, V, A, F, 2 (two-photon abs.), 3 (three-photon), ...
!!----------------------------------------------------------------------------
  subroutine add_Process(P, F, S, str, value, E, i, j, jp, initial_occ, &
    initial_multp, initial_csf_num, initial_state_num, initial_irrep, &
    final_multp, final_csf_num, final_state_num, final_irrep)
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S
    character(len=*), intent(in) :: str
    real(long), intent(in) :: value, E
    integer, intent(in) :: i      ! i, j, jp: orbital index for config(:,:)
    integer, intent(in), optional :: j, jp
    integer, intent(in), optional :: initial_occ(:)
    integer, intent(in), optional :: initial_multp, final_multp
    integer, intent(in), optional :: initial_csf_num, final_csf_num
    integer, intent(in), optional :: initial_state_num, final_state_num
    integer, intent(in), optional :: initial_irrep, final_irrep
    integer :: k
    !
    ! increase # of processes
    ! and put this process into Process_data
    !
    S%N_process = S%N_process + 1

    if (S%N_process .gt. size(S%tag)) then
      S%N_process = S%N_process - 1
      call print_Process(S, 6)
      stop "ERROR in add_Process(): N_process exceeds N_proc_max."
    end if
    k = S%N_process
    S%value(k) = value
    S%E_eV(k) = nint(E*AU2EV)    ! nint(): round to nearest integer
    if (allocated(S%E_au)) S%E_au(k) = E
    if (present(initial_occ)) &
      S%initial_occ(:, k) = initial_occ(:)
    if (present(initial_multp)) &
      S%initial_spin_multp(k) = initial_multp
    if (present(initial_csf_num)) &
      S%initial_csf_num(k) = initial_csf_num
    if (present(final_multp)) &
      S%final_spin_multp(k) = final_multp
    if (present(initial_csf_num)) &
      S%final_csf_num(k) = final_csf_num
    if (present(initial_state_num)) &
      S%initial_state_num(k) = initial_state_num
    if (present(final_state_num)) &
      S%final_state_num(k) = final_state_num
    if (present(final_irrep)) &
      S%final_irrep(k) = final_irrep
    if (present(initial_irrep)) &
      S%initial_irrep(k) = initial_irrep

    !
    ! set up the label for the process
    !
    select case (str)
     case ("P", "photoionization", "photoabsorption")  ! i ionized
      S%tag(k) = "P"
      S%occ_down(1, k) = i

     case ("A", "auger", "Auger")                      ! j -> i, jp ionized
      if (.not. (present(j) .and. present(jp))) &
        stop "ERROR when add_Process() is called with A."
      S%tag(k) = "A"
      S%occ_up(1, k) = i
      S%occ_down(2, k) = jp
      S%occ_down(1, k) = j

     case ("F", "fluorescence", "fluo")        ! j -> i
      if (.not. present(j)) stop "ERROR when add_Process() is called with F."

      S%tag(k) = "F"
      S%occ_up(1, k) = i
      S%occ_down(1, k) = j

    end select
    if (P%yes_average_rates) &
      call make_equivalent_configuration(F, S, k)

  contains
    subroutine make_equivalent_configuration(F, S, k)
      type(WF_data), intent(in) :: F

      type(Process_data), intent(inout) :: S
      integer, intent(in) :: k
      integer :: j, l

      if (allocated(F%av_occ)) then
        select case (S%tag(k))
         case ("P")
          av_loop_P: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_down(1, k))) then
              do l = size(F%av_occ(j)%MO), 1, -1
                if (S%occ(F%av_occ(j)%MO(l)) .gt. 0) then
                  S%occ_down(1, k) = F%av_occ(j)%MO(l)
                  exit av_loop_P
                end if
              end do
            end if
          end do av_loop_P

         case ("F")
          av_loop_F1: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_down(1, k))) then
              do l = size(F%av_occ(j)%MO), 1, -1
                if (S%occ(F%av_occ(j)%MO(l)) .gt. 0) then
                  S%occ_down(1, k) = F%av_occ(j)%MO(l)
                  exit av_loop_F1
                end if
              end do
            end if
          end do av_loop_F1
          av_loop_F2: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_up(1, k))) then
              do l = 1, size(F%av_occ(j)%MO), 1
                if (S%occ(F%av_occ(j)%MO(l)) .lt. 2) then
                  S%occ_up(1, k) = F%av_occ(j)%MO(l)
                  exit av_loop_F2
                end if
              end do
            end if
          end do av_loop_F2
         case ("A")

          av_loop_A1: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_down(1, k))) then
              do l = size(F%av_occ(j)%MO), 1, -1
                if (S%occ(F%av_occ(j)%MO(l)) .gt. 0) then
                  S%occ_down(1, k) = F%av_occ(j)%MO(l)
                  exit av_loop_A1
                end if
              end do
            end if
          end do av_loop_A1

          av_loop_A2: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_down(2, k))) then
              do l = size(F%av_occ(j)%MO), 1, -1
                if ((S%occ(F%av_occ(j)%MO(l)) .gt. 0 .and. F%av_occ(j)%MO(l) .ne. S%occ_down(1, k)) &
                  .or. (S%occ(F%av_occ(j)%MO(l)) .gt. 1)) then
                  S%occ_down(2, k) = F%av_occ(j)%MO(l)
                  exit av_loop_A2
                end if
              end do
            end if
          end do av_loop_A2

          av_loop_A3: do j = 1, size(F%av_occ)
            if (ANY(F%av_occ(j)%MO .eq. S%occ_up(1, k))) then
              do l = 1, size(F%av_occ(j)%MO), 1
                if (S%occ(F%av_occ(j)%MO(l)) .lt. 2) then
                  S%occ_up(1, k) = F%av_occ(j)%MO(l)
                  exit av_loop_A3
                end if
              end do
            end if
          end do av_loop_A3
        end select
      end if

    end subroutine make_equivalent_configuration

  end subroutine add_Process

  subroutine sum_equivalent_configuration(S)
    !! sum up process data over equivalent configurations
    type(Process_data), intent(inout):: S
    integer :: i, j
    real(long) :: sum_E_eV
    integer :: n_equiv
    integer :: imultp_i, jmultp_i
    integer :: imultp_f, jmultp_f
    integer :: icsf_num, jcsf_num
    imultp_i = -1
    jmultp_i = -1
    imultp_f = -1
    jmultp_f = -1
    icsf_num = -1
    jcsf_num = -1
    do i = 1, S%N_process
      sum_E_eV = S%E_eV(i)
      n_equiv = 1
      if (allocated(S%initial_spin_multp)) then
        imultp_i = S%initial_spin_multp(i)
        imultp_f = S%final_spin_multp(i)
        icsf_num = S%initial_csf_num(i)
      end if
      do j = i + 1, S%N_process
        if (allocated(S%initial_spin_multp)) then
          jmultp_i = S%initial_spin_multp(j)
          jmultp_f = S%final_spin_multp(j)
          jcsf_num = S%initial_csf_num(j)
        end if
        if (S%tag(i) .eq. S%tag(j) .and. S%tag(j) .ne. "X" .and. ALL(S%occ_down(:, i) .eq. S%occ_down(:, j)) &
          .and. ALL(S%occ_up(:, i) .eq. S%occ_up(:, j)) .and. imultp_i .eq. jmultp_i .and. imultp_f .eq. jmultp_f) then
          n_equiv = n_equiv + 1
          ! sum up all transition rates
          S%value(i) = S%value(i) + S%value(j)
          sum_E_eV = sum_E_eV + ONE*S%E_eV(j)
          S%tag(j) = "X"
          S%value(j) = ZERO
        end if
      end do
      ! average the transition energy
      S%E_eV(i) = nint(ONE*sum_E_eV/n_equiv)
    end do
  end subroutine sum_equivalent_configuration

  function create_Process(occ, N_proc_max, spin, ci) result(S)
    !! creates process structure
    type(Process_data) :: S
    integer, intent(in) :: occ(:), N_proc_max
    logical, intent(in), optional :: spin
    logical, intent(in), optional :: ci
    logical :: yes_spin, yes_CI
    yes_spin = .FALSE.
    yes_CI = .FALSE.
    if (present(spin)) yes_spin = spin
    if (present(ci)) yes_ci = ci

    !type(Config_tree), intent(in), target :: Q0
    S%N_process = 0       ! it will be increased whenever proc. is added
    allocate (S%occ(size(occ)))
    allocate (S%tag(N_proc_max), S%value(N_proc_max), S%E_eV(N_proc_max))
    allocate (S%occ_up(1, N_proc_max), S%occ_down(3, N_proc_max))
    if (yes_spin) then
      allocate (S%initial_occ(size(occ), N_proc_max))
      allocate (S%initial_csf_num(N_proc_max))
      allocate (S%initial_spin_multp(N_proc_max))
      allocate (S%final_csf_num(N_proc_max))
      allocate (S%final_spin_multp(N_proc_max))
      S%initial_csf_num(:) = 0
      S%final_csf_num(:) = 0
      S%initial_spin_multp(:) = 0
      S%final_spin_multp(:) = 0
    end if
    if (yes_ci) then
      allocate (S%initial_spin_multp(N_proc_max))
      allocate (S%final_spin_multp(N_proc_max))
      allocate (S%initial_irrep(N_proc_max))
      allocate (S%final_irrep(N_proc_max))
      allocate (S%initial_state_num(N_proc_max))
      allocate (S%final_state_num(N_proc_max))
      allocate (S%E_au(N_proc_max))
      S%initial_spin_multp(:) = 0
      S%final_spin_multp(:) = 0
      S%initial_irrep(:) = 0
      S%final_irrep(:) = 0
      S%initial_state_num(:) = 0
      S%final_state_num(:) = 0
      S%E_au(:) = ZERO
    end if
    S%value(:) = ZERO
    S%occ(:) = occ(:)
    S%occ_up(:, :) = 0
    S%occ_down(:, :) = 0
  end function create_Process

  subroutine purge_Process(S)
    !! purges process data
    type(Process_data), intent(inout) :: S
    if (allocated(S%occ)) deallocate (S%occ)
    if (allocated(S%tag)) deallocate (S%tag)
    if (allocated(S%value)) deallocate (S%value)
    if (allocated(S%E_eV)) deallocate (S%E_eV)
    if (allocated(S%occ_up)) deallocate (S%occ_up)
    if (allocated(S%occ_down)) deallocate (S%occ_down)
    if (allocated(S%initial_occ)) deallocate (S%initial_occ)
    if (allocated(S%final_csf_num)) deallocate (S%final_csf_num)
    if (allocated(S%initial_csf_num)) deallocate (S%initial_csf_num)
    if (allocated(S%final_spin_multp)) deallocate (S%final_spin_multp)
    if (allocated(S%initial_spin_multp)) deallocate (S%initial_spin_multp)
    if (allocated(S%final_state_num)) deallocate (S%final_state_num)
    if (allocated(S%initial_state_num)) deallocate (S%initial_state_num)
    if (allocated(S%initial_irrep)) deallocate (S%initial_irrep)
    if (allocated(S%final_irrep)) deallocate (S%final_irrep)
    if (allocated(S%E_au)) deallocate (S%E_au)
  end subroutine purge_Process

  subroutine print_Process(S, OUTPUT)
    type(Process_data), intent(in) :: S
    integer, intent(in) :: OUTPUT
    integer :: k
    !  integer*8 :: i_initial, i_final
    integer :: i_initial, i_final
    character(len=1000) :: s1, s2, s3, s4
    character(len=2010) :: description
    integer :: occ_final(size(S%occ))

    write (s1, "(1000i1)") S%occ(:)
    call occ_to_index(S%occ, i_initial)
    do k = 1, S%N_process
      if (allocated(S%initial_occ)) then
        write (s1, "(1000i1)") S%initial_occ(:, k)
      end if
      if (S%tag(k) .eq. "X") cycle
      !     if(S%value(k)==ZERO)cycle
      occ_final(:) = occ_process(S, k)

      call occ_to_index(occ_final, i_final)
      write (s2, "(1000i1)") occ_final(:)
      write (s3, "(2i12)") i_initial, i_final
      write (s4, "(es24.7)") S%value(k)
      if (allocated(S%final_spin_multp)) then
        write( description, "(' # ',a,i3,i3,' -> ',a,i3,i3)" ) trim(s1), S%initial_spin_multp(k),S%initial_csf_num(k),trim(s2),S%final_spin_multp(k),S%final_csf_num(k)
      else
        write (description, "(' # ',a,' -> ',a)") trim(s1), trim(s2)
      end if
      write (OUTPUT, "(x,a,a24,a24,x,i6,a)") &
        S%tag(k), trim(s3), s4, S%E_eV(k), trim(description)
    end do

  end subroutine print_Process

  function occ_Process(S, k) result(occ_final)
    !! returns the final occupation numbers for a specific
    type(Process_data), intent(in) :: S
    integer, intent(in) :: k
    integer :: occ_final(size(S%occ))
    integer :: i, j
    occ_final(:) = S%occ(:)
    if (allocated(S%initial_occ)) then
      occ_final(:) = S%initial_occ(:, k)
    else
      occ_final(:) = S%occ(:)
    end if
    !
    ! checking a subshell for decreasing occupation #
    !
    do j = 1, size(S%occ_down, dim=1)
      i = S%occ_down(j, k)
      if (i .ne. 0) occ_final(i) = occ_final(i) - 1
    end do
    !
    ! checking a subshell for increasing occupation #
    !
    do j = 1, size(S%occ_up, dim=1)
      i = S%occ_up(j, k)
      if (i .ne. 0) then
        occ_final(i) = occ_final(i) + 1
      end if
    end do

  end function occ_Process

end module process
