!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

program project_gto_basis
  !! author: Ludger Inhester
  !! small tool to project molecular orbital coefficients from one GTO basis set into another
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use wavefunction
  use gaussian
  use rwfile
  use constants
  use stdio
  use gps
  use interpolate
  use eigen_lapack
  use matrix
  use packed_index
  implicit none

  type(Param) :: P
  type(Atom_set), allocatable :: A(:)
  type(Grid3D) :: G
  type(Basis) :: B1
  type(Basis) :: B2
  type(WF_data) :: F1
  type(WF_data) :: F2

  real(long), allocatable :: env12(:)
  integer, allocatable :: atm12(:, :)
  integer, allocatable :: bas12(:, :)
  integer, allocatable :: first_bf_Of_Shell12(:)
  integer, allocatable :: shellInSet(:)
  integer, allocatable :: bfInSet(:)

  integer :: sh, bf
  integer :: nJoint
  real(long), allocatable :: Sjoint(:, :)
  real(long), allocatable :: S1(:, :)
  real(long), allocatable :: M(:, :)
  real(long), allocatable :: M2(:, :)
  real(long), allocatable :: S2(:, :)
  real(long), allocatable :: S21(:, :)
  integer, allocatable :: ind1(:), ind2(:)
  integer :: j

  call read_parameters(P, "static")
  call print_parameters(P, "electronic structure of molecules")
  !
  ! construct molecular coordinate
  !
  call init_index_packed()
  call construct_Atom_set(P, A)
  call construct_Grid3D(P, A, G)

  ! make here a joint basis construction!
  call construct_Basis(A, G, B1, &
                       P%EPS_overlap, &
                       P%EPS_overlap_single, &
                       yes_sym=P%yes_sym, &
                       gto_file_path=P%gto_file_path, &
                       yes_gradient=.False., &
                       basisSetOverlapLim=P%basisSetOverlapLim)
  if (P%yes_verbose) &
    call print_calc_info(P, A, G, B1)
  call construct_WF_data(P, A, G, B1, F1)
  call read_wf(F1, F1%LC_Ini, P%wfFileName)

  call construct_Basis(A, G, B2, &
                       P%EPS_overlap, &
                       P%EPS_overlap_single, &
                       yes_sym=P%yes_sym, &
                       gto_file_path=P%gto_file_path2, &
                       yes_gradient=.False., &
                       basisSetOverlapLim=P%basisSetOverlapLim)
  if (P%yes_verbose) &
    call print_calc_info(P, A, G, B2)
  call construct_WF_data(P, A, G, B2, F2)

  call construct_Gaussian_basis_set(A, env12, atm12, bas12, &
                                    filename=P%gto_file_path, filename2=P%gto_file_path2, shellInSet=shellInSet)
  allocate (first_bf_of_shell12(size(bas12, 2)))
  nJoint = sum(2*bas12(2, :) + 1)
  allocate (bfInSet(nJoint))
  bfInSet = 0
  bf = 1
  do sh = 1, size(bas12, 2) ! loop over shells
    first_bf_of_shell12(sh) = bf
    bfInSet(bf:bf + 2*bas12(2, sh)) = shellInSet(sh)
    bf = bf + 2*bas12(2, sh) + 1
  end do

  allocate (Sjoint(nJoint, nJoint))
  call calculate_overlap_Gaussian_basis_set(env12, atm12, bas12, first_bf_of_shell12, Sjoint)

  allocate (ind1(count((/bfInset .eq. 1/))))
  allocate (ind2(count((/bfInset .eq. 2/))))
  ind1 = pack((/(j, j=1, size(bfInset, 1))/), bfInset .eq. 1)
  ind2 = pack((/(j, j=1, size(bfInset, 1))/), bfInset .eq. 2)
  allocate (S21(size(ind2), size(ind1)))
  allocate (S1(size(ind1), size(ind1)))
  allocate (S2(size(ind2), size(ind2)))
  S21 = Sjoint(ind2(:), ind1(:))
  S1 = Sjoint(ind1(:), ind1(:))
  S2 = Sjoint(ind2(:), ind2(:))

  ! projection from basis 1 to basis 2 is
  ! F2%LC  = S2^{-1} .m. S21 .m. F1%LC_ini

  !call print_matrix(F1%LC)
  !call print_matrix(S21.m.F1%LC)
  !call print_matrix(B2%SInv.m.S21.m.F1%LC_ini)
  allocate (M(size(S21, 1), size(F1%LC_ini, 2)))
  !call print_matrix(F1%LC_ini)
  !call print_matrix(S21)
  !call print_matrix(S21.m.F1%LC_ini)
  M = S21.m.F1%LC_ini
  allocate (M2(size(B2%Sinv, 1), size(M, 2)))
  M2 = B2%SInv.m.M
  F2%LC(:, 1:size(M2, 2)) = M2
  !F2%LC = B2%SInv.m.(S21.m.F1%LC_ini)
  ! orthonormalize orbitals (and fill up possible remaining orbitals)
  call orthonormalizeOrbitals(B2, F2%LC)

  call save_wf(F2%LC, "wf2.dat")

end program project_gto_basis
