/*
 XMOLECULE electronic structure toolkit
 Copyright 2023 Ludger Inhester
 Ludger Inhester, Center for Free-Electron Laser Science, DESY
 ludger.inhester@cfel.de

 This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
 See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
*/

#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
//! Python-xmolecule interface
//! author: Ludger Inhester

#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>
#include <numpy/arrayobject.h>

#include "procmem.h"
#include "python_xmolecule_interface.h"
#include "xmolecule_interface.h"

#if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
// Routines from libifcore
void for_rtl_init_(int *, char **);
int for_rtl_finish_();
// Wrappers
#define fc_init(argc, argv) \
  for_rtl_init_(&argc, argv)
#define fc_finish() \
  for_rtl_finish_()
#elif defined(__GNUC__)
// Routines from libgfortran
void _gfortran_set_args(int, char **);
// Wrappers
#define fc_init(argc, argv) \
  _gfortran_set_args(argc, argv)
#define fc_finish()

#endif

// maximal length of inputfilepath
#define LEN_INPUTFILENAME 1024
// maximal length of errormessage in exception
#define LEN_ERRMSG 512

/*
   xmol object member function
*/

static void
xmol_dealloc(xmol *self)
{
  xmolecule_purge(self->xmol_data);
  free(self->inputfile);
  free(self->xmol_data);
  PyTypeObject *tp = Py_TYPE(self);
  tp->tp_free(self);
  Py_DECREF(tp);
}

static int
xmol_setup(xmol *self, PyObject *args, PyObject *kwds)
{
  int argc;
  char **argv;
  char default_inputfile[] = "input.in";
  char *argInputfile = NULL;
  static char *kwlist[] = {"input", NULL};
  int len = 0;
  self->xmol_data = malloc(sizeof(void *));
  PyArg_ParseTupleAndKeywords(args, kwds, "|s", kwlist, &argInputfile);
  if (argInputfile == NULL)
  {
    len = strlen(default_inputfile) + 1;
    self->inputfile = malloc(sizeof(char) * len);
    strncpy(self->inputfile, default_inputfile, len);
  }
  else
  {
    len = (strlen(argInputfile) + 1);
    self->inputfile = malloc(sizeof(char) * len);
    strncpy(self->inputfile, argInputfile, len);
  }
  argc = 1;
  argv = malloc(sizeof(char *) * argc);
  argv[0] = "python_xmolecule";
  fc_init(argc, argv);
  len = len - 1;
  xmolecule_setup(self->xmol_data, self->inputfile, &len);
  fc_finish();
  free(argv);
  return (0);
}

static PyObject *
xmol_init(xmol *self, PyObject *args)
{
  xmolecule_init(self->xmol_data);
  self->natom = 0;
  xmolecule_get_geom(self->xmol_data, NULL, &(self->natom));
  self->locc = 0;
  xmolecule_get_occ(self->xmol_data, NULL, &(self->locc));
  self->nmo = 0;
  self->nao = 0;
  xmolecule_get_mo_coef(self->xmol_data, NULL, &(self->nmo), &(self->nao));
  xmolecule_get_energy(self->xmol_data, &self->energy);
  self->nstates = 1;
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_read_chkpt(xmol *self, PyObject *args)
{
  xmolecule_read_chkpt(self->xmol_data);
  self->natom = 0;
  xmolecule_get_geom(self->xmol_data, NULL, &(self->natom));
  self->locc = 0;
  xmolecule_get_occ(self->xmol_data, NULL, &(self->locc));
  self->nmo = 0;
  self->nao = 0;
  xmolecule_get_mo_coef(self->xmol_data, NULL, &(self->nmo), &(self->nao));
  xmolecule_get_energy(self->xmol_data, &self->energy);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_write_density_cube(xmol *self, PyObject *args)
{
  xmolecule_write_density_cube(self->xmol_data);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_write_molden(xmol *self, PyObject *args, PyObject *kwds)
{
  char default_outputfile[] = "out.molden";
  static char *kwlist[] = {"file", NULL};
  char *argOutputfile = NULL;
  int len;
  PyArg_ParseTupleAndKeywords(args, kwds, "|s", kwlist, &argOutputfile);
  if (argOutputfile == NULL)
    argOutputfile = default_outputfile;
  len = strlen(argOutputfile);
  xmolecule_write_molden(self->xmol_data, argOutputfile, &len);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_write_mo_cube(xmol *self, PyObject *args)
{
  int mo;
  PyArg_ParseTuple(args, "i", &mo);
  xmolecule_write_mo_cube(self->xmol_data, &mo);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_write_chkpt(xmol *self, PyObject *args)
{
  xmolecule_write_chkpt(self->xmol_data);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_getstate(xmol *self, PyObject *args)
{
  int res;
  char errmsg[LEN_ERRMSG];
  xmolecule_write_chkpt(self->xmol_data);
  FILE *datafile = fopen("chkpt.dat", "r");
  if (datafile == NULL)
  {
    PyErr_SetString(PyExc_Exception, "Cannot open file chkpt.dat");
    return NULL;
  }
  fseek(datafile, 0, SEEK_END);
  long len = ftell(datafile);
  char *data = malloc((len + 1) * sizeof(char));
  fseek(datafile, 0, SEEK_SET);
  res = fread(data, sizeof(char), len, datafile);
  fclose(datafile);
  if (res == 0)
  {
    free(data);
    PyErr_SetString(PyExc_Exception, "Error reading chkpt.dat");
    return NULL;
  }
  data[len] = '\0';

  FILE *inputfile = fopen(self->inputfile, "r");
  if (inputfile == NULL)
  {
    free(data);
    snprintf(errmsg, LEN_ERRMSG, "Error opening %s", self->inputfile);
    PyErr_SetString(PyExc_Exception, errmsg);
    return NULL;
  }
  fseek(inputfile, 0, SEEK_END);
  long ilen = ftell(inputfile);
  char *idata = malloc((ilen + 1) * sizeof(char));
  fseek(inputfile, 0, SEEK_SET);
  res = fread(idata, sizeof(char), ilen, inputfile);
  fclose(inputfile);
  if (res == 0)
  {
    free(data);
    free(idata);
    snprintf(errmsg, LEN_ERRMSG, "Error reading file %s", self->inputfile);
    PyErr_SetString(PyExc_Exception, errmsg);
    return NULL;
  }
  idata[ilen] = '\0';

  PyObject *stateData = PyByteArray_FromStringAndSize(data, len);
  PyObject *inputData = PyByteArray_FromStringAndSize(idata, ilen);
  free(data);
  free(idata);

  return Py_BuildValue("(O,O)", stateData, inputData);
}

static PyObject *
xmol_setstate(xmol *self, PyObject *args)
{
  PyObject *stateData;
  PyObject *inputData;
  char errmsg[LEN_ERRMSG];
  char *inputfilePath = "restoreinput.in";
  char *chkptfilePath = "chkpt.dat";
  Py_buffer buffer;
  int retval = PyArg_ParseTuple(args, "(OO)", &stateData, &inputData);
  if (!retval)
  {
    PyErr_SetString(PyExc_ValueError, "Bad arguments");
    return NULL;
  }
  PyObject_GetBuffer(stateData, &buffer, PyBUF_C_CONTIGUOUS);
  FILE *datafile = fopen(chkptfilePath, "w");
  if (datafile != NULL)
  {
    fwrite(buffer.buf, 1, buffer.len, datafile);
    fclose(datafile);
  }
  else
  {
    snprintf(errmsg, LEN_ERRMSG, "Cannot open file %s", chkptfilePath);
    PyErr_SetString(PyExc_Exception, errmsg);
    return NULL;
  }
  PyBuffer_Release(&buffer);

  PyObject_GetBuffer(inputData, &buffer, PyBUF_C_CONTIGUOUS);
  FILE *inputfile = fopen(inputfilePath, "w");
  if (inputfile != NULL)
  {
    fwrite(buffer.buf, 1, buffer.len, inputfile);
    fclose(inputfile);
  }
  else
  {
    snprintf(errmsg, LEN_ERRMSG, "Cannot open file %s", inputfilePath);
    PyErr_SetString(PyExc_Exception, errmsg);
    return NULL;
  }
  PyBuffer_Release(&buffer);
  self->inputfile = inputfilePath;
  PyObject *kwlist = Py_BuildValue("{s:s}", "input", inputfilePath);
  PyObject *setup_args = Py_BuildValue("()");
  xmol_setup(self, setup_args, kwlist);
  Py_DECREF(kwlist);
  Py_DECREF(setup_args);

  PyObject *ret = xmol_read_chkpt(self, NULL);
  return ret;
}

static PyObject *
xmol_hole_occs_energies_gradients_nacs(xmol *self, PyObject *args)
{
  int *occs;
  double *energies;
  double *gradients;
  double *nacs;
  int N_exc = self->locc;
  long int occdim[] = {self->locc, N_exc};
  long int endim[] = {N_exc};
  long int graddim[] = {3 * self->natom, N_exc};
  long int nacdim[] = {3 * self->natom, N_exc, N_exc};
  PyArrayObject *en_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  energies = PyArray_DATA(en_arr);
  PyArrayObject *occ_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                           2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  occs = PyArray_DATA(occ_arr2);
  PyArrayObject *grad_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type,
                           PyArray_DescrFromType(NPY_DOUBLE),
                           2, graddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  gradients = PyArray_DATA(grad_arr2);
  PyArrayObject *nac_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           3, nacdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  nacs = PyArray_DATA(nac_arr2);

  xmolecule_hole_occs_energies_gradients_nacs(self->xmol_data, occs,
                                              energies, gradients, nacs,
                                              &N_exc, &(self->natom),
                                              &(self->locc));
  if (N_exc < 0)
  {
    N_exc = -N_exc;
    self->nstates = N_exc;
    Py_CLEAR(en_arr);
    Py_CLEAR(occ_arr2);
    Py_CLEAR(grad_arr2);
    Py_CLEAR(nac_arr2);
    occdim[1] = N_exc;
    endim[0] = N_exc;
    graddim[1] = N_exc;
    nacdim[1] = N_exc;
    nacdim[2] = N_exc;
    en_arr = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    energies = (double *)PyArray_DATA(en_arr);
    occ_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                             2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    occs = (int *)PyArray_DATA(occ_arr2);
    grad_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             2, graddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    gradients = (double *)PyArray_DATA(grad_arr2);
    nac_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             3, nacdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    nacs = (double *)PyArray_DATA(nac_arr2);
    xmolecule_hole_occs_energies_gradients_nacs(self->xmol_data, occs,
                                                energies, gradients, nacs,
                                                &N_exc, &(self->natom),
                                                &(self->locc));
  }
  PyArrayObject *occ_arr = (PyArrayObject *)PyArray_SwapAxes(occ_arr2, 0, 1);
  Py_CLEAR(occ_arr2);
  PyArrayObject *grad_arr = (PyArrayObject *)PyArray_SwapAxes(grad_arr2, 0, 1);
  Py_CLEAR(grad_arr2);
  PyArrayObject *nac_arr = (PyArrayObject *)PyArray_SwapAxes(nac_arr2, 0, 2);
  Py_CLEAR(nac_arr2);
  PyObject *ret = Py_BuildValue("(N,N,N,N)", occ_arr, en_arr, grad_arr, nac_arr);
  return (ret);
}

static PyObject *
xmol_part_occs_energies_gradients_nacs(xmol *self, PyObject *args)
{
  int *occs;
  double *energies;
  double *gradients;
  double *nacs;
  int N_exc = self->nstates;
  long int occdim[] = {self->locc, N_exc};
  long int endim[] = {N_exc};
  long int graddim[] = {3 * self->natom, N_exc};
  long int nacdim[] = {3 * self->natom, N_exc, N_exc};
  PyArrayObject *en_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  energies = PyArray_DATA(en_arr);
  PyArrayObject *occ_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                           2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  occs = PyArray_DATA(occ_arr2);
  PyArrayObject *grad_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type,
                           PyArray_DescrFromType(NPY_DOUBLE),
                           2, graddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  gradients = PyArray_DATA(grad_arr2);
  PyArrayObject *nac_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           3, nacdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  nacs = PyArray_DATA(nac_arr2);
  xmolecule_part_occs_energies_gradients_nacs(self->xmol_data, occs,
                                              energies, gradients, nacs,
                                              &N_exc, &(self->natom),
                                              &(self->locc));
  if (N_exc < 0)
  {
    N_exc = -N_exc;
    self->nstates = N_exc;
    Py_CLEAR(en_arr);
    Py_CLEAR(occ_arr2);
    Py_CLEAR(grad_arr2);
    Py_CLEAR(nac_arr2);
    occdim[1] = N_exc;
    endim[0] = N_exc;
    graddim[1] = N_exc;
    nacdim[1] = N_exc;
    nacdim[2] = N_exc;
    en_arr = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    energies = (double *)PyArray_DATA(en_arr);
    occ_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                             2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    occs = (int *)PyArray_DATA(occ_arr2);
    grad_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             2, graddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    gradients = (double *)PyArray_DATA(grad_arr2);
    nac_arr2 = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             3, nacdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    nacs = (double *)PyArray_DATA(nac_arr2);
    xmolecule_part_occs_energies_gradients_nacs(self->xmol_data, occs,
                                                energies, gradients, nacs,
                                                &N_exc, &(self->natom),
                                                &(self->locc));
  }
  PyArrayObject *occ_arr = (PyArrayObject *)PyArray_SwapAxes(occ_arr2, 0, 1);
  Py_CLEAR(occ_arr2);
  PyArrayObject *grad_arr = (PyArrayObject *)PyArray_SwapAxes(grad_arr2, 0, 1);
  Py_CLEAR(grad_arr2);
  PyArrayObject *nac_arr = (PyArrayObject *)PyArray_SwapAxes(nac_arr2, 0, 2);
  Py_CLEAR(nac_arr2);
  PyObject *ret = Py_BuildValue("(N,N,N,N)", occ_arr, en_arr, grad_arr, nac_arr);
  return (ret);
}

// TODO: rewrite using Pyarrays
static PyObject *
xmol_hole_occs_energies_efields(xmol *self, PyObject *args)
{
  int i, j;
  int *occs;
  double *energies;
  double *efields;
  double *upots;
  double pos[3];
  int N_exc = self->nstates;
  int ret;

  PyObject *PyPos;
  ret = PyArg_ParseTuple(args, "O", &PyPos);
  if (!ret)
  {
    PyErr_SetString(PyExc_TypeError, "Bad arguments");
    return NULL;
  }
  if (PyList_Check(PyPos))
  {
    if (PyList_Size(PyPos) != 3)
    {
      PyErr_SetString(PyExc_TypeError, "List must have length 3");
      return NULL;
    }
    // argument is a list
    for (i = 0; i < 3; i++)
    {
      PyObject *item;
      item = PyList_GetItem(PyPos, i);
      pos[i] = PyFloat_AsDouble(item);
    }
  }
  else if (PyArray_CheckExact(PyPos))
  {
    if (PyArray_NDIM((PyArrayObject *)PyPos) != 1 && PyArray_SHAPE((PyArrayObject *)PyPos)[0] != 3)
    {
      PyErr_SetString(PyExc_TypeError, "Array must have shape (3)");
      return NULL;
    }
    // argument is an array
    for (i = 0; i < 3; i++)
    {
      pos[i] = *((double *)PyArray_GETPTR1((PyArrayObject *)PyPos, i));
    }
  }
  else
  {
    PyErr_SetString(PyExc_TypeError, "Argument must be either list or array");
    return NULL;
  }
  N_exc = self->nstates;
  long int occdim[] = {self->locc, N_exc};
  long int endim[] = {N_exc};
  long int efielddim[] = {3, N_exc};
  long int upotdim[] = {N_exc};

  //occs = malloc(sizeof(int) * self->locc * N_exc);
  //energies = malloc(sizeof(double) * N_exc);
  //efields = malloc(sizeof(double) * 3 * N_exc);
  //upots = malloc(sizeof(double) * N_exc);
  PyArrayObject *occ_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                           2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  occs = (int *)PyArray_DATA(occ_arr2);
  PyArrayObject *en_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  energies = (double *)PyArray_DATA(en_arr);
  PyArrayObject *efield_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             2, efielddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  efields = PyArray_DATA(efield_arr2);
  PyArrayObject * upot_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, upotdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  upots = PyArray_DATA(upot_arr);
  xmolecule_hole_occs_energies_efield(self->xmol_data, pos, occs, energies,
                                      efields, upots, &N_exc, &(self->locc));
  if (N_exc < 0)
  {
    N_exc = -N_exc;
    self->nstates = N_exc;
    Py_CLEAR(en_arr);
    Py_CLEAR(occ_arr2);
    Py_CLEAR(efield_arr2);
    Py_CLEAR(upot_arr);
    occdim[1] = N_exc;
    endim[0] = N_exc;
    efielddim[1] = N_exc;
    upotdim[0] = N_exc;
    occ_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                           2, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    occs = (int *)PyArray_DATA(occ_arr2);
    en_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             1, endim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    energies = (double *)PyArray_DATA(en_arr);
    efield_arr2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             2, efielddim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    efields = PyArray_DATA(efield_arr2);
    upot_arr = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, upotdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    upots = PyArray_DATA(upot_arr);
    xmolecule_hole_occs_energies_efield(self->xmol_data, pos, occs,
                                        energies, efields, upots, &N_exc,
                                        &(self->locc));
  }
  PyArrayObject *occ_arr = (PyArrayObject *)PyArray_SwapAxes(occ_arr2, 0, 1);
  Py_CLEAR(occ_arr2);
  PyArrayObject *efield_arr = (PyArrayObject *)PyArray_SwapAxes(efield_arr2, 0, 1);
  Py_CLEAR(efield_arr2);
  PyObject *retval = Py_BuildValue("(N,N,N,N)", occ_arr, en_arr, efield_arr, upot_arr);
  return (retval);
}

static PyObject *
xmol_CIS_transdens(xmol *self, PyObject *Ostate)
{
  int state;
  long int AOdims[2];
  AOdims[0] = self->nao;
  AOdims[1] = self->nao;
  PyArg_ParseTuple(Ostate, "i", &state);
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  PyArrayObject *transdens = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           2, AOdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  double *tdens = (double *)PyArray_DATA(transdens);
  xmolecule_get_cis_transdens(self->xmol_data, tdens, &state, &self->nao);
  return (PyObject *)transdens;
}

static PyObject *
xmol_CIS_energy_gradient_nacs(xmol *self, PyObject *Ostate)
{
  int state=0;
  double *gradient, *nacs, *energies;
  PyArg_ParseTuple(Ostate, "i", &state);
  xmolecule_get_energy(self->xmol_data, &self->energy);
  long int edims[1];
  edims[0] = self->nstates;
  PyArrayObject *en = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, edims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  energies = (double *)PyArray_DATA(en);
  xmolecule_cis_energies(self->xmol_data, energies, &self->nstates);
  if (self->nstates < 0)
  {
    self->nstates *= -1;
    edims[0] = self->nstates;
    en = (PyArrayObject *)
        PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                             1, edims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    energies = (double *)PyArray_DATA(en);
    xmolecule_cis_energies(self->xmol_data, energies, &self->nstates);
  }
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  long int gdims[1];
  gdims[0] = 3 * self->natom;
  PyArrayObject *grad_array = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, gdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  gradient = (double *)PyArray_DATA(grad_array);
  long int dims[3];
  dims[0] = 3 * self->natom;
  dims[1] = self->nstates;
  dims[2] = self->nstates;
  PyArrayObject *nac_array2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           3, dims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  nacs = PyArray_DATA(nac_array2);
  xmolecule_get_cis_gradient_nacs(self->xmol_data, &state, gradient,
                                  &(self->natom), nacs, &(self->nstates));
  // returned array shoud have the shape: nstates, nstates, 3*natoms
  // -> swap axes 0 and 2
  PyArrayObject *nac_array = (PyArrayObject *)PyArray_SwapAxes(nac_array2, 0, 2);
  Py_CLEAR(nac_array2);
  PyObject *ret = Py_BuildValue("(N,N,N)", en, grad_array, nac_array);
  return (ret);
}

static PyObject *
xmol_CIS_energy_gradient(xmol *self, PyObject *Ostate)
{
  int state;
  double *gradient;
  state = PyLong_AsLong(Ostate);
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  long int gdims[1];
  gdims[0] = 3 * self->natom;
  PyArrayObject *grad_array = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, gdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  gradient = (double *)PyArray_DATA(grad_array);
  xmolecule_get_cis_gradient(self->xmol_data, &state, gradient,
                             &(self->natom));
  return ((PyObject *)grad_array);
}

static PyObject *
xmol_gradient(xmol *self, PyObject *args)
{
  double *gradient;
  long int gdims[1];
  gdims[0] = 3 * self->natom;
  PyArrayObject *grad_array = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, gdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  gradient = (double *)PyArray_DATA(grad_array);
  xmolecule_get_gradient(self->xmol_data, gradient, &(self->natom));
  return ((PyObject *)grad_array);
}

static PyObject *
xmol_get_geom(xmol *self, PyObject *args)
{
  double *geom;
  long int dims[] = {3 * self->natom};
  PyArrayObject *py_geom = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, dims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  geom = PyArray_DATA(py_geom);
  xmolecule_get_geom(self->xmol_data, geom, &(self->natom));
  return (PyObject *)py_geom;
}

// TODO: rewrite using pyarray
static PyObject *
xmol_excitation_energies(xmol *self, PyObject *args)
{
#define MAX_EXC 2000
  int N_exc = MAX_EXC;
  int i, j;
  int *f_occs;
  double *E;
  f_occs = malloc(sizeof(int) * self->locc * N_exc);
  E = malloc(sizeof(double) * N_exc);
  xmolecule_excitation_energies(self->xmol_data, f_occs, E, &N_exc,
                                &(self->locc));
  if (N_exc > MAX_EXC)
  {
    f_occs = realloc(f_occs, sizeof(int) * self->locc * N_exc);
    E = realloc(E, sizeof(double) * N_exc);
    xmolecule_excitation_energies(self->xmol_data, f_occs, E, &N_exc,
                                  &(self->locc));
  }
  PyObject *list = PyList_New(N_exc);
  for (i = 0; i < N_exc; i++)
  {
    PyObject *final_occ = PyList_New(self->locc);
    for (j = 0; j < self->locc; j++)
    {
      PyList_SetItem(final_occ, j,
                     Py_BuildValue("i", f_occs[self->locc * i + j]));
    }

    PyObject *entry;
    entry = Py_BuildValue("(O,d)", final_occ, E[i]);
    PyList_SetItem(list, i, entry);
  }
  free(E);
  free(f_occs);
  return list;
}

// TODO: rewrite using Pyarray
static PyObject *
xmol_CIS_energies(xmol *self, PyObject *state_obj)
{
  int i;
  double *E;
  E = malloc(sizeof(double) * self->nstates);
  xmolecule_cis_energies(self->xmol_data, E, &self->nstates);
  if (self->nstates < 0)
  {
    self->nstates *= -1;
    E = realloc(E, sizeof(double) * self->nstates);
    xmolecule_cis_energies(self->xmol_data, E, &self->nstates);
  }
  PyObject *list = PyList_New(self->nstates);
  for (i = 0; i < self->nstates; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", E[i]));
  }
  free(E);
  return list;
}

static PyObject *
xmol_occ2cistate(xmol *self, PyObject *args)
{
  int *occ;
  int i;
  int numLines = PyList_Size(args);
  if (numLines != self->locc)
  {
    PyErr_SetString(PyExc_ValueError, "Wrong len of occ list");
    return NULL;
  }
  occ = malloc(sizeof(int) * self->locc);
  for (i = 0; i < self->locc; i++)
  {
    PyObject *item;
    item = PyList_GetItem(args, i);
    occ[i] = PyLong_AsLong(item);
  }

  int state;
  occ2cistate(self->xmol_data, occ, &self->locc, &state);
  free(occ);
  return (Py_BuildValue("i", state - 1));
}

static PyObject *
xmol_cistate2occ(xmol *self, PyObject *state_arg)
{
  int *occ;
  int i;
  int state;
  state = (int)PyLong_AsLong(state_arg) + 1;
  if (state < 1 || state > self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  occ = malloc(sizeof(int) * self->locc);
  cistate2occ(self->xmol_data, &state, occ, &self->locc);

  PyObject *list = PyList_New(self->locc);
  for (i = 0; i < self->locc; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("i", occ[i]));
  }
  free(occ);
  return list;
}

static PyObject *
xmol_CI_energies(xmol *self, PyObject *args)
{
  int i, state_given, state;
  double *E;
  state_given = PyArg_ParseTuple(args, "i", &state);

  E = malloc(sizeof(double) * self->nstates);
  xmolecule_ci_energies(self->xmol_data, E, &self->nstates);
  if (self->nstates < 0)
  {
    self->nstates *= -1;
    E = realloc(E, sizeof(double) * self->nstates);
    xmolecule_ci_energies(self->xmol_data, E, &self->nstates);
  }
  if (state_given)
  {
    state++;
    if (state < 1 || state > self->nstates)
    {
      PyErr_SetString(PyExc_ValueError, "Bad state index");
      return NULL;
    }
    xmolecule_set_ci_state(self->xmol_data, &state);
  }
  PyObject *list = PyList_New(self->nstates);
  for (i = 0; i < self->nstates; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", E[i]));
  }
  free(E);
  return (PyObject *)list;
}

static PyObject *
xmol_CI_Auger(xmol *self, PyObject *args)
{
#define MAX_PROCESS 10000
  static int N_process = MAX_PROCESS;
  int N_process1;
  int i;
  int *f_irrep, *f_multp, *f_index;
  double *rate;
  double *energy;

  f_irrep = calloc(sizeof(int), N_process);
  f_multp = calloc(sizeof(int), N_process);
  f_index = calloc(sizeof(int), N_process);
  energy = calloc(sizeof(double), N_process);
  rate = calloc(sizeof(double), N_process);
  N_process1 = N_process;
  xmolecule_get_auger_ci(self->xmol_data, f_irrep, f_multp, f_index, rate,
                         energy, &N_process1);

  if (N_process1 > N_process)
  {
    printf("more than %d processes (%d)\n", N_process, N_process1);
    printf("reallocating memory\n");
    fflush(stdout);
    N_process = N_process1;
    // in case there are more then MAX_PROCESS processes,
    // reallocate more memory and recall get_process()
    f_irrep = realloc(f_irrep, sizeof(int) * N_process);
    f_multp = realloc(f_multp, sizeof(int) * N_process);
    f_index = realloc(f_index, sizeof(int) * N_process);
    energy = realloc(energy, sizeof(double) * N_process);
    rate = realloc(rate, sizeof(double) * N_process);
    xmolecule_get_auger_ci(self->xmol_data, f_irrep, f_multp, f_index,
                           rate, energy, &N_process1);
  }
  int k = 0;
  N_process = N_process1;
  PyObject *list = PyList_New(N_process);
  for (i = 0; i < N_process; i++)
  {
    PyObject *entry;
    entry = Py_BuildValue("(i,i,i,f,f)",
                          f_irrep[i],
                          f_multp[i], f_index[i], rate[i], energy[i]);
    PyList_SetItem(list, k, entry);
    k++;
  }
  free(f_irrep);
  free(f_multp);
  free(f_index);
  free(energy);
  free(rate);

  if (N_process < MAX_PROCESS)
  {
    N_process = MAX_PROCESS;
  }
  return list;
}

static PyObject *
xmol_CI_olap(xmol *self, PyObject *args)
{
  double *O;
  long int olapdims[] = {self->nstates, self->nstates};
  PyArrayObject *py_olap = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                                 2, olapdims, NULL, NULL,
                                                                 NPY_ARRAY_C_CONTIGUOUS, NULL);
  O = PyArray_DATA(py_olap);
  xmolecule_ci_overlaps(self->xmol_data, O, &self->nstates);
  if (self->nstates < 0)
  {
    self->nstates *= -1;
    olapdims[0] = self->nstates;
    olapdims[1] = self->nstates;
    Py_CLEAR(py_olap);
    py_olap = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                    2, olapdims, NULL, NULL,
                                                    NPY_ARRAY_C_CONTIGUOUS, NULL);
    O = PyArray_DATA(py_olap);
    xmolecule_ci_overlaps(self->xmol_data, O, &self->nstates);
  }
  return (PyObject *)py_olap;
}

static PyObject *
xmol_orbital_energies(xmol *self, PyObject *args)
{
  double *oe;
  long int moedims[] = {self->nmo};
  PyArrayObject *py_moe = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                                1, moedims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  oe = PyArray_DATA(py_moe);
  xmolecule_orbital_energies(self->xmol_data, oe, &(self->nmo));

  return (PyObject *)py_moe;
}

static PyObject *
xmol_get_occ(xmol *self, PyObject *args)
{
  int *occ;
  int i;
  occ = malloc(sizeof(int) * self->locc);
  xmolecule_get_occ(self->xmol_data, occ, &(self->locc));
  PyObject *list = PyList_New(self->locc);
  for (i = 0; i < self->locc; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("i", occ[i]));
  }
  free(occ);
  return list;
}

static PyObject *
xmol_population(xmol *self, PyObject *args)
{

  double *pop;
  long int popdims[] = {self->nmo, self->natom};
  PyArrayObject *py_pop = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                                2, popdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  pop = PyArray_DATA(py_pop);
  xmolecule_get_population(self->xmol_data, pop, &(self->nmo),
                           &(self->natom));
  return (PyObject *)py_pop;
}

static PyObject *
xmol_nao_conf(xmol *self, PyObject *args)
{
  char *nao_string;
  nao_string = malloc(sizeof(char) * 512 * self->natom);
  xmolecule_get_nao_conf(self->xmol_data, nao_string, &(self->natom));
  PyObject *py_nao_string = Py_BuildValue("s", nao_string);
  free(nao_string);
  return (py_nao_string);
}

static PyObject *
xmol_set_mo_coef(xmol *self, PyObject *args)
{
  double *coef;
  int i, j;
  char error_msg[LEN_ERRMSG];

  PyArrayObject *py_coef;
  if (!PyArg_ParseTuple(args, "O!",
                        &PyArray_Type, &py_coef))
  {
    PyErr_SetString(PyExc_TypeError, "Bad argument");
    return NULL;
  }
  if (PyArray_NDIM(py_coef) != 1 || PyArray_SHAPE(py_coef)[0] != self->nao || PyArray_SHAPE(py_coef)[1] != self->nmo)
  {
    snprintf(error_msg, LEN_ERRMSG, "shape of coefficient matrix should be (%d,%d)", self->nao, self->nmo);
    PyErr_SetString(PyExc_TypeError, error_msg);
    return NULL;
  }
  coef = malloc(sizeof(double) * self->nao * self->nmo);
  for (i = 0; i < self->nmo; i++)
    for (j = 0; j < self->nao; j++)
      coef[i * self->nmo + j] =
          *(double *)((char *)PyArray_DATA(py_coef) +
                      j * PyArray_STRIDES(py_coef)[0] +
                      i * PyArray_STRIDES(py_coef)[1]);
  xmolecule_set_mo_coef(self->xmol_data, coef, &(self->nao), &(self->nmo));
  free(coef);
  Py_INCREF(Py_None);
  return Py_None;
}
static PyObject *
xmol_ao_overlap(xmol *self, PyObject *args)
{
  double *S;
  long int dims[] = {self->nao, self->nao};
  PyArrayObject *py_S = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                              2, dims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  S = PyArray_DATA(py_S);
  xmolecule_get_overlapmatrix(self->xmol_data, S, &self->nao);
  return ((PyObject *)py_S);
}

static PyObject *
xmol_mo_coef(xmol *self, PyObject *args)
{
  double *coef;
  long int gdims[2];
  gdims[1] = self->nao;
  gdims[0] = self->nmo;
  PyArrayObject *py_coef = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                                 2, gdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  coef = PyArray_DATA(py_coef);
  xmolecule_get_mo_coef(self->xmol_data, coef, &(self->nmo), &(self->nao));
  return ((PyObject *)py_coef);
}

static PyObject *
xmol_mullikencharge(xmol *self, PyObject *args)
{
  double *charge;
  int i;
  charge = malloc(sizeof(double) * self->natom);
  xmolecule_get_mullikencharge(self->xmol_data, charge, &(self->natom));
  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", charge[i]));
  }
  free(charge);
  return list;
}

static PyObject *
xmol_loewdincharge(xmol *self, PyObject *args)
{
  double *charge;
  int i;
  charge = malloc(sizeof(double) * self->natom);
  xmolecule_get_loewdincharge(self->xmol_data, charge, &(self->natom));
  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", charge[i]));
  }
  free(charge);
  return list;
}

static PyObject *
xmol_spin_analysis(xmol *self, PyObject *args)
{
  long int dims[1];
  dims[0] = 3;
  PyArrayObject *py_center = (PyArrayObject *)PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                                                                   1, dims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  double *center = PyArray_DATA(py_center);
  double rog;
  xmolecule_get_spin_analysis(self->xmol_data, center, &rog);
  PyObject *py_rog = Py_BuildValue("d", rog);
  PyObject *ret = Py_BuildValue("(N,N)", py_center, py_rog);
  return (PyObject *)ret;
}

static PyObject *
xmol_os(xmol *self, PyObject *args)
{
  double *os, *tDip;
  long int csdims[] = {self->nmo, self->nmo};
  long int tddims[] = {self->nmo, self->nmo, 3};

  PyArrayObject *py_os = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           2, csdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  PyArrayObject *py_tdip = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           3, tddims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  os = PyArray_DATA(py_os);
  tDip = PyArray_DATA(py_tdip);
  xmolecule_get_oscillator_strength(self->xmol_data, os, tDip, &(self->nmo));
  PyArrayObject *py_os_T = (PyArrayObject *)PyArray_SwapAxes(py_os, 0, 1);
  Py_CLEAR(py_os);
  PyObject *ret = Py_BuildValue("(N,N)", py_os_T, py_tdip);
  return (ret);
}

static PyObject *
xmol_cis_os(xmol *self, PyObject *args)
{
  double *aCS, *tDip;
  int state;
  int retval = PyArg_ParseTuple(args, "i", &state);
  if (!retval)
  {
    PyErr_SetString(PyExc_TypeError, "Bad argument");
    return NULL;
  }
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  long int csdims[] = {self->nstates};
  long int tddims[] = {self->nstates, 3};
  PyArrayObject *py_acs = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           1, csdims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  PyArrayObject *py_tdip = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           2, tddims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  aCS = PyArray_DATA(py_acs);
  tDip = PyArray_DATA(py_tdip);
  xmolecule_get_cis_oscillator_strength(self->xmol_data, &state, aCS, tDip,
                                        &(self->nstates));
  PyObject *ret = Py_BuildValue("(N,N)", py_acs, py_tdip);
  return ret;
}

static PyObject *
xmol_cis_mullikenpop(xmol *self, PyObject *args, PyObject *kwds)
{
  double *charge;
  int i;
  int state = 0;
  char *typ = NULL;
  char default_typ = 'F';
  static char *kwlist[] = {"state", "typ", NULL};
  PyArg_ParseTupleAndKeywords(args, kwds, "i|s", kwlist,
                              &state, &typ);
  if (typ == NULL)
    typ = &default_typ;
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  charge = malloc(sizeof(double) * self->natom);
  if (state == 0)
    xmolecule_get_mullikencharge(self->xmol_data, charge, &(self->natom));
  else
    xmolecule_get_cis_mullikenpop(self->xmol_data, charge,
                                  &(self->natom), &state, &(typ[0]));
  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", charge[i]));
  }
  free(charge);
  return list;
}

static PyObject *
xmol_cis_loewdinpop(xmol *self, PyObject *args, PyObject *kwds)
{
  double *charge;
  int i;
  int state;
  char *typ = NULL;
  char default_typ = 'F';
  static char *kwlist[] = {"state", "typ", NULL};
  PyArg_ParseTupleAndKeywords(args, kwds, "i|s", kwlist, &state, &typ);
  if (typ == NULL)
    typ = &default_typ;
  if (state < 0 || state >= self->nstates)
  {
    PyErr_SetString(PyExc_ValueError, "Bad state index");
    return NULL;
  }
  if (state == 0)
    return xmol_loewdincharge(self, args);
  charge = malloc(sizeof(double) * self->natom);
  xmolecule_get_cis_loewdinpop(self->xmol_data, charge, &(self->natom),
                               &state, &(typ[0]));
  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", charge[i]));
  }
  free(charge);
  return list;
}

static PyObject *
xmol_fluorescencerates(xmol *self, PyObject *args)
{
  double *fluorates;
  long int fldims[] = {self->nmo, self->nmo};
  PyArrayObject *py_fl2 = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           2, fldims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  fluorates = PyArray_DATA(py_fl2);
  xmolecule_get_fluorescencerates(self->xmol_data, fluorates, &(self->nmo));
  PyArrayObject *py_fl = (PyArrayObject *)PyArray_SwapAxes(py_fl2, 0, 1);
  Py_CLEAR(py_fl2);
  return (PyObject *)py_fl;
}

static PyObject *
xmol_bondorder(xmol *self, PyObject *args)
{
  double *BO;
  long int bodims[] = {self->natom, self->natom};
  PyArrayObject *py_bo = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_DOUBLE),
                           2, bodims, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
  BO = PyArray_DATA(py_bo);
  xmolecule_get_bondorder(self->xmol_data, BO, &(self->natom));
  return (PyObject *)py_bo;
}

static PyObject *
xmol_set_geom(xmol *self, PyObject *args, PyObject *kwds)
{
  double *geom;
  int i;
  PyObject *new_geom;
  char error_msg[LEN_ERRMSG];
  static char *kwlist[] = {"geom", "overlap", "reinit", NULL};
  int yes_overlap, yes_reinit;
  yes_overlap = 0;
  yes_reinit = 0;
  int ret = PyArg_ParseTupleAndKeywords(args, kwds, "O|ii:set_geom", kwlist,
                                        &new_geom, &yes_overlap, &yes_reinit);
  if (!ret)
  {
    PyErr_SetString(PyExc_ValueError, "Bad arguments");
    return NULL;
  }
  xmolecule_set_calculate_ci_overlap(self->xmol_data, &yes_overlap);

  geom = malloc(sizeof(double) * 3 * self->natom);
  if (PyList_Check(new_geom))
  {
    int numLines = PyList_Size(new_geom);
    if (numLines != 3 * self->natom)
    {
      snprintf(error_msg, LEN_ERRMSG, "Geometry list should be %d ", 3 * self->natom);
      PyErr_SetString(PyExc_ValueError, error_msg);
      return NULL;
    }
    for (i = 0; i < 3 * self->natom; i++)
    {
      PyObject *item;
      item = PyList_GetItem(new_geom, i);
      geom[i] = PyFloat_AsDouble(item);
    }
  }
  else if (PyArray_CheckExact(new_geom))
  {
    if (PyArray_NDIM((PyArrayObject *)new_geom) != 1 && PyArray_SHAPE((PyArrayObject *)new_geom)[0] != 3 * self->natom)
    {
      snprintf(error_msg, LEN_ERRMSG, "Geometry array should be of shape (%d) ", 3 * self->natom);
      PyErr_SetString(PyExc_ValueError, error_msg);
      return NULL;
    }
    for (i = 0; i < 3 * self->natom; i++)
    {
      geom[i] = *((double *)PyArray_GETPTR1((PyArrayObject *)new_geom, i));
    }
  }
  if (yes_reinit)
    xmolecule_set_reinit(self->xmol_data, &yes_reinit);
  xmolecule_set_geom(self->xmol_data, geom, &(self->natom));
  free(geom);
  xmolecule_get_energy(self->xmol_data, &self->energy);

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_set_occ(xmol *self, PyObject *args)
{
  int *occ;
  int i, ret;
  char error_msg[LEN_ERRMSG];
  int numLines;
  PyObject *PyOcc;
  ret = PyArg_ParseTuple(args, "O", &PyOcc);
  if (!ret)
  {
    PyErr_SetString(PyExc_TypeError, "Argument should be list/array of occupation numbers");
    return NULL;
  }
  if (PyList_Check(PyOcc)){
    numLines = PyList_Size(PyOcc);
    if (numLines != self->locc)
    {
      snprintf(error_msg, LEN_ERRMSG, "Occupation number list should have length %d not %d", self->locc, numLines);
      PyErr_SetString(PyExc_ValueError, error_msg);
      return NULL;
    }
    occ = malloc(sizeof(int) * self->locc);
    for (i = 0; i < self->locc; i++)
    {
      PyObject *item;
      item = PyList_GetItem(PyOcc, i);
      occ[i] = PyLong_AsLong(item);
      if (occ[i] < 0 || occ[i] > 2)
      {
        PyErr_SetString(PyExc_ValueError, "occupation number must be 0, 1, or 2");
      }
    }
  } else if (PyArray_CheckExact(PyOcc)){
    if (PyArray_NDIM((PyArrayObject *)PyOcc) != 1 && PyArray_SHAPE((PyArrayObject *)PyOcc)[0] != self->locc)
    {
      snprintf(error_msg, LEN_ERRMSG, "Occupation number array should have shape (%d) ", self->locc);
      PyErr_SetString(PyExc_ValueError, error_msg);
      return NULL;
    }
    occ = malloc(sizeof(int) * self->locc);
    for (i = 0; i < self->locc; i++)
    {
      occ[i] = *((int *)PyArray_GETPTR1((PyArrayObject *)PyOcc, i));
      if (occ[i] < 0 || occ[i] > 2)
      {
        PyErr_SetString(PyExc_ValueError, "occupation number must be 0, 1, or 2");
      }
    }
  }
  xmolecule_set_occ(self->xmol_data, occ, &(self->locc));
  free(occ);
  xmolecule_get_energy(self->xmol_data, &self->energy);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_set_multp(xmol *self, PyObject *multp_arg)
{
  int multp;
  multp = PyLong_AsLong(multp_arg);
  xmolecule_set_multp(self->xmol_data, &multp);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_set_irrep(xmol *self, PyObject *multp_arg)
{
  int irrep;
  irrep = PyLong_AsLong(multp_arg);
  xmolecule_set_irrep(self->xmol_data, &irrep);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
xmol_mass(xmol *self, PyObject *args)
{
  double *mass;
  int i;
  mass = calloc(sizeof(double), self->natom);
  xmolecule_get_mass(self->xmol_data, mass, &(self->natom));
  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("d", mass[i]));
  }
  free(mass);
  return (list);
}

static PyObject *
xmol_atomlabels(xmol *self, PyObject *args)
{
  char *labels;
  int i;
  labels = calloc(sizeof(char), 6 * self->natom);
  xmolecule_get_atomlabels(self->xmol_data, labels, &(self->natom));

  PyObject *list = PyList_New(self->natom);
  for (i = 0; i < self->natom; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("s#", labels + i * 6, 6));
  }
  free(labels);
  return (list);
}

static PyObject *
xmol_aolabels(xmol *self, PyObject *args)
{
  char *labels;
  int i;
  labels = calloc(sizeof(char), 10 * self->nao);
  xmolecule_get_ao_labels(self->xmol_data, labels, &(self->nao));
  PyObject *list = PyList_New(self->nao);
  for (i = 0; i < self->nao; i++)
  {
    PyList_SetItem(list, i, Py_BuildValue("s#", labels + i * 10, 10));
  }
  free(labels);
  return (list);
}

static PyObject *
xmol_process(xmol *self, PyObject *args)
{
#define MAX_PROCESS 10000
  static int N_process = MAX_PROCESS;
  int N_process1;
  int i, j;
  int *f_occs;
  int *E_eV;
  double *value;
  unsigned char *tag;
  int N_Xprocess;
  f_occs = calloc(sizeof(int), self->locc * N_process);
  E_eV = calloc(sizeof(int), N_process);
  value = calloc(sizeof(double), N_process);
  tag = calloc(sizeof(unsigned char), N_process);
  N_process1 = N_process;
  xmolecule_get_process(self->xmol_data, f_occs, E_eV, value, tag,
                        &N_process1, &self->locc);
  if (N_process1 > N_process)
  {
    printf("more than %d processes (%d)\n", N_process, N_process1);
    printf("reallocating memory\n");
    fflush(stdout);
    N_process = N_process1;
    // in case there are more then MAX_PROCESS processes,
    // reallocate more memory and recall get_process()
    f_occs = realloc(f_occs, sizeof(int) * self->locc * N_process);
    E_eV = realloc(E_eV, sizeof(int) * N_process);
    value = realloc(value, sizeof(double) * N_process);
    tag = realloc(tag, sizeof(unsigned char) * N_process);
    xmolecule_get_process(self->xmol_data, f_occs, E_eV, value, tag,
                          &N_process, &self->locc);
  }
  N_process = N_process1;
  N_Xprocess = 0;
  for (i = 0; i < N_process; i++)
  {
    if (tag[i] == 'X')
    {
      N_Xprocess++;
    }
  }
  int k = 0;
  PyObject *list = PyList_New(N_process - N_Xprocess);
  for (i = 0; i < N_process; i++)
  {
    if (tag[i] == 'X')
      continue;
    PyObject *final_occ = PyList_New(self->locc);
    for (j = 0; j < self->locc; j++){
      PyList_SetItem(final_occ, j, Py_BuildValue("i", f_occs[self->locc * i + j]));
    }
    /*
    long int occdim[] = {self->locc};
    PyArrayObject *final_occ = (PyArrayObject *)
      PyArray_NewFromDescr(&PyArray_Type, PyArray_DescrFromType(NPY_INT),
                           1, occdim, NULL, NULL, NPY_ARRAY_F_CONTIGUOUS, NULL);
    for (j = 0; j < self->locc; j++){
      ((int*)PyArray_DATA(final_occ))[j] = f_occs[self->locc * i + j];
    }
    */
    PyObject *entry;
    entry = Py_BuildValue("(c,O,d,i)",
                          tag[i], final_occ, value[i], E_eV[i]);
    PyList_SetItem(list, k, entry);
    k++;
  }
  free(f_occs);
  free(E_eV);
  free(value);
  free(tag);

  if (N_process < MAX_PROCESS)
  {
    N_process = MAX_PROCESS;
  }
  return list;
}

/*
  object type methods and member arrays
*/
static PyMethodDef xmol_methods[] = {
    {"init", (PyCFunction)xmol_init, METH_NOARGS,
     "initialize xmolecule"},
    {"__setstate__", (PyCFunction)xmol_setstate, METH_VARARGS,
     "__setstate__ implementation"},
    {"__getstate__", (PyCFunction)xmol_getstate, METH_VARARGS,
     "__getstate__ implementaion"},
    {"get_geom", (PyCFunction)xmol_get_geom, METH_NOARGS,
     "returns geometry"},
    {"set_geom", (PyCFunction)xmol_set_geom, METH_VARARGS | METH_KEYWORDS,
     "sets geometry"},
    {"get_occ", (PyCFunction)xmol_get_occ, METH_NOARGS,
     "get occupation number"},
    {"set_occ", (PyCFunction)xmol_set_occ, METH_VARARGS | METH_KEYWORDS, "sets occupation number"},
    {"set_multp", (PyCFunction)xmol_set_multp, METH_O,
     "sets spin multiplicity"},
    {"set_irrep", (PyCFunction)xmol_set_irrep, METH_O, "sets irrep"},
    {"mass", (PyCFunction)xmol_mass, METH_NOARGS, "returns atom masses"},
    {"atom_labels", (PyCFunction)xmol_atomlabels, METH_NOARGS,
     "returns atom labels"},
    {"ao_labels", (PyCFunction)xmol_aolabels, METH_NOARGS, "returns ao labels"},
    {"process", (PyCFunction)xmol_process, METH_NOARGS,
     "returns process list"},
    {"write_molden", (PyCFunction)xmol_write_molden, METH_VARARGS | METH_KEYWORDS,
     "write molden file"},
    {"write_density_cube", (PyCFunction)xmol_write_density_cube, METH_NOARGS,
     "write cube file with e-density"},
    {"write_mo_cube", (PyCFunction)xmol_write_mo_cube, METH_VARARGS,
     "write cube file with molecule orbital; argument 1..n specifies orbital index"},
    {"gradient", (PyCFunction)xmol_gradient, METH_NOARGS,
     "returns gradient"},
    {"mullikencharge", (PyCFunction)xmol_mullikencharge, METH_NOARGS,
     "returns mulliken charge analysis"},
    {"spin_analysis", (PyCFunction)xmol_spin_analysis, METH_NOARGS,
     "returns center of spin and radius of gyration"},
    {"CIS_mullikenpop", (PyCFunction)xmol_cis_mullikenpop, METH_VARARGS | METH_KEYWORDS,
     "returns mulliken charge analysis. cis state number as argument"},
    {"loewdincharge", (PyCFunction)xmol_loewdincharge, METH_NOARGS,
     "returns loewdin charge analysis"},
    {"CIS_loewdinpop", (PyCFunction)xmol_cis_loewdinpop, METH_VARARGS | METH_KEYWORDS,
     "returns loewdin charge. cis state number as argument"},
    {"CIS_os", (PyCFunction)xmol_cis_os, METH_VARARGS,
     "returns CIS oscillator strength"},
    {"os", (PyCFunction)xmol_os, METH_NOARGS,
     "returns oscillator strength"},
    {"fluorescencerates", (PyCFunction)xmol_fluorescencerates, METH_NOARGS,
     "returns fluorescence rates"},
    {"read_chkpt", (PyCFunction)xmol_read_chkpt, METH_NOARGS,
     "read chkpt.dat"},
    {"write_chkpt", (PyCFunction)xmol_write_chkpt, METH_NOARGS,
     "write chkpt.dat"},
    {"bondorder", (PyCFunction)xmol_bondorder, METH_NOARGS,
     "returns bondorder analysis"},
    {"population", (PyCFunction)xmol_population, METH_NOARGS,
     "returns population analysis"},
    {"orbital_energies", (PyCFunction)xmol_orbital_energies, METH_NOARGS,
     "returns orbital energies"},
    {"nao_conf", (PyCFunction)xmol_nao_conf, METH_NOARGS,
     "returns nao basis configuration"},
    {"mo_coef", (PyCFunction)xmol_mo_coef, METH_NOARGS, "returns mo coef C"},
    {"ao_overlap", (PyCFunction)xmol_ao_overlap, METH_NOARGS, "returns ao overlap matrix"},
    {"excitation_energies", (PyCFunction)xmol_excitation_energies,
     METH_NOARGS, "returns excitation energies"},
    {"hole_properties", (PyCFunction)xmol_hole_occs_energies_gradients_nacs,
     METH_NOARGS, "returns hole state prop"},
    {"part_properties", (PyCFunction)xmol_part_occs_energies_gradients_nacs,
     METH_NOARGS, "returns particle state prop"},
    {"hole_efields", (PyCFunction)xmol_hole_occs_energies_efields,
     METH_VARARGS,
     "returns hole state efields"},
    {"CI_energies", (PyCFunction)xmol_CI_energies, METH_VARARGS,
     "returns list of CI energies "},
    {"CI_olap", (PyCFunction)xmol_CI_olap, METH_NOARGS,
     "returns matrix of CI state overlaps to previous geometry "},
    {"CI_Auger", (PyCFunction)xmol_CI_Auger, METH_NOARGS,
     "returns Auger decay rates of CI state"},
    {"occ2CIstate", (PyCFunction)xmol_occ2cistate, METH_O,
     "returns a random state index based on probabilities from the weight on occupation number"},
    {"CIstate2occ", (PyCFunction)xmol_cistate2occ, METH_O,
     "returns the occupation number of the dominant csf"},
    {"CIS_energies", (PyCFunction)xmol_CIS_energies, METH_NOARGS,
     "returns list of CIS energies "},
    {"CIS_energy_gradient", (PyCFunction)xmol_CIS_energy_gradient, METH_O,
     "returns gradients for CIS state (argument is state index)"},
    {"CIS_energy_gradient_nacs", (PyCFunction)xmol_CIS_energy_gradient_nacs,
     METH_VARARGS,
     "returns gradients & nacs for CIS state (argument is state index)"},
    {"CIS_transdens", (PyCFunction)xmol_CIS_transdens,
     METH_VARARGS,
     "returns trans density for CIS state (argument is state index)"},
    {"set_mo_coef", (PyCFunction)xmol_set_mo_coef, METH_VARARGS, "sets the LCAO orbital coefficients"},
    {NULL} /* Sentinel */
};

static PyMemberDef xmol_members[] = {
    {"energy", T_DOUBLE, offsetof(xmol, energy), 0, "SCF energy"},
    {NULL} /* Sentinel */
};

/*
  module and object type initialization
*/
#if PY_MAJOR_VERSION == 2
/*
    python 2.x implementation for module and object type
*/
static PyTypeObject XmolType = {
    PyObject_HEAD_INIT(NULL) 0,                        /*ob_size */
    "xmol",                                            /*tp_name */
    sizeof(xmol),                                      /*tp_basicsize */
    0,                                                 /*tp_itemsize */
    (destructor)xmol_dealloc,                          /*tp_dealloc */
    0,                                                 /*tp_print */
    0,                                                 /*tp_getattr */
    0,                                                 /*tp_setattr */
    0,                                                 /*tp_compare */
    0,                                                 /*tp_repr */
    0,                                                 /*tp_as_number */
    0,                                                 /*tp_as_sequence */
    0,                                                 /*tp_as_mapping */
    0,                                                 /*tp_hash */
    0,                                                 /*tp_call */
    0,                                                 /*tp_str */
    0,                                                 /*tp_getattro */
    0,                                                 /*tp_setattro */
    0,                                                 /*tp_as_buffer */
    Py_TPFLAGS_DEFAULT,                                /*tp_flags */
    (const char *)"xmol object type to run xmolecule", /* tp_doc */
    0,                                                 /* tp_traverse */
    0,                                                 /* tp_clear */
    0,                                                 /* tp_richcompare */
    0,                                                 /* tp_weaklistoffset */
    0,                                                 /* tp_iter */
    0,                                                 /* tp_iternext */
    xmol_methods,                                      /* tp_methods */
    xmol_members,                                      /* tp_members */
    0,                                                 /* tp_getset */
    0,                                                 /* tp_base */
    0,                                                 /* tp_dict */
    0,                                                 /* tp_descr_get */
    0,                                                 /* tp_descr_set */
    0,                                                 /* tp_dictoffset */
    (initproc)xmol_setup,                              /* tp_init */
    0,                                                 /* tp_alloc */
    0,                                                 /* tp_new */
};

void initpython_xmolecule_interface(void)
{
  PyObject *m;
  XmolType.tp_new = PyType_GenericNew;
  if (PyType_Ready(&XmolType) < 0)
    return;
  m = Py_InitModule3("python_xmolecule_interface", NULL,
                     "The python-xmolecule-interface.");
  Py_INCREF(&XmolType);
  if (PyModule_AddObject(m, "xmol", (PyObject *)&XmolType) < 0)
  {
    Py_DECREF(&XmolType);
    Py_DECREF(m);
    return;
  }
  import_array();
  return;
}

#elif PY_MAJOR_VERSION == 3
/*
    python 3.x implementation for module and object type
*/
static struct PyModuleDef python_xmolecule_interface_module = {
    PyModuleDef_HEAD_INIT,
    "python_xmolecule_interface",
    "python 3.x Interface to xmolecule",
    -1,
};

static PyType_Slot slots[] = {
    {Py_tp_init, (initproc)xmol_setup},
    {Py_tp_dealloc, (destructor)xmol_dealloc},
    {Py_tp_methods, xmol_methods},
    {Py_tp_members, xmol_members},
    {0, NULL},
};

static PyType_Spec spec = {
    "python_xmolecule_interface.xmol",
    sizeof(xmol),
    0,
    Py_TPFLAGS_DEFAULT,
    slots,
};

PyMODINIT_FUNC
PyInit_python_xmolecule_interface(void)
{
  PyObject *m;
  m = PyModule_Create(&python_xmolecule_interface_module);
  if (m == NULL)
  {
    PyErr_SetString(PyExc_Exception, "Error loading module");
    return NULL;
  }
  PyObject *XmolType = PyType_FromSpec(&spec);
  if (PyType_Ready((PyTypeObject *)XmolType) < 0)
  {
    PyErr_SetString(PyExc_Exception, "Error creating xmol type");
    return NULL;
  }
  Py_INCREF(XmolType);
  if (PyModule_AddObject(m, "xmol", (PyObject *)XmolType) < 0)
  {
    Py_DECREF(XmolType);
    Py_DECREF(m);
    PyErr_SetString(PyExc_Exception, "Error  adding object");
    return NULL;
  }
  import_array();
  return m;
}

#else
#error PY_MAJOR_VERSION NOT 2 or 3
#endif
