!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module read_param
  !! author: Sang-Kil Son
  !! author: Yajiang Hao
  !! author: Ludger Inhester
  !!
  !! READ_PARAM module for atomic system and polyatomic molecular system
  !!   written by Son, Sang-Kil in Jan. 2005, Jun. 2005
  !!   re-written by Son, Sang-Kil in Oct. 2007 for polyatomic molecules
  !!   re-written by Son, Sang-Kil and Hao, Yajiang in Mar. 2014 for xmolecule
  !!   adapated by Ludger Inhester
  !! USAGE:
  !!   type(Param) :: P
  !!   call read_parameters( P )
  !!   call print_parameters( P, "your title" )
  !!
  !! SUBROUTINES:
  !! - SR read_parameters( P, calc_string ): read all parameters
  !! - SR read_[*]_param( P, ARG )
  !!   [*]: molecule, grid, XFEL, DFT, MC, process, plasma, and misc
  !! - SR print_parameters( P, title ): print out all relevant parameters
  !! - SR print_[*]_param( P )
  !!   [*]: molecule, grid, XFEL, DFT, MC, process, plasma, and misc
  !! - SR print_raw_param( P ): print all parameters as they are
  !!
  use constants
  use stdio
  use atom
  use stdio
  use iso_c_binding
  implicit none
  type(c_ptr), bind(C, name="version_str") :: version_str_ptr
  type(c_ptr), bind(C, name="build_time_str") :: build_time_str_ptr

  ! auxilliary structure for average_occ
  type average_occ
    !! index of MOs for which we average occupation number
    integer, allocatable :: MO(:)
  end type average_occ
  type :: GAS
    !! general active space specifications
    !! orbitals and occupation number limits
    integer, allocatable :: orbitals(:)
    integer :: nmin
    integer :: nmax
  end type GAS

  type Param
    !! Param: all parameters to be used in main program
    !>
    !! Calculation type -------------------------------------------------------
    !!   - 0 : static (electronic structure)
    !!   - 1 : ionization dynamics simulated by direct rate equaiton (REQ) solver
    !!   - 2 : ionization dynamics simulated by Monte Carlo on the fly (MC_FLY)
    !!   - 3 : ionization and molecular dynamcis by Monte Carlo on the fly
    !!
    integer :: calc_type
    !
    ! Molecular system -------------------------------------------------------
    !   molecular geometry and charge, and so on
    !   allocatable variables with "_nuc" are arrays run thru 1 ... N_nuc
    !
    !   EXAMPLE of specifying molecular geometry and basis functions
    !     O  0.0000   0.0000   0.0000  2s1p  O.nao
    !     H  0.0000  -1.4309   1.1084  1s    H.nao
    !     H  0.0000   1.4309   1.1084  1s    H.nao
    !   Here, 2s1p for O and 1s for H correspond to the minimal basis set.
    !   2s1p means two s orbitals (1s, 2s) and one p orbital (2p). => N_nao = 3
    !   One p orbital has 3 basis functions (2px, 2py, 2pz).
    !   Therefore, # of basis functions from O is 2*1 + 1*3 = 5. => N_basis = 5
    !
    character(len=80) :: system_name    !! name of the system
    integer :: N_nuc                    !! num. of nuclei
    integer :: charge                   !! total charge of molecule
    integer :: N_elec                   !! num. of electrons
    integer :: N_a, N_b                 !! num. of alpha and beta electrons
    integer :: N_orb                    !! num. of orbitals (or subshells)
    real(long), allocatable :: x_nuc(:), y_nuc(:), z_nuc(:), mass_nuc(:) !! position and mass of the nuclei
    integer, allocatable :: charge_nuc(:) !! integer charge of the nuclei
    real(long), allocatable :: fcharge_nuc(:), sigma_nuc(:) !! fractional charge of the nuclei,
    !! spatial width (stddev) of the nuclei charge
    real(long) :: E_nuc                 !! nucleus-nucleus repulsion energy
    !
    ! specifying electronic configuration via occupation numbers
    !   1) non-relativistic atomic calc. :   2   2   6   1  <-- N_occ(:)
    !                                      1s  2s  2p  3s
    !
    !   2) relativistic atomic calc. :      2      2      2      4      1
    !                                  1s1/2  2s1/2  2p1/2  2p3/2  3s1/2
    !
    !   3) molecular calc. :    2    2    2    2    2    1
    !                        1st  2nd  3rd  4th  5th  6th
    !      it's only energy-ordered because we use no molecular symmetry.
    !
    integer, allocatable :: N_occ(:)        !! num. of occupation in [1...N_orb]
    real(long), allocatable :: F_occ(:)        !! num. of occupation  (fractional) in [1...N_orb]
    real(long), allocatable :: F_occ_a(:)        !! num. of alpha occupation in [1...N_orb]
    real(long), allocatable :: F_occ_b(:)        !! num. of beta occupation in [1...N_orb]
    !
    ! Basis functions to be read in read_molecule_param() --------------------
    !   We employ NAO (Numerical Atomic Orbital) as basis functions
    !     for molecular calculations.
    !   N_nao_table(:,:) : specifying num. of NAO functions per l and each atom
    !     (0:MAX_lmax,1:N_nuc) where MAX_lmax = max. of lmax for all atoms
    !     Remember that 2s1p means two s orbitals and one p orbital,
    !     so num. of basis functions becomes 5!!!
    !   N_basis : total num. of basis functions, considering l
    !
    !   EXAMPLE of basis_string
    !     2s1p at i-th atom: two s orbitals and one p orbitals (px, py, pz)
    !       N_nao = 3;  N_basis = 2*1 + 1*3 = 5
    !       N_nao_table(l=0,i) = 2;  N_nao_table(l=1,i) = 1
    !     5s3p1d at i-th atom: 5 s orbitals, 3 p orbitals, and 1 d orbital
    !       N_nao = 9;  N_basis = 5*1 + 3*3 + 1*5 = 19
    !       N_nao_table(0:2,i) = ( 5 3 1 )
    !
    !   EXAMPLE of the NAO basis function file:
    !     200  1.0  50.0        # N_AG, L_AG, rmax_AG
    !     2  5                  # lmax and how many NAOs for given l
    !     ... (5 lines with 200 NAO values at 200 radial grid points for l=0)
    !     ... (5 lines with 200 NAO values at 200 radial grid points for l=1)
    !     ... (5 lines with 200 NAO values at 200 radial grid points for l=2)
    !     EOF
    !
    character(len=80), allocatable  :: nao_name(:) !! the basis name printed out
    character(len=5), allocatable  :: nuc_name(:) !! the name of the atom as specified in the input file
    character(len=512), allocatable :: nao_file_nuc(:)  !! filename, (1:N_nuc)
    integer, allocatable :: N_nao_table(:, :)            !! (0:lmax,1:N_nuc)
    integer :: N_basis                                  !! total num. of basis func.
    !
    ! Grid system ------------------------------------------------------------
    !   without suffix : representive value (all the same for individual atoms)
    !   with suffix _AG : for individual atomic grids
    !   with suffix _GG : for global grids
    !   - N : num. of radial grids
    !   - lmax : define the quality of angular grids, num. of ang. grids ~ lmax^2
    !   - L : mapping parameters for radial grids
    !   - rmax : maximum radius for radial grids
    !
    integer :: N_grid                   !! total number of molecular grids
    integer :: N, lmax !! number of radial grid points, number of max. angular moment defining the angular grid points
    real(long) :: L, rmax
    integer, allocatable :: N_AG(:), lmax_AG(:)         !! (1:N_nuc)
    real(long), allocatable :: L_AG(:), rmax_AG(:)      !! (1:N_nuc)
    integer :: N_GG, lmax_GG
    real(long) :: L_GG, rmax_GG
    real(long) :: dx_GG         ! dx if global grids are Cartesian
    character(len=20) :: type_radial, type_radial_GG
    character(len=20) :: type_angular, type_angular_GG
    logical :: yes_overlap, yes_squeeze, yes_global, yes_atomic, yes_cartesian
    !
    ! DFT and SCF parameters -------------------------------------------------
    !   x_type and c_type may have the integer index from
    !   FUNCTIONAL_X_NAME(:) and FUNCTIONAL_C_NAME(:).
    !
    integer :: x_type, c_type   !! exchange and correlation functional names
    real(long) :: alpha, beta   !! parameters for the particular functional
    logical :: yes_OEP_KLI      !! turn on or off the OEP-KLI procedure
    logical :: yes_Latter       !! turn on or off the Latter tail correction
    integer :: SCF_max_iter             !! the maximum iteration of SCF
    real(long) :: MCSCF_gradient_conv   !! the convergence of SCF with E diff
    real(long) :: SCF_energy_conv       !! the convergence of SCF with E diff
    real(long) :: SCF_density_conv      !! the conv. of SCF with density diff
    real(long) :: SCF_mixing            !! mixing ratio of old and new pot.
    logical :: yes_load_pot     !! load the potential from file
    integer :: lmax_MP  !! lmax for multipole expansion in direct Coulomb term
    real(long) :: SCF_SR_den_conv       !! the conv. of density before DIIS
    real(long) :: SCF_SR_pot_conv       !! the conv. of eff. pot. before DIIS
    logical :: yes_DIIS                 !! turn on DIIS
    logical :: yes_CPHFDIIS             !! turn on/off CPHFDIIS
    integer :: DIIS_nvector     !! num. of stored vector in DIIS
    logical :: yes_chk_fid_ini  !! conserve the Ini. MO symmetry of during SCF
    logical :: yes_population   !! Mulliken atomic population analysis
    logical :: yes_population_L   !! population analysis resolved for L basis functions=S,P,D
    logical :: yes_stability    !! stability analysis
    logical :: yes_LevelShift   !! turn on LevelShifting
    logical :: yes_LevelShiftDynamic    !! turn on Dynamic LevelShifting
    real(long) :: LShift        !! level shift
    real(long) :: SwitchToDIISCriterion(2)  !! 28.07.2015 : (2):density deviation threshold is added
    integer :: SwitchToDIISIter
    real(long) :: SwitchToRSTRCTCriterion
    real(long) :: SwitchLevelShiftOffCriterion
    logical :: yes_RSTRCT       ! restriction of orbitals interchanges
    logical :: yes_MOM          ! MOM
    integer :: MOMtype          ! MOM type
    !
    ! freeze(:)     : MO indices which are frozen during the SCF
    ! localize(:)   : MO indices which are localized after SCF
    ! fix_nuc_MO(:) : MO indices which are sorted such that
    !                 they are dominantly on specific nucleus
    ! fix_nuc_n(:)  : nuclei indices for sorting MOs such that
    !                 they are dominantly on specific nucleus
    integer, allocatable :: freeze(:), localize(:)
    integer, allocatable :: final_hole(:)
    integer, allocatable :: cis_hole(:)
    integer, allocatable :: fix_nuc_MO(:), fix_nuc_n(:)
    type(average_occ), allocatable :: av_occ(:)
    type(GAS), allocatable :: gas(:)
    type(GAS), allocatable :: final_gas(:)
    type(GAS), allocatable :: ras(:)
    integer, allocatable :: active_space_MO(:)
    integer, allocatable :: final_active_space_MO(:)
    logical :: yes_auto_average

    logical :: yes_average_rates
    logical :: yes_average_csf_rates
    logical :: yes_all_occs ! compute rates for all occs within average scheme
    logical :: yes_all_multps
    logical :: yes_sym
    logical :: yes_nat_orb ! compute natural orbitals
    integer, allocatable :: nat_trans_orb(:) ! array of final state indices for which natural transition orbitals are calculated
    integer :: state_irrep ! number of irreducible representation of state
    integer, allocatable :: irrep_MO(:)        ! index of irreducible representation per MO

    !
    ! truncation control parameters
    !   EPS_00 := EPS0 in Struct. Dyn. 2, 041707 (2015)
    !   EPS_lm := EPS1 in Struct. Dyn. 2, 041707 (2015)
    !   EPS_r  := EPS2, which is not used in the above paper
    !   r_cutoff : cut-off radius to be included in the potential calculation
    !              w.r.t a certain molecular grid point
    !              WARNING -- it is NOT r_c in (A4) in the above paper
    !
    real(long) :: EPS_overlap   ! finding local atoms for phi_i phi_j
    real(long) :: EPS_overlap_single  ! finding local atoms for phi_i^2
    real(long) :: EPS_00        ! EPS0: accuracy control in pot_J, Q_A/r
    real(long) :: EPS_lm        ! EPS1: accuracy control in pot_J, d_lm/d_00
    real(long) :: EPS_r         ! EPS2: accuracy control in pot_J, d_lm(r)/d_lm
    real(long) :: r_cutoff      ! cut-off radius for potential calculation
    !    real(long) :: r_elec_stat   ! another cut-off radius for electrostatic calc
    logical :: yes_cutoff       ! use r_cutoff, NOTE: yes_cutoff=on for crystal
    !               ===> NOT ANYMORE... Apr. 2, 2019
    real(long) :: EPS_Auger     ! truncation for Auger rate calculation
    real(long) :: EPS_teint  ! twoelectron Integral screening

    real(long) :: R_softcore    ! Radius for softcore potential (Coulomb interaction)
    logical :: yes_overlap_twocenter    ! S(i,j) calc. by two-center integral

    real(long) :: basisSetOverlapLim ! reduce the number of basis functions by neglecting eigenvectors of the overlap matrix with eigenvalues below this value
    !
    ! beyond conventional DFT: thermal HF, RHF, ROHF, and UHF
    !
    logical :: yes_HF           ! do a Hartree Fock calculation
    logical :: yes_MCSCF        ! do MCSCF calculation
    logical :: yes_CIS          ! calculate excited states single excitation configuration interaaction
    logical :: yes_CI           ! perform general configuration interaction calculation
    integer :: mcscf_type       ! integer for different option for mcscf: steepest descent: 1, quasi-newton BFGS: 2, full newton step: 3
    integer :: multp            ! spin multiplicity
    integer :: state            ! nr of state
    integer :: nstates            ! number of states
    real(long), allocatable :: sa_weights(:) ! state average weights
    logical :: yes_state_average ! perform state averaging
    real(long), allocatable :: ci_coef(:) ! fixed ci coefficients for mcscf
    logical :: yes_openshell    ! do we have open shell?
    integer :: ROHF_variant ! 1=Roothaan (default), 2=Davidson (e.g. as in molpro), 3=Guest&Saunders
    logical :: yes_unrestricted    ! do we have open shell?
    real(long) :: T_eV          ! temperature (in eV)
    logical :: yes_thermal      ! finite-temp SCF (pop. according to Fermi dist)
    logical :: yes_broadening   ! pop. convoluted with gaussians
    logical :: yes_direct_scf   ! (for Gaussians) direct scf instead of precalculated teints
    logical :: yes_mmap_teints  ! (for Gaussians) placed precomputed teints into memory-mapped scratch file
    ! Koopmanns' theorem mode
    logical :: yes_valgs_occ      ! compute ground state (valence closed shell) and then remove an electron to obtain the actual configuration
    logical :: yes_valgs_occ_part ! instead of removing an electron; add an electron
    integer :: Ncore            ! number of core orbitals: these keep fixed occupation regardless of termperature
    !
    ! yes_DDIFF: reuse old densities for constructing v_J / Fock matrix
    ! n_diff_save: num. of saved den. matrices / coulomb pot. from prev. iterations
    !
    logical :: yes_DDIFF
    integer :: n_ddiff_save
    !
    ! gradient calculation
    !
    logical :: yes_gradient
    logical :: yes_orbital_energy_gradient
    logical :: fix_geom
    !
    ! misc. for electronic structure calculation
    !
    logical :: yes_atomic_potential     ! use atomic pot. for NAOs
    !
    ! Crystalline parameters -------------------------------------------------
    !   For implementing the periodic condition in the potential calculation,
    !   r_cutoff must be assigned, at least r_cutoff < min(unitcell_{x,y,z})/2
    !
    logical :: yes_periodic                             ! use periodicity
    real(long) :: unitcell_x, unitcell_y, unitcell_z    ! size of unitcell
    logical :: yes_periodic_fuzzy       ! for Kota's fuzzy gradient scheme
    integer :: N_tr_max                 ! max num. of translations to be included
    !
    ! Process parameters -----------------------------------------------------
    !
    logical :: yes_test_fallback ! this switches off the new fallback routine
    logical :: yes_freezeBadOverlap ! rather take frozen orbitals instead of bad overlap
    logical :: yes_ionization_dynamics ! do we have ionization dynamics at all?
    logical :: yes_shakeup      ! shakeup probability
    logical :: yes_fluorescence ! fluorescence rate
    logical :: yes_auger        ! Auger (Coster-Kronig) decay rate
    logical :: yes_process_spin ! all transition rates are calculated for spin configurations
    logical :: yes_process_cont   ! all transition rates are calculated with plane wave continuum
    logical :: yes_process_CI   ! all transition rates are calculated with CI states
    logical :: yes_process_CIS   ! all transition rates are calculated with CI states (including single excitations)
    logical :: yes_process_CIS_ini   ! all transition rates are calculated with CI states for the initial state (including single excitations)
    logical :: yes_process_CISD   ! all transition rates are calculated with CI states (including single+double excitations)
    logical :: yes_process_CI_relaxed   ! all transition rates are calculated with CI states and relaxed orbitals
    integer :: process_CI_relaxed_variant ! 1 or 2: calculate the transition matrix elements between the two orbital sets or instead make a projection
    integer :: CIS_limit(2) = (/-1, -1/) ! highest orbital index to which additional excitation is included
    real(long) :: CIS_E_limit = -HUGE(1) ! lowest energy limit for which additional single excitation are included
    character(len=512) :: process_relaxed_wf_filename ! filename to read in relaxed orbital coefficients
    logical :: yes_auger_random ! switch to radomly scale Auger rates up and down
    logical :: yes_auger_random_2 ! switch to radomly scale Auger rates up and down
    logical :: yes_pcs          ! photoionization cross section
    logical :: yes_absorption   ! absorption rate
    logical :: yes_tpcs         ! direct two-photon ionization cross section
    logical :: yes_spectra      ! print out electron and photon spectra
    logical :: yes_absorptioncs ! compute also absorption cross section based on provided photon energy / spectral stdev
    logical :: yes_dispersion   ! dispersion correction to elastic x-ray scatt.
    logical :: yes_compton      ! Compton scattering
    logical :: yes_scattering   ! elastic x-ray scattering form factor
    integer :: N_proc_max       ! max. num. of processes, used for MC on the fly
    logical :: yes_qmmm         ! skip nuclear interaction between 2 fractional charges (for QM/MM)

    !
    ! Plasma parameters ------------------------------------------------------
    !
    logical :: yes_screening
    !
    ! XFEL parameters --------------------------------------------------------
    ! most of values are specified with a given unit,
    ! but they are stored in a.u. unless specified.
    ! * pulse_type : temporal shape of the x-ray beam
    !     0: ramp, 1: sin^2, 2: chirp, 3: Gaussian, 4: trapezoidal
    !     5: cos, 6: flat-top, 7: two Gaussians
    !     * Note: 3, 6, and 7 are relevant to x-ray pulses.
    !             others were used for laser parameters (obsolete).
    ! * pulse_duration: temporal width in a.u.
    !   3) Gaussian pulse shape
    !      pulse_duration is given by FWHM (Full Width at Half Maximum)
    !      FWHM is given in fs from the input, but stored in a.u.
    !      t_start_au = - (pulse_multiple / 2) * FWHM
    !      t_end_au   = + (pulse_multiple / 2) * FWHM
    !   7) flat-top pulse shape
    !      FWHM & pulse_multiple are no longer relevant
    !      t_start_au = 0
    !      t_end_au   = pulse_duration
    ! * dt_au : time step in a.u.
    ! * width : spatial width in a.u., given in micron from the input
    ! * n_photons : number of photons
    ! * fluence := n_photons / width**2
    !           Here we don't consider a spatial shape (usually Gaussian).
    !           Instead we will calculate dynamics at a single fluence point.
    !           The spatial shape will be taken care by the volume integration.
    !
    integer :: pulse_type
    integer :: N_time   ! num. of temporal grid points
    real(long) :: PE    ! photon energy in a.u., but input given in eV
    real(long) :: PE_width    ! photon energy bandwidth (stdev) in a.u., but input given in eV
    real(long) :: FWHM, pulse_duration  ! in a.u., but input given in fs
    real(long) :: pulse_multiple        ! in multiple of FWHM
    real(long) :: dt_au, t_start_au, t_end_au
    real(long) :: t_pulse_offset        ! center of Gaussian pulse (default=0)
    real(long) :: n_photons     ! num. of photons
    real(long) :: width         ! in a.u., but input given in micron
    real(long) :: fluence       ! == n_photons / width**2, in a.u.
    real(long) :: direction(3)
    real(long) :: pol_axis(3)
    !
    ! Monte-Carlo control parameters -----------------------------------------
    !
    integer :: N_MC_max = -1      ! limits MC step if it is set>0 (default=nolimit)
    real(long) :: MC_conv       ! convergence threshold
    integer :: MC_max_events    ! max num. of events
    integer :: MC_step          ! print out & check convergence at evey MC_step
    real(long) :: MC_dt_ratio   ! dt = 1 / (sum of rates) * MC_dt_ratio
    logical :: yes_pathway      ! print out each trajectory in MC
    integer :: n_pathway        ! print out trajectory at every n trajectory

    real(long)::MC_ratesum_cutoff   ! if (rate_sum < this ) MC step exits.
    logical :: yes_reUse_PreviousResults ! if(yes) saves ProcData and SCF results into FILE
    integer :: MC_fileTableSizeMax, &    ! num. of Nocc to be kept in Table   if(num>this) oldest of the least-visited will be delted
      MC_nfileSaveMax, &    ! num. of Files to be kept
      MC_nVisitThreshold, &    ! if( (num. of occurrance of the same Nocc) >= this )
    !   save it into file. (1=save on the first occurrance)
      MC_NprocessToBeSaved      ! Realistic num. of processes to be kept in file
    Integer ::   MC_dumpOutput           ! output unit for error-dump (for re-starting MC trajectory)
    Real(long):: MC_maxNfaultRatio       ! max tolerance ratio for (num. of Error MC step)/(max MC step)
    logical :: yes_adapt_for_all_AO ! all atomic orbitals are adapted to the electronic configuration
    ! otherwise only core orbital configuration is used in atomic orbital calculation
    integer ::   MC_nStepUpl_Cache  ! max num. of steps to be cached : 0=initial state only,  1= up to 1st step from intl Nocc...
    integer, allocatable :: MC_seeds(:)  ! MC seed for restart
    ! -- every MC step starts out with same conditions provided by
    character(len=512) :: wfFileName    ! LC guess
    character(len=512) :: wfaFileName    ! LC guess
    character(len=512) :: wfbFileName    ! LC guess
    character(len=512) :: basisFileName ! basis
    character(len=512) :: ref_wf    ! LC guess
    character(len=512) :: ref_basis    ! LC guess
    integer :: SCF_N_iter_Large     ! if N_iter exceeds this value Warning is issued
    !
    ! Misc. parameters -------------------------------------------------------
    !
    integer :: OUTPUT = STDOUT
    integer :: OUTPUT2 = STDOUT ! used for back-compatibility
    logical :: yes_verbose      ! determine wordy outputs or not
    integer :: print_level      ! choose the degree of verbose outputs
    integer :: nmo_print !number of orbitals to print
    logical :: yes_test_only, yes_gaussian, yes_print_wf
    logical :: yes_ReadIni, yes_SaveIni, yes_ReadShell, yes_CoreIni
    logical :: yes_plot_potential       ! print out r-dep. atomic potentails
    logical :: yes_plot_density         ! print out r-dep. atomic densities
    character(len=512) :: filename, file_out, file_out2, dir_womersley
    character(len=512) :: file_data     ! file for molecular data
    character(len=:), allocatable :: gto_file_path    ! Gaussian basis set file path, this needs to be rather larger because in may contain for each atom the full path
    character(len=:), allocatable :: gto_file_path2    ! Gaussian basis set file path, this needs to be rather larger because in may contain for each atom the full path
    character(len=:), allocatable :: gto_file_path_nuc(:)    ! Gaussian basis set file for each atom separately specified
    character(len=:), allocatable :: gto_file_path2_nuc(:)    ! Gaussian basis set file for each atom separately specified
    character(len=256) :: gto_dir    ! directory containing gaussian basis sets
    character(len=256) :: ini_File_Path
    character(len=256) :: XATOM_PATH
    character(len=256) :: WORK_DIR_PATH
    character(len=512) :: molden_file    ! LC guess
    real(long), allocatable :: vini(:)
    integer :: Fluorescence_Energy_Threshold_eV  ! MC transition cut-off:any 'F' process with E_eV < (this value) is removed from the table
    real(long) :: avgSqrOvlp_TOL(2) = (/-ONE, -ONE/)   ! TOL of  avgSqrOvlp_w_refMO (see wavefunction_mol)  set <0 to switch off
    ! obtained val < (1) : Error   < (2) : warn    (DEFAULT value yields neither)
    real(long) :: guess_overlap_TOL(2) = (/0.50_long, -ONE/) ! TOL of F%guess_overlap**2,           (1):Error (2):Warn*  * try-fallback if you fail in EITHER stop MC/MD iff you still fail in (1)
    real(long) :: MO_fidelity_TOL(2) = (/-ONE, -ONE/)   ! TOL of min(F%fidelity(:)). MD-only** (1):Error (2) Warn*  ** only for MD (as we want to keep continuity of (excited state) PESs )
    logical :: yes_dump_ints ! dump transformed el integrals
    logical :: yes_use_arpack
    logical :: yes_use_krylov
    logical :: yes_use_davidson
    integer :: davBlock ! davidson block size
    integer :: davMax ! maximum matrix size before reset
    logical :: yes_density_cube ! write density cube file
    logical :: yes_spin_density_cube ! write density cube file
    integer :: Nvoxel ! number of voxels in cubefile
    integer, allocatable :: mo_cube(:) ! mo cube file
    logical :: yes_SOSCF
    real(long) :: SOSCF_start(3)    ! starting orbGrad/density convergence and deadline iter

    real(long), allocatable :: SCF_acceptable_convergence(:) ! acceptable convergence of Energy/Density : in dynamical calculations, we may want to set -SCF_xxx_conv- tighter than necs. to make better convergence in the next step
    ! this parameter allows to apply looser criterion if the former is unreachable

    real(long) :: tStepMD                         ! nuclear timeStep for MD (fs)
    real(long), allocatable :: internuc_dist_upl_au(:), internuc_dist_lwl_au(:) ! Upper/Lower limit of internuclear distance(s); if ANY of them is satisfied,
    ! nuclear dynamics is regarded to be done
    ! Formally set all n_nuc*(n_nuc-1)/2 values, where values <=0 are ignored.
    real(long) :: MD_maxtime                      ! max time of MD (fs) (<0 means un-limited)
    Character(len=256) :: pulse_file_path    ! pulse file path
    Integer ::                     pulse_file_offset  ! offset of pulse file number : we use pulsefile offset, offset+1,... offset+(N_MC-1)

    real(long) :: max_theta
  end type Param

  ! ----------------
  ! global constants
  ! ----------------
  character(len=10), parameter :: FUNCTIONAL_X_NAME(7) = (/ &
    "LSDA  ", &   ! Local Spin Density Approximation (N_a /= N_b)
    "LDA   ", &   ! Local Density Approximation (N_a == N_b)
    "Xa    ", &   ! X-alpha: == LDA without correlation if alpha = 2/3
    "LB    ", &   ! R. van Leeuwen and E. J. Baerends
    "LBa   ", &   ! improved LB containing alpha and beta empirical parameters
    "B     ", &   ! Becke potential
    "Slater"/) ! Slater potential == X-alpha with alpha = 1.00
  character(len=10), parameter :: FUNCTIONAL_C_NAME(5) = (/ &
    "none", &
    "CA  ", &   ! Ceperley and Alder (1980)
    "VWN ", &   ! Vosko, Wilk, and Nusair potential (1980)
    "PW  ", &   ! Perdew and Wang potential (1992)
    "LYP "/)   ! Lee, Yang, and Parr potential

  private
  public :: GAS, Param, average_occ, read_parameters, print_parameters, dump_parameters, purge_parameters, &
    put_minimal_basis_set, put_basis_function, &
    FUNCTIONAL_X_NAME, FUNCTIONAL_C_NAME
contains

  subroutine read_parameters(P, calc_string, filename)
    !! read all parameters
    type(Param), intent(inout) :: P
    character(len=*), intent(in), optional :: calc_string
    character(len=*), intent(in), optional :: filename
    character(len=100) :: string
    type(Argument) :: ARG
    logical, allocatable :: is_parsed(:)
    logical, allocatable :: is_parsed2(:)
    integer :: i
    !
    ! set calc_type
    !
    P%calc_type = 0               ! default: electronic structure calculation
    if (present(calc_string)) then
      string = calc_string
      call convert_to_lowercase(string)
      select case (string)
       case ("static", "statics", "electronic structure", "elec_structure", &
         "elec_struct", "structure")
        P%calc_type = 0
       case ("dynamic", "dynamics", "electronic dynamics", "elec_dynamics", &
         "elec_dyn", "ionization dynamics", "ion_dynamics", "ion_dyn", &
         "req", "rate_equation", "rate_eq")
        P%calc_type = 1
       case ("mc", "monte-carlo", "monte caro", "mc_fly", "mc on the fly", &
         "monte-carlo on the fly", "monte carlo on the fly")
        P%calc_type = 2
       case ("full", "ionization and molecular", "ionization and nuclear", &
         "md", "molecular dynamics")
        P%calc_type = 3
       case default;
        call error_message("calc_type="//calc_string//" is not supported.")
      end select
    end if
    !
    ! read data from STDIN or input file through get_input( ARG )
    !
    if (present(filename)) then
      call get_input(ARG, filename2=filename)
    else
      call get_input(ARG)
    end if
    allocate (is_parsed2(ARG%N))
    allocate (is_parsed(ARG%N))
    is_parsed(:) = .False.
    call read_molecule_param(P, ARG, is_parsed2)
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_process_param(P, ARG, is_parsed2)
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_grid_param(P, ARG, is_parsed2(:))
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_DFT_param(P, ARG, is_parsed2(:))
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_XFEL_param(P, ARG, is_parsed2(:))
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_MC_param(P, ARG, is_parsed2(:))
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_crystal_param(P, ARG, is_parsed2(:))     ! should be after read_molecule_param()
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)
    call read_misc_param(P, ARG, is_parsed2(:))
    is_parsed(:) = is_parsed(:) .or. is_parsed2(:)

    if (P%calc_type .gt. 0) P%print_level = 1  !! added 31.Mar.2015: minimize print-outs for dynamical calc
    write (P%OUTPUT, '("# ",a)') repeat("-", 64)
    write (P%OUTPUT, '("# Summary of Input Parameters: ")')
    do i = 1, ARG%N
      if (ARG%option(i) .eq. "") cycle
      if (.not. is_parsed(i)) then
        write (P%OUTPUT, '("# Dear Xmolecule User, I do not understand the following input:")')
        write (P%OUTPUT, '("# >> ",a,"=",a," << ")') trim(ARG%option(i) (2:)), trim(ARG%string(i))
        stop
      end if
      if (ARG%option(i) .eq. "-") then
        write (P%OUTPUT, '("# ",a)') trim(ARG%string(i))
      else
        write (P%OUTPUT, '("# ",a,"=",a)') trim(ARG%option(i) (2:)), trim(ARG%string(i))
      end if
    end do
    if (P%yes_test_only) then
      call print_parameters(P)
      stop
    end if
    call purge_ARG(ARG)
  end subroutine read_parameters

  subroutine read_molecule_param(P, ARG, is_parsed)
    !! read molecular system parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    real(long), parameter :: P_MINIMUM_R = 1e-4_long
    integer :: i, j, k, l, N_nuc, N_nuc_count
    integer :: pos1, pos2
    integer :: INFO, INFO2
    character(len=80) :: nuc_name, basis_string
    character(len=80), allocatable :: basis_string_nuc(:)
    character(len=512) :: nao_file, occ_string
    real(long) :: x, y, z, R, D
    logical :: yes_angstrom
    !
    ! Do you need more than s, p, d, f orbitals?
    ! Now, the limit is l <= 10: s, p, d, f, g, h, i, j, k, l, m orbitals
    !
    integer, parameter :: MAX_lmax_nao = 10
    !
    ! set initial values for numbers of electrons
    !
    P%basisSetOverlapLim = ZERO
    P%yes_gaussian = .false.
    P%gto_file_path = ""
    call getenv("XMOLECULE_BASIS_DIR", P%gto_dir)
    P%gto_dir=trim(P%gto_dir)//'/'
    P%system_name = ""
    P%charge = 0
    P%N_elec = -1         ! it remains -1 if not specified in input files
    P%N_a = -1            ! ditto --> N_a is not used in HFS calculation
    P%N_b = -1            ! ditto --> N_b is not used in HFS calculation
    P%multp = -1           ! ditto --> multiplicity is not considered
    occ_string = ""
    basis_string = ""
    nao_file = ""
    yes_angstrom = .false.        ! geometry is given in atomic unit (bohr)
    P%yes_qmmm = .false.   ! calculates nuclear repulsion between partial charges
    !
    ! find # of nuclei and set # of electrons
    !
    N_nuc_count = 0
    N_nuc = 1     ! an atom if no geometry is specified
    R = ZERO      ! internuclear distance used if no geometry is specified
    nuc_name = ""
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-name", "-system", "-system_name"); P%system_name = trim(ARG%string(i));
       case ("-c", "-charge"); P%charge = int(ARG%value(i))
       case ("-n_e", "-n_elec", "-e", "-elec"); P%N_elec = int(ARG%value(i))
       case ("-n_a", "-na", "-n_alpha"); P%N_a = int(ARG%value(i))
       case ("-n_b", "-nb", "-n_beta"); P%N_b = int(ARG%value(i))
       case ("-atom", "-a", "-nuc", "-nucleus"); nuc_name = trim(ARG%string(i))
       case ("-n_atom", "-n_nuc"); N_nuc = int(ARG%value(i))
       case ("-")      ! find # of nuclei
        read (ARG%string(i), *, iostat=INFO) nuc_name, x, y, z
        if (INFO .ge. 0) then
          N_nuc_count = N_nuc_count + 1
        else
          is_parsed(i) = .False.
        end if
       case ("-occ", "-occupation"); occ_string = trim(ARG%string(i))
       case ("-basis", "-basis_string", "-nao", "-nao_string")
        basis_string = trim(ARG%string(i))
       case ("-basis_file", "-nao_file"); nao_file = trim(ARG%string(i))
       case ("-r"); R = ARG%value(i)
       case ("-singlet"); P%multp = 1
       case ("-doublet"); P%multp = 2
       case ("-triplet"); P%multp = 3
       case ("-quadruplet"); P%multp = 4
       case ("-quintuplet"); P%multp = 5
       case ("-multp", "-mult", "-multi", "-multiplicity"); P%multp = int(ARG%value(i))
       case ("-unit")
        if (ARG%string(i) .eq. "angstrom" .or. ARG%string(i) .eq. "angs" .or. &
          ARG%string(i) .eq. "Angstrom" .or. ARG%string(i) .eq. " Angs") then
          yes_angstrom = .true.
        else if (ARG%string(i) .eq. "bohr" .or. ARG%string(i) .eq. "au" ) then
          yes_angstrom = .false.
        else
          write(P%OUTPUT,'("unit ",a, "not understood")') trim(ARG%string(i))
          stop
        end if
       case ("-angstrom", "-angs")
        yes_angstrom = ARG%yes(i)
       case ("-basisoverlaplim", "-basissetoverlaplim"); P%basisSetOverlapLim = ARG%value(i)
       case ("-gaussian_basis", "-gto")
        P%yes_gaussian = .true.
        call parse_gto(P, ARG%string(i), P%gto_file_path, P%gto_file_path_nuc)
       case ("-gaussian_basis2", "-gto2")
        call parse_gto(P, ARG%string(i), P%gto_file_path2, P%gto_file_path2_nuc)
       case ("-qmmm"); P%yes_qmmm = ARG%yes(i)
       case default; is_parsed(i) = .False.
      end select
    end do
    if (N_nuc_count .gt. 0) N_nuc = N_nuc_count    ! overwritten with counted N_nuc
    !
    ! memory allocation
    !
    P%N_nuc = N_nuc
    allocate (P%x_nuc(N_nuc), P%y_nuc(N_nuc), P%z_nuc(N_nuc))
    allocate (P%charge_nuc(N_nuc), P%mass_nuc(N_nuc), P%fcharge_nuc(N_nuc), P%sigma_nuc(N_nuc))
    allocate (P%nao_file_nuc(N_nuc))
    allocate (P%N_nao_table(0:MAX_lmax_nao, 1:N_nuc))
    allocate (P%nao_name(N_nuc))
    allocate (P%nuc_name(N_nuc))
    allocate (basis_string_nuc(N_nuc))
    P%nao_file_nuc(:) = ""
    basis_string_nuc(:) = ""
    !
    ! read out x, y, z, and charge of nuclei: the 2nd sweep!!
    !   eg) C  1.0  0.0 -1.0 -> ( 1,0, 0.0, -1.0 ), charge=6, mass=12.01
    !   information of charge and mass can be obtained from atom.f90
    !
    j = 0
    do i = 1, ARG%N
      if (ARG%option(i) .ne. "-") cycle
      read (ARG%string(i), *, iostat=INFO) nuc_name, x, y, z
      if (INFO .ge. 0) then
        j = j + 1
        P%x_nuc(j) = x
        P%y_nuc(j) = y
        P%z_nuc(j) = z
        call convert_to_uppercase(nuc_name) ! convert atom types into uppercase
        P%nuc_name(j) = trim(nuc_name)
        P%sigma_nuc(j) = -ONE ! default value for point charges, better to compare with 0 float
        if (nuc_name .eq. "X") then
          P%charge_nuc(j) = 0
          P%mass_nuc(j) = ZERO
          read (ARG%string(i), *, iostat=INFO2) nuc_name, x, y, z, &
            P%fcharge_nuc(j)
          !    this is a plan for counterpoise correction (not yet working)
          !      else if(nuc_name(1:1)==".")then
          !        call get_atomic_species( nuc_name(2:), P%charge_nuc(j), P%mass_nuc(j))
          !        P%fcharge_nuc(j)=ZERO
        else if (nuc_name .eq. "XG") then
          P%charge_nuc(j) = 0
          P%mass_nuc(j) = ZERO
          read (ARG%string(i), *, iostat=INFO2) nuc_name, x, y, z, &
            P%fcharge_nuc(j), P%sigma_nuc(j)
        else
          call get_atomic_species(nuc_name, P%charge_nuc(j), P%mass_nuc(j))
          P%fcharge_nuc(j) = dble(P%charge_nuc(j))
        end if
        !
        ! read basis_string and nao_filename (if they are given)
        !
        if (basis_string_nuc(j) .eq. "" .and. nuc_name(1:1) .ne. "X") &
          read (ARG%string(i), *, iostat=INFO2) nuc_name, x, y, &
          z, basis_string_nuc(j), P%nao_file_nuc(j)
      end if
    end do
    if (P%yes_gaussian .and. allocated(P%gto_file_path_nuc)) then
      if (size(P%gto_file_path_nuc, 1) .ne. count(P%nuc_name .ne. "X")) &
        call error_message("number of atoms and number of gaussian basis set specification does not match")
    end if
    !
    ! if no geometry is specified, we may consider
    !   1) atom,
    !   2) diatomic molecule, or
    !   3) triangular-shaped triatomic molecule.
    !
    if (j .eq. 0) then                    ! if no geometry is specified
      if (nuc_name .eq. "") then          ! put the H atom
        P%charge_nuc(:) = 1
        P%mass_nuc(:) = ATOMIC_MASS(1)
        P%nuc_name(:) = "H"
      else                                ! put the specified atom
        P%charge_nuc(:) = get_atomic_number(nuc_name)
        P%mass_nuc(:) = ATOMIC_MASS(P%charge_nuc(:))
      end if
      if (N_nuc .eq. 1) then              ! for the atomic case
        P%x_nuc(1) = ZERO; P%y_nuc(1) = ZERO; P%z_nuc(1) = ZERO
      else if (N_nuc .eq. 2) then         ! for homonuclear diatomic molecules
        if (R .eq. ZERO) then
          if (P%charge_nuc(1) .eq. 1) R = 1.400_long  ! for H2+
          if (P%charge_nuc(1) .eq. 7) R = 2.072_long  ! for N2
          if (P%charge_nuc(1) .eq. 9) R = 2.668_long  ! for F2
        end if
        D = R*HALF
        P%x_nuc(1) = ZERO; P%y_nuc(1) = ZERO; P%z_nuc(1) = D
        P%x_nuc(2) = ZERO; P%y_nuc(2) = ZERO; P%z_nuc(2) = -D
      else if (N_nuc .eq. 3) then         ! for triangular H3++
        if (R .eq. ZERO) R = 1.68_long    ! default internuclear distance = 1.68
        D = R/sqrt(THREE)
        P%x_nuc(1) = D; P%y_nuc(1) = ZERO; P%z_nuc(1) = ZERO
        P%x_nuc(2) = -HALF*D; P%y_nuc(2) = HALF*R; P%z_nuc(2) = ZERO
        P%x_nuc(3) = -HALF*D; P%y_nuc(3) = -HALF*R; P%z_nuc(3) = ZERO
      else
        call error_message("When no geometry is specified, N_atom <= 3")
      end if
      !
      ! set up fcharge_nuc... who introduced P%fcharge_nuc?!
      !
      P%fcharge_nuc(:) = dble(P%charge_nuc(:))
    end if
    !
    ! unit conversion: from angstrom to a.u.
    !
    if (yes_angstrom) then
      P%x_nuc(:) = P%x_nuc(:)*ANG2AU
      P%y_nuc(:) = P%y_nuc(:)*ANG2AU
      P%z_nuc(:) = P%z_nuc(:)*ANG2AU
    end if
    !
    ! set up basis functions (and a basis file) for each atom
    !
    if (basis_string .ne. "") then
      pos1 = 1
      do k = 1, P%N_nuc
        if (k .lt. P%N_nuc) then
          pos2 = index(basis_string(pos1:), ",")
          basis_string_nuc(k) = basis_string(pos1:pos1 + pos2 - 2)
        else
          pos2 = 0
          basis_string_nuc(k) = basis_string(pos1:)
        end if
        pos1 = pos1 + pos2
      end do
      !    basis_string_nuc(:) = basis_string
    end if
    do j = 1, N_nuc
      if (basis_string_nuc(j) .ne. "" .and. P%charge_nuc(j) .ne. 0) then
        P%nao_name(j) = basis_string_nuc(j)
      else
        P%nao_name(j) = "minimal"
      end if
    end do
    if (nao_file .ne. "") P%nao_file_nuc(:) = nao_file
    do j = 1, N_nuc
      if (P%charge_nuc(j) .eq. 0) then
        P%N_nao_table(:, j) = 0
      else
        if (P%yes_gaussian .and. allocated(P%gto_file_path_nuc)) then
          call put_gto_basis_set(P%charge_nuc(j), P%gto_file_path_nuc(j), P%N_nao_table(0:, j))
        else if (P%yes_gaussian) then
          call put_gto_basis_set(P%charge_nuc(j), P%gto_file_path, P%N_nao_table(0:, j))
        else if (basis_string_nuc(j) .eq. "") then
          call put_minimal_basis_set(P%charge_nuc(j), P%N_nao_table(0:, j))
        else
          call put_basis_function(basis_string_nuc(j), P%N_nao_table(0:, j))
        end if
        if (P%nao_file_nuc(j) .eq. "" .and. .not. P%yes_gaussian ) &
          P%nao_file_nuc(j) = trim(ATOMIC_NAME(P%charge_nuc(j)))//".nao"
      end if
    end do
    deallocate (basis_string_nuc)

    do i = 1, N_nuc
      do j = i + 1, N_nuc
        if (sqrt((P%x_nuc(i) - P%x_nuc(j))**2 + &
          (P%y_nuc(i) - P%y_nuc(j))**2 + &
          (P%z_nuc(i) - P%z_nuc(j))**2) .lt. P_MINIMUM_R) &
          stop 'NUCLEI ARE ON TOP OF EACH OTHER!'
      end do
      if(sum(P%N_nao_table(0:, i)) == 0 .and. P%charge_nuc(i) /= 0)then
        write(STDERR, '(a,x,i4,": ",a)') "There are no basis function for atom ", i, P%nuc_name(i)
        stop 'BASIS FUNCTIONS MISSING'
      end if
    end do
    !
    ! calculate total # of basis functions
    ! considering l: currently it's given as 2l+1 (5D).
    !                however, you may consider the following (6D).
    !        5D   6D
    !    s    1    1
    !    p    3    3 : ( x y z )
    !    d    5    6 : ( xx yy zz xy yz zx )
    !    f    7   10 : ( xxx yyy zzz xxy xxz xyy xyz xzz yyz yzz )
    !
    P%N_basis = 0
    do l = 0, MAX_lmax_nao
      P%N_basis = P%N_basis + (2*l + 1)*sum(P%N_nao_table(l, :))
    end do
    !
    ! determine # of electrons
    !
    if (P%N_a .ne. -1 .and. P%N_b .ne. -1) P%N_elec = P%N_a + P%N_b
    if (P%N_elec .eq. -1) P%N_elec = sum(P%charge_nuc(:)) - P%charge
    !
    ! determine N_orb (different orbitals or subshells)
    ! and the electronic configuration (occupation numbers in diff. orbitals)
    ! - if occ_string is specified, then P%N_elec is overwritten.
    ! - if occ_string is not given, then
    !   - if the atomic case, follow the Aufbau principle.
    !   - if the molecular case, P%N_occ(:) is given as 2 for all.
    !
    if (occ_string .ne. "") then       ! when occ_string is specified
      call convert_to_lowercase(occ_string)
      P%N_orb = len_trim(occ_string)
      allocate (P%N_occ(P%N_orb))
      do j = 1, P%N_orb
        select case (occ_string(j:j))
         case ("a"); P%N_occ(j) = 10
         case ("b"); P%N_occ(j) = 11
         case ("c"); P%N_occ(j) = 12
         case ("d"); P%N_occ(j) = 13
         case ("e"); P%N_occ(j) = 14  ! let's cover the case of f orbitals.
         case ("f"); P%N_occ(j) = 15  ! well... it is like hex. number.
         case default    ! 1...9 becomes an integer; otherwise it is zero.
          read (occ_string(j:j), *, iostat=INFO) P%N_occ(j)
        end select
      end do
      P%N_elec = sum(P%N_occ(:))
    else                                  ! when occ_string is not specified
      if (N_nuc .eq. 1 .and. .FALSE.) then              ! the is the atomic case
        !
        ! P%N_orb is determined and P%N_occ is allocated
        ! inside the Aufbau_principle subroutine.
        !
        call Aufbau_principle(P%N_elec, P%N_orb, P%N_occ)
      else                                ! this is the molecular case
        P%N_orb = int(P%N_elec/2)     ! N_orb = half of N_elec
        if (P%multp .eq. 3 .and. mod(P%N_elec, 2) .eq. 0) P%N_orb = P%N_orb + 1
        if (mod(P%N_elec, 2) .ne. 0) P%N_orb = P%N_orb + 1
        allocate (P%N_occ(P%N_orb))
        P%N_occ(:) = 2                    ! doubly occupied if N_elec is even
        if (mod(P%N_elec, 2) .ne. 0) P%N_occ(P%N_orb) = 1  ! singly occupied
        if (P%multp .eq. 3 .and. mod(P%N_elec, 2) .eq. 0) P%N_occ(P%N_elec/2 - 1:) = 1
      end if
    end if

    !
    ! Up to now, N_elec is determined. Let's set up charge according to N_elec.
    !
    P%charge = sum(P%charge_nuc(:)) - P%N_elec
    !
    ! set up multiplicity and determine spin-up and -down electrons
    ! ... they may be not used in HFS calculations anyway.
    !
    if (P%multp .eq. -1 .or. mod(P%multp - 1, 2) .ne. mod(P%N_elec, 2)) then
      P%multp = mod(P%N_elec, 2) + 1
    end if
    if (P%N_a .ne. -1) then
      P%N_b = P%N_elec - P%N_a
    else if (P%N_b .ne. -1) then
      P%N_a = P%N_elec - P%N_b
    else
      P%N_a = int((P%N_elec + (P%multp - 1))/2)
      P%N_b = int((P%N_elec - (P%multp - 1))/2)
    end if
    if (P%N_a .lt. P%N_b) then     ! make N_a >= N_b always
      P%N_a = P%N_b
      P%N_b = P%N_elec - P%N_b
    end if
    if (P%N_elec .lt. 0 .or. P%N_a .lt. 0 .or. P%N_b .lt. 0) &
      call error_message("The number of electrons should be greater than zero.")
    !
    ! calculate nucleus-nucleus repulsion energy
    !
    P%E_nuc = nuclear_repulsion_E(P%charge_nuc(:), &
      P%x_nuc(:), P%y_nuc(:), P%z_nuc(:))
  end subroutine read_molecule_param

  subroutine parse_gto(P, string, gto_file_path, gto_file_path_nuc)
    type(Param), intent(in) :: P
    character(len=:), allocatable, intent(out) :: gto_file_path
    character(len=:), allocatable, intent(out) :: gto_file_path_nuc(:)
    character(len=*), intent(in) :: string

    character(len=:), allocatable :: temp_gto_nuc(:)
    logical :: file_exists
    integer :: j, l

    if (index(string, ',') .ne. 0) then
      ! separate basis set for each atom
      j = size(parse_string_array(string, ','), 1)
      temp_gto_nuc = parse_string_array(string, ',')
      l = 0
      do j = 1, size(temp_gto_nuc)
        l = max((len_trim(P%gto_dir) + len_trim(temp_gto_nuc(j)) + 7), l)
      end do
      allocate (character(l) :: gto_file_path_nuc(size(temp_gto_nuc)))
      allocate (character(l*size(gto_file_path_nuc)) :: gto_file_path)

      do j = 1, size(temp_gto_nuc)
        gto_file_path_nuc(j) = trim(temp_gto_nuc(j))
        INQUIRE (FILE=gto_file_path_nuc(j), EXIST=file_exists)
        if (.not. file_exists) then
          gto_file_path_nuc(j) = trim(P%gto_dir)//trim(gto_file_path_nuc(j))
          INQUIRE (FILE=gto_file_path_nuc(j), EXIST=file_exists)
          if (.not. file_exists) then
            gto_file_path_nuc(j) = trim(gto_file_path_nuc(j))//'.gbs'
            INQUIRE (FILE=gto_file_path_nuc(j), EXIST=file_exists)
            if (.not. file_exists) then
              call error_message("Basis set not found")
            end if
          end if
        end if
        if (j .eq. 1) then
          gto_file_path = trim(gto_file_path_nuc(j))
        else
          gto_file_path = trim(gto_file_path)//","//trim(gto_file_path_nuc(j))
        end if
      end do! j
      deallocate(temp_gto_nuc)
    else ! string does not contain ',', one basis set for all atoms
      allocate (character(len_trim(P%gto_dir) + len_trim(string) + 6) :: gto_file_path)
      gto_file_path = string
      INQUIRE (FILE=gto_file_path, EXIST=file_exists)
      if (.not. file_exists) then
        gto_file_path = trim(P%gto_dir)//trim(gto_file_path)
        INQUIRE (FILE=gto_file_path, EXIST=file_exists)
        if (.not. file_exists) then
          gto_file_path = trim(gto_file_path)//'.gbs'
          INQUIRE (FILE=gto_file_path, EXIST=file_exists)
          if (.not. file_exists) then
            call error_message("Basis set not found")
          end if
        end if
      end if
    end if
  end subroutine parse_gto

  subroutine put_gto_basis_set(atom_charge, filename, N_per_l)
    !! this is equivalent to put_minimal_basis_set
    !! returns an array of shell per angularmomentum    
    integer, intent(in) :: atom_charge
    character(len=*), intent(in) :: filename
    integer, intent(out) :: N_per_l(0:)
    integer :: status
    integer :: MAX_lmax
    integer, parameter :: BASIS_DATA = 13
    integer :: i
    character(len=80) :: line
    character(len=6) :: atom_name
    character(len=4) :: sym
    logical :: parseAtom
    integer :: n_prim
    real(long) :: norm
    real(long) :: exp, coef, coef_p
    parseAtom = .False.
    N_per_l(0:) = 0
    MAX_lmax = size(N_per_l) - 1
    open (BASIS_DATA, file=trim(filename), action="read", form="formatted")
    read_loop: do
      read (BASIS_DATA, "(a)", iostat=status) line
      if (status .lt. 0) exit read_loop
      if (line(1:1)=='!') cycle read_loop
      if (trim(line) =='') cycle read_loop
      if (trim(line) .eq. "****") then
        parseAtom = .False.
        cycle read_loop
      end if
      if (.not. parseAtom) then
        backspace (BASIS_DATA)
        read (BASIS_DATA, "(a,i10)", iostat=status) atom_name, i
        if (status .lt. 0) exit read_loop
        parseAtom = .True.
        if (atom_name .eq. ATOMIC_NAME(atom_charge)) then ! atom name matches
          read_bf: do
            read (BASIS_DATA, "(a4,i3,f7.2)", iostat=status) sym, n_prim, norm
            if (status .lt. 0 .or. trim(sym) .eq. "****") then
              backspace (BASIS_DATA)
              exit read_bf
            end if
            call convert_to_lowercase(sym)
            select case (sym)
             case ("s"); N_per_l(0) = N_per_l(0) + 1
             case ("p"); N_per_l(1) = N_per_l(1) + 1
             case ("sp"); N_per_l(0) = N_per_l(0) + 1; N_per_l(1) = N_per_l(1) + 1
             case ("d"); N_per_l(2) = N_per_l(2) + 1
             case ("f"); N_per_l(3) = N_per_l(3) + 1
             case ("g"); N_per_l(4) = N_per_l(4) + 1
             case ("h"); N_per_l(5) = N_per_l(5) + 1
             case ("i"); N_per_l(6) = N_per_l(6) + 1
             case default;
              backspace (BASIS_DATA)
              read (BASIS_DATA, "(a)", iostat=status) line
              write (STDERR, *) 'line:'
              write (STDERR, *) line
              write (STDERR, *) 'ERROR unrecognized AM'
              stop
            end select
            do i = 1, n_prim
              read (BASIS_DATA, "(f15.7,8x,f15.7,8x,f15.7)", iostat=status) exp, coef, coef_p
            end do
          end do read_bf
        end if
      end if
    end do read_loop
    close (BASIS_DATA)
  end subroutine put_gto_basis_set

  subroutine put_minimal_basis_set(charge, N_nao_each)
    !! places minimal nao functions and counts the shells
    integer, intent(in) :: charge
    integer, intent(out) :: N_nao_each(0:)
    integer :: nl, n, l, num, MAX_lmax_nao
    MAX_lmax_nao = size(N_nao_each) - 1
    N_nao_each(0:) = 0
    !
    ! let's do the Aufbau principle --> See atom.f90
    !
    nl = 0; num = 0
    loop: do
      nl = nl + 1               ! nl = n + l
      do n = 1, nl
        l = nl - n
        if (l .gt. MAX_lmax_nao) stop "put_minimal_basis_set(): l is too high!"
        if (l .ge. n) cycle
        N_nao_each(l) = N_nao_each(l) + 1
        num = num + (4*l + 2)
        if (num .ge. charge) exit loop
      end do
    end do loop
  end subroutine put_minimal_basis_set

  subroutine put_basis_function(basis_string, N_nao_each)
    !! counts the number of shells for the basis functions
    character(len=*), intent(in) :: basis_string
    integer, intent(out) :: N_nao_each(0:)
    integer :: i, l, s, MAX_lmax_nao, INFO
    character(len=1) :: a
    character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
    MAX_lmax_nao = size(N_nao_each) - 1
    N_nao_each(0:) = 0
    s = 1
    do i = 1, len(trim(basis_string))
      a = basis_string(i:i)
      if (index(lowercase, a) .ne. 0) then
        select case (a)
         case ('s'); l = 0
         case ('p'); l = 1
         case ('d'); l = 2
         case default; l = iachar(a) - iachar('f') + 3
        end select
        if (l .gt. MAX_lmax_nao) stop "put_basis_function(): l is too high!"
        !
        ! read the number right before spdf characters
        !
        read (basis_string(s:i - 1), *, iostat=INFO) N_nao_each(l)
        s = i + 1
      end if
    end do
  end subroutine put_basis_function

  subroutine read_grid_param(P, ARG, is_parsed)
    !! read grid parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i
    !
    ! set initial grid parameters: #1 for global and #2 for atomic
    !
    P%N = 30
    P%L = ONE          ! mapping parameter for radial grids
    P%rmax = 50.0_long  ! the maximum radius for the atomic grids
    P%lmax = 10           ! # of partial waves for each atom
    P%N_GG = 0
    P%L_GG = -ONE
    P%rmax_GG = -ONE
    P%lmax_GG = 0
    P%yes_overlap = .false.
    P%yes_squeeze = .false.
    P%yes_global = .false.
    P%yes_atomic = .true.
    P%yes_cartesian = .false.
    P%type_radial = "LGL"
    P%type_radial_GG = "LGL"
    !P%type_radial     = "algebraic"
    !P%type_radial_GG  = "algebraic"
    P%type_angular = "lebedev"
    P%type_angular_GG = "lebedev"
    P%dx_GG = ONE
    !
    ! determine grid parameters
    !
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-n", "-n_a", "-n_atomic"); P%N = int(ARG%value(i))
       case ("-ng", "-n_g", "-n_global", &
         "-ngg"); P%N_GG = int(ARG%value(i))
       case ("-l", "-l_a", "-l_atomic"); P%L = int(ARG%value(i))
       case ("-lg", "-l_g", "-l_global", &
         "-lgg"); P%L_GG = int(ARG%value(i))
       case ("-rmax", "-rmax_a", "-rmax_atomic", "-rm");
        P%rmax = ARG%value(i)
       case ("-rmaxg", "-rmax_g", "-rmax_global", &
         "-rmax_gg", "-rm_gg")
        P%rmax_GG = ARG%value(i)
       case ("-lmax", "-lmax_a", "-lmax_atomic", "-lm", &
         "-lmax_ag", "-lmax_ang")
        P%lmax = int(ARG%value(i))
       case ("-lmaxg", "-lmax_g", "-lmax_global", "-lm_g", &
         "-lmaxgg", "-lm_gg")
        P%lmax_GG = int(ARG%value(i))
       case ("-overlap"); P%yes_overlap = ARG%yes(i)
       case ("-squeeze"); P%yes_squeeze = ARG%yes(i)
       case ("-radial", "-radial_a", "-radial_atomic")
        P%type_radial = trim(ARG%string(i))
       case ("-angular", "-angular_a", "-angular_atomic")
        P%type_angular = trim(ARG%string(i))
       case ("-cartesian", "-cart", "-box", "-uniform")
        P%yes_global = ARG%yes(i)
        P%yes_cartesian = ARG%yes(i)
       case ("-global", "-background", "-encapsule", "-encapsuled", &
         "-encapsulate", "-encapsulated", "-embed", "-embeded")
        P%yes_global = ARG%yes(i)
        select case (ARG%string(i))
         case ("box", "uniform", "cart", "cartesian", &
           "rect", "rectangular"); P%yes_cartesian = ARG%yes(i)
         case default; P%type_radial_GG = trim(ARG%string(i))
        end select
       case ("-atomic", "-individual"); P%yes_atomic = ARG%yes(i)
        if (P%yes_atomic .eqv. .false.) P%yes_global = .true.
       case ("-noatomic", "-noindividual"); P%yes_global = .true.
        P%yes_atomic = .false.
       case ("-OCE", "-oce"); P%yes_global = .true.
        P%yes_atomic = .false.
        P%yes_cartesian = .false.
       case ("-radial_G", "-radial_g", "-radial_global", &
         "-radial_GG", "-radial_gg"); P%yes_global = .true.
        P%type_radial_GG = trim(ARG%string(i))
       case ("-angular_G", "-angular_g", "-angular_global", &
         "-angular_GG", "-angular_gg"); P%yes_global = .true.
        P%type_angular_GG = trim(ARG%string(i))
       case ("-dx", "-dx_G", "-dx_g", "-dx_GG", "-dx_gg", "-dx_global")
        P%dx_GG = ARG%value(i)
       case default; is_parsed(i) = .False.
      end select
    end do
    if (P%N_nuc .eq. 1) then      ! use atomic parameters for the atomic case
      P%yes_global = .true.
      P%yes_atomic = .false.
      if (P%N_GG .eq. 0) P%N_GG = P%N
      if (P%L_GG .eq. -ONE) P%L_GG = P%L
      if (P%rmax_GG .eq. -ONE) P%rmax_GG = P%rmax
      if (P%lmax_GG .eq. 0) P%lmax_GG = P%lmax
    else                          ! set initial parameters for global grids
      if (P%N_GG .eq. 0) P%N_GG = 10
      if (P%L_GG .eq. -ONE) P%L_GG = THREE
      if (P%rmax_GG .eq. -ONE) P%rmax_GG = ZERO
      if (P%lmax_GG .eq. 0) P%lmax_GG = 8
    end if
    if (.not. P%yes_global .and. .not. P%yes_atomic) &
      call error_message("There are no grids at all!")
    !  if ( P%yes_global .and. P%yes_overlap == .false. .and. P%rmax == ZERO ) &
    !    call error_message( "rmax should not be zero when global is used." )
    if (P%yes_overlap .and. P%yes_squeeze) &
      call error_message("overlap and squeeze may not be used at the same time.")
    !
    ! put the same angular grids for all atoms at this moment!!!
    !
    allocate (P%lmax_AG(P%N_nuc))
    P%lmax_AG(:) = P%lmax
  end subroutine read_grid_param

  subroutine read_XFEL_param(P, ARG, is_parsed)
    !! read XFEL parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i, keyln, le, koff
    !
    ! initialize
    !
    P%pulse_type = 3                      ! 3: Gaussian
    P%PE = 12000.0_long*EV2AU           ! photon energy = 12 keV
    P%PE_width = 0.01*P%PE              ! photon energy spectral width (stdev)
    P%N_time = 20000
    P%FWHM = 10_long*1e-15_long*SEC2AU       ! == 10 fs
    P%pulse_multiple = 8.0_long           ! 8 times of FWHM
    P%pulse_duration = -ONE               ! to be calculated
    P%t_pulse_offset = ZERO
    P%t_start_au = ZERO                   ! to be calculated
    P%t_end_au = ZERO                     ! to be calculated
    P%dt_au = ZERO                        ! to be calculated
    P%n_photons = 1e12_long               ! # of photons
    P%width = ONE*1.0e-6_long*M2AU    ! == 1 micrometer
    P%fluence = -ONE                      ! to be calculated by n_photons/width**2
    P%direction = (/ZERO, ZERO, ONE/)               ! photon momentum direction (default along z-axis)
    P%pol_axis = (/ONE, ZERO, ZERO/)                ! photon polarization axis (default along x-axis)
    P%max_theta = 3.142_long
    P%yes_spectra = .FALSE.
    P%fix_geom = .False.
    !
    ! set up parameters
    !
    P%pulse_file_path = ""
    is_parsed(:) = .True.
    keyln = 0
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-laser", "-x_ray", "-pulse", "-pulse_shape", "-pulse_type")
        !
        ! Special cases : read_in (type=10) and spiky (type=9)
        ! you also need to specify pulse file path
        !
        if (ARG%string(i) (1:8) .eq. "read_in:") then; P%pulse_type = 10
          keyln = 8
        end if
        if (ARG%string(i) (1:6) .eq. "spiky:") then; P%pulse_type = 9
          keyln = 6
        end if
        if (P%pulse_type .ge. 9) then
          koff = index(ARG%string(i), '-')
          if (koff .le. 0) STOP "specify offset"
          read (ARG%string(i) (koff + 1:), *) P%pulse_file_offset
          le = (koff - 1) - keyln
          if (le .gt. len(P%pulse_file_path)) STOP "too large pathlen"
          P%pulse_file_path(1:le) = ARG%string(i) (keyln + 1:keyln + le)

        end if
        select case (ARG%string(i))
         case ("gaussian"); P%pulse_type = 3
         case ("flat-top", "flattop"); P%pulse_type = 6
        end select
       case ("-PE", "-pe", "-photon", "-photon_E", "-photon_energy", &
         "-omega", "-omega_in"); P%PE = ARG%value(i)*EV2AU
       case ("-pe_width", "-pe_stdev"); P%PE_width = ARG%value(i)*EV2AU
       case ("-pe_fwhm"); P%PE_width = ARG%value(i) * EV2AU / TWO / sqrt( TWO * log( TWO ) )
       case ("-direction", "-dir");
        select case (ARG%string(i))
         case ("x", "X"); P%direction = (/ONE, ZERO, ZERO/)
         case ("y", "Y"); P%direction = (/ZERO, ONE, ZERO/)
         case ("z", "Z"); P%direction = (/ZERO, ZERO, ONE/)
        end select
       case ("-pol", "-pol_axis");
        select case (ARG%string(i))
         case ("x", "X"); P%pol_axis = (/ONE, ZERO, ZERO/)
         case ("y", "Y"); P%pol_axis = (/ZERO, ONE, ZERO/)
         case ("z", "Z"); P%pol_axis = (/ZERO, ZERO, ONE/)
        end select
       case ("-fix_geom"); P%fix_geom = ARG%yes(i)
       case ("-N_time", "-Ntime"); P%N_time = int(ARG%value(i))
       case ("-t_start", "-t_s", "-t_start_au"); P%t_start_au = ARG%value(i)
       case ("-t_end", "-t_e", "-t_end_au"); P%t_end_au = ARG%value(i)
       case ("-dt", "-dt_au"); P%dt_au = ARG%value(i)
       case ("-FWHM", "-fwhm", "-tau", "-temporal_width")
        P%FWHM = ARG%value(i)*1.0e-15_long*SEC2AU           ! fs -> a.u.
       case ("-pulse_duration", "-duration")
        P%pulse_duration = ARG%value(i)*1.0e-15_long*SEC2AU ! fs -> a.u.
       case ("-pulse_multiple", "-multiple", "-p_multple", "-p_mult", &
         "-pulse_length", "-length")
        P%pulse_multiple = ARG%value(i)
       case ("-photons", "-N_photons", "-n_photons", "-N_photon", "-n_photon")
        P%n_photons = ARG%value(i)
       case ("-width", "-spatial_width", "-focal_width")
        P%width = ARG%value(i)*1.0e-6_long*M2AU             ! micron -> a.u.
       case ("-area", "-spatial_area", "-focal_area")
        P%width = sqrt(ARG%value(i))*1.0e-6_long*M2AU     ! micron -> a.u.
       case ("-fluence"); P%fluence = ARG%value(i)        ! in a.u.
       case ("-pulse_offset")
        P%t_pulse_offset = ARG%value(i)*1.0e-15_long*SEC2AU ! fs -> a.u.
       case ("-spectra")
        P%yes_spectra = .TRUE.
       case default;
        is_parsed(i) = .False.
      end select
    end do
    !
    ! assign (or re-assign) some parameters if unspecified
    !
    if (P%pulse_duration .lt. ZERO) P%pulse_duration = P%FWHM  ! default: 10 fs
    if (P%pulse_type .eq. 3) then         ! 3: Gaussian
      if (P%t_start_au .eq. ZERO .and. P%t_end_au .eq. ZERO) then
        P%t_start_au = -(P%pulse_multiple*HALF)*P%pulse_duration
        P%t_end_au = +(P%pulse_multiple*HALF)*P%pulse_duration
      end if
      if (P%t_pulse_offset .ne. ZERO) then
        P%t_start_au = P%t_pulse_offset + P%t_start_au
        P%t_end_au = P%t_pulse_offset + P%t_end_au
      end if
    else if (P%pulse_type .eq. 6) then    ! 6: flat-top
      if (P%t_start_au .eq. ZERO .and. P%t_end_au .eq. ZERO) then
        P%t_start_au = ZERO
        P%t_end_au = P%pulse_duration
      else
        P%pulse_duration = P%t_end_au - P%t_start_au
      end if
    else if (P%pulse_type .eq. 9) then    ! 9: Spiky
      !
      ! special treatment for pulse_type 9 (spiky) and 10 (read_in)
      !
      if (P%t_start_au .eq. ZERO .and. P%t_end_au .eq. ZERO) then
        P%t_start_au = P%t_pulse_offset
        P%t_end_au = P%t_pulse_offset + P%pulse_duration
      end if
    else if (P%pulse_type .eq. 10) then    ! 10: Readin
      if (P%t_start_au .eq. ZERO .and. P%t_end_au .eq. ZERO) then
        P%t_start_au = P%t_pulse_offset
        P%t_end_au = P%t_pulse_offset + P%pulse_duration
      end if
    end if
    if (P%dt_au .eq. ZERO) then
      P%dt_au = (P%t_end_au - P%t_start_au)/dble(P%N_time)
    else
      P%N_time = int((P%t_end_au - P%t_start_au)/P%dt_au)
    end if
    if (P%fluence .lt. ZERO) then
      P%fluence = P%n_photons/P%width**2
    else
      P%n_photons = P%fluence*P%width**2
    end if
  end subroutine read_XFEL_param

  subroutine read_DFT_param(P, ARG, is_parsed)
    !! read DFT parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    character(len=20) :: pot_name, x_name, c_name
    character(len=2048) :: buffer
    character(len=:), allocatable :: strarray(:)
    integer, allocatable :: intarray(:)
    integer :: i, j, pos1, pos2, npos, k, n, le
    logical :: yes_alpha, yes_beta
    real(long) :: dum
    !
    ! set initial values: the DFT type (XC potential)
    !
    pot_name = ""
    x_name = "Slater"
    c_name = "none"
    P%alpha = ZERO; yes_alpha = .false.
    P%beta = ZERO; yes_beta = .false.
    P%yes_OEP_KLI = .false.
    P%yes_Latter = .false.
    P%yes_load_pot = .false.
    P%lmax_MP = 5                 ! for multipole expansion
    !
    ! set initial SCF parameters
    !
    P%SCF_max_iter = 100
    P%MCSCF_gradient_conv = 1.0e-4_long
    P%SCF_energy_conv = 1.0e-6_long
    P%SCF_density_conv = 1.0e-6_long
    P%SCF_mixing = 0.2_long       ! it means 20% of new and 80% of old
    ! this value will be automatically adapted during SCF
    P%SCF_SR_den_conv = 1.0e-2_long
    P%SCF_SR_pot_conv = 1.0e-2_long
    P%yes_DIIS = .true.
    P%yes_CPHFDIIS = .true.
    P%yes_direct_scf = .false.
    P%yes_mmap_teints = .false.
    P%yes_DDIFF = .false.
    P%n_ddiff_save = 8
    P%DIIS_nvector = 8
    P%yes_chk_fid_ini = .false.
    P%yes_population = .false.
    P%yes_population_L = .false.
    P%yes_stability = .false.
    P%yes_dump_ints = .false.
    P%yes_use_arpack = .false.
    P%yes_use_davidson = .false.
    P%yes_use_krylov = .false.
    P%davBlock = 10
    P%davMax = 200
    P%yes_density_cube = .false.
    P%yes_spin_density_cube = .false.
    P%Nvoxel = 100*100*100
    P%yes_LevelShift = .false.
    P%yes_LevelShiftDynamic = .false.
    P%SwitchToDIISCriterion(:) = ZERO ! energy/density criterion for switching to diis: 0.0 means never
    P%SwitchToDIISIter = HUGE(P%SwitchToDIISIter) ! deadline at which DIIS is forced to start: HUGE means no deadline
    P%SwitchToRSTRCTCriterion = ZERO ! energy criterion for switching to RSTRCT: 0.0 means never
    P%SwitchLevelShiftOffCriterion = ZERO ! energy criterion for switching Level-Shifting Off: 0.0 means never
    P%yes_RSTRCT = .false.
    P%yes_MOM = .false.
    P%MOMtype = 3 ! default MOM type

    P%yes_SOSCF = .false.
    P%SOSCF_start = (/1.0e-03_long, 1.0e-03_long, 50.0_long/)    ! (1) orbital Gradient threshold to start SOSCF ; (3) iteration number deadline  i.e. orbGrad<(1).or.NINT(3)<=iter  start SOSCF

    P%LShift = ZERO
    P%EPS_overlap = 1e-6_long  ! accuracy control for overlap pairs
    P%EPS_overlap_single = ZERO       ! same as EPS_overlap as a default
    P%EPS_00 = 0.01_long      ! accuracy control: Q_A / r < EPS_00   <May.14.2015> !Please be careful of this EPS.. possibly zero!
    P%EPS_lm = 0.5e-4_long; ! accuracy control: 1% for d_lm / d_00  <Feb.25.2015> modified from "1e-3_long"
    !                       due to Energy convergency difficulty in large-size calculations
    P%EPS_r = 1e-5_long ! accuracy control: 0.01% for d_lm(r) / d_lm(rmax)

    P%EPS_Auger = 1e-6_long ! accuracy control for Auger calculation if product of MO coefficients
    ! smaller than that, skip calculation
    P%EPS_teint = 1e-9_long ! Schwarz twoelectron integral screening
    P%r_cutoff = ZERO     ! default: not assigned
    !  P%r_elec_stat = ZERO  ! default: not assigned
    P%yes_cutoff = .false.
    P%yes_overlap_twocenter = .false.
    P%R_softcore = ZERO   ! default: not assigned

    P%state = 1 ! default 1st state
    P%nstates = -1 ! default -1 number of states
    P%yes_HF = .false.
    P%yes_CIS = .false.
    P%yes_CI = .false.
    P%yes_MCSCF = .false.
    P%yes_state_average = .false.
    P%mcscf_type = 2 ! default perform quasi newton bfgs steps
    P%T_eV = ZERO         ! default: temperature T_eV = 0 K = 0 eV
    P%Ncore = -1         ! default: -1 number of core orbitals will be automatically detected
    P%yes_thermal = .false.
    P%yes_broadening = .false.
    P%yes_valgs_occ = .false.
    P%yes_valgs_occ_part = .false.
    P%yes_gradient = .false.
    P%yes_orbital_energy_gradient = .false.
    P%yes_atomic_potential = .false.

    P%yes_nat_orb = .False.
    P%yes_sym = .FALSE.
    P%state_irrep = -1
    P%yes_auto_average = .FALSE.
    P%yes_average_rates = .FALSE.
    P%yes_average_csf_rates = .True.
    P%yes_all_occs = .FALSE.
    P%yes_all_multps = .TRUE.
    P%yes_openshell = .False.
    P%ROHF_variant = 1
    P%yes_unrestricted = .False.
    !
    ! determine DFT and SCF parameters
    !
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
        !
        ! determine the DFT type
        !
       case ("-direct")
        P%yes_direct_scf = ARG%yes(i)
      case ("-mmap_teints")
        P%yes_mmap_teints = ARG%yes(i)
       case ("-t", "-type", "-xc_type", "-xc", "-dft", "-functional")
        pot_name = trim(ARG%string(i))
       case ("-x_type", "-x", "-exchange"); x_name = trim(ARG%string(i))
       case ("-c_type", "-c", "-correlation"); c_name = trim(ARG%string(i))
       case ("-alpha"); P%alpha = ARG%value(i); yes_alpha = .true.
       case ("-beta"); P%beta = ARG%value(i); yes_beta = .true.
       case ("-kli", "-oep", "-sic", "-oep-kli", "-oep-kli-sic")
        P%yes_OEP_KLI = ARG%yes(i)
       case ("-tail", "-latter"); P%yes_Latter = ARG%yes(i)
       case ("-pot", "-potential", "-load_pot", "-load_potential");
        P%yes_load_pot = ARG%yes(i)
       case ("-lmax_mp"); P%lmax_MP = int(ARG%value(i))
        !
        ! determine SCF parameters
        !
       case ("-iteration", "-iter", "-max", "-scf_max", "-scf_max_iter")
        P%SCF_max_iter = int(ARG%value(i))
       case ("-noiteration", "-noiter", "-no_iteration", "-no_iter")
        P%SCF_max_iter = 0
       case ("-scf", "-scf_conv", "-scf_energy_conv", "-energy_conv", &
         "-conv", "-real_conv", "-ener_conv")
        P%SCF_energy_conv = ARG%value(i)
       case ("-scf_density_conv", "-density_conv", "-dens_conv")
        P%SCF_density_conv = ARG%value(i)
       case ("-mcscf_gradient_conv", "-gradient_conv", "-grad_conv")
        P%MCSCF_gradient_conv = ARG%value(i)
       case ("-scf_mixing", "-mixing"); P%SCF_mixing = ARG%value(i)
       case ("-scf_sr_den_conv", "-scf_sr_den")
        P%SCF_SR_den_conv = ARG%value(i)
       case ("-scf_sr_pot_conv", "-scf_sr_pot")
        P%SCF_SR_pot_conv = ARG%value(i)

       case ("-scf_acceptable_convergence", "-acceptable")
        if (.not. allocated(P%SCF_acceptable_convergence)) allocate (P%SCF_acceptable_convergence(2));
        P%SCF_acceptable_convergence = ZERO;
        buffer = ARG%string(i);
        pos1 = 1;
        n = 0;
        le = len(trim(buffer))
        do while (pos1 .lt. le);
          pos2 = index(buffer(pos1:), ",");
          if (pos2 .lt. 1) then
            pos2 = le;
          else;
            pos2 = pos1 + pos2 - 2;
          end if   !! if index==2 ',' is at pos1+pos2-1.
          n = n + 1;
          read (buffer(pos1:pos2), *) P%SCF_acceptable_convergence(n)
          pos1 = pos2 + 2
        end do
        if (n .eq. 1) P%SCF_acceptable_convergence(2) = P%SCF_acceptable_convergence(1)              !!  regarded as the same

       case ("-diis"); P%yes_DIIS = ARG%yes(i)
       case ("-cphfdiis"); P%yes_CPHFDIIS = ARG%yes(i)
       case ("-nodiis"); P%yes_DIIS = .not. ARG%yes(i)
       case ("-ddiff"); P%yes_DDIFF = ARG%yes(i)
       case ("-levelshift"); P%yes_LevelShift = .true.
        P%Lshift = ARG%value(i)
       case ("-switchtodiis")

        buffer = ARG%string(i); pos1 = 1; n = 0; le = len(trim(buffer))
        do while (pos1 .lt. le);
          pos2 = index(buffer(pos1:), ","); if (pos2 .lt. 1) then; pos2 = le; else; pos2 = pos1 + pos2 - 2; end if   !! if index==2 ',' is at pos1+pos2-1.
          read (buffer(pos1:pos2), *) dum;
          if (dum .lt. ONE) then;
            n = n + 1;
            P%SwitchToDIISCriterion(n) = dum;
          else;
            P%SwitchToDIISIter = NINT(dum);
          end if
          pos1 = pos2 + 2; !next to comma
        end do
        if (n .eq. 1) P%SwitchToDIISCriterion(2) = HUGE(P%SwitchToDIISCriterion(2))    !! invalidate the second condition
       case ("-switchtorstrct")
        P%SwitchToRSTRCTCriterion = ARG%value(i)
       case ("-switchlevelshiftoff")
        P%SwitchLevelShiftOffCriterion = ARG%value(i)
       case ("-rstrct"); P%yes_RSTRCT = ARG%yes(i)
       case ("-mom"); P%yes_MOM = ARG%yes(i)
       case ("-momtype"); P%MOMtype = int(ARG%value(i))
       case ("-openshell"); P%yes_openshell = ARG%yes(i)
       case ("-rohf_variant"); P%ROHF_variant = int(ARG%value(i))
       case ("-soscf");
        P%yes_SOSCF = ARG%yes(i)
        P%yes_DIIS = .not. ARG%yes(i)
       case ("-soscf_start")
        buffer = ARG%string(i); pos1 = 1; n = 0; le = len(trim(buffer))
        do while (pos1 .lt. le);
          pos2 = index(buffer(pos1:), ","); if (pos2 .lt. 1) then; pos2 = le; else; pos2 = pos1 + pos2 - 2; end if   !! if index==2 ',' is at pos1+pos2-1.
          read (buffer(pos1:pos2), *) dum;
          if (dum .lt. ONE) then;
            n = n + 1;
            P%SOSCF_start(n) = dum;
          else;
            P%SOSCF_start(3) = NINT(dum);
          end if
          pos1 = pos2 + 2; !next to comma
        end do
        if (n .eq. 1) P%SOSCF_start(2) = 9.999e+03                                        !! invalidate the second condition

       case ("-f_occ", "-frac_occ", "-frac_occupation");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%f_occ(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%f_occ(j)
          else
            read (buffer(pos1:), *) P%f_occ(j)
          end if
          pos1 = pos2 + pos1
        end do

       case ("-f_occ_a", "-frac_occ_a", "-frac_occupation_a");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%f_occ_a(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%f_occ_a(j)
          else
            read (buffer(pos1:), *) P%f_occ_a(j)
          end if
          pos1 = pos2 + pos1
        end do

       case ("-f_occ_b", "-frac_occ_b", "-frac_occupation_b");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%f_occ_b(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%f_occ_b(j)
          else
            read (buffer(pos1:), *) P%f_occ_b(j)
          end if
          pos1 = pos2 + pos1
        end do

       case ("-freeze");
        buffer = ARG%string(i)
        if (trim(buffer) .eq. "all") then
          allocate (P%freeze(P%N_basis))
          do pos1 = 1, P%N_basis
            P%freeze(pos1) = pos1
          end do
        else
          npos = 0
          pos1 = 1
          do
            pos2 = index(buffer(pos1:), ",")
            IF (pos2 .eq. 0) THEN
              npos = npos + 1
              exit
            end if
            npos = npos + 1
            pos1 = pos2 + pos1
          end do
          allocate (P%freeze(npos))
          pos1 = 1
          do j = 1, npos
            pos2 = index(buffer(pos1:), ",")
            if (j .lt. npos) then
              read (buffer(pos1:pos1 + pos2 - 2), *) P%freeze(j)
            else
              read (buffer(pos1:), *) P%freeze(j)
            end if
            pos1 = pos2 + pos1
          end do! j
        end if
       case ("-sym");
        P%yes_sym = ARG%yes(i)
       case ("-state_irrep");
        P%state_irrep = int(ARG%value(i))
       case ("-irrep_mo");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%irrep_MO(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%irrep_MO(j)
          else
            read (buffer(pos1:), *) P%irrep_MO(j)
          end if
          pos1 = pos2 + pos1
        end do
       case ("-localize");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do

        allocate (P%localize(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%localize(j)
          else
            read (buffer(pos1:), *) P%localize(j)
          end if
          pos1 = pos2 + pos1
        end do
       case ("-fix_nuc");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0 .or. pos2 .gt. index(buffer(pos1:), ":")) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%fix_nuc_MO(npos))
        pos1 = 1
        do j = 1, npos
          if (j .lt. npos) then
            pos2 = index(buffer(pos1:), ",")
            read (buffer(pos1:pos1 + pos2 - 2), *) P%fix_nuc_MO(j)
          else
            pos2 = index(buffer(pos1:), ":")
            read (buffer(pos1:pos1 + pos2 - 2), *) P%fix_nuc_MO(j)
          end if
          pos1 = pos2 + pos1
        end do
        allocate (P%fix_nuc_n(npos))
        do j = 1, npos
          if (j .lt. npos) then
            pos2 = index(buffer(pos1:), ",")
            read (buffer(pos1:pos1 + pos2 - 2), *) P%fix_nuc_n(j)
          else
            read (buffer(pos1:), *) P%fix_nuc_n(j)
          end if
          pos1 = pos2 + pos1
        end do
       case ("-auto_average");
        P%yes_auto_average = ARG%yes(i)
       case ("-average_rates");
        P%yes_average_rates = ARG%yes(i)
       case ("-average_csf_rates");
        P%yes_average_csf_rates = ARG%yes(i)
       case ("-all_occs");
        P%yes_all_occs = ARG%yes(i)
       case ("-all_multps");
        P%yes_all_multps = ARG%yes(i)
       case ("-active_space_mo");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer(:), ',', '-'), 1)
        allocate (P%active_space_MO(k))
        P%active_space_MO(:) = parse_int_array(buffer(:), ',', '-')
       case ("-final_active_space_mo");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer(:), ',', '-'), 1)
        allocate (P%final_active_space_MO(k))
        P%final_active_space_MO(:) = parse_int_array(buffer(:), ',', '-')
       case ("-general_active_space");
        buffer = ARG%string(i)
        k = size(parse_string_array(buffer(:), ':'), 1)
        allocate (P%gas(k))
        allocate (character(100) :: strarray(k))
        strarray(:) = ""
        strarray = parse_string_array(buffer(:), ':')
        do j = 1, k
          n = size(parse_int_array(strarray(j)//" ", ',', '-'), 1)
          allocate (intarray(n))
          intarray = parse_int_array(strarray(j)//" ", ',', '-')
          P%gas(j)%nmin = intarray(1)
          P%gas(j)%nmax = intarray(2)
          allocate (P%gas(j)%orbitals(n - 2))
          P%gas(j)%orbitals(:) = intarray(3:)
          deallocate (intarray)
        end do
        deallocate (strarray)
      case ("-final_general_active_space", "-auger_general_active_space");
        buffer = ARG%string(i)
        k = size(parse_string_array(buffer(:), ':'), 1)
        allocate (P%final_gas(k))
        allocate (character(100) :: strarray(k))
        strarray(:) = ""
        strarray = parse_string_array(buffer(:), ':')
        do j = 1, k
          n = size(parse_int_array(strarray(j)//" ", ',', '-'), 1)
          allocate (intarray(n))
          intarray = parse_int_array(strarray(j)//" ", ',', '-')
          P%final_gas(j)%nmin = intarray(1)
          P%final_gas(j)%nmax = intarray(2)
          allocate (P%final_gas(j)%orbitals(n - 2))
          P%final_gas(j)%orbitals(:) = intarray(3:)
          deallocate (intarray)
        end do
        deallocate (strarray)
       case ("-restricted_active_space");
        buffer = ARG%string(i)
        k = size(parse_string_array(buffer(:), ':'), 1)
        allocate (P%ras(k))
        allocate (character(100) :: strarray(k))
        strarray(:) = ""
        strarray = parse_string_array(buffer(:), ':')
        do j = 1, k
          n = size(parse_int_array(strarray(j), ','), 1)
          allocate (intarray(n))
          intarray = parse_int_array(strarray(j), ',')
          P%ras(j)%nmin = intarray(1)
          P%ras(j)%nmax = intarray(2)
          allocate (P%ras(j)%orbitals(n - 2))
          P%ras(j)%orbitals(:) = intarray(3:)
          deallocate (intarray)
        end do
        deallocate (strarray)
       case ("-average_occ_mo");
        ! parse the average_occ_MO string:
        ! first count the number of average_MO_spaces
        ! separated by ':'
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ":")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        ! allocate array of av_occ structures
        allocate (P%av_occ(npos))
        ! for each entry in ':' separated string
        ! parse ',' separated list of MO indices
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ":")
          if (j .lt. npos) then
            k = size(parse_int_array(buffer(pos1:pos1 + pos2 - 2), ','), 1)
            allocate (P%av_occ(j)%MO(k))
            P%av_occ(j)%MO(:) = parse_int_array(buffer(pos1:pos1 + pos2 - 2), ',')
          else
            k = size(parse_int_array(buffer(pos1:), ','), 1)
            allocate (P%av_occ(j)%MO(k))
            P%av_occ(j)%MO(:) = parse_int_array(buffer(pos1:), ',')
          end if
          pos1 = pos2 + pos1
        end do

       case ("-levelshiftdynamic"); P%yes_LevelShift = .True.
        P%yes_LevelShiftDynamic = .True.
        P%Lshift = ARG%value(i)
       case ("-diis_nvec"); P%DIIS_nvector = int(ARG%value(i))
       case ("-mo_fid_ini"); P%yes_chk_fid_ini = ARG%yes(i)
       case ("-pop", "-population", "-mulliken", "-mulliken_population");
        P%yes_population = ARG%yes(i)
       case ("-pop_l", "-population_l");
        P%yes_population_L = ARG%yes(i)
        P%yes_population = ARG%yes(i)
       case ("-stab", "-stability"); P%yes_stability = ARG%yes(i);
       case ("-unrestricted"); P%yes_unrestricted = ARG%yes(i)
       case ("-hf"); P%yes_HF = ARG%yes(i)
       case ("-cis");
        P%yes_CIS = ARG%yes(i);
       case ("-ci"); P%yes_CI = ARG%yes(i)
       case ("-mcscf"); P%yes_MCSCF = ARG%yes(i)
       case ("-nat_orb"); P%yes_nat_orb = ARG%yes(i)
       case ("-nat_trans_orb");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer(:), ',', '-'), 1)
        allocate (P%nat_trans_orb(k))
        P%nat_trans_orb(:) = parse_int_array(buffer(:), ',', '-')
       case ("-mcscf_type"); P%mcscf_type = int(ARG%value(i))
       case ("-ci_coef");
        k = size(parse_real_array(ARG%string(i), ','), 1)
        allocate (P%ci_coef(k))
        P%ci_coef = parse_real_array(ARG%string(i), ',')
       case ("-state_average");
        P%yes_state_average = .True.
        if( trim(ARG%string(i)) == 'all')then
          ! we do not allocate P%SA_weights; this implies we average over all states equally
        else
          k = size(parse_real_array(ARG%string(i), ','), 1)
          allocate (P%SA_weights(k))
          P%SA_weights = parse_real_array(ARG%string(i), ',')
          P%SA_weights(:) = P%SA_weights(:)/sum(P%SA_weights)
        end if
       case ("-state"); P%state = int(ARG%value(i))
       case ("-nstates"); P%nstates = int(ARG%value(i))
       case ("-arpack"); P%yes_use_arpack = ARG%yes(i)
       case ("-davidson"); P%yes_use_davidson = ARG%yes(i)
       case ("-krylov"); P%yes_use_krylov = ARG%yes(i)
       case ("-davidson_block"); P%davBlock = int(ARG%value(i))
       case ("-davidson_max"); P%davMax = int(ARG%value(i))
       case ("-dump_ints"); P%yes_dump_ints = ARG%yes(i)
       case ("-Nvoxel","-nvoxel"); P%nvoxel = int(ARG%value(i))
       case ("-density_cube"); P%yes_density_cube = ARG%yes(i)
       case ("-spin_density_cube"); P%yes_spin_density_cube = ARG%yes(i)
       case ("-mo_cube");
        buffer = ARG%string(i)
        npos = 0
        pos1 = 1
        do
          pos2 = index(buffer(pos1:), ",")
          IF (pos2 .eq. 0) THEN
            npos = npos + 1
            exit
          end if
          npos = npos + 1
          pos1 = pos2 + pos1
        end do
        allocate (P%mo_cube(npos))
        pos1 = 1
        do j = 1, npos
          pos2 = index(buffer(pos1:), ",")
          if (j .lt. npos) then
            read (buffer(pos1:pos1 + pos2 - 2), *) P%mo_cube(j)
          else
            read (buffer(pos1:), *) P%mo_cube(j)
          end if
          pos1 = pos2 + pos1
        end do
       case ("-t_ev", "-temp_ev", "-temperature_ev", &
         "-t_e", "-t_elec", "-temp", "-temperature")
        P%T_eV = ARG%value(i)
        P%yes_thermal = .true.
       case ("-t_k", "-temp_k", "-temperature_k")
        P%T_eV = ARG%value(i)/EV2K
        P%yes_thermal = .true.
       case ("-ncore", "-n_core")
        P%Ncore = int(ARG%value(i))
       case ("-broadening")
        P%yes_broadening = .true.
        P%T_eV = ARG%value(i)
       case ("-valgs_occ", "-gs_occ", "-val_gs_occ", "-gsocc")
        P%yes_valgs_occ = ARG%yes(i)
        P%yes_valgs_occ_part = .not. ARG%yes(i)
      case ("-valgs_occ_part", "-gs_occ_part", "-val_gs_occ_part", "-gsocc_part")
        P%yes_valgs_occ = ARG%yes(i)
        P%yes_valgs_occ_part = ARG%yes(i)
        !
        ! accuracy control
        !   EPS_overlap, EPS_single: AO pair on A ==> EPS
        !   EPS_00 : Q_A / r = sqrt(4pi) d_00 / r ==> EPS0
        !   EPS_lm : d_lm / d_00                  ==> EPS1
        !   EPS_r  : d_lm(r) / d_lm(rmax)         ==> EPS2
        !
       case ("-eps_auger")
        P%EPS_Auger = ARG%value(i)
       case ("-eps_teint")
        P%EPS_teint = ARG%value(i)
       case ("-eps", "-eps_o", "-eps_pair", &
         "-eps_overlap", &
         "-eps_offdiagonal", "-eps_ij")
        P%EPS_overlap = ARG%value(i)
       case ("-eps_single", "-eps_overlap1", &
         "-eps_overlap_single", &
         "-eps_diagonal", "-eps_ii")
        P%EPS_overlap_single = ARG%value(i)
       case ("-eps_lm", "-eps1"); P%EPS_lm = ARG%value(i)
       case ("-eps_r", "-eps2"); P%EPS_r = ARG%value(i)
       case ("-eps_00", "-eps0"); P%EPS_00 = ARG%value(i)
       case ("-r_cutoff", "-cutoff_r", "-cutoff_radius", "-r_c", "-rc", &
         "-r_short_range", "-r_short", "-r_shortrange", "-r_c1", "-rc1")
        P%r_cutoff = ARG%value(i)
        if (P%r_cutoff .ne. ZERO) P%yes_cutoff = .true.
       case ("-cutoff"); P%yes_cutoff = ARG%yes(i)

       case ("-overlap_twocenter"); P%yes_overlap_twocenter = ARG%yes(i)
       case ("-r_softcore"); P%R_softcore = ARG%value(i)
        !
        ! gradient calculation
        !
       case ("-gradient", "-grad", "-grad_e_tot", "-grad_total_e")
        P%yes_gradient = ARG%yes(i)
       case ("-orbital_energy_gradient", "-orb_grad", "-grad_e_orb", "-cegrad")
        P%yes_orbital_energy_gradient = ARG%yes(i)
        !
        ! misc.
        !
       case ("-atomic_potential", "-atomic_pot", "-atom_pot")
        P%yes_atomic_potential = ARG%yes(i)
       case default;
        is_parsed(i) = .False.
      end select
    end do
    !
    ! specify DFT xc name
    !
    if (pot_name .ne. "") then
      if (pot_name .eq. "BLYP") then
        x_name = "B"
        c_name = "LYP"
      else
        x_name = pot_name
        c_name = "none"
      end if
    end if
    !
    ! specify DFT xc functionalARG%option(i) /= "-"
    !
    P%x_type = 0; P%c_type = 0
    do j = 1, size(FUNCTIONAL_X_NAME)
      if (x_name .eq. FUNCTIONAL_X_NAME(j)) P%x_type = j
    end do
    do j = 1, size(FUNCTIONAL_C_NAME)
      if (c_name .eq. FUNCTIONAL_C_NAME(j)) P%c_type = j
    end do
    if (P%x_type .eq. 0) call error_message &
      (trim(x_name)//" is not applicable for the exchange functional.")
    if (P%c_type .eq. 0) call error_message &
      (trim(c_name)//" is not applicable for the correlation functional.")
    !
    ! initialize parameters for the particular func. if they are not specified
    !
    select case (FUNCTIONAL_X_NAME(P%x_type))
     case ("Xa")
      if (.not. yes_alpha) P%alpha = TWO/THREE
     case ("LB")
      if (.not. yes_beta) P%beta = 0.05_long
     case ("LBa")
      if (.not. yes_alpha) P%alpha = 1.19_long
      if (.not. yes_beta) P%beta = 0.01_long
    end select
    !
    ! set SCF convergence: it can be specified by either 1e-5 or 5.
    ! especially in the command line, -SCF_conv=1e-5 doesn't work
    ! instead, you may use -SCF_conv=0.00001 or -SCF_conv=5
    !
    if (P%SCF_energy_conv .gt. ONE) P%SCF_energy_conv = 10_long**(-P%SCF_energy_conv)
    if (P%SCF_density_conv .gt. ONE) P%SCF_density_conv = 10_long**(-P%SCF_density_conv)
    !
    ! set EPS_overlap_single
    !
    if (P%EPS_overlap_single .eq. ZERO) P%EPS_overlap_single = P%EPS_overlap
    ! Set open shell
    if (ANY(P%N_occ .eq. 1)) then
      P%yes_openshell = .TRUE.
    end if
    ! If ground state occupation, shell may not be open
    if (P%yes_valgs_occ ) then
      P%yes_openshell = .FALSE.
    end if
  end subroutine read_DFT_param

  subroutine read_MC_param(P, ARG, is_parsed)
    !! read Monte-Carlo control parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i, k
    real(long) :: dbuf2(2)
    character(len=2048) :: buffer
    !
    ! initialize
    !
    P%MC_max_events = 100000
    P%MC_step = 100
    P%MC_conv = 1.0e-4_long     !  convergence criteria for MC-averaged quantities  (max-dev < this value)
    P%MC_dt_ratio = 0.1_long    !  (this value)/ rateSum  is the max estimate of dt ( also consider  dt_au  due to  pulse shape )
    P%yes_pathway = .false.
    P%yes_adapt_for_all_AO = .false.

    P%yes_reUse_PreviousResults = .false.
    P%MC_fileTableSizeMax = 2048
    P%MC_nfileSaveMax = 1024
    P%MC_nVisitThreshold = 1
    P%MC_NprocessToBeSaved = 200

    P%MC_dumpOutput = 99
    P%MC_maxNfaultRatio = 0.10_long

    P%MC_nStepUpl_Cache = 3          ! max # of steps to be cached : 0=initial state only,  1= up to 1st step from intl Nocc...

    P%wfFileName = ""
    P%wfaFileName = ""
    P%wfbFileName = ""
    P%basisFileName = ""
    P%ref_wf = ""
    P%ref_basis = ""

    P%Fluorescence_Energy_Threshold_eV = 10    ! MC transition cut-off:any 'F' process with E_eV < (this value) is removed from the table
    P%MC_ratesum_cutoff = 0.024E-06  ! process with more than a nano sec events will be cut off

    P%N_MC_max = -1               ! limits MC step if it is set>0  if it is negative, no limit.  (default=nolimit)

    P%SCF_N_iter_Large = P%SCF_max_iter + 100  ! by default no warning is issued
    P%XATOM_PATH = ""
    call getenv("XATOM_PATH", P%XATOM_PATH)
    P%WORK_DIR_PATH = ""
    P%ini_File_Path = ""
    P%avgSqrOvlp_TOL = (/-ONE, -ONE/)
    P%guess_overlap_TOL = (/0.50_long, -ONE/)
    if (P%yes_valgs_occ) P%guess_overlap_TOL = (/0., 0./)
    P%MO_fidelity_TOL = (/-ONE, -ONE/)
    P%N_pathway = 0            ! 0 := no pathway print out

    P%tStepMD = 0.24_long  ! default : 10 a.u.
    P%MD_maxtime = -ONE    ! default : unlimited

    !
    ! set up parameters
    !
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-n_mc_max"); P%N_MC_max = int(ARG%value(i))
       case ("-mc_conv"); P%MC_conv = ARG%value(i)
       case ("-mc_max_events", "-max_events"); P%MC_max_events = int(ARG%value(i))
       case ("-mc_step", "-step", "-every"); P%MC_step = int(ARG%value(i))
       case ("-mc_dt_ratio", "-dt_ratio", "-ratio")
        P%MC_dt_ratio = ARG%value(i)
       case ("-adapt_all_ao"); P%yes_adapt_for_all_AO = ARG%yes(i)
       case ("-pathway", "-traj", "-trajectory", "-mc_pathway", &
         "-mc_traj", "-mc_trajectory"); !  P%yes_pathway = ARG%yes(i)
        ! we accept Integer input
        select case (ARG%string(i) (1:1));
         case ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9');
          P%N_pathway = int(ARG%value(i));
        end select;
        if (P%N_pathway .le. 0) P%yes_pathway = ARG%yes(i)
        if (P%yes_pathway) P%n_pathway = 1           ! this is formally equivalent to  always-yes
        ! write(*,*)"#yes_Pathway:",P%N_pathway,P%yes_pathway;
       case ("-mc_ratesum_cutoff"); P%MC_ratesum_cutoff = ARG%value(i)
       case ("-reuse_previousresults", "-mc_cache"); P%yes_reUse_PreviousResults = ARG%yes(i)
       case ("-mc_filetablesizemax"); P%MC_fileTableSizeMax = int(ARG%value(i))
       case ("-mc_nfilesavemax"); P%MC_nfileSaveMax = int(ARG%value(i))
       case ("-mc_nvisitthreshold"); P%MC_nVisitThreshold = int(ARG%value(i))
       case ("-mc_nprocesstobesaved"); P%MC_NprocessToBeSaved = int(ARG%value(i))

       case ("-mc_dumpoutput", "-mc_dump"); P%MC_dumpOutput = int(ARG%value(i))
       case ("-mc_maxnfaultratio"); P%MC_maxNfaultRatio = ARG%value(i)
       case ("-scf_n_iter_large"); P%SCF_N_iter_Large = int(ARG%value(i))
       case ("-mc_nstepupl_cache"); P%MC_nStepUpl_Cache = int(ARG%value(i))
       case ("-seed", "-mc_seeds");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer, ','), 1)
        allocate (P%MC_seeds(k))
        P%MC_seeds = parse_int_array(buffer, ',')
       case ("-tstepmd"); P%tStepMD = ARG%value(i)
       case ("-md_maxtime"); P%MD_maxtime = ARG%value(i)
       case ("-internuc_dist_upl", "-internuc_dist_upl_au");
        allocate (P%internuc_dist_upl_au(P%N_nuc*(P%N_nuc - 1)/2));
        buffer = ARG%string(i)
        P%internuc_dist_upl_au = parse_real_array(buffer, ',')
       case ("-internuc_dist_lwl_au");
        allocate (P%internuc_dist_lwl_au(P%N_nuc*(P%N_nuc - 1)/2));
        buffer = ARG%string(i)
        P%internuc_dist_lwl_au = parse_real_array(buffer, ',')
       case ("-vini");
        allocate (P%vini(3*P%N_nuc));
        buffer = ARG%string(i)
        P%vini = parse_real_array(buffer, ',')
       case ("-xatom_path", "-xatom_exe"); P%XATOM_PATH = trim(arg%String(i));
       case ("-ini_file_path", "-inifile"); P%ini_File_Path = trim(arg%String(i));
       case ("-work_dir_path", "-work_dir"); P%WORK_DIR_PATH = trim(arg%String(i));
       case ("-fluorescence_energy_threshold_in_ev", "-fluorescence_threshold_ev", "-f_threshold_ev");
        P%Fluorescence_Energy_Threshold_eV = int(ARG%value(i))
       case ("-guess_overlap_tol");
        buffer = ARG%string(i)
        k = size(parse_real_array(buffer, ','))
        if (k .gt. 2) stop 'error: only one or two arguments for option guess_overlap_tol !'
        dbuf2 = parse_real_array(buffer, ',')
        if (k .eq. 1) then
          P%guess_overlap_TOL(1) = dbuf2(1)
        else
          P%guess_overlap_TOL(:) = dbuf2(:)
        end if
       case ("-avgsqrovlp_tol");
        buffer = ARG%string(i)
        k = size(parse_real_array(buffer, ','))
        if (k .gt. 2) stop 'error: only one or two arguments for option avgsqrovlp_tol !'
        dbuf2 = parse_real_array(buffer, ',')
        if (k .eq. 1) then
          P%avgSqrOvlp_TOL(1) = dbuf2(1)
        else
          P%avgSqrOvlp_TOL(:) = dbuf2(:)
        end if
        !
        ! Seriously, is MO_fidelity relevant to Monte-Carlo calculation?
        ! Same questions for most of above options, are they really relevant to MC?
        !
       case ("-mo_fidelity_tol", "-mo_fidelity");
        buffer = ARG%string(i)
        k = size(parse_real_array(buffer, ','))
        if (k .gt. 2) stop 'error: only one or two arguments for option mo_fidelity_tol !'
        dbuf2 = parse_real_array(buffer, ',')
        if (k .eq. 1) then
          P%MO_fidelity_TOL(1) = dbuf2(1)
        else
          P%MO_fidelity_TOL(:) = dbuf2(:)
        end if
       case default; is_parsed(i) = .False.
      end select
    end do
  end subroutine read_MC_param

  subroutine read_process_param(P, ARG, is_parsed)
    !! read process parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i, k
    character(len=2048) :: buffer

    !
    ! initialize
    !
    is_parsed(:) = .True.
    P%yes_scattering = .False.
    P%yes_compton = .False.
    P%yes_ionization_dynamics = .False.
    P%yes_test_fallback = .False.
    P%yes_freezeBadOverlap = .False.
    P%yes_shakeup = .false.
    P%yes_auger = .false.
    P%yes_auger_random = .false.
    P%yes_auger_random_2 = .false.
    P%yes_process_spin = .false.
    P%yes_process_CI = .false.
    P%yes_process_cont = .false.
    P%yes_process_CI_relaxed = .false.
    P%process_CI_relaxed_variant = 1
    ! 1: means calculate transitions matrix elements between different orbital sets (works only with absorption)
    ! 2: means calculate transitions matrix elements in initial orbital set and project configurations into final orbital set configurations
    P%process_relaxed_wf_filename = 'wf_relaxed.dat'
    P%yes_process_CIS = .false.
    P%yes_process_CIS_ini = .false.
    P%yes_process_CISD = .false.
    P%yes_fluorescence = .false.
    P%yes_pcs = .false.
    P%yes_absorption = .false.
    P%yes_absorptioncs = .false.
    P%N_proc_max = 10000  ! max # of processes, used for MC on the fly
    !
    ! set up parameters
    !
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-scattering"); P%yes_scattering = ARG%yes(i)
       case ("-compton"); P%yes_compton = ARG%yes(i)
       case ("-auger"); P%yes_auger = ARG%yes(i)
       case ("-auger_random"); P%yes_auger_random = ARG%yes(i)
       case ("-auger_random2"); P%yes_auger_random_2 = ARG%yes(i)
       case ("-process_spin"); P%yes_process_spin = ARG%yes(i)
       case ("-ionization_dynamics"); P%yes_ionization_dynamics = ARG%yes(i)
       case ("-test_fallback"); P%yes_test_fallback = ARG%yes(i)
       case ("-freeze_bad_overlap"); P%yes_freezeBadOverlap = ARG%yes(i)
       case ("-final_hole");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer, ','), 1)
        allocate (P%final_hole(k))
        P%final_hole = parse_int_array(buffer, ',')
       case ("-cis_hole");
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer, ','), 1)
        allocate (P%cis_hole(k))
        P%cis_hole = parse_int_array(buffer, ',')
       case ("-process_ci");
        P%yes_process_CI = ARG%yes(i)
       case ("-process_cont");
        P%yes_process_cont = ARG%yes(i)
       case ("-process_cis_lim");
        P%yes_process_CI = .True.
        P%yes_process_CIS = .True.
        buffer = ARG%string(i)
        k = size(parse_int_array(buffer, ','), 1)
        if (k .eq. 1) then
          P%CIS_limit(2) = int(ARG%value(i))
          P%CIS_limit(1) = -1
        else if (k .eq. 2) then
          P%CIS_limit(:) = parse_int_array(buffer(:), ',')
        end if
       case ("-process_cis_e_lim");
        P%yes_process_CI = .True.
        P%yes_process_CIS = .True.
        P%CIS_E_limit = ARG%value(i)*EV2AU
       case ("-process_cis_ini");
        P%yes_process_CI = ARG%yes(i);
        P%yes_process_CIS_ini = ARG%yes(i)
       case ("-process_cis");
        P%yes_process_CI = ARG%yes(i);
        P%yes_process_CIS = ARG%yes(i)
       case ("-process_cisd");
        P%yes_process_CI = ARG%yes(i);
        P%yes_process_CISD = ARG%yes(i)
       case ("-process_ci_relaxed_variant");
        P%process_CI_relaxed_variant = int(ARG%value(i))
       case ("-process_ci_relaxed");
        P%yes_process_CI = ARG%yes(i);
        P%yes_process_CI_relaxed = ARG%yes(i);
       case ("-process_relaxed_wf")
        P%process_relaxed_wf_filename = trim(ARG%string(i))
        P%yes_process_CI_relaxed = .True.;
        P%yes_process_CI = .True.;
       case ("-shakeup", "-shake_up", "-shake")
        P%yes_shakeup = ARG%yes(i)
       case ("-fluorescence", "-fluo")
        P%yes_fluorescence = ARG%yes(i)
       case ("-rate", "-rates", "-decay"); P%yes_fluorescence = ARG%yes(i)
        P%yes_auger = ARG%yes(i)
       case ("-pcs", "-photo", "-photoionization", "-photoion")
        P%yes_pcs = ARG%yes(i)
       case ("-absorption"); P%yes_absorption = ARG%yes(i)
       case ("-absorptioncs"); P%yes_absorption = ARG%yes(i); P%yes_absorptioncs = ARG%yes(i)
       case ("-n_process", "-n_process_max", "-n_proc", "-n_proc_max")
        P%N_proc_max = int(ARG%value(i))
       case default; is_parsed(i) = .False.
      end select
    end do
  end subroutine read_process_param

  subroutine read_crystal_param(P, ARG, is_parsed)
    !! read crystalline parameters
    !! - it should be used after read_molecular_param()
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i
    real(long) :: r_c, r_c_relative    ! short-range cutoff (inner region)
    !  real(long) :: r_es, r_es_relative  ! long-range cutoff (electrostatic region)
    logical :: yes_lattice
    P%yes_periodic = .false.      ! default: no crystal calculation
    P%yes_periodic_fuzzy = .false.
    P%N_tr_max = 27               ! default: 27=3^3 operations in translation
    yes_lattice = .false.
    r_c_relative = ZERO
    !  r_es_relative = ZERO
    P%unitcell_x = 10_long        ! default unitcell size: 10 a.u.
    P%unitcell_y = 10_long        ! default unitcell size: 10 a.u.
    P%unitcell_z = 10_long        ! default unitcell size: 10 a.u.
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("-unit")
        if (ARG%string(i) .eq. "lattice" .or. ARG%string(i) .eq. "unitcell" .or. &
          ARG%string(i) .eq. "relative") yes_lattice = .true.
       case ("-periodic", "-periodicity", "-crystal", "-crystalline")
        P%yes_periodic = ARG%yes(i)
       case ("-periodic_fuzzy", "-gradient_fuzzy")
        P%yes_periodic_fuzzy = ARG%yes(i)
       case ("-unitcell", "-unitcell_size", "-unitsize", &
         "-lattice", "-lattice_parameter", "-lattice_param")
        P%unitcell_x = ARG%value(i)
        P%unitcell_y = ARG%value(i)
        P%unitcell_z = ARG%value(i)
       case ("-unitcell_x", "-unitcell_size_x", "-unitsize_x", &
         "-unit_x", "-cell_x", "-lattice_x")
        P%unitcell_x = ARG%value(i)
       case ("-unitcell_y", "-unitcell_size_y", "-unitsize_y", &
         "-unit_y", "-cell_y", "-lattice_y")
        P%unitcell_y = ARG%value(i)
       case ("-unitcell_z", "-unitcell_size_z", "-unitsize_z", &
         "-unit_z", "-cell_z", "-lattice_z")
        P%unitcell_z = ARG%value(i)
       case ("-translation", "-n_translation", "-max_translation", &
         "-n_max_translation", "-n_tr", "-n_max_tr", "-n_tr_max")
        P%N_tr_max = int(ARG%value(i))
        !
        ! ACHTUNG: r_cutoff and r_elec_stat are assigned in SR read_DFT_param
        !
       case ("-r_cutoff_relative", "-cutoff_r_relative", &
         "-cutoff_radius_relative", "-r_c_relative", "-rc_relative", &
         "-r_cutoff_rel", "-cutoff_r_rel", "-r_c_rel", "-rc_rel", &
         "-r_short_range_relative", "-r_short_range_rel", &
         "-r_short_relative", "-r_short_rel", &
         "-r_shortrange_relative", "-r_shortrange_rel")
        r_c_relative = ARG%value(i)
        if (r_c_relative .ne. ZERO) P%yes_cutoff = .true.

       case default; is_parsed(i) = .False.
      end select
    end do
    if (P%yes_periodic) then
      !
      ! If the periodic condition is used, P%r_cutoff must be assigned.
      ! P%r_cutoff should be smaller than a half of the unitcell length
      !                                    ==> NOT TRUE ANYMORE
      ! because a nucleus is counted only ONCE within r < P%r_cutoff.
      !                                    ==> NO MORE MINIMUM IMAGE CONVENTION
      !
      r_c = min(P%unitcell_x, P%unitcell_y, P%unitcell_z)/TWO
      if (P%r_cutoff .eq. ZERO) then
        P%r_cutoff = r_c
        if (r_c_relative .gt. ZERO) P%r_cutoff = P%r_cutoff*r_c_relative
      end if

      if (P%EPS_r .ne. ZERO) &
        call error_message("EPS2=0 is required for periodic calculations.")
    end if
    !
    ! re-adjust molecular geometry -- read_molecule_param() should be called first
    ! and re-calculate nuclear_repulsion_E (WITHOUT periodicity YET!!!)
    !
    if (yes_lattice) then
      P%x_nuc(:) = P%x_nuc(:)*P%unitcell_x
      P%y_nuc(:) = P%y_nuc(:)*P%unitcell_y
      P%z_nuc(:) = P%z_nuc(:)*P%unitcell_z
      P%E_nuc = nuclear_repulsion_E(P%charge_nuc(:), &
        P%x_nuc(:), P%y_nuc(:), P%z_nuc(:))
    end if
  end subroutine read_crystal_param

  subroutine read_misc_param(P, ARG, is_parsed)
    !! read misc. parameters
    type(Param), intent(inout) :: P
    type(Argument), intent(in) :: ARG
    logical, intent(out) :: is_parsed(ARG%N)
    integer :: i, j
    integer, parameter :: OUTPUT = 10
    integer, parameter :: OUTPUT2 = 11
    character(len=*), parameter :: file_default = "results"
    character(len=80) :: file_root
    !
    ! initialize filenames
    !
    P%nmo_print = -1
    P%OUTPUT = STDOUT             ! if not assigned, OUTPUT is STDOUT
    P%OUTPUT2 = STDOUT            ! OUTPUT2 is used for back-compatibility
    file_root = ""
    P%file_out = ""
    P%file_out2 = ""
    P%molden_file = ""
    P%yes_verbose = .false.
    P%yes_plot_potential = .false.
    P%yes_plot_density = .false.
    P%print_level = 2     ! >=2 : standard, 1 : minimal, 0 : zero (error only)
    P%yes_test_only = .false.
    P%yes_print_wf = .false.
    P%yes_ReadIni = .false.
    P%yes_ReadShell = .false.
    P%yes_CoreIni = .false.
    P%yes_SaveIni = .false.
    P%file_data = "process.dat"   ! used for molecular data
    P%wfFileName = "wf.dat"
    P%wfaFileName = "wfa.dat"
    P%wfbFileName = "wfb.dat"
    P%basisFileName = "basis.dat"
    !
    ! determine computational parameters
    !
    is_parsed(:) = .True.
    do i = 1, ARG%N
      select case (ARG%option(i))
       case ("", "-i", "-file", "-input", "-filename", "-root")
        file_root = trim(ARG%string(i))
       case ("-o", "-output"); P%OUTPUT = OUTPUT
        P%file_out = trim(ARG%string(i))
       case ("-o2", "-output2"); P%OUTPUT2 = OUTPUT2
        P%file_out2 = trim(ARG%string(i))
       case ("-v", "-verbose", "--verbose"); P%yes_verbose = ARG%yes(i)
       case ("-print_level"); P%print_level = int(ARG%value(i))
       case ("-nmo_print"); P%nmo_print = int(ARG%value(i))
       case ("-test"); P%yes_test_only = ARG%yes(i)
       case ("-wf", "-wavefunction"); P%yes_print_wf = ARG%yes(i)
       case ("-dump", "--dump"); call show_input(ARG)
       case ("-rini", "-inir"); P%yes_ReadIni = ARG%yes(i)
       case ("-coreini"); P%yes_CoreIni = ARG%yes(i)
       case ("-sini", "-inis"); P%yes_SaveIni = ARG%yes(i)
       case ("-rinish", "-inishr"); P%yes_ReadShell = ARG%yes(i)
       case ("-d", "+d", "-dat", "-data", "-process", &
         "-file_data", "-file_dat"); P%file_data = trim(ARG%string(i))
       case ("-file_wf", "-wffilename"); P%wfFileName = trim(ARG%string(i))
       case ("-file_wfa", "-wfafilename"); P%wfaFileName = trim(ARG%string(i))
       case ("-file_wfb", "-wfbfilename"); P%wfbFileName = trim(ARG%string(i))
       case ("-ref_wf", "-ref"); P%ref_wf = trim(ARG%string(i))
       case ("-ref_basis"); P%ref_basis = trim(ARG%string(i))
       case ("-plot")
        if (ARG%string(i) .eq. "potential" .or. ARG%string(i) .eq. "pot") then
          P%yes_plot_potential = .true.
        else if (ARG%string(i) .eq. "density" .or. ARG%string(i) .eq. "den") then
          P%yes_plot_density = .true.
        end if
       case ("-plot_pot", "-plot_potential"); P%yes_plot_potential = ARG%yes(i)
       case ("-plot_den", "-plot_density"); P%yes_plot_density = ARG%yes(i)
       case ("-molden");
        P%molden_file = trim(ARG%string(i))
       case default; is_parsed(i) = .False.
      end select
    end do
    !
    ! set filename and open output
    !
    if (file_root .eq. "" .and. P%file_out .ne. "") file_root = trim(P%file_out)
    P%filename = file_root
    j = index(file_root, ".", .true.)
    if (j .ne. 0) P%filename = file_root(:j - 1)   ! delete the extension
    if (P%filename .eq. "") P%filename = file_default
    if (P%file_out .eq. "") P%file_out = trim(P%filename)//".out"
    if (P%OUTPUT .ne. STDOUT) open (P%OUTPUT, file=P%file_out)
    if (P%file_out2 .eq. "") P%file_out2 = trim(P%filename)//".out2"
    if (P%OUTPUT2 .ne. STDOUT) open (P%OUTPUT2, file=P%file_out2)
  end subroutine read_misc_param

  subroutine print_parameters(P, title, IOUT)
    !! print all parameters
    type(Param), intent(in) :: P
    character(len=*), intent(in), optional :: title
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    character(len=80) :: string
    character(len=80) :: build_time_str
    character(len=80) :: version_str
    call C_F_string(build_time_str_ptr, build_time_str )
    call C_F_string(version_str_ptr, version_str )
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    !
    ! print the title
    !
    string = "XMOLECULE"
    if (present(title)) then
      string = trim(string)//": "//title
    else if (P%calc_type .eq. 0) then
      string = trim(string)//": electronic structure"
    else if (P%calc_type .eq. 1) then
      string = trim(string)//": ionization dynamics"
    else if (P%calc_type .eq. 2) then
      string = trim(string)//": ionization dynamics by Monte-Carlo"
    else if (P%calc_type .eq. 3) then
      string = trim(string)//": ionization and nuclear dynamics by Monte-Carlo"
    end if
    write (OUTPUT, "('# ',a)") repeat("-", len_trim(string))
    write (OUTPUT, "('# ',a)") trim(string) // ' GIT Version: ' // trim(version_str)
    write (OUTPUT, "('# build_time: ',a)") build_time_str
    write (OUTPUT, "('# ',a)") repeat("-", len_trim(string))
    !
    ! print out all parameters
    !
    call print_molecule_param(P, OUTPUT)
    !call print_grid_param( P, OUTPUT )   ! print grid info with A(:) and G later
    call print_DFT_param(P, OUTPUT)
    if (P%calc_type .ne. 0) then
      call print_XFEL_param(P, OUTPUT)
      if (P%calc_type .ge. 2 ) &   ! MC fly: ion. only or ion. & nuc. dynamics
        call print_MC_param(P, OUTPUT)
    else
      if (P%yes_pcs .or. P%yes_absorptioncs) &
        write (OUTPUT, "('# Photon energy (eV) = ',f12.2)") P%PE*AU2EV
      if (P%yes_absorptioncs) &
        write (OUTPUT, "('# Photon spectral width (stdev) (eV) = ',f12.2,x,'(FWHM) ',f12.2)") &
          P%PE_width * AU2EV, P%PE_width * TWO*sqrt(TWO*log(TWO)) * AU2EV
      if (P%yes_scattering .or. P%yes_compton) then
        write (OUTPUT, "('# Photon energy (eV) = ',f9.2)") P%PE*AU2EV
        write (OUTPUT, "('#   Pulse direction  = ',3f9.2)") P%direction
        write (OUTPUT, "('#   Pol axis  = ',3f9.2)") P%pol_axis
      end if
    end if
    if (P%yes_periodic) call print_crystal_param(P, OUTPUT)
    call print_misc_param(P, OUTPUT)
  end subroutine print_parameters

  subroutine print_molecule_param(P, IOUT)
    !! print molecule parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: i, j, OUTPUT
    real(long) :: dist(P%N_nuc, P%N_nuc)
    integer :: minidx(2)
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    if (P%N_nuc .gt. 1) then
      write (OUTPUT, "('# Molecular geometry (a.u.):')")
      write (OUTPUT, "('#   ',2x,a4,a10,5a15)") "Z", "mass", "x", "y", "z", "charge", "zeta"
      do i = 1, P%N_nuc
        if (P%charge_nuc(i) .ne. 0) then
          write (OUTPUT, "('#   ',a2,i4,f10.4,4f15.8,a15)") &
            ATOMIC_NAME(P%charge_nuc(i)), P%charge_nuc(i), P%mass_nuc(i), &
            P%x_nuc(i), P%y_nuc(i), P%z_nuc(i), P%fcharge_nuc(i), '---'
        else if (P%charge_nuc(i) .eq. 0 .and. P%sigma_nuc(i) .lt. ZERO) then
          write (OUTPUT, "('#   ',a2,a4,f10.4,4f15.8, a15)") &
            "X", "---", P%mass_nuc(i), &
            P%x_nuc(i), P%y_nuc(i), P%z_nuc(i), P%fcharge_nuc(i), '---'
        else if (P%charge_nuc(i) .eq. 0 .and. P%sigma_nuc(i) .gt. ZERO) then
          write (OUTPUT, "('#   ',a2,a4,f10.4,5f15.8)") &
            "XG", "---", P%mass_nuc(i), &
            P%x_nuc(i), P%y_nuc(i), P%z_nuc(i), P%fcharge_nuc(i), P%sigma_nuc(i)
        end if
      end do
      write (OUTPUT, "('#')")
      write (OUTPUT, "('#   the number of nuclei       = ',i11)") P%N_nuc
      write (OUTPUT, "('#   nuclear repulsion energy   = ',f20.8)", &
        advance="no") P%E_nuc
      if (P%yes_periodic) then
        write (OUTPUT, "(1x,a)") "(non-periodic calc.)"
      else
        write (OUTPUT, "()")
      end if
      do i = 1, P%N_nuc
        do j = i + 1, P%N_nuc
          dist(i, j) = dsqrt((P%x_nuc(i) - P%x_nuc(j))**2 + (P%y_nuc(i) - P%y_nuc(j))**2 + (P%z_nuc(i) - P%z_nuc(j))**2)
          dist(j, i) = dist(i, j)
        end do
        dist(i, i) = huge(dist(i, i))
      end do
      minidx = minloc(dist)
      write (OUTPUT, "('#   smallest distance among nuclei=',f20.8, 4x, 'for atom indices',x,i5,x,i5)") &
        dist(minidx(1), minidx(2)), minidx(1), minidx(2)
    else
      if (P%charge .eq. 0) then
        write (OUTPUT, "('# Atom: ',a2)") ATOMIC_NAME(P%charge_nuc(1))
      else
        write (OUTPUT, "('# Atomic ion: ',a)", advance="no") &
          trim(ATOMIC_NAME(P%charge_nuc(1)))
        if (P%charge .gt. 0) then
          write (OUTPUT, "(a)") repeat("+", abs(P%charge))
        else
          write (OUTPUT, "(a)") repeat("-", abs(P%charge))
        end if
      end if
    end if
    write (OUTPUT, "('#   total charge of the system = ',sp,i11)") P%charge
    write (OUTPUT, "('#   the number of electrons    = ',i11)") P%N_elec
    write (OUTPUT, "('#   occupation numbers : ',50i1)") P%N_occ(:)

    if (P%yes_sym) &
      write (OUTPUT, "('#   symmetry : yes')")


    write (OUTPUT, "('#')")
  end subroutine print_molecule_param

!----------------------------------------------------------------------------
! print grid parameters
!----------------------------------------------------------------------------
!subroutine print_grid_param( P, IOUT )
!  type(Param), intent(in) :: P
!  integer, intent(in), optional :: IOUT
!  integer :: OUTPUT
!  OUTPUT = P%OUTPUT;    if ( present(IOUT) ) OUTPUT = IOUT
!  write(OUTPUT, "('# Grid: ')", advance="no")
!  if ( P%yes_atomic ) then
!    if ( P%yes_global ) write(OUTPUT, "('encapsulated ')", advance="no")
!    if ( P%yes_overlap ) write(OUTPUT, "('overlapping ')", advance="no")
!    if ( P%yes_squeeze ) write(OUTPUT, "('squeezed ')", advance="no")
!    write(OUTPUT, "('composite grid system')", advance="no")
!    if ( P%yes_global ) then; write(OUTPUT, "(' (GIG)')")
!    else; write(OUTPUT, "(' (GPG)')"); end if
!    if ( P%N_nuc > 1 ) then
!      write(OUTPUT, "('#   Individual atomic grid')")
!    end if
!    write(OUTPUT, "('#   - radial:  N = ',i6,' | rmax = ')", advance="no") P%N
!    if ( P%rmax == ZERO ) then; write(OUTPUT, "('  N/A')", advance="no")
!    else; write(OUTPUT, "(f5.1)", advance="no") P%rmax; end if
!    write(OUTPUT, "(' | L = ',f5.2,' | mapping = ',a)") P%L, trim(P%type_radial)
!    write(OUTPUT, "('#   - angular: lmax = ',i3,' | type = ',a)", advance="no") &
!      P%lmax, trim(P%type_angular)
!    write(OUTPUT, "(' | lmax_MP = ',i3)")  P%lmax_MP    ! it's not a grid param!
!  else
!    if ( P%yes_cartesian ) then; write(OUTPUT, "('uniform 3D Cartesian')")
!    else; write(OUTPUT, "('OCE (one-center expansion)')"); end if
!  end if
!  if ( P%yes_global ) then
!    write(OUTPUT, "('#   Global grid')")
!    if ( P%yes_cartesian ) then
!      write(OUTPUT, "('#   - cartesian: dx = ',f5.2)") P%dx_GG
!    else
!      write(OUTPUT, "('#   - radial:  N = ',i6)", advance="no") P%N_GG
!      write(OUTPUT, "(' | rmax = ')", advance="no")
!      if ( P%rmax_GG == ZERO ) then; write(OUTPUT, "('  N/A')", advance="no")
!      else; write(OUTPUT, "(f5.1)", advance="no") P%rmax_GG; end if
!      write(OUTPUT, "(' | L = ',f5.2)", advance="no") P%L_GG
!      write(OUTPUT, "(' | mapping = ',a)") trim(P%type_radial_GG)
!      write(OUTPUT, "('#   - angular: lmax = ',i3,' | type = ',a)") &
!        P%lmax_GG, trim(P%type_angular_GG)
!    end if
!  end if
!  write(OUTPUT, "('#')")
!  if ( P%yes_verbose ) then
!    ! ...
!  end if
!end subroutine print_grid_param

  subroutine print_XFEL_param(P, IOUT)
    !! print XFEL parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    write (OUTPUT, "('# X-ray parameters:')")
    if (P%fluence .gt. ZERO) then
      write (OUTPUT, "('#   Number of photons = ',es10.2)") P%n_photons
      write (OUTPUT, "('#   Spatial width (micron) = ',f8.3)") &
        P%width*AU2M/1.0e-6_long
      write (OUTPUT, "('#   Fluence (ph/micron^2) = ',es10.2,"// &
        "'  [ ',f10.2,' a.u.^-2 ]')") &
        P%fluence/(AU2M/1.0e-6_long)**2, P%fluence
      write (OUTPUT, "('#   Photon energy (eV) = ',f9.2)") P%PE*AU2EV
    end if
    if (P%pulse_type .eq. 3) then         ! Gaussian
      write (OUTPUT, "('#   Gaussian envelope')")
      write (OUTPUT, "('#   Temporal width, tau (fs) [FWHM] = ',f8.1)") &
        P%FWHM*AU2SEC/1.0e-15_long
      write (OUTPUT, "('#   Pulse length = ',f5.1,' times of tau')") &
        P%pulse_multiple
    else if (P%pulse_type .eq. 6) then    ! flat-top
      write (OUTPUT, "('#   Flat-top envelope')")
      write (OUTPUT, "('#   Pulse duration (fs) = ',f8.1)") &
        (P%t_end_au - P%t_start_au)*AU2SEC/1.0e-15_long
    end if
    write (OUTPUT, "('#   Number of temporal grid points = ',i8)") P%N_time
    write (OUTPUT, "('#   dt = ',f10.5,' (a.u.) = ',f10.5,' (fs)')") &
      P%dt_au, P%dt_au*AU2SEC/1.0e-15_long
    write (OUTPUT, "('#')")
  end subroutine print_XFEL_param

  subroutine print_MC_param(P, IOUT)
    !! print Monte-Carlo parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    write (OUTPUT, "('# Monte-Carlo parameters:')")
    write (OUTPUT, "('#   max events = ',i10)") P%MC_max_events
    write (OUTPUT, "('#   convergence for population = ',es8.1)") P%MC_conv
    write (OUTPUT, "('#   print and check convergence "// &
      "at every step of ',i5,' events')") P%MC_step
    write (OUTPUT, "('#   dt ratio = ',es8.1)") P%MC_dt_ratio
    write (OUTPUT, "('#   * dt is given by 1/(sum of rates) times dt_ratio.')")

    if (P%yes_reUse_PreviousResults) then
      write (OUTPUT, "('#   Caching is turned on')")
    else
      write (OUTPUT, "('#   Caching is turned off')")
    end if

    write (OUTPUT, "('#')")
  end subroutine print_MC_param

  subroutine print_DFT_param(P, IOUT)
    !! print DFT parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    !
    ! DFT xc type and parameters
    !
    write (OUTPUT, "('# electronic structure method:')")
    if (P%yes_MCSCF) then
      write (OUTPUT, "('#   MCSCF spin-multiplicity',i4)") P%multp
    else if (P%yes_HF .and. .not. P%yes_openshell) then
      write (OUTPUT, "('#   Hartree-Fock ')")
    else if (P%yes_HF .and. P%yes_openshell .and. .not. P%yes_unrestricted) then
      write (OUTPUT, "('#   Restricted open shell Hartree-Fock ')")
    else if (P%yes_HF .and. P%yes_openshell .and. P%yes_unrestricted) then
      write (OUTPUT, "('#   Unrestricted open shell Hartree-Fock ')")
    else
      write (OUTPUT, "('#   correlation functional: ',a)") &
        trim(FUNCTIONAL_C_NAME(P%c_type))
      write (OUTPUT, "('#   exchange functional:    ',a)", advance="no") &
        trim(FUNCTIONAL_X_NAME(P%x_type))
      if (P%yes_OEP_KLI) write (OUTPUT, "(' (OEP-KLI-SIC)')", advance="no")
      if (P%yes_Latter) &
        write (OUTPUT, "(' (Latter tail correction)')", advance="no")
      write (OUTPUT, "()")
      select case (FUNCTIONAL_X_NAME(P%x_type))
       case ("Xa"); write (OUTPUT, "('#     alpha = ',f6.3)") P%alpha
       case ("LB"); write (OUTPUT, "('#     beta  = ',f6.3)") P%beta
       case ("LBa"); write (OUTPUT, "('#     alpha = ',f6.3)") P%alpha
        write (OUTPUT, "('#     beta  = ',f6.3)") P%beta
      end select
    end if
    !
    ! parameters related to accuracy control
    !
    write (OUTPUT, "('#   accuracy controls in direct Coulomb potentials: ')")
    write (OUTPUT, "('#   - EPS0 for Q_A / r            = ',es9.1)") P%EPS_00
    write (OUTPUT, "('#   - EPS1 for d_lm / d_00        = ',es9.1)") P%EPS_lm
    write (OUTPUT, "('#   - EPS2 for d_lm(r) / d_lm     = ',es9.1)") P%EPS_r
    if (P%yes_cutoff) &
      write (OUTPUT, "('#   - cutoff radius (a.u.) = ',f8.2)") P%r_cutoff
    write (OUTPUT, "('#   accuracy controls in matrix element integrals: ')")
    write (OUTPUT, "('#   - EPS(_ij) for overlap pairs  = ',es9.1)") P%EPS_overlap
    write (OUTPUT, "('#   - EPS(_ii) for overlap single = ',es9.1)") &
      P%EPS_overlap_single
    write (OUTPUT, "('#   Softcore Coulomb potential radius (a.u.) = ',f8.2)") P%R_softcore
    write (OUTPUT, "('#')")
    !
    ! parameters related to SCF
    !
    if (P%SCF_max_iter .gt. 0) then
      write (OUTPUT, "('# SCF procedure:')")
      write (OUTPUT, "('#   max iteration = ',i4)") P%SCF_max_iter
      write (OUTPUT, "('#   convergence for energy  = ',es8.1)") P%SCF_energy_conv
      write (OUTPUT, "('#   convergence for density = ',es8.1)") P%SCF_density_conv
      if (.not. P%yes_DIIS) then
        write (OUTPUT, "('#   mixing ratio = ',f6.3,' of new potentials')") &
          P%SCF_mixing
      end if
      write (OUTPUT, "('#   Subspace size in DIIS  = ',i4)") P%DIIS_nvector
      if (P%yes_DIIS) then
        write (OUTPUT, "('#   SCF method: DIIS!')")
      else
        write (OUTPUT, "('#   SCF method: SR + DIIS!')")
      end if
    else
      write (OUTPUT, "('# NO SCF procedure: max iteration = 0')")
    end if
    write (OUTPUT, "('#')")
  end subroutine print_DFT_param

  subroutine print_crystal_param(P, IOUT)
    !! print crystalline parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    write (OUTPUT, "('# Crystalline calculation:')")
    write (OUTPUT, "('#   - unitcell size  (a.u.) = ',3f11.5)") &
      P%unitcell_x, P%unitcell_y, P%unitcell_z
    write (OUTPUT, "('#   - yes_cutoff: ',l1)") P%yes_cutoff
    if (P%yes_cutoff) &
      write (OUTPUT, "('#   - cutoff radius (a.u.)  = ',f11.5)") P%r_cutoff
    !  write(OUTPUT, "('#   - r_elec_stat (a.u.)    = ',f11.5)") P%r_elec_stat
    write (OUTPUT, "('#')")
  end subroutine print_crystal_param

  subroutine print_misc_param(P, IOUT)
    !! print misc. parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    logical :: yes_output
    integer :: OUTPUT
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    yes_output = .false.
    if (P%OUTPUT .ne. STDOUT) yes_output = .true.
    write (OUTPUT, "('# Filenames:')")
    if (P%yes_unrestricted) then
      write (OUTPUT, "('#   - wfa:   ',a)") trim(P%wfaFileName)
      write (OUTPUT, "('#   - wfb:   ',a)") trim(P%wfbFileName)
    else
      write (OUTPUT, "('#   - wf:    ',a)") trim(P%wfFileName)
    end if
    write (OUTPUT, "('#   - basis: ',a)") trim(P%basisFileName)

    if (yes_output) write (OUTPUT, "('#   output: ',a)") trim(P%file_out)
    if (P%yes_load_pot) &
      write (OUTPUT, "('#   potential: ',a)") trim(P%filename)//".pot"
    write (OUTPUT, "('#   process: ',a)") trim(P%file_data)
    write (OUTPUT, "('#')")
  end subroutine print_misc_param

  subroutine dump_parameters(P, IOUT)
    !! print all parameters
    type(Param), intent(in) :: P
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT, i
    OUTPUT = P%OUTPUT; if (present(IOUT)) OUTPUT = IOUT
    !
    ! calculation type
    !
    write (OUTPUT, "(a,x,a)") "Calculation type:"
    write (OUTPUT, "(a,x,i4)") "  calc_type = ", P%calc_type
    !
    ! molecular system: read_molecule_param
    !
    write (OUTPUT, "()")
    write (OUTPUT, "(a)") "Molecular system:"
    write (OUTPUT, "(a,x,a)") "  system_name = ", trim(P%system_name)
    write (OUTPUT, "(a,i4)") "  N_nuc = ", P%N_nuc
    write (OUTPUT, "(10x,a10,4a15)") "charge", "mass", "x", "y", "z"
    do i = 1, P%N_nuc
      write (OUTPUT, "(i8,': ',i10,4f15.8)") i, P%charge_nuc(i), P%mass_nuc(i), &
        P%x_nuc(i), P%y_nuc(i), P%z_nuc(i)
    end do
    write (OUTPUT, "(a,x,i4)") "  charge = ", P%charge
    write (OUTPUT, "(a,x,i4)") "  N_elec = ", P%N_elec
    write (OUTPUT, "(a,x,i4)") "  N_a = ", P%N_a
    write (OUTPUT, "(a,x,i4)") "  N_b = ", P%N_b
    write (OUTPUT, "(a,x,i4)") "* N_orb = ", P%N_orb
    write (OUTPUT, "(1x,a,100i1)") "  N_occ(:) = ", P%N_occ(:)
    write (OUTPUT, "(a,f15.8)") "* E_nuc = ", P%E_nuc
    !
    ! basis function info: read_molecule_param
    !
    write (OUTPUT, "()")
    write (OUTPUT, "(a)") "Basis function:"
    write (OUTPUT, "(a)") "  nao_file_nuc:"
    do i = 1, P%N_nuc
      write (OUTPUT, "(i8,': ',a)") i, trim(P%nao_file_nuc(i))
    end do
    write (OUTPUT, "(a)") "  N_nao_table:"
    do i = 1, P%N_nuc
      write (OUTPUT, "(i8,': ',100i3)") i, P%N_nao_table(0:, i)
    end do
    write (OUTPUT, "(a,x,i4)") "* N_basis = ", P%N_basis
    !
    ! grid system: read_grid_param
    !
    !
    ! XFEL parameters: read_XFEL_param
    !
    !
    ! DFT and SCF parameters: read_DFT_param
    !
    !
    ! MC parameters
    ! process parameters
    ! plasma parameters
    ! misc parameters
    !
  end subroutine dump_parameters

  subroutine Aufbau_principle(N_elec, N_orb, N_occ)
    !! Aufbau principle or Madelung's rule for the non-relativistic atomic cases    
    integer, intent(in) :: N_elec
    integer, intent(out) :: N_orb
    integer, allocatable, intent(out) :: N_occ(:)
    integer :: nl, n, l, i, num
    nl = 0; i = 0; num = 0
    loop: do
      nl = nl + 1         ! nl = n + l
      do n = 1, nl
        l = nl - n
        if (l .ge. n) cycle
        i = i + 1
        num = num + (4*l + 2)
        if (num .ge. N_elec) exit loop
      end do
    end do loop
    N_orb = i             ! # of different orbitals
    allocate (N_occ(N_orb))
    nl = 0; i = 0; num = 0
    loop2: do
      nl = nl + 1
      do n = 1, nl
        l = nl - n
        if (l .ge. n) cycle
        i = i + 1
        N_occ(i) = 4*l + 2
        num = num + N_occ(i)
        if (num .ge. N_elec) then
          num = num - (4*l + 2)
          N_occ(i) = N_elec - num ! put the rest
          exit loop2
        end if
      end do
    end do loop2
  end subroutine Aufbau_principle

  function nuclear_repulsion_E(charge_nuc, x_nuc, y_nuc, z_nuc) result(E_nuc)
    !! calculate nuclear repulsion energy -- NOT including periodic condition
    integer, intent(in) :: charge_nuc(:)
    real(long), intent(in) :: x_nuc(:), y_nuc(:), z_nuc(:)
    real(long) :: E_nuc
    real(long) :: r_ij
    integer :: N_nuc, i, j
    N_nuc = size(charge_nuc)
    E_nuc = ZERO
    do i = 1, N_nuc
      do j = i + 1, N_nuc
        r_ij = sqrt((x_nuc(i) - x_nuc(j))**2 + &
          (y_nuc(i) - y_nuc(j))**2 + &
          (z_nuc(i) - z_nuc(j))**2)
        E_nuc = E_nuc + charge_nuc(i)*charge_nuc(j)/r_ij
      end do
    end do
  end function nuclear_repulsion_E

  subroutine purge_parameters(P)
    !! purge parameters
    type(Param), intent(inout) :: P

    deallocate (P%x_nuc)
    deallocate (P%y_nuc)
    deallocate (P%z_nuc)
    deallocate (P%charge_nuc)
    deallocate (P%mass_nuc)
    deallocate (P%fcharge_nuc)
    deallocate (P%nao_file_nuc)
    deallocate (P%N_nao_table)
    deallocate (P%nao_name)
    deallocate (P%nuc_name)
    deallocate (P%N_occ)
    deallocate (P%lmax_AG)
    if (allocated(P%gto_file_path)) deallocate (P%gto_file_path)
    if (allocated(P%gto_file_path2)) deallocate (P%gto_file_path2)
    if (allocated(P%sigma_nuc)) deallocate(P%sigma_nuc)
    if (allocated(P%active_space_MO)) deallocate(P%active_space_MO)
    if (allocated(P%nat_trans_orb)) deallocate(P%nat_trans_orb)
  end subroutine purge_parameters

end module read_param
