!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module scattering
  !! author: Ludger Inhester
  !!
  !! This module is for the calculation of scattering
  !!
  use constants
  use read_param
  use atom_in_mol
  use basis_func
  use wavefunction
  use molecular_grid
  use averaging

  implicit none
  private
  public :: calc_elastic_scattering, calc_Waller_Hartree_scattering
contains

  !> calculates the
  !! absolute square Fourier transform of the electron density
  !! on a cartesian reciprocal grid
  subroutine calc_elastic_scattering(P, G, F)
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(WF_data), intent(in) :: F

    complex(long), allocatable :: wave(:)
    real(long), allocatable ::elastic_scattering(:)
    complex(long) :: dens_F_transform

    integer ::i, j, l, N
    real(long):: alpha

    real(long) :: K_in(3), K_in_normalized(3), Q(3), pol_in(3), K_in_x_pol_in(3)
    integer :: N_theta, N_phi
    real(long), allocatable :: theta(:), phi(:)
    real(long), allocatable:: K_out(:, :)
    integer :: OUTPUT
    OUTPUT = P%OUTPUT

    alpha = ONE/SPEED_C_AU
    N_theta = 30
    N_phi = 30

    ! prepare the grid in reciprocal space for the momentum transfer:
    ! incomming photon momentum vector
    !
    K_in_normalized(:) = P%direction(:)/sqrt(sum(P%direction(:)**2))
    K_in(:) = K_in_normalized*alpha*P%PE
    pol_in(:) = P%pol_axis(:)/sqrt(sum(P%pol_axis(:)**2))
    K_in_x_pol_in(1) = (K_in_normalized(2)*pol_in(3) - K_in_normalized(3)*pol_in(2))
    K_in_x_pol_in(2) = (-K_in_normalized(1)*pol_in(3) + K_in_normalized(3)*pol_in(1))
    K_in_x_pol_in(3) = (K_in_normalized(1)*pol_in(2) - K_in_normalized(2)*pol_in(1))

    ! sample the outgoing photon momentum: on a sphere with radius alpha*P%PE
    N = N_theta*N_phi
    allocate (theta(N_theta))
    allocate (phi(N_phi))
    allocate (K_out(N, 3))
    theta(:) = (/(dble(j)*(PI)/(N_theta + 1), j=1, N_theta)/)
    phi(:) = (/(dble(j)*(TWO*PI)/(N_phi), j=1, N_phi)/)
    do i = 1, N_theta
      do j = 1, N_phi
        l = (i - 1)*N_phi + j
        K_out(l, :) = alpha*P%PE*( &
          sin(theta(i))*cos(phi(j))*pol_in(:) &
          + sin(theta(i))*sin(phi(j))*K_in_x_pol_in(:) &
          + cos(theta(i))*K_in_normalized(:))
      end do
    end do

    allocate (wave(G%N_grid))
    allocate (elastic_scattering(N))

    ! perform the integration
    do l = 1, N
      Q(:) = K_out(l, :) - K_in(:)
      wave(:) = exp((0, 1.)*(G%x(:)*Q(1) + G%y(:)*Q(2) + G%z(:)*Q(3)))
      dens_F_transform = sum(G%weight(:)*F%rho(:)*wave(:))
      elastic_scattering(l) = abs(dens_F_transform)
    end do

    write (OUTPUT, '("# K_in = ",3f15.10)') K_in
    write (OUTPUT, '("# Elastic scattering: cross section is given by static structure factor |F(Q)|^2*Thomsonscattering")')
    write (OUTPUT, '("# we assume photon energy out ~= photon energy in  theta,pi are spherical coordinates for the K_out direction")')
    write(OUTPUT,'("# ",4a20,3a20)') "angle(K_in,K_out)","angle(pol_in,K_out)","|q|","Thomson cs","Thomson cs*|F(q)|^2","|F(q)|^2"

    do i = 1, N_theta
      do j = 1, N_phi
        l = (i - 1)*N_phi + j
        Q(:) = K_out(l, :) - K_in(:)
        write (OUTPUT, '(3f20.10,3es20.10)') theta_angle(K_in, K_out(l, :)), phi_angle(K_in, pol_in, K_out(l, :)), sqrt(sum(Q(:)**2)), &
          Thomson(pol_in, K_out(l, :)), Thomson(pol_in, K_out(l, :))*elastic_scattering(l)**2, elastic_scattering(l)**2
      end do
      ! print additionally value for the first phi value + 2 PI to `close the circle'
      ! j=1
      ! l=(i-1)*N_phi+j
      ! Q(:)=K_out(l,:)-K_in(:)
      ! write(OUTPUT,'(3f20.10,3es20.10)') theta_angle(K_in,K_out(l,:)),phi_angle(K_in,pol_in,K_out(l,:))+TWO*PI,sqrt(sum(Q(:)**2)),&
      !   Thomson(K_in,pol_in,K_out(l,:)),Thomson(K_in,pol_in,K_out(l,:))*elastic_scattering(l),elastic_scattering(l)

      write (OUTPUT, *) ' '
    end do

    deallocate (theta)
    deallocate (phi)
    deallocate (K_out)
    deallocate (wave)
    deallocate (elastic_scattering)
  end subroutine calc_elastic_scattering

  
  subroutine calc_Waller_Hartree_scattering(P, B, G, F)
    !! calculates inelastic scattering in the Waller_Hartree approximation
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(Grid3D), intent(in) :: G

    type(WF_data), intent(in) :: F

    real(long), allocatable:: K_out(:, :)

    real(long), allocatable ::waller_hartree_scattering(:)
    real(long), allocatable ::elastic_scattering(:)
    complex(long) :: dens_F_transform

    integer ::i, j, l, N
    integer :: mo1, mo2

    real(long) :: alpha
    complex(long), allocatable :: wave(:)
    !complex(long), allocatable :: temp(:)
    !complex(long) :: Fvalue

    real(long), allocatable :: orb1(:), orb2(:)

    !real(long), allocatable :: mo_factor(:,:)

    real(long) :: K_in(3), Q(3), pol_in(3), K_in_normalized(3), K_in_x_pol_in(3)
    !real(long) :: dtheta, dphi
    integer :: N_theta, N_phi
    real(long), allocatable :: theta(:), phi(:)
    integer :: OUTPUT
    OUTPUT = P%OUTPUT

    alpha = ONE/SPEED_C_AU
    N_theta = 30
    N_phi = 30

    ! incomming photon momentum vector
    K_in_normalized(:) = P%direction(:)/sqrt(sum(P%direction(:)**2))
    K_in(:) = K_in_normalized*alpha*P%PE
    pol_in(:) = P%pol_axis(:)/sqrt(sum(P%pol_axis(:)**2))
    K_in_x_pol_in(1) = (K_in_normalized(2)*pol_in(3) - K_in_normalized(3)*pol_in(2))
    K_in_x_pol_in(2) = (-K_in_normalized(1)*pol_in(3) + K_in_normalized(3)*pol_in(1))
    K_in_x_pol_in(3) = (K_in_normalized(1)*pol_in(2) - K_in_normalized(2)*pol_in(1))

    ! sample the outgoing photon momentum: on a sphere with radius alpha*P%PE
    N = N_theta*N_phi
    allocate (theta(N_theta))
    allocate (phi(N_phi))
    allocate (K_out(N, 3))
    theta(:) = (/(dble(j - 1)*(PI)/(N_theta + 1), j=1, N_theta)/)
    phi(:) = (/(dble(j)*(TWO*PI)/(N_phi), j=1, N_phi)/)
    do i = 1, N_theta
      do j = 1, N_phi
        l = (i - 1)*N_phi + j
        K_out(l, :) = alpha*P%PE*( &
          sin(theta(i))*cos(phi(j))*pol_in(:) &
          + sin(theta(i))*sin(phi(j))*K_in_x_pol_in(:) &
          + cos(theta(i))*K_in_normalized(:))
      end do
    end do

    allocate (waller_hartree_scattering(N))
    allocate (elastic_scattering(N))
    allocate (wave(G%N_grid))

    ! perform the integration for elastic scattering
    do l = 1, N
      Q(:) = K_out(l, :) - K_in(:)
      wave(:) = exp((0, 1.)*(G%x(:)*Q(1) + G%y(:)*Q(2) + G%z(:)*Q(3)))
      dens_F_transform = sum(G%weight(:)*F%rho(:)*wave(:))
      elastic_scattering(l) = abs(dens_F_transform)**2
    end do

    ! !Variant1: naive implementation
    allocate (orb1(G%N_grid))
    allocate (orb2(G%N_grid))
    waller_hartree_scattering(:) = sum(F%N_occ)
    do mo1 = 1, B%N_basis
      if (.not. is_occupied(P, F, F%N_occ, mo1)) cycle
      do i = 1, G%N_grid
        orb1(i) = sum(F%LC(:, mo1)*B%func(i, :))
      end do
      do mo2 = mo1, B%N_basis
        if (.not. is_occupied(P, F, F%N_occ, mo2)) cycle
        do i = 1, G%N_grid
          orb2(i) = sum(F%LC(:, mo2)*B%func(i, :))
        end do
        do l = 1, N
          Q(:) = K_out(l, :) - K_in(:)
          wave(:) = exp((0, 1.)*(G%x(:)*Q(1) + G%y(:)*Q(2) + G%z(:)*Q(3)))
          if (mo1 .eq. mo2) then
            waller_hartree_scattering(l) = waller_hartree_scattering(l) - &
              TWO*stat_factor_occupied(P, F, F%N_occ, mo1) &
              *abs(sum(G%weight(:)*orb1(:)*orb2(:)*wave(:)))**2
          else
            waller_hartree_scattering(l) = waller_hartree_scattering(l) - &
              FOUR*stat_factor_pair_occupied(P, F, F%N_occ, mo1, mo2) &
              *abs(sum(G%weight(:)*orb1(:)*orb2(:)*wave(:)))**2
          end if
        end do
      end do
    end do
    deallocate (orb1)
    deallocate (orb2)


    write (OUTPUT, '("# K_in = ",3f15.10)') K_in
    write (OUTPUT, '("# pol_in = ",3f15.10)') pol_in
    write(OUTPUT,'("# Waller Hartree inelastic scattering: cross section is given by static structure factor S(Q)*Thomsonscattering")')
    write (OUTPUT, '("# we assume photon energy out ~= photon energy in ")')
    write(OUTPUT,'("# ",4a20,3a20)') "angle(K_in,K_out)","angle(pol_in,K_out)","|q|","Thomson cs","Thomson cs*S(q)","S(q)","S(q)+|f(q)|^2"

    do i = 1, N_theta
      do j = 1, N_phi
        l = (i - 1)*N_phi + j
        Q(:) = K_out(l, :) - K_in(:)
        write (OUTPUT, '(3f20.10,4es20.10)') theta_angle(K_in, K_out(l, :)), phi_angle(K_in, pol_in, K_out(l, :)), sqrt(sum(Q(:)**2)), &
          Thomson(pol_in,K_out(l,:)),Thomson(pol_in,K_out(l,:))*waller_hartree_scattering(l),waller_hartree_scattering(l),waller_hartree_scattering(l)+elastic_scattering(l)
      end do
      ! print additionally value for the first phi value + 2 PI to `close the circle'
      ! j=1
      ! l=(i-1)*N_phi+j
      ! Q(:)=K_out(l,:)-K_in(:)
      ! write(OUTPUT,'(3f20.10,3es20.10)') theta_angle(K_in,K_out(l,:)),phi_angle(K_in,pol_in,K_out(l,:))+TWO*PI,sqrt(sum(Q(:)**2)),&
      !  Thomson(K_in,pol_in,K_out(l,:)),Thomson(K_in,pol_in,K_out(l,:))*waller_hartree_scattering(l),waller_hartree_scattering(l)

      write (OUTPUT, *) ' '
    end do

    deallocate (theta)
    deallocate (phi)
    deallocate (K_out)
    deallocate (wave)
    deallocate (waller_hartree_scattering)
    deallocate (elastic_scattering)
  end subroutine calc_Waller_Hartree_scattering

  function Thomson(pol_in, K_out) result(cs)
    real(long), intent(in) :: pol_in(3), K_out(3)
    real(long) :: cs, alpha

    ! Thomson cross section is given by
    ! alpha^4 (1 - (e_in * k_out)^2 )
    ! e_in is the incident polarization axis
    ! k_out the outgoing photon momentum
    ! (all normalized vectors)

    alpha = ONE/SPEED_C_AU
    !
    !    cs=alpha**4*HALF*(ONE + costheta**2) for average polarizations
    !
    cs = alpha**4 * (ONE - (sum(pol_in(:)*K_out(:))**2/(sum(pol_in(:)**2)*sum(K_out(:)**2))))
    !   this is
    !    alpha**4 * (ONE - sin(theta)**2 * cos(phi)**2)
    !   where phi is along the pol axis
    !   averaging over phi obtains [  int cos(phi)**2 d phi / 2 pi = 1/2 ]
    !   (1 - sin(theta)**2 /2 ) =  1 - 1/2 + cos(theta)**2/2 = 1/2 (1+costheta**2)
  end function Thomson

  function theta_angle(K_in, K_out) result(theta)
    real(long), intent(in) :: K_in(3), K_out(3)
    real(long) :: costheta, theta
    costheta = sum(K_in(:)*K_out(:))/sqrt(sum(K_in(:)*K_in(:))*sum(K_out(:)*K_out(:)))

    theta = acos(costheta)
  end function theta_angle

  function phi_angle(K_in, pol_in, K_out) result(phi)
    real(long), intent(in) :: K_in(3), K_out(3), pol_in(3)
    real(long) :: phi, K_in_x_pol_in(3)

    K_in_x_pol_in(1) = (K_in(2)*pol_in(3) - K_in(3)*pol_in(2))
    K_in_x_pol_in(2) = (-K_in(1)*pol_in(3) + K_in(3)*pol_in(1))
    K_in_x_pol_in(3) = (K_in(1)*pol_in(2) - K_in(2)*pol_in(1))
    K_in_x_pol_in(:) = K_in_x_pol_in(:)/sqrt(sum(K_in_x_pol_in(:)**2))

    phi = atan2(sum(K_in_x_pol_in(:)*K_out(:)), sum(pol_in(:)*K_out(:)))
    if (phi .lt. ZERO) phi = phi + 2*PI
  end function phi_angle

end module scattering
