!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module scf
  !! author: Ludger Inhester
  !! author: Yajian Hao
  !!
  !! HFS and HF calculation for the molecular system using the LCAO
  !! written by Yajiang Hao, Ludger Inhester in 2014-2016 at DESY CFEL
  !! Mulliken atomic population analysis was written by Sang-Kil, Son in Sep 2014.
  !! This module contains the core part of DFT calculation, i.e. SCF procedure.
  !! During SCF we use DIIS to accelate the convergency.

  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use eigen_lapack
  use functional_xc
  use operations
  use neweffpot
  use rwfile
  use finitetemp
  use hartreefock
  use population
  use localization
  use packed_index
  !  use soscf
  use secondorder_scf
  use stdio
  implicit none
  private
  public :: SCF_procedure, print_orbitals, stability_analysis, construct_h0_matrix, prepare_initial_guess
contains

  subroutine SCF_procedure(P, A, G, B, F)
    !! main SCF prodedure
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F

    real(long) :: ovlMin, sqrOvl
    integer    :: minAT
    integer :: N          ! # of basis functions
    integer :: N_MO       ! # of occupied MO
    integer :: OUTPUT = STDOUT
    integer :: i, j, k, iter                                    ! counters
    logical :: yes_converged = .False.     ! check whether it's converged or not
    character(len=80) :: string
    integer :: iter_no_diis         ! iterations before DIIS has been switched on
    integer :: iter_wo_SOSCF  ! iSOSCF : first iter w SOSCF (default=-1 means off, -2 means 'never re-enter SOSCF')
    real(long) :: DIISerr, homo_lumo_gap   ! helpful quantities to compute during the SCF
    real(long) :: alpha                    ! The damping/mixing coefficient
    integer*8 :: int_time      ! timer
    real :: timer, timer_step
    real :: time_init, time_before_SCF, time_during_SCF, time_after_SCF
    real :: time_SCF(0:7)
    real(long) :: acceptable_convergenceAt_old(3, 3)
    real(long) :: dTHR, eTHR
    ! these variables are for second order SCF:
    type(indep_mo_pair), allocatable :: pairs(:)  ! orbital pairs indices
    real(long), allocatable :: theta(:) ! rotation  angle
    real(long), allocatable :: gradient(:) ! orbital rotation gradient
    real(long), allocatable :: inv_hessian(:, :) ! inverse rotation hessian
    ! for SOCSF unrestricted
    real(long), allocatable :: theta_beta(:) ! rotation  angle
    real(long), allocatable :: gradient_beta(:) ! orbital rotation gradient
    real(long), allocatable :: inv_hessian_beta(:,:) ! inverse rotation hessian
    !
    real(long) :: dE ! energy difference between steps
    integer :: actual_occ(size(F%N_occ))
    integer :: actual_multp
    real(long) :: f_occ_save(size(F%N_occ))

    !
    ! Koopmans and open shell ground state does not work
    !
    if (P%yes_openshell .and. P%yes_valgs_occ) then
      write (P%OUTPUT, *) "SCF procedure: Koopmans and open shell ground state does not work"
      stop
    end if

    if (P%yes_verbose) then
      write (P%OUTPUT, *) '# F%iSCFstate=', F%iSCFstate
    end if
    !
    ! initialize parameters:
    !
    N = P%N_basis
    N_MO = F%N_MO
    OUTPUT = P%OUTPUT
    actual_multp = F%multp
    if (P%yes_valgs_occ) then
      actual_occ = F%N_occ
      actual_multp = F%multp
      F%multp = 1
      !if (int(sum(A(:)%charge) - sum(actual_occ)) == 1) then
      if (.not. P%yes_valgs_occ_part) then
        F%N_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.True.)
      else
        F%N_occ = groundstate_occ(actual_occ, Ncore=F%Ncore, add_electrons=.False.)
      end if
      if (allocated(F%av_occ)) then
        do i = 1, size(F%av_occ)
          do j = 1, size(F%av_occ(i)%MO)
            do k = j + 1, size(F%av_occ(i)%MO)
              if (F%N_occ(F%av_occ(i)%MO(j)) .eq. 2 .and. F%N_occ(F%av_occ(i)%MO(k)) .eq. 0) then
                F%N_occ(F%av_occ(i)%MO(j)) = 1
                F%N_occ(F%av_occ(i)%MO(k)) = 1
              end if
            end do! k
          end do! j
        end do! i
      end if
    end if
    !
    ! find the highest occupied molecular orbital (HOMO)
    !
    moloop: do i = size(F%n_occ), 1, -1
      if (allocated(F%av_occ)) then
        do j = 1, size(F%av_occ)
          if (any(F%av_occ(j)%MO(:) .eq. i)) exit moloop
        end do
      end if
      if (F%n_occ(i) .ne. 0) exit moloop
    end do moloop
    F%homo_idx = i
    ! set the fractional occupations
    if (.not. F%f_occ_given) then
      if(P%yes_unrestricted)then
        where(F%N_occ(:) .ge. 1)
          F%f_occ_a(:) = ONE
        elsewhere
          F%f_occ_a(:) = ZERO
        end where
        where(F%N_occ(:) .eq. 2)
          F%f_occ_b(:) = ONE
        elsewhere
          F%f_occ_b(:) = ZERO
        end where
        F%F_occ(:) = F%F_occ_a(:) + F%f_occ_b(:)  
      else      
        F%f_occ = ONE * F%n_occ
      end if
    end if
    ! average over occupations if av_occ is given
    if (allocated(F%av_occ)) then
      do j = 1, size(F%av_occ)
        F%F_occ(F%av_occ(j)%MO(:)) = ONE/size(F%av_occ(j)%MO)*(sum(F%f_occ(F%av_occ(j)%MO)))
      end do
    end if
    if ((P%yes_verbose)) then
      if(P%yes_unrestricted) then
        write(OUTPUT,"('# f_occ_a=',10f5.2)") F%f_occ_a
        write(OUTPUT,"('# f_occ_b=',10f5.2)") F%f_occ_b
      else
        write (OUTPUT, "('# f_occ=',10f5.2)") F%f_occ
      end if
      if (allocated(F%frozen_MO)) &
        write (OUTPUT, "('# freeze=',10i5)") F%frozen_MO
    end if
    f_occ_save = F%f_occ
    !
    ! check CPU time
    !
    call check_time(int_time, timer)            ! ============================
    time_init = timer     ! check time even before entering SCF_procedure()
    time_SCF(0:) = ZERO
    call check_time(int_time, timer_step)       ! ============================
    time_SCF(0) = time_SCF(0) - timer_step
    ! alpha is the mixing / damping parameter: start with the given P%SCF_mixing
    alpha = P%SCF_mixing
    ! Initialize LevelShift vector
    if (P%yes_LevelShift) then
      do k = 1, size(F%LC, 2) ! sum over orbitals
        ! if orbital k is valence,i.e. not in the occupation string -> shift it up by LShift
        if (k .gt. F%homo_idx) then
          F%LevelShift(k) = P%lshift
        else
          F%LevelShift(k) = 0
        end if
      end do
    end if
    !
    ! v_ext: nuclear-electron interaction
    !
    F%v_ext(:) = potential_ext(G, A)
    !  call print_atomic_quantity( A, G, F%v_ext, P%OUTPUT )
    call construct_h0_matrix(G, B, F)
    call prepare_initial_guess(P, A, G, B, F)
    if (P%yes_sym) then
      do i = 1, B%nirrep
        write (P%OUTPUT, '("# irrep ",i4," n_bf ",i4)') i, count(B%irrep_sym_bfr .eq. i)
      end do
      write (P%OUTPUT, '("# initial irrep-order of MOs: ")')
      write (P%OUTPUT, '(i1)', advance='no') F%irrep_MO(1)
      do i = 2, size(F%irrep_MO)
        write (P%OUTPUT, '(",",i1)', advance='no') F%irrep_MO(i)
      end do
      write (P%OUTPUT, '()', advance='yes')
    end if

    if (P%yes_soscf) then
      if (P%yes_sym) then
        call prepare_independent_pairs(P, F, pairs, F%irrep_MO)
      else
        call prepare_independent_pairs(P, F, pairs)
      end if
      allocate (theta(size(pairs, 1)))
      allocate (gradient(size(pairs, 1)))
      allocate (inv_hessian(size(pairs, 1), size(pairs, 1)))
      inv_hessian(:, :) = ZERO
      if(P%yes_unrestricted)then
        allocate(theta_beta(size(pairs,1)))
        allocate(gradient_beta(size(pairs,1)))
        allocate(inv_hessian_beta(size(pairs,1),size(pairs,1)))
        inv_hessian_beta(:,:)=ZERO
      end if
    end if
    ! calculate the projector into the orbitals that are not frozen
    if (allocated(F%frozen_MO)) then
      F%P_not_frozen(:, :) = ZERO
      do i = 1, N
        do j = 1, N
          do k = 1, size(F%LC, 2)
            if (ANY(F%frozen_MO .eq. k)) cycle
            F%P_not_frozen(i, j) = F%P_not_frozen(i, j) + F%LC(i, k)*sum(B%S(j, :)*F%LC(:, k))
          end do
        end do
      end do
    end if
    ! Initialize variables for switching the methods during SCF
    iter_no_diis = 0
    iter_wo_SOSCF = 0
    homo_lumo_gap = ZERO
    F%switch_diis_on = .false.
    F%switch_soscf_on = P%yes_soscf
    F%switch_rstrct_on = .false.
    F%switch_levelshift_off = .false.
    F%E = ZERO
    F%d_rho = ZERO
    F%E_nuc = nuclear_repulsion_energy(A, P)
    if (G%yes_periodic) F%E_nuc = nuclear_repulsion_energy_crystal(G, A)
    if (P%yes_verbose) then
      write (OUTPUT, "('# SCFITER: ',a4,2a20,2a10)") &
        "iter", "E(0)_elec", "E(1)_elec", "E_diff", "D_diff"
    end if
    ! If an initial guess was provided print out some useful information
    if (P%yes_ReadIni .or. F%iSCFstate .gt. 0) then
      if (P%yes_unrestricted) then
        F%fidelity = fidelity(F%LC_a, F%STrans, F%LC_Ini)
        F%fidelityb = fidelity(F%LC_b, F%STrans, F%LC_Ini_b)
      else
        F%fidelity = fidelity(F%LC, F%STrans, F%LC_Ini)
      end if
      if (P%yes_verbose .and. .not. P%yes_unrestricted) then
        write (OUTPUT, "('# GUESSINFO: Prod.Overlap with initial guess (start): ',2x,f15.4)") &
        prodOverlap(P, F, F%STrans, F%LC_Ini)
        write (OUTPUT, "('# GUESSINFO: Fidelity: ',2x,10f7.4)") F%fidelity(:)
      end if
      if (P%yes_verbose .and. P%yes_unrestricted) then
        write (OUTPUT, "('# GUESSINFO: Prod.Overlap with initial guess (start): ',2x,f15.4)") &
        prodOverlapUnrestricted(P, F, F%STrans, F%LC_Ini, F%LC_ini_b)
        write (OUTPUT, "('# GUESSINFO: FidelityA: ',2x,10f7.4)") F%fidelity(:)
        write (OUTPUT, "('# GUESSINFO: FidelityB: ',2x,10f7.4)") F%fidelityb(:)
      end if
      ! density difference with the input guess
      if (P%yes_verbose) &
        write (OUTPUT, "('# GUESSINFO: Sqrt( Int (Delta rho ^2) ) : ',2x,f15.4)") &
        sqrt(integral(G, (F%rho(:) - F%rho_ini(:))**2))
      ! fidelity
    end if

    call check_time(int_time, timer_step)     ! ==========================
    time_SCF(0) = time_SCF(0) + timer_step      ! accumulate time for step 0
    time_before_SCF = timer_step
    ! ===========================================================================
    ! SCF (self-consistence field) procedure
    !   0. prepare the inital guess: coefficients & density matrix
    !   1. calculate the new density
    !   2. calculate the effective potential
    !   3. assemble the Hamiltonian matrix and do damping/DIIS
    !   4. Solve Eigenvalue Problem
    !   5. calculate energy; check the SCF convergence; if it's converged, exit SCF
    !   5a. (verbose) print some quantities
    !   6. Adjust parameters / switch methods
    !   7. repeat with step 1.
    ! ===========================================================================
    ! ===========================================================================
    ! NOTE from Sang-Kil (as of Jan. 5, 2017)
    !   Replace "write(*" and "write(6" with "write(STDERR"
    !   Please think what kind of information you are printing:
    !   - is it useful for all users? --> standard output file (P%OUTPUT)
    !   - or only useful for you?     --> STDERR or print it with verbose option
    !   Currently there have been no careful thoughts on printing options.
    !   Therefore, all unthoughtful printings (6 or *) are redirected to STDERR.
    !   In P, P%OUTPUT2 is defined to redirect all xmolecule_mc outputs,
    !         but it's not used here in xmolecule.
    !
    if (allocated(P%SCF_acceptable_convergence)) then
      if (.not. allocated(F%acceptable_convergenceAt)) then; allocate (F%acceptable_convergenceAt(3, 3))
      else
        if (allocated(F%apply_looser_threshold)) then
          acceptable_convergenceAt_old(:, :) = F%acceptable_convergenceAt(:, :)
          do i = 3, 1, -1
            do j = 3, 1, -1
              if (F%apply_looser_threshold(j, i)) then
                write (STDERR, '(a,i4)') "#SCF:apply_lower at:", F%acceptable_convergenceAt(j, i)
                exit
              end if
            end do
          end do
        end if
      end if
      F%acceptable_convergenceAt(:, :) = -1 ! invalid
    end if
    !
    SCF: do iter = 1, P%SCF_max_iter
      ! -------------------------------------------------------------------------
      ! 1. Compute the new density
      ! -------------------------------------------------------------------------
      call check_time(int_time, timer_step)       ! ============================
      time_SCF(1) = time_SCF(1) - timer_step

      if ((P%yes_thermal .and. iter .gt. 1) .or. (P%yes_thermal .and. F%iSCFstate .gt. 0)) then
        call fermi_occupation(P, F)
      end if
      F%D_old2 = F%D_old
      F%D_old = F%D
      if(P%yes_unrestricted)then
        call density_matrix(F%LC_a, F%f_occ_a, F%Da)
        call density_matrix(F%LC_b, F%f_occ_b, F%Db)
        F%D = F%Da + F%Db
      else
        call density_matrix(F%LC, F%f_occ, F%D)
      end if
      if (.not. P%yes_HF) then
         if(P%yes_unrestricted)then
            stop "HFS + unrestricted not implemented, yet"
         end if
        F%rho_old(:) = F%rho(:)
        F%rho(:) = density(G, B, F%D)
      end if
      call check_time(int_time, timer_step)     ! ==========================
      time_SCF(1) = time_SCF(1) + timer_step      ! accumulate time for step 0
      ! -------------------------------------------------------------------------
      ! 2. Calculate new potential
      ! -------------------------------------------------------------------------
      time_SCF(2) = time_SCF(2) - timer_step
      if (.not. P%yes_HF) then
        F%v_old(:) = F%v_eff(:)
        if (len(trim(P%gto_file_path)) .gt. 0) then
          ! if we use gaussian orbitals: skip calculating v_J
          F%v_xc(:) = potential_xc(P, F%rho)
          F%v_eff(:) = F%v_ext(:) + F%v_xc(:)
        else
          F%v_J(:) = potential_J(P, G, A, F%rho, F%Q_A, F%dipole_A)
          F%v_xc(:) = potential_xc(P, F%rho)
          F%v_eff(:) = F%v_ext(:) + F%v_J(:) + F%v_xc(:)
        end if
      end if
      call check_time(int_time, timer_step)     ! ==========================
      time_SCF(2) = time_SCF(2) + timer_step      ! accumulate time for step 2
      ! -------------------------------------------------------------------------
      ! 3. Assemble the Hamiltonian Matrix and do damping/ DIIS
      ! -------------------------------------------------------------------------
      time_SCF(3) = time_SCF(3) - timer_step      ! accumulate time for step 3
      F%H_old = F%H
      if (P%yes_HF) then
        if (.not. P%yes_openshell .and. .not. P%yes_unrestricted) then
          call construct_Fock_Matrix(P, B, F)
        else
          if (P%yes_unrestricted) then
            call construct_UHF_Fock_Matrix(P, B, F, F%Da, F%Db, F%Focka, F%Fockb)
            F%H = F%focka
            F%LC = F%LC_a
            F%D = F%Da + F%Db
          else
            call construct_ROHF_Fock_Matrix(P, B, F, variant=1)
          end if
        end if
      else
        call construct_Hamiltonian(P, G, B, F)
      end if
      if(.not. P%yes_unrestricted)then
        if (allocated(F%frozen_MO)) then
          DIISerr = maxval(abs( &
            B%XXts.tm.F%P_not_frozen.tm. (((F%H(:, :) .m.F%D(:, :)) .m.B%S(:, :)) &
            .sub. ((B%S(:, :) .m.F%D(:, :)) .m.F%H(:, :))) .m.F%P_not_frozen.m.B%XXts))
        else
          DIISerr = maxval(abs( &
            B%XXts.tm. ((F%H(:, :) .m.F%D(:, :) .m.B%S(:, :)) &
            .sub. (B%S(:, :) .m.F%D(:, :) .m.F%H(:, :))).m.B%XXts))
        end if
      else
        if(allocated(F%frozen_MO))then
          DIISerr = maxval(abs(&
               (B%XXts .tm. F%P_not_frozen .tm. ((( F%focka(:,:) .m. F%Da(:,:) ) .m. B%S(:,:)) &
               .sub. ((B%S(:,:) .m. F%Da(:,:) ) .m. F%focka(:,:) ) ) .m. F%P_not_frozen .m. B%XXts) + &
               (B%XXts .tm. F%P_not_frozen .tm. ((( F%fockb(:,:) .m. F%Db(:,:) ) .m. B%S(:,:)) &
               .sub. ((B%S(:,:) .m. F%Db(:,:) ) .m. F%fockb(:,:) ) ) .m. F%P_not_frozen .m. B%XXts) &
               ))
        else
          DIISerr = maxval(abs(& 
               (B%XXts .tm. (( F%focka(:,:) .m. F%Da(:,:) .m. B%S(:,:)) &
               .sub. (B%S(:,:) .m. F%Da(:,:) .m. F%focka(:,:) )) .m. B%XXts) + &
               (B%XXts .tm. (( F%fockb(:,:) .m. F%Db(:,:) .m. B%S(:,:)) &
               .sub. (B%S(:,:) .m. F%Db(:,:) .m. F%fockb(:,:) )) .m. B%XXts) &
               ))
        end if
      end if
      ! SOSCF step
      if (P%yes_soscf .and. F%switch_soscf_on) then
        if (.not. P%yes_unrestricted) then
          call make_soscf_step(F%H, F%LC, pairs, theta, gradient, inv_hessian)
          if (P%yes_verbose) &
             write (OUTPUT, '("# max gradient:",e15.5) ') maxval(abs(gradient(:)))
          if (P%yes_sym) then
            F%LC = B%sym_bf(:, :) .tm.F%LC
            do i = 1, size(F%LC, 2)
              do j = 1, size(F%LC, 1)
                if (F%irrep_MO(i) .ne. B%irrep_sym_bf(j)) then
                  if (abs(F%LC(j, i)) .gt. 1e-5) then
                    write (STDERR, '("# Warning unsymmetric orbital after rot ",i4,i4,es15.7)') i, j, F%LC(j, i)
                    F%LC(j, i) = ZERO
                 end if
                end if
              end do
            end do
            F%LC = B%sym_bf(:, :) .m.F%LC
          end if
          do i=1,size(F%LC,2)
            F%CE(i) = sum(F%LC(:,i) * (F%H(:,:) .m. F%LC(:,i)))
          end do
          ! if we have very small gradients < 10*P%SCF_energy_conv switch back to usual iterations
          if (sqrt(sum(gradient(:)**2)) .lt. 10*P%SCF_energy_conv) then
            if (P%yes_verbose) write (OUTPUT, '("# switch soscf off")')
            F%switch_soscf_on = .FALSE.
            iter_wo_SOSCF = 0
          end if
        else
          ! here we make soscf steps for alpha and beta part independently.
          ! this disregards any coupling between them
          ! @todo: make an implementation for true unrestricted SOSCF
          ! alpha part
          call make_soscf_step(F%Focka, F%LC_a, pairs, theta, gradient, inv_hessian)
          if(P%yes_verbose)&
            write(OUTPUT,'("# max gradient (alpha):",e15.5) ') maxval(abs(gradient(:)))
          if(P%yes_verbose)&
           write(OUTPUT,'("# max step (alpha):",e15.5) ') maxval(abs(theta(:)))
          if(P%yes_sym)then
            F%LC_a = B%sym_bf(:,:) .tm. F%LC_a
            do i = 1,size(F%LC,2)
              do j = 1,size(F%LC,1)
                if( F%irrep_MO(i) /= B%irrep_sym_bf(j) )then
                  if(abs(F%LC_a(j,i)) >1e-5  )then
                    write(STDERR,'("# Warning unsymmetric orbital after rot ",i4,i4,es15.7)') i,j,F%LC_a(j,i)
                    F%LC_a(j,i)=ZERO
                  end if
                end if
              end do! i
            end do! j
            F%LC_a= B%sym_bf(:,:) .m. F%LC_a
          end if
          do i=1,size(F%LC,2)
            F%CE_a(i)=sum(F%LC_a(:,i) * (F%focka(:,:) .m. F%LC_a(:,i)))
          end do
          ! beta part
          call make_soscf_step(F%Fockb, F%LC_b, pairs, theta_beta, gradient_beta, inv_hessian_beta)
          if(P%yes_verbose)&
            write(OUTPUT,'("# max gradient (beta):",e15.5) ') maxval(abs(gradient_beta(:)))
          if(P%yes_verbose)&
            write(OUTPUT,'("# max step (beta):",e15.5) ') maxval(abs(theta_beta(:)))
          if(P%yes_sym)then
            F%LC_b = B%sym_bf(:,:) .tm. F%LC_b
            do i = 1,size(F%LC,2)
              do j = 1,size(F%LC,1)
                if( F%irrep_MO(i) /= B%irrep_sym_bf(j) )then
                  if(abs(F%LC_b(j,i)) >1e-5  )then
                    write(STDERR,'("# Warning unsymmetric orbital after rot ",i4,i4,es15.7)') i,j,F%LC_b(j,i)
                    F%LC_b(j,i)=ZERO
                  end if
                end if
              end do
            end do
            F%LC_b = B%sym_bf(:,:) .m. F%LC_b
          end if
          do i=1,size(F%LC,2)
            F%CE_b(i)=sum(F%LC_b(:,i) * (F%fockb(:,:) .m. F%LC_b(:,i)))
          end do
          if (sqrt(sum(gradient(:)**2)) .lt. 10*P%SCF_energy_conv .and. &
              sqrt(sum(gradient_beta(:)**2)) .lt. 10*P%SCF_energy_conv) then
            if (P%yes_verbose) write (OUTPUT, '("# switch soscf off")')
              F%switch_soscf_on = .FALSE.
            iter_wo_SOSCF = 0
          end if
        end if        
      end if
      if (P%yes_DIIS .or. F%switch_diis_on) then
        if(P%yes_unrestricted)then
          call DIIS_unrestricted(P, B, F, iter - iter_no_diis)
        else
          call DIIS(P, B, F, iter - iter_no_diis)
        end if
      else if (iter .gt. 1 .and. .not. P%yes_SOSCF .and. .not. P%yes_unrestricted) then
        ! Damping only if no DIIS/ SOSCF        
        F%H = alpha*F%H + (ONE - alpha)*F%H_old
        F%D = alpha*F%D + (ONE - alpha)*F%D_old
      end if
      call check_time(int_time, timer_step)     ! ==========================
      time_SCF(3) = time_SCF(3) + timer_step      ! accumulate time for step 3

      !! if( .NOT. (P%yes_soscf .and. iter>3 .and. ( iSOSCF>0 .or. F%d_rho < P%SOSCF_start(2) )) ) then  << this is wrong : we have to check if SOSCF is actually started (see return value of soscf_driver)
      if (.not. F%switch_soscf_on) then                        ! avoid Solve_OE_Eigensystem when using SOSCF
        ! -------------------------------------------------------------------------
        ! 4. Solve Eigenvalue Problem
        ! -------------------------------------------------------------------------
        time_SCF(4) = time_SCF(4) - timer_step      ! check CPU time for step 4
        call Solve_OE_Eigensystem(P, F)
        call check_time(int_time, timer_step)
        time_SCF(4) = time_SCF(4) + timer_step
      end if ! .NOT. P%yes_soscf ...

      if (P%yes_sym .and. .not. allocated(P%irrep_MO) .and. (.NOT. P%yes_MOM .or. P%yes_RSTRCT)) then
        if(P%yes_unrestricted)then
          call reorder_irrep_by_CE(P, B, F, HALF * (F%CE_b + F%CE_a))
        else
          call reorder_irrep_by_CE(P, B, F, F%CE)
        end if
      end if

      ! -------------------------------------------------------------------------
      ! 5. Calculate energy and check the SCF convergence; if it's converged, exit SCF
      ! -------------------------------------------------------------------------
      time_SCF(5) = time_SCF(5) - timer_step      ! check time for step 5a & 5
      F%E_old2 = F%E_old
      F%E_old = F%E
      call calculate_energy(P, G, B, F)
      !
      if (P%yes_broadening .and. iter .gt. 1) then
        F%F_occ = f_occ_save
        call occ_broadening(P, F, P%T_ev*EV2AU)
      end if
      ! -------------------------------------------------------------------------
      !   5a. (verbose): print some quantities
      ! -------------------------------------------------------------------------
      ! compute homo lumo gap: homo is the last orbital specified in occ
      if (F%homo_idx .gt. 0) homo_lumo_gap = F%CE(min(F%homo_idx + 1, F%N_basis)) - F%CE(F%homo_idx)
      ! if initial guess was provided, calculate fidelity
      if (P%yes_ReadIni .or. F%iSCFstate .gt. 0) then
        if(P%yes_unrestricted)then
          F%fidelity = fidelity(F%LC_a, F%STrans, F%LC_Ini)
          F%fidelityb = fidelity(F%LC_b, F%STrans, F%LC_Ini_b)
        else
          F%fidelity = fidelity(F%LC, F%STrans, F%LC_Ini)
        end if
      end if
      if (P%yes_verbose) then
        write (OUTPUT, "('# SCFINFO: DIIS error',e9.2)") DIISerr
        if(P%yes_unrestricted)then
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: occ  A',10f8.2)") &
               pack(F%CE_a,F%N_occ(:)>0)
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: unoccA',10f8.2)") &
               pack(F%CE_a,F%N_occ(:)==0)
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: occ  B',10f8.2)") &
               pack(F%CE_b,F%N_occ(:)==2)
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: unoccB',10f8.2)") &
               pack(F%CE_b,F%N_occ(:)<2)
        else
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: occ   ',10f8.2)") &
               F%CE(1:F%homo_idx)
          write (OUTPUT, "('# SCFINFO: MO eigenvalues: unocc ',10f8.2)") &
               F%CE(F%homo_idx + 1:size(F%CE, 1))
        end if
        if (F%homo_idx .gt. 0 .and. .not. P%yes_unrestricted) &
          write (OUTPUT, "('# SCFINFO: HOMO-LUMO gap',f9.4)") &
             F%CE(min(F%homo_idx + 1, size(F%CE, 1))) - F%CE(F%homo_idx)
        if (P%yes_thermal) &
          write (OUTPUT, "('# SCFINFO: chem_pot',f9.4)") F%chem_pot
        if (P%yes_LevelShift) &
          write (OUTPUT, "('# SCFINFO: level shifts:',10f6.2)") F%LevelShift
        if (P%yes_thermal .or. P%yes_LevelShift) &
          write (OUTPUT, "('# SCFINFO:  occupations:',10f6.2)") F%f_occ
        if (P%yes_HF) then
          write (OUTPUT, "('# SCFITER: ',i4,2f20.8,2es10.2)") iter, &
            F%E_sum_orb, F%E, F%E_old - F%E, maxval(abs(F%D_old(:, :) - F%D(:, :)))
        else
          write (OUTPUT, "('# SCFITER: ',i4,2f20.8,2es10.2)") iter, &
            F%E_sum_orb, F%E, F%E_old - F%E, maxval(abs(F%rho_old(:) - F%rho(:)))
        end if

      end if

      call check_time(int_time, timer_step)     ! ==========================
      time_SCF(5) = time_SCF(5) + timer_step      ! accumulate time for step 5

      F%d_rho_old = F%d_rho
      if (P%yes_HF) then
        F%d_rho = maxval(abs(F%D_old(:, :) - F%D(:, :)))
      else
        F%d_rho = maxval(abs(F%rho_old(:) - F%rho(:)))
      end if

      dE = dabs(F%E - F%E_old)
      if (.not. F%switch_soscf_on .and. P%yes_SOSCF) &
        iter_wo_soscf = iter_wo_soscf + 1
      if (check_convergence(P, dE, F%d_rho) .and. &
        (.not. P%yes_SOSCF .or. .not. F%switch_soscf_on .and. iter_wo_soscf .gt. 2)) exit SCF
      if (allocated(P%SCF_acceptable_convergence)) then
        dTHR = P%SCF_acceptable_convergence(2)/10;
        dTHRcheck: do i = 1, 3; eTHR = P%SCF_acceptable_convergence(1)/10
          do j = 1, 3
            if (F%acceptable_convergenceAt(j, i) .gt. 0) exit dTHRcheck; ! already achieved
            !! if( allocated(F%apply_looser_threshold) )then; if( F%apply_looser_threshold(j,i) .and. iter>=acceptable_convergenceAt_old(j,i)-1 .and. iter<acceptable_convergenceAt_old(j,i)+6 )then;
            !!    write(*,'(a,i4,2f12.6,a,2es12.4,a,2i2)')"#SCF:apply_lower check convergence around expected point:",iter, dE, F%d_rho,"/",eTHR,dTHR,":",j,i
            !! endif; endif
            if (dE .lt. eTHR .and. F%d_rho .lt. dTHR) then
              F%acceptable_convergenceAt(j, i) = iter
              if (allocated(F%apply_looser_threshold)) then
                if (F%apply_looser_threshold(j, i)) then
                  write (STDERR, '(a,I4,2es14.6)') "#SCF:#!W converged with looser threshold", iter, dE, F%d_rho;
                  exit SCF
                end if
              end if
              if (P%yes_verbose) &
                write(STDERR,'(A,i4,2f12.6,a,2es12.4,a,2i2)')"#SCF:# achieved an acceptable convergence:",iter,dE,F%d_rho,"/",eTHR,dTHR,";",j,i
            end if
            eTHR = eTHR*sqrt(10.0);
          end do
          dTHR = dTHR*sqrt(10.0);
        end do dTHRcheck
      end if

      ! -------------------------------------------------------------------------
      ! 6. Adjust parameters / switch methods
      ! -------------------------------------------------------------------------
      time_SCF(6) = time_SCF(6) - timer_step      ! check CPU time for step 6

      ! Dynamical Levelshifting
      if (P%yes_LevelShiftDynamic .and. P%yes_LevelShift &
        .and. (.not. F%switch_levelshift_off)) then
        ! adjust the levelshift slightly such that homo_lumo_gap approaches P%Lshift
        do k = 1, F%N_basis ! sum over orbitals
          ! if orbital k is valence -> shift up by LShift
          if (k .gt. F%homo_idx) then
            F%LevelShift(k) = F%LevelShift(k) + 0.1*(P%LShift - homo_lumo_gap - F%LevelShift(k))
            if (F%LevelShift(k) .lt. ZERO) F%LevelShift(k) = ZERO
          else
            F%LevelShift(k) = ZERO
          end if
        end do
      end if

      if (.not. P%yes_DIIS .and. (.not. F%switch_diis_on) .and. (.not. F%switch_SOSCF_on) .and. .not. P%yes_unrestricted) then

        ! This is a very simple way to adjust the mixing/ damping parameter for convergence acceleration->
        ! We modify alpha in three cases:
        ! (1) flipping (alternating energy changes. energy and density changes do not decrease significantly. )
        ! (2) slow convergence (non-alternating energy changes. energy changes decrease but not significantly, density changes do not increase)
        ! (3) nearly converged (energy and density changes are close to the convergence criterion)
        if (iter .gt. 3 .and. (F%E - F%E_old)*(F%E_old - F%E_old2) .lt. 0 &
          .and. abs(F%E - F%E_old)/abs(F%E_old - F%E_old2) .gt. 0.5 &
          .and. F%d_rho/F%d_rho_old .gt. 0.5 .and. alpha .gt. 0.1) then
          ! flipping: we adjust the damping/mixing factor alpha
          alpha = 0.6*alpha
          if (alpha .lt. 0.1) alpha = 0.1
          if (P%yes_verbose) &
            write (OUTPUT, "('#  flipping: adjusting mixing coefficient to ',f7.3)") alpha
        end if
        if (iter .gt. 3 .and. (F%E - F%E_old)*(F%E_old - F%E_old2) .gt. 0 .and. &
          abs(F%E - F%E_old)/abs(F%E_old - F%E_old2) .gt. 0.5 .and. &
          abs(F%E - F%E_old)/abs(F%E_old - F%E_old2) .lt. ONE .and. &
          F%d_rho/F%d_rho_old .lt. ONE .and. alpha .lt. ONE) then
          alpha = 1.2*alpha
          if (alpha .gt. ONE) alpha = ONE
          if (P%yes_verbose) &
            write (OUTPUT, "('#  slow convergence rate: adjusting mixing coefficient to ',f7.3)") alpha
        end if
        if (iter .gt. 3 .and. F%d_rho .lt. 10*P%SCF_density_conv &
          .and. abs(F%E - F%E_old) .gt. 10*P%SCF_energy_conv .and. alpha .lt. ONE) then
          alpha = 1.2*alpha
          if (alpha .gt. ONE) alpha = ONE
          if (P%yes_verbose) &
            write (OUTPUT, "('#  nearly converged : adjusting mixing coefficient to ',f7.3)") alpha
        end if
      else
        ! no adjustment for DIIS
      end if

      ! Switch RSTRCT on
      if (.not. F%switch_rstrct_on .and. &
        abs(F%E_old - F%E) .lt. P%SwitchToRSTRCTCriterion) then
        ! we are close to convergence: switch RSTRCT on
        F%switch_RSTRCT_on = .true.
        if (P%yes_verbose) write (OUTPUT, *) '#  Restricting orbitals interchanges: on'
      end if

      ! if(.not. P%yes_soscf ) then  ---- it appears that SOSCF/DIIS is not mutually exclusive ----
      if (.not. F%switch_SOSCF_on) then
        ! Switch DIIS on
        if (.not. F%switch_diis_on .and. &
          ((abs(F%E_old - F%E) .lt. P%SwitchToDIISCriterion(1) .and. F%d_rho .lt. P%SwitchToDIISCriterion(2)) &
          .or. iter .gt. P%SwitchToDIISIter)) then
          ! if( iter> P%SwitchToDIISIter ) write(output,'(a,i4,2es12.4)')"#SCF:switch_converger:force switch DIIS..", iter, abs(F%E_old-F%E), F%d_rho
          ! we are close to convergence: switch DIIS on and Level shifting off
          ! level shift does not go well together with DIIS: switch Level shifting off
          if (allocated(F%LevelShift)) F%LevelShift(:) = ZERO
          F%switch_LevelShift_off = .true.
          F%switch_diis_on = .true.
          iter_no_diis = iter
          if (P%yes_verbose) write (OUTPUT, *) '#  DIIS: on'
        end if

        ! Switch DIIS off     : additional constraint "iter_no_DIIS + 15" prevents too early exit (in case DIIS is started due to the last condition "iter > P%SwitchToDIISIter")
        if (F%switch_diis_on .and. (iter>iter_no_DIIS + 15 ) .and. abs(F%E_old-F%E) > 10*P%SwitchToDIISCriterion(1) .and. (F%E_old-F%E_old2) > 10*P%SwitchToDIISCriterion(1) ) then
          ! switch DIIS off and Level shifting on
          ! level shift does not go well together with DIIS: switch Level shifting off
          if (P%yes_LevelShift) then
            do k = 1, F%N_basis ! sum over orbitals
              ! if orbital k is valence,i.e. not in the occupation string -> shift it up by LShift
              if (k .gt. F%homo_idx) then
                F%LevelShift(k) = P%lshift
              else
                F%LevelShift(k) = 0
              end if
            end do
            F%switch_LevelShift_off = .FALSE.
          end if
          F%switch_diis_on = .FALSE.
          iter_no_diis = 0
          if (P%yes_verbose) write (OUTPUT, *) '#  DIIS: off'
        end if
      end if

      ! Switch level shift off
      if (P%yes_LevelShift .and. (.not. F%switch_LevelShift_off) &
        .and. abs(F%E_old - F%E) .lt. P%SwitchLevelShiftOffCriterion) then
        ! we are close to convergence: switch LevelShift off
        F%LevelShift(:) = ZERO
        F%switch_LevelShift_off = .true.
        if (P%yes_verbose) write (OUTPUT, *) '#  Level shifting: off'
      end if

      call check_time(int_time, timer_step)     ! ==========================
      time_SCF(6) = time_SCF(6) + timer_step      ! accumulate time for step 6

    end do SCF
    ! ===========================================================================
    ! the end of the SCF procedure
    ! ===========================================================================
    call check_time(int_time, timer)
    time_during_SCF = timer - time_before_SCF
    ! if open shell with ROHF switch to the right ROHF variant (in case it is not already Roothan):
    ! Re-calculate the fock matrix and set the new orbitals and eigenvalues (without recalculating the energy)
    !
    if (P%yes_HF .and. P%yes_openshell .and. .NOT. P%yes_unrestricted) then
      call ROHF_canonicalize(P, B, F)
    end if

    !
    ! Do a final reordering according to initial guess
    !
    if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. (P%yes_chk_fid_ini .or. P%yes_MOM .or. P%yes_RSTRCT)) then
      if(P%yes_unrestricted)then
        ! beta part
        F%LC=F%LC_b
        F%CE=F%CE_b
        F%f_occ=F%f_occ_b
        call reorder_final(P,F,F%LC_Ini,F%STrans)
        F%fidelityb = F%fidelity
        F%LC_b=F%LC
        F%CE_b=F%CE
        F%f_occ_b=F%f_occ
        ! alpha part
        F%LC=F%LC_a
        F%CE=F%CE_a
        F%f_occ=F%f_occ_a
        call reorder_final(P,F,F%LC_Ini,F%STrans)
        F%LC_a=F%LC
        F%CE_a=F%CE
        F%F_occ_a=F%F_occ
      else
        call reorder_final(P,F,F%LC_Ini,F%STrans)
      end if
    end if

    !
    ! Print out SCF convergency information
    !
    yes_converged = .False.
    if (iter .lt. P%SCF_max_iter) yes_converged = .true.

    if (P%yes_verbose .and. P%print_level .gt. 1) then
      write (OUTPUT, "('# ',a)") repeat("-", 78)
      write (OUTPUT, "('# ',10x,2f20.8,a10)", advance="no") F%E_sum_orb, F%E
      if (yes_converged) then
        write (OUTPUT, "(a10)") "converged"
      else
        write (OUTPUT, "(a10)") "not conv."
      end if
      write (OUTPUT, "('#')")
    end if
    if (P%print_level .gt. 1) then
      string = ""
      if (iter .gt. 1) then
        if (yes_converged) then
          write (OUTPUT, "('# SCF converged after ',i4,' iterations')") iter
          string = ""
        else
          write (OUTPUT, "('# SCF NOT converged after ',i4,' iterations!!!')") iter
          string = "*** NOT CONVERGED ***"
        end if
        write (OUTPUT, "('#   energy convergence  = ',es10.3)") dE
        write (OUTPUT, "('#   density convergence = ',es10.3)") F%d_rho

        write (OUTPUT, "('#')")
      end if
    end if

    if(P%yes_valgs_occ .and. .not. F%f_occ_given )then
      write(P%OUTPUT, "('#   occ=',99i1)") F%N_occ
      write(P%OUTPUT, "('#   Total Energy    = ',f16.8,'  [ ',f12.2,' eV ]')") F%E_tot, F%E_tot * AU2EV
      F%multp = actual_multp
      F%Db = F%D
      F%Da = F%D
      do i=1,size(F%N_occ)
        if(actual_occ(i)-F%N_occ(i)/=0)then
          F%E=F%E+F%CE(i)*(actual_occ(i)-F%N_occ(i))

          F%Da(:,:) = F%Da(:,:) + (actual_occ(i)-F%N_occ(i)) * &
            spread(F%LC(:,i), dim=2, ncopies=size(F%LC,1)) * &
            spread(F%LC(:,i), dim=1, ncopies=size(F%LC,1))
        end if
      end do
      F%E_tot = F%E + F%E_nuc
      F%N_occ = actual_occ
      write(P%OUTPUT, "('#   occ=',99i1)") F%N_occ
      write(P%OUTPUT, "('#   Total Energy    = ',f16.8,'  [ ',f12.2,' eV ]')") F%E_tot, F%E_tot * AU2EV

    end if
    F%yes_converged = yes_converged

    if (P%print_level .gt. 1) then
      !
      ! print out calcualted data related to the periodic boundary condition
      !
      if (P%yes_periodic) call print_crystalline_data(G, F, OUTPUT)
      !
      ! take into account the long-range interaction, if r_cutoff is used
      !
      if (P%yes_cutoff) then
        F%E_long_range(:) = long_range_energies(P, G, F)
        F%E_ext = F%E_ext + F%E_long_range(1)
        F%E_J = F%E_J + F%E_long_range(2)
        F%E_nuc = F%E_nuc + F%E_long_range(3)
        F%E = F%E_sum_orb - F%E_J + (F%E_x + F%E_c) - F%E_xc_int
        F%E_Ts = F%E - F%E_J - F%E_x - F%E_c - F%E_ext       ! kinetic energy of electrons
        F%E_tot = F%E + F%E_nuc
      end if
      !
      ! print dipole moments, energetics, and orbitals
      !
      call print_dipole_moments(P, G, A, B, F, OUTPUT)
      call print_energies(P, F, string)
      if(P%yes_unrestricted)then
        F%LC = F%LC_b
        F%CE = F%CE_b
        string = "*** unrestricted beta orbitals ***"
        call print_orbitals(P,B,F,trim(string))
        F%LC = F%LC_a
        F%CE = F%CE_a
        string = "*** unrestricted alpha orbitals ***"
        call print_orbitals(P,B,F,trim(string))
      else
        call print_orbitals(P,B,F,string)
      end if    
    end if
    F%prod_overlap = ZERO

    if (P%yes_ReadIni .or. F%iSCFstate .gt. 0) then
      if (P%yes_unrestricted) then
        F%fidelity = fidelity(F%LC_a, F%STrans, F%LC_Ini)
        F%fidelityb = fidelity(F%LC_b, F%STrans, F%LC_Ini_b)
        F%prod_overlap = prodOverlapUnrestricted(P, F, F%STrans, F%LC_Ini, F%LC_ini_b)
      else
        F%fidelity = fidelity(F%LC, F%STrans, F%LC_Ini)
        F%prod_overlap = prodOverlap(P, F, F%STrans, F%LC_Ini)
      end if
      if (P%yes_verbose .and. .not. P%yes_unrestricted) then
        write (OUTPUT, "('# GUESSINFO: Prod.Overlap with initial guess (end): ',2x,f15.4)") &
        prodOverlap(P, F, F%STrans, F%LC_Ini)
        write (OUTPUT, "('# GUESSINFO: Fidelity: ',2x,10f7.4)") F%fidelity(:)
      end if
      if (P%yes_verbose .and. P%yes_unrestricted) then
        write (OUTPUT, "('# GUESSINFO: Prod.Overlap with initial guess (end): ',2x,f15.4)") &
        prodOverlapUnrestricted(P, F, F%STrans, F%LC_Ini, F%LC_ini_b)
        write (OUTPUT, "('# GUESSINFO: FidelityA: ',2x,10f7.4)") F%fidelity(:)
        write (OUTPUT, "('# GUESSINFO: FidelityB: ',2x,10f7.4)") F%fidelityb(:)
      end if
      ! density difference with the input guess
      if (P%yes_verbose) &
        write (OUTPUT, "('# GUESSINFO: Sqrt( Int (Delta rho ^2) ) : ',2x,f15.4)") &
        sqrt(integral(G, (F%rho(:) - F%rho_ini(:))**2))
      ! fidelity
        if(F%prod_overlap < 0.5 )&
        write(OUTPUT,"('# WARNING: Product overlap with initial guess is smaller than 0.5 !')")
    end if

    ! additional output for MC --------------------
    if (yes_converged) then
      F%ierr_o = 0
      F%N_iter_o = iter
      F%avgSqrOvlp_w_refMO_o = ZERO
      if (F%iSCFstate .gt. 0) then
        ovlMIN = 9.99E+03
        minAT = -1
        do i = 1, B%Nr_basis
          sqrovl = dot_product(F%LC(:, i), matmul(F%STrans(:, :), F%LC_INI(:, i)))**2
          if (sqrovl .lt. ovlMIN) then
            ovlMIN = sqrovl
            minAT = i
          end if
          F%avgSqrOvlp_w_refMO_o = F%avgSqrOvlp_w_refMO_o + sqrovl
        end do
        F%avgSqrOvlp_w_refMO_o = F%avgSqrOvlp_w_refMO_o/B%N_basis
        if (P%print_level .gt. 1) write (OUTPUT, "('#ovlp_w_refMO:',f12.6,' MIN:',f12.6,' for:',I4)") &
          F%avgSqrOvlp_w_refMO_o, ovlMIN, minAT
      end if
    else
      F%ierr_o = 1
      F%N_iter_o = iter
      F%avgSqrOvlp_w_refMO_o = -ONE
    end if

    !! --- minimal print option for MC ---
    if (P%print_level .le. 1) then
      if (.not. yes_converged) then
        if (.not. F%F_occ_given) then
          write (OUTPUT, "('# SCF NOT converged:',' after ',i4,' iterations. ','  occ=  ',9999i1)") &
            iter, F%N_occ(:)
        else
          write (OUTPUT, "('# SCF NOT converged:',' after ',i4,' iterations. ','  occ=  ',9999f5.1)") &
            iter, F%f_occ(:)

        end if
        string = "*** NOT CONVERGED ***"
      else
        if (P%print_level .eq. 1) then
          if (.not. F%F_occ_given) then
            write (OUTPUT, "('# SCF converged: E=',f20.4,' after ',i4,' iterations. ','prod_overlap=',f12.5,'  occ=  ',9999i1)") &
              F%E_tot, iter, F%prod_overlap, F%N_occ(:)
          else
            write (OUTPUT, "('# SCF converged: E=',f20.4,' after ',i4,' iterations. ','prod_overlap=',f12.5,'  occ=  ',9999f5.1)") &
              F%E_tot, iter, F%prod_overlap, F%f_occ(:)
          end if
        end if

      end if
    end if

    !
    ! print a summary
    ! HOMO and LUMO should be redefined according to occupation numbers
    !
    if (P%print_level .gt. 1) then
      if(.NOT.P%yes_unrestricted)then
        if ( string /= "" ) write(OUTPUT, "('# ',a)") trim(string)
        write(OUTPUT, "('# |SCF| ',2000f10.2)") F%CE(1:F%N_MO) * AU2EV
        if(F%homo_idx .gt. 0 )then
          write(OUTPUT, "('# |HOMO|',f10.2)") F%CE(F%homo_idx) * AU2EV
          write(OUTPUT, "('# |LUMO|',f10.2)") F%CE(min(F%homo_idx+1,B%Nr_basis))*AU2EV
        end if
      else
        string = "*** beta orbitals ***"
        write(OUTPUT, "('# ',a)") trim(string)
        write(OUTPUT, "('# |SCF| ',2000f10.2)") F%CE_b(1:F%N_MO) * AU2EV
        write(OUTPUT, "('# |LUMO|',f10.2)") F%CE_b(min(F%homo_idx+1,B%Nr_basis))*AU2EV
        string = "*** alpha orbitals ***"
        write(OUTPUT, "('# ',a)") trim(string)
        write(OUTPUT, "('# |SCF| ',2000f10.2)") F%CE_a(1:F%N_MO) * AU2EV
        write(OUTPUT, "('# |LUMO|',f10.2)") F%CE_a(min(F%homo_idx+1,B%Nr_basis))*AU2EV
      end if
      !
      ! print out CPU time analysis
      !
      call check_time(int_time, timer)    ! timer: total CPU time
      time_after_SCF = timer - time_before_SCF - time_during_SCF
      write(OUTPUT, "('#')")
      write(OUTPUT, "('# CPU time analysis:')")
      call print_t( OUTPUT, "before SCF ---------", time_before_SCF, timer )
      call print_t( OUTPUT, "  construct A,G,B,F ", time_init, timer )
      call print_t( OUTPUT, "  0. initial guess  ", time_SCF(0), timer )
      call print_t( OUTPUT, "during SCF ---------", time_during_SCF, timer, iter )
      call print_t( OUTPUT, "  1. new density    ", time_SCF(1), timer, iter )
      call print_t( OUTPUT, "  2. new potential  ", time_SCF(2), timer, iter )
      call print_t( OUTPUT, "  3. construct H    ", time_SCF(3), timer, iter )
      call print_t( OUTPUT, "  4. eigenvalue     ", time_SCF(4), timer, iter )
      call print_t( OUTPUT, "  5. print / check  ", time_SCF(5), timer, iter )
      call print_t( OUTPUT, "  6. miscellaneous  ", time_SCF(6), timer, iter )
      call print_t( OUTPUT, "after SCF ----------", time_after_SCF, timer )
      write(OUTPUT, "('#   Total CPU time -----',f12.2,' sec')") timer
      write(OUTPUT, "('#')")
    end if ! print_level > 1

    ! do localization
    if(allocated(P%localize))then
      ! only do localization within irrep
      if(P%yes_sym)then
        do i=1,B%nirrep
          if(any(F%irrep_MO(P%localize(:))==i))then
            call localize_orbitals(P, G, B, F, &
                 pack( [ (P%localize(j),j=1,size(P%localize,1)) ],&
                 F%irrep_MO(P%localize(:))==i))
          end if
        end do! i
      else
        call localize_orbitals(P, G, B, F, P%localize)
      end if
      if(allocated(P%fix_nuc_MO) .and. allocated(P%fix_nuc_n)) then
        do i = 1, size(P%fix_nuc_MO)
          if ( .not. any(P%localize(:) == P%fix_nuc_MO(i))) then
            write(OUTPUT, "('All MO in the fix_nuc set have to be in the localize set.')")
            stop
          end if
        end do
        call fix_orbitals_on_nuclei(P, B, F, P%fix_nuc_MO, P%fix_nuc_n)
      end if
      if ( P%print_level>1) then
        string="*** Localized Orbitals *** "
        call print_orbitals(P,B,F,string)       
      end if
    end if

    !
    ! save the current results for the future usage
    !
    if (P%yes_SaveIni .and. yes_converged .and. F%iSCFstate .eq. 0) then
      F%rho(:) = density(G, B, F%D)
      if (P%yes_unrestricted) then
        call save_wf(F%LC_a, filepath=P%wfaFileName)
        call save_wf(F%LC_b, filepath=P%wfbFileName)
      else
        call save_wf(F%LC)
      end if
      if(.not. P%yes_gaussian)&
        call save_basis(B)
    end if

    if (P%yes_soscf) then
      deallocate(pairs)
      deallocate(theta)
      deallocate(gradient)
      deallocate(inv_hessian)
      if(P%yes_unrestricted)then
        deallocate(theta_beta)
        deallocate(gradient_beta)
        deallocate(inv_hessian_beta)
      end if
    end if
  contains
    subroutine print_t(OUTPUT, string, t, t_tot, iter)
      integer, intent(in) :: OUTPUT
      character(len=20), intent(in) :: string
      real, intent(in) :: t, t_tot
      integer, intent(in), optional :: iter
      write (OUTPUT, "('#   ',a20,f12.2,' sec')", advance="no") string, t
      if (string(20:20) .ne. '-' ) &
        write (OUTPUT, "(' (',f5.1,'%)')", advance="no") t/t_tot*100_long
      if (present(iter)) then
        if (string(20:20) .eq. '-') then
          write (OUTPUT, "(6x,' / ',i5,' iter =')", advance="no") iter
        else
          write (OUTPUT, "(12x)", advance="no")
        end if
        write (OUTPUT, "(f10.2,' sec/iter')", advance="no") t/dble(iter)
      end if
      write (OUTPUT, "()")
    end subroutine print_t
  end subroutine SCF_procedure

!----------------------------------------------------------------------------
! check convergence with total energy and all points of density
!----------------------------------------------------------------------------
  function check_convergence(P, DeltaE, Deltarho) result(yes_conv)
    type(Param), intent(in) :: P
    real(long), intent(in) :: DeltaE, Deltarho
    logical :: yes_conv
    if (abs(DeltaE) .lt. P%SCF_energy_conv .and. abs(Deltarho) .lt. P%SCF_density_conv) then
      yes_conv = .true.
    else
      yes_conv = .false.
    end if
  end function check_convergence

  subroutine construct_Hamiltonian(P, G, B, F)
    implicit none
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long) :: teint
    integer :: i, j, k, l
    integer(8) :: ij, kl, ijkl
    integer :: i_nuc, j_nuc
    integer :: N_tr
    real(long) :: S_mu_nu_A, R_AB
    real(long) :: H_inside        ! short-range: treated with v_eff(:)
    real(long) :: H_outside       ! long-range: treated with electrostatic forces
    real(long) :: monopole, dipole, Z_eff, d0
    real(long) :: R0(3)           ! position vector without translation
    real(long) :: T(3)            ! translation vector
    real(long) :: r_c_EPS_00
    logical :: yes_dipole(G%N_nuc)        ! whether dipole interaction included
    N_tr = 1
    if (G%yes_periodic) N_tr = G%N_tr
    if (P%EPS_00 .eq. ZERO) r_c_EPS_00 = HUGE(r_c_EPS_00)
    do i_nuc = 1, G%N_nuc
      if (all(abs(F%dipole_A(1:3, i_nuc)) .lt. EPS)) then
        yes_dipole(i_nuc) = .false.
      else
        yes_dipole(i_nuc) = .true.
      end if
    end do
    !
    ! all prescreening are considered in braket() in basis_func.f90
    ! ==> as of May 7, 2019 (Sang-Kil)
    ! braket() uses B%yes_overlap to prescreen integerals.
    ! however, S_ij ~ exp(-r) but H_ij ~ 1/r.
    ! slowly-changing H_ij should not be screened as the same way as S_ij.
    ! instead of using braket(), here we explicit calculate H_ij
    ! only if B%yes_elec_stat is okay, then we approximate H_ij as
    ! electrostatic Coulomb force with point charges [ F%Q_A(:) ]
    !
    loop_mu: do i = 1, B%N_basis
      loop_nu: do j = i, B%N_basis
        !
        ! short-range region (r <= r_cutoff if G%yes_cutoff):
        ! full calculation with the calculated effective potential
        ! - effective potential evaluated up to r <= r_cutoff (if G%yes_cutoff)
        ! - all truncations included using B%yes_overlap
        !
        H_inside = braket(G, B, i, j, F%v_eff(:))
        F%H(i, j) = B%HKin(i, j) + H_inside
        !
        ! long-range region (r > r_cutoff):
        ! evaulate additional terms with electrostatic interaction
        ! with monopole-monopole interaction (point charge or monopole contrib.)
        ! and monopole-dipole interaction (dipole contribution)
        !
        d0 = ZERO
        ij = int(index_packed(i, j))                   ! i --> mu and j --> nu
        if (G%yes_cutoff .and. B%yes_overlap(0, ij)) then
          H_outside = ZERO
          fixed_A: do i_nuc = 1, G%N_nuc            ! it's NOT summed over all A's
            if (B%yes_overlap(i_nuc, ij) .eqv. .false.) cycle fixed_A
            S_mu_nu_A = B%overlap(i_nuc, ij)         ! overlap of mu and nu on A
            loop_B: do j_nuc = 1, G%N_nuc              ! scan over all B's
              Z_eff = F%Q_A(j_nuc) - G%Z_A(j_nuc)     ! effective charge at j_nuc
              if (P%EPS_00 .gt. ZERO ) &                ! screening consistently
                r_c_EPS_00 = F%Q_A(j_nuc)/P%EPS_00  !   in potential_J_atom
              !
              ! d0 calculation: dipole even without translation
              !   d = vec_d . vec_R
              !     = vec_d . ( vec_R0 + mx vec_Tx + my vec_Ty + mz vec_Tz )
              !     = vec_d . vec_R0 + mx d_x L_x + my d_y L_y + mz d_z L_z
              !       -------------- = d0
              !   where vec_d = ( d_x, d_y, d_z )
              !         vec_Tx=(L_x,0,0); vec_Ty=(0,L_y,0); vec_Tz=(0,0,L_z)
              ! REMEMBER: G%RP needs to consider sign and translation
              !
              if (yes_dipole(j_nuc)) then
                R0(:) = G%RP(1:3, index_packed(i_nuc, j_nuc))
                if (j_nuc .lt. i_nuc) R0(:) = -R0(:)     ! vector direction corrected
                if (G%yes_periodic) &                 ! needs to be translated
                  call translation_in_crystal(G, R0(1), R0(2), R0(3))
                d0 = sum(R0(1:3)*F%dipole_A(1:3, j_nuc))
              end if
              loop_tr: do k = 1, N_tr                   ! N_tr = 1 if no periodic
                !
                ! ACHTUNG:
                !   scanning order for i_nuc, j_nuc, k is important
                !   bond_distance() has been modified to accommodate k->j_nuc->i_nuc
                !   on June 19, 2019
                !   the bond_distance is symmetric, thus
                !     bond_distance(G,i_nuc,j_nuc,k) == bond_distance(G,j_nuc,i_nuc,k)
                !   however, the latter is about 2 times slower than the former!!!
                !
                R_AB = bond_distance(G, i_nuc, j_nuc, k)  ! <-- good scanning order
                if (R_AB .le. G%r_cutoff) cycle loop_tr ! skip the short-range
                !
                ! consider monopole contribution only
                !
                monopole = Z_eff/R_AB
                H_outside = H_outside + monopole*S_mu_nu_A
                !
                ! consider dipole contribution
                !   n_vec := a normalized vector from j_nuc pointing i_nuc
                !   n_vec(:) = obtain_nuclear_position( G, i_nuc, j_nuc, k ) / R_AB
                !   dipole = sum( n_vec(:) * F%dipole_A(:,j_nuc) ) / R_AB**2
                ! however, here is a simpler form (see the above):
                !   d = d0 + mx d_x L_x + my d_y L_y + mz d_z L_z
                !   dipole = d / R_AB**3
                !
                if (yes_dipole(j_nuc)) then
                  if (R_AB .gt. r_c_EPS_00) exit loop_tr ! skip R_AB>r_c given by EPS_00
                  T = (/G%ix_tr(k, j_nuc, i_nuc)*G%unitcell_x, &
                    G%iy_tr(k, j_nuc, i_nuc)*G%unitcell_y, &
                    G%iz_tr(k, j_nuc, i_nuc)*G%unitcell_z/)
                  dipole = (d0 + sum(T(1:3)*F%dipole_A(1:3, j_nuc)))/R_AB**3
                  H_outside = H_outside + dipole*S_mu_nu_A
                end if
              end do loop_tr
            end do loop_B
          end do fixed_A
          F%H(i, j) = F%H(i, j) + H_outside
        end if
        if (i .ne. j) F%H(j, i) = F%H(i, j)   ! make it symmetric
      end do loop_nu
    end do loop_mu
    !
    ! below is for the Gaussian basis set
    !
    !  when we use gto basis, add the Coulomb potential here
    ! if(allocated(B%teint))then
    !   do i = 1, B%N_basis
    !     do j = i, B%N_basis
    !       ij=index_packed(i,j)
    !       do k = 1, B%N_basis
    !         l=k
    !         kl=index_packed(k,l)
    !         ijkl=index_packed(ij,kl)
    !         H(i,j)=H(i,j)+B%teint(ijkl)*D(k,l)
    !         do l = k+1, B%N_basis
    !           kl=index_packed(k,l)
    !           ijkl=index_packed(ij,kl)
    !           H(i,j)=H(i,j)+B%teint(ijkl)*TWO*D(k,l)
    !         end do! l
    !       end do! k
    !       if ( i /= j ) H(j,i) = H(i,j)
    !     end do
    !   end do
    ! end if

    ! This is a faster variant for the above:
    ! loop over teint entries instead over H indices
    if(P%yes_gaussian .and. associated(B%teint) )then ! i<j, k<l, and ij<kl
      do i = 1, B%N_basis 
        do j = i, B%N_basis
          ij = index_packed(i, j)
          do k = 1, B%N_basis
            do l = k, B%N_basis
              kl = index_packed(k, l)
              if (ij .lt. kl) cycle

              ijkl = index_packed(ij, kl)
              teint = B%teint(ijkl)

              ! 4 indices equal
              if (i .eq. j .and. k .eq. l .and. i .eq. k) then! (ii|ii)
                F%H(i, i) = F%H(i, i) + F%D(i, i)*teint
                ! 3 indices equal
              else if (i .eq. j .and. j .eq. k) then ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
                F%H(i, i) = F%H(i, i) + TWO*F%D(i, l)*teint
                if (i .lt. l) then  ! i is always smaller than l because ij < kl and i<j and k<l
                  F%H(i, l) = F%H(i, l) + F%D(i, i)*teint
                else
                  F%H(l, i) = F%H(l, i) + F%D(i, i)*teint
                end if
              else if (i .eq. j .and. j .eq. l) then ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
                F%H(i, i) = F%H(i, i) + TWO*F%D(k, i)*teint
                if (i .lt. k) then
                  F%H(i, k) = F%H(i, k) + F%D(i, i)*teint
                else
                  F%H(k, i) = F%H(k, i) + F%D(i, i)*teint
                end if
              else if (i .eq. k .and. k .eq. l) then ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
                F%H(i, j) = F%H(i, j) + F%D(i, i)*teint
                F%H(i, i) = F%H(i, i) + TWO*F%D(i, j)*teint
              else if (j .eq. k .and. k .eq. l) then ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
                F%H(i, j) = F%H(i, j) + F%D(j, j)*teint
                F%H(j, j) = F%H(j, j) + TWO*F%D(i, j)*teint
                ! two indices pairwise equal
              else if (i .eq. j .and. k .eq. l) then ! (ii|kk) (kk|ii)
                F%H(i, i) = F%H(i, i) + F%D(k, k)*teint
                F%H(k, k) = F%H(k, k) + F%D(i, i)*teint
              else if (i .eq. k .and. j .eq. l) then ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
                F%H(i, j) = F%H(i, j) + TWO*F%D(i, j)*teint
                ! two indices equal
              else if (i .eq. j) then ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
                F%H(i, i) = F%H(i, i) + TWO*F%D(k, l)*teint
                F%H(k, l) = F%H(k, l) + F%D(i, i)*teint
              else if (k .eq. l) then ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
                F%H(i, j) = F%H(i, j) + F%D(k, k)*teint
                F%H(k, k) = F%H(k, k) + TWO*F%D(i, j)*teint
              else if (i .eq. k .and. j .ne. l) then ! (ij|il) (ji|il) (ij|li) (ji|li) (il|ij) (il|ji) (li|ij) (li|ji)
                F%H(i, j) = F%H(i, j) + TWO*F%D(i, l)*teint
                if (i .lt. l) then
                  F%H(i, l) = F%H(i, l) + TWO*F%D(i, j)*teint
                else
                  F%H(l, i) = F%H(l, i) + TWO*F%D(i, j)*teint
                end if
              else if (i .ne. k .and. j .eq. l) then ! (ij|kj) (ji|kj) (ij|jk) (ji|jk) (kj|ij) (kj|ji) (jk|ij) (jk|ji)
                F%H(i, j) = F%H(i, j) + TWO*F%D(k, j)*teint
                if (k .lt. j) then
                  F%H(k, j) = F%H(k, j) + TWO*F%D(i, j)*teint
                else
                  F%H(j, k) = F%H(j, k) + TWO*F%D(i, j)*teint
                end if
                ! no indices equal
              else ! (ij|kl) (ji|kl) (ij|lk) (ji|lk) (kl|ij) (kl|ji) (lk|ij) (lk|ji)
                F%H(i, j) = F%H(i, j) + TWO*F%D(k, l)*teint
                F%H(k, l) = F%H(k, l) + TWO*F%D(i, j)*teint
              end if
            end do! l
          end do! k
        end do! j
      end do! i
      do i = 1, B%N_basis
        do j = i + 1, B%N_basis
          F%H(j, i) = F%H(i, j)
        end do
      end do
    else if(.not. associated(B%teint) .and. P%yes_gaussian)then
      call add_electrostatic_interaction(B%env, B%atm, B%bas, B%first_bf_of_shell, F%D, F%H)
    end if! if(.not. allocated(B%teint) .and. len(trim(P%gto_file_path) )> 0)then
  end subroutine construct_Hamiltonian

  subroutine calculate_energy(P, G, B, F)
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer :: N_MO
    real(long) :: entropy
    N_MO = F%N_MO
    if(P%yes_unrestricted)then
      F%E_sum_orb = matTraceProduct(F%Da(:,:), F%Focka(:,:)) + matTraceProduct(F%Db(:,:), F%Fockb(:,:))
    else
      F%E_sum_orb = MatTraceProduct(F%D(:,:), F%H(:,:))
    end if
    F%E_x = ZERO
    F%E_c = ZERO
    F%E_xc_int = ZERO
    F%E_J = ZERO

    if (.not. P%yes_HF) then
      F%E_ext = integral(G, F%rho(:)*F%v_ext(:))
      F%E_J = HALF*integral(G, F%rho(:)*F%v_J(:))
      !E_xc = functional_xc_test( G, F%rho )
      F%E_x = functional_x(P, G, F%rho)
      F%E_c = functional_c(P, G, F%rho)
      F%E_xc_int = integral(G, F%rho(:)*F%v_xc(:))
      !
      ! The expression with MatTrace is advantageous because
      ! 1) it doesn't need any nonlocal potential like v_J; it has only E_xc_int,
      ! 2) it is the same expression with the HF case.
      ! [NOTE: what about "H .m. D"?  Is this matrix multiplication expensive?]
      ! Anyhow, it uses H0... For periodic calculations, H may be constructed
      ! in the short- and long-range interactions differently. In this case,
      ! v_ext in H0 becomes different from v_ext in H.
      ! To avoid this problem, let's use the simple expression with F%E_sum_orb,
      ! which is computed anyway...
      ! For Gaussians we avoid calculating some things on the grid
      ! It's cheaper to directly calculate quantities like H0
      ! therefore for Gaussians we do it the 'Trace' way
      ! Since gaussians and periodic boundary conditions do not work together at the moment
      ! this should be fine
      ! It turns out that the two equations
      ! (Trace and sum over orbitals) are not equivalent during orbital optimization
      ! for some reason ( because energy is connected to decisions on changes on the damping factor
      ! switching on/off DIIS) the Trace variant shows better convergence behavior
      if(P%yes_gaussian)then
        F%E_Ts = matTraceProduct(B%Hkin, F%D)
        F%E=HALF * (matTraceProduct(F%H0, F%D) + matTraceProduct(F%H(:,:), F%D(:,:) ) ) + - HALF * F%E_xc_int + ( F%E_x + F%E_c )
      else
        F%E_Ts = F%E - F%E_J - F%E_x - F%E_c - F%E_ext
        if (P%yes_periodic) then
          F%E = F%E_sum_orb - F%E_J + (F%E_x + F%E_c) - F%E_xc_int
        else
          F%E=HALF * (matTraceProduct(F%H0, F%D) + matTraceProduct(F%H(:,:), F%D(:,:) ) ) + &
               - HALF * F%E_xc_int + ( F%E_x + F%E_c ) 
        end if

      end if
    else ! yes HF
      if(.not. P%yes_openshell)then
        F%E = HALF * matTraceProduct(F%H0, F%D) + HALF * matTraceProduct(F%D, F%H)
        if(P%yes_gaussian)then
          F%E_ext = matTraceProduct(B%vext, F%D)
          F%E_Ts = matTraceProduct(B%hkin, F%D)
        else
          F%E_J = HALF*integral(G, F%rho(:)*F%v_J(:))
          F%E_ext = integral(G, F%rho(:)*F%v_ext(:))
          F%E_Ts = F%E - F%E_J - F%E_ext       ! kinetic energy of electrons
        end if
      else
        if(.not. P%yes_unrestricted)then
          ! ROHF
          call density_matrix_alpha( P, F, F%Da)
          call density_matrix_beta( P, F, F%Db)
        else
          ! UHF
          call density_matrix( F%LC_a, F%f_occ_a, F%Da)
          call density_matrix( F%LC_b, F%f_occ_b, F%Db)
        end if
        F%E = HALF*matTraceProduct(F%H0, (F%Da+F%Db)) + HALF*MatTraceProduct(F%FockA,F%Da) + HALF*MatTraceProduct(F%FockB,F%Db)
        if(P%yes_gaussian)then
          F%E_ext = matTraceProduct(B%vext, F%Da + F%Db)
          F%E_Ts = matTraceProduct(B%hkin, F%Da + F%Db)
        else
          F%E_J = HALF * integral(G, F%rho(:)*F%v_J(:))
          F%E_ext = integral( G, F%rho(:)*F%v_ext(:))
          F%E_Ts = F%E - F%E_J - F%E_ext       ! kinetic energy of electrons
        end if
      end if
    end if! yes_HF
    F%E_tot = F%E + F%E_nuc
    if (P%yes_thermal) then
      entropy = calculate_entropy(F%F_occ)
      F%E_tot = F%E_tot - P%T_eV/AU2EV*entropy ! free energy = E - T * S
    end if
  end subroutine calculate_energy

  subroutine print_orbitals(P, B, F, string)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    character(*), intent(in) :: string
    integer :: is, ie, i, j, k
    integer, parameter :: MAX_COLUMNS = 6
    integer :: N_MO, N_MO_print
    N_MO = F%homo_idx
    if (P%NMO_print .gt. 0) then
      N_MO_print = min(size(F%LC, 2), P%NMO_print)
    else
      N_MO_print = min(size(F%LC, 2), N_MO + 10)
    end if
!  N_MO_print=F%N_basis
    write (P%OUTPUT, "('# Printing ',i4,'out of ',i4,'orbitals')") N_MO_print, size(F%LC, 2)
    write (P%OUTPUT, "('# Molecular orbital energies: ',a)") trim(string)
    write (P%OUTPUT, "('#   ',a3,a6,a8,a16,2a12)", advance="no") &
      "idx", "n_occ", "f_occ", "E(a.u.)", "E(eV)"
    if (P%yes_sym) write (P%OUTPUT, "(a12)", advance="no") "irrep"
    write (P%OUTPUT, "()")
    do j = 1, N_MO_print
      write (P%OUTPUT, "('# ',i5,i6,f8.5,f16.8,f12.2,f12.3)", advance="no") j, &
        F%N_occ(j), F%f_occ(j), F%CE(j), F%CE(j)*AU2EV
      if (P%yes_sym) write (P%OUTPUT, "(i12)", advance="no") F%irrep_MO(j)
      write (P%OUTPUT, "()")
    end do
    write (P%OUTPUT, "('#')")
    !
    ! print MO coefficients
    !
    write (P%OUTPUT, "('# Molecular orbital coefficients: ',a)") trim(string)
    do i = 1, N_MO_print, MAX_COLUMNS
      is = i
      ie = i + MAX_COLUMNS - 1; if (ie .gt. N_MO_print) ie = N_MO_print
      write (P%OUTPUT, "('#   ',a10,2x,100i10)") "MO index", (/(k, k=is, ie)/)
      !write(P%OUTPUT, "('#',a9,100i10)") "dgnrt", dgnrt(is:ie,1)
      write (P%OUTPUT, "('#   ',a10,2x,100f10.3)") "energy", F%CE(is:ie)
      if (P%yes_ReadIni .or. F%iSCFstate .gt. 0 ) then
        if(.not. P%yes_unrestricted)then
          write (P%OUTPUT, "('#   ',a10,2x,100f10.3)") "fidelity", F%fidelity(is:ie)
        end if
      end if
      if (allocated(F%irrep_MO)) &
        write (P%OUTPUT, "('#   ',a10,2x,100i10)") "irrep", F%irrep_MO(is:ie)
      write (P%OUTPUT, "('#   ',10x,2x,a)") repeat('-', 10*(ie - is + 1))
      do j = 1, B%N_basis
        write (P%OUTPUT, "('#   ',a10,2x,100f10.6)") B%label(j), F%LC(j, is:ie)
      end do
      write (P%OUTPUT, "('#')")
    end do
    if (P%yes_sym) then
      write (P%OUTPUT, '("# final irrep-order of MOs: ")', advance='no')
      write (P%OUTPUT, '(i1)', advance='no') F%irrep_MO(1)
      do i = 2, size(F%irrep_MO)
        write (P%OUTPUT, '(",",i1)', advance='no') F%irrep_MO(i)
      end do
      write (P%OUTPUT, '()', advance='yes')
    end if
  end subroutine print_orbitals

  subroutine print_energies(P, F, string)
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    character(len=80), intent(in) :: string
    real(long) :: E_atom, entropy
    real(long) :: E0, E0_total, E1, E1_total
    integer :: OUTPUT
    OUTPUT = P%OUTPUT
    !
    ! print energies
    !  - zeroth-order electronic energy: E0
    !      E0 = SUM{e_orbital}
    !
    !  - electronic energy: E
    !      E = SUM{e_orbital} - J[rho] + E_xc[rho] - INT{ v_xc(r) * rho(r) }dr
    !          -------------- ------------------------------------------------
    !                  ==> E0                       ==> first-order correction
    !        where SUM{e_orbital} = T_s[rho] + INT{ v_eff(r) rho(r) }dr
    !              J[rho] = 1/2 * INT{ rho(r) * rho(r') / |r-r'| }dr dr'
    !      or
    !      E = Ts[rho] + INT{ v_ext(r) * rho(r) }dr + J[rho] + E_xc[rho]
    !          ------------------------------------   ------------------
    !                            ==> one-electron E   ==> two-electron E
    !    [ACHTUNG] Be careful with the sign of v_xc and E_xc in the above eqs.
    !    In this code, v_xc and E_xc already contain the minus sign.
    !
    !  - total energy: E_total = electronic energy + nuclear repulsion energy
    !
    E0 = F%E_sum_orb              ! 0th-order elec. E
    E0_total = E0 + F%E_nuc       ! 0th-order total E
    if (P%yes_cutoff) &         ! only F%E_nuc has been elec-stat corrected
      E0_total = E0_total + sum(F%E_long_range(1:2))     ! adding two elec terms
    E1 = F%E                      ! 1st-order elec. E
    E1_total = F%E_tot            ! 1st-order total E
    write (OUTPUT, "('# Energies: ', a)") trim(string)
    write (OUTPUT, "('# - zeroth-order energy (a.u.):')")
    write (OUTPUT, "('#   SUM of E_orb   = ',f16.8)") F%E_sum_orb
    if (P%yes_cutoff) &
      write (OUTPUT, "('#   [long-range]   = ',f16.8)") sum(F%E_long_range(1:2))
    write (OUTPUT, "('#   Nuc-Nuc pot E  = ',f16.8)") F%E_nuc
    write (OUTPUT, "('#   -----------------',a)") repeat("-", 16)
    write (OUTPUT, "('#   Total E(0)     = ',f16.8,'  [ ',f12.2,' eV ]')") &
      E0_total, E0_total*AU2EV
    write (OUTPUT, "('#')")
    write (OUTPUT, "('# - first-order energy (a.u.):')")
    write (OUTPUT, "('#   Kinetic E      = ',f16.8)") F%E_Ts
    write (OUTPUT, "('#   Nuc-Elec pot E = ',f16.8)", advance="no") F%E_ext
    write (OUTPUT, "('    One-Elec E = ',f16.8)") F%E_Ts + F%E_ext
    write (OUTPUT, "('#')")
    if (P%N_elec .gt. 1) then
      write (OUTPUT, "('#   Hartree E      = ',f16.8)") F%E_J
      write (OUTPUT, "('#   Exchange E     = ',f16.8)") F%E_x
      write (OUTPUT, "('#   Correlation E  = ',f16.8)", advance="no") F%E_c
      write (OUTPUT, "('    Two-Elec E = ',f16.8)") F%E_J + F%E_x + F%E_c
      write (OUTPUT, "('#                    ',16x)", advance="no")
      write (OUTPUT, "('    -------------',a)") repeat("-", 16)
    end if
    write (OUTPUT, "('#   Nuc-Nuc pot E  = ',f16.8)", advance="no") F%E_nuc
    write (OUTPUT, "('    Tot Elec E = ',f16.8)") E1
    write (OUTPUT, "('#   -----------------',a)") repeat("-", 16)
    if (P%yes_thermal) then
      entropy = calculate_entropy(F%F_occ)
      write (OUTPUT, "('#   Total Energy   = ',f16.8,'  [ ',f12.2,' eV ]')") &
        (E1_total + P%T_eV/AU2EV*entropy), (E1_total + P%T_eV/AU2EV*entropy)*AU2EV
      write (OUTPUT, "('#   Entropy=',f16.8)") entropy
      write (OUTPUT, "('#   Total Free Energy   = ',f16.8,'  [ ',f12.2,' eV ]')") &
        E1_total, E1_total*AU2EV
    else
      write (OUTPUT, "('#   Total Energy   = ',f16.8,'  [ ',f12.2,' eV ]')") &
        E1_total, E1_total*AU2EV
    end if
    write (OUTPUT, "('#')")

    if (P%N_elec .gt. 1) then
      write (OUTPUT, "('#        ',4a13)") &
        "KE(a.u.)", "E_ne(a.u.)", "E_ee(a.u.)", "E_nn(a.u.)"
      write (OUTPUT, "('# |E_COM|',4es13.5)") F%E_Ts, F%E_ext, &
        F%E_J + F%E_x + F%E_c, F%E_nuc
      write (OUTPUT, "('#')")
      write (OUTPUT, "('#        ',2a12)") "E(0)", "E(1)"
    end if
    write (OUTPUT, "('# |E_TOT|',f12.2)", advance="no") E0_total*AU2EV
    if (P%N_elec .gt. 1) write (OUTPUT, "(f12.2)", advance="no") E1_total*AU2EV
    write (OUTPUT, "()")
    write (OUTPUT, "('#')")
    !
    ! treatment for short- and long-range interactions, if r_cutoff is used
    !   short-range: inner region, r < P%r_cutoff
    !   long-range : electrostatic region, r > P%r_cutoff
    !
    if (P%yes_cutoff) then
      write (OUTPUT, "('# - above energies are calculated "// &
        "in two regions with one cutoff radius')")
      write (OUTPUT, "('#   short-range (r <= r_c) : "// &
        "full quantum interaction included')")
      write (OUTPUT, "('#   long-range  (r_c < r)  : "// &
        "monopole and dipole interactions included')")
      write (OUTPUT, "('#   r_c  (a.u.) = ',f8.2)") P%r_cutoff
      write (OUTPUT, "('#                            ',3a16)") &
        "short-range", "long-range", "whole"
      write (OUTPUT, "('#   nuc-elec energy, E_ext = ',3f16.5)") & ! E_ne -> (1)
        F%E_ext - F%E_long_range(1), F%E_long_range(1), F%E_ext
      write (OUTPUT, "('#   nuc-nuc energy,  E_nuc = ',3f16.5)") & ! E_nn -> (3)
        F%E_nuc - F%E_long_range(3), F%E_long_range(3), F%E_nuc
      write (OUTPUT, "('#   elec-elec energy,  E_J = ',3f16.5)") & ! E_ee -> (2)
        F%E_J - F%E_long_range(2), F%E_long_range(2), F%E_J
      write (OUTPUT, "('#                     E_xc = ',f16.5,16x,f16.5)") &
        F%E_x + F%E_c, F%E_x + F%E_c
      write (OUTPUT, "('#   kinetic energy,   E_Ts = ',f16.5,16x,f16.5)") &
        F%E_Ts, F%E_Ts
      write (OUTPUT, "('#   ',a)") repeat("-", 73)
      write (OUTPUT, "('#   total energy,  E_total = ',3f16.5)") &
        F%E_tot - sum(F%E_long_range(:)), sum(F%E_long_range(:)), F%E_tot
      write (OUTPUT, "('#')")
    end if
    !
    ! print the energy per atom for periodic calculations
    !
    if (P%yes_periodic) then
      E_atom = F%E_tot/dble(F%N_nuc)
      write (OUTPUT, "('# - N of atoms within the unitcell     = ',i10)") F%N_nuc
      write (OUTPUT, "('#   total energy per atom  = ',f16.8,"// &
        "'  [ ',f12.2,' eV ]')") E_atom, E_atom*AU2EV
      write (OUTPUT, "('#')")
    end if
  end subroutine print_energies

  subroutine prepare_initial_guess(P, A, G, B, F)
    !! prepares the initial guess 
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long), dimension(B%Nr_basis, B%Nr_basis) :: Hmat   ! internal H matrix
    integer :: i, j, k
    integer :: N, N_MO

    N = P%N_basis
    N_MO = F%N_MO
    !
    ! Option for restarting from  already-converged(=2) or good-old(=1) MOs ...
    !
    if (F%iSCFstate .gt. 0) then
      if (F%iSCFstate .eq. 1 .and. .not. P%yes_gaussian) then
        do i = 1, P%N_basis
          do j = 1, P%N_basis
            F%STrans(i, j) = integral(G, B%func(:, i)*F%basis_ini(:, j))
          end do
        end do
      else
        F%STrans = B%S
      end if

      if (P%yes_unrestricted) then
        F%LC_a = B%SInv.m.F%STrans.m.F%LC_Ini(:, :)
        F%LC_b(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini_b(:, :)
        call prepare_orbitals(B, F, F%LC_a, P%yes_sym)
        call prepare_orbitals(B, F, F%LC_b, P%yes_sym)
        call density_matrix( F%LC_a, F%f_occ_a, F%Da)
        call density_matrix( F%LC_b, F%f_occ_b, F%Db)
        F%D = F%Da + F%Db
      else
        F%LC(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini(:, :)
        call prepare_orbitals(B, F, F%LC, P%yes_sym)
        call density_matrix(F%LC, F%f_occ, F%D)
      end if

      F%rho_ini = density(G, B, F%D)
      F%rho(:) = F%rho_ini(:)

    else ! iSCFstate ==0
      !
      ! set up the initial guess for the density
      !
      if (P%yes_CoreIni) then
        F%D = ZERO
        F%rho(:) = ZERO
        F%v_eff(:) = F%v_ext(:)
        if (P%yes_HF) then
          F%H = F%H0
        else
          call construct_Hamiltonian(P, G, B, F)
        end if
        !Hmat(:,:) = B%SInvSqrt(:,:) .tm. F%H .m. B%SInvSqrt(:,:)
        Hmat(:, :) = B%X(:, :) .tm.F%H.m.B%X(:, :)
        if (P%yes_sym) then
          F%irrep_MO = B%irrep_sym_bfr
          call get_eigenvalues_symmetry(Hmat, F%CE, F%irrep_MO)
        else
          call get_eigenvalues(Hmat, F%CE)
        end if
        !F%LC = B%SInvSqrt(:,:) .m. Hmat(:,:)
        F%LC = B%X(:, :) .m. Hmat(:, :)
        call density_matrix(F%LC, F%f_occ, F%D)
        if(P%yes_unrestricted)then
          F%LC_a = F%LC
          F%LC_b = F%LC
          F%Da = HALF * F%D
          F%Db = F%Da
        end if
        F%rho_ini = density(G, B, F%D)
        F%rho(:) = F%rho_ini
        call calculate_energy(P, G, B, F)
      else if (P%yes_ReadIni .and. .not. P%yes_ReadShell) then
        ! read in the coefficients
        if(P%yes_unrestricted)then
          call read_wf(F, F%LC_Ini, P%wfaFileName)
          call read_wf(F, F%LC_Ini_b, P%wfbFileName)          
          F%LC_a(:, :) = F%LC_Ini(:, :)
          F%LC_b(:, :) = F%LC_Ini_b(:, :)
        else
          call read_wf(F, F%LC_Ini, P%wfFileName)
          F%LC(:, :) = F%LC_Ini(:, :)
        end if
        ! we need this for the fidelity checking and MOM:
        F%STrans = B%S
        ! Do a Gram-Schmidt orthogonalization to prepare the initial guess
        ! in case it is not orthonormal, (e.g. it was
        ! calculated on a different basis or molecular geometry)
        if(P%yes_unrestricted)then
          call prepare_orbitals(B, F, F%LC_a, P%yes_sym)
          call prepare_orbitals(B, F, F%LC_b, P%yes_sym)
          call density_matrix( F%LC_a, F%f_occ_a, F%Da)
          call density_matrix( F%LC_b, F%f_occ_b, F%Db)
          F%D = F%Da + F%Db
        else
          F%LC(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini(:, :)
          call prepare_orbitals(B, F, F%LC, P%yes_sym)
          call density_matrix(F%LC, F%f_occ, F%D)
        end if
        F%rho_ini = density(G, B, F%D)
      
      else if (P%yes_ReadIni .and. P%yes_ReadShell) then
        call read_basis(F, F%basis_ini, P%basisFileName)
        if(P%yes_unrestricted)then
          call read_wf(F, F%LC_Ini, P%wfaFileName)
          call read_wf(F, F%LC_Ini_b, P%wfbFileName)
        else
          call read_wf(F, F%LC_Ini, P%wfFileName)
        end if

        ! Calculate the overlap matrix STrans between the two basis sets
        do i = 1, P%N_basis
          do j = 1, P%N_basis
            F%STrans(i, j) = integral(G, B%func(:, i)*F%basis_ini(:, j))
          end do
        end do
        if (P%yes_sym) then
          if (allocated(P%irrep_MO)) &
            F%irrep_MO = P%irrep_MO
        end if        
        ! Basis transformation from the given basis into the new basis
        ! basis is non-orthogonal so we need Sinv=S^-1
        if(P%yes_unrestricted)then
          F%LC_a(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini(:, :)
          call prepare_orbitals(B, F, F%LC_a, P%yes_sym)
          F%LC_b(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini_b(:, :)
          call prepare_orbitals(B, F, F%LC_b, P%yes_sym)
          call density_matrix( F%LC_a, F%f_occ_a, F%Da)
          call density_matrix( F%LC_b, F%f_occ_b, F%Db)
          F%D = F%Da + F%Db          
        else
          F%LC(:, :) = B%SInv.m.F%STrans.m.F%LC_Ini(:, :)
          call prepare_orbitals(B, F, F%LC, P%yes_sym)
          call density_matrix(F%LC, F%f_occ, F%D)
        end if
        if(P%yes_unrestricted)then
          ! calculate the density from provided basis set
          F%rho_ini = ZERO
          do k = 1, N_MO          ! 1 ... # of occupied MOs
            do i = 1, F%N_basis     ! 1 ... # of atomic orbitals
              do j = 1, F%N_basis     ! 1 ... # of atomic orbitals
                F%rho_ini(:) = F%rho_ini(:) + &
                  (F%f_occ_a(k) * F%LC_Ini(i, k)*F%LC_Ini(j, k) + &
                  F%f_occ_b(k) * F%LC_Ini_b(i, k)*F%LC_Ini_b(j, k) ) &
                  * F%basis_ini(:, i)*F%basis_ini(:, j)
              end do
            end do
          end do          
        else
          ! calculate the density from provided basis set
          F%rho_ini = ZERO
          do k = 1, N_MO          ! 1 ... # of occupied MOs
            do i = 1, F%N_basis     ! 1 ... # of atomic orbitals
              do j = 1, F%N_basis     ! 1 ... # of atomic orbitals
                F%rho_ini(:) = F%rho_ini(:) + &
                  F%f_occ(k)*F%LC_Ini(i, k)*F%LC_Ini(j, k)* &
                  F%basis_ini(:, i)*F%basis_ini(:, j)
              end do
            end do
          end do
        end if

      else if (.not. P%yes_ReadIni .and. P%yes_ReadShell) then
        ! only the basis was given
        ! don't know what to do here
        stop 'Please provide an initial guess for the coefficients'
      else
        ! default: we start from a real space density provided by initial_density()
        ! do a zeroth step to get mo's and density matrix
        F%rho(:) = initial_density(P, G, A, B)
        F%D = initial_density_matrix(P, A, B)
        F%rho_ini = F%rho
        if (.not. P%yes_HF) then
          if (len(trim(P%gto_file_path)) .gt. 0) then
            ! if we use gaussian orbitals: skip calculating v_J
            F%v_xc(:) = potential_xc(P, F%rho)
            F%v_eff(:) = F%v_ext(:) + F%v_xc(:)
          else
            F%v_J(:) = potential_J(P, G, A, F%rho, F%Q_A, F%dipole_A)
            F%v_xc(:) = potential_xc(P, F%rho)
            F%v_eff(:) = F%v_ext(:) + F%v_J(:) + F%v_xc(:)
          end if
        end if
        if (P%yes_HF) then
          !if(.not. P%yes_openshell)then
          ! start with a closed shell fock matrix (even if we have open shell)
          call construct_Fock_Matrix(P, B, F)
          !else
          !  call construct_ROHF_Fock_Matrix(P,A,G,B,F,variant=1)
          !end if
        else
          call construct_Hamiltonian(P, G, B, F)
        end if
        !call construct_Hamiltonian( P, G, B, F )

        Hmat(:,:) = B%X .tm. F%H .m. B%X
        if(P%yes_sym)then
          call get_eigenvalues_symmetry(Hmat, F%CE, B%irrep_sym_bfr)
          F%irrep_MO = B%irrep_sym_bfr
        else
          call get_eigenvalues( Hmat, F%CE )  
        end if
        F%LC = B%X(:, :) .m.Hmat(:, :)
        if (P%yes_sym) then
          if (allocated(P%irrep_MO)) then
            call reorder_irrep(P, B, F, P%irrep_MO)
          else
            call reorder_irrep_by_CE(P, B, F, F%CE)
          end if
        end if
        call density_matrix(F%LC, F%f_occ, F%D)
        if(P%yes_unrestricted)then
          F%LC_a = F%LC
          F%LC_b = F%LC
          F%D_a = HALF * F%D
          F%D_b = HALF * F%D
        end if        
      end if
      F%rho(:) = density(G, B, F%D)
    end if ! iSCFstate==0

    contains
    subroutine prepare_orbitals(B, F, LC, yes_sym)
      type(Basis), intent(in) :: B
      type(WF_data), intent(in) :: F
      real(long), intent(inout)  :: LC(:,:)
      logical, intent(in) :: yes_sym

      integer, allocatable :: ndx(:), ndxB(:)
      real(long), allocatable :: projector(:, :)
      integer :: i, j
      if (yes_sym) then
        do i = 1, B%nirrep
          allocate (ndx(count(F%irrep_MO(:) .eq. i)))
          allocate (ndxB(count(B%irrep_sym_bf(:) .eq. i)))
          allocate (projector(size(LC, 1), size(LC, 1)))
          ndx = pack([(j, j=1, size(F%irrep_MO, 1))], F%irrep_MO .eq. i)
          ndxB = pack([(j, j=1, size(B%irrep_sym_bf, 1))], B%irrep_sym_bf .eq. i)
          projector = B%sym_bf(:, ndxB) .mt.B%sym_bf(:, ndxB)
          LC(:, ndx) = projector.m.LC(:, ndx)
          deallocate (projector)
          deallocate (ndx)
          deallocate (ndxB)
        end do
        call orthonormalizeOrbitals(B, LC, F%irrep_MO)
      else
        call orthonormalizeOrbitals(B, LC)
      end if
    end subroutine prepare_orbitals

  end subroutine prepare_initial_guess

  subroutine reorder_irrep_by_CE(P, B, F, CE)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long) :: CE(:)
    logical :: mk(size(CE, 1))
    integer :: irrep_MO(size(CE))
    integer :: i, j
    ! sort orbitals according to orbital energy

    mk(:) = .TRUE.
    do i = 1, size(CE)
      j = minloc(CE, dim=1, mask=mk)
      irrep_MO(i) = F%irrep_MO(j)
      mk(j) = .False.
    end do

    if (.not. all(irrep_MO .eq. F%irrep_MO)) then
      if (P%yes_verbose) &
        write (P%OUTPUT, *) '# reorder irreps according to orbital energy'
      call reorder_irrep(P, B, F, irrep_MO)
    end if
  end subroutine reorder_irrep_by_CE

  subroutine reorder_irrep(P, B, F, new_irrep_MO)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: new_irrep_MO(:)

    integer :: i, j, k
    real(long) :: temp

    do i = 1, B%nirrep
      if (count(new_irrep_MO .eq. i) .ne. count(F%irrep_mo .eq. i)) then
        write (P%OUTPUT, *) 'irrep', i, count(new_irrep_MO .eq. i), count(F%irrep_mo .eq. i)
        stop "irreps do not match!"
      end if
    end do

    do i = 1, B%Nr_basis
      if (new_irrep_MO(i) .ne. F%irrep_MO(i)) then
        do k = i + 1, B%Nr_basis
          if (F%irrep_MO(k) .eq. new_irrep_MO(i)) exit
        end do
        if (k .eq. B%Nr_basis .and. F%irrep_MO(k) .ne. new_irrep_MO(i)) &
          stop 'error reordering orbitals according to irrep_MO'
        ! swap i <-> k
        do j = 1, B%N_basis
          temp = F%LC(j, i)
          F%LC(j, i) = F%LC(j, k)
          F%LC(j, k) = temp
        end do
        temp = F%CE(i)
        F%CE(i) = F%CE(k)
        F%CE(k) = temp
        if(P%yes_unrestricted)then
          do j =1, B%N_basis
            temp=F%LC_a(j,i)
            F%LC_a(j,i) = F%LC_a(j,k)
            F%LC_a(j,k) = temp
            temp=F%LC_b(j,i)
            F%LC_b(j,i) = F%LC_b(j,k)
            F%LC_b(j,k) = temp
          end do
          temp = F%CE_a(i)
          F%CE_a(i) = F%CE_a(k)
          F%CE_a(k) = temp
          temp = F%CE_b(i)
          F%CE_b(i) = F%CE_b(k)
          F%CE_b(k) = temp
        end if        
        j = F%irrep_MO(i)
        F%irrep_MO(i) = F%irrep_MO(k)
        F%irrep_MO(k) = j
      end if

    end do
  end subroutine reorder_irrep

  subroutine Solve_OE_Eigensystem(P, F)
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    integer i, k, j, i_spin
    integer :: N, N_MO, N_MO_nf
    real(long), dimension(size(F%LC, 2), size(F%LC, 2)) :: Hmat      ! internal H matrix
    real(long), allocatable :: Hmat_nf(:,:)      ! internal H matrix
    integer, allocatable :: nf_index(:), irrep_MO_nf(:)
    real(long), allocatable :: CE_nf(:)
    logical :: do_eigensolver
    N = P%N_basis
    N_MO = size(Hmat, DIM=1)
    do_eigensolver = .True.
    i_spin = 1
    ! This loop is executed only once for restricted SCF and twice for unrestricted SCF
    do while (do_eigensolver)
      if(P%yes_unrestricted .and. i_spin == 1) then
        ! set H and LC for beta spin-orbitals
        F%H = F%Fockb
        F%LC = F%LC_b
      else if (P%yes_unrestricted .and. i_spin == 2) then
        ! set H and LC for alpha spin-orbitals
        F%H = F%Focka
        F%LC = F%LC_a
      end if
      ! transform H into MO space
      Hmat = F%LC(:, :) .tm.F%H(:, :) .m.F%LC(:, :)
      ! Add Level Shifts
      if (P%yes_LevelShift) forall (i=1:size(F%LC, 2)) Hmat(i, i) = Hmat(i, i) + F%LevelShift(i)
      if( allocated(F%frozen_MO) ) then
        N_MO_nf = N_MO-size(F%frozen_MO)
        if (N_MO_nf > 0) then 
          ! Build temporary arrays for non-frozen MO
          allocate(nf_index(N_MO_nf), CE_nf(N_MO_nf))
          if (P%yes_sym) then
            allocate(irrep_MO_nf(N_MO_nf))
          end if
          j=1
          do i = 1, N_MO
            if (any(F%frozen_MO(:) == i)) then
              F%CE(i) =  Hmat(i,i)
            else
              nf_index(j) = i
              if (P%yes_sym) then
                irrep_MO_nf(j) = F%irrep_MO(i)
              end if
              j = j+1 
            end if
          end do 
          allocate(Hmat_nf(N_MO_nf,N_MO_nf))
          Hmat_nf(:,:) = Hmat(nf_index(:),nf_index(:))
          ! Solve eigenvalue problem for non-frozen MO
          if (P%yes_sym) then
            call get_eigenvalues_symmetry(Hmat_nf, CE_nf, irrep_MO_nf)
          else
            call get_eigenvalues(Hmat_nf, CE_nf)  
          end if
          ! Reconstruct Hmat and CE arrays for all MO
          Hmat(:,:) = ZERO
          Hmat(nf_index,nf_index) = Hmat_nf(:,:)
          F%CE(nf_index) = CE_nf(:)
          do i = 1, size(F%frozen_MO)
            Hmat(F%frozen_MO(i), F%frozen_MO(i)) = ONE
          end do
          deallocate(nf_index,Hmat_nf,CE_nf)
          if (P%yes_sym) then
            deallocate(irrep_MO_nf)
          end if
        else 
          ! Deal with freeze all
          do i = 1, N_MO
            F%CE(i) =  Hmat(i,i)
            Hmat(:,i) = ZERO
            Hmat(i,i) = ONE
          end do 
        end if
      else
        ! solve eigenvalue problem for all MO
        if (P%yes_sym) then
          call get_eigenvalues_symmetry(Hmat, F%CE, F%irrep_MO)
        else
          call get_eigenvalues( Hmat, F%CE )  
        end if
      end if
      ! substract the applied Level Shift from the eigenvalues
      if (P%yes_LevelShift) F%CE(:) = F%CE(:) - F%LevelShift(:)
      ! restrict orbital interchanges
      ! like in GAMESS-US RSTRCT option
      if (P%yes_RSTRCT .or. F%switch_RSTRCT_on) &
        call reorder_RSTRCT(P, Hmat, F%CE)
      ! transform eigenvector back into bf space
      F%LC(:, :) = F%LC(:, :) .m. Hmat(:, :)
      ! @todo: make an unrestricted version for MOM
      if (P%yes_unrestricted .and. i_spin == 1) then
         if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_MOM) &
           call reorder_MOM(P, F, F%LC_ini_b, F%f_occ_b, F%STrans)
         if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_chk_fid_ini) &
           call reorder_MOFID(P, F, F%LC_ini_b, F%STrans)      
      else if (P%yes_unrestricted .and. i_spin == 2) then
        if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_MOM) &
          call reorder_MOM(P, F, F%LC_ini,  F%f_occ_a, F%STrans)
        if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_chk_fid_ini) &
          call reorder_MOFID(P, F, F%LC_ini, F%STrans)
      else
        if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_MOM) &
          call reorder_MOM(P, F, F%LC_ini, F%f_occ, F%STrans)
        if ((P%yes_ReadIni .or. F%iSCFstate .gt. 0) .and. P%yes_chk_fid_ini) &
          call reorder_MOFID(P, F, F%LC_ini, F%STrans)
      end if
      ! Phase correction for orbitals:
      ! coefficients are positive on the largest component
      ! this avoids phase flips in the orbitals (just cosmetics)
      ! -> it's easier to compare output when phase is defined
      do i =1,size(F%LC,2)
        k=maxloc( abs(F%LC(:,i)),1)
        if(F%LC(k,i) < 0 )  F%LC(:,i)=-F%LC(:,i)
      end do
      if(P%yes_unrestricted .and. i_spin == 1) then
          ! assign results to beta part
          F%LC_b = F%LC
          F%CE_b = F%CE
          i_spin = 2
      else if (P%yes_unrestricted .and. i_spin == 2) then
          ! assign results to alpha part
          F%LC_a = F%LC
          F%CE_a = F%CE      
          do_eigensolver = .False.
      else
        do_eigensolver = .False.
      end if
    end do
  contains
    subroutine reorder_MOM(P, F, LCref, f_occ, SI)
      !! This is the subroutine to rearrange MOs according to best match to the
      !! MO space of the occupied guess orbitals -> The Maximum Overlap Method
      type(Param), intent(in) :: P
      type(WF_data), intent(inout) :: F
      real(long), dimension(:, :), intent(in) :: LCref ! The reference MO coefficients
      real(long) :: f_occ(:)
      real(long), dimension(:, :), intent(in) :: SI   ! The overlap matrix SI(i,j) between the
      ! basis function i and the reference
      ! basis function j
      real(long), dimension(size(F%LC, 2), size(LCref, 2)) :: O  ! the MO-MO overlap matrix
      real(long), allocatable :: Proj(:, :)  ! for each occupation number 0,1, and 2, and special separately
      integer :: i, j, k, occ, occi, occj, occk, opt
      integer :: N
      real(long) :: temp
      N = size(F%LC, 2)
      ! compute the overlap matrix to the reference MO
      do i = 1, N
        do j = 1, N
          O(i, j) = dot_product(F%LC(:, i), MATMUL(SI, LCref(:, j)))
        end do
      end do
      if (allocated(F%av_occ)) then
        allocate (Proj(F%Nr_basis, 3 + size(F%av_occ)))
      else
        allocate (Proj(F%Nr_basis, 3))
      end if
      ! compute the overlap matrix to the reference MO
      do i = 1, N
        do j = 1, N
          O(i, j) = dot_product(F%LC(:, i), MATMUL(SI, LCref(:, j)))
        end do
      end do
      ! compute the Projections
      do i = 1, N
        Proj(i, :) = ZERO
        do j = 1, N
          occ = occ_space(F, j)
          ! projection into the orbital subspace of respective occupation
          !
          if (P%MOMtype .eq. 1) then
            ! original version
            Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)
          else if (P%MOMtype .eq. 2) then
            ! GAMESS version
            Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)
          else if (P%MOMtype .eq. 3) then
            ! Q-chem version
            Proj(i, occ + 1) = Proj(i, occ + 1) + O(i, j)*O(i, j)
          end if
        end do
      end do
      if (P%MOMtype .eq. 1) then
        ! original version
        Proj(:, :) = abs(Proj(:, :))
      end if
      ! loop over all orbital pairs:
      ! consider all orbital swaps (i <-> j )
      ! and find the orbital swaps that gives the largest gain in the MOM criterion
      do
        opt = 1
        do i = N, 1, -1
          occi = occ_space(F, i)
          k = i
          occk = occi
          do j = N, i + 1, -1
            if (P%yes_valgs_occ .and. j .gt. F%Ncore .and. i .gt. F%Ncore) cycle
            if (allocated(F%irrep_MO)) then
              if (F%irrep_MO(i) .ne. F%irrep_MO(j)) cycle
            end if
            occj = occ_space(F, j)
            if (occi .eq. occj) cycle
            ! old version
            !if( Proj(i,occi+1) + Proj(j,occj+1) < Proj(i,occj+1) + Proj(j,occi+1) )then
            if( f_occ(i) * log(Proj(i,occi+1)) + f_occ(j) * log(Proj(j,occj+1)) < f_occ(j) * log(Proj(i,occj+1)) + f_occ(i) * log(Proj(j,occi+1)) )then
              k = j
              occk = occj
              exit
            end if
          end do ! j loop
          if (i .ne. k) then
            ! swap the orbitals i <-> k , orbitaleigenvalues and projections
            if (P%yes_verbose) write (P%OUTPUT, "('# MOM: swapping MOs',i3,i3)") i, k
            if(P%yes_verbose) write(P%OUTPUT,"('# MOM: orbital ',i3,'occ space ',i3,f5.1,'->','occ space',i3,f5.1)") i,occi,Proj(i,occi+1),occj,Proj(i,occj+1)
            if(P%yes_verbose) write(P%OUTPUT,"('# MOM: orbital ',i3,'occ space ',i3,f5.1,'->','occ space',i3,f5.1)") j,occj,Proj(j,occj+1),occi,Proj(j,occi+1)
            do j = 1, size(F%LC, 1)
              temp = F%LC(j, i)
              F%LC(j, i) = F%LC(j, k)
              F%LC(j, k) = temp
            end do
            temp = F%CE(i)
            F%CE(i) = F%CE(k)
            F%CE(k) = temp
            do j = 1, size(Proj(:, :), 2)
              temp = Proj(i, j)
              Proj(i, j) = Proj(k, j)
              Proj(k, j) = temp
            end do
            opt = 0
          end if ! if i /= k
        end do ! i loop
        if (opt .eq. 1) exit
      end do
      ! reorder valence electrons according to energy
      if (P%yes_valgs_occ) then
        do i = F%Ncore + 1, N
          do k = i + 1, N
            if (allocated(F%irrep_MO)) then
              if (F%irrep_MO(i) .ne. F%irrep_MO(k)) cycle
            end if
            if (F%CE(i) .gt. F%CE(k)) then
              do j = 1, size(F%LC, 1)
                temp = F%LC(j, i)
                F%LC(j, i) = F%LC(j, k)
                F%LC(j, k) = temp
              end do
              temp = F%CE(i)
              F%CE(i) = F%CE(k)
              F%CE(k) = temp
              do j = 1, size(Proj(:, :), 2)
                temp = Proj(i, j)
                Proj(i, j) = Proj(k, j)
                Proj(k, j) = temp
              end do!j
            end if
          end do!k
        end do!i
      end if
      if (P%yes_verbose) then
        do i = 1, size(Proj(:, :), 2)
          write (P%OUTPUT, "('# MOM: ')", advance='no')
          if (i .eq. 1) then
            write (P%OUTPUT, "(' occ=',i5)", advance='no') 0
          else if (i .eq. 2) then
            write (P%OUTPUT, "(' occ=',i5)", advance='no') 1
          else if (i .eq. 3) then
            write (P%OUTPUT, "(' occ=',i5)", advance='no') 2
          else
            write (P%OUTPUT, "(' occ=',f5.2)", advance='no') F_occ(F%av_occ(i - 3)%MO(1))
          end if
          do j = 1, size(Proj(:, :), 1)
            if (occ_space(F, j) + 1 .eq. i) &
              write (P%OUTPUT, "( f7.3,'(',i3,')')", advance='no') Proj(j, i), j ! MO j occ space i
          end do
          write (P%OUTPUT, "()")
        end do
        write (P%OUTPUT, "('# MOM Prod_i P_i(n_i)^{n_i} = ',f15.5)") prodOverlap(P, F, SI, LCref)
      end if
    end subroutine reorder_MOM

    subroutine reorder_RSTRCT(P, C, CE)
      !! This is restriction of orbital interchanges as done in GAMESS-US by the RSTRCT option
      !! arguments
      !! P: Parameter structure
      !! C: the MO orbital coeficient after diagonalizing H in MO space
      !!    (i.e. the rotation of MOs)
      !! CE: the eigenvalues
      !! it just reorders the eigenvectors such that C is
      !! similar to a diagonal matrix
      type(Param), intent(in) :: P
      real(long), dimension(:, :), intent(inout) :: C
      real(long), dimension(:), intent(inout) :: CE
      integer i, j, k
      integer N
      real(long) temp
      N = size(CE)
      do i = 1, N
        k = i
        do j = 1, N
          if (abs(C(i, j)) .gt. abs(C(i, k)) .and. abs(C(i, j)) .gt. abs(C(j, j))) then
            k = j
          end if
        end do
        if ((k .ne. i)) then
          ! there is another eigenvector with larger amplitude for component i
          ! so swap k <-> i
          if (P%yes_verbose) write (P%OUTPUT, "('# RSTRCT: swapping MOs',i3,i3)") i, k
          do j = 1, N
            temp = C(j, i)
            C(j, i) = C(j, k)
            C(j, k) = temp
          end do
          temp = CE(i)
          CE(i) = CE(k)
          CE(k) = temp
        end if
      end do
    end subroutine reorder_RSTRCT

    subroutine reorder_MOFID(P, F, Cref, SI)
      !! This is the subroutine to rearrange MOs according to best match to the 
      !! guess MOs: 'Orbital Fidelity'. Orbitals are reordered according to O(i,j)**2
      !!      type(Param), intent(in) :: P      
      type(Param), intent(in) :: P
      type(WF_data), intent(inout) :: F
      real(long), dimension(:, :), intent(in) :: Cref ! The reference MO coefficients
      real(long), dimension(:, :), intent(in) :: SI   ! The overlap matrix SI(i,j) between the
      ! basis function i and the reference
      ! basis function j
      real(long), dimension(size(F%LC, 2), size(Cref, 2)) :: O  ! the MO-MO overlap matrix
      real(long) :: temp
      integer :: i, j, k, l, mu
      integer :: N
      integer, dimension(2) :: ik
      N = size(F%LC, 2)
      ! compute the overlap matrix to the reference MO
      ! Note: Left index is actual orbital, right index is reference orbital
      do i = 1, N
        do j = 1, N
          O(i, j) = dot_product(F%LC(:, i), MATMUL(SI, Cref(:, j)))
        end do
      end do
      ! we go through the reference orbitals i and pick the orbital j
      ! which has largest |Oij*Oij|
      do l = 1, N ! loop over orbitals
        ik = maxloc(O(:, :)**2)
        i = ik(1)
        k = ik(2)
        ! reference orbital k best fits current orbital i
        ! delete all entries referring to reference orbital k
        O(:, k) = ZERO
        O(i, :) = ZERO
        !    ! if we have same occupancy: don't interchange orbitals (except if final_reordering=.TRUE. was given)
        if (i .ne. k) then
          !       ! swap the orbitals i and k, the orbitaleigenvalues CE(i), CE(k) and the overlap O(i,:) and O(k,:)
          if (P%yes_verbose) write (P%OUTPUT, "('#  MOFID: swapping MOs',i3,i3,2f10.3)") i, k, F%CE(i), F%CE(k)
          do mu = 1, size(F%LC, 1)
            temp = F%LC(mu, i)
            F%LC(mu, i) = F%LC(mu, k)
            F%LC(mu, k) = temp
          end do
          temp = F%CE(i)
          F%CE(i) = F%CE(k)
          F%CE(k) = temp
          do j = 1, N
            temp = O(i, j)
            O(i, j) = O(k, j)
            O(k, j) = temp
          end do
        end if ! if i /= k
      end do ! l loop
      ! reorder valence electrons according to energy
      if (P%yes_valgs_occ) then
        do i = F%Ncore + 1, N
          do k = i + 1, N
            if (F%CE(i) .gt. F%CE(k)) then
              do j = 1, size(F%LC, 2)
                temp = F%LC(j, i)
                F%LC(j, i) = F%LC(j, k)
                F%LC(j, k) = temp
              end do
              temp = F%CE(i)
              F%CE(i) = F%CE(k)
              F%CE(k) = temp
              do j = 1, N
                temp = O(i, j)
                O(i, j) = O(k, j)
                O(k, j) = temp
              end do
            end if
          end do
        end do
      end if
      do i = 1, N
        do j = 1, N
          O(i, j) = dot_product(F%LC(:, i), MATMUL(SI, Cref(:, j)))
        end do
      end do
      do i = 1, N
        F%fidelity(i) = O(i, i)*O(i, i)
      end do
      if (P%yes_verbose) then
        write (P%OUTPUT, "('# MOFID: ')", advance='no')
        do j = 1, size(O, 1)
          write (P%OUTPUT, "( f7.3,'(',i3,')')", advance='no') O(j, j)**2, j ! MO j occ space i
          if (mod(j, 10) .eq. 0 .and. j .ne. size(O, 1)) then
            write (P%OUTPUT, "()", advance='yes')
            write (P%OUTPUT, "('# MOFID: ')", advance='no')
          end if
        end do
        write (P%OUTPUT, "()", advance='yes')
      end if
    end subroutine reorder_MOFID

  end subroutine Solve_OE_Eigensystem

  subroutine construct_h0_matrix(G, B, F)
    !!
    !! calculates the H0 matrix for given basis B: hkin+v_ext contribution
    !!
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer :: i, j
    if (allocated(B%vext)) then
      F%H0(:, :) = B%Hkin(:, :) + B%vext(:, :)
    else
      do i = 1, B%N_basis
        do j = i, B%N_basis
          F%H0(i, j) = B%Hkin(i, j) + braket(G, B, i, j, F%v_ext)
          if (i .ne. j) F%H0(j, i) = F%H0(i, j)
        end do
      end do
    end if
  end subroutine construct_h0_matrix

  subroutine ROHF_canonicalize(P, B, F)
    !! ROHF canonicalization
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long), allocatable :: H_double(:, :), CE_double(:)
    real(long), allocatable :: H_open(:, :), CE_open(:)
    real(long), allocatable :: H_virt(:, :), CE_virt(:)
    real(long) :: Hmat(size(F%LC, 2), size(F%LC, 2))
    integer, allocatable :: double(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    logical :: is_frozen(size(F%LC, 2))
    integer :: j
    integer :: Ndouble, Nvirt, Nopen
    is_frozen(:) = .False.
    if (allocated(F%frozen_MO)) then
      is_frozen(F%frozen_MO) = .True.
    end if
    allocate (active(count(F%N_occ(:) .eq. 1 .and. .not. is_frozen)))
    allocate (double(count(F%N_occ(:) .eq. 2 .and. .not. is_frozen)))
    allocate (virtual(count(F%N_occ(:) .eq. 0 .and. .not. is_frozen)))
    active(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 1 .and. .not. is_frozen)
    double(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 2 .and. .not. is_frozen)
    virtual(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 0 .and. .not. is_frozen)
    Ndouble = size(double)
    Nopen = size(active)
    Nvirt = size(virtual)
    allocate (H_double(Ndouble, Ndouble))
    allocate (CE_double(Ndouble))
    allocate (H_open(Nopen, Nopen))
    allocate (CE_open(Nopen))
    allocate (H_virt(Nvirt, Nvirt))
    allocate (CE_virt(Nvirt))
    call construct_ROHF_Fock_Matrix(P, B, F, variant=P%ROHF_variant)
    Hmat = F%LC .tm. F%H(:, :) .m. F%LC
    H_double = Hmat(double, double)
    H_open = Hmat(active, active)
    H_virt = Hmat(virtual, virtual)
    if (Ndouble .gt. 0) then
      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_double, CE_double, F%irrep_MO(double(:)))
      else
        call get_eigenvalues(H_double, CE_double)
      end if
      !call reorder_RSTRCT(P,H_double,CE_double)
      F%LC(:, double(:)) = matmul(F%LC(:, double(:)), H_double(:, :))
      F%CE(double(:)) = CE_double
    end if
    if (Nopen .gt. 0) then
      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_open, CE_open, F%irrep_MO(active(:)))
      else
        call get_eigenvalues(H_open, CE_open)
      end if
      !call reorder_RSTRCT(P,H_open,CE_open)
      F%LC(:, active) = matmul(F%LC(:, active(:)), H_open(:, :))
      F%CE(active) = CE_open
    end if
    if (Nvirt .gt. 0) then
      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_virt, CE_virt, F%irrep_MO(virtual(:)))
      else
        call get_eigenvalues(H_virt, CE_virt)
      end if
      !call reorder_RSTRCT(P,H_virt,CE_virt)
      F%LC(:, virtual) = matmul(F%LC(:, virtual(:)), H_virt(:, :))
      F%CE(virtual) = CE_virt
    end if
    deallocate (H_open)
    deallocate (CE_open)
    deallocate (H_double)
    deallocate (CE_double)
    deallocate (H_virt)
    deallocate (CE_virt)
  end subroutine ROHF_canonicalize

  !>
  !! This is the subroutine to rearrange MOs according to best match to the
  !! guess MOs: 'Orbital Fidelity'. Orbitals are reordered according to O(i,j)**2
  !!
  subroutine reorder_final(P, F, Cref, SI)
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), dimension(:, :), intent(in) :: Cref ! The reference MO coefficients
    real(long), dimension(:, :), intent(in) :: SI   ! The overlap matrix SI(i,j) between the
    ! basis function i and the reference
    ! basis function j
    real(long), dimension(P%N_basis, P%N_basis) :: O  ! the MO-MO overlap matrix
    integer i, j, k, occk, mu
    integer N, occi
    real(long) temp
    N = size(F%LC, 2)
    ! compute the overlap matrix to the reference MO
    ! Note: Left index is actual orbital, right index is reference orbital
    do i = 1, N
      do j = 1, N
        O(i, j) = dot_product(F%LC(:, i), MATMUL(SI, Cref(:, j)))
      end do
    end do
    ! we go through the reference orbitals i and pick the orbital j
    ! which has largest |Oij*Oij|
    do i = 1, N ! loop over reference orbitals
      ! find the largest Projection O(i,j)**2 into
      ! reference orbital: selection sort algorithm
      k = i
      do j = i, N ! loop over actual orbitals
        if (P%yes_valgs_occ .and. i .gt. F%Ncore .and. j .gt. F%Ncore) cycle
        ! don't touch an orbital pair if they are supposed to be localized later
        ! for reasons I don't understand (in combination with fix_nuc) this can cause problems
        if (allocated(P%localize)) then
          if (any(i .eq. P%localize(:)) .and. any(j .eq. P%localize(:))) cycle
        end if
        if (allocated(F%irrep_MO)) then
          if (F%irrep_MO(i) .ne. F%irrep_MO(j)) cycle
        end if
        ! look for an orbital with larger overlap:
        if (O(j, i)*O(j, i) .gt. O(k, i)*O(k, i)) then
          k = j
        end if
      end do
      occi = occ_space(F, i)
      occk = occ_space(F, k)
      if (i .ne. k .and. (occi .eq. occk)) then
        ! swap the orbitals i and k, the orbitaleigenvalues CE(i), CE(k) and the overlap O(i,:) and O(k,:)
        if (P%yes_verbose) write (P%OUTPUT, "('#  MOFID: swapping MOs',i3,i3,2f10.3)") i, k, F%CE(i), F%CE(k)
        do mu = 1, size(F%LC, 1)
          temp = F%LC(mu, i)
          F%LC(mu, i) = F%LC(mu, k)
          F%LC(mu, k) = temp
        end do
        temp = F%CE(i)
        F%CE(i) = F%CE(k)
        F%CE(k) = temp
        temp = F%F_occ(i)
        F%F_occ(i) = F%F_occ(k)
        F%F_occ(k) = temp
        do j = 1, N
          temp = O(i, j)
          O(i, j) = O(k, j)
          O(k, j) = temp
        end do
      end if ! if i /= k
    end do ! i loop
    ! reorder valence electrons according to energy
    if (P%yes_valgs_occ) then
      do i = F%Ncore + 1, N
        do k = i + 1, N
          if (allocated(F%irrep_MO)) then
            if (F%irrep_MO(i) .ne. F%irrep_MO(k)) cycle
          end if
          if (F%CE(i) .gt. F%CE(k)) then
            do j = 1, size(F%LC, 1)
              temp = F%LC(j, i)
              F%LC(j, i) = F%LC(j, k)
              F%LC(j, k) = temp
            end do
            temp = F%CE(i)
            F%CE(i) = F%CE(k)
            F%CE(k) = temp
            do j = 1, N
              temp = O(i, j)
              O(i, j) = O(k, j)
              O(k, j) = temp
            end do
          end if
        end do
      end do
    end if
    do i = 1, N
      F%fidelity(i) = O(i, i)*O(i, i)
    end do
    if (P%yes_verbose) then
      write (P%OUTPUT, "('# Overlap^2 to Reference   ')")
      write (P%OUTPUT, "('#  ',10i7)") 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
      do i = 1, N
        write (P%OUTPUT, "('#  ',10f7.3)") O(i, :)**2
      end do
      write (P%OUTPUT, "('# Overlap^2 diagonal  ')")
      do i = 1, N
        write (P%OUTPUT, "('#  ',i0,10f7.3)") i, O(i, i)**2
      end do
    end if
  end subroutine reorder_final

  subroutine stability_analysis(P, A, G, B, F)
    !! performs stability analysis by inspecting numerical derivatives of the Fock matrix
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long) :: temp
    real(long), parameter :: angle = 0.0005
    type(indep_mo_pair), allocatable :: pairs(:)
    integer :: i, j, k
    real(long), dimension(B%Nr_basis, B%Nr_basis) :: Hmat      ! internal H matrix
    real(long), dimension(B%N_basis, B%Nr_basis) :: LCsave      ! internal H matrix
    real(long), allocatable :: hessian(:, :)
    real(long), allocatable :: theta(:)
    real(long), allocatable :: gradient(:, :)
    real(long), allocatable :: hessian_ev(:)
    LCsave = F%LC
    call prepare_independent_pairs(P, F, pairs)
    allocate (hessian(size(pairs, 1), size(pairs, 1)))
    allocate (theta(size(pairs, 1)))
    allocate (gradient(size(pairs, 1), 2))
    allocate (hessian_ev(size(pairs, 1)))
    F%v_ext(:) = potential_ext(G, A)
    do i = 1, size(pairs, 1)
      theta = zero
      do k = 1, 2
        theta(i) = angle*(2*k - 3) ! 2*1-3=-1, 2*2-4=1
        F%LC = LCsave
        call rotate_orbitals(theta, pairs, F%LC)
        call density_matrix(F%LC, F%f_occ, F%D)
        if (.not. P%yes_HF) then
          F%rho(:) = density(G, B, F%D)
          if (len(trim(P%gto_file_path)) .gt. 0) then
            F%v_xc(:) = potential_xc(P, F%rho)
            F%v_eff(:) = F%v_ext(:) + F%v_xc(:)
          else
            F%v_J(:) = potential_J(P, G, A, F%rho)
            F%v_xc(:) = potential_xc(P, F%rho)
            F%v_eff(:) = F%v_ext(:) + F%v_J(:) + F%v_xc(:)
          end if
        end if
        if (P%yes_HF) then
          if (.not. P%yes_openshell) then
            call construct_Fock_Matrix(P, B, F)
          else
            call construct_ROHF_Fock_Matrix(P, B, F, variant=1)
          end if
        else
          call construct_Hamiltonian(P, G, B, F)
        end if
        Hmat = F%LC(:, :) .tm.F%H(:, :) .m.F%LC(:, :)
        do j = 1, size(pairs)
          gradient(j, k) = Hmat(pairs(j)%mo1, pairs(j)%mo2)
        end do
      end do
      hessian(i, :) = (gradient(:, 2) - gradient(:, 1))/(angle*TWO)
    end do
    ! symmetriz hessian
    do i = 1, size(pairs, 1)
      do j = i + 1, size(pairs, 1)
        temp = HALF*(hessian(i, j) + hessian(j, i))
        hessian(i, j) = temp
        hessian(j, i) = temp
      end do
    end do
    call get_eigenvalues(hessian, hessian_ev)
    write (P%OUTPUT, '("# ORBITAL ROTATION HESSIAN EIGENVALUES",10f10.3)') hessian_ev
    k = minloc(abs(hessian_ev(:)), 1)
    write (P%OUTPUT, '("# SMALLEST ORBITAL ROTATION HESSIAN EIGENVALUES",es15.7)') hessian_ev(k)
    do k = 1, size(pairs, 1)
      if (abs(hessian_ev(k)) .gt. 0.1) cycle
      write (P%OUTPUT, '("# ORBITAL ROTATION HESSIAN EIGENVECTOR FOR EIGENVALUE",f15.5)') hessian_ev(k)
      do i = 1, size(pairs, 1)
        if (abs(hessian(i, k)) .lt. 0.01) cycle
        write (P%OUTPUT, '("#",x,i4,i4,x)', advance="no") pairs(i)%MO1, pairs(i)%MO2
        write (P%OUTPUT, '("#",x,f10.3)') hessian(i, k)
      end do
    end do
    deallocate (hessian)
    deallocate (hessian_ev)
    deallocate (gradient)
    deallocate (theta)
    deallocate (pairs)
  end subroutine stability_analysis

  subroutine print_crystalline_data(G, F, OUTPUT)
    type(Grid3D), intent(in) :: G
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    integer :: i
    integer :: Z_sum      ! total nuclear charge
    real(long) :: Q_int   ! total elec. charge calculated from density
    real(long) :: Q_sum   ! total elec. charge calculated from point charges
    Z_sum = nint(sum(G%Z_A(:)))
    Q_int = integral(G, F%rho(:))
    Q_sum = sum(F%Q_A(:))
    write (OUTPUT, "('# Crystalline calculation:')")
    write (OUTPUT, "('#   - N of atoms in the unitcell         = ',i10)") F%N_nuc
    write (OUTPUT, "('#   - sum of nuclear charges (Z_sum)     = ',i10)") Z_sum
    write (OUTPUT, "('#   - int of electron densities          = ',f17.6)") Q_int
    write (OUTPUT, "('#   - sum of elec. point charges (Q_sum) = ',f17.6)") Q_sum
    write (OUTPUT, "('#   - quasi-neutrality (Z_sum - Q_sum)   = ',es14.6)") &
      dble(Z_sum) - Q_sum
    write (OUTPUT, "('#   - unitcell volume calc. by int       = ',es14.6)") &
      integral(G, (/(ONE, i=1, G%N_grid)/))
    write (OUTPUT, "('#   - unitcell volume by lattice param.  = ',es14.6)") &
      G%unitcell_x*G%unitcell_y*G%unitcell_z
    write (OUTPUT, "('#')")
  end subroutine print_crystalline_data

  subroutine print_dipole_moments(P, G, A, B, F, OUTPUT)
    type(Param), intent(in) :: P
    type(Grid3D), intent(in) :: G
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    integer :: i_nuc, j_nuc, k
    real(long) :: d_x, d_y, d_z
    real(long) :: Z_eff, QN_rc, R_AB, Z_eff_rc
    real(long), allocatable :: tdipole_bf(:,:,:)
    real(long) :: traceDS

    if(P%yes_gaussian)then
      allocate(tdipole_bf(size(F%D, 1), size(F%D, 1), 3))
      call calc_trans_dipole_moment_bf(P, B, G, tdipole_bf)
      traceDS = matTrace(F%D(:,:).m.B%S(:,:))
      !
      ! we have to correct for the fact that dipole moment integrals are calculated with respect to the center
      ! given in B%env(2:4)
      !
      d_x = sum(A(:)%fcharge * A(:)%x) - sum(tdipole_bf(:,:,1) * F%D(:,:)) - B%env(2) * traceDS
      d_y = sum(A(:)%fcharge * A(:)%y) - sum(tdipole_bf(:,:,2) * F%D(:,:)) - B%env(3) * traceDS
      d_z = sum(A(:)%fcharge * A(:)%z) - sum(tdipole_bf(:,:,3) * F%D(:,:)) - B%env(4) * traceDS
      deallocate(tdipole_bf)
    else
      !
      ! print partial charges and calculate the dipole based on point charges
      !
      write (OUTPUT, "('# Partial charges based on atomic grids:')")
      write (OUTPUT, "('#   ',5x,2a10,a14)", advance="no") &
          "Z_nuc", "Q_nuc", "Eff. charge"
      if (G%yes_cutoff) write (OUTPUT, "(a14)", advance="no") "Z_eff(r<r_c)"
      write (OUTPUT, "()")
      d_x = ZERO; d_y = ZERO; d_z = ZERO
      QN_rc = ZERO
      do i_nuc = 1, F%N_nuc
        Z_eff = G%Z_A(i_nuc) - F%Q_A(i_nuc)
        write (OUTPUT, "('#   ',a5,2f10.3,es14.5)", advance="no") &
            A(i_nuc)%label, A(i_nuc)%fcharge, F%Q_A(i_nuc), Z_eff
        d_x = d_x + Z_eff*G%XX(i_nuc)
        d_y = d_y + Z_eff*G%YY(i_nuc)
        d_z = d_z + Z_eff*G%ZZ(i_nuc)
        if (G%yes_cutoff) then
          Z_eff_rc = ZERO
          loop_B: do j_nuc = 1, F%N_nuc
            loop_tr: do k = 1, G%N_tr
              R_AB = bond_distance(G, i_nuc, j_nuc, k)  ! <-- good scanning order
              if (R_AB .le. G%r_cutoff) cycle loop_tr   ! skip the short-range region
              Z_eff_rc = Z_eff_rc + G%Z_A(j_nuc) - F%Q_A(j_nuc)
            end do loop_tr
          end do loop_B
          write (OUTPUT, "(es14.5)", advance="no") Z_eff_rc
          QN_rc = QN_rc + Z_eff_rc
        end if
        write(OUTPUT, "()")
      end do
      if ( G%yes_cutoff ) then              ! effective charge within r_cutoff
        QN_rc = QN_rc / dble(F%N_nuc)       ! quasi-neutrality per atom
        write(OUTPUT, "('# Quasi-neutrality per atom within r_c = ',3x,es13.5)") &
            QN_rc
      end if
      write(OUTPUT, "('#')")
      !
      ! print dipole moments
      !
      write(OUTPUT, "('# Dipole moment (a.u.):',5x,4a13)") "x","y","z", "magnitude"
      write(OUTPUT, "('# - based on point charges: ',4es13.5)") &
          d_x, d_y, d_z, sqrt( d_x**2 + d_y**2 + d_z**2 )
      !
      ! calculate dipole moments based on electron density
      !
      d_x = sum( G%Z_A(:) * G%XX(:) ) - integral( G, G%x(:) * F%rho(:) )
      d_y = sum( G%Z_A(:) * G%YY(:) ) - integral( G, G%y(:) * F%rho(:) )
      d_z = sum( G%Z_A(:) * G%ZZ(:) ) - integral( G, G%z(:) * F%rho(:) )
      write(OUTPUT, "('# - based on elec. density: ',4es13.5)") &
      d_x, d_y, d_z, sqrt( d_x**2 + d_y**2 + d_z**2 )
    end if
    write(OUTPUT, "('# |DIPOLE| ',17x,3es13.5)") d_x, d_y, d_z
    write(OUTPUT, "('# |DIPOLE_DEBEYE| ',10x,3es13.5)") d_x*AU2DEB, d_y*AU2DEB, d_z*AU2DEB
    write(OUTPUT, "('# with respect to (xyz) ',4x,3es13.5)") ZERO,ZERO,ZERO
    write(OUTPUT, "('#')")
  end subroutine print_dipole_moments

end module SCF
