!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module secondorder_scf
  !! author: Ludger Inhester
  !!
  !! second-order SCF module
  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use eigen_lapack
  use functional_xc
  use operations
  use averaging
  use matrix

  implicit none
  type :: indep_mo_pair
    !! pair of molecular orbitals
    integer :: mo1
    integer :: mo2
  end type indep_mo_pair
  private
  public :: indep_mo_pair, prepare_independent_pairs, rotate_orbitals, BFGS_Hessian, make_soscf_step
contains
  subroutine prepare_independent_pairs(P, F, pairs, irrep)
    !! fills the structure indep_mo_pairs containing pairs of mos
    !! forming independent SCF optimization parameters
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    type(indep_mo_pair), allocatable, intent(out) :: pairs(:)
    integer, intent(in), optional :: irrep(:)
    integer ::i, j, n
    n = 0
    do i = 1, size(F%N_occ, 1)
      do j = i + 1, size(F%N_occ, 1)
        if (allocated(F%frozen_MO)) then
          if (any(F%frozen_MO(:) .eq. i)) cycle
          if (any(F%frozen_MO(:) .eq. j)) cycle
        end if
        if (F%N_occ(i) .ne. F%N_occ(j) .and. .not. is_equivalent(P, F, i, j)) then
          if (present(irrep)) then
            if (irrep(i) .ne. irrep(j)) cycle
          end if
          n = n + 1
        end if
      end do
    end do
    allocate (pairs(n))
    n = 0
    do i = 1, size(F%N_occ, 1)
      do j = i + 1, size(F%N_occ, 1)
        if (allocated(F%frozen_MO)) then
          if (any(F%frozen_MO(:) .eq. i)) cycle
          if (any(F%frozen_MO(:) .eq. j)) cycle
        end if
        if (F%N_occ(i) .ne. F%N_occ(j) .and. .not. is_equivalent(P, F, i, j)) then
          if (present(irrep)) then
            if (irrep(i) .ne. irrep(j)) cycle
          end if
          n = n + 1
          pairs(n)%mo1 = i
          pairs(n)%mo2 = j
        end if
      end do
    end do
  end subroutine prepare_independent_pairs

  subroutine rotate_orbitals(theta, pairs, C)
    !! rotates orbital pairs by given angle theta    
    real(long), intent(in) :: theta(:)
    type(indep_mo_pair), intent(in) :: pairs(:)
    real(long), intent(out) :: C(:, :)
    real(long) :: costheta, sintheta, temp
    integer :: i, pair, mo1, mo2
    do pair = 1, size(pairs, 1)
      mo1 = pairs(pair)%mo1
      mo2 = pairs(pair)%mo2
      costheta = cos(theta(pair))
      sintheta = sin(theta(pair))
      do i = 1, size(C, 1)
        temp = C(i, mo1)*costheta + C(i, mo2)*sintheta
        C(i, mo2) = -C(i, mo1)*sintheta + C(i, mo2)*costheta
        C(i, mo1) = temp
      end do
    end do
  end subroutine rotate_orbitals

!>  Broyden-Fletcher-Goldfarb-Shanno (BFGS) method to approximate the inverse hessian
!!  by previous step and this step gradient and previous step gradient, and previous estimated hessian
  subroutine BFGS_Hessian(old_step, gradient, old_gradient, old_inv_hessian, inv_hessian)
    real(long), intent(in) :: old_step(:)
    real(long), intent(in) :: gradient(:)
    real(long), intent(in) :: old_gradient(:)
    real(long), intent(in) :: old_inv_hessian(:, :)
    real(long), intent(out) :: inv_hessian(size(old_inv_hessian, 1), size(old_inv_hessian, 1))

    real(long) :: alpha
    real(long) :: v(size(old_step, 1))
    real(long) :: d(size(old_step, 1))
    real(long) :: dv
    integer :: i, j

    d = gradient(:) - old_gradient(:)
    alpha = ONE/sum(old_step(:)*d(:))
    do i = 1, size(old_step, 1)
      v(i) = sum(old_inv_hessian(i, :)*d(:))
    end do
    inv_hessian = old_inv_hessian

    dv = sum(d(:)*v(:))
    do i = 1, size(old_step, 1)
      do j = 1, size(old_step, 1)
        inv_hessian(i, j) = inv_hessian(i, j) + &
                            +(ONE + alpha*dv)*alpha*old_step(i)*old_step(j) &
                            - alpha*(old_step(i)*v(j) + old_step(j)*v(i))
      end do
    end do

  end subroutine BFGS_hessian

  subroutine make_soscf_step(H, LC, pairs, theta, gradient, inv_hessian)
    real(long), intent(in) :: H(:,:)  !! fock matrix or hamiltoninan
    real(long), intent(inout) :: LC(:,:) !! mo coefficients
    type(indep_mo_pair), intent(in)  :: pairs(:) !! pair structure
    real(long), intent(inout) :: theta(:) !! rotation angles
    real(long), intent(inout) :: gradient(:) !! gradients
    real(long), intent(inout) :: inv_hessian(:,:) !! inv_hessian
    real(long) :: old_theta(size(theta,1))
    real(long) :: old_gradient(size(gradient,1))
    real(long) :: old_inv_hessian(size(inv_hessian,1), size(inv_hessian,2))
    real(long) :: Hmat(size(LC,1), size(LC,1))
    integer :: i

    old_theta = theta
    old_gradient(:) = gradient(:)
    old_inv_hessian(:,:) = inv_hessian(:,:)
    Hmat = LC(:,:) .tm. H(:,:) .m. LC(:,:)
    ! gradient = -H(i,j)
    do i=1,size(pairs,1)
      gradient(i) = - Hmat(pairs(i)%mo1, pairs(i)%mo2)
    end do
    if(maxval(abs(old_inv_hessian)) < 1e-10_long)then
      inv_hessian(:,:) = ZERO
      do i=1,size(pairs,1)
        inv_hessian(i,i) = ONE/(Hmat(pairs(i)%mo1,pairs(i)%mo1)-Hmat(pairs(i)%mo2,pairs(i)%mo2))
      end do
    else
      ! Broyden-Fletcher-Goldfarb-Shanno  approximation for the hessian matrix
      call BFGS_Hessian(old_theta, gradient, old_gradient, old_inv_hessian, inv_hessian)
    end if
    ! do a Newton step
    theta = - inv_hessian.m.gradient
    ! rotate orbitals by angle theta
    call rotate_orbitals(theta, pairs, LC)
  end subroutine make_soscf_step

end module secondorder_scf
