!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module sph_harmonics
  !! author: Sang-Kil Son
  !!
  !! Spherical Harmonics module
  !!   written by Son, Sang-Kil in Jun. 2006
  !!   the derivatives of the real spherical harmonics have been added by Ludger Inhester Apr. 2017
  !! Notation convention:
  !!   r     : radius
  !!   theta : polar angle (latitude)
  !!   phi   : azimuthal angle (longitude)
  !!
  !! * This is the same as Arfken's book
  !!   and same as http://mathworld.wolfram.com/SphericalHarmonic.html
  !!   but different from http://mathworld.wolfram.com/SphericalCoordinates.html
  !!
  !!                            2l+1 (l-m)!
  !!   Y_l^m(theta,phi) = sqrt{ ---- ------ } P_l^m( cos theta ) exp( i m phi )
  !!                             4pi (l+m)!
  !!
  !!   Y_l^(-m)(theta,phi) = (-1)^m conjg( Y_l^m(theta,phi) )
  !!
  !! Be careful for a phase sign: many texts use a different phase convention.
  !!
  use constants
  use legendre
  implicit none
  interface sph_harmonics_complex
    !! complex spherical harmonics
    module procedure sph_harmonics_imag
  end interface
  private
  public :: sph_harmonics_real, factorial2, grad_sph_harmonic_Y_real, xyz_to_sph
contains

  !>----------------------------------------------------------------------------
  !! calculate real spherical harmonics:
  !! this is closely connected to sph_harmonics_prefactor() in molecular_grid.f90
  !!   sph_harmonics_real( l, m, theta, phi ) :=
  !!     G%Y_lm_prefactor(lm) * (2l+1) / 4pi *
  !!     associated_Legendre_P( l, |m|, cos(theta) ) * [ ... ]
  !!     [ ... ] = 1                 for m = 0
  !!             = cos( |m| * phi )  for m > 0
  !!             = sin( |m| * phi )  for m < 0
  !!----------------------------------------------------------------------------
  function sph_harmonics_real(l, m, theta, phi) result(Y)
    real(long) :: Y
    integer, intent(in) :: l, m
    real(long), intent(in) :: theta, phi
    real(long) :: p
    p = sqrt(dble(2*l + 1)/TWO/factorial2(l - abs(m) + 1, l + abs(m)))* &
        associated_Legendre_P(l, abs(m), cos(theta))
    !
    ! The real harmonics are derived from the complex functions
    ! by replacing exp( i m phi ) / sqrt( 2 PI ) with
    ! sin( |m| phi ) / sqrt(PI) or cos ( |m| phi ) / sqrt(PI)
    ! and a proper phase sign
    !
    if (m .eq. 0) then
      Y = p/sqrt(TWO*PI)
    else if (m .gt. 0) then
      Y = p/sqrt(PI)*cos(dble(m)*phi)
      !Y = p / sqrt(PI) * sin( dble( abs(m) ) * phi )
      !if ( mod(m,2) == 0 ) Y = -Y
    else
      Y = p/sqrt(PI)*sin(dble(-m)*phi)
      !Y = p / sqrt(PI) * cos( dble( abs(m) ) * phi )
      !if ( mod(m,2) /= 0 ) Y = -Y
    end if
  end function sph_harmonics_real

  !> This function computes d/dtheta Y_lm(theta,phi)
  function sph_harmonics_real_dtheta(l, m, theta, phi) result(Y)
    real(long) :: Y
    integer, intent(in) :: l, m
    real(long), intent(in) :: theta, phi
    real(long) :: dp
    real(long) :: l1, l2
    real(long) :: dl
    ! We use this formula for calulcate the derivative of the Legendre polynomials
    ! 1/sqrt(1-x^2) d/dx Pl|m|(x) = -1/(2)*((l+m)*(l-m+1)*P_l,|m|-1 (x) - P_l,|m|+1(x))
    if (m .eq. 0) then
      if (l .eq. 0) then
        l1 = ZERO
      else
        l1 = -ONE/((l + 1)*l)*associated_Legendre_P(l, +1, cos(theta))
      end if
    else
      l1=associated_Legendre_P( l, abs(m)-1, cos(theta) )
    end if  
    if(abs(m)==l)then
      l2=ZERO
    else
      l2=associated_Legendre_P( l, abs(m)+1, cos(theta) )
    end if
    dl=-HALF*((l+abs(m))*(l-abs(m)+1)*l1-l2)
    dp=sqrt( dble(2*l+1) / TWO / factorial2( l-abs(m)+1, l+abs(m) ) ) * dl

    if (m .eq. 0) then
      Y = dp/sqrt(TWO*PI)
    else if (m .gt. 0) then
      Y = dp/sqrt(PI)*cos(dble(m)*phi)
      !Y = p / sqrt(PI) * sin( dble( abs(m) ) * phi )
      !if ( mod(m,2) == 0 ) Y = -Y
    else
      Y = dp/sqrt(PI)*sin(dble(-m)*phi)
      !Y = p / sqrt(PI) * cos( dble( abs(m) ) * phi )
      !if ( mod(m,2) /= 0 ) Y = -Y
    end if
  end function sph_harmonics_real_dtheta

  !> This function computes 1/sin(theta) d/dphi Y_lm(theta,phi)
  function sph_harmonics_real_dphi_over_sintheta(l, m, theta, phi) result(Y)
    real(long) :: Y
    integer, intent(in) :: l, m
    real(long), intent(in) :: theta, phi
    real(long) :: p
    if (m .eq. 0) then
      p = ZERO
    else
      ! we use this formula for the factor sintheta * Plm(x)
      ! 1/sqrt(1-x^2) Pl|m|(x) = -1/(2|m|)*(P_l+1,|m|+1 (x) + (l-|m|+1)(l-|m|+2) P_l+1,|m|-1(x))
      p = sqrt(dble(2*l + 1)/TWO/factorial2(l - abs(m) + 1, l + abs(m)))* &
          -ONE/(TWO*abs(m))*( &
          associated_Legendre_P(l + 1, abs(m) + 1, cos(theta)) &
          + (l - abs(m) + 1)*(l - abs(m) + 2)*associated_Legendre_P(l + 1, abs(m) - 1, cos(theta)) &
          )
    end if
    !
    ! The real harmonics are derived from the complex functions
    ! by replacing exp( i m phi ) / sqrt( 2 PI ) with
    ! sin( |m| phi ) / sqrt(PI) or cos ( |m| phi ) / sqrt(PI)
    ! and a proper phase sign
    !
    if (m .eq. 0) then
      Y = ZERO
    else if (m .gt. 0) then
      Y = p/sqrt(PI)*dble(m)*(-sin(dble(m)*phi))
      !Y = p / sqrt(PI) * sin( dble( abs(m) ) * phi )
      !if ( mod(m,2) == 0 ) Y = -Y
    else
      Y = p/sqrt(PI)*dble(abs(m))*cos(dble(abs(m))*phi)
      !Y = p / sqrt(PI) * cos( dble( abs(m) ) * phi )
      !if ( mod(m,2) /= 0 ) Y = -Y
    end if
  end function sph_harmonics_real_dphi_over_sintheta

  function grad_sph_harmonic_Y_real(l, m, theta, phi) result(gY)
    !! gradient of real sph harmonic
    real(long) :: gY(3)
    integer, intent(in) :: l, m
    real(long), intent(in) :: theta, phi

    real(long) :: Y1, Y2
    real(long) :: sin_theta, cos_theta, sin_phi, cos_phi
    sin_theta = sin(theta)
    cos_theta = cos(theta)
    sin_phi = sin(phi)
    cos_phi = cos(phi)

    ! Y1 =dY/dtheta
    Y1 = sph_harmonics_real_dtheta(l, m, theta, phi)
    ! Y2 = 1/sin(theta) dY/dphi
    Y2 = sph_harmonics_real_dphi_over_sintheta(l, m, theta, phi)

    ! cartesian gradient grad Y_lm is
    !  dY/dtheta * e_theta + dY/dphi * e_phi
    gY(1) = Y1*cos_theta*cos_phi + Y2*(-sin_phi)
    gY(2) = Y1*cos_theta*sin_phi + Y2*cos_phi
    gY(3) = Y1*(-sin_theta)

  end function grad_sph_harmonic_Y_real

  function sph_harmonics_imag(l, m, theta, phi) result(Y)
    complex(long) :: Y
    integer, intent(in) :: l, m
    real(long), intent(in) :: theta, phi
    real(long) :: p
    p = sqrt(dble(2*l + 1)/TWO/factorial2(l - abs(m) + 1, l + abs(m)))* &
        associated_Legendre_P(l, abs(m), cos(theta))
    if (m .eq. 0) then
      Y = p/sqrt(TWO*PI)
    else
      Y = p/sqrt(TWO*PI)*exp(IMAG_ONE*dble(m)*phi)
      if (m .lt. 0 .and. mod(m, 2) .ne. 0) Y = -Y
    end if
  end function sph_harmonics_imag

  !>----------------------------------------------------------------------------
  !! factorial(n,m) = m! / n!      for n < m
  !!                = 1            otherwise
  !!       thus, factorial(1,2) = 2; factorial(2,1) = 1
  !!----------------------------------------------------------------------------
  pure function factorial2(n, n2) result(f)
    real(long) :: f
    integer, intent(in) :: n
    integer, intent(in), optional :: n2
    integer :: i
    f = ONE
    if (present(n2)) then
     if (n .le. n2) then
       do i = n, n2
         f = f*dble(i)
       end do
     end if
   else
     do i = 1, n   
       f = f * dble(i)
     end do
   end if
end function factorial2

  subroutine xyz_to_sph(x, y, z, r, theta, phi)
    !! translates x,y,z into r, theta, phi
    real(long), intent(in) :: x, y, z
    real(long), intent(out) :: r, theta, phi
    r = sqrt(x**2 + y**2 + z**2)    ! r > 0
    theta = acos(z/r)                 ! 0 <= theta <= pi
    phi = atan2(y, x)                 ! 0 <= phi < 2 pi
    if (phi .lt. ZERO) phi = phi + TWO*PI
  end subroutine xyz_to_sph

  subroutine sph_to_xyz(r, theta, phi, x, y, z)
    !! translates r, theta, phi into x,y,z
    real(long), intent(in) :: r, theta, phi
    real(long), intent(out) :: x, y, z
    x = r*sin(theta)*cos(phi)
    y = r*sin(theta)*sin(phi)
    z = r*cos(theta)
  end subroutine sph_to_xyz

end module sph_harmonics
