!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module spin_coupling
  !! author: Ludger Inhester
  !!
  !! This module is for the calculation of spin configurations
  !! according to genealogical coupling
  !! written by Ludger Inhester in 2016 at DESY CFEL
  !!
  use constants
  use anglib
  use matrix
  implicit none

  type :: UMatrixTable
    !!  structure to keep precompute spin coupling matrices
    integer :: NopenMax !! maximum numer of open shells
    integer :: multp !! spin multiplicity
    integer, allocatable :: f(:) !! number of spin configurations for multp for given number of open shells
    integer, allocatable :: indN(:) !! starting index for n open shells
    real(long), allocatable :: Umatrix(:) !! array containing Umarix data
    real(long), allocatable :: UCycleMatrix(:) !! array containing Umarix data
  end type UMatrixTable
  private
  public :: UMatrixTable, constructUMatrixTable, purge_UMatrixTable, getUMatrixFromTable, getUCycleMatrixFromTable, makeCycleUij, &
    one_particle_dm, s_two_particle_dm, &
    num_csf, csfstring, csfdet, detstring, spin_string, Nopen2csfs, Nopen2dets, &
    compare_det
contains

  subroutine constructUMatrixTable(NopenMax, multp, UT)
    !! make a table of coupling matrices for a set of transpositions of
    !! a set of maximal NopenMax un-paired electrons
    integer, intent(in) :: NopenMax
    integer, intent(in) :: multp
    type(UmatrixTable), intent(out) :: UT
    integer :: n, i, mu, nu
    real(long), allocatable :: U(:, :)
    UT%multp = multp
    UT%NOpenMax = NopenMax
    allocate (UT%f(0:NOpenMax))
    allocate (UT%indn(NOpenMax))
    i = 1
    if (multp.eq.1) then
      UT%f(0) = 1
    else
      UT%f(0) = 0
    end if
    do n = 1, NopenMax
      UT%indn(n) = i
      UT%f(n) = num_csf(n, multp)
      i = i + (n*(n + 1)/2)*UT%f(n)**2
      ! n x (n+1)/2 possible transpositions time f^2 matrix size
    end do
    allocate (UT%Umatrix(i))
    i = 1
    do n = 1, NopenMax
      if (UT%f(n) .eq. 0) cycle
      allocate (U(UT%f(n), UT%f(n)))
      do mu = 1, n
        do nu = 1, mu
          call Umatrix(n, multp, nu, mu, U)
          UT%Umatrix(i:i + UT%f(n)**2) = pack(U, .True.)
          i = i + UT%f(n)**2
        end do
      end do
      deallocate (U)
    end do

    allocate (UT%UCycleMatrix(i))
    i = 1
    do n = 1, NopenMax
      if (UT%f(n) .eq. 0) cycle
      allocate (U(UT%f(n), UT%f(n)))
      do mu = 1, n
        do nu = 1, mu
          call makeCycleUij(UT, mu, nu, n, U, yes_cache=.False.)
          UT%UCycleMatrix(i:i + UT%f(n)**2) = pack(U, .True.)
          i = i + UT%f(n)**2
        end do
      end do
      deallocate (U)
    end do
  end subroutine constructUMatrixTable

  subroutine purge_UMatrixTable(UT)
    !! deallocates Umatrixtable structure
    type(UmatrixTable), intent(inout) :: UT
    deallocate (UT%f)
    deallocate (UT%indN)
    deallocate (UT%Umatrix)
    deallocate (UT%UCycleMatrix)
  end subroutine purge_UMatrixTable

  pure subroutine getUMatrixFromTable(UT, n, mu, nu, U)
    !! returns the coupling matrix for translocation mu-<nu
    !! using the previously prepared table UT
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: n, mu, nu
    real(long), intent(out) :: U(:, :)
    integer :: i
    if (mu .gt. nu) then
      i = UT%indn(n) + (nu - 1 + mu*(mu - 1)/2)*UT%f(n)**2
    else
      i = UT%indn(n) + (mu - 1 + nu*(nu - 1)/2)*UT%f(n)**2
    end if
    U = reshape(UT%Umatrix(i:i + UT%f(n)**2), (/UT%f(n), UT%f(n)/))
  end subroutine getUMatrixFromTable

  pure subroutine getUCycleMatrixFromTable(UT, n, mu, nu, U)
    !! returns the coupling matrix for translocation mu-<nu
    !! using the previously prepared table UT
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: n, mu, nu
    real(long), intent(out) :: U(:, :)
    integer :: i
    if (mu .gt. nu) then
      i = UT%indn(n) + (nu - 1 + mu*(mu - 1)/2)*UT%f(n)**2
      U = reshape(UT%UCycleMatrix(i:i + UT%f(n)**2), (/UT%f(n), UT%f(n)/))
    else
      i = UT%indn(n) + (mu - 1 + nu*(nu - 1)/2)*UT%f(n)**2
      U = transpose(reshape(UT%UCycleMatrix(i:i + UT%f(n)**2), (/UT%f(n), UT%f(n)/)))
    end if

  end subroutine getUCycleMatrixFromTable

  
  pure subroutine one_particle_dm(dm, occ, dets, csf)
    !! returns the one-particle density matrix
    real(long), intent(out) :: dm(:, :)
    integer, intent(in) :: occ(:)
    integer, intent(in) :: dets(:, :)
    integer, intent(in) :: csf(:)
    integer :: only_left(1), only_left_spin(1)
    integer :: only_right(1), only_right_spin(1)
    integer :: i, j
    integer :: coef_sign
    dm(:, :) = ZERO
    do i = 1, size(occ)
      if (occ(i) .eq. 2) dm(i, i) = TWO
      if (occ(i) .eq. 1) dm(i, i) = ONE
    end do
    do i = 1, size(dets, 2)
      do j = i + 1, size(dets, 2)
        !        write(STDERR,"(a,2x,a)") trim(detstring(occ,dets(:,i))),trim(detstring(occ,dets(:,j)))
        call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, occ, dets(:, i), occ, dets(:, j))
        if (coef_sign .eq. 0) cycle
        if (count(only_left_spin(:) .ne. 0) .ne. 1 .or. count(only_right_spin(:) .ne. 0) .ne. 1) cycle
        if (only_left_spin(1) .ne. only_right_spin(1)) cycle
        !        write(STDERR,*) only_left(1),only_right(1)
        dm(only_left(1), only_right(1)) = &
          dm(only_left(1), only_right(1)) + coef_sign*csfdet(csf(:), dets(:, i))*csfdet(csf(:), dets(:, j))
        dm(only_right(1), only_left(1)) = &
          dm(only_right(1), only_left(1)) + coef_sign*csfdet(csf(:), dets(:, i))*csfdet(csf(:), dets(:, j))
      end do

    end do
  end subroutine one_particle_dm

  subroutine s_two_particle_dm(s_tpdm, occ, dets, csf)
    !! symmetrized two particle density matrix
    real(long), intent(out) :: s_tpdm(:)
    integer, intent(in) :: occ(:)
    integer, intent(in) :: dets(:, :)
    integer, intent(in) :: csf(:)
    real(long), allocatable :: tpdm(:, :, :, :)
    integer :: p, q, r, s
    integer :: smax
    integer :: t

    allocate (tpdm(size(occ, 1), size(occ, 1), size(occ, 1), size(occ, 1)))
    call two_particle_dm(tpdm, occ, dets, csf)
    t = 1
    do p = 1, size(tpdm, 1)
      do q = 1, p
        do r = 1, p
          if (r .eq. p) then
            smax = q
          else
            smax = r
          end if
          do s = 1, smax
            s_tpdm(t) = (tpdm(p, q, r, s) + tpdm(q, p, r, s) + tpdm(p, q, s, r) + tpdm(q, p, s, r))/FOUR
            t = t + 1
          end do
        end do
      end do
    end do
    deallocate (tpdm)
  end subroutine s_two_particle_dm

  pure subroutine two_particle_dm(tpdm, occ, dets, csf)
    ! This has to be tested further
    real(long), intent(out) :: tpdm(:, :, :, :)
    integer, intent(in) :: occ(:)
    integer, intent(in) :: dets(:, :)
    integer, intent(in) :: csf(:)
    integer :: only_left(2), only_left_spin(2)
    integer :: only_right(2), only_right_spin(2)
    integer :: i, j, mo1, mo2, oi
    integer :: coef_sign
    integer :: moa, mob, moc, mod
    integer :: spin_of_mo(size(occ, 1))

    real(long) :: coef1, coef2
    tpdm(:, :, :, :) = ZERO

    do i = 1, size(dets, 2)
      !      write(STDERR,"(a)") trim(detstring(occ,dets(:,i)))
      coef1 = csfdet(csf(:), dets(:, i))

      oi = 0
      do mo1 = 1, size(occ(:), 1)
        if (occ(mo1) .eq. 1) then
          oi = oi + 1
          spin_of_mo(mo1) = open_spin(dets(:, i), oi)
        else
          spin_of_mo(mo1) = 0
        end if
      end do

      do mo1 = 1, size(occ(:), 1)
        if (occ(mo1) .eq. 2) &
          tpdm(mo1, mo1, mo1, mo1) = tpdm(mo1, mo1, mo1, mo1) + TWO*coef1**2
        if (occ(mo1) .eq. 1) &
          tpdm(mo1, mo1, mo1, mo1) = tpdm(mo1, mo1, mo1, mo1) + ZERO*coef1**2
        do mo2 = mo1 + 1, size(occ(:), 1)
          if (occ(mo1) .eq. 2 .and. occ(mo2) .eq. 2) then
            tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + FOUR*coef1**2
            tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + FOUR*coef1**2
            tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - TWO*coef1**2
            tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - TWO*coef1**2
          else if (occ(mo1) .eq. 1 .and. occ(mo2) .eq. 2) then
            tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + TWO*coef1**2
            tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + TWO*coef1**2
            tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef1**2
            tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef1**2
          else if (occ(mo1) .eq. 2 .and. occ(mo2) .eq. 1) then
            tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + TWO*coef1**2
            tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + TWO*coef1**2
            tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef1**2
            tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef1**2
          else if (occ(mo1) .eq. 1 .and. occ(mo2) .eq. 1) then
            tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + ONE*coef1**2
            tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + ONE*coef1**2
            if (spin_of_mo(mo1) .eq. spin_of_mo(mo2)) then
              tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef1**2
              tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef1**2
            end if
          end if
        end do
      end do

      ! loop over all determinant pairs
      do j = i + 1, size(dets, 2)
        coef2 = csfdet(csf(:), dets(:, j))
        !        write(STDERR,"(a,2x,a)") trim(detstring(occ,dets(:,i))),trim(detstring(occ,dets(:,j)))
        call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, occ, dets(:, i), occ, dets(:, j))
        if (coef_sign .eq. 0) cycle
        if (count(only_left(:) .ne. 0) .eq. 1 .and. count(only_right(:) .ne. 0) .eq. 1 .and. only_left_spin(1) .eq. only_right_spin(1)) then
          ! determinant differ by one electron: this may not happen with simple spin configurations
          moa = only_left(1)
          mob = only_right(1)
          do mo1 = 1, size(occ(:), 1)
            if (mo1 .eq. moa .or. mo1 .eq. mob) cycle
            if (occ(mo1) .eq. 2) then
              tpdm(mo1, mo1, moa, mob) = tpdm(mo1, mo1, moa, mob) + coef_sign*coef1*coef2*TWO
              tpdm(moa, mob, mo1, mo1) = tpdm(moa, mob, mo1, mo1) + coef_sign*coef1*coef2*TWO
              tpdm(mo1, mob, moa, mo1) = tpdm(mo1, mob, moa, mo1) - coef_sign*coef1*coef2*ONE
              tpdm(moa, mo1, mo1, mob) = tpdm(moa, mo1, mo1, mob) - coef_sign*coef1*coef2*ONE
            else if (occ(mo1) .eq. 1) then
              tpdm(mo1, mo1, moa, mob) = tpdm(mo1, mo1, moa, mob) + coef_sign*coef1*coef2*occ(mo1)
              tpdm(moa, mob, mo1, mo1) = tpdm(moa, mob, mo1, mo1) + coef_sign*coef1*coef2*occ(mo1)
              if (spin_of_mo(mo1) .eq. only_right_spin(1)) then
                tpdm(mo1, mob, moa, mo1) = tpdm(mo1, mob, moa, mo1) - coef_sign*coef1*coef2*occ(mo1)
                tpdm(moa, mo1, mo1, mob) = tpdm(moa, mo1, mo1, mob) - coef_sign*coef1*coef2*occ(mo1)
              end if
            end if
          end do
        else if (count(only_left(:) .ne. 0) .eq. 2 &
          .and. count(only_right(:) .ne. 0 ) .eq. 2 &
          .and. only_left_spin(1) .eq. only_right_spin(1) &
          .and. only_left_spin(2) .eq. only_right_spin(2)) then
          ! determinant differ by two electrons: <ab|ab>
          moa = only_left(1)
          mob = only_left(2)
          moc = only_right(1)
          mod = only_right(2)
          tpdm(moa, moc, mob, mod) = tpdm(moa, moc, mob, mod) + coef_sign*coef1*coef2
          tpdm(mob, mod, moa, moc) = tpdm(mob, mod, moa, moc) + coef_sign*coef1*coef2
          tpdm(moc, moa, mod, mob) = tpdm(moc, moa, mod, mob) + coef_sign*coef1*coef2
          tpdm(mod, mob, moc, moa) = tpdm(mod, mob, moc, moa) + coef_sign*coef1*coef2
          if (only_left_spin(1) .eq. only_left_spin(2)) then
            tpdm(moa, mod, mob, moc) = tpdm(moa, mod, mob, moc) - coef_sign*coef1*coef2
            tpdm(mob, moc, moa, mod) = tpdm(mob, moc, moa, mod) - coef_sign*coef1*coef2
            tpdm(mod, moa, moc, mob) = tpdm(mod, moa, moc, mob) - coef_sign*coef1*coef2
            tpdm(moc, mob, mod, moa) = tpdm(moc, mob, mod, moa) - coef_sign*coef1*coef2
          end if
        else if (count(only_left_spin(:) .ne. 0) .eq. 2 &
          .and. count(only_right_spin(:) .ne. 0) .eq. 2 &
          .and. only_left_spin(2) .eq. only_right_spin(1) &
          .and. only_left_spin(1) .eq. only_right_spin(2)) then
          ! determinant differ by two electrons: <ab|ba>
          moa = only_left(1)
          mob = only_left(2)
          moc = only_right(1)
          mod = only_right(2)
          tpdm(moa, mod, mob, moc) = tpdm(moa, mod, mob, moc) - coef_sign*coef1*coef2
          tpdm(mob, moc, moa, mod) = tpdm(mob, moc, moa, mod) - coef_sign*coef1*coef2
          tpdm(mod, moa, moc, mob) = tpdm(mod, moa, moc, mob) - coef_sign*coef1*coef2
          tpdm(moc, mob, mod, moa) = tpdm(moc, mob, mod, moa) - coef_sign*coef1*coef2
          if (only_left_spin(1) .eq. only_left_spin(2)) then
            tpdm(moa, moc, mob, mod) = tpdm(moa, moc, mob, mod) + coef_sign*coef1*coef2
            tpdm(mob, mod, moa, moc) = tpdm(mob, mod, moa, moc) + coef_sign*coef1*coef2
            tpdm(moc, moa, mod, mob) = tpdm(moc, moa, mod, mob) + coef_sign*coef1*coef2
            tpdm(mod, mob, moc, moa) = tpdm(mod, mob, moc, moa) + coef_sign*coef1*coef2
          end if

        end if
      end do
    end do

  contains
    pure function open_spin(det, i)
      integer, intent(in) :: det(:)
      integer, intent(in) :: i
      integer :: open_spin
      if (i .eq. 1) then
        open_spin = det(i)
      else
        open_spin = det(i) - det(i - 1)
      end if
    end function open_spin

  end subroutine two_particle_dm


  pure subroutine Nopen2dets(Nopen, multp, dets)
    !! routine calculates the genealogical pathes for
    !! all high spin determinants for given multiplicity multp
    !! with Nopen open shells.
    !! pathes are returned in arrays dets(1:Nopen,:)
    !! example pathes for multp=2 and Nopen=3
    !! 1,2,1 means spin  up up down
    !! 1,0,1 means spin  up down up
    !!-1,0,1 means spin  down up up  
    integer, intent(in) :: Nopen
    integer, intent(in) :: multp
    integer, intent(out), allocatable :: dets(:, :)
    integer :: Ndet, Ndets
    integer :: two_S
    integer :: i, j, k, l
    two_S = (multp - 1)
    if (Nopen .eq. 0 .and. two_S .eq. 0) then
      ! trivial case: one closed shell determinant
      Ndets = 1
      allocate (dets(Nopen, Ndets))
    else
      ! count the number of possible determinants
      Ndet = 1
      do j = 1, Nopen
        Ndet = Ndet*2
      end do
      ! l=0
      ! do i=1,Ndet
      !   k=(i-1)
      !   sum=0
      !   do j=1,Nopen
      !     sum=sum+ (1-2*mod(k,2))
      !     k=(k-mod(k,2))/2
      !   end do
      !   if(two_S==sum ) l=l+1
      ! end do
      ! Ndets=l
      ! if(l/=nint(binom(Nopen,(Nopen-two_S)/2 )))then
      !   write(STDERR,*) 'Nopen',Nopen,'multp',multp
      !   write(STDERR,*) l,nint(binom(Nopen,(Nopen-two_S)/2 ))
      !   stop
      ! end if
      if ((Nopen + two_S)/2 .gt. Nopen) then
        Ndets = 0
      else
        Ndets = nint(binom(Nopen, (Nopen + two_S)/2))
      end if
      ! allocate and fill the determinants
      allocate (dets(Nopen, Ndets))
      l = 1
      do i = 1, Ndet
        if (l .gt. Ndets) exit
        !k=(i-1)
        k = i
        do j = 1, Nopen
          if (k .le. 2**(Nopen - j)) then
            dets(j, l) = 1
          else
            dets(j, l) = -1
            k = k - 2**(Nopen - j)
          end if
          !dets(j,l)=1-2*mod(k,2)
          !k=(k-mod(k,2))/2
        end do
        do j = 2, Nopen
          dets(j, l) = dets(j, l) + dets(j - 1, l)
        end do
        if (dets(Nopen, l) .eq. two_S) then
          l = l + 1
        else
          dets(:, l) = 0
        end if
      end do
    end if
  end subroutine Nopen2dets


  pure subroutine Nopen2csfs(Nopen, multp, csfs)
    !! routine calculates the genealogical pathes for
    !! all configuration state functions for given multiplicity multp
    !! with Nopen open shells.
    !! pathes are returned in arrays csfs(1:Nopen, :)
    !! example pathes for multp=2 and Nopen=3
    !! there are two spin configurations:
    !! 1,2,1 means spin configuration '//\'
    !! 1,0,1 means spin configuration '/\/'  
    integer, intent(in) :: Nopen
    integer, intent(in) :: multp
    integer, intent(out), allocatable :: csfs(:, :)
    integer :: i, j, k, l
    integer :: Ncsfs
    integer :: Ndet
    integer :: two_S

    two_S = (multp - 1)
    if (Nopen .eq. 0 .and. two_S .eq. 0) then
      ! trivial case: closed shell
      Ncsfs = 1
      allocate (csfs(Nopen, Ncsfs))
    else
      ! Ndeterminants
      Ndet = 1
      do j = 1, Nopen
        Ndet = Ndet*2
      end do
      Ndet = Ndet/2
      Ncsfs = num_csf(Nopen, two_S + 1)
      ! fill csfs
      allocate (csfs(Nopen, Ncsfs))
      l = 1
      iloop2: do i = 1, Ndet
        if (l .gt. Ncsfs) exit
        !        k=(i-1)*2  ! the first entry is always 1
        k = i
        do j = 1, Nopen
          if (k .le. 2**(Nopen - j)) then
            csfs(j, l) = 1
          else
            csfs(j, l) = -1
            k = k - 2**(Nopen - j)
          end if
          !          csfs(j,l)=1-2*mod(k,2)
          !          k=(k-mod(k,2))/2
        end do
        do j = 2, Nopen
          csfs(j, l) = csfs(j, l) + csfs(j - 1, l)
          if (csfs(j, l) .lt. 0) cycle iloop2
        end do
        if (csfs(Nopen, l) .eq. two_S) then
          l = l + 1
        else
          csfs(:, l) = 0
        end if
      end do iloop2
    end if

    !do l=1,size(csfs,2)
    !  write(6,'(i,x,99i1)') l,csfs(:,l)
    !end do
    !stop
  end subroutine Nopen2csfs

  pure function num_csf(Nopen, multp)
    !! returns the number of possible spin configurations
    integer :: Num_csf !! number of configurations
    integer, intent(in) :: Nopen !! number of open shells
    integer, intent(in) :: multp !! spin multiplicity
    integer :: two_S
    if (multp .lt. 1) then
      Num_csf = 0
      return
    end if

    two_S = (multp - 1)
    if (mod(Nopen + 1, 2) .ne. mod(multp, 2)) then
      Num_csf = 0
      return
    end if

    if (Nopen .eq. 0 .and. two_S .eq. 0) then
      Num_csf = 1
    else
      Num_csf = nint((two_S + ONE)/((Nopen + two_S)*HALF + ONE)*binom(Nopen, (Nopen - two_S)/2))
    end if
  end function num_csf

!> returns the coefficient for determinant specified by determinant path P
!! in csf specified by path T
  real(long) pure function csfdet(T, P)
    integer, intent(in) :: T(:) ! specifies path for csf
    integer, intent(in) :: P(:) ! specifies path for det
    integer :: N
    integer :: i
    N = size(T, 1)
    CSFDET = 1.0
    if (N .eq. 0) then
      CSFDET = 1.
    else
      CSFDET = CSFDET*coupling(T(1), P(1), 1, P(1))
      do i = 2, N
        CSFDET = CSFDET*coupling(T(i), P(i), T(i) - T(i - 1), P(i) - P(i - 1))
      end do
    end if
  contains
    real(long) pure function coupling(two_S, two_M, two_t, two_p)
      integer, intent(in) :: two_S, two_M, two_t, two_p
      coupling = ONE
      if (two_t .eq. 1) then
        if (two_S + two_p*two_M .lt. 0) then
          coupling = ZERO
        else
          coupling = sqrt(dble(two_S + two_p*two_M)/(TWO*two_S))
        end if
      else if (two_t .eq. -1) then
        if (two_S + TWO - two_p*two_M .lt. 0) then
          coupling = ZERO
        else
          coupling = -two_p*sqrt(dble(two_S + TWO - two_p*two_M)/(TWO*(two_S + TWO)))
        end if
      end if
    end function coupling
  end function csfdet

!> returns a determinant given in genealogical path as alpha and beta strings plus sign
  pure subroutine spin_string(occ, det, spin_string_a, spin_string_b, sign)
    integer, intent(in) :: occ(:)
    integer, intent(in) :: det(:)
    integer, intent(out) :: spin_string_a(:)
    integer, intent(out) :: spin_string_b(:)
    integer, intent(inout) :: sign

    integer :: i, iopen
    integer :: ja, jb
    integer :: flips
    flips = 0
    iopen = 1
    ja = 1
    jb = 1
    do i = 1, size(occ)
      if (occ(i) .eq. 2) then
        spin_string_a(ja) = i
        ja = ja + 1
        flips = flips + (jb - 1)
        spin_string_b(jb) = i
        jb = jb + 1
      else if (occ(i) .eq. 1) then
        if (iopen .eq. 1) then
          if (det(1) .eq. 1) then
            spin_string_a(ja) = i
            flips = flips + (jb - 1)
            ja = ja + 1
          else if (det(1) .eq. -1) then
            spin_string_b(jb) = i
            jb = jb + 1
          end if
        else
          if (det(iopen) - det(iopen - 1) .eq. 1) then
            spin_string_a(ja) = i
            flips = flips + (jb - 1)
            ja = ja + 1
          else if (det(iopen) - det(iopen - 1) .eq. -1) then
            spin_string_b(jb) = i
            jb = jb + 1
          end if! if det(iopen)-det(iopen-1)
        end if! if iopen==
        iopen = iopen + 1
      end if! if (occ==)
    end do! i
    if (mod(flips, 2) .ne. 0) then
      sign = -1*sign
    end if
  end subroutine spin_string

  pure function detstring(occ, det) result(string)
    !! returns a readable string for printing out determinant specified by
    !! occupation and determinant path det
    integer, intent(in) :: occ(:)
    integer, intent(in) :: det(:)
    character(len=100) :: string
    integer :: i, j
    i = 1
    string = ""
    do j = 1, size(occ, 1)
      if (occ(j) .eq. 2) then
        string = trim(string)//"2"
      else if (occ(j) .eq. 0) then
        string = trim(string)//"0"
      else if (occ(j) .eq. 1) then
        if (i .eq. 1) then
          if (det(i) .eq. 1) then
            string = trim(string)//"a"
          else if (det(1) .eq. -1) then
            string = trim(string)//"b"
          end if
        else if (det(i) - det(i - 1) .eq. 1) then
          string = trim(string)//"a"
        else if (det(i) - det(i - 1) .eq. -1) then
          string = trim(string)//"b"
        end if
        i = i + 1
      end if
    end do
  end function detstring


  pure subroutine compare_det(coef_sign,only_left,only_left_spin,only_right,only_right_spin,left_occ,left_det,right_occ,right_det)
    !! compares two determinants specified by determinant path left_det(1:Nopen) right_det(1:Nopen)
    !! and occupation left_occ(:) and right_occ(:), respectively.
    !! only_left(1:M) and only_right(1:M) returns the orbital indices of electrons that are
    !! only present in left and right determinant
    !! only_left_spin(1:M) and only_right_spin(1:M) the respective spin of the electrons
    !! coef_sign returns the sign produced by flips to bring the determinants into
    !! best congruence. The array length M can be arbitrarily.
    !! When the number of electrons
    !! only present in one determinant exceeds M comparison is quit and coef_sign=0 is returned
    integer, intent(out) :: coef_sign
    integer, intent(out) :: only_left(:)
    integer, intent(out) :: only_right(:)
    integer, intent(out) :: only_left_spin(:)
    integer, intent(out) :: only_right_spin(:)

    integer, intent(in) :: left_det(:)
    integer, intent(in) :: right_det(:)
    integer, intent(in) :: left_occ(:)
    integer, intent(in) :: right_occ(:)

    integer :: i, onlyl, onlyr, open_li, open_ri
    integer :: open_spin_left, open_spin_right
    integer :: count
    coef_sign = 1
    only_left(:) = 0
    only_right(:) = 0
    only_left_spin(:) = 0
    only_right_spin(:) = 0
    if (size(left_occ, 1) .ne. size(right_occ, 1)) then
      coef_sign = 0
      return
      !    stop 'left_occ and right_occ should have same dimension'
    end if

    open_li = 0 ! index for left_det(:)
    open_ri = 0 ! index for right_det(:)
    onlyr = 0 ! index for only_right(:)
    onlyl = 0 ! index for only_left(:)
    count = 0 ! counter of electrons that are present in both determinants

    ! The determinants are in order 1a 1b 2a 2b .... na mb , with n<m
    orbital: do i = 1, size(left_occ, 1)
      if (left_occ(i) .eq. 2 .and. right_occ(i) .eq. 2) then
        ! left and right doubly occupied
        count = count + 2
      else if (left_occ(i) .eq. 0 .and. right_occ(i) .eq. 0) then
        ! left and right empty
      else if (left_occ(i) .eq. 1 .and. right_occ(i) .eq. 1) then
        ! left and right singly occupied
        open_li = open_li + 1
        open_spin_left = open_spin(left_det, open_li)
        open_ri = open_ri + 1
        open_spin_right = open_spin(right_det, open_ri)
        if (open_spin_left .eq. open_spin_right) then
          ! same spin
          count = count + 1
        else
          ! different spin
          onlyl = onlyl + 1
          if (onlyl .gt. size(only_left)) then
            coef_sign = 0
            exit orbital
          end if
          only_left(onlyl) = i
          only_left_spin(onlyl) = open_spin_left
          onlyr = onlyr + 1
          if (onlyr .gt. size(only_right)) then
            coef_sign = 0
            exit orbital
          end if
          only_right(onlyr) = i
          only_right_spin(onlyr) = open_spin_right
        end if
      else if (left_occ(i) .eq. 2 .and. right_occ(i) .eq. 0) then
        ! left double occupied right empty
        onlyl = onlyl + 1
        if (onlyl .gt. size(only_left)) then
          coef_sign = 0
          exit orbital
        end if
        only_left(onlyl) = i
        only_left_spin(onlyl) = 1
        onlyl = onlyl + 1
        if (onlyl .gt. size(only_left)) then
          coef_sign = 0
          exit orbital
        end if
        only_left(onlyl) = i
        only_left_spin(onlyl) = -1
      else if (left_occ(i) .eq. 0 .and. right_occ(i) .eq. 2) then
        ! left empty right double occupied
        onlyr = onlyr + 1
        if (onlyr .gt. size(only_right)) then
          coef_sign = 0
          exit orbital
        end if
        only_right(onlyr) = i
        only_right_spin(onlyr) = 1
        onlyr = onlyr + 1
        if (onlyr .gt. size(only_right)) then
          coef_sign = 0
          exit orbital
        end if
        only_right(onlyr) = i
        only_right_spin(onlyr) = -1
      else if (left_occ(i) .eq. 2 .and. right_occ(i) .eq. 1) then
        ! left double right singly occupied
        open_ri = open_ri + 1
        open_spin_right = open_spin(right_det, open_ri)
        onlyl = onlyl + 1
        if (onlyl .gt. size(only_left)) then
          coef_sign = 0
          exit orbital
        end if
        only_left(onlyl) = i
        only_left_spin(onlyl) = -open_spin_right
        if (only_left_spin(onlyl) .eq. 1 .and. mod(count, 2) .eq. 1) coef_sign = -coef_sign
        if (only_left_spin(onlyl) .eq. -1 .and. mod(count + 1, 2) .eq. 1) coef_sign = -coef_sign
        count = count + 1
      else if (left_occ(i) .eq. 1 .and. right_occ(i) .eq. 2) then
        ! left singly right doubly occupied
        open_li = open_li + 1
        open_spin_left = open_spin(left_det, open_li)
        onlyr = onlyr + 1
        if (onlyr .gt. size(only_right)) then
          coef_sign = 0
          exit orbital
        end if
        only_right(onlyr) = i
        only_right_spin(onlyr) = -open_spin_left
        if (only_right_spin(onlyr) .eq. 1 .and. mod(count, 2) .eq. 1) coef_sign = -coef_sign
        if (only_right_spin(onlyr) .eq. -1 .and. mod(count + 1, 2) .eq. 1) coef_sign = -coef_sign
        count = count + 1
      else if (left_occ(i) .eq. 0 .and. right_occ(i) .eq. 1) then
        ! left empty, right singly occupied
        open_ri = open_ri + 1
        open_spin_right = open_spin(right_det, open_ri)
        onlyr = onlyr + 1
        if (onlyr .gt. size(only_right)) then
          coef_sign = 0
          exit orbital
        end if
        only_right(onlyr) = i
        only_right_spin(onlyr) = open_spin_right
        if (mod(count, 2) .eq. 1) coef_sign = -coef_sign
      else if (left_occ(i) .eq. 1 .and. right_occ(i) .eq. 0) then
        ! left singly, right empty
        open_li = open_li + 1
        open_spin_left = open_spin(left_det, open_li)
        onlyl = onlyl + 1
        if (onlyl .gt. size(only_left)) then
          coef_sign = 0
          exit orbital
        end if
        only_left(onlyl) = i
        only_left_spin(onlyl) = open_spin_left
        if (mod(count, 2) .eq. 1) coef_sign = -coef_sign
      end if
    end do orbital

  end subroutine compare_det

  pure function open_spin(det, i)
    integer, intent(in) :: det(:)
    integer, intent(in) :: i
    integer :: open_spin
    if (i .eq. 1) then
      open_spin = det(i)
    else
      open_spin = det(i) - det(i - 1)
    end if
  end function open_spin

  pure function latticePermutation(csf) result(l)
    integer, intent(in) :: csf(:)
    integer :: l(size(csf, 1))
    l(1) = 1
    l(2:) = csf(2:) - csf(:size(csf, 1) - 1)
    l(:) = (1 - l(:))/2 + 1
  end function latticePermutation

!>
!! this subroutine returns the representation matrices of the symmetric group
!! for elementary transpositions
!! see Manley & Gerrat CPC 31, 75
!!
  pure recursive subroutine Umatrix(Nopen, multp, mu, nu, U)
    integer, intent(in) :: Nopen
    integer, intent(in) :: multp
    integer, intent(in) :: mu, nu
    real(long), intent(out) :: U(:, :)
    real(long), allocatable :: U1(:, :)
    real(long), allocatable :: U2(:, :)
    real(long) :: rho
    integer :: k, l
    integer Ncsf
    integer, allocatable :: csfs(:, :)
    integer, allocatable :: lp(:, :)
    integer, allocatable :: lpSwapped(:)
    Ncsf = num_csf(Nopen, multp)
    !allocate(U(Ncsf,Ncsf))
    if (Nopen .eq. 0) then
      U(1, 1) = ONE
    else if (mu .eq. nu) then
      U = ZERO
      do k = 1, Ncsf
        U(k, k) = ONE
      end do
    else if (mu .gt. nu) then
      call Umatrix(Nopen, multp, nu, mu, U)
    else if (nu .eq. mu + 1) then
      call Nopen2csfs(Nopen, multp, csfs)
      allocate (lp(size(csfs, 1), size(csfs, 2)))
      allocate (lpSwapped(size(csfs, 1)))
      do k = 1, Ncsf
        lp(:, k) = latticePermutation(csfs(:, k))
      end do
      U = ZERO
      do k = 1, Ncsf
        if (lp(mu, k) .eq. lp(mu + 1, k)) then
          U(k, k) = -ONE
        else if (count(lp(:mu + 1, k) .eq. 1) .eq. count(lp(:mu + 1, k) .eq. 2)) then
          U(k, k) = ONE
        else
          rho = ONE/dble(abs(count(lp(:mu + 1, k) .eq. 1) - count(lp(:mu + 1, k) .eq. 2)) + 1)
          lpSwapped = lp(:, k)
          lpSwapped(mu) = lp(mu + 1, k)
          lpSwapped(mu + 1) = lp(mu, k)
          do l = k + 1, Ncsf
            if (all(lpSwapped .eq. lp(:, l))) then
              U(k, k) = rho
              U(l, l) = -rho
              U(k, l) = -sqrt(ONE - rho**2)
              U(l, k) = U(k, l)
            end if
          end do! l
        end if
      end do! k
      deallocate (lp)
      deallocate (lpSwapped)
    else if (nu .gt. mu + 1) then
      allocate (U1(Ncsf, Ncsf))
      allocate (U2(Ncsf, Ncsf))
      call Umatrix(Nopen, multp, mu, mu + 1, U1)
      call Umatrix(Nopen, multp, mu + 1, nu, U2)
      U = matmul(U1, matmul(U2, U1))
      deallocate (U1)
      deallocate (U2)
    end if

  end subroutine Umatrix

  subroutine printPermutation(Pk)
    integer, intent(in) :: Pk(:)
    integer :: N
    integer, allocatable :: P(:)
    integer, allocatable :: Pindex(:)
    integer :: i
    N = size(Pk, 1)
    allocate (P(N))
    allocate (Pindex(N))

    P(:) = 0
    Pindex(:) = (/(i, i=1, N)/)
    do i = N, 1, -1
      Pindex = pack((/(i, i=1, N)/), P(:) .eq. 0)
      P(Pindex(Pk(i))) = i
    end do
    deallocate (Pindex)
    do i = 1, N
      write (6, '(i2,x)', advance="no") Pk(i)
    end do
    write (6, '()', advance="yes")

    do i = 1, N
      write (6, '(i2,x)', advance="no") i
    end do
    write (6, '()', advance="yes")
    do i = 1, N
      write (6, '(i2,x)', advance="no") P(i)
    end do
    write (6, '()', advance="yes")
    deallocate (P)
  end subroutine printPermutation

  pure subroutine makeCycleUij(UT, oi, oj, Nsingle, U, yes_cache)
    !! computes coupling matrix for transposition uj->ui
    !! oj is moved down to position oi
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: oi, oj, Nsingle
    real(long), intent(out) :: U(:, :)
    logical, intent(in), optional :: yes_cache
    logical :: cache
    integer :: multp
    integer :: i, lo, up
    integer :: Pk(Nsingle)
    cache = .True.
    if(present(yes_cache)) cache = yes_cache
    if(cache)then
      if (oi .gt. oj) then
        i = UT%indn(Nsingle) + (oj - 1 + oi*(oi - 1)/2) * UT%f(Nsingle)**2
        U = reshape(UT%UCycleMatrix(i:i + UT%f(Nsingle)**2), (/UT%f(Nsingle), UT%f(Nsingle)/))
      else
        i = UT%indn(Nsingle) + (oi - 1 + oj*(oj - 1)/2) * UT%f(Nsingle)**2
        U = transpose(reshape(UT%UCycleMatrix(i:i + UT%f(Nsingle)**2), (/UT%f(Nsingle), UT%f(Nsingle)/)))
      end if
      !call getUCycleMatrixFromTable(UT, Nsingle, oi, oj, U)
      return
    end if
    multp = UT%multp
    if (oi .eq. oj) then
      ! return identity
      U(:, :) = 0
      do i = 1, size(U, 1)
        U(i, i) = 1
      end do
    else
      up = max(oj, oi)
      lo = min(oi, oj)
      !seq(:)=(/ (i,i=1,Nsingle) )/
      !seq(up) = lo
      !seq(lo:up-1) = seq(lo:up-1) + 1
      ! swapping the order of open shell electrons
      ! seq=1, 2, ..., oi,   ..., oj-1, oj, oj+1, ...
      !     1, 2, ..., oi+1, ..., oj,   oi, oj+1, ...
      ! New order:
      !     1, 2, ..., oj, oi, oi+1,, ... oj-1, oj+1, ...
      !                     \________________/
      ! Permutation (oi,...,oj) = (oi,oi+1) (oi+1,oi+2) ... (oj-1,oj)

      !Pk= seq2Permutation(size(seq), seq)
      Pk(1:lo) = (/(i, i=1, lo)/)
      Pk(lo + 1:up) = (/(i - 1, i=lo + 1, up)/)
      Pk(up + 1:Nsingle) = (/(i, i=up + 1, Nsingle)/)
      call makeCycleU(UT, Pk, U)
      if (oi .gt. oj) U = transpose(U)
    end if

  end subroutine makeCycleUij

  pure subroutine makeCycleU(UT, Pk, U)
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: Pk(:)
    real(long), intent(out) :: U(:, :)
    integer :: N
    integer :: i
    real(long) U1(size(U, 1), size(U, 2))
    N = size(Pk, 1)
    U(:, :) = ZERO
    forall (i=1:size(U, 1))
      U(i, i) = ONE
    end forall
    do i = N, 1, -1
      if (Pk(i) .eq. i) cycle
      !call Umatrix(N, multp, i, Pk(i), U1)
      call getUMatrixFromTable(UT, N, i, Pk(i), U1)
      U = matmul(U1, U)
      !U = U1 .m. U
    end do
  end subroutine makeCycleU

  pure function index2Permutation(N, index) result(k)
    integer, intent(in) :: N
    integer, intent(in) :: index
    integer :: ind, i
    integer :: k(N)
    ind = index
    do i = N, 1, -1
      k(i) = i - int((ind - 1)/int(factorial(i - 1)))
      ind = ind - (i - k(i))*int(factorial(i - 1))
    end do
  end function index2Permutation

  pure function seq2Permutation(N, seq) result(k)
    integer, intent(in) :: N
    integer, intent(in) :: seq(N)
    integer :: P(N), Pindex(N)
    integer :: k(N)
    integer :: i, j
    P(:) = seq(:)
    Pindex(:) = 1
    do i = N, 1, -1
      do j = 1, N
        if (P(j) .eq. i) exit
      end do
      k(i) = j ! element i is at jth position
      if (j .lt. N) then
        P(j:i - 1) = P(j + 1:i)
        P(i) = 0
      end if
    end do
  end function seq2Permutation

  pure function csfstring(occ, csf) result(string)
    !! returns printable string for a given spin configuration
    integer, intent(in) :: csf(:)
    integer, intent(in) :: occ(:)
    character(len=size(occ)) :: string
    integer :: i, j

    string = ""
    i = 1
    do j = 1, size(occ, 1)
      if (occ(j) .eq. 2) then
        string = trim(string)//"2"
      else if (occ(j) .eq. 0) then
        string = trim(string)//"0"
      else if (occ(j) .eq. 1) then
        if (i .eq. 1) then
          string = trim(string)//"/"
        else if (csf(i) .gt. csf(i - 1)) then
          string = trim(string)//"/"
        else
          string = trim(string)//'\'
        end if
        i = i + 1
      end if
    end do
  end function csfstring

  subroutine testUmatrix(N, multp)
    integer, intent(in) :: multp
    integer, intent(in) :: N
    real(long), allocatable :: U(:, :)
    real(long), allocatable :: Usum(:, :)

    integer, allocatable :: Pk(:)
    integer :: i, j
    type(UmatrixTable) :: UT

    allocate (Pk(N))
    call constructUMatrixTable(N, multp, UT)

    call Umatrix(N, multp, 1, 1, Usum)
    Usum(:, :) = ZERO
    do j = 1, int(factorial(N))
      write (6, *) 'j=', j
      Pk(:) = 0
      Pk(:) = index2Permutation(N, j)
      write (6, '(99i1)') Pk(:)
      call printPermutation(Pk)
      call makeCycleU(UT, Pk, U)
      Usum = Usum + U**2
      write (6, *) 'U='
      do i = 1, size(U, 1)
        write (6, '(99f7.2)') U(i, :)
      end do

    end do

    write (6, *) 'sum |U_ij|^2== N!/f_{NS} ?'
    do i = 1, size(Usum, 1)
      do j = 1, size(Usum, 1)
        write (6, '(i4,i4,f17.2,x,"==",f17.2)') i, j, Usum(i, j), factorial(N)/size(Usum, 1)
        if (abs(Usum(i, j) - factorial(N)/size(Usum, 1)) .gt. 1e-4) then
          stop
        end if

      end do
    end do

  end subroutine testUmatrix

end module spin_coupling

