!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module stdio
  !! author: Sang-Kil Son
  !!
  !! STDIO : Standard Input/Output module
  !!   written by Son, Sang-Kil in Mar. 2005
  !! This module requires two intrinsic subroutines: getarg() and getenv()
  !!
  !! Data type
  !! - Argument : for command-line & input file options
  !!
  !! Functions(F) and subroutines(SR)
  !! - SR error_message( string ) : print error messages and stop
  !!
  !! - SR get_options( ARG ) : get options from command line
  !! - SR get_input( ARG )   : get options from input file and command line
  !!   ex) foobar -N 30  or  -N:30  or -N=30  or -N30  or +N 30
  !!       foobar -N 10-30
  !!       foobar file.in  or  -N 30 file.in
  !!       * if no option, file.in will be taken as input file
  !! - SR parse_option( ARG, option_name, string or value )
  !! - SR show_input( ARG )  : show options from input file and command line
  !!
  !! - SR footnote() : print out CPU_time, current time, and hostname
  !!
  !! - SR check_time( counter ) : check elapsed time
  !! - F job_time() : return job time converted from counter with hostname
  !!   . F job_time_integer( counter ) : integer value from check_time()
  !!   . F job_time_real( CPU_time ) : real value from CPU_time()
  !!
  !! - F today() : return the date string of today
  !! - F hostname() : return the hostname
  !!
  use constants
  use iso_c_binding
  implicit none
  type :: Argument
    !! general command line argument
    integer :: N
    character(len=2048), dimension(:), allocatable :: option, string
    real(long), dimension(:), allocatable :: value, value2
    logical, dimension(:), allocatable :: yes
  end type Argument
  interface job_time
    !! returns job time
    module procedure job_time_integer8, job_time_real
  end interface
  interface parse_option
    module procedure parse_option_string, &
      parse_option_real, parse_option_integer, &
      parse_option_logical
  end interface

  interface
    function check_mem_current() bind(c, name="getCurrentRSS") result(mem)
      !! returns current memory
      use :: iso_c_binding, only:c_size_t
      implicit none
      integer(c_size_t) :: mem
    end function check_mem_current

    function check_mem_peak() bind(c, name="getPeakRSS") result(mem)
      !! returns peak memory      
      use :: iso_c_binding, only:c_size_t
      implicit none
      integer(c_size_t) :: mem
    end function check_mem_peak
  end interface
  private
  public :: Argument, check_mem_current, check_mem_peak, job_time, parse_option, &
    parse_int_array, parse_string_array, parse_real_array, get_input, error_message, &
    timestamp, check_time, convert_to_lowercase, convert_to_uppercase, show_input, footnote, &
    init_time, init_checkpoint, checkpoint, c_f_string, purge_ARG, user_time
contains

  subroutine purge_ARG(ARG)
    type(Argument), intent(inout) :: ARG    
    deallocate(ARG%option)
    deallocate(ARG%string)
    deallocate(ARG%value)
    deallocate(ARG%value2)
    deallocate(ARG%yes)
  end subroutine purge_ARG

  subroutine error_message(s)
    !! show error message and stop the program
    character(len=*), optional :: s
    if (present(s)) write (STDERR, *) "ERROR: "//s
    stop
  end subroutine error_message

!>----------------------------------------------------------------------------
!! get options from command lines and the input file and put them into ARG
!!----------------------------------------------------------------------------
!! if filename2 is specified, override reading of command line options.
!! instead only read from filename2
  subroutine get_input(ARG, filename2)
    type(Argument), intent(out) :: ARG
    character(len=*), intent(in), optional :: filename2
    type(Argument) :: ARG_CL
    integer :: i, j, n, n_line, status
    character(len=2048) :: filename, line
    integer :: INPUT_FILE = 9
    integer, parameter :: MAX_N_LINE = 20000
    logical :: yes_exist
    logical :: itsopen
    inquire (unit=INPUT_FILE, opened=itsopen)
    if (itsopen) then
      close (INPUT_FILE)
    end if

    !
    ! First, get options from command line
    !
    if (present(filename2)) then
      allocate (ARG_CL%option(1))
      allocate (ARG_CL%string(1))
      allocate (ARG_CL%value(1))
      allocate (ARG_CL%value2(1))
      allocate (ARG_CL%yes(1))
      ARG_CL%n = 1
      ARG_CL%option(1) = ""
      ARG_CL%string(1) = filename2
    else
      call get_options(ARG_CL)
    end if

    !
    ! find filename in ARG_CL
    !
    n_line = 0    ! total # line of file
    arg_index: do i = 1, ARG_CL%N
      select case (ARG_CL%option(i))
      case ("")
        filename = ARG_CL%string(i)
        if (filename .eq. "") then
          INPUT_FILE = STDIN
          n_line = MAX_N_LINE   ! add additional lines of input file
        else
          inquire (file=filename, exist=yes_exist)
          if (.not. yes_exist) &
            call error_message(trim(filename)//" is not found.")
          open (INPUT_FILE, file=filename, action="read", form="formatted")
          n_line = 0
          do
            read (INPUT_FILE, "(a)", iostat=status) line
            if (status .lt. 0) exit
            n_line = n_line + 1
          end do
          rewind (INPUT_FILE)
        end if
        exit arg_index
      end select
    end do arg_index
    !
    ! memory allocation
    !
    n = ARG_CL%N + n_line
    allocate (ARG%option(n), ARG%string(n), &
              ARG%value(n), ARG%value2(n), ARG%yes(n))
    !
    ! read data from input file or STDIN
    !
    n = 0         ! # meaningful line
    if (n_line .gt. 0) then
      do
        read (INPUT_FILE, "(a)", iostat=status) line
        if (status .lt. 0) exit
        !
        ! put this line into ARG
        !
        if (len_trim(line) .ne. 0 .and. line(1:1) .ne. "#") then
          n = n + 1
          if (n .gt. ARG_CL%N + MAX_N_LINE &
              ) call error_message("The length of input file is too long.")
          j = index(line, "=")
          if (j .eq. 0) j = index(line, ":")
          ARG%option(n) = "-"//adjustl(line(1:j - 1))
          ARG%string(n) = adjustl(line(j + 1:))
          read (ARG%string(n), *, iostat=status) ARG%value(n)
          ARG%yes(n) = determine_yes_no(ARG%string(n))
        end if
      end do
      rewind (INPUT_FILE)
    end if
    !
    ! copy ARG_CL to ARG
    !
    ARG%N = ARG_CL%N + n
    do i = n + 1, ARG%N
      ARG%option(i) = ARG_CL%option(i - n)
      ARG%string(i) = ARG_CL%string(i - n)
      ARG%value(i) = ARG_CL%value(i - n)
      ARG%value2(i) = ARG_CL%value2(i - n)
      ARG%yes(i) = ARG_CL%yes(i - n)
    end do
    !
    ! deallocate ARG_CL
    !
    deallocate (ARG_CL%option, ARG_CL%string, &
                ARG_CL%value, ARG_CL%value2, ARG_CL%yes)
    do i = 1, ARG%N
      call convert_to_lowercase(ARG%option(i))
    end do

  end subroutine get_input

  subroutine convert_to_lowercase(string)
    !! convert string into lowercase
    character(len=*), intent(inout) :: string
    integer :: i, j
    character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
    do i = 1, len_trim(string)
      j = index(uppercase, string(i:i))
      if (j .gt. 0) string(i:i) = lowercase(j:j)
    end do
  end subroutine convert_to_lowercase

  subroutine convert_to_uppercase(string)
    !! convert string into uppercase
    character(len=*), intent(inout) :: string
    integer :: i, j
    character(len=*), parameter :: uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    character(len=*), parameter :: lowercase = "abcdefghijklmnopqrstuvwxyz"
    do i = 1, len_trim(string)
      j = index(lowercase, string(i:i))
      if (j .gt. 0) string(i:i) = uppercase(j:j)
    end do
  end subroutine convert_to_uppercase

!>----------------------------------------------------------------------------
!! get command-line options and put them into ARG
!!----------------------------------------------------------------------------
  subroutine get_options(ARG)
    type(Argument), intent(out) :: ARG
    integer :: narg, i, k, n
    integer :: iargc, status
    real(long) :: value
    character(len=2048) :: s, s2, s3
    narg = iargc()
    allocate (ARG%option(narg), ARG%string(narg))
    allocate (ARG%value(narg), ARG%value2(narg), ARG%yes(narg))
    ARG%option(:) = ""
    ARG%value(:) = ZERO
    ARG%value2(:) = ZERO
    ARG%string(:) = ""
    ARG%yes(:) = .false.
    i = 0
    do n = 1, narg
      i = i + 1
      if (i .gt. narg) exit
      call getarg(i, s)
      if (s(1:1) .eq. "-" .or. s(1:1) .eq. "+") then
        !
        ! option sign started with - or +
        !
        s2 = ""; value = ZERO
        if (i + 1 .le. narg) call getarg(i + 1, s2)
        read (s2, *, iostat=status) value
        if (i + 1 .gt. narg .or. (s2(1:1) .eq. "-" .and. value .eq. ZERO) &
            .or. (s2(1:1) .eq. "+" .and. value .eq. ZERO)) then
          !
          ! next option follows the option
          ! so, we need to find some arguments inside the option
          !
          k = index(s, "=")
          if (k .eq. 0) k = index(s, ":")
          if (k .ne. 0) then
            ARG%option(n) = s(:k - 1)      ! the case of -N=30 or -N:30
            ARG%string(n) = s(k + 1:)
            read (s(k + 1:), *, iostat=status) ARG%value(n)
          else
            read (s(3:), *, iostat=status) ARG%value(n)
            if (ARG%value(n) .ne. ZERO) then
              ARG%option(n) = s(1:2)      ! the case of -N30
              ARG%string(n) = s(3:)
            else if (s(3:3) .eq. "0") then
              ARG%option(n) = s(1:2)      ! the case of -N0
              ARG%string(n) = s(3:)
            else
              ARG%option(n) = s           ! the case of -N
              ARG%string(n) = ""
            end if
          end if
        else
          !
          ! there is an argument before next option
          !
          ARG%option(n) = s               ! the case of -N 30
          ARG%string(n) = s2
          read (s2, *, iostat=status) ARG%value(n)
          i = i + 1
        end if
      else
        !
        ! no option sign
        !
        ARG%option(n) = ""
        ARG%string(n) = s
        read (s, *, iostat=status) ARG%value(n)
      end if
      !
      ! check the range such as "30-40", "30:40", "30~40"
      ! exclude exponential expressions like "1e-10" or "1E-10"
      !
      s3 = ARG%string(n)
      k = scan(s3, "-~,:;")
      if (k .gt. 1) then
        if (s3(k - 1:k) .ne. "e-" .and. s3(k - 1:k) .ne. "E-") then
          read (s3(:k - 1), *, iostat=status) ARG%value(n)
          read (s3(k + 1:), *, iostat=status) ARG%value2(n)
        end if
      end if
      !
      ! set ARG%yes
      !
      ARG%yes(n) = determine_yes_no(ARG%string(n))
    end do
    ARG%N = n - 1
  end subroutine get_options

!>----------------------------------------------------------------------------
!! show all options from the input file and the command line
!!----------------------------------------------------------------------------
  subroutine show_input(ARG)
    type(Argument), intent(in) :: ARG
    integer :: i
    do i = 1, ARG%N
      if (ARG%option(i) .eq. "-") then
        write (STDERR, "(a)") trim(ARG%string(i))
      else if (ARG%option(i) (1:1) .eq. "-") then
        write (STDERR, "(a,'= ',a)") trim(ARG%option(i) (2:)), &
          trim(ARG%string(i))
      else if (ARG%option(i) (1:1) .eq. "+") then
        write (STDERR, "('# command option: ',a,' ',a)") trim(ARG%option(i)), &
          trim(ARG%string(i))
      else
        write (STDERR, "('# command option: ',a)") trim(ARG%string(i))
      end if
    end do
  end subroutine show_input

!>----------------------------------------------------------------------------
!! get one option value or string from ARG
!!   The first chracter of ARG%option(i) should be + or -.
!!   That's why it is taken by ARG%option(i)(2:).
!!----------------------------------------------------------------------------
  subroutine parse_option_string(ARG, option_name, string)
    type(Argument), intent(in) :: ARG
    character(len=*), intent(in) :: option_name
    character(len=*), intent(out) :: string
    integer :: i
    do i = 1, ARG%N
      if (ARG%option(i) (2:) .eq. option_name) string = ARG%string(i)
    end do
  end subroutine parse_option_string

  subroutine parse_option_real(ARG, option_name, value, value2)
    type(Argument), intent(in) :: ARG
    character(len=*), intent(in) :: option_name
    real(long), intent(out) :: value
    real(long), intent(out), optional :: value2
    integer :: i
    do i = 1, ARG%N
      if (ARG%option(i) (2:) .eq. option_name) then
        value = ARG%value(i)
        if (present(value2)) value2 = ARG%value2(i)
      end if
    end do
  end subroutine parse_option_real

  subroutine parse_option_integer(ARG, option_name, value, value2)
    type(Argument), intent(in) :: ARG
    character(len=*), intent(in) :: option_name
    integer, intent(out) :: value
    integer, intent(out), optional :: value2
    integer :: i
    do i = 1, ARG%N
      if (ARG%option(i) (2:) .eq. option_name) then
        value = int(ARG%value(i))
        if (present(value2)) value2 = int(ARG%value2(i))
      end if
    end do
  end subroutine parse_option_integer

  subroutine parse_option_logical(ARG, option_name, yes)
    type(Argument), intent(in) :: ARG
    character(len=*), intent(in) :: option_name
    logical, intent(inout) :: yes
    integer :: i
    do i = 1, ARG%N
      if (ARG%option(i) (2:) .eq. option_name &
          ) yes = determine_yes_no(ARG%string(i), yes)
    end do
  end subroutine parse_option_logical

  function determine_yes_no(string, yes_default) result(yes)
    logical :: yes
    character(len=*), intent(in) :: string
    logical, intent(in), optional :: yes_default
    character(len=5) :: word
    if (present(yes_default)) then
      yes = yes_default
    else
      yes = .false.
    end if
    read (string, "(a)") word
    select case (word)
    case ("yes", "YES", "Yes", "Y", "on", "ON", "On", &
          "true", "TRUE", "True", "T", ""); yes = .true.
    case ("no", "NO", "No", "N", "off", "OFF", "Off", &
          "false", "FALSE", "False", "F"); yes = .false.
    case default; ! no change
    end select
  end function determine_yes_no

  function parse_string_array(buffer, delim) result(array)
    !! parses a string array
    character(*), intent(in) :: buffer
    character, intent(in) :: delim
    integer :: max_len
    character(len=:), allocatable :: array(:)
    integer :: pos1, npos, j
    integer :: pos2
    ! count the number of entries
    npos = 0
    pos1 = 1
    max_len = 0
    do
      pos2 = index(buffer(pos1:), delim)
      max_len = max(max_len, pos2)
      IF (pos2 .eq. 0) THEN
        npos = npos + 1
        max_len = max(max_len, len(trim(buffer)) + 1)
        exit
      end if
      npos = npos + 1
      pos1 = pos2 + pos1
    end do
    allocate (character(max_len) :: array(npos))
    array(:) = ""
    pos1 = 1
    do j = 1, npos
      pos2 = index(buffer(pos1:), delim)
      if (j .lt. npos) then
        array(j) = buffer(pos1:pos1 + pos2 - 2)
      else
        array(j) = buffer(pos1:)
      end if
      pos1 = pos2 + pos1
    end do
  end function parse_string_array

  function parse_int_array(buffer, delim, rangeDelim) result(array)
    !! parses an integer array
    character(*), intent(in) :: buffer
    character, intent(in) :: delim
    character, intent(in), optional :: rangeDelim
    integer, allocatable :: array(:)
    integer :: pos1, npos, j, k
    integer :: pos2, pos3
    integer :: r1, r2
    ! count the number of entries
    npos = 0
    pos1 = 1
    pos3 = 0
    do
      if (present(rangeDelim)) then
        pos3 = index(buffer(pos1:), rangeDelim)
      end if
      pos2 = index(buffer(pos1:), delim)
      if (pos3 .gt. 0 .and. (pos2 .gt. pos3 .or. pos2 .eq. 0)) then
        ! next comes a integer range: from -- to
        read (buffer(pos1:pos1 + pos3 - 2), *) r1
        if (pos2 .eq. 0) then
          ! range is terminated by end of string
          read (buffer(pos1 + pos3:), *) r2
          npos = npos + r2 - r1 + 1
          exit
        else
          ! range is terminated by delim
          read (buffer(pos1 + pos3:pos1 + pos2 - 2), *) r2
          npos = npos + r2 - r1 + 1
          pos1 = pos1 + pos2
        end if
      else
        ! next comma separated integer
        if (pos2 .eq. 0) then
          ! terminated by the end of the string
          npos = npos + 1
          exit
        else
          ! terminated by delim
          npos = npos + 1
          pos1 = pos2 + pos1
        end if
      end if
    end do

    allocate (array(npos))

    pos1 = 1
    j = 1
    do
      if (present(rangeDelim)) then
        pos3 = index(buffer(pos1:), rangeDelim)
      end if
      pos2 = index(buffer(pos1:), delim)
      if (pos3 .gt. 0 .and. (pos2 .gt. pos3 .or. pos2 .eq. 0)) then
        ! next comes a integer range: from r1 -- to r2
        read (buffer(pos1:pos1 + pos3 - 2), *) r1
        if (pos2 .eq. 0) then
          ! range is terminated at the end of the string
          read (buffer(pos1 + pos3:), *) r2
        else
          ! range is terminated by delim
          read (buffer(pos1 + pos3:pos1 + pos2 - 2), *) r2
          pos1 = pos2 + pos1
        end if
        do k = r1, r2
          array(j) = k
          j = j + 1
        end do
        if (pos2 .eq. 0) exit
      else
        ! next integer value
        if (pos2 .eq. 0) then
          ! terminated by end-of-string
          read (buffer(pos1:), *) array(j)
          j = j + 1
          exit
        else
          ! terminated by delimiter
          read (buffer(pos1:pos1 + pos2 - 2), *) array(j)
          j = j + 1
        end if
        pos1 = pos2 + pos1
      end if

      if (j .gt. npos) exit
    end do
  end function parse_int_array

  function parse_real_array(buffer, delim) result(array)
    !! parses a float array
    character(*), intent(in) :: buffer
    character, intent(in) :: delim
    real(long), allocatable :: array(:)

    integer :: pos1, npos, j
    integer :: pos2

    ! count the number of entries
    npos = 0
    pos1 = 1
    do
      pos2 = index(buffer(pos1:), delim)
      IF (pos2 .eq. 0) THEN
        npos = npos + 1
        exit
      end if
      npos = npos + 1
      pos1 = pos2 + pos1
    end do
    allocate (array(npos))
    pos1 = 1
    do j = 1, npos
      pos2 = index(buffer(pos1:), ",")
      if (j .lt. npos) then
        read (buffer(pos1:pos1 + pos2 - 2), *) array(j)
      else
        read (buffer(pos1:), *) array(j)
      end if
      pos1 = pos2 + pos1
    end do
  end function parse_real_array

  !>----------------------------------------------------------------------------
  !! return the date string of today  ex) 26-Jun-2004
  !!----------------------------------------------------------------------------
  function today() result(string)
    integer :: v(8), y, m, d, hh, mm, ss
    character(len=20) :: string
    character(len=3), dimension(12), parameter :: month = &
                                                  (/"Jan", "Feb", "Mar", "Apr", "May", "Jun", &
                                                    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"/)
    call date_and_time(values=v)
    y = v(1); m = v(2); d = v(3)
    hh = v(5); mm = v(6); ss = v(7)
    write (string, "(i2,'-',a3,'-',i4,x,i2,':',i2.2,':',i2.2)") &
      d, month(m), y, hh, mm, ss
  end function today

  !>----------------------------------------------------------------------------
  !! Use this procedure when you want a check point.
  !! But, for initilization, you must put this without counter
  !!      at the very begining of your program.
  !! ex)    check_time()
  !!        ...
  !!        check_time( c, t )
  !!        print *, job_time( c )
  !!
  !! Here is an issue of SYSTEM_CLOCK( count, count_rate, count_max )
  !!
  !! From the GNU Fortran Compiler webpage:
  !! "Note that the millisecond resolution of the kind=4 version implies that
  !!  the COUNT will wrap around in roughly 25 days."
  !!
  !! In the Intel Fortran Compiler, count_rate is 10000 for integer(4).
  !! Since the range of integer(4) is -2147483648 ~ 2147483647,
  !! count in integer(4) can handle only 429496 seconds, i.e., about 5 days!
  !!
  !! To avoid this issue, let's use the kind=8 version, i.e., integer(8).
  !!----------------------------------------------------------------------------
  subroutine check_time(c, t)
    integer*8, intent(out), optional :: c ! used for system_clock
    real, intent(out), optional :: t      ! used for CPU_time
    integer*8 :: c1
    real :: t1
    integer*8, save :: c0 = 0
    real, save :: t0 = ZERO
    logical, save :: yes_started = .false.
    call system_clock(c1)       ! it must give real elapsed time
    call CPU_time(t1)           ! it must give user(CPU) time, but not with PGI
    if (present(c) .and. present(t)) then
      if (yes_started .eqv. .false.) &
        call error_message("check_time should be used without counter "// &
                           "at the begining of your program.")
      c = c1 - c0
      t = t1 - t0
    else
      yes_started = .true.        ! initialization: save c1 and t1 into c0 and t0
      c0 = c1
      t0 = t1
    end if
  end subroutine check_time

  !>----------------------------------------------------------------------------
  !! checkpoint(): more convenient procedure than check_time()
  !!   An array time(0:) is a table containing all checkpoint times
  !!   - time(0) : reference time stamp
  !!   - time(1) : elapsed time for the first checkpoint
  !!   - time(2) : elapsed time for the second checkpoint
  !!   - ...
  !! ex) call init_checkpoint( time ) : initialize a time array
  !!     ... (do some calculations) ...
  !!     call checkpoint( time ) : make a checkpoint with the time array
  !!
  !!     call print_checkpoint( time, string, iout ) :
  !!       make a checkpoint with the time array and print "string" on iout
  !!----------------------------------------------------------------------------
  subroutine checkpoint(time, counter)
    !! initialize a time array
    real, intent(inout) :: time(0:)
    integer, intent(in), optional :: counter
    real :: t1
    integer :: i
    call CPU_time(t1)
    if (present(counter)) then
      if (counter .eq. 0) then
        time(0) = t1      ! store the initial time stamp
        time(1:) = ZERO   ! initialize
        return
      end if
      i = counter
    else
      loop: do i = 1, size(time)          ! search zero
        if (time(i) .eq. ZERO) exit loop
      end do loop
    end if
    if (i .gt. size(time) &
        ) call error_message("checkpoint(): increase N of time(:)")
    t1 = t1 - time(0)                     ! substract the initial time stamp
    t1 = t1 - sum(time(1:))             ! substract all previous steps
    if (t1 .eq. ZERO) t1 = tiny(t1)       ! put an infinitesimal value
    time(i) = time(i) + t1                ! store or accumulate the elapsed time
  end subroutine checkpoint

  subroutine print_checkpoint(time, string, iout)
    !! print checkpoint
    real, intent(inout) :: time(0:)
    character(len=*), intent(in) :: string
    integer, intent(in), optional :: iout
    integer :: i, OUTPUT = STDERR
    real :: t1
    if (present(iout)) OUTPUT = iout
    call CPU_time(t1)
    t1 = t1 - time(0)                     ! substract the initial time stamp
    loop: do i = 1, size(time)            ! search zero
      if (time(i) .eq. ZERO) exit loop
      t1 = t1 - time(i)                   ! substract the previous step
    end do loop
    if (i .gt. size(time) &
        ) call error_message("print_checkpoint(): increase N of time(:)")
    if (t1 .eq. ZERO) t1 = tiny(t1)       ! put an infinitesimal value
    time(i) = t1                          ! store the elapsed time
    write (OUTPUT, "('# CHECKPOINT ',a,' : ',a)") string, trim(timestamp(t1))
  end subroutine print_checkpoint

  subroutine init_checkpoint(time)
    real, intent(inout) :: time(0:)
    call checkpoint(time, 0)
  end subroutine init_checkpoint

  !>----------------------------------------------------------------------------
  !! return the time string converted from counter
  !! ex) counter=100000000 with integer(8)
  !!     --> 100 sec --> 1 minute 40.00 seconds
  !!----------------------------------------------------------------------------
  function job_time_integer8(counter) result(string)
    integer*8, intent(in) :: counter
    integer*8 :: c, c_rate
    integer :: hh, mm, mss, time
    real :: ss
    character(len=60) :: string
    call system_clock(c, c_rate)
    mss = int(mod(counter, c_rate))
    time = int(counter/c_rate)
    ss = mod(time, 60) + real(mss)/real(c_rate)
    time = time/60; mm = mod(time, 60)
    time = time/60; hh = time
    string = get_job_time(hh, mm, ss)
  end function job_time_integer8

  function job_time_real(time) result(string)
    real, intent(in) :: time
    integer :: hh, mm, itime
    real :: ss
    character(len=60) :: string
    itime = int(time/60)
    mm = mod(itime, 60)
    itime = itime/60; hh = itime
    ss = time - real(hh*60*60 + mm*60)
    string = get_job_time(hh, mm, ss)
  end function job_time_real

  function get_job_time(hh, mm, ss) result(string)
    character(len=60) :: string
    integer :: hh, mm
    real :: ss
    if (hh .ne. 0) then
      write (string, "(i8,x,a,x,i2,x,a,x,f5.2,x,'seconds')") &
        hh, trim(plural("hour", hh)), mm, trim(plural("minute", mm)), ss
      string = adjustl(string)
    else if (mm .ne. 0) then
      write (string, "(i2,x,a,x,f5.2,x,'seconds')") &
        mm, trim(plural("minute", mm)), ss
    else
      write (string, "(f5.2,x,'seconds')") ss
    end if
    string = trim(string)//" on "//trim(hostname())
  contains
    function plural(s, i) result(p)
      character(len=*), intent(in) :: s
      integer, intent(in) :: i
      character(len=len(s) + 1) :: p
      p = s
      if (i .gt. 1) p = trim(p)//"s"
    end function plural
  end function get_job_time

  function timestamp(time) result(string)
    real, intent(in) :: time
    integer :: hh, mm, ss, mss, itime
    character(len=60) :: string
    itime = int(time)
    mss = int((time - real(itime))*100)
    ss = mod(itime, 60)
    itime = itime/60; mm = mod(itime, 60)
    itime = itime/60; hh = itime
    !
    ! return a string of "12345:67:89.00" <-- H:MM:SS.ss, string length = 14
    !
    write (string, "(i5,':',i2.2,':',i2.2,'.',i2.2)") hh, mm, ss, mss
  end function timestamp

  !>----------------------------------------------------------------------------
  !! return hostname from environment variable of HOSTNAME
  !!----------------------------------------------------------------------------
  function hostname() result(string)
    character(len=20) :: string
    call hostnm(string)
    if (string .eq. "") call getenv("HOSTNAME", string)
    if (string .eq. "") call getenv("HOST", string)
  end function hostname

  !>----------------------------------------------------------------------------
  !! print out CPU_time, current time, and hostname
  !!   With PGI compiler, CPU_time gives real time instead of CPU time.
  !!   What can I do for it? Huh?
  !! USAGE)  call init_time()
  !!         ...
  !!         call footnote()
  !!----------------------------------------------------------------------------
  subroutine init_time()
    call check_time()
  end subroutine init_time

  function user_time(counter) result(ss)
    integer*8, intent(in) :: counter
    integer*8 :: c, c_rate
    real :: ss
    call system_clock(c, c_rate)
    ss = real(counter)/real(c_rate)
  end function user_time

  subroutine footnote(IOUT)
    !! prints footnote
    integer, intent(in), optional :: IOUT
    integer :: OUTPUT
    integer*8 :: mem
    integer*8 :: c        ! used for system_clock
    real :: t             ! used for CPU_time
    character(len=80) :: string_today, string_real, string_CPU
    OUTPUT = STDOUT; if (present(IOUT)) OUTPUT = IOUT
    call check_time(c, t)
    string_today = today()
    string_real = job_time(c)
    string_CPU = job_time(t)
    write (OUTPUT, "('#'/'# Current date/time  : ',a)") trim(string_today)
    write (OUTPUT, "('# elapsed real time  : ',a)") trim(string_real)
    write (OUTPUT, "('# Job time(CPU_time) : ',a)") trim(string_CPU)

    mem = check_mem_peak()
    if (mem .gt. 1024_long*1024_long*1024_long*1024_long) then
      write (OUTPUT, "('# MEMORY : ',f12.3,'TB')") mem/1024.0_long/1024.0_long/1024_long/1024_long
    else if (mem .gt. 1024_long*1024_long*1024_long) then
      write (OUTPUT, "('# MEMORY : ',f12.3,'GB')") mem/1024.0_long/1024.0_long/1024_long
    else if (mem .gt. 1024_long*1024_long) then
      write (OUTPUT, "('# MEMORY : ',f12.3,'MB')") mem/1024.0_long/1024.0_long
    else if (mem .gt. 1024_long) then
      write (OUTPUT, "('# MEMORY : ',f12.3,'kB')") mem/1024.0_long
    else
      write (OUTPUT, "('# MEMORY : ',f12.3,'B')") mem
    end if
    !
    ! for OpenMP machines
    !
    !call getenv( "OMP_NUM_THREADS", string )
    !write(OUTPUT, "('# Job time(CPU_time) : ',a,' (x ',a,' CPUs)')") &
    !  trim(string_CPU), trim(string)
  end subroutine footnote

! subroutine flush( OUTPUT )
!   integer, intent(in) :: OUTPUT
!   character(len=2048) :: filename
!   logical :: yes_file
!   inquire( unit=OUTPUT, named=yes_file, name=filename )
!   close( OUTPUT )
!   if ( yes_file .and. filename /= "stdout" ) then
!     open( OUTPUT, file=trim(filename), status="old", position="append" )
!   else
!     open( OUTPUT )
!   end if
! end subroutine flush

  subroutine C_F_string(cStrPtr, fStr )
    !! converts a c_char pointer array cStrPtr to a fortran string fStr
    type(c_ptr), intent(in) :: cStrPtr !! c pointer to a char* array
    character(len=*), intent(out) :: fStr !! fortan string
    character(kind=c_char,len=1), dimension(:), pointer :: pchars
    integer :: i
    call c_f_pointer(cStrPtr, pchars, [huge(0)])
    do i=1,len(fStr)
      if(pchars(i) == C_NULL_char) exit
      fStr(i:i) = pchars(i)
    end do
    if(i<len(fStr))&
      fStr(i:) = ' '
  end subroutine C_F_string

end module stdio
