!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module wavefunction
  !! author: Sang-Kil Son
  !!
  !! All about wavefunctions (actually orbitals) for the molecular system
  !!   written by Yajiang Hao in 2014
  !!   editted by Ludger Inhester, Kota Hanasaki, Sang-Kil Son in 2014-2015
  !!
  !! USAGE:
  !!   type(WF_data) :: F
  !!   call construct_WF_data( P, A, G, B, F )  ! create F
  !!   call SCF_procedure( P, A, G, B, F )      ! calculate MO and E (in scf.f90)
  !!   call calculate_gradient( P, A, G, B, F ) ! calculate grad. (in gradient.f90)
  !!
  !! WF_data contains all information of SCF results.
  !! The main quantities are:
  !! - total energy = F%E_tot
  !! - molecular orbitals: stored in F%LC(:,:)
  !! - orbital energies: stored in F%CE(:)
  !! - gradients: stored in F%grad(:)
  !!
  use constants
  use sph_harmonics
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use interpolate
  use matrix
  use atom
  use stdio
  implicit none
  type :: WF_data
    !! wave function data
    integer :: N_nuc            ! # of nuclei
    integer :: N_basis          ! # of basis functions
    integer :: Nr_basis         ! # of basis functions
    integer :: N_grid           ! # of molecular grid points
    integer :: N_MO             ! # of occupied MO
    integer :: homo_idx         ! index of HOMO
    !
    ! basis-function-based quantities: (1:N_basis) or (1:N_basis,1:N_basis)
    !   f_occ(k) : fractional occupation (only used for finite temp. calc.)
    !   CE(k) : orbital energy for the k-th MO
    !   LC(i,k) : MO coefficient of the i-th basis for the k-th MO
    !   LevelShift(k) : level shift for individual orbitals
    !
    integer :: Nopen
    integer :: multp
    integer :: csf_num

    real(long), allocatable :: fidelity(:)      ! fidelity of MO
    real(long), allocatable :: fidelityb(:)      ! fidelity of MO beta
    integer, allocatable :: n_occ(:)            ! integer occupations
    integer, allocatable :: n_occ_old(:)            ! integer occupations
    integer, allocatable :: irrep_MO(:)            ! irreps of MO
    integer :: state_irrep
    real(long), allocatable :: f_occ(:)         ! fractional occupations
    real(long), allocatable :: D(:, :)           ! the density matrix
    real(long), allocatable :: D_old(:, :)       ! the density matrix old
    real(long), allocatable :: D_old2(:, :)      ! the density matrix old
    real(long), allocatable :: CE(:)            ! orbital energy
    real(long), allocatable :: LC(:, :)          ! MO coefficient
    real(long), allocatable :: H(:, :)           ! one particle hamiltonian
    real(long), allocatable :: H0(:, :)          ! kinetic+v_ext part
    real(long), allocatable :: H0_MO(:, :)       ! kinetic+v_ext part in MO basis
    real(long), allocatable :: H_old(:, :)       ! one particle hamiltonian old
    real(long), allocatable :: LevelShift(:)    ! only used for P%yes_LevelShift
    ! unrestricted
    integer, allocatable :: n_occ_a(:), n_occ_b(:)           ! integer occupations
    real(long), allocatable :: f_occ_a(:), f_occ_b(:)         ! fractional occupations
    real(long), allocatable :: H_a(:, :), H_b(:, :)           ! one particle hamiltonian
    real(long), allocatable :: LC_a(:, :), LC_b(:, :)           ! one particle hamiltonian
    real(long), allocatable :: CE_a(:), CE_b(:)            ! orbital energy
    real(long), allocatable :: D_a(:, :), D_b(:, :)
    real(long), allocatable :: D_old_a(:, :), D_old_b(:, :)       ! the density matrix old
    real(long), allocatable :: H_old_a(:, :), H_old_b(:, :)       ! one particle hamiltonian old
    !
    ! Variables for open-shell HF: alpha- and beta-spin, (1:N_basis,1:N_basis)
    !
    real(long), allocatable :: Focka(:, :), Fockb(:, :)
    real(long), allocatable :: Da(:, :), Db(:, :)
    !
    ! data needed for the DIIS: (1:N_basis,1:N_basis,1:P%DIIS_nvector)
    !
    real(long), allocatable :: Errsave(:, :, :)   ! array of previous commutators
    real(long), allocatable :: Hsave(:, :, :)     ! array of previous H
    !
    ! additional data needed for DDIFF (density difference)
    !
    integer :: nsave = 0
    real(long), allocatable :: Dsave(:, :, :)     ! array of previous D
    real(long), allocatable :: v_Jsave(:, :)     ! array of previous coulomb pot
    !
    ! real-space (grid-based) quantities: (1:N_grid)
    !
    real(long), allocatable :: rho(:), rho_old(:)       ! real-space density
    real(long), allocatable :: v_eff(:), v_old(:)       ! real-space potential
    real(long), allocatable :: v_ext(:)                 ! external potential
    real(long), allocatable :: v_J(:)                   ! direct Coulomb pot.
    real(long), allocatable :: v_xc(:), v_xc_old(:)      ! exchange-corr. pot.
    ! unrestricted variants
    real(long), allocatable :: spin_density(:)     ! real-space spin-density
    real(long), allocatable :: v_xc_a(:), v_xc_b(:)      ! exchange-corr. pot.
    real(long), allocatable :: v_eff_a(:), v_eff_b(:)      ! real-space potential
    !
    ! Variables related to initial guess
    !   rho_ini(j) : density at the j-th grid
    !   LC_ini(i,k) : MO coefficient of the i-th basis for the k-th MO
    !   basis_ini(j,i) : basis func. at the j-th grid for the i-th basis func.
    !   STrans(:,:) : overlap matrix to initial guess basis functions
    !
    real(long), allocatable :: rho_ini(:)       ! always used
    real(long), allocatable :: LC_ini(:, :)      ! only used if P%yes_ReadIni
    real(long), allocatable :: LC_ini_b(:, :)    ! only used if P%yes_ReadIni
    real(long), allocatable :: basis_ini(:, :)   ! only used if P%yes_ReadShell
    real(long), allocatable :: STrans(:, :)      ! only used if P%yes_ReadIni
    !
    ! SCF energies: see calculate_energy in scf.f90
    !
    real(long) :: E, E_old, E_old2      ! electronic energy
    real(long) :: d_rho, d_rho_old          ! maximum density variation
    real(long) :: E_nuc                 ! nuc-nuc energy
    real(long) :: E_tot                 ! total energy
    real(long) :: E_ext, E_J, E_x, E_c, E_xc_int, E_Ts, E_sum_orb
    real(long) :: chem_pot              ! chemical potential
    real(long) :: E_long_range(3)       ! electrostatic E corrections
    ! 1: nuc-elec, 2: elec-elec, 3: nuc-nuc
    !
    ! atomic quantities: (1:N_nuc)
    !
    real(long), allocatable :: partial_charge(:)  ! Z - #elec. in atomic density
    real(long), allocatable :: Q_A(:)   ! store elec. den. as a point charge
    real(long), allocatable :: dipole_A(:, :)    ! dipole_A(1:3,i_nuc)
    !
    ! switches to control the behaviour during SCF
    !
    logical :: switch_diis_on, switch_rstrct_on, switch_levelshift_off, switch_soscf_on
    !    use MOM
    logical :: yes_mom
    !
    ! for gradient calculation, (1:N_nuc*3), 3 for x,y,z
    !   grad(:) : total energy gradient
    !   grad_orb(:) : orbital energy gradient
    !
    real(long), allocatable :: grad(:)
    !
    ! for MC calculation
    !   iSCFstate = 0 : default
    !             < 0 : error
    !             = 1 : old (good but not necessarily consistent with B)
    !             = 2 : good
    !   ierr_o, N_iter_o : returns SCF result
    !   avgSqrOvlp_w_refMO_o : returns Overlap with reference MO (or Fidelity?)
    !
    ! sangkil: renaming suggestion
    !   ierr_o -> ierr  ..... is it the same as iSCFstate?
    !   N_iter_o -> N_iter
    !   avgSqrOvlp_w_refMO_o -> avg_overlap
    !
    integer*1 :: iSCFstate = 0
    integer :: ierr_o, N_iter_o
    real(long) :: avgSqrOvlp_w_refMO_o

    real(long) :: prod_overlap                    ! contains the product overlap:  < guess | SCF_result >

    real(long), allocatable :: partialPopulation_M(:) ! Mulliken Population
    real(long), allocatable :: partialPopulation_L(:) ! Lowdin Population
    real(long), allocatable :: partialPopulation_G(:) ! Grid-based Population
    real(long), allocatable :: partialPopulation_G1(:)! Grid-based Population (sharp cut)

    ! expansion coefficients in minimal basis (for GTO basis)
    real(long), allocatable :: Omin(:, :)
    real(long), allocatable :: Smin(:, :)
    integer, allocatable :: min_info(:, :) ! info on minimal basis:
    ! min_info(:,1)=atom index
    ! min_info(:,2)=n
    ! min_info(:,3)=l
    ! min_info(:,4)=m
    real(long), allocatable :: CE_grad(:, :)          ! [B%N_basis][# of direction ( 3*P%N_nuc )]   Orbital energy gradient. By default un-allocated
    logical :: f_occ_given                          ! set to true when the fractional occupation numbers f_occ is explicitly given
    !! XXX real(long),allocatable :: partialPopulation(:)  ! Loewdin or Mulliken population. Removed on 28.07.2015 to replace with 4 types of charges

    integer, allocatable :: acceptable_convergenceAt(:, :) ! first iter at which acceptable convergence is achieved ( dE< acceptable_convergence(1){ /10,/sqrt(10),/1} and  drho< acceptable_convergence(2){/10,/sqrt(10),/1} are checked
    ! this is to be allocated at scf_procedure iff P%SCF_acceptable_convergence is allocated.  Default <0 means un-reached
    logical, allocatable :: apply_looser_threshold(:, :)   ! this is to be allocated at -MC driver code- checking  corresponding acceptable_convergenceAt(:,:) >0

    type(average_occ), allocatable :: av_occ(:)
    real(long), allocatable :: P_not_frozen(:, :) ! Projector into frozen orbitals
    integer, allocatable :: frozen_MO(:) ! frozen MOs indices
    real(long) :: random_scale ! random factor with which the Auger rates are multiplied
    real(long), allocatable :: teint(:) ! transformed two electron integrals
    real(long), allocatable :: teint_three_active(:) ! transformed two electron integrals
    real(long), allocatable :: teint_abij(:) ! transformed two electron integrals (two indices active)
    real(long), allocatable :: teint_abcd(:) ! transformed two electron integrals (4 indices active)
    real(long), allocatable :: teint_aibj(:) ! transformed two electron integrals (two indices active)
    real(long), allocatable :: teint_ccij(:) ! transformed two electron integrals (two indices core)
    real(long), allocatable :: teint_cicj(:) ! transformed two electron integrals (two indices core)
    real(long), allocatable :: teint_cdij(:) ! two core two virtual (for CIS)
    real(long), allocatable :: teint_cidj(:) ! two core two virtual (for CIS)
    real(long), allocatable :: teint_cdei(:) ! three core one virtual
    integer :: Ncore ! number of core orbitals in molecule

    real(long), allocatable :: E_cis(:) ! array of energy excitations for excited states
    real(long), allocatable :: C_cis(:, :) ! coefficients of excited states
    real(long), allocatable :: C_cis_old(:, :) ! CIS coefficients of excited states (old)
    real(long), allocatable :: LC_old(:, :)    ! orbital coefficients (old)
    integer, allocatable :: CIS_excitation(:, :) ! double array describing excited state configuration
    ! describes excitation from orbital (i,1) to (i,2)

    logical :: yes_calculate_CI_overlap
    logical :: yes_reinit
    real(long), allocatable :: E_ci(:) ! array of energy excitations for ci states
    real(long), allocatable :: C_ci(:, :) ! coefficients of excited states
    real(long), allocatable :: O_ci(:, :) ! coefficients of excited states
    real(long), allocatable :: occ_overlap(:) ! overlap of states to main occ in F%occ
    integer, allocatable :: occs_CI(:, :) ! occ of csfs
    real(long), allocatable :: D_CI(:, :) ! opdm for a state
    integer :: state
    logical :: yes_converged ! is this converged?
  end type WF_data
  private
  public :: WF_data, construct_WF_data, purge_WF_data, random_gaussian, &
    print_projection_info, print_density_of_states, construct_overlap2minimalbasis, adapt_wf_nr_basis
contains

  subroutine construct_WF_data(P, A, G, B, F)
    !! create WF_data from P, A, G, and B
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(out) :: F
    integer :: i, j, ierr
    integer :: N_basis, N_grid
    !
    ! set up basic numbers
    !
    if (P%yes_auger_random) then
      F%random_scale = random_gaussian()
    end if
    F%yes_reinit = .False.
    F%f_occ_given = .FALSE.
    N_basis = B%N_basis
    N_grid = G%N_grid
    F%N_basis = N_basis
    F%Nr_basis = B%Nr_basis
    F%N_grid = N_grid
    F%N_nuc = size(A)
    F%N_MO = min(size(P%N_occ), F%Nr_basis)
    ! this assigns the number of core orbitals in the molecule
    F%Ncore = 0
    do i = 1, size(A)
      if (A(i)%charge .le. 2) then ! Z=1,2
        F%Ncore = F%Ncore + 0 ! no core for H,He
      else if (A(i)%charge .le. 10) then  ! Z=3-10
        F%Ncore = F%Ncore + 1 ! K Shell for first row
      else if (A(i)%charge .le. 18) then ! Z=11-18
        F%Ncore = F%Ncore + 1 + 1 + 3 ! K and L1, L2 Shell for second row
      else if (A(i)%charge .le. 29) then ! Z=19-28
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 ! 1s, 2s, 2p, 3s, 3p for third row
      else if (A(i)%charge .le. 30) then ! Z=29-30
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 + 5 ! 1s, 2s, 2p, 3s, 3p, 3d
      else if (A(i)%charge .le. 36) then ! Z=31-36
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 + 5 + 1 ! 1s, 2s, 2p, 3s, 3p, 3d, 4s
      else if (A(i)%charge .le. 48) then ! Z=37-48
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 + 5 + 1 + 3 ! 1s, 2s, 2p, 3s, 3p, 3d, 4s, 4p
      else if (A(i)%charge .le. 54) then ! Z=49-54
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 + 5 + 1 + 3 + 5 ! 1s, 2s, 2p, 3s, 3p, 3d, 4s, 4p, 4d
      else  ! Z=55..
        F%Ncore = F%Ncore + 1 + 1 + 3 + 1 + 3 + 5 + 1 + 3 + 5 + 1 + 3 ! 1s, 2s, 2p, 3s, 3p, 3d, 4s, 4p, 4d, 5s, 5p
      end if
    end do
    if (P%Ncore .ne. -1 .and. P%Ncore .ne. F%Ncore) then
      write (P%OUTPUT, "('# Number of core orbitals is overwritten. value from auto-detection:',i4,x,'specified:',i4)") &
        F%Ncore, P%Ncore
      F%Ncore = P%Ncore
    end if

    !
    ! set up occupation numbers
    !
    allocate (F%n_occ(B%Nr_basis), F%f_occ(B%Nr_basis))
    allocate (F%n_occ_old(B%Nr_basis))
    F%n_occ(:) = 0
    do i = 1, min(size(F%N_occ), size(P%n_occ))
      F%n_occ(i) = P%n_occ(i)     ! copy the occupation string from P
    end do
    ! unrestricted
    if (P%yes_unrestricted) then
      allocate (F%n_occ_a(B%Nr_basis), F%f_occ_a(B%Nr_basis))
      allocate (F%n_occ_b(B%Nr_basis), F%f_occ_b(B%Nr_basis))
    end if
    if (allocated(P%F_occ)) then
      F%f_occ_given = .TRUE.
      F%f_occ(:) = 0
      do i = 1, size(P%F_occ)
        F%F_occ(i) = P%F_occ(i)
      end do
    else
      F%f_occ_given = .FALSE.
    end if

    F%Nopen = count(F%N_occ(:) .eq. 1)
    F%multp = P%multp
    F%csf_num = P%state
    F%state = P%state

    F%yes_MOM = P%yes_MOM

    if (P%yes_sym) then
      allocate (F%irrep_MO(B%Nr_basis))
      if (allocated(P%irrep_MO)) &
        F%irrep_MO = P%irrep_MO
      F%state_irrep = P%state_irrep
    else
      F%state_irrep = -1
    end if
    !
    ! find the highest occupied molecular orbital (HOMO)
    !
    moloop: do i = size(F%n_occ), 1, -1
      if (allocated(P%av_occ)) then
        do j = 1, size(P%av_occ)
          if (any(P%av_occ(j)%MO(:) .eq. i)) exit moloop
        end do
      end if
      if (F%n_occ(i) .ne. 0) exit moloop
    end do moloop
    F%homo_idx = i
    !
    ! memory allocation for N_basis-based quantities
    !
    allocate (F%CE(B%Nr_basis), F%LC(N_basis, B%Nr_basis))
    allocate (F%fidelity(B%Nr_basis))
    allocate (F%D(N_basis, N_basis), F%D_old(F%N_basis, N_basis))
    allocate (F%D_old2(F%N_basis, N_basis))
    allocate (F%H(N_basis, N_basis), F%H_old(N_basis, N_basis))
    allocate (F%H0(N_basis, N_basis))
    allocate (F%H0_MO(B%Nr_basis, B%Nr_basis))
    F%D=ZERO
    F%H=ZERO
    ! unrestricted
    if (P%yes_unrestricted) then
      allocate (F%CE_a(B%Nr_basis))
      allocate (F%CE_b(B%Nr_basis))
      allocate (F%LC_a(N_basis, B%Nr_basis))
      allocate (F%LC_b(N_basis, B%Nr_basis))
      allocate (F%H_a(N_basis, N_basis))
      allocate (F%H_b(N_basis, N_basis))
      allocate (F%H_old_a(N_basis, N_basis))
      allocate (F%H_old_b(N_basis, N_basis))
      allocate (F%D_a(N_basis, N_basis))
      allocate (F%D_b(N_basis, N_basis))
      allocate (F%D_old_a(N_basis, N_basis))
      allocate (F%D_old_b(N_basis, N_basis))
    end if

    if (P%yes_HF .and. (P%yes_openshell .or. P%yes_unrestricted)) then
      allocate (F%Focka(N_basis, N_basis))
      allocate (F%Fockb(N_basis, N_basis))
      allocate (F%Da(N_basis, N_basis))
      allocate (F%Db(N_basis, N_basis))
    else if (P%yes_HF .and. P%yes_valgs_occ) then
      allocate (F%Da(N_basis, N_basis))
      allocate (F%Db(N_basis, N_basis))
    end if

    !
    ! memory allocation for N_nuc-based quantities
    !
    allocate (F%partial_charge(F%N_nuc))
    allocate (F%Q_A(F%N_nuc))
    F%Q_A = ZERO
    allocate (F%dipole_A(3, F%N_nuc))     ! 1:3 --> x, y, z
    F%dipole_A = ZERO
    !
    ! allocate DIIS space
    !
    if (P%yes_unrestricted) then
      allocate (F%Hsave(N_basis, N_basis, 2 * P%DIIS_nvector), stat=ierr)
      allocate (F%Errsave(N_basis, N_basis, 2 * P%DIIS_nvector), stat=ierr)
    else
      allocate (F%Hsave(N_basis, N_basis, P%DIIS_nvector), stat=ierr)
      allocate (F%Errsave(N_basis, N_basis, P%DIIS_nvector), stat=ierr)
    end if
    if (ierr .gt. 0) stop "Allocate Errsave failed"
    !
    ! allocate memory for DDIFF (density difference)
    !
    if (P%yes_DDIFF) then
      allocate (F%Dsave(N_basis, N_basis, P%n_ddiff_save), stat=ierr)
      if (ierr .gt. 0) stop "Allocate Dsave failed"
      allocate (F%v_Jsave(N_grid, P%n_ddiff_save), stat=ierr)
      if (ierr .gt. 0) stop "Allocate v_Jsave failed"
    end if
    !
    ! memory allocation for N_grid-based quantities
    !
    allocate (F%rho(N_grid), F%rho_old(N_grid))
    allocate (F%v_eff(N_grid), F%v_old(N_grid))
    allocate (F%v_ext(N_grid))
    allocate (F%v_J(N_grid))
    allocate (F%v_xc(N_grid))
    !
    ! unrestricted
    if (P%yes_unrestricted) then
      allocate (F%spin_density(N_grid))
      allocate (F%v_xc_a(N_grid))
      allocate (F%v_xc_b(N_grid))
      allocate (F%v_eff_a(N_grid))
      allocate (F%v_eff_b(N_grid))
    end if
    !
    ! memory allocation for initial guess
    !
    allocate (F%rho_ini(N_grid))

    allocate (F%LC_ini(B%N_basis, B%Nr_basis))
    if (P%yes_unrestricted) then
      allocate (F%LC_ini_b(B%N_basis, B%Nr_basis))
      allocate (F%fidelityb(B%Nr_basis))
    end if
    allocate (F%STrans(N_basis, N_basis))
    allocate (F%basis_ini(N_grid, N_basis))

    !
    ! memory allocation for level shifting
    !
    if (P%yes_LevelShift) allocate (F%LevelShift(B%Nr_basis))
    !
    ! memory allocation for gradients: N_nuc * 3 times (x,y,z)
    !
    if (P%yes_gradient) allocate (F%grad(F%N_nuc*3))

    ! copy averaging information giving in P structure

    if (allocated(P%av_occ)) then
      allocate (F%av_occ(size(P%av_occ)))
      do j = 1, size(F%av_occ)
        allocate (F%av_occ(j)%MO(size(P%av_occ(j)%MO)))
        F%av_occ(j)%MO(:) = P%av_occ(j)%MO(:)
      end do
    end if

    if (allocated(P%freeze)) then
      allocate (F%P_not_frozen(N_basis, N_basis))
      allocate (F%frozen_MO(size(P%freeze)))
      F%frozen_MO = P%freeze
    end if

    !
  end subroutine construct_WF_data

  subroutine adapt_WF_Nr_basis(B, F)
    !! in case the number of basis functions has changed dynamically
    !! due to out-projection of high overlap matrix eigenvalues
    !! we here adapt the wavefunction object, so that we can smoothly continue
    !! the calculation
    type(Basis), intent(in)  :: B 
    type(WF_data), intent(inout) :: F
    integer :: i, c, j, n, cOld, cMin, cMax
    integer, allocatable :: old_irrep_MO(:), ndx(:), ndxOld(:)
    real(long), allocatable :: old_LC(:, :), old_LC_ini(:, :), old_LC_a(:, :), old_LC_b(:, :), old_H0_MO(:, :)
    real(long), allocatable :: old_CE(:), old_CE_a(:), old_CE_b(:), old_levelshift(:), old_fidelity(:)
    if (allocated(F%irrep_MO)) then
      allocate (old_irrep_MO(size(F%irrep_MO, 1)))
      old_irrep_MO = F%irrep_MO
      call update_int_1array(F%irrep_MO, B%Nr_basis, 0)
      do i = 1, B%nirrep
        c = count(B%irrep_sym_bfr(:) .eq. i)
        cOld = count(old_irrep_MO(:) .eq. i)
        n = 0
        j = 1
        do
          if (j .gt. size(old_irrep_MO, 1) .and. F%irrep_MO(j) .eq. 0) then
            F%irrep_MO(j) = i
            n = n + 1
          else if (old_irrep_MO(j) .eq. i .or. F%irrep_MO(j) .eq. 0) then
            F%irrep_MO(j) = i
            n = n + 1
          end if
          if (n .eq. c) exit
          j = j + 1
          if (j .gt. size(F%irrep_MO, 1)) then
            exit
          end if
        end do
      end do
      allocate (old_CE(size(F%CE, 1)))
      old_CE = F%CE
      allocate (old_LC(size(F%LC, 1), size(F%LC, 2)))
      old_LC = F%LC
      allocate (old_LC_ini(size(F%LC_ini, 1), size(F%LC_ini, 2)))
      old_LC_ini = F%LC_ini
      allocate (old_fidelity(size(F%fidelity, 1)))
      old_fidelity = F%fidelity
      allocate (old_H0_MO(size(F%H0_MO, 1), size(F%H0_MO, 2)))
      old_H0_MO = F%H0_MO
      if (allocated(F%LevelShift)) then
        allocate (old_LevelShift(size(F%LevelShift, 1)))
        old_levelshift = F%levelshift
      end if
      if (allocated(F%CE_a)) then
        allocate (old_CE_a(size(F%CE_a, 1)))
        old_CE_a = F%CE_a
        allocate (old_CE_b(size(F%CE_b, 1)))
        old_CE_b = F%CE_b
        allocate (old_LC_a(size(F%LC_a, 1), size(F%LC_a, 2)))
        old_LC_a = F%LC_a
        allocate (old_LC_b(size(F%LC_b, 1), size(F%LC_b, 2)))
        old_LC_b = F%LC_b
      end if
    end if

    call update_int_1array(F%N_occ, B%Nr_basis, 0)
    call update_int_1array(F%N_occ_old, B%Nr_basis, 0)
    call update_int_1array(F%N_occ_a, B%Nr_basis, 0)
    call update_int_1array(F%N_occ_b, B%Nr_basis, 0)
    call update_real_1array(F%f_occ, B%Nr_basis, ZERO)
    call update_real_1array(F%CE, B%Nr_basis, ZERO)
    call update_real_2array(F%LC, F%N_basis, B%Nr_basis, ZERO)
    call update_real_2array(F%LC_ini, F%N_basis, B%Nr_basis, ZERO)
    call update_real_1array(F%fidelity, B%Nr_basis, ZERO)
    call update_real_2array(F%H0_MO, B%Nr_basis, B%Nr_basis, ZERO)
    call update_real_1array(F%CE_a, B%Nr_basis, ZERO)
    call update_real_1array(F%CE_b, B%Nr_basis, ZERO)
    call update_real_2array(F%LC_a, F%N_basis, B%Nr_basis, ZERO)
    call update_real_2array(F%LC_b, F%N_basis, B%Nr_basis, ZERO)
    call update_real_1array(F%LevelShift, B%Nr_basis, ZERO)

    ! for irrep_MO and LC_ini it's more complicated because we cannot cut away the last entry or add some default values at the end
    if (allocated(F%irrep_MO)) then
      do i = 1, B%nirrep
        c = count(B%irrep_sym_bfr(:) .eq. i)
        cOld = count(old_irrep_MO(:) .eq. i)
        allocate (ndxOld(cOld))
        allocate (ndx(c))
        ndxOld = pack([(j, j=1, size(old_irrep_MO, 1))], old_irrep_MO .eq. i)
        ndx = pack([(j, j=1, size(B%irrep_sym_bfr(:), 1))], F%irrep_MO .eq. i)
        ! move LC_ini keeping symmetry constraints
        cMin = min(c, cOld)
        cMax = max(c, cOld)
        F%LC(:, ndx(:)) = ZERO
        F%LC_ini(:, ndx(:)) = ZERO
        F%fidelity(ndx(:)) = ZERO
        F%H0_MO(ndx(:), ndx(:)) = ZERO
        F%LC(:, ndx(:cMin)) = old_LC(:, ndxOld(:cMin))
        F%LC_ini(:, ndx(:cMin)) = old_LC_ini(:, ndxOld(:cMin))
        F%fidelity(ndx(:cMin)) = old_fidelity(ndxOld(:cMin))
        F%H0_MO(ndx(:cMin), ndx(:cMin)) = old_H0_MO(ndx(:cMin), ndx(:cMin))
        if (allocated(F%LevelShift)) then
          F%LevelShift(ndx(:)) = ZERO
          F%LevelShift(ndx(:cMin)) = old_LevelShift(ndxOld(:cMin))
        end if
        if (allocated(F%CE_a)) then
          F%CE_a(ndx(:)) = ZERO
          F%CE_b(ndx(:)) = ZERO
          F%LC_a(:, ndx(:)) = ZERO
          F%LC_b(:, ndx(:)) = ZERO
          F%CE_a(ndx(:cMin)) = old_CE_a(ndxOld(:cMin))
          F%CE_b(ndx(:cMin)) = old_CE_b(ndxOld(:cMin))
          F%LC_a(:, ndx(:cMin)) = old_LC_a(:, ndxOld(:cMin))
          F%LC_b(:, ndx(:cMin)) = old_LC_b(:, ndxOld(:cMin))
        end if
        deallocate (ndx)
        deallocate (ndxOld)
      end do
      deallocate (old_CE)
      deallocate (old_LC)
      deallocate (old_LC_ini)
      deallocate (old_fidelity)
      deallocate (old_H0_MO)
      if (allocated(F%CE_a)) then
        deallocate (old_CE_a)
        deallocate (old_CE_b)
        deallocate (old_LC_a)
        deallocate (old_LC_b)
      end if
      if (allocated(F%LevelShift)) deallocate (old_LevelShift)
      deallocate (old_irrep_MO)
    end if
    F%Nr_basis = B%Nr_basis

    if (allocated(F%irrep_MO)) then
      F%LC_ini = B%sym_bf(:, :) .tm.F%LC_ini
      do i = 1, B%Nr_basis
        do j = 1, B%N_basis
          if (F%irrep_MO(i) .ne. B%irrep_sym_bf(j)) then
            if (abs(F%LC_ini(j, i)) .gt. 1e-10) then
              write (STDERR, '("# Warning unsymmetric orbital guess ",i4,i4,es15.7)') i, j, F%LC_ini(j, i)
              stop
            end if
          end if
        end do
      end do
      F%LC_ini = B%sym_bf(:, :) .m.F%LC_ini
    end if

  contains
    subroutine update_int_1array(array, len, default)
      integer, allocatable, intent(inout) :: array(:)
      integer, intent(in) :: len
      integer, intent(in), optional :: default
      integer :: def
      integer, allocatable :: newarray(:)
      integer :: lenOld
      def = 0
      if (present(default)) def = default
      if (.not. allocated(array)) return
      lenOld = size(array, 1)
      allocate (newarray(len))
      newarray(:) = def
      newarray(:min(len, lenOld)) = array(:min(len, lenOld))
      call move_alloc(newarray, array)
    end subroutine update_int_1array

    subroutine update_real_1array(array, len, default)
      real(long), allocatable, intent(inout) :: array(:)
      real(long), allocatable :: newarray(:)
      integer, intent(in) :: len
      real(long), intent(in), optional :: default
      real(long) :: def
      integer :: lenOld
      def = ZERO
      if (present(default)) def = default
      if (.not. allocated(array)) return
      lenOld = size(array, 1)
      allocate (newarray(len))
      newarray(:) = def
      newarray(:min(len, lenOld)) = array(:min(len, lenOld))
      call move_alloc(newarray, array)
    end subroutine update_real_1array

    subroutine update_real_2array(array, len1, len2, default)
      real(long), allocatable, intent(inout) :: array(:, :)
      real(long), allocatable :: newarray(:, :)
      integer, intent(in) :: len1, len2
      real(long), intent(in), optional :: default
      real(long) :: def
      integer :: len1Old, len2Old
      def = ZERO
      if (present(default)) def = default
      if (.not. allocated(array)) return
      len1Old = size(array, 1)
      len2Old = size(array, 2)
      allocate (newarray(len1, len2))
      newarray(:, :) = def
      newarray(:min(len1, len1Old), :min(len2, len2Old)) = array(:min(len1, len1Old), :min(len2, len2Old))
      call move_alloc(newarray, array)
    end subroutine update_real_2array

  end subroutine adapt_WF_Nr_basis

  function random_gaussian() result(rn)
    real(long) :: x1, x2, r, w, rn
    ! Box-Muller algorithm
    do
      call RANDOM_NUMBER(r)
      x1 = (TWO*r - ONE)
      call RANDOM_NUMBER(r)
      x2 = (TWO*r - ONE)
      w = x1**2 + x2**2
      if (w .lt. ONE) exit
    end do
    w = sqrt((-TWO*log(w))/w)
    rn = x1*w
  end function random_gaussian

  subroutine print_density_of_states(F, OUTPUT, dE_eV)
    !! print out density of states (DOS)
    type(WF_data), intent(in) :: F
    integer, intent(in) :: OUTPUT
    real(long), intent(in), optional :: dE_eV
    integer :: E, E_min, E_max, dE        ! unit: eV, we are using INTEGER
    integer, allocatable :: E_bin(:)      ! unit: eV
    integer :: i, j, N
    dE = 1        ! 1 eV as default
    if (present(dE_eV)) dE = int(dE_eV)
    if (dE .eq. 0) call error_message("dE_eV should be > 0!")
    E_min = int(minval(F%CE(:))*AU2EV)
    E_max = int(maxval(F%CE(:))*AU2EV)
    N = int(dble(E_max - E_min)/dble(dE))
    allocate (E_bin(0:N))
    E_bin(:) = 0
    do i = 1, size(F%CE)
      E = int(F%CE(i)*AU2EV)
      j = int(dble(E - E_min)/dble(dE))
      E_bin(j) = E_bin(j) + 1
    end do
    write (OUTPUT, "('# Density of states (DOS):')")
    write (OUTPUT, "('#',a9,a10)") "E(eV)", "count"
    do j = 0, N
      write (OUTPUT, "(2i10)") E_min + j*dE, E_bin(j)
    end do
    write (OUTPUT, "('#',a9,i10)") "sum:", sum(E_bin(0:N))
    write (OUTPUT, "(/)")
    deallocate (E_bin)
  end subroutine print_density_of_states

  subroutine purge_WF_data(F)
    !! purges Wave function data
    type(WF_data), intent(inout) :: F
    integer :: j
    if (allocated(F%teint)) deallocate (F%teint)
    if (allocated(F%teint_abij)) deallocate (F%teint_abij)
    if (allocated(F%teint_aibj)) deallocate (F%teint_aibj)
    if (allocated(F%teint_cdij)) deallocate (F%teint_cdij)
    if (allocated(F%teint_cidj)) deallocate (F%teint_cidj)
    if (allocated(F%E_cis)) deallocate (F%E_cis)
    if (allocated(F%C_cis)) deallocate (F%C_cis)
    if (allocated(F%C_cis_old)) deallocate (F%C_cis_old)
    if (allocated(F%LC_old)) deallocate (F%LC_old)
    if (allocated(F%CIS_excitation)) deallocate (F%CIS_excitation)

    deallocate (F%n_occ)
    deallocate (F%n_occ_old)
    deallocate (F%f_occ)

    deallocate (F%CE)
    deallocate (F%LC)

    deallocate (F%fidelity)
    deallocate (F%D)
    deallocate (F%D_old)
    deallocate (F%D_old2)
    deallocate (F%H)
    deallocate (F%H_old)
    deallocate (F%H0)
    deallocate (F%H0_MO)

    if (allocated(F%partial_charge)) deallocate (F%partial_charge)
    if (allocated(F%Q_A)) deallocate (F%Q_A)
    if (allocated(F%dipole_A)) deallocate (F%dipole_A)
    if (allocated(F%Focka)) deallocate (F%Focka)
    if (allocated(F%Fockb)) deallocate (F%Fockb)
    if (allocated(F%Da)) deallocate (F%Da)
    if (allocated(F%Db)) deallocate (F%Db)

    !
    ! deallocate DIIS space
    !
    deallocate (F%Hsave)
    deallocate (F%Errsave)

    !
    ! deallocate memory for DDIFF (density difference)
    !
    if (allocated(F%Dsave)) deallocate (F%Dsave)
    if (allocated(F%Dsave)) deallocate (F%v_Jsave)
    !
    ! memory allocation for N_grid-based quantities
    !
    deallocate (F%rho)
    deallocate (F%rho_old)
    deallocate (F%v_eff)
    deallocate (F%v_old)
    deallocate (F%v_ext)
    deallocate (F%v_J)
    deallocate (F%v_xc)
    !
    ! memory allocation for initial guess
    !
    deallocate (F%rho_ini)
    if (allocated(F%LC_ini)) deallocate (F%LC_ini)
    if (allocated(F%STrans)) deallocate (F%STrans)
    if (allocated(F%basis_ini)) deallocate (F%basis_ini)

    !
    ! memory allocation for level shifting
    !
    if (allocated(F%LevelShift)) deallocate (F%LevelShift)
    !
    ! memory allocation for gradients: N_nuc * 3 times (x,y,z)
    !
    if (allocated(F%grad)) deallocate (F%grad)

    ! symmetry stuff
    if (allocated(F%irrep_MO)) deallocate (F%irrep_MO)

    if (allocated(F%av_occ)) then
      do j = 1, size(F%av_occ)
        deallocate (F%av_occ(j)%MO)
      end do
      deallocate (F%av_occ)
    end if

    if (allocated(F%frozen_MO)) then
      deallocate (F%P_not_frozen)
      deallocate (F%frozen_MO)
    end if

    if (allocated(F%Omin)) &
      deallocate (F%Omin)
    if (allocated(F%Smin)) &
      deallocate (F%Smin)
    if (allocated(F%min_info)) &
      deallocate (F%min_info)

    if (allocated(F%E_ci)) &
      deallocate (F%E_ci)
    if (allocated(F%C_ci)) &
      deallocate (F%C_ci)
    if (allocated(F%D_ci)) &
      deallocate (F%D_ci)
    if (allocated(F%O_ci)) &
      deallocate (F%O_ci)
    if (allocated(F%occs_ci)) &
      deallocate (F%occs_ci)

  end subroutine purge_WF_data

  subroutine construct_overlap2minimalbasis(P, A, G, B, F)
    !! prepares the overlap matrix of the current molecular orbitals
    !! to a minimal orbital set based on numerical atomic orbitals    
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F

    type(Atom_set), allocatable :: Amin(:)
    type(Grid3D) :: Gmin
    type(Basis) :: Bmin
    type(Param) :: Pmin
    real(long), allocatable :: O(:, :)
    real(long), allocatable :: M(:, :)
    logical :: file_exists
    integer :: i, j
    Pmin = P
    do i = 1, P%N_nuc
      Pmin%gto_file_path = ""
      Pmin%yes_gaussian = .False.
      Pmin%nao_file_nuc(i) = trim(Pmin%nuc_name(i))//".nao"
      INQUIRE (FILE=Pmin%nao_file_nuc(i), EXIST=file_exists)
      if (.not. file_exists) &
        Pmin%nao_file_nuc(i) = trim(ATOMIC_NAME(P%charge_nuc(i)))//".nao"
      Pmin%x_nuc(i) = A(i)%x
      Pmin%y_nuc(i) = A(i)%y
      Pmin%z_nuc(i) = A(i)%z
      if (trim(P%nao_name(i)) .eq. 'minimal') then
        call put_minimal_basis_set(Pmin%charge_nuc(i), Pmin%N_nao_table(0:, i))
      else
        call put_basis_function(Pmin%nao_name(i), Pmin%N_nao_table(0:, i))
      end if
    end do
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Contructing minimal basis from nao:")')
    call construct_Atom_set(Pmin, Amin)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# ",a10,x,a15,x,a15,x,a15)') "atom", "nao file", "tdipole file", "auger file"
    do i = 1, P%N_nuc
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# ",a10,x,a15,x,a15,x,a15)') &
        trim(Pmin%nuc_name(i)), trim(Pmin%nao_file_nuc(i)), trim(A(i)%tdipole_file_nuc), trim(A(i)%auger_file_nuc)
      if (Amin(i)%N_r .ne. A(i)%N_r .or. &
          Amin(i)%mapping_L .ne. A(i)%mapping_L .or. &
          Amin(i)%rmax .ne. A(i)%rmax) then
        write (P%OUTPUT, '("# grid parameters for atom ",x,a,x,"different: ")') trim(Pmin%nuc_name(i))
        write (P%OUTPUT, '("#   minimal basis from file :",x,a)') trim(Pmin%nao_file_nuc(i))
        if (trim(P%nao_file_nuc(i)) .eq. '') then
          write (P%OUTPUT, '("#   original basis          :",x,a)') &
            "gaussians basis set"
        else
          write (P%OUTPUT, '("#   original basis from file:",x,a)') &
            trim(P%nao_file_nuc(i))
        end if
        write (P%OUTPUT, '("#   ",a10,x,a15,x,a15)') "", "minimal basis", "original basis"
        write (P%OUTPUT, '("#   ",a10,x,i15,x,i15)') "N_r = ", Amin(i)%N_r, A(i)%N_r
        write (P%OUTPUT, '("#   ",a10,x,f15.5,x,f15.5)') "L = ", Amin(i)%mapping_L, A(i)%mapping_L
        write (P%OUTPUT, '("#   ",a10,x,f15.5,x,f15.5)') "rmax = ", Amin(i)%rmax, A(i)%rmax
      end if ! if Amin(i)%N_r /= A(i)%N_r .or. ...
    end do
    call construct_Grid3D(Pmin, Amin, Gmin)
    call construct_Basis(Amin, Gmin, Bmin, Pmin%EPS_overlap, Pmin%EPS_overlap_single, yes_sym=P%yes_sym, yes_gradient=.False.)
    if (size(G%x) .ne. size(Gmin%x) .or. &
        size(G%y) .ne. size(Gmin%y) .or. &
        size(G%z) .ne. size(Gmin%z)) then
      stop "different grid size"
    end if
    if (ANY(G%x .ne. Gmin%x) .or. ANY(G%y .ne. Gmin%y) .or. ANY(G%z .ne. Gmin%z)) then
      stop "different grid points"
    end if
    call print_calc_info(Pmin, Amin, Gmin, Bmin)
    allocate (O(Bmin%N_basis, B%N_basis))
    allocate (M(Bmin%N_basis, B%N_basis))
    do i = 1, Bmin%N_basis
      do j = 1, B%N_basis
        O(i, j) = sum(Bmin%func(:, i)*B%func(:, j)*G%weight(:))
      end do
    end do

    do i = 1, Bmin%N_basis
      do j = 1, B%N_basis
        M(i, j) = sum(Bmin%SInv(i, :)*O(:, j))
      end do
    end do

    if (.not. allocated(F%Omin)) &
      allocate (F%Omin(Bmin%N_basis, B%N_basis))
    if (.not. allocated(F%min_info)) &
      allocate (F%min_info(Bmin%N_basis, 5))
    if (.not. allocated(F%Smin)) &
      allocate (F%Smin(Bmin%N_basis, Bmin%N_basis))
    F%min_info(:, :) = Bmin%info(:, :)
    F%Omin = matmul(M, F%LC)
    F%Smin = Bmin%S

    deallocate (M)
    deallocate (O)

    call purge_Grid3D(Gmin)
    call purge_Basis(Bmin)

  end subroutine construct_overlap2minimalbasis

  subroutine print_projection_info(info, Omin, Smin, atom, IOUT)
    !! print projection information
    integer, intent(in) :: info(:, :)
    real(long), intent(in) :: Omin(:, :)
    real(long), intent(in) :: Smin(:, :)
    integer, intent(in) :: atom
    integer, intent(in) :: IOUT

    integer, allocatable :: index_bf(:)
    integer :: i, j
    allocate (index_bf(count(info(:, 1) .eq. atom)))
    index_bf = pack((/(i, i=1, size(info, 1))/), info(:, 1) .eq. atom)
    write (IOUT, '("# for atom ",i5)', advance="yes") atom
    write (IOUT, '("#",a15)', advance="no") "relevant orbs. "
    do j = 1, size(Omin, 2)
      if (sum(Omin(index_bf(:), j)**2) .gt. 0.1) then
        write (IOUT, '(x,a4,i4)', advance="no") "MO", j
      end if
    end do
    write (IOUT, '()')
    write (IOUT, '("#",a15)', advance="no") "norm in atom"
    do j = 1, size(Omin, 2)
      if (sum(Omin(index_bf(:), j)**2) .gt. 0.1) then
       write (IOUT, '(x,f8.5)', advance="no") sum(matmul(Smin(index_bf(:), index_bf(:)), Omin(index_bf(:), j))*Omin(index_bf(:), j))
      end if
    end do
    write (IOUT, '()')
    write (IOUT, '("#",a15)', advance="no") "norm (minbasis)"
    do j = 1, size(Omin, 2)
      if (sum(Omin(index_bf(:), j)**2) .gt. 0.1) then
        write (IOUT, '(x,f8.5)', advance="no") sum(matmul(Smin(:, :), Omin(:, j))*Omin(:, j))
      end if
    end do
    write (IOUT, '()')
    do i = 1, size(Omin, 1)
      write (IOUT, '("#",a3,x,i4,x,i2,x,i2,x)', advance="no") "min", info(i, 1), info(i, 3), info(i, 4)
      do j = 1, size(Omin, 2)
        if (sum(Omin(index_bf(:), j)**2) .gt. 0.1) then
          write (IOUT, '(x,f8.3)', advance="no") Omin(i, j)
        end if
      end do
      write (IOUT, '()')
    end do
    deallocate (index_bf)
  end subroutine print_projection_info

end module wavefunction
