!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!
!! main program file of XMOLECULE

program xmolecule
  !! author: Ludger Inhester
  !! author: Sang-Kil Son
  !! author: Yajiang Hai 
  !! author: Kota Hanasaki
  !! author: Caroline Arnold
  !! main program of XMOLECULE: an integrated toolkit for X-ray and molecular physics
  
  use read_param
  use atom_in_mol
  use molecular_grid
  use basis_func
  use wavefunction
  use process
  use augerrates
  use photoionization
  use fluorescence
  use absorption
  use population
  use scf
  use nuclear_derivatives
  use scattering
  use mcscf
  use gradient
  use cis
  use cubefile
  use molden
  use stdio
  implicit none
  type(Param) :: P !! parameters
  type(Atom_set), allocatable :: A(:) !! atom structures
  type(Grid3D) :: G !! molecular grid
  type(Basis) :: B !! basis functions
  type(WF_data) :: F !! wave function data
  type(Process_data) :: S !! process data

  character(len=2048) :: filename !! filename variable
  integer :: i !!
  real :: T_time(0:14), time_sum        !! CPU time table --------------------
  integer, parameter :: T_PRT = 1       !!   read and print data
  integer, parameter :: T_A = 2       !!   construct atom set
  integer, parameter :: T_MG = 3       !!   construct molecular grid
  integer, parameter :: T_B = 4       !!   construct basis set
  integer, parameter :: T_TEINT = 5       !!   precalculate GTO integrals
  integer, parameter :: T_WF = 6       !!   construct WF_data
  integer, parameter :: T_SCF = 7       !!   SCF procedure
  integer, parameter :: T_CI = 8       !!   CI procedure
  integer, parameter :: T_R = 9       !!   calculate rate and cross section
  integer, parameter :: T_POP = 10      !!   population analysis
  integer, parameter :: T_GR = 11      !!   gradient calculation
  integer, parameter :: T_OG = 12      !!   orbital gradient calculation
  integer, parameter :: T_S = 13      !!   scattering calculation
  integer, parameter :: T_CS = 14      !!   compton calculation
  call init_time()
  call init_checkpoint(T_time)

  !
  ! read parameters
  !
  call read_parameters(P, "static")
  call print_parameters(P, "electronic structure of molecules")
  call checkpoint(T_time, T_PRT)      ! CPU time: read and print parameters
  !
  ! construct molecular coordinate
  !
  call init_index_packed()
  call construct_Atom_set(P, A)
  call checkpoint(T_time, T_A)        ! CPU time: construct atom set
  call construct_Grid3D(P, A, G)
  call checkpoint(T_time, T_MG)       ! CPU time: construct molecular grid
  if (P%yes_periodic) call find_neighboring_atoms(G, A, P%OUTPUT)
  call checkpoint(T_time, T_PRT)      ! CPU time: print data
  if (P%yes_gaussian) then
    call construct_Basis(A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym, gto_file_path=P%gto_file_path, &
      yes_gradient=P%yes_gradient, basisSetOverlapLim=P%basisSetOverlapLim)
  else
    call construct_Basis(A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym, yes_gradient=P%yes_gradient, &
      basisSetOverlapLim=P%basisSetOverlapLim)
  end if
  call checkpoint(T_time, T_B)        ! CPU time: construct basis set
  call print_calc_info(P, A, G, B)
  if (P%yes_gaussian .and. .NOT. P%yes_direct_scf) then
    write (P%OUTPUT, '("# calculating two electron integrals (gto): ...")')
    call precalculate_twoel_GTO(P, B)
    call checkpoint(T_time, T_TEINT)    ! CPU time: pre-calculate GTO integrals
    write (P%OUTPUT, '("# done")')
    if (P%yes_verbose) &
      write (P%OUTPUT, "('# time ',f10.2,' secs')") T_time(T_TEINT)
  end if

  call construct_WF_data(P, A, G, B, F)
  call checkpoint(T_time, T_WF)       ! CPU time: construct WF_data

  !
  ! to check constructions
  !
  !call dump_Atom_set( A )
  !call dump_Grid3D( G )
  !call dump_Basis( B )

  if (P%yes_MCSCF) then
    if (.not. associated(B%teint) .and. .not. P%yes_direct_scf) then
      write (P%OUTPUT, '("# calculating two electron integrals: (grid)...")')
      call precalculate_twoel(P, A, G, B)
      call checkpoint(T_time, T_TEINT)  ! CPU time: pre-calculate twoel integrals
    end if
    call MCSCF_procedure(P, A, G, B, F)
  else if (P%yes_HF) then
    if (.not. associated(B%teint) .and. .not. P%yes_direct_scf) then
      write (P%OUTPUT, '("# calculating two electron integrals: (grid)...")')
      call precalculate_twoel(P, A, G, B)
      call checkpoint(T_time, T_TEINT)  ! CPU time: pre-calculate twoel integrals
    end if
    call SCF_procedure(P, A, G, B, F)
!  else if(P%yes_unrestricted) then
!    call SCF_procedure_unrestricted( P, A, G, B, F )
  else
    call SCF_procedure(P, A, G, B, F)
  end if
  call checkpoint(T_time, T_SCF)      ! CPU time: SCF procedure

  if (.not. F%yes_converged) then
    stop 'not converged'
  end if

  if (P%yes_CIS) then
    call CIS_procedure(P, A, G, B, F)
  else if (P%yes_CI) then
    call CI_procedure(P, A, B, F, OUTPUT=P%OUTPUT)
    !call test_CI_procedure( P, A, G, B, F, OUTPUT=P%OUTPUT)
  end if
  call checkpoint(T_time, T_CI)  ! CPU: CI procedure (P%yes_CI .or. P%yes_CIS)

  !
  ! print out DOS if the periodic calculation has been done
  !
  if (P%yes_periodic) call print_density_of_states(F, P%OUTPUT)
  !
  ! print out atomic potentials or atomic densities, which are
  !   spherically averaged and constructed from the molecular pot / den.
  !
  if (P%yes_plot_potential) then
    write (P%OUTPUT, "('# Atomic potential for each atom')")
    call print_atomic_quantity(A, G, F%v_eff, P%OUTPUT)
  end if
  if (P%yes_plot_density) then
    write (P%OUTPUT, "('# Atomic density for each atom')")
    call print_atomic_quantity(A, G, F%rho, P%OUTPUT)
  end if
  call checkpoint(T_time, T_PRT)      ! CPU time: read and print parameters

  !
  ! dump transformed two-body electron integrals
  !
  if (P%yes_dump_ints) then
    if (.NOT. associated(B%teint)) then
      call precalculate_twoel(P, A, G, B)
      call checkpoint(T_time, T_TEINT)
    end if
    F%H0_MO = F%LC(:, :) .tm.F%H0(:, :) .m.F%LC(:, :)
    call transform_twoel(P, B, F, P%OUTPUT)
    open (unit=55, file="INTDUMP", status='UNKNOWN', ACTION='WRITE')
    call dump_twoel(P, F, 55)
    close (55)
    call checkpoint(T_time, T_TEINT)    ! CPU time: pre-calculate GTO integrals
  end if

  !
  ! cross section or rate calculation
  !
  if (P%yes_auger .or. P%yes_fluorescence .or. P%yes_pcs .or. P%yes_absorption) then
    if ((P%yes_auger .or. P%yes_pcs) .and. P%yes_gaussian) then
      call construct_overlap2minimalbasis(P, A, G, B, F)
    end if
    if (P%yes_auger_random) then
      call random_seed()
    end if
    S = create_Process(P%N_occ, P%N_proc_max, spin=P%yes_process_spin, ci=P%yes_process_CI)
    if( P%yes_auger .and. P%yes_process_spin .and. P%yes_average_csf_rates .and. .not. P%yes_process_CI)  call calcAugerRates_spin_avg( P, A, B, F, S )
    if (P%yes_auger .and. P%yes_process_spin .and. .NOT. P%yes_average_csf_rates) call calcAugerRates_spin(P, A, B, F, S)

    if (P%yes_fluorescence .and. P%yes_process_CI) then
      call calc_Fluorescence_Rates_CI(P, A, B, F, G)
    end if
    if (P%yes_auger .and. P%yes_process_CI) then
      call calcAugerRates_CI(P, A, B, F, G)
    end if
    if (P%yes_pcs .and. P%yes_process_CI) then
      call calculate_photoIonization_CS_CI(P, A, B, F, G)
    end if
    if (P%yes_auger .and. .not. P%yes_process_spin .and. .not. P%yes_process_CI) call calcAugerRates(P, A, B, F, S)
    if( P%yes_fluorescence .and. P%yes_process_spin .and. .not. P%yes_average_csf_rates .and. .not. P%yes_process_CI) call calc_Fluorescence_Rates_spin( P, B, F, G, S )
    if (P%yes_fluorescence .and. P%yes_process_spin .and. P%yes_average_csf_rates.and. .not. P%yes_process_CI) call calc_Fluorescence_Rates_spin_avg(P, B, F, G, S)
    if (P%yes_fluorescence .and. .not. P%yes_process_spin .and. .not. P%yes_process_CI) call calc_Fluorescence_Rates(P, B, F, G, S)
    if (P%yes_pcs .and. P%yes_process_spin .and. .not. P%yes_average_csf_rates .and. .not. P%yes_process_CI) call calculate_photoIonization_CS_spin(P, A, B, F, S)
    if (P%yes_pcs .and. P%yes_process_spin .and. P%yes_average_csf_rates .and. .not. P%yes_process_CI) call calculate_photoIonization_CS_spin_avg(P, A, B, F, S)
    if (P%yes_pcs .and. .not. P%yes_process_spin) call calculate_photoIonization_CS(P, A, B, F, S)
    if (P%yes_absorption .and. P%yes_process_CI) then
      call calc_oscillator_strength_CI(P, A, B, F, G)
    else if (P%yes_absorption .and. P%yes_CIS) then
      call calc_oscillator_strength_CIS(P, A, B, F, G)
    else if (P%yes_absorption .and. .not. P%yes_process_spin) then
      call calc_oscillator_strength(P, B, F, G)
    end if

    if (.not. P%yes_process_CI) then
      if (P%yes_average_rates .or. P%yes_average_csf_rates) &
        call sum_equivalent_configuration(S)
      open (unit=55, file=P%file_data, status='UNKNOWN', ACTION='WRITE', position='append')
      call print_Process(S, 55)
      close (55)
    end if
  end if
  if (P%yes_stability) &
    call stability_analysis(P, A, G, B, F)

  ! shake up
  !
  if (P%yes_shakeup .and. P%yes_ReadIni) then
    call shakeup_analysis(P%OUTPUT, P, B, F)
  end if

  call checkpoint(T_time, T_R)        ! CPU time: calc. rates & cross sections

  !
  ! atomic population analysis: Mulliken and Loewdin
  !
  if (P%yes_population) then
    if (P%yes_MCSCF .or. P%yes_CI .or. P%yes_CIS) then
      call mulliken_population_analysis(A, B, F, P%OUTPUT, D_given=F%D_CI)
      call loewdin_population_analysis(A, B, F, P%OUTPUT, D_given=F%D_CI)
      call density_analysis(P, G, B, F%D_CI, P%OUTPUT)
    else
      call mulliken_population_analysis(A, B, F, P%OUTPUT)
      call loewdin_population_analysis(A, B, F, P%OUTPUT)
      call atomic_density_population_analysis(G, A, F, P%OUTPUT)
      call density_analysis(P, G, B, F%D, P%OUTPUT)
      if (P%yes_openshell .or. (P%yes_valgs_occ .and. any(F%N_occ==1))) then
        call spin_analysis(P, G, B, F%Da, F%Db, P%OUTPUT)
      end if
    end if
    call Gross_population_analysis(P, A, B, F, P%OUTPUT)
    if (P%yes_population_L) then
      call Gross_population_analysis(P, A, B, F, P%OUTPUT, sym_l=0)
      call Gross_population_analysis(P, A, B, F, P%OUTPUT, sym_l=1)
      call Gross_population_analysis(P, A, B, F, P%OUTPUT, sym_l=2)
    end if
    call bondOrder_analysis(P, B, F%D, P%OUTPUT)
    call fragment_analysis(P, B, F, P%OUTPUT)
  end if
  call checkpoint(T_time, T_POP) ! CPU: population analysis (P%yes_population)

  !
  ! SCF energy gradient
  !
  if (P%yes_gradient) then
    if (P%yes_gaussian) then
      if (P%yes_valgs_occ) then
        if (count(F%N_occ .eq. 1) .eq. 1 .and. sum(F%N_occ) + 1 .eq. sum(A(:)%charge)) then
          call calculate_hole_energies_gradients_nacs(P, A, B, F, OUTPUT=P%OUTPUT)
        else if (count(F%N_occ .eq. 1) .eq. 1 .and. sum(F%N_occ) - 1 .eq. sum(A(:)%charge)) then
          call calculate_hole_energies_gradients_nacs(P, A, B, F, particles=.True., OUTPUT=P%OUTPUT)
        else
          call calculate_Koopmann_energy_gradient(P, A, B, F)
        end if
      else if (P%yes_CIS) then
        call calculate_CIS_gradient_GTO(P, A, B, F, state=P%state, yes_calc_nacs=.True., OUTPUT=P%OUTPUT)
      else
        call calculate_energy_gradient_GTO(P, A, G, B, F, OUTPUT=P%OUTPUT)
      end if
    end if
    if (.not. P%yes_gaussian .and. .not. P%yes_HF) then
      call calculate_energy_gradient(P, A, G, B, F)
    end if
  end if
  call checkpoint(T_time, T_GR) ! CPU: gradient calculation (P%yes_gradient)

  if (P%yes_orbital_energy_gradient) then
    if (P%yes_gaussian .and. P%yes_gaussian) then
      call calculate_orbital_gradient_GTO(P, A, B, F, OUTPUT=P%OUTPUT)
    else
      stop "orbital energy gradients only implemented for GTO + HF"
    end if
  end if
  call checkpoint(T_time, T_OG) ! CPU: (P%yes_orbital_energy_gradient)

  !
  ! elastic scattering
  !
  !  if(P%yes_scattering)&
  !    call calc_transform(P,A,B,F,G)
  if (P%yes_scattering) &
    call calc_elastic_scattering(P, G, F)
  call checkpoint(T_time, T_S) ! CPU: scattering (P%yes_scattering)

  !
  ! Waller Hartree scattering
  !
  if (P%yes_compton) &
    call calc_Waller_Hartree_scattering(P, B, G, F)
!  if(P%yes_compton) &
!    call calc_Waller_Hartree_scattering_from_atom(P,A,B,F,G)
  call checkpoint(T_time, T_CS) ! CPU: compton scattering (P%yes_compton)

  if (P%yes_density_cube) then
    if (P%yes_CI) then
      call write_cube_file(P, A, B, F, "density.cube", D=F%D_CI)
    else
      call write_cube_file(P, A, B, F, "density.cube")
    end if
  end if
  if (P%yes_spin_density_cube) then
    call write_cube_file(P, A, B, F, "spin_density.cube", D=F%Da - F%Db)
  end if
  if (allocated(P%mo_cube)) then
    do i = 1, size(P%mo_cube)
      write (filename, '("mo_",i0.4,".cube")') P%mo_cube(i)
      write (P%OUTPUT, '("# writing ",a)') trim(filename)
      call write_cube_file(P, A, B, F, trim(filename), mo=P%mo_cube(i))
    end do
  end if
  call checkpoint(T_time, T_PRT)      ! CPU time: read and print parameters
  if (P%molden_file .ne. "") then
    call write_molden_file(P, A, B, F, P%molden_file)
  end if
  !
  ! CPU time table ----------------------------------------------------------
  !
  time_sum = sum(T_time(1:))
  write (P%OUTPUT, "('# Computational time table:')")
  write (P%OUTPUT, "('# ',a)") repeat("-", 54)
  call write_T("read and print data      ", T_time(T_PRT), time_sum)
  call write_T("construct atom set       ", T_time(T_A), time_sum)
  call write_T("construct molecular grid ", T_time(T_MG), time_sum)
  call write_T("construct basis set      ", T_time(T_B), time_sum)
  if (P%yes_gaussian .and. .NOT. P%yes_direct_scf) &
    call write_T("precalculate teints    ", T_time(T_TEINT), time_sum)
  call write_T("construct WF_data        ", T_time(T_WF), time_sum)
  call write_T("SCF procedure            ", T_time(T_SCF), time_sum)
  if (P%yes_CI .or. P%yes_CIS) &
    call write_T("CI procedure             ", T_time(T_CI), time_sum)
  if (P%yes_auger .or. P%yes_fluorescence .or. P%yes_pcs .or. &
    P%yes_absorption .or. P%yes_stability .or. P%yes_shakeup) &
    call write_T("rate and cross section   ", T_time(T_R), time_sum)
  if (P%yes_population) &
    call write_T("population analysis      ", T_time(T_POP), time_sum)
  if (P%yes_gradient) &
    call write_T("gradient calculation     ", T_time(T_GR), time_sum)
  if (P%yes_orbital_energy_gradient) &
    call write_T("orbital gradient calc.   ", T_time(T_OG), time_sum)
  if (P%yes_scattering) &
    call write_T("scattering calculation   ", T_time(T_S), time_sum)
  if (P%yes_compton) &
    call write_T("compton calculation      ", T_time(T_CS), time_sum)
  write (P%OUTPUT, "('# ',a)") repeat("-", 54)
  write (P%OUTPUT, "('# |CPU_TIME|',17x,a14,f13.2)") &
    timestamp(time_sum), time_sum

  !if ( P%yes_print_wf ) call print_bound( P, F )

  call purge_WF_data(F)
  call purge_Basis(B)
  call purge_Grid3D(G)
  call purge_Atom_set(A)
  call purge_Process(S)
#ifdef __INTEL_MKL__
  call mkl_free_buffers()
#endif
  !
  ! print footnote
  !
  call footnote(P%OUTPUT)
  call purge_parameters(P)
contains
  subroutine write_T(string, time, time_S)
    character(len=*), intent(in) :: string
    real, intent(in) :: time, time_S
    write (P%OUTPUT, "('# - ',a25,a14,4x,'(',f6.2,'%)')") &
      string, timestamp(time), time/time_S*100_long
  end subroutine write_T
end program xmolecule
