!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!
module xmolecule_interface
  !! author: Ludger Inhester
  !!
  !! XMOLECULE_interface: an integrated toolkit for X-ray and molecular physics
  !! subroutines providing an interface to other programs
  !!
  use read_param
  use atom_in_mol
  use molecular_grid
  use wavefunction
  use process
  use augerrates
  use photoionization
  use fluorescence
  use absorption
  use scf
  use rwfile
  use operations
  use gradient
  use cubefile
  use molden
  use cis
  use iso_c_binding
  use mcscf
  use population
  use averaging
  use atom
  use stdio
  implicit none
  type xmol_data
    type(Param), allocatable:: P
    type(Atom_set), allocatable :: A(:)
    type(Grid3D), allocatable :: G
    type(Basis), allocatable :: B
    type(WF_data), allocatable :: F
    type(Process_data), allocatable :: S
  end type xmol_data
  private
  public :: xmolecule_setup, xmolecule_purge, xmolecule_init, &
    xmolecule_get_energy, xmolecule_get_mass, xmolecule_get_atomlabels, &
    xmolecule_orbital_energies, xmolecule_hole_occs_energies_efield, &
    xmolecule_hole_occs_energies_gradients_nacs, xmolecule_part_occs_energies_gradients_nacs, &
    xmolecule_excitation_energies, xmolecule_CIS_energies, xmolecule_set_ci_state, &
    xmolecule_ci_overlaps, xmolecule_CI_energies, xmolecule_get_CIS_gradient, &
    xmolecule_get_CIS_gradient_nacs, xmolecule_get_cis_transdens, xmolecule_get_gradient, &
    xmolecule_get_auger_ci, xmolecule_get_process, xmolecule_get_population, xmolecule_get_ao_labels, &
    xmolecule_get_overlapmatrix, xmolecule_get_mo_coef, xmolecule_set_mo_coef, xmolecule_get_nao_conf, &
    xmolecule_get_mullikencharge, xmolecule_get_loewdincharge, xmolecule_get_cis_mullikenpop, &
    xmolecule_get_cis_loewdinpop, xmolecule_get_bondorder, xmolecule_get_CIS_oscillator_strength, &
    xmolecule_get_oscillator_strength, xmolecule_get_fluorescencerates, xmolecule_write_molden, &
    xmolecule_write_density_cube, xmolecule_write_mo_cube, xmolecule_get_geom, &
    xmolecule_set_calculate_ci_overlap, xmolecule_set_geom, xmolecule_write_chkpt, xmolecule_read_chkpt, &
    xmolecule_get_occ, xmolecule_set_multp, xmolecule_set_irrep, cistate2occ, occ2cistate, &
    xmolecule_set_occ, xmolecule_get_spin_analysis
contains

  subroutine original_averaging(P, av_occ)
    type(Param), intent(in) :: P
    type(average_occ), allocatable, intent(inout) :: av_occ(:)
    integer :: j
    if (allocated(P%av_occ)) then
      do j = 1, size(av_occ)
        deallocate (av_occ(j)%MO)
      end do
      deallocate (av_occ)
      allocate (av_occ(size(P%av_occ)))
      do j = 1, size(av_occ)
        allocate (av_occ(j)%MO(size(P%av_occ(j)%MO)))
        av_occ(j)%MO(:) = P%av_occ(j)%MO(:)
      end do
    end if
  end subroutine original_averaging

  subroutine automatic_averaging(P, B, F, av_occ)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    type(average_occ), allocatable, intent(inout) :: av_occ(:)
    ! energy range in hartree in which orbitals can be considered degenerate
    real(long), parameter :: energy_degenerate_criterion = 0.05
    ! minimal population on an atomic subshell for averaging
    real(long), parameter :: subshell_pop_criterion = 0.8
    integer :: i, j, k, av_occ_index
    integer :: i_a, i_n, i_l
    integer :: j_a, j_n, j_l

    integer, allocatable :: ind(:, :)
    integer, allocatable :: ind_n(:)
    integer :: n
    real(long) :: i_popnl, j_popnl
    integer :: IOUT
    IOUT = P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    allocate (ind(F%N_basis, F%N_basis))
    allocate (ind_n(F%N_basis))

    n = 1
    ind(1, 1) = -1
    ind(:, :) = -1
    do i = 1, F%N_basis ! loop over all MO indices
      if (ANY(ind(:, :) .eq. i)) cycle ! when it is already in some averaging scheme: skip it
      call dominant_subshell_MO(B, F%LC(:, i), i_a, i_n, i_l, i_popnl)

      av_occ_index = -1
      if (allocated(P%av_occ)) then
        do k = 1, size(P%av_occ)
          if (ANY(P%av_occ(k)%MO .eq. i)) then
            av_occ_index = k
          end if
        end do
      end if

      ind_n(n) = 1 ! number of MO's in this group
      ind(:, n) = -1
      ind(1, n) = i ! MO i is in the group n
      do j = i + 1, F%N_basis ! loop over every other MO
        call dominant_subshell_MO(B, F%LC(:, j), j_a, j_n, j_l, j_popnl)
        if (abs(F%CE(i) - F%CE(j)) .lt. energy_degenerate_criterion .and. &
          i_popnl .gt. subshell_pop_criterion .and. j_popnl .gt. subshell_pop_criterion .and. &
          i_a .eq. j_a .and. i_n .eq. j_n .and. i_l .eq. j_l) then  ! MO i and MO j match: add MO j to group n
          ind_n(n) = ind_n(n) + 1
          ind(ind_n(n), n) = j
        else if (.FALSE. .and. allocated(P%av_occ)) then
          if (av_occ_index .ne. -1) then ! MO i is in av_occ group
            if (ANY(P%av_occ(av_occ_index)%MO .eq. j)) then ! and if MO j is in the same P%av_occ group: add MO j to group n
              ind_n(n) = ind_n(n) + 1
              ind(ind_n(n), n) = j
            end if
          end if
        end if
      end do ! j loop
      ! if group n is not trivial (contains more than one MO) increase counter n
      if (ind_n(n) .gt. 1) n = n + 1
    end do ! i loop
    n = n - 1

    ! reallocation of av_occ array
    if (allocated(av_occ)) then
      do j = 1, size(av_occ)
        deallocate (av_occ(j)%MO)
      end do
      deallocate (av_occ)
    end if
    allocate (av_occ(n))
    write (IOUT, "('# Average over MOs: ')")
    do i = 1, n
      write (IOUT, "('# ',99i4)") ind(1:ind_n(i), i)
      allocate (av_occ(i)%MO(ind_n(i)))
      av_occ(i)%MO(1:ind_n(i)) = ind(1:ind_n(i), i)
    end do

    do j = 1, size(F%av_occ)
      F%F_occ(F%av_occ(j)%MO(:)) = ONE/size(F%av_occ(j)%MO)*(sum(F%f_occ(F%av_occ(j)%MO)))
    end do

    deallocate (ind, ind_n)
  end subroutine automatic_averaging

  !> subroutine initial_calculate_electrons(P, A, B, G, F)
  !! calculates the electronic structure
  !! A(:),B,G,F structure has to be prepared in advance
  !! the naos specified in P%file_nao_nuc(:) are used for the calculation
  !! This subroutine can use an initial guess from file
  !! using the options in P structure
  !!
  subroutine initial_calculate_electrons(P, A, B, G, F)
    type(Param), intent(in) :: P
    type(Atom_set), intent(inout) :: A(:)
    type(Basis), intent(inout)  :: B
    type(Grid3D), intent(inout) :: G
    type(WF_data), intent(inout) :: F
    integer :: i
    integer :: IOUT
    IOUT = P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    ! throw away old stuff
    call purge_WF_data(F)
    call purge_Grid3D(G)
    call purge_Basis(B)

    ! set basis function to those specified in P
    do i = 1, size(A)
      A(i)%nao_file_nuc = P%nao_file_nuc(i)
      A(i)%hole_string = ""
      A(i)%conf_string = ""
      A(i)%x = P%x_nuc(i)
      A(i)%y = P%y_nuc(i)
      A(i)%z = P%z_nuc(i)
      if (P%nao_file_nuc(i) .eq. "" .and. P%yes_gaussian) then
        A(i)%N_r = P%N
        A(i)%mapping_L = P%L
        A(i)%rmax = P%rmax
        if (P%charge_nuc(i) .ne. 0) then
          A(i)%nao_file_nuc = trim(ATOMIC_NAME(A(i)%charge))//".nao"
          A(i)%auger_file_nuc = trim(ATOMIC_NAME(A(i)%charge))//".auger"
          A(i)%tdipole_file_nuc = trim(ATOMIC_NAME(A(i)%charge))//".tdipole"
        end if
      end if
    end do
    ! construct new
    if (.not. P%yes_gaussian) &
      call update_Atom_set_with_NAO(A)

    call construct_Grid3D(P, A, G)
    if (P%yes_gaussian) then
      call construct_Basis( A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym ,gto_file_path=P%gto_file_path, yes_gradient=.True., basisSetOverlapLim=P%basisSetOverlapLim)
    else
      call construct_Basis( A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym, yes_gradient=.True., basisSetOverlapLim=P%basisSetOverlapLim )
    end if
    if (P%yes_gaussian .and. .NOT. P%yes_direct_scf) then
      call precalculate_twoel_GTO(P, B)
    end if

    call construct_WF_data(P, A, G, B, F)
    ! do a first SCF with the basis set from P structure
    ! create basis function from them and do an SCF with the
    ! new basis functions.
    ! This ensures that transition amplitudes, basis functions, and solution match
    F%iSCFstate = 0
    call SCF_procedure(P, A, G, B, F)
    F%iSCFstate = 1
    F%LC_ini = F%LC
    if(P%yes_unrestricted) then
      F%LC_ini = F%LC_a
      F%LC_ini_b = F%LC_b
    end if
    F%basis_ini = B%func
    if (P%yes_gaussian .and. P%yes_ionization_dynamics) then
      call construct_overlap2minimalbasis(P, A, G, B, F)
      call calculate_electrons(P, A, B, G, F, set_new_geometry=.FALSE., set_new_naos=.TRUE., freeze=.FALSE.)
    end if
    if (.NOT. P%yes_gaussian) then
      call calculate_electrons(P, A, B, G, F, set_new_geometry=.FALSE., set_new_naos=.TRUE., freeze=.FALSE.)
      !    call make_naos(P,A,B,G,F)
      !    call update_Atom_set_with_NAO( A )
      !    call SCF_procedure( P, A, G, B, F )
    end if
    if (P%yes_auger_random) then
      F%random_scale = random_gaussian()
      write (IOUT, *) '# new random scale:', F%random_scale
    end if

  end subroutine initial_calculate_electrons

  subroutine calculate_electrons_with_partial_freezing(P, A, B, G, F, set_new_geometry)
    type(Param), intent(in) :: P
    type(Atom_set), intent(inout) :: A(:)
    type(Basis), intent(inout)  :: B
    type(Grid3D), intent(inout) :: G
    type(WF_data), intent(inout) :: F
    logical, intent(in), optional :: set_new_geometry
    integer, allocatable :: not_frozen(:)
    integer, allocatable :: partial_freeze(:)
    real(long), allocatable :: LC_save(:, :)
    real(long), allocatable :: B_save(:, :)
    integer :: i, j, n
    logical :: new_geom = .True.
    integer, parameter :: NSTEPS = 10
    integer :: IOUT
    type(WF_data) :: F_save
    logical :: converged
    integer*1 :: ifault
    IOUT = P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    if (present(set_new_geometry)) new_geom = set_new_geometry

    allocate (B_save(G%N_grid, B%N_basis))
    allocate (LC_save(B%N_basis, size(F%LC_ini, 2)))

    B_save = F%basis_ini
    LC_save = F%LC_ini
    converged = .FALSE.

    do n = 1, NSTEPS
      ! keep all MOs frozen
      call calculate_electrons(P, A, B, G, F, &
        set_new_geometry=.true., set_new_naos=.true., freeze=.true.)
      ! select a subset of orbitals
      call select_MO_subset(F, not_frozen)
      ! make an array of frozen MOs
      allocate (partial_freeze(size(F%CE) - size(not_frozen)))
      j = 1
      do i = 1, size(F%CE)
        if (any(i .eq. not_frozen(:))) cycle
        partial_freeze(j) = i
        j = j + 1
      end do
      write (IOUT, *) '   FALLBACK: SELECTED ORBITAL SUBSET:', not_frozen(:)
      ! 1. optimize all the remaining orbitals, keep the subset frozen
      F%basis_ini = B%func
      F%LC_ini = F%LC
      F%iSCFstate = 1
      write (IOUT, *) '   FALLBACK: 1. WITHOUT SUBSET'
      call calculate_electrons(P, A, B, G, F, &
        set_new_geometry=new_geom, set_new_naos=.true., &
        freeze=.True., partial_freeze=not_frozen)
      ! 2. optimize the subset (without MOM), keep the remaining orbitals frozen
      if (F%ierr_o .ne. 1) then
        F%basis_ini = B%func
        F%LC_ini = F%LC
      end if
      F%yes_mom = .False.
      write (IOUT, *) '   FALLBACK: 2. SUBSET ONLY'
      F_save%prod_overlap = ZERO
      call calculate_electrons(P, A, B, G, F, &
        set_new_geometry=new_geom, set_new_naos=.true., &
        freeze=.True., partial_freeze=partial_freeze)
      call check_result(F, P, ifault)
      if (ifault .ne. -1) then
        write (IOUT, '("#   FALLBACK: keeping this result")')
        converged = .TRUE.
        F_save = F
      end if
      ! in case we fail (and len not_frozen==2) rotate the guess orbitals by -pi/4
      if (ifault .ne. 0 .and. size(not_frozen) .eq. 2) then
        write (IOUT, *) '   FALLBACK: 2. SUBSET ONLY (ROTATE BY -PI/4)'
        F%LC(:, not_frozen(1)) = cos(-PI/FOUR)*F%LC_ini(:, not_frozen(1)) + sin(-PI/FOUR)*F%LC_ini(:, not_frozen(2))
        F%LC(:, not_frozen(2)) = -sin(-PI/FOUR)*F%LC_ini(:, not_frozen(1)) + cos(-PI/FOUR)*F%LC_ini(:, not_frozen(2))
        F%LC_ini = F%LC
        call calculate_electrons(P, A, B, G, F, &
          set_new_geometry=new_geom, set_new_naos=.true., &
          freeze=.True., partial_freeze=partial_freeze)
        F%prod_overlap = prodOverlap(P, F, F%STrans, LC_save)
        write (IOUT, '("# FALLBACK: prod overlap to guess ",f15.5)') F%prod_overlap
        call check_result(F, P, ifault)
        if (ifault .ne. -1 .and. F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          converged = .TRUE.
          F_save = F
        end if
      end if
      ! in case we fail (and len not_frozen==2) rotate the guess orbitals by +pi/2 (i.e. +pi/4 to the initial guess orbital)
      if (ifault .ne. 0 .and. size(not_frozen) .eq. 2) then
        write (IOUT, *) '   FALLBACK: 2. SUBSET ONLY (ROTATE BY +PI/4)'
        F%LC(:, not_frozen(1)) = cos(PI/TWO)*F%LC_ini(:, not_frozen(1)) + sin(PI/TWO)*F%LC_ini(:, not_frozen(2))
        F%LC(:, not_frozen(2)) = -sin(PI/TWO)*F%LC_ini(:, not_frozen(1)) + cos(PI/TWO)*F%LC_ini(:, not_frozen(2))
        F%LC_ini = F%LC
        call calculate_electrons(P, A, B, G, F, &
          set_new_geometry=new_geom, set_new_naos=.true., &
          freeze=.True., partial_freeze=partial_freeze)
        F%prod_overlap = prodOverlap(P, F, F%STrans, LC_save)
        write (IOUT, '("# FALLBACK: prod overlap to guess ",f15.5)') F%prod_overlap
        call check_result(F, P, ifault)
        if (ifault .ne. -1 .and. F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          converged = .TRUE.
          F_save = F
        end if
      end if
      if (ifault .ne. 0 .and. converged) then
        write (IOUT, '("#   FALLBACK: RESTORING PREVIOUS RESULT")')
        F = F_save
        F%LC_ini = F%LC
        F%iSCFstate = 2
        call calculate_electrons(P, A, B, G, F, set_new_geometry=.FALSE., &
          freeze=.TRUE., partial_freeze=partial_freeze)
        F%prod_overlap = F_save%prod_overlap
      end if

      ! in case we still fail (and len not_frozen==2) rotate the guess orbitals back by pi/4 (i.e. 0 degree to the initial guess orbital)
      if (.not. converged .and. size(not_frozen) .eq. 2) then
        F%LC(:, not_frozen(1)) = cos(PI/FOUR)*F%LC_ini(:, not_frozen(1)) + sin(PI/FOUR)*F%LC_ini(:, not_frozen(2))
        F%LC(:, not_frozen(2)) = -sin(PI/FOUR)*F%LC_ini(:, not_frozen(1)) + cos(PI/FOUR)*F%LC_ini(:, not_frozen(2))
        F%LC_ini = F%LC
      end if

      ! 3. optimize all orbitals
      deallocate (partial_freeze)
      deallocate (not_frozen)
      F%yes_mom = P%yes_mom
      if (F%ierr_o .ne. 1) then
        F%LC_ini = F%LC
        F%basis_ini = B%func
      end if
      write (IOUT, *) '   FALLBACK: 3. NO RESTRICTION'
      call calculate_electrons(P, A, B, G, F, &
        set_new_geometry=new_geom, set_new_naos=.true., &
        freeze=.False.)
      if (F%ierr_o .ne. 1) exit
    end do

    if (.not. P%yes_gaussian) then
      do i = 1, B%N_basis
        do j = 1, B%N_basis
          F%STrans(i, j) = integral(G, B%func(:, i)*B_save(:, j))
        end do
      end do
    else
      F%STrans = B%S
    end if
    F%prod_overlap = prodOverlap(P, F, F%STrans, LC_save)
    F%LC_ini = LC_save ! reset LC_ini to the initially given LC_ini
    F%basis_ini = B_save
    deallocate (B_save)
    deallocate (LC_save)

  end subroutine calculate_electrons_with_partial_freezing

  subroutine calculate_electrons_in_smooth_steps(P, A, B, G, F, set_new_geometry, auto_average, partial_freezing)
    type(Param), intent(inout) :: P
    type(Atom_set), intent(inout) :: A(:)
    type(Basis), intent(inout)  :: B
    type(Grid3D), intent(inout) :: G
    type(WF_data), intent(inout) :: F

    logical, intent(in), optional :: set_new_geometry
    logical, intent(in), optional :: auto_average
    logical, intent(in), optional :: partial_freezing
    logical :: yes_auto_average = .FALSE.
    logical :: yes_partial_freezing = .FALSE.
    integer :: from_occ(size(F%N_occ))
    integer :: to_occ(size(F%N_occ))
    integer :: i, j, k, l
    integer, parameter :: nstep = 10

    real(long), allocatable :: LC_save(:, :)
    real(long), allocatable :: Strans_save(:, :)
    real(long), allocatable :: B_save(:, :)

    integer :: IOUT
    IOUT = P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    if (present(auto_average)) yes_auto_average = auto_average
    if (present(partial_freezing)) yes_partial_freezing = partial_freezing

    allocate (B_save(G%N_grid, B%N_basis))
    allocate (Strans_save(B%N_basis, B%N_basis))
    allocate (LC_save(B%N_basis, B%N_basis))

    from_occ = F%N_occ_old
    to_occ = F%N_occ

    F%F_occ_given = .TRUE.
    B_save = F%basis_ini
    LC_save = F%LC_ini

    do i = 0, nstep
      ! we set the fractional occupation number explicitly:
      ! 1. f_occ=from_occ + i/n_step * (to_occ - from_occ)
      ! 2. if yes_gs_occ first convert occs to groundstate and then do smooth steps
      ! 3. perform averaging
      F%N_occ = to_occ
      if (2*i .lt. nstep) then
        F%N_occ = from_occ
      else
        F%N_occ = to_occ
      end if
      F%f_occ = dble(from_occ) + dble(i)/dble(nstep)*(dble(to_occ) - dble(from_occ))
      if (P%yes_valgs_occ) then
        F%f_occ=dble(groundstate_occ(from_occ,Ncore=F%Ncore, add_electrons=.not. P%yes_valgs_occ_part)) &
          + dble(i)/dble(nstep) * ( &
              dble(groundstate_occ(to_occ,Ncore=F%Ncore, add_electrons=.not. P%yes_valgs_occ_part))&
            - dble(groundstate_occ(from_occ,Ncore=F%Ncore, add_electrons=.not. P%yes_valgs_occ_part)) )
      end if
      if (allocated(F%av_occ)) then
        do j = 1, size(F%av_occ)
          F%F_occ(F%av_occ(j)%MO(:)) = ONE/size(F%av_occ(j)%MO)*(sum(F%f_occ(F%av_occ(j)%MO)))
        end do
      end if
      if (F%ierr_o .ne. 1) then
        ! if converged set LC_ini and basis_ini to the converged bf
        F%LC_ini = F%LC
        F%basis_ini = B%func
        if (yes_auto_average) &
          call automatic_averaging(P, B, F, F%av_occ)
      else
        ! if not converged use basis_ini, LC_ini to perform auto average and determine basis set
        F%LC = F%LC_ini
        B%func = F%basis_ini
        do k = 1, B%N_basis
          do l = k, B%N_basis
            B%S(k, l) = integral(G, B%func(:, k)*B%func(:, l))
            B%S(l, k) = B%S(k, l)
          end do
        end do
        if (yes_auto_average) &
          call automatic_averaging(P, B, F, F%av_occ)
      end if

      call calculate_electrons(P, A, B, G, F, set_new_geometry, set_new_naos=.true.)
      if (F%ierr_o .eq. 1 .and. yes_partial_freezing) then
        ! embed partial freezing here 20200925
        call calculate_electrons_with_partial_freezing(P, A, B, G, F, set_new_geometry=set_new_geometry)
      end if
      if (F%ierr_o .eq. 1) then
        call calculate_electrons(P, A, B, G, F, set_new_geometry, set_new_naos=.false., freeze=.true.)
        F%ierr_o = 1
      end if
      if (.not. P%yes_gaussian) then
        do k = 1, B%N_basis
          do l = 1, B%N_basis
            STrans_save(k, l) = integral(G, B%func(:, k)*B_save(:, l))
          end do
        end do
      else
        STrans_save = B%S
      end if
      if (F%ierr_o .eq. 1) then
        write (IOUT, "('# step ',i3,' failed')") i
      else
        write (IOUT, "('# step ',i3,' prod overlap to start',f15.5)") i, prodOverlap(P, F, STrans_save, LC_save)
      end if
    end do

    F%F_occ_given = .FALSE.
    if (F%ierr_o .ne. 1) then
      F%LC_ini = F%LC
      F%basis_ini = B%func
      if (yes_auto_average) &
        call automatic_averaging(P, B, F, F%av_occ)
    else
      ! if not converged use basis_ini, LC_ini to determine basis set
      F%LC = F%LC_ini
      B%func = F%basis_ini
      do k = 1, B%N_basis
        do l = k, B%N_basis
          B%S(k, l) = integral(G, B%func(:, k)*B%func(:, l))
          B%S(l, k) = B%S(k, l)
        end do
      end do
    end if

    F%N_occ = to_occ
    F%iSCFstate = 1
    call calculate_electrons(P, A, B, G, F, set_new_geometry, set_new_naos=.true.)
    if (.not. P%yes_gaussian) then
      do i = 1, B%N_basis
        do j = 1, B%N_basis
          F%STrans(i, j) = integral(G, B%func(:, i)*B_save(:, j))
        end do
      end do
    else
      F%STrans = B%S
    end if

    F%prod_overlap = prodOverlap(P, F, F%STrans, LC_save)
    if (F%ierr_o .eq. 1) then
      write (IOUT, "('# smooth steps failed')")
    else
      write (IOUT, "('# smooth steps done prod. overlap ',f15.5)") F%prod_overlap
    end if

    ! after we are done return to original averaging
    ! if (yes_auto_average)call original_averaging(P,A,B,F,F%av_occ)

    F%LC_ini = LC_save ! reset LC_ini to the initially given LC_ini
    F%basis_ini = B_save

    deallocate (B_save)
    deallocate (LC_save)
    deallocate (Strans_save)
  end subroutine calculate_electrons_in_smooth_steps

  !> subroutine calculate_electrons(P, A, B, G, F)
  !! calculates the electronic structure
  !! P,A,B,G,F structure are prepared from a previous calculation
  !! A,B,G,F are adapted to the new geometry and new wave function
  !! F has been adapted such that:
  !! F%N_occ has changed
  !! F%basis_ini and LC_ini needs to be set by data from a previous run, e.g.
  !! F%basis_ini=B%func
  !! F%LC_ini=F%LC
  !! F%iSCFstate=1 for a calculation that automatically prepares and use the naos
  !!               on basis of the information from the current basis in B
  !! F%iSCFstate=2 for a calculation that uses the hole string defined in A(i)%hole_string
  !!               to prepare and use the naos
  !!
  !! optional freeze=True/False freeze all mo's
  !! optional freeze=True combined with partial_freeze=(/2,4,.../) freeze mo's specified in partial_freeze
  !! optional set_new_geometry: construct new grid
  !! optional set_new_naos: determine new basis functions
  subroutine calculate_electrons(P, A, B, G, F, set_new_geometry, set_new_naos, freeze, partial_freeze)
    type(Param), intent(in) :: P
    type(Atom_set), intent(inout) :: A(:)
    type(Basis), intent(inout)  :: B
    type(Grid3D), intent(inout) :: G
    type(WF_data), intent(inout) :: F
    logical, intent(in), optional :: set_new_geometry
    logical, intent(in), optional :: set_new_naos
    logical, intent(in), optional :: freeze
    integer, intent(in), optional :: partial_freeze(:)
    logical :: new_naos, new_geometry, yes_freeze
    integer :: i, j

    if (present(set_new_geometry)) then
      new_geometry = set_new_geometry
    else
      ! we assume a new geometry if not specified
      new_geometry = .TRUE.
    end if
    if (present(freeze)) then
      yes_freeze = freeze
    else
      yes_freeze = .FALSE.
    end if
    new_naos = .TRUE.
    if (present(set_new_naos)) &
      new_naos = set_new_naos
    ! update F%f_occ to derive the new basis functions
    if (.NOT. F%f_occ_given) then
      F%f_occ = ONE*F%n_occ
      if (allocated(F%av_occ)) then
        do j = 1, size(F%av_occ)
          F%F_occ(F%av_occ(j)%MO(:)) = ONE/size(F%av_occ(j)%MO)*(sum(F%N_occ(F%av_occ(j)%MO)))
        end do
      end if
    end if
    if (new_naos .and. .NOT. (P%yes_gaussian .and. .not. P%yes_ionization_dynamics)) &
      call make_naos(P, A, B, F, new_naos)
    ! only neccessary when basis functions have changed
    if (new_naos .and. .NOT. P%yes_gaussian) call update_Atom_set_with_NAO(A)
    if (new_geometry) then
      !call purge_Grid3D( G )
      !call construct_Grid3D( P, A, G )
      call reconstruct_Grid3D(A, G)
    end if
    if (new_naos .OR. new_geometry .OR. .TRUE.) then
      call purge_Basis(B)
      if (P%yes_gaussian) then
        call construct_Basis( A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym ,gto_file_path=P%gto_file_path, yes_gradient=.True., basisSetOverlapLim=P%basisSetOverlapLim)
      else
        call construct_Basis( A, G, B, P%EPS_overlap, P%EPS_overlap_single, yes_sym=P%yes_sym ,yes_gradient=.True., basisSetOverlapLim=P%basisSetOverlapLim)
      end if
      ! in case of large overlap the number of orbitals might change
      ! in this case we need to update the wavefunction data
      if (B%Nr_basis .ne. F%Nr_basis) then
        write (P%OUTPUT, '("# Number of bf changes from ",i5," to ", i5)') F%Nr_basis, B%Nr_basis
        call adapt_WF_Nr_basis(B, F)
      end if

      if (P%yes_gaussian .and. .NOT. P%yes_direct_scf) then
        call precalculate_twoel_GTO(P, B)
      end if
    end if
    if (yes_freeze) then
      if (present(partial_freeze)) then
        allocate (F%frozen_MO(size(partial_freeze)))
        allocate (F%P_not_frozen(B%N_basis, B%N_basis))
        F%frozen_MO(:) = partial_freeze(:)
      else
        allocate (F%frozen_MO(B%N_basis))
        allocate (F%P_not_frozen(B%N_basis, B%N_basis))
        F%frozen_MO(:) = (/(i, i=1, B%N_basis)/)
      end if
    end if

    call SCF_procedure(P, A, G, B, F)

    if (P%yes_gaussian .and. P%yes_ionization_dynamics) &
      call construct_overlap2minimalbasis(P, A, G, B, F)

    if (yes_freeze) then
      deallocate (F%frozen_MO)
      deallocate (F%P_not_frozen)
    end if
    call print_calc_info(P, A, G, B)    ! defined in the wavefunction module

  end subroutine calculate_electrons

  !> This subroutine calculates rates after
  !! a call to calculate_electrons
  subroutine calculate_rates(P, A, B, G, F, S)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in)  :: B
    type(Grid3D), intent(in) :: G
    type(WF_data), intent(in) :: F
    type(Process_data), intent(inout) :: S
    !integer :: N_occ(size(F%N_occ))
    !  if ( P%yes_gaussian ) return
    call purge_Process(S)
    S = create_Process(F%N_occ, P%N_proc_max, spin=P%yes_process_spin, ci=P%yes_process_ci)
    if (P%yes_process_spin) then
      call calcAugerRates_spin_avg(P, A, B, F, S)
      call calc_Fluorescence_Rates_spin_avg(P, B, F, G, S)
      call calculate_photoIonization_CS_spin_avg(P, A, B, F, S)
    else
      call calcAugerRates(P, A, B, F, S)
      call calc_Fluorescence_Rates(P, B, F, G, S)
      call calculate_photoIonization_CS(P, A, B, F, S)
    end if
    if (P%yes_average_rates .or. P%yes_average_csf_rates) &
      call sum_equivalent_configuration(S)

    !open(unit=55,file=P%file_data,status='UNKNOWN',ACTION='WRITE',position = 'append')
    !close(55)

    !  call print_process(S,P%OUTPUT)
  end subroutine calculate_rates

  !> subroutine make_naos(P, A, B, G, F)
  !! performs xatom calculations for each atom
  !! for the basis functions
  !! for the auger radial matrix elements
  !! for the transition dipole matrix elements
  !! the filenames where the data for each atom is stored
  !! is
  !! A(i)%nao_file_nuc
  !! A(i)%auger_file_nuc
  !! A(i)%tdipole_file_nuc
  !!
  subroutine make_naos(P, A, B, F, new_naos)
    type(Param), intent(in) :: P
    type(Atom_set), intent(inout) :: A(:)
    type(Basis), intent(in)  :: B
    type(WF_data), intent(in) :: F
    logical, intent(inout), optional :: new_naos

    integer :: i, ind, j
    !character(len=512), parameter :: XATOM_DEFAULT_PATH ="/work/linhest/XATOM/src/xatom"
    character(len=512) :: XATOM_PATH = ""
    character(len=128) :: PE_string, xatom_param_string
    character(len=128) :: hole_cmd_string, hole_string
    character(len=128) :: conf_cmd_string, conf_string

    character(len=1500) :: cmd_line

    integer, parameter:: MAX_CACHE = 100
    character(len=512), dimension(MAX_CACHE), static :: cached_nao_files
    character(len=512), dimension(MAX_CACHE), static :: cached_auger_files
    character(len=512), dimension(MAX_CACHE), static :: cached_tdipole_files
    integer, static :: ncached = 0

    logical :: nao_is_cached = .FALSE.
    logical :: file_exist
    logical :: file_exist1
    logical :: file_exist2
    logical :: file_exist3

    integer :: IOUT
    IOUT = P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    if (present(new_naos)) new_naos = .FALSE.

    inquire (file=P%XATOM_PATH, EXIST=file_exist)
    if (file_exist) then
      XATOM_PATH = P%XATOM_PATH
    else
      write (STDERR, '("Not found: ",a)') P%XATOM_PATH
      stop 'Please specify xatom path'
    end if
    !
    ! prepare the basis functions:
    !

    ! get the string defining the photo electron energy range
    PE_string = ""
    if (F%iSCFstate .eq. 1) &
      call get_PE_string(P, A, F, PE_string)
    if (len(trim(PE_string)) .eq. 0) &
      write (PE_string, '("-PE_start ",f12.2," -PE_end ",f12.2," -dE ",f12.2 )' ) &
      10., P%PE*AU2EV + 500, 10.

    !write(IOUT,"('# PE_string=',a)") trim(PE_string)

    do i = 1, P%N_nuc
      ! special case for hydrogen: for hydrogen we assume naos are there
      ! and we don't calculate tdipole auger matrix element: by this we can save grid points
      if (A(i)%charge .eq. 1) cycle

      ! only get new hole_string if F%iSCFstate==1 else use the conf_string in A(i)
      if (F%iSCFstate .eq. 1) then
        hole_string = ""
        call get_hole_string(P, B, F, i, hole_string)
        conf_string = ""
        call get_conf_string(P, A, B, F, i, conf_string)
        if (A(i)%hole_string .ne. hole_string) then
          A(i)%hole_string = hole_string
          if (present(new_naos)) new_naos = .TRUE.
        end if
        if (A(i)%conf_string .ne. conf_string) then
          A(i)%conf_string = conf_string
          if (present(new_naos)) new_naos = .TRUE.
        end if
      else
        if (present(new_naos)) new_naos = .TRUE.
        ! we take the conf string that is
        ! specified in A(i)%conf_string
      end if

      ! set the file names and cmd options
      ind = index(A(i)%nao_file_nuc, ".nao") - 1
      if (len(trim(A(i)%conf_string)) .eq. 0) then
        stop "no conf string set in A structure"
      else
        hole_cmd_string = " -hole "//trim(A(i)%hole_string)
        conf_cmd_string = " -conf "//trim(A(i)%conf_string)
        A(i)%nao_file_nuc = trim(A(i)%nao_file_nuc(1:ind))//".nao.conf_"//trim(A(i)%conf_string)
        A(i)%auger_file_nuc = trim(A(i)%auger_file_nuc(1:ind))//".auger.conf_"//trim(A(i)%conf_string)
        A(i)%tdipole_file_nuc = trim(A(i)%tdipole_file_nuc(1:ind))//".tdipole.conf_"//trim(A(i)%conf_string)
      end if
      nao_is_cached = .FALSE.
      ! if the nao is already there skip the calculation
      do j = 1, ncached
        if (A(i)%nao_file_nuc .eq. cached_nao_files(j)) then
          nao_is_cached = .TRUE.
          exit
        end if
      end do
      if (nao_is_cached) cycle

      inquire (file=trim(A(i)%nao_file_nuc), EXIST=file_exist1)
      inquire (file=trim(A(i)%auger_file_nuc), EXIST=file_exist2)
      inquire (file=trim(A(i)%tdipole_file_nuc), EXIST=file_exist3)
      if (file_exist1 .and. file_exist2 .and. file_exist3) then
        write (IOUT, "('# xatom result files:',a,' already exist')") &
          trim(A(i)%nao_file_nuc)
      else

        ! perform xatom call
        ! parameters and options
        ! nev=maxval(A(i)%N_nao_table)
        ! lmax=maxval(A(i)%l_AO)
        ! N=A(i)%N_r
        ! rmax=A(i)%rmax
        ! rm=1e-6 (fixed)
        ! latter=off
        ! silent=yes
        write (xatom_param_string, '(" -silent -nev ",i4," -lmax ",i4," -latter off -rm 1.e-6 -N ",i4, " -rmax ",f15.3,"  ")') &
          maxval(A(i)%N_nao_table), &
          maxval(A(i)%l_AO), &
          A(i)%N_r, &
          A(i)%rmax

        cmd_line = trim(XATOM_PATH)// &
          " -s "//trim(ATOMIC_NAME(A(i)%charge))// &
        !        trim(hole_cmd_string) //&
          trim(conf_cmd_string)// &
          trim(xatom_param_string)// &
          " -nao -file_nao "//trim(A(i)%nao_file_nuc)// &
          " -transition auger -file_auger "//trim(A(i)%auger_file_nuc)// &
          " -transition transitiondipole "//trim(PE_string)//" -file_tdipole "//trim(A(i)%tdipole_file_nuc)

        ind = index(P%nao_file_nuc(i), "ext")
        if (ind .ge. 1) &
          cmd_line = trim(cmd_line)//" -polarization "

        write (IOUT, "('# CALLING:',100a)") trim(cmd_line)
        call system(cmd_line)
      end if

      ! add to cache list
      if (ncached .lt. MAX_CACHE) then
        ncached = ncached + 1
        cached_nao_files(ncached) = A(i)%nao_file_nuc
        cached_auger_files(ncached) = A(i)%auger_file_nuc
        cached_tdipole_files(ncached) = A(i)%tdipole_file_nuc
      else
        ! remove first entry
        !      cmd_line=trim("rm ") // cached_nao_files(1)
        !      write(IOUT,"('# CALLING:',100a)") trim(cmd_line)
        !      call system(cmd_line)
        !      cmd_line=trim("rm ") // cached_auger_files(1)
        !      write(IOUT,"('# CALLING:',100a)") trim(cmd_line)
        !      call system(cmd_line)
        !      cmd_line=trim("rm ") // cached_tdipole_files(1)
        !      write(IOUT,"('# CALLING:',100a)") trim(cmd_line)
        !      call system(cmd_line)
        ! cycle
        do j = 1, ncached - 1
          cached_nao_files(j) = cached_nao_files(j + 1)
          cached_auger_files(j) = cached_auger_files(j + 1)
          cached_tdipole_files(j) = cached_tdipole_files(j + 1)
        end do
        ! add as last entry
        cached_nao_files(ncached) = A(i)%nao_file_nuc
        cached_auger_files(ncached) = A(i)%auger_file_nuc
        cached_tdipole_files(ncached) = A(i)%tdipole_file_nuc
      end if
    end do ! loop over the atoms

  end subroutine make_naos

!> subroutine get_PE_string(P,A,B,G,F,PE_string )
!! this subroutine returns the string defining the range of photoelectrons we consider in the xatom calculation
!! (e.g. '-PE_start 11000 -PE_end 12000 -dE 10' for photoelectrons from 11keV to 12keV)
!! for performing xatom calculation  (adopted from Kotas original implementation)
  subroutine get_PE_string(P, A, F, PE_string)
    type(Param), intent(in) :: P     ! the usual structures
    type(Atom_set), intent(in) :: A(:)
    type(WF_data), intent(in) :: F
    character(len=128), intent(inout) :: PE_string ! the hole string to return
    real(long) dEmin_eV, omega_AU, Emin_ev, Emax_eV, Emargin_eV

    dEmin_eV = 10.0_long                                    ! XXX TODO
    omega_AU = P%PE
    Emargin_eV = 200_long

    Emax_eV = (max(maxval(F%CE(:)), ZERO) + omega_AU)/EV2AU
    !  Emin_ev= ( minval(F%CE(:) + omega_AU, mask=(F%CE(:) + omega_AU + Emargin_eV > ZERO ) ) )/ EV2AU
    ! there is a trend that with higher charge states the photoelectron energy lowers
    ! we take into account this effect by a shift of 30% (Kotas idea)
    !  Emin_ev=0.7_long*Emin_eV

    ! 30 % seemed to be not enough: 50 %
    !  Emin_ev=HALF*Emin_eV

    ! additional savety margin of 100 eV

    !  Emin_eV=max( Emin_eV - Emargin_eV, ZERO )

    ! (2*pi/rmax)**2 is the continuum energy that has in the neutral case one period
    !  Emin_ev=(4*PI/minval( A(:)%rmax,mask=A(:)%charge>1 ))**2*AU2EV
    Emin_ev = (10*PI/minval(A(:)%rmax, mask=A(:)%charge .gt. 1))**2*AU2EV

    write (PE_string, '("-PE_start ",f12.2," -PE_end ",f12.2," -dE ",f12.2 )') Emin_ev, Emax_ev, dEmin_eV

  end subroutine get_PE_string

!> subroutine get_hole_string(P,B,G,F,nucleus_i,hole_string )
!! this subroutine returns the hole_string (e.g. '1s1' for a single K-shell vacancy)
!! for performing xatom calculation for atom nucleus_i
!! according to the gross shell population of the current F structure
  subroutine get_hole_string(P, B, F, nucleus_i, hole_string)
    type(Param), intent(in) :: P     ! the usual structures
    type(Basis), intent(in)  :: B
    type(WF_data), intent(in) :: F
    integer, intent(in):: nucleus_i ! index of the nucleus to consider
    character(len=80), intent(inout) :: hole_string ! the hole string to return
    integer ::  i, j
    integer :: bf, n, l, occ ! bf: basis function index, n,l : atomic shell indices, occ: rounded occupation of shell
    real(long) :: gross_pop_shell ! the gross population of the shell (n,l)
    character(len=8) :: substr  ! substring for string manipulation
    integer :: neutral_occ
    integer, allocatable :: info(:, :)
    hole_string = ""

    if (P%yes_gaussian) then
      allocate (info(size(F%min_info, 1), size(F%min_info, 2)))
      info = F%min_info
    else
      allocate (info(size(B%info, 1), size(B%info, 2)))
      info = B%info
    end if

    ! loop over all bf
    bf = 1
    do
      if (bf .gt. size(info, 1)) exit
      n = info(bf, 3)
      l = info(bf, 4)
      neutral_occ = neutral_occ_aufbau(info(bf, 2), n, l)

      !  bf is located on atom i and n is <= 1 (i.e. core orbitals = K-shell)
      ! B%info(bf,2) > 1 (no hydrogen atoms!)

      if (info(bf, 1) .eq. nucleus_i .and. &
        (P%yes_adapt_for_all_AO .or. &
        (n .le. 1 .and. info(bf, 2) .gt. 2) &  ! for the first row only K-Shell
        .or. (n .le. 2 .and. info(bf, 2) .gt. 10) &  ! for the second row K and L1 and L2 Shell
        .or. (n .le. 3 .and. l .le. 1 .and. info(bf, 2) .gt. 18) &  ! for the third row K, L1, L2, M1,M2 Shell
        .or. (n .le. 4 .and. l .le. 2 .and. info(bf, 2) .gt. 36)) &  ! for the fourth row n=1,2,3,4 and l=s,p,d
        ) then
        ! calculate the gross population of this (n,l) shell
        ! the next 2*l+1 basis functions are members of this shell
        ! bf+j runs from bf, ... , bf+(2*l)  (in total: 2*l+1)
        gross_pop_shell = ZERO
        do j = 0, 2*l
          if (P%yes_gaussian) then
            do i = 1, size(info, 1)
              gross_pop_shell = gross_pop_shell + F%f_occ(i)*F%Omin(bf + j, i)*sum(F%Omin(:, i)*F%Smin(bf + j, :))
            end do
          else
            do i = 1, F%N_basis
              gross_pop_shell = gross_pop_shell + F%f_occ(i)*F%LC(bf + j, i)*sum(F%LC(:, i)*B%S(bf + j, :))
            end do
          end if! P%yes_gaussian
        end do! j
        occ = 0
        occ = nint(gross_pop_shell)
        if (neutral_occ .gt. occ) then
          ! the 'hole occ' is the total shell occupation neutral_occ - occ
          write (substr, '(i2)') neutral_occ - occ
          if (len(trim(hole_string)) .eq. 0) then
            hole_string = trim(subshell_name(n, l))//adjustl(substr)
          else
            hole_string = trim(hole_string)//"_"//trim(subshell_name(n, l))//adjustl(substr)
          end if
        end if
      end if

      ! go to the next nl shell, i.e. bf=2*l+1
      bf = bf + (2*l + 1)
    end do
    deallocate (info)
  contains
    function neutral_occ_aufbau(Z, n, l) result(occ)
      integer, intent(in):: Z, n, l
      integer :: occ
      integer :: n_i, l_i
      integer :: num
      occ = 0
      num = 0
      do n_i = 1, n
        do l_i = 0, n_i - 1
          occ = 2*(2*l_i + 1)
          if (num + occ .le. Z) then
            num = num + occ
          else
            occ = Z - num
            num = num + occ
          end if
          if (num .eq. Z .and. n_i .eq. n .and. l_i .eq. l) return
          if (n_i .eq. n .and. l_i .eq. l) return
        end do
      end do
    end function neutral_occ_aufbau
  end subroutine get_hole_string

!> subroutine get_conf_string(P,B,G,F,nucleus_i,hole_string )
!! this subroutine returns the conf_string (e.g. '1s1_2s2...' for a single K-shell vacancy)
!! for performing xatom calculation for atom nucleus_i
!! according to the gross shell population of the current F structure
!! subroutine get_conf_string(P,A,B,G,F,nucleus_i,conf_string )
  subroutine get_conf_string(P, A, B, F, nucleus_i, conf_string)
    type(Param), intent(in) :: P     ! the usual structures
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in)  :: B
    type(WF_data), intent(in) :: F
    integer, intent(in):: nucleus_i ! index of the nucleus to consider
    character(len=80), intent(inout) :: conf_string ! the hole string to return
    integer ::  i, j
    integer :: bf, n, l, occ ! bf: basis function index, n,l : atomic shell indices, occ: rounded occupation of shell
    real(long) :: gross_pop_shell ! the gross population of the shell (n,l)
    character(len=8) :: substr  ! substring for string manipulation
    integer, allocatable :: info(:, :)
    conf_string = ""

    if (P%yes_gaussian) then
      allocate (info(size(F%min_info, 1), size(F%min_info, 2)))
      info = F%min_info
    else
      allocate (info(size(B%info, 1), size(B%info, 2)))
      info = B%info
    end if

    ! loop over all bf
    bf = 1
    do
      if (bf .gt. size(info, 1)) exit
      n = info(bf, 3)
      l = info(bf, 4)

      !  bf is located on atom i and n is <= 1 (i.e. core orbitals = K-shell)
      ! B%info(bf,2) > 1 (no hydrogen atoms!)

      if (info(bf, 1) .eq. nucleus_i .and. &
        (P%yes_adapt_for_all_AO .or. (n .le. 1 .and. info(bf, 2) .gt. 2) &  ! for the first row only K-Shell
        .or. (n .le. 2 .and. info(bf, 2) .gt. 10) &  ! for the second row K and L1 and L2 Shell
        .or. (n .le. 3 .and. l .le. 1 .and. info(bf, 2) .gt. 18) &  ! for the third row K, L1, L2, M1,M2 Shell
        .or. (n .le. 4 .and. l .le. 2 .and. info(bf, 2) .gt. 36)) &  ! for the fourth row n=1,2,3,4 and l=s,p,d
        ) then
        ! calculate the gross population of this (n,l) shell
        ! the next 2*l+1 basis functions are members of this shell
        ! bf+j runs from bf, ... , bf+(2*l)  (in total: 2*l+1)
        gross_pop_shell = ZERO
        do j = 0, 2*l
          if (P%yes_gaussian) then
            do i = 1, size(info, 1)
              gross_pop_shell = gross_pop_shell + F%f_occ(i)*F%Omin(bf + j, i)*sum(F%Omin(:, i)*F%Smin(bf + j, :))
            end do
          else
            do i = 1, F%N_basis
              gross_pop_shell = gross_pop_shell + F%f_occ(i)*F%LC(bf + j, i)*sum(F%LC(:, i)*B%S(bf + j, :))
            end do
          end if! P%yes_gaussian
        end do! j
        occ = nint(gross_pop_shell)

        ! in case of strong linear dependency of the basis functions
        ! occ may be smaller than 0 or larger than 2*(2*l+1)
        if (occ .gt. 2*(2*l + 1)) occ = 2*(2*l + 1)
        if (occ .lt. 0) occ = 0

        write (substr, '(i2)') occ
        if (len(trim(conf_string)) .eq. 0) then
          conf_string = trim(subshell_name(n, l))//adjustl(substr)
        else
          conf_string = trim(conf_string)//"_"//trim(subshell_name(n, l))//adjustl(substr)
        end if

      else if (info(bf, 1) .eq. nucleus_i) then
        ! if it is not considered as a core shell, add the configuration that is specified in A(nucleus_i)%config
        do j = 1, size(A(nucleus_i)%config, 1)
          if (A(nucleus_i)%config(j, 1) .eq. n .and. A(nucleus_i)%config(j, 2) .eq. l) then
            write (substr, '(i2)') A(nucleus_i)%config(j, 3)
            if (len(trim(conf_string)) .eq. 0) then
              conf_string = trim(subshell_name(n, l))//adjustl(substr)
            else
              conf_string = trim(conf_string)//"_"//trim(subshell_name(n, l))//adjustl(substr)
            end if
            exit ! we have found the right config entry, so exit loop
          end if
          ! else it is a polarization basis function
          ! or a virtual that is not specified in configuration string
        end do
      end if!

      ! go to the next nl shell, i.e. bf=2*l+1
      bf = bf + (2*l + 1)
    end do
    deallocate (info)
  end subroutine get_conf_string

  subroutine xmolecule_setup(Xdata_cptr, filename, lenFilename) bind(c)
    !! sets up the xmol object
    !! reads in the input parameters and prepares for the calculation
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    character(c_char), intent(inout) :: filename(lenFilename) !! input filename
    integer(c_int), intent(in) :: lenFilename !! length of input filename
    character(len=lenFilename) :: filename_string
    !type(Param) :: P
    !type(Atom_set), allocatable :: A(:)
    !type(Grid3D) :: G
    !type(Basis) :: B
    !type(WF_data) :: F
    integer :: i
    logical :: file_exist
    character(len=1500) :: cmd_line
    character(len=512) :: XATOM_PATH = ""
    integer :: IOUT
    filename_string = transfer(filename, filename_string)
    allocate (Xdata)
    allocate (Xdata%P)
    allocate (Xdata%G)
    allocate (Xdata%B)
    allocate (Xdata%F)
    allocate (Xdata%S)
    call init_time()
    call read_parameters(Xdata%P, calc_string="full", filename=filename_string)
    call print_parameters(Xdata%P, "electronic structure of molecules")
    !Xdata%P = P
    IOUT = Xdata%P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2
    ! check for existance of xatom
    ! if nao files not exist create nao files with default parameters
    ! inquire(file=P%XATOM_PATH,EXIST=file_exist)
    ! if (file_exist) then
    !   XATOM_PATH=P%XATOM_PATH
    ! else
    !   write(STDERR,'("Not found: ",a)') P%XATOM_PATH
    !   stop 'Please specify xatom path'
    ! end if
    ! do i=1,size(P%nao_file_nuc)
    !   inquire(file=P%nao_file_nuc(i), exist=file_exist)
    !   if ( .NOT. file_exist) then
    !     cmd_line=trim(XATOM_PATH) // &
    !       " -s " // trim( ATOMIC_NAME( P%charge_nuc(i) ) ) //&
    !       " -silent -latter off -rm 1.e-6 -N "//&
    !       " -nao -file_nao " // trim(P%nao_file_nuc(i))

    !     write(IOUT,"('# CALLING:',100a)") trim(cmd_line)
    !     call system(cmd_line)
    !   end if
    ! end do
    call init_index_packed()
    call construct_Atom_set(Xdata%P, Xdata%A)
    call construct_Grid3D(Xdata%P, Xdata%A, Xdata%G)
    if (Xdata%P%yes_gaussian) then
      call construct_Basis( Xdata%A, Xdata%G, Xdata%B, Xdata%P%EPS_overlap, Xdata%P%EPS_overlap_single, yes_sym=Xdata%P%yes_sym ,gto_file_path=Xdata%P%gto_file_path, yes_gradient=.True., basisSetOverlapLim=Xdata%P%basisSetOverlapLim)
    else
      call construct_Basis( Xdata%A, Xdata%G, Xdata%B, Xdata%P%EPS_overlap, Xdata%P%EPS_overlap_single, yes_sym=Xdata%P%yes_sym ,yes_gradient=.True., basisSetOverlapLim=Xdata%P%basisSetOverlapLim)
    end if
    if (Xdata%P%yes_gaussian .and. .NOT. Xdata%P%yes_direct_scf) then
      call precalculate_twoel_GTO(Xdata%P, Xdata%B)
    end if

    call construct_WF_data(Xdata%P, Xdata%A, Xdata%G, Xdata%B, Xdata%F)
    !call print_calc_info( Xdata%P, Xdata%A, Xdata%G, Xdata%B )    ! defined in the wavefunction module

    Xdata_cptr = C_LOC(Xdata)
  end subroutine xmolecule_setup

  subroutine xmolecule_purge(Xdata_cptr) bind(c)
    !! clears all the allocated data
    !! purges Xdata structure
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    CALL C_F_POINTER(Xdata_cptr, Xdata)

    call purge_Grid3D(Xdata%G)
    deallocate (Xdata%G)
    call purge_Basis(Xdata%B)
    deallocate (Xdata%B)
    call purge_WF_data(Xdata%F)
    deallocate (Xdata%F)
    deallocate (Xdata%A)
    deallocate (Xdata%P)
    deallocate (Xdata)
  end subroutine xmolecule_purge

  subroutine xmolecule_init(Xdata_cptr) bind(c)
    !! initializes calculations
    !! performs the first electronic structure calculation
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call original_averaging(Xdata%P, Xdata%F%av_occ)
    call initial_calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F)
    if (Xdata%P%yes_CI) then
      call CI_procedure(Xdata%P, Xdata%A(:), Xdata%B, Xdata%F, OUTPUT=0)
    end if
  end subroutine xmolecule_init

  subroutine xmolecule_get_energy(Xdata_cptr, energy) bind(c)
    !! returns energy
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: energy
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    energy = Xdata%F%E_tot
  end subroutine xmolecule_get_energy

  subroutine xmolecule_get_mass(Xdata_cptr, mass, l_mass) bind(c)
    !! returns atomic masses
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: mass(1:l_mass)
    integer(c_int), intent(inout) :: l_mass
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (l_mass .ne. size(Xdata%P%mass_nuc)) then
      l_mass = size(Xdata%P%mass_nuc)
    else
      mass = Xdata%P%mass_nuc(:)
    end if
  end subroutine xmolecule_get_mass

  subroutine xmolecule_get_atomlabels(Xdata_cptr, labels, l_labels) bind(c)
    !! returns atom labels
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    character(c_char), intent(inout) :: labels(6*l_labels)
    integer(c_int), intent(inout) :: l_labels
    integer :: i, j
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (l_labels .ne. size(Xdata%A)) then
      l_labels = size(Xdata%A)
    else
      do i = 1, l_labels
        do j = 1, 6
          labels((i - 1)*6 + j:(i - 1)*6 + j) = Xdata%A(i)%label(j:j)
        end do
      end do
    end if
  end subroutine xmolecule_get_atomlabels

  subroutine xmolecule_orbital_energies(Xdata_cptr, CE, l_CE) bind(c)
    !! returns orbital energies
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: CE(l_CE)
    integer(c_int), intent(inout) :: l_CE
    integer :: Norb
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Norb = size(Xdata%F%CE)
    if (Norb .ne. l_CE) then
      l_CE = Norb
    else
      CE = Xdata%F%CE
    end if
  end subroutine xmolecule_orbital_energies

  subroutine xmolecule_hole_occs_energies_efield(Xdata_cptr, pos, occs, energies, efields, upots, l_E, l_occ) bind(c)
    !! returns for hole configurations energies and occupation patterns in the presence of electric field based on Koopmann's theorem
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(in) :: pos(3)
    integer(c_int), intent(inout) :: occs(l_occ, l_E)
    real(c_double), intent(inout) :: energies(l_E)
    real(c_double), intent(inout) :: efields(3, l_E)
    real(c_double), intent(inout) :: upots(l_E)
    integer(c_int), intent(inout) :: l_E !! length of energies
    integer(c_int), intent(inout) :: l_occ !! length of occ
    integer, allocatable :: gs_occ(:)
    integer :: i
    integer :: Norb, Nval
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_excitation_energies: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ)) stop 'xmolecule_excitation_energies: occ has wrong len'
    allocate (gs_occ(size(Xdata%F%N_occ)))
    Norb = size(Xdata%F%CE)
    gs_occ = groundstate_occ(Xdata%F%N_occ(:), Ncore=Xdata%F%Ncore)
    if (.not. Xdata%P%yes_valgs_occ .or. sum(gs_occ) .ne. sum(Xdata%F%N_occ(:)) + 1 ) &
      stop 'only works with Koopmann s theorem states!'

    Nval = count(gs_occ .eq. 2) - Xdata%F%Ncore
    if (l_E .ne. Nval) then
      l_E = -Nval
      return
    end if
    call calculate_hole_energies_Efield(Xdata%P, Xdata%A, Xdata%B, Xdata%F, pos, occs, energies, efields, upots)
    do i = 1, size(occs, 2)
      if (all(occs(:, i) .eq. -1)) then
        l_E = i - 1
        exit
      end if
    end do
    deallocate (gs_occ)
  end subroutine xmolecule_hole_occs_energies_efield

  subroutine xmolecule_hole_occs_energies_gradients_nacs(Xdata_cptr, occs, energies, gradients, nacs, l_E, l_gradient, l_occ) bind(c)
    !! returns for hole configurations energies, occupation patterns, gradients and nacs in the presence of electric field based on Koopmann's theorem
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: occs(l_occ, l_E)
    real(c_double), intent(inout) :: energies(l_E)
    real(c_double), intent(inout) :: gradients(3*l_gradient, l_E)
    real(c_double), intent(inout) :: nacs(3*l_gradient, l_E, l_E)
    integer(c_int), intent(inout) :: l_E
    integer(c_int), intent(inout) :: l_gradient
    integer(c_int), intent(inout) :: l_occ
    integer :: gs_occ(l_occ)
    integer :: i
    integer :: Norb, Nval
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ) + 2) &
      stop 'xmolecule_excitation_energies: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ)) &
      stop 'xmolecule_excitation_energies: occ has wrong len'
    Norb = size(Xdata%F%CE)
    gs_occ = groundstate_occ(Xdata%F%N_occ(:), Ncore=Xdata%F%Ncore)
    if (.not. Xdata%P%yes_valgs_occ .or. sum(gs_occ) .ne. sum(Xdata%F%N_occ(:)) + 1 ) &
      stop 'only works with Koopmann s theorem states!'
    Nval = count(gs_occ .eq. 2) - Xdata%F%Ncore
    if (l_E .ne. Nval) then
      l_E = -Nval
      return
    end if
    call calculate_hole_energies_gradients_nacs(Xdata%P, Xdata%A, Xdata%B, Xdata%F, &
      occs, energies, gradients, nacs)
    do i = 1, size(occs, 2)
      if (all(occs(:, i) .eq. -1)) then
        l_E = i - 1
        exit
      end if
    end do
    if (l_E .ne. size(nacs, 2)) then
      stop 'invalid value of l_E'
    end if
  end subroutine xmolecule_hole_occs_energies_gradients_nacs

  subroutine xmolecule_part_occs_energies_gradients_nacs(Xdata_cptr, occs, energies, gradients, nacs, l_E, l_gradient, l_occ) bind(c)
    !! returns for particle configurations energies, occupation patterns, gradients and nacs in the presence of electric field based on Koopmann's theorem
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: occs(l_occ, l_E)
    real(c_double), intent(inout) :: energies(l_E)
    real(c_double), intent(inout) :: gradients(3*l_gradient, l_E)
    real(c_double), intent(inout) :: nacs(3*l_gradient, l_E, l_E)
    integer(c_int), intent(inout) :: l_E
    integer(c_int), intent(inout) :: l_gradient
    integer(c_int), intent(inout) :: l_occ
    integer :: gs_occ(l_occ)
    integer :: Norb, Nvirt, i
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_excitation_energies: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ)) stop 'xmolecule_excitation_energies: occ has wrong len'
    Norb = size(Xdata%F%CE)
    gs_occ = groundstate_occ(Xdata%F%N_occ(:), Ncore=Xdata%F%Ncore, add_electrons=.False.)
    if (.not. Xdata%P%yes_valgs_occ .or. sum(gs_occ) .ne. sum(Xdata%F%N_occ(:)) - 1 ) &
      stop 'only works with Koopmann s theorem states!'
    Nvirt = count(gs_occ .eq. 0)
    if (l_E .ne. Nvirt) then
      l_E = -Nvirt
      return
    end if
    call calculate_hole_energies_gradients_nacs(Xdata%P, Xdata%A, Xdata%B, Xdata%F, occs, energies, gradients,nacs, particles=.True.)
    do i = 1, size(occs, 2)
      if (all(occs(:, i) .eq. -1)) then
        l_E = i - 1
        exit
      end if
    end do
    if (l_E .ne. size(nacs, 2)) then
      stop 'invalid value of l_E'
    end if
  end subroutine xmolecule_part_occs_energies_gradients_nacs

  subroutine xmolecule_excitation_energies(Xdata_cptr, occs, energies, l_E, l_occ) bind(c)
    !! returns excitation energies based on MO energies
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: energies(l_E)
    integer(c_int), intent(inout) :: occs(l_occ, l_E)
    integer(c_int), intent(inout) :: l_E
    integer(c_int), intent(inout) :: l_occ
    integer, allocatable :: gs_occ(:)
    integer :: i, j, k
    integer :: Norb
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_excitation_energies: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ)) stop 'xmolecule_excitation_energies: occ has wrong len'
    Norb = size(Xdata%F%CE)
    allocate (gs_occ(size(Xdata%F%N_occ)))
    gs_occ = groundstate_occ(Xdata%F%N_occ(:), Ncore=Xdata%F%Ncore)

    if (Xdata%P%yes_valgs_occ .and. sum(gs_occ) .eq. sum(Xdata%F%N_occ(:)) + 1) then
      ! we have hole state Koopmann's theorem
      k = 1
      do i = Norb, Xdata%F%Ncore + 1, -1
        if (gs_occ(i) .ne. 2) cycle
        occs(1:size(Xdata%F%N_occ), k) = gs_occ(:)
        occs(i, k) = 1
        Energies(k) = Xdata%F%E - Xdata%F%CE(i)
        k = k + 1
      end do
    else
      k = 1
      do i = Xdata%F%Ncore + 1, Norb
        if (Xdata%F%N_occ(i) .eq. 0) cycle
        do j = Xdata%F%Ncore + 1, Norb
          if (Xdata%F%N_occ(j) .eq. 2) cycle
          occs(1:size(Xdata%F%N_occ), k) = Xdata%F%N_occ(:)
          if (Xdata%P%yes_process_spin) then
            occs(size(Xdata%F%N_occ) + 1, k) = Xdata%F%multp
            occs(size(Xdata%F%N_occ) + 2, k) = Xdata%F%csf_num
          end if
          occs(j, k) = occs(j, k) + 1
          occs(i, k) = occs(i, k) - 1
          Energies(k) = Xdata%F%CE(j) - Xdata%F%CE(i)
          k = k + 1
        end do
      end do
    end if
    l_E = k - 1
    deallocate (gs_occ)
  end subroutine xmolecule_excitation_energies

  subroutine xmolecule_CIS_energies(Xdata_cptr, energies, l_E) bind(c)
    !! returns CIS excitation energies
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: energies(l_E)
    integer(c_int), intent(inout) :: l_E
    integer :: l
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call CIS_procedure(Xdata%P, Xdata%A, Xdata%G, Xdata%B, Xdata%F)
    l = min(l_E, size(Xdata%F%E_cis) + 1)
    energies(1) = Xdata%F%E_tot
    energies(2:l) = Xdata%F%E_tot + Xdata%F%E_cis(1:l - 1)
    if (l_E .ne. size(Xdata%F%E_cis) + 1) l_E = -(size(Xdata%F%E_cis) + 1)
  end subroutine xmolecule_CIS_energies

  subroutine xmolecule_set_ci_state(Xdata_cptr, state) bind(c)
    !! set CI state index
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Xdata%F%state = state
  end subroutine xmolecule_set_ci_state

  subroutine xmolecule_ci_overlaps(Xdata_cptr, Olap, l_E) bind(c)
    !! calculated CI state overlaps
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: Olap(l_E, l_E)
    integer(c_int), intent(inout) :: l_E
    integer :: l
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (.not. allocated(Xdata%F%O_ci) .or. .not. Xdata%F%yes_calculate_CI_overlap) then
      Olap = ZERO
    else
      l = min(l_E, size(Xdata%F%O_ci, 1))
      if (l_E .lt. size(Xdata%F%O_ci, 1)) l_E = -(size(Xdata%F%O_ci, 1))
      Olap(1:l, 1:l) = Xdata%F%O_ci(1:l, 1:l)
    end if
  end subroutine xmolecule_ci_overlaps

  subroutine xmolecule_CI_energies(Xdata_cptr, energies, l_E) bind(c)
    !! get CI state energies
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: energies(l_E)
    integer(c_int), intent(inout) :: l_E
    integer :: l
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    l = min(l_E, size(Xdata%F%E_ci))
    energies(1:l) = Xdata%F%E_ci(1:l)
    if (l_E .ne. size(Xdata%F%E_ci)) l_E = -(size(Xdata%F%E_ci))
  end subroutine xmolecule_CI_energies

  subroutine xmolecule_get_CIS_gradient(Xdata_cptr, state, grad, l_grad) bind(c)
    !! compute the CIS gradient
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    real(c_double), intent(inout) :: grad(3*l_grad)
    integer(c_int), intent(inout) :: l_grad
    integer :: Natom
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Natom = size(Xdata%A)
    if (l_grad .ne. Natom) stop 'gradient has wrong len'
    if (Xdata%P%yes_gaussian .and. Xdata%P%yes_HF) then
      call calculate_CIS_gradient_GTO(Xdata%P, Xdata%A, Xdata%B, Xdata%F, state, gradients_out=grad)
    else
      stop 'CIS gradients only possible with gaussian+RHF'
    end if
  end subroutine xmolecule_get_CIS_gradient

  subroutine xmolecule_get_CIS_gradient_nacs(Xdata_cptr, state, grad, l_grad, nacs, l_nacs) bind(c)
    !! compute CIS gradients and nacs
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    real(c_double), intent(inout) :: grad(3*l_grad)
    integer(c_int), intent(inout) :: l_grad
    real(c_double), intent(inout) :: nacs(3*l_grad, l_nacs, l_nacs)
    integer(c_int), intent(inout) :: l_nacs

    integer :: Natom
    CALL C_F_POINTER(Xdata_cptr, Xdata)

    Natom = size(Xdata%A)
    if (l_grad .ne. Natom) stop 'gradient has wrong len'
    if (l_nacs .ne. size(Xdata%F%E_CIS, 1) + 1) then
      l_nacs = -(size(Xdata%F%E_CIS, 1) + 1)
      !     write(6,*) 'size of CIS_excitation',size(Xdata%F%E_CIS,1)
      !     write(6,*) 'size of l_nacs',l_nacs
      return
    else
      l_nacs = size(Xdata%F%E_CIS, 1) + 1
    end if

    if (Xdata%P%yes_gaussian .and. Xdata%P%yes_HF) then
      call calculate_cis_gradient_GTO(Xdata%P, Xdata%A, Xdata%B, Xdata%F, state, gradients_out=grad,&
        nacs_out=nacs,yes_calc_nacs=.True.)
    else
      stop 'CIS gradients only possible with gaussian+RHF'
    end if
  end subroutine xmolecule_get_cis_gradient_nacs

  subroutine xmolecule_get_cis_transdens(Xdata_cptr, tdens, state, l_ao) bind(c)
    !! compute the cis state transition density
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    real(c_double), intent(inout) :: tdens(l_ao, l_ao)
    integer(c_int), intent(in) :: l_ao
    integer :: k, l
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (state .eq. 0) then
      tdens(:, :) = ZERO
    else
      ! first index is particle; second index is hole
      forall (k=1:size(Xdata%F%LC, 1), l=1:size(Xdata%F%LC, 1)) &
        tdens(k, l) = sum(Xdata%F%C_cis(:, state)*Xdata%F%LC(k, Xdata%F%CIS_excitation(2, :))*Xdata%F%LC(l, Xdata%F%CIS_excitation(1, :)))
    end if
  end subroutine xmolecule_get_CIS_transdens

  subroutine xmolecule_get_gradient(Xdata_cptr, grad, l_grad) bind(c)
    !! compute the HFS, HF, or ROHF gradient
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: grad(3, l_grad)
    integer(c_int), intent(inout) :: l_grad
    integer :: i
    integer :: Natom
    CALL C_F_POINTER(Xdata_cptr, Xdata)

    if (Xdata%P%fix_geom) then
      grad(:, :) = ZERO
    else
      Natom = size(Xdata%A)
      if (l_grad .ne. Natom) stop 'gradient has wrong len'
      if (Xdata%P%yes_gaussian) then
        if (Xdata%P%yes_valgs_occ) then
          call calculate_Koopmann_energy_gradient(Xdata%P, Xdata%A, Xdata%B, Xdata%F)
        else
          call calculate_energy_gradient_GTO(Xdata%P, Xdata%A, Xdata%G, Xdata%B, Xdata%F)
        end if
      end if
      if (.not. Xdata%P%yes_gaussian .and. .not. Xdata%P%yes_HF) then
        call calculate_energy_gradient(Xdata%P, Xdata%A, Xdata%G, Xdata%B, Xdata%F)
      end if
      do i = 1, l_grad
        grad(1, i) = Xdata%F%grad(3*(i - 1) + 1)
        grad(2, i) = Xdata%F%grad(3*(i - 1) + 2)
        grad(3, i) = Xdata%F%grad(3*(i - 1) + 3)
      end do
    end if! fix_geom
    !  call calculate_dEdR(gradient, Xdata%P,Xdata%A,Xdata%G,Xdata%B,Xdata%F, Jnuc_want, Natom)
    !  deallocate(Jnuc_want)

  end subroutine xmolecule_get_gradient

  subroutine xmolecule_get_auger_ci(Xdata_cptr, f_irrep, f_multp, f_index, rate, energy, N_process) bind(c)
    !! return auger decay rates based on configuration interation
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: f_irrep(N_process)
    integer(c_int), intent(inout) :: f_multp(N_process)
    integer(c_int), intent(inout) :: f_index(N_process)
    real(c_double), intent(inout) :: rate(N_process)     ! cross sec or rate in a.u., 1...N_p
    real(c_double), intent(inout):: energy(N_process)         ! transition energy in eV, 1...N_proc
    integer(c_int), intent(inout) :: N_process
    type(Process_data) :: S
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    S = create_Process(Xdata%F%N_occ, Xdata%P%N_proc_max, spin=Xdata%P%yes_process_spin, ci=Xdata%P%yes_process_ci)

    call calcAugerRates_CI(Xdata%P, Xdata%A, Xdata%B, Xdata%F, Xdata%G)
    stop "not implemented"
    if (S%N_process .gt. N_process) then
      N_process = S%N_process
    else
      N_process = S%N_process
      if (N_process .gt. 0) then
        rate(1:N_process) = S%value(1:N_process)
        energy(1:N_process) = S%E_au(1:N_process)
        f_irrep(1:N_process) = S%final_irrep(1:N_process)
        f_multp(1:N_process) = S%final_spin_multp(1:N_process)
        f_index(1:N_process) = S%final_state_num(1:N_process)
      end if
    end if
    call purge_Process(S)
  end subroutine xmolecule_get_auger_ci

  subroutine xmolecule_get_process(Xdata_cptr, final_occs, E_eV, val, tag, N_process, l_occ) bind(c)
    !! compute process data
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    character(c_char), intent(inout) :: tag(N_process) ! process name tag, 1...N_proc
    real(c_double), intent(inout) :: val(N_process)     ! cross sec or rate in a.u., 1...N_p
    integer(c_int), intent(inout):: E_eV(N_process)         ! transition energy in eV, 1...N_proc
    integer(c_int), intent(inout):: final_occs(l_occ, N_process)
    integer(c_int), intent(inout) :: N_process
    integer(c_int), intent(inout) :: l_occ
    type(Process_data) :: S
    integer :: i
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_set_occ: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ .ne. size(Xdata%F%N_occ)) stop 'xmolecule_set_occ: occ has wrong len'

    call calculate_rates(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, S)

    if (S%N_process .gt. N_process) then
      N_process = S%N_process
    else
      N_process = S%N_process
      final_occs(:, :) = 0
      do i = 1, N_process
        final_occs(1:size(Xdata%F%N_occ), i) = occ_Process(S, i)
        if (Xdata%P%yes_process_spin) then
          final_occs(size(Xdata%F%N_occ) + 1, i) = S%final_spin_multp(i)
          final_occs(size(Xdata%F%N_occ) + 2, i) = S%final_csf_num(i)
        end if
      end do

      E_eV(:) = 0
      val(:) = ZERO
      tag(:) = " "

      if (N_process .gt. 0) then
        E_eV(1:N_process) = S%E_ev(1:N_process)
        val(1:N_process) = S%value(1:N_process)
        tag(1:N_process) = S%tag(1:N_process)
      end if
    end if

  end subroutine xmolecule_get_process

  subroutine xmolecule_get_population(Xdata_cptr, pop, l_mo, l_atom) bind(c)
    !! compute orbital population analysis
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: l_mo, l_atom
    real(c_double), intent(inout):: pop(l_mo, l_atom)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call gross_population_analysis(Xdata%P, Xdata%A, Xdata%B, Xdata%F, OUTPUT=Xdata%P%OUTPUT, Gross_pop=pop)
  end subroutine xmolecule_get_population

  subroutine xmolecule_get_ao_labels(Xdata_cptr, ao_labels, l_ao) bind(c)
    !! return atomic orbital labels
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: l_ao
    character(c_char), intent(inout):: ao_labels(l_ao*10)
    integer :: j, i
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (l_ao .ne. size(Xdata%F%LC, 1)) then
      l_ao = size(Xdata%F%LC, 1)
    else
      ao_labels(:) = ''
      do j = 1, l_ao
        do i = 1, 10
          ao_labels((j - 1)*10 + i) = Xdata%B%label(j) (i:i)
        end do
      end do
    end if
  end subroutine xmolecule_get_ao_labels

  subroutine xmolecule_get_overlapmatrix(Xdata_cptr, S_ao, l_ao) bind(c)
    !! return overlap matrix
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: l_ao
    real(c_double), intent(inout):: S_ao(l_ao, l_ao)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (l_ao .ne. size(Xdata%F%LC, 1)) then
      l_ao = size(Xdata%F%LC, 1)
    else
      S_ao(:, :) = Xdata%B%S(:, :)
    end if
  end subroutine xmolecule_get_overlapmatrix

  subroutine xmolecule_get_mo_coef(Xdata_cptr, mo_coef, l_mo, l_ao) bind(c)
    !! return LCAO coefficients
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: l_mo, l_ao
    real(c_double), intent(inout):: mo_coef(l_ao, l_mo)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (l_ao .ne. size(Xdata%F%LC, 1) .or. l_mo .ne. size(Xdata%F%LC, 2)) then
      l_ao = size(Xdata%F%LC, 1)
      l_mo = size(Xdata%F%LC, 2)
    else
      mo_coef(:, :) = Xdata%F%LC(:, :)
    end if
  end subroutine xmolecule_get_mo_coef

  subroutine xmolecule_set_mo_coef(Xdata_cptr, mo_coef, l_ao, l_mo) bind(c)
    !! set LCAO coefficients
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(in):: mo_coef(l_ao, l_mo)
    integer(c_int), intent(in) :: l_mo, l_ao
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Xdata%F%LC_ini(:, :) = mo_coef(:, :)
    Xdata%F%LC(:, :) = mo_coef(:, :)
  end subroutine xmolecule_set_mo_coef

  subroutine xmolecule_get_nao_conf(Xdata_cptr, nao_conf_string, l_atom) bind(c)
    !! returns numerical atomic orbital configuration string
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: l_atom
    character(c_char), intent(inout):: nao_conf_string(l_atom * 512)
    integer :: i, j_start, j_end, c
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    j_start = 0
    j_end = 0
    nao_conf_string = ''
    do i = 1, l_atom
      j_end = j_start + len(trim(Xdata%A(i)%conf_string))
      do c = 1, len(trim(Xdata%A(i)%conf_string))
        nao_conf_string(j_start + c) = Xdata%A(i)%conf_string(c:c)
      end do
      nao_conf_string(j_end + 1) = ','
      j_start = j_end + 1
    end do
    nao_conf_string(j_end + 1) = char(0) ! add '\0' character to terminate string in C
  end subroutine xmolecule_get_nao_conf

  subroutine xmolecule_get_mullikencharge(Xdata_cptr, charge, l_charge) bind(c)
    !! returns mulliken charges
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: charge(l_charge)
    integer(c_int), intent(inout) :: l_charge
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call Mulliken_population(Xdata%A, Xdata%B, Xdata%F, charge)
    charge(:) = -charge(:) + dble(Xdata%A(:)%charge)
  end subroutine xmolecule_get_mullikencharge

  subroutine xmolecule_get_loewdincharge(Xdata_cptr, charge, l_charge) bind(c)
    !! returns loewdin charges
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: charge(l_charge)
    integer(c_int), intent(inout) :: l_charge
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call Loewdin_population(Xdata%A, Xdata%B, Xdata%F, charge)
    charge(:) = -charge(:) + dble(Xdata%A(:)%charge)
  end subroutine xmolecule_get_loewdincharge

  subroutine xmolecule_get_cis_mullikenpop(Xdata_cptr, charge, l_charge, state, type) bind(c)
    !! returns mulliken population for CIS
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: charge(l_charge)
    integer(c_int), intent(inout) :: l_charge
    integer(c_int), intent(inout) :: state
    character(c_char), intent(in) :: type
    real(long), allocatable :: D(:, :)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    allocate (D(Xdata%B%N_basis, Xdata%B%N_basis))
    D = cis_densitymatrix(Xdata%B, Xdata%F, state, type=type)
    call Mulliken_population(Xdata%A, Xdata%B, Xdata%F, charge, dmatrix=D)
    deallocate (D)
    if (type .eq. 'F') charge(:) = -charge(:) + dble(Xdata%A(:)%charge)
  end subroutine xmolecule_get_cis_mullikenpop

  subroutine xmolecule_get_cis_loewdinpop(Xdata_cptr, charge, l_charge, state, type) bind(c)
    !! returns Loewdin population for CIS
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: charge(l_charge)
    integer(c_int), intent(inout) :: l_charge
    integer(c_int), intent(inout) :: state
    character(c_char), intent(in) :: type
    real(long), allocatable :: D(:, :)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    allocate (D(Xdata%B%N_basis, Xdata%B%N_basis))
    D = cis_densitymatrix(Xdata%B, Xdata%F, state, type=type)
    call Loewdin_population(Xdata%A, Xdata%B, Xdata%F, charge, dmatrix=D)
    deallocate (D)
    if (type .eq. 'F') charge(:) = -charge(:) + dble(Xdata%A(:)%charge)
  end subroutine xmolecule_get_cis_loewdinpop

  subroutine xmolecule_get_bondorder(Xdata_cptr, BO, n_atom) bind(c)
    !! returns bond order analysis
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: BO(n_atom, n_atom)
    integer(c_int), intent(inout) :: n_atom
    real(long), allocatable :: D(:, :)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    allocate (D(Xdata%B%N_basis, Xdata%B%N_basis))
    if (allocated(Xdata%F%D_CI)) then
      D = Xdata%F%D_CI
    else
      call density_matrix(Xdata%F%LC, dble(Xdata%F%N_occ), D)
    end if
    call bondOrder_analysis(Xdata%P, Xdata%B, D, 0, BO)
    deallocate (D)
  end subroutine xmolecule_get_bondorder

  subroutine xmolecule_get_spin_analysis(Xdata_cptr, center, rog) bind(c)
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout):: center(3)
    real(c_double), intent(inout):: rog
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call spin_analysis(Xdata%P, Xdata%G, Xdata%B, Xdata%F%Da, Xdata%F%Db, 0, center, rog)
  end subroutine xmolecule_get_spin_analysis

  subroutine xmolecule_get_cis_oscillator_strength(Xdata_cptr, state, os, tDIP, nstates) bind(c)
    !! returns cis oscillator strengths
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    real(c_double), intent(inout) :: os(nstates)
    real(c_double), intent(inout) :: tDip(nstates, 3)
    integer(c_int), intent(inout) :: nstates
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (state .gt. nstates) stop 'wrong state number'
    if (state .lt. 0) stop 'wrong state number'
    if (.not. allocated(Xdata%F%E_cis)) &
      stop "CIS calculation has not been initiated yet."
    call calc_oscillator_strength_CIS(Xdata%P, Xdata%A, Xdata%B, Xdata%F, Xdata%G, state_in=state, os=os, tDip=tDip)
  end subroutine xmolecule_get_cis_oscillator_strength

  subroutine xmolecule_get_oscillator_strength(Xdata_cptr, os, tDip, nmo) bind(c)
    !! returns oscillator strength
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: os(nmo, nmo)
    real(c_double), intent(inout) :: tDip(nmo, nmo, 3)
    integer(c_int), intent(inout) :: nmo
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call calc_oscillator_strength(Xdata%P, Xdata%B, Xdata%F, Xdata%G, os, tDip)
  end subroutine xmolecule_get_oscillator_strength

  subroutine xmolecule_get_fluorescencerates(Xdata_cptr, fluorates, nmo) bind(c)
    !! returns fluorescence rates
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: fluorates(nmo, nmo)
    integer(c_int), intent(inout) :: nmo
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call calc_Fluorescence_rates(Xdata%P, Xdata%B, Xdata%F, Xdata%G, Xdata%S, fluorates)

  end subroutine xmolecule_get_fluorescencerates

  subroutine xmolecule_write_molden(Xdata_cptr, filename, lenFilename) bind(c)
    !! writes molden file
    type(c_ptr), intent(inout) :: Xdata_cptr
    character(c_char), intent(inout) :: filename(lenFilename) !! input filename
    integer(c_int), intent(in) :: lenFilename !! length of input filename
    character(len=lenFilename) :: filename_string
    integer :: IOUT
    type(xmol_data), pointer :: Xdata
    filename_string = transfer(filename, filename_string)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call write_molden_file(Xdata%P, Xdata%A, Xdata%B, Xdata%F, filename_string)
  end subroutine xmolecule_write_molden

  subroutine xmolecule_write_density_cube(Xdata_cptr) bind(c)
    !! writes density cube file
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call write_cube_file(Xdata%P, Xdata%A, Xdata%B, Xdata%F, "dens.cube")
  end subroutine xmolecule_write_density_cube

  subroutine xmolecule_write_mo_cube(Xdata_cptr, mo) bind(c)
    !! writes mo cub file
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: mo
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    call write_cube_file(Xdata%P, Xdata%A, Xdata%B, Xdata%F, "mo.cube", mo)
  end subroutine xmolecule_write_mo_cube

  subroutine xmolecule_get_geom(Xdata_cptr, geom, l_geom) bind(c)
    !! returns geometry
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(inout) :: geom(3, l_geom)
    integer(c_int), intent(inout) :: l_geom
    integer :: i
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    do i = 1, l_geom
      geom(1, i) = Xdata%A(i)%x
      geom(2, i) = Xdata%A(i)%y
      geom(3, i) = Xdata%A(i)%z
    end do
    l_geom = size(Xdata%A, 1)
  end subroutine xmolecule_get_geom

  subroutine xmolecule_set_reinit(Xdata_cptr, yes) bind(c)
    !! triggers reinitialization for the next set_geom
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: yes
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (yes .eq. 0) then
      Xdata%F%yes_reinit = .False.
    else
      Xdata%F%yes_reinit = .True.
    end if
  end subroutine xmolecule_set_reinit

  subroutine xmolecule_set_calculate_ci_overlap(Xdata_cptr, yes) bind(c)
    !! triggers calculation of CI overlap
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: yes
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (yes .eq. 0) then
      Xdata%F%yes_calculate_CI_overlap = .False.
    else
      Xdata%F%yes_calculate_CI_overlap = .True.
    end if
  end subroutine xmolecule_set_calculate_ci_overlap

  subroutine xmolecule_set_geom(Xdata_cptr, geom, l_geom) bind(c)
    !! sets geometry
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    real(c_double), intent(in) :: geom(3, l_geom)
    integer(c_int), intent(inout) :: l_geom
    integer, parameter :: nstep = 5
    Real(long), allocatable:: geom_save(:, :)
    type(WF_data) :: F_old, F_save
    real(long), allocatable :: B_func_old(:, :)
    logical :: poor_but_converged = .false.
    integer*1 :: ifault
    integer :: i, j, Natom
    integer :: aa
    character(len=20) :: fb_type_save
    character(len=20) :: fb_type
    integer :: IOUT
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    IOUT = Xdata%P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2
    allocate (geom_save(3, size(Xdata%A)))
    Natom = size(Xdata%A)
    if (l_geom .ne. Natom) stop 'geom has wrong len'

    if (Xdata%P%yes_verbose) &
      write (Xdata%P%OUTPUT, '("# set geometry")')
    poor_but_converged = .FALSE.
    F_old = Xdata%F
    allocate (B_func_old(Xdata%G%N_grid, Xdata%B%N_basis))
    B_func_old = Xdata%B%func(:, :)
    ! save the old geometry
    geom_save(1, 1:Natom) = Xdata%A(1:Natom)%x
    geom_save(2, 1:Natom) = Xdata%A(1:Natom)%y
    geom_save(3, 1:Natom) = Xdata%A(1:Natom)%z

    Xdata%A(1:Natom)%x = geom(1, 1:Natom)
    Xdata%A(1:Natom)%y = geom(2, 1:Natom)
    Xdata%A(1:Natom)%z = geom(3, 1:Natom)
    if (Xdata%F%yes_reinit) then
      Xdata%F%iSCFstate = 0
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F)
      F_old = Xdata%F
      B_func_old = Xdata%B%func(:, :)
      Xdata%F%yes_reinit = .False.
    end if
    Xdata%F%iSCFstate = 2
    Xdata%F%LC_ini = F_old%LC
    if (Xdata%P%yes_unrestricted) &
      Xdata%F%LC_ini_b = F_old%LC_b
    Xdata%F%basis_ini = B_func_old

    fb_type = "bare SCF"
    call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, &
      set_new_geometry=.true., set_new_naos=.false., &
      freeze=.false.)
    call check_result(Xdata%F, Xdata%P, ifault)
    if (ifault .ne. 0) write (IOUT, '("# FALLBACK  GEOM STEP: ")')
    if (ifault .gt. 0) then
      write (IOUT, '("# FALLBACK: keeping this result")')
      poor_but_converged = .TRUE.
      F_save = Xdata%F
      fb_type_save = fb_type
    end if

    if (ifault .ne. 0) then
      !    loop_auto_avg: do aa=1,2 ! disabled auto-averaging
      loop_auto_avg: do aa = 1, 1
        Xdata%F = F_old
        Xdata%F%iSCFstate = 2
        if (Xdata%P%yes_unrestricted) &
          Xdata%F%LC_ini_b = F_old%LC_b
        Xdata%F%basis_ini = B_func_old
        if (aa .eq. 1) write (IOUT, '("# FALLBACK -> SMOOTH STEPS")')
        if (aa .eq. 1) fb_type = "smooth"
        if (aa .eq. 2) write (IOUT, '("# FALLBACK -> SMOOTH STEPS WITH AUTO AVERAGING")')
        if (aa .eq. 2) fb_type = "smooth+avg"
        nsteps: do i = 0, nstep
          if (aa .eq. 2) then
            call automatic_averaging(Xdata%P, Xdata%B, Xdata%F, Xdata%F%av_occ)
          end if
          Xdata%A(1:Natom)%x = geom_save(1, 1:Natom) + ONE*i/nstep*(geom(1, 1:Natom) - geom_save(1, 1:Natom))
          Xdata%A(1:Natom)%y = geom_save(2, 1:Natom) + ONE*i/nstep*(geom(2, 1:Natom) - geom_save(2, 1:Natom))
          Xdata%A(1:Natom)%z = geom_save(3, 1:Natom) + ONE*i/nstep*(geom(3, 1:Natom) - geom_save(3, 1:Natom))
          write (IOUT, '("# FALLBACK STEP:",i2)') i
          call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, set_new_geometry=.true., set_new_naos=.false.)
          call check_result(Xdata%F, Xdata%P, ifault)
          if (ifault .eq. 0) then ! only if this step gets a really good result use it for the next guess
            Xdata%F%LC_ini = Xdata%F%LC
            if (Xdata%P%yes_unrestricted) &
              Xdata%F%LC_ini_b = Xdata%F%LC_b                  
            Xdata%F%basis_ini = Xdata%B%func
            if (Xdata%P%yes_unrestricted) then
              Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
              Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
              Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
            else
              Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
              Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
            end if
          else
            if(Xdata%P%yes_unrestricted) then
              Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
              Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
              Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
            else
              Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
              Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
            end if
          end if
        end do nsteps
        call check_result(Xdata%F, Xdata%P, ifault)
        !write(IOUT,'("# FALLBACK: ifault =",i3)') ifault
        if (ifault .eq. 0) then
          if(Xdata%P%yes_unrestricted) then
            Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
            Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
            Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
          else
            Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
            Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
          end if
          write (IOUT, '("# FALLBACK: smooth steps done. prod overlap",f15.5)') Xdata%F%prod_overlap
          exit loop_auto_avg
        else if (ifault .gt. 0) then
          if(Xdata%P%yes_unrestricted) then
            Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)**2
            Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
            Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
          else
            Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
            Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
          end if
          write (IOUT, '("# FALLBACK: smooth steps done. prod overlap",f15.5)') Xdata%F%prod_overlap
          if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
            write (IOUT, '("# FALLBACK: keeping this result")')
            F_save = Xdata%F
            poor_but_converged = .TRUE.
            fb_type_save = fb_type
          end if
        else if (ifault .eq. -1) then
          Xdata%F%prod_overlap = ZERO
          Xdata%F%fidelity(:) = ZERO
          Xdata%F%fidelityb(:) = ZERO
          write (IOUT, '("# FALLBACK: smooth steps failed")')
        else if (ifault .lt. -1) then
          if(Xdata%P%yes_unrestricted) then
            Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
            Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
            Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
          else
            Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
            Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
          end if
          write (IOUT, '("# FALLBACK: failed. low prod overlap",f15.5)') Xdata%F%prod_overlap
        end if
      end do loop_auto_avg
    end if

    ! NEW FALLBACK STRATEGY HERE
    if (ifault .ne. 0 .and. .NOT. Xdata%P%yes_test_fallback) then
      write (IOUT, '("#   FALLBACK: PARTIAL FREEZING MOS + DISABLING MOM")')
      fb_type = "partial freezing"
      Xdata%F = F_old
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      Xdata%A(1:Natom)%x = geom(1, 1:Natom)
      Xdata%A(1:Natom)%y = geom(2, 1:Natom)
      Xdata%A(1:Natom)%z = geom(3, 1:Natom)
      Xdata%F%iSCFstate = 2
      Xdata%F%LC_ini = F_old%LC
      if (Xdata%P%yes_unrestricted) &
        Xdata%F%LC_ini_b = F_old%LC_b
      Xdata%F%basis_ini = B_func_old
      call calculate_electrons_with_partial_freezing(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F)
      call check_result(Xdata%F, Xdata%P, ifault)

      if (ifault .eq. 0) then
        if(Xdata%P%yes_unrestricted) then
          Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
          Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
          Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
        else
          Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
          Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
        end if
        write (IOUT, '("# FALLBACK: partial freezing+releasing done. prod overlap=",f15.5)') Xdata%F%prod_overlap
      else if (ifault .gt. 0) then
        if(Xdata%P%yes_unrestricted) then
          Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
          Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
          Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
        else
          Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
          Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
        end if
        write (IOUT, '("# FALLBACK: partial freezing+releasing done. prod overlap=",f15.5)') Xdata%F%prod_overlap
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("# FALLBACK: keeping this result")')
          F_save = Xdata%F
          poor_but_converged = .TRUE.
          fb_type_save = fb_type
        end if
      else if (ifault .eq. -1) then
        Xdata%F%prod_overlap = ZERO
        Xdata%F%fidelity(:) = ZERO
        Xdata%F%fidelityb(:) = ZERO
        write (IOUT, '("# FALLBACK: partial freezing+releasing failed")')
      else if (ifault .lt. -1) then
        if(Xdata%P%yes_unrestricted) then
          Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)**2
          Xdata%F%fidelity = fidelity(Xdata%F%LC_a, Xdata%B%S, F_old%LC_a)
          Xdata%F%fidelityb = fidelity(Xdata%F%LC_b, Xdata%B%S, F_old%LC_b)
        else
          Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
          Xdata%F%fidelity = fidelity(Xdata%F%LC, Xdata%B%S, F_old%LC)
        end if
        write (IOUT, '("# FALLBACK: partial freezing failed. low prod overlap", f15.5)') &
          Xdata%F%prod_overlap
      end if
    end if
    if (ifault .ne. 0) then
      fb_type = "freeze"
      write (IOUT, '("# FALLBACK: FREEZING MOS")')
      Xdata%F = F_old
      Xdata%F%iSCFstate = 2
      Xdata%F%LC_ini = F_old%LC
      if (Xdata%P%yes_unrestricted) &
        Xdata%F%LC_ini_b = F_old%LC_b
      Xdata%F%basis_ini = B_func_old
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, freeze=.TRUE.)
      !call Gross_population_analysis( Xdata%A, Xdata%B, Xdata%F, Xdata%P%OUTPUT )

      call check_result(Xdata%F, Xdata%P, ifault)
      !write(IOUT,'("# FALLBACK: ifault =",i3)') ifault
    else if (ifault .ne. 0 .and. poor_but_converged .and. .FALSE.) then  ! disabled: only use results that are properly converged otherwise freeze
      write (IOUT, '("# FALLBACK: RESTORING PREVIOUS RESULT")')
      Xdata%F = F_save
      Xdata%F%LC_ini = Xdata%F%LC
      if (Xdata%P%yes_unrestricted) &
        Xdata%F%LC_ini_b = Xdata%F%LC_b      
      Xdata%F%iSCFstate = 2
      fb_type = fb_type_save
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, set_new_geometry=.FALSE., freeze=.FALSE.)
      call check_result(Xdata%F, Xdata%P, ifault)
      if(Xdata%P%yes_unrestricted) then
        Xdata%F%prod_overlap = prodOverlapUnrestricted(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC_a, F_old%LC_b)
      else
        Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%B%S, F_old%LC)
      end if
    end if
    deallocate (B_func_old)

    if (Xdata%P%yes_CI) then
      if (allocated(F_old%C_CI)) then
        call CI_procedure(Xdata%P, Xdata%A(:), Xdata%B, Xdata%F, F2=F_old)
      else
        call CI_procedure(Xdata%P, Xdata%A(:), Xdata%B, Xdata%F)
      end if
    end if
    ! store old orbital coefficients and CIS coefficients for overlap calculations
    if (allocated(F_old%C_cis)) then
      if (allocated(Xdata%F%LC_old)) deallocate (Xdata%F%LC_old)
      if (allocated(Xdata%F%C_cis_old)) deallocate (Xdata%F%C_cis_old)
      allocate (Xdata%F%C_cis_old(size(F_old%C_cis, 1), size(F_old%C_cis, 2)))
      allocate (Xdata%F%LC_old(size(F_old%LC, 1), size(F_old%LC, 2)))
      Xdata%F%LC_old = F_old%LC
      Xdata%F%C_cis_old = F_old%C_cis
    end if

    write (IOUT, '("# GEOM STEP: ",a20,"prod overlap",f15.5, x, "E=",f10.3 )') &
      fb_type, Xdata%F%prod_overlap, Xdata%F%E_tot
    deallocate (geom_save)
  end subroutine xmolecule_set_geom

  subroutine xmolecule_write_chkpt(Xdata_cptr) bind(c)
    !! writes chkpt file
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer :: i, RWStatus
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    open (unit=90, form='unformatted', file="chkpt.dat", IOSTAT=RWStatus)
    do i = 1, size(Xdata%A)
      write (90) Xdata%A(i)%x, Xdata%A(i)%y, Xdata%A(i)%z, Xdata%A(i)%conf_string
    end do
    write (90) Xdata%F%N_occ
    if (allocated(Xdata%F%av_occ)) then
      write (90) size(Xdata%F%av_occ, 1)
      do i = 1, size(Xdata%F%av_occ, 1)
        write (90) size(Xdata%F%av_occ(i)%MO, 1)
        write (90) Xdata%F%av_occ(i)%MO(:)
      end do
    else
      write (90) 0
    end if
    write (90) Xdata%F%LC
    close (90)
    call save_wf(Xdata%F%LC, "chkpt_wf.dat")
    !call save_basis(Xdata%B,"chkpt_basis.dat" )

    open (unit=91, form='formatted', file="chkpt_input.in", IOSTAT=RWStatus)
    do i = 1, size(Xdata%A)
      write(91,'(a6,x,f15.5,f15.5,f15.5,x,a,x,a)') Xdata%A(i)%label,Xdata%A(i)%x,Xdata%A(i)%y,Xdata%A(i)%z,trim(Xdata%P%nao_name(i)),trim(Xdata%A(i)%nao_file_nuc)
    end do
    write (91, '("occ=",99i1)') Xdata%F%N_occ(1:Xdata%F%homo_idx)
    close (91)

  end subroutine xmolecule_write_chkpt

  subroutine xmolecule_read_chkpt(Xdata_cptr) bind(c)
    !! reads chkpt file
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer :: i, j, n, RWStatus
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    open (unit=90, form='unformatted', file="chkpt.dat", IOSTAT=RWStatus, status='OLD', ACTION='READ')

    do i = 1, size(Xdata%A)
      read (90) Xdata%A(i)%x, Xdata%A(i)%y, Xdata%A(i)%z, Xdata%A(i)%conf_string
    end do

    if (allocated(Xdata%F%av_occ)) then
      do j = 1, size(Xdata%F%av_occ)
        deallocate (Xdata%F%av_occ(j)%MO)
      end do
    end if

    read (90) Xdata%F%N_occ
    read (90) n
    if (n .ne. 0) then
      if (allocated(Xdata%F%av_occ)) &
        deallocate (Xdata%F%av_occ)
      allocate (Xdata%F%av_occ(n))
      do i = 1, size(Xdata%F%av_occ, 1)
        read (90) n
        allocate (Xdata%F%av_occ(i)%MO(n))
        read (90) Xdata%F%av_occ(i)%MO(:)
      end do
    end if
    read (90) Xdata%F%LC
    close (90)

    Xdata%F%LC_ini = Xdata%F%LC
    Xdata%F%iSCFstate = 2
    call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, &
      set_new_geometry=.TRUE., freeze=.FALSE., &
      set_new_naos=.TRUE.)
  end subroutine xmolecule_read_chkpt

  subroutine xmolecule_get_occ(Xdata_cptr, occ, l_occ) bind(c)
    !! returns occupation numbers
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(inout) :: occ(l_occ)
    integer(c_int), intent(inout) :: l_occ
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (Xdata%P%yes_process_spin) then
      if (l_occ .eq. size(Xdata%F%N_occ) + 2) then
        occ(size(Xdata%F%N_occ) + 1) = Xdata%F%multp
        occ(size(Xdata%F%N_occ) + 2) = Xdata%F%csf_num
      end if
      l_occ = size(Xdata%F%N_occ) + 2
    else
      if (l_occ .gt. 0) occ(1:min(l_occ, size(Xdata%F%N_occ))) = Xdata%F%N_occ(1:size(Xdata%F%N_occ))
      l_occ = size(Xdata%F%N_occ)
    end if
  end subroutine xmolecule_get_occ

  subroutine xmolecule_set_multp(Xdata_cptr, multp_in) bind(c)
    !! set spin multiplicity
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: multp_in
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Xdata%F%multp = multp_in
  end subroutine xmolecule_set_multp

  subroutine xmolecule_set_irrep(Xdata_cptr, irrep_in) bind(c)
    !! set irreducible representation
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: irrep_in
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    Xdata%F%state_irrep = irrep_in
  end subroutine xmolecule_set_irrep

  subroutine cistate2occ(Xdata_cptr, state, occ, l_occ) bind(c)
    !! returns the dominant configuration of the CI state with state index state
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: state
    integer(c_int), intent(inout) :: occ(l_occ)
    integer(c_int), intent(inout) :: l_occ
    integer :: dominant_csf, i
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (.not. allocated(Xdata%F%C_ci)) &
      stop "Please run CI calculation first"

    write (6, *) '# CI state', state
    do i = 1, size(Xdata%F%C_ci, 1)
      if (Xdata%F%C_ci(i, state)**2 .gt. 0.01) write (6, "('# ',f15.3,x,99i2)") Xdata%F%C_ci(i, state)**2, Xdata%F%occs_CI(:, i)
    end do

    dominant_csf = maxloc(Xdata%F%C_ci(:, state)**2, dim=1)
    l_occ = size(Xdata%F%occs_ci, 1)

    occ = Xdata%F%occs_ci(:, dominant_csf)
  end subroutine cistate2occ

  subroutine occ2cistate(Xdata_cptr, occ_in, l_occ_in, state) bind(c)
    !! randomly selects a CI states according to overlap with given occupation
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: occ_in(l_occ_in)
    integer(c_int), intent(in) :: l_occ_in
    integer(c_int), intent(out) :: state
    integer(c_int), allocatable :: occ(:)
    integer :: i, j
    real(long) :: r
    real(long), allocatable :: csf_occ_overlap(:)
    real(long), allocatable :: state_occ_overlap(:)
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    if (.not. allocated(Xdata%F%C_ci)) &
      stop "Please run CI calculation first"
    allocate (occ(size(Xdata%F%N_occ)))
    if (Xdata%P%yes_process_spin .and. l_occ_in .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_set_occ: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ_in .ne. size(Xdata%F%N_occ)) stop 'xmolecule_set_occ: occ has wrong len'
    occ(:) = occ_in(1:size(occ))

    allocate (csf_occ_overlap(size(Xdata%F%C_ci, 1)))
    allocate (state_occ_overlap(size(Xdata%F%C_ci, 1)))

    write (6, '("# ",a,x,99i1)') 'occ=', occ
    do j = 1, size(Xdata%F%C_ci, 1)
      if (all(occ(:) .eq. Xdata%F%occs_CI(:, j))) then
        csf_occ_overlap(j) = ONE
      else
        csf_occ_overlap(j) = ZERO
      end if
    end do

    do i = 1, size(Xdata%F%C_ci, 2)
      state_occ_overlap(i) = sum(Xdata%F%C_ci(:, i)**2*csf_occ_overlap(:))
    end do

    state = 0
    call RANDOM_NUMBER(r)

    r = r*sum(state_occ_overlap(:))
    write (6, *) "r=", r
    do i = 1, size(state_occ_overlap)
      if (state_occ_overlap(i) .gt. 0.001) write (6, *) '#', i, state_occ_overlap(i)
    end do
    write (6, *) "sum", sum(state_occ_overlap(:))
    write (6, *) "N_ci", size(state_occ_overlap)

    do i = 1, size(state_occ_overlap)
      if (r .gt. state_occ_overlap(i)) then
        r = r - state_occ_overlap(i)
      else
        state = i
        write (6, *) '# state ', i, state_occ_overlap(i)
        exit
      end if
    end do

    if (state .eq. 0) stop "error state==0"
    if (size(state_occ_overlap) .eq. 0) stop "Nstate==0"

    deallocate (csf_occ_overlap)
    deallocate (state_occ_overlap)
    deallocate (occ)
  end subroutine occ2cistate

  recursive subroutine xmolecule_set_occ(Xdata_cptr, occ_in, l_occ_in) bind(c)
    !! sets occupation numbers
    type(c_ptr), intent(inout) :: Xdata_cptr
    type(xmol_data), pointer :: Xdata
    integer(c_int), intent(in) :: occ_in(l_occ_in)
    integer(c_int), intent(in) :: l_occ_in

    type(WF_data) :: F_old, F_save
    real(long), allocatable :: B_func_old(:, :)
    logical :: converged = .FALSE.
    integer*1 :: ifault
    integer :: i, j
    character(len=30) :: fb_type_save
    character(len=30) :: fb_type
    integer :: IOUT
    integer, allocatable :: occ(:)
    integer :: multp, csf_num

    integer :: occ_shake(l_occ_in)
    integer, allocatable :: occs(:, :)
    real(long), allocatable :: overlap2(:)
    real(long), allocatable :: LCrot(:, :)
    integer, allocatable :: active_space(:)
    real(long) :: r
    CALL C_F_POINTER(Xdata_cptr, Xdata)
    ! if occupation is the same as before return
    if (ALL(Xdata%F%N_occ .eq. occ_in)) return

    if (Xdata%P%yes_verbose) &
      write (Xdata%P%OUTPUT, '("# set occ")')
    allocate (occ(size(Xdata%F%N_occ, 1)))

    IOUT = Xdata%P%OUTPUT2      ! all outputs are redirected to P%OUTPUT2

    if (Xdata%P%yes_process_spin .and. l_occ_in .ne. size(Xdata%F%N_occ) + 2) stop 'xmolecule_set_occ: occ has wrong len'
    if (.NOT. Xdata%P%yes_process_spin .and. l_occ_in .ne. size(Xdata%F%N_occ)) stop 'xmolecule_set_occ: occ has wrong len'

    occ(:) = occ_in(1:size(occ))
    multp = -1
    csf_num = -1
    if (Xdata%P%yes_process_spin) then
      multp = occ_in(size(occ) + 1)
      csf_num = occ_in(size(occ) + 2)
    end if

    converged = .FALSE.
    F_old = Xdata%F
    allocate (B_func_old(Xdata%G%N_grid, Xdata%B%N_basis))
    B_func_old = Xdata%B%func(:, :)

    Xdata%F%LC_ini = F_old%LC
    Xdata%F%basis_ini(:, :) = B_func_old
    Xdata%F%iSCFstate = 1
    Xdata%F%N_occ(:) = occ(:)
    if (Xdata%P%yes_sym .and. allocated(Xdata%P%irrep_MO)) then
      Xdata%F%state_irrep = 1
      do i = 1, size(Xdata%F%N_occ)
        if (Xdata%F%N_occ(i) .eq. 1 ) &
          Xdata%F%state_irrep = Xdata%B%irrep_prod(Xdata%F%state_irrep, Xdata%F%irrep_MO(i))
      end do
      write (IOUT, '("# ","new irrep ",i4)') Xdata%F%state_irrep
    end if
    if (Xdata%P%yes_process_spin) then
      Xdata%F%multp = multp
      Xdata%F%csf_num = csf_num
    end if
    fb_type = "bare SCF"
    call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, freeze=.FALSE., set_new_geometry=.false., set_new_naos=.true.)

    !  if(Xdata%P%yes_shakeup)then
    !    do i = 1, size(Xdata%F%N_occ)
    !      do j = 1, size(Xdata%F%N_occ)
    !        Xdata%F%STrans(i,j) = integral( Xdata%G, Xdata%B%func(:,i) * B_func_old(:,j) )
    !      end do
    !    end do
    !    call active_space_occs(Xdata%F%N_occ,Xdata%P%active_space_MO,occs)
    !    allocate(overlap2(size(occs,2)))
    !    overlap2=occ_overlap2(Xdata%F%N_occ,occs,Xdata%F%LC,F_old%LC,Xdata%F%STrans)
    !    Xdata%F%guess_overlap=sqrt(sum(overlap2(:)))
    !    deallocate(overlap2)
    !    deallocate(occs)
    !  end if

    F_save%prod_overlap = ZERO

    call check_result(Xdata%F, Xdata%P, ifault)
    if (Xdata%P%yes_shakeup) then
      Xdata%F%prod_overlap = prodOverlap(Xdata%P, Xdata%F, Xdata%F%STrans, F_old%LC)
    end if
    if (ifault .ne. 0) write (IOUT, '("#   FALLBACK ELEC STEP: ")')
    if (ifault .ne. -1) then
      write (IOUT, '("#   FALLBACK: keeping this result")')
      converged = .TRUE.
      F_save = Xdata%F
      fb_type_save = fb_type
    end if

    if (ifault .ne. 0) then
      write (IOUT, '("#   FALLBACK: SMOOTH STEPS")')
      fb_type = "smooth"
      Xdata%F = F_old
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old
      Xdata%F%LC = F_old%LC
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, &
        set_new_geometry=.false., set_new_naos=.true., freeze=.true.)
      Xdata%F%N_occ = occ
      if (Xdata%P%yes_process_spin) then
        Xdata%F%multp = multp
        Xdata%F%csf_num = csf_num
      end if
      Xdata%F%N_occ_old = F_old%N_occ
      call calculate_electrons_in_smooth_steps(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, auto_average=.FALSE.)
      call check_result(Xdata%F, Xdata%P, ifault)
      if (ifault .eq. -1) then
        write (IOUT, '("#   FALLBACK: smooth steps failed")')
      else if (ifault .ne. -1) then
        write (IOUT, '("#   FALLBACK: smooth steps done. prod overlap",f15.5)') &
          Xdata%F%prod_overlap
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          F_save = Xdata%F
          converged = .TRUE.
          fb_type_save = fb_type
        end if
      end if
    end if

    ! disabled fallback
    if (ifault .ne. 0 .and. .FALSE.) then
      write (IOUT, '("#   FALLBACK: SMOOTH STEPS + RSTRCT")')
      fb_type = "smooth"
      Xdata%F = F_old
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old
      Xdata%F%LC = F_old%LC
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, &
        set_new_geometry=.false., set_new_naos=.true., freeze=.true.)
      Xdata%F%N_occ = occ
      if (Xdata%P%yes_process_spin) then
        Xdata%F%multp = multp
        Xdata%F%csf_num = csf_num
      end if
      Xdata%F%N_occ_old = F_old%N_occ
      Xdata%F%yes_mom = .False.
      Xdata%F%switch_rstrct_on = .True.
      call calculate_electrons_in_smooth_steps(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, auto_average=.FALSE.)
      Xdata%F%yes_mom = .True.
      Xdata%F%switch_rstrct_on = .False.

      call check_result(Xdata%F, Xdata%P, ifault)
      if (ifault .eq. -1) then
        write (IOUT, '("#   FALLBACK: smooth steps + rstrct failed")')
      else if (ifault .ne. -1) then
        write (IOUT, '("#   FALLBACK: smooth steps + rstrct done. prod overlap",f15.5)') &
          Xdata%F%prod_overlap
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          F_save = Xdata%F
          converged = .TRUE.
          fb_type_save = fb_type
        end if
      end if
    end if

    !  if ( ifault /= 0 .and. .not. Xdata%P%yes_valgs_occ .and. .FALSE. ) then ! commented out auto-averaging
    if (ifault .ne. 0 .and. .not. Xdata%P%yes_valgs_occ .and. .FALSE.) then
      write (IOUT, '("#   FALLBACK: SMOOTH STEPS + AUTO AVERAGING")')
      fb_type = "smooth+avg"
      Xdata%F = F_old
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old
      Xdata%F%LC = F_old%LC
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, &
        set_new_geometry=.false., set_new_naos=.true., freeze=.true.)
      Xdata%F%N_occ = occ
      if (Xdata%P%yes_process_spin) then
        Xdata%F%multp = multp
        Xdata%F%csf_num = csf_num
      end if
      Xdata%F%N_occ_old = F_old%N_occ
      call calculate_electrons_in_smooth_steps(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, auto_average=.TRUE.)
      write (IOUT, '("#   FALLBACK: smooth steps done. prod overlap",f15.5)') &
        Xdata%F%prod_overlap
      call check_result(Xdata%F, Xdata%P, ifault)
      if (ifault .eq. -1) then
        write (IOUT, '("#   FALLBACK: smooth steps+AA failed")')
      else if (ifault .ne. -1) then
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          F_save = Xdata%F
          converged = .TRUE.
          fb_type_save = fb_type
        end if
      end if
    end if

    if (ifault .ne. 0 .and. .NOT. converged .and. .NOT. Xdata%P%yes_test_fallback) then
      write (IOUT, '("#   FALLBACK: PARTIAL FREEZING MOS + DISABLING MOM")')
      fb_type = "partial freezing"
      Xdata%F = F_old
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do

      Xdata%F%N_occ = occ
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old

      call calculate_electrons_with_partial_freezing(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F)
      call check_result(Xdata%F, Xdata%P, ifault)
      if (ifault .eq. -1) then
        write (IOUT, '("# FALLBACK: partial freezing+releasing failed")')
      else if (ifault .ne. -1) then
        write (IOUT, '("# FALLBACK: partial freezing+releasing done. prod overlap",f15.5)') &
          Xdata%F%prod_overlap
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          F_save = Xdata%F
          fb_type_save = fb_type
          converged = .TRUE.
        end if
      end if
    end if

    ! another level: combine partial freeying and smooth steps (disabled)
    if (ifault .ne. 0 .and. .NOT. converged .and. .NOT. Xdata%P%yes_test_fallback .and. .FALSE.) then
      write (IOUT, '("#   FALLBACK: SMOOTH STEPS + PARTIAL FREEZING MOS + DISABLING MOM")')
      fb_type = "smooth steps+partial freezing"
      Xdata%F = F_old
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      Xdata%F%N_occ = occ
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old
      call calculate_electrons_in_smooth_steps(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, auto_average=.FALSE., partial_freezing=.TRUE.)
      call check_result(Xdata%F, Xdata%P, ifault)
      if (ifault .eq. -1) then
        write (IOUT, '("# FALLBACK: partial freezing+releasing failed")')
      else if (ifault .ne. -1) then
        write (IOUT, '("# FALLBACK: partial freezing+releasing done. prod overlap", f15.5)') &
          Xdata%F%prod_overlap
        if (Xdata%F%prod_overlap .gt. F_save%prod_overlap) then
          write (IOUT, '("#   FALLBACK: keeping this result")')
          F_save = Xdata%F
          fb_type_save = fb_type
          converged = .TRUE.
        end if
      end if
    end if

    if (ifault .ne. 0 .and. ((.NOT. converged) .or. (Xdata%P%yes_freezeBadOverlap .and. ifault .ne. 0))) then
      write (IOUT, '("#   FALLBACK: FREEZING MOS")')
      fb_type = "freeze"
      Xdata%F = F_old
      Xdata%F%N_occ = occ
      if (Xdata%P%yes_process_spin) then
        Xdata%F%multp = multp
        Xdata%F%csf_num = csf_num
      end if
      Xdata%F%LC_ini = F_old%LC
      Xdata%F%basis_ini = B_func_old
      Xdata%F%LC = F_old%LC
      Xdata%F%iSCFstate = 1
      Xdata%B%func = B_func_old
      do i = 1, Xdata%B%N_basis
        do j = i, Xdata%B%N_basis
          Xdata%B%S(i, j) = integral(Xdata%G, Xdata%F%basis_ini(:, i)*Xdata%F%basis_ini(:, j))
          Xdata%B%S(j, i) = Xdata%B%S(i, j)
        end do
      end do
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, set_new_naos=.TRUE., freeze=.TRUE.)
      call check_result(Xdata%F, Xdata%P, ifault)
      !write(IOUT,'("# FALLBACK: ifault =",i3)') ifault
    end if
    if (ifault .ne. 0 .and. converged) then
      write (IOUT, '("#   FALLBACK: RESTORING PREVIOUS RESULT")')
      fb_type = fb_type_save
      Xdata%F = F_save
      Xdata%F%LC_ini = Xdata%F%LC
      Xdata%F%iSCFstate = 2
      call calculate_electrons(Xdata%P, Xdata%A, Xdata%B, Xdata%G, Xdata%F, set_new_geometry=.FALSE., freeze=.FALSE.)
      Xdata%F%prod_overlap = F_save%prod_overlap
      !write(IOUT,'("# FALLBACK: ifault =",i3)') ifault
    end if

    write (IOUT, '("# ELEC STEP: ",a20,x,"prod overlap",f15.5,x, "E=",f10.3)') &
      fb_type, Xdata%F%prod_overlap, Xdata%F%E_tot
    Xdata%F%N_occ_old = Xdata%F%N_occ

    ! Now we have a solution (may not be converged)

    if (Xdata%P%yes_shakeup) then
      if (.not. Xdata%P%yes_gaussian) then
        do i = 1, Xdata%B%N_basis
          do j = 1, Xdata%B%N_basis
            Xdata%F%STrans(i, j) = integral(Xdata%G, Xdata%B%func(:, i)*B_func_old(:, j))
          end do
        end do
      else
        Xdata%F%STrans = Xdata%B%S
      end if
      !    call single_excitation_occs(Xdata%F%N_occ,occs,lower_limit=Xdata%F%Ncore+1)

      allocate (LCrot(size(Xdata%F%LC, 1), size(Xdata%F%LC, 2)))
      LCrot = Xdata%F%LC
      if (allocated(Xdata%F%av_occ)) &
        call rotate_av_orbitals(Xdata%F%av_occ, F_old%LC, Xdata%F%Strans, LCrot)
      call auto_active_space(Xdata%F%N_occ, LCrot, F_old%LC, Xdata%F%Strans, active_space, Xdata%F%Ncore)
      write (IOUT, '("# ",a,x,99i3)') 'active_space:', active_space
      call active_space_occs(Xdata%F%N_occ, active_space, occs)
      !    call active_space_occs(Xdata%F%N_occ,Xdata%P%active_space_MO,occs)
      allocate (overlap2(size(occs, 2)))
      !    call occ_overlap2(Xdata%F%N_occ,occs,LCrot,F_old%LC,Xdata%F%STrans,overlap2)
      call occ_overlap2_highspin_core(Xdata%F%N_occ, occs, LCrot, F_old%LC, Xdata%F%STrans, Xdata%F%Ncore, overlap2)
      !    overlap2=occ_overlap2_high_spin(Xdata%F%N_occ,occs,LCrot,F_old%LC,Xdata%F%STrans)
      deallocate (active_space)
      call RANDOM_NUMBER(r)
      r = r*sum(overlap2(:))

      do i = 1, size(occs, 2)
        if (overlap2(i) .gt. 0.1 .or. is_equivalent_occ(Xdata%F, Xdata%F%N_occ, occs(:, i) )) &
          write (IOUT, '("#SHAKE ",f15.5,x,999i1)') overlap2(i), occs(:, i)
      end do

      write (IOUT, '("#SHAKE sum=",f15.5)') sum(overlap2(:))

      shakeup: do i = 1, size(occs, 2)
        if (r .le. overlap2(i)) then
          if (ANY(occ .ne. occs(:, i))) then
            if (allocated(Xdata%F%av_occ)) then
              if (is_equivalent_occ(Xdata%F, occ, occs(:, i))) then
                exit shakeup
              end if
            end if
            write (IOUT, '("#SHAKE ","shake up from: ",99i1)') occ
            write (IOUT, '("#SHAKE ","shake up to  : ",99i1)') occs(:, i)
            occ_shake(1:size(occs, 1)) = occs(:, i)
            if (Xdata%P%yes_process_spin) then
              occ_shake(size(occs, 1) + 1) = Xdata%F%multp
              occ_shake(size(occs, 1) + 2) = Xdata%F%csf_num
            end if
            call xmolecule_set_occ(Xdata_cptr, occ_shake, l_occ_in)
          else
            write (IOUT, '("#SHAKE ","no shake up ")')
          end if
          exit shakeup
        end if
        r = r - overlap2(i)
      end do shakeup
      deallocate (occs)
      deallocate (overlap2)
      deallocate (LCrot)
    end if! shake up

    deallocate (B_func_old)

    if (Xdata%P%yes_CI) then
      call CI_procedure(Xdata%P, Xdata%A(:), Xdata%B, Xdata%F)
      !call CI_olap(Xdata%P,Xdata%B,Xdata%F,F_old,Xdata%F%O_CI,OUTPUT=1)
    end if

    deallocate (occ)
  end subroutine xmolecule_set_occ

  subroutine select_MO_subset(F, not_frozen)
    type(WF_data), intent(in) :: F
    integer, intent(out), allocatable :: not_frozen(:)
    integer :: i, j, k, l
    real(long) :: delta = ZERO
    real(long) :: H(size(F%LC, 2), size(F%LC, 2))
    real(long), parameter :: EPSILON = 1e-4

    ! Strategy: find the pair of MOs that have different occupation (are not part of the same averaging block) and the larges non-diagonal Fock matrix element
    allocate (not_frozen(2))

    H = F%LC(:, :) .tm.F%H(:, :) .m.F%LC(:, :)
    !  do i=1,size(F%CE)
    !    write(6,'(99f10.3)') H(i,:)
    !  end do

    delta = ZERO
    do i = F%Ncore + 1, size(F%CE)
      PAIR: do j = i + 1, size(F%CE)
        ! skip orbital pairs in the same averaging region:
        if (allocated(F%av_occ)) then
          do k = 1, size(F%av_occ)
            if (any(F%av_occ(k)%MO(:) .eq. i) .and. any(F%av_occ(k)%MO(:) .eq. j)) cycle PAIR
          end do
        end if
        !      write(6,*) i,j,abs(H(i,j)),F%N_occ(i)/=F%N_occ(j)
        if (abs(H(i, j))/(abs(H(i, i) - H(j, j)) + EPSILON) .gt. delta .and. F%N_occ(i) .ne. F%N_occ(j)) then
          delta = abs(H(i, j))/(abs(H(i, i) - H(j, j)) + EPSILON)
          not_frozen(1) = i
          not_frozen(2) = j
        end if
      end do PAIR
    end do

    ! in case the selected orbitals are part of an averaging block,
    ! include the whole block
    if (allocated(F%av_occ)) then
      do k = 1, size(F%av_occ)
        if (any(F%av_occ(k)%MO(:) .eq. not_frozen(1))) exit
      end do
      do l = 1, size(F%av_occ)
        if (any(F%av_occ(l)%MO(:) .eq. not_frozen(2))) exit
      end do
      if (k .le. size(F%av_occ) .and. l .gt. size(F%av_occ)) then
        j = not_frozen(2)
        deallocate (not_frozen)
        allocate (not_frozen(size(F%av_occ(k)%MO) + 1))
        not_frozen(1:size(F%av_occ(k)%MO)) = F%av_occ(k)%MO(:)
        not_frozen(size(F%av_occ(k)%MO) + 1) = j
      else if (k .gt. size(F%av_occ) .and. l .le. size(F%av_occ)) then
        i = not_frozen(1)
        deallocate (not_frozen)
        allocate (not_frozen(size(F%av_occ(l)%MO) + 1))
        not_frozen(1) = i
        not_frozen(2:size(F%av_occ(l)%MO)) = F%av_occ(l)%MO(:)
      else if (k .le. size(F%av_occ) .and. l .le. size(F%av_occ)) then
        i = not_frozen(1)
        j = not_frozen(2)
        deallocate (not_frozen)
        allocate (not_frozen(size(F%av_occ(k)%MO) + size(F%av_occ(l)%MO)))
        not_frozen(1:size(F%av_occ(k)%MO)) = F%av_occ(k)%MO(:)
        not_frozen(size(F%av_occ(k)%MO) + 1:) = F%av_occ(l)%MO(:)
      end if

    end if

  end subroutine select_MO_subset

  subroutine check_result(F, P, iFault)
    type(WF_data) :: F
    type(Param), intent(in) :: P     ! the usual structure
    integer*1, intent(inout) :: iFault
    ifault = 0
    if (F%ierr_o .ne. 0) then
      ifault = -1 ! not converged
    else if (F%prod_overlap .lt. P%guess_overlap_TOL(1)) then
      ifault = -2 ! converged but with error tolerance
    else if (F%prod_overlap .lt. P%guess_overlap_TOL(2)) then
      ifault = 2  ! converged but with warning tolerance
    end if
  end subroutine check_result

end module xmolecule_interface
