#!/usr/bin/env python3

import os
import re
import unittest
from subprocess import PIPE, Popen, call
from scipy.special import erf
import numpy as np


class TestEmbedding(unittest.TestCase):
    xmol = ""
    xatom = ""
    xmol_input = "input.in"
    calc_dict = {
        "HF": "yes",
        "gto": "6-31G_star_star",
        "XATOM_PATH": xatom,
        "Sini": "yes",
        "coreini": "yes",
        "Rini": "yes",
        "EPS0": "0.",
        "EPS1": "0.",
        "occ": "222220",
        "SCF_energy_conv": "1E-7",
        "SCF_density_conv": "1E-7",
        "gradient": "yes",
    }

    @classmethod
    def setUpClass(cls):
        cls.xatom = os.getenv("XATOM_PATH")
        if not os.path.isfile(cls.xatom):
            print("Please specify xatom path!")
            exit(-1)
        print("Using xatom =", cls.xatom)

        test_dir = os.path.dirname(os.path.realpath(__file__))
        cls.xmol = os.path.join(test_dir, "../src/xmolecule")
        print("Using xmolecule =", cls.xmol)
        print("")

    @staticmethod
    def write_xmolecule_input_file(atoms_dict, params_dict, filename="input.in"):
        """This function writes the input file to run xmolecule

        Parameters
        ----------
        atoms_dict : Dict
            The dictionary containing the atoms and their coordinates,
            the key is the atom name and the value is a list of the
            coordinates.
        params_dict : Dict
            The dictionary containing the parameters for the calculation,
            the key is the parameter name and the value is the value of
            the parameter.
        filename : str, optional
            Input file name, by default "input.in"
        """

        with open(filename, "w") as f:
            for atom, coord in atoms_dict.items():
                coord_fstring = "".join(["{c:<20.10f} ".format(c=c) for c in coord])
                f.write("{atom:2}  {coord_fstring}\n".format(atom=atom,coord_fstring=coord_fstring))

            for key, value in params_dict.items():
                f.write("{key}={value}\n".format(key=key,value=value))

    def coulomb_interaction_fcharge_Na(self, param_test, q, s=None):
        """
        This functions checks that the interaction between a fractional
        charge and a Na^+ ion behaves like the Coulomb interaction
        between two point charges at large distances.
        """

        coord_dict = {"Na": [0.0, 0.0, 0.0]}
        distances = [np.inf] + list(np.linspace(10, 1, 81))
        energies = np.zeros(len(distances))
        gradients = np.zeros((len(distances), 3))

        for n, dist in enumerate(distances):

            # Copy dictionaries and modify them for each run
            coord_write = coord_dict.copy()
            param_write = self.calc_dict.copy()
            if param_test["mmap"]:
                param_write["mmap_teints"] = "yes"
            else:
                param_write["mmap_teints"] = "no"
            if param_test["HF"]:
                param_write["HF"] = "yes"
            else:
                param_write.pop("HF")
            if param_test["gto"]:
                param_write["gto"] = "6-31G_star_star"
            else:
                param_write.pop("gto")
                param_write["XATOM_PATH"] = self.xatom
            if n == 0:
                param_write["coreini"] = "yes"
                param_write.pop("Rini")
            elif s is None:
                param_write.pop("coreini")
                coord_write["X"] = [0.0, 0.0, dist, q]
            elif s is not None:
                param_write.pop("coreini")
                coord_write["XG"] = [0.0, 0.0, dist, q, s]

            if len(coord_write) == 1 and "gto" not in param_write:
                # If only Na is present and no GTOs are used run XATOM
                cmd = "rm Na.nao"
                call(cmd, shell=True)
                cmd = (
                    self.xatom
                    + " -silent -latter off -s Na -charge 1 -rmax  10.0 -N 50 -nao -file_nao Na.nao -nev 3"
                )
                print("calling " + cmd)
                call(cmd, shell=True)

            # Write input file
            self.write_xmolecule_input_file(
                coord_write, param_write, filename=self.xmol_input
            )

            # Run xmolecule and capture stdout
            cmd = "{xmol} {xmol_input}".format(
                xmol=self.xmol,
                xmol_input=self.xmol_input)
            pipe = Popen(cmd, stdout=PIPE, shell=True)
            out = pipe.communicate()[0].decode("UTF8")
            str_en = re.search("Total Energy.*", out)
            str_grad = re.search("GRAD\| Na1.*", out)
            pipe.stdout.close()

            # Save energies and gradients
            energies[n] = float(str_en.group(0).split()[3])
            if str_grad != None:
                gradients[n, :] = [
                    -float(str_grad.group(0).split()[2]),
                    -float(str_grad.group(0).split()[3]),
                    -float(str_grad.group(0).split()[4]),
                ]
            else:
                gradients[n, :] = [0.0, 0.0, 0.0]

        ## Print results and perform tests
        print("# Parameters HF={p1} gto={p2}".format(
            p1=param_test['HF'], p2=param_test['gto']))
        print("# Energy for no point charge: {e:11.7e} eV".format(
            e=energies[0]))
        print("")

        print("# Deviation to the pure Coulombic interaction")
        print("# {d:^12s} {E:^12s} {qd:^12s} {diff:^12s}".format(
            d='d',E='E',qd='q/d',diff='diff'))
        for i in range(3, len(distances)):
            en_inter = energies[i] - energies[0]
            coul_inter = q / distances[i]
            print(
                "{d:12.6f} {en_inter:12.6e} {coul_inter:12.6e} {ec:12.6e}".format(
                    d=distances[i],
                    en_inter=en_inter,
                    coul_inter=coul_inter,
                    ec=(en_inter - coul_inter)
                )
            )
            if distances[i] > 6.5 and distances[i] < 10.0:
                self.assertAlmostEqual(en_inter, coul_inter, 5)
        print("")

        print("# Difference Numerical - Analytical gradient")
        print("# {d:^12s} {ng:^12s} {ag:^12s} {diff:^12s}".format(d='d',ng='num grad', ag='anal grad', diff='diff'))
        for i in range(3, len(distances)):
            num_grad_z = (energies[i] - energies[i - 1]) / (
                distances[i] - distances[i - 1]
            )
            anal_grad_z = gradients[i, 2]
            print(
                "{distances:12.6f} {num_grad_z:12.6e} {anal_grad_z:12.6e} {na:12.6e}".format(
                    distances=distances[i],
                    num_grad_z=num_grad_z,
                    anal_grad_z=anal_grad_z,
                    na=(num_grad_z - anal_grad_z)
            ))

            # if( param['gto'] or (not param['gto'] and not param['HF'] ) ):
            #     self.assertAlmostEqual(num_grad_z, anal_grad_z, 2)
        print("")

    def test_coulomb_point_charge_Na(self):
        """
        Calls the function coulomb_interaction_fcharge_Na for the system
        Na^+ and a point charge for different types of calculations.
        """

        print("Testing Na^+ plus Gaussian distributed charge")
        q = 0.1  # Fractional charge
        param_loop = [
            {"HF": True, "gto": True, "mmap": False},
            {"HF": True, "gto": False, "mmap": False},
            {"HF": False, "gto": True, "mmap": False},
            {"HF": False, "gto": False, "mmap": False},
            {"HF": True, "gto": True, "mmap": True},
            {"HF": True, "gto": False, "mmap": True},
            {"HF": False, "gto": True, "mmap": True},
            {"HF": False, "gto": False, "mmap": True},
        ]

        for param in param_loop:
            self.coulomb_interaction_fcharge_Na(param, q)

        return

    def test_coulomb_gaussian_charge_Na(self):
        """
        Calls the function coulomb_interaction_fcharge_Na for the system
        Na^+ and a Gaussian charge distribution for different types of
        calculations.
        """

        print("Testing Na^+ plus Gaussian charge distribution")
        q = 0.1  # Fractional charge
        s = 0.5  # Gaussian width
        param_loop = [
            {"HF": True, "gto": True, "mmap": False},
            {"HF": True, "gto": False, "mmap": False},
            {"HF": False, "gto": True, "mmap": False},
            {"HF": False, "gto": False, "mmap": False},
            {"HF": True, "gto": True, "mmap": True},
            {"HF": True, "gto": False, "mmap": True},
            {"HF": False, "gto": True, "mmap": True},
            {"HF": False, "gto": False, "mmap": True},
        ]

        for param in param_loop:
            self.coulomb_interaction_fcharge_Na(param, q, s=s)

        return

    def test_gaussian_nuclear_repulsion(self, mmap=False):
        q = 0.1  # Fractional charge
        s = 0.5  # Gaussian width
        coord_dict = {"H": [0.0, 0.0, 0.0], "XG": [0.0, 0.0, 0.0, q, s]}
        distances = np.linspace(0.01, 5, 81)
        energies = np.zeros(len(distances))
        gradients = np.zeros((len(distances), 3))

        for n, dist in enumerate(distances):

            coord_write = coord_dict.copy()
            param_write = self.calc_dict.copy()
            if mmap:
                param_write["mmap_teints"] = "yes"
            param_write["occ"] = "000000"
            coord_dict["XG"][2] = dist

            # Write input file
            self.write_xmolecule_input_file(
                coord_write, param_write, filename=self.xmol_input
            )

            # Run xmolecule and capture stdout
            cmd = "{xmol} {xmol_input}".format(
                xmol=self.xmol,
                xmol_input=self.xmol_input)
            pipe = Popen(cmd, stdout=PIPE, shell=True)
            out = pipe.communicate()[0].decode("UTF8")
            str_en = re.search("Total Energy.*", out)
            str_grad = re.search("GRAD\| H1.*", out)
            pipe.stdout.close()

            # Save energies and gradients
            energies[n] = float(str_en.group(0).split()[3])
            if str_grad != None:
                gradients[n, :] = [
                    float(str_grad.group(0).split()[2]),
                    float(str_grad.group(0).split()[3]),
                    float(str_grad.group(0).split()[4]),
                ]
            else:
                gradients[n, :] = [0.0, 0.0, 0.0]

        ## Print results and perform tests ##
        print("# Energy deviation from analytical solution")
        print("# {d:^12s} {xmolecule:^12s} {analytic:^12s} {diff:^12s}".format(d='d', xmolecule='xmolecule', analytic='analytic', diff='diff'))
        for n, dist in enumerate(distances):
            en_coul_anal = (q / dist) * erf(dist/np.sqrt(2*s**2))
            print(
                "{dist:12.6f} {energies:12.6e} {en_coul_anal:12.6e} {ec:12.6e}".format(
                    dist=dist,
                    energies=energies[n],
                    en_coul_anal=en_coul_anal,
                    ec=(energies[n] - en_coul_anal)
                )
            )
            self.assertAlmostEqual(energies[n], en_coul_anal, 7)
        print("")

        print("# Gradient deviation from analytical solution")
        print("# {d:^12s} {xmolecule:^12s} {analytic:^12s} {diff:^12s}".format(d='d', xmolecule='xmolecule', analytic='analytic', diff='diff'))
        for n, dist in enumerate(distances):
            grad_coul_anal = -(q / dist**2) * (
                2 / np.sqrt(2 * np.pi * s**2) * dist * np.exp(-dist**2/(2*s**2)) - erf(dist/np.sqrt(2*s**2))
            )
            grad_xmol = gradients[n, 2]
            print(
                "{dist:12.6f} {grad_xmol:12.6e} {grad_coul_anal:12.6e} {diff:12.6e}".format(
                    dist=dist,
                    grad_xmol=grad_xmol,
                    grad_coul_anal=grad_coul_anal,
                    diff=(grad_xmol - grad_coul_anal)
                )
            )
            self.assertAlmostEqual(grad_xmol, grad_coul_anal, 7)
        print("")

    def test_gaussian_nuclear_repulsion_mmap(self, mmap=False):
        self.test_gaussian_nuclear_repulsion(mmap=True)

    def test_gaussian_charge_water_HF_GTO(self, mmap=False):
        """
        Calculates the energy and gradient of a system with a Ne atom
        with two fractional charges at 1.0 and -1.0 bohr from the
        center of the Ne atom. The charges are Gaussian distributed
        with and a point charge. The obtained values are compared to
        the reference values (GIT VERSION).
        """

        energy_ref = -126.19881440
        gradient_Ne_reg = [0.0, 0.0, -0.55055485]
        gradient_X_reg = [0.0, 0.0, 2.79111676]
        gradient_XG_reg = [0.0, 0.0, -2.24056191]

        coord_dict = {
            "Ne": [0.0, 0.0, 0.0],
            "X": [0.0, 0.0, -1.0, 1],
            "XG": [0.0, 0.0, 1.0, 1, 0.6],
        }

        # Xmolecule parameters
        param_dict = self.calc_dict.copy()
        if mmap:
            param_dict["mmap_teints"] = "yes"
        param_dict["gto"] = "6-31G"
        # Remove unncessary parameters
        param_dict.pop("Sini")
        param_dict.pop("Rini")
        param_dict.pop("coreini")

        # Write input file
        self.write_xmolecule_input_file(
            coord_dict, param_dict, filename=self.xmol_input
        )

        # Run xmolecule and capture stdout
        cmd = "{xmol} {xmol_input}".format(
            xmol=self.xmol,
            xmol_input=self.xmol_input)
        pipe = Popen(cmd, stdout=PIPE, shell=True)
        out = pipe.communicate()[0].decode("UTF8")
        str_en = re.search("Total Energy.*", out)
        str_grad_Ne = re.search("GRAD\| Ne.*", out)
        str_grad_X = re.search("GRAD\| X.*", out)
        str_grad_XG = re.search("GRAD\| XG.*", out)
        pipe.stdout.close()

        energy = float(str_en.group(0).split()[3])
        gradient_Ne = [float(el) for el in str_grad_Ne.group(0).split()[2:5]]
        gradient_X = [float(el) for el in str_grad_X.group(0).split()[2:5]]
        gradient_XG = [float(el) for el in str_grad_XG.group(0).split()[2:5]]

        # Assert energy and gradients
        self.assertAlmostEqual(energy, energy_ref, 7)
        self.assertAlmostEqual(gradient_Ne[2], gradient_Ne_reg[2], 7)
        self.assertAlmostEqual(gradient_X[2], gradient_X_reg[2], 7)
        self.assertAlmostEqual(gradient_XG[2], gradient_XG_reg[2], 7)

    def test_gaussian_charge_water_HF_GTO_mmap(self):
        self.test_gaussian_charge_water_HF_GTO(mmap=True)

def getSuite():
    suite = unittest.TestSuite()

    suite.addTest(TestEmbedding("test_coulomb_point_charge_Na"))
    suite.addTest(TestEmbedding("test_coulomb_gaussian_charge_Na"))
    suite.addTest(TestEmbedding("test_gaussian_nuclear_repulsion"))
    suite.addTest(TestEmbedding("test_gaussian_charge_water_HF_GTO"))

    return suite


if __name__ == "__main__":
    unittest.main()
