#!/usr/bin/python


import os
from sys import exit
import unittest
import re
from subprocess import PIPE,Popen,call



class TestIodineRates(unittest.TestCase):

    xmol=""
    xatom=""

    def setUp(self):
        dir=os.getcwd()
        self.xatom=os.getenv("XATOM_PATH")
        if not os.path.isfile(self.xatom):
             print('Please specify xatom path!')
             exit(-1)
        print("Using xatom=",self.xatom  )

        self.xmol=dir+"/../src/xmolecule"
        print("Using xmolecule=",self.xmol)

    def test_IodineRates(self):
        # This Test calculates all states for the oxygen atom, calculated all transition rates and runs xmolecule_req
        # to compare the final charge state population 
        def atom_conf_to_occ_array(atom_conf):        
            occ_array=[]
            for n in atom_conf:
                for i in range(1,n,2):
                    occ_array.append(2)
                if(n%2==1):
                    occ_array.append(1)
            return(occ_array)

        def occ_string_to_occ_array(occ_string):        
            occ_array=[]
            for s in occ_string:
                occ_array.append(int(s))
            return(occ_array)

        def occ_array_to_atom_array(occ_array):        
            atom_array=[]

            atom_array.append(sum(occ_array[0:1])) #1s
            atom_array.append(sum(occ_array[1:2])) #2s
            atom_array.append(sum(occ_array[2:5])) # 2p
            atom_array.append(sum(occ_array[5:6])) # 3s
            atom_array.append(sum(occ_array[6:9])) # 3p
            atom_array.append(sum(occ_array[14:15])) # 4s
            atom_array.append(sum(occ_array[9:14])) # 3d
            atom_array.append(sum(occ_array[15:18])) # 4p
            atom_array.append(sum(occ_array[23:24])) # 5s
            atom_array.append(sum(occ_array[18:23])) # 4d
            atom_array.append(sum(occ_array[24:27])) # 5p

            return(atom_array)

        print('Testing rates for Iodine for all single hole configurations, this may take a long time...')

        
        ##############################
        # set up input file
        f = open('input.in', 'w')
        f.write('system=Iodine\n')
        f.write('N_nuc= 3\n')
        f.write('I     0.0000000000        0.0000000000        0.0000000000    5s4p2d I.nao\n')
        f.write('average_occ_MO=3,4,5:7,8,9:10,11,12,13,14:16,17,18:19,20,21,22,23:25,26,27\n')
        f.write('average_rates=yes\n')
        f.close()
        

        atomholeconfs=["1s0", "2s1","2p1", "3s1", "3p1", "3d1","4s1","4p1", "4d1", "5s1", "5p1"]

        occstrings=["222222222222222222222222221", # 1s0
                    "212222222222222222222222221", # 2s1
                    "222212222222222222222222221", # 2p1
                    "222221222222222222222222221", # 3s1
                    "222222221222222222222222221", # 3p1                   
                    "222222222222212222222222221", # 3d1
                    "222222222222221222222222221", # 4s1
                    "222222222222222221222222221", # 4p1
                    "222222222222222222222212221", # 4d1
                    "222222222222222222222221221", # 5s1
                    "222222222222222222222222220"  # 5p1
                    ]
        xmol_options=["-SIni",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      ]

        #################################
        # loop over different hole configurations
        for n in range(0,len(atomholeconfs)):
            atomholeconf=atomholeconfs[n]
            occstring=occstrings[n]
            xmol_option=xmol_options[n]

            #prepare xatom input files
            print('Calculating atomic configuration '+atomholeconf)
            cmd=self.xatom+' -s I -N 50 -silent -rmax 40 -latter false  -nao -file_nao I.nao -transition transitiondipole -PE_start 5000 -PE_end 12000 -dE 10 -file_tdipole I.tdipole -transition auger -file_auger I.auger -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)


            # run xmolecule
            cmd=self.xmol+' input.in '+ xmol_option +' -pcs -fluorescence -auger -occ '+ occstring + '| tee output.tmp| grep -v ^# > process.dat'
            print('  Running',cmd )
            os.system(cmd)

            # compare with xatom process data
            os.system('rm xatom_process.dat')
            cmd=self.xatom+' -s I -N 50 -silent -rmax 40 -latter false -auger  -pcs -fluorescence -process xatom_process.dat -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)
            f1 = open('process.dat', 'r')
            f2 = open('xatom_process.dat', 'r')        
            process1=[]
            process2=[]
            while True :
                line1=f1.readline()
                if not line1: break
                tag=line1.split()[0]
                rate=float(line1.split()[3])
                # skip rate smaller than 1e-20, in xatom some transitions are excluded due to symmetry
                if rate<1e-20: continue 
                energy=int(line1.split()[4])            
                iniocc=occ_string_to_occ_array(line1.split()[6])
                finalocc=occ_string_to_occ_array(line1.split()[8])
                iniconf=occ_array_to_atom_array(iniocc)        
                finalconf=occ_array_to_atom_array(finalocc)        
                process1.append((tag,rate,energy,iniconf,finalconf))

            while True :
                line2=f2.readline()
                if not line2: break
                if len(line2.split()) < 1: break
                tag=line2.split()[0]
                if tag=="V": tag="P"
                rate=float(line2.split()[3])
                energy=int(line2.split()[4])
                iniconf=list(map(int,line2.split()[6:17]))
                finalconf=list(map(int,line2.split()[18:29]))
                iniocc=atom_conf_to_occ_array(iniconf)
                finalocc=atom_conf_to_occ_array(finalconf)
                process2.append((tag,rate,energy,iniconf,finalconf))

            f1.close
            f2.close
            Nprocess_xmolecule=len(process1)
            Nprocess_xatom=len(process2)

            self.assertEqual(Nprocess_xmolecule,Nprocess_xatom)        
            for p1 in process1:
                (tag1,rate1,energy1,iniconf1,finalconf1)=p1
                found=False
                for p2 in process2:
                    (tag2,rate2,energy2,iniconf2,finalconf2)=p2
                    if(tag1==tag2 and iniconf1==iniconf2 and finalconf1==finalconf2 ):
                        print('   relative difference in rate/cross section',abs(2.*(rate1-rate2)/(rate1+rate2)))
                        
                        self.assertAlmostEqual(rate1,rate2,places=4,msg="rate for %s %s->%s " % (tag1,str(iniconf1),str(finalconf2)))
                        self.assertEqual(energy1,energy2,msg="energy for %s %s->%s " % (tag1,str(iniconf1),str(finalconf2)))
                        found=True
                self.assertTrue(found,msg="matching process for %s %s->%s not found" % (tag1,str(iniconf1),str(finalconf2)) )

    
        #  end of loop
        ####################################





    def test_IodineRates_Spin(self):
        # This Test calculates all states for the oxygen atom, calculated all transition rates and runs xmolecule_req
        # to compare the final charge state population 
        def atom_conf_to_occ_array(atom_conf):        
            occ_array=[]
            for n in atom_conf:
                for i in range(1,n,2):
                    occ_array.append(2)
                if(n%2==1):
                    occ_array.append(1)
            return(occ_array)

        def occ_string_to_occ_array(occ_string):        
            occ_array=[]
            for s in occ_string:
                occ_array.append(int(s))
            return(occ_array)

        def occ_array_to_atom_array(occ_array):        
            atom_array=[]

            atom_array.append(sum(occ_array[0:1])) #1s
            atom_array.append(sum(occ_array[1:2])) #2s
            atom_array.append(sum(occ_array[2:5])) # 2p
            atom_array.append(sum(occ_array[5:6])) # 3s
            atom_array.append(sum(occ_array[6:9])) # 3p
            atom_array.append(sum(occ_array[14:15])) # 4s
            atom_array.append(sum(occ_array[9:14])) # 3d
            atom_array.append(sum(occ_array[15:18])) # 4p
            atom_array.append(sum(occ_array[23:24])) # 5s
            atom_array.append(sum(occ_array[18:23])) # 4d
            atom_array.append(sum(occ_array[24:27])) # 5p

            return(atom_array)

        print('Testing rates for Iodine for all single hole configurations, this may take a long time...')

        
        ##############################
        # set up input file
        f = open('input.in', 'w')
        f.write('system=Iodine\n')
        f.write('N_nuc= 3\n')
        f.write('I     0.0000000000        0.0000000000        0.0000000000    5s4p2d I.nao\n')
        f.write('average_occ_MO=3,4,5:7,8,9:10,11,12,13,14:16,17,18:19,20,21,22,23:25,26,27\n')
        f.write('all_occs=yes\n')
        f.write('process_spin=yes\n')
        f.write('N_process=100000\n')
        f.close()
        

        atomholeconfs=["1s0", "2s1","2p1", "3s1", "3p1", "3d1","4s1","4p1", "4d1", "5s1", "5p1"]

        occstrings=["222222222222222222222222221", # 1s0
                    "212222222222222222222222221", # 2s1
                    "222212222222222222222222221", # 2p1
                    "222221222222222222222222221", # 3s1
                    "222222221222222222222222221", # 3p1                   
                    "222222222222212222222222221", # 3d1
                    "222222222222221222222222221", # 4s1
                    "222222222222222221222222221", # 4p1
                    "222222222222222222222212221", # 4d1
                    "222222222222222222222221221", # 5s1
                    "222222222222222222222222220"  # 5p1
                    ]
        xmol_options=["-SIni",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      ]

        #################################
        # loop over different hole configurations
        for n in range(0,len(atomholeconfs)):
            atomholeconf=atomholeconfs[n]
            occstring=occstrings[n]
            xmol_option=xmol_options[n]

            #prepare xatom input files
            print('Calculating atomic configuration '+atomholeconf)
            cmd=self.xatom+' -s I -N 50 -silent -rmax 40 -latter false  -nao -file_nao I.nao -transition transitiondipole -PE_start 5000 -PE_end 12000 -dE 10 -file_tdipole I.tdipole -transition auger -file_auger I.auger -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)


            # run xmolecule
            cmd=self.xmol+' input.in '+ xmol_option +' -pcs -fluorescence -auger -occ '+ occstring 
            print('  Running',cmd )
            p=Popen(cmd,stdout=PIPE, shell=True)
            out=p.communicate()[0].decode('UTF8')

            m=re.search('Average Total Augerrate.*',out)
            xmolecule_Arate=float(m.group(0).split()[3])           
            m=re.search('Average Total Fluor.*',out)
            xmolecule_Frate=float(m.group(0).split()[4])           
            m=re.search('Average Total CS.*',out)
            xmolecule_Pcs=float(m.group(0).split()[3])           

            # compare with xatom process data
            cmd=self.xatom+' -s I -N 50  -rmax 40 -latter false -auger  -pcs -fluorescence -process xatom_process.dat -hole ' +atomholeconf
            print('  Running',cmd )
            p=Popen(cmd,stdout=PIPE, shell=True)
            out=p.communicate()[0].decode('UTF8')
            m=re.search('Total A rate =.*',out)
            if m==None:
                xatom_Arate=0.0
            else:
                xatom_Arate=float(m.group(0).split()[4])
            m=re.search('Total F rate =.*',out)
            if m==None:
                xatom_Frate=0.0
            else:
                xatom_Frate=float(m.group(0).split()[4])
            m=re.search('Total P cs =.*',out)
            if m==None:
                xatom_Pcs=0.0
            else:
                xatom_Pcs=float(m.group(0).split()[5])

            if(xmolecule_Arate>0.0):
                print('relative difference in avg. total A rates', abs(2*(xatom_Arate-xmolecule_Arate)/(xatom_Arate+xmolecule_Arate)))
            if(xmolecule_Frate>0.0):
                print('relative difference in avg. total F rates', abs(2*(xatom_Frate-xmolecule_Frate)/(xatom_Frate+xmolecule_Frate)))
            if(xmolecule_Pcs>0.0):
                print('relative difference in avg. total P cs', abs(2*(xatom_Pcs-xmolecule_Pcs)/(xatom_Pcs+xmolecule_Pcs)))

            self.assertAlmostEqual(xatom_Arate,xmolecule_Arate,places=4,msg="Total A rate for %s " % (occstring))
            self.assertAlmostEqual(xatom_Frate,xmolecule_Frate,places=4,msg="Total F rate for %s " % (occstring))
            self.assertAlmostEqual(xatom_Pcs,xmolecule_Pcs,places=4,msg="Total P cs for %s " % (occstring))

    
        #  end of loop
        ####################################





            
if __name__ == "__main__": 
    unittest.main()


