#!/usr/bin/python


import os
from sys import exit
import unittest
import re
from subprocess import PIPE,Popen,call



class TestOxygenRates(unittest.TestCase):

    xmol=""
    xatom=""

    def setUp(self):
        dir=os.getcwd()
        self.xatom=os.getenv("XATOM_PATH")
        if not os.path.isfile(self.xatom):
             print('Please specify xatom path!')
             exit(-1)
        print("Using xatom=",self.xatom)

        self.xmol=dir+"/../src/xmolecule"
        print("Using xmolecule=",self.xmol)


    def test_OxygenRates(self):
        # This Test calculates all states for the oxygen atom, calculated all transition rates and runs xmolecule_req
        # to compare the final charge state population 
        def atom_conf_to_occ_array(atom_conf):        
            occ_array=[]
            for n in atom_conf:
                for i in range(1,n,2):
                    occ_array.append(2)
                if(n%2==1):
                    occ_array.append(1)
            return(occ_array)

        def occ_string_to_occ_array(occ_string):        
            occ_array=[]
            for s in occ_string:
                occ_array.append(int(s))
            return(occ_array)

        def occ_array_to_atom_array(occ_array):        
            atom_array=[]

            atom_array.append(sum(occ_array[0:1])) #1s
            atom_array.append(sum(occ_array[1:2])) #2s
            atom_array.append(sum(occ_array[2:5])) # 2p

            return(atom_array)

        print('Testing rates for Oxygen atom for several holes, this may take some time...')

        
        ##############################
        # set up input file
        f = open('input.in', 'w')
        f.write('system=Oxygen\n')
        f.write('N_nuc= 1\n')
        f.write('O     0.0000000000        0.0000000000        0.0000000000    2s1p O.nao\n')
        f.write('average_occ_MO=3,4,5\n')
        f.write('average_rates=yes\n')
        f.close()
        

        atomholeconfs=["1s0","1s1","1s2","1s1_2s1","1s1_2p1","1s2_2p2","1s2_2p2","1s2_2p2"]
#        atomholeconfs=["1s0","1s1","1s2","1s1","1s1","1s2","1s2","1s2"]

        occstrings=["22220",
                    "12220",
                    "02220",
                    "11220",
                    "12210",
                    "02200",
                    "02110",
                    "02002"]


        xmol_options=["-SIni",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh"]

        #################################
        # loop over different hole configurations
        for n in range(0,len(atomholeconfs)):
            atomholeconf=atomholeconfs[n]
            occstring=occstrings[n]
            xmol_option=xmol_options[n]

            #prepare xatom input files
            print('Calculating atomic configuration '+atomholeconf)
            cmd=self.xatom+' -s O -N 50 -silent -rmax 20 -rm 1e-6 -latter false  -nao -file_nao O.nao -transition transitiondipole -PE_start 5000 -PE_end 12000 -dE 10 -file_tdipole O.tdipole -transition auger -file_auger O.auger -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)


            # run xmolecule
            cmd=self.xmol+' input.in '+ xmol_option +' -pcs -fluorescence -auger -occ '+ occstring + '| tee output.tmp| grep -v ^# |grep -v singular > process.dat'
            print('  Running',cmd )
            os.system(cmd)

            # compare with xatom process data
            os.system('rm xatom_process.dat')
            cmd=self.xatom+' -s O -N 50 -silent -rmax 20 -rm 1e-6 -latter false -auger -pcs -fluorescence -process xatom_process.dat -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)
            f1 = open('process.dat', 'r')
            f2 = open('xatom_process.dat', 'r')        
            process1=[]
            process2=[]
            while True :
                line1=f1.readline()
                if not line1: break
                tag=line1.split()[0]
                rate=float(line1.split()[3])
                # skip rate smaller than 1e-20, in xatom some transitions are excluded due to symmetry, which in 
                if rate<1e-20: continue 
                energy=int(line1.split()[4])            
                iniocc=occ_string_to_occ_array(line1.split()[6])
                finalocc=occ_string_to_occ_array(line1.split()[8])
                iniconf=occ_array_to_atom_array(iniocc)        
                finalconf=occ_array_to_atom_array(finalocc)        
                process1.append((tag,rate,energy,iniconf,finalconf))

            while True :
                line2=f2.readline()
                if not line2: break
                if len(line2.split()) < 1: break
                tag=line2.split()[0]
                if tag=="V": tag="P"
                rate=float(line2.split()[3])
                energy=int(line2.split()[4])
                iniconf=list(map(int,line2.split()[6:9]))
                finalconf=list(map(int,line2.split()[10:13]))
                iniocc=atom_conf_to_occ_array(iniconf)
                finalocc=atom_conf_to_occ_array(finalconf)
                process2.append((tag,rate,energy,iniconf,finalconf))

            f1.close
            f2.close
            Nprocess_xmolecule=len(process1)
            Nprocess_xatom=len(process2)

            self.assertEqual(Nprocess_xmolecule,Nprocess_xatom)        
            for p1 in process1:
                (tag1,rate1,energy1,iniconf1,finalconf1)=p1
                found=False
                for p2 in process2:
                    (tag2,rate2,energy2,iniconf2,finalconf2)=p2
                    if(tag1==tag2 and iniconf1==iniconf2 and finalconf1==finalconf2 ):
                        self.assertAlmostEqual(rate1,rate2,places=4,msg="Rate for %s %s->%s " % (tag1,str(iniconf1),str(finalconf2)))
                        self.assertEqual(energy1,energy2,msg="Energy for %s %s->%s " % (tag1,str(iniconf1),str(finalconf2)))
                        found=True
                self.assertTrue(found,msg="Matching Process for %s %s->%s not found" % (tag1,str(iniconf1),str(finalconf2)) )

    
        #  end of loop
        ####################################






    def test_OxygenRates_Spin(self):
        # This Test calculates all states for the oxygen atom, calculated all transition rates and runs xmolecule_req
        # to compare the final charge state population 
        def atom_conf_to_occ_array(atom_conf):        
            occ_array=[]
            for n in atom_conf:
                for i in range(1,n,2):
                    occ_array.append(2)
                if(n%2==1):
                    occ_array.append(1)
            return(occ_array)

        def occ_string_to_occ_array(occ_string):        
            occ_array=[]
            for s in occ_string:
                occ_array.append(int(s))
            return(occ_array)

        def occ_array_to_atom_array(occ_array):        
            atom_array=[]

            atom_array.append(sum(occ_array[0:1])) #1s
            atom_array.append(sum(occ_array[1:2])) #2s
            atom_array.append(sum(occ_array[2:5])) # 2p

            return(atom_array)

        print('Testing spin-resolved Lifetimes for Oxygen atom for several holes, this may take some time...')

        
        ##############################
        # set up input file
        f = open('input.in', 'w')
        f.write('system=Oxygen\n')
        f.write('N_nuc= 1\n')
        f.write('O     0.0000000000        0.0000000000        0.0000000000    2s1p O.nao\n')
        f.write('average_occ_MO=3,4,5\n')
        f.write('all_occs=yes\n')
        f.write('process_spin=yes\n')
        f.close()
        

        atomholeconfs=["1s0","1s1","1s2","1s1_2s1","1s1_2p1","1s2_2p2","1s2_2p2","1s2_2p2"]

        occstrings=["22220",
                    "12220",
                    "02220",
                    "11220",
                    "12210",
                    "02200",
                    "02110",
                    "02002"]


        xmol_options=["-SIni",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh",
                      "-MOM -RIni -RIniSh"]

        #################################
        # loop over different hole configurations
        for n in range(0,len(atomholeconfs)):
            atomholeconf=atomholeconfs[n]
            occstring=occstrings[n]
            xmol_option=xmol_options[n]

            #prepare xatom input files
            print('Calculating atomic configuration '+atomholeconf)
            cmd=self.xatom+' -s O -N 50 -silent -rmax 20 -rm 1e-6 -latter false  -nao -file_nao O.nao -transition transitiondipole -PE_start 5000 -PE_end 12000 -dE 10 -file_tdipole O.tdipole -transition auger -file_auger O.auger -hole ' +atomholeconf
            print('  Running',cmd )
            os.system(cmd)

            # run xmolecule
            cmd=self.xmol+' input.in '+ xmol_option +' -pcs -fluorescence -auger -occ '+ occstring 
            print('  Running',cmd )
            p=Popen(cmd,stdout=PIPE, shell=True)
            out=p.communicate()[0].decode('UTF8')

            m=re.search('Average Total Augerrate.*',out)
            xmolecule_Arate=float(m.group(0).split()[3])           
            m=re.search('Average Total Fluor.*',out)
            xmolecule_Frate=float(m.group(0).split()[4])           
            m=re.search('Average Total CS.*',out)
            xmolecule_Pcs=float(m.group(0).split()[3])           

            # compare with xatom process data
            cmd=self.xatom+' -s O -N 50  -rmax 20 -rm 1e-6 -latter false -auger -pcs -fluorescence -process xatom_process.dat -hole ' +atomholeconf 
            print('  Running',cmd )
            p=Popen(cmd,stdout=PIPE, shell=True)
            out=p.communicate()[0].decode('UTF8')
            m=re.search('Total A rate =.*',out)
            if m==None:
                xatom_Arate=0.0
            else:
                xatom_Arate=float(m.group(0).split()[4])
            m=re.search('Total F rate =.*',out)
            if m==None:
                xatom_Frate=0.0
            else:
                xatom_Frate=float(m.group(0).split()[4])
            m=re.search('Total P cs =.*',out)
            if m==None:
                xatom_Pcs=0.0
            else:
                xatom_Pcs=float(m.group(0).split()[5])

            if(xmolecule_Arate>0.0):
                print('relative difference in avg. total A rates', abs(2*(xatom_Arate-xmolecule_Arate)/(xatom_Arate+xmolecule_Arate)))
            if(xmolecule_Frate>0.0):
                print('relative difference in avg. total F rates', abs(2*(xatom_Frate-xmolecule_Frate)/(xatom_Frate+xmolecule_Frate)))
            if(xmolecule_Pcs>0.0):
                print('relative difference in avg. total P cs', abs(2*(xatom_Pcs-xmolecule_Pcs)/(xatom_Pcs+xmolecule_Pcs)))

            self.assertAlmostEqual(xatom_Arate,xmolecule_Arate,places=4,msg="Total A rate for %s " % (occstring))
            self.assertAlmostEqual(xatom_Frate,xmolecule_Frate,places=4,msg="Total F rate for %s " % (occstring))
            self.assertAlmostEqual(xatom_Pcs,xmolecule_Pcs,places=4,msg="Total P cs for %s " % (occstring))

    
        #  end of loop
        ####################################



            
if __name__ == "__main__": 
    unittest.main()


