#!/usr/bin/python


import os
from sys import exit,stdout
import unittest
import re
from python_xmolecule_interface import xmol
import random
import numpy as np
from math import sqrt,acos
import numdifftools as nd

class water(unittest.TestCase):

    xmol=""
    xatom=""

    def setUp(self):
        dir=os.getcwd()
        self.xatom=os.getenv("XATOM_PATH")
        if self.xatom is None:
             print('Please specify xatom path!')
             return
        if not os.path.isfile(self.xatom):
             print('Please specify xatom path!')
             exit(-1)
        print("Using xatom=" + self.xatom)


    def test_water_HF_GTO_CIS(self):
        print('Checking finite differences using HF+CIS with 6-31G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=6-31G.gbs\n')
        f.write('HF=yes\n')
        f.write('ncore=0\n')
        f.write('SCF_energy_conv=1e-12\n')
        f.write('SCF_density_conv=1e-12\n')
        f.write('eps_teint=0.0\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000        \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000        \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000        \n')
        f.close()

        XM_cis=xmol()
        XM_cis.init()
        state=6
        geom=np.array(XM_cis.get_geom())
        E0=XM_cis.CIS_energies()[state]
        grad=np.array(XM_cis.CIS_energy_gradient(state))
        d_geometry=np.array(geom)
        epsilon=1e-4
        n_grad=[]
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM_cis.set_geom(d_geometry.tolist())
                E1=XM_cis.CIS_energies()[state]
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e %e" % (i,num_grad,grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)



    def test_water_HF_GTO_FD(self):
        print(' Checking finite differences using HF with STO-3G basis ')
        
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geometry.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)
            d_geometry[i]-=epsilon




    def test_water_HFS_GTO_FD(self):
        print(' Checking finite differences using HFS with STO-3G basis ')
        os.system('rm -rf O.nao H.nao O.nao.* H.nao.*')        

        # set up input file
        # set up input file
        f = open('input.in', 'w')
        f.write('system=water\n')
        f.write('gto=STO-3G.gbs\n')
        #f.write('lmax=%d\n' % 12)
        #f.write('n=%d\n' % 100)
        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()

        # distorted from optimium geometry: 10degree bending mode
        # and 10 percent antisymmetric stretching
        d=1.86971269191
        d1=1.0*d
        d2=1.1*d
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d1*np.sin(a/2.),d1*np.cos(a/2.),0]
        geometry[2]=[ d2*np.sin(a/2.),d2*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geom.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)
            d_geometry[i]-=epsilon

        
    def test_water_HFS_NAO_FD(self):
        print(' Checking finite differences using HFS with STO-3G basis ')
        os.system('rm -rf O.nao H.nao O.nao.* H.nao.*')
        
        # set up input file
        # set up input file
        f = open('input.in', 'w')
        f.write('system=water\n')
        #f.write('lmax=%d\n' % 12)
        f.write('n=%d\n' % 50)
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('SCF_energy_conv=1e-6\n')
        f.write('SCF_density_conv=1e-6\n')

        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()

        # distorted from optimium geometry: 10degree bending mode
        # and 10 percent antisymmetric stretching
        d=1.86971269191
        d1=1.0*d
        d2=1.1*d
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d1*np.sin(a/2.),d1*np.cos(a/2.),0]
        geometry[2]=[ d2*np.sin(a/2.),d2*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-3
        d_geometry=np.zeros(9)
        d_geometry=geom.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)
            d_geometry[i]-=epsilon


    def test_water_HFS_GTO_finiteT_FD(self):
        print(' Checking finite differences using HFS with STO-3G basis and Temp=10eV')
        os.system('rm -rf O.nao H.nao O.nao.* H.nao.*')

        # set up input file
        # set up input file
        
        f = open('input.in', 'w')
        f.write('system=water\n')
        f.write('gto=STO-3G.gbs\n')
        f.write('T_eV=10\n')
        f.write('n=100\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')


        #f.write('lmax=%d\n' % 12)
        #f.write('n=%d\n' % 100)
        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()

        # distorted from optimium geometry: 10degree bending mode
        # and 10 percent antisymmetric stretching
        d=1.86971269191
        d1=1.0*d
        d2=1.1*d
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d1*np.sin(a/2.),d1*np.cos(a/2.),0]
        geometry[2]=[ d2*np.sin(a/2.),d2*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geom.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)
            d_geometry[i]-=epsilon



    def test_water_ROHF_GTO_FD(self):
        print('Checking finite differences using ROHF with 6-31G basis ')
#        os.system('cp ../O_STO-3G.nao .')
#        os.system('cp ../H_STO-3G.nao .')
        os.system('cp ../6-31G.gbs .')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=6-31G.gbs\n')
#        f.write('gto=6-311G_star_star.gbs\n')
        f.write('HF=yes\n')
        f.write('occ=22221\n')
        f.write('MOM=yes\n')
        f.write('SCF_energy_conv=1e-12\n')
        f.write('SCF_density_conv=1e-12\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000      \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000      \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000      \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+0.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0.2]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geometry.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)

    def test_water_UHF_GTO_FD(self):
        print('Checking finite differences using UHF with 6-31G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=6-31G\n')
        f.write('HF=yes\n')
        f.write('unrestricted=yes\n')
        f.write('occ=122220\n')
        f.write('guess_overlap_tol=0,0\n')
        f.write('SCF_energy_conv=1e-12\n')
        f.write('SCF_density_conv=1e-12\n')
        f.write('O       0.0000000000        0.0000000000  0.0000000000      \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000      \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000      \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+0.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0.2]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geometry.reshape(9)
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy
                d_geometry[i]-=x
                return(E1)
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)

    def test_HeH_UHF_GTO_FD(self):
        print('Checking finite differences using UHF with 6-31G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=6-31G\n')
        f.write('HF=yes\n')
        f.write('unrestricted=yes\n')
        f.write('SCF_energy_conv=1e-12\n')
        f.write('SCF_density_conv=1e-12\n')
        f.write('He      0.0000000000        0.0000000000  -1.5000000000      \n')
        f.write('H       0.0000000000        0.0000000000   1.5000000000 \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.])
        mass=np.array([1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.])

        for d in [4.0, 3.0, 2.0]:
            print(f"d={d}")
            geometry=np.zeros((2,3))
            geometry[0]=[0.,0.,-d/2.]
            geometry[1]=[0.,0.,d/2.]
            XM.set_geom(geometry.reshape(6).tolist())
            geom=np.array(XM.get_geom())
            E0=XM.energy
            grad=np.array(XM.gradient())

            epsilon=1e-4
            d_geometry=np.zeros(6)
            d_geometry=geometry.reshape(6)
            for i in range(0,len(d_geometry)):
                def f(x):
                    d_geometry[i]+=x
                    XM.set_geom(d_geometry.tolist())
                    E1=XM.energy
                    d_geometry[i]-=x
                    return(E1)
                df = nd.Derivative(f, step=epsilon)
                num_grad = df(0.)
                print ("Numerical gradient along coordinate %d : %e" % (i,num_grad))
                print ("Analytical gradient along coordinate %d : %e" % (i,grad[i]))
                print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
                self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)

    def test_water_HF_GTO_Koopmann_FD(self):
        print('Checking finite differences using HF-Koopmann with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('gs_occ=yes\n')
        f.write('occ=222121\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000      \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000      \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000      \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        E0=XM.energy
        grad=np.array(XM.gradient())

        epsilon=1e-4
        d_geometry=np.zeros(9)
        d_geometry=geometry.reshape(9)
        
        for i in range(0,len(d_geometry)):
            def f(x):
                d_geometry[i]+=x                
                XM.set_geom(d_geometry.tolist())
                E1=XM.energy            
                d_geometry[i]-=x
                return(E1)                
            df = nd.Derivative(f, step=epsilon)
            num_grad = df(0.)
            print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
            self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)

    def test_water_HF_GTO_Koopmann_hole_FD(self):
        print('Checking finite differences using HF-Koopmann hole with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('gs_occ=yes\n')
        f.write('N_core=0\n')
        f.write('occ=222210\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()

        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())

        hp=XM.hole_properties()
        for i in range(hp[0].shape[0]):
            #for p in hp:
            occ = hp[0][i]
            E0 = hp[1][i]
            g = hp[2][i]
            nac = hp[3][i]
            #occ,E0,g,nac=p
            XM.set_occ(occ.tolist())
            #print occ,E0,XM.energy
            self.assertTrue( abs(E0-XM.energy) < 1e-6)
            grad2=np.array(XM.gradient())
            grad=np.array(g)
            self.assertTrue( abs(sum((grad-grad2)**2)) < 1e-6)

            epsilon=1e-4
            d_geometry=np.zeros(9)
            d_geometry=geometry.reshape(9)
            for i in range(0,len(d_geometry)):
                def f(x):
                    d_geometry[i]+=x                
                    XM.set_geom(d_geometry.tolist())
                    E1=XM.energy            
                    d_geometry[i]-=x
                    return(E1)                
                df = nd.Derivative(f, step=epsilon)
                num_grad = df(0.)
                print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
                self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)


    def test_water_HF_GTO_Koopmann_part_FD(self):
        print('Checking finite differences using HF-Koopmann particle with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('gs_occ_part=yes\n')
        f.write('N_core=0\n')
        f.write('verbose=no\n')
        f.write('occ=2222210\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        
        f.write('O       0.0000000000        0.0000000000  0.0000000000       \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000       \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000       \n')
        f.close()
        XM=xmol()
        XM.init()

        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())

        hp=XM.part_properties()

        for i in range(hp[0].shape[0]):

            #for p in hp:
            occ = hp[0][i]
            E0 = hp[1][i]
            g = hp[2][i]
            nac = hp[3][i]
            XM.set_occ(occ.tolist())
            print(occ,E0,XM.energy)
            self.assertTrue( abs(E0-XM.energy) < 1e-6)
            grad2=np.array(XM.gradient())
            grad=np.array(g)            
            self.assertTrue( abs(sum((grad-grad2)**2)) < 1e-6)
            
            epsilon=1e-4
            d_geometry=np.zeros(9)
            d_geometry=geometry.reshape(9)
            for i in range(0,len(d_geometry)):
                def f(x):
                    d_geometry[i]+=x                
                    XM.set_geom(d_geometry.tolist())
                    E1=XM.energy            
                    d_geometry[i]-=x
                    return(E1)                
                df = nd.Derivative(f, step=epsilon)
                num_grad = df(0.)
                print ("Numerical-analytical gradient along coordinate %d : %e" % (i,num_grad-grad[i]))
                self.assertTrue( abs(num_grad-grad[i]) < 10*epsilon)

    #@unittest.skip("some problem?")
    def test_water_HF_GTO_Koopmann_hole_Efield(self):
        print('Checking calculated electric field for HF-Koopmann hole with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('gs_occ=yes\n')
        f.write('N_core=0\n')
        f.write('occ=222210\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000        \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000        \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000        \n')
        f.close()
        XM=xmol()
        print("initializing")
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])

        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        
        Nstep=100
        dphi=2*np.pi/Nstep

        l=XM.hole_efields(np.array([0.,0.,0.]).tolist())
        l=XM.hole_efields(np.array([0.,0.,0.]))
        occs, energies, efields, upots = l
        Nholes=len(energies)
        for d in [0.2,0.4,1.0]:
            I=np.zeros(Nholes)
            for phi in np.arange(0.,2*np.pi,dphi):
                l=XM.hole_efields([d*np.cos(phi),d*np.sin(phi),0.])
                occs, energies, efields, upots = l
                for j in range(0,Nholes):
                    occ,E,efield,upot=occs[j], energies[j], efields[j], upots[j]
                    I[j]+=(efield[0]*-np.sin(phi)+efield[1]*np.cos(phi))*d * dphi    
            for j in range(0,len(l)):
                print("Integrated closed loop for hole state %d at distance %f = %e" % (j,d,I[j]))
                self.assertTrue( I[j] < 1e-5)            

    @unittest.skip("takes time")
    def test_water_HF_GTO(self):
        print('Checking gradients (energy drift in MD simulation) using HF with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
        f.write('HF=yes\n')
        f.write('SCF_energy_conv=1e-10\n')
        f.write('SCF_density_conv=1e-10\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000        2s1p O_STO-3G.nao\n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000        1s   H_STO-3G.nao\n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000        1s   H_STO-3G.nao\n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad_n=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        Nsteps=100
        t=0.
        dt=0.5
        geom_n=np.zeros(9)
        pot_energy=0.
        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        energy_deviation=[]
        energy_kin=[]
        grad=np.array(XM.gradient())
        energy_start=pot_energy
        # velocity verlet code:
        for i in range (0, Nsteps):
            pot_energy=XM.energy
            kin_energy=sum(0.5*mass*vel**2)
            if(i==0):
                energy_start=pot_energy+kin_energy
            else:
                #print t,pot_energy+kin_energy-energy_start,pot_energy-energy_start,kin_energy,
                geometry=np.array(geom).reshape(3,3)
                d12=geometry[0]-geometry[1]
                d13=geometry[0]-geometry[2]
                a=acos(np.dot(d12,d13)/sqrt(np.dot(d12,d12)*np.dot(d13,d13)))
                #print sqrt(np.dot(d12,d12)),sqrt(np.dot(d13,d13)),a*180./np.pi
                stdout.flush()
                energy_deviation.append(abs(pot_energy+kin_energy-energy_start))
                energy_kin.append(abs(kin_energy))
            geom_n=geom+vel*dt + (-grad/mass*(dt**2)* 0.5)
            XM.set_geom(geom_n.tolist())
            grad_n=np.array(XM.gradient())
            vel_n= vel - 0.5*(grad_n+grad)/mass *dt
            geom=geom_n            
            vel=vel_n
            grad=grad_n
            t=t+dt
        print("# Maximum energy deviation {:f} after {:d} steps dt={:f}".format(max(energy_deviation),Nsteps,dt))
        print("# Maximum kin energy {:f}".format(max(energy_kin)))
        print("# max energy dev/ max kin energy {:f}".format(max(energy_deviation)/max(energy_kin)))
        self.assertTrue( max(energy_deviation)/max(energy_kin) < 0.05)

    #@unittest.skip("takes time")
    def test_water_HF_GTO_Koopmann(self):
        print('Checking gradients (energy drift in MD simulation) using HF with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G\n')
        f.write('HF=yes\n')
        f.write('gs_occ=yes\n')
        f.write('guess_overlap_tol=0.0,0.0\n')
        f.write('occ=22221\n')
        f.write('N=1\n')
#        f.write('verbose=yes\n')
        f.write('SCF_energy_conv=1e-8\n')
        f.write('SCF_density_conv=1e-8\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000  \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000  \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000  \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad_n=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        Nsteps=100
        t=0.
        dt=0.1
        geom_n=np.zeros(9)
        pot_energy=0.
        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        energy_deviation=[]
        energy_kin=[]
        grad=np.array(XM.gradient())
        energy_start=pot_energy
        # velocity verlet code:
        for i in range (0, Nsteps):
            pot_energy=XM.energy
            kin_energy=sum(0.5*mass*vel**2)
            if(i==0):
                energy_start=pot_energy+kin_energy
            else:
                #print t,pot_energy+kin_energy-energy_start,pot_energy-energy_start,kin_energy,
                geometry=np.array(geom).reshape(3,3)
                d12=geometry[0]-geometry[1]
                d13=geometry[0]-geometry[2]
                a=acos(np.dot(d12,d13)/sqrt(np.dot(d12,d12)*np.dot(d13,d13)))
                #print sqrt(np.dot(d12,d12)),sqrt(np.dot(d13,d13)),a*180./np.pi
                stdout.flush()
                energy_deviation.append(abs(pot_energy+kin_energy-energy_start))
                energy_kin.append(abs(kin_energy))
            geom_n=geom+vel*dt + (-grad/mass*(dt**2)* 0.5)
            XM.set_geom(geom_n.tolist())
            grad_n=np.array(XM.gradient())
            vel_n= vel - 0.5*(grad_n+grad)/mass *dt
            geom=geom_n            
            vel=vel_n
            grad=grad_n
            t=t+dt
        print("# Maximum energy deviation {:f} after {:d} steps dt={:f}".format(max(energy_deviation),Nsteps,dt))
        print("# Maximum kin energy {:f}".format(max(energy_kin)))
        print("# max energy dev/ max kin energy {:f}".format(max(energy_deviation)/max(energy_kin)))
        self.assertTrue( max(energy_deviation)/max(energy_kin) < 0.01)


    def test_water_HF_GTO_Koopmann_part(self):
        print('Checking gradients (energy drift in MD simulation) using HF with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G\n')
        f.write('HF=yes\n')
        f.write('gs_occ_part=yes\n')
        f.write('guess_overlap_tol=0.0,0.0\n')
        f.write('occ=222221\n')
        f.write('N=1\n')
#        f.write('verbose=yes\n')
        f.write('SCF_energy_conv=1e-8\n')
        f.write('SCF_density_conv=1e-8\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000  \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000  \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000  \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad_n=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        Nsteps=100
        t=0.
        dt=0.1
        geom_n=np.zeros(9)
        pot_energy=0.
        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        energy_deviation=[]
        energy_kin=[]
        grad=np.array(XM.gradient())
        energy_start=pot_energy
        # velocity verlet code:
        for i in range (0, Nsteps):
            pot_energy=XM.energy
            kin_energy=sum(0.5*mass*vel**2)
            if(i==0):
                energy_start=pot_energy+kin_energy
            else:
                #print t,pot_energy+kin_energy-energy_start,pot_energy-energy_start,kin_energy,
                geometry=np.array(geom).reshape(3,3)
                d12=geometry[0]-geometry[1]
                d13=geometry[0]-geometry[2]
                a=acos(np.dot(d12,d13)/sqrt(np.dot(d12,d12)*np.dot(d13,d13)))
                #print sqrt(np.dot(d12,d12)),sqrt(np.dot(d13,d13)),a*180./np.pi
                stdout.flush()
                energy_deviation.append(abs(pot_energy+kin_energy-energy_start))
                energy_kin.append(abs(kin_energy))
            geom_n=geom+vel*dt + (-grad/mass*(dt**2)* 0.5)
            XM.set_geom(geom_n.tolist())
            grad_n=np.array(XM.gradient())
            vel_n= vel - 0.5*(grad_n+grad)/mass *dt
            geom=geom_n            
            vel=vel_n
            grad=grad_n
            t=t+dt
        print("# Maximum energy deviation {:f} after {:d} steps dt={:f}".format(max(energy_deviation),Nsteps,dt))
        print("# Maximum kin energy {:f}".format(max(energy_kin)))
        print("# max energy dev/ max kin energy {:f}".format(max(energy_deviation)/max(energy_kin)))
        self.assertTrue( max(energy_deviation)/max(energy_kin) < 0.01)

    @unittest.skip("takes time")
    def test_water_HFS_GTO(self):
        print('Checking gradients (energy drift in MD simulation) using HFS with STO-3G basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('gto=STO-3G.gbs\n')
#        f.write('HF=yes\n')
        f.write('SCF_energy_conv=1e-8\n')
        f.write('SCF_density_conv=1e-8\n')
        f.write('lmax=12\n')
        f.write('N=50\n')
        f.write('rmax=20\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000      \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000      \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000      \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad_n=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])            
        Nsteps=100
        t=0.
        dt=0.5
        geom_n=np.zeros(9)
        pot_energy=0.
        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        energy_deviation=[]
        energy_kin=[]
        grad=np.array(XM.gradient())
        energy_start=pot_energy
        # velocity verlet code:
        for i in range (0, Nsteps):
            pot_energy=XM.energy
            kin_energy=sum(0.5*mass*vel**2)
            if(i==0):
                energy_start=pot_energy+kin_energy
            else:
                #print t,pot_energy+kin_energy-energy_start,pot_energy-energy_start,kin_energy,
                geometry=np.array(geom).reshape(3,3)
                d12=geometry[0]-geometry[1]
                d13=geometry[0]-geometry[2]
                a=acos(np.dot(d12,d13)/sqrt(np.dot(d12,d12)*np.dot(d13,d13)))
                #print sqrt(np.dot(d12,d12)),sqrt(np.dot(d13,d13)),a*180./np.pi
                stdout.flush()
                energy_deviation.append(abs(pot_energy+kin_energy-energy_start))
                energy_kin.append(abs(kin_energy))
            geom_n=geom+vel*dt + (-grad/mass*(dt**2)* 0.5)
            XM.set_geom(geom_n.tolist())
            grad_n=np.array(XM.gradient())
            vel_n= vel - 0.5*(grad_n+grad)/mass *dt
            geom=geom_n            
            vel=vel_n
            grad=grad_n
            t=t+dt
        print("# Maximum energy deviation {:f} after {:d} steps dt={:f}".format(max(energy_deviation),Nsteps,dt))
        print("# Maximum kin energy {:f}".format(max(energy_kin)))
        print("# max energy dev/ max kin energy {:f}".format(max(energy_deviation)/max(energy_kin)))
        self.assertTrue( max(energy_deviation)/max(energy_kin) < 0.05)


    @unittest.skip("takes time")
    def test_water_HFS_AO(self):
        print('Checking gradients (energy drift in MD simulation) using HFS with AO grid basis ')
        # set up input file
        f = open('input.in', 'w')
        f.write('energy_conv=1e-8\n')
        f.write('density_conv=1e-8\n')
        f.write('EPS0=0.0\n')
        f.write('EPS1=0.0\n')
        f.write('lmax=17\n')
        f.write('N=50\n')
        f.write('rmax=20\n')
        f.write('XATOM_PATH=%s\n' % self.xatom)
        f.write('O       0.0000000000        0.0000000000  0.0000000000      2s1p O.nao \n')
        f.write('H      -1.4309301866        1.1083919440  0.0000000000      1s   H.nao \n')
        f.write('H       1.4309301866        1.1083919440  0.0000000000      1s   H.nao \n')
        f.close()
        XM=xmol()
        XM.init()
        vel =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        vel_n =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        mass=np.array([16.,16.,16.,1.,1.,1.,1.,1.,1.])
        geom =np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        grad_n=np.array([0.,0.,0.,0.,0.,0.,0.,0.,0.])
        Nsteps=100
        t=0.
        dt=0.5
        geom_n=np.zeros(9)
        pot_energy=0.
        # distorted optimium geometry: 10degree bending mode
        d=1.86971269191
        a=1.74579831808+10.*np.pi/180.
        geometry=np.zeros((3,3))
        geometry[0]=[0.,0.,0.]
        geometry[1]=[-d*np.sin(a/2.),d*np.cos(a/2.),0]
        geometry[2]=[ d*np.sin(a/2.),d*np.cos(a/2.),0]
        XM.set_geom(geometry.reshape(9).tolist())
        geom=np.array(XM.get_geom())
        energy_deviation=[]
        energy_kin=[]
        grad=np.array(XM.gradient())
        energy_start=pot_energy
        # velocity verlet code:
        for i in range (0, Nsteps):
            pot_energy=XM.energy
            kin_energy=sum(0.5*mass*vel**2)
            if(i==0):
                energy_start=pot_energy+kin_energy
            else:
                #print t,pot_energy+kin_energy-energy_start,pot_energy-energy_start,kin_energy,
                geometry=np.array(geom).reshape(3,3)
                d12=geometry[0]-geometry[1]
                d13=geometry[0]-geometry[2]
                a=acos(np.dot(d12,d13)/sqrt(np.dot(d12,d12)*np.dot(d13,d13)))
                #print sqrt(np.dot(d12,d12)),sqrt(np.dot(d13,d13)),a*180./np.pi
                stdout.flush()
                energy_deviation.append(abs(pot_energy+kin_energy-energy_start))
                energy_kin.append(abs(kin_energy))
            geom_n=geom+vel*dt + (-grad/mass*(dt**2)* 0.5)
            XM.set_geom(geom_n.tolist())
            grad_n=np.array(XM.gradient())
            vel_n= vel - 0.5*(grad_n+grad)/mass *dt
            geom=geom_n            
            vel=vel_n
            grad=grad_n
            t=t+dt
        print("# Maximum energy deviation {:f} after {:d} steps dt={:f}".format(max(energy_deviation),Nsteps,dt))
        print("# Maximum kin energy {:f}".format(max(energy_kin)))
        print("# max energy dev/ max kin energy {:f}".format(max(energy_deviation)/max(energy_kin)))


        self.assertTrue( max(energy_deviation)/max(energy_kin) < 0.05)
        os.system('rm -rf water_tmp')
            
if __name__ == "__main__": 
    unittest.main()


