!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

#ifndef DONT_REORDERP
#define REORDERP
#endif

#ifdef _OPENMP
#define SETLOCK  call omp_set_lock(lock);
#define UNSETLOCK call omp_unset_lock(lock);
#else
#define SETLOCK
#define UNSETLOCK
#endif


module gaussian
  !! author: Ludger Inhester
  !! read Gaussian Basis set and wavefunction
  !! interfaces to libcint to calculate integrals
  use constants
  use atom
  use atom_in_mol
  use matrix
  use packed_index
  use stdio
#ifdef _OPENMP
  use omp_lib
#endif
  implicit none
  private
  public :: print_GTO_info, construct_Gaussian_basis_set, calculate_teints_gaussian_basis_set, calculate_Hkin_Gaussian_basis_set, &
    calculate_Vcoul_Gaussian_basis_set, calculate_Vext_Gaussian_basis_set, construct_Fock_Matrix_direct, construct_UHF_Fock_Matrix_direct, &
    add_electrostatic_interaction, calculate_1contracted_teints_direct, jk_gradient, two_electron_gradient, two_electron_gradient_ROHF, &
    overlap_gradient, h0_gradient, Rinv_gradient, one_electron_gradient, calculate_overlap_Gaussian_basis_set, calculate_dipole_Gaussian_basis_set, calculate_quadrupole_Gaussian_basis_set, &
    testLibCint, overlap_energy_gradient, overlap_energy_gradient_ROHF, overlap_energy_gradient_UHF, rinv_gradient_int


contains

  subroutine print_GTO_info(bas, atm, env, OUTPUT)
    !! prints info on Gaussian basis set
    real(long), external :: CINTgto_norm
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: atm(:, :)
    real(long), intent(in) :: env(:)
    integer, intent(in) :: OUTPUT

    integer :: sh, bf, at, l, n_prim, i
    character(len=5) :: s, str, strl
    real(long) :: alpha, coef
    integer :: n
    integer(8) :: nij, nijkl
    n = sum(2*bas(2, :) + 1)
    nij = n*(n + 1)/2
    nijkl = nij*(nij + 1)/2
    write (OUTPUT, '("# ",a5,a7,x,a5,x,a3,a15,a15)') 'bf', 'atom', 'shell', 'l', 'exponent', 'coefficient'
    bf = 1
    do sh = 1, size(bas, 2) ! loop over shells
      at = bas(1, sh) + 1
      l = bas(2, sh)    ! l: the orbital angular momentum quantum number
      n_prim = bas(3, sh) ! number of primitives
      write (s, "(i4)") at
      str = trim(ATOMIC_NAME(atm(1, at)))//trim(adjustl(s))
      select case (l)
       case (0); strl = "S"
       case (1); strl = "P"
       case (2); strl = "D"
       case (3); strl = "F"
       case (4); strl = "G"
       case (5); strl = "H"
       case (6); strl = "I"
       case default; write (strl, '("i2")') l
      end select
      do i = 1, n_prim
        alpha = env(bas(6, sh) + i)
        coef = env(bas(7, sh) + i)
        if (i .eq. 1) then
          write (OUTPUT, '("# ",i5,a7,x,i5,x,a3,es15.5,es15.5)') bf, str, sh, strl, alpha, coef/CINTgto_norm(l, alpha)
        else
          write (OUTPUT, '("# ",5x,7x,x,5x,x,3x,es15.5,es15.5)') alpha, coef/CINTgto_norm(l, alpha)
        end if
      end do
      bf = bf + 2*l + 1
    end do
    write (OUTPUT, '("# Number of shells ",i9)') size(bas, 2)
    write (OUTPUT, '("# Number of basis functions ",i9)') n
    write (OUTPUT, '("# Number of teints to store ",i20)') nijkl
  end subroutine print_GTO_info

  subroutine construct_Gaussian_basis_set(A, env, atm, bas, filename, filename2, shellInSet)
    !! constructs Gaussian basis set
    integer, external :: CINTcgto_spheric
    integer, external :: CINT1e_ovlp_sph
    real(long), external :: CINTgto_norm
    type(Atom_set), intent(in) :: A(:)
    real(long), allocatable, intent(inout) :: env(:)
    integer, allocatable, intent(inout) :: atm(:, :)
    integer, allocatable, intent(inout) :: bas(:, :)
    character(len=*), intent(in) :: filename
    character(len=*), intent(in), optional :: filename2
    integer, intent(out), optional, allocatable :: shellInSet(:)
    integer, parameter :: MAX_ENV_OFFSET = 100000 ! this parameter must be large enough to keep all geometry data + exponents + coefficients
    integer, parameter :: LMAX = 6 ! maximum angular momentum
    integer, parameter :: BASIS_DATA = 13
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer :: i, bf
    integer :: aindex
    integer :: env_offset = 1
    integer :: nshell, sh, shOld
    real(long), allocatable :: data(:, :)
    integer :: shls(2)
    integer :: l
    character(len=:), allocatable :: filenames(:)
    character(len=:), allocatable :: filenames2(:)
    real(long) :: centerOfNucCharge(3)
    integer(8) :: opt_ovlp
    ! set the 'common center' to the center of nuclear charges
    centerOfNucCharge(:) = ZERO
    do i = 1, size(A, 1)
      centerOfNucCharge(1) = centerOfNucCharge(1) + A(i)%charge*A(i)%x
      centerOfNucCharge(2) = centerOfNucCharge(2) + A(i)%charge*A(i)%y
      centerOfNucCharge(3) = centerOfNucCharge(3) + A(i)%charge*A(i)%z
    end do
    centerOfNucCharge(:) = centerOfNucCharge(:)/sum(A(:)%charge)
    if (allocated(env)) deallocate (env)
    if (allocated(atm)) deallocate (atm)
    if (allocated(bas)) deallocate (bas)
    if (index(filename, ',') .ne. 0 ) &
      filenames = parse_string_array(filename, ',')
    if (present(filename2)) then
      if (index(filename2, ',') .ne. 0 ) &
        filenames2 = parse_string_array(filename2, ',')
    end if
    allocate (env(MAX_ENV_OFFSET))
    allocate (atm(6, size(A)))
    env(:) = ZERO
    env(2:4) = centerOfNucCharge(:)
    env_offset = 21! start at index 21 (as it is done in the examples in libcint)
    ! first count the number of contracted basis functions
    sh = 0
    do aindex = 1, size(A)
      if (A(aindex)%charge .eq. 0) cycle
      ! count the number of shells
      if (index(filename, ',') .ne. 0) then
        call add_gbs(filenames(aindex), ATOMIC_NAME(A(aindex)%charge), aindex, sh)
      else
        call add_gbs(filename, ATOMIC_NAME(A(aindex)%charge), aindex, sh)
      end if
      if (present(filename2)) then
        if (index(filename2, ',') .ne. 0) then
          call add_gbs(filenames2(aindex), ATOMIC_NAME(A(aindex)%charge), aindex, sh)
        else
          call add_gbs(filename2, ATOMIC_NAME(A(aindex)%charge), aindex, sh)
        end if
      end if
    end do ! aindex loop
    nshell = sh
    allocate (bas(8, nshell))
    if (present(filename2) .and. present(shellInSet)) then
      allocate (shellInSet(nshell))
    end if
    ! reread and save the exponents and contraction coefficients
    sh = 0
    do aindex = 1, size(A)
      atm(1, aindex) = A(aindex)%charge ! nuclear charge
      atm(2, aindex) = env_offset      ! atom coordinates:
      env(env_offset + 1) = A(aindex)%x
      env(env_offset + 2) = A(aindex)%y
      env(env_offset + 3) = A(aindex)%z
      env_offset = env_offset + 3

      ! atom(3,:): nuclear model
      !          =4 means fractional charge + Gaussian
      !          =3 means fractional charge
      !          =2 means Gaussian
      !          =1 (and any other) means point_nuc
      ! atom(4,:): parameter
      !          for fractional charge: env_offset to fractional charge
      !          for Gaussian: zeta parameter (zeta=1/(2*sigma**2))
      !          else ignored
      if (A(aindex)%charge .eq. 0 .and. A(aindex)%sigma .lt. ZERO) then
        atm(3, aindex) = 3
        atm(4, aindex) = -1   ! reserved / unused
        atm(5, aindex) = env_offset
        env(env_offset + 1) = A(aindex)%fcharge
        env_offset = env_offset + 1
      else if (A(aindex)%charge .eq. 0 .and. A(aindex)%sigma .gt. ZERO) then
        atm(3, aindex) = 4
        atm(4, aindex) = env_offset
        env(env_offset + 1) = 0.5 / (A(aindex)%sigma**2)
        atm(5, aindex) = env_offset + 1
        env(env_offset + 2) = A(aindex)%fcharge
        env_offset = env_offset + 2
      else
        atm(3, aindex) = 1
        atm(4, aindex) = -1   ! ignored
        atm(5, aindex) = -1   ! reserved / unused
      end if
      atm(6, aindex) = -1   ! reserved / unused

      if (A(aindex)%charge .eq. 0) cycle
      shOld = sh
      if (index(filename, ',') .ne. 0) then
        call add_gbs(filenames(aindex), ATOMIC_NAME(A(aindex)%charge), aindex, sh, env, bas, env_offset)
      else
        call add_gbs(filename, ATOMIC_NAME(A(aindex)%charge), aindex, sh, env, bas, env_offset)
      end if
      if (present(filename2)) then
        if (present(shellInSet)) shellInSet(shOld + 1:sh) = 1
        shOld = sh
        if (index(filename2, ',') .ne. 0) then
          call add_gbs(filenames2(aindex), ATOMIC_NAME(A(aindex)%charge), aindex, sh, env, bas, env_offset)
        else
          call add_gbs(filename2, ATOMIC_NAME(A(aindex)%charge), aindex, sh, env, bas, env_offset)
        end if
        if (present(shellInSet)) shellInSet(shOld + 1:sh) = 2
      end if
    end do ! aindex loop
    call cint1e_ovlp_sph_optimizer(opt_ovlp, atm, size(atm, 2), bas, size(bas, 2), env)
    ! Normalize the basis functions:
    bf = 1
    do sh = 1, nshell
      shls(1) = sh - 1
      shls(2) = sh - 1
      allocate (data(2*bas(2, sh) + 1, 2*bas(2, sh) + 1))
      data = ZERO
      i = CINT1e_ovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt_ovlp)
#ifdef REORDERP
      ! reorder p type bf
      if (bas(2, sh) .eq. 1) then
        data = matmul(transpose(reorder_p), data)
        data = matmul(data, reorder_p)
      end if
#endif
      do i = 1, bas(3, sh)
        env(bas(7, sh) + i) = env(bas(7, sh) + i)/sqrt(data(1, 1))
      end do
      deallocate (data)
    end do
    call CINTdel_optimizer(opt_ovlp)
  contains
    subroutine add_gbs(gto_filename, atomname, atindex, shell, my_env, my_bas, my_env_offset)
      !! adds the basis functions for atom atomname in basis set definition file filename
      !! to the arrays env and bas
      !! aindex is a running index for each atom starting from 1
      !! sh is the current shell index
      !! env_offset is the current offset in env array
      !! if env, bas, and env_offset is omitted, it just counts the number of shells
      character(len=*), intent(in) :: gto_filename
      character(len=*), intent(in) :: atomname
      integer, intent(inout) :: shell
      integer, intent(in) :: atindex
      real(long), intent(inout), optional :: my_env(:)
      integer, intent(inout), optional :: my_bas(:, :)
      integer, intent(inout), optional :: my_env_offset
      integer :: ind, status
      character(len=4) :: sym
      character(len=6) :: atom_name
      integer :: n_prim
      real(long) :: norm
      character(len=100) :: line
      real(long) :: exp, coef, coef_p
      integer :: lines
      logical :: parseAtom
      parseAtom = .False.
      open (BASIS_DATA, file=trim(gto_filename), action="read", form="formatted", access="sequential")
      rewind (BASIS_DATA)
      read_loop_A: do
        read (BASIS_DATA, "(a)", iostat=status) line
        if (status .lt. 0) exit read_loop_A
        if (line(1:1)=='!') cycle read_loop_A
        if (trim(line) =='') cycle read_loop_A
        if (trim(line) .eq. "****") then
          parseAtom = .False.
          cycle read_loop_A
        end if
        if (.not. parseAtom) then
          backspace (BASIS_DATA)
          read (BASIS_DATA, "(a,i10)", iostat=status) atom_name, ind
          if (status .lt. 0) exit read_loop_A
          parseAtom = .True.
          if (atom_name .eq. atomname) then ! atom name matches
            read_L: do l = 0, LMAX
              lines = 0
              read_bf: do
                read (BASIS_DATA, "(a4,i3,f7.2)", iostat=status) sym, n_prim, norm
                if (status .lt. 0 .or. trim(sym) .eq. "****") then
                  backspace (BASIS_DATA)
                  exit read_bf
                end if
                lines = lines + 1
                if ( &
                  (trim(sym) .eq. "SP" .and. l .eq. 0) .or. &
                  (trim(sym) .eq. "SP" .and. l .eq. 1) .or. &
                  (trim(sym) .eq. "S" .and.  l .eq. 0) .or. &
                  (trim(sym) .eq. "P" .and. l .eq. 1) .or. &
                  (trim(sym) .eq. "D" .and. l .eq. 2) .or. &
                  (trim(sym) .eq. "F" .and. l .eq. 3) .or. &
                  (trim(sym) .eq. "G" .and. l .eq. 4) .or. &
                  (trim(sym) .eq. "H" .and. l .eq. 5) .or. &
                  (trim(sym) .eq. "I" .and. l .eq. 6)) then
                  shell = shell + 1
                  if (present(my_bas)) then
                    my_bas(1, shell) = atindex - 1 ! atom index                bas(1,sh)=aindex-1 ! atom index
                    my_bas(2, shell) = l   ! l=l
                    my_bas(3, shell) = n_prim ! number of primitives
                    my_bas(4, shell) = 1 ! number of contractions for these primitives
                    my_bas(5, shell) = 0 ! kappa for spinors
                    my_bas(6, shell) = my_env_offset ! offset for exponents
                    my_bas(7, shell) = my_env_offset + n_prim ! offset for contraction coefficients
                    my_bas(8, shell) = 0
                    if (my_env_offset + 3*n_prim .gt. MAX_ENV_OFFSET) then
                      stop 'increase MAX_ENV_OFFSET'
                    end if
                  end if
                  do ind = 1, n_prim
                    if (trim(sym) .eq. "SP") then
                      read (BASIS_DATA, *, iostat=status) exp, coef, coef_p
                    else
                      read (BASIS_DATA, *, iostat=status) exp, coef
                    end if
                    lines = lines + 1
                    if (present(my_env)) then
                      my_env(my_bas(6, shell) + ind) = exp
                      if (trim(sym) .eq. "SP" .and. l .eq. 1) then
                        my_env(my_bas(7, shell) + ind) = coef_p*CINTgto_norm(my_bas(2, shell), exp)
                      else
                        my_env(my_bas(7, shell) + ind) = coef*CINTgto_norm(my_bas(2, shell), exp)
                      end if
                    end if
                  end do
                  if (present(my_env_offset)) &
                    my_env_offset = my_env_offset + 3*n_prim
                else
                  ! skip over the not matching angular momentum
                  do ind = 1, n_prim
                    if (trim(sym) .eq. "SP") then
                      read (BASIS_DATA, *, iostat=status) exp, coef, coef_p
                    else
                      read (BASIS_DATA, *, iostat=status) exp, coef
                    end if
                    lines = lines + 1
                  end do
                end if
              end do read_bf
              ! go back the read lines
              do ind = 1, lines
                backspace (BASIS_DATA)
              end do! i
            end do read_L ! loop over l
          end if! atom name matched
        end if!
      end do read_loop_A
      close (BASIS_DATA)
    end subroutine add_gbs
  end subroutine construct_Gaussian_basis_set

  subroutine calculate_Hkin_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, hkin)
    !! calculates the kinetic energy matrix
    integer, external :: CINT1e_kin_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: hkin(:, :)
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    integer i, j
    real(long), allocatable :: data(:, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    hkin(:, :) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        i = CINT1e_kin_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (bas(2, sh1) .eq. 1) then
          data = matmul(transpose(reorder_p), data)
        end if
        if (bas(2, sh2) .eq. 1) then
          data = matmul(data, reorder_p)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1
            ! the libc libray uses a different phase convention:
            ! multiply with (-1)^m1 (-1)^m2
            if (mod(m1 + m2, 2) .eq. 0) then
              hkin(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = data(i, j)
              hkin(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = data(i, j)
            else
              hkin(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = -data(i, j)
              hkin(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = -data(i, j)
            end if
          end do
        end do
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
  end subroutine calculate_Hkin_Gaussian_basis_set

  subroutine calculate_Vcoul_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, pos, vcoul)
    !! calculates the external potential energy matrix
    integer, external :: cint1e_rinv_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: pos(3)
    real(long), intent(out) :: vcoul(:, :)
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    integer i, j
    real(long), allocatable :: my_env(:)
    real(long), allocatable :: data(:, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    vcoul(:, :) = ZERO
    allocate (my_env(size(env, 1)))
    my_env(:) = env(:)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        my_env(5) = pos(1) !x
        my_env(6) = pos(2) !y
        my_env(7) = pos(3) !z

        i = cint1e_rinv_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), my_env, opt)
        if (i .ne. 0) then
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data = matmul(transpose(reorder_p), data)
          end if
          if (bas(2, sh2) .eq. 1) then
            data = matmul(data, reorder_p)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              ! the libc libray uses a different phase convention:
              ! multiply with (-1)^m1 (-1)^m2
              if (mod(m1 + m2, 2) .eq. 0) then
                vcoul(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = data(i, j)
                vcoul(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = data(i, j)
              else
                vcoul(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = -data(i, j)
                vcoul(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = -data(i, j)
              end if
            end do
          end do
        end if ! i/=0
        deallocate (data)
      end do! sh2
    end do! sh1
    deallocate (my_env)
    call CINTdel_optimizer(opt)
  end subroutine calculate_Vcoul_Gaussian_basis_set

  subroutine testLibCint
    !! make some simple tests for the library
    !! it checks that the correct order of p basis functions is used
    integer, external :: CINT1e_nuc_sph
    real(long), external :: CINTgto_norm
    real(long), allocatable :: env(:)
    integer, allocatable :: atm(:, :)
    integer, allocatable :: bas(:, :)
    integer, parameter :: MAX_ENV_OFFSET = 100000 ! this parameter must be large enough to keep all geometry data + exponents + coefficients
    integer, parameter :: LMAX = 6 ! maximum angular momentum
    integer, parameter :: BASIS_DATA = 13
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    real(long), allocatable :: data(:, :)
    integer shls(2)
    integer :: env_offset
    integer :: i
    real(long), parameter :: exp = 0.1_long
    integer(8) :: opt_nuc
    allocate (env(MAX_ENV_OFFSET))
    allocate (atm(6, 2))
    allocate (bas(8, 2))
    env(:) = ZERO
    atm(:, :) = 0
    bas(:, :) = 0
    env_offset = 21! start at index 21 (as it is done in the examples in libcint)

    atm(1, 1) = 1 ! nuclear charge
    atm(2, 1) = env_offset      ! atom coordinates:
    env(env_offset + 1) = 0.
    env(env_offset + 2) = 0.
    env(env_offset + 3) = 0.
    env_offset = env_offset + 3
    atm(3, 1) = 1    ! nuclear model
    atm(4, 1) = 0 ! nuclear charge distribution parameter offset
    atm(5, 1) = -1   ! unused
    atm(6, 1) = -1   ! unused

    atm(1, 2) = 1 ! nuclear charge
    atm(2, 2) = env_offset      ! atom coordinates:
    env(env_offset + 1) = 2.
    env(env_offset + 2) = 0.
    env(env_offset + 3) = 0.
    env_offset = env_offset + 3
    atm(3, 2) = 1    ! nuclear model
    atm(4, 2) = 0   ! nuclear charge distribution parameter offset
    atm(5, 2) = -1   ! unused
    atm(6, 2) = -1   ! unused

    bas(1, 1) = 0 !atom index
    bas(2, 1) = 1 ! l=p
    bas(3, 1) = 1 ! number of primitives
    bas(4, 1) = 1 ! number of contractions for these primitives
    bas(5, 1) = 0 ! kappa for spinors
    bas(6, 1) = env_offset ! offset for exponents
    bas(7, 1) = env_offset + 1 ! offset for contraction coefficient
    bas(8, 1) = 0
    env_offset = env_offset + 3
    env(bas(6, 1) + 1) = exp ! exponent
    env(bas(7, 1) + 1) = 1.0*CINTgto_norm(bas(2, 1), dble(1.0)) ! contraction coeff

    bas(1, 2) = 0 !atom index
    bas(2, 2) = 1 ! l=p
    bas(3, 2) = 1 ! number of primitives
    bas(4, 2) = 1 ! number of contractions for these primitives
    bas(5, 2) = 0 ! kappa for spinors
    bas(6, 2) = env_offset ! offset for exponents
    bas(7, 2) = env_offset + 1 ! offset for contraction coefficient
    bas(8, 2) = 0
    env_offset = env_offset + 3
    env(bas(6, 2) + 1) = exp ! exponent
    env(bas(7, 2) + 1) = 1.0*CINTgto_norm(bas(2, 2), dble(1.0)) ! contraction coeff
    call cint1e_nuc_sph_optimizer(opt_nuc, atm, size(atm, 2), bas, size(bas, 2), env)
    shls(1) = 1
    shls(2) = 1
    allocate (data(3, 3))
    data = ZERO
    i = CINT1e_nuc_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt_nuc)
#ifdef REORDERP
    data = matmul(transpose(reorder_p), data)
    data = matmul(data, reorder_p)
#endif

    ! nuclei has been oriented along the x-axis
    ! (px | vnuc | px) /=  (py | vnuc | py) == (pz | vnuc | pz)
    ! xmolecule order is y,z,x
    if (abs(data(1, 1) - data(2, 2)) .gt. 1e-10) then
      stop "mismatch in p orbital order/libcint library:"// &
        "Either compile libcint with option "// &
        "PYPZPX OR don't use option '-DDONT_REORDERP' in xmolecule"

    end if
    deallocate (data)
    deallocate (bas)
    deallocate (atm)
    call CINTdel_optimizer(opt_nuc)
  end subroutine testLibCint

  subroutine calculate_Vext_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, vext)
    !! computes the interaction matrix elements of the electrons with nuclei potential
    integer, external :: CINT1e_nuc_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: vext(:, :)
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    integer i, j
    real(long), allocatable :: data(:, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer(8) :: opt_nuc
    call cint1e_nuc_sph_optimizer(opt_nuc, atm, size(atm, 2), bas, size(bas, 2), env)
    vext(:, :) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        i = CINT1e_nuc_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt_nuc)
        if (i .ne. 0) then
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data = matmul(transpose(reorder_p), data)
          end if
          if (bas(2, sh2) .eq. 1) then
            data = matmul(data, reorder_p)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              ! the libc libray uses a different phase convention:
              ! multiply with (-1)^m1 (-1)^m2
              if (mod(m1 + m2, 2) .eq. 0) then
                vext(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = data(i, j)
                vext(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = data(i, j)
              else
                vext(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = -data(i, j)
                vext(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = -data(i, j)
              end if
            end do
          end do
        end if! i/=0
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt_nuc)
  end subroutine calculate_Vext_Gaussian_basis_set

  subroutine calculate_overlap_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, S)
    !! computes the overlap matrix
    integer, external :: CINT1e_ovlp_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: S(:, :)
    integer :: i, j
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    integer(8) :: opt_ovlp
    real(long), allocatable :: data(:, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    call cint1e_ovlp_sph_optimizer(opt_ovlp, atm, size(atm, 2), bas, size(bas, 2), env)
    S(:, :) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n1 = 2*bas(2, sh1) + 1
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        i = CINT1e_ovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt_ovlp)
#ifdef REORDERP
        if (bas(2, sh1) .eq. 1) then
          data = matmul(transpose(reorder_p), data)
        end if
        if (bas(2, sh2) .eq. 1) then
          data = matmul(data, reorder_p)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1
            if (mod(m1 + m2, 2) .eq. 0) then
              S(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = data(i, j)
              S(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = data(i, j)
            else
              S(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1) = -data(i, j)
              S(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1) = -data(i, j)
            end if
          end do
        end do
        deallocate (data)
      end do
    end do
    call CINTdel_optimizer(opt_ovlp)
  end subroutine calculate_overlap_Gaussian_basis_set

  subroutine calculate_teints_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, N_basis, teint, accuracy)
    !! computes all two-electron integrals
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    integer, intent(in) :: N_basis
    real(long), intent(inout) :: teint(:)
    real(long), optional :: accuracy
    integer :: sh1, sh2, sh3, sh4
    integer(8) :: sh12, sh34
    integer :: bf1, bf2, bf3, bf4
    integer :: n1, n2, n3, n4
    integer shls(4)
    real(long), allocatable :: data(:, :, :, :)
    integer :: i, j, k, l, res
    integer(8) :: ij, kl, nij, nijkl, ijkl, ijmax, ijmin, klmin, klmax, imax, jmax, kmin, lmin, imin, jmin, lmax, kmax
    integer :: m1, m2, m3, m4
    integer(8) :: opt
    integer :: err
    real(long) :: epsilonI, epsilonIsq
    real(long), allocatable :: schwarz(:)
    if (present(accuracy)) then
      epsilonI = accuracy
      epsilonIsq = accuracy**2
    else
      epsilonI = ZERO
      epsilonIsq = ZERO
    end if

    allocate (schwarz(size(bas, 2)*(size(bas, 2) + 1)/2))
    nij = N_basis*(N_basis + 1)/2
    nijkl = nij*(nij + 1)/2
    teint(:) = ZERO
    schwarz(:) = ZERO
    !$omp parallel private(sh1,sh2,sh12,n1,n2,res,shls,opt,data) default(none)  shared(atm, bas, env, schwarz)
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    ! first make integral screening: calculate maximum integrals of shell pairs (sh1 sh2| sh1 sh2)
    !$omp do schedule(static,1)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        allocate (data(n1, n2, n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        shls(3) = sh1 - 1
        shls(4) = sh2 - 1
        res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (res .ne. 0) then ! if return value is zero integrals are zero by symmetry
          schwarz(sh12) = maxval(abs(data))
        else
          schwarz(sh12) = ZERO
        end if
        deallocate (data)
      end do! sh2
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(opt)
    !$omp end parallel
    ! end of schwarz screening: Now do the actual calculations:

    !$omp parallel private(sh1,sh2,sh3,sh4,sh12,sh34,n1,n2,n3,n4,imax,jmax,imin,jmin,ijmax,ijmin,kmin,lmin,kmax,lmax,klmin,klmax, res, shls, bf1,bf2,bf3,bf4, i,j,k,l, m1,m2,m3,m4, ij, kl, ijkl, opt, data) default(none)  shared(epsilonIsq, schwarz, atm, bas, env, teint, first_bf_of_shell)
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    !$omp do schedule(static,1)
    do sh1 = 1, size(bas, 2)
      n1 = 2*bas(2, sh1) + 1
      sh12 = sh1*(sh1 - 1)/2
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh12 + 1
        sh34 = 0
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          do sh4 = 1, sh3 ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            !sh34 = sh3*(sh3-1)/2+sh4
            sh34 = sh34 + 1
            if (schwarz(sh12)*schwarz(sh34) .lt. epsilonIsq) cycle
            imin = first_bf_of_shell(sh1)
            jmin = first_bf_of_shell(sh2)
            imax = imin + n1 - 1
            jmax = jmin + n2 - 1
            ijmax = jmax + imax * (imax - 1)/2
            ijmin = jmin + imin * (imin - 1)/2
            kmin = first_bf_of_shell(sh3)
            lmin = first_bf_of_shell(sh4)
            kmax = kmin + n3 -1
            lmax = lmin + n4 -1
            klmin = lmin + kmin * (kmin - 1)/2
            klmax = lmax + kmax * (kmax - 1)/2
            ! this prescreens condition ij<kl
            if (ijmax .lt. klmin) cycle
            allocate (data(n1, n2, n3, n4))
            shls(1) = sh1 - 1
            shls(2) = sh2 - 1
            shls(3) = sh3 - 1
            shls(4) = sh4 - 1
            data = ZERO
            res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
            if (res .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
              allocate (data_temp(n1, n2, n3, n4))
              ! libc orders the p functions as x,y,z
              ! we use y,z,x so we need to reorder the p basis functions
              if (bas(2, sh1) .eq. 1) then
                data_temp(1, :, :, :) = data(2, :, :, :)
                data_temp(2, :, :, :) = data(3, :, :, :)
                data_temp(3, :, :, :) = data(1, :, :, :)
                data = data_temp
              end if
              if (bas(2, sh2) .eq. 1) then
                data_temp(:, 1, :, :) = data(:, 2, :, :)
                data_temp(:, 2, :, :) = data(:, 3, :, :)
                data_temp(:, 3, :, :) = data(:, 1, :, :)
                data = data_temp
              end if
              if (bas(2, sh3) .eq. 1) then
                data_temp(:, :, 1, :) = data(:, :, 2, :)
                data_temp(:, :, 2, :) = data(:, :, 3, :)
                data_temp(:, :, 3, :) = data(:, :, 1, :)
                data = data_temp
              end if
              if (bas(2, sh4) .eq. 1) then
                data_temp(:, :, :, 1) = data(:, :, :, 2)
                data_temp(:, :, :, 2) = data(:, :, :, 3)
                data_temp(:, :, :, 3) = data(:, :, :, 1)
                data = data_temp
              end if
              deallocate (data_temp)
#endif
              do bf1 = 1, size(data, 1)
                m1 = bf1 - bas(2, sh1) - 1
                i = first_bf_of_shell(sh1) + bf1 - 1
                do bf2 = 1, size(data, 2)
                  m2 = bf2 - bas(2, sh2) - 1
                  j = first_bf_of_shell(sh2) + bf2 - 1
                  if (i .lt. j) cycle
                  ij = j + i*(i - 1)/2
                  do bf3 = 1, size(data, 3)
                    m3 = bf3 - bas(2, sh3) - 1
                    k = first_bf_of_shell(sh3) + bf3 - 1
                    do bf4 = 1, size(data, 4)
                      l = first_bf_of_shell(sh4) + bf4 - 1
                      m4 = bf4 - bas(2, sh4) - 1
                      if (k .lt. l) cycle
                      kl = l + k*(k - 1)/2
                      if (ij .lt. kl) cycle
                      ijkl = kl + ij*(ij - 1)/2
                      ! the libc libray uses a different phase convention:
                      ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                      if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                        teint(ijkl) = data(bf1, bf2, bf3, bf4)
                      else
                        teint(ijkl) = -data(bf1, bf2, bf3, bf4)
                      end if
                    end do! l
                  end do! k
                end do! j
              end do! i
            end if! res/=0
            deallocate (data)
          end do ! sh4
        end do! sh3
      end do! sh2
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(opt)
    !$omp end parallel

  end subroutine calculate_teints_Gaussian_basis_set

  subroutine calculate_1contracted_teints_direct(env, atm, bas, first_bf_of_shell, N_basis, LC, c_teint, accuracy)
    !! returns two-electron integrals, where the first index has been contracted with LCAO orbital coefficients LC
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    integer, intent(in) :: N_basis
    real(long), intent(in) :: LC(:, :)
    real(long), intent(out) :: c_teint(size(LC, 2), N_basis, N_basis*(N_basis + 1)/2)
    real(long), optional :: accuracy
    real(long) :: teint
    integer :: sh1, sh2, sh3, sh4
    integer(8) :: sh12, sh34
    integer :: bf1, bf2, bf3, bf4
    integer :: n1, n2, n3, n4
    integer shls(4)
    real(long), allocatable :: data(:, :, :, :)
    integer :: i, j, k, l, res
    integer(8) :: ij, kl, nij, nijkl, ijkl, ijmax, klmin, imax, jmax, kmin, lmin, imin, jmin, kmax, lmax
    integer :: m1, m2, m3, m4, c
    integer(8) :: opt = 0
    real(long) :: epsilonI, epsilonIsq
    real(long), allocatable :: schwarz(:)
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    real(long), allocatable :: c_teintckij(:,:,:,:), c_teintclij(:,:,:,:), c_teintcikl(:,:,:,:), c_teintcjkl(:,:,:,:)
    if (present(accuracy)) then
      epsilonI = accuracy
      epsilonIsq = accuracy**2
    else
      epsilonI = ZERO
      epsilonIsq = ZERO
    end if
    if (size(LC, 1) .ne. N_basis) stop "LC has wrong dimensions"
    allocate (schwarz(size(bas, 2)*(size(bas, 2) + 1)/2))
    nij = N_basis*(N_basis + 1)/2
    nijkl = nij*(nij + 1)/2
    c_teint(:, :, :) = ZERO
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    ! first make integral screening: calculate maximum integrals of shell pairs (sh1 sh2| sh1 sh2)
    schwarz(:) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        allocate (data(n1, n2, n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        shls(3) = sh1 - 1
        shls(4) = sh2 - 1
        res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (res .ne. 0) then ! if return value is zero integrals are zero by symmetry
          schwarz(sh12) = maxval(abs(data))
        else
          schwarz(sh12) = ZERO
        end if
        deallocate (data)
      end do! sh2
    end do! sh1
    ! end of schwarz screening: Now do the actual calculations:
    call CINTdel_optimizer(opt)
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    !!@TODO:parallelization doesn't work her with large basis sets
    !$omp parallel private(sh1,sh2,sh3,sh4,sh12,sh34,n1,n2,n3,n4,imin,imax,jmin,jmax,ijmax,kmin,kmax,lmin,lmax,klmin, res, shls, bf1,bf2,bf3,bf4, i,j,k,l, m1,m2,m3,m4, ij, kl, ijkl, opt, data,teint, c_teintckij, c_teintclij, c_teintcikl, c_teintcjkl) default(none)  shared(epsilonIsq, schwarz, atm, bas, env, first_bf_of_shell, LC, c_teint, lock)
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    !$omp do schedule(guided)
    do sh1 = 1, size(bas, 2)
      n1 = 2*bas(2, sh1) + 1
      sh12 = sh1*(sh1 - 1)/2
      imin = first_bf_of_shell(sh1)
      imax = first_bf_of_shell(sh1) + n1 - 1
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        jmin = first_bf_of_shell(sh2)
        jmax = first_bf_of_shell(sh2) + n2 - 1
        sh12 = sh12 + 1
        sh34 = 0
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          kmin = first_bf_of_shell(sh3)
          kmax = first_bf_of_shell(sh3) + n3 - 1
          do sh4 = 1, sh3 ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            lmin = first_bf_of_shell(sh4)
            lmax = first_bf_of_shell(sh4) + n4 - 1
            !sh34 = sh3*(sh3-1)/2+sh4
            sh34 = sh34 + 1
            if (schwarz(sh12)*schwarz(sh34) .lt. epsilonIsq) cycle
            ! this prescreens condition ij<kl
            ijmax = jmax + imax*(imax - 1)/2
            kmin = first_bf_of_shell(sh3)
            lmin = first_bf_of_shell(sh4)
            klmin = lmin + kmin*(kmin - 1)/2
            if (ijmax .lt. klmin) cycle
            allocate (data(n1, n2, n3, n4))
            shls(1) = sh1 - 1
            shls(2) = sh2 - 1
            shls(3) = sh3 - 1
            shls(4) = sh4 - 1
            data = ZERO
            res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
            if (res .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
              allocate (data_temp(n1, n2, n3, n4))
              ! libc orders the p functions as x,y,z
              ! we use y,z,x so we need to reorder the p basis functions
              if (bas(2, sh1) .eq. 1) then
                data_temp(1, :, :, :) = data(2, :, :, :)
                data_temp(2, :, :, :) = data(3, :, :, :)
                data_temp(3, :, :, :) = data(1, :, :, :)
                data = data_temp
              end if
              if (bas(2, sh2) .eq. 1) then
                data_temp(:, 1, :, :) = data(:, 2, :, :)
                data_temp(:, 2, :, :) = data(:, 3, :, :)
                data_temp(:, 3, :, :) = data(:, 1, :, :)
                data = data_temp
              end if
              if (bas(2, sh3) .eq. 1) then
                data_temp(:, :, 1, :) = data(:, :, 2, :)
                data_temp(:, :, 2, :) = data(:, :, 3, :)
                data_temp(:, :, 3, :) = data(:, :, 1, :)
                data = data_temp
              end if
              if (bas(2, sh4) .eq. 1) then
                data_temp(:, :, :, 1) = data(:, :, :, 2)
                data_temp(:, :, :, 2) = data(:, :, :, 3)
                data_temp(:, :, :, 3) = data(:, :, :, 1)
                data = data_temp
              end if
              deallocate (data_temp)
#endif
              allocate(c_teintckij(size(LC,2), n3, n2, n1)); c_teintckij(:,:,:,:) = ZERO
              allocate(c_teintclij(size(LC,2), n4, n2, n1)); c_teintclij(:,:,:,:) = ZERO
              allocate(c_teintcjkl(size(LC,2), n2, n4, n3)); c_teintcjkl(:,:,:,:) = ZERO
              allocate(c_teintcikl(size(LC,2), n1, n4, n3)); c_teintcikl(:,:,:,:) = ZERO
              do bf1 = 1, size(data, 1)
                m1 = bf1 - bas(2, sh1) - 1
                i = first_bf_of_shell(sh1) + bf1 - 1
                do bf2 = 1, size(data, 2)
                  m2 = bf2 - bas(2, sh2) - 1
                  j = first_bf_of_shell(sh2) + bf2 - 1
                  if (i .lt. j) cycle
                  ij = j + i*(i - 1)/2
                  do bf3 = 1, size(data, 3)
                    m3 = bf3 - bas(2, sh3) - 1
                    k = first_bf_of_shell(sh3) + bf3 - 1
                    do bf4 = 1, size(data, 4)
                      l = first_bf_of_shell(sh4) + bf4 - 1
                      m4 = bf4 - bas(2, sh4) - 1
                      if (k .lt. l) cycle
                      !kl = l + k*(k - 1)/2
                      kl = index_packed(k,l)
                      if (ij .lt. kl) cycle
                      ijkl = kl + ij*(ij - 1)/2
                      ! the libc libray uses a different phase convention:
                      ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                      if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                        teint = data(bf1, bf2, bf3, bf4)
                      else
                        teint = -data(bf1, bf2, bf3, bf4)
                      end if
                      ! klij -> clij
                      do c = 1, size(LC, 2)
                        !c_teint(c, k, ij) = c_teint(c, k, ij) + teint*LC(l, c)
                        c_teintckij(c, bf3, bf2, bf1) = c_teintckij(c, bf3, bf2, bf1) + teint * LC(l, c)
                      end do ! c
                      if (k .ne. l) then
                        ! lkij -> ckij
                        do c = 1, size(LC, 2)
                          !c_teint(c, l, ij) = c_teint(c, l, ij) + teint*LC(k, c)
                          c_teintclij(c, bf4, bf2, bf1) = c_teintclij(c, bf4, bf2, bf1) + teint * LC(k, c)
                        end do ! c
                      end if
                      if (ij .ne. kl) then
                        ! ijkl -> cjkl
                        do c = 1, size(LC, 2)
                          !c_teint(c, j, kl) = c_teint(c, j, kl) + teint*LC(i, c)
                          c_teintcjkl(c, bf2, bf4, bf3) = c_teintcjkl(c, bf2, bf4, bf3) + teint * LC(i, c)
                        end do ! c
                        if (i .ne. j) then
                          ! jikl -> cikl
                          do c = 1, size(LC, 2)
                            !c_teint(c, i, kl) = c_teint(c, i, kl) + teint*LC(j, c)
                            c_teintcikl(c, bf1, bf4, bf3) = c_teintcikl(c, bf1, bf4, bf3) + teint * LC(j, c)
                          end do ! c
                        end if
                      end if
                    end do! l
                  end do! k
                end do! j
              end do! i
              SETLOCK
              do bf1 = 1, size(data, 1)
                i = imin + bf1 - 1
                ij = i*(i - 1)/2 + jmin - 1
                forall(bf2 = 1:min(size(data, 2), imin - jmin + bf1 ))
                  c_teint(:, kmin:kmax, ij + bf2) = c_teint(:, kmin:kmax, ij + bf2) + c_teintckij(:, :, bf2, bf1)
                  c_teint(:, lmin:lmax, ij + bf2) = c_teint(:, lmin:lmax, ij + bf2) + c_teintclij(:, :, bf2, bf1)
                end forall
              end do
              do bf3 = 1, size(data, 3)
                k = first_bf_of_shell(sh3) + bf3 - 1
                kl = k*(k - 1)/2 + lmin - 1
                forall(bf4 = 1:min(size(data, 4), kmin - lmin + bf3 ))
                  c_teint(:, jmin:jmax, kl + bf4) = c_teint(:, jmin:jmax, kl + bf4) + c_teintcjkl(:, :, bf4, bf3)
                  c_teint(:, imin:imax, kl + bf4) = c_teint(:, imin:imax, kl + bf4) + c_teintcikl(:, :, bf4, bf3)
                end forall
              end do
              UNSETLOCK
              c_teintcikl = ZERO
              c_teintcjkl = ZERO
              deallocate(c_teintckij)
              deallocate(c_teintclij)
              deallocate(c_teintcjkl)
              deallocate(c_teintcikl)
            end if! res/=0
            deallocate (data)
          end do ! sh4
        end do! sh3
      end do! sh2
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(opt)
    !$omp end parallel
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif
  end subroutine calculate_1contracted_teints_direct

  subroutine calculate_dipole_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, dipole)
    !! calculates dipole matrix elements
    integer, external :: CINT1e_r_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: dipole(:, :, :)
    integer :: i, j
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    real(long), allocatable :: data(:, :, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    dipole = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n1 = 2*bas(2, sh1) + 1
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2, 3))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        i = CINT1e_r_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (bas(2, sh1) .eq. 1) then
          data(:, :, 1) = matmul(transpose(reorder_p), data(:, :, 1))
          data(:, :, 2) = matmul(transpose(reorder_p), data(:, :, 2))
          data(:, :, 3) = matmul(transpose(reorder_p), data(:, :, 3))
        end if
        if (bas(2, sh2) .eq. 1) then
          data(:, :, 1) = matmul(data(:, :, 1), reorder_p)
          data(:, :, 2) = matmul(data(:, :, 2), reorder_p)
          data(:, :, 3) = matmul(data(:, :, 3), reorder_p)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1
            if (mod(m1 + m2, 2) .eq. 0) then
              dipole(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, :) = data(i, j, :)
              dipole(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, :) = data(i, j, :)
            else
              dipole(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, :) = -data(i, j, :)
              dipole(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, :) = -data(i, j, :)
            end if
          end do
        end do
        deallocate (data)
      end do
    end do
    call CINTdel_optimizer(opt)
  end subroutine calculate_dipole_Gaussian_basis_set

  subroutine calculate_quadrupole_Gaussian_basis_set(env, atm, bas, first_bf_of_shell, quadrupole)
    !! calculates quadrupole matrix elements
    !! CINT_1e_rr_sph
    integer, external :: CINT1e_rr_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: quadrupole(:, :, :, :)
    integer :: i, j, p1, p2
    integer :: sh1, sh2
    integer :: n1, n2, m1, m2
    integer shls(2)
    real(long), allocatable :: data(:, :, :, :)
#ifdef REORDERP
    real(long), parameter :: reorder_p(3, 3) = reshape((/ZERO, ONE, ZERO, ZERO, ZERO, ONE, ONE, ZERO, ZERO/), (/3, 3/))
#endif
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    quadrupole = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n1 = 2*bas(2, sh1) + 1
        n2 = 2*bas(2, sh2) + 1
        allocate (data(n1, n2, 3, 3))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        i = CINT1e_rr_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (bas(2, sh1) .eq. 1) then
          forall(p1=1:3, p2=1:3)
            data(:, :, p1, p2) = matmul(transpose(reorder_p), data(:, :, p1, p2))
          end forall
        end if
        if (bas(2, sh2) .eq. 1) then
          forall(p1=1:3, p2=1:3)
            data(:, :, p1, p2) = matmul(data(:, :, p1, p2), reorder_p)
          end forall
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1
            if (mod(m1 + m2, 2) .eq. 0) then
              quadrupole(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, :, :) = data(i, j, :, :)
              quadrupole(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, :, :) = data(i, j, :, :)
            else
              quadrupole(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, :, :) = -data(i, j, :, :)
              quadrupole(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, :, :) = -data(i, j, :, :)
            end if
          end do
        end do
        deallocate (data)
      end do
    end do
    call CINTdel_optimizer(opt)
  end subroutine calculate_quadrupole_Gaussian_basis_set


  subroutine construct_Fock_Matrix_direct(env, atm, bas, first_bf_of_shell, Dmat, H0, Fout, accuracy)
    !! This procedure constructs the RHF Fock Matrix
    !! assumes that density matrix is symmetric
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: Dmat(:, :)
    real(long), intent(in) :: H0(:, :)
    real(long), intent(out) :: Fout(:, :)
    real(long), optional :: accuracy
    real(long), dimension(size(bas, 2)*(size(bas, 2) + 1)/2):: schwarz ! separate array for integrals (ij|ij)
    integer :: i, j, k, l, ij, kl
    integer :: ijmax, klmin, imin, jmin, imax, jmax, kmin, lmin, kmax, lmax
    integer :: bf1, bf2, bf3, bf4
    integer :: sh1, sh2, sh3, sh4
    integer(8) :: sh12, sh34
    integer :: m1, m2, m3, m4
    integer :: n1, n2, n3, n4
    integer :: shls(4)
    integer(8) :: opt = 0
    real(long), allocatable :: data(:, :, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :, :)
#endif
    real(long), allocatable :: Fjj(:), Fkk(:), Flk(:,:), Fjk(:,:), Fkj(:,:), Fjl(:,:), Flj(:,:), Fik(:,:), Fxi(:,:)
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    integer :: res
    real(long) :: teint, epsilonI, epsilonIsq
    schwarz = ZERO
    if (present(accuracy)) then
      epsilonI = accuracy
      epsilonIsq = accuracy**2
    else
      epsilonI = ZERO
      epsilonIsq = ZERO
    end if
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    ! first make integral screening: calculate maximum integrals of shell pairs (sh1 sh2| sh1 sh2)
    schwarz(:) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        allocate (data(n1, n2, n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        shls(3) = sh1 - 1
        shls(4) = sh2 - 1
        res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (res .ne. 0) then ! if return value is zero integrals are zero by symmetry
          schwarz(sh12) = maxval(abs(data))
        else
          schwarz(sh12) = ZERO
        end if
        deallocate (data)
      end do! sh2
    end do! sh1
    ! end of schwarz screening: Now do the actual calculations:
    ! in the loop we use local copies of Fock matrix blocks and add them later to Fout
    ! this why it can be more efficiently parallelized
    call CINTdel_optimizer(opt)
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    Fout = ZERO
    !$omp parallel private(i,j,k,l,ij,kl,teint,opt,n1,n2,n3,n4,sh12,sh34,bf1,bf2,bf3,bf4,imin,imax,jmin,jmax,ijmax,kmin,lmin,kmax,lmax,klmin,shls,res,m1,m2,m3,m4,data,Fjj,Fkk,Flk,Fjk,Fkj,Fjl,Flj,Fik,Fxi) default(none) shared(env,atm,Dmat,schwarz, epsilonIsq,bas,first_bf_of_shell, lock, Fout)
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    !$omp do schedule(guided)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      imin = first_bf_of_shell(sh1)
      imax = first_bf_of_shell(sh1) + n1 - 1
      allocate(Fxi(imax,n1)); Fxi(:,:) = ZERO
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        jmin = first_bf_of_shell(sh2)
        jmax = first_bf_of_shell(sh2) + n2 - 1
        ijmax = jmax + imax*(imax - 1)/2        
        allocate(Fjj(n2)); Fjj(:) = ZERO
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          kmin = first_bf_of_shell(sh3)
          kmax = first_bf_of_shell(sh3) + n3 - 1
          allocate(Fkk(n3)); Fkk(:) = ZERO
          allocate(Fjk(n2,n3)); Fjk(:,:) = ZERO
          allocate(Fkj(n3,n2)); Fkj(:,:) = ZERO
          allocate(Fik(n1,n3)); Fik(:,:) = ZERO
          do sh4 = 1, sh3 ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            sh34 = sh3*(sh3 - 1)/2 + sh4
            if (schwarz(sh12)*schwarz(sh34) .lt. epsilonIsq) cycle
            lmin = first_bf_of_shell(sh4)
            lmax = first_bf_of_shell(sh4) + n4 - 1
            klmin = lmin + kmin*(kmin - 1)/2
            ! this prescreens condition ij<kl
            if (ijmax .lt. klmin) cycle
            allocate (data(n1, n2, n3, n4))
            data = ZERO
            shls(1) = sh1 - 1
            shls(2) = sh2 - 1
            shls(3) = sh3 - 1
            shls(4) = sh4 - 1
            res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
            if (res .ne. 0) then! if return value is zero integrals are zero by symmetry
              ! libc orders the p functions as x,y,z
              ! we use y,z,x so we need to reorder the p basis functions
#ifdef REORDERP
              allocate (data_temp(n1, n2, n3, n4))
              if (bas(2, sh1) .eq. 1) then
                data_temp(1, :, :, :) = data(2, :, :, :)
                data_temp(2, :, :, :) = data(3, :, :, :)
                data_temp(3, :, :, :) = data(1, :, :, :)
                data = data_temp
              end if
              if (bas(2, sh2) .eq. 1) then
                data_temp(:, 1, :, :) = data(:, 2, :, :)
                data_temp(:, 2, :, :) = data(:, 3, :, :)
                data_temp(:, 3, :, :) = data(:, 1, :, :)
                data = data_temp
              end if
              if (bas(2, sh3) .eq. 1) then
                data_temp(:, :, 1, :) = data(:, :, 2, :)
                data_temp(:, :, 2, :) = data(:, :, 3, :)
                data_temp(:, :, 3, :) = data(:, :, 1, :)
                data = data_temp
              end if
              if (bas(2, sh4) .eq. 1) then
                data_temp(:, :, :, 1) = data(:, :, :, 2)
                data_temp(:, :, :, 2) = data(:, :, :, 3)
                data_temp(:, :, :, 3) = data(:, :, :, 1)
                data = data_temp
              end if
              deallocate (data_temp)
#endif
              allocate(Flk(n4,n3)); Flk(:,:) = ZERO
              allocate(Fjl(n2,n4)); Fjl(:,:) = ZERO
              allocate(Flj(n4,n2)); Flj(:,:) = ZERO
              do bf1 = 1, size(data, 1)
                m1 = bf1 - bas(2, sh1) - 1
                i = first_bf_of_shell(sh1) + bf1 - 1
                do bf2 = 1, size(data, 2)
                  m2 = bf2 - bas(2, sh2) - 1
                  j = first_bf_of_shell(sh2) + bf2 - 1
                  if (i .lt. j) cycle
                  ij = i*(i - 1)/2 + j
                  do bf3 = 1, size(data, 3)
                    m3 = bf3 - bas(2, sh3) - 1
                    k = first_bf_of_shell(sh3) + bf3 - 1
                    do bf4 = 1, size(data, 4)
                      l = first_bf_of_shell(sh4) + bf4 - 1
                      m4 = bf4 - bas(2, sh4) - 1
                      if (k .lt. l) cycle
                      kl = l + k*(k - 1)/2
                      if (ij .lt. kl) cycle
                      ! the libc libray uses a different phase convention:
                      ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                      if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                        teint = data(bf1, bf2, bf3, bf4)
                      else
                        teint = -data(bf1, bf2, bf3, bf4)
                      end if
                      if (abs(teint) .eq. ZERO) cycle
                      if (i .eq. j .and. k .eq. l .and. i .eq. k) then! (ii|ii)
                        !Fout(i, i) = Fout(i, i) + HALF*Dmat(i, i)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + HALF*Dmat(i, i)*teint
                      else if (i .eq. j .and. j .eq. k) then
                        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
                        !Fout(i, i) = Fout(i, i) + Dmat(l, i)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + Dmat(l, i)*teint                        
                        !Fout(l, i) = Fout(l, i) + HALF*Dmat(i, i)*teint
                        Fxi(l, bf1) = Fxi(l, bf1) + HALF*Dmat(i, i)*teint                        
                      else if (i .eq. k .and. k .eq. l) then ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
                        !Fout(i, i) = Fout(i, i) + Dmat(i, j)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + Dmat(i, j)*teint
                        !Fout(j, i) = Fout(j, i) + HALF*Dmat(i, i)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + HALF*Dmat(i, i)*teint
                      else if (i .eq. j .and. j .eq. l) then ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
                        !Fout(i, i) = Fout(i, i) + Dmat(k, i)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + Dmat(k, i)*teint
                        !Fout(k, i) = Fout(k, i) + HALF*Dmat(i, i)*teint
                        Fxi(k, bf1) = Fxi(k, bf1) + HALF*Dmat(i, i)*teint
                      else if (j .eq. k .and. k .eq. l) then ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
                        !Fout(j, j) = Fout(j, j) + Dmat(i, j)*teint
                        Fjj(bf2) = Fjj(bf2) + Dmat(i, j)*teint
                        !Fout(j, i) = Fout(j, i) + HALF*Dmat(j, j)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + HALF*Dmat(j, j)*teint
                      else if (i .eq. j .and. k .eq. l) then
                        ! (ii|kk) (kk|ii)
                        !Fout(i, i) = Fout(i, i) + Dmat(k, k)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + Dmat(k, k)*teint
                        !Fout(k, k) = Fout(k, k) + Dmat(i, i)*teint
                        Fkk(bf3) = Fkk(bf3) + Dmat(i, i)*teint
                        !Fout(k, i) = Fout(k, i) - HALF*Dmat(k, i)*teint
                        Fxi(k, bf1) = Fxi(k, bf1) - HALF*Dmat(k, i)*teint
                      else if (i .eq. k .and. j .eq. l) then
                        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
                        !Fout(i, i) = Fout(i, i) - HALF*Dmat(j, j)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) - HALF*Dmat(j, j)*teint
                        !Fout(j, j) = Fout(j, j) - HALF*Dmat(i, i)*teint
                        Fjj(bf2) = Fjj(bf2) - HALF*Dmat(i, i)*teint
                        !Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(j, i)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + THREEHALF*Dmat(j, i)*teint
                      else if (i .eq. j) then ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
                        !Fout(i, i) = Fout(i, i) + TWO*Dmat(k, l)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) + TWO*Dmat(k, l)*teint
                        !Fout(l, k) = Fout(l, k) + Dmat(i, i)*teint
                        Flk(bf4,bf3) = Flk(bf4,bf3) + Dmat(i, i)*teint
                        if (i .lt. k) then
                          !Fout(i, k) = Fout(i, k) - HALF*Dmat(i, l)*teint
                          Fik(bf1,bf3) = Fik(bf1,bf3) - HALF*Dmat(i, l)*teint
                        else
                          !Fout(k, i) = Fout(k, i) - HALF*Dmat(i, l)*teint
                          Fxi(k, bf1) = Fxi(k, bf1) - HALF*Dmat(i, l)*teint
                        end if
                        !Fout(l, i) = Fout(l, i) - HALF*Dmat(i, k)*teint
                        Fxi(l, bf1) = Fxi(l, bf1) - HALF*Dmat(i, k)*teint
                      else if (k .eq. l) then ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
                        !Fout(k, k) = Fout(k, k) + TWO*Dmat(i, j)*teint
                        Fkk(bf3) = Fkk(bf3) + TWO*Dmat(i, j)*teint
                        !Fout(j, i) = Fout(j, i) + Dmat(k, k)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + Dmat(k, k)*teint
                        !Fout(k, i) = Fout(k, i) - HALF*Dmat(j, k)*teint
                        Fxi(k, bf1) = Fxi(k, bf1) - HALF*Dmat(j, k)*teint
                        if (j .lt. k) then
                          !Fout(j, k) = Fout(j, k) - HALF*Dmat(i, k)*teint
                          Fjk(bf2, bf3) = Fjk(bf2, bf3) - HALF*Dmat(i, k)*teint
                        else
                          !Fout(k, j) = Fout(k, j) - HALF*Dmat(i, k)*teint
                          Fkj(bf3, bf2) = Fkj(bf3, bf2) - HALF*Dmat(i, k)*teint
                        end if
                      else if (i .eq. k) then ! (ij|il) (ji|il) (ij|li) (ji|li) (il|ij) (il|ji) (li|ij) (li|ji)
                        !Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(i, l)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + THREEHALF*Dmat(i, l)*teint
                        !Fout(l, i) = Fout(l, i) + THREEHALF*Dmat(i, j)*teint
                        Fxi(l, bf1) = Fxi(l, bf1) + THREEHALF*Dmat(i, j)*teint
                        !Fout(i, i) = Fout(i, i) - Dmat(j, l)*teint
                        Fxi(i, bf1) = Fxi(i, bf1) - Dmat(j, l)*teint
                        if (j .lt. l) then
                          !Fout(j, l) = Fout(j, l) - HALF*Dmat(i, i)*teint
                          Fjl(bf2, bf4) = Fjl(bf2, bf4) - HALF*Dmat(i, i)*teint
                        else
                          !Fout(l, j) = Fout(l, j) - HALF*Dmat(i, i)*teint
                          Flj(bf4, bf2) = Flj(bf4, bf2) - HALF*Dmat(i, i)*teint
                        end if
                      else if (j .eq. l) then ! (ij|kj) (ji|kj) (ij|jk) (ji|jk) (kj|ij) (kj|ji) (jk|ij) (jk|ji)
                        !Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(k, j)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + THREEHALF*Dmat(k, j)*teint
                        if (k .lt. j) then
                          !Fout(k, j) = Fout(k, j) + THREEHALF*Dmat(i, j)*teint
                          Fkj(bf3, bf2) = Fkj(bf3, bf2) + THREEHALF*Dmat(i, j)*teint
                        else
                          !Fout(j, k) = Fout(j, k) + THREEHALF*Dmat(i, j)*teint
                          Fjk(bf2, bf3) = Fjk(bf2, bf3) + THREEHALF*Dmat(i, j)*teint
                        end if
                        if (k .lt. i) then
                          !Fout(k, i) = Fout(k, i) - HALF*Dmat(j, j)*teint
                          Fxi(k, bf1) = Fxi(k, bf1) - HALF*Dmat(j, j)*teint
                        else
                          !Fout(i, k) = Fout(i, k) - HALF*Dmat(j, j)*teint
                          Fik(bf1, bf3) = Fik(bf1, bf3) - HALF*Dmat(j, j)*teint
                        end if
                        !Fout(j, j) = Fout(j, j) - Dmat(i, k)*teint
                        Fjj(bf2) = Fjj(bf2) - Dmat(i, k)*teint
                      else if (j .eq. k) then ! (ij|jl) (ji|jl) (ij|lj) (ji|lj) (jl|ij) (jl|ji) (lj|ij) (lj|ji)
                        !Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(j, l)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + THREEHALF*Dmat(j, l)*teint                        
                        if (j .lt. l) then
                          !Fout(j, l) = Fout(j, l) + THREEHALF*Dmat(i, j)*teint
                          Fjl(bf2, bf4) = Fjl(bf2, bf4) + THREEHALF*Dmat(i, j)*teint
                        else
                          !Fout(l, j) = Fout(l, j) + THREEHALF*Dmat(j, i)*teint
                          Flj(bf4, bf2) = Flj(bf4, bf2) + THREEHALF*Dmat(j, i)*teint
                        end if
                        !Fout(l, i) = Fout(l, i) - HALF*Dmat(j, j)*teint
                        Fxi(l, bf1) = Fxi(l, bf1) - HALF*Dmat(j, j)*teint
                        !Fout(j, j) = Fout(j, j) - Dmat(i, l)*teint
                        Fjj(bf2) = Fjj(bf2) - Dmat(i, l)*teint
                      else ! (ij|kl) (ji|kl) (ij|lk) (ji|lk) (kl|ij) (kl|ji) (lk|ij) (lk|ji)
                        !Fout(j, i) = Fout(j, i) + TWO*Dmat(k, l)*teint
                        Fxi(j, bf1) = Fxi(j, bf1) + TWO*Dmat(k, l)*teint
                        !Fout(l, k) = Fout(l, k) + TWO*Dmat(i, j)*teint
                        Flk(bf4, bf3) = Flk(bf4, bf3) + TWO*Dmat(i, j)*teint
                        !Fout(k, i) = Fout(k, i) - HALF*Dmat(j, l)*teint
                        Fxi(k, bf1) = Fxi(k, bf1) - HALF*Dmat(j, l)*teint
                        !Fout(l, i) = Fout(l, i) - HALF*Dmat(j, k)*teint
                        Fxi(l, bf1) = Fxi(l, bf1) - HALF*Dmat(j, k)*teint
                        if (j .lt. k) then
                          !Fout(j, k) = Fout(j, k) - HALF*Dmat(i, l)*teint
                          Fjk(bf2, bf3) = Fjk(bf2, bf3) - HALF*Dmat(i, l)*teint
                        else
                          !Fout(k, j) = Fout(k, j) - HALF*Dmat(i, l)*teint
                          Fkj(bf3, bf2) = Fkj(bf3, bf2) - HALF*Dmat(i, l)*teint
                        end if
                        if (j .lt. l) then
                          !Fout(j, l) = Fout(j, l) - HALF*Dmat(i, k)*teint
                          Fjl(bf2, bf4) = Fjl(bf2, bf4) - HALF*Dmat(i, k)*teint
                        else
                          !Fout(l, j) = Fout(l, j) - HALF*Dmat(i, k)*teint
                          Flj(bf4, bf2) = Flj(bf4, bf2) - HALF*Dmat(i, k)*teint
                        end if
                      end if! all indices different
                    end do! bf4
                  end do! bf3
                end do! bf2
              end do! bf1
              SETLOCK
              if(jmin .le. lmax)&
                   Fout(jmin:jmax, lmin:lmax) = Fout(jmin:jmax, lmin:lmax) + Fjl(:,:)
              if(lmin .le. jmax)&
                   Fout(lmin:lmax, jmin:jmax) = Fout(lmin:lmax, jmin:jmax) + Flj(:,:)
              if(lmin .le. kmax)&
                   Fout(lmin:lmax, kmin:kmax) = Fout(lmin:lmax, kmin:kmax) + Flk(:,:)
              UNSETLOCK
              deallocate (Flk)
              deallocate (Fjl)
              deallocate (Flj)
            end if ! res/=0
            deallocate (data)

          end do! sh4
          SETLOCK
          !Fout(kmin:kmax, kmin:kmax) = Fout(kmin:kmax, kmin:kmax) + Fkk(:,:)
          forall(bf3=1:n3)
            Fout(kmin + bf3 - 1, kmin + bf3 - 1) = Fout(kmin + bf3 -1, kmin + bf3 -1) + Fkk(bf3)
          end forall
          if(jmin .le. kmax)&
               Fout(jmin:jmax, kmin:kmax) = Fout(jmin:jmax, kmin:kmax) + Fjk(:,:)
          if(kmin .le. jmax)&
               Fout(kmin:kmax, jmin:jmax) = Fout(kmin:kmax, jmin:jmax) + Fkj(:,:)
          if(imin .le. kmax)&
               Fout(imin:imax, kmin:kmax) = Fout(imin:imax, kmin:kmax) + Fik(:,:)
          UNSETLOCK
          deallocate(Fkk)
          deallocate(Fjk)
          deallocate(Fkj)
          deallocate(Fik)
        end do! sh3
        SETLOCK
        !Fout(jmin:jmax, jmin:jmax) = Fout(jmin:jmax, jmin:jmax) + Fjj(:,:)
        forall(bf2=1:n2)
          Fout(jmin+bf2-1, jmin+bf2-1) = Fout(jmin+bf2-1, jmin+bf2-1) + Fjj(bf2)
        end forall
        UNSETLOCK
        deallocate(Fjj)
      end do! sh2
      SETLOCK
      Fout(1:imax, imin:imax) = &
           Fout(1:imax, imin:imax) + Fxi(1:imax,:)
      UNSETLOCK
      deallocate(Fxi)
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(opt)
    !$omp end parallel
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif    
    Fout = H0 + Fout
    do i = 1, size(Fout, 1)
      do j = 1, i
        Fout(i, j) = Fout(j, i)
      end do
    end do
  end subroutine construct_Fock_Matrix_direct

  subroutine construct_UHF_Fock_Matrix_direct(env, atm, bas, first_bf_of_shell, Dmata, Dmatb, H0, FoutA, FoutB, accuracy)
    !! construct the UHF Fock Matrix
    !! assumes that density matrix is symmetric
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: Dmata(:, :)
    real(long), intent(in) :: Dmatb(:, :)
    real(long), intent(in) :: H0(:, :)
    real(long), intent(out) :: FoutA(:, :)
    real(long), intent(out) :: FoutB(:, :)
    real(long), optional :: accuracy
    real(long), dimension(size(bas, 2)*(size(bas, 2) + 1)/2):: schwarz ! separate array for integrals (ij|ij)
    integer :: i, j, k, l, ij, kl
    integer :: ijmax, klmin, imin, jmin, imax, jmax, kmin, lmin, kmax, lmax
    integer :: bf1, bf2, bf3, bf4
    integer :: sh1, sh2, sh3, sh4
    integer(8) :: sh12, sh34
    integer :: m1, m2, m3, m4
    integer :: n1, n2, n3, n4
    integer :: shls(4)
    integer(8) :: opt
    real(long), allocatable :: data(:, :, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :, :)
#endif
    integer :: res
    real(long) :: teint, epsilonI, epsilonIsq
    real(long) :: Dmat(size(Dmata, 1), size(Dmata, 2))
    real(long), allocatable :: Fajj(:), Fakk(:), Falk(:,:), Fajk(:,:), Fakj(:,:), Fajl(:,:), Falj(:,:), Faik(:,:), Faxi(:,:)
    real(long), allocatable :: Fbjj(:), Fbkk(:), Fblk(:,:), Fbjk(:,:), Fbkj(:,:), Fbjl(:,:), Fblj(:,:), Fbik(:,:), Fbxi(:,:)
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    schwarz = ZERO
    if (present(accuracy)) then
      epsilonI = accuracy
      epsilonIsq = accuracy**2
    else
      epsilonI = ZERO
      epsilonIsq = ZERO
    end if
    Dmat = Dmata + Dmatb
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    ! first make integral screening: calculate maximum integrals of shell pairs (sh1 sh2| sh1 sh2)
    schwarz(:) = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        allocate (data(n1, n2, n1, n2))
        data = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        shls(3) = sh1 - 1
        shls(4) = sh2 - 1
        res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (res .ne. 0) then ! if return value is zero integrals are zero by symmetry
          schwarz(sh12) = maxval(abs(data))
        else
          schwarz(sh12) = ZERO
        end if
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
    ! end of schwarz screening: Now do the actual calculations:
    FoutA = ZERO
    FoutB = ZERO
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    !$omp parallel private(i,j,k,l,ij,kl,teint,opt,n1,n2,n3,n4,sh12,sh34,bf1,bf2,bf3,bf4,imin,imax,jmin,jmax,ijmax,kmin,lmin,kmax,lmax,klmin,shls,res,m1,m2,m3,m4,data,Fajj,Fakk,Falk,Fajk,Fakj,Fajl,Falj,Faik,Faxi,Fbjj,Fbkk,Fblk,Fbjk,Fbkj,Fbjl,Fblj,Fbik,Fbxi) default(none) shared(env,atm,Dmat,DmatA,DmatB,schwarz,epsilonIsq,bas,first_bf_of_shell,lock,FoutA,FoutB)
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    !$omp do schedule(guided)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      imin = first_bf_of_shell(sh1)
      imax = first_bf_of_shell(sh1) + n1 - 1
      allocate(Faxi(imax,n1)); Faxi(:,:) = ZERO
      allocate(Fbxi(imax,n1)); Fbxi(:,:) = ZERO
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        jmin = first_bf_of_shell(sh2)
        jmax = first_bf_of_shell(sh2) + n2 - 1
        ijmax = jmax + imax*(imax - 1)/2
        allocate(Fajj(n2)); Fajj(:) = ZERO
        allocate(Fbjj(n2)); Fbjj(:) = ZERO
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          kmin = first_bf_of_shell(sh3)
          kmax = first_bf_of_shell(sh3) + n3 - 1
          allocate(Fakk(n3)); Fakk(:) = ZERO
          allocate(Fajk(n2,n3)); Fajk(:,:) = ZERO
          allocate(Fakj(n3,n2)); Fakj(:,:) = ZERO
          allocate(Faik(n1,n3)); Faik(:,:) = ZERO
          allocate(Fbkk(n3)); Fbkk(:) = ZERO
          allocate(Fbjk(n2,n3)); Fbjk(:,:) = ZERO
          allocate(Fbkj(n3,n2)); Fbkj(:,:) = ZERO
          allocate(Fbik(n1,n3)); Fbik(:,:) = ZERO
          do sh4 = 1, sh3 ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            sh34 = sh3*(sh3 - 1)/2 + sh4
            if (schwarz(sh12)*schwarz(sh34) .lt. epsilonIsq) cycle
            ! this prescreens condition ij<kl
            lmin = first_bf_of_shell(sh4)
            lmax = first_bf_of_shell(sh4) + n4 - 1
            klmin = lmin + kmin*(kmin - 1)/2
            if (ijmax .lt. klmin) cycle
            allocate (data(n1, n2, n3, n4))
            data = ZERO
            shls(1) = sh1 - 1
            shls(2) = sh2 - 1
            shls(3) = sh3 - 1
            shls(4) = sh4 - 1
            res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
            if (res .ne. 0) then! if return value is zero integrals are zero by symmetry
              ! libc orders the p functions as x,y,z
              ! we use y,z,x so we need to reorder the p basis functions
#ifdef REORDERP
              allocate (data_temp(n1, n2, n3, n4))
              if (bas(2, sh1) .eq. 1) then
                data_temp(1, :, :, :) = data(2, :, :, :)
                data_temp(2, :, :, :) = data(3, :, :, :)
                data_temp(3, :, :, :) = data(1, :, :, :)
                data = data_temp
              end if
              if (bas(2, sh2) .eq. 1) then
                data_temp(:, 1, :, :) = data(:, 2, :, :)
                data_temp(:, 2, :, :) = data(:, 3, :, :)
                data_temp(:, 3, :, :) = data(:, 1, :, :)
                data = data_temp
              end if
              if (bas(2, sh3) .eq. 1) then
                data_temp(:, :, 1, :) = data(:, :, 2, :)
                data_temp(:, :, 2, :) = data(:, :, 3, :)
                data_temp(:, :, 3, :) = data(:, :, 1, :)
                data = data_temp
              end if
              if (bas(2, sh4) .eq. 1) then
                data_temp(:, :, :, 1) = data(:, :, :, 2)
                data_temp(:, :, :, 2) = data(:, :, :, 3)
                data_temp(:, :, :, 3) = data(:, :, :, 1)
                data = data_temp
              end if
              deallocate (data_temp)
#endif
              allocate(Falk(n4,n3)); Falk(:,:) = ZERO
              allocate(Fajl(n2,n4)); Fajl(:,:) = ZERO
              allocate(Falj(n4,n2)); Falj(:,:) = ZERO
              allocate(Fblk(n4,n3)); Fblk(:,:) = ZERO
              allocate(Fbjl(n2,n4)); Fbjl(:,:) = ZERO
              allocate(Fblj(n4,n2)); Fblj(:,:) = ZERO
              do bf1 = 1, size(data, 1)
                m1 = bf1 - bas(2, sh1) - 1
                i = first_bf_of_shell(sh1) + bf1 - 1
                do bf2 = 1, size(data, 2)
                  m2 = bf2 - bas(2, sh2) - 1
                  j = first_bf_of_shell(sh2) + bf2 - 1
                  if (i .lt. j) cycle
                  ij = i*(i - 1)/2 + j
                  do bf3 = 1, size(data, 3)
                    m3 = bf3 - bas(2, sh3) - 1
                    k = first_bf_of_shell(sh3) + bf3 - 1
                    do bf4 = 1, size(data, 4)
                      l = first_bf_of_shell(sh4) + bf4 - 1
                      m4 = bf4 - bas(2, sh4) - 1
                      if (k .lt. l) cycle
                      kl = l + k*(k - 1)/2
                      if (ij .lt. kl) cycle
                      ! the libc libray uses a different phase convention:
                      ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                      if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                        teint = data(bf1, bf2, bf3, bf4)
                      else
                        teint = -data(bf1, bf2, bf3, bf4)
                      end if
                      if (abs(teint) .eq. ZERO) cycle
                      if (i .eq. j .and. k .eq. l .and. i .eq. k) then! (ii|ii)
                        !FoutA(i, i) = FoutA(i, i) + DmatB(i, i)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + DmatB(i, i)*teint
                        !FoutB(i, i) = FoutB(i, i) + DmatA(i, i)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + DmatA(i, i)*teint
                      else if (i .eq. j .and. j .eq. k) then
                        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
                        !FoutA(i, i) = FoutA(i, i) + TWO*DmatB(l, i)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + TWO*DmatB(l, i)*teint
                        !FoutB(i, i) = FoutB(i, i) + TWO*DmatA(l, i)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + TWO*DmatA(l, i)*teint
                        !FoutA(l, i) = FoutA(l, i) + DmatB(i, i)*teint
                        Faxi(l, bf1) = Faxi(l, bf1) + DmatB(i, i)*teint
                        !FoutB(l, i) = FoutB(l, i) + DmatA(i, i)*teint
                        Fbxi(l, bf1) = Fbxi(l, bf1) + DmatA(i, i)*teint
                      else if (i .eq. k .and. k .eq. l) then ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
                        !FoutA(i, i) = FoutA(i, i) + TWO*DmatB(i, j)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + TWO*DmatB(i, j)*teint
                        !FoutB(i, i) = FoutB(i, i) + TWO*DmatA(i, j)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + TWO*DmatA(i, j)*teint
                        !FoutA(j, i) = FoutA(j, i) + DmatB(i, i)*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + DmatB(i, i)*teint
                        !FoutB(j, i) = FoutB(j, i) + DmatA(i, i)*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + DmatA(i, i)*teint
                      else if (i .eq. j .and. j .eq. l) then ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
                        !FoutA(i, i) = FoutA(i, i) + TWO*DmatB(k, i)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + TWO*DmatB(k, i)*teint
                        !FoutB(i, i) = FoutB(i, i) + TWO*DmatA(k, i)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + TWO*DmatA(k, i)*teint
                        !FoutA(k, i) = FoutA(k, i) + DmatB(i, i)*teint
                        Faxi(k, bf1) = Faxi(k, bf1) + DmatB(i, i)*teint
                        !FoutB(k, i) = FoutB(k, i) + DmatA(i, i)*teint
                        Fbxi(k, bf1) = Fbxi(k, bf1) + DmatA(i, i)*teint
                      else if (j .eq. k .and. k .eq. l) then ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
                        !FoutA(j, j) = FoutA(j, j) + TWO*DmatB(i, j)*teint
                        Fajj(bf2) = Fajj(bf2) + TWO*DmatB(i, j)*teint
                        !FoutB(j, j) = FoutB(j, j) + TWO*DmatA(i, j)*teint
                        Fbjj(bf2) = Fbjj(bf2) + TWO*DmatA(i, j)*teint
                        !FoutA(j, i) = FoutA(j, i) + DmatB(j, j)*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + DmatB(j, j)*teint
                        !FoutB(j, i) = FoutB(j, i) + DmatA(j, j)*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + DmatA(j, j)*teint
                      else if (i .eq. j .and. k .eq. l) then
                        ! (ii|kk) (kk|ii)
                        !FoutA(i, i) = FoutA(i, i) + Dmat(k, k)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + Dmat(k, k)*teint
                        !FoutB(i, i) = FoutB(i, i) + Dmat(k, k)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + Dmat(k, k)*teint
                        !FoutA(k, k) = FoutA(k, k) + Dmat(i, i)*teint
                        Fakk(bf3) = Fakk(bf3) + Dmat(i, i)*teint
                        !FoutB(k, k) = FoutB(k, k) + Dmat(i, i)*teint
                        Fbkk(bf3) = Fbkk(bf3) + Dmat(i, i)*teint
                        !FoutA(k, i) = FoutA(k, i) - DmatA(k, i)*teint
                        Faxi(k, bf1) = Faxi(k, bf1) - DmatA(k, i)*teint
                        !FoutB(k, i) = FoutB(k, i) - DmatB(k, i)*teint
                        Fbxi(k, bf1) = Fbxi(k, bf1) - DmatB(k, i)*teint
                      else if (i .eq. k .and. j .eq. l) then
                        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
                        !FoutA(i, i) = FoutA(i, i) - DmatA(j, j)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) - DmatA(j, j)*teint
                        !FoutB(i, i) = FoutB(i, i) - DmatB(j, j)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) - DmatB(j, j)*teint
                        !FoutA(j, j) = FoutA(j, j) - DmatA(i, i)*teint
                        Fajj(bf2) = Fajj(bf2) - DmatA(i, i)*teint
                        !FoutB(j, j) = FoutB(j, j) - DmatB(i, i)*teint
                        Fbjj(bf2) = Fbjj(bf2) - DmatB(i, i)*teint
                        !FoutA(j, i) = FoutA(j, i) + (TWO*DmatB(j, i) + DmatA(j, i))*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + (TWO*DmatB(j, i) + DmatA(j, i))*teint
                        !FoutB(j, i) = FoutB(j, i) + (TWO*DmatA(j, i) + DmatB(j, i))*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + (TWO*DmatA(j, i) + DmatB(j, i))*teint
                      else if (i .eq. j) then ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
                        !FoutA(i, i) = FoutA(i, i) + TWO*Dmat(k, l)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) + TWO*Dmat(k, l)*teint
                        !FoutB(i, i) = FoutB(i, i) + TWO*Dmat(k, l)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) + TWO*Dmat(k, l)*teint
                        !FoutA(l, k) = FoutA(l, k) + Dmat(i, i)*teint
                        Falk(bf4, bf3) = Falk(bf4, bf3) + Dmat(i, i)*teint
                        !FoutB(l, k) = FoutB(l, k) + Dmat(i, i)*teint
                        Fblk(bf4, bf3) = Fblk(bf4, bf3) + Dmat(i, i)*teint
                        if (i .lt. k) then
                          !FoutA(i, k) = FoutA(i, k) - DmatA(i, l)*teint
                          Faik(bf1, bf3) = Faik(bf1, bf3) - DmatA(i, l)*teint
                          !FoutB(i, k) = FoutB(i, k) - DmatB(i, l)*teint
                          Fbik(bf1, bf3) = Fbik(bf1, bf3) - DmatB(i, l)*teint
                        else
                          !FoutA(k, i) = FoutA(k, i) - DmatA(i, l)*teint
                          Faxi(k, bf1) = Faxi(k, bf1) - DmatA(i, l)*teint
                          !FoutB(k, i) = FoutB(k, i) - DmatB(i, l)*teint
                          Fbxi(k, bf1) = Fbxi(k, bf1) - DmatB(i, l)*teint
                        end if
                        !FoutA(l, i) = FoutA(l, i) - DmatA(i, k)*teint
                        Faxi(l, bf1) = Faxi(l, bf1) - DmatA(i, k)*teint
                        !FoutB(l, i) = FoutB(l, i) - DmatB(i, k)*teint
                        Fbxi(l, bf1) = Fbxi(l, bf1) - DmatB(i, k)*teint
                      else if (k .eq. l) then ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
                        !FoutA(k, k) = FoutA(k, k) + TWO*Dmat(i, j)*teint
                        Fakk(bf3) = Fakk(bf3) + TWO*Dmat(i, j)*teint
                        !FoutB(k, k) = FoutB(k, k) + TWO*Dmat(i, j)*teint
                        Fbkk(bf3) = Fbkk(bf3) + TWO*Dmat(i, j)*teint
                        !FoutA(j, i) = FoutA(j, i) + Dmat(k, k)*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + Dmat(k, k)*teint
                        !FoutB(j, i) = FoutB(j, i) + Dmat(k, k)*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + Dmat(k, k)*teint
                        !FoutA(k, i) = FoutA(k, i) - DmatA(j, k)*teint
                        Faxi(k, bf1) = Faxi(k, bf1) - DmatA(j, k)*teint
                        !FoutB(k, i) = FoutB(k, i) - DmatB(j, k)*teint
                        Fbxi(k, bf1) = Fbxi(k, bf1) - DmatB(j, k)*teint
                        if (j .lt. k) then
                          !FoutA(j, k) = FoutA(j, k) - DmatA(i, k)*teint
                          Fajk(bf2, bf3) = Fajk(bf2, bf3) - DmatA(i, k)*teint
                          !FoutB(j, k) = FoutB(j, k) - DmatB(i, k)*teint
                          Fbjk(bf2, bf3) = Fbjk(bf2, bf3) - DmatB(i, k)*teint
                        else
                          !FoutA(k, j) = FoutA(k, j) - DmatA(i, k)*teint
                          Fakj(bf3, bf2) = Fakj(bf3, bf2) - DmatA(i, k)*teint
                          !FoutB(k, j) = FoutB(k, j) - DmatB(i, k)*teint
                          Fbkj(bf3, bf2) = Fbkj(bf3, bf2) - DmatB(i, k)*teint
                        end if
                      else if (i .eq. k) then ! (ij|il) (ji|il) (ij|li) (ji|li) (il|ij) (il|ji) (li|ij) (li|ji)
                        !FoutA(j, i) = FoutA(j, i) + (TWO*DmatB(i, l) + DmatA(i, l))*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + (TWO*DmatB(i, l) + DmatA(i, l))*teint
                        !FoutB(j, i) = FoutB(j, i) + (TWO*DmatA(i, l) + DmatB(i, l))*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + (TWO*DmatA(i, l) + DmatB(i, l))*teint
                        !FoutA(l, i) = FoutA(l, i) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                        Faxi(l, bf1) = Faxi(l, bf1) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                        !FoutB(l, i) = FoutB(l, i) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                        Fbxi(l, bf1) = Fbxi(l, bf1) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                        !FoutA(i, i) = FoutA(i, i) - TWO*DmatA(j, l)*teint
                        Faxi(i, bf1) = Faxi(i, bf1) - TWO*DmatA(j, l)*teint
                        !FoutB(i, i) = FoutB(i, i) - TWO*DmatB(j, l)*teint
                        Fbxi(i, bf1) = Fbxi(i, bf1) - TWO*DmatB(j, l)*teint
                        if (j .lt. l) then
                          !FoutA(j, l) = FoutA(j, l) - DmatA(i, i)*teint
                          Fajl(bf2, bf4) = Fajl(bf2, bf4) - DmatA(i, i)*teint
                          !FoutB(j, l) = FoutB(j, l) - DmatB(i, i)*teint
                          Fbjl(bf2, bf4) = Fbjl(bf2, bf4) - DmatB(i, i)*teint
                        else
                          !FoutA(l, j) = FoutA(l, j) - DmatA(i, i)*teint
                          Falj(bf4, bf2) = Falj(bf4, bf2) - DmatA(i, i)*teint
                          !FoutB(l, j) = FoutB(l, j) - DmatB(i, i)*teint
                          Fblj(bf4, bf2) = Fblj(bf4, bf2) - DmatB(i, i)*teint
                        end if
                      else if (j .eq. l) then ! (ij|kj) (ji|kj) (ij|jk) (ji|jk) (kj|ij) (kj|ji) (jk|ij) (jk|ji)
                        !FoutA(j, i) = FoutA(j, i) + (TWO*DmatB(k, j) + DmatA(k, j))*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + (TWO*DmatB(k, j) + DmatA(k, j))*teint
                        !FoutB(j, i) = FoutB(j, i) + (TWO*DmatA(k, j) + DmatB(k, j))*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + (TWO*DmatA(k, j) + DmatB(k, j))*teint
                        if (k .lt. j) then
                          !FoutA(k, j) = FoutA(k, j) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          Fakj(bf3, bf2) = Fakj(bf3, bf2) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          !FoutB(k, j) = FoutB(k, j) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                          Fbkj(bf3, bf2) = Fbkj(bf3, bf2) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                        else
                          !FoutA(j, k) = FoutA(j, k) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          Fajk(bf2, bf3) = Fajk(bf2, bf3) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          !FoutB(j, k) = FoutB(j, k) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                          Fbjk(bf2, bf3) = Fbjk(bf2, bf3) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                        end if
                        if (k .lt. i) then
                          !FoutA(k, i) = FoutA(k, i) - DmatA(j, j)*teint
                          Faxi(k, bf1) = Faxi(k, bf1) - DmatA(j, j)*teint
                          !FoutB(k, i) = FoutB(k, i) - DmatB(j, j)*teint
                          Fbxi(k, bf1) = Fbxi(k, bf1) - DmatB(j, j)*teint
                        else
                          !FoutA(i, k) = FoutA(i, k) - DmatA(j, j)*teint
                          Faik(bf1, bf3) = Faik(bf1, bf3) - DmatA(j, j)*teint
                          !FoutB(i, k) = FoutB(i, k) - DmatB(j, j)*teint
                          Fbik(bf1, bf3) = Fbik(bf1, bf3) - DmatB(j, j)*teint
                        end if
                        !FoutA(j, j) = FoutA(j, j) - TWO*DmatA(i, k)*teint
                        Fajj(bf2) = Fajj(bf2) - TWO*DmatA(i, k)*teint
                        !FoutB(j, j) = FoutB(j, j) - TWO*DmatB(i, k)*teint
                        Fbjj(bf2) = Fbjj(bf2) - TWO*DmatB(i, k)*teint
                      else if (j .eq. k) then ! (ij|jl) (ji|jl) (ij|lj) (ji|lj) (jl|ij) (jl|ji) (lj|ij) (lj|ji)
                        !FoutA(j, i) = FoutA(j, i) + (TWO*DmatB(j, l) + DmatA(j, l))*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + (TWO*DmatB(j, l) + DmatA(j, l))*teint
                        !FoutB(j, i) = FoutB(j, i) + (TWO*DmatA(j, l) + DmatB(j, l))*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + (TWO*DmatA(j, l) + DmatB(j, l))*teint
                        if (j .lt. l) then
                          !FoutA(j, l) = FoutA(j, l) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          Fajl(bf3, bf4) = Fajl(bf3, bf4) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          !FoutB(j, l) = FoutB(j, l) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                          Fbjl(bf3, bf4) = Fbjl(bf3, bf4) + (TWO*DmatB(i, j) + DmatB(i, j))*teint
                        else
                          !FoutA(l, j) = FoutA(l, j) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          Falj(bf4, bf2) = Falj(bf4, bf2) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
                          !FoutB(l, j) = FoutB(l, j) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                          Fblj(bf4, bf2) = Fblj(bf4, bf2) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
                        end if
                        !FoutA(l, i) = FoutA(l, i) - DmatA(j, j)*teint
                        Faxi(l, bf1) = Faxi(l, bf1) - DmatA(j, j)*teint
                        !FoutB(l, i) = FoutB(l, i) - DmatB(j, j)*teint
                        Fbxi(l, bf1) = Fbxi(l, bf1) - DmatB(j, j)*teint
                        !FoutA(j, j) = FoutA(j, j) - TWO*DmatA(i, l)*teint
                        Fajj(bf2) = Fajj(bf2) - TWO*DmatA(i, l)*teint
                        !FoutB(j, j) = FoutB(j, j) - TWO*DmatB(i, l)*teint
                        Fbjj(bf2) = Fbjj(bf2) - TWO*DmatB(i, l)*teint
                      else ! (ij|kl) (ji|kl) (ij|lk) (ji|lk) (kl|ij) (kl|ji) (lk|ij) (lk|ji)
                        !FoutA(j, i) = FoutA(j, i) + TWO*Dmat(k, l)*teint
                        Faxi(j, bf1) = Faxi(j, bf1) + TWO*Dmat(k, l)*teint
                        !FoutB(j, i) = FoutB(j, i) + TWO*Dmat(k, l)*teint
                        Fbxi(j, bf1) = Fbxi(j, bf1) + TWO*Dmat(k, l)*teint
                        !FoutA(l, k) = FoutA(l, k) + TWO*Dmat(i, j)*teint
                        Falk(bf4, bf3) = Falk(bf4, bf3) + TWO*Dmat(i, j)*teint
                        !FoutB(l, k) = FoutB(l, k) + TWO*Dmat(i, j)*teint
                        Fblk(bf4, bf3) = Fblk(bf4, bf3) + TWO*Dmat(i, j)*teint
                        !FoutA(k, i) = FoutA(k, i) - DmatA(j, l)*teint
                        Faxi(k, bf1) = Faxi(k, bf1) - DmatA(j, l)*teint
                        !FoutB(k, i) = FoutB(k, i) - DmatB(j, l)*teint
                        Fbxi(k, bf1) = Fbxi(k, bf1) - DmatB(j, l)*teint
                        !FoutA(l, i) = FoutA(l, i) - DmatA(j, k)*teint
                        Faxi(l, bf1) = Faxi(l, bf1) - DmatA(j, k)*teint
                        !FoutB(l, i) = FoutB(l, i) - DmatB(j, k)*teint
                        Fbxi(l, bf1) = Fbxi(l, bf1) - DmatB(j, k)*teint
                        if (j .lt. k) then
                          !FoutA(j, k) = FoutA(j, k) - DmatA(i, l)*teint
                          Fajk(bf2, bf3) = Fajk(bf2, bf3) - DmatA(i, l)*teint
                          !FoutB(j, k) = FoutB(j, k) - DmatB(i, l)*teint
                          Fbjk(bf2, bf3) = Fbjk(bf2, bf3) - DmatB(i, l)*teint
                        else
                          !FoutA(k, j) = FoutA(k, j) - DmatA(i, l)*teint
                          Fakj(bf3, bf2) = Fakj(bf3, bf2) - DmatA(i, l)*teint
                          !FoutB(k, j) = FoutB(k, j) - DmatB(i, l)*teint
                          Fbkj(bf3, bf2) = Fbkj(bf3, bf2) - DmatB(i, l)*teint
                        end if
                        if (j .lt. l) then
                          !FoutA(j, l) = FoutA(j, l) - DmatA(i, k)*teint
                          Fajl(bf2, bf4) = Fajl(bf2, bf4) - DmatA(i, k)*teint
                          !FoutB(j, l) = FoutB(j, l) - DmatB(i, k)*teint
                          Fbjl(bf2, bf4) = Fbjl(bf2, bf4) - DmatB(i, k)*teint
                        else
                          !FoutA(l, j) = FoutA(l, j) - DmatA(i, k)*teint
                          Falj(bf4, bf2) = Falj(bf4, bf2) - DmatA(i, k)*teint
                          !FoutB(l, j) = FoutB(l, j) - DmatB(i, k)*teint
                          Fblj(bf4, bf2) = Fblj(bf4, bf2) - DmatB(i, k)*teint
                        end if
                      end if! all indices different
                    end do! bf4
                  end do! bf3
                end do! bf2
              end do! bf1
              SETLOCK
              if(jmin .le. lmax) then
                FoutA(jmin:jmax, lmin:lmax) = FoutA(jmin:jmax, lmin:lmax) + Fajl(:,:)
                FoutB(jmin:jmax, lmin:lmax) = FoutB(jmin:jmax, lmin:lmax) + Fbjl(:,:)
              end if
              if(lmin .le. jmax) then
                FoutA(lmin:lmax, jmin:jmax) = FoutA(lmin:lmax, jmin:jmax) + Falj(:,:)
                FoutB(lmin:lmax, jmin:jmax) = FoutB(lmin:lmax, jmin:jmax) + Fblj(:,:)
              end if
              if(lmin .le. kmax) then
                FoutA(lmin:lmax, kmin:kmax) = FoutA(lmin:lmax, kmin:kmax) + Falk(:,:)
                FoutB(lmin:lmax, kmin:kmax) = FoutB(lmin:lmax, kmin:kmax) + Fblk(:,:)
              end if
              UNSETLOCK
              deallocate (Falk)
              deallocate (Fajl)
              deallocate (Falj)
              deallocate (Fblk)
              deallocate (Fbjl)
              deallocate (Fblj)
            end if ! res/=0
            deallocate (data)
          end do! sh4
          SETLOCK
          !FoutA(kmin:kmax, kmin:kmax) = FoutA(kmin:kmax, kmin:kmax) + Fakk(:,:)
          forall(bf3=1:n3)
            FoutA(kmin + bf3 - 1, kmin + bf3 - 1) = FoutA(kmin + bf3 -1, kmin + bf3 -1) + Fakk(bf3)
          end forall
          !FoutB(kmin:kmax, kmin:kmax) = FoutB(kmin:kmax, kmin:kmax) + Fbkk(:,:)
          forall(bf3=1:n3)
            FoutB(kmin + bf3 - 1, kmin + bf3 - 1) = FoutB(kmin + bf3 -1, kmin + bf3 -1) + Fbkk(bf3)
          end forall
          if(jmin .le. kmax) then
               FoutA(jmin:jmax, kmin:kmax) = FoutA(jmin:jmax, kmin:kmax) + Fajk(:,:)
               FoutB(jmin:jmax, kmin:kmax) = FoutB(jmin:jmax, kmin:kmax) + Fbjk(:,:)
          end if
          if(kmin .le. jmax) then
               FoutA(kmin:kmax, jmin:jmax) = FoutA(kmin:kmax, jmin:jmax) + Fakj(:,:)
               FoutB(kmin:kmax, jmin:jmax) = FoutB(kmin:kmax, jmin:jmax) + Fbkj(:,:)
          end if
          if(imin .le. kmax) then
               FoutA(imin:imax, kmin:kmax) = FoutA(imin:imax, kmin:kmax) + Faik(:,:)
               FoutB(imin:imax, kmin:kmax) = FoutB(imin:imax, kmin:kmax) + Fbik(:,:)
          end if
          UNSETLOCK
          deallocate (Fakk)
          deallocate (Fbkk)
          deallocate (Fajk)
          deallocate (Fbjk)
          deallocate (Fakj)
          deallocate (Fbkj)
          deallocate (Faik)
          deallocate (Fbik)
        end do! sh3
        SETLOCK
        !FoutA(jmin:jmax, jmin:jmax) = FoutA(jmin:jmax, jmin:jmax) + Fajj(:, :)
        forall(bf2=1:n2)
          FoutA(jmin + bf2 - 1, jmin + bf2 - 1) = FoutA(jmin + bf2 - 1, jmin + bf2 - 1) + Fajj(bf2)
        end forall
        !FoutB(jmin:jmax, jmin:jmax) = FoutB(jmin:jmax, jmin:jmax) + Fbjj(:, :)
        forall(bf2=1:n2)
          FoutB(jmin + bf2 - 1, jmin + bf2 - 1) = FoutB(jmin + bf2 - 1, jmin + bf2 - 1) + Fbjj(bf2)
        end forall
        UNSETLOCK
        deallocate (Fajj)
        deallocate (Fbjj)
      end do! sh2
      SETLOCK
      FoutA(1:imax, imin:imax) = FoutA(1:imax, imin:imax) + Faxi(1:imax,:)
      FoutB(1:imax, imin:imax) = FoutB(1:imax, imin:imax) + Fbxi(1:imax,:)
      UNSETLOCK
      deallocate (Faxi)
      deallocate (Fbxi)
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(opt)
    !$omp end parallel
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif
    FoutA = H0 + FoutA
    FoutB = H0 + FoutB
    do i = 1, size(FoutA, 1)
      do j = 1, i
        FoutA(i, j) = FoutA(j, i)
        FoutB(i, j) = FoutB(j, i)
      end do
    end do
  end subroutine construct_UHF_Fock_Matrix_direct

  subroutine add_electrostatic_interaction(env, atm, bas, first_bf_of_shell, D, H)
    !! this is for the direct SCF GTO calculation
    !! adds all electrostatic interaction matrix elements to the one-electron Hamiltonian H
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: D(:, :)
    real(long), intent(inout) :: H(:, :)
    integer :: i, j, k, l, ij, kl
    integer :: ijmax, klmin, imax, jmax, kmin, lmin
    integer :: bf1, bf2, bf3, bf4
    integer :: sh1, sh2, sh3, sh4
    integer :: m1, m2, m3, m4
    integer :: n1, n2, n3, n4
    integer :: shls(4)
    integer(8) :: opt
    real(long), allocatable :: data(:, :, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :, :)
#endif
    integer :: res
    real(long) :: teint
    ! loop through the index combinations i<=j, k<=l, and ij=>kl
    ! this implies
    ! sh1<=sh2, sh3<=sh4
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = sh1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          do sh4 = sh3, size(bas, 2) ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            ! this prescreens condition ij<kl
            imax = first_bf_of_shell(sh1) + n1 - 1
            jmax = first_bf_of_shell(sh2) + n2 - 1
            ijmax = imax + jmax*(jmax - 1)/2
            kmin = first_bf_of_shell(sh3)
            lmin = first_bf_of_shell(sh4)
            klmin = kmin + lmin*(lmin - 1)/2
            if (ijmax .lt. klmin) cycle
            allocate (data(n1, n2, n3, n4))
            data = ZERO
            shls(1) = sh1 - 1
            shls(2) = sh2 - 1
            shls(3) = sh3 - 1
            shls(4) = sh4 - 1
            res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
            if (res .ne. 0) then! if return value is zero integrals are zero by symmetry
              ! libc orders the p functions as x,y,z
              ! we use y,z,x so we need to reorder the p basis functions
#ifdef REORDERP
            allocate (data_temp(n1, n2, n3, n4))
            if (bas(2, sh1) .eq. 1) then
                data_temp(1, :, :, :) = data(2, :, :, :)
                data_temp(2, :, :, :) = data(3, :, :, :)
                data_temp(3, :, :, :) = data(1, :, :, :)
                data = data_temp
              end if
              if (bas(2, sh2) .eq. 1) then
                data_temp(:, 1, :, :) = data(:, 2, :, :)
                data_temp(:, 2, :, :) = data(:, 3, :, :)
                data_temp(:, 3, :, :) = data(:, 1, :, :)
                data = data_temp
              end if
              if (bas(2, sh3) .eq. 1) then
                data_temp(:, :, 1, :) = data(:, :, 2, :)
                data_temp(:, :, 2, :) = data(:, :, 3, :)
                data_temp(:, :, 3, :) = data(:, :, 1, :)
                data = data_temp
              end if
              if (bas(2, sh4) .eq. 1) then
                data_temp(:, :, :, 1) = data(:, :, :, 2)
                data_temp(:, :, :, 2) = data(:, :, :, 3)
                data_temp(:, :, :, 3) = data(:, :, :, 1)
                data = data_temp
              end if
              deallocate (data_temp)
#endif
              do bf1 = 1, size(data, 1)
                m1 = bf1 - bas(2, sh1) - 1
                i = first_bf_of_shell(sh1) + bf1 - 1
                do bf2 = 1, size(data, 2)
                  m2 = bf2 - bas(2, sh2) - 1
                  j = first_bf_of_shell(sh2) + bf2 - 1
                  if (i .gt. j) cycle
                  ij = i + j*(j - 1)/2
                  do bf3 = 1, size(data, 3)
                    m3 = bf3 - bas(2, sh3) - 1
                    k = first_bf_of_shell(sh3) + bf3 - 1
                    do bf4 = 1, size(data, 4)
                      l = first_bf_of_shell(sh4) + bf4 - 1
                      m4 = bf4 - bas(2, sh4) - 1
                      if (k .gt. l) cycle
                      kl = k + l*(l - 1)/2
                      if (ij .lt. kl) cycle
                      ! the libc libray uses a different phase convention:
                      ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                      if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                        teint = data(bf1, bf2, bf3, bf4)
                      else
                        teint = -data(bf1, bf2, bf3, bf4)
                      end if
                      ! 4 indices equal
                      if (i .eq. j .and. k .eq. l .and. i .eq. k) then! (ii|ii)
                        H(i, i) = H(i, i) + D(i, i)*teint
                        ! 3 indices equal
                      else if (i .eq. j .and. j .eq. k) then ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
                        H(i, i) = H(i, i) + TWO*D(i, l)*teint
                        if (i .lt. l) then
                          H(i, l) = H(i, l) + D(i, i)*teint
                        else
                          H(l, i) = H(l, i) + D(i, i)*teint
                        end if
                      else if (i .eq. j .and. j .eq. l) then ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
                        H(i, i) = H(i, i) + TWO*D(k, i)*teint
                        if (i .lt. k) then
                          H(i, k) = H(i, k) + D(i, i)*teint
                        else
                          H(k, i) = H(k, i) + D(i, i)*teint
                        end if
                      else if (i .eq. k .and. k .eq. l) then ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
                        H(i, j) = H(i, j) + D(i, i)*teint
                        H(i, i) = H(i, i) + TWO*D(i, j)*teint
                      else if (j .eq. k .and. k .eq. l) then ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
                        H(i, j) = H(i, j) + D(j, j)*teint
                        H(j, j) = H(j, j) + TWO*D(i, j)*teint
                        ! two indices pairwise equal
                      else if (i .eq. j .and. k .eq. l) then ! (ii|kk) (kk|ii)
                        H(i, i) = H(i, i) + D(k, k)*teint
                        H(k, k) = H(k, k) + D(i, i)*teint
                      else if (i .eq. k .and. j .eq. l) then ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
                        H(i, j) = H(i, j) + TWO*D(i, j)*teint
                        ! two indices equal
                      else if (i .eq. j) then ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
                        H(i, i) = H(i, i) + TWO*D(k, l)*teint
                        H(k, l) = H(k, l) + D(i, i)*teint
                      else if (k .eq. l) then ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
                        H(i, j) = H(i, j) + D(k, k)*teint
                        H(k, k) = H(k, k) + TWO*D(i, j)*teint
                      else if (i .eq. k .and. j .ne. l) then ! (ij|il) (ji|il) (ij|li) (ji|li) (il|ij) (il|ji) (li|ij) (li|ji)
                        H(i, j) = H(i, j) + TWO*D(i, l)*teint
                        if (i .lt. l) then ! i is always smaller than l
                          H(i, l) = H(i, l) + TWO*D(i, j)*teint
                        else
                          H(l, i) = H(l, i) + TWO*D(i, j)*teint
                        end if
                      else if (i .ne. k .and. j .eq. l) then ! (ij|kj) (ji|kj) (ij|jk) (ji|jk) (kj|ij) (kj|ji) (jk|ij) (jk|ji)
                        H(i, j) = H(i, j) + TWO*D(k, j)*teint
                        if (k .lt. j) then
                          H(k, j) = H(k, j) + TWO*D(i, j)*teint
                        else
                          H(j, k) = H(j, k) + TWO*D(i, j)*teint
                        end if
                        ! no indices equal
                      else ! (ij|kl) (ji|kl) (ij|lk) (ji|lk) (kl|ij) (kl|ji) (lk|ij) (lk|ji)
                        H(i, j) = H(i, j) + TWO*D(k, l)*teint
                        H(k, l) = H(k, l) + TWO*D(i, j)*teint
                      end if
                    end do! l
                  end do! k
                end do! j
              end do! i
            end if! res/=0
            deallocate (data)
          end do ! sh4
        end do! sh3
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
    do i = 1, size(H, 1)
      do j = i + 1, size(H, 1)
        H(j, i) = H(i, j)
      end do
    end do
  end subroutine add_electrostatic_interaction

  subroutine overlap_gradient(env, atm, bas, first_bf_of_shell, Ograd)
    !! compute gradient of overlap matrix
    integer, external :: CINT1e_ipovlp_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: Ograd(:, :, :)

    integer :: i, j
    integer :: mu
    integer :: atom1, atom2
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: shls(2)
    real(long), allocatable :: data(:, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :)
#endif
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    Ograd(:, :, :) = ZERO

    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        allocate (data(n1, n2, 3))
        data = ZERO
        i = cint1e_ipovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        data = -data ! sign from library
#ifdef REORDERP
        allocate (data_temp(n1, n2, 3))
        if (i .ne. 0) then! if return value is zero integrals are zero by symmetry
          if (bas(2, sh1) .eq. 1) then
            data_temp(1, :, :) = data(2, :, :)
            data_temp(2, :, :) = data(3, :, :)
            data_temp(3, :, :) = data(1, :, :)
            data = data_temp
          end if
          if (bas(2, sh2) .eq. 1) then
            data_temp(:, 1, :) = data(:, 2, :)
            data_temp(:, 2, :) = data(:, 3, :)
            data_temp(:, 3, :) = data(:, 1, :)
            data = data_temp
          end if
        end if
        deallocate (data_temp)
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1

            if (mod(m1 + m2, 2) .eq. 0) then
              do mu = 1, 3 ! mu=x,y,z
                Ograd(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) = &
                  Ograd(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) + data(i, j, mu)
                Ograd(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) = &
                  Ograd(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) + data(i, j, mu)
              end do
            else
              do mu = 1, 3 ! mu=x,y,z
                Ograd(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) = &
                  Ograd(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) - data(i, j, mu)
                Ograd(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) = &
                  Ograd(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) - data(i, j, mu)
              end do
            end if
          end do ! n2
        end do! n1
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
  end subroutine overlap_gradient

  subroutine overlap_energy_gradient_ROHF(env, atm, bas, first_bf_of_shell, LC, Fa, Fb, n_occ, gradient)
    !! computes the gradient due to basis overlap
    integer, external :: CINT1e_ipovlp_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: LC(:, :)
    real(long), intent(in) :: Fa(:, :)
    real(long), intent(in) :: Fb(:, :)
    integer, intent(in) :: n_occ(:)
    real(long), intent(out) :: gradient(:)

    integer :: i, j
    integer :: mu, nu
    integer :: atom1, atom2
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: shls(2)
    real(long), allocatable :: data(:, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :)
#endif
    real(long), allocatable :: W(:, :)
    real(long), allocatable :: Fat(:, :)
    real(long), allocatable :: Fbt(:, :)
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    allocate (Fat(size(LC, 2), size(LC, 2)))
    allocate (Fbt(size(LC, 2), size(LC, 2)))
    Fat = LC.tm.Fa.m.LC
    Fbt = LC.tm.Fb.m.LC
    gradient(:) = ZERO
    allocate (W(size(LC, 1), size(LC, 1)))
    W(:, :) = ZERO

    do mu = 1, size(LC, 1)
      do nu = 1, size(LC, 1)
        do i = 1, size(n_occ)
          do j = 1, size(n_occ)
            if (n_occ(i) .gt. 0 .and. n_occ(j) .gt. 0) W(mu, nu) = W(mu, nu) + LC(mu, i)*LC(nu, j)*Fat(i, j)
          end do
        end do
        do i = 1, size(n_occ)
          do j = 1, size(n_occ)
            if (n_occ(i) .eq. 2 .and. n_occ(j) .eq. 2) W(mu, nu) = W(mu, nu) + LC(mu, i)*LC(nu, j)*Fbt(i, j)
          end do
        end do
      end do
    end do

    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        allocate (data(n1, n2, 3))
        data = ZERO
        i = cint1e_ipovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (i .ne. 0) then! if return value is zero integrals are zero by symmetry
          allocate (data_temp(n1, n2, 3))
          if (bas(2, sh1) .eq. 1) then
            data_temp(1, :, :) = data(2, :, :)
            data_temp(2, :, :) = data(3, :, :)
            data_temp(3, :, :) = data(1, :, :)
            data = data_temp
          end if
          if (bas(2, sh2) .eq. 1) then
            data_temp(:, 1, :) = data(:, 2, :)
            data_temp(:, 2, :) = data(:, 3, :)
            data_temp(:, 3, :) = data(:, 1, :)
            data = data_temp
          end if
          deallocate (data_temp)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1

            if (mod(m1 + m2, 2) .eq. 0) then
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1)* &
                  (data(i, j, mu))
              end do
            else
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1)* &
                  (-data(i, j, mu))
              end do
            end if
          end do ! n2
        end do! n1
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
  end subroutine overlap_energy_gradient_ROHF

  subroutine overlap_energy_gradient_UHF(env, atm, bas, first_bf_of_shell, LCa, LCb, Fa, Fb, n_occ, gradient)
    !! computes the gradient due to basis overlap
    integer, external :: CINT1e_ipovlp_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: LCa(:, :)
    real(long), intent(in) :: LCb(:, :)
    real(long), intent(in) :: Fa(:, :)
    real(long), intent(in) :: Fb(:, :)
    integer, intent(in) :: n_occ(:)
    real(long), intent(out) :: gradient(:)

    integer :: i, j
    integer :: mu, nu
    integer :: atom1, atom2
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: shls(2)
    real(long), allocatable :: data(:, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :)
#endif
    real(long), allocatable :: W(:, :)
    real(long), allocatable :: Fat(:, :)
    real(long), allocatable :: Fbt(:, :)
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)

    allocate (Fat(size(LCa, 2), size(LCa, 2)))
    allocate (Fbt(size(LCb, 2), size(LCb, 2)))
    Fat = LCa.tm.Fa.m.LCa
    Fbt = LCb.tm.Fb.m.LCb
    gradient(:) = ZERO
    allocate (W(size(LCa, 1), size(LCa, 1)))
    W(:, :) = ZERO

    do mu = 1, size(LCa, 1)
      do nu = 1, size(LCa, 1)
        do i = 1, size(n_occ)
          do j = 1, size(n_occ)
            if (n_occ(i) .gt. 0 .and. n_occ(j) .gt. 0) W(mu, nu) = W(mu, nu) + LCa(mu, i)*LCa(nu, j)*Fat(i, j)
          end do
        end do
        do i = 1, size(n_occ)
          do j = 1, size(n_occ)
            if (n_occ(i) .eq. 2 .and. n_occ(j) .eq. 2) W(mu, nu) = W(mu, nu) + LCb(mu, i)*LCb(nu, j)*Fbt(i, j)
          end do
        end do
      end do
    end do

    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        allocate (data(n1, n2, 3))
        data = ZERO
        i = cint1e_ipovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (i .ne. 0) then! if return value is zero integrals are zero by symmetry
          allocate (data_temp(n1, n2, 3))
          if (bas(2, sh1) .eq. 1) then
            data_temp(1, :, :) = data(2, :, :)
            data_temp(2, :, :) = data(3, :, :)
            data_temp(3, :, :) = data(1, :, :)
            data = data_temp
          end if
          if (bas(2, sh2) .eq. 1) then
            data_temp(:, 1, :) = data(:, 2, :)
            data_temp(:, 2, :) = data(:, 3, :)
            data_temp(:, 3, :) = data(:, 1, :)
            data = data_temp
          end if
          deallocate (data_temp)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1

            if (mod(m1 + m2, 2) .eq. 0) then
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1)* &
                  (data(i, j, mu))
              end do
            else
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1)* &
                  (-data(i, j, mu))
              end do
            end if
          end do ! n2
        end do! n1
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
  end subroutine overlap_energy_gradient_UHF

  subroutine overlap_energy_gradient(env, atm, bas, first_bf_of_shell, LC, CE, f_occ, gradient)
    !! computes derivative of overlap matrix with respect to geometry
    integer, external :: CINT1e_ipovlp_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: LC(:, :)
    real(long), intent(in) :: CE(:)
    real(long), intent(in) :: f_occ(:)
    real(long), intent(out) :: gradient(:)

    integer :: i, j
    integer :: mu, nu
    integer :: atom1, atom2
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: shls(2)
    real(long), allocatable :: data(:, :, :)
#ifdef REORDERP
    real(long), allocatable :: data_temp(:, :, :)
#endif
    real(long), allocatable :: W(:, :)
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    gradient(:) = ZERO
    allocate (W(size(LC, 1), size(LC, 1)))
    W(:, :) = ZERO
    do i = 1, size(f_occ)
      do mu = 1, size(LC, 1)
        do nu = 1, size(LC, 1)
          W(mu, nu) = W(mu, nu) + LC(mu, i)*LC(nu, i)*CE(i)*f_occ(i)
        end do
      end do
    end do
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        allocate (data(n1, n2, 3))
        data = ZERO
        i = cint1e_ipovlp_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
#ifdef REORDERP
        if (i .ne. 0) then! if return value is zero integrals are zero by symmetry
          allocate (data_temp(n1, n2, 3))
          if (bas(2, sh1) .eq. 1) then
            data_temp(1, :, :) = data(2, :, :)
            data_temp(2, :, :) = data(3, :, :)
            data_temp(3, :, :) = data(1, :, :)
            data = data_temp
          end if
          if (bas(2, sh2) .eq. 1) then
            data_temp(:, 1, :) = data(:, 2, :)
            data_temp(:, 2, :) = data(:, 3, :)
            data_temp(:, 3, :) = data(:, 1, :)
            data = data_temp
          end if
          deallocate (data_temp)
        end if
#endif
        do i = 1, n1
          m1 = i - bas(2, sh1) - 1
          do j = 1, n2
            m2 = j - bas(2, sh2) - 1

            if (mod(m1 + m2, 2) .eq. 0) then
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, &
                  first_bf_of_shell(sh2) + j - 1)*data(i, j, mu)
              end do
            else
              do mu = 1, 3 ! mu=x,y,z
                gradient((atom1 - 1)*3 + mu) = &
                  gradient((atom1 - 1)*3 + mu) + TWO*W(first_bf_of_shell(sh1) + i - 1, &
                  first_bf_of_shell(sh2) + j - 1)*(-data(i, j, mu))
              end do
            end if
          end do ! n2
        end do! n1
        deallocate (data)
      end do! sh2
    end do! sh1
    call CINTdel_optimizer(opt)
  end subroutine overlap_energy_gradient

  subroutine one_electron_gradient(env, atm, bas, first_bf_of_shell, D, gradient)
    !! returns the gradient resulting from one-electron contributions (h0)
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: D(:, :)
    real(long), intent(out) :: gradient(:)
    integer :: i, j, k
    real(long), allocatable :: h_grad(:, :, :)
    allocate (h_grad(size(D, 1), size(D, 1), 3*size(atm, 2)))
    call h0_gradient(env, atm, bas, first_bf_of_shell, h_grad)
    gradient(:) = ZERO
    do k = 1, size(atm, 2)*3
      do i = 1, size(D, 1)
        do j = 1, size(D, 2)
          gradient(k) = gradient(k) + D(i, j)*h_grad(i, j, k)
        end do
      end do
    end do
    deallocate (h_grad)
  end subroutine one_electron_gradient

  subroutine Rinv_gradient(env, atm, bas, first_bf_of_shell, pos, at, gradient)
    !! calculates $$\frac{d}{dR} \chi_i 1/\vec{R}_i \chi_j$$
    !! where R is given by pos(1:3) and has atom index 'at'    
    integer, external :: CINT1e_iprinv_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: pos(:)
    integer, intent(in) :: at
    real(long), intent(out) :: gradient(:, :, :)
    integer :: ret1
    integer :: i, j
    integer :: mu
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: atom1, atom2
    integer :: shls(2)
    real(long) :: factor
    real(long), allocatable :: my_env(:)
    real(long), allocatable :: data1(:, :, :)
    integer(8) :: opt
    allocate (my_env(size(env, 1)))
    my_env(:) = env(:)
    my_env(5) = pos(1) !x
    my_env(6) = pos(2) !y
    my_env(7) = pos(3) !z
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), my_env)
    gradient(:, :, :) = ZERO
    ! 2. loop over shell pairs
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        allocate (data1(n1, n2, 3))
        data1 = ZERO
        ! R' is specified via my_env(5:7)

        ! compute <grad sh1| rinv | sh2> +  <sh1| rinv | grad sh2>
        ! and compute < sh1 | grad Z/|r-R'| | sh2 > = <grad sh1| rinv | sh2> + < sh1| rinv |grad sh2>
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        ret1 = cint1e_iprinv_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), my_env, opt)
        if (ret1 .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data1 = reorder_data_p(data1, 1)
          end if
          if (bas(2, sh2) .eq. 1) then
            data1 = reorder_data_p(data1, 2)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              if (mod(m1 + m2, 2) .eq. 0) then
                factor = ONE
              else
                factor = -ONE
              end if
              do mu = 1, 3 ! mu=x,y,z
                gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) = &
                  gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) + &
                  factor*(-data1(i, j, mu))
                gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) = &
                  gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) + &
                  factor*(-data1(i, j, mu))

                gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (at - 1)*3 + mu) = &
                  gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (at - 1)*3 + mu) + factor*data1(i, j, mu)
                gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (at - 1)*3 + mu) = &
                  gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (at - 1)*3 + mu) + factor*data1(i, j, mu)

              end do
            end do ! j=1,n2
          end do! i=1,n1
        end if! if ret1/=0 and ret2/=0
        deallocate (data1)
      end do! sh2
    end do! sh1
#ifdef REORDERP
  contains
    function reorder_data_p(data, p_index) result(r_data)
      integer, intent(in) :: p_index
      real(long), intent(in) :: data(:, :, :)
      real(long) :: r_data(size(data, 1), size(data, 2), size(data, 3))
      if (size(data, p_index) .ne. 3) stop "error in reorder_data_p"
      if (p_index .eq. 1) then
        r_data(1, :, :) = data(2, :, :)
        r_data(2, :, :) = data(3, :, :)
        r_data(3, :, :) = data(1, :, :)
      else if (p_index .eq. 2) then
        r_data(:, 1, :) = data(:, 2, :)
        r_data(:, 2, :) = data(:, 3, :)
        r_data(:, 3, :) = data(:, 1, :)
      end if
    end function reorder_data_p
#endif
    call CINTdel_optimizer(opt)
  end subroutine Rinv_gradient



  subroutine rinv_gradient_int(env, atm, bas, first_bf_of_shell, pos, gradient)
    !! calculates  $$\langle \chi_i \frac{d}{dR_i} (1/\vec{R}_i)  \cih_j \rangle$$
    !! where R_ is given by pos(1:3)    
    integer, external :: CINT1e_iprinv_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: pos(:)
    real(long), intent(out) :: gradient(:, :, :)
    integer :: ret1
    integer :: i, j
    integer :: mu
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: atom1, atom2
    integer :: shls(2)
    real(long) :: factor
    real(long), allocatable :: my_env(:)
    real(long), allocatable :: data1(:, :, :)
    integer(8) :: opt
    allocate (my_env(size(env, 1)))
    my_env(:) = env(:)
    my_env(5) = pos(1) !x
    my_env(6) = pos(2) !y
    my_env(7) = pos(3) !z
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), my_env)
    gradient(:, :, :) = ZERO
    ! 2. loop over shell pairs
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      atom1 = bas(1, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        ! calculate  < sh1 | grad Z/|r-R_atom| | sh2 >
        !          = < grad sh1 |Z/|r-R_atom| sh2>+<sh1 |Z/|r-R_atom| grad sh2>
        ! Rinv=Ratom is specified via my_env(5:7)
        allocate (data1(n1, n2, 3))
        data1 = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        ret1 = cint1e_iprinv_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), my_env, opt)

        if (ret1 .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data1 = reorder_data_p(data1, 1)
          end if
          if (bas(2, sh2) .eq. 1) then
            data1 = reorder_data_p(data1, 2)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              if (mod(m1 + m2, 2) .eq. 0) then
                factor = ONE
              else
                factor = -ONE
              end if
              do mu = 1, 3 ! mu=x,y,z
                gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, mu) = &
                  gradient(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, mu) + &
                  factor*(data1(i, j, mu))
                gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, mu) = &
                  gradient(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, mu) + &
                  factor*(data1(i, j, mu))
              end do
            end do ! j=1,n2
          end do! i=1,n1
        end if! if ret1/=0
        deallocate (data1)
      end do! sh2
    end do! sh1
#ifdef REORDERP
  contains
    function reorder_data_p(data, p_index) result(r_data)
      integer, intent(in) :: p_index
      real(long), intent(in) :: data(:, :, :)
      real(long) :: r_data(size(data, 1), size(data, 2), size(data, 3))
      if (size(data, p_index) .ne. 3) stop "error in reorder_data_p"
      if (p_index .eq. 1) then
        r_data(1, :, :) = data(2, :, :)
        r_data(2, :, :) = data(3, :, :)
        r_data(3, :, :) = data(1, :, :)
      else if (p_index .eq. 2) then
        r_data(:, 1, :) = data(:, 2, :)
        r_data(:, 2, :) = data(:, 3, :)
        r_data(:, 3, :) = data(:, 1, :)
      end if
    end function reorder_data_p
#endif
    call CINTdel_optimizer(opt)
  end subroutine rinv_gradient_int

  subroutine two_electron_gradient(env, atm, bas, first_bf_of_shell, D, gradient, Fock, accuracy)
    !! returns the gradient resulting from two-electron contributions 
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: D(:, :)
    real(long), intent(out) :: gradient(:)
    logical, intent(in), optional :: Fock
    real(long), optional, intent(in) :: accuracy

    logical :: yes_Fock
    real(long), allocatable :: Jd(:, :, :)
    real(long), allocatable :: Kd(:, :, :)
    integer :: i, j, k
    real(long) :: epsilonI
    epsilonI = ZERO
    if (present(accuracy)) epsilonI = accuracy
    yes_Fock = .FALSE.
    if (present(Fock)) yes_Fock = Fock
    gradient(:) = ZERO
    allocate (Jd(size(D, 1), size(D, 1), 3*size(atm, 2)))
    allocate (Kd(size(D, 1), size(D, 1), 3*size(atm, 2)))
    call jk_gradient(env, atm, bas, first_bf_of_shell, D, Jd, Kd, symmetricD=.True., accuracy=epsilonI)
    gradient(:) = ZERO
    do concurrent(k=1:size(atm, 2)*3)
      do j = 1, size(D, 1)
        do i = 1, size(D, 1)
          if (yes_Fock) then
            gradient(k) = gradient(k) + HALF*D(i, j)*(Jd(i, j, k) - HALF*Kd(i, j, k))
          else
            gradient(k) = gradient(k) + HALF*D(i, j)*(Jd(i, j, k))
          end if
        end do
      end do
    end do
    deallocate (Jd, Kd)
  end subroutine two_electron_gradient

  subroutine two_electron_gradient_ROHF(env, atm, bas, first_bf_of_shell, Da, Db, gradient)
    !! returns the gradient resulting from two-electron contributions for ROHF
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: Da(:, :)
    real(long), intent(in) :: Db(:, :)
    real(long), intent(out) :: gradient(:)
    real(long), allocatable :: Jda(:, :, :)
    real(long), allocatable :: Kda(:, :, :)
    real(long), allocatable :: Jdb(:, :, :)
    real(long), allocatable :: Kdb(:, :, :)
    integer :: i, j, k
    gradient(:) = ZERO

    allocate (Jda(size(Da, 1), size(Da, 1), 3*size(atm, 2)))
    allocate (Kda(size(Da, 1), size(Da, 1), 3*size(atm, 2)))
    allocate (Jdb(size(Da, 1), size(Da, 1), 3*size(atm, 2)))
    allocate (Kdb(size(Da, 1), size(Da, 1), 3*size(atm, 2)))
    call jk_gradient(env, atm, bas, first_bf_of_shell, Da, Jda, Kda)
    call jk_gradient(env, atm, bas, first_bf_of_shell, Db, Jdb, Kdb)
    gradient(:) = ZERO
    do k = 1, size(atm, 2)*3
      do j = 1, size(Da, 1)
        do i = 1, size(Da, 1)
          gradient(k) = gradient(k) + HALF*Da(i, j)*(Jda(i, j, k) + Jdb(i, j, k) - Kda(i, j, k))
          gradient(k) = gradient(k) + HALF*Db(i, j)*(Jda(i, j, k) + Jdb(i, j, k) - Kdb(i, j, k))
        end do
      end do
    end do
    deallocate (Jda, Kda)
    deallocate (Jdb, Kdb)
  end subroutine two_electron_gradient_ROHF

  subroutine jk_gradient(env, atm, bas, first_bf_of_shell, D, Jd, Kd, symmetricD, accuracy)
    !! computes derivatives of J and K
    ! use the optional parameter symmetricD=False, if D is asymmetric (e.g. for CIS gradients)!
    integer, external :: cint2e_ip1_sph
    integer, external :: cint2e_ip1ip2_sph
    integer, external :: cint2e_sph
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(in) :: D(:, :)
    real(long), intent(out) :: Jd(:, :, :)
    real(long), intent(out) :: Kd(:, :, :)

    logical, optional, intent(in) :: symmetricD
    real(long), optional, intent(in) :: accuracy

    real(long) :: Jd2(size(Jd, 3), size(Jd, 1), size(Jd, 2))
    real(long) :: Kd2(size(Kd, 3), size(Kd, 1), size(Kd, 2))
    
    real(long), allocatable :: Jdlk(:, :, :), Jdjk(:, :, :), Jdlj(:, :, :), Jdli(:,:,:), Jdji(:,:,:)
    real(long), allocatable :: Jdii(:,:), Jdjj(:, :), Jdkk(:, :)
    real(long), allocatable :: Kdij(:, :, :), Kdik(:, :, :), Kdil(:, :, :)
    real(long), allocatable :: Kdji(:, :, :), Kdjk(:, :, :), Kdjl(:, :, :)
    real(long), allocatable :: Kdki(:, :, :), Kdkj(:, :, :)
    real(long), allocatable :: Kdli(:, :, :), Kdlj(:, :, :)
    real(long), allocatable :: Kdii(:, :), Kdjj(:, :)
    logical :: yesDSym
    real(long) :: Dsym(size(D, 1), size(D, 2))
    integer :: i, j, k, l
    integer(8) :: ij, kl
    integer :: bf1, bf2, bf3, bf4
    integer :: sh1, sh2, sh3, sh4
    integer(8) :: sh12, sh34
    integer :: m1, m2, m3, m4
    integer :: n1, n2, n3, n4
    integer :: shls(4)
    integer(8) :: opt
    integer(8) :: optip
    integer(8) :: optipip
    real(long), allocatable :: data(:, :, :, :)
    real(long), allocatable :: data12(:, :, :, :, :, :)
    real(long), allocatable :: data1(:, :, :, :, :)
    real(long), allocatable :: data2(:, :, :, :, :)
    real(long), allocatable :: data3(:, :, :, :, :)
    real(long), allocatable :: data4(:, :, :, :, :)
    integer :: res, res1, res2, res3, res4
    integer :: atom1, atom2, atom3, atom4
    real(long) :: teintd(12)
    integer(8) :: ijmax, klmin, imin, jmin, imax, jmax, kmin, lmin, kmax, lmax
    integer :: a1i, a1e, a2i, a2e, a3i, a3e, a4i, a4e
    real(long) :: epsilonI, epsilonIsq
    real(long) :: schwarz(size(bas, 2) * (size(bas, 2) + 1)/2)
    real(long) :: schwarzD(size(bas, 2), size(bas, 2))
    logical :: skipShell, needsUpdate
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    epsilonI = ZERO
    if (present(accuracy)) epsilonI = accuracy
    epsilonIsq = epsilonI**2
    yesDSym = .True.
    if (present(symmetricD)) &
      yesDSym = symmetricD
    Dsym = D + transpose(D)
    Jd(:, :, :) = ZERO
    Kd(:, :, :) = ZERO
    Jd2(:, :, :) = ZERO
    Kd2(:, :, :) = ZERO
    opt = 0
    call cint2e_sph_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    optipip = 0
    call cint2e_ip1ip2_sph_optimizer(optipip, atm, size(atm, 2), bas, size(bas, 2), env)
    ! Schwarz screening:
    ! |(sh1',sh2| sh3,sh4)| < sqrt(|(sh1',sh2| sh1',sh2)| * |(sh3,sh4| sh3,sh4)|)
    ! so we need integrals of type (sh1',sh2| sh1',sh2) and (sh1,sh2| sh1,sh2)
    schwarz = ZERO
    schwarzD = ZERO
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      do sh2 = 1, size(bas, 2) ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        shls(3) = sh1 - 1
        shls(4) = sh2 - 1
        if (sh1 .ge. sh2) then
          ! calculate (sh1 sh2| sh1 sh2)
          allocate (data(n1, n2, n1, n2))
          data = ZERO
          sh12 = sh1*(sh1 - 1)/2 + sh2
          res = cint2e_sph(data, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
          if (res .ne. 0) then
            schwarz(sh12) = maxval(abs(data))
          else
            schwarz(sh12) = ZERO
          end if
          deallocate (data)
        end if
        ! calculate (sh1' sh2| sh1' sh2)
        allocate (data12(n1, n2, n1, n2, 3, 3))
        data12 = ZERO
        res1 = cint2e_ip1ip2_sph(data12, shls, atm, size(atm, 2), bas, size(bas, 2), env, optipip) ! (sh1' sh2 | sh1' sh2)
        if (res1 .ne. 0) then
          schwarzD(sh1, sh2) = maxval(abs(data12))
        else
          schwarzD(sh1, sh2) = ZERO
        end if
        deallocate (data12)
      end do
    end do
    call CINTdel_optimizer(opt)
    call CINTdel_optimizer(optipip)
    ! end of preparations for Schwarz screening
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    ! loop through the index combinations i<=j, k<=l, and ij=>kl
    ! this implies
    ! sh1<=sh2, sh3<=sh4
    !$omp parallel default(private) shared(lock,schwarzD,schwarz,bas,atm,env,first_bf_of_shell,D,Dsym,yesDSym,epsilonIsq,Jd2,Kd2)
    !!$omp private(sh1,sh2,sh3,sh4,sh12,sh34,n1,n2,n3,n4,i,j,k,l,ij,kl,data1,data2,data3,data4,atom1,atom2,atom3,atom4,a1i,a1e,a2i,a2e,a3i,a3e,a4i,a4e,shls,res1,res2,res3,res4,bf1,bf2,bf3,bf4,imin,imax,jmin,jmax,ijmax,kmin,lmin,kmax,lmax,klmin,m1,m2,m3,m4,teintd,optip) copyin(data1,data2,data3,data4)
    !!$omp 
    optip = 0
    call cint2e_ip1_sph_optimizer(optip, atm, size(atm, 2), bas, size(bas, 2), env)
    !$omp do schedule(guided)
    do sh1 = 1, size(bas, 2) ! loop over shells
      n1 = 2*bas(2, sh1) + 1
      imin = first_bf_of_shell(sh1)
      imax = first_bf_of_shell(sh1) + n1 - 1
      allocate (Jdii(12, n1)); Jdii(:,:) = ZERO
      allocate (Kdii(12, n1)); Kdii(:,:) = ZERO
      do sh2 = 1, sh1 ! loop over shells
        n2 = 2*bas(2, sh2) + 1
        jmin = first_bf_of_shell(sh2)
        jmax = first_bf_of_shell(sh2) + n2 - 1
        sh12 = sh1*(sh1 - 1)/2 + sh2
        ! idea make Jlocal(12,N,N) and Klocal(12,N,N)
        allocate (Jdjj(12, n2)); Jdjj(:,:) = ZERO
        allocate (Jdji(12, n2, n1)); Jdji(:,:,:) = ZERO
        allocate (Kdjj(12, n2)); Kdjj(:,:) = ZERO
        allocate (Kdij(12, n1, n2)); Kdij(:,:,:) = ZERO
        allocate (Kdji(12, n2, n1)); Kdji(:,:,:) = ZERO
        do sh3 = 1, size(bas, 2) ! loop over shells
          n3 = 2*bas(2, sh3) + 1
          kmin = first_bf_of_shell(sh3)
          kmax = first_bf_of_shell(sh3) + n3 - 1
          allocate (Jdkk(12, n3)); Jdkk(:,:) = ZERO
          allocate (Kdik(12, n1, n3)); Kdik(:,:,:) = ZERO
          allocate (Kdki(12, n3, n1)); Kdki(:,:,:) = ZERO
          allocate (Kdjk(12, n2, n3)); Kdjk(:,:,:) = ZERO
          allocate (Kdkj(12, n3, n2)); Kdkj(:,:,:) = ZERO
          allocate (Jdjk(12, n2, n3)); Jdjk(:,:,:) = ZERO
          needsUpdate = .False.
          do sh4 = 1, sh3 ! loop over shells
            n4 = 2*bas(2, sh4) + 1
            sh34 = sh3*(sh3 - 1)/2 + sh4
            ! this prescreens condition ij>=kl
            ! if max(ij) < min(kl) cycle
            lmin = first_bf_of_shell(sh4)
            lmax = first_bf_of_shell(sh4) + n4 - 1
            ijmax = index_packed(imax, jmax)
            klmin = index_packed(kmin, lmin)
            SkipShell = (ijmax .lt. klmin)
            if (.not. SkipShell) &
                 SkipShell = ((schwarzD(sh1, sh2) + schwarzD(sh2, sh1))*schwarz(sh34) .lt. epsilonIsq &
                 .and. (schwarzD(sh3, sh4) + schwarzD(sh4, sh3))*schwarz(sh12) .lt. epsilonIsq)
            if (.not. SkipShell) then
              allocate (Jdli(12, n4, n1)); Jdli(:,:,:) = ZERO
              allocate (Jdlj(12, n4, n2)); Jdlj(:,:,:) = ZERO
              allocate (Jdlk(12, n4, n3)); Jdlk(:,:,:) = ZERO
              allocate (Kdil(12, n1, n4)); Kdil(:,:,:) = ZERO
              allocate (Kdli(12, n4, n1)); Kdli(:,:,:) = ZERO
              allocate (Kdjl(12, n2, n4)); Kdjl(:,:,:) = ZERO
              allocate (Kdlj(12, n4, n2)); Kdlj(:,:,:) = ZERO
              allocate (data1(n1, n2, n3, n4, 3))
              allocate (data2(n2, n1, n3, n4, 3))
              allocate (data3(n3, n4, n1, n2, 3))
              allocate (data4(n4, n3, n1, n2, 3))
              data1 = ZERO
              data2 = ZERO
              data3 = ZERO
              data4 = ZERO
              atom1 = bas(1, sh1) + 1
              atom2 = bas(1, sh2) + 1
              atom3 = bas(1, sh3) + 1
              atom4 = bas(1, sh4) + 1
              a1i = (atom1 - 1)*3 + 1
              a1e = a1i + 2
              a2i = (atom2 - 1)*3 + 1
              a2e = a2i + 2
              a3i = (atom3 - 1)*3 + 1
              a3e = a3i + 2
              a4i = (atom4 - 1)*3 + 1
              a4e = a4i + 2
              ! calculate grad on sh1
              shls(1) = sh1 - 1
              shls(2) = sh2 - 1
              shls(3) = sh3 - 1
              shls(4) = sh4 - 1
              res1 = cint2e_ip1_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), env, optip) ! (sh1' sh2 | sh3 sh4)
              if (sh1 .eq. sh2) then
                data2 = data1 ! (sh2' sh1| sh3 sh4) = (sh1' sh2| sh3 sh4)
                res2 = res1
              else
                ! calculate grad on sh2
                shls(1) = sh2 - 1
                shls(2) = sh1 - 1
                shls(3) = sh3 - 1
                shls(4) = sh4 - 1
                res2 = cint2e_ip1_sph(data2, shls, atm, size(atm, 2), bas, size(bas, 2), env, optip) ! (sh2' sh1 | sh3 sh4)
              end if
              if (sh1 .eq. sh3 .and. sh2 .eq. sh4) then
                data3 = data1 ! (sh3' sh4| sh1 sh2) = (sh1' sh2| sh3 sh4)
                res3 = res1
              else if (sh2 .eq. sh3 .and. sh1 .eq. sh4) then
                data3 = data2 ! (sh3' sh4| sh1 sh2) = (sh2' sh1| sh3 sh4)
                res3 = res2
              else
                ! calculate grad on sh3
                shls(1) = sh3 - 1
                shls(2) = sh4 - 1
                shls(3) = sh1 - 1
                shls(4) = sh2 - 1
                res3 = cint2e_ip1_sph(data3, shls, atm, size(atm, 2), bas, size(bas, 2), env, optip)   ! (sh3' sh4 | sh1 sh2)
              end if
              if (sh4 .eq. sh1 .and. sh3 .eq. sh2) then
                data4 = data1 ! (sh4' sh3 | sh1 sh2) = ! (sh1' sh2 | sh3 sh4)
                res4 = res1
              else if (sh4 .eq. sh2 .and. sh3 .eq. sh1) then
                data4 = data2 ! (sh4' sh3 | sh1 sh2) = ! (sh2' sh1 | sh3 sh4)
                res4 = res2
              else if (sh4 .eq. sh3) then
                data4 = data3 ! (sh4' sh3 | sh1 sh2) = ! (sh3' sh4 | sh1 sh2)
                res4 = res3
              else
                ! calculate grad on sh4
                shls(1) = sh4 - 1
                shls(2) = sh3 - 1
                shls(3) = sh1 - 1
                shls(4) = sh2 - 1
                res4 = cint2e_ip1_sph(data4, shls, atm, size(atm, 2), bas, size(bas, 2), env, optip) ! (sh4' sh3 | sh1 sh2)
              end if
              if (res1 .ne. 0 .or. res2 .ne. 0 .or. res3 .ne. 0 .or. res4 .ne. 0) then! if return value is zero integrals are zero by symmetry
                ! libcint orders the p functions as x,y,z
                ! we use y,z,x so we need to reorder the p basis functions
                needsUpdate = .True.
#ifdef REORDERP
                if (bas(2, sh1) .eq. 1) then
                  data1 = reorder_data_p(data1, 1)
                  data2 = reorder_data_p(data2, 2)
                  data3 = reorder_data_p(data3, 3)
                  data4 = reorder_data_p(data4, 3)
                end if
                if (bas(2, sh2) .eq. 1) then
                  data1 = reorder_data_p(data1, 2)
                  data2 = reorder_data_p(data2, 1)
                  data3 = reorder_data_p(data3, 4)
                  data4 = reorder_data_p(data4, 4)
                end if
                if (bas(2, sh3) .eq. 1) then
                  data1 = reorder_data_p(data1, 3)
                  data2 = reorder_data_p(data2, 3)
                  data3 = reorder_data_p(data3, 1)
                  data4 = reorder_data_p(data4, 2)
                end if
                if (bas(2, sh4) .eq. 1) then
                  data1 = reorder_data_p(data1, 4)
                  data2 = reorder_data_p(data2, 4)
                  data3 = reorder_data_p(data3, 2)
                  data4 = reorder_data_p(data4, 1)
                end if
#endif
                ! loop over basis functions combination that are within the shell combinations sh1,sh2,sh3,sh4
                do bf1 = 1, size(data1, 1)
                  m1 = bf1 - bas(2, sh1) - 1
                  i = first_bf_of_shell(sh1) + bf1 - 1
                  do bf2 = 1, size(data1, 2)
                    m2 = bf2 - bas(2, sh2) - 1
                    j = first_bf_of_shell(sh2) + bf2 - 1
                    if (i .lt. j) cycle
                    ij = index_packed(i, j)
                    do bf3 = 1, size(data1, 3)
                      m3 = bf3 - bas(2, sh3) - 1
                      k = first_bf_of_shell(sh3) + bf3 - 1
                      do bf4 = 1, size(data1, 4)
                        l = first_bf_of_shell(sh4) + bf4 - 1
                        m4 = bf4 - bas(2, sh4) - 1
                        if (k .lt. l) cycle
                        kl = index_packed(k, l)
                        if (ij .lt. kl) cycle
                        ! the libc libray uses a different phase convention:
                        ! multiply with (-1)^m1 (-1)^m2 (-1)^m3 (-1)^m4
                        if (mod(m1 + m2 + m3 + m4, 2) .eq. 0) then
                          ! somehow there must be an additional factor of -1
                          teintd(1:3) = -data1(bf1, bf2, bf3, bf4, :) ! (i'j | k l )
                          teintd(4:6) = -data2(bf2, bf1, bf3, bf4, :) ! (i j'| k l )
                          teintd(7:9) = -data3(bf3, bf4, bf1, bf2, :) ! (i j | k'l )
                          teintd(10:12) = -data4(bf4, bf3, bf1, bf2, :) ! (i j | k l')
                        else
                          teintd(1:3) = data1(bf1, bf2, bf3, bf4, :) ! (i'j | k l )
                          teintd(4:6) = data2(bf2, bf1, bf3, bf4, :) ! (i j'| k l )
                          teintd(7:9) = data3(bf3, bf4, bf1, bf2, :) ! (i j | k'l )
                          teintd(10:12) = data4(bf4, bf3, bf1, bf2, :) ! (i j | k l')
                        end if
                        ! integral screening
                        !if ( all(abs(teint1) < epsilonI) .and. &
                        !     all(abs(teint2) < epsilonI) .and. &
                        !     all(abs(teint3) < epsilonI) .and. &
                        !     all(abs(teint4) < epsilonI)) cycle
                        ! Now add the contributions to Jd or Kd
                        ! 4 indices equal
                        if (i .eq. j .and. k .eq. l .and. i .eq. k) then! (ii|ii)
                          !J(i,i)=J(i,i)+D(i,i)*teint
                          Jdii(:, bf1) = Jdii(:, bf1) + D(i, i) * teintd(:)
                          !K(i,i)=K(i,i)+D(i,i)*teint
                          !Kdloc(:, i, i) = Kdloc(:, i, i) + D(i, i) * teintd(:)
                          Kdii(:, bf1) = Kdii(:, bf1) + D(i, i) * teintd(:)
                          ! 3 indices equal
                        else if (i .eq. j .and. j .eq. k) then ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
                          !J(i,l)=J(i,l)+D(i,i)*teint
                          Jdli(:, bf4, bf1) = Jdli(:, bf4, bf1) + D(i, i) * teintd(:)
                          !K(i,l)=K(i,l)+D(i,i)*teint
                          !Kdloc(:, l, i) = Kdloc(:, l, i) + D(i, i)*teintd(:)
                          Kdli(:, bf4, bf1) = Kdli(:, bf4, bf1) + D(i, i) * teintd(:)
                          !J(i,i)=J(i,i)+D(i,l)*teint
                          !J(i,i)=J(i,i)+D(l,i)*teint
                          Jdii(:, bf1) = Jdii(:, bf1) + Dsym(i, l) * teintd(:)
                          !K(i,i)=K(i,i)+D(i,l)*teint
                          !K(i,i)=K(i,i)+D(l,i)*teint
                          !Kdloc(:, i, i) = Kdloc(:, i, i) + Dsym(i, l)*teintd(:)
                          Kdii(:, bf1) = Kdii(:, bf1) + Dsym(i, l) * teintd(:)
                          if (.not. yesDSym) then
                            !K(l,i)=K(l,i)+D(i,i)*teint
                            !Kdloc(:, i, l) = Kdloc(:, i, l) + D(i, i)*teintd(:)
                            Kdil(:, bf1, bf4) = Kdil(:, bf1, bf4) + D(i, i) * teintd(:)
                          end if
                        else if (i .eq. k .and. k .eq. l) then ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
                          !J(i,j)=J(i,j)+D(i,i)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + D(i, i) * teintd(:)
                          !K(i,j)=K(i,j)+D(i,i)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(i, i)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(i, i) * teintd(:)
                          !J(i,i)=J(i,i)+D(i,j)*teint
                          !J(i,i)=J(i,i)+D(j,i)*teint
                          Jdii(:, bf1) = Jdii(:, bf1) + Dsym(i, j) * teintd(:)
                          !K(i,i)=K(i,i)+D(j,i)*teint
                          !K(i,i)=K(i,i)+D(i,j)*teint
                          !Kdloc(:, i, i) = Kdloc(:, i, i) + Dsym(i, j)*teintd(:)
                          Kdii(:, bf1) = Kdii(:, bf1) + Dsym(i, j) * teintd(:)
                          if (.not. yesDSym) then
                            !K(j,i)=K(j,i)+D(i,i)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(i, i)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(i, i) * teintd(:)
                          end if
                        else if (j .eq. k .and. k .eq. l) then ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
                          !J(i,j)=J(i,j)+D(j,j)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + D(j, j) * teintd(:)
                          !K(i,j)=K(i,j)+D(j,j)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(j, j)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(j, j) * teintd(:)
                          !J(j,j)=J(j,j)+D(i,j)*teint
                          !J(j,j)=J(j,j)+D(j,i)*teint
                          Jdjj(:, bf2) = Jdjj(:, bf2) + Dsym(i, j) * teintd(:)
                          !K(j,j)=K(j,j)+D(i,j)*teint
                          !K(j,j)=K(j,j)+D(j,i)*teint
                          !Kdloc(:, j, j) = Kdloc(:, j, j) + Dsym(i, j)*teintd(:)
                          Kdjj(:, bf2) = Kdjj(:, bf2) + Dsym(i, j) * teintd(:)
                          if (.not. yesDSym) then
                            !K(j,i)=K(j,i)+D(j,j)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(j, j)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(j, j) * teintd(:)
                          end if
                          ! two indices pairwise equal
                        else if (i .eq. j .and. k .eq. l) then ! (ii|kk) (kk|ii)
                          !J(i,i)=J(i,i)+D(k,k)*teint
                          Jdii(:, bf1) = Jdii(:, bf1) + D(k, k) * teintd(:)
                          !J(k,k)=J(k,k)+D(i,i)*teint
                          Jdkk(:, bf3) = Jdkk(:, bf3) + D(i, j) * teintd(:)                        
                          !K(i,k)=K(i,k)+D(i,k)*teint
                          !Kdloc(:, k, i) = Kdloc(:, k, i) + D(k, i)*teintd(:)
                          Kdki(:, bf3, bf1) = Kdki(:, bf3, bf1) + D(k, i) * teintd(:)
                          if (.not. yesDSym) then
                            !K(k,i)=K(k,i)+D(k,i)*teint
                            !Kdloc(:, i, k) = Kdloc(:, i, k) + D(i, k)*teintd(:)
                            Kdik(:, bf1, bf3) = Kdik(:, bf1, bf3) + D(i, k) * teintd(:)
                          end if
                        else if (i .eq. k .and. j .eq. l) then ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
                          !J(i,j)=J(i,j)+D(j,i)*teint
                          !J(i,j)=J(i,j)+D(i,j)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + Dsym(j, i)*teintd(:)
                          !K(i,j)=K(i,j)+D(j,i)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(i, j)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(i, j) * teintd(:)
                          !K(i,i)=K(i,i)+D(j,j)*teint
                          !Kdloc(:, i, i) = Kdloc(:, i, i) + D(j, j)*teintd(:)
                          Kdii(:, bf1) = Kdii(:, bf1) + D(j, j) * teintd(:)
                          !K(j,j)=K(j,j)+D(i,i)*teint
                          !Kdloc(:, j, j) = Kdloc(:, j, j) + D(i, i)*teintd(:)
                          Kdjj(:, bf2) = Kdjj(:, bf2) + D(i, i) * teintd(:)
                          if (.not. yesDSym) then
                            !K(j,i)=K(j,i)+D(i,j)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(j, i)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(j, i) * teintd(:)
                          end if
                          ! two indices equal
                        else if (i .eq. j) then ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
                          !J(i,i)=J(i,i)+D(k,l)*teint
                          !J(i,i)=J(i,i)+D(l,k)*teint
                          Jdii(:, bf1) = Jdii(:, bf1) + Dsym(k, l) * teintd(:)
                          !J(k,l)=J(k,l)+D(i,i)*teint
                          Jdlk(:, bf4, bf3) = Jdlk(:, bf4, bf3) + D(i, i) * teintd(:)
                          !K(i,k)=K(i,k)+D(i,l)*teint
                          !Kdloc(:, k, i) = Kdloc(:, k, i) + D(l, i)*teintd(:)
                          Kdki(:, bf3, bf1) = Kdki(:, bf3, bf1) + D(l, i) * teintd(:)
                          !K(i,l)=K(i,l)+D(i,k)*teint
                          !Kdloc(:, l, i) = Kdloc(:, l, i) + D(k, i)*teintd(:)
                          Kdli(:, bf4, bf1) = Kdli(:, bf4, bf1) + D(k, i) * teintd(:)
                          if (.not. yesDSym) then
                            !K(k,i)=K(k,i)+D(i,l)*teint
                            !Kdloc(:, i, k) = Kdloc(:, i, k) + D(i, l)*teintd(:)
                            Kdik(:, bf1, bf3) = Kdik(:, bf1, bf3) + D(i, l) * teintd(:)
                            !K(l,i)=K(l,i)+D(k,i)*teint
                            !Kdloc(:, i, l) = Kdloc(:, i, l) + D(i, k)*teintd(:)
                            Kdil(:, bf1, bf4) = Kdil(:, bf1, bf4) + D(i, k) * teintd(:)
                          end if
                        else if (k .eq. l) then ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
                          !J(i,j)=J(i,j)+D(k,k)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + D(k, k) * teintd(:)
                          !J(k,k)=J(k,k)+D(i,j)*teint
                          !J(k,k)=J(k,k)+D(j,i)*teint
                          Jdkk(:, bf3) = Jdkk(:, bf3) + Dsym(j, i) * teintd(:)
                          !if(i>k)then
                          !K(i,k)=K(i,k)+D(j,k)*teint
                          !Kdloc(:, k, i) = Kdloc(:, k, i) + D(k, j)*teintd(:)
                          Kdki(:, bf3, bf1) = Kdki(:, bf3, bf1) + D(k, j) * teintd(:)
                          if (.not. yesDSym) then
                            !K(k,i)=K(k,i)+D(k,j)*teint
                            !Kdloc(:, i, k) = Kdloc(:, i, k) + D(j, k)*teintd(:)
                            Kdik(:, bf1, bf3) = Kdik(:, bf1, bf3) + D(j, k) * teintd(:)
                          end if
                          if (j .lt. k .or. .not. yesDSym) then
                            !K(k,j)=K(k,j)+D(i,k)*teint
                            !Kdloc(:, j, k) = Kdloc(:, j, k) + D(i, k)*teintd(:)
                            Kdjk(:, bf2, bf3) = Kdjk(:, bf2, bf3) + D(i, k) * teintd(:)
                          end if
                          if (j .gt. k .or. .not. yesDSym) then
                            !K(j,k)=K(j,k)+D(i,k)*teint
                            !Kdloc(:, k, j) = Kdloc(:, k, j) + D(k, i)*teintd(:)
                            Kdkj(:, bf3, bf2) = Kdkj(:, bf3, bf2) + D(k, i) * teintd(:)
                          end if
                        else if (i .eq. k) then ! (ij|il) (ji|il) (ij|li) (ji|li) (il|ij) (il|ji) (li|ij) (li|ji)
                          !J(i,j)=J(i,j)+D(i,l)*teint
                          !J(i,j)=J(i,j)+D(l,i)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + Dsym(i, l) * teintd(:)
                          !K(i,j)=K(i,j)+D(l,i)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(i, l)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(i, l) * teintd(:)
                          !J(i,l)=J(i,l)+D(i,j)*teint
                          !J(i,l)=J(i,l)+D(j,i)*teint
                          Jdli(:, bf4, bf1) = Jdli(:, bf4, bf1) + Dsym(i, j) * teintd(:)
                          !K(i,l)=K(i,l)+D(j,i)*teint
                          !Kdloc(:, l, i) = Kdloc(:, l, i) + D(i, j)*teintd(:)
                          Kdli(:, bf4, bf1) = Kdli(:, bf4, bf1) + D(i, j) * teintd(:)
                          !K(i,i)=K(i,i)+D(j,l)*teint
                          !K(i,i)=K(i,i)+D(l,j)*teint
                          !Kdloc(:, i, i) = Kdloc(:, i, i) + Dsym(j, l)*teintd(:)
                          Kdii(:, bf1) = Kdii(:, bf1) + Dsym(j, l) * teintd(:)
                          !K(j,l)=K(j,l)+D(i,i)*teint
                          !Kdloc(:, l, j) = Kdloc(:, l, j) + D(i, i)*teintd(:)
                          Kdlj(:, bf4, bf2) = Kdlj(:, bf4, bf2) + D(i, i) * teintd(:)
                          if (.not. yesDSym) then
                            !K(j,i)=K(j,i)+D(i,l)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(l, i)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(l, i) * teintd(:)
                            !K(l,i)=K(l,i)+D(i,j)*teint
                            !Kdloc(:, i, l) = Kdloc(:, i, l) + D(j, i)*teintd(:)
                            Kdil(:, bf1, bf4) = Kdil(:, bf1, bf4) + D(j, i) * teintd(:)
                            !K(l,j)=K(l,j)+D(i,i)*teint
                            !Kdloc(:, j, l) = Kdloc(:, j, l) + D(i, i)*teintd(:)
                            Kdjl(:, bf2, bf4) = Kdjl(:, bf2, bf4) + D(i, i) * teintd(:)
                          end if
                        else if (j .eq. l) then ! (ij|kj) (ji|kj) (ij|jk) (ji|jk) (kj|ij) (kj|ji) (jk|ij) (jk|ji)
                          !J(i,j)=J(i,j)+D(k,j)*teint
                          !J(i,j)=J(i,j)+D(j,k)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + Dsym(k, j) * teintd(:)
                          !K(i,j)=K(i,j)+D(j,k)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(k, j)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(k, j) * teintd(:)
                          !J(k,j)=J(k,j)+D(i,j)*teint
                          !J(k,j)=J(k,j)+D(j,i)*teint
                          Jdjk(:, bf2, bf3) = Jdjk(:, bf2, bf3) + Dsym(i, j)*teintd(:)
                          !K(k,j)=K(k,j)+D(j,i)*teint
                          !Kdloc(:, j, k) = Kdloc(:, j, k) + D(i, j)*teintd(:)
                          Kdjk(:, bf2, bf3) = Kdjk(:, bf2, bf3) + D(i, j) * teintd(:)
                          !K(i,k)=K(i,k)+D(j,j)*teint
                          !Kdloc(:, k, i) = Kdloc(:, k, i) + D(j, j)*teintd(:)
                          Kdki(:, bf3, bf1) = Kdki(:, bf3, bf1) + D(j, j) * teintd(:)
                          !K(j,j)=K(j,j)+D(i,k)*teint
                          !K(j,j)=K(j,j)+D(i,k)*teint
                          !Kdloc(:, j, j) = Kdloc(:, j, j) + Dsym(k, i)*teintd(:)
                          Kdjj(:, bf2) = Kdjj(:, bf2) + Dsym(k, i) * teintd(:)
                          if (.not. yesDSym) then
                            !K(k,i)=K(k,i)+D(j,j)*teint
                            !Kdloc(:, i, k) = Kdloc(:, i, k) + D(j, j)*teintd(:)
                            Kdik(:, bf1, bf3) = Kdik(:, bf1, bf3) + D(j, j) * teintd(:)
                            !K(j,k)=K(j,k)+D(i,j)*teint
                            !Kdloc(:, k, j) = Kdloc(:, k, j) + D(j, i)*teintd(:)
                            Kdkj(:, bf3, bf2) = Kdkj(:, bf3, bf2) + D(j, i) * teintd(:)
                            !K(j,i)=K(j,i)+D(k,j)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(j, k)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(j, k)*teintd(:)
                          end if
                        else if (j .eq. k) then ! (ij|jl) (ji|jl) (ij|lj) (ji|lj) (jl|ij) (jl|ji) (lj|ij) (lj|ji)
                          !J(i,j)=J(i,j)+D(l,j)*teint
                          !J(i,j)=J(i,j)+D(j,l)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + Dsym(j, l) * teintd(:)
                          !K(i,j)=K(i,j)+D(j,l)*teint
                          !Kdloc(:, j, i) = Kdloc(:, j, i) + D(l, j)*teintd(:)
                          Kdji(:, bf2, bf1) = Kdji(:, bf2, bf1) + D(l, j) * teintd(:)
                          !J(j,l)=J(j,l)+D(i,j)*teint
                          !J(j,l)=J(j,l)+D(j,i)*teint
                          Jdlj(:, bf4, bf2) = Jdlj(:, bf4, bf2) + Dsym(i, j) * teintd(:)
                          !K(j,l)=K(j,l)+D(i,j)*teint
                          !Kdloc(:, l, j) = Kdloc(:, l, j) + D(j, i)*teintd(:)
                          Kdlj(:, bf4, bf2) = Kdlj(:, bf4, bf2) + D(j, i) * teintd(:)
                          !K(j,j)=K(j,j)+D(i,l)*teint
                          !K(j,j)=K(j,j)+D(l,i)*teint
                          !Kdloc(:, j, j) = Kdloc(:, j, j) + Dsym(i, l)*teintd(:)
                          Kdjj(:, bf2) = Kdjj(:, bf2) + Dsym(i, l) * teintd(:)
                          !K(i,l)=K(i,l)+D(j,j)*teint
                          !Kdloc(:, l, i) = Kdloc(:, l, i) + D(j, j)*teintd(:)
                          Kdli(:, bf4, bf1) = Kdli(:, bf4, bf1) + D(j, j) * teintd(:)
                          if (.not. yesDSym) then
                            !K(j,i)=K(j,i)+D(l,j)*teint
                            !Kdloc(:, i, j) = Kdloc(:, i, j) + D(j, l)*teintd(:)
                            Kdij(:, bf1, bf2) = Kdij(:, bf1, bf2) + D(j, l) * teintd(:)
                            !K(l,j)=K(l,j)+D(j,i)*teint
                            !Kdloc(:, j, l) = Kdloc(:, j, l) + D(i, j)*teintd(:)
                            Kdjl(:, bf2, bf4) = Kdjl(:, bf2, bf4) + D(i, j) * teintd(:)
                            !K(l,i)=K(l,i)+D(j,j)*teint
                            !Kdloc(:, i, l) = Kdloc(:, i, l) + D(j, j)*teintd(:)
                            Kdil(:, bf1, bf4) = Kdil(:, bf1, bf4) + D(j, j) * teintd(:)
                          end if
                          ! no indices equal
                        else ! (ij|kl) (ji|kl) (ij|lk) (ji|lk) (kl|ij) (kl|ji) (lk|ij) (lk|ji)
                          !J(i,j)=J(i,j)+D(k,l)*teint
                          !J(i,j)=J(i,j)+D(l,k)*teint
                          Jdji(:, bf2, bf1) = Jdji(:, bf2, bf1) + Dsym(k, l) * teintd(:)
                          !J(k,l)=J(k,l)+D(j,i)*teint
                          !J(k,l)=J(k,l)+D(i,j)*teint
                          Jdlk(:, bf4, bf3) = Jdlk(:, bf4, bf3) + Dsym(i, j) * teintd(:)
                          !K(i,k)=K(i,k)+D(j,l)*teint
                          !Kdloc(:, k, i) = Kdloc(:, k, i) + D(l, j)*teintd(:)
                          Kdki(:, bf3, bf1) = Kdki(:, bf3, bf1) + D(l, j) * teintd(:)
                          !K(i,l)=K(i,l)+D(j,k)*teint
                          !Kdloc(:, l, i) = Kdloc(:, l, i) + D(k, j)*teintd(:)
                          Kdli(:, bf4, bf1) = Kdli(:, bf4, bf1) + D(k, j) * teintd(:)
                          if (.not. yesDSym) then
                            !K(k,i)=K(k,i)+D(l,j)*teint
                            !Kdloc(:, i, k) = Kdloc(:, i, k) + D(j, l)*teintd(:)
                            Kdik(:, bf1, bf3) = Kdik(:, bf1, bf3) + D(j, l) * teintd(:)
                            !K(l,i)=K(l,i)+D(k,j)*teint
                            !Kdloc(:, i, l) = Kdloc(:, i, l) + D(j, k)*teintd(:)
                            Kdil(:, bf1, bf4) = Kdil(:, bf1, bf4) + D(j, k) * teintd(:)
                          end if
                          if (j .gt. k .or. .not. yesDSym) then
                            !K(j,k)=K(j,k)+D(i,l)*teint
                            !Kdloc(:, k, j) = Kdloc(:, k, j) + D(l, i)*teintd(:)
                            Kdkj(:, bf3, bf2) = Kdkj(:, bf3, bf2) + D(l, i) * teintd(:)
                          end if
                          if (j .lt. k .or. .not. yesDSym) then
                            !K(k,j)=K(k,j)+D(l,i)*teint
                            !Kdloc(:, j, k) = Kdloc(:, j, k) + D(i, l)*teintd(:)
                            Kdjk(:, bf2, bf3) = Kdjk(:, bf2, bf3) + D(i, l) * teintd(:)
                          end if
                          if (j .gt. l .or. .not. yesDSym) then
                            !K(j,l)=K(j,l)+D(i,k)*teint
                            !Kdloc(:, l, j) = Kdloc(:, l, j) + D(k, i)*teintd(:)
                            Kdlj(:, bf4, bf2) = Kdlj(:, bf4, bf2) + D(k, i) * teintd(:)
                          end if
                          if (j .lt. l .or. .not. yesDSym) then
                            !K(l,j)=K(l,j)+D(k,i)*teint
                            !Kdloc(:, j, l) = Kdloc(:, j, l) + D(i, k)*teintd(:)
                            Kdjl(:, bf2, bf4) = Kdjl(:, bf2, bf4) + D(i, k) * teintd(:)
                          end if
                        end if
                      end do! l
                    end do! k
                  end do! j
                end do! i
                ! do update on those that depend on innermost l loop
                SETLOCK
                !Jdlk
                Jd2(a1i:a1e, lmin:lmax, kmin:kmax) = Jd2(a1i:a1e, lmin:lmax, kmin:kmax) + Jdlk(1:3, :, :)
                Jd2(a2i:a2e, lmin:lmax, kmin:kmax) = Jd2(a2i:a2e, lmin:lmax, kmin:kmax) + Jdlk(4:6, :, :)
                Jd2(a3i:a3e, lmin:lmax, kmin:kmax) = Jd2(a3i:a3e, lmin:lmax, kmin:kmax) + Jdlk(7:9, :, :)
                Jd2(a4i:a4e, lmin:lmax, kmin:kmax) = Jd2(a4i:a4e, lmin:lmax, kmin:kmax) + Jdlk(10:12, :, :)
                !Jdlj
                Jd2(a1i:a1e, lmin:lmax, jmin:jmax) = Jd2(a1i:a1e, lmin:lmax, jmin:jmax) + Jdlj(1:3, :, :)
                Jd2(a2i:a2e, lmin:lmax, jmin:jmax) = Jd2(a2i:a2e, lmin:lmax, jmin:jmax) + Jdlj(4:6, :, :)
                Jd2(a3i:a3e, lmin:lmax, jmin:jmax) = Jd2(a3i:a3e, lmin:lmax, jmin:jmax) + Jdlj(7:9, :, :)
                Jd2(a4i:a4e, lmin:lmax, jmin:jmax) = Jd2(a4i:a4e, lmin:lmax, jmin:jmax) + Jdlj(10:12, :, :)
                !Jdli
                Jd2(a1i:a1e, lmin:lmax, imin:imax) = Jd2(a1i:a1e, lmin:lmax, imin:imax) + Jdli(1:3, :, :)
                Jd2(a2i:a2e, lmin:lmax, imin:imax) = Jd2(a2i:a2e, lmin:lmax, imin:imax) + Jdli(4:6, :, :)
                Jd2(a3i:a3e, lmin:lmax, imin:imax) = Jd2(a3i:a3e, lmin:lmax, imin:imax) + Jdli(7:9, :, :)
                Jd2(a4i:a4e, lmin:lmax, imin:imax) = Jd2(a4i:a4e, lmin:lmax, imin:imax) + Jdli(10:12, :, :)
                if(.not. yesDSym)then
                  !Kdil
                  Kd2(a1i:a1e, imin:imax, lmin:lmax) = Kd2(a1i:a1e, imin:imax, lmin:lmax) + Kdil(1:3, :, :)
                  Kd2(a2i:a2e, imin:imax, lmin:lmax) = Kd2(a2i:a2e, imin:imax, lmin:lmax) + Kdil(4:6, :, :)
                  Kd2(a3i:a3e, imin:imax, lmin:lmax) = Kd2(a3i:a3e, imin:imax, lmin:lmax) + Kdil(7:9, :, :)
                  Kd2(a4i:a4e, imin:imax, lmin:lmax) = Kd2(a4i:a4e, imin:imax, lmin:lmax) + Kdil(10:12, :, :)
                end if
                !Kdli
                Kd2(a1i:a1e, lmin:lmax, imin:imax) = Kd2(a1i:a1e, lmin:lmax, imin:imax) + Kdli(1:3, :, :)
                Kd2(a2i:a2e, lmin:lmax, imin:imax) = Kd2(a2i:a2e, lmin:lmax, imin:imax) + Kdli(4:6, :, :)
                Kd2(a3i:a3e, lmin:lmax, imin:imax) = Kd2(a3i:a3e, lmin:lmax, imin:imax) + Kdli(7:9, :, :)
                Kd2(a4i:a4e, lmin:lmax, imin:imax) = Kd2(a4i:a4e, lmin:lmax, imin:imax) + Kdli(10:12, :, :)
                if (lmin .le. jmax .or. .not. yesDSym) then
                  !Kdlj
                  Kd2(a1i:a1e, lmin:lmax, jmin:jmax) = Kd2(a1i:a1e, lmin:lmax, jmin:jmax) + Kdlj(1:3, :, :)
                  Kd2(a2i:a2e, lmin:lmax, jmin:jmax) = Kd2(a2i:a2e, lmin:lmax, jmin:jmax) + Kdlj(4:6, :, :)
                  Kd2(a3i:a3e, lmin:lmax, jmin:jmax) = Kd2(a3i:a3e, lmin:lmax, jmin:jmax) + Kdlj(7:9, :, :)
                  Kd2(a4i:a4e, lmin:lmax, jmin:jmax) = Kd2(a4i:a4e, lmin:lmax, jmin:jmax) + Kdlj(10:12, :, :)
                end if
                if (jmin .le. lmax .or. .not. yesDSym) then
                  !Kdjl
                  Kd2(a1i:a1e, jmin:jmax, lmin:lmax) = Kd2(a1i:a1e, jmin:jmax, lmin:lmax) + Kdjl(1:3, :, :)
                  Kd2(a2i:a2e, jmin:jmax, lmin:lmax) = Kd2(a2i:a2e, jmin:jmax, lmin:lmax) + Kdjl(4:6, :, :)
                  Kd2(a3i:a3e, jmin:jmax, lmin:lmax) = Kd2(a3i:a3e, jmin:jmax, lmin:lmax) + Kdjl(7:9, :, :)
                  Kd2(a4i:a4e, jmin:jmax, lmin:lmax) = Kd2(a4i:a4e, jmin:jmax, lmin:lmax) + Kdjl(10:12, :, :)
                end if
                UNSETLOCK
                Jdlk(:,:,:) = ZERO
                Jdlj(:,:,:) = ZERO                
                Kdil(:,:,:) = ZERO         
                Kdli(:,:,:) = ZERO
                Kdlj(:,:,:) = ZERO
                Kdjl(:,:,:) = ZERO
              end if! res/=0
              deallocate (data4)
              deallocate (data3)
              deallocate (data2)
              deallocate (data1)
              deallocate (Kdlj)
              deallocate (Kdjl)
              deallocate (Kdli)
              deallocate (Kdil)
              deallocate (Jdlk)
              deallocate (Jdlj)
              deallocate (Jdli)
            end if! SkipShell
            ! if either shell loop 4 has ended OR the next shell 4 is on a different atom
            ! AND we have at least once added some values to any Kdxy Jdxy arrays for the current atoms
            ! do an update on the global array Jd2
            if( (sh3==sh4 .or. bas(1, min(sh4+1, sh3)) /= bas(1,sh4)) .and. needsUpdate) then
              SETLOCK
              !Jdii
              forall (bf1=1:n1)
                Jd2(a1i:a1e, imin+bf1-1, imin+bf1-1) = Jd2(a1i:a1e, imin+bf1-1, imin+bf1-1) + Jdii(1:3, bf1)
                Jd2(a2i:a2e, imin+bf1-1, imin+bf1-1) = Jd2(a2i:a2e, imin+bf1-1, imin+bf1-1) + Jdii(4:6, bf1)
                Jd2(a3i:a3e, imin+bf1-1, imin+bf1-1) = Jd2(a3i:a3e, imin+bf1-1, imin+bf1-1) + Jdii(7:9, bf1)
                Jd2(a4i:a4e, imin+bf1-1, imin+bf1-1) = Jd2(a4i:a4e, imin+bf1-1, imin+bf1-1) + Jdii(10:12, bf1)
              end forall
              !Jdjj
              forall (bf2=1:n2)
                Jd2(a1i:a1e, jmin+bf2-1, jmin+bf2-1) = Jd2(a1i:a1e, jmin+bf2-1, jmin+bf2-1) + Jdjj(1:3, bf2)
                Jd2(a2i:a2e, jmin+bf2-1, jmin+bf2-1) = Jd2(a2i:a2e, jmin+bf2-1, jmin+bf2-1) + Jdjj(4:6, bf2)
                Jd2(a3i:a3e, jmin+bf2-1, jmin+bf2-1) = Jd2(a3i:a3e, jmin+bf2-1, jmin+bf2-1) + Jdjj(7:9, bf2)
                Jd2(a4i:a4e, jmin+bf2-1, jmin+bf2-1) = Jd2(a4i:a4e, jmin+bf2-1, jmin+bf2-1) + Jdjj(10:12, bf2)
              end forall
              !Jdkk
              forall (bf3=1:n3)
                Jd2(a1i:a1e, kmin+bf3-1, kmin+bf3-1) = Jd2(a1i:a1e, kmin+bf3-1, kmin+bf3-1) + Jdkk(1:3, bf3)
                Jd2(a2i:a2e, kmin+bf3-1, kmin+bf3-1) = Jd2(a2i:a2e, kmin+bf3-1, kmin+bf3-1) + Jdkk(4:6, bf3)
                Jd2(a3i:a3e, kmin+bf3-1, kmin+bf3-1) = Jd2(a3i:a3e, kmin+bf3-1, kmin+bf3-1) + Jdkk(7:9, bf3)
                Jd2(a4i:a4e, kmin+bf3-1, kmin+bf3-1) = Jd2(a4i:a4e, kmin+bf3-1, kmin+bf3-1) + Jdkk(10:12, bf3)
              end forall
              !Jdji
              Jd2(a1i:a1e, jmin:jmax, imin:imax) = Jd2(a1i:a1e, jmin:jmax, imin:imax) + Jdji(1:3, :, :)
              Jd2(a2i:a2e, jmin:jmax, imin:imax) = Jd2(a2i:a2e, jmin:jmax, imin:imax) + Jdji(4:6, :, :)
              Jd2(a3i:a3e, jmin:jmax, imin:imax) = Jd2(a3i:a3e, jmin:jmax, imin:imax) + Jdji(7:9, :, :)
              Jd2(a4i:a4e, jmin:jmax, imin:imax) = Jd2(a4i:a4e, jmin:jmax, imin:imax) + Jdji(10:12, :, :)
              !Jdjk
              Jd2(a1i:a1e, jmin:jmax, kmin:kmax) = Jd2(a1i:a1e, jmin:jmax, kmin:kmax) + Jdjk(1:3, :, :)
              Jd2(a2i:a2e, jmin:jmax, kmin:kmax) = Jd2(a2i:a2e, jmin:jmax, kmin:kmax) + Jdjk(4:6, :, :)
              Jd2(a3i:a3e, jmin:jmax, kmin:kmax) = Jd2(a3i:a3e, jmin:jmax, kmin:kmax) + Jdjk(7:9, :, :)
              Jd2(a4i:a4e, jmin:jmax, kmin:kmax) = Jd2(a4i:a4e, jmin:jmax, kmin:kmax) + Jdjk(10:12, :, :)
              ! Kdii
              do bf1=1,n1
                Kd2(a1i:a1e, imin+bf1-1, imin+bf1-1) = Kd2(a1i:a1e, imin+bf1-1, imin+bf1-1) + Kdii(1:3, bf1)
                Kd2(a2i:a2e, imin+bf1-1, imin+bf1-1) = Kd2(a2i:a2e, imin+bf1-1, imin+bf1-1) + Kdii(4:6, bf1)
                Kd2(a3i:a3e, imin+bf1-1, imin+bf1-1) = Kd2(a3i:a3e, imin+bf1-1, imin+bf1-1) + Kdii(7:9, bf1)
                Kd2(a4i:a4e, imin+bf1-1, imin+bf1-1) = Kd2(a4i:a4e, imin+bf1-1, imin+bf1-1) + Kdii(10:12, bf1)
              end do
              ! Kdjj
              forall (bf2=1:n2)
                Kd2(a1i:a1e, jmin+bf2-1, jmin+bf2-1) = Kd2(a1i:a1e, jmin+bf2-1, jmin+bf2-1) + Kdjj(1:3, bf2)
                Kd2(a2i:a2e, jmin+bf2-1, jmin+bf2-1) = Kd2(a2i:a2e, jmin+bf2-1, jmin+bf2-1) + Kdjj(4:6, bf2)
                Kd2(a3i:a3e, jmin+bf2-1, jmin+bf2-1) = Kd2(a3i:a3e, jmin+bf2-1, jmin+bf2-1) + Kdjj(7:9, bf2)
                Kd2(a4i:a4e, jmin+bf2-1, jmin+bf2-1) = Kd2(a4i:a4e, jmin+bf2-1, jmin+bf2-1) + Kdjj(10:12, bf2)
              end forall
              !Kdki
              Kd2(a1i:a1e, kmin:kmax, imin:imax) = Kd2(a1i:a1e, kmin:kmax, imin:imax) + Kdki(1:3, :, :)
              Kd2(a2i:a2e, kmin:kmax, imin:imax) = Kd2(a2i:a2e, kmin:kmax, imin:imax) + Kdki(4:6, :, :)
              Kd2(a3i:a3e, kmin:kmax, imin:imax) = Kd2(a3i:a3e, kmin:kmax, imin:imax) + Kdki(7:9, :, :)
              Kd2(a4i:a4e, kmin:kmax, imin:imax) = Kd2(a4i:a4e, kmin:kmax, imin:imax) + Kdki(10:12, :, :)
              !Kdji
              Kd2(a1i:a1e, jmin:jmax, imin:imax) = Kd2(a1i:a1e, jmin:jmax, imin:imax) + Kdji(1:3, :, :)
              Kd2(a2i:a2e, jmin:jmax, imin:imax) = Kd2(a2i:a2e, jmin:jmax, imin:imax) + Kdji(4:6, :, :)
              Kd2(a3i:a3e, jmin:jmax, imin:imax) = Kd2(a3i:a3e, jmin:jmax, imin:imax) + Kdji(7:9, :, :)
              Kd2(a4i:a4e, jmin:jmax, imin:imax) = Kd2(a4i:a4e, jmin:jmax, imin:imax) + Kdji(10:12, :, :)
              if(.not. yesDSym)then
                !Kdij
                Kd2(a1i:a1e, imin:imax, jmin:jmax) = Kd2(a1i:a1e, imin:imax, jmin:jmax) + Kdij(1:3, :, :)
                Kd2(a2i:a2e, imin:imax, jmin:jmax) = Kd2(a2i:a2e, imin:imax, jmin:jmax) + Kdij(4:6, :, :)
                Kd2(a3i:a3e, imin:imax, jmin:jmax) = Kd2(a3i:a3e, imin:imax, jmin:jmax) + Kdij(7:9, :, :)
                Kd2(a4i:a4e, imin:imax, jmin:jmax) = Kd2(a4i:a4e, imin:imax, jmin:jmax) + Kdij(10:12, :, :)
                !Kdik
                Kd2(a1i:a1e, imin:imax, kmin:kmax) = Kd2(a1i:a1e, imin:imax, kmin:kmax) + Kdik(1:3, :, :)
                Kd2(a2i:a2e, imin:imax, kmin:kmax) = Kd2(a2i:a2e, imin:imax, kmin:kmax) + Kdik(4:6, :, :)
                Kd2(a3i:a3e, imin:imax, kmin:kmax) = Kd2(a3i:a3e, imin:imax, kmin:kmax) + Kdik(7:9, :, :)
                Kd2(a4i:a4e, imin:imax, kmin:kmax) = Kd2(a4i:a4e, imin:imax, kmin:kmax) + Kdik(10:12, :, :)
              end if
              if (jmin .le. kmax .or. .not. yesDSym) then
                !Kdjk
                Kd2(a1i:a1e, jmin:jmax, kmin:kmax) = Kd2(a1i:a1e, jmin:jmax, kmin:kmax) + Kdjk(1:3, :, :)
                Kd2(a2i:a2e, jmin:jmax, kmin:kmax) = Kd2(a2i:a2e, jmin:jmax, kmin:kmax) + Kdjk(4:6, :, :)
                Kd2(a3i:a3e, jmin:jmax, kmin:kmax) = Kd2(a3i:a3e, jmin:jmax, kmin:kmax) + Kdjk(7:9, :, :)
                Kd2(a4i:a4e, jmin:jmax, kmin:kmax) = Kd2(a4i:a4e, jmin:jmax, kmin:kmax) + Kdjk(10:12, :, :)
              end if
              if (kmin .le. jmax .or. .not. yesDSym) then
                !Kdkj
                Kd2(a1i:a1e, kmin:kmax, jmin:jmax) = Kd2(a1i:a1e, kmin:kmax, jmin:jmax) + Kdkj(1:3, :, :)
                Kd2(a2i:a2e, kmin:kmax, jmin:jmax) = Kd2(a2i:a2e, kmin:kmax, jmin:jmax) + Kdkj(4:6, :, :)
                Kd2(a3i:a3e, kmin:kmax, jmin:jmax) = Kd2(a3i:a3e, kmin:kmax, jmin:jmax) + Kdkj(7:9, :, :)
                Kd2(a4i:a4e, kmin:kmax, jmin:jmax) = Kd2(a4i:a4e, kmin:kmax, jmin:jmax) + Kdkj(10:12, :, :)
              end if
              UNSETLOCK
              Jdii(:,:) = ZERO
              Jdjj(:,:) = ZERO
              Jdkk(:,:) = ZERO
              Jdji(:,:,:) = ZERO
              Jdjk(:,:,:) = ZERO
              Kdii(:,:) = ZERO
              Kdjj(:,:) = ZERO
              Kdki(:,:,:) = ZERO
              if(.not. yesDSym)then
                Kdij(:,:,:) = ZERO
                Kdik(:,:,:) = ZERO
              end if
              Kdji(:,:,:) = ZERO
              if (jmin .le. kmax .or. .not. yesDSym) &
                   Kdjk(:,:,:) = ZERO
              if (kmin .le. jmax .or. .not. yesDSym) &
                   Kdkj(:,:,:) = ZERO
              needsUpdate = .False.
            end if
          end do ! sh4             
          deallocate (Jdjk)
          deallocate (Kdkj)
          deallocate (Kdjk)
          deallocate (Kdki)
          deallocate (Kdik)
          deallocate (Jdkk)
        end do! sh3
        deallocate (Kdji)
        deallocate (Kdij)
        deallocate (Kdjj)
        deallocate (Jdji)
        deallocate (Jdjj)
      end do! sh2
      deallocate (Kdii)
      deallocate (Jdii)
    end do! sh1
    !$omp end do
    call CINTdel_optimizer(optip)
    !$omp end parallel
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif       
    do i = 1, size(Jd, 1)
      do j = 1, i
        Jd2(:, i, j) = Jd2(:, j, i)
        if (yesDSym) &
          Kd2(:, i, j) = Kd2(:, j, i)
      end do
    end do

    Jd = reshape(Jd2, shape=shape(Jd), order=(/3, 1, 2/))
    Kd = reshape(Kd2, shape=shape(Jd), order=(/3, 1, 2/))
#ifdef REORDERP
  contains
    function reorder_data_p(dat, p_index) result(r_data)
      integer, intent(in) :: p_index
      real(long), intent(in) :: dat(:, :, :, :, :)
      real(long) :: r_data(size(dat, 1), size(dat, 2), size(dat, 3), size(dat, 4), size(dat, 5))
      if (size(dat, p_index) .ne. 3) stop "error in reorder_data_p"
      if (p_index .eq. 1) then
        r_data(1, :, :, :, :) = dat(2, :, :, :, :)
        r_data(2, :, :, :, :) = dat(3, :, :, :, :)
        r_data(3, :, :, :, :) = dat(1, :, :, :, :)
      else if (p_index .eq. 2) then
        r_data(:, 1, :, :, :) = dat(:, 2, :, :, :)
        r_data(:, 2, :, :, :) = dat(:, 3, :, :, :)
        r_data(:, 3, :, :, :) = dat(:, 1, :, :, :)
      else if (p_index .eq. 3) then
        r_data(:, :, 1, :, :) = dat(:, :, 2, :, :)
        r_data(:, :, 2, :, :) = dat(:, :, 3, :, :)
        r_data(:, :, 3, :, :) = dat(:, :, 1, :, :)
      else if (p_index .eq. 4) then
        r_data(:, :, :, 1, :) = dat(:, :, :, 2, :)
        r_data(:, :, :, 2, :) = dat(:, :, :, 3, :)
        r_data(:, :, :, 3, :) = dat(:, :, :, 1, :)
      end if
    end function reorder_data_p
#endif
  end subroutine jk_gradient

  subroutine h0_gradient(env, atm, bas, first_bf_of_shell, h_grad)
    !! computes gradient of h0
    integer, external :: CINT1e_ipkin_sph
    integer, external :: CINT1e_ipnuc_sph
    integer, external :: CINT1e_iprinv_sph
    integer, external :: CINT1e_nuc_sph
    integer, external :: CINTcgto_spheric
    real(long), intent(in) :: env(:)
    integer, intent(in) :: atm(:, :)
    integer, intent(in) :: bas(:, :)
    integer, intent(in) :: first_bf_of_shell(:)
    real(long), intent(out) :: h_grad(:, :, :)
    integer :: ret1
    integer :: i, j, k
    integer :: mu
    integer :: atom1, atom2, at
    integer :: sh1, sh2
    integer :: n1, n2
    integer :: m1, m2
    integer :: shls(2)
    real(long) :: factor
    real(long), allocatable :: my_env(:)
    real(long), allocatable :: data1(:, :, :)
    real(long) :: v(size(h_grad, 1), size(h_grad, 1), 3*size(atm, 2))
    real(long) :: t(size(h_grad, 1), size(h_grad, 1), 3*size(atm, 2))
    integer(8) :: opt
    call cintinit_optimizer(opt, atm, size(atm, 2), bas, size(bas, 2), env)
    t = ZERO
    v = ZERO
    h_grad = ZERO

    h_grad(:, :, :) = ZERO
    ! loop over shell pairs
    ! no race condition, because different sh1 implies different matrix elements in v and t
    ! not true: for pairs of shells sh1,sh2 == sh2,sh1 if shells are on the same atom, there is a race condition
    ! todo: make 3 parallelized loops for sh1>sh2, sh2>sh1, and sh1==sh2
    !!$omp parallel shared(atm, bas, env, first_bf_of_shell) private(sh1,sh2,atom,atom1,atom2,shls,ret1,n1,n2,data1,m1,m2,i,j,factor,my_env,mu) default(none) reduction(+:v,t)
    !!$omp do
    do sh1 = 1, size(bas, 2) ! loop over shells
      do sh2 = 1, size(bas, 2) ! loop over shells
        n1 = 2*bas(2, sh1) + 1
        atom1 = bas(1, sh1) + 1
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        allocate (data1(n1, n2, 3))
        data1 = ZERO
        ! compute grad <  sh1 | h_kin | sh2 > and
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        ret1 = cint1e_ipkin_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (ret1 .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data1 = reorder_data_p(data1, 1)
          end if
          if (bas(2, sh2) .eq. 1) then
            data1 = reorder_data_p(data1, 2)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              if (mod(m1 + m2, 2) .eq. 0) then
                factor = ONE
              else
                factor = -ONE
              end if
              do mu = 1, 3 ! mu=x,y,z
                t(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) = &
                  t(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) + factor*(-data1(i, j, mu))
                t(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) = &
                  t(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) + factor*(-data1(i, j, mu))
              end do
            end do ! j=1,n2
          end do! i=1,n1
        end if! if ret1/=0 and ret2/=0
        ! compute grad (<  sh1 | V_nuc | sh2 > )
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        ret1 = cint1e_ipnuc_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), env, opt)
        if (ret1 .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
          if (bas(2, sh1) .eq. 1) then
            data1 = reorder_data_p(data1, 1)
          end if
          if (bas(2, sh2) .eq. 1) then
            data1 = reorder_data_p(data1, 2)
          end if
#endif
          do i = 1, n1
            m1 = i - bas(2, sh1) - 1
            do j = 1, n2
              m2 = j - bas(2, sh2) - 1
              if (mod(m1 + m2, 2) .eq. 0) then
                factor = ONE
              else
                factor = -ONE
              end if
              do mu = 1, 3 ! mu=x,y,z              !
                v(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) = &
                  v(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (atom1 - 1)*3 + mu) + factor*(-data1(i, j, mu))
                v(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) = &
                  v(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (atom1 - 1)*3 + mu) + factor*(-data1(i, j, mu))
              end do
            end do ! j=1,n2
          end do! i=1,n1
        end if! if i/=0 and j/=0
        deallocate (data1)
      end do! sh2
    end do! sh1
    !!$omp end do
    !!$omp end parallel

    ! this loop is not so trivial parallelizable
    allocate (my_env(size(env, 1)))
    do sh1 = 1, size(bas, 2) ! loop over shells
      do sh2 = 1, size(bas, 2) ! loop over shells
        n1 = 2*bas(2, sh1) + 1
        atom1 = bas(1, sh1) + 1
        n2 = 2*bas(2, sh2) + 1
        atom2 = bas(1, sh2) + 1
        allocate (data1(n1, n2, 3))
        data1 = ZERO
        shls(1) = sh1 - 1
        shls(2) = sh2 - 1
        !  compute < sh1 | grad V_nuc | sh2 >
        my_env(:) = env(:)
        do at = 1, size(atm, 2)
          ! calculate < sh1 | grad Z/|r-R_atom| | sh2 >
          ! Rinv=Ratom is specified via my_env(5:7)
          my_env(5) = env(atm(2, at) + 1) !x
          my_env(6) = env(atm(2, at) + 2) !y
          my_env(7) = env(atm(2, at) + 3) !z
          if (atm(4, at) .gt. ZERO) then
            my_env(8) = env(atm(4, at) + 1)
          else
            my_env(8) = -ONE
          end if
          !        ret1=cint1e_nablarinv_sph(data1, shls, atm, size(atm,2), bas, size(bas,2), my_env)
          ret1 = cint1e_iprinv_sph(data1, shls, atm, size(atm, 2), bas, size(bas, 2), my_env, opt)
          if (ret1 .ne. 0) then! if return value is zero integrals are zero by symmetry
#ifdef REORDERP
            if (bas(2, sh1) .eq. 1) then
              data1 = reorder_data_p(data1, 1)
            end if
            if (bas(2, sh2) .eq. 1) then
              data1 = reorder_data_p(data1, 2)
            end if
#endif
            do i = 1, n1
              m1 = i - bas(2, sh1) - 1
              do j = 1, n2
                m2 = j - bas(2, sh2) - 1
                if (mod(m1 + m2, 2) .eq. 0) then
                  factor = ONE
                else
                  factor = -ONE
                end if
                if (atm(3, at) .eq. 3 .or. atm(3, at) .eq. 4) then ! partial charges
                  factor = factor*env(atm(5, at) + 1)
                else
                  factor = factor*dble(atm(1, at))
                end if
                do mu = 1, 3 ! mu=x,y,z
                  v(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (at - 1)*3 + mu) = &
                    v(first_bf_of_shell(sh1) + i - 1, first_bf_of_shell(sh2) + j - 1, (at - 1)*3 + mu) + factor*(-data1(i, j, mu))
                  v(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (at - 1)*3 + mu) = &
                    v(first_bf_of_shell(sh2) + j - 1, first_bf_of_shell(sh1) + i - 1, (at - 1)*3 + mu) + factor*(-data1(i, j, mu))
                end do
              end do ! j=1,n2
            end do! i=1,n1
          end if! if ret/=0
        end do! atom
        deallocate (data1)
      end do! sh2
    end do! sh1
    deallocate (my_env)

    do k = 1, size(atm, 2)*3
      do i = 1, size(v, 1)
        do j = 1, size(v, 2)
          h_grad(i, j, k) = t(i, j, k) + v(i, j, k)
        end do
      end do
    end do
#ifdef REORDERP
  contains
    function reorder_data_p(data, p_index) result(r_data)
      integer, intent(in) :: p_index
      real(long), intent(in) :: data(:, :, :)
      real(long) :: r_data(size(data, 1), size(data, 2), size(data, 3))
      if (size(data, p_index) .ne. 3) stop "error in reorder_data_p"
      if (p_index .eq. 1) then
        r_data(1, :, :) = data(2, :, :)
        r_data(2, :, :) = data(3, :, :)
        r_data(3, :, :) = data(1, :, :)
      else if (p_index .eq. 2) then
        r_data(:, 1, :) = data(:, 2, :)
        r_data(:, 2, :) = data(:, 3, :)
        r_data(:, 3, :) = data(:, 1, :)
      end if
    end function reorder_data_p
#endif

    call CINTdel_optimizer(opt)
  end subroutine h0_gradient

end module Gaussian
