!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!
#ifdef _OPENMP
#define SETLOCK  call omp_set_lock(lock);
#define UNSETLOCK call omp_unset_lock(lock);
#else
#define SETLOCK
#define UNSETLOCK
#endif

module hartreefock
  !! author: Ludger Inhester
  !!
  !! module to construct Fock matrix for RHF and ROHF calculations
  !! and various orbital transformation routines
  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use eigen_lapack
  use operations
  use neweffpot
  use spin_coupling
  use packed_index
#ifdef _OPENMP
  use omp_lib
#endif
  use gaussian
  use basis_func
  use stdio
  use matrix
  implicit none

contains

  subroutine construct_UHF_Fock_Matrix(P, B, F, Dmata, Dmatb, Focka, Fockb)
    !! constructs UHF Fock Matrix
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long), intent(in), dimension(B%N_basis, B%N_basis):: Dmata, Dmatb
    real(long), intent(inout), dimension(B%N_basis, B%N_basis):: Focka, Fockb
    real(long), dimension(B%N_basis*(B%N_basis + 1)/2):: teint_schwarz
    real(long), dimension(B%N_basis, B%N_basis):: Dmat
    real(long) :: teint
    integer :: i, j, k, l
    integer(8) :: ij, kl, ijkl
    real(long) :: epsilonIsq ! threshold for skipping integrals. Integrals smaller than sqrt(epsilonI2) are skipped

    if (P%yes_direct_SCF) then
    call construct_UHF_Fock_Matrix_direct(B%env,B%atm,B%bas,B%first_bf_of_shell,Dmata,Dmatb,B%H0,FockA, FockB, accuracy=P%eps_teint)
      return
    end if
    epsilonIsq = P%EPS_teint**2
    FockA(:, :) = F%H0(:, :)
    FockB(:, :) = F%H0(:, :)
    Dmat = DmatA + DmatB

    ! loop over all unique twoelectronintegral indices
    ! i.e. i>j, k>l and i*(i-1)/2+j > k*(k-1)/2+l
    ! for each i,j pair calculate the potential
    ! and for each k,l pair do the contraction

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! all the same (ii|ii)
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = index_packed(i, i)
      ijkl = index_packed(ij, ij)
      teint = B%teint(ijkl)
      FockA(i, i) = FockA(i, i) + DmatB(i, i)*teint
      FockB(i, i) = FockB(i, i) + DmatA(i, i)*teint
      teint_schwarz(ij) = abs(teint)
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      do j = 1, i - 1
        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
        ij = index_packed(i, i)
        kl = index_packed(i, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        FockA(i, i) = FockA(i, i) + TWO*DmatB(j, i)*teint
        FockB(i, i) = FockB(i, i) + TWO*DmatA(j, i)*teint
        FockA(j, i) = FockA(j, i) + DmatB(i, i)*teint
        FockB(j, i) = FockB(j, i) + DmatA(i, i)*teint
        ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
        ij = index_packed(i, j)
        kl = index_packed(j, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        FockA(j, i) = FockA(j, i) + DmatB(j, j)*teint
        FockA(j, j) = FockA(j, j) + TWO*DmatB(j, i)*teint
        FockB(j, i) = FockB(j, i) + DmatA(j, j)*teint
        FockB(j, j) = FockB(j, j) + TWO*DmatA(j, i)*teint
      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! two pairwise equal i>k
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i + 1)/2
      do k = 1, i - 1
        kl = k*(k + 1)/2
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! (ii|kk) (kk|ii)
        FockA(i, i) = FockA(i, i) + Dmat(k, k)*teint
        FockA(k, k) = FockA(k, k) + Dmat(i, i)*teint
        FockA(k, i) = FockA(k, i) - DmatA(k, i)*teint
        FockB(i, i) = FockB(i, i) + Dmat(k, k)*teint
        FockB(k, k) = FockB(k, k) + Dmat(i, i)*teint
        FockB(k, i) = FockB(k, i) - DmatB(k, i)*teint
      end do
    end do
    !$omp end do
    ! two pairwise equal i>j
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        ijkl = index_packed(ij, ij)
        teint = B%teint(ijkl)
        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
        FockA(j, i) = FockA(j, i) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
        FockA(i, i) = FockA(i, i) - DmatA(j, j)*teint
        FockA(j, j) = FockA(j, j) - DmatA(i, i)*teint
        FockB(j, i) = FockB(j, i) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
        FockB(i, i) = FockB(i, i) - DmatB(j, j)*teint
        FockB(j, j) = FockB(j, j) - DmatB(i, i)*teint
        teint_schwarz(ij) = abs(teint)
      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! three different indices
    ! i=j, l<k<i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      j = i
      ij = i*(i + 1)/2
      !ij=index_packed(i,j)
      do k = 1, i - 1
        kl = k*(k - 1)/2
        do l = 1, k - 1
          kl = kl + 1
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
            FockA(i, i) = FockA(i, i) + TWO*Dmat(l, k)*teint
            FockA(l, k) = FockA(l, k) + Dmat(i, i)*teint
            FockA(k, i) = FockA(k, i) - DmatA(l, i)*teint
            FockA(l, i) = FockA(l, i) - DmatA(k, i)*teint
            FockB(i, i) = FockB(i, i) + TWO*Dmat(l, k)*teint
            FockB(l, k) = FockB(l, k) + Dmat(i, i)*teint
            FockB(k, i) = FockB(k, i) - DmatB(l, i)*teint
            FockB(l, i) = FockB(l, i) - DmatB(k, i)*teint
          end if
        end do! l loop
      end do! k loop
    end do!i loop
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! k=l, l=k<j<i or j<l=k<i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        !ij=index_packed(i,j)
        ij = ij + 1
        do k = 1, j - 1
          l = k
          kl = k*(k + 1)/2
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          FockA(j, i) = FockA(j, i) + Dmat(k, k)*teint
          FockA(k, k) = FockA(k, k) + TWO*Dmat(i, j)*teint
          FockA(k, i) = FockA(k, i) - DmatA(k, j)*teint
          FockA(k, j) = FockA(k, j) - DmatA(i, k)*teint
          FockB(j, i) = FockB(j, i) + Dmat(k, k)*teint
          FockB(k, k) = FockB(k, k) + TWO*Dmat(i, j)*teint
          FockB(k, i) = FockB(k, i) - DmatB(k, j)*teint
          FockB(k, j) = FockB(k, j) - DmatB(i, k)*teint
        end do
        do k = j + 1, i - 1
          l = k
          kl = k*(k + 1)/2
          !kl=index_packed(k,l)
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          FockA(j, i) = FockA(j, i) + Dmat(k, k)*teint
          FockA(k, k) = FockA(k, k) + TWO*Dmat(i, j)*teint
          FockA(k, i) = FockA(k, i) - DmatA(k, j)*teint
          FockA(j, k) = FockA(j, k) - DmatA(i, k)*teint
          FockB(j, i) = FockB(j, i) + Dmat(k, k)*teint
          FockB(k, k) = FockB(k, k) + TWO*Dmat(i, j)*teint
          FockB(k, i) = FockB(k, i) - DmatB(k, j)*teint
          FockB(j, k) = FockB(j, k) - DmatB(i, k)*teint
        end do
      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! i=k, l<j<k=i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        !ij=index_packed(i,j)
        k = i
        kl = k*(k - 1)/2
        do l = 1, j - 1
          kl = kl + 1
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|il) (ji|il) (ij|li) (ji|li)
            ! (il|ij) (il|ji) (li|ij) (li|ji)
            FockA(j, i) = FockA(j, i) + (TWO*DmatB(i, l) + DmatA(i, l))*teint
            FockA(l, i) = FockA(l, i) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
            FockA(i, i) = FockA(i, i) - TWO*DmatA(j, l)*teint
            FockA(l, j) = FockA(l, j) - DmatA(i, i)*teint
            FockB(j, i) = FockB(j, i) + (TWO*DmatA(i, l) + DmatB(i, l))*teint
            FockB(l, i) = FockB(l, i) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
            FockB(i, i) = FockB(i, i) - TWO*DmatB(j, l)*teint
            FockB(l, j) = FockB(l, j) - DmatB(i, i)*teint
          end if
        end do
      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! j=l,  j=l<k<i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        !ij=index_packed(i,j)
        l = j
        do k = j + 1, i - 1
          !kl=index_packed(k,l)
          kl = k*(k - 1)/2 + l
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
            ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
            FockA(j, i) = FockA(j, i) + (TWO*DmatB(k, j) + DmatA(k, j))*teint
            FockA(j, k) = FockA(j, k) + (TWO*DmatB(i, j) + DmatA(i, j))*teint
            FockA(j, j) = FockA(j, j) - TWO*DmatA(i, k)*teint
            FockA(k, i) = FockA(k, i) - DmatA(j, j)*teint
            FockB(j, i) = FockB(j, i) + (TWO*DmatA(k, j) + DmatB(k, j))*teint
            FockB(j, k) = FockB(j, k) + (TWO*DmatA(i, j) + DmatB(i, j))*teint
            FockB(j, j) = FockB(j, j) - TWO*DmatB(i, k)*teint
            FockB(k, i) = FockB(k, i) - DmatB(j, j)*teint
          end if
        end do

      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! j=k, l<k=j<i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        k = j
        kl = k*(k - 1)/2
        do l = 1, k - 1
          !kl=index_packed(k,l)
          kl = kl + 1
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|jl) (ji|jl) (ij|lj) (ji|lj)
            ! (jl|ij) (jl|ji) (lj|ij) (lj|ji)
            FockA(j, i) = FockA(j, i) + (TWO*DmatB(l, j) + DmatA(l, j))*teint
            FockA(l, j) = FockA(l, j) + (TWO*DmatB(j, i) + DmatA(j, i))*teint
            FockA(l, i) = FockA(l, i) - DmatA(j, j)*teint
            FockA(j, j) = FockA(j, j) - TWO*DmatA(l, i)*teint
            FockB(j, i) = FockB(j, i) + (TWO*DmatA(l, j) + DmatB(l, j))*teint
            FockB(l, j) = FockB(l, j) + (TWO*DmatA(j, i) + DmatB(j, i))*teint
            FockB(l, i) = FockB(l, i) - DmatB(j, j)*teint
            FockB(j, j) = FockB(j, j) - TWO*DmatB(l, i)*teint
          end if
        end do
      end do
    end do
    !$omp end do
    !$omp end parallel

    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,DmatA,DmatB,teint_schwarz, epsilonIsq) reduction(+:FockA,FockB)
    ! no indices equal
    ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
    ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
    ! i>j, k<i, l<k
    ! three sub-cases:
    ! l<k<j<i
    ! l<j<k<i
    ! j<l<k<i
    !$omp do schedule(static,1)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        ! l<k<j<i
        do k = 1, j - 1
          kl = k*(k - 1)/2
          do l = 1, k - 1
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)

              FockA(j, i) = FockA(j, i) + TWO*Dmat(k, l)*teint
              FockA(l, k) = FockA(l, k) + TWO*Dmat(j, i)*teint
              FockA(k, i) = FockA(k, i) - DmatA(l, j)*teint
              FockA(l, i) = FockA(l, i) - DmatA(k, j)*teint
              FockA(k, j) = FockA(k, j) - DmatA(l, i)*teint
              FockA(l, j) = FockA(l, j) - DmatA(k, i)*teint
              FockB(j, i) = FockB(j, i) + TWO*Dmat(k, l)*teint
              FockB(l, k) = FockB(l, k) + TWO*Dmat(j, i)*teint
              FockB(k, i) = FockB(k, i) - DmatB(l, j)*teint
              FockB(l, i) = FockB(l, i) - DmatB(k, j)*teint
              FockB(k, j) = FockB(k, j) - DmatB(l, i)*teint
              FockB(l, j) = FockB(l, j) - DmatB(k, i)*teint
            end if
          end do ! l loop
        end do! k loop
        do k = j + 1, i - 1
          kl = k*(k - 1)/2
          do l = 1, j - 1
            ! l<j<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)
              FockA(j, i) = FockA(j, i) + TWO*Dmat(l, k)*teint
              FockA(l, k) = FockA(l, k) + TWO*Dmat(j, i)*teint
              FockA(k, i) = FockA(k, i) - DmatA(l, j)*teint
              FockA(l, i) = FockA(l, i) - DmatA(j, k)*teint
              FockA(j, k) = FockA(j, k) - DmatA(l, i)*teint
              FockA(l, j) = FockA(l, j) - DmatA(k, i)*teint
              FockB(j, i) = FockB(j, i) + TWO*Dmat(l, k)*teint
              FockB(l, k) = FockB(l, k) + TWO*Dmat(j, i)*teint
              FockB(k, i) = FockB(k, i) - DmatB(l, j)*teint
              FockB(l, i) = FockB(l, i) - DmatB(j, k)*teint
              FockB(j, k) = FockB(j, k) - DmatB(l, i)*teint
              FockB(l, j) = FockB(l, j) - DmatB(k, i)*teint
            end if
          end do! l loop
        end do! k loop
        do k = j + 1, i - 1
          kl = k*(k - 1)/2 + j
          do l = j + 1, k - 1
            ! j<l<k<i
            ! kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)
              FockA(j, i) = FockA(j, i) + TWO*Dmat(l, k)*teint
              FockA(l, k) = FockA(l, k) + TWO*Dmat(j, i)*teint
              FockA(k, i) = FockA(k, i) - DmatA(j, l)*teint
              FockA(l, i) = FockA(l, i) - DmatA(j, k)*teint
              FockA(j, k) = FockA(j, k) - DmatA(l, i)*teint
              FockA(j, l) = FockA(j, l) - DmatA(k, i)*teint
              FockB(j, i) = FockB(j, i) + TWO*Dmat(l, k)*teint
              FockB(l, k) = FockB(l, k) + TWO*Dmat(j, i)*teint
              FockB(k, i) = FockB(k, i) - DmatB(j, l)*teint
              FockB(l, i) = FockB(l, i) - DmatB(j, k)*teint
              FockB(j, k) = FockB(j, k) - DmatB(l, i)*teint
              FockB(j, l) = FockB(j, l) - DmatB(k, i)*teint
            end if
          end do! l loop
        end do! k loop
      end do!j loop
    end do!i loop
    !$omp end do
    !$omp end parallel
    do i = 1, P%N_basis
      do j = 1, i - 1
        FockA(i, j) = FockA(j, i)
        FockB(i, j) = FockB(j, i)
      end do
    end do
  end subroutine construct_UHF_Fock_Matrix

  subroutine construct_Fock_Matrix2(h0, teints, D, Fock)
    !! returns RHF Fock matrix
    implicit none
    real(long), intent(in) :: h0(:, :)
    real(long), intent(in) :: teints(:)
    real(long), intent(in), dimension(size(h0, 1), size(h0, 1)) :: D
    real(long), intent(out) :: Fock(size(h0, 1), size(h0, 1))
    real(long), dimension(size(h0, 1), size(h0, 1)):: J_Matrix, K_Matrix
    call construct_J_Matrix(teints, D, J_Matrix)
    call construct_K_Matrix(teints, D, K_Matrix)
    Fock = h0 + J_Matrix - HALF*K_Matrix
  end subroutine construct_Fock_Matrix2

  subroutine construct_ROHF_Fock_Matrix2(P, B, F)
    !! returns ROHF Fock matrix for high spin half filled open shell
    !! alternative ROHF Fock matrix routine
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F

    integer :: i, j
    real(long), dimension(P%N_basis, P%N_basis):: Fock
    real(long), dimension(P%N_basis, P%N_basis):: temp, D_temp
    real(long), dimension(B%N_basis, B%Nr_basis):: X

    integer :: n_shell

    real(long), allocatable, dimension(:, :, :):: D_shell
    real(long), allocatable, dimension(:, :, :):: Fock_shell

    real(long), allocatable :: a_factor(:), b_factor(:, :)
    real(long), allocatable :: occ(:), mu(:, :), lambda(:, :)
    real(long) :: Acc, Bcc, Aoo, Boo, Avv, Bvv
    integer, allocatable :: open_idx(:)

    X = B%S.m.F%LC(:, :)

    if (P%multp .eq. count(F%N_occ .eq. 1) + 1) then
      n_shell = 2
      allocate (a_factor(n_shell))
      allocate (b_factor(n_shell, n_shell))
      allocate (occ(n_shell))
      allocate (mu(n_shell, n_shell))
      allocate (lambda(n_shell, n_shell))
      allocate (D_shell(B%N_basis, B%N_basis, n_shell))
      allocate (Fock_shell(B%N_basis, B%N_basis, n_shell))
      ! high spin case
      ! J coupling
      mu(1, 1) = ONE
      mu(1, 2) = ONE
      mu(2, 1) = ONE
      mu(2, 2) = ONE
      ! K coupling
      lambda(1, 1) = HALF
      lambda(1, 2) = HALF
      lambda(2, 1) = HALF
      lambda(2, 2) = ONE
      ! occupations
      occ(1) = TWO
      occ(2) = ONE
      call density_matrix_closed(P, F, D_shell(:, :, 1))
      call density_matrix_open(P, F, D_shell(:, :, 2))
    else if (P%multp .eq. 1 .and. count(F%N_occ .eq. 1) .eq. 2) then
      n_shell = 3
      allocate (open_idx(count(F%N_occ(:) .eq. 1)))
      open_idx = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 1)
      allocate (a_factor(n_shell))
      allocate (b_factor(n_shell, n_shell))
      allocate (occ(n_shell))
      allocate (mu(n_shell, n_shell))
      allocate (lambda(n_shell, n_shell))
      allocate (D_shell(B%N_basis, B%N_basis, n_shell))
      allocate (Fock_shell(B%N_basis, B%N_basis, n_shell))
      ! J coupling
      mu(1, 1) = ONE
      mu(1, 2) = ONE
      mu(1, 3) = ONE
      mu(2, 1) = ONE
      mu(2, 2) = ONE
      mu(2, 3) = ONE
      mu(3, 1) = ONE
      mu(3, 2) = ONE
      mu(3, 3) = ONE

      ! K coupling
      lambda(1, 1) = HALF
      lambda(3, 1) = HALF
      lambda(2, 1) = HALF

      lambda(1, 2) = HALF
      lambda(2, 2) = ONE
      lambda(3, 2) = -ONE

      lambda(1, 3) = HALF
      lambda(2, 3) = -ONE
      lambda(3, 3) = ONE

      ! occupations
      occ(1) = TWO
      occ(2) = ONE
      occ(3) = ONE
      call density_matrix_closed(P, F, D_shell(:, :, 1))
      do i = 1, B%N_basis
        do j = 1, B%N_basis
          D_shell(i, j, 2) = F%LC(i, open_idx(1))*F%LC(j, open_idx(1))
          D_shell(i, j, 3) = F%LC(i, open_idx(2))*F%LC(j, open_idx(2))
        end do
      end do
    end if

    do i = 1, n_shell
      Fock_shell(:, :, i) = F%h0(:, :)
      D_temp(:, :) = ZERO
      do j = 1, n_shell
        D_temp(:, :) = D_temp(:, :) + occ(j)*mu(j, i)*D_shell(:, :, j)
      end do
      call construct_J_Matrix(B%teint, D_temp, temp)
      Fock_shell(:, :, i) = Fock_shell(:, :, i) + temp(:, :)
      D_temp(:, :) = ZERO
      do j = 1, n_shell
        D_temp(:, :) = D_temp(:, :) + occ(j)*lambda(j, i)*D_shell(:, :, j)
      end do
      call construct_K_Matrix(B%teint, D_temp, temp)
      Fock_shell(:, :, i) = Fock_shell(:, :, i) - temp(:, :)

      temp = F%LC(:, :) .tm.Fock_shell(:, :, i) .m.F%LC(:, :)
      Fock_shell(:, :, i) = temp
    end do

    if (P%multp .eq. count(F%N_occ .eq. 1) + 1) then
      ! See J Chem Phys 125 204110 Table II for different coupling coefficients
      ! ! Guest & Saunders
      ! Acc=HALF
      ! Bcc=HALF
      ! Aoo=HALF
      ! Boo=HALF
      ! Avv=HALF
      ! Bvv=HALF
      ! ! Roothan
      Acc = -HALF
      Bcc = THREE*HALF
      Aoo = HALF
      Boo = HALF
      Avv = THREE*HALF
      Bvv = -HALF
      ! ! Davidson
      !  Acc=HALF
      !  Bcc=HALF
      !  Aoo=ONE
      !  Boo=ZERO
      !  Avv=ONE
      !  Bvv=ZERO

      a_factor(1) = ONE ! this is an arbitraty non-zero coupling coefficient scaling the closed virtual orthogonality constraint
      a_factor(2) = ONE ! this is an arbitraty non-zero coupling coefficient scaling the open virtual orthogonality constraint
      b_factor(1, 2) = ONE ! this is an non-zero coupling coefficient scaling the open-closed orthogonality constraint
      b_factor(2, 1) = -ONE  ! they must be antisymmetric

      do i = 1, P%N_basis
        do j = 1, P%N_basis
          ! cc
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = Acc*Fock_shell(i, j, 2) + Bcc*(TWO*Fock_shell(i, j, 1) - Fock_shell(i, j, 2))
          ! oo
          if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 1) &
            Fock(i, j) = Aoo*Fock_shell(i, j, 2) + Boo*(TWO*Fock_shell(i, j, 1) - Fock_shell(i, j, 2))
          ! vv
          if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = Avv*Fock_shell(i, j, 2) + Bvv*(TWO*Fock_shell(i, j, 1) - Fock_shell(i, j, 2))
          ! cv
          if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = a_factor(1)*Fock_shell(i, j, 1)
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = a_factor(1)*Fock_shell(i, j, 1)
          ! ov
          if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = a_factor(2)*Fock_shell(i, j, 2)
          if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 1) &
            Fock(i, j) = a_factor(2)*Fock_shell(i, j, 2)
          ! co
          if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = b_factor(1, 2)*(occ(1)*Fock_shell(i, j, 1) - occ(2)*Fock_shell(i, j, 2))
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 1) &
            Fock(i, j) = b_factor(2, 1)*(occ(2)*Fock_shell(i, j, 2) - occ(1)*Fock_shell(i, j, 1))
        end do
      end do

    else if (P%multp .eq. 1 .and. count(F%N_occ .eq. 1) .eq. 2) then
      do i = 1, P%N_basis
        do j = 1, P%N_basis
          ! cc
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = Fock_shell(i, j, 1)
          ! vv
          if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = Fock_shell(i, j, 1)
          ! o1 - o1
          if (i .eq. open_idx(1) .and. j .eq. open_idx( 1)) &
            Fock(i, j) = Fock_shell(i, j, 2)
          ! o2 - o2
          if (i .eq. open_idx(2) .and. j .eq. open_idx( 2)) &
            Fock(i, j) = Fock_shell(i, j, 2)
          ! c - v
          if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = Fock_shell(i, j, 1)
          ! v - c
          if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = Fock_shell(i, j, 1)
          ! v - o1
          if (F%N_occ(i) .eq. 0 .and. j .eq. open_idx( 1)) &
            Fock(i, j) = Fock_shell(i, j, 2)
          ! o1 - v
          if (i .eq. open_idx(1) .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = Fock_shell(i, j, 2)
          ! v - o2
          if (F%N_occ(i) .eq. 0 .and. j .eq. open_idx( 2)) &
            Fock(i, j) = Fock_shell(i, j, 3)
          ! o2 - v
          if (i .eq. open_idx(2) .and. F%N_occ(j) .eq. 0) &
            Fock(i, j) = Fock_shell(i, j, 3)
          ! c - o1
          if (F%N_occ(i) .eq. 2 .and. j .eq. open_idx( 1)) &
            Fock(i, j) = Fock_shell(i, j, 2) - Fock_shell(i, j, 1)
          ! o1 - c
          if (i .eq. open_idx(1) .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = Fock_shell(i, j, 1) - Fock_shell(i, j, 2)
          ! c - o2
          if (F%N_occ(i) .eq. 2 .and. j .eq. open_idx( 2)) &
            Fock(i, j) = Fock_shell(i, j, 3) - Fock_shell(i, j, 1)
          ! o2 - c
          if (i .eq. open_idx(2) .and. F%N_occ(j) .eq. 2) &
            Fock(i, j) = Fock_shell(i, j, 1) - Fock_shell(i, j, 3)
          ! o1 - o2
          if (i .eq. open_idx(1) .and. j .eq. open_idx( 2)) &
            Fock(i, j) = Fock_shell(i, j, 2) - Fock_shell(i, j, 3)
          ! o2 - o1
          if (i .eq. open_idx(1) .and. j .eq. open_idx( 2)) &
            Fock(i, j) = Fock_shell(i, j, 3) - Fock_shell(i, j, 2)
        end do
      end do
    end if

    ! transform back into bf space
    !  F%H=(B%S .m. F%LC(:,:)) .m. Fock .mt. (B%S .m. F%LC(:,:) )
    F%H = X.m.Fock.mt.X
    F%FockA = X.m.Fock_shell(:, :, 2) .mt.X
    F%FockB = X.m. (TWO*Fock_shell(:, :, 1) - Fock_shell(:, :, 2)) .mt.X

  end subroutine construct_ROHF_Fock_Matrix2

  subroutine construct_ROHF_Fock_Matrix(P, B, F, variant)
    !!
    !! returns ROHF Fock matrix for high spin half filled open shell
    !! variant=1 (default) Roothan
    !! variant=2 Davidson
    !! variant=3 Guest & Saunders
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in), optional :: variant

    integer :: i, j
    integer :: ROHF_variant = 1
    real(long), dimension(size(F%LC, 2), size(F%LC, 2)):: Focka, Fockb, Fock

    real(long), dimension(size(F%LC, 1), size(F%LC, 2)):: X
    real(long) :: Acc, Bcc, Aoo, Boo, Avv, Bvv

    if (present(variant)) &
      ROHF_variant = variant
    call density_matrix_alpha(P, F, F%Da)
    call density_matrix_beta(P, F, F%Db)

    call construct_UHF_Fock_Matrix(P, B, F, F%Da, F%Db, F%Focka, F%Fockb)

    ! transform into MO space
    FockA = F%LC(:, :) .tm.F%FockA.m.F%LC(:, :)
    FockB = F%LC(:, :) .tm.F%FockB.m.F%LC(:, :)
    X = B%S.m.F%LC(:, :)

    ! See J Chem Phys 125 204110 Table II for different coupling coefficients

    if (variant .eq. 2) then
      ! ! Davidson ( as in molpro)
      Acc = HALF
      Bcc = HALF
      Aoo = ONE
      Boo = ZERO
      Avv = ONE
      Bvv = ZERO
    else if (variant .eq. 1) then
      ! ! Roothan
      Acc = -HALF
      Bcc = THREE*HALF
      Aoo = HALF
      Boo = HALF
      Avv = THREE*HALF
      Bvv = -HALF
    else if (variant .eq. 3) then
      ! ! Guest & Saunders
      Acc = HALF
      Bcc = HALF
      Aoo = HALF
      Boo = HALF
      Avv = HALF
      Bvv = HALF
    else
      stop "unknown ROHF variant"
    end if

    ! beta is spin minority
    ! alpha is spin majority
    do i = 1, size(F%N_occ, 1)
      do j = 1, size(F%N_occ, 1)

        ! cc
        if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 2) Fock(i, j) = Acc*FockA(i, j) + Bcc*FockB(i, j)
        ! oo
        if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 1) Fock(i, j) = Aoo*FockA(i, j) + Boo*FockB(i, j)
        ! vv
        if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 0) Fock(i, j) = Avv*FockA(i, j) + Bvv*FockB(i, j)

        ! cv
        if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 2) Fock(i, j) = HALF*(FockA(i, j) + FockB(i, j))
        if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 0) Fock(i, j) = HALF*(FockA(i, j) + FockB(i, j))
        ! ov
        if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 2) Fock(i, j) = FockB(i, j)
        if (F%N_occ(i) .eq. 2 .and. F%N_occ(j) .eq. 1) Fock(i, j) = FockB(i, j)
        ! co
        if (F%N_occ(i) .eq. 1 .and. F%N_occ(j) .eq. 0) Fock(i, j) = FockA(i, j)
        if (F%N_occ(i) .eq. 0 .and. F%N_occ(j) .eq. 1) Fock(i, j) = FockA(i, j)
      end do
    end do

    ! transform back into bf space
    !  F%H=(B%S .m. F%LC(:,:)) .m. Fock .mt. (B%S .m. F%LC(:,:) )
    F%H = X.m.Fock.mt.X

  end subroutine construct_ROHF_Fock_Matrix

  subroutine construct_Fock_Matrix(P, B, F, use_D_Diff, Dmat_given)
    !! This function constructs the RHF Fock Matrix
    !! assumes that density matrix is symmetric
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    real(long) :: teint
    logical, intent(in), optional    :: use_D_diff
    real(long), intent(in), optional, dimension(P%N_basis, P%N_basis) :: Dmat_given
    real(long), dimension(F%nsave) :: coef
    real(long), dimension(P%N_basis, P%N_basis):: Dmat
    logical  :: yes_use_D_diff = .False.
    integer :: i, j, k, l, count
    integer(8) :: ij, kl, ijkl
    integer :: nonzero_diff, nonzero_orig
    real(long) :: epsilonD ! threshold for Density matrix elements
    real(long) :: epsilonI ! threshold for 2e integrals
    !  real(long) :: teint, teint_ijkl,teint_ikjl,teint_iljk
    real(long), dimension(B%N_basis*(B%N_basis + 1)/2):: teint_schwarz
    real(long) :: epsilonIsq ! threshold for skipping integrals. Integrals smaller than sqrt(epsilonI2) are skipped
    real(long) :: Fout(B%N_basis, B%N_basis)

    epsilonI = P%EPS_teint
    epsilonIsq = P%EPS_teint**2

    if (present(use_D_diff)) yes_use_D_diff = use_D_diff
    ! This is the same as below
    ! call construct_J_Matrix(P, B, F, J_matrix, F%D)
    ! call construct_K_Matrix(P, B, F, K_matrix, F%D)
    ! F%H=F%H0+J_matrix-HALF*K_matrix
    ! return


    epsilonD = P%SCF_density_conv/10.
    if (yes_use_D_diff) then

      !     do k=1,P%DIIS_nvector
      !        RHS(k)=ZERO
      !        do l=1,P%DIIS_nvector
      !           BMatrix(k,l)=ZERO
      !           do i=1,P%N_basis
      !              do j=1,i
      !                 if(k==l)&
      !                      RHS(k)=RHS(k)+F%Dsave(i,j,k)*F%D(i,j)
      !                 BMatrix(k,l)=BMatrix(k,l)+F%Dsave(i,j,k)*F%Dsave(i,j,l)
      !              end do
      !           end do
      !        end do
      !     end do

      !     call solve_AxB(BMatrix, RHS, coef, ierr)

      !     do k=1,P%DIIS_nvector
      !        write(STDERR,'(10f7.3)') BMatrix(k,:)
      !     end do
      !     write(STDERR,*) coef

      !     F%H=F%H0(:,:)
      !     Dmat=F%D
      !     do k=1,P%DIIS_nvector
      !        F%H=F%H+  coef(k) * (F%Hsave(:,:,k)-F%H0(:,:))
      !        Dmat= Dmat - coef(k) * F%Dsave(:,:,k)
      !     end do

      call find_optimal_deltaD(P, F, coef)

      write (STDERR, *) coef
      F%H = F%H0(:, :)
      Dmat = F%D
      do k = 1, F%nsave
        F%H = F%H + coef(k)*(F%Hsave(:, :, k) - F%H0(:, :))
        Dmat = Dmat - coef(k)*F%Dsave(:, :, k)
      end do

      nonzero_orig = 0
      do k = 1, F%n_basis
        do l = k, F%n_basis
          if (abs(F%D(k, l)) .gt. epsilonD) nonzero_orig = nonzero_orig + 1
        end do
      end do

      nonzero_diff = 0
      do k = 1, F%n_basis
        do l = k, F%n_basis
          if (abs(Dmat(k, l)) .gt. epsilonD) nonzero_diff = nonzero_diff + 1
        end do
      end do

      if (P%yes_verbose) &
        write (P%OUTPUT, "('#DDIFF: relation nonzero Dmatrix elements diff/total',f8.3,i5,i5)") &
        ONE*nonzero_diff/nonzero_orig, nonzero_diff, nonzero_orig

      ! If no density matrix elements are remaining skip further calculations
      if (nonzero_diff .eq. 0) return

    else
      F%H(:, :) = F%H0(:, :)
      if (present(Dmat_given)) then
        Dmat = Dmat_given
      else
        Dmat = F%D
      end if
    end if
    if (P%yes_direct_scf) then
      call construct_Fock_Matrix_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, Dmat, B%H0, Fout, accuracy=P%eps_teint)
      F%H = Fout
      return
    end if
    count = 0
    Fout(:,:) = ZERO
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! all the same (ii|ii)
    !$omp do schedule(static)
    do i = 1, B%N_basis
      ij = index_packed(i, i)
      ijkl = index_packed(ij, ij)
      teint = B%teint(ijkl)
      Fout(i, i) = Fout(i, i) + Dmat(i, i) * HALF * teint
      teint_schwarz(ij) = abs(teint)
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! two pairwise equal i>j
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i * (i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        ijkl = index_packed(ij, ij)
        teint = B%teint(ijkl)
        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
        Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(j, i) * teint
        Fout(i, i) = Fout(i, i) - HALF*Dmat(j, j)*teint
        Fout(j, j) = Fout(j, j) - HALF*Dmat(i, i)*teint
        teint_schwarz(ij) = abs(teint)
      end do
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      do j = 1, i - 1
        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
        ij = index_packed(i, i)
        kl = index_packed(i, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        Fout(j, i) = Fout(j, i) + HALF*Dmat(i, i) * teint
        Fout(i, i) = Fout(i, i) + Dmat(j, i) * teint
        ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
        ij = index_packed(i, j)
        kl = index_packed(j, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        Fout(j, i) = Fout(j, i) + HALF*Dmat(j, j) * teint
        Fout(j, j) = Fout(j, j) + Dmat(j, i) * teint
      end do
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! two pairwise equal i>k
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i*(i + 1)/2
      do k = 1, i - 1
        kl = k*(k + 1)/2
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! (ii|kk) (kk|ii)
        Fout(i, i) = Fout(i, i) + Dmat(k, k)*teint
        Fout(k, k) = Fout(k, k) + Dmat(i, i)*teint
        Fout(k, i) = Fout(k, i) - HALF * Dmat(k, i) * teint
      end do
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! three different indices
    ! i=j, l<k<i
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      j = i
      ij = i*(i + 1)/2
      !ij=index_packed(i,j)
      do k = 1, i - 1
        kl = k*(k - 1)/2
        do l = 1, k - 1
          kl = kl + 1
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
            Fout(i, i) = Fout(i, i) + TWO*Dmat(l, k)*teint
            Fout(l, k) = Fout(l, k) + Dmat(i, i)*teint
            Fout(k, i) = Fout(k, i) - HALF*Dmat(l, i)*teint
            Fout(l, i) = Fout(l, i) - HALF*Dmat(k, i)*teint
          end if
        end do! l loop
      end do! k loop
    end do!i loop
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)

    ! k=l, l=k<j<i or j<l=k<i
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        !ij=index_packed(i,j)
        ij = ij + 1
        do k = 1, j - 1
          l = k
          kl = k*(k + 1)/2
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          Fout(j, i) = Fout(j, i) + Dmat(k, k)*teint
          Fout(k, k) = Fout(k, k) + TWO*Dmat(j, i)*teint
          Fout(k, i) = Fout(k, i) - HALF*Dmat(k, j)*teint
          Fout(k, j) = Fout(k, j) - HALF*Dmat(k, i)*teint
        end do
        do k = j + 1, i - 1
          l = k
          kl = k*(k + 1)/2
          !kl=index_packed(k,l)
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          Fout(j, i) = Fout(j, i) + Dmat(k, k)*teint
          Fout(k, k) = Fout(k, k) + TWO*Dmat(j, i)*teint
          Fout(k, i) = Fout(k, i) - HALF*Dmat(j, k)*teint
          Fout(j, k) = Fout(j, k) - HALF*Dmat(k, i)*teint
       end do
      end do!j
    end do! i
    !$omp end do
    ! i=k, l<j<k=i
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        !ij=index_packed(i,j)
        k = i
        kl = k*(k - 1)/2
        do l = 1, j - 1
          kl = kl + 1
          !kl=index_packed(k,l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|il) (ji|il) (ij|li) (ji|li)
            ! (il|ij) (il|ji) (li|ij) (li|ji)
            Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(l, i)*teint
            Fout(l, i) = Fout(l, i) + THREEHALF*Dmat(j, i)*teint
            Fout(l, j) = Fout(l, j) - HALF*Dmat(i, i)*teint
            Fout(i, i) = Fout(i, i) - Dmat(l, j)*teint
          end if
        end do! l
      end do
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! j=l,  j=l<k<i
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        !ij=index_packed(i,j)
        l = j
        do k = j + 1, i - 1
          !kl=index_packed(k,l)
          kl = k*(k - 1)/2 + l
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
            ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
            Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(j, k)*teint
            Fout(j, k) = Fout(j, k) + THREEHALF*Dmat(j, i)*teint
            Fout(k, i) = Fout(k, i) - HALF*Dmat(j, j)*teint
            Fout(j, j) = Fout(j, j) - Dmat(k, i)*teint
          end if
        end do! k
      end do! j
    end do
    !$omp end do
    !$omp end parallel
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    ! j=k, l<k=j<i
    !$omp do schedule(dynamic, 10)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        k = j
        kl = k*(k - 1)/2
        do l = 1, k - 1
          !kl=index_packed(k,l)
          kl = kl + 1
          if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
            ijkl = index_packed(ij, kl)
            teint = B%teint(ijkl)
            ! (ij|jl) (ji|jl) (ij|lj) (ji|lj)
            ! (jl|ij) (jl|ji) (lj|ij) (lj|ji)
            Fout(j, i) = Fout(j, i) + THREEHALF*Dmat(l, j)*teint
            Fout(l, j) = Fout(l, j) + THREEHALF*Dmat(j, i)*teint
            Fout(l, i) = Fout(l, i) - HALF*Dmat(j, j)*teint
            Fout(j, j) = Fout(j, j) - Dmat(l, i)*teint
          end if
        end do! l
      end do! j
    end do
    !$omp end do
    !$omp end parallel
    ! no indices equal
    ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
    ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
    ! i>j, k<i, l<k
    ! three sub-cases:
    ! l<k<j<i
    ! l<j<k<i
    ! j<l<k<i
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint) default(none) shared(B,Dmat,teint_schwarz, epsilonIsq) reduction(+:Fout)
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      ij = i*(i - 1)/2
      do j = 1, i - 1
        ij = ij + 1
        ! l<k<j<i
        do k = 1, j - 1
          kl = k*(k - 1)/2
          do l = 1, k - 1
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)
              Fout(j, i) = Fout(j, i) + TWO*Dmat(l, k)*teint
              Fout(l, k) = Fout(l, k) + TWO*Dmat(j, i)*teint
              Fout(k, i) = Fout(k, i) - HALF*Dmat(l, j)*teint
              Fout(l, i) = Fout(l, i) - HALF*Dmat(k, j)*teint
              Fout(k, j) = Fout(k, j) - HALF*Dmat(l, i)*teint
              Fout(l, j) = Fout(l, j) - HALF*Dmat(k, i)*teint
            end if
          end do ! l loop
        end do! k loop

        do k = j + 1, i - 1
          kl = k*(k - 1)/2
          do l = 1, j - 1
            ! l<j<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)
              Fout(j, i) = Fout(j, i) + TWO*Dmat(l, k)*teint
              Fout(l, k) = Fout(l, k) + TWO*Dmat(j, i)*teint
              Fout(k, i) = Fout(k, i) - HALF*Dmat(l, j)*teint
              Fout(l, i) = Fout(l, i) - HALF*Dmat(j, k)*teint
              Fout(j, k) = Fout(j, k) - HALF*Dmat(l, i)*teint
              Fout(l, j) = Fout(l, j) - HALF*Dmat(k, i)*teint
            end if
          end do! l loop
          kl = k*(k - 1)/2 + j
          do l = j + 1, k - 1
            ! j<l<k<i
            ! kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .gt. epsilonIsq) then
              ijkl = ij*(ij - 1)/2 + kl
              teint = B%teint(ijkl)
              Fout(j, i) = Fout(j, i) + TWO*Dmat(l, k)*teint
              Fout(l, k) = Fout(l, k) + TWO*Dmat(j, i)*teint
              Fout(k, i) = Fout(k, i) - HALF*Dmat(j, l)*teint
              Fout(l, i) = Fout(l, i) - HALF*Dmat(j, k)*teint
              Fout(j, k) = Fout(j, k) - HALF*Dmat(l, i)*teint
              Fout(j, l) = Fout(j, l) - HALF*Dmat(k, i)*teint
            end if
          end do! l loop
        end do! k loop
      end do!j loop
    end do!i loop
    !$omp end do
    !$omp end parallel
    F%H = F%H + Fout
    do i = 1, P%N_basis
      do j = 1, i
        F%H(i, j) = F%H(j, i)
      end do
    end do
  end subroutine construct_Fock_Matrix

  subroutine construct_M_Matrix(P, F, DU, M)
    !! this subroutine returns the matrix
    !! $$M_{ij c}= \sum_o^{\text{occ orbitals}} \sum_{k} \frac{d u_{ko} ({\bf R}) }{d {\bf R}} \left( 4 (i j | k o) - (i  k | j o) -  (i  o | j k) \right)$$
    !! it expects to find the MO-space twoelectron integrals in F%teint
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    real(long), intent(in), dimension(:, :, :):: DU ! input the derivative of orbital rotation
    real(long), intent(out), dimension(:, :, :):: M ! resulting M matrix
    integer :: i, j, k, o
    integer(8) :: ij, ko, ijko
    integer(8) :: ik, jo, ikjo
    integer(8) :: io, jk, iojk
    real(long) :: teint_sum
    if (.not. allocated(F%teint)) then
      stop "Please first calculate and perform transformation of all teints"
    end if
    if (any(F%N_occ .eq. 1) .or. P%yes_HF .eqv. .False.) then
      stop "only works for closed shell Hartree Fock"
    end if
    do i = 1, size(F%LC, 1)
      do j = i, size(F%LC, 1)
        M(i, j, :) = ZERO
        ij = index_packed(i, j)
        do k = 1, size(F%LC, 1)
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          do o = 1, size(F%LC, 1)
            if (F%N_occ(o) .eq. 0) cycle
            ko = index_packed(k, o)
            jo = index_packed(j, o)
            io = index_packed(i, o)
            ijko = index_packed(ij, ko)
            ikjo = index_packed(ik, jo)
            iojk = index_packed(io, jk)
            teint_sum = (FOUR*F%teint(ijko) - F%teint(ikjo) - F%teint(iojk))
            M(i, j, :) = M(i, j, :) + DU(k, o, :)*teint_sum
          end do! o
        end do! k
        M(j, i, :) = M(i, j, :)
      end do!j
    end do!i

  end subroutine construct_M_Matrix

  subroutine construct_Mov_Matrix(P, F, DU, M)
    !! this subroutine returns the matrix
    !! $$M_{ij c}= \sum_o^{\text{occ orbitals}} \sum_{k}^{\text{virt. orbitals}} \frac{d u_{ko} ({\bf R}) }{d {\bf R}} \left( 4 (i j | k o) - (i  k | j o) -  (i  o | j k) \right)$$
    !! it expects to find the MO-space twoelectron integrals in F%teint
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    real(long), intent(in), dimension(:, :, :):: DU ! input the derivative of orbital rotation
    real(long), intent(out), dimension(:, :, :):: M ! resulting M matrix

    real(long) :: M2(size(DU, 3), size(DU, 1), size(DU, 2)) ! reshaped M matrix
    real(long) :: DU2(size(DU, 3), size(DU, 1), size(DU, 2)) ! reshaped DU matrix
    integer :: i, j, k, o
    integer(8) :: ij, ko, ijko
    integer(8) :: ik, jo, ikjo
    integer(8) :: io, jk, iojk
    real(long) :: teint_sum

    if (.not. allocated(F%teint)) then
      stop "Please first calculate and perform transformation of all teints"
    end if
    if (any(F%N_occ .eq. 1) .or. P%yes_HF .eqv. .False.) then
      stop "only works for closed shell Hartree Fockk"
    end if

    DU2 = reshape(source=DU, shape=shape(DU2), order=(/2, 3, 1/))
    if (.not. allocated(F%teint)) then
      stop "Please first calculate and perform transformation of all teints"
    end if
    if (any(F%N_occ .eq. 1) .or. P%yes_HF .eqv. .False.) then
      stop "only works for closed shell Hartree Fock"
    end if
    do i = 1, size(F%LC, 1)
      do j = i, size(F%LC, 1)
        M2(:, i, j) = ZERO
        ij = index_packed(i, j)
        do k = 1, size(F%LC, 1)
          if (F%N_occ(k) .eq. 2) cycle
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          do o = 1, size(F%LC, 1)
            if (F%N_occ(o) .eq. 0) cycle
            ko = index_packed(k, o)
            jo = index_packed(j, o)
            io = index_packed(i, o)
            ijko = index_packed(ij, ko)
            ikjo = index_packed(ik, jo)
            iojk = index_packed(io, jk)
            teint_sum = (FOUR*F%teint(ijko) - F%teint(ikjo) - F%teint(iojk))
            M2(:, i, j) = M2(:, i, j) + DU2(:, k, o)*teint_sum
          end do! o, occupied orbitals
        end do! k, virtual orbitals
        M2(:, j, i) = M2(:, i, j)
      end do!j
    end do!i

    ! reshape: new order is 3,1,2
    M(:, :, :) = reshape(source=M2, shape=shape(M), order=(/3, 1, 2/))

  end subroutine construct_Mov_Matrix

  subroutine construct_Mov_Matrix2(P, A, B, F, DU, M)
    !! this subroutine returns the matrix
    !! $$M_{ij c}= \sum_o^{\text{occ orbitals}} \sum_{k}^{\text{virt. orbitals}} \frac{d u_{ko} ({\bf R}) }{d {\bf R}} \left( 4 (i j | k o) - (i  k | j o) -  (i  o | j k) \right)$$
    !! idea: employ a density difference and screen low contributions
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(in), dimension(:, :, :):: DU ! input the derivative of orbital rotation
    real(long), intent(out), dimension(:, :, :):: M ! resulting M matrix

    integer :: i, j, k, l, o, v, ai, ae
    integer(8) :: ij, kl, ijkl
    real(long) :: X(size(F%LC, 1), size(F%LC, 1))
    real(long) :: teint
    ! real(long) :: duTrans_sumSym(size(DU,3),size(F%LC,1),size(F%LC,1))
    real(long), allocatable :: duHalfTrans(:, :, :), duTrans(:, :, :), duTrans_sumSym(:, :, :)
    integer, allocatable :: idx_occ(:), idx_virt(:)
    real(long), allocatable :: X2(:, :, :), DU2(:, :, :) ! reshaped X and DO matrices
    real(long), dimension(size(F%LC, 1)*(size(F%LC, 1) + 1)/2):: teint_schwarz ! separate array for integrals (ij|ij)
    logical :: screenDU(size(F%LC, 1), size(F%LC, 1))
    real(long) :: epsilonIsq ! threshold for skipping integrals. Integrals smaller than sqrt(epsilonI2) are skipped
    real(long) :: epsilonDU ! threshold for skipping DU entries
    real :: timer
    real :: time_init, time_last_step
    integer*8 :: int_time, int_time_last   ! timer

    real(long), allocatable :: Xji(:), Xki(:), Xkj(:), Xjk(:), Xxi(:,:), Xxj(:,:), Xxk(:,:), Xjx(:,:)
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    call check_time(int_time, timer)
    time_init = timer

    epsilonIsq = P%SCF_energy_conv**2
    epsilonDU = P%SCF_density_conv

    if (any(F%N_occ .eq. 1) .or. P%yes_HF .eqv. .False.) then
      stop "only works for closed shell Hartree Fock"
    end if

    allocate (idx_occ(count((/(F%N_occ(:) .eq. 2)/))))
    allocate (idx_virt(count((/(F%N_occ(:) .eq. 0)/))))
    idx_occ = pack((/(j, j=1, size(F%N_occ, 1))/), (/(F%N_occ(:) .eq. 2)/))
    idx_virt = pack((/(j, j=1, size(F%N_occ, 1))/), (/(F%N_occ(:) .eq. 0)/))

    allocate (DU2(size(DU,3), size(DU, 1), size(DU, 2)))
    allocate (duTrans_sumSym(size(DU,3), size(F%LC, 1), size(F%LC, 1)))
    allocate (X2(size(DU,3), size(F%LC, 1), size(F%LC, 1)))
    DU2=reshape(source=DU,shape=shape(DU2),order=(/2,3,1/))
    duTrans_sumSym(:, :, :) = ZERO
    screenDU(:, :) = .False.
    call check_time(int_time, timer)
    time_last_step = timer
    allocate (duHalfTrans(size(DU2, 1), size(F%LC, 1), size(idx_occ, 1)))
    duHalfTrans(:, :, :) = ZERO
    do k = 1, size(F%LC, 1)
      do o = 1, size(idx_occ)
        do v = 1, size(idx_virt)
          duHalfTrans(:, k, o) = duHalfTrans(:, k, o) + F%LC(k, idx_virt(v)) * DU2(:, idx_virt(v), idx_occ(o))
        end do! v
      end do! o
    end do! k
    deallocate (DU2)
    allocate (duTrans(size(duHalfTrans, 1), size(F%LC, 1), size(F%LC, 1)))
    duTrans(:, :, :) = ZERO
    do k = 1, size(F%LC, 1)
      do l = 1, size(F%LC, 1)
        do o = 1, size(idx_occ)
          duTrans(:, k, l) = duTrans(:, k, l) + F%LC(l, idx_occ(o))*duHalfTrans(:, k, o)
        end do! v
      end do! l
    end do! k
    deallocate (duHalfTrans)
    deallocate (idx_virt)
    deallocate (idx_occ)

    do k = 1, size(F%LC, 1)
      do l = 1, size(F%LC, 1)
        duTrans_sumSym(:, l, k) = duTrans(:, k, l) + duTrans(:, l, k)
        if (maxval(abs(duTrans_sumSym(:, l, k))) .lt. epsilonDU) then
          screenDU(l, k) = .True.
          screenDU(k, l) = .True.
        end if
      end do! l
    end do! k

    deallocate (duTrans)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for duTrans",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! the following loops take the most computing time
    X2(:, :, :) = ZERO
    do i = 1, size(F%LC, 1)
      ij = index_packed(i, i)
      ijkl = index_packed(ij, ij)
      teint = B%teint(ijkl)
      X2(:, i, i) = X2(:, i, i) + duTrans_sumSym(:, i, i)*teint
      teint_schwarz(ij) = abs(teint)
    end do

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for iiii entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! three indices same one unique (either larger or smaller than the others)
    do i = 1, P%N_basis
      do j = 1, i - 1
        ij = index_packed(i, i)
        kl = index_packed(i, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
        X2(:, i, i) = X2(:, i, i) + TWO*duTrans_sumSym(:, j, i)*teint
        X2(:, j, i) = X2(:, j, i) + duTrans_sumSym(:, i, i)*teint
        ij = index_packed(i, j)
        kl = index_packed(j, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
        X2(:, j, i) = X2(:, j, i) + duTrans_sumSym(:, j, j)*teint
        X2(:, j, j) = X2(:, j, j) + TWO*duTrans_sumSym(:, j, i)*teint
      end do
    end do

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for iii-j entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! two pairwise equal i>j
    do i = 1, P%N_basis
      do j = 1, i - 1
        ij = index_packed(i, j)
        ijkl = index_packed(ij, ij)
        teint = B%teint(ijkl)
        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
        X2(:, j, i) = X2(:, j, i) + THREE*duTrans_sumSym(:, j, i)*teint
        X2(:, i, i) = X2(:, i, i) - duTrans_sumSym(:, j, j)*teint
        X2(:, j, j) = X2(:, j, j) - duTrans_sumSym(:, i, i)*teint
        teint_schwarz(ij) = abs(teint)
      end do
    end do

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for ijij entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! two pairwise equal i>k
    do i = 1, P%N_basis
      do k = 1, i - 1
        ij = index_packed(i, i)
        kl = index_packed(k, k)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! (ii|kk) (kk|ii)
        X2(:, i, i) = X2(:, i, i) + TWO*duTrans_sumSym(:, k, k)*teint
        X2(:, k, k) = X2(:, k, k) + TWO*duTrans_sumSym(:, i, i)*teint
        X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, k, i)*teint
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for iijj entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer
    int_time_last = int_time
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    ! three different indices
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint, Xxi, Xxj, Xxk, Xjx)  default(none) shared(lock,B,duTrans_sumSym,teint_schwarz,epsilonIsq,X2)
    allocate (Xxi(size(X2,1), size(X2,2)), Xxj(size(X2,1), size(X2,2)), Xxk(size(X2,1), size(X2,2)), Xjx(size(X2,1), size(X2,2)))
    ! i=j, l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      j = i
      ij = index_packed(i, j)
      Xxi(:,:) = ZERO
      do k = 1, i - 1
        Xxk(:, :) = ZERO
        do l = 1, k - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
          !X2(:, i, i) = X2(:, i, i) + FOUR*duTrans_sumSym(:, l, k)*teint
          Xxi(:, i) = Xxi(:, i) + FOUR*duTrans_sumSym(:, l, k)*teint
          !X2(:, l, k) = X2(:, l, k) + TWO*duTrans_sumSym(:, i, i)*teint
          Xxk(:, l) = Xxk(:, l) + TWO*duTrans_sumSym(:, i, i)*teint
          !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, l, i)*teint
          Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, l, i)*teint
          !X2(:, l, i) = X2(:, l, i) - duTrans_sumSym(:, k, i)*teint
          Xxi(:, l) = Xxi(:, l) - duTrans_sumSym(:, k, i)*teint
        end do! l loop
        SETLOCK
        X2(:, 1:k-1, k) = X2(:, 1:k-1, k) + Xxk(:, 1:k-1)
        UNSETLOCK
      end do! k loop
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:, 1:i)
      UNSETLOCK
    end do!i loop
    !$omp end do

    ! k=l, l=k<j<i or j<l=k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
        Xxk(:,:) = ZERO
        ij = index_packed(i, j)
        do k = 1, j - 1
          l = k
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          !X2(:, j, i) = X2(:, j, i) + TWO*duTrans_sumSym(:, k, k)*teint
          Xxi(:, j) = Xxi(:, j) + TWO*duTrans_sumSym(:, k, k)*teint
          !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, k, j)*teint
          Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, k, j)*teint
          !X2(:, k, k) = X2(:, k, k) + FOUR*duTrans_sumSym(:, j, i)*teint
          Xxk(:, k) = Xxk(:, k) + FOUR*duTrans_sumSym(:, j, i)*teint
          !X2(:, k, j) = X2(:, k, j) - duTrans_sumSym(:, k, i)*teint
          Xxj(:, k) = Xxj(:, k) - duTrans_sumSym(:, k, i)*teint
        end do
        do k = j + 1, i - 1
          l = k
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          !X2(:, j, i) = X2(:, j, i) + TWO*duTrans_sumSym(:, k, k)*teint
          Xxi(:, j) = Xxi(:, j) + TWO*duTrans_sumSym(:, k, k)*teint
          !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, k, j)*teint
          Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, k, j)*teint
          !X2(:, k, k) = X2(:, k, k) + FOUR*duTrans_sumSym(:, j, i)*teint
          Xxk(:, k) = Xxk(:, k) + FOUR*duTrans_sumSym(:, j, i)*teint
          !X2(:, j, k) = X2(:, j, k) - duTrans_sumSym(:, i, k)*teint
          Xjx(:, k) = Xjx(:, k) - duTrans_sumSym(:, i, k)*teint
        end do!k
        SETLOCK
        forall (k=1:i-1)
          X2(:, k, k) = X2(:, k, k) + Xxk(:, k)
        end forall
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        X2(:, j, j+1:i-1) = X2(:, j, j+1:i-1) + Xjx(:,j+1:i-1)
        UNSETLOCK
      end do!j
      SETLOCK
      X2(:, 1:i-1, i) = X2(:, 1:i-1, i) + Xxi(:,1:i-1)
      UNSETLOCK
    end do!i
    !$omp end do

    ! i=k, l<j<k=i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        ij = index_packed(i, j)
        k = i
        do l = 1, j - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|il) (ji|il) (ij|li) (ji|li)
          ! (il|ij) (il|ji) (li|ij) (li|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*duTrans_sumSym(:, l, i)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*duTrans_sumSym(:, l, i)*teint
          !X2(:, l, i) = X2(:, l, i) + THREE*duTrans_sumSym(:, j, i)*teint
          Xxi(:, l) = Xxi(:, l) + THREE*duTrans_sumSym(:, j, i)*teint
          !X2(:, l, j) = X2(:, l, j) - duTrans_sumSym(:, i, i)*teint
          Xxj(:, l) = Xxj(:, l) - duTrans_sumSym(:, i, i)*teint
          !X2(:, i, i) = X2(:, i, i) - TWO*duTrans_sumSym(:, l, j)*teint
          Xxi(:, i) = Xxi(:, i) - TWO*duTrans_sumSym(:, l, j)*teint
        end do
        SETLOCK
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !$omp end do
    ! j=l,  j=l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xjx(:,:) = ZERO
        ij = index_packed(i, j)
        l = j
        do k = j + 1, i - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)

          ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
          ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*duTrans_sumSym(:, k, j)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*duTrans_sumSym(:, k, j)*teint
          !X2(:, j, k) = X2(:, j, k) + THREE*duTrans_sumSym(:, j, i)*teint
          Xjx(:, k) = Xjx(:, k) + THREE*duTrans_sumSym(:, j, i)*teint
          !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, j, j)*teint
          Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, j, j)*teint
          !X2(:, j, j) = X2(:, j, j) - TWO*duTrans_sumSym(:, k, i)*teint
          Xjx(:, j) = Xjx(:, j) - TWO*duTrans_sumSym(:, k, i)*teint
        end do
        SETLOCK
        X2(:, j, j:i-1) = X2(:, j, j:i-1) + Xjx(:, j:i-1)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !$omp end do

    ! j=k, l<k=j<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        ij = index_packed(i, j)
        k = j
        do l = 1, k - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|jl) (ji|jl) (ij|lj) (ji|lj)
          ! (jl|ij) (jl|ji) (lj|ij) (lj|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*duTrans_sumSym(:, l, j)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*duTrans_sumSym(:, l, j)*teint
          !X2(:, l, j) = X2(:, l, j) + THREE*duTrans_sumSym(:, j, i)*teint
          Xxj(:, l) = Xxj(:, l) + THREE*duTrans_sumSym(:, j, i)*teint
          !X2(:, l, i) = X2(:, l, i) - duTrans_sumSym(:, j, j)*teint
          Xxi(:, l) = Xxi(:, l) - duTrans_sumSym(:, j, j)*teint
          !X2(:, j, j) = X2(:, j, j) - TWO*duTrans_sumSym(:, l, i)*teint
          Xxj(:, j) = Xxj(:, j) - TWO*duTrans_sumSym(:, l, i)*teint
        end do
        SETLOCK
        X2(:, 1:j, j) = X2(:, 1:j, j) + Xxj(:,1:j)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !$omp end do

    ! no indices equal
    ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
    ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
    ! i>j, k<i, l<k
    ! three sub-cases:
    ! l<k<j<i
    ! l<j<k<i
    ! j<l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      ij = i*(i - 1)/2
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
        !ij=i*(i-1)/2+j
        ij = ij + 1
        ! l<k<j<i
        kl = 0
        do k = 1, j - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2
          do l = 1, k - 1
            kl = kl + 1
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*duTrans_sumSym(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*duTrans_sumSym(:, l, k)*teint
            !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, l, j)*teint
            Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, l, j)*teint
            !X2(:, k, j) = X2(:, k, j) - duTrans_sumSym(:, l, i)*teint
            Xxj(:, k) = Xxj(:, k) - duTrans_sumSym(:, l, i)*teint
            !X2(:, l, i) = X2(:, l, i) - duTrans_sumSym(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - duTrans_sumSym(:, j, k)*teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*duTrans_sumSym(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*duTrans_sumSym(:, j, i)*teint
            !X2(:, l, j) = X2(:, l, j) - duTrans_sumSym(:, k, i)*teint
            Xxj(:, l) = Xxj(:, l) - duTrans_sumSym(:, k, i)*teint
          end do ! l loop
          SETLOCK
          X2(:, 1:k-1, k) = X2(:, 1:k-1, k) + Xxk(:,1:k-1)
          UNSETLOCK
          XXk(:,1:k-1) = ZERO
        end do! k loop
        !kl=(j+1)*j/2
        do k = j + 1, i - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2
          do l = 1, j - 1
            ! l<j<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*duTrans_sumSym(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*duTrans_sumSym(:, l, k)*teint
            !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, l, j)*teint
            Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, l, j)*teint
            !X2(:, j, k) = X2(:, j, k) - duTrans_sumSym(:, l, i)*teint
            Xxk(:, j) = Xxk(:, j) - duTrans_sumSym(:, l, i)*teint
            !X2(:, l, i) = X2(:, l, i) - duTrans_sumSym(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - duTrans_sumSym(:, j, k) * teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*duTrans_sumSym(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*duTrans_sumSym(:, j, i)*teint
            !X2(:, l, j) = X2(:, l, j) - duTrans_sumSym(:, k, i)*teint
            Xxj(:, l) = Xxj(:, l) - duTrans_sumSym(:, k, i)*teint
          end do! l loop
          SETLOCK
          X2(:, 1:j, k) = X2(:, 1:j, k) + Xxk(:,1:j)
          UNSETLOCK
          Xxk(:, 1:j) = ZERO
        end do! k loop
        !kl=(j+1)*j/2 + j
        do k = j + 1, i - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2 + j
          do l = j + 1, k - 1
            ! j<l<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k-1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*duTrans_sumSym(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*duTrans_sumSym(:, l, k)*teint
            !X2(:, k, i) = X2(:, k, i) - duTrans_sumSym(:, j, l)*teint
            Xxi(:, k) = Xxi(:, k) - duTrans_sumSym(:, j, l)*teint
            !X2(:, j, k) = X2(:, j, k) - duTrans_sumSym(:, l, i)*teint
            Xxk(:, j) = Xxk(:, j) - duTrans_sumSym(:, l, i)*teint
            !X2(:, l, i) = X2(:, l, i) - duTrans_sumSym(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - duTrans_sumSym(:, j, k)*teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*duTrans_sumSym(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*duTrans_sumSym(:, j, i)*teint
            !X2(:, j, l) = X2(:, j, l) - duTrans_sumSym(:, k, i)*teint
            Xjx(:, l) = Xjx(:, l) - duTrans_sumSym(:, k, i)*teint
          end do! l loop
          SETLOCK
          X2(:, j : k - 1, k) = X2(:, j : k - 1, k) + Xxk(:,j : k - 1)
          UNSETLOCK
          Xxk(:,j : k - 1) = ZERO
        end do! k loop
        SETLOCK
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        X2(:, j, j+1:) = X2(:, j, j+1:) + Xjx(:,j+1:)
        UNSETLOCK
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
      end do!j loop
      SETLOCK
      X2(:, 1:i-1, i) = X2(:, 1:i-1, i) + Xxi(:, 1:i-1)
      UNSETLOCK
      Xxi(:,:) = ZERO
    end do!i loop
    !$omp end do
    deallocate (Xxi, Xxj, Xxk, Xjx)
    !$omp end parallel
    deallocate (duTrans_sumSym)
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","time for iikl and ijkl entries",f15.3,"sec.  real time:",f15.3,"sec. ")') &
        timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time

    do i = 1, P%N_basis
      do j = 1, i
        X2(:, i, j) = X2(:, j, i)
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","symmetrize X2",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    M(:, :, :) = ZERO
    do i = 1, size(DU,3)
      X(:, :) = X2(i, :, :)
      M(:, :, i) = F%LC .tm. X(:, :) .m. F%LC
    end do!c
    deallocate (X2)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Mov ","transform X",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

  end subroutine construct_Mov_Matrix2

  subroutine construct_Moo_Matrix(P, F, DO, M)
    !! this subroutine returns the matrix
    !! $$M_{ij c}= \sum_o^{\text{occ orbitals}} \sum^{\text{occ orbitals}}_{p} \frac{d O_{op} ({\bf R}) }{d {\bf R}} \left( 2 (i j | o p) - (i  o | j p) \right)$$
    !! it expects to find the MO-space twoelectron integrals in F%teint
    type(Param), intent(in) :: P
    type(WF_data), intent(in) :: F
    real(long), intent(in), dimension(:, :, :):: DO ! input the derivative of overlap matrix
    real(long), intent(out), dimension(:, :, :):: M ! resulting M matrix
    real(long) :: M2(size(DO, 3), size(DO, 1), size(DO, 2)) ! reshaped M matrix
    real(long) :: DO2(size(DO, 3), size(DO, 1), size(DO, 2)) ! reshaped DO matrix

    integer :: i, j, q, o
    integer(8) :: ij, oq, ijoq
    integer(8) :: io, jq, iojq
    real(long) :: teint_sum

    if (.not. allocated(F%teint)) then
      stop "Please first calculate and perform transformation of all teints"
    end if
    if (any(F%N_occ .eq. 1) .or. .not. P%yes_HF) then
      stop "only works for closed shell Hartree Fock"
    end if

    DO2 = reshape(source=DO, shape=shape(DO2), order=(/2, 3, 1/))
    do i = 1, size(F%LC, 1)
      do j = i, size(F%LC, 1)
        M2(:, i, j) = ZERO
        ij = index_packed(i, j)
        do o = 1, size(F%LC, 1)
          if (F%N_occ(o) .eq. 0) cycle
          io = index_packed(i, o)
          do q = 1, size(F%LC, 1)
            if (F%N_occ(q) .eq. 0) cycle
            oq = index_packed(o, q)
            jq = index_packed(j, q)
            ijoq = index_packed(ij, oq)
            iojq = index_packed(io, jq)
            teint_sum = (TWO*F%teint(ijoq) - F%teint(iojq))
            M2(:, i, j) = M2(:, i, j) - DO2(:, o, q) * teint_sum
          end do! o, occupied
        end do! q, occupied
        M2(:, j, i) = M2(:, i, j)
      end do!j
    end do!i
    ! order is 3,1,2
    M(:, :, :) = reshape(source=M2, shape=shape(M), order=(/3, 1, 2/))

  end subroutine construct_Moo_Matrix

  subroutine construct_Moo_Matrix2(P, A, B, F, DO, M)
    !! this subroutine returns the matrix
    !! $$ M_{ij c}= \sum_o^{\text{occ orbitals}} \sum^{\text{occ orbitals}}_{p} - \frac{d O_{op} ({\bf R}) }{d {\bf R}} \left( 2 (i j | o p) - (i  o | j p) \right) $$
    !! it doesn't need transformed integrals
    !! TODO parallize!
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    real(long), intent(in), dimension(:, :, :):: DO ! input the derivative of overlap matrix
    real(long), intent(out), dimension(:, :, :):: M ! resulting M matrix

    integer :: i, j, k, l, o1, o2
    integer(8) :: ij, kl, ijkl
    real(long) :: teint
    real(long), allocatable :: doTrans_sum(:, :, :)
    real(long), allocatable :: doHalfTrans(:, :, :)
    integer, allocatable :: idx_occ(:)
    !integer :: dimr, rind_thresh(size(DO, 3))
    real(long) :: X(size(F%LC, 1), size(F%LC, 1)) ! untrasformed M matrix
    real(long), allocatable :: X2(:, :, :) ! reshaped X Matrix
    real(long), allocatable :: DO2(:, :, :) ! reshaped DO matrix
    real(long), dimension(P%N_basis*(P%N_basis + 1)/2):: teint_schwarz ! separate array for integrals (ij|ij)
    real(long) :: epsilonIsq ! threshold for skipping integrals. Integrals smaller than sqrt(epsilonI2) are skipped
    real(long) :: epsilonDO ! threshold for skipping DO entries
    logical :: screenDO(size(F%LC, 1), size(F%LC, 1))
    real :: timer
    real :: time_init, time_last_step
    integer*8 :: int_time, int_time_last      ! timer
    real(long), allocatable :: Xji(:), Xki(:), Xkj(:), Xjk(:), Xxi(:,:), Xxj(:,:), Xxk(:,:), Xjx(:,:)
#ifdef _OPENMP
    integer (kind=omp_lock_kind) :: lock
#endif
    call check_time(int_time, timer)
    time_init = timer

    epsilonIsq = P%SCF_energy_conv**2
    epsilonDO = P%SCF_density_conv
    if (any(F%N_occ .eq. 1) .or. .not. P%yes_HF) then
      stop "only works for closed shell Hartree Fock"
    end if

    allocate (idx_occ(count((/(F%N_occ(:) .eq. 2)/))))
    idx_occ = pack((/(j, j=1, size(F%N_occ, 1))/), (/(F%N_occ(:) .eq. 2)/))

    ! Create DO2 matrix eliminating the rows of DO that are not point charges
    !dimr = 0
    !do i = 1, size(A, 1)
    !  ! atom = int(i/3) + 1
    !  if (A(i)%charge .ne. 0) then
    !    rind_thresh(dimr + 1:dimr + 3) = (/3*i - 2, 3*i - 1, 3*i/)
    !    dimr = dimr + 3
    !  end if
    !end do
    allocate (DO2(size(DO,3), size(DO, 1), size(DO, 2)))
    allocate (doHalfTrans(size(DO2, 1), size(F%LC, 1), size(idx_occ, 1)))
    allocate (doTrans_sum(size(DO2, 1), size(F%LC, 1), size(F%LC, 1)))
    allocate (X2(size(DO,3), size(F%LC, 1), size(F%LC, 1)))
    DO2=reshape(source=DO(:,:,:),shape=shape(DO2),order=(/2,3,1/))
    call check_time(int_time, timer)
    time_last_step = timer
    int_time_last = int_time
    doHalfTrans(:, :, :) = ZERO
    do k = 1, size(F%LC, 1)
      do o2 = 1, size(idx_occ)
        do o1 = 1, size(idx_occ)
          doHalfTrans(:, k, o2) = doHalfTrans(:, k, o2) + F%LC(k, idx_occ(o1))*DO2(:, idx_occ(o1), idx_occ(o2))
        end do! o1
      end do! o2
    end do! k
    deallocate (DO2)
    doTrans_sum(:, :, :) = ZERO
    do k = 1, size(F%LC, 1)
      do l = 1, size(F%LC, 1)
        do o2 = 1, size(idx_occ)
          doTrans_sum(:, k, l) = doTrans_sum(:, k, l) - F%LC(l, idx_occ(o2))*doHalfTrans(:, k, o2)
        end do! o2
        if (maxval(abs(doTrans_sum(:, k, l))) .lt. epsilonDO) then
          screenDO(l, k) = .True.
          screenDO(k, l) = .True.
        end if
      end do! l
    end do! k
    deallocate (doHalfTrans)
    deallocate (idx_occ)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for doTrans",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer
    int_time_last = int_time

    X2(:, :, :) = ZERO

    do i = 1, B%N_basis
      ij = index_packed(i, i)
      ijkl = index_packed(ij, ij)
      teint = B%teint(ijkl)
      X2(:, i, i) = X2(:, i, i) + doTrans_sum(:, i, i)*teint
      teint_schwarz(ij) = abs(teint)
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for iiii entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! three indices same one unique (either larger or smaller than the others)
    do i = 1, B%N_basis
      do j = 1, i - 1
        ij = index_packed(i, i)
        kl = index_packed(i, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
        X2(:, i, i) = X2(:, i, i) + TWO*doTrans_sum(:, j, i)*teint
        X2(:, j, i) = X2(:, j, i) + doTrans_sum(:, i, i)*teint
        ij = index_packed(i, j)
        kl = index_packed(j, j)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
        X2(:, j, i) = X2(:, j, i) + doTrans_sum(:, j, j)*teint
        X2(:, j, j) = X2(:, j, j) + TWO*doTrans_sum(:, j, i)*teint
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for iii-j entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! two pairwise equal i>j
    do i = 1, B%N_basis
      do j = 1, i - 1
        ij = index_packed(i, j)
        ijkl = index_packed(ij, ij)
        teint = B%teint(ijkl)
        ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
        X2(:, j, i) = X2(:, j, i) + THREE*doTrans_sum(:, j, i)*teint
        X2(:, i, i) = X2(:, i, i) - doTrans_sum(:, j, j)*teint
        X2(:, j, j) = X2(:, j, j) - doTrans_sum(:, i, i)*teint
        teint_schwarz(ij) = abs(teint)
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for ijij entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    ! two pairwise equal i>k
    do i = 1, B%N_basis
      do k = 1, i - 1
        ij = index_packed(i, i)
        kl = index_packed(k, k)
        ijkl = index_packed(ij, kl)
        teint = B%teint(ijkl)
        ! (ii|kk) (kk|ii)
        X2(:, i, i) = X2(:, i, i) + TWO*doTrans_sum(:, k, k)*teint
        X2(:, k, k) = X2(:, k, k) + TWO*doTrans_sum(:, i, i)*teint
        X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, k, i)*teint
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for iijj entries",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer
    int_time_last = int_time
#ifdef _OPENMP
    call omp_init_lock(lock)
#endif
    ! three different indices
    !$omp parallel private(i,j,k,l,ij,kl,ijkl,teint, Xxi, Xxj, Xxk, Xjx)  default(none) shared(lock,B,doTrans_sum,teint_schwarz,epsilonIsq,X2)
    allocate (Xxi(size(X2,1), size(X2,2)), Xxj(size(X2,1), size(X2,2)), Xxk(size(X2,1), size(X2,2)), Xjx(size(X2,1), size(X2,2)))
    ! i=j, l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      j = i
      ij = index_packed(i, j)
      Xxi(:,:) = ZERO
      do k = 1, i - 1
        Xxk(:, :) = ZERO
        do l = 1, k - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
          !X2(:, i, i) = X2(:, i, i) + FOUR*doTrans_sum(:, l, k)*teint
          Xxi(:, i) = Xxi(:, i) + FOUR*doTrans_sum(:, l, k)*teint
          !X2(:, l, k) = X2(:, l, k) + TWO*doTrans_sum(:, i, i)*teint
          Xxk(:, l) = Xxk(:, l) + TWO*doTrans_sum(:, i, i)*teint
          !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, l, i)*teint
          Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, l, i)*teint
          !X2(:, l, i) = X2(:, l, i) - doTrans_sum(:, k, i)*teint
          Xxi(:, l) = Xxi(:, l) - doTrans_sum(:, k, i)*teint
        end do! l loop
        SETLOCK
        X2(:, 1:k-1, k) = X2(:, 1:k-1, k) + Xxk(:, 1:k-1)
        UNSETLOCK
      end do! k loop
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:, 1:i)
      UNSETLOCK
    end do!i loop
    !$omp end do
    ! k=l, l=k<j<i or j<l=k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
        Xxk(:,:) = ZERO
        ij = index_packed(i, j)
        do k = 1, j - 1
          l = k
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          !X2(:, j, i) = X2(:, j, i) + TWO*doTrans_sum(:, k, k)*teint
          Xxi(:, j) = Xxi(:, j) + TWO*doTrans_sum(:, k, k)*teint
          !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, k, j)*teint
          Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, k, j)*teint
          !X2(:, k, k) = X2(:, k, k) + FOUR*doTrans_sum(:, j, i)*teint
          Xxk(:, k) = Xxk(:, k) + FOUR*doTrans_sum(:, j, i)*teint
          !X2(:, k, j) = X2(:, k, j) - doTrans_sum(:, k, i)*teint
          Xxj(:, k) = Xxj(:, k) - doTrans_sum(:, k, i)*teint
        end do
        do k = j + 1, i - 1
          l = k
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
          !X2(:, j, i) = X2(:, j, i) + TWO*doTrans_sum(:, k, k)*teint
          Xxi(:, j) = Xxi(:, j) + TWO*doTrans_sum(:, k, k)*teint
          !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, k, j)*teint
          Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, k, j)*teint
          !X2(:, k, k) = X2(:, k, k) + FOUR*doTrans_sum(:, j, i)*teint
          Xxk(:, k) = Xxk(:, k) + FOUR*doTrans_sum(:, j, i)*teint
          !X2(:, j, k) = X2(:, j, k) - doTrans_sum(:, i, k)*teint
          Xjx(:, k) = Xjx(:, k) - doTrans_sum(:, i, k)*teint
        end do! k
        SETLOCK
        forall (k=1:i-1)
          X2(:, k, k) = X2(:, k, k) + Xxk(:, k)
        end forall
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        X2(:, j, j+1:i-1) = X2(:, j, j+1:i-1) + Xjx(:,j+1:i-1)
        UNSETLOCK
      end do!j
      SETLOCK
      X2(:, 1:i-1, i) = X2(:, 1:i-1, i) + Xxi(:,1:i-1)
      UNSETLOCK
    end do! i
    !$omp end do

    ! i=k, l<j<k=i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        ij = index_packed(i, j)
        k = i
        do l = 1, j - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|il) (ji|il) (ij|li) (ji|li)
          ! (il|ij) (il|ji) (li|ij) (li|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*doTrans_sum(:, l, i)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*doTrans_sum(:, l, i)*teint
          !X2(:, l, i) = X2(:, l, i) + THREE*doTrans_sum(:, j, i)*teint
          Xxi(:, l) = Xxi(:, l) + THREE*doTrans_sum(:, j, i)*teint
          !X2(:, l, j) = X2(:, l, j) - doTrans_sum(:, i, i)*teint
          Xxj(:, l) = Xxj(:, l) - doTrans_sum(:, i, i)*teint
          !X2(:, i, i) = X2(:, i, i) - TWO*doTrans_sum(:, l, j)*teint
          Xxi(:, i) = Xxi(:, i) - TWO*doTrans_sum(:, l, j)*teint
        end do
        SETLOCK
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !$omp end do
    ! j=l,  j=l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xjx(:,:) = ZERO
        ij = index_packed(i, j)
        l = j
        do k = j + 1, i - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)

          ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
          ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*doTrans_sum(:, k, j)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*doTrans_sum(:, k, j)*teint
          !X2(:, j, k) = X2(:, j, k) + THREE*doTrans_sum(:, j, i)*teint
          Xjx(:, k) = Xjx(:, k) + THREE*doTrans_sum(:, j, i)*teint
          !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, j, j)*teint
          Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, j, j)*teint
          !X2(:, j, j) = X2(:, j, j) - TWO*doTrans_sum(:, k, i)*teint
          Xjx(:, j) = Xjx(:, j) - TWO*doTrans_sum(:, k, i)*teint
        end do
        SETLOCK
        X2(:, j, j:i-1) = X2(:, j, j:i-1) + Xjx(:, j:i-1)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !$omp end do
    ! j=k, l<k=j<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        ij = index_packed(i, j)
        k = j
        do l = 1, k - 1
          kl = index_packed(k, l)
          if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
          ijkl = index_packed(ij, kl)
          teint = B%teint(ijkl)
          ! (ij|jl) (ji|jl) (ij|lj) (ji|lj)
          ! (jl|ij) (jl|ji) (lj|ij) (lj|ji)
          !X2(:, j, i) = X2(:, j, i) + THREE*doTrans_sum(:, l, j)*teint
          Xxi(:, j) = Xxi(:, j) + THREE*doTrans_sum(:, l, j)*teint
          !X2(:, l, j) = X2(:, l, j) + THREE*doTrans_sum(:, j, i)*teint
          Xxj(:, l) = Xxj(:, l) + THREE*doTrans_sum(:, j, i)*teint
          !X2(:, l, i) = X2(:, l, i) - doTrans_sum(:, j, j)*teint
          Xxi(:, l) = Xxi(:, l) - doTrans_sum(:, j, j)*teint
          !X2(:, j, j) = X2(:, j, j) - TWO*doTrans_sum(:, l, i)*teint
          Xxj(:, j) = Xxj(:, j) - TWO*doTrans_sum(:, l, i)*teint
        end do
        SETLOCK
        X2(:, 1:j, j) = X2(:, 1:j, j) + Xxj(:,1:j)
        UNSETLOCK
      end do
      SETLOCK
      X2(:, 1:i, i) = X2(:, 1:i, i) + Xxi(:,1:i)
      UNSETLOCK
    end do
    !omp end do

    ! no indices equal
    ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
    ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
    ! i>j, k<i, l<k
    ! three sub-cases:
    ! l<k<j<i
    ! l<j<k<i
    ! j<l<k<i
    !$omp do schedule(guided)
    do i = 1, B%N_basis
      Xxi(:,:) = ZERO
      ij = i*(i - 1)/2
      do j = 1, i - 1
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
        ij = ij + 1
        !ij = i*(i - 1)/2 + j
        ! l<k<j<i
        kl = 0
        do k = 1, j - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2
          do l = 1, k - 1
            kl = kl + 1
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k - 1)/2 + l
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*doTrans_sum(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*doTrans_sum(:, l, k)*teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*doTrans_sum(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*doTrans_sum(:, j, i)*teint
            !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, l, j)*teint
            Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, l, j)*teint
            !X2(:, l, i) = X2(:, l, i) - doTrans_sum(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - doTrans_sum(:, j, k)*teint            
            !X2(:, k, j) = X2(:, k, j) - doTrans_sum(:, l, i)*teint
            Xxj(:, k) = Xxj(:, k) - doTrans_sum(:, l, i)*teint
            !X2(:, l, j) = X2(:, l, j) - doTrans_sum(:, k, i)*teint
            Xxj(:, l) = Xxj(:, l) - doTrans_sum(:, k, i)*teint
          end do ! l loop
          SETLOCK
          X2(:, 1:k-1, k) = X2(:, 1:k-1, k) + Xxk(:,1:k-1)
          UNSETLOCK
        end do! k loop
        do k = j + 1, i - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2
          do l = 1, j - 1
            ! l<j<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k - 1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*doTrans_sum(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*doTrans_sum(:, l, k)*teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*doTrans_sum(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*doTrans_sum(:, j, i)*teint
            !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, l, j)*teint
            Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, l, j)*teint
            !X2(:, l, i) = X2(:, l, i) - doTrans_sum(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - doTrans_sum(:, j, k) * teint
            !X2(:, j, k) = X2(:, j, k) - doTrans_sum(:, l, i)*teint
            Xxk(:, j) = Xxk(:, j) - doTrans_sum(:, l, i)*teint
            !X2(:, l, j) = X2(:, l, j) - doTrans_sum(:, k, i)*teint
            Xxj(:, l) = Xxj(:, l) - doTrans_sum(:, k, i)*teint
          end do! l loop
          SETLOCK
          X2(:, 1:j, k) = X2(:, 1:j, k) + Xxk(:,1:j)
          UNSETLOCK
        end do! k loop
        do k = j + 1, i - 1
          Xxk(:,:) = ZERO
          kl = k*(k - 1)/2 + j
          do l = j + 1, k - 1
            ! j<l<k<i
            !kl=index_packed(k,l)
            !ijkl=index_packed(ij,kl)
            !kl = k*(k - 1)/2 + l
            kl = kl + 1
            if (teint_schwarz(ij)*teint_schwarz(kl) .le. epsilonIsq) cycle
            ijkl = ij*(ij - 1)/2 + kl
            teint = B%teint(ijkl)
            !X2(:, j, i) = X2(:, j, i) + FOUR*doTrans_sum(:, l, k)*teint
            Xxi(:, j) = Xxi(:, j) + FOUR*doTrans_sum(:, l, k)*teint
            !X2(:, l, k) = X2(:, l, k) + FOUR*doTrans_sum(:, j, i)*teint
            Xxk(:, l) = Xxk(:, l) + FOUR*doTrans_sum(:, j, i)*teint
            !X2(:, k, i) = X2(:, k, i) - doTrans_sum(:, j, l)*teint
            Xxi(:, k) = Xxi(:, k) - doTrans_sum(:, j, l)*teint
            !X2(:, l, i) = X2(:, l, i) - doTrans_sum(:, j, k)*teint
            Xxi(:, l) = Xxi(:, l) - doTrans_sum(:, j, k)*teint
            !X2(:, j, k) = X2(:, j, k) - doTrans_sum(:, l, i)*teint
            Xxk(:, j) = Xxk(:, j) - doTrans_sum(:, l, i)*teint
            !X2(:, j, l) = X2(:, j, l) - doTrans_sum(:, k, i)*teint
            Xjx(:, l) = Xjx(:, l) - doTrans_sum(:, k, i)*teint
          end do! l loop
          SETLOCK
          X2(:, j : k - 1, k) = X2(:, j : k - 1, k) + Xxk(:,j : k - 1)
          UNSETLOCK
          Xxk(:,j : k - 1) = ZERO
        end do! k loop
        SETLOCK
        X2(:, 1:j-1, j) = X2(:, 1:j-1, j) + Xxj(:,1:j-1)
        X2(:, j, j+1:) = X2(:, j, j+1:) + Xjx(:,j+1:)
        UNSETLOCK
        Xxj(:,:) = ZERO
        Xjx(:,:) = ZERO
      end do!j loop
      SETLOCK
      X2(:, 1:i-1, i) = X2(:, 1:i-1, i) + Xxi(:, 1:i-1)
      UNSETLOCK
      Xxi(:,:) = ZERO
    end do!i loop
    !$omp end do
    deallocate (Xxi, Xxj, Xxk, Xjx)
    !$omp end parallel
    deallocate (doTrans_sum)
#ifdef _OPENMP
    call omp_destroy_lock(lock)
#endif
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","time for iikl and ijkl entries",f15.3,"sec.  real time:",f15.3,"sec. ")') &
        timer - time_last_step, user_time(int_time - int_time_last)
    time_last_step = timer
    int_time_last = int_time
    do i = 1, B%N_basis
      do j = 1, i
        X2(:, i, j) = X2(:, j, i)
      end do
    end do
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","symmetrize X2",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    M(:, :, :) = ZERO
    do i = 1, size(DO,3)
      X(:, :) = X2(i, :, :)
      M(:, :, i) = F%LC .tm. X(:, :) .m.F%LC
    end do
    deallocate (X2)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# Moo ","transform X",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer
  end subroutine construct_Moo_Matrix2


  subroutine construct_J_Matrix(teints, Dmat, J_matrix)
    !! This function constructs the J Matrix
    real(long), intent(in) :: teints(:)
    real(long), intent(in), dimension(:, :):: Dmat
    real(long), intent(out), dimension(size(Dmat, 1), size(Dmat, 2)):: J_matrix

    real(long) :: Dmatsym(size(Dmat, 1), size(Dmat, 2))
    real(long) :: teint
    integer :: i, j, k, l
    integer(8) :: ij, kl, ijkl
    integer :: N
    Dmatsym = Dmat + transpose(Dmat)
    N = size(Dmat, 1)
    ! loop over all unique twoelectronintegral indices
    ! i.e. \(i>j, k>l\) and \(i(i-1)/2+j > k(k-1)/2+l\)
    ! for each i,j pair calculate the potential
    ! and for each k,l pair do the contraction
    J_matrix(:, :) = ZERO
    do i = 1, N
      do j = 1, i
        ij = index_packed(i, j)
        do k = 1, N
          do l = 1, k
            kl = index_packed(k, l)
            if (ij .lt. kl) cycle
            ijkl = index_packed(ij, kl)
            teint = teints(ijkl)
            ! 4 indices equal
            if (i .eq. j .and. k .eq. l .and. i .eq. k) then
              ! (ii|ii)
              J_matrix(i, i) = J_matrix(i, i) + Dmat(i, i)*teint
              ! 3 indices equal
            else if (i .eq. j .and. j .eq. k) then
              ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
              J_matrix(i, i) = J_matrix(i, i) + Dmatsym(l, i)*teint
              J_matrix(i, l) = J_matrix(i, l) + Dmat(i, i)*teint
              J_matrix(l, i) = J_matrix(l, i) + Dmat(i, i)*teint
            else if (i .eq. j .and. j .eq. l) then
              ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
              J_matrix(i, i) = J_matrix(i, i) + Dmatsym(i, k)*teint
              J_matrix(i, k) = J_matrix(i, k) + Dmat(i, i)*teint
              J_matrix(k, i) = J_matrix(k, i) + Dmat(i, i)*teint
            else if (i .eq. k .and. k .eq. l) then
              ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmat(i, i)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmat(i, i)*teint
              J_matrix(i, i) = J_matrix(i, i) + Dmatsym(j, i)*teint
            else if (j .eq. k .and. k .eq. l) then
              ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmat(j, j)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmat(j, j)*teint
              J_matrix(j, j) = J_matrix(j, j) + Dmatsym(j, i)*teint
              ! two indices pairwise equal
            else if (i .eq. j .and. k .eq. l) then
              ! (ii|kk) (kk|ii)
              J_matrix(i, i) = J_matrix(i, i) + Dmat(k, k)*teint
              J_matrix(k, k) = J_matrix(k, k) + Dmat(i, i)*teint
            else if (i .eq. k .and. j .eq. l) then
              ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmatsym(i, j)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmatsym(i, j)*teint
              ! two indices equal
            else if (i .eq. j) then
              ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
              J_matrix(i, i) = J_matrix(i, i) + Dmatsym(l, k)*teint
              J_matrix(k, l) = J_matrix(k, l) + Dmat(i, i)*teint
              J_matrix(l, k) = J_matrix(l, k) + Dmat(i, i)*teint
            else if (k .eq. l) then
              ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmat(k, k)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmat(k, k)*teint
              J_matrix(k, k) = J_matrix(k, k) + Dmatsym(j, i)*teint
            else if (i .eq. k .and. j .ne. l) then
              ! (ij|il) (ji|il) (ij|li) (ji|li)
              ! (il|ij) (il|ji) (li|ij) (li|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmatsym(l, i)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmatsym(l, i)*teint
              J_matrix(i, l) = J_matrix(i, l) + Dmatsym(j, i)*teint
              J_matrix(l, i) = J_matrix(l, i) + Dmatsym(j, i)*teint
            else if (i .ne. k .and. j .eq. l) then
              ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
              ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmatsym(j, k)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmatsym(j, k)*teint
              J_matrix(k, j) = J_matrix(k, j) + Dmatsym(j, i)*teint
              J_matrix(j, k) = J_matrix(j, k) + Dmatsym(j, i)*teint
              ! no indices equal
            else
              ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
              ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
              J_matrix(i, j) = J_matrix(i, j) + Dmatsym(l, k)*teint
              J_matrix(j, i) = J_matrix(j, i) + Dmatsym(l, k)*teint
              J_matrix(k, l) = J_matrix(k, l) + Dmatsym(j, i)*teint
              J_matrix(l, k) = J_matrix(l, k) + Dmatsym(j, i)*teint
            end if
          end do ! l loop
        end do! k loop
      end do!j loop
    end do!i loop

  end subroutine construct_J_Matrix

  subroutine construct_K_Matrix(teints, Dmat, K_matrix)
    !! construct the K matrix
    real(long), intent(in) :: teints(:)
    real(long), intent(in), dimension(:, :):: Dmat
    real(long), intent(out), dimension(size(Dmat, 1), size(Dmat, 2)):: K_matrix
    integer :: i, j, k, l
    integer(8) :: ij, kl, ijkl
    real(long) :: teint
    integer :: N
    real(long) :: Dmatsym(size(Dmat, 1), size(Dmat, 2))
    N = size(Dmat, 1)
    Dmatsym = Dmat + transpose(Dmat)
    K_matrix(:, :) = ZERO
    ! loop over all unique twoelectronintegral indices
    ! i.e. i>j, k>l and i*(i-1)/2+j > k*(k-1)/2+l
    ! for each i,j pair calculate the potential
    ! and for each k,l pair do the contraction
    do i = 1, N
      do j = 1, i
        ij = index_packed(i, j)
        !if ( B%yes_overlap(0,ij) .eqv. .false. ) cycle
        do k = 1, N
          do l = 1, k
            kl = index_packed(k, l)
            if (ij .lt. kl) cycle
            ijkl = index_packed(ij, kl)
            teint = teints(ijkl)
            ! 4 indices equal
            if (i .eq. j .and. k .eq. l .and. i .eq. k) then
              ! (ii|ii)
              K_matrix(i, i) = K_matrix(i, i) + Dmat(i, i)*teint
              ! 3 indices equal
            else if (i .eq. j .and. j .eq. k) then
              ! l unique: (ii|il) (ii|li) (il|ii) (li|ii)
              K_matrix(i, l) = K_matrix(i, l) + Dmat(i, i)*teint
              K_matrix(l, i) = K_matrix(l, i) + Dmat(i, i)*teint
              K_matrix(i, i) = K_matrix(i, i) + Dmatsym(l, i)*teint
            else if (i .eq. j .and. j .eq. l) then
              ! k unique (ii|ki) (ii|ik) (ik|ii) (ki|ii)
              K_matrix(i, i) = K_matrix(i, i) + Dmatsym(k, i)*teint
              K_matrix(i, k) = K_matrix(i, k) + Dmat(i, i)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(i, i)*teint
            else if (i .eq. k .and. k .eq. l) then
              ! j unique: (ij|ii) (ji|ii) (ii|ij) (ii|ji)
              K_matrix(i, i) = K_matrix(i, i) + Dmatsym(i, j)*teint
              K_matrix(j, i) = K_matrix(j, i) + Dmat(i, i)*teint
              K_matrix(i, j) = K_matrix(i, j) + Dmat(i, i)*teint
            else if (j .eq. k .and. k .eq. l) then
              ! i unique: (ij|jj) (ji|jj) (jj|ij) (jj|ji)
              K_matrix(j, j) = K_matrix(j, j) + Dmatsym(j, i)*teint
              K_matrix(i, j) = K_matrix(i, j) + Dmat(j, j)*teint
              K_matrix(j, i) = K_matrix(j, i) + Dmat(j, j)*teint
              ! two indices pairwise equal
            else if (i .eq. j .and. k .eq. l .and. .True.) then
              ! (ii|kk) (kk|ii)
              K_matrix(i, k) = K_matrix(i, k) + Dmat(i, k)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(k, i)*teint
            else if (i .eq. k .and. j .eq. l) then
              ! (ij|ij) (ij|ji) (ji|ij) (ji|ji)
              K_matrix(i, j) = K_matrix(i, j) + Dmat(j, i)*teint
              K_matrix(i, i) = K_matrix(i, i) + Dmat(j, j)*teint
              K_matrix(j, j) = K_matrix(j, j) + Dmat(i, i)*teint
              K_matrix(j, i) = K_matrix(j, i) + Dmat(i, j)*teint
              ! two indices equal
            else if (i .eq. j) then
              ! (ii|kl) (ii|lk) (kl|ii) (lk|ii)
              K_matrix(i, k) = K_matrix(i, k) + Dmat(i, l)*teint
              K_matrix(i, l) = K_matrix(i, l) + Dmat(i, k)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(l, i)*teint
              K_matrix(l, i) = K_matrix(l, i) + Dmat(k, i)*teint
            else if (k .eq. l) then
              ! (ij|kk) (ji|kk) (kk|ij) (kk|ji)
              K_matrix(i, k) = K_matrix(i, k) + Dmat(j, k)*teint
              K_matrix(j, k) = K_matrix(j, k) + Dmat(i, k)*teint
              K_matrix(k, j) = K_matrix(k, j) + Dmat(k, i)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(k, j)*teint
            else if (i .eq. k .and. j .ne. l) then
              ! (ij|il) (ji|il) (ij|li) (ji|li)
              ! (il|ij) (il|ji) (li|ij) (li|ji)
              K_matrix(i, i) = K_matrix(i, i) + Dmatsym(l, j)*teint
              K_matrix(j, i) = K_matrix(j, i) + Dmat(i, l)*teint
              K_matrix(i, l) = K_matrix(i, l) + Dmat(j, i)*teint
              K_matrix(j, l) = K_matrix(j, l) + Dmat(i, i)*teint
              K_matrix(i, j) = K_matrix(i, j) + Dmat(l, i)*teint
              K_matrix(l, i) = K_matrix(l, i) + Dmat(i, j)*teint
              K_matrix(l, j) = K_matrix(l, j) + Dmat(i, i)*teint
            else if (i .ne. k .and. j .eq. l) then
              ! (ij|kj) (ji|kj) (ij|jk) (ji|jk)
              ! (kj|ij) (kj|ji) (jk|ij) (jk|ji)
              K_matrix(i, k) = K_matrix(i, k) + Dmat(j, j)*teint
              K_matrix(j, k) = K_matrix(j, k) + Dmat(i, j)*teint
              K_matrix(i, j) = K_matrix(i, j) + Dmat(j, k)*teint
              K_matrix(j, j) = K_matrix(j, j) + Dmatsym(k, i)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(j, j)*teint
              K_matrix(k, j) = K_matrix(k, j) + Dmat(j, i)*teint
              K_matrix(j, i) = K_matrix(j, i) + Dmat(k, j)*teint
              ! no indices equal
            else
              ! (ij|kl) (ji|kl) (ij|lk) (ji|lk)
              ! (kl|ij) (kl|ji) (lk|ij) (lk|ji)
              K_matrix(i, k) = K_matrix(i, k) + Dmat(j, l)*teint
              K_matrix(j, k) = K_matrix(j, k) + Dmat(i, l)*teint
              K_matrix(i, l) = K_matrix(i, l) + Dmat(j, k)*teint
              K_matrix(j, l) = K_matrix(j, l) + Dmat(i, k)*teint
              K_matrix(k, i) = K_matrix(k, i) + Dmat(l, j)*teint
              K_matrix(k, j) = K_matrix(k, j) + Dmat(l, i)*teint
              K_matrix(l, i) = K_matrix(l, i) + Dmat(k, j)*teint
              K_matrix(l, j) = K_matrix(l, j) + Dmat(k, i)*teint
            end if
          end do
        end do
      end do
    end do
    !  do i=1,P%N_basis
    !    do j=1,i
    !      K_matrix(j,i)=K_matrix(i,j)
    !    end do
    !  end do
  end subroutine construct_K_Matrix


  subroutine density_matrix_alpha(P, F, D)
    !! spin-alpha density matrix where alpha is the majority spin population
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), intent(inout), dimension(P%N_basis, P%N_basis):: D
    integer :: i, j, k, N
    N = F%N_basis
    D = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, size(F%LC, 2)
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          if (F%N_occ(k) .eq. 2 .or. F%N_occ(k) .eq. &
              1) D(i, j) = D(i, j) + F%LC(i, k)*F%LC(j, k)
          if (i .ne. j) D(j, i) = D(i, j)
        end do
      end do
    end do
  end subroutine density_matrix_alpha


  subroutine density_matrix_beta(P, F, D)
    !! spin-beta density matrix where alpha is the majority spin population
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), intent(inout), dimension(P%N_basis, P%N_basis):: D
    integer :: i, j, k, N

    N = F%N_basis
    D = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, size(F%LC, 2)
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          if (F%N_occ(k) .eq. 2 &
              ) D(i, j) = D(i, j) + F%LC(i, k)*F%LC(j, k)
          if (i .ne. j) D(j, i) = D(i, j)
        end do
      end do
    end do
  end subroutine density_matrix_beta


  subroutine density_matrix_closed(P, F, D)
    !! closed shell part density matrix
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), intent(inout), dimension(P%N_basis, P%N_basis):: D

    integer :: i, j, k, N

    N = F%N_basis
    D = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, size(F%LC, 2)
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          if (F%N_occ(k) .eq. 2 )&
               D(i, j) = D(i, j) + F%LC(i, k)*F%LC(j, k)
          if (i .ne. j) D(j, i) = D(i, j)
        end do
      end do
    end do
  end subroutine density_matrix_closed

  subroutine density_matrix_open(P, F, D)
    !! open shell part of the density matrix
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    real(long), intent(inout), dimension(P%N_basis, P%N_basis):: D

    integer :: i, j, k, N

    N = F%N_basis
    D = ZERO
    ! if not yes_thermal
    ! note that F%n_occ is supposed to be the same as F%f_occ
    do k = 1, N
      do j = 1, N    ! 1 ... # of atomic orbitals
        do i = j, N     ! 1 ... # of atomic orbitals
          if (F%N_occ(k) .eq. 1 )&
              D(i, j) = D(i, j) + F%LC(i, k)*F%LC(j, k)
          if (i .ne. j) D(j, i) = D(i, j)
        end do
      end do
    end do
  end subroutine density_matrix_open

  subroutine precalculate_twoel(P, A, G, B)
    !! precalculate all 2el matrix elements
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(inout) :: B

    real(long), dimension(G%N_grid) :: rhoij
    real(long), dimension(G%N_grid) :: vij
    integer :: i, j, k, l, gs, ge, nuc
    integer(8) :: ij, kl, nij, nijkl, ijkl

    !  integer :: NORB,NELEC,MS2,ISYM
    !  integer , dimension(10) :: ORBSYM

    !  namelist / FCI / NORB,NELEC,MS2,ORBSYM,ISYM

    !  NORB=P%N_basis
    !  NELEC=sum(P%N_occ)
    !  MS2=0
    !  ORBSYM(:)=1
    !  ISYM=1

    !  write(OUTPUT,nml=FCI)

    if (P%EPS_00 .gt. TINY(P%EPS_00)) then
      write (STDERR, *) 'Please use option  -eps0 0. '
      stop
    end if

    nij = P%N_basis*(P%N_basis + 1)/2
    nijkl = nij*(nij + 1)/2
    allocate (B%teint(nijkl))

    do i = 1, P%N_basis
      do j = 1, i
        ij = index_packed(i, j)

        ! Screen according to yes_overlap. This is adapted from subroutine density
        rhoij(:) = ZERO
        do nuc = 1, G%N_nuc
          if (B%yes_overlap(nuc, ij) .eqv. .false.) cycle
          gs = G%i_start(nuc)
          ge = G%i_end(nuc)
          rhoij(gs:ge) = B%func(gs:ge, i)*B%func(gs:ge, j)
        end do

        vij = potential_J(P, G, A, rhoij)

        do k = 1, P%N_basis
          do l = 1, k
            kl = index_packed(k, l)
            if (ij .lt. kl) cycle
            ijkl = index_packed(ij, kl)
            B%teint(ijkl) = braket(G, B, k, l, vij)
          end do! l loop
        end do! k loop
      end do ! j loop
    end do ! i loop

  end subroutine precalculate_twoel

  subroutine transform_twoel(P, B, F, OUTPUT)
    !! transform all 2el matrix elements into MO space
    type(Param), intent(in) :: P
    type(Basis), intent(inout) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: temp(:, :, :)
    real(long), allocatable :: temp2(:, :)

    !real(long), allocatable :: temp3(:,:,:,:)
    !real(long) :: testval
    integer :: i, j, k, l
    integer :: ni
    integer(8) :: nij, nijkl, ij, kl, ijkl
    integer :: r, s, t, u
    integer(8) :: rs, tu, rstu
    real :: timer
    real :: time_init, time_done
    integer*8 :: int_time      ! timer
    call check_time(int_time, timer)
    time_init = timer

    ni = P%N_basis
    nij = P%N_basis*(P%N_basis + 1)/2
    nijkl = nij*(nij + 1)/2
    if (.not. allocated(F%teint)) allocate (F%teint(nijkl))
    F%teint(:) = ZERO
    allocate (temp(ni, ni, nij))
    allocate (temp2(nij, nij))

    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space")')
    ! This is the N^8 contraction alorithm
    ! which gives the same as the lower N^5 alorithm
    ! result is stored in temp3(i,j,k,l)
    ! allocate(temp3(ni,ni,ni,ni))
    ! temp3(:,:,:,:)=ZERO
    ! do i=1,P%N_basis
    !    do j=1,P%N_basis
    !       do k=1,P%N_basis
    !          do l=1,P%N_basis
    !             do r=1,P%N_basis
    !                do s=1,P%N_basis
    !                   rs=index_packed(r,s)
    !                   do t=1,P%N_basis
    !                      do u=1,P%N_basis
    !                         tu=index_packed(t,u)
    !                         rstu=index_packed(rs,tu)
    !                         temp3(i,j,k,l)=temp3(i,j,k,l)+&
    !                              B%teint(rstu)*F%LC(r,i)*F%LC(s,j)*F%LC(t,k)*F%LC(u,l)
    !                      end do
    !                   end do
    !                end do
    !             end do
    !          end do
    !       end do
    !    end do
    ! end do
    ! ! N^5 contraction alorithm
    ! contract over i
    temp(:, :, :) = ZERO
    do tu = 1, nij
      do s = 1, P%N_basis
        do i = 1, P%N_basis
          do r = 1, P%N_basis
            rs = index_packed(r, s)
            rstu = index_packed(rs, tu)
            temp(i, s, tu) = temp(i, s, tu) + B%teint(rstu)*F%LC(r, i)
          end do
        end do
      end do
    end do
    temp2 = ZERO
    do tu = 1, nij
      do i = 1, P%N_basis
        do j = 1, i
          ij = index_packed(i, j)
          do s = 1, P%N_basis
            temp2(ij, tu) = temp2(ij, tu) + &
                            temp(i, s, tu)*F%LC(s, j)
          end do
        end do
      end do
    end do
    temp2 = reshape(source=temp2, shape=shape(temp2), order=(/2, 1/))
    temp = ZERO
    do ij = 1, nij
      do u = 1, P%N_basis
        do k = 1, P%N_basis
          do t = 1, P%N_basis
            tu = index_packed(t, u)
            temp(k, u, ij) = temp(k, u, ij) + &
                             temp2(tu, ij)*F%LC(t, k)
          end do
        end do
      end do
    end do
    temp = reshape(source=temp, shape=shape(temp), order=(/2, 1, 3/))
    F%teint = ZERO
    do ij = 1, nij
      do k = 1, P%N_basis
        do l = 1, k
          kl = index_packed(k, l)
          if (ij .lt. kl) cycle
          ijkl = index_packed(ij, kl)
          do u = 1, P%N_basis
            F%teint(ijkl) = F%teint(ijkl) + &
                            temp(u, k, ij)*F%LC(u, l)
          end do
        end do
      end do
    end do
    deallocate (temp)
    deallocate (temp2)

    call check_time(int_time, timer)
    time_done = timer

    if (P%yes_verbose) &
      write (OUTPUT, '("# ","time",f15.3,"sec ")') time_done - time_init
  end subroutine transform_twoel

  subroutine transform_twoel_allactive(P, Ba, F, active, OUTPUT)
    !! transform all active 2el matrix elements into MO space
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: Ba
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: active(:)
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: temp(:, :, :)
    real(long), allocatable :: temp2(:, :), temp2T(:,:)
    real(long) :: LCact(size(F%LC, 1), size(active, 1))
    real(long), allocatable :: temp3(:,:)
    !real(long) :: testval
    integer :: ni, nc
    integer(8) :: ab, cd, abcd
    integer(8) ::  nij, nabcd, ncd, nci
    integer :: r, s, t, u, a, b, c, d
    integer(8) :: rs, tu, rstu
    real :: timer
    real :: time_init, time_done
    integer*8 :: int_time      ! timer
    call check_time(int_time, timer)
    time_init = timer

    nc = size(active)
    ni = Ba%N_basis
    nci = nc*ni
    ncd = nc*(nc + 1)/2
    nij = ni*(ni + 1)/2
    nabcd = ncd*(ncd + 1)/2
    if (allocated(F%teint_abcd)) then
      if (size(F%teint_abcd, 1) .ne. nabcd) deallocate (F%teint_abcd)
    end if
    if (.not. allocated(F%teint_abcd)) allocate (F%teint_abcd(nabcd))
    F%teint_abcd(:) = ZERO
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space (active,active,active,active)")')
    ! N^5 contraction alorithm
    ! first transformation (rs|tu) -> (rs|tc)
    !allocate (temp(nc, ni, nij))
    !call contraction_ijkl2ijka(P, B%teint, LCactive, n, temp)
    LCact(:, :) = F%LC(:, active(:))
    allocate (temp2(ncd, nij))
    temp2(:, :) = ZERO
    if (P%yes_direct_scf) then
      allocate (temp(nc, ni, nij))
      call calculate_1contracted_teints_direct(Ba%env, Ba%atm, Ba%bas, Ba%first_bf_of_shell, size(LCact, 1), &
                                               LCact, temp, accuracy=P%eps_teint)
      ! 2nd transformation A (rs|tc) -> (rs|dc)
      !$omp parallel default(none) shared(nc, nij, ni, LCact, temp, temp2) private(rs, c, d, cd)
      !$omp do
      do rs = 1, nij
        temp2(:, rs) = ZERO
        do c = 1, nc
          do d = 1, c
            cd = index_packed(c, d)
            temp2(cd, rs) = sum(temp(c, :, rs) * LCact(:, d))
          end do! d
        end do! c
      end do! rs
      !$omp end do
      !$omp end parallel
      deallocate (temp)
    else
      !$omp parallel default(none) shared(nc, nij, ni, LCact, Ba, temp2) private(rs, cd, tu, c, d, t, u, rstu, temp3)
      allocate(temp3(nc,ni))
      !$omp do
      do rs = 1, nij
        temp3(:,:) = ZERO
        do t = 1, ni
          do c = 1, nc
            do u = 1, ni
              tu = index_packed(t, u)
              rstu = index_packed(rs, tu)
              temp3(c, t) = temp3(c, t) + Ba%teint(rstu) * LCact(u,c)
            end do! u
          end do! c
        end do! t
        ! 2nd transformation A (rs|tc) -> (rs|dc)
        do c = 1, nc
          do d = 1, c
            cd = index_packed(c, d)
            temp2(cd, rs) = sum(temp3(c, :) * LCact(:, d))
          end do! d
        end do! c
      end do! rs
      !$omp end do
      deallocate(temp3)
      !$omp end parallel
    end if
    ! transpose the transformed <-> untransformed
    allocate(temp2T(size(temp2,2), size(temp2,1)))
    temp2T = transpose(temp2)
    deallocate(temp2)
    ! 3rd transformation A (cd|rs) -> (cd|rb)
    F%teint_abcd(:) = ZERO
    !$omp parallel default(none) shared(nc, ncd, ni, LCact, Ba, F, temp2T) private(rs, ab, cd, abcd, a, b, r, s, temp3)
    allocate(temp3(ni,nc))
    !$omp do
    do cd = 1, ncd
      temp3(:,:) = ZERO
      do r = 1, ni
        do b = 1, nc
          do s = 1, ni
            rs = index_packed(r, s)
            temp3(r, b) = temp3(r, b) + temp2T(rs, cd) * LCact(s, b)
          end do !s
        end do! b
      end do! r
      ! 4th transformation  (cd|rb) -> (cd|ab)
      do a = 1, nc
        do b = 1, a
          ab = index_packed(a, b)
          if (ab .lt. cd) cycle
          abcd = index_packed(ab, cd)
          F%teint_abcd(abcd) = sum(temp3(:, b) * LCact(:, a))
        end do! b
      end do! a
    end do! cd
    !$omp end do
    deallocate(temp3)
    !$omp end parallel
    deallocate (temp2T)
    call check_time(int_time, timer)
    time_done = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","time",f15.3,"sec ")') time_done - time_init
  end subroutine transform_twoel_allactive


  subroutine transform_twoel_two_core(P, B, F, core, OUTPUT)
    !! transform 2el matrix elements into MO space
    !! only those of type (cc|ij) and (ci|cj)
    !! where two electrons are member of the array core    
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: core(:)
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: temp(:, :, :)
    real(long), allocatable :: temp2(:, :)
    real(long), allocatable :: temp3(:, :)
    real(long), allocatable :: temp4(:, :, :)
    ! index a,b is for active
    ! index i,j is for general
    ! index r,s,t,u is for basis functions
    integer :: i, j, ni
    integer :: c, nc
    integer(8) :: ij
    integer(8) :: ci
    integer(8) :: cicj, ccij
    integer :: r, s, t, u
    integer(8) :: rs, tu, rstu, rt
    integer(8) :: nij, nci
    integer(8) :: nccij, ncicj
    real :: timer
    real :: time_init, time_done
    integer*8 :: int_time      ! timer

    call check_time(int_time, timer)
    time_init = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space two core")')

    nc = size(core)
    ni = B%N_basis
    nij = ni*(ni + 1)/2
    nci = nc*ni
    nccij = nc*nij
    ncicj = nc*nij
    if (allocated(F%teint_ccij) .and. &
        size(F%teint_ccij, 1) .ne. nccij) deallocate (F%teint_ccij)
    if (.not. allocated(F%teint_ccij)) allocate (F%teint_ccij(nccij))
    if (allocated(F%teint_cicj) .and. &
        size(F%teint_cicj, 1) .ne. ncicj) deallocate (F%teint_cicj)
    if (.not. allocated(F%teint_cicj)) allocate (F%teint_cicj(ncicj))
    F%teint_ccij(:) = ZERO
    F%teint_cicj(:) = ZERO
    ! ! first transformation (rs|tu) -> (rs|tc)
    allocate (temp(nc, ni, nij))
    temp(:, :, :) = ZERO
    do rs = 1, nij
      do t = 1, ni
        do c = 1, nc
          ci = core(c)
          temp(c, t, rs) = ZERO
          do u = 1, ni
            tu = index_packed(t, u)
            rstu = index_packed(tu, rs)
            temp(c, t, rs) = temp(c, t, rs) + B%teint(rstu)*F%LC(u, ci)
          end do! u
        end do! c
      end do! t
    end do! rs
    ! 2nd transformation A (rs|tc) -> (rs|cc)
    allocate (temp2(nc, nij))
    temp2(:, :) = ZERO
    do r = 1, ni
      do s = 1, r
        rs = index_packed(r, s)
        do c = 1, nc
          temp2(c, rs) = ZERO
          do t = 1, ni
            temp2(c, rs) = temp2(c, rs) + temp(c, t, rs)*F%LC(t, core(c))
          end do! t
        end do! c
      end do! s
    end do! r
    ! 2nd transformation B (sr|ct) -> (ct|cr)
    allocate (temp3(nc, nij))
    temp3(:, :) = ZERO
    do r = 1, ni
      do t = 1, ni
        rt = index_packed(r, t)
        do c = 1, nc
          temp3(c, rt) = ZERO
          do s = 1, ni
            rs = index_packed(r, s)
            temp3(c, rt) = temp3(c, rt) + temp(c, t, rs)*F%LC(s, core(c))
          end do! s
        end do ! c
      end do! t
    end do! r
    deallocate (temp)
    ! 3rd transformation A (cc|rs) -> (cc|rj)
    allocate (temp(ni, ni, nc))
    temp(:, :, :) = ZERO
    do c = 1, nc
      do r = 1, ni
        do j = 1, ni
          temp(r, j, c) = ZERO
          do s = 1, ni
            rs = index_packed(r, s)
            temp(r, j, c) = temp(r, j, c) + temp2(c, rs)*F%LC(s, j)
          end do !s
        end do! j
      end do! r
    end do! c
    deallocate (temp2)
    ! 3rd transformation B  (ct|cr) -> (ct|cj)
    allocate (temp4(ni, ni, nc))
    temp4(:, :, :) = ZERO
    do c = 1, nc
      do t = 1, ni
        do j = 1, ni
          temp4(t, j, c) = ZERO
          do r = 1, ni
            rt = index_packed(r, t)
            temp4(t, j, c) = temp4(t, j, c) + temp3(c, rt)*F%LC(r, j)
          end do! r
        end do! j
      end do! t
    end do! c
    deallocate (temp3)
    ! 4th transformation  (cc|rj) -> (cc|ij)
    F%teint_ccij(:) = ZERO
    do c = 1, nc
      do i = 1, ni
        do j = 1, i
          ij = index_packed(i, j)
          ccij = (c - 1)*nij + ij
          F%teint_ccij(ccij) = ZERO
          do r = 1, ni
            F%teint_ccij(ccij) = F%teint_ccij(ccij) + temp(r, j, c)*F%LC(r, i)
          end do! s
        end do! j
      end do! i
    end do! c
    deallocate (temp)
    ! 4th transformation B   (ct|cj)  -> (ci|cj)
    F%teint_cicj(:) = ZERO
    do c = 1, nc
      do i = 1, ni
        do j = 1, ni
          ij = index_packed(i, j)
          cicj = (c - 1)*nij + ij
          F%teint_cicj(cicj) = ZERO
          do t = 1, ni
            F%teint_cicj(cicj) = F%teint_cicj(cicj) + temp4(t, j, c)*F%LC(t, i)
          end do! t
        end do! j
      end do! i
    end do! c
    deallocate (temp4)
    call check_time(int_time, timer)
    time_done = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","time",f15.3,"sec ")') time_done - time_init
  end subroutine transform_twoel_two_core


  subroutine transform_twoel_two_active(P, B, F, active, OUTPUT)
    !! transform 2el matrix elements into MO space
    !! only those where two electrons are member of the array active    
    type(Param), intent(in) :: P
    type(Basis), intent(inout) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: active(:)
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: temp(:, :, :), tempM(:, :)
    real(long), allocatable :: temp2A(:, :), temp2Ap(:, :)
    real(long), allocatable :: temp3B(:, :, :), temp3(:), temp4(:,:)
    ! index a,b is for active
    ! index i,j is for general
    ! index r,s,t,u is for basis functions
    integer :: i, j, ni
    integer(8) :: ij
    integer :: c, d, nc
    integer(8) :: ci, dj, cd
    integer(8) :: cidj, cdij
    integer :: r, s, t
    integer(8) :: rs
    integer(8) :: ct, dr, ctdr
    integer(8) ::  nij, ncd, nci
    integer(8) :: ncdij, ncidj

    real(long) :: LCactT(size(active), size(F%LC, 1))
    real(long) :: LCact(size(F%LC, 1), size(active))

    real :: timer, timer_step
    real :: time_init, time_done
    integer*8 :: int_time      ! timer

    LCact = F%LC(:, active(:))
    LCactT = transpose(LCact)

    call check_time(int_time, timer)
    time_init = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space two active")')
    nc = size(active)
    ncd = nc*(nc + 1)/2
    ni = B%N_basis
    nij = ni*(ni + 1)/2
    nci = nc*ni
    ncdij = ncd*nij
    ncidj = nci*(nci + 1)/2
    if (.not. allocated(F%teint_abij)) allocate (F%teint_abij(ncdij))
    if (.not. allocated(F%teint_aibj)) allocate (F%teint_aibj(ncidj))
    F%teint_abij(:) = ZERO
    F%teint_aibj(:) = ZERO
    if (size(active) .eq. 0) return

    ! first transformation (rs|tu) -> (rs|tc)
    ! For each rs tuple, expand tu-packed storage into tu-full storage
    ! and perform matrix-matrix multiplication
    allocate (temp(nc, ni, nij))
    temp(:, :, :) = ZERO
    if (P%yes_direct_SCF) then
      call calculate_1contracted_teints_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, B%N_basis, &
                                               LCact, temp, accuracy=P%eps_teint)
    else
      call contraction_ijkl2ijka(P, B%teint, LCact(:, :), ni, temp)
    end if
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","1st step ",f15.3,x,"sec")') timer - time_init
    timer_step = timer

    ! 2nd transformation A (rs|tc) -> (rs|dc)
    allocate (temp2A(nc*nc, nij))
    call contraction_ijka2ijba(P, temp, LCact(:, :), temp2A)
    ! A: go back to packed storage
    allocate (temp2Ap(ncd, nij))
    temp2Ap(:, :) = ZERO
    
    !$omp parallel default(none) shared(temp2Ap, temp2A, nc, nij) private(rs, cd, c, d)
    !$omp do
    do rs=1,nij
      cd = 0
      do c = 1, nc
        do d = 1, c
          cd = cd + 1
          temp2Ap(cd, rs) = temp2A(d + (c - 1) * nc, rs)
        end do! d
      end do! c
    end do !rs
    !$omp end do
    !$omp end parallel
    deallocate (temp2A)

    ! allocate (temp2Ap(ncd, nij))
    ! temp2Ap(:, :) = ZERO
    ! do r = 1, ni
    !   do s = 1, r
    !     rs = index_packed(r, s)
    !     do c = 1, nc
    !       do d = 1, c
    !         cd = index_packed(c, d)
    !         temp2Ap(cd, rs) = ZERO
    !         do t = 1, ni
    !           temp2Ap(cd, rs) = temp2Ap(cd, rs) + temp(c, t, rs)*F%LC(t, active(d))
    !         end do! t
    !       end do! d
    !     end do! c
    !   end do! s
    ! end do! r
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","2nd  stepA ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    allocate (temp3(ncidj))
    ! temp3(:) = ZERO
    ! do c = 1, nc
    !   do t = 1, ni
    !     ct = (c - 1)*ni + t
    !     do d = 1, nc
    !       do r = 1, ni
    !         dr = (d - 1)*ni + r
    !         ctdr = index_packed(ct, dr)
    !         if(ct>dr) cycle
    !         temp3(ctdr) = ZERO
    !         do s = 1, ni
    !           rs = index_packed(r, s)
    !           temp3(ctdr) = temp3(ctdr) + temp(c, t, rs)*F%LC(s, active(d))
    !         end do!s
    !       end do!r
    !     end do ! d
    !   end do! t
    ! end do! c

    allocate (temp3B(nc, ni, nc*ni))
    !$omp parallel default(none) shared(temp, temp3B, temp3, LCactT, ni, nc) private(ct,rs,dr,c,t,r,s,ctdr,tempM)
    allocate (tempM(ni, ni))
    !$omp do schedule(static) collapse(2)
    do c = 1, nc
      do t = 1, ni
        ct = (c - 1)*ni + t
        do r = 1, ni
          do s = 1, ni
            rs = index_packed(r, s)
            tempM(r, s) = temp(c, t, rs)
            tempM(s, r) = tempM(r, s)
          end do
        end do
        temp3B(:, :, ct) = LCactT(:, :) .m.tempM(:, :)
      end do! ct
    end do
    !$omp end do
    deallocate (tempM)
    !$omp end parallel

    ! ! go back to packed storage
    do c = 1, nc
      do t = 1, ni
        ct = (c - 1)*ni + t
        do d = 1, nc
          do r = 1, ni
            dr = (d - 1)*ni + r
            if (ct .gt. dr) cycle
            ctdr = index_packed(ct, dr)
            temp3(ctdr) = temp3B(d, r, ct)
          end do
        end do
      end do
    end do
    deallocate (temp3B)

    deallocate (temp)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","2nd  stepB ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 3rd transformation A (cd|rs) -> (cd|rj)
    allocate (temp(ni, ni, ncd))
    !allocate (temp2ApT( nij, ncd))
    !temp2ApT = transpose(temp2Ap)
    !deallocate(temp2Ap)
    !allocate (temp3A(ni, ni, ncd))
    !call contraction_ijab2icab(P, temp2ApT, F%LC, ni, temp3A)
    !deallocate(temp2ApT)

    temp(:, :, :) = ZERO
    do c = 1, nc
      do d = 1, c
        cd = index_packed(c, d)
        do r = 1, ni
          do j = 1, ni
            temp(r, j, cd) = ZERO
            do s = 1, ni
              rs = index_packed(r, s)
              temp(r, j, cd) = temp(r, j, cd) + temp2Ap(cd, rs)*F%LC(s, j)
            end do !s
          end do! j
        end do! r
      end do! d
    end do! c
    deallocate (temp2Ap)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","3rd  stepA ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 3rd transformation B  (ct|dr) -> (ct|dj)
    ! transpose temp2B
    ! allocate (temp2BT(nij, ncd))
    ! temp2BT = transpose(temp2B)
    ! deallocate(temp2B)
    ! allocate (temp3B(nc, ni, nc*ni))
    ! call contraction_ijab2icab(P, temp2BT, LCact, ni, temp3B)
    ! deallocate (temp2BT)

    allocate (temp4(nci, nci))
    temp4(:, :) = ZERO
    do c = 1, nc
      do t = 1, ni
        ct = (c - 1)*ni + t
        do d = 1, nc
          do j = 1, ni
            dj = (d - 1)*ni + j
            temp4(ct, dj) = ZERO
            do r = 1, ni
              dr = (d - 1)*ni + r
              ctdr = index_packed(ct, dr)
              temp4(ct, dj) = temp4(ct, dj) + temp3(ctdr)*F%LC(r, j)
            end do! r
          end do! j
        end do! d
      end do! t
    end do! c
    deallocate (temp3)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","3rd  stepB ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 4th transformation A (ab|rj) -> (ab|ij)
    ! allocate (temp4A(ni*ni, ncd))
    ! call contraction_icab2cdab(P, temp3A, F%LC, ni, temp4A)
    ! deallocate (temp3A)
    ! ! go back to packed storage
    ! F%teint_abij(:) = ZERO
    ! do cd = 1, ncd
    !   do i = 1, ni
    !     do j = 1, i
    !       ij = index_packed(i, j)
    !       cdij = ij + (cd - 1)*nij
    !       F%teint_abij(cdij) = temp4A((i - 1)*ni + j, cd)
    !     end do! i
    !   end do! j
    ! end do! c
    ! deallocate (temp4A)

    F%teint_abij(:) = ZERO
    do c = 1, nc
      do d = 1, c
        cd = index_packed(c, d)
        do i = 1, ni
          do j = 1, i
            ij = index_packed(i, j)
            cdij = (cd - 1)*nij + ij
            F%teint_abij(cdij) = ZERO
            do r = 1, ni
              F%teint_abij(cdij) = F%teint_abij(cdij) + temp(r, j, cd)*F%LC(r, i)
            end do! s
          end do! i
        end do! j
      end do! d
    end do! c
    deallocate (temp)
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","4th  stepA ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 4th transformation B   (ct|dj)  -> (ci|dj)

    ! allocate (temp4B(nc*ni, nc*ni))
    ! call contraction_icab2cdab(P, temp3B, F%LC, ni, temp4B)
    ! deallocate (temp3B)
    ! ! go back to packed storage
    ! F%teint_aibj(:) = ZERO
    ! do j = 1, ni
    !   do d = 1, nc
    !     dj = (d - 1)*ni + j
    !     do i = 1, ni
    !       do c = 1, nc
    !         ci = (c - 1)*ni + i
    !         if (dj > ci) cycle
    !         cidj = index_packed(ci, dj)
    !         F%teint_aibj(cidj) = temp4B((i - 1)*nc + c, (j - 1)*nc + d)
    !       end do! i
    !     end do! c
    !   end do! d
    ! end do
    ! deallocate(temp4B)

    do c = 1, nc
      do i = 1, ni
        ci = (c - 1)*ni + i
        do d = 1, nc
          do j = 1, ni
            dj = (d - 1)*ni + j
            if (dj .gt. ci) cycle
            cidj = index_packed(dj, ci)
            F%teint_aibj(cidj) = ZERO
            do t = 1, ni
              ct = (c - 1)*ni + t
              F%teint_aibj(cidj) = F%teint_aibj(cidj) + temp4(ct, dj)*F%LC(t, i)
            end do! t
          end do! j
        end do! d
      end do! i
    end do! c
    deallocate (temp4)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","4th  stepB ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    call check_time(int_time, timer)
    time_done = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","time",f15.3,"sec ")') time_done - time_init
  end subroutine transform_twoel_two_active

  subroutine transform_twoel_three_core_one_virtual(P, B, F, core, virtual, OUTPUT)
    !! transform all 2el matrix elements into MO space
    !! only those where two electrons are member of the array core and two electrons are in the array virtual (as required for CIS)
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: core(:)
    integer, intent(in) :: virtual(:)
    integer, intent(in) :: OUTPUT

    integer :: i
    integer :: c, d, e
    integer(8) :: ie, cd
    integer :: r, s, t, u, ni, nc
    integer(8) :: rs, tu, rstu
    integer(8) :: cdei
    integer(8) :: ncd, nci
    integer(8) :: ncdei, nn
    integer :: n
    real(long) :: LCcoreT(size(core), size(F%LC, 1))
    real(long) :: LCvirtualT(size(virtual), size(F%LC, 1))
    real(long) :: LCcore(size(F%LC, 1), size(core))
    real(long) :: LCvirtual(size(F%LC, 1), size(virtual))
    real(long), allocatable :: temp(:, :, :)
    real(long), allocatable :: temp2(:, :)

    real :: timer, timer_step
    real :: time_init
    integer*8 :: int_time      ! timer

    if (size(core) == 0) return
    if (size(virtual) == 0) return

    LCcore = F%LC(:, core(:))
    LCvirtual = F%LC(:, virtual(:))
    LCcoreT = transpose(LCcore)
    LCvirtualT = transpose(LCvirtual)

    n = B%N_basis
    nn = n*(n + 1)/2
    nc = size(core)
    ncd = nc*(nc + 1)/2
    ni = size(virtual)
    nci = nc*ni
    ncdei = ncd*nci

    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space three core one virtual")')

    if (allocated(F%teint_cdei)) deallocate (F%teint_cdei)
    allocate (F%teint_cdei(ncdei))
    F%teint_cdei(:) = ZERO

    call check_time(int_time, timer)
    time_init = timer
    timer_step = timer

    ! first transformation (rs|tu) -> (rs|tc)
    allocate (temp(nc, n, nn))

    if (P%yes_direct_scf) then
      call calculate_1contracted_teints_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, B%N_basis, &
                                               LCcore, temp, accuracy=P%eps_teint)
    else
      temp(:, :, :) = ZERO
      !$omp parallel default(none) shared(B, temp, LCcoreT, nn, n) private(rs, t, u, tu, rstu, temp2)
      allocate (temp2(n, n))
      !$omp do schedule(static)
      do rs = 1, nn
        do t = 1, n
          do u = 1, t
            tu = index_packed(t, u)
            rstu = index_packed(rs, tu)
            temp2(t, u) = B%teint(rstu)
            temp2(u, t) = B%teint(rstu)
          end do ! u
        end do ! t
        temp(:, :, rs) = LCcoreT(:, :) .m. temp2(:, :)
      end do! rs
      !$omp end do
      deallocate (temp2)
      !$omp end parallel
    end if

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","1st step ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 2nd transformation (rs|tc) -> (rs|dc)
    allocate (temp2(ncd, nn))
    temp2(:, :) = ZERO
    !$omp parallel default(none) shared(temp2, temp, n, nc, ncd, core, F) private(rs, s, r, cd)
    !$omp do schedule(static)
    do r = 1, n
      do s = 1, r
        rs = index_packed(r, s)
        do c = 1, nc
          do d = 1, c
            cd = index_packed(c, d)
            temp2(cd, rs) = ZERO
            do t = 1, n
              temp2(cd, rs) = temp2(cd, rs) + temp(c, t, rs)*F%LC(t, core(d))
            end do! t
          end do! d
        end do! c
      end do! s
    end do! r
    !$omp end do
    !$omp end parallel
    deallocate (temp)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","2nd step ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 3rd transformation (cd|rs) -> (cd|re)
    allocate (temp(n, nc, ncd))
    temp(:, :, :) = ZERO
    !do c = 1, nc
      !do d = 1, c
        !cd = index_packed(c, d)
    !$omp parallel default(none) shared(temp2, temp, n, nc, ncd, core, F) private(rs, e, s, r, cd)
    !$omp do schedule(static)
    do cd = 1, ncd
      do r = 1, n
        do e = 1, nc
          do s = 1, n
            rs = index_packed(r, s)
            temp(r, e, cd) = temp(r, e, cd) + temp2(cd, rs)*F%LC(s, core(e))
          end do !s
        end do! e
      end do! r
    end do !cd
    !$omp end do
    !$omp end parallel
    !end do! d
    !end do! c
    deallocate (temp2)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","3rd step ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

    ! 4th transformation (cd|rs) -> (cd|re)
    F%teint_cdei(:) = ZERO
    do c = 1, nc
      do d = 1, c
        cd = index_packed(c, d)
        do e = 1, nc
          do i = 1, ni
            ie = (i + (e - 1)*ni)
            cdei = ie + (cd - 1)*nci
            do r = 1, n
              F%teint_cdei(cdei) = F%teint_cdei(cdei) + temp(r, e, cd)*F%LC(r, virtual(i))
            end do !r
          end do! i
        end do! e
      end do! d
    end do! c
    deallocate (temp)

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","4th step ",f15.3,x,"sec")') timer - timer_step
    timer_step = timer

  end subroutine transform_twoel_three_core_one_virtual

  subroutine contraction_ijkl2ijka(P, teint, LC, n, teint_ijka)
    !! contracts a packed storage teints set with dimension n along the first coordinate with coefficients LC
    !! $$(ij| ka) = \sum_l C_{la} (ij| kl)$$
    type(Param), intent(in) :: P
    real(long), intent(in) :: teint(:)
    real(long), intent(in) :: LC(:, :)
    integer, intent(in) :: n
    real(long), intent(out) :: teint_ijka(size(LC, 2), n, n*(n + 1)/2)
    integer :: k, l
    integer(8) :: ij, kl, ijkl, nn
    real(long) :: LCT(size(LC, 2), size(LC, 1))
    real(long), allocatable :: temp(:, :)
    real :: timer_start, timer_step
    integer*8 :: int_time      ! timer
    if (P%yes_verbose) &
      call check_time(int_time, timer_start)
    nn = n*(n + 1)/2
    LCT = transpose(LC)

    !$omp parallel do default(none) shared(teint_ijka, LCT, teint,n,nn) private(kl,ij,k,l,ijkl,temp) schedule(static)
    do ij = 1, nn
      allocate (temp(n, n))
      kl = 0
      do k = 1, n
        do l = 1, k
          kl = kl + 1
          ijkl = index_packed(ij, kl)
          temp(k, l) = teint(ijkl)
          temp(l, k) = temp(k, l)
        end do
      end do
      teint_ijka(:, :, ij) = LCT(:, :) .m.temp(:, :)
      deallocate (temp)
    end do
    !$omp end parallel do

    if (P%yes_verbose) then
      call check_time(int_time, timer_step)
      write (P%OUTPUT, '("# ",4x,"(ijkl)->(ijka) ",f15.3,x,"sec")') timer_step - timer_start
    end if
  end subroutine contraction_ijkl2ijka

  subroutine contraction_ijka2ijba(P, teint_ijka, LC, teint_ijba)
    !! contracts a packed storage teints set with dimension n along the first coordinate with coefficients LC
    !! $$(ab |ij) = \sum_l C_{kb} (ka|ij)$$
    type(Param), intent(in) :: P
    real(long), intent(in) :: teint_ijka(:, :, :)
    real(long), intent(in) :: LC(:, :)
    real(long), intent(out) :: teint_ijba(size(teint_ijka, 1)*size(LC, 2), size(teint_ijka, 3))
    integer(8) :: ij, nn
    real :: timer_start, timer_step
    integer*8 :: int_time      ! timer
    if (size(teint_ijka, 2) .ne. size(LC, 1) ) stop 'error'
    if (P%yes_verbose) &
      call check_time(int_time, timer_start)
    nn = size(teint_ijka, 3)

    !$omp parallel workshare
    forall (ij=1:nn)
      teint_ijba(:, ij) = pack(matmul(teint_ijka(:, :, ij), LC), .TRUE.)
    end forall
    !$omp end parallel workshare

    if (P%yes_verbose) then
      call check_time(int_time, timer_step)
      write (P%OUTPUT, '("# ",4x,"(ijka)->(ijba) ",f15.3,x,"sec")') timer_step - timer_start
    end if
  end subroutine contraction_ijka2ijba

  !> contracts a packed storage teints set with dimension n along the first coordinate with coefficients LC
  !! $$ (ic| ab) = \sum_j C_{jc} (ij| ab) $$
  !!
  subroutine contraction_ijab2icab(P, teint_ijab, LC, n, teint_icab)
    type(Param), intent(in) :: P
    real(long), intent(in) :: teint_ijab(:, :)
    real(long), intent(in) :: LC(:, :)
    integer, intent(in) :: n
    real(long), intent(out) :: teint_icab(size(LC, 2), n, size(teint_ijab, 2))
    integer :: k, l
    integer(8) :: ab, kl
    real(long) :: LCT(size(LC, 2), size(LC, 1))
    real(long), allocatable :: temp(:, :)
    real :: timer_start, timer_step
    integer*8 :: int_time      ! timer
    if (P%yes_verbose) &
      call check_time(int_time, timer_start)
    LCT = transpose(LC)
    !$omp parallel default(none) shared(teint_icab, LCT, teint_ijab,n) private(kl,ab,k,l,temp)
    allocate (temp(n, n))
    !$omp do
    do ab = 1, size(teint_ijab, 2)
      kl = 0
      do k = 1, n
        do l = 1, k
          kl = kl + 1
          temp(k, l) = teint_ijab(kl, ab)
          temp(l, k) = temp(k, l)
        end do
      end do
      teint_icab(:, :, ab) = LCT(:, :) .m.temp(:, :)
    end do
    !$omp end do
    deallocate (temp)
    !$omp end parallel
    if (P%yes_verbose) then
      call check_time(int_time, timer_step)
      write (P%OUTPUT, '("# ",4x,"(ijab)->(icab) ",f15.3,x,"sec")') timer_step - timer_start
    end if
  end subroutine contraction_ijab2icab

  subroutine contraction_icab2cdab(P, teint_icab, LC, teint_cdab)
    !! contracts a packed storage teints set with dimension n along the first coordinate with coefficients LC
    !! $$ (cd |ab) = \sum_i C_{id} (ic|ab)$$
    type(Param), intent(in) :: P
    real(long), intent(in) :: teint_icab(:, :, :)
    real(long), intent(in) :: LC(:, :)
    real(long), intent(out) :: teint_cdab(size(teint_icab, 1)*size(LC, 2), size(teint_icab, 3))
    integer(8) :: ij
    real :: timer_start, timer_step
    integer*8 :: int_time      ! timer
    if (size(teint_icab, 2) .ne. size(LC, 1) &
        ) stop 'error'
    if (P%yes_verbose) &
      call check_time(int_time, timer_start)
    !$omp parallel workshare
    forall (ij=1:size(teint_icab, 3))
      teint_cdab(:, ij) = pack(matmul(teint_icab(:, :, ij), LC), .TRUE.)
    end forall
    !$omp end parallel workshare
    if (P%yes_verbose) then
      call check_time(int_time, timer_step)
      write (P%OUTPUT, '("# ",4x,"(icab)->(cdab) ",f15.3,x,"sec")') timer_step - timer_start
    end if 
  end subroutine contraction_icab2cdab

  !> subroutine to transform all 2el matrix elements into MO space
  !! where two electrons are member of the array core and two electrons are in the array virtual (as required for CIS)
  subroutine transform_twoel_two_core_two_virtual(P, B, F, core, virtual, OUTPUT)
    !use ifport
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: core(:)
    integer, intent(in) :: virtual(:)
    integer, intent(in) :: OUTPUT
    real(long), allocatable :: temp(:, :, :), temp2(:, :), temp2A(:, :), temp2B(:, :), temp2AT(:, :)
    real(long), allocatable :: temp3A(:, :, :), temp4A(:, :), temp3B(:, :, :), temp4B(:, :)

    real(long) :: LCcore(size(F%LC, 1), size(core))
    real(long) :: LCvirtual(size(F%LC, 1), size(virtual))
    ! index a,b is for core
    ! index i,j is for virtual
    ! index r,s,t,u is for basis functions

    integer :: i, j, n
    integer :: c, d
    integer(8) :: ij, cd, ci, dj
    integer(8) :: cidj, cdij
    integer :: ni, nc
    integer(8) :: rs
    integer(8) :: nij, ncd, nci
    integer(8) :: ncdij, ncidj, nn, ncrds, ncr
    real :: timer, timer_step
    real :: time_init, time_done
    integer*8 :: int_time      ! timer

    call check_time(int_time, timer)
    time_init = timer
    timer_step = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","transforming two-electron integrals into MO space two core two virtual")')
    n = B%N_basis
    nn = n*(n + 1)/2
    nc = size(core)
    ncd = nc*(nc + 1)/2
    ni = size(virtual)
    nij = ni*(ni + 1)/2
    nci = nc*ni
    ncr = nc*n
    ncdij = ncd*nij
    ncidj = nci*(nci + 1)/2
    ncrds = ncr*(ncr + 1)/2
    if (allocated(F%teint_cdij)) deallocate (F%teint_cdij)
    if (allocated(F%teint_cidj)) deallocate (F%teint_cidj)
    allocate (F%teint_cdij(ncdij))
    allocate (F%teint_cidj(ncidj))
    if(nc==0 .or. ni==0)&
      return

    F%teint_cdij(:) = ZERO
    F%teint_cidj(:) = ZERO

    LCcore = F%LC(:, core(:))
    LCvirtual = F%LC(:, virtual(:))

    ! first transformation (rs|tu) -> (rs|tc)
    ! this is the most expensive step
    allocate (temp(nc, n, nn))
    if (P%yes_direct_scf) then
      call calculate_1contracted_teints_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, B%N_basis, &
                                               LCcore, temp, accuracy=P%eps_teint)
    else
      call contraction_ijkl2ijka(P, B%teint, LCcore, n, temp)
    end if
    ! A second transformation (rs|tc) -> (rs|cd)
    allocate (temp2(nc*nc, nn))
    call contraction_ijka2ijba(P, temp, LCcore, temp2)
    ! A: go back to packed storage
    allocate (temp2A(ncd, nn))
    temp2A(:, :) = ZERO
    !$omp parallel default(none) shared(temp2, temp2A, nn, n, nc) private(cd, c, d)
    !$omp do schedule(static)
    do rs=1,nn
      cd = 0
      do c = 1, nc
        do d = 1, c
          cd = cd + 1
          temp2A(cd, rs) = temp2(d + (c - 1)*nc, rs)
        end do! d
      end do! c
    end do !rs
    !$omp end do
    !$omp end parallel
    deallocate (temp2)

    ! B second transformation (rs|tc) -> (rs|ci)
    allocate (temp2(nc*ni, nn))
    call contraction_ijka2ijba(P, temp, LCvirtual, temp2)

    ! release temp
    deallocate (temp)
    ! transpose temp2A
    allocate (temp2AT(nn, ncd))
    temp2AT = transpose(temp2A)
    ! A 3rd transformation
    allocate (temp3A(ni, n, ncd))
    call contraction_ijab2icab(P, temp2AT, LCvirtual, n, temp3A)
    deallocate (temp2AT)
    ! A 4th transformation
    allocate (temp4A(ni*ni, ncd))
    call contraction_icab2cdab(P, temp3A, LCvirtual, temp4A)
    deallocate (temp3A)
    ! go back to packed storage
    F%teint_cdij(:) = ZERO
    do cd = 1, ncd
      do i = 1, ni
        do j = 1, i
          ij = index_packed(i, j)
          cdij = ij + (cd - 1)*nij
          F%teint_cdij(cdij) = temp4A((i - 1)*ni + j, cd)
        end do! i
      end do! j
    end do! c
    deallocate (temp4A)

    ! reshape B
    allocate (temp2B(nn, nc*ni))
    temp2B = transpose(temp2)
    deallocate (temp2)

    ! B 3rd transformation
    allocate (temp3B(nc, n, nc*ni))
    call contraction_ijab2icab(P, temp2B, LCcore, n, temp3B)
    deallocate (temp2B)

    ! B 4th transformation
    allocate (temp4B(nc*ni, nc*ni))
    call contraction_icab2cdab(P, temp3B, LCvirtual, temp4B)
    deallocate (temp3B)

    F%teint_cidj(:) = ZERO
    do j = 1, ni
      do d = 1, nc
        dj = (d - 1)*ni + j
        do i = 1, ni
          do c = 1, nc
            ci = (c - 1)*ni + i
            if (dj .gt. ci) cycle
            cidj = index_packed(ci, dj)
            F%teint_cidj(cidj) = temp4B((i - 1)*nc + c, (j - 1)*nc + d)
          end do! i
        end do! c
      end do! d
    end do

    call check_time(int_time, timer)
    time_done = timer
    if (P%yes_verbose) &
      write (OUTPUT, '("# ","time",f15.3,"sec ")') time_done - time_init
  end subroutine transform_twoel_two_core_two_virtual

  subroutine twoel_allactive_from_two_active(Ba, F, active)
    !! obtain all-active twoelectrons from transformed two-active two-electron integrals
    type(Basis), intent(in) :: Ba
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: active(:)
    integer :: i, j, a, b, nc, ni
    integer(8) :: ncd, nabcd, nij
    integer(8) :: ij, ab, cd, abcd, cdij
    nc = size(active)
    ncd = nc*(nc + 1)/2
    nabcd = ncd*(ncd + 1)/2
    ni = Ba%N_basis
    nij = ni*(ni + 1)/2

    if (allocated(F%teint_abcd)) then
      if (size(F%teint_abcd, 1) .ne. nabcd) deallocate (F%teint_abcd)
    end if
    if (.not. allocated(F%teint_abcd)) allocate (F%teint_abcd(nabcd))
    F%teint_abcd(:) = ZERO

    do cd = 1, ncd
      do a = 1, nc
        do b = 1, a
          ab = index_packed(a, b)
          if (ab .lt. cd) cycle
          abcd = index_packed(ab, cd)
          i = active(a)
          j = active(b)
          ij = index_packed(i, j)
          cdij = (cd - 1)*nij + ij
          F%teint_abcd(abcd) = F%teint_abij(cdij)
        end do
      end do
    end do

  end subroutine twoel_allactive_from_two_active

  subroutine dump_twoel(P, F, OUTPUT)
    !! print all two-electron integrals
    type(Param), intent(in) :: P
    type(WF_Data), intent(inout) :: F
    integer, intent(in) :: OUTPUT

    integer :: i, j, k, l
    integer(8) :: ij, kl, ijkl
    integer :: NORB, NELEC, MS2
    integer, dimension(size(F%LC,2)) :: ORBSYM
    integer :: ISYM
    ! this is a hack araound:
    ! the original data format of molpro contains
    ! a namelist formatting with an allocatable array
    ! impossible to have in f95
    ! namelist /FCI/ NORB,NELEC,MS2,ORBSYM(1:NORB),ISYM
    NORB = size(F%LC,2)
    NELEC = P%N_elec
    MS2 = 0
    ORBSYM(:) = 1
    ISYM = 1
    write (OUTPUT, "('&FCI NORB=',i3,',NELEC=',i3,',MS2=',i2,',' )") NORB, NELEC, MS2
    write (OUTPUT, "(' ORBSYM=',100(i1,',') )") ORBSYM(:)
    write (OUTPUT, "(' ISYM=',i1)") ISYM
    write (OUTPUT, "('&END')")
    !  write(OUTPUT,nml=FCI)
    ! &FCI NORB=  7,NELEC= 10,MS2= 0,
    !  ORBSYM=1,1,1,1,1,1,1,
    !  ISYM=1
    ! &END

    do i = 1, size(F%LC,2)
      do j = 1, i
        ij = index_packed(i, j)
        do k = 1, size(F%LC,2)
          do l = 1, k
            kl = index_packed(k, l)
            if (ij .lt. kl) cycle

            ijkl = index_packed(ij, kl)
            write (OUTPUT, "(es20.10,3x,4(x,i3))") F%teint(ijkl), i, j, k, l

          end do
        end do
      end do
    end do

    do i = 1, size(F%LC,2)
      do j = 1, i
        ij = index_packed(i, j)
        write (OUTPUT, "(es20.10,3x,4(x,i3))") F%H0_MO(i, j), i, j, 0, 0
      end do
    end do

    write (OUTPUT, "(es20.10,3x,4(x,i3))") P%E_nuc, 0, 0, 0, 0
  end subroutine dump_twoel

end module hartreefock
