!
! XMOLECULE electronic structure toolkit
! Copyright 2023 Ludger Inhester
! Ludger Inhester, Center for Free-Electron Laser Science, DESY
! ludger.inhester@cfel.de
!
! This file is part of XMOLECULE, which is distributed under the terms of the GNU General Public License version 3.
! See LICENSE.txt or https://www.gnu.org/licenses/gpl-3.0.html#license-text for a full license text.
!

module mcscf
  !! author: Ludger Inhester
  !!
  !! MCSCF module
  !! This is a first attempt to incorporate MCSCF into xmolecule

  use constants
  use read_param
  use wavefunction
  use atom_in_mol
  use molecular_grid
  use eigen_lapack
  use functional_xc
  use operations
  use neweffpot
  use rwfile
  use hartreefock
  use matrix
  use scf
  use averaging
  use secondorder_scf
  use anglib
  use molden
  use mergesort
  use localization
  implicit none

  type :: CSF
    !! defines a configuration state function
    integer, pointer :: dets(:, :)
    integer, pointer :: csf(:)
    integer, pointer :: occ(:)
    integer :: irrep
  end type CSF

  private
  public :: CSF, construct_csf_array, purge_csf_array, CI_procedure, MCSCF_procedure, &
    calc_CIMatrix, calc_CIMatrix_diagonal, calc_CIMatrix_matrix_multp, calc_CIMatrix_vector_multp, &
    calculate_F_inactive, calculate_F_inactive_direct, overlap_csfs, one_particle_dm_CI2, &
    occ_overlap2, overlap_operator_csfs, trans_dens_csf, reorder_RSTRCT, &
    opdm_matrix_subblock, make_nat_trans_orbitals

#ifdef __INTEL_COMPILER
   public :: findloc !! todo: move this to operations or matrix?
#endif
contains

  subroutine occ_overlap2(occ, occs, LC1, LC2, S12, overlap2)
    !! calculates squared overlaps of a configuration occ with orbital coefficients LC1 (averaged over spin)
    !! to a set of configurations occs with orbital coefficients LC2 
    integer, intent(in) :: occ(:) !! occupation numbers (1)
    real(long), intent(in) :: LC1(:, :) !! orbital coefficients (1)
    real(long), intent(in) :: LC2(:, :) !! orbital coefficients (2)
    real(long), intent(in) :: S12(:, :) !! overlap between basis functions (1)->(2)
    integer, intent(in) :: occs(:, :) !! target occupation numbers (2)
    real(long), intent(out) :: overlap2(size(occs, 2)) !! result overlap squared
    integer :: i, j
    real(long), dimension(:, :), allocatable :: Oalpha! one particle overlap matrix for the double occupied space
    real(long), dimension(:, :), allocatable :: Obeta! one particle overlap matrix for the double+single occupied space
    real(long), dimension(:, :), allocatable :: O! one particle overlap matrix all orbitals
    integer :: Nclosed, Nopen, Nalpha, Nbeta ! Number of single / double occupied MOs, alpha/beta MOs
    integer :: Nclosed2, Nopen2 ! Number of single / double occupied MOs
    integer, allocatable :: index_alpha1(:, :)
    integer, allocatable :: index_beta1(:, :)

    integer, allocatable :: index_alpha2(:)
    integer, allocatable :: index_beta2(:)
    integer, allocatable :: dets1(:, :) ! spin determinants
    integer, allocatable :: dets2(:, :) ! spin determinants
    integer:: multp
    integer :: sign1, sign2
    integer :: d1, d2
    real(long):: olap2_multp
    real(long):: olap
    integer :: N
    allocate (O(size(occ, 1), size(occ, 1)))
    ! calculate all MO overlaps
    do i = 1, size(occ, 1)
      do j = 1, size(occ, 1)
        O(i, j) = dot_product(LC1(:, i), MATMUL(S12, LC2(:, j)))
      end do
    end do
    Nclosed = count(occ(:) .eq. 2)
    Nopen = count(occ(:) .eq. 1)
    N = 0
    overlap2 = ZERO
    ! loop over all possible spin multiplicities for elec. configuration occ
    do multp = Nopen + 1, 1, -2
      olap2_multp = ZERO
      ! genealogical construction of determinants and spin configurations
      call Nopen2dets(Nopen, multp, dets1)
      !      write(6,*) 'multp',multp,'Nopen',Nopen,'n dets',size(dets1,2)
      if (size(dets1, 2) > 0) then
        if (Nopen > 0) then
          Nalpha = Nclosed + (Nopen + dets1(Nopen, 1))/2
          Nbeta = Nclosed + (Nopen - dets1(Nopen, 1))/2
        else
          Nalpha = Nclosed
          Nbeta = Nclosed
        end if
        allocate (Oalpha(Nalpha, Nalpha))
        allocate (Obeta(Nbeta, Nbeta))
        allocate (index_alpha1(Nalpha, size(dets1, 2)))
        allocate (index_beta1(Nbeta, size(dets1, 2)))
        do d1 = 1, size(dets1, 2) ! loop over all dets configurations for occ
          sign1 = 1
          call spin_string(occ, dets1(:, d1), index_alpha1(:, d1), index_beta1(:, d1), sign1)
        end do
        allocate (index_alpha2(Nalpha))
        allocate (index_beta2(Nbeta))
        do d1 = 1, size(dets1, 2) ! loop over all dets configurations for occ
          sign1 = 1
          do i = 1, size(occs, 2)
            ! loop over elec. configurations in occs
            Nclosed2 = count(occs(:, i) .eq. 2)
            Nopen2 = count(occs(:, i) .eq. 1)
            ! genealogical construction of determinants and spin configurations for occs(:,i)
            call Nopen2dets(Nopen2, multp, dets2)
            !            call spin_string(occ,dets1(:,d1),index_alpha1,index_beta1,sign1)
            do d2 = 1, size(dets2, 2) ! loop over all dets configurations for occs(:,i)
              sign2 = 1
              call spin_string(occs(:, i), dets2(:, d2), index_alpha2, index_beta2, sign2)
              !              Oalpha(1:Nalpha,1:Nalpha)=O(index_alpha1(:),index_alpha2(:))
              !              Obeta(1:Nbeta,1:Nbeta)=O(index_beta1(:),index_beta2(:))
              Oalpha(1:Nalpha, 1:Nalpha) = O(index_alpha1(:, d1), index_alpha2(:))
              Obeta(1:Nbeta, 1:Nbeta) = O(index_beta1(:, d1), index_beta2(:))
              olap = matrix_determinant(Oalpha)*matrix_determinant(Obeta)
              if (multp > 1) then
                overlap2(i) = overlap2(i) + multp*olap**2 !
              else
                overlap2(i) = overlap2(i) + olap**2 !
              end if
            end do! d2
          end do! i loop
          deallocate (dets2)
        end do ! d1
      end if! size(dets1,2)>0
      deallocate (Oalpha)
      deallocate (Obeta)
      deallocate (index_alpha1)
      deallocate (index_alpha2)
      deallocate (index_beta1)
      deallocate (index_beta2)
      deallocate (dets1)
      if (multp > 1) then
        N = N + multp*size(dets1, 2) ! count the total multiplicity of all the spin configurations for occ(:)
      else
        N = N + 1*size(dets1, 2) ! count the total multiplicity of all the spin configurations for occ(:)
      end if
    end do! multp
    ! we have summed over all initial and all final spin configurations
    ! Now, divide by number of initial spin configurations
    overlap2(:) = overlap2(:)/N
    deallocate (O)
  end subroutine occ_overlap2

  subroutine construct_csf_array(P, B, F, multp, state_irrep, occs, csf_array)
    !! construcst an array of CSF types for a given set of occupation configurations, spin multiplicity, and irreducible represention
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    type(WF_data), intent(in) :: F
    integer, intent(in) :: multp
    integer, intent(in) :: state_irrep ! is ignored if P%yes_sym=False. Use -1 for any irrep.
    integer, target, intent(in) :: occs(:, :)
    type(CSF), intent(out), allocatable :: csf_array(:)

    integer :: io, l, ll, Nopen, i, j
    integer, allocatable, target :: irrep_occs(:)
    integer, allocatable, target :: csfs(:, :)
    integer, allocatable, target :: dets(:, :)
    integer :: N_csfs

    if (P%yes_sym) &
      allocate (irrep_occs(size(occs, 2)))

    N_csfs = 0
    ! count configuration state functions
    do io = 1, size(occs, 2)
      if (P%yes_sym) &
        irrep_occs(io) = 1
      Nopen = count(occs(:, io) .eq. 1)
      if (multp - 1 > Nopen) cycle
      if (mod(multp - 1, 2) .ne. mod(Nopen, 2)) cycle
      if (P%yes_sym) then
        do i = 1, size(occs(:, io))
          if (occs(i, io) .eq. 1 .and. P%yes_sym) &
            irrep_occs(io) = B%irrep_prod(irrep_occs(io), F%irrep_MO(i))
        end do
        if (state_irrep > 0 .and. irrep_occs(io) .ne. state_irrep) cycle
      end if
      N_csfs = N_csfs + num_csf(Nopen, multp)
    end do
    ! allocate array
    allocate (csf_array(N_csfs))
    if (N_csfs .eq. 0) return

    ! fill csf arrays
    l = 1
    do io = 1, size(occs, 2)
      if (P%yes_sym) then
        if (state_irrep > 0 .and. irrep_occs(io) .ne. state_irrep) cycle
      end if
      Nopen = count(occs(:, io) .eq. 1)
      if (multp - 1 > Nopen) cycle
      if (mod(multp - 1, 2) .ne. mod(Nopen, 2)) cycle
      call Nopen2csfs(Nopen, multp, csfs)
      call Nopen2dets(Nopen, multp, dets)
      if (size(csfs, 2) > 0) then
        allocate (csf_array(l)%dets(size(dets, 1), size(dets, 2)))
        csf_array(l)%dets = dets
        ll = l
        do j = 1, size(csfs, 2)
          csf_array(l)%occ => occs(:, io)
          if (l .ne. ll) &
            csf_array(l)%dets => csf_array(ll)%dets
          allocate (csf_array(l)%csf(size(csfs, 1)))
          csf_array(l)%csf = csfs(:, j)
          l = l + 1
        end do
      end if
      deallocate (csfs)
      deallocate (dets)
    end do! io
    if (allocated(irrep_occs)) &
      deallocate (irrep_occs)
  end subroutine construct_csf_array

  subroutine purge_csf_array(csf_array)
    !! deallocates and purges array of csf structures
    type(CSF), intent(inout), allocatable :: csf_array(:)
    integer :: i, j
    do i = 1, size(csf_array)
      if (associated(csf_array(i)%dets)) then
        deallocate (csf_array(i)%dets)
        nullify (csf_array(i)%dets)
        do j = i + 1, size(csf_array)
          if (any(csf_array(i)%occ .ne. csf_array(j)%occ)) exit
          nullify (csf_array(j)%dets)
        end do
      end if
      deallocate (csf_array(i)%csf)
    end do
  end subroutine purge_csf_array

  subroutine test_CI_procedure(P, A, B, F, OUTPUT)
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(inout) :: B
    type(WF_data), intent(inout) :: F
    integer, intent(in), optional :: OUTPUT
    integer :: N_csfs, Ncsfi, Ncsfj
    integer, allocatable, target :: occs(:, :)
    integer, allocatable :: occs_temp(:, :)
    type(CSF), allocatable :: csf_array(:)
    real(long), allocatable :: CI_matrix(:, :)
    real(long), allocatable :: CI_matrix2(:, :)
    real(long), allocatable :: CI_energy(:)
    real(long), allocatable :: CI_energy2(:)
    real(long), allocatable :: CI_block(:, :)
    real(long), allocatable :: CI_block_T(:, :)
    integer :: IOUT
    integer :: i, j, k, l, ii, jj
    real :: timer
    real :: time_init, time_last_step
    integer*8 :: int_time      ! timer
    integer :: multp

    integer, allocatable :: core(:)
    integer, allocatable :: occupied(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    type(UmatrixTable) :: UT

    if (present(OUTPUT)) then
      IOUT = OUTPUT
    else
      IOUT = P%OUTPUT
    end if

    multp = P%multp
    F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)

    if (P%yes_verbose) &
      write (IOUT, '("# ","test CI procedure: ")')

    call check_time(int_time, timer)
    time_init = timer
    time_last_step = timer
    if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, occs)
    else if (allocated(P%active_space_MO) .and. allocated(P%ras)) then
      call active_space_occs(F%N_occ, P%active_space_MO, occs_temp)
      call filter_ras_occs(P%ras, occs_temp, occs)
    else if (allocated(P%active_space_MO)) then
      call active_space_occs(F%N_occ, P%active_space_MO, occs)
    else
      allocate (occs(size(F%N_occ, 1), 1))
      occs(:, 1) = F%N_occ(:)
    end if

    if (P%yes_sym .and. F%state_irrep > 0) then
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, occs, occs_temp)
      deallocate (occs)
      allocate (occs(size(occs_temp, 1), size(occs_temp, 2)))
      occs = occs_temp
      deallocate (occs_temp)
    end if

    allocate (active( &
      count((/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))))
    allocate (core( &
      count((/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))))
    allocate (virtual( &
      count((/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))))
    allocate (occupied(size(active, 1) + size(core, 1)))
    active = pack((/(j, j=1, size(occs, 1))/), &
      (/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))
    core = pack((/(j, j=1, size(occs, 1))/), &
      (/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))
    virtual = pack((/(j, j=1, size(occs, 1))/), &
      (/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))
    occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))

    call check_time(int_time, timer)
    time_last_step = timer

    allocate (F_core(B%Nr_basis, B%Nr_basis))
    !allocate (F_core(B%N_basis, B%N_basis))
    !call transform_twoel_two_active(P, B ,F, active, STDOUT )
    !call twoel_allactive_from_two_active(P, B ,F, active, STDOUT)
    call transform_twoel_allactive(P, B, F, active, STDOUT)
    call calculate_F_inactive(P, B, core, F%LC, F_core)
    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","time for preparing integrals",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    call construct_csf_array(P, B, F, F%multp, P%state_irrep, occs, csf_array)
    N_csfs = size(csf_array, 1)
    if (P%yes_verbose) &
      write (P%OUTPUT, "('# N_csfs: ',i7)") N_csfs

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","time for constructing csfs",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    allocate (CI_matrix(N_csfs, N_csfs))
    allocate (CI_energy(N_csfs))
    allocate (CI_matrix2(N_csfs, N_csfs))
    allocate (CI_energy2(N_csfs))
    CI_matrix2 = ZERO
    if (P%yes_verbose) &
      write (IOUT, '("# computing CI matrix")')
    ! compute CI matrix
    do i = 1, size(csf_array, 1)
      do j = 1, i
        if (sum(abs(csf_array(i)%occ - csf_array(j)%occ)) > 4) then
          CI_matrix(i, j) = ZERO
        else
          CI_matrix(i, j) = CI_matrix_element_CSFaa( &
            F%H0_MO(active, active), &
            F%teint_abcd, &
            csf_array(i), &
            csf_array(j), &
            active, &
            F_core(active, active), &
            E_core)
        end if
        CI_matrix(j, i) = CI_matrix(i, j)
      end do
    end do

    if (P%yes_verbose) &
      write (IOUT, '("# done")')
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","time for calculating CI matrix",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer
    call constructUMatrixTable(maxval(count(occs .eq. 1, dim=1)), F%multp, UT)

    ii = 0
    do i = 1, size(occs, 2)
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, size(occs, 2)
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then

          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            !call hamiltonian_matrix_subblock(occs(:,j), occs(:,i), multp, F%H0, F%teint, CI_block_T, OUTPUT=P%OUTPUT)
            call hamiltonian_matrix_subblock(occs(active(:), j), occs(active(:), i), F%multp, &
              F_core(active, active), F%teint_abcd, E_core, UT, CI_block_T, OUTPUT=P%OUTPUT)

            allocate (CI_block(size(CI_block_T, 2), size(CI_block_T, 1)))
            CI_block = transpose(CI_block_T)
            deallocate (CI_block_T)
          else
            !call hamiltonian_matrix_subblock(occs(:,i), occs(:,j), multp, F%H0, F%teint, CI_block, OUTPUT=P%OUTPUT)
            call hamiltonian_matrix_subblock(occs(active(:), i), occs(active(:), j), F%multp, &
              F_core(active, active), F%teint_abcd, E_core, UT, CI_block, OUTPUT=P%OUTPUT)
          end if

          if (any(abs(CI_matrix(ii + 1:ii + size(CI_block, 1), jj + 1:jj + size(CI_block, 2)) - CI_block(:, :)) > 1e-7)) then

            write (6, *) 'block'
            write (6, '(99i1)') occs(:, i)
            do k = 1, Ncsfi
              write (6, *) ii + k, trim(csfstring(csf_array(ii + k)%occ, csf_array(ii + k)%csf(:)))
            end do
            write (6, '(99i1)') occs(:, j)
            do k = 1, Ncsfj
              write (6, *) jj + k, trim(csfstring(csf_array(jj + k)%occ, csf_array(jj + k)%csf(:)))
            end do
            write (6, *) 'CI block (new)'
            do l = 1, size(CI_block, 2)
              write (6, '(i10)', advance='no') jj + l
            end do
            write (6, '()', advance='yes')
            do k = 1, size(CI_block, 1)
              write (6, '(i4,99es15.5)') ii + k, CI_block(k, :)
            end do
            write (6, *) 'CI block (should be)'
            do l = 1, size(CI_block, 2)
              write (6, '(i10)', advance='no') jj + l
            end do
            write (6, '()', advance='yes')
            do k = 1, size(CI_block, 1)
              write (6, '(i4,99es15.5)') ii + k, CI_matrix(ii + k, jj + 1:jj + size(CI_block, 2))
            end do
            do k = 1, size(CI_block, 1)
              do l = 1, size(CI_block, 2)
                if (abs(CI_matrix(ii + k, jj + l) - CI_block(k, l)) > 5e-7) then
                  if (abs(CI_matrix(ii + k, jj + l) + CI_block(k, l)) < 5e-7) then
                    write (6, *) 'sign matrix element different', ii + k, jj + l, CI_matrix(ii + k, jj + l), CI_block(k, l)
                  else
                    write (6, '(a25,i8,i8,es15.5,es15.5,es15.5)') &
                      'matrix element different', ii + k, jj + l, CI_matrix(ii + k, jj + l), &
                      CI_block(k, l), CI_matrix(ii + k, jj + l) - CI_block(k, l)
                  end if
                  write (6, '(i8,x,a)', advance="yes") ii + k, trim(csfstring(csf_array(ii + k)%occ, csf_array(ii + k)%csf(:)))
                  write (6, '(i8,x,a)', advance="yes") jj + l, trim(csfstring(csf_array(jj + l)%occ, csf_array(jj + l)%csf(:)))
                  stop 'matrix element different'
                end if
              end do
            end do
          end if
          CI_matrix2(ii + 1:ii + size(CI_block, 1), jj + 1:jj + size(CI_block, 2)) = CI_block(:, :)
          deallocate (CI_block)
        end if
        jj = jj + Ncsfj
      end do
      ii = ii + Ncsfi
    end do

    ! do i=1,size(csf_array,1)
    !   write(6,'(i10)',advance='no') i
    ! end do
    ! write(6,'()',advance='yes')
    ! do i=1,size(csf_array,1)
    !   write(6,'(i4,99f10.4)') i,CI_matrix(i,:)
    ! end do

    ! do i=1,size(csf_array,1)
    !   write(6,'(i10)',advance='no') i
    ! end do
    ! write(6,'()',advance='yes')
    ! do i=1,size(csf_array,1)
    !    write(6,'(i4,99f10.4)') i,CI_matrix2(i,:)
    ! end do

    call get_eigenvalues(CI_matrix, CI_energy)
    call get_eigenvalues(CI_matrix2, CI_energy2)

    do i = 1, size(CI_matrix, 2)
      write (IOUT, '("# state",i7,3x,"CI_energy=",f15.8)') i, CI_energy(i) + nuclear_repulsion_energy(A, P)
      write (IOUT, '("# state",i7,3x,"CI_energy=",f15.8)') i, CI_energy2(i) + nuclear_repulsion_energy(A, P)
      if (abs(CI_energy(i) - CI_energy2(i)) > 1e-7) then
        write (6, *) 'CI_energy different', abs(CI_energy(i) - CI_energy2(i))
      end if

      do j = 1, size(CI_matrix, 1)
        write (IOUT, '("# ",es20.7,x,a)', advance="yes") CI_matrix(j, i), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
        write (IOUT, '("# ",es20.7,x,a)', advance="yes") CI_matrix2(j, i), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
      end do
    end do

    deallocate (core)
    deallocate (active)
    deallocate (occupied)
    deallocate (virtual)
    deallocate (F_core)

  end subroutine test_CI_procedure

  subroutine CI_procedure(P, A, B, F, F2, OUTPUT)
    !! performs CI calculations
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Basis), intent(inout) :: B
    type(WF_data), intent(inout) :: F
    type(WF_data), intent(in), optional :: F2
    integer, intent(in), optional :: OUTPUT
    integer :: N_csfs
    integer, allocatable, target :: occs(:, :)
    integer, allocatable :: occs_temp(:, :)
    type(CSF), allocatable :: csf_array(:)
    real(long), allocatable :: csf_c(:), csf_c2(:, :)
    real(long), allocatable :: CI_matrix(:, :)
    real(long), allocatable :: CI_energy(:)
    real(long), allocatable :: opdm(:, :)
    real(long), allocatable :: csf_O(:, :)
    real(long) :: weight
    integer :: IOUT
    integer :: i, j
    real :: timer
    real :: time_init, time_done, time_last_step
    integer*8 :: int_time      ! timer
    real(long), allocatable :: CI_diagonal(:)
    integer, allocatable :: core(:)
    integer, allocatable :: occupied(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    real(long), allocatable :: F_core(:, :)
    integer, allocatable :: idx(:)
    real(long) :: E_core
    real(long) :: epsilon
    type(UmatrixTable) :: UT
    integer :: nintern
    character(len=512) :: moldenfilename ! filename
    integer :: nev
    if (present(OUTPUT)) then
      IOUT = OUTPUT
    else
      IOUT = P%OUTPUT
    end if

    F%H0_MO = F%LC(:, :) .tm. B%H0(:, :) .m.F%LC(:, :)

    if (P%yes_verbose) &
      write (IOUT, '("# ","CI procedure: ")')

    call check_time(int_time, timer)
    time_init = timer
    time_last_step = timer
    if (allocated(P%gas)) then
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, occs)
    else if (allocated(P%active_space_MO) .and. allocated(P%ras)) then
      call active_space_occs(F%N_occ, P%active_space_MO, occs_temp)
      call filter_ras_occs(P%ras, occs_temp, occs)
    else if (allocated(P%active_space_MO)) then
      call active_space_occs(F%N_occ, P%active_space_MO, occs)
    else
      allocate (occs(size(F%N_occ, 1), 1))
      occs(:, 1) = F%N_occ(:)
    end if

    if (P%yes_sym .and. F%state_irrep > 0) then
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, occs, occs_temp)
      deallocate (occs)
      allocate (occs(size(occs_temp, 1), size(occs_temp, 2)))
      occs = occs_temp
      deallocate (occs_temp)
    end if

    allocate (active( &
      count((/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))))
    allocate (core( &
      count((/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))))
    allocate (virtual( &
      count((/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))))
    allocate (occupied(size(active, 1) + size(core, 1)))
    active = pack((/(j, j=1, size(occs, 1))/), &
      (/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))
    core = pack((/(j, j=1, size(occs, 1))/), &
      (/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))
    virtual = pack((/(j, j=1, size(occs, 1))/), &
      (/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))
    occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
      (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))
    nintern = size(active)
    do j = size(active), 1, -1
      if (any(sum(occs(active(j:), :), dim=1) > 1)) exit
    end do
    nintern = j
    write (P%OUTPUT, '("#     core:",99i4)') core
    write (P%OUTPUT, '("#   active:",99i4)') active
    write (P%OUTPUT, '("# internal:",99i4)') active(:nintern)
    write (P%OUTPUT, '("# external:",99i4)') active(nintern + 1:)

    allocate (F_core(B%Nr_basis, B%Nr_basis))
    if (P%yes_direct_scf) then
      call calculate_F_inactive_direct(B, core, F%LC, F_core, P%EPS_teint)
    else
      call calculate_F_inactive(P, B, core, F%LC, F_core)
    end if

    E_core = ZERO
    do i = 1, size(core)
      E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
    end do

    write (6, '("# maximum number of open shells: ",i8)') maxval(count(occs .eq. 1, dim=1))
    call constructUMatrixTable(maxval(count(occs .eq. 1, dim=1)), F%multp, UT)
    call construct_csf_array(P, B, F, F%multp, F%state_irrep, occs, csf_array)
    N_csfs = size(csf_array, 1)
    if (P%yes_verbose) &
      write (P%OUTPUT, "('# N_csfs: ',i7)") N_csfs

    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","time for constructing csfs",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    allocate (csf_c(N_csfs))
    allocate (CI_energy(N_csfs))

    call transform_twoel_allactive(P, B, F, active(:nintern), P%OUTPUT)
    if (nintern < size(active)) then
      call transform_twoel_two_core_two_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
      call transform_twoel_three_core_one_virtual(P, B, F, active(:nintern), active(nintern + 1:), P%OUTPUT)
    end if
    call check_time(int_time, timer)
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","time for transforming teints",f15.3,"sec ")') timer - time_last_step
    time_last_step = timer

    if (P%yes_use_arpack .or. P%yes_use_davidson .or. P%yes_use_krylov) then
      allocate (CI_diagonal(N_csfs))
      call calc_CIMatrix_diagonal(UT, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        occs(active(:), :), &
        CI_diagonal, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)

      nev = max(P%nstates, 1)
      allocate (csf_c2(N_csfs, nev))
      allocate (idx(size(CI_diagonal)))
      call merge_argsort(CI_diagonal, idx)
      ! take a starting configuration with lowest diagonal element
      csf_c2 = ZERO
      j = 0
      do while (j < nev)
        j = j + 1
        csf_c2(idx(j), j) = ONE
      end do
      deallocate (idx)
      do i = 1, nev
        do j = 1, size(csf_c2, 1)
          if (csf_c2(j, i) .eq. ZERO) cycle
          write (IOUT, '("# Starting guess: ",i4,x,es20.7)', advance="yes") i, ci_diagonal(j)
          write (IOUT, '("# ",es20.7,x,a)', advance="yes") &
            csf_c2(j, i), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
        end do
      end do
      epsilon = 1e-6
      if (P%yes_use_arpack) then
        ! call arpack
#ifdef WITH_ARPACK
        call get_some_eigenvalues_arpack(N_csfs, nev, CI_energy(1:nev), csf_c2, matrix_vector_multp)
#else
        stop 'not compiled with arpack support'
#endif
      else if (P%yes_use_davidson) then
        call get_some_eigenvalues_davpack(N_csfs, nev, P%davblock, max(P%davMax, nev), &
          CI_energy(1:nev), csf_c2, ci_diagonal, matrix_matrix_multp, epsilon, OUTPUT)
      else if (P%yes_use_krylov) then
#ifdef WITH_KRYLOV        
        call get_some_eigenvalues_libKrylov(N_csfs, nev, nev, CI_energy(1:nev), csf_c2, ci_diagonal, matrix_matrix_multp)
#else
          stop "not compiled with libkrylov support"
#endif
      end if

      deallocate (CI_diagonal)

      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CI procedure: ","time for arnoldi iterations",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer

      do i = 1, nev
        weight = ZERO
        do j = 1, size(csf_c2, 1)
          if (all(csf_array(j)%occ(:) .eq. F%N_occ(:))) &
            weight = weight + csf_c2(j, i)**2
        end do
        if (IOUT > 1) then
          write (IOUT, '("# state",i7,3x,"CI_energy=",f15.8)') i, CI_energy(i) + nuclear_repulsion_energy(A, P)
          write (IOUT, '("# ","weight = ",f10.3)', advance='no') weight
          write (IOUT, '(" elec. configuration ",99i1)') F%N_occ(:)
        end if
        do j = 1, size(csf_c2, 1)
          if (csf_c2(j, i)**2 > 1e-4 .and. IOUT > 1) &
            write (IOUT, '("# ",es20.7,x,a)', advance="yes") csf_c2(j, i), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
        end do
      end do
      if (nintern < size(active)) then
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        if(allocated(F%teint_cdei)) &
          deallocate (F%teint_cdei)
      end if
      csf_c(:) = csf_c2(:, F%state)
      deallocate (csf_c2)
    else
      ! calculate full CI matrix and use mkl dsyev
      allocate (CI_matrix(N_csfs, N_csfs))
      call calc_CIMatrix(UT, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        occs(active(:), :), &
        CI_matrix, nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei)

      if (nintern < size(active)) then
        deallocate (F%teint_cdij)
        deallocate (F%teint_cidj)
        if(allocated(F%teint_cdei)) &
          deallocate (F%teint_cdei)
      end if

      if (P%yes_verbose) &
        write (IOUT, '("# done")')
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CI procedure: ","time for calculating CI matrix",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer

      if (N_csfs .eq. 1) then
        CI_energy = CI_matrix(1, 1)
        CI_matrix(1, 1) = ONE
      else
        ! solve it
        if (P%yes_verbose) &
          write (IOUT, '("# solving CI matrix")')
        call get_eigenvalues(CI_matrix, CI_energy)
        if (P%yes_verbose) &
          write (IOUT, '("# done")')

      end if
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CI procedure: ","time for solving CI matrix",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer
      do i = 1, size(CI_matrix, 2)
        weight = ZERO
        do j = 1, size(CI_matrix, 1)
          if (all(csf_array(j)%occ(:) .eq. F%N_occ(:))) &
            weight = weight + CI_matrix(j, i)**2
        end do
        if (IOUT > 1) then
          write (IOUT, '("# state",i7,3x,"CI_energy=",f15.8)') i, CI_energy(i) + nuclear_repulsion_energy(A, P)
          if (P%yes_verbose) then
            write (IOUT, '("# ","weight = ",f10.3)', advance='no') weight
            write (IOUT, '(" elec. configuration ",99i1)') F%N_occ(:)
          end if
        end if
        if (P%yes_verbose .and. IOUT > 1) then
          do j = 1, size(CI_matrix, 1)
            if (CI_matrix(j, i)**2 > 1e-4) &
              write (IOUT, '("# ",es20.7,x,a)', advance="yes") CI_matrix(j, i), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
          end do
        end if
      end do

      csf_c(:) = CI_matrix(:, F%state)
    end if

    ! if(P%yes_verbose)&
    !   write(IOUT,'("# computing CI matrix")')
    ! call calc_CIMatrix(UT, E_core, F_core(active(:),active(:)), F%teint_abcd, occs(active(:),:), CI_matrix)

    write (IOUT, *) "" 
    write (IOUT, '("# Target state",i7,3x,"CI_energy=",f15.8)') F%state, CI_energy(F%state) + nuclear_repulsion_energy(A, P)
    do j = 1, size(csf_c, 1)
      if (csf_c(j)**2 > 1e-4 .and. IOUT > 1) &
        write (IOUT, '("# ",es20.7,x,a)', advance="yes") csf_c(j), trim(csfstring(csf_array(j)%occ, csf_array(j)%csf(:)))
    end do

    if (P%yes_nat_orb) then
      call check_time(int_time, timer)
      time_last_step = timer

      allocate (opdm(B%N_basis, B%N_basis))
      call one_particle_dm_CI2(UT, occs(:, :), csf_c(:), opdm)

      call natural_orbitals(P, F, active, opdm, ZERO*F_core, F_core)
      if (P%molden_file .ne. "") then
        write (moldenfilename, '("natorb.molden")')
        call write_molden_file(P, A, B, F, moldenfilename)
      end if
      call print_orbitals(P, B, F, "natural orbitals")
      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CI procedure: ","time for opdm matrix",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer

    end if

    !max(P%nstates,1)
    if (.not. P%yes_use_arpack .and. .not. P%yes_use_davidson .and. .not. P%yes_use_krylov) then

      if (.not. allocated(F%E_ci)) then
        allocate (F%E_ci(N_csfs))
      else
        deallocate (F%E_ci)
        allocate (F%E_ci(N_csfs))
      end if
      if (.not. allocated(F%C_ci)) then
        allocate (F%C_ci(N_csfs, N_csfs))
      else
        deallocate (F%C_ci)
        allocate (F%C_ci(N_csfs, N_csfs))
      end if
      if (.not. allocated(F%O_ci)) then
        allocate (F%O_ci(N_csfs, N_csfs))
      else
        deallocate (F%O_ci)
        allocate (F%O_ci(N_csfs, N_csfs))
      end if
      if (.not. allocated(F%occs_CI)) then
        allocate (F%occs_CI(size(occs, 1), N_csfs))
      else
        deallocate (F%occs_CI)
        allocate (F%occs_CI(size(occs, 1), N_csfs))
      end if
      if (.not. allocated(F%D_CI)) then
        allocate (F%D_CI(B%N_basis, B%N_basis))
      end if
      F%C_ci(:, :) = CI_Matrix(:, :)
      F%E_ci(:) = CI_energy(:) + nuclear_repulsion_energy(A, P)
      do i = 1, size(F%C_ci, 1)
        F%occs_CI(:, i) = csf_array(i)%occ(:)
      end do
      if (allocated(opdm)) &
        F%D_CI = F%LC.m. (opdm.mt.F%LC)
      deallocate (CI_Matrix)
    end if
    deallocate (CI_energy)
    if (allocated(opdm)) &
      deallocate (opdm)

    if (present(F2) .and. F%yes_calculate_CI_overlap) then
      call check_time(int_time, timer)
      time_last_step = timer

      allocate (csf_O(size(csf_array, 1), size(csf_array, 1)))
      csf_O = ZERO
      call overlap_csfs(P, csf_array, F%LC, F2%LC, B%S, csf_O)
      !    call overlap_csfs(P,csf_array,F%LC,F2%LC,B%S,csf_O)
      do i = 1, size(F%C_CI, 1)
        do j = 1, size(F%C_CI, 2)
          F%O_CI(i, j) = dot_product(F%C_CI(:, i), MATMUL(csf_O(:, :), F2%C_CI(:, j)))
        end do
      end do
      deallocate (csf_O)

      call check_time(int_time, timer)
      if (P%yes_verbose) &
        write (P%OUTPUT, '("# CI procedure: ","time for olap matrix",f15.3,"sec ")') timer - time_last_step
      time_last_step = timer

    end if

    call purge_UMatrixTable(UT)
    call purge_csf_array(csf_array)
    deallocate (occs)

    call check_time(int_time, timer)
    time_done = timer
    if (P%yes_verbose) &
      write (P%OUTPUT, '("# CI procedure: ","total time",f15.3,"sec ")') time_done - time_init
    deallocate (F_core)
    deallocate (active)
    deallocate (virtual)
    deallocate (core)
    deallocate (occupied)

  contains
    function matrix_vector_multp(n, Cin, Cout) result(error)
      integer, intent(in) :: n
      real(long), intent(in) :: Cin(n)
      real(long), intent(out) :: Cout(n)
      integer :: error
      Cout = Cin
      call calc_CIMatrix_vector_multp(UT, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        occs(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_vector_multp

    function matrix_matrix_multp(n, m, Cin, Cout) result(error)
      integer, intent(in) :: n
      integer, intent(in) :: m
      real(long), intent(in) :: Cin(n, m)
      real(long), intent(out) :: Cout(n, m)
      integer :: error
      call calc_CIMatrix_matrix_multp(UT, E_core, F_core(active(:), active(:)), &
        F%teint_abcd, &
        occs(active(:), :), &
        Cin, &
        Cout, &
        nintern=nintern, &
        teint_iiee=F%teint_cdij, &
        teint_ieie=F%teint_cidj, &
        teint_iiie=F%teint_cdei, accuracy=P%eps_teint)
      error = 0
    end function matrix_matrix_multp

  end subroutine CI_procedure

  
  subroutine calc_CIMatrix(UT, E_core, F_core, teint_abcd, occs, CI_matrix, nintern, teint_iiee, teint_ieie, teint_iiie)
    !! calculates the configuration interaction matrix
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(in) :: E_core, F_core(:, :)
    real(long), intent(in) :: teint_abcd(:)
    integer, intent(in) :: occs(:, :)
    real(long), intent(out) :: CI_matrix(:, :)
    integer, optional, intent(in) :: nintern
    real(long), optional, intent(in) :: teint_iiee(:)
    real(long), optional, intent(in) :: teint_ieie(:)
    real(long), optional, intent(in) :: teint_iiie(:)

    integer :: i, j, ii, jj
    integer :: multp
    integer :: Ncsfi, Ncsfj
    real(long), allocatable :: CI_block(:, :), CI_block_T(:, :)

    multp = UT%multp
    CI_matrix = ZERO
    ii = 0
    do i = 1, size(occs, 2)
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, i
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then
          allocate (CI_block(Ncsfi, Ncsfj))
          CI_block = ZERO
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (CI_block_T(Ncsfj, Ncsfi))
            CI_block_T = ZERO
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, j), occs(:, i), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, &
                UT, nintern, CI_block_T)
            else
              call hamiltonian_matrix_subblock(occs(:, j), occs(:, i), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block_T)
            end if
            CI_block = transpose(CI_block_T)
            deallocate (CI_block_T)
          else
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, i), occs(:, j), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, &
                UT, nintern, CI_block)
            else
              call hamiltonian_matrix_subblock(occs(:, i), occs(:, j), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block)
            end if
          end if
          CI_matrix(ii + 1:ii + size(CI_block, 1), jj + 1:jj + size(CI_block, 2)) = CI_block(:, :)
          if (i .ne. j) &
            CI_matrix(jj + 1:jj + size(CI_block, 2), ii + 1:ii + size(CI_block, 1)) = transpose(CI_block(:, :))
          deallocate (CI_block)
        end if
        jj = jj + Ncsfj
      end do! j
      ii = ii + Ncsfi
    end do ! i
  end subroutine calc_CIMatrix

  subroutine calc_CIMatrix_diagonal(UT, E_core, F_core, teint_abcd, occs, CI_matrix_diagonal, &
    nintern, teint_iiee, teint_ieie, teint_iiie)
    !! calculates the configuration interaction matrix (only diagonal elements)
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(in) :: E_core, F_core(:, :)
    real(long), intent(in) :: teint_abcd(:)
    integer, intent(in) :: occs(:, :)
    real(long), intent(out) :: CI_matrix_diagonal(:)
    integer, optional, intent(in) :: nintern
    real(long), optional, intent(in) :: teint_iiee(:)
    real(long), optional, intent(in) :: teint_ieie(:)
    real(long), optional, intent(in) :: teint_iiie(:)
    integer :: ii_i(size(occs, 2) + 1)
    integer :: i, ii, i1
    integer :: multp
    integer :: Ncsfi
    real(long), allocatable :: CI_block(:, :)

    multp = UT%multp
    CI_matrix_diagonal = ZERO
    ! precalculate the offset index for a given occ
    ii = 0
    do i = 1, size(occs, 2) + 1
      ii_i(i) = ii
      if (i .le. size(occs, 2)) &
        ii = ii + num_csf(count(occs(:, i) .eq. 1), multp)
    end do

    !$omp parallel private(i,i1,Ncsfi,ii,ci_block) shared(ii_i,UT,occs,multp,F_core,teint_abcd,teint_iiee,teint_ieie, teint_iiie, E_core, nintern, CI_Matrix_diagonal) default(none)
    !$omp do 
    do i = 1, size(occs, 2)
      !Ncsfi = Num_csf(count(occs(:,i).eq.1), multp)
      ii = ii_i(i)
      Ncsfi = ii_i(i + 1) - ii
      if (Ncsfi .eq. 0) cycle
      allocate (CI_block(Ncsfi, Ncsfi))
      CI_block = ZERO
      if (present(nintern)) then
        call hamiltonian_matrix_subblock_int_ext(occs(:, i), occs(:, i), multp, F_core(:, :), &
          teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, UT, nintern, CI_block)
      else
        call hamiltonian_matrix_subblock(occs(:, i), occs(:, i), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block)
      end if
      forall (i1=1:Ncsfi)
        CI_matrix_diagonal(ii + i1) = CI_block(i1, i1)
      end forall
      deallocate (CI_block)
      !ii = ii + Ncsfi
    end do ! i
    !$omp end do
    !$omp end parallel

  end subroutine calc_CIMatrix_diagonal

  subroutine calc_CIMatrix_vector_multp(UT, E_core, F_core, teint_abcd, occs, C, C_out, &
    nintern, teint_iiee, teint_ieie, teint_iiie, accuracy)
    !! calculates the product of the CI matrix with a coefficient vector
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(in) :: E_core, F_core(:, :)
    real(long), intent(in) :: teint_abcd(:)
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: C(:) !! CI expansion coefficients
    real(long), intent(out) :: C_out(size(C, 1)) !! result: CI Matrix x C
    integer, optional, intent(in) :: nintern
    real(long), optional, intent(in) :: teint_iiee(:)
    real(long), optional, intent(in) :: teint_ieie(:)
    real(long), optional, intent(in) :: teint_iiie(:)
    real(long), optional, intent(in) :: accuracy
    real(long) :: epsilon
    integer :: i, j, ii, jj, i1, j1
    integer :: multp
    integer :: Ncsfi, Ncsfj
    real(long), allocatable :: CI_block(:, :), CI_block_T(:, :)

    integer :: ii_i(size(occs, 2) + 1)

    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_done
    logical :: verbose
    verbose = .TRUE.
    if (verbose) then
      call check_time(int_time, timer)            !
      time_init = timer     !
    end if

    epsilon = ZERO
    if (present(accuracy)) epsilon = accuracy

    C_out = ZERO
    multp = UT%multp

    ! precalculate the offset index for a given occ
    ii = 0
    do i = 1, size(occs, 2) + 1
      ii_i(i) = ii
      if (i .le. size(occs, 2)) &
        ii = ii + num_csf(count(occs(:, i) .eq. 1), multp)
    end do
    !$omp parallel private(i,j,i1,j1,Ncsfi,Ncsfj,ii,jj,ci_block,ci_block_T,timer,int_time) shared(verbose,ii_i,epsilon,UT,occs,C,multp,F_core,teint_abcd,teint_iiee,teint_ieie, teint_iiie, E_core, nintern) reduction(+:C_out) default(none)
    !$omp do schedule(guided)
    do i = 1, size(occs, 2)
      !Ncsfi = Num_csf(count(occs(:,i).eq.1), multp)
      ii = ii_i(i)
      Ncsfi = ii_i(i + 1) - ii
      if (Ncsfi .eq. 0) cycle
      do j = 1, i
        !Ncsfj = Num_csf(count(occs(:,j).eq.1), multp)
        jj = ii_i(j)
        Ncsfj = ii_i(j + 1) - jj
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then
          if (all(abs(C(jj + 1:jj + Ncsfj)) < epsilon) .and. all(abs(C(ii + 1:ii + Ncsfi)) < epsilon) .and. i .ne. j) cycle
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (CI_block_T(Ncsfj, Ncsfi))
            CI_block_T = ZERO
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, j), occs(:, i), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, UT, &
                nintern, CI_block_T)
            else
              call hamiltonian_matrix_subblock(occs(:, j), occs(:, i), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block_T)
            end if
            if (ii .eq. jj) then
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + CI_block_T(i1, i1)*C(ii + i1)
              !sum over j1>i1
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + sum(CI_block_T(i1 + 1:, i1)*C(jj + 1 + i1:jj + Ncsfj))
              !sum over i1<j1
              forall (j1=1:Ncsfj) &
                C_out(jj + j1) = C_out(jj + j1) + sum(CI_block_T(j1, :j1 - 1)*C(ii + 1:ii + j1 - 1))
            else
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + sum(CI_block_T(:, i1)*C(jj + 1:jj + Ncsfj))
              forall (j1=1:Ncsfj) &
                C_out(jj + j1) = C_out(jj + j1) + sum(CI_block_T(j1, :)*C(ii + 1:ii + Ncsfi))
            end if
            deallocate (CI_block_T)
          else
            allocate (CI_block(Ncsfi, Ncsfj))
            CI_block = ZERO
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, i), occs(:, j), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, &
                E_core, UT, nintern, CI_block)
            else
              call hamiltonian_matrix_subblock(occs(:, i), occs(:, j), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block)
            end if

            if (ii .eq. jj) then
              ! this means CI_block(i,j) .eq. CI_block(j,i)
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + CI_block(i1, i1)*C(ii + i1)
              ! sum over j1> i1
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + sum(CI_block(i1, i1 + 1:)*C(jj + 1 + i1:jj + Ncsfj))
              !sum over i1<j1
              forall (j1=1:Ncsfj) &
                C_out(jj + j1) = C_out(jj + j1) + sum(CI_block(:j1 - 1, j1)*C(ii + 1:ii + j1 - 1))
            else
              ! do i1=1,size(CI_block,1)
              !   do j1=1,size(CI_block,2)
              !     if(ii+i1 .le. jj+j1)cycle
              !     C_out(ii+i1) = C_out(ii+i1) + CI_block(i1,j1) * C(jj+j1)
              !     C_out(jj+j1) = C_out(jj+j1) + CI_block(i1,j1) * C(ii+i1)
              !   end do
              ! end do
              forall (i1=1:Ncsfi) &
                C_out(ii + i1) = C_out(ii + i1) + sum(CI_block(i1, :)*C(jj + 1:jj + Ncsfj))
              forall (j1=1:Ncsfj) &
                C_out(jj + j1) = C_out(jj + j1) + sum(CI_block(:, j1)*C(ii + 1:ii + Ncsfi))
            end if
            deallocate (CI_block)
          end if!
        end if! diff < 4
      end do! j
    end do ! i
    !$omp end do
    !$omp end parallel

    if (verbose) then
      call check_time(int_time, timer)
      time_done = timer
      write (6, '("# CI procedure: vector matrix multiplication ","time",f15.3,"sec ")') time_done - time_init
    end if
  end subroutine calc_CIMatrix_vector_multp

  subroutine calc_CIMatrix_matrix_multp(UT, E_core, F_core, teint_abcd, occs, C, C_out, &       
       nintern, teint_iiee, teint_ieie, teint_iiie, accuracy)
    !! calculates the matrix-matrix product of the CI matrix with coefficients
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(in) :: E_core, F_core(:, :)
    real(long), intent(in) :: teint_abcd(:)
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: C(:, :) !! CI expansion coefficient matrix
    real(long), intent(out) :: C_out(size(C, 1), size(C, 2)) !! result CI Matrix x C
    integer, optional, intent(in) :: nintern
    real(long), optional, intent(in) :: teint_iiee(:)
    real(long), optional, intent(in) :: teint_ieie(:)
    real(long), optional, intent(in) :: teint_iiie(:)
    real(long), optional, intent(in) :: accuracy
    real(long) :: epsilon
    integer :: i, j, ii, jj, i1, j1, k
    integer :: multp
    integer :: Ncsfi, Ncsfj
    real(long), allocatable :: CI_block(:, :), CI_block_T(:, :)

    integer :: ii_i(size(occs, 2) + 1)
    integer :: nvec
    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_done
    logical :: verbose
    logical :: iCoefSmallerEps

    verbose = .TRUE.
    if (verbose) then
      call check_time(int_time, timer)
      time_init = timer     !
    end if

    epsilon = ZERO
    if (present(accuracy)) epsilon = accuracy

    nvec = size(C, 2)
    C_out(:, :) = ZERO
    multp = UT%multp

    ! precalculate the offset index for a given occ
    ii = 0
    do i = 1, size(occs, 2) + 1
      ii_i(i) = ii
      if (i .le. size(occs, 2)) &
           ii = ii + UT%f(count(occs(:, i) .eq. 1))
           !ii = ii + num_csf(count(occs(:, i) .eq. 1), multp)
    end do

    ! diagonal terms (same occs)
    !$omp parallel private(k,i,j,i1,j1,Ncsfi,ii,ci_block) shared(ii_i,epsilon,UT,occs,C,nvec,multp,F_core,teint_abcd,teint_iiee,teint_ieie, teint_iiie, E_core, nintern, C_out)  default(none)
    !$omp do schedule(guided)
    diagloop: do i = 1, size(occs, 2)
      ii = ii_i(i)
      Ncsfi = ii_i(i + 1) - ii
      if (Ncsfi .eq. 0) cycle diagloop
      allocate (CI_block(Ncsfi, Ncsfi))
      CI_block = ZERO
      if (present(nintern)) then
        call hamiltonian_matrix_subblock_int_ext(occs(:, i), occs(:, i), multp, F_core(:, :), &
          teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, UT, &
          nintern, CI_block)
      else
        call hamiltonian_matrix_subblock(occs(:, i), occs(:, i), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block)
      end if
      forall (i1=1:Ncsfi, k=1:nvec) &
        C_out(ii + i1, k) = C_out(ii + i1, k) + CI_block(i1, i1)*C(ii + i1, k)
      !sum over j1>i1
      forall (i1=1:Ncsfi, k=1:nvec) &
        C_out(ii + i1, k) = C_out(ii + i1, k) + sum(CI_block(i1 + 1:, i1) * C(ii + 1 + i1:ii + Ncsfi, k))
      !sum over i1<j1
      forall (j1=1:Ncsfi, k=1:nvec) &
        C_out(ii + j1, k) = C_out(ii + j1, k) + sum(CI_block(j1, :j1 - 1) * C(ii + 1:ii + j1 - 1, k))
      deallocate(CI_block)
    end do diagloop
    !$omp end do
    !$omp end parallel

    !$omp parallel private(iCoefSmallerEps,k,i,j,i1,j1,Ncsfi,Ncsfj,ii,jj,ci_block,ci_block_T,timer,int_time) shared(verbose,ii_i,epsilon,UT,occs,C,multp,F_core,teint_abcd,teint_iiee,teint_ieie, teint_iiie, E_core, nintern,nvec, C_out)  default(none)
    !$omp do schedule(guided)
    iloop: do i = 1, size(occs, 2)
      !Ncsfi = Num_csf(count(occs(:,i).eq.1), multp)
      ii = ii_i(i)
      Ncsfi = ii_i(i + 1) - ii
      if (Ncsfi .eq. 0) cycle iloop
      jloop: do j = 1, size(occs, 2)
        if(j .eq. i) cycle
        jj = ii_i(j)
        Ncsfj = ii_i(j + 1) - jj
        if (Ncsfj .eq. 0) cycle jloop
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then
          if (all(abs(C(jj + 1:jj + Ncsfj, :)) < epsilon)) cycle jloop
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (CI_block_T(Ncsfj, Ncsfi))
            CI_block_T = ZERO
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, j), occs(:, i), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, &
                E_core, UT, nintern, CI_block_T)
            else
              call hamiltonian_matrix_subblock(occs(:, j), occs(:, i), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block_T)
            end if
            forall (i1=1:Ncsfi, k=1:nvec) &
                 C_out(ii + i1, k) = C_out(ii + i1, k) + sum(CI_block_T(:, i1)*C(jj + 1:jj + Ncsfj, k))
            deallocate (CI_block_T)
          else
            allocate (CI_block(Ncsfi, Ncsfj))
            CI_block = ZERO
            if (present(nintern)) then
              call hamiltonian_matrix_subblock_int_ext(occs(:, i), occs(:, j), multp, F_core(:, :), &
                teint_abcd, teint_iiee, teint_ieie, teint_iiie, E_core, UT, &
                nintern, CI_block)
            else
              call hamiltonian_matrix_subblock(occs(:, i), occs(:, j), multp, F_core(:, :), teint_abcd, E_core, UT, CI_block)
            end if
            forall (i1=1:Ncsfi, k=1:nvec) &
                 C_out(ii + i1, k) = C_out(ii + i1, k) + sum(CI_block(i1, :) * C(jj + 1:jj + Ncsfj, k))
            deallocate (CI_block)
          end if!
        end if! diff < 4
      end do jloop! j
    end do iloop ! i
    !$omp end do
    !$omp end parallel
    if (verbose) then
      call check_time(int_time, timer)
      time_done = timer
      write (6, '("# CI procedure: matrix matrix multiplication ","time",f15.3,"sec ")') time_done - time_init
    end if
  end subroutine calc_CIMatrix_matrix_multp

  !> calculated overlaps of configuration state functions
  !! for different orbital sets
  subroutine overlap_csfs(P, csf_array, LC1, LC2, S12, csf_O)
    type(Param), intent(in) :: P
    type(CSF), intent(in) :: csf_array(:) ! array for csfs final state
    real(long), intent(in) :: LC1(:, :)
    real(long), intent(in) :: LC2(:, :)
    real(long), intent(in) :: S12(:, :)
    real(long), intent(out) :: csf_O(:, :) ! array for csfs final state

    real(long), dimension(:, :), allocatable :: O! one particle overlap matrix all orbitals
    integer c1, c2, d1, d2
    integer, allocatable :: index_alpha(:)
    integer, allocatable :: index_beta(:)
    integer :: i, j
    real(long) :: coef, coef1, coef2
    integer :: sign1, sign2
    real(long):: det

    ! for testing
    !real(long), dimension(:,:),allocatable :: Oalpha! one particle overlap matrix for the double occupied space
    !real(long), dimension(:,:),allocatable :: Obeta! one particle overlap matrix for the double+single occupied space

    integer, allocatable :: string_start(:)
    integer, allocatable :: string_sign(:)
    real(long), pointer :: detOverlapsA(:, :)
    real(long), pointer :: detOverlapsB(:, :)

    integer, pointer :: unique_stringsA(:, :)
    integer, pointer :: unique_stringsB(:, :)

    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_step

    call check_time(int_time, timer)
    time_init = timer     !
    time_step = timer
    allocate (O(size(LC1, 2), size(LC1, 2)))
    ! calculate all MO overlaps
    do i = 1, size(LC1, 2)
      do j = 1, size(LC2, 2)
        O(i, j) = dot_product(LC1(:, i), MATMUL(S12, LC2(:, j)))
      end do
    end do

    call make_string_table(P, csf_array, unique_stringsA, unique_stringsB, index_alpha, index_beta, string_start, string_sign)
    if (.not. ASSOCIATED(unique_stringsA, unique_stringsB)) then
      allocate (detOverlapsA(size(unique_stringsA, 2), size(unique_stringsA, 2)))
      call calc_stringsOverlap(unique_stringsA(:, :), O, detOverlapsA)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for overlap strings a',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      allocate (detOverlapsB(size(unique_stringsB, 2), size(unique_stringsB, 2)))
      call calc_stringsOverlap(unique_stringsB(:, :), O, detOverlapsB)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for overlap strings b',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
    else
      allocate (detOverlapsA(size(unique_stringsA, 2), size(unique_stringsA, 2)))
      call calc_stringsOverlap(unique_stringsA(:, :), O, detOverlapsA)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for overlap strings a+b',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      detOverlapsB => detOverlapsA
    end if

    if (P%yes_verbose) then
      call check_time(int_time, timer)
      time_step = timer
    end if
    ! compile the configuration state function overlaps
    !do concurrent(c1=1:size(csf_array)) local(coef1, coef2,coef, sign1, sign2, det)
    do c1=1,size(csf_array)
      do c2=1,size(csf_array)
        csf_O(c1, c2) = ZERO
        do d1 = 1, size(csf_array(c1)%dets, 2)
          coef1 = csfdet(csf_array(c1)%csf, csf_array(c1)%dets(:, d1))
          if (abs(coef1) < 1e-10) cycle
          do d2 = 1, size(csf_array(c2)%dets, 2)
            coef2 = csfdet(csf_array(c2)%csf, csf_array(c2)%dets(:, d2))
            if (abs(coef2) < 1e-10) cycle
            det = detOverlapsA(index_alpha(string_start(c1) + d1), index_alpha(string_start(c2) + d2))
            det = det*detOverlapsB(index_beta(string_start(c1) + d1), index_beta(string_start(c2) + d2))
            ! !! test
            ! call spin_string(csf_array(c1)%occ, csf_array(c1)%dets(:,d1), stringa(:), stringb(:), sign1)
            ! call spin_string(csf_array(c2)%occ, csf_array(c2)%dets(:,d2), stringa2(:), stringb2(:), sign2)
            ! Oalpha(1:Nalpha,1:Nalpha) = O(stringa, stringa2 )
            ! Obeta(1:Nbeta,1:Nbeta)=O(stringb ,stringb2 )
            ! if(abs(det - matrix_determinant(Oalpha) * matrix_determinant(Obeta)) > 1e-10) then
            !   write(6,*) 'conf 1:',c1,d1
            !   write(6,*) 'conf 2:',c2,d2

            !   write(6,*) 'det1:',trim(detstring(csf_array(c1)%occ,csf_array(c1)%dets(:,d1)))
            !   write(6,*) 'det2:',trim(detstring(csf_array(c2)%occ,csf_array(c2)%dets(:,d2)))
            !   write(6,*) 'Oalpha:'
            !   write(6,*) 'strings ',index_alpha(string_start(c1) + d1),index_alpha(string_start(c2) + d2)
            !   write(6,*) 'u. string1 ',unique_stringsA(:,index_alpha(string_start(c1) + d1))
            !   write(6,*) 'u. string2 ',unique_stringsA(:,index_alpha(string_start(c2) + d2))

            !   call print_matrix(Oalpha)
            !   write(6,*) 'det=',matrix_determinant(Oalpha),&
            !        detOverlapsA(index_alpha(string_start(c1) + d1), index_alpha(string_start(c2) + d2))
            !   write(6,*) 'Obeta:'
            !   write(6,*) 'strings ',index_beta(string_start(c1) + d1), index_beta(string_start(c2) + d2)
            !   write(6,*) 'u. string1 ',unique_stringsB(:,index_beta(string_start(c1) + d1))
            !   write(6,*) 'u. string2 ',unique_stringsB(:,index_beta(string_start(c2) + d2))

            !   call print_matrix(Obeta)
            !   write(6,*) 'det=',matrix_determinant(Obeta),&
            !        detOverlapsB(index_beta(string_start(c1) + d1), index_beta(string_start(c2) + d2))
            !   stop
            ! end if
            coef = coef1*coef2
            sign1 = string_sign(string_start(c1) + d1)
            sign2 = string_sign(string_start(c2) + d2)
            csf_O(c1, c2) = csf_O(c1, c2) + sign1*sign2*coef*det
          end do! d2
        end do! d1
        !      if(csf_O(c1,c2)**2 > 0.1)&
        !        write(6,*) 'CSF overlap > 0.1',c1,c2,csf_O(c1,c2)**2
      end do! c2
    end do! c1
    if (P%yes_verbose) then
      call check_time(int_time, timer)
      write (P%OUTPUT, "('# CSFS-overlap time for compiling overlaps ',f15.3,'sec')") timer - time_step
      time_step = timer     !
    end if

    deallocate (string_sign)
    deallocate (string_start)
    deallocate (index_alpha)
    deallocate (index_beta)
    deallocate (O)
    if (associated(unique_stringsB, unique_stringsA)) then
      deallocate (unique_stringsA)
      nullify (unique_stringsB)
      deallocate (detOverlapsA)
      nullify (detOverlapsA)
    else
      deallocate (unique_stringsA)
      deallocate (unique_stringsB)
      deallocate (detOverlapsA)
      deallocate (detOverlapsB)
    end if

    ! for testing
    !deallocate(Oalpha)
    !deallocate(Obeta)
    !deallocate(stringa2)
    !deallocate(stringb2)

  contains
    subroutine calc_stringsOverlap(strings, OMO, overlap)
      implicit none
      integer, intent(in) :: strings(:, :)
      real(long), intent(in) :: OMO(:, :) ! orbital overlap
      real(long), intent(inout) :: overlap(:, :)
      integer :: i1, i2
      integer :: nstrings, N
      integer :: sign
      real(long) :: On(size(strings, 1), size(strings, 1)) ! overlap matrix
      real(long) :: Om(size(strings, 1) - 1, size(strings, 1) - 1) ! overlap matrix of dimension N-1
      real(long) :: cofactors(size(strings, 1)) ! stores co-factors for Laplacian expansion
      nstrings = size(strings, 2)
      N = size(strings, 1)
      !$omp parallel private(i1,i2,i,sign,On,Om,cofactors) shared(nstrings,N,strings,overlap,OMO) default(none)
      !$omp do schedule(guided)
      do i1 = 1, nstrings
        do i2 = 1, nstrings
          if (.True.) then
            if (i2 .eq. 1) then
              ! when i2.eq.1
              ! calculate new cofactors
              sign = 2*mod(N - 1, 2) - 1 ! -1^j
              On(1:N, 1:N) = OMO(strings(:, i1), strings(:, i2))
              do i = 1, N
                sign = -sign ! -1^(j+i)
                Om(1:i - 1, 1:N - 1) = On(1:i - 1, 1:N - 1)
                Om(i:N - 1, 1:N - 1) = On(i + 1:N, 1:N - 1)
                cofactors(i) = sign*matrix_determinant(Om)
              end do! i
            else if (any(strings(:N - 1, i2) .ne. strings(:N - 1, i2 - 1))) then
              ! when the right determinant differs in the first N-1 entries, recalculate co-factors of the determinant
              ! otherwise (i.e., the 1...N-1 entries are the same), reuse the previous cofactors
              sign = 2*mod(N - 1, 2) - 1 ! =-1^(N)
              On(1:N, 1:N) = OMO(strings(:, i1), strings(:, i2))
              do i = 1, N
                sign = -sign ! -1^(j+i)
                Om(1:i - 1, 1:N - 1) = On(1:i - 1, 1:N - 1)
                Om(i:N - 1, 1:N - 1) = On(i + 1:N, 1:N - 1)
                cofactors(i) = sign*matrix_determinant(Om)
              end do! i
            end if
            ! calculate determinant using Laplace expansion using precalculated co-factors along the last column
            overlap(i1, i2) = ZERO
            do i = 1, N
              overlap(i1, i2) = overlap(i1, i2) + &
                OMO(strings(i, i1), strings(N, i2))*cofactors(i)
            end do
          else
            overlap(i1, i2) = matrix_determinant(OMO(strings(:, i1), strings(:, i2)))
          end if
          ! check
          !   if( abs(overlap(i1,i2)- matrix_determinant(OMO(strings(:,i1), strings(:,i2))) ) > 1e-10)then
          !     write(6,*) i1,':', strings(:,i1)
          !     write(6,*) i2,':', strings(:,i2)
          !     call print_matrix(OMO(strings(:,i1), strings(:,i2)))
          !     do i=1,N
          !       write(6,*) OMO(strings(i, i1), strings(N, i2)), cofactors(i)
          !     end do
          !     write(6,*) overlap(i1,i2)
          !     write(6,*) matrix_determinant(OMO(strings(:,i1), strings(:,i2)))
          !     stop
          !   end if
        end do! i2
      end do! i1
      !$omp end do
      !$omp end parallel
    end subroutine calc_stringsOverlap
  end subroutine overlap_csfs


  subroutine overlap_operator_csfs(P, csf_array1, csf_array2, LC1, LC2, S12, OP1, csf_O)
    !! calculated overlaps of configuration state functions
    !! for different orbital sets    
    type(Param), intent(in) :: P
    type(CSF), intent(in) :: csf_array1(:) !! array for csfs 1
    type(CSF), intent(in) :: csf_array2(:) !! array for csfs 2
    real(long), intent(in) :: LC1(:, :) !! orbitals 1
    real(long), intent(in) :: LC2(:, :) !! orbitals 2
    real(long), intent(in) :: S12(:, :) !! overlap (basis set)
    real(long), intent(in) :: OP1(:, :, :) !! operator (basis set)
    real(long), intent(out) :: csf_O(size(csf_array1, 1), size(csf_array2, 1), size(OP1, 3)) !! array for final csfs operator

    real(long), dimension(:, :), allocatable :: O! one particle overlap matrix all orbitals
    real(long), dimension(:, :, :), allocatable :: OP! one particle operator matrix all orbitals
    integer c1, c2, d1, d2
    integer :: i, j, k
    real(long) :: coef, coef1, coef2
    integer :: sign1, sign2
    real(long):: det1, det2

    integer, allocatable :: string1_start(:)
    integer, allocatable :: string1_sign(:)
    integer, allocatable :: string2_start(:)
    integer, allocatable :: string2_sign(:)
    real(long), pointer :: detOverlapsA(:, :)
    real(long), pointer :: detOpOverlapsA(:, :, :)
    real(long), pointer :: detOverlapsB(:, :)
    real(long), pointer :: detOpOverlapsB(:, :, :)
    integer, pointer :: unique_strings1A(:, :)
    integer, pointer :: unique_strings1B(:, :)
    integer, pointer :: unique_strings2A(:, :)
    integer, pointer :: unique_strings2B(:, :)
    integer, allocatable :: index1_alpha(:)
    integer, allocatable :: index1_beta(:)
    integer, allocatable :: index2_alpha(:)
    integer, allocatable :: index2_beta(:)

    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_step

    call check_time(int_time, timer)
    time_init = timer     !
    time_step = timer
    allocate (O(size(LC1, 2), size(LC2, 2)))
    allocate (OP(size(LC1, 2), size(LC2, 2), size(OP1, 3)))
    ! calculate all MO overlaps
    do i = 1, size(LC1, 2)
      do j = 1, size(LC2, 2)
        O(i, j) = dot_product(LC1(:, i), MATMUL(S12, LC2(:, j)))
        do k = 1, size(OP1, 3)
          OP(i, j, k) = dot_product(LC1(:, i), MATMUL(OP1(:, :, k), LC2(:, j)))
        end do
      end do
    end do

    call make_string_table(P, csf_array1, unique_strings1A, unique_strings1B, &
      index1_alpha, index1_beta, string1_start, string1_sign)
    call make_string_table(P, csf_array2, unique_strings2A, unique_strings2B, &
      index2_alpha, index2_beta, string2_start, string2_sign)
    if (.not. ASSOCIATED(unique_strings1A, unique_strings1B) .or. .TRUE.) then

      allocate (detOverlapsA(size(unique_strings1A, 2), size(unique_strings2A, 2)))
      allocate (detOpOverlapsA(size(unique_strings1A, 2), size(unique_strings2A, 2), size(OP1, 3)))
      call calc_stringsOperator1(unique_strings1A, unique_strings2A, O, OP, detOpOverlapsA, detOverlapsA)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for operator strings a',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      allocate (detOverlapsB(size(unique_strings1B, 2), size(unique_strings2B, 2)))
      allocate (detOpOverlapsB(size(unique_strings1B, 2), size(unique_strings2B, 2), size(OP1, 3)))
      call calc_stringsOperator1(unique_strings1B, unique_strings2B, O, OP, detOpOverlapsB, detOverlapsB)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for operator strings b',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
    else
      stop 'not implemented'
      !allocate(detOverlapsA(size(unique_stringsA,2),size(unique_stringsA,2)))
      !call calc_stringsOverlap(unique_stringsA(:,:), O, detOverlapsA)
      !if(P%yes_verbose)then
      !  call check_time( int_time, timer )
      !  write(P%OUTPUT,"('# CSFS-overlap time for overlap strings a+b',f15.3,'sec')") timer-time_step
      !  time_step = timer     !
      !end if
      !detOverlapsB => detOverlapsA
    end if

    if (P%yes_verbose) then
      call check_time(int_time, timer)
      time_step = timer
    end if
    ! compile the configuration state function overlaps
    do k = 1, size(OP1, 3)
      do c1 = 1, size(csf_array1)
        do c2 = 1, size(csf_array2)
          csf_O(c1, c2, k) = ZERO
          do d1 = 1, size(csf_array1(c1)%dets, 2)
            coef1 = csfdet(csf_array1(c1)%csf, csf_array1(c1)%dets(:, d1))
            if (abs(coef1) < 1e-10) cycle
            do d2 = 1, size(csf_array2(c2)%dets, 2)
              coef2 = csfdet(csf_array2(c2)%csf, csf_array2(c2)%dets(:, d2))
              if (abs(coef2) < 1e-10) cycle

              if (size(unique_strings1A, 1) .eq. 0 .and. size(unique_strings1B, 1) .eq. 0) then
                cycle
              else if (size(unique_strings1A, 1) .eq. 0) then
                det1 = detOpOverlapsB(index1_beta(string1_start(c1) + d1), index2_beta(string2_start(c2) + d2), k)
                det2 = ZERO
              else if (size(unique_strings1B, 1) .eq. 0) then
                det1 = ZERO
                det2 = detOpOverlapsA(index1_alpha(string1_start(c1) + d1), index2_alpha(string2_start(c2) + d2), k)
              else
                ! < det_a det_b| op | det_a' det_b'> = <det_a|det_a'> <det_b|op|det_b'> + <det_a|op|det_a'> <det_b|det_b'>
                det1 = detOverlapsA(index1_alpha(string1_start(c1) + d1), index2_alpha(string2_start(c2) + d2))
                det1 = det1*detOpOverlapsB(index1_beta(string1_start(c1) + d1), index2_beta(string2_start(c2) + d2), k)
                det2 = detOverlapsB(index1_beta(string1_start(c1) + d1), index2_beta(string2_start(c2) + d2))
                det2 = det2*detOpOverlapsA(index1_alpha(string1_start(c1) + d1), index2_alpha(string2_start(c2) + d2), k)
              end if

              coef = coef1*coef2
              sign1 = string1_sign(string1_start(c1) + d1)
              sign2 = string2_sign(string2_start(c2) + d2)
              csf_O(c1, c2, k) = csf_O(c1, c2, k) + sign1*sign2*coef*(det1 + det2)
            end do ! d2
          end do! d1
        end do! c2
      end do! c1
    end do! k
    if (P%yes_verbose) then
      call check_time(int_time, timer)
      write (P%OUTPUT, "('# CSFS-overlap time for operator ',f15.3,'sec')") timer - time_step
      time_step = timer     !
    end if

    deallocate (string1_sign)
    deallocate (string1_start)
    deallocate (index1_alpha)
    deallocate (index1_beta)
    deallocate (string2_sign)
    deallocate (string2_start)
    deallocate (index2_alpha)
    deallocate (index2_beta)
    deallocate (O)
    deallocate (OP)
    if (associated(unique_strings1B, unique_strings1A)) then
      deallocate (unique_strings1A)
      nullify (unique_strings1A)
      deallocate (detOverlapsA)
      nullify (detOverlapsA)
    else
      deallocate (unique_strings1A)
      deallocate (unique_strings1B)
      deallocate (detOverlapsA)
      deallocate (detOverlapsB)
    end if
    if (associated(unique_strings2B, unique_strings2A)) then
      deallocate (unique_strings2A)
      nullify (unique_strings2A)
      nullify (detOverlapsA)
    else
      deallocate (unique_strings2A)
      deallocate (unique_strings2B)
    end if
  contains
    ! calculated overlaps of configuration state functions
    subroutine calc_stringsOperator1(strings1, strings2, OMO, OPO, OPoverlap, overlap)
      implicit none
      integer, intent(in) :: strings1(:, :)
      integer, intent(in) :: strings2(:, :)
      real(long), intent(in) :: OMO(:, :) ! orbital overlap
      real(long), intent(in) :: OPO(:, :, :) ! one-particle operator
      real(long), intent(inout) :: overlap(size(strings1, 2), size(strings2, 2))
      real(long), intent(inout) :: OPoverlap(size(strings1, 2), size(strings2, 2), size(OPO, 3))
      integer :: i1, i2
      integer :: nstrings1, nstrings2, N, sign
      real(long) :: Ov(size(strings1, 1), size(strings2, 1)) ! overlap matrix
      real(long) :: Opm(size(strings1, 1), size(strings2, 1), size(OPO, 3)) ! operator matrix
      real(long) :: Om(size(strings1, 1) - 1, size(strings2, 1) - 1) ! overlap matrix of dimension N-1
      real(long) :: cofactors(size(strings1, 1)) ! stores co-factors for Laplacian expansion
      real(long) :: cofactorsOp(size(strings1, 1), size(OPO, 3)) ! stores co-factors for Laplacian expansion, where operator as been applied
      logical :: yes_debug = .False.
      real(long) :: temp
      nstrings1 = size(strings1, 2)
      nstrings2 = size(strings2, 2)
      N = size(strings1, 1)
      if (size(strings1, 1) .ne. size(strings2, 1)) &
        stop 'string length not the same!'
      !$omp parallel private(i1,i2,i,Ov,cofactors,cofactorsOP,sign) shared(nstrings1,nstrings2,N,strings1,strings2,overlap,OPoverlap,OMO,OPO,yes_debug) default(private)
      !$omp do schedule(guided)
      do i1 = 1, nstrings1
        do i2 = 1, nstrings2
          if (.NOT. yes_debug) then
            ! if we are at the first determinant OR
            ! when the right determinant differs in the first N-1 entries, recalculate co-factors of the determinant
            ! otherwise (i.e., the 1...N-1 entries are the same), reuse the previous cofactors
            ! cofactor for overlap and N-th column and i-th row removed
            ! cofactors(i) = sign * determinant of
            !            <1|1'>   <1|2'>   ...   <1|i'> ...   <1|N-1'>
            !            <2|1'>   <2|2'>   ...   <2|i'>
            !            ...
            !            <i-1|1'> <i-1|2'> ... <i-1|i'> ... <i-1|N-1'>
            !            <i+1|1'> <i+1|2'> ... <i+1|i'> ... <i+1|N-1'>
            !             ...
            !            <N|1'>    <N|2'> ...    <N|i'> ...   <N|N-1'>
            ! cofactor for 1-p-operator: overlap with N-th column and i-th row removed, j-th column replaced with operator
            ! cofactorsOp(i) = sign * sum_j determinant of
            !            <1|1'>   <1|2'>   ...   <1|i'> ...    <1|op|j'> ...   <1|N-1'>
            !            <2|1'>   <2|2'>   ...   <2|i'>        <2|op|j'>
            !            ...
            !            <i-1|1'> <i-1|2'> ... <i-1|i'> ...  <i-1|op|j'> ... <i-1|N-1'>
            !            <i+1|1'> <i+1|2'> ... <i+1|i'> ...  <i+1|op|j'> ... <i+1|N-1'>
            !             ...
            !            <N|1'>    <N|2'> ...    <N|i'> ...    <N|op|j'>  ...   <N|N-1'>
            if (i2 .eq. 1) then
              N = size(strings1(:, i1), 1)
              sign = 2*mod(N - 1, 2) - 1 ! -1^j
              Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
              Opm(1:N, 1:N, :) = OPO(strings1(:, i1), strings2(:, i2), :)
              do i = 1, N
                sign = -sign ! -1^(j+i)
                Om(1:i - 1, 1:N - 1) = Ov(1:i - 1, 1:N - 1)
                Om(i:N - 1, 1:N - 1) = Ov(i + 1:N, 1:N - 1)
                cofactors(i) = sign*matrix_determinant(Om)
                do k = 1, size(OPO, 3)
                  cofactorsOp(i, k) = ZERO
                  do j = 1, N - 1
                    Om(1:i - 1, 1:N - 1) = Ov(1:i - 1, 1:N - 1)
                    Om(i:N - 1, 1:N - 1) = Ov(i + 1:N, 1:N - 1)
                    Om(1:i - 1, j) = Opm(1:i - 1, j, k)
                    Om(i:N - 1, j) = Opm(i + 1:N, j, k)
                    cofactorsOp(i, k) = cofactorsOp(i, k) + sign*matrix_determinant(Om)
                  end do! j
                end do!k
              end do! i
            else if (any(strings2(:N - 1, i2) .ne. strings2(:N - 1, i2 - 1))) then
              N = size(strings1(:, i1), 1)
              sign = 2*mod(N - 1, 2) - 1 ! -1^j
              Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
              Opm(1:N, 1:N, :) = OPO(strings1(:, i1), strings2(:, i2), :)
              do i = 1, N
                sign = -sign ! -1^(j+i)
                Om(1:i - 1, 1:N - 1) = Ov(1:i - 1, 1:N - 1)
                Om(i:N - 1, 1:N - 1) = Ov(i + 1:N, 1:N - 1)
                cofactors(i) = sign*matrix_determinant(Om)
                do k = 1, size(OPO, 3)
                  cofactorsOp(i, k) = ZERO
                  do j = 1, N - 1
                    Om(1:i - 1, 1:N - 1) = Ov(1:i - 1, 1:N - 1)
                    Om(i:N - 1, 1:N - 1) = Ov(i + 1:N, 1:N - 1)
                    Om(1:i - 1, j) = Opm(1:i - 1, j, k)
                    Om(i:N - 1, j) = Opm(i + 1:N, j, k)
                    cofactorsOp(i, k) = cofactorsOp(i, k) + sign*matrix_determinant(Om)
                  end do! j
                end do!k
              end do! i
            end if
            ! calculate determinant using Laplace expansion using precalculated co-factors along the last column
            OPoverlap(i1, i2, :) = ZERO
            overlap(i1, i2) = ZERO
            do i = 1, N
              overlap(i1, i2) = overlap(i1, i2) + &
                OMO(strings1(i, i1), strings2(N, i2))*cofactors(i)
              do k = 1, size(OPO, 3)
                OPoverlap(i1, i2, k) = OPoverlap(i1, i2, k) + &
                  OPO(strings1(i, i1), strings2(N, i2), k)*cofactors(i) + &
                  OMO(strings1(i, i1), strings2(N, i2))*cofactorsOp(i, k)
              end do
            end do
          else ! for debugging
            Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
            overlap(i1, i2) = matrix_determinant(Ov)
            OPoverlap(i1, i2, :) = ZERO
            do k = 1, size(OPO, 3)
              do j = 1, N
                Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
                Ov(:, j) = Opo(strings1(:, i1), strings2(j, i2), k)
                OPoverlap(i1, i2, k) = OPoverlap(i1, i2, k) + matrix_determinant(Ov)
              end do
            end do
          end if
          !!!!!!!!!!!!!!!!!!!!!!
          ! ! check
          !
          if (yes_debug) then
            Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
            if (abs(overlap(i1, i2) - matrix_determinant(OMO(strings1(:, i1), strings2(:, i2)))) > 1e-10) then
              write (6, *) i1, ':', strings1(:, i1)
              write (6, *) i2, ':', strings2(:, i2)
              write (6, *) 'Overlap'
              call print_matrix(Ov)
              write (6, *) overlap(i1, i2), matrix_determinant(Ov)
              stop
            end if
            do k = 1, size(OPO, 3)
              temp = ZERO
              do j = 1, N
                Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
                Ov(:, j) = Opo(strings1(:, i1), strings2(j, i2), k)
                temp = temp + matrix_determinant(Ov)
              end do
              if (abs(OPoverlap(i1, i2, k) - temp) > 1e-10) then
                write (6, *) i1, ':', strings1(:, i1)
                write (6, *) i2, ':', strings2(:, i2)
                write (6, *) 'Operator'
                do j = 1, N
                  Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
                  Ov(:, j) = Opo(strings1(:, i1), strings2(j, i2), k)
                  call print_matrix(Ov)
                end do
                write (6, *) OPoverlap(i1, i2, k), temp
                stop
              end if
            end do

            write (6, *) i1, ':', strings1(:, i1)
            write (6, *) i2, ':', strings2(:, i2)
            write (6, *) 'Overlap:'
            Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
            call print_matrix(Ov)
            write (6, *) overlap(i1, i2)

            do k = 1, size(OPO, 3)
              write (6, *) 'Operator:', k
              do j = 1, N
                Ov(1:N, 1:N) = OMO(strings1(:, i1), strings2(:, i2))
                Ov(:, j) = Opo(strings1(:, i1), strings2(j, i2), k)
                call print_matrix(Ov)
              end do
              write (6, *) OPoverlap(i1, i2, k)
            end do
          end if

        end do! i2
      end do! i1
      !$omp end do
      !$omp end parallel
    end subroutine calc_stringsOperator1
  end subroutine overlap_operator_csfs

  subroutine make_string_table(P, csf_array, unique_stringsA, unique_stringsB, index_alpha, index_beta, string_start, string_sign)    
    type(Param), intent(in) :: P
    type(CSF), intent(in) :: csf_array(:) ! array of csfs
    integer, pointer, intent(out) :: unique_stringsA(:, :)
    integer, pointer, intent(out) :: unique_stringsB(:, :)
    integer, allocatable, intent(out) :: index_alpha(:)
    integer, allocatable, intent(out) :: index_beta(:)
    integer, allocatable, intent(out) :: string_start(:)
    integer, allocatable, intent(out) :: string_sign(:)
    integer c1, d1
    integer :: i, Nclosed, Nopen, Nalpha, Nbeta, ja, jb
    integer :: ndet
    integer, pointer :: last_dets(:, :)
    integer, pointer :: unique_stringsA_tmp(:, :)
    integer, pointer :: unique_stringsB_tmp(:, :)
    integer, allocatable :: stringa(:)
    integer, allocatable :: stringb(:)
    integer, allocatable :: table(:)
    integer, allocatable :: rtable(:)
    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_step
    last_dets => null()
    call check_time(int_time, timer)
    time_init = timer     !
    time_step = timer
    c1 = 1
    d1 = 1
    Nclosed = count(csf_array(c1)%occ .eq. 2)
    Nopen = count(csf_array(c1)%occ .eq. 1)
    if (Nopen > 0) then
      Nalpha = Nclosed + (Nopen + csf_array(c1)%dets(Nopen, d1))/2
      Nbeta = Nclosed + (Nopen - csf_array(c1)%dets(Nopen, d1))/2
    else
      Nalpha = Nclosed
      Nbeta = Nclosed
    end if

    ! count the number of determinants
    ndet = 0
    do c1 = 1, size(csf_array)
      if (associated(last_dets, csf_array(c1)%dets)) cycle
      ndet = ndet + size(csf_array(c1)%dets, 2)
      last_dets => csf_array(c1)%dets
    end do

    allocate (index_alpha(ndet))
    allocate (index_beta(ndet))
    allocate (stringa(Nalpha))
    allocate (stringb(Nbeta))
    allocate (string_sign(ndet))
    allocate (string_start(size(csf_array)))
    allocate (unique_stringsA(Nalpha, ndet))
    if (Nalpha .eq. Nbeta) then
      unique_stringsB => unique_stringsA
    else
      allocate (unique_stringsB(Nbeta, ndet))
    end if
    if (P%yes_verbose) then
      call check_time(int_time, timer)
      write (P%OUTPUT, "('# CSFS-overlap time preparation ',f15.3,'sec')") timer - time_init
      time_step = timer     !
    end if

    ja = 0
    jb = 0
    ndet = 0
    nullify (last_dets)
    do c1 = 1, size(csf_array)
      !write(6,*) trim(csfstring(csf_array(c1)%occ,csf_array(c1)%csf(:)))
      ! when the set of determinants is the same, use the earlier strings
      if (associated(last_dets, csf_array(c1)%dets)) then
        string_start(c1) = ndet - size(csf_array(c1)%dets, 2)
      else
        string_start(c1) = ndet
        detloop: do d1 = 1, size(csf_array(c1)%dets, 2)
          string_sign(ndet + d1) = 1
          call spin_string(csf_array(c1)%occ, csf_array(c1)%dets(:, d1), stringa(:), stringb(:), string_sign(ndet + d1))
          index_alpha(string_start(c1) + d1) = 0
          index_beta(string_start(c1) + d1) = 0
          ! see if there is already this string in the unique_strings arrays.
          ! if not, add it
          do i = 1, ja
            if (all(unique_stringsA(:, i) .eq. stringa(:))) then
              index_alpha(string_start(c1) + d1) = i
              exit
            end if
          end do! i
          if (index_alpha(string_start(c1) + d1) .eq. 0) then
            ja = ja + 1
            if (associated(unique_stringsA, unique_stringsB)) jb = ja
            unique_stringsA(:, ja) = stringa(:)
            index_alpha(string_start(c1) + d1) = ja
          end if
          do i = 1, jb
            if (all(unique_stringsB(:, i) .eq. stringb(:))) then
              index_beta(string_start(c1) + d1) = i
              exit
            end if
          end do! i
          if (index_beta(string_start(c1) + d1) .eq. 0) then
            jb = jb + 1
            if (associated(unique_stringsA, unique_stringsB)) ja = jb
            unique_stringsB(:, jb) = stringb(:)
            index_beta(string_start(c1) + d1) = jb
          end if
        end do detloop
        ndet = ndet + size(csf_array(c1)%dets, 2)
        last_dets => csf_array(c1)%dets
      end if
    end do! c1
    if (P%yes_verbose) then
      write (6, "('# number of unique strings',i10,i10)") ja, jb
      call check_time(int_time, timer)
      write (P%OUTPUT, "('# CSFS-overlap time for finding unique strings ',f15.3,'sec')") timer - time_step
      time_step = timer     !
    end if
    ! shorten the arrays unique_strings
    allocate (unique_stringsA_tmp(Nalpha, ja), source=unique_stringsA(:, 1:ja))
    if (Nalpha .eq. Nbeta) then
      deallocate (unique_stringsA)
      nullify (unique_stringsA)
      unique_stringsA => unique_stringsA_tmp
      unique_stringsB => unique_stringsA_tmp
      nullify (unique_stringsA_tmp)
    else
      allocate (unique_stringsB_tmp(Nbeta, jb), source=unique_stringsB(:, 1:jb))
      deallocate (unique_stringsA)
      deallocate (unique_stringsB)
      nullify (unique_stringsA)
      nullify (unique_stringsB)
      unique_stringsA => unique_stringsA_tmp
      unique_stringsB => unique_stringsB_tmp
      nullify (unique_stringsA_tmp)
      nullify (unique_stringsB_tmp)
    end if

    ! sort the strings in increasing order, adapt index_alpha and index_beta for the new order
    ! and compute all string overlaps
    if (Nalpha .ne. Nbeta) then
      allocate (table(ja))
      allocate (rtable(ja))
      table(:) = (/(i, i=1, ja)/)
      call quicksort_strings(unique_stringsA(:, :ja), table)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for sorting unique strings a',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      do i = 1, ja
        rtable(table(i)) = i
      end do
      index_alpha = rtable(index_alpha)
      deallocate (table)
      deallocate (rtable)
      allocate (table(jb))
      allocate (rtable(jb))
      table(:) = (/(i, i=1, jb)/)
      call quicksort_strings(unique_stringsB(:, :jb), table)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for sorting unique strings b',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      do i = 1, jb
        rtable(table(i)) = i
      end do
      index_beta = rtable(index_beta)
      deallocate (table)
      deallocate (rtable)
    else
      ! sort the strings in increasing order
      allocate (table(ja))
      allocate (rtable(ja))
      table(:) = (/(i, i=1, ja)/)
      call quicksort_strings(unique_stringsA(:, 1:ja), table)
      if (P%yes_verbose) then
        call check_time(int_time, timer)
        write (P%OUTPUT, "('# CSFS-overlap time for sorting unique strings a+b',f15.3,'sec')") timer - time_step
        time_step = timer     !
      end if
      do i = 1, ja
        rtable(table(i)) = i
      end do
      index_alpha = rtable(index_alpha)
      index_beta = rtable(index_beta)
      deallocate (table)
      deallocate (rtable)
    end if
  contains
    recursive subroutine quicksort_strings(strings, tabl)
      implicit none
      integer, intent(inout) :: strings(:, :)
      integer, intent(inout) :: tabl(:)
      integer :: first = 1, last
      integer ::  k, j, t
      integer :: string_tmp(size(strings, 1))
      integer :: string_x(size(strings, 1))
      last = size(strings, 2)
      string_x = strings(:, int((last + first)/2))
      k = first
      j = last
      do
        do while (string_is_smaller_than(strings(:, k), string_x))
          k = k + 1
        end do
        do while (string_is_smaller_than(string_x, strings(:, j)))
          j = j - 1
        end do
        if (k .ge. j) exit
        ! swap i and j
        string_tmp(:) = strings(:, k)
        strings(:, k) = strings(:, j)
        strings(:, j) = string_tmp
        t = tabl(k)
        tabl(k) = tabl(j)
        tabl(j) = t
        k = k + 1
        j = j - 1
      end do
      if (first < k - 1) call quicksort_strings(strings(:, first:k - 1), tabl(first:k - 1))
      if (j + 1 < last) call quicksort_strings(strings(:, j + 1:last), tabl(j + 1:last))
    end subroutine quicksort_strings

    function string_is_smaller_than(string1, string2)
      implicit none
      integer :: string1(:)
      integer :: string2(:)
      integer :: k
      logical :: string_is_smaller_than
      do k = 1, size(string1)
        if (string1(k) .ne. string2(k)) then
          string_is_smaller_than = string1(k) < string2(k)
          return
        end if
      end do
      string_is_smaller_than = .False.
    end function string_is_smaller_than
  end subroutine make_string_table

  subroutine trans_dens_csf(csf1, csf2, tdens)
    !! calculates the transition density between two csfs
    type(CSF), intent(in) :: csf1 !! csf1
    type(CSF), intent(in) :: csf2 !! csf2
    integer :: d1, d2
    real(long), intent(inout), allocatable :: tdens(:, :) !! transition density
    integer :: nmo1
    integer :: nmo2
    integer :: only_left(1), only_left_spin(1) ! array for orbital indices for determinant comparison
    integer :: only_right(1), only_right_spin(1) ! array for spin indices for det
    real(long) :: coef
    integer :: coef_sign

    nmo1 = size(csf1%occ, 1)
    nmo2 = size(csf2%occ, 1)
    tdens = ZERO
    do d1 = 1, size(csf1%dets, 2)
      do d2 = 1, size(csf2%dets, 2)
        coef = csfdet(csf1%csf, csf1%dets(:, d1))* &
          csfdet(csf2%csf, csf2%dets(:, d2))
        if (coef .eq. ZERO) cycle
        call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, &
          csf1%occ, csf1%dets(:, d1), &
          csf2%occ, csf2%dets(:, d2))
        coef = coef*coef_sign
        if (coef_sign .ne. 0 .and. count(only_left(:) .ne. 0) .eq. 1 .and. count(only_right(:) .ne. 0) .eq. 1 .and. &
          only_left_spin(1) .eq. only_right_spin(1)) then
          tdens(only_left(1), only_right(1)) = &
            tdens(only_left(1), only_right(1)) + coef
        end if
      end do! d2
    end do! d1
  end subroutine trans_dens_csf

  subroutine MCSCF_procedure(P, A, G, B, F)
    !! MCSCF procedure
    type(Param), intent(in) :: P
    type(Atom_set), intent(in) :: A(:)
    type(Grid3D), intent(in) :: G
    type(Basis), intent(inout) :: B
    type(WF_data), intent(inout) :: F

    integer :: Nopen
    integer :: multp
    integer :: it, inner_it
    integer :: i, j, k, l
    integer(8) :: ll
    integer, allocatable, target :: csfs(:, :)
    integer, allocatable, target :: dets(:, :)
    integer, allocatable, target :: occs(:, :)
    integer, allocatable, target :: irrep_occs(:)
    integer, allocatable :: occs_temp(:, :)
    type(UmatrixTable) :: UT

    real(long), allocatable :: opdm(:, :)
    real(long), allocatable :: F_act(:, :)
    real(long), allocatable :: F_core(:, :)
    real(long) :: E_core
    real(long), allocatable :: old_opdm(:, :)
    real(long), allocatable :: h0(:, :)
    real(long), allocatable :: LC_save(:, :)
    real(long), allocatable :: Lag(:, :)
    type(indep_mo_pair), allocatable :: pairs(:)
    real(long), allocatable :: gradient(:)
    real(long), allocatable :: old_gradient(:)
    real(long), allocatable :: Hessian(:, :)
    real(long), allocatable :: Hessian_copy(:, :)
    real(long), allocatable :: inv_Hessian(:, :)
    real(long), allocatable :: old_inv_Hessian(:, :)
    real(long), allocatable :: theta_step(:)
    real(long), allocatable :: theta(:)
    real(long), allocatable :: old_theta_step(:)
    real(long), allocatable :: old_theta(:, :)
    real(long), allocatable :: old_err(:, :)
    real(long), allocatable :: last_theta(:)
    real(long), allocatable :: s_tpdm(:)
    real(long), allocatable :: s_tpdmaa(:)

    type(CSF), allocatable :: csf_array(:)
    real(long), allocatable :: csf_c(:)
    real(long), allocatable :: CI_matrix(:, :)
    real(long), allocatable :: CI_energy(:)

    integer*8 :: int_time      ! timer
    real :: timer
    real :: time_init, time_inner_it, time_step

    type(LU_matrix) :: LU_hess
    real(long) :: hess_det, nominator, denominator
    real(long) :: energy, old_energy, expected_energy
    real(long) :: max_theta_step

    integer :: n_two
    integer :: Ncsfs
    integer :: OUTPUT
    integer :: n_diis = 0

    integer, allocatable :: core(:)
    integer, allocatable :: occupied(:)
    integer, allocatable :: active(:)
    integer, allocatable :: virtual(:)
    integer, allocatable :: virtual_nFz(:)
    integer, allocatable :: active_nFz(:)
    integer, allocatable :: core_nFz(:)

    real(long) :: gamma = 0.1
    real(long) :: lambda = 0.0
    real(long) :: norm
    ! for linesearch
    integer, parameter :: NLS = 30
    integer :: ls_step
    real(long) :: Energy_lambda(NLS)
    real(long) :: gradient_lambda(NLS)
    real(long) :: lambda_values(NLS)
    real(long), allocatable :: gradient_ls(:)
    real(long) :: delta_energy_CI
    logical :: yes_activeactive = .FALSE. ! include rotations among active orbitals
    real(long), parameter :: conv_gradient = 1e-4
    character(len=512) :: moldenfilename ! filename
    logical :: yes_averageOverAllStates
    yes_averageOverAllStates = P%yes_state_average .and. .not. allocated(P%SA_weights)
    call check_time(int_time, timer) 
    time_init = timer
    time_step = timer
    OUTPUT = P%OUTPUT
    allocate (h0(B%N_basis, B%N_basis))
    !
    ! find the highest occupied molecular orbital (HOMO)
    !
    moloop: do i = size(F%n_occ), 1, -1
      if (allocated(F%av_occ)) then
        do j = 1, size(F%av_occ)
          if (any(F%av_occ(j)%MO(:) .eq. i)) exit moloop
        end do
      end if
      if (F%n_occ(i) .ne. 0) exit moloop
    end do moloop
    F%homo_idx = i
    F%f_occ = ONE*F%n_occ

    F%v_ext(:) = potential_ext(G, A)
    call construct_h0_matrix(G, B, F)
    h0 = F%h0

    if (allocated(P%gas)) then
      write (P%OUTPUT, '("# ",a16)') 'orbital spaces:'
      write (P%OUTPUT, '("# ",a5,a5,a16)') 'min', 'max', 'orbitals'
      do i = 1, size(P%gas)
        write (P%OUTPUT, '("# ",i5,i5,99i4)') P%gas(i)%nmin, P%gas(i)%nmax, P%gas(i)%orbitals(:)
      end do
      call general_active_space_occs(size(F%N_occ), sum(F%N_occ), P%gas, occs)
      do i = 1, size(occs, 2)
        write (P%OUTPUT, '("# ",99i1)') occs(:, i)
      end do

      allocate (active( &
        count((/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))))
      allocate (core( &
        count((/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))))
      allocate (virtual( &
        count((/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))))
      allocate (occupied(size(active, 1) + size(core, 1)))
      active = pack((/(j, j=1, size(occs, 1))/), &
        (/(any(occs(j, :) < 2) .and. any(occs(j, :) > 0), j=1, size(occs, 1))/))
      core = pack((/(j, j=1, size(occs, 1))/), &
        (/(all(occs(j, :) .eq. 2), j=1, size(occs, 1))/))
      virtual = pack((/(j, j=1, size(occs, 1))/), &
        (/(all(occs(j, :) .eq. 0), j=1, size(occs, 1))/))
      occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
        (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))
      yes_activeactive = .True.
    else if (allocated(P%active_space_MO)) then
      allocate (active(size(P%active_space_MO) + count(F%N_occ(:) .eq. 1) - count(F%N_occ(P%active_space_MO) .eq. 1)))
      active(:) = pack((/(j, j=1, size(F%N_occ, 1))/), &
        (/((any(P%active_space_MO .eq. j) .or. F%N_occ(j) .eq. 1), j=1, size(F%N_occ, 1))/))
      allocate (core(count(F%N_occ(:) .eq. 2) - count(F%N_occ(active) .eq. 2)))
      allocate (virtual(count(F%N_occ(:) .eq. 0) - count(F%N_occ(active) .eq. 0)))
      allocate (occupied(size(core, 1) + size(active, 1)))
      core(:) = pack((/(j, j=1, size(F%N_occ, 1))/), (/((F%N_occ(j) .eq. 2 .and. all(active .ne. j)), j=1, size(F%N_occ, 1))/))
      virtual(:) = pack((/(j, j=1, size(F%N_occ, 1))/), (/((F%N_occ(j) .eq. 0 .and. all(active .ne. j)), j=1, size(F%N_occ, 1))/))
      occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), (/((any(core .eq. j) .or. any(active .eq. j)), j=1, size(F%N_occ, 1))/))
      call active_space_occs(F%N_occ, P%active_space_MO, occs)
    else
      allocate (active(count(F%N_occ(:) .eq. 1)))
      allocate (core(count(F%N_occ(:) .eq. 2)))
      allocate (occupied(count(F%N_occ(:) > 0)))
      allocate (virtual(count(F%N_occ(:) .eq. 0)))
      active(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 1)
      core(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 2)
      virtual(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) .eq. 0)
      occupied(:) = pack((/(j, j=1, size(F%N_occ, 1))/), F%N_occ(:) > 0)
      allocate (occs(size(F%N_occ), 1))
      occs(:, 1) = F%N_occ(:)
    end if

    if (P%yes_sym .and. F%state_irrep > 0) then
      call filter_sym_occs(B%irrep_prod, F%irrep_MO, F%state_irrep, occs, occs_temp)
      deallocate (occs)
      allocate (occs(size(occs_temp, 1), size(occs_temp, 2)))
      occs = occs_temp
      deallocate (occs_temp)
    end if

    write (P%OUTPUT, '("# ",a16)', advance='no') 'active orbitals'
    write (P%OUTPUT, '(99i4)') active(:)
    write (P%OUTPUT, '("# ",a16)', advance='no') 'core orbitals'
    write (P%OUTPUT, '(99i4)') core(:)
    write (P%OUTPUT, '("# ",a16)', advance='no') 'virtual orbitals'
    write (P%OUTPUT, '(99i4)') virtual(:)
    write (P%OUTPUT, '("# ",a16)', advance='no') 'occupied orbitals'
    write (P%OUTPUT, '(99i4)') occupied(:)

    if (allocated(F%frozen_MO)) then
      write (P%OUTPUT, '("# ",a16)', advance='no') 'frozen orbitals'
      write (P%OUTPUT, '(99i4)') F%frozen_MO
      allocate (active_nFz(count((/(all(active(j) .ne. F%frozen_MO(:)), j=1, size(active))/))))
      active_nFz = pack(active, (/(all(active(j) .ne. F%frozen_MO(:)), j=1, size(active))/))
      allocate (core_nFz(count((/(all(core(j) .ne. F%frozen_MO(:)), j=1, size(core))/))))
      core_nFz = pack(core, (/(all(core(j) .ne. F%frozen_MO(:)), j=1, size(core))/))
      allocate (virtual_nFz(count((/(all(virtual(j) .ne. F%frozen_MO(:)), j=1, size(virtual))/))))
      virtual_nFz = pack(virtual, (/(all(virtual(j) .ne. F%frozen_MO(:)), j=1, size(virtual))/))

      write (P%OUTPUT, '("# ",a16)', advance='no') 'active-nFz'
      write (P%OUTPUT, '(99i4)') active_nFZ(:)
      write (P%OUTPUT, '("# ",a16)', advance='no') 'core-nFz'
      write (P%OUTPUT, '(99i4)') core_nFZ(:)
      write (P%OUTPUT, '("# ",a16)', advance='no') 'virtual-nFz'
      write (P%OUTPUT, '(99i4)') virtual_nFZ(:)

    else
      allocate (active_nFz(size(active)))
      allocate (virtual_nFz(size(virtual)))
      allocate (core_nFz(size(core)))
      virtual_nFz = virtual
      active_nFz = active
      core_nFz = core
    end if

    multp = F%multp
    write (6, '("# maximum number of open shells: ",i8)') maxval(count(occs .eq. 1, dim=1))
    call constructUMatrixTable(maxval(count(occs .eq. 1, dim=1)), F%multp, UT)

    Ncsfs = 0
    if (P%yes_sym) then
      allocate (irrep_occs(size(occs, 2)))
      if (allocated(P%irrep_MO)) &
        F%irrep_MO = P%irrep_MO
    end if

    do k = 1, size(occs, 2)
      Nopen = count(occs(:, k) .eq. 1)
      if (P%yes_sym) then
        irrep_occs(k) = 1
        do i = 1, B%N_basis
          if (occs(i, k) .eq. 1 .and. P%yes_sym) then
            irrep_occs(k) = B%irrep_prod(irrep_occs(k), F%irrep_MO(i))
          end if
        end do
        if (P%state_irrep > 0 .and. irrep_occs(k) .ne. P%state_irrep) cycle
      end if

      write (OUTPUT, '("#occs:  ",99i1)', advance="no") occs(:, k)
      write (OUTPUT, '(x,"Nopen=",i4,x,"multp",i4)') Nopen, multp
      call Nopen2csfs(Nopen, multp, csfs)
      call Nopen2dets(Nopen, multp, dets)
      if (size(csfs, 2) > 0) then
        Ncsfs = Ncsfs + size(csfs, 2)
        do j = 1, size(csfs, 2)
          write (OUTPUT, '("#  CSF: ",a)') trim(csfstring(occs(:, k), csfs(:, j)))
          do i = 1, size(dets, 2)
            write (OUTPUT, "('#     ',es13.5,1x)", advance='no') csfdet(csfs(:, j), dets(:, i))
            write (OUTPUT, "(a)") detstring(occs(:, k), dets(:, i))
          end do
        end do
      end if
      deallocate (csfs)
      deallocate (dets)
    end do
    if (Ncsfs .eq. 0) then
      stop 'No configurations possible. check occ / multp!'
    end if
    allocate (csf_array(Ncsfs))
    allocate (csf_c(Ncsfs))
    allocate (CI_matrix(Ncsfs, Ncsfs))
    allocate (CI_energy(Ncsfs))

    l = 1
    do k = 1, size(occs, 2)
      if (P%yes_sym) then
        if (P%state_irrep > 0 .and. irrep_occs(k) .ne. P%state_irrep) cycle
      end if
      Nopen = count(occs(:, k) .eq. 1)
      call Nopen2csfs(Nopen, multp, csfs)
      call Nopen2dets(Nopen, multp, dets)
      if (size(csfs, 2) > 0) then
        allocate (csf_array(l)%dets(size(dets, 1), size(dets, 2)))
        csf_array(l)%dets = dets
        ll = l
        do j = 1, size(csfs, 2)
          csf_array(l)%occ => occs(:, k)
          if (P%yes_sym) &
            csf_array(l)%irrep = irrep_occs(k)
          if (l .ne. ll) &
            csf_array(l)%dets => csf_array(ll)%dets
          allocate (csf_array(l)%csf(size(csfs, 1)))
          csf_array(l)%csf = csfs(:, j)
          l = l + 1
        end do
      end if
      deallocate (csfs)
      deallocate (dets)
    end do

    if (P%yes_verbose) then
      write (OUTPUT, '("# ","Configurations:")')
      do k = 1, size(csf_array, 1)
        write (OUTPUT, '("#",99i1)', advance="no") csf_array(k)%occ
        if (P%yes_sym) &
          write (OUTPUT, '(x,"irrep=",i3)', advance="no") csf_array(k)%irrep
        write (OUTPUT, '(x,a)', advance="yes") trim(csfstring(csf_array(k)%occ, csf_array(k)%csf))
        !    do i=1,size(csf_array(k)%dets,2)
        !      write(OUTPUT,"('#     ',es13.5,1x)",advance='no') CSFDET(csf_array(k)%csf,csf_array(k)%dets(:,i))
        !      write(OUTPUT,"(a)") detstring(csf_array(k)%occ,csf_array(k)%dets(:,i))
        !    end do
      end do
    end if

    Nopen = 0
    do i = 1, B%N_basis
      if (occs(i, 1) .eq. 1) Nopen = Nopen + 1
    end do
    call prepare_initial_guess(P, A, G, B, F)

    allocate (opdm(B%N_basis, B%N_basis))
    allocate (old_opdm(B%N_basis, B%N_basis))
    old_opdm(:, :) = ZERO

    n_two = (size(active) + 1)*size(active)/2
    allocate (s_tpdmaa((n_two + 1)*n_two/2))

    if (P%mcscf_type .eq. 3) then
      n_two = (B%N_basis + 1)*B%N_basis/2
      allocate (s_tpdm((n_two + 1)*n_two/2))
      allocate (Lag(B%N_basis, B%N_basis))
    end if

    allocate (LC_save(B%N_basis, B%N_basis))
    allocate (F_act(B%N_basis, B%N_basis))
    allocate (F_core(B%N_basis, B%N_basis))

    if (P%yes_sym) then
      if (allocated(P%freeze)) then
        call prepare_independent_pairs_active_space(active, F%N_occ, pairs, &
          irrep=P%irrep_MO, freeze=F%frozen_MO, activeactive=yes_activeactive)
      else
        call prepare_independent_pairs_active_space(active, F%N_occ, pairs, irrep=P%irrep_MO, activeactive=yes_activeactive)
      end if
    else
      if (allocated(P%freeze)) then
        call prepare_independent_pairs_active_space(active, F%N_occ, pairs, freeze=f%frozen_MO, activeactive=yes_activeactive)
      else
        call prepare_independent_pairs_active_space(active, F%N_occ, pairs, activeactive=yes_activeactive)
      end if
    end if

    if (P%yes_verbose) then
      write (OUTPUT, *) '#independent pairs of MO rotation:'
      write (OUTPUT, '("# ",a4,a4,a4)') "ind", "MO1", "MO2"
      do i = 1, size(pairs)
        write (OUTPUT, '("# ",i4,i4,i4)') i, pairs(i)%mo1, pairs(i)%mo2
      end do
    end if
    allocate (gradient(size(pairs, 1)))
    gradient = ZERO
    allocate (gradient_ls(size(pairs, 1)))
    allocate (theta_step(size(pairs, 1)))
    allocate (theta(size(pairs, 1)))
    allocate (last_theta(size(pairs, 1)))
    allocate (Hessian(size(pairs, 1), size(pairs, 1)))
    allocate (Hessian_copy(size(pairs, 1), size(pairs, 1)))

    ! for BFGS
    allocate (inv_Hessian(size(pairs, 1), size(pairs, 1)))
    allocate (old_inv_Hessian(size(pairs, 1), size(pairs, 1)))
    ! for BFGS
    allocate (old_gradient(size(pairs, 1)))
    allocate (old_theta_step(size(pairs, 1)))
    ! for DIIS
    allocate (old_theta(size(pairs, 1), P%DIIS_nvector))
    allocate (old_err(size(pairs, 1), P%DIIS_nvector))

    old_energy = ZERO
    energy = ZERO
    theta = ZERO
    LC_save = F%LC

    if (P%yes_verbose) then
      call check_time(int_time, timer)
      write (P%OUTPUT, "('# MCSCF time preparation ',f15.3,'sec')") timer - time_init
      time_init = timer     !
    end if
    opdm = ZERO
    write (OUTPUT, '("# ",a5,x,a4,x,a15,x,a15,x,a15,x,a15)') &
      "iter", "inner_it", "RMS grad", "max rotation", "energy diff", "dens diff"
    OUTER_ITERATIONS: do it = 1, P%SCF_max_iter!

      call check_time(int_time, timer)
      time_init = timer     !

      F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)
      call transform_twoel_two_active(P, B, F, active, STDOUT)
      call twoel_allactive_from_two_active(B, F, active)
      call calculate_F_active(active, opdm, F%teint_abij, F%teint_aibj, F_act)
      call calculate_F_inactive(P, B, core, F%LC, F_core)

      E_core = ZERO
      do i = 1, size(core)
        E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
      end do
      if (allocated(P%ci_coef)) then
        csf_c(:) = P%ci_coef(:)
        norm = sqrt(sum(csf_c(:)**2))
        csf_c(:) = csf_c(:)/norm
        if (P%yes_verbose) then
          write (OUTPUT, '("# fixed coefficients:")')
          do j = 1, size(CI_matrix, 1)
            if (abs(csf_c(j)) < 1e-3) cycle
            write (OUTPUT, '("#",a)', advance="no") trim(csfstring(csf_array(j)%occ, csf_array(j)%csf))
            write (OUTPUT, '(f15.8)') csf_c(j)
          end do
        end if
      else
        if (P%yes_verbose) &
          write (OUTPUT, '("# calculating CI matrix ",i10,"x",i10)') size(csf_array, 1), size(csf_array, 1)
        call calc_CIMatrix(UT, E_core, F_core(active(:), active(:)), F%teint_abcd, occs(active(:), :), CI_matrix)

        ! do i=1,size(csf_array,1)
        !   do j=1,i
        !   CI_matrix(i,j)=CI_matrix_element_CSFaa(&
        !        F%H0_MO(active, active),&
        !        F%teint_abcd,&
        !        csf_array(i),&
        !        csf_array(j),&
        !        core,&
        !        active,&
        !        F_core(active,active),&
        !        E_core)
        !     CI_matrix(j,i)=CI_matrix(i,j)
        !   end do
        ! end do
        if (P%yes_verbose) &
          write (OUTPUT, '("# solving CI matrix")')
        call get_eigenvalues(CI_matrix, CI_energy)

        !if(CI_matrix(1,1)<0)CI_matrix(:,:)=-CI_matrix(:,:)

        csf_c = CI_matrix(:, P%state)

        if (P%yes_verbose) then
          do j = 1, size(CI_matrix, 1)
            if (abs(csf_c(j)) < 1e-3) cycle
            write (OUTPUT, '("#",a)', advance="no") trim(csfstring(csf_array(j)%occ, csf_array(j)%csf))
            write (OUTPUT, '(f15.8)') csf_c(j)
          end do
          write (OUTPUT, '("#","CI_energy=",f15.8)') CI_energy(P%state) + nuclear_repulsion_energy(A, P)
        end if
      end if
      old_opdm = opdm

      if (P%yes_verbose) &
        write (OUTPUT, '("# calculating tpdm ")')
      if (P%yes_state_average) then
        ! state-averaged mcscf
        if(allocated( P%SA_weights)) then
          write (P%OUTPUT, '("# ",a,99f5.2)') 'state weights', P%SA_weights
          call one_particle_dm_CI2_SA(UT, occs, CI_matrix(:, 1:size(P%SA_weights)), P%SA_weights(:), opdm)
          if (P%mcscf_type .eq. 3) &
            call s_two_particle_dm_CI_SA(s_tpdm, CI_matrix(:, 1:size(P%SA_weights)), P%SA_weights, csf_array)
          call two_particle_dm_CI2_SA(UT, occs(active(:), :), CI_matrix(:, 1:size(P%SA_weights)), P%SA_weights(:), s_tpdmaa)
        else
          write (P%OUTPUT, '("# ",a,x,a)') 'state weights', 'all states are weighted equally'
          call one_particle_dm_CI2_SA(UT, occs, CI_matrix(:, :), (/((ONE/size(CI_matrix,2)), j=1, size(CI_matrix,2))/), opdm)
          if (P%mcscf_type .eq. 3) &
            call s_two_particle_dm_CI_SA(s_tpdm, CI_matrix(:, :), (/((ONE/size(CI_matrix,2)), j=1, size(CI_matrix,2))/), csf_array)
          call two_particle_dm_CI2_SA(UT, occs(active(:), :), CI_matrix(:, :), &
            (/((ONE/size(CI_matrix,2)), j=1, size(CI_matrix,2))/), s_tpdmaa)
        end if
      else
        !call one_particle_dm_CI(opdm,csf_c,csf_array )
        call one_particle_dm_CI2(UT, occs, csf_c, opdm)
        if (P%mcscf_type .eq. 3) &
          call s_two_particle_dm_CI(s_tpdm, csf_c, csf_array)
        !call s_two_particle_dm_CI(s_tpdmaa,csf_c,csf_array, active)
        call two_particle_dm_CI2(UT, occs(active(:), :), csf_c, s_tpdmaa)
      end if
      old_energy = energy
      energy = calculate_MCSCF_energy(opdm, F%H0_MO, s_tpdmaa, &
        F%teint_abij, F%teint_aibj, F_core, core, active) + nuclear_repulsion_energy(A, P)

      write (OUTPUT, '("##",i5,x,4x,x,es15.8,x,15x,x,es15.8,x,es15.8)') &
        it, sqrt(sum(gradient(:)**2)), energy - old_energy, maxval(abs(opdm - old_opdm))
      if (abs(energy - old_energy) < P%SCF_energy_conv .and. &
        maxval(abs(opdm - old_opdm)) < P%SCF_density_conv .and. it > 1 .and. &
        (inner_it < 2 .or. yes_averageOverAllStates).and. &
        sqrt(sum(gradient(:)**2)) < P%MCSCF_gradient_conv) &
        exit OUTER_ITERATIONS

      delta_energy_CI = abs(energy - old_energy)

      theta(:) = ZERO
      last_theta = ZERO
      LC_save = F%LC
      n_diis = 0
      old_theta(:, :) = ZERO
      old_err(:, :) = ZERO
      old_gradient = ZERO
      old_theta_step(:) = ZERO
      old_inv_hessian = ZERO
      inner_it = 1
      lambda = ONE
      ! inner iterations
      INNER_ITERATIONS: do inner_it = 1, 200

        old_theta_step(:) = theta(:) - last_theta(:)

        if (P%yes_verbose) then
          call check_time(int_time, timer)
          time_inner_it = timer     !
          time_step = timer
        end if

        call calculate_F_active(active, opdm, F%teint_abij, F%teint_aibj, F_act)
        call calculate_F_inactive(P, B, core, F%LC, F_core)
        if (P%yes_verbose) then
          call check_time(int_time, timer)
          write (P%OUTPUT, "('# inner it time for F_act, F_core ',f15.3,'sec')") timer - time_step
          time_step = timer     !
        end if

        call calculate_gradient(F_act, F_core, opdm, s_tpdmaa, F%teint_abij, core, active, virtual, pairs, gradient)
        ! do i=1,size(active,1)
        !   write(6,'(99f10.5)') opdm(active(i),active(:))
        ! end do
        !   do i=1,size(gradient)
        !     write(OUTPUT,'("#",i4,i4,es15.3,es15.3)') pairs(i)%MO1,pairs(i)%MO2, gradient(i), gradient_ls(i)
        !   end do

        if (P%yes_verbose) then
          call check_time(int_time, timer)
          write (P%OUTPUT, "('# inner it time for gradient ',f15.3,'sec')") timer - time_step
          time_step = timer     !
        end if

        if (sqrt(sum(gradient(:)**2)) < P%MCSCF_gradient_conv) then
          if (size(gradient) > 0) &
            write (OUTPUT, '("# ",i5,x,i4,x,es15.8,x,15x,x,15x,x,15x)') it, inner_it, sqrt(sum(gradient(:)**2))
          exit INNER_ITERATIONS
        end if

        if (P%mcscf_type .eq. 3) then
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            time_step = timer     !
          end if
          call transform_twoel(P, B, F, STDOUT)
          call calculate_Lagrangian(opdm, F%H0_MO, s_tpdm, F%teint, Lag)
          call calculate_Hessian(opdm, F%H0_MO, s_tpdm, F%teint, Lag, pairs, Hessian)
          !        do i=1,size(gradient)
          !          do j=1,size(gradient)
          !            write(OUTPUT,'("#",i4,i4,2x,i4,i4,es15.3)') pairs(i)%MO1,pairs(i)%MO2,pairs(j)%MO1,pairs(j)%MO2,Hessian(i,j)
          !          end do
          !        end do
          do
          !          LU_hess=LU_decompose(Hessian)
          !          hess_det=LU_determinant(LU_hess)
            Hessian_copy = Hessian
            hess_det = matrix_determinant(Hessian_copy)
            write (OUTPUT, *) '# Hessian determinant', hess_det
            if (abs(hess_det) > 0.00001) exit
            write (OUTPUT, *) '# LevelShifting Hessian with', 0.01
            do i = 1, size(hessian, 1)
              hessian(i, i) = hessian(i, i) + 0.01
            end do
          end do
          LU_hess = LU_decompose(Hessian)
          hess_det = LU_determinant(LU_hess)
          theta_step(:) = -LU_solve(LU_hess, gradient(:))
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for Hessian ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if

        else if (P%mcscf_type .eq. 2) then
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            time_step = timer     !
          end if

          ! first iteration: diagonal Hessian
          ! all other iterations: quasi Newton using BFGS Hessian
          if (inner_it .eq. 1) then
            ! for the first iteration: invert the diagonal Hessian matrix

            !          call calculate_diagonal_Hessian(opdm, F%H0_MO, s_tpdm, F%teint, Lag, pairs, Hessian)
            !          call calculate_approx_diagonal_Hessian(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,Lag,pairs, Hessian)
            ! call calculate_diagonal_Hessian2(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,Lag,pairs, Hessian_temp)
            ! call calculate_Hessian(opdm, F%H0_MO, s_tpdm, F%teint, Lag, pairs, Hessian_temp2)

            !          call calculate_diagonal_Hessian2(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,Lag,pairs, Hessian)
            !call calculate_approx_diagonal_Hessian_old(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,pairs, Hessian)
            !call calculate_approx_diagonal_Hessian(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint_abij,pairs, Hessian)
            call calculate_approx_diagonal_Hessianaa(core, active, virtual, F_act, F_core, opdm, &
              F%H0_MO, s_tpdmaa, F%teint_abij, pairs, Hessian)


            inv_hessian = ZERO
            do i = 1, size(theta_step, 1)
              inv_hessian(i, i) = ONE/Hessian(i, i)
              nominator = -gradient(i)

              if (abs(Hessian(i, i)) < 1.0e-2) then
                write (OUTPUT, '("# Warning Hessian diagonal small !", 2i4,es20.7)') pairs(i)%MO1, pairs(i)%MO2, Hessian(i, i)
                denominator = 1.0e-2
                inv_hessian(i, i) = 1.0e-2
              else if (Hessian(i, i) < ZERO) then
                write (OUTPUT, '("# Warning Hessian diagonal negative !", 2i4,es20.7)') pairs(i)%MO1, pairs(i)%MO2, Hessian(i, i)
                denominator = -Hessian(i, i)
              else
                denominator = Hessian(i, i)
              end if
              theta_step(i) = nominator/denominator
            end do
          else
            ! BFGS step
            call BFGS_Hessian(old_theta_step, gradient, old_gradient, old_inv_hessian, inv_hessian)
            do i = 1, size(theta_step, 1)
              theta_step(i) = -sum(inv_hessian(i, :)*gradient(:))
            end do
            !          LU_hess=LU_decompose(inv_Hessian)
            !          hess_det=LU_determinant(LU_hess)
            Hessian_copy = inv_Hessian
            hess_det = matrix_determinant(Hessian_copy)
            !write(OUTPUT,'("# inverse hessian determinant: ",es)') hess_det
            if (hess_det < ZERO .or. isnan(hess_det)) then
              !            call calculate_diagonal_Hessian(opdm, F%H0_MO, s_tpdm, F%teint, Lag, pairs, Hessian)
              !call calculate_approx_diagonal_Hessian_old(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,pairs, Hessian)
              !call calculate_approx_diagonal_Hessian(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint_abij,pairs, Hessian)
              call calculate_approx_diagonal_Hessianaa(core, active, virtual, F_act, F_core, opdm, &
                F%H0_MO, s_tpdmaa, F%teint_abij, pairs, Hessian)
              !            call calculate_diagonal_Hessian2(core,active,virtual,F_act,F_core,opdm, F%H0_MO, s_tpdm, F%teint,Lag,pairs, Hessian)
              inv_hessian = ZERO
              do i = 1, size(theta_step, 1)
                inv_hessian(i, i) = ONE/Hessian(i, i)
                nominator = -gradient(i)
                if (abs(Hessian(i, i)) < 1.0e-2) then
                  write (OUTPUT, '("# Warning Hessian diagonal small !", 2i4,es20.7)') pairs(i)%MO1, pairs(i)%MO2, Hessian(i, i)
                  denominator = 1.0e-2
                else if (Hessian(i, i) < ZERO) then
                  write (OUTPUT, '("# Warning Hessian diagonal negative !", 2i4,es20.7)') pairs(i)%MO1, pairs(i)%MO2, Hessian(i, i)
                  denominator = -Hessian(i, i)
                else
                  denominator = Hessian(i, i)
                end if
                theta_step(i) = nominator/denominator
              end do
              ! steepest descent
              !gamma=sum(old_theta_step(:)*(gradient(:)-old_gradient(:)))/sum((gradient(:)-old_gradient(:))**2)
              !theta_step(:)=-gamma*gradient(:)
            end if
          end if ! if it.eq.1
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for BFGS Hessian ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if

        else if (P%mcscf_type .eq. 1) then
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            time_step = timer     !
          end if

          ! steepest descent
          if (inner_it > 1) then
            gamma = sum(old_theta_step(:)*(gradient(:) - old_gradient(:)))/sum((gradient(:) - old_gradient(:))**2)
            theta_step(:) = -gamma*gradient(:)
          else
            theta_step(:) = -gamma*gradient(:)
          end if

          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for steepest descent ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if
        else
          stop "unknown mcscf type"
        end if! choose mcscf type
        max_theta_step = maxval(abs(theta_step))
        if (max_theta_step > 0.2) then
          write (OUTPUT, '("# scaling the step")')
          theta_step = theta_step/max_theta_step*0.2
        end if
        old_gradient = gradient
        old_inv_hessian = inv_hessian

        last_theta = theta(:)
        old_energy = energy

        lambda = ONE
        ls_step = 1
        Energy_lambda(:) = ZERO
        lambda_values(:) = ZERO
        Energy_lambda(ls_step) = energy
        lambda_values(ls_step) = ZERO
        gradient_lambda(ls_step) = sum(gradient(:)*theta_step(:))
        LINESEARCH_LOOP: do
          theta(:) = last_theta(:) + lambda*theta_step(:)
          expected_energy = old_energy + lambda*sum(gradient(:)*theta_step(:))
          F%LC = LC_save

          if (P%yes_verbose) then
            call check_time(int_time, timer)
            time_step = timer     !
          end if

          call rotate_orbitals(theta, pairs, F%LC)
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for rotate orbitals ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if

          F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)
          !call transform_twoel(P, B ,F, STDOUT )
          !if(P%yes_verbose)then
          !  call check_time( int_time, timer )
          !  write(P%OUTPUT,"('# inner it time for transform teints ',f15.3,'sec')") timer-time_step
          !  time_step = timer     !
          !end if

          call transform_twoel_two_active(P, B, F, active, STDOUT)
          call twoel_allactive_from_two_active(B, F, active)
          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for transform teints ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if

          call calculate_F_active(active, opdm, F%teint_abij, F%teint_aibj, F_act)
          call calculate_F_inactive(P, B, core, F%LC, F_core)
          energy = calculate_MCSCF_energy(opdm, F%H0_MO, s_tpdmaa, &
            F%teint_abij, F%teint_aibj, F_core, core, active) + nuclear_repulsion_energy(A, P)

          if (P%yes_verbose) then
            call check_time(int_time, timer)
            write (P%OUTPUT, "('# inner it time for calc energy ',f15.3,'sec')") timer - time_step
            time_step = timer     !
          end if

          call calculate_gradient(F_act, F_core, opdm, s_tpdmaa, F%teint_abij, core, active, virtual, pairs, gradient_ls)

          ls_step = ls_step + 1
          Energy_lambda(ls_step) = energy
          lambda_values(ls_step) = lambda
          gradient_lambda(ls_step) = sum(gradient_ls(:)*theta_step(:))

          if (P%yes_verbose) then
            if (ls_step > 2) then
              write (6, "('#Armijo 1 c1=0.1:', es20.7,'<=',es20.7, l)") &
                Energy_lambda(ls_step) - Energy_lambda(ls_step - 1), 0.1*lambda_values(ls_step - 1)*gradient_lambda(ls_step - 1), &
                Energy_lambda(ls_step) - Energy_lambda(ls_step - 1) .le. 0.1*lambda_values(ls_step - 1)*gradient_lambda(ls_step - 1)
              write (6, "('#Wolfe condition c2=0.9:', es20.7,'<=',es20.7, l)") &
                abs(gradient_lambda(ls_step)), 0.9*abs(gradient_lambda(ls_step - 1)), &
                abs(gradient_lambda(ls_step)) .le. 0.9*abs(gradient_lambda(ls_step - 1))
            end if
            write (6, "('#(g-gold)**2=',es20.7)") sum((gradient_ls(:) - old_gradient(:))**2)
            write (6, "('#|(g)|',es20.7)") sqrt(sum(gradient_ls(:)**2))
            write (6, "('#(g*g_old)/|g||g_old|',es20.7)") &
              sum(gradient_ls(:)*old_gradient(:))/sqrt(sum(gradient_ls(:)**2)*sum(old_gradient(:)**2))
            !        write(6,*) 'old_step*g_old',sum(old_theta_step(:)*old_gradient(:))/sqrt(sum(old_theta_step(:)**2)*sum(old_gradient(:)**2) )
            write (6, "('#Expected Step:',es20.7)") expected_energy - old_energy
            write (6, "('#Got:',es20.7)") energy - old_energy
            write (6, "('#Got-Exp:      ',es20.7)") energy - expected_energy
            write (6, "('#Ratio:        ',es20.7)") abs(expected_energy - old_energy)/abs(energy - old_energy)
            write (6, "('#step length',es20.7)") lambda*sqrt(sum(theta_step**2))
            write (6, "('#estimated 2nd deriv. ',es20.7)") TWO*(energy - expected_energy)/(lambda**2*sum(theta_step**2))
            write (6, "('#estimated 1st deriv. ',es20.7)") (expected_energy - old_energy)/(lambda*sqrt(sum(theta_step**2)))
            write (6, "('#basic step length:',es20.7)") sqrt(sum(theta_step(:)**2))
            write (6, "('#line search steps: ')")
            write (6, '("#LS ",a4,a15,a25,a15,a15)') 'step', 'lambda', 'Energy', 'gradient*step', '|gradient|'
            do i = 1, ls_step
              write (6, '("#LS ",i4,es15.5,f25.10,es15.5,es15.5)') &
                i, lambda_values(i), Energy_lambda(i), gradient_lambda(i), &
                abs(gradient_lambda(i))/sqrt(sum(theta_step(:)**2))
            end do
          end if! verbose
          ! four criterions for abort:
          ! 1. we reach the desired convergence of 1e-4
          if (abs(gradient_lambda(ls_step))/sqrt(sum(theta_step(:)**2)) < P%MCSCF_gradient_conv) exit LINESEARCH_LOOP
          ! 2. the step size becomes too small
          if (lambda*sqrt(sum(theta_step(:)**2)) < P%MCSCF_gradient_conv) exit LINESEARCH_LOOP
          ! 3. too many iterations
          if (ls_step .eq. NLS) exit LINESEARCH_LOOP
          ! 4. if mcscf_type=1 or  mcscf_type=3
          if (P%mcscf_type .eq. 1 .or. P%mcscf_type .eq. 3) exit LINESEARCH_LOOP
          ! 5. Armijoo rules
          !if( energy_lambda(ls_step) - energy_lambda(ls_step-1)<= 0.1 * lambda_values(ls_step-1) * gradient_lambda(ls_step-1) .and. &
          !     -gradient_lambda(ls_step)<= -0.9*gradient_lambda(ls_step-1)) exit LINESEARCH_LOOP
          ! 6. Wolfe condition
          !if( ls_step>2 .and. energy_lambda(ls_step) <= energy_lambda(ls_step-1) + 0.1 * lambda_values(ls_step-1) * gradient_lambda(ls_step-1) .and. &
          !     abs(gradient_lambda(ls_step)) <= 0.9*abs(gradient_lambda(ls_step-1))) exit LINESEARCH_LOOP
          ! 2. the step size becomes too large
          if (lambda > 3) exit LINESEARCH_LOOP

          lambda = linesearch(energy_lambda, gradient_lambda, lambda_values, ls_step)
        end do LINESEARCH_LOOP
        theta_step = theta(:) - last_theta
        gradient = gradient_ls
        ! save current theta, theta_step, gradient and inv_hessian for later usage
        if (n_diis < P%DIIS_nvector) &
          n_diis = n_diis + 1
        do i = n_diis, 2, -1
          old_theta(:, i) = old_theta(:, i - 1)
          old_err(:, i) = old_err(:, i - 1)
        end do
        old_theta(:, 1) = theta(:)
        old_err(:, 1) = theta(:) - last_theta(:)
        ! DIIS does not work here
        ! if(n_diis>1 .and. P%yes_DIIS .and. P%mcscf_type.eq.3 .and. sqrt(sum(gradient(:)**2)) > 1e-4)then
        !   theta(:)=diis_step(old_theta,old_err,n_diis)
        ! end if
        if (size(gradient) > 0) &
          write (OUTPUT, '("# ",i5,x,i4,x,es15.8,x,es15.8,x,es15.8,x,15x,i4)') &
          it, inner_it, sqrt(sum(gradient(:)**2)), sqrt(sum(theta_step(:)**2)), energy - old_energy, ls_step
        if (sqrt(sum(theta_step(:)**2)) < P%MCSCF_gradient_conv .and. inner_it > 3) then
          exit INNER_ITERATIONS
        end if

        ! if energy changes are 5% of the energy changes due to CI, exit outerloop - only if activeactive=False
        if (abs(energy - old_energy) < 0.05*delta_energy_CI .and. .not. yes_activeactive .and. .not. yes_averageOverAllStates) then
          exit INNER_ITERATIONS
        end if
        ! if enery is change less than 1e-6, make an outer iteration
        !if( abs(energy - old_energy) < 1e-6 .and. inner_it >10)then
        !  exit INNER_ITERATIONS
        !end if

        ! if energy rises, better make an outer iteration step
        if (energy - old_energy > 0) &
          exit INNER_ITERATIONS

      end do INNER_ITERATIONS ! inner iterations

      !call canonicalize_orbitals(P, F, core, active, virtual, F_core, F_act, J_act,K_act)
    end do OUTER_ITERATIONS! it


    F%yes_converged = .TRUE.
    energy = calculate_MCSCF_energy(opdm, F%H0_MO, s_tpdmaa, &
      F%teint_abij, F%teint_aibj, F_core, core, active) + nuclear_repulsion_energy(A, P)
    !  write(OUTPUT,'("CI_energy=",f15.8)')  CI_energy(P%state) + nuclear_repulsion_energy( A, P )
    write (OUTPUT, '("#MCSCF_energy=",f15.8)') energy
    do j = 1, size(CI_matrix, 1)
      if (abs(csf_c(j)) < 1e-3) cycle
      write (OUTPUT, '("#",a)', advance="no") trim(csfstring(csf_array(j)%occ, csf_array(j)%csf))
      !    write(OUTPUT,'(f15.8)')  CI_matrix(j,P%state)
      write (OUTPUT, '(f15.8)') csf_c(j)
    end do

    call canonicalize_orbitals(P, F, core, active, virtual, core_nFz, active_nFz, virtual_nFz, &
      F_core, F_act, opdm, s_tpdmaa, yes_completeActiveSpace=allocated(P%active_space_MO))

    ! perform final CI with canonical orbitals
    if (.not. allocated(P%ci_coef)) then
      F%H0_MO = F%LC(:, :) .tm.B%H0(:, :) .m.F%LC(:, :)
      call transform_twoel_two_active(P, B, F, active, STDOUT)
      call twoel_allactive_from_two_active(B, F, active)
      call calculate_F_active(active, opdm, F%teint_abij, F%teint_aibj, F_act)
      call calculate_F_inactive(P, B, core, F%LC, F_core)
      E_core = ZERO
      do i = 1, size(core)
        E_core = E_core + F_core(core(i), core(i)) + F%H0_MO(core(i), core(i))
      end do
      call calc_CIMatrix(UT, E_core, F_core(active(:), active(:)), F%teint_abcd, occs(active(:), :), CI_matrix)

      call get_eigenvalues(CI_matrix, CI_energy)
      do j = 1, size(CI_matrix, 1)
        write (OUTPUT, '("# state nr ",i4,f15.8)') j, CI_energy(j) + nuclear_repulsion_energy(A, P)
      end do
      csf_c = CI_matrix(:, P%state)
      write (OUTPUT, '("#CI_energy=",f15.8,x, "iter=",i4)') CI_energy(P%state) + nuclear_repulsion_energy(A, P), it
      do j = 1, size(CI_matrix, 1)
        if (abs(csf_c(j)) < 1e-3) cycle
        write (OUTPUT, '("#",a)', advance="no") trim(csfstring(csf_array(j)%occ, csf_array(j)%csf))
        !    write(OUTPUT,'(f15.8)')  CI_matrix(j,P%state)
        write (OUTPUT, '(f15.8)') csf_c(j)
      end do
    end if

    if (P%yes_state_average) then
      !call one_particle_dm_CI_SA(opdm,CI_matrix(:,1:size(P%SA_weights)),P%SA_weights,csf_array )
      if(allocated( P%SA_weights)) then
        call one_particle_dm_CI2_SA(UT, occs, CI_matrix(:, 1:size(P%SA_weights)), P%SA_weights(:), opdm)
      else
        call one_particle_dm_CI2_SA(UT, occs, CI_matrix(:, :), (/(ONE/size(CI_matrix,2), j=1, size(CI_matrix,2))/), opdm)
      end if
    else
      call one_particle_dm_CI2(UT, occs(:, :), csf_c, opdm)
      !call one_particle_dm_CI(opdm,csf_c,csf_array )
    end if

    ! put density matrix into F%D
    F%D = F%LC.m.opdm.mt.F%LC(:, :)
    !
    F%f_occ(virtual(:)) = ZERO
    F%f_occ(core(:)) = TWO
    if (P%yes_nat_orb) then
      call natural_orbitals(P, F, active, opdm, F_act, F_core)
      if (P%molden_file .ne. "") then
        write (moldenfilename, '("natorb.molden")')
        call write_molden_file(P, A, B, F, moldenfilename)
      end if
      call print_orbitals(P, B, F, "natural orbitals")
    else
      do i = 1, size(active)
        F%f_occ(active(i)) = opdm(active(i), active(i))
      end do
      call print_orbitals(P, B, F, "canonical orbitals")
    end if

    ! do localization
    if (allocated(P%localize)) then
      call localize_orbitals(P, G, B, F, P%localize)
      if (allocated(P%fix_nuc_MO) .and. allocated(P%fix_nuc_n)) &
        call fix_orbitals_on_nuclei(P, B, F, P%fix_nuc_MO, P%fix_nuc_n)
      if (P%print_level > 1) then
        call print_orbitals(P, B, F, "*** Localized Orbitals *** ")
      end if
    end if

    if (P%yes_SaveIni) then
      call save_wf(F%LC)
    end if

    do i = 1, size(csf_array)
      if (associated(csf_array(i)%dets)) then
        deallocate (csf_array(i)%dets)
        nullify (csf_array(i)%dets)
        do j = i + 1, size(csf_array)
          if (any(csf_array(i)%occ .ne. csf_array(j)%occ)) exit
          nullify (csf_array(j)%dets)
        end do
      end if
      deallocate (csf_array(i)%csf)
    end do
    deallocate (occs)
    deallocate (csf_c)
    deallocate (old_theta)
    deallocate (old_theta_step)
    deallocate (old_gradient)
    deallocate (old_inv_Hessian)
    deallocate (Hessian)
    deallocate (Hessian_copy)
    deallocate (theta)
    deallocate (theta_step)
    deallocate (gradient)
    deallocate (LC_save)

    if (P%mcscf_type .eq. 3) then
      deallocate (s_tpdm)
      deallocate (Lag)
    end if
    deallocate (s_tpdmaa)
    deallocate (opdm)
    deallocate (CI_energy)
    deallocate (CI_matrix)
    deallocate (h0)

    deallocate (active)
    deallocate (virtual)
    deallocate (core)
    deallocate (occupied)
    if (allocated(active_nFz)) &
      deallocate (active_nFz)
    if (allocated(virtual_nFz)) &
      deallocate (virtual_nFz)
    if (allocated(core_nFz)) &
      deallocate (core_nFz)
    call purge_UMatrixTable(UT)

  end subroutine MCSCF_procedure

  subroutine canonicalize_orbitals(P, F, core, active, virtual, core_nFz, active_nFz, virtual_nFz, F_core, F_act, opdm, s_tpdmaa, yes_completeActiveSpace)
    !! canonicalize orbitals
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: core(:) !! index array for core orbitals
    integer, intent(in) :: active(:) !! index array for active orbitals
    integer, intent(in) :: virtual(:) !! index array for virtual orbitals
    integer, intent(in) :: core_nFz(:) !! index array for core orbitals that are not frozen
    integer, intent(in) :: active_nFz(:) !! index array for active orbitals that are not frozen
    integer, intent(in) :: virtual_nFz(:) !! index array for virtual orbitals that are not frozen
    real(long), intent(in) :: F_act(:, :) !! active Fock matrix
    real(long), intent(in) :: F_core(:, :) !! core Fock matrix
    real(long), intent(in) :: opdm(:, :) !! one-particle density matrix
    real(long), intent(in) :: s_tpdmaa(:) !! symmetrized two-particle density matrix
    logical, intent(in), optional :: yes_completeActiveSpace !! do a re-diagonalization of the active-space orbitals
    logical :: completeActiveSpace
    real(long) :: genFock(size(F_act, 1), size(F_act, 2))
    integer :: i

    real(long), allocatable :: H_virt(:, :), CE_virt(:)
    real(long), allocatable :: H_core(:, :), CE_core(:)
    real(long), allocatable :: H_active(:, :), CE_active(:)

    real(long) :: LC(size(F%LC, 1), size(F%LC, 2))
    real(long) :: CE(size(F%CE, 1))
    completeActiveSpace = .False.
    if(present(yes_completeActiveSpace)) completeActiveSpace = yes_completeActiveSpace
    call calculate_genFock(F_act, F_core, opdm, s_tpdmaa, F%teint_abij, core, active, virtual, genFock)

    LC(:, :) = F%LC
    ! canonicalize core and virtual orbitals
    if (size(virtual) > 0) then
      allocate (H_virt(size(virtual), size(virtual)))
      H_virt = F_core(virtual(:), virtual(:)) + F_act(virtual(:), virtual(:))
      do i = 1, size(H_virt, 1)
        CE(virtual(i)) = H_virt(i, i)
      end do
      deallocate (H_virt)
    end if
    if (size(virtual_nFz) > 0) then
      allocate (H_virt(size(virtual_nFz), size(virtual_nFz)))
      allocate (CE_virt(size(virtual_nFz)))
      ! Davidson
      !H_virt=F_core(virtual_nFz(:),virtual_nFz(:))+ (J_act(virtual_nFz(:),virtual_nFz(:))-ONE*K_act(virtual_nFz(:),virtual_nFz(:)))
      ! Guest Saunders
      H_virt = F_core(virtual_nFz(:), virtual_nFz(:)) + F_act(virtual_nFz(:), virtual_nFz(:))
      !H_virt = genFock(virtual_nFz,virtual_nFz)
      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_virt, CE_virt, F%irrep_mo(virtual_nFz(:)))
      else
        call get_eigenvalues(H_virt, CE_virt)
      end if
      !call reorder_RSTRCT(P,H_virt,CE_virt)
      LC(:, virtual_nFz(:)) = matmul(F%LC(:, virtual_nFz(:)), H_virt(:, :))
      CE(virtual_nFz) = CE_virt
      deallocate (CE_virt)
      deallocate (H_virt)
    end if

    if (size(active) > 0) then
      allocate (H_active(size(active), size(active)))
      H_active = F_core(active(:), active(:)) + (F_act(active(:), active(:)))
      do i = 1, size(H_active, 1)
        CE(active(i)) = H_active(i, i)
      end do
      deallocate (H_active)
    end if

    if (size(active_nFz) > 0 .and. completeActiveSpace) then
      allocate (H_active(size(active_nFz), size(active_nFz)))
      allocate (CE_active(size(active_nFz)))
      !H_active = genFock(active_nFz,active_nFz)
      ! Davidson
      !H_active=F_core(active_nFz(:),active_nFz(:))+ONE*(J_act(active_nFz(:),active_nFz(:))-ONE*K_act(active_nFz(:),active_nFz(:)))
      ! Guest Saunders
      H_active = F_core(active_nFz(:), active_nFz(:)) + (F_act(active_nFz(:), active_nFz(:)))

      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_active, CE_active, F%irrep_mo(active_nFz(:)))
      else
        call get_eigenvalues(H_active, CE_active)
        !call reorder_RSTRCT(P,H_active,CE_active)
      end if

      LC(:, active_nFz(:)) = matmul(F%LC(:, active_nFz(:)), H_active(:, :))
      CE(active_nFz) = CE_active
      deallocate (H_active)
      deallocate (CE_active)
    end if

    if (size(core) > 0) then
      allocate (H_core(size(core), size(core)))
      H_core = F_core(core(:), core(:)) + F_act(core(:), core(:))
      do i = 1, size(H_core, 1)
        CE(core(i)) = H_core(i, i)
      end do
      deallocate (H_core)
    end if

    if (size(core_nFz) > 0) then
      allocate (H_core(size(core_nFz), size(core_nFz)))
      allocate (CE_core(size(core_nFz)))

      !H_core = HALF*genFock(core_nFz,core_nFz)
      H_core = F_core(core_nFz(:), core_nFz(:)) + F_act(core_nFz(:), core_nFz(:))
      if (P%yes_sym) then
        call get_eigenvalues_symmetry(H_core, CE_core, F%irrep_mo(core_nFz(:)))
      else
        call get_eigenvalues(H_core, CE_core)
        !call reorder_RSTRCT(P,H_core,CE_core)
      end if
      LC(:, core_nFz(:)) = matmul(F%LC(:, core_nFz(:)), H_core(:, :))
      CE(core_nFz) = CE_core
      deallocate (H_core)
      deallocate (CE_core)
    end if

    F%LC = LC
    F%CE = CE
  end subroutine canonicalize_orbitals

  subroutine natural_orbitals(P, F, active, opdm, F_act, F_core)
    !! prepares natural orbitals
    type(Param), intent(in) :: P
    type(WF_data), intent(inout) :: F
    integer, intent(in) :: active(:)
    real(long), intent(in) :: opdm(:, :)
    real(long) :: opdm_active(size(active), size(active))
    real(long) :: nocc_active(size(active))
    real(long), intent(in) :: F_act(:, :)
    real(long), intent(in) :: F_core(:, :)
    real(long) :: LC(size(F%LC, 1), size(F%LC, 2))
    real(long) :: F_eff(size(active), size(active))
    integer :: i
    opdm_active = opdm(active, active)
    F_eff = F_act(active, active) + F_core(active, active)
    if (P%yes_sym .and. P%state_irrep > 0) then
      call get_eigenvalues_symmetry(opdm_active, nocc_active, F%irrep_mo(active))
    else
      call get_eigenvalues(opdm_active, nocc_active)
    end if
    call reorder_RSTRCT(opdm_active, nocc_active)
    LC = F%LC
    F%F_occ(active) = nocc_active
    LC(:, active) = matmul(F%LC(:, active), opdm_active)
    F_eff = opdm_active.tm.F_eff.m.opdm_active
    do i = 1, size(opdm_active, 1)
      F%CE(active(i)) = F_eff(i, i)
    end do
    F%LC = LC
  end subroutine natural_orbitals

  ! calculates the Hessian matrix for the rotation of orbitals pairs \theta_ij, orbital pairs are
  ! given in array pairs
  ! This subroutine is adapted from psi3
  ! Note that it works with the symmetrized two particle density matrix!
  ! d^2 E / (Theta_{pq} Theta_{rs}) =
  !   y_{pqrs} - y_{qprs} - y_{pqsr} + y_{qpsr}
  !   + 1/2 delta_{ps} ( x_{qr} + x_{rq} )
  !   - 1/2 delta_{qs} ( x_{pr} + x_{rp} )
  !   - 1/2 delta_{pr} ( x_{qs} + x_{sq} )
  !   + 1/2 delta_{qr} ( x_{ps} + x_{sp} )
  ! x is the Lagrangian
  ! y_{pqrs} = gamma_{qs}
  !          + \sum_{mn} ( 2 Gamma_{qsmn} (pr|mn) + 4 Gamma_{qmsn} (pm|rn) )
  subroutine calculate_Hessian(opdm, oeints, s_tpdm, teints, Lag, pairs, Hessian)
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: oeints(:, :)
    !  real(long), intent(in) :: tpdm(:,:,:,:)
    real(long), intent(in) :: s_tpdm(:)
    real(long), intent(in) :: teints(:)
    real(long), intent(in) :: Lag(:, :)
    type(indep_mo_pair), intent(in) :: pairs(:)
    real(long), intent(out) :: Hessian(:, :)
    !  real(long), allocatable :: s_tpdm(:)
    integer :: pair1, pair2
    integer :: p, q, r, s, m, n
    integer(8) :: rs, pq, pm, qm, ps, qs, pr, qr, mn, rn, sn, qrmn, qmrn, psmn, pmsn, prmn, pmrn, qsmn, qmsn
    real(long) :: y

    !write(OUTPUT,'("# Hessian")')
    ! assemble Hessian
    Hessian = ZERO
    do pair1 = 1, size(pairs, 1)
      p = pairs(pair1)%mo1
      q = pairs(pair1)%mo2
      pq = index_packed(p, q)
      do pair2 = 1, pair1
        r = pairs(pair2)%mo1
        s = pairs(pair2)%mo2
        rs = index_packed(r, s)
        pr = index_packed(p, r)
        ps = index_packed(p, s)
        qr = index_packed(q, r)
        qs = index_packed(q, s)
        Hessian(pair1, pair2) = ZERO
        if (p .eq. s) &
          Hessian(pair1, pair2) = Hessian(pair1, pair2) + HALF*(Lag(q, r) + Lag(r, q))
        if (q .eq. s) &
          Hessian(pair1, pair2) = Hessian(pair1, pair2) - HALF*(Lag(p, r) + Lag(r, p))
        if (p .eq. r) &
          Hessian(pair1, pair2) = Hessian(pair1, pair2) - HALF*(Lag(q, s) + Lag(s, q))
        if (q .eq. r) &
          Hessian(pair1, pair2) = Hessian(pair1, pair2) + HALF*(Lag(p, s) + Lag(s, p))

          !      write(OUTPUT,'(4i4,f15.6)') p,q,r,s,Hessian(pair1,pair2)

        y = ZERO
        ! y_pqrs
        y = y + oeints(p, r)*opdm(q, s)
        !      write(OUTPUT,'("<p|h|r> * <q|s> ",f15.6)') oeints(p,r)*opdm(q,s)
        !      write(OUTPUT,'("<p|h|r> ",f15.6)') oeints(p,r)
        !      write(OUTPUT,'("<q|s> ",f15.6)') opdm(q,s)
        !      write(OUTPUT,'("<q|h|s> ",f15.6)') oeints(q,s)
        !      write(OUTPUT,'("<p|r> ",f15.6)') opdm(p,r)

        do m = 1, size(opdm, 1)
          qm = index_packed(q, m)
          pm = index_packed(p, m)
          do n = 1, size(opdm, 1)
            mn = index_packed(m, n)
            rn = index_packed(r, n)
            sn = index_packed(s, n)
            qsmn = index_packed(qs, mn)
            qmsn = index_packed(qm, sn)
            prmn = index_packed(pr, mn)
            pmrn = index_packed(pm, rn)
            y = y + ONE*s_tpdm(qsmn)*teints(prmn)
            y = y + TWO*s_tpdm(qmsn)*teints(pmrn)
          end do
        end do
        !      write(OUTPUT,'("y_pqrs",f15.6)') y
        Hessian(pair1, pair2) = Hessian(pair1, pair2) + y
        ! y_qprs
        y = ZERO
        y = y + oeints(q, r)*opdm(p, s)
        do m = 1, size(opdm, 1)
          qm = index_packed(q, m)
          pm = index_packed(p, m)
          do n = 1, size(opdm, 1)
            mn = index_packed(m, n)
            rn = index_packed(r, n)
            sn = index_packed(s, n)
            psmn = index_packed(ps, mn)
            pmsn = index_packed(pm, sn)
            qrmn = index_packed(qr, mn)
            qmrn = index_packed(qm, rn)
            y = y + ONE*s_tpdm(psmn)*teints(qrmn)
            y = y + TWO*s_tpdm(pmsn)*teints(qmrn)
          end do
        end do
        !      write(OUTPUT,'("y_qprs",f15.6)') -y
        Hessian(pair1, pair2) = Hessian(pair1, pair2) - y
        ! y_pqsr
        y = ZERO
        y = y + oeints(p, s)*opdm(q, r)
        do m = 1, size(opdm, 1)
          qm = index_packed(q, m)
          pm = index_packed(p, m)
          do n = 1, size(opdm, 1)
            mn = index_packed(m, n)
            rn = index_packed(r, n)
            sn = index_packed(s, n)
            qrmn = index_packed(qr, mn)
            qmrn = index_packed(qm, rn)
            psmn = index_packed(ps, mn)
            pmsn = index_packed(pm, sn)
            y = y + ONE*s_tpdm(qrmn)*teints(psmn)
            y = y + TWO*s_tpdm(qmrn)*teints(pmsn)
          end do
        end do
        !      write(OUTPUT,'("y_pqsr",f15.6)') -y
        Hessian(pair1, pair2) = Hessian(pair1, pair2) - y
        ! y_qpsr
        y = ZERO
        y = y + oeints(q, s)*opdm(p, r)
        do m = 1, size(opdm, 1)
          qm = index_packed(q, m)
          pm = index_packed(p, m)
          do n = 1, size(opdm, 1)
            mn = index_packed(m, n)
            rn = index_packed(r, n)
            sn = index_packed(s, n)
            prmn = index_packed(pr, mn)
            pmrn = index_packed(pm, rn)
            qsmn = index_packed(qs, mn)
            qmsn = index_packed(qm, sn)
            y = y + ONE*s_tpdm(prmn)*teints(qsmn)
            y = y + TWO*s_tpdm(pmrn)*teints(qmsn)
          end do
        end do
        !      write(OUTPUT,'("y_qpsr",f15.6)') y
        Hessian(pair1, pair2) = Hessian(pair1, pair2) + y

        Hessian(pair1, pair2) = TWO*Hessian(pair1, pair2)
        Hessian(pair2, pair1) = Hessian(pair1, pair2)
      !      write(OUTPUT,'(4i4,f15.7)') p,q,r,s,Hessian(pair1,pair2)
      end do
    end do
    !  deallocate(s_tpdm)
  end subroutine calculate_Hessian



  subroutine calculate_approx_diagonal_Hessianaa(core, active, virtual, F_act, F_core, opdm, &
    oeints, s_tpdmaa, teints_abij, pairs, Hessian)
    integer, intent(in) :: core(:)
    integer, intent(in) :: active(:)
    integer, intent(in) :: virtual(:)

    real(long), intent(in) :: F_act(:, :)
    real(long), intent(in) :: F_core(:, :)
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: oeints(:, :)
    real(long), intent(in) :: s_tpdmaa(:)
    real(long), intent(in) :: teints_abij(:)
    type(indep_mo_pair), intent(in) :: pairs(:)
    real(long), intent(out) :: Hessian(:, :)
    integer :: pair1
    integer :: a, i, p, q, t, u, v, w
    integer(8) :: tu, vw, tt, ii, aa, ai, tuaa
    integer(8) :: tuvw, aaii, aiai
    integer :: np
    integer(8) :: npq
    np = size(oeints, 1)
    npq = np*(np + 1)/2
    ! assemble Hessian
    Hessian = ZERO
    do pair1 = 1, size(pairs, 1)
      p = pairs(pair1)%mo1
      q = pairs(pair1)%mo2
      if ((ANY(p .eq. core) .and. any(q .eq. virtual)) .or. (ANY(p .eq. virtual) .and. any(q .eq. core))) then
        if (ANY(p .eq. core) .and. any(q .eq. virtual)) then
          i = p
          a = q
        else
          i = q
          a = p
        end if
        ai = index_packed(a, i)
        aa = index_packed(a, a)
        ii = index_packed(i, i)
        aiai = index_packed(ai, ai)
        aaii = index_packed(aa, ii)
        Hessian(pair1, pair1) = FOUR*(F_core(a, a) + F_act(a, a) - F_core(i, i) - F_act(i, i))
        !      +THREE * teints(aiai) - teints(aaii) )
      else if ((ANY(p .eq. virtual) .and. any(q .eq. active)) .or. (ANY(p .eq. active) .and. any(q .eq. virtual))) then
        if (ANY(p .eq. virtual) .and. any(q .eq. active)) then
          a = p
          t = q
        else
          t = p
          a = q
        end if
        tt = index_packed(t, t)
        aa = index_packed(a, a)
        Hessian(pair1, pair1) = TWO*opdm(t, t)*(F_core(a, a) + F_act(a, a))
        do u = 1, size(active)
          tuaa = index_packed(findloc(active, t, dim=1), u)
          tu = index_packed(t, active(u))
          Hessian(pair1, pair1) = Hessian(pair1, pair1) - TWO*opdm(t, active(u))*F_core(t, active(u))
          do v = 1, size(active)
            do w = 1, size(active)
              vw = index_packed(v, w)
              tuvw = index_packed(tuaa, vw)
              Hessian(pair1, pair1) = Hessian(pair1, pair1) - TWO*s_tpdmaa(tuvw)*teints_abij((vw - 1)*npq + tu)
            end do! w
          end do! v
        end do! u
      else if ((ANY(p .eq. active) .and. any(q .eq. core)) .or. (ANY(p .eq. core) .and. any(q .eq. active))) then
        if (ANY(p .eq. active) .and. any(q .eq. core)) then
          t = p ! active
          i = q ! core
        else
          t = q
          i = p
        end if
        Hessian(pair1, pair1) = FOUR*(F_core(t, t) + F_act(t, t) - F_core(i, i) - F_act(i, i))!
        Hessian(pair1, pair1) = Hessian(pair1, pair1) + TWO*opdm(t, t)*(F_core(i, i) + F_act(i, i))!
        Hessian(pair1, pair1) = Hessian(pair1, pair1) - TWO*sum(opdm(t, active(:))*F_core(t, active(:)))!
        do u = 1, size(active)
          tu = index_packed(t, active(u))
          tuaa = index_packed(findloc(active, t, dim=1), u)
          do v = 1, size(active)
            do w = 1, size(active)
              vw = index_packed(v, w)
              tuvw = index_packed(tuaa, vw)
              Hessian(pair1, pair1) = Hessian(pair1, pair1) - TWO*s_tpdmaa(tuvw)*teints_abij((vw - 1)*npq + tu)!
            end do! w
          end do! v
        end do! u
      end if

    end do! pair1
  end subroutine calculate_approx_diagonal_Hessianaa

  subroutine calculate_F_active(active, opdm, teints_abij, teints_aibj, F_active)
    integer, intent(in) :: active(:)
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: teints_abij(:)
    real(long), intent(in) :: teints_aibj(:)
    real(long), intent(out) :: F_active(:, :)
    real(long) :: T1, T2
    integer :: p, q, u, v
    integer(8) :: pq, pu
    integer(8) :: pquv, puqv
    integer :: np
    integer(8) :: npq
    np = size(opdm, 1)
    npq = np*(np + 1)/2
    do p = 1, size(opdm, 1)
      do q = 1, size(opdm, 1)
        pq = index_packed(p, q)
        F_active(p, q) = ZERO
        do u = 1, size(active, 1)
          pu = index_packed(p, active(u))
          do v = 1, size(active, 1)
            pquv = (index_packed(u, v) - 1)*npq + pq
            puqv = index_packed((u - 1)*np + p, (v - 1)*np + q)
            T1 = teints_abij(pquv)
            T2 = teints_aibj(puqv)
            F_active(p, q) = F_active(p, q) + opdm(active(u), active(v))*(T1 - HALF*T2)
          end do
        end do
      end do
    end do
  end subroutine calculate_F_active


  subroutine calculate_F_inactive(P, B, core, LC, F_inactive)
    !! return the inactive Fock matrix
    !! i.e. from the inactive orbital's density matrix
    type(Param), intent(in) :: P
    type(Basis), intent(in) :: B
    integer, intent(in) :: core(:)
    real(long), intent(in) :: LC(:, :)
    real(long) :: Dcore(size(LC, 1), size(LC, 1))
    real(long) :: F_inactive_AO(size(LC, 1), size(LC, 1))
    real(long), intent(out) :: F_inactive(size(LC, 2), size(LC, 2))
    integer :: j, k
    real(long) :: LCcore(size(LC, 1), size(core, 1))
    real(long) :: LCcoreT(size(core, 1), size(LC, 1))
    LCcore = LC(:, core(:))
    LCcoreT = transpose(LCcore)
    Dcore(:, :) = ZERO
    do j = 1, size(LC, 1)
      do k = 1, size(LC, 1)
        Dcore(j, k) = Dcore(j, k) + TWO*sum(LCcoreT(:, j)*LCcoreT(:, k))
      end do
    end do
    if (P%yes_direct_scf) then
      call construct_Fock_Matrix_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, Dcore, B%h0, F_inactive_AO, accuracy=P%eps_teint)
    else
      call construct_Fock_Matrix2(B%h0, B%teint, Dcore, F_inactive_AO)
    end if
    F_inactive = LC.tm.F_inactive_AO.m.LC
  end subroutine calculate_F_inactive

  subroutine calculate_F_inactive_direct(B, core, LC, F_inactive, accuracy)
    !! return the inactive Fock matrix
    !! i.e. from the inactive orbital's density matrix
    !! in a direct-SCF context (without pre-calculated teints)
    type(Basis), intent(in) :: B
    integer, intent(in) :: core(:)
    real(long), intent(in) :: LC(:, :)
    real(long) :: Dcore(size(LC, 1), size(LC, 1))
    real(long) :: F_inactive_AO(size(LC, 1), size(LC, 1))
    real(long), intent(out) :: F_inactive(size(LC, 2), size(LC, 2))
    real(long), optional :: accuracy
    integer :: j, k

    real(long) :: LCcore(size(LC, 1), size(core, 1))
    real(long) :: LCcoreT(size(core, 1), size(LC, 1))
    real(long) :: epsilon
    epsilon = ZERO
    if (present(accuracy)) &
      epsilon = accuracy
    LCcore = LC(:, core(:))
    LCcoreT = transpose(LCcore)
    Dcore(:, :) = ZERO
    do j = 1, size(LC, 1)
      do k = 1, size(LC, 1)
        Dcore(j, k) = Dcore(j, k) + TWO*sum(LCcoreT(:, j)*LCcoreT(:, k))
      end do
    end do
    call construct_Fock_Matrix_direct(B%env, B%atm, B%bas, B%first_bf_of_shell, Dcore, B%H0, F_inactive_AO, accuracy=accuracy)
    !call construct_Fock_Matrix2(oeints, teints, Dcore, F_inactive_AO)
    F_inactive = LC.tm.F_inactive_AO.m.LC
  end subroutine calculate_F_inactive_direct

! for given one particle density matrix and twoparticle density matrix
! calculates the Lagrangian using the oeints and teints
  subroutine calculate_Lagrangian(opdm, oeints, s_tpdm, teints, Lag, virtual)
    integer :: i, j, k, l, m
    integer(8) :: im, kl, imkl, jm, jmkl
    ! see http://vergil.chemistry.gatech.edu/notes/cigrad/cigrad.pdf
    ! http://vergil.chemistry.gatech.edu/notes/mcscf.pdf
    real(long), intent(in) :: opdm(:, :)
      !  real(long), intent(in) :: tpdm(:,:,:,:)
    real(long), intent(in) :: s_tpdm(:)

    real(long), intent(in) :: oeints(:, :)
    real(long), intent(in) :: teints(:)
    real(long), intent(out) :: Lag(:, :)
    integer, optional, intent(in) :: virtual(:)

    real(long) :: OEsum
    real(long) :: TEsum

    real(long), allocatable :: OE_Lag(:, :), TE_Lag(:, :)
    allocate (OE_Lag(size(oeints, 1), size(oeints, 1)))
    allocate (TE_Lag(size(oeints, 1), size(oeints, 1)))

    do i = 1, size(oeints, 1)
      do j = 1, size(oeints, 1)
        OEsum = ZERO
        TEsum = ZERO
        OEsum = sum(opdm(:, j)*oeints(:, i))
        do m = 1, size(oeints, 1)
          if (present(virtual)) then
            if (ANY(m .eq. virtual)) cycle
          end if
          im = index_packed(i, m)
          jm = index_packed(j, m)

          !        if(j,m) is in core
          !        TEsum+=sum_kl (delta_jm s_tpdm(jj kl)) * teints(imkl)
          do k = 1, size(oeints, 1)
            if (present(virtual)) then
              if (ANY(k .eq. virtual)) cycle
            end if
            do l = 1, size(oeints, 1)
              if (present(virtual)) then
                if (ANY(l .eq. virtual)) cycle
              end if

              kl = index_packed(k, l)
              imkl = index_packed(im, kl)
              jmkl = index_packed(jm, kl)
              !            TEsum=TEsum + HALF*(tpdm(m,k,j,l)+tpdm(j,k,m,l)) * teints(imkl)
              !            TEsum=TEsum + HALF*HALF*(tpdm(m,j,k,l)+tpdm(j,m,k,l)+tpdm(m,j,l,k)+tpdm(j,m,l,k)) * teints(imkl)

              TEsum = TEsum + (s_tpdm(jmkl))*teints(imkl)
            end do
          end do
        end do
        OE_Lag(i, j) = OEsum
        TE_Lag(i, j) = TEsum
        Lag(i, j) = (OEsum + TEsum)
      end do
    end do

    ! write(OUTPUT,'("# OE_Lagrangian")')
    ! do i=1,size(OE_Lag,1)
    !   write(OUTPUT,'("#",99f15.10)') OE_Lag(i,:)
    ! end do
    ! write(OUTPUT,'("# TE_Lagrangian")')
    ! do i=1,size(OE_Lag,1)
    !   write(OUTPUT,'("#",99f15.10)') TE_Lag(i,:)
    ! end do

    deallocate (OE_Lag)
    deallocate (TE_Lag)
  end subroutine calculate_Lagrangian

  function calculate_MCSCF_energy(opdm, oeints, s_tpdmaa, teints_abij, teints_aibj, F_inactive, core, active) result(energy)
    integer :: i, j, k, l, ni
    integer(8) :: kl, ij, ijkl, ii, iikl, ikil, nij, ijaa, klaa
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: oeints(:, :)
    !  real(long), intent(in) :: tpdm(:,:,:,:)
    real(long), intent(in) :: s_tpdmaa(:)
    !real(long), intent(in) :: teints(:)
    real(long), intent(in) :: teints_abij(:)
    real(long), intent(in) :: teints_aibj(:)
    !real(long), intent(in) :: teints_ccij(:)
    real(long), intent(in) :: F_inactive(:, :)
    integer, intent(in) :: core(:)
    integer, intent(in) :: active(:)
    real(long) :: energy, oesum, tesum
    energy = ZERO
    oesum = ZERO
    tesum = ZERO
    do i = 1, size(oeints, 1)
      do j = 1, size(oeints, 1)
        oesum = oesum + opdm(i, j)*oeints(i, j)
      end do
    end do

    ni = size(oeints, 1)
    nij = ni*(ni + 1)/2
    do i = 1, size(core, 1)
      tesum = tesum + (F_inactive(core(i), core(i)) - oeints(core(i), core(i)))
    end do

    ! core - active part
    do i = 1, size(core, 1)
      ii = index_packed(core(i), core(i))

      do k = 1, size(active, 1)
        do l = 1, size(active, 1)
          kl = index_packed(k, l)
          iikl = (kl - 1)*nij + ii
          ikil = index_packed((k - 1)*ni + core(i), (l - 1)*ni + core(i))
          tesum = tesum + teints_abij(iikl)*TWO*opdm(active(k), active(l)) ! (iikl)
          tesum = tesum + teints_aibj(ikil)*(-ONE)*opdm(active(k), active(l)) ! (ikil)+(kili)+(kiil)+(ikli)
        end do
      end do
    end do

    ! active-active part
    do i = 1, size(active)
      do j = 1, size(active)
        ij = index_packed(active(i), active(j))
        ijaa = index_packed(i, j)
        do k = 1, size(active)
          do l = 1, size(active)
            kl = index_packed(active(k), active(l))
            klaa = index_packed(k, l)
            ijkl = index_packed(ijaa, klaa)
            tesum = tesum + HALF*s_tpdmaa(ijkl)*teints_abij((ijaa - 1)*nij + kl)
          end do! l
        end do! k
      end do!j
    end do! i

    energy = oesum + tesum
  end function calculate_MCSCF_energy

  subroutine one_particle_dm_CI2(UT, occs, csf_c, dm)
    !! computes the one-particle density matrix
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: csf_c(:)
    real(long), intent(out) :: dm(size(occs, 1), size(occs, 1))
    integer :: Ncsfi, Ncsfj
    integer :: i, j, k, l, ii, jj, iold
    integer :: a, b
    integer :: multp

    real(long), allocatable :: dm_block(:, :, :, :)
    real(long), allocatable :: dm_block_T(:, :, :, :)
    real(long) :: dm_part(size(occs, 1), size(occs, 1))
    !integer :: OUTPUT= 6

    multp = UT%multp
    dm = ZERO
    ii = 0
!$omp parallel private(i,Ncsfi,Ncsfj,dm_block,dm_block_T,dm_part,ii,jj,iold) shared(UT,occs,csf_c,multp) reduction(+:dm) default(none)
    iold = 0
    ii = 0
!$omp do schedule(dynamic) ! monotonic
    do i = 1, size(occs, 2)
      if (iold .ne. i - 1) then
        ! in case we start with a value i larger than 0 (and haven't done before iteration i-1), first calculate the right ii index
        ii = 0
        do j = 1, i - 1
          ii = ii + num_csf(count(occs(:, j) .eq. 1), multp)
        end do
      end if
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, i
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (maxval(abs(csf_c(ii + 1:ii + Ncsfi)))*maxval(abs(csf_c(jj + 1:jj + Ncsfj))) .eq. ZERO) cycle

        if (sum(abs(occs(:, i) - occs(:, j))) .le. 2) then
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (dm_block_T(Ncsfj, Ncsfi, size(occs, 1), size(occs, 1)))
            dm_block_T(:, :, :, :) = ZERO
            call opdm_matrix_subblock(occs(:, j), occs(:, i), multp, UT, dm_block_T)
            dm_part = ZERO
            do a = 1, size(dm, 1)
              do b = 1, size(dm, 1)
                do l = 1, Ncsfi
                  dm_part(a, b) = dm_part(a, b) + csf_c(ii + l)*sum(csf_c(jj + 1:jj + Ncsfj)*dm_block_T(:, l, a, b))
                end do !l
              end do!b
            end do!a
            deallocate (dm_block_T)
          else
            allocate (dm_block(Ncsfi, Ncsfj, size(occs, 1), size(occs, 1)))
            dm_block(:, :, :, :) = ZERO
            call opdm_matrix_subblock(occs(:, i), occs(:, j), multp, UT, dm_block)
            dm_part = ZERO
            do a = 1, size(dm, 1)
              do b = 1, size(dm, 1)
                do k = 1, Ncsfj
                  dm_part(a, b) = dm_part(a, b) + csf_c(jj + k)*sum(csf_c(ii + 1:ii + Ncsfi)*dm_block(:, k, a, b))
                end do !k
              end do!b
            end do!a
            deallocate (dm_block)
          end if
          if (i .ne. j) then
            dm = dm + dm_part + transpose(dm_part)
          else
            dm = dm + dm_part
          end if! i.ne.j
        end if! occs diff < 2
        jj = jj + Ncsfj
      end do ! j
      ii = ii + Ncsfi
      iold = i
    end do! i
!$omp end do
!$omp end parallel
  end subroutine one_particle_dm_CI2

  subroutine one_particle_dm_CI2_SA(UT, occs, csf_c, weights, dm)
    !! computes the one-particle density matrix for state-averaging
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: csf_c(:, :)
    real(long), intent(in) :: weights(:)
    real(long), intent(out) :: dm(size(occs, 1), size(occs, 1))
    integer :: Ncsfi, Ncsfj
    integer :: i, j, k, l, ii, jj
    integer :: multp
    real(long) :: w
    real(long), allocatable :: dm_block(:, :, :, :)
    real(long), allocatable :: dm_block_T(:, :, :, :)
    real(long) :: dm_part(size(occs, 1), size(occs, 1))
    !integer :: OUTPUT= 6

    multp = UT%multp
    dm = ZERO
    ii = 0
    do i = 1, size(occs, 2)
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, i
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 2) then
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (dm_block_T(Ncsfj, Ncsfi, size(occs, 1), size(occs, 1)))
            dm_block_T(:, :, :, :) = ZERO
            call opdm_matrix_subblock(occs(:, j), occs(:, i), multp, UT, dm_block_T)
            dm_part = ZERO
            do l = 1, Ncsfi
              do k = 1, Ncsfj
                w = sum(csf_c(ii + l, :)*csf_c(jj + k, :)*weights(:))
                dm_part(:, :) = dm_part(:, :) + w*dm_block_T(k, l, :, :)
              end do !l
            end do !k
            deallocate (dm_block_T)
          else
            allocate (dm_block(Ncsfi, Ncsfj, size(occs, 1), size(occs, 1)))
            dm_block(:, :, :, :) = ZERO
            call opdm_matrix_subblock(occs(:, i), occs(:, j), multp, UT, dm_block)
            dm_part = ZERO
            do l = 1, Ncsfi
              do k = 1, Ncsfj
                w = sum(csf_c(ii + l, :)*csf_c(jj + k, :)*weights(:))
                dm_part(:, :) = dm_part(:, :) + w*dm_block(l, k, :, :)
              end do !k
            end do !l
            deallocate (dm_block)
          end if
          if (i .ne. j) then
            dm = dm + dm_part + transpose(dm_part)
          else
            dm = dm + dm_part
          end if! i.ne.j
        end if! occs diff < 2
        jj = jj + Ncsfj
      end do ! j
      ii = ii + Ncsfi
    end do! i
  end subroutine one_particle_dm_CI2_SA

  subroutine two_particle_dm_CI2(UT, occs, csf_c, s_tpdm)
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: csf_c(:)
    real(long), intent(out) :: s_tpdm(:)
    integer :: multp
    integer :: i, j, k, l, ii, jj
    !integer :: OUTPUT=6
    real(long), allocatable :: s_tpdm_block(:, :, :)
    integer :: n_two, n_four, n_one
    integer :: Ncsfi, Ncsfj
    n_one = size(occs, 1)
    n_two = (n_one + 1)*n_one/2
    n_four = (n_two + 1)*n_two/2
    s_tpdm = ZERO
    multp = UT%multp
    ii = 0
    do i = 1, size(occs, 2)
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, i
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (s_tpdm_block(Ncsfj, Ncsfi, n_four))
            s_tpdm_block=ZERO
            call tpdm_matrix_subblock(occs(:, j), occs(:, i), UT, s_tpdm_block)
            if (i .ne. j) then
              do k=1,Ncsfj
                do l=1,Ncsfi
                  s_tpdm(:) = s_tpdm(:) + TWO*csf_c(jj + k)*csf_c(ii + l)*s_tpdm_block(k, l, :)
                end do !k
              end do!l
            else
              do k=1,Ncsfj
                do l=1,Ncsfi
                  s_tpdm(:) = s_tpdm(:) + csf_c(jj + k)*csf_c(ii + l)*s_tpdm_block(k, l, :)
                end do !k
              end do!l
            end if! i.ne.j
            deallocate (s_tpdm_block)
          else
            allocate (s_tpdm_block(Ncsfi, Ncsfj, n_four))
            s_tpdm_block = ZERO
            call tpdm_matrix_subblock(occs(:, i), occs(:, j), UT, s_tpdm_block)
            
            ! do a reshape with s_tpdm ???
            ! -> reshaping does not speed up the computations significantly
            !allocate(s_tpdm_block_reshaped(N_four, Ncsfi, Ncsfj))
            !s_tpdm_block_reshaped = reshape(s_tpdm_block, shape=shape(s_tpdm_block_reshaped), order=(/2, 3, 1/) )
            if (i .ne. j) then
              do k=1,Ncsfj
                do l=1,Ncsfi
                  s_tpdm(:) = s_tpdm(:) + TWO*csf_c(jj + k)*csf_c(ii + l)*s_tpdm_block(l, k, :)
                end do !l
              end do!k
            else
              do k=1,Ncsfj
                do l=1,Ncsfi
                  s_tpdm(:) = s_tpdm(:) + csf_c(jj + k)*csf_c(ii + l)*s_tpdm_block(l, k, :)
                end do! l
              end do! k
            end if! i.ne.j
            deallocate (s_tpdm_block)
          end if
        end if! sum(abs(occs(:,i)-occs(:,j))) < 4
        jj = jj + Ncsfj
      end do
      ii = ii + Ncsfi
    end do
  end subroutine two_particle_dm_CI2

  subroutine two_particle_dm_CI2_SA(UT, occs, csf_c, weights, s_tpdm)
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: occs(:, :)
    real(long), intent(in) :: csf_c(:, :)
    real(long), intent(in) :: weights(:)
    real(long), intent(out) :: s_tpdm(:)
    real(long) :: w
    integer :: multp
    integer :: i, j, k, l, ii, jj
    !integer :: OUTPUT=6
    real(long), allocatable :: s_tpdm_block(:, :, :)
    integer :: n_two, n_four, n_one
    integer :: Ncsfi, Ncsfj
    n_one = size(occs, 1)
    n_two = (n_one + 1)*n_one/2
    n_four = (n_two + 1)*n_two/2
    s_tpdm = ZERO
    multp = UT%multp
    ii = 0
    do i = 1, size(occs, 2)
      Ncsfi = num_csf(count(occs(:, i) .eq. 1), multp)
      if (Ncsfi .eq. 0) cycle
      jj = 0
      do j = 1, i
        Ncsfj = num_csf(count(occs(:, j) .eq. 1), multp)
        if (Ncsfj .eq. 0) cycle
        if (sum(abs(occs(:, i) - occs(:, j))) .le. 4) then
          if (count(occs(:, i) .eq. 1) < count(occs(:, j) .eq. 1)) then
            allocate (s_tpdm_block(Ncsfj, Ncsfi, n_four))
            call tpdm_matrix_subblock(occs(:, j), occs(:, i), UT, s_tpdm_block)
            if (i .ne. j) then
              do k=1,Ncsfj
                do l=1,Ncsfi
                  w = TWO*sum(csf_c(jj + k, :)*csf_c(ii + l, :)*weights(:))
                  s_tpdm(:) = s_tpdm(:) + w*s_tpdm_block(k, l, :)
                end do !k
              end do!l
            else
              do k=1,Ncsfj
                do l=1,Ncsfi
                  w = sum(csf_c(jj + k, :)*csf_c(ii + l, :)*weights(:))
                  s_tpdm(:) = s_tpdm(:) + w*s_tpdm_block(k, l, :)
                end do !k
              end do!l
            end if! i.ne.j
            deallocate (s_tpdm_block)
          else
            allocate (s_tpdm_block(Ncsfi, Ncsfj, n_four))
            s_tpdm_block = ZERO
            call tpdm_matrix_subblock(occs(:, i), occs(:, j), UT, s_tpdm_block)
            ! do a reshape with s_tpdm ???
            ! -> reshaping does not speed up the computations significantly
            !allocate(s_tpdm_block_reshaped(N_four, Ncsfi, Ncsfj))
            !s_tpdm_block_reshaped = reshape(s_tpdm_block, shape=shape(s_tpdm_block_reshaped), order=(/2, 3, 1/) )
            if (i .ne. j) then
              do k=1,Ncsfj
                do l=1,Ncsfi
                  w = TWO*sum(csf_c(jj + k, :)*csf_c(ii + l, :)*weights(:))
                  s_tpdm(:) = s_tpdm(:) + w*s_tpdm_block(l, k, :)
                end do !l
              end do!k
            else
              do k=1,Ncsfj
                do l=1,Ncsfi
                  w = sum(csf_c(jj + k, :)*csf_c(ii + l, :)*weights(:))
                  s_tpdm(:) = s_tpdm(:) + w*s_tpdm_block(l, k, :)
                end do! l
              end do! k
            end if! i.ne.j
            deallocate (s_tpdm_block)
          end if
        end if! sum(abs(occs(:,i)-occs(:,j))) < 4
        jj = jj + Ncsfj
      end do
      ii = ii + Ncsfi
    end do
  end subroutine two_particle_dm_CI2_SA

  subroutine s_two_particle_dm_CI(s_tpdmaa, csf_c, csf_array, active)
    real(long), intent(out) :: s_tpdmaa(:)
    real(long), intent(in) :: csf_c(:)
    type(CSF), intent(in) :: csf_array(:)
    integer, optional, intent(in) :: active(:)
    integer :: csf_a, csf_b
    real(long) :: coef1, coef2
    integer :: a, b
    integer :: t, p, q, r, s
    integer :: smax
    integer :: nmo
    real(long), allocatable :: tpdm(:, :, :, :)
    if (present(active)) then
      nmo = size(csf_array(1)%occ(active))
    else
      nmo = size(csf_array(1)%occ(:))
    end if
    allocate (tpdm(nmo, nmo, nmo, nmo))
    tpdm = ZERO
    do csf_a = 1, size(csf_array, 1)
      do a = 1, size(csf_array(csf_a)%dets, 2)
        coef1 = csf_c(csf_a)*csfdet(csf_array(csf_a)%csf, csf_array(csf_a)%dets(:, a))
        do csf_b = 1, size(csf_array, 1)
          do b = 1, size(csf_array(csf_b)%dets, 2)
            coef2 = csf_c(csf_b)*csfdet(csf_array(csf_b)%csf, csf_array(csf_b)%dets(:, b))
            if (coef1 .eq. ZERO .or. coef2 .eq. ZERO) cycle
            if (present(active)) then
              call two_particle_dm_two_det(tpdm, &
                csf_array(csf_a)%occ(active), csf_array(csf_b)%occ(active), &
                csf_array(csf_a)%dets(:, a), csf_array(csf_b)%dets(:, b), &
                coef1*coef2)
            else
              call two_particle_dm_two_det(tpdm, &
                csf_array(csf_a)%occ(:), csf_array(csf_b)%occ(:), &
                csf_array(csf_a)%dets(:, a), csf_array(csf_b)%dets(:, b), &
                coef1*coef2)
            end if
          end do
        end do
      end do
    end do

    s_tpdmaa(:) = ZERO
    t = 1
    do p = 1, size(tpdm, 1)
      do q = 1, p
        do r = 1, p
          if (r .eq. p) then
            smax = q
          else
            smax = r
          end if
          do s = 1, smax
            s_tpdmaa(t) = FOURTH*(tpdm(p, q, r, s) + tpdm(q, p, r, s) + tpdm(p, q, s, r) + tpdm(q, p, s, r))
            t = t + 1
          end do
        end do
      end do
    end do
    deallocate (tpdm)
  end subroutine s_two_particle_dm_CI

  subroutine s_two_particle_dm_CI_SA(s_tpdmaa, csf_c, weights, csf_array, active)
    real(long), intent(out) :: s_tpdmaa(:)
    real(long), intent(in) :: csf_c(:, :)
    real(long), intent(in) :: weights(:)
    type(CSF), intent(in) :: csf_array(:)
    integer, optional, intent(in) :: active(:)
    integer :: csf_a, csf_b
    real(long) :: coef12, coef1, coef2
    integer :: a, b
    integer :: t, p, q, r, s
    integer :: smax
    integer :: nmo
    real(long), allocatable :: tpdm(:, :, :, :)
    if (present(active)) then
      nmo = size(csf_array(1)%occ(active))
    else
      nmo = size(csf_array(1)%occ(:))
    end if

    allocate (tpdm(nmo, nmo, nmo, nmo))
    tpdm = ZERO
    do csf_a = 1, size(csf_array, 1)
      do a = 1, size(csf_array(csf_a)%dets, 2)
        coef1 = csfdet(csf_array(csf_a)%csf, csf_array(csf_a)%dets(:, a))
        do csf_b = 1, size(csf_array, 1)
          do b = 1, size(csf_array(csf_b)%dets, 2)
            coef2 = csfdet(csf_array(csf_b)%csf, csf_array(csf_b)%dets(:, b))
            coef12 = sum(weights(:)*csf_c(csf_b, :)*csf_c(csf_a, :))*coef1*coef2
            if (coef12 .eq. ZERO) cycle
            if (present(active)) then
              call two_particle_dm_two_det(tpdm, &
                csf_array(csf_a)%occ(active), csf_array(csf_b)%occ(active), &
                csf_array(csf_a)%dets(:, a), csf_array(csf_b)%dets(:, b), &
                coef12)
            else
              call two_particle_dm_two_det(tpdm, &
                csf_array(csf_a)%occ(:), csf_array(csf_b)%occ(:), &
                csf_array(csf_a)%dets(:, a), csf_array(csf_b)%dets(:, b), &
                coef12)
            end if

          end do
        end do
      end do
    end do

    s_tpdmaa(:) = ZERO
    t = 1
    do p = 1, size(tpdm, 1)
      do q = 1, p
        do r = 1, p
          if (r .eq. p) then
            smax = q
          else
            smax = r
          end if
          do s = 1, smax
            s_tpdmaa(t) = FOURTH*(tpdm(p, q, r, s) + tpdm(q, p, r, s) + tpdm(p, q, s, r) + tpdm(q, p, s, r))
            !          write(6,*) s_tpdm(t),p,q,r,s,t
            t = t + 1
          end do
        end do
      end do
    end do
    deallocate (tpdm)
  end subroutine s_two_particle_dm_CI_SA

  subroutine two_particle_dm_det(tpdm, occ, det, coef_sq)
    real(long), intent(inout) :: tpdm(:, :, :, :)
    integer, intent(in) :: occ(:)
    integer, intent(in) :: det(:)
    real(long), intent(in) :: coef_sq
    integer :: oi
    integer :: mo1, mo2
    integer :: spin_of_mo(size(occ, 1))
    oi = 0
    do mo1 = 1, size(occ(:), 1)
      if (occ(mo1) .eq. 1) then
        oi = oi + 1
        spin_of_mo(mo1) = open_spin(det(:), oi)
      else
        spin_of_mo(mo1) = 0
      end if
    end do
    do mo1 = 1, size(occ(:), 1)
      if (occ(mo1) .eq. 2) then
        !      write(OUTPUT,'(3x,4i1,f)') mo1,mo1,mo1,mo1,TWO*coef_sq
        tpdm(mo1, mo1, mo1, mo1) = tpdm(mo1, mo1, mo1, mo1) + TWO*coef_sq
      end if
      if (occ(mo1) .eq. 1) then
        tpdm(mo1, mo1, mo1, mo1) = tpdm(mo1, mo1, mo1, mo1) + ZERO*coef_sq
      end if
      do mo2 = mo1 + 1, size(occ(:), 1)
        if (occ(mo1) .eq. 2 .and. occ(mo2) .eq. 2) then
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,mo2,mo2,FOUR*coef_sq
          tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + FOUR*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo2,mo1,mo1,FOUR*coef_sq
          tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + FOUR*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo2,mo2,mo1,-TWO*coef_sq
          tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - TWO*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo1,mo1,mo2,-TWO*coef_sq
          tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - TWO*coef_sq
        else if (occ(mo1) .eq. 1 .and. occ(mo2) .eq. 2) then
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,mo2,mo2,TWO*coef_sq
          tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + TWO*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo2,mo1,mo1,TWO*coef_sq
          tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + TWO*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo2,mo2,mo1,-ONE*coef_sq
          tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo1,mo1,mo2,-ONE*coef_sq
          tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef_sq
        else if (occ(mo1) .eq. 2 .and. occ(mo2) .eq. 1) then
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,mo2,mo2,TWO*coef_sq
          tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + TWO*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo2,mo1,mo1,TWO*coef_sq
          tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + TWO*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo2,mo2,mo1,-ONE*coef_sq
          tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo1,mo1,mo2,-ONE*coef_sq
          tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef_sq
        else if (occ(mo1) .eq. 1 .and. occ(mo2) .eq. 1) then
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,mo2,mo2,ONE*coef_sq
          tpdm(mo1, mo1, mo2, mo2) = tpdm(mo1, mo1, mo2, mo2) + ONE*coef_sq
          !        write(OUTPUT,'(3x,4i1,f)') mo2,mo2,mo1,mo1,ONE*coef_sq
          tpdm(mo2, mo2, mo1, mo1) = tpdm(mo2, mo2, mo1, mo1) + ONE*coef_sq
          if (spin_of_mo(mo1) .eq. spin_of_mo(mo2)) then
            !          write(OUTPUT,'(3x,4i1,f)') mo1,mo2,mo2,mo1,-ONE*coef_sq
            tpdm(mo1, mo2, mo2, mo1) = tpdm(mo1, mo2, mo2, mo1) - ONE*coef_sq
            !          write(OUTPUT,'(3x,4i1,f)') mo2,mo1,mo1,mo2,-ONE*coef_sq
            tpdm(mo2, mo1, mo1, mo2) = tpdm(mo2, mo1, mo1, mo2) - ONE*coef_sq
          end if
        end if
      end do
    end do
  contains
    function open_spin(detv, i)
      integer :: detv(:)
      integer :: i
      integer :: open_spin
      if (i .eq. 1) then
        open_spin = detv(i)
      else
        open_spin = detv(i) - detv(i - 1)
      end if
    end function open_spin
  end subroutine two_particle_dm_det

! varient of the above employing E_core and F_core
  function CI_matrix_elementaa(oeints, teints, occ1, occ2, det1, det2, F_core, E_core) result(E)
    real(long), intent(in) :: oeints(:, :)
    real(long), intent(in) :: teints(:)
    integer, intent(in) :: occ1(:)
    integer, intent(in) :: occ2(:)
    integer, intent(in) :: det1(:)
    integer, intent(in) :: det2(:)
    real(long), intent(in) :: F_core(:, :)
    real(long), intent(in) :: E_core
    real(long) :: E

    integer :: only_left(2), only_left_spin(2)
    integer :: only_right(2), only_right_spin(2)
    integer :: coef_sign
    integer :: spin_of_mo(size(occ1, 1))
    integer :: oi
    integer :: i, j, k, l
    integer(8) :: iijj, ijij, ii, jj, ij, kk, il, jl, kl, ik, jk, ijkk, ikjk, ijkl, ikjl, iljk
    integer :: OUTPUT = 6
    logical :: verbose = .False.
      !  write(OUTPUT,"('<',a,'|H|',a,'>')") trim(detstring(occ1(:),det1(:))),trim(detstring(occ2(:),det2(:)))
    E = ZERO
    oi = 0
    do k = 1, size(occ1(:), 1)
      if (occ1(k) .eq. 1) then
        oi = oi + 1
        spin_of_mo(k) = open_spin(det1(:), oi)
      else
        spin_of_mo(k) = 0
      end if
    end do
    call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, occ1, det1, occ2, det2)
    !  write(OUTPUT,*) 'coef_sign=',coef_sign
    if (coef_sign .eq. 0) then
      !    write(OUTPUT,"(3x,'=0')")
      ! if there are more than two differences return
    else if (count(only_left(:) .ne. 0) .eq. 0 .and. count(only_right(:) .ne. 0) .eq. 0) then
      E = E_core
      do i = 1, size(oeints, 1)
        ii = index_packed(i, i)
        if (verbose) write (OUTPUT, "(3x,'+',i2,a,2i2, f10.4)") occ1(i), ' fcore', i, i, occ1(i)*F_core(i, i)
        E = E + occ1(i)*F_core(i, i)
        do j = 1, size(oeints, 1)
          jj = index_packed(j, j)
          ij = index_packed(i, j)
          iijj = index_packed(ii, jj)
          ijij = index_packed(ij, ij)
          if (occ1(i) .eq. 2 .and. occ1(j) .eq. 2) then
            !if(verbose) write(OUTPUT,"(3x,'+2.0*',4i1,'-1.0*',4i1, x,f10.4)") i,i,j,j, i,j,j,i, coef_sign*TWO*teints(iijj)-coef_sign*ONE*teints(ijij)
            E = E + coef_sign*TWO*teints(iijj) - coef_sign*ONE*teints(ijij)
          else if (occ1(i) .eq. 2 .and. occ1(j) .eq. 1) then
            !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1,'-0.5*',4i1)") i,i,j,j, i,j,j,i
            E = E + coef_sign*ONE*teints(iijj) - coef_sign*HALF*teints(ijij)
          else if (occ1(i) .eq. 1 .and. occ1(j) .eq. 2) then
            !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1,'-0.5*',4i1)") i,i,j,j, i,j,j,i
            E = E + coef_sign*ONE*teints(iijj) - coef_sign*HALF*teints(ijij)
          else if (occ1(i) .eq. 1 .and. occ1(j) .eq. 1) then
            !if(verbose) write(OUTPUT,"(3x,'+0.5*',4i1,'-0.0*',4i1)") i,i,j,j, i,j,j,i
            E = E + coef_sign*HALF*teints(iijj)
            if (spin_of_mo(i) .eq. spin_of_mo(j)) then
              E = E - coef_sign*HALF*teints(ijij)
              !if(verbose) write(OUTPUT,"(3x,'+0.0*',4i1,'-0.5*',4i1)") i,i,j,j, i,j,j,i
            end if
          end if
        end do
      end do
    else if (count(only_left(:) .ne. 0) .eq. 1 .and. &
      count(only_right(:) .ne. 0) .eq. 1 .and. only_left_spin(1) .eq. only_right_spin(1)) then
      ! left determinant has one and right determinant has one electron unique, same spin
      !    write(OUTPUT,*) 'spin of excitation:',only_left_spin(1),only_right_spin(1)
      i = only_left(1)
      j = only_right(1)
      !    write(OUTPUT,"(3x,'only left:',i2,i2,'only right:',i2,i2 )") i,only_left_spin(1),j,only_right_spin(1)
      ij = index_packed(i, j)

      if (verbose) write (OUTPUT, "(3x,'+',i2,a,2i2, f10.4)") coef_sign, ' fcore', i, i, coef_sign*F_core(i, j)
      E = coef_sign*F_core(i, j)

      do k = 1, size(occ1, 1)
        kk = index_packed(k, k)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ijkk = index_packed(ij, kk)
        ikjk = index_packed(ik, jk)
        if (occ1(k) .eq. 2) then
          !if(verbose) write(OUTPUT,"(3x,'+2.0*',4i1,'-1.0*',4i1)") i,j,k,k, i,k,k,j
          E = E + coef_sign*TWO*teints(ijkk) - coef_sign*ONE*teints(ikjk)
        else if (occ1(k) .eq. 1) then
          !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1,'-0.0*',4i1)") i,j,k,k, i,k,k,j
          E = E + coef_sign*ONE*teints(ijkk)
          if (spin_of_mo(k) .eq. only_left_spin(1)) then
            if (verbose) write (OUTPUT, "(3x,'+0.0*',4i1,'-1.0*',4i1)") i, j, k, k, i, k, k, j
            E = E - coef_sign*ONE*teints(ikjk)
          end if
        end if
      end do
    else if (count(only_left(:) .ne. 0) .eq. 2 .and. count(only_right(:) .ne. 0) .eq. 2 &
      .and. only_left_spin(1) .eq. only_right_spin(1) .and. only_left_spin(2) .eq. only_right_spin(2)) then
      ! determinant differ by two electrons: <ab|ab>
      !write(OUTPUT,"(3x,'only left:',2i2,2i2,'only right:',2i2,2i2 )") only_left,only_left_spin,only_right,only_right_spin
      i = only_left(1)
      j = only_right(1)
      k = only_left(2)
      l = only_right(2)
      ij = index_packed(i, j)
      kl = index_packed(k, l)
      ik = index_packed(i, k)
      il = index_packed(i, l)
      jk = index_packed(j, k)
      jl = index_packed(j, l)
      ijkl = index_packed(ij, kl)
      ikjl = index_packed(ik, jl)
      iljk = index_packed(il, jk)
      !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1)") i,j,k,l
      E = E + coef_sign*teints(ijkl)
      if (only_left_spin(1) .eq. only_left_spin(2)) then
        !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1)") i,l,j,k
        E = E - coef_sign*teints(iljk)
      end if
    else if (count(only_left_spin(:) .ne. 0) .eq. 2 .and. count(only_right_spin(:) .ne. 0) .eq. 2 &
      .and. only_left_spin(2) .eq. only_right_spin(1) .and. only_left_spin(1) .eq. only_right_spin(2)) then
      !write(OUTPUT,"(3x,'only left:',2i2,2i2,'only right:',2i2,2i2 )") only_left,only_left_spin,only_right,only_right_spin
      ! determinant differ by two electrons: <ab|ba>
      i = only_left(1)
      j = only_right(1)
      k = only_left(2)
      l = only_right(2)
      ij = index_packed(i, j)
      kl = index_packed(k, l)
      ik = index_packed(i, k)
      il = index_packed(i, l)
      jk = index_packed(j, k)
      jl = index_packed(j, l)
      ijkl = index_packed(ij, kl)
      ikjl = index_packed(ik, jl)
      iljk = index_packed(il, jk)
      !if(verbose) write(OUTPUT,"(3x,'-1.0*',4i1)") i,l,j,k
      E = E - coef_sign*teints(iljk)
      if (only_left_spin(1) .eq. only_left_spin(2)) then
        !if(verbose) write(OUTPUT,"(3x,'+1.0*',4i1)") i,j,k,l
        E = E + coef_sign*teints(ijkl)
      end if
    else
    !    write(OUTPUT,"(3x,'only left:',2i2,2i2,'only right:',2i2,2i2 )") only_left,only_left_spin,only_right,only_right_spin
    end if
  contains
    function open_spin(det, ind)
      integer :: det(:)
      integer :: ind
      integer :: open_spin
      if (ind .eq. 1) then
        open_spin = det(ind)
      else
        open_spin = det(ind) - det(ind - 1)
      end if
    end function open_spin
  end function CI_matrix_elementaa

  function CI_matrix_element_CSFaa(oeints_ab, teints_abcd, csf1, csf2, active, F_core, E_core) result(E)
    real(long), intent(in) :: oeints_ab(:, :)
    real(long), intent(in) :: teints_abcd(:)
    type(CSF), intent(in) :: csf1
    type(CSF), intent(in) :: csf2
    integer, intent(in) :: active(:)
    real(long), intent(in) :: F_core(:, :)
    real(long), intent(in) :: E_core
    real(long) :: E
    integer :: i, j
    real(long) :: coef1, coef2
    E = ZERO
    do i = 1, size(csf1%dets, 2)
      coef1 = csfdet(csf1%csf, csf1%dets(:, i))
      do j = 1, size(csf2%dets, 2)
        coef2 = csfdet(csf2%csf, csf2%dets(:, j))
        E = E + coef1*coef2*CI_matrix_elementaa( &
          oeints_ab, &
          teints_abcd, &
          csf1%occ(active), &
          csf2%occ(active), &
          csf1%dets(:, i), &
          csf2%dets(:, j), &
          F_core, &
          E_core)
      end do
    end do
  end function CI_matrix_element_CSFaa

!> calculates the hamiltonian matrix elements for all configurations between occupations occ1 and occ2
!! occ1 and occ2 must be swapped such that there are more or equal single occupated orbitals in occ1 than in occ2
  subroutine hamiltonian_matrix_subblock(occ1, occ2, multp, oeints, teints, E_core, UT, H, OUTPUT)
    integer, intent(in) :: occ1(:), occ2(:)
    integer, intent(in) :: multp
    real(long), intent(in) :: oeints(:, :)
    real(long), intent(in) :: teints(:)
    real(long), intent(in) :: E_core
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(out), allocatable :: H(:, :)
    integer, intent(in), optional :: OUTPUT
    logical :: yes_verbose
    integer :: IOUT
    real(long), allocatable :: A(:, :), U(:, :), U3(:, :), U2(:, :), U1(:, :)
    integer :: Ndiff, Ncsf1, Ncsf2, Nsingle, Nsingle2
    integer :: i, j, k, l
    integer(8) :: ii, jj, kk, kl, ij, ik, jk, il, jl, iijj, ijij, ijii, ijjj, ijkk, ikjk, iljk, ikjl, ijkl
    integer :: oi, oj, ok, ol
    integer :: uocc_left(4), socc_left(4), docc_left(4)
    integer :: uocc_right(4), socc_right(4), docc_right(4)
    integer :: count_uocc_left, count_socc_left, count_docc_left
    real(long) :: temp

    yes_verbose = .False.
    IOUT = 6
    if (present(OUTPUT)) then
      IOUT = OUTPUT
      yes_verbose = .True.
    end if
    Nsingle = count(occ1 .eq. 1)
    Nsingle2 = count(occ2 .eq. 1)
    if (Nsingle2 > Nsingle) then
      stop "occ2 has more singles than occ1. This is not allowed!"
    end if
    !Ncsf1 = num_csf(Nsingle, multp)
    !Ncsf2 = num_csf(Nsingle2, multp)
    Ncsf1 = UT%f(Nsingle)
    Ncsf2 = UT%f(Nsingle2)

    allocate (H(Ncsf1, Ncsf2))
    H(:, :) = ZERO
    Ndiff = sum(abs(occ1 - occ2))/2
    if (Ndiff > 2) then
      return
    end if

    uocc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    uocc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    count_uocc_left = count(uocc_left .ne. 0)
    count_socc_left = count(socc_left .ne. 0)
    count_docc_left = count(docc_left .ne. 0)

    if (yes_verbose) then
      write (IOUT, '("# ",a,i5,a,i5)') 'Ndiff=', Ndiff
      write (IOUT, '("# occ1",x,99i1)') occ1
      write (IOUT, '("# occ2",x,99i1)') occ2
      write (IOUT, '("# multp: ",i4)') multp
      write (IOUT, '("# Num_csf1: ",i4)') Ncsf1
      write (IOUT, '("# Num_csf2: ",i4)') Ncsf2
    end if
    allocate (A(Ncsf1, Ncsf2))
    allocate (U(Ncsf1, Ncsf1))
    allocate (U1(Ncsf1, Ncsf1))
    allocate (U2(Ncsf1, Ncsf1))
    allocate (U3(Ncsf1, Ncsf1))
    A = ZERO
    H = ZERO
    if (Ndiff .eq. 0) then
      if (yes_verbose) write (IOUT, "('# case diagonal')")
      temp = E_core
      do i = 1, size(occ1)
        ii = index_packed(i, i)
        if (occ1(i) > 0) then
          temp = temp + occ1(i)*oeints(i, i)
        end if
        if (occ1(i) .eq. 2) then
          iijj = index_packed(ii, ii)
          temp = temp + teints(iijj)
        end if
        ii = index_packed(i, i)
        do j = 1, i - 1
          jj = index_packed(j, j)
          ij = index_packed(i, j)
          iijj = index_packed(ii, jj)
          ijij = index_packed(ij, ij)
          if (occ1(i)*occ1(j) > 0) then
            temp = temp + occ1(i)*occ1(j)*teints(iijj)
          end if
          if (occ1(i)*occ1(j) .eq. 1) then
            oi = i - count(occ1(:i) .ne. 1)
            oj = j - count(occ1(:j) .ne. 1)
            call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
            H(:, :) = H(:, :) + U(1:Ncsf1, 1:Ncsf2)*teints(ijij)
          else if (occ1(i)*occ1(j) > 0) then
            temp = temp - HALF*occ1(i)*occ1(j)*teints(ijij)
          end if
        end do
      end do

      do i = 1, size(H, 1)
        H(i, i) = H(i, i) + temp
      end do
    else if (Ndiff .eq. 1) then
      if ((count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 0) .or. &
        (count_uocc_left .eq. 0 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1)) then
        if (yes_verbose) write (IOUT, "('# case A1')")
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        call makeCycleUij(UT, oi, oj, Nsingle, U1)
        if (count_uocc_left .eq. 0) then
          ! 2,1 -> 1,2
          A(:, :) = -U1(1:Ncsf1, 1:Ncsf1)
        else
          ! 0,1 -> 1,0
          A(:, :) = U1(1:Ncsf1, 1:Ncsf1)
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! 1,1 -> 2,0  A(2)
        if (yes_verbose) write (IOUT, "('# case A2')")
        i = socc_left(2)
        j = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = transpose(matmul(U1, U)) ! P1 = (2...oi) (1...oj)
        ! Why transpose ?
        A(:, :) = sqrt(2.)*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
      end if
      ii = index_packed(i, i)
      jj = index_packed(j, j)
      ij = index_packed(i, j)
      iijj = index_packed(ii, jj)
      ijij = index_packed(ij, ij)
      ijii = index_packed(ij, ii)
      ijjj = index_packed(ij, jj)
      temp = oeints(i, j)
      if (occ1(j) .eq. 2) then
        temp = temp + teints(ijjj)
      end if
      if (occ1(i) .eq. 2) then
        temp = temp + teints(ijii)
      end if
      if (occ2(j) .eq. 2) then
        temp = temp + teints(ijjj)
      end if
      if (occ2(i) .eq. 2) then
        temp = temp + teints(ijii)
      end if
      do k = 1, size(occ2)
        if (k .eq. i .or. k .eq. j) cycle
        if (occ2(k) .eq. 0) cycle
        kk = index_packed(k, k)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ijkk = index_packed(ij, kk)
        ikjk = index_packed(ik, jk)
        temp = temp + occ1(k)*teints(ijkk)
        if (occ1(k) .eq. 2) then
          temp = temp - teints(ikjk)
        end if
      end do
      H(:, :) = H(:, :) + A(:, :)*temp

      if (count_socc_left .eq. 1) then
        ! A1 case
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          ikjk = index_packed(ik, jk)
          call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          !call Umatrix(Nsingle, multp, oi, ok, U)

          U = matmul(U, U1) ! P0 = (i,k) P1
          if (count_docc_left .eq. 0) then
            U2 = U ! P2 = P0
          else
            U2 = U + U1 ! P2 = P1 + P0
          end if
          A(:, :) = U2(1:Ncsf1, 1:Ncsf2)
          temp = teints(ikjk)
          H(:, :) = H(:, :) + A(:, :)*temp
        end do! k
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! A(2) case
        ! oi and oj are singly occupied left
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          ikjk = index_packed(ik, jk)
          if (occ2(j) .eq. 0) then ! we need P2
            !call Umatrix(Nsingle, multp, oj, ok, U)
            call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
          else if (occ2(j) .eq. 2) then ! we need P2'
            !call Umatrix(Nsingle, multp, oi, ok, U)
            call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          end if
          ! P0 = (oj,k) * P1
          ! P0' = (oi,k) * P1
          U2 = matmul(U, U1) ! P2 = P0; P2' = P0'
          A(:, :) = sqrt(2.)*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
          temp = teints(ikjk)
          H(:, :) = H(:, :) + A(:, :)*temp
        end do! k
      end if! A(1) or A(2)
    else if (Ndiff .eq. 2) then
      if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 0 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case A3')")
        ! 2,0 -> 0,2
        ! identity matrix
        U(:, :) = 0
        do k = 1, size(U, 1)
          U(k, k) = 1
        end do
        i = uocc_left(1)
        j = docc_left(1)
        ij = index_packed(i, j)
        ijij = index_packed(ij, ij)
        temp = teints(ijij)
        A(:, :) = U(:, :)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case B1')")
        !2,1,0 -> 1,0,2
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        if (docc_left(1) .eq. uocc_right(1)) then
          k = docc_left(1)
        else
          k = docc_right(1)
        end if
        ! i is singly occupied left
        ! j is singly occupied right
        ! k is doubly occupied on one side and empty on the other
        call makeCycleUij(UT, oi, oj, Nsingle, U2)
        A(:, :) = -U2(1:Ncsf1, 1:Ncsf2)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ikjk = index_packed(ik, jk)
        temp = teints(ikjk)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case B2')")
        i = socc_left(2)
        j = socc_left(1)
        if (count_docc_left .eq. 1) then
          k = docc_left(1)
        else if (count(docc_right .ne. 0) .eq. 1) then
          k = docc_right(1)
        end if
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U2)
        U2 = transpose(matmul(U2, U)) ! P2 = (1...oj) (2...oi) ^T
        if (count_docc_left .eq. 1) then
          U2 = -U2
        end if
        A(:, :) = sqrt(2.)*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ikjk = index_packed(ik, jk)
        temp = teints(ikjk)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 2) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 2 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case C1-3')")
        ! 1,1,x,x -> x,x,1,1 (c1)
        ! 1,x,1,x -> x,1,x,1 (c2)
        ! x,1,1,x -> 1,x,x,1 (c3)
        i = socc_left(2)
        j = socc_left(1)
        l = socc_right(1)
        k = socc_right(2)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ2(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        jk = index_packed(j, k)
        jl = index_packed(j, l)
        kl = index_packed(k, l)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)
        if (ol .le. oj) then
          call makeCycleUij(UT, oi, ol, Nsingle, U)
          call makeCycleUij(UT, oj + 1, ok, Nsingle, U2)
          U2 = matmul(U, U2)  ! P2 = (i...l) (j+1...k)
          !call Umatrix(Nsingle, multp, oi, oj, U)
          call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
          U3 = matmul(U, U2) ! P3
        else
          call makeCycleUij(UT, oi, ol + 1, Nsingle, U)
          call makeCycleUij(UT, oj, ok, Nsingle, U2)
          U2 = matmul(U, U2) ! P2 = (i...l+1) (j...k)
          call makeCycleUij(UT, oj, ol, Nsingle, U3)
          call makeCycleUij(UT, oi, ok, Nsingle, U)
          U3 = matmul(U, U3) ! P3 = (i...k) (j...l)
        end if
        U1 = U2 + U3 ! P1 = P2 + P3
        if (count(uocc_left > 0) .eq. 1 .and. count(socc_left > 0) .eq. 2 .and. count(docc_left > 0) .eq. 1) then
          if (occ2(i) .eq. occ1(k)) then
            A = -U3
            H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
          else
            A = -U2
            H(:, :) = H(:, :) + A(:, :)*teints(iljk)
          end if
          A = U1
          H(:, :) = H(:, :) + A(:, :)*teints(ijkl)
        else
          A = U3
          H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
          A = U2
          H(:, :) = H(:, :) + A(:, :)*teints(iljk)
        end if
      else if (count(occ2 .eq. 2) - count(occ1 .eq. 2) .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! x,1,1,1 -> x,1,x,x (c4)
        ! 1,x,1,1 -> x,1,x,x (c5)
        ! 1,1,x,1 -> x,x,1,x (c6)
        ! 1,1,1,x -> x,x,1,x (c7)
        i = socc_left(3)
        j = socc_left(2)
        k = socc_left(1)
        l = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        kl = index_packed(k, l)
        jl = index_packed(j, l)
        jk = index_packed(j, k)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = matmul(U1, U) !   (2...oi) (1...oj)
        U1 = transpose(U1)
        if (l .ge. k) then
          call makeCycleUij(UT, ok + 2, ol + 2, Nsingle, U)
        else
          call makeCycleUij(UT, ol + 2, ok + 2, Nsingle, U)
          U = transpose(U)
        end if
        U1 = matmul(U1, U) !   (2...oi) (1...oj) (ok..ol)
        !call Umatrix(Nsingle, multp, oi, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
        U2 = matmul(U, U1) ! P2 = (i,k) P1
        !call Umatrix(Nsingle, multp, oj, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1) ! P3 = (j,k) P1
        if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 1) then
          if (occ2(k) .eq. 2) then
            A(:, :) = -sqrt(2.)*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P1
            H(:, :) = H(:, :) + A(:, :)*teints(ijkl)
          else
            A = -sqrt(2.)*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P2
            H(:, :) = H(:, :) + A(:, :)*teints(iljk)
          end if
          if (occ2(j) .eq. 2) then
            A = sqrt(2.)*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P3' = P1+P2
            H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
          else
            A = sqrt(2.)*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P2'= P1+P3
            H(:, :) = H(:, :) + A(:, :)*teints(iljk)
          end if
        else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 0) then
          if (occ2(k) .eq. 2) then
            A(:, :) = sqrt(2.)*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            H(:, :) = H(:, :) + A(:, :)*teints(iljk)
          else
            A = sqrt(2.)*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
            H(:, :) = H(:, :) + A(:, :)*teints(ijkl)
          end if
          if (occ2(j) .eq. 2) then
            A = sqrt(2.)*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            H(:, :) = H(:, :) + A(:, :)*teints(iljk)
          else
            A = sqrt(2.)*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
            H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
          end if
        else
          stop 'impossible'
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 4 .and. count_docc_left .eq. 0) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! 1,1,1,1 -> x,x,x,x (c8)
        i = socc_left(4)
        j = socc_left(3)
        k = socc_left(2)
        l = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ1(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        kl = index_packed(k, l)
        jl = index_packed(j, l)
        jk = index_packed(j, k)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)

        call makeCycleUij(UT, 1, ol, Nsingle, U1)
        call makeCycleUij(UT, 2, ok, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 3, oj, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 4, oi, Nsingle, U)
        U1 = matmul(U, U1)
        U1 = transpose(U1)
        ! P1 = (4..oi)(3..oj)(2..ok)(1..ol)^T
        !call Umatrix(Nsingle, multp, oj, ol, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ol, U)
        U2 = matmul(U, U1)  ! P2 = (j,l) P1
        !call Umatrix(Nsingle, multp, oj, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1)  ! P3 = (j,k) P1
        if (occ2(i) .eq. occ2(j)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
        else
          A = TWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
          H(:, :) = H(:, :) + A(:, :)*teints(ijkl)
        end if
        if (occ2(j) .eq. occ2(k)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          H(:, :) = H(:, :) + A(:, :)*teints(ikjl)
        else
          A = TWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
          H(:, :) = H(:, :) + A(:, :)*teints(iljk)
        end if
      else! we shouldn't reach this point
        stop "reached an impossible situation!"
      end if
    else  ! if Ndiff>2
      ! this case is covered already earlier. We should never reach this point
      H = ZERO
    end if
    ! deallocation
    deallocate (A)
    deallocate (U)
    deallocate (U1)
    deallocate (U2)
    deallocate (U3)
  end subroutine hamiltonian_matrix_subblock

!> calculates the one-particle density matrix elements for all configurations between occupations occ1 and occ2
!! occ1 and occ2 must be swapped such that there are more or equal single occupated orbitals in occ1 than in occ2
  subroutine opdm_matrix_subblock(occ1, occ2, multp, UT, dm, OUTPUT)
    integer, intent(in) :: occ1(:), occ2(:)
    integer, intent(in) :: multp
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(out), allocatable :: dm(:, :, :, :)
    integer, intent(in), optional :: OUTPUT
    logical :: yes_verbose
    integer :: IOUT
    real(long), allocatable :: A(:, :), U(:, :), U1(:, :)
    integer :: Ndiff, Ncsf1, Ncsf2, Nsingle, Nsingle2
    integer :: i, j
    integer :: oi, oj
    integer :: uocc_left(4), socc_left(4), docc_left(4)
    integer :: uocc_right(4), socc_right(4), docc_right(4)
    integer :: count_uocc_left, count_socc_left, count_docc_left

    yes_verbose = .False.
    IOUT = 6
    if (present(OUTPUT)) then
      IOUT = OUTPUT
      yes_verbose = .True.
    end if
    Nsingle = count(occ1 .eq. 1)
    Nsingle2 = count(occ2 .eq. 1)
    if (Nsingle2 > Nsingle) then
      stop "occ2 has more singles than occ1. This is not allowed!"
    end if
    !Ncsf1 = num_csf(Nsingle, multp)
    !Ncsf2 = num_csf(Nsingle2, multp)
    Ncsf1 = UT%f(Nsingle)
    Ncsf2 = UT%f(Nsingle2)
    allocate (dm(Ncsf1, Ncsf2, size(occ1, 1), size(occ1, 1)))
    dm(:, :, :, :) = ZERO
    Ndiff = sum(abs(occ1 - occ2))/2
    if (yes_verbose) then
      write (IOUT, '("# ",a,i5,a,i5)') 'Ndiff=', Ndiff
      write (IOUT, '("# occ1",x,99i1)') occ1
      write (IOUT, '("# occ2",x,99i1)') occ2
      write (IOUT, '("# multp: ",i4)') multp
      write (IOUT, '("# Num_csf1: ",i4)') Ncsf1
      write (IOUT, '("# Num_csf2: ",i4)') Ncsf2
    end if
    if (Ndiff > 1) then
      return
    end if

    uocc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    uocc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    count_uocc_left = count(uocc_left .ne. 0)
    count_socc_left = count(socc_left .ne. 0)
    count_docc_left = count(docc_left .ne. 0)


    allocate (A(Ncsf1, Ncsf2))
    allocate (U(Ncsf1, Ncsf1))
    allocate (U1(Ncsf1, Ncsf1))
    A = ZERO
    i = 0
    j = 0
    if (Ndiff .eq. 0) then
      if (yes_verbose) write (IOUT, "('# case diagonal')")
      do i = 1, Ncsf1
        do j = 1, size(occ1)
          dm(i, i, j, j) = dm(i, i, j, j) + dble(occ1(j))
        end do
      end do
    else if (Ndiff .eq. 1) then
      if ((count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 0) .or. &
        (count_uocc_left .eq. 0 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1)) then
        if (yes_verbose) write (IOUT, "('# case A1')")
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        call makeCycleUij(UT, oi, oj, Nsingle, U1)
        if (count_uocc_left .eq. 0) then
          ! 2,1 -> 1,2
          A(:, :) = -U1(1:Ncsf1, 1:Ncsf1)
        else
          ! 0,1 -> 1,0
          A(:, :) = U1(1:Ncsf1, 1:Ncsf1)
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! 1,1 -> 2,0  A(2)
        if (yes_verbose) write (IOUT, "('# case A2')")
        i = socc_left(2)
        j = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = transpose(matmul(U1, U)) ! P1 = (2...oi) (1...oj)
        ! Why transpose ?
        A(:, :) = SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
      else
        ! impossible
      end if

      if (occ1(i) < occ2(i)) then
        dm(:, :, j, i) = dm(:, :, j, i) + A(:, :)
      else
        dm(:, :, i, j) = dm(:, :, i, j) + A(:, :)
      end if
    else  ! if Ndiff>1
      ! this case is covered already earlier. We should never reach this point
      dm = ZERO
    end if
    ! deallocation
    deallocate (A)
    deallocate (U)
    deallocate (U1)
  end subroutine opdm_matrix_subblock

!> calculates the hamiltonian matrix elements for all configurations between occupations occ1 and occ2
!! occ1 and occ2 must be swapped such that there are more or equal single occupated orbitals in occ1 than in occ2
!! Variant for intern, extern:
!! nintern specifies orbital indices of active indices that can have any occupation
!! only one electron is allowed for the orbitals with indices larger than nintern
  subroutine hamiltonian_matrix_subblock_int_ext(occ1, occ2, multp, oeints, &
    teints, teints_iiee, teints_ieie, teints_iiie, E_core, &
    UT, nintern, H, OUTPUT)
    integer, intent(in) :: occ1(:), occ2(:)
    integer, intent(in) :: multp
    real(long), intent(in) :: oeints(:, :)
    real(long), intent(in) :: teints(:)
    real(long), intent(in) :: teints_iiee(:)
    real(long), intent(in) :: teints_ieie(:)
    real(long), intent(in) :: teints_iiie(:)
    real(long), intent(in) :: E_core
    type(UmatrixTable), intent(in) :: UT
    integer, intent(in) :: nintern
    real(long), intent(inout) :: H(:, :)
    integer, intent(in), optional :: OUTPUT
    logical :: yes_verbose
    integer :: IOUT
    real(long), allocatable :: A(:, :), B(:, :), U(:, :), U3(:, :), U2(:, :), U1(:, :)
    real(long) :: temp
    integer :: Ndiff, Ncsf1, Ncsf2, Nsingle, Nsingle2
    integer :: i, j, k, l
    integer(8) :: ii
    integer :: oi, oj, ok, ol
    integer :: uocc_left(4), socc_left(4), docc_left(4)
    integer :: uocc_right(4), socc_right(4), docc_right(4)
    integer :: count_uocc_left, count_socc_left, count_docc_left
    integer :: nactive, nextern

    nactive = size(oeints, 1)
    nextern = nactive - nintern
    yes_verbose = .False.
    IOUT = 6
    Ndiff = sum(abs(occ1 - occ2))/2
    if (Ndiff > 2) then
      return
    end if
    if (present(OUTPUT)) then
      IOUT = OUTPUT
      yes_verbose = .True.
    end if
    Nsingle = count(occ1 .eq. 1)
    Nsingle2 = count(occ2 .eq. 1)
    if (Nsingle2 > Nsingle) then
      stop "occ2 has more singles than occ1. This is not allowed!"
    end if
    Ncsf1 = UT%f(Nsingle)
    Ncsf2 = UT%f(Nsingle2)
    if (size(H, 1) .ne. Ncsf1 .or. size(H, 2) .ne. Ncsf2) then
      stop 'wrong dimension'
    end if
    H(:, :) = ZERO

    uocc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    uocc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    count_uocc_left = count(uocc_left .ne. 0)
    count_socc_left = count(socc_left .ne. 0)
    count_docc_left = count(docc_left .ne. 0)

    if (yes_verbose) then
      write (IOUT, '("# ",a,i5,a,i5)') 'Ndiff=', Ndiff
      write (IOUT, '("# occ1",x,99i1)') occ1
      write (IOUT, '("# occ2",x,99i1)') occ2
      write (IOUT, '("# multp: ",i4)') multp
      write (IOUT, '("# Num_csf1: ",i4)') Ncsf1
      write (IOUT, '("# Num_csf2: ",i4)') Ncsf2
    end if
    if (Ndiff .eq. 0) then
      allocate (U(Ncsf1, Ncsf1))
      H = ZERO
      if (yes_verbose) write (IOUT, "('# case diagonal')")
      temp = E_core
      do i = 1, size(occ1)
        ii = index_packed(i, i)
        temp = temp + occ1(i)*oeints(i, i)
        if (occ1(i) .eq. 2) then
          ! i can only be .le. nintern
          !iijj=index_packed(ii,ii)
          !temp = temp + teints(iijj)
          temp = temp + get_twoel(i, i, i, i, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        end if
        do j = 1, i - 1
          if (occ1(i)*occ1(j) > 0) then
            temp = temp + occ1(i)*occ1(j)* &
              get_twoel(i, i, j, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
          if (occ1(i)*occ1(j) .eq. 1) then
            oi = i - count(occ1(:i) .ne. 1)
            oj = j - count(occ1(:j) .ne. 1)
            call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
            H(:, :) = H(:, :) + U(1:Ncsf1, 1:Ncsf2)* &
              get_twoel(i, j, i, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          else if (occ1(i)*occ1(j) > 0) then
            temp = temp - HALF*occ1(i)*occ1(j)* &
              get_twoel(i, j, i, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
        end do
      end do

      do i = 1, size(H, 1)
        H(i, i) = H(i, i) + temp
      end do
      ! deallocation
      deallocate (U)
    else if (Ndiff .eq. 1) then
      allocate (A(Ncsf1, Ncsf2))
      allocate (U(Ncsf1, Ncsf1))
      allocate (U1(Ncsf1, Ncsf1))
      allocate (U2(Ncsf1, Ncsf1))
      A = ZERO
      H = ZERO

      if ((count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 0) .or. &
        (count_uocc_left .eq. 0 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1)) then
        if (yes_verbose) write (IOUT, "('# case A1')")
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        call makeCycleUij(UT, oi, oj, Nsingle, U1)
        if (count_uocc_left .eq. 0) then
          ! 2,1 -> 1,2
          A(:, :) = -U1(1:Ncsf1, 1:Ncsf1)
        else
          ! 0,1 -> 1,0
          A(:, :) = U1(1:Ncsf1, 1:Ncsf1)
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! 1,1 -> 2,0  A(2)
        if (yes_verbose) write (IOUT, "('# case A2')")
        i = socc_left(2)
        j = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = transpose(matmul(U1, U)) ! P1 = (2...oi) (1...oj)
        ! Why transpose ?
        A(:, :) = SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
      end if
      temp = oeints(i, j)
      if (occ1(j) .eq. 2) then
        temp = temp + get_twoel(i, j, j, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
      end if
      if (occ1(i) .eq. 2) then
        temp = temp + get_twoel(i, j, i, i, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
      end if
      if (occ2(j) .eq. 2) then
        temp = temp + get_twoel(i, j, j, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
      end if
      if (occ2(i) .eq. 2) then
        temp = temp + get_twoel(i, j, i, i, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
      end if
      do k = 1, size(occ2)
        if (k .eq. i .or. k .eq. j) cycle
        if (occ2(k) .eq. 0) cycle
        temp = temp + occ1(k)*get_twoel(i, j, k, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        if (occ1(k) .eq. 2) then
          temp = temp - get_twoel(i, k, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        end if
      end do
      H(:, :) = H(:, :) + A(:, :)*temp

      if (count_socc_left .eq. 1) then
        ! A1 case
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          U = matmul(U, U1) ! P0 = (i,k) P1
          if (count_docc_left .eq. 0) then
            U2 = U ! P2 = P0
          else
            U2 = U + U1 ! P2 = P1 + P0
          end if
          A(:, :) = U2(1:Ncsf1, 1:Ncsf2)
          temp = get_twoel(i, k, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          H(:, :) = H(:, :) + A(:, :)*temp
        end do! k
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! A(2) case
        ! oi and oj are singly occupied left
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          if (occ2(j) .eq. 0) then ! we need P2
            call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
          else if (occ2(j) .eq. 2) then ! we need P2'
            call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          end if
          ! P0 = (oj,k) * P1
          ! P0' = (oi,k) * P1
          U2 = matmul(U, U1) ! P2 = P0; P2' = P0'
          A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
          temp = get_twoel(i, k, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          H(:, :) = H(:, :) + A(:, :)*temp
        end do! k
      end if! A(1) or A(2)
      ! deallocation
      deallocate (A)
      deallocate (U)
      deallocate (U1)
      deallocate (U2)
    else if (Ndiff .eq. 2) then
      allocate (A(Ncsf1, Ncsf2))
      allocate (B(Ncsf1, Ncsf2))
      allocate (U(Ncsf1, Ncsf1))
      allocate (U1(Ncsf1, Ncsf1))
      allocate (U2(Ncsf1, Ncsf1))
      allocate (U3(Ncsf1, Ncsf1))
      A = ZERO
      H = ZERO

      if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 0 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case A3')")
        ! 2,0 -> 0,2
        ! identity matrix
        U(:, :) = 0
        do k = 1, size(U, 1)
          U(k, k) = 1
        end do
        i = uocc_left(1)
        j = docc_left(1)
        temp = get_twoel(i, j, i, j, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        A(:, :) = U(:, :)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case B1')")
        !2,1,0 -> 1,0,2
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        if (docc_left(1) .eq. uocc_right(1)) then
          k = docc_left(1)
        else
          k = docc_right(1)
        end if
        ! i is singly occupied left
        ! j is singly occupied right
        ! k is doubly occupied on one side and empty on the other
        call makeCycleUij(UT, oi, oj, Nsingle, U2)
        A(:, :) = -U2(1:Ncsf1, 1:Ncsf2)
        temp = get_twoel(i, k, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case B2')")
        i = socc_left(2)
        j = socc_left(1)
        if (count_docc_left .eq. 1) then
          k = docc_left(1)
        else if (count(docc_right .ne. 0) .eq. 1) then
          k = docc_right(1)
        end if
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U2)
        U2 = transpose(matmul(U2, U)) ! P2 = (1...oj) (2...oi) ^T
        if (count_docc_left .eq. 1) then
          U2 = -U2
        end if
        A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
        temp = get_twoel(i, k, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        H(:, :) = H(:, :) + A(:, :)*temp
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 2) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 2 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case C1-3')")
        ! 1,1,x,x -> x,x,1,1 (c1)
        ! 1,x,1,x -> x,1,x,1 (c2)
        ! x,1,1,x -> 1,x,x,1 (c3)
        i = socc_left(2)
        j = socc_left(1)
        l = socc_right(1)
        k = socc_right(2)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ2(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        if (ol .le. oj) then
          call makeCycleUij(UT, oi, ol, Nsingle, U)
          call makeCycleUij(UT, oj + 1, ok, Nsingle, U2)
          U2 = matmul(U, U2)  ! P2 = (i...l) (j+1...k)
          !call Umatrix(Nsingle, multp, oi, oj, U)
          call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
          U3 = matmul(U, U2) ! P3
        else
          call makeCycleUij(UT, oi, ol + 1, Nsingle, U)
          call makeCycleUij(UT, oj, ok, Nsingle, U2)
          U2 = matmul(U, U2) ! P2 = (i...l+1) (j...k)
          call makeCycleUij(UT, oj, ol, Nsingle, U3)
          call makeCycleUij(UT, oi, ok, Nsingle, U)
          U3 = matmul(U, U3) ! P3 = (i...k) (j...l)
        end if
        U1 = U2 + U3 ! P1 = P2 + P3
        if (count(uocc_left > 0) .eq. 1 .and. count(socc_left > 0) .eq. 2 .and. count(docc_left > 0) .eq. 1) then
          if (occ2(i) .eq. occ1(k)) then
            A = -U3
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          else
            A = -U2
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
          B = U1
          H(:, :) = H(:, :) + B(:, :)*get_twoel(i, j, k, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        else
          A = U3
          B = U2
          H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          H(:, :) = H(:, :) + B(:, :)*get_twoel(j, k, i, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          !H(:,:) = H(:,:) + A(:,:) * teints(ikjl)
          !H(:,:) = H(:,:) + B(:,:) * teints(iljk)
        end if
      else if (count(occ2 .eq. 2) - count(occ1 .eq. 2) .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! x,1,1,1 -> x,1,x,x (c4)
        ! 1,x,1,1 -> x,1,x,x (c5)
        ! 1,1,x,1 -> x,x,1,x (c6)
        ! 1,1,1,x -> x,x,1,x (c7)
        i = socc_left(3)
        j = socc_left(2)
        k = socc_left(1)
        l = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = matmul(U1, U) !   (2...oi) (1...oj)
        U1 = transpose(U1)
        if (l .ge. k) then
          call makeCycleUij(UT, ok + 2, ol + 2, Nsingle, U)
        else
          call makeCycleUij(UT, ol + 2, ok + 2, Nsingle, U)
          U = transpose(U)
        end if
        U1 = matmul(U1, U) !   (2...oi) (1...oj) (ok..ol)
        !call Umatrix(Nsingle, multp, oi, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
        U2 = matmul(U, U1) ! P2 = (i,k) P1
        !call Umatrix(Nsingle, multp, oj, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1) ! P3 = (j,k) P1
        if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 1) then
          if (occ2(k) .eq. 2) then
            A(:, :) = -SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P1
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, j, k, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          else
            A = -SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P2
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
          if (occ2(j) .eq. 2) then
            A = SQRTTWO*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P3' = P1+P2
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
            !H(:,:) = H(:,:) + A(:,:) * teints(ikjl)
          else ! here occ2(i) must be 2, so i cannot be extern
            A = SQRTTWO*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P2'= P1+P3
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
            !H(:,:) = H(:,:) + A(:,:) * teints(iljk)
          end if
        else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 0) then
          if (occ2(k) .eq. 2) then
            A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
            !H(:,:) = H(:,:) + A(:,:) * teints(iljk)
          else
            A = SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
            !H(:,:) = H(:,:) + A(:,:) * teints(ijkl)
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, j, k, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
          if (occ2(j) .eq. 2) then
            A = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            !H(:,:) = H(:,:) + A(:,:) * teints(iljk)
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          else
            A = SQRTTWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
            !H(:,:) = H(:,:) + A(:,:) * teints(ikjl)
            H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          end if
        else
          stop 'impossible'
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 4 .and. count_docc_left .eq. 0) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! 1,1,1,1 -> x,x,x,x (c8)
        i = socc_left(4)
        j = socc_left(3)
        k = socc_left(2)
        l = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ1(:l) .ne. 1)

        call makeCycleUij(UT, 1, ol, Nsingle, U1)
        call makeCycleUij(UT, 2, ok, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 3, oj, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 4, oi, Nsingle, U)
        U1 = matmul(U, U1)
        U1 = transpose(U1)
        ! P1 = (4..oi)(3..oj)(2..ok)(1..ol)^T
        !call Umatrix(Nsingle, multp, oj, ol, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ol, U)
        U2 = matmul(U, U1)  ! P2 = (j,l) P1
        !call Umatrix(Nsingle, multp, oj, ok, U)
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1)  ! P3 = (j,k) P1
        if (occ2(i) .eq. occ2(j)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          !H(:,:) = H(:,:) + A(:,:) * teints(ikjl)
          H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        else
          A = TWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
          !H(:,:) = H(:,:) + A(:,:) * teints(ijkl)
          H(:, :) = H(:, :) + A(:, :)*get_twoel(i, j, k, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
        end if
        if (occ2(j) .eq. occ2(k)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          H(:, :) = H(:, :) + A(:, :)*get_twoel(i, k, j, l, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          !H(:,:) = H(:,:) + A(:,:) * teints(ikjl)
        else
          A = TWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
          H(:, :) = H(:, :) + A(:, :)*get_twoel(i, l, j, k, teints, teints_iiee, teints_ieie, teints_iiie, nintern, nextern)
          !H(:,:) = H(:,:) + A(:,:) * teints(iljk)
        end if
      else! we shouldn't reach this point
        stop "reached an impossible situation!"
      end if
      ! deallocation
      deallocate (A)
      deallocate (B)
      deallocate (U)
      deallocate (U1)
      deallocate (U2)
      deallocate (U3)
    else  ! if Ndiff>2
      ! this case is covered already earlier. We should never reach this point
      H = ZERO
    end if

  contains
    pure function get_twoel(i, j, k, l, teints_iiii, teints_iiee, teints_ieie, teints_iiie, nintern, nextern) result(teint)
      integer, intent(in) :: i, j, k, l, nintern, nextern
      real(long), intent(in) :: teints_iiii(:), teints_iiee(:), teints_ieie(:), teints_iiie(:)
      integer(8) :: ij, kl, ijkl, nee
      real(long) :: teint
      if (i > nintern .and. j > nintern) then
        nee = nextern*(nextern + 1)/2
        ij = index_packed(i - nintern, j - nintern)
        kl = index_packed(k, l)
        ijkl = ij + (kl - 1)*nee
        teint = teints_iiee(ijkl)
      else if (k > nintern .and. l > nintern) then
        nee = nextern*(nextern + 1)/2
        kl = index_packed(k - nintern, l - nintern)
        ij = index_packed(i, j)
        ijkl = kl + (ij - 1)*nee
        teint = teints_iiee(ijkl)
      else if (i > nintern .and. k > nintern) then
        ij = (j - 1)*nextern + i - nintern
        kl = (l - 1)*nextern + k - nintern
        ijkl = index_packed(ij, kl)
        teint = teints_ieie(ijkl)
      else if (i > nintern .and. l > nintern) then
        ij = (j - 1)*nextern + i - nintern
        kl = (k - 1)*nextern + l - nintern
        ijkl = index_packed(ij, kl)
        teint = teints_ieie(ijkl)
      else if (j > nintern .and. k > nintern) then
        ij = (i - 1)*nextern + j - nintern
        kl = (l - 1)*nextern + k - nintern
        ijkl = index_packed(ij, kl)
        teint = teints_ieie(ijkl)
      else if (j > nintern .and. l > nintern) then
        ij = (i - 1)*nextern + j - nintern
        kl = (k - 1)*nextern + l - nintern
        ijkl = index_packed(ij, kl)
        teint = teints_ieie(ijkl)
      else if (i > nintern) then
        ij = (j - 1)*nextern + i - nintern
        kl = index_packed(k, l)
        ijkl = (ij) + (kl - 1)*(nextern*nintern)
        teint = teints_iiie(ijkl)
      else if (j > nintern) then
        ij = (i - 1)*nextern + j - nintern
        kl = index_packed(k, l)
        ijkl = (ij) + (kl - 1)*(nextern*nintern)
        teint = teints_iiie(ijkl)
      else if (k > nintern) then
        kl = (l - 1)*nextern + k - nintern
        ij = index_packed(i, j)
        ijkl = (kl) + (ij - 1)*(nextern*nintern)
        teint = teints_iiie(ijkl)
      else if (l > nintern) then
        kl = (k - 1)*nextern + l - nintern
        ij = index_packed(i, j)
        ijkl = (kl) + (ij - 1)*(nextern*nintern)
        teint = teints_iiie(ijkl)
      else if (i .le. nintern .and. j .le. nintern .and. k .le. nintern .and. l .le. nintern) then
        ij = index_packed(i, j)
        kl = index_packed(k, l)
        ijkl = index_packed(ij, kl)
        teint = teints_iiii(ijkl)
      else
        !stop 'impossible case'
      end if

      ! ij = index_packed(i,j)
      ! kl = index_packed(k,l)
      ! ijkl = index_packed(ij,kl)
      ! if(abs(teints_iiii(ijkl) - teint)>1e-5)then
      !   write(6,'(4i4,2es)') i,j,k,l,teints_iiii(ijkl),teint
      !   stop
      ! end if

    end function get_twoel
  end subroutine hamiltonian_matrix_subblock_int_ext

!> calculates the matrix elements of the symmetrized two particle density matrix
!! 1/8 (< phi_a | (c_i^+ c_j^+ c_k c_l) + (c_j^+ c_i^+ c_k c_l) + (c_j^+ c_i^+ c_l c_k) + (c_i^+ c_j^+ c_l c_k)| phi_b >
!! + < phi_b | (c_i^+ c_j^+ c_k c_l) + (c_j^+ c_i^+ c_k c_l) + (c_j^+ c_i^+ c_l c_k) + (c_i^+ c_j^+ c_l c_k)| phi_a > )
!! for all phi_a and phi_b specified in occ1 and occ2
!! occ1 and occ2 must be swapped such that there are more or equal single occupated orbitals in occ1 than in occ2
  subroutine tpdm_matrix_subblock(occ1, occ2, UT, stpdm, OUTPUT)
    integer, intent(in) :: occ1(:), occ2(:)
    type(UmatrixTable), intent(in) :: UT
    real(long), intent(out), allocatable :: stpdm(:, :, :)
    integer, intent(in), optional :: OUTPUT
    logical :: yes_verbose
    integer :: IOUT
    integer :: multp

    real(long), allocatable :: A(:, :), U(:, :), U3(:, :), U2(:, :), U1(:, :)
    integer :: Ndiff, Ncsf1, Ncsf2, Nsingle, Nsingle2
    integer :: i, j, k, l, m
    integer(8) :: ii, jj, kk, kl, ij, ik, jk, il, jl, iijj, ijij, ijii, ijjj, ijkk, ikjk, iljk, ikjl, ijkl
    integer :: oi, oj, ok, ol
    integer :: uocc_left(4), socc_left(4), docc_left(4)
    integer :: uocc_right(4), socc_right(4), docc_right(4)
    integer :: count_uocc_left, count_socc_left, count_docc_left
    integer :: n_one, n_two, n_four

    n_one = size(occ1, 1)
    n_two = (n_one + 1)*n_one/2
    n_four = (n_two + 1)*n_two/2

    multp = UT%multp
    yes_verbose = .False.
    IOUT = 6
    if (present(OUTPUT)) then
      IOUT = OUTPUT
      yes_verbose = .True.
    end if
    Nsingle = count(occ1 .eq. 1)
    Nsingle2 = count(occ2 .eq. 1)
    if (Nsingle2 > Nsingle) then
      stop "occ2 has more single occupied orbitals than occ1. This is not allowed!"
    end if
    !Ncsf1 = num_csf(Nsingle, multp)
    !Ncsf2 = num_csf(Nsingle2, multp)
    Ncsf1 = UT%f(Nsingle)
    Ncsf2 = UT%f(Nsingle2)
    allocate (stpdm(Ncsf1, Ncsf2, n_four))
    stpdm = ZERO
    Ndiff = sum(abs(occ1 - occ2))/2
    if (Ndiff > 2) then
      return
    end if
    uocc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_left = pack((/(j, j=1, size(occ1, 1))/), occ1(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    uocc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 0 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    socc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 1 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    docc_right = pack((/(j, j=1, size(occ1, 1))/), occ2(:) .eq. 2 .and. occ1(:) .ne. occ2(:), vector=(/0, 0, 0, 0/))
    count_uocc_left = count(uocc_left .ne. 0)
    count_socc_left = count(socc_left .ne. 0)
    count_docc_left = count(docc_left .ne. 0)
    if (yes_verbose) then
      write (IOUT, '("# ",a,i5,a,i5)') 'Ndiff=', Ndiff
      write (IOUT, '("# occ1",x,99i1)') occ1
      write (IOUT, '("# occ2",x,99i1)') occ2
      write (IOUT, '("# multp: ",i4)') multp
      write (IOUT, '("# Num_csf1: ",i4)') Ncsf1
      write (IOUT, '("# Num_csf2: ",i4)') Ncsf2
    end if
    allocate (A(Ncsf1, Ncsf2))
    allocate (U(Ncsf1, Ncsf1))
    allocate (U1(Ncsf1, Ncsf1))
    allocate (U2(Ncsf1, Ncsf1))
    allocate (U3(Ncsf1, Ncsf1))
    A = ZERO
    if (Ndiff .eq. 0) then
      if (yes_verbose) write (IOUT, "('# case diagonal')")
      do i = 1, size(occ1)
        ii = index_packed(i, i)
        if (occ1(i) .eq. 2) then
          iijj = index_packed(ii, ii)
          do m = 1, Ncsf1
            stpdm(m, m, iijj) = stpdm(m, m, iijj) + TWO
          end do
        end if
        ii = index_packed(i, i)
        do j = 1, i - 1
          jj = index_packed(j, j)
          ij = index_packed(i, j)
          iijj = index_packed(ii, jj)
          ijij = index_packed(ij, ij)
          if (occ1(i)*occ1(j) > 0) then
            do m = 1, Ncsf1
              stpdm(m, m, iijj) = stpdm(m, m, iijj) + dble(occ1(i)*occ1(j))
            end do
          end if
          if (occ1(i)*occ1(j) .eq. 1) then
            oi = i - count(occ1(:i) .ne. 1)
            oj = j - count(occ1(:j) .ne. 1)
            call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
            stpdm(:, :, ijij) = stpdm(:, :, ijij) + HALF*U(1:Ncsf1, 1:Ncsf2)
          else if (occ1(i)*occ1(j) > 0) then
            do m = 1, Ncsf1
              stpdm(m, m, ijij) = stpdm(m, m, ijij) - FOURTH*dble(occ1(i)*occ1(j))
            end do
          end if
        end do
      end do
    else if (Ndiff .eq. 1) then
      if ((count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 0) .or. &
        (count_uocc_left .eq. 0 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1)) then
        if (yes_verbose) write (IOUT, "('# case A1')")
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        call makeCycleUij(UT, oi, oj, Nsingle, U1)
        if (count_uocc_left .eq. 0) then
          ! 2,1 -> 1,2
          A(:, :) = -U1(1:Ncsf1, 1:Ncsf1)
        else
          ! 0,1 -> 1,0
          A(:, :) = U1(1:Ncsf1, 1:Ncsf1)
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! 1,1 -> 2,0  A(2)
        if (yes_verbose) write (IOUT, "('# case A2')")
        i = socc_left(2)
        j = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = transpose(matmul(U1, U)) ! P1 = (2...oi) (1...oj)
        ! Why transpose ?
        A(:, :) = SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
      end if
      ii = index_packed(i, i)
      jj = index_packed(j, j)
      ij = index_packed(i, j)
      iijj = index_packed(ii, jj)
      ijij = index_packed(ij, ij)
      ijii = index_packed(ij, ii)
      ijjj = index_packed(ij, jj)
      if (occ1(j) .eq. 2) then
        stpdm(:, :, ijjj) = stpdm(:, :, ijjj) + HALF*A(:, :)
      end if
      if (occ1(i) .eq. 2) then
        stpdm(:, :, ijii) = stpdm(:, :, ijii) + HALF*A(:, :)
      end if
      if (occ2(j) .eq. 2) then
        stpdm(:, :, ijjj) = stpdm(:, :, ijjj) + HALF*A(:, :)
      end if
      if (occ2(i) .eq. 2) then
        stpdm(:, :, ijii) = stpdm(:, :, ijii) + HALF*A(:, :)
      end if
      do k = 1, size(occ2)
        if (k .eq. i .or. k .eq. j) cycle
        if (occ2(k) .eq. 0) cycle
        kk = index_packed(k, k)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ijkk = index_packed(ij, kk)
        ikjk = index_packed(ik, jk)
        stpdm(:, :, ijkk) = stpdm(:, :, ijkk) + HALF*dble(occ1(k))*A(:, :)
        if (occ1(k) .eq. 2) then
          stpdm(:, :, ikjk) = stpdm(:, :, ikjk) - FOURTH*A(:, :)
        end if
      end do
      if (count_socc_left .eq. 1) then
        ! A1 case
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          ikjk = index_packed(ik, jk)
          call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          U = matmul(U, U1) ! P0 = (i,k) P1
          if (count_docc_left .eq. 0) then
            U2 = U ! P2 = P0
          else
            U2 = U + U1 ! P2 = P1 + P0
          end if
          A(:, :) = U2(1:Ncsf1, 1:Ncsf2)
          stpdm(:, :, ikjk) = stpdm(:, :, ikjk) + FOURTH*A(:, :)
        end do! k
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0) then
        ! A(2) case
        ! oi and oj are singly occupied left
        do k = 1, size(occ2)
          if (k .eq. i .or. k .eq. j) cycle
          if (occ1(k) .ne. 1 .or. occ2(k) .ne. 1) cycle
          ok = k - count(occ1(:k) .ne. 1)
          ik = index_packed(i, k)
          jk = index_packed(j, k)
          ikjk = index_packed(ik, jk)
          if (occ2(j) .eq. 0) then ! we need P2
            call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
          else if (occ2(j) .eq. 2) then ! we need P2'
            call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
          end if
          ! P0 = (oj,k) * P1
          ! P0' = (oi,k) * P1
          U2 = matmul(U, U1) ! P2 = P0; P2' = P0'
          A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
          stpdm(:, :, ikjk) = stpdm(:, :, ikjk) + FOURTH*A(:, :)
        end do! k
      end if! A(1) or A(2)
    else if (Ndiff .eq. 2) then
      if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 0 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case A3')")
        ! 2,0 -> 0,2
        ! identity matrix
        U(:, :) = 0
        do k = 1, size(U, 1)
          U(k, k) = 1
        end do
        i = uocc_left(1)
        j = docc_left(1)
        ij = index_packed(i, j)
        ijij = index_packed(ij, ij)
        A(:, :) = U(:, :)
        stpdm(:, :, ijij) = stpdm(:, :, ijij) + HALF*A(:, :)
      else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 1 .and. count_docc_left .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case B1')")
        !2,1,0 -> 1,0,2
        i = socc_left(1)
        j = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ2(:j) .ne. 1)
        if (docc_left(1) .eq. uocc_right(1)) then
          k = docc_left(1)
        else
          k = docc_right(1)
        end if
        ! i is singly occupied left
        ! j is singly occupied right
        ! k is doubly occupied on one side and empty on the other
        call makeCycleUij(UT, oi, oj, Nsingle, U2)
        A(:, :) = -U2(1:Ncsf1, 1:Ncsf2)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ikjk = index_packed(ik, jk)
        stpdm(:, :, ikjk) = stpdm(:, :, ikjk) + FOURTH*A(:, :)
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case B2')")
        i = socc_left(2)
        j = socc_left(1)
        if (count_docc_left .eq. 1) then
          k = docc_left(1)
        else if (count(docc_right .ne. 0) .eq. 1) then
          k = docc_right(1)
        end if
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U2)
        U2 = transpose(matmul(U2, U)) ! P2 = (1...oj) (2...oi) ^T
        if (count_docc_left .eq. 1) then
          U2 = -U2
        end if
        A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)
        ik = index_packed(i, k)
        jk = index_packed(j, k)
        ikjk = index_packed(ik, jk)
        stpdm(:, :, ikjk) = stpdm(:, :, ikjk) + FOURTH*A(:, :)
      else if ((count_uocc_left .eq. 0 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 2) .or. &
        (count_uocc_left .eq. 1 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 1) .or. &
        (count_uocc_left .eq. 2 .and. count_socc_left .eq. 2 .and. count_docc_left .eq. 0)) then
        if (yes_verbose) write (IOUT, "('# case C1-3')")
        ! 1,1,x,x -> x,x,1,1 (c1)
        ! 1,x,1,x -> x,1,x,1 (c2)
        ! x,1,1,x -> 1,x,x,1 (c3)
        i = socc_left(2)
        j = socc_left(1)
        l = socc_right(1)
        k = socc_right(2)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ2(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        jk = index_packed(j, k)
        jl = index_packed(j, l)
        kl = index_packed(k, l)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)
        if (ol .le. oj) then
          call makeCycleUij(UT, oi, ol, Nsingle, U)
          call makeCycleUij(UT, oj + 1, ok, Nsingle, U2)
          U2 = matmul(U, U2)  ! P2 = (i...l) (j+1...k)
          call getUMatrixFromTable(UT, Nsingle, oi, oj, U)
          U3 = matmul(U, U2) ! P3
        else
          call makeCycleUij(UT, oi, ol + 1, Nsingle, U)
          call makeCycleUij(UT, oj, ok, Nsingle, U2)
          U2 = matmul(U, U2) ! P2 = (i...l+1) (j...k)
          call makeCycleUij(UT, oj, ol, Nsingle, U3)
          call makeCycleUij(UT, oi, ok, Nsingle, U)
          U3 = matmul(U, U3) ! P3 = (i...k) (j...l)
        end if
        U1 = U2 + U3 ! P1 = P2 + P3
        if (count(uocc_left > 0) .eq. 1 .and. count(socc_left > 0) .eq. 2 .and. count(docc_left > 0) .eq. 1) then
          if (occ2(i) .eq. occ1(k)) then
            A = -U3
            stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
          else
            A = -U2
            stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
          end if
          A = U1
          stpdm(:, :, ijkl) = stpdm(:, :, ijkl) + FOURTH*A(:, :)
        else
          A = U3
          stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
          A = U2
          stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
        end if
      else if (count(occ2 .eq. 2) - count(occ1 .eq. 2) .eq. 1) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! x,1,1,1 -> x,1,x,x (c4)
        ! 1,x,1,1 -> x,1,x,x (c5)
        ! 1,1,x,1 -> x,x,1,x (c6)
        ! 1,1,1,x -> x,x,1,x (c7)
        i = socc_left(3)
        j = socc_left(2)
        k = socc_left(1)
        l = socc_right(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ2(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        kl = index_packed(k, l)
        jl = index_packed(j, l)
        jk = index_packed(j, k)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)
        call makeCycleUij(UT, 1, oj, Nsingle, U)
        call makeCycleUij(UT, 2, oi, Nsingle, U1)
        U1 = matmul(U1, U) !   (2...oi) (1...oj)
        U1 = transpose(U1)
        if (l .ge. k) then
          call makeCycleUij(UT, ok + 2, ol + 2, Nsingle, U)
        else
          call makeCycleUij(UT, ol + 2, ok + 2, Nsingle, U)
          U = transpose(U)
        end if
        U1 = matmul(U1, U) !   (2...oi) (1...oj) (ok..ol)
        call getUMatrixFromTable(UT, Nsingle, oi, ok, U)
        U2 = matmul(U, U1) ! P2 = (i,k) P1
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1) ! P3 = (j,k) P1
        if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 1) then
          if (occ2(k) .eq. 2) then
            A(:, :) = -SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P1
            stpdm(:, :, ijkl) = stpdm(:, :, ijkl) + FOURTH*A(:, :)
          else
            A = -SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! - P2
            stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
          end if
          if (occ2(j) .eq. 2) then
            A = SQRTTWO*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P3' = P1+P2
            stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
          else
            A = SQRTTWO*(U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) + U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1)) ! P2'= P1+P3
            stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
          end if
        else if (count_uocc_left .eq. 1 .and. count_socc_left .eq. 3 .and. count_docc_left .eq. 0) then
          if (occ2(k) .eq. 2) then
            A(:, :) = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
          else
            A = SQRTTWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
            !H(:,:) = H(:,:) + A(:,:) * teints(ijkl)
            stpdm(:, :, ijkl) = stpdm(:, :, ijkl) + FOURTH*A(:, :)
          end if
          if (occ2(j) .eq. 2) then
            A = SQRTTWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
            stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
          else
            A = SQRTTWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
            stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
          end if
        else
          stop 'impossible'
        end if
      else if (count_uocc_left .eq. 0 .and. count_socc_left .eq. 4 .and. count_docc_left .eq. 0) then
        if (yes_verbose) write (IOUT, "('# case C4-7')")
        ! 1,1,1,1 -> x,x,x,x (c8)
        i = socc_left(4)
        j = socc_left(3)
        k = socc_left(2)
        l = socc_left(1)
        oi = i - count(occ1(:i) .ne. 1)
        oj = j - count(occ1(:j) .ne. 1)
        ok = k - count(occ1(:k) .ne. 1)
        ol = l - count(occ1(:l) .ne. 1)
        ik = index_packed(i, k)
        il = index_packed(i, l)
        ij = index_packed(i, j)
        kl = index_packed(k, l)
        jl = index_packed(j, l)
        jk = index_packed(j, k)
        ikjl = index_packed(ik, jl)
        iljk = index_packed(il, jk)
        ijkl = index_packed(ij, kl)
        call makeCycleUij(UT, 1, ol, Nsingle, U1)
        call makeCycleUij(UT, 2, ok, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 3, oj, Nsingle, U)
        U1 = matmul(U, U1)
        call makeCycleUij(UT, 4, oi, Nsingle, U)
        U1 = matmul(U, U1)
        U1 = transpose(U1)
        ! P1 = (4..oi)(3..oj)(2..ok)(1..ol)^T
        call getUMatrixFromTable(UT, Nsingle, oj, ol, U)
        U2 = matmul(U, U1)  ! P2 = (j,l) P1
        call getUMatrixFromTable(UT, Nsingle, oj, ok, U)
        U3 = matmul(U, U1)  ! P3 = (j,k) P1
        if (occ2(i) .eq. occ2(j)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
        else
          A = TWO*U1(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P1
          stpdm(:, :, ijkl) = stpdm(:, :, ijkl) + FOURTH*A(:, :)
        end if
        if (occ2(j) .eq. occ2(k)) then
          A = TWO*U3(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P3
          stpdm(:, :, ikjl) = stpdm(:, :, ikjl) + FOURTH*A(:, :)
        else
          A = TWO*U2(1:Ncsf1, Ncsf1 - Ncsf2 + 1:Ncsf1) ! P2
          stpdm(:, :, iljk) = stpdm(:, :, iljk) + FOURTH*A(:, :)
        end if
      else! we shouldn't reach this point
        stop "reached an impossible situation!"
      end if
    else  ! if Ndiff>2
      ! this case is covered already earlier. We should never reach this point
      !H = ZERO
    end if
    ! deallocation
    deallocate (A)
    deallocate (U)
    deallocate (U1)
    deallocate (U2)
    deallocate (U3)
  end subroutine tpdm_matrix_subblock

  subroutine two_particle_dm_two_det(tpdm, occ1, occ2, det1, det2, coef12)
    real(long), intent(inout) :: tpdm(:, :, :, :)
    integer, intent(in) :: occ1(:)
    integer, intent(in) :: occ2(:)
    integer, intent(in) :: det1(:)
    integer, intent(in) :: det2(:)
    !  real(long), intent(in) :: coef1
    !  real(long), intent(in) :: coef2
    real(long), intent(in) :: coef12

    integer :: only_left(2), only_left_spin(2)
    integer :: only_right(2), only_right_spin(2)
    integer :: coef_sign
    integer :: oi, spin_mo1
    integer :: mo_a, mo_b, mo_c, mo_d, mo1
    call compare_det(coef_sign, only_left, only_left_spin, only_right, only_right_spin, occ1, det1, occ2, det2)
    if (coef_sign .eq. 0) then
      ! if there are more than two differences return
    else if (count(only_left(:) .ne. 0) .eq. 0 .and. &
      count(only_right(:) .ne. 0) .eq. 0) then
      ! if there are zero differences call two_particle_dm_det
      call two_particle_dm_det(tpdm, occ1, det1, coef12)
    else if (count(only_left(:) .ne. 0) .eq. 1 .and. &
      count(only_right(:) .ne. 0) .eq. 1 .and. &
      only_left_spin(1) .eq. only_right_spin(1)) then
      ! left determinant has one and right determinant has one electron unique, same spin
      oi = 0
      do mo1 = 1, size(occ1, 1)
        if (mo1 .eq. only_left(1) .and. occ1(mo1) .eq. 2) then
          !        write(OUTPUT,'(3x,4i1,f)') only_left(1),only_right(1),mo1,mo1,coef_sign*ONE*coef12
          tpdm(only_left(1), only_right(1), mo1, mo1) = tpdm(only_left(1), only_right(1), mo1, mo1) + coef_sign*coef12*ONE
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,only_left(1),only_right(1),coef_sign*ONE*coef12
          tpdm(mo1, mo1, only_left(1), only_right(1)) = tpdm(mo1, mo1, only_left(1), only_right(1)) + coef_sign*coef12*ONE
        else if (occ1(mo1) .eq. 2) then
          !        write(OUTPUT,'(3x,4i1,f)') only_left(1),only_right(1),mo1,mo1,coef_sign*TWO*coef12
          tpdm(only_left(1), only_right(1), mo1, mo1) = tpdm(only_left(1), only_right(1), mo1, mo1) + coef_sign*coef12*TWO
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,only_left(1),only_right(1),coef_sign*TWO*coef12
          tpdm(mo1, mo1, only_left(1), only_right(1)) = tpdm(mo1, mo1, only_left(1), only_right(1)) + coef_sign*coef12*TWO
          !        write(OUTPUT,'(3x,4i1,f)') mo1,only_right(1),only_left(1),mo1,-coef_sign*ONE*coef12
          tpdm(mo1, only_right(1), only_left(1), mo1) = tpdm(mo1, only_right(1), only_left(1), mo1) - coef_sign*coef12*ONE
          !        write(OUTPUT,'(3x,4i1,f)') only_left(1),mo1,mo1,only_right(1),-coef_sign*ONE*coef12
          tpdm(only_left(1), mo1, mo1, only_right(1)) = tpdm(only_left(1), mo1, mo1, only_right(1)) - coef_sign*coef12*ONE
        else if (occ1(mo1) .eq. 1) then
          oi = oi + 1
          if (oi .eq. 1) then
            spin_mo1 = det1(oi)
          else
            spin_mo1 = det1(oi) - det1(oi - 1)
          end if
          !        write(OUTPUT,'(3x,4i1,f)') only_left(1),only_right(1),mo1,mo1,coef_sign*ONE*coef12
          tpdm(only_left(1), only_right(1), mo1, mo1) = tpdm(only_left(1), only_right(1), mo1, mo1) + coef_sign*coef12*occ1(mo1)
          !        write(OUTPUT,'(3x,4i1,f)') mo1,mo1,only_left(1),only_right(1),coef_sign*ONE*coef12
          tpdm(mo1, mo1, only_left(1), only_right(1)) = tpdm(mo1, mo1, only_left(1), only_right(1)) + coef_sign*coef12*occ1(mo1)
          if (spin_mo1 .eq. only_right_spin(1)) then
            !          write(OUTPUT,'(3x,4i1,f)') mo1,only_right(1),only_left(1),mo1,-coef_sign*ONE*coef12
            tpdm(mo1, only_right(1), only_left(1), mo1) = tpdm(mo1, only_right(1), only_left(1), mo1) - coef_sign*coef12*occ1(mo1)
            !          write(OUTPUT,'(3x,4i1,f)') only_left(1),mo1,mo1,only_right(1),-coef_sign*ONE*coef12
            tpdm(only_left(1), mo1, mo1, only_right(1)) = tpdm(only_left(1), mo1, mo1, only_right(1)) - coef_sign*coef12*occ1(mo1)
          end if
        end if
      end do
    else if (count(only_left(:) .ne. 0) .eq. 2 .and. count(only_right(:) .ne. 0) .eq. 2 &
      .and. only_left_spin(1) .eq. only_right_spin(1) .and. only_left_spin(2) .eq. only_right_spin(2)) then

      !    if(mod( (sum(occ1)-1)*(sum(occ1)-2)/2 ,2) .eq.1 )&
      !      coef_sign=coef_sign*-1
      !    write(OUTPUT,"(3x,'only left:',2i2,2i2,'only right:',2i2,2i2 )") only_left(:),only_left_spin(:),only_right(:),only_right_spin(:)
      ! determinant differ by two electrons: <ab|ab>
      mo_a = only_left(1)
      mo_b = only_left(2)
      mo_c = only_right(1)
      mo_d = only_right(2)
      !    write(OUTPUT,'(3x,4i1,f)') mo_a,mo_c,mo_b,mo_d,coef_sign*ONE*coef12
      tpdm(mo_a, mo_c, mo_b, mo_d) = tpdm(mo_a, mo_c, mo_b, mo_d) + coef_sign*coef12
      !    write(6,'(3x,4i1,f)') mo_b,mo_d,mo_a,mo_c,coef_sign*ONE*coef12
      tpdm(mo_b, mo_d, mo_a, mo_c) = tpdm(mo_b, mo_d, mo_a, mo_c) + coef_sign*coef12

      if (only_left_spin(1) .eq. only_left_spin(2)) then
        !      write(6,'(3x,4i1,f)') mo_a,mo_d,mo_b,mo_c,-coef_sign*ONE*coef12
        tpdm(mo_a, mo_d, mo_b, mo_c) = tpdm(mo_a, mo_d, mo_b, mo_c) - coef_sign*coef12
        !      write(6,'(3x,4i1,f)') mo_b,mo_c,mo_a,mo_d,-coef_sign*ONE*coef12
        tpdm(mo_b, mo_c, mo_a, mo_d) = tpdm(mo_b, mo_c, mo_a, mo_d) - coef_sign*coef12
      end if
    else if (count(only_left_spin(:) .ne. 0) .eq. 2 .and. count(only_right_spin(:) .ne. 0) .eq. 2 &
      .and. only_left_spin(2) .eq. only_right_spin(1) .and. only_left_spin(1) .eq. only_right_spin(2)) then
      !    write(6,"(3x,'only left:',2i2,2i2,'only right:',2i2,2i2 )") only_left(:),only_left_spin(:),only_right(:),only_right_spin(:)
      ! determinant differ by two electrons: <ab|ba>
      mo_a = only_left(1)
      mo_b = only_left(2)
      mo_c = only_right(1)
      mo_d = only_right(2)
      !    write(6,'(3x,4i1,f)') mo_a,mo_d,mo_b,mo_c,-coef_sign*ONE*coef12
      tpdm(mo_a, mo_d, mo_b, mo_c) = tpdm(mo_a, mo_d, mo_b, mo_c) - coef_sign*coef12
      !    write(6,'(3x,4i1,f)') mo_b,mo_c,mo_a,mo_d,-coef_sign*ONE*coef12
      tpdm(mo_b, mo_c, mo_a, mo_d) = tpdm(mo_b, mo_c, mo_a, mo_d) - coef_sign*coef12
      if (only_left_spin(1) .eq. only_left_spin(2)) then
        !      write(6,'(3x,4i1,f)') mo_a,mo_c,mo_b,mo_d,+coef_sign*ONE*coef12
        tpdm(mo_a, mo_c, mo_b, mo_d) = tpdm(mo_a, mo_c, mo_b, mo_d) + coef_sign*coef12
        !      write(6,'(3x,4i1,f)') mo_b,mo_d,mo_a,mo_c,+coef_sign*ONE*coef12
        tpdm(mo_b, mo_d, mo_a, mo_c) = tpdm(mo_b, mo_d, mo_a, mo_c) + coef_sign*coef12
      end if
    end if
  end subroutine two_particle_dm_two_det

  subroutine calculate_gradient(F_act, F_core, opdm, s_tpdmaa, teints_abij, core, active, virtual, pairs, gradient)
    real(long), intent(in) :: F_act(:, :)
    real(long), intent(in) :: F_core(:, :)
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: s_tpdmaa(:)
    real(long), intent(in) :: teints_abij(:)
    integer, intent(in) :: core(:)
    integer, intent(in) :: active(:)
    integer, intent(in) :: virtual(:)
    type(indep_mo_pair), intent(in) :: pairs(:)
    real(long), intent(out) :: gradient(:)
    integer :: mo1, mo2, pair
    real(long) :: genFock(size(F_act, 1), size(F_act, 2))

    call calculate_genFock(F_act, F_core, opdm, s_tpdmaa, teints_abij, core, active, virtual, genFock)

    gradient(:) = ZERO
    do pair = 1, size(pairs, 1)
      mo1 = pairs(pair)%MO1
      mo2 = pairs(pair)%MO2
      gradient(pair) = TWO*(genFock(mo1, mo2) - genFock(mo2, mo1))
    end do
  end subroutine calculate_gradient

  !> calculates the generalized Fock matrix
  !!
  subroutine calculate_genFock(F_act, F_core, opdm, s_tpdmaa, teints_abij, core, active, virtual, genFock)
    real(long), intent(in) :: F_act(:, :)
    real(long), intent(in) :: F_core(:, :)
    real(long), intent(in) :: opdm(:, :)
    real(long), intent(in) :: s_tpdmaa(:)
    real(long), intent(in) :: teints_abij(:)
    integer, intent(in) :: core(:)
    integer, intent(in) :: active(:)
    integer, intent(in) :: virtual(:)
    real(long), intent(out) :: genFock(:, :)
    integer :: v, w, x, y, n, ni, nv, nw, vi
    integer(8) :: vw, xy, nij, nvw, nwxy, vwxy
    ni = size(F_act, 1)
    nij = ni*(ni + 1)/2
    nv = size(active, 1)
    nvw = nv*(nv + 1)/2
    genFock = ZERO
    ! core-general part
    genFock(core(:), :) = TWO*(F_act(core(:), :) + F_core(core(:), :))
    ! active-general part
    do v = 1, nv
      vi = active(v)
      do n = 1, ni
        genFock(vi, n) = genFock(vi, n) + &
          sum(F_core(n, active(:))*opdm(vi, active(:)))
        do w = 1, nv
          vw = index_packed(v, w)
          nw = int(index_packed(active(w), n))
          do x = 1, nv
            do y = 1, nv
              xy = index_packed(x, y)
              vwxy = index_packed(vw, xy)
              nwxy = (xy - 1)*nij + nw
              genFock(vi, n) = genFock(vi, n) + &
                ONE*s_tpdmaa(vwxy)*teints_abij(nwxy)
            end do! y
          end do! x
        end do! w
      end do! n
    end do! v
    ! virtual-general part
    genFock(virtual(:), :) = ZERO
  end subroutine calculate_genFock

  subroutine reorder_RSTRCT(C, CE)
    !! it just reorders the eigenvectors such that C is most
    !! similar to a diagonal matrix
    real(long), dimension(:, :), intent(inout) :: C !! eigenvectors
    real(long), dimension(:), intent(inout) :: CE !! eigenvalues
    integer i, j, k
    integer N
    real(long) temp
    N = size(CE)
    do i = 1, N
      k = i
      do j = 1, N
        if (abs(C(i, j)) > abs(C(i, k)) .and. abs(C(i, j)) > abs(C(j, j))) then
          k = j
        end if
      end do
      if ((k .ne. i)) then
        ! swap k <-> i
        do j = 1, N
          temp = C(j, i)
          C(j, i) = C(j, k)
          C(j, k) = temp
        end do
        temp = CE(i)
        CE(i) = CE(k)
        CE(k) = temp
      end if
    end do
    do i = 1, N
      if (C(i, i) < ZERO) then
        C(:, i) = -C(:, i)
      end if
    end do

  end subroutine

  subroutine prepare_independent_pairs_active_space(active_space_MO, N_occ, pairs, irrep, freeze, activeactive)
    integer, intent(in) :: active_space_MO(:)
    integer, intent(in) :: N_occ(:)
    type(indep_mo_pair), allocatable, intent(out) :: pairs(:)
    integer, intent(in), optional :: irrep(:)
    integer, intent(in), optional :: freeze(:)
    logical, intent(in), optional :: activeactive
    logical :: yes_activeactive = .False.
    integer ::i, j, n
    if (present(activeactive)) yes_activeactive = activeactive
    n = 0
    do i = 1, size(N_occ, 1)
      do j = i + 1, size(N_occ, 1)
        if (present(freeze)) then
          if (any(freeze .eq. i)) cycle
          if (any(freeze .eq. j)) cycle
        end if
        if (present(irrep)) then
          if (irrep(i) .ne. irrep(j)) cycle
        end if
        if ((.not. ANY(active_space_MO .eq. i) .and. .not. ANY(active_space_MO .eq. j) .and. N_occ(j) .ne. N_occ(i)) &  ! unocc-occ
          .or. (.NOT. ANY(active_space_MO .eq. i) .and. ANY(active_space_MO .eq. j)) &  ! active - non active
          .or. (ANY(active_space_MO .eq. i) .and. .not. ANY(active_space_MO .eq. j)) &  ! non active -active
          .or. (ANY(active_space_MO .eq. i) .and. ANY(active_space_MO .eq. j) .and. yes_activeactive) &  ! active -active
          ) then
          n = n + 1
        end if
      end do
    end do
    allocate (pairs(n))
    n = 0

    do i = 1, size(N_occ, 1)
      do j = i + 1, size(N_occ, 1)
        if (present(freeze)) then
          if (any(freeze .eq. i)) cycle
          if (any(freeze .eq. j)) cycle
        end if
        if (present(irrep)) then
          if (irrep(i) .ne. irrep(j)) cycle
        end if
        if ((.not. ANY(active_space_MO .eq. i) .and. .not. ANY(active_space_MO .eq. j) .and. N_occ(j) .ne. N_occ(i)) &  ! unocc-occ
          .or. (.NOT. ANY(active_space_MO .eq. i) .and. ANY(active_space_MO .eq. j)) &  ! active - non active
          .or. (ANY(active_space_MO .eq. i) .and. .not. ANY(active_space_MO .eq. j)) &  ! non active -active
          .or. (ANY(active_space_MO .eq. i) .and. ANY(active_space_MO .eq. j) .and. yes_activeactive) &  ! active -active
          ) then
          n = n + 1
          pairs(n)%mo2 = i
          pairs(n)%mo1 = j
        end if
      end do
    end do
  end subroutine prepare_independent_pairs_active_space

  function linesearch(energy, gradient, lambda, n) result(opt)
    real(long), intent(in) :: energy(:)
    real(long), intent(in) :: gradient(:)
    real(long), intent(in) :: lambda(:)
    integer, intent(in) :: n
    integer :: i, j
    real(long) :: opt
    real(long) :: H, G

    opt = ZERO
    if (all(gradient(n)*gradient(1:n) > 0)) then
      !write(6,*) 'all same sign: new lambda',lambda(n)+(lambda(n)-lambda(1))
      H = (gradient(n) - gradient(n - 1))/(lambda(n) - lambda(n - 1))
      G = (energy(n) - energy(n - 1))/(lambda(n) - lambda(n - 1))
      if (G*gradient(n) < ZERO .and. .False.) then
        !write(6,*) 'energy rising even if we go along negative gradient! make a very small step'
        opt = lambda(1) + 1./10*abs(lambda(n) - lambda(n - 1))
      else if (abs(G/gradient(n)) - ONE > 1.5) then
        !write(6,*) 'decreasing step',G/gradient(n)
        opt = lambda(1) + abs(gradient(n)/G)*abs(lambda(n) - lambda(n - 1))
      else
        if (H < 0) then
          !write(6,*) 'H<ZERO but energy decreasing; going forward'
          opt = lambda(n) + lambda(n) - lambda(n - 1)
        else if (H > 0) then
          !write(6,*) 'H>ZERO; Newton step'
          opt = lambda(n) - gradient(n)/H
        end if
      end if
    else
      if (gradient(n)*gradient(1) < ZERO) then
        !write(6,*) 'gradient sign switch; bisecting'
        i = n
        j = minloc(abs(lambda(n) - lambda(1:n)), dim=1, mask=gradient(1)*gradient(1:n) > 0)
        opt = abs(1/gradient(i))/(abs(1/gradient(i)) + abs(1/gradient(j)))*lambda(i) + &
          abs(1/gradient(j))/(abs(1/gradient(i)) + abs(1/gradient(j)))*lambda(j)
      else
        !write(6,*) 'gradient sign switch; bisecting'
        i = minloc(abs(gradient(1:n)), dim=1, mask=gradient(n)*gradient(1:n) < 0)
        opt = abs(1/gradient(i))/(abs(1/gradient(i)) + abs(1/gradient(n)))*lambda(i) + &
          abs(1/gradient(n))/(abs(1/gradient(i)) + abs(1/gradient(n)))*lambda(n)

        !opt=0.5*(lambda(n)+lambda(i))
        if (opt < 1e-4) &
          opt = ZERO
      end if
    end if
  end function linesearch

!>------------------------------------------------------------------------------
!! subroutine make_nat_trans_orbitals
!! computes natural orbitals from the provided transition density tdens
!! hole_orbital returns the hole orbital,
!! part_orbital returns the particle orbital expansion coefficients
!! OUTPUT is the unit number for output
  subroutine make_nat_trans_orbitals(OUTPUT, tdens, hole_orbital, part_orbital)
    integer, intent(in) :: OUTPUT
    real(long), intent(in) :: tdens(:, :)
    real(long), intent(out) :: hole_orbital(size(tdens, 1))
    real(long), intent(out) :: part_orbital(size(tdens, 2))
    real(long) :: tdens2O(size(tdens, 1), size(tdens, 1))
    real(long) :: tdens2V(size(tdens, 2), size(tdens, 2))
    real(long) :: tdens2O_ev(size(tdens, 1))
    real(long) :: tdens2V_ev(size(tdens, 2))
    integer :: i

    tdens2O = tdens.mt.tdens
    call get_eigenvalues(tdens2O, tdens2O_ev)
    hole_orbital = tdens2O(:, size(tdens2O, 1))
    write (OUTPUT, '("# ",7x,"Natural orbital hole sqrt lambda=",f10.5)') sqrt(tdens2O_ev(size(tdens2O, 1)))
    do i = 1, size(tdens2O, 1)
      if (abs(hole_orbital(i)) .gt. 0.001) write (OUTPUT, '("# ",7x,i4,x,f10.5)') i, hole_orbital(i)
    end do
    tdens2V = tdens.tm.tdens
    call get_eigenvalues(tdens2V, tdens2V_ev)
    part_orbital = tdens2V(:, size(tdens2V, 1))
    write (OUTPUT, '("# ",7x,"Natural orbital particle sqrt lambda=",f10.5)') sqrt(tdens2V_ev(size(tdens2V, 1)))
    do i = 1, size(tdens2V, 1)
      if (abs(part_orbital(i)) .gt. 0.001) write (OUTPUT, '("# ",7x,i4,x,f10.5)') i, part_orbital(i)
    end do
  end subroutine make_nat_trans_orbitals

#ifdef NEEDSFINDLOC
  integer function findloc(a, val, dim) result(i)
    !! returns index of a given value in an integer array
    !! todo: check version number. in 2018? findloc is there in intel compiler
    !! when does it appear in gfortran?
    integer :: a(:), val
    integer, optional :: dim
    do i = 1, size(a)
      if (a(i) .eq. val) exit
    end do
    if ((i > size(a))) then
      i = 0
      return
    end if
  end function findloc
#endif
end module mcscf
